(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 11.3' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       158,          7]
NotebookDataLength[  34827603,     811519]
NotebookOptionsPosition[  34261438,     799803]
NotebookOutlinePosition[  34270437,     800003]
CellTagsIndexPosition[  34269803,     799984]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["xAct`xCoba`", "Section",
 FontSize->24,
 CellTags->{
  "S5.18.1", "1.1"},ExpressionUUID->"dbbba50a-4c94-44b0-a4d2-1993a97edb6b"],

Cell[TextData[{
 StyleBox["xCoba`", "MR"],
 " is the ",
 StyleBox["xAct`", "MR"],
 " package for tensor component computations. These are computations in which \
tensors are expressed as collections of scalar components (or coefficients) \
in a pre-defined basis of vector fields (or tensor products of them). Those \
components can be numbers or functions of the scalar fields of a given \
coordinate system or chart. It is, therefore, complementary to the abstract \
(aka geometric) approach of ",
 StyleBox["xTensor`", "MR"],
 ", where all computations are basis-free and coordinate-free. In ",
 StyleBox["xCoba`", "MR"],
 " the word \"basis\" is fully equivalent to the word \"frame\"."
}], "Text",
 CellChangeTimes->{{3.728649299105685*^9, 3.728649300206122*^9}, {
  3.728649345471011*^9, 3.7286495134044733`*^9}, {3.728649577666924*^9, 
  3.72864967506115*^9}, {3.728649774569929*^9, 3.728649797799415*^9}, {
  3.728650131879994*^9, 3.728650147623506*^9}, {3.728661019153554*^9, 
  3.728661081654855*^9}, {3.728733891302291*^9, 3.728733892939979*^9}, {
  3.790862853064907*^9, 
  3.7908629434119053`*^9}},ExpressionUUID->"3ea432db-0842-4a82-933f-\
4fc79004767f"],

Cell[TextData[{
 StyleBox["xCoba`", "MR"],
 " allows the user to define bases on one or more vector bundles and to \
handle basis vectors, using basis indices notation. It performs component \
calculations such as expanding a tensor in a specified basis, changing the \
basis of an expression or tracing the contraction of basis dummies. ",
 StyleBox["xCoba`", "MR"],
 " stores and handles component values efficiently, making full use of tensor \
symmetries. The package knows how to express derivatives and brackets of \
basis vectors in terms of Christoffel and torsion tensors and how to assign \
values to the components of a tensor. And many other things."
}], "Text",
 CellChangeTimes->{{3.728649299105685*^9, 3.728649300206122*^9}, {
  3.728649345471011*^9, 3.728649363276018*^9}, {3.7286499841202393`*^9, 
  3.7286499928149977`*^9}},
 CellID->981026496,ExpressionUUID->"0338f2fc-a904-4892-9b19-3ff6702765bf"],

Cell[TextData[{
 StyleBox["xCoba`", "MR"],
 " is fully compatible with ",
 StyleBox["xTensor`", "MR"],
 ". In fact, many computations in ",
 StyleBox["xCoba`", "MR"],
 " are performed using the ",
 StyleBox["xTensor`", "MR"],
 " functions. Note, however, that ",
 StyleBox["xTensor`", "MR"],
 " is  independent of ",
 StyleBox["xCoba`", "MR"],
 "."
}], "Text",
 CellChangeTimes->{{3.728649997245592*^9, 3.728650105457034*^9}, {
  3.728650186500956*^9, 3.728650253201866*^9}, {3.7286502897693043`*^9, 
  3.728650315223659*^9}, {3.728650373588835*^9, 3.728650395851543*^9}, {
  3.72873393607399*^9, 3.72873393692181*^9}},
 CellID->1715514340,ExpressionUUID->"fa9ac34b-3607-4207-92e9-185e063bce35"],

Cell[TextData[{
 "A crucial ingredient is the formalism of ",
 StyleBox["marked indices ",
  FontSlant->"Italic"],
 "and particularly the treatment of frame dependent objects. Perhaps the most \
conspicuous example is the covariant parallel derivative, which extends the \
concept of ordinary derivative as described on Wald's ",
 StyleBox["General Relativity.",
  FontSlant->"Italic"],
 " This reference has been the main inspiration for much of the notation, and \
the user is advised to keep it in mind when dealing with some of the non \
standard ideas employed by this package."
}], "Text",
 CellChangeTimes->{{3.728645682029468*^9, 3.728645684935066*^9}, 
   3.728650323793211*^9, {3.728650405923553*^9, 3.728650407811368*^9}, {
   3.728733946314147*^9, 3.7287339777201633`*^9}},
 CellID->2099645976,ExpressionUUID->"6474d4f7-8d4a-4dea-b3d1-8af5c08e2840"],

Cell["\<\
There are several ways of storing the components of a tensor, and xCoba` \
implements three of them:\
\>", "Text",
 CellChangeTimes->{{3.728650474519134*^9, 3.728650519215756*^9}, {
  3.7286527422542887`*^9, 3.728652753453808*^9}, {3.790863026057303*^9, 
  3.7908630366811*^9}},
 CellID->711920896,ExpressionUUID->"5f424d0a-16ab-4859-b710-5e7bb8907abd"],

Cell["\t1) Explicit linear combinations of Basis tensors.", "Text",
 CellChangeTimes->{{3.728650474519134*^9, 3.728650540756831*^9}, {
  3.728652758421358*^9, 3.728652758797284*^9}},
 CellID->519866281,ExpressionUUID->"87731305-ca71-46f2-ab1e-543f7913297c"],

Cell["\t2) A matrix (or array) based approach, using CTensor[...].", "Text",
 CellChangeTimes->{{3.7286505483317842`*^9, 3.7286505509083014`*^9}, {
  3.728652764661096*^9, 3.72865279018821*^9}, {3.728652825906341*^9, 
  3.7286528355450773`*^9}},
 CellID->1569558782,ExpressionUUID->"8e592d57-865d-451c-9c9e-525017076f79"],

Cell["\t3) Implicit storage of components of an abstract tensor.", "Text",
 CellChangeTimes->{{3.728652795587613*^9, 3.728652817778964*^9}},
 CellID->1902448593,ExpressionUUID->"57b046f0-6234-4e17-b5ad-ed61f3702da2"],

Cell["\<\
Each method has pros and cons, and you'll need to decide which method is most \
convenient for your computation. In general, the CTensor[...] approach is the \
easiest for simple manipulations, like computing all curvature tensors of a \
metric field expressed in a given coordinate system. The implicit storage \
method can be useful when working with computations that still have a \
strongly abstract flavor, or a high degree of symmetry. The explicit linear \
combinations can be useful for typesetting results. xCoba` provides functions \
to work in each of these representations, and also to change from one to \
another. This notebook contains descriptions of the three approaches and \
their relations.\
\>", "Text",
 CellChangeTimes->{{3.7286528488694687`*^9, 3.728653106702353*^9}, {
  3.7287340172566223`*^9, 3.728734083963386*^9}},
 CellID->166422542,ExpressionUUID->"ca3b484e-a3a6-4023-91b1-a60eef8264fe"],

Cell[TextData[{
 "Historically (in the development of ",
 StyleBox["xAct`", "MR"],
 "), method 1) was developed first, then method 3) and finally method 2) \
above. In hindsight, the CTensor[...] method 2) is the most natural and \
simplest of them, and we recommend you to start with it if you are new to ",
 StyleBox["xCoba`", "MR"],
 ". You can right now go to section 3 of this document."
}], "Text",
 CellChangeTimes->{{3.790863109490128*^9, 3.790863258626314*^9}, {
  3.790866420158791*^9, 3.790866466799429*^9}, {3.790867225498068*^9, 
  3.79086723724268*^9}},
 CellID->1209417041,ExpressionUUID->"d824ff7a-9884-45ba-bc77-126663bb22d5"],

Cell[CellGroupData[{

Cell["Load the package", "Subsection",
 CellChangeTimes->{{3.7908686241758003`*^9, 3.790868626368066*^9}, {
  3.790876643724308*^9, 
  3.790876645090822*^9}},ExpressionUUID->"c400feeb-afb2-4fb8-b75a-\
bc95a6dfa74c"],

Cell[CellGroupData[{

Cell[TextData[{
 "This loads the package from the default directory, for example $HOME/.",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 "/AddOns/Applications/xAct/ for a single-user installation under Linux. ",
 StyleBox["xTensor`", "MR"],
 " and ",
 StyleBox["xPerm` ", "MR"],
 "are automatically loaded. See these [",
 ButtonBox["Installation notes",
  BaseStyle->"Hyperlink",
  ButtonData->{
    URL["http://xact.es/download/install"], None},
  ButtonNote->"http://xact.es/download/install"],
 "] for help with installation."
}], "MathCaption",
 CellChangeTimes->{{3.728645731347241*^9, 3.72864573257948*^9}, {
  3.790866494535139*^9, 3.7908665173645697`*^9}, {3.7908665493546343`*^9, 
  3.7908665629689617`*^9}},
 CellTags->{
  "S5.18.1", "1.3"},ExpressionUUID->"905e69e0-0b65-4680-a63d-32011e1e7f2e"],

Cell[CellGroupData[{

Cell["<<xAct`xCoba`", "Input",
 CellTags->"S5.18.1",
 CellLabel->"In[1]:=",ExpressionUUID->"230b9d8c-8eb7-4fe3-9ce6-d08664c83251"],

Cell[CellGroupData[{

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{
  3.728645581071208*^9, 3.728655115243815*^9, 3.728655380154409*^9, 
   3.728660560061676*^9, 3.7286709929670486`*^9, 3.728674862207717*^9, 
   3.72873264668158*^9, 3.728732781528521*^9, 3.728732965353595*^9, 
   3.72873318559013*^9, 3.728733364720545*^9, 3.728742301210548*^9, 
   3.728743280643224*^9, 3.728743379944036*^9, 3.7287469076880608`*^9, 
   3.728749404278715*^9, 3.7287499227757874`*^9, 3.728750072885293*^9, 
   3.728750171031398*^9, 3.728750281089324*^9, 3.728750506187701*^9, 
   3.7287505659627943`*^9, 3.728750714044882*^9, {3.728750972665595*^9, 
   3.728750996561256*^9}, 3.728751603861548*^9, 3.7287593455663757`*^9, 
   3.728813252044345*^9, 3.728827368522168*^9, 3.728835952688724*^9, 
   3.728836285758855*^9, 3.728842985295206*^9, 3.728845607700259*^9, 
   3.733367186582423*^9, 3.790862251030995*^9, 3.790862558159748*^9, 
   3.790866581734527*^9, 3.7908689999729023`*^9, 3.7908712361143723`*^9, 
   3.790874168786983*^9, 3.7908742777079897`*^9, 3.7908750366484833`*^9, 
   3.790875820427225*^9, 3.790876581445723*^9, 3.790876651280979*^9, 
   3.7908767015020237`*^9, 3.791377987312237*^9, 3.7913816674423523`*^9, 
   3.79138963148276*^9, 3.791389715986176*^9, 3.791389779608074*^9, 
   3.79139065953981*^9, 3.79227312741565*^9},
 CellTags->"S5.18.1",
 CellLabel->"During evaluation of In[1]:=",
 CellID->1179299757,ExpressionUUID->"ba9e778d-7b75-4d49-badd-8ace32f2ff65"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`xPerm`  version \"\>", 
   "\[InvisibleSpace]", "\<\"1.2.3\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2015", ",", "8", ",", "23"}], "}"}]}],
  SequenceForm["Package xAct`xPerm`  version ", "1.2.3", ", ", {2015, 8, 23}],
  
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645581071208*^9, 3.728655115243815*^9, 3.728655380154409*^9, 
   3.728660560061676*^9, 3.7286709929670486`*^9, 3.728674862207717*^9, 
   3.72873264668158*^9, 3.728732781528521*^9, 3.728732965353595*^9, 
   3.72873318559013*^9, 3.728733364720545*^9, 3.728742301210548*^9, 
   3.728743280643224*^9, 3.728743379944036*^9, 3.7287469076880608`*^9, 
   3.728749404278715*^9, 3.7287499227757874`*^9, 3.728750072885293*^9, 
   3.728750171031398*^9, 3.728750281089324*^9, 3.728750506187701*^9, 
   3.7287505659627943`*^9, 3.728750714044882*^9, {3.728750972665595*^9, 
   3.728750996561256*^9}, 3.728751603861548*^9, 3.7287593455663757`*^9, 
   3.728813252044345*^9, 3.728827368522168*^9, 3.728835952688724*^9, 
   3.728836285758855*^9, 3.728842985295206*^9, 3.728845607700259*^9, 
   3.733367186582423*^9, 3.790862251030995*^9, 3.790862558159748*^9, 
   3.790866581734527*^9, 3.7908689999729023`*^9, 3.7908712361143723`*^9, 
   3.790874168786983*^9, 3.7908742777079897`*^9, 3.7908750366484833`*^9, 
   3.790875820427225*^9, 3.790876581445723*^9, 3.790876651280979*^9, 
   3.7908767015020237`*^9, 3.791377987312237*^9, 3.7913816674423523`*^9, 
   3.79138963148276*^9, 3.791389715986176*^9, 3.791389779608074*^9, 
   3.79139065953981*^9, 3.792273127427191*^9},
 CellTags->"S5.18.1",
 CellLabel->"During evaluation of In[1]:=",
 CellID->1364437736,ExpressionUUID->"81c1f1fe-1d92-4100-b475-ed767052ad80"],

Cell[BoxData["\<\"CopyRight (C) 2003-2020, Jose M. Martin-Garcia, under the \
General Public License.\"\>"], "Print",
 CellChangeTimes->{
  3.728645581071208*^9, 3.728655115243815*^9, 3.728655380154409*^9, 
   3.728660560061676*^9, 3.7286709929670486`*^9, 3.728674862207717*^9, 
   3.72873264668158*^9, 3.728732781528521*^9, 3.728732965353595*^9, 
   3.72873318559013*^9, 3.728733364720545*^9, 3.728742301210548*^9, 
   3.728743280643224*^9, 3.728743379944036*^9, 3.7287469076880608`*^9, 
   3.728749404278715*^9, 3.7287499227757874`*^9, 3.728750072885293*^9, 
   3.728750171031398*^9, 3.728750281089324*^9, 3.728750506187701*^9, 
   3.7287505659627943`*^9, 3.728750714044882*^9, {3.728750972665595*^9, 
   3.728750996561256*^9}, 3.728751603861548*^9, 3.7287593455663757`*^9, 
   3.728813252044345*^9, 3.728827368522168*^9, 3.728835952688724*^9, 
   3.728836285758855*^9, 3.728842985295206*^9, 3.728845607700259*^9, 
   3.733367186582423*^9, 3.790862251030995*^9, 3.790862558159748*^9, 
   3.790866581734527*^9, 3.7908689999729023`*^9, 3.7908712361143723`*^9, 
   3.790874168786983*^9, 3.7908742777079897`*^9, 3.7908750366484833`*^9, 
   3.790875820427225*^9, 3.790876581445723*^9, 3.790876651280979*^9, 
   3.7908767015020237`*^9, 3.791377987312237*^9, 3.7913816674423523`*^9, 
   3.79138963148276*^9, 3.791389715986176*^9, 3.791389779608074*^9, 
   3.79139065953981*^9, 3.792273127439725*^9},
 CellTags->"S5.18.1",
 CellLabel->"During evaluation of In[1]:=",
 CellID->1274646286,ExpressionUUID->"375f14ff-36e7-4c99-b124-e9243169232e"],

Cell[BoxData["\<\"Connecting to external mac executable...\"\>"], "Print",
 CellChangeTimes->{
  3.728645581071208*^9, 3.728655115243815*^9, 3.728655380154409*^9, 
   3.728660560061676*^9, 3.7286709929670486`*^9, 3.728674862207717*^9, 
   3.72873264668158*^9, 3.728732781528521*^9, 3.728732965353595*^9, 
   3.72873318559013*^9, 3.728733364720545*^9, 3.728742301210548*^9, 
   3.728743280643224*^9, 3.728743379944036*^9, 3.7287469076880608`*^9, 
   3.728749404278715*^9, 3.7287499227757874`*^9, 3.728750072885293*^9, 
   3.728750171031398*^9, 3.728750281089324*^9, 3.728750506187701*^9, 
   3.7287505659627943`*^9, 3.728750714044882*^9, {3.728750972665595*^9, 
   3.728750996561256*^9}, 3.728751603861548*^9, 3.7287593455663757`*^9, 
   3.728813252044345*^9, 3.728827368522168*^9, 3.728835952688724*^9, 
   3.728836285758855*^9, 3.728842985295206*^9, 3.728845607700259*^9, 
   3.733367186582423*^9, 3.790862251030995*^9, 3.790862558159748*^9, 
   3.790866581734527*^9, 3.7908689999729023`*^9, 3.7908712361143723`*^9, 
   3.790874168786983*^9, 3.7908742777079897`*^9, 3.7908750366484833`*^9, 
   3.790875820427225*^9, 3.790876581445723*^9, 3.790876651280979*^9, 
   3.7908767015020237`*^9, 3.791377987312237*^9, 3.7913816674423523`*^9, 
   3.79138963148276*^9, 3.791389715986176*^9, 3.791389779608074*^9, 
   3.79139065953981*^9, 3.792273127450056*^9},
 CellTags->"S5.18.1",
 CellLabel->"During evaluation of In[1]:=",
 CellID->1830749977,ExpressionUUID->"5f20e29b-4673-4992-a466-a9e4352a4854"],

Cell[BoxData["\<\"Connection established.\"\>"], "Print",
 CellChangeTimes->{
  3.728645581071208*^9, 3.728655115243815*^9, 3.728655380154409*^9, 
   3.728660560061676*^9, 3.7286709929670486`*^9, 3.728674862207717*^9, 
   3.72873264668158*^9, 3.728732781528521*^9, 3.728732965353595*^9, 
   3.72873318559013*^9, 3.728733364720545*^9, 3.728742301210548*^9, 
   3.728743280643224*^9, 3.728743379944036*^9, 3.7287469076880608`*^9, 
   3.728749404278715*^9, 3.7287499227757874`*^9, 3.728750072885293*^9, 
   3.728750171031398*^9, 3.728750281089324*^9, 3.728750506187701*^9, 
   3.7287505659627943`*^9, 3.728750714044882*^9, {3.728750972665595*^9, 
   3.728750996561256*^9}, 3.728751603861548*^9, 3.7287593455663757`*^9, 
   3.728813252044345*^9, 3.728827368522168*^9, 3.728835952688724*^9, 
   3.728836285758855*^9, 3.728842985295206*^9, 3.728845607700259*^9, 
   3.733367186582423*^9, 3.790862251030995*^9, 3.790862558159748*^9, 
   3.790866581734527*^9, 3.7908689999729023`*^9, 3.7908712361143723`*^9, 
   3.790874168786983*^9, 3.7908742777079897`*^9, 3.7908750366484833`*^9, 
   3.790875820427225*^9, 3.790876581445723*^9, 3.790876651280979*^9, 
   3.7908767015020237`*^9, 3.791377987312237*^9, 3.7913816674423523`*^9, 
   3.79138963148276*^9, 3.791389715986176*^9, 3.791389779608074*^9, 
   3.79139065953981*^9, 3.792273127482608*^9},
 CellTags->"S5.18.1",
 CellLabel->"During evaluation of In[1]:=",
 CellID->269169223,ExpressionUUID->"29c5c7c7-d6f2-43fa-82bd-1ac7b2c2935e"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{
  3.728645581071208*^9, 3.728655115243815*^9, 3.728655380154409*^9, 
   3.728660560061676*^9, 3.7286709929670486`*^9, 3.728674862207717*^9, 
   3.72873264668158*^9, 3.728732781528521*^9, 3.728732965353595*^9, 
   3.72873318559013*^9, 3.728733364720545*^9, 3.728742301210548*^9, 
   3.728743280643224*^9, 3.728743379944036*^9, 3.7287469076880608`*^9, 
   3.728749404278715*^9, 3.7287499227757874`*^9, 3.728750072885293*^9, 
   3.728750171031398*^9, 3.728750281089324*^9, 3.728750506187701*^9, 
   3.7287505659627943`*^9, 3.728750714044882*^9, {3.728750972665595*^9, 
   3.728750996561256*^9}, 3.728751603861548*^9, 3.7287593455663757`*^9, 
   3.728813252044345*^9, 3.728827368522168*^9, 3.728835952688724*^9, 
   3.728836285758855*^9, 3.728842985295206*^9, 3.728845607700259*^9, 
   3.733367186582423*^9, 3.790862251030995*^9, 3.790862558159748*^9, 
   3.790866581734527*^9, 3.7908689999729023`*^9, 3.7908712361143723`*^9, 
   3.790874168786983*^9, 3.7908742777079897`*^9, 3.7908750366484833`*^9, 
   3.790875820427225*^9, 3.790876581445723*^9, 3.790876651280979*^9, 
   3.7908767015020237`*^9, 3.791377987312237*^9, 3.7913816674423523`*^9, 
   3.79138963148276*^9, 3.791389715986176*^9, 3.791389779608074*^9, 
   3.79139065953981*^9, 3.792273127490374*^9},
 CellTags->"S5.18.1",
 CellLabel->"During evaluation of In[1]:=",
 CellID->1326570379,ExpressionUUID->"a1fdd9a4-466b-4b62-96b9-9d9112a747f6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`xTensor`  version \"\>", 
   "\[InvisibleSpace]", "\<\"1.1.4\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2020", ",", "2", ",", "16"}], "}"}]}],
  SequenceForm[
  "Package xAct`xTensor`  version ", "1.1.4", ", ", {2020, 2, 16}],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645581071208*^9, 3.728655115243815*^9, 3.728655380154409*^9, 
   3.728660560061676*^9, 3.7286709929670486`*^9, 3.728674862207717*^9, 
   3.72873264668158*^9, 3.728732781528521*^9, 3.728732965353595*^9, 
   3.72873318559013*^9, 3.728733364720545*^9, 3.728742301210548*^9, 
   3.728743280643224*^9, 3.728743379944036*^9, 3.7287469076880608`*^9, 
   3.728749404278715*^9, 3.7287499227757874`*^9, 3.728750072885293*^9, 
   3.728750171031398*^9, 3.728750281089324*^9, 3.728750506187701*^9, 
   3.7287505659627943`*^9, 3.728750714044882*^9, {3.728750972665595*^9, 
   3.728750996561256*^9}, 3.728751603861548*^9, 3.7287593455663757`*^9, 
   3.728813252044345*^9, 3.728827368522168*^9, 3.728835952688724*^9, 
   3.728836285758855*^9, 3.728842985295206*^9, 3.728845607700259*^9, 
   3.733367186582423*^9, 3.790862251030995*^9, 3.790862558159748*^9, 
   3.790866581734527*^9, 3.7908689999729023`*^9, 3.7908712361143723`*^9, 
   3.790874168786983*^9, 3.7908742777079897`*^9, 3.7908750366484833`*^9, 
   3.790875820427225*^9, 3.790876581445723*^9, 3.790876651280979*^9, 
   3.7908767015020237`*^9, 3.791377987312237*^9, 3.7913816674423523`*^9, 
   3.79138963148276*^9, 3.791389715986176*^9, 3.791389779608074*^9, 
   3.79139065953981*^9, 3.792273127503167*^9},
 CellTags->"S5.18.1",
 CellLabel->"During evaluation of In[1]:=",
 CellID->493136699,ExpressionUUID->"d6981527-ad0a-4b72-b1de-9d2aaccab228"],

Cell[BoxData["\<\"CopyRight (C) 2002-2020, Jose M. Martin-Garcia, under the \
General Public License.\"\>"], "Print",
 CellChangeTimes->{
  3.728645581071208*^9, 3.728655115243815*^9, 3.728655380154409*^9, 
   3.728660560061676*^9, 3.7286709929670486`*^9, 3.728674862207717*^9, 
   3.72873264668158*^9, 3.728732781528521*^9, 3.728732965353595*^9, 
   3.72873318559013*^9, 3.728733364720545*^9, 3.728742301210548*^9, 
   3.728743280643224*^9, 3.728743379944036*^9, 3.7287469076880608`*^9, 
   3.728749404278715*^9, 3.7287499227757874`*^9, 3.728750072885293*^9, 
   3.728750171031398*^9, 3.728750281089324*^9, 3.728750506187701*^9, 
   3.7287505659627943`*^9, 3.728750714044882*^9, {3.728750972665595*^9, 
   3.728750996561256*^9}, 3.728751603861548*^9, 3.7287593455663757`*^9, 
   3.728813252044345*^9, 3.728827368522168*^9, 3.728835952688724*^9, 
   3.728836285758855*^9, 3.728842985295206*^9, 3.728845607700259*^9, 
   3.733367186582423*^9, 3.790862251030995*^9, 3.790862558159748*^9, 
   3.790866581734527*^9, 3.7908689999729023`*^9, 3.7908712361143723`*^9, 
   3.790874168786983*^9, 3.7908742777079897`*^9, 3.7908750366484833`*^9, 
   3.790875820427225*^9, 3.790876581445723*^9, 3.790876651280979*^9, 
   3.7908767015020237`*^9, 3.791377987312237*^9, 3.7913816674423523`*^9, 
   3.79138963148276*^9, 3.791389715986176*^9, 3.791389779608074*^9, 
   3.79139065953981*^9, 3.792273127511488*^9},
 CellTags->"S5.18.1",
 CellLabel->"During evaluation of In[1]:=",
 CellID->1009106320,ExpressionUUID->"605e2495-5085-4665-913a-70e8c1c4e280"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{
  3.728645581071208*^9, 3.728655115243815*^9, 3.728655380154409*^9, 
   3.728660560061676*^9, 3.7286709929670486`*^9, 3.728674862207717*^9, 
   3.72873264668158*^9, 3.728732781528521*^9, 3.728732965353595*^9, 
   3.72873318559013*^9, 3.728733364720545*^9, 3.728742301210548*^9, 
   3.728743280643224*^9, 3.728743379944036*^9, 3.7287469076880608`*^9, 
   3.728749404278715*^9, 3.7287499227757874`*^9, 3.728750072885293*^9, 
   3.728750171031398*^9, 3.728750281089324*^9, 3.728750506187701*^9, 
   3.7287505659627943`*^9, 3.728750714044882*^9, {3.728750972665595*^9, 
   3.728750996561256*^9}, 3.728751603861548*^9, 3.7287593455663757`*^9, 
   3.728813252044345*^9, 3.728827368522168*^9, 3.728835952688724*^9, 
   3.728836285758855*^9, 3.728842985295206*^9, 3.728845607700259*^9, 
   3.733367186582423*^9, 3.790862251030995*^9, 3.790862558159748*^9, 
   3.790866581734527*^9, 3.7908689999729023`*^9, 3.7908712361143723`*^9, 
   3.790874168786983*^9, 3.7908742777079897`*^9, 3.7908750366484833`*^9, 
   3.790875820427225*^9, 3.790876581445723*^9, 3.790876651280979*^9, 
   3.7908767015020237`*^9, 3.791377987312237*^9, 3.7913816674423523`*^9, 
   3.79138963148276*^9, 3.791389715986176*^9, 3.791389779608074*^9, 
   3.79139065953981*^9, 3.792273127605134*^9},
 CellTags->"S5.18.1",
 CellLabel->"During evaluation of In[1]:=",
 CellID->1371559881,ExpressionUUID->"0cc75c7a-dfed-4d24-a707-c5d025a7c427"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`xCoba`  version \"\>", 
   "\[InvisibleSpace]", "\<\"0.8.5\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2020", ",", "2", ",", "16"}], "}"}]}],
  SequenceForm["Package xAct`xCoba`  version ", "0.8.5", ", ", {2020, 2, 16}],
  
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645581071208*^9, 3.728655115243815*^9, 3.728655380154409*^9, 
   3.728660560061676*^9, 3.7286709929670486`*^9, 3.728674862207717*^9, 
   3.72873264668158*^9, 3.728732781528521*^9, 3.728732965353595*^9, 
   3.72873318559013*^9, 3.728733364720545*^9, 3.728742301210548*^9, 
   3.728743280643224*^9, 3.728743379944036*^9, 3.7287469076880608`*^9, 
   3.728749404278715*^9, 3.7287499227757874`*^9, 3.728750072885293*^9, 
   3.728750171031398*^9, 3.728750281089324*^9, 3.728750506187701*^9, 
   3.7287505659627943`*^9, 3.728750714044882*^9, {3.728750972665595*^9, 
   3.728750996561256*^9}, 3.728751603861548*^9, 3.7287593455663757`*^9, 
   3.728813252044345*^9, 3.728827368522168*^9, 3.728835952688724*^9, 
   3.728836285758855*^9, 3.728842985295206*^9, 3.728845607700259*^9, 
   3.733367186582423*^9, 3.790862251030995*^9, 3.790862558159748*^9, 
   3.790866581734527*^9, 3.7908689999729023`*^9, 3.7908712361143723`*^9, 
   3.790874168786983*^9, 3.7908742777079897`*^9, 3.7908750366484833`*^9, 
   3.790875820427225*^9, 3.790876581445723*^9, 3.790876651280979*^9, 
   3.7908767015020237`*^9, 3.791377987312237*^9, 3.7913816674423523`*^9, 
   3.79138963148276*^9, 3.791389715986176*^9, 3.791389779608074*^9, 
   3.79139065953981*^9, 3.7922731276138773`*^9},
 CellTags->"S5.18.1",
 CellLabel->"During evaluation of In[1]:=",
 CellID->717453069,ExpressionUUID->"6f2083df-3a6f-45b1-993a-3ba0cc722b11"],

Cell[BoxData["\<\"CopyRight (C) 2005-2020, David Yllanes and Jose M. \
Martin-Garcia, under the General Public License.\"\>"], "Print",
 CellChangeTimes->{
  3.728645581071208*^9, 3.728655115243815*^9, 3.728655380154409*^9, 
   3.728660560061676*^9, 3.7286709929670486`*^9, 3.728674862207717*^9, 
   3.72873264668158*^9, 3.728732781528521*^9, 3.728732965353595*^9, 
   3.72873318559013*^9, 3.728733364720545*^9, 3.728742301210548*^9, 
   3.728743280643224*^9, 3.728743379944036*^9, 3.7287469076880608`*^9, 
   3.728749404278715*^9, 3.7287499227757874`*^9, 3.728750072885293*^9, 
   3.728750171031398*^9, 3.728750281089324*^9, 3.728750506187701*^9, 
   3.7287505659627943`*^9, 3.728750714044882*^9, {3.728750972665595*^9, 
   3.728750996561256*^9}, 3.728751603861548*^9, 3.7287593455663757`*^9, 
   3.728813252044345*^9, 3.728827368522168*^9, 3.728835952688724*^9, 
   3.728836285758855*^9, 3.728842985295206*^9, 3.728845607700259*^9, 
   3.733367186582423*^9, 3.790862251030995*^9, 3.790862558159748*^9, 
   3.790866581734527*^9, 3.7908689999729023`*^9, 3.7908712361143723`*^9, 
   3.790874168786983*^9, 3.7908742777079897`*^9, 3.7908750366484833`*^9, 
   3.790875820427225*^9, 3.790876581445723*^9, 3.790876651280979*^9, 
   3.7908767015020237`*^9, 3.791377987312237*^9, 3.7913816674423523`*^9, 
   3.79138963148276*^9, 3.791389715986176*^9, 3.791389779608074*^9, 
   3.79139065953981*^9, 3.7922731276215487`*^9},
 CellTags->"S5.18.1",
 CellLabel->"During evaluation of In[1]:=",
 CellID->113052778,ExpressionUUID->"12c7c00d-d500-4221-8f3f-9b085dda45e2"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{
  3.728645581071208*^9, 3.728655115243815*^9, 3.728655380154409*^9, 
   3.728660560061676*^9, 3.7286709929670486`*^9, 3.728674862207717*^9, 
   3.72873264668158*^9, 3.728732781528521*^9, 3.728732965353595*^9, 
   3.72873318559013*^9, 3.728733364720545*^9, 3.728742301210548*^9, 
   3.728743280643224*^9, 3.728743379944036*^9, 3.7287469076880608`*^9, 
   3.728749404278715*^9, 3.7287499227757874`*^9, 3.728750072885293*^9, 
   3.728750171031398*^9, 3.728750281089324*^9, 3.728750506187701*^9, 
   3.7287505659627943`*^9, 3.728750714044882*^9, {3.728750972665595*^9, 
   3.728750996561256*^9}, 3.728751603861548*^9, 3.7287593455663757`*^9, 
   3.728813252044345*^9, 3.728827368522168*^9, 3.728835952688724*^9, 
   3.728836285758855*^9, 3.728842985295206*^9, 3.728845607700259*^9, 
   3.733367186582423*^9, 3.790862251030995*^9, 3.790862558159748*^9, 
   3.790866581734527*^9, 3.7908689999729023`*^9, 3.7908712361143723`*^9, 
   3.790874168786983*^9, 3.7908742777079897`*^9, 3.7908750366484833`*^9, 
   3.790875820427225*^9, 3.790876581445723*^9, 3.790876651280979*^9, 
   3.7908767015020237`*^9, 3.791377987312237*^9, 3.7913816674423523`*^9, 
   3.79138963148276*^9, 3.791389715986176*^9, 3.791389779608074*^9, 
   3.79139065953981*^9, 3.792273127630357*^9},
 CellTags->"S5.18.1",
 CellLabel->"During evaluation of In[1]:=",
 CellID->1702495898,ExpressionUUID->"4ac41860-f22b-4a85-8cf3-1e92c262da90"],

Cell[BoxData["\<\"These packages come with ABSOLUTELY NO WARRANTY; for \
details type Disclaimer[]. This is free software, and you are welcome to \
redistribute it under certain conditions. See the General Public License for \
details.\"\>"], "Print",
 CellChangeTimes->{
  3.728645581071208*^9, 3.728655115243815*^9, 3.728655380154409*^9, 
   3.728660560061676*^9, 3.7286709929670486`*^9, 3.728674862207717*^9, 
   3.72873264668158*^9, 3.728732781528521*^9, 3.728732965353595*^9, 
   3.72873318559013*^9, 3.728733364720545*^9, 3.728742301210548*^9, 
   3.728743280643224*^9, 3.728743379944036*^9, 3.7287469076880608`*^9, 
   3.728749404278715*^9, 3.7287499227757874`*^9, 3.728750072885293*^9, 
   3.728750171031398*^9, 3.728750281089324*^9, 3.728750506187701*^9, 
   3.7287505659627943`*^9, 3.728750714044882*^9, {3.728750972665595*^9, 
   3.728750996561256*^9}, 3.728751603861548*^9, 3.7287593455663757`*^9, 
   3.728813252044345*^9, 3.728827368522168*^9, 3.728835952688724*^9, 
   3.728836285758855*^9, 3.728842985295206*^9, 3.728845607700259*^9, 
   3.733367186582423*^9, 3.790862251030995*^9, 3.790862558159748*^9, 
   3.790866581734527*^9, 3.7908689999729023`*^9, 3.7908712361143723`*^9, 
   3.790874168786983*^9, 3.7908742777079897`*^9, 3.7908750366484833`*^9, 
   3.790875820427225*^9, 3.790876581445723*^9, 3.790876651280979*^9, 
   3.7908767015020237`*^9, 3.791377987312237*^9, 3.7913816674423523`*^9, 
   3.79138963148276*^9, 3.791389715986176*^9, 3.791389779608074*^9, 
   3.79139065953981*^9, 3.7922731276385193`*^9},
 CellTags->"S5.18.1",
 CellLabel->"During evaluation of In[1]:=",
 CellID->776445058,ExpressionUUID->"2dcd86c4-fb1e-4819-a823-74f6e956c7b2"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{
  3.728645581071208*^9, 3.728655115243815*^9, 3.728655380154409*^9, 
   3.728660560061676*^9, 3.7286709929670486`*^9, 3.728674862207717*^9, 
   3.72873264668158*^9, 3.728732781528521*^9, 3.728732965353595*^9, 
   3.72873318559013*^9, 3.728733364720545*^9, 3.728742301210548*^9, 
   3.728743280643224*^9, 3.728743379944036*^9, 3.7287469076880608`*^9, 
   3.728749404278715*^9, 3.7287499227757874`*^9, 3.728750072885293*^9, 
   3.728750171031398*^9, 3.728750281089324*^9, 3.728750506187701*^9, 
   3.7287505659627943`*^9, 3.728750714044882*^9, {3.728750972665595*^9, 
   3.728750996561256*^9}, 3.728751603861548*^9, 3.7287593455663757`*^9, 
   3.728813252044345*^9, 3.728827368522168*^9, 3.728835952688724*^9, 
   3.728836285758855*^9, 3.728842985295206*^9, 3.728845607700259*^9, 
   3.733367186582423*^9, 3.790862251030995*^9, 3.790862558159748*^9, 
   3.790866581734527*^9, 3.7908689999729023`*^9, 3.7908712361143723`*^9, 
   3.790874168786983*^9, 3.7908742777079897`*^9, 3.7908750366484833`*^9, 
   3.790875820427225*^9, 3.790876581445723*^9, 3.790876651280979*^9, 
   3.7908767015020237`*^9, 3.791377987312237*^9, 3.7913816674423523`*^9, 
   3.79138963148276*^9, 3.791389715986176*^9, 3.791389779608074*^9, 
   3.79139065953981*^9, 3.792273127646978*^9},
 CellTags->"S5.18.1",
 CellLabel->"During evaluation of In[1]:=",
 CellID->1601450634,ExpressionUUID->"76c3b7cc-ab4f-4624-a2ae-cf7a524c3985"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Hide dollar-indices, as usual (see ",
 StyleBox["xTensor`", "MR"],
 " documentation for more details on this):"
}], "MathCaption",
 CellChangeTimes->{{3.728732722010008*^9, 3.728732765425928*^9}},
 CellID->899063648,ExpressionUUID->"07b7acc4-9757-4c71-ab8a-974855d85ede"],

Cell[BoxData[
 RowBox[{
  RowBox[{"$PrePrint", "=", "ScreenDollarIndices"}], ";"}]], "Input",
 CellChangeTimes->{{3.7287327697928743`*^9, 3.72873277018363*^9}},
 CellLabel->"In[2]:=",
 CellID->867791644,ExpressionUUID->"542ec53a-331d-4856-a7d7-e70cc5cb75bc"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["0. xCoba` essentials", "Subsection",
 CellChangeTimes->{{3.728656510283742*^9, 3.728656511217067*^9}, {
  3.7286598570138693`*^9, 3.728659868855068*^9}, {3.728660932877496*^9, 
  3.7286609335000553`*^9}, {3.790868628359874*^9, 3.790868629072007*^9}, {
  3.790876656416953*^9, 
  3.790876657937037*^9}},ExpressionUUID->"1d1a3bbf-f661-4f3d-8fe4-\
621aa6bd670b"],

Cell[TextData[{
 "A session with ",
 StyleBox["xCoba`", "MR"],
 " begins just like one with ",
 StyleBox["xTensor`", "MR"],
 ", defining one or more manifolds and vector bundles and the objects living \
on them."
}], "Text",
 CellChangeTimes->{
  3.72864627574824*^9, {3.728655278473693*^9, 
   3.72865527924363*^9}},ExpressionUUID->"f0fe6d4f-06bd-416b-9c68-\
4dbe13a5980f"],

Cell[CellGroupData[{

Cell[TextData[{
 "Define a 3d manifold ",
 StyleBox["M3", "MR"],
 ", with indices ",
 StyleBox["a...f",
  FontFamily->"Courier",
  FontWeight->"Regular"],
 ":"
}], "MathCaption",
 CellChangeTimes->{{3.790866613985384*^9, 
  3.79086662220093*^9}},ExpressionUUID->"d73d0932-aee1-47f0-97a5-\
1d9b24e8bb84"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M3", ",", " ", "3", ",", " ", 
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f"}], "}"}]}], 
  "]"}]], "Input",
 CellLabel->"In[3]:=",ExpressionUUID->"2227dbb2-4f25-4016-a854-ecb257ec0cc6"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefManifold", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"manifold \"\>", "\[InvisibleSpace]", "M3", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefManifold, ": Defining ", 
   "manifold ", $CellContext`M3, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728746913041642*^9, 3.728749407468087*^9, 3.7287499235469503`*^9, 
   3.728750073635009*^9, 3.728750174189872*^9, 3.728750281855173*^9, 
   3.7287505072177753`*^9, 3.728750570020459*^9, 3.728750715090887*^9, {
   3.7287509734858227`*^9, 3.728751000013081*^9}, 3.728751604536048*^9, 
   3.7288132529344807`*^9, 3.7288273693552027`*^9, 3.7288429914226313`*^9, 
   3.728845608707136*^9, 3.733367198261632*^9, 3.790862572598646*^9, 
   3.790866605593536*^9, 3.792273141701139*^9},
 CellLabel->"During evaluation of In[3]:=",
 CellID->293029645,ExpressionUUID->"90952141-66c3-40b6-9e8b-4a485d2388ca"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefVBundle", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle \"\>", "\[InvisibleSpace]", "TangentM3", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefVBundle, ": Defining ", 
   "vbundle ", $CellContext`TangentM3, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728746913041642*^9, 3.728749407468087*^9, 3.7287499235469503`*^9, 
   3.728750073635009*^9, 3.728750174189872*^9, 3.728750281855173*^9, 
   3.7287505072177753`*^9, 3.728750570020459*^9, 3.728750715090887*^9, {
   3.7287509734858227`*^9, 3.728751000013081*^9}, 3.728751604536048*^9, 
   3.7288132529344807`*^9, 3.7288273693552027`*^9, 3.7288429914226313`*^9, 
   3.728845608707136*^9, 3.733367198261632*^9, 3.790862572598646*^9, 
   3.790866605593536*^9, 3.7922731417097*^9},
 CellLabel->"During evaluation of In[3]:=",
 CellID->769240944,ExpressionUUID->"ddb35440-7399-42e7-b857-c7fce7c011ee"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Define a complex vector bundle (the warnings are purely informative and \
harmless):\
\>", "MathCaption",
 CellChangeTimes->{
  3.7286552824253073`*^9, {3.728732672908708*^9, 3.728732681972042*^9}, {
   3.728732994022616*^9, 
   3.7287330054606743`*^9}},ExpressionUUID->"2f4e63a1-9698-4648-be5f-\
74b3a9d261cf"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefVBundle", "[", 
  RowBox[{"InnerC", ",", "M3", ",", "2", ",", 
   RowBox[{"{", 
    RowBox[{"A", ",", "B", ",", "C", ",", "D", ",", "F", ",", "G"}], "}"}], 
   ",", 
   RowBox[{"Dagger", "\[Rule]", "Complex"}]}], "]"}]], "Input",
 CellLabel->"In[4]:=",ExpressionUUID->"1c61b0aa-a0ba-45a1-b437-0e5a59a551ed"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefVBundle", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle \"\>", "\[InvisibleSpace]", "InnerC", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefVBundle, ": Defining ", 
   "vbundle ", $CellContext`InnerC, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728732789216392*^9, 3.728733007375355*^9, 3.728733189752511*^9, 
   3.7287333656471233`*^9, 3.728742302205514*^9, 3.7287432820370502`*^9, 
   3.728743381304302*^9, 3.728746914762179*^9, 3.7287494080059566`*^9, 
   3.7287499237170897`*^9, 3.7287500737875147`*^9, 3.728750174415072*^9, 
   3.728750282026334*^9, 3.728750507387027*^9, 3.72875057027335*^9, 
   3.728750715260644*^9, {3.7287509736369467`*^9, 3.72875100021126*^9}, 
   3.72875160470079*^9, 3.7288132531123457`*^9, 3.728827369570734*^9, 
   3.728842992769582*^9, 3.728845608939727*^9, 3.733367199611528*^9, 
   3.790862574415738*^9, 3.790866684316745*^9, 3.792273143027075*^9},
 CellLabel->"During evaluation of In[4]:=",
 CellID->1034147936,ExpressionUUID->"9fe35c8a-fa01-4731-b10d-4e8e8ceaf3c2"],

Cell[BoxData[
 TemplateBox[{
  "ValidateSymbol", "capital", 
   "\"System name \\!\\(\\*RowBox[{\\\"\\\\\\\"C\\\\\\\"\\\"}]\\) is \
overloaded as an abstract index.\"", 2, 4, 1, 19668387891272167742, "Local", 
   "xAct`xCore`ValidateSymbol"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{
  3.728645867902925*^9, 3.728655158876857*^9, 3.728655384910871*^9, 
   3.728660570057692*^9, 3.7286710165454817`*^9, 3.728674863441874*^9, 
   3.728732667790407*^9, 3.7287327892828617`*^9, 3.728733007445066*^9, 
   3.72873318985763*^9, 3.728733365719778*^9, 3.728742302337237*^9, 
   3.728743282128933*^9, 3.728743381368064*^9, 3.728746914857394*^9, 
   3.728749408072154*^9, 3.728749923760768*^9, 3.728750073879168*^9, 
   3.7287501744942293`*^9, 3.728750282076119*^9, 3.7287505074364223`*^9, 
   3.7287505703431187`*^9, 3.728750715309421*^9, {3.728750973685526*^9, 
   3.7287510002788773`*^9}, 3.728751604786333*^9, 3.728813253206449*^9, 
   3.7288273696884193`*^9, 3.72884299288723*^9, 3.728845609088861*^9, 
   3.7333671996694517`*^9, 3.7908625744404907`*^9, 3.790866684333572*^9, 
   3.79227314305369*^9},
 CellLabel->"During evaluation of In[4]:=",
 CellID->1339205181,ExpressionUUID->"7695401f-4718-42c0-b04f-8e56c8635877"],

Cell[BoxData[
 TemplateBox[{
  "ValidateSymbol", "capital", 
   "\"System name \\!\\(\\*RowBox[{\\\"\\\\\\\"D\\\\\\\"\\\"}]\\) is \
overloaded as an abstract index.\"", 2, 4, 2, 19668387891272167742, "Local", 
   "xAct`xCore`ValidateSymbol"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{
  3.728645867902925*^9, 3.728655158876857*^9, 3.728655384910871*^9, 
   3.728660570057692*^9, 3.7286710165454817`*^9, 3.728674863441874*^9, 
   3.728732667790407*^9, 3.7287327892828617`*^9, 3.728733007445066*^9, 
   3.72873318985763*^9, 3.728733365719778*^9, 3.728742302337237*^9, 
   3.728743282128933*^9, 3.728743381368064*^9, 3.728746914857394*^9, 
   3.728749408072154*^9, 3.728749923760768*^9, 3.728750073879168*^9, 
   3.7287501744942293`*^9, 3.728750282076119*^9, 3.7287505074364223`*^9, 
   3.7287505703431187`*^9, 3.728750715309421*^9, {3.728750973685526*^9, 
   3.7287510002788773`*^9}, 3.728751604786333*^9, 3.728813253206449*^9, 
   3.7288273696884193`*^9, 3.72884299288723*^9, 3.728845609088861*^9, 
   3.7333671996694517`*^9, 3.7908625744404907`*^9, 3.790866684333572*^9, 
   3.7922731430656977`*^9},
 CellLabel->"During evaluation of In[4]:=",
 CellID->245372860,ExpressionUUID->"e3d52473-9aff-4678-a8ae-d3522aa85672"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefVBundle", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"conjugated vbundle \"\>", "\[InvisibleSpace]", 
   "InnerC\[Dagger]", "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Assuming fixed anti-isomorphism between InnerC \
and InnerC\[Dagger]\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefVBundle, ": Defining ", 
   "conjugated vbundle ", $CellContext`InnerC\[Dagger], ". ", 
   "Assuming fixed anti-isomorphism between InnerC and InnerC\[Dagger]"],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728732789216392*^9, 3.728733007375355*^9, 3.728733189752511*^9, 
   3.7287333656471233`*^9, 3.728742302205514*^9, 3.7287432820370502`*^9, 
   3.728743381304302*^9, 3.728746914762179*^9, 3.7287494080059566`*^9, 
   3.7287499237170897`*^9, 3.7287500737875147`*^9, 3.728750174415072*^9, 
   3.728750282026334*^9, 3.728750507387027*^9, 3.72875057027335*^9, 
   3.728750715260644*^9, {3.7287509736369467`*^9, 3.72875100021126*^9}, 
   3.72875160470079*^9, 3.7288132531123457`*^9, 3.728827369570734*^9, 
   3.728842992769582*^9, 3.728845608939727*^9, 3.733367199611528*^9, 
   3.790862574415738*^9, 3.790866684316745*^9, 3.792273143077354*^9},
 CellLabel->"During evaluation of In[4]:=",
 CellID->813055780,ExpressionUUID->"726ea625-8b95-43ae-9d74-519d49f6d3d6"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Define a contravariant vector ",
 StyleBox["v", "MR"],
 " and a covariant symmetric tensor ",
 StyleBox["T", "MR"],
 ":"
}], "MathCaption",ExpressionUUID->"144e099c-20ee-46c4-b579-e9eda0162447"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"v", "[", "a", "]"}], ",", "M3"}], "]"}], "\[IndentingNewLine]", 
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", "M3", ",", 
   RowBox[{"Antisymmetric", "[", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}]}], "}"}], "]"}]}], "]"}]}], "Input",
 CellLabel->"In[5]:=",ExpressionUUID->"fa47a144-19e3-4f11-8851-55e3f7cfe830"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"v", "[", "a", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`v[$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.7287327908309793`*^9, 3.7287330150798683`*^9, 3.72873319130202*^9, 
   3.728733365935651*^9, 3.728742302604347*^9, 3.728743284012957*^9, 
   3.7287433815971527`*^9, 3.728746916401211*^9, 3.728749409729851*^9, 
   3.728749923918819*^9, 3.728750074081847*^9, 3.7287501748165493`*^9, 
   3.728750282448073*^9, 3.7287505076363583`*^9, 3.728750570506785*^9, 
   3.728750715507*^9, {3.728750973919486*^9, 3.728751000439788*^9}, 
   3.728751604992632*^9, 3.7288132534313*^9, 3.728827369938023*^9, 
   3.728842994170582*^9, 3.728845611657659*^9, 3.733367201295232*^9, 
   3.790862580791296*^9, 3.790866702489691*^9, 3.792273144126815*^9},
 CellLabel->"During evaluation of In[5]:=",
 CellID->211920697,ExpressionUUID->"73315d18-a979-4e26-a785-4c75b58f02e3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"T", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`T[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.7287327908309793`*^9, 3.7287330150798683`*^9, 3.72873319130202*^9, 
   3.728733365935651*^9, 3.728742302604347*^9, 3.728743284012957*^9, 
   3.7287433815971527`*^9, 3.728746916401211*^9, 3.728749409729851*^9, 
   3.728749923918819*^9, 3.728750074081847*^9, 3.7287501748165493`*^9, 
   3.728750282448073*^9, 3.7287505076363583`*^9, 3.728750570506785*^9, 
   3.728750715507*^9, {3.728750973919486*^9, 3.728751000439788*^9}, 
   3.728751604992632*^9, 3.7288132534313*^9, 3.728827369938023*^9, 
   3.728842994170582*^9, 3.728845611657659*^9, 3.733367201295232*^9, 
   3.790862580791296*^9, 3.790866702489691*^9, 3.7922731441398487`*^9},
 CellLabel->"During evaluation of In[5]:=",
 CellID->1223267753,ExpressionUUID->"10f2167a-ccaa-4478-86bd-151a6a2d36fd"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Define a Riemann symmetric tensor and a covariant vector on ",
 StyleBox["InnerC:", "MR"]
}], "MathCaption",
 CellChangeTimes->{
  3.7286552866653337`*^9},ExpressionUUID->"9e87e110-73d4-42a6-a9b0-\
9820a7c58a00"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"R", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "B"}], ",", 
     RowBox[{"-", "C"}], ",", 
     RowBox[{"-", "D"}]}], "]"}], ",", "M3", ",", 
   RowBox[{"RiemannSymmetric", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "2", ",", "3", ",", "4"}], "}"}], "]"}], ",", 
   RowBox[{"Dagger", "\[Rule]", "Complex"}]}], "]"}]], "Input",
 CellLabel->"In[7]:=",ExpressionUUID->"1aa374b7-92b8-4bb2-b014-cf39b8e0f0d2"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"R", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "B"}], ",", 
     RowBox[{"-", "C"}], ",", 
     RowBox[{"-", "D"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`R[-$CellContext`A, -$CellContext`B, -C, -D], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728732792164723*^9, 3.72873301786403*^9, 3.728733192700284*^9, 
   3.7287333661043043`*^9, 3.728742302834033*^9, 3.7287432841887074`*^9, 
   3.7287433834456463`*^9, 3.728746917914863*^9, 3.7287494108604794`*^9, 
   3.728749924068158*^9, 3.728750074236538*^9, 3.728750175184826*^9, 
   3.728750284228787*^9, 3.7287505099060297`*^9, 3.728750570672619*^9, 
   3.7287507173700647`*^9, {3.728750974090043*^9, 3.728751000623879*^9}, 
   3.728751605168685*^9, 3.7288132536292467`*^9, 3.728827370153162*^9, 
   3.728842995434134*^9, 3.7288456118472*^9, 3.733367203663684*^9, 
   3.7908625820867233`*^9, 3.79086670524874*^9, 3.792273145507202*^9},
 CellLabel->"During evaluation of In[7]:=",
 CellID->2123278630,ExpressionUUID->"f17adc8a-1554-437a-b866-8a0058efb006"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"R\[Dagger]", "[", 
    RowBox[{
     RowBox[{"-", "A\[Dagger]"}], ",", 
     RowBox[{"-", "B\[Dagger]"}], ",", 
     RowBox[{"-", "C\[Dagger]"}], ",", 
     RowBox[{"-", "D\[Dagger]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`R\[Dagger][-$CellContext`A\[Dagger], \
-$CellContext`B\[Dagger], -$CellContext`C\[Dagger], -$CellContext`D\[Dagger]],
    ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728732792164723*^9, 3.72873301786403*^9, 3.728733192700284*^9, 
   3.7287333661043043`*^9, 3.728742302834033*^9, 3.7287432841887074`*^9, 
   3.7287433834456463`*^9, 3.728746917914863*^9, 3.7287494108604794`*^9, 
   3.728749924068158*^9, 3.728750074236538*^9, 3.728750175184826*^9, 
   3.728750284228787*^9, 3.7287505099060297`*^9, 3.728750570672619*^9, 
   3.7287507173700647`*^9, {3.728750974090043*^9, 3.728751000623879*^9}, 
   3.728751605168685*^9, 3.7288132536292467`*^9, 3.728827370153162*^9, 
   3.728842995434134*^9, 3.7288456118472*^9, 3.733367203663684*^9, 
   3.7908625820867233`*^9, 3.79086670524874*^9, 3.792273145516654*^9},
 CellLabel->"During evaluation of In[7]:=",
 CellID->1208172211,ExpressionUUID->"430e5b8f-b19e-4f46-bc80-bddc73bf21bd"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"u", "[", 
    RowBox[{"-", "A"}], "]"}], ",", "M3", ",", 
   RowBox[{"Dagger", "\[Rule]", "Complex"}]}], "]"}]], "Input",
 CellLabel->"In[8]:=",ExpressionUUID->"09246897-571f-4402-a8c3-88dac122abee"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"u", "[", 
    RowBox[{"-", "A"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`u[-$CellContext`A], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728732793031885*^9, 3.7287330193964357`*^9, 3.7287331934176197`*^9, 
   3.728733367841752*^9, 3.728742304764203*^9, 3.728743284364119*^9, 
   3.728743383592803*^9, 3.728746918733553*^9, 3.728749411556912*^9, 
   3.728749924201953*^9, 3.728750075996698*^9, 3.7287501753276587`*^9, 
   3.728750284611602*^9, 3.7287505103051558`*^9, 3.728750570813385*^9, 
   3.728750717507382*^9, {3.728750975804295*^9, 3.7287510007569*^9}, 
   3.728751605335957*^9, 3.728813255344324*^9, 3.728827371940276*^9, 
   3.72884299626847*^9, 3.728845612040966*^9, 3.733367204313841*^9, 
   3.7908625832073307`*^9, 3.790866722977416*^9, 3.792273146039297*^9},
 CellLabel->"During evaluation of In[8]:=",
 CellID->370703793,ExpressionUUID->"9cb4191c-52e6-4546-aa2d-de81d64e02a0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"u\[Dagger]", "[", 
    RowBox[{"-", "A\[Dagger]"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`u\[Dagger][-$CellContext`A\[Dagger]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728732793031885*^9, 3.7287330193964357`*^9, 3.7287331934176197`*^9, 
   3.728733367841752*^9, 3.728742304764203*^9, 3.728743284364119*^9, 
   3.728743383592803*^9, 3.728746918733553*^9, 3.728749411556912*^9, 
   3.728749924201953*^9, 3.728750075996698*^9, 3.7287501753276587`*^9, 
   3.728750284611602*^9, 3.7287505103051558`*^9, 3.728750570813385*^9, 
   3.728750717507382*^9, {3.728750975804295*^9, 3.7287510007569*^9}, 
   3.728751605335957*^9, 3.728813255344324*^9, 3.728827371940276*^9, 
   3.72884299626847*^9, 3.728845612040966*^9, 3.733367204313841*^9, 
   3.7908625832073307`*^9, 3.790866722977416*^9, 3.7922731460509443`*^9},
 CellLabel->"During evaluation of In[8]:=",
 CellID->568589004,ExpressionUUID->"f5d1b321-f009-4aeb-a074-92b6b7447389"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Remember that in ",
 StyleBox["xTensor`", "MR"],
 " we use abstract indices:"
}], "MathCaption",
 CellChangeTimes->{
  3.728655291680807*^9, {3.7908667412210817`*^9, 
   3.790866743469027*^9}},ExpressionUUID->"8df2b989-60fa-483e-92c7-\
b9c8d36c5750"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"T", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}]}], "]"}], 
  RowBox[{"v", "[", "b", "]"}]}]], "Input",
 CellLabel->"In[9]:=",ExpressionUUID->"3e510aa4-9ca5-4a69-b048-4e760800cdd1"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {" ", " "},
          {"a", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[-$CellContext`a, -$CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"b"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{
  3.7286458767146587`*^9, 3.7286551885318947`*^9, 3.728655390286623*^9, 
   3.728660571033349*^9, 3.728671028487309*^9, 3.728674865970592*^9, 
   3.7287326974532633`*^9, 3.7287327944319696`*^9, 3.728733022430253*^9, 
   3.728733195485608*^9, 3.728733368093307*^9, 3.7287423049633427`*^9, 
   3.7287432845473337`*^9, 3.7287433837479687`*^9, 3.728746921440851*^9, 
   3.728749412796199*^9, 3.7287499243361073`*^9, 3.728750076212714*^9, 
   3.728750175528617*^9, 3.728750284795129*^9, 3.7287505104627028`*^9, 
   3.72875057095914*^9, 3.7287507176556673`*^9, {3.728750976001012*^9, 
   3.728751000892148*^9}, 3.728751607089539*^9, 3.728813255544886*^9, 
   3.728827372169869*^9, 3.728842997551614*^9, 3.728845613230689*^9, 
   3.733367205486898*^9, 3.7908625846351757`*^9, 3.790866730650879*^9, 
   3.792273147399817*^9},
 CellLabel->"Out[9]=",
 CellID->1751099190,ExpressionUUID->"e0637ed9-f69f-441c-b979-9f0b0300d0da"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "A basis (or frame) is a set of independent vector fields on a vbundle, with \
an associated dual basis of vector fields. We only declare the name of the \
basis and the associated cnumbers (\"component numbers\"), integers that will \
identify the individual vectors/covectors. ",
 StyleBox["xCoba`", "MR"],
 " then constructs a number of geometric objects associated to the basis. We \
will review them in detail further down."
}], "Text",
 CellChangeTimes->{{3.728655215908573*^9, 3.7286552396036386`*^9}, {
  3.7286570835519457`*^9, 3.728657160267198*^9}, {3.7287330407074423`*^9, 
  3.728733115239875*^9}},ExpressionUUID->"f5748fd6-e0f0-4e93-ad14-\
30ba9184170e"],

Cell[CellGroupData[{

Cell[TextData[{
 "Now ",
 StyleBox["xCoba`", "MR"],
 "  lets us define a basis, called ",
 StyleBox["fred", "MR"],
 ". We have to specify its vector bundle and cnumbers. Several related \
objects are also defined (parallel derivative and associated tensors and \
densities):"
}], "MathCaption",
 CellChangeTimes->{{3.728655296528775*^9, 3.728655300136568*^9}, {
   3.728733121247657*^9, 3.728733164717702*^9}, {3.728733202092903*^9, 
   3.728733204091763*^9}, {3.7287469313471613`*^9, 3.728746932106833*^9}, 
   3.7908667908275957`*^9},ExpressionUUID->"e67021f0-f9cd-4b75-947a-\
e00601bd1833"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefBasis", "[", 
  RowBox[{"fred", ",", "TangentM3", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2"}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728733172278639*^9, 3.728733172669983*^9}, {
  3.7287469345143843`*^9, 3.728746935106546*^9}},
 CellLabel->"In[10]:=",ExpressionUUID->"087432d1-54fc-4625-bf20-fe69ae86adb8"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefBasis", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"basis \"\>", "\[InvisibleSpace]", "fred", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xCoba`DefBasis, ": Defining ", "basis ", $CellContext`fred, 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.7286458789989557`*^9, 3.728655348131331*^9, 3.728655396054265*^9, 
   3.728660571390258*^9, 3.728671031709518*^9, 3.728674866121977*^9, 
   3.728732798434576*^9, 3.728733169099123*^9, 3.728733233489567*^9, 
   3.7287333682539043`*^9, 3.728742305097023*^9, 3.728743284688232*^9, 
   3.7287433838776407`*^9, 3.728746936980541*^9, 3.728749414926165*^9, 
   3.728749926212223*^9, 3.728750076520802*^9, 3.728750175670993*^9, 
   3.728750284975143*^9, 3.728750510583269*^9, 3.728750571074135*^9, 
   3.7287507177754803`*^9, {3.728750976137418*^9, 3.7287510010086803`*^9}, 
   3.728751607274138*^9, 3.728813255662562*^9, 3.7288273723153343`*^9, 
   3.7288429998448687`*^9, 3.728845613410329*^9, 3.7333672104145803`*^9, 
   3.790862589001452*^9, 3.7908668047211323`*^9, 3.792273149394065*^9},
 CellLabel->"During evaluation of In[10]:=",
 CellID->1009788355,ExpressionUUID->"07145505-31b4-4972-8458-6a60642c1427"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"parallel derivative \"\>", "\[InvisibleSpace]", 
   RowBox[{"PDfred", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "parallel derivative ", 
   $CellContext`PDfred[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.7286458789989557`*^9, 3.728655348131331*^9, 3.728655396054265*^9, 
   3.728660571390258*^9, 3.728671031709518*^9, 3.728674866121977*^9, 
   3.728732798434576*^9, 3.728733169099123*^9, 3.728733233489567*^9, 
   3.7287333682539043`*^9, 3.728742305097023*^9, 3.728743284688232*^9, 
   3.7287433838776407`*^9, 3.728746936980541*^9, 3.728749414926165*^9, 
   3.728749926212223*^9, 3.728750076520802*^9, 3.728750175670993*^9, 
   3.728750284975143*^9, 3.728750510583269*^9, 3.728750571074135*^9, 
   3.7287507177754803`*^9, {3.728750976137418*^9, 3.7287510010086803`*^9}, 
   3.728751607274138*^9, 3.728813255662562*^9, 3.7288273723153343`*^9, 
   3.7288429998448687`*^9, 3.728845613410329*^9, 3.7333672104145803`*^9, 
   3.790862589001452*^9, 3.7908668047211323`*^9, 3.792273149407629*^9},
 CellLabel->"During evaluation of In[10]:=",
 CellID->2097103891,ExpressionUUID->"a1676383-3aae-43cd-8644-a9771f721ad4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"TorsionPDfred", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "torsion tensor ", 
   $CellContext`TorsionPDfred[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.7286458789989557`*^9, 3.728655348131331*^9, 3.728655396054265*^9, 
   3.728660571390258*^9, 3.728671031709518*^9, 3.728674866121977*^9, 
   3.728732798434576*^9, 3.728733169099123*^9, 3.728733233489567*^9, 
   3.7287333682539043`*^9, 3.728742305097023*^9, 3.728743284688232*^9, 
   3.7287433838776407`*^9, 3.728746936980541*^9, 3.728749414926165*^9, 
   3.728749926212223*^9, 3.728750076520802*^9, 3.728750175670993*^9, 
   3.728750284975143*^9, 3.728750510583269*^9, 3.728750571074135*^9, 
   3.7287507177754803`*^9, {3.728750976137418*^9, 3.7287510010086803`*^9}, 
   3.728751607274138*^9, 3.728813255662562*^9, 3.7288273723153343`*^9, 
   3.7288429998448687`*^9, 3.728845613410329*^9, 3.7333672104145803`*^9, 
   3.790862589001452*^9, 3.7908668047211323`*^9, 3.7922731494221487`*^9},
 CellLabel->"During evaluation of In[10]:=",
 CellID->1523481473,ExpressionUUID->"d2d6785c-0e39-4a2b-bc5b-b3eb57923967"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelPDfred", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "non-symmetric Christoffel tensor ", 
   $CellContext`ChristoffelPDfred[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.7286458789989557`*^9, 3.728655348131331*^9, 3.728655396054265*^9, 
   3.728660571390258*^9, 3.728671031709518*^9, 3.728674866121977*^9, 
   3.728732798434576*^9, 3.728733169099123*^9, 3.728733233489567*^9, 
   3.7287333682539043`*^9, 3.728742305097023*^9, 3.728743284688232*^9, 
   3.7287433838776407`*^9, 3.728746936980541*^9, 3.728749414926165*^9, 
   3.728749926212223*^9, 3.728750076520802*^9, 3.728750175670993*^9, 
   3.728750284975143*^9, 3.728750510583269*^9, 3.728750571074135*^9, 
   3.7287507177754803`*^9, {3.728750976137418*^9, 3.7287510010086803`*^9}, 
   3.728751607274138*^9, 3.728813255662562*^9, 3.7288273723153343`*^9, 
   3.7288429998448687`*^9, 3.728845613410329*^9, 3.7333672104145803`*^9, 
   3.790862589001452*^9, 3.7908668047211323`*^9, 3.792273149430114*^9},
 CellLabel->"During evaluation of In[10]:=",
 CellID->705593530,ExpressionUUID->"8838fbb9-2a5c-4beb-95b7-1ec14e0063be"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Riemann tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RiemannPDfred", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Riemann tensor ", 
   $CellContext`RiemannPDfred[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, $CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.7286458789989557`*^9, 3.728655348131331*^9, 3.728655396054265*^9, 
   3.728660571390258*^9, 3.728671031709518*^9, 3.728674866121977*^9, 
   3.728732798434576*^9, 3.728733169099123*^9, 3.728733233489567*^9, 
   3.7287333682539043`*^9, 3.728742305097023*^9, 3.728743284688232*^9, 
   3.7287433838776407`*^9, 3.728746936980541*^9, 3.728749414926165*^9, 
   3.728749926212223*^9, 3.728750076520802*^9, 3.728750175670993*^9, 
   3.728750284975143*^9, 3.728750510583269*^9, 3.728750571074135*^9, 
   3.7287507177754803`*^9, {3.728750976137418*^9, 3.7287510010086803`*^9}, 
   3.728751607274138*^9, 3.728813255662562*^9, 3.7288273723153343`*^9, 
   3.7288429998448687`*^9, 3.728845613410329*^9, 3.7333672104145803`*^9, 
   3.790862589001452*^9, 3.7908668047211323`*^9, 3.792273149437521*^9},
 CellLabel->"During evaluation of In[10]:=",
 CellID->505759976,ExpressionUUID->"1b918961-8ade-4cf5-82cd-352e6e27a8bf"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciPDfred", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Ricci tensor ", 
   $CellContext`RicciPDfred[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.7286458789989557`*^9, 3.728655348131331*^9, 3.728655396054265*^9, 
   3.728660571390258*^9, 3.728671031709518*^9, 3.728674866121977*^9, 
   3.728732798434576*^9, 3.728733169099123*^9, 3.728733233489567*^9, 
   3.7287333682539043`*^9, 3.728742305097023*^9, 3.728743284688232*^9, 
   3.7287433838776407`*^9, 3.728746936980541*^9, 3.728749414926165*^9, 
   3.728749926212223*^9, 3.728750076520802*^9, 3.728750175670993*^9, 
   3.728750284975143*^9, 3.728750510583269*^9, 3.728750571074135*^9, 
   3.7287507177754803`*^9, {3.728750976137418*^9, 3.7287510010086803`*^9}, 
   3.728751607274138*^9, 3.728813255662562*^9, 3.7288273723153343`*^9, 
   3.7288429998448687`*^9, 3.728845613410329*^9, 3.7333672104145803`*^9, 
   3.790862589001452*^9, 3.7908668047211323`*^9, 3.792273149444989*^9},
 CellLabel->"During evaluation of In[10]:=",
 CellID->567641806,ExpressionUUID->"1769bf0c-6ad3-486d-9254-70352e1c8bb2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaUpfred", "[", 
    RowBox[{"a", ",", "b", ",", "c"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric +1 density ", 
   $CellContext`etaUpfred[$CellContext`a, $CellContext`b, $CellContext`c], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.7286458789989557`*^9, 3.728655348131331*^9, 3.728655396054265*^9, 
   3.728660571390258*^9, 3.728671031709518*^9, 3.728674866121977*^9, 
   3.728732798434576*^9, 3.728733169099123*^9, 3.728733233489567*^9, 
   3.7287333682539043`*^9, 3.728742305097023*^9, 3.728743284688232*^9, 
   3.7287433838776407`*^9, 3.728746936980541*^9, 3.728749414926165*^9, 
   3.728749926212223*^9, 3.728750076520802*^9, 3.728750175670993*^9, 
   3.728750284975143*^9, 3.728750510583269*^9, 3.728750571074135*^9, 
   3.7287507177754803`*^9, {3.728750976137418*^9, 3.7287510010086803`*^9}, 
   3.728751607274138*^9, 3.728813255662562*^9, 3.7288273723153343`*^9, 
   3.7288429998448687`*^9, 3.728845613410329*^9, 3.7333672104145803`*^9, 
   3.790862589001452*^9, 3.7908668047211323`*^9, 3.792273149452388*^9},
 CellLabel->"During evaluation of In[10]:=",
 CellID->1233473468,ExpressionUUID->"af12c497-c6c3-4f87-9f79-d34ab399bb97"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaDownfred", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric -1 density ", 
   $CellContext`etaDownfred[-$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.7286458789989557`*^9, 3.728655348131331*^9, 3.728655396054265*^9, 
   3.728660571390258*^9, 3.728671031709518*^9, 3.728674866121977*^9, 
   3.728732798434576*^9, 3.728733169099123*^9, 3.728733233489567*^9, 
   3.7287333682539043`*^9, 3.728742305097023*^9, 3.728743284688232*^9, 
   3.7287433838776407`*^9, 3.728746936980541*^9, 3.728749414926165*^9, 
   3.728749926212223*^9, 3.728750076520802*^9, 3.728750175670993*^9, 
   3.728750284975143*^9, 3.728750510583269*^9, 3.728750571074135*^9, 
   3.7287507177754803`*^9, {3.728750976137418*^9, 3.7287510010086803`*^9}, 
   3.728751607274138*^9, 3.728813255662562*^9, 3.7288273723153343`*^9, 
   3.7288429998448687`*^9, 3.728845613410329*^9, 3.7333672104145803`*^9, 
   3.790862589001452*^9, 3.7908668047211323`*^9, 3.792273149460125*^9},
 CellLabel->"During evaluation of In[10]:=",
 CellID->1328932185,ExpressionUUID->"6baa3bcf-ed20-447d-b484-08e577adc695"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
These are the basis vector fields. We use covariant C-indices (\"component \
indices\") marked with their basis, with form {cnumber, -basis}:\
\>", "MathCaption",
 CellChangeTimes->{{3.7286554213388577`*^9, 3.7286554302267714`*^9}, {
  3.728656587316477*^9, 3.7286566273477287`*^9}, {3.728656659105844*^9, 
  3.728656670849522*^9}, {3.790866865777133*^9, 3.790866876200891*^9}},
 CellID->1847620262,ExpressionUUID->"bfbb7878-6d87-45eb-9b1f-8134d0c9e7df"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"basis", "=", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"Basis", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "fred"}]}], "}"}], ",", "a"}], "]"}], ",", 
    RowBox[{"Basis", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "fred"}]}], "}"}], ",", "a"}], "]"}], ",", 
    RowBox[{"Basis", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"2", ",", 
        RowBox[{"-", "fred"}]}], "}"}], ",", "a"}], "]"}]}], "}"}]}]], "Input",\

 CellChangeTimes->{{3.728655433054002*^9, 3.728655473937039*^9}, {
  3.728655514407632*^9, 3.728655516206547*^9}, {3.728733242650382*^9, 
  3.7287332602723618`*^9}, {3.728746947243972*^9, 3.728746952129867*^9}},
 CellLabel->"In[11]:=",
 CellID->1127027992,ExpressionUUID->"184ddbd8-4220-455a-bebc-cb681598f84f"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`fred}, $CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`fred}, $CellContext`a],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{{3.728655462000415*^9, 3.728655474457041*^9}, 
   3.728655516531454*^9, 3.728660571825045*^9, 3.7286710330549717`*^9, 
   3.7286748665619583`*^9, 3.728732799816126*^9, 3.728733261926977*^9, 
   3.728733368588394*^9, 3.728742305394815*^9, 3.728743285013411*^9, 
   3.728743384176815*^9, 3.7287469527783337`*^9, 3.7287494163473053`*^9, 
   3.728749926494989*^9, 3.7287500769651737`*^9, 3.728750175896924*^9, 
   3.7287502853269243`*^9, 3.728750511073656*^9, 3.728750571339703*^9, 
   3.728750717993685*^9, {3.728750976350565*^9, 3.728751001246626*^9}, 
   3.728751607577166*^9, 3.728813255942058*^9, 3.728827372638554*^9, 
   3.728843001219489*^9, 3.72884561372703*^9, 3.7333672119989758`*^9, 
   3.790862590535096*^9, 3.7908668271189413`*^9, 3.79227315095926*^9},
 CellLabel->"Out[11]=",
 CellID->1870663923,ExpressionUUID->"867453c7-31ee-48dc-8797-b58f5cef4fe1"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"BasisArray", "[", "fred", "]"}], "[", "a", "]"}]], "Input",
 CellChangeTimes->{{3.728733265265188*^9, 3.728733269140704*^9}, {
  3.7287469546834707`*^9, 3.7287469551694403`*^9}},
 CellLabel->"In[12]:=",
 CellID->1216174007,ExpressionUUID->"cde85017-b1a2-4b16-af78-9a532fb54ba8"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`fred}, $CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`fred}, $CellContext`a],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{
  3.728645911058243*^9, 3.728659447223671*^9, 3.728660571999341*^9, 
   3.728671033670998*^9, 3.728674866709922*^9, 3.728732800612605*^9, 
   3.728733271009583*^9, 3.7287333687759123`*^9, 3.728742305525177*^9, 
   3.728743285162301*^9, 3.728743384329052*^9, 3.7287469556495457`*^9, 
   3.728749417012884*^9, 3.728749926617692*^9, 3.7287500771045313`*^9, 
   3.728750176031145*^9, 3.728750285508523*^9, 3.728750511331139*^9, 
   3.728750571478397*^9, 3.728750718103443*^9, {3.728750976527994*^9, 
   3.728751001385909*^9}, 3.728751607692135*^9, 3.728813256047917*^9, 
   3.7288273727741003`*^9, 3.728843001934779*^9, 3.7288456139106503`*^9, 
   3.733367212698457*^9, 3.7908625913323717`*^9, 3.79086682930262*^9, 
   3.792273151710997*^9},
 CellLabel->"Out[12]=",
 CellID->823175240,ExpressionUUID->"773b7ba5-ac1b-4482-8aa3-a01368a66620"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
And these are the basis covector fields. We use contravariant C-indices of \
the form {cnumber, basis}:\
\>", "MathCaption",
 CellChangeTimes->{{3.728655484409545*^9, 3.7286554903999767`*^9}, {
  3.72865555160491*^9, 3.728655553004799*^9}, {3.7286566170037727`*^9, 
  3.728656666801887*^9}},
 CellID->1297931365,ExpressionUUID->"6c7c536b-1f29-4165-95e7-d244a9731e07"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"cobasis", "=", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"Basis", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", "fred"}], "}"}]}], "]"}], ",", 
    RowBox[{"Basis", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", "fred"}], "}"}]}], "]"}], ",", 
    RowBox[{"Basis", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"{", 
       RowBox[{"2", ",", "fred"}], "}"}]}], "]"}]}], "}"}]}]], "Input",
 CellChangeTimes->{{3.7286554961356153`*^9, 3.728655543709876*^9}, {
   3.728733275641111*^9, 3.728733284360139*^9}, 3.7287333470137243`*^9, {
   3.728746959955346*^9, 3.728746964441264*^9}},
 CellLabel->"In[13]:=",
 CellID->659237437,ExpressionUUID->"e5a12cc0-15d1-49e0-a791-2092a633badc"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`fred}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`fred}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`fred}],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{{3.728655523231439*^9, 3.7286555442025003`*^9}, 
   3.72866057246726*^9, 3.728671034688511*^9, 3.728674866859187*^9, 
   3.728732801898532*^9, 3.728733287196742*^9, {3.7287333478856897`*^9, 
   3.72873336924189*^9}, 3.728742305638713*^9, 3.7287432853015833`*^9, 
   3.7287433844761047`*^9, 3.728746965111393*^9, 3.728749418096479*^9, 
   3.728749926724791*^9, 3.728750077235261*^9, 3.728750176182753*^9, 
   3.728750285948215*^9, 3.728750511659299*^9, 3.728750571608736*^9, 
   3.728750718226863*^9, {3.728750976723653*^9, 3.7287510015101547`*^9}, 
   3.7287516078344927`*^9, 3.728813256177154*^9, 3.72882737292388*^9, 
   3.7288430030518723`*^9, 3.7288456140718527`*^9, 3.7333672137929697`*^9, 
   3.790862595054039*^9, 3.790866923596253*^9, 3.7922731547611647`*^9},
 CellLabel->"Out[13]=",
 CellID->467570674,ExpressionUUID->"9b553423-f706-4774-919b-b3b99757f0e5"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"BasisArray", "[", "fred", "]"}], "[", 
  RowBox[{"-", "a"}], "]"}]], "Input",
 CellChangeTimes->{{3.728659454681059*^9, 3.728659493663294*^9}, {
  3.72873329165604*^9, 3.728733295218252*^9}, {3.728746967450945*^9, 
  3.728746968000918*^9}},
 CellLabel->"In[14]:=",
 CellID->1565089844,ExpressionUUID->"08b1c38d-b280-4ba5-9542-498233de0309"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`fred}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`fred}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`fred}],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{
  3.728645911058243*^9, {3.728659455765608*^9, 3.728659494614534*^9}, 
   3.7286605726554947`*^9, 3.7286710352686777`*^9, 3.728674867012429*^9, 
   3.7287328031389713`*^9, 3.728733296230047*^9, {3.728733348636352*^9, 
   3.728733369572157*^9}, 3.7287423057719584`*^9, 3.7287432854466*^9, 
   3.728743384626629*^9, 3.72874696848363*^9, 3.7287494187800407`*^9, 
   3.7287499268513412`*^9, 3.728750077368993*^9, 3.7287501763580847`*^9, 
   3.728750286131349*^9, 3.728750511777478*^9, 3.728750571762067*^9, 
   3.728750718350115*^9, {3.7287509768307056`*^9, 3.7287510016252117`*^9}, 
   3.7287516079667788`*^9, 3.7288132562848253`*^9, 3.7288273730496197`*^9, 
   3.728843003731093*^9, 3.728845614233707*^9, 3.733367214431109*^9, 
   3.790862595767466*^9, 3.790866927050166*^9, 3.79227315541717*^9},
 CellLabel->"Out[14]=",
 CellID->605431428,ExpressionUUID->"e113f92b-1c0b-45c5-abd3-69bddb4ade84"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The so-called ",
 StyleBox["parallel derivative",
  FontSlant->"Italic"],
 " of the ",
 StyleBox["fred", "MR"],
 " basis (denoted ",
 StyleBox["PDfred", "MR"],
 ") is a covariant derivative operator with the fundamental property of \
giving zero on all elements of the basis and its cobasis:"
}], "MathCaption",
 CellChangeTimes->{{3.728655584668852*^9, 3.728655607274728*^9}, {
   3.7286558176967297`*^9, 3.728655858167245*^9}, {3.728655988865336*^9, 
   3.728655988961316*^9}, {3.728656051438716*^9, 3.7286560766764812`*^9}, 
   3.728656941245068*^9, {3.728733310608329*^9, 3.728733315366745*^9}, {
   3.7908669597263927`*^9, 3.790866965062149*^9}},
 CellID->1988507208,ExpressionUUID->"30041a36-35a2-422b-b446-4df5385dfd6c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PDfred", "[", 
   RowBox[{"-", "b"}], "]"}], "/@", "basis"}]], "Input",
 CellChangeTimes->{{3.728655791316927*^9, 3.7286557990634413`*^9}, {
  3.728733324726336*^9, 3.7287333271524982`*^9}, {3.728746971626563*^9, 
  3.728746973264761*^9}},
 CellLabel->"In[15]:=",
 CellID->1795804312,ExpressionUUID->"080a21f4-b11e-4035-8174-656b65cd33df"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0", ",", "0", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{
  3.7286557996652613`*^9, 3.728660572842967*^9, 3.7286710377564917`*^9, 
   3.728674867167346*^9, 3.7287328068672523`*^9, {3.728733327850851*^9, 
   3.7287333698887377`*^9}, 3.728742305905595*^9, 3.728743285581959*^9, 
   3.728743384775867*^9, 3.72874697381513*^9, 3.728749420505151*^9, 
   3.728749926996276*^9, 3.728750077498836*^9, 3.72875017649741*^9, 
   3.728750286469523*^9, 3.728750512158341*^9, 3.728750571963519*^9, 
   3.728750718450172*^9, {3.7287509769087143`*^9, 3.728751001780058*^9}, 
   3.728751608097495*^9, 3.728813256405746*^9, 3.728827373172605*^9, 
   3.7288430051687403`*^9, 3.728845614383861*^9, 3.733367215732592*^9, 
   3.7908625970352182`*^9, 3.7908669938739557`*^9, 3.7922731571405573`*^9},
 CellLabel->"Out[15]=",
 CellID->892606450,ExpressionUUID->"7b18a9c0-66e5-43fe-9f12-0f3afd5acc02"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PDfred", "[", 
   RowBox[{"-", "b"}], "]"}], "/@", "cobasis"}]], "Input",
 CellChangeTimes->{{3.728655802055221*^9, 3.728655808977315*^9}, {
  3.728733330526189*^9, 3.7287333329763403`*^9}, {3.728746975722322*^9, 
  3.7287469767845716`*^9}},
 CellLabel->"In[16]:=",
 CellID->1945905506,ExpressionUUID->"e38a628e-0602-41d9-af35-a7b5321a68a3"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0", ",", "0", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{
  3.728655809472327*^9, 3.7286605731302767`*^9, 3.7286710383589993`*^9, 
   3.728674867317415*^9, 3.728732808033497*^9, {3.728733333818017*^9, 
   3.72873337017198*^9}, 3.728742306022086*^9, 3.7287432857184258`*^9, 
   3.728743384896708*^9, 3.728746977324727*^9, 3.728749421239315*^9, 
   3.728749927108137*^9, 3.728750077632869*^9, 3.728750176660335*^9, 
   3.728750286797181*^9, 3.728750512468728*^9, 3.728750572109757*^9, 
   3.728750718573182*^9, {3.7287509770030193`*^9, 3.728751001898838*^9}, 
   3.728751608229992*^9, 3.728813256512205*^9, 3.7288273733169317`*^9, 
   3.7288430057926407`*^9, 3.728845614532275*^9, 3.733367216363984*^9, 
   3.7908625976776953`*^9, 3.7908669957894573`*^9, 3.792273157652977*^9},
 CellLabel->"Out[16]=",
 CellID->993820116,ExpressionUUID->"0bb558b1-33dd-4e0d-8955-07f90454dab9"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "As any other covariant derivative operator, it has curvature and torsion. \
Parallel derivatives always have zero curvature, but they have torsion if \
they are associated to non-coordinate (or anholonomic) bases. In ",
 StyleBox["xCoba`", "MR"],
 " the defined bases are non-coordinate by default:"
}], "MathCaption",
 CellChangeTimes->{{3.7286568087702217`*^9, 3.728656895535421*^9}, {
  3.728657179641828*^9, 3.7286572275036793`*^9}},
 CellID->1341226455,ExpressionUUID->"e3936a74-132d-44fc-b228-0fe43db46f04"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Riemann", "[", "PDfred", "]"}], "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}], ",", "d"}], "]"}]], "Input",
 CellChangeTimes->{{3.7286568998667603`*^9, 3.72865690711069*^9}, {
  3.7287333747251883`*^9, 3.728733376815939*^9}, {3.728746980072507*^9, 
  3.728746981272419*^9}},
 CellLabel->"In[17]:=",
 CellID->1937961720,ExpressionUUID->"23d79355-2836-42ce-a4e2-308af96ae7ff"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{
  3.7286569075810328`*^9, 3.728660573290009*^9, 3.728671039577153*^9, 
   3.728674867487274*^9, 3.72873280973309*^9, 3.7287333777732267`*^9, 
   3.728742306159916*^9, 3.728743285866784*^9, 3.728743385028717*^9, 
   3.728746982732587*^9, 3.728749422563686*^9, 3.728749927290986*^9, 
   3.7287500777681417`*^9, 3.728750176770134*^9, 3.728750286960092*^9, 
   3.7287505126199617`*^9, 3.7287505722797747`*^9, 3.72875071869069*^9, {
   3.72875097712186*^9, 3.7287510020002937`*^9}, 3.72875160836451*^9, 
   3.728813256636408*^9, 3.7288273734419928`*^9, 3.728843007109387*^9, 
   3.7288456147547092`*^9, 3.7333672183820753`*^9, 3.7908625989397917`*^9, 
   3.790867021722987*^9, 3.7922731591901493`*^9},
 CellLabel->"Out[17]=",
 CellID->294662927,ExpressionUUID->"a2c4d7de-cf2f-4c7c-8e2a-c17c918d58e7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Torsion", "[", "PDfred", "]"}], "[", 
  RowBox[{"a", ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.7286569087143793`*^9, 3.728656915486318*^9}, {
  3.728733379757504*^9, 3.728733381295504*^9}, {3.728746984442017*^9, 
  3.7287469857200203`*^9}},
 CellLabel->"In[18]:=",
 CellID->1385272999,ExpressionUUID->"9a072d86-7c51-46a6-a794-ea1ce43b63f9"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox["T",
       FontColor->RGBColor[1., 0., 0.]], 
      StyleBox[GridBox[{
         {"a", " ", " "},
         {" ", "b", "c"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`TorsionPDfred[$CellContext`a, -$CellContext`b, -$CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.728656915895322*^9, 3.728660573440927*^9, 3.728671040155109*^9, 
   3.728674867638186*^9, 3.728732810666657*^9, 3.728733382156805*^9, 
   3.728742306292121*^9, 3.728743286000317*^9, 3.7287433851643877`*^9, 
   3.728746986130619*^9, 3.7287494231916647`*^9, 3.728749927394438*^9, 
   3.7287500779197693`*^9, 3.72875017689419*^9, 3.728750287418477*^9, 
   3.728750512770618*^9, 3.728750572426141*^9, 3.7287507188415127`*^9, {
   3.728750977236865*^9, 3.728751002100812*^9}, 3.728751608500063*^9, 
   3.728813256755705*^9, 3.72882737358352*^9, 3.728843007818997*^9, 
   3.728845615178458*^9, 3.733367219048399*^9, 3.790862599586281*^9, 
   3.7908670227245493`*^9, 3.792273159813241*^9},
 CellLabel->"Out[18]=",
 CellID->1257509919,ExpressionUUID->"acc7e043-d07c-4b5b-8f6f-ea6ceaccec25"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Each basis has a color (red by default), used to identify its associated \
objects in StandardForm. Define another basis, colored blue:\
\>", "MathCaption",
 CellChangeTimes->{{3.7286553028803997`*^9, 3.728655358878181*^9}, {
  3.728656088189817*^9, 
  3.728656102211955*^9}},ExpressionUUID->"9291b8b6-e93e-46b0-889f-\
10bfad8faacc"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefBasis", "[", 
  RowBox[{"blues", ",", "TangentM3", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2"}], "}"}], ",", 
   RowBox[{"BasisColor", "\[Rule]", "Blue"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728655362000798*^9, 3.72865536249697*^9}, {
   3.7287333933173723`*^9, 3.728733407178443*^9}, {3.728747009402038*^9, 
   3.728747010455083*^9}, 3.728749224921836*^9},
 CellLabel->"In[19]:=",ExpressionUUID->"01f5cb31-2e45-4c6f-a741-9ffabb0fc8e2"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefBasis", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"basis \"\>", "\[InvisibleSpace]", "blues", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xCoba`DefBasis, ": Defining ", "basis ", $CellContext`blues, 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.7286458808338213`*^9, 3.728655352350588*^9, 3.728656105652667*^9, 
   3.728660573642881*^9, 3.728671045107004*^9, 3.728674867807619*^9, 
   3.728732812164307*^9, 3.728733409396785*^9, 3.728742306408434*^9, 
   3.728743286153721*^9, 3.7287433853131237`*^9, 3.728747012484117*^9, 
   3.728749424327572*^9, 3.728749927526626*^9, 3.728750078069125*^9, 
   3.728750177076168*^9, 3.728750287577373*^9, 3.728750512885186*^9, 
   3.728750572558249*^9, 3.728750718969363*^9, {3.72875097735047*^9, 
   3.728751002205101*^9}, 3.728751608630973*^9, 3.728813256877846*^9, 
   3.72882737373587*^9, 3.728843009027207*^9, 3.728845615342039*^9, 
   3.733367220212503*^9, 3.790862603802002*^9, 3.790867032499701*^9, 
   3.7922731615125628`*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->175165393,ExpressionUUID->"38372027-d060-4742-b26b-76a3b75a5413"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"parallel derivative \"\>", "\[InvisibleSpace]", 
   RowBox[{"PDblues", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "parallel derivative ", 
   $CellContext`PDblues[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.7286458808338213`*^9, 3.728655352350588*^9, 3.728656105652667*^9, 
   3.728660573642881*^9, 3.728671045107004*^9, 3.728674867807619*^9, 
   3.728732812164307*^9, 3.728733409396785*^9, 3.728742306408434*^9, 
   3.728743286153721*^9, 3.7287433853131237`*^9, 3.728747012484117*^9, 
   3.728749424327572*^9, 3.728749927526626*^9, 3.728750078069125*^9, 
   3.728750177076168*^9, 3.728750287577373*^9, 3.728750512885186*^9, 
   3.728750572558249*^9, 3.728750718969363*^9, {3.72875097735047*^9, 
   3.728751002205101*^9}, 3.728751608630973*^9, 3.728813256877846*^9, 
   3.72882737373587*^9, 3.728843009027207*^9, 3.728845615342039*^9, 
   3.733367220212503*^9, 3.790862603802002*^9, 3.790867032499701*^9, 
   3.792273161525831*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1952163761,ExpressionUUID->"75914c48-15b9-42b6-8572-8aac4d274b9c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"TorsionPDblues", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "torsion tensor ", 
   $CellContext`TorsionPDblues[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.7286458808338213`*^9, 3.728655352350588*^9, 3.728656105652667*^9, 
   3.728660573642881*^9, 3.728671045107004*^9, 3.728674867807619*^9, 
   3.728732812164307*^9, 3.728733409396785*^9, 3.728742306408434*^9, 
   3.728743286153721*^9, 3.7287433853131237`*^9, 3.728747012484117*^9, 
   3.728749424327572*^9, 3.728749927526626*^9, 3.728750078069125*^9, 
   3.728750177076168*^9, 3.728750287577373*^9, 3.728750512885186*^9, 
   3.728750572558249*^9, 3.728750718969363*^9, {3.72875097735047*^9, 
   3.728751002205101*^9}, 3.728751608630973*^9, 3.728813256877846*^9, 
   3.72882737373587*^9, 3.728843009027207*^9, 3.728845615342039*^9, 
   3.733367220212503*^9, 3.790862603802002*^9, 3.790867032499701*^9, 
   3.792273161540303*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->761172261,ExpressionUUID->"f8cff853-766e-4995-9938-2f77693c6975"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelPDblues", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "non-symmetric Christoffel tensor ", 
   $CellContext`ChristoffelPDblues[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.7286458808338213`*^9, 3.728655352350588*^9, 3.728656105652667*^9, 
   3.728660573642881*^9, 3.728671045107004*^9, 3.728674867807619*^9, 
   3.728732812164307*^9, 3.728733409396785*^9, 3.728742306408434*^9, 
   3.728743286153721*^9, 3.7287433853131237`*^9, 3.728747012484117*^9, 
   3.728749424327572*^9, 3.728749927526626*^9, 3.728750078069125*^9, 
   3.728750177076168*^9, 3.728750287577373*^9, 3.728750512885186*^9, 
   3.728750572558249*^9, 3.728750718969363*^9, {3.72875097735047*^9, 
   3.728751002205101*^9}, 3.728751608630973*^9, 3.728813256877846*^9, 
   3.72882737373587*^9, 3.728843009027207*^9, 3.728845615342039*^9, 
   3.733367220212503*^9, 3.790862603802002*^9, 3.790867032499701*^9, 
   3.792273161547595*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->462105448,ExpressionUUID->"b6999010-c6fe-4aa8-859a-375782b264df"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Riemann tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RiemannPDblues", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Riemann tensor ", 
   $CellContext`RiemannPDblues[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, $CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.7286458808338213`*^9, 3.728655352350588*^9, 3.728656105652667*^9, 
   3.728660573642881*^9, 3.728671045107004*^9, 3.728674867807619*^9, 
   3.728732812164307*^9, 3.728733409396785*^9, 3.728742306408434*^9, 
   3.728743286153721*^9, 3.7287433853131237`*^9, 3.728747012484117*^9, 
   3.728749424327572*^9, 3.728749927526626*^9, 3.728750078069125*^9, 
   3.728750177076168*^9, 3.728750287577373*^9, 3.728750512885186*^9, 
   3.728750572558249*^9, 3.728750718969363*^9, {3.72875097735047*^9, 
   3.728751002205101*^9}, 3.728751608630973*^9, 3.728813256877846*^9, 
   3.72882737373587*^9, 3.728843009027207*^9, 3.728845615342039*^9, 
   3.733367220212503*^9, 3.790862603802002*^9, 3.790867032499701*^9, 
   3.79227316155514*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1305396984,ExpressionUUID->"cf96c5f8-8290-430f-8ec7-a5aeac738910"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciPDblues", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Ricci tensor ", 
   $CellContext`RicciPDblues[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.7286458808338213`*^9, 3.728655352350588*^9, 3.728656105652667*^9, 
   3.728660573642881*^9, 3.728671045107004*^9, 3.728674867807619*^9, 
   3.728732812164307*^9, 3.728733409396785*^9, 3.728742306408434*^9, 
   3.728743286153721*^9, 3.7287433853131237`*^9, 3.728747012484117*^9, 
   3.728749424327572*^9, 3.728749927526626*^9, 3.728750078069125*^9, 
   3.728750177076168*^9, 3.728750287577373*^9, 3.728750512885186*^9, 
   3.728750572558249*^9, 3.728750718969363*^9, {3.72875097735047*^9, 
   3.728751002205101*^9}, 3.728751608630973*^9, 3.728813256877846*^9, 
   3.72882737373587*^9, 3.728843009027207*^9, 3.728845615342039*^9, 
   3.733367220212503*^9, 3.790862603802002*^9, 3.790867032499701*^9, 
   3.792273161562517*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1114172336,ExpressionUUID->"80b02b7a-4e63-4408-ada1-22c125276511"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaUpblues", "[", 
    RowBox[{"a", ",", "b", ",", "c"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric +1 density ", 
   $CellContext`etaUpblues[$CellContext`a, $CellContext`b, $CellContext`c], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.7286458808338213`*^9, 3.728655352350588*^9, 3.728656105652667*^9, 
   3.728660573642881*^9, 3.728671045107004*^9, 3.728674867807619*^9, 
   3.728732812164307*^9, 3.728733409396785*^9, 3.728742306408434*^9, 
   3.728743286153721*^9, 3.7287433853131237`*^9, 3.728747012484117*^9, 
   3.728749424327572*^9, 3.728749927526626*^9, 3.728750078069125*^9, 
   3.728750177076168*^9, 3.728750287577373*^9, 3.728750512885186*^9, 
   3.728750572558249*^9, 3.728750718969363*^9, {3.72875097735047*^9, 
   3.728751002205101*^9}, 3.728751608630973*^9, 3.728813256877846*^9, 
   3.72882737373587*^9, 3.728843009027207*^9, 3.728845615342039*^9, 
   3.733367220212503*^9, 3.790862603802002*^9, 3.790867032499701*^9, 
   3.7922731615697107`*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1984532959,ExpressionUUID->"25c1bc30-3e76-4558-9191-872f42012ae4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaDownblues", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric -1 density ", 
   $CellContext`etaDownblues[-$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.7286458808338213`*^9, 3.728655352350588*^9, 3.728656105652667*^9, 
   3.728660573642881*^9, 3.728671045107004*^9, 3.728674867807619*^9, 
   3.728732812164307*^9, 3.728733409396785*^9, 3.728742306408434*^9, 
   3.728743286153721*^9, 3.7287433853131237`*^9, 3.728747012484117*^9, 
   3.728749424327572*^9, 3.728749927526626*^9, 3.728750078069125*^9, 
   3.728750177076168*^9, 3.728750287577373*^9, 3.728750512885186*^9, 
   3.728750572558249*^9, 3.728750718969363*^9, {3.72875097735047*^9, 
   3.728751002205101*^9}, 3.728751608630973*^9, 3.728813256877846*^9, 
   3.72882737373587*^9, 3.728843009027207*^9, 3.728845615342039*^9, 
   3.733367220212503*^9, 3.790862603802002*^9, 3.790867032499701*^9, 
   3.792273161577112*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1476920356,ExpressionUUID->"bfdf68ce-aaaf-4e5a-8413-d7db42039835"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
For example, these are the parallel derivatives of the bases acting on a \
given tensor field:\
\>", "MathCaption",
 CellChangeTimes->{{3.728656290588045*^9, 3.7286563303043003`*^9}, {
  3.728656744030398*^9, 3.7286567638451138`*^9}, {3.7286569685473757`*^9, 
  3.7286569810274267`*^9}},
 CellID->1984038762,ExpressionUUID->"391bcf87-e02f-46a0-ae84-04de15fddfa1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PDfred", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"T", "[", 
   RowBox[{
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}]}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.728733416906411*^9, 3.728733418173409*^9}, {
  3.728747016598736*^9, 3.728747017886704*^9}},
 CellLabel->"In[20]:=",
 CellID->1760885965,ExpressionUUID->"f14ab5d7-76cc-4ee3-958e-22377b581561"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox[
     StyleBox["\[ScriptCapitalD]",
      FontColor->RGBColor[1., 0., 0.]], "a"], GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {" ", " "},
          {"b", "c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`PDfred[-$CellContext`a][
   $CellContext`T[-$CellContext`b, -$CellContext`c]],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.728645884210881*^9, 3.728656293543068*^9, 3.728660574295596*^9, 
   3.728671046357476*^9, 3.728674868236816*^9, 3.72873281400082*^9, 
   3.728733418876144*^9, 3.728742306691835*^9, 3.728743286534622*^9, 
   3.7287433856599903`*^9, 3.7287470186269627`*^9, 3.728749426030559*^9, 
   3.728749927769022*^9, 3.7287500782793627`*^9, 3.728750177340556*^9, 
   3.7287502881189833`*^9, 3.72875051313247*^9, 3.728750572793954*^9, 
   3.728750719222075*^9, {3.728750977633602*^9, 3.7287510024074707`*^9}, 
   3.728751608934576*^9, 3.728813257174419*^9, 3.7288273740535583`*^9, 
   3.728843010586295*^9, 3.7288456156288223`*^9, 3.733367225698143*^9, 
   3.7908626051158333`*^9, 3.790867036523588*^9, 3.792273162895042*^9},
 CellLabel->"Out[20]=",
 CellID->1154765805,ExpressionUUID->"e25a6674-5c55-45a7-b6e2-a963129aa111"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PDblues", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"T", "[", 
   RowBox[{
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}]}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.72865633937044*^9, 3.728656342800857*^9}, {
   3.728733422337994*^9, 3.728733423409343*^9}, {3.728747021480248*^9, 
   3.7287470251343*^9}, 3.728749236697857*^9},
 CellLabel->"In[21]:=",
 CellID->320834710,ExpressionUUID->"1a2d5ac5-51d4-43f0-888a-8404a262a17e"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox[
     StyleBox["\[ScriptCapitalD]",
      FontColor->RGBColor[0., 0., 1.]], "a"], GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {" ", " "},
          {"b", "c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`PDblues[-$CellContext`a][
   $CellContext`T[-$CellContext`b, -$CellContext`c]],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.7286563433078003`*^9, 3.72866057444617*^9, 3.728671046973516*^9, 
   3.7286748683889503`*^9, 3.728732814966547*^9, 3.72873342419563*^9, 
   3.728742306908111*^9, 3.728743286666613*^9, 3.728743385877739*^9, 
   3.728747026369301*^9, 3.728749426726337*^9, 3.728749927911965*^9, 
   3.7287500783927097`*^9, 3.72875017747672*^9, 3.728750288276533*^9, 
   3.728750513285879*^9, 3.7287505729418907`*^9, 3.728750719371892*^9, {
   3.728750977768615*^9, 3.728751002498704*^9}, 3.7287516090841703`*^9, 
   3.7288132573247023`*^9, 3.7288273742018414`*^9, 3.728843011320195*^9, 
   3.728845615801454*^9, 3.7333672268815393`*^9, 3.790862605788739*^9, 
   3.790867037907772*^9, 3.792273163534154*^9},
 CellLabel->"Out[21]=",
 CellID->1691980261,ExpressionUUID->"d0bed620-799c-42e4-b848-a09a51795b7f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "We now define bases on the complex vector bundle. This generates additional \
\"inner\" objects for its parallel derivative (",
 StyleBox["FRiemann", "MR"],
 " and ",
 StyleBox["AChristoffel", "MR"],
 "):"
}], "MathCaption",
 CellChangeTimes->{
  3.728655315455933*^9, {3.728656991700653*^9, 3.72865699204287*^9}, {
   3.728657042944353*^9, 
   3.728657044400608*^9}},ExpressionUUID->"b08378bc-0f01-42bb-8aa5-\
c2e25a22ee2e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefBasis", "[", 
  RowBox[{"fcomp", ",", "InnerC", ",", 
   RowBox[{"{", 
    RowBox[{"4", ",", "5"}], "}"}], ",", 
   RowBox[{"Dagger", "\[Rule]", "Complex"}], ",", 
   RowBox[{"BasisColor", "\[Rule]", "Green"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728656111851475*^9, 3.728656112699613*^9}, {
  3.728733438180216*^9, 3.7287334383688593`*^9}, {3.728747037893715*^9, 
  3.728747039485759*^9}},
 CellLabel->"In[22]:=",ExpressionUUID->"4116992a-0df4-4d2e-9008-e0fb7d61d8fd"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefBasis", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"basis \"\>", "\[InvisibleSpace]", "fcomp", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xCoba`DefBasis, ": Defining ", "basis ", $CellContext`fcomp, 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645882419693*^9, 3.728656279753868*^9, 3.7286605745966597`*^9, 
   3.728671049477675*^9, 3.728674868560582*^9, 3.728732817198344*^9, 
   3.728733440196548*^9, 3.728742307385651*^9, 3.7287432868208437`*^9, 
   3.728743385999811*^9, 3.728747043815258*^9, 3.728749428123103*^9, 
   3.7287499280330343`*^9, 3.728750078538865*^9, 3.728750177616901*^9, 
   3.7287502886486177`*^9, 3.728750513436042*^9, 3.728750573076346*^9, 
   3.7287507194950857`*^9, {3.7287509779871817`*^9, 3.728751002599765*^9}, 
   3.728751609230563*^9, 3.72881325745619*^9, 3.728827374359118*^9, 
   3.728843012899191*^9, 3.728845615976507*^9, 3.733367228512617*^9, 
   3.790862607019595*^9, 3.790867044632689*^9, 3.792273164933652*^9},
 CellLabel->"During evaluation of In[22]:=",
 CellID->94044733,ExpressionUUID->"6591214d-41a9-4dbe-b28a-16b494d0bf93"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"parallel derivative \"\>", "\[InvisibleSpace]", 
   RowBox[{"PDfcomp", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "parallel derivative ", 
   $CellContext`PDfcomp[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645882419693*^9, 3.728656279753868*^9, 3.7286605745966597`*^9, 
   3.728671049477675*^9, 3.728674868560582*^9, 3.728732817198344*^9, 
   3.728733440196548*^9, 3.728742307385651*^9, 3.7287432868208437`*^9, 
   3.728743385999811*^9, 3.728747043815258*^9, 3.728749428123103*^9, 
   3.7287499280330343`*^9, 3.728750078538865*^9, 3.728750177616901*^9, 
   3.7287502886486177`*^9, 3.728750513436042*^9, 3.728750573076346*^9, 
   3.7287507194950857`*^9, {3.7287509779871817`*^9, 3.728751002599765*^9}, 
   3.728751609230563*^9, 3.72881325745619*^9, 3.728827374359118*^9, 
   3.728843012899191*^9, 3.728845615976507*^9, 3.733367228512617*^9, 
   3.790862607019595*^9, 3.790867044632689*^9, 3.79227316494691*^9},
 CellLabel->"During evaluation of In[22]:=",
 CellID->61863339,ExpressionUUID->"877fe4cc-7525-490d-8268-30a02429f64c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"TorsionPDfcomp", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "torsion tensor ", 
   $CellContext`TorsionPDfcomp[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645882419693*^9, 3.728656279753868*^9, 3.7286605745966597`*^9, 
   3.728671049477675*^9, 3.728674868560582*^9, 3.728732817198344*^9, 
   3.728733440196548*^9, 3.728742307385651*^9, 3.7287432868208437`*^9, 
   3.728743385999811*^9, 3.728747043815258*^9, 3.728749428123103*^9, 
   3.7287499280330343`*^9, 3.728750078538865*^9, 3.728750177616901*^9, 
   3.7287502886486177`*^9, 3.728750513436042*^9, 3.728750573076346*^9, 
   3.7287507194950857`*^9, {3.7287509779871817`*^9, 3.728751002599765*^9}, 
   3.728751609230563*^9, 3.72881325745619*^9, 3.728827374359118*^9, 
   3.728843012899191*^9, 3.728845615976507*^9, 3.733367228512617*^9, 
   3.790862607019595*^9, 3.790867044632689*^9, 3.79227316496134*^9},
 CellLabel->"During evaluation of In[22]:=",
 CellID->355053425,ExpressionUUID->"2d346f63-5728-446f-a198-415a4ad27ea8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelPDfcomp", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "non-symmetric Christoffel tensor ", 
   $CellContext`ChristoffelPDfcomp[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645882419693*^9, 3.728656279753868*^9, 3.7286605745966597`*^9, 
   3.728671049477675*^9, 3.728674868560582*^9, 3.728732817198344*^9, 
   3.728733440196548*^9, 3.728742307385651*^9, 3.7287432868208437`*^9, 
   3.728743385999811*^9, 3.728747043815258*^9, 3.728749428123103*^9, 
   3.7287499280330343`*^9, 3.728750078538865*^9, 3.728750177616901*^9, 
   3.7287502886486177`*^9, 3.728750513436042*^9, 3.728750573076346*^9, 
   3.7287507194950857`*^9, {3.7287509779871817`*^9, 3.728751002599765*^9}, 
   3.728751609230563*^9, 3.72881325745619*^9, 3.728827374359118*^9, 
   3.728843012899191*^9, 3.728845615976507*^9, 3.733367228512617*^9, 
   3.790862607019595*^9, 3.790867044632689*^9, 3.792273164968862*^9},
 CellLabel->"During evaluation of In[22]:=",
 CellID->1673342609,ExpressionUUID->"fbf8c4c1-c8a7-4985-a24c-274ce27758b7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Riemann tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RiemannPDfcomp", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Riemann tensor ", 
   $CellContext`RiemannPDfcomp[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, $CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645882419693*^9, 3.728656279753868*^9, 3.7286605745966597`*^9, 
   3.728671049477675*^9, 3.728674868560582*^9, 3.728732817198344*^9, 
   3.728733440196548*^9, 3.728742307385651*^9, 3.7287432868208437`*^9, 
   3.728743385999811*^9, 3.728747043815258*^9, 3.728749428123103*^9, 
   3.7287499280330343`*^9, 3.728750078538865*^9, 3.728750177616901*^9, 
   3.7287502886486177`*^9, 3.728750513436042*^9, 3.728750573076346*^9, 
   3.7287507194950857`*^9, {3.7287509779871817`*^9, 3.728751002599765*^9}, 
   3.728751609230563*^9, 3.72881325745619*^9, 3.728827374359118*^9, 
   3.728843012899191*^9, 3.728845615976507*^9, 3.733367228512617*^9, 
   3.790862607019595*^9, 3.790867044632689*^9, 3.7922731649764214`*^9},
 CellLabel->"During evaluation of In[22]:=",
 CellID->423388351,ExpressionUUID->"a4b76c2e-1ce9-4892-8bed-93608127618b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciPDfcomp", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Ricci tensor ", 
   $CellContext`RicciPDfcomp[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645882419693*^9, 3.728656279753868*^9, 3.7286605745966597`*^9, 
   3.728671049477675*^9, 3.728674868560582*^9, 3.728732817198344*^9, 
   3.728733440196548*^9, 3.728742307385651*^9, 3.7287432868208437`*^9, 
   3.728743385999811*^9, 3.728747043815258*^9, 3.728749428123103*^9, 
   3.7287499280330343`*^9, 3.728750078538865*^9, 3.728750177616901*^9, 
   3.7287502886486177`*^9, 3.728750513436042*^9, 3.728750573076346*^9, 
   3.7287507194950857`*^9, {3.7287509779871817`*^9, 3.728751002599765*^9}, 
   3.728751609230563*^9, 3.72881325745619*^9, 3.728827374359118*^9, 
   3.728843012899191*^9, 3.728845615976507*^9, 3.733367228512617*^9, 
   3.790862607019595*^9, 3.790867044632689*^9, 3.79227316498409*^9},
 CellLabel->"During evaluation of In[22]:=",
 CellID->1274692746,ExpressionUUID->"08d1647f-09ff-4012-afe7-469fc346d5ca"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"nonsymmetric AChristoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"AChristoffelPDfcomp", "[", 
    RowBox[{"A", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "C"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "nonsymmetric AChristoffel tensor ", 
   $CellContext`AChristoffelPDfcomp[$CellContext`A, -$CellContext`b, -C], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645882419693*^9, 3.728656279753868*^9, 3.7286605745966597`*^9, 
   3.728671049477675*^9, 3.728674868560582*^9, 3.728732817198344*^9, 
   3.728733440196548*^9, 3.728742307385651*^9, 3.7287432868208437`*^9, 
   3.728743385999811*^9, 3.728747043815258*^9, 3.728749428123103*^9, 
   3.7287499280330343`*^9, 3.728750078538865*^9, 3.728750177616901*^9, 
   3.7287502886486177`*^9, 3.728750513436042*^9, 3.728750573076346*^9, 
   3.7287507194950857`*^9, {3.7287509779871817`*^9, 3.728751002599765*^9}, 
   3.728751609230563*^9, 3.72881325745619*^9, 3.728827374359118*^9, 
   3.728843012899191*^9, 3.728845615976507*^9, 3.733367228512617*^9, 
   3.790862607019595*^9, 3.790867044632689*^9, 3.792273164991181*^9},
 CellLabel->"During evaluation of In[22]:=",
 CellID->464119550,ExpressionUUID->"69c6d938-4cd9-4a55-ad85-28d5d297fe0c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"nonsymmetric AChristoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"AChristoffelPDfcomp\[Dagger]", "[", 
    RowBox[{"A\[Dagger]", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "C\[Dagger]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "nonsymmetric AChristoffel tensor ", 
   $CellContext`AChristoffelPDfcomp\[Dagger][$CellContext`A\[Dagger], \
-$CellContext`b, -$CellContext`C\[Dagger]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645882419693*^9, 3.728656279753868*^9, 3.7286605745966597`*^9, 
   3.728671049477675*^9, 3.728674868560582*^9, 3.728732817198344*^9, 
   3.728733440196548*^9, 3.728742307385651*^9, 3.7287432868208437`*^9, 
   3.728743385999811*^9, 3.728747043815258*^9, 3.728749428123103*^9, 
   3.7287499280330343`*^9, 3.728750078538865*^9, 3.728750177616901*^9, 
   3.7287502886486177`*^9, 3.728750513436042*^9, 3.728750573076346*^9, 
   3.7287507194950857`*^9, {3.7287509779871817`*^9, 3.728751002599765*^9}, 
   3.728751609230563*^9, 3.72881325745619*^9, 3.728827374359118*^9, 
   3.728843012899191*^9, 3.728845615976507*^9, 3.733367228512617*^9, 
   3.790862607019595*^9, 3.790867044632689*^9, 3.7922731649984827`*^9},
 CellLabel->"During evaluation of In[22]:=",
 CellID->796670946,ExpressionUUID->"19502156-09e8-4f67-a721-f2862d1f3b09"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing FRiemann tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"FRiemannPDfcomp", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "C"}], ",", "D"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "vanishing FRiemann tensor ", 
   $CellContext`FRiemannPDfcomp[-$CellContext`a, -$CellContext`b, -C, D], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645882419693*^9, 3.728656279753868*^9, 3.7286605745966597`*^9, 
   3.728671049477675*^9, 3.728674868560582*^9, 3.728732817198344*^9, 
   3.728733440196548*^9, 3.728742307385651*^9, 3.7287432868208437`*^9, 
   3.728743385999811*^9, 3.728747043815258*^9, 3.728749428123103*^9, 
   3.7287499280330343`*^9, 3.728750078538865*^9, 3.728750177616901*^9, 
   3.7287502886486177`*^9, 3.728750513436042*^9, 3.728750573076346*^9, 
   3.7287507194950857`*^9, {3.7287509779871817`*^9, 3.728751002599765*^9}, 
   3.728751609230563*^9, 3.72881325745619*^9, 3.728827374359118*^9, 
   3.728843012899191*^9, 3.728845615976507*^9, 3.733367228512617*^9, 
   3.790862607019595*^9, 3.790867044632689*^9, 3.792273165005732*^9},
 CellLabel->"During evaluation of In[22]:=",
 CellID->88050377,ExpressionUUID->"6253fa69-9408-4fcb-bef8-27bd65716fa3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing FRiemann tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"FRiemannPDfcomp\[Dagger]", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "C\[Dagger]"}], ",", "D\[Dagger]"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "vanishing FRiemann tensor ", 
   $CellContext`FRiemannPDfcomp\[Dagger][-$CellContext`a, -$CellContext`b, \
-$CellContext`C\[Dagger], $CellContext`D\[Dagger]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645882419693*^9, 3.728656279753868*^9, 3.7286605745966597`*^9, 
   3.728671049477675*^9, 3.728674868560582*^9, 3.728732817198344*^9, 
   3.728733440196548*^9, 3.728742307385651*^9, 3.7287432868208437`*^9, 
   3.728743385999811*^9, 3.728747043815258*^9, 3.728749428123103*^9, 
   3.7287499280330343`*^9, 3.728750078538865*^9, 3.728750177616901*^9, 
   3.7287502886486177`*^9, 3.728750513436042*^9, 3.728750573076346*^9, 
   3.7287507194950857`*^9, {3.7287509779871817`*^9, 3.728751002599765*^9}, 
   3.728751609230563*^9, 3.72881325745619*^9, 3.728827374359118*^9, 
   3.728843012899191*^9, 3.728845615976507*^9, 3.733367228512617*^9, 
   3.790862607019595*^9, 3.790867044632689*^9, 3.792273165013291*^9},
 CellLabel->"During evaluation of In[22]:=",
 CellID->246453456,ExpressionUUID->"4bd00293-c989-41f8-9630-c31ed377a1c2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefBasis", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"basis \"\>", "\[InvisibleSpace]", 
   "fcomp\[Dagger]", "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xCoba`DefBasis, ": Defining ", 
   "basis ", $CellContext`fcomp\[Dagger], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645882419693*^9, 3.728656279753868*^9, 3.7286605745966597`*^9, 
   3.728671049477675*^9, 3.728674868560582*^9, 3.728732817198344*^9, 
   3.728733440196548*^9, 3.728742307385651*^9, 3.7287432868208437`*^9, 
   3.728743385999811*^9, 3.728747043815258*^9, 3.728749428123103*^9, 
   3.7287499280330343`*^9, 3.728750078538865*^9, 3.728750177616901*^9, 
   3.7287502886486177`*^9, 3.728750513436042*^9, 3.728750573076346*^9, 
   3.7287507194950857`*^9, {3.7287509779871817`*^9, 3.728751002599765*^9}, 
   3.728751609230563*^9, 3.72881325745619*^9, 3.728827374359118*^9, 
   3.728843012899191*^9, 3.728845615976507*^9, 3.733367228512617*^9, 
   3.790862607019595*^9, 3.790867044632689*^9, 3.7922731650207*^9},
 CellLabel->"During evaluation of In[22]:=",
 CellID->1794763576,ExpressionUUID->"b3e02462-15be-444c-a299-3476d3805d80"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaUpfcomp", "[", 
    RowBox[{"A", ",", "B"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric +1 density ", 
   $CellContext`etaUpfcomp[$CellContext`A, $CellContext`B], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645882419693*^9, 3.728656279753868*^9, 3.7286605745966597`*^9, 
   3.728671049477675*^9, 3.728674868560582*^9, 3.728732817198344*^9, 
   3.728733440196548*^9, 3.728742307385651*^9, 3.7287432868208437`*^9, 
   3.728743385999811*^9, 3.728747043815258*^9, 3.728749428123103*^9, 
   3.7287499280330343`*^9, 3.728750078538865*^9, 3.728750177616901*^9, 
   3.7287502886486177`*^9, 3.728750513436042*^9, 3.728750573076346*^9, 
   3.7287507194950857`*^9, {3.7287509779871817`*^9, 3.728751002599765*^9}, 
   3.728751609230563*^9, 3.72881325745619*^9, 3.728827374359118*^9, 
   3.728843012899191*^9, 3.728845615976507*^9, 3.733367228512617*^9, 
   3.790862607019595*^9, 3.790867044632689*^9, 3.7922731650278*^9},
 CellLabel->"During evaluation of In[22]:=",
 CellID->1036116070,ExpressionUUID->"dc918e01-1f1f-4b35-87ba-a2e99dfbfefe"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaUpfcomp\[Dagger]", "[", 
    RowBox[{"A\[Dagger]", ",", "B\[Dagger]"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric +1 density ", 
   $CellContext`etaUpfcomp\[Dagger][$CellContext`A\[Dagger], $CellContext`B\
\[Dagger]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645882419693*^9, 3.728656279753868*^9, 3.7286605745966597`*^9, 
   3.728671049477675*^9, 3.728674868560582*^9, 3.728732817198344*^9, 
   3.728733440196548*^9, 3.728742307385651*^9, 3.7287432868208437`*^9, 
   3.728743385999811*^9, 3.728747043815258*^9, 3.728749428123103*^9, 
   3.7287499280330343`*^9, 3.728750078538865*^9, 3.728750177616901*^9, 
   3.7287502886486177`*^9, 3.728750513436042*^9, 3.728750573076346*^9, 
   3.7287507194950857`*^9, {3.7287509779871817`*^9, 3.728751002599765*^9}, 
   3.728751609230563*^9, 3.72881325745619*^9, 3.728827374359118*^9, 
   3.728843012899191*^9, 3.728845615976507*^9, 3.733367228512617*^9, 
   3.790862607019595*^9, 3.790867044632689*^9, 3.792273165035139*^9},
 CellLabel->"During evaluation of In[22]:=",
 CellID->57894967,ExpressionUUID->"f1fb491a-d09d-4286-bc60-e89eff9af817"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaDownfcomp", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "B"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric -1 density ", 
   $CellContext`etaDownfcomp[-$CellContext`A, -$CellContext`B], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645882419693*^9, 3.728656279753868*^9, 3.7286605745966597`*^9, 
   3.728671049477675*^9, 3.728674868560582*^9, 3.728732817198344*^9, 
   3.728733440196548*^9, 3.728742307385651*^9, 3.7287432868208437`*^9, 
   3.728743385999811*^9, 3.728747043815258*^9, 3.728749428123103*^9, 
   3.7287499280330343`*^9, 3.728750078538865*^9, 3.728750177616901*^9, 
   3.7287502886486177`*^9, 3.728750513436042*^9, 3.728750573076346*^9, 
   3.7287507194950857`*^9, {3.7287509779871817`*^9, 3.728751002599765*^9}, 
   3.728751609230563*^9, 3.72881325745619*^9, 3.728827374359118*^9, 
   3.728843012899191*^9, 3.728845615976507*^9, 3.733367228512617*^9, 
   3.790862607019595*^9, 3.790867044632689*^9, 3.792273165042238*^9},
 CellLabel->"During evaluation of In[22]:=",
 CellID->230618278,ExpressionUUID->"3fd0c586-9037-4ff6-8575-cdfb0d270a5d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaDownfcomp\[Dagger]", "[", 
    RowBox[{
     RowBox[{"-", "A\[Dagger]"}], ",", 
     RowBox[{"-", "B\[Dagger]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric -1 density ", 
   $CellContext`etaDownfcomp\[Dagger][-$CellContext`A\[Dagger], \
-$CellContext`B\[Dagger]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645882419693*^9, 3.728656279753868*^9, 3.7286605745966597`*^9, 
   3.728671049477675*^9, 3.728674868560582*^9, 3.728732817198344*^9, 
   3.728733440196548*^9, 3.728742307385651*^9, 3.7287432868208437`*^9, 
   3.728743385999811*^9, 3.728747043815258*^9, 3.728749428123103*^9, 
   3.7287499280330343`*^9, 3.728750078538865*^9, 3.728750177616901*^9, 
   3.7287502886486177`*^9, 3.728750513436042*^9, 3.728750573076346*^9, 
   3.7287507194950857`*^9, {3.7287509779871817`*^9, 3.728751002599765*^9}, 
   3.728751609230563*^9, 3.72881325745619*^9, 3.728827374359118*^9, 
   3.728843012899191*^9, 3.728845615976507*^9, 3.733367228512617*^9, 
   3.790862607019595*^9, 3.790867044632689*^9, 3.792273165049589*^9},
 CellLabel->"During evaluation of In[22]:=",
 CellID->1940899158,ExpressionUUID->"b3e11ff3-0898-461b-bd7d-bca31dda6d13"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "Once we have defined bases, the first task is to construct tensors. The \
simplest way is to build linear combinations of basis vectors with constant \
coefficients. ",
 StyleBox["xCoba`", "MR"],
 "  also offers a matrix/array representation using the head CTensor."
}], "Text",
 CellChangeTimes->{{3.728657543834518*^9, 3.728657738761075*^9}},
 CellID->1031311492,ExpressionUUID->"41b7ddbd-e6b9-4e91-b209-172a0ce81027"],

Cell[CellGroupData[{

Cell["Linear combination of the basis vector fields:", "MathCaption",
 CellChangeTimes->{{3.7286577709429693`*^9, 3.728657784998665*^9}},
 CellID->567493020,ExpressionUUID->"8cf80a25-7610-4ee8-bf5f-1a09bd97c61b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{
    RowBox[{"Sqrt", "[", "3", "]"}], ",", 
    RowBox[{"4", "/", "5"}], ",", 
    RowBox[{"-", "2"}]}], "}"}], ".", "basis"}]], "Input",
 CellChangeTimes->{{3.728657788410364*^9, 3.7286578097334127`*^9}, {
  3.728657853236046*^9, 3.728657867611246*^9}, {3.728659175198987*^9, 
  3.728659176333827*^9}, {3.728659221868815*^9, 3.728659223515777*^9}},
 CellLabel->"In[23]:=",
 CellID->798724419,ExpressionUUID->"939b2cc8-339b-4dbf-aef5-bc4fc457a978"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   SqrtBox["3"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`fred}, $CellContext`a],
    Editable->False]}], "+", 
  FractionBox[
   RowBox[{"4", " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
     Editable->False]}], "5"], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`fred}, $CellContext`a],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{{3.72865778932733*^9, 3.728657809966426*^9}, 
   3.728657868703639*^9, 3.728658207995817*^9, 3.728659176775906*^9, 
   3.7286592239652157`*^9, 3.7286605755591707`*^9, 3.728671053211897*^9, 
   3.7286748692553368`*^9, 3.7287328200185328`*^9, 3.728733451197308*^9, 
   3.728742308937552*^9, 3.72874328733191*^9, 3.72874338644768*^9, 
   3.728747054089987*^9, 3.728749433731832*^9, 3.728749928348205*^9, 
   3.7287500788663197`*^9, 3.728750178418771*^9, 3.728750289768194*^9, 
   3.728750513765011*^9, 3.728750573396962*^9, 3.728750719963344*^9, {
   3.7287509782892437`*^9, 3.72875100289567*^9}, 3.728751609642486*^9, 
   3.728813257911372*^9, 3.728827374841497*^9, 3.728843015870945*^9, 
   3.72884561639473*^9, 3.7333672316606627`*^9, 3.790862611318912*^9, 
   3.790867047190349*^9, 3.7922731669994917`*^9},
 CellLabel->"Out[23]=",
 CellID->951812572,ExpressionUUID->"acf2015e-6d38-40d4-9196-ed53dd263eb3"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Find a particular component using the dual basis:", "MathCaption",
 CellChangeTimes->{{3.728657831036209*^9, 3.728657849931662*^9}},
 CellID->1224072815,ExpressionUUID->"00a20cd5-14c5-4389-b2e8-c7fec37e866d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", " ", 
  RowBox[{"Basis", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"{", 
     RowBox[{"1", ",", "fred"}], "}"}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.728657873326367*^9, 3.728657888115551*^9}, {
  3.728733455176523*^9, 3.728733457483597*^9}, {3.728747056596898*^9, 
  3.7287470578517923`*^9}},
 CellLabel->"In[24]:=",
 CellID->146109740,ExpressionUUID->"0fb10a9e-1a6e-46b7-ace1-db410f868873"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]]},
          {"a", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`fred}],
   Editable->False], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     SqrtBox["3"], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", "a"},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{0, -$CellContext`fred}, $CellContext`a],
      Editable->False]}], "+", 
    FractionBox[
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", "a"},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
       Editable->False]}], "5"], "-", 
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", "a"},
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{2, -$CellContext`fred}, $CellContext`a],
      Editable->False]}]}], ")"}]}]], "Output",
 CellChangeTimes->{
  3.728657889820837*^9, 3.728658209943397*^9, 3.728659178229656*^9, 
   3.728659225134118*^9, 3.728660575724956*^9, 3.728671054345994*^9, 
   3.7286748693983183`*^9, 3.728732821369384*^9, 3.728733459498345*^9, 
   3.728742309140935*^9, 3.728743287485834*^9, 3.728743386563595*^9, 
   3.728747058608076*^9, 3.7287494349934187`*^9, 3.72874992849739*^9, 
   3.728750079048293*^9, 3.728750179020624*^9, 3.72875028996517*^9, 
   3.728750513925475*^9, 3.72875057354738*^9, 3.7287507202221537`*^9, {
   3.7287509784361897`*^9, 3.728751003003475*^9}, 3.72875160980897*^9, 
   3.728813258060451*^9, 3.728827374992092*^9, 3.728843017320829*^9, 
   3.7288456165726633`*^9, 3.733367232951432*^9, 3.790862614084567*^9, 
   3.790867053046297*^9, 3.792273168468376*^9},
 CellLabel->"Out[24]=",
 CellID->407381183,ExpressionUUID->"cafb6616-daab-42b6-bb4c-c077fc2b5f1a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Contraction of basis elements is not automatic and needs to be forced with \
ContractBasis:\
\>", "MathCaption",
 CellChangeTimes->{{3.7286579086421127`*^9, 3.7286579419502583`*^9}},
 CellID->1517941996,ExpressionUUID->"276f5939-64c7-4551-81d1-62b1cfe2aefe"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ContractBasis"}]], "Input",
 CellChangeTimes->{{3.7286578914688883`*^9, 3.728657894305317*^9}},
 CellLabel->"In[25]:=",
 CellID->873925861,ExpressionUUID->"b40a2a60-a516-441e-b0b1-92132c68a089"],

Cell[BoxData[
 FractionBox["4", "5"]], "Output",
 CellChangeTimes->{
  3.7286578947395573`*^9, 3.72865821091333*^9, 3.728659179618195*^9, 
   3.728659226186515*^9, 3.728660575871538*^9, 3.728671055390955*^9, 
   3.72867486953607*^9, 3.728732822800817*^9, 3.7287334613976803`*^9, 
   3.7287423094180937`*^9, 3.728743287617667*^9, 3.728743386679995*^9, 
   3.728747060263914*^9, 3.72874943642666*^9, 3.72874992868524*^9, 
   3.72875007920193*^9, 3.7287501796193247`*^9, 3.728750290131489*^9, 
   3.728750514087226*^9, 3.7287505736927567`*^9, 3.728750720331127*^9, {
   3.7287509785733356`*^9, 3.72875100309855*^9}, 3.7287516099625483`*^9, 
   3.728813258205059*^9, 3.72882737513629*^9, 3.728843018672751*^9, 
   3.728845616760726*^9, 3.733367234267839*^9, 3.790862615530878*^9, 
   3.79086705456151*^9, 3.7922731697002*^9},
 CellLabel->"Out[25]=",
 CellID->896289178,ExpressionUUID->"69738db6-24e3-4311-8947-118ab01abcf0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We can do the same thing with abstract tensors. This is the 1-component of \
the vector v:\
\>", "MathCaption",
 CellChangeTimes->{{3.728657870771105*^9, 3.7286578715551453`*^9}, {
  3.728657968591552*^9, 
  3.728657995292942*^9}},ExpressionUUID->"893ff924-3a92-4a50-bdcb-\
73ecee828f56"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"v", "[", "a", "]"}], 
  RowBox[{"Basis", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"{", 
     RowBox[{"1", ",", "fred"}], "}"}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.728657997345171*^9, 3.728658004140646*^9}, {
  3.728733465472506*^9, 3.728733471539266*^9}, {3.728747063468745*^9, 
  3.728747064084592*^9}},
 CellLabel->"In[26]:=",
 CellID->495245388,ExpressionUUID->"88eb222a-0a82-4e42-b087-30b334c8c6b3"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]]},
          {"a", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`fred}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"a"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`a],
   Editable->False]}]], "Output",
 CellChangeTimes->{
  3.728658004849516*^9, {3.728658094277974*^9, 3.728658104546372*^9}, 
   3.728658212313325*^9, 3.728659180781633*^9, 3.728659227234871*^9, 
   3.728660576038485*^9, 3.728671056424767*^9, 3.72867486968263*^9, 
   3.72873282418576*^9, 3.728733473266857*^9, 3.728742309989181*^9, 
   3.728743287765223*^9, 3.728743386797634*^9, 3.72874706484513*^9, 
   3.7287494382328672`*^9, 3.728749928845009*^9, 3.728750079381485*^9, 
   3.7287501805235367`*^9, 3.728750290294507*^9, 3.72875051421445*^9, 
   3.7287505738264103`*^9, 3.728750720437146*^9, {3.728750978719502*^9, 
   3.72875100318469*^9}, 3.728751610098958*^9, 3.7288132583588533`*^9, 
   3.728827375290291*^9, 3.728843019971199*^9, 3.728845616944168*^9, 
   3.733367235483755*^9, 3.790862617500257*^9, 3.790867055810647*^9, 
   3.792273171245263*^9},
 CellLabel->"Out[26]=",
 CellID->29071767,ExpressionUUID->"ece8551b-c1c4-421a-8c1a-ff37b29d4dd9"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ContractBasis"}]], "Input",
 CellChangeTimes->{{3.72865800629854*^9, 3.728658009988182*^9}},
 CellLabel->"In[27]:=",
 CellID->354917875,ExpressionUUID->"e352fa8f-f991-40f4-9f4a-696eb0edc24f"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"v", 
      StyleBox[GridBox[{
         {
          StyleBox["1",
           FontColor->RGBColor[1., 0., 0.]]},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`v[{1, $CellContext`fred}],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.728658010352304*^9, {3.728658094899235*^9, 3.728658105132345*^9}, 
   3.7286582128967*^9, 3.728659181365068*^9, 3.728659227820901*^9, 
   3.728660576169086*^9, 3.7286710570891438`*^9, 3.728674869822386*^9, 
   3.728732825035427*^9, 3.728733474349777*^9, 3.7287423101119633`*^9, 
   3.7287432879005547`*^9, 3.728743386946542*^9, 3.728747067076078*^9, 
   3.728749438932197*^9, 3.728749928955408*^9, 3.728750079552868*^9, 
   3.7287501809401407`*^9, 3.728750290457869*^9, 3.72875051437143*^9, 
   3.728750573975359*^9, 3.728750720543305*^9, {3.72875097885708*^9, 
   3.728751003279801*^9}, 3.72875161026226*^9, 3.728813258508492*^9, 
   3.7288273754386168`*^9, 3.728843020673112*^9, 3.728845617110462*^9, 
   3.7333672362671623`*^9, 3.7908626204197206`*^9, 3.790867056530903*^9, 
   3.79227317180599*^9},
 CellLabel->"Out[27]=",
 CellID->2128758162,ExpressionUUID->"5b33e57d-752f-4b2f-8f8b-45257cd01ecd"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Such object is a scalar, and not a vector. It is a scalar component of the \
vector v, and hence has no free index:\
\>", "MathCaption",
 CellChangeTimes->{{3.728658026652529*^9, 3.728658058410124*^9}, {
  3.728658113574903*^9, 3.72865812752693*^9}},
 CellID->1926011727,ExpressionUUID->"01922e37-a050-4631-968d-54ca32a1b501"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", 
  RowBox[{"IndicesOf", "[", "Free", "]"}]}]], "Input",
 CellChangeTimes->{{3.7286580605327177`*^9, 3.7286581024114122`*^9}},
 CellLabel->"In[28]:=",
 CellID->1726618588,ExpressionUUID->"2fd3c73e-8945-46f6-9f02-1723d21557c5"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   StyleBox[
    RowBox[{"{", "}"}],
    FontColor->RGBColor[0, 0, 1]],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[],
  Editable->False]], "Output",
 CellChangeTimes->{{3.728658069691043*^9, 3.728658106140317*^9}, 
   3.7286582139586277`*^9, 3.728659182397478*^9, 3.728659228887414*^9, 
   3.7286605763310623`*^9, 3.728671058132729*^9, 3.728674869971005*^9, 
   3.72873282631888*^9, 3.72873347573483*^9, 3.728742310270246*^9, 
   3.728743288053254*^9, 3.728743387083701*^9, 3.7287470696638937`*^9, 
   3.728749440154785*^9, 3.728749929065579*^9, 3.728750079686417*^9, 
   3.728750181135494*^9, 3.7287502907730417`*^9, 3.728750514518721*^9, 
   3.728750574122034*^9, 3.7287507206435547`*^9, {3.7287509789696693`*^9, 
   3.728751003359228*^9}, 3.72875161041416*^9, 3.72881325863787*^9, 
   3.728827375586524*^9, 3.728843021904561*^9, 3.728845617282915*^9, 
   3.733367237484853*^9, 3.790862621855618*^9, 3.790867057610441*^9, 
   3.792273172924511*^9},
 CellLabel->"Out[28]=",
 CellID->734337630,ExpressionUUID->"039476bc-8c05-46fb-9df3-55717c3c1d3a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["These are all the components of a tensor in given bases:", "MathCaption",
 CellChangeTimes->{{3.728659553726243*^9, 3.7286595657803907`*^9}},
 CellID->2040884250,ExpressionUUID->"f315b260-ffa6-404c-8aba-49421fe2cd0d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentArray", "[", 
  RowBox[{"T", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "blues"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "blues"}]}], "}"}]}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.728733483096191*^9, 3.728733485543358*^9}, {
  3.7287470730764*^9, 3.728747078931922*^9}, {3.728749237674259*^9, 
  3.728749238371763*^9}},
 CellLabel->"In[29]:=",
 CellID->1532779866,ExpressionUUID->"b852019f-6038-4c28-98ce-8b2f112ff081"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["0",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`blues}, {0, -$CellContext`blues}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`blues}, {1, -$CellContext`blues}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`blues}, {2, -$CellContext`blues}],
      Editable->False]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["0",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`blues}, {0, -$CellContext`blues}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`blues}, {1, -$CellContext`blues}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`blues}, {2, -$CellContext`blues}],
      Editable->False]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["0",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, -$CellContext`blues}, {0, -$CellContext`blues}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, -$CellContext`blues}, {1, -$CellContext`blues}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, -$CellContext`blues}, {2, -$CellContext`blues}],
      Editable->False]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{
  3.728645913465144*^9, 3.7286595675123873`*^9, 3.728660576828341*^9, 
   3.728671059242107*^9, 3.7286748701285877`*^9, 3.728732830383156*^9, 
   3.7287334882366533`*^9, 3.728742310674667*^9, 3.7287432882158747`*^9, 
   3.7287433872182207`*^9, 3.7287470796259117`*^9, 3.728749441610855*^9, 
   3.7287499292007093`*^9, 3.728750079860327*^9, 3.728750181540159*^9, 
   3.728750291063883*^9, 3.728750514761248*^9, 3.7287505742767773`*^9, 
   3.7287507207557917`*^9, {3.728750979139906*^9, 3.728751003466155*^9}, 
   3.728751610572269*^9, 3.728813258799909*^9, 3.7288273757476797`*^9, 
   3.728843023180586*^9, 3.728845617462277*^9, 3.733367238851947*^9, 
   3.790862623028624*^9, 3.7908670586740637`*^9, 3.792273174667753*^9},
 CellLabel->"Out[29]=",
 CellID->1881806026,ExpressionUUID->"0e33bec5-87a2-48d0-b3e5-c0bcca9df27e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Notice that T is a symmetric tensor, so this can be simplified\
\>", "MathCaption",
 CellID->217378450,ExpressionUUID->"31d51e36-85ae-433d-b70d-5fef2a5210bc"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Simplification", "[", "%", "]"}], "//", "MatrixForm"}]], "Input",
 CellLabel->"In[30]:=",
 CellID->609925603,ExpressionUUID->"95c84cfd-f695-49df-bf33-878fbd0f1e50"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {"0", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`blues}, {1, -$CellContext`blues}],
       Editable->False], 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`blues}, {2, -$CellContext`blues}],
       Editable->False]},
     {
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`blues}, {1, -$CellContext`blues}],
        Editable->False]}], "0", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`blues}, {2, -$CellContext`blues}],
       Editable->False]},
     {
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`blues}, {2, -$CellContext`blues}],
        Editable->False]}], 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{1, -$CellContext`blues}, {2, -$CellContext`blues}],
        Editable->False]}], "0"}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Baseline}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{
  3.728645914748652*^9, 3.7286595706313143`*^9, 3.728660576972238*^9, 
   3.728671061709361*^9, 3.728674870276717*^9, 3.728732831891994*^9, 
   3.728733490115657*^9, 3.728742310814097*^9, 3.72874328838453*^9, 
   3.728743387409368*^9, 3.7287470844769583`*^9, 3.728749442933326*^9, 
   3.728749929317546*^9, 3.728750080021885*^9, 3.7287501816988297`*^9, 
   3.728750291231916*^9, 3.728750514970084*^9, 3.728750574405529*^9, 
   3.72875072088616*^9, {3.72875097933547*^9, 3.7287510035829782`*^9}, 
   3.728751610792466*^9, 3.7288132589439993`*^9, 3.728827375891399*^9, 
   3.7288430244731693`*^9, 3.7288456176781816`*^9, 3.733367240083146*^9, 
   3.790862624139817*^9, 3.79086705970389*^9, 3.792273176769701*^9},
 CellLabel->"Out[30]//MatrixForm=",
 CellID->2026104355,ExpressionUUID->"0dd1d415-ef67-429d-a517-377fc94aee8a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["These are all contractions of the basis and cobasis vectors:", \
"MathCaption",
 CellChangeTimes->{{3.728658261080299*^9, 3.728658275112071*^9}},
 CellID->467947572,ExpressionUUID->"9b6da2d7-1380-4536-96d6-bf6760d56437"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Outer", "[", 
   RowBox[{"Times", ",", "basis", ",", "cobasis"}], "]"}], "//", 
  "MatrixForm"}]], "Input",
 CellChangeTimes->{{3.728658238158003*^9, 3.728658252599633*^9}},
 CellLabel->"In[31]:=",
 CellID->39003799,ExpressionUUID->"c683d903-b92a-41aa-b2c1-fddf615ddfe1"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]]},
               {"a", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`fred}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", "a"},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[{0, -$CellContext`fred}, $CellContext`a],
        Editable->False]}], 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]},
               {"a", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`fred}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", "a"},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[{0, -$CellContext`fred}, $CellContext`a],
        Editable->False]}], 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {"a", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`fred}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", "a"},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[{0, -$CellContext`fred}, $CellContext`a],
        Editable->False]}]},
     {
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]]},
               {"a", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`fred}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", "a"},
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
        Editable->False]}], 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]},
               {"a", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`fred}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", "a"},
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
        Editable->False]}], 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {"a", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`fred}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", "a"},
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
        Editable->False]}]},
     {
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]]},
               {"a", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`fred}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", "a"},
               {
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[{2, -$CellContext`fred}, $CellContext`a],
        Editable->False]}], 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]},
               {"a", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`fred}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", "a"},
               {
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[{2, -$CellContext`fred}, $CellContext`a],
        Editable->False]}], 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {"a", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`fred}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", "a"},
               {
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[{2, -$CellContext`fred}, $CellContext`a],
        Editable->False]}]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Baseline}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{{3.728658245946588*^9, 3.728658253660502*^9}, 
   3.7286591834475327`*^9, 3.728659230237402*^9, 3.728660577133449*^9, 
   3.728671063292205*^9, 3.728674870453608*^9, 3.728732833408968*^9, 
   3.72873349196168*^9, 3.7287423109711437`*^9, 3.7287432890450363`*^9, 
   3.7287433876032953`*^9, 3.728747087112566*^9, 3.7287494442828503`*^9, 
   3.728749929454063*^9, 3.7287500801840677`*^9, 3.72875018186316*^9, 
   3.728750291752371*^9, 3.7287505151023273`*^9, 3.728750574547233*^9, 
   3.728750721014017*^9, {3.72875097951621*^9, 3.7287510037077436`*^9}, 
   3.7287516109768963`*^9, 3.728813259091239*^9, 3.728827376057448*^9, 
   3.728843025715432*^9, 3.728845617915887*^9, 3.7333672414016457`*^9, 
   3.7908626277834682`*^9, 3.790867063192526*^9, 3.7922731780510883`*^9},
 CellLabel->"Out[31]//MatrixForm=",
 CellID->432518983,ExpressionUUID->"b9b7c6be-1a3a-4740-8de9-5c594af361aa"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ContractBasis"}]], "Input",
 CellChangeTimes->{{3.728658285587792*^9, 3.7286582889984303`*^9}},
 CellLabel->"In[32]:=",
 CellID->1702767342,ExpressionUUID->"bc7e61e9-0384-4762-83c1-7f23e3b88429"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "1"}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{
  3.728658289756238*^9, 3.728659184981235*^9, 3.7286592315704517`*^9, 
   3.7286605772726393`*^9, 3.7286710638933783`*^9, 3.728674870586783*^9, 
   3.728732834320591*^9, 3.728733493254899*^9, 3.72874231110819*^9, 
   3.7287432896049843`*^9, 3.728743387712571*^9, 3.728747088127081*^9, 
   3.7287494452493896`*^9, 3.728749929626946*^9, 3.728750080352744*^9, 
   3.728750181997426*^9, 3.728750291964533*^9, 3.728750515236052*^9, 
   3.728750574710142*^9, 3.728750721099382*^9, {3.7287509796551733`*^9, 
   3.728751003807551*^9}, 3.7287516111154737`*^9, 3.7288132592387877`*^9, 
   3.728827376202216*^9, 3.7288430265226507`*^9, 3.728845618076868*^9, 
   3.733367242752063*^9, 3.790862628370533*^9, 3.790867063912924*^9, 
   3.7922731786366653`*^9},
 CellLabel->"Out[32]=",
 CellID->1505882069,ExpressionUUID->"c79d527e-e5d6-4fd2-821f-4965540672c0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Our example vector can also be represented in CTensor notation. The fact that \
these are components in the polar basis is represented by the red vertical \
bars:\
\>", "MathCaption",
 CellChangeTimes->{{3.728658451016865*^9, 3.7286584651032963`*^9}, {
  3.7286588851888103`*^9, 3.728658919258182*^9}, {3.728659587435247*^9, 
  3.7286595881634483`*^9}},
 CellID->242311782,ExpressionUUID->"20f59a56-cf63-49ad-ba67-2f2baef4bca1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"w", "=", 
  RowBox[{"CTensor", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"Sqrt", "[", "3", "]"}], ",", 
      RowBox[{"4", "/", "5"}], ",", 
      RowBox[{"-", "2"}]}], "}"}], ",", 
    RowBox[{"{", "fred", "}"}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.728657788410364*^9, 3.7286578097334127`*^9}, {
  3.728657853236046*^9, 3.728657867611246*^9}, {3.728658767619624*^9, 
  3.728658778624563*^9}, {3.72865896515481*^9, 3.728658966648027*^9}, {
  3.7286591886861887`*^9, 3.728659190381384*^9}, {3.728659234396261*^9, 
  3.728659235979086*^9}, {3.728733497566725*^9, 3.728733497943125*^9}, {
  3.728747091971465*^9, 3.72874709252317*^9}},
 CellLabel->"In[33]:=",
 CellID->1584502302,ExpressionUUID->"b64f669d-7de5-4b8d-b960-22ad52163c34"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     SqrtBox["3"], ",", 
     FractionBox["4", "5"], ",", 
     RowBox[{"-", "2"}]}], "}"}], ",", 
   RowBox[{"{", "fred", "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{{3.72865778932733*^9, 3.728657809966426*^9}, 
   3.728657868703639*^9, 3.728658207995817*^9, 3.728658466410768*^9, 
   3.728658779122549*^9, 3.728658829308704*^9, 3.7286589670606537`*^9, 
   3.728659190811661*^9, 3.728659236403413*^9, 3.728660577400886*^9, 
   3.728671066262951*^9, 3.728674870738194*^9, 3.728732835887046*^9, 
   3.7287335040387783`*^9, 3.728742311225581*^9, 3.728743289730118*^9, 
   3.7287433878340178`*^9, 3.728747093315234*^9, 3.728749446574032*^9, 
   3.728749929916246*^9, 3.728750080519561*^9, 3.728750182287417*^9, 
   3.728750292127923*^9, 3.728750515372609*^9, 3.728750574922373*^9, 
   3.728750721186805*^9, {3.728750979792142*^9, 3.728751003894363*^9}, 
   3.728751611263776*^9, 3.72881325938962*^9, 3.72882737634579*^9, 
   3.728843028217855*^9, 3.7288456182490807`*^9, 3.733367244151692*^9, 
   3.790862629602626*^9, 3.790867065228609*^9, 3.7922731796454077`*^9},
 CellLabel->"Out[33]=",
 CellID->183612468,ExpressionUUID->"94fe6a1d-0e46-4ec6-aa6c-9d9252c0296c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"w", "[", "a", "]"}]], "Input",
 CellChangeTimes->{{3.72865896869092*^9, 3.728658969159752*^9}},
 CellLabel->"In[34]:=",
 CellID->1820321917,ExpressionUUID->"6b084ed5-c180-42e2-9db7-0e32ed98bca7"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            SqrtBox["3"],
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", 
              SqrtBox["3"]}]],
           Annotation[#, {1} -> 3^Rational[1, 2], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            FractionBox["4", "5"],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              FractionBox["4", "5"]}]],
           Annotation[#, {2} -> Rational[4, 5], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{"-", "2"}],
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", 
              RowBox[{"-", "2"}]}]],
           Annotation[#, {3} -> -2, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{3^Rational[1, 2], 
    Rational[4, 5], -2}, {$CellContext`fred}, 0][$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.728658969584231*^9, 3.728659191882553*^9, 3.72865923728759*^9, 
   3.7286605777129517`*^9, 3.728671067263575*^9, 3.72867487115812*^9, 
   3.728732836853332*^9, 3.728733505504985*^9, 3.7287423113612013`*^9, 
   3.728743289868075*^9, 3.728743387965412*^9, 3.728747094730979*^9, 
   3.728749447383438*^9, 3.7287499300370092`*^9, 3.728750080687973*^9, 
   3.728750182647161*^9, 3.728750292670869*^9, 3.72875051550168*^9, 
   3.728750575089509*^9, 3.7287507213014174`*^9, {3.728750979941448*^9, 
   3.728751003984441*^9}, 3.72875161144671*^9, 3.72881325954033*^9, 
   3.7288273765217247`*^9, 3.728843028939458*^9, 3.728845618477373*^9, 
   3.733367244866397*^9, 3.790862630340108*^9, 3.790867066031636*^9, 
   3.792273182750264*^9},
 CellLabel->"Out[34]=",
 CellID->1733239118,ExpressionUUID->"47b89554-787b-481f-9f5b-0aac11630c57"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Convert to an explicit linear combination:", "MathCaption",
 CellChangeTimes->{{3.7286589338730593`*^9, 3.728658941745254*^9}},
 CellID->632631793,ExpressionUUID->"26b0cced-a85b-4400-ba9d-162cfaf7c110"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ToBasisExpand"}]], "Input",
 CellChangeTimes->{{3.728658822556219*^9, 3.728658825686892*^9}},
 CellLabel->"In[35]:=",
 CellID->1842212869,ExpressionUUID->"1104ab5c-dc5f-4a57-b1ff-8f6b8500797e"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   SqrtBox["3"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`fred}, $CellContext`a],
    Editable->False]}], "+", 
  FractionBox[
   RowBox[{"4", " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
     Editable->False]}], "5"], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`fred}, $CellContext`a],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{{3.728658826546609*^9, 3.7286588300070343`*^9}, 
   3.728658972421937*^9, 3.728659193901593*^9, 3.7286592384717417`*^9, 
   3.728660577990355*^9, 3.7286710683254347`*^9, 3.728674872158225*^9, 
   3.7287328381538467`*^9, 3.728733507170361*^9, 3.728742311493371*^9, 
   3.72874328998131*^9, 3.7287433880841618`*^9, 3.7287470971759033`*^9, 
   3.728749448667193*^9, 3.728749930155632*^9, 3.728750080839747*^9, 
   3.728750183011454*^9, 3.728750292879649*^9, 3.728750516352376*^9, 
   3.728750575322447*^9, 3.7287507214070377`*^9, {3.728750980111847*^9, 
   3.728751004086618*^9}, 3.7287516115835238`*^9, 3.728813259676899*^9, 
   3.728827376675194*^9, 3.7288430302727947`*^9, 3.728845618732238*^9, 
   3.7333672467856417`*^9, 3.790862631485632*^9, 3.79086707071535*^9, 
   3.792273183929689*^9},
 CellLabel->"Out[35]=",
 CellID->671775273,ExpressionUUID->"eb5a6790-4533-4e5c-9712-8d06838a7ae7"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "CTensor objects can be manipulated like tensor heads in ",
 StyleBox["xTensor`", "MR"],
 "."
}], "MathCaption",
 CellChangeTimes->{{3.7286589896785583`*^9, 3.728659037996649*^9}},
 CellID->174587974,ExpressionUUID->"a3116752-c77a-4c3e-a713-23368952637d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"xTensorQ", "[", "w", "]"}]], "Input",
 CellChangeTimes->{{3.728659058700526*^9, 3.728659061629056*^9}},
 CellLabel->"In[36]:=",
 CellID->777404065,ExpressionUUID->"efe69f1c-6189-4eee-a396-338e69fd47ae"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{
  3.728659062053235*^9, 3.728659196150271*^9, 3.728659239654976*^9, 
   3.728660578142432*^9, 3.728671069345858*^9, 3.728674872901696*^9, 
   3.728732839403862*^9, 3.728733510472055*^9, 3.728742311629909*^9, 
   3.7287432901777143`*^9, 3.7287433882337017`*^9, 3.728747098394784*^9, 
   3.728749449759087*^9, 3.7287499302673388`*^9, 3.728750081001747*^9, 
   3.7287501832368803`*^9, 3.728750293051141*^9, 3.7287505164725513`*^9, 
   3.7287505754389477`*^9, 3.728750721510827*^9, {3.7287509802557793`*^9, 
   3.7287510041780767`*^9}, 3.728751611951913*^9, 3.728813259821253*^9, 
   3.728827376820032*^9, 3.728843031461504*^9, 3.728845619010659*^9, 
   3.733367248069642*^9, 3.790862632719934*^9, 3.7908670718477983`*^9, 
   3.792273185251446*^9},
 CellLabel->"Out[36]=",
 CellID->1028234363,ExpressionUUID->"cf8ae64f-a76c-4633-ac77-a64d36eb2d63"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", "a", "]"}], 
  RowBox[{"w", "[", "b", "]"}]}]], "Input",
 CellChangeTimes->{{3.7286590735504*^9, 3.728659076330945*^9}},
 CellLabel->"In[37]:=",
 CellID->870075231,ExpressionUUID->"f7dd621f-2615-4f58-a013-ea6a4f177477"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["3",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "3"}]],
           Annotation[#, {1, 1} -> 3, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{"4", " ", 
              SqrtBox["3"]}], "5"],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{"4", " ", 
                 SqrtBox["3"]}], "5"]}]],
           
           Annotation[#, {1, 2} -> Rational[4, 5] 3^Rational[1, 2], 
            "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{"-", "2"}], " ", 
             SqrtBox["3"]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{"-", "2"}], " ", 
                SqrtBox["3"]}]}]],
           Annotation[#, {1, 3} -> (-2) 3^Rational[1, 2], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{"4", " ", 
              SqrtBox["3"]}], "5"],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{"4", " ", 
                 SqrtBox["3"]}], "5"]}]],
           
           Annotation[#, {2, 1} -> Rational[4, 5] 3^Rational[1, 2], 
            "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox["16", "25"],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              FractionBox["16", "25"]}]],
           Annotation[#, {2, 2} -> Rational[16, 25], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{"-", 
             FractionBox["8", "5"]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", 
              RowBox[{"-", 
                FractionBox["8", "5"]}]}]],
           Annotation[#, {2, 3} -> Rational[-8, 5], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{"-", "2"}], " ", 
             SqrtBox["3"]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{"-", "2"}], " ", 
                SqrtBox["3"]}]}]],
           Annotation[#, {3, 1} -> (-2) 3^Rational[1, 2], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{"-", 
             FractionBox["8", "5"]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{"-", 
                FractionBox["8", "5"]}]}]],
           Annotation[#, {3, 2} -> Rational[-8, 5], "Tooltip"]& ], 
          TagBox[
           TooltipBox["4",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", "4"}]],
           Annotation[#, {3, 3} -> 4, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 4 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 4 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a", "b"},
         {" ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{
    3, Rational[4, 5] 3^Rational[1, 2], (-2) 3^Rational[1, 2]}, {
    Rational[4, 5] 3^Rational[1, 2], 
     Rational[16, 25], 
     Rational[-8, 5]}, {(-2) 3^Rational[1, 2], 
     Rational[-8, 5], 4}}, {$CellContext`fred, $CellContext`fred}, 
   0][$CellContext`a, $CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.728659076824127*^9, 3.72865919674898*^9, 3.72865924025385*^9, 
   3.728660578427897*^9, 3.728671070008979*^9, 3.728674873945915*^9, 
   3.728732840186411*^9, 3.7287335115719748`*^9, 3.728742311984723*^9, 
   3.7287432905187798`*^9, 3.7287433883996153`*^9, 3.728747099060446*^9, 
   3.728749450366997*^9, 3.728749930383419*^9, 3.728750081171165*^9, 
   3.728750183394473*^9, 3.7287502934780483`*^9, 3.728750516618408*^9, 
   3.728750575562221*^9, 3.728750721601364*^9, {3.728750980409465*^9, 
   3.728751004263761*^9}, 3.7287516129276447`*^9, 3.728813259982071*^9, 
   3.7288273769731293`*^9, 3.7288430321735983`*^9, 3.728845619186076*^9, 
   3.733367248718882*^9, 3.790862633387103*^9, 3.790867072483173*^9, 
   3.7922731858141413`*^9},
 CellLabel->"Out[37]=",
 CellID->2121223330,ExpressionUUID->"33403d98-cb72-4542-b224-17b6450c97fb"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Basis vectors and transition matrices are represented by the two-index object \
Basis. Without a metric, only the down-up configuration is allowed:\
\>", "MathCaption",ExpressionUUID->"e0f55b9e-3dcd-4936-9ae4-6a3451845cfe"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Basis", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"4", ",", 
     RowBox[{"-", "fcomp"}]}], "}"}], ",", "A"}], "]"}]], "Input",
 CellChangeTimes->{{3.72873351551768*^9, 3.728733515717142*^9}, {
  3.728747101658777*^9, 3.7287471030267887`*^9}},
 CellLabel->"In[38]:=",ExpressionUUID->"d4c9e50d-7192-45ff-8f8d-7cc00c86e048"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"e", 
      StyleBox[GridBox[{
         {" ", "A"},
         {
          StyleBox["4",
           FontColor->RGBColor[0., 1., 0.]], " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`Basis[{4, -$CellContext`fcomp}, $CellContext`A],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.7286458898499613`*^9, 3.728658332478682*^9, 3.728659115660843*^9, 
   3.728659201854439*^9, 3.728659242338132*^9, 3.7286605788780327`*^9, 
   3.728671072326453*^9, 3.728674874430497*^9, 3.7287328416203613`*^9, 
   3.7287335221569433`*^9, 3.728742312161118*^9, 3.7287432906847*^9, 
   3.728743388533638*^9, 3.7287471037450037`*^9, 3.7287494516511793`*^9, 
   3.7287499305023947`*^9, 3.728750081336788*^9, 3.72875018364561*^9, 
   3.728750293628276*^9, 3.728750516750753*^9, 3.728750575721716*^9, 
   3.728750721696929*^9, {3.728750980635593*^9, 3.728751004343829*^9}, 
   3.728751613460923*^9, 3.728813260107193*^9, 3.7288273771046257`*^9, 
   3.7288430339393587`*^9, 3.728845619351062*^9, 3.733367250318191*^9, 
   3.7908626348435802`*^9, 3.790867074761552*^9, 3.7922731880128193`*^9},
 CellLabel->"Out[38]=",
 CellID->1277912894,ExpressionUUID->"00e4b102-90f5-4370-8a91-349e7401633d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Catch", "@", 
  RowBox[{"Basis", "[", 
   RowBox[{"A", ",", 
    RowBox[{"{", 
     RowBox[{"4", ",", 
      RowBox[{"-", "fcomp"}]}], "}"}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.728733524245178*^9, 3.728733524380929*^9}, {
  3.72874710664259*^9, 3.7287471077064543`*^9}},
 CellLabel->"In[39]:=",ExpressionUUID->"a78b18f2-7258-4c1a-bc84-9fb076646eaf"],

Cell[BoxData[
 TemplateBox[{
  "MetricsOfVBundle", "missing", 
   "\"There is no \\!\\(\\*RowBox[{\\\"\\\\\\\"metric\\\\\\\"\\\"}]\\) in \\!\
\\(\\*InterpretationBox[StyleBox[\\\"InnerC\\\", Rule[AutoSpacing, False], \
Rule[ShowAutoStyles, False]], InnerC, Rule[Editable, False]]\\).\"", 2, 39, 3,
    19668387891272167742, "Local", "xAct`xTensor`MetricsOfVBundle"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{
  3.728645891665702*^9, 3.728658333350918*^9, 3.7286591167546873`*^9, 
   3.7286592025318117`*^9, 3.728659243023828*^9, 3.728660579021009*^9, 
   3.728671072902687*^9, 3.72867487499222*^9, 3.728732842519462*^9, 
   3.728733524819257*^9, 3.728742312322235*^9, 3.728743290814465*^9, 
   3.7287433886846123`*^9, 3.728747108047368*^9, 3.7287494522659883`*^9, 
   3.728749930617119*^9, 3.7287500814763823`*^9, 3.728750183851701*^9, 
   3.728750293780211*^9, 3.728750516857692*^9, 3.728750575888358*^9, 
   3.7287507217758427`*^9, {3.728750980819462*^9, 3.728751004431987*^9}, 
   3.7287516136137037`*^9, 3.728813260238896*^9, 3.728827377252462*^9, 
   3.7288430346426487`*^9, 3.728845619510648*^9, 3.733367251034429*^9, 
   3.7908626355377607`*^9, 3.790867077664089*^9, 3.792273188715226*^9},
 CellLabel->"During evaluation of In[39]:=",
 CellID->407528486,ExpressionUUID->"549e158b-4fa1-4245-9e29-d7e06fd9582d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Basis", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "fred"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"b", ",", "blues"}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728733529133051*^9, 3.728733537759077*^9}, {
   3.728747109930544*^9, 3.7287471147063007`*^9}, 3.728749239245758*^9},
 CellLabel->"In[40]:=",ExpressionUUID->"574efbd0-14d4-469e-a113-a3a4d8886874"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"e", 
      StyleBox[GridBox[{
         {" ", 
          StyleBox["b",
           FontColor->RGBColor[0., 0., 1.]]},
         {
          StyleBox["a",
           FontColor->RGBColor[1., 0., 0.]], " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`Basis[{-$CellContext`a, -$CellContext`fred}, {$CellContext`b, \
$CellContext`blues}],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.728645894400773*^9, 3.7286591178395557`*^9, 3.728659203167418*^9, 
   3.728659243655389*^9, 3.728660579628969*^9, 3.728671073576277*^9, 
   3.728674875448032*^9, 3.728732843354403*^9, 3.728733538660166*^9, 
   3.728742312620281*^9, 3.728743290946706*^9, 3.728743389059758*^9, 
   3.7287471151789503`*^9, 3.728749452979878*^9, 3.7287499307038*^9, 
   3.728750081632181*^9, 3.7287501840027027`*^9, 3.728750294135548*^9, 
   3.728750516990486*^9, 3.72875057604432*^9, 3.728750721924171*^9, {
   3.728750980957382*^9, 3.728751004544126*^9}, 3.72875161379213*^9, 
   3.7288132603811502`*^9, 3.728827377392501*^9, 3.728843035340128*^9, 
   3.728845619711618*^9, 3.73336725310111*^9, 3.790862637971806*^9, 
   3.792273190467808*^9},
 CellLabel->"Out[40]=",
 CellID->1751994733,ExpressionUUID->"1b59040e-08cc-44cb-a994-4b7e08a720a6"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Hence, this is the vector w in a different basis:", "MathCaption",
 CellChangeTimes->{{3.728659247106689*^9, 3.728659268609988*^9}},
 CellID->1417706722,ExpressionUUID->"63abdba6-c302-4ed1-9c3c-74bd5ee4104a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCTensor", "[", 
  RowBox[{"w", ",", 
   RowBox[{"{", "blues", "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728659125467942*^9, 3.728659147079515*^9}, {
   3.728659205757403*^9, 3.7286592135086718`*^9}, {3.728733550758011*^9, 
   3.728733554157062*^9}, {3.728747119955625*^9, 3.728747120769897*^9}, 
   3.728749240270226*^9},
 CellLabel->"In[41]:=",
 CellID->1090855304,ExpressionUUID->"faa06ec4-e845-446a-9328-1060eea2ebad"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      RowBox[{
       SqrtBox["3"], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["0",
                 FontColor->RGBColor[0., 0., 1.]]},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[{0, -$CellContext`fred}, {0, $CellContext`blues}],
        Editable->False]}], "+", 
      FractionBox[
       RowBox[{"4", " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]]},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{1, -$CellContext`fred}, {0, $CellContext`blues}],
         
         Editable->False]}], "5"], "-", 
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["0",
                 FontColor->RGBColor[0., 0., 1.]]},
               {
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[{2, -$CellContext`fred}, {0, $CellContext`blues}],
        Editable->False]}]}], ",", 
     RowBox[{
      RowBox[{
       SqrtBox["3"], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]]},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[{0, -$CellContext`fred}, {1, $CellContext`blues}],
        Editable->False]}], "+", 
      FractionBox[
       RowBox[{"4", " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["1",
                  FontColor->RGBColor[0., 0., 1.]]},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{1, -$CellContext`fred}, {1, $CellContext`blues}],
         
         Editable->False]}], "5"], "-", 
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]]},
               {
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[{2, -$CellContext`fred}, {1, $CellContext`blues}],
        Editable->False]}]}], ",", 
     RowBox[{
      RowBox[{
       SqrtBox["3"], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[{0, -$CellContext`fred}, {2, $CellContext`blues}],
        Editable->False]}], "+", 
      FractionBox[
       RowBox[{"4", " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[0., 0., 1.]]},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{1, -$CellContext`fred}, {2, $CellContext`blues}],
         
         Editable->False]}], "5"], "-", 
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]},
               {
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[{2, -$CellContext`fred}, {2, $CellContext`blues}],
        Editable->False]}]}]}], "}"}], ",", 
   RowBox[{"{", "blues", "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{{3.728659137756296*^9, 3.728659147554789*^9}, {
   3.7286592037845783`*^9, 3.728659213998067*^9}, {3.728659270055814*^9, 
   3.728659272842313*^9}, 3.728660579878415*^9, 3.7286710749598703`*^9, 
   3.728674875925165*^9, 3.728732844550922*^9, 3.728733555159945*^9, 
   3.728742312898328*^9, 3.728743291085745*^9, 3.7287433894135313`*^9, 
   3.728747121380472*^9, 3.728749454110663*^9, 3.728749930884413*^9, 
   3.728750081816392*^9, 3.7287501844686813`*^9, 3.728750294547535*^9, 
   3.7287505171508408`*^9, 3.72875057620463*^9, 3.728750722035533*^9, {
   3.728750981069841*^9, 3.728751004650337*^9}, 3.728751614300568*^9, 
   3.728813260530862*^9, 3.728827377547081*^9, 3.728843036503633*^9, 
   3.7288456199006863`*^9, 3.733367254836533*^9, 3.79086264100975*^9, 
   3.790867080055656*^9, 3.792273194180238*^9},
 CellLabel->"Out[41]=",
 CellID->754733583,ExpressionUUID->"5f4fec61-8038-4013-841b-cd34297ac554"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Hover over the \"blobs\" to see the vector components. Note the new color of \
the vertical bars:\
\>", "MathCaption",
 CellChangeTimes->{{3.7286592836506166`*^9, 3.728659312760132*^9}},
 CellID->709935661,ExpressionUUID->"1c30ac6f-acd3-4454-b9d3-6a09f8f82b70"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "[", "a", "]"}]], "Input",
 CellChangeTimes->{{3.72865927628489*^9, 3.728659276961568*^9}},
 CellLabel->"In[42]:=",
 CellID->1422333728,ExpressionUUID->"15f54362-4420-4f40-8fc2-7d59c9ba8288"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  SqrtBox["3"], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"e", 
                    StyleBox[
                    
                    GridBox[{{
                    " ", "\!\(\*StyleBox[\"0\",FontColor->RGBColor[0., 0., \
1.]]\)"}, {"\!\(\*StyleBox[\"0\",FontColor->RGBColor[1., 0., 0.]]\)", " "}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   
                   xAct`xTensor`Basis[{0, -$CellContext`fred}, {
                    0, $CellContext`blues}], Editable -> False]}], "+", 
                FractionBox[
                 RowBox[{"4", " ", 
                   InterpretationBox[
                    StyleBox[
                    GridBox[{{"e", 
                    StyleBox[
                    
                    GridBox[{{
                    " ", "\!\(\*StyleBox[\"0\",FontColor->RGBColor[0., 0., \
1.]]\)"}, {"\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0., 0.]]\)", " "}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    xAct`xTensor`Basis[{1, -$CellContext`fred}, {
                    0, $CellContext`blues}], Editable -> False]}], "5"], "-", 
                
                RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"e", 
                    StyleBox[
                    
                    GridBox[{{
                    " ", "\!\(\*StyleBox[\"0\",FontColor->RGBColor[0., 0., \
1.]]\)"}, {"\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0., 0.]]\)", " "}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   
                   xAct`xTensor`Basis[{2, -$CellContext`fred}, {
                    0, $CellContext`blues}], Editable -> False]}]}]}]],
           
           Annotation[#, {1} -> 
            3^Rational[1, 2] 
              xAct`xTensor`Basis[{0, -$CellContext`fred}, {
                0, $CellContext`blues}] + 
             Rational[4, 5] 
              xAct`xTensor`Basis[{1, -$CellContext`fred}, {
                0, $CellContext`blues}] - 2 
             xAct`xTensor`Basis[{2, -$CellContext`fred}, {
               0, $CellContext`blues}], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  SqrtBox["3"], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"e", 
                    StyleBox[
                    
                    GridBox[{{
                    " ", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[0., 0., \
1.]]\)"}, {"\!\(\*StyleBox[\"0\",FontColor->RGBColor[1., 0., 0.]]\)", " "}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   
                   xAct`xTensor`Basis[{0, -$CellContext`fred}, {
                    1, $CellContext`blues}], Editable -> False]}], "+", 
                FractionBox[
                 RowBox[{"4", " ", 
                   InterpretationBox[
                    StyleBox[
                    GridBox[{{"e", 
                    StyleBox[
                    
                    GridBox[{{
                    " ", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[0., 0., \
1.]]\)"}, {"\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0., 0.]]\)", " "}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    xAct`xTensor`Basis[{1, -$CellContext`fred}, {
                    1, $CellContext`blues}], Editable -> False]}], "5"], "-", 
                
                RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"e", 
                    StyleBox[
                    
                    GridBox[{{
                    " ", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[0., 0., \
1.]]\)"}, {"\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0., 0.]]\)", " "}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   
                   xAct`xTensor`Basis[{2, -$CellContext`fred}, {
                    1, $CellContext`blues}], Editable -> False]}]}]}]],
           
           Annotation[#, {2} -> 
            3^Rational[1, 2] 
              xAct`xTensor`Basis[{0, -$CellContext`fred}, {
                1, $CellContext`blues}] + 
             Rational[4, 5] 
              xAct`xTensor`Basis[{1, -$CellContext`fred}, {
                1, $CellContext`blues}] - 2 
             xAct`xTensor`Basis[{2, -$CellContext`fred}, {
               1, $CellContext`blues}], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  SqrtBox["3"], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"e", 
                    StyleBox[
                    
                    GridBox[{{
                    " ", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[0., 0., \
1.]]\)"}, {"\!\(\*StyleBox[\"0\",FontColor->RGBColor[1., 0., 0.]]\)", " "}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   
                   xAct`xTensor`Basis[{0, -$CellContext`fred}, {
                    2, $CellContext`blues}], Editable -> False]}], "+", 
                FractionBox[
                 RowBox[{"4", " ", 
                   InterpretationBox[
                    StyleBox[
                    GridBox[{{"e", 
                    StyleBox[
                    
                    GridBox[{{
                    " ", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[0., 0., \
1.]]\)"}, {"\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0., 0.]]\)", " "}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    xAct`xTensor`Basis[{1, -$CellContext`fred}, {
                    2, $CellContext`blues}], Editable -> False]}], "5"], "-", 
                
                RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"e", 
                    StyleBox[
                    
                    GridBox[{{
                    " ", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[0., 0., \
1.]]\)"}, {"\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0., 0.]]\)", " "}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   
                   xAct`xTensor`Basis[{2, -$CellContext`fred}, {
                    2, $CellContext`blues}], Editable -> False]}]}]}]],
           
           Annotation[#, {3} -> 
            3^Rational[1, 2] 
              xAct`xTensor`Basis[{0, -$CellContext`fred}, {
                2, $CellContext`blues}] + 
             Rational[4, 5] 
              xAct`xTensor`Basis[{1, -$CellContext`fred}, {
                2, $CellContext`blues}] - 2 
             xAct`xTensor`Basis[{2, -$CellContext`fred}, {
               2, $CellContext`blues}], "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[0, 0, 1], 2 -> RGBColor[0, 0, 1]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{
   3^Rational[1, 2] 
     xAct`xTensor`Basis[{0, -$CellContext`fred}, {0, $CellContext`blues}] + 
    Rational[4, 5] 
     xAct`xTensor`Basis[{1, -$CellContext`fred}, {0, $CellContext`blues}] - 2 
    xAct`xTensor`Basis[{2, -$CellContext`fred}, {0, $CellContext`blues}], 
    3^Rational[1, 2] 
     xAct`xTensor`Basis[{0, -$CellContext`fred}, {1, $CellContext`blues}] + 
    Rational[4, 5] 
     xAct`xTensor`Basis[{1, -$CellContext`fred}, {1, $CellContext`blues}] - 2 
    xAct`xTensor`Basis[{2, -$CellContext`fred}, {1, $CellContext`blues}], 
    3^Rational[1, 2] 
     xAct`xTensor`Basis[{0, -$CellContext`fred}, {2, $CellContext`blues}] + 
    Rational[4, 5] 
     xAct`xTensor`Basis[{1, -$CellContext`fred}, {2, $CellContext`blues}] - 2 
    xAct`xTensor`Basis[{2, -$CellContext`fred}, {
      2, $CellContext`blues}]}, {$CellContext`blues}, 0][$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.7286592773614264`*^9, 3.728660580063014*^9, 3.728671076309702*^9, 
   3.728674876761455*^9, 3.728732845738203*^9, 3.728733556792618*^9, 
   3.728742313419415*^9, 3.728743291384695*^9, 3.728743389854595*^9, 
   3.728747125664263*^9, 3.7287494556864033`*^9, 3.72874993103216*^9, 
   3.7287500819943523`*^9, 3.728750185117831*^9, 3.7287502952206182`*^9, 
   3.728750517338655*^9, 3.7287505763871727`*^9, 3.728750722127281*^9, {
   3.7287509812516603`*^9, 3.728751004757245*^9}, 3.728751614818686*^9, 
   3.728813260682721*^9, 3.728827377700304*^9, 3.728843037787189*^9, 
   3.728845620080158*^9, 3.7333672574684*^9, 3.790862642464549*^9, 
   3.790867081152453*^9, 3.7922731954534483`*^9},
 CellLabel->"Out[42]=",
 CellID->635845039,ExpressionUUID->"a8ab833a-16d1-4f97-a83a-ae7075a4d5e7"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
ContractBasis is quite general and allows the user to specify which indices \
should be acted upon. The default is total contraction:\
\>", "MathCaption",
 CellChangeTimes->{{3.7286593507920218`*^9, 
  3.728659357125814*^9}},ExpressionUUID->"732e82b6-b346-4d4a-8cb7-\
4804076352ee"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr", "=", 
  RowBox[{
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "fred"}]}], "}"}], ",", "a"}], "]"}], 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"-", "c"}], ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", "blues"}], "}"}]}], "]"}], 
   RowBox[{"v", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", "c", ",", 
     RowBox[{"{", 
      RowBox[{"d", ",", "fred"}], "}"}]}], "]"}], 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "d"}], ",", 
       RowBox[{"-", "fred"}]}], "}"}], ",", "e"}], "]"}]}]}]], "Input",
 CellChangeTimes->{{3.728733562883897*^9, 3.7287336096098824`*^9}, {
   3.728747128939732*^9, 3.728747161496373*^9}, 3.7287492445139303`*^9},
 CellLabel->"In[43]:=",ExpressionUUID->"644c1d17-43de-4b30-933f-75681cb99f09"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["1",
            FontColor->RGBColor[0., 0., 1.]]},
          {"c", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`blues}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", "a"},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", "e"},
          {
           StyleBox["d",
            FontColor->RGBColor[1., 0., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{-$CellContext`d, -$CellContext`fred}, $CellContext`e],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" ", " ", "c", 
           StyleBox["d",
            FontColor->RGBColor[1., 0., 0.]]},
          {"a", "b", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`a, -$CellContext`b, $CellContext`c, \
{$CellContext`d, $CellContext`fred}],
   Editable->False]}]], "Output",
 CellChangeTimes->{
  3.728645901215798*^9, 3.7286593590402937`*^9, 3.72866058025765*^9, 
   3.728671077575845*^9, 3.7286748770658703`*^9, 3.728732852304935*^9, 
   3.728733610232348*^9, 3.728742313561075*^9, 3.7287432915512114`*^9, 
   3.7287433899992857`*^9, 3.7287471619578857`*^9, 3.7287494567594147`*^9, 
   3.72874993115726*^9, 3.7287500821826067`*^9, 3.728750185319347*^9, 
   3.7287502954016447`*^9, 3.728750517536615*^9, 3.728750576529258*^9, 
   3.728750722221685*^9, {3.728750981418046*^9, 3.728751004859355*^9}, 
   3.7287516149816217`*^9, 3.728813260826335*^9, 3.728827377857832*^9, 
   3.72884303927351*^9, 3.7288456202500963`*^9, 3.733367258953932*^9, 
   3.790862643662071*^9, 3.790867091450263*^9, 3.792273196632375*^9},
 CellLabel->"Out[43]=",
 CellID->1421076894,ExpressionUUID->"65cd8321-03db-4867-aad4-cadea8023d4a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ContractBasis", "[", "expr", "]"}]], "Input",
 CellLabel->"In[44]:=",ExpressionUUID->"4023367d-aa79-4d9f-abeb-f2900830add1"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"v", 
      StyleBox[GridBox[{
         {" ", " ", 
          StyleBox["1",
           FontColor->RGBColor[0., 0., 1.]], "e"},
         {
          StyleBox["1",
           FontColor->RGBColor[1., 0., 0.]], "b", " ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`v[{1, -$CellContext`fred}, -$CellContext`b, {
   1, $CellContext`blues}, $CellContext`e],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.728645902037877*^9, 3.728659359857806*^9, 3.728660580393265*^9, 
   3.7286710781947403`*^9, 3.728674877218143*^9, 3.728732853288909*^9, 
   3.728733611215819*^9, 3.7287423139089403`*^9, 3.728743291683559*^9, 
   3.728743390132901*^9, 3.728747162846347*^9, 3.72874945740127*^9, 
   3.728749931271481*^9, 3.728750082352669*^9, 3.728750185657567*^9, 
   3.728750295551732*^9, 3.7287505176720333`*^9, 3.728750576713509*^9, 
   3.728750722317079*^9, {3.728750981540331*^9, 3.728751004949976*^9}, 
   3.72875161567697*^9, 3.728813260998289*^9, 3.728827377988513*^9, 
   3.728843039990844*^9, 3.728845620410428*^9, 3.733367259653534*^9, 
   3.790862644290967*^9, 3.790867092265725*^9, 3.7922731973160343`*^9},
 CellLabel->"Out[44]=",
 CellID->2075856550,ExpressionUUID->"4dccd27d-c450-4d22-8572-f86837b10363"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ContractBasis", "[", 
  RowBox[{"expr", ",", "fred"}], "]"}]], "Input",
 CellChangeTimes->{{3.728733614481164*^9, 3.728733614801462*^9}, {
  3.7287471669907722`*^9, 3.728747167487653*^9}},
 CellLabel->"In[45]:=",ExpressionUUID->"b3f1b95a-f13a-49eb-83bd-261e42316d5b"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["1",
            FontColor->RGBColor[0., 0., 1.]]},
          {"c", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`blues}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", "a"},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" ", " ", "c", "e"},
          {"a", "b", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`a, -$CellContext`b, $CellContext`c, \
$CellContext`e],
   Editable->False]}]], "Output",
 CellChangeTimes->{
  3.7286459028012047`*^9, 3.7286593605550632`*^9, 3.728660580544837*^9, 
   3.7286710788592587`*^9, 3.728674877400217*^9, 3.7287328541561117`*^9, 
   3.72873361561948*^9, 3.728742314178133*^9, 3.7287432918330936`*^9, 
   3.728743390300458*^9, 3.728747167939932*^9, 3.728749458017366*^9, 
   3.728749931366468*^9, 3.728750082521471*^9, 3.728750186070795*^9, 
   3.728750295731461*^9, 3.728750517853341*^9, 3.728750576878702*^9, 
   3.728750722426601*^9, {3.728750981675538*^9, 3.72875100506754*^9}, 
   3.7287516166572638`*^9, 3.728813261140973*^9, 3.728827378141533*^9, 
   3.728843040674385*^9, 3.7288456205910387`*^9, 3.733367260352812*^9, 
   3.7908626448588543`*^9, 3.7908670929054747`*^9, 3.792273197886033*^9},
 CellLabel->"Out[45]=",
 CellID->875006227,ExpressionUUID->"91944c9a-d826-4f99-a471-80f305b052aa"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ContractBasis", "[", 
  RowBox[{"expr", ",", "a"}], "]"}]], "Input",
 CellLabel->"In[46]:=",ExpressionUUID->"185b485b-f451-4e4c-b55c-c113a375f0c1"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["1",
            FontColor->RGBColor[0., 0., 1.]]},
          {"c", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`blues}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", "e"},
          {
           StyleBox["d",
            FontColor->RGBColor[1., 0., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{-$CellContext`d, -$CellContext`fred}, $CellContext`e],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" ", " ", "c", 
           StyleBox["d",
            FontColor->RGBColor[1., 0., 0.]]},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], "b", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{
    1, -$CellContext`fred}, -$CellContext`b, $CellContext`c, {$CellContext`d, \
$CellContext`fred}],
   Editable->False]}]], "Output",
 CellChangeTimes->{
  3.728645903539863*^9, 3.728659361175002*^9, 3.7286605807039557`*^9, 
   3.728671079493909*^9, 3.728674877547068*^9, 3.7287328549888487`*^9, 
   3.728733617135695*^9, 3.728742314674972*^9, 3.7287432920024557`*^9, 
   3.728743390434216*^9, 3.728747169719655*^9, 3.728749458584783*^9, 
   3.728749931518736*^9, 3.7287500827750263`*^9, 3.7287501866217947`*^9, 
   3.72875029648604*^9, 3.7287505180042343`*^9, 3.728750577026417*^9, 
   3.728750722528451*^9, {3.728750981809354*^9, 3.728751005178564*^9}, 
   3.728751616842252*^9, 3.728813261293439*^9, 3.728827378291399*^9, 
   3.728843041373919*^9, 3.72884562072771*^9, 3.733367261038096*^9, 
   3.790862645954204*^9, 3.790867093488693*^9, 3.792273198406361*^9},
 CellLabel->"Out[46]=",
 CellID->595358560,ExpressionUUID->"32da72cf-b3aa-4c46-8928-aa1bb46437f1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The inverse function is SeparateBasis, also quite general. It can undo the \
action of ContractBasis and also change the basis of the whole expression\
\>", "MathCaption",ExpressionUUID->"f191f0cd-83f9-4208-b040-9a8523661f0f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"cexpr", "=", 
  RowBox[{"ContractBasis", "[", "expr", "]"}]}]], "Input",
 CellLabel->"In[47]:=",ExpressionUUID->"9562eb05-9c3e-4c35-a176-2bcd9d36f40c"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"v", 
      StyleBox[GridBox[{
         {" ", " ", 
          StyleBox["1",
           FontColor->RGBColor[0., 0., 1.]], "e"},
         {
          StyleBox["1",
           FontColor->RGBColor[1., 0., 0.]], "b", " ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`v[{1, -$CellContext`fred}, -$CellContext`b, {
   1, $CellContext`blues}, $CellContext`e],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.728645905035679*^9, 3.728659365325197*^9, 3.728660581008718*^9, 
   3.7286710807604628`*^9, 3.728674877710493*^9, 3.728732856421146*^9, 
   3.728733620051135*^9, 3.728742314804782*^9, 3.7287432921466208`*^9, 
   3.728743390566123*^9, 3.728747173773511*^9, 3.7287494597346563`*^9, 
   3.728749931637566*^9, 3.728750082916168*^9, 3.728750186807109*^9, 
   3.728750296636909*^9, 3.728750518139744*^9, 3.7287505771768637`*^9, 
   3.728750722627792*^9, {3.728750981925397*^9, 3.728751005267762*^9}, 
   3.728751617038905*^9, 3.728813261439868*^9, 3.7288273784386663`*^9, 
   3.728843042791192*^9, 3.728845620901767*^9, 3.733367262437214*^9, 
   3.790862647499134*^9, 3.7908670952850523`*^9, 3.7922731998988733`*^9},
 CellLabel->"Out[47]=",
 CellID->2130434817,ExpressionUUID->"f9e38686-cad3-4fef-9d05-3e9dcb22de92"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SeparateBasis", "[", "AIndex", "]"}], "[", "cexpr", "]"}]], "Input",\

 CellLabel->"In[48]:=",ExpressionUUID->"c0e15c0e-7246-4677-83f1-5e4102bfd863"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["1",
            FontColor->RGBColor[0., 0., 1.]]},
          {"c", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`blues}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", "a"},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" ", " ", "c", "e"},
          {"a", "b", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`a, -$CellContext`b, $CellContext`c, \
$CellContext`e],
   Editable->False]}]], "Output",
 CellChangeTimes->{
  3.728645906085597*^9, 3.7286593660723057`*^9, 3.728660581446041*^9, 
   3.728671081394004*^9, 3.728674877881874*^9, 3.728732858138887*^9, 
   3.728733621401819*^9, 3.7287423154035683`*^9, 3.728743292348319*^9, 
   3.728743390682749*^9, 3.728747176957244*^9, 3.72874946033469*^9, 
   3.72874993175078*^9, 3.728750083099104*^9, 3.728750187284821*^9, 
   3.728750296817696*^9, 3.728750518284646*^9, 3.728750577339946*^9, 
   3.7287507227682056`*^9, {3.728750982036283*^9, 3.728751005347109*^9}, 
   3.728751617239727*^9, 3.728813261592263*^9, 3.728827378606987*^9, 
   3.728843043507699*^9, 3.728845621058146*^9, 3.733367263036324*^9, 
   3.790862648075068*^9, 3.790867096721552*^9, 3.792273200397128*^9},
 CellLabel->"Out[48]=",
 CellID->2018184273,ExpressionUUID->"f0fed273-4c5d-46ef-b7b7-031066f8ea49"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SeparateBasis", "[", "blues", "]"}], "[", "cexpr", "]"}]], "Input",\

 CellChangeTimes->{{3.7287336243046627`*^9, 3.7287336257550917`*^9}, {
   3.728747179993021*^9, 3.7287471806872*^9}, 3.728749246765738*^9},
 CellLabel->"In[49]:=",ExpressionUUID->"5f27a659-edaa-4817-9ca6-f92b5b99c917"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["c",
            FontColor->RGBColor[0., 0., 1.]]},
          {"b", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[-$CellContext`b, {$CellContext`c, $CellContext`blues}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["a",
            FontColor->RGBColor[0., 0., 1.]]},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{
    1, -$CellContext`fred}, {$CellContext`a, $CellContext`blues}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["1",
            FontColor->RGBColor[0., 0., 1.]]},
          {
           StyleBox["d",
            FontColor->RGBColor[0., 0., 1.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{-$CellContext`d, -$CellContext`blues}, {
    1, $CellContext`blues}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", "e"},
          {
           StyleBox["f",
            FontColor->RGBColor[0., 0., 1.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{-$CellContext`f, -$CellContext`blues}, $CellContext`e],
   
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" ", " ", 
           StyleBox["d",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox["f",
            FontColor->RGBColor[0., 0., 1.]]},
          {
           StyleBox["a",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox["c",
            FontColor->RGBColor[0., 0., 1.]], " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{-$CellContext`a, -$CellContext`blues}, {-$CellContext`c, \
-$CellContext`blues}, {$CellContext`d, $CellContext`blues}, {$CellContext`f, \
$CellContext`blues}],
   Editable->False]}]], "Output",
 CellChangeTimes->{
  3.7286459069592752`*^9, 3.728659366756983*^9, 3.728660581639659*^9, 
   3.728671082025057*^9, 3.728674878066738*^9, 3.728732859606332*^9, 
   3.728733626418371*^9, 3.728742315869981*^9, 3.7287432924850683`*^9, 
   3.728743390850091*^9, 3.728747181188478*^9, 3.728749460947112*^9, 
   3.728749931868095*^9, 3.7287500832581377`*^9, 3.728750187569511*^9, 
   3.728750297029105*^9, 3.728750518437439*^9, 3.7287505775005627`*^9, 
   3.728750722865849*^9, {3.728750982212698*^9, 3.728751005451541*^9}, 
   3.728751617396041*^9, 3.728813261744954*^9, 3.728827378758469*^9, 
   3.728843044291744*^9, 3.7288456212939568`*^9, 3.733367263834042*^9, 
   3.790862648622868*^9, 3.790867097980218*^9, 3.792273202672491*^9},
 CellLabel->"Out[49]=",
 CellID->2072055645,ExpressionUUID->"26da2d4c-7e0a-4ead-9100-485ae7d95a59"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Another important features is the ability to trace basis contractions,\
\>", "MathCaption",
 CellChangeTimes->{{3.728659376404819*^9, 
  3.728659381884268*^9}},ExpressionUUID->"600bb412-90c2-4297-af29-\
984398da8d42"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TraceBasisDummy", "[", 
  RowBox[{
   RowBox[{"v", "[", 
    RowBox[{"{", 
     RowBox[{"a", ",", "blues"}], "}"}], "]"}], 
   RowBox[{"T", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "blues"}]}], "}"}], ",", 
     RowBox[{"-", "b"}]}], "]"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728733632337534*^9, 3.728733634383157*^9}, {
  3.728747184464964*^9, 3.728747190454898*^9}, {3.728749249641384*^9, 
  3.728749251005344*^9}},
 CellLabel->"In[50]:=",ExpressionUUID->"2b19ef62-7e0e-4584-ac0f-0a8cd5e21383"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]], "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{0, -$CellContext`blues}, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[{0, $CellContext`blues}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{1, -$CellContext`blues}, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[{1, $CellContext`blues}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{2, -$CellContext`blues}, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[{2, $CellContext`blues}],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{
  3.7286459096361647`*^9, 3.728659383525722*^9, 3.728660581941255*^9, 
   3.7286710837773314`*^9, 3.728674878397744*^9, 3.728732864957357*^9, 
   3.728733635137611*^9, 3.728742315996902*^9, 3.728743292717317*^9, 
   3.728743390966918*^9, 3.728747191171215*^9, 3.728749462223446*^9, 
   3.728749931984376*^9, 3.728750083433025*^9, 3.728750187757642*^9, 
   3.728750297615917*^9, 3.7287505185502977`*^9, 3.7287505776903143`*^9, 
   3.728750722966165*^9, {3.728750982391608*^9, 3.728751005563882*^9}, 
   3.728751617613662*^9, 3.728813261894397*^9, 3.728827378911272*^9, 
   3.7288430460244493`*^9, 3.728845621475374*^9, 3.7333672650225697`*^9, 
   3.790862650020854*^9, 3.790867105302608*^9, 3.792273204857037*^9},
 CellLabel->"Out[50]=",
 CellID->454205248,ExpressionUUID->"6b47f00a-c810-4918-b610-c9100445e6a5"]
}, Open  ]]
}, Open  ]],

Cell["\<\
Finally, we will have a quick look at the third method to store components, \
which is useful for partially-abstract computations, or cases with a high \
degree of symmetry.\
\>", "Text",
 CellChangeTimes->{{3.7286596164167624`*^9, 3.72865967513535*^9}},
 CellID->1541888545,ExpressionUUID->"afbccaae-ae0f-4f20-8cca-5c862826a72a"],

Cell[CellGroupData[{

Cell["\<\
We can assign rules to the components of a tensor taking the symmetries into \
account\
\>", "MathCaption",ExpressionUUID->"deb17fcf-f095-41f1-a81c-cd04b791fe2d"],

Cell[BoxData[
 RowBox[{
  RowBox[{"values", "=", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"i", "-", "j"}], ",", 
     RowBox[{"{", 
      RowBox[{"i", ",", "3"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"j", ",", "3"}], "}"}]}], "]"}]}], ";"}]], "Input",
 CellLabel->"In[51]:=",ExpressionUUID->"33350187-25c2-431b-b909-dc8b7a9fb8a4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AllComponentValues", "[", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "fred"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "fred"}]}], "}"}]}], "]"}], ",", " ", "values"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.728733642834337*^9, 3.7287336488737307`*^9}, {
  3.7287471964484987`*^9, 3.728747199238364*^9}},
 CellLabel->"In[52]:=",ExpressionUUID->"1b4d9a3d-c4f5-483c-be9a-93895f145913"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`fred}, {0, -$CellContext`fred}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`T[{0, -$CellContext`fred}, {
     0, -$CellContext`fred}] -> 0, " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.7286459173172073`*^9, 3.728659686300744*^9, 3.728660582228342*^9, 
   3.728671086075186*^9, 3.7286748786260967`*^9, 3.7287328676708927`*^9, 
   3.728733649639093*^9, 3.728742316161644*^9, 3.728743292919387*^9, 
   3.728743391180121*^9, 3.728747199672443*^9, 3.7287494655488586`*^9, 
   3.7287499321812153`*^9, 3.728750083671248*^9, 3.728750188389699*^9, 
   3.728750298129628*^9, 3.728750518820197*^9, 3.7287505779621572`*^9, 
   3.728750723090013*^9, {3.7287509825734453`*^9, 3.728751005661495*^9}, 
   3.728751617832035*^9, 3.728813262155266*^9, 3.728827379123863*^9, 
   3.7288430492376633`*^9, 3.728845621854383*^9, 3.733367267635683*^9, 
   3.79086265235045*^9, 3.790867107345881*^9, 3.7922732069714518`*^9},
 CellLabel->"During evaluation of In[52]:=",
 CellID->944573012,ExpressionUUID->"f40e06a7-3aa6-4c4b-98df-2e976f11afc6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`fred}, {1, -$CellContext`fred}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", "1"}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added independent rule ", $CellContext`T[{0, -$CellContext`fred}, {
     1, -$CellContext`fred}] -> -1, " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.7286459173172073`*^9, 3.728659686300744*^9, 3.728660582228342*^9, 
   3.728671086075186*^9, 3.7286748786260967`*^9, 3.7287328676708927`*^9, 
   3.728733649639093*^9, 3.728742316161644*^9, 3.728743292919387*^9, 
   3.728743391180121*^9, 3.728747199672443*^9, 3.7287494655488586`*^9, 
   3.7287499321812153`*^9, 3.728750083671248*^9, 3.728750188389699*^9, 
   3.728750298129628*^9, 3.728750518820197*^9, 3.7287505779621572`*^9, 
   3.728750723090013*^9, {3.7287509825734453`*^9, 3.728751005661495*^9}, 
   3.728751617832035*^9, 3.728813262155266*^9, 3.728827379123863*^9, 
   3.7288430492376633`*^9, 3.728845621854383*^9, 3.733367267635683*^9, 
   3.79086265235045*^9, 3.790867107345881*^9, 3.7922732070026007`*^9},
 CellLabel->"During evaluation of In[52]:=",
 CellID->1239494060,ExpressionUUID->"0ded7113-e553-4f40-9e8f-b69fc788e421"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`fred}, {2, -$CellContext`fred}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", "2"}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added independent rule ", $CellContext`T[{0, -$CellContext`fred}, {
     2, -$CellContext`fred}] -> -2, " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.7286459173172073`*^9, 3.728659686300744*^9, 3.728660582228342*^9, 
   3.728671086075186*^9, 3.7286748786260967`*^9, 3.7287328676708927`*^9, 
   3.728733649639093*^9, 3.728742316161644*^9, 3.728743292919387*^9, 
   3.728743391180121*^9, 3.728747199672443*^9, 3.7287494655488586`*^9, 
   3.7287499321812153`*^9, 3.728750083671248*^9, 3.728750188389699*^9, 
   3.728750298129628*^9, 3.728750518820197*^9, 3.7287505779621572`*^9, 
   3.728750723090013*^9, {3.7287509825734453`*^9, 3.728751005661495*^9}, 
   3.728751617832035*^9, 3.728813262155266*^9, 3.728827379123863*^9, 
   3.7288430492376633`*^9, 3.728845621854383*^9, 3.733367267635683*^9, 
   3.79086265235045*^9, 3.790867107345881*^9, 3.7922732070222397`*^9},
 CellLabel->"During evaluation of In[52]:=",
 CellID->1585290520,ExpressionUUID->"b375287e-5673-4831-892e-0cdd341ceb22"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`fred}, {0, -$CellContext`fred}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`fred}, {1, -$CellContext`fred}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`T[{1, -$CellContext`fred}, {
     0, -$CellContext`fred}] -> -$CellContext`T[{0, -$CellContext`fred}, {
      1, -$CellContext`fred}], " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.7286459173172073`*^9, 3.728659686300744*^9, 3.728660582228342*^9, 
   3.728671086075186*^9, 3.7286748786260967`*^9, 3.7287328676708927`*^9, 
   3.728733649639093*^9, 3.728742316161644*^9, 3.728743292919387*^9, 
   3.728743391180121*^9, 3.728747199672443*^9, 3.7287494655488586`*^9, 
   3.7287499321812153`*^9, 3.728750083671248*^9, 3.728750188389699*^9, 
   3.728750298129628*^9, 3.728750518820197*^9, 3.7287505779621572`*^9, 
   3.728750723090013*^9, {3.7287509825734453`*^9, 3.728751005661495*^9}, 
   3.728751617832035*^9, 3.728813262155266*^9, 3.728827379123863*^9, 
   3.7288430492376633`*^9, 3.728845621854383*^9, 3.733367267635683*^9, 
   3.79086265235045*^9, 3.790867107345881*^9, 3.792273207057208*^9},
 CellLabel->"During evaluation of In[52]:=",
 CellID->148161311,ExpressionUUID->"117862d2-3d6e-41f6-8750-2c856039f956"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`fred}, {1, -$CellContext`fred}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`T[{1, -$CellContext`fred}, {
     1, -$CellContext`fred}] -> 0, " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.7286459173172073`*^9, 3.728659686300744*^9, 3.728660582228342*^9, 
   3.728671086075186*^9, 3.7286748786260967`*^9, 3.7287328676708927`*^9, 
   3.728733649639093*^9, 3.728742316161644*^9, 3.728743292919387*^9, 
   3.728743391180121*^9, 3.728747199672443*^9, 3.7287494655488586`*^9, 
   3.7287499321812153`*^9, 3.728750083671248*^9, 3.728750188389699*^9, 
   3.728750298129628*^9, 3.728750518820197*^9, 3.7287505779621572`*^9, 
   3.728750723090013*^9, {3.7287509825734453`*^9, 3.728751005661495*^9}, 
   3.728751617832035*^9, 3.728813262155266*^9, 3.728827379123863*^9, 
   3.7288430492376633`*^9, 3.728845621854383*^9, 3.733367267635683*^9, 
   3.79086265235045*^9, 3.790867107345881*^9, 3.792273207089044*^9},
 CellLabel->"During evaluation of In[52]:=",
 CellID->1213607104,ExpressionUUID->"61a662ad-2b6c-403a-852d-0fc368e7de74"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`fred}, {2, -$CellContext`fred}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", "1"}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added independent rule ", $CellContext`T[{1, -$CellContext`fred}, {
     2, -$CellContext`fred}] -> -1, " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.7286459173172073`*^9, 3.728659686300744*^9, 3.728660582228342*^9, 
   3.728671086075186*^9, 3.7286748786260967`*^9, 3.7287328676708927`*^9, 
   3.728733649639093*^9, 3.728742316161644*^9, 3.728743292919387*^9, 
   3.728743391180121*^9, 3.728747199672443*^9, 3.7287494655488586`*^9, 
   3.7287499321812153`*^9, 3.728750083671248*^9, 3.728750188389699*^9, 
   3.728750298129628*^9, 3.728750518820197*^9, 3.7287505779621572`*^9, 
   3.728750723090013*^9, {3.7287509825734453`*^9, 3.728751005661495*^9}, 
   3.728751617832035*^9, 3.728813262155266*^9, 3.728827379123863*^9, 
   3.7288430492376633`*^9, 3.728845621854383*^9, 3.733367267635683*^9, 
   3.79086265235045*^9, 3.790867107345881*^9, 3.792273207098261*^9},
 CellLabel->"During evaluation of In[52]:=",
 CellID->294917722,ExpressionUUID->"94dccdfb-5140-42e6-8dad-f0c21f3083d8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, -$CellContext`fred}, {0, -$CellContext`fred}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`fred}, {2, -$CellContext`fred}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`T[{2, -$CellContext`fred}, {
     0, -$CellContext`fred}] -> -$CellContext`T[{0, -$CellContext`fred}, {
      2, -$CellContext`fred}], " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.7286459173172073`*^9, 3.728659686300744*^9, 3.728660582228342*^9, 
   3.728671086075186*^9, 3.7286748786260967`*^9, 3.7287328676708927`*^9, 
   3.728733649639093*^9, 3.728742316161644*^9, 3.728743292919387*^9, 
   3.728743391180121*^9, 3.728747199672443*^9, 3.7287494655488586`*^9, 
   3.7287499321812153`*^9, 3.728750083671248*^9, 3.728750188389699*^9, 
   3.728750298129628*^9, 3.728750518820197*^9, 3.7287505779621572`*^9, 
   3.728750723090013*^9, {3.7287509825734453`*^9, 3.728751005661495*^9}, 
   3.728751617832035*^9, 3.728813262155266*^9, 3.728827379123863*^9, 
   3.7288430492376633`*^9, 3.728845621854383*^9, 3.733367267635683*^9, 
   3.79086265235045*^9, 3.790867107345881*^9, 3.792273207132406*^9},
 CellLabel->"During evaluation of In[52]:=",
 CellID->289896378,ExpressionUUID->"745c9ebc-aecf-46b3-8510-3b5bd218c97f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, -$CellContext`fred}, {1, -$CellContext`fred}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`fred}, {2, -$CellContext`fred}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`T[{2, -$CellContext`fred}, {
     1, -$CellContext`fred}] -> -$CellContext`T[{1, -$CellContext`fred}, {
      2, -$CellContext`fred}], " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.7286459173172073`*^9, 3.728659686300744*^9, 3.728660582228342*^9, 
   3.728671086075186*^9, 3.7286748786260967`*^9, 3.7287328676708927`*^9, 
   3.728733649639093*^9, 3.728742316161644*^9, 3.728743292919387*^9, 
   3.728743391180121*^9, 3.728747199672443*^9, 3.7287494655488586`*^9, 
   3.7287499321812153`*^9, 3.728750083671248*^9, 3.728750188389699*^9, 
   3.728750298129628*^9, 3.728750518820197*^9, 3.7287505779621572`*^9, 
   3.728750723090013*^9, {3.7287509825734453`*^9, 3.728751005661495*^9}, 
   3.728751617832035*^9, 3.728813262155266*^9, 3.728827379123863*^9, 
   3.7288430492376633`*^9, 3.728845621854383*^9, 3.733367267635683*^9, 
   3.79086265235045*^9, 3.790867107345881*^9, 3.792273207167738*^9},
 CellLabel->"During evaluation of In[52]:=",
 CellID->1795513650,ExpressionUUID->"267332f1-8507-493f-af6a-97b1e2fcec76"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, -$CellContext`fred}, {2, -$CellContext`fred}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`T[{2, -$CellContext`fred}, {
     2, -$CellContext`fred}] -> 0, " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.7286459173172073`*^9, 3.728659686300744*^9, 3.728660582228342*^9, 
   3.728671086075186*^9, 3.7286748786260967`*^9, 3.7287328676708927`*^9, 
   3.728733649639093*^9, 3.728742316161644*^9, 3.728743292919387*^9, 
   3.728743391180121*^9, 3.728747199672443*^9, 3.7287494655488586`*^9, 
   3.7287499321812153`*^9, 3.728750083671248*^9, 3.728750188389699*^9, 
   3.728750298129628*^9, 3.728750518820197*^9, 3.7287505779621572`*^9, 
   3.728750723090013*^9, {3.7287509825734453`*^9, 3.728751005661495*^9}, 
   3.728751617832035*^9, 3.728813262155266*^9, 3.728827379123863*^9, 
   3.7288430492376633`*^9, 3.728845621854383*^9, 3.733367267635683*^9, 
   3.79086265235045*^9, 3.790867107345881*^9, 3.7922732071767597`*^9},
 CellLabel->"During evaluation of In[52]:=",
 CellID->757707906,ExpressionUUID->"f79d217b-b6e5-41ae-b9ba-1d1a39fb69dd"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`fred}, {0, -$CellContext`fred}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`fred}, {0, -$CellContext`fred}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`fred}, {1, -$CellContext`fred}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`fred}, {1, -$CellContext`fred}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`fred}, {0, -$CellContext`fred}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`fred}, {2, -$CellContext`fred}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`fred}, {1, -$CellContext`fred}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{1, -$CellContext`fred}, {2, -$CellContext`fred}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`fred}, {2, -$CellContext`fred}],
       Editable->False], "\[Rule]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`fred}, {1, -$CellContext`fred}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", "1"}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`fred}, {2, -$CellContext`fred}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", "2"}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`fred}, {2, -$CellContext`fred}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", "1"}]}]}], "}"}]}], "]"}]], "Output",
 CellChangeTimes->{
  3.7286459179746923`*^9, 3.728659687048736*^9, 3.728660583362002*^9, 
   3.728671086866268*^9, 3.7286748794919367`*^9, 3.728732868206554*^9, 
   3.728733650319686*^9, 3.7287423167203703`*^9, 3.728743293496194*^9, 
   3.728743391961341*^9, 3.728747200459404*^9, 3.728749466357004*^9, 
   3.728749932922831*^9, 3.728750084574027*^9, 3.728750189459407*^9, 
   3.728750299470415*^9, 3.728750519591885*^9, 3.728750578749645*^9, 
   3.7287507235549517`*^9, {3.7287509832135067`*^9, 3.728751006208733*^9}, 
   3.7287516186252823`*^9, 3.728813262969472*^9, 3.728827379901928*^9, 
   3.7288430502302103`*^9, 3.7288456227529097`*^9, 3.733367268219865*^9, 
   3.790862652599327*^9, 3.7908671076560497`*^9, 3.792273207218918*^9},
 CellLabel->"Out[52]=",
 CellID->218749432,ExpressionUUID->"bb8f41fd-7280-4b15-8402-87665ef58891"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The FoldedRule notation represents a way of storing nested rules:\
\>", "MathCaption",
 CellChangeTimes->{{3.728659694973914*^9, 3.7286597475960093`*^9}},
 CellID->1885112655,ExpressionUUID->"4bdb2f65-538f-4d0c-ac66-cb62449142e9"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ColumnForm", "/@", 
  RowBox[{"TensorValues", "[", "T", "]"}]}]], "Input",
 CellLabel->"In[53]:=",ExpressionUUID->"00f32037-3bee-42f0-a788-d81b78c5e3ff"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   InterpretationBox[GridBox[{
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`fred}, {0, -$CellContext`fred}],
         Editable->False], "\[Rule]", "0"}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, -$CellContext`fred}, {0, -$CellContext`fred}],
         Editable->False], "\[Rule]", 
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"T", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["0",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["1",
                   FontColor->RGBColor[1., 0., 0.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
               FontSize->9]}
            },
            GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
            
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T[{0, -$CellContext`fred}, {1, -$CellContext`fred}],
          Editable->False]}]}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, -$CellContext`fred}, {1, -$CellContext`fred}],
         Editable->False], "\[Rule]", "0"}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{2, -$CellContext`fred}, {0, -$CellContext`fred}],
         Editable->False], "\[Rule]", 
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"T", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["0",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["2",
                   FontColor->RGBColor[1., 0., 0.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
               FontSize->9]}
            },
            GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
            
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T[{0, -$CellContext`fred}, {2, -$CellContext`fred}],
          Editable->False]}]}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{2, -$CellContext`fred}, {1, -$CellContext`fred}],
         Editable->False], "\[Rule]", 
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"T", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["1",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["2",
                   FontColor->RGBColor[1., 0., 0.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
               FontSize->9]}
            },
            GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
            
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T[{1, -$CellContext`fred}, {2, -$CellContext`fred}],
          Editable->False]}]}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{2, -$CellContext`fred}, {2, -$CellContext`fred}],
         Editable->False], "\[Rule]", "0"}]}
     },
     BaselinePosition->{Baseline, {1, 1}},
     GridBoxAlignment->{"Columns" -> {{Left}}, "Rows" -> {{Baseline}}}],
    ColumnForm[{$CellContext`T[{0, -$CellContext`fred}, {
        0, -$CellContext`fred}] -> 
      0, $CellContext`T[{1, -$CellContext`fred}, {
        0, -$CellContext`fred}] -> -$CellContext`T[{0, -$CellContext`fred}, {
         1, -$CellContext`fred}], $CellContext`T[{1, -$CellContext`fred}, {
        1, -$CellContext`fred}] -> 
      0, $CellContext`T[{2, -$CellContext`fred}, {
        0, -$CellContext`fred}] -> -$CellContext`T[{0, -$CellContext`fred}, {
         2, -$CellContext`fred}], $CellContext`T[{2, -$CellContext`fred}, {
        1, -$CellContext`fred}] -> -$CellContext`T[{1, -$CellContext`fred}, {
         2, -$CellContext`fred}], $CellContext`T[{2, -$CellContext`fred}, {
        2, -$CellContext`fred}] -> 0}],
    Editable->False], ",", 
   InterpretationBox[GridBox[{
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`fred}, {1, -$CellContext`fred}],
         Editable->False], "\[Rule]", 
        RowBox[{"-", "1"}]}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`fred}, {2, -$CellContext`fred}],
         Editable->False], "\[Rule]", 
        RowBox[{"-", "2"}]}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, -$CellContext`fred}, {2, -$CellContext`fred}],
         Editable->False], "\[Rule]", 
        RowBox[{"-", "1"}]}]}
     },
     BaselinePosition->{Baseline, {1, 1}},
     GridBoxAlignment->{"Columns" -> {{Left}}, "Rows" -> {{Baseline}}}],
    ColumnForm[{$CellContext`T[{0, -$CellContext`fred}, {
        1, -$CellContext`fred}] -> -1, $CellContext`T[{
        0, -$CellContext`fred}, {
        2, -$CellContext`fred}] -> -2, $CellContext`T[{
        1, -$CellContext`fred}, {2, -$CellContext`fred}] -> -1}],
    Editable->False]}], "]"}]], "Output",
 CellChangeTimes->{
  3.728645920101096*^9, 3.7286605835473127`*^9, 3.728671088071703*^9, 
   3.7286748798499327`*^9, 3.728732870520162*^9, 3.728733651825803*^9, 
   3.728742316859425*^9, 3.728743293667642*^9, 3.728743392147851*^9, 
   3.728747202889242*^9, 3.728749467512702*^9, 3.7287499330807123`*^9, 
   3.728750084744596*^9, 3.728750189669278*^9, 3.728750299686432*^9, 
   3.7287505197821074`*^9, 3.7287505789226007`*^9, 3.728750723659313*^9, {
   3.728750983379961*^9, 3.728751006302421*^9}, 3.728751618897305*^9, 
   3.7288132631516123`*^9, 3.72882738013489*^9, 3.728843051852195*^9, 
   3.7288456229477463`*^9, 3.7333672699365463`*^9, 3.790862654237136*^9, 
   3.790867108821273*^9, 3.7922732091408653`*^9},
 CellLabel->"Out[53]=",
 CellID->2010904542,ExpressionUUID->"5701f1c6-5f93-4f66-8f51-66aba91a25dd"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
As we can see, dependent and independent components are stored in separate \
sublists. We can change basis:\
\>", "MathCaption",ExpressionUUID->"201596bc-06bb-4556-922e-4c117a9b7035"],

Cell[BoxData[
 RowBox[{
  RowBox[{"$CVVerbose", "=", "False"}], ";"}]], "Input",
 CellLabel->"In[54]:=",ExpressionUUID->"26b472bf-3a6f-4dd6-a1a9-016ae202193e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ChangeComponents", "[", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{
     RowBox[{"-", 
      RowBox[{"{", 
       RowBox[{"a", ",", "blues"}], "}"}]}], ",", 
     RowBox[{"-", 
      RowBox[{"{", 
       RowBox[{"b", ",", "blues"}], "}"}]}]}], "]"}], ",", 
   RowBox[{"T", "[", 
    RowBox[{
     RowBox[{"-", 
      RowBox[{"{", 
       RowBox[{"a", ",", "fred"}], "}"}]}], ",", 
     RowBox[{"-", 
      RowBox[{"{", 
       RowBox[{"b", ",", "fred"}], "}"}]}]}], "]"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728733657823721*^9, 3.728733666718472*^9}, {
  3.728747207703905*^9, 3.728747214613381*^9}, {3.728749252135845*^9, 
  3.72874925297509*^9}},
 CellLabel->"In[55]:=",ExpressionUUID->"a88a7724-b78c-4527-b786-30607311eee0"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Computed \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["a",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{-$CellContext`a, -$CellContext`fred}, {-$CellContext`b, \
-$CellContext`blues}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["c",
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{-$CellContext`b, -$CellContext`blues}, \
{$CellContext`c, $CellContext`fred}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["a",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["c",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{-$CellContext`a, -$CellContext`fred}, {-$CellContext`c, \
-$CellContext`fred}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" in \"\>", 
   "\[InvisibleSpace]", "0.039167`5.044465301491588", 
   "\[InvisibleSpace]", "\<\" Seconds\"\>"}],
  SequenceForm[
  "Computed ", $CellContext`T[{-$CellContext`a, -$CellContext`fred}, \
{-$CellContext`b, -$CellContext`blues}] -> 
   xAct`xTensor`Basis[{-$CellContext`b, -$CellContext`blues}, \
{$CellContext`c, $CellContext`fred}] $CellContext`T[{-$CellContext`a, \
-$CellContext`fred}, {-$CellContext`c, -$CellContext`fred}], " in ", 
   0.039167`5.044465301491588, " Seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.7286459228511667`*^9, 3.728659757075322*^9, 3.728660584151334*^9, 
   3.7286710898935127`*^9, 3.7286748800471563`*^9, 3.728732872855349*^9, 
   3.728733667311123*^9, 3.728742317094062*^9, 3.72874329391508*^9, 
   3.728743392411622*^9, 3.728747215006261*^9, 3.7287494692934647`*^9, 
   3.728749933270899*^9, 3.728750085060267*^9, 3.728750190236019*^9, 
   3.728750300144363*^9, 3.728750520034485*^9, 3.728750579189823*^9, 
   3.7287507238179903`*^9, {3.728750983606666*^9, 3.72875100644411*^9}, 
   3.728751619130589*^9, 3.7288132633851347`*^9, 3.7288273803699703`*^9, 
   3.728843054036162*^9, 3.728845623215534*^9, 3.733367272118264*^9, 
   3.790862656260172*^9, 3.790867110543455*^9, 3.7922732113029346`*^9},
 CellLabel->"During evaluation of In[55]:=",
 CellID->1593342428,ExpressionUUID->"b7e58551-71c7-4f5a-95ec-1dd7aeb8cc65"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Computed \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{-$CellContext`a, -$CellContext`blues}, {-$CellContext`b, \
-$CellContext`blues}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["c",
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["a",
               FontColor->RGBColor[0., 0., 1.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{-$CellContext`a, -$CellContext`blues}, \
{$CellContext`c, $CellContext`fred}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["c",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{-$CellContext`c, -$CellContext`fred}, {-$CellContext`b, \
-$CellContext`blues}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" in \"\>", 
   "\[InvisibleSpace]", "0.05134`5.162000857831396", 
   "\[InvisibleSpace]", "\<\" Seconds\"\>"}],
  SequenceForm[
  "Computed ", $CellContext`T[{-$CellContext`a, -$CellContext`blues}, \
{-$CellContext`b, -$CellContext`blues}] -> 
   xAct`xTensor`Basis[{-$CellContext`a, -$CellContext`blues}, \
{$CellContext`c, $CellContext`fred}] $CellContext`T[{-$CellContext`c, \
-$CellContext`fred}, {-$CellContext`b, -$CellContext`blues}], " in ", 
   0.05134`5.162000857831396, " Seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.7286459228511667`*^9, 3.728659757075322*^9, 3.728660584151334*^9, 
   3.7286710898935127`*^9, 3.7286748800471563`*^9, 3.728732872855349*^9, 
   3.728733667311123*^9, 3.728742317094062*^9, 3.72874329391508*^9, 
   3.728743392411622*^9, 3.728747215006261*^9, 3.7287494692934647`*^9, 
   3.728749933270899*^9, 3.728750085060267*^9, 3.728750190236019*^9, 
   3.728750300144363*^9, 3.728750520034485*^9, 3.728750579189823*^9, 
   3.7287507238179903`*^9, {3.728750983606666*^9, 3.72875100644411*^9}, 
   3.728751619130589*^9, 3.7288132633851347`*^9, 3.7288273803699703`*^9, 
   3.728843054036162*^9, 3.728845623215534*^9, 3.733367272118264*^9, 
   3.790862656260172*^9, 3.790867110543455*^9, 3.792273211356633*^9},
 CellLabel->"During evaluation of In[55]:=",
 CellID->185109306,ExpressionUUID->"d8326ce1-787c-4f52-ac74-31037b577a5a"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["0",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`blues}, {0, -$CellContext`blues}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["0",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`blues}, {0, -$CellContext`blues}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`blues}, {1, -$CellContext`blues}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`blues}, {1, -$CellContext`blues}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["0",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`blues}, {0, -$CellContext`blues}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`blues}, {2, -$CellContext`blues}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`blues}, {1, -$CellContext`blues}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{1, -$CellContext`blues}, {2, -$CellContext`blues}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`blues}, {2, -$CellContext`blues}],
       Editable->False], "\[Rule]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`blues}, {1, -$CellContext`blues}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]},
                {
                 StyleBox["1",
                  FontColor->RGBColor[0., 0., 1.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{1, -$CellContext`blues}, {0, $CellContext`fred}],
         
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`blues}, {0, -$CellContext`fred}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {
                 StyleBox["1",
                  FontColor->RGBColor[0., 0., 1.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{1, -$CellContext`blues}, {1, $CellContext`fred}],
         
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`blues}, {1, -$CellContext`fred}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {
                 StyleBox["1",
                  FontColor->RGBColor[0., 0., 1.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{1, -$CellContext`blues}, {2, $CellContext`fred}],
         
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`blues}, {2, -$CellContext`fred}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`blues}, {2, -$CellContext`blues}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]},
                {
                 StyleBox["2",
                  FontColor->RGBColor[0., 0., 1.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{2, -$CellContext`blues}, {0, $CellContext`fred}],
         
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`blues}, {0, -$CellContext`fred}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {
                 StyleBox["2",
                  FontColor->RGBColor[0., 0., 1.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{2, -$CellContext`blues}, {1, $CellContext`fred}],
         
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`blues}, {1, -$CellContext`fred}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {
                 StyleBox["2",
                  FontColor->RGBColor[0., 0., 1.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{2, -$CellContext`blues}, {2, $CellContext`fred}],
         
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`blues}, {2, -$CellContext`fred}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`blues}, {2, -$CellContext`blues}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"e", 
              StyleBox[GridBox[{
                 {" ", 
                  StyleBox["0",
                   FontColor->RGBColor[1., 0., 0.]]},
                 {
                  StyleBox["2",
                   FontColor->RGBColor[0., 0., 1.]], " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
               FontSize->9]}
            },
            GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
            
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Basis[{2, -$CellContext`blues}, {0, $CellContext`fred}],
          Editable->False]}], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[0., 0., 1.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`fred}, {1, -$CellContext`blues}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {
                 StyleBox["2",
                  FontColor->RGBColor[0., 0., 1.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{2, -$CellContext`blues}, {1, $CellContext`fred}],
         
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[0., 0., 1.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, -$CellContext`blues}, {1, -$CellContext`fred}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {
                 StyleBox["2",
                  FontColor->RGBColor[0., 0., 1.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{2, -$CellContext`blues}, {2, $CellContext`fred}],
         
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[0., 0., 1.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, -$CellContext`blues}, {2, -$CellContext`fred}],
         Editable->False]}]}]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`fred}, {0, -$CellContext`blues}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`blues}, {0, -$CellContext`fred}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`fred}, {0, -$CellContext`blues}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`blues}, {1, -$CellContext`fred}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`fred}, {0, -$CellContext`blues}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`blues}, {2, -$CellContext`fred}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`fred}, {1, -$CellContext`blues}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{1, -$CellContext`blues}, {1, -$CellContext`fred}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`fred}, {1, -$CellContext`blues}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{1, -$CellContext`blues}, {2, -$CellContext`fred}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`fred}, {2, -$CellContext`blues}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{2, -$CellContext`blues}, {2, -$CellContext`fred}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`blues}, {0, -$CellContext`fred}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`fred}, {1, -$CellContext`blues}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`blues}, {0, -$CellContext`fred}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`fred}, {2, -$CellContext`blues}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`blues}, {1, -$CellContext`fred}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{1, -$CellContext`fred}, {2, -$CellContext`blues}],
        Editable->False]}]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`blues}, {0, -$CellContext`fred}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"e", 
              StyleBox[GridBox[{
                 {" ", 
                  StyleBox["1",
                   FontColor->RGBColor[1., 0., 0.]]},
                 {
                  StyleBox["0",
                   FontColor->RGBColor[0., 0., 1.]], " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
               FontSize->9]}
            },
            GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
            
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Basis[{0, -$CellContext`blues}, {1, $CellContext`fred}],
          Editable->False]}], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`fred}, {1, -$CellContext`fred}],
         Editable->False]}], "-", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{0, -$CellContext`blues}, {2, $CellContext`fred}],
         
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`fred}, {2, -$CellContext`fred}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`blues}, {1, -$CellContext`fred}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]},
                {
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{0, -$CellContext`blues}, {0, $CellContext`fred}],
         
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`fred}, {1, -$CellContext`fred}],
         Editable->False]}], "-", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{0, -$CellContext`blues}, {2, $CellContext`fred}],
         
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, -$CellContext`fred}, {2, -$CellContext`fred}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`blues}, {2, -$CellContext`fred}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]},
                {
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{0, -$CellContext`blues}, {0, $CellContext`fred}],
         
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`fred}, {2, -$CellContext`fred}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{0, -$CellContext`blues}, {1, $CellContext`fred}],
         
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, -$CellContext`fred}, {2, -$CellContext`fred}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`fred}, {1, -$CellContext`blues}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {
                 StyleBox["1",
                  FontColor->RGBColor[0., 0., 1.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{1, -$CellContext`blues}, {1, $CellContext`fred}],
         
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`fred}, {1, -$CellContext`fred}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {
                 StyleBox["1",
                  FontColor->RGBColor[0., 0., 1.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{1, -$CellContext`blues}, {2, $CellContext`fred}],
         
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`fred}, {2, -$CellContext`fred}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`blues}, {1, -$CellContext`fred}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]},
                {
                 StyleBox["1",
                  FontColor->RGBColor[0., 0., 1.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{1, -$CellContext`blues}, {0, $CellContext`fred}],
         
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`fred}, {1, -$CellContext`fred}],
         Editable->False]}], "-", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {
                 StyleBox["1",
                  FontColor->RGBColor[0., 0., 1.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{1, -$CellContext`blues}, {2, $CellContext`fred}],
         
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, -$CellContext`fred}, {2, -$CellContext`fred}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`blues}, {2, -$CellContext`fred}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]},
                {
                 StyleBox["1",
                  FontColor->RGBColor[0., 0., 1.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{1, -$CellContext`blues}, {0, $CellContext`fred}],
         
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`fred}, {2, -$CellContext`fred}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {
                 StyleBox["1",
                  FontColor->RGBColor[0., 0., 1.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{1, -$CellContext`blues}, {1, $CellContext`fred}],
         
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, -$CellContext`fred}, {2, -$CellContext`fred}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`fred}, {2, -$CellContext`blues}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {
                 StyleBox["2",
                  FontColor->RGBColor[0., 0., 1.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{2, -$CellContext`blues}, {1, $CellContext`fred}],
         
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`fred}, {1, -$CellContext`fred}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {
                 StyleBox["2",
                  FontColor->RGBColor[0., 0., 1.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{2, -$CellContext`blues}, {2, $CellContext`fred}],
         
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`fred}, {2, -$CellContext`fred}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`fred}, {2, -$CellContext`blues}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"e", 
              StyleBox[GridBox[{
                 {" ", 
                  StyleBox["0",
                   FontColor->RGBColor[1., 0., 0.]]},
                 {
                  StyleBox["2",
                   FontColor->RGBColor[0., 0., 1.]], " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
               FontSize->9]}
            },
            GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
            
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Basis[{2, -$CellContext`blues}, {0, $CellContext`fred}],
          Editable->False]}], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`fred}, {1, -$CellContext`fred}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {
                 StyleBox["2",
                  FontColor->RGBColor[0., 0., 1.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{2, -$CellContext`blues}, {2, $CellContext`fred}],
         
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, -$CellContext`fred}, {2, -$CellContext`fred}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`blues}, {2, -$CellContext`fred}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]},
                {
                 StyleBox["2",
                  FontColor->RGBColor[0., 0., 1.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{2, -$CellContext`blues}, {0, $CellContext`fred}],
         
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`fred}, {2, -$CellContext`fred}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {
                 StyleBox["2",
                  FontColor->RGBColor[0., 0., 1.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{2, -$CellContext`blues}, {1, $CellContext`fred}],
         
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, -$CellContext`fred}, {2, -$CellContext`fred}],
         Editable->False]}]}]}]}], "}"}]}], "]"}]], "Output",
 CellChangeTimes->{
  3.7286459230374117`*^9, 3.728659757294372*^9, 3.72866058435069*^9, 
   3.728671090099367*^9, 3.728674880297851*^9, 3.728732873036181*^9, 
   3.728733667516769*^9, 3.728742317317684*^9, 3.728743294132987*^9, 
   3.7287433926221437`*^9, 3.728747215201643*^9, 3.728749469484234*^9, 
   3.7287499334980917`*^9, 3.728750085275119*^9, 3.7287501904443293`*^9, 
   3.728750300799444*^9, 3.728750520212735*^9, 3.7287505795030327`*^9, 
   3.7287507239856157`*^9, {3.728750983805768*^9, 3.7287510066182213`*^9}, 
   3.728751619348427*^9, 3.72881326365734*^9, 3.72882738066547*^9, 
   3.728843054295719*^9, 3.728845623469989*^9, 3.733367272289229*^9, 
   3.790862656431302*^9, 3.790867110708832*^9, 3.792273211539281*^9},
 CellLabel->"Out[55]=",
 CellID->1575785086,ExpressionUUID->"721f3b7c-2116-4398-8363-82c4c54b088e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ColumnForm", "/@", 
  RowBox[{"TensorValues", "[", 
   RowBox[{"T", ",", " ", 
    RowBox[{"{", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "blues"}], ",", " ", 
       RowBox[{"-", "blues"}]}], "}"}], "}"}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.728733670591663*^9, 3.7287336726387777`*^9}, {
  3.728747218039291*^9, 3.7287472208292637`*^9}, {3.7287492540175743`*^9, 
  3.728749254682846*^9}},
 CellLabel->"In[56]:=",ExpressionUUID->"bdde1cee-8adf-4e1d-aaf1-a71df7604a34"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   InterpretationBox[GridBox[{
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]], 
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`blues}, {0, -$CellContext`blues}],
         Editable->False], "\[Rule]", "0"}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[0., 0., 1.]], 
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, -$CellContext`blues}, {0, -$CellContext`blues}],
         Editable->False], "\[Rule]", 
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"T", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["0",
                   FontColor->RGBColor[0., 0., 1.]], 
                  StyleBox["1",
                   FontColor->RGBColor[0., 0., 1.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
               FontSize->9]}
            },
            GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
            
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T[{0, -$CellContext`blues}, {1, -$CellContext`blues}],
          Editable->False]}]}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[0., 0., 1.]], 
                 StyleBox["1",
                  FontColor->RGBColor[0., 0., 1.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, -$CellContext`blues}, {1, -$CellContext`blues}],
         Editable->False], "\[Rule]", "0"}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["2",
                  FontColor->RGBColor[0., 0., 1.]], 
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{2, -$CellContext`blues}, {0, -$CellContext`blues}],
         Editable->False], "\[Rule]", 
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"T", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["0",
                   FontColor->RGBColor[0., 0., 1.]], 
                  StyleBox["2",
                   FontColor->RGBColor[0., 0., 1.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
               FontSize->9]}
            },
            GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
            
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T[{0, -$CellContext`blues}, {2, -$CellContext`blues}],
          Editable->False]}]}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["2",
                  FontColor->RGBColor[0., 0., 1.]], 
                 StyleBox["1",
                  FontColor->RGBColor[0., 0., 1.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{2, -$CellContext`blues}, {1, -$CellContext`blues}],
         Editable->False], "\[Rule]", 
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"T", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["1",
                   FontColor->RGBColor[0., 0., 1.]], 
                  StyleBox["2",
                   FontColor->RGBColor[0., 0., 1.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
               FontSize->9]}
            },
            GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
            
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T[{1, -$CellContext`blues}, {2, -$CellContext`blues}],
          Editable->False]}]}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["2",
                  FontColor->RGBColor[0., 0., 1.]], 
                 StyleBox["2",
                  FontColor->RGBColor[0., 0., 1.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{2, -$CellContext`blues}, {2, -$CellContext`blues}],
         Editable->False], "\[Rule]", "0"}]}
     },
     BaselinePosition->{Baseline, {1, 1}},
     GridBoxAlignment->{"Columns" -> {{Left}}, "Rows" -> {{Baseline}}}],
    ColumnForm[{$CellContext`T[{0, -$CellContext`blues}, {
        0, -$CellContext`blues}] -> 
      0, $CellContext`T[{1, -$CellContext`blues}, {
        0, -$CellContext`blues}] -> -$CellContext`T[{
         0, -$CellContext`blues}, {1, -$CellContext`blues}], $CellContext`T[{
        1, -$CellContext`blues}, {1, -$CellContext`blues}] -> 
      0, $CellContext`T[{2, -$CellContext`blues}, {
        0, -$CellContext`blues}] -> -$CellContext`T[{
         0, -$CellContext`blues}, {2, -$CellContext`blues}], $CellContext`T[{
        2, -$CellContext`blues}, {
        1, -$CellContext`blues}] -> -$CellContext`T[{
         1, -$CellContext`blues}, {2, -$CellContext`blues}], $CellContext`T[{
        2, -$CellContext`blues}, {2, -$CellContext`blues}] -> 0}],
    Editable->False], ",", 
   InterpretationBox[GridBox[{
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]], 
                 StyleBox["1",
                  FontColor->RGBColor[0., 0., 1.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`blues}, {1, -$CellContext`blues}],
         Editable->False], "\[Rule]", 
        RowBox[{
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"e", 
               StyleBox[GridBox[{
                  {" ", 
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {
                   StyleBox["1",
                    FontColor->RGBColor[0., 0., 1.]], " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                FontSize->9]}
             },
             GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           
           xAct`xTensor`Basis[{1, -$CellContext`blues}, {
            0, $CellContext`fred}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"T", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["0",
                    FontColor->RGBColor[0., 0., 1.]], 
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                FontSize->9]}
             },
             GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`T[{0, -$CellContext`blues}, {0, -$CellContext`fred}],
           Editable->False]}], "+", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"e", 
               StyleBox[GridBox[{
                  {" ", 
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {
                   StyleBox["1",
                    FontColor->RGBColor[0., 0., 1.]], " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                FontSize->9]}
             },
             GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           
           xAct`xTensor`Basis[{1, -$CellContext`blues}, {
            1, $CellContext`fred}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"T", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["0",
                    FontColor->RGBColor[0., 0., 1.]], 
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                FontSize->9]}
             },
             GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`T[{0, -$CellContext`blues}, {1, -$CellContext`fred}],
           Editable->False]}], "+", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"e", 
               StyleBox[GridBox[{
                  {" ", 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {
                   StyleBox["1",
                    FontColor->RGBColor[0., 0., 1.]], " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                FontSize->9]}
             },
             GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           
           xAct`xTensor`Basis[{1, -$CellContext`blues}, {
            2, $CellContext`fred}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"T", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["0",
                    FontColor->RGBColor[0., 0., 1.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                FontSize->9]}
             },
             GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`T[{0, -$CellContext`blues}, {2, -$CellContext`fred}],
           Editable->False]}]}]}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]], 
                 StyleBox["2",
                  FontColor->RGBColor[0., 0., 1.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`blues}, {2, -$CellContext`blues}],
         Editable->False], "\[Rule]", 
        RowBox[{
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"e", 
               StyleBox[GridBox[{
                  {" ", 
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {
                   StyleBox["2",
                    FontColor->RGBColor[0., 0., 1.]], " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                FontSize->9]}
             },
             GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           
           xAct`xTensor`Basis[{2, -$CellContext`blues}, {
            0, $CellContext`fred}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"T", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["0",
                    FontColor->RGBColor[0., 0., 1.]], 
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                FontSize->9]}
             },
             GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`T[{0, -$CellContext`blues}, {0, -$CellContext`fred}],
           Editable->False]}], "+", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"e", 
               StyleBox[GridBox[{
                  {" ", 
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {
                   StyleBox["2",
                    FontColor->RGBColor[0., 0., 1.]], " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                FontSize->9]}
             },
             GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           
           xAct`xTensor`Basis[{2, -$CellContext`blues}, {
            1, $CellContext`fred}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"T", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["0",
                    FontColor->RGBColor[0., 0., 1.]], 
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                FontSize->9]}
             },
             GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`T[{0, -$CellContext`blues}, {1, -$CellContext`fred}],
           Editable->False]}], "+", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"e", 
               StyleBox[GridBox[{
                  {" ", 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {
                   StyleBox["2",
                    FontColor->RGBColor[0., 0., 1.]], " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                FontSize->9]}
             },
             GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           
           xAct`xTensor`Basis[{2, -$CellContext`blues}, {
            2, $CellContext`fred}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"T", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["0",
                    FontColor->RGBColor[0., 0., 1.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                FontSize->9]}
             },
             GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`T[{0, -$CellContext`blues}, {2, -$CellContext`fred}],
           Editable->False]}]}]}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[0., 0., 1.]], 
                 StyleBox["2",
                  FontColor->RGBColor[0., 0., 1.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, -$CellContext`blues}, {2, -$CellContext`blues}],
         Editable->False], "\[Rule]", 
        RowBox[{
         RowBox[{
          RowBox[{"-", 
           InterpretationBox[
            StyleBox[GridBox[{
               {"e", 
                StyleBox[GridBox[{
                   {" ", 
                    StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]]},
                   {
                    StyleBox["2",
                    FontColor->RGBColor[0., 0., 1.]], " "}
                  },
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                 FontSize->9]}
              },
              
              GridBoxAlignment->{
               "Columns" -> {{Center}}, "Rows" -> {{Center}}},
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.27999999999999997`], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.27999999999999997`]}, "Rows" -> {
                  Offset[0.2], {
                   Offset[0.4]}, 
                  Offset[0.2]}}],
             ShowAutoStyles->False,
             AutoSpacing->False],
            
            xAct`xTensor`Basis[{2, -$CellContext`blues}, {
             0, $CellContext`fred}],
            Editable->False]}], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"T", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["1",
                    FontColor->RGBColor[0., 0., 1.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                FontSize->9]}
             },
             GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`T[{0, -$CellContext`fred}, {1, -$CellContext`blues}],
           Editable->False]}], "+", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"e", 
               StyleBox[GridBox[{
                  {" ", 
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {
                   StyleBox["2",
                    FontColor->RGBColor[0., 0., 1.]], " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                FontSize->9]}
             },
             GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           
           xAct`xTensor`Basis[{2, -$CellContext`blues}, {
            1, $CellContext`fred}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"T", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["1",
                    FontColor->RGBColor[0., 0., 1.]], 
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                FontSize->9]}
             },
             GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`T[{1, -$CellContext`blues}, {1, -$CellContext`fred}],
           Editable->False]}], "+", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"e", 
               StyleBox[GridBox[{
                  {" ", 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {
                   StyleBox["2",
                    FontColor->RGBColor[0., 0., 1.]], " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                FontSize->9]}
             },
             GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           
           xAct`xTensor`Basis[{2, -$CellContext`blues}, {
            2, $CellContext`fred}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"T", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["1",
                    FontColor->RGBColor[0., 0., 1.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                FontSize->9]}
             },
             GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`T[{1, -$CellContext`blues}, {2, -$CellContext`fred}],
           Editable->False]}]}]}]}
     },
     BaselinePosition->{Baseline, {1, 1}},
     GridBoxAlignment->{"Columns" -> {{Left}}, "Rows" -> {{Baseline}}}],
    ColumnForm[{$CellContext`T[{0, -$CellContext`blues}, {
        1, -$CellContext`blues}] -> 
      xAct`xTensor`Basis[{1, -$CellContext`blues}, {
          0, $CellContext`fred}] $CellContext`T[{0, -$CellContext`blues}, {
          0, -$CellContext`fred}] + 
       xAct`xTensor`Basis[{1, -$CellContext`blues}, {
          1, $CellContext`fred}] $CellContext`T[{0, -$CellContext`blues}, {
          1, -$CellContext`fred}] + 
       xAct`xTensor`Basis[{1, -$CellContext`blues}, {
          2, $CellContext`fred}] $CellContext`T[{0, -$CellContext`blues}, {
          2, -$CellContext`fred}], $CellContext`T[{0, -$CellContext`blues}, {
        2, -$CellContext`blues}] -> 
      xAct`xTensor`Basis[{2, -$CellContext`blues}, {
          0, $CellContext`fred}] $CellContext`T[{0, -$CellContext`blues}, {
          0, -$CellContext`fred}] + 
       xAct`xTensor`Basis[{2, -$CellContext`blues}, {
          1, $CellContext`fred}] $CellContext`T[{0, -$CellContext`blues}, {
          1, -$CellContext`fred}] + 
       xAct`xTensor`Basis[{2, -$CellContext`blues}, {
          2, $CellContext`fred}] $CellContext`T[{0, -$CellContext`blues}, {
          2, -$CellContext`fred}], $CellContext`T[{1, -$CellContext`blues}, {
        2, -$CellContext`blues}] -> -
        xAct`xTensor`Basis[{2, -$CellContext`blues}, {
          0, $CellContext`fred}] $CellContext`T[{0, -$CellContext`fred}, {
          1, -$CellContext`blues}] + 
       xAct`xTensor`Basis[{2, -$CellContext`blues}, {
          1, $CellContext`fred}] $CellContext`T[{1, -$CellContext`blues}, {
          1, -$CellContext`fred}] + 
       xAct`xTensor`Basis[{2, -$CellContext`blues}, {
          2, $CellContext`fred}] $CellContext`T[{1, -$CellContext`blues}, {
          2, -$CellContext`fred}]}],
    Editable->False]}], "]"}]], "Output",
 CellChangeTimes->{
  3.728645924666836*^9, 3.728659762206048*^9, 3.7286605853308163`*^9, 
   3.728671091742529*^9, 3.728674880475525*^9, 3.728732874218255*^9, 
   3.728733673300496*^9, 3.728742317480967*^9, 3.7287432943153353`*^9, 
   3.728743392781143*^9, 3.728747221333743*^9, 3.728749471109683*^9, 
   3.728749933735169*^9, 3.728750085508163*^9, 3.728750190671772*^9, 
   3.7287503010340233`*^9, 3.7287505204723167`*^9, 3.728750579702359*^9, 
   3.728750724121296*^9, {3.728750984039693*^9, 3.728751006751008*^9}, 
   3.72875161954198*^9, 3.7288132638334303`*^9, 3.72882738087258*^9, 
   3.728843056056176*^9, 3.7288456236959667`*^9, 3.733367273716549*^9, 
   3.79086265757546*^9, 3.790867112029004*^9, 3.7922732129124537`*^9},
 CellLabel->"Out[56]=",
 CellID->50503431,ExpressionUUID->"1e9cb0c5-48c4-4d68-9765-c1690ca4190e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Clean up:", "MathCaption",
 CellChangeTimes->{{3.7286597672185*^9, 3.7286597745066967`*^9}, 
   3.728660495866705*^9},ExpressionUUID->"22be9d13-4b88-494d-afe2-\
0b5ea855fdca"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DeleteTensorValues", "[", "T", "]"}]], "Input",
 CellLabel->"In[57]:=",ExpressionUUID->"d394f28a-2df6-4f92-80a6-e57a1477b212"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Deleted values for tensor \"\>", "\[InvisibleSpace]", "T", 
   "\[InvisibleSpace]", "\<\", derivatives \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", "}"}], "\[InvisibleSpace]", "\<\" and bases \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "blues"}], ",", 
      RowBox[{"-", "blues"}]}], "}"}], "}"}], 
   "\[InvisibleSpace]", "\<\".\"\>"}],
  SequenceForm[
  "Deleted values for tensor ", $CellContext`T, ", derivatives ", {}, 
   " and bases ", {{-$CellContext`blues, -$CellContext`blues}}, "."],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645926770521*^9, 3.7286597758742113`*^9, 3.728660585525004*^9, 
   3.728671093018279*^9, 3.728674880619116*^9, 3.728732876381125*^9, 
   3.728733676557001*^9, 3.728742317613348*^9, 3.728743294432661*^9, 
   3.728743392914095*^9, 3.728747224415332*^9, 3.7287494726411963`*^9, 
   3.728749933919656*^9, 3.728750085682232*^9, 3.7287501908476467`*^9, 
   3.728750301470889*^9, 3.728750520604165*^9, 3.728750579861247*^9, 
   3.728750724218624*^9, {3.728750984191975*^9, 3.728751006851103*^9}, 
   3.728751619682251*^9, 3.7288132639870253`*^9, 3.728827381030262*^9, 
   3.728843057543241*^9, 3.728845623887203*^9, 3.733367275888467*^9, 
   3.790862658836885*^9, 3.790867113306862*^9, 3.792273214557021*^9},
 CellLabel->"During evaluation of In[57]:=",
 CellID->345137860,ExpressionUUID->"ba0bb955-80e9-47ee-9fec-a0d468038c04"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Deleted values for tensor \"\>", "\[InvisibleSpace]", "T", 
   "\[InvisibleSpace]", "\<\", derivatives \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", "}"}], "\[InvisibleSpace]", "\<\" and bases \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "blues"}], ",", 
       RowBox[{"-", "fred"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "fred"}], ",", 
       RowBox[{"-", "blues"}]}], "}"}]}], "}"}], 
   "\[InvisibleSpace]", "\<\".\"\>"}],
  SequenceForm[
  "Deleted values for tensor ", $CellContext`T, ", derivatives ", {}, 
   " and bases ", {{-$CellContext`blues, -$CellContext`fred}, \
{-$CellContext`fred, -$CellContext`blues}}, "."],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645926770521*^9, 3.7286597758742113`*^9, 3.728660585525004*^9, 
   3.728671093018279*^9, 3.728674880619116*^9, 3.728732876381125*^9, 
   3.728733676557001*^9, 3.728742317613348*^9, 3.728743294432661*^9, 
   3.728743392914095*^9, 3.728747224415332*^9, 3.7287494726411963`*^9, 
   3.728749933919656*^9, 3.728750085682232*^9, 3.7287501908476467`*^9, 
   3.728750301470889*^9, 3.728750520604165*^9, 3.728750579861247*^9, 
   3.728750724218624*^9, {3.728750984191975*^9, 3.728751006851103*^9}, 
   3.728751619682251*^9, 3.7288132639870253`*^9, 3.728827381030262*^9, 
   3.728843057543241*^9, 3.728845623887203*^9, 3.733367275888467*^9, 
   3.790862658836885*^9, 3.790867113306862*^9, 3.7922732145682983`*^9},
 CellLabel->"During evaluation of In[57]:=",
 CellID->378682473,ExpressionUUID->"e59a450a-0dc6-4238-806d-87eb4d5235d9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Deleted values for tensor \"\>", "\[InvisibleSpace]", "T", 
   "\[InvisibleSpace]", "\<\", derivatives \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", "}"}], "\[InvisibleSpace]", "\<\" and bases \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "fred"}], ",", 
      RowBox[{"-", "fred"}]}], "}"}], "}"}], 
   "\[InvisibleSpace]", "\<\".\"\>"}],
  SequenceForm[
  "Deleted values for tensor ", $CellContext`T, ", derivatives ", {}, 
   " and bases ", {{-$CellContext`fred, -$CellContext`fred}}, "."],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645926770521*^9, 3.7286597758742113`*^9, 3.728660585525004*^9, 
   3.728671093018279*^9, 3.728674880619116*^9, 3.728732876381125*^9, 
   3.728733676557001*^9, 3.728742317613348*^9, 3.728743294432661*^9, 
   3.728743392914095*^9, 3.728747224415332*^9, 3.7287494726411963`*^9, 
   3.728749933919656*^9, 3.728750085682232*^9, 3.7287501908476467`*^9, 
   3.728750301470889*^9, 3.728750520604165*^9, 3.728750579861247*^9, 
   3.728750724218624*^9, {3.728750984191975*^9, 3.728751006851103*^9}, 
   3.728751619682251*^9, 3.7288132639870253`*^9, 3.728827381030262*^9, 
   3.728843057543241*^9, 3.728845623887203*^9, 3.733367275888467*^9, 
   3.790862658836885*^9, 3.790867113306862*^9, 3.7922732145789757`*^9},
 CellLabel->"During evaluation of In[57]:=",
 CellID->1522277650,ExpressionUUID->"387379d4-cd80-41ec-b4f1-67f669fa4421"]
}, Open  ]]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"$CVVerbose", "=", "True"}], ";"}]], "Input",
 CellLabel->"In[58]:=",ExpressionUUID->"12dc5e47-bb9e-4b22-82f4-8c45d98d3b38"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"UndefBasis", "/@", 
   RowBox[{"{", 
    RowBox[{"fred", ",", "blues"}], "}"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.7287336824000797`*^9, 3.7287336849020233`*^9}, {
   3.728747233092782*^9, 3.728747236764454*^9}, 3.7287492602227173`*^9},
 CellLabel->"In[59]:=",ExpressionUUID->"68321ca2-7f23-414c-afb7-abf188a74b94"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "ChristoffelPDfred"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "non-symmetric Christoffel tensor", " ", $CellContext`ChristoffelPDfred],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645939324555*^9, 3.728660585841178*^9, 3.728671095157638*^9, 
   3.728674880967157*^9, 3.728732879161681*^9, 3.728733686924255*^9, 
   3.728742318008175*^9, 3.7287432947505503`*^9, 3.728743393197723*^9, 
   3.728747237262712*^9, 3.7287494745303164`*^9, 3.7287499348201714`*^9, 
   3.728750085987439*^9, 3.7287501915322237`*^9, 3.7287503020600357`*^9, 
   3.728750520800502*^9, 3.728750580123959*^9, 3.72875072434544*^9, {
   3.7287509844380207`*^9, 3.728751006976675*^9}, 3.728751619944504*^9, 
   3.728813264274679*^9, 3.728827381326201*^9, 3.7288430592053967`*^9, 
   3.728845624172392*^9, 3.733367277519153*^9, 3.790862660646812*^9, 
   3.790867115247067*^9, 3.792273216824662*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->668658622,ExpressionUUID->"20f30a94-0934-46fd-8929-a5a2e99cdfb1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciPDfred"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Ricci tensor", 
   " ", $CellContext`RicciPDfred],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645939324555*^9, 3.728660585841178*^9, 3.728671095157638*^9, 
   3.728674880967157*^9, 3.728732879161681*^9, 3.728733686924255*^9, 
   3.728742318008175*^9, 3.7287432947505503`*^9, 3.728743393197723*^9, 
   3.728747237262712*^9, 3.7287494745303164`*^9, 3.7287499348201714`*^9, 
   3.728750085987439*^9, 3.7287501915322237`*^9, 3.7287503020600357`*^9, 
   3.728750520800502*^9, 3.728750580123959*^9, 3.72875072434544*^9, {
   3.7287509844380207`*^9, 3.728751006976675*^9}, 3.728751619944504*^9, 
   3.728813264274679*^9, 3.728827381326201*^9, 3.7288430592053967`*^9, 
   3.728845624172392*^9, 3.733367277519153*^9, 3.790862660646812*^9, 
   3.790867115247067*^9, 3.792273216837693*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->363535203,ExpressionUUID->"248387c7-e78f-4d05-a0b9-51691f210971"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannPDfred"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannPDfred],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645939324555*^9, 3.728660585841178*^9, 3.728671095157638*^9, 
   3.728674880967157*^9, 3.728732879161681*^9, 3.728733686924255*^9, 
   3.728742318008175*^9, 3.7287432947505503`*^9, 3.728743393197723*^9, 
   3.728747237262712*^9, 3.7287494745303164`*^9, 3.7287499348201714`*^9, 
   3.728750085987439*^9, 3.7287501915322237`*^9, 3.7287503020600357`*^9, 
   3.728750520800502*^9, 3.728750580123959*^9, 3.72875072434544*^9, {
   3.7287509844380207`*^9, 3.728751006976675*^9}, 3.728751619944504*^9, 
   3.728813264274679*^9, 3.728827381326201*^9, 3.7288430592053967`*^9, 
   3.728845624172392*^9, 3.733367277519153*^9, 3.790862660646812*^9, 
   3.790867115247067*^9, 3.7922732168517437`*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->345181106,ExpressionUUID->"3a692f01-3260-412e-b8d1-c7b37b3a5ece"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TorsionPDfred"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionPDfred],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645939324555*^9, 3.728660585841178*^9, 3.728671095157638*^9, 
   3.728674880967157*^9, 3.728732879161681*^9, 3.728733686924255*^9, 
   3.728742318008175*^9, 3.7287432947505503`*^9, 3.728743393197723*^9, 
   3.728747237262712*^9, 3.7287494745303164`*^9, 3.7287499348201714`*^9, 
   3.728750085987439*^9, 3.7287501915322237`*^9, 3.7287503020600357`*^9, 
   3.728750520800502*^9, 3.728750580123959*^9, 3.72875072434544*^9, {
   3.7287509844380207`*^9, 3.728751006976675*^9}, 3.728751619944504*^9, 
   3.728813264274679*^9, 3.728827381326201*^9, 3.7288430592053967`*^9, 
   3.728845624172392*^9, 3.733367277519153*^9, 3.790862660646812*^9, 
   3.790867115247067*^9, 3.792273216858735*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->1105517995,ExpressionUUID->"beaafcbc-10ea-4767-8fbe-18c113cf1607"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"parallel derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "PDfred"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "parallel derivative", 
   " ", $CellContext`PDfred],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645939324555*^9, 3.728660585841178*^9, 3.728671095157638*^9, 
   3.728674880967157*^9, 3.728732879161681*^9, 3.728733686924255*^9, 
   3.728742318008175*^9, 3.7287432947505503`*^9, 3.728743393197723*^9, 
   3.728747237262712*^9, 3.7287494745303164`*^9, 3.7287499348201714`*^9, 
   3.728750085987439*^9, 3.7287501915322237`*^9, 3.7287503020600357`*^9, 
   3.728750520800502*^9, 3.728750580123959*^9, 3.72875072434544*^9, {
   3.7287509844380207`*^9, 3.728751006976675*^9}, 3.728751619944504*^9, 
   3.728813264274679*^9, 3.728827381326201*^9, 3.7288430592053967`*^9, 
   3.728845624172392*^9, 3.733367277519153*^9, 3.790862660646812*^9, 
   3.790867115247067*^9, 3.7922732168661823`*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->412428121,ExpressionUUID->"4e9ac448-47ff-41ae-bfcf-fe983ff89c68"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "etaUpfred"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric +1 density",
    " ", $CellContext`etaUpfred],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645939324555*^9, 3.728660585841178*^9, 3.728671095157638*^9, 
   3.728674880967157*^9, 3.728732879161681*^9, 3.728733686924255*^9, 
   3.728742318008175*^9, 3.7287432947505503`*^9, 3.728743393197723*^9, 
   3.728747237262712*^9, 3.7287494745303164`*^9, 3.7287499348201714`*^9, 
   3.728750085987439*^9, 3.7287501915322237`*^9, 3.7287503020600357`*^9, 
   3.728750520800502*^9, 3.728750580123959*^9, 3.72875072434544*^9, {
   3.7287509844380207`*^9, 3.728751006976675*^9}, 3.728751619944504*^9, 
   3.728813264274679*^9, 3.728827381326201*^9, 3.7288430592053967`*^9, 
   3.728845624172392*^9, 3.733367277519153*^9, 3.790862660646812*^9, 
   3.790867115247067*^9, 3.7922732168735523`*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->1749700778,ExpressionUUID->"02f628a7-617b-437f-b33a-242efa0f90ee"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "etaDownfred"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric -1 density",
    " ", $CellContext`etaDownfred],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645939324555*^9, 3.728660585841178*^9, 3.728671095157638*^9, 
   3.728674880967157*^9, 3.728732879161681*^9, 3.728733686924255*^9, 
   3.728742318008175*^9, 3.7287432947505503`*^9, 3.728743393197723*^9, 
   3.728747237262712*^9, 3.7287494745303164`*^9, 3.7287499348201714`*^9, 
   3.728750085987439*^9, 3.7287501915322237`*^9, 3.7287503020600357`*^9, 
   3.728750520800502*^9, 3.728750580123959*^9, 3.72875072434544*^9, {
   3.7287509844380207`*^9, 3.728751006976675*^9}, 3.728751619944504*^9, 
   3.728813264274679*^9, 3.728827381326201*^9, 3.7288430592053967`*^9, 
   3.728845624172392*^9, 3.733367277519153*^9, 3.790862660646812*^9, 
   3.790867115247067*^9, 3.792273216880509*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->1727397475,ExpressionUUID->"2fcee940-f0a3-41db-9456-a0d0f7637d7b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefBasis", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"basis\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "fred"}],
  SequenceForm[
  "** ", xAct`xCoba`UndefBasis, ": Undefined ", "basis", 
   " ", $CellContext`fred],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645939324555*^9, 3.728660585841178*^9, 3.728671095157638*^9, 
   3.728674880967157*^9, 3.728732879161681*^9, 3.728733686924255*^9, 
   3.728742318008175*^9, 3.7287432947505503`*^9, 3.728743393197723*^9, 
   3.728747237262712*^9, 3.7287494745303164`*^9, 3.7287499348201714`*^9, 
   3.728750085987439*^9, 3.7287501915322237`*^9, 3.7287503020600357`*^9, 
   3.728750520800502*^9, 3.728750580123959*^9, 3.72875072434544*^9, {
   3.7287509844380207`*^9, 3.728751006976675*^9}, 3.728751619944504*^9, 
   3.728813264274679*^9, 3.728827381326201*^9, 3.7288430592053967`*^9, 
   3.728845624172392*^9, 3.733367277519153*^9, 3.790862660646812*^9, 
   3.790867115247067*^9, 3.7922732168875303`*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->537938532,ExpressionUUID->"e1a9d93f-cb91-4b8d-b880-c6072502cebf"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "ChristoffelPDblues"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "non-symmetric Christoffel tensor", " ", $CellContext`ChristoffelPDblues],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645939324555*^9, 3.728660585841178*^9, 3.728671095157638*^9, 
   3.728674880967157*^9, 3.728732879161681*^9, 3.728733686924255*^9, 
   3.728742318008175*^9, 3.7287432947505503`*^9, 3.728743393197723*^9, 
   3.728747237262712*^9, 3.7287494745303164`*^9, 3.7287499348201714`*^9, 
   3.728750085987439*^9, 3.7287501915322237`*^9, 3.7287503020600357`*^9, 
   3.728750520800502*^9, 3.728750580123959*^9, 3.72875072434544*^9, {
   3.7287509844380207`*^9, 3.728751006976675*^9}, 3.728751619944504*^9, 
   3.728813264274679*^9, 3.728827381326201*^9, 3.7288430592053967`*^9, 
   3.728845624172392*^9, 3.733367277519153*^9, 3.790862660646812*^9, 
   3.790867115247067*^9, 3.7922732168945007`*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->226753454,ExpressionUUID->"c4c06884-8f62-47df-be2f-38528d36d8a9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciPDblues"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Ricci tensor", 
   " ", $CellContext`RicciPDblues],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645939324555*^9, 3.728660585841178*^9, 3.728671095157638*^9, 
   3.728674880967157*^9, 3.728732879161681*^9, 3.728733686924255*^9, 
   3.728742318008175*^9, 3.7287432947505503`*^9, 3.728743393197723*^9, 
   3.728747237262712*^9, 3.7287494745303164`*^9, 3.7287499348201714`*^9, 
   3.728750085987439*^9, 3.7287501915322237`*^9, 3.7287503020600357`*^9, 
   3.728750520800502*^9, 3.728750580123959*^9, 3.72875072434544*^9, {
   3.7287509844380207`*^9, 3.728751006976675*^9}, 3.728751619944504*^9, 
   3.728813264274679*^9, 3.728827381326201*^9, 3.7288430592053967`*^9, 
   3.728845624172392*^9, 3.733367277519153*^9, 3.790862660646812*^9, 
   3.790867115247067*^9, 3.792273216901402*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->1409311600,ExpressionUUID->"06cf6d5b-0259-4d12-b4c3-f9e9f5001c9f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannPDblues"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannPDblues],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645939324555*^9, 3.728660585841178*^9, 3.728671095157638*^9, 
   3.728674880967157*^9, 3.728732879161681*^9, 3.728733686924255*^9, 
   3.728742318008175*^9, 3.7287432947505503`*^9, 3.728743393197723*^9, 
   3.728747237262712*^9, 3.7287494745303164`*^9, 3.7287499348201714`*^9, 
   3.728750085987439*^9, 3.7287501915322237`*^9, 3.7287503020600357`*^9, 
   3.728750520800502*^9, 3.728750580123959*^9, 3.72875072434544*^9, {
   3.7287509844380207`*^9, 3.728751006976675*^9}, 3.728751619944504*^9, 
   3.728813264274679*^9, 3.728827381326201*^9, 3.7288430592053967`*^9, 
   3.728845624172392*^9, 3.733367277519153*^9, 3.790862660646812*^9, 
   3.790867115247067*^9, 3.792273216908408*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->1702918437,ExpressionUUID->"6d55ee9c-dee6-4f87-9475-0d996d30c3bf"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TorsionPDblues"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionPDblues],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645939324555*^9, 3.728660585841178*^9, 3.728671095157638*^9, 
   3.728674880967157*^9, 3.728732879161681*^9, 3.728733686924255*^9, 
   3.728742318008175*^9, 3.7287432947505503`*^9, 3.728743393197723*^9, 
   3.728747237262712*^9, 3.7287494745303164`*^9, 3.7287499348201714`*^9, 
   3.728750085987439*^9, 3.7287501915322237`*^9, 3.7287503020600357`*^9, 
   3.728750520800502*^9, 3.728750580123959*^9, 3.72875072434544*^9, {
   3.7287509844380207`*^9, 3.728751006976675*^9}, 3.728751619944504*^9, 
   3.728813264274679*^9, 3.728827381326201*^9, 3.7288430592053967`*^9, 
   3.728845624172392*^9, 3.733367277519153*^9, 3.790862660646812*^9, 
   3.790867115247067*^9, 3.792273216915245*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->1435607090,ExpressionUUID->"3aa4c10f-7e90-48df-88dd-8b44919948ef"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"parallel derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "PDblues"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "parallel derivative", 
   " ", $CellContext`PDblues],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645939324555*^9, 3.728660585841178*^9, 3.728671095157638*^9, 
   3.728674880967157*^9, 3.728732879161681*^9, 3.728733686924255*^9, 
   3.728742318008175*^9, 3.7287432947505503`*^9, 3.728743393197723*^9, 
   3.728747237262712*^9, 3.7287494745303164`*^9, 3.7287499348201714`*^9, 
   3.728750085987439*^9, 3.7287501915322237`*^9, 3.7287503020600357`*^9, 
   3.728750520800502*^9, 3.728750580123959*^9, 3.72875072434544*^9, {
   3.7287509844380207`*^9, 3.728751006976675*^9}, 3.728751619944504*^9, 
   3.728813264274679*^9, 3.728827381326201*^9, 3.7288430592053967`*^9, 
   3.728845624172392*^9, 3.733367277519153*^9, 3.790862660646812*^9, 
   3.790867115247067*^9, 3.7922732169222193`*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->1269587585,ExpressionUUID->"a91eee8c-adc2-438b-aeea-29b261414380"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "etaUpblues"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric +1 density",
    " ", $CellContext`etaUpblues],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645939324555*^9, 3.728660585841178*^9, 3.728671095157638*^9, 
   3.728674880967157*^9, 3.728732879161681*^9, 3.728733686924255*^9, 
   3.728742318008175*^9, 3.7287432947505503`*^9, 3.728743393197723*^9, 
   3.728747237262712*^9, 3.7287494745303164`*^9, 3.7287499348201714`*^9, 
   3.728750085987439*^9, 3.7287501915322237`*^9, 3.7287503020600357`*^9, 
   3.728750520800502*^9, 3.728750580123959*^9, 3.72875072434544*^9, {
   3.7287509844380207`*^9, 3.728751006976675*^9}, 3.728751619944504*^9, 
   3.728813264274679*^9, 3.728827381326201*^9, 3.7288430592053967`*^9, 
   3.728845624172392*^9, 3.733367277519153*^9, 3.790862660646812*^9, 
   3.790867115247067*^9, 3.7922732169293528`*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->561024503,ExpressionUUID->"1edd1ea1-3722-4b18-b71c-67793ef58a57"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "etaDownblues"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric -1 density",
    " ", $CellContext`etaDownblues],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645939324555*^9, 3.728660585841178*^9, 3.728671095157638*^9, 
   3.728674880967157*^9, 3.728732879161681*^9, 3.728733686924255*^9, 
   3.728742318008175*^9, 3.7287432947505503`*^9, 3.728743393197723*^9, 
   3.728747237262712*^9, 3.7287494745303164`*^9, 3.7287499348201714`*^9, 
   3.728750085987439*^9, 3.7287501915322237`*^9, 3.7287503020600357`*^9, 
   3.728750520800502*^9, 3.728750580123959*^9, 3.72875072434544*^9, {
   3.7287509844380207`*^9, 3.728751006976675*^9}, 3.728751619944504*^9, 
   3.728813264274679*^9, 3.728827381326201*^9, 3.7288430592053967`*^9, 
   3.728845624172392*^9, 3.733367277519153*^9, 3.790862660646812*^9, 
   3.790867115247067*^9, 3.792273216936311*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->1685611591,ExpressionUUID->"01f1e567-7657-4842-84b1-3c72151fde10"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefBasis", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"basis\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "blues"}],
  SequenceForm[
  "** ", xAct`xCoba`UndefBasis, ": Undefined ", "basis", 
   " ", $CellContext`blues],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645939324555*^9, 3.728660585841178*^9, 3.728671095157638*^9, 
   3.728674880967157*^9, 3.728732879161681*^9, 3.728733686924255*^9, 
   3.728742318008175*^9, 3.7287432947505503`*^9, 3.728743393197723*^9, 
   3.728747237262712*^9, 3.7287494745303164`*^9, 3.7287499348201714`*^9, 
   3.728750085987439*^9, 3.7287501915322237`*^9, 3.7287503020600357`*^9, 
   3.728750520800502*^9, 3.728750580123959*^9, 3.72875072434544*^9, {
   3.7287509844380207`*^9, 3.728751006976675*^9}, 3.728751619944504*^9, 
   3.728813264274679*^9, 3.728827381326201*^9, 3.7288430592053967`*^9, 
   3.728845624172392*^9, 3.733367277519153*^9, 3.790862660646812*^9, 
   3.790867115247067*^9, 3.7922732169431887`*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->469738713,ExpressionUUID->"c9d2e7e6-026e-4677-970f-b7095935672b"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Removing a complex basis also removes its complex conjugate:", \
"MathCaption",ExpressionUUID->"4fbaadb2-6371-4cd5-811a-1b7897c8e955"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"UndefBasis", "[", "fcomp", "]"}], ";"}]], "Input",
 CellChangeTimes->{{3.728733690697093*^9, 3.728733690837076*^9}, {
  3.7287472403500013`*^9, 3.728747241651298*^9}},
 CellLabel->"In[60]:=",ExpressionUUID->"a590ad3f-c9d9-49c4-bf1d-7488a8bfff9d"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"nonsymmetric AChristoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "AChristoffelPDfcomp\[Dagger]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "nonsymmetric AChristoffel tensor", 
   " ", $CellContext`AChristoffelPDfcomp\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645940536729*^9, 3.728660586170697*^9, 3.7286710969258337`*^9, 
   3.728674881732094*^9, 3.728732882198947*^9, 3.72873369130713*^9, 
   3.728742318595891*^9, 3.728743295627975*^9, 3.72874339368121*^9, 
   3.728747243015862*^9, 3.72874947617679*^9, 3.7287499351204863`*^9, 
   3.728750086642294*^9, 3.728750192197423*^9, 3.7287503026850367`*^9, 
   3.7287505211625357`*^9, 3.728750580457086*^9, 3.728750724629656*^9, {
   3.728750984756345*^9, 3.728751007258465*^9}, 3.7287516202781153`*^9, 
   3.728813264727947*^9, 3.728827381808064*^9, 3.728843061102787*^9, 
   3.728845624726934*^9, 3.733367279436389*^9, 3.7908626620637217`*^9, 
   3.790867116622571*^9, 3.7922732182599154`*^9},
 CellLabel->"During evaluation of In[60]:=",
 CellID->908849206,ExpressionUUID->"58c7e838-4903-439e-9fb2-a114e17da76c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"nonsymmetric AChristoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "AChristoffelPDfcomp"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "nonsymmetric AChristoffel tensor", " ", $CellContext`AChristoffelPDfcomp],
  
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645940536729*^9, 3.728660586170697*^9, 3.7286710969258337`*^9, 
   3.728674881732094*^9, 3.728732882198947*^9, 3.72873369130713*^9, 
   3.728742318595891*^9, 3.728743295627975*^9, 3.72874339368121*^9, 
   3.728747243015862*^9, 3.72874947617679*^9, 3.7287499351204863`*^9, 
   3.728750086642294*^9, 3.728750192197423*^9, 3.7287503026850367`*^9, 
   3.7287505211625357`*^9, 3.728750580457086*^9, 3.728750724629656*^9, {
   3.728750984756345*^9, 3.728751007258465*^9}, 3.7287516202781153`*^9, 
   3.728813264727947*^9, 3.728827381808064*^9, 3.728843061102787*^9, 
   3.728845624726934*^9, 3.733367279436389*^9, 3.7908626620637217`*^9, 
   3.790867116622571*^9, 3.792273218270548*^9},
 CellLabel->"During evaluation of In[60]:=",
 CellID->2124907778,ExpressionUUID->"75e6d2cd-bc1d-4eb4-819c-bfabfc64903d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "ChristoffelPDfcomp"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "non-symmetric Christoffel tensor", " ", $CellContext`ChristoffelPDfcomp],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645940536729*^9, 3.728660586170697*^9, 3.7286710969258337`*^9, 
   3.728674881732094*^9, 3.728732882198947*^9, 3.72873369130713*^9, 
   3.728742318595891*^9, 3.728743295627975*^9, 3.72874339368121*^9, 
   3.728747243015862*^9, 3.72874947617679*^9, 3.7287499351204863`*^9, 
   3.728750086642294*^9, 3.728750192197423*^9, 3.7287503026850367`*^9, 
   3.7287505211625357`*^9, 3.728750580457086*^9, 3.728750724629656*^9, {
   3.728750984756345*^9, 3.728751007258465*^9}, 3.7287516202781153`*^9, 
   3.728813264727947*^9, 3.728827381808064*^9, 3.728843061102787*^9, 
   3.728845624726934*^9, 3.733367279436389*^9, 3.7908626620637217`*^9, 
   3.790867116622571*^9, 3.792273218285263*^9},
 CellLabel->"During evaluation of In[60]:=",
 CellID->671975236,ExpressionUUID->"fd077c86-fbf7-4501-adfc-a0a64b2aeb62"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"FRiemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "FRiemannPDfcomp\[Dagger]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "FRiemann tensor", 
   " ", $CellContext`FRiemannPDfcomp\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645940536729*^9, 3.728660586170697*^9, 3.7286710969258337`*^9, 
   3.728674881732094*^9, 3.728732882198947*^9, 3.72873369130713*^9, 
   3.728742318595891*^9, 3.728743295627975*^9, 3.72874339368121*^9, 
   3.728747243015862*^9, 3.72874947617679*^9, 3.7287499351204863`*^9, 
   3.728750086642294*^9, 3.728750192197423*^9, 3.7287503026850367`*^9, 
   3.7287505211625357`*^9, 3.728750580457086*^9, 3.728750724629656*^9, {
   3.728750984756345*^9, 3.728751007258465*^9}, 3.7287516202781153`*^9, 
   3.728813264727947*^9, 3.728827381808064*^9, 3.728843061102787*^9, 
   3.728845624726934*^9, 3.733367279436389*^9, 3.7908626620637217`*^9, 
   3.790867116622571*^9, 3.792273218292466*^9},
 CellLabel->"During evaluation of In[60]:=",
 CellID->271291879,ExpressionUUID->"9a71faa4-0069-4eca-ac0a-35762dfdbd84"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"FRiemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "FRiemannPDfcomp"}],
  
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "FRiemann tensor", 
   " ", $CellContext`FRiemannPDfcomp],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645940536729*^9, 3.728660586170697*^9, 3.7286710969258337`*^9, 
   3.728674881732094*^9, 3.728732882198947*^9, 3.72873369130713*^9, 
   3.728742318595891*^9, 3.728743295627975*^9, 3.72874339368121*^9, 
   3.728747243015862*^9, 3.72874947617679*^9, 3.7287499351204863`*^9, 
   3.728750086642294*^9, 3.728750192197423*^9, 3.7287503026850367`*^9, 
   3.7287505211625357`*^9, 3.728750580457086*^9, 3.728750724629656*^9, {
   3.728750984756345*^9, 3.728751007258465*^9}, 3.7287516202781153`*^9, 
   3.728813264727947*^9, 3.728827381808064*^9, 3.728843061102787*^9, 
   3.728845624726934*^9, 3.733367279436389*^9, 3.7908626620637217`*^9, 
   3.790867116622571*^9, 3.792273218299695*^9},
 CellLabel->"During evaluation of In[60]:=",
 CellID->494827772,ExpressionUUID->"280b1ad0-faba-4ec0-9170-733c59f08bde"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciPDfcomp"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Ricci tensor", 
   " ", $CellContext`RicciPDfcomp],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645940536729*^9, 3.728660586170697*^9, 3.7286710969258337`*^9, 
   3.728674881732094*^9, 3.728732882198947*^9, 3.72873369130713*^9, 
   3.728742318595891*^9, 3.728743295627975*^9, 3.72874339368121*^9, 
   3.728747243015862*^9, 3.72874947617679*^9, 3.7287499351204863`*^9, 
   3.728750086642294*^9, 3.728750192197423*^9, 3.7287503026850367`*^9, 
   3.7287505211625357`*^9, 3.728750580457086*^9, 3.728750724629656*^9, {
   3.728750984756345*^9, 3.728751007258465*^9}, 3.7287516202781153`*^9, 
   3.728813264727947*^9, 3.728827381808064*^9, 3.728843061102787*^9, 
   3.728845624726934*^9, 3.733367279436389*^9, 3.7908626620637217`*^9, 
   3.790867116622571*^9, 3.7922732183073063`*^9},
 CellLabel->"During evaluation of In[60]:=",
 CellID->1513682820,ExpressionUUID->"7916fa80-7af1-46ba-a398-be16b67faa9c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannPDfcomp"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannPDfcomp],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645940536729*^9, 3.728660586170697*^9, 3.7286710969258337`*^9, 
   3.728674881732094*^9, 3.728732882198947*^9, 3.72873369130713*^9, 
   3.728742318595891*^9, 3.728743295627975*^9, 3.72874339368121*^9, 
   3.728747243015862*^9, 3.72874947617679*^9, 3.7287499351204863`*^9, 
   3.728750086642294*^9, 3.728750192197423*^9, 3.7287503026850367`*^9, 
   3.7287505211625357`*^9, 3.728750580457086*^9, 3.728750724629656*^9, {
   3.728750984756345*^9, 3.728751007258465*^9}, 3.7287516202781153`*^9, 
   3.728813264727947*^9, 3.728827381808064*^9, 3.728843061102787*^9, 
   3.728845624726934*^9, 3.733367279436389*^9, 3.7908626620637217`*^9, 
   3.790867116622571*^9, 3.792273218314466*^9},
 CellLabel->"During evaluation of In[60]:=",
 CellID->1375873378,ExpressionUUID->"ac404752-7bb4-45f2-bd4f-e13dce0e2894"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TorsionPDfcomp"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionPDfcomp],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645940536729*^9, 3.728660586170697*^9, 3.7286710969258337`*^9, 
   3.728674881732094*^9, 3.728732882198947*^9, 3.72873369130713*^9, 
   3.728742318595891*^9, 3.728743295627975*^9, 3.72874339368121*^9, 
   3.728747243015862*^9, 3.72874947617679*^9, 3.7287499351204863`*^9, 
   3.728750086642294*^9, 3.728750192197423*^9, 3.7287503026850367`*^9, 
   3.7287505211625357`*^9, 3.728750580457086*^9, 3.728750724629656*^9, {
   3.728750984756345*^9, 3.728751007258465*^9}, 3.7287516202781153`*^9, 
   3.728813264727947*^9, 3.728827381808064*^9, 3.728843061102787*^9, 
   3.728845624726934*^9, 3.733367279436389*^9, 3.7908626620637217`*^9, 
   3.790867116622571*^9, 3.7922732183215723`*^9},
 CellLabel->"During evaluation of In[60]:=",
 CellID->199953150,ExpressionUUID->"08e5e904-64cf-4446-8929-9d98c375bfe9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"parallel derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "PDfcomp"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "parallel derivative", 
   " ", $CellContext`PDfcomp],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645940536729*^9, 3.728660586170697*^9, 3.7286710969258337`*^9, 
   3.728674881732094*^9, 3.728732882198947*^9, 3.72873369130713*^9, 
   3.728742318595891*^9, 3.728743295627975*^9, 3.72874339368121*^9, 
   3.728747243015862*^9, 3.72874947617679*^9, 3.7287499351204863`*^9, 
   3.728750086642294*^9, 3.728750192197423*^9, 3.7287503026850367`*^9, 
   3.7287505211625357`*^9, 3.728750580457086*^9, 3.728750724629656*^9, {
   3.728750984756345*^9, 3.728751007258465*^9}, 3.7287516202781153`*^9, 
   3.728813264727947*^9, 3.728827381808064*^9, 3.728843061102787*^9, 
   3.728845624726934*^9, 3.733367279436389*^9, 3.7908626620637217`*^9, 
   3.790867116622571*^9, 3.792273218328581*^9},
 CellLabel->"During evaluation of In[60]:=",
 CellID->1948168142,ExpressionUUID->"135da206-eff4-4f71-9c57-c702f6732246"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefBasis", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"basis\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "fcomp\[Dagger]"}],
  SequenceForm[
  "** ", xAct`xCoba`UndefBasis, ": Undefined ", "basis", 
   " ", $CellContext`fcomp\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645940536729*^9, 3.728660586170697*^9, 3.7286710969258337`*^9, 
   3.728674881732094*^9, 3.728732882198947*^9, 3.72873369130713*^9, 
   3.728742318595891*^9, 3.728743295627975*^9, 3.72874339368121*^9, 
   3.728747243015862*^9, 3.72874947617679*^9, 3.7287499351204863`*^9, 
   3.728750086642294*^9, 3.728750192197423*^9, 3.7287503026850367`*^9, 
   3.7287505211625357`*^9, 3.728750580457086*^9, 3.728750724629656*^9, {
   3.728750984756345*^9, 3.728751007258465*^9}, 3.7287516202781153`*^9, 
   3.728813264727947*^9, 3.728827381808064*^9, 3.728843061102787*^9, 
   3.728845624726934*^9, 3.733367279436389*^9, 3.7908626620637217`*^9, 
   3.790867116622571*^9, 3.792273218335402*^9},
 CellLabel->"During evaluation of In[60]:=",
 CellID->178796785,ExpressionUUID->"a335989a-acd0-4c4e-a100-aa9806c2f73b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "etaUpfcomp\[Dagger]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric +1 density",
    " ", $CellContext`etaUpfcomp\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645940536729*^9, 3.728660586170697*^9, 3.7286710969258337`*^9, 
   3.728674881732094*^9, 3.728732882198947*^9, 3.72873369130713*^9, 
   3.728742318595891*^9, 3.728743295627975*^9, 3.72874339368121*^9, 
   3.728747243015862*^9, 3.72874947617679*^9, 3.7287499351204863`*^9, 
   3.728750086642294*^9, 3.728750192197423*^9, 3.7287503026850367`*^9, 
   3.7287505211625357`*^9, 3.728750580457086*^9, 3.728750724629656*^9, {
   3.728750984756345*^9, 3.728751007258465*^9}, 3.7287516202781153`*^9, 
   3.728813264727947*^9, 3.728827381808064*^9, 3.728843061102787*^9, 
   3.728845624726934*^9, 3.733367279436389*^9, 3.7908626620637217`*^9, 
   3.790867116622571*^9, 3.79227321834212*^9},
 CellLabel->"During evaluation of In[60]:=",
 CellID->707943342,ExpressionUUID->"aed33cd4-05ac-4471-965f-d8ba2f029330"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "etaUpfcomp"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric +1 density",
    " ", $CellContext`etaUpfcomp],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645940536729*^9, 3.728660586170697*^9, 3.7286710969258337`*^9, 
   3.728674881732094*^9, 3.728732882198947*^9, 3.72873369130713*^9, 
   3.728742318595891*^9, 3.728743295627975*^9, 3.72874339368121*^9, 
   3.728747243015862*^9, 3.72874947617679*^9, 3.7287499351204863`*^9, 
   3.728750086642294*^9, 3.728750192197423*^9, 3.7287503026850367`*^9, 
   3.7287505211625357`*^9, 3.728750580457086*^9, 3.728750724629656*^9, {
   3.728750984756345*^9, 3.728751007258465*^9}, 3.7287516202781153`*^9, 
   3.728813264727947*^9, 3.728827381808064*^9, 3.728843061102787*^9, 
   3.728845624726934*^9, 3.733367279436389*^9, 3.7908626620637217`*^9, 
   3.790867116622571*^9, 3.792273218349063*^9},
 CellLabel->"During evaluation of In[60]:=",
 CellID->1344344614,ExpressionUUID->"a9132ed3-6fb9-47c3-932e-719611ee79da"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "etaDownfcomp\[Dagger]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric -1 density",
    " ", $CellContext`etaDownfcomp\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645940536729*^9, 3.728660586170697*^9, 3.7286710969258337`*^9, 
   3.728674881732094*^9, 3.728732882198947*^9, 3.72873369130713*^9, 
   3.728742318595891*^9, 3.728743295627975*^9, 3.72874339368121*^9, 
   3.728747243015862*^9, 3.72874947617679*^9, 3.7287499351204863`*^9, 
   3.728750086642294*^9, 3.728750192197423*^9, 3.7287503026850367`*^9, 
   3.7287505211625357`*^9, 3.728750580457086*^9, 3.728750724629656*^9, {
   3.728750984756345*^9, 3.728751007258465*^9}, 3.7287516202781153`*^9, 
   3.728813264727947*^9, 3.728827381808064*^9, 3.728843061102787*^9, 
   3.728845624726934*^9, 3.733367279436389*^9, 3.7908626620637217`*^9, 
   3.790867116622571*^9, 3.792273218355761*^9},
 CellLabel->"During evaluation of In[60]:=",
 CellID->724477411,ExpressionUUID->"7acfa328-d399-4e22-b87f-326da21303c5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "etaDownfcomp"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric -1 density",
    " ", $CellContext`etaDownfcomp],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645940536729*^9, 3.728660586170697*^9, 3.7286710969258337`*^9, 
   3.728674881732094*^9, 3.728732882198947*^9, 3.72873369130713*^9, 
   3.728742318595891*^9, 3.728743295627975*^9, 3.72874339368121*^9, 
   3.728747243015862*^9, 3.72874947617679*^9, 3.7287499351204863`*^9, 
   3.728750086642294*^9, 3.728750192197423*^9, 3.7287503026850367`*^9, 
   3.7287505211625357`*^9, 3.728750580457086*^9, 3.728750724629656*^9, {
   3.728750984756345*^9, 3.728751007258465*^9}, 3.7287516202781153`*^9, 
   3.728813264727947*^9, 3.728827381808064*^9, 3.728843061102787*^9, 
   3.728845624726934*^9, 3.733367279436389*^9, 3.7908626620637217`*^9, 
   3.790867116622571*^9, 3.7922732183625937`*^9},
 CellLabel->"During evaluation of In[60]:=",
 CellID->59568187,ExpressionUUID->"d781ce1f-bf37-42fc-9cae-fb17be06238d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefBasis", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"basis\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "fcomp"}],
  SequenceForm[
  "** ", xAct`xCoba`UndefBasis, ": Undefined ", "basis", 
   " ", $CellContext`fcomp],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728645940536729*^9, 3.728660586170697*^9, 3.7286710969258337`*^9, 
   3.728674881732094*^9, 3.728732882198947*^9, 3.72873369130713*^9, 
   3.728742318595891*^9, 3.728743295627975*^9, 3.72874339368121*^9, 
   3.728747243015862*^9, 3.72874947617679*^9, 3.7287499351204863`*^9, 
   3.728750086642294*^9, 3.728750192197423*^9, 3.7287503026850367`*^9, 
   3.7287505211625357`*^9, 3.728750580457086*^9, 3.728750724629656*^9, {
   3.728750984756345*^9, 3.728751007258465*^9}, 3.7287516202781153`*^9, 
   3.728813264727947*^9, 3.728827381808064*^9, 3.728843061102787*^9, 
   3.728845624726934*^9, 3.733367279436389*^9, 3.7908626620637217`*^9, 
   3.790867116622571*^9, 3.792273218369534*^9},
 CellLabel->"During evaluation of In[60]:=",
 CellID->435139407,ExpressionUUID->"f3fb8537-e3a5-4b4e-bb32-66165ebb774d"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"UndefTensor", "/@", 
   RowBox[{"{", 
    RowBox[{"R", ",", "T", ",", "u", ",", "v"}], "}"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.728750107915875*^9, 3.728750156061103*^9}},
 CellLabel->"In[61]:=",
 CellID->891057657,ExpressionUUID->"624217e7-ddf7-407e-b693-f792e56ca870"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "R\[Dagger]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`R\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728750122635363*^9, 3.728750156557139*^9, 3.7287501925812273`*^9, 
   3.728750303300208*^9, 3.728750521528466*^9, 3.7287505807593527`*^9, 
   3.728750724928062*^9, {3.728750985058617*^9, 3.7287510075418243`*^9}, 
   3.728751620777418*^9, 3.728813265182197*^9, 3.728827382390902*^9, 
   3.7288430631029367`*^9, 3.728845625160882*^9, 3.733367281102833*^9, 
   3.790862663395739*^9, 3.790867117657794*^9, 3.792273219749674*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->457359816,ExpressionUUID->"72b19dfe-b5bc-4474-a567-bcf8a3ea3795"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "R"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`R],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728750122635363*^9, 3.728750156557139*^9, 3.7287501925812273`*^9, 
   3.728750303300208*^9, 3.728750521528466*^9, 3.7287505807593527`*^9, 
   3.728750724928062*^9, {3.728750985058617*^9, 3.7287510075418243`*^9}, 
   3.728751620777418*^9, 3.728813265182197*^9, 3.728827382390902*^9, 
   3.7288430631029367`*^9, 3.728845625160882*^9, 3.733367281102833*^9, 
   3.790862663395739*^9, 3.790867117657794*^9, 3.79227321976206*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->1002454899,ExpressionUUID->"d70573e5-0ebd-41ec-98be-4d74fa53556a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728750122635363*^9, 3.728750156557139*^9, 3.7287501925812273`*^9, 
   3.728750303300208*^9, 3.728750521528466*^9, 3.7287505807593527`*^9, 
   3.728750724928062*^9, {3.728750985058617*^9, 3.7287510075418243`*^9}, 
   3.728751620777418*^9, 3.728813265182197*^9, 3.728827382390902*^9, 
   3.7288430631029367`*^9, 3.728845625160882*^9, 3.733367281102833*^9, 
   3.790862663395739*^9, 3.790867117657794*^9, 3.792273219776425*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->1250276778,ExpressionUUID->"95773e13-5f2d-4533-80ed-72963d3056a2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "u\[Dagger]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`u\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728750122635363*^9, 3.728750156557139*^9, 3.7287501925812273`*^9, 
   3.728750303300208*^9, 3.728750521528466*^9, 3.7287505807593527`*^9, 
   3.728750724928062*^9, {3.728750985058617*^9, 3.7287510075418243`*^9}, 
   3.728751620777418*^9, 3.728813265182197*^9, 3.728827382390902*^9, 
   3.7288430631029367`*^9, 3.728845625160882*^9, 3.733367281102833*^9, 
   3.790862663395739*^9, 3.790867117657794*^9, 3.792273219783746*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->274321951,ExpressionUUID->"4a7d9b88-d68a-477f-9e07-98b40b2f9e9c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "u"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`u],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728750122635363*^9, 3.728750156557139*^9, 3.7287501925812273`*^9, 
   3.728750303300208*^9, 3.728750521528466*^9, 3.7287505807593527`*^9, 
   3.728750724928062*^9, {3.728750985058617*^9, 3.7287510075418243`*^9}, 
   3.728751620777418*^9, 3.728813265182197*^9, 3.728827382390902*^9, 
   3.7288430631029367`*^9, 3.728845625160882*^9, 3.733367281102833*^9, 
   3.790862663395739*^9, 3.790867117657794*^9, 3.792273219791134*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->2029123995,ExpressionUUID->"61b7f67c-4cc6-4d40-93cb-0ad1c804ea4e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "v"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`v],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728750122635363*^9, 3.728750156557139*^9, 3.7287501925812273`*^9, 
   3.728750303300208*^9, 3.728750521528466*^9, 3.7287505807593527`*^9, 
   3.728750724928062*^9, {3.728750985058617*^9, 3.7287510075418243`*^9}, 
   3.728751620777418*^9, 3.728813265182197*^9, 3.728827382390902*^9, 
   3.7288430631029367`*^9, 3.728845625160882*^9, 3.733367281102833*^9, 
   3.790862663395739*^9, 3.790867117657794*^9, 3.792273219798318*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->1446669605,ExpressionUUID->"cb85afb9-2b5e-489e-bd3c-73d1ef5cc3ce"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefVBundle", "[", "InnerC", "]"}]], "Input",
 CellChangeTimes->{{3.7287500333255167`*^9, 3.728750042034931*^9}},
 CellLabel->"In[62]:=",
 CellID->1401790690,ExpressionUUID->"caf0ef7e-cbbb-4142-809d-3a6f314d6837"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefVBundle", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"conjugated vbundle\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["InnerC\[Dagger]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`InnerC\[Dagger],
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefVBundle, ": Undefined ", "conjugated vbundle", 
   " ", $CellContext`InnerC\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728750133502331*^9, 3.728750193124095*^9, 3.728750303771426*^9, 
   3.728750521805451*^9, 3.728750581037147*^9, 3.728750725163906*^9, {
   3.728750985386902*^9, 3.728751007790182*^9}, 3.728751621094442*^9, 
   3.728813265533646*^9, 3.728827382736196*^9, 3.728843064817408*^9, 
   3.728845625474742*^9, 3.733367282462525*^9, 3.790862665019261*^9, 
   3.7908671187294807`*^9, 3.792273221245615*^9},
 CellLabel->"During evaluation of In[62]:=",
 CellID->1941333954,ExpressionUUID->"6dfdefbf-aa77-42dd-a57a-1ab493d7024b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefVBundle", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["InnerC",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`InnerC,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefVBundle, ": Undefined ", "vbundle", 
   " ", $CellContext`InnerC],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728750133502331*^9, 3.728750193124095*^9, 3.728750303771426*^9, 
   3.728750521805451*^9, 3.728750581037147*^9, 3.728750725163906*^9, {
   3.728750985386902*^9, 3.728751007790182*^9}, 3.728751621094442*^9, 
   3.728813265533646*^9, 3.728827382736196*^9, 3.728843064817408*^9, 
   3.728845625474742*^9, 3.733367282462525*^9, 3.790862665019261*^9, 
   3.7908671187294807`*^9, 3.7922732213076887`*^9},
 CellLabel->"During evaluation of In[62]:=",
 CellID->1280716007,ExpressionUUID->"59a72445-1349-44df-a013-dcd667e416a6"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M3", "]"}]], "Input",
 CellChangeTimes->{{3.728750135437042*^9, 3.728750140151712*^9}},
 CellLabel->"In[63]:=",
 CellID->1216022970,ExpressionUUID->"014fba38-4fd9-4108-a282-a171ba8b66e5"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefVBundle", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]M3",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentM3,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefVBundle, ": Undefined ", "vbundle", 
   " ", $CellContext`TangentM3],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728750162341569*^9, 3.728750193468371*^9, 3.728750304321321*^9, 
   3.72875052217933*^9, 3.7287505813532248`*^9, 3.728750725472004*^9, {
   3.7287509857227087`*^9, 3.728751008074291*^9}, 3.728751621364987*^9, 
   3.728813265886754*^9, 3.728827383053286*^9, 3.7288430655928698`*^9, 
   3.7288456257819366`*^9, 3.73336728327675*^9, 3.7908626659023037`*^9, 
   3.79086711933471*^9, 3.792273222028212*^9},
 CellLabel->"During evaluation of In[63]:=",
 CellID->2059295894,ExpressionUUID->"8dc4848f-120b-4c67-8a45-21cc4bda9bed"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefManifold", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"manifold\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "M3"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefManifold, ": Undefined ", "manifold", 
   " ", $CellContext`M3],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.728750162341569*^9, 3.728750193468371*^9, 3.728750304321321*^9, 
   3.72875052217933*^9, 3.7287505813532248`*^9, 3.728750725472004*^9, {
   3.7287509857227087`*^9, 3.728751008074291*^9}, 3.728751621364987*^9, 
   3.728813265886754*^9, 3.728827383053286*^9, 3.7288430655928698`*^9, 
   3.7288456257819366`*^9, 3.73336728327675*^9, 3.7908626659023037`*^9, 
   3.79086711933471*^9, 3.792273222038924*^9},
 CellLabel->"During evaluation of In[63]:=",
 CellID->1739635406,ExpressionUUID->"b86d2204-c8b1-4aab-a275-3b6bc4ddf341"]
}, Open  ]]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{
  "basis", ",", "cobasis", ",", "expr", ",", "cexpr", ",", "values", ",", "i",
    ",", "j", ",", "w"}], "]"}]], "Input",
 CellChangeTimes->{{3.7287502118481293`*^9, 3.72875021828246*^9}, {
  3.7287502530488234`*^9, 3.728750256974098*^9}},
 CellLabel->"In[64]:=",
 CellID->94413090,ExpressionUUID->"5241413d-cefc-4d83-9d9f-25c95dd98a30"]
}, Open  ]]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{

Cell["1. Algebra and bases", "Section",
 CellFrame->{{0, 0}, {0, 0.5}},
 CellChangeTimes->{{3.7286627354220657`*^9, 3.728662748172317*^9}, {
  3.728664037492599*^9, 3.72866403885896*^9}, {3.7286704747855062`*^9, 
  3.728670480415387*^9}, {3.728670889883876*^9, 3.728670893540101*^9}, {
  3.728673410238123*^9, 3.728673425886174*^9}},
 FontSize->18,
 FontColor->GrayLevel[0],
 CellTags->{"S5.18.1", "1.4"},
 CellID->853073756,ExpressionUUID->"9f2804d8-0164-470b-a13e-c7b68c3d7339"],

Cell[CellGroupData[{

Cell["1.1. Bases", "Subsection",
 CellChangeTimes->{{3.7286627354220657`*^9, 3.728662748172317*^9}, {
  3.728673412590672*^9, 3.728673427462188*^9}},
 CellTags->{
  "S5.18.1", "1.4"},ExpressionUUID->"86385e57-5f81-479d-b8c0-73ad6467ad57"],

Cell[CellGroupData[{

Cell["1.1.1. DefBasis", "Subsubsection",
 CellChangeTimes->{{3.728673437021533*^9, 
  3.72867343714992*^9}},ExpressionUUID->"a8c121ed-a412-44f3-a13a-\
a9fd7d2138af"],

Cell["\<\
This section will explain how to define and work with new bases, without \
assuming that they have an underlying coordinate chart. We will focus on the \
algebraic manipulations, ignoring the fact that our basic objects are \
actually fields, except for a few examples involving differentiation. Later \
on in this notebook, we will show how to work with charts, and how to perform \
derivatives of several kinds.\
\>", "Text",
 CellChangeTimes->{{3.728671118769095*^9, 3.728671210301458*^9}},
 CellTags->{
  "S5.18.1", "1.5"},ExpressionUUID->"6ea21273-ba95-4a75-a84b-683255bd3833"],

Cell[TextData[{
 StyleBox["DefBasis\t\t\t", "DisplayMath"],
 StyleBox["Define a basis", "Text"],
 StyleBox["\nVBundleOfBasis\t\t", "DisplayMath"],
 StyleBox["Vector bundle on which a basis lives\n", "Text"],
 StyleBox["PDOfBasis\t\t\t", "DisplayMath"],
 StyleBox["Parallel derivative associated to a given basis\n", "Text"],
 StyleBox["$Bases\t\t\t\t", "DisplayMath"],
 StyleBox["List of currently defined bases\n", "Text"],
 StyleBox["BasisQ\t\t\t\t", "DisplayMath"],
 StyleBox["Check existence of a given basis name", "Text"]
}], "DefinitionBox",
 CellChangeTimes->{{3.7287423824023037`*^9, 
  3.728742388520258*^9}},ExpressionUUID->"a204ebe7-8a70-466f-b93e-\
31f2286ed32e"],

Cell["Definition of a basis.", "Text",ExpressionUUID->"061818b3-e6ff-40f8-b9bb-0589f86c83f6"],

Cell[CellGroupData[{

Cell["Declare a manifold and some tensors:", "MathCaption",
 CellChangeTimes->{{3.728750346653975*^9, 3.728750350652412*^9}, {
  3.7287505907691298`*^9, 3.728750594465168*^9}},
 CellID->1431971475,ExpressionUUID->"f33a8b9b-febb-4147-996a-db46f665cd76"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M3", ",", "3", ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f"}], "}"}]}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.728750351919133*^9, 3.728750377187203*^9}},
 CellLabel->"In[65]:=",
 CellID->571909325,ExpressionUUID->"ba46731a-9074-4342-a71d-f8619fbc0b8d"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefManifold", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"manifold \"\>", "\[InvisibleSpace]", "M3", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefManifold, ": Defining ", 
   "manifold ", $CellContext`M3, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287503777488403`*^9, 3.728750522525712*^9, 
  3.728750585714631*^9, 3.728750759631687*^9, 3.7287510437107277`*^9, 
  3.728751621587109*^9, 3.7287593548827887`*^9, 3.7288132661487617`*^9, 
  3.72882738334347*^9, 3.728835974288519*^9, 3.728836291877338*^9, 
  3.728843080377976*^9, 3.728845626062332*^9, 3.7333672998433867`*^9, 
  3.792273239925037*^9},
 CellLabel->"During evaluation of In[65]:=",
 CellID->60992304,ExpressionUUID->"5d07142e-6384-4562-9e1e-d025fad18b87"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefVBundle", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle \"\>", "\[InvisibleSpace]", "TangentM3", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefVBundle, ": Defining ", 
   "vbundle ", $CellContext`TangentM3, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287503777488403`*^9, 3.728750522525712*^9, 
  3.728750585714631*^9, 3.728750759631687*^9, 3.7287510437107277`*^9, 
  3.728751621587109*^9, 3.7287593548827887`*^9, 3.7288132661487617`*^9, 
  3.72882738334347*^9, 3.728835974288519*^9, 3.728836291877338*^9, 
  3.728843080377976*^9, 3.728845626062332*^9, 3.7333672998433867`*^9, 
  3.792273239933625*^9},
 CellLabel->"During evaluation of In[65]:=",
 CellID->747953709,ExpressionUUID->"0f02e7b7-61d9-469b-bf54-2a3dd697325e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"v", "[", "a", "]"}], ",", "M3"}], "]"}]], "Input",
 CellChangeTimes->{{3.7287505537674723`*^9, 3.728750557011115*^9}},
 CellLabel->"In[66]:=",
 CellID->96268193,ExpressionUUID->"10df15cf-8da1-4d20-86a2-afc8aa52cda7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"v", "[", "a", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`v[$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{{3.728750557520774*^9, 3.7287505863204317`*^9}, 
   3.728750759799919*^9, 3.72875104391705*^9, 3.728751621764145*^9, 
   3.728759355134914*^9, 3.7288132663200197`*^9, 3.728827383549553*^9, 
   3.728835974593898*^9, 3.728836292103294*^9, 3.728843080618816*^9, 
   3.72884562626477*^9, 3.733367300910368*^9, 3.79227324080305*^9},
 CellLabel->"During evaluation of In[66]:=",
 CellID->550003913,ExpressionUUID->"7d02a7aa-abf8-489c-a5aa-ab4bb15a0e9b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", "M3", ",", 
   RowBox[{"Antisymmetric", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "2"}], "}"}], "]"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.72875044344412*^9, 3.728750485742198*^9}},
 CellLabel->"In[67]:=",
 CellID->1818717525,ExpressionUUID->"cea71f12-8a6f-4852-9035-62fd44fd7cd3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"T", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`T[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{{3.7287504952935467`*^9, 3.728750525108186*^9}, 
   3.728750586936494*^9, 3.728750759940381*^9, 3.728751044057124*^9, 
   3.7287516218973217`*^9, 3.728759355283832*^9, 3.728813266481922*^9, 
   3.728827383715893*^9, 3.728835974767*^9, 3.728836292248*^9, 
   3.728843080784718*^9, 3.7288456264292393`*^9, 3.733367301626313*^9, 
   3.792273241401979*^9},
 CellLabel->"During evaluation of In[67]:=",
 CellID->633278035,ExpressionUUID->"bb212a4a-ecac-4a84-9bd7-94724e98603a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Declare a complex vector bundle and a vector field on it:", \
"MathCaption",
 CellChangeTimes->{{3.7287509198272753`*^9, 3.728750936353177*^9}},
 CellID->1079830948,ExpressionUUID->"a4b54d25-5faf-45cd-a737-f5b18e4b5737"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefVBundle", "[", 
  RowBox[{"InnerC", ",", "M3", ",", "2", ",", 
   RowBox[{"{", 
    RowBox[{"A", ",", "B", ",", "C", ",", "D", ",", "F", ",", "G"}], "}"}], 
   ",", 
   RowBox[{"Dagger", "\[Rule]", "Complex"}]}], "]"}]], "Input",
 CellLabel->"In[68]:=",
 CellID->1538122804,ExpressionUUID->"9fd9142d-7e83-4041-9d92-9bddd7058dc4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefVBundle", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle \"\>", "\[InvisibleSpace]", "InnerC", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefVBundle, ": Defining ", 
   "vbundle ", $CellContext`InnerC, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728732789216392*^9, 3.728733007375355*^9, 
  3.728733189752511*^9, 3.7287333656471233`*^9, 3.728742302205514*^9, 
  3.7287432820370502`*^9, 3.728743381304302*^9, 3.728746914762179*^9, 
  3.7287494080059566`*^9, 3.7287499237170897`*^9, 3.7287500737875147`*^9, 
  3.728750174415072*^9, 3.728750282026334*^9, 3.728750507387027*^9, 
  3.72875057027335*^9, 3.728750698097638*^9, 3.7287507601412487`*^9, 
  3.7287510441944437`*^9, 3.728751622111302*^9, 3.728759355443617*^9, 
  3.728813266634886*^9, 3.728827383881782*^9, 3.728835974965375*^9, 
  3.728836292392392*^9, 3.7288430809653788`*^9, 3.7288456265978403`*^9, 
  3.733367302792327*^9, 3.792273242325447*^9},
 CellLabel->"During evaluation of In[68]:=",
 CellID->151982334,ExpressionUUID->"7bf9964c-4d03-4b5e-9c1c-06b204350078"],

Cell[BoxData[
 TemplateBox[{
  "ValidateSymbol", "capital", 
   "\"System name \\!\\(\\*RowBox[{\\\"\\\\\\\"C\\\\\\\"\\\"}]\\) is \
overloaded as an abstract index.\"", 2, 68, 4, 19668387891272167742, "Local", 
   "xAct`xCore`ValidateSymbol"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728645867902925*^9, 3.728655158876857*^9, 
  3.728655384910871*^9, 3.728660570057692*^9, 3.7286710165454817`*^9, 
  3.728674863441874*^9, 3.728732667790407*^9, 3.7287327892828617`*^9, 
  3.728733007445066*^9, 3.72873318985763*^9, 3.728733365719778*^9, 
  3.728742302337237*^9, 3.728743282128933*^9, 3.728743381368064*^9, 
  3.728746914857394*^9, 3.728749408072154*^9, 3.728749923760768*^9, 
  3.728750073879168*^9, 3.7287501744942293`*^9, 3.728750282076119*^9, 
  3.7287505074364223`*^9, 3.7287505703431187`*^9, 3.7287506981274548`*^9, 
  3.728750760171526*^9, 3.7287510442101183`*^9, 3.728751622127942*^9, 
  3.7287593555714617`*^9, 3.728813266651669*^9, 3.728827383900041*^9, 
  3.728835975125285*^9, 3.7288362924418583`*^9, 3.728843081002459*^9, 
  3.728845626617558*^9, 3.73336730281894*^9, 3.79227324233578*^9},
 CellLabel->"During evaluation of In[68]:=",
 CellID->1009031255,ExpressionUUID->"323404f2-2d8c-420a-837f-915d9896e2cd"],

Cell[BoxData[
 TemplateBox[{
  "ValidateSymbol", "capital", 
   "\"System name \\!\\(\\*RowBox[{\\\"\\\\\\\"D\\\\\\\"\\\"}]\\) is \
overloaded as an abstract index.\"", 2, 68, 5, 19668387891272167742, "Local", 
   "xAct`xCore`ValidateSymbol"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728645867902925*^9, 3.728655158876857*^9, 
  3.728655384910871*^9, 3.728660570057692*^9, 3.7286710165454817`*^9, 
  3.728674863441874*^9, 3.728732667790407*^9, 3.7287327892828617`*^9, 
  3.728733007445066*^9, 3.72873318985763*^9, 3.728733365719778*^9, 
  3.728742302337237*^9, 3.728743282128933*^9, 3.728743381368064*^9, 
  3.728746914857394*^9, 3.728749408072154*^9, 3.728749923760768*^9, 
  3.728750073879168*^9, 3.7287501744942293`*^9, 3.728750282076119*^9, 
  3.7287505074364223`*^9, 3.7287505703431187`*^9, 3.7287506981274548`*^9, 
  3.728750760171526*^9, 3.7287510442101183`*^9, 3.728751622127942*^9, 
  3.7287593555714617`*^9, 3.728813266651669*^9, 3.728827383900041*^9, 
  3.728835975125285*^9, 3.7288362924418583`*^9, 3.728843081002459*^9, 
  3.728845626617558*^9, 3.73336730281894*^9, 3.792273242346336*^9},
 CellLabel->"During evaluation of In[68]:=",
 CellID->109942426,ExpressionUUID->"dd418d3b-7032-4c51-bb3d-679661bd6bc2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefVBundle", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"conjugated vbundle \"\>", "\[InvisibleSpace]", 
   "InnerC\[Dagger]", "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Assuming fixed anti-isomorphism between InnerC \
and InnerC\[Dagger]\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefVBundle, ": Defining ", 
   "conjugated vbundle ", $CellContext`InnerC\[Dagger], ". ", 
   "Assuming fixed anti-isomorphism between InnerC and InnerC\[Dagger]"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728732789216392*^9, 3.728733007375355*^9, 
  3.728733189752511*^9, 3.7287333656471233`*^9, 3.728742302205514*^9, 
  3.7287432820370502`*^9, 3.728743381304302*^9, 3.728746914762179*^9, 
  3.7287494080059566`*^9, 3.7287499237170897`*^9, 3.7287500737875147`*^9, 
  3.728750174415072*^9, 3.728750282026334*^9, 3.728750507387027*^9, 
  3.72875057027335*^9, 3.728750698097638*^9, 3.7287507601412487`*^9, 
  3.7287510441944437`*^9, 3.728751622111302*^9, 3.728759355443617*^9, 
  3.728813266634886*^9, 3.728827383881782*^9, 3.728835974965375*^9, 
  3.728836292392392*^9, 3.7288430809653788`*^9, 3.7288456265978403`*^9, 
  3.733367302792327*^9, 3.792273242357745*^9},
 CellLabel->"During evaluation of In[68]:=",
 CellID->966417362,ExpressionUUID->"f4166d09-5e98-4f98-b63a-fa5b279c6a5a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"u", "[", 
    RowBox[{"-", "A"}], "]"}], ",", "M3", ",", 
   RowBox[{"Dagger", "\[Rule]", "Complex"}]}], "]"}]], "Input",
 CellLabel->"In[69]:=",
 CellID->593059525,ExpressionUUID->"41fee987-76df-4dd9-899d-b2e054ca23ea"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"u", "[", 
    RowBox[{"-", "A"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`u[-$CellContext`A], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287510446740217`*^9, 3.72875162231525*^9, 
  3.728759355794941*^9, 3.7288132668477507`*^9, 3.72882738412694*^9, 
  3.728835975419291*^9, 3.7288362926715097`*^9, 3.728843081383877*^9, 
  3.728845626833881*^9, 3.733367303759838*^9, 3.792273242939406*^9},
 CellLabel->"During evaluation of In[69]:=",
 CellID->300424984,ExpressionUUID->"c67147a3-a54d-464a-a700-9bb268ac1e6e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"u\[Dagger]", "[", 
    RowBox[{"-", "A\[Dagger]"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`u\[Dagger][-$CellContext`A\[Dagger]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287510446740217`*^9, 3.72875162231525*^9, 
  3.728759355794941*^9, 3.7288132668477507`*^9, 3.72882738412694*^9, 
  3.728835975419291*^9, 3.7288362926715097`*^9, 3.728843081383877*^9, 
  3.728845626833881*^9, 3.733367303759838*^9, 3.792273242950436*^9},
 CellLabel->"During evaluation of In[69]:=",
 CellID->508651991,ExpressionUUID->"4551fc82-a1c5-4198-9f78-b4c56e9941f5"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
To define a basis we only have to provide a name, a vector bundle and a list \
of cnumbers (\"component numbers\"), whose length must be the dimension of \
the vector bundle. The cnumbers can include 0 and even negative integers:\
\>", "MathCaption",
 CellChangeTimes->{{3.7286611194299307`*^9, 3.728661143387035*^9}, {
  3.72874241716654*^9, 3.728742418102738*^9}},
 CellTags->{
  "S5.18.1", "1.6"},ExpressionUUID->"1b913b5b-14e7-4071-af65-bf81fe188bf6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefBasis", "[", 
  RowBox[{"fred", ",", "TangentM3", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2"}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728742430302266*^9, 3.7287424306157227`*^9}, {
  3.7287472611266336`*^9, 3.728747261659307*^9}},
 CellLabel->"In[70]:=",ExpressionUUID->"c2ec987c-3aab-44cd-bc09-7bfa61defc86"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefBasis", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"basis \"\>", "\[InvisibleSpace]", "fred", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xCoba`DefBasis, ": Defining ", "basis ", $CellContext`fred, 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286611082289352`*^9, 3.728671231038711*^9, 
  3.7286748823136253`*^9, 3.7287424393287687`*^9, 3.72874329617754*^9, 
  3.728743394128528*^9, 3.728747262265957*^9, 3.7287494852097387`*^9, 
  3.728749935431992*^9, 3.7287503828126183`*^9, 3.728750527946123*^9, 
  3.7287505881209917`*^9, 3.7287507606555843`*^9, 3.7287510451404133`*^9, 
  3.7287516224984713`*^9, 3.728759356113003*^9, 3.7288132670155363`*^9, 
  3.728827384314583*^9, 3.7288359757332907`*^9, 3.728836292865864*^9, 
  3.7288430821123743`*^9, 3.728845627120904*^9, 3.733367305188802*^9, 
  3.7922732441733303`*^9},
 CellLabel->"During evaluation of In[70]:=",
 CellID->575210575,ExpressionUUID->"a8d5bf5e-654f-497c-9a04-76d29474aa88"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"parallel derivative \"\>", "\[InvisibleSpace]", 
   RowBox[{"PDfred", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "parallel derivative ", 
   $CellContext`PDfred[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286611082289352`*^9, 3.728671231038711*^9, 
  3.7286748823136253`*^9, 3.7287424393287687`*^9, 3.72874329617754*^9, 
  3.728743394128528*^9, 3.728747262265957*^9, 3.7287494852097387`*^9, 
  3.728749935431992*^9, 3.7287503828126183`*^9, 3.728750527946123*^9, 
  3.7287505881209917`*^9, 3.7287507606555843`*^9, 3.7287510451404133`*^9, 
  3.7287516224984713`*^9, 3.728759356113003*^9, 3.7288132670155363`*^9, 
  3.728827384314583*^9, 3.7288359757332907`*^9, 3.728836292865864*^9, 
  3.7288430821123743`*^9, 3.728845627120904*^9, 3.733367305188802*^9, 
  3.7922732441865*^9},
 CellLabel->"During evaluation of In[70]:=",
 CellID->1740238878,ExpressionUUID->"eb85d954-50b7-4900-b351-f4d20e285f43"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"TorsionPDfred", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "torsion tensor ", 
   $CellContext`TorsionPDfred[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286611082289352`*^9, 3.728671231038711*^9, 
  3.7286748823136253`*^9, 3.7287424393287687`*^9, 3.72874329617754*^9, 
  3.728743394128528*^9, 3.728747262265957*^9, 3.7287494852097387`*^9, 
  3.728749935431992*^9, 3.7287503828126183`*^9, 3.728750527946123*^9, 
  3.7287505881209917`*^9, 3.7287507606555843`*^9, 3.7287510451404133`*^9, 
  3.7287516224984713`*^9, 3.728759356113003*^9, 3.7288132670155363`*^9, 
  3.728827384314583*^9, 3.7288359757332907`*^9, 3.728836292865864*^9, 
  3.7288430821123743`*^9, 3.728845627120904*^9, 3.733367305188802*^9, 
  3.792273244200592*^9},
 CellLabel->"During evaluation of In[70]:=",
 CellID->1615034053,ExpressionUUID->"03a370a9-fb76-4e76-b865-9ea0dbdb6c32"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelPDfred", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "non-symmetric Christoffel tensor ", 
   $CellContext`ChristoffelPDfred[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286611082289352`*^9, 3.728671231038711*^9, 
  3.7286748823136253`*^9, 3.7287424393287687`*^9, 3.72874329617754*^9, 
  3.728743394128528*^9, 3.728747262265957*^9, 3.7287494852097387`*^9, 
  3.728749935431992*^9, 3.7287503828126183`*^9, 3.728750527946123*^9, 
  3.7287505881209917`*^9, 3.7287507606555843`*^9, 3.7287510451404133`*^9, 
  3.7287516224984713`*^9, 3.728759356113003*^9, 3.7288132670155363`*^9, 
  3.728827384314583*^9, 3.7288359757332907`*^9, 3.728836292865864*^9, 
  3.7288430821123743`*^9, 3.728845627120904*^9, 3.733367305188802*^9, 
  3.792273244207863*^9},
 CellLabel->"During evaluation of In[70]:=",
 CellID->1831514338,ExpressionUUID->"adccf27a-8b56-42cb-8fee-198c0d5b1f37"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Riemann tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RiemannPDfred", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Riemann tensor ", 
   $CellContext`RiemannPDfred[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, $CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286611082289352`*^9, 3.728671231038711*^9, 
  3.7286748823136253`*^9, 3.7287424393287687`*^9, 3.72874329617754*^9, 
  3.728743394128528*^9, 3.728747262265957*^9, 3.7287494852097387`*^9, 
  3.728749935431992*^9, 3.7287503828126183`*^9, 3.728750527946123*^9, 
  3.7287505881209917`*^9, 3.7287507606555843`*^9, 3.7287510451404133`*^9, 
  3.7287516224984713`*^9, 3.728759356113003*^9, 3.7288132670155363`*^9, 
  3.728827384314583*^9, 3.7288359757332907`*^9, 3.728836292865864*^9, 
  3.7288430821123743`*^9, 3.728845627120904*^9, 3.733367305188802*^9, 
  3.792273244215954*^9},
 CellLabel->"During evaluation of In[70]:=",
 CellID->230882668,ExpressionUUID->"1a770054-12c1-453b-910d-b5ab710aab7e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciPDfred", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Ricci tensor ", 
   $CellContext`RicciPDfred[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286611082289352`*^9, 3.728671231038711*^9, 
  3.7286748823136253`*^9, 3.7287424393287687`*^9, 3.72874329617754*^9, 
  3.728743394128528*^9, 3.728747262265957*^9, 3.7287494852097387`*^9, 
  3.728749935431992*^9, 3.7287503828126183`*^9, 3.728750527946123*^9, 
  3.7287505881209917`*^9, 3.7287507606555843`*^9, 3.7287510451404133`*^9, 
  3.7287516224984713`*^9, 3.728759356113003*^9, 3.7288132670155363`*^9, 
  3.728827384314583*^9, 3.7288359757332907`*^9, 3.728836292865864*^9, 
  3.7288430821123743`*^9, 3.728845627120904*^9, 3.733367305188802*^9, 
  3.792273244223461*^9},
 CellLabel->"During evaluation of In[70]:=",
 CellID->2089534594,ExpressionUUID->"ca77aa55-0b76-4fc8-9afc-7a6f9c6a0e0e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaUpfred", "[", 
    RowBox[{"a", ",", "b", ",", "c"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric +1 density ", 
   $CellContext`etaUpfred[$CellContext`a, $CellContext`b, $CellContext`c], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286611082289352`*^9, 3.728671231038711*^9, 
  3.7286748823136253`*^9, 3.7287424393287687`*^9, 3.72874329617754*^9, 
  3.728743394128528*^9, 3.728747262265957*^9, 3.7287494852097387`*^9, 
  3.728749935431992*^9, 3.7287503828126183`*^9, 3.728750527946123*^9, 
  3.7287505881209917`*^9, 3.7287507606555843`*^9, 3.7287510451404133`*^9, 
  3.7287516224984713`*^9, 3.728759356113003*^9, 3.7288132670155363`*^9, 
  3.728827384314583*^9, 3.7288359757332907`*^9, 3.728836292865864*^9, 
  3.7288430821123743`*^9, 3.728845627120904*^9, 3.733367305188802*^9, 
  3.792273244230814*^9},
 CellLabel->"During evaluation of In[70]:=",
 CellID->1007841967,ExpressionUUID->"b4441d93-bff9-4ca3-95fd-c5a3278d2942"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaDownfred", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric -1 density ", 
   $CellContext`etaDownfred[-$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286611082289352`*^9, 3.728671231038711*^9, 
  3.7286748823136253`*^9, 3.7287424393287687`*^9, 3.72874329617754*^9, 
  3.728743394128528*^9, 3.728747262265957*^9, 3.7287494852097387`*^9, 
  3.728749935431992*^9, 3.7287503828126183`*^9, 3.728750527946123*^9, 
  3.7287505881209917`*^9, 3.7287507606555843`*^9, 3.7287510451404133`*^9, 
  3.7287516224984713`*^9, 3.728759356113003*^9, 3.7288132670155363`*^9, 
  3.728827384314583*^9, 3.7288359757332907`*^9, 3.728836292865864*^9, 
  3.7288430821123743`*^9, 3.728845627120904*^9, 3.733367305188802*^9, 
  3.792273244238061*^9},
 CellLabel->"During evaluation of In[70]:=",
 CellID->1596415480,ExpressionUUID->"42829db3-b635-4c4b-9dab-be7eaa8913ec"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Several other objects are automatically defined: the parallel derivative and \
its torsion, Christoffel, Riemann and Ricci tensors. We shall say more about \
them later on. This is the name of the parallel derivative of the defined \
basis:\
\>", "MathCaption",
 CellChangeTimes->{
  3.728663915144733*^9, {3.728742450013453*^9, 3.7287425031058197`*^9}},
 CellTags->{
  "S5.18.1", "1.7"},ExpressionUUID->"f50dbadd-96cf-49cb-b613-f6c7131614b5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"PDOfBasis", "[", "fred", "]"}]], "Input",
 CellChangeTimes->{{3.728742474803959*^9, 3.728742475909643*^9}, {
  3.728747264588997*^9, 3.728747265083156*^9}},
 CellLabel->"In[71]:=",ExpressionUUID->"64d49cf8-473b-4b55-aed2-ceadd1bec358"],

Cell[BoxData["PDfred"], "Output",
 CellChangeTimes->{3.728661158947282*^9, 3.72867123135399*^9, 
  3.728674882680056*^9, 3.728742476741095*^9, 3.728743296601721*^9, 
  3.728743394446211*^9, 3.728747265401724*^9, 3.728749486905335*^9, 
  3.7287499356893187`*^9, 3.728750384396605*^9, 3.728750530298643*^9, 
  3.728750598066757*^9, 3.728750761497398*^9, 3.728751045535391*^9, 
  3.728751622761219*^9, 3.728759356490272*^9, 3.72881326728673*^9, 
  3.728827384612341*^9, 3.728835976077777*^9, 3.728836293159688*^9, 
  3.728843082701517*^9, 3.728845627561737*^9, 3.7333673069269342`*^9, 
  3.792273245801257*^9},
 CellLabel->"Out[71]=",
 CellID->251193709,ExpressionUUID->"8842468a-5a67-4cf0-897b-27804e44b84a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Each basis has a color, used to identify its associated indices and objects \
in StandardForm. The default for new bases is red, but we can specify a \
different color:\
\>", "MathCaption",
 CellChangeTimes->{
  3.7286611631859007`*^9, {3.728742523225885*^9, 
   3.728742523801676*^9}},ExpressionUUID->"f4b5fb15-fa95-40f5-af81-\
04cddf29abc6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefBasis", "[", 
  RowBox[{"blues", ",", "TangentM3", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2"}], "}"}], ",", 
   RowBox[{"BasisColor", "\[Rule]", "Blue"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728661174785022*^9, 3.728661182787078*^9}, {
   3.728742527433813*^9, 3.7287425277786617`*^9}, {3.728742558297847*^9, 
   3.728742558781736*^9}, {3.7287472711248703`*^9, 3.7287472718746767`*^9}, 
   3.728749268079259*^9},
 CellLabel->"In[72]:=",ExpressionUUID->"5b5205ac-cdcd-453f-b226-1215b1249018"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefBasis", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"basis \"\>", "\[InvisibleSpace]", "blues", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xCoba`DefBasis, ": Defining ", "basis ", $CellContext`blues, 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728661179154333*^9, 3.728671231480196*^9, 
  3.7286748828358517`*^9, 3.728742559486743*^9, 3.728743296733789*^9, 
  3.728743394579537*^9, 3.72874727248982*^9, 3.728749488249411*^9, 
  3.7287499358344307`*^9, 3.7287503857765017`*^9, 3.7287505315481997`*^9, 
  3.7287505994012747`*^9, 3.728750762009864*^9, 3.7287510456985197`*^9, 
  3.728751622914558*^9, 3.728759356797165*^9, 3.728813267431576*^9, 
  3.728827384782134*^9, 3.728835976244763*^9, 3.728836293302853*^9, 
  3.728843083129384*^9, 3.728845627720441*^9, 3.7333673085732183`*^9, 
  3.79227324734207*^9},
 CellLabel->"During evaluation of In[72]:=",
 CellID->731076079,ExpressionUUID->"02efbdff-6e89-493d-ae2c-f05edf8adb62"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"parallel derivative \"\>", "\[InvisibleSpace]", 
   RowBox[{"PDblues", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "parallel derivative ", 
   $CellContext`PDblues[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728661179154333*^9, 3.728671231480196*^9, 
  3.7286748828358517`*^9, 3.728742559486743*^9, 3.728743296733789*^9, 
  3.728743394579537*^9, 3.72874727248982*^9, 3.728749488249411*^9, 
  3.7287499358344307`*^9, 3.7287503857765017`*^9, 3.7287505315481997`*^9, 
  3.7287505994012747`*^9, 3.728750762009864*^9, 3.7287510456985197`*^9, 
  3.728751622914558*^9, 3.728759356797165*^9, 3.728813267431576*^9, 
  3.728827384782134*^9, 3.728835976244763*^9, 3.728836293302853*^9, 
  3.728843083129384*^9, 3.728845627720441*^9, 3.7333673085732183`*^9, 
  3.792273247354578*^9},
 CellLabel->"During evaluation of In[72]:=",
 CellID->1441715266,ExpressionUUID->"a7f64574-17de-47b5-a533-f16476dda035"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"TorsionPDblues", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "torsion tensor ", 
   $CellContext`TorsionPDblues[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728661179154333*^9, 3.728671231480196*^9, 
  3.7286748828358517`*^9, 3.728742559486743*^9, 3.728743296733789*^9, 
  3.728743394579537*^9, 3.72874727248982*^9, 3.728749488249411*^9, 
  3.7287499358344307`*^9, 3.7287503857765017`*^9, 3.7287505315481997`*^9, 
  3.7287505994012747`*^9, 3.728750762009864*^9, 3.7287510456985197`*^9, 
  3.728751622914558*^9, 3.728759356797165*^9, 3.728813267431576*^9, 
  3.728827384782134*^9, 3.728835976244763*^9, 3.728836293302853*^9, 
  3.728843083129384*^9, 3.728845627720441*^9, 3.7333673085732183`*^9, 
  3.792273247369157*^9},
 CellLabel->"During evaluation of In[72]:=",
 CellID->850486561,ExpressionUUID->"6ed28012-b6c2-442e-a554-54363d70a86a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelPDblues", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "non-symmetric Christoffel tensor ", 
   $CellContext`ChristoffelPDblues[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728661179154333*^9, 3.728671231480196*^9, 
  3.7286748828358517`*^9, 3.728742559486743*^9, 3.728743296733789*^9, 
  3.728743394579537*^9, 3.72874727248982*^9, 3.728749488249411*^9, 
  3.7287499358344307`*^9, 3.7287503857765017`*^9, 3.7287505315481997`*^9, 
  3.7287505994012747`*^9, 3.728750762009864*^9, 3.7287510456985197`*^9, 
  3.728751622914558*^9, 3.728759356797165*^9, 3.728813267431576*^9, 
  3.728827384782134*^9, 3.728835976244763*^9, 3.728836293302853*^9, 
  3.728843083129384*^9, 3.728845627720441*^9, 3.7333673085732183`*^9, 
  3.792273247376731*^9},
 CellLabel->"During evaluation of In[72]:=",
 CellID->476476295,ExpressionUUID->"5b03e026-057f-46f2-9000-f2a2ad0d9e12"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Riemann tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RiemannPDblues", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Riemann tensor ", 
   $CellContext`RiemannPDblues[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, $CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728661179154333*^9, 3.728671231480196*^9, 
  3.7286748828358517`*^9, 3.728742559486743*^9, 3.728743296733789*^9, 
  3.728743394579537*^9, 3.72874727248982*^9, 3.728749488249411*^9, 
  3.7287499358344307`*^9, 3.7287503857765017`*^9, 3.7287505315481997`*^9, 
  3.7287505994012747`*^9, 3.728750762009864*^9, 3.7287510456985197`*^9, 
  3.728751622914558*^9, 3.728759356797165*^9, 3.728813267431576*^9, 
  3.728827384782134*^9, 3.728835976244763*^9, 3.728836293302853*^9, 
  3.728843083129384*^9, 3.728845627720441*^9, 3.7333673085732183`*^9, 
  3.792273247384388*^9},
 CellLabel->"During evaluation of In[72]:=",
 CellID->170570402,ExpressionUUID->"2650f2e1-4809-43fc-b8b3-9948068612ed"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciPDblues", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Ricci tensor ", 
   $CellContext`RicciPDblues[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728661179154333*^9, 3.728671231480196*^9, 
  3.7286748828358517`*^9, 3.728742559486743*^9, 3.728743296733789*^9, 
  3.728743394579537*^9, 3.72874727248982*^9, 3.728749488249411*^9, 
  3.7287499358344307`*^9, 3.7287503857765017`*^9, 3.7287505315481997`*^9, 
  3.7287505994012747`*^9, 3.728750762009864*^9, 3.7287510456985197`*^9, 
  3.728751622914558*^9, 3.728759356797165*^9, 3.728813267431576*^9, 
  3.728827384782134*^9, 3.728835976244763*^9, 3.728836293302853*^9, 
  3.728843083129384*^9, 3.728845627720441*^9, 3.7333673085732183`*^9, 
  3.792273247391569*^9},
 CellLabel->"During evaluation of In[72]:=",
 CellID->2033561316,ExpressionUUID->"c93d6a16-03ca-4db1-a236-7c10768f8686"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaUpblues", "[", 
    RowBox[{"a", ",", "b", ",", "c"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric +1 density ", 
   $CellContext`etaUpblues[$CellContext`a, $CellContext`b, $CellContext`c], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728661179154333*^9, 3.728671231480196*^9, 
  3.7286748828358517`*^9, 3.728742559486743*^9, 3.728743296733789*^9, 
  3.728743394579537*^9, 3.72874727248982*^9, 3.728749488249411*^9, 
  3.7287499358344307`*^9, 3.7287503857765017`*^9, 3.7287505315481997`*^9, 
  3.7287505994012747`*^9, 3.728750762009864*^9, 3.7287510456985197`*^9, 
  3.728751622914558*^9, 3.728759356797165*^9, 3.728813267431576*^9, 
  3.728827384782134*^9, 3.728835976244763*^9, 3.728836293302853*^9, 
  3.728843083129384*^9, 3.728845627720441*^9, 3.7333673085732183`*^9, 
  3.792273247399048*^9},
 CellLabel->"During evaluation of In[72]:=",
 CellID->862796007,ExpressionUUID->"6734124c-c4b0-46b9-a37e-9c6ff3d3c624"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaDownblues", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric -1 density ", 
   $CellContext`etaDownblues[-$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728661179154333*^9, 3.728671231480196*^9, 
  3.7286748828358517`*^9, 3.728742559486743*^9, 3.728743296733789*^9, 
  3.728743394579537*^9, 3.72874727248982*^9, 3.728749488249411*^9, 
  3.7287499358344307`*^9, 3.7287503857765017`*^9, 3.7287505315481997`*^9, 
  3.7287505994012747`*^9, 3.728750762009864*^9, 3.7287510456985197`*^9, 
  3.728751622914558*^9, 3.728759356797165*^9, 3.728813267431576*^9, 
  3.728827384782134*^9, 3.728835976244763*^9, 3.728836293302853*^9, 
  3.728843083129384*^9, 3.728845627720441*^9, 3.7333673085732183`*^9, 
  3.792273247406353*^9},
 CellLabel->"During evaluation of In[72]:=",
 CellID->1202905105,ExpressionUUID->"63cd069e-036f-42ff-ab7e-b11dda6e66a4"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["A basis has several associated definitions and UpValues:", "MathCaption",
 CellChangeTimes->{{3.72866118914478*^9, 3.7286611913286448`*^9}, 
   3.728742589917898*^9},ExpressionUUID->"588ab9ea-12b6-4cd9-be95-\
2b53ccebebd7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "fred"}]], "Input",
 CellChangeTimes->{{3.728742592278576*^9, 3.728742593752141*^9}, {
  3.72874727480256*^9, 3.728747275290606*^9}},
 CellLabel->"In[73]:=",ExpressionUUID->"9bfb9365-41e0-450e-9e89-c458ee1bfa00"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   FrameBox[
    DynamicModuleBox[{System`InformationDump`open$$ = True, 
     System`InformationDump`mouseOver$$ = False}, 
     PaneSelectorBox[{True->
      TagBox[GridBox[{
         {
          ItemBox[
           PaneBox[
            StyleBox["\<\" Symbol\"\>", "InformationTitleText",
             StripOnInput->False,
             BaseStyle -> None],
            FrameMargins->{{4, 0}, {-1, 1}}],
           BaseStyle->"InformationTitleBackground",
           StripOnInput->False], 
          ItemBox["\<\"\"\>",
           BaseStyle->"InformationTitleBackground",
           StripOnInput->False]},
         {
          ItemBox[
           PaneBox[
            StyleBox["\<\"Global`fred\"\>", "InformationUsageText",
             StripOnInput->False,
             LineSpacing->{1.5, 1.5, 3.}],
            FrameMargins->{{10, 10}, {8, 10}}],
           BaseStyle->"InformationUsageSubtitleBackground",
           StripOnInput->False], 
          ItemBox["\<\"\"\>",
           BaseStyle->"InformationUsageSubtitleBackground",
           StripOnInput->False]},
         {
          PaneBox[GridBox[{
             {
              
              DynamicModuleBox[{System`InformationDump`open$$ = {
               False, False, False, False, False, False, False, False, False, 
                False, False, False}}, 
               StyleBox[GridBox[{
                  {
                   TagBox[
                    TooltipBox[
                    
                    StyleBox["\<\" UpValue Definitions\"\>", 
                    "InformationRowLabel",
                    StripOnInput->False],
                    "\"UpValues\"",
                    TooltipStyle->"TextStyling"],
                    Annotation[#, "UpValues", "Tooltip"]& ], GridBox[{
                    {
                    RowBox[{
                    RowBox[{"BasisQ", "[", "fred", "]"}], "^=", "True"}]},
                    {
                    RowBox[{
                    RowBox[{"BasisColor", "[", "fred", "]"}], "^=", 
                    RowBox[{"RGBColor", "[", 
                    RowBox[{"1", ",", "0", ",", "0"}], "]"}]}]},
                    {
                    RowBox[{
                    RowBox[{"VBundleOfBasis", "[", "fred", "]"}], "^=", 
                    InterpretationBox[
                    StyleBox["\[DoubleStruckCapitalT]M3",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`TangentM3,
                    Editable->False]}]},
                    {
                    RowBox[{"fred", "/:", 
                    RowBox[{"CNumbersOf", "[", 
                    RowBox[{"fred", ",", 
                    InterpretationBox[
                    StyleBox["\[DoubleStruckCapitalT]M3",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`TangentM3,
                    Editable->False]}], "]"}], "=", 
                    RowBox[{"{", 
                    RowBox[{"0", ",", "1", ",", "2"}], "}"}]}]},
                    {
                    RowBox[{
                    RowBox[{"etaOrientation", "[", "fred", "]"}], "^=", "1"}]},
                    {
                    RowBox[{
                    RowBox[{"DependenciesOfBasis", "[", "fred", "]"}], "^=", 
                    RowBox[{"{", "M3", "}"}]}]},
                    {
                    RowBox[{
                    RowBox[{"DefInfo", "[", "fred", "]"}], "^=", 
                    RowBox[{"{", 
                    RowBox[{"\<\"basis\"\>", ",", "\<\"\"\>"}], "}"}]}]},
                    {
                    RowBox[{
                    RowBox[{"HostsOf", "[", "fred", "]"}], "^=", 
                    RowBox[{"{", 
                    InterpretationBox[
                    StyleBox["\[DoubleStruckCapitalT]M3",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`TangentM3,
                    Editable->False], "}"}]}]},
                    {
                    RowBox[{
                    RowBox[{"PDOfBasis", "[", "fred", "]"}], "^=", "PDfred"}]},
                    {
                    RowBox[{
                    RowBox[{"ServantsOf", "[", "fred", "]"}], "^=", 
                    RowBox[{"{", 
                    RowBox[{"PDfred", ",", "etaUpfred", ",", "etaDownfred"}], 
                    "}"}]}]},
                    {
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Dagger", "[", "fred", "]"}],
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    xAct`xTensor`Dagger[$CellContext`fred],
                    Editable->False], "^=", "fred"}]}
                    },
                    DefaultBaseStyle->"Column",
                    GridBoxAlignment->{"Columns" -> {{Left}}},
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}]},
                  {
                   TagBox[
                    TooltipBox[
                    StyleBox["\<\" Full Name\"\>", "InformationRowLabel",
                    StripOnInput->False],
                    "\"FullName\"",
                    TooltipStyle->"TextStyling"],
                    
                    Annotation[#, "FullName", 
                    "Tooltip"]& ], "\<\"Global`fred\"\>"}
                 },
                 AutoDelete->False,
                 GridBoxAlignment->{"Columns" -> {Right, Left}},
                 GridBoxDividers->None,
                 GridBoxItemSize->{"Columns" -> {Automatic, Automatic}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.5599999999999999]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.8]}, 
                    Offset[0.2]}}], "DialogStyle",
                StripOnInput->False],
               DynamicModuleValues:>{}]}
            },
            DefaultBaseStyle->"Column",
            GridBoxAlignment->{"Columns" -> {{Left}}},
            GridBoxDividers->{"Columns" -> {{False}}, "Rows" -> {{False}}},
            
            GridBoxItemSize->{
             "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.5599999999999999]}, 
                Offset[0.27999999999999997`]}, "Rows" -> {
                Offset[0.2], {
                 Offset[3.6]}, 
                Offset[0.2]}}],
           FrameMargins->6], ""},
         {
          ItemBox[
           TagBox[
            ButtonBox[
             PaneSelectorBox[{False->
              
              DynamicBox[FEPrivate`FrontEndResource[
               "FEBitmaps", "UpPointerOpener"],
               ImageSizeCache->{10., {2., 8.}}], True->
              
              DynamicBox[FEPrivate`FrontEndResource[
               "FEBitmaps", "UpPointerOpenerHot"],
               ImageSizeCache->{10., {2., 8.}}]}, Dynamic[
              System`InformationDump`mouseOver$$]],
             Alignment->Left,
             Appearance->{"Default" -> None},
             
             ButtonFunction:>FEPrivate`Set[
              System`InformationDump`open$$, False],
             Evaluator->Automatic,
             FrameMargins->{{9, 0}, {0, 0}},
             ImageMargins->0,
             ImageSize->Full,
             Method->"Preemptive"],
            
            EventHandlerTag[{
             "MouseEntered" :> 
              FEPrivate`Set[System`InformationDump`mouseOver$$, True], 
              "MouseExited" :> 
              FEPrivate`Set[System`InformationDump`mouseOver$$, False], 
              Method -> "Preemptive", PassEventsDown -> Automatic, 
              PassEventsUp -> True}]],
           BaseStyle->"InformationTitleBackground",
           StripOnInput->False], "\[SpanFromLeft]"}
        },
        AutoDelete->False,
        FrameStyle->Directive[
          GrayLevel[0.8], 
          Thickness[Tiny]],
        GridBoxAlignment->{"Columns" -> {Left, Right}, "Rows" -> {{Center}}},
        GridBoxDividers->{
         "Columns" -> {{None}}, "Rows" -> {False, {True}, False}},
        GridBoxItemSize->{
         "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
       "Grid"], False->
      TagBox[GridBox[{
         {
          ItemBox[
           PaneBox[
            StyleBox["\<\" Symbol\"\>", "InformationTitleText",
             StripOnInput->False],
            FrameMargins->{{4, 0}, {-1, 1}}],
           BaseStyle->"InformationTitleBackground",
           StripOnInput->False], 
          ItemBox["\<\"\"\>",
           BaseStyle->"InformationTitleBackground",
           StripOnInput->False]},
         {
          ItemBox[
           PaneBox[
            StyleBox["\<\"Global`fred\"\>", "InformationUsageText",
             StripOnInput->False,
             LineSpacing->{1.5, 1.5, 3.}],
            FrameMargins->{{10, 10}, {8, 10}}],
           BaseStyle->"InformationUsageSubtitleBackground",
           StripOnInput->False], 
          ItemBox["\<\"\"\>",
           BaseStyle->"InformationUsageSubtitleBackground",
           StripOnInput->False]},
         {
          ItemBox[
           TagBox[
            ButtonBox[
             PaneSelectorBox[{False->
              
              DynamicBox[FEPrivate`FrontEndResource[
               "FEBitmaps", "DownPointerOpener"]], True->
              
              DynamicBox[FEPrivate`FrontEndResource[
               "FEBitmaps", "DownPointerOpenerHot"]]}, Dynamic[
              System`InformationDump`mouseOver$$]],
             Alignment->Left,
             Appearance->{"Default" -> None},
             ButtonFunction:>FEPrivate`Set[
              System`InformationDump`open$$, True],
             Evaluator->Automatic,
             FrameMargins->{{9, 0}, {0, 0}},
             ImageMargins->0,
             ImageSize->Full,
             Method->"Preemptive"],
            
            EventHandlerTag[{
             "MouseEntered" :> 
              FEPrivate`Set[System`InformationDump`mouseOver$$, True], 
              "MouseExited" :> 
              FEPrivate`Set[System`InformationDump`mouseOver$$, False], 
              Method -> "Preemptive", PassEventsDown -> Automatic, 
              PassEventsUp -> True}]],
           BaseStyle->"InformationTitleBackground",
           StripOnInput->False], "\[SpanFromLeft]"}
        },
        AutoDelete->False,
        FrameStyle->Directive[
          GrayLevel[0.8], 
          Thickness[Tiny]],
        GridBoxAlignment->{"Columns" -> {Left, Right}, "Rows" -> {{Center}}},
        GridBoxDividers->{
         "Columns" -> {{None}}, "Rows" -> {False, {True}, False}},
        GridBoxItemSize->{
         "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
       "Grid"]}, Dynamic[System`InformationDump`open$$],
      BaselinePosition->Baseline,
      FrameMargins->0,
      ImageSize->Automatic],
     DynamicModuleValues:>{}],
    BaseStyle->"InformationGridFrame",
    StripOnInput->False], "InformationGridPlain",
   StripOnInput->False],
  InformationData[
   Association[
   "ObjectType" -> "Symbol", "Usage" -> "Global`fred", "Documentation" -> 
    None, "OwnValues" -> None, "UpValues" -> 
    Information`InformationValueForm[
     UpValues, $CellContext`fred, {
      xAct`xTensor`BasisQ[$CellContext`fred] -> True, 
       xAct`xCoba`BasisColor[$CellContext`fred] -> RGBColor[1, 0, 0], 
       xAct`xCoba`VBundleOfBasis[$CellContext`fred] -> $CellContext`TangentM3,
        xAct`xCoba`CNumbersOf[$CellContext`fred, $CellContext`TangentM3] -> {
        0, 1, 2}, xAct`xCoba`etaOrientation[$CellContext`fred] -> 1, 
       xAct`xCoba`DependenciesOfBasis[$CellContext`fred] -> {$CellContext`M3},
        xAct`xTensor`DefInfo[$CellContext`fred] -> {"basis", ""}, 
       xAct`xTensor`HostsOf[$CellContext`fred] -> {$CellContext`TangentM3}, 
       xAct`xCoba`PDOfBasis[$CellContext`fred] -> $CellContext`PDfred, 
       xAct`xTensor`ServantsOf[$CellContext`fred] -> {$CellContext`PDfred, \
$CellContext`etaUpfred, $CellContext`etaDownfred}, 
       xAct`xTensor`Dagger[$CellContext`fred] -> $CellContext`fred}], 
    "DownValues" -> None, "SubValues" -> None, "DefaultValues" -> None, 
    "NValues" -> None, "FormatValues" -> None, "Options" -> None, 
    "Attributes" -> {}, "FullName" -> "Global`fred"], False]]], "Output",
 CellChangeTimes->{3.792273248829608*^9},
 CellLabel->"Out[73]=",
 CellID->835398157,ExpressionUUID->"0a613360-10e5-4875-84c5-2b42866c06b7"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["This is the list of bases we have defined so far:", "MathCaption",
 CellChangeTimes->{{3.7286612083358994`*^9, 3.7286612207838573`*^9}},
 CellID->1062134855,ExpressionUUID->"2c95fb17-daae-496b-90bd-14b7cceced56"],

Cell[CellGroupData[{

Cell[BoxData["$Bases"], "Input",
 CellLabel->"In[74]:=",ExpressionUUID->"284de3dd-8b5c-47e4-a943-90d6d74f0e55"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"fred", ",", "blues"}], "}"}]], "Output",
 CellChangeTimes->{3.7286612048268747`*^9, 3.728671232082341*^9, 
  3.728674883500795*^9, 3.728742606243246*^9, 3.728743297961096*^9, 
  3.728743395152429*^9, 3.728747278904173*^9, 3.7287494934410763`*^9, 
  3.728749936335544*^9, 3.728750390630744*^9, 3.728750535411242*^9, 
  3.728750602202997*^9, 3.728750762873687*^9, 3.728751046527026*^9, 
  3.728751623380911*^9, 3.728759357964581*^9, 3.728813267972953*^9, 
  3.728827385428608*^9, 3.728835977050632*^9, 3.7288362938927603`*^9, 
  3.728843084205617*^9, 3.7288456282451344`*^9, 3.733367311293479*^9, 
  3.7922732509842377`*^9},
 CellLabel->"Out[74]=",
 CellID->1057914556,ExpressionUUID->"0a0e6d75-ec94-4060-aaf6-6f57961a95f1"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["1.1.2. Basis objects", "Subsubsection",
 CellChangeTimes->{{3.728661774391335*^9, 3.728661777357946*^9}, {
   3.728661817699605*^9, 3.728661818283976*^9}, 3.72866278132205*^9, {
   3.728673438717552*^9, 
   3.728673438909156*^9}},ExpressionUUID->"7883ce9d-6f1e-4b4c-826d-\
d82119e15c0c"],

Cell[TextData[{
 "Now that we have bases, we can use basis vectors and work with tensor \
components. Two new kinds of indices are needed:\n\t ",
 StyleBox["{a, basisname}", "MR"],
 " :  Basis index (",
 StyleBox["BIndex", "MR"],
 "),  ",
 StyleBox["a", "MR"],
 " is an abstract index.\n\t ",
 StyleBox["{i, basisname}", "MR"],
 " :  Component index (",
 StyleBox["CIndex", "MR"],
 "), ",
 StyleBox["i", "MR"],
 " is a valid component number.\nWhen we define a basis in the tangent \
bundle, its dual in the cotangent bundle is also automatically defined. This \
dual basis is called -basisname and has the same cnumbers as the original. \
The notation for covariant bc-indices is \n\t",
 StyleBox["{-a, -basisname}", "MR"],
 " :  Basis index (",
 StyleBox["BIndex", "MR"],
 "),  ",
 StyleBox["a", "MR"],
 " is an abstract index.\n\t",
 StyleBox["{i,  -basisname}", "MR"],
 " :  Component index (",
 StyleBox["CIndex", "MR"],
 "), ",
 StyleBox["i", "MR"],
 " is a valid coordinate number.\nNotice how a covariant b-index is the \
negative of the corresponding contravariant index, but this is not true for \
c-indices. This apparent inconsistency is a conscious choice: it allows us to \
use zero and negative cnumbers (the latter being common in treatments of \
angular momentum, tensor harmonics, etc.)\nNOTE: In versions 0.3 and 0.4 we \
used b-indices like {a, polar}, {-a, polar} and c-indices like {1, polar} and \
{-1, polar}. This notational change is ",
 StyleBox["not",
  FontSlant->"Italic"],
 " backwards compatible. We apologize to those users of ",
 StyleBox["xCoba`", "MR"],
 " who now need to change lots of signs in previously developed notebooks, \
but we feel that the use of this new notation will pay off."
}], "Text",
 CellChangeTimes->{{3.7286612759669437`*^9, 3.728661295501946*^9}, {
  3.728661333340712*^9, 
  3.728661353426968*^9}},ExpressionUUID->"225c0d65-6dec-4c7f-aeb8-\
0a7353325878"],

Cell[CellGroupData[{

Cell["Several functions are provided to check validity:", "MathCaption",
 CellChangeTimes->{{3.728661376151986*^9, 
  3.728661378991946*^9}},ExpressionUUID->"536dc41c-e7b4-4863-968d-\
339931e28a80"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndicesOfVBundle", "[", "TangentM3", "]"}]], "Input",
 CellLabel->"In[75]:=",ExpressionUUID->"587060d8-a9c0-4eb1-8df3-ad17ce14a4b4"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f"}], "}"}], 
   ",", 
   RowBox[{"{", 
    RowBox[{"f1", ",", "f2", ",", "f3"}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728661380379287*^9, 3.728671237060877*^9, 
  3.728674883717046*^9, 3.728742622895011*^9, 3.7287432984615498`*^9, 
  3.728743395276918*^9, 3.7287472862186193`*^9, 3.728749497710363*^9, 
  3.728749936533288*^9, 3.728750410991712*^9, 3.728750538197644*^9, 
  3.728750605802652*^9, 3.728750766786337*^9, 3.728751057183301*^9, 
  3.7287516235902853`*^9, 3.7288132681020727`*^9, 3.728827385598845*^9, 
  3.728835977188724*^9, 3.7288362940430727`*^9, 3.728843089501257*^9, 
  3.728845628395335*^9, 3.733367320426853*^9, 3.7922732653392982`*^9},
 CellLabel->"Out[75]=",
 CellID->2021201992,ExpressionUUID->"d0f83860-d6e6-4dee-995d-326572fc3736"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"indexlist", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"a", ",", "fred"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "blues"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"w", ",", "fred"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"A", ",", "fred"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", "fred"}], "}"}], ",", "a", ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "blues"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"5", ",", "fred"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "1"}], ",", 
       RowBox[{"-", "fred"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"a", ",", "other"}], "}"}]}], "}"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.728742626429675*^9, 3.728742688873519*^9}, {
  3.728747288236272*^9, 3.728747304105721*^9}, {3.728749277465746*^9, 
  3.7287492783811903`*^9}},
 CellLabel->"In[76]:=",ExpressionUUID->"d4b0cb4a-45b6-4991-a4c1-ca8aae86f333"],

Cell[CellGroupData[{

Cell[BoxData["$Bases"], "Input",
 CellLabel->"In[77]:=",ExpressionUUID->"73543f66-ce0c-41c3-8759-deacdeff54c5"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"fred", ",", "blues"}], "}"}]], "Output",
 CellChangeTimes->{3.728661383210031*^9, 3.728671237265472*^9, 
  3.728674883916779*^9, 3.7287426951716137`*^9, 3.728743298950094*^9, 
  3.728743395531502*^9, 3.728747306790636*^9, 3.728749498942534*^9, 
  3.7287499368185167`*^9, 3.7287504111597*^9, 3.728750539447094*^9, 
  3.7287506070751333`*^9, 3.728750767036315*^9, 3.7287510573349743`*^9, 
  3.728751623879921*^9, 3.728813268240901*^9, 3.728827385828704*^9, 
  3.728835977365494*^9, 3.728836294257658*^9, 3.728843089754072*^9, 
  3.7288456286331577`*^9, 3.733367322044083*^9, 3.792273266897358*^9},
 CellLabel->"Out[77]=",
 CellID->1364513160,ExpressionUUID->"dc41ff6e-c6f1-43cd-bef7-b553e32bd387"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"BIndexQ", "/@", " ", "indexlist"}]], "Input",
 CellLabel->"In[78]:=",ExpressionUUID->"32dc7866-a948-4db3-8202-b25604fd70b5"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
  "True", ",", "True", ",", "False", ",", "True", ",", "False", ",", "False", 
   ",", "False", ",", "False", ",", "False", ",", "False"}], "}"}]], "Output",\

 CellChangeTimes->{3.728661384313463*^9, 3.728671237398191*^9, 
  3.728674884129939*^9, 3.728742697189395*^9, 3.728743299070634*^9, 
  3.7287433957253036`*^9, 3.7287473077585783`*^9, 3.7287494995308113`*^9, 
  3.728749936951621*^9, 3.728750411293066*^9, 3.728750540022892*^9, 
  3.7287506077680397`*^9, 3.7287507671760187`*^9, 3.728751057484213*^9, 
  3.7287516240146646`*^9, 3.728813268401772*^9, 3.728827385981289*^9, 
  3.72883597750557*^9, 3.728836294410036*^9, 3.72884308989904*^9, 
  3.7288456287823563`*^9, 3.733367322610421*^9, 3.792273267472179*^9},
 CellLabel->"Out[78]=",
 CellID->1429685618,ExpressionUUID->"66ab0f5a-1c2e-4039-b052-ea86545a4864"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"CIndexQ", " ", "/@", " ", "indexlist"}]], "Input",
 CellLabel->"In[79]:=",ExpressionUUID->"58c5b5e2-6c70-47e7-b949-f203cb85b8c0"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
  "False", ",", "False", ",", "False", ",", "False", ",", "True", ",", 
   "False", ",", "True", ",", "False", ",", "False", ",", "False"}], 
  "}"}]], "Output",
 CellChangeTimes->{3.7286613853437147`*^9, 3.7286712375336933`*^9, 
  3.7286748843441877`*^9, 3.728742698233316*^9, 3.728743299197647*^9, 
  3.728743395864582*^9, 3.728747308605625*^9, 3.728749500173131*^9, 
  3.728749937067589*^9, 3.72875041144209*^9, 3.728750540659316*^9, 
  3.728750608355386*^9, 3.7287507674415293`*^9, 3.728751057616477*^9, 
  3.728751624163352*^9, 3.728813268552442*^9, 3.728827386131575*^9, 
  3.7288359776385317`*^9, 3.728836294559916*^9, 3.728843090218813*^9, 
  3.728845628930691*^9, 3.7333673232462053`*^9, 3.792273268019999*^9},
 CellLabel->"Out[79]=",
 CellID->149829661,ExpressionUUID->"5654b148-4d9f-41e5-b225-0b18652bf811"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["An extra function selects both basis and component indices:", \
"MathCaption",
 CellChangeTimes->{{3.728742702770005*^9, 
  3.7287427033133507`*^9}},ExpressionUUID->"8e17a1d0-fcfa-45a7-88e1-\
e8335e2d68e4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"BCIndexQ", " ", "/@", " ", "indexlist"}]], "Input",
 CellLabel->"In[80]:=",ExpressionUUID->"0c9a0486-7d17-47ea-a628-990f4e38afee"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
  "True", ",", "True", ",", "False", ",", "True", ",", "True", ",", "False", 
   ",", "True", ",", "False", ",", "False", ",", "False"}], "}"}]], "Output",
 CellChangeTimes->{3.728661388443357*^9, 3.728671237664977*^9, 
  3.728674884516561*^9, 3.728742700418931*^9, 3.728743299338346*^9, 
  3.728743396081163*^9, 3.728747310040349*^9, 3.728749501890305*^9, 
  3.7287499372011433`*^9, 3.728750411575324*^9, 3.7287505416944036`*^9, 
  3.7287506094350843`*^9, 3.728750768042274*^9, 3.7287510578421297`*^9, 
  3.728751624326755*^9, 3.728813269074976*^9, 3.728827386374284*^9, 
  3.7288359777718678`*^9, 3.728836294709413*^9, 3.7288430905625353`*^9, 
  3.7288456290799828`*^9, 3.733367324262833*^9, 3.792273269011567*^9},
 CellLabel->"Out[80]=",
 CellID->440272853,ExpressionUUID->"6117b2df-5ad9-475d-a5f1-ed2323824370"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "On occasion, it is interesting to know whether an index is either an ",
 StyleBox["AIndex", "MR"],
 " or a ",
 StyleBox["BIndex", "MR"],
 ". Those are the only types that may be contracted, which makes them \
important for functions such as ",
 StyleBox["ContractBasis", "MR"],
 ". We will call them `contractible indices'."
}], "MathCaption",
 CellChangeTimes->{{3.733367330880289*^9, 
  3.733367347207659*^9}},ExpressionUUID->"129c6ba2-51dd-4a92-9ef7-\
b79a233f7ea6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ABIndexQ", " ", "/@", " ", "indexlist"}]], "Input",
 CellLabel->"In[81]:=",ExpressionUUID->"ea5d79e8-f11c-4409-8f0f-490801d63df1"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
  "True", ",", "True", ",", "False", ",", "True", ",", "False", ",", "True", 
   ",", "False", ",", "False", ",", "False", ",", "False"}], "}"}]], "Output",\

 CellChangeTimes->{3.728661393245131*^9, 3.728671237792642*^9, 
  3.728674884652648*^9, 3.728742719658112*^9, 3.7287432995539703`*^9, 
  3.72874339619477*^9, 3.728747311337358*^9, 3.728749504371229*^9, 
  3.7287499373178053`*^9, 3.7287504117082357`*^9, 3.728750543961832*^9, 
  3.728750611968759*^9, 3.7287507684900846`*^9, 3.728751058237167*^9, 
  3.72875162445859*^9, 3.728813269564513*^9, 3.728827386606957*^9, 
  3.728835977908737*^9, 3.728836294859911*^9, 3.728843091133651*^9, 
  3.728845629230131*^9, 3.73336735143302*^9, 3.792273270205676*^9},
 CellLabel->"Out[81]=",
 CellID->1691975529,ExpressionUUID->"50245daf-a32c-4598-9962-e0ce436a60dd"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "Notice that we can get a ",
 StyleBox["False", "MR"],
 " output for several reasons: The basis, abstract index or cnumber might not \
exist or they might belong to different vbundles."
}], "MathCaption",ExpressionUUID->"444eeeea-77d5-4d25-a3ad-21940ae1a5b2"],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["If we use these indices inside of tensors, ", "Text"],
 StyleBox["Mathematica", "Text",
  FontSlant->"Italic"],
 StyleBox[" will display them according to ", "Text"],
 StyleBox["BasisColor", "Text", "MR"],
 StyleBox[",", "Text"]
}], "MathCaption",ExpressionUUID->"eebf6268-a2c4-4344-8477-950d6fc8c617"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"T", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"1", ",", 
      RowBox[{"-", "blues"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{"2", ",", 
      RowBox[{"-", "blues"}]}], "}"}]}], "]"}], " ", 
  RowBox[{"v", "[", 
   RowBox[{"{", 
    RowBox[{"a", ",", "fred"}], "}"}], "]"}], " "}]], "Input",
 CellChangeTimes->{
  3.728742736366063*^9, {3.728747315633*^9, 3.7287473206807327`*^9}, {
   3.728749281682767*^9, 3.728749282308937*^9}},
 CellLabel->"In[82]:=",ExpressionUUID->"268191a3-370e-44fb-995b-3764026a42ee"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {" ", " "},
          {
           StyleBox["1",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox["2",
            FontColor->RGBColor[0., 0., 1.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[{1, -$CellContext`blues}, {2, -$CellContext`blues}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {
           StyleBox["a",
            FontColor->RGBColor[1., 0., 0.]]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{$CellContext`a, $CellContext`fred}],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728661405712282*^9, 3.728671237928385*^9, 
  3.728674884827883*^9, 3.728742741010296*^9, 3.72874329968436*^9, 
  3.728743396330844*^9, 3.728747321620275*^9, 3.728749506169181*^9, 
  3.728749937435382*^9, 3.728750411847303*^9, 3.728750545643959*^9, 
  3.728750614459724*^9, 3.728750768676738*^9, 3.728751058381624*^9, 
  3.728751624593524*^9, 3.728813269791626*^9, 3.7288273868110857`*^9, 
  3.728835978083827*^9, 3.728836295035866*^9, 3.728843091307954*^9, 
  3.728845629385003*^9, 3.733367355876244*^9, 3.792273271989026*^9},
 CellLabel->"Out[82]=",
 CellID->148472211,ExpressionUUID->"27651885-7304-4ec0-813a-a8321c7c48fc"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "The Basis tensor is formally xTensor's delta tensor, but we prefer to \
separate Basis in xCoba`. Depending on its indices, the meaning will change, \
and in particular when both its indices are abstract, it will autoconvert \
into delta:\n\t- ",
 StyleBox["Basis[ AIndex, BCIndex]", "MR"],
 " :\tBasis vector (or dual basis covector).\n\t- ",
 StyleBox["Basis[ AIndex, AIndex]", "MR"],
 ": \tIdentity tensor in the corresponding vbundle (turns into \[Delta]).\n\t\
- ",
 StyleBox["Basis[ BCIndex, BCIndex]", "MR"],
 ":\tIf the indices belong to different bases, it represents a basis change.\n\
\t\t\t\t\t\tIf they belong to the same basis, we have a Kronecker delta \
(products of the elements of dual bases)."
}], "Text",
 CellChangeTimes->{{3.728661455805543*^9, 3.728661530090159*^9}, {
  3.728661561007251*^9, 3.7286615803757763`*^9}, {3.728742762246937*^9, 
  3.728742764231371*^9}},ExpressionUUID->"d34d2147-8c98-4d66-b59f-\
1f302fc13ae7"],

Cell[CellGroupData[{

Cell["Basis is output as an e, except when it converts into delta:", \
"MathCaption",
 CellChangeTimes->{{3.728661588286262*^9, 
  3.728661606446005*^9}},ExpressionUUID->"76a5a604-83dc-4878-831c-\
146276a3514a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "fred"}]}], "}"}], ",", "a"}], "]"}], ",", " ", 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", "b"}], "]"}], ",", " ", 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "blues"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"b", ",", "fred"}], "}"}]}], "]"}], ",", 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "blues"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"b", ",", "blues"}], "}"}]}], "]"}], ",", "  ", 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "fred"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", "fred"}], "}"}]}], "]"}], ",", " ", 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "fred"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"2", ",", "fred"}], "}"}]}], "]"}]}], "}"}]], "Input",
 CellChangeTimes->{{3.728742778262514*^9, 3.7287427941028233`*^9}, {
  3.728747328600923*^9, 3.728747346711525*^9}, {3.72874928319238*^9, 
  3.7287492843490763`*^9}},
 CellLabel->"In[83]:=",ExpressionUUID->"468fa3bd-ce86-4aa7-ad63-104127e19477"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {" ", "b"},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`delta[-$CellContext`a, $CellContext`b],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["b",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["a",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`a, -$CellContext`blues}, \
{$CellContext`b, $CellContext`fred}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["b",
             FontColor->RGBColor[0., 0., 1.]]},
           {
            StyleBox["a",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`a, -$CellContext`blues}, \
{$CellContext`b, $CellContext`blues}],
    Editable->False], ",", "1", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.728661608258806*^9, 3.728671238071991*^9, 
  3.728674884988592*^9, 3.728742810486932*^9, 3.728743299802369*^9, 
  3.7287433964477377`*^9, 3.728747347143012*^9, 3.728749509213345*^9, 
  3.7287499375631027`*^9, 3.7287504119943542`*^9, 3.728750616952544*^9, 
  3.728750768855591*^9, 3.728751058519384*^9, 3.7287516247447977`*^9, 
  3.728813269964263*^9, 3.728827386973833*^9, 3.728835978219914*^9, 
  3.72883629518077*^9, 3.728843091735166*^9, 3.7288456295356627`*^9, 
  3.733367360166727*^9, 3.792273274712234*^9},
 CellLabel->"Out[83]=",
 CellID->842480991,ExpressionUUID->"0a565396-33c1-43ce-96a1-a91ac2a6c7ac"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
A Basis object with two indices of the same character (covariant or \
contravariant) is converted into a metric tensor. If we  have defined more \
than one metric, the first one is used (see xTensorDoc.nb for details).\
\>", "MathCaption",
 CellChangeTimes->{{3.7286616215969887`*^9, 
  3.7286616343163424`*^9}},ExpressionUUID->"ddd22766-b3c0-42af-a36b-\
3184ffae9202"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Catch", "@", 
  RowBox[{"Basis", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}]}], "]"}]}]], "Input",
 CellLabel->"In[84]:=",ExpressionUUID->"c578bbd3-765d-4ce0-adc2-c97919d73d2b"],

Cell[BoxData[
 TemplateBox[{
  "MetricsOfVBundle", "missing", 
   "\"There is no \\!\\(\\*RowBox[{\\\"\\\\\\\"metric\\\\\\\"\\\"}]\\) in \\!\
\\(\\*InterpretationBox[StyleBox[\\\"\[DoubleStruckCapitalT]M3\\\", \
Rule[AutoSpacing, False], Rule[ShowAutoStyles, False]], TangentM3, \
Rule[Editable, False]]\\).\"", 2, 84, 6, 19668387891272167742, "Local", 
   "xAct`xTensor`MetricsOfVBundle"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728661636908689*^9, 3.72867123822229*^9, 
  3.7286748853104687`*^9, 3.728742819965699*^9, 3.7287432999438057`*^9, 
  3.72874339656462*^9, 3.728747349521413*^9, 3.7287495106740294`*^9, 
  3.7287499376858253`*^9, 3.728750412142022*^9, 3.728750618184095*^9, 
  3.7287507691354733`*^9, 3.728751058664926*^9, 3.72875162491118*^9, 
  3.728813270140737*^9, 3.728827387130657*^9, 3.728835978354167*^9, 
  3.728836295326893*^9, 3.728843091896617*^9, 3.7288456296777763`*^9, 
  3.733367361516884*^9, 3.792273276544488*^9},
 CellLabel->"During evaluation of In[84]:=",
 CellID->585816200,ExpressionUUID->"3cec4acc-8b23-4480-b915-4b8c9e34d618"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefMetric", "[", 
  RowBox[{
   RowBox[{"-", "1"}], ",", 
   RowBox[{"metricg", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", "CD", ",", 
   RowBox[{"SymbolOfCovD", "->", 
    RowBox[{"{", 
     RowBox[{"\"\<;\>\"", ",", "\"\<\[EmptyDownTriangle]\>\""}], "}"}]}], ",", 
   RowBox[{"PrintAs", "->", "\"\<g\>\""}]}], "]"}]], "Input",
 CellLabel->"In[85]:=",ExpressionUUID->"adf56673-76fc-41bd-93bf-ad8c5a2b94f4"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric metric tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"metricg", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric metric tensor ", 
   $CellContext`metricg[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728661638455361*^9, 3.728671238335267*^9, 
  3.728674885590446*^9, 3.7287428220135603`*^9, 3.728743300081644*^9, 
  3.728743396682912*^9, 3.728747351111747*^9, 3.728749511320087*^9, 
  3.7287499378179483`*^9, 3.728750412318598*^9, 3.728750618744329*^9, 
  3.728750769588572*^9, 3.7287510588105593`*^9, 3.7287516250453987`*^9, 
  3.728813270305621*^9, 3.728827387290801*^9, 3.7288359784983664`*^9, 
  3.728836295487876*^9, 3.728843092171422*^9, 3.728845629860628*^9, 
  3.7333673621128187`*^9, 3.792273277153592*^9},
 CellLabel->"During evaluation of In[85]:=",
 CellID->1085112334,ExpressionUUID->"20864c41-33b6-4a0f-95c9-c285e9fcee58"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric tensor \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"epsilonmetricg", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric tensor ", 
   $CellContext`epsilonmetricg[-$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728661638455361*^9, 3.728671238335267*^9, 
  3.728674885590446*^9, 3.7287428220135603`*^9, 3.728743300081644*^9, 
  3.728743396682912*^9, 3.728747351111747*^9, 3.728749511320087*^9, 
  3.7287499378179483`*^9, 3.728750412318598*^9, 3.728750618744329*^9, 
  3.728750769588572*^9, 3.7287510588105593`*^9, 3.7287516250453987`*^9, 
  3.728813270305621*^9, 3.728827387290801*^9, 3.7288359784983664`*^9, 
  3.728836295487876*^9, 3.728843092171422*^9, 3.728845629860628*^9, 
  3.7333673621128187`*^9, 3.7922732771639633`*^9},
 CellLabel->"During evaluation of In[85]:=",
 CellID->1051867214,ExpressionUUID->"730608cc-55a0-4046-b25d-11ea92fada14"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"CD", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "covariant derivative ", 
   $CellContext`CD[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728661638455361*^9, 3.728671238335267*^9, 
  3.728674885590446*^9, 3.7287428220135603`*^9, 3.728743300081644*^9, 
  3.728743396682912*^9, 3.728747351111747*^9, 3.728749511320087*^9, 
  3.7287499378179483`*^9, 3.728750412318598*^9, 3.728750618744329*^9, 
  3.728750769588572*^9, 3.7287510588105593`*^9, 3.7287516250453987`*^9, 
  3.728813270305621*^9, 3.728827387290801*^9, 3.7288359784983664`*^9, 
  3.728836295487876*^9, 3.728843092171422*^9, 3.728845629860628*^9, 
  3.7333673621128187`*^9, 3.7922732771793013`*^9},
 CellLabel->"During evaluation of In[85]:=",
 CellID->646883594,ExpressionUUID->"085a7d4b-6ee8-44a3-bb4f-0b7713915333"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing torsion tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TorsionCD", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing torsion tensor ", 
   $CellContext`TorsionCD[$CellContext`a, -$CellContext`b, -$CellContext`c], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728661638455361*^9, 3.728671238335267*^9, 
  3.728674885590446*^9, 3.7287428220135603`*^9, 3.728743300081644*^9, 
  3.728743396682912*^9, 3.728747351111747*^9, 3.728749511320087*^9, 
  3.7287499378179483`*^9, 3.728750412318598*^9, 3.728750618744329*^9, 
  3.728750769588572*^9, 3.7287510588105593`*^9, 3.7287516250453987`*^9, 
  3.728813270305621*^9, 3.728827387290801*^9, 3.7288359784983664`*^9, 
  3.728836295487876*^9, 3.728843092171422*^9, 3.728845629860628*^9, 
  3.7333673621128187`*^9, 3.792273277187532*^9},
 CellLabel->"During evaluation of In[85]:=",
 CellID->1610343244,ExpressionUUID->"24824b51-6c3b-4f1c-b8ad-d70855c9d57b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelCD", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Christoffel tensor ", 
   $CellContext`ChristoffelCD[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728661638455361*^9, 3.728671238335267*^9, 
  3.728674885590446*^9, 3.7287428220135603`*^9, 3.728743300081644*^9, 
  3.728743396682912*^9, 3.728747351111747*^9, 3.728749511320087*^9, 
  3.7287499378179483`*^9, 3.728750412318598*^9, 3.728750618744329*^9, 
  3.728750769588572*^9, 3.7287510588105593`*^9, 3.7287516250453987`*^9, 
  3.728813270305621*^9, 3.728827387290801*^9, 3.7288359784983664`*^9, 
  3.728836295487876*^9, 3.728843092171422*^9, 3.728845629860628*^9, 
  3.7333673621128187`*^9, 3.792273277195475*^9},
 CellLabel->"During evaluation of In[85]:=",
 CellID->302498767,ExpressionUUID->"b2b85861-ec88-4330-a61c-904933638c6a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Riemann tensor ", 
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728661638455361*^9, 3.728671238335267*^9, 
  3.728674885590446*^9, 3.7287428220135603`*^9, 3.728743300081644*^9, 
  3.728743396682912*^9, 3.728747351111747*^9, 3.728749511320087*^9, 
  3.7287499378179483`*^9, 3.728750412318598*^9, 3.728750618744329*^9, 
  3.728750769588572*^9, 3.7287510588105593`*^9, 3.7287516250453987`*^9, 
  3.728813270305621*^9, 3.728827387290801*^9, 3.7288359784983664`*^9, 
  3.728836295487876*^9, 3.728843092171422*^9, 3.728845629860628*^9, 
  3.7333673621128187`*^9, 3.792273277203349*^9},
 CellLabel->"During evaluation of In[85]:=",
 CellID->1002984520,ExpressionUUID->"a3ca23a7-b1dd-404b-b157-bf8fc6796c80"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric Ricci tensor ", 
   $CellContext`RicciCD[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728661638455361*^9, 3.728671238335267*^9, 
  3.728674885590446*^9, 3.7287428220135603`*^9, 3.728743300081644*^9, 
  3.728743396682912*^9, 3.728747351111747*^9, 3.728749511320087*^9, 
  3.7287499378179483`*^9, 3.728750412318598*^9, 3.728750618744329*^9, 
  3.728750769588572*^9, 3.7287510588105593`*^9, 3.7287516250453987`*^9, 
  3.728813270305621*^9, 3.728827387290801*^9, 3.7288359784983664`*^9, 
  3.728836295487876*^9, 3.728843092171422*^9, 3.728845629860628*^9, 
  3.7333673621128187`*^9, 3.792273277211154*^9},
 CellLabel->"During evaluation of In[85]:=",
 CellID->1561682337,ExpressionUUID->"83a4a250-88ed-450e-a417-71a9d511e8b3"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Riemann automatically replaced \
by Ricci.\"\>"], "Print",
 CellChangeTimes->{3.728661638455361*^9, 3.728671238335267*^9, 
  3.728674885590446*^9, 3.7287428220135603`*^9, 3.728743300081644*^9, 
  3.728743396682912*^9, 3.728747351111747*^9, 3.728749511320087*^9, 
  3.7287499378179483`*^9, 3.728750412318598*^9, 3.728750618744329*^9, 
  3.728750769588572*^9, 3.7287510588105593`*^9, 3.7287516250453987`*^9, 
  3.728813270305621*^9, 3.728827387290801*^9, 3.7288359784983664`*^9, 
  3.728836295487876*^9, 3.728843092171422*^9, 3.728845629860628*^9, 
  3.7333673621128187`*^9, 3.792273277218889*^9},
 CellLabel->"During evaluation of In[85]:=",
 CellID->649904325,ExpressionUUID->"90d12eb2-3f31-40e3-ac3f-3dd78d9c7516"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"RicciScalarCD", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Ricci scalar ", 
   $CellContext`RicciScalarCD[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728661638455361*^9, 3.728671238335267*^9, 
  3.728674885590446*^9, 3.7287428220135603`*^9, 3.728743300081644*^9, 
  3.728743396682912*^9, 3.728747351111747*^9, 3.728749511320087*^9, 
  3.7287499378179483`*^9, 3.728750412318598*^9, 3.728750618744329*^9, 
  3.728750769588572*^9, 3.7287510588105593`*^9, 3.7287516250453987`*^9, 
  3.728813270305621*^9, 3.728827387290801*^9, 3.7288359784983664`*^9, 
  3.728836295487876*^9, 3.728843092171422*^9, 3.728845629860628*^9, 
  3.7333673621128187`*^9, 3.7922732772262373`*^9},
 CellLabel->"During evaluation of In[85]:=",
 CellID->840161633,ExpressionUUID->"69fb9bd7-0128-4d12-8f6d-626f92afc21a"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Ricci automatically replaced \
by RicciScalar.\"\>"], "Print",
 CellChangeTimes->{3.728661638455361*^9, 3.728671238335267*^9, 
  3.728674885590446*^9, 3.7287428220135603`*^9, 3.728743300081644*^9, 
  3.728743396682912*^9, 3.728747351111747*^9, 3.728749511320087*^9, 
  3.7287499378179483`*^9, 3.728750412318598*^9, 3.728750618744329*^9, 
  3.728750769588572*^9, 3.7287510588105593`*^9, 3.7287516250453987`*^9, 
  3.728813270305621*^9, 3.728827387290801*^9, 3.7288359784983664`*^9, 
  3.728836295487876*^9, 3.728843092171422*^9, 3.728845629860628*^9, 
  3.7333673621128187`*^9, 3.7922732772339287`*^9},
 CellLabel->"During evaluation of In[85]:=",
 CellID->891586806,ExpressionUUID->"28354677-d7f6-4b2d-bce1-2bb26931fe8c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Einstein tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"EinsteinCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Einstein tensor ", 
   $CellContext`EinsteinCD[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728661638455361*^9, 3.728671238335267*^9, 
  3.728674885590446*^9, 3.7287428220135603`*^9, 3.728743300081644*^9, 
  3.728743396682912*^9, 3.728747351111747*^9, 3.728749511320087*^9, 
  3.7287499378179483`*^9, 3.728750412318598*^9, 3.728750618744329*^9, 
  3.728750769588572*^9, 3.7287510588105593`*^9, 3.7287516250453987`*^9, 
  3.728813270305621*^9, 3.728827387290801*^9, 3.7288359784983664`*^9, 
  3.728836295487876*^9, 3.728843092171422*^9, 3.728845629860628*^9, 
  3.7333673621128187`*^9, 3.7922732772411203`*^9},
 CellLabel->"During evaluation of In[85]:=",
 CellID->1906043323,ExpressionUUID->"64173909-59c4-4153-996e-5da798772610"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Weyl tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"WeylCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Weyl tensor ", 
   $CellContext`WeylCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728661638455361*^9, 3.728671238335267*^9, 
  3.728674885590446*^9, 3.7287428220135603`*^9, 3.728743300081644*^9, 
  3.728743396682912*^9, 3.728747351111747*^9, 3.728749511320087*^9, 
  3.7287499378179483`*^9, 3.728750412318598*^9, 3.728750618744329*^9, 
  3.728750769588572*^9, 3.7287510588105593`*^9, 3.7287516250453987`*^9, 
  3.728813270305621*^9, 3.728827387290801*^9, 3.7288359784983664`*^9, 
  3.728836295487876*^9, 3.728843092171422*^9, 3.728845629860628*^9, 
  3.7333673621128187`*^9, 3.792273277247841*^9},
 CellLabel->"During evaluation of In[85]:=",
 CellID->856567362,ExpressionUUID->"f46be9e9-b6fa-4e8e-81b9-9edeb1cb37f5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric TFRicci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TFRicciCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric TFRicci tensor ", 
   $CellContext`TFRicciCD[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728661638455361*^9, 3.728671238335267*^9, 
  3.728674885590446*^9, 3.7287428220135603`*^9, 3.728743300081644*^9, 
  3.728743396682912*^9, 3.728747351111747*^9, 3.728749511320087*^9, 
  3.7287499378179483`*^9, 3.728750412318598*^9, 3.728750618744329*^9, 
  3.728750769588572*^9, 3.7287510588105593`*^9, 3.7287516250453987`*^9, 
  3.728813270305621*^9, 3.728827387290801*^9, 3.7288359784983664`*^9, 
  3.728836295487876*^9, 3.728843092171422*^9, 3.728845629860628*^9, 
  3.7333673621128187`*^9, 3.792273277255437*^9},
 CellLabel->"During evaluation of In[85]:=",
 CellID->1744767293,ExpressionUUID->"d8c0ab6f-ac98-4a67-bc2e-352c2fd26079"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Kretschmann scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"KretschmannCD", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Kretschmann scalar ", 
   $CellContext`KretschmannCD[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728661638455361*^9, 3.728671238335267*^9, 
  3.728674885590446*^9, 3.7287428220135603`*^9, 3.728743300081644*^9, 
  3.728743396682912*^9, 3.728747351111747*^9, 3.728749511320087*^9, 
  3.7287499378179483`*^9, 3.728750412318598*^9, 3.728750618744329*^9, 
  3.728750769588572*^9, 3.7287510588105593`*^9, 3.7287516250453987`*^9, 
  3.728813270305621*^9, 3.728827387290801*^9, 3.7288359784983664`*^9, 
  3.728836295487876*^9, 3.728843092171422*^9, 3.728845629860628*^9, 
  3.7333673621128187`*^9, 3.79227327726305*^9},
 CellLabel->"During evaluation of In[85]:=",
 CellID->414693666,ExpressionUUID->"2d333f68-4b95-4dd0-9ae8-9d758b94bf28"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RiemannToWeylRules for dim \"\>", 
   "\[InvisibleSpace]", "3"}],
  SequenceForm["** DefCovD:  Computing RiemannToWeylRules for dim ", 3],
  Editable->False]], "Print",
 CellChangeTimes->{3.728661638455361*^9, 3.728671238335267*^9, 
  3.728674885590446*^9, 3.7287428220135603`*^9, 3.728743300081644*^9, 
  3.728743396682912*^9, 3.728747351111747*^9, 3.728749511320087*^9, 
  3.7287499378179483`*^9, 3.728750412318598*^9, 3.728750618744329*^9, 
  3.728750769588572*^9, 3.7287510588105593`*^9, 3.7287516250453987`*^9, 
  3.728813270305621*^9, 3.728827387290801*^9, 3.7288359784983664`*^9, 
  3.728836295487876*^9, 3.728843092171422*^9, 3.728845629860628*^9, 
  3.7333673621128187`*^9, 3.792273277270741*^9},
 CellLabel->"During evaluation of In[85]:=",
 CellID->1172009947,ExpressionUUID->"d85048b7-1dbb-4b7e-bfdb-0d10114c74f3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RicciToTFRicci for dim \"\>", 
   "\[InvisibleSpace]", "3"}],
  SequenceForm["** DefCovD:  Computing RicciToTFRicci for dim ", 3],
  Editable->False]], "Print",
 CellChangeTimes->{3.728661638455361*^9, 3.728671238335267*^9, 
  3.728674885590446*^9, 3.7287428220135603`*^9, 3.728743300081644*^9, 
  3.728743396682912*^9, 3.728747351111747*^9, 3.728749511320087*^9, 
  3.7287499378179483`*^9, 3.728750412318598*^9, 3.728750618744329*^9, 
  3.728750769588572*^9, 3.7287510588105593`*^9, 3.7287516250453987`*^9, 
  3.728813270305621*^9, 3.728827387290801*^9, 3.7288359784983664`*^9, 
  3.728836295487876*^9, 3.728843092171422*^9, 3.728845629860628*^9, 
  3.7333673621128187`*^9, 3.792273277278487*^9},
 CellLabel->"During evaluation of In[85]:=",
 CellID->1251168945,ExpressionUUID->"be26bf36-91cf-408c-b230-e70b976bad1e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RicciToEinsteinRules for dim \"\>", 
   "\[InvisibleSpace]", "3"}],
  SequenceForm["** DefCovD:  Computing RicciToEinsteinRules for dim ", 3],
  Editable->False]], "Print",
 CellChangeTimes->{3.728661638455361*^9, 3.728671238335267*^9, 
  3.728674885590446*^9, 3.7287428220135603`*^9, 3.728743300081644*^9, 
  3.728743396682912*^9, 3.728747351111747*^9, 3.728749511320087*^9, 
  3.7287499378179483`*^9, 3.728750412318598*^9, 3.728750618744329*^9, 
  3.728750769588572*^9, 3.7287510588105593`*^9, 3.7287516250453987`*^9, 
  3.728813270305621*^9, 3.728827387290801*^9, 3.7288359784983664`*^9, 
  3.728836295487876*^9, 3.728843092171422*^9, 3.728845629860628*^9, 
  3.7333673621128187`*^9, 3.792273277286379*^9},
 CellLabel->"During evaluation of In[85]:=",
 CellID->236587191,ExpressionUUID->"7f4ec9d4-004d-43b9-b4ea-e93fd7c48ff6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density \"\>", "\[InvisibleSpace]", 
   RowBox[{"Detmetricg", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Determinant.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "weight +2 density ", 
   $CellContext`Detmetricg[], ". ", "Determinant."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728661638455361*^9, 3.728671238335267*^9, 
  3.728674885590446*^9, 3.7287428220135603`*^9, 3.728743300081644*^9, 
  3.728743396682912*^9, 3.728747351111747*^9, 3.728749511320087*^9, 
  3.7287499378179483`*^9, 3.728750412318598*^9, 3.728750618744329*^9, 
  3.728750769588572*^9, 3.7287510588105593`*^9, 3.7287516250453987`*^9, 
  3.728813270305621*^9, 3.728827387290801*^9, 3.7288359784983664`*^9, 
  3.728836295487876*^9, 3.728843092171422*^9, 3.728845629860628*^9, 
  3.7333673621128187`*^9, 3.792273277324883*^9},
 CellLabel->"During evaluation of In[85]:=",
 CellID->1328688540,ExpressionUUID->"59ba8c7e-b9f9-4809-8a9e-0544a8ce1967"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", " ", 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"a", ",", "blues"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"b", ",", "blues"}], "}"}]}], "]"}], ",", " ", 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"a", ",", "blues"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"b", ",", "fred"}], "}"}]}], "]"}]}], "}"}]], "Input",
 CellChangeTimes->{{3.728742826924197*^9, 3.728742832397194*^9}, {
  3.728747354520946*^9, 3.7287473612949533`*^9}, {3.728749285331777*^9, 
  3.7287492866894617`*^9}},
 CellLabel->"In[86]:=",ExpressionUUID->"030e90da-d725-49ed-9995-ef89684cd6c2"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`a, -$CellContext`b],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {
            StyleBox["a",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["b",
             FontColor->RGBColor[0., 0., 1.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{$CellContext`a, $CellContext`blues}, \
{$CellContext`b, $CellContext`blues}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {
            StyleBox["a",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["b",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{$CellContext`a, $CellContext`blues}, \
{$CellContext`b, $CellContext`fred}],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728661643063468*^9, 3.7286712389318123`*^9, 
  3.728674886450713*^9, 3.728742832988544*^9, 3.728743300703643*^9, 
  3.728743397231599*^9, 3.728747361945566*^9, 3.728749513191443*^9, 
  3.728749938149901*^9, 3.72875041266987*^9, 3.728750620369935*^9, 
  3.728750770134644*^9, 3.728751059212175*^9, 3.7287516254445744`*^9, 
  3.7288132707477713`*^9, 3.728827387757455*^9, 3.728835979265053*^9, 
  3.728836295919467*^9, 3.728843093517049*^9, 3.728845630462747*^9, 
  3.7333673636276827`*^9, 3.7922732787005*^9},
 CellLabel->"Out[86]=",
 CellID->1499578617,ExpressionUUID->"6976e279-f426-4d9d-8788-c3acb4c0b61f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Now other index configurations are allowed, but transformed automatically \
into the original down-up configuration:\
\>", "MathCaption",ExpressionUUID->"eef2cdde-1048-4cb2-bd37-94b0ab1e03a8"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"Basis", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"a", ",", "fred"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "blues"}]}], "}"}]}], "]"}]}], "}"}]], "Input",
 CellChangeTimes->{{3.728742836595674*^9, 3.728742838573037*^9}, {
   3.7287473657749567`*^9, 3.728747368214006*^9}, 3.728749287750205*^9},
 CellLabel->"In[87]:=",ExpressionUUID->"08565ed1-7cb9-4ead-a740-a5523f686629"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {" ", "a"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`delta[-$CellContext`b, $CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["a",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["b",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`b, -$CellContext`blues}, \
{$CellContext`a, $CellContext`fred}],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.7286616456259317`*^9, 3.7286712390782747`*^9, 
  3.728674886587406*^9, 3.728742839151299*^9, 3.7287433008352003`*^9, 
  3.728743397365224*^9, 3.728747368809691*^9, 3.728749514253943*^9, 
  3.728749938265368*^9, 3.7287504128263206`*^9, 3.7287506214396267`*^9, 
  3.728750770350238*^9, 3.728751059397326*^9, 3.728751625595461*^9, 
  3.728813270914156*^9, 3.728827387913743*^9, 3.72883597949817*^9, 
  3.728836296070532*^9, 3.728843094367774*^9, 3.728845630603524*^9, 
  3.7333673651003923`*^9, 3.7922732808982162`*^9},
 CellLabel->"Out[87]=",
 CellID->1069677556,ExpressionUUID->"eafae9e1-9734-41d3-a6fd-34b24cdc25fe"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
When a vector bundle has a metric, delta and Basis inherit the symmetry of \
that metric.\
\>", "MathCaption",
 CellChangeTimes->{{3.728661657851058*^9, 3.7286616843458*^9}},
 CellID->1116390537,ExpressionUUID->"95ca4695-08b7-4804-943d-72d469c0fe01"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SymmetryGroupOfTensor", "[", 
  RowBox[{"Basis", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"1", ",", "fred"}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "2"}], ",", "blues"}], "}"}]}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.7287428438849154`*^9, 3.728742847392952*^9}, {
   3.728747372366438*^9, 3.728747375670146*^9}, 3.728749290467914*^9},
 CellLabel->"In[88]:=",ExpressionUUID->"ba477352-b254-44b7-bfd8-56f7ed5f643b"],

Cell[BoxData[
 RowBox[{"StrongGenSet", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "2"}], "}"}], ",", 
   RowBox[{"GenSet", "[", 
    InterpretationBox[
     StyleBox[
      RowBox[{"(", "1", ",", "2", ")"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xPerm`Cycles[{1, 2}],
     Editable->False], "]"}]}], "]"}]], "Output",
 CellChangeTimes->{
  3.728661646465723*^9, 3.728671239238368*^9, 3.728674886745989*^9, 
   3.7287428480546017`*^9, 3.728743300967203*^9, 3.7287433974813633`*^9, {
   3.728747373554792*^9, 3.728747376134741*^9}, 3.728749515402854*^9, 
   3.72874993841537*^9, 3.728750412972868*^9, 3.72875062250322*^9, 
   3.7287507704975357`*^9, 3.7287510596445436`*^9, 3.728751625761207*^9, 
   3.72881327107134*^9, 3.728827388095648*^9, 3.728835979677929*^9, 
   3.728836296259377*^9, 3.7288430945576153`*^9, 3.728845630769259*^9, 
   3.733367366332876*^9, 3.792273282450046*^9},
 CellLabel->"Out[88]=",
 CellID->1181479492,ExpressionUUID->"d06c8e45-c4de-4288-a3e0-739ab116cb2f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefMetric", "[", "metricg", "]"}]], "Input",
 CellLabel->"In[89]:=",ExpressionUUID->"824892bd-03f6-421f-811c-8a38c1b6e85c"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Detmetricg"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "weight +2 density", 
   " ", $CellContext`Detmetricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286616909860077`*^9, 3.728671239373364*^9, 
  3.728674886881982*^9, 3.728742849824275*^9, 3.728743301136413*^9, 
  3.7287433976309443`*^9, 3.728747378762335*^9, 3.728749516005363*^9, 
  3.728749938548609*^9, 3.72875041310998*^9, 3.728750623064906*^9, 
  3.72875077066613*^9, 3.72875105978465*^9, 3.728751625894623*^9, 
  3.728813271234043*^9, 3.7288273882579536`*^9, 3.7288359798659163`*^9, 
  3.728836296401224*^9, 3.728843094801313*^9, 3.728845630904367*^9, 
  3.733367366864629*^9, 3.792273283000696*^9},
 CellLabel->"During evaluation of In[89]:=",
 CellID->25712005,ExpressionUUID->"86b6cda9-7f37-4e8b-8382-f637182db9fb"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "ChristoffelCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Christoffel tensor", " ", $CellContext`ChristoffelCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286616909860077`*^9, 3.728671239373364*^9, 
  3.728674886881982*^9, 3.728742849824275*^9, 3.728743301136413*^9, 
  3.7287433976309443`*^9, 3.728747378762335*^9, 3.728749516005363*^9, 
  3.728749938548609*^9, 3.72875041310998*^9, 3.728750623064906*^9, 
  3.72875077066613*^9, 3.72875105978465*^9, 3.728751625894623*^9, 
  3.728813271234043*^9, 3.7288273882579536`*^9, 3.7288359798659163`*^9, 
  3.728836296401224*^9, 3.728843094801313*^9, 3.728845630904367*^9, 
  3.733367366864629*^9, 3.792273283012747*^9},
 CellLabel->"During evaluation of In[89]:=",
 CellID->497454988,ExpressionUUID->"6f0e02c4-dd25-4aaa-b37f-b0bfcdd644e5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Einstein tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "EinsteinCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Einstein tensor", " ", $CellContext`EinsteinCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286616909860077`*^9, 3.728671239373364*^9, 
  3.728674886881982*^9, 3.728742849824275*^9, 3.728743301136413*^9, 
  3.7287433976309443`*^9, 3.728747378762335*^9, 3.728749516005363*^9, 
  3.728749938548609*^9, 3.72875041310998*^9, 3.728750623064906*^9, 
  3.72875077066613*^9, 3.72875105978465*^9, 3.728751625894623*^9, 
  3.728813271234043*^9, 3.7288273882579536`*^9, 3.7288359798659163`*^9, 
  3.728836296401224*^9, 3.728843094801313*^9, 3.728845630904367*^9, 
  3.733367366864629*^9, 3.7922732830272083`*^9},
 CellLabel->"During evaluation of In[89]:=",
 CellID->572145545,ExpressionUUID->"8d8d962a-ce87-460b-a5d4-2fc73a144f5e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Kretschmann scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "KretschmannCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Kretschmann scalar", 
   " ", $CellContext`KretschmannCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286616909860077`*^9, 3.728671239373364*^9, 
  3.728674886881982*^9, 3.728742849824275*^9, 3.728743301136413*^9, 
  3.7287433976309443`*^9, 3.728747378762335*^9, 3.728749516005363*^9, 
  3.728749938548609*^9, 3.72875041310998*^9, 3.728750623064906*^9, 
  3.72875077066613*^9, 3.72875105978465*^9, 3.728751625894623*^9, 
  3.728813271234043*^9, 3.7288273882579536`*^9, 3.7288359798659163`*^9, 
  3.728836296401224*^9, 3.728843094801313*^9, 3.728845630904367*^9, 
  3.733367366864629*^9, 3.792273283034501*^9},
 CellLabel->"During evaluation of In[89]:=",
 CellID->1765887196,ExpressionUUID->"7ae9f1cf-7849-4987-969b-58ca16e587b3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric Ricci tensor", 
   " ", $CellContext`RicciCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286616909860077`*^9, 3.728671239373364*^9, 
  3.728674886881982*^9, 3.728742849824275*^9, 3.728743301136413*^9, 
  3.7287433976309443`*^9, 3.728747378762335*^9, 3.728749516005363*^9, 
  3.728749938548609*^9, 3.72875041310998*^9, 3.728750623064906*^9, 
  3.72875077066613*^9, 3.72875105978465*^9, 3.728751625894623*^9, 
  3.728813271234043*^9, 3.7288273882579536`*^9, 3.7288359798659163`*^9, 
  3.728836296401224*^9, 3.728843094801313*^9, 3.728845630904367*^9, 
  3.733367366864629*^9, 3.792273283041925*^9},
 CellLabel->"During evaluation of In[89]:=",
 CellID->1042101632,ExpressionUUID->"c4324fc8-77d0-49e2-9f2f-914cc234a4fe"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciScalarCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Ricci scalar", 
   " ", $CellContext`RicciScalarCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286616909860077`*^9, 3.728671239373364*^9, 
  3.728674886881982*^9, 3.728742849824275*^9, 3.728743301136413*^9, 
  3.7287433976309443`*^9, 3.728747378762335*^9, 3.728749516005363*^9, 
  3.728749938548609*^9, 3.72875041310998*^9, 3.728750623064906*^9, 
  3.72875077066613*^9, 3.72875105978465*^9, 3.728751625894623*^9, 
  3.728813271234043*^9, 3.7288273882579536`*^9, 3.7288359798659163`*^9, 
  3.728836296401224*^9, 3.728843094801313*^9, 3.728845630904367*^9, 
  3.733367366864629*^9, 3.792273283049142*^9},
 CellLabel->"During evaluation of In[89]:=",
 CellID->1872987739,ExpressionUUID->"cc17e843-3040-4928-9075-5f5208017b0c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286616909860077`*^9, 3.728671239373364*^9, 
  3.728674886881982*^9, 3.728742849824275*^9, 3.728743301136413*^9, 
  3.7287433976309443`*^9, 3.728747378762335*^9, 3.728749516005363*^9, 
  3.728749938548609*^9, 3.72875041310998*^9, 3.728750623064906*^9, 
  3.72875077066613*^9, 3.72875105978465*^9, 3.728751625894623*^9, 
  3.728813271234043*^9, 3.7288273882579536`*^9, 3.7288359798659163`*^9, 
  3.728836296401224*^9, 3.728843094801313*^9, 3.728845630904367*^9, 
  3.733367366864629*^9, 3.792273283056467*^9},
 CellLabel->"During evaluation of In[89]:=",
 CellID->1489631647,ExpressionUUID->"36b9b1d6-7bbe-4b9c-8b47-d80a532b9b09"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric TFRicci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TFRicciCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric TFRicci tensor",
    " ", $CellContext`TFRicciCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286616909860077`*^9, 3.728671239373364*^9, 
  3.728674886881982*^9, 3.728742849824275*^9, 3.728743301136413*^9, 
  3.7287433976309443`*^9, 3.728747378762335*^9, 3.728749516005363*^9, 
  3.728749938548609*^9, 3.72875041310998*^9, 3.728750623064906*^9, 
  3.72875077066613*^9, 3.72875105978465*^9, 3.728751625894623*^9, 
  3.728813271234043*^9, 3.7288273882579536`*^9, 3.7288359798659163`*^9, 
  3.728836296401224*^9, 3.728843094801313*^9, 3.728845630904367*^9, 
  3.733367366864629*^9, 3.792273283064126*^9},
 CellLabel->"During evaluation of In[89]:=",
 CellID->874734403,ExpressionUUID->"a15af2ba-cb1c-434c-8594-3e4729bdf953"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TorsionCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286616909860077`*^9, 3.728671239373364*^9, 
  3.728674886881982*^9, 3.728742849824275*^9, 3.728743301136413*^9, 
  3.7287433976309443`*^9, 3.728747378762335*^9, 3.728749516005363*^9, 
  3.728749938548609*^9, 3.72875041310998*^9, 3.728750623064906*^9, 
  3.72875077066613*^9, 3.72875105978465*^9, 3.728751625894623*^9, 
  3.728813271234043*^9, 3.7288273882579536`*^9, 3.7288359798659163`*^9, 
  3.728836296401224*^9, 3.728843094801313*^9, 3.728845630904367*^9, 
  3.733367366864629*^9, 3.792273283071553*^9},
 CellLabel->"During evaluation of In[89]:=",
 CellID->2135547506,ExpressionUUID->"8abc0fef-1867-4b06-bf48-37fbfa2a04e0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Weyl tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "WeylCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Weyl tensor", 
   " ", $CellContext`WeylCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286616909860077`*^9, 3.728671239373364*^9, 
  3.728674886881982*^9, 3.728742849824275*^9, 3.728743301136413*^9, 
  3.7287433976309443`*^9, 3.728747378762335*^9, 3.728749516005363*^9, 
  3.728749938548609*^9, 3.72875041310998*^9, 3.728750623064906*^9, 
  3.72875077066613*^9, 3.72875105978465*^9, 3.728751625894623*^9, 
  3.728813271234043*^9, 3.7288273882579536`*^9, 3.7288359798659163`*^9, 
  3.728836296401224*^9, 3.728843094801313*^9, 3.728845630904367*^9, 
  3.733367366864629*^9, 3.7922732830785427`*^9},
 CellLabel->"During evaluation of In[89]:=",
 CellID->1252741031,ExpressionUUID->"83e698ab-3f75-469d-826e-d588eca2e1e2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "CD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "covariant derivative", 
   " ", $CellContext`CD],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286616909860077`*^9, 3.728671239373364*^9, 
  3.728674886881982*^9, 3.728742849824275*^9, 3.728743301136413*^9, 
  3.7287433976309443`*^9, 3.728747378762335*^9, 3.728749516005363*^9, 
  3.728749938548609*^9, 3.72875041310998*^9, 3.728750623064906*^9, 
  3.72875077066613*^9, 3.72875105978465*^9, 3.728751625894623*^9, 
  3.728813271234043*^9, 3.7288273882579536`*^9, 3.7288359798659163`*^9, 
  3.728836296401224*^9, 3.728843094801313*^9, 3.728845630904367*^9, 
  3.733367366864629*^9, 3.79227328308562*^9},
 CellLabel->"During evaluation of In[89]:=",
 CellID->888832829,ExpressionUUID->"1a73b07e-d5a8-4dd5-8a4d-3815091c9092"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "epsilonmetricg"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric tensor", 
   " ", $CellContext`epsilonmetricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286616909860077`*^9, 3.728671239373364*^9, 
  3.728674886881982*^9, 3.728742849824275*^9, 3.728743301136413*^9, 
  3.7287433976309443`*^9, 3.728747378762335*^9, 3.728749516005363*^9, 
  3.728749938548609*^9, 3.72875041310998*^9, 3.728750623064906*^9, 
  3.72875077066613*^9, 3.72875105978465*^9, 3.728751625894623*^9, 
  3.728813271234043*^9, 3.7288273882579536`*^9, 3.7288359798659163`*^9, 
  3.728836296401224*^9, 3.728843094801313*^9, 3.728845630904367*^9, 
  3.733367366864629*^9, 3.7922732830938663`*^9},
 CellLabel->"During evaluation of In[89]:=",
 CellID->717108471,ExpressionUUID->"2b3c0178-692c-4b1f-b092-37441bde7ea6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric metric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "metricg"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric metric tensor", 
   " ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286616909860077`*^9, 3.728671239373364*^9, 
  3.728674886881982*^9, 3.728742849824275*^9, 3.728743301136413*^9, 
  3.7287433976309443`*^9, 3.728747378762335*^9, 3.728749516005363*^9, 
  3.728749938548609*^9, 3.72875041310998*^9, 3.728750623064906*^9, 
  3.72875077066613*^9, 3.72875105978465*^9, 3.728751625894623*^9, 
  3.728813271234043*^9, 3.7288273882579536`*^9, 3.7288359798659163`*^9, 
  3.728836296401224*^9, 3.728843094801313*^9, 3.728845630904367*^9, 
  3.733367366864629*^9, 3.79227328310098*^9},
 CellLabel->"During evaluation of In[89]:=",
 CellID->763845133,ExpressionUUID->"2ec132e4-e810-40f8-a567-f4ec448c5324"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell["More on Basis contraction.", "Text",
 CellChangeTimes->{{3.728662296961779*^9, 3.7286623058884783`*^9}},
 CellID->608198706,ExpressionUUID->"45cdac70-2586-456c-8eba-efb49ce35c18"],

Cell[CellGroupData[{

Cell["\<\
An important issue is when to contract Basis objects with other tensors and \
with one another:\
\>", "MathCaption",ExpressionUUID->"559dd06d-60d4-449f-9b2b-128865ba1876"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ContractBasis", "/@", 
  RowBox[{"{", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"Basis", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"{", 
        RowBox[{"b", ",", "fred"}], "}"}]}], "]"}], 
     RowBox[{"Basis", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "c"}], ",", 
         RowBox[{"-", "blues"}]}], "}"}], ",", "a"}], "]"}]}], ",", 
    "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"Basis", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"1", ",", 
         RowBox[{"-", "fred"}]}], "}"}], ",", "a"}], "]"}], 
     RowBox[{"Basis", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"{", 
        RowBox[{"1", ",", "fred"}], "}"}]}], "]"}]}], ",", 
    "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"Basis", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "b"}], ",", 
         RowBox[{"-", "fred"}]}], "}"}], ",", "a"}], "]"}], 
     RowBox[{"Basis", "[", 
      RowBox[{
       RowBox[{"-", "c"}], ",", 
       RowBox[{"{", 
        RowBox[{"b", ",", "fred"}], "}"}]}], "]"}]}]}], "}"}], " "}]], "Input",\

 CellChangeTimes->{{3.728742860058667*^9, 3.728742874419442*^9}, {
   3.728747381119966*^9, 3.7287473925735693`*^9}, 3.728749291415968*^9},
 CellLabel->"In[90]:=",ExpressionUUID->"6c5caee3-0d93-45be-8241-a5e44aed8f86"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["b",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["c",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`c, -$CellContext`blues}, \
{$CellContext`b, $CellContext`fred}],
    Editable->False], ",", "1", ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {" ", "a"},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`delta[-$CellContext`c, $CellContext`a],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728662313013785*^9, 3.728671240050469*^9, 
  3.7286748873911037`*^9, 3.7287428752240343`*^9, 3.728743301518968*^9, 
  3.728743398029992*^9, 3.728747393028531*^9, 3.728749518736701*^9, 
  3.728749938816696*^9, 3.7287504133999643`*^9, 3.728750625119927*^9, 
  3.728750771428582*^9, 3.7287510600848627`*^9, 3.7287516261596737`*^9, 
  3.728813271597406*^9, 3.728827388638134*^9, 3.728835980381736*^9, 
  3.728836296874257*^9, 3.728843095585246*^9, 3.728845631255974*^9, 
  3.7333673687315693`*^9, 3.792273285377809*^9},
 CellLabel->"Out[90]=",
 CellID->1166833561,ExpressionUUID->"b96e7312-cc11-40be-a003-87053f5eefad"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "It would be desirable to automatize this behaviour sometimes. We can do it \
with the command ",
 StyleBox["AutomaticBasisContractionStart[]", "MR"],
 ":"
}], "MathCaption",ExpressionUUID->"ae04f47c-7779-45d2-a8ff-6c1756179d6c"],

Cell[TextData[{
 StyleBox["AutomaticBasisContractionStart\t\t", "DisplayMath"],
 StyleBox["Start automatic contraction of Basis objects", "Text"],
 StyleBox["\nAutomaticBasisContractionStop\t\t", "DisplayMath"],
 StyleBox["Stop automatic contraction of Basis objects", "Text"]
}], "DefinitionBox",ExpressionUUID->"3c54f885-109b-4734-90ee-89cccb8c4639"],

Cell[CellGroupData[{

Cell["Contraction of Basis objects.", "Text",
 CellChangeTimes->{
  3.728742883208961*^9},ExpressionUUID->"62a19e41-3d53-4baf-b6d8-\
2693966b5396"],

Cell[BoxData[
 RowBox[{"AutomaticBasisContractionStart", "[", "]"}]], "Input",
 CellLabel->"In[91]:=",ExpressionUUID->"c4f766fd-2ed4-4910-97f5-65ffeb5631e6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{
    RowBox[{
     RowBox[{"Basis", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"{", 
        RowBox[{"b", ",", "fred"}], "}"}]}], "]"}], 
     RowBox[{"Basis", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "c"}], ",", 
         RowBox[{"-", "blues"}]}], "}"}], ",", "a"}], "]"}]}], ",", 
    "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"Basis", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"1", ",", 
         RowBox[{"-", "fred"}]}], "}"}], ",", "a"}], "]"}], 
     RowBox[{"Basis", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"{", 
        RowBox[{"1", ",", "fred"}], "}"}]}], "]"}]}], ",", 
    "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"Basis", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "b"}], ",", 
         RowBox[{"-", "fred"}]}], "}"}], ",", "a"}], "]"}], 
     RowBox[{"Basis", "[", 
      RowBox[{
       RowBox[{"-", "c"}], ",", 
       RowBox[{"{", 
        RowBox[{"b", ",", "fred"}], "}"}]}], "]"}]}]}], "}"}], " "}]], "Input",\

 CellChangeTimes->{{3.72874289605062*^9, 3.7287429144335527`*^9}, {
   3.728747398240921*^9, 3.728747411348606*^9}, 3.728749292282937*^9},
 CellLabel->"In[92]:=",ExpressionUUID->"2fbe94f4-4d1b-4784-a55c-b1cbdd7c364b"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["b",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["c",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`c, -$CellContext`blues}, \
{$CellContext`b, $CellContext`fred}],
    Editable->False], ",", "1", ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {" ", "a"},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`delta[-$CellContext`c, $CellContext`a],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{
  3.728662324080068*^9, 3.728671240414238*^9, 3.7286748876006393`*^9, 
   3.72874291529953*^9, 3.728743301704653*^9, 3.728743398215102*^9, {
   3.728747412134615*^9, 3.72874741843577*^9}, 3.728749523401845*^9, 
   3.7287499390151997`*^9, 3.728750413683279*^9, 3.7287506281685247`*^9, 
   3.7287507718333397`*^9, 3.7287510602848253`*^9, 3.728751626368258*^9, 
   3.7288132717890663`*^9, 3.7288273889719877`*^9, 3.7288359806593246`*^9, 
   3.7288362970918016`*^9, 3.728843095875654*^9, 3.728845631464579*^9, 
   3.733367372285383*^9, 3.792273289425418*^9},
 CellLabel->"Out[92]=",
 CellID->2082240699,ExpressionUUID->"c3808519-8e25-4b73-a60c-6e99495e05b8"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"AutomaticBasisContractionStop", "[", "]"}]], "Input",
 CellLabel->"In[93]:=",ExpressionUUID->"0af93778-d4a6-4e12-b277-a1c839f8ca7a"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
When the contraction is between a Basis object and a different tensor, it is \
only performed if the former is a formal Kronecker delta or an identity \
tensor,\
\>", "MathCaption",ExpressionUUID->"5781a117-9b27-41d3-b919-37216dd0c452"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"Basis", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", "b"}], "]"}], 
    RowBox[{"T", "[", 
     RowBox[{
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}]}], "]"}]}], ",", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"Basis", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "a"}], ",", 
        RowBox[{"-", "fred"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"b", ",", "fred"}], "}"}]}], "]"}], " ", 
    RowBox[{"v", "[", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "fred"}]}], "}"}], "]"}]}], ",", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"Basis", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "a"}], ",", 
        RowBox[{"-", "fred"}]}], "}"}], ",", "b"}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "b"}], "]"}]}], ",", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"Basis", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "a"}], ",", 
        RowBox[{"-", "blues"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"b", ",", "fred"}], "}"}]}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "fred"}]}], "}"}], "]"}]}]}], "}"}]], "Input",
 CellChangeTimes->{{3.728742924855788*^9, 3.728742939776351*^9}, {
   3.728747422140414*^9, 3.72874744076338*^9}, 3.728749293139597*^9},
 CellLabel->"In[94]:=",ExpressionUUID->"c210dfb2-afb9-46f3-a745-60fff0813869"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[-$CellContext`a, -$CellContext`c],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {
            StyleBox["a",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[{-$CellContext`a, -$CellContext`fred}],
    Editable->False], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "b"},
            {
             StyleBox["a",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{-$CellContext`a, -$CellContext`fred}, $CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[-$CellContext`b],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["b",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{-$CellContext`a, -$CellContext`blues}, \
{$CellContext`b, $CellContext`fred}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["b",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[{-$CellContext`b, -$CellContext`fred}],
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728662327129163*^9, 3.728671240780653*^9, 
  3.72867488781712*^9, 3.728742940636722*^9, 3.72874330190103*^9, 
  3.7287433984302607`*^9, 3.728747441689652*^9, 3.7287495267160187`*^9, 
  3.728749939234132*^9, 3.728750413991691*^9, 3.728750630268087*^9, 
  3.728750772740086*^9, 3.728751060493637*^9, 3.7287516267009697`*^9, 
  3.7288132721199512`*^9, 3.728827389194763*^9, 3.728835980885695*^9, 
  3.728836297309002*^9, 3.728843096142869*^9, 3.72884563165425*^9, 
  3.733367374069059*^9, 3.792273291015128*^9},
 CellLabel->"Out[94]=",
 CellID->287309547,ExpressionUUID->"5cd7c505-3574-49db-8ba4-46d017cedfc2"]
}, Open  ]]
}, Open  ]],

Cell["\<\
To force the contraction in the last two cases we need a new function (next \
section). Notice how the third contraction is equivalent to expressing a \
tensor in its components and the fourth represents a basis change. \
\>", "MathCaption",ExpressionUUID->"afc732bd-e4d0-4f84-86fc-025f87d783b6"],

Cell["\<\
Basis formatting. We can define a special name  for the elements of one \
particular basis. \
\>", "Text",
 CellChangeTimes->{{3.728662341462467*^9, 
  3.728662344246704*^9}},ExpressionUUID->"97a4ad94-1860-4290-af3b-\
c6e824da2462"],

Cell[CellGroupData[{

Cell["Specify the format of a given basis vector object:", "MathCaption",
 CellChangeTimes->{{3.728742961958776*^9, 3.7287429757809553`*^9}},
 CellID->1298301299,ExpressionUUID->"d496bdab-0792-4b4d-9968-c0b15f004228"],

Cell[BoxData[
 RowBox[{"FormatBasis", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "fred"}], "}"}], ",", "\"\<n\>\""}], "]"}]], "Input",
 CellChangeTimes->{{3.728742979741571*^9, 3.728742980046142*^9}, {
  3.72874744812201*^9, 3.7287474486828537`*^9}},
 CellLabel->"In[95]:=",ExpressionUUID->"642d0f71-619f-4f53-b444-a151fe89bc4c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", "fred"}], "}"}]}], "]"}], ",", 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "fred"}]}], "}"}], ",", "a"}], "]"}], ",", 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"{", 
      RowBox[{"2", ",", "fred"}], "}"}]}], "]"}], ",", 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "blues"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", "fred"}], "}"}]}], "]"}], ",", 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "blues"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"2", ",", "fred"}], "}"}]}], "]"}]}], "}"}]], "Input",
 CellChangeTimes->{{3.728742982326457*^9, 3.7287429965416327`*^9}, {
  3.728747450468731*^9, 3.728747467026029*^9}, {3.7287492940733232`*^9, 
  3.728749294900197*^9}},
 CellLabel->"In[96]:=",ExpressionUUID->"f4b7b0ee-0d87-4a63-9e33-8eed6897eb48"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`fred}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`fred}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`blues}, {1, $CellContext`fred}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`blues}, {2, $CellContext`fred}],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{
  3.728662346833909*^9, 3.728671240989102*^9, 3.728674888066058*^9, 
   3.728742997563932*^9, 3.72874330208559*^9, 3.7287433986190557`*^9, {
   3.728747467586706*^9, 3.728747495128203*^9}, 3.728749529597672*^9, 
   3.728749939431364*^9, 3.728750414184514*^9, 3.728750633389709*^9, 
   3.728750772991611*^9, 3.7287510607120113`*^9, 3.728751626869123*^9, 
   3.72881327243596*^9, 3.728827389444408*^9, 3.72883598111872*^9, 
   3.728836297506646*^9, 3.728843096936516*^9, 3.728845631857239*^9, 
   3.7333673773480673`*^9, 3.792273297396103*^9},
 CellLabel->"Out[96]=",
 CellID->672679339,ExpressionUUID->"6a19051b-3782-46e5-9dea-9304defd5093"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"FormatBasis", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"2", ",", "fred"}], "}"}], ",", "\"\<l\>\""}], "]"}]], "Input",
 CellChangeTimes->{{3.728743005677487*^9, 3.7287430062692013`*^9}, {
  3.728747470513936*^9, 3.7287474710657673`*^9}},
 CellLabel->"In[97]:=",ExpressionUUID->"7a64673b-5605-4ea8-916c-bdfaed8ea27b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", "fred"}], "}"}]}], "]"}], ",", 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "fred"}]}], "}"}], ",", "a"}], "]"}], ",", 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"{", 
      RowBox[{"2", ",", "fred"}], "}"}]}], "]"}], ",", 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "blues"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", "fred"}], "}"}]}], "]"}], ",", 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "blues"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"2", ",", "fred"}], "}"}]}], "]"}]}], "}"}]], "Input",
 CellChangeTimes->{{3.728743009557383*^9, 3.728743024492346*^9}, {
  3.7287474765393057`*^9, 3.728747490056826*^9}, {3.7287492957987127`*^9, 
  3.7287492966769543`*^9}},
 CellLabel->"In[98]:=",ExpressionUUID->"de4efe2a-1ed1-42c8-99e1-d829a5090ffc"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`fred}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"l", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`fred}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`blues}, {1, $CellContext`fred}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"l", 
        StyleBox[GridBox[{
           {" "},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`blues}, {2, $CellContext`fred}],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{
  3.7286623488675413`*^9, 3.728671241772698*^9, 3.728674888434207*^9, 
   3.728743025496443*^9, 3.728743302349186*^9, 3.7287433988530407`*^9, {
   3.72874749097241*^9, 3.728747496645485*^9}, 3.728749530797489*^9, 
   3.728749939741631*^9, 3.728750414392314*^9, 3.728750634553755*^9, 
   3.728750773241551*^9, 3.7287510609281187`*^9, 3.72875162703512*^9, 
   3.728813272666911*^9, 3.728827389745471*^9, 3.72883598136464*^9, 
   3.728836297707798*^9, 3.7288430977035923`*^9, 3.728845632046257*^9, 
   3.733367379081904*^9, 3.792273298566042*^9},
 CellLabel->"Out[98]=",
 CellID->565320818,ExpressionUUID->"141fdadf-d032-42ca-bbd2-df91a85cc813"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We remove the special names:", "MathCaption",ExpressionUUID->"3961005f-d594-472d-a0a8-f641e0c1f219"],

Cell[BoxData[{
 RowBox[{"FormatBasis", "[", 
  RowBox[{"{", 
   RowBox[{"1", ",", "fred"}], "}"}], "]"}], "\[IndentingNewLine]", 
 RowBox[{"FormatBasis", "[", 
  RowBox[{"{", 
   RowBox[{"2", ",", "fred"}], "}"}], "]"}]}], "Input",
 CellChangeTimes->{{3.728743029074996*^9, 3.728743036323586*^9}, {
  3.728747504168293*^9, 3.7287475083840237`*^9}},
 CellLabel->"In[99]:=",ExpressionUUID->"c2d9f1d9-d136-44b0-8428-cecbe9730339"]
}, Open  ]],

Cell["Two more details.", "Text",
 CellChangeTimes->{{3.728662368726519*^9, 3.728662374908482*^9}},
 CellID->2035023769,ExpressionUUID->"1955df8b-2777-4c89-abe1-1b77490b65ad"],

Cell[CellGroupData[{

Cell["\<\
A bookkeeping detail: Basis is considered to be a tensor whose master is \
Symbol, so it cannot be undefined:\
\>", "MathCaption",
 CellChangeTimes->{{3.7286617212001247`*^9, 3.728661734072157*^9}},
 CellID->1177372917,ExpressionUUID->"d2ca40a6-8be7-49b9-90cc-ebe5e6d31bfa"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"xTensorQ", "[", "Basis", "]"}]], "Input",
 CellLabel->"In[101]:=",
 CellID->1192292561,ExpressionUUID->"dac3c695-c2c7-4a8a-b803-c698bd4c3fa0"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.7286616955645533`*^9, 3.728671241900662*^9, 
  3.728674888651722*^9, 3.728743045494307*^9, 3.728743302549265*^9, 
  3.7287433990652437`*^9, 3.728747512686635*^9, 3.728749533279435*^9, 
  3.7287499399755363`*^9, 3.728750414587826*^9, 3.7287506387111053`*^9, 
  3.728750773439419*^9, 3.728751061158033*^9, 3.7287516272106857`*^9, 
  3.728813272915758*^9, 3.7288273899838552`*^9, 3.7288359815849733`*^9, 
  3.7288362978935137`*^9, 3.7288430983555193`*^9, 3.7288456322616777`*^9, 
  3.733367382035584*^9, 3.792273304457622*^9},
 CellLabel->"Out[101]=",
 CellID->580559433,ExpressionUUID->"918b136f-d475-4861-99fb-967c71819511"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"MasterOf", "[", "Basis", "]"}]], "Input",
 CellLabel->"In[102]:=",
 CellID->691162570,ExpressionUUID->"c69d5739-836b-4e1c-a08d-588bbde19b69"],

Cell[BoxData["Symbol"], "Output",
 CellChangeTimes->{3.728661696565592*^9, 3.72867124223378*^9, 
  3.72867488878331*^9, 3.728743046278872*^9, 3.728743302679347*^9, 
  3.72874339919886*^9, 3.728747513414658*^9, 3.728749533875742*^9, 
  3.728749940086069*^9, 3.728750414740555*^9, 3.728750639338141*^9, 
  3.7287507735903378`*^9, 3.728751061324078*^9, 3.728751627325673*^9, 
  3.7288132730366087`*^9, 3.728827390149475*^9, 3.7288359817429037`*^9, 
  3.7288362980439672`*^9, 3.728843098630637*^9, 3.72884563244127*^9, 
  3.7333673826018353`*^9, 3.79227330507928*^9},
 CellLabel->"Out[102]=",
 CellID->586231575,ExpressionUUID->"59bcfa76-d232-4b84-a09d-0dda13e0d380"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Basis", "MR"],
 " and ",
 StyleBox["Dir", "MR"],
 " act as inverses of each other"
}], "MathCaption",
 CellID->630267367,ExpressionUUID->"f2ae76fc-7560-46ee-a7b6-829e61f9f246"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Dir", "[", 
  RowBox[{"Basis", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"{", 
     RowBox[{"b", ",", "fred"}], "}"}]}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.7287430500736017`*^9, 3.728743053387002*^9}, {
  3.7287475148399267`*^9, 3.72874751539183*^9}},
 CellLabel->"In[103]:=",
 CellID->1514290165,ExpressionUUID->"71005db8-5dd4-4d14-b9af-e20758159a12"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"b", ",", "fred"}], "}"}]], "Output",
 CellChangeTimes->{3.7286712427877703`*^9, 3.7286748889017677`*^9, 
  3.72874305416835*^9, 3.7287433028006697`*^9, 3.728743399331657*^9, 
  3.728747520300786*^9, 3.728749534893056*^9, 3.7287499402200117`*^9, 
  3.728750414874661*^9, 3.728750640581684*^9, 3.728750773737246*^9, 
  3.7287510614911346`*^9, 3.728751627465412*^9, 3.728813273171417*^9, 
  3.728827390294929*^9, 3.72883598189392*^9, 3.7288362981798697`*^9, 
  3.728843099069454*^9, 3.7288456326284037`*^9, 3.733367383635747*^9, 
  3.792273306151977*^9},
 CellLabel->"Out[103]=",
 CellID->1901460078,ExpressionUUID->"6483f584-b440-4cb1-a3c6-23150536e660"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Basis", "[", 
  RowBox[{
   RowBox[{"Dir", "[", 
    RowBox[{"v", "[", "a", "]"}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", "fred"}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728743055921344*^9, 3.728743057122847*^9}, {
  3.7287475170015182`*^9, 3.728747517463586*^9}},
 CellLabel->"In[104]:=",
 CellID->1956465519,ExpressionUUID->"a2eedcf1-0da5-400a-93c9-8ea6c4278b6d"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"v", 
      StyleBox[GridBox[{
         {
          StyleBox["1",
           FontColor->RGBColor[1., 0., 0.]]},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`v[{1, $CellContext`fred}],
  Editable->False]], "Output",
 CellChangeTimes->{3.72867124312745*^9, 3.728674889209538*^9, 
  3.728743058262854*^9, 3.728743302918087*^9, 3.728743399513651*^9, 
  3.7287475210435963`*^9, 3.72874953546037*^9, 3.728749940337488*^9, 
  3.728750415072947*^9, 3.7287506412380047`*^9, 3.728750774227255*^9, 
  3.728751061657795*^9, 3.72875162761244*^9, 3.72881327344273*^9, 
  3.72882739046871*^9, 3.728835982045106*^9, 3.7288362983163*^9, 
  3.7288430995530987`*^9, 3.728845632752743*^9, 3.733367384187002*^9, 
  3.792273306695842*^9},
 CellLabel->"Out[104]=",
 CellID->1729986636,ExpressionUUID->"9605cf6b-c627-400f-9946-e6ac066aa69d"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["1.1.3. Bases on inner bundles and complex bases", "Subsubsection",
 CellChangeTimes->{{3.728673439853546*^9, 
  3.728673440141943*^9}},ExpressionUUID->"ad588e59-8951-4a3e-ae93-\
8d65e3ea4266"],

Cell[TextData[{
 "Starting with version 0.9, ",
 StyleBox["xTensor`", "MR"],
 " is capable of dealing with complex tensors and vector bundles. The \
function ",
 StyleBox["DefBasis", "MR"],
 " has been extended accordingly to allow the user to define complex bases, \
through the option ",
 StyleBox["Dagger, ", "MR"],
 "taking the values ",
 StyleBox["Real", "MR"],
 " or ",
 StyleBox["Complex", "MR"],
 "."
}], "Text",
 CellChangeTimes->{{3.728662468554126*^9, 3.728662483048169*^9}, 
   3.728662515014694*^9},ExpressionUUID->"0358f8a9-76cc-4ae8-b6ae-\
4d70b6cd4a68"],

Cell[CellGroupData[{

Cell["A basis on a real vbundle is real by default:", "MathCaption",
 CellChangeTimes->{{3.728662501151197*^9, 3.7286625106152067`*^9}},
 CellID->2018963650,ExpressionUUID->"363948c7-d854-4309-8b3c-4c45d486a48c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Dagger", "[", "fred", "]"}]], "Input",
 CellChangeTimes->{{3.728743089450519*^9, 3.7287430898012342`*^9}, {
  3.728747537545662*^9, 3.728747538030662*^9}},
 CellLabel->
  "In[105]:=",ExpressionUUID->"116011f7-42f4-407e-b277-46c132541d53"],

Cell[BoxData["fred"], "Output",
 CellChangeTimes->{3.72867124692056*^9, 3.728674889482587*^9, 
  3.728743090679674*^9, 3.728743303036099*^9, 3.728743399699605*^9, 
  3.728747538465747*^9, 3.728749538569736*^9, 3.728749940452485*^9, 
  3.728750655066505*^9, 3.7287507776077757`*^9, 3.728751061818139*^9, 
  3.728751627727195*^9, 3.728813273728304*^9, 3.728827390610705*^9, 
  3.728835982189221*^9, 3.728836298446171*^9, 3.728843105674286*^9, 
  3.728845632931336*^9, 3.733367392634997*^9, 3.79227331397898*^9},
 CellLabel->"Out[105]=",
 CellID->1047717905,ExpressionUUID->"ca4e5d1f-ed96-45b4-b670-9efd0a82ca40"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
But let us see what happens when we define a basis on a complex bundle:\
\>", "MathCaption",
 CellChangeTimes->{{3.728662527157831*^9, 
  3.728662538510088*^9}},ExpressionUUID->"9fcdcbff-e456-4f29-a900-\
bb73000f05da"],

Cell[CellGroupData[{

Cell[BoxData["$Bases"], "Input",
 CellLabel->
  "In[106]:=",ExpressionUUID->"6e4346ad-9575-41b4-8d31-d5fcedfe1699"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"fred", ",", "blues"}], "}"}]], "Output",
 CellChangeTimes->{3.728662535464829*^9, 3.728671247081367*^9, 
  3.728674889632959*^9, 3.7287430921335077`*^9, 3.728743303150292*^9, 
  3.728743399832601*^9, 3.72874754083609*^9, 3.728749540031995*^9, 
  3.728749940552361*^9, 3.728750656924406*^9, 3.728750777774437*^9, 
  3.728751061951644*^9, 3.728751627865032*^9, 3.728813273863812*^9, 
  3.728827390854671*^9, 3.7288359823345423`*^9, 3.728836298577326*^9, 
  3.728843105853525*^9, 3.7288456330477047`*^9, 3.73336739279274*^9, 
  3.7922733153777637`*^9},
 CellLabel->"Out[106]=",
 CellID->423408786,ExpressionUUID->"67d2a4c7-7224-45f3-a8c4-e2be128e51a6"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefBasis", "[", 
  RowBox[{"fcomp", ",", "InnerC", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"+", "1"}]}], "}"}], ",", 
   RowBox[{"Dagger", "\[Rule]", "Complex"}], ",", 
   RowBox[{"BasisColor", "\[Rule]", " ", "Green"}]}], "]"}]], "Input",
 CellChangeTimes->{
  3.728743101663217*^9, {3.7287431841429234`*^9, 3.728743184995275*^9}, {
   3.728747543366444*^9, 3.728747544510332*^9}},
 CellLabel->
  "In[107]:=",ExpressionUUID->"1ab94e36-e437-4575-ba98-2bcc75ddb04f"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefBasis", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"basis \"\>", "\[InvisibleSpace]", "fcomp", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xCoba`DefBasis, ": Defining ", "basis ", $CellContext`fcomp, 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728662543911641*^9, 3.728671247284732*^9, 
  3.728674889773672*^9, 3.728743103159237*^9, 3.728743303269644*^9, 
  3.728743399965954*^9, 3.7287475449351387`*^9, 3.728749540708436*^9, 
  3.7287499406822033`*^9, 3.728750777972083*^9, 3.7287510620820227`*^9, 
  3.7287516279936934`*^9, 3.728813274031293*^9, 3.728827391107604*^9, 
  3.728835982479816*^9, 3.728836298736219*^9, 3.728843106132284*^9, 
  3.7288456331746893`*^9, 3.7333673929673634`*^9, 3.792273316016864*^9},
 CellLabel->"During evaluation of In[107]:=",
 CellID->457973117,ExpressionUUID->"1fe6afb1-6fea-4550-b90b-9be5cd4a70b8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"parallel derivative \"\>", "\[InvisibleSpace]", 
   RowBox[{"PDfcomp", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "parallel derivative ", 
   $CellContext`PDfcomp[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728662543911641*^9, 3.728671247284732*^9, 
  3.728674889773672*^9, 3.728743103159237*^9, 3.728743303269644*^9, 
  3.728743399965954*^9, 3.7287475449351387`*^9, 3.728749540708436*^9, 
  3.7287499406822033`*^9, 3.728750777972083*^9, 3.7287510620820227`*^9, 
  3.7287516279936934`*^9, 3.728813274031293*^9, 3.728827391107604*^9, 
  3.728835982479816*^9, 3.728836298736219*^9, 3.728843106132284*^9, 
  3.7288456331746893`*^9, 3.7333673929673634`*^9, 3.792273316025237*^9},
 CellLabel->"During evaluation of In[107]:=",
 CellID->572786571,ExpressionUUID->"f04e4bf4-9f57-4464-9ced-98b50742825e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"TorsionPDfcomp", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "torsion tensor ", 
   $CellContext`TorsionPDfcomp[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728662543911641*^9, 3.728671247284732*^9, 
  3.728674889773672*^9, 3.728743103159237*^9, 3.728743303269644*^9, 
  3.728743399965954*^9, 3.7287475449351387`*^9, 3.728749540708436*^9, 
  3.7287499406822033`*^9, 3.728750777972083*^9, 3.7287510620820227`*^9, 
  3.7287516279936934`*^9, 3.728813274031293*^9, 3.728827391107604*^9, 
  3.728835982479816*^9, 3.728836298736219*^9, 3.728843106132284*^9, 
  3.7288456331746893`*^9, 3.7333673929673634`*^9, 3.7922733160361347`*^9},
 CellLabel->"During evaluation of In[107]:=",
 CellID->1802192943,ExpressionUUID->"907c69a4-1708-4938-9db6-bfada43db1fb"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelPDfcomp", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "non-symmetric Christoffel tensor ", 
   $CellContext`ChristoffelPDfcomp[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728662543911641*^9, 3.728671247284732*^9, 
  3.728674889773672*^9, 3.728743103159237*^9, 3.728743303269644*^9, 
  3.728743399965954*^9, 3.7287475449351387`*^9, 3.728749540708436*^9, 
  3.7287499406822033`*^9, 3.728750777972083*^9, 3.7287510620820227`*^9, 
  3.7287516279936934`*^9, 3.728813274031293*^9, 3.728827391107604*^9, 
  3.728835982479816*^9, 3.728836298736219*^9, 3.728843106132284*^9, 
  3.7288456331746893`*^9, 3.7333673929673634`*^9, 3.7922733160438967`*^9},
 CellLabel->"During evaluation of In[107]:=",
 CellID->1347435713,ExpressionUUID->"a2f9a11a-b4d9-4715-bc6e-a335fbba43e1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Riemann tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RiemannPDfcomp", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Riemann tensor ", 
   $CellContext`RiemannPDfcomp[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, $CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728662543911641*^9, 3.728671247284732*^9, 
  3.728674889773672*^9, 3.728743103159237*^9, 3.728743303269644*^9, 
  3.728743399965954*^9, 3.7287475449351387`*^9, 3.728749540708436*^9, 
  3.7287499406822033`*^9, 3.728750777972083*^9, 3.7287510620820227`*^9, 
  3.7287516279936934`*^9, 3.728813274031293*^9, 3.728827391107604*^9, 
  3.728835982479816*^9, 3.728836298736219*^9, 3.728843106132284*^9, 
  3.7288456331746893`*^9, 3.7333673929673634`*^9, 3.792273316051384*^9},
 CellLabel->"During evaluation of In[107]:=",
 CellID->1136970776,ExpressionUUID->"f37883bf-afd3-4da4-9162-78a3bff78891"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciPDfcomp", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Ricci tensor ", 
   $CellContext`RicciPDfcomp[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728662543911641*^9, 3.728671247284732*^9, 
  3.728674889773672*^9, 3.728743103159237*^9, 3.728743303269644*^9, 
  3.728743399965954*^9, 3.7287475449351387`*^9, 3.728749540708436*^9, 
  3.7287499406822033`*^9, 3.728750777972083*^9, 3.7287510620820227`*^9, 
  3.7287516279936934`*^9, 3.728813274031293*^9, 3.728827391107604*^9, 
  3.728835982479816*^9, 3.728836298736219*^9, 3.728843106132284*^9, 
  3.7288456331746893`*^9, 3.7333673929673634`*^9, 3.792273316058969*^9},
 CellLabel->"During evaluation of In[107]:=",
 CellID->758341226,ExpressionUUID->"2dd05365-f479-407b-9293-aa89f3c4b10b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"nonsymmetric AChristoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"AChristoffelPDfcomp", "[", 
    RowBox[{"A", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "C"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "nonsymmetric AChristoffel tensor ", 
   $CellContext`AChristoffelPDfcomp[$CellContext`A, -$CellContext`b, -C], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728662543911641*^9, 3.728671247284732*^9, 
  3.728674889773672*^9, 3.728743103159237*^9, 3.728743303269644*^9, 
  3.728743399965954*^9, 3.7287475449351387`*^9, 3.728749540708436*^9, 
  3.7287499406822033`*^9, 3.728750777972083*^9, 3.7287510620820227`*^9, 
  3.7287516279936934`*^9, 3.728813274031293*^9, 3.728827391107604*^9, 
  3.728835982479816*^9, 3.728836298736219*^9, 3.728843106132284*^9, 
  3.7288456331746893`*^9, 3.7333673929673634`*^9, 3.792273316066289*^9},
 CellLabel->"During evaluation of In[107]:=",
 CellID->125540437,ExpressionUUID->"89fc858c-e1d2-4591-94c7-2f7392997ed6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"nonsymmetric AChristoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"AChristoffelPDfcomp\[Dagger]", "[", 
    RowBox[{"A\[Dagger]", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "C\[Dagger]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "nonsymmetric AChristoffel tensor ", 
   $CellContext`AChristoffelPDfcomp\[Dagger][$CellContext`A\[Dagger], \
-$CellContext`b, -$CellContext`C\[Dagger]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728662543911641*^9, 3.728671247284732*^9, 
  3.728674889773672*^9, 3.728743103159237*^9, 3.728743303269644*^9, 
  3.728743399965954*^9, 3.7287475449351387`*^9, 3.728749540708436*^9, 
  3.7287499406822033`*^9, 3.728750777972083*^9, 3.7287510620820227`*^9, 
  3.7287516279936934`*^9, 3.728813274031293*^9, 3.728827391107604*^9, 
  3.728835982479816*^9, 3.728836298736219*^9, 3.728843106132284*^9, 
  3.7288456331746893`*^9, 3.7333673929673634`*^9, 3.792273316073503*^9},
 CellLabel->"During evaluation of In[107]:=",
 CellID->1129183305,ExpressionUUID->"29ef0b62-c1d3-4d21-a3ca-3b01a13cad7c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing FRiemann tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"FRiemannPDfcomp", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "C"}], ",", "D"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "vanishing FRiemann tensor ", 
   $CellContext`FRiemannPDfcomp[-$CellContext`a, -$CellContext`b, -C, D], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728662543911641*^9, 3.728671247284732*^9, 
  3.728674889773672*^9, 3.728743103159237*^9, 3.728743303269644*^9, 
  3.728743399965954*^9, 3.7287475449351387`*^9, 3.728749540708436*^9, 
  3.7287499406822033`*^9, 3.728750777972083*^9, 3.7287510620820227`*^9, 
  3.7287516279936934`*^9, 3.728813274031293*^9, 3.728827391107604*^9, 
  3.728835982479816*^9, 3.728836298736219*^9, 3.728843106132284*^9, 
  3.7288456331746893`*^9, 3.7333673929673634`*^9, 3.792273316080656*^9},
 CellLabel->"During evaluation of In[107]:=",
 CellID->870818596,ExpressionUUID->"f4e15e2e-bfd3-4a97-89c1-303eb6ba9bb7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing FRiemann tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"FRiemannPDfcomp\[Dagger]", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "C\[Dagger]"}], ",", "D\[Dagger]"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "vanishing FRiemann tensor ", 
   $CellContext`FRiemannPDfcomp\[Dagger][-$CellContext`a, -$CellContext`b, \
-$CellContext`C\[Dagger], $CellContext`D\[Dagger]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728662543911641*^9, 3.728671247284732*^9, 
  3.728674889773672*^9, 3.728743103159237*^9, 3.728743303269644*^9, 
  3.728743399965954*^9, 3.7287475449351387`*^9, 3.728749540708436*^9, 
  3.7287499406822033`*^9, 3.728750777972083*^9, 3.7287510620820227`*^9, 
  3.7287516279936934`*^9, 3.728813274031293*^9, 3.728827391107604*^9, 
  3.728835982479816*^9, 3.728836298736219*^9, 3.728843106132284*^9, 
  3.7288456331746893`*^9, 3.7333673929673634`*^9, 3.792273316088019*^9},
 CellLabel->"During evaluation of In[107]:=",
 CellID->747088667,ExpressionUUID->"9f04e2e3-b9b3-4946-a0f0-b8ac8bfed560"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefBasis", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"basis \"\>", "\[InvisibleSpace]", 
   "fcomp\[Dagger]", "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xCoba`DefBasis, ": Defining ", 
   "basis ", $CellContext`fcomp\[Dagger], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728662543911641*^9, 3.728671247284732*^9, 
  3.728674889773672*^9, 3.728743103159237*^9, 3.728743303269644*^9, 
  3.728743399965954*^9, 3.7287475449351387`*^9, 3.728749540708436*^9, 
  3.7287499406822033`*^9, 3.728750777972083*^9, 3.7287510620820227`*^9, 
  3.7287516279936934`*^9, 3.728813274031293*^9, 3.728827391107604*^9, 
  3.728835982479816*^9, 3.728836298736219*^9, 3.728843106132284*^9, 
  3.7288456331746893`*^9, 3.7333673929673634`*^9, 3.792273316095182*^9},
 CellLabel->"During evaluation of In[107]:=",
 CellID->2129825907,ExpressionUUID->"f251092d-e7c1-460a-a962-b24da7424004"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaUpfcomp", "[", 
    RowBox[{"A", ",", "B"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric +1 density ", 
   $CellContext`etaUpfcomp[$CellContext`A, $CellContext`B], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728662543911641*^9, 3.728671247284732*^9, 
  3.728674889773672*^9, 3.728743103159237*^9, 3.728743303269644*^9, 
  3.728743399965954*^9, 3.7287475449351387`*^9, 3.728749540708436*^9, 
  3.7287499406822033`*^9, 3.728750777972083*^9, 3.7287510620820227`*^9, 
  3.7287516279936934`*^9, 3.728813274031293*^9, 3.728827391107604*^9, 
  3.728835982479816*^9, 3.728836298736219*^9, 3.728843106132284*^9, 
  3.7288456331746893`*^9, 3.7333673929673634`*^9, 3.7922733161024857`*^9},
 CellLabel->"During evaluation of In[107]:=",
 CellID->1726590753,ExpressionUUID->"a30acda1-8557-4aa1-9ef3-c2dafa6648b1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaUpfcomp\[Dagger]", "[", 
    RowBox[{"A\[Dagger]", ",", "B\[Dagger]"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric +1 density ", 
   $CellContext`etaUpfcomp\[Dagger][$CellContext`A\[Dagger], $CellContext`B\
\[Dagger]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728662543911641*^9, 3.728671247284732*^9, 
  3.728674889773672*^9, 3.728743103159237*^9, 3.728743303269644*^9, 
  3.728743399965954*^9, 3.7287475449351387`*^9, 3.728749540708436*^9, 
  3.7287499406822033`*^9, 3.728750777972083*^9, 3.7287510620820227`*^9, 
  3.7287516279936934`*^9, 3.728813274031293*^9, 3.728827391107604*^9, 
  3.728835982479816*^9, 3.728836298736219*^9, 3.728843106132284*^9, 
  3.7288456331746893`*^9, 3.7333673929673634`*^9, 3.792273316109995*^9},
 CellLabel->"During evaluation of In[107]:=",
 CellID->2011747407,ExpressionUUID->"06551c1b-5de5-495f-bbde-b719b93b5835"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaDownfcomp", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "B"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric -1 density ", 
   $CellContext`etaDownfcomp[-$CellContext`A, -$CellContext`B], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728662543911641*^9, 3.728671247284732*^9, 
  3.728674889773672*^9, 3.728743103159237*^9, 3.728743303269644*^9, 
  3.728743399965954*^9, 3.7287475449351387`*^9, 3.728749540708436*^9, 
  3.7287499406822033`*^9, 3.728750777972083*^9, 3.7287510620820227`*^9, 
  3.7287516279936934`*^9, 3.728813274031293*^9, 3.728827391107604*^9, 
  3.728835982479816*^9, 3.728836298736219*^9, 3.728843106132284*^9, 
  3.7288456331746893`*^9, 3.7333673929673634`*^9, 3.792273316117756*^9},
 CellLabel->"During evaluation of In[107]:=",
 CellID->1456131081,ExpressionUUID->"a4c6dfa7-f6b5-42f5-89d3-0ec85a88d69b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaDownfcomp\[Dagger]", "[", 
    RowBox[{
     RowBox[{"-", "A\[Dagger]"}], ",", 
     RowBox[{"-", "B\[Dagger]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric -1 density ", 
   $CellContext`etaDownfcomp\[Dagger][-$CellContext`A\[Dagger], \
-$CellContext`B\[Dagger]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728662543911641*^9, 3.728671247284732*^9, 
  3.728674889773672*^9, 3.728743103159237*^9, 3.728743303269644*^9, 
  3.728743399965954*^9, 3.7287475449351387`*^9, 3.728749540708436*^9, 
  3.7287499406822033`*^9, 3.728750777972083*^9, 3.7287510620820227`*^9, 
  3.7287516279936934`*^9, 3.728813274031293*^9, 3.728827391107604*^9, 
  3.728835982479816*^9, 3.728836298736219*^9, 3.728843106132284*^9, 
  3.7288456331746893`*^9, 3.7333673929673634`*^9, 3.7922733161252613`*^9},
 CellLabel->"During evaluation of In[107]:=",
 CellID->471437155,ExpressionUUID->"232a1dd5-83a2-4ef5-9b78-2969fe6ecd84"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData["$Bases"], "Input",
 CellLabel->
  "In[108]:=",ExpressionUUID->"f096b0d0-eb4b-4c4a-90ca-80a3e343c1ed"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"fred", ",", "blues", ",", "fcomp", ",", "fcomp\[Dagger]"}], 
  "}"}]], "Output",
 CellChangeTimes->{3.728662545449069*^9, 3.7286712476994133`*^9, 
  3.728674890334281*^9, 3.72874310587364*^9, 3.72874330368539*^9, 
  3.7287434003972397`*^9, 3.7287475470835943`*^9, 3.728749542759932*^9, 
  3.728749940948069*^9, 3.7287507784555483`*^9, 3.728751062512973*^9, 
  3.728751628343998*^9, 3.728813274366044*^9, 3.728827391512039*^9, 
  3.7288359829991837`*^9, 3.728836299135511*^9, 3.728843106974292*^9, 
  3.7288456339592543`*^9, 3.733367393354104*^9, 3.792273317094853*^9},
 CellLabel->"Out[108]=",
 CellID->1466307460,ExpressionUUID->"7fb28bc8-daaa-4dee-9c4e-755017aafaf8"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
A conjugate basis has also appeared on the conjugate manifold:\
\>", "MathCaption",ExpressionUUID->"48eceb47-8d82-4491-9731-a9bb32511f7d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Dagger", "/@", 
  RowBox[{"{", 
   RowBox[{"fcomp", ",", "fcomp\[Dagger]"}], "}"}]}]], "Input",
 CellChangeTimes->{{3.728743118302342*^9, 3.7287431238460617`*^9}, {
  3.728747556285782*^9, 3.728747559565571*^9}},
 CellLabel->
  "In[109]:=",ExpressionUUID->"8e18fc1a-a81f-41ee-b09f-83a64b62a11e"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"fcomp\[Dagger]", ",", "fcomp"}], "}"}]], "Output",
 CellChangeTimes->{3.728662554682727*^9, 3.728671247847705*^9, 
  3.7286748904856*^9, 3.728743124777381*^9, 3.728743303870226*^9, 
  3.728743400529509*^9, 3.728747559969225*^9, 3.7287495440568943`*^9, 
  3.728749941091353*^9, 3.7287507787583*^9, 3.728751062666411*^9, 
  3.728751628481544*^9, 3.728813274516375*^9, 3.7288273916617737`*^9, 
  3.7288359831446457`*^9, 3.728836299305315*^9, 3.728843107672044*^9, 
  3.728845634178636*^9, 3.733367393604959*^9, 3.792273318518443*^9},
 CellLabel->"Out[109]=",
 CellID->922155191,ExpressionUUID->"0230dce8-6f63-485a-b7b2-737fcb3524aa"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"VBundleOfBasis", "/@", 
  RowBox[{"{", 
   RowBox[{"fcomp", ",", "fcomp\[Dagger]"}], "}"}]}]], "Input",
 CellChangeTimes->{{3.728743129093815*^9, 3.728743131365672*^9}, {
  3.7287475628694*^9, 3.728747567301271*^9}},
 CellLabel->
  "In[110]:=",ExpressionUUID->"0b33ed36-0ed6-4b22-bfd8-24aa09f7868d"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox["InnerC",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`InnerC,
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["InnerC\[Dagger]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`InnerC\[Dagger],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.72866255575674*^9, 3.728671247996912*^9, 
  3.728674890636238*^9, 3.728743131938429*^9, 3.728743303986672*^9, 
  3.7287434006464243`*^9, 3.7287475676011744`*^9, 3.728749544644074*^9, 
  3.728749941277614*^9, 3.72875077894623*^9, 3.728751062795731*^9, 
  3.728751628614292*^9, 3.7288132746642027`*^9, 3.728827391834931*^9, 
  3.7288359832967577`*^9, 3.728836299441401*^9, 3.728843107915455*^9, 
  3.72884563434454*^9, 3.7333673938364067`*^9, 3.792273319046935*^9},
 CellLabel->"Out[110]=",
 CellID->1233431399,ExpressionUUID->"bddbd679-9d17-4954-a94c-53dcc07bedcb"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Notice how negative cnumbers do not denote covariant indices:\
\>", "MathCaption",ExpressionUUID->"fad86ace-6ac0-438b-90ff-51ac842864bd"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "1"}], ",", "fcomp"}], "}"}]}], "]"}], ",", 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "1"}], ",", "fcomp\[Dagger]"}], "}"}]}], "]"}]}], 
  "}"}]], "Input",
 CellChangeTimes->{{3.728743139190913*^9, 3.7287431419094763`*^9}, {
  3.728747570485351*^9, 3.728747574565057*^9}},
 CellLabel->
  "In[111]:=",ExpressionUUID->"5f4077eb-97ac-4911-aeb8-548b4e7365c5"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox[
             RowBox[{"-", "1"}],
             FontColor->RGBColor[0., 1., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {-1, $CellContext`fcomp}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox[
             RowBox[{"-", "1"}],
             FontColor->RGBColor[0., 1., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {-1, $CellContext`fcomp\[Dagger]}],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728662561074698*^9, 3.72867124819037*^9, 
  3.72867489081181*^9, 3.7287431425161877`*^9, 3.7287433041026773`*^9, 
  3.7287434007634573`*^9, 3.7287475748662577`*^9, 3.728749547061842*^9, 
  3.728749941431819*^9, 3.728750779102298*^9, 3.728751062951744*^9, 
  3.728751628788262*^9, 3.728813274800503*^9, 3.728827392297591*^9, 
  3.7288359834615097`*^9, 3.7288362996033573`*^9, 3.728843108115877*^9, 
  3.72884563451518*^9, 3.7333673940040913`*^9, 3.7922733212005463`*^9},
 CellLabel->"Out[111]=",
 CellID->400987425,ExpressionUUID->"abf35b60-cdb0-421f-bde7-d59c5d7dec6e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The basis objects of both bases look the same, because ", "MathCaption",ExpressionUUID->"aec10345-992f-4fb8-8bd2-a8516d0afd05"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Dagger", "[", "Basis", "]"}]], "Input",
 CellLabel->
  "In[112]:=",ExpressionUUID->"e8d7e545-e76e-49fd-aeff-5c74e2b64b91"],

Cell[BoxData["Basis"], "Output",
 CellChangeTimes->{3.728662570374631*^9, 3.72867124834468*^9, 
  3.728674890951342*^9, 3.7287431504622087`*^9, 3.728743304218072*^9, 
  3.728743400896872*^9, 3.728747576999234*^9, 3.728749548144659*^9, 
  3.728749941565023*^9, 3.7287507792485857`*^9, 3.72875106307299*^9, 
  3.728751628986322*^9, 3.728813275033065*^9, 3.7288273925626917`*^9, 
  3.72883598361465*^9, 3.728836299775648*^9, 3.728843108349691*^9, 
  3.728845634680299*^9, 3.733367394154217*^9, 3.792273322904826*^9},
 CellLabel->"Out[112]=",
 CellID->862925962,ExpressionUUID->"4d52fca8-9aca-40a2-bb98-9bbfe18442ad"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Dagger", "[", 
  RowBox[{"Basis", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "1"}], ",", "fcomp"}], "}"}]}], "]"}], "]"}]], "Input",
 CellChangeTimes->{
  3.728743157901415*^9, {3.7287475791568823`*^9, 3.728747580316689*^9}},
 CellLabel->
  "In[113]:=",ExpressionUUID->"cc27bc8f-e58e-4c31-8834-5742d763e399"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"e", 
      StyleBox[GridBox[{
         {" ", 
          StyleBox[
           RowBox[{"-", "1"}],
           FontColor->RGBColor[0., 1., 0.]]},
         {"a", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`Basis[-$CellContext`a, {-1, $CellContext`fcomp\[Dagger]}],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.728662571365798*^9, 3.728671248478361*^9, 3.7286748911839848`*^9, {
   3.728743153464117*^9, 3.728743158467505*^9}, 3.7287433043362207`*^9, 
   3.728743401015444*^9, 3.728747580817154*^9, 3.728749548761578*^9, 
   3.728749941703971*^9, 3.728750779402073*^9, 3.728751063216125*^9, 
   3.728751629133195*^9, 3.728813275283862*^9, 3.7288273926952143`*^9, 
   3.728835983777828*^9, 3.728836299922529*^9, 3.728843108898394*^9, 
   3.728845634864781*^9, 3.733367394403808*^9, 3.7922733234006033`*^9},
 CellLabel->"Out[113]=",
 CellID->1859674700,ExpressionUUID->"ed9ba09d-085e-42cb-8b14-4b30ac14095c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"InputForm", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[114]:=",ExpressionUUID->"e84e2d6c-619b-4f6e-aa83-69e0a2ed03d7"],

Cell["Basis[-a, {-1, fcomp\[Dagger]}]", "Output",
 CellChangeTimes->{3.728662572752851*^9, 3.728671248818705*^9, 
  3.728674891300931*^9, 3.728743159865275*^9, 3.728743304452919*^9, 
  3.728743401147974*^9, 3.728747583006897*^9, 3.7287495493685217`*^9, 
  3.7287499418338947`*^9, 3.728750779532933*^9, 3.7287510633490334`*^9, 
  3.72875162926143*^9, 3.7288132754474287`*^9, 3.728827392844061*^9, 
  3.728835983922563*^9, 3.72883630007096*^9, 3.7288431091977577`*^9, 
  3.728845635030485*^9, 3.733367394686515*^9, 3.792273324951456*^9},
 CellLabel->"Out[114]//InputForm=",
 CellID->1593238590,ExpressionUUID->"1647086e-e1f9-4937-bf8d-073506ad91da"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "and their cnumbers are also defined to be the same. This can be changed, \
with the new function ",
 StyleBox["DaggerBCIndex. ", "MR"],
 "Remember how in ",
 StyleBox["xTensor`", "MR"],
 " the conjugates of indices were given by "
}], "MathCaption",
 CellChangeTimes->{
  3.7287431704520893`*^9},ExpressionUUID->"1de26540-5893-43fa-820e-\
af6c3d3d8df5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DaggerIndex", "/@", 
  RowBox[{"{", 
   RowBox[{"a", ",", "b", ",", "A", ",", "B"}], "}"}]}]], "Input",
 CellLabel->
  "In[115]:=",ExpressionUUID->"c3bdfedb-e251-433b-8db0-dabdca1732d1"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"a", ",", "b", ",", "A\[Dagger]", ",", "B\[Dagger]"}], 
  "}"}]], "Output",
 CellChangeTimes->{3.728662585419776*^9, 3.728671249404047*^9, 
  3.728674891436489*^9, 3.728743162099724*^9, 3.728743304585217*^9, 
  3.7287434012634363`*^9, 3.728747585533781*^9, 3.7287495506050243`*^9, 
  3.728749941953538*^9, 3.72875077965828*^9, 3.728751063496058*^9, 
  3.7287516294149303`*^9, 3.728813275695641*^9, 3.728827392996203*^9, 
  3.728835984078121*^9, 3.728836300227271*^9, 3.728843109417149*^9, 
  3.728845635230579*^9, 3.7333673948859463`*^9, 3.792273327512177*^9},
 CellLabel->"Out[115]=",
 CellID->1702226930,ExpressionUUID->"a80ca744-ea41-4a25-a079-4c9d3637d6bd"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "But we cannot change the definition associated to ",
 StyleBox["DaggerIndex[{1,fcomp}]", "MR"],
 ",  because ",
 StyleBox["complex", "MR"],
 " is too deep to be given an upvalue. Let us see how we can use  ",
 StyleBox["DaggerCIndex", "MR"],
 " to solve this with an example. We want to complexify a real basis:"
}], "MathCaption",
 CellChangeTimes->{{3.7287475934042387`*^9, 
  3.7287475973957453`*^9}},ExpressionUUID->"22b71b56-9c3d-4ee4-a1c7-\
04b8a788662e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefBasis", "[", 
  RowBox[{"test", ",", "TangentM3", ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", "2", ",", "3"}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[116]:=",ExpressionUUID->"8695ff25-fb91-4c86-b442-70e743784f59"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefBasis", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"basis \"\>", "\[InvisibleSpace]", "test", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xCoba`DefBasis, ": Defining ", "basis ", $CellContext`test, 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728662591992484*^9, 3.728671249713957*^9, 
  3.7286748915900917`*^9, 3.7287432021954117`*^9, 3.7287433047520227`*^9, 
  3.728743401381711*^9, 3.72874760149131*^9, 3.728749551841444*^9, 
  3.728749942084701*^9, 3.728750779964903*^9, 3.72875106364469*^9, 
  3.728751629543634*^9, 3.728813276110716*^9, 3.728827393136383*^9, 
  3.728835984227166*^9, 3.728836300381048*^9, 3.728843109666819*^9, 
  3.728845635421936*^9, 3.733367395051529*^9, 3.79227332979818*^9},
 CellLabel->"During evaluation of In[116]:=",
 CellID->1921187312,ExpressionUUID->"b201662e-eebe-4b95-85a6-ec15854e888b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"parallel derivative \"\>", "\[InvisibleSpace]", 
   RowBox[{"PDtest", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "parallel derivative ", 
   $CellContext`PDtest[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728662591992484*^9, 3.728671249713957*^9, 
  3.7286748915900917`*^9, 3.7287432021954117`*^9, 3.7287433047520227`*^9, 
  3.728743401381711*^9, 3.72874760149131*^9, 3.728749551841444*^9, 
  3.728749942084701*^9, 3.728750779964903*^9, 3.72875106364469*^9, 
  3.728751629543634*^9, 3.728813276110716*^9, 3.728827393136383*^9, 
  3.728835984227166*^9, 3.728836300381048*^9, 3.728843109666819*^9, 
  3.728845635421936*^9, 3.733367395051529*^9, 3.79227332981013*^9},
 CellLabel->"During evaluation of In[116]:=",
 CellID->1978520139,ExpressionUUID->"a0fa3a78-07ed-4240-9c37-9cd7be9ba9d5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"TorsionPDtest", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "torsion tensor ", 
   $CellContext`TorsionPDtest[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728662591992484*^9, 3.728671249713957*^9, 
  3.7286748915900917`*^9, 3.7287432021954117`*^9, 3.7287433047520227`*^9, 
  3.728743401381711*^9, 3.72874760149131*^9, 3.728749551841444*^9, 
  3.728749942084701*^9, 3.728750779964903*^9, 3.72875106364469*^9, 
  3.728751629543634*^9, 3.728813276110716*^9, 3.728827393136383*^9, 
  3.728835984227166*^9, 3.728836300381048*^9, 3.728843109666819*^9, 
  3.728845635421936*^9, 3.733367395051529*^9, 3.792273329824707*^9},
 CellLabel->"During evaluation of In[116]:=",
 CellID->1351186025,ExpressionUUID->"0f147443-fec2-47dc-b334-fca4e7f13c3c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelPDtest", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "non-symmetric Christoffel tensor ", 
   $CellContext`ChristoffelPDtest[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728662591992484*^9, 3.728671249713957*^9, 
  3.7286748915900917`*^9, 3.7287432021954117`*^9, 3.7287433047520227`*^9, 
  3.728743401381711*^9, 3.72874760149131*^9, 3.728749551841444*^9, 
  3.728749942084701*^9, 3.728750779964903*^9, 3.72875106364469*^9, 
  3.728751629543634*^9, 3.728813276110716*^9, 3.728827393136383*^9, 
  3.728835984227166*^9, 3.728836300381048*^9, 3.728843109666819*^9, 
  3.728845635421936*^9, 3.733367395051529*^9, 3.7922733298319798`*^9},
 CellLabel->"During evaluation of In[116]:=",
 CellID->1891438797,ExpressionUUID->"1e0eaa49-6100-47bc-aa3c-a60e317716fe"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Riemann tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RiemannPDtest", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Riemann tensor ", 
   $CellContext`RiemannPDtest[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, $CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728662591992484*^9, 3.728671249713957*^9, 
  3.7286748915900917`*^9, 3.7287432021954117`*^9, 3.7287433047520227`*^9, 
  3.728743401381711*^9, 3.72874760149131*^9, 3.728749551841444*^9, 
  3.728749942084701*^9, 3.728750779964903*^9, 3.72875106364469*^9, 
  3.728751629543634*^9, 3.728813276110716*^9, 3.728827393136383*^9, 
  3.728835984227166*^9, 3.728836300381048*^9, 3.728843109666819*^9, 
  3.728845635421936*^9, 3.733367395051529*^9, 3.79227332983949*^9},
 CellLabel->"During evaluation of In[116]:=",
 CellID->211434638,ExpressionUUID->"ddf3c89c-36f6-4c37-863e-a1d366313c2e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciPDtest", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Ricci tensor ", 
   $CellContext`RicciPDtest[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728662591992484*^9, 3.728671249713957*^9, 
  3.7286748915900917`*^9, 3.7287432021954117`*^9, 3.7287433047520227`*^9, 
  3.728743401381711*^9, 3.72874760149131*^9, 3.728749551841444*^9, 
  3.728749942084701*^9, 3.728750779964903*^9, 3.72875106364469*^9, 
  3.728751629543634*^9, 3.728813276110716*^9, 3.728827393136383*^9, 
  3.728835984227166*^9, 3.728836300381048*^9, 3.728843109666819*^9, 
  3.728845635421936*^9, 3.733367395051529*^9, 3.792273329847333*^9},
 CellLabel->"During evaluation of In[116]:=",
 CellID->1644008728,ExpressionUUID->"73e3eab5-ce14-4da6-926a-c17ef6a89f9d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaUptest", "[", 
    RowBox[{"a", ",", "b", ",", "c"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric +1 density ", 
   $CellContext`etaUptest[$CellContext`a, $CellContext`b, $CellContext`c], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728662591992484*^9, 3.728671249713957*^9, 
  3.7286748915900917`*^9, 3.7287432021954117`*^9, 3.7287433047520227`*^9, 
  3.728743401381711*^9, 3.72874760149131*^9, 3.728749551841444*^9, 
  3.728749942084701*^9, 3.728750779964903*^9, 3.72875106364469*^9, 
  3.728751629543634*^9, 3.728813276110716*^9, 3.728827393136383*^9, 
  3.728835984227166*^9, 3.728836300381048*^9, 3.728843109666819*^9, 
  3.728845635421936*^9, 3.733367395051529*^9, 3.792273329854677*^9},
 CellLabel->"During evaluation of In[116]:=",
 CellID->1330089194,ExpressionUUID->"fe01269b-06ca-4693-bffe-6dda69c1177e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaDowntest", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric -1 density ", 
   $CellContext`etaDowntest[-$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728662591992484*^9, 3.728671249713957*^9, 
  3.7286748915900917`*^9, 3.7287432021954117`*^9, 3.7287433047520227`*^9, 
  3.728743401381711*^9, 3.72874760149131*^9, 3.728749551841444*^9, 
  3.728749942084701*^9, 3.728750779964903*^9, 3.72875106364469*^9, 
  3.728751629543634*^9, 3.728813276110716*^9, 3.728827393136383*^9, 
  3.728835984227166*^9, 3.728836300381048*^9, 3.728843109666819*^9, 
  3.728845635421936*^9, 3.733367395051529*^9, 3.792273329861988*^9},
 CellLabel->"During evaluation of In[116]:=",
 CellID->1651801935,ExpressionUUID->"f2ff2998-b67c-42f3-aec8-034a8fa6a80d"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "DaggerCIndex"}]], "Input",
 CellLabel->
  "In[117]:=",ExpressionUUID->"33bacd2f-f679-4de7-812e-cc96471a0dd4"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   FrameBox[
    DynamicModuleBox[{System`InformationDump`open$$ = False, 
     System`InformationDump`mouseOver$$ = False}, 
     PaneSelectorBox[{True->
      TagBox[GridBox[{
         {
          ItemBox[
           PaneBox[
            StyleBox["\<\" Symbol\"\>", "InformationTitleText",
             StripOnInput->False,
             BaseStyle -> None],
            FrameMargins->{{4, 0}, {-1, 1}}],
           BaseStyle->"InformationTitleBackground",
           StripOnInput->False], 
          ItemBox["\<\"\"\>",
           BaseStyle->"InformationTitleBackground",
           StripOnInput->False]},
         {
          ItemBox[
           PaneBox[
            
            StyleBox["\<\"DaggerCIndex[basis, cindex] returns the conjugated \
index to the cindex (assumed to belong to basis or -basis), in general a Dir \
expression. This function should be handled through upvalues for basis.\"\>", 
             "InformationUsageText",
             StripOnInput->False,
             LineSpacing->{1.5, 1.5, 3.}],
            FrameMargins->{{10, 10}, {8, 10}}],
           BaseStyle->"InformationUsageSubtitleBackground",
           StripOnInput->False], 
          ItemBox["\<\"\"\>",
           BaseStyle->"InformationUsageSubtitleBackground",
           StripOnInput->False]},
         {
          PaneBox[GridBox[{
             {
              
              DynamicModuleBox[{System`InformationDump`open$$ = {
               False, False, False, False, False, False, False, False, False, 
                False, False, False}}, 
               StyleBox[GridBox[{
                  {
                   TagBox[
                    TooltipBox[
                    StyleBox["\<\" Definitions\"\>", "InformationRowLabel",
                    StripOnInput->False],
                    "\"Definitions\"",
                    TooltipStyle->"TextStyling"],
                    Annotation[#, "Definitions", "Tooltip"]& ], GridBox[{
                    {
                    RowBox[{
                    RowBox[{"DaggerCIndex", "[", "]"}], ":=", 
                    RowBox[{"Message", "[", 
                    RowBox[{
                    StyleBox[
                    RowBox[{"DaggerCIndex", "::", "argrx"}], "MessageName"], 
                    ",", "DaggerCIndex", ",", "0", ",", "2"}], "]"}]}]},
                    {
                    RowBox[{
                    RowBox[{"DaggerCIndex", "[", 
                    RowBox[{
                    RowBox[{"-", "xAct`xCoba`Private`basis_"}], ",", 
                    "xAct`xCoba`Private`cindex_"}], "]"}], ":=", 
                    RowBox[{"DaggerCIndex", "[", 
                    RowBox[{
                    "xAct`xCoba`Private`basis", ",", 
                    "xAct`xCoba`Private`cindex"}], "]"}]}]},
                    {
                    RowBox[{
                    RowBox[{"DaggerCIndex", "[", 
                    RowBox[{"_", ",", 
                    RowBox[{"{", 
                    RowBox[{
                    "xAct`xCoba`Private`a_Integer", ",", 
                    "xAct`xCoba`Private`basis_"}], "}"}]}], "]"}], ":=", 
                    RowBox[{"{", 
                    RowBox[{"xAct`xCoba`Private`a", ",", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Dagger", "[", "xAct`xCoba`Private`basis", "]"}],
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    xAct`xTensor`Dagger[xAct`xCoba`Private`basis],
                    Editable->False]}], "}"}]}]},
                    {
                    RowBox[{
                    RowBox[{"DaggerCIndex", "[", 
                    RowBox[{"_", ",", 
                    RowBox[{
                    RowBox[{"xAct`xCoba`Private`basis_", "?", "BasisQ"}], "[",
                     "xAct`xCoba`Private`a_Integer", "]"}]}], "]"}], ":=", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Dagger", "[", "xAct`xCoba`Private`basis", "]"}],
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    xAct`xTensor`Dagger[xAct`xCoba`Private`basis],
                    Editable->False], "[", "xAct`xCoba`Private`a", "]"}]}]},
                    {
                    RowBox[{
                    RowBox[{"DaggerCIndex", "[", "_", "]"}], ":=", 
                    RowBox[{"Message", "[", 
                    RowBox[{
                    StyleBox[
                    RowBox[{"DaggerCIndex", "::", "argr"}], "MessageName"], 
                    ",", "DaggerCIndex", ",", "2"}], "]"}]}]},
                    {
                    RowBox[{
                    RowBox[{"DaggerCIndex", "[", 
                    RowBox[{"_", ",", "_", ",", "xAct`xCore`Private`x__"}], 
                    "]"}], ":=", 
                    RowBox[{"Message", "[", 
                    RowBox[{
                    StyleBox[
                    RowBox[{"DaggerCIndex", "::", "argrx"}], "MessageName"], 
                    ",", "DaggerCIndex", ",", 
                    RowBox[{"2", "+", 
                    RowBox[{"Length", "[", 
                    RowBox[{"{", "xAct`xCore`Private`x", "}"}], "]"}]}], ",", 
                    "2"}], "]"}]}]}
                    },
                    DefaultBaseStyle->"Column",
                    GridBoxAlignment->{"Columns" -> {{Left}}},
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}]},
                  {
                   TagBox[
                    TooltipBox[
                    StyleBox["\<\" Attributes\"\>", "InformationRowLabel",
                    StripOnInput->False],
                    "\"Attributes\"",
                    TooltipStyle->"TextStyling"],
                    Annotation[#, "Attributes", "Tooltip"]& ], 
                   RowBox[{"{", "Protected", "}"}]},
                  {
                   TagBox[
                    TooltipBox[
                    StyleBox["\<\" Full Name\"\>", "InformationRowLabel",
                    StripOnInput->False],
                    "\"FullName\"",
                    TooltipStyle->"TextStyling"],
                    
                    Annotation[#, "FullName", 
                    "Tooltip"]& ], "\<\"xAct`xCoba`DaggerCIndex\"\>"}
                 },
                 AutoDelete->False,
                 GridBoxAlignment->{"Columns" -> {Right, Left}},
                 GridBoxDividers->None,
                 GridBoxItemSize->{"Columns" -> {Automatic, Automatic}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.5599999999999999]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.8]}, 
                    Offset[0.2]}}], "DialogStyle",
                StripOnInput->False],
               DynamicModuleValues:>{}]}
            },
            DefaultBaseStyle->"Column",
            GridBoxAlignment->{"Columns" -> {{Left}}},
            GridBoxDividers->{"Columns" -> {{False}}, "Rows" -> {{False}}},
            
            GridBoxItemSize->{
             "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.5599999999999999]}, 
                Offset[0.27999999999999997`]}, "Rows" -> {
                Offset[0.2], {
                 Offset[3.6]}, 
                Offset[0.2]}}],
           FrameMargins->6], ""},
         {
          ItemBox[
           TagBox[
            ButtonBox[
             PaneSelectorBox[{False->
              
              DynamicBox[FEPrivate`FrontEndResource[
               "FEBitmaps", "UpPointerOpener"]], True->
              
              DynamicBox[FEPrivate`FrontEndResource[
               "FEBitmaps", "UpPointerOpenerHot"]]}, Dynamic[
              System`InformationDump`mouseOver$$]],
             Alignment->Left,
             Appearance->{"Default" -> None},
             
             ButtonFunction:>FEPrivate`Set[
              System`InformationDump`open$$, False],
             Evaluator->Automatic,
             FrameMargins->{{9, 0}, {0, 0}},
             ImageMargins->0,
             ImageSize->Full,
             Method->"Preemptive"],
            
            EventHandlerTag[{
             "MouseEntered" :> 
              FEPrivate`Set[System`InformationDump`mouseOver$$, True], 
              "MouseExited" :> 
              FEPrivate`Set[System`InformationDump`mouseOver$$, False], 
              Method -> "Preemptive", PassEventsDown -> Automatic, 
              PassEventsUp -> True}]],
           BaseStyle->"InformationTitleBackground",
           StripOnInput->False], "\[SpanFromLeft]"}
        },
        AutoDelete->False,
        FrameStyle->Directive[
          GrayLevel[0.8], 
          Thickness[Tiny]],
        GridBoxAlignment->{"Columns" -> {Left, Right}, "Rows" -> {{Center}}},
        GridBoxDividers->{
         "Columns" -> {{None}}, "Rows" -> {False, {True}, False}},
        GridBoxItemSize->{
         "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
       "Grid"], False->
      TagBox[GridBox[{
         {
          ItemBox[
           PaneBox[
            StyleBox["\<\" Symbol\"\>", "InformationTitleText",
             StripOnInput->False],
            FrameMargins->{{4, 0}, {-1, 1}}],
           BaseStyle->"InformationTitleBackground",
           StripOnInput->False], 
          ItemBox["\<\"\"\>",
           BaseStyle->"InformationTitleBackground",
           StripOnInput->False]},
         {
          ItemBox[
           PaneBox[
            
            StyleBox["\<\"DaggerCIndex[basis, cindex] returns the conjugated \
index to the cindex (assumed to belong to basis or -basis), in general a Dir \
expression. This function should be handled through upvalues for basis.\"\>", 
             "InformationUsageText",
             StripOnInput->False,
             LineSpacing->{1.5, 1.5, 3.}],
            FrameMargins->{{10, 10}, {8, 10}}],
           BaseStyle->"InformationUsageSubtitleBackground",
           StripOnInput->False], 
          ItemBox["\<\"\"\>",
           BaseStyle->"InformationUsageSubtitleBackground",
           StripOnInput->False]},
         {
          ItemBox[
           TagBox[
            ButtonBox[
             PaneSelectorBox[{False->
              
              DynamicBox[FEPrivate`FrontEndResource[
               "FEBitmaps", "DownPointerOpener"],
               ImageSizeCache->{10., {2., 8.}}], True->
              
              DynamicBox[FEPrivate`FrontEndResource[
               "FEBitmaps", "DownPointerOpenerHot"],
               ImageSizeCache->{10., {2., 8.}}]}, Dynamic[
              System`InformationDump`mouseOver$$]],
             Alignment->Left,
             Appearance->{"Default" -> None},
             
             ButtonFunction:>FEPrivate`Set[
              System`InformationDump`open$$, True],
             Evaluator->Automatic,
             FrameMargins->{{9, 0}, {0, 0}},
             ImageMargins->0,
             ImageSize->Full,
             Method->"Preemptive"],
            
            EventHandlerTag[{
             "MouseEntered" :> 
              FEPrivate`Set[System`InformationDump`mouseOver$$, True], 
              "MouseExited" :> 
              FEPrivate`Set[System`InformationDump`mouseOver$$, False], 
              Method -> "Preemptive", PassEventsDown -> Automatic, 
              PassEventsUp -> True}]],
           BaseStyle->"InformationTitleBackground",
           StripOnInput->False], "\[SpanFromLeft]"}
        },
        AutoDelete->False,
        FrameStyle->Directive[
          GrayLevel[0.8], 
          Thickness[Tiny]],
        GridBoxAlignment->{"Columns" -> {Left, Right}, "Rows" -> {{Center}}},
        GridBoxDividers->{
         "Columns" -> {{None}}, "Rows" -> {False, {True}, False}},
        GridBoxItemSize->{
         "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
       "Grid"]}, Dynamic[System`InformationDump`open$$],
      BaselinePosition->Baseline,
      FrameMargins->0,
      ImageSize->Automatic],
     DynamicModuleValues:>{}],
    BaseStyle->"InformationGridFrame",
    StripOnInput->False], "InformationGridPlain",
   StripOnInput->False],
  InformationData[
   Association[
   "ObjectType" -> "Symbol", "Usage" -> 
    "DaggerCIndex[basis, cindex] returns the conjugated index to the cindex \
(assumed to belong to basis or -basis), in general a Dir expression. This \
function should be handled through upvalues for basis.", "Documentation" -> 
    None, "OwnValues" -> None, "UpValues" -> None, "DownValues" -> 
    Information`InformationValueForm[
     DownValues, xAct`xCoba`DaggerCIndex, {
      xAct`xCoba`DaggerCIndex[] :> Message[
         MessageName[xAct`xCoba`DaggerCIndex, "argrx"], 
         xAct`xCoba`DaggerCIndex, 0, 2], 
       xAct`xCoba`DaggerCIndex[-Pattern[xAct`xCoba`Private`basis, 
           Blank[]], 
         Pattern[xAct`xCoba`Private`cindex, 
          Blank[]]] :> 
       xAct`xCoba`DaggerCIndex[
        xAct`xCoba`Private`basis, xAct`xCoba`Private`cindex], 
       xAct`xCoba`DaggerCIndex[
         Blank[], {
          Pattern[xAct`xCoba`Private`a, 
           Blank[Integer]], 
          Pattern[xAct`xCoba`Private`basis, 
           Blank[]]}] :> {xAct`xCoba`Private`a, 
         xAct`xTensor`Dagger[xAct`xCoba`Private`basis]}, 
       xAct`xCoba`DaggerCIndex[
         Blank[], 
         PatternTest[
          Pattern[xAct`xCoba`Private`basis, 
           Blank[]], xAct`xTensor`BasisQ][
          Pattern[xAct`xCoba`Private`a, 
           Blank[Integer]]]] :> 
       xAct`xTensor`Dagger[xAct`xCoba`Private`basis][xAct`xCoba`Private`a], 
       xAct`xCoba`DaggerCIndex[
         Blank[]] :> Message[
         MessageName[xAct`xCoba`DaggerCIndex, "argr"], 
         xAct`xCoba`DaggerCIndex, 2], xAct`xCoba`DaggerCIndex[
         Blank[], 
         Blank[], 
         Pattern[xAct`xCore`Private`x, 
          BlankSequence[]]] :> Message[
         MessageName[xAct`xCoba`DaggerCIndex, "argrx"], 
         xAct`xCoba`DaggerCIndex, 2 + Length[{xAct`xCore`Private`x}], 2]}], 
    "SubValues" -> None, "DefaultValues" -> None, "NValues" -> None, 
    "FormatValues" -> None, "Options" -> None, "Attributes" -> {Protected}, 
    "FullName" -> "xAct`xCoba`DaggerCIndex"], False]]], "Output",
 CellChangeTimes->{3.792273331301209*^9},
 CellLabel->"Out[117]=",
 CellID->1287333306,ExpressionUUID->"70d492b0-9c07-41e9-bb22-859d844a16e4"]
}, Open  ]],

Cell[BoxData[{
 RowBox[{
  RowBox[{"test", "/:", 
   RowBox[{"DaggerCIndex", "[", 
    RowBox[{"test", ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", "test"}], "}"}]}], "]"}], "=", 
   RowBox[{"{", 
    RowBox[{"2", ",", "test"}], "}"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"test", "/:", 
   RowBox[{"DaggerCIndex", "[", 
    RowBox[{"test", ",", 
     RowBox[{"{", 
      RowBox[{"2", ",", "test"}], "}"}]}], "]"}], "=", 
   RowBox[{"{", 
    RowBox[{"1", ",", "test"}], "}"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"test", "/:", 
   RowBox[{"DaggerCIndex", "[", 
    RowBox[{"test", ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "test"}]}], "}"}]}], "]"}], "=", 
   RowBox[{"{", 
    RowBox[{"2", ",", 
     RowBox[{"-", "test"}]}], "}"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"test", "/:", 
   RowBox[{"DaggerCIndex", "[", 
    RowBox[{"test", ",", 
     RowBox[{"{", 
      RowBox[{"2", ",", 
       RowBox[{"-", "test"}]}], "}"}]}], "]"}], "=", 
   RowBox[{"{", 
    RowBox[{"1", ",", 
     RowBox[{"-", "test"}]}], "}"}]}], ";"}]}], "Input",
 CellLabel->
  "In[118]:=",ExpressionUUID->"6990f634-9fbb-423e-90c8-0366ca3959de"]
}, Open  ]],

Cell[CellGroupData[{

Cell["And now:", "MathCaption",
 CellChangeTimes->{
  3.728662597786602*^9},ExpressionUUID->"2a3e7300-524a-47ae-9d03-\
2d4426b648ff"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Dagger", "/@", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"Basis", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", "test"}], "}"}]}], "]"}], ",", 
    RowBox[{"Basis", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"2", ",", 
        RowBox[{"-", "test"}]}], "}"}], ",", "a"}], "]"}]}], "}"}]}]], "Input",\

 CellLabel->
  "In[122]:=",ExpressionUUID->"5107a35c-cee4-4852-bf67-585e2e2e102e"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`test}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`test}, $CellContext`a],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728662598929804*^9, 3.728671250398526*^9, 
  3.7286748922505074`*^9, 3.7287432100997057`*^9, 3.728743305752893*^9, 
  3.728743402025063*^9, 3.728747606352209*^9, 3.728749555598225*^9, 
  3.728749942675518*^9, 3.7287507816376953`*^9, 3.72875106419559*^9, 
  3.7287516300940447`*^9, 3.728813276826453*^9, 3.728827393745614*^9, 
  3.728835985106072*^9, 3.728836300995817*^9, 3.728843110882763*^9, 
  3.7288456360833893`*^9, 3.73336739581744*^9, 3.79227333626521*^9},
 CellLabel->"Out[122]=",
 CellID->1835792792,ExpressionUUID->"a7a6821f-d72a-4815-aced-d68b5a5658b7"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
If we define a complex tensor on M3 we can see that this works\
\>", "MathCaption",ExpressionUUID->"c74d384d-0b63-428c-a277-971b4d88f69b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"J", "[", "a", "]"}], ",", "M3", ",", 
   RowBox[{"Dagger", "\[Rule]", " ", "Complex"}]}], "]"}]], "Input",
 CellLabel->
  "In[123]:=",ExpressionUUID->"346fa2a8-49aa-40ec-8727-9793e6fab7f1"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"J", "[", "a", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`J[$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728662600651246*^9, 3.7286712505475492`*^9, 
  3.728674892384047*^9, 3.728743213731908*^9, 3.728743305884169*^9, 
  3.728743402150305*^9, 3.728747608346651*^9, 3.7287495566299047`*^9, 
  3.728749942865288*^9, 3.7287507817918*^9, 3.728751064345154*^9, 
  3.728751630278948*^9, 3.7288132771157017`*^9, 3.728827393962731*^9, 
  3.7288359852176743`*^9, 3.728836301138102*^9, 3.728843111289583*^9, 
  3.728845636251433*^9, 3.733367395952888*^9, 3.792273338129738*^9},
 CellLabel->"During evaluation of In[123]:=",
 CellID->1706961440,ExpressionUUID->"28e6af69-97a6-4888-ad68-8277aba051bd"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"J\[Dagger]", "[", "a", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`J\[Dagger][$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728662600651246*^9, 3.7286712505475492`*^9, 
  3.728674892384047*^9, 3.728743213731908*^9, 3.728743305884169*^9, 
  3.728743402150305*^9, 3.728747608346651*^9, 3.7287495566299047`*^9, 
  3.728749942865288*^9, 3.7287507817918*^9, 3.728751064345154*^9, 
  3.728751630278948*^9, 3.7288132771157017`*^9, 3.728827393962731*^9, 
  3.7288359852176743`*^9, 3.728836301138102*^9, 3.728843111289583*^9, 
  3.728845636251433*^9, 3.733367395952888*^9, 3.792273338138747*^9},
 CellLabel->"During evaluation of In[123]:=",
 CellID->666881807,ExpressionUUID->"014c562f-73b8-4614-b07b-5e8df844cb66"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Dagger", "[", 
  RowBox[{"J", "[", 
   RowBox[{"{", 
    RowBox[{"2", ",", "test"}], "}"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[124]:=",ExpressionUUID->"9ca6c30f-993a-40e0-b2e8-04159f7b92ff"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"J\[Dagger]", 
      StyleBox[GridBox[{
         {
          StyleBox["1",
           FontColor->RGBColor[1., 0., 0.]]},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`J\[Dagger][{1, $CellContext`test}],
  Editable->False]], "Output",
 CellChangeTimes->{3.7286626015496798`*^9, 3.7286712511715803`*^9, 
  3.728674892555069*^9, 3.728743214899925*^9, 3.728743306068276*^9, 
  3.728743402346798*^9, 3.728747609319138*^9, 3.728749557196179*^9, 
  3.728749943034581*^9, 3.728750782293539*^9, 3.728751064541338*^9, 
  3.7287516304494267`*^9, 3.72881327728329*^9, 3.728827394227569*^9, 
  3.728835985382608*^9, 3.728836301305984*^9, 3.7288431119524117`*^9, 
  3.728845636465294*^9, 3.73336739646026*^9, 3.792273338697386*^9},
 CellLabel->"Out[124]=",
 CellID->2138518791,ExpressionUUID->"20064156-7601-419d-b885-3c07ebe18bea"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We can even establish more complicated relations, if we remember that Basis \
and Dir act as inverses of each other\
\>", "MathCaption",ExpressionUUID->"b998c5c5-aab8-4ba3-aca9-9822439a4e2d"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"test", "/:", 
   RowBox[{"DaggerCIndex", "[", 
    RowBox[{"test", ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", "test"}], "}"}]}], "]"}], "=", 
   RowBox[{"Dir", "[", 
    RowBox[{"I", " ", 
     RowBox[{"Basis", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"{", 
        RowBox[{"2", ",", "test"}], "}"}]}], "]"}]}], "]"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"test", "/:", 
   RowBox[{"DaggerCIndex", "[", 
    RowBox[{"test", ",", 
     RowBox[{"{", 
      RowBox[{"2", ",", "test"}], "}"}]}], "]"}], "=", 
   RowBox[{"Dir", "[", 
    RowBox[{"I", " ", 
     RowBox[{"Basis", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"{", 
        RowBox[{"1", ",", "test"}], "}"}]}], "]"}]}], "]"}]}], 
  ";"}]}], "Input",
 CellLabel->
  "In[125]:=",ExpressionUUID->"f29c6c77-f41d-452f-af3e-304caf504daa"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Dagger", "[", 
  RowBox[{"J", "[", 
   RowBox[{"{", 
    RowBox[{"2", ",", "test"}], "}"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[127]:=",ExpressionUUID->"40b02f95-2a4e-48df-86a9-6cadabd90bd2"],

Cell[BoxData[
 RowBox[{"\[ImaginaryI]", " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"J\[Dagger]", 
       StyleBox[GridBox[{
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`J\[Dagger][{1, $CellContext`test}],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728662603705291*^9, 3.728671251553388*^9, 
  3.728674892757125*^9, 3.728743219134955*^9, 3.728743306345037*^9, 
  3.728743402515512*^9, 3.728747611694564*^9, 3.72874955877943*^9, 
  3.72874994321653*^9, 3.7287507827941303`*^9, 3.728751064896092*^9, 
  3.728751630773662*^9, 3.7288132774823914`*^9, 3.728827394411394*^9, 
  3.728835985554162*^9, 3.72883630152188*^9, 3.728843113003737*^9, 
  3.7288456366992683`*^9, 3.733367396668973*^9, 3.7922733404328003`*^9},
 CellLabel->"Out[127]=",
 CellID->711058643,ExpressionUUID->"3947039e-d028-4533-b693-5b72b9db7ef1"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"UndefBasis", "[", "test", "]"}], ";"}]], "Input",
 CellLabel->
  "In[128]:=",ExpressionUUID->"ac129649-f83e-453c-babf-02995386158b"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "ChristoffelPDtest"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "non-symmetric Christoffel tensor", " ", $CellContext`ChristoffelPDtest],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286626045098763`*^9, 3.728671251705339*^9, 
  3.728674892884242*^9, 3.728743220050411*^9, 3.7287433065214567`*^9, 
  3.7287434026481123`*^9, 3.728747612287574*^9, 3.728749559360145*^9, 
  3.728749943353883*^9, 3.728750782957477*^9, 3.728751065067802*^9, 
  3.728751630893968*^9, 3.728813277649822*^9, 3.728827394546196*^9, 
  3.7288359856773148`*^9, 3.7288363016877937`*^9, 3.7288431132055492`*^9, 
  3.728845636847151*^9, 3.7333673968064833`*^9, 3.792273341556437*^9},
 CellLabel->"During evaluation of In[128]:=",
 CellID->1866498034,ExpressionUUID->"c36167c3-2d5a-42ae-8ed3-44c817f6f55f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciPDtest"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Ricci tensor", 
   " ", $CellContext`RicciPDtest],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286626045098763`*^9, 3.728671251705339*^9, 
  3.728674892884242*^9, 3.728743220050411*^9, 3.7287433065214567`*^9, 
  3.7287434026481123`*^9, 3.728747612287574*^9, 3.728749559360145*^9, 
  3.728749943353883*^9, 3.728750782957477*^9, 3.728751065067802*^9, 
  3.728751630893968*^9, 3.728813277649822*^9, 3.728827394546196*^9, 
  3.7288359856773148`*^9, 3.7288363016877937`*^9, 3.7288431132055492`*^9, 
  3.728845636847151*^9, 3.7333673968064833`*^9, 3.7922733415684547`*^9},
 CellLabel->"During evaluation of In[128]:=",
 CellID->1938825709,ExpressionUUID->"7097503f-b615-40d6-b6e3-38ca750d199a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannPDtest"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannPDtest],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286626045098763`*^9, 3.728671251705339*^9, 
  3.728674892884242*^9, 3.728743220050411*^9, 3.7287433065214567`*^9, 
  3.7287434026481123`*^9, 3.728747612287574*^9, 3.728749559360145*^9, 
  3.728749943353883*^9, 3.728750782957477*^9, 3.728751065067802*^9, 
  3.728751630893968*^9, 3.728813277649822*^9, 3.728827394546196*^9, 
  3.7288359856773148`*^9, 3.7288363016877937`*^9, 3.7288431132055492`*^9, 
  3.728845636847151*^9, 3.7333673968064833`*^9, 3.7922733415826263`*^9},
 CellLabel->"During evaluation of In[128]:=",
 CellID->2074315232,ExpressionUUID->"4c7f35cd-420f-4809-800a-abb39e54752e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TorsionPDtest"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionPDtest],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286626045098763`*^9, 3.728671251705339*^9, 
  3.728674892884242*^9, 3.728743220050411*^9, 3.7287433065214567`*^9, 
  3.7287434026481123`*^9, 3.728747612287574*^9, 3.728749559360145*^9, 
  3.728749943353883*^9, 3.728750782957477*^9, 3.728751065067802*^9, 
  3.728751630893968*^9, 3.728813277649822*^9, 3.728827394546196*^9, 
  3.7288359856773148`*^9, 3.7288363016877937`*^9, 3.7288431132055492`*^9, 
  3.728845636847151*^9, 3.7333673968064833`*^9, 3.792273341589675*^9},
 CellLabel->"During evaluation of In[128]:=",
 CellID->766578826,ExpressionUUID->"f861fcff-37b1-4048-9d3e-d264b52c6d15"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"parallel derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "PDtest"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "parallel derivative", 
   " ", $CellContext`PDtest],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286626045098763`*^9, 3.728671251705339*^9, 
  3.728674892884242*^9, 3.728743220050411*^9, 3.7287433065214567`*^9, 
  3.7287434026481123`*^9, 3.728747612287574*^9, 3.728749559360145*^9, 
  3.728749943353883*^9, 3.728750782957477*^9, 3.728751065067802*^9, 
  3.728751630893968*^9, 3.728813277649822*^9, 3.728827394546196*^9, 
  3.7288359856773148`*^9, 3.7288363016877937`*^9, 3.7288431132055492`*^9, 
  3.728845636847151*^9, 3.7333673968064833`*^9, 3.792273341596857*^9},
 CellLabel->"During evaluation of In[128]:=",
 CellID->1135930229,ExpressionUUID->"2ad2bbb2-1fd7-466f-b734-af4e57cc52e2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "etaUptest"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric +1 density",
    " ", $CellContext`etaUptest],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286626045098763`*^9, 3.728671251705339*^9, 
  3.728674892884242*^9, 3.728743220050411*^9, 3.7287433065214567`*^9, 
  3.7287434026481123`*^9, 3.728747612287574*^9, 3.728749559360145*^9, 
  3.728749943353883*^9, 3.728750782957477*^9, 3.728751065067802*^9, 
  3.728751630893968*^9, 3.728813277649822*^9, 3.728827394546196*^9, 
  3.7288359856773148`*^9, 3.7288363016877937`*^9, 3.7288431132055492`*^9, 
  3.728845636847151*^9, 3.7333673968064833`*^9, 3.792273341604622*^9},
 CellLabel->"During evaluation of In[128]:=",
 CellID->449736973,ExpressionUUID->"76106789-5b14-4fda-958e-43fc1a3cd5f2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "etaDowntest"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric -1 density",
    " ", $CellContext`etaDowntest],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286626045098763`*^9, 3.728671251705339*^9, 
  3.728674892884242*^9, 3.728743220050411*^9, 3.7287433065214567`*^9, 
  3.7287434026481123`*^9, 3.728747612287574*^9, 3.728749559360145*^9, 
  3.728749943353883*^9, 3.728750782957477*^9, 3.728751065067802*^9, 
  3.728751630893968*^9, 3.728813277649822*^9, 3.728827394546196*^9, 
  3.7288359856773148`*^9, 3.7288363016877937`*^9, 3.7288431132055492`*^9, 
  3.728845636847151*^9, 3.7333673968064833`*^9, 3.7922733416123734`*^9},
 CellLabel->"During evaluation of In[128]:=",
 CellID->1734351418,ExpressionUUID->"f2f2eaf2-db82-4ae3-9505-fdc9d78f79dc"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefBasis", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"basis\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "test"}],
  SequenceForm[
  "** ", xAct`xCoba`UndefBasis, ": Undefined ", "basis", 
   " ", $CellContext`test],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286626045098763`*^9, 3.728671251705339*^9, 
  3.728674892884242*^9, 3.728743220050411*^9, 3.7287433065214567`*^9, 
  3.7287434026481123`*^9, 3.728747612287574*^9, 3.728749559360145*^9, 
  3.728749943353883*^9, 3.728750782957477*^9, 3.728751065067802*^9, 
  3.728751630893968*^9, 3.728813277649822*^9, 3.728827394546196*^9, 
  3.7288359856773148`*^9, 3.7288363016877937`*^9, 3.7288431132055492`*^9, 
  3.728845636847151*^9, 3.7333673968064833`*^9, 3.79227334161996*^9},
 CellLabel->"During evaluation of In[128]:=",
 CellID->1448741595,ExpressionUUID->"bb1a7bb2-4c97-4b61-aab9-4a3c853a3aae"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["1.1.4. BasisArray", "Subsubsection",
 CellChangeTimes->{{3.728673440948658*^9, 
  3.728673441228715*^9}},ExpressionUUID->"2293933f-3cd3-45d8-9482-\
8786a1dadffa"],

Cell[TextData[{
 StyleBox["BasisArray\t\t\t", "DisplayMath"],
 StyleBox["List of basis vectors", "Text"]
}], "DefinitionBox",ExpressionUUID->"ae30a1d2-ed84-443a-b53c-1ab209af4509"],

Cell["Listing basis objects", "Text",ExpressionUUID->"71caf07e-a74c-4e54-b6d5-a6d4918e2d7d"],

Cell[CellGroupData[{

Cell[TextData[{
 "The function ",
 StyleBox["BasisArray", "MR"],
 " returns a list of all basis vectors of a given basis:"
}], "MathCaption",
 CellChangeTimes->{
  3.728662629057212*^9},ExpressionUUID->"36f5ef7f-e782-4794-96de-\
a953abadd8ac"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"BasisArray", "[", "blues", "]"}], "[", "a", "]"}]], "Input",
 CellChangeTimes->{{3.728743232041184*^9, 3.7287432325706367`*^9}, {
   3.7287476192189407`*^9, 3.728747620194909*^9}, 3.7287493010675297`*^9},
 CellLabel->
  "In[129]:=",ExpressionUUID->"3c419301-418c-4d4c-9a96-ce367f785028"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`blues}, $CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`blues}, $CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`blues}, $CellContext`a],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728662630718644*^9, 3.728671255566935*^9, 
  3.728674893267625*^9, 3.728743249286961*^9, 3.728743306786578*^9, 
  3.7287434029489317`*^9, 3.7287476209716377`*^9, 3.728749565010408*^9, 
  3.728749943599453*^9, 3.728750787158956*^9, 3.728751065300852*^9, 
  3.728751631162703*^9, 3.728813277915422*^9, 3.728827394822345*^9, 
  3.728835985994619*^9, 3.728836301989148*^9, 3.728843120517087*^9, 
  3.728845637195084*^9, 3.733367401628311*^9, 3.792273347459838*^9},
 CellLabel->"Out[129]=",
 CellID->627601471,ExpressionUUID->"19a17f7a-49ba-4bf5-81ee-6f1138d02cea"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"BasisArray", "[", "fcomp", "]"}], "[", "A", "]"}]], "Input",
 CellChangeTimes->{
  3.728743235200674*^9, {3.728747622988144*^9, 3.728747624306611*^9}},
 CellLabel->
  "In[130]:=",ExpressionUUID->"1a09cf95-6c5f-4394-b6e3-bab07722f3da"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "A"},
           {
            StyleBox[
             RowBox[{"-", "1"}],
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-1, -$CellContext`fcomp}, $CellContext`A],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "A"},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`fcomp}, $CellContext`A],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728662631418577*^9, 3.728671255753602*^9, 
  3.728674893469021*^9, 3.728743250158434*^9, 3.7287433069203377`*^9, 
  3.728743403082686*^9, 3.728747624788251*^9, 3.7287495652013817`*^9, 
  3.728749943721011*^9, 3.728750787344207*^9, 3.728751065451335*^9, 
  3.728751631316011*^9, 3.728813278056772*^9, 3.728827394965919*^9, 
  3.7288359860850687`*^9, 3.728836302105842*^9, 3.728843120716978*^9, 
  3.728845637528639*^9, 3.733367401752824*^9, 3.792273347518866*^9},
 CellLabel->"Out[130]=",
 CellID->1805732680,ExpressionUUID->"bf559b56-6d36-45a5-9c5a-3f4aceada8a1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["If we give several indices, the product basis is displayed:", \
"MathCaption",
 CellChangeTimes->{
  3.728662633441082*^9},ExpressionUUID->"3dcea435-b176-45a1-886a-\
4e555c34ee13"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"BasisArray", "[", 
   RowBox[{"fred", ",", "fred"}], "]"}], "[", 
  RowBox[{"a", ",", "b"}], "]"}]], "Input",
 CellChangeTimes->{{3.7287432379546537`*^9, 3.728743239698331*^9}, {
  3.7287476272666197`*^9, 3.728747629962434*^9}},
 CellLabel->
  "In[131]:=",ExpressionUUID->"96e1e910-c066-44ab-a78c-bc9ba62e0a4c"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", "a"},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{0, -$CellContext`fred}, $CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", "b"},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{0, -$CellContext`fred}, $CellContext`b],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", "a"},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{0, -$CellContext`fred}, $CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", "b"},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`b],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", "a"},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{0, -$CellContext`fred}, $CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", "b"},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{2, -$CellContext`fred}, $CellContext`b],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", "b"},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{0, -$CellContext`fred}, $CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", "a"},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", "a"},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", "b"},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`b],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", "a"},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", "b"},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{2, -$CellContext`fred}, $CellContext`b],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", "b"},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{0, -$CellContext`fred}, $CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", "a"},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{2, -$CellContext`fred}, $CellContext`a],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", "b"},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", "a"},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{2, -$CellContext`fred}, $CellContext`a],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", "a"},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{2, -$CellContext`fred}, $CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", "b"},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{2, -$CellContext`fred}, $CellContext`b],
       Editable->False]}]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.7286626345827093`*^9, 3.728671255978422*^9, 
  3.7286748936159554`*^9, 3.72874325173666*^9, 3.728743307068747*^9, 
  3.728743403265562*^9, 3.728747630422699*^9, 3.728749565422147*^9, 
  3.728749943865258*^9, 3.728750787583197*^9, 3.7287510656293507`*^9, 
  3.728751631495434*^9, 3.728813278187914*^9, 3.728827395120206*^9, 
  3.7288359862023697`*^9, 3.728836302303639*^9, 3.728843120977862*^9, 
  3.728845637693138*^9, 3.7333674018816223`*^9, 3.792273347581558*^9},
 CellLabel->"Out[131]=",
 CellID->710253356,ExpressionUUID->"68815388-709c-4091-8ba1-78b25af8ae7f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We can mix bases:", "MathCaption",
 CellChangeTimes->{
  3.728662635840891*^9},ExpressionUUID->"7cfbeaad-18f5-437b-8e09-\
4ca35d55f9ef"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"BasisArray", "[", 
    RowBox[{"fred", ",", "blues", ",", "fcomp"}], "]"}], "[", 
   RowBox[{"a", ",", "c", ",", 
    RowBox[{"-", "B"}]}], "]"}], "//", "MatrixForm"}]], "Input",
 CellChangeTimes->{{3.728743241833324*^9, 3.728743246561861*^9}, {
   3.7287476326280937`*^9, 3.728747637898013*^9}, 3.728749302018683*^9},
 CellLabel->
  "In[132]:=",ExpressionUUID->"b397e55b-f385-4af2-9549-8033b90016e2"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      RowBox[{"(", "\[NoBreak]", 
       TagBox[GridBox[{
          {
           RowBox[{
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", 
                    StyleBox[
                    RowBox[{"-", "1"}],
                    FontColor->RGBColor[0., 1., 0.]]},
                    {"B", " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[-$CellContext`B, {-1, $CellContext`fcomp}],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "c"},
                    {
                    StyleBox["0",
                    FontColor->RGBColor[0., 0., 1.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{0, -$CellContext`blues}, $CellContext`c],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "a"},
                    {
                    StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{0, -$CellContext`fred}, $CellContext`a],
             Editable->False]}]},
          {
           RowBox[{
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[0., 1., 0.]]},
                    {"B", " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[-$CellContext`B, {1, $CellContext`fcomp}],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "c"},
                    {
                    StyleBox["0",
                    FontColor->RGBColor[0., 0., 1.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{0, -$CellContext`blues}, $CellContext`c],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "a"},
                    {
                    StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{0, -$CellContext`fred}, $CellContext`a],
             Editable->False]}]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Baseline}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.5599999999999999]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        Column], "\[NoBreak]", ")"}], 
      RowBox[{"(", "\[NoBreak]", 
       TagBox[GridBox[{
          {
           RowBox[{
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", 
                    StyleBox[
                    RowBox[{"-", "1"}],
                    FontColor->RGBColor[0., 1., 0.]]},
                    {"B", " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[-$CellContext`B, {-1, $CellContext`fcomp}],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "a"},
                    {
                    StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{0, -$CellContext`fred}, $CellContext`a],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "c"},
                    {
                    StyleBox["1",
                    FontColor->RGBColor[0., 0., 1.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{1, -$CellContext`blues}, $CellContext`c],
             Editable->False]}]},
          {
           RowBox[{
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[0., 1., 0.]]},
                    {"B", " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[-$CellContext`B, {1, $CellContext`fcomp}],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "a"},
                    {
                    StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{0, -$CellContext`fred}, $CellContext`a],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "c"},
                    {
                    StyleBox["1",
                    FontColor->RGBColor[0., 0., 1.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{1, -$CellContext`blues}, $CellContext`c],
             Editable->False]}]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Baseline}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.5599999999999999]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        Column], "\[NoBreak]", ")"}], 
      RowBox[{"(", "\[NoBreak]", 
       TagBox[GridBox[{
          {
           RowBox[{
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", 
                    StyleBox[
                    RowBox[{"-", "1"}],
                    FontColor->RGBColor[0., 1., 0.]]},
                    {"B", " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[-$CellContext`B, {-1, $CellContext`fcomp}],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "a"},
                    {
                    StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{0, -$CellContext`fred}, $CellContext`a],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "c"},
                    {
                    StyleBox["2",
                    FontColor->RGBColor[0., 0., 1.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{2, -$CellContext`blues}, $CellContext`c],
             Editable->False]}]},
          {
           RowBox[{
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[0., 1., 0.]]},
                    {"B", " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[-$CellContext`B, {1, $CellContext`fcomp}],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "a"},
                    {
                    StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{0, -$CellContext`fred}, $CellContext`a],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "c"},
                    {
                    StyleBox["2",
                    FontColor->RGBColor[0., 0., 1.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{2, -$CellContext`blues}, $CellContext`c],
             Editable->False]}]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Baseline}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.5599999999999999]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        Column], "\[NoBreak]", ")"}]},
     {
      RowBox[{"(", "\[NoBreak]", 
       TagBox[GridBox[{
          {
           RowBox[{
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", 
                    StyleBox[
                    RowBox[{"-", "1"}],
                    FontColor->RGBColor[0., 1., 0.]]},
                    {"B", " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[-$CellContext`B, {-1, $CellContext`fcomp}],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "c"},
                    {
                    StyleBox["0",
                    FontColor->RGBColor[0., 0., 1.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{0, -$CellContext`blues}, $CellContext`c],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "a"},
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
             Editable->False]}]},
          {
           RowBox[{
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[0., 1., 0.]]},
                    {"B", " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[-$CellContext`B, {1, $CellContext`fcomp}],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "c"},
                    {
                    StyleBox["0",
                    FontColor->RGBColor[0., 0., 1.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{0, -$CellContext`blues}, $CellContext`c],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "a"},
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
             Editable->False]}]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Baseline}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.5599999999999999]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        Column], "\[NoBreak]", ")"}], 
      RowBox[{"(", "\[NoBreak]", 
       TagBox[GridBox[{
          {
           RowBox[{
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", 
                    StyleBox[
                    RowBox[{"-", "1"}],
                    FontColor->RGBColor[0., 1., 0.]]},
                    {"B", " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[-$CellContext`B, {-1, $CellContext`fcomp}],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "c"},
                    {
                    StyleBox["1",
                    FontColor->RGBColor[0., 0., 1.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{1, -$CellContext`blues}, $CellContext`c],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "a"},
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
             Editable->False]}]},
          {
           RowBox[{
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[0., 1., 0.]]},
                    {"B", " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[-$CellContext`B, {1, $CellContext`fcomp}],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "c"},
                    {
                    StyleBox["1",
                    FontColor->RGBColor[0., 0., 1.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{1, -$CellContext`blues}, $CellContext`c],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "a"},
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
             Editable->False]}]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Baseline}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.5599999999999999]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        Column], "\[NoBreak]", ")"}], 
      RowBox[{"(", "\[NoBreak]", 
       TagBox[GridBox[{
          {
           RowBox[{
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", 
                    StyleBox[
                    RowBox[{"-", "1"}],
                    FontColor->RGBColor[0., 1., 0.]]},
                    {"B", " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[-$CellContext`B, {-1, $CellContext`fcomp}],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "a"},
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "c"},
                    {
                    StyleBox["2",
                    FontColor->RGBColor[0., 0., 1.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{2, -$CellContext`blues}, $CellContext`c],
             Editable->False]}]},
          {
           RowBox[{
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[0., 1., 0.]]},
                    {"B", " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[-$CellContext`B, {1, $CellContext`fcomp}],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "a"},
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "c"},
                    {
                    StyleBox["2",
                    FontColor->RGBColor[0., 0., 1.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{2, -$CellContext`blues}, $CellContext`c],
             Editable->False]}]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Baseline}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.5599999999999999]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        Column], "\[NoBreak]", ")"}]},
     {
      RowBox[{"(", "\[NoBreak]", 
       TagBox[GridBox[{
          {
           RowBox[{
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", 
                    StyleBox[
                    RowBox[{"-", "1"}],
                    FontColor->RGBColor[0., 1., 0.]]},
                    {"B", " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[-$CellContext`B, {-1, $CellContext`fcomp}],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "c"},
                    {
                    StyleBox["0",
                    FontColor->RGBColor[0., 0., 1.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{0, -$CellContext`blues}, $CellContext`c],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "a"},
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{2, -$CellContext`fred}, $CellContext`a],
             Editable->False]}]},
          {
           RowBox[{
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[0., 1., 0.]]},
                    {"B", " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[-$CellContext`B, {1, $CellContext`fcomp}],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "c"},
                    {
                    StyleBox["0",
                    FontColor->RGBColor[0., 0., 1.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{0, -$CellContext`blues}, $CellContext`c],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "a"},
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{2, -$CellContext`fred}, $CellContext`a],
             Editable->False]}]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Baseline}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.5599999999999999]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        Column], "\[NoBreak]", ")"}], 
      RowBox[{"(", "\[NoBreak]", 
       TagBox[GridBox[{
          {
           RowBox[{
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", 
                    StyleBox[
                    RowBox[{"-", "1"}],
                    FontColor->RGBColor[0., 1., 0.]]},
                    {"B", " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[-$CellContext`B, {-1, $CellContext`fcomp}],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "c"},
                    {
                    StyleBox["1",
                    FontColor->RGBColor[0., 0., 1.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{1, -$CellContext`blues}, $CellContext`c],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "a"},
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{2, -$CellContext`fred}, $CellContext`a],
             Editable->False]}]},
          {
           RowBox[{
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[0., 1., 0.]]},
                    {"B", " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[-$CellContext`B, {1, $CellContext`fcomp}],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "c"},
                    {
                    StyleBox["1",
                    FontColor->RGBColor[0., 0., 1.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{1, -$CellContext`blues}, $CellContext`c],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "a"},
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{2, -$CellContext`fred}, $CellContext`a],
             Editable->False]}]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Baseline}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.5599999999999999]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        Column], "\[NoBreak]", ")"}], 
      RowBox[{"(", "\[NoBreak]", 
       TagBox[GridBox[{
          {
           RowBox[{
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", 
                    StyleBox[
                    RowBox[{"-", "1"}],
                    FontColor->RGBColor[0., 1., 0.]]},
                    {"B", " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[-$CellContext`B, {-1, $CellContext`fcomp}],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "c"},
                    {
                    StyleBox["2",
                    FontColor->RGBColor[0., 0., 1.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{2, -$CellContext`blues}, $CellContext`c],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "a"},
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{2, -$CellContext`fred}, $CellContext`a],
             Editable->False]}]},
          {
           RowBox[{
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[0., 1., 0.]]},
                    {"B", " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[-$CellContext`B, {1, $CellContext`fcomp}],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "c"},
                    {
                    StyleBox["2",
                    FontColor->RGBColor[0., 0., 1.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{2, -$CellContext`blues}, $CellContext`c],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", "a"},
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xTensor`Basis[{2, -$CellContext`fred}, $CellContext`a],
             Editable->False]}]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Baseline}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.5599999999999999]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        Column], "\[NoBreak]", ")"}]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Baseline}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{3.728662637101242*^9, 3.7286712561746473`*^9, 
  3.728674893812711*^9, 3.728743252922406*^9, 3.7287433072348557`*^9, 
  3.7287434035022383`*^9, 3.728747638416152*^9, 3.728749565617111*^9, 
  3.728749944055521*^9, 3.7287507877982283`*^9, 3.728751065845194*^9, 
  3.728751631711618*^9, 3.728813278420416*^9, 3.7288273957265863`*^9, 
  3.728835986360104*^9, 3.72883630259731*^9, 3.728843121231201*^9, 
  3.728845637869211*^9, 3.7333674020527153`*^9, 3.792273347680484*^9},
 CellLabel->"Out[132]//MatrixForm=",
 CellID->1514044266,ExpressionUUID->"bac0b3ea-0565-4a8d-8420-9d5c09c4e000"]
}, Open  ]]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["1.2. Basis indices", "Subsection",
 CellChangeTimes->{{3.728662754011717*^9, 3.728662756723887*^9}, {
  3.728673416534724*^9, 
  3.728673428653996*^9}},ExpressionUUID->"72218c58-dd15-4055-929e-\
32e81a9d0e94"],

Cell[CellGroupData[{

Cell["1.2.1. IndicesOf", "Subsubsection",
 CellChangeTimes->{{3.728673447277226*^9, 
  3.728673447413394*^9}},ExpressionUUID->"36e48e02-82e5-4221-87be-\
c7807dcd35b4"],

Cell[TextData[{
 "In ",
 StyleBox["xTensor`", "MR"],
 ", the main user level function to find indices in an expression was ",
 StyleBox["IndicesOf[selectors][expression]", "MR"],
 ". Let us start by reviewin that function."
}], "Text",
 CellChangeTimes->{{3.728743417978653*^9, 
  3.728743438911489*^9}},ExpressionUUID->"54d297ff-67ce-4972-a3f7-\
040388e78001"],

Cell[CellGroupData[{

Cell["Define several tensors:", "MathCaption",
 CellChangeTimes->{{3.728662867607512*^9, 3.728662873854418*^9}},
 CellID->266904421,ExpressionUUID->"b30e1f89-0463-43bb-95c2-da41d44b9bab"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"TT", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", "c", ",", "d", ",", 
     RowBox[{"-", "e"}]}], "]"}], ",", "M3"}], "]"}], "\[IndentingNewLine]", 
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"U", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", "c", ",", "d"}], "]"}], ",", "M3"}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"S", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", "B"}], "]"}], ",", "M3", ",", 
   RowBox[{"Dagger", "\[Rule]", "Complex"}]}], "]"}]}], "Input",
 CellLabel->
  "In[133]:=",ExpressionUUID->"12c43a21-58ec-4b82-9c1a-a7c05b17d05a"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"TT", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", "c", ",", "d", ",", 
     RowBox[{"-", "e"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`TT[-$CellContext`a, -$CellContext`b, $CellContext`c, \
$CellContext`d, -$CellContext`e], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728743478958407*^9, 3.728747651284718*^9, 
  3.728749569519003*^9, 3.728749944165229*^9, 3.728750797945162*^9, 
  3.728751069959074*^9, 3.7287516318716927`*^9, 3.7288132785833197`*^9, 
  3.728827395953147*^9, 3.728835986455418*^9, 3.728836302739628*^9, 
  3.728843127676228*^9, 3.728845637998889*^9, 3.7333674100780687`*^9, 
  3.792273354649199*^9},
 CellLabel->"During evaluation of In[133]:=",
 CellID->920137883,ExpressionUUID->"53693de6-650a-4baf-b4ee-a251cd53ac4c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"U", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", "c", ",", "d"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`U[-$CellContext`a, -$CellContext`b, $CellContext`c, \
$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728743478958407*^9, 3.728747651284718*^9, 
  3.728749569519003*^9, 3.728749944165229*^9, 3.728750797945162*^9, 
  3.728751069959074*^9, 3.7287516318716927`*^9, 3.7288132785833197`*^9, 
  3.728827395953147*^9, 3.728835986455418*^9, 3.728836302739628*^9, 
  3.728843127676228*^9, 3.728845637998889*^9, 3.7333674100780687`*^9, 
  3.7922733546599607`*^9},
 CellLabel->"During evaluation of In[133]:=",
 CellID->727657534,ExpressionUUID->"ded9321e-8b15-4c15-ad44-83e7972f6b4a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"S", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", "B"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`S[-$CellContext`A, $CellContext`B], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728743478958407*^9, 3.728747651284718*^9, 
  3.728749569519003*^9, 3.728749944165229*^9, 3.728750797945162*^9, 
  3.728751069959074*^9, 3.7287516318716927`*^9, 3.7288132785833197`*^9, 
  3.728827395953147*^9, 3.728835986455418*^9, 3.728836302739628*^9, 
  3.728843127676228*^9, 3.728845637998889*^9, 3.7333674100780687`*^9, 
  3.7922733546783*^9},
 CellLabel->"During evaluation of In[133]:=",
 CellID->1968287920,ExpressionUUID->"4ba7f0ca-08a0-42d3-b61b-7593946a5ab2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"S\[Dagger]", "[", 
    RowBox[{
     RowBox[{"-", "A\[Dagger]"}], ",", "B\[Dagger]"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`S\[Dagger][-$CellContext`A\[Dagger], $CellContext`B\[Dagger]],
    ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728743478958407*^9, 3.728747651284718*^9, 
  3.728749569519003*^9, 3.728749944165229*^9, 3.728750797945162*^9, 
  3.728751069959074*^9, 3.7287516318716927`*^9, 3.7288132785833197`*^9, 
  3.728827395953147*^9, 3.728835986455418*^9, 3.728836302739628*^9, 
  3.728843127676228*^9, 3.728845637998889*^9, 3.7333674100780687`*^9, 
  3.7922733546857977`*^9},
 CellLabel->"During evaluation of In[133]:=",
 CellID->1176973052,ExpressionUUID->"10d789d3-6d8e-4f77-8192-91530ffedbec"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Construct an expression containing them:", "MathCaption",
 CellChangeTimes->{{3.7286628855483027`*^9, 3.728662900476931*^9}},
 CellID->617006311,ExpressionUUID->"bc2c03a8-6f62-45fc-ae7a-4e54c61d7dad"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"TT", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "fred"}]}], "}"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"{", 
       RowBox[{"2", ",", "blues"}], "}"}], ",", "c", ",", 
      RowBox[{"-", "a"}]}], "]"}], 
    RowBox[{"Basis", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "fred"}]}], "}"}], ",", "a"}], "]"}]}], "+", 
   RowBox[{
    RowBox[{"Basis", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "fred"}]}], "}"}], ",", "a"}], "]"}], 
    RowBox[{"Basis", "[", 
     RowBox[{
      RowBox[{"-", "d"}], ",", 
      RowBox[{"{", 
       RowBox[{"2", ",", "blues"}], "}"}]}], "]"}], 
    RowBox[{"U", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", "d", ",", 
      RowBox[{"{", 
       RowBox[{"e", ",", "fred"}], "}"}]}], "]"}], 
    RowBox[{"Basis", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "e"}], ",", 
        RowBox[{"-", "fred"}]}], "}"}], ",", "c"}], "]"}], 
    RowBox[{"S", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "A"}], ",", 
        RowBox[{"-", "fcomp"}]}], "}"}], ",", "B"}], "]"}], 
    RowBox[{"S", "[", 
     RowBox[{
      RowBox[{"-", "B"}], ",", 
      RowBox[{"{", 
       RowBox[{"A", ",", "fcomp"}], "}"}]}], "]"}]}]}]}]], "Input",
 CellChangeTimes->{{3.728743485312006*^9, 3.728743507948966*^9}, {
  3.72874765346628*^9, 3.7287476726091223`*^9}, {3.728749302968383*^9, 
  3.728749303666894*^9}},
 CellLabel->
  "In[136]:=",ExpressionUUID->"75f1bb86-54a6-4869-b103-46477260b76c"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"TT", 
        StyleBox[GridBox[{
           {" ", " ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], "c", " "},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], "b", " ", " ", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TT[{1, -$CellContext`fred}, -$CellContext`b, {
     2, $CellContext`blues}, $CellContext`c, -$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {2, $CellContext`blues}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "c"},
           {
            StyleBox["e",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`e, -$CellContext`fred}, $CellContext`c],
    
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["A",
             FontColor->RGBColor[0., 1., 0.]]},
           {"B", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[-$CellContext`B, {$CellContext`A, $CellContext`fcomp}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", "B"},
           {
            StyleBox["A",
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[{-$CellContext`A, -$CellContext`fcomp}, $CellContext`B],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", "d", 
            StyleBox["e",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", "b", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, -$CellContext`b, $CellContext`d, \
{$CellContext`e, $CellContext`fred}],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728662901700288*^9, 3.7286712664372473`*^9, 
  3.728674894261183*^9, 3.728743307564918*^9, 3.728743509266629*^9, 
  3.7287476730941467`*^9, 3.728749570080853*^9, 3.7287499443667803`*^9, 
  3.72875079824466*^9, 3.728751070240017*^9, 3.728751632045661*^9, 
  3.728813278798461*^9, 3.728827396164927*^9, 3.7288359868571243`*^9, 
  3.728836302933477*^9, 3.7288431280525303`*^9, 3.728845638320331*^9, 
  3.7333674103728724`*^9, 3.792273354761351*^9},
 CellLabel->"Out[136]=",
 CellID->914212447,ExpressionUUID->"fce32b15-e43a-46a0-9615-96117e84f21d"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Find different types of indices (See xTensorDoc.nb, section 9.1):\
\>", "MathCaption",
 CellChangeTimes->{{3.7286629064843082`*^9, 3.7286629334287148`*^9}},
 CellID->1719940314,ExpressionUUID->"fd814381-e900-4682-a039-e257f3076cd5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IndicesOf", "[", "Free", "]"}], "[", "expr", "]"}]], "Input",
 CellLabel->
  "In[137]:=",ExpressionUUID->"563f9246-0de7-43ec-be51-90801d6f26e9"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{
     RowBox[{"-", "b"}], 
     RowBox[{",", "\[ThinSpace]"}], "c"}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[-$CellContext`b, $CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{3.728662913025135*^9, 3.728671266601688*^9, 
  3.72867489440442*^9, 3.728743307678035*^9, 3.728743511172538*^9, 
  3.728747675370242*^9, 3.72874957060299*^9, 3.728749944486382*^9, 
  3.728750798402842*^9, 3.728751070455977*^9, 3.72875163219838*^9, 
  3.728813278938417*^9, 3.7288273964052773`*^9, 3.728835987046899*^9, 
  3.7288363030575123`*^9, 3.728843128292523*^9, 3.728845638605145*^9, 
  3.73336741052242*^9, 3.792273354816433*^9},
 CellLabel->"Out[137]=",
 CellID->2070586091,ExpressionUUID->"97a1ee93-2ef3-4e72-80c9-58a69a31ff16"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IndicesOf", "[", "TangentM3", "]"}], "[", "expr", "]"}]], "Input",
 CellLabel->
  "In[138]:=",ExpressionUUID->"62e747e6-e704-464b-bb1f-b730bf6e975b"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "fred"}]}], "}"}], 
     RowBox[{",", "\[ThinSpace]"}], "a", 
     RowBox[{",", "\[ThinSpace]"}], 
     RowBox[{"-", "b"}], 
     RowBox[{",", "\[ThinSpace]"}], 
     RowBox[{"{", 
      RowBox[{"2", ",", "blues"}], "}"}], 
     RowBox[{",", "\[ThinSpace]"}], "c", 
     RowBox[{",", "\[ThinSpace]"}], 
     RowBox[{"-", "a"}], 
     RowBox[{",", "\[ThinSpace]"}], 
     RowBox[{"-", "d"}], 
     RowBox[{",", "\[ThinSpace]"}], 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "e"}], ",", 
       RowBox[{"-", "fred"}]}], "}"}], 
     RowBox[{",", "\[ThinSpace]"}], "d", 
     RowBox[{",", "\[ThinSpace]"}], 
     RowBox[{"{", 
      RowBox[{"e", ",", "fred"}], "}"}]}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[{
   1, -$CellContext`fred}, $CellContext`a, -$CellContext`b, {
   2, $CellContext`blues}, $CellContext`c, -$CellContext`a, -$CellContext`d, \
{-$CellContext`e, -$CellContext`fred}, $CellContext`d, {$CellContext`e, \
$CellContext`fred}],
  Editable->False]], "Output",
 CellChangeTimes->{3.728662913828676*^9, 3.728671266755533*^9, 
  3.728674894617045*^9, 3.728743307820076*^9, 3.728743593027919*^9, 
  3.728747677214147*^9, 3.728749570785131*^9, 3.72874994463627*^9, 
  3.728750798562683*^9, 3.728751070842411*^9, 3.728751632348804*^9, 
  3.7288132791450663`*^9, 3.728827396710535*^9, 3.7288359872319098`*^9, 
  3.728836303204278*^9, 3.728843128483032*^9, 3.728845638822239*^9, 
  3.733367410804085*^9, 3.7922733548683777`*^9},
 CellLabel->"Out[138]=",
 CellID->367931802,ExpressionUUID->"95bd047b-053c-4fcf-ac10-2ff84c0d45ed"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IndicesOf", "[", "InnerC", "]"}], "[", "expr", "]"}]], "Input",
 CellLabel->
  "In[139]:=",ExpressionUUID->"59e57e41-bcdb-4c5a-9aa8-72cd878ffcb4"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{
     RowBox[{"-", "B"}], 
     RowBox[{",", "\[ThinSpace]"}], 
     RowBox[{"{", 
      RowBox[{"A", ",", "fcomp"}], "}"}], 
     RowBox[{",", "\[ThinSpace]"}], 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "A"}], ",", 
       RowBox[{"-", "fcomp"}]}], "}"}], 
     RowBox[{",", "\[ThinSpace]"}], "B"}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[-$CellContext`B, {$CellContext`A, \
$CellContext`fcomp}, {-$CellContext`A, -$CellContext`fcomp}, $CellContext`B],
  Editable->False]], "Output",
 CellChangeTimes->{3.728662914593055*^9, 3.728671266905149*^9, 
  3.7286748948291893`*^9, 3.7287433079364643`*^9, 3.728743561278638*^9, 
  3.728747679534916*^9, 3.7287495710710783`*^9, 3.7287499447430773`*^9, 
  3.728750798729145*^9, 3.728751071169552*^9, 3.728751632542307*^9, 
  3.7288132792897577`*^9, 3.728827396902165*^9, 3.728835987413169*^9, 
  3.728836303304405*^9, 3.728843129123221*^9, 3.728845638990073*^9, 
  3.73336741112834*^9, 3.792273354926408*^9},
 CellLabel->"Out[139]=",
 CellID->1224376501,ExpressionUUID->"4c94c0b5-6bbd-4c47-a71e-2d47ccce7641"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IndicesOf", "[", 
   RowBox[{"U", ",", "Free"}], "]"}], "[", "expr", "]"}]], "Input",
 CellLabel->
  "In[140]:=",ExpressionUUID->"f28ab0e5-e513-4521-b6db-96ee845e594e"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{
     RowBox[{"-", "a"}], 
     RowBox[{",", "\[ThinSpace]"}], 
     RowBox[{"-", "b"}], 
     RowBox[{",", "\[ThinSpace]"}], "d", 
     RowBox[{",", "\[ThinSpace]"}], 
     RowBox[{"{", 
      RowBox[{"e", ",", "fred"}], "}"}]}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[-$CellContext`a, -$CellContext`b, $CellContext`d, \
{$CellContext`e, $CellContext`fred}],
  Editable->False]], "Output",
 CellChangeTimes->{3.7286629153060007`*^9, 3.728671267056593*^9, 
  3.7286748949666767`*^9, 3.728743308055807*^9, 3.72874356514459*^9, 
  3.728747681223322*^9, 3.7287495715528507`*^9, 3.7287499448858843`*^9, 
  3.728750798894746*^9, 3.728751071493341*^9, 3.728751632778537*^9, 
  3.7288132795339403`*^9, 3.728827397113237*^9, 3.72883598759137*^9, 
  3.728836303421398*^9, 3.728843129490836*^9, 3.72884563916651*^9, 
  3.73336741124319*^9, 3.7922733549728603`*^9},
 CellLabel->"Out[140]=",
 CellID->907546753,ExpressionUUID->"1a11710d-b6d0-43d4-b395-a359a7a706c1"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IndicesOf", "[", 
   RowBox[{"Free", ",", "U"}], "]"}], "[", "expr", "]"}]], "Input",
 CellLabel->
  "In[141]:=",ExpressionUUID->"6696e00e-8955-4351-bcf8-101b3ef78416"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{"-", "b"}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[-$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.728662916075876*^9, 3.728671267221127*^9, 
  3.7286748951207848`*^9, 3.7287433081892776`*^9, 3.7287435669116783`*^9, 
  3.728747682745392*^9, 3.728749572068906*^9, 3.728749945019732*^9, 
  3.728750799053091*^9, 3.728751071671332*^9, 3.72875163293152*^9, 
  3.728813279655222*^9, 3.7288273972670717`*^9, 3.7288359878207912`*^9, 
  3.7288363035866117`*^9, 3.7288431296943502`*^9, 3.728845639337706*^9, 
  3.733367411396675*^9, 3.792273355023271*^9},
 CellLabel->"Out[141]=",
 CellID->1709416677,ExpressionUUID->"2e00e1a5-bb06-4b0d-a32d-34cd52619aa7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IndicesOf", "[", 
   RowBox[{"{", 
    RowBox[{"U", ",", "Free"}], "}"}], "]"}], "[", "expr", "]"}]], "Input",
 CellLabel->
  "In[142]:=",ExpressionUUID->"571c8d35-a5c3-4086-8f79-78a9067155db"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{
     RowBox[{"-", "a"}], 
     RowBox[{",", "\[ThinSpace]"}], 
     RowBox[{"-", "b"}], 
     RowBox[{",", "\[ThinSpace]"}], "d", 
     RowBox[{",", "\[ThinSpace]"}], 
     RowBox[{"{", 
      RowBox[{"e", ",", "fred"}], "}"}], 
     RowBox[{",", "\[ThinSpace]"}], 
     RowBox[{"-", "b"}], 
     RowBox[{",", "\[ThinSpace]"}], "c"}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[-$CellContext`a, -$CellContext`b, $CellContext`d, \
{$CellContext`e, $CellContext`fred}, -$CellContext`b, $CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{3.728662916911298*^9, 3.72867126740421*^9, 
  3.728674895301733*^9, 3.728743308324449*^9, 3.728743568726738*^9, 
  3.728747683590425*^9, 3.7287495722686777`*^9, 3.7287499451521873`*^9, 
  3.728750799628355*^9, 3.7287510718454638`*^9, 3.728751633095399*^9, 
  3.728813279771282*^9, 3.728827397413208*^9, 3.7288359879944897`*^9, 
  3.728836303708001*^9, 3.7288431299179697`*^9, 3.728845639523444*^9, 
  3.733367411537825*^9, 3.792273355073821*^9},
 CellLabel->"Out[142]=",
 CellID->1129860773,ExpressionUUID->"3d57c3fe-bf7f-4b17-8abd-bba27b51330c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["xCoba`", "MR"],
 " adds more possibilities. We can now use the types ",
 StyleBox["BIndex", "MR"],
 " and ",
 StyleBox["CIndex", "MR"],
 " as selectors, or search for a particular basis"
}], "MathCaption",
 CellChangeTimes->{{3.728662928825243*^9, 
  3.728662936274993*^9}},ExpressionUUID->"8fd8c5cb-8915-4e3c-8568-\
8993147aff59"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IndicesOf", "[", 
   RowBox[{"{", 
    RowBox[{"BIndex", ",", "CIndex"}], "}"}], "]"}], "[", "expr", 
  "]"}]], "Input",
 CellLabel->
  "In[143]:=",ExpressionUUID->"5011d870-ae08-4b36-b633-1db8d8650a02"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "e"}], ",", 
       RowBox[{"-", "fred"}]}], "}"}], 
     RowBox[{",", "\[ThinSpace]"}], 
     RowBox[{"{", 
      RowBox[{"A", ",", "fcomp"}], "}"}], 
     RowBox[{",", "\[ThinSpace]"}], 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "A"}], ",", 
       RowBox[{"-", "fcomp"}]}], "}"}], 
     RowBox[{",", "\[ThinSpace]"}], 
     RowBox[{"{", 
      RowBox[{"e", ",", "fred"}], "}"}], 
     RowBox[{",", "\[ThinSpace]"}], 
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "fred"}]}], "}"}], 
     RowBox[{",", "\[ThinSpace]"}], 
     RowBox[{"{", 
      RowBox[{"2", ",", "blues"}], "}"}]}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[{-$CellContext`e, -$CellContext`fred}, \
{$CellContext`A, $CellContext`fcomp}, {-$CellContext`A, -$CellContext`fcomp}, \
{$CellContext`e, $CellContext`fred}, {1, -$CellContext`fred}, {
   2, $CellContext`blues}],
  Editable->False]], "Output",
 CellChangeTimes->{3.728662940575639*^9, 3.7286712675460443`*^9, 
  3.7286748954532957`*^9, 3.728743308498334*^9, 3.728743625990293*^9, 
  3.728747687315961*^9, 3.7287495726695642`*^9, 3.728749945265636*^9, 
  3.728750799899115*^9, 3.728751072013143*^9, 3.728751633249147*^9, 
  3.7288132798893843`*^9, 3.728827397563284*^9, 3.728835988183255*^9, 
  3.7288363039104433`*^9, 3.728843130098372*^9, 3.728845639729144*^9, 
  3.7333674116890383`*^9, 3.7922733551230516`*^9},
 CellLabel->"Out[143]=",
 CellID->1519605037,ExpressionUUID->"abbd73f1-d10a-435d-a708-cbeaa6fc6875"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IndicesOf", "[", "fred", "]"}], "[", "expr", "]"}]], "Input",
 CellChangeTimes->{{3.728743629407703*^9, 3.728743629752121*^9}, {
  3.728747690286623*^9, 3.728747690791609*^9}},
 CellLabel->
  "In[144]:=",ExpressionUUID->"fe18a01b-35ef-4ad2-b5e0-e0ca9eb665c1"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "fred"}]}], "}"}], 
     RowBox[{",", "\[ThinSpace]"}], 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "e"}], ",", 
       RowBox[{"-", "fred"}]}], "}"}], 
     RowBox[{",", "\[ThinSpace]"}], 
     RowBox[{"{", 
      RowBox[{"e", ",", "fred"}], "}"}]}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[{
   1, -$CellContext`fred}, {-$CellContext`e, -$CellContext`fred}, \
{$CellContext`e, $CellContext`fred}],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.728662941563719*^9, 3.7286712676785994`*^9, 3.728674895599922*^9, 
   3.728743308706205*^9, {3.728743627191276*^9, 3.728743630232864*^9}, 
   3.7287476912151833`*^9, 3.728749573069787*^9, 3.7287499454021*^9, 
   3.728750800310749*^9, 3.728751072176577*^9, 3.7287516333997717`*^9, 
   3.7288132800208473`*^9, 3.728827397718237*^9, 3.7288359883308287`*^9, 
   3.728836304037706*^9, 3.728843130373419*^9, 3.728845639914359*^9, 
   3.733367411861854*^9, 3.792273355173645*^9},
 CellLabel->"Out[144]=",
 CellID->2126326735,ExpressionUUID->"9e8a4101-d72e-4af7-9de9-7de849439cb2"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Consider now the following example", "MathCaption",ExpressionUUID->"02557eb9-26fd-4a8a-b9f5-6532564d2409"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IndicesOf", "[", 
   RowBox[{"Basis", ",", "fred"}], "]"}], "[", "expr", "]"}]], "Input",
 CellChangeTimes->{{3.72874363329669*^9, 3.728743633591782*^9}, {
  3.728747693752948*^9, 3.7287476942235394`*^9}},
 CellLabel->
  "In[145]:=",ExpressionUUID->"dc0547a7-f11c-4b82-922a-ee1c47f03739"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "fred"}]}], "}"}], 
     RowBox[{",", "\[ThinSpace]"}], 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "e"}], ",", 
       RowBox[{"-", "fred"}]}], "}"}]}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[{
   1, -$CellContext`fred}, {-$CellContext`e, -$CellContext`fred}],
  Editable->False]], "Output",
 CellChangeTimes->{3.728662944481868*^9, 3.7286712678132353`*^9, 
  3.728674895754097*^9, 3.7287433089043617`*^9, 3.728743634169718*^9, 
  3.72874769455389*^9, 3.728749573331471*^9, 3.728749945536181*^9, 
  3.728750800606406*^9, 3.728751072346672*^9, 3.728751633548877*^9, 
  3.728813280148264*^9, 3.7288273978946257`*^9, 3.728835988483926*^9, 
  3.728836304172531*^9, 3.7288431305621758`*^9, 3.728845640079959*^9, 
  3.733367412038066*^9, 3.7922733552346687`*^9},
 CellLabel->"Out[145]=",
 CellID->898065418,ExpressionUUID->"06c35b40-d3f0-45f3-98f1-d04cd820da48"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IndicesOf", "[", 
   RowBox[{"Basis", "[", "fred", "]"}], "]"}], "[", "expr", "]"}]], "Input",
 CellChangeTimes->{{3.728743635664445*^9, 3.7287436359753857`*^9}, {
  3.728747696577207*^9, 3.7287476971993856`*^9}},
 CellLabel->
  "In[146]:=",ExpressionUUID->"277857e2-9a66-4a12-8f05-5ddea2f68d9d"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{"a", 
     RowBox[{",", "\[ThinSpace]"}], "c", 
     RowBox[{",", "\[ThinSpace]"}], 
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "fred"}]}], "}"}], 
     RowBox[{",", "\[ThinSpace]"}], 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "e"}], ",", 
       RowBox[{"-", "fred"}]}], "}"}]}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[$CellContext`a, $CellContext`c, {
   1, -$CellContext`fred}, {-$CellContext`e, -$CellContext`fred}],
  Editable->False]], "Output",
 CellChangeTimes->{3.728662945861377*^9, 3.7286712679470377`*^9, 
  3.728674895902524*^9, 3.728743309020855*^9, 3.728743636666313*^9, 
  3.7287476979181967`*^9, 3.728749573519548*^9, 3.728749945670128*^9, 
  3.728750800744646*^9, 3.72875107247948*^9, 3.728751633683181*^9, 
  3.728813280288496*^9, 3.7288273980456953`*^9, 3.7288359886347027`*^9, 
  3.728836304333148*^9, 3.728843130737877*^9, 3.728845640451013*^9, 
  3.73336741220759*^9, 3.792273355282111*^9},
 CellLabel->"Out[146]=",
 CellID->1270409210,ExpressionUUID->"0b50deb8-c36b-4b3b-85b3-7ad2f4a5934c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The first selects all the indices appearing in Basis objects and then \
selects from those only the ones that belong to polar. The second selects \
only those indices that belong to a Basis object of the polar basis, but \
returns the pair, not only the ",
 StyleBox["BCIndex", "MR"],
 "."
}], "MathCaption",ExpressionUUID->"8416f90a-b65f-49bd-ba1a-fa0b3e1edeaf"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IndicesOf", "[", 
   RowBox[{"Basis", "[", "fred", "]"}], "]"}], "[", 
  RowBox[{"Basis", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "fred"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{"b", ",", "blues"}], "}"}]}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.728743642952241*^9, 3.728743654126987*^9}, {
   3.728747700800893*^9, 3.728747707014974*^9}, 3.7287493083194723`*^9},
 CellLabel->
  "In[147]:=",ExpressionUUID->"018e1f6f-3e95-40bc-ade0-65018c63ae4b"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "fred"}]}], "}"}], 
     RowBox[{",", "\[ThinSpace]"}], 
     RowBox[{"{", 
      RowBox[{"b", ",", "blues"}], "}"}]}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[{-$CellContext`a, -$CellContext`fred}, \
{$CellContext`b, $CellContext`blues}],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.728662950030306*^9, 3.7286712680957212`*^9, 3.728674896053019*^9, 
   3.728743309156048*^9, {3.728743646068892*^9, 3.728743654842252*^9}, 
   3.728747707552477*^9, 3.728749573662181*^9, 3.7287499458040752`*^9, 
   3.728750800910693*^9, 3.728751072630424*^9, 3.7287516338336773`*^9, 
   3.728813280421391*^9, 3.728827398203546*^9, 3.728835988792111*^9, 
   3.728836304473757*^9, 3.728843130952777*^9, 3.7288456408305187`*^9, 
   3.733367412355309*^9, 3.792273355342181*^9},
 CellLabel->"Out[147]=",
 CellID->1485173996,ExpressionUUID->"a40e4244-2a8e-4398-8b7d-bd2eaaebb36b"]
}, Open  ]]
}, Open  ]],

Cell["\<\
This function will be used heavily to control the behaviour of ContractBasis \
and SeparateBasis.\
\>", "MathCaption",ExpressionUUID->"3a797d89-c1c6-4c23-b19b-e006f338f34a"]
}, Closed]],

Cell[CellGroupData[{

Cell["1.2.2. ContractBasis", "Subsubsection",
 CellChangeTimes->{{3.728673448516897*^9, 
  3.728673448668441*^9}},ExpressionUUID->"dd869560-b307-4804-b0a3-\
2322cee31c1d"],

Cell[TextData[{
 StyleBox["ContractBasis", "MR"],
 " is responsible for all contractions of ",
 StyleBox["Basis", "MR"],
 " with any other object. "
}], "Text",ExpressionUUID->"b51bd2f9-84ff-4d6d-8263-772889b99dba"],

Cell[TextData[{
 StyleBox["ContractBasis[expr, indices]\t\t", "DisplayMath"],
 StyleBox["Perform the contractions of ", "Text"],
 StyleBox["Basis", "Text", "MR"],
 StyleBox[" objects involving the specified dummy indices", "Text"]
}], "DefinitionBox",ExpressionUUID->"a175050a-6be7-4d1d-be7f-847bee268215"],

Cell["Contraction of Basis objects with tensors.", "Text",ExpressionUUID->"293181a8-1fe7-4235-90ab-19e6b5ae9c1e"],

Cell[TextData[{
 "We have complete control over which dummy indices should be eliminated. We \
can select them in several ways, all of which are eventually converted into a \
list of indices with head ",
 StyleBox["IndexList", "MR"],
 ":\n\t- ",
 StyleBox["IndexList[...]", "MR"],
 ",\t\t\nbe careful not to use the normal ",
 StyleBox["Mathematica ",
  FontSlant->"Italic"],
 StyleBox["List", "MR"],
 ", which could be mistaken for a ",
 StyleBox["BCIndex", "MR"],
 ". ",
 StyleBox["IndicesOf[selectors]", "MR"],
 " as a second argument is treated as ",
 StyleBox["IndicesOf[selectors][expr]", "MR"],
 " and gives great flexibility, but it is sometimes too long for simple \
cases. To handle them, several shorthands are defined: \n\t- ",
 StyleBox["ContractBasis[expr,basis] \t= \
ContractBasis[expr,IndicesOf[basis]]", "MR"],
 ",\n\t- ",
 StyleBox["ContractBasis[expr,vbundle]\t= \
ContractBasis[expr,IndicesOf[vbundle]]", "MR"],
 ",\n\t- ",
 StyleBox["ContractBasis[expr,tensor]\t= \
ContractBasis[expr,IndicesOf[tensor]]", "MR"],
 ",\n\t- ",
 StyleBox["ContractBasis[expr,index]\t= ContractBasis[expr,IndexList[index]]",
   "MR"],
 " .\nIt is important to keep in mind that the second argument gives the list \
of indices to be contracted, not their pairs in the Basis object. \nAll of \
this is best illustrated through some examples."
}], "Text",
 CellChangeTimes->{{3.7286629733615017`*^9, 
  3.728662989425746*^9}},ExpressionUUID->"f27ff5fb-0bdc-4e46-b3a9-\
8c578d8f3cda"],

Cell[CellGroupData[{

Cell["Take again example expressions:", "MathCaption",
 CellChangeTimes->{{3.72866299287225*^9, 3.728663009839904*^9}},
 CellID->1402959324,ExpressionUUID->"08360204-917c-4a99-af4d-cb1071d41802"],

Cell[CellGroupData[{

Cell[BoxData["expr"], "Input",
 CellLabel->
  "In[148]:=",ExpressionUUID->"1165e8c9-df89-4549-85d6-a51c3503f62c"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"TT", 
        StyleBox[GridBox[{
           {" ", " ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], "c", " "},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], "b", " ", " ", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TT[{1, -$CellContext`fred}, -$CellContext`b, {
     2, $CellContext`blues}, $CellContext`c, -$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {2, $CellContext`blues}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "c"},
           {
            StyleBox["e",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`e, -$CellContext`fred}, $CellContext`c],
    
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["A",
             FontColor->RGBColor[0., 1., 0.]]},
           {"B", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[-$CellContext`B, {$CellContext`A, $CellContext`fcomp}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", "B"},
           {
            StyleBox["A",
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[{-$CellContext`A, -$CellContext`fcomp}, $CellContext`B],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", "d", 
            StyleBox["e",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", "b", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, -$CellContext`b, $CellContext`d, \
{$CellContext`e, $CellContext`fred}],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728663011165794*^9, 3.7286712717804623`*^9, 
  3.728674896215891*^9, 3.72874330932238*^9, 3.728743677514431*^9, 
  3.728747718455544*^9, 3.728749578624978*^9, 3.7287499459502163`*^9, 
  3.728750804701466*^9, 3.7287510792015457`*^9, 3.728751633988937*^9, 
  3.728813280541718*^9, 3.728827398360956*^9, 3.728835988937056*^9, 
  3.7288363046334057`*^9, 3.728843135552659*^9, 3.728845641007641*^9, 
  3.733367418455378*^9, 3.7922733697722692`*^9},
 CellLabel->"Out[148]=",
 CellID->1639099083,ExpressionUUID->"d4801b87-9bd0-40c0-91ca-1e0023a879b4"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr2", "=", 
  RowBox[{
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "e"}], ",", 
       RowBox[{"-", "blues"}]}], "}"}], ",", "b"}], "]"}], 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", "fcomp"}], "}"}]}], "]"}], " ", 
   RowBox[{"S", "[", 
    RowBox[{
     RowBox[{"-", "B"}], ",", "A"}], "]"}], 
   RowBox[{"v", "[", 
    RowBox[{"{", 
     RowBox[{"e", ",", "blues"}], "}"}], "]"}]}]}]], "Input",
 CellChangeTimes->{{3.7287436818228903`*^9, 3.728743686853614*^9}, {
  3.7287477223842*^9, 3.728747728861279*^9}, {3.7287493121841*^9, 
  3.728749312876054*^9}},
 CellLabel->
  "In[149]:=",ExpressionUUID->"decbdad5-9c12-40a9-98d0-ed0418969392"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["1",
            FontColor->RGBColor[0., 1., 0.]]},
          {"A", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[-$CellContext`A, {1, $CellContext`fcomp}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", "b"},
          {
           StyleBox["e",
            FontColor->RGBColor[0., 0., 1.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{-$CellContext`e, -$CellContext`blues}, $CellContext`b],
   
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {" ", "A"},
          {"B", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[-$CellContext`B, $CellContext`A],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {
           StyleBox["e",
            FontColor->RGBColor[0., 0., 1.]]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{$CellContext`e, $CellContext`blues}],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728663011764852*^9, 3.728671271911983*^9, 
  3.7286748964178877`*^9, 3.72874330948427*^9, 3.728743687485941*^9, 
  3.7287477297563143`*^9, 3.728749578760478*^9, 3.7287499460729513`*^9, 
  3.7287508048732*^9, 3.7287510793498383`*^9, 3.728751634134891*^9, 
  3.7288132807003403`*^9, 3.72882739862958*^9, 3.728835989103211*^9, 
  3.728836304779057*^9, 3.728843135750628*^9, 3.728845641193734*^9, 
  3.733367418601719*^9, 3.792273370611444*^9},
 CellLabel->"Out[149]=",
 CellID->632927560,ExpressionUUID->"38100f1d-fe38-4c4b-8933-453d4db65395"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The function called with just one argument performs all the possible \
contractions:\
\>", "MathCaption",
 CellChangeTimes->{{3.728663020086968*^9, 
  3.7286630207992764`*^9}},ExpressionUUID->"5e0ded7b-692e-4726-a6b0-\
de8c852fadc3"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ContractBasis", "[", "expr", "]"}]], "Input",
 CellLabel->
  "In[150]:=",ExpressionUUID->"beec4951-1537-41a4-83b8-a728fc481f7e"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"TT", 
       StyleBox[GridBox[{
          {" ", " ", 
           StyleBox["2",
            FontColor->RGBColor[0., 0., 1.]], "c", " "},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], "b", " ", " ", 
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`TT[{1, -$CellContext`fred}, -$CellContext`b, {
    2, $CellContext`blues}, $CellContext`c, {1, -$CellContext`fred}],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["A",
             FontColor->RGBColor[0., 1., 0.]]},
           {"B", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[-$CellContext`B, {$CellContext`A, $CellContext`fcomp}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", "B"},
           {
            StyleBox["A",
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[{-$CellContext`A, -$CellContext`fcomp}, $CellContext`B],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], "c"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], "b", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[{1, -$CellContext`fred}, -$CellContext`b, {
     2, $CellContext`blues}, $CellContext`c],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728663014720232*^9, 3.728671272065366*^9, 
  3.728674896617235*^9, 3.728743309635664*^9, 3.728743690316615*^9, 
  3.728747731219652*^9, 3.728749578888653*^9, 3.728749946205847*^9, 
  3.72875080504277*^9, 3.728751079516765*^9, 3.72875163427899*^9, 
  3.728813280933338*^9, 3.728827398878858*^9, 3.728835989252557*^9, 
  3.7288363049492817`*^9, 3.728843135893608*^9, 3.7288456413746223`*^9, 
  3.733367418738909*^9, 3.792273371642367*^9},
 CellLabel->"Out[150]=",
 CellID->1423645728,ExpressionUUID->"20d4c8fe-a444-4a40-8384-9188bad474b1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
If we choose to give the indices directly, we can specifiy either element of \
the pair:\
\>", "MathCaption",
 CellChangeTimes->{
  3.728663022687049*^9},ExpressionUUID->"599a2e5f-bd4a-4bd9-8bcd-\
9dd0c1cc6739"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"ContractBasis", "[", 
       RowBox[{"expr2", ",", " ", "A"}], "]"}], ",", " ", 
      RowBox[{"ContractBasis", "[", 
       RowBox[{"expr2", ",", 
        RowBox[{"-", "A"}]}], "]"}]}], "}"}], ",", "\[IndentingNewLine]", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"ContractBasis", "[", 
       RowBox[{"expr2", ",", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{"-", "e"}], ",", 
          RowBox[{"-", "blues"}]}], "}"}]}], "]"}], ",", " ", 
      RowBox[{"ContractBasis", "[", 
       RowBox[{"expr2", ",", 
        RowBox[{"{", 
         RowBox[{"e", ",", "blues"}], "}"}]}], "]"}]}], "}"}]}], "}"}], "//", 
  "TableForm"}]], "Input",
 CellChangeTimes->{{3.728743698140293*^9, 3.728743702236883*^9}, {
  3.728747734966063*^9, 3.728747738388536*^9}, {3.72874931378502*^9, 
  3.728749314411675*^9}},
 CellLabel->
  "In[151]:=",ExpressionUUID->"b34949bf-5c43-4c79-b7fa-befbb412bf18"],

Cell[BoxData[
 TagBox[GridBox[{
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", "b"},
              {
               StyleBox["e",
                FontColor->RGBColor[0., 0., 1.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{-$CellContext`e, -$CellContext`blues}, \
$CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"S", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]},
              {"B", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`S[-$CellContext`B, {1, $CellContext`fcomp}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["e",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{$CellContext`e, $CellContext`blues}],
       Editable->False]}], 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", "b"},
              {
               StyleBox["e",
                FontColor->RGBColor[0., 0., 1.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{-$CellContext`e, -$CellContext`blues}, \
$CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"S", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]},
              {"B", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`S[-$CellContext`B, {1, $CellContext`fcomp}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["e",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{$CellContext`e, $CellContext`blues}],
       Editable->False]}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]},
              {"A", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[-$CellContext`A, {1, $CellContext`fcomp}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"S", 
           StyleBox[GridBox[{
              {" ", "A"},
              {"B", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`S[-$CellContext`B, $CellContext`A],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"b"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[$CellContext`b],
       Editable->False]}], 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]},
              {"A", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[-$CellContext`A, {1, $CellContext`fcomp}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"S", 
           StyleBox[GridBox[{
              {" ", "A"},
              {"B", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`S[-$CellContext`B, $CellContext`A],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"b"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[$CellContext`b],
       Editable->False]}]}
   },
   GridBoxAlignment->{"Columns" -> {{Left}}, "Rows" -> {{Baseline}}},
   GridBoxSpacings->{"Columns" -> {
       Offset[0.27999999999999997`], {
        Offset[2.0999999999999996`]}, 
       Offset[0.27999999999999997`]}, "Rows" -> {
       Offset[0.2], {
        Offset[0.4]}, 
       Offset[0.2]}}],
  Function[BoxForm`e$, 
   TableForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{3.7286630162568293`*^9, 3.7286712721918716`*^9, 
  3.728674896778658*^9, 3.728743309757697*^9, 3.728743703584848*^9, 
  3.728747739002524*^9, 3.728749579017353*^9, 3.728749946345387*^9, 
  3.728750805177804*^9, 3.728751079667246*^9, 3.728751634442272*^9, 
  3.7288132811508827`*^9, 3.7288273990418253`*^9, 3.7288359894122257`*^9, 
  3.728836305153501*^9, 3.728843136505499*^9, 3.728845641566951*^9, 
  3.733367418866384*^9, 3.792273372680544*^9},
 CellLabel->"Out[151]//TableForm=",
 CellID->461810322,ExpressionUUID->"66168d59-6dd8-4555-8956-a12873168123"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Nothing happens if we give an invalid index:", "MathCaption",
 CellChangeTimes->{
  3.7286630257186127`*^9},ExpressionUUID->"68e86b2c-1ddd-43cf-ac86-\
bc034db2766b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ContractBasis", "[", 
  RowBox[{"expr2", ",", "C"}], "]"}]], "Input",
 CellLabel->
  "In[152]:=",ExpressionUUID->"bacd5051-b924-46f6-9450-99424ec88845"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["1",
            FontColor->RGBColor[0., 1., 0.]]},
          {"A", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[-$CellContext`A, {1, $CellContext`fcomp}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", "b"},
          {
           StyleBox["e",
            FontColor->RGBColor[0., 0., 1.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{-$CellContext`e, -$CellContext`blues}, $CellContext`b],
   
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {" ", "A"},
          {"B", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[-$CellContext`B, $CellContext`A],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {
           StyleBox["e",
            FontColor->RGBColor[0., 0., 1.]]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{$CellContext`e, $CellContext`blues}],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7286630176183662`*^9, 3.728671272327167*^9, 
  3.7286748969234657`*^9, 3.728743309952489*^9, 3.7287437127167997`*^9, 
  3.728747742739894*^9, 3.728749579434037*^9, 3.7287499465011044`*^9, 
  3.7287508056167*^9, 3.728751079790455*^9, 3.728751634596306*^9, 
  3.728813281284278*^9, 3.728827399190762*^9, 3.7288359895526867`*^9, 
  3.728836305296734*^9, 3.728843136701837*^9, 3.7288456417746964`*^9, 
  3.733367419057459*^9, 3.7922733744917803`*^9},
 CellLabel->"Out[152]=",
 CellID->1960327329,ExpressionUUID->"81f251a8-e610-4c77-a851-255407ceed5b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["More simple examples:", "MathCaption",
 CellChangeTimes->{
  3.72866302884656*^9},ExpressionUUID->"2c3c5f61-9573-4923-a647-fac8d8887e51"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ContractBasis", "[", 
  RowBox[{"expr", ",", 
   RowBox[{"IndicesOf", "[", "AIndex", "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[153]:=",ExpressionUUID->"b09fb82c-77e5-447b-ba60-68ae1b883283"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"TT", 
       StyleBox[GridBox[{
          {" ", " ", 
           StyleBox["2",
            FontColor->RGBColor[0., 0., 1.]], "c", " "},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], "b", " ", " ", 
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`TT[{1, -$CellContext`fred}, -$CellContext`b, {
    2, $CellContext`blues}, $CellContext`c, {1, -$CellContext`fred}],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "c"},
           {
            StyleBox["e",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`e, -$CellContext`fred}, $CellContext`c],
    
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["A",
             FontColor->RGBColor[0., 1., 0.]]},
           {"B", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[-$CellContext`B, {$CellContext`A, $CellContext`fcomp}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", "B"},
           {
            StyleBox["A",
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[{-$CellContext`A, -$CellContext`fcomp}, $CellContext`B],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["e",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], "b", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[{1, -$CellContext`fred}, -$CellContext`b, {
     2, $CellContext`blues}, {$CellContext`e, $CellContext`fred}],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728663037470421*^9, 3.72867127245844*^9, 
  3.7286748970527763`*^9, 3.728743310103649*^9, 3.7287437157167263`*^9, 
  3.728747744524109*^9, 3.728749579722011*^9, 3.7287499466501427`*^9, 
  3.728750805889895*^9, 3.728751079906003*^9, 3.7287516347274647`*^9, 
  3.72881328144967*^9, 3.728827399325493*^9, 3.728835989737007*^9, 
  3.7288363055551167`*^9, 3.728843136935309*^9, 3.728845642113414*^9, 
  3.7333674194750557`*^9, 3.792273376288932*^9},
 CellLabel->"Out[153]=",
 CellID->936333049,ExpressionUUID->"86303904-5c21-4f9d-bfb7-6b089adfd258"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ContractBasis", "[", 
  RowBox[{"expr", ",", 
   RowBox[{"IndicesOf", "[", 
    RowBox[{"TT", ",", "U"}], "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[154]:=",ExpressionUUID->"99d0a71c-0244-4532-9286-09bccdf841fc"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"TT", 
       StyleBox[GridBox[{
          {" ", " ", 
           StyleBox["2",
            FontColor->RGBColor[0., 0., 1.]], "c", " "},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], "b", " ", " ", 
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`TT[{1, -$CellContext`fred}, -$CellContext`b, {
    2, $CellContext`blues}, $CellContext`c, {1, -$CellContext`fred}],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {2, $CellContext`blues}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "c"},
           {
            StyleBox["e",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`e, -$CellContext`fred}, $CellContext`c],
    
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["A",
             FontColor->RGBColor[0., 1., 0.]]},
           {"B", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[-$CellContext`B, {$CellContext`A, $CellContext`fcomp}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", "B"},
           {
            StyleBox["A",
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[{-$CellContext`A, -$CellContext`fcomp}, $CellContext`B],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", "d", 
            StyleBox["e",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], "b", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[{
     1, -$CellContext`fred}, -$CellContext`b, $CellContext`d, \
{$CellContext`e, $CellContext`fred}],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728663037699325*^9, 3.728671272585891*^9, 
  3.728674897210112*^9, 3.728743310246863*^9, 3.7287437175356817`*^9, 
  3.728747745288149*^9, 3.7287495803859777`*^9, 3.7287499468000174`*^9, 
  3.72875080608013*^9, 3.728751080051708*^9, 3.728751634879304*^9, 
  3.7288132815794067`*^9, 3.7288273994760733`*^9, 3.728835989896276*^9, 
  3.728836305699483*^9, 3.728843137127388*^9, 3.7288456422738657`*^9, 
  3.7333674196359577`*^9, 3.792273376825521*^9},
 CellLabel->"Out[154]=",
 CellID->1746856133,ExpressionUUID->"84af7c29-383b-4271-8029-f214c3813ee5"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ContractBasis", "[", 
  RowBox[{"expr", ",", 
   RowBox[{"IndicesOf", "[", 
    RowBox[{"{", 
     RowBox[{"TT", ",", "U"}], "}"}], "]"}]}], "]"}]], "Input",
 CellChangeTimes->{3.728747755444659*^9},
 CellLabel->
  "In[155]:=",ExpressionUUID->"a779f61a-671e-4dee-93fb-92ea30e20d03"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"TT", 
       StyleBox[GridBox[{
          {" ", " ", 
           StyleBox["2",
            FontColor->RGBColor[0., 0., 1.]], "c", " "},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], "b", " ", " ", 
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`TT[{1, -$CellContext`fred}, -$CellContext`b, {
    2, $CellContext`blues}, $CellContext`c, {1, -$CellContext`fred}],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["A",
             FontColor->RGBColor[0., 1., 0.]]},
           {"B", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[-$CellContext`B, {$CellContext`A, $CellContext`fcomp}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", "B"},
           {
            StyleBox["A",
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[{-$CellContext`A, -$CellContext`fcomp}, $CellContext`B],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], "c"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], "b", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[{1, -$CellContext`fred}, -$CellContext`b, {
     2, $CellContext`blues}, $CellContext`c],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728663037841983*^9, 3.728671272725999*^9, 
  3.72867489738348*^9, 3.7287433103788137`*^9, 3.7287437188805304`*^9, 
  3.728747745983366*^9, 3.728749580586484*^9, 3.72874994692204*^9, 
  3.728750806236944*^9, 3.728751080173587*^9, 3.7287516350181627`*^9, 
  3.728813281743771*^9, 3.728827399628521*^9, 3.728835990054654*^9, 
  3.728836305884104*^9, 3.728843137858159*^9, 3.728845642452681*^9, 
  3.733367419973032*^9, 3.792273377843425*^9},
 CellLabel->"Out[155]=",
 CellID->289727438,ExpressionUUID->"b4b29332-f625-4201-87cc-d6e56764fcc8"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ContractBasis", "[", 
  RowBox[{"expr2", ",", "InnerC"}], "]"}]], "Input",
 CellLabel->
  "In[156]:=",ExpressionUUID->"cab229d0-47cc-4953-8742-1d727eb4a378"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", "b"},
          {
           StyleBox["e",
            FontColor->RGBColor[0., 0., 1.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{-$CellContext`e, -$CellContext`blues}, $CellContext`b],
   
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["1",
            FontColor->RGBColor[0., 1., 0.]]},
          {"B", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[-$CellContext`B, {1, $CellContext`fcomp}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {
           StyleBox["e",
            FontColor->RGBColor[0., 0., 1.]]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{$CellContext`e, $CellContext`blues}],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728663037992858*^9, 3.7286712728591213`*^9, 
  3.728674897590745*^9, 3.728743310479224*^9, 3.728743720265397*^9, 
  3.728747746600493*^9, 3.728749580887897*^9, 3.728749947065178*^9, 
  3.728750806402104*^9, 3.728751080305382*^9, 3.7287516351846313`*^9, 
  3.728813281917762*^9, 3.72882739974617*^9, 3.728835990197287*^9, 
  3.728836306050655*^9, 3.7288431380736923`*^9, 3.7288456426144333`*^9, 
  3.7333674200910254`*^9, 3.792273378327318*^9},
 CellLabel->"Out[156]=",
 CellID->1601519298,ExpressionUUID->"de69afdd-43a9-4e89-af8e-32e0f9a11fc9"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ContractBasis", "[", 
  RowBox[{"expr", ",", 
   RowBox[{"IndexList", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "d"}]}], "]"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728747758588607*^9, 3.7287477637724743`*^9}},
 CellLabel->
  "In[157]:=",ExpressionUUID->"0c58b5b6-43b9-49d8-a8b3-dff1eba80588"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"TT", 
       StyleBox[GridBox[{
          {" ", " ", 
           StyleBox["2",
            FontColor->RGBColor[0., 0., 1.]], "c", " "},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], "b", " ", " ", 
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`TT[{1, -$CellContext`fred}, -$CellContext`b, {
    2, $CellContext`blues}, $CellContext`c, {1, -$CellContext`fred}],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "c"},
           {
            StyleBox["e",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`e, -$CellContext`fred}, $CellContext`c],
    
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["A",
             FontColor->RGBColor[0., 1., 0.]]},
           {"B", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[-$CellContext`B, {$CellContext`A, $CellContext`fcomp}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", "B"},
           {
            StyleBox["A",
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[{-$CellContext`A, -$CellContext`fcomp}, $CellContext`B],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["e",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], "b", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[{1, -$CellContext`fred}, -$CellContext`b, {
     2, $CellContext`blues}, {$CellContext`e, $CellContext`fred}],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728663038151451*^9, 3.728671272957246*^9, 
  3.72867489773435*^9, 3.7287433105825853`*^9, 3.7287437236136427`*^9, 
  3.728747747632903*^9, 3.7287495810843983`*^9, 3.728749947187538*^9, 
  3.728750806563682*^9, 3.728751080462019*^9, 3.728751635345593*^9, 
  3.7288132821167717`*^9, 3.728827399891308*^9, 3.72883599035998*^9, 
  3.728836306197598*^9, 3.728843138283201*^9, 3.728845642796091*^9, 
  3.733367420241695*^9, 3.792273379329015*^9},
 CellLabel->"Out[157]=",
 CellID->123988376,ExpressionUUID->"5436f05b-e70d-41e7-843f-f253ca6bf695"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ContractBasis", "[", 
  RowBox[{"expr", ",", "fred"}], "]"}]], "Input",
 CellChangeTimes->{{3.7287437289625807`*^9, 3.728743729563363*^9}, {
  3.728747748925027*^9, 3.728747767004155*^9}},
 CellLabel->
  "In[158]:=",ExpressionUUID->"cdc7ec7d-f32f-4eba-87ea-2b6961f5e2f6"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"TT", 
        StyleBox[GridBox[{
           {" ", " ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], "c", " "},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], "b", " ", " ", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TT[{1, -$CellContext`fred}, -$CellContext`b, {
     2, $CellContext`blues}, $CellContext`c, -$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {2, $CellContext`blues}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["A",
             FontColor->RGBColor[0., 1., 0.]]},
           {"B", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[-$CellContext`B, {$CellContext`A, $CellContext`fcomp}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", "B"},
           {
            StyleBox["A",
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[{-$CellContext`A, -$CellContext`fcomp}, $CellContext`B],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", "d", "c"},
           {"a", "b", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, -$CellContext`b, $CellContext`d, \
$CellContext`c],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728663038303135*^9, 3.7286712731156797`*^9, 
  3.728674897884893*^9, 3.728743310714995*^9, 3.7287437311048393`*^9, 
  3.7287477507367983`*^9, 3.728749581265134*^9, 3.7287499473328876`*^9, 
  3.7287508067169456`*^9, 3.728751080664998*^9, 3.728751635478985*^9, 
  3.728813282339802*^9, 3.728827400146914*^9, 3.7288359905085897`*^9, 
  3.728836306352727*^9, 3.728843138488131*^9, 3.7288456429723797`*^9, 
  3.733367420445752*^9, 3.7922733797935667`*^9},
 CellLabel->"Out[158]=",
 CellID->1187002912,ExpressionUUID->"b3de59a9-ebab-4297-a0b8-01fe1a58ebaf"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Notice how only the index",
 StyleBox[" {-e, -bas}", "MR"],
 " has been supressed. Nothing has been done to ",
 StyleBox["a", "MR"],
 ", even though its pair in the Basis object belongs to",
 StyleBox[" polar", "MR"],
 ".  If we want to act on that index, we can use the following to contract \
everything that appears in a Basis object that has at least one polar index:"
}], "MathCaption",
 CellChangeTimes->{{3.7286630442301197`*^9, 3.7286630664931383`*^9}, {
  3.7287437382501793`*^9, 
  3.728743738769786*^9}},ExpressionUUID->"48d0d17c-29ba-4a7d-979a-\
86078e7b6c5d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ContractBasis", "[", 
  RowBox[{"expr", ",", " ", 
   RowBox[{"IndicesOf", "[", 
    RowBox[{"Basis", "[", "fred", "]"}], "]"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.72874374257813*^9, 3.728743742946794*^9}, {
  3.728747772221841*^9, 3.7287477727396317`*^9}},
 CellLabel->
  "In[159]:=",ExpressionUUID->"c82642e5-1e49-4451-bfe6-e5a36e43835f"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"TT", 
       StyleBox[GridBox[{
          {" ", " ", 
           StyleBox["2",
            FontColor->RGBColor[0., 0., 1.]], "c", " "},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], "b", " ", " ", 
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`TT[{1, -$CellContext`fred}, -$CellContext`b, {
    2, $CellContext`blues}, $CellContext`c, {1, -$CellContext`fred}],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {2, $CellContext`blues}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["A",
             FontColor->RGBColor[0., 1., 0.]]},
           {"B", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[-$CellContext`B, {$CellContext`A, $CellContext`fcomp}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", "B"},
           {
            StyleBox["A",
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[{-$CellContext`A, -$CellContext`fcomp}, $CellContext`B],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", "d", "c"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], "b", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[{
     1, -$CellContext`fred}, -$CellContext`b, $CellContext`d, $CellContext`c],
    
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728663045440043*^9, 3.728671273328673*^9, 
  3.72867489802955*^9, 3.728743310852707*^9, 3.728743743732864*^9, 
  3.728747773193441*^9, 3.728749581868882*^9, 3.728749947448266*^9, 
  3.728750806886918*^9, 3.728751080805984*^9, 3.728751635617979*^9, 
  3.728813282567163*^9, 3.728827400348795*^9, 3.728835990682992*^9, 
  3.7288363065177317`*^9, 3.7288431391522493`*^9, 3.7288456431635447`*^9, 
  3.733367420588579*^9, 3.792273381213282*^9},
 CellLabel->"Out[159]=",
 CellID->1585944226,ExpressionUUID->"aace69ec-f9c7-456c-8ca8-c05f165d18ef"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Or this to contract only abstract indices whose pair in ",
 StyleBox["Basis", "MR"],
 " belongs to ",
 StyleBox["bas", "MR"],
 ":"
}], "MathCaption",
 CellChangeTimes->{{3.728663056942094*^9, 3.728663087236*^9}, {
  3.728743751442189*^9, 
  3.72874375206485*^9}},ExpressionUUID->"569c28bb-8349-40a3-b0d1-\
b6b865edb553"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ContractBasis", "[", 
  RowBox[{"expr", ",", " ", 
   RowBox[{"IndicesOf", "[", 
    RowBox[{
     RowBox[{"Basis", "[", "fred", "]"}], ",", "AIndex"}], "]"}]}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.728743747091444*^9, 3.728743747498706*^9}, {
  3.728747774709148*^9, 3.728747775339307*^9}},
 CellLabel->
  "In[160]:=",ExpressionUUID->"b9a6929d-0c6a-49b0-946b-da0a9ec9aaaf"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"TT", 
       StyleBox[GridBox[{
          {" ", " ", 
           StyleBox["2",
            FontColor->RGBColor[0., 0., 1.]], "c", " "},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], "b", " ", " ", 
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`TT[{1, -$CellContext`fred}, -$CellContext`b, {
    2, $CellContext`blues}, $CellContext`c, {1, -$CellContext`fred}],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {2, $CellContext`blues}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "c"},
           {
            StyleBox["e",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`e, -$CellContext`fred}, $CellContext`c],
    
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["A",
             FontColor->RGBColor[0., 1., 0.]]},
           {"B", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[-$CellContext`B, {$CellContext`A, $CellContext`fcomp}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", "B"},
           {
            StyleBox["A",
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[{-$CellContext`A, -$CellContext`fcomp}, $CellContext`B],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", "d", 
            StyleBox["e",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], "b", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[{
     1, -$CellContext`fred}, -$CellContext`b, $CellContext`d, \
{$CellContext`e, $CellContext`fred}],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.72866307174539*^9, 3.728671273446652*^9, 
  3.728674898198309*^9, 3.728743310996438*^9, 3.7287437567657137`*^9, 
  3.728747775741579*^9, 3.7287495822031517`*^9, 3.728749947598312*^9, 
  3.728750807040139*^9, 3.7287510809230223`*^9, 3.728751635789316*^9, 
  3.728813282696127*^9, 3.728827400493875*^9, 3.72883599081738*^9, 
  3.7288363066633387`*^9, 3.728843139469502*^9, 3.728845643381034*^9, 
  3.7333674207361403`*^9, 3.792273382566044*^9},
 CellLabel->"Out[160]=",
 CellID->1243667434,ExpressionUUID->"8f52a453-6e25-4200-8bd6-c7af3faded62"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "Notice that some specifications can be redundant; ",
 StyleBox["IndicesOf[Basis, x]", "MR"],
 " is equivalent to ",
 StyleBox["IndicesOf[x]", "MR"],
 ", because all the indices that can be contracted by this function already \
have to appear in ",
 StyleBox["Basis", "MR"],
 "."
}], "MathCaption",
 CellChangeTimes->{{3.728663081093527*^9, 
  3.728663089579892*^9}},ExpressionUUID->"38deea00-fb20-49eb-bf64-\
c902147e552d"],

Cell[CellGroupData[{

Cell[TextData[{
 "Contraction of ",
 StyleBox["Basis", "MR"],
 " objects can be performed with ",
 StyleBox["ContractBasis", "MR"],
 ", or automatized:"
}], "MathCaption",ExpressionUUID->"813e3dcc-82d7-422c-bd47-760da955b048"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Basis", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"{", 
     RowBox[{"b", ",", "blues"}], "}"}]}], "]"}], 
  RowBox[{"Basis", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "fred"}]}], "}"}], ",", "a"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.728743769288903*^9, 3.728743775089315*^9}, {
   3.7287477789734898`*^9, 3.728747781771467*^9}, 3.728749315335889*^9},
 CellLabel->
  "In[161]:=",ExpressionUUID->"f1bb2e5f-6686-4721-b0f9-77d625b4c071"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]},
          {"a", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[-$CellContext`a, {$CellContext`b, $CellContext`blues}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", "a"},
          {
           StyleBox["c",
            FontColor->RGBColor[1., 0., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{-$CellContext`c, -$CellContext`fred}, $CellContext`a],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7286630934629393`*^9, 3.728671273578199*^9, 
  3.728674898351202*^9, 3.728743311135695*^9, 3.728743775633046*^9, 
  3.728747782128055*^9, 3.728749582394476*^9, 3.728749947734995*^9, 
  3.728750807170431*^9, 3.728751081053824*^9, 3.72875163603017*^9, 
  3.728813282850644*^9, 3.728827400646784*^9, 3.728835990977695*^9, 
  3.728836306809739*^9, 3.728843139686997*^9, 3.72884564358099*^9, 
  3.7333674208521*^9, 3.79227338422324*^9},
 CellLabel->"Out[161]=",
 CellID->1016337740,ExpressionUUID->"d99cedcd-a65c-4246-9ff3-df70a31031b6"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ContractBasis", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[162]:=",ExpressionUUID->"f53d44cd-fdad-49ff-b468-16a992b8d49f"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"e", 
      StyleBox[GridBox[{
         {" ", 
          StyleBox["b",
           FontColor->RGBColor[0., 0., 1.]]},
         {
          StyleBox["c",
           FontColor->RGBColor[1., 0., 0.]], " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`Basis[{-$CellContext`c, -$CellContext`fred}, {$CellContext`b, \
$CellContext`blues}],
  Editable->False]], "Output",
 CellChangeTimes->{3.728663094148765*^9, 3.728671273694584*^9, 
  3.72867489850734*^9, 3.728743311285429*^9, 3.72874377655488*^9, 
  3.7287477840080347`*^9, 3.728749582649171*^9, 3.728749947870941*^9, 
  3.728750807318522*^9, 3.7287510811691723`*^9, 3.728751636166449*^9, 
  3.728813283001474*^9, 3.728827400796179*^9, 3.728835991130169*^9, 
  3.728836306972303*^9, 3.728843139872521*^9, 3.728845643763938*^9, 
  3.733367421076902*^9, 3.7922733857577457`*^9},
 CellLabel->"Out[162]=",
 CellID->1903012753,ExpressionUUID->"fe08c6a8-73df-4f90-97bb-bcabb7244d6b"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"AutomaticBasisContractionStart", "[", "]"}]], "Input",
 CellLabel->
  "In[163]:=",ExpressionUUID->"aebd8276-7a99-409a-9470-ce6aecc5708d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Basis", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"{", 
     RowBox[{"b", ",", "blues"}], "}"}]}], "]"}], 
  RowBox[{"Basis", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "fred"}]}], "}"}], ",", "a"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.72874379308004*^9, 3.728743795336542*^9}, {
   3.728747787093163*^9, 3.728747790283039*^9}, 3.728749316302886*^9},
 CellLabel->
  "In[164]:=",ExpressionUUID->"6d53468b-5e7c-4b13-9b1c-7528bc2f41ef"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"e", 
      StyleBox[GridBox[{
         {" ", 
          StyleBox["b",
           FontColor->RGBColor[0., 0., 1.]]},
         {
          StyleBox["c",
           FontColor->RGBColor[1., 0., 0.]], " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`Basis[{-$CellContext`c, -$CellContext`fred}, {$CellContext`b, \
$CellContext`blues}],
  Editable->False]], "Output",
 CellChangeTimes->{3.728663095660439*^9, 3.7286712738496943`*^9, 
  3.728674898863269*^9, 3.728743311486979*^9, 3.7287437959371567`*^9, 
  3.728747790791176*^9, 3.728749583487628*^9, 3.728749947996559*^9, 
  3.7287508075168047`*^9, 3.728751081318787*^9, 3.728751636361573*^9, 
  3.7288132831669807`*^9, 3.72882740101143*^9, 3.728835991285769*^9, 
  3.728836307255384*^9, 3.7288431411400747`*^9, 3.728845643999836*^9, 
  3.733367421457368*^9, 3.792273387861184*^9},
 CellLabel->"Out[164]=",
 CellID->144068692,ExpressionUUID->"787a4d7d-3b0e-4527-b665-ed5a6eac435e"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"AutomaticBasisContractionStop", "[", "]"}]], "Input",
 CellLabel->
  "In[165]:=",ExpressionUUID->"88fed2ab-7212-47f8-a927-88b042fc7695"]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The default behaviour of ",
 StyleBox["ContractBasis", "MR"],
 " disregards contractions involving derivatives. We must use the option ",
 StyleBox["OverDerivatives", "MR"],
 ":"
}], "MathCaption",ExpressionUUID->"fcac0608-b255-4d8e-bc6d-905276cafd3f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefCovD", "[", 
  RowBox[{
   RowBox[{"CD", "[", 
    RowBox[{"-", "a"}], "]"}], ",", 
   RowBox[{"SymbolOfCovD", "->", 
    RowBox[{"{", 
     RowBox[{"\"\<#\>\"", ",", "\"\<D\>\""}], "}"}]}]}], "]"}]], "Input",
 CellLabel->
  "In[166]:=",ExpressionUUID->"a509d96c-81ed-492a-a246-819a74a5c111"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"CD", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "covariant derivative ", 
   $CellContext`CD[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728663099782902*^9, 3.72867127402682*^9, 
  3.728674899034264*^9, 3.7287433116474524`*^9, 3.728743803009328*^9, 
  3.7287477954501057`*^9, 3.7287495838810987`*^9, 3.728749948199922*^9, 
  3.728750807749278*^9, 3.7287510819504766`*^9, 3.72875163656514*^9, 
  3.7288132834033422`*^9, 3.728827401236884*^9, 3.7288359914725237`*^9, 
  3.7288363075564823`*^9, 3.728843141443959*^9, 3.728845644301548*^9, 
  3.7333674216751823`*^9, 3.792273390391007*^9},
 CellLabel->"During evaluation of In[166]:=",
 CellID->1934053673,ExpressionUUID->"c50841bb-4de4-4204-8883-245644f2340c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing torsion tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TorsionCD", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing torsion tensor ", 
   $CellContext`TorsionCD[$CellContext`a, -$CellContext`b, -$CellContext`c], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728663099782902*^9, 3.72867127402682*^9, 
  3.728674899034264*^9, 3.7287433116474524`*^9, 3.728743803009328*^9, 
  3.7287477954501057`*^9, 3.7287495838810987`*^9, 3.728749948199922*^9, 
  3.728750807749278*^9, 3.7287510819504766`*^9, 3.72875163656514*^9, 
  3.7288132834033422`*^9, 3.728827401236884*^9, 3.7288359914725237`*^9, 
  3.7288363075564823`*^9, 3.728843141443959*^9, 3.728845644301548*^9, 
  3.7333674216751823`*^9, 3.792273390403359*^9},
 CellLabel->"During evaluation of In[166]:=",
 CellID->1327601119,ExpressionUUID->"77d2a23d-4b82-48b0-8a1f-210a52c4287c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelCD", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Christoffel tensor ", 
   $CellContext`ChristoffelCD[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728663099782902*^9, 3.72867127402682*^9, 
  3.728674899034264*^9, 3.7287433116474524`*^9, 3.728743803009328*^9, 
  3.7287477954501057`*^9, 3.7287495838810987`*^9, 3.728749948199922*^9, 
  3.728750807749278*^9, 3.7287510819504766`*^9, 3.72875163656514*^9, 
  3.7288132834033422`*^9, 3.728827401236884*^9, 3.7288359914725237`*^9, 
  3.7288363075564823`*^9, 3.728843141443959*^9, 3.728845644301548*^9, 
  3.7333674216751823`*^9, 3.792273390418318*^9},
 CellLabel->"During evaluation of In[166]:=",
 CellID->636914703,ExpressionUUID->"ee3f8f7e-19c8-4a88-a81d-7cb9077e1618"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"Antisymmetric only in the first pair.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Riemann tensor ", 
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
$CellContext`d], ". ", "Antisymmetric only in the first pair."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728663099782902*^9, 3.72867127402682*^9, 
  3.728674899034264*^9, 3.7287433116474524`*^9, 3.728743803009328*^9, 
  3.7287477954501057`*^9, 3.7287495838810987`*^9, 3.728749948199922*^9, 
  3.728750807749278*^9, 3.7287510819504766`*^9, 3.72875163656514*^9, 
  3.7288132834033422`*^9, 3.728827401236884*^9, 3.7288359914725237`*^9, 
  3.7288363075564823`*^9, 3.728843141443959*^9, 3.728845644301548*^9, 
  3.7333674216751823`*^9, 3.79227339042622*^9},
 CellLabel->"During evaluation of In[166]:=",
 CellID->1566916673,ExpressionUUID->"b751e93b-4111-4347-8024-a11eaa173b43"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "non-symmetric Ricci tensor ", 
   $CellContext`RicciCD[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728663099782902*^9, 3.72867127402682*^9, 
  3.728674899034264*^9, 3.7287433116474524`*^9, 3.728743803009328*^9, 
  3.7287477954501057`*^9, 3.7287495838810987`*^9, 3.728749948199922*^9, 
  3.728750807749278*^9, 3.7287510819504766`*^9, 3.72875163656514*^9, 
  3.7288132834033422`*^9, 3.728827401236884*^9, 3.7288359914725237`*^9, 
  3.7288363075564823`*^9, 3.728843141443959*^9, 3.728845644301548*^9, 
  3.7333674216751823`*^9, 3.792273390433702*^9},
 CellLabel->"During evaluation of In[166]:=",
 CellID->576559950,ExpressionUUID->"cdde46d8-f613-40ab-83f1-20a19ec3a244"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Riemann automatically replaced \
by Ricci.\"\>"], "Print",
 CellChangeTimes->{3.728663099782902*^9, 3.72867127402682*^9, 
  3.728674899034264*^9, 3.7287433116474524`*^9, 3.728743803009328*^9, 
  3.7287477954501057`*^9, 3.7287495838810987`*^9, 3.728749948199922*^9, 
  3.728750807749278*^9, 3.7287510819504766`*^9, 3.72875163656514*^9, 
  3.7288132834033422`*^9, 3.728827401236884*^9, 3.7288359914725237`*^9, 
  3.7288363075564823`*^9, 3.728843141443959*^9, 3.728845644301548*^9, 
  3.7333674216751823`*^9, 3.792273390440721*^9},
 CellLabel->"During evaluation of In[166]:=",
 CellID->796864386,ExpressionUUID->"cc36fc9a-4372-4e87-b9a5-e99e1a5987d1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Basis", "[", 
   RowBox[{
    RowBox[{"-", "b"}], ",", 
    RowBox[{"{", 
     RowBox[{"1", ",", "blues"}], "}"}]}], "]"}], 
  RowBox[{
   RowBox[{"CD", "[", 
    RowBox[{"-", "a"}], "]"}], "[", 
   RowBox[{"S", "[", 
    RowBox[{"c", ",", "b"}], "]"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.72874380755916*^9, 3.728743808047893*^9}, {
   3.7287477973965807`*^9, 3.728747798146764*^9}, 3.7287493170364943`*^9},
 CellLabel->
  "In[167]:=",ExpressionUUID->"4e2226a7-a0bc-494a-be9f-4f39fa95d6d0"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["1",
            FontColor->RGBColor[0., 0., 1.]]},
          {"b", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`blues}],
   Editable->False], " ", 
  RowBox[{"(", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["D", "a"], GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {"c", "b"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`CD[-$CellContext`a][
     $CellContext`S[$CellContext`c, $CellContext`b]],
    Editable->False], ")"}]}]], "Output",
 CellChangeTimes->{3.728663100532262*^9, 3.7286712742487392`*^9, 
  3.728674899351944*^9, 3.7287433119078083`*^9, 3.7287438088213787`*^9, 
  3.728747798930056*^9, 3.728749584144321*^9, 3.728749948389165*^9, 
  3.7287508080652933`*^9, 3.7287510821525593`*^9, 3.728751636785783*^9, 
  3.72881328360844*^9, 3.72882740149787*^9, 3.72883599176202*^9, 
  3.7288363078109694`*^9, 3.728843141992927*^9, 3.728845644678824*^9, 
  3.7333674221215687`*^9, 3.79227339144145*^9},
 CellLabel->"Out[167]=",
 CellID->1119621641,ExpressionUUID->"714d8fc7-1a51-4741-acba-885aa17a77f9"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ContractBasis", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[168]:=",ExpressionUUID->"a78ed31d-cde4-49e7-8fe8-9d86255c992b"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["1",
            FontColor->RGBColor[0., 0., 1.]]},
          {"b", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`blues}],
   Editable->False], " ", 
  RowBox[{"(", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["D", "a"], GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {"c", "b"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`CD[-$CellContext`a][
     $CellContext`S[$CellContext`c, $CellContext`b]],
    Editable->False], ")"}]}]], "Output",
 CellChangeTimes->{3.728663101178075*^9, 3.728671274364765*^9, 
  3.7286748994867487`*^9, 3.728743312042774*^9, 3.7287438100907173`*^9, 
  3.7287478004901867`*^9, 3.7287495843137608`*^9, 3.7287499485218077`*^9, 
  3.72875080830048*^9, 3.728751082456458*^9, 3.728751636934523*^9, 
  3.728813283767205*^9, 3.7288274016792097`*^9, 3.728835991914384*^9, 
  3.728836307988325*^9, 3.728843142541194*^9, 3.7288456448219*^9, 
  3.733367422270143*^9, 3.7922733919423943`*^9},
 CellLabel->"Out[168]=",
 CellID->1336700114,ExpressionUUID->"46e5ca8e-6f5f-44d1-8769-a1f6f69d908f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ContractBasis", "[", 
  RowBox[{"%", ",", 
   RowBox[{"OverDerivatives", "\[Rule]", " ", "True"}]}], "]"}]], "Input",
 CellLabel->
  "In[169]:=",ExpressionUUID->"906d7d83-6d13-4c13-8ba4-eef6383ade75"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"ChristoffelCDPDblues", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`ChristoffelCDPDblues[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728743812092696*^9, 3.728747802159236*^9, 
  3.728749584623296*^9, 3.728749948634161*^9, 3.7287508084288588`*^9, 
  3.728751082582986*^9, 3.72875163708454*^9, 3.728813283921212*^9, 
  3.728827401835664*^9, 3.728835992054041*^9, 3.728836308148189*^9, 
  3.728843142727701*^9, 3.7288456449807653`*^9, 3.733367422406664*^9, 
  3.792273392999098*^9},
 CellLabel->"During evaluation of In[169]:=",
 CellID->2113793198,ExpressionUUID->"9b2c4b76-8f19-4d8d-b408-bae86cd577fa"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"D", ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCDPDblues[{
     1, $CellContext`blues}, -$CellContext`a, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"c", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[$CellContext`c, $CellContext`b],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["D", "a"], GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"c", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`a][
    $CellContext`S[$CellContext`c, {1, $CellContext`blues}]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728663101877411*^9, 3.728671274547736*^9, 
  3.728674899652532*^9, 3.728743312177422*^9, 3.728743812197616*^9, 
  3.7287478022259083`*^9, 3.728749584892544*^9, 3.7287499486671762`*^9, 
  3.728750808496327*^9, 3.728751082616646*^9, 3.728751637133222*^9, 
  3.728813283972328*^9, 3.728827401880311*^9, 3.7288359921027822`*^9, 
  3.728836308192698*^9, 3.728843142825169*^9, 3.728845645035654*^9, 
  3.733367422459038*^9, 3.792273393024742*^9},
 CellLabel->"Out[169]=",
 CellID->700306465,ExpressionUUID->"a7f97c6a-5a08-4795-8ec5-ad4bf39529aa"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
A new Christoffel tensor is automatically defined. We shall see more about \
the interplay between bases and derivatives in Section 5. Any type of \
derivate exhibits the same behaviour:\
\>", "MathCaption",
 CellChangeTimes->{{3.7286631068750067`*^9, 
  3.728663108347053*^9}},ExpressionUUID->"5c1fa4a1-08a8-4b37-9e9a-\
ebfb7eabe0b3"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"lieexpr", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"LieD", "[", 
     RowBox[{"v", "[", "a", "]"}], "]"}], "[", 
    RowBox[{"T", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}]}], "]"}], "]"}], 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "fred"}]}], "}"}], ",", "a"}], "]"}], 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"3", ",", 
       RowBox[{"-", "fred"}]}], "}"}], ",", "b"}], "]"}]}]}]], "Input",
 CellChangeTimes->{{3.7287438238388443`*^9, 3.728743826207314*^9}, {
  3.728747808860269*^9, 3.72874781135443*^9}},
 CellLabel->
  "In[170]:=",ExpressionUUID->"695685a0-4fa9-4847-a927-b748e0277606"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", "a"},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", "b"},
          {
           StyleBox["3",
            FontColor->RGBColor[1., 0., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{3, -$CellContext`fred}, $CellContext`b],
   Editable->False], " ", 
  RowBox[{"(", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[ScriptCapitalL]", "v"], GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`LieD[
     $CellContext`v[$CellContext`a]][
     $CellContext`T[-$CellContext`a, -$CellContext`b]],
    Editable->False], ")"}]}]], "Output",
 CellChangeTimes->{3.728663110550679*^9, 3.728671274682137*^9, 
  3.728674899802196*^9, 3.728743312286374*^9, 3.728743826974349*^9, 
  3.728747811756318*^9, 3.728749585283352*^9, 3.728749948804955*^9, 
  3.728750808607388*^9, 3.728751082755324*^9, 3.728751637268841*^9, 
  3.72881328416673*^9, 3.728827402229391*^9, 3.728835992249226*^9, 
  3.728836308388891*^9, 3.728843143875637*^9, 3.728845645166185*^9, 
  3.733367422705554*^9, 3.792273394528611*^9},
 CellLabel->"Out[170]=",
 CellID->444699952,ExpressionUUID->"04e9e040-b068-4a24-afe4-3ec3ca15dca1"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{
    RowBox[{"ContractBasis", "[", "lieexpr", "]"}], ",", 
    RowBox[{"ContractBasis", "[", 
     RowBox[{"lieexpr", ",", 
      RowBox[{"OverDerivatives", "\[Rule]", "True"}]}], "]"}]}], "}"}], "//", 
  "TableForm"}]], "Input",
 CellChangeTimes->{3.728747815633573*^9},
 CellLabel->
  "In[171]:=",ExpressionUUID->"c35b6d88-8db8-43f9-b960-14330ddf37b3"],

Cell[BoxData[
 TagBox[
  TagBox[GridBox[{
     {
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", "a"},
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", "b"},
               {
                StyleBox["3",
                 FontColor->RGBColor[1., 0., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[{3, -$CellContext`fred}, $CellContext`b],
        Editable->False], " ", 
       RowBox[{"(", 
        InterpretationBox[
         StyleBox[
          RowBox[{
           SubscriptBox["\[ScriptCapitalL]", "v"], GridBox[{
             {"T", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {"a", "b"}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
               FontSize->9]}
            },
            GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
            
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}}]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`LieD[
          $CellContext`v[$CellContext`a]][
          $CellContext`T[-$CellContext`a, -$CellContext`b]],
         Editable->False], ")"}]}]},
     {
      RowBox[{
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[ScriptCapitalL]", "v"], GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["3",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`LieD[
         $CellContext`v[$CellContext`a]][
         $CellContext`T[{1, -$CellContext`fred}, {3, -$CellContext`fred}]],
        Editable->False], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {"a", 
                 StyleBox["3",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[-$CellContext`a, {3, -$CellContext`fred}],
         Editable->False], " ", 
        RowBox[{"(", 
         InterpretationBox[
          StyleBox[
           RowBox[{
            SubscriptBox[
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]], GridBox[{
              {"v", 
               StyleBox[GridBox[{
                  {"a"},
                  {" "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                FontSize->9]}
             },
             GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}}]}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`PDfred[{1, -$CellContext`fred}][
           $CellContext`v[$CellContext`a]],
          Editable->False], ")"}]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], "b"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, -$CellContext`fred}, -$CellContext`b],
         Editable->False], " ", 
        RowBox[{"(", 
         InterpretationBox[
          StyleBox[
           RowBox[{
            SubscriptBox[
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]], GridBox[{
              {"v", 
               StyleBox[GridBox[{
                  {"b"},
                  {" "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                FontSize->9]}
             },
             GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}}]}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`PDfred[{3, -$CellContext`fred}][
           $CellContext`v[$CellContext`b]],
          Editable->False], ")"}]}]}]}
    },
    GridBoxAlignment->{"Columns" -> {{Left}}, "Rows" -> {{Baseline}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.5599999999999999]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   Column],
  Function[BoxForm`e$, 
   TableForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{3.728663111298308*^9, 3.728671274818078*^9, 
  3.728674899950275*^9, 3.728743312433391*^9, 3.7287438315121*^9, 
  3.728747818494254*^9, 3.7287495855210323`*^9, 3.728749948915821*^9, 
  3.7287508087314177`*^9, 3.728751082908436*^9, 3.7287516374491587`*^9, 
  3.728813284423177*^9, 3.728827402412491*^9, 3.728835992385984*^9, 
  3.7288363086299343`*^9, 3.728843144225772*^9, 3.728845645294862*^9, 
  3.733367423025365*^9, 3.792273395148541*^9},
 CellLabel->"Out[171]//TableForm=",
 CellID->1942107452,ExpressionUUID->"2ac7433a-d156-4835-aa9d-c4caf79623e3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"paramexpr", "=", 
  RowBox[{
   RowBox[{"OverDot", "[", 
    RowBox[{"T", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}]}], "]"}], "]"}], 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "fred"}]}], "}"}], ",", "a"}], "]"}], 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"3", ",", 
       RowBox[{"-", "fred"}]}], "}"}], ",", "b"}], "]"}]}]}]], "Input",
 CellChangeTimes->{{3.728743839725857*^9, 3.728743841846367*^9}, {
  3.7287478206979923`*^9, 3.7287478228976603`*^9}},
 CellLabel->
  "In[172]:=",ExpressionUUID->"a3b7e8ea-ba13-437a-8cd7-83b01c5608d3"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", "a"},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", "b"},
          {
           StyleBox["3",
            FontColor->RGBColor[1., 0., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{3, -$CellContext`fred}, $CellContext`b],
   Editable->False], " ", 
  OverscriptBox[
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[-$CellContext`a, -$CellContext`b],
    Editable->False], "."]}]], "Output",
 CellChangeTimes->{3.728663112023747*^9, 3.7286712749438477`*^9, 
  3.7286749000673323`*^9, 3.72874331261899*^9, 3.728743842607437*^9, 
  3.728747823399844*^9, 3.728749586072657*^9, 3.728749949053047*^9, 
  3.728750808883502*^9, 3.7287510830544033`*^9, 3.7287516376017437`*^9, 
  3.728813284567219*^9, 3.7288274025708437`*^9, 3.728835992544775*^9, 
  3.72883630887173*^9, 3.728843144436438*^9, 3.728845645437591*^9, 
  3.7333674232255297`*^9, 3.792273396325379*^9},
 CellLabel->"Out[172]=",
 CellID->2060783373,ExpressionUUID->"b0cfb436-f367-4a19-9a9b-941d7ab9e511"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{
    RowBox[{"ContractBasis", "[", "paramexpr", "]"}], ",", 
    RowBox[{"ContractBasis", "[", 
     RowBox[{"paramexpr", ",", 
      RowBox[{"OverDerivatives", "\[Rule]", "True"}]}], "]"}]}], "}"}], "//", 
  "TableForm"}]], "Input",
 CellChangeTimes->{3.728747827346963*^9},
 CellLabel->
  "In[173]:=",ExpressionUUID->"12d1ebb5-23a2-436d-b259-9600e185e618"],

Cell[BoxData[
 TagBox[
  TagBox[GridBox[{
     {
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", "a"},
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", "b"},
               {
                StyleBox["3",
                 FontColor->RGBColor[1., 0., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[{3, -$CellContext`fred}, $CellContext`b],
        Editable->False], " ", 
       OverscriptBox[
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {"a", "b"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[-$CellContext`a, -$CellContext`b],
         Editable->False], "."]}]},
     {
      RowBox[{
       OverscriptBox[
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["3",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, -$CellContext`fred}, {3, -$CellContext`fred}],
         Editable->False], "."], "-", 
       RowBox[{
        OverscriptBox[
         InterpretationBox[
          StyleBox[GridBox[{
             {"e", 
              StyleBox[GridBox[{
                 {" ", "a"},
                 {
                  StyleBox["1",
                   FontColor->RGBColor[1., 0., 0.]], " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
               FontSize->9]}
            },
            GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
            
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
          Editable->False], "."], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {"a", 
                 StyleBox["3",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[-$CellContext`a, {3, -$CellContext`fred}],
         Editable->False]}], "-", 
       RowBox[{
        OverscriptBox[
         InterpretationBox[
          StyleBox[GridBox[{
             {"e", 
              StyleBox[GridBox[{
                 {" ", "b"},
                 {
                  StyleBox["3",
                   FontColor->RGBColor[1., 0., 0.]], " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
               FontSize->9]}
            },
            GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
            
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Basis[{3, -$CellContext`fred}, $CellContext`b],
          Editable->False], "."], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], "b"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, -$CellContext`fred}, -$CellContext`b],
         Editable->False]}]}]}
    },
    GridBoxAlignment->{"Columns" -> {{Left}}, "Rows" -> {{Baseline}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.5599999999999999]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   Column],
  Function[BoxForm`e$, 
   TableForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{3.7286631129477463`*^9, 3.7286712750708838`*^9, 
  3.728674900237917*^9, 3.7287433127663*^9, 3.728743848930964*^9, 
  3.728747827879977*^9, 3.728749586282051*^9, 3.728749949166296*^9, 
  3.728750809009508*^9, 3.728751083208933*^9, 3.728751637744975*^9, 
  3.728813284697267*^9, 3.728827402748296*^9, 3.728835992703635*^9, 
  3.728836309018961*^9, 3.728843144637005*^9, 3.728845645593237*^9, 
  3.7333674233733664`*^9, 3.792273397077754*^9},
 CellLabel->"Out[173]//TableForm=",
 CellID->1155650882,ExpressionUUID->"43454c6d-c37a-4568-903c-d898194c0f15"]
}, Open  ]],

Cell[BoxData[{
 RowBox[{"expr2", "=."}], "\[IndentingNewLine]", 
 RowBox[{"lieexpr", "=."}], "\[IndentingNewLine]", 
 RowBox[{"paramexpr", "=."}]}], "Input",
 CellLabel->
  "In[174]:=",ExpressionUUID->"89bd548d-8fbf-49f7-9f6d-9118b9fbd62c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefCovD", "[", "CD", "]"}]], "Input",
 CellLabel->
  "In[177]:=",ExpressionUUID->"6e7a8336-94b9-420f-85e2-5aa421faf43b"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "ChristoffelCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Christoffel tensor", " ", $CellContext`ChristoffelCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728663115090858*^9, 3.728671275276232*^9, 
  3.728674900453342*^9, 3.72874331294737*^9, 3.7287438556866426`*^9, 
  3.728747830215596*^9, 3.7287495870215797`*^9, 3.728749949438222*^9, 
  3.728750809152625*^9, 3.728751083342888*^9, 3.7287516380953827`*^9, 
  3.728813284920105*^9, 3.7288274029688272`*^9, 3.728835992918973*^9, 
  3.728836309239647*^9, 3.7288431449215717`*^9, 3.7288456458152447`*^9, 
  3.733367423721635*^9, 3.7922733991843243`*^9},
 CellLabel->"During evaluation of In[177]:=",
 CellID->1169328056,ExpressionUUID->"f95facd7-cef4-4f6a-85ec-5a840cfe5520"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "ChristoffelCDPDblues"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`ChristoffelCDPDblues],
  Editable->False]], "Print",
 CellChangeTimes->{3.728663115090858*^9, 3.728671275276232*^9, 
  3.728674900453342*^9, 3.72874331294737*^9, 3.7287438556866426`*^9, 
  3.728747830215596*^9, 3.7287495870215797`*^9, 3.728749949438222*^9, 
  3.728750809152625*^9, 3.728751083342888*^9, 3.7287516380953827`*^9, 
  3.728813284920105*^9, 3.7288274029688272`*^9, 3.728835992918973*^9, 
  3.728836309239647*^9, 3.7288431449215717`*^9, 3.7288456458152447`*^9, 
  3.733367423721635*^9, 3.792273399194531*^9},
 CellLabel->"During evaluation of In[177]:=",
 CellID->1273783495,ExpressionUUID->"c68201fc-d704-4ce8-9020-5db364fafca1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "non-symmetric Ricci tensor", " ", $CellContext`RicciCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728663115090858*^9, 3.728671275276232*^9, 
  3.728674900453342*^9, 3.72874331294737*^9, 3.7287438556866426`*^9, 
  3.728747830215596*^9, 3.7287495870215797`*^9, 3.728749949438222*^9, 
  3.728750809152625*^9, 3.728751083342888*^9, 3.7287516380953827`*^9, 
  3.728813284920105*^9, 3.7288274029688272`*^9, 3.728835992918973*^9, 
  3.728836309239647*^9, 3.7288431449215717`*^9, 3.7288456458152447`*^9, 
  3.733367423721635*^9, 3.7922733992082977`*^9},
 CellLabel->"During evaluation of In[177]:=",
 CellID->214723522,ExpressionUUID->"013cacbb-9ee9-4079-b240-2825dca0872f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728663115090858*^9, 3.728671275276232*^9, 
  3.728674900453342*^9, 3.72874331294737*^9, 3.7287438556866426`*^9, 
  3.728747830215596*^9, 3.7287495870215797`*^9, 3.728749949438222*^9, 
  3.728750809152625*^9, 3.728751083342888*^9, 3.7287516380953827`*^9, 
  3.728813284920105*^9, 3.7288274029688272`*^9, 3.728835992918973*^9, 
  3.728836309239647*^9, 3.7288431449215717`*^9, 3.7288456458152447`*^9, 
  3.733367423721635*^9, 3.792273399215461*^9},
 CellLabel->"During evaluation of In[177]:=",
 CellID->1085707294,ExpressionUUID->"819540d9-5a4b-4888-b299-67394baec7e2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TorsionCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728663115090858*^9, 3.728671275276232*^9, 
  3.728674900453342*^9, 3.72874331294737*^9, 3.7287438556866426`*^9, 
  3.728747830215596*^9, 3.7287495870215797`*^9, 3.728749949438222*^9, 
  3.728750809152625*^9, 3.728751083342888*^9, 3.7287516380953827`*^9, 
  3.728813284920105*^9, 3.7288274029688272`*^9, 3.728835992918973*^9, 
  3.728836309239647*^9, 3.7288431449215717`*^9, 3.7288456458152447`*^9, 
  3.733367423721635*^9, 3.7922733992225103`*^9},
 CellLabel->"During evaluation of In[177]:=",
 CellID->313941699,ExpressionUUID->"d157dd89-ab67-490f-afed-b3cdc243a4bc"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "CD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "covariant derivative", 
   " ", $CellContext`CD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728663115090858*^9, 3.728671275276232*^9, 
  3.728674900453342*^9, 3.72874331294737*^9, 3.7287438556866426`*^9, 
  3.728747830215596*^9, 3.7287495870215797`*^9, 3.728749949438222*^9, 
  3.728750809152625*^9, 3.728751083342888*^9, 3.7287516380953827`*^9, 
  3.728813284920105*^9, 3.7288274029688272`*^9, 3.728835992918973*^9, 
  3.728836309239647*^9, 3.7288431449215717`*^9, 3.7288456458152447`*^9, 
  3.733367423721635*^9, 3.792273399229508*^9},
 CellLabel->"During evaluation of In[177]:=",
 CellID->50814414,ExpressionUUID->"288f7f26-6d97-4546-b49b-d68e3c6cd5f9"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "The input expression is automatically expanded (with ",
 StyleBox["Expand", "MR"],
 ") before applying ",
 StyleBox["ContractBasis", "MR"],
 ". It would be better, in principle, to contract the bases only locally, but \
this is not yet implemented.\nAt the moment, the user cannot specify in which \
order the indices should be contracted. An option ",
 StyleBox["ContractFirst", "MR"],
 " is planned but not yet coded."
}], "Text",
 CellChangeTimes->{
  3.728663126866399*^9},ExpressionUUID->"7cf8717d-6da8-422d-acdd-\
5ec0734cc807"]
}, Closed]],

Cell[CellGroupData[{

Cell["1.2.3. SeparateBasis", "Subsubsection",
 CellChangeTimes->{{3.72867344958838*^9, 
  3.728673449764204*^9}},ExpressionUUID->"2f7db809-afa0-432a-9b5f-\
d4dd5996996f"],

Cell[TextData[{
 StyleBox["SeparateBasis[basis][expr, indices]\t\t", "DisplayMath"],
 StyleBox["Expand the selected indices of a given expression in the specified \
basis ", "Text"]
}], "DefinitionBox",ExpressionUUID->"cf26f1ac-868d-4dad-bc22-c41330818fa7"],

Cell["Expansion of an expression in a given basis.", "Text",ExpressionUUID->"a3d63aae-bc1a-4150-a064-1f924728add3"],

Cell[TextData[{
 StyleBox["SeparateBasis", "MR"],
 " works in a similar way to ",
 StyleBox["ContractBasis", "MR"],
 ", but now we have two brackets. In the second one we can specify which \
indices should be `separated' from their original tensor. In the first one we \
say what should remain in their place (either ",
 StyleBox["AIndex", "MR"],
 " or a basis). "
}], "Text",ExpressionUUID->"63e7b682-5352-437f-8a38-97b2f00904ca"],

Cell[CellGroupData[{

Cell["Take another example expression:", "MathCaption",
 CellChangeTimes->{{3.728663164866006*^9, 3.7286631731444283`*^9}},
 CellID->1734699235,ExpressionUUID->"35b53f9e-4c74-4fa9-a59c-ecf125a3deff"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"TT", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "fred"}]}], "}"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"{", 
       RowBox[{"2", ",", "blues"}], "}"}], ",", "c", ",", 
      RowBox[{"-", "a"}]}], "]"}], 
    RowBox[{"Basis", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "fred"}]}], "}"}], ",", "a"}], "]"}]}], "+", 
   RowBox[{
    RowBox[{"Basis", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "fred"}]}], "}"}], ",", "a"}], "]"}], 
    RowBox[{"Basis", "[", 
     RowBox[{
      RowBox[{"-", "d"}], ",", 
      RowBox[{"{", 
       RowBox[{"2", ",", "blues"}], "}"}]}], "]"}], 
    RowBox[{"U", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", "d", ",", 
      RowBox[{"{", 
       RowBox[{"e", ",", "fred"}], "}"}]}], "]"}], 
    RowBox[{"Basis", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "e"}], ",", 
        RowBox[{"-", "fred"}]}], "}"}], ",", "c"}], "]"}]}]}]}]], "Input",
 CellChangeTimes->{{3.728743885206805*^9, 3.7287439042679863`*^9}, {
  3.728747846609357*^9, 3.728747861560281*^9}, {3.728749324186049*^9, 
  3.728749324875772*^9}},
 CellLabel->
  "In[178]:=",ExpressionUUID->"1272182b-62ed-4834-a933-57c104d4b164"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"TT", 
        StyleBox[GridBox[{
           {" ", " ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], "c", " "},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], "b", " ", " ", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TT[{1, -$CellContext`fred}, -$CellContext`b, {
     2, $CellContext`blues}, $CellContext`c, -$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {2, $CellContext`blues}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "c"},
           {
            StyleBox["e",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`e, -$CellContext`fred}, $CellContext`c],
    
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", "d", 
            StyleBox["e",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", "b", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, -$CellContext`b, $CellContext`d, \
{$CellContext`e, $CellContext`fred}],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728663174416699*^9, 3.7286712824067783`*^9, 
  3.72867490070669*^9, 3.7287433131797943`*^9, 3.728743905372628*^9, 
  3.728747862661521*^9, 3.728749590878407*^9, 3.7287499497006598`*^9, 
  3.728750814811941*^9, 3.728751102028264*^9, 3.728751638296752*^9, 
  3.728813285162055*^9, 3.72882740321819*^9, 3.728835993193042*^9, 
  3.72883630946417*^9, 3.728843158975603*^9, 3.728845646081662*^9, 
  3.733367430091831*^9, 3.792273417632221*^9},
 CellLabel->"Out[178]=",
 CellID->2077686682,ExpressionUUID->"1d216f50-cf0d-46e1-b395-af3ba862f6cb"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"cexpr", "=", 
  RowBox[{"ContractBasis", "[", "expr", "]"}]}]], "Input",
 CellLabel->
  "In[179]:=",ExpressionUUID->"548a95c6-39fa-4cdd-b119-5937f69198f7"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"TT", 
       StyleBox[GridBox[{
          {" ", " ", 
           StyleBox["2",
            FontColor->RGBColor[0., 0., 1.]], "c", " "},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], "b", " ", " ", 
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`TT[{1, -$CellContext`fred}, -$CellContext`b, {
    2, $CellContext`blues}, $CellContext`c, {1, -$CellContext`fred}],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"U", 
       StyleBox[GridBox[{
          {" ", " ", 
           StyleBox["2",
            FontColor->RGBColor[0., 0., 1.]], "c"},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], "b", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`U[{1, -$CellContext`fred}, -$CellContext`b, {
    2, $CellContext`blues}, $CellContext`c],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728663175163403*^9, 3.728671282589719*^9, 
  3.728674900868689*^9, 3.728743313286272*^9, 3.728743907026011*^9, 
  3.728747864378812*^9, 3.728749591048254*^9, 3.728749949838718*^9, 
  3.72875081497607*^9, 3.72875110221124*^9, 3.7287516384172373`*^9, 
  3.7288132853197536`*^9, 3.728827403365364*^9, 3.7288359933367987`*^9, 
  3.7288363095915318`*^9, 3.728843159193825*^9, 3.728845646247662*^9, 
  3.733367430207837*^9, 3.792273418663286*^9},
 CellLabel->"Out[179]=",
 CellID->1669015989,ExpressionUUID->"fcf22ba5-7f9c-47a6-83e3-5d3a430f0158"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SeparateBasis", "[", "AIndex", "]"}], "[", "cexpr", "]"}]], "Input",\

 CellChangeTimes->{{3.728743927091468*^9, 3.72874392748831*^9}},
 CellLabel->
  "In[180]:=",ExpressionUUID->"5bece986-0760-4793-a3a9-7b5708f5f7d4"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {2, $CellContext`blues}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "e"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"TT", 
        StyleBox[GridBox[{
           {" ", " ", "d", "c", " "},
           {"a", "b", " ", " ", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TT[-$CellContext`a, -$CellContext`b, $CellContext`d, \
$CellContext`c, -$CellContext`e],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {2, $CellContext`blues}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", "d", "c"},
           {"a", "b", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, -$CellContext`b, $CellContext`d, \
$CellContext`c],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{
  3.728663176370295*^9, 3.728671282771544*^9, 3.728674901068562*^9, 
   3.728743313457999*^9, {3.728743912723126*^9, 3.728743945426394*^9}, 
   3.728747865100656*^9, 3.728749591401835*^9, 3.728749949950965*^9, 
   3.728750815186393*^9, 3.728751102429246*^9, 3.7287516385462418`*^9, 
   3.728813285464039*^9, 3.728827403579996*^9, 3.7288359934898157`*^9, 
   3.7288363097335978`*^9, 3.728843159405739*^9, 3.728845646394927*^9, 
   3.7333674303612556`*^9, 3.7922734193014393`*^9},
 CellLabel->"Out[180]=",
 CellID->1409327831,ExpressionUUID->"f69450ac-f1a0-4236-b82d-569c68fcc130"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Simplification", "[", "%", "]"}]], "Input",
 CellChangeTimes->{{3.728743947912505*^9, 3.728743955732843*^9}},
 CellLabel->
  "In[181]:=",ExpressionUUID->"8164679e-d859-4574-a001-ab58fa03c245"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["2",
            FontColor->RGBColor[0., 0., 1.]]},
          {"a", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`blues}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", "d"},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`d],
   Editable->False], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", "e"},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`e],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"TT", 
          StyleBox[GridBox[{
             {" ", " ", "a", "c", " "},
             {"d", "b", " ", " ", "e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TT[-$CellContext`d, -$CellContext`b, $CellContext`a, \
$CellContext`c, -$CellContext`e],
      Editable->False]}], "+", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"U", 
         StyleBox[GridBox[{
            {" ", " ", "a", "c"},
            {"d", "b", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`U[-$CellContext`d, -$CellContext`b, $CellContext`a, \
$CellContext`c],
     Editable->False]}], ")"}]}]], "Output",
 CellChangeTimes->{
  3.728663177098195*^9, 3.7286712829358997`*^9, 3.728674901255007*^9, 
   3.7287433135750637`*^9, {3.728743930323833*^9, 3.7287439566230383`*^9}, 
   3.7287478657128067`*^9, 3.728749591920096*^9, 3.728749950099045*^9, 
   3.72875081535963*^9, 3.728751102949999*^9, 3.728751638668673*^9, 
   3.72881328563348*^9, 3.7288274037422943`*^9, 3.728835993644795*^9, 
   3.728836309927886*^9, 3.728843159621855*^9, 3.728845646564481*^9, 
   3.7333674305090857`*^9, 3.792273419812639*^9},
 CellLabel->"Out[181]=",
 CellID->279075165,ExpressionUUID->"a3421b9e-a7cb-48fe-b88d-a990b05485b5"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
But we may also want to perform a change of basis. This is done as follows\
\>", "MathCaption",ExpressionUUID->"af622d27-d7d8-4beb-86f0-1a2ab685a91b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SeparateBasis", "[", "blues", "]"}], "[", "cexpr", "]"}]], "Input",\

 CellChangeTimes->{{3.728743963953397*^9, 3.72874396791857*^9}, {
   3.7287478694393997`*^9, 3.728747870143303*^9}, 3.7287493258380117`*^9},
 CellLabel->
  "In[182]:=",ExpressionUUID->"20a5e2fb-3ad2-413c-a702-f83641f675d6"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["d",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {$CellContext`d, $CellContext`blues}],
    
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["a",
             FontColor->RGBColor[0., 0., 1.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{
     1, -$CellContext`fred}, {$CellContext`a, $CellContext`blues}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["f1",
             FontColor->RGBColor[0., 0., 1.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{
     1, -$CellContext`fred}, {$CellContext`f1, $CellContext`blues}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {
            StyleBox["e",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`e, -$CellContext`blues}, {
     2, $CellContext`blues}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "c"},
           {
            StyleBox["f",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`f, -$CellContext`blues}, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"TT", 
        StyleBox[GridBox[{
           {" ", " ", 
            StyleBox["e",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["f",
             FontColor->RGBColor[0., 0., 1.]], " "},
           {
            StyleBox["a",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["d",
             FontColor->RGBColor[0., 0., 1.]], " ", " ", 
            StyleBox["f1",
             FontColor->RGBColor[0., 0., 1.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TT[{-$CellContext`a, -$CellContext`blues}, {-$CellContext`d, \
-$CellContext`blues}, {$CellContext`e, $CellContext`blues}, {$CellContext`f, \
$CellContext`blues}, {-$CellContext`f1, -$CellContext`blues}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["d",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {$CellContext`d, $CellContext`blues}],
    
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["a",
             FontColor->RGBColor[0., 0., 1.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{
     1, -$CellContext`fred}, {$CellContext`a, $CellContext`blues}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {
            StyleBox["e",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`e, -$CellContext`blues}, {
     2, $CellContext`blues}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "c"},
           {
            StyleBox["f",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`f, -$CellContext`blues}, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", 
            StyleBox["e",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["f",
             FontColor->RGBColor[0., 0., 1.]]},
           {
            StyleBox["a",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["d",
             FontColor->RGBColor[0., 0., 1.]], " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[{-$CellContext`a, -$CellContext`blues}, {-$CellContext`d,
       -$CellContext`blues}, {$CellContext`e, $CellContext`blues}, \
{$CellContext`f, $CellContext`blues}],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728663184088327*^9, 3.7286712831020412`*^9, 
  3.728674901397585*^9, 3.72874331380646*^9, 3.728743968442474*^9, 
  3.7287478707331944`*^9, 3.7287495921112757`*^9, 3.72874995026729*^9, 
  3.728750815542727*^9, 3.728751103225436*^9, 3.728751638809885*^9, 
  3.7288132857826443`*^9, 3.728827403941143*^9, 3.7288359937879257`*^9, 
  3.7288363101419363`*^9, 3.7288431598193483`*^9, 3.728845646693261*^9, 
  3.73336743102462*^9, 3.7922734210091753`*^9},
 CellLabel->"Out[182]=",
 CellID->1165567163,ExpressionUUID->"7a34c1e8-71cc-4ff5-b36a-9ce932cd333b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Simplification", "[", "%", "]"}]], "Input",
 CellChangeTimes->{{3.728743973311306*^9, 3.728743977167922*^9}},
 CellLabel->
  "In[183]:=",ExpressionUUID->"e1fcc168-85f5-446b-b0b9-2731f2001ca8"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["a",
            FontColor->RGBColor[0., 0., 1.]]},
          {"b", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[-$CellContext`b, {$CellContext`a, $CellContext`blues}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["f",
            FontColor->RGBColor[0., 0., 1.]]},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{
    1, -$CellContext`fred}, {$CellContext`f, $CellContext`blues}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", "c"},
          {
           StyleBox["d",
            FontColor->RGBColor[0., 0., 1.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{-$CellContext`d, -$CellContext`blues}, $CellContext`c],
   
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["2",
            FontColor->RGBColor[0., 0., 1.]]},
          {
           StyleBox["e",
            FontColor->RGBColor[0., 0., 1.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{-$CellContext`e, -$CellContext`blues}, {
    2, $CellContext`blues}],
   Editable->False], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["f1",
               FontColor->RGBColor[0., 0., 1.]]},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{
       1, -$CellContext`fred}, {$CellContext`f1, $CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"TT", 
          StyleBox[GridBox[{
             {" ", " ", 
              StyleBox["e",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["d",
               FontColor->RGBColor[0., 0., 1.]], " "},
             {
              StyleBox["f",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["a",
               FontColor->RGBColor[0., 0., 1.]], " ", " ", 
              StyleBox["f1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TT[{-$CellContext`f, -$CellContext`blues}, \
{-$CellContext`a, -$CellContext`blues}, {$CellContext`e, $CellContext`blues}, \
{$CellContext`d, $CellContext`blues}, {-$CellContext`f1, -$CellContext`blues}],
      Editable->False]}], "+", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"U", 
         StyleBox[GridBox[{
            {" ", " ", 
             StyleBox["e",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]]},
            {
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`U[{-$CellContext`f, -$CellContext`blues}, {-$CellContext`a, \
-$CellContext`blues}, {$CellContext`e, $CellContext`blues}, {$CellContext`d, \
$CellContext`blues}],
     Editable->False]}], ")"}]}]], "Output",
 CellChangeTimes->{3.7286631847503567`*^9, 3.728671283247098*^9, 
  3.728674901522379*^9, 3.728743313942835*^9, 3.728743977902967*^9, 
  3.728747871526704*^9, 3.728749592424479*^9, 3.728749950433887*^9, 
  3.7287508161642303`*^9, 3.728751103443327*^9, 3.728751638942452*^9, 
  3.728813286000019*^9, 3.7288274041168013`*^9, 3.728835993947585*^9, 
  3.728836310327475*^9, 3.7288431606611233`*^9, 3.728845646850769*^9, 
  3.733367431171801*^9, 3.792273421667903*^9},
 CellLabel->"Out[183]=",
 CellID->966044531,ExpressionUUID->"1e212e50-f90c-420a-ae1b-804b79d0d977"]
}, Open  ]]
}, Open  ]],

Cell["Now all tensor indices belong to the specified basis.", "MathCaption",
 CellChangeTimes->{{3.728663198472712*^9, 3.72866319971918*^9}},
 CellID->2091127534,ExpressionUUID->"3b71e45c-0a1c-4907-a514-80a783e8c4b0"],

Cell[CellGroupData[{

Cell[TextData[{
 "We have the same syntax as in ",
 StyleBox["ContractBasis", "MR"],
 " for the second argument:"
}], "MathCaption",
 CellChangeTimes->{{3.728663213750051*^9, 
  3.728663219285885*^9}},ExpressionUUID->"0150f988-a717-4c0e-aad7-\
d10771d03726"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SeparateBasis", "[", "AIndex", "]"}], "[", 
  RowBox[{"cexpr", ",", 
   RowBox[{"{", 
    RowBox[{"4", ",", "blues"}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.7287439953037567`*^9, 3.728743995679036*^9}, {
   3.728747876487*^9, 3.7287478771830893`*^9}, 3.728749326671093*^9},
 CellLabel->
  "In[184]:=",ExpressionUUID->"c13c403d-395d-4efb-94d6-7973939f84f4"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"TT", 
       StyleBox[GridBox[{
          {" ", " ", 
           StyleBox["2",
            FontColor->RGBColor[0., 0., 1.]], "c", " "},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], "b", " ", " ", 
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`TT[{1, -$CellContext`fred}, -$CellContext`b, {
    2, $CellContext`blues}, $CellContext`c, {1, -$CellContext`fred}],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"U", 
       StyleBox[GridBox[{
          {" ", " ", 
           StyleBox["2",
            FontColor->RGBColor[0., 0., 1.]], "c"},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], "b", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`U[{1, -$CellContext`fred}, -$CellContext`b, {
    2, $CellContext`blues}, $CellContext`c],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728663221113062*^9, 3.7286712834640827`*^9, 
  3.7286749016677847`*^9, 3.728743314240777*^9, 3.728743996206026*^9, 
  3.728747877673833*^9, 3.7287495925950747`*^9, 3.728749950570953*^9, 
  3.72875081637811*^9, 3.72875110359938*^9, 3.728751639078404*^9, 
  3.728813286218667*^9, 3.728827404265697*^9, 3.728835994081532*^9, 
  3.7288363104647713`*^9, 3.728843160921865*^9, 3.728845646989284*^9, 
  3.733367431429491*^9, 3.792273423399088*^9},
 CellLabel->"Out[184]=",
 CellID->133944636,ExpressionUUID->"8f4c9473-cf3a-46cc-a0d3-97e07d926ee7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SeparateBasis", "[", "AIndex", "]"}], "[", 
  RowBox[{"cexpr", ",", 
   RowBox[{"IndicesOf", "[", "blues", "]"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.7287439985919733`*^9, 3.7287439989399443`*^9}, {
   3.728747879216832*^9, 3.728747879830783*^9}, 3.7287493273184433`*^9},
 CellLabel->
  "In[185]:=",ExpressionUUID->"117cbe02-7573-4b99-86f4-cce3bdb2e2ae"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`blues}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"TT", 
        StyleBox[GridBox[{
           {" ", " ", "a", "c", " "},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], "b", " ", " ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TT[{
     1, -$CellContext`fred}, -$CellContext`b, $CellContext`a, $CellContext`c, \
{1, -$CellContext`fred}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`blues}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", "a", "c"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], "b", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[{
     1, -$CellContext`fred}, -$CellContext`b, $CellContext`a, $CellContext`c],
    
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.72866322171522*^9, 3.728671283645924*^9, 
  3.728674901803842*^9, 3.728743314391651*^9, 3.7287439995291843`*^9, 
  3.728747880289*^9, 3.728749593171*^9, 3.728749950688672*^9, 
  3.728750816829886*^9, 3.728751104114324*^9, 3.728751639211591*^9, 
  3.728813286375731*^9, 3.728827404447122*^9, 3.7288359942395*^9, 
  3.728836310593894*^9, 3.728843161138939*^9, 3.728845647224126*^9, 
  3.733367431607966*^9, 3.792273424463252*^9},
 CellLabel->"Out[185]=",
 CellID->1477250969,ExpressionUUID->"b94be748-8b20-466a-b603-ec8f8ec7ad47"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SeparateBasis", "[", "AIndex", "]"}], "[", 
  RowBox[{"cexpr", ",", 
   RowBox[{"IndicesOf", "[", "CIndex", "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[186]:=",ExpressionUUID->"a9359986-8cb3-41b6-ac27-22a1fc133872"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {2, $CellContext`blues}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "e"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"TT", 
        StyleBox[GridBox[{
           {" ", " ", "d", "c", " "},
           {"a", "b", " ", " ", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TT[-$CellContext`a, -$CellContext`b, $CellContext`d, \
$CellContext`c, -$CellContext`e],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {2, $CellContext`blues}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", "d", "c"},
           {"a", "b", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, -$CellContext`b, $CellContext`d, \
$CellContext`c],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728663222310565*^9, 3.728671283792646*^9, 
  3.728674901938335*^9, 3.728743314546207*^9, 3.728744004405684*^9, 
  3.7287478822194147`*^9, 3.728749593518612*^9, 3.728749950838234*^9, 
  3.728750817363805*^9, 3.7287511044427156`*^9, 3.7287516393459187`*^9, 
  3.7288132865001698`*^9, 3.7288274047265873`*^9, 3.728835994384535*^9, 
  3.7288363107795067`*^9, 3.7288431613560133`*^9, 3.728845647439254*^9, 
  3.733367431857011*^9, 3.792273424945182*^9},
 CellLabel->"Out[186]=",
 CellID->1543138871,ExpressionUUID->"7317b7cf-663b-4ce1-8324-f6f25fcca5b9"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
When separating component indices, the function selects only those with valid \
coordinate numbers:\
\>", "MathCaption",
 CellChangeTimes->{{3.7286632241656027`*^9, 
  3.7286632244616833`*^9}},ExpressionUUID->"750d3378-e6a2-4a32-b351-\
2b3ab9c43cbd"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"v", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "fred"}], "}"}], "]"}], 
   RowBox[{"v", "[", 
    RowBox[{"{", 
     RowBox[{"2", ",", "fred"}], "}"}], "]"}], 
   RowBox[{"v", "[", 
    RowBox[{"{", 
     RowBox[{"7", ",", "fred"}], "}"}], "]"}]}], "//", 
  RowBox[{"SeparateBasis", "[", "AIndex", "]"}]}]], "Input",
 CellChangeTimes->{{3.728744017397813*^9, 3.728744021509976*^9}, {
  3.7287478842645893`*^9, 3.7287478882064466`*^9}},
 CellLabel->
  "In[187]:=",ExpressionUUID->"d59cbb71-0bfe-491e-9023-bb6ba33d23cb"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]]},
          {"a", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`fred}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["2",
            FontColor->RGBColor[1., 0., 0.]]},
          {"b", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`fred}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"a"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"b"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {
           StyleBox["7",
            FontColor->RGBColor[1., 0., 0.]]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{7, $CellContext`fred}],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728663226296311*^9, 3.7286712839440317`*^9, 
  3.7286749020852613`*^9, 3.72874331471382*^9, 3.728744022313785*^9, 
  3.7287478887738647`*^9, 3.72874959372252*^9, 3.728749950972891*^9, 
  3.728750817562345*^9, 3.72875110464837*^9, 3.728751639496512*^9, 
  3.728813286634178*^9, 3.728827404881271*^9, 3.72883599453854*^9, 
  3.728836310912809*^9, 3.7288431621160173`*^9, 3.7288456475729237`*^9, 
  3.733367432003793*^9, 3.792273426169119*^9},
 CellLabel->"Out[187]=",
 CellID->1668739797,ExpressionUUID->"4205d0f8-5153-4df4-9f84-8653bbc7422b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We must be careful when working with scalar functions of scalars, lest we get \
meaningless outputs, such as powers of tensor products:\
\>", "MathCaption",
 CellChangeTimes->{
  3.728663229197392*^9},ExpressionUUID->"98de541f-0321-403b-8282-\
8fd22f924a10"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"v", "[", 
   RowBox[{"{", 
    RowBox[{"1", ",", "fred"}], "}"}], "]"}], "//", 
  RowBox[{"SeparateBasis", "[", "AIndex", "]"}]}]], "Input",
 CellChangeTimes->{{3.7287440288226833`*^9, 3.728744029309675*^9}, {
  3.728747890344105*^9, 3.728747890782249*^9}},
 CellLabel->
  "In[188]:=",ExpressionUUID->"3c6ec5dd-41b6-44f6-9482-a090eb9545f4"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]]},
          {"a", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`fred}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"a"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`a],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728663230395639*^9, 3.72867128407758*^9, 
  3.728674902217338*^9, 3.7287433149071903`*^9, 3.7287440312002573`*^9, 
  3.728747891104973*^9, 3.7287495939208107`*^9, 3.728749951182963*^9, 
  3.7287508177150307`*^9, 3.7287511048245173`*^9, 3.728751639676751*^9, 
  3.728813286751265*^9, 3.728827405047646*^9, 3.7288359946678677`*^9, 
  3.728836311045084*^9, 3.7288431625278273`*^9, 3.7288456477031813`*^9, 
  3.733367432214287*^9, 3.792273427862135*^9},
 CellLabel->"Out[188]=",
 CellID->2116524650,ExpressionUUID->"c07842ee-433d-498e-8820-509b2b6eb15c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"v", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "fred"}], "}"}], "]"}], 
   RowBox[{"v", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "fred"}], "}"}], "]"}]}], "//", 
  RowBox[{"SeparateBasis", "[", "AIndex", "]"}]}]], "Input",
 CellChangeTimes->{{3.728744033788046*^9, 3.7287440375409813`*^9}, {
  3.7287478929201927`*^9, 3.728747895237998*^9}},
 CellLabel->
  "In[189]:=",ExpressionUUID->"614ccb8a-3133-45a4-9207-1a9346de3694"],

Cell[BoxData[
 SuperscriptBox[
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{1, $CellContext`fred}],
   Editable->False], "2"]], "Output",
 CellChangeTimes->{3.728663230996005*^9, 3.728671284215716*^9, 
  3.728674902340604*^9, 3.728743315055085*^9, 3.728744038382916*^9, 
  3.728747895807766*^9, 3.728749594101077*^9, 3.728749951320107*^9, 
  3.728750817909144*^9, 3.728751105008654*^9, 3.7287516398431807`*^9, 
  3.72881328686714*^9, 3.7288274052135477`*^9, 3.7288359947964687`*^9, 
  3.728836311179101*^9, 3.728843162738806*^9, 3.728845647835169*^9, 
  3.733367432474581*^9, 3.792273428364567*^9},
 CellLabel->"Out[189]=",
 CellID->1493055970,ExpressionUUID->"c836bac3-1cd4-4be4-85cc-333860fe808d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%%", " ", "^", "2"}]], "Input",
 CellLabel->
  "In[190]:=",ExpressionUUID->"df320ab0-6523-4844-9657-18b498e9fd6e"],

Cell[BoxData[
 RowBox[{
  SuperscriptBox[
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"f$9271", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`f$9271, {1, $CellContext`fred}],
    Editable->False], "2"], " ", 
  SuperscriptBox[
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"f$9271"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`f$9271],
    Editable->False], "2"]}]], "Output",
 CellChangeTimes->{3.728663231725615*^9, 3.728671284350449*^9, 
  3.728674902470661*^9, 3.7287433152159367`*^9, 3.728744039687871*^9, 
  3.728747896570108*^9, 3.728749594554093*^9, 3.728749951449985*^9, 
  3.7287508180649767`*^9, 3.7287511051920843`*^9, 3.7287516399652777`*^9, 
  3.728813286973216*^9, 3.72882740539813*^9, 3.728835994930071*^9, 
  3.728836311358829*^9, 3.728843162955448*^9, 3.7288456479852953`*^9, 
  3.733367432624062*^9, 3.7922734295339737`*^9},
 CellLabel->"Out[190]=",
 CellID->941986412,ExpressionUUID->"231ee5ee-48b5-4210-902f-90db21d99bf3"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The safe way to manipulate such expressions is via the ",
 StyleBox["Scalar", "MR"],
 " head:"
}], "MathCaption",
 CellChangeTimes->{{3.7286632339891253`*^9, 
  3.7286632343325787`*^9}},ExpressionUUID->"4a041000-bef7-4041-aadd-\
e82e849a574e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Scalar", "[", 
   RowBox[{"v", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "fred"}], "}"}], "]"}], "]"}], 
  RowBox[{"Scalar", "[", 
   RowBox[{"v", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "fred"}], "}"}], "]"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.7287440541882677`*^9, 3.728744059786496*^9}, {
  3.728747899398151*^9, 3.728747901741927*^9}},
 CellLabel->
  "In[191]:=",ExpressionUUID->"a0d93d66-1d28-4f23-b1dc-77e13cf8d6b1"],

Cell[BoxData[
 SuperscriptBox[
  InterpretationBox[
   StyleBox[
    RowBox[{
     StyleBox["(",
      FontColor->RGBColor[1, 0, 0]], 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{1, $CellContext`fred}],
      Editable->False], 
     StyleBox[")",
      FontColor->RGBColor[1, 0, 0]]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Scalar[
    $CellContext`v[{1, $CellContext`fred}]],
   Editable->False], "2"]], "Output",
 CellChangeTimes->{3.728663235412464*^9, 3.728671284482861*^9, 
  3.7286749026200867`*^9, 3.72874331532543*^9, 3.7287440658684673`*^9, 
  3.728747902388307*^9, 3.728749594985111*^9, 3.7287499515864983`*^9, 
  3.728750818214415*^9, 3.7287511053427896`*^9, 3.7287516400863237`*^9, 
  3.72881328710596*^9, 3.728827405564492*^9, 3.728835995068254*^9, 
  3.728836311475738*^9, 3.7288431631263933`*^9, 3.7288456481260843`*^9, 
  3.7333674327764874`*^9, 3.7922734331941967`*^9},
 CellLabel->"Out[191]=",
 CellID->300289431,ExpressionUUID->"6a1c9a0e-7e2c-475c-bfa5-5068437f7e28"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", 
  RowBox[{"SeparateBasis", "[", "blues", "]"}]}]], "Input",
 CellChangeTimes->{{3.728744061900717*^9, 3.728744062264105*^9}, {
   3.7287479040954742`*^9, 3.72874790483786*^9}, 3.728749328300708*^9},
 CellLabel->
  "In[192]:=",ExpressionUUID->"b6f6df80-bd50-4a40-b68d-5042d28062f9"],

Cell[BoxData[
 SuperscriptBox[
  InterpretationBox[
   StyleBox[
    RowBox[{
     StyleBox["(",
      FontColor->RGBColor[1, 0, 0]], 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{1, $CellContext`fred}],
      Editable->False], 
     StyleBox[")",
      FontColor->RGBColor[1, 0, 0]]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Scalar[
    $CellContext`v[{1, $CellContext`fred}]],
   Editable->False], "2"]], "Output",
 CellChangeTimes->{3.728663236079261*^9, 3.728671284645269*^9, 
  3.728674902770996*^9, 3.728743315439351*^9, 3.7287440667865562`*^9, 
  3.728747905325137*^9, 3.728749595161807*^9, 3.7287499517538424`*^9, 
  3.728750818362625*^9, 3.728751105501089*^9, 3.7287516402328*^9, 
  3.7288132872509813`*^9, 3.72882740576228*^9, 3.728835995198235*^9, 
  3.7288363116762533`*^9, 3.728843163762068*^9, 3.728845648256156*^9, 
  3.73336743302509*^9, 3.792273433730762*^9},
 CellLabel->"Out[192]=",
 CellID->1241121440,ExpressionUUID->"0e74474e-5fb6-471a-a734-c8f051915c28"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", 
  RowBox[{
   RowBox[{"Scalar", "[", "scalar_", "]"}], "\[RuleDelayed]", " ", 
   RowBox[{"Scalar", "[", 
    RowBox[{
     RowBox[{"SeparateBasis", "[", "blues", "]"}], "[", "scalar", "]"}], 
    "]"}]}]}]], "Input",
 CellChangeTimes->{{3.728744070131301*^9, 3.7287440705717793`*^9}, {
   3.728747909127335*^9, 3.728747909868451*^9}, 3.728749329187169*^9},
 CellLabel->
  "In[193]:=",ExpressionUUID->"3cf7be0a-3efb-4bca-bda3-295dd56715b3"],

Cell[BoxData[
 SuperscriptBox[
  InterpretationBox[
   StyleBox[
    RowBox[{
     StyleBox["(",
      FontColor->RGBColor[1, 0, 0]], 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{-$CellContext`a, -$CellContext`blues}, {
        1, $CellContext`fred}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{$CellContext`a, $CellContext`blues}],
       Editable->False]}], 
     StyleBox[")",
      FontColor->RGBColor[1, 0, 0]]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Scalar[
   xAct`xTensor`Basis[{-$CellContext`a, -$CellContext`blues}, {
      1, $CellContext`fred}] $CellContext`v[{$CellContext`a, \
$CellContext`blues}]],
   Editable->False], "2"]], "Output",
 CellChangeTimes->{3.728663236742922*^9, 3.728671284801283*^9, 
  3.728674902920618*^9, 3.728743315608817*^9, 3.728744071992042*^9, 
  3.7287479103757544`*^9, 3.728749595301793*^9, 3.7287499519006357`*^9, 
  3.7287508185148783`*^9, 3.728751105667885*^9, 3.728751640350806*^9, 
  3.728813287401247*^9, 3.7288274059411173`*^9, 3.728835995338673*^9, 
  3.728836311893959*^9, 3.728843164623775*^9, 3.728845648380205*^9, 
  3.733367433298826*^9, 3.7922734342608147`*^9},
 CellLabel->"Out[193]=",
 CellID->1893679720,ExpressionUUID->"6f3bc8d0-92a0-42b0-b48d-111d698ee5e9"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "If a component index is repeated, we may need to apply ",
 StyleBox["SeparateBasis", "MR"],
 " several times:"
}], "MathCaption",
 CellChangeTimes->{{3.728663238988908*^9, 
  3.728663240309353*^9}},ExpressionUUID->"d622fb7a-a083-4ed9-a722-\
d6bc7b296e1e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr4", "=", 
  RowBox[{"T", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"1", ",", 
      RowBox[{"-", "fred"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{"1", ",", 
      RowBox[{"-", "fred"}]}], "}"}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.728744079355933*^9, 3.728744090009165*^9}, {
  3.728747912695608*^9, 3.728747914869269*^9}},
 CellLabel->
  "In[194]:=",ExpressionUUID->"3a47f130-7b80-48af-bfe4-fdffe2b8e5b5"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"T", 
      StyleBox[GridBox[{
         {" ", " "},
         {
          StyleBox["1",
           FontColor->RGBColor[1., 0., 0.]], 
          StyleBox["1",
           FontColor->RGBColor[1., 0., 0.]]}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`T[{1, -$CellContext`fred}, {1, -$CellContext`fred}],
  Editable->False]], "Output",
 CellChangeTimes->{3.728663241480397*^9, 3.7286712849379177`*^9, 
  3.728674903073415*^9, 3.728743315812944*^9, 3.728744108925407*^9, 
  3.728747915227124*^9, 3.728749595632992*^9, 3.728749952038205*^9, 
  3.728750818669423*^9, 3.728751105821662*^9, 3.7287516404842253`*^9, 
  3.728813287517503*^9, 3.7288274061005373`*^9, 3.728835995479603*^9, 
  3.7288363120288*^9, 3.728843164822515*^9, 3.72884564851045*^9, 
  3.7333674334418383`*^9, 3.792273435212376*^9},
 CellLabel->"Out[194]=",
 CellID->1292102491,ExpressionUUID->"d14cd1a0-6d50-4705-89b7-491163652aaf"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", 
  RowBox[{"SeparateBasis", "[", "AIndex", "]"}]}]], "Input",
 CellLabel->
  "In[195]:=",ExpressionUUID->"9ba56b7c-cab6-4c3a-ac3c-0938578f2d7e"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", "a"},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", "b"},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {" ", " "},
          {"a", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[-$CellContext`a, -$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{
  3.7286632422296867`*^9, 3.728671285068437*^9, 3.728674903187726*^9, 
   3.728743315959894*^9, {3.7287441054145107`*^9, 3.728744109818117*^9}, 
   3.728747916643712*^9, 3.728749596189513*^9, 3.7287499521505423`*^9, 
   3.728750819095861*^9, 3.72875110596621*^9, 3.728751640599524*^9, 
   3.728813287649963*^9, 3.728827406274405*^9, 3.7288359956141443`*^9, 
   3.728836312160178*^9, 3.728843165039933*^9, 3.728845648637105*^9, 
   3.733367433648641*^9, 3.792273435751055*^9},
 CellLabel->"Out[195]=",
 CellID->1252954152,ExpressionUUID->"a582cc06-2e82-4fbc-8d90-4c9079aadc00"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["It works, but if we specify the index:", "MathCaption",
 CellChangeTimes->{{3.728663243924008*^9, 
  3.728663244388739*^9}},ExpressionUUID->"56557915-6da7-4235-bb14-\
4ca4a504aab7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SeparateBasis", "[", "AIndex", "]"}], "[", 
  RowBox[{"expr4", ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", 
     RowBox[{"-", "fred"}]}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{
  3.7287441009452543`*^9, {3.728747919077051*^9, 3.728747919540966*^9}},
 CellLabel->
  "In[196]:=",ExpressionUUID->"a184a7d2-2e33-4725-8f83-aa8413e16b2c"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", "a"},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {" ", " "},
          {"a", 
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[-$CellContext`a, {1, -$CellContext`fred}],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728663245613999*^9, 3.7286712852164173`*^9, 
  3.72867490331846*^9, 3.7287433161246767`*^9, 3.728744123544303*^9, 
  3.728747920077636*^9, 3.7287495963694487`*^9, 3.7287499522714148`*^9, 
  3.728750819292346*^9, 3.7287511061059523`*^9, 3.728751640732603*^9, 
  3.728813287851083*^9, 3.728827406446835*^9, 3.728835995747025*^9, 
  3.728836312295662*^9, 3.728843165255814*^9, 3.728845648773655*^9, 
  3.733367433957999*^9, 3.792273437349799*^9},
 CellLabel->"Out[196]=",
 CellID->1244693224,ExpressionUUID->"aa3a2aca-4d78-4b02-a2ef-e78e4760f824"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["{1, -fred}", "MR"],
 " appears twice in the same tensor, so we need to reapply ",
 StyleBox["SeparateBasis", "MR"],
 " (or pass it twice to the function). TODO: Is this correct?:"
}], "MathCaption",
 CellChangeTimes->{{3.728663248348638*^9, 3.728663249620529*^9}, {
  3.728747924782755*^9, 
  3.728747951299535*^9}},ExpressionUUID->"c919089f-7209-47b9-9346-\
895425605ddc"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SeparateBasis", "[", "AIndex", "]"}], "[", 
  RowBox[{"%", ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", 
     RowBox[{"-", "fred"}]}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728744129912511*^9, 3.728744133332665*^9}, {
  3.728747921614811*^9, 3.728747922556794*^9}},
 CellLabel->
  "In[197]:=",ExpressionUUID->"ba14f68a-792e-46b1-a095-539345b5bc0e"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", "a"},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", "b"},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {" ", " "},
          {"a", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[-$CellContext`a, -$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728663250648176*^9, 3.728671285367091*^9, 
  3.728674903448752*^9, 3.728743316287383*^9, 3.728744134428835*^9, 
  3.7287479233908653`*^9, 3.728749596507544*^9, 3.728749952422155*^9, 
  3.728750819581649*^9, 3.728751106239698*^9, 3.728751640849107*^9, 
  3.728813288035625*^9, 3.728827406615498*^9, 3.72883599588305*^9, 
  3.728836312454925*^9, 3.728843165455805*^9, 3.7288456489078083`*^9, 
  3.733367434077029*^9, 3.792273438878456*^9},
 CellLabel->"Out[197]=",
 CellID->928540201,ExpressionUUID->"ef6c99e0-b9d5-45c6-a200-910077606d9c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SeparateBasis", "[", "AIndex", "]"}], "[", 
  RowBox[{"expr4", ",", 
   RowBox[{"IndexList", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "fred"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "fred"}]}], "}"}]}], "]"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.7287441427280703`*^9, 3.728744147976247*^9}, {
  3.728747929516767*^9, 3.7287479316766567`*^9}},
 CellLabel->
  "In[198]:=",ExpressionUUID->"d7cb9dd0-f551-47bf-a944-0ea92fd31017"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", "a"},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", "b"},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{1, -$CellContext`fred}, $CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {" ", " "},
          {"a", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[-$CellContext`a, -$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728663251297659*^9, 3.728671285562068*^9, 
  3.728674903630513*^9, 3.728743316465641*^9, 3.728744149243312*^9, 
  3.7287479343098383`*^9, 3.728749596722316*^9, 3.728749952575232*^9, 
  3.728750819779779*^9, 3.728751106389082*^9, 3.728751641016313*^9, 
  3.728813288144209*^9, 3.728827406872306*^9, 3.728835996032954*^9, 
  3.7288363126231127`*^9, 3.7288431656291237`*^9, 3.728845649036478*^9, 
  3.733367434226169*^9, 3.7922734395185328`*^9},
 CellLabel->"Out[198]=",
 CellID->1530963271,ExpressionUUID->"e68c0503-91a2-45c4-8f9a-78603ababc17"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["1.2.4. ToBasis", "Subsubsection",
 CellChangeTimes->{{3.728673450573042*^9, 
  3.7286734507892427`*^9}},ExpressionUUID->"f4d26f62-4c68-41c1-a3de-\
23359605fc6a"],

Cell[TextData[{
 StyleBox["FreeToBasis[basis][expr,indices]\t\t\t", "DisplayMath"],
 StyleBox["Convert free indices into basis indices\n", "Text"],
 StyleBox["FreeToBasis[basis][expr,indices]\t\t\t", "DisplayMath"],
 StyleBox["Change the basis of pairs of dummies", "Text"],
 StyleBox["\nToBasis[basis][expr,indices]\t\t\t    ", "DisplayMath"],
 StyleBox["Replace the indices in an expression by indices in the given \
basis", "Text"]
}], "DefinitionBox",ExpressionUUID->"49cac7fa-2261-4dc6-a325-a1061ea3c0a4"],

Cell["Safe replacement of basis indices.", "Text",ExpressionUUID->"d3bf3c50-64a0-4062-a9c8-417ec7e88489"],

Cell[CellGroupData[{

Cell["\<\
We may want to replace the indices in an expression by indices in a different \
basis.  Sometimes this is easy enough:\
\>", "MathCaption",
 CellChangeTimes->{
  3.728663267067759*^9},ExpressionUUID->"e3fdd907-e478-40d1-ab6c-\
1cbc7e57fee9"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SeparateBasis", "[", "fred", "]"}], "[", 
  RowBox[{
   RowBox[{"v", "[", "a", "]"}], 
   RowBox[{"v", "[", 
    RowBox[{"-", "a"}], "]"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728744172143615*^9, 3.728744172830991*^9}, {
  3.728747955885948*^9, 3.728747956346986*^9}},
 CellLabel->
  "In[199]:=",ExpressionUUID->"7794ffa4-c11e-4cc8-b842-54c07da3ad4d"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["b",
            FontColor->RGBColor[1., 0., 0.]]},
          {"a", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[-$CellContext`a, {$CellContext`b, $CellContext`fred}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", "a"},
          {
           StyleBox["c",
            FontColor->RGBColor[1., 0., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{-$CellContext`c, -$CellContext`fred}, $CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {
           StyleBox["b",
            FontColor->RGBColor[1., 0., 0.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{-$CellContext`b, -$CellContext`fred}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {
           StyleBox["c",
            FontColor->RGBColor[1., 0., 0.]]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{$CellContext`c, $CellContext`fred}],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728663268398988*^9, 3.7286712889363127`*^9, 
  3.728674903758086*^9, 3.728743316636869*^9, 3.7287441739171133`*^9, 
  3.728747956779812*^9, 3.728749600002178*^9, 3.728749952767283*^9, 
  3.728750834261421*^9, 3.7287511145635157`*^9, 3.728751641150243*^9, 
  3.7288132882548723`*^9, 3.728827407163886*^9, 3.728835996182344*^9, 
  3.728836312749338*^9, 3.728843171928685*^9, 3.728845649202405*^9, 
  3.733367442633747*^9, 3.7922734474520607`*^9},
 CellLabel->"Out[199]=",
 CellID->1009678446,ExpressionUUID->"8aad4887-a5c7-46e9-86ee-cb29054b7018"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ContractBasis"}]], "Input",
 CellLabel->
  "In[200]:=",ExpressionUUID->"4e8f5d8c-8453-45f7-b668-f4c988a37246"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {
           StyleBox["a",
            FontColor->RGBColor[1., 0., 0.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{-$CellContext`a, -$CellContext`fred}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {
           StyleBox["a",
            FontColor->RGBColor[1., 0., 0.]]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{$CellContext`a, $CellContext`fred}],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728663269350452*^9, 3.7286712890869102`*^9, 
  3.728674903888303*^9, 3.728743316775115*^9, 3.728744181876185*^9, 
  3.728747958060308*^9, 3.728749600167247*^9, 3.728749952999701*^9, 
  3.728750834429264*^9, 3.7287511147459106`*^9, 3.728751641283783*^9, 
  3.728813288353423*^9, 3.728827407304269*^9, 3.728835996330565*^9, 
  3.728836312877108*^9, 3.7288431721407337`*^9, 3.7288456493852863`*^9, 
  3.733367442792952*^9, 3.792273447515333*^9},
 CellLabel->"Out[200]=",
 CellID->249863328,ExpressionUUID->"7e3607ea-24a5-4bb0-b76e-5105cba73936"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "And it may seem that the ",
 StyleBox["ReplaceIndex", "MR"],
 " function could take care of the more difficult situations. But we must \
watch for pitfalls such as:"
}], "MathCaption",
 CellChangeTimes->{
  3.728663274899472*^9},ExpressionUUID->"4b6b7cbd-5696-4509-b12e-\
a02c39c81ccc"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ReplaceIndex", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"v", "[", "a", "]"}], 
    RowBox[{
     RowBox[{"PD", "[", 
      RowBox[{"-", "b"}], "]"}], "[", 
     RowBox[{"v", "[", 
      RowBox[{"-", "a"}], "]"}], "]"}]}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"a", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"a", ",", "fred"}], "}"}]}], ",", 
     RowBox[{
      RowBox[{"-", "a"}], "\[Rule]", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "a"}], ",", 
        RowBox[{"-", "fred"}]}], "}"}]}]}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.72874419042202*^9, 3.7287441941264057`*^9}, {
  3.728747960276929*^9, 3.7287479625393953`*^9}},
 CellLabel->
  "In[201]:=",ExpressionUUID->"138037b0-7585-4eb4-bb77-896672055746"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {
           StyleBox["a",
            FontColor->RGBColor[1., 0., 0.]]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{$CellContext`a, $CellContext`fred}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "b"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {
            StyleBox["a",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`b][
    $CellContext`v[{-$CellContext`a, -$CellContext`fred}]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728663276216927*^9, 3.7286712892745953`*^9, 
  3.728674904003986*^9, 3.7287433169111433`*^9, 3.728744195703588*^9, 
  3.728747963161359*^9, 3.728749600333692*^9, 3.7287499531212397`*^9, 
  3.7287508345961514`*^9, 3.728751114931621*^9, 3.7287516414366198`*^9, 
  3.728813288470952*^9, 3.728827407437394*^9, 3.7288359964909983`*^9, 
  3.728836313015564*^9, 3.7288431723579273`*^9, 3.728845649568321*^9, 
  3.7333674429111958`*^9, 3.79227344756875*^9},
 CellLabel->"Out[201]=",
 CellID->1122423811,ExpressionUUID->"729c95e3-047f-4326-8ca0-d6c20cd22a3d"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "This is wrong, because\n\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SuperscriptBox["v", "b"], 
     RowBox[{
      SubscriptBox["\[Del]", "a"], " ", 
      SubscriptBox["v", "b"]}]}], " ", "=", " ", 
    RowBox[{"(", 
     RowBox[{
      SuperscriptBox["v", 
       StyleBox["c",
        FontColor->RGBColor[1, 0, 0]]], 
      SubscriptBox[
       SuperscriptBox["e", 
        RowBox[{" ", "b"}]], 
       StyleBox["c",
        FontColor->RGBColor[1, 0, 0]]]}]}]}], TraditionalForm]],
  ExpressionUUID->"3d5b87a0-8cc7-40d8-81b6-e02df9114407"],
 ") ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["\[Del]", "a"], 
     RowBox[{"(", 
      RowBox[{
       SuperscriptBox[
        SubscriptBox["e", "b"], 
        StyleBox["d",
         FontColor->RGBColor[1, 0, 0]]], 
       SubscriptBox["v", 
        StyleBox["d",
         FontColor->RGBColor[1, 0, 0]]]}], ")"}]}], " ", "\[NotEqual]", " "}],
    TraditionalForm]],ExpressionUUID->
  "83f19831-f621-4d84-8b4c-570de93bc683"],
 Cell[BoxData[
  FormBox[
   RowBox[{
    SuperscriptBox["v", 
     StyleBox["b",
      FontColor->RGBColor[1, 0, 0]]], 
    RowBox[{
     SubscriptBox["\[Del]", "a"], " ", 
     SubscriptBox["v", 
      StyleBox["b",
       FontColor->RGBColor[1, 0, 0]]]}]}], TraditionalForm]],ExpressionUUID->
  "1d79f0c7-f37c-4a15-807c-141c5e9e9eab"]
}], "MathCaption",
 CellChangeTimes->{{3.72866331313833*^9, 
  3.7286633138826647`*^9}},ExpressionUUID->"44236b51-0290-4270-863a-\
bbeeda07f574"],

Cell[CellGroupData[{

Cell[TextData[{
 "We can arrive at the correct result via a roundabout way, with ",
 StyleBox["Contract", "MR"],
 " and ",
 StyleBox["SeparateBasis:", "MR"]
}], "MathCaption",
 CellChangeTimes->{{3.728663300252371*^9, 3.728663305218025*^9}},
 CellID->1618843451,ExpressionUUID->"cabba59a-a17b-4596-930b-c32af474c8d6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SeparateBasis", "[", "fred", "]"}], "[", 
  RowBox[{
   RowBox[{"v", "[", "a", "]"}], 
   RowBox[{
    RowBox[{"PD", "[", 
     RowBox[{"-", "b"}], "]"}], "[", 
    RowBox[{"v", "[", 
     RowBox[{"-", "a"}], "]"}], "]"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728744206678141*^9, 3.728744207595689*^9}, {
  3.728747965996553*^9, 3.728747966466937*^9}},
 CellLabel->
  "In[202]:=",ExpressionUUID->"783dabf5-b623-4e01-b026-a817ba1b992c"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", "a"},
          {
           StyleBox["c",
            FontColor->RGBColor[1., 0., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{-$CellContext`c, -$CellContext`fred}, $CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {
           StyleBox["c",
            FontColor->RGBColor[1., 0., 0.]]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{$CellContext`c, $CellContext`fred}],
   Editable->False], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["e",
               FontColor->RGBColor[1., 0., 0.]]},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[-$CellContext`b, {$CellContext`e, $CellContext`fred}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["d",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["e",
               FontColor->RGBColor[1., 0., 0.]], "a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelPDfred[{$CellContext`d, $CellContext`fred}, \
{-$CellContext`e, -$CellContext`fred}, -$CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {
              StyleBox["d",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{-$CellContext`d, -$CellContext`fred}],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["d",
               FontColor->RGBColor[1., 0., 0.]]},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[-$CellContext`a, {$CellContext`d, $CellContext`fred}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["f",
               FontColor->RGBColor[1., 0., 0.]]},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[-$CellContext`b, {$CellContext`f, $CellContext`fred}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", 
         StyleBox["f",
          FontColor->RGBColor[1., 0., 0.]]], GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["d",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[{-$CellContext`f, -$CellContext`fred}][
       $CellContext`v[{-$CellContext`d, -$CellContext`fred}]],
      Editable->False]}]}], ")"}]}]], "Output",
 CellChangeTimes->{3.728663324345949*^9, 3.728671289413801*^9, 
  3.728674904169784*^9, 3.728743317070236*^9, 3.7287479668325233`*^9, 
  3.728749600955299*^9, 3.728749953251396*^9, 3.7287508347939453`*^9, 
  3.728751115397202*^9, 3.7287516415838137`*^9, 3.728813288602961*^9, 
  3.728827407568215*^9, 3.728835996639628*^9, 3.7288363131776123`*^9, 
  3.728843172713995*^9, 3.72884564971418*^9, 3.733367443027567*^9, 
  3.792273447631373*^9},
 CellLabel->"Out[202]=",
 CellID->1080636229,ExpressionUUID->"dfa840b2-95ad-440d-a0a5-12df50654ada"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ContractBasis", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[203]:=",ExpressionUUID->"8af79e68-da8d-42a9-9a27-7a792b8acd3b"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         StyleBox["\[ScriptCapitalD]",
          FontColor->RGBColor[1., 0., 0.]], "]"}], 
        StyleBox[GridBox[{
           {
            StyleBox["c",
             FontColor->RGBColor[1., 0., 0.]], " ", " "},
           {" ", "b", 
            StyleBox["a",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelPDfred[{$CellContext`c, $CellContext`fred}, \
-$CellContext`b, {-$CellContext`a, -$CellContext`fred}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {
            StyleBox["a",
             FontColor->RGBColor[1., 0., 0.]]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[{$CellContext`a, $CellContext`fred}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {
            StyleBox["c",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[{-$CellContext`c, -$CellContext`fred}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {
            StyleBox["a",
             FontColor->RGBColor[1., 0., 0.]]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[{$CellContext`a, $CellContext`fred}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["a",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`v[{-$CellContext`a, -$CellContext`fred}]],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728663325096855*^9, 3.728671289562372*^9, 
  3.7286749042934*^9, 3.728743317226448*^9, 3.728747969546893*^9, 
  3.728749601118381*^9, 3.7287499533843*^9, 3.728750834942754*^9, 
  3.728751115800832*^9, 3.7287516417497263`*^9, 3.72881328875673*^9, 
  3.728827407712874*^9, 3.728835996785057*^9, 3.7288363133588667`*^9, 
  3.728843173555442*^9, 3.728845649859317*^9, 3.7333674431240797`*^9, 
  3.7922734476879263`*^9},
 CellLabel->"Out[203]=",
 CellID->983898124,ExpressionUUID->"bfdbadb0-90c3-42eb-b57f-14d63fcc6667"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The functions *ToBasis automate this process:", "MathCaption",
 CellChangeTimes->{
  3.728663329424713*^9},ExpressionUUID->"9b0daaf3-a596-4e65-b3ea-\
140a1c50b392"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"DummyToBasis", "[", "fred", "]"}], "[", 
   RowBox[{
    RowBox[{"v", "[", "a", "]"}], 
    RowBox[{
     RowBox[{"PD", "[", 
      RowBox[{"-", "b"}], "]"}], "[", 
     RowBox[{"v", "[", 
      RowBox[{"-", "a"}], "]"}], "]"}]}], "]"}], "//", 
  "Simplification"}]], "Input",
 CellChangeTimes->{{3.728744219148332*^9, 3.728744230688871*^9}, {
  3.728747977012287*^9, 3.728747977930249*^9}},
 CellLabel->
  "In[204]:=",ExpressionUUID->"a0c224ae-9878-4336-834a-43c7737e4589"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         StyleBox["\[ScriptCapitalD]",
          FontColor->RGBColor[1., 0., 0.]], "]"}], 
        StyleBox[GridBox[{
           {
            StyleBox["a",
             FontColor->RGBColor[1., 0., 0.]], " ", " "},
           {" ", "b", 
            StyleBox["c",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelPDfred[{$CellContext`a, $CellContext`fred}, \
-$CellContext`b, {-$CellContext`c, -$CellContext`fred}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {
            StyleBox["a",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[{-$CellContext`a, -$CellContext`fred}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {
            StyleBox["c",
             FontColor->RGBColor[1., 0., 0.]]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[{$CellContext`c, $CellContext`fred}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {
            StyleBox["a",
             FontColor->RGBColor[1., 0., 0.]]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[{$CellContext`a, $CellContext`fred}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["a",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`v[{-$CellContext`a, -$CellContext`fred}]],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{
  3.728663331211638*^9, 3.72867128971108*^9, 3.728674904428852*^9, 
   3.7287433174118834`*^9, {3.728744221307715*^9, 3.728744231391253*^9}, 
   3.7287479789473953`*^9, 3.7287496016045017`*^9, 3.728749953487886*^9, 
   3.7287508350809593`*^9, 3.728751115973569*^9, 3.728751641871139*^9, 
   3.7288132889060707`*^9, 3.7288274078582478`*^9, 3.728835996943076*^9, 
   3.7288363135422173`*^9, 3.728843174111834*^9, 3.728845649994981*^9, 
   3.733367443348596*^9, 3.792273447752658*^9},
 CellLabel->"Out[204]=",
 CellID->751688168,ExpressionUUID->"407abdcd-be0b-4152-8eb1-13fa5e644893"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Notice how the derivative of the Basis object has been replaced by the \
appropriate Christoffel. Further examples:\
\>", "MathCaption",ExpressionUUID->"b8cb3dc7-e44f-403f-91ef-f6e1b1b75d23"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr", "=", 
  RowBox[{
   RowBox[{"S", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", "B"}], "]"}], 
   RowBox[{"v", "[", "a", "]"}], 
   RowBox[{"v", "[", "c", "]"}], 
   RowBox[{"v", "[", 
    RowBox[{"-", "c"}], "]"}]}]}]], "Input",
 CellLabel->
  "In[205]:=",ExpressionUUID->"35176443-6dda-4458-b171-3ebd84f30d7d"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {" ", "B"},
          {"A", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[-$CellContext`A, $CellContext`B],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"a"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`c],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"c"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`c],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7286633383281384`*^9, 3.728671289849052*^9, 
  3.728674904559095*^9, 3.72874331756048*^9, 3.7287442395271673`*^9, 
  3.728744285539342*^9, 3.728747981883909*^9, 3.72874960180538*^9, 
  3.728749953618031*^9, 3.728750835234326*^9, 3.728751116175981*^9, 
  3.728751642005293*^9, 3.728813289037627*^9, 3.728827407983973*^9, 
  3.728835997083041*^9, 3.7288363136792707`*^9, 3.72884317432362*^9, 
  3.7288456501389523`*^9, 3.733367443666719*^9, 3.792273447802779*^9},
 CellLabel->"Out[205]=",
 CellID->2124227922,ExpressionUUID->"1a803500-a9e0-467b-9fb9-ba5de17fe454"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"FreeToBasis", "[", "blues", "]"}], "[", "%", "]"}]], "Input",
 CellChangeTimes->{{3.728744242828311*^9, 3.7287442436116247`*^9}, {
   3.7287479833799257`*^9, 3.7287479840899563`*^9}, 3.728749330082061*^9},
 CellLabel->
  "In[206]:=",ExpressionUUID->"3f141297-d6f5-49e2-b848-0750c29390f3"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {" ", "B"},
          {"A", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[-$CellContext`A, $CellContext`B],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`c],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"c"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`c],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {
           StyleBox["a",
            FontColor->RGBColor[0., 0., 1.]]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{$CellContext`a, $CellContext`blues}],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.72866333901333*^9, 3.728671290481298*^9, 
  3.728674904693162*^9, 3.728743317686653*^9, 3.728744244524535*^9, 
  3.728744286718424*^9, 3.728747984581665*^9, 3.728749601978895*^9, 
  3.7287499537506866`*^9, 3.728750835764901*^9, 3.728751116342368*^9, 
  3.728751642138258*^9, 3.728813289159987*^9, 3.728827408146447*^9, 
  3.728835997238*^9, 3.728836313846559*^9, 3.728843174541082*^9, 
  3.72884565027493*^9, 3.733367443859282*^9, 3.792273447862529*^9},
 CellLabel->"Out[206]=",
 CellID->2130537326,ExpressionUUID->"a716f73a-252b-4341-aa72-574323edbaad"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToBasis", "[", "blues", "]"}], "[", "%", "]"}]], "Input",
 CellChangeTimes->{{3.728744247186451*^9, 3.7287442483061237`*^9}, {
   3.728747986155675*^9, 3.7287479868339443`*^9}, 3.728749330763679*^9},
 CellLabel->
  "In[207]:=",ExpressionUUID->"42648f61-fcd0-42eb-b575-711558801ea7"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {" ", "B"},
          {"A", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[-$CellContext`A, $CellContext`B],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {
           StyleBox["a",
            FontColor->RGBColor[0., 0., 1.]]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{$CellContext`a, $CellContext`blues}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{-$CellContext`b, -$CellContext`blues}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{$CellContext`b, $CellContext`blues}],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728663339607588*^9, 3.728671290974699*^9, 
  3.7286749048220863`*^9, 3.728743317867764*^9, 3.728744249001994*^9, 
  3.728744287902669*^9, 3.7287479872838306`*^9, 3.728749602456056*^9, 
  3.728749953872786*^9, 3.728750835943674*^9, 3.728751116914106*^9, 
  3.728751642281556*^9, 3.728813289276197*^9, 3.728827408297968*^9, 
  3.728835997382346*^9, 3.728836313998402*^9, 3.728843174715681*^9, 
  3.728845650396686*^9, 3.733367444004942*^9, 3.792273447924224*^9},
 CellLabel->"Out[207]=",
 CellID->798508004,ExpressionUUID->"d406b170-b4d6-4948-9f77-ac39b1200ec7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"FreeToBasis", "[", "fcomp", "]"}], "[", "%", "]"}]], "Input",
 CellChangeTimes->{
  3.728744252554297*^9, {3.728747990897737*^9, 3.7287479920016623`*^9}},
 CellLabel->
  "In[208]:=",ExpressionUUID->"c51abc58-6e66-4aad-a733-e35a4def8f49"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["B",
            FontColor->RGBColor[0., 1., 0.]]},
          {
           StyleBox["A",
            FontColor->RGBColor[0., 1., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[{-$CellContext`A, -$CellContext`fcomp}, {$CellContext`B, \
$CellContext`fcomp}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {
           StyleBox["a",
            FontColor->RGBColor[0., 0., 1.]]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{$CellContext`a, $CellContext`blues}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{-$CellContext`b, -$CellContext`blues}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{$CellContext`b, $CellContext`blues}],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728663340208885*^9, 3.728671291441922*^9, 
  3.728674904961482*^9, 3.728743318007451*^9, 3.728744253455823*^9, 
  3.7287442887864523`*^9, 3.728747992585712*^9, 3.7287496029607058`*^9, 
  3.728749953999362*^9, 3.728750836107617*^9, 3.7287511172819767`*^9, 
  3.728751642387649*^9, 3.728813289389207*^9, 3.728827408482785*^9, 
  3.728835997523992*^9, 3.728836314132392*^9, 3.728843175464389*^9, 
  3.728845650542616*^9, 3.733367444127673*^9, 3.792273447984149*^9},
 CellLabel->"Out[208]=",
 CellID->898713125,ExpressionUUID->"de3382db-0fd3-4069-8170-f505c22e79a3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToBasis", "[", "fred", "]"}], "[", 
  RowBox[{"%", ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "blues"}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728744258994485*^9, 3.728744281297986*^9}, {
   3.728747995169649*^9, 3.7287479983373117`*^9}, 3.7287493315193367`*^9},
 CellLabel->
  "In[209]:=",ExpressionUUID->"510cb4de-9c2e-4923-8076-a8d443655978"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["B",
            FontColor->RGBColor[0., 1., 0.]]},
          {
           StyleBox["A",
            FontColor->RGBColor[0., 1., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[{-$CellContext`A, -$CellContext`fcomp}, {$CellContext`B, \
$CellContext`fcomp}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {
           StyleBox["a",
            FontColor->RGBColor[1., 0., 0.]]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{$CellContext`a, $CellContext`fred}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{-$CellContext`b, -$CellContext`blues}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{$CellContext`b, $CellContext`blues}],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728663340774884*^9, 3.728671291708797*^9, 
  3.728674905089025*^9, 3.7287433181805677`*^9, 3.728744289826075*^9, 
  3.72874799879843*^9, 3.728749603102792*^9, 3.728749954150195*^9, 
  3.728750836271646*^9, 3.7287511174833593`*^9, 3.7287516425499077`*^9, 
  3.72881328949475*^9, 3.728827408658145*^9, 3.728835997665015*^9, 
  3.7288363142763033`*^9, 3.728843175646999*^9, 3.7288456506897993`*^9, 
  3.733367444383091*^9, 3.792273448033102*^9},
 CellLabel->"Out[209]=",
 CellID->1419002524,ExpressionUUID->"26885443-f6cf-4901-a218-565f8431b25f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Many times we have expressions such as:", "MathCaption",
 CellChangeTimes->{
  3.728663351911763*^9},ExpressionUUID->"4d150a0a-98fe-44f6-9332-\
b686ccab10ca"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"T", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "fred"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "fred"}]}], "}"}]}], "]"}], 
  RowBox[{"v", "[", 
   RowBox[{"{", 
    RowBox[{"a", ",", "fred"}], "}"}], "]"}], 
  RowBox[{"v", "[", 
   RowBox[{"{", 
    RowBox[{"b", ",", "fred"}], "}"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.728744296984803*^9, 3.728744307318844*^9}, {
  3.728748001139216*^9, 3.72874800704103*^9}},
 CellLabel->
  "In[210]:=",ExpressionUUID->"d030aa1d-3ed5-4714-a87c-a14bd742e5c3"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {" ", " "},
          {
           StyleBox["a",
            FontColor->RGBColor[1., 0., 0.]], 
           StyleBox["b",
            FontColor->RGBColor[1., 0., 0.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[{-$CellContext`a, -$CellContext`fred}, {-$CellContext`b, \
-$CellContext`fred}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {
           StyleBox["a",
            FontColor->RGBColor[1., 0., 0.]]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{$CellContext`a, $CellContext`fred}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {
           StyleBox["b",
            FontColor->RGBColor[1., 0., 0.]]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{$CellContext`b, $CellContext`fred}],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728663343195956*^9, 3.72867129184632*^9, 
  3.7286749051993027`*^9, 3.7287433183247137`*^9, 3.728744308050682*^9, 
  3.728748007450532*^9, 3.7287496036719227`*^9, 3.728749954272884*^9, 
  3.728750836763976*^9, 3.728751117807726*^9, 3.7287516427002172`*^9, 
  3.728813289611944*^9, 3.728827408793488*^9, 3.728835997821952*^9, 
  3.7288363143993073`*^9, 3.728843175875133*^9, 3.728845650823798*^9, 
  3.7333674446438503`*^9, 3.792273448092935*^9},
 CellLabel->"Out[210]=",
 CellID->1369520933,ExpressionUUID->"1c8bea4e-fe21-48fc-a0e8-eb5dff17bbd8"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "where we had defined ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["T", "ab"], TraditionalForm]],ExpressionUUID->
  "29d053c9-9b27-4517-8399-d8b48b20d009"],
 " to be antisymmetric. Therefore this is zero:"
}], "MathCaption",ExpressionUUID->"7805f1c1-5858-47c3-8ad5-c4afd7924bf9"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Simplification", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[211]:=",ExpressionUUID->"a0fbd442-4b22-4416-b9dd-3b93f462f04e"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.7286633461257887`*^9, 3.728671291989855*^9, 
  3.7286749053549957`*^9, 3.728743318486854*^9, 3.728744311855497*^9, 
  3.7287480091031227`*^9, 3.728749603973624*^9, 3.728749954367857*^9, 
  3.7287508370656853`*^9, 3.728751118114162*^9, 3.7287516428195343`*^9, 
  3.728813289771603*^9, 3.7288274089510193`*^9, 3.728835997965558*^9, 
  3.728836314531395*^9, 3.728843176074349*^9, 3.728845650953027*^9, 
  3.733367444744834*^9, 3.792273448144682*^9},
 CellLabel->"Out[211]=",
 CellID->808592385,ExpressionUUID->"1de95ba4-e756-40e8-9ab8-488ef849c095"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["However, if we have:", "MathCaption",
 CellChangeTimes->{
  3.7286633494078493`*^9},ExpressionUUID->"4f19cd8f-8d6a-4354-a8ec-\
b69554592185"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"v", "[", 
    RowBox[{"-", "a"}], "]"}], 
   RowBox[{"v", "[", "a", "]"}]}], "-", 
  RowBox[{
   RowBox[{"v", "[", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "fred"}]}], "}"}], "]"}], 
   RowBox[{"v", "[", 
    RowBox[{"{", 
     RowBox[{"a", ",", "fred"}], "}"}], "]"}]}]}]], "Input",
 CellChangeTimes->{{3.728744317149809*^9, 3.72874431965417*^9}, {
  3.728748011266757*^9, 3.728748013496834*^9}},
 CellLabel->
  "In[212]:=",ExpressionUUID->"cae7f8f3-de97-45a6-8b50-bf5248677cc3"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {
            StyleBox["a",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[{-$CellContext`a, -$CellContext`fred}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {
            StyleBox["a",
             FontColor->RGBColor[1., 0., 0.]]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[{$CellContext`a, $CellContext`fred}],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728663354524167*^9, 3.728671292128254*^9, 
  3.7286749054754543`*^9, 3.728743318625399*^9, 3.728744322419191*^9, 
  3.7287480140381527`*^9, 3.728749604157895*^9, 3.7287499545153008`*^9, 
  3.728750837245163*^9, 3.728751118431095*^9, 3.728751642955398*^9, 
  3.728813289889471*^9, 3.728827409086697*^9, 3.728835998100047*^9, 
  3.728836314693761*^9, 3.7288431767184563`*^9, 3.728845651090139*^9, 
  3.7333674448775673`*^9, 3.792273448202237*^9},
 CellLabel->"Out[212]=",
 CellID->735696479,ExpressionUUID->"7821570f-f7bd-4ff2-8c62-c7420f0d720e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Simplification", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[213]:=",ExpressionUUID->"ecb5c111-4734-47cb-8cb6-9909b5fb5b65"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {
            StyleBox["a",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[{-$CellContext`a, -$CellContext`fred}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {
            StyleBox["a",
             FontColor->RGBColor[1., 0., 0.]]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[{$CellContext`a, $CellContext`fred}],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728663355227797*^9, 3.72867129229046*^9, 
  3.7286749056107397`*^9, 3.728743318855324*^9, 3.7287443246723003`*^9, 
  3.728748014850709*^9, 3.7287496045068693`*^9, 3.728749954659195*^9, 
  3.728750837800799*^9, 3.728751118565621*^9, 3.7287516430870934`*^9, 
  3.728813290010865*^9, 3.728827409199092*^9, 3.7288359982530003`*^9, 
  3.728836314830311*^9, 3.7288431769908867`*^9, 3.7288456512436934`*^9, 
  3.733367445161484*^9, 3.792273448255024*^9},
 CellLabel->"Out[213]=",
 CellID->1787366774,ExpressionUUID->"d058777a-d008-455f-a1c9-d37a46a3f4c6"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
we have to change the whole expression to a single basis first:\
\>", "MathCaption",
 CellChangeTimes->{
  3.72866335789552*^9},ExpressionUUID->"adc861ab-91dc-4aa3-b123-da287061da6e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToBasis", "[", "AIndex", "]"}], "[", "%", "]"}]], "Input",
 CellLabel->
  "In[214]:=",ExpressionUUID->"c5d8897b-1b09-4de0-a41a-8be3149078ba"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728663359455875*^9, 3.7286712924236727`*^9, 
  3.728674905740296*^9, 3.728743318980095*^9, 3.728744328816621*^9, 
  3.7287480174612703`*^9, 3.728749604773488*^9, 3.7287499547565203`*^9, 
  3.728750837998012*^9, 3.728751118731019*^9, 3.728751643219501*^9, 
  3.7288132901236973`*^9, 3.728827409323389*^9, 3.728835998384478*^9, 
  3.728836314961362*^9, 3.728843177174057*^9, 3.7288456514923973`*^9, 
  3.733367445291747*^9, 3.792273448304736*^9},
 CellLabel->"Out[214]=",
 CellID->1272116382,ExpressionUUID->"4f83d7d0-375e-40dc-a8e8-6d755dcc56d1"]
}, Open  ]]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["1.3. Tracing contractions", "Subsection",
 CellChangeTimes->{
  3.728663810590992*^9, {3.7286734186465273`*^9, 3.72867342989332*^9}},
 CellID->431306534,ExpressionUUID->"991f8040-8df9-46b6-8ade-ffc5320a5f72"],

Cell[TextData[{
 "One of the most important functions of ",
 StyleBox["xCoba` ", "MR"],
 " is ",
 StyleBox["TraceBasisDummy", "MR"],
 ". We already know how to express a tensor on a given basis, as a \
contraction with ",
 StyleBox["Basis", "MR"],
 " objects. With this new function, we can expand that contraction as a sum \
of tensor components."
}], "Text",
 CellChangeTimes->{{3.728669850409487*^9, 3.728669854356661*^9}},
 CellID->1211608313,ExpressionUUID->"2df945df-a6d7-48c6-b307-6b14d39e4deb"],

Cell[TextData[{
 StyleBox["TraceBasisDummy\t\t\t", "DisplayMath"],
 StyleBox["Expand dummy basis indices into their coordinate ranges", "Text"]
}], "DefinitionBox",
 CellID->1060975737,ExpressionUUID->"f6d10f94-e232-4aa3-b888-8933390cc274"],

Cell["Tracing basis contractions.", "Text",
 CellID->1262490718,ExpressionUUID->"7e4350b2-9c7c-4a3b-b021-569d2d7eeacd"],

Cell[TextData[{
 "Again, there is complete control over which dummies should be expanded. The \
complete syntax is ",
 StyleBox["TraceBasisDummy[expr, indices]", "MR"],
 ", where indices is one of the following:\n\t- A single ",
 StyleBox["BIndex", "MR"],
 ".\n\t- A list of basis indices with head Indices.\n\t- A vbundle (all basis \
dummies belonging to it will we expanded).\n\t- A basis \t"
}], "Text",
 CellID->1594783279,ExpressionUUID->"1cc9e05a-d0e6-4d57-9f23-19b2e24166b1"],

Cell[CellGroupData[{

Cell["This is the simplest example:", "MathCaption",
 CellChangeTimes->{{3.728669865275415*^9, 3.7286698923064413`*^9}},
 CellID->471215995,ExpressionUUID->"f45ab70c-e978-404d-876f-9767a6147762"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TraceBasisDummy", "[", 
  RowBox[{
   RowBox[{"v", "[", 
    RowBox[{"{", 
     RowBox[{"a", ",", "blues"}], "}"}], "]"}], 
   RowBox[{"v", "[", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "blues"}]}], "}"}], "]"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.7287480300881863`*^9, 3.7287480329920893`*^9}, {
  3.728749332416678*^9, 3.728749332986751*^9}},
 CellLabel->"In[215]:=",
 CellID->1543065066,ExpressionUUID->"4cfad6c5-a962-4aa3-82c5-805a365eb0ce"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[{0, -$CellContext`blues}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[{0, $CellContext`blues}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[{1, -$CellContext`blues}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[{1, $CellContext`blues}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[{2, -$CellContext`blues}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[{2, $CellContext`blues}],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{{3.728669868637611*^9, 3.7286698945172567`*^9}, 
   3.7286713038879967`*^9, 3.728674905862668*^9, 3.728743319126711*^9, 
   3.728748033474265*^9, 3.728749607314517*^9, 3.728749954903339*^9, 
   3.728750844499654*^9, 3.7287511291724777`*^9, 3.728751643367936*^9, 
   3.728813290239232*^9, 3.728827409463421*^9, 3.728835998527288*^9, 
   3.728836315098544*^9, 3.728843184946677*^9, 3.728845651886425*^9, 
   3.733367453125596*^9, 3.792273463853486*^9},
 CellLabel->"Out[215]=",
 CellID->1483245410,ExpressionUUID->"f134bd0a-8241-43b0-9978-3458cc23a22f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Further examples with a more complicated expression:", "MathCaption",
 CellChangeTimes->{{3.7286699000097322`*^9, 3.72866990968178*^9}},
 CellID->75238861,ExpressionUUID->"0f268c5c-6ec9-49fb-aa5e-f7edbac73476"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"traceexpr", "=", 
  RowBox[{
   RowBox[{"S", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "A"}], ",", 
       RowBox[{"-", "fcomp"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"A", ",", "fcomp"}], "}"}]}], "]"}], 
   RowBox[{"T", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "fred"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "blues"}]}], "}"}]}], "]"}], 
   RowBox[{"v", "[", 
    RowBox[{"{", 
     RowBox[{"a", ",", "fred"}], "}"}], "]"}], 
   RowBox[{"v", "[", 
    RowBox[{"{", 
     RowBox[{"b", ",", "blues"}], "}"}], "]"}]}]}]], "Input",
 CellChangeTimes->{{3.7287480365679626`*^9, 3.728748055831164*^9}, {
  3.7287493338714247`*^9, 3.7287493344536743`*^9}},
 CellLabel->"In[216]:=",
 CellID->779171946,ExpressionUUID->"b2c72ff6-9813-4ac2-948d-73d07a4eaff0"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["A",
            FontColor->RGBColor[0., 1., 0.]]},
          {
           StyleBox["A",
            FontColor->RGBColor[0., 1., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[{-$CellContext`A, -$CellContext`fcomp}, {$CellContext`A, \
$CellContext`fcomp}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {" ", " "},
          {
           StyleBox["a",
            FontColor->RGBColor[1., 0., 0.]], 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[{-$CellContext`a, -$CellContext`fred}, {-$CellContext`b, \
-$CellContext`blues}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {
           StyleBox["a",
            FontColor->RGBColor[1., 0., 0.]]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{$CellContext`a, $CellContext`fred}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{$CellContext`b, $CellContext`blues}],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7286699108807993`*^9, 3.728671304055625*^9, 
  3.728674906009906*^9, 3.7287433192509613`*^9, 3.728748056325992*^9, 
  3.728749607475677*^9, 3.728749955038563*^9, 3.7287508446956797`*^9, 
  3.728751129359357*^9, 3.7287516435441847`*^9, 3.728813290358902*^9, 
  3.728827409581115*^9, 3.728835998706197*^9, 3.7288363152262383`*^9, 
  3.728843185159169*^9, 3.728845652008081*^9, 3.733367453267687*^9, 
  3.792273463920856*^9},
 CellLabel->"Out[216]=",
 CellID->915431494,ExpressionUUID->"865ea503-f017-4505-81b6-fbd9169f1ef9"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"TraceBasisDummy", "[", "traceexpr", "]"}], "//", 
  "Simplify"}]], "Input",
 CellLabel->"In[217]:=",
 CellID->773194116,ExpressionUUID->"6bdaf458-95d9-43db-bfa8-c9f8513b195c"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox[
              RowBox[{"-", "1"}],
              FontColor->RGBColor[0., 1., 0.]]},
            {
             StyleBox[
              RowBox[{"-", "1"}],
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`S[{-1, -$CellContext`fcomp}, {-1, $CellContext`fcomp}],
     Editable->False], "+", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`S[{1, -$CellContext`fcomp}, {1, $CellContext`fcomp}],
     Editable->False]}], ")"}], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`fred}, {0, -$CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{0, $CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{0, $CellContext`fred}],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`fred}, {1, -$CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{0, $CellContext`fred}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{1, $CellContext`blues}],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`fred}, {0, -$CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{0, $CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{1, $CellContext`fred}],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`fred}, {1, -$CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{1, $CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{1, $CellContext`fred}],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`fred}, {2, -$CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{0, $CellContext`fred}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{2, $CellContext`blues}],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`fred}, {2, -$CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{1, $CellContext`fred}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{2, $CellContext`blues}],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, -$CellContext`fred}, {0, -$CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{0, $CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{2, $CellContext`fred}],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, -$CellContext`fred}, {1, -$CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{1, $CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{2, $CellContext`fred}],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, -$CellContext`fred}, {2, -$CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{2, $CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{2, $CellContext`fred}],
      Editable->False]}]}], ")"}]}]], "Output",
 CellChangeTimes->{3.728669911624845*^9, 3.7286713042470093`*^9, 
  3.7286749061481867`*^9, 3.728743319467375*^9, 3.728748058614851*^9, 
  3.728749607669875*^9, 3.728749955141904*^9, 3.728750844894843*^9, 
  3.728751129562991*^9, 3.7287516436752872`*^9, 3.728813290488751*^9, 
  3.7288274097398157`*^9, 3.728835998878707*^9, 3.728836315412733*^9, 
  3.728843185365019*^9, 3.728845652176749*^9, 3.733367453420726*^9, 
  3.79227346401015*^9},
 CellLabel->"Out[217]=",
 CellID->1084272550,ExpressionUUID->"a37ce3c7-2d08-4ff1-90fe-8641328c07cd"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"TraceBasisDummy", "[", 
   RowBox[{"traceexpr", ",", "TangentM3"}], "]"}], "//", 
  "Simplify"}]], "Input",
 CellLabel->"In[218]:=",
 CellID->1769420663,ExpressionUUID->"a6765f3c-d1d0-4f93-ad97-7e1d954c995f"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["A",
            FontColor->RGBColor[0., 1., 0.]]},
          {
           StyleBox["A",
            FontColor->RGBColor[0., 1., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[{-$CellContext`A, -$CellContext`fcomp}, {$CellContext`A, \
$CellContext`fcomp}],
   Editable->False], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`fred}, {0, -$CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{0, $CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{0, $CellContext`fred}],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`fred}, {1, -$CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{0, $CellContext`fred}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{1, $CellContext`blues}],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`fred}, {0, -$CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{0, $CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{1, $CellContext`fred}],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`fred}, {1, -$CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{1, $CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{1, $CellContext`fred}],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`fred}, {2, -$CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{0, $CellContext`fred}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{2, $CellContext`blues}],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`fred}, {2, -$CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{1, $CellContext`fred}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{2, $CellContext`blues}],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, -$CellContext`fred}, {0, -$CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{0, $CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{2, $CellContext`fred}],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, -$CellContext`fred}, {1, -$CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{1, $CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{2, $CellContext`fred}],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, -$CellContext`fred}, {2, -$CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{2, $CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{2, $CellContext`fred}],
      Editable->False]}]}], ")"}]}]], "Output",
 CellChangeTimes->{3.728669912313346*^9, 3.7286713044208727`*^9, 
  3.7286749063293333`*^9, 3.728743319592259*^9, 3.72874806194628*^9, 
  3.728749608324108*^9, 3.728749955258635*^9, 3.728750845074977*^9, 
  3.728751129765456*^9, 3.72875164393362*^9, 3.72881329064926*^9, 
  3.728827409934431*^9, 3.72883599903423*^9, 3.728836315593276*^9, 
  3.728843185599861*^9, 3.7288456523359118`*^9, 3.733367453559538*^9, 
  3.792273464098386*^9},
 CellLabel->"Out[218]=",
 CellID->1970003055,ExpressionUUID->"8e78d937-5a83-49bc-b6c7-2894f278cc14"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"TraceBasisDummy", "[", 
   RowBox[{"traceexpr", ",", "fred"}], "]"}], "//", "Simplify"}]], "Input",
 CellChangeTimes->{{3.7287480640063667`*^9, 3.728748065558014*^9}},
 CellLabel->"In[219]:=",
 CellID->656775127,ExpressionUUID->"669472bb-3a73-4d05-950e-eb920b8035dd"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["A",
            FontColor->RGBColor[0., 1., 0.]]},
          {
           StyleBox["A",
            FontColor->RGBColor[0., 1., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[{-$CellContext`A, -$CellContext`fcomp}, {$CellContext`A, \
$CellContext`fcomp}],
   Editable->False], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{
       0, -$CellContext`fred}, {-$CellContext`b, -$CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{0, $CellContext`fred}],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{
       1, -$CellContext`fred}, {-$CellContext`b, -$CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{1, $CellContext`fred}],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{
       2, -$CellContext`fred}, {-$CellContext`b, -$CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{2, $CellContext`fred}],
      Editable->False]}]}], ")"}], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{$CellContext`b, $CellContext`blues}],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728669912991357*^9, 3.728671305193616*^9, 
  3.728674906438115*^9, 3.728743319790514*^9, 3.728748066126779*^9, 
  3.7287496085674963`*^9, 3.7287499553489523`*^9, 3.7287508457824183`*^9, 
  3.728751129983696*^9, 3.728751644122727*^9, 3.728813290779911*^9, 
  3.728827410080262*^9, 3.728835999186967*^9, 3.7288363157611103`*^9, 
  3.728843186381065*^9, 3.728845652479891*^9, 3.733367453677747*^9, 
  3.792273464159018*^9},
 CellLabel->"Out[219]=",
 CellID->2085959586,ExpressionUUID->"cde15701-7247-4360-a018-850f6a1ca19d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"TraceBasisDummy", "[", 
   RowBox[{"traceexpr", ",", "blues"}], "]"}], "//", "Simplify"}]], "Input",
 CellChangeTimes->{{3.72874806895855*^9, 3.7287480695500526`*^9}, 
   3.728749335321701*^9},
 CellLabel->"In[220]:=",
 CellID->1101717314,ExpressionUUID->"fa37887c-091e-41cc-821a-5b5f2ef981b4"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["A",
            FontColor->RGBColor[0., 1., 0.]]},
          {
           StyleBox["A",
            FontColor->RGBColor[0., 1., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[{-$CellContext`A, -$CellContext`fcomp}, {$CellContext`A, \
$CellContext`fcomp}],
   Editable->False], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["a",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{-$CellContext`a, -$CellContext`fred}, {
       0, -$CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{0, $CellContext`blues}],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["a",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{-$CellContext`a, -$CellContext`fred}, {
       1, -$CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{1, $CellContext`blues}],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["a",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{-$CellContext`a, -$CellContext`fred}, {
       2, -$CellContext`blues}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{2, $CellContext`blues}],
      Editable->False]}]}], ")"}], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {
           StyleBox["a",
            FontColor->RGBColor[1., 0., 0.]]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[{$CellContext`a, $CellContext`fred}],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728669913641521*^9, 3.7286713054243526`*^9, 
  3.728674906589961*^9, 3.7287433199917316`*^9, 3.728748070257715*^9, 
  3.728749608746283*^9, 3.7287499554828043`*^9, 3.7287508460470457`*^9, 
  3.728751130155991*^9, 3.7287516442956533`*^9, 3.7288132909974613`*^9, 
  3.728827410202561*^9, 3.728835999336035*^9, 3.7288363159032097`*^9, 
  3.728843186736487*^9, 3.7288456526141777`*^9, 3.733367453793283*^9, 
  3.792273464228982*^9},
 CellLabel->"Out[220]=",
 CellID->1052224627,ExpressionUUID->"c1da8ff8-47f2-4cc4-bdb2-8b6e91b4f5fd"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["With derivatives:", "MathCaption",
 CellChangeTimes->{{3.728669916688908*^9, 3.7286699173129053`*^9}},
 CellID->1379951860,ExpressionUUID->"1bd3ccea-d4be-438c-83ce-b0aee7b42768"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TraceBasisDummy", "[", 
  RowBox[{
   RowBox[{"PD", "[", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "fred"}]}], "}"}], "]"}], "[", 
   RowBox[{"v", "[", 
    RowBox[{"{", 
     RowBox[{"b", ",", "fred"}], "}"}], "]"}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.7287480792076063`*^9, 3.728748081373763*^9}},
 CellLabel->"In[221]:=",
 CellID->27523420,ExpressionUUID->"d0e6eeb6-355d-47c0-ac4d-28d98ef14f38"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", 
      StyleBox["0",
       FontColor->RGBColor[1., 0., 0.]]], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[{0, -$CellContext`fred}][
    $CellContext`v[{0, $CellContext`fred}]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", 
      StyleBox["1",
       FontColor->RGBColor[1., 0., 0.]]], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[{1, -$CellContext`fred}][
    $CellContext`v[{1, $CellContext`fred}]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", 
      StyleBox["2",
       FontColor->RGBColor[1., 0., 0.]]], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[{2, -$CellContext`fred}][
    $CellContext`v[{2, $CellContext`fred}]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728669918416731*^9, 3.728671305607786*^9, 
  3.7286749067115927`*^9, 3.728743320153665*^9, 3.728748081796682*^9, 
  3.72874960891945*^9, 3.728749955604347*^9, 3.72875084673388*^9, 
  3.728751130320572*^9, 3.728751644464028*^9, 3.728813291170125*^9, 
  3.728827410320662*^9, 3.728835999480905*^9, 3.728836316060174*^9, 
  3.728843186975065*^9, 3.7288456528623667`*^9, 3.733367453942087*^9, 
  3.792273464284607*^9},
 CellLabel->"Out[221]=",
 CellID->211472944,ExpressionUUID->"f87cbf0a-34ba-4a07-bda3-9a54f355232b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{" ", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     RowBox[{"u", "[", "B", "]"}], "+", 
     RowBox[{
      RowBox[{"S", "[", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{"1", ",", 
          RowBox[{"-", "fcomp"}]}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{"A", ",", "fcomp"}], "}"}]}], "]"}], 
      RowBox[{"S", "[", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{
          RowBox[{"-", "A"}], ",", 
          RowBox[{"-", "fcomp"}]}], "}"}], ",", "B"}], "]"}]}]}], ")"}], "+", 
   " ", 
   RowBox[{
    RowBox[{
     RowBox[{"PD", "[", 
      RowBox[{"{", 
       RowBox[{"A", ",", "fcomp"}], "}"}], "]"}], "[", 
     RowBox[{"S", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "A"}], ",", 
         RowBox[{"-", "fcomp"}]}], "}"}], ",", "B"}], "]"}], "]"}], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"u", "[", 
       RowBox[{"{", 
        RowBox[{"C", ",", "fcomp"}], "}"}], "]"}], "+", 
      RowBox[{
       RowBox[{"S", "[", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           RowBox[{"-", "D"}], ",", 
           RowBox[{"-", "fcomp"}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"C", ",", "fcomp"}], "}"}]}], "]"}], 
       RowBox[{"u", "[", 
        RowBox[{"{", 
         RowBox[{"D", ",", "fcomp"}], "}"}], "]"}]}]}], ")"}], 
    RowBox[{"u", "[", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "C"}], ",", 
       RowBox[{"-", "fcomp"}]}], "}"}], "]"}]}]}]}]], "Input",
 CellChangeTimes->{{3.7287480843816*^9, 3.728748099540729*^9}},
 CellLabel->"In[222]:=",
 CellID->1694225837,ExpressionUUID->"15ae4422-299c-4bbd-b911-ae6871e39793"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["A",
             FontColor->RGBColor[0., 1., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[{
     1, -$CellContext`fcomp}, {$CellContext`A, $CellContext`fcomp}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", "B"},
           {
            StyleBox["A",
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[{-$CellContext`A, -$CellContext`fcomp}, $CellContext`B],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"u", 
       StyleBox[GridBox[{
          {"B"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`u[$CellContext`B],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"u", 
        StyleBox[GridBox[{
           {" "},
           {
            StyleBox["C",
             FontColor->RGBColor[0., 1., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`u[{-C, -$CellContext`fcomp}],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"u", 
          StyleBox[GridBox[{
             {
              StyleBox["C",
               FontColor->RGBColor[0., 1., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[{C, $CellContext`fcomp}],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"S", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["C",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["D",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`S[{-D, -$CellContext`fcomp}, {C, $CellContext`fcomp}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"u", 
           StyleBox[GridBox[{
              {
               StyleBox["D",
                FontColor->RGBColor[0., 1., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[{D, $CellContext`fcomp}],
       Editable->False]}]}], ")"}], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SuperscriptBox["\[PartialD]", 
       StyleBox["A",
        FontColor->RGBColor[0., 1., 0.]]], GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {" ", "B"},
            {
             StyleBox["A",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[{$CellContext`A, $CellContext`fcomp}][
     $CellContext`S[{-$CellContext`A, -$CellContext`fcomp}, $CellContext`B]],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.7286699196712008`*^9, 3.7286713064755077`*^9, 
  3.728674906855814*^9, 3.7287433202992477`*^9, 3.728748100449011*^9, 
  3.728749610023291*^9, 3.728749955753448*^9, 3.728750846975134*^9, 
  3.7287511305472383`*^9, 3.728751644624826*^9, 3.728813291293785*^9, 
  3.728827410441512*^9, 3.7288359996416197`*^9, 3.728836316202324*^9, 
  3.728843187612689*^9, 3.728845653042403*^9, 3.733367454095626*^9, 
  3.792273464356484*^9},
 CellLabel->"Out[222]=",
 CellID->140334023,ExpressionUUID->"baf485b2-cdb8-45aa-876b-0701c23368dd"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TraceBasisDummy", "@", "%"}]], "Input",
 CellLabel->"In[223]:=",
 CellID->459528220,ExpressionUUID->"c016f05a-e9a1-4621-9174-3ee8d407c4eb"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", "B"},
           {
            StyleBox[
             RowBox[{"-", "1"}],
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[{-1, -$CellContext`fcomp}, $CellContext`B],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox[
             RowBox[{"-", "1"}],
             FontColor->RGBColor[0., 1., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[{1, -$CellContext`fcomp}, {-1, $CellContext`fcomp}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", "B"},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[{1, -$CellContext`fcomp}, $CellContext`B],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 1., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[{1, -$CellContext`fcomp}, {1, $CellContext`fcomp}],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"u", 
       StyleBox[GridBox[{
          {"B"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`u[$CellContext`B],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"u", 
        StyleBox[GridBox[{
           {" "},
           {
            StyleBox[
             RowBox[{"-", "1"}],
             FontColor->RGBColor[0., 1., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`u[{-1, -$CellContext`fcomp}],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"u", 
          StyleBox[GridBox[{
             {
              StyleBox[
               RowBox[{"-", "1"}],
               FontColor->RGBColor[0., 1., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[{-1, $CellContext`fcomp}],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"S", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox[
                RowBox[{"-", "1"}],
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox[
                RowBox[{"-", "1"}],
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`S[{-1, -$CellContext`fcomp}, {-1, $CellContext`fcomp}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"u", 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"-", "1"}],
                FontColor->RGBColor[0., 1., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[{-1, $CellContext`fcomp}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"S", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox[
                RowBox[{"-", "1"}],
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`S[{1, -$CellContext`fcomp}, {-1, $CellContext`fcomp}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"u", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[{1, $CellContext`fcomp}],
       Editable->False]}]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{
        SuperscriptBox["\[PartialD]", 
         StyleBox[
          RowBox[{"-", "1"}],
          FontColor->RGBColor[0., 1., 0.]]], GridBox[{
          {"S", 
           StyleBox[GridBox[{
              {" ", "B"},
              {
               StyleBox[
                RowBox[{"-", "1"}],
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[{-1, $CellContext`fcomp}][
       $CellContext`S[{-1, -$CellContext`fcomp}, $CellContext`B]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SuperscriptBox["\[PartialD]", 
         StyleBox["1",
          FontColor->RGBColor[0., 1., 0.]]], GridBox[{
          {"S", 
           StyleBox[GridBox[{
              {" ", "B"},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[{1, $CellContext`fcomp}][
       $CellContext`S[{1, -$CellContext`fcomp}, $CellContext`B]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"u", 
        StyleBox[GridBox[{
           {" "},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 1., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`u[{1, -$CellContext`fcomp}],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"S", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox[
                RowBox[{"-", "1"}],
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`S[{-1, -$CellContext`fcomp}, {1, $CellContext`fcomp}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"u", 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"-", "1"}],
                FontColor->RGBColor[0., 1., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[{-1, $CellContext`fcomp}],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"u", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 1., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[{1, $CellContext`fcomp}],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"S", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`S[{1, -$CellContext`fcomp}, {1, $CellContext`fcomp}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"u", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[{1, $CellContext`fcomp}],
       Editable->False]}]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{
        SuperscriptBox["\[PartialD]", 
         StyleBox[
          RowBox[{"-", "1"}],
          FontColor->RGBColor[0., 1., 0.]]], GridBox[{
          {"S", 
           StyleBox[GridBox[{
              {" ", "B"},
              {
               StyleBox[
                RowBox[{"-", "1"}],
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[{-1, $CellContext`fcomp}][
       $CellContext`S[{-1, -$CellContext`fcomp}, $CellContext`B]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SuperscriptBox["\[PartialD]", 
         StyleBox["1",
          FontColor->RGBColor[0., 1., 0.]]], GridBox[{
          {"S", 
           StyleBox[GridBox[{
              {" ", "B"},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[{1, $CellContext`fcomp}][
       $CellContext`S[{1, -$CellContext`fcomp}, $CellContext`B]],
      Editable->False]}], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728669922252746*^9, 3.728671306677682*^9, 
  3.72867490699127*^9, 3.7287433204502163`*^9, 3.728748102229064*^9, 
  3.728749610223332*^9, 3.728749955899817*^9, 3.7287508472725277`*^9, 
  3.7287511307488623`*^9, 3.728751644752063*^9, 3.7288132914491167`*^9, 
  3.7288274105971403`*^9, 3.728835999797946*^9, 3.7288363163471537`*^9, 
  3.728843187898109*^9, 3.728845653235464*^9, 3.733367454235414*^9, 
  3.792273464442214*^9},
 CellLabel->"Out[223]=",
 CellID->656880155,ExpressionUUID->"da51204e-3115-4125-b770-d20c0df386fa"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Simplify", "[", "%", "]"}]], "Input",
 CellLabel->"In[224]:=",
 CellID->1391926505,ExpressionUUID->"b398098e-6222-4089-acc5-c6ec0bb51ef1"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", "B"},
           {
            StyleBox[
             RowBox[{"-", "1"}],
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[{-1, -$CellContext`fcomp}, $CellContext`B],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox[
             RowBox[{"-", "1"}],
             FontColor->RGBColor[0., 1., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[{1, -$CellContext`fcomp}, {-1, $CellContext`fcomp}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", "B"},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[{1, -$CellContext`fcomp}, $CellContext`B],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 1., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[{1, -$CellContext`fcomp}, {1, $CellContext`fcomp}],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"u", 
       StyleBox[GridBox[{
          {"B"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`u[$CellContext`B],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"u", 
        StyleBox[GridBox[{
           {" "},
           {
            StyleBox[
             RowBox[{"-", "1"}],
             FontColor->RGBColor[0., 1., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`u[{-1, -$CellContext`fcomp}],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{"1", "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"S", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox[
                  RowBox[{"-", "1"}],
                  FontColor->RGBColor[0., 1., 0.]]},
                {
                 StyleBox[
                  RowBox[{"-", "1"}],
                  FontColor->RGBColor[0., 1., 0.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`S[{-1, -$CellContext`fcomp}, {-1, $CellContext`fcomp}],
         Editable->False]}], ")"}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"u", 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"-", "1"}],
                FontColor->RGBColor[0., 1., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[{-1, $CellContext`fcomp}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"S", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox[
                RowBox[{"-", "1"}],
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`S[{1, -$CellContext`fcomp}, {-1, $CellContext`fcomp}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"u", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[{1, $CellContext`fcomp}],
       Editable->False]}]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{
        SuperscriptBox["\[PartialD]", 
         StyleBox[
          RowBox[{"-", "1"}],
          FontColor->RGBColor[0., 1., 0.]]], GridBox[{
          {"S", 
           StyleBox[GridBox[{
              {" ", "B"},
              {
               StyleBox[
                RowBox[{"-", "1"}],
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[{-1, $CellContext`fcomp}][
       $CellContext`S[{-1, -$CellContext`fcomp}, $CellContext`B]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SuperscriptBox["\[PartialD]", 
         StyleBox["1",
          FontColor->RGBColor[0., 1., 0.]]], GridBox[{
          {"S", 
           StyleBox[GridBox[{
              {" ", "B"},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[{1, $CellContext`fcomp}][
       $CellContext`S[{1, -$CellContext`fcomp}, $CellContext`B]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"u", 
        StyleBox[GridBox[{
           {" "},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 1., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`u[{1, -$CellContext`fcomp}],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"S", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox[
                RowBox[{"-", "1"}],
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`S[{-1, -$CellContext`fcomp}, {1, $CellContext`fcomp}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"u", 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"-", "1"}],
                FontColor->RGBColor[0., 1., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[{-1, $CellContext`fcomp}],
       Editable->False]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"1", "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"S", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["1",
                  FontColor->RGBColor[0., 1., 0.]]},
                {
                 StyleBox["1",
                  FontColor->RGBColor[0., 1., 0.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`S[{1, -$CellContext`fcomp}, {1, $CellContext`fcomp}],
         Editable->False]}], ")"}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"u", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[{1, $CellContext`fcomp}],
       Editable->False]}]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{
        SuperscriptBox["\[PartialD]", 
         StyleBox[
          RowBox[{"-", "1"}],
          FontColor->RGBColor[0., 1., 0.]]], GridBox[{
          {"S", 
           StyleBox[GridBox[{
              {" ", "B"},
              {
               StyleBox[
                RowBox[{"-", "1"}],
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[{-1, $CellContext`fcomp}][
       $CellContext`S[{-1, -$CellContext`fcomp}, $CellContext`B]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SuperscriptBox["\[PartialD]", 
         StyleBox["1",
          FontColor->RGBColor[0., 1., 0.]]], GridBox[{
          {"S", 
           StyleBox[GridBox[{
              {" ", "B"},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[{1, $CellContext`fcomp}][
       $CellContext`S[{1, -$CellContext`fcomp}, $CellContext`B]],
      Editable->False]}], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728669923068334*^9, 3.728671306871726*^9, 
  3.728674907141953*^9, 3.7287433205958147`*^9, 3.728748103843605*^9, 
  3.728749610420054*^9, 3.728749956020595*^9, 3.728750847492627*^9, 
  3.7287511309296207`*^9, 3.728751644964299*^9, 3.728813291594194*^9, 
  3.728827410746101*^9, 3.728835999954357*^9, 3.728836316516487*^9, 
  3.728843188206366*^9, 3.728845653397662*^9, 3.733367454414003*^9, 
  3.792273464525296*^9},
 CellLabel->"Out[224]=",
 CellID->2118819505,ExpressionUUID->"0a9150e0-e700-4399-ad72-3049ec6d28d7"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Another example:", "MathCaption",
 CellChangeTimes->{3.728669926944439*^9},
 CellID->1558102094,ExpressionUUID->"71ed5452-40e5-4df4-95c2-e2ea371592e0"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TraceBasisDummy", "[", 
  RowBox[{
   RowBox[{"PDfred", "[", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "blues"}]}], "}"}], "]"}], "[", 
   RowBox[{"v", "[", 
    RowBox[{"{", 
     RowBox[{"a", ",", "blues"}], "}"}], "]"}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.728748113716324*^9, 3.728748123843709*^9}, {
  3.728749336043593*^9, 3.7287493365818863`*^9}},
 CellLabel->"In[225]:=",
 CellID->606302340,ExpressionUUID->"8b967ba1-e237-4ef9-a33f-b37a6e022d0d"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox[
      StyleBox["\[ScriptCapitalD]",
       FontColor->RGBColor[1., 0., 0.]], 
      StyleBox["0",
       FontColor->RGBColor[0., 0., 1.]]], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`PDfred[{0, -$CellContext`blues}][
    $CellContext`v[{0, $CellContext`blues}]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox[
      StyleBox["\[ScriptCapitalD]",
       FontColor->RGBColor[1., 0., 0.]], 
      StyleBox["1",
       FontColor->RGBColor[0., 0., 1.]]], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`PDfred[{1, -$CellContext`blues}][
    $CellContext`v[{1, $CellContext`blues}]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox[
      StyleBox["\[ScriptCapitalD]",
       FontColor->RGBColor[1., 0., 0.]], 
      StyleBox["2",
       FontColor->RGBColor[0., 0., 1.]]], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`PDfred[{2, -$CellContext`blues}][
    $CellContext`v[{2, $CellContext`blues}]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.72866992800522*^9, 3.728671307029174*^9, 
  3.728674907271343*^9, 3.728743320740497*^9, 3.728748124367655*^9, 
  3.72874961059774*^9, 3.7287499561551027`*^9, 3.7287508476599483`*^9, 
  3.72875113111644*^9, 3.7287516455391397`*^9, 3.728813291738801*^9, 
  3.728827410877468*^9, 3.728836000098776*^9, 3.728836316644738*^9, 
  3.7288431889868793`*^9, 3.728845653524334*^9, 3.733367454546537*^9, 
  3.792273464591057*^9},
 CellLabel->"Out[225]=",
 CellID->1425585981,ExpressionUUID->"1440bf6b-660b-41ab-aa26-12381fb44d9e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"SeparateBasis", "[", "AIndex", "]"}], "[", "%", "]"}], "//", 
  "Expand"}]], "Input",
 CellLabel->"In[226]:=",
 CellID->1609809464,ExpressionUUID->"80849fe7-b459-4f78-a114-20adf5e0c1f8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"ChristoffelPDbluesPDfred", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`ChristoffelPDbluesPDfred[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728748126816881*^9, 3.728749610780635*^9, 
  3.728749956271208*^9, 3.7287508478334913`*^9, 3.728751131291375*^9, 
  3.7287516457807302`*^9, 3.728813291833787*^9, 3.728827410997855*^9, 
  3.728836000232377*^9, 3.7288363168077383`*^9, 3.72884318947606*^9, 
  3.728845653651153*^9, 3.733367454680045*^9, 3.7922734646438427`*^9},
 CellLabel->"During evaluation of In[226]:=",
 CellID->348533088,ExpressionUUID->"9401ed46-fc0a-4328-829f-3c4f0645ee71"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "b"},
           {
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`blues}, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[0., 0., 1.]], ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]], " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelPDbluesPDfred[{
     0, $CellContext`blues}, -$CellContext`b, -$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "b"},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`blues}, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[0., 0., 1.]], ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelPDbluesPDfred[{
     1, $CellContext`blues}, -$CellContext`b, -$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "b"},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`blues}, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[0., 0., 1.]], ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
        StyleBox[GridBox[{
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelPDbluesPDfred[{
     2, $CellContext`blues}, -$CellContext`b, -$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`blues}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "b"},
           {
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`blues}, $CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox[
        StyleBox["\[ScriptCapitalD]",
         FontColor->RGBColor[1., 0., 0.]], "b"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDfred[-$CellContext`b][
      $CellContext`v[$CellContext`a]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`blues}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "b"},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`blues}, $CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox[
        StyleBox["\[ScriptCapitalD]",
         FontColor->RGBColor[1., 0., 0.]], "b"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDfred[-$CellContext`b][
      $CellContext`v[$CellContext`a]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`blues}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "b"},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`blues}, $CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox[
        StyleBox["\[ScriptCapitalD]",
         FontColor->RGBColor[1., 0., 0.]], "b"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDfred[-$CellContext`b][
      $CellContext`v[$CellContext`a]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728669928741796*^9, 3.728671307251361*^9, 
  3.7286749074778852`*^9, 3.728743320880674*^9, 3.728748126905306*^9, 
  3.728749610839019*^9, 3.728749956326913*^9, 3.7287508478825817`*^9, 
  3.728751131347896*^9, 3.728751645877923*^9, 3.728813291898107*^9, 
  3.728827411077916*^9, 3.728836000269689*^9, 3.728836316926865*^9, 
  3.7288431895871468`*^9, 3.728845653723381*^9, 3.733367454726541*^9, 
  3.792273464707293*^9},
 CellLabel->"Out[226]=",
 CellID->1607704647,ExpressionUUID->"2dfd0cc7-a164-4cb8-b1d6-8b99cc06e35c"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["1.4. Arrays of components", "Subsection",
 CellChangeTimes->{{3.728670536781797*^9, 3.728670537332263*^9}, {
  3.728671340208577*^9, 3.728671345447023*^9}, {3.728673420750449*^9, 
  3.728673431229848*^9}},
 CellID->307506738,ExpressionUUID->"91c17308-5319-495d-97c0-9bf71d378eba"],

Cell[CellGroupData[{

Cell["1.4.1. ComponentArray and TableOfComponents", "Subsubsection",
 CellChangeTimes->{
  3.728671575550845*^9, {3.728673460212575*^9, 3.728673460324917*^9}},
 CellID->334234630,ExpressionUUID->"256f3d06-bb3e-426b-a9bd-fc1f6559166b"],

Cell[TextData[{
 StyleBox["ComponentArray\t\t\t", "DisplayMath"],
 StyleBox["Generate a list of components \n", "Text"],
 StyleBox["TableOfComponents\t\t", "DisplayMath"],
 StyleBox["Generate a list of components, with many more options", "Text"]
}], "DefinitionBox",
 CellID->1811689505,ExpressionUUID->"6ce642fe-f734-4948-8fd7-f924ee187d73"],

Cell["Arrays of components", "Text",
 CellID->2042363369,ExpressionUUID->"dea6c7aa-d716-456e-b91e-d9cb04e92dd5"],

Cell[TextData[{
 "The commands described in this subsection essentially generate lists of \
c-indices from free b-indices (or free a-indices through ",
 StyleBox["ToBasis", "MR"],
 "). The basic command is ",
 StyleBox["ComponentArray", "MR"],
 ",."
}], "Text",
 CellChangeTimes->{{3.728671364463563*^9, 3.728671365133699*^9}},
 CellID->622529135,ExpressionUUID->"f0d186fb-8c5a-4b75-b6c7-098431a50e20"],

Cell[CellGroupData[{

Cell["\<\
Construct the list of components of a vector in a given basis:\
\>", "MathCaption",
 CellChangeTimes->{{3.7286713701413193`*^9, 3.7286713824292107`*^9}},
 CellID->1942919801,ExpressionUUID->"9a6f8d2c-b8e1-4c5b-aabc-8fca5f369f95"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentArray", "[", 
  RowBox[{"v", "[", 
   RowBox[{"{", 
    RowBox[{"a", ",", "fred"}], "}"}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.728748160581015*^9, 3.728748161082078*^9}},
 CellLabel->"In[227]:=",
 CellID->306843761,ExpressionUUID->"fbe10e4b-8c78-431f-9403-531838f0b8aa"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[{0, $CellContext`fred}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[{1, $CellContext`fred}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[{2, $CellContext`fred}],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.7286713840809526`*^9, 3.7286749076089563`*^9, 
  3.728743320974186*^9, 3.728748161616395*^9, 3.7287496144511547`*^9, 
  3.728749956421143*^9, 3.728751138163239*^9, 3.728751646017804*^9, 
  3.728813292039867*^9, 3.7288274112176247`*^9, 3.728836000418169*^9, 
  3.72883631706439*^9, 3.728843197376691*^9, 3.728845653845398*^9, 
  3.733367462232519*^9, 3.792273471573875*^9},
 CellLabel->"Out[227]=",
 CellID->1698417242,ExpressionUUID->"b3dfb33d-7d85-470f-8a1a-ecc5b9d24bb8"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "NOTE: Up to version 0.6, the role of ",
 StyleBox["ComponentArray", "MR"],
 " was played by a different function, called ",
 StyleBox["Components.", "MR"]
}], "MathCaption",
 CellID->1347482286,ExpressionUUID->"553bb5fe-d911-4ff7-b4ab-8d080b46dd11"],

Cell[CellGroupData[{

Cell["\<\
If the object to expand is a tensor (and only in this case), we can use a \
shorter notation:\
\>", "MathCaption",
 CellChangeTimes->{3.728671396644285*^9},
 CellID->1919723187,ExpressionUUID->"2ed23899-9c44-4230-9e5f-0f4e08aa5f9f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentArray", "[", 
  RowBox[{"T", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "fred"}], ",", 
     RowBox[{"-", "fred"}]}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728748164931493*^9, 3.728748166889673*^9}},
 CellLabel->"In[228]:=",
 CellID->993291381,ExpressionUUID->"2d38aaff-2117-4c3d-95df-c21801dd1d99"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`fred}, {0, -$CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`fred}, {1, -$CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`fred}, {2, -$CellContext`fred}],
      Editable->False]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`fred}, {0, -$CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`fred}, {1, -$CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`fred}, {2, -$CellContext`fred}],
      Editable->False]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, -$CellContext`fred}, {0, -$CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, -$CellContext`fred}, {1, -$CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, -$CellContext`fred}, {2, -$CellContext`fred}],
      Editable->False]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728671398812212*^9, 3.7286749077338467`*^9, 
  3.728743321174183*^9, 3.728748167287958*^9, 3.728749614633626*^9, 
  3.728749956552713*^9, 3.728751138373003*^9, 3.728751646158636*^9, 
  3.728813292154814*^9, 3.728827411317936*^9, 3.728836000576301*^9, 
  3.7288363172004623`*^9, 3.728843197640319*^9, 3.728845653964571*^9, 
  3.733367462393733*^9, 3.792273471651279*^9},
 CellLabel->"Out[228]=",
 CellID->945950370,ExpressionUUID->"118dfd7e-7320-444e-8ba6-790cde7bf667"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We can mix bases:", "MathCaption",
 CellChangeTimes->{{3.728671404931847*^9, 3.728671406595648*^9}},
 CellID->511983772,ExpressionUUID->"6c5d0be9-ecf8-47b9-96ae-2d17aec35761"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentArray", "[", 
  RowBox[{"T", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "fred"}], ",", 
     RowBox[{"-", "blues"}]}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728748169267353*^9, 3.728748172240056*^9}, 
   3.728749342911891*^9},
 CellLabel->"In[229]:=",
 CellID->2091766122,ExpressionUUID->"b7c939b0-aea5-4e57-90e6-7f9e9e2da16d"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`fred}, {0, -$CellContext`blues}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`fred}, {1, -$CellContext`blues}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`fred}, {2, -$CellContext`blues}],
      Editable->False]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`fred}, {0, -$CellContext`blues}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`fred}, {1, -$CellContext`blues}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`fred}, {2, -$CellContext`blues}],
      Editable->False]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, -$CellContext`fred}, {0, -$CellContext`blues}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, -$CellContext`fred}, {1, -$CellContext`blues}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, -$CellContext`fred}, {2, -$CellContext`blues}],
      Editable->False]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728671407878373*^9, 3.728674907881707*^9, 
  3.72874332130976*^9, 3.728748172704029*^9, 3.728749614812111*^9, 
  3.7287499566873713`*^9, 3.728751138539048*^9, 3.7287516462841578`*^9, 
  3.728813292272297*^9, 3.728827411451585*^9, 3.7288360007356777`*^9, 
  3.7288363173527822`*^9, 3.7288431978620996`*^9, 3.728845654098394*^9, 
  3.733367462509252*^9, 3.792273471722356*^9},
 CellLabel->"Out[229]=",
 CellID->766429849,ExpressionUUID->"069b0125-5679-40ea-af91-526e32de6654"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Component indices are not expanded:", "MathCaption",
 CellChangeTimes->{3.728671413899356*^9},
 CellID->1478395,ExpressionUUID->"b6bb5788-32c2-4569-9371-f56d936d44f5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentArray", "[", 
  RowBox[{"T", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "fred"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{"1", ",", 
      RowBox[{"-", "fred"}]}], "}"}]}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.728748175219145*^9, 3.72874817761729*^9}},
 CellLabel->"In[230]:=",
 CellID->1225064648,ExpressionUUID->"b1d13d53-7038-4cf0-8cf5-dd071c81d4de"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{0, -$CellContext`fred}, {1, -$CellContext`fred}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{1, -$CellContext`fred}, {1, -$CellContext`fred}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{2, -$CellContext`fred}, {1, -$CellContext`fred}],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.72867141519095*^9, 3.7286749080253687`*^9, 
  3.7287433214534607`*^9, 3.728748177957554*^9, 3.728749615408486*^9, 
  3.72874995679375*^9, 3.728751138691856*^9, 3.728751646430319*^9, 
  3.728813292372498*^9, 3.728827411585722*^9, 3.7288360008829603`*^9, 
  3.728836317481399*^9, 3.728843198673315*^9, 3.72884565422392*^9, 
  3.7333674626287603`*^9, 3.79227347178335*^9},
 CellLabel->"Out[230]=",
 CellID->779557437,ExpressionUUID->"fc66019e-b4e6-4e29-8aa0-ed717e68d75b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We can specify which indices to expand:", "MathCaption",
 CellChangeTimes->{3.728671420178986*^9},
 CellID->279831134,ExpressionUUID->"689bb334-967c-4dd7-86ab-5e512751f968"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentArray", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"T", "[", 
     RowBox[{
      RowBox[{"-", 
       RowBox[{"{", 
        RowBox[{"a", ",", "fred"}], "}"}]}], ",", 
      RowBox[{"-", 
       RowBox[{"{", 
        RowBox[{"b", ",", "fred"}], "}"}]}]}], "]"}], " ", 
    RowBox[{"v", "[", 
     RowBox[{"{", 
      RowBox[{"c", ",", "blues"}], "}"}], "]"}]}], ",", 
   RowBox[{"IndexList", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "fred"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"c", ",", "blues"}], "}"}]}], "]"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728748180107232*^9, 3.7287481894650097`*^9}, {
  3.728749343792663*^9, 3.728749344571169*^9}},
 CellLabel->"In[231]:=",
 CellID->132682208,ExpressionUUID->"0a486f8c-c973-4305-949d-cb8b149c4f2b"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["b",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{
        0, -$CellContext`fred}, {-$CellContext`b, -$CellContext`fred}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{0, $CellContext`blues}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["b",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{
        0, -$CellContext`fred}, {-$CellContext`b, -$CellContext`fred}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{1, $CellContext`blues}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["b",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{
        0, -$CellContext`fred}, {-$CellContext`b, -$CellContext`fred}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{2, $CellContext`blues}],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["b",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{
        1, -$CellContext`fred}, {-$CellContext`b, -$CellContext`fred}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{0, $CellContext`blues}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["b",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{
        1, -$CellContext`fred}, {-$CellContext`b, -$CellContext`fred}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{1, $CellContext`blues}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["b",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{
        1, -$CellContext`fred}, {-$CellContext`b, -$CellContext`fred}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{2, $CellContext`blues}],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["b",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{
        2, -$CellContext`fred}, {-$CellContext`b, -$CellContext`fred}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{0, $CellContext`blues}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["b",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{
        2, -$CellContext`fred}, {-$CellContext`b, -$CellContext`fred}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{1, $CellContext`blues}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["b",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{
        2, -$CellContext`fred}, {-$CellContext`b, -$CellContext`fred}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{2, $CellContext`blues}],
       Editable->False]}]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728671421384153*^9, 3.7286749081736107`*^9, 
  3.728743321566247*^9, 3.7287481899873962`*^9, 3.7287496155619307`*^9, 
  3.728749956903899*^9, 3.728751139568926*^9, 3.728751646575639*^9, 
  3.728813292484947*^9, 3.728827411721583*^9, 3.728836001028981*^9, 
  3.728836317608458*^9, 3.7288431990200253`*^9, 3.728845654372141*^9, 
  3.733367462744845*^9, 3.792273471858428*^9},
 CellLabel->"Out[231]=",
 CellID->224113312,ExpressionUUID->"d02bb1e6-0567-454f-a791-20266acfbe65"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentArray", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"T", "[", 
     RowBox[{
      RowBox[{"-", 
       RowBox[{"{", 
        RowBox[{"a", ",", "fred"}], "}"}]}], ",", 
      RowBox[{"-", 
       RowBox[{"{", 
        RowBox[{"b", ",", "fred"}], "}"}]}]}], "]"}], " ", 
    RowBox[{"v", "[", 
     RowBox[{"{", 
      RowBox[{"c", ",", "blues"}], "}"}], "]"}]}], ",", 
   RowBox[{"IndicesOf", "[", "T", "]"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.7287481919385777`*^9, 3.7287481990583982`*^9}, 
   3.72874934537096*^9},
 CellLabel->"In[232]:=",
 CellID->357065364,ExpressionUUID->"3057850f-0f17-4feb-807b-ca34a1f9ab76"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`fred}, {0, -$CellContext`fred}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{$CellContext`c, $CellContext`blues}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`fred}, {1, -$CellContext`fred}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{$CellContext`c, $CellContext`blues}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`fred}, {2, -$CellContext`fred}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{$CellContext`c, $CellContext`blues}],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`fred}, {0, -$CellContext`fred}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{$CellContext`c, $CellContext`blues}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`fred}, {1, -$CellContext`fred}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{$CellContext`c, $CellContext`blues}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`fred}, {2, -$CellContext`fred}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{$CellContext`c, $CellContext`blues}],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`fred}, {0, -$CellContext`fred}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{$CellContext`c, $CellContext`blues}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`fred}, {1, -$CellContext`fred}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{$CellContext`c, $CellContext`blues}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`fred}, {2, -$CellContext`fred}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{$CellContext`c, $CellContext`blues}],
       Editable->False]}]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{
  3.728671422121765*^9, 3.7286749083072443`*^9, 3.7287433217148733`*^9, {
   3.728748196938038*^9, 3.728748199523316*^9}, 3.7287496162591763`*^9, 
   3.728749957054669*^9, 3.728751139945642*^9, 3.728751646694422*^9, 
   3.72881329260404*^9, 3.7288274118775473`*^9, 3.728836001196907*^9, 
   3.728836317798483*^9, 3.728843199238874*^9, 3.7288456545240726`*^9, 
   3.733367462944648*^9, 3.792273471936545*^9},
 CellLabel->"Out[232]=",
 CellID->2133601593,ExpressionUUID->"a91dd48a-16c7-4dc6-a925-e3a5196d97a0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["But c-indices are never expanded:", "MathCaption",
 CellChangeTimes->{{3.728671425170817*^9, 3.72867142677072*^9}},
 CellID->436142200,ExpressionUUID->"8d58c171-da7d-45d8-b0a6-7912e456cb50"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentArray", "[", 
  RowBox[{
   RowBox[{"v", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "fred"}], "}"}], "]"}], ",", 
   RowBox[{"IndexList", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "fred"}], "}"}], "]"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728748202256241*^9, 3.7287482046237307`*^9}},
 CellLabel->"In[233]:=",
 CellID->880268189,ExpressionUUID->"67ea0e34-d820-4e63-8a77-573160d1b72e"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"v", 
      StyleBox[GridBox[{
         {
          StyleBox["1",
           FontColor->RGBColor[1., 0., 0.]]},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`v[{1, $CellContext`fred}],
  Editable->False]], "Output",
 CellChangeTimes->{3.728671428036895*^9, 3.728674908544602*^9, 
  3.728743321840966*^9, 3.728748205010784*^9, 3.728749616765636*^9, 
  3.728749957187785*^9, 3.728751140146388*^9, 3.7287516468140793`*^9, 
  3.728813292755021*^9, 3.7288274120021887`*^9, 3.7288360013444147`*^9, 
  3.728836317962322*^9, 3.728843199460099*^9, 3.7288456546704884`*^9, 
  3.733367463168701*^9, 3.792273471995929*^9},
 CellLabel->"Out[233]=",
 CellID->760035945,ExpressionUUID->"74ebc574-a49b-4abf-838a-0943aa930244"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The order of indices matters:", "MathCaption",
 CellID->1189385151,ExpressionUUID->"f0abf66e-b1a5-4810-83c6-c23c22ef2a8d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentArray", "[", 
  RowBox[{"v", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"a", ",", "fred"}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{"b", ",", "fred"}], "}"}]}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.728748207176359*^9, 3.728748209271873*^9}},
 CellLabel->"In[234]:=",
 CellID->1218446581,ExpressionUUID->"bfaaaab2-2530-4843-aaeb-76765b2a8291"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{0, $CellContext`fred}, {0, $CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{0, $CellContext`fred}, {1, $CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{0, $CellContext`fred}, {2, $CellContext`fred}],
      Editable->False]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{1, $CellContext`fred}, {0, $CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{1, $CellContext`fred}, {1, $CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{1, $CellContext`fred}, {2, $CellContext`fred}],
      Editable->False]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{2, $CellContext`fred}, {0, $CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{2, $CellContext`fred}, {1, $CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{2, $CellContext`fred}, {2, $CellContext`fred}],
      Editable->False]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728671434208651*^9, 3.7286749088651237`*^9, 
  3.728743321942219*^9, 3.728749617196711*^9, 3.7287499572817373`*^9, 
  3.72875114036687*^9, 3.7287516469635563`*^9, 3.728813292871694*^9, 
  3.7288274121200447`*^9, 3.728836001487913*^9, 3.728836318086776*^9, 
  3.7288432007140713`*^9, 3.728845654816008*^9, 3.733367463612067*^9, 
  3.792273472071608*^9},
 CellLabel->"Out[234]=",
 CellID->691395259,ExpressionUUID->"65074ebc-72ed-4bd5-8614-f735d2b96a01"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentArray", "[", 
  RowBox[{"v", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"b", ",", "fred"}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{"a", ",", "fred"}], "}"}]}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.72874821062536*^9, 3.728748212927594*^9}},
 CellLabel->"In[235]:=",
 CellID->505395773,ExpressionUUID->"4b73b2d4-a383-4978-9f40-9b49cda636fc"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{0, $CellContext`fred}, {0, $CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{1, $CellContext`fred}, {0, $CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{2, $CellContext`fred}, {0, $CellContext`fred}],
      Editable->False]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{0, $CellContext`fred}, {1, $CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{1, $CellContext`fred}, {1, $CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{2, $CellContext`fred}, {1, $CellContext`fred}],
      Editable->False]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{0, $CellContext`fred}, {2, $CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{1, $CellContext`fred}, {2, $CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{2, $CellContext`fred}, {2, $CellContext`fred}],
      Editable->False]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728671436249634*^9, 3.7286749089820633`*^9, 
  3.7287433220762453`*^9, 3.728748213475788*^9, 3.7287496174415617`*^9, 
  3.728749957418251*^9, 3.7287511406998577`*^9, 3.728751647128005*^9, 
  3.7288132930035763`*^9, 3.728827412251717*^9, 3.7288360016394367`*^9, 
  3.728836318219225*^9, 3.728843200946307*^9, 3.728845654965426*^9, 
  3.733367463761429*^9, 3.792273472142355*^9},
 CellLabel->"Out[235]=",
 CellID->246104096,ExpressionUUID->"1ac7588d-5955-434c-91b4-98438a2b4f99"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["ComponentArray", "MR"],
 " avoids clashes with dummy indices (notice the use of Scalar):"
}], "MathCaption",
 CellID->251955103,ExpressionUUID->"39b01977-996d-4a31-8694-545b4e1eac8b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentArray", "[", 
  RowBox[{
   RowBox[{"Scalar", "[", 
    RowBox[{
     RowBox[{"v", "[", 
      RowBox[{"{", 
       RowBox[{"a", ",", "fred"}], "}"}], "]"}], 
     RowBox[{"v", "[", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "a"}], ",", 
        RowBox[{"-", "fred"}]}], "}"}], "]"}]}], "]"}], 
   RowBox[{"v", "[", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "fred"}]}], "}"}], "]"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728748215657268*^9, 3.728748221935136*^9}},
 CellLabel->"In[236]:=",
 CellID->1919147884,ExpressionUUID->"e4ad6a50-5f85-4073-a814-57f064618d51"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       StyleBox["(",
        FontColor->RGBColor[1, 0, 0]], 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {" "},
                {
                 StyleBox["a",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{-$CellContext`a, -$CellContext`fred}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["a",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{$CellContext`a, $CellContext`fred}],
         Editable->False]}], 
       StyleBox[")",
        FontColor->RGBColor[1, 0, 0]]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Scalar[$CellContext`v[{-$CellContext`a, \
-$CellContext`fred}] $CellContext`v[{$CellContext`a, $CellContext`fred}]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[{0, -$CellContext`fred}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       StyleBox["(",
        FontColor->RGBColor[1, 0, 0]], 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {" "},
                {
                 StyleBox["a",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{-$CellContext`a, -$CellContext`fred}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["a",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{$CellContext`a, $CellContext`fred}],
         Editable->False]}], 
       StyleBox[")",
        FontColor->RGBColor[1, 0, 0]]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Scalar[$CellContext`v[{-$CellContext`a, \
-$CellContext`fred}] $CellContext`v[{$CellContext`a, $CellContext`fred}]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[{1, -$CellContext`fred}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       StyleBox["(",
        FontColor->RGBColor[1, 0, 0]], 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {" "},
                {
                 StyleBox["a",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{-$CellContext`a, -$CellContext`fred}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["a",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{$CellContext`a, $CellContext`fred}],
         Editable->False]}], 
       StyleBox[")",
        FontColor->RGBColor[1, 0, 0]]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Scalar[$CellContext`v[{-$CellContext`a, \
-$CellContext`fred}] $CellContext`v[{$CellContext`a, $CellContext`fred}]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[{2, -$CellContext`fred}],
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.72867144332218*^9, 3.728674909137657*^9, 
  3.728743322263077*^9, 3.728748222425851*^9, 3.728749617641158*^9, 
  3.728749957571431*^9, 3.728751141217173*^9, 3.728751647264408*^9, 
  3.728813293134851*^9, 3.728827412377021*^9, 3.7288360017784643`*^9, 
  3.728836318393175*^9, 3.7288432011735487`*^9, 3.7288456550998096`*^9, 
  3.7333674639074163`*^9, 3.792273472217237*^9},
 CellLabel->"Out[236]=",
 CellID->218037350,ExpressionUUID->"7441d518-0d03-4be5-9660-16ec035aae00"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "If we want more flexibility, we can use ",
 StyleBox["TableOfComponents", "MR"],
 ". This function, whose syntax mimics that of the standard ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " ",
 StyleBox["Table", "MR"],
 ", is much more flexible, but also slower. The reason for this is that ",
 StyleBox["TableOfComponents", "MR"],
 " uses ",
 StyleBox["ToBasis", "MR"],
 " internally, which performs many checks.  Now we specifiy `iterators' on \
basis indices as additional arguments. If the specified index",
 StyleBox[" {a, basis}", "MR"],
 " was already in the expression, this just tells the function when to expand \
it (i.e. in what order with respect to the remaining indices). If it was not \
present but ",
 StyleBox["a", "MR"],
 " appeared as an abstract index, the system first turns it into a basis \
index and then expands it:"
}], "Text",
 CellID->1677179967,ExpressionUUID->"17311fc4-58bc-40ff-b377-8e9c024377d4"],

Cell[CellGroupData[{

Cell["Components of a tensor:", "MathCaption",
 CellChangeTimes->{{3.728671465736947*^9, 3.728671469449094*^9}},
 CellID->1292931946,ExpressionUUID->"6f5d28cd-880a-48d9-ac6e-f60137683129"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TableOfComponents", "[", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{
     RowBox[{"-", 
      RowBox[{"{", 
       RowBox[{"a", ",", "fred"}], "}"}]}], ",", 
     RowBox[{"-", 
      RowBox[{"{", 
       RowBox[{"b", ",", "fred"}], "}"}]}]}], "]"}], ",", 
   RowBox[{"-", 
    RowBox[{"{", 
     RowBox[{"a", ",", "fred"}], "}"}]}], ",", 
   RowBox[{"-", 
    RowBox[{"{", 
     RowBox[{"b", ",", "fred"}], "}"}]}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728748225993267*^9, 3.728748231270944*^9}},
 CellLabel->"In[237]:=",
 CellID->493354847,ExpressionUUID->"4fa29e41-8938-4aee-bf26-95bd3ff19ed3"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`fred}, {0, -$CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`fred}, {1, -$CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`fred}, {2, -$CellContext`fred}],
      Editable->False]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`fred}, {0, -$CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`fred}, {1, -$CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`fred}, {2, -$CellContext`fred}],
      Editable->False]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, -$CellContext`fred}, {0, -$CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, -$CellContext`fred}, {1, -$CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, -$CellContext`fred}, {2, -$CellContext`fred}],
      Editable->False]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728671470708962*^9, 3.72867490928337*^9, 
  3.728743322451272*^9, 3.728748231726149*^9, 3.7287496178353653`*^9, 
  3.728749957688471*^9, 3.728751141686677*^9, 3.728751647416174*^9, 
  3.728813293256015*^9, 3.728827412502222*^9, 3.728836001921897*^9, 
  3.728836318537271*^9, 3.728843202451536*^9, 3.728845655229763*^9, 
  3.733367464609322*^9, 3.792273472294238*^9},
 CellLabel->"Out[237]=",
 CellID->946639668,ExpressionUUID->"be918f10-a1e2-4492-8e73-0f5f98995554"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TableOfComponents", "[", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", 
   RowBox[{"-", 
    RowBox[{"{", 
     RowBox[{"a", ",", "fred"}], "}"}]}], ",", 
   RowBox[{"-", 
    RowBox[{"{", 
     RowBox[{"b", ",", "fred"}], "}"}]}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728748233952306*^9, 3.7287482388546352`*^9}},
 CellLabel->"In[238]:=",
 CellID->738928068,ExpressionUUID->"240e0b44-c50a-420e-bf9a-537cba251d9c"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`fred}, {0, -$CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`fred}, {1, -$CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`fred}, {2, -$CellContext`fred}],
      Editable->False]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`fred}, {0, -$CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`fred}, {1, -$CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`fred}, {2, -$CellContext`fred}],
      Editable->False]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, -$CellContext`fred}, {0, -$CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, -$CellContext`fred}, {1, -$CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, -$CellContext`fred}, {2, -$CellContext`fred}],
      Editable->False]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728671471392598*^9, 3.728674909417317*^9, 
  3.728743322605654*^9, 3.728748239443735*^9, 3.728749618564131*^9, 
  3.728749957810782*^9, 3.728751141962102*^9, 3.728751647561982*^9, 
  3.728813293378243*^9, 3.728827412636283*^9, 3.7288360020764723`*^9, 
  3.728836318687743*^9, 3.728843202662019*^9, 3.728845655364756*^9, 
  3.733367464765006*^9, 3.792273472369886*^9},
 CellLabel->"Out[238]=",
 CellID->1614043100,ExpressionUUID->"f86f11db-9eb2-438c-85dc-3a7ddc1bb92a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Notice how basis indices mimic the behaviour of ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " iterators. We have been able to expand an expression which contained only \
abstract indices. This is equivalent to doing:"
}], "MathCaption",
 CellChangeTimes->{3.728671474368849*^9},
 CellID->1420824871,ExpressionUUID->"4799cf53-93c2-4548-b41c-02cfdba862f4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToBasis", "[", "fred", "]"}], "[", 
  RowBox[{"T", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}]}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.728748242271817*^9, 3.728748242774135*^9}},
 CellLabel->"In[239]:=",
 CellID->1932935904,ExpressionUUID->"c0a5f5b7-a144-4d2f-8e81-4657f1b78b56"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"T", 
      StyleBox[GridBox[{
         {" ", " "},
         {
          StyleBox["a",
           FontColor->RGBColor[1., 0., 0.]], 
          StyleBox["b",
           FontColor->RGBColor[1., 0., 0.]]}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`T[{-$CellContext`a, -$CellContext`fred}, {-$CellContext`b, \
-$CellContext`fred}],
  Editable->False]], "Output",
 CellChangeTimes->{3.7286714755476713`*^9, 3.728674909539322*^9, 
  3.728743322818674*^9, 3.7287482431289167`*^9, 3.728749618924347*^9, 
  3.728749957920411*^9, 3.728751142162635*^9, 3.728751647700996*^9, 
  3.7288132934880867`*^9, 3.728827412758305*^9, 3.728836002213447*^9, 
  3.728836318828266*^9, 3.7288432028781424`*^9, 3.728845655486554*^9, 
  3.7333674649680233`*^9, 3.792273472433866*^9},
 CellLabel->"Out[239]=",
 CellID->208952796,ExpressionUUID->"56698c26-ab0f-4d5b-8dcf-d83f178043e6"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentArray", "[", "%", "]"}]], "Input",
 CellLabel->"In[240]:=",
 CellID->933449985,ExpressionUUID->"917eaa03-5db4-4ca2-9623-69cdf0b8e81b"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`fred}, {0, -$CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`fred}, {1, -$CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`fred}, {2, -$CellContext`fred}],
      Editable->False]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`fred}, {0, -$CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`fred}, {1, -$CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`fred}, {2, -$CellContext`fred}],
      Editable->False]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, -$CellContext`fred}, {0, -$CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, -$CellContext`fred}, {1, -$CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, -$CellContext`fred}, {2, -$CellContext`fred}],
      Editable->False]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728671476260687*^9, 3.728674909702829*^9, 
  3.728743322943756*^9, 3.728748244849114*^9, 3.72874961925522*^9, 
  3.7287499580145893`*^9, 3.728751142335991*^9, 3.728751647840724*^9, 
  3.728813293587212*^9, 3.7288274128779793`*^9, 3.728836002383499*^9, 
  3.728836319024743*^9, 3.7288432040164948`*^9, 3.728845655636343*^9, 
  3.733367465345388*^9, 3.7922734725085487`*^9},
 CellLabel->"Out[240]=",
 CellID->733879527,ExpressionUUID->"7ec7bc3e-19d3-4d0e-90f0-5e292ecdbcd4"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
However, we cannot change the basis of a BIndex already in the expression:\
\>", "MathCaption",
 CellChangeTimes->{3.728671481160225*^9},
 CellID->792790203,ExpressionUUID->"33fc9932-6e5c-4c8e-92a2-a825703ad7cc"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TableOfComponents", "[", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "fred"}]}], "}"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "blues"}]}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.7287482468959427`*^9, 3.7287482502379417`*^9}, 
   3.728749346813799*^9},
 CellLabel->"In[241]:=",
 CellID->1436395833,ExpressionUUID->"2407c5a8-5327-46f5-b5e1-7f08c87af4f0"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"T", 
      StyleBox[GridBox[{
         {" ", " "},
         {
          StyleBox["a",
           FontColor->RGBColor[1., 0., 0.]], "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`T[{-$CellContext`a, -$CellContext`fred}, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.7286714821883183`*^9, 3.728674909823511*^9, 
  3.728743323109071*^9, 3.728748250762381*^9, 3.728749619454109*^9, 
  3.728749958124642*^9, 3.72875114280402*^9, 3.728751647984015*^9, 
  3.728813293710915*^9, 3.728827413006935*^9, 3.7288360025140944`*^9, 
  3.728836319161438*^9, 3.7288432042446136`*^9, 3.728845655771043*^9, 
  3.733367465495447*^9, 3.792273472568255*^9},
 CellLabel->"Out[241]=",
 CellID->1314625568,ExpressionUUID->"559049d9-77dc-42cf-b4d7-0261c413ac6d"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "If we do not specify any index, the argument is returned unchanged (compare \
with ",
 StyleBox["ComponentArray", "MR"],
 "):"
}], "MathCaption",
 CellID->1655453907,ExpressionUUID->"ab0f5ff5-09f9-4ff7-8007-ef3c2c6c2119"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TableOfComponents", "[", 
  RowBox[{"v", "[", 
   RowBox[{"{", 
    RowBox[{"a", ",", "fred"}], "}"}], "]"}], " ", "]"}]], "Input",
 CellChangeTimes->{{3.728748254127534*^9, 3.728748254565769*^9}},
 CellLabel->"In[242]:=",
 CellID->415684417,ExpressionUUID->"065bcdc8-d017-464e-adf7-deb839def043"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"v", 
      StyleBox[GridBox[{
         {
          StyleBox["a",
           FontColor->RGBColor[1., 0., 0.]]},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`v[{$CellContext`a, $CellContext`fred}],
  Editable->False]], "Output",
 CellChangeTimes->{3.728671487160816*^9, 3.728674909939664*^9, 
  3.728743323256686*^9, 3.728748255129406*^9, 3.7287496200647497`*^9, 
  3.7287499582697144`*^9, 3.728751143484943*^9, 3.7287516481805677`*^9, 
  3.7288132938564157`*^9, 3.728827413152452*^9, 3.728836002670817*^9, 
  3.728836319294531*^9, 3.728843204477043*^9, 3.728845655920558*^9, 
  3.7333674656445293`*^9, 3.7922734726281567`*^9},
 CellLabel->"Out[242]=",
 CellID->1600161040,ExpressionUUID->"00746fdf-45ef-44ba-adde-1508d1b7117e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{" ", 
  RowBox[{"ComponentArray", "[", 
   RowBox[{"v", "[", 
    RowBox[{"{", 
     RowBox[{"a", ",", "fred"}], "}"}], "]"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.7287482572713737`*^9, 3.728748257717671*^9}},
 CellLabel->"In[243]:=",
 CellID->1670230257,ExpressionUUID->"cb10128b-ed67-43b4-b609-bb86f72e1731"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[{0, $CellContext`fred}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[{1, $CellContext`fred}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[{2, $CellContext`fred}],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.7286714878757763`*^9, 3.7286749100725327`*^9, 
  3.7287433233879633`*^9, 3.728748258078566*^9, 3.728749620240603*^9, 
  3.728749958404331*^9, 3.7287511436611958`*^9, 3.728751648312583*^9, 
  3.7288132939734383`*^9, 3.7288274133106947`*^9, 3.728836002845586*^9, 
  3.7288363194270897`*^9, 3.728843204957714*^9, 3.72884565610191*^9, 
  3.733367465797553*^9, 3.7922734726987963`*^9},
 CellLabel->"Out[243]=",
 CellID->968887899,ExpressionUUID->"29e66191-ea04-4afa-9e86-e5b0c6269509"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The order is again important:", "MathCaption",
 CellChangeTimes->{3.728671492551804*^9},
 CellID->1493729722,ExpressionUUID->"8c2b2c07-3538-4cd4-8d47-454bd16e15a3"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TableOfComponents", "[", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{"a", ",", "b"}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "fred"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"b", ",", "fred"}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.7287482595751534`*^9, 3.728748262565436*^9}},
 CellLabel->"In[244]:=",
 CellID->1031604224,ExpressionUUID->"b88d5f8e-c836-49de-9a87-be58be26e1dd"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, $CellContext`fred}, {0, $CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, $CellContext`fred}, {1, $CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, $CellContext`fred}, {2, $CellContext`fred}],
      Editable->False]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, $CellContext`fred}, {0, $CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, $CellContext`fred}, {1, $CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, $CellContext`fred}, {2, $CellContext`fred}],
      Editable->False]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, $CellContext`fred}, {0, $CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, $CellContext`fred}, {1, $CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, $CellContext`fred}, {2, $CellContext`fred}],
      Editable->False]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728671493610955*^9, 3.728674910380188*^9, 
  3.728743323565233*^9, 3.72874826402414*^9, 3.728749620453319*^9, 
  3.728749958515077*^9, 3.728751144234973*^9, 3.728751648464246*^9, 
  3.728813294089703*^9, 3.728827413450541*^9, 3.728836003009054*^9, 
  3.728836319555196*^9, 3.728843205949601*^9, 3.728845656215138*^9, 
  3.733367465926509*^9, 3.792273472774425*^9},
 CellLabel->"Out[244]=",
 CellID->1877906939,ExpressionUUID->"e11567ce-a08a-40be-8258-f1cf71fe7c9b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TableOfComponents", "[", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{"a", ",", "b"}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"b", ",", "fred"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "fred"}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728748265694745*^9, 3.728748269428916*^9}},
 CellLabel->"In[245]:=",
 CellID->1855656537,ExpressionUUID->"58a23fdb-c11a-43b0-9fb9-c07e8f192fd9"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, $CellContext`fred}, {0, $CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, $CellContext`fred}, {0, $CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, $CellContext`fred}, {0, $CellContext`fred}],
      Editable->False]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, $CellContext`fred}, {1, $CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, $CellContext`fred}, {1, $CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, $CellContext`fred}, {1, $CellContext`fred}],
      Editable->False]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, $CellContext`fred}, {2, $CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, $CellContext`fred}, {2, $CellContext`fred}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, $CellContext`fred}, {2, $CellContext`fred}],
      Editable->False]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728671495723659*^9, 3.7286749105816298`*^9, 
  3.728743323753043*^9, 3.7287482699399652`*^9, 3.728749620635275*^9, 
  3.728749958613258*^9, 3.72875114485163*^9, 3.728751648600526*^9, 
  3.728813294211544*^9, 3.7288274135785637`*^9, 3.7288360031541443`*^9, 
  3.72883631974387*^9, 3.7288432061975193`*^9, 3.728845656346684*^9, 
  3.7333674664148283`*^9, 3.792273472844969*^9},
 CellLabel->"Out[245]=",
 CellID->414763814,ExpressionUUID->"af24a6a9-25d2-4280-882b-05c919f9b520"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "As ",
 StyleBox["ToBasis", "MR"],
 " is able to handle derivatives, we can include them in the arguments of ",
 StyleBox["TableOfComponents:", "MR"]
}], "MathCaption",
 CellChangeTimes->{3.7286715010714703`*^9},
 CellID->924250181,ExpressionUUID->"96741730-4511-4908-81d1-46844edb4f98"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"TableOfComponents", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"PD", "[", 
       RowBox[{"-", "a"}], "]"}], "[", 
      RowBox[{"v", "[", "b", "]"}], "]"}], ",", 
     RowBox[{"-", 
      RowBox[{"{", 
       RowBox[{"a", ",", "fred"}], "}"}]}], ",", 
     RowBox[{"{", 
      RowBox[{"b", ",", "blues"}], "}"}]}], "]"}], "//", 
   "ScreenDollarIndices"}], "//", "MatrixForm"}]], "Input",
 CellChangeTimes->{{3.7287482723428183`*^9, 3.728748305235218*^9}, 
   3.728749347640298*^9},
 CellLabel->"In[246]:=",
 CellID->1123573316,ExpressionUUID->"a01c276b-3c48-433e-8d5c-d316b1fec3d7"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {
              RowBox[{"\[CapitalGamma]", "[", 
               StyleBox["\[ScriptCapitalD]",
                FontColor->RGBColor[0., 0., 1.]], "]"}], 
              StyleBox[GridBox[{
                 {
                  StyleBox["0",
                   FontColor->RGBColor[0., 0., 1.]], " ", " "},
                 {" ", 
                  StyleBox["0",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["a",
                   FontColor->RGBColor[0., 0., 1.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
               FontSize->9]}
            },
            GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
            
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`ChristoffelPDblues[{0, $CellContext`blues}, {
           0, -$CellContext`fred}, {-$CellContext`a, -$CellContext`blues}],
          Editable->False]}], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["a",
                  FontColor->RGBColor[0., 0., 1.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{$CellContext`a, $CellContext`blues}],
         Editable->False]}], "+", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[PartialD]", 
           StyleBox["0",
            FontColor->RGBColor[1., 0., 0.]]], GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`PD[{0, -$CellContext`fred}][
         $CellContext`v[{0, $CellContext`blues}]],
        Editable->False]}], 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {
              RowBox[{"\[CapitalGamma]", "[", 
               StyleBox["\[ScriptCapitalD]",
                FontColor->RGBColor[0., 0., 1.]], "]"}], 
              StyleBox[GridBox[{
                 {
                  StyleBox["1",
                   FontColor->RGBColor[0., 0., 1.]], " ", " "},
                 {" ", 
                  StyleBox["0",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["a",
                   FontColor->RGBColor[0., 0., 1.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
               FontSize->9]}
            },
            GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
            
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`ChristoffelPDblues[{1, $CellContext`blues}, {
           0, -$CellContext`fred}, {-$CellContext`a, -$CellContext`blues}],
          Editable->False]}], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["a",
                  FontColor->RGBColor[0., 0., 1.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{$CellContext`a, $CellContext`blues}],
         Editable->False]}], "+", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[PartialD]", 
           StyleBox["0",
            FontColor->RGBColor[1., 0., 0.]]], GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[0., 0., 1.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`PD[{0, -$CellContext`fred}][
         $CellContext`v[{1, $CellContext`blues}]],
        Editable->False]}], 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {
              RowBox[{"\[CapitalGamma]", "[", 
               StyleBox["\[ScriptCapitalD]",
                FontColor->RGBColor[0., 0., 1.]], "]"}], 
              StyleBox[GridBox[{
                 {
                  StyleBox["2",
                   FontColor->RGBColor[0., 0., 1.]], " ", " "},
                 {" ", 
                  StyleBox["0",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["a",
                   FontColor->RGBColor[0., 0., 1.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
               FontSize->9]}
            },
            GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
            
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`ChristoffelPDblues[{2, $CellContext`blues}, {
           0, -$CellContext`fred}, {-$CellContext`a, -$CellContext`blues}],
          Editable->False]}], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["a",
                  FontColor->RGBColor[0., 0., 1.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{$CellContext`a, $CellContext`blues}],
         Editable->False]}], "+", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[PartialD]", 
           StyleBox["0",
            FontColor->RGBColor[1., 0., 0.]]], GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["2",
                  FontColor->RGBColor[0., 0., 1.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`PD[{0, -$CellContext`fred}][
         $CellContext`v[{2, $CellContext`blues}]],
        Editable->False]}]},
     {
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {
              RowBox[{"\[CapitalGamma]", "[", 
               StyleBox["\[ScriptCapitalD]",
                FontColor->RGBColor[0., 0., 1.]], "]"}], 
              StyleBox[GridBox[{
                 {
                  StyleBox["0",
                   FontColor->RGBColor[0., 0., 1.]], " ", " "},
                 {" ", 
                  StyleBox["1",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["a",
                   FontColor->RGBColor[0., 0., 1.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
               FontSize->9]}
            },
            GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
            
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`ChristoffelPDblues[{0, $CellContext`blues}, {
           1, -$CellContext`fred}, {-$CellContext`a, -$CellContext`blues}],
          Editable->False]}], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["a",
                  FontColor->RGBColor[0., 0., 1.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{$CellContext`a, $CellContext`blues}],
         Editable->False]}], "+", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[PartialD]", 
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]]], GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`PD[{1, -$CellContext`fred}][
         $CellContext`v[{0, $CellContext`blues}]],
        Editable->False]}], 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {
              RowBox[{"\[CapitalGamma]", "[", 
               StyleBox["\[ScriptCapitalD]",
                FontColor->RGBColor[0., 0., 1.]], "]"}], 
              StyleBox[GridBox[{
                 {
                  StyleBox["1",
                   FontColor->RGBColor[0., 0., 1.]], " ", " "},
                 {" ", 
                  StyleBox["1",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["a",
                   FontColor->RGBColor[0., 0., 1.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
               FontSize->9]}
            },
            GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
            
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`ChristoffelPDblues[{1, $CellContext`blues}, {
           1, -$CellContext`fred}, {-$CellContext`a, -$CellContext`blues}],
          Editable->False]}], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["a",
                  FontColor->RGBColor[0., 0., 1.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{$CellContext`a, $CellContext`blues}],
         Editable->False]}], "+", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[PartialD]", 
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]]], GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[0., 0., 1.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`PD[{1, -$CellContext`fred}][
         $CellContext`v[{1, $CellContext`blues}]],
        Editable->False]}], 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {
              RowBox[{"\[CapitalGamma]", "[", 
               StyleBox["\[ScriptCapitalD]",
                FontColor->RGBColor[0., 0., 1.]], "]"}], 
              StyleBox[GridBox[{
                 {
                  StyleBox["2",
                   FontColor->RGBColor[0., 0., 1.]], " ", " "},
                 {" ", 
                  StyleBox["1",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["a",
                   FontColor->RGBColor[0., 0., 1.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
               FontSize->9]}
            },
            GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
            
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`ChristoffelPDblues[{2, $CellContext`blues}, {
           1, -$CellContext`fred}, {-$CellContext`a, -$CellContext`blues}],
          Editable->False]}], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["a",
                  FontColor->RGBColor[0., 0., 1.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{$CellContext`a, $CellContext`blues}],
         Editable->False]}], "+", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[PartialD]", 
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]]], GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["2",
                  FontColor->RGBColor[0., 0., 1.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`PD[{1, -$CellContext`fred}][
         $CellContext`v[{2, $CellContext`blues}]],
        Editable->False]}]},
     {
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {
              RowBox[{"\[CapitalGamma]", "[", 
               StyleBox["\[ScriptCapitalD]",
                FontColor->RGBColor[0., 0., 1.]], "]"}], 
              StyleBox[GridBox[{
                 {
                  StyleBox["0",
                   FontColor->RGBColor[0., 0., 1.]], " ", " "},
                 {" ", 
                  StyleBox["2",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["a",
                   FontColor->RGBColor[0., 0., 1.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
               FontSize->9]}
            },
            GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
            
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`ChristoffelPDblues[{0, $CellContext`blues}, {
           2, -$CellContext`fred}, {-$CellContext`a, -$CellContext`blues}],
          Editable->False]}], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["a",
                  FontColor->RGBColor[0., 0., 1.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{$CellContext`a, $CellContext`blues}],
         Editable->False]}], "+", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[PartialD]", 
           StyleBox["2",
            FontColor->RGBColor[1., 0., 0.]]], GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`PD[{2, -$CellContext`fred}][
         $CellContext`v[{0, $CellContext`blues}]],
        Editable->False]}], 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {
              RowBox[{"\[CapitalGamma]", "[", 
               StyleBox["\[ScriptCapitalD]",
                FontColor->RGBColor[0., 0., 1.]], "]"}], 
              StyleBox[GridBox[{
                 {
                  StyleBox["1",
                   FontColor->RGBColor[0., 0., 1.]], " ", " "},
                 {" ", 
                  StyleBox["2",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["a",
                   FontColor->RGBColor[0., 0., 1.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
               FontSize->9]}
            },
            GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
            
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`ChristoffelPDblues[{1, $CellContext`blues}, {
           2, -$CellContext`fred}, {-$CellContext`a, -$CellContext`blues}],
          Editable->False]}], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["a",
                  FontColor->RGBColor[0., 0., 1.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{$CellContext`a, $CellContext`blues}],
         Editable->False]}], "+", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[PartialD]", 
           StyleBox["2",
            FontColor->RGBColor[1., 0., 0.]]], GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[0., 0., 1.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`PD[{2, -$CellContext`fred}][
         $CellContext`v[{1, $CellContext`blues}]],
        Editable->False]}], 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {
              RowBox[{"\[CapitalGamma]", "[", 
               StyleBox["\[ScriptCapitalD]",
                FontColor->RGBColor[0., 0., 1.]], "]"}], 
              StyleBox[GridBox[{
                 {
                  StyleBox["2",
                   FontColor->RGBColor[0., 0., 1.]], " ", " "},
                 {" ", 
                  StyleBox["2",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["a",
                   FontColor->RGBColor[0., 0., 1.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
               FontSize->9]}
            },
            GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
            
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`ChristoffelPDblues[{2, $CellContext`blues}, {
           2, -$CellContext`fred}, {-$CellContext`a, -$CellContext`blues}],
          Editable->False]}], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["a",
                  FontColor->RGBColor[0., 0., 1.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{$CellContext`a, $CellContext`blues}],
         Editable->False]}], "+", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[PartialD]", 
           StyleBox["2",
            FontColor->RGBColor[1., 0., 0.]]], GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["2",
                  FontColor->RGBColor[0., 0., 1.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`PD[{2, -$CellContext`fred}][
         $CellContext`v[{2, $CellContext`blues}]],
        Editable->False]}]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Baseline}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{3.7286715023937683`*^9, 3.728674910751134*^9, 
  3.72874332393915*^9, 3.728748275545122*^9, 3.7287483057927628`*^9, 
  3.7287496207919064`*^9, 3.7287499587447643`*^9, 3.7287511450469313`*^9, 
  3.7287516487888107`*^9, 3.728813294372806*^9, 3.728827413740324*^9, 
  3.728836003344092*^9, 3.728836319929907*^9, 3.728843206477178*^9, 
  3.7288456565104513`*^9, 3.733367466659717*^9, 3.792273472933734*^9},
 CellLabel->"Out[246]//MatrixForm=",
 CellID->203503736,ExpressionUUID->"ebe3a2eb-9c86-4e02-a6ac-b18fd376157b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TraceBasisDummy", "[", "%", "]"}]], "Input",
 CellLabel->"In[247]:=",
 CellID->1725122528,ExpressionUUID->"674c249b-e869-4535-aec4-6678bee89a06"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox[GridBox[{
            {
             RowBox[{"\[CapitalGamma]", "[", 
              StyleBox["\[ScriptCapitalD]",
               FontColor->RGBColor[0., 0., 1.]], "]"}], 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]], " ", " "},
                {" ", 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`ChristoffelPDblues[{0, $CellContext`blues}, {
          0, -$CellContext`fred}, {0, -$CellContext`blues}],
         Editable->False]}], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[{0, $CellContext`blues}],
        Editable->False]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelPDblues[{0, $CellContext`blues}, {
         0, -$CellContext`fred}, {1, -$CellContext`blues}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[{1, $CellContext`blues}],
        Editable->False]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelPDblues[{0, $CellContext`blues}, {
         0, -$CellContext`fred}, {2, -$CellContext`blues}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[{2, $CellContext`blues}],
        Editable->False]}], "+", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SubscriptBox["\[PartialD]", 
          StyleBox["0",
           FontColor->RGBColor[1., 0., 0.]]], GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`PD[{0, -$CellContext`fred}][
        $CellContext`v[{0, $CellContext`blues}]],
       Editable->False]}], ",", 
     RowBox[{
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox[GridBox[{
            {
             RowBox[{"\[CapitalGamma]", "[", 
              StyleBox["\[ScriptCapitalD]",
               FontColor->RGBColor[0., 0., 1.]], "]"}], 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[0., 0., 1.]], " ", " "},
                {" ", 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`ChristoffelPDblues[{1, $CellContext`blues}, {
          0, -$CellContext`fred}, {0, -$CellContext`blues}],
         Editable->False]}], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[{0, $CellContext`blues}],
        Editable->False]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelPDblues[{1, $CellContext`blues}, {
         0, -$CellContext`fred}, {1, -$CellContext`blues}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[{1, $CellContext`blues}],
        Editable->False]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelPDblues[{1, $CellContext`blues}, {
         0, -$CellContext`fred}, {2, -$CellContext`blues}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[{2, $CellContext`blues}],
        Editable->False]}], "+", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SubscriptBox["\[PartialD]", 
          StyleBox["0",
           FontColor->RGBColor[1., 0., 0.]]], GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`PD[{0, -$CellContext`fred}][
        $CellContext`v[{1, $CellContext`blues}]],
       Editable->False]}], ",", 
     RowBox[{
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox[GridBox[{
            {
             RowBox[{"\[CapitalGamma]", "[", 
              StyleBox["\[ScriptCapitalD]",
               FontColor->RGBColor[0., 0., 1.]], "]"}], 
             StyleBox[GridBox[{
                {
                 StyleBox["2",
                  FontColor->RGBColor[0., 0., 1.]], " ", " "},
                {" ", 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`ChristoffelPDblues[{2, $CellContext`blues}, {
          0, -$CellContext`fred}, {0, -$CellContext`blues}],
         Editable->False]}], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[{0, $CellContext`blues}],
        Editable->False]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelPDblues[{2, $CellContext`blues}, {
         0, -$CellContext`fred}, {1, -$CellContext`blues}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[{1, $CellContext`blues}],
        Editable->False]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelPDblues[{2, $CellContext`blues}, {
         0, -$CellContext`fred}, {2, -$CellContext`blues}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[{2, $CellContext`blues}],
        Editable->False]}], "+", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SubscriptBox["\[PartialD]", 
          StyleBox["0",
           FontColor->RGBColor[1., 0., 0.]]], GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`PD[{0, -$CellContext`fred}][
        $CellContext`v[{2, $CellContext`blues}]],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox[GridBox[{
            {
             RowBox[{"\[CapitalGamma]", "[", 
              StyleBox["\[ScriptCapitalD]",
               FontColor->RGBColor[0., 0., 1.]], "]"}], 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]], " ", " "},
                {" ", 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`ChristoffelPDblues[{0, $CellContext`blues}, {
          1, -$CellContext`fred}, {0, -$CellContext`blues}],
         Editable->False]}], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[{0, $CellContext`blues}],
        Editable->False]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelPDblues[{0, $CellContext`blues}, {
         1, -$CellContext`fred}, {1, -$CellContext`blues}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[{1, $CellContext`blues}],
        Editable->False]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelPDblues[{0, $CellContext`blues}, {
         1, -$CellContext`fred}, {2, -$CellContext`blues}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[{2, $CellContext`blues}],
        Editable->False]}], "+", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SubscriptBox["\[PartialD]", 
          StyleBox["1",
           FontColor->RGBColor[1., 0., 0.]]], GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`PD[{1, -$CellContext`fred}][
        $CellContext`v[{0, $CellContext`blues}]],
       Editable->False]}], ",", 
     RowBox[{
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox[GridBox[{
            {
             RowBox[{"\[CapitalGamma]", "[", 
              StyleBox["\[ScriptCapitalD]",
               FontColor->RGBColor[0., 0., 1.]], "]"}], 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[0., 0., 1.]], " ", " "},
                {" ", 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`ChristoffelPDblues[{1, $CellContext`blues}, {
          1, -$CellContext`fred}, {0, -$CellContext`blues}],
         Editable->False]}], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[{0, $CellContext`blues}],
        Editable->False]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelPDblues[{1, $CellContext`blues}, {
         1, -$CellContext`fred}, {1, -$CellContext`blues}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[{1, $CellContext`blues}],
        Editable->False]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelPDblues[{1, $CellContext`blues}, {
         1, -$CellContext`fred}, {2, -$CellContext`blues}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[{2, $CellContext`blues}],
        Editable->False]}], "+", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SubscriptBox["\[PartialD]", 
          StyleBox["1",
           FontColor->RGBColor[1., 0., 0.]]], GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`PD[{1, -$CellContext`fred}][
        $CellContext`v[{1, $CellContext`blues}]],
       Editable->False]}], ",", 
     RowBox[{
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox[GridBox[{
            {
             RowBox[{"\[CapitalGamma]", "[", 
              StyleBox["\[ScriptCapitalD]",
               FontColor->RGBColor[0., 0., 1.]], "]"}], 
             StyleBox[GridBox[{
                {
                 StyleBox["2",
                  FontColor->RGBColor[0., 0., 1.]], " ", " "},
                {" ", 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`ChristoffelPDblues[{2, $CellContext`blues}, {
          1, -$CellContext`fred}, {0, -$CellContext`blues}],
         Editable->False]}], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[{0, $CellContext`blues}],
        Editable->False]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelPDblues[{2, $CellContext`blues}, {
         1, -$CellContext`fred}, {1, -$CellContext`blues}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[{1, $CellContext`blues}],
        Editable->False]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelPDblues[{2, $CellContext`blues}, {
         1, -$CellContext`fred}, {2, -$CellContext`blues}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[{2, $CellContext`blues}],
        Editable->False]}], "+", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SubscriptBox["\[PartialD]", 
          StyleBox["1",
           FontColor->RGBColor[1., 0., 0.]]], GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`PD[{1, -$CellContext`fred}][
        $CellContext`v[{2, $CellContext`blues}]],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox[GridBox[{
            {
             RowBox[{"\[CapitalGamma]", "[", 
              StyleBox["\[ScriptCapitalD]",
               FontColor->RGBColor[0., 0., 1.]], "]"}], 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]], " ", " "},
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`ChristoffelPDblues[{0, $CellContext`blues}, {
          2, -$CellContext`fred}, {0, -$CellContext`blues}],
         Editable->False]}], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[{0, $CellContext`blues}],
        Editable->False]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelPDblues[{0, $CellContext`blues}, {
         2, -$CellContext`fred}, {1, -$CellContext`blues}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[{1, $CellContext`blues}],
        Editable->False]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelPDblues[{0, $CellContext`blues}, {
         2, -$CellContext`fred}, {2, -$CellContext`blues}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[{2, $CellContext`blues}],
        Editable->False]}], "+", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SubscriptBox["\[PartialD]", 
          StyleBox["2",
           FontColor->RGBColor[1., 0., 0.]]], GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`PD[{2, -$CellContext`fred}][
        $CellContext`v[{0, $CellContext`blues}]],
       Editable->False]}], ",", 
     RowBox[{
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox[GridBox[{
            {
             RowBox[{"\[CapitalGamma]", "[", 
              StyleBox["\[ScriptCapitalD]",
               FontColor->RGBColor[0., 0., 1.]], "]"}], 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[0., 0., 1.]], " ", " "},
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`ChristoffelPDblues[{1, $CellContext`blues}, {
          2, -$CellContext`fred}, {0, -$CellContext`blues}],
         Editable->False]}], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[{0, $CellContext`blues}],
        Editable->False]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelPDblues[{1, $CellContext`blues}, {
         2, -$CellContext`fred}, {1, -$CellContext`blues}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[{1, $CellContext`blues}],
        Editable->False]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelPDblues[{1, $CellContext`blues}, {
         2, -$CellContext`fred}, {2, -$CellContext`blues}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[{2, $CellContext`blues}],
        Editable->False]}], "+", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SubscriptBox["\[PartialD]", 
          StyleBox["2",
           FontColor->RGBColor[1., 0., 0.]]], GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`PD[{2, -$CellContext`fred}][
        $CellContext`v[{1, $CellContext`blues}]],
       Editable->False]}], ",", 
     RowBox[{
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox[GridBox[{
            {
             RowBox[{"\[CapitalGamma]", "[", 
              StyleBox["\[ScriptCapitalD]",
               FontColor->RGBColor[0., 0., 1.]], "]"}], 
             StyleBox[GridBox[{
                {
                 StyleBox["2",
                  FontColor->RGBColor[0., 0., 1.]], " ", " "},
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[0., 0., 1.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
              FontSize->9]}
           },
           GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`ChristoffelPDblues[{2, $CellContext`blues}, {
          2, -$CellContext`fred}, {0, -$CellContext`blues}],
         Editable->False]}], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[{0, $CellContext`blues}],
        Editable->False]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelPDblues[{2, $CellContext`blues}, {
         2, -$CellContext`fred}, {1, -$CellContext`blues}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[{1, $CellContext`blues}],
        Editable->False]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelPDblues[{2, $CellContext`blues}, {
         2, -$CellContext`fred}, {2, -$CellContext`blues}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[{2, $CellContext`blues}],
        Editable->False]}], "+", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SubscriptBox["\[PartialD]", 
          StyleBox["2",
           FontColor->RGBColor[1., 0., 0.]]], GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`PD[{2, -$CellContext`fred}][
        $CellContext`v[{2, $CellContext`blues}]],
       Editable->False]}]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728671503809122*^9, 3.728674910942719*^9, 
  3.728743324103361*^9, 3.728748312513291*^9, 3.7287496215879917`*^9, 
  3.728749958887137*^9, 3.728751146181731*^9, 3.728751648964793*^9, 
  3.7288132945434017`*^9, 3.728827413928339*^9, 3.728836003551755*^9, 
  3.728836320166152*^9, 3.728843206794034*^9, 3.728845656720356*^9, 
  3.733367466844223*^9, 3.79227347306147*^9},
 CellLabel->"Out[247]=",
 CellID->1493324928,ExpressionUUID->"9ae8bc88-ba16-4153-8a72-3e3fd9fe41ae"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "If we wanted to obtain this result from ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], " ", 
      SuperscriptBox["v", "b"]}], " ", "using", " ", 
     StyleBox["ComponentArray",
      FontFamily->"Courier"]}], ",", " ", 
    RowBox[{
    "we", " ", "would", " ", "have", " ", "to", " ", "be", " ", "much", " ", 
     "more", " ", 
     RowBox[{"careful", ".", " "}]}]}], TraditionalForm]],ExpressionUUID->
  "bc18fc9e-44da-4c18-9aed-e9012f35fdfe"]
}], "MathCaption",
 CellID->1307962225,ExpressionUUID->"4d8a8a9f-a481-4b6e-add1-68202ae98216"],

Cell[CellGroupData[{

Cell["Several vbundles:", "MathCaption",
 CellChangeTimes->{3.728671516094545*^9},
 CellID->1278504883,ExpressionUUID->"9ed37ed2-f8da-44e5-9c2a-5bf264ca1685"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TableOfComponents", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"v", "[", "a", "]"}], " ", 
    RowBox[{"u", "[", "A", "]"}]}], ",", " ", 
   RowBox[{"{", 
    RowBox[{"A", ",", "fcomp"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "blues"}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.72874831918683*^9, 3.728748321682487*^9}, 
   3.728749348404161*^9},
 CellLabel->"In[248]:=",
 CellID->110196699,ExpressionUUID->"00771092-e168-4e7e-9fa9-56e7de2b0166"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"u", 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"-", "1"}],
                FontColor->RGBColor[0., 1., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[{-1, $CellContext`fcomp}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{0, $CellContext`blues}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"u", 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"-", "1"}],
                FontColor->RGBColor[0., 1., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[{-1, $CellContext`fcomp}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{1, $CellContext`blues}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"u", 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"-", "1"}],
                FontColor->RGBColor[0., 1., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[{-1, $CellContext`fcomp}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{2, $CellContext`blues}],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"u", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[{1, $CellContext`fcomp}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{0, $CellContext`blues}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"u", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[{1, $CellContext`fcomp}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{1, $CellContext`blues}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"u", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[{1, $CellContext`fcomp}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{2, $CellContext`blues}],
       Editable->False]}]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728671517528159*^9, 3.72867491108456*^9, 
  3.72874332432487*^9, 3.728748322149275*^9, 3.7287496219942007`*^9, 
  3.728749958998665*^9, 3.728751146413595*^9, 3.7287516492686987`*^9, 
  3.7288132946540413`*^9, 3.728827414062829*^9, 3.7288360037143993`*^9, 
  3.728836320288702*^9, 3.728843207425775*^9, 3.7288456568675127`*^9, 
  3.7333674670132217`*^9, 3.792273473136998*^9},
 CellLabel->"Out[248]=",
 CellID->670682407,ExpressionUUID->"7f05672e-95a7-4ac1-8bef-7dc2a4793217"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
These functions do not take the symmetries into account automatically, but we \
can simplify:\
\>", "MathCaption",
 CellChangeTimes->{3.7286715231021757`*^9, 3.733367472779055*^9},
 CellID->101228641,ExpressionUUID->"be925756-e373-4e37-906a-7fad107dbc08"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"ComponentArray", "[", 
    RowBox[{"T", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"a", ",", "fred"}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"b", ",", "fred"}], "}"}]}], "]"}], "]"}], "//", 
   "ToCanonical"}], "//", "MatrixForm"}]], "Input",
 CellChangeTimes->{{3.7287483241483583`*^9, 3.728748327842355*^9}},
 CellLabel->"In[249]:=",
 CellID->542720863,ExpressionUUID->"bdf73f85-7693-432f-be78-2125d0961215"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {"0", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, $CellContext`fred}, {1, $CellContext`fred}],
       Editable->False], 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, $CellContext`fred}, {2, $CellContext`fred}],
       Editable->False]},
     {
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, $CellContext`fred}, {1, $CellContext`fred}],
        Editable->False]}], "0", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, $CellContext`fred}, {2, $CellContext`fred}],
       Editable->False]},
     {
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, $CellContext`fred}, {2, $CellContext`fred}],
        Editable->False]}], 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{1, $CellContext`fred}, {2, $CellContext`fred}],
        Editable->False]}], "0"}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Baseline}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{3.728671524349931*^9, 3.728674911380134*^9, 
  3.728743324478451*^9, 3.728748328239881*^9, 3.7287496221667643`*^9, 
  3.728749959103977*^9, 3.72875114658739*^9, 3.7287516494067287`*^9, 
  3.7288132947821712`*^9, 3.72882741419106*^9, 3.728836003881256*^9, 
  3.7288363208728037`*^9, 3.7288432086685658`*^9, 3.728845656994191*^9, 
  3.73336746735211*^9, 3.7922734732057037`*^9},
 CellLabel->"Out[249]//MatrixForm=",
 CellID->17551346,ExpressionUUID->"049e18c5-a65e-4ee0-b410-e703b2943f67"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "Currently there is no function to construct a single component, because the \
notation must be nearly as long as the direct construction itself! ",
 Cell[BoxData[
  FormBox[
   StyleBox["Coba", "MR"], TraditionalForm]],ExpressionUUID->
  "85aea221-05ee-438c-8443-6a30be28e7f0"],
 " can be used to do that, in case the same component must be constructed for \
several different tensors."
}], "Text",
 CellChangeTimes->{{3.728671537917933*^9, 3.728671539070142*^9}},
 CellID->1928806140,ExpressionUUID->"9a804f81-43a4-4c2e-a320-bbe5d64d36c8"]
}, Closed]],

Cell["1.4.2. ThreadComponentArray. TODOcument", "Subsubsection",
 CellChangeTimes->{{3.72866399283917*^9, 3.728663999804419*^9}, {
  3.728671576811432*^9, 3.728671576931362*^9}, {3.728673461460437*^9, 
  3.728673461620924*^9}},
 CellID->1179143515,ExpressionUUID->"1370a689-69b9-4350-86c5-fcf7ccf3e772"]
}, Closed]],

Cell[CellGroupData[{

Cell["1.5. Tensor densities; \[Eta] and \[Epsilon] tensors and determinants", \
"Subsection",
 CellChangeTimes->{{3.728663818308868*^9, 3.728663848059477*^9}, {
  3.7286705385882177`*^9, 3.728670538740139*^9}, {3.728673422694458*^9, 
  3.728673432348941*^9}},ExpressionUUID->"bf8a23c1-c6a3-4c9a-9947-\
82b3277d77db"],

Cell[CellGroupData[{

Cell["1.5.1. The \[Eta] tensors and the Jacobian", "Subsubsection",
 CellChangeTimes->{{3.728663829532053*^9, 3.728663849803274*^9}, {
  3.728670543491893*^9, 3.728670543859935*^9}, {3.7286734676598263`*^9, 
  3.728673467780589*^9}},ExpressionUUID->"fb3b1d67-18c9-47cc-8570-\
50ccfec54332"],

Cell[TextData[{
 StyleBox["EtaUp  \t\t\t\t", "DisplayMath"],
 StyleBox["Define a basis", "Text"],
 StyleBox["\nEtaDown\t\t\t\t", "DisplayMath"],
 StyleBox["VBundle on which a basis lives\n", "Text"],
 StyleBox["Jacobian\t\t\t", "DisplayMath"],
 StyleBox["Parallel derivative associated to the given basis\n", "Text"],
 StyleBox["WeightOf\t\t\t", "DisplayMath"],
 StyleBox["Weight of an expression as a linear combination of basis names", 
  "Text"]
}], "DefinitionBox",ExpressionUUID->"74ab3fbb-d0e0-409d-8631-e0e5123e9fac"],

Cell["Definition of a basis.", "Text",ExpressionUUID->"04f2cbef-3c0a-4be6-9e9d-d48ec264fe11"],

Cell[TextData[{
 "The \[Eta] tensors represent the totally antisymmetric products of all \
basis (co)vectors:\n\t\t",
 Cell[BoxData[
  FormBox[
   SuperscriptBox[Cell[TextData[Cell[BoxData[
     FormBox[
      OverscriptBox["\[Eta]", 
       StyleBox["~",
        FontColor->RGBColor[1, 0, 0]]], TraditionalForm]],ExpressionUUID->
     "bdec2b0a-1523-4d71-aac7-fc6fea6a3c94"]],ExpressionUUID->
     "f6ba45bc-7f17-49a5-8381-133d2e08a0e6"], 
    RowBox[{
     SubscriptBox["a", "1"], "\[CenterEllipsis]", "  ", 
     SubscriptBox["a", "n"]}]], TraditionalForm]],ExpressionUUID->
  "502bdcbe-7fcd-4042-a655-331d322ebe2e"],
 "= ",
 StyleBox["n",
  FontSlant->"Italic"],
 "! ",
 Cell[BoxData[
  FormBox[
   SubscriptBox[
    SuperscriptBox["e", 
     RowBox[{"[", 
      SubscriptBox["a", "1"]}]], 
    StyleBox["1",
     FontColor->RGBColor[1, 0, 0]]], TraditionalForm]],ExpressionUUID->
  "10b644cc-ef86-438d-a35a-7dac40dec90c"],
 " \[CenterEllipsis] ",
 Cell[BoxData[
  FormBox[
   SubscriptBox[
    SuperscriptBox["e", 
     RowBox[{
      SubscriptBox["a", "n"], "]"}]], 
    StyleBox["n",
     FontColor->RGBColor[1, 0, 0]]], TraditionalForm]],ExpressionUUID->
  "b2e74dfb-8321-4a12-8871-c3f5c2917d35"],
 ", \t\t",
 Cell[BoxData[
  FormBox[
   SubscriptBox[Cell[TextData[Cell[BoxData[
     FormBox[
      UnderscriptBox["\[Eta]", 
       StyleBox["~",
        FontColor->RGBColor[1, 0, 0]]], TraditionalForm]],ExpressionUUID->
     "bdec2b0a-1523-4d71-aac7-fc6fea6a3c94"]],ExpressionUUID->
     "f6ba45bc-7f17-49a5-8381-133d2e08a0e6"], 
    RowBox[{
     SubscriptBox["a", "1"], "\[CenterEllipsis]", "  ", 
     SubscriptBox["a", "n"]}]], TraditionalForm]],ExpressionUUID->
  "0767f9ec-2846-4b26-9bad-50df11547618"],
 "= ",
 StyleBox["n",
  FontSlant->"Italic"],
 "! ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox[
    SubscriptBox["e", 
     RowBox[{"[", 
      SubscriptBox["a", "1"]}]], 
    StyleBox["1",
     FontColor->RGBColor[1, 0, 0]]], TraditionalForm]],ExpressionUUID->
  "5d4b8501-ba83-4756-898d-d99962919096"],
 " \[CenterEllipsis]  ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox[
    SubscriptBox["e", 
     RowBox[{
      SubscriptBox["a", "n"], "]"}]], 
    StyleBox["n",
     FontColor->RGBColor[1, 0, 0]]], TraditionalForm]],ExpressionUUID->
  "52b159f0-60d9-4c27-9966-5933d19e74eb"],
 "\nIn ",
 StyleBox["xCoba`", "MR"],
 ", these objects are automatically defined for each basis. "
}], "Text",ExpressionUUID->"d47a1a03-58bf-4b62-ae1c-5c235726fe35"],

Cell[CellGroupData[{

Cell["Totally antisymmetric densities for a given basis:", "MathCaption",
 CellChangeTimes->{{3.7286699846537647`*^9, 3.728670008789217*^9}, {
  3.7286700533298197`*^9, 3.728670080258039*^9}},
 CellID->1161546665,ExpressionUUID->"14b3003d-f554-43e0-977f-89e94c4e3efe"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"etaUpfred", "[", 
    RowBox[{"a", ",", "b", ",", "c"}], "]"}], ",", 
   RowBox[{"etaDownfred", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}]}], "}"}]], "Input",
 CellChangeTimes->{{3.728748341113676*^9, 3.728748346625393*^9}},
 CellLabel->
  "In[250]:=",ExpressionUUID->"ed4d8f17-15ea-47dd-bcab-170938f72991"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        OverscriptBox["\[Eta]", 
         StyleBox["~",
          FontColor->RGBColor[1, 0, 0]]], 
        StyleBox[GridBox[{
           {"a", "b", "c"},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`etaUpfred[$CellContext`a, $CellContext`b, $CellContext`c],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        UnderscriptBox["\[Eta]", 
         StyleBox["~",
          FontColor->RGBColor[1, 0, 0]]], 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"a", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`etaDownfred[-$CellContext`a, -$CellContext`b, \
-$CellContext`c],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728670010256935*^9, 3.728670180483173*^9, 
  3.728670310587813*^9, 3.728671584762743*^9, 3.728674911580784*^9, 
  3.728743324609325*^9, 3.728748347105888*^9, 3.7287496268873*^9, 
  3.728749959208276*^9, 3.728751157132765*^9, 3.728751649538288*^9, 
  3.7288132948901*^9, 3.7288274143354187`*^9, 3.728836004018483*^9, 
  3.728836321030897*^9, 3.728843220513522*^9, 3.728845657152334*^9, 
  3.733367482295369*^9, 3.792273487506834*^9},
 CellLabel->"Out[250]=",
 CellID->964835428,ExpressionUUID->"e7c34ff0-99e6-4d6d-b313-41f914f3529f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The basis is marked with an over or undertilde, whose significance we shall \
see below. Their components in the natural basis are very simple:\
\>", "MathCaption",ExpressionUUID->"185d76a5-30d5-4edc-a9f5-f96b6ac9c6aa"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"etaUpfred", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "fred"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"2", ",", "fred"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"3", ",", "fred"}], "}"}]}], "]"}], ",", "\[IndentingNewLine]", 
   
   RowBox[{"etaUpfred", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "fred"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"3", ",", "fred"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"2", ",", "fred"}], "}"}]}], "]"}], ",", "\[IndentingNewLine]", 
   
   RowBox[{"etaUpfred", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"3", ",", "fred"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", "fred"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"2", ",", "fred"}], "}"}]}], "]"}], ",", "\[IndentingNewLine]", 
   
   RowBox[{"etaUpfred", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "fred"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", "fred"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", "fred"}], "}"}]}], "]"}]}], "}"}]], "Input",
 CellChangeTimes->{{3.7287483508588467`*^9, 3.72874838342385*^9}},
 CellLabel->
  "In[251]:=",ExpressionUUID->"4396c0c7-a62d-4ef6-a1c0-45f107b5a961"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"1", ",", 
   RowBox[{"-", "1"}], ",", "1", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.7286700251137457`*^9, 3.728671584947259*^9, 
  3.7286749117220507`*^9, 3.728743324754252*^9, 3.728748383881031*^9, 
  3.728749627041582*^9, 3.728749959324502*^9, 3.72875115728973*^9, 
  3.728751649669057*^9, 3.728813295002202*^9, 3.7288274144498672`*^9, 
  3.728836004157168*^9, 3.72883632121029*^9, 3.728843220713006*^9, 
  3.728845657270832*^9, 3.7333674824140177`*^9, 3.7922734875662737`*^9},
 CellLabel->"Out[251]=",
 CellID->340019899,ExpressionUUID->"8d31b179-551a-41b0-8873-0d11a4a2a404"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["And they are related to the generalised Kronecker delta:", "MathCaption",ExpressionUUID->"37fba9d6-71e7-452b-a456-01a40c8cdd3a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"etaDownfred", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}]}], "]"}], 
  RowBox[{"etaUpfred", "[", 
   RowBox[{"d", ",", "e", ",", "f"}], "]"}], " "}]], "Input",
 CellChangeTimes->{{3.728748388087571*^9, 3.728748392471561*^9}},
 CellLabel->
  "In[252]:=",ExpressionUUID->"e0b6311e-131d-4b16-861f-a2732904b3a0"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"\[Delta]", 
      StyleBox[GridBox[{
         {" ", " ", " ", "d", "e", "f"},
         {"a", "b", "c", " ", " ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`Gdelta[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
$CellContext`d, $CellContext`e, $CellContext`f],
  Editable->False]], "Output",
 CellChangeTimes->{3.72867003248414*^9, 3.7286715850942707`*^9, 
  3.728674911857893*^9, 3.728743324891079*^9, 3.7287483929987497`*^9, 
  3.728749627188569*^9, 3.728749959423993*^9, 3.728751157444481*^9, 
  3.728751649804253*^9, 3.728813295121743*^9, 3.728827414551198*^9, 
  3.728836004314186*^9, 3.728836321362741*^9, 3.72884322089648*^9, 
  3.7288456574065*^9, 3.733367482549287*^9, 3.792273487621863*^9},
 CellLabel->"Out[252]=",
 CellID->260457826,ExpressionUUID->"f5d7bc51-3def-45a2-85b2-1a52b54ef786"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "The presence of a metric provides additional relations, as we shall see \
later. Each basis has a different pair of \[Eta] tensors but they are all \
related through the Jacobians of the transformations. Because of this, \[Eta] \
are generally considered not tensors, but tensor densities with weight = \
\[PlusMinus] basis. In ",
 StyleBox["xCoba`", "MR"],
 " we can also consider them tensors, but dependent on a reference basis. If \
we change the reference basis (e.g., from etaUppolar to etaUpcartesian), we \
have to include Jacobians in the transformation. However, if we don't change \
the reference basis (e.g., from etaUppolar[{a,polar},{b, polar}, {c,polar}] \
to etaUppolar[{a,cartesian}, {b,cartesian}, {c, cartesian}]) they transform \
as tensors. This is analogous to our treatment of Christoffel tensors."
}], "Text",
 CellChangeTimes->{{3.728670096081162*^9, 
  3.728670152222863*^9}},ExpressionUUID->"e3602400-dc7f-4031-b615-\
65c120129823"],

Cell[CellGroupData[{

Cell["Consider this example:", "MathCaption",
 CellChangeTimes->{{3.7286701390066023`*^9, 3.728670143927023*^9}},
 CellID->1185430083,ExpressionUUID->"e59b8230-2ee5-42ce-a6c0-1632b86a9e9f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"etaUpfred", "[", 
   RowBox[{"a", ",", "b", ",", "c"}], "]"}], " ", 
  RowBox[{"etaDownblues", "[", 
   RowBox[{
    RowBox[{"-", "d"}], ",", 
    RowBox[{"-", "e"}], ",", 
    RowBox[{"-", "f"}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.728748395839053*^9, 3.728748401342956*^9}, 
   3.7287493502007017`*^9},
 CellLabel->
  "In[253]:=",ExpressionUUID->"a54f58ac-59ae-4bb6-ad3e-b65d3bdb6919"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       UnderscriptBox["\[Eta]", 
        StyleBox["~",
         FontColor->RGBColor[0, 0, 1]]], 
       StyleBox[GridBox[{
          {" ", " ", " "},
          {"d", "e", "f"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`etaDownblues[-$CellContext`d, -$CellContext`e, \
-$CellContext`f],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       OverscriptBox["\[Eta]", 
        StyleBox["~",
         FontColor->RGBColor[1, 0, 0]]], 
       StyleBox[GridBox[{
          {"a", "b", "c"},
          {" ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`etaUpfred[$CellContext`a, $CellContext`b, $CellContext`c],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7286702901995087`*^9, 3.7286703277185717`*^9, 
  3.728671585225504*^9, 3.728674911988743*^9, 3.728743325040968*^9, 
  3.728748401831388*^9, 3.728749627473755*^9, 3.728749959521137*^9, 
  3.728751157577244*^9, 3.72875164993403*^9, 3.728813295239181*^9, 
  3.728827414668948*^9, 3.728836004475925*^9, 3.7288363215125427`*^9, 
  3.728843221363309*^9, 3.728845657553028*^9, 3.733367482646468*^9, 
  3.7922734876757803`*^9},
 CellLabel->"Out[253]=",
 CellID->943008996,ExpressionUUID->"25cb10e8-1285-4442-acef-570e0f7ca5b7"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
This product is a density of weight +1 if we change the basis polar and of \
weight -1 if we change cartesian. In order to convey all this information, we \
represent weigths as linear combinations of the names of the bases:\
\>", "MathCaption",
 CellChangeTimes->{{3.728670272199699*^9, 
  3.728670285959838*^9}},ExpressionUUID->"7bb252c9-8ba4-4561-8a4e-\
846ece9e4d4d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"WeightOf", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[254]:=",ExpressionUUID->"02facd93-f3a9-43b6-979b-1f7a27c3d8a4"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", "blues"}], "+", "fred"}]], "Output",
 CellChangeTimes->{3.728670292750895*^9, 3.728670329467754*^9, 
  3.7286715853552923`*^9, 3.728674912122794*^9, 3.728743325171607*^9, 
  3.728748405599909*^9, 3.728749627703281*^9, 3.7287499596403303`*^9, 
  3.728751157725252*^9, 3.728751650072158*^9, 3.728813295352097*^9, 
  3.728827414782785*^9, 3.728836004635786*^9, 3.728836321661598*^9, 
  3.7288432218185387`*^9, 3.7288456576697493`*^9, 3.733367482818327*^9, 
  3.7922734877216263`*^9},
 CellLabel->"Out[254]=",
 CellID->722880912,ExpressionUUID->"4314cf9e-c8ee-43ae-964f-16f9212cb173"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "If we worked with a single basis, a weight ",
 StyleBox["n",
  FontSlant->"Italic"],
 " *",
 StyleBox["basisname", "MR"],
 " could be understood as the traditional integer weight ",
 StyleBox["n",
  FontSlant->"Italic"]
}], "MathCaption",ExpressionUUID->"34dddd89-5f25-4f76-b7bc-a8ee870a622e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"WeightOf", "[", 
  RowBox[{
   RowBox[{"etaUpfred", "[", 
    RowBox[{"a", ",", "b", ",", "c"}], "]"}], " ", 
   RowBox[{"etaUpfred", "[", 
    RowBox[{"d", ",", "e", ",", "f"}], "]"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728748409478752*^9, 3.7287484140944843`*^9}},
 CellLabel->
  "In[255]:=",ExpressionUUID->"78ae2b9b-dbd0-42d4-b2c3-18e9aa6b2481"],

Cell[BoxData[
 RowBox[{"2", " ", "fred"}]], "Output",
 CellChangeTimes->{3.728670331421989*^9, 3.728671585578971*^9, 
  3.72867491225494*^9, 3.728743325306254*^9, 3.7287484145378*^9, 
  3.728749627833127*^9, 3.728749959740272*^9, 3.728751157998699*^9, 
  3.7287516502045107`*^9, 3.728813295469331*^9, 3.728827414906907*^9, 
  3.728836004763771*^9, 3.7288363218125677`*^9, 3.728843222081087*^9, 
  3.728845657803467*^9, 3.7333674830511703`*^9, 3.7922734877759457`*^9},
 CellLabel->"Out[255]=",
 CellID->1144496905,ExpressionUUID->"a23da994-ae82-4ffb-83dc-9c4ff32800cf"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The Jacobian is represented as", "Text",ExpressionUUID->"9ed8ea1c-31fa-49ab-8bcb-9e158bf253da"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Jacobian", "[", 
   RowBox[{"blues", ",", "fred"}], "]"}], "[", "]"}]], "Input",
 CellChangeTimes->{{3.728748417830709*^9, 3.72874842019788*^9}, 
   3.728749354090314*^9},
 CellLabel->
  "In[256]:=",ExpressionUUID->"5d5b0b0a-de0a-4fd2-b045-31f1c0ba3e54"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>", "\[InvisibleSpace]", 
   RowBox[{"Jacobianbluesfred", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "", 
   $CellContext`Jacobianbluesfred[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728748420634597*^9, 3.728749627949717*^9, 
  3.728749959851079*^9, 3.728751158436325*^9, 3.728751650334216*^9, 
  3.728813295582685*^9, 3.728827415030851*^9, 3.728836004921782*^9, 
  3.728836321960047*^9, 3.728843222253745*^9, 3.728845657934849*^9, 
  3.7333674832167253`*^9, 3.792273487823332*^9},
 CellLabel->"During evaluation of In[256]:=",
 CellID->548456156,ExpressionUUID->"21665ddb-161d-406c-a456-be506dc15de6"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   UnderscriptBox[
    OverscriptBox["J", 
     StyleBox["~",
      FontColor->RGBColor[0, 0, 1]]], 
    StyleBox["~",
     FontColor->RGBColor[1, 0, 0]]],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Jacobianbluesfred[],
  Editable->False]], "Output",
 CellChangeTimes->{3.7286703332998247`*^9, 3.728671586167886*^9, 
  3.7286749124209967`*^9, 3.728743325478663*^9, 3.728748420652157*^9, 
  3.728749627966134*^9, 3.728749959870327*^9, 3.728751158454108*^9, 
  3.7287516503536654`*^9, 3.728813295603465*^9, 3.728827415054614*^9, 
  3.728836004943618*^9, 3.728836321987294*^9, 3.728843222276063*^9, 
  3.728845657959165*^9, 3.7333674832334843`*^9, 3.792273487834279*^9},
 CellLabel->"Out[256]=",
 CellID->902800968,ExpressionUUID->"8241cd92-b3c7-46bd-9c43-6717a3fdb1d3"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
It is a scalar with weight +1 with respect to cartesian and -1 with respect \
to polar, so it has two tildes of different colours.\
\>", "MathCaption",ExpressionUUID->"431b34bd-cb1f-478e-8796-fb15ee589f08"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"WeightOf", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[257]:=",ExpressionUUID->"271545d4-e9c0-43fa-94bd-a37c24f9b26d"],

Cell[BoxData[
 RowBox[{"blues", "-", "fred"}]], "Output",
 CellChangeTimes->{3.728670336985979*^9, 3.728671586751782*^9, 
  3.728674912540073*^9, 3.728743325624661*^9, 3.728748423786683*^9, 
  3.7287496283721867`*^9, 3.728749959973968*^9, 3.728751158606455*^9, 
  3.728751650487162*^9, 3.7288132957203693`*^9, 3.728827415173235*^9, 
  3.72883600508567*^9, 3.728836322142974*^9, 3.728843222425103*^9, 
  3.7288456580869493`*^9, 3.733367483329784*^9, 3.7922734878913183`*^9},
 CellLabel->"Out[257]=",
 CellID->1403903121,ExpressionUUID->"501e1d33-d5f6-4510-9a62-54e7b00f34dc"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"Jacobian", "[", 
    RowBox[{"blues", ",", "fred"}], "]"}], "[", "]"}], "*", 
  RowBox[{
   RowBox[{"Jacobian", "[", 
    RowBox[{"fred", ",", "blues"}], "]"}], "[", "]"}]}]], "Input",
 CellChangeTimes->{{3.7287484265035057`*^9, 3.728748435597313*^9}, {
  3.728749358334024*^9, 3.728749358967559*^9}},
 CellLabel->
  "In[258]:=",ExpressionUUID->"24596909-d185-49f0-976b-5dc0cf7df19e"],

Cell[BoxData["1"], "Output",
 CellChangeTimes->{3.728670339320879*^9, 3.728671586899951*^9, 
  3.728674912672673*^9, 3.728743325810277*^9, 3.7287484362183113`*^9, 
  3.728749628645359*^9, 3.72874996009086*^9, 3.728751158857567*^9, 
  3.7287516506206617`*^9, 3.728813295843527*^9, 3.7288274152850924`*^9, 
  3.728836005225264*^9, 3.7288363222952623`*^9, 3.72884322259158*^9, 
  3.7288456582364807`*^9, 3.733367483496088*^9, 3.792273487934677*^9},
 CellLabel->"Out[258]=",
 CellID->996925058,ExpressionUUID->"652af091-a720-43ff-8b23-d266bbbf4dd2"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["xCoba`", "MR"],
 "  knows how to compute derivatives of Jacobians"
}], "MathCaption",ExpressionUUID->"b99d7649-2df5-4466-97e6-6e483720fc64"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"PD", "[", 
    RowBox[{"-", "a"}], "]"}], "[", 
   RowBox[{
    RowBox[{"Jacobian", "[", 
     RowBox[{"blues", ",", "fred"}], "]"}], "[", "]"}], "]"}], "//", 
  "ScreenDollarIndices"}]], "Input",
 CellChangeTimes->{{3.728748440246842*^9, 3.7287484439810038`*^9}, 
   3.728749360075082*^9},
 CellLabel->
  "In[259]:=",ExpressionUUID->"c4351d2b-cfa6-4687-b65f-6ab086b02c07"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[0., 0., 1.]], ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
        StyleBox[GridBox[{
           {"b", " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelPDbluesPDfred[$CellContext`b, -$CellContext`a, \
-$CellContext`b],
    Editable->False]}], " ", 
  InterpretationBox[
   StyleBox[
    UnderscriptBox[
     OverscriptBox["J", 
      StyleBox["~",
       FontColor->RGBColor[0, 0, 1]]], 
     StyleBox["~",
      FontColor->RGBColor[1, 0, 0]]],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Jacobianbluesfred[],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728670348686851*^9, 3.728671587046874*^9, 
  3.728674912807398*^9, 3.728743325977075*^9, 3.7287484444019403`*^9, 
  3.728749628830327*^9, 3.728749960192029*^9, 3.728751159227942*^9, 
  3.728751650754743*^9, 3.7288132959894876`*^9, 3.728827415408124*^9, 
  3.728836005381989*^9, 3.728836322443844*^9, 3.728843223014773*^9, 
  3.7288456583563843`*^9, 3.7333674837133923`*^9, 3.792273487992374*^9},
 CellLabel->"Out[259]=",
 CellID->652035912,ExpressionUUID->"450246e7-fcfc-418b-a043-a43f526c1b7b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"LieD", "[", 
    RowBox[{"v", "[", "a", "]"}], "]"}], "[", 
   RowBox[{
    RowBox[{"Jacobian", "[", 
     RowBox[{"blues", ",", "fred"}], "]"}], "[", "]"}], "]"}], "//", 
  "ScreenDollarIndices"}]], "Input",
 CellChangeTimes->{{3.728748447822567*^9, 3.7287484506205387`*^9}, 
   3.728749360696756*^9},
 CellLabel->
  "In[260]:=",ExpressionUUID->"b84656c0-60d8-43ee-a023-996654ffbd93"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    UnderscriptBox[
     OverscriptBox["J", 
      StyleBox["~",
       FontColor->RGBColor[0, 0, 1]]], 
     StyleBox["~",
      FontColor->RGBColor[1, 0, 0]]],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Jacobianbluesfred[],
   Editable->False], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]]},
              {"a", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[-$CellContext`a, {$CellContext`b, \
$CellContext`blues}],
       Editable->False]}], " ", 
     RowBox[{"(", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SubscriptBox[
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox["b",
           FontColor->RGBColor[0., 0., 1.]]], GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {"a"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PDblues[{-$CellContext`b, -$CellContext`blues}][
        $CellContext`v[$CellContext`a]],
       Editable->False], ")"}]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", "a"},
             {
              StyleBox["c",
               FontColor->RGBColor[1., 0., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{-$CellContext`c, -$CellContext`fred}, \
$CellContext`a],
      Editable->False], " ", 
     RowBox[{"(", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SubscriptBox[
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[1., 0., 0.]], "a"], GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["c",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PDfred[-$CellContext`a][
        $CellContext`v[{$CellContext`c, $CellContext`fred}]],
       Editable->False], ")"}]}]}], ")"}]}]], "Output",
 CellChangeTimes->{3.7286703496357927`*^9, 3.728671587233218*^9, 
  3.728674912941164*^9, 3.728743326147401*^9, 3.728748450988184*^9, 
  3.728749629316369*^9, 3.728749960304288*^9, 3.728751159816908*^9, 
  3.728751650917398*^9, 3.7288132960897284`*^9, 3.7288274155455227`*^9, 
  3.728836005537561*^9, 3.72883632259533*^9, 3.728843224069539*^9, 
  3.7288456585026073`*^9, 3.73336748380752*^9, 3.792273488041759*^9},
 CellLabel->"Out[260]=",
 CellID->158522343,ExpressionUUID->"cfcf1ec8-4ca0-47af-bc5b-05c0a835f9b9"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ContractBasis", "[", 
   RowBox[{"%", ",", 
    RowBox[{"OverDerivatives", "\[Rule]", "True"}]}], "]"}], "//", 
  "Simplify"}]], "Input",
 CellLabel->
  "In[261]:=",ExpressionUUID->"5a98f0b8-352a-4f23-af7c-bf6e646fdc12"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    UnderscriptBox[
     OverscriptBox["J", 
      StyleBox["~",
       FontColor->RGBColor[0, 0, 1]]], 
     StyleBox["~",
      FontColor->RGBColor[1, 0, 0]]],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Jacobianbluesfred[],
   Editable->False], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"-", 
     RowBox[{"(", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SubscriptBox[
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox["b",
           FontColor->RGBColor[0., 0., 1.]]], GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {
                StyleBox["b",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PDblues[{-$CellContext`b, -$CellContext`blues}][
        $CellContext`v[{$CellContext`b, $CellContext`blues}]],
       Editable->False], ")"}]}], "+", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox[
        StyleBox["\[ScriptCapitalD]",
         FontColor->RGBColor[1., 0., 0.]], 
        StyleBox["c",
         FontColor->RGBColor[1., 0., 0.]]], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["c",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDfred[{-$CellContext`c, -$CellContext`fred}][
      $CellContext`v[{$CellContext`c, $CellContext`fred}]],
     Editable->False]}], ")"}]}]], "Output",
 CellChangeTimes->{3.728670350288319*^9, 3.728671587421577*^9, 
  3.728674913073295*^9, 3.728743326274417*^9, 3.728748453936343*^9, 
  3.728749629487898*^9, 3.728749960439065*^9, 3.7287511599854116`*^9, 
  3.728751651053056*^9, 3.7288132961920967`*^9, 3.728827415671682*^9, 
  3.7288360056676598`*^9, 3.7288363227453012`*^9, 3.728843224265073*^9, 
  3.7288456586411123`*^9, 3.733367483909152*^9, 3.792273488094461*^9},
 CellLabel->"Out[261]=",
 CellID->1405296521,ExpressionUUID->"ccb9b423-b386-413e-9da1-1329d0aa03e8"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["1.5.2. Determinants", "Subsubsection",
 CellChangeTimes->{{3.72866383169213*^9, 3.7286638519310713`*^9}, {
  3.728670544963614*^9, 3.728670545051759*^9}, {3.728673468932129*^9, 
  3.728673469068563*^9}},ExpressionUUID->"5517a51f-3f93-4b9d-b5fc-\
8bd9fd3d7328"],

Cell[TextData[{
 StyleBox["Determinant\t\t\t", "DisplayMath"],
 StyleBox["Determinant of a tensor of arbitrary rank", "Text"]
}], "DefinitionBox",
 CellChangeTimes->{
  3.8278334063370523`*^9},ExpressionUUID->"ce361ebb-e766-46f2-9ed6-\
359ba6ef7034"],

Cell["Determinants", "Text",ExpressionUUID->"fed624a6-ada2-4fc2-9ac7-d778d9a990c1"],

Cell[CellGroupData[{

Cell["\<\
We can define the determinant of a tensor in any given basis through the \
\[Eta] tensors:\
\>", "MathCaption",ExpressionUUID->"1ee6647d-fcf0-431d-89d3-dd6e56aa114a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Determinant", "[", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{"a", ",", "b"}], "]"}], ",", "fred"}], "]"}]], "Input",
 CellChangeTimes->{{3.728748462583021*^9, 3.728748463083766*^9}},
 CellLabel->
  "In[262]:=",ExpressionUUID->"046b7ec2-b105-4604-b965-62d269f2b606"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", "6"], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       UnderscriptBox["\[Eta]", 
        StyleBox["~",
         FontColor->RGBColor[1, 0, 0]]], 
       StyleBox[GridBox[{
          {" ", " ", " "},
          {"a", "c", "e"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`etaDownfred[-$CellContext`a, -$CellContext`c, -$CellContext`e],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       UnderscriptBox["\[Eta]", 
        StyleBox["~",
         FontColor->RGBColor[1, 0, 0]]], 
       StyleBox[GridBox[{
          {" ", " ", " "},
          {"b", "d", "f"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`etaDownfred[-$CellContext`b, -$CellContext`d, -$CellContext`f],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"a", "b"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[$CellContext`a, $CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"c", "d"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[$CellContext`c, $CellContext`d],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"e", "f"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[$CellContext`e, $CellContext`f],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7286703720006313`*^9, 3.728671592403536*^9, 
  3.728674913204527*^9, 3.7287433264389267`*^9, 3.7287484635132713`*^9, 
  3.728749960543789*^9, 3.728751163850298*^9, 3.728751651191609*^9, 
  3.7288132963057203`*^9, 3.7288274157970123`*^9, 3.728836005819973*^9, 
  3.728836322916732*^9, 3.728843229027981*^9, 3.728845658770472*^9, 
  3.7333675027675037`*^9, 3.7922734931541967`*^9},
 CellLabel->"Out[262]=",
 CellID->1137660123,ExpressionUUID->"5c8574a2-f98f-458b-9f6a-b7bc131c45a5"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"WeightOf", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[263]:=",ExpressionUUID->"68c42f31-701f-4bee-b53c-795b37a11c55"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", "2"}], " ", "fred"}]], "Output",
 CellChangeTimes->{3.72867037297845*^9, 3.728671592602455*^9, 
  3.7286749133222113`*^9, 3.728743326553973*^9, 3.728748465940275*^9, 
  3.728749960654914*^9, 3.728751164035919*^9, 3.728751651333404*^9, 
  3.728813296422131*^9, 3.728827415920388*^9, 3.7288360059606113`*^9, 
  3.7288363231011667`*^9, 3.72884322923096*^9, 3.728845658919772*^9, 
  3.7333675028821583`*^9, 3.7922734932184668`*^9},
 CellLabel->"Out[263]=",
 CellID->1606778020,ExpressionUUID->"7777eebe-9070-4b80-a118-ee765acb7cbb"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Notice how the determinant depends on the basis, unless there are the same \
number of contravariant and covariant indices:\
\>", "MathCaption",
 CellChangeTimes->{{3.728670379747571*^9, 
  3.72867038034758*^9}},ExpressionUUID->"f78e620b-acda-4245-b00f-\
2488b34148f7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"T2", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", "M3", ",", 
   RowBox[{"PrintAs", "\[Rule]", " ", "\"\<T\>\""}]}], "]"}]], "Input",
 CellLabel->
  "In[264]:=",ExpressionUUID->"cda11638-b2eb-4771-ba40-f008d61c2e6f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"T2", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`T2[$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286703817468*^9, 3.728671592780036*^9, 
  3.7286749134564447`*^9, 3.728743326671267*^9, 3.728748468131248*^9, 
  3.728749960773644*^9, 3.72875116420334*^9, 3.7287516514677963`*^9, 
  3.728813296536545*^9, 3.728827416034101*^9, 3.728836006117331*^9, 
  3.728836323290147*^9, 3.7288432294256363`*^9, 3.7288456590534678`*^9, 
  3.733367503002145*^9, 3.792273493270707*^9},
 CellLabel->"During evaluation of In[264]:=",
 CellID->511321115,ExpressionUUID->"94fdcefb-5765-447b-8ea6-2d75f4a67c55"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Determinant", "[", 
  RowBox[{
   RowBox[{"T2", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", "fred"}], "]"}]], "Input",
 CellChangeTimes->{{3.7287484698535423`*^9, 3.7287484703717413`*^9}},
 CellLabel->
  "In[265]:=",ExpressionUUID->"e0ad47c0-69fe-45f1-92a4-f060b74ff0cd"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   FractionBox["1", "3"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"a", " "},
           {" ", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T2[$CellContext`a, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"b", " "},
           {" ", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T2[$CellContext`b, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"c", " "},
           {" ", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T2[$CellContext`c, -$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "3"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"a", " "},
           {" ", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T2[$CellContext`a, -$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"b", " "},
           {" ", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T2[$CellContext`b, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"c", " "},
           {" ", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T2[$CellContext`c, -$CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "6"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"a", " "},
           {" ", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T2[$CellContext`a, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"b", " "},
           {" ", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T2[$CellContext`b, -$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"c", " "},
           {" ", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T2[$CellContext`c, -$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "6"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"a", " "},
           {" ", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T2[$CellContext`a, -$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"b", " "},
           {" ", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T2[$CellContext`b, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"c", " "},
           {" ", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T2[$CellContext`c, -$CellContext`c],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.7286703828216753`*^9, 3.7286715929849567`*^9, 
  3.728674913611912*^9, 3.728743326813878*^9, 3.7287484707871304`*^9, 
  3.72874996091145*^9, 3.728751165187043*^9, 3.728751651647854*^9, 
  3.728813296666399*^9, 3.728827416184472*^9, 3.728836006291769*^9, 
  3.7288363234946117`*^9, 3.728843230801852*^9, 3.728845659198772*^9, 
  3.733367503126793*^9, 3.7922734933426943`*^9},
 CellLabel->"Out[265]=",
 CellID->1691908158,ExpressionUUID->"c3a75e22-732d-4004-9bd3-8dd35fb07671"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"WeightOf", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[266]:=",ExpressionUUID->"57e16bda-76e0-483c-a625-10b95c98af0f"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728670383590355*^9, 3.728671593285857*^9, 
  3.728674913755302*^9, 3.728743326957507*^9, 3.728748472742032*^9, 
  3.7287499610050364`*^9, 3.728751165403256*^9, 3.728751651769556*^9, 
  3.728813296785904*^9, 3.728827416306316*^9, 3.7288360064411097`*^9, 
  3.728836323654097*^9, 3.7288432310858107`*^9, 3.728845659354039*^9, 
  3.733367503266953*^9, 3.7922734933908567`*^9},
 CellLabel->"Out[266]=",
 CellID->1069441579,ExpressionUUID->"81515935-6e43-4529-a547-22f014e03517"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
If a tensor has an odd number of indices, its determinant is zero:\
\>", "MathCaption",
 CellChangeTimes->{
  3.7286703882831297`*^9},ExpressionUUID->"c4608ec9-3c78-44ef-abf4-\
25d9a8472eeb"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"T3", "[", 
    RowBox[{"a", ",", "b", ",", "c"}], "]"}], ",", "M3", ",", 
   RowBox[{"PrintAs", "\[Rule]", " ", "\"\<T\>\""}]}], "]"}]], "Input",
 CellLabel->
  "In[267]:=",ExpressionUUID->"e86cb3df-05dd-41d8-ac6a-a89ad460ef59"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"T3", "[", 
    RowBox[{"a", ",", "b", ",", "c"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`T3[$CellContext`a, $CellContext`b, $CellContext`c], ". ", ""],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.728670389558528*^9, 3.7286715938199587`*^9, 
  3.7286749138913527`*^9, 3.728743327106645*^9, 3.728748475378729*^9, 
  3.7287499611246233`*^9, 3.728751165570314*^9, 3.728751651886614*^9, 
  3.728813296903171*^9, 3.728827416403016*^9, 3.728836006601118*^9, 
  3.728836323826735*^9, 3.728843231292109*^9, 3.728845659484486*^9, 
  3.733367503416334*^9, 3.792273493439432*^9},
 CellLabel->"During evaluation of In[267]:=",
 CellID->1813976510,ExpressionUUID->"1b44471d-711a-4483-a5b9-c83d386f498c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Determinant", "[", 
  RowBox[{
   RowBox[{"T3", "[", 
    RowBox[{"a", ",", "b", ",", "c"}], "]"}], ",", "blues"}], "]"}]], "Input",\

 CellChangeTimes->{{3.728748477180602*^9, 3.7287484777954206`*^9}, 
   3.728749361449823*^9},
 CellLabel->
  "In[268]:=",ExpressionUUID->"ce66d50b-96ce-4366-88e7-6b95d92ec609"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", "6"], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       UnderscriptBox["\[Eta]", 
        StyleBox["~",
         FontColor->RGBColor[0, 0, 1]]], 
       StyleBox[GridBox[{
          {" ", " ", " "},
          {"a", "d", "f1"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`etaDownblues[-$CellContext`a, -$CellContext`d, \
-$CellContext`f1],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       UnderscriptBox["\[Eta]", 
        StyleBox["~",
         FontColor->RGBColor[0, 0, 1]]], 
       StyleBox[GridBox[{
          {" ", " ", " "},
          {"b", "e", "f2"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`etaDownblues[-$CellContext`b, -$CellContext`e, \
-$CellContext`f2],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       UnderscriptBox["\[Eta]", 
        StyleBox["~",
         FontColor->RGBColor[0, 0, 1]]], 
       StyleBox[GridBox[{
          {" ", " ", " "},
          {"c", "f", "f3"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`etaDownblues[-$CellContext`c, -$CellContext`f, \
-$CellContext`f3],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"a", "b", "c"},
          {" ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T3[$CellContext`a, $CellContext`b, $CellContext`c],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"d", "e", "f"},
          {" ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T3[$CellContext`d, $CellContext`e, $CellContext`f],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"f1", "f2", "f3"},
          {" ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T3[$CellContext`f1, $CellContext`f2, $CellContext`f3],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7286703910489397`*^9, 3.728671594122151*^9, 
  3.728674914039226*^9, 3.728743327303488*^9, 3.72874847824238*^9, 
  3.728749961245771*^9, 3.728751165770466*^9, 3.7287516520079927`*^9, 
  3.728813297011063*^9, 3.72882741653125*^9, 3.728836006750703*^9, 
  3.7288363240075808`*^9, 3.72884323152045*^9, 3.728845659610579*^9, 
  3.7333675036364927`*^9, 3.792273493502963*^9},
 CellLabel->"Out[268]=",
 CellID->1825350758,ExpressionUUID->"d8aaef10-c90d-4fbe-af02-3a7945b42916"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCanonical", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[269]:=",ExpressionUUID->"717aa808-c21c-473c-af3f-f0f8e2fe6458"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728670391825135*^9, 3.728671594283904*^9, 
  3.728674914173649*^9, 3.728743327461998*^9, 3.728748480379766*^9, 
  3.728749961372636*^9, 3.728751165935973*^9, 3.728751652118746*^9, 
  3.728813297120755*^9, 3.728827416654682*^9, 3.728836006903596*^9, 
  3.728836324180806*^9, 3.7288432317377367`*^9, 3.728845659736987*^9, 
  3.733367503890842*^9, 3.792273493555167*^9},
 CellLabel->"Out[269]=",
 CellID->1866292311,ExpressionUUID->"dfdfc557-41c2-4c38-a24a-04cf5e19d0d7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"T4", "[", 
    RowBox[{"a", ",", "b", ",", "c", ",", 
     RowBox[{"-", "d"}]}], "]"}], ",", "M3", ",", 
   RowBox[{"PrintAs", "\[Rule]", " ", "\"\<T\>\""}]}], "]"}]], "Input",
 CellLabel->
  "In[270]:=",ExpressionUUID->"63ef81ea-9788-45f6-bc02-779cf7eff6b9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"T4", "[", 
    RowBox[{"a", ",", "b", ",", "c", ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`T4[$CellContext`a, $CellContext`b, $CellContext`c, \
-$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728670393221095*^9, 3.7286715944364977`*^9, 
  3.728674914306179*^9, 3.728743327652094*^9, 3.728748482341395*^9, 
  3.728749961473918*^9, 3.7287511661042023`*^9, 3.728751652255419*^9, 
  3.7288132972201757`*^9, 3.7288274167688627`*^9, 3.728836007088993*^9, 
  3.728836324346279*^9, 3.728843231942986*^9, 3.72884565987484*^9, 
  3.733367504093351*^9, 3.792273493614723*^9},
 CellLabel->"During evaluation of In[270]:=",
 CellID->628722895,ExpressionUUID->"0038d2a2-a361-465b-a662-665bbe416ee8"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Determinant", "[", 
   RowBox[{
    RowBox[{"T4", "[", 
     RowBox[{"a", ",", "b", ",", "c", ",", 
      RowBox[{"-", "d"}]}], "]"}], ",", "blues"}], "]"}], "//", 
  "ToCanonical"}]], "Input",
 CellChangeTimes->{{3.728748483580381*^9, 3.7287484841950483`*^9}, 
   3.7287493622756443`*^9},
 CellLabel->
  "In[271]:=",ExpressionUUID->"dbedcf26-963c-4331-b9fa-fe6e6c6933ea"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   FractionBox["1", "3"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        UnderscriptBox["\[Eta]", 
         StyleBox["~",
          FontColor->RGBColor[0, 0, 1]]], 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"b", "e", "f2"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`etaDownblues[-$CellContext`b, -$CellContext`e, \
-$CellContext`f2],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        UnderscriptBox["\[Eta]", 
         StyleBox["~",
          FontColor->RGBColor[0, 0, 1]]], 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"c", "f", "f3"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`etaDownblues[-$CellContext`c, -$CellContext`f, \
-$CellContext`f3],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"a", "b", "c", " "},
           {" ", " ", " ", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T4[$CellContext`a, $CellContext`b, $CellContext`c, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"d", "e", "f", " "},
           {" ", " ", " ", "f1"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T4[$CellContext`d, $CellContext`e, $CellContext`f, \
-$CellContext`f1],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"f1", "f2", "f3", " "},
           {" ", " ", " ", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T4[$CellContext`f1, $CellContext`f2, $CellContext`f3, \
-$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        UnderscriptBox["\[Eta]", 
         StyleBox["~",
          FontColor->RGBColor[0, 0, 1]]], 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"b", "e", "f2"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`etaDownblues[-$CellContext`b, -$CellContext`e, \
-$CellContext`f2],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        UnderscriptBox["\[Eta]", 
         StyleBox["~",
          FontColor->RGBColor[0, 0, 1]]], 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"c", "f", "f3"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`etaDownblues[-$CellContext`c, -$CellContext`f, \
-$CellContext`f3],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"a", "b", "c", " "},
           {" ", " ", " ", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T4[$CellContext`a, $CellContext`b, $CellContext`c, \
-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"d", "e", "f", " "},
           {" ", " ", " ", "f1"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T4[$CellContext`d, $CellContext`e, $CellContext`f, \
-$CellContext`f1],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"f1", "f2", "f3", " "},
           {" ", " ", " ", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T4[$CellContext`f1, $CellContext`f2, $CellContext`f3, \
-$CellContext`d],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "6"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        UnderscriptBox["\[Eta]", 
         StyleBox["~",
          FontColor->RGBColor[0, 0, 1]]], 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"b", "e", "f2"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`etaDownblues[-$CellContext`b, -$CellContext`e, \
-$CellContext`f2],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        UnderscriptBox["\[Eta]", 
         StyleBox["~",
          FontColor->RGBColor[0, 0, 1]]], 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"c", "f", "f3"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`etaDownblues[-$CellContext`c, -$CellContext`f, \
-$CellContext`f3],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"a", "b", "c", " "},
           {" ", " ", " ", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T4[$CellContext`a, $CellContext`b, $CellContext`c, \
-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"d", "e", "f", " "},
           {" ", " ", " ", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T4[$CellContext`d, $CellContext`e, $CellContext`f, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"f1", "f2", "f3", " "},
           {" ", " ", " ", "f1"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T4[$CellContext`f1, $CellContext`f2, $CellContext`f3, \
-$CellContext`f1],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728670395612712*^9, 3.72867159546*^9, 3.7286749152331*^9,
   3.728743327803164*^9, 3.7287484854053183`*^9, 3.7287499624144897`*^9, 
  3.728751167069385*^9, 3.728751653167213*^9, 3.72881329810942*^9, 
  3.728827417646289*^9, 3.728836008112145*^9, 3.728836325385541*^9, 
  3.728843233084774*^9, 3.728845660864455*^9, 3.733367505060858*^9, 
  3.792273493760167*^9},
 CellLabel->"Out[271]=",
 CellID->1326153025,ExpressionUUID->"c4c9d5c7-dc02-463e-8f98-a55a051b5f74"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Even if we know that the determinant of a given object is going to be basis \
independent, we must provide some basis:\
\>", "MathCaption",ExpressionUUID->"89958e6f-5fbd-4853-a72e-6a4f7ed97fb2"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"Determinant", "[", 
    RowBox[{"delta", "[", 
     RowBox[{"a", ",", 
      RowBox[{"-", "b"}]}], "]"}], "]"}], ",", " ", 
   RowBox[{"Determinant", "[", 
    RowBox[{
     RowBox[{"delta", "[", 
      RowBox[{"a", ",", 
       RowBox[{"-", "b"}]}], "]"}], ",", "fred"}], "]"}]}], "}"}]], "Input",
 CellChangeTimes->{{3.72874848939496*^9, 3.728748490178731*^9}},
 CellLabel->
  "In[272]:=",ExpressionUUID->"e6c42b3a-8b82-4028-bdcd-f312501f4739"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"Determinant", "[", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {"a", " "},
            {" ", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`delta[$CellContext`a, -$CellContext`b],
     Editable->False], "]"}], ",", "1"}], "}"}]], "Output",
 CellChangeTimes->{3.7286704011803837`*^9, 3.72867159568145*^9, 
  3.728674915359578*^9, 3.728743327905566*^9, 3.728748490493862*^9, 
  3.728749962523612*^9, 3.7287511672823668`*^9, 3.7287516533041573`*^9, 
  3.728813298228282*^9, 3.728827417785646*^9, 3.728836008275041*^9, 
  3.728836325551958*^9, 3.728843233604061*^9, 3.728845660991474*^9, 
  3.733367505217865*^9, 3.792273493809404*^9},
 CellLabel->"Out[272]=",
 CellID->2068602609,ExpressionUUID->"5634f899-488f-47d1-b3e6-23cbd878590c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"UndefTensor", " ", "/@", " ", 
   RowBox[{"{", 
    RowBox[{"T2", ",", "T3", ",", "T4"}], "}"}]}], ";"}]], "Input",
 CellLabel->
  "In[273]:=",ExpressionUUID->"cfc3cd91-deba-4600-88a6-33745a56dfd6"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "T2"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`T2],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286704020477667`*^9, 3.7286715958306227`*^9, 
  3.728674915503469*^9, 3.728743328025218*^9, 3.7287484924607277`*^9, 
  3.7287499626371*^9, 3.728751167442008*^9, 3.7287516534314003`*^9, 
  3.728813298336175*^9, 3.728827417911641*^9, 3.728836008418909*^9, 
  3.728836325707951*^9, 3.7288432337899923`*^9, 3.7288456611032867`*^9, 
  3.733367505336726*^9, 3.792273493859112*^9},
 CellLabel->"During evaluation of In[273]:=",
 CellID->1391288180,ExpressionUUID->"bbc3bfd4-4779-4636-8405-7b21fd588dbc"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "T3"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`T3],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286704020477667`*^9, 3.7286715958306227`*^9, 
  3.728674915503469*^9, 3.728743328025218*^9, 3.7287484924607277`*^9, 
  3.7287499626371*^9, 3.728751167442008*^9, 3.7287516534314003`*^9, 
  3.728813298336175*^9, 3.728827417911641*^9, 3.728836008418909*^9, 
  3.728836325707951*^9, 3.7288432337899923`*^9, 3.7288456611032867`*^9, 
  3.733367505336726*^9, 3.7922734938693867`*^9},
 CellLabel->"During evaluation of In[273]:=",
 CellID->1578492724,ExpressionUUID->"665c9837-2fbb-4aeb-ba7a-00ef22965ea5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "T4"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`T4],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286704020477667`*^9, 3.7286715958306227`*^9, 
  3.728674915503469*^9, 3.728743328025218*^9, 3.7287484924607277`*^9, 
  3.7287499626371*^9, 3.728751167442008*^9, 3.7287516534314003`*^9, 
  3.728813298336175*^9, 3.728827417911641*^9, 3.728836008418909*^9, 
  3.728836325707951*^9, 3.7288432337899923`*^9, 3.7288456611032867`*^9, 
  3.733367505336726*^9, 3.792273493880909*^9},
 CellLabel->"During evaluation of In[273]:=",
 CellID->1861720877,ExpressionUUID->"2be4aebe-ae6e-4589-9dc7-04764a8bd818"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["1.5.3. The determinant of the metric and the \[Epsilon] tensor", \
"Subsubsection",
 CellChangeTimes->{{3.7286638341400023`*^9, 3.728663853555193*^9}, {
  3.728670546107547*^9, 3.7286705464359207`*^9}, {3.72867347002808*^9, 
  3.728673470196589*^9}},ExpressionUUID->"08ed6c8f-96d6-4476-b7e7-\
d08c642f69c5"],

Cell[TextData[{
 StyleBox["epsilonToetaDown  \t\t\t\t", "DisplayMath"],
 StyleBox["Transform an \[Epsilon] tensor into an etaUp tensor", "Text"],
 StyleBox["\nepsilonToetaUp\t\t\t\t\t", "DisplayMath"],
 StyleBox["Transform an \[Epsilon] tensor into an etaDown tensor\n", "Text"],
 StyleBox["etaDownToepsilon\t\t\t\t", "DisplayMath"],
 StyleBox["Transform an etaDown tensor into an \[Epsilon] tensor\n", "Text"],
 StyleBox["etaUpToepsilon\t\t\t\t\t", "DisplayMath"],
 StyleBox["Transform an etaUp tensor into an \[Epsilon] tensor\n", "Text"],
 StyleBox["epsilonOrientation\t\t\t\t", "DisplayMath"],
 StyleBox["Factor between the \[Epsilon] tensor and the eta tensors", "Text"]
}], "DefinitionBox",ExpressionUUID->"8c8d68fd-236d-44bf-af7b-3c24667ca857"],

Cell["Determinant of the metric; relation between \[Eta] and \[Epsilon] \
tensors", "Text",
 CellChangeTimes->{
  3.82783341295997*^9},ExpressionUUID->"5105a549-923b-4ea5-a757-3a48359d7237"],

Cell[CellGroupData[{

Cell[TextData[{
 "The determinant of the metric is a weight 2 density. When a metric is \
defined, its determinant in the priviledged basis ",
 StyleBox["AIndex", "MR"],
 " is also defined. We also parametrize the orientation of the volume form of \
the metric with respect to the eta tensors of a basis:"
}], "MathCaption",ExpressionUUID->"ce172b65-2844-4ea8-b50c-3d50b2e93fcf"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefMetric", "[", 
  RowBox[{
   RowBox[{"-", "1"}], ",", 
   RowBox[{"metric", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", " ", "cd", ",", 
   RowBox[{"{", 
    RowBox[{"\"\<|\>\"", ",", "\"\<D\>\""}], "}"}], ",", 
   RowBox[{"PrintAs", "\[Rule]", " ", "\"\<g\>\""}], ",", 
   RowBox[{"epsilonOrientationInBasis", "->", 
    RowBox[{"{", 
     RowBox[{"blues", ",", "\[Sigma]"}], "}"}]}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728748505203576*^9, 3.7287485058084803`*^9}, 
   3.72874936373693*^9},
 CellLabel->
  "In[274]:=",ExpressionUUID->"dc93e8bb-70c6-43e2-a7b8-02251ad53bf8"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric metric tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"metric", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric metric tensor ", 
   $CellContext`metric[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728748506698633*^9, 3.728749635482196*^9, 
  3.728749962819729*^9, 3.7287511729638567`*^9, 3.7287516536139917`*^9, 
  3.7288132985430107`*^9, 3.728827418130249*^9, 3.7288360086334887`*^9, 
  3.728836325956868*^9, 3.72884323962825*^9, 3.728845661347815*^9, 
  3.733367510651116*^9, 3.792273499911399*^9},
 CellLabel->"During evaluation of In[274]:=",
 CellID->220771680,ExpressionUUID->"12cf3852-20ca-4b93-a629-98596fe330a0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric tensor \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"epsilonmetric", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric tensor ", 
   $CellContext`epsilonmetric[-$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728748506698633*^9, 3.728749635482196*^9, 
  3.728749962819729*^9, 3.7287511729638567`*^9, 3.7287516536139917`*^9, 
  3.7288132985430107`*^9, 3.728827418130249*^9, 3.7288360086334887`*^9, 
  3.728836325956868*^9, 3.72884323962825*^9, 3.728845661347815*^9, 
  3.733367510651116*^9, 3.79227349992587*^9},
 CellLabel->"During evaluation of In[274]:=",
 CellID->1805367391,ExpressionUUID->"1089b79d-0f1e-40f7-8e01-7e507598e02a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"cd", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "covariant derivative ", 
   $CellContext`cd[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728748506698633*^9, 3.728749635482196*^9, 
  3.728749962819729*^9, 3.7287511729638567`*^9, 3.7287516536139917`*^9, 
  3.7288132985430107`*^9, 3.728827418130249*^9, 3.7288360086334887`*^9, 
  3.728836325956868*^9, 3.72884323962825*^9, 3.728845661347815*^9, 
  3.733367510651116*^9, 3.792273499946129*^9},
 CellLabel->"During evaluation of In[274]:=",
 CellID->1013292074,ExpressionUUID->"8a0a04c4-67af-4a95-b85b-89a1c813c827"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing torsion tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"Torsioncd", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing torsion tensor ", 
   $CellContext`Torsioncd[$CellContext`a, -$CellContext`b, -$CellContext`c], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728748506698633*^9, 3.728749635482196*^9, 
  3.728749962819729*^9, 3.7287511729638567`*^9, 3.7287516536139917`*^9, 
  3.7288132985430107`*^9, 3.728827418130249*^9, 3.7288360086334887`*^9, 
  3.728836325956868*^9, 3.72884323962825*^9, 3.728845661347815*^9, 
  3.733367510651116*^9, 3.792273499954308*^9},
 CellLabel->"During evaluation of In[274]:=",
 CellID->854567008,ExpressionUUID->"acd942e6-516d-4aa8-bd06-32ea0e885eaf"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"Christoffelcd", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Christoffel tensor ", 
   $CellContext`Christoffelcd[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728748506698633*^9, 3.728749635482196*^9, 
  3.728749962819729*^9, 3.7287511729638567`*^9, 3.7287516536139917`*^9, 
  3.7288132985430107`*^9, 3.728827418130249*^9, 3.7288360086334887`*^9, 
  3.728836325956868*^9, 3.72884323962825*^9, 3.728845661347815*^9, 
  3.733367510651116*^9, 3.792273499962491*^9},
 CellLabel->"During evaluation of In[274]:=",
 CellID->337072320,ExpressionUUID->"54f2a242-e079-4505-9005-8aab8226bb34"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"Riemanncd", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Riemann tensor ", 
   $CellContext`Riemanncd[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728748506698633*^9, 3.728749635482196*^9, 
  3.728749962819729*^9, 3.7287511729638567`*^9, 3.7287516536139917`*^9, 
  3.7288132985430107`*^9, 3.728827418130249*^9, 3.7288360086334887`*^9, 
  3.728836325956868*^9, 3.72884323962825*^9, 3.728845661347815*^9, 
  3.733367510651116*^9, 3.792273499970208*^9},
 CellLabel->"During evaluation of In[274]:=",
 CellID->112621454,ExpressionUUID->"a22bf354-1fe1-4bf2-9fb8-91f4211cb35f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"Riccicd", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric Ricci tensor ", 
   $CellContext`Riccicd[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728748506698633*^9, 3.728749635482196*^9, 
  3.728749962819729*^9, 3.7287511729638567`*^9, 3.7287516536139917`*^9, 
  3.7288132985430107`*^9, 3.728827418130249*^9, 3.7288360086334887`*^9, 
  3.728836325956868*^9, 3.72884323962825*^9, 3.728845661347815*^9, 
  3.733367510651116*^9, 3.7922734999771347`*^9},
 CellLabel->"During evaluation of In[274]:=",
 CellID->895684371,ExpressionUUID->"d88ea636-1de4-4542-88d9-a7e380e683c0"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Riemann automatically replaced \
by Ricci.\"\>"], "Print",
 CellChangeTimes->{3.728748506698633*^9, 3.728749635482196*^9, 
  3.728749962819729*^9, 3.7287511729638567`*^9, 3.7287516536139917`*^9, 
  3.7288132985430107`*^9, 3.728827418130249*^9, 3.7288360086334887`*^9, 
  3.728836325956868*^9, 3.72884323962825*^9, 3.728845661347815*^9, 
  3.733367510651116*^9, 3.7922734999849167`*^9},
 CellLabel->"During evaluation of In[274]:=",
 CellID->2054341574,ExpressionUUID->"f779c985-a204-420c-b48b-630f38aaf9f9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"RicciScalarcd", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Ricci scalar ", 
   $CellContext`RicciScalarcd[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728748506698633*^9, 3.728749635482196*^9, 
  3.728749962819729*^9, 3.7287511729638567`*^9, 3.7287516536139917`*^9, 
  3.7288132985430107`*^9, 3.728827418130249*^9, 3.7288360086334887`*^9, 
  3.728836325956868*^9, 3.72884323962825*^9, 3.728845661347815*^9, 
  3.733367510651116*^9, 3.792273499992525*^9},
 CellLabel->"During evaluation of In[274]:=",
 CellID->76757752,ExpressionUUID->"65de0311-db07-41fd-a4d6-206e365601ee"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Ricci automatically replaced \
by RicciScalar.\"\>"], "Print",
 CellChangeTimes->{3.728748506698633*^9, 3.728749635482196*^9, 
  3.728749962819729*^9, 3.7287511729638567`*^9, 3.7287516536139917`*^9, 
  3.7288132985430107`*^9, 3.728827418130249*^9, 3.7288360086334887`*^9, 
  3.728836325956868*^9, 3.72884323962825*^9, 3.728845661347815*^9, 
  3.733367510651116*^9, 3.792273499999135*^9},
 CellLabel->"During evaluation of In[274]:=",
 CellID->1577349664,ExpressionUUID->"b387ee00-c2b7-48f9-a4a0-a06511077084"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Einstein tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"Einsteincd", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Einstein tensor ", 
   $CellContext`Einsteincd[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728748506698633*^9, 3.728749635482196*^9, 
  3.728749962819729*^9, 3.7287511729638567`*^9, 3.7287516536139917`*^9, 
  3.7288132985430107`*^9, 3.728827418130249*^9, 3.7288360086334887`*^9, 
  3.728836325956868*^9, 3.72884323962825*^9, 3.728845661347815*^9, 
  3.733367510651116*^9, 3.7922735000070057`*^9},
 CellLabel->"During evaluation of In[274]:=",
 CellID->1977664280,ExpressionUUID->"a07d72c8-2dd7-4bbf-b03c-ca868444d488"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Weyl tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"Weylcd", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Weyl tensor ", 
   $CellContext`Weylcd[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728748506698633*^9, 3.728749635482196*^9, 
  3.728749962819729*^9, 3.7287511729638567`*^9, 3.7287516536139917`*^9, 
  3.7288132985430107`*^9, 3.728827418130249*^9, 3.7288360086334887`*^9, 
  3.728836325956868*^9, 3.72884323962825*^9, 3.728845661347815*^9, 
  3.733367510651116*^9, 3.792273500015333*^9},
 CellLabel->"During evaluation of In[274]:=",
 CellID->1999149341,ExpressionUUID->"0d721454-298f-446e-b758-17cb9bb2bedb"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric TFRicci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TFRiccicd", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric TFRicci tensor ", 
   $CellContext`TFRiccicd[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728748506698633*^9, 3.728749635482196*^9, 
  3.728749962819729*^9, 3.7287511729638567`*^9, 3.7287516536139917`*^9, 
  3.7288132985430107`*^9, 3.728827418130249*^9, 3.7288360086334887`*^9, 
  3.728836325956868*^9, 3.72884323962825*^9, 3.728845661347815*^9, 
  3.733367510651116*^9, 3.7922735000235167`*^9},
 CellLabel->"During evaluation of In[274]:=",
 CellID->173833225,ExpressionUUID->"541f3685-75f9-4cd6-94c9-76b8b8428b78"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Kretschmann scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"Kretschmanncd", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Kretschmann scalar ", 
   $CellContext`Kretschmanncd[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728748506698633*^9, 3.728749635482196*^9, 
  3.728749962819729*^9, 3.7287511729638567`*^9, 3.7287516536139917`*^9, 
  3.7288132985430107`*^9, 3.728827418130249*^9, 3.7288360086334887`*^9, 
  3.728836325956868*^9, 3.72884323962825*^9, 3.728845661347815*^9, 
  3.733367510651116*^9, 3.7922735000321074`*^9},
 CellLabel->"During evaluation of In[274]:=",
 CellID->1037252655,ExpressionUUID->"c79da98e-8db3-4437-8f75-a2df17e40c4f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RiemannToWeylRules for dim \"\>", 
   "\[InvisibleSpace]", "3"}],
  SequenceForm["** DefCovD:  Computing RiemannToWeylRules for dim ", 3],
  Editable->False]], "Print",
 CellChangeTimes->{3.728748506698633*^9, 3.728749635482196*^9, 
  3.728749962819729*^9, 3.7287511729638567`*^9, 3.7287516536139917`*^9, 
  3.7288132985430107`*^9, 3.728827418130249*^9, 3.7288360086334887`*^9, 
  3.728836325956868*^9, 3.72884323962825*^9, 3.728845661347815*^9, 
  3.733367510651116*^9, 3.792273500039359*^9},
 CellLabel->"During evaluation of In[274]:=",
 CellID->1980609886,ExpressionUUID->"668fe0d6-1e96-4567-9ffe-8d1e1d63497e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RicciToTFRicci for dim \"\>", 
   "\[InvisibleSpace]", "3"}],
  SequenceForm["** DefCovD:  Computing RicciToTFRicci for dim ", 3],
  Editable->False]], "Print",
 CellChangeTimes->{3.728748506698633*^9, 3.728749635482196*^9, 
  3.728749962819729*^9, 3.7287511729638567`*^9, 3.7287516536139917`*^9, 
  3.7288132985430107`*^9, 3.728827418130249*^9, 3.7288360086334887`*^9, 
  3.728836325956868*^9, 3.72884323962825*^9, 3.728845661347815*^9, 
  3.733367510651116*^9, 3.792273500046151*^9},
 CellLabel->"During evaluation of In[274]:=",
 CellID->2113825502,ExpressionUUID->"e50dd1e1-627d-47b8-a5fb-6d3f16646765"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RicciToEinsteinRules for dim \"\>", 
   "\[InvisibleSpace]", "3"}],
  SequenceForm["** DefCovD:  Computing RicciToEinsteinRules for dim ", 3],
  Editable->False]], "Print",
 CellChangeTimes->{3.728748506698633*^9, 3.728749635482196*^9, 
  3.728749962819729*^9, 3.7287511729638567`*^9, 3.7287516536139917`*^9, 
  3.7288132985430107`*^9, 3.728827418130249*^9, 3.7288360086334887`*^9, 
  3.728836325956868*^9, 3.72884323962825*^9, 3.728845661347815*^9, 
  3.733367510651116*^9, 3.7922735000538263`*^9},
 CellLabel->"During evaluation of In[274]:=",
 CellID->1243239793,ExpressionUUID->"1c26a392-4e91-4378-b46e-660f8aa71e26"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density \"\>", "\[InvisibleSpace]", 
   RowBox[{"Detmetric", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Determinant.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "weight +2 density ", 
   $CellContext`Detmetric[], ". ", "Determinant."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728748506698633*^9, 3.728749635482196*^9, 
  3.728749962819729*^9, 3.7287511729638567`*^9, 3.7287516536139917`*^9, 
  3.7288132985430107`*^9, 3.728827418130249*^9, 3.7288360086334887`*^9, 
  3.728836325956868*^9, 3.72884323962825*^9, 3.728845661347815*^9, 
  3.733367510651116*^9, 3.792273500061393*^9},
 CellLabel->"During evaluation of In[274]:=",
 CellID->115315641,ExpressionUUID->"1dececb5-a6c8-46ee-910a-3be27b28079f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Detmetric", "[", "]"}]], "Input",
 CellLabel->
  "In[275]:=",ExpressionUUID->"5e412601-0c66-493f-b76b-c632608e689d"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   OverscriptBox["g", 
    OverscriptBox[
     StyleBox["~",
      FontColor->RGBColor[0, 0, 0]], 
     StyleBox["~",
      FontColor->RGBColor[0, 0, 0]]]],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Detmetric[],
  Editable->False]], "Output",
 CellChangeTimes->{3.728670424668317*^9, 3.7286716021590652`*^9, 
  3.7286749163384953`*^9, 3.7287433283074427`*^9, 3.7287485102124777`*^9, 
  3.7287496359958897`*^9, 3.728749963134247*^9, 3.72875117346178*^9, 
  3.728751653982265*^9, 3.7288132989020147`*^9, 3.728827418630899*^9, 
  3.728836009124195*^9, 3.728836326424309*^9, 3.7288432408051167`*^9, 
  3.728845661866374*^9, 3.7333675111198397`*^9, 3.792273500153081*^9},
 CellLabel->"Out[275]=",
 CellID->1079728693,ExpressionUUID->"86b01d45-8954-45c9-a3f3-6a446eada101"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["In any other basis:", "MathCaption",ExpressionUUID->"52d194c2-576a-4d83-907e-d0f691c74315"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Determinant", "[", 
   RowBox[{"metric", ",", "fred"}], "]"}], "[", "]"}]], "Input",
 CellChangeTimes->{{3.728748512867516*^9, 3.728748513313608*^9}},
 CellLabel->
  "In[276]:=",ExpressionUUID->"a1372ca5-53aa-45be-aaf1-202a458cd20b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density \"\>", "\[InvisibleSpace]", 
   RowBox[{"Detmetricfred", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Determinant.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "weight +2 density ", 
   $CellContext`Detmetricfred[], ". ", "Determinant."],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287485136308002`*^9, 3.728749643461133*^9, 
  3.728749963256164*^9, 3.728751173855628*^9, 3.728751654119377*^9, 
  3.728813299036016*^9, 3.728827418814692*^9, 3.7288360092750177`*^9, 
  3.7288363266743298`*^9, 3.728843241443462*^9, 3.7288456620498953`*^9, 
  3.7333675115073347`*^9, 3.792273500210033*^9},
 CellLabel->"During evaluation of In[276]:=",
 CellID->763326101,ExpressionUUID->"27c93d52-f301-4620-b6c9-5ec74655f7b6"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   OverscriptBox["g", 
    OverscriptBox[
     StyleBox["~",
      FontColor->RGBColor[1, 0, 0]], 
     StyleBox["~",
      FontColor->RGBColor[1, 0, 0]]]],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Detmetricfred[],
  Editable->False]], "Output",
 CellChangeTimes->{3.728670426020297*^9, 3.7286716023546677`*^9, 
  3.728674916531829*^9, 3.728743328462296*^9, 3.728748513648756*^9, 
  3.728749643478203*^9, 3.728749963270344*^9, 3.728751173873406*^9, 
  3.728751654135865*^9, 3.728813299054452*^9, 3.7288274188357477`*^9, 
  3.728836009300892*^9, 3.728836326696278*^9, 3.728843241468335*^9, 
  3.728845662071618*^9, 3.7333675115238953`*^9, 3.792273500223905*^9},
 CellLabel->"Out[276]=",
 CellID->154472329,ExpressionUUID->"ae418ec8-30cf-4a7f-ad0f-060878bd1755"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Once we have a metric, we can build new tensors (with zero weight) from the \
etaUp and etaDown tensors. These new \[Epsilon] tensors depend only on the \
metric:\
\>", "MathCaption",
 CellChangeTimes->{
  3.72867043552107*^9},ExpressionUUID->"da6a3990-a9a3-402b-8954-32ea402bca5e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"epsilonmetric", "[", 
  RowBox[{"a", ",", "b", ",", "c"}], "]"}]], "Input",
 CellLabel->
  "In[277]:=",ExpressionUUID->"22500a6b-149e-4031-aaca-fb4f124bfce5"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"\[Epsilon]g", 
      StyleBox[GridBox[{
         {"a", "b", "c"},
         {" ", " ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`epsilonmetric[$CellContext`a, $CellContext`b, $CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{3.7286704366516533`*^9, 3.72867160251301*^9, 
  3.728674916655364*^9, 3.728743328605523*^9, 3.728748516293858*^9, 
  3.7287496448073187`*^9, 3.7287499633894787`*^9, 3.728751174488139*^9, 
  3.7287516542689123`*^9, 3.728813299201096*^9, 3.728827418971282*^9, 
  3.728836009454915*^9, 3.7288363268643923`*^9, 3.728843241749498*^9, 
  3.728845662223881*^9, 3.733367511774762*^9, 3.792273500274781*^9},
 CellLabel->"Out[277]=",
 CellID->2056187927,ExpressionUUID->"417e2e6b-0351-489f-a732-904772e59c2b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"WeightOf", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[278]:=",ExpressionUUID->"f971d30c-b3be-4291-9a5a-bc71f2a15b84"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.7286704373146267`*^9, 3.7286716026685743`*^9, 
  3.7286749168030233`*^9, 3.728743328740383*^9, 3.728748517395459*^9, 
  3.728749645368593*^9, 3.7287499635352993`*^9, 3.728751174653966*^9, 
  3.7287516544532623`*^9, 3.728813299385256*^9, 3.7288274190979424`*^9, 
  3.728836009598062*^9, 3.728836327118355*^9, 3.728843241933928*^9, 
  3.728845662365077*^9, 3.7333675119195023`*^9, 3.7922735003310957`*^9},
 CellLabel->"Out[278]=",
 CellID->1043641565,ExpressionUUID->"38476243-64d2-413f-9e6d-527e966c4786"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"etaUp", "[", "blues", "]"}], "[", 
   RowBox[{"a", ",", "b", ",", "c"}], "]"}], "/.", 
  RowBox[{"etaUpToepsilon", "[", 
   RowBox[{"blues", ",", "metric"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.72874851994704*^9, 3.7287485227292013`*^9}, {
  3.728749364642071*^9, 3.728749365524075*^9}},
 CellLabel->
  "In[279]:=",ExpressionUUID->"4316b4ea-0107-4db2-9dc5-626be467ebdb"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density \"\>", "\[InvisibleSpace]", 
   RowBox[{"Detmetricblues", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Determinant.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "weight +2 density ", 
   $CellContext`Detmetricblues[], ". ", "Determinant."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728748523453823*^9, 3.728749645968648*^9, 
  3.728749963702135*^9, 3.72875117480842*^9, 3.72875165460165*^9, 
  3.728813299518766*^9, 3.728827419255136*^9, 3.7288360097519617`*^9, 
  3.728836327405076*^9, 3.72884324222153*^9, 3.728845662517704*^9, 
  3.733367512087075*^9, 3.7922735003790007`*^9},
 CellLabel->"During evaluation of In[279]:=",
 CellID->1984837906,ExpressionUUID->"da78f669-2bdc-4065-a824-380a655aa1c0"],

Cell[BoxData[
 RowBox[{"-", 
  FractionBox[
   RowBox[{
    SqrtBox[
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["g", 
         OverscriptBox[
          StyleBox["~",
           FontColor->RGBColor[0, 0, 1]], 
          StyleBox["~",
           FontColor->RGBColor[0, 0, 1]]]],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Detmetricblues[],
       Editable->False]}]], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Epsilon]g", 
         StyleBox[GridBox[{
            {"a", "b", "c"},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`epsilonmetric[$CellContext`a, $CellContext`b, \
$CellContext`c],
     Editable->False]}], "\[Sigma]"]}]], "Output",
 CellChangeTimes->{3.7286704381176767`*^9, 3.728671602883369*^9, 
  3.728674916988083*^9, 3.728743328880433*^9, 3.7287485235303383`*^9, 
  3.7287496460237913`*^9, 3.728749963718958*^9, 3.72875117486792*^9, 
  3.7287516546187*^9, 3.728813299540885*^9, 3.728827419277317*^9, 
  3.728836009782902*^9, 3.7288363274261627`*^9, 3.7288432423898582`*^9, 
  3.728845662539598*^9, 3.733367512191922*^9, 3.792273500396572*^9},
 CellLabel->"Out[279]=",
 CellID->159713644,ExpressionUUID->"739959c9-3395-4e0c-af25-62471b9f2540"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", 
  RowBox[{"epsilonToetaDown", "[", 
   RowBox[{"metric", ",", "blues"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.728748525962963*^9, 3.728748526520977*^9}, 
   3.728749368975387*^9},
 CellLabel->
  "In[280]:=",ExpressionUUID->"0539c1c0-5a28-485e-9b91-d30ca7209aa5"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    OverscriptBox["g", 
     OverscriptBox[
      StyleBox["~",
       FontColor->RGBColor[0, 0, 1]], 
      StyleBox["~",
       FontColor->RGBColor[0, 0, 1]]]],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Detmetricblues[],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       UnderscriptBox["\[Eta]", 
        StyleBox["~",
         FontColor->RGBColor[0, 0, 1]]], 
       StyleBox[GridBox[{
          {"a", "b", "c"},
          {" ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`etaDownblues[$CellContext`a, $CellContext`b, $CellContext`c],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728670439315012*^9, 3.7286716035325823`*^9, 
  3.728674917122781*^9, 3.728743329029052*^9, 3.7287485304767323`*^9, 
  3.7287496474189463`*^9, 3.728749963839673*^9, 3.728751175248954*^9, 
  3.728751654745501*^9, 3.728813299703236*^9, 3.728827419416347*^9, 
  3.7288360099167747`*^9, 3.7288363276095543`*^9, 3.728843243039102*^9, 
  3.728845662691594*^9, 3.733367512493003*^9, 3.7922735004568644`*^9},
 CellLabel->"Out[280]=",
 CellID->2100139605,ExpressionUUID->"bd7ad3bb-31be-45e5-a910-0af47c6e5c96"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%%", "/.", 
  RowBox[{"epsilonToetaUp", "[", 
   RowBox[{"metric", ",", "blues"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.728748528066579*^9, 3.728748528633094*^9}, 
   3.728749369575933*^9},
 CellLabel->
  "In[281]:=",ExpressionUUID->"27c8ea32-d4df-4a30-9d91-66111c5675b4"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      OverscriptBox["\[Eta]", 
       StyleBox["~",
        FontColor->RGBColor[0, 0, 1]]], 
      StyleBox[GridBox[{
         {"a", "b", "c"},
         {" ", " ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`etaUpblues[$CellContext`a, $CellContext`b, $CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{3.728670440131382*^9, 3.7286716036650963`*^9, 
  3.7286749172705517`*^9, 3.7287433292044067`*^9, 3.728748531398119*^9, 
  3.728749648801202*^9, 3.728749963967782*^9, 3.728751175772224*^9, 
  3.728751654881996*^9, 3.728813299842313*^9, 3.728827419564733*^9, 
  3.728836010082581*^9, 3.728836327807907*^9, 3.7288432432332687`*^9, 
  3.7288456628642263`*^9, 3.733367512651615*^9, 3.792273500506*^9},
 CellLabel->"Out[281]=",
 CellID->1005119143,ExpressionUUID->"d506648c-4660-4180-b745-8dc38590bb0b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The relative orientation of the volume form with respect to the eta tensors \
of a basis is stored in the following function:\
\>", "MathCaption",ExpressionUUID->"c4e154b2-0aeb-4571-b328-ffadd6d548a0"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"epsilonOrientation", "[", 
  RowBox[{"metric", ",", "blues"}], "]"}]], "Input",
 CellChangeTimes->{{3.728748533576858*^9, 3.728748534120748*^9}, 
   3.72874937037632*^9},
 CellLabel->
  "In[282]:=",ExpressionUUID->"5becb215-6358-4923-b8b9-22e02bced121"],

Cell[BoxData["\[Sigma]"], "Output",
 CellChangeTimes->{3.7286704472317343`*^9, 3.728671603853889*^9, 
  3.728674917436536*^9, 3.728743329381081*^9, 3.7287485347816896`*^9, 
  3.728749649885633*^9, 3.728749964119038*^9, 3.728751175952379*^9, 
  3.7287516550132523`*^9, 3.7288132999675083`*^9, 3.728827419731258*^9, 
  3.72883601021563*^9, 3.728836327980287*^9, 3.728843243639215*^9, 
  3.728845663030904*^9, 3.733367512803212*^9, 3.792273500563859*^9},
 CellLabel->"Out[282]=",
 CellID->931069099,ExpressionUUID->"cf83ba9f-5a92-4588-8364-f04d6f579fcd"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefMetric", "[", "metric", "]"}]], "Input",
 CellChangeTimes->{{3.728751185652645*^9, 3.728751189238616*^9}},
 CellLabel->"In[283]:=",
 CellID->859527540,ExpressionUUID->"6d715067-b856-4a86-8dd0-72127094ca28"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Detmetricblues"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "weight +2 density", 
   " ", $CellContext`Detmetricblues],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287511899016037`*^9, 3.728751655147818*^9, 
  3.728813300108953*^9, 3.7288274198959103`*^9, 3.72883601036698*^9, 
  3.7288363281744413`*^9, 3.728843244133231*^9, 3.7288456632158327`*^9, 
  3.733367512954897*^9, 3.792273500619803*^9},
 CellLabel->"During evaluation of In[283]:=",
 CellID->1912494851,ExpressionUUID->"8ead0fa0-a7b3-423e-b0ff-0d1e795c0fc3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Detmetricfred"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "weight +2 density", 
   " ", $CellContext`Detmetricfred],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287511899016037`*^9, 3.728751655147818*^9, 
  3.728813300108953*^9, 3.7288274198959103`*^9, 3.72883601036698*^9, 
  3.7288363281744413`*^9, 3.728843244133231*^9, 3.7288456632158327`*^9, 
  3.733367512954897*^9, 3.792273500631983*^9},
 CellLabel->"During evaluation of In[283]:=",
 CellID->1913216108,ExpressionUUID->"70f2e64e-6b0f-440a-9afd-f117c0027aa0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Detmetric"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "weight +2 density", 
   " ", $CellContext`Detmetric],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287511899016037`*^9, 3.728751655147818*^9, 
  3.728813300108953*^9, 3.7288274198959103`*^9, 3.72883601036698*^9, 
  3.7288363281744413`*^9, 3.728843244133231*^9, 3.7288456632158327`*^9, 
  3.733367512954897*^9, 3.7922735006450033`*^9},
 CellLabel->"During evaluation of In[283]:=",
 CellID->1150480625,ExpressionUUID->"b1f51af3-05ea-41af-9daa-cea7066b3010"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Christoffelcd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Christoffel tensor", " ", $CellContext`Christoffelcd],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287511899016037`*^9, 3.728751655147818*^9, 
  3.728813300108953*^9, 3.7288274198959103`*^9, 3.72883601036698*^9, 
  3.7288363281744413`*^9, 3.728843244133231*^9, 3.7288456632158327`*^9, 
  3.733367512954897*^9, 3.792273500653934*^9},
 CellLabel->"During evaluation of In[283]:=",
 CellID->185106787,ExpressionUUID->"85513c16-34a6-4d9d-846f-4d00831b9c6a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Einstein tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Einsteincd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Einstein tensor", " ", $CellContext`Einsteincd],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287511899016037`*^9, 3.728751655147818*^9, 
  3.728813300108953*^9, 3.7288274198959103`*^9, 3.72883601036698*^9, 
  3.7288363281744413`*^9, 3.728843244133231*^9, 3.7288456632158327`*^9, 
  3.733367512954897*^9, 3.792273500662105*^9},
 CellLabel->"During evaluation of In[283]:=",
 CellID->1533448253,ExpressionUUID->"bd7a2902-aec0-4e91-b06c-ac89a501f889"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Kretschmann scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Kretschmanncd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Kretschmann scalar", 
   " ", $CellContext`Kretschmanncd],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287511899016037`*^9, 3.728751655147818*^9, 
  3.728813300108953*^9, 3.7288274198959103`*^9, 3.72883601036698*^9, 
  3.7288363281744413`*^9, 3.728843244133231*^9, 3.7288456632158327`*^9, 
  3.733367512954897*^9, 3.792273500670382*^9},
 CellLabel->"During evaluation of In[283]:=",
 CellID->713540524,ExpressionUUID->"6b9d1ba8-79b7-4a83-8a4a-bd7a3bd1761a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Riccicd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric Ricci tensor", 
   " ", $CellContext`Riccicd],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287511899016037`*^9, 3.728751655147818*^9, 
  3.728813300108953*^9, 3.7288274198959103`*^9, 3.72883601036698*^9, 
  3.7288363281744413`*^9, 3.728843244133231*^9, 3.7288456632158327`*^9, 
  3.733367512954897*^9, 3.7922735006787786`*^9},
 CellLabel->"During evaluation of In[283]:=",
 CellID->926902020,ExpressionUUID->"6ec40b58-4209-4998-8190-91056e8bbd79"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciScalarcd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Ricci scalar", 
   " ", $CellContext`RicciScalarcd],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287511899016037`*^9, 3.728751655147818*^9, 
  3.728813300108953*^9, 3.7288274198959103`*^9, 3.72883601036698*^9, 
  3.7288363281744413`*^9, 3.728843244133231*^9, 3.7288456632158327`*^9, 
  3.733367512954897*^9, 3.7922735006862783`*^9},
 CellLabel->"During evaluation of In[283]:=",
 CellID->595874802,ExpressionUUID->"0055860d-d7a8-4ea7-84ac-413edc4db234"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Riemanncd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`Riemanncd],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287511899016037`*^9, 3.728751655147818*^9, 
  3.728813300108953*^9, 3.7288274198959103`*^9, 3.72883601036698*^9, 
  3.7288363281744413`*^9, 3.728843244133231*^9, 3.7288456632158327`*^9, 
  3.733367512954897*^9, 3.7922735006939573`*^9},
 CellLabel->"During evaluation of In[283]:=",
 CellID->1151551253,ExpressionUUID->"7d5052ec-b0d7-4750-b76b-0400efe8a8ab"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric TFRicci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TFRiccicd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric TFRicci tensor",
    " ", $CellContext`TFRiccicd],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287511899016037`*^9, 3.728751655147818*^9, 
  3.728813300108953*^9, 3.7288274198959103`*^9, 3.72883601036698*^9, 
  3.7288363281744413`*^9, 3.728843244133231*^9, 3.7288456632158327`*^9, 
  3.733367512954897*^9, 3.7922735007012253`*^9},
 CellLabel->"During evaluation of In[283]:=",
 CellID->999282407,ExpressionUUID->"c447ee50-8a51-4bdd-b6b6-74c5f23167fe"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Torsioncd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`Torsioncd],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287511899016037`*^9, 3.728751655147818*^9, 
  3.728813300108953*^9, 3.7288274198959103`*^9, 3.72883601036698*^9, 
  3.7288363281744413`*^9, 3.728843244133231*^9, 3.7288456632158327`*^9, 
  3.733367512954897*^9, 3.7922735007085114`*^9},
 CellLabel->"During evaluation of In[283]:=",
 CellID->1617294909,ExpressionUUID->"a25e2713-996c-4177-8fb3-b38b9f1e2685"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Weyl tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Weylcd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Weyl tensor", 
   " ", $CellContext`Weylcd],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287511899016037`*^9, 3.728751655147818*^9, 
  3.728813300108953*^9, 3.7288274198959103`*^9, 3.72883601036698*^9, 
  3.7288363281744413`*^9, 3.728843244133231*^9, 3.7288456632158327`*^9, 
  3.733367512954897*^9, 3.792273500715878*^9},
 CellLabel->"During evaluation of In[283]:=",
 CellID->1175015484,ExpressionUUID->"853d1766-95bf-4d37-92bd-b299f7a71590"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "cd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "covariant derivative", 
   " ", $CellContext`cd],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287511899016037`*^9, 3.728751655147818*^9, 
  3.728813300108953*^9, 3.7288274198959103`*^9, 3.72883601036698*^9, 
  3.7288363281744413`*^9, 3.728843244133231*^9, 3.7288456632158327`*^9, 
  3.733367512954897*^9, 3.792273500722967*^9},
 CellLabel->"During evaluation of In[283]:=",
 CellID->225621776,ExpressionUUID->"8a3a4415-e547-4bbd-a098-6cd416d90f81"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "epsilonmetric"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric tensor", 
   " ", $CellContext`epsilonmetric],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287511899016037`*^9, 3.728751655147818*^9, 
  3.728813300108953*^9, 3.7288274198959103`*^9, 3.72883601036698*^9, 
  3.7288363281744413`*^9, 3.728843244133231*^9, 3.7288456632158327`*^9, 
  3.733367512954897*^9, 3.792273500730113*^9},
 CellLabel->"During evaluation of In[283]:=",
 CellID->1716552277,ExpressionUUID->"40cb1ffb-5755-4a82-9a17-f77b235bcb17"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric metric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "metric"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric metric tensor", 
   " ", $CellContext`metric],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287511899016037`*^9, 3.728751655147818*^9, 
  3.728813300108953*^9, 3.7288274198959103`*^9, 3.72883601036698*^9, 
  3.7288363281744413`*^9, 3.728843244133231*^9, 3.7288456632158327`*^9, 
  3.733367512954897*^9, 3.792273500737122*^9},
 CellLabel->"During evaluation of In[283]:=",
 CellID->798805741,ExpressionUUID->"c74045aa-c41f-485e-8dad-06c98e5ab153"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Closed]]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{

Cell["2. Differentiation and charts", "Section",
 CellFrame->{{0, 0}, {0, 0.5}},
 CellChangeTimes->{{3.7286627354220657`*^9, 3.728662748172317*^9}, {
  3.728664037492599*^9, 3.72866403885896*^9}, {3.7286641258241453`*^9, 
  3.728664135390469*^9}, {3.7286704844786243`*^9, 3.728670486574984*^9}, {
  3.728670883885109*^9, 3.72867088543647*^9}, {3.7286734750119667`*^9, 
  3.728673475252084*^9}},
 FontSize->18,
 FontColor->GrayLevel[0],
 CellTags->{"S5.18.1", "1.4"},
 CellID->256747671,ExpressionUUID->"2c215748-a3b5-4b64-a43d-34271bae9190"],

Cell[CellGroupData[{

Cell["2.1. Parallel derivatives and Christoffel tensors", "Subsection",
 CellChangeTimes->{{3.728663816469069*^9, 3.728663844365477*^9}, {
  3.7286705484436493`*^9, 3.728670548795568*^9}, {3.728673477851577*^9, 
  3.728673479131962*^9}},
 CellID->2126653677,ExpressionUUID->"7d2ec21e-6cc6-48a6-86ff-67582dbed596"],

Cell[CellGroupData[{

Cell["2.1.1. Partial derivatives and parallel derivatives", "Subsubsection",
 CellChangeTimes->{{3.728663829532053*^9, 3.728663849803274*^9}, {
  3.728670543491893*^9, 3.728670543859935*^9}, {3.72867306644833*^9, 
  3.728673079878126*^9}, {3.7286734857870407`*^9, 3.728673489676098*^9}},
 CellID->1988076142,ExpressionUUID->"d90e3d53-1dd9-4a77-84ed-9cf49e0d4bec"],

Cell["\<\
xTensor` performs abstract (coordinate-free) computations, so there is no \
real notion of chart in xTensor`. However, xTensor` has a PD operator, mainly \
to be able to convert covariant derivatives of tensors into \"partial \
derivatives\" of tensors plus Christoffel terms. Following Wald, we interpret \
PD as a arbitrary (but fixed) choice of a flat (zero curvature) and \
torsionless covariant derivative. Under such conditions, it is possible to \
find (at least locally) a coordinate system in which such PD covariant \
derivative corresponds directly to the partial derivatives with respect to \
the scalar fields of the chart. In that sense, the arbitrary choice of PD \
represents an arbitrary choice of chart, but we never specify any other chart \
in xTensor`, so there is no notion of \"coordinate change\". xTensor` is (by \
design) very limited regarding coordinate systems, precisely to make sure \
that all your computations are coordinate free.\
\>", "Text",
 CellChangeTimes->{{3.728671702711429*^9, 3.7286720182154493`*^9}},
 CellID->205791133,ExpressionUUID->"689c8572-10e6-4141-bb7a-9591821d06e3"],

Cell["\<\
xCoba` now allows you to declare any number of charts on each given manifold, \
and transition functions among them. Again, this is actually only a local \
statement, and the Manifold objects of xTensor are just open submanifold; \
xAct is not yet able to perform global computations by itself. Associated to \
each chart we will define a \"partial derivate\", which will be implemented \
as a flat and torsionless covariant derivative as well.\
\>", "Text",
 CellChangeTimes->{{3.728672033967682*^9, 3.728672246485546*^9}},
 CellID->1280456169,ExpressionUUID->"a335faf2-7e92-4082-a8c9-6b68db5c1fb2"],

Cell["\<\
xCoba` goes a bit beyond that and generalizes the concept of \"partial \
derivative\" associated to a coordinate system or frame. We will introduce \
the concept of \"parallel derivative\" to any basis of vector fields, both \
coordinated or not. In the former case the concept of \"parallel derivative\" \
collapses to that of \"partial derivative\". In the latter case the \
\"parallel derivative\" is still flat, but now has torsion. We will see that \
such torsion is equivalent to the Lie brackets of the basis vectors.\
\>", "Text",
 CellChangeTimes->{{3.72867226152398*^9, 3.728672396390621*^9}},
 CellID->415868790,ExpressionUUID->"9bf3975a-c17c-403c-b3bd-5288c83dc305"],

Cell["\<\
Finally, any pair of covariant derivatives is related by a Christoffel \
tensor, so in xCoba we will handle routinely Christoffel tensors relating two \
parallel derivatives. These will be Christoffel tensors with two different \
colored derivatives.\
\>", "Text",
 CellChangeTimes->{{3.72867241680641*^9, 3.72867248689818*^9}},
 CellID->133220817,ExpressionUUID->"e3dace28-2d9a-44f7-96f3-83baf0ba3a82"],

Cell["\<\
We use PD... both to mean \"Partial Derivative\" and \"Parallel Derivative\", \
depending on the context. We find this amusing, and hope that doesn't lead to \
confusion.\
\>", "Text",
 CellChangeTimes->{{3.728672667169649*^9, 3.728672727830968*^9}},
 CellID->624875189,ExpressionUUID->"2e52efe7-11c6-4a61-a1e9-e3d4a0c7f884"],

Cell[CellGroupData[{

Cell["\<\
We begin by remembering that each basis has its own parallel derivative\
\>", "MathCaption",
 CellID->2098012318,ExpressionUUID->"ce5ba03b-5b2f-4862-b94c-877bd9c79e3d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"PDfred", "[", 
     RowBox[{"-", "a"}], "]"}], "[", 
    RowBox[{"v", "[", "b", "]"}], "]"}], ",", 
   RowBox[{
    RowBox[{"PDblues", "[", 
     RowBox[{"-", "a"}], "]"}], "[", 
    RowBox[{"T", "[", 
     RowBox[{
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}]}], "]"}], "]"}]}], "}"}]], "Input",
 CellChangeTimes->{{3.728748552130513*^9, 3.7287485614316807`*^9}, 
   3.7287493718167143`*^9},
 CellLabel->"In[284]:=",
 CellID->1759028533,ExpressionUUID->"3a064422-9148-449e-9be6-89df4a5ab782"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox[
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[1., 0., 0.]], "a"], GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"b"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDfred[-$CellContext`a][
     $CellContext`v[$CellContext`b]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox[
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[0., 0., 1.]], "a"], GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {"b", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDblues[-$CellContext`a][
     $CellContext`T[-$CellContext`b, -$CellContext`c]],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.7286725016879377`*^9, 3.728674917585697*^9, 
  3.728748562149437*^9, 3.728749684593812*^9, 3.728749964254449*^9, 
  3.7287512100119553`*^9, 3.728751655481162*^9, 3.7288133004529943`*^9, 
  3.728827420265903*^9, 3.728836010796384*^9, 3.728836345343937*^9, 
  3.728843262938987*^9, 3.728845663600649*^9, 3.733367527135302*^9, 
  3.792273540359583*^9},
 CellLabel->"Out[284]=",
 CellID->643272850,ExpressionUUID->"93c3e456-f8bf-4ec2-8949-c8baeff63ee5"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
defined so that it vanishes on every vector field of that basis (this is the \
origin of the notion of \"parallelism\"):\
\>", "MathCaption",
 CellChangeTimes->{{3.7286725080976057`*^9, 3.7286725387037*^9}},
 CellID->1098520583,ExpressionUUID->"fdef649d-074a-4c51-ad0f-7cd8f41a5a14"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PDfred", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"Basis", "[", 
   RowBox[{
    RowBox[{"-", "c"}], ",", 
    RowBox[{"{", 
     RowBox[{"b", ",", "fred"}], "}"}]}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.728748564736871*^9, 3.7287485680544767`*^9}},
 CellLabel->"In[285]:=",
 CellID->918525222,ExpressionUUID->"5cc9e161-76d5-4b61-9513-888fa51ea295"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728672540273348*^9, 3.728674917752277*^9, 
  3.728748568366909*^9, 3.7287496847426167`*^9, 3.728749964368163*^9, 
  3.728751210149623*^9, 3.728751655601686*^9, 3.7288133006207123`*^9, 
  3.728827420430339*^9, 3.728836010931756*^9, 3.7288363455429707`*^9, 
  3.7288432642501802`*^9, 3.728845663753106*^9, 3.733367527285079*^9, 
  3.792273541890389*^9},
 CellLabel->"Out[285]=",
 CellID->2098680252,ExpressionUUID->"5d551068-876d-4e8e-830a-94fcdfcc0eae"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["This derivative is always flat, with vanishing Riemann :", "MathCaption",
 CellChangeTimes->{{3.728672551934729*^9, 3.728672556662323*^9}},
 CellID->119380206,ExpressionUUID->"7ab848b3-00c2-4524-85bf-00b7037475b7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"RiemannPDfred", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}], ",", 
   RowBox[{"-", "d"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.7287485709590178`*^9, 3.728748572399085*^9}},
 CellLabel->"In[286]:=",
 CellID->673395944,ExpressionUUID->"add5d5c5-6bef-48d4-b90c-b4c8d6d0f3bb"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728672558351481*^9, 3.728674917888337*^9, 
  3.728748572843502*^9, 3.728749684863408*^9, 3.728749964488912*^9, 
  3.7287512103003263`*^9, 3.728751655734556*^9, 3.728813300815753*^9, 
  3.7288274205814543`*^9, 3.728836011063385*^9, 3.72883634571371*^9, 
  3.728843265368532*^9, 3.7288456639008703`*^9, 3.733367527420041*^9, 
  3.792273542937006*^9},
 CellLabel->"Out[286]=",
 CellID->394869652,ExpressionUUID->"55cad4f8-66a9-4863-b987-bd72ed215b64"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["But it has torsion, unless the basis is coordinated:", "MathCaption",
 CellChangeTimes->{{3.7286725666535873`*^9, 3.728672567541924*^9}},
 CellID->39383645,ExpressionUUID->"263d1081-aecd-4950-aa5f-93539cbf0f42"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TorsionPDfred", "[", 
  RowBox[{"a", ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728672586158119*^9, 3.7286725867163353`*^9}, {
  3.728748575176313*^9, 3.728748576399001*^9}},
 CellLabel->"In[287]:=",
 CellID->495958239,ExpressionUUID->"230d09b7-c761-454f-9243-7effafd41658"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox["T",
       FontColor->RGBColor[1., 0., 0.]], 
      StyleBox[GridBox[{
         {"a", " ", " "},
         {" ", "b", "c"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`TorsionPDfred[$CellContext`a, -$CellContext`b, -$CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{{3.728672569484911*^9, 3.728672587218607*^9}, 
   3.728674918038289*^9, 3.7287485768201113`*^9, 3.728749684991351*^9, 
   3.7287499646049833`*^9, 3.728751210459476*^9, 3.7287516558686247`*^9, 
   3.728813300981427*^9, 3.728827420733109*^9, 3.728836011203533*^9, 
   3.728836345925848*^9, 3.728843266501292*^9, 3.728845664053832*^9, 
   3.733367527559805*^9, 3.792273543929577*^9},
 CellLabel->"Out[287]=",
 CellID->810790651,ExpressionUUID->"22333f5d-76a5-4d5d-b367-fe0782a39e89"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
This torsion tensor is equivalent to the Lie brackets of the basis fields:\
\>", "MathCaption",
 CellChangeTimes->{{3.72867259743808*^9, 3.728672613196266*^9}},
 CellID->181064170,ExpressionUUID->"94cbf179-96e8-4b59-abef-6b2f5b088226"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Bracket", "[", 
   RowBox[{
    RowBox[{"Basis", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "b"}], ",", 
        RowBox[{"-", "fred"}]}], "}"}], ",", "b"}], "]"}], ",", 
    RowBox[{"Basis", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "c"}], ",", 
        RowBox[{"-", "fred"}]}], "}"}], ",", "b"}], "]"}]}], "]"}], "[", "a", 
  "]"}]], "Input",
 CellChangeTimes->{{3.728672576813734*^9, 3.7286725787327003`*^9}, {
  3.7286726224125967`*^9, 3.728672626379681*^9}, {3.728748578624596*^9, 
  3.7287485808308573`*^9}},
 CellLabel->"In[288]:=",
 CellID->1652359408,ExpressionUUID->"bd534649-f103-42c2-9546-84860f0b898b"],

Cell[BoxData[
 RowBox[{"-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       StyleBox["T",
        FontColor->RGBColor[1., 0., 0.]], 
       StyleBox[GridBox[{
          {"a", " ", " "},
          {" ", 
           StyleBox["b",
            FontColor->RGBColor[1., 0., 0.]], 
           StyleBox["c",
            FontColor->RGBColor[1., 0., 0.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`TorsionPDfred[$CellContext`a, {-$CellContext`b, \
-$CellContext`fred}, {-$CellContext`c, -$CellContext`fred}],
   Editable->False]}]], "Output",
 CellChangeTimes->{
  3.7286725791576557`*^9, {3.7286726168110113`*^9, 3.7286726269792*^9}, 
   3.728674918186446*^9, 3.728748581171721*^9, 3.728749685262805*^9, 
   3.7287499647688026`*^9, 3.728751210760124*^9, 3.728751655997093*^9, 
   3.7288133011180553`*^9, 3.728827420867291*^9, 3.7288360113506527`*^9, 
   3.7288363461161118`*^9, 3.728843267669807*^9, 3.728845664239827*^9, 
   3.733367527702132*^9, 3.7922735452977324`*^9},
 CellLabel->"Out[288]=",
 CellID->2003355641,ExpressionUUID->"f15c97fb-e47a-4646-93b3-b7a2ca881b48"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "In ",
 StyleBox["xTensor` ", "MR"],
 "we can define the Christoffel tensor connecting any two covariant \
derivatives (",
 StyleBox["cf. ",
  FontSlant->"Italic"],
 "Section 6 of xTensorDoc.nb). In particular, this works with the parallel \
derivatives of our bases:"
}], "MathCaption",
 CellChangeTimes->{{3.728672644506626*^9, 3.728672659130252*^9}},
 CellID->1601747045,ExpressionUUID->"b823a468-74ae-499b-b2ce-d8811f2d93da"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Christoffel", "[", 
   RowBox[{"PDblues", ",", "PDfred"}], "]"}], "[", 
  RowBox[{"a", ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728672765647002*^9, 3.7286727743890753`*^9}, {
   3.72874858320008*^9, 3.728748589157393*^9}, 3.728749373193589*^9},
 CellLabel->"In[289]:=",
 CellID->1855070170,ExpressionUUID->"1b02c939-8220-40fd-8277-1be5bb3b3fe2"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"\[CapitalGamma]", "[", 
       RowBox[{
        StyleBox["\[ScriptCapitalD]",
         FontColor->RGBColor[0., 0., 1.]], ",", 
        StyleBox["\[ScriptCapitalD]",
         FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
      StyleBox[GridBox[{
         {"a", " ", " "},
         {" ", "b", "c"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`ChristoffelPDbluesPDfred[$CellContext`a, -$CellContext`b, \
-$CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{{3.728672745346299*^9, 3.7286727748944187`*^9}, 
   3.728674918370273*^9, 3.728748589572462*^9, 3.72874968560781*^9, 
   3.728749964918391*^9, 3.7287512109418573`*^9, 3.728751656144601*^9, 
   3.728813301287201*^9, 3.728827421032216*^9, 3.728836011505423*^9, 
   3.728836346298684*^9, 3.72884326911975*^9, 3.728845664382987*^9, 
   3.733367527894252*^9, 3.7922735465930433`*^9},
 CellLabel->"Out[289]=",
 CellID->589546690,ExpressionUUID->"39f806ad-0b7e-4b61-befa-21273b143f0e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellChangeTimes->{{3.7286727547927322`*^9, 3.7286727606844187`*^9}},
 CellLabel->"In[290]:=",
 CellID->381890607,ExpressionUUID->"47e62dc5-946c-470b-8a9d-2b81ad0d5d11"],

Cell["ChristoffelPDbluesPDfred[a, -b, -c]", "Output",
 CellChangeTimes->{{3.728672761366311*^9, 3.72867277576792*^9}, 
   3.7286749185845823`*^9, 3.728748592405418*^9, 3.7287496857441463`*^9, 
   3.728749965036146*^9, 3.7287512110913563`*^9, 3.728751656350666*^9, 
   3.728813301452121*^9, 3.728827421165004*^9, 3.72883601164767*^9, 
   3.728836346480488*^9, 3.7288432698524942`*^9, 3.728845664616726*^9, 
   3.733367528152362*^9, 3.79227354715368*^9},
 CellLabel->"Out[290]//InputForm=",
 CellID->766197533,ExpressionUUID->"87225785-11fa-42bf-9274-4420a3e7df2d"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
As usual, Christoffel is antisymmetric in its derivative arguments, and \
automatically reverse them for canonicalization:\
\>", "MathCaption",
 CellChangeTimes->{{3.728672807828321*^9, 3.728672835505986*^9}},
 CellID->1447927465,ExpressionUUID->"0f0b984e-cf9a-4287-9f11-ee05fd1856b0"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Christoffel", "[", 
   RowBox[{"PDfred", ",", "PDblues"}], "]"}], "[", 
  RowBox[{"a", ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.7286727965971527`*^9, 3.728672805402841*^9}, {
   3.728748595525897*^9, 3.7287486032054663`*^9}, 3.7287493789355927`*^9},
 CellLabel->"In[291]:=",
 CellID->1934179853,ExpressionUUID->"94f0faea-ca78-4f0f-bf70-7cdbb9cdadcf"],

Cell[BoxData[
 RowBox[{"-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"\[CapitalGamma]", "[", 
        RowBox[{
         StyleBox["\[ScriptCapitalD]",
          FontColor->RGBColor[0., 0., 1.]], ",", 
         StyleBox["\[ScriptCapitalD]",
          FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
       StyleBox[GridBox[{
          {"a", " ", " "},
          {" ", "b", "c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`ChristoffelPDbluesPDfred[$CellContext`a, -$CellContext`b, \
-$CellContext`c],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728672836689334*^9, 3.72867491873571*^9, 
  3.728748603757416*^9, 3.728749686056175*^9, 3.728749965172864*^9, 
  3.728751211273828*^9, 3.728751656515503*^9, 3.72881330159032*^9, 
  3.7288274213173122`*^9, 3.728836011804617*^9, 3.728836346675767*^9, 
  3.72884327170292*^9, 3.7288456648445873`*^9, 3.733367528388183*^9, 
  3.792273548601664*^9},
 CellLabel->"Out[291]=",
 CellID->703900149,ExpressionUUID->"0a9bbfcc-42f5-475c-9bdb-52f4fb1500dd"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellChangeTimes->{{3.728672843811977*^9, 3.728672845769153*^9}},
 CellLabel->"In[292]:=",
 CellID->157390978,ExpressionUUID->"9c102dd0-45e6-4d19-9bd6-6a4255bdf6aa"],

Cell["-ChristoffelPDbluesPDfred[a, -b, -c]", "Output",
 CellChangeTimes->{3.728672846179039*^9, 3.728674918885602*^9, 
  3.728748605409337*^9, 3.728749686460569*^9, 3.728749965334557*^9, 
  3.728751211440981*^9, 3.728751656648131*^9, 3.728813301735385*^9, 
  3.728827421448019*^9, 3.7288360119493027`*^9, 3.7288363468589687`*^9, 
  3.7288432724533873`*^9, 3.728845665017321*^9, 3.733367528503848*^9, 
  3.792273549166574*^9},
 CellLabel->"Out[292]//InputForm=",
 CellID->1710995331,ExpressionUUID->"a5fa7ce9-bad0-41c4-be1f-04b08bdf91a0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Notice how the Christoffel connecting any covariant derivative to the \
fiducial PD is defined with the derivative, in our case by ",
 StyleBox["DefBasis", "MR"],
 "."
}], "MathCaption",
 CellChangeTimes->{{3.7286728560904016`*^9, 3.728672858129053*^9}},
 CellID->1067825639,ExpressionUUID->"9402fcd7-8e53-49d0-9b7e-296a5cc65a55"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Christoffel", "[", 
   RowBox[{"PDfred", ",", "PD"}], "]"}], "[", 
  RowBox[{"a", ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.7287486078296223`*^9, 3.7287486092532*^9}},
 CellLabel->"In[293]:=",
 CellID->424676694,ExpressionUUID->"c21c3b2e-fd13-464f-aafd-723aed0d1cfe"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"\[CapitalGamma]", "[", 
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[1., 0., 0.]], "]"}], 
      StyleBox[GridBox[{
         {"a", " ", " "},
         {" ", "b", "c"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`ChristoffelPDfred[$CellContext`a, -$CellContext`b, \
-$CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{3.728672859978459*^9, 3.728674919037634*^9, 
  3.728748609602903*^9, 3.728749686840098*^9, 3.728749965471828*^9, 
  3.728751211593129*^9, 3.728751656785248*^9, 3.728813301932622*^9, 
  3.728827421582296*^9, 3.728836012082677*^9, 3.728836347030765*^9, 
  3.728843273772214*^9, 3.728845665168254*^9, 3.733367528693448*^9, 
  3.79227355049583*^9},
 CellLabel->"Out[293]=",
 CellID->1613886747,ExpressionUUID->"181d16c0-8a0b-4608-8d05-aafcf3286a4e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Any derivative of any ",
 StyleBox["Basis", "MR"],
 " object can be translated into a component of Christoffel tensors relating \
that derivative to the PDs of the bases involved. This is one of the \
fundamental building blocks of xCoba`:"
}], "MathCaption",
 CellChangeTimes->{{3.7286728722156982`*^9, 3.728672900911186*^9}, 
   3.728748619532588*^9},
 CellID->10119455,ExpressionUUID->"19b003db-2a0b-421c-9aa2-4feae25b1e29"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"Basis", "[", 
   RowBox[{
    RowBox[{"-", "b"}], ",", 
    RowBox[{"{", 
     RowBox[{"1", ",", "fred"}], "}"}]}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.728748611735093*^9, 3.728748612197146*^9}},
 CellLabel->"In[294]:=",
 CellID->426472072,ExpressionUUID->"43768d32-c248-4828-ad5f-3132364c879b"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"\[CapitalGamma]", "[", 
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[1., 0., 0.]], "]"}], 
      StyleBox[GridBox[{
         {
          StyleBox["1",
           FontColor->RGBColor[1., 0., 0.]], " ", " "},
         {" ", "a", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`ChristoffelPDfred[{
   1, $CellContext`fred}, -$CellContext`a, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.728672906933257*^9, 3.728674919169766*^9, 
  3.728748614403857*^9, 3.72874968709655*^9, 3.728749965601438*^9, 
  3.7287512117930107`*^9, 3.728751656930643*^9, 3.728813302135994*^9, 
  3.7288274217588577`*^9, 3.72883601221537*^9, 3.728836347243225*^9, 
  3.7288432752201347`*^9, 3.7288456653669243`*^9, 3.7333675289598713`*^9, 
  3.792273551928212*^9},
 CellLabel->"Out[294]=",
 CellID->661105440,ExpressionUUID->"31ead0a5-1954-4ac9-9b78-bdd62ae8f740"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"Basis", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "fred"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{"c", ",", "blues"}], "}"}]}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.728748621510294*^9, 3.7287486239064093`*^9}, 
   3.728749382113204*^9},
 CellLabel->"In[295]:=",
 CellID->119643057,ExpressionUUID->"92e47895-2c66-4af9-bebc-03607ccf188e"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"\[CapitalGamma]", "[", 
       RowBox[{
        StyleBox["\[ScriptCapitalD]",
         FontColor->RGBColor[0., 0., 1.]], ",", 
        StyleBox["\[ScriptCapitalD]",
         FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
      StyleBox[GridBox[{
         {
          StyleBox["c",
           FontColor->RGBColor[0., 0., 1.]], " ", " "},
         {" ", "a", 
          StyleBox["b",
           FontColor->RGBColor[1., 0., 0.]]}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`ChristoffelPDbluesPDfred[{$CellContext`c, $CellContext`blues}, \
-$CellContext`a, {-$CellContext`b, -$CellContext`fred}],
  Editable->False]], "Output",
 CellChangeTimes->{3.728672908914378*^9, 3.72867491933494*^9, 
  3.728748624358423*^9, 3.728749687239887*^9, 3.728749965753055*^9, 
  3.728751212343521*^9, 3.728751657081427*^9, 3.728813302303309*^9, 
  3.728827421980006*^9, 3.7288360123518143`*^9, 3.728836347415296*^9, 
  3.7288432759870977`*^9, 3.7288456655159388`*^9, 3.7333675291731243`*^9, 
  3.79227355244676*^9},
 CellLabel->"Out[295]=",
 CellID->1947674454,ExpressionUUID->"f7745f14-70be-4fa9-bb56-b723d583e0df"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["With another parallel derivative:", "MathCaption",
 CellChangeTimes->{{3.728672987628174*^9, 3.728672988362382*^9}},
 CellID->1882859465,ExpressionUUID->"57c1cfa8-53cb-43e2-8292-9c1f7f321955"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PDfred", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"Basis", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"1", ",", 
      RowBox[{"-", "fred"}]}], "}"}], ",", "b"}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.728748627348443*^9, 3.7287486302283792`*^9}},
 CellLabel->"In[296]:=",
 CellID->2047489710,ExpressionUUID->"be81e70c-e5dc-4837-95c1-5b26f43d10fd"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728672990576849*^9, 3.728674919486823*^9, 
  3.7287486307935963`*^9, 3.7287496873958178`*^9, 3.728749966133214*^9, 
  3.728751212541235*^9, 3.728751657214693*^9, 3.728813302438035*^9, 
  3.728827422198983*^9, 3.728836012502481*^9, 3.728836347597928*^9, 
  3.7288432772292137`*^9, 3.728845665667232*^9, 3.733367529337884*^9, 
  3.792273553703189*^9},
 CellLabel->"Out[296]=",
 CellID->734409410,ExpressionUUID->"44753240-9e8a-4c91-8aee-420af15d78fd"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PDfred", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"Basis", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"1", ",", 
      RowBox[{"-", "blues"}]}], "}"}], ",", "b"}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.728748632964198*^9, 3.72874863749999*^9}, 
   3.728749383210792*^9},
 CellLabel->"In[297]:=",
 CellID->1501720041,ExpressionUUID->"e309059a-55a6-4953-9ba8-a284c870ded0"],

Cell[BoxData[
 RowBox[{"-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"\[CapitalGamma]", "[", 
        RowBox[{
         StyleBox["\[ScriptCapitalD]",
          FontColor->RGBColor[0., 0., 1.]], ",", 
         StyleBox["\[ScriptCapitalD]",
          FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
       StyleBox[GridBox[{
          {"b", " ", " "},
          {" ", "a", 
           StyleBox["1",
            FontColor->RGBColor[0., 0., 1.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`ChristoffelPDbluesPDfred[$CellContext`b, -$CellContext`a, {
    1, -$CellContext`blues}],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728672991790614*^9, 3.728674919621299*^9, 
  3.728748637958571*^9, 3.7287496875260344`*^9, 3.7287499664919786`*^9, 
  3.7287512127053137`*^9, 3.7287516573521976`*^9, 3.728813302586151*^9, 
  3.728827422352476*^9, 3.728836012642828*^9, 3.728836347781872*^9, 
  3.728843278120603*^9, 3.7288456658158903`*^9, 3.733367529561252*^9, 
  3.792273554246997*^9},
 CellLabel->"Out[297]=",
 CellID->1063436316,ExpressionUUID->"47993797-d38e-42a2-80f5-b391cc78e27d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PDblues", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"Basis", "[", 
   RowBox[{
    RowBox[{"-", "b"}], ",", 
    RowBox[{"{", 
     RowBox[{"1", ",", "fred"}], "}"}]}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.728672930276903*^9, 3.7286729418277283`*^9}, {
   3.728748640165506*^9, 3.728748644683774*^9}, 3.7287493840104227`*^9},
 CellLabel->"In[298]:=",
 CellID->331410073,ExpressionUUID->"aaf95caf-d338-4cb0-b9ee-244e582e83e4"],

Cell[BoxData[
 RowBox[{"-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"\[CapitalGamma]", "[", 
        RowBox[{
         StyleBox["\[ScriptCapitalD]",
          FontColor->RGBColor[0., 0., 1.]], ",", 
         StyleBox["\[ScriptCapitalD]",
          FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
       StyleBox[GridBox[{
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], " ", " "},
          {" ", "a", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`ChristoffelPDbluesPDfred[{
    1, $CellContext`fred}, -$CellContext`a, -$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{{3.728672906933257*^9, 3.728672942774508*^9}, 
   3.7286729986249437`*^9, 3.728674919784843*^9, 3.728748645128807*^9, 
   3.728749687651897*^9, 3.7287499668011847`*^9, 3.728751212851164*^9, 
   3.728751657514962*^9, 3.728813302769446*^9, 3.728827422499247*^9, 
   3.728836012782136*^9, 3.728836347966353*^9, 3.728843278936513*^9, 
   3.728845665967411*^9, 3.733367529753378*^9, 3.792273555336109*^9},
 CellLabel->"Out[298]=",
 CellID->513451764,ExpressionUUID->"e8e27757-f9ad-4c5f-b72e-262a66baf916"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PDblues", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"Basis", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "fred"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{"c", ",", "blues"}], "}"}]}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.7286729455313387`*^9, 3.728672947628838*^9}, {
  3.7287486472833567`*^9, 3.728748653835457*^9}, {3.728749384801496*^9, 
  3.728749385676532*^9}},
 CellLabel->"In[299]:=",
 CellID->2084745468,ExpressionUUID->"82109728-ed67-43e0-acab-4dfb1eeb1763"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"\[CapitalGamma]", "[", 
       RowBox[{
        StyleBox["\[ScriptCapitalD]",
         FontColor->RGBColor[0., 0., 1.]], ",", 
        StyleBox["\[ScriptCapitalD]",
         FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
      StyleBox[GridBox[{
         {
          StyleBox["c",
           FontColor->RGBColor[0., 0., 1.]], " ", " "},
         {" ", "a", 
          StyleBox["b",
           FontColor->RGBColor[1., 0., 0.]]}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`ChristoffelPDbluesPDfred[{$CellContext`c, $CellContext`blues}, \
-$CellContext`a, {-$CellContext`b, -$CellContext`fred}],
  Editable->False]], "Output",
 CellChangeTimes->{3.728672908914378*^9, 3.728672948134693*^9, 
  3.728672999396946*^9, 3.728674919935863*^9, 3.728748654375793*^9, 
  3.7287496878567944`*^9, 3.7287499672524242`*^9, 3.728751213000889*^9, 
  3.728751657836185*^9, 3.728813302936141*^9, 3.7288274226568527`*^9, 
  3.728836012916641*^9, 3.72883634814904*^9, 3.728843279853693*^9, 
  3.728845666115531*^9, 3.733367529887031*^9, 3.7922735559279747`*^9},
 CellLabel->"Out[299]=",
 CellID->1826832301,ExpressionUUID->"ab3cb86f-c295-4133-b79a-5270fefce4dd"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["2.1.2. Parallel derivatives of complex bases", "Subsubsection",
 CellChangeTimes->{{3.728663829532053*^9, 3.728663849803274*^9}, {
  3.728670543491893*^9, 3.728670543859935*^9}, {3.72867306644833*^9, 
  3.728673079878126*^9}, {3.728673109894463*^9, 3.728673122644684*^9}, {
  3.728673490986658*^9, 3.7286734932669477`*^9}},
 CellID->1300805695,ExpressionUUID->"2eb07e5a-875d-4e3b-b26f-439bd095966f"],

Cell[CellGroupData[{

Cell["\<\
Complex bases have FRiemann and AChristoffel tensors, as explained in \
xTensorDoc, section 6.8:\
\>", "MathCaption",
 CellChangeTimes->{{3.7286730065775747`*^9, 3.728673007897552*^9}},
 CellID->1295802205,ExpressionUUID->"02b878e0-a54e-47e8-a3d2-bc1fb60150a7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "AChristoffelPDfcomp"}]], "Input",
 CellChangeTimes->{3.72874866367563*^9},
 CellLabel->"In[300]:=",
 CellID->915995208,ExpressionUUID->"28927013-dfbb-49f7-b3a4-87a2b787d4d8"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   FrameBox[
    DynamicModuleBox[{System`InformationDump`open$$ = True, 
     System`InformationDump`mouseOver$$ = False}, 
     PaneSelectorBox[{True->
      TagBox[GridBox[{
         {
          ItemBox[
           PaneBox[
            StyleBox["\<\" Symbol\"\>", "InformationTitleText",
             StripOnInput->False,
             BaseStyle -> None],
            FrameMargins->{{4, 0}, {-1, 1}}],
           BaseStyle->"InformationTitleBackground",
           StripOnInput->False], 
          ItemBox["\<\"\"\>",
           BaseStyle->"InformationTitleBackground",
           StripOnInput->False]},
         {
          ItemBox[
           PaneBox[
            
            StyleBox["\<\"Global`AChristoffelPDfcomp\"\>", 
             "InformationUsageText",
             StripOnInput->False,
             LineSpacing->{1.5, 1.5, 3.}],
            FrameMargins->{{10, 10}, {8, 10}}],
           BaseStyle->"InformationUsageSubtitleBackground",
           StripOnInput->False], 
          ItemBox["\<\"\"\>",
           BaseStyle->"InformationUsageSubtitleBackground",
           StripOnInput->False]},
         {
          PaneBox[GridBox[{
             {
              
              DynamicModuleBox[{System`InformationDump`open$$ = {
               False, False, False, False, False, False, False, False, False, 
                False, False, False}}, 
               StyleBox[GridBox[{
                  {
                   TagBox[
                    TooltipBox[
                    
                    StyleBox["\<\" UpValue Definitions\"\>", 
                    "InformationRowLabel",
                    StripOnInput->False],
                    "\"UpValues\"",
                    TooltipStyle->"TextStyling"],
                    Annotation[#, "UpValues", "Tooltip"]& ], GridBox[{
                    {
                    RowBox[{
                    RowBox[{
                    "SymmetryGroupOfTensor", "[", "AChristoffelPDfcomp", 
                    "]"}], "^=", 
                    RowBox[{"StrongGenSet", "[", 
                    RowBox[{
                    RowBox[{"{", "}"}], ",", 
                    RowBox[{"GenSet", "[", "]"}]}], "]"}]}]},
                    {
                    RowBox[{
                    RowBox[{"xTensorQ", "[", "AChristoffelPDfcomp", "]"}], "^=",
                     "True"}]},
                    {
                    RowBox[{
                    RowBox[{
                    "SlotsOfTensor", "[", "AChristoffelPDfcomp", "]"}], "^=", 
                    
                    RowBox[{"{", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["InnerC",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`InnerC,
                    Editable->False], ",", 
                    RowBox[{"-", 
                    InterpretationBox[
                    StyleBox["\[DoubleStruckCapitalT]M3",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`TangentM3,
                    Editable->False]}], ",", 
                    RowBox[{"-", 
                    InterpretationBox[
                    StyleBox["InnerC",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`InnerC,
                    Editable->False]}]}], "}"}]}]},
                    {
                    RowBox[{
                    RowBox[{
                    "DependenciesOfTensor", "[", "AChristoffelPDfcomp", "]"}],
                     "^=", 
                    RowBox[{"{", "M3", "}"}]}]},
                    {
                    RowBox[{
                    RowBox[{"MasterOf", "[", "AChristoffelPDfcomp", "]"}], "^=",
                     "PDfcomp"}]},
                    {
                    RowBox[{
                    RowBox[{"HostsOf", "[", "AChristoffelPDfcomp", "]"}], "^=", 
                    RowBox[{"{", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["InnerC",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`InnerC,
                    Editable->False], ",", "M3", ",", 
                    InterpretationBox[
                    StyleBox["\[DoubleStruckCapitalT]M3",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`TangentM3,
                    Editable->False]}], "}"}]}]},
                    {
                    RowBox[{
                    RowBox[{"DefInfo", "[", "AChristoffelPDfcomp", "]"}], "^=", 
                    RowBox[{"{", 
                    
                    RowBox[{"\<\"nonsymmetric AChristoffel tensor\"\>", 
                    ",", "\<\"\"\>"}], "}"}]}]},
                    {
                    RowBox[{
                    RowBox[{"TensorID", "[", "AChristoffelPDfcomp", "]"}], "^=", 
                    RowBox[{"{", 
                    RowBox[{"AChristoffel", ",", "PDfcomp", ",", "PD"}], 
                    "}"}]}]},
                    {
                    RowBox[{
                    RowBox[{"PrintAs", "[", "AChristoffelPDfcomp", "]"}], 
                    "^=", "\<\"A[\\!\\(\\*StyleBox[\\\"\[ScriptCapitalD]\\\",\
FontColor->RGBColor[0., 1., 0.]]\\)]\"\>"}]},
                    {
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Dagger", "[", "AChristoffelPDfcomp", "]"}],
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    xAct`xTensor`Dagger[$CellContext`AChristoffelPDfcomp],
                    Editable->False], "^=", 
                    "AChristoffelPDfcomp\[Dagger]"}]},
                    {
                    RowBox[{
                    RowBox[{"ServantsOf", "[", "AChristoffelPDfcomp", "]"}], "^=", 
                    RowBox[{"{", "AChristoffelPDfcomp\[Dagger]", "}"}]}]}
                    },
                    DefaultBaseStyle->"Column",
                    GridBoxAlignment->{"Columns" -> {{Left}}},
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}]},
                  {
                   TagBox[
                    TooltipBox[
                    StyleBox["\<\" Full Name\"\>", "InformationRowLabel",
                    StripOnInput->False],
                    "\"FullName\"",
                    TooltipStyle->"TextStyling"],
                    
                    Annotation[#, "FullName", 
                    "Tooltip"]& ], "\<\"Global`AChristoffelPDfcomp\"\>"}
                 },
                 AutoDelete->False,
                 GridBoxAlignment->{"Columns" -> {Right, Left}},
                 GridBoxDividers->None,
                 GridBoxItemSize->{"Columns" -> {Automatic, Automatic}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.5599999999999999]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.8]}, 
                    Offset[0.2]}}], "DialogStyle",
                StripOnInput->False],
               DynamicModuleValues:>{}]}
            },
            DefaultBaseStyle->"Column",
            GridBoxAlignment->{"Columns" -> {{Left}}},
            GridBoxDividers->{"Columns" -> {{False}}, "Rows" -> {{False}}},
            
            GridBoxItemSize->{
             "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.5599999999999999]}, 
                Offset[0.27999999999999997`]}, "Rows" -> {
                Offset[0.2], {
                 Offset[3.6]}, 
                Offset[0.2]}}],
           FrameMargins->6], ""},
         {
          ItemBox[
           TagBox[
            ButtonBox[
             PaneSelectorBox[{False->
              
              DynamicBox[FEPrivate`FrontEndResource[
               "FEBitmaps", "UpPointerOpener"],
               ImageSizeCache->{10., {2., 8.}}], True->
              
              DynamicBox[FEPrivate`FrontEndResource[
               "FEBitmaps", "UpPointerOpenerHot"],
               ImageSizeCache->{10., {2., 8.}}]}, Dynamic[
              System`InformationDump`mouseOver$$]],
             Alignment->Left,
             Appearance->{"Default" -> None},
             
             ButtonFunction:>FEPrivate`Set[
              System`InformationDump`open$$, False],
             Evaluator->Automatic,
             FrameMargins->{{9, 0}, {0, 0}},
             ImageMargins->0,
             ImageSize->Full,
             Method->"Preemptive"],
            
            EventHandlerTag[{
             "MouseEntered" :> 
              FEPrivate`Set[System`InformationDump`mouseOver$$, True], 
              "MouseExited" :> 
              FEPrivate`Set[System`InformationDump`mouseOver$$, False], 
              Method -> "Preemptive", PassEventsDown -> Automatic, 
              PassEventsUp -> True}]],
           BaseStyle->"InformationTitleBackground",
           StripOnInput->False], "\[SpanFromLeft]"}
        },
        AutoDelete->False,
        FrameStyle->Directive[
          GrayLevel[0.8], 
          Thickness[Tiny]],
        GridBoxAlignment->{"Columns" -> {Left, Right}, "Rows" -> {{Center}}},
        GridBoxDividers->{
         "Columns" -> {{None}}, "Rows" -> {False, {True}, False}},
        GridBoxItemSize->{
         "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
       "Grid"], False->
      TagBox[GridBox[{
         {
          ItemBox[
           PaneBox[
            StyleBox["\<\" Symbol\"\>", "InformationTitleText",
             StripOnInput->False],
            FrameMargins->{{4, 0}, {-1, 1}}],
           BaseStyle->"InformationTitleBackground",
           StripOnInput->False], 
          ItemBox["\<\"\"\>",
           BaseStyle->"InformationTitleBackground",
           StripOnInput->False]},
         {
          ItemBox[
           PaneBox[
            
            StyleBox["\<\"Global`AChristoffelPDfcomp\"\>", 
             "InformationUsageText",
             StripOnInput->False,
             LineSpacing->{1.5, 1.5, 3.}],
            FrameMargins->{{10, 10}, {8, 10}}],
           BaseStyle->"InformationUsageSubtitleBackground",
           StripOnInput->False], 
          ItemBox["\<\"\"\>",
           BaseStyle->"InformationUsageSubtitleBackground",
           StripOnInput->False]},
         {
          ItemBox[
           TagBox[
            ButtonBox[
             PaneSelectorBox[{False->
              
              DynamicBox[FEPrivate`FrontEndResource[
               "FEBitmaps", "DownPointerOpener"]], True->
              
              DynamicBox[FEPrivate`FrontEndResource[
               "FEBitmaps", "DownPointerOpenerHot"]]}, Dynamic[
              System`InformationDump`mouseOver$$]],
             Alignment->Left,
             Appearance->{"Default" -> None},
             
             ButtonFunction:>FEPrivate`Set[
              System`InformationDump`open$$, True],
             Evaluator->Automatic,
             FrameMargins->{{9, 0}, {0, 0}},
             ImageMargins->0,
             ImageSize->Full,
             Method->"Preemptive"],
            
            EventHandlerTag[{
             "MouseEntered" :> 
              FEPrivate`Set[System`InformationDump`mouseOver$$, True], 
              "MouseExited" :> 
              FEPrivate`Set[System`InformationDump`mouseOver$$, False], 
              Method -> "Preemptive", PassEventsDown -> Automatic, 
              PassEventsUp -> True}]],
           BaseStyle->"InformationTitleBackground",
           StripOnInput->False], "\[SpanFromLeft]"}
        },
        AutoDelete->False,
        FrameStyle->Directive[
          GrayLevel[0.8], 
          Thickness[Tiny]],
        GridBoxAlignment->{"Columns" -> {Left, Right}, "Rows" -> {{Center}}},
        GridBoxDividers->{
         "Columns" -> {{None}}, "Rows" -> {False, {True}, False}},
        GridBoxItemSize->{
         "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
       "Grid"]}, Dynamic[System`InformationDump`open$$],
      BaselinePosition->Baseline,
      FrameMargins->0,
      ImageSize->Automatic],
     DynamicModuleValues:>{}],
    BaseStyle->"InformationGridFrame",
    StripOnInput->False], "InformationGridPlain",
   StripOnInput->False],
  InformationData[
   Association[
   "ObjectType" -> "Symbol", "Usage" -> "Global`AChristoffelPDfcomp", 
    "Documentation" -> None, "OwnValues" -> None, "UpValues" -> 
    Information`InformationValueForm[
     UpValues, $CellContext`AChristoffelPDfcomp, {
      xAct`xTensor`SymmetryGroupOfTensor[$CellContext`AChristoffelPDfcomp] -> 
       xAct`xPerm`StrongGenSet[{}, 
         xAct`xPerm`GenSet[]], 
       xAct`xTensor`xTensorQ[$CellContext`AChristoffelPDfcomp] -> True, 
       xAct`xTensor`SlotsOfTensor[$CellContext`AChristoffelPDfcomp] -> \
{$CellContext`InnerC, -$CellContext`TangentM3, -$CellContext`InnerC}, 
       xAct`xTensor`DependenciesOfTensor[$CellContext`AChristoffelPDfcomp] -> \
{$CellContext`M3}, 
       xAct`xTensor`MasterOf[$CellContext`AChristoffelPDfcomp] -> \
$CellContext`PDfcomp, 
       xAct`xTensor`HostsOf[$CellContext`AChristoffelPDfcomp] -> \
{$CellContext`InnerC, $CellContext`M3, $CellContext`TangentM3}, 
       xAct`xTensor`DefInfo[$CellContext`AChristoffelPDfcomp] -> {
        "nonsymmetric AChristoffel tensor", ""}, 
       xAct`xTensor`TensorID[$CellContext`AChristoffelPDfcomp] -> {
        xAct`xTensor`AChristoffel, $CellContext`PDfcomp, xAct`xTensor`PD}, 
       xAct`xTensor`PrintAs[$CellContext`AChristoffelPDfcomp] -> 
       "A[\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[0., 1., \
0.]]\)]", 
       xAct`xTensor`Dagger[$CellContext`AChristoffelPDfcomp] -> \
$CellContext`AChristoffelPDfcomp\[Dagger], 
       xAct`xTensor`ServantsOf[$CellContext`AChristoffelPDfcomp] -> \
{$CellContext`AChristoffelPDfcomp\[Dagger]}}], "DownValues" -> None, 
    "SubValues" -> None, "DefaultValues" -> None, "NValues" -> None, 
    "FormatValues" -> None, "Options" -> None, "Attributes" -> {}, "FullName" -> 
    "Global`AChristoffelPDfcomp"], False]]], "Output",
 CellChangeTimes->{3.792273561704673*^9},
 CellLabel->"Out[300]=",
 CellID->470484424,ExpressionUUID->"ed10973b-578b-41d0-9210-b2e73506fb02"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The FRiemann is also zero", "MathCaption",
 CellID->579705646,ExpressionUUID->"49a61b71-c68e-47cb-a0d0-a8670ab0e5a7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FRiemannPDfcomp", "[", 
  RowBox[{
   RowBox[{"-", "A"}], ",", 
   RowBox[{"-", "B"}], ",", 
   RowBox[{"-", "C"}], ",", 
   RowBox[{"-", "D"}]}], "]"}]], "Input",
 CellChangeTimes->{3.728748669178358*^9},
 CellLabel->"In[301]:=",
 CellID->2126969530,ExpressionUUID->"b140e9d1-d653-4799-92e6-ae367a8be29b"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.7286730110756617`*^9, 3.728674920321459*^9, 
  3.728748669547724*^9, 3.728749696681161*^9, 3.7287499676536217`*^9, 
  3.728751218057479*^9, 3.7287516581474333`*^9, 3.728813303373885*^9, 
  3.7288274234372168`*^9, 3.728836013277399*^9, 3.728836348632159*^9, 
  3.72884328973629*^9, 3.728845666600918*^9, 3.7333675341221733`*^9, 
  3.7922735617818403`*^9},
 CellLabel->"Out[301]=",
 CellID->396925914,ExpressionUUID->"8bda260c-48ed-4366-a9fe-16b5ce611d6f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Derivatives of complex bases are also replaced by the corresponding \
Christoffels\
\>", "MathCaption",
 CellID->910769021,ExpressionUUID->"3afa6f5b-6a5e-4e06-8963-9eccd57600c6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"Basis", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "A"}], ",", 
      RowBox[{"-", "fcomp"}]}], "}"}], ",", "C"}], "]"}], "]"}]], "Input",
 CellChangeTimes->{3.728748672203967*^9},
 CellLabel->"In[302]:=",
 CellID->31500131,ExpressionUUID->"e9367cb9-4529-4a81-8893-b026ecfe8de2"],

Cell[BoxData[
 RowBox[{"-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"A", "[", 
        StyleBox["\[ScriptCapitalD]",
         FontColor->RGBColor[0., 1., 0.]], "]"}], 
       StyleBox[GridBox[{
          {"C", " ", " "},
          {" ", "a", 
           StyleBox["A",
            FontColor->RGBColor[0., 1., 0.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`AChristoffelPDfcomp[
   C, -$CellContext`a, {-$CellContext`A, -$CellContext`fcomp}],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728673015763026*^9, 3.7286749204711246`*^9, 
  3.728748672926983*^9, 3.72874969682543*^9, 3.728749967945477*^9, 
  3.728751218192811*^9, 3.72875165825238*^9, 3.728813303561289*^9, 
  3.728827423605811*^9, 3.7288360134170837`*^9, 3.728836348828122*^9, 
  3.7288432910220823`*^9, 3.728845666816527*^9, 3.733367534257814*^9, 
  3.792273561831664*^9},
 CellLabel->"Out[302]=",
 CellID->1049629016,ExpressionUUID->"f54b3353-f9de-4a44-aa2c-bb5067d30fa3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"Basis", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "A\[Dagger]"}], ",", 
      RowBox[{"-", "fcomp"}]}], "}"}], ",", "C\[Dagger]"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{3.728748674178396*^9},
 CellLabel->"In[303]:=",
 CellID->1198209940,ExpressionUUID->"5345e42e-6496-44e7-b4aa-77f1521b4896"],

Cell[BoxData[
 RowBox[{"-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{
        RowBox[{"A", "[", 
         StyleBox["\[ScriptCapitalD]",
          FontColor->RGBColor[0., 1., 0.]], "]"}], "\[Dagger]"}], 
       StyleBox[GridBox[{
          {"C\[Dagger]", " ", " "},
          {" ", "a", 
           StyleBox["A\[Dagger]",
            FontColor->RGBColor[0., 1., 0.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`AChristoffelPDfcomp\[Dagger][$CellContext`C\[Dagger], \
-$CellContext`a, {-$CellContext`A\[Dagger], -$CellContext`fcomp}],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728673016448737*^9, 3.728674920620337*^9, 
  3.728748674695506*^9, 3.7287496969735937`*^9, 3.728749968080339*^9, 
  3.728751218334743*^9, 3.728751658364901*^9, 3.728813303743558*^9, 
  3.728827423842516*^9, 3.728836013549449*^9, 3.728836349017345*^9, 
  3.7288432917890472`*^9, 3.7288456669503098`*^9, 3.733367534388493*^9, 
  3.792273561879979*^9},
 CellLabel->"Out[303]=",
 CellID->1684195454,ExpressionUUID->"a75083be-1580-4d8c-8e2b-9022b880888e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"Basis", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "fcomp"}]}], "}"}], ",", "b"}], "]"}], "]"}]], "Input",
 CellChangeTimes->{3.728748676618145*^9},
 CellLabel->"In[304]:=",
 CellID->512724465,ExpressionUUID->"55b92aa5-5d43-4250-b633-90a82bd0106c"],

Cell[BoxData[
 RowBox[{"-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"\[CapitalGamma]", "[", 
        StyleBox["\[ScriptCapitalD]",
         FontColor->RGBColor[0., 1., 0.]], "]"}], 
       StyleBox[GridBox[{
          {"b", " ", " "},
          {" ", "a", 
           StyleBox["c",
            FontColor->RGBColor[0., 1., 0.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`ChristoffelPDfcomp[$CellContext`b, -$CellContext`a, \
{-$CellContext`c, -$CellContext`fcomp}],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728673017848868*^9, 3.728674920753587*^9, 
  3.728748677064981*^9, 3.7287496971200953`*^9, 3.7287499681833773`*^9, 
  3.7287512185140657`*^9, 3.728751658463332*^9, 3.728813303903529*^9, 
  3.728827423999754*^9, 3.728836013713558*^9, 3.7288363492010393`*^9, 
  3.7288432925562363`*^9, 3.728845667071686*^9, 3.7333675345933247`*^9, 
  3.792273561922613*^9},
 CellLabel->"Out[304]=",
 CellID->291044271,ExpressionUUID->"2aac526e-6fda-431c-9476-442f7e01ef0d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PDfred", "[", 
   RowBox[{"-", "b"}], "]"}], "[", 
  RowBox[{"Basis", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "A"}], ",", 
      RowBox[{"-", "fcomp"}]}], "}"}], ",", "C"}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.728748679050004*^9, 3.728748682201932*^9}},
 CellLabel->"In[305]:=",
 CellID->113791411,ExpressionUUID->"edf9c63b-bf73-4ad2-8ebb-2e2b049ecfb3"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"AChristoffelPDfcompPDfred", "[", 
    RowBox[{"A", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "C"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`AChristoffelPDfcompPDfred[$CellContext`A, -$CellContext`b, -
    C], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286730185629396`*^9, 3.7286749209925957`*^9, 
  3.7287486825217953`*^9, 3.7287496972629547`*^9, 3.7287499683185663`*^9, 
  3.728751218877084*^9, 3.7287516585693703`*^9, 3.72881330402142*^9, 
  3.728827424182749*^9, 3.728836013851919*^9, 3.72883634937668*^9, 
  3.728843293339492*^9, 3.728845667199903*^9, 3.733367534811849*^9, 
  3.792273561964923*^9},
 CellLabel->"During evaluation of In[305]:=",
 CellID->1760798478,ExpressionUUID->"bf126c11-bf2c-4775-bd11-5636046bc134"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"AChristoffelPDfcompPDfred\[Dagger]", "[", 
    RowBox[{"A\[Dagger]", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "C\[Dagger]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`AChristoffelPDfcompPDfred\[Dagger][$CellContext`A\[Dagger], \
-$CellContext`b, -$CellContext`C\[Dagger]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286730185629396`*^9, 3.7286749209925957`*^9, 
  3.7287486825217953`*^9, 3.7287496972629547`*^9, 3.7287499683185663`*^9, 
  3.728751218877084*^9, 3.7287516585693703`*^9, 3.72881330402142*^9, 
  3.728827424182749*^9, 3.728836013851919*^9, 3.72883634937668*^9, 
  3.728843293339492*^9, 3.728845667199903*^9, 3.733367534811849*^9, 
  3.792273561976379*^9},
 CellLabel->"During evaluation of In[305]:=",
 CellID->1415364086,ExpressionUUID->"808f03f8-4da9-4de7-88df-db0fc52af4f1"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"A", "[", 
        RowBox[{
         StyleBox["\[ScriptCapitalD]",
          FontColor->RGBColor[0., 1., 0.]], ",", 
         StyleBox["\[ScriptCapitalD]",
          FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
       StyleBox[GridBox[{
          {"C", " ", " "},
          {" ", "b", 
           StyleBox["A",
            FontColor->RGBColor[0., 1., 0.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`AChristoffelPDfcompPDfred[
   C, -$CellContext`b, {-$CellContext`A, -$CellContext`fcomp}],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728673018681288*^9, 3.728674921052535*^9, 
  3.7287486826320457`*^9, 3.7287496973666983`*^9, 3.7287499683542223`*^9, 
  3.7287512191101294`*^9, 3.728751658604012*^9, 3.728813304060976*^9, 
  3.72882742423175*^9, 3.728836013901147*^9, 3.728836349517765*^9, 
  3.728843293459317*^9, 3.728845667249612*^9, 3.7333675349227962`*^9, 
  3.792273561988944*^9},
 CellLabel->"Out[305]=",
 CellID->347835583,ExpressionUUID->"8173b221-428d-423e-9eb0-5f58ed8515cb"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Both basis indices must belong to the same vbundle:", "MathCaption",
 CellID->1432168523,ExpressionUUID->"d1a0bdf6-5132-45f4-bf5d-f6c74f709798"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Catch", "@", 
  RowBox[{
   RowBox[{"PD", "[", 
    RowBox[{"-", "a"}], "]"}], "[", 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "c"}], ",", 
       RowBox[{"-", "fcomp"}]}], "}"}], ",", "B"}], "]"}], "]"}]}]], "Input",
 CellChangeTimes->{3.7287486856739397`*^9},
 CellLabel->"In[306]:=",
 CellID->1459650485,ExpressionUUID->"cb45c973-6612-4d7f-bc9e-37996d2db8b9"],

Cell[BoxData[
 TemplateBox[{
  "Christoffel", "inds13", 
   "\"Indices 1 and 3 of (A)Christoffel must belong to the same vbundle.\"", 
   2, 306, 7, 19668387891272167742, "Local", "xAct`xTensor`Christoffel"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728673021583768*^9, 3.7286749212129383`*^9, 
  3.728748686149846*^9, 3.728749697505445*^9, 3.728749968501244*^9, 
  3.72875121927151*^9, 3.7287516587534027`*^9, 3.728813304187418*^9, 
  3.728827424483233*^9, 3.728836014061686*^9, 3.728836349719756*^9, 
  3.72884329499037*^9, 3.728845667399128*^9, 3.733367535037731*^9, 
  3.792273562035246*^9},
 CellLabel->"During evaluation of In[306]:=",
 CellID->622156347,ExpressionUUID->"1047f1e4-7d75-4d73-8a6d-26cf33ebccfc"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["2.1.3. Ricci rotation coefficients", "Subsubsection",
 CellChangeTimes->{{3.728673137787421*^9, 3.728673140011092*^9}, {
  3.7286734943547688`*^9, 3.728673495003137*^9}},
 CellID->556197953,ExpressionUUID->"807c75e1-46b5-4a17-93cf-4e90f5048288"],

Cell[TextData[{
 "Ricci rotation coefficients are not supported as separate objects, because \
they do not add any new information.",
 StyleBox[" xCoba`", "MR"],
 "  limits itself to converting  ",
 StyleBox["RicciRotation[covd,basis]", "MR"],
 " into ",
 StyleBox["Christoffel[covd, PDofBasis[basis]].", "MR"]
}], "Text",
 CellChangeTimes->{{3.728673148835059*^9, 3.7286731522274446`*^9}},
 CellID->22680680,ExpressionUUID->"559bb3f7-da02-49cb-a073-b44f096c2e4b"],

Cell[CellGroupData[{

Cell["\<\
Ricci rotation coefficients of a covariant derivative in a given (coordinated \
or not) frame:\
\>", "MathCaption",
 CellChangeTimes->{{3.7286731614023743`*^9, 3.7286731971372128`*^9}},
 CellID->1494094561,ExpressionUUID->"6dde03ad-f4cb-456c-8dd3-537e34b2f126"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"RicciRotation", "[", 
   RowBox[{"PDfred", ",", "blues"}], "]"}], "[", 
  RowBox[{"a", ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728748694705296*^9, 3.728748700577098*^9}, 
   3.728749386826086*^9},
 CellLabel->"In[307]:=",
 CellID->1089583241,ExpressionUUID->"65855468-2aef-4fe0-910f-2d4049df56ec"],

Cell[BoxData[
 RowBox[{"-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"\[CapitalGamma]", "[", 
        RowBox[{
         StyleBox["\[ScriptCapitalD]",
          FontColor->RGBColor[0., 0., 1.]], ",", 
         StyleBox["\[ScriptCapitalD]",
          FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
       StyleBox[GridBox[{
          {"a", " ", " "},
          {" ", "b", "c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`ChristoffelPDbluesPDfred[$CellContext`a, -$CellContext`b, \
-$CellContext`c],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728673199114688*^9, 3.7286749213531313`*^9, 
  3.728748701318726*^9, 3.728749704758872*^9, 3.72874996864045*^9, 
  3.728751224777053*^9, 3.7287516588830023`*^9, 3.728813304309362*^9, 
  3.728827424632864*^9, 3.728836014228973*^9, 3.728836349926968*^9, 
  3.7288433017115173`*^9, 3.728845667550579*^9, 3.733367541924734*^9, 
  3.792273567530016*^9},
 CellLabel->"Out[307]=",
 CellID->65755722,ExpressionUUID->"257d3d18-77de-458d-b7b5-732bf3d44b7a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"RicciRotation", "[", "PDblues", "]"}], "[", 
  RowBox[{"a", ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.7287487031466503`*^9, 3.728748706304779*^9}, 
   3.728749388595064*^9},
 CellLabel->"In[308]:=",
 CellID->1114549266,ExpressionUUID->"f52ae658-260a-489e-9e40-d7e302c6845e"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"\[CapitalGamma]", "[", 
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[0., 0., 1.]], "]"}], 
      StyleBox[GridBox[{
         {"a", " ", " "},
         {" ", "b", "c"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`ChristoffelPDblues[$CellContext`a, -$CellContext`b, \
-$CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{3.728673199898958*^9, 3.728674921468522*^9, 
  3.7287487068204737`*^9, 3.728749704894971*^9, 3.7287499687688837`*^9, 
  3.7287512249395647`*^9, 3.728751658997685*^9, 3.728813304454094*^9, 
  3.728827424782168*^9, 3.7288360143674192`*^9, 3.728836350100136*^9, 
  3.7288433018952627`*^9, 3.7288456677508707`*^9, 3.733367542055812*^9, 
  3.792273567585883*^9},
 CellLabel->"Out[308]=",
 CellID->1349825096,ExpressionUUID->"86029925-ea73-4413-9626-355929b7ecec"]
}, Open  ]]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["2.2. Charts", "Subsection",
 CellChangeTimes->{
  3.728670946657796*^9, {3.7286734807787313`*^9, 
   3.728673481371875*^9}},ExpressionUUID->"5d1d6849-a99b-4de9-a3d0-\
d37ac7639baf"],

Cell[CellGroupData[{

Cell["2.2.1. Define a chart", "Subsubsection",
 CellChangeTimes->{{3.728663829532053*^9, 3.728663849803274*^9}, {
  3.728670543491893*^9, 3.728670543859935*^9}, {3.72867306644833*^9, 
  3.728673079878126*^9}, {3.7286734857870407`*^9, 3.728673489676098*^9}, {
  3.728759263849457*^9, 3.72875926969563*^9}},
 CellID->2089523020,ExpressionUUID->"d93f6ed4-496a-4c48-9f24-6fb16b80c08d"],

Cell["\<\
A chart (or coordinate system) in xCoba` is just a particular type of basis: \
a coordinated basis. The main difference is that together with the frame we \
also declare a number of scalar fields, the coordinate scalar fields of the \
chart, whose gradients are aligned with the frame. Any other scalar field \
will be expressible as a scalar function (in the sense of the ScalarFunction \
type of xTensor`) of those coordinate scalar fields.\
\>", "Text",
 CellChangeTimes->{{3.7286743858352537`*^9, 3.728674520740087*^9}, {
  3.728759277191424*^9, 3.728759315998025*^9}},
 CellID->840881529,ExpressionUUID->"0eaf4081-2e47-47bf-ae56-ae4e7f43f715"],

Cell[TextData[{
 StyleBox["DefChart\t\t\t", "DisplayMath"],
 StyleBox["Define a basis", "Text"],
 StyleBox["\n$Bases\t\t\t\t", "DisplayMath"],
 StyleBox["List of currently defined bases\n", "Text"],
 StyleBox["ChartQ\t\t\t\t", "DisplayMath"],
 StyleBox["Check existence of a given basis name", "Text"]
}], "DefinitionBox",ExpressionUUID->"82eb9af8-9886-4026-b35c-e76f693a5e10"],

Cell["Definition of a chart.", "Text",
 CellChangeTimes->{
  3.7286745341784897`*^9},ExpressionUUID->"44455346-3c8a-4578-b8d3-\
77884d5e069c"],

Cell[CellGroupData[{

Cell["\<\
Declare a coordinate system called csys, whose scalar fields as x[], y[] and \
z[]. As usual in xAct, we need the empty pairs of brackets to denote scalar \
fields:\
\>", "MathCaption",
 CellChangeTimes->{{3.7286745552352943`*^9, 3.728674630854826*^9}, {
  3.728748749289596*^9, 3.728748749655054*^9}},
 CellID->1009234187,ExpressionUUID->"6bc23be5-6940-43bb-8453-b172c78be34b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefChart", "[", 
  RowBox[{"csys", ",", "M3", ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", "2", ",", "3"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"x", "[", "]"}], ",", 
     RowBox[{"y", "[", "]"}], ",", 
     RowBox[{"z", "[", "]"}]}], "}"}], ",", 
   RowBox[{"ChartColor", "\[Rule]", "Orange"}]}], "]"}]], "Input",
 CellChangeTimes->{
  3.728674610031507*^9, {3.7286746471793547`*^9, 3.7286746501834793`*^9}, {
   3.7287487534708548`*^9, 3.728748753952252*^9}, {3.728813464315523*^9, 
   3.728813465201313*^9}, 3.7288135057356367`*^9, {3.728813950155806*^9, 
   3.7288139510112267`*^9}},
 CellLabel->
  "In[309]:=",ExpressionUUID->"fe27af13-9dff-48e3-81e1-f935a4a3101f"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefChart", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"chart \"\>", "\[InvisibleSpace]", "csys", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xCoba`DefChart, ": Defining ", "chart ", $CellContext`csys, 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728674615751286*^9, 3.728674653682967*^9, 
  3.7286749215842857`*^9, 3.728748755793079*^9, 3.7287497255857153`*^9, 
  3.7287499689342527`*^9, 3.728751230978325*^9, 3.7287516591137857`*^9, 
  3.728759366462978*^9, 3.7288133045683117`*^9, 3.728813474750271*^9, 
  3.728813951774508*^9, 3.728827424928256*^9, 3.728836014516368*^9, 
  3.728836350287994*^9, 3.728843310859083*^9, 3.728845667895321*^9, 
  3.7333675485386257`*^9, 3.792273574939508*^9},
 CellLabel->"During evaluation of In[309]:=",
 CellID->1224112186,ExpressionUUID->"f64b1fff-b928-42e9-8a18-5600e25f4154"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"x", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "coordinate scalar ", 
   $CellContext`x[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728674615751286*^9, 3.728674653682967*^9, 
  3.7286749215842857`*^9, 3.728748755793079*^9, 3.7287497255857153`*^9, 
  3.7287499689342527`*^9, 3.728751230978325*^9, 3.7287516591137857`*^9, 
  3.728759366462978*^9, 3.7288133045683117`*^9, 3.728813474750271*^9, 
  3.728813951774508*^9, 3.728827424928256*^9, 3.728836014516368*^9, 
  3.728836350287994*^9, 3.728843310859083*^9, 3.728845667895321*^9, 
  3.7333675485386257`*^9, 3.792273574954063*^9},
 CellLabel->"During evaluation of In[309]:=",
 CellID->760171842,ExpressionUUID->"5921a264-7dd4-47c7-bac0-8d78834a6fae"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"y", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "coordinate scalar ", 
   $CellContext`y[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728674615751286*^9, 3.728674653682967*^9, 
  3.7286749215842857`*^9, 3.728748755793079*^9, 3.7287497255857153`*^9, 
  3.7287499689342527`*^9, 3.728751230978325*^9, 3.7287516591137857`*^9, 
  3.728759366462978*^9, 3.7288133045683117`*^9, 3.728813474750271*^9, 
  3.728813951774508*^9, 3.728827424928256*^9, 3.728836014516368*^9, 
  3.728836350287994*^9, 3.728843310859083*^9, 3.728845667895321*^9, 
  3.7333675485386257`*^9, 3.792273574970091*^9},
 CellLabel->"During evaluation of In[309]:=",
 CellID->827932491,ExpressionUUID->"9241dc63-3e65-4665-adcb-5eef1df6b6d9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"z", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "coordinate scalar ", 
   $CellContext`z[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728674615751286*^9, 3.728674653682967*^9, 
  3.7286749215842857`*^9, 3.728748755793079*^9, 3.7287497255857153`*^9, 
  3.7287499689342527`*^9, 3.728751230978325*^9, 3.7287516591137857`*^9, 
  3.728759366462978*^9, 3.7288133045683117`*^9, 3.728813474750271*^9, 
  3.728813951774508*^9, 3.728827424928256*^9, 3.728836014516368*^9, 
  3.728836350287994*^9, 3.728843310859083*^9, 3.728845667895321*^9, 
  3.7333675485386257`*^9, 3.792273574977837*^9},
 CellLabel->"During evaluation of In[309]:=",
 CellID->1514827324,ExpressionUUID->"062018ed-ec1b-44a9-959a-766b646cab39"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefMapping", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"mapping \"\>", "\[InvisibleSpace]", "csys", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefMapping, ": Defining ", 
   "mapping ", $CellContext`csys, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728674615751286*^9, 3.728674653682967*^9, 
  3.7286749215842857`*^9, 3.728748755793079*^9, 3.7287497255857153`*^9, 
  3.7287499689342527`*^9, 3.728751230978325*^9, 3.7287516591137857`*^9, 
  3.728759366462978*^9, 3.7288133045683117`*^9, 3.728813474750271*^9, 
  3.728813951774508*^9, 3.728827424928256*^9, 3.728836014516368*^9, 
  3.728836350287994*^9, 3.728843310859083*^9, 3.728845667895321*^9, 
  3.7333675485386257`*^9, 3.792273574985239*^9},
 CellLabel->"During evaluation of In[309]:=",
 CellID->1284199283,ExpressionUUID->"eed51d51-187e-4d69-a316-4b2f93dc3415"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefMapping", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"inverse mapping \"\>", "\[InvisibleSpace]", 
   "icsys", "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefMapping, ": Defining ", 
   "inverse mapping ", $CellContext`icsys, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728674615751286*^9, 3.728674653682967*^9, 
  3.7286749215842857`*^9, 3.728748755793079*^9, 3.7287497255857153`*^9, 
  3.7287499689342527`*^9, 3.728751230978325*^9, 3.7287516591137857`*^9, 
  3.728759366462978*^9, 3.7288133045683117`*^9, 3.728813474750271*^9, 
  3.728813951774508*^9, 3.728827424928256*^9, 3.728836014516368*^9, 
  3.728836350287994*^9, 3.728843310859083*^9, 3.728845667895321*^9, 
  3.7333675485386257`*^9, 3.792273574992597*^9},
 CellLabel->"During evaluation of In[309]:=",
 CellID->1326697031,ExpressionUUID->"1d115aca-6210-453d-8672-359dad2c9dea"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"mapping differential tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"dicsys", "[", 
    RowBox[{
     RowBox[{"-", "\[FormalA]"}], ",", "icsysa"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "mapping differential tensor ", 
   $CellContext`dicsys[-\[FormalA], $CellContext`icsysa], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728674615751286*^9, 3.728674653682967*^9, 
  3.7286749215842857`*^9, 3.728748755793079*^9, 3.7287497255857153`*^9, 
  3.7287499689342527`*^9, 3.728751230978325*^9, 3.7287516591137857`*^9, 
  3.728759366462978*^9, 3.7288133045683117`*^9, 3.728813474750271*^9, 
  3.728813951774508*^9, 3.728827424928256*^9, 3.728836014516368*^9, 
  3.728836350287994*^9, 3.728843310859083*^9, 3.728845667895321*^9, 
  3.7333675485386257`*^9, 3.7922735749998817`*^9},
 CellLabel->"During evaluation of In[309]:=",
 CellID->474293216,ExpressionUUID->"e7ac1d4d-864f-4c4f-9020-cbe363c68666"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"mapping differential tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"dcsys", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", "csys\[FormalA]"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "mapping differential tensor ", 
   $CellContext`dcsys[-$CellContext`a, $CellContext`csys\[FormalA]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728674615751286*^9, 3.728674653682967*^9, 
  3.7286749215842857`*^9, 3.728748755793079*^9, 3.7287497255857153`*^9, 
  3.7287499689342527`*^9, 3.728751230978325*^9, 3.7287516591137857`*^9, 
  3.728759366462978*^9, 3.7288133045683117`*^9, 3.728813474750271*^9, 
  3.728813951774508*^9, 3.728827424928256*^9, 3.728836014516368*^9, 
  3.728836350287994*^9, 3.728843310859083*^9, 3.728845667895321*^9, 
  3.7333675485386257`*^9, 3.792273575007539*^9},
 CellLabel->"During evaluation of In[309]:=",
 CellID->2134267295,ExpressionUUID->"fec38bb0-48a0-4f57-ba22-c2002229749a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefBasis", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"basis \"\>", "\[InvisibleSpace]", "csys", 
   "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Coordinated basis.\"\>"}],
  SequenceForm[
  "** ", xAct`xCoba`DefBasis, ": Defining ", "basis ", $CellContext`csys, 
   ". ", "Coordinated basis."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728674615751286*^9, 3.728674653682967*^9, 
  3.7286749215842857`*^9, 3.728748755793079*^9, 3.7287497255857153`*^9, 
  3.7287499689342527`*^9, 3.728751230978325*^9, 3.7287516591137857`*^9, 
  3.728759366462978*^9, 3.7288133045683117`*^9, 3.728813474750271*^9, 
  3.728813951774508*^9, 3.728827424928256*^9, 3.728836014516368*^9, 
  3.728836350287994*^9, 3.728843310859083*^9, 3.728845667895321*^9, 
  3.7333675485386257`*^9, 3.7922735750149593`*^9},
 CellLabel->"During evaluation of In[309]:=",
 CellID->1211071224,ExpressionUUID->"fdffe5de-3496-4f0e-b023-0f9571fcfc92"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"parallel derivative \"\>", "\[InvisibleSpace]", 
   RowBox[{"PDcsys", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "parallel derivative ", 
   $CellContext`PDcsys[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728674615751286*^9, 3.728674653682967*^9, 
  3.7286749215842857`*^9, 3.728748755793079*^9, 3.7287497255857153`*^9, 
  3.7287499689342527`*^9, 3.728751230978325*^9, 3.7287516591137857`*^9, 
  3.728759366462978*^9, 3.7288133045683117`*^9, 3.728813474750271*^9, 
  3.728813951774508*^9, 3.728827424928256*^9, 3.728836014516368*^9, 
  3.728836350287994*^9, 3.728843310859083*^9, 3.728845667895321*^9, 
  3.7333675485386257`*^9, 3.792273575022581*^9},
 CellLabel->"During evaluation of In[309]:=",
 CellID->624055502,ExpressionUUID->"c3a1c054-e44c-4c77-bdc1-6e0d27a2c80b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing torsion tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TorsionPDcsys", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing torsion tensor ", 
   $CellContext`TorsionPDcsys[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728674615751286*^9, 3.728674653682967*^9, 
  3.7286749215842857`*^9, 3.728748755793079*^9, 3.7287497255857153`*^9, 
  3.7287499689342527`*^9, 3.728751230978325*^9, 3.7287516591137857`*^9, 
  3.728759366462978*^9, 3.7288133045683117`*^9, 3.728813474750271*^9, 
  3.728813951774508*^9, 3.728827424928256*^9, 3.728836014516368*^9, 
  3.728836350287994*^9, 3.728843310859083*^9, 3.728845667895321*^9, 
  3.7333675485386257`*^9, 3.792273575029962*^9},
 CellLabel->"During evaluation of In[309]:=",
 CellID->190690166,ExpressionUUID->"a9068d7b-ae68-4e3c-9c58-eb0f20075b7b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelPDcsys", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Christoffel tensor ", 
   $CellContext`ChristoffelPDcsys[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728674615751286*^9, 3.728674653682967*^9, 
  3.7286749215842857`*^9, 3.728748755793079*^9, 3.7287497255857153`*^9, 
  3.7287499689342527`*^9, 3.728751230978325*^9, 3.7287516591137857`*^9, 
  3.728759366462978*^9, 3.7288133045683117`*^9, 3.728813474750271*^9, 
  3.728813951774508*^9, 3.728827424928256*^9, 3.728836014516368*^9, 
  3.728836350287994*^9, 3.728843310859083*^9, 3.728845667895321*^9, 
  3.7333675485386257`*^9, 3.792273575037284*^9},
 CellLabel->"During evaluation of In[309]:=",
 CellID->884018638,ExpressionUUID->"02fecf1e-c5c5-4e74-9a22-4925f6d98dec"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Riemann tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RiemannPDcsys", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Riemann tensor ", 
   $CellContext`RiemannPDcsys[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, $CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728674615751286*^9, 3.728674653682967*^9, 
  3.7286749215842857`*^9, 3.728748755793079*^9, 3.7287497255857153`*^9, 
  3.7287499689342527`*^9, 3.728751230978325*^9, 3.7287516591137857`*^9, 
  3.728759366462978*^9, 3.7288133045683117`*^9, 3.728813474750271*^9, 
  3.728813951774508*^9, 3.728827424928256*^9, 3.728836014516368*^9, 
  3.728836350287994*^9, 3.728843310859083*^9, 3.728845667895321*^9, 
  3.7333675485386257`*^9, 3.7922735750450706`*^9},
 CellLabel->"During evaluation of In[309]:=",
 CellID->1409378920,ExpressionUUID->"30416c28-3b39-4e5a-86f8-c77c14a7c760"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciPDcsys", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Ricci tensor ", 
   $CellContext`RicciPDcsys[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728674615751286*^9, 3.728674653682967*^9, 
  3.7286749215842857`*^9, 3.728748755793079*^9, 3.7287497255857153`*^9, 
  3.7287499689342527`*^9, 3.728751230978325*^9, 3.7287516591137857`*^9, 
  3.728759366462978*^9, 3.7288133045683117`*^9, 3.728813474750271*^9, 
  3.728813951774508*^9, 3.728827424928256*^9, 3.728836014516368*^9, 
  3.728836350287994*^9, 3.728843310859083*^9, 3.728845667895321*^9, 
  3.7333675485386257`*^9, 3.792273575052826*^9},
 CellLabel->"During evaluation of In[309]:=",
 CellID->686882030,ExpressionUUID->"005c0e2a-e836-424f-8716-cbd90e18bc18"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaUpcsys", "[", 
    RowBox[{"a", ",", "b", ",", "c"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric +1 density ", 
   $CellContext`etaUpcsys[$CellContext`a, $CellContext`b, $CellContext`c], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728674615751286*^9, 3.728674653682967*^9, 
  3.7286749215842857`*^9, 3.728748755793079*^9, 3.7287497255857153`*^9, 
  3.7287499689342527`*^9, 3.728751230978325*^9, 3.7287516591137857`*^9, 
  3.728759366462978*^9, 3.7288133045683117`*^9, 3.728813474750271*^9, 
  3.728813951774508*^9, 3.728827424928256*^9, 3.728836014516368*^9, 
  3.728836350287994*^9, 3.728843310859083*^9, 3.728845667895321*^9, 
  3.7333675485386257`*^9, 3.79227357506041*^9},
 CellLabel->"During evaluation of In[309]:=",
 CellID->1701675585,ExpressionUUID->"f62814ab-0dda-41ee-a3bf-7c494d966fa2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaDowncsys", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric -1 density ", 
   $CellContext`etaDowncsys[-$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728674615751286*^9, 3.728674653682967*^9, 
  3.7286749215842857`*^9, 3.728748755793079*^9, 3.7287497255857153`*^9, 
  3.7287499689342527`*^9, 3.728751230978325*^9, 3.7287516591137857`*^9, 
  3.728759366462978*^9, 3.7288133045683117`*^9, 3.728813474750271*^9, 
  3.728813951774508*^9, 3.728827424928256*^9, 3.728836014516368*^9, 
  3.728836350287994*^9, 3.728843310859083*^9, 3.728845667895321*^9, 
  3.7333675485386257`*^9, 3.792273575067965*^9},
 CellLabel->"During evaluation of In[309]:=",
 CellID->2021829996,ExpressionUUID->"4360915f-4179-4f2c-90af-3c3b2b33cc8c"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The frame or basis of the chart has the same name. Its partial derivative is:\
\
\>", "MathCaption",
 CellChangeTimes->{{3.72867466575815*^9, 3.728674696467753*^9}},
 CellID->1206102974,ExpressionUUID->"0149e487-24df-4112-88c1-01a3d4f80003"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"PDOfBasis", "[", "csys", "]"}]], "Input",
 CellChangeTimes->{{3.728674697941228*^9, 3.728674707475024*^9}, {
  3.728748760270349*^9, 3.7287487613983393`*^9}},
 CellLabel->"In[310]:=",
 CellID->968864843,ExpressionUUID->"4dc55368-35fb-462e-8fad-26a59c8f7e6d"],

Cell[BoxData["PDcsys"], "Output",
 CellChangeTimes->{3.728674708138804*^9, 3.7286749220835876`*^9, 
  3.72874876205838*^9, 3.7287497259915047`*^9, 3.72874996923855*^9, 
  3.728751231398212*^9, 3.728751659431201*^9, 3.728759369368182*^9, 
  3.728813304937305*^9, 3.728813477552287*^9, 3.728813954361804*^9, 
  3.7288274253043222`*^9, 3.728836014944372*^9, 3.728836350678053*^9, 
  3.728843312775276*^9, 3.7288456682908983`*^9, 3.733367548971919*^9, 
  3.792273575151634*^9},
 CellLabel->"Out[310]=",
 CellID->1262996291,ExpressionUUID->"0f2bcd17-63ce-42fa-8ee7-81cc57a09b74"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Notice that now the torsion tensor vanishes and the Christoffel is symmetric.\
\
\>", "MathCaption",ExpressionUUID->"fdbf253a-036e-4b02-8002-2fad677fc882"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Torsion", "[", "PDcsys", "]"}], "[", 
  RowBox[{"a", ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.7286747151144753`*^9, 3.7286747181067657`*^9}, {
  3.7287487659659977`*^9, 3.728748767753706*^9}},
 CellLabel->
  "In[311]:=",ExpressionUUID->"97dc6d95-07e3-4293-a9e6-80739e43d581"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.7286747187554398`*^9, 3.728674922202756*^9, 
  3.7287487681993523`*^9, 3.728749726127124*^9, 3.728749969389655*^9, 
  3.728751231534273*^9, 3.7287516595517893`*^9, 3.728759370967626*^9, 
  3.728813305067898*^9, 3.728813478718979*^9, 3.728813955629243*^9, 
  3.7288274254378347`*^9, 3.72883601508139*^9, 3.728836350832745*^9, 
  3.7288433140570297`*^9, 3.728845668483439*^9, 3.733367549209805*^9, 
  3.79227357520924*^9},
 CellLabel->"Out[311]=",
 CellID->1450295889,ExpressionUUID->"937b8699-50bf-4341-994f-647c82fc212f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Bracket", "[", 
   RowBox[{
    RowBox[{"Basis", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "csys"}]}], "}"}], ",", "b"}], "]"}], ",", 
    RowBox[{"Basis", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"2", ",", 
        RowBox[{"-", "csys"}]}], "}"}], ",", "b"}], "]"}]}], "]"}], "[", "a", 
  "]"}]], "Input",
 CellChangeTimes->{{3.728674721058447*^9, 3.7286747327379513`*^9}, {
  3.728748769919778*^9, 3.72874877227808*^9}},
 CellLabel->
  "In[312]:=",ExpressionUUID->"57395fc5-5ba1-4375-8214-746988fc9497"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.72867473357367*^9, 3.728674922336952*^9, 
  3.728748772775812*^9, 3.728749726243553*^9, 3.728749969539049*^9, 
  3.728751231648952*^9, 3.7287516596667967`*^9, 3.7287593717892637`*^9, 
  3.7288133051869993`*^9, 3.7288134793683853`*^9, 3.728813956294676*^9, 
  3.728827425567247*^9, 3.728836015236248*^9, 3.72883635104375*^9, 
  3.7288433147591753`*^9, 3.7288456686156893`*^9, 3.7333675495455503`*^9, 
  3.792273575264426*^9},
 CellLabel->"Out[312]=",
 CellID->1183612973,ExpressionUUID->"056773b8-cdd3-40f1-a6eb-44f0357f42fc"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["TODO: Why is csys repeated?", "MathCaption",
 CellChangeTimes->{{3.728674944849308*^9, 3.728674961615156*^9}, {
  3.7287593785546923`*^9, 3.728759382794426*^9}},
 CellID->1917134628,ExpressionUUID->"29ef79f2-7c73-47dc-ba08-6a7ecaf4c298"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ChartsOfManifold", "[", "M3", "]"}]], "Input",
 CellLabel->
  "In[313]:=",ExpressionUUID->"5f515c29-fc9a-4e38-bf4e-a667c28835e0"],

Cell[BoxData[
 RowBox[{"{", "csys", "}"}]], "Output",
 CellChangeTimes->{3.728674738760294*^9, 3.728674922519051*^9, 
  3.7287487752533493`*^9, 3.7287497263787727`*^9, 3.728749969672037*^9, 
  3.728751231900982*^9, 3.7287516597860518`*^9, 3.728759384049623*^9, 
  3.728813305303528*^9, 3.728813480408283*^9, 3.728813957463113*^9, 
  3.728827425715337*^9, 3.728836015367872*^9, 3.728836351199623*^9, 
  3.728843315976447*^9, 3.728845668766878*^9, 3.7333675497063293`*^9, 
  3.7922735753187943`*^9},
 CellLabel->"Out[313]=",
 CellID->842215050,ExpressionUUID->"f4182371-c175-4a18-b48a-445f9ff692ef"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ChartQ", " ", "/@", " ", 
  RowBox[{"{", 
   RowBox[{"csys", ",", "fred"}], "}"}]}]], "Input",
 CellChangeTimes->{{3.72867475247285*^9, 3.728674756666709*^9}, {
  3.728748776853952*^9, 3.728748779573257*^9}},
 CellLabel->
  "In[314]:=",ExpressionUUID->"ab9e62f9-5c1d-46df-9c3d-4d7781053935"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"True", ",", "False"}], "}"}]], "Output",
 CellChangeTimes->{3.728674757414494*^9, 3.728674922638335*^9, 
  3.728748779924282*^9, 3.728749726511079*^9, 3.7287499698845043`*^9, 
  3.7287512324288883`*^9, 3.728751659919791*^9, 3.728759385698059*^9, 
  3.728813305420295*^9, 3.72881348105066*^9, 3.728813958111677*^9, 
  3.728827425854445*^9, 3.728836015497444*^9, 3.7288363513498287`*^9, 
  3.728843316675672*^9, 3.728845668906518*^9, 3.7333675499450083`*^9, 
  3.7922735753763857`*^9},
 CellLabel->"Out[314]=",
 CellID->1043627973,ExpressionUUID->"aa76bf58-6c95-41ba-8d5e-ddaf1ba1b910"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"BasisQ", " ", "/@", " ", 
  RowBox[{"{", 
   RowBox[{"csys", ",", "fred"}], "}"}]}]], "Input",
 CellChangeTimes->{{3.728674760248493*^9, 3.7286747613831472`*^9}, {
  3.7287487813257837`*^9, 3.7287487856306667`*^9}},
 CellLabel->
  "In[315]:=",ExpressionUUID->"da1a54cb-6658-4366-b610-770d278eb8de"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"True", ",", "True"}], "}"}]], "Output",
 CellChangeTimes->{3.728674762104776*^9, 3.7286749227699966`*^9, 
  3.7287487837616453`*^9, 3.72874972662663*^9, 3.728749970114472*^9, 
  3.728751232686573*^9, 3.728751660035303*^9, 3.728759387399218*^9, 
  3.7288133055529623`*^9, 3.7288134817010937`*^9, 3.728813958716507*^9, 
  3.728827425988894*^9, 3.728836015632001*^9, 3.7288363514997253`*^9, 
  3.7288433173731213`*^9, 3.728845669071231*^9, 3.733367550325303*^9, 
  3.7922735754289207`*^9},
 CellLabel->"Out[315]=",
 CellID->1756397162,ExpressionUUID->"f63bb3da-6350-4d8f-8af0-c1c3ed30abcd"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The derivative of a scalar field is a (coordinated) basis vector field:\
\>", "MathCaption",
 CellChangeTimes->{{3.7286747728317137`*^9, 3.728674794782421*^9}},
 CellID->1747104672,ExpressionUUID->"0642d060-3083-49ee-a7c1-1b43c16925d2"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"z", "[", "]"}], "]"}]], "Input",
 CellLabel->
  "In[316]:=",ExpressionUUID->"05271e5b-d22a-4176-90a6-b2e7d4f66d74"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"e", 
      StyleBox[GridBox[{
         {" ", 
          StyleBox["3",
           FontColor->RGBColor[1., 0.5, 0.]]},
         {"a", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`csys}],
  Editable->False]], "Output",
 CellChangeTimes->{3.72867476781098*^9, 3.728674922902606*^9, 
  3.7287487890212307`*^9, 3.728749726744773*^9, 3.7287499702568493`*^9, 
  3.7287512330326357`*^9, 3.728751660152069*^9, 3.728813305669779*^9, 
  3.728813482851063*^9, 3.728813960029499*^9, 3.728827426122901*^9, 
  3.728836015767551*^9, 3.728836351651122*^9, 3.728843319077046*^9, 
  3.72884566920226*^9, 3.733367550442318*^9, 3.792273575488482*^9},
 CellLabel->"Out[316]=",
 CellID->456890072,ExpressionUUID->"9d63c6e7-e948-4110-99e6-d5b4385fd01d"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
It does not matter which derivative do we actually use, because all covariant \
derivatives coincide on scalars:\
\>", "MathCaption",
 CellChangeTimes->{{3.728674798630528*^9, 3.72867482085387*^9}},
 CellID->1681265147,ExpressionUUID->"033280e2-a975-4a84-bd5d-0f10c7d41c0b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PDcsys", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"z", "[", "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.728674822199306*^9, 3.728674826645218*^9}, {
  3.728748791044774*^9, 3.728748792007152*^9}},
 CellLabel->"In[317]:=",
 CellID->1960971562,ExpressionUUID->"d1de9970-be80-44cf-8f94-b24ef9909869"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"e", 
      StyleBox[GridBox[{
         {" ", 
          StyleBox["3",
           FontColor->RGBColor[1., 0.5, 0.]]},
         {"a", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`csys}],
  Editable->False]], "Output",
 CellChangeTimes->{3.7286748270841627`*^9, 3.728674923053941*^9, 
  3.72874879248967*^9, 3.728749726851615*^9, 3.728749970418367*^9, 
  3.728751233477644*^9, 3.7287516602531347`*^9, 3.7287594054180117`*^9, 
  3.728813305796968*^9, 3.728813490789832*^9, 3.728813962029008*^9, 
  3.728827426249881*^9, 3.728836015899583*^9, 3.728836351800516*^9, 
  3.728843320643571*^9, 3.728845669332563*^9, 3.733367550780198*^9, 
  3.792273575546815*^9},
 CellLabel->"Out[317]=",
 CellID->1697402079,ExpressionUUID->"3705f40e-9737-4961-b1eb-d946a0a0ccfc"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
It is still typeset using the standard typesetting of frame fields. Undefine \
the chart and define it again, with a more \"coordinated\" formatting:\
\>", "MathCaption",
 CellChangeTimes->{{3.728759407449394*^9, 3.728759436104024*^9}, {
  3.728759904938377*^9, 3.72875993551993*^9}},
 CellID->810815228,ExpressionUUID->"571ced46-780a-4e73-a257-3c042132bf5e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefChart", "[", "csys", "]"}]], "Input",
 CellChangeTimes->{{3.7287598178320293`*^9, 3.728759823158008*^9}},
 CellLabel->"In[318]:=",
 CellID->1169243091,ExpressionUUID->"645abf26-050f-492f-baf9-1a48454f0cf5"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "x"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "coordinate scalar", 
   " ", $CellContext`x],
  Editable->False]], "Print",
 CellChangeTimes->{3.728759823616356*^9, 3.7288133059498663`*^9, 
  3.728813492142671*^9, 3.728813904459449*^9, 3.7288139442322407`*^9, 
  3.72882742645573*^9, 3.728836016092347*^9, 3.728836352019011*^9, 
  3.728843322175136*^9, 3.728845669505144*^9, 3.733367550927487*^9, 
  3.7922735756586514`*^9},
 CellLabel->"During evaluation of In[318]:=",
 CellID->1063975005,ExpressionUUID->"e05f4522-a60a-4b5c-bcd6-f19eaf6fdb4d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "y"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "coordinate scalar", 
   " ", $CellContext`y],
  Editable->False]], "Print",
 CellChangeTimes->{3.728759823616356*^9, 3.7288133059498663`*^9, 
  3.728813492142671*^9, 3.728813904459449*^9, 3.7288139442322407`*^9, 
  3.72882742645573*^9, 3.728836016092347*^9, 3.728836352019011*^9, 
  3.728843322175136*^9, 3.728845669505144*^9, 3.733367550927487*^9, 
  3.7922735756736*^9},
 CellLabel->"During evaluation of In[318]:=",
 CellID->131580466,ExpressionUUID->"f497c4a5-d407-49ed-8674-1d50a83fa9eb"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "z"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "coordinate scalar", 
   " ", $CellContext`z],
  Editable->False]], "Print",
 CellChangeTimes->{3.728759823616356*^9, 3.7288133059498663`*^9, 
  3.728813492142671*^9, 3.728813904459449*^9, 3.7288139442322407`*^9, 
  3.72882742645573*^9, 3.728836016092347*^9, 3.728836352019011*^9, 
  3.728843322175136*^9, 3.728845669505144*^9, 3.733367550927487*^9, 
  3.7922735756894493`*^9},
 CellLabel->"During evaluation of In[318]:=",
 CellID->1712219299,ExpressionUUID->"c162a5c9-c0ae-4157-890a-271f1e5e25ba"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"mapping differential tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "dicsys"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "mapping differential tensor", " ", $CellContext`dicsys],
  Editable->False]], "Print",
 CellChangeTimes->{3.728759823616356*^9, 3.7288133059498663`*^9, 
  3.728813492142671*^9, 3.728813904459449*^9, 3.7288139442322407`*^9, 
  3.72882742645573*^9, 3.728836016092347*^9, 3.728836352019011*^9, 
  3.728843322175136*^9, 3.728845669505144*^9, 3.733367550927487*^9, 
  3.792273575696063*^9},
 CellLabel->"During evaluation of In[318]:=",
 CellID->988888493,ExpressionUUID->"823e7470-850c-4326-b016-e0923c53207c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefMapping", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"inverse mapping\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "icsys"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefMapping, ": Undefined ", "inverse mapping", 
   " ", $CellContext`icsys],
  Editable->False]], "Print",
 CellChangeTimes->{3.728759823616356*^9, 3.7288133059498663`*^9, 
  3.728813492142671*^9, 3.728813904459449*^9, 3.7288139442322407`*^9, 
  3.72882742645573*^9, 3.728836016092347*^9, 3.728836352019011*^9, 
  3.728843322175136*^9, 3.728845669505144*^9, 3.733367550927487*^9, 
  3.79227357573057*^9},
 CellLabel->"During evaluation of In[318]:=",
 CellID->872957718,ExpressionUUID->"7719fc84-fe53-43b1-b90c-2754ab8f45e1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"mapping differential tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "dcsys"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "mapping differential tensor", " ", $CellContext`dcsys],
  Editable->False]], "Print",
 CellChangeTimes->{3.728759823616356*^9, 3.7288133059498663`*^9, 
  3.728813492142671*^9, 3.728813904459449*^9, 3.7288139442322407`*^9, 
  3.72882742645573*^9, 3.728836016092347*^9, 3.728836352019011*^9, 
  3.728843322175136*^9, 3.728845669505144*^9, 3.733367550927487*^9, 
  3.792273575740675*^9},
 CellLabel->"During evaluation of In[318]:=",
 CellID->192090122,ExpressionUUID->"571f9ae2-3175-408d-841e-6c895dd78c19"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "ChristoffelPDcsys"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Christoffel tensor", " ", $CellContext`ChristoffelPDcsys],
  Editable->False]], "Print",
 CellChangeTimes->{3.728759823616356*^9, 3.7288133059498663`*^9, 
  3.728813492142671*^9, 3.728813904459449*^9, 3.7288139442322407`*^9, 
  3.72882742645573*^9, 3.728836016092347*^9, 3.728836352019011*^9, 
  3.728843322175136*^9, 3.728845669505144*^9, 3.733367550927487*^9, 
  3.792273575747346*^9},
 CellLabel->"During evaluation of In[318]:=",
 CellID->790759013,ExpressionUUID->"047f0038-92a6-4c30-89f0-99235577ceb3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciPDcsys"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Ricci tensor", 
   " ", $CellContext`RicciPDcsys],
  Editable->False]], "Print",
 CellChangeTimes->{3.728759823616356*^9, 3.7288133059498663`*^9, 
  3.728813492142671*^9, 3.728813904459449*^9, 3.7288139442322407`*^9, 
  3.72882742645573*^9, 3.728836016092347*^9, 3.728836352019011*^9, 
  3.728843322175136*^9, 3.728845669505144*^9, 3.733367550927487*^9, 
  3.792273575753777*^9},
 CellLabel->"During evaluation of In[318]:=",
 CellID->1657923855,ExpressionUUID->"993a695d-a0a0-4bcd-9225-9313770a5368"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannPDcsys"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannPDcsys],
  Editable->False]], "Print",
 CellChangeTimes->{3.728759823616356*^9, 3.7288133059498663`*^9, 
  3.728813492142671*^9, 3.728813904459449*^9, 3.7288139442322407`*^9, 
  3.72882742645573*^9, 3.728836016092347*^9, 3.728836352019011*^9, 
  3.728843322175136*^9, 3.728845669505144*^9, 3.733367550927487*^9, 
  3.7922735757606077`*^9},
 CellLabel->"During evaluation of In[318]:=",
 CellID->1125911160,ExpressionUUID->"11d08324-3f6c-4ab8-ac47-068f2a499603"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TorsionPDcsys"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionPDcsys],
  Editable->False]], "Print",
 CellChangeTimes->{3.728759823616356*^9, 3.7288133059498663`*^9, 
  3.728813492142671*^9, 3.728813904459449*^9, 3.7288139442322407`*^9, 
  3.72882742645573*^9, 3.728836016092347*^9, 3.728836352019011*^9, 
  3.728843322175136*^9, 3.728845669505144*^9, 3.733367550927487*^9, 
  3.792273575767735*^9},
 CellLabel->"During evaluation of In[318]:=",
 CellID->1710452403,ExpressionUUID->"b8e9b156-5c3e-4729-87ee-b730b3aca8da"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"parallel derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "PDcsys"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "parallel derivative", 
   " ", $CellContext`PDcsys],
  Editable->False]], "Print",
 CellChangeTimes->{3.728759823616356*^9, 3.7288133059498663`*^9, 
  3.728813492142671*^9, 3.728813904459449*^9, 3.7288139442322407`*^9, 
  3.72882742645573*^9, 3.728836016092347*^9, 3.728836352019011*^9, 
  3.728843322175136*^9, 3.728845669505144*^9, 3.733367550927487*^9, 
  3.792273575774942*^9},
 CellLabel->"During evaluation of In[318]:=",
 CellID->1357438479,ExpressionUUID->"6f788af2-b1c7-4987-9f35-ced71ba62bba"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "etaUpcsys"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric +1 density",
    " ", $CellContext`etaUpcsys],
  Editable->False]], "Print",
 CellChangeTimes->{3.728759823616356*^9, 3.7288133059498663`*^9, 
  3.728813492142671*^9, 3.728813904459449*^9, 3.7288139442322407`*^9, 
  3.72882742645573*^9, 3.728836016092347*^9, 3.728836352019011*^9, 
  3.728843322175136*^9, 3.728845669505144*^9, 3.733367550927487*^9, 
  3.79227357578205*^9},
 CellLabel->"During evaluation of In[318]:=",
 CellID->1749734472,ExpressionUUID->"7a082e34-c8bf-418d-9e3b-f079b3575780"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "etaDowncsys"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric -1 density",
    " ", $CellContext`etaDowncsys],
  Editable->False]], "Print",
 CellChangeTimes->{3.728759823616356*^9, 3.7288133059498663`*^9, 
  3.728813492142671*^9, 3.728813904459449*^9, 3.7288139442322407`*^9, 
  3.72882742645573*^9, 3.728836016092347*^9, 3.728836352019011*^9, 
  3.728843322175136*^9, 3.728845669505144*^9, 3.733367550927487*^9, 
  3.792273575789686*^9},
 CellLabel->"During evaluation of In[318]:=",
 CellID->146208886,ExpressionUUID->"2536e67a-3350-4c03-95a3-8087cd59528d"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefChart", "[", 
  RowBox[{"csys", ",", "M3", ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", "2", ",", "3"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"x", "[", "]"}], ",", 
     RowBox[{"y", "[", "]"}], ",", 
     RowBox[{"z", "[", "]"}]}], "}"}], ",", 
   RowBox[{"ChartColor", "\[Rule]", "Orange"}], ",", 
   RowBox[{"FormatBasis", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"\"\<Partials\>\"", ",", "\"\<Differentials\>\""}], "}"}]}]}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.728759835759212*^9, 3.7287598709496117`*^9}, {
  3.728813497777852*^9, 3.72881350087189*^9}, {3.7288139079431763`*^9, 
  3.72881390887495*^9}},
 CellLabel->"In[319]:=",
 CellID->1189915789,ExpressionUUID->"46ec0d42-03d5-4767-800e-ea425709de77"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefChart", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"chart \"\>", "\[InvisibleSpace]", "csys", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xCoba`DefChart, ": Defining ", "chart ", $CellContext`csys, 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287598718896017`*^9, 3.728813306235182*^9, 
  3.728813511133918*^9, 3.728813909571615*^9, 3.728827426867485*^9, 
  3.728836016444899*^9, 3.728836352455557*^9, 3.728843323803298*^9, 
  3.728845669882051*^9, 3.7333675512452*^9, 3.792273575848734*^9},
 CellLabel->"During evaluation of In[319]:=",
 CellID->611454834,ExpressionUUID->"d0310852-a8fc-4fc7-86cb-da2588576e75"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"x", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "coordinate scalar ", 
   $CellContext`x[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287598718896017`*^9, 3.728813306235182*^9, 
  3.728813511133918*^9, 3.728813909571615*^9, 3.728827426867485*^9, 
  3.728836016444899*^9, 3.728836352455557*^9, 3.728843323803298*^9, 
  3.728845669882051*^9, 3.7333675512452*^9, 3.7922735758685913`*^9},
 CellLabel->"During evaluation of In[319]:=",
 CellID->1012144143,ExpressionUUID->"2e073ffe-30ad-4594-9c29-b60cbe2d6df9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"y", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "coordinate scalar ", 
   $CellContext`y[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287598718896017`*^9, 3.728813306235182*^9, 
  3.728813511133918*^9, 3.728813909571615*^9, 3.728827426867485*^9, 
  3.728836016444899*^9, 3.728836352455557*^9, 3.728843323803298*^9, 
  3.728845669882051*^9, 3.7333675512452*^9, 3.7922735758926687`*^9},
 CellLabel->"During evaluation of In[319]:=",
 CellID->888643514,ExpressionUUID->"9aa75be1-6ae7-4026-afb2-586c4b7dede6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"z", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "coordinate scalar ", 
   $CellContext`z[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287598718896017`*^9, 3.728813306235182*^9, 
  3.728813511133918*^9, 3.728813909571615*^9, 3.728827426867485*^9, 
  3.728836016444899*^9, 3.728836352455557*^9, 3.728843323803298*^9, 
  3.728845669882051*^9, 3.7333675512452*^9, 3.792273575904203*^9},
 CellLabel->"During evaluation of In[319]:=",
 CellID->1830258560,ExpressionUUID->"cdd6c079-0b49-4477-b718-a42dbd888b05"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefMapping", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"mapping \"\>", "\[InvisibleSpace]", "csys", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefMapping, ": Defining ", 
   "mapping ", $CellContext`csys, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287598718896017`*^9, 3.728813306235182*^9, 
  3.728813511133918*^9, 3.728813909571615*^9, 3.728827426867485*^9, 
  3.728836016444899*^9, 3.728836352455557*^9, 3.728843323803298*^9, 
  3.728845669882051*^9, 3.7333675512452*^9, 3.792273575915461*^9},
 CellLabel->"During evaluation of In[319]:=",
 CellID->599858292,ExpressionUUID->"39db6774-a18c-47e8-95f8-c18c05a90538"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefMapping", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"inverse mapping \"\>", "\[InvisibleSpace]", 
   "icsys", "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefMapping, ": Defining ", 
   "inverse mapping ", $CellContext`icsys, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287598718896017`*^9, 3.728813306235182*^9, 
  3.728813511133918*^9, 3.728813909571615*^9, 3.728827426867485*^9, 
  3.728836016444899*^9, 3.728836352455557*^9, 3.728843323803298*^9, 
  3.728845669882051*^9, 3.7333675512452*^9, 3.792273575926722*^9},
 CellLabel->"During evaluation of In[319]:=",
 CellID->1530074626,ExpressionUUID->"0c729604-30f8-4f63-acc1-8d2c0db30ea9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"mapping differential tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"dicsys", "[", 
    RowBox[{
     RowBox[{"-", "\[FormalA]"}], ",", "icsysa"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "mapping differential tensor ", 
   $CellContext`dicsys[-\[FormalA], $CellContext`icsysa], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287598718896017`*^9, 3.728813306235182*^9, 
  3.728813511133918*^9, 3.728813909571615*^9, 3.728827426867485*^9, 
  3.728836016444899*^9, 3.728836352455557*^9, 3.728843323803298*^9, 
  3.728845669882051*^9, 3.7333675512452*^9, 3.792273575936935*^9},
 CellLabel->"During evaluation of In[319]:=",
 CellID->1995050004,ExpressionUUID->"dd09ccdc-8dab-4c61-976e-c39f6856f7b9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"mapping differential tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"dcsys", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", "csys\[FormalA]"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "mapping differential tensor ", 
   $CellContext`dcsys[-$CellContext`a, $CellContext`csys\[FormalA]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287598718896017`*^9, 3.728813306235182*^9, 
  3.728813511133918*^9, 3.728813909571615*^9, 3.728827426867485*^9, 
  3.728836016444899*^9, 3.728836352455557*^9, 3.728843323803298*^9, 
  3.728845669882051*^9, 3.7333675512452*^9, 3.792273575946066*^9},
 CellLabel->"During evaluation of In[319]:=",
 CellID->2143236617,ExpressionUUID->"267fa9be-bdc8-4602-8489-a39bbba78c40"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefBasis", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"basis \"\>", "\[InvisibleSpace]", "csys", 
   "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Coordinated basis.\"\>"}],
  SequenceForm[
  "** ", xAct`xCoba`DefBasis, ": Defining ", "basis ", $CellContext`csys, 
   ". ", "Coordinated basis."],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287598718896017`*^9, 3.728813306235182*^9, 
  3.728813511133918*^9, 3.728813909571615*^9, 3.728827426867485*^9, 
  3.728836016444899*^9, 3.728836352455557*^9, 3.728843323803298*^9, 
  3.728845669882051*^9, 3.7333675512452*^9, 3.7922735759552727`*^9},
 CellLabel->"During evaluation of In[319]:=",
 CellID->1634610788,ExpressionUUID->"152e6ef6-67d1-4ae5-9391-144e3cb3a6b8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"parallel derivative \"\>", "\[InvisibleSpace]", 
   RowBox[{"PDcsys", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "parallel derivative ", 
   $CellContext`PDcsys[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287598718896017`*^9, 3.728813306235182*^9, 
  3.728813511133918*^9, 3.728813909571615*^9, 3.728827426867485*^9, 
  3.728836016444899*^9, 3.728836352455557*^9, 3.728843323803298*^9, 
  3.728845669882051*^9, 3.7333675512452*^9, 3.792273575963687*^9},
 CellLabel->"During evaluation of In[319]:=",
 CellID->145217845,ExpressionUUID->"fbfdd177-fba2-4890-89ab-cc47a721fa7a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing torsion tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TorsionPDcsys", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing torsion tensor ", 
   $CellContext`TorsionPDcsys[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287598718896017`*^9, 3.728813306235182*^9, 
  3.728813511133918*^9, 3.728813909571615*^9, 3.728827426867485*^9, 
  3.728836016444899*^9, 3.728836352455557*^9, 3.728843323803298*^9, 
  3.728845669882051*^9, 3.7333675512452*^9, 3.792273575972538*^9},
 CellLabel->"During evaluation of In[319]:=",
 CellID->1134897923,ExpressionUUID->"e3ef4c78-e613-4e63-be2e-3b2f8e401c26"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelPDcsys", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Christoffel tensor ", 
   $CellContext`ChristoffelPDcsys[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287598718896017`*^9, 3.728813306235182*^9, 
  3.728813511133918*^9, 3.728813909571615*^9, 3.728827426867485*^9, 
  3.728836016444899*^9, 3.728836352455557*^9, 3.728843323803298*^9, 
  3.728845669882051*^9, 3.7333675512452*^9, 3.792273575981057*^9},
 CellLabel->"During evaluation of In[319]:=",
 CellID->279639207,ExpressionUUID->"0a86ff13-d8af-47d3-a264-21e77869d924"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Riemann tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RiemannPDcsys", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Riemann tensor ", 
   $CellContext`RiemannPDcsys[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, $CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287598718896017`*^9, 3.728813306235182*^9, 
  3.728813511133918*^9, 3.728813909571615*^9, 3.728827426867485*^9, 
  3.728836016444899*^9, 3.728836352455557*^9, 3.728843323803298*^9, 
  3.728845669882051*^9, 3.7333675512452*^9, 3.7922735759893312`*^9},
 CellLabel->"During evaluation of In[319]:=",
 CellID->1201932413,ExpressionUUID->"7233746a-8c44-42b7-bf23-bf180043fc81"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciPDcsys", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Ricci tensor ", 
   $CellContext`RicciPDcsys[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287598718896017`*^9, 3.728813306235182*^9, 
  3.728813511133918*^9, 3.728813909571615*^9, 3.728827426867485*^9, 
  3.728836016444899*^9, 3.728836352455557*^9, 3.728843323803298*^9, 
  3.728845669882051*^9, 3.7333675512452*^9, 3.792273575997426*^9},
 CellLabel->"During evaluation of In[319]:=",
 CellID->1646881609,ExpressionUUID->"da51a13a-7e8b-4352-aa03-75fde2fba632"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaUpcsys", "[", 
    RowBox[{"a", ",", "b", ",", "c"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric +1 density ", 
   $CellContext`etaUpcsys[$CellContext`a, $CellContext`b, $CellContext`c], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287598718896017`*^9, 3.728813306235182*^9, 
  3.728813511133918*^9, 3.728813909571615*^9, 3.728827426867485*^9, 
  3.728836016444899*^9, 3.728836352455557*^9, 3.728843323803298*^9, 
  3.728845669882051*^9, 3.7333675512452*^9, 3.792273576005048*^9},
 CellLabel->"During evaluation of In[319]:=",
 CellID->222476280,ExpressionUUID->"e358d9e9-7a39-4c69-be48-be28126ef9d7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaDowncsys", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric -1 density ", 
   $CellContext`etaDowncsys[-$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287598718896017`*^9, 3.728813306235182*^9, 
  3.728813511133918*^9, 3.728813909571615*^9, 3.728827426867485*^9, 
  3.728836016444899*^9, 3.728836352455557*^9, 3.728843323803298*^9, 
  3.728845669882051*^9, 3.7333675512452*^9, 3.792273576012496*^9},
 CellLabel->"During evaluation of In[319]:=",
 CellID->389808533,ExpressionUUID->"252e6a61-3522-4418-a9c9-3bf07af1f9ff"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Now these are the coordinate vector fields of the frame, in a typesetting \
that reminds us of their \"directional derivative\" meaning:\
\>", "MathCaption",
 CellChangeTimes->{{3.728759944056485*^9, 3.728759955344122*^9}, {
  3.7287602037832727`*^9, 3.728760235859098*^9}},
 CellID->1470402736,ExpressionUUID->"63652355-9688-4749-83b3-9c5e73407244"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"basis", "=", 
  RowBox[{
   RowBox[{"BasisArray", "[", "csys", "]"}], "[", "a", "]"}]}]], "Input",
 CellChangeTimes->{{3.72875988289143*^9, 3.728759888530981*^9}, {
  3.728759985706764*^9, 3.728759986726553*^9}},
 CellLabel->"In[320]:=",
 CellID->1556741925,ExpressionUUID->"c32bdf03-3459-4454-a7d1-dd381c6f730f"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        FractionBox["\[PartialD]", 
         RowBox[{"\[PartialD]", "x"}]], 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`csys}, $CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        FractionBox["\[PartialD]", 
         RowBox[{"\[PartialD]", "y"}]], 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`csys}, $CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        FractionBox["\[PartialD]", 
         RowBox[{"\[PartialD]", "z"}]], 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`csys}, $CellContext`a],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.7287598890073977`*^9, 3.728759987408491*^9, 
  3.728813306536767*^9, 3.7288135131544*^9, 3.728813911432745*^9, 
  3.72882742720998*^9, 3.728836016847184*^9, 3.728836352907651*^9, 
  3.728843325793874*^9, 3.728845670332925*^9, 3.733367551660479*^9, 
  3.7922735760771303`*^9},
 CellLabel->"Out[320]=",
 CellID->1686890781,ExpressionUUID->"05302c71-fdb5-4b09-b086-1751ae0d7ec0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["And these are the coordinate covector fields of the frame:", \
"MathCaption",
 CellChangeTimes->{{3.728759958535903*^9, 3.728759968175482*^9}},
 CellID->1922101666,ExpressionUUID->"712b40e7-02fd-4eea-adc3-8e858e94d9b0"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"cobasis", "=", 
  RowBox[{
   RowBox[{"BasisArray", "[", "csys", "]"}], "[", 
   RowBox[{"-", "a"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.7287598921135483`*^9, 3.728759896721829*^9}, {
  3.728759988535306*^9, 3.728759990390338*^9}},
 CellLabel->"In[321]:=",
 CellID->1196988032,ExpressionUUID->"4f8374b5-f626-47aa-bb7f-1fd38eee6997"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"dx", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`csys}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"dy", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`csys}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"dz", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`csys}],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728759897189481*^9, 3.728759990808138*^9, 
  3.7288133066539497`*^9, 3.72881351463827*^9, 3.72881391243251*^9, 
  3.728827427326313*^9, 3.728836016984392*^9, 3.728836353123848*^9, 
  3.728843327143715*^9, 3.728845670531933*^9, 3.733367551840822*^9, 
  3.792273576136889*^9},
 CellLabel->"Out[321]=",
 CellID->494248573,ExpressionUUID->"1eed1a2d-e1df-4794-a291-55e40587c278"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The cobasis can be obtained with any derivative operator:", \
"MathCaption",
 CellChangeTimes->{{3.728760109546154*^9, 3.728760124816135*^9}},
 CellID->1981624549,ExpressionUUID->"de30eb0d-09dd-4b2e-bdde-d3357f5e2e27"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"-", "a"}], "]"}], "/@", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"x", "[", "]"}], ",", 
    RowBox[{"y", "[", "]"}], ",", 
    RowBox[{"z", "[", "]"}]}], "}"}]}]], "Input",
 CellChangeTimes->{{3.728760074452353*^9, 3.7287600961055937`*^9}},
 CellLabel->"In[322]:=",
 CellID->79371859,ExpressionUUID->"1a865b04-696f-44ce-b33f-19ec534977ed"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"dx", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`csys}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"dy", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`csys}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"dz", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`csys}],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{{3.728760080145653*^9, 3.728760096543372*^9}, 
   3.7288133067704287`*^9, 3.728813515869375*^9, 3.728813913521278*^9, 
   3.728827427450243*^9, 3.728836017126157*^9, 3.728836353336318*^9, 
   3.728843328777075*^9, 3.728845670679953*^9, 3.7333675519877443`*^9, 
   3.792273576217202*^9},
 CellLabel->"Out[322]=",
 CellID->369019815,ExpressionUUID->"564650e2-d4b1-4877-9823-8c019fe8ff1b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Take any scalar function and compute its gradient covector:", \
"MathCaption",
 CellChangeTimes->{{3.728760566869066*^9, 3.728760580483162*^9}, {
  3.728760673382712*^9, 3.7287606760387363`*^9}},
 CellID->342164511,ExpressionUUID->"662820fe-7e52-4032-aedd-ddbc79e45f2b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefScalarFunction", "[", "func", "]"}]], "Input",
 CellChangeTimes->{{3.728760585054123*^9, 3.728760595410371*^9}},
 CellLabel->"In[323]:=",
 CellID->1115327372,ExpressionUUID->"75ab1fbd-2e11-4d3e-b62b-b24d39ba103b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefScalarFunction", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"scalar function \"\>", "\[InvisibleSpace]", 
   "func", "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefScalarFunction, ": Defining ", 
   "scalar function ", $CellContext`func, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728760597053225*^9, 3.72881330691815*^9, 
  3.728827427549015*^9, 3.7288360172880507`*^9, 3.728836353586976*^9, 
  3.728843330310925*^9, 3.728845670816105*^9, 3.7333675521228523`*^9, 
  3.792273576265637*^9},
 CellLabel->"During evaluation of In[323]:=",
 CellID->183258169,ExpressionUUID->"675753f9-b01b-494b-ad75-00deee28831a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"grad", "=", 
  RowBox[{
   RowBox[{"PDcsys", "[", 
    RowBox[{"-", "a"}], "]"}], "[", 
   RowBox[{"func", "[", 
    RowBox[{
     RowBox[{"x", "[", "]"}], ",", 
     RowBox[{"y", "[", "]"}], ",", 
     RowBox[{"z", "[", "]"}]}], "]"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.72876059880632*^9, 3.72876062092083*^9}, {
  3.728760807346549*^9, 3.72876080828055*^9}},
 CellLabel->"In[324]:=",
 CellID->1740807702,ExpressionUUID->"dcebf824-8a2d-4ae4-a5c1-4fd39b00dec5"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"dz", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`csys}],
    Editable->False], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["func",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`func,
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "0", ",", "1"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["x",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`x[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["y",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`y[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["z",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`z[],
      Editable->False]}], "]"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"dy", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`csys}],
    Editable->False], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["func",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`func,
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "1", ",", "0"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["x",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`x[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["y",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`y[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["z",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`z[],
      Editable->False]}], "]"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"dx", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`csys}],
    Editable->False], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["func",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`func,
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"1", ",", "0", ",", "0"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["x",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`x[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["y",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`y[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["z",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`z[],
      Editable->False]}], "]"}]}]}]], "Output",
 CellChangeTimes->{{3.7287606046495543`*^9, 3.728760621426284*^9}, 
   3.728760809005278*^9, 3.728813307043704*^9, 3.7288274276826963`*^9, 
   3.7288360174388237`*^9, 3.728836353772678*^9, 3.7288433310374603`*^9, 
   3.7288456709841547`*^9, 3.73336755251182*^9, 3.7922735763330097`*^9},
 CellLabel->"Out[324]=",
 CellID->528496585,ExpressionUUID->"ee975225-9b93-4155-b3e1-8449ba8ad760"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["2.2.2. Coordinate expressions", "Subsubsection",
 CellChangeTimes->{{3.728663829532053*^9, 3.728663849803274*^9}, {
  3.728670543491893*^9, 3.728670543859935*^9}, {3.72867306644833*^9, 
  3.728673079878126*^9}, {3.7286734857870407`*^9, 3.728673489676098*^9}, {
  3.728759263849457*^9, 3.72875926969563*^9}, {3.7288128856373863`*^9, 
  3.728812910778777*^9}},
 CellID->618036468,ExpressionUUID->"06d8f700-ca08-4499-b174-8fbdcbd1bca2"],

Cell["\<\
xCoba` follows xTensor` in the fact that all expressions are full tensor \
fields, and not function representations in a given chart. That is, we will \
always work with functions of scalar fields, like Sin[x[]], and not just \
functions of variables, like Sin[x]. It is very important to pay attention to \
this, because it is a frequent mistake to forget the empty pairs of brackets \
when working with xCoba`.\
\>", "Text",
 CellChangeTimes->{{3.728812913695751*^9, 3.7288130748835573`*^9}},
 CellID->1917193779,ExpressionUUID->"16c82575-a5f5-4d92-ab74-5bb9b2b9e00c"],

Cell["\<\
xCoba` provides the general scalar function object InChart to convert a \
scalar field into a scalar function in a given chart. Using InChart is not \
really required in standard explicit computations. See many examples further \
down in subsection Curvature computations of the section CTensor framework, \
none of which uses InChart. Essentially, InChart is a composition operator, \
composing a scalar field with a chart, producing a function of coordinates (a \
\"scalar function\" in xTensor's terminology).\
\>", "Text",
 CellChangeTimes->{{3.728813184487438*^9, 3.728813216316896*^9}, {
   3.728846092838039*^9, 3.728846129673686*^9}, {3.728846162047709*^9, 
   3.728846295681094*^9}, 3.7333675611923733`*^9},
 CellID->177552500,ExpressionUUID->"d4a04d56-b5bd-439e-9477-c68fc3ace364"],

Cell[CellGroupData[{

Cell["This is a scalar field on M3:", "MathCaption",
 CellChangeTimes->{{3.728813329479272*^9, 3.728813349174572*^9}},
 CellID->582647065,ExpressionUUID->"da01fe8f-18a0-46df-a371-0a10461d0da8"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"u", "[", "]"}]], "Input",
 CellChangeTimes->{{3.728813219687796*^9, 3.728813223108273*^9}, {
  3.72881335481728*^9, 3.7288133561101437`*^9}},
 CellLabel->"In[325]:=",
 CellID->1224311109,ExpressionUUID->"332a3e92-f1cf-4176-b233-1551e1a92613"],

Cell[BoxData[
 InterpretationBox[
  StyleBox["u",
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`u[],
  Editable->False]], "Output",
 CellChangeTimes->{3.7288132235183496`*^9, 3.7288133071751633`*^9, 
  3.728813356517745*^9, 3.728813445383758*^9, 3.7288135280280457`*^9, 
  3.728813917889847*^9, 3.728827427833652*^9, 3.728836017582819*^9, 
  3.728836353970089*^9, 3.728843339078989*^9, 3.7288456711196213`*^9, 
  3.733367565989151*^9, 3.792273588119995*^9},
 CellLabel->"Out[325]=",
 CellID->833008401,ExpressionUUID->"403beafd-f0e7-40c6-9307-50d58f5295a7"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
This is an equivalent field, expressing the field u[] as a scalar function of \
the scalars of a chart on M3. Note the underlining:\
\>", "MathCaption",
 CellChangeTimes->{{3.7288133621419153`*^9, 3.728813394764584*^9}, {
  3.7288135338299713`*^9, 3.7288135448457613`*^9}},
 CellID->188171900,ExpressionUUID->"dd68cd28-2904-45eb-846b-9117ed84270c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", 
  RowBox[{"InChart", "[", "csys", "]"}]}]], "Input",
 CellChangeTimes->{{3.728813447210281*^9, 3.728813454193494*^9}},
 CellLabel->"In[326]:=",
 CellID->1431482550,ExpressionUUID->"f157f915-70f8-4b3d-b640-3c27f0119c65"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    UnderscriptBox["u", 
     StyleBox["_",
      FontColor->RGBColor[1, 0.5, 0]]],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xCoba`InChart[
    $CellContext`u[], $CellContext`csys],
   Editable->False], "[", 
  RowBox[{
   InterpretationBox[
    StyleBox["x",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`x[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["y",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`y[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["z",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`z[],
    Editable->False]}], "]"}]], "Output",
 CellChangeTimes->{3.7288134546197853`*^9, 3.7288135292088757`*^9, 
  3.728813918972187*^9, 3.728827427971204*^9, 3.7288360177212133`*^9, 
  3.728836354376079*^9, 3.728843340295084*^9, 3.7288456712516127`*^9, 
  3.733367566176292*^9, 3.7922735895364857`*^9},
 CellLabel->"Out[326]=",
 CellID->1784566159,ExpressionUUID->"b0c3b5f0-9c1a-4d6c-b1ab-822b67dda615"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellChangeTimes->{{3.728813556764539*^9, 3.7288135587798557`*^9}},
 CellLabel->"In[327]:=",
 CellID->2028784213,ExpressionUUID->"37d6cbac-c31a-425b-ab20-704fa8613a08"],

Cell["InChart[u[], csys][x[], y[], z[]]", "Output",
 CellChangeTimes->{3.7288135597789803`*^9, 3.728813926455379*^9, 
  3.7288274280976467`*^9, 3.7288360178657103`*^9, 3.728836355586059*^9, 
  3.72884334105026*^9, 3.728845671388435*^9, 3.7333675663101397`*^9, 
  3.7922735901754723`*^9},
 CellLabel->"Out[327]//InputForm=",
 CellID->1094520716,ExpressionUUID->"0472a165-0158-4aac-84a2-dd01fa605ee5"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["u and InChart[u[], csys] are different types of objects:", "MathCaption",
 CellChangeTimes->{{3.728813596651155*^9, 3.728813613066394*^9}, {
  3.72881366719458*^9, 3.728813668943417*^9}},
 CellID->2105785373,ExpressionUUID->"76c2dc90-6cb8-46e9-8d3d-59bf99af9c95"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"xTensorQ", "/@", 
  RowBox[{"{", 
   RowBox[{"u", ",", 
    RowBox[{"InChart", "[", 
     RowBox[{
      RowBox[{"u", "[", "]"}], ",", "csys"}], "]"}]}], "}"}]}]], "Input",
 CellChangeTimes->{{3.728813615538437*^9, 3.728813649450162*^9}},
 CellLabel->"In[328]:=",
 CellID->19999940,ExpressionUUID->"e1cdf93c-5e88-453c-a556-8366dd461e03"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"True", ",", "False"}], "}"}]], "Output",
 CellChangeTimes->{{3.728813621069611*^9, 3.728813649903812*^9}, 
   3.728813927913357*^9, 3.728827428216194*^9, 3.728836018005651*^9, 
   3.728836355745364*^9, 3.7288433422677383`*^9, 3.728845671518306*^9, 
   3.73336756644366*^9, 3.792273591177796*^9},
 CellLabel->"Out[328]=",
 CellID->1057668632,ExpressionUUID->"47bf97f0-e7a6-4e2d-bd33-63cfca1e7f78"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ScalarFunctionQ", "/@", 
  RowBox[{"{", 
   RowBox[{"u", ",", 
    RowBox[{"InChart", "[", 
     RowBox[{
      RowBox[{"u", "[", "]"}], ",", "csys"}], "]"}]}], "}"}]}]], "Input",
 CellChangeTimes->{{3.728813624392311*^9, 3.728813657696249*^9}},
 CellLabel->"In[329]:=",
 CellID->87718649,ExpressionUUID->"e3ec0f34-fa8a-4428-892e-756f312e2bb4"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"False", ",", "True"}], "}"}]], "Output",
 CellChangeTimes->{{3.728813637530959*^9, 3.728813658224285*^9}, 
   3.7288139285899067`*^9, 3.728827428347533*^9, 3.728836018157734*^9, 
   3.7288363563533287`*^9, 3.728843342962726*^9, 3.728845671651202*^9, 
   3.7333675665778217`*^9, 3.79227359176507*^9},
 CellLabel->"Out[329]=",
 CellID->361095595,ExpressionUUID->"5d7fe42c-3200-4f47-9288-06fc1704757c"]
}, Open  ]]
}, Open  ]],

Cell["\<\
The field u[] could never be replaced by the function Function[{x, y, z}, \
Sin[x+y+z]], but the scalar function InChart[u[], csys] can be replaced by \
such function.\
\>", "MathCaption",
 CellChangeTimes->{{3.7288137183906727`*^9, 3.7288137803868713`*^9}},
 CellID->1622491755,ExpressionUUID->"ffaf3618-4430-4a6a-be7c-f66250c7de20"],

Cell["\<\
InChart can be used when one wants to perform chart computations with \
symbolic tensor components, but it is not needed with explict functions.\
\>", "Text",
 CellChangeTimes->{{3.7288138112904377`*^9, 3.7288138527833033`*^9}},
 CellID->799280702,ExpressionUUID->"b1430c5d-7ac8-453c-981d-f7c559b4f1c8"],

Cell[CellGroupData[{

Cell["CTensor of components of a tensor field v:", "MathCaption",
 CellChangeTimes->{{3.7288140077293177`*^9, 3.728814016255932*^9}, {
  3.728836596390792*^9, 3.728836596470874*^9}},
 CellID->2013765975,ExpressionUUID->"a313a714-5727-41ad-b943-37c58125bd94"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"vector", "=", 
  RowBox[{
   RowBox[{"ToCTensor", "[", 
    RowBox[{"v", ",", 
     RowBox[{"{", "csys", "}"}]}], "]"}], "[", "a", "]"}]}]], "Input",
 CellChangeTimes->{{3.728813862700075*^9, 3.728813885189576*^9}, {
  3.728825288492765*^9, 3.728825289203237*^9}, {3.7288365839153357`*^9, 
  3.728836609087072*^9}},
 CellLabel->"In[330]:=",
 CellID->920876781,ExpressionUUID->"3d8e0e67-93d3-4378-b698-1a0aa95a1a12"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            InterpretationBox[
             StyleBox[GridBox[{
                {"v", 
                 StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`v[{1, $CellContext`csys}],
             Editable->False],
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", 
              InterpretationBox[
               StyleBox[
                GridBox[{{"v", 
                   StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                 ColumnSpacings -> 0.05, RowAlignments -> Center], 
                AutoSpacing -> False, ShowAutoStyles -> False], 
               $CellContext`v[{1, $CellContext`csys}], Editable -> False]}]],
           
           Annotation[#, {1} -> $CellContext`v[{1, $CellContext`csys}], 
            "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            InterpretationBox[
             StyleBox[GridBox[{
                {"v", 
                 StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`v[{2, $CellContext`csys}],
             Editable->False],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              InterpretationBox[
               StyleBox[
                GridBox[{{"v", 
                   StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                 ColumnSpacings -> 0.05, RowAlignments -> Center], 
                AutoSpacing -> False, ShowAutoStyles -> False], 
               $CellContext`v[{2, $CellContext`csys}], Editable -> False]}]],
           
           Annotation[#, {2} -> $CellContext`v[{2, $CellContext`csys}], 
            "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            InterpretationBox[
             StyleBox[GridBox[{
                {"v", 
                 StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`v[{3, $CellContext`csys}],
             Editable->False],
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", 
              InterpretationBox[
               StyleBox[
                GridBox[{{"v", 
                   StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                 ColumnSpacings -> 0.05, RowAlignments -> Center], 
                AutoSpacing -> False, ShowAutoStyles -> False], 
               $CellContext`v[{3, $CellContext`csys}], Editable -> False]}]],
           
           Annotation[#, {3} -> $CellContext`v[{3, $CellContext`csys}], 
            "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0.5, 0], 2 -> RGBColor[1, 0.5, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{
    $CellContext`v[{1, $CellContext`csys}], 
    $CellContext`v[{2, $CellContext`csys}], 
    $CellContext`v[{3, $CellContext`csys}]}, {$CellContext`csys}, 
   0][$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{{3.728813879650221*^9, 3.728813885805361*^9}, 
   3.728813930307811*^9, 3.7288140788624496`*^9, {3.728814845010572*^9, 
   3.728814931003628*^9}, {3.728814986513379*^9, 3.728815008413872*^9}, 
   3.7288152091089907`*^9, 3.728815601543112*^9, {3.728815646059751*^9, 
   3.728815674896533*^9}, 3.728820199863452*^9, 3.728825289756854*^9, 
   3.728826164632386*^9, 3.728827428500448*^9, 3.728827570489872*^9, {
   3.728827602902502*^9, 3.728827646921155*^9}, 3.728828096340921*^9, 
   3.72883601831933*^9, 3.728836356741312*^9, 3.728836467758109*^9, {
   3.72883659767412*^9, 3.72883660969293*^9}, 3.728843345160159*^9, 
   3.728845671791244*^9, 3.733367566743209*^9, 3.792273597860653*^9},
 CellLabel->"Out[330]=",
 CellID->196066505,ExpressionUUID->"0cb62bed-8bba-4e7d-b533-d7877623588b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Differentiating with the parallel derivative of the frame is equivalent to \
differentiating the component fields:\
\>", "MathCaption",
 CellChangeTimes->{{3.728845714749569*^9, 3.728845767242043*^9}},
 CellID->237444473,ExpressionUUID->"a830cae6-a0f0-46d6-8986-f18ab42e36dd"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PDcsys", "[", 
   RowBox[{"-", "b"}], "]"}], "[", "vector", "]"}]], "Input",
 CellChangeTimes->{{3.728836469124797*^9, 3.728836517510007*^9}, {
  3.728836611312229*^9, 3.728836616565943*^9}},
 CellLabel->"In[331]:=",
 CellID->1975108063,ExpressionUUID->"1129d404-c72b-4020-be0d-f655bf6f5a07"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            InterpretationBox[
             StyleBox[
              RowBox[{
               SubscriptBox[
                StyleBox["\[ScriptCapitalD]",
                 FontColor->RGBColor[1., 0.5, 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0.5, 0.]]], GridBox[{
                 {"v", 
                  StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}]}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`PDcsys[{1, -$CellContext`csys}][
              $CellContext`v[{1, $CellContext`csys}]],
             Editable->False],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
              InterpretationBox[
               StyleBox[
                RowBox[{
                  SubscriptBox[
                  "\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1.\
, 0.5, 0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"], 
                  GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                   ColumnSpacings -> 0.05, RowAlignments -> Center]}], 
                AutoSpacing -> False, ShowAutoStyles -> False], 
               $CellContext`PDcsys[{1, -$CellContext`csys}][
                $CellContext`v[{1, $CellContext`csys}]], Editable -> False]}]],
           
           Annotation[#, {1, 
             1} -> $CellContext`PDcsys[{1, -$CellContext`csys}][
              $CellContext`v[{1, $CellContext`csys}]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            InterpretationBox[
             StyleBox[
              RowBox[{
               SubscriptBox[
                StyleBox["\[ScriptCapitalD]",
                 FontColor->RGBColor[1., 0.5, 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0.5, 0.]]], GridBox[{
                 {"v", 
                  StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}]}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`PDcsys[{2, -$CellContext`csys}][
              $CellContext`v[{1, $CellContext`csys}]],
             Editable->False],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
              InterpretationBox[
               StyleBox[
                RowBox[{
                  SubscriptBox[
                  "\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1.\
, 0.5, 0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"], 
                  GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                   ColumnSpacings -> 0.05, RowAlignments -> Center]}], 
                AutoSpacing -> False, ShowAutoStyles -> False], 
               $CellContext`PDcsys[{2, -$CellContext`csys}][
                $CellContext`v[{1, $CellContext`csys}]], Editable -> False]}]],
           
           Annotation[#, {1, 
             2} -> $CellContext`PDcsys[{2, -$CellContext`csys}][
              $CellContext`v[{1, $CellContext`csys}]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            InterpretationBox[
             StyleBox[
              RowBox[{
               SubscriptBox[
                StyleBox["\[ScriptCapitalD]",
                 FontColor->RGBColor[1., 0.5, 0.]], 
                StyleBox["3",
                 FontColor->RGBColor[1., 0.5, 0.]]], GridBox[{
                 {"v", 
                  StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}]}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`PDcsys[{3, -$CellContext`csys}][
              $CellContext`v[{1, $CellContext`csys}]],
             Editable->False],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", 
              InterpretationBox[
               StyleBox[
                RowBox[{
                  SubscriptBox[
                  "\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1.\
, 0.5, 0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"], 
                  GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                   ColumnSpacings -> 0.05, RowAlignments -> Center]}], 
                AutoSpacing -> False, ShowAutoStyles -> False], 
               $CellContext`PDcsys[{3, -$CellContext`csys}][
                $CellContext`v[{1, $CellContext`csys}]], Editable -> False]}]],
           
           Annotation[#, {1, 
             3} -> $CellContext`PDcsys[{3, -$CellContext`csys}][
              $CellContext`v[{1, $CellContext`csys}]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            InterpretationBox[
             StyleBox[
              RowBox[{
               SubscriptBox[
                StyleBox["\[ScriptCapitalD]",
                 FontColor->RGBColor[1., 0.5, 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0.5, 0.]]], GridBox[{
                 {"v", 
                  StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}]}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`PDcsys[{1, -$CellContext`csys}][
              $CellContext`v[{2, $CellContext`csys}]],
             Editable->False],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
              InterpretationBox[
               StyleBox[
                RowBox[{
                  SubscriptBox[
                  "\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1.\
, 0.5, 0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"], 
                  GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                   ColumnSpacings -> 0.05, RowAlignments -> Center]}], 
                AutoSpacing -> False, ShowAutoStyles -> False], 
               $CellContext`PDcsys[{1, -$CellContext`csys}][
                $CellContext`v[{2, $CellContext`csys}]], Editable -> False]}]],
           
           Annotation[#, {2, 
             1} -> $CellContext`PDcsys[{1, -$CellContext`csys}][
              $CellContext`v[{2, $CellContext`csys}]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            InterpretationBox[
             StyleBox[
              RowBox[{
               SubscriptBox[
                StyleBox["\[ScriptCapitalD]",
                 FontColor->RGBColor[1., 0.5, 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0.5, 0.]]], GridBox[{
                 {"v", 
                  StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}]}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`PDcsys[{2, -$CellContext`csys}][
              $CellContext`v[{2, $CellContext`csys}]],
             Editable->False],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              InterpretationBox[
               StyleBox[
                RowBox[{
                  SubscriptBox[
                  "\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1.\
, 0.5, 0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"], 
                  GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                   ColumnSpacings -> 0.05, RowAlignments -> Center]}], 
                AutoSpacing -> False, ShowAutoStyles -> False], 
               $CellContext`PDcsys[{2, -$CellContext`csys}][
                $CellContext`v[{2, $CellContext`csys}]], Editable -> False]}]],
           
           Annotation[#, {2, 
             2} -> $CellContext`PDcsys[{2, -$CellContext`csys}][
              $CellContext`v[{2, $CellContext`csys}]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            InterpretationBox[
             StyleBox[
              RowBox[{
               SubscriptBox[
                StyleBox["\[ScriptCapitalD]",
                 FontColor->RGBColor[1., 0.5, 0.]], 
                StyleBox["3",
                 FontColor->RGBColor[1., 0.5, 0.]]], GridBox[{
                 {"v", 
                  StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}]}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`PDcsys[{3, -$CellContext`csys}][
              $CellContext`v[{2, $CellContext`csys}]],
             Editable->False],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", 
              InterpretationBox[
               StyleBox[
                RowBox[{
                  SubscriptBox[
                  "\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1.\
, 0.5, 0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"], 
                  GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                   ColumnSpacings -> 0.05, RowAlignments -> Center]}], 
                AutoSpacing -> False, ShowAutoStyles -> False], 
               $CellContext`PDcsys[{3, -$CellContext`csys}][
                $CellContext`v[{2, $CellContext`csys}]], Editable -> False]}]],
           
           Annotation[#, {2, 
             3} -> $CellContext`PDcsys[{3, -$CellContext`csys}][
              $CellContext`v[{2, $CellContext`csys}]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            InterpretationBox[
             StyleBox[
              RowBox[{
               SubscriptBox[
                StyleBox["\[ScriptCapitalD]",
                 FontColor->RGBColor[1., 0.5, 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0.5, 0.]]], GridBox[{
                 {"v", 
                  StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}]}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`PDcsys[{1, -$CellContext`csys}][
              $CellContext`v[{3, $CellContext`csys}]],
             Editable->False],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", 
              InterpretationBox[
               StyleBox[
                RowBox[{
                  SubscriptBox[
                  "\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1.\
, 0.5, 0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"], 
                  GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                   ColumnSpacings -> 0.05, RowAlignments -> Center]}], 
                AutoSpacing -> False, ShowAutoStyles -> False], 
               $CellContext`PDcsys[{1, -$CellContext`csys}][
                $CellContext`v[{3, $CellContext`csys}]], Editable -> False]}]],
           
           Annotation[#, {3, 
             1} -> $CellContext`PDcsys[{1, -$CellContext`csys}][
              $CellContext`v[{3, $CellContext`csys}]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            InterpretationBox[
             StyleBox[
              RowBox[{
               SubscriptBox[
                StyleBox["\[ScriptCapitalD]",
                 FontColor->RGBColor[1., 0.5, 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0.5, 0.]]], GridBox[{
                 {"v", 
                  StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}]}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`PDcsys[{2, -$CellContext`csys}][
              $CellContext`v[{3, $CellContext`csys}]],
             Editable->False],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", 
              InterpretationBox[
               StyleBox[
                RowBox[{
                  SubscriptBox[
                  "\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1.\
, 0.5, 0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"], 
                  GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                   ColumnSpacings -> 0.05, RowAlignments -> Center]}], 
                AutoSpacing -> False, ShowAutoStyles -> False], 
               $CellContext`PDcsys[{2, -$CellContext`csys}][
                $CellContext`v[{3, $CellContext`csys}]], Editable -> False]}]],
           
           Annotation[#, {3, 
             2} -> $CellContext`PDcsys[{2, -$CellContext`csys}][
              $CellContext`v[{3, $CellContext`csys}]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            InterpretationBox[
             StyleBox[
              RowBox[{
               SubscriptBox[
                StyleBox["\[ScriptCapitalD]",
                 FontColor->RGBColor[1., 0.5, 0.]], 
                StyleBox["3",
                 FontColor->RGBColor[1., 0.5, 0.]]], GridBox[{
                 {"v", 
                  StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}]}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`PDcsys[{3, -$CellContext`csys}][
              $CellContext`v[{3, $CellContext`csys}]],
             Editable->False],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", 
              InterpretationBox[
               StyleBox[
                RowBox[{
                  SubscriptBox[
                  "\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1.\
, 0.5, 0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"], 
                  GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                   ColumnSpacings -> 0.05, RowAlignments -> Center]}], 
                AutoSpacing -> False, ShowAutoStyles -> False], 
               $CellContext`PDcsys[{3, -$CellContext`csys}][
                $CellContext`v[{3, $CellContext`csys}]], Editable -> False]}]],
           
           Annotation[#, {3, 
             3} -> $CellContext`PDcsys[{3, -$CellContext`csys}][
              $CellContext`v[{3, $CellContext`csys}]], "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0.5, 0], 4 -> RGBColor[1, 0.5, 0]}, 
          "RowsIndexed" -> {
           1 -> RGBColor[1, 0.5, 0], 4 -> RGBColor[1, 0.5, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a", " "},
         {" ", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{
     $CellContext`PDcsys[{1, -$CellContext`csys}][
      $CellContext`v[{1, $CellContext`csys}]], 
     $CellContext`PDcsys[{2, -$CellContext`csys}][
      $CellContext`v[{1, $CellContext`csys}]], 
     $CellContext`PDcsys[{3, -$CellContext`csys}][
      $CellContext`v[{1, $CellContext`csys}]]}, {
     $CellContext`PDcsys[{1, -$CellContext`csys}][
      $CellContext`v[{2, $CellContext`csys}]], 
     $CellContext`PDcsys[{2, -$CellContext`csys}][
      $CellContext`v[{2, $CellContext`csys}]], 
     $CellContext`PDcsys[{3, -$CellContext`csys}][
      $CellContext`v[{2, $CellContext`csys}]]}, {
     $CellContext`PDcsys[{1, -$CellContext`csys}][
      $CellContext`v[{3, $CellContext`csys}]], 
     $CellContext`PDcsys[{2, -$CellContext`csys}][
      $CellContext`v[{3, $CellContext`csys}]], 
     $CellContext`PDcsys[{3, -$CellContext`csys}][
      $CellContext`v[{
       3, $CellContext`csys}]]}}, {$CellContext`csys, -$CellContext`csys}, 
   0][$CellContext`a, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{{3.728836476702044*^9, 3.728836518085503*^9}, 
   3.728836616999979*^9, 3.7288433460184793`*^9, 3.728845671981615*^9, 
   3.733367566949792*^9, 3.792273601873419*^9},
 CellLabel->"Out[331]=",
 CellID->325228131,ExpressionUUID->"ceadf765-97e7-4461-afe0-49789ec3a977"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Using any other derivative, including the parallel derivative of any other \
frame, introduces Christoffel terms:\
\>", "MathCaption",
 CellChangeTimes->{{3.728845770659192*^9, 3.728845795584992*^9}},
 CellID->1221432916,ExpressionUUID->"cd40ec7e-cdcd-43ed-bc8a-a554d3ceeb01"],

Cell[BoxData[
 RowBox[{
  RowBox[{"$LargeComponentSize", "=", "3000"}], ";"}]], "Input",
 CellChangeTimes->{{3.7288366794365673`*^9, 3.7288366873545*^9}},
 CellLabel->"In[332]:=",
 CellID->187134635,ExpressionUUID->"ddf03eb3-189a-4fcb-b242-efa467c9c420"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PDblues", "[", 
   RowBox[{"-", "b"}], "]"}], "[", "vector", "]"}]], "Input",
 CellChangeTimes->{{3.728836469124797*^9, 3.728836517510007*^9}, {
  3.728836550841219*^9, 3.72883655451684*^9}, {3.7288366229275427`*^9, 
  3.7288366259683523`*^9}, {3.728845800997794*^9, 3.728845805650465*^9}},
 CellLabel->"In[333]:=",
 CellID->974452277,ExpressionUUID->"e36cd802-d6fc-476d-8923-fac85c609a25"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"ChristoffelPDbluesPDcsys", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`ChristoffelPDbluesPDcsys[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728843351046611*^9, 3.7288456722990932`*^9, 
  3.728845808023147*^9, 3.7333675672442827`*^9, 3.792273614885078*^9},
 CellLabel->"During evaluation of In[333]:=",
 CellID->948957470,ExpressionUUID->"2a8366f6-e586-4262-bc74-9541ad0acb7e"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
                1, -$CellContext`csys}, {1, -$CellContext`csys}],
               Editable->False], " ", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`v[{1, $CellContext`csys}],
               Editable->False]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
                1, -$CellContext`csys}, {2, -$CellContext`csys}],
               Editable->False], " ", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`v[{2, $CellContext`csys}],
               Editable->False]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
                1, -$CellContext`csys}, {3, -$CellContext`csys}],
               Editable->False], " ", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`v[{3, $CellContext`csys}],
               Editable->False]}], "+", 
             InterpretationBox[
              StyleBox[
               RowBox[{
                SubscriptBox[
                 StyleBox["\[ScriptCapitalD]",
                  FontColor->RGBColor[0., 0., 1.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0.5, 0.]]], GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}]}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`PDblues[{1, -$CellContext`csys}][
               $CellContext`v[{1, $CellContext`csys}]],
              Editable->False]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    1, $CellContext`csys}, {1, -$CellContext`csys}, {
                    1, -$CellContext`csys}], Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`v[{1, $CellContext`csys}], Editable -> 
                   False]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    1, $CellContext`csys}, {1, -$CellContext`csys}, {
                    2, -$CellContext`csys}], Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`v[{2, $CellContext`csys}], Editable -> 
                   False]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    1, $CellContext`csys}, {1, -$CellContext`csys}, {
                    3, -$CellContext`csys}], Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`v[{3, $CellContext`csys}], Editable -> 
                   False]}], "+", 
                InterpretationBox[
                 StyleBox[
                  RowBox[{
                    SubscriptBox[
                    "\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[\
0., 0., 1.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"], 
                    
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center]}], 
                  AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`PDblues[{1, -$CellContext`csys}][
                  $CellContext`v[{1, $CellContext`csys}]], Editable -> 
                 False]}]}]],
           
           Annotation[#, {1, 
             1} -> $CellContext`ChristoffelPDbluesPDcsys[{
                1, $CellContext`csys}, {1, -$CellContext`csys}, {
                1, -$CellContext`csys}] $CellContext`v[{
                1, $CellContext`csys}] + \
$CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
                1, -$CellContext`csys}, {
                2, -$CellContext`csys}] $CellContext`v[{
                2, $CellContext`csys}] + \
$CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
                1, -$CellContext`csys}, {
                3, -$CellContext`csys}] $CellContext`v[{
                3, $CellContext`csys}] + $CellContext`PDblues[{
               1, -$CellContext`csys}][
               $CellContext`v[{1, $CellContext`csys}]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
                2, -$CellContext`csys}, {1, -$CellContext`csys}],
               Editable->False], " ", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`v[{1, $CellContext`csys}],
               Editable->False]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
                2, -$CellContext`csys}, {2, -$CellContext`csys}],
               Editable->False], " ", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`v[{2, $CellContext`csys}],
               Editable->False]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
                2, -$CellContext`csys}, {3, -$CellContext`csys}],
               Editable->False], " ", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`v[{3, $CellContext`csys}],
               Editable->False]}], "+", 
             InterpretationBox[
              StyleBox[
               RowBox[{
                SubscriptBox[
                 StyleBox["\[ScriptCapitalD]",
                  FontColor->RGBColor[0., 0., 1.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0.5, 0.]]], GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}]}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`PDblues[{2, -$CellContext`csys}][
               $CellContext`v[{1, $CellContext`csys}]],
              Editable->False]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    1, $CellContext`csys}, {2, -$CellContext`csys}, {
                    1, -$CellContext`csys}], Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`v[{1, $CellContext`csys}], Editable -> 
                   False]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    1, $CellContext`csys}, {2, -$CellContext`csys}, {
                    2, -$CellContext`csys}], Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`v[{2, $CellContext`csys}], Editable -> 
                   False]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    1, $CellContext`csys}, {2, -$CellContext`csys}, {
                    3, -$CellContext`csys}], Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`v[{3, $CellContext`csys}], Editable -> 
                   False]}], "+", 
                InterpretationBox[
                 StyleBox[
                  RowBox[{
                    SubscriptBox[
                    "\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[\
0., 0., 1.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"], 
                    
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center]}], 
                  AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`PDblues[{2, -$CellContext`csys}][
                  $CellContext`v[{1, $CellContext`csys}]], Editable -> 
                 False]}]}]],
           
           Annotation[#, {1, 
             2} -> $CellContext`ChristoffelPDbluesPDcsys[{
                1, $CellContext`csys}, {2, -$CellContext`csys}, {
                1, -$CellContext`csys}] $CellContext`v[{
                1, $CellContext`csys}] + \
$CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
                2, -$CellContext`csys}, {
                2, -$CellContext`csys}] $CellContext`v[{
                2, $CellContext`csys}] + \
$CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
                2, -$CellContext`csys}, {
                3, -$CellContext`csys}] $CellContext`v[{
                3, $CellContext`csys}] + $CellContext`PDblues[{
               2, -$CellContext`csys}][
               $CellContext`v[{1, $CellContext`csys}]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
                3, -$CellContext`csys}, {1, -$CellContext`csys}],
               Editable->False], " ", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`v[{1, $CellContext`csys}],
               Editable->False]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
                3, -$CellContext`csys}, {2, -$CellContext`csys}],
               Editable->False], " ", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`v[{2, $CellContext`csys}],
               Editable->False]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
                3, -$CellContext`csys}, {3, -$CellContext`csys}],
               Editable->False], " ", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`v[{3, $CellContext`csys}],
               Editable->False]}], "+", 
             InterpretationBox[
              StyleBox[
               RowBox[{
                SubscriptBox[
                 StyleBox["\[ScriptCapitalD]",
                  FontColor->RGBColor[0., 0., 1.]], 
                 StyleBox["3",
                  FontColor->RGBColor[1., 0.5, 0.]]], GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}]}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`PDblues[{3, -$CellContext`csys}][
               $CellContext`v[{1, $CellContext`csys}]],
              Editable->False]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    1, $CellContext`csys}, {3, -$CellContext`csys}, {
                    1, -$CellContext`csys}], Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`v[{1, $CellContext`csys}], Editable -> 
                   False]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    1, $CellContext`csys}, {3, -$CellContext`csys}, {
                    2, -$CellContext`csys}], Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`v[{2, $CellContext`csys}], Editable -> 
                   False]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    1, $CellContext`csys}, {3, -$CellContext`csys}, {
                    3, -$CellContext`csys}], Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`v[{3, $CellContext`csys}], Editable -> 
                   False]}], "+", 
                InterpretationBox[
                 StyleBox[
                  RowBox[{
                    SubscriptBox[
                    "\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[\
0., 0., 1.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"], 
                    
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center]}], 
                  AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`PDblues[{3, -$CellContext`csys}][
                  $CellContext`v[{1, $CellContext`csys}]], Editable -> 
                 False]}]}]],
           
           Annotation[#, {1, 
             3} -> $CellContext`ChristoffelPDbluesPDcsys[{
                1, $CellContext`csys}, {3, -$CellContext`csys}, {
                1, -$CellContext`csys}] $CellContext`v[{
                1, $CellContext`csys}] + \
$CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
                3, -$CellContext`csys}, {
                2, -$CellContext`csys}] $CellContext`v[{
                2, $CellContext`csys}] + \
$CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
                3, -$CellContext`csys}, {
                3, -$CellContext`csys}] $CellContext`v[{
                3, $CellContext`csys}] + $CellContext`PDblues[{
               3, -$CellContext`csys}][
               $CellContext`v[{1, $CellContext`csys}]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
                1, -$CellContext`csys}, {1, -$CellContext`csys}],
               Editable->False], " ", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`v[{1, $CellContext`csys}],
               Editable->False]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
                1, -$CellContext`csys}, {2, -$CellContext`csys}],
               Editable->False], " ", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`v[{2, $CellContext`csys}],
               Editable->False]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
                1, -$CellContext`csys}, {3, -$CellContext`csys}],
               Editable->False], " ", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`v[{3, $CellContext`csys}],
               Editable->False]}], "+", 
             InterpretationBox[
              StyleBox[
               RowBox[{
                SubscriptBox[
                 StyleBox["\[ScriptCapitalD]",
                  FontColor->RGBColor[0., 0., 1.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0.5, 0.]]], GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}]}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`PDblues[{1, -$CellContext`csys}][
               $CellContext`v[{2, $CellContext`csys}]],
              Editable->False]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    2, $CellContext`csys}, {1, -$CellContext`csys}, {
                    1, -$CellContext`csys}], Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`v[{1, $CellContext`csys}], Editable -> 
                   False]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    2, $CellContext`csys}, {1, -$CellContext`csys}, {
                    2, -$CellContext`csys}], Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`v[{2, $CellContext`csys}], Editable -> 
                   False]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    2, $CellContext`csys}, {1, -$CellContext`csys}, {
                    3, -$CellContext`csys}], Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`v[{3, $CellContext`csys}], Editable -> 
                   False]}], "+", 
                InterpretationBox[
                 StyleBox[
                  RowBox[{
                    SubscriptBox[
                    "\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[\
0., 0., 1.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"], 
                    
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center]}], 
                  AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`PDblues[{1, -$CellContext`csys}][
                  $CellContext`v[{2, $CellContext`csys}]], Editable -> 
                 False]}]}]],
           
           Annotation[#, {2, 
             1} -> $CellContext`ChristoffelPDbluesPDcsys[{
                2, $CellContext`csys}, {1, -$CellContext`csys}, {
                1, -$CellContext`csys}] $CellContext`v[{
                1, $CellContext`csys}] + \
$CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
                1, -$CellContext`csys}, {
                2, -$CellContext`csys}] $CellContext`v[{
                2, $CellContext`csys}] + \
$CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
                1, -$CellContext`csys}, {
                3, -$CellContext`csys}] $CellContext`v[{
                3, $CellContext`csys}] + $CellContext`PDblues[{
               1, -$CellContext`csys}][
               $CellContext`v[{2, $CellContext`csys}]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
                2, -$CellContext`csys}, {1, -$CellContext`csys}],
               Editable->False], " ", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`v[{1, $CellContext`csys}],
               Editable->False]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
                2, -$CellContext`csys}, {2, -$CellContext`csys}],
               Editable->False], " ", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`v[{2, $CellContext`csys}],
               Editable->False]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
                2, -$CellContext`csys}, {3, -$CellContext`csys}],
               Editable->False], " ", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`v[{3, $CellContext`csys}],
               Editable->False]}], "+", 
             InterpretationBox[
              StyleBox[
               RowBox[{
                SubscriptBox[
                 StyleBox["\[ScriptCapitalD]",
                  FontColor->RGBColor[0., 0., 1.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0.5, 0.]]], GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}]}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`PDblues[{2, -$CellContext`csys}][
               $CellContext`v[{2, $CellContext`csys}]],
              Editable->False]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    2, $CellContext`csys}, {2, -$CellContext`csys}, {
                    1, -$CellContext`csys}], Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`v[{1, $CellContext`csys}], Editable -> 
                   False]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    2, $CellContext`csys}, {2, -$CellContext`csys}, {
                    2, -$CellContext`csys}], Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`v[{2, $CellContext`csys}], Editable -> 
                   False]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    2, $CellContext`csys}, {2, -$CellContext`csys}, {
                    3, -$CellContext`csys}], Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`v[{3, $CellContext`csys}], Editable -> 
                   False]}], "+", 
                InterpretationBox[
                 StyleBox[
                  RowBox[{
                    SubscriptBox[
                    "\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[\
0., 0., 1.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"], 
                    
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center]}], 
                  AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`PDblues[{2, -$CellContext`csys}][
                  $CellContext`v[{2, $CellContext`csys}]], Editable -> 
                 False]}]}]],
           
           Annotation[#, {2, 
             2} -> $CellContext`ChristoffelPDbluesPDcsys[{
                2, $CellContext`csys}, {2, -$CellContext`csys}, {
                1, -$CellContext`csys}] $CellContext`v[{
                1, $CellContext`csys}] + \
$CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
                2, -$CellContext`csys}, {
                2, -$CellContext`csys}] $CellContext`v[{
                2, $CellContext`csys}] + \
$CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
                2, -$CellContext`csys}, {
                3, -$CellContext`csys}] $CellContext`v[{
                3, $CellContext`csys}] + $CellContext`PDblues[{
               2, -$CellContext`csys}][
               $CellContext`v[{2, $CellContext`csys}]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
                3, -$CellContext`csys}, {1, -$CellContext`csys}],
               Editable->False], " ", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`v[{1, $CellContext`csys}],
               Editable->False]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
                3, -$CellContext`csys}, {2, -$CellContext`csys}],
               Editable->False], " ", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`v[{2, $CellContext`csys}],
               Editable->False]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
                3, -$CellContext`csys}, {3, -$CellContext`csys}],
               Editable->False], " ", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`v[{3, $CellContext`csys}],
               Editable->False]}], "+", 
             InterpretationBox[
              StyleBox[
               RowBox[{
                SubscriptBox[
                 StyleBox["\[ScriptCapitalD]",
                  FontColor->RGBColor[0., 0., 1.]], 
                 StyleBox["3",
                  FontColor->RGBColor[1., 0.5, 0.]]], GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}]}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`PDblues[{3, -$CellContext`csys}][
               $CellContext`v[{2, $CellContext`csys}]],
              Editable->False]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    2, $CellContext`csys}, {3, -$CellContext`csys}, {
                    1, -$CellContext`csys}], Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`v[{1, $CellContext`csys}], Editable -> 
                   False]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    2, $CellContext`csys}, {3, -$CellContext`csys}, {
                    2, -$CellContext`csys}], Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`v[{2, $CellContext`csys}], Editable -> 
                   False]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    2, $CellContext`csys}, {3, -$CellContext`csys}, {
                    3, -$CellContext`csys}], Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`v[{3, $CellContext`csys}], Editable -> 
                   False]}], "+", 
                InterpretationBox[
                 StyleBox[
                  RowBox[{
                    SubscriptBox[
                    "\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[\
0., 0., 1.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"], 
                    
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center]}], 
                  AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`PDblues[{3, -$CellContext`csys}][
                  $CellContext`v[{2, $CellContext`csys}]], Editable -> 
                 False]}]}]],
           
           Annotation[#, {2, 
             3} -> $CellContext`ChristoffelPDbluesPDcsys[{
                2, $CellContext`csys}, {3, -$CellContext`csys}, {
                1, -$CellContext`csys}] $CellContext`v[{
                1, $CellContext`csys}] + \
$CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
                3, -$CellContext`csys}, {
                2, -$CellContext`csys}] $CellContext`v[{
                2, $CellContext`csys}] + \
$CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
                3, -$CellContext`csys}, {
                3, -$CellContext`csys}] $CellContext`v[{
                3, $CellContext`csys}] + $CellContext`PDblues[{
               3, -$CellContext`csys}][
               $CellContext`v[{2, $CellContext`csys}]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
                1, -$CellContext`csys}, {1, -$CellContext`csys}],
               Editable->False], " ", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`v[{1, $CellContext`csys}],
               Editable->False]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
                1, -$CellContext`csys}, {2, -$CellContext`csys}],
               Editable->False], " ", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`v[{2, $CellContext`csys}],
               Editable->False]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
                1, -$CellContext`csys}, {3, -$CellContext`csys}],
               Editable->False], " ", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`v[{3, $CellContext`csys}],
               Editable->False]}], "+", 
             InterpretationBox[
              StyleBox[
               RowBox[{
                SubscriptBox[
                 StyleBox["\[ScriptCapitalD]",
                  FontColor->RGBColor[0., 0., 1.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0.5, 0.]]], GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}]}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`PDblues[{1, -$CellContext`csys}][
               $CellContext`v[{3, $CellContext`csys}]],
              Editable->False]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    3, $CellContext`csys}, {1, -$CellContext`csys}, {
                    1, -$CellContext`csys}], Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`v[{1, $CellContext`csys}], Editable -> 
                   False]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    3, $CellContext`csys}, {1, -$CellContext`csys}, {
                    2, -$CellContext`csys}], Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`v[{2, $CellContext`csys}], Editable -> 
                   False]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    3, $CellContext`csys}, {1, -$CellContext`csys}, {
                    3, -$CellContext`csys}], Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`v[{3, $CellContext`csys}], Editable -> 
                   False]}], "+", 
                InterpretationBox[
                 StyleBox[
                  RowBox[{
                    SubscriptBox[
                    "\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[\
0., 0., 1.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"], 
                    
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center]}], 
                  AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`PDblues[{1, -$CellContext`csys}][
                  $CellContext`v[{3, $CellContext`csys}]], Editable -> 
                 False]}]}]],
           
           Annotation[#, {3, 
             1} -> $CellContext`ChristoffelPDbluesPDcsys[{
                3, $CellContext`csys}, {1, -$CellContext`csys}, {
                1, -$CellContext`csys}] $CellContext`v[{
                1, $CellContext`csys}] + \
$CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
                1, -$CellContext`csys}, {
                2, -$CellContext`csys}] $CellContext`v[{
                2, $CellContext`csys}] + \
$CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
                1, -$CellContext`csys}, {
                3, -$CellContext`csys}] $CellContext`v[{
                3, $CellContext`csys}] + $CellContext`PDblues[{
               1, -$CellContext`csys}][
               $CellContext`v[{3, $CellContext`csys}]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
                2, -$CellContext`csys}, {1, -$CellContext`csys}],
               Editable->False], " ", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`v[{1, $CellContext`csys}],
               Editable->False]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
                2, -$CellContext`csys}, {2, -$CellContext`csys}],
               Editable->False], " ", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`v[{2, $CellContext`csys}],
               Editable->False]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
                2, -$CellContext`csys}, {3, -$CellContext`csys}],
               Editable->False], " ", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`v[{3, $CellContext`csys}],
               Editable->False]}], "+", 
             InterpretationBox[
              StyleBox[
               RowBox[{
                SubscriptBox[
                 StyleBox["\[ScriptCapitalD]",
                  FontColor->RGBColor[0., 0., 1.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0.5, 0.]]], GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}]}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`PDblues[{2, -$CellContext`csys}][
               $CellContext`v[{3, $CellContext`csys}]],
              Editable->False]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    3, $CellContext`csys}, {2, -$CellContext`csys}, {
                    1, -$CellContext`csys}], Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`v[{1, $CellContext`csys}], Editable -> 
                   False]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    3, $CellContext`csys}, {2, -$CellContext`csys}, {
                    2, -$CellContext`csys}], Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`v[{2, $CellContext`csys}], Editable -> 
                   False]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    3, $CellContext`csys}, {2, -$CellContext`csys}, {
                    3, -$CellContext`csys}], Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`v[{3, $CellContext`csys}], Editable -> 
                   False]}], "+", 
                InterpretationBox[
                 StyleBox[
                  RowBox[{
                    SubscriptBox[
                    "\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[\
0., 0., 1.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"], 
                    
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center]}], 
                  AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`PDblues[{2, -$CellContext`csys}][
                  $CellContext`v[{3, $CellContext`csys}]], Editable -> 
                 False]}]}]],
           
           Annotation[#, {3, 
             2} -> $CellContext`ChristoffelPDbluesPDcsys[{
                3, $CellContext`csys}, {2, -$CellContext`csys}, {
                1, -$CellContext`csys}] $CellContext`v[{
                1, $CellContext`csys}] + \
$CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
                2, -$CellContext`csys}, {
                2, -$CellContext`csys}] $CellContext`v[{
                2, $CellContext`csys}] + \
$CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
                2, -$CellContext`csys}, {
                3, -$CellContext`csys}] $CellContext`v[{
                3, $CellContext`csys}] + $CellContext`PDblues[{
               2, -$CellContext`csys}][
               $CellContext`v[{3, $CellContext`csys}]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
                3, -$CellContext`csys}, {1, -$CellContext`csys}],
               Editable->False], " ", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`v[{1, $CellContext`csys}],
               Editable->False]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
                3, -$CellContext`csys}, {2, -$CellContext`csys}],
               Editable->False], " ", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`v[{2, $CellContext`csys}],
               Editable->False]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
                3, -$CellContext`csys}, {3, -$CellContext`csys}],
               Editable->False], " ", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`v[{3, $CellContext`csys}],
               Editable->False]}], "+", 
             InterpretationBox[
              StyleBox[
               RowBox[{
                SubscriptBox[
                 StyleBox["\[ScriptCapitalD]",
                  FontColor->RGBColor[0., 0., 1.]], 
                 StyleBox["3",
                  FontColor->RGBColor[1., 0.5, 0.]]], GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}]}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`PDblues[{3, -$CellContext`csys}][
               $CellContext`v[{3, $CellContext`csys}]],
              Editable->False]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    3, $CellContext`csys}, {3, -$CellContext`csys}, {
                    1, -$CellContext`csys}], Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`v[{1, $CellContext`csys}], Editable -> 
                   False]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    3, $CellContext`csys}, {3, -$CellContext`csys}, {
                    2, -$CellContext`csys}], Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`v[{2, $CellContext`csys}], Editable -> 
                   False]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    3, $CellContext`csys}, {3, -$CellContext`csys}, {
                    3, -$CellContext`csys}], Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`v[{3, $CellContext`csys}], Editable -> 
                   False]}], "+", 
                InterpretationBox[
                 StyleBox[
                  RowBox[{
                    SubscriptBox[
                    "\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[\
0., 0., 1.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"], 
                    
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center]}], 
                  AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`PDblues[{3, -$CellContext`csys}][
                  $CellContext`v[{3, $CellContext`csys}]], Editable -> 
                 False]}]}]],
           
           Annotation[#, {3, 
             3} -> $CellContext`ChristoffelPDbluesPDcsys[{
                3, $CellContext`csys}, {3, -$CellContext`csys}, {
                1, -$CellContext`csys}] $CellContext`v[{
                1, $CellContext`csys}] + \
$CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
                3, -$CellContext`csys}, {
                2, -$CellContext`csys}] $CellContext`v[{
                2, $CellContext`csys}] + \
$CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
                3, -$CellContext`csys}, {
                3, -$CellContext`csys}] $CellContext`v[{
                3, $CellContext`csys}] + $CellContext`PDblues[{
               3, -$CellContext`csys}][
               $CellContext`v[{3, $CellContext`csys}]], "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0.5, 0], 4 -> RGBColor[1, 0.5, 0]}, 
          "RowsIndexed" -> {
           1 -> RGBColor[1, 0.5, 0], 4 -> RGBColor[1, 0.5, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a", " "},
         {" ", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{$CellContext`ChristoffelPDbluesPDcsys[{
        1, $CellContext`csys}, {1, -$CellContext`csys}, {
        1, -$CellContext`csys}] $CellContext`v[{
        1, $CellContext`csys}] + $CellContext`ChristoffelPDbluesPDcsys[{
        1, $CellContext`csys}, {1, -$CellContext`csys}, {
        2, -$CellContext`csys}] $CellContext`v[{
        2, $CellContext`csys}] + $CellContext`ChristoffelPDbluesPDcsys[{
        1, $CellContext`csys}, {1, -$CellContext`csys}, {
        3, -$CellContext`csys}] $CellContext`v[{
        3, $CellContext`csys}] + $CellContext`PDblues[{1, -$CellContext`csys}][
       $CellContext`v[{
        1, $CellContext`csys}]], $CellContext`ChristoffelPDbluesPDcsys[{
        1, $CellContext`csys}, {2, -$CellContext`csys}, {
        1, -$CellContext`csys}] $CellContext`v[{
        1, $CellContext`csys}] + $CellContext`ChristoffelPDbluesPDcsys[{
        1, $CellContext`csys}, {2, -$CellContext`csys}, {
        2, -$CellContext`csys}] $CellContext`v[{
        2, $CellContext`csys}] + $CellContext`ChristoffelPDbluesPDcsys[{
        1, $CellContext`csys}, {2, -$CellContext`csys}, {
        3, -$CellContext`csys}] $CellContext`v[{
        3, $CellContext`csys}] + $CellContext`PDblues[{2, -$CellContext`csys}][
       $CellContext`v[{
        1, $CellContext`csys}]], $CellContext`ChristoffelPDbluesPDcsys[{
        1, $CellContext`csys}, {3, -$CellContext`csys}, {
        1, -$CellContext`csys}] $CellContext`v[{
        1, $CellContext`csys}] + $CellContext`ChristoffelPDbluesPDcsys[{
        1, $CellContext`csys}, {3, -$CellContext`csys}, {
        2, -$CellContext`csys}] $CellContext`v[{
        2, $CellContext`csys}] + $CellContext`ChristoffelPDbluesPDcsys[{
        1, $CellContext`csys}, {3, -$CellContext`csys}, {
        3, -$CellContext`csys}] $CellContext`v[{
        3, $CellContext`csys}] + $CellContext`PDblues[{3, -$CellContext`csys}][
       $CellContext`v[{
        1, $CellContext`csys}]]}, {$CellContext`ChristoffelPDbluesPDcsys[{
        2, $CellContext`csys}, {1, -$CellContext`csys}, {
        1, -$CellContext`csys}] $CellContext`v[{
        1, $CellContext`csys}] + $CellContext`ChristoffelPDbluesPDcsys[{
        2, $CellContext`csys}, {1, -$CellContext`csys}, {
        2, -$CellContext`csys}] $CellContext`v[{
        2, $CellContext`csys}] + $CellContext`ChristoffelPDbluesPDcsys[{
        2, $CellContext`csys}, {1, -$CellContext`csys}, {
        3, -$CellContext`csys}] $CellContext`v[{
        3, $CellContext`csys}] + $CellContext`PDblues[{1, -$CellContext`csys}][
       $CellContext`v[{
        2, $CellContext`csys}]], $CellContext`ChristoffelPDbluesPDcsys[{
        2, $CellContext`csys}, {2, -$CellContext`csys}, {
        1, -$CellContext`csys}] $CellContext`v[{
        1, $CellContext`csys}] + $CellContext`ChristoffelPDbluesPDcsys[{
        2, $CellContext`csys}, {2, -$CellContext`csys}, {
        2, -$CellContext`csys}] $CellContext`v[{
        2, $CellContext`csys}] + $CellContext`ChristoffelPDbluesPDcsys[{
        2, $CellContext`csys}, {2, -$CellContext`csys}, {
        3, -$CellContext`csys}] $CellContext`v[{
        3, $CellContext`csys}] + $CellContext`PDblues[{2, -$CellContext`csys}][
       $CellContext`v[{
        2, $CellContext`csys}]], $CellContext`ChristoffelPDbluesPDcsys[{
        2, $CellContext`csys}, {3, -$CellContext`csys}, {
        1, -$CellContext`csys}] $CellContext`v[{
        1, $CellContext`csys}] + $CellContext`ChristoffelPDbluesPDcsys[{
        2, $CellContext`csys}, {3, -$CellContext`csys}, {
        2, -$CellContext`csys}] $CellContext`v[{
        2, $CellContext`csys}] + $CellContext`ChristoffelPDbluesPDcsys[{
        2, $CellContext`csys}, {3, -$CellContext`csys}, {
        3, -$CellContext`csys}] $CellContext`v[{
        3, $CellContext`csys}] + $CellContext`PDblues[{3, -$CellContext`csys}][
       $CellContext`v[{
        2, $CellContext`csys}]]}, {$CellContext`ChristoffelPDbluesPDcsys[{
        3, $CellContext`csys}, {1, -$CellContext`csys}, {
        1, -$CellContext`csys}] $CellContext`v[{
        1, $CellContext`csys}] + $CellContext`ChristoffelPDbluesPDcsys[{
        3, $CellContext`csys}, {1, -$CellContext`csys}, {
        2, -$CellContext`csys}] $CellContext`v[{
        2, $CellContext`csys}] + $CellContext`ChristoffelPDbluesPDcsys[{
        3, $CellContext`csys}, {1, -$CellContext`csys}, {
        3, -$CellContext`csys}] $CellContext`v[{
        3, $CellContext`csys}] + $CellContext`PDblues[{1, -$CellContext`csys}][
       $CellContext`v[{
        3, $CellContext`csys}]], $CellContext`ChristoffelPDbluesPDcsys[{
        3, $CellContext`csys}, {2, -$CellContext`csys}, {
        1, -$CellContext`csys}] $CellContext`v[{
        1, $CellContext`csys}] + $CellContext`ChristoffelPDbluesPDcsys[{
        3, $CellContext`csys}, {2, -$CellContext`csys}, {
        2, -$CellContext`csys}] $CellContext`v[{
        2, $CellContext`csys}] + $CellContext`ChristoffelPDbluesPDcsys[{
        3, $CellContext`csys}, {2, -$CellContext`csys}, {
        3, -$CellContext`csys}] $CellContext`v[{
        3, $CellContext`csys}] + $CellContext`PDblues[{2, -$CellContext`csys}][
       $CellContext`v[{
        3, $CellContext`csys}]], $CellContext`ChristoffelPDbluesPDcsys[{
        3, $CellContext`csys}, {3, -$CellContext`csys}, {
        1, -$CellContext`csys}] $CellContext`v[{
        1, $CellContext`csys}] + $CellContext`ChristoffelPDbluesPDcsys[{
        3, $CellContext`csys}, {3, -$CellContext`csys}, {
        2, -$CellContext`csys}] $CellContext`v[{
        2, $CellContext`csys}] + $CellContext`ChristoffelPDbluesPDcsys[{
        3, $CellContext`csys}, {3, -$CellContext`csys}, {
        3, -$CellContext`csys}] $CellContext`v[{
        3, $CellContext`csys}] + $CellContext`PDblues[{3, -$CellContext`csys}][
       $CellContext`v[{
        3, $CellContext`csys}]]}}, {$CellContext`csys, -$CellContext`csys}, 
   0][$CellContext`a, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{{3.728836476702044*^9, 3.728836518085503*^9}, 
   3.728836556285392*^9, 3.728836626607774*^9, {3.728836684706974*^9, 
   3.728836688460067*^9}, 3.728836742646254*^9, 3.72884335122005*^9, 
   3.7288456724656887`*^9, 3.7288458082167597`*^9, 3.7333675673629007`*^9, 
   3.792273615009704*^9},
 CellLabel->"Out[333]=",
 CellID->1914635668,ExpressionUUID->"1bbf0498-83e7-47d6-ba42-041db35cbec4"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"$LargeComponentSize", "=", "1000"}], ";"}]], "Input",
 CellChangeTimes->{{3.728836696779812*^9, 3.7288366969461946`*^9}},
 CellLabel->"In[334]:=",
 CellID->707327288,ExpressionUUID->"7257562e-d25c-4c8d-b2b3-ac9b30e6d313"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We can reproduce the same computations using InChart scalar functions:\
\>", "MathCaption",
 CellChangeTimes->{{3.72884586849328*^9, 3.728845869380991*^9}, {
  3.728845909820978*^9, 3.728845921387244*^9}},
 CellID->1891731754,ExpressionUUID->"b7842a20-c47e-4c2e-a794-033c358db9df"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ivector", "=", 
  RowBox[{"vector", "//", 
   RowBox[{"ToInChart", "[", "csys", "]"}]}]}]], "Input",
 CellChangeTimes->{{3.728845872875189*^9, 3.728845888444456*^9}},
 CellLabel->"In[335]:=",
 CellID->642506035,ExpressionUUID->"108353b5-ea70-46cd-95b7-6ceb2ec6efde"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{
             InterpretationBox[
              StyleBox[
               UnderscriptBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                StyleBox["_",
                 FontColor->RGBColor[1, 0.5, 0]]],
               ShowAutoStyles->False,
               AutoSpacing->False],
              xAct`xCoba`InChart[
               $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", 
              RowBox[{
                InterpretationBox[
                 StyleBox[
                  UnderscriptBox[
                   GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                   "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                  AutoSpacing -> False, ShowAutoStyles -> False], 
                 xAct`xCoba`InChart[
                  $CellContext`v[{1, $CellContext`csys}], $CellContext`csys], 
                 Editable -> False], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`x[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`y[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`z[], Editable -> False]}], "]"}]}]],
           Annotation[#, {1} -> xAct`xCoba`InChart[
              $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             InterpretationBox[
              StyleBox[
               UnderscriptBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                StyleBox["_",
                 FontColor->RGBColor[1, 0.5, 0]]],
               ShowAutoStyles->False,
               AutoSpacing->False],
              xAct`xCoba`InChart[
               $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              RowBox[{
                InterpretationBox[
                 StyleBox[
                  UnderscriptBox[
                   GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                   "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                  AutoSpacing -> False, ShowAutoStyles -> False], 
                 xAct`xCoba`InChart[
                  $CellContext`v[{2, $CellContext`csys}], $CellContext`csys], 
                 Editable -> False], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`x[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`y[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`z[], Editable -> False]}], "]"}]}]],
           Annotation[#, {2} -> xAct`xCoba`InChart[
              $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             InterpretationBox[
              StyleBox[
               UnderscriptBox[GridBox[{
                  {"v", 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                StyleBox["_",
                 FontColor->RGBColor[1, 0.5, 0]]],
               ShowAutoStyles->False,
               AutoSpacing->False],
              xAct`xCoba`InChart[
               $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", 
              RowBox[{
                InterpretationBox[
                 StyleBox[
                  UnderscriptBox[
                   GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                   "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                  AutoSpacing -> False, ShowAutoStyles -> False], 
                 xAct`xCoba`InChart[
                  $CellContext`v[{3, $CellContext`csys}], $CellContext`csys], 
                 Editable -> False], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`x[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`y[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`z[], Editable -> False]}], "]"}]}]],
           Annotation[#, {3} -> xAct`xCoba`InChart[
              $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]], "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0.5, 0], 2 -> RGBColor[1, 0.5, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{
    xAct`xCoba`InChart[
     $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
     $CellContext`x[], 
     $CellContext`y[], 
     $CellContext`z[]], 
    xAct`xCoba`InChart[
     $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
     $CellContext`x[], 
     $CellContext`y[], 
     $CellContext`z[]], 
    xAct`xCoba`InChart[
     $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
     $CellContext`x[], 
     $CellContext`y[], 
     $CellContext`z[]]}, {$CellContext`csys}, 0][$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{{3.7288458811540327`*^9, 3.7288458891181097`*^9}, 
   3.7288460102251883`*^9, 3.733367567710024*^9, 3.7922736191970873`*^9},
 CellLabel->"Out[335]=",
 CellID->666835233,ExpressionUUID->"2e95a346-7095-4948-96f3-833d2d797722"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PDcsys", "[", 
   RowBox[{"-", "b"}], "]"}], "[", "ivector", "]"}]], "Input",
 CellChangeTimes->{{3.72884589094239*^9, 3.7288459008357677`*^9}},
 CellLabel->"In[336]:=",
 CellID->272106681,ExpressionUUID->"f041a4f8-6ae6-47fe-b39d-f06b36fbb170"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox[
                UnderscriptBox[GridBox[{
                   {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                  },
                  
                  GridBoxAlignment->{
                   "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                 StyleBox["_",
                  FontColor->RGBColor[1, 0.5, 0]]],
                ShowAutoStyles->False,
                AutoSpacing->False],
               xAct`xCoba`InChart[
                $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
               Editable->False], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"1", ",", "0", ",", "0"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                   UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                   AutoSpacing -> False, ShowAutoStyles -> False], 
                  xAct`xCoba`InChart[
                   $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
                   Editable -> False], 
                 TagBox[
                  RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                 MultilineFunction -> None], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`x[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`y[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`z[], Editable -> False]}], "]"}]}]],
           Annotation[#, {1, 1} -> Derivative[1, 0, 0][
              xAct`xCoba`InChart[
               $CellContext`v[{1, $CellContext`csys}], $CellContext`csys]][
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox[
                UnderscriptBox[GridBox[{
                   {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                  },
                  
                  GridBoxAlignment->{
                   "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                 StyleBox["_",
                  FontColor->RGBColor[1, 0.5, 0]]],
                ShowAutoStyles->False,
                AutoSpacing->False],
               xAct`xCoba`InChart[
                $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
               Editable->False], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"0", ",", "1", ",", "0"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                   UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                   AutoSpacing -> False, ShowAutoStyles -> False], 
                  xAct`xCoba`InChart[
                   $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
                   Editable -> False], 
                 TagBox[
                  RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                 MultilineFunction -> None], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`x[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`y[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`z[], Editable -> False]}], "]"}]}]],
           Annotation[#, {1, 2} -> Derivative[0, 1, 0][
              xAct`xCoba`InChart[
               $CellContext`v[{1, $CellContext`csys}], $CellContext`csys]][
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox[
                UnderscriptBox[GridBox[{
                   {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                  },
                  
                  GridBoxAlignment->{
                   "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                 StyleBox["_",
                  FontColor->RGBColor[1, 0.5, 0]]],
                ShowAutoStyles->False,
                AutoSpacing->False],
               xAct`xCoba`InChart[
                $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
               Editable->False], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"0", ",", "0", ",", "1"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", 
              RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                   UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                   AutoSpacing -> False, ShowAutoStyles -> False], 
                  xAct`xCoba`InChart[
                   $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
                   Editable -> False], 
                 TagBox[
                  RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                 MultilineFunction -> None], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`x[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`y[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`z[], Editable -> False]}], "]"}]}]],
           Annotation[#, {1, 3} -> Derivative[0, 0, 1][
              xAct`xCoba`InChart[
               $CellContext`v[{1, $CellContext`csys}], $CellContext`csys]][
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox[
                UnderscriptBox[GridBox[{
                   {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                  },
                  GridBoxAlignment->{
                   "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                 StyleBox["_",
                  FontColor->RGBColor[1, 0.5, 0]]],
                ShowAutoStyles->False,
                AutoSpacing->False],
               xAct`xCoba`InChart[
                $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
               Editable->False], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"1", ",", "0", ",", "0"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                   UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                   AutoSpacing -> False, ShowAutoStyles -> False], 
                  xAct`xCoba`InChart[
                   $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
                   Editable -> False], 
                 TagBox[
                  RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                 MultilineFunction -> None], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`x[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`y[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`z[], Editable -> False]}], "]"}]}]],
           Annotation[#, {2, 1} -> Derivative[1, 0, 0][
              xAct`xCoba`InChart[
               $CellContext`v[{2, $CellContext`csys}], $CellContext`csys]][
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox[
                UnderscriptBox[GridBox[{
                   {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                  },
                  
                  GridBoxAlignment->{
                   "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                 StyleBox["_",
                  FontColor->RGBColor[1, 0.5, 0]]],
                ShowAutoStyles->False,
                AutoSpacing->False],
               xAct`xCoba`InChart[
                $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
               Editable->False], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"0", ",", "1", ",", "0"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                   UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                   AutoSpacing -> False, ShowAutoStyles -> False], 
                  xAct`xCoba`InChart[
                   $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
                   Editable -> False], 
                 TagBox[
                  RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                 MultilineFunction -> None], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`x[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`y[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`z[], Editable -> False]}], "]"}]}]],
           Annotation[#, {2, 2} -> Derivative[0, 1, 0][
              xAct`xCoba`InChart[
               $CellContext`v[{2, $CellContext`csys}], $CellContext`csys]][
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox[
                UnderscriptBox[GridBox[{
                   {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                  },
                  
                  GridBoxAlignment->{
                   "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                 StyleBox["_",
                  FontColor->RGBColor[1, 0.5, 0]]],
                ShowAutoStyles->False,
                AutoSpacing->False],
               xAct`xCoba`InChart[
                $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
               Editable->False], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"0", ",", "0", ",", "1"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", 
              RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                   UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                   AutoSpacing -> False, ShowAutoStyles -> False], 
                  xAct`xCoba`InChart[
                   $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
                   Editable -> False], 
                 TagBox[
                  RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                 MultilineFunction -> None], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`x[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`y[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`z[], Editable -> False]}], "]"}]}]],
           Annotation[#, {2, 3} -> Derivative[0, 0, 1][
              xAct`xCoba`InChart[
               $CellContext`v[{2, $CellContext`csys}], $CellContext`csys]][
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox[
                UnderscriptBox[GridBox[{
                   {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                  },
                  
                  GridBoxAlignment->{
                   "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                 StyleBox["_",
                  FontColor->RGBColor[1, 0.5, 0]]],
                ShowAutoStyles->False,
                AutoSpacing->False],
               xAct`xCoba`InChart[
                $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
               Editable->False], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"1", ",", "0", ",", "0"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                   UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                   AutoSpacing -> False, ShowAutoStyles -> False], 
                  xAct`xCoba`InChart[
                   $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
                   Editable -> False], 
                 TagBox[
                  RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                 MultilineFunction -> None], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`x[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`y[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`z[], Editable -> False]}], "]"}]}]],
           Annotation[#, {3, 1} -> Derivative[1, 0, 0][
              xAct`xCoba`InChart[
               $CellContext`v[{3, $CellContext`csys}], $CellContext`csys]][
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox[
                UnderscriptBox[GridBox[{
                   {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                  },
                  
                  GridBoxAlignment->{
                   "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                 StyleBox["_",
                  FontColor->RGBColor[1, 0.5, 0]]],
                ShowAutoStyles->False,
                AutoSpacing->False],
               xAct`xCoba`InChart[
                $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
               Editable->False], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"0", ",", "1", ",", "0"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                   UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                   AutoSpacing -> False, ShowAutoStyles -> False], 
                  xAct`xCoba`InChart[
                   $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
                   Editable -> False], 
                 TagBox[
                  RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                 MultilineFunction -> None], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`x[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`y[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`z[], Editable -> False]}], "]"}]}]],
           Annotation[#, {3, 2} -> Derivative[0, 1, 0][
              xAct`xCoba`InChart[
               $CellContext`v[{3, $CellContext`csys}], $CellContext`csys]][
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox[
                UnderscriptBox[GridBox[{
                   {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                  },
                  
                  GridBoxAlignment->{
                   "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                 StyleBox["_",
                  FontColor->RGBColor[1, 0.5, 0]]],
                ShowAutoStyles->False,
                AutoSpacing->False],
               xAct`xCoba`InChart[
                $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
               Editable->False], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"0", ",", "0", ",", "1"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", 
              RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                   UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                   AutoSpacing -> False, ShowAutoStyles -> False], 
                  xAct`xCoba`InChart[
                   $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
                   Editable -> False], 
                 TagBox[
                  RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                 MultilineFunction -> None], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`x[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`y[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`z[], Editable -> False]}], "]"}]}]],
           Annotation[#, {3, 3} -> Derivative[0, 0, 1][
              xAct`xCoba`InChart[
               $CellContext`v[{3, $CellContext`csys}], $CellContext`csys]][
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]], "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0.5, 0], 4 -> RGBColor[1, 0.5, 0]}, 
          "RowsIndexed" -> {
           1 -> RGBColor[1, 0.5, 0], 4 -> RGBColor[1, 0.5, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a", " "},
         {" ", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{
     Derivative[1, 0, 0][
      xAct`xCoba`InChart[
       $CellContext`v[{1, $CellContext`csys}], $CellContext`csys]][
      $CellContext`x[], 
      $CellContext`y[], 
      $CellContext`z[]], 
     Derivative[0, 1, 0][
      xAct`xCoba`InChart[
       $CellContext`v[{1, $CellContext`csys}], $CellContext`csys]][
      $CellContext`x[], 
      $CellContext`y[], 
      $CellContext`z[]], 
     Derivative[0, 0, 1][
      xAct`xCoba`InChart[
       $CellContext`v[{1, $CellContext`csys}], $CellContext`csys]][
      $CellContext`x[], 
      $CellContext`y[], 
      $CellContext`z[]]}, {
     Derivative[1, 0, 0][
      xAct`xCoba`InChart[
       $CellContext`v[{2, $CellContext`csys}], $CellContext`csys]][
      $CellContext`x[], 
      $CellContext`y[], 
      $CellContext`z[]], 
     Derivative[0, 1, 0][
      xAct`xCoba`InChart[
       $CellContext`v[{2, $CellContext`csys}], $CellContext`csys]][
      $CellContext`x[], 
      $CellContext`y[], 
      $CellContext`z[]], 
     Derivative[0, 0, 1][
      xAct`xCoba`InChart[
       $CellContext`v[{2, $CellContext`csys}], $CellContext`csys]][
      $CellContext`x[], 
      $CellContext`y[], 
      $CellContext`z[]]}, {
     Derivative[1, 0, 0][
      xAct`xCoba`InChart[
       $CellContext`v[{3, $CellContext`csys}], $CellContext`csys]][
      $CellContext`x[], 
      $CellContext`y[], 
      $CellContext`z[]], 
     Derivative[0, 1, 0][
      xAct`xCoba`InChart[
       $CellContext`v[{3, $CellContext`csys}], $CellContext`csys]][
      $CellContext`x[], 
      $CellContext`y[], 
      $CellContext`z[]], 
     Derivative[0, 0, 1][
      xAct`xCoba`InChart[
       $CellContext`v[{3, $CellContext`csys}], $CellContext`csys]][
      $CellContext`x[], 
      $CellContext`y[], 
      $CellContext`z[]]}}, {$CellContext`csys, -$CellContext`csys}, 
   0][$CellContext`a, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.7288459015491343`*^9, 3.728846011067089*^9, 
  3.733367568024934*^9, 3.792273620011808*^9},
 CellLabel->"Out[336]=",
 CellID->1920945982,ExpressionUUID->"57eb0879-8175-4bf1-92e0-b58c233809ea"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"$LargeComponentSize", "=", "4000"}], ";"}]], "Input",
 CellChangeTimes->{{3.72884595257054*^9, 3.728845956825097*^9}},
 CellLabel->"In[337]:=",
 CellID->1092795901,ExpressionUUID->"100a758a-a268-4ea6-9907-d6fca04654e0"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PDblues", "[", 
   RowBox[{"-", "b"}], "]"}], "[", "ivector", "]"}]], "Input",
 CellChangeTimes->{{3.728845935288577*^9, 3.728845941225828*^9}},
 CellLabel->"In[338]:=",
 CellID->408925064,ExpressionUUID->"347ee434-edd7-457e-8e13-f90f11524c9f"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
                1, -$CellContext`csys}, {1, -$CellContext`csys}],
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
                1, -$CellContext`csys}, {2, -$CellContext`csys}],
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
                1, -$CellContext`csys}, {3, -$CellContext`csys}],
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"1", ",", "0", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    1, $CellContext`csys}, {1, -$CellContext`csys}, {
                    1, -$CellContext`csys}], Editable -> False], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    1, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    1, $CellContext`csys}, {1, -$CellContext`csys}, {
                    2, -$CellContext`csys}], Editable -> False], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    2, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    1, $CellContext`csys}, {1, -$CellContext`csys}, {
                    3, -$CellContext`csys}], Editable -> False], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    3, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    1, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                   MultilineFunction -> None], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}]],
           
           Annotation[#, {1, 
             1} -> $CellContext`ChristoffelPDbluesPDcsys[{
                1, $CellContext`csys}, {1, -$CellContext`csys}, {
                1, -$CellContext`csys}] xAct`xCoba`InChart[
                $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
                1, $CellContext`csys}, {1, -$CellContext`csys}, {
                2, -$CellContext`csys}] xAct`xCoba`InChart[
                $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
                1, $CellContext`csys}, {1, -$CellContext`csys}, {
                3, -$CellContext`csys}] xAct`xCoba`InChart[
                $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + Derivative[1, 0, 0][
               xAct`xCoba`InChart[
                $CellContext`v[{1, $CellContext`csys}], $CellContext`csys]][
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
                2, -$CellContext`csys}, {1, -$CellContext`csys}],
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
                2, -$CellContext`csys}, {2, -$CellContext`csys}],
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
                2, -$CellContext`csys}, {3, -$CellContext`csys}],
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"0", ",", "1", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    1, $CellContext`csys}, {2, -$CellContext`csys}, {
                    1, -$CellContext`csys}], Editable -> False], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    1, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    1, $CellContext`csys}, {2, -$CellContext`csys}, {
                    2, -$CellContext`csys}], Editable -> False], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    2, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    1, $CellContext`csys}, {2, -$CellContext`csys}, {
                    3, -$CellContext`csys}], Editable -> False], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    3, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    1, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                   MultilineFunction -> None], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}]],
           
           Annotation[#, {1, 
             2} -> $CellContext`ChristoffelPDbluesPDcsys[{
                1, $CellContext`csys}, {2, -$CellContext`csys}, {
                1, -$CellContext`csys}] xAct`xCoba`InChart[
                $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
                1, $CellContext`csys}, {2, -$CellContext`csys}, {
                2, -$CellContext`csys}] xAct`xCoba`InChart[
                $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
                1, $CellContext`csys}, {2, -$CellContext`csys}, {
                3, -$CellContext`csys}] xAct`xCoba`InChart[
                $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + Derivative[0, 1, 0][
               xAct`xCoba`InChart[
                $CellContext`v[{1, $CellContext`csys}], $CellContext`csys]][
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
                3, -$CellContext`csys}, {1, -$CellContext`csys}],
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
                3, -$CellContext`csys}, {2, -$CellContext`csys}],
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
                3, -$CellContext`csys}, {3, -$CellContext`csys}],
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"0", ",", "0", ",", "1"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    1, $CellContext`csys}, {3, -$CellContext`csys}, {
                    1, -$CellContext`csys}], Editable -> False], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    1, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    1, $CellContext`csys}, {3, -$CellContext`csys}, {
                    2, -$CellContext`csys}], Editable -> False], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    2, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    1, $CellContext`csys}, {3, -$CellContext`csys}, {
                    3, -$CellContext`csys}], Editable -> False], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    3, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    1, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                   MultilineFunction -> None], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}]],
           
           Annotation[#, {1, 
             3} -> $CellContext`ChristoffelPDbluesPDcsys[{
                1, $CellContext`csys}, {3, -$CellContext`csys}, {
                1, -$CellContext`csys}] xAct`xCoba`InChart[
                $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
                1, $CellContext`csys}, {3, -$CellContext`csys}, {
                2, -$CellContext`csys}] xAct`xCoba`InChart[
                $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
                1, $CellContext`csys}, {3, -$CellContext`csys}, {
                3, -$CellContext`csys}] xAct`xCoba`InChart[
                $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + Derivative[0, 0, 1][
               xAct`xCoba`InChart[
                $CellContext`v[{1, $CellContext`csys}], $CellContext`csys]][
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
                1, -$CellContext`csys}, {1, -$CellContext`csys}],
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
                1, -$CellContext`csys}, {2, -$CellContext`csys}],
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
                1, -$CellContext`csys}, {3, -$CellContext`csys}],
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"1", ",", "0", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    2, $CellContext`csys}, {1, -$CellContext`csys}, {
                    1, -$CellContext`csys}], Editable -> False], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    1, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    2, $CellContext`csys}, {1, -$CellContext`csys}, {
                    2, -$CellContext`csys}], Editable -> False], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    2, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    2, $CellContext`csys}, {1, -$CellContext`csys}, {
                    3, -$CellContext`csys}], Editable -> False], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    3, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    2, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                   MultilineFunction -> None], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}]],
           
           Annotation[#, {2, 
             1} -> $CellContext`ChristoffelPDbluesPDcsys[{
                2, $CellContext`csys}, {1, -$CellContext`csys}, {
                1, -$CellContext`csys}] xAct`xCoba`InChart[
                $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
                2, $CellContext`csys}, {1, -$CellContext`csys}, {
                2, -$CellContext`csys}] xAct`xCoba`InChart[
                $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
                2, $CellContext`csys}, {1, -$CellContext`csys}, {
                3, -$CellContext`csys}] xAct`xCoba`InChart[
                $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + Derivative[1, 0, 0][
               xAct`xCoba`InChart[
                $CellContext`v[{2, $CellContext`csys}], $CellContext`csys]][
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
                2, -$CellContext`csys}, {1, -$CellContext`csys}],
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
                2, -$CellContext`csys}, {2, -$CellContext`csys}],
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
                2, -$CellContext`csys}, {3, -$CellContext`csys}],
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"0", ",", "1", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    2, $CellContext`csys}, {2, -$CellContext`csys}, {
                    1, -$CellContext`csys}], Editable -> False], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    1, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    2, $CellContext`csys}, {2, -$CellContext`csys}, {
                    2, -$CellContext`csys}], Editable -> False], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    2, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    2, $CellContext`csys}, {2, -$CellContext`csys}, {
                    3, -$CellContext`csys}], Editable -> False], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    3, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    2, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                   MultilineFunction -> None], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}]],
           
           Annotation[#, {2, 
             2} -> $CellContext`ChristoffelPDbluesPDcsys[{
                2, $CellContext`csys}, {2, -$CellContext`csys}, {
                1, -$CellContext`csys}] xAct`xCoba`InChart[
                $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
                2, $CellContext`csys}, {2, -$CellContext`csys}, {
                2, -$CellContext`csys}] xAct`xCoba`InChart[
                $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
                2, $CellContext`csys}, {2, -$CellContext`csys}, {
                3, -$CellContext`csys}] xAct`xCoba`InChart[
                $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + Derivative[0, 1, 0][
               xAct`xCoba`InChart[
                $CellContext`v[{2, $CellContext`csys}], $CellContext`csys]][
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
                3, -$CellContext`csys}, {1, -$CellContext`csys}],
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
                3, -$CellContext`csys}, {2, -$CellContext`csys}],
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
                3, -$CellContext`csys}, {3, -$CellContext`csys}],
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"0", ",", "0", ",", "1"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    2, $CellContext`csys}, {3, -$CellContext`csys}, {
                    1, -$CellContext`csys}], Editable -> False], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    1, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    2, $CellContext`csys}, {3, -$CellContext`csys}, {
                    2, -$CellContext`csys}], Editable -> False], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    2, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    2, $CellContext`csys}, {3, -$CellContext`csys}, {
                    3, -$CellContext`csys}], Editable -> False], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    3, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    2, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                   MultilineFunction -> None], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}]],
           
           Annotation[#, {2, 
             3} -> $CellContext`ChristoffelPDbluesPDcsys[{
                2, $CellContext`csys}, {3, -$CellContext`csys}, {
                1, -$CellContext`csys}] xAct`xCoba`InChart[
                $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
                2, $CellContext`csys}, {3, -$CellContext`csys}, {
                2, -$CellContext`csys}] xAct`xCoba`InChart[
                $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
                2, $CellContext`csys}, {3, -$CellContext`csys}, {
                3, -$CellContext`csys}] xAct`xCoba`InChart[
                $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + Derivative[0, 0, 1][
               xAct`xCoba`InChart[
                $CellContext`v[{2, $CellContext`csys}], $CellContext`csys]][
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
                1, -$CellContext`csys}, {1, -$CellContext`csys}],
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
                1, -$CellContext`csys}, {2, -$CellContext`csys}],
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
                1, -$CellContext`csys}, {3, -$CellContext`csys}],
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"1", ",", "0", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    3, $CellContext`csys}, {1, -$CellContext`csys}, {
                    1, -$CellContext`csys}], Editable -> False], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    1, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    3, $CellContext`csys}, {1, -$CellContext`csys}, {
                    2, -$CellContext`csys}], Editable -> False], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    2, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    3, $CellContext`csys}, {1, -$CellContext`csys}, {
                    3, -$CellContext`csys}], Editable -> False], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    3, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    3, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                   MultilineFunction -> None], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}]],
           
           Annotation[#, {3, 
             1} -> $CellContext`ChristoffelPDbluesPDcsys[{
                3, $CellContext`csys}, {1, -$CellContext`csys}, {
                1, -$CellContext`csys}] xAct`xCoba`InChart[
                $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
                3, $CellContext`csys}, {1, -$CellContext`csys}, {
                2, -$CellContext`csys}] xAct`xCoba`InChart[
                $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
                3, $CellContext`csys}, {1, -$CellContext`csys}, {
                3, -$CellContext`csys}] xAct`xCoba`InChart[
                $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + Derivative[1, 0, 0][
               xAct`xCoba`InChart[
                $CellContext`v[{3, $CellContext`csys}], $CellContext`csys]][
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
                2, -$CellContext`csys}, {1, -$CellContext`csys}],
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
                2, -$CellContext`csys}, {2, -$CellContext`csys}],
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
                2, -$CellContext`csys}, {3, -$CellContext`csys}],
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"0", ",", "1", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    3, $CellContext`csys}, {2, -$CellContext`csys}, {
                    1, -$CellContext`csys}], Editable -> False], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    1, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    3, $CellContext`csys}, {2, -$CellContext`csys}, {
                    2, -$CellContext`csys}], Editable -> False], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    2, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    3, $CellContext`csys}, {2, -$CellContext`csys}, {
                    3, -$CellContext`csys}], Editable -> False], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    3, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    3, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                   MultilineFunction -> None], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}]],
           
           Annotation[#, {3, 
             2} -> $CellContext`ChristoffelPDbluesPDcsys[{
                3, $CellContext`csys}, {2, -$CellContext`csys}, {
                1, -$CellContext`csys}] xAct`xCoba`InChart[
                $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
                3, $CellContext`csys}, {2, -$CellContext`csys}, {
                2, -$CellContext`csys}] xAct`xCoba`InChart[
                $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
                3, $CellContext`csys}, {2, -$CellContext`csys}, {
                3, -$CellContext`csys}] xAct`xCoba`InChart[
                $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + Derivative[0, 1, 0][
               xAct`xCoba`InChart[
                $CellContext`v[{3, $CellContext`csys}], $CellContext`csys]][
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
                3, -$CellContext`csys}, {1, -$CellContext`csys}],
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
                3, -$CellContext`csys}, {2, -$CellContext`csys}],
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox[GridBox[{
                  {
                   RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                   StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
                3, -$CellContext`csys}, {3, -$CellContext`csys}],
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"0", ",", "0", ",", "1"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    3, $CellContext`csys}, {3, -$CellContext`csys}, {
                    1, -$CellContext`csys}], Editable -> False], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    1, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    3, $CellContext`csys}, {3, -$CellContext`csys}, {
                    2, -$CellContext`csys}], Editable -> False], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    2, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelPDbluesPDcsys[{
                    3, $CellContext`csys}, {3, -$CellContext`csys}, {
                    3, -$CellContext`csys}], Editable -> False], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    3, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    3, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                   MultilineFunction -> None], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}]],
           
           Annotation[#, {3, 
             3} -> $CellContext`ChristoffelPDbluesPDcsys[{
                3, $CellContext`csys}, {3, -$CellContext`csys}, {
                1, -$CellContext`csys}] xAct`xCoba`InChart[
                $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
                3, $CellContext`csys}, {3, -$CellContext`csys}, {
                2, -$CellContext`csys}] xAct`xCoba`InChart[
                $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
                3, $CellContext`csys}, {3, -$CellContext`csys}, {
                3, -$CellContext`csys}] xAct`xCoba`InChart[
                $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + Derivative[0, 0, 1][
               xAct`xCoba`InChart[
                $CellContext`v[{3, $CellContext`csys}], $CellContext`csys]][
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]], "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0.5, 0], 4 -> RGBColor[1, 0.5, 0]}, 
          "RowsIndexed" -> {
           1 -> RGBColor[1, 0.5, 0], 4 -> RGBColor[1, 0.5, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a", " "},
         {" ", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{$CellContext`ChristoffelPDbluesPDcsys[{
        1, $CellContext`csys}, {1, -$CellContext`csys}, {
        1, -$CellContext`csys}] xAct`xCoba`InChart[
        $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
        1, $CellContext`csys}, {1, -$CellContext`csys}, {
        2, -$CellContext`csys}] xAct`xCoba`InChart[
        $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
        1, $CellContext`csys}, {1, -$CellContext`csys}, {
        3, -$CellContext`csys}] xAct`xCoba`InChart[
        $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + Derivative[1, 0, 0][
       xAct`xCoba`InChart[
        $CellContext`v[{1, $CellContext`csys}], $CellContext`csys]][
       $CellContext`x[], 
       $CellContext`y[], 
       $CellContext`z[]], $CellContext`ChristoffelPDbluesPDcsys[{
        1, $CellContext`csys}, {2, -$CellContext`csys}, {
        1, -$CellContext`csys}] xAct`xCoba`InChart[
        $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
        1, $CellContext`csys}, {2, -$CellContext`csys}, {
        2, -$CellContext`csys}] xAct`xCoba`InChart[
        $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
        1, $CellContext`csys}, {2, -$CellContext`csys}, {
        3, -$CellContext`csys}] xAct`xCoba`InChart[
        $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + Derivative[0, 1, 0][
       xAct`xCoba`InChart[
        $CellContext`v[{1, $CellContext`csys}], $CellContext`csys]][
       $CellContext`x[], 
       $CellContext`y[], 
       $CellContext`z[]], $CellContext`ChristoffelPDbluesPDcsys[{
        1, $CellContext`csys}, {3, -$CellContext`csys}, {
        1, -$CellContext`csys}] xAct`xCoba`InChart[
        $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
        1, $CellContext`csys}, {3, -$CellContext`csys}, {
        2, -$CellContext`csys}] xAct`xCoba`InChart[
        $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
        1, $CellContext`csys}, {3, -$CellContext`csys}, {
        3, -$CellContext`csys}] xAct`xCoba`InChart[
        $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + Derivative[0, 0, 1][
       xAct`xCoba`InChart[
        $CellContext`v[{1, $CellContext`csys}], $CellContext`csys]][
       $CellContext`x[], 
       $CellContext`y[], 
       $CellContext`z[]]}, {$CellContext`ChristoffelPDbluesPDcsys[{
        2, $CellContext`csys}, {1, -$CellContext`csys}, {
        1, -$CellContext`csys}] xAct`xCoba`InChart[
        $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
        2, $CellContext`csys}, {1, -$CellContext`csys}, {
        2, -$CellContext`csys}] xAct`xCoba`InChart[
        $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
        2, $CellContext`csys}, {1, -$CellContext`csys}, {
        3, -$CellContext`csys}] xAct`xCoba`InChart[
        $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + Derivative[1, 0, 0][
       xAct`xCoba`InChart[
        $CellContext`v[{2, $CellContext`csys}], $CellContext`csys]][
       $CellContext`x[], 
       $CellContext`y[], 
       $CellContext`z[]], $CellContext`ChristoffelPDbluesPDcsys[{
        2, $CellContext`csys}, {2, -$CellContext`csys}, {
        1, -$CellContext`csys}] xAct`xCoba`InChart[
        $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
        2, $CellContext`csys}, {2, -$CellContext`csys}, {
        2, -$CellContext`csys}] xAct`xCoba`InChart[
        $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
        2, $CellContext`csys}, {2, -$CellContext`csys}, {
        3, -$CellContext`csys}] xAct`xCoba`InChart[
        $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + Derivative[0, 1, 0][
       xAct`xCoba`InChart[
        $CellContext`v[{2, $CellContext`csys}], $CellContext`csys]][
       $CellContext`x[], 
       $CellContext`y[], 
       $CellContext`z[]], $CellContext`ChristoffelPDbluesPDcsys[{
        2, $CellContext`csys}, {3, -$CellContext`csys}, {
        1, -$CellContext`csys}] xAct`xCoba`InChart[
        $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
        2, $CellContext`csys}, {3, -$CellContext`csys}, {
        2, -$CellContext`csys}] xAct`xCoba`InChart[
        $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
        2, $CellContext`csys}, {3, -$CellContext`csys}, {
        3, -$CellContext`csys}] xAct`xCoba`InChart[
        $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + Derivative[0, 0, 1][
       xAct`xCoba`InChart[
        $CellContext`v[{2, $CellContext`csys}], $CellContext`csys]][
       $CellContext`x[], 
       $CellContext`y[], 
       $CellContext`z[]]}, {$CellContext`ChristoffelPDbluesPDcsys[{
        3, $CellContext`csys}, {1, -$CellContext`csys}, {
        1, -$CellContext`csys}] xAct`xCoba`InChart[
        $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
        3, $CellContext`csys}, {1, -$CellContext`csys}, {
        2, -$CellContext`csys}] xAct`xCoba`InChart[
        $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
        3, $CellContext`csys}, {1, -$CellContext`csys}, {
        3, -$CellContext`csys}] xAct`xCoba`InChart[
        $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + Derivative[1, 0, 0][
       xAct`xCoba`InChart[
        $CellContext`v[{3, $CellContext`csys}], $CellContext`csys]][
       $CellContext`x[], 
       $CellContext`y[], 
       $CellContext`z[]], $CellContext`ChristoffelPDbluesPDcsys[{
        3, $CellContext`csys}, {2, -$CellContext`csys}, {
        1, -$CellContext`csys}] xAct`xCoba`InChart[
        $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
        3, $CellContext`csys}, {2, -$CellContext`csys}, {
        2, -$CellContext`csys}] xAct`xCoba`InChart[
        $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
        3, $CellContext`csys}, {2, -$CellContext`csys}, {
        3, -$CellContext`csys}] xAct`xCoba`InChart[
        $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + Derivative[0, 1, 0][
       xAct`xCoba`InChart[
        $CellContext`v[{3, $CellContext`csys}], $CellContext`csys]][
       $CellContext`x[], 
       $CellContext`y[], 
       $CellContext`z[]], $CellContext`ChristoffelPDbluesPDcsys[{
        3, $CellContext`csys}, {3, -$CellContext`csys}, {
        1, -$CellContext`csys}] xAct`xCoba`InChart[
        $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
        3, $CellContext`csys}, {3, -$CellContext`csys}, {
        2, -$CellContext`csys}] xAct`xCoba`InChart[
        $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + $CellContext`ChristoffelPDbluesPDcsys[{
        3, $CellContext`csys}, {3, -$CellContext`csys}, {
        3, -$CellContext`csys}] xAct`xCoba`InChart[
        $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + Derivative[0, 0, 1][
       xAct`xCoba`InChart[
        $CellContext`v[{3, $CellContext`csys}], $CellContext`csys]][
       $CellContext`x[], 
       $CellContext`y[], 
       $CellContext`z[]]}}, {$CellContext`csys, -$CellContext`csys}, 
   0][$CellContext`a, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{{3.7288459419434357`*^9, 3.728846012772623*^9}, 
   3.7333675683105783`*^9, 3.792273621628491*^9},
 CellLabel->"Out[338]=",
 CellID->813641839,ExpressionUUID->"93713607-6a67-4cc1-b8e0-6c3808b12271"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", 
  RowBox[{"ToInChart", "[", "csys", "]"}]}]], "Input",
 CellChangeTimes->{{3.728845984679652*^9, 3.728845990623561*^9}},
 CellLabel->"In[339]:=",
 CellID->2033365647,ExpressionUUID->"c71a330d-14b5-4c5d-8f2d-264cf926b500"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`ChristoffelPDbluesPDcsys[{
                  1, $CellContext`csys}, {1, -$CellContext`csys}, {
                  1, -$CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`ChristoffelPDbluesPDcsys[{
                  1, $CellContext`csys}, {1, -$CellContext`csys}, {
                  2, -$CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`ChristoffelPDbluesPDcsys[{
                  1, $CellContext`csys}, {1, -$CellContext`csys}, {
                  3, -$CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"1", ",", "0", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`ChristoffelPDbluesPDcsys[{
                    1, $CellContext`csys}, {1, -$CellContext`csys}, {
                    1, -$CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    1, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`ChristoffelPDbluesPDcsys[{
                    1, $CellContext`csys}, {1, -$CellContext`csys}, {
                    2, -$CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    2, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`ChristoffelPDbluesPDcsys[{
                    1, $CellContext`csys}, {1, -$CellContext`csys}, {
                    3, -$CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    3, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    1, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                   MultilineFunction -> None], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}]],
           Annotation[#, {1, 1} -> xAct`xCoba`InChart[
                $CellContext`ChristoffelPDbluesPDcsys[{
                 1, $CellContext`csys}, {1, -$CellContext`csys}, {
                 1, -$CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] xAct`xCoba`InChart[
                $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + xAct`xCoba`InChart[
                $CellContext`ChristoffelPDbluesPDcsys[{
                 1, $CellContext`csys}, {1, -$CellContext`csys}, {
                 2, -$CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] xAct`xCoba`InChart[
                $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + xAct`xCoba`InChart[
                $CellContext`ChristoffelPDbluesPDcsys[{
                 1, $CellContext`csys}, {1, -$CellContext`csys}, {
                 3, -$CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] xAct`xCoba`InChart[
                $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + Derivative[1, 0, 0][
               xAct`xCoba`InChart[
                $CellContext`v[{1, $CellContext`csys}], $CellContext`csys]][
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`ChristoffelPDbluesPDcsys[{
                  1, $CellContext`csys}, {2, -$CellContext`csys}, {
                  1, -$CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`ChristoffelPDbluesPDcsys[{
                  1, $CellContext`csys}, {2, -$CellContext`csys}, {
                  2, -$CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`ChristoffelPDbluesPDcsys[{
                  1, $CellContext`csys}, {2, -$CellContext`csys}, {
                  3, -$CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"0", ",", "1", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`ChristoffelPDbluesPDcsys[{
                    1, $CellContext`csys}, {2, -$CellContext`csys}, {
                    1, -$CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    1, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`ChristoffelPDbluesPDcsys[{
                    1, $CellContext`csys}, {2, -$CellContext`csys}, {
                    2, -$CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    2, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`ChristoffelPDbluesPDcsys[{
                    1, $CellContext`csys}, {2, -$CellContext`csys}, {
                    3, -$CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    3, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    1, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                   MultilineFunction -> None], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}]],
           Annotation[#, {1, 2} -> xAct`xCoba`InChart[
                $CellContext`ChristoffelPDbluesPDcsys[{
                 1, $CellContext`csys}, {2, -$CellContext`csys}, {
                 1, -$CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] xAct`xCoba`InChart[
                $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + xAct`xCoba`InChart[
                $CellContext`ChristoffelPDbluesPDcsys[{
                 1, $CellContext`csys}, {2, -$CellContext`csys}, {
                 2, -$CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] xAct`xCoba`InChart[
                $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + xAct`xCoba`InChart[
                $CellContext`ChristoffelPDbluesPDcsys[{
                 1, $CellContext`csys}, {2, -$CellContext`csys}, {
                 3, -$CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] xAct`xCoba`InChart[
                $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + Derivative[0, 1, 0][
               xAct`xCoba`InChart[
                $CellContext`v[{1, $CellContext`csys}], $CellContext`csys]][
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`ChristoffelPDbluesPDcsys[{
                  1, $CellContext`csys}, {3, -$CellContext`csys}, {
                  1, -$CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`ChristoffelPDbluesPDcsys[{
                  1, $CellContext`csys}, {3, -$CellContext`csys}, {
                  2, -$CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`ChristoffelPDbluesPDcsys[{
                  1, $CellContext`csys}, {3, -$CellContext`csys}, {
                  3, -$CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"0", ",", "0", ",", "1"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`ChristoffelPDbluesPDcsys[{
                    1, $CellContext`csys}, {3, -$CellContext`csys}, {
                    1, -$CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    1, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`ChristoffelPDbluesPDcsys[{
                    1, $CellContext`csys}, {3, -$CellContext`csys}, {
                    2, -$CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    2, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`ChristoffelPDbluesPDcsys[{
                    1, $CellContext`csys}, {3, -$CellContext`csys}, {
                    3, -$CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    3, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    1, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                   MultilineFunction -> None], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}]],
           Annotation[#, {1, 3} -> xAct`xCoba`InChart[
                $CellContext`ChristoffelPDbluesPDcsys[{
                 1, $CellContext`csys}, {3, -$CellContext`csys}, {
                 1, -$CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] xAct`xCoba`InChart[
                $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + xAct`xCoba`InChart[
                $CellContext`ChristoffelPDbluesPDcsys[{
                 1, $CellContext`csys}, {3, -$CellContext`csys}, {
                 2, -$CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] xAct`xCoba`InChart[
                $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + xAct`xCoba`InChart[
                $CellContext`ChristoffelPDbluesPDcsys[{
                 1, $CellContext`csys}, {3, -$CellContext`csys}, {
                 3, -$CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] xAct`xCoba`InChart[
                $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + Derivative[0, 0, 1][
               xAct`xCoba`InChart[
                $CellContext`v[{1, $CellContext`csys}], $CellContext`csys]][
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`ChristoffelPDbluesPDcsys[{
                  2, $CellContext`csys}, {1, -$CellContext`csys}, {
                  1, -$CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`ChristoffelPDbluesPDcsys[{
                  2, $CellContext`csys}, {1, -$CellContext`csys}, {
                  2, -$CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`ChristoffelPDbluesPDcsys[{
                  2, $CellContext`csys}, {1, -$CellContext`csys}, {
                  3, -$CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"1", ",", "0", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`ChristoffelPDbluesPDcsys[{
                    2, $CellContext`csys}, {1, -$CellContext`csys}, {
                    1, -$CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    1, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`ChristoffelPDbluesPDcsys[{
                    2, $CellContext`csys}, {1, -$CellContext`csys}, {
                    2, -$CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    2, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`ChristoffelPDbluesPDcsys[{
                    2, $CellContext`csys}, {1, -$CellContext`csys}, {
                    3, -$CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    3, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    2, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                   MultilineFunction -> None], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}]],
           Annotation[#, {2, 1} -> xAct`xCoba`InChart[
                $CellContext`ChristoffelPDbluesPDcsys[{
                 2, $CellContext`csys}, {1, -$CellContext`csys}, {
                 1, -$CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] xAct`xCoba`InChart[
                $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + xAct`xCoba`InChart[
                $CellContext`ChristoffelPDbluesPDcsys[{
                 2, $CellContext`csys}, {1, -$CellContext`csys}, {
                 2, -$CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] xAct`xCoba`InChart[
                $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + xAct`xCoba`InChart[
                $CellContext`ChristoffelPDbluesPDcsys[{
                 2, $CellContext`csys}, {1, -$CellContext`csys}, {
                 3, -$CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] xAct`xCoba`InChart[
                $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + Derivative[1, 0, 0][
               xAct`xCoba`InChart[
                $CellContext`v[{2, $CellContext`csys}], $CellContext`csys]][
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`ChristoffelPDbluesPDcsys[{
                  2, $CellContext`csys}, {2, -$CellContext`csys}, {
                  1, -$CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`ChristoffelPDbluesPDcsys[{
                  2, $CellContext`csys}, {2, -$CellContext`csys}, {
                  2, -$CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`ChristoffelPDbluesPDcsys[{
                  2, $CellContext`csys}, {2, -$CellContext`csys}, {
                  3, -$CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"0", ",", "1", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`ChristoffelPDbluesPDcsys[{
                    2, $CellContext`csys}, {2, -$CellContext`csys}, {
                    1, -$CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    1, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`ChristoffelPDbluesPDcsys[{
                    2, $CellContext`csys}, {2, -$CellContext`csys}, {
                    2, -$CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    2, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`ChristoffelPDbluesPDcsys[{
                    2, $CellContext`csys}, {2, -$CellContext`csys}, {
                    3, -$CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    3, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    2, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                   MultilineFunction -> None], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}]],
           Annotation[#, {2, 2} -> xAct`xCoba`InChart[
                $CellContext`ChristoffelPDbluesPDcsys[{
                 2, $CellContext`csys}, {2, -$CellContext`csys}, {
                 1, -$CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] xAct`xCoba`InChart[
                $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + xAct`xCoba`InChart[
                $CellContext`ChristoffelPDbluesPDcsys[{
                 2, $CellContext`csys}, {2, -$CellContext`csys}, {
                 2, -$CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] xAct`xCoba`InChart[
                $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + xAct`xCoba`InChart[
                $CellContext`ChristoffelPDbluesPDcsys[{
                 2, $CellContext`csys}, {2, -$CellContext`csys}, {
                 3, -$CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] xAct`xCoba`InChart[
                $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + Derivative[0, 1, 0][
               xAct`xCoba`InChart[
                $CellContext`v[{2, $CellContext`csys}], $CellContext`csys]][
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`ChristoffelPDbluesPDcsys[{
                  2, $CellContext`csys}, {3, -$CellContext`csys}, {
                  1, -$CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`ChristoffelPDbluesPDcsys[{
                  2, $CellContext`csys}, {3, -$CellContext`csys}, {
                  2, -$CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`ChristoffelPDbluesPDcsys[{
                  2, $CellContext`csys}, {3, -$CellContext`csys}, {
                  3, -$CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"0", ",", "0", ",", "1"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`ChristoffelPDbluesPDcsys[{
                    2, $CellContext`csys}, {3, -$CellContext`csys}, {
                    1, -$CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    1, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`ChristoffelPDbluesPDcsys[{
                    2, $CellContext`csys}, {3, -$CellContext`csys}, {
                    2, -$CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    2, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`ChristoffelPDbluesPDcsys[{
                    2, $CellContext`csys}, {3, -$CellContext`csys}, {
                    3, -$CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    3, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    2, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                   MultilineFunction -> None], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}]],
           Annotation[#, {2, 3} -> xAct`xCoba`InChart[
                $CellContext`ChristoffelPDbluesPDcsys[{
                 2, $CellContext`csys}, {3, -$CellContext`csys}, {
                 1, -$CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] xAct`xCoba`InChart[
                $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + xAct`xCoba`InChart[
                $CellContext`ChristoffelPDbluesPDcsys[{
                 2, $CellContext`csys}, {3, -$CellContext`csys}, {
                 2, -$CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] xAct`xCoba`InChart[
                $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + xAct`xCoba`InChart[
                $CellContext`ChristoffelPDbluesPDcsys[{
                 2, $CellContext`csys}, {3, -$CellContext`csys}, {
                 3, -$CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] xAct`xCoba`InChart[
                $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + Derivative[0, 0, 1][
               xAct`xCoba`InChart[
                $CellContext`v[{2, $CellContext`csys}], $CellContext`csys]][
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`ChristoffelPDbluesPDcsys[{
                  3, $CellContext`csys}, {1, -$CellContext`csys}, {
                  1, -$CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`ChristoffelPDbluesPDcsys[{
                  3, $CellContext`csys}, {1, -$CellContext`csys}, {
                  2, -$CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`ChristoffelPDbluesPDcsys[{
                  3, $CellContext`csys}, {1, -$CellContext`csys}, {
                  3, -$CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"1", ",", "0", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`ChristoffelPDbluesPDcsys[{
                    3, $CellContext`csys}, {1, -$CellContext`csys}, {
                    1, -$CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    1, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`ChristoffelPDbluesPDcsys[{
                    3, $CellContext`csys}, {1, -$CellContext`csys}, {
                    2, -$CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    2, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`ChristoffelPDbluesPDcsys[{
                    3, $CellContext`csys}, {1, -$CellContext`csys}, {
                    3, -$CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    3, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    3, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                   MultilineFunction -> None], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}]],
           Annotation[#, {3, 1} -> xAct`xCoba`InChart[
                $CellContext`ChristoffelPDbluesPDcsys[{
                 3, $CellContext`csys}, {1, -$CellContext`csys}, {
                 1, -$CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] xAct`xCoba`InChart[
                $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + xAct`xCoba`InChart[
                $CellContext`ChristoffelPDbluesPDcsys[{
                 3, $CellContext`csys}, {1, -$CellContext`csys}, {
                 2, -$CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] xAct`xCoba`InChart[
                $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + xAct`xCoba`InChart[
                $CellContext`ChristoffelPDbluesPDcsys[{
                 3, $CellContext`csys}, {1, -$CellContext`csys}, {
                 3, -$CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] xAct`xCoba`InChart[
                $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + Derivative[1, 0, 0][
               xAct`xCoba`InChart[
                $CellContext`v[{3, $CellContext`csys}], $CellContext`csys]][
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`ChristoffelPDbluesPDcsys[{
                  3, $CellContext`csys}, {2, -$CellContext`csys}, {
                  1, -$CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`ChristoffelPDbluesPDcsys[{
                  3, $CellContext`csys}, {2, -$CellContext`csys}, {
                  2, -$CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`ChristoffelPDbluesPDcsys[{
                  3, $CellContext`csys}, {2, -$CellContext`csys}, {
                  3, -$CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"0", ",", "1", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`ChristoffelPDbluesPDcsys[{
                    3, $CellContext`csys}, {2, -$CellContext`csys}, {
                    1, -$CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    1, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`ChristoffelPDbluesPDcsys[{
                    3, $CellContext`csys}, {2, -$CellContext`csys}, {
                    2, -$CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    2, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`ChristoffelPDbluesPDcsys[{
                    3, $CellContext`csys}, {2, -$CellContext`csys}, {
                    3, -$CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    3, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    3, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                   MultilineFunction -> None], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}]],
           Annotation[#, {3, 2} -> xAct`xCoba`InChart[
                $CellContext`ChristoffelPDbluesPDcsys[{
                 3, $CellContext`csys}, {2, -$CellContext`csys}, {
                 1, -$CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] xAct`xCoba`InChart[
                $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + xAct`xCoba`InChart[
                $CellContext`ChristoffelPDbluesPDcsys[{
                 3, $CellContext`csys}, {2, -$CellContext`csys}, {
                 2, -$CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] xAct`xCoba`InChart[
                $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + xAct`xCoba`InChart[
                $CellContext`ChristoffelPDbluesPDcsys[{
                 3, $CellContext`csys}, {2, -$CellContext`csys}, {
                 3, -$CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] xAct`xCoba`InChart[
                $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + Derivative[0, 1, 0][
               xAct`xCoba`InChart[
                $CellContext`v[{3, $CellContext`csys}], $CellContext`csys]][
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`ChristoffelPDbluesPDcsys[{
                  3, $CellContext`csys}, {3, -$CellContext`csys}, {
                  1, -$CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{1, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`ChristoffelPDbluesPDcsys[{
                  3, $CellContext`csys}, {3, -$CellContext`csys}, {
                  2, -$CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{2, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[0., 0., 1.]], ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0.5, 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`ChristoffelPDbluesPDcsys[{
                  3, $CellContext`csys}, {3, -$CellContext`csys}, {
                  3, -$CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox[
                 UnderscriptBox[GridBox[{
                    {"v", 
                    StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0.5, 0.]]},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                    FontSize->9]}
                   },
                   
                   GridBoxAlignment->{
                    "Columns" -> {{Center}}, "Rows" -> {{Center}}},
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}}], 
                  StyleBox["_",
                   FontColor->RGBColor[1, 0.5, 0]]],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                xAct`xCoba`InChart[
                 $CellContext`v[{3, $CellContext`csys}], $CellContext`csys],
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"0", ",", "0", ",", "1"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`ChristoffelPDbluesPDcsys[{
                    3, $CellContext`csys}, {3, -$CellContext`csys}, {
                    1, -$CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    1, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`ChristoffelPDbluesPDcsys[{
                    3, $CellContext`csys}, {3, -$CellContext`csys}, {
                    2, -$CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    2, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\!\(\*StyleBox[\"\[ScriptCapitalD]\",\
FontColor->RGBColor[0., 0., \
1.]]\),\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1., 0.5, 0.]]\
\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", " ", " "}, {" ", 
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`ChristoffelPDbluesPDcsys[{
                    3, $CellContext`csys}, {3, -$CellContext`csys}, {
                    3, -$CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    3, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    UnderscriptBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0.5, \
0.]]\)"}, {" "}}, RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    "\*StyleBox[\"_\",FontColor->RGBColor[1, 0.5, 0]]"], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xCoba`InChart[
                    $CellContext`v[{
                    3, $CellContext`csys}], $CellContext`csys], Editable -> 
                    False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                   MultilineFunction -> None], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}]],
           Annotation[#, {3, 3} -> xAct`xCoba`InChart[
                $CellContext`ChristoffelPDbluesPDcsys[{
                 3, $CellContext`csys}, {3, -$CellContext`csys}, {
                 1, -$CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] xAct`xCoba`InChart[
                $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + xAct`xCoba`InChart[
                $CellContext`ChristoffelPDbluesPDcsys[{
                 3, $CellContext`csys}, {3, -$CellContext`csys}, {
                 2, -$CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] xAct`xCoba`InChart[
                $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + xAct`xCoba`InChart[
                $CellContext`ChristoffelPDbluesPDcsys[{
                 3, $CellContext`csys}, {3, -$CellContext`csys}, {
                 3, -$CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] xAct`xCoba`InChart[
                $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + Derivative[0, 0, 1][
               xAct`xCoba`InChart[
                $CellContext`v[{3, $CellContext`csys}], $CellContext`csys]][
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]], "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0.5, 0], 4 -> RGBColor[1, 0.5, 0]}, 
          "RowsIndexed" -> {
           1 -> RGBColor[1, 0.5, 0], 4 -> RGBColor[1, 0.5, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a", " "},
         {" ", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{xAct`xCoba`InChart[
        $CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
         1, -$CellContext`csys}, {
         1, -$CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] xAct`xCoba`InChart[
        $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + xAct`xCoba`InChart[
        $CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
         1, -$CellContext`csys}, {
         2, -$CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] xAct`xCoba`InChart[
        $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + xAct`xCoba`InChart[
        $CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
         1, -$CellContext`csys}, {
         3, -$CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] xAct`xCoba`InChart[
        $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + Derivative[1, 0, 0][
       xAct`xCoba`InChart[
        $CellContext`v[{1, $CellContext`csys}], $CellContext`csys]][
       $CellContext`x[], 
       $CellContext`y[], 
       $CellContext`z[]], xAct`xCoba`InChart[
        $CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
         2, -$CellContext`csys}, {
         1, -$CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] xAct`xCoba`InChart[
        $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + xAct`xCoba`InChart[
        $CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
         2, -$CellContext`csys}, {
         2, -$CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] xAct`xCoba`InChart[
        $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + xAct`xCoba`InChart[
        $CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
         2, -$CellContext`csys}, {
         3, -$CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] xAct`xCoba`InChart[
        $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + Derivative[0, 1, 0][
       xAct`xCoba`InChart[
        $CellContext`v[{1, $CellContext`csys}], $CellContext`csys]][
       $CellContext`x[], 
       $CellContext`y[], 
       $CellContext`z[]], xAct`xCoba`InChart[
        $CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
         3, -$CellContext`csys}, {
         1, -$CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] xAct`xCoba`InChart[
        $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + xAct`xCoba`InChart[
        $CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
         3, -$CellContext`csys}, {
         2, -$CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] xAct`xCoba`InChart[
        $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + xAct`xCoba`InChart[
        $CellContext`ChristoffelPDbluesPDcsys[{1, $CellContext`csys}, {
         3, -$CellContext`csys}, {
         3, -$CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] xAct`xCoba`InChart[
        $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + Derivative[0, 0, 1][
       xAct`xCoba`InChart[
        $CellContext`v[{1, $CellContext`csys}], $CellContext`csys]][
       $CellContext`x[], 
       $CellContext`y[], 
       $CellContext`z[]]}, {xAct`xCoba`InChart[
        $CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
         1, -$CellContext`csys}, {
         1, -$CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] xAct`xCoba`InChart[
        $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + xAct`xCoba`InChart[
        $CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
         1, -$CellContext`csys}, {
         2, -$CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] xAct`xCoba`InChart[
        $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + xAct`xCoba`InChart[
        $CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
         1, -$CellContext`csys}, {
         3, -$CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] xAct`xCoba`InChart[
        $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + Derivative[1, 0, 0][
       xAct`xCoba`InChart[
        $CellContext`v[{2, $CellContext`csys}], $CellContext`csys]][
       $CellContext`x[], 
       $CellContext`y[], 
       $CellContext`z[]], xAct`xCoba`InChart[
        $CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
         2, -$CellContext`csys}, {
         1, -$CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] xAct`xCoba`InChart[
        $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + xAct`xCoba`InChart[
        $CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
         2, -$CellContext`csys}, {
         2, -$CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] xAct`xCoba`InChart[
        $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + xAct`xCoba`InChart[
        $CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
         2, -$CellContext`csys}, {
         3, -$CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] xAct`xCoba`InChart[
        $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + Derivative[0, 1, 0][
       xAct`xCoba`InChart[
        $CellContext`v[{2, $CellContext`csys}], $CellContext`csys]][
       $CellContext`x[], 
       $CellContext`y[], 
       $CellContext`z[]], xAct`xCoba`InChart[
        $CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
         3, -$CellContext`csys}, {
         1, -$CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] xAct`xCoba`InChart[
        $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + xAct`xCoba`InChart[
        $CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
         3, -$CellContext`csys}, {
         2, -$CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] xAct`xCoba`InChart[
        $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + xAct`xCoba`InChart[
        $CellContext`ChristoffelPDbluesPDcsys[{2, $CellContext`csys}, {
         3, -$CellContext`csys}, {
         3, -$CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] xAct`xCoba`InChart[
        $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + Derivative[0, 0, 1][
       xAct`xCoba`InChart[
        $CellContext`v[{2, $CellContext`csys}], $CellContext`csys]][
       $CellContext`x[], 
       $CellContext`y[], 
       $CellContext`z[]]}, {xAct`xCoba`InChart[
        $CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
         1, -$CellContext`csys}, {
         1, -$CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] xAct`xCoba`InChart[
        $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + xAct`xCoba`InChart[
        $CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
         1, -$CellContext`csys}, {
         2, -$CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] xAct`xCoba`InChart[
        $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + xAct`xCoba`InChart[
        $CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
         1, -$CellContext`csys}, {
         3, -$CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] xAct`xCoba`InChart[
        $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + Derivative[1, 0, 0][
       xAct`xCoba`InChart[
        $CellContext`v[{3, $CellContext`csys}], $CellContext`csys]][
       $CellContext`x[], 
       $CellContext`y[], 
       $CellContext`z[]], xAct`xCoba`InChart[
        $CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
         2, -$CellContext`csys}, {
         1, -$CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] xAct`xCoba`InChart[
        $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + xAct`xCoba`InChart[
        $CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
         2, -$CellContext`csys}, {
         2, -$CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] xAct`xCoba`InChart[
        $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + xAct`xCoba`InChart[
        $CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
         2, -$CellContext`csys}, {
         3, -$CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] xAct`xCoba`InChart[
        $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + Derivative[0, 1, 0][
       xAct`xCoba`InChart[
        $CellContext`v[{3, $CellContext`csys}], $CellContext`csys]][
       $CellContext`x[], 
       $CellContext`y[], 
       $CellContext`z[]], xAct`xCoba`InChart[
        $CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
         3, -$CellContext`csys}, {
         1, -$CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] xAct`xCoba`InChart[
        $CellContext`v[{1, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + xAct`xCoba`InChart[
        $CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
         3, -$CellContext`csys}, {
         2, -$CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] xAct`xCoba`InChart[
        $CellContext`v[{2, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + xAct`xCoba`InChart[
        $CellContext`ChristoffelPDbluesPDcsys[{3, $CellContext`csys}, {
         3, -$CellContext`csys}, {
         3, -$CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] xAct`xCoba`InChart[
        $CellContext`v[{3, $CellContext`csys}], $CellContext`csys][
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]] + Derivative[0, 0, 1][
       xAct`xCoba`InChart[
        $CellContext`v[{3, $CellContext`csys}], $CellContext`csys]][
       $CellContext`x[], 
       $CellContext`y[], 
       $CellContext`z[]]}}, {$CellContext`csys, -$CellContext`csys}, 
   0][$CellContext`a, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{{3.7288459914233847`*^9, 3.7288460136174726`*^9}, 
   3.733367568720933*^9, 3.79227362285432*^9},
 CellLabel->"Out[339]=",
 CellID->1732991312,ExpressionUUID->"637b4e77-2b5b-477b-84f9-c661f90c87d8"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"$LargeComponentSize", "=", "1000"}], ";"}]], "Input",
 CellLabel->"In[340]:=",
 CellID->1259202711,ExpressionUUID->"be026b14-5f83-4716-b938-6e1495299782"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["2.3. Clean up", "Subsection",
 CellChangeTimes->{
  3.728670946657796*^9, {3.7286734807787313`*^9, 3.728673481371875*^9}, {
   3.7287512652274923`*^9, 3.728751271641602*^9}},
 CellID->969574339,ExpressionUUID->"a6f3d01d-d3d5-4e38-9231-6295fbe85441"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefChart", "[", "csys", "]"}]], "Input",
 CellChangeTimes->{{3.728751308795063*^9, 3.7287513279988728`*^9}},
 CellLabel->"In[341]:=",
 CellID->99866145,ExpressionUUID->"44b8029e-b934-4298-a4b9-efd86e20f302"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "x"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "coordinate scalar", 
   " ", $CellContext`x],
  Editable->False]], "Print",
 CellChangeTimes->{3.72875132844075*^9, 3.728751660429084*^9, 
  3.7288433664199963`*^9, 3.733367582475669*^9, 3.792273636073277*^9},
 CellLabel->"During evaluation of In[341]:=",
 CellID->378126311,ExpressionUUID->"3d8e4379-848c-48be-840a-dac9f0762408"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "y"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "coordinate scalar", 
   " ", $CellContext`y],
  Editable->False]], "Print",
 CellChangeTimes->{3.72875132844075*^9, 3.728751660429084*^9, 
  3.7288433664199963`*^9, 3.733367582475669*^9, 3.792273636088706*^9},
 CellLabel->"During evaluation of In[341]:=",
 CellID->764797504,ExpressionUUID->"c082fa0a-b726-4a60-9769-2e08f28ec031"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "z"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "coordinate scalar", 
   " ", $CellContext`z],
  Editable->False]], "Print",
 CellChangeTimes->{3.72875132844075*^9, 3.728751660429084*^9, 
  3.7288433664199963`*^9, 3.733367582475669*^9, 3.792273636105174*^9},
 CellLabel->"During evaluation of In[341]:=",
 CellID->1262022433,ExpressionUUID->"f16f9c8e-1db8-4179-9f31-ae410beaf31d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"mapping differential tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "dicsys"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "mapping differential tensor", " ", $CellContext`dicsys],
  Editable->False]], "Print",
 CellChangeTimes->{3.72875132844075*^9, 3.728751660429084*^9, 
  3.7288433664199963`*^9, 3.733367582475669*^9, 3.7922736361119213`*^9},
 CellLabel->"During evaluation of In[341]:=",
 CellID->115050012,ExpressionUUID->"5e4ea8db-3d98-4a2a-96fe-a0c481de03ea"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefMapping", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"inverse mapping\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "icsys"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefMapping, ": Undefined ", "inverse mapping", 
   " ", $CellContext`icsys],
  Editable->False]], "Print",
 CellChangeTimes->{3.72875132844075*^9, 3.728751660429084*^9, 
  3.7288433664199963`*^9, 3.733367582475669*^9, 3.7922736361473713`*^9},
 CellLabel->"During evaluation of In[341]:=",
 CellID->910269384,ExpressionUUID->"e1bee64c-a166-40f2-8545-021845ade5c9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"mapping differential tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "dcsys"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "mapping differential tensor", " ", $CellContext`dcsys],
  Editable->False]], "Print",
 CellChangeTimes->{3.72875132844075*^9, 3.728751660429084*^9, 
  3.7288433664199963`*^9, 3.733367582475669*^9, 3.792273636157991*^9},
 CellLabel->"During evaluation of In[341]:=",
 CellID->224035660,ExpressionUUID->"de24e7c8-7777-4364-a244-bbfc5584ab5f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "ChristoffelPDbluesPDcsys"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`ChristoffelPDbluesPDcsys],
  Editable->False]], "Print",
 CellChangeTimes->{3.72875132844075*^9, 3.728751660429084*^9, 
  3.7288433664199963`*^9, 3.733367582475669*^9, 3.792273636167101*^9},
 CellLabel->"During evaluation of In[341]:=",
 CellID->828504429,ExpressionUUID->"71d2dd17-006c-4e9b-9d57-96452b3a00fa"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "ChristoffelPDcsys"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Christoffel tensor", " ", $CellContext`ChristoffelPDcsys],
  Editable->False]], "Print",
 CellChangeTimes->{3.72875132844075*^9, 3.728751660429084*^9, 
  3.7288433664199963`*^9, 3.733367582475669*^9, 3.792273636174238*^9},
 CellLabel->"During evaluation of In[341]:=",
 CellID->389971055,ExpressionUUID->"b0f58539-2f5d-49c9-a1f1-33bdeb8f3b33"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciPDcsys"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Ricci tensor", 
   " ", $CellContext`RicciPDcsys],
  Editable->False]], "Print",
 CellChangeTimes->{3.72875132844075*^9, 3.728751660429084*^9, 
  3.7288433664199963`*^9, 3.733367582475669*^9, 3.792273636180799*^9},
 CellLabel->"During evaluation of In[341]:=",
 CellID->123430741,ExpressionUUID->"5f8fbb76-e854-4a5c-944c-38a59116d387"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannPDcsys"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannPDcsys],
  Editable->False]], "Print",
 CellChangeTimes->{3.72875132844075*^9, 3.728751660429084*^9, 
  3.7288433664199963`*^9, 3.733367582475669*^9, 3.792273636188221*^9},
 CellLabel->"During evaluation of In[341]:=",
 CellID->31260985,ExpressionUUID->"46888d52-80a7-4c1e-82ed-dbe11f5793e2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TorsionPDcsys"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionPDcsys],
  Editable->False]], "Print",
 CellChangeTimes->{3.72875132844075*^9, 3.728751660429084*^9, 
  3.7288433664199963`*^9, 3.733367582475669*^9, 3.792273636196095*^9},
 CellLabel->"During evaluation of In[341]:=",
 CellID->1417365027,ExpressionUUID->"9166e738-7b29-41c5-9c0f-c3d111674cb2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"parallel derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "PDcsys"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "parallel derivative", 
   " ", $CellContext`PDcsys],
  Editable->False]], "Print",
 CellChangeTimes->{3.72875132844075*^9, 3.728751660429084*^9, 
  3.7288433664199963`*^9, 3.733367582475669*^9, 3.7922736362036943`*^9},
 CellLabel->"During evaluation of In[341]:=",
 CellID->1765396265,ExpressionUUID->"ff11bf5b-4de3-43ff-a44e-0d314adadbbd"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "etaUpcsys"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric +1 density",
    " ", $CellContext`etaUpcsys],
  Editable->False]], "Print",
 CellChangeTimes->{3.72875132844075*^9, 3.728751660429084*^9, 
  3.7288433664199963`*^9, 3.733367582475669*^9, 3.7922736362115602`*^9},
 CellLabel->"During evaluation of In[341]:=",
 CellID->1380958903,ExpressionUUID->"566c7e23-7fbb-4a16-9b55-5cfb6a58fe29"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "etaDowncsys"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric -1 density",
    " ", $CellContext`etaDowncsys],
  Editable->False]], "Print",
 CellChangeTimes->{3.72875132844075*^9, 3.728751660429084*^9, 
  3.7288433664199963`*^9, 3.733367582475669*^9, 3.792273636218841*^9},
 CellLabel->"During evaluation of In[341]:=",
 CellID->1920509592,ExpressionUUID->"9e7f8367-fd62-4738-b271-57c4ba2e3d88"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
TODO: It should not be necessary to remove explicitly this object:\
\>", "MathCaption",
 CellChangeTimes->{{3.728751405164706*^9, 3.728751422690716*^9}},
 CellID->2108054115,ExpressionUUID->"9cf0465d-82eb-4df9-8609-4ea9f9b2e166"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefTensor", "[", "Jacobianbluesfred", "]"}]], "Input",
 CellChangeTimes->{{3.728751367892559*^9, 3.728751376956616*^9}},
 CellLabel->"In[342]:=",
 CellID->243573842,ExpressionUUID->"f06ecfab-57d9-4895-9908-9edb0ac82f65"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "Jacobianbluesfred"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "", 
   " ", $CellContext`Jacobianbluesfred],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751377590625*^9, 3.728751660862019*^9, 
  3.72884336827109*^9, 3.7333675827851343`*^9, 3.7922736362931843`*^9},
 CellLabel->"During evaluation of In[342]:=",
 CellID->1325498334,ExpressionUUID->"7b20dc61-927e-4deb-9215-a82c1609ec6a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"UndefBasis", "/@", 
   RowBox[{"{", 
    RowBox[{"fred", ",", "blues", ",", "fcomp"}], "}"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.7287513393578053`*^9, 3.728751349717901*^9}, 
   3.7287513952049913`*^9},
 CellLabel->"In[343]:=",
 CellID->1956939754,ExpressionUUID->"ac1c6c0e-5fe9-4853-a5cb-92b5102b8670"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "AChristoffelPDfcompPDfred\[Dagger]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`AChristoffelPDfcompPDfred\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.792273636363002*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->1802629207,ExpressionUUID->"31d76366-1b55-49cf-ac6d-118371752bf1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "AChristoffelPDfcompPDfred"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`AChristoffelPDfcompPDfred],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.7922736363827972`*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->89790173,ExpressionUUID->"4fa6e652-b96c-4216-803a-9012a310ab76"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "ChristoffelPDbluesPDfred"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`ChristoffelPDbluesPDfred],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.792273636405855*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->1569917417,ExpressionUUID->"1b9adb9d-fbe1-40dd-b6ad-897c036701ac"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "ChristoffelPDfred"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "non-symmetric Christoffel tensor", " ", $CellContext`ChristoffelPDfred],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.792273636417506*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->1617940477,ExpressionUUID->"0ad3bf7a-1460-4647-84c8-b3fd01d6b6f8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciPDfred"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Ricci tensor", 
   " ", $CellContext`RicciPDfred],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.792273636429172*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->1287658625,ExpressionUUID->"2d98467d-854e-4d5f-b0eb-022054bcd636"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannPDfred"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannPDfred],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.7922736364406643`*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->1485799556,ExpressionUUID->"4314aba7-a4c3-486a-8e01-2addb0c65a4a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TorsionPDfred"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionPDfred],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.792273636452141*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->893290376,ExpressionUUID->"a9a3ed9e-a4a1-4a50-bf13-418522c67c39"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"parallel derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "PDfred"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "parallel derivative", 
   " ", $CellContext`PDfred],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.7922736364634247`*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->473173255,ExpressionUUID->"3b27cab7-c6e8-49d6-8bf5-8a740973a1cc"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "etaUpfred"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric +1 density",
    " ", $CellContext`etaUpfred],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.792273636474634*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->490951944,ExpressionUUID->"beedab78-6764-4d35-b04d-d97a51a01560"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "etaDownfred"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric -1 density",
    " ", $CellContext`etaDownfred],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.7922736364858007`*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->797151034,ExpressionUUID->"e47ed9db-9876-4699-ba83-67944cbe107b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefBasis", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"basis\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "fred"}],
  SequenceForm[
  "** ", xAct`xCoba`UndefBasis, ": Undefined ", "basis", 
   " ", $CellContext`fred],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.792273636497175*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->1714438452,ExpressionUUID->"c6b91a09-bdfc-46e8-8760-be5c91ddb5f2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "ChristoffelPDblues"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "non-symmetric Christoffel tensor", " ", $CellContext`ChristoffelPDblues],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.792273636508377*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->1778970965,ExpressionUUID->"de2ad920-57d7-43ec-9d9c-526ba4b5f0ce"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciPDblues"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Ricci tensor", 
   " ", $CellContext`RicciPDblues],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.792273636519594*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->1897675221,ExpressionUUID->"3d5e47df-a34a-40dc-a8c3-eb69fa2f8928"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannPDblues"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannPDblues],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.792273636530871*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->1947797750,ExpressionUUID->"561ebe83-fda0-49e3-8209-d1ce5ff71b8a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TorsionPDblues"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionPDblues],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.792273636542314*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->396069382,ExpressionUUID->"ae6ecc2e-7122-493f-af1c-a921e1267319"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"parallel derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "PDblues"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "parallel derivative", 
   " ", $CellContext`PDblues],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.792273636553423*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->1686281221,ExpressionUUID->"32cbf2a8-30f2-45b0-8ff9-0a2c8eaef468"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "etaUpblues"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric +1 density",
    " ", $CellContext`etaUpblues],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.79227363656455*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->986791888,ExpressionUUID->"9c186931-f868-44fa-ae80-b5aed1c0291d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "etaDownblues"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric -1 density",
    " ", $CellContext`etaDownblues],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.792273636575844*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->2142539482,ExpressionUUID->"0782ec81-69ee-4ac3-a41a-48bfc566f1cf"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefBasis", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"basis\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "blues"}],
  SequenceForm[
  "** ", xAct`xCoba`UndefBasis, ": Undefined ", "basis", 
   " ", $CellContext`blues],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.792273636586906*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->655281078,ExpressionUUID->"895b11f0-2b36-461c-a6af-020f9aab0f77"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"nonsymmetric AChristoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "AChristoffelPDfcomp\[Dagger]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "nonsymmetric AChristoffel tensor", 
   " ", $CellContext`AChristoffelPDfcomp\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.792273636598064*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->1012936130,ExpressionUUID->"a451ace1-2b26-452a-9763-fdb595b55be1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"nonsymmetric AChristoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "AChristoffelPDfcomp"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "nonsymmetric AChristoffel tensor", " ", $CellContext`AChristoffelPDfcomp],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.7922736366093407`*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->1314667141,ExpressionUUID->"a6391ba2-4ddf-4129-9891-32c82b77f749"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "ChristoffelPDfcomp"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "non-symmetric Christoffel tensor", " ", $CellContext`ChristoffelPDfcomp],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.792273636620323*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->151394804,ExpressionUUID->"6d3ad999-44d3-45b6-beb4-a3b3bb2f4a91"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"FRiemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "FRiemannPDfcomp\[Dagger]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "FRiemann tensor", 
   " ", $CellContext`FRiemannPDfcomp\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.7922736366313267`*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->1871832780,ExpressionUUID->"5c2503e7-4ab3-4958-b3e8-9ece8900033e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"FRiemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "FRiemannPDfcomp"}],
  
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "FRiemann tensor", 
   " ", $CellContext`FRiemannPDfcomp],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.792273636642158*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->1405588557,ExpressionUUID->"c070b5b2-bf54-4286-8a3d-44642d7128f5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciPDfcomp"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Ricci tensor", 
   " ", $CellContext`RicciPDfcomp],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.7922736366533957`*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->1406760499,ExpressionUUID->"05f9f6b7-a5bd-4ba0-b358-c4c869da0a64"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannPDfcomp"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannPDfcomp],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.792273636664269*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->1776236870,ExpressionUUID->"803327b6-6b6d-4188-b626-d5da29f08034"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TorsionPDfcomp"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionPDfcomp],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.79227363667516*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->1042897143,ExpressionUUID->"fe003b6b-eaca-4424-9110-54177131e550"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"parallel derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "PDfcomp"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "parallel derivative", 
   " ", $CellContext`PDfcomp],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.792273636684978*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->210755587,ExpressionUUID->"ebe2e56b-a789-4631-99e1-dc7d8ee8cbfa"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefBasis", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"basis\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "fcomp\[Dagger]"}],
  SequenceForm[
  "** ", xAct`xCoba`UndefBasis, ": Undefined ", "basis", 
   " ", $CellContext`fcomp\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.792273636694871*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->968616806,ExpressionUUID->"5afdfb2f-a144-4024-9d1d-5707a7f74d59"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "etaUpfcomp\[Dagger]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric +1 density",
    " ", $CellContext`etaUpfcomp\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.792273636704006*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->1616614182,ExpressionUUID->"94e0e5bd-333c-4c43-9695-1e484d4db09b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "etaUpfcomp"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric +1 density",
    " ", $CellContext`etaUpfcomp],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.792273636712737*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->471455030,ExpressionUUID->"dd4c70ba-ef36-4135-bc61-9d363526874f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "etaDownfcomp\[Dagger]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric -1 density",
    " ", $CellContext`etaDownfcomp\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.792273636720975*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->1677515427,ExpressionUUID->"f43191a0-945a-4c20-968f-f5fb12ab8e7b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "etaDownfcomp"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric -1 density",
    " ", $CellContext`etaDownfcomp],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.792273636741054*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->1836463773,ExpressionUUID->"1b63a560-eb93-433a-8530-b701ede44fd1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefBasis", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"basis\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "fcomp"}],
  SequenceForm[
  "** ", xAct`xCoba`UndefBasis, ": Undefined ", "basis", 
   " ", $CellContext`fcomp],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751380221341*^9, 3.728751660977154*^9, 
  3.728843369695499*^9, 3.733367582908372*^9, 3.7922736367539053`*^9},
 CellLabel->"During evaluation of In[343]:=",
 CellID->1811658127,ExpressionUUID->"fc2fce22-f260-4425-9b6a-64a8c48ea6c6"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"UndefTensor", "/@", 
   RowBox[{"{", 
    RowBox[{
    "u", ",", "v", ",", "S", ",", "T", ",", "TT", ",", "U", ",", "J"}], 
    "}"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.728751459772893*^9, 3.7287514945274477`*^9}},
 CellLabel->"In[344]:=",
 CellID->1884771571,ExpressionUUID->"4c0d419c-480d-4aec-a43e-0b6a465060f6"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "u\[Dagger]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`u\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{{3.728751466128125*^9, 3.7287514953387737`*^9}, 
   3.728751661691924*^9, 3.728843371627293*^9, 3.733367583409745*^9, 
   3.792273636816329*^9},
 CellLabel->"During evaluation of In[344]:=",
 CellID->1514993323,ExpressionUUID->"0aee5ed0-b372-4dbf-b4cc-e7f3969e558a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "u"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`u],
  Editable->False]], "Print",
 CellChangeTimes->{{3.728751466128125*^9, 3.7287514953387737`*^9}, 
   3.728751661691924*^9, 3.728843371627293*^9, 3.733367583409745*^9, 
   3.7922736368306313`*^9},
 CellLabel->"During evaluation of In[344]:=",
 CellID->1926660829,ExpressionUUID->"1e5351e4-9807-4fb2-97fc-9f833ea36f62"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "v"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`v],
  Editable->False]], "Print",
 CellChangeTimes->{{3.728751466128125*^9, 3.7287514953387737`*^9}, 
   3.728751661691924*^9, 3.728843371627293*^9, 3.733367583409745*^9, 
   3.792273636844751*^9},
 CellLabel->"During evaluation of In[344]:=",
 CellID->1630123537,ExpressionUUID->"405d48d5-32b8-417b-88a4-0c29246bb9cc"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "S\[Dagger]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`S\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{{3.728751466128125*^9, 3.7287514953387737`*^9}, 
   3.728751661691924*^9, 3.728843371627293*^9, 3.733367583409745*^9, 
   3.7922736368546457`*^9},
 CellLabel->"During evaluation of In[344]:=",
 CellID->2037401580,ExpressionUUID->"c17cbc38-8ac1-4df5-b76f-b025d06cd6e3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "S"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`S],
  Editable->False]], "Print",
 CellChangeTimes->{{3.728751466128125*^9, 3.7287514953387737`*^9}, 
   3.728751661691924*^9, 3.728843371627293*^9, 3.733367583409745*^9, 
   3.792273636864682*^9},
 CellLabel->"During evaluation of In[344]:=",
 CellID->981603645,ExpressionUUID->"f743b2d6-27bb-49bc-b74b-9c6b7c9ea840"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{{3.728751466128125*^9, 3.7287514953387737`*^9}, 
   3.728751661691924*^9, 3.728843371627293*^9, 3.733367583409745*^9, 
   3.792273636873271*^9},
 CellLabel->"During evaluation of In[344]:=",
 CellID->843085261,ExpressionUUID->"68926355-1ed8-41c2-8727-3577f0eab204"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "TT"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`TT],
  Editable->False]], "Print",
 CellChangeTimes->{{3.728751466128125*^9, 3.7287514953387737`*^9}, 
   3.728751661691924*^9, 3.728843371627293*^9, 3.733367583409745*^9, 
   3.7922736368817787`*^9},
 CellLabel->"During evaluation of In[344]:=",
 CellID->636878721,ExpressionUUID->"3975f6c0-c659-43c0-bcec-2f4a59def999"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "U"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`U],
  Editable->False]], "Print",
 CellChangeTimes->{{3.728751466128125*^9, 3.7287514953387737`*^9}, 
   3.728751661691924*^9, 3.728843371627293*^9, 3.733367583409745*^9, 
   3.7922736368897123`*^9},
 CellLabel->"During evaluation of In[344]:=",
 CellID->962167199,ExpressionUUID->"8d2649c1-e8b4-47fd-9dd9-329c9265faa9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "J\[Dagger]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`J\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{{3.728751466128125*^9, 3.7287514953387737`*^9}, 
   3.728751661691924*^9, 3.728843371627293*^9, 3.733367583409745*^9, 
   3.792273636897428*^9},
 CellLabel->"During evaluation of In[344]:=",
 CellID->592251683,ExpressionUUID->"ff314803-898d-498a-b26a-65f10e408ea8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "J"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`J],
  Editable->False]], "Print",
 CellChangeTimes->{{3.728751466128125*^9, 3.7287514953387737`*^9}, 
   3.728751661691924*^9, 3.728843371627293*^9, 3.733367583409745*^9, 
   3.792273636905242*^9},
 CellLabel->"During evaluation of In[344]:=",
 CellID->387332336,ExpressionUUID->"a4a9349c-2d5a-43f3-8b5c-67d0d178771b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefVBundle", "[", "InnerC", "]"}]], "Input",
 CellChangeTimes->{{3.7287514281094227`*^9, 3.7287514478652687`*^9}},
 CellLabel->"In[345]:=",
 CellID->1036853606,ExpressionUUID->"671ea665-b64c-4921-b7b1-8e211f238e61"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefVBundle", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"conjugated vbundle\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["InnerC\[Dagger]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`InnerC\[Dagger],
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefVBundle, ": Undefined ", "conjugated vbundle", 
   " ", $CellContext`InnerC\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751468071309*^9, 3.728751662065617*^9, 
  3.7288433732372503`*^9, 3.733367583741948*^9, 3.7922736370366573`*^9},
 CellLabel->"During evaluation of In[345]:=",
 CellID->871637095,ExpressionUUID->"9b26f3ae-b264-470e-b857-33c466e6303e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefVBundle", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["InnerC",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`InnerC,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefVBundle, ": Undefined ", "vbundle", 
   " ", $CellContext`InnerC],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751468071309*^9, 3.728751662065617*^9, 
  3.7288433732372503`*^9, 3.733367583741948*^9, 3.792273637099872*^9},
 CellLabel->"During evaluation of In[345]:=",
 CellID->1618699478,ExpressionUUID->"af73adc1-b161-415a-a8d9-451e8784ceed"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M3", "]"}]], "Input",
 CellChangeTimes->{{3.7287515020581207`*^9, 3.728751505584279*^9}},
 CellLabel->"In[346]:=",
 CellID->2131866533,ExpressionUUID->"8d318d61-9dce-42d1-ae6d-d70f7b7729f1"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefVBundle", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]M3",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentM3,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefVBundle, ": Undefined ", "vbundle", 
   " ", $CellContext`TangentM3],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751506166246*^9, 3.728751662332601*^9, 
  3.728843374311479*^9, 3.733367584185658*^9, 3.792273637277717*^9},
 CellLabel->"During evaluation of In[346]:=",
 CellID->1159286550,ExpressionUUID->"9e3342ee-d13e-4ca3-bd13-30c7027c7991"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefManifold", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"manifold\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "M3"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefManifold, ": Undefined ", "manifold", 
   " ", $CellContext`M3],
  Editable->False]], "Print",
 CellChangeTimes->{3.728751506166246*^9, 3.728751662332601*^9, 
  3.728843374311479*^9, 3.733367584185658*^9, 3.79227363728782*^9},
 CellLabel->"During evaluation of In[346]:=",
 CellID->9916619,ExpressionUUID->"55882254-fda9-413d-bc0b-6e7a024951ca"]
}, Open  ]]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{
  "expr", ",", "expr2", ",", "expr4", ",", "cexpr", ",", "lieexpr", ",", 
   "paramexpr", ",", "traceexpr", ",", "indexlist", ",", "S", ",", "scalar", 
   ",", "u", ",", "w", ",", "\[Sigma]", ",", "other", ",", "$Chars"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.7287515177290382`*^9, 3.728751573282249*^9}},
 CellLabel->"In[347]:=",
 CellID->710107198,ExpressionUUID->"f528cfe3-6a20-4b86-af2b-20f4763fc3b1"]
}, Open  ]]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{

Cell["3. CTensor framework", "Section",
 CellFrame->{{0, 0}, {0, 0.5}},
 CellChangeTimes->{{3.7286627354220657`*^9, 3.728662748172317*^9}, {
   3.728664037492599*^9, 3.72866403885896*^9}, {3.7286641258241453`*^9, 
   3.7286641646769876`*^9}, {3.728670489894273*^9, 3.728670492830665*^9}, {
   3.7286733828242407`*^9, 3.7286733849035788`*^9}, {3.728674011059067*^9, 
   3.72867401136334*^9}, 3.728675271498732*^9},
 FontSize->18,
 FontColor->GrayLevel[0],
 CellTags->{"S5.18.1", "1.4"},
 CellID->1107613592,ExpressionUUID->"c70e460c-3a92-47d2-8cd3-6c82c14c37b0"],

Cell[CellGroupData[{

Cell["3.1. Basic ideas", "Subsection",
 CellChangeTimes->{{3.7286740191226053`*^9, 3.7286740193550167`*^9}, {
  3.728675274000371*^9, 3.728675274144375*^9}},
 CellID->946395098,ExpressionUUID->"cfab1a5a-5365-45b1-864f-ba0610d8efba"],

Cell[CellGroupData[{

Cell["3.1.1. Setup", "Subsubsection",
 CellChangeTimes->{{3.72867402065863*^9, 3.728674020882242*^9}, {
  3.733368000688815*^9, 3.733368001192829*^9}},
 CellID->1798522404,ExpressionUUID->"472e654a-d1cf-4d81-a709-584acdb0da3e"],

Cell["\<\
Here we start again from scratch, even loading xCoba` again, so that this \
section of the documentation is fully independent from the rest.\
\>", "Text",
 CellChangeTimes->{{3.72867531910424*^9, 3.728675358123992*^9}},
 CellID->1386590022,ExpressionUUID->"027b15b6-0fa6-42f9-94ca-278bdab4349b"],

Cell[CellGroupData[{

Cell["Load xCoba`, which automatically loads xTensor` and xPerm`:", \
"MathCaption",
 CellChangeTimes->{{3.7286753616124353`*^9, 3.728675373724103*^9}},
 CellID->214212395,ExpressionUUID->"80694448-4b61-4cd3-b8f6-a785cd1fd441"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<xAct`xCoba`\>\"", "]"}]], "Input",
 CellChangeTimes->{{3.7287517190049963`*^9, 3.7287517241087017`*^9}},
 CellLabel->"In[1]:=",
 CellID->1933712503,ExpressionUUID->"4f49134d-a716-4b08-9e2b-586e1bb6a149"],

Cell[CellGroupData[{

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{
  3.730955390508546*^9, 3.73095618235896*^9, 3.731114903331139*^9, 
   3.733367623601452*^9, 3.733367883463982*^9, 3.7333681160777493`*^9, 
   3.7908672650431623`*^9, 3.792273707291259*^9, {3.79451790837276*^9, 
   3.79451793318016*^9}, 3.7945180110442867`*^9, 3.794518118779554*^9, 
   3.7945187770442333`*^9, 3.794519056766591*^9, 3.794519292279009*^9, 
   3.794519992055606*^9, 3.794520177316565*^9, 3.794520311486889*^9, 
   3.794520413301036*^9, 3.794520582455724*^9, 3.794520658568473*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->2064527593,ExpressionUUID->"238bf73a-58f0-47d2-8191-1e3f9a4d6be9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`xPerm`  version \"\>", 
   "\[InvisibleSpace]", "\<\"1.2.3\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2015", ",", "8", ",", "23"}], "}"}]}],
  SequenceForm["Package xAct`xPerm`  version ", "1.2.3", ", ", {2015, 8, 23}],
  
  Editable->False]], "Print",
 CellChangeTimes->{
  3.730955390508546*^9, 3.73095618235896*^9, 3.731114903331139*^9, 
   3.733367623601452*^9, 3.733367883463982*^9, 3.7333681160777493`*^9, 
   3.7908672650431623`*^9, 3.792273707291259*^9, {3.79451790837276*^9, 
   3.79451793318016*^9}, 3.7945180110442867`*^9, 3.794518118779554*^9, 
   3.7945187770442333`*^9, 3.794519056766591*^9, 3.794519292279009*^9, 
   3.794519992055606*^9, 3.794520177316565*^9, 3.794520311486889*^9, 
   3.794520413301036*^9, 3.794520582455724*^9, 3.794520658576541*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1621970972,ExpressionUUID->"06b4e268-5bc1-4bfd-8093-6c49e848a057"],

Cell[BoxData["\<\"CopyRight (C) 2003-2020, Jose M. Martin-Garcia, under the \
General Public License.\"\>"], "Print",
 CellChangeTimes->{
  3.730955390508546*^9, 3.73095618235896*^9, 3.731114903331139*^9, 
   3.733367623601452*^9, 3.733367883463982*^9, 3.7333681160777493`*^9, 
   3.7908672650431623`*^9, 3.792273707291259*^9, {3.79451790837276*^9, 
   3.79451793318016*^9}, 3.7945180110442867`*^9, 3.794518118779554*^9, 
   3.7945187770442333`*^9, 3.794519056766591*^9, 3.794519292279009*^9, 
   3.794519992055606*^9, 3.794520177316565*^9, 3.794520311486889*^9, 
   3.794520413301036*^9, 3.794520582455724*^9, 3.7945206585948753`*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->308711386,ExpressionUUID->"2c21ad01-669e-4913-9ef7-200c07bcb297"],

Cell[BoxData["\<\"Connecting to external mac executable...\"\>"], "Print",
 CellChangeTimes->{
  3.730955390508546*^9, 3.73095618235896*^9, 3.731114903331139*^9, 
   3.733367623601452*^9, 3.733367883463982*^9, 3.7333681160777493`*^9, 
   3.7908672650431623`*^9, 3.792273707291259*^9, {3.79451790837276*^9, 
   3.79451793318016*^9}, 3.7945180110442867`*^9, 3.794518118779554*^9, 
   3.7945187770442333`*^9, 3.794519056766591*^9, 3.794519292279009*^9, 
   3.794519992055606*^9, 3.794520177316565*^9, 3.794520311486889*^9, 
   3.794520413301036*^9, 3.794520582455724*^9, 3.7945206586045313`*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->191773350,ExpressionUUID->"ebf6df46-287d-4f05-a00b-ff382be2dc77"],

Cell[BoxData["\<\"Connection established.\"\>"], "Print",
 CellChangeTimes->{
  3.730955390508546*^9, 3.73095618235896*^9, 3.731114903331139*^9, 
   3.733367623601452*^9, 3.733367883463982*^9, 3.7333681160777493`*^9, 
   3.7908672650431623`*^9, 3.792273707291259*^9, {3.79451790837276*^9, 
   3.79451793318016*^9}, 3.7945180110442867`*^9, 3.794518118779554*^9, 
   3.7945187770442333`*^9, 3.794519056766591*^9, 3.794519292279009*^9, 
   3.794519992055606*^9, 3.794520177316565*^9, 3.794520311486889*^9, 
   3.794520413301036*^9, 3.794520582455724*^9, 3.794520658643313*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1909222950,ExpressionUUID->"973b7c77-470a-4775-8ac6-b27808953bc9"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{
  3.730955390508546*^9, 3.73095618235896*^9, 3.731114903331139*^9, 
   3.733367623601452*^9, 3.733367883463982*^9, 3.7333681160777493`*^9, 
   3.7908672650431623`*^9, 3.792273707291259*^9, {3.79451790837276*^9, 
   3.79451793318016*^9}, 3.7945180110442867`*^9, 3.794518118779554*^9, 
   3.7945187770442333`*^9, 3.794519056766591*^9, 3.794519292279009*^9, 
   3.794519992055606*^9, 3.794520177316565*^9, 3.794520311486889*^9, 
   3.794520413301036*^9, 3.794520582455724*^9, 3.79452065865464*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->609467176,ExpressionUUID->"cc4276cf-3902-4566-afb8-477a66374115"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`xTensor`  version \"\>", 
   "\[InvisibleSpace]", "\<\"1.1.4\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2020", ",", "2", ",", "16"}], "}"}]}],
  SequenceForm[
  "Package xAct`xTensor`  version ", "1.1.4", ", ", {2020, 2, 16}],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.730955390508546*^9, 3.73095618235896*^9, 3.731114903331139*^9, 
   3.733367623601452*^9, 3.733367883463982*^9, 3.7333681160777493`*^9, 
   3.7908672650431623`*^9, 3.792273707291259*^9, {3.79451790837276*^9, 
   3.79451793318016*^9}, 3.7945180110442867`*^9, 3.794518118779554*^9, 
   3.7945187770442333`*^9, 3.794519056766591*^9, 3.794519292279009*^9, 
   3.794519992055606*^9, 3.794520177316565*^9, 3.794520311486889*^9, 
   3.794520413301036*^9, 3.794520582455724*^9, 3.79452065866269*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1965314489,ExpressionUUID->"becf4ddd-2a56-45d2-9ef9-06ad42303e0e"],

Cell[BoxData["\<\"CopyRight (C) 2002-2020, Jose M. Martin-Garcia, under the \
General Public License.\"\>"], "Print",
 CellChangeTimes->{
  3.730955390508546*^9, 3.73095618235896*^9, 3.731114903331139*^9, 
   3.733367623601452*^9, 3.733367883463982*^9, 3.7333681160777493`*^9, 
   3.7908672650431623`*^9, 3.792273707291259*^9, {3.79451790837276*^9, 
   3.79451793318016*^9}, 3.7945180110442867`*^9, 3.794518118779554*^9, 
   3.7945187770442333`*^9, 3.794519056766591*^9, 3.794519292279009*^9, 
   3.794519992055606*^9, 3.794520177316565*^9, 3.794520311486889*^9, 
   3.794520413301036*^9, 3.794520582455724*^9, 3.794520658669855*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->594642116,ExpressionUUID->"ee7be762-1375-4e16-b6ef-8115af8119a5"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{
  3.730955390508546*^9, 3.73095618235896*^9, 3.731114903331139*^9, 
   3.733367623601452*^9, 3.733367883463982*^9, 3.7333681160777493`*^9, 
   3.7908672650431623`*^9, 3.792273707291259*^9, {3.79451790837276*^9, 
   3.79451793318016*^9}, 3.7945180110442867`*^9, 3.794518118779554*^9, 
   3.7945187770442333`*^9, 3.794519056766591*^9, 3.794519292279009*^9, 
   3.794519992055606*^9, 3.794520177316565*^9, 3.794520311486889*^9, 
   3.794520413301036*^9, 3.794520582455724*^9, 3.794520658752675*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1908634121,ExpressionUUID->"59a9b218-0d29-445d-b8b2-bdb0b4e8f0f0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`xCoba`  version \"\>", 
   "\[InvisibleSpace]", "\<\"0.8.5\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2020", ",", "2", ",", "16"}], "}"}]}],
  SequenceForm["Package xAct`xCoba`  version ", "0.8.5", ", ", {2020, 2, 16}],
  
  Editable->False]], "Print",
 CellChangeTimes->{
  3.730955390508546*^9, 3.73095618235896*^9, 3.731114903331139*^9, 
   3.733367623601452*^9, 3.733367883463982*^9, 3.7333681160777493`*^9, 
   3.7908672650431623`*^9, 3.792273707291259*^9, {3.79451790837276*^9, 
   3.79451793318016*^9}, 3.7945180110442867`*^9, 3.794518118779554*^9, 
   3.7945187770442333`*^9, 3.794519056766591*^9, 3.794519292279009*^9, 
   3.794519992055606*^9, 3.794520177316565*^9, 3.794520311486889*^9, 
   3.794520413301036*^9, 3.794520582455724*^9, 3.794520658764826*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1450436408,ExpressionUUID->"ae4cb934-042e-4f05-9fe2-240c31aa8971"],

Cell[BoxData["\<\"CopyRight (C) 2005-2020, David Yllanes and Jose M. \
Martin-Garcia, under the General Public License.\"\>"], "Print",
 CellChangeTimes->{
  3.730955390508546*^9, 3.73095618235896*^9, 3.731114903331139*^9, 
   3.733367623601452*^9, 3.733367883463982*^9, 3.7333681160777493`*^9, 
   3.7908672650431623`*^9, 3.792273707291259*^9, {3.79451790837276*^9, 
   3.79451793318016*^9}, 3.7945180110442867`*^9, 3.794518118779554*^9, 
   3.7945187770442333`*^9, 3.794519056766591*^9, 3.794519292279009*^9, 
   3.794519992055606*^9, 3.794520177316565*^9, 3.794520311486889*^9, 
   3.794520413301036*^9, 3.794520582455724*^9, 3.7945206587736483`*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1397832159,ExpressionUUID->"7942d65e-0060-46cb-8910-619832f5857a"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{
  3.730955390508546*^9, 3.73095618235896*^9, 3.731114903331139*^9, 
   3.733367623601452*^9, 3.733367883463982*^9, 3.7333681160777493`*^9, 
   3.7908672650431623`*^9, 3.792273707291259*^9, {3.79451790837276*^9, 
   3.79451793318016*^9}, 3.7945180110442867`*^9, 3.794518118779554*^9, 
   3.7945187770442333`*^9, 3.794519056766591*^9, 3.794519292279009*^9, 
   3.794519992055606*^9, 3.794520177316565*^9, 3.794520311486889*^9, 
   3.794520413301036*^9, 3.794520582455724*^9, 3.7945206587831173`*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->2041481780,ExpressionUUID->"f5f7a00c-13a4-423d-9400-2f581a9251e3"],

Cell[BoxData["\<\"These packages come with ABSOLUTELY NO WARRANTY; for \
details type Disclaimer[]. This is free software, and you are welcome to \
redistribute it under certain conditions. See the General Public License for \
details.\"\>"], "Print",
 CellChangeTimes->{
  3.730955390508546*^9, 3.73095618235896*^9, 3.731114903331139*^9, 
   3.733367623601452*^9, 3.733367883463982*^9, 3.7333681160777493`*^9, 
   3.7908672650431623`*^9, 3.792273707291259*^9, {3.79451790837276*^9, 
   3.79451793318016*^9}, 3.7945180110442867`*^9, 3.794518118779554*^9, 
   3.7945187770442333`*^9, 3.794519056766591*^9, 3.794519292279009*^9, 
   3.794519992055606*^9, 3.794520177316565*^9, 3.794520311486889*^9, 
   3.794520413301036*^9, 3.794520582455724*^9, 3.7945206587923393`*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->838048341,ExpressionUUID->"41a801ce-9042-4abd-b585-27c286a67312"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{
  3.730955390508546*^9, 3.73095618235896*^9, 3.731114903331139*^9, 
   3.733367623601452*^9, 3.733367883463982*^9, 3.7333681160777493`*^9, 
   3.7908672650431623`*^9, 3.792273707291259*^9, {3.79451790837276*^9, 
   3.79451793318016*^9}, 3.7945180110442867`*^9, 3.794518118779554*^9, 
   3.7945187770442333`*^9, 3.794519056766591*^9, 3.794519292279009*^9, 
   3.794519992055606*^9, 3.794520177316565*^9, 3.794520311486889*^9, 
   3.794520413301036*^9, 3.794520582455724*^9, 3.794520658800932*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1880710161,ExpressionUUID->"aea99583-d203-4040-90e6-18176664c3f4"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Deactivate def/undef messages:", "MathCaption",
 CellID->250012957,ExpressionUUID->"ba4cfccf-3bea-427b-948f-b95e4af8a646"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"$PrePrint", "=", "ScreenDollarIndices"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"$DefInfoQ", "=", "False"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"$UndefInfoQ", "=", "False"}], ";"}]}], "Input",
 CellLabel->"In[2]:=",
 CellID->1489754767,ExpressionUUID->"45ba71ad-414f-42a4-96f6-e583ec3c5bdd"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Prepare a manifold:", "MathCaption",
 CellID->796528596,ExpressionUUID->"fb2335a8-373e-42b1-90b8-c948ac480124"],

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M", ",", "4", ",", 
   RowBox[{"IndexRange", "[", 
    RowBox[{"a", ",", "n"}], "]"}]}], "]"}]], "Input",
 CellLabel->"In[5]:=",
 CellID->1990541221,ExpressionUUID->"4ad7960d-7007-4692-863f-930f5a0d7f32"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
and two bases in its tangent bundle, named after their respective colors:\
\>", "MathCaption",
 CellChangeTimes->{{3.728751750147954*^9, 3.728751760915051*^9}},
 CellID->1526048381,ExpressionUUID->"e804e41b-5d36-4e6c-9caa-f12377cbd1fd"],

Cell[BoxData[
 RowBox[{"DefBasis", "[", 
  RowBox[{"red", ",", "TangentM", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2", ",", "3"}], "}"}], ",", 
   RowBox[{"BasisColor", "\[Rule]", "Red"}]}], "]"}]], "Input",
 CellLabel->"In[6]:=",
 CellID->897943346,ExpressionUUID->"327ea09f-44ec-4a84-b8be-f0f742efd06c"],

Cell[BoxData[
 RowBox[{"DefBasis", "[", 
  RowBox[{"blue", ",", "TangentM", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2", ",", "3"}], "}"}], ",", 
   RowBox[{"BasisColor", "\[Rule]", "Blue"}]}], "]"}]], "Input",
 CellLabel->"In[7]:=",
 CellID->1366228753,ExpressionUUID->"badef58d-46ec-4a80-9d47-da2d4681e4c4"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.1.2. Algebra", "Subsubsection",
 CellChangeTimes->{{3.728674021826518*^9, 3.728674022082863*^9}, {
  3.733368002544642*^9, 3.733368002680707*^9}},
 CellID->1311497947,ExpressionUUID->"9b46d7d3-8a5f-4551-986f-82ef61f7d1f9"],

Cell[TextData[{
 "The basic construction in ",
 StyleBox["xTensor",
  FontSlant->"Italic"],
 " is a tensor field of the form head[indices]. ",
 StyleBox["xCoba",
  FontSlant->"Italic"],
 " now accepts the same structure, but the head can be an expression \
containing a tensor described by an array of components in a given list of \
bases."
}], "Text",
 CellID->573842421,ExpressionUUID->"ba489325-b663-4a95-be0d-1a95805dc5ef"],

Cell[CellGroupData[{

Cell[TextData[{
 "This a vector field. An additional argument is automatically added, the ",
 StyleBox["added tensor weight",
  FontSlant->"Italic"],
 ", zero by default:"
}], "MathCaption",
 CellChangeTimes->{{3.728675446736493*^9, 3.728675447184754*^9}},
 CellID->220511070,ExpressionUUID->"dfcfc4c4-9698-43b3-a3c3-8a540aaf0ec7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"v", "=", 
  RowBox[{"CTensor", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"1", ",", "2", ",", "3", ",", "4"}], "}"}], ",", 
    RowBox[{"{", "red", "}"}]}], "]"}]}]], "Input",
 CellLabel->"In[8]:=",
 CellID->1720262415,ExpressionUUID->"bb010995-42b4-4f80-9e4f-bafffbc77444"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "2", ",", "3", ",", "4"}], "}"}], ",", 
   RowBox[{"{", "red", "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{
  3.728675450523308*^9, 3.7287534552349167`*^9, 3.72884344368132*^9, 
   3.730955413382681*^9, 3.730956190878937*^9, 3.7311149415007067`*^9, 
   3.733367636356691*^9, 3.7333678843166323`*^9, 3.733368117127791*^9, 
   3.790867304142208*^9, 3.792273721203518*^9, {3.794517909696333*^9, 
   3.794517933931875*^9}, 3.794518012532152*^9, 3.794518119971178*^9, 
   3.794518778162059*^9, 3.7945190579479427`*^9, 3.794519293529892*^9, 
   3.7945200031615877`*^9, 3.7945201780857162`*^9, 3.7945203127906322`*^9, 
   3.794520414574726*^9, 3.794520583672452*^9, 3.794520659830744*^9},
 CellLabel->"Out[8]=",
 CellID->283875734,ExpressionUUID->"e4ba63b3-8498-4684-a5e6-b7f721134bf9"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"xTensorQ", "[", "v", "]"}]], "Input",
 CellLabel->"In[9]:=",
 CellID->203842424,ExpressionUUID->"f2682149-1957-4a40-a01b-88c5b9f66701"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{
  3.728675451807439*^9, 3.728753455478801*^9, 3.7288434439251328`*^9, 
   3.7309554154018993`*^9, 3.730956190981359*^9, 3.731114943801209*^9, 
   3.7333676365183773`*^9, 3.733367884429693*^9, 3.733368117260767*^9, 
   3.790867306725608*^9, 3.792273722195887*^9, {3.794517909746091*^9, 
   3.794517933985223*^9}, 3.794518013269021*^9, 3.794518120021016*^9, 
   3.794518778207672*^9, 3.7945190579954653`*^9, 3.794519293579266*^9, 
   3.7945200032389174`*^9, 3.794520178136207*^9, 3.7945203128403482`*^9, 
   3.7945204146221333`*^9, 3.794520583721245*^9, 3.794520659880026*^9},
 CellLabel->"Out[9]=",
 CellID->1538281351,ExpressionUUID->"19618354-fc1a-4c08-ad4b-ba3144d1e511"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "By default the object is not formatted. It is typeset when we add the \
index, as usual in ",
 StyleBox["xTensor",
  FontSlant->"Italic"],
 ":"
}], "MathCaption",
 CellChangeTimes->{{3.728675457815989*^9, 3.728675457952219*^9}},
 CellID->1251341869,ExpressionUUID->"550badcb-df8b-483d-95cf-f86f2f3e17c1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"v", "[", "a", "]"}]], "Input",
 CellLabel->"In[10]:=",
 CellID->995437212,ExpressionUUID->"369f4b69-df35-47ed-afca-7d62dc9483b4"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", "1"}]],
           Annotation[#, {1} -> 1, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["2",
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", "2"}]],
           Annotation[#, {2} -> 2, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["3",
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", "3"}]],
           Annotation[#, {3} -> 3, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["4",
            RowBox[{
              RowBox[{"{", "4", "}"}], "\[Rule]", "4"}]],
           Annotation[#, {4} -> 4, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{1, 2, 3, 4}, {$CellContext`red}, 0][$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.728675459980867*^9, 3.7287534556917257`*^9, 3.7288434442428713`*^9, 
   3.730955421350073*^9, 3.7309561910950203`*^9, 3.731114950324847*^9, 
   3.733367636635314*^9, 3.733367884565138*^9, 3.733368117461815*^9, 
   3.790867312846445*^9, 3.7922737233229856`*^9, {3.794517909817258*^9, 
   3.794517934054654*^9}, 3.794518013337923*^9, 3.794518120074597*^9, 
   3.794518778259378*^9, 3.794519058058423*^9, 3.7945192936378508`*^9, 
   3.7945200032984533`*^9, 3.794520178211232*^9, 3.794520312905629*^9, 
   3.79452041468548*^9, 3.7945205837957478`*^9, 3.794520659944921*^9},
 CellLabel->"Out[10]=",
 CellID->339040024,ExpressionUUID->"57650730-1273-43ad-a0b8-ca4c4a2ca047"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The red vertical bars indicate that components correspond to the red basis, \
whose color is Red:\
\>", "MathCaption",
 CellID->218785765,ExpressionUUID->"82388376-0b05-4bde-9135-6b747fb6cb4e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"BasisColor", "[", "red", "]"}]], "Input",
 CellLabel->"In[11]:=",
 CellID->640348691,ExpressionUUID->"364cddd3-42df-4ac6-aa5b-3ee5b54c2668"],

Cell[BoxData[
 InterpretationBox[
  ButtonBox[
   TooltipBox[
    GraphicsBox[{
      {GrayLevel[0], RectangleBox[{0, 0}]}, 
      {GrayLevel[0], RectangleBox[{1, -1}]}, 
      {RGBColor[1, 0, 0], RectangleBox[{0, -1}, {2, 1}]}},
     AspectRatio->1,
     DefaultBaseStyle->"ColorSwatchGraphics",
     Frame->True,
     FrameStyle->RGBColor[0.6666666666666666, 0., 0.],
     FrameTicks->None,
     ImageSize->
      Dynamic[{Automatic, 1.35 CurrentValue["FontCapHeight"]/
        AbsoluteCurrentValue[Magnification]}],
     PlotRangePadding->None],
    StyleBox[
     RowBox[{"RGBColor", "[", 
       RowBox[{"1", ",", "0", ",", "0"}], "]"}], NumberMarks -> False]],
   Appearance->None,
   BaseStyle->{},
   BaselinePosition->Baseline,
   ButtonFunction:>With[{Typeset`box$ = EvaluationBox[]}, 
     If[
      Not[
       AbsoluteCurrentValue["Deployed"]], 
      SelectionMove[Typeset`box$, All, Expression]; 
      FrontEnd`Private`$ColorSelectorInitialAlpha = 1; 
      FrontEnd`Private`$ColorSelectorInitialColor = RGBColor[1, 0, 0]; 
      FrontEnd`Private`$ColorSelectorUseMakeBoxes = True; 
      MathLink`CallFrontEnd[
        FrontEnd`AttachCell[Typeset`box$, 
         FrontEndResource["RGBColorValueSelector"], {0, {Left, Bottom}}, {
         Left, Top}, 
         "ClosingActions" -> {
          "SelectionDeparture", "ParentChanged", "EvaluatorQuit"}]]]],
   DefaultBaseStyle->{},
   Evaluator->Automatic,
   Method->"Preemptive"],
  RGBColor[1, 0, 0],
  Editable->False,
  Selectable->False]], "Output",
 CellChangeTimes->{
  3.728675466528844*^9, 3.7287534558583727`*^9, 3.728843444713129*^9, 
   3.730955423907007*^9, 3.7309561912095633`*^9, 3.731114959750484*^9, 
   3.733367636794379*^9, 3.733367884700239*^9, 3.73336811916654*^9, 
   3.790867317899048*^9, 3.79227372441889*^9, {3.794517909895342*^9, 
   3.794517934123983*^9}, 3.7945180134183817`*^9, 3.794518120140326*^9, 
   3.794518778333714*^9, 3.794519058133575*^9, 3.794519293712459*^9, 
   3.794520003371098*^9, 3.794520178287047*^9, 3.794520312983708*^9, 
   3.7945204147635813`*^9, 3.794520583870164*^9, 3.794520660019969*^9},
 CellLabel->"Out[11]=",
 CellID->971567877,ExpressionUUID->"86145cdf-2d12-403c-987f-705e1f7bdf4a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["This is a 2-tensor:", "MathCaption",
 CellChangeTimes->{{3.7286754716632023`*^9, 3.7286754718154984`*^9}},
 CellID->976056709,ExpressionUUID->"1a1231b2-1997-466a-871c-b6bc12482ad7"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"SeedRandom", "[", "100", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"S", "=", 
  RowBox[{"CTensor", "[", 
   RowBox[{
    RowBox[{"RandomInteger", "[", 
     RowBox[{"10", ",", 
      RowBox[{"{", 
       RowBox[{"4", ",", "4"}], "}"}]}], "]"}], ",", 
    RowBox[{"{", 
     RowBox[{"red", ",", " ", 
      RowBox[{"-", "red"}]}], "}"}]}], "]"}]}]}], "Input",
 CellLabel->"In[12]:=",
 CellID->2105012377,ExpressionUUID->"ee75c62f-41e4-40c7-a7d0-3058eb522f64"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"0", ",", "3", ",", "4", ",", "7"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"3", ",", "10", ",", "1", ",", "3"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"4", ",", "5", ",", "6", ",", "2"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", "8", ",", "9", ",", "10"}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"red", ",", 
     RowBox[{"-", "red"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{
  3.728675473112475*^9, 3.728753456000218*^9, 3.728843444939745*^9, 
   3.730955426413097*^9, 3.730956191314999*^9, 3.731114973148767*^9, 
   3.7333676369354277`*^9, 3.733367884832323*^9, 3.733368119345962*^9, 
   3.790867325560615*^9, 3.792273725706457*^9, {3.794517910629974*^9, 
   3.794517934170635*^9}, 3.794518013469898*^9, 3.794518120185239*^9, 
   3.7945187783766212`*^9, 3.794519058180771*^9, 3.794519293754797*^9, 
   3.7945200034117413`*^9, 3.7945201783351*^9, 3.794520313729801*^9, 
   3.794520414939473*^9, 3.794520583921122*^9, 3.7945206601958523`*^9},
 CellLabel->"Out[13]=",
 CellID->1823140598,ExpressionUUID->"0d40a737-723d-4c4c-a389-faa57a8fddcc"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"S", "[", 
  RowBox[{"a", ",", 
   RowBox[{"-", "b"}]}], "]"}]], "Input",
 CellLabel->"In[14]:=",
 CellID->181610089,ExpressionUUID->"9bae844a-c14c-465f-88ca-1865faf2359c"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["3",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "3"}]],
           Annotation[#, {1, 2} -> 3, "Tooltip"]& ], 
          TagBox[
           TooltipBox["4",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", "4"}]],
           Annotation[#, {1, 3} -> 4, "Tooltip"]& ], 
          TagBox[
           TooltipBox["7",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "4"}], "}"}], "\[Rule]", "7"}]],
           Annotation[#, {1, 4} -> 7, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["3",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "3"}]],
           Annotation[#, {2, 1} -> 3, "Tooltip"]& ], 
          TagBox[
           TooltipBox["10",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", "10"}]],
           Annotation[#, {2, 2} -> 10, "Tooltip"]& ], 
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {2, 3} -> 1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["3",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "4"}], "}"}], "\[Rule]", "3"}]],
           Annotation[#, {2, 4} -> 3, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["4",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "4"}]],
           Annotation[#, {3, 1} -> 4, "Tooltip"]& ], 
          TagBox[
           TooltipBox["5",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", "5"}]],
           Annotation[#, {3, 2} -> 5, "Tooltip"]& ], 
          TagBox[
           TooltipBox["6",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", "6"}]],
           Annotation[#, {3, 3} -> 6, "Tooltip"]& ], 
          TagBox[
           TooltipBox["2",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "4"}], "}"}], "\[Rule]", "2"}]],
           Annotation[#, {3, 4} -> 2, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "1"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {4, 1} -> 1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["8",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "2"}], "}"}], "\[Rule]", "8"}]],
           Annotation[#, {4, 2} -> 8, "Tooltip"]& ], 
          TagBox[
           TooltipBox["9",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "3"}], "}"}], "\[Rule]", "9"}]],
           Annotation[#, {4, 3} -> 9, "Tooltip"]& ], 
          TagBox[
           TooltipBox["10",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "4"}], "}"}], "\[Rule]", "10"}]],
           Annotation[#, {4, 4} -> 10, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a", " "},
         {" ", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{0, 3, 4, 7}, {3, 10, 1, 3}, {4, 5, 6, 2}, {1, 8, 9, 
   10}}, {$CellContext`red, -$CellContext`red}, 
   0][$CellContext`a, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.7286754741589403`*^9, 3.728753456142952*^9, 3.728843445158709*^9, 
   3.730955427476548*^9, 3.730956191425696*^9, 3.7311149766012363`*^9, 
   3.73336763706809*^9, 3.733367886730303*^9, 3.7333681194618807`*^9, 
   3.79086732695156*^9, 3.792273726616872*^9, {3.794517910680908*^9, 
   3.794517934219431*^9}, 3.794518013523616*^9, 3.794518120239338*^9, 
   3.794518778424054*^9, 3.794519058230049*^9, 3.7945192944996243`*^9, 
   3.79452000345651*^9, 3.794520178389374*^9, 3.794520313784369*^9, 
   3.794520415546238*^9, 3.794520584696513*^9, 3.794520660795899*^9},
 CellLabel->"Out[14]=",
 CellID->1227355190,ExpressionUUID->"0691241f-6c09-402f-a8de-aa49f903c14f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Note that the standard representation of a matrix {{1}, {2}} and a vector {1, \
2} coincide in MatrixForm, but not in this CTensor formatting:\
\>", "MathCaption",
 CellID->883783078,ExpressionUUID->"ff66743e-e05c-43ac-921f-ce20d5926555"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"MatrixForm", "[", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", "1", "}"}], ",", 
      RowBox[{"{", "2", "}"}]}], "}"}], "]"}], ",", 
   RowBox[{"MatrixForm", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "2"}], "}"}], "]"}]}], "}"}]], "Input",
 CellLabel->"In[15]:=",
 CellID->1745289294,ExpressionUUID->"06197b6d-cc5b-4a64-a46d-c7ed6b9eca4c"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   TagBox[
    RowBox[{"(", "\[NoBreak]", GridBox[{
       {"1"},
       {"2"}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Baseline}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.7]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}], "\[NoBreak]", ")"}],
    Function[BoxForm`e$, 
     MatrixForm[BoxForm`e$]]], ",", 
   TagBox[
    RowBox[{"(", "\[NoBreak]", 
     TagBox[GridBox[{
        {"1"},
        {"2"}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Baseline}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.5599999999999999]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      Column], "\[NoBreak]", ")"}],
    Function[BoxForm`e$, 
     MatrixForm[BoxForm`e$]]]}], "}"}]], "Output",
 CellChangeTimes->{
  3.728675481357127*^9, 3.728753456284071*^9, 3.7288434453594*^9, 
   3.730955429313015*^9, 3.7309561915321913`*^9, 3.731114996931541*^9, 
   3.733367637203269*^9, 3.733367886896556*^9, 3.733368119545128*^9, 
   3.790867336741935*^9, 3.792273728299842*^9, {3.7945179107265577`*^9, 
   3.794517934264278*^9}, 3.7945180135712357`*^9, 3.7945181204239492`*^9, 
   3.7945187784665213`*^9, 3.79451905840913*^9, 3.794519294544176*^9, 
   3.7945200034972477`*^9, 3.7945201784377203`*^9, 3.7945203138331947`*^9, 
   3.794520415586546*^9, 3.7945205847450542`*^9, 3.794520660845091*^9},
 CellLabel->"Out[15]=",
 CellID->1578248895,ExpressionUUID->"9fdf83f2-ec29-4867-9a7f-a6822de4b643"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"CTensor", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"{", "1", "}"}], ",", 
        RowBox[{"{", "2", "}"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"red", ",", "blue"}], "}"}]}], "]"}], "[", 
    RowBox[{"a", ",", "b"}], "]"}], ",", 
   RowBox[{
    RowBox[{"CTensor", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"1", ",", "2"}], "}"}], ",", 
      RowBox[{"{", "red", "}"}]}], "]"}], "[", "a", "]"}]}], "}"}]], "Input",
 CellLabel->"In[16]:=",
 CellID->864809235,ExpressionUUID->"3c8b5549-4c31-4254-a61d-cbc6895a198b"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        StyleBox[GridBox[{
           {
            TagBox[
             TooltipBox["1",
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "1"}]],
             Annotation[#, {1, 1} -> 1, "Tooltip"]& ]},
           {
            TagBox[
             TooltipBox["2",
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "2"}]],
             Annotation[#, {2, 1} -> 2, "Tooltip"]& ]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[0, 0, 1], 3 -> RGBColor[0, 0, 1]}}],
         FontSize->Small], 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xCoba`CTensor[{{1}, {2}}, {$CellContext`red, $CellContext`blue}, 
     0][$CellContext`a, $CellContext`b],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        StyleBox[GridBox[{
           {
            TagBox[
             TooltipBox["1",
              RowBox[{
                RowBox[{"{", "1", "}"}], "\[Rule]", "1"}]],
             Annotation[#, {1} -> 1, "Tooltip"]& ]},
           {
            TagBox[
             TooltipBox["2",
              RowBox[{
                RowBox[{"{", "2", "}"}], "\[Rule]", "2"}]],
             Annotation[#, {2} -> 2, "Tooltip"]& ]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
         FontSize->Small], 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xCoba`CTensor[{1, 2}, {$CellContext`red}, 0][$CellContext`a],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{
  3.7286754822877407`*^9, 3.7287534564887047`*^9, 3.728843446018837*^9, 
   3.730955430040996*^9, 3.730956191661071*^9, 3.7311150013570433`*^9, 
   3.733367637294586*^9, 3.733367887013174*^9, 3.733368119627675*^9, 
   3.790867339130694*^9, 3.792273728940489*^9, {3.7945179107776613`*^9, 
   3.794517934319145*^9}, 3.794518013697073*^9, 3.794518121046014*^9, 
   3.7945187786506577`*^9, 3.794519059015482*^9, 3.794519294596849*^9, 
   3.794520003542235*^9, 3.794520178497678*^9, 3.794520313885997*^9, 
   3.794520415633973*^9, 3.794520584796447*^9, 3.794520660901834*^9},
 CellLabel->"Out[16]=",
 CellID->2058654168,ExpressionUUID->"781377a5-b169-4d13-a56a-88b51089b5bd"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The vertical bars correspond to the first index and the horizontal bars \
correspond to the second index. If more indices are present then we nest the \
colored bars, as MatrixForm does. Use the variable $CTensorRepresentation to \
switch between typesetting forms:\
\>", "MathCaption",
 CellID->2013660344,ExpressionUUID->"8d332b9e-5919-446c-8474-ae15637e990c"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"$CTensorRepresentation", "=", "\"\<MatrixForm\>\""}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"CTensor", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"{", "1", "}"}], ",", 
        RowBox[{"{", "2", "}"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"{", "3", "}"}], ",", 
        RowBox[{"{", "4", "}"}]}], "}"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{"red", ",", "red", ",", "blue"}], "}"}]}], "]"}], "[", 
  RowBox[{"a", ",", "b", ",", "c"}], "]"}]}], "Input",
 CellLabel->"In[17]:=",
 CellID->1394608535,ExpressionUUID->"a876c047-9e06-4ecf-8305-a95f794dffe3"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[
       TagBox[
        RowBox[{"(", "\[NoBreak]", GridBox[{
           {
            RowBox[{"(", "\[NoBreak]", 
             TagBox[GridBox[{
                {
                 TagBox[
                  TooltipBox["1",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    "1"}]],
                  Annotation[#, {1, 1, 1} -> 1, "Tooltip"]& ]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Baseline}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.5599999999999999]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              Column], "\[NoBreak]", ")"}], 
            RowBox[{"(", "\[NoBreak]", 
             TagBox[GridBox[{
                {
                 TagBox[
                  TooltipBox["2",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    "2"}]],
                  Annotation[#, {1, 2, 1} -> 2, "Tooltip"]& ]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Baseline}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.5599999999999999]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              Column], "\[NoBreak]", ")"}]},
           {
            RowBox[{"(", "\[NoBreak]", 
             TagBox[GridBox[{
                {
                 TagBox[
                  TooltipBox["3",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    "3"}]],
                  Annotation[#, {2, 1, 1} -> 3, "Tooltip"]& ]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Baseline}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.5599999999999999]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              Column], "\[NoBreak]", ")"}], 
            RowBox[{"(", "\[NoBreak]", 
             TagBox[GridBox[{
                {
                 TagBox[
                  TooltipBox["4",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    "4"}]],
                  Annotation[#, {2, 2, 1} -> 4, "Tooltip"]& ]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Baseline}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.5599999999999999]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              Column], "\[NoBreak]", ")"}]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Baseline}}},
          
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.7]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}], "\[NoBreak]", ")"}],
        Function[BoxForm`e$, 
         MatrixForm[BoxForm`e$]]],
       StripOnInput->False,
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a", "b", "c"},
         {" ", " ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{{1}, {2}}, {{3}, {
   4}}}, {$CellContext`red, $CellContext`red, $CellContext`blue}, 
   0][$CellContext`a, $CellContext`b, $CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.728675489593014*^9, 3.728753456621922*^9, 3.7288434463594637`*^9, 
   3.730955432696208*^9, 3.730956191766995*^9, 3.731115025409672*^9, 
   3.733367637525769*^9, 3.7333678871473227`*^9, 3.733368119745957*^9, 
   3.790867358571974*^9, 3.792273730822054*^9, {3.794517910827959*^9, 
   3.794517934365835*^9}, 3.794518013752194*^9, 3.794518121105689*^9, 
   3.7945187792587013`*^9, 3.794519059081307*^9, 3.7945192946423063`*^9, 
   3.7945200035883293`*^9, 3.794520178550956*^9, 3.794520313948731*^9, 
   3.7945204156800737`*^9, 3.794520584854933*^9, 3.79452066095367*^9},
 CellLabel->"Out[18]=",
 CellID->1691564759,ExpressionUUID->"f41a813f-75f2-41c0-b18e-8e6994936e81"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"$CTensorRepresentation", "=", "\"\<ColorBasisForm\>\""}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"CTensor", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"{", "1", "}"}], ",", 
        RowBox[{"{", "2", "}"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"{", "3", "}"}], ",", 
        RowBox[{"{", "4", "}"}]}], "}"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{"red", ",", "red", ",", "blue"}], "}"}]}], "]"}], "[", 
  RowBox[{"a", ",", "b", ",", "c"}], "]"}]}], "Input",
 CellLabel->"In[19]:=",
 CellID->41836081,ExpressionUUID->"90abb9a3-0c97-46ef-9ffa-c83640251cad"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {GridBox[{
            {
             TagBox[
              TooltipBox["1",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", "1"}]],
              Annotation[#, {1, 1, 1} -> 1, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 2 -> RGBColor[0, 0, 1]}}], GridBox[{
            {
             TagBox[
              TooltipBox["2",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", "2"}]],
              Annotation[#, {1, 2, 1} -> 2, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 2 -> RGBColor[0, 0, 1]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["3",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", "3"}]],
              Annotation[#, {2, 1, 1} -> 3, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 2 -> RGBColor[0, 0, 1]}}], GridBox[{
            {
             TagBox[
              TooltipBox["4",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", "4"}]],
              Annotation[#, {2, 2, 1} -> 4, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 2 -> RGBColor[0, 0, 1]}}]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a", "b", "c"},
         {" ", " ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{{1}, {2}}, {{3}, {
   4}}}, {$CellContext`red, $CellContext`red, $CellContext`blue}, 
   0][$CellContext`a, $CellContext`b, $CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.7286754906574097`*^9, 3.728753456760559*^9, 3.728843446630731*^9, 
   3.730955433813099*^9, 3.73095619189356*^9, 3.73111502876307*^9, 
   3.733367637975149*^9, 3.733367887279882*^9, 3.733368119862097*^9, 
   3.7908673623328047`*^9, 3.792273732171048*^9, {3.7945179108673058`*^9, 
   3.794517934417447*^9}, 3.794518013806497*^9, 3.794518121156024*^9, 
   3.794518779312386*^9, 3.7945190591321173`*^9, 3.794519294693118*^9, 
   3.794520003633067*^9, 3.794520178600793*^9, 3.794520313994548*^9, 
   3.7945204157237873`*^9, 3.794520584906959*^9, 3.794520661008959*^9},
 CellLabel->"Out[20]=",
 CellID->1740385527,ExpressionUUID->"0368f66e-0e1b-444c-be15-14ad4e1feb44"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "Operations with c-tensors proceed as usual in ",
 StyleBox["xAct",
  FontSlant->"Italic"],
 ", using indices to indicate contractions or transpositions."
}], "Text",
 CellID->1201126994,ExpressionUUID->"53a71cf1-904b-4dc7-8010-0f08167ca4b5"],

Cell[CellGroupData[{

Cell["Compute the trace of a tensor:", "MathCaption",
 CellChangeTimes->{{3.7286754985018787`*^9, 3.728675498678266*^9}},
 CellID->995106358,ExpressionUUID->"7852e2f9-ab1f-4064-83e2-5ff9c0c028ae"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"S", "[", 
  RowBox[{"a", ",", 
   RowBox[{"-", "a"}]}], "]"}]], "Input",
 CellLabel->"In[21]:=",
 CellID->149916070,ExpressionUUID->"a67eab13-4a6b-409a-a45a-12babe0aa5c3"],

Cell[BoxData["26"], "Output",
 CellChangeTimes->{
  3.728675499762279*^9, 3.728753456909568*^9, 3.728843446995265*^9, 
   3.730955436319991*^9, 3.730956192080285*^9, 3.7311150403180513`*^9, 
   3.7333676381853867`*^9, 3.7333678873806143`*^9, 3.733368119944961*^9, 
   3.790867373928945*^9, 3.79227373710493*^9, {3.794517910907523*^9, 
   3.794517934461644*^9}, 3.794518013858364*^9, 3.794518121203272*^9, 
   3.7945187793546267`*^9, 3.794519059179193*^9, 3.7945192947345448`*^9, 
   3.79452000367454*^9, 3.794520178649091*^9, 3.79452031409522*^9, 
   3.794520415767412*^9, 3.794520584957103*^9, 3.794520661054936*^9},
 CellLabel->"Out[21]=",
 CellID->1932280149,ExpressionUUID->"67f4a88b-b26b-4e10-bb6d-3709e494b8bf"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Contract two tensors:", "MathCaption",
 CellChangeTimes->{{3.728675504557749*^9, 3.728675504701726*^9}},
 CellID->547428114,ExpressionUUID->"8dde0237-1f17-4f2f-86a5-39612bfa4cd7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"S", "[", 
   RowBox[{"a", ",", 
    RowBox[{"-", "b"}]}], "]"}], 
  RowBox[{"v", "[", "b", "]"}]}]], "Input",
 CellLabel->"In[22]:=",
 CellID->804368250,ExpressionUUID->"9f49d0df-c75f-48b9-8b60-273887d52513"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["46",
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", "46"}]],
           Annotation[#, {1} -> 46, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["38",
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", "38"}]],
           Annotation[#, {2} -> 38, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["40",
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", "40"}]],
           Annotation[#, {3} -> 40, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["84",
            RowBox[{
              RowBox[{"{", "4", "}"}], "\[Rule]", "84"}]],
           Annotation[#, {4} -> 84, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{46, 38, 40, 84}, {$CellContext`red}, 0][$CellContext`a],
  
  Editable->False]], "Output",
 CellChangeTimes->{
  3.7286755058378077`*^9, 3.72875345706614*^9, 3.728843447735674*^9, 
   3.730955437444701*^9, 3.7309561922612047`*^9, 3.731115044316956*^9, 
   3.733367638394793*^9, 3.733367887512847*^9, 3.733368120061904*^9, 
   3.790867376840082*^9, 3.7922737399376793`*^9, {3.794517910953548*^9, 
   3.7945179345157948`*^9}, 3.794518013908505*^9, 3.794518121250146*^9, 
   3.794518779402111*^9, 3.794519059229909*^9, 3.79451929477565*^9, 
   3.794520003720148*^9, 3.794520178702695*^9, 3.794520314144329*^9, 
   3.794520415811129*^9, 3.794520585058463*^9, 3.7945206611626*^9},
 CellLabel->"Out[22]=",
 CellID->1584754309,ExpressionUUID->"933eb438-a6f8-4ecb-b412-2529e3835d47"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Multiply two tensors (their tensor product):", "MathCaption",
 CellChangeTimes->{{3.728675510133267*^9, 3.7286755103016443`*^9}},
 CellID->982665710,ExpressionUUID->"df974585-84f4-45e0-99d6-73c491c73ac9"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"T", "[", 
   RowBox[{"a_", ",", "b_"}], "]"}], "=", 
  RowBox[{
   RowBox[{"v", "[", "a", "]"}], 
   RowBox[{"v", "[", "b", "]"}]}]}]], "Input",
 CellLabel->"In[23]:=",
 CellID->1513342540,ExpressionUUID->"aee76edd-edf5-4fcf-9c9c-a5d705996605"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {1, 1} -> 1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["2",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "2"}]],
           Annotation[#, {1, 2} -> 2, "Tooltip"]& ], 
          TagBox[
           TooltipBox["3",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", "3"}]],
           Annotation[#, {1, 3} -> 3, "Tooltip"]& ], 
          TagBox[
           TooltipBox["4",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "4"}], "}"}], "\[Rule]", "4"}]],
           Annotation[#, {1, 4} -> 4, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["2",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "2"}]],
           Annotation[#, {2, 1} -> 2, "Tooltip"]& ], 
          TagBox[
           TooltipBox["4",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", "4"}]],
           Annotation[#, {2, 2} -> 4, "Tooltip"]& ], 
          TagBox[
           TooltipBox["6",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", "6"}]],
           Annotation[#, {2, 3} -> 6, "Tooltip"]& ], 
          TagBox[
           TooltipBox["8",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "4"}], "}"}], "\[Rule]", "8"}]],
           Annotation[#, {2, 4} -> 8, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["3",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "3"}]],
           Annotation[#, {3, 1} -> 3, "Tooltip"]& ], 
          TagBox[
           TooltipBox["6",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", "6"}]],
           Annotation[#, {3, 2} -> 6, "Tooltip"]& ], 
          TagBox[
           TooltipBox["9",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", "9"}]],
           Annotation[#, {3, 3} -> 9, "Tooltip"]& ], 
          TagBox[
           TooltipBox["12",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "4"}], "}"}], "\[Rule]", "12"}]],
           Annotation[#, {3, 4} -> 12, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["4",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "1"}], "}"}], "\[Rule]", "4"}]],
           Annotation[#, {4, 1} -> 4, "Tooltip"]& ], 
          TagBox[
           TooltipBox["8",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "2"}], "}"}], "\[Rule]", "8"}]],
           Annotation[#, {4, 2} -> 8, "Tooltip"]& ], 
          TagBox[
           TooltipBox["12",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "3"}], "}"}], "\[Rule]", "12"}]],
           Annotation[#, {4, 3} -> 12, "Tooltip"]& ], 
          TagBox[
           TooltipBox["16",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "4"}], "}"}], "\[Rule]", "16"}]],
           Annotation[#, {4, 4} -> 16, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a", "b"},
         {" ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{1, 2, 3, 4}, {2, 4, 6, 8}, {3, 6, 9, 12}, {4, 8, 12, 
   16}}, {$CellContext`red, $CellContext`red}, 
   0][$CellContext`a, $CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.728675512625367*^9, 3.728753457182906*^9, 3.7288434480291023`*^9, 
   3.730955438608396*^9, 3.730956192366589*^9, 3.731115047297064*^9, 
   3.733367638597386*^9, 3.733367887679399*^9, 3.73336812016247*^9, 
   3.7908673854339046`*^9, 3.792273742777335*^9, {3.794517911057316*^9, 
   3.794517934566642*^9}, 3.794518013961843*^9, 3.7945181213025227`*^9, 
   3.794518779450006*^9, 3.7945190592797413`*^9, 3.794519294906025*^9, 
   3.794520003766335*^9, 3.794520178758417*^9, 3.794520314196259*^9, 
   3.794520415907526*^9, 3.794520585108676*^9, 3.794520661218052*^9},
 CellLabel->"Out[23]=",
 CellID->1873481269,ExpressionUUID->"ab6c91cd-e15e-4611-b7fb-d012567b613c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["This is a symmetric tensor:", "MathCaption",
 CellChangeTimes->{{3.7286755281644793`*^9, 3.728675528332745*^9}},
 CellID->890263150,ExpressionUUID->"80575164-8a5b-4a91-8827-7f2e3f950d02"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"T", "[", 
   RowBox[{"a", ",", "b"}], "]"}], "\[Equal]", 
  RowBox[{"T", "[", 
   RowBox[{"b", ",", "a"}], "]"}]}]], "Input",
 CellLabel->"In[24]:=",
 CellID->1134193401,ExpressionUUID->"4bc57f11-a773-40dd-86d9-811d78347615"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       StyleBox[GridBox[{
          {
           TagBox[
            TooltipBox["1",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "1"}]],
            Annotation[#, {1, 1} -> 1, "Tooltip"]& ], 
           TagBox[
            TooltipBox["2",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "2"}]],
            Annotation[#, {1, 2} -> 2, "Tooltip"]& ], 
           TagBox[
            TooltipBox["3",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", "3"}]],
            Annotation[#, {1, 3} -> 3, "Tooltip"]& ], 
           TagBox[
            TooltipBox["4",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"1", ",", "4"}], "}"}], "\[Rule]", "4"}]],
            Annotation[#, {1, 4} -> 4, "Tooltip"]& ]},
          {
           TagBox[
            TooltipBox["2",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "2"}]],
            Annotation[#, {2, 1} -> 2, "Tooltip"]& ], 
           TagBox[
            TooltipBox["4",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", "4"}]],
            Annotation[#, {2, 2} -> 4, "Tooltip"]& ], 
           TagBox[
            TooltipBox["6",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", "6"}]],
            Annotation[#, {2, 3} -> 6, "Tooltip"]& ], 
           TagBox[
            TooltipBox["8",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"2", ",", "4"}], "}"}], "\[Rule]", "8"}]],
            Annotation[#, {2, 4} -> 8, "Tooltip"]& ]},
          {
           TagBox[
            TooltipBox["3",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "3"}]],
            Annotation[#, {3, 1} -> 3, "Tooltip"]& ], 
           TagBox[
            TooltipBox["6",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", "6"}]],
            Annotation[#, {3, 2} -> 6, "Tooltip"]& ], 
           TagBox[
            TooltipBox["9",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", "9"}]],
            Annotation[#, {3, 3} -> 9, "Tooltip"]& ], 
           TagBox[
            TooltipBox["12",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"3", ",", "4"}], "}"}], "\[Rule]", "12"}]],
            Annotation[#, {3, 4} -> 12, "Tooltip"]& ]},
          {
           TagBox[
            TooltipBox["4",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"4", ",", "1"}], "}"}], "\[Rule]", "4"}]],
            Annotation[#, {4, 1} -> 4, "Tooltip"]& ], 
           TagBox[
            TooltipBox["8",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"4", ",", "2"}], "}"}], "\[Rule]", "8"}]],
            Annotation[#, {4, 2} -> 8, "Tooltip"]& ], 
           TagBox[
            TooltipBox["12",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"4", ",", "3"}], "}"}], "\[Rule]", "12"}]],
            Annotation[#, {4, 3} -> 12, "Tooltip"]& ], 
           TagBox[
            TooltipBox["16",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"4", ",", "4"}], "}"}], "\[Rule]", "16"}]],
            Annotation[#, {4, 4} -> 16, "Tooltip"]& ]}
         },
         AutoDelete->False,
         GridBoxDividers->{
          "ColumnsIndexed" -> {
            1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}],
        FontSize->Small], 
       StyleBox[GridBox[{
          {"a", "b"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xCoba`CTensor[{{1, 2, 3, 4}, {2, 4, 6, 8}, {3, 6, 9, 12}, {4, 8, 12, 
    16}}, {$CellContext`red, $CellContext`red}, 
    0][$CellContext`a, $CellContext`b],
   Editable->False], "\[Equal]", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       StyleBox[GridBox[{
          {
           TagBox[
            TooltipBox["1",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "1"}]],
            Annotation[#, {1, 1} -> 1, "Tooltip"]& ], 
           TagBox[
            TooltipBox["2",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "2"}]],
            Annotation[#, {1, 2} -> 2, "Tooltip"]& ], 
           TagBox[
            TooltipBox["3",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", "3"}]],
            Annotation[#, {1, 3} -> 3, "Tooltip"]& ], 
           TagBox[
            TooltipBox["4",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"1", ",", "4"}], "}"}], "\[Rule]", "4"}]],
            Annotation[#, {1, 4} -> 4, "Tooltip"]& ]},
          {
           TagBox[
            TooltipBox["2",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "2"}]],
            Annotation[#, {2, 1} -> 2, "Tooltip"]& ], 
           TagBox[
            TooltipBox["4",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", "4"}]],
            Annotation[#, {2, 2} -> 4, "Tooltip"]& ], 
           TagBox[
            TooltipBox["6",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", "6"}]],
            Annotation[#, {2, 3} -> 6, "Tooltip"]& ], 
           TagBox[
            TooltipBox["8",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"2", ",", "4"}], "}"}], "\[Rule]", "8"}]],
            Annotation[#, {2, 4} -> 8, "Tooltip"]& ]},
          {
           TagBox[
            TooltipBox["3",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "3"}]],
            Annotation[#, {3, 1} -> 3, "Tooltip"]& ], 
           TagBox[
            TooltipBox["6",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", "6"}]],
            Annotation[#, {3, 2} -> 6, "Tooltip"]& ], 
           TagBox[
            TooltipBox["9",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", "9"}]],
            Annotation[#, {3, 3} -> 9, "Tooltip"]& ], 
           TagBox[
            TooltipBox["12",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"3", ",", "4"}], "}"}], "\[Rule]", "12"}]],
            Annotation[#, {3, 4} -> 12, "Tooltip"]& ]},
          {
           TagBox[
            TooltipBox["4",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"4", ",", "1"}], "}"}], "\[Rule]", "4"}]],
            Annotation[#, {4, 1} -> 4, "Tooltip"]& ], 
           TagBox[
            TooltipBox["8",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"4", ",", "2"}], "}"}], "\[Rule]", "8"}]],
            Annotation[#, {4, 2} -> 8, "Tooltip"]& ], 
           TagBox[
            TooltipBox["12",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"4", ",", "3"}], "}"}], "\[Rule]", "12"}]],
            Annotation[#, {4, 3} -> 12, "Tooltip"]& ], 
           TagBox[
            TooltipBox["16",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"4", ",", "4"}], "}"}], "\[Rule]", "16"}]],
            Annotation[#, {4, 4} -> 16, "Tooltip"]& ]}
         },
         AutoDelete->False,
         GridBoxDividers->{
          "ColumnsIndexed" -> {
            1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}],
        FontSize->Small], 
       StyleBox[GridBox[{
          {"b", "a"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xCoba`CTensor[{{1, 2, 3, 4}, {2, 4, 6, 8}, {3, 6, 9, 12}, {4, 8, 12, 
    16}}, {$CellContext`red, $CellContext`red}, 
    0][$CellContext`b, $CellContext`a],
   Editable->False]}]], "Output",
 CellChangeTimes->{
  3.7286755295649157`*^9, 3.728753457339127*^9, 3.7288434487970963`*^9, 
   3.730955439727563*^9, 3.730956192482012*^9, 3.731115052492839*^9, 
   3.733367638898735*^9, 3.733367887848672*^9, 3.7333681202784443`*^9, 
   3.7908673887370653`*^9, 3.7922737448707314`*^9, {3.7945179111057587`*^9, 
   3.794517934623646*^9}, 3.79451801401764*^9, 3.7945181214270687`*^9, 
   3.794518779496229*^9, 3.79451905939601*^9, 3.7945192949570723`*^9, 
   3.7945200038308163`*^9, 3.794520178815001*^9, 3.794520314254654*^9, 
   3.794520415970984*^9, 3.794520585189296*^9, 3.794520661274991*^9},
 CellLabel->"Out[24]=",
 CellID->1194455769,ExpressionUUID->"20c3eb96-c8c9-4da7-859b-375de3ef604d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ToCanonical"}]], "Input",
 CellLabel->"In[25]:=",
 CellID->631607572,ExpressionUUID->"377e1db9-7891-4652-8842-8ffa96e79bea"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{
  3.728675530631819*^9, 3.728753457523169*^9, 3.7288434491652403`*^9, 
   3.730955440447505*^9, 3.730956192595237*^9, 3.7311150655708017`*^9, 
   3.733367638996192*^9, 3.733367887948285*^9, 3.7333681203822927`*^9, 
   3.790867390267949*^9, 3.792273746488638*^9, {3.794517911157618*^9, 
   3.794517934669702*^9}, 3.794518014072508*^9, 3.794518121472679*^9, 
   3.794518779590334*^9, 3.7945190594448957`*^9, 3.794519295001918*^9, 
   3.79452000387477*^9, 3.794520179590477*^9, 3.794520314296218*^9, 
   3.794520416013651*^9, 3.794520585240108*^9, 3.794520661322229*^9},
 CellLabel->"Out[25]=",
 CellID->540977427,ExpressionUUID->"9bfd89fc-258f-4e95-bc6e-1830c51128d1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["In sums there is automatic reordering of indices:", "MathCaption",
 CellChangeTimes->{{3.7286755354602547`*^9, 3.728675535612438*^9}},
 CellID->2137509914,ExpressionUUID->"78ebf05d-2dd5-4d24-b5ee-7147c7357c83"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"T", "[", 
   RowBox[{"a", ",", "b"}], "]"}], "-", 
  RowBox[{"T", "[", 
   RowBox[{"b", ",", "a"}], "]"}]}]], "Input",
 CellLabel->"In[26]:=",
 CellID->2022677582,ExpressionUUID->"cdc353ac-6316-466e-b211-c5be91430bc5"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{
  3.7286755366456137`*^9, 3.728753457659789*^9, 3.7288434493314323`*^9, 
   3.730955441491143*^9, 3.73095619269694*^9, 3.731115068359783*^9, 
   3.733367639100366*^9, 3.733367888031074*^9, 3.733368120495987*^9, 
   3.7908673978359213`*^9, 3.7922737477464123`*^9, {3.794517911202278*^9, 
   3.79451793533913*^9}, 3.794518014128221*^9, 3.794518121515593*^9, 
   3.794518779636202*^9, 3.794519059491939*^9, 3.7945192950442753`*^9, 
   3.7945200039174232`*^9, 3.794520179637085*^9, 3.794520314342942*^9, 
   3.7945204160560904`*^9, 3.794520585294876*^9, 3.794520661369647*^9},
 CellLabel->"Out[26]=",
 CellID->1909337838,ExpressionUUID->"76d22231-c9b0-473a-855a-eb3047ac4937"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Note that the previous expression returned 0, and not a CTensor with an \
array of zeros. This is to follow the standard behaviour of ",
 StyleBox["xTensor",
  FontSlant->"Italic"],
 ":"
}], "MathCaption",
 CellChangeTimes->{{3.728675544156136*^9, 3.72867554430761*^9}},
 CellID->564890773,ExpressionUUID->"1bde43d6-a4c5-4029-a6b4-6ab1fcdd0a78"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"v", "[", "a", "]"}], 
   RowBox[{"v", "[", "b", "]"}]}], "-", 
  RowBox[{
   RowBox[{"v", "[", "b", "]"}], 
   RowBox[{"v", "[", "a", "]"}]}]}]], "Input",
 CellLabel->"In[27]:=",
 CellID->94018424,ExpressionUUID->"824d6f7a-d54c-4fd9-8067-ef883f205eaf"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{
  3.728753457807764*^9, 3.728843449831109*^9, 3.73095544256207*^9, 
   3.730956192798074*^9, 3.731115075834507*^9, 3.7333676392337503`*^9, 
   3.73336788814856*^9, 3.733368120578047*^9, 3.790867405705461*^9, 
   3.7922737503459177`*^9, {3.794517911243023*^9, 3.7945179353839827`*^9}, 
   3.7945180141865797`*^9, 3.7945181215568438`*^9, 3.794518779676478*^9, 
   3.794519059543346*^9, 3.7945192950914516`*^9, 3.7945200039603043`*^9, 
   3.794520179687388*^9, 3.7945203143883667`*^9, 3.794520416097968*^9, 
   3.7945205853448553`*^9, 3.7945206614177017`*^9},
 CellLabel->"Out[27]=",
 CellID->392906145,ExpressionUUID->"d7437574-ba2a-40fb-81b3-0f4ebc75ea44"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "The CTensor notation is an alternative to the \[OpenCurlyDoubleQuote]linear \
combination notation\[CloseCurlyDoubleQuote] in ",
 StyleBox["xCoba",
  FontSlant->"Italic"],
 ". It is possible to change between both notations."
}], "Text",
 CellID->1154629751,ExpressionUUID->"747ae260-a1b6-4609-9b86-15aef1a6daf6"],

Cell[CellGroupData[{

Cell["\<\
Expand a CTensor into a linear combination of bases objects, and then back:\
\>", "MathCaption",
 CellChangeTimes->{{3.7286755548353662`*^9, 3.7286755804258957`*^9}},
 CellID->1199990765,ExpressionUUID->"c551201f-38e5-4db8-8a2a-6c017e559b11"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"v", "[", "a", "]"}]], "Input",
 CellLabel->"In[28]:=",
 CellID->1225858378,ExpressionUUID->"26b0e0d3-c027-4600-9700-fa68f8132440"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", "1"}]],
           Annotation[#, {1} -> 1, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["2",
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", "2"}]],
           Annotation[#, {2} -> 2, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["3",
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", "3"}]],
           Annotation[#, {3} -> 3, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["4",
            RowBox[{
              RowBox[{"{", "4", "}"}], "\[Rule]", "4"}]],
           Annotation[#, {4} -> 4, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{1, 2, 3, 4}, {$CellContext`red}, 0][$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.728675556413795*^9, 3.728753457960579*^9, 3.7288434505973873`*^9, 
   3.73095544556291*^9, 3.730956192902398*^9, 3.7311150818346357`*^9, 
   3.73336763951455*^9, 3.733367888231453*^9, 3.733368120660985*^9, 
   3.7908674121267643`*^9, 3.7922737569539003`*^9, {3.794517911289469*^9, 
   3.794517935428439*^9}, 3.794518014248289*^9, 3.7945181216010447`*^9, 
   3.79451877972128*^9, 3.794519059589971*^9, 3.794519295137301*^9, 
   3.7945200040056953`*^9, 3.7945201797379303`*^9, 3.79452031443897*^9, 
   3.7945204161382732`*^9, 3.794520585397373*^9, 3.7945206614679537`*^9},
 CellLabel->"Out[28]=",
 CellID->61364490,ExpressionUUID->"da597dc5-0162-4028-9e2c-616f923cc33c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"BasisExpand", "[", 
  RowBox[{"%", ",", 
   RowBox[{"{", "red", "}"}]}], "]"}]], "Input",
 CellLabel->"In[29]:=",
 CellID->1329054869,ExpressionUUID->"0e1e1d1a-5b59-41c8-aa07-2e59074d1599"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", "a"},
          {
           StyleBox["0",
            FontColor->RGBColor[1., 0., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{0, -$CellContext`red}, $CellContext`a],
   Editable->False], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`red}, $CellContext`a],
    Editable->False]}], "+", 
  RowBox[{"3", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`red}, $CellContext`a],
    Editable->False]}], "+", 
  RowBox[{"4", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`red}, $CellContext`a],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{
  3.728675558294766*^9, 3.728753458121972*^9, 3.7288434509633493`*^9, 
   3.730955446411839*^9, 3.730956193058544*^9, 3.7311150833835487`*^9, 
   3.7333676396999702`*^9, 3.7333678883486443`*^9, 3.7333681207613497`*^9, 
   3.790867413344054*^9, 3.792273757762892*^9, {3.7945179113376617`*^9, 
   3.794517935487075*^9}, 3.794518014317898*^9, 3.79451812165339*^9, 
   3.794518779769758*^9, 3.79451905964602*^9, 3.79451929518781*^9, 
   3.794520004056786*^9, 3.794520179797019*^9, 3.7945203144913588`*^9, 
   3.794520416188653*^9, 3.794520585461747*^9, 3.794520661532131*^9},
 CellLabel->"Out[29]=",
 CellID->560832870,ExpressionUUID->"e18cd6e0-bbe7-45a7-b1d3-f11537557b80"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%%", "//", "ToBasisExpand"}]], "Input",
 CellLabel->"In[30]:=",
 CellID->1694720142,ExpressionUUID->"6012be2c-d682-43c6-bc7c-15395d03a97b"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", "a"},
          {
           StyleBox["0",
            FontColor->RGBColor[1., 0., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{0, -$CellContext`red}, $CellContext`a],
   Editable->False], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`red}, $CellContext`a],
    Editable->False]}], "+", 
  RowBox[{"3", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`red}, $CellContext`a],
    Editable->False]}], "+", 
  RowBox[{"4", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`red}, $CellContext`a],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{
  3.7286755603906593`*^9, 3.728753458273004*^9, 3.728843451143325*^9, 
   3.730955447327018*^9, 3.730956193241171*^9, 3.731115084703246*^9, 
   3.733367639848325*^9, 3.7333678884274693`*^9, 3.733368120861435*^9, 
   3.7908674155584593`*^9, 3.792273760740499*^9, {3.79451791138225*^9, 
   3.794517935532445*^9}, 3.794518014385807*^9, 3.7945181217016153`*^9, 
   3.794518779819655*^9, 3.7945190596965714`*^9, 3.794519295242584*^9, 
   3.7945200041102753`*^9, 3.794520179845435*^9, 3.794520314542398*^9, 
   3.7945204162497797`*^9, 3.7945205855245934`*^9, 3.794520661587514*^9},
 CellLabel->"Out[30]=",
 CellID->612319407,ExpressionUUID->"47fc1ad9-9bf0-48bc-b6b5-b1a568e2e7f6"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FromBasisExpand", "[", "%", "]"}]], "Input",
 CellChangeTimes->{{3.792273764836111*^9, 3.792273765756147*^9}},
 CellLabel->"In[31]:=",
 CellID->757225,ExpressionUUID->"41f49a22-28d4-46bb-a046-07ffd40d91f8"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", "1"}]],
           Annotation[#, {1} -> 1, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["2",
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", "2"}]],
           Annotation[#, {2} -> 2, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["3",
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", "3"}]],
           Annotation[#, {3} -> 3, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["4",
            RowBox[{
              RowBox[{"{", "4", "}"}], "\[Rule]", "4"}]],
           Annotation[#, {4} -> 4, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{1, 2, 3, 4}, {$CellContext`red}, 0][$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.7286755660950327`*^9, 3.7287534584103327`*^9, 3.728843451295847*^9, 
   3.730955452730246*^9, 3.7309561933548326`*^9, 3.731115090970282*^9, 
   3.7333676399499693`*^9, 3.733367888515101*^9, 3.733368120961026*^9, 
   3.79086742317054*^9, 3.792273766593338*^9, {3.7945179114308243`*^9, 
   3.7945179355752897`*^9}, 3.7945180144429607`*^9, 3.7945181217438927`*^9, 
   3.79451877986639*^9, 3.794519059747944*^9, 3.794519295288041*^9, 
   3.7945200041566887`*^9, 3.794520179967259*^9, 3.794520314587335*^9, 
   3.7945204162929497`*^9, 3.794520585579295*^9, 3.794520661641758*^9},
 CellLabel->"Out[31]=",
 CellID->510637025,ExpressionUUID->"35827a2f-00f0-434e-ba68-3ad3eb9ab176"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.1.3. A metric field", "Subsubsection",
 CellChangeTimes->{{3.728674022994522*^9, 3.728674023274963*^9}, {
  3.733368003800891*^9, 3.733368003919387*^9}},
 CellID->1506731884,ExpressionUUID->"5fe8e3e7-f13c-43e2-bdf9-ad5e84894f10"],

Cell["\<\
This subsection will use a different chart and metric, namely Schwarzschild.\
\>", "Text",
 CellChangeTimes->{{3.731115160421138*^9, 3.731115186906599*^9}, 
   3.790867441915625*^9},
 CellID->1773480832,ExpressionUUID->"0a56d410-e643-4c32-8dac-2a8047ae7755"],

Cell[CellGroupData[{

Cell["Define a coordinate system:", "MathCaption",
 CellChangeTimes->{{3.7286756009530897`*^9, 3.728675601129382*^9}},
 CellID->515608964,ExpressionUUID->"33924e69-beea-43f0-b611-01a8304e95f8"],

Cell[BoxData[
 RowBox[{"DefChart", "[", 
  RowBox[{"ch", ",", "M", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2", ",", "3"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"t", "[", "]"}], ",", 
     RowBox[{"r", "[", "]"}], ",", 
     RowBox[{"\[Theta]", "[", "]"}], ",", 
     RowBox[{"\[Phi]", "[", "]"}]}], "}"}], ",", 
   RowBox[{"ChartColor", "\[Rule]", "Brown"}]}], "]"}]], "Input",
 CellLabel->"In[32]:=",
 CellID->743342303,ExpressionUUID->"44cc2207-fb49-42f2-8bc4-a7a4563ee3b2"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Now define a metric field, the Schwarzschild metric:", "MathCaption",
 CellChangeTimes->{{3.728675610912695*^9, 3.7286756110888987`*^9}},
 CellID->1441711922,ExpressionUUID->"12226d3a-a740-4439-9528-f8a71024138d"],

Cell[BoxData[
 RowBox[{"DefConstantSymbol", "[", 
  RowBox[{"mass", ",", 
   RowBox[{"PrintAs", "\[Rule]", "\"\<M\>\""}]}], "]"}]], "Input",
 CellLabel->"In[33]:=",
 CellID->794283953,ExpressionUUID->"9016df49-04ad-4bd0-83e6-214869a2d228"],

Cell[BoxData[
 RowBox[{
  RowBox[{"met", "=", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"DiagonalMatrix", "[", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "1"}], "+", 
         RowBox[{"2", 
          RowBox[{"mass", "/", 
           RowBox[{"r", "[", "]"}]}]}]}], ",", 
        RowBox[{"1", "/", 
         RowBox[{"(", 
          RowBox[{"1", "-", 
           RowBox[{"2", 
            RowBox[{"mass", "/", 
             RowBox[{"r", "[", "]"}]}]}]}], ")"}]}], ",", 
        RowBox[{
         RowBox[{"r", "[", "]"}], "^", "2"}], ",", 
        RowBox[{
         RowBox[{
          RowBox[{"r", "[", "]"}], "^", "2"}], 
         RowBox[{
          RowBox[{"Sin", "[", 
           RowBox[{"\[Theta]", "[", "]"}], "]"}], "^", "2"}]}]}], "}"}], 
      "]"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "ch"}], ",", 
       RowBox[{"-", "ch"}]}], "}"}]}], "]"}]}], ";"}]], "Input",
 CellLabel->"In[34]:=",
 CellID->772048319,ExpressionUUID->"33a14312-4241-4671-9b30-785890809b72"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"met", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}]}], "]"}]], "Input",
 CellLabel->"In[35]:=",
 CellID->719902259,ExpressionUUID->"a1c20b71-2eb1-47cf-af1a-cee79ce5d739"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{"-", "1"}], "+", 
             FractionBox[
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`mass,
                Editable->False]}], 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{"-", "1"}], "+", 
                FractionBox[
                 RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], 
                 InterpretationBox[
                  StyleBox[
                  "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`r[], Editable -> False]]}]}]],
           
           Annotation[#, {1, 1} -> -1 + 2 $CellContext`mass/$CellContext`r[], 
            "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox["1", 
             RowBox[{"1", "-", 
              FractionBox[
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`mass,
                 Editable->False]}], 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]]}]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              FractionBox["1", 
               RowBox[{"1", "-", 
                 FractionBox[
                  RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], 
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False]]}]]}]],
           
           Annotation[#, {2, 
             2} -> (1 - 2 $CellContext`mass/$CellContext`r[])^(-1), 
            "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox[
                "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                $CellContext`r[], Editable -> False], "2"]}]],
           Annotation[#, {3, 3} -> $CellContext`r[]^2, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"], " ", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], "2"]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "4"}], "}"}], "\[Rule]", 
              RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                  "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`r[], Editable -> False], "2"], " ", 
                SuperscriptBox[
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                   InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                 "2"]}]}]],
           Annotation[#, {4, 4} -> $CellContext`r[]^2 Sin[
                $CellContext`\[Theta][]]^2, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[0.6, 0.4, 0.2], 5 -> RGBColor[0.6, 0.4, 0.2]}, 
          "RowsIndexed" -> {
           1 -> RGBColor[0.6, 0.4, 0.2], 5 -> RGBColor[0.6, 0.4, 0.2]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " "},
         {"a", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{-1 + 2 $CellContext`mass/$CellContext`r[], 0, 0, 0}, {
    0, (1 - 2 $CellContext`mass/$CellContext`r[])^(-1), 0, 0}, {
    0, 0, $CellContext`r[]^2, 0}, {0, 0, 0, $CellContext`r[]^2 Sin[
        $CellContext`\[Theta][]]^2}}, {-$CellContext`ch, -$CellContext`ch}, 
   0][-$CellContext`a, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.728675614570578*^9, 3.7287534664660063`*^9, 3.728843460577059*^9, 
   3.730955468634659*^9, 3.730956193717885*^9, 3.731115204921974*^9, 
   3.733367647102477*^9, 3.73336788890082*^9, 3.733368121293325*^9, 
   3.790867452693145*^9, 3.792273789088664*^9, {3.794517911573339*^9, 
   3.794517936324622*^9}, 3.794518014627788*^9, 3.794518121911418*^9, 
   3.7945187800066147`*^9, 3.79451905990331*^9, 3.794519295427374*^9, 
   3.7945200043793287`*^9, 3.794520180128693*^9, 3.7945203147377243`*^9, 
   3.7945204164434633`*^9, 3.794520585741695*^9, 3.794520661822433*^9},
 CellLabel->"Out[35]=",
 CellID->931825763,ExpressionUUID->"12567fc9-ea55-40a2-8bcc-5f4494002338"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
That tensor does not have any information about being a metric. In particular \
we cannot yet use it to raise and lower indices:\
\>", "MathCaption",
 CellChangeTimes->{{3.731115290080184*^9, 3.731115318294155*^9}},
 CellID->1485801604,ExpressionUUID->"f89f3030-a151-4409-af11-20385c37d108"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"met", "[", 
   RowBox[{"a", ",", "b"}], "]"}], "//", "Catch"}]], "Input",
 CellChangeTimes->{{3.731115319738065*^9, 3.731115325597781*^9}},
 CellLabel->"In[36]:=",
 CellID->1196101862,ExpressionUUID->"f4208aa9-ac67-4566-b81e-067df1103fc0"],

Cell[BoxData[
 TemplateBox[{
  "MetricsOfVBundle", "missing", 
   "\"There is no \\!\\(\\*RowBox[{\\\"\\\\\\\"metric\\\\\\\"\\\"}]\\) in \\!\
\\(\\*InterpretationBox[StyleBox[\\\"\[DoubleStruckCapitalT]M\\\", \
Rule[AutoSpacing, False], Rule[ShowAutoStyles, False]], TangentM, \
Rule[Editable, False]]\\).\"", 2, 36, 1, 19683117308123869115, "Local", 
   "xAct`xTensor`MetricsOfVBundle"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{{3.731115321786879*^9, 3.731115326087393*^9}, 
   3.733367647256508*^9, 3.73336788911875*^9, 3.733368121398917*^9, 
   3.7908674602309523`*^9, 3.792273790131481*^9, {3.794517911630246*^9, 
   3.7945179363779993`*^9}, 3.7945180148121243`*^9, 3.7945181219654512`*^9, 
   3.794518780055272*^9, 3.794519059970068*^9, 3.7945192954867783`*^9, 
   3.794520004428812*^9, 3.794520180195681*^9, 3.794520314787225*^9, 
   3.7945204165044937`*^9, 3.794520585812105*^9, 3.794520661877304*^9},
 CellLabel->"During evaluation of In[36]:=",
 CellID->1744844817,ExpressionUUID->"5026d175-a6a7-4eff-abf8-6b64b4309f11"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
This is a contravariant vector, because we are using basis ch, rather than \
basis -ch for a covariant vector:\
\>", "MathCaption",
 CellChangeTimes->{{3.7311154105867567`*^9, 3.731115443521899*^9}},
 CellID->1766165890,ExpressionUUID->"1b18e42a-51fe-4579-8ede-1a32ab800da7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"v", "=", 
  RowBox[{"CTensor", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"0", ",", 
      RowBox[{"r", "[", "]"}], ",", "0", ",", "0"}], "}"}], ",", 
    RowBox[{"{", "ch", "}"}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.731115332193253*^9, 3.731115347334496*^9}, {
  3.731115377464089*^9, 3.731115382971672*^9}},
 CellLabel->"In[37]:=",
 CellID->2064427616,ExpressionUUID->"c81140ce-3664-4751-bd18-ead8f98a983e"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0", ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", "ch", "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{
  3.7311153506319*^9, 3.7311153956962843`*^9, 3.733367647386559*^9, 
   3.7333678892360363`*^9, 3.733368121516445*^9, 3.790867469947216*^9, 
   3.7922737915752363`*^9, {3.794517911675222*^9, 3.7945179364262114`*^9}, 
   3.79451801486841*^9, 3.794518122018239*^9, 3.7945187800991917`*^9, 
   3.794519060019753*^9, 3.794519295528309*^9, 3.7945200044690647`*^9, 
   3.794520180241791*^9, 3.7945203148275013`*^9, 3.794520416549163*^9, 
   3.794520585863248*^9, 3.7945206619218893`*^9},
 CellLabel->"Out[37]=",
 CellID->1717639534,ExpressionUUID->"25645139-4869-4123-b364-8c9d80a14b13"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"v", "[", "a", "]"}]], "Input",
 CellChangeTimes->{{3.731115351873135*^9, 3.7311153544767227`*^9}},
 CellLabel->"In[38]:=",
 CellID->1367265467,ExpressionUUID->"ae362dcc-94e1-4324-be24-5c32f923de5c"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              InterpretationBox[
               StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> False], 
               $CellContext`r[], Editable -> False]}]],
           Annotation[#, {2} -> $CellContext`r[], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", "4", "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4} -> 0, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{0, 
    $CellContext`r[], 0, 0}, {$CellContext`ch}, 0][$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.731115355155909*^9, 3.731115397009707*^9, 3.7333676475021057`*^9, 
   3.7333678893857927`*^9, 3.733368121633013*^9, 3.790867470916614*^9, 
   3.792273792073091*^9, {3.794517911726552*^9, 3.794517936467064*^9}, 
   3.7945180149267263`*^9, 3.7945181220720367`*^9, 3.794518780149357*^9, 
   3.79451906007025*^9, 3.794519295577506*^9, 3.794520004512014*^9, 
   3.7945201802963743`*^9, 3.7945203148767548`*^9, 3.794520416599251*^9, 
   3.794520585919868*^9, 3.794520661971919*^9},
 CellLabel->"Out[38]=",
 CellID->1892464964,ExpressionUUID->"78630d43-01fb-44ff-95b6-6cbdb0b55bf4"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"v", "[", 
   RowBox[{"-", "a"}], "]"}], "//", "Catch"}]], "Input",
 CellChangeTimes->{{3.731115356463869*^9, 3.731115358884501*^9}},
 CellLabel->"In[39]:=",
 CellID->206460103,ExpressionUUID->"7aae0d79-4f5d-4dbc-a5c2-2e203db308ed"],

Cell[BoxData[
 TemplateBox[{
  "MetricsOfVBundle", "missing", 
   "\"There is no \\!\\(\\*RowBox[{\\\"\\\\\\\"metric\\\\\\\"\\\"}]\\) in \\!\
\\(\\*InterpretationBox[StyleBox[\\\"\[DoubleStruckCapitalT]M\\\", \
Rule[AutoSpacing, False], Rule[ShowAutoStyles, False]], TangentM, \
Rule[Editable, False]]\\).\"", 2, 39, 2, 19683117308123869115, "Local", 
   "xAct`xTensor`MetricsOfVBundle"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{
  3.731115359377205*^9, 3.7311153988494062`*^9, 3.733367647724102*^9, 
   3.733367889502824*^9, 3.7333681217494287`*^9, 3.790867471928893*^9, 
   3.792273792828895*^9, {3.794517911771819*^9, 3.7945179365144787`*^9}, 
   3.7945180149762897`*^9, 3.7945181221215057`*^9, 3.794518780191422*^9, 
   3.794519060117358*^9, 3.7945192956186028`*^9, 3.794520004548723*^9, 
   3.794520180364463*^9, 3.794520314916974*^9, 3.7945204166386423`*^9, 
   3.79452058597195*^9, 3.794520662014533*^9},
 CellLabel->"During evaluation of In[39]:=",
 CellID->278354231,ExpressionUUID->"ba3b9d83-cfa0-4ed8-8735-72a75d52249e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Of course, you can manually use the metric to lower indices:", \
"MathCaption",
 CellChangeTimes->{{3.7311154105867567`*^9, 3.731115443521899*^9}, {
  3.7908675028501*^9, 3.7908675264637117`*^9}},
 CellID->1338376509,ExpressionUUID->"917465d0-14df-486c-b94e-dade6b7360e3"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"met", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}]}], "]"}], 
  RowBox[{"v", "[", "b", "]"}]}]], "Input",
 CellChangeTimes->{{3.731115332193253*^9, 3.731115347334496*^9}, {
  3.731115377464089*^9, 3.731115382971672*^9}, {3.790867529016938*^9, 
  3.790867533537118*^9}},
 CellLabel->"In[40]:=",
 CellID->1350268085,ExpressionUUID->"311a57e5-e35b-453e-b9e3-ef8341ea4996"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            FractionBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], 
             RowBox[{"1", "-", 
              FractionBox[
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`mass,
                 Editable->False]}], 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]]}]],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              FractionBox[
               InterpretationBox[
                StyleBox[
                "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                $CellContext`r[], Editable -> False], 
               RowBox[{"1", "-", 
                 FractionBox[
                  RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], 
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False]]}]]}]],
           
           Annotation[#, {
             2} -> (1 - 
               2 $CellContext`mass/$CellContext`r[])^(-1) $CellContext`r[], 
            "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", "4", "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4} -> 0, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" "},
         {"a"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{
   0, (1 - 2 $CellContext`mass/$CellContext`r[])^(-1) $CellContext`r[], 0, 
    0}, {-$CellContext`ch}, 0][-$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.7311153506319*^9, 3.7311153956962843`*^9, 3.733367647386559*^9, 
   3.7333678892360363`*^9, 3.733368121516445*^9, 3.790867469947216*^9, 
   3.790867534473448*^9, 3.792273794376608*^9, {3.794517911814517*^9, 
   3.7945179365602207`*^9}, 3.7945180150400333`*^9, 3.7945181221789703`*^9, 
   3.79451878024606*^9, 3.794519060165064*^9, 3.794519295667225*^9, 
   3.794520004596633*^9, 3.794520180408197*^9, 3.7945203149646*^9, 
   3.794520416688428*^9, 3.79452058604494*^9, 3.794520662075132*^9},
 CellLabel->"Out[40]=",
 CellID->1080177251,ExpressionUUID->"558e2c03-3c8b-46e3-a4ee-e1ec20ee136a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We now declare the met is the metric on the vbundle of the coordinated frame \
of ch (this is roughly the equivalent of DefMetric in xTensor). In some cases \
SetCMetric can unambiguously compute the signature from the matrix of \
components, but in general it cannot, so we need to specify the signature as \
a triple {number of +1s, number of -1s, number of 0s}:\
\>", "MathCaption",
 CellChangeTimes->{{3.728675622072378*^9, 3.72867562223173*^9}, {
  3.7311154698317947`*^9, 3.7311155297381783`*^9}, {3.79086756234429*^9, 
  3.790867628452179*^9}},
 CellID->1778393041,ExpressionUUID->"3047ac7b-30a6-464c-a599-83c73e27ebc7"],

Cell[BoxData[
 RowBox[{"SetCMetric", "[", 
  RowBox[{"met", ",", "ch", ",", 
   RowBox[{"SignatureOfMetric", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"3", ",", "1", ",", "0"}], "}"}]}]}], "]"}]], "Input",
 CellLabel->"In[41]:=",
 CellID->774441141,ExpressionUUID->"8035bd06-daf9-49bf-b046-36cfda00156e"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Now that we have a metric, raising and lowering indices automatically \
produces the expected result:\
\>", "MathCaption",
 CellChangeTimes->{{3.72867562895199*^9, 3.7286756291202393`*^9}, {
  3.731115560933786*^9, 3.731115563101631*^9}},
 CellID->133872320,ExpressionUUID->"20fba6c2-8ca2-4fb7-86b3-a1c42ab48ed1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"v", "[", "a", "]"}], ",", 
   RowBox[{"v", "[", 
    RowBox[{"-", "a"}], "]"}]}], "}"}]], "Input",
 CellChangeTimes->{{3.731115540686391*^9, 3.731115567893421*^9}},
 CellLabel->"In[42]:=",
 CellID->1576703831,ExpressionUUID->"79d081b3-782e-4756-9af2-ca0a547a5ae2"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        StyleBox[GridBox[{
           {
            TagBox[
             TooltipBox["0",
              RowBox[{
                RowBox[{"{", "1", "}"}], "\[Rule]", "0"}]],
             Annotation[#, {1} -> 0, "Tooltip"]& ]},
           {
            TagBox[
             TooltipBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False],
              RowBox[{
                RowBox[{"{", "2", "}"}], "\[Rule]", 
                InterpretationBox[
                 StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`r[], Editable -> False]}]],
             Annotation[#, {2} -> $CellContext`r[], "Tooltip"]& ]},
           {
            TagBox[
             TooltipBox["0",
              RowBox[{
                RowBox[{"{", "3", "}"}], "\[Rule]", "0"}]],
             Annotation[#, {3} -> 0, "Tooltip"]& ]},
           {
            TagBox[
             TooltipBox["0",
              RowBox[{
                RowBox[{"{", "4", "}"}], "\[Rule]", "0"}]],
             Annotation[#, {4} -> 0, "Tooltip"]& ]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}],
         FontSize->Small], 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xCoba`CTensor[{0, 
      $CellContext`r[], 0, 0}, {$CellContext`ch}, 0][$CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        StyleBox[GridBox[{
           {
            TagBox[
             TooltipBox["0",
              RowBox[{
                RowBox[{"{", "1", "}"}], "\[Rule]", "0"}]],
             Annotation[#, {1} -> 0, "Tooltip"]& ]},
           {
            TagBox[
             TooltipBox[
              FractionBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], 
               RowBox[{"1", "-", 
                FractionBox[
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`mass,
                   Editable->False]}], 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]]}]],
              RowBox[{
                RowBox[{"{", "2", "}"}], "\[Rule]", 
                FractionBox[
                 InterpretationBox[
                  StyleBox[
                  "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`r[], Editable -> False], 
                 RowBox[{"1", "-", 
                   FractionBox[
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]]}]],
             
             Annotation[#, {
               2} -> (1 - 
                 2 $CellContext`mass/$CellContext`r[])^(-1) $CellContext`r[], 
              "Tooltip"]& ]},
           {
            TagBox[
             TooltipBox["0",
              RowBox[{
                RowBox[{"{", "3", "}"}], "\[Rule]", "0"}]],
             Annotation[#, {3} -> 0, "Tooltip"]& ]},
           {
            TagBox[
             TooltipBox["0",
              RowBox[{
                RowBox[{"{", "4", "}"}], "\[Rule]", "0"}]],
             Annotation[#, {4} -> 0, "Tooltip"]& ]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}],
         FontSize->Small], 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xCoba`CTensor[{
     0, (1 - 2 $CellContext`mass/$CellContext`r[])^(-1) $CellContext`r[], 0, 
      0}, {-$CellContext`ch}, 0][-$CellContext`a],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{
  3.7286756304168*^9, 3.728753466719215*^9, 3.728843463168593*^9, 
   3.730955475161502*^9, 3.730956193931532*^9, {3.731115545359721*^9, 
   3.731115572498822*^9}, 3.733367648102648*^9, 3.733367889801869*^9, 
   3.733368121948951*^9, 3.7908676386690483`*^9, 3.7922737982108927`*^9, {
   3.7945179119580708`*^9, 3.7945179366837482`*^9}, 3.794518015157805*^9, 
   3.7945181223139973`*^9, 3.7945187803704567`*^9, 3.79451906029252*^9, 
   3.794519295774599*^9, 3.794520004709215*^9, 3.7945201805205793`*^9, 
   3.79452031514423*^9, 3.794520416804084*^9, 3.794520586178864*^9, 
   3.794520662225387*^9},
 CellLabel->"Out[42]=",
 CellID->1859789466,ExpressionUUID->"f63760fc-bf78-43f0-9a89-214d5b6b8ab0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["This is a contraction:", "MathCaption",
 CellID->556313662,ExpressionUUID->"e3d392ed-88f0-4fe7-ad8d-f8e953eafb82"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"v", "[", "a", "]"}], 
  RowBox[{"v", "[", 
   RowBox[{"-", "a"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.731115581877224*^9, 3.731115583597047*^9}},
 CellLabel->"In[43]:=",
 CellID->1967401843,ExpressionUUID->"ac8d6240-9f51-4643-9779-8389ff9fd036"],

Cell[BoxData[
 FractionBox[
  SuperscriptBox[
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False], "2"], 
  RowBox[{"1", "-", 
   FractionBox[
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`mass,
      Editable->False]}], 
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False]]}]]], "Output",
 CellChangeTimes->{
  3.728675641573112*^9, 3.72875346701488*^9, 3.728843465485214*^9, 
   3.730955478278717*^9, 3.730956194146173*^9, 3.731115584022462*^9, 
   3.733367648219386*^9, 3.733367889918771*^9, 3.733368122080062*^9, 
   3.7908676414040403`*^9, 3.79227379939884*^9, {3.7945179120770607`*^9, 
   3.794517936727437*^9}, 3.794518015210493*^9, 3.7945181223668747`*^9, 
   3.7945187804126997`*^9, 3.794519060422271*^9, 3.794519295821858*^9, 
   3.794520004749542*^9, 3.794520180564629*^9, 3.794520315205687*^9, 
   3.794520416846587*^9, 3.794520586230001*^9, 3.7945206622710743`*^9},
 CellLabel->"Out[43]=",
 CellID->857072977,ExpressionUUID->"2fe4ac5f-0c4d-4980-95a3-3995b940bcb4"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["If we raise one index of the metric we get the identity:", "MathCaption",
 CellID->1754513065,ExpressionUUID->"1e38c929-950f-4da9-bc9b-11bb38f71e5f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"met", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", "b"}], "]"}]], "Input",
 CellLabel->"In[44]:=",
 CellID->1003126498,ExpressionUUID->"f674f4b9-10b0-423d-a6a0-2675dd1f329d"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {1, 1} -> 1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {2, 2} -> 1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {3, 3} -> 1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "4"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {4, 4} -> 1, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[0.6, 0.4, 0.2], 5 -> RGBColor[0.6, 0.4, 0.2]}, 
          "RowsIndexed" -> {
           1 -> RGBColor[0.6, 0.4, 0.2], 5 -> RGBColor[0.6, 0.4, 0.2]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", "b"},
         {"a", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{1, 0, 0, 0}, {0, 1, 0, 0}, {0, 0, 1, 0}, {0, 0, 0, 
   1}}, {-$CellContext`ch, $CellContext`ch}, 
   0][-$CellContext`a, $CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.728675648662574*^9, 3.728753467160966*^9, 3.728843466795478*^9, 
   3.7309554795035543`*^9, 3.730956194279152*^9, 3.731115590940771*^9, 
   3.73336764833814*^9, 3.733367890052656*^9, 3.7333681221819067`*^9, 
   3.7908676466364193`*^9, 3.792273800593377*^9, {3.7945179121277437`*^9, 
   3.7945179367720423`*^9}, 3.794518015262186*^9, 3.794518122503883*^9, 
   3.7945187804603767`*^9, 3.794519060473482*^9, 3.794519295965386*^9, 
   3.7945200047967176`*^9, 3.794520180614873*^9, 3.79452031525595*^9, 
   3.794520416934709*^9, 3.794520586290669*^9, 3.794520662356243*^9},
 CellLabel->"Out[44]=",
 CellID->1652704010,ExpressionUUID->"d54654a7-fc53-4fe3-a8db-230fb0001128"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["This is the inverse metric:", "MathCaption",
 CellChangeTimes->{{3.731115595708486*^9, 3.731115600604542*^9}},
 CellID->831538599,ExpressionUUID->"c5604c5c-bf1a-40ff-b5d2-facbc74fc77c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"met", "[", 
  RowBox[{"a", ",", "b"}], "]"}]], "Input",
 CellLabel->"In[45]:=",
 CellID->1663699122,ExpressionUUID->"b0042f75-63be-46c9-9578-e315b3c8e7cd"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            FractionBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], 
             RowBox[{
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`mass,
                Editable->False]}], "-", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
              FractionBox[
               InterpretationBox[
                StyleBox[
                "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                $CellContext`r[], Editable -> False], 
               RowBox[{
                 RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], "-", 
                 InterpretationBox[
                  StyleBox[
                  "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`r[], Editable -> False]}]]}]],
           
           Annotation[#, {1, 
             1} -> (2 $CellContext`mass - $CellContext`r[])^(-1) \
$CellContext`r[], "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{"1", "-", 
             FractionBox[
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`mass,
                Editable->False]}], 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{"1", "-", 
                FractionBox[
                 RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], 
                 InterpretationBox[
                  StyleBox[
                  "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`r[], Editable -> False]]}]}]],
           
           Annotation[#, {2, 2} -> 1 - 2 $CellContext`mass/$CellContext`r[], 
            "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox["1", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", 
              FractionBox["1", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`r[], Editable -> False], "2"]]}]],
           Annotation[#, {3, 3} -> $CellContext`r[]^(-2), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox[
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Csc",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Csc,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], "2"], 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "4"}], "}"}], "\[Rule]", 
              FractionBox[
               SuperscriptBox[
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"Csc"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Csc, Editable -> False], "[", 
                  InterpretationBox[
                   StyleBox[
                   "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`\[Theta][], Editable -> False], "]"}], "2"], 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`r[], Editable -> False], "2"]]}]],
           Annotation[#, {4, 4} -> Csc[
                $CellContext`\[Theta][]]^2 $CellContext`r[]^(-2), 
            "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[0.6, 0.4, 0.2], 5 -> RGBColor[0.6, 0.4, 0.2]}, 
          "RowsIndexed" -> {
           1 -> RGBColor[0.6, 0.4, 0.2], 5 -> RGBColor[0.6, 0.4, 0.2]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a", "b"},
         {" ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{(
       2 $CellContext`mass - $CellContext`r[])^(-1) $CellContext`r[], 0, 0, 
     0}, {0, 1 - 2 $CellContext`mass/$CellContext`r[], 0, 0}, {
    0, 0, $CellContext`r[]^(-2), 0}, {0, 0, 0, Csc[
        $CellContext`\[Theta][]]^2 $CellContext`r[]^(-2)}}, {$CellContext`ch, \
$CellContext`ch}, 0][$CellContext`a, $CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.728675649379076*^9, 3.7287534673910637`*^9, 3.7288434675614643`*^9, 
   3.7309554802352448`*^9, 3.730956194395062*^9, 3.731115602603587*^9, 
   3.733367648454612*^9, 3.733367890200292*^9, 3.733368122343433*^9, 
   3.790867648718597*^9, 3.792273801746409*^9, {3.794517912303384*^9, 
   3.7945179368241873`*^9}, 3.794518015317706*^9, 3.7945181225581512`*^9, 
   3.7945187805152893`*^9, 3.794519060527013*^9, 3.794519296009829*^9, 
   3.794520004845532*^9, 3.7945201806651506`*^9, 3.79452031531495*^9, 
   3.794520416985145*^9, 3.794520586352178*^9, 3.79452066240476*^9},
 CellLabel->"Out[45]=",
 CellID->1442805772,ExpressionUUID->"2f3378e6-ad73-46a8-bb3e-cc1d967f658c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Without using SetCMetric, it could have also been computed in this form:\
\>", "MathCaption",
 CellChangeTimes->{{3.731115616979823*^9, 3.7311156385150843`*^9}},
 CellID->1163447446,ExpressionUUID->"93a60267-1455-4185-a87d-ed2027600ec4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Inv", "[", "met", "]"}]], "Input",
 CellChangeTimes->{{3.731115606860257*^9, 3.7311156092278967`*^9}},
 CellLabel->"In[46]:=",
 CellID->892729497,ExpressionUUID->"c860eb0c-9a8b-4802-b522-03b09d55ad95"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       FractionBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], 
        RowBox[{
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`mass,
           Editable->False]}], "-", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]}]], ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", 
       RowBox[{"1", "-", 
        FractionBox[
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`mass,
           Editable->False]}], 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]]}], ",", "0", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", 
       FractionBox["1", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]], ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", "0", ",", 
       FractionBox[
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Csc",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Csc,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"], 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"ch", ",", "ch"}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{
  3.731115612280665*^9, 3.733367648550318*^9, 3.733367890352141*^9, 
   3.733368122468131*^9, 3.7908676544036827`*^9, 3.7922738029112387`*^9, {
   3.794517912346243*^9, 3.794517936880116*^9}, 3.794518015372789*^9, 
   3.794518122609177*^9, 3.79451878062362*^9, 3.79451906066578*^9, 
   3.794519296058302*^9, 3.794520004898538*^9, 3.7945201807082443`*^9, 
   3.794520315357123*^9, 3.794520417052713*^9, 3.79452058640467*^9, 
   3.794520662451804*^9},
 CellLabel->"Out[46]=",
 CellID->1998751727,ExpressionUUID->"a07ca585-c570-493b-9633-633e220431c8"]
}, Open  ]]
}, Open  ]],

Cell["Some properties of the metric.", "Text",
 CellChangeTimes->{{3.731115160421138*^9, 3.731115186906599*^9}, 
   3.790867441915625*^9, {3.790867756472289*^9, 3.790867765432922*^9}},
 CellID->1924939986,ExpressionUUID->"430ba7bc-1b38-47a7-8273-eefbe1cf70d3"],

Cell[CellGroupData[{

Cell["\<\
The determinant is a density in a given basis, so one needs to specify the \
basis in which you want that density:\
\>", "MathCaption",
 CellChangeTimes->{{3.728675654182659*^9, 3.728675654342785*^9}, {
   3.7908676837545147`*^9, 3.7908677106746597`*^9}, 3.790867769880591*^9},
 CellID->1382136984,ExpressionUUID->"c320da86-3313-4ba5-9a5a-16f5c9bba7fc"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Determinant", "[", 
  RowBox[{"met", ",", "ch"}], "]"}]], "Input",
 CellLabel->"In[47]:=",
 CellID->245680489,ExpressionUUID->"84701a0f-e630-4b41-b650-e6f9c2d642fa"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"-", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "4"]}], " ", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"]}], ",", 
   RowBox[{"{", "}"}], ",", 
   RowBox[{"2", " ", "ch"}]}], "]"}]], "Output",
 CellChangeTimes->{
  3.731115657261827*^9, 3.73336764875675*^9, 3.7333678904532957`*^9, 
   3.73336812255801*^9, 3.790867661899293*^9, 3.792273804777055*^9, {
   3.794517912392432*^9, 3.794517936923242*^9}, 3.794518015421523*^9, 
   3.7945181226879673`*^9, 3.794518780676341*^9, 3.794519060710309*^9, 
   3.7945192961042757`*^9, 3.794520004945046*^9, 3.7945201807535686`*^9, 
   3.794520315409008*^9, 3.794520417090706*^9, 3.7945205864537086`*^9, 
   3.794520662493492*^9},
 CellLabel->"Out[47]=",
 CellID->2081069315,ExpressionUUID->"cd975ac3-843b-4d1e-ae72-5d226663d733"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
This is the signature of the metric, as was specified (or computed) in input:\
\
\>", "MathCaption",
 CellChangeTimes->{{3.790867782752799*^9, 3.7908677981918097`*^9}},
 CellID->57914906,ExpressionUUID->"775aa667-ed41-4cc8-b564-9907d1035f62"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SignatureOfMetric", "[", "met", "]"}]], "Input",
 CellLabel->"In[48]:=",
 CellID->1483626726,ExpressionUUID->"36cbc39b-7a6b-4c65-9ce0-08d1e04bfa3b"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"3", ",", "1", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{
  3.7286756566577597`*^9, 3.728753467826663*^9, 3.728843470145184*^9, 
   3.730955482727166*^9, 3.7309561946980143`*^9, 3.7311156597995977`*^9, 
   3.733367649167345*^9, 3.733367890784828*^9, 3.733368122799387*^9, 
   3.79086771799254*^9, 3.7922738064063387`*^9, {3.7945179124324617`*^9, 
   3.7945179369620457`*^9}, 3.7945180154743032`*^9, 3.794518122742835*^9, 
   3.794518780719207*^9, 3.794519060759492*^9, 3.79451929615353*^9, 
   3.794520004984046*^9, 3.79452018079401*^9, 3.794520315451696*^9, 
   3.7945204171303663`*^9, 3.794520586505185*^9, 3.794520662535359*^9},
 CellLabel->"Out[48]=",
 CellID->465818516,ExpressionUUID->"c61daeaa-30e5-41ab-9cf3-f2cedeefa15f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The sign of the metric is obtained from the signature, not from the \
expression of the determinant itself:\
\>", "MathCaption",
 CellChangeTimes->{{3.790867773056582*^9, 3.7908678186151047`*^9}},
 CellID->2095173567,ExpressionUUID->"9ef2b72b-f452-45e2-98a4-862fcdd745c3"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SignDetOfMetric", "[", "met", "]"}]], "Input",
 CellLabel->"In[49]:=",
 CellID->242647365,ExpressionUUID->"ad867df0-2c12-4615-8ea4-b4dd3b19e4a6"],

Cell[BoxData[
 RowBox[{"-", "1"}]], "Output",
 CellChangeTimes->{
  3.728675656044964*^9, 3.728753467693686*^9, 3.728843469433915*^9, 
   3.730955482125618*^9, 3.730956194594159*^9, 3.73111565896145*^9, 
   3.733367649026122*^9, 3.7333678906355667`*^9, 3.7333681226812553`*^9, 
   3.790867714008971*^9, 3.792273807781456*^9, {3.7945179124836283`*^9, 
   3.794517937002199*^9}, 3.794518015521976*^9, 3.794518122790901*^9, 
   3.7945187807936974`*^9, 3.794519060803194*^9, 3.794519296194895*^9, 
   3.794520005025042*^9, 3.794520180839004*^9, 3.7945203154925117`*^9, 
   3.7945204171710653`*^9, 3.794520586551001*^9, 3.794520662573267*^9},
 CellLabel->"Out[49]=",
 CellID->1433905097,ExpressionUUID->"7a98a1e7-44ef-472d-ab50-8cbf8ee11518"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.1.4. Covariant derivatives", "Subsubsection",
 CellChangeTimes->{{3.728674024194376*^9, 3.7286740245307503`*^9}, {
  3.733368004952485*^9, 3.7333680050726357`*^9}},
 CellID->2079037655,ExpressionUUID->"d30fc830-5584-44b4-8bae-d349dd7fcddc"],

Cell["\<\
In the same way that an abstract tensor T has a \"component version\" \
CTensor[...], a covariant derivative CD also has a component version \
CCovD[...]. In the same way that we can change from a covariant derivative to \
another one using Christoffel tensors, we can also change from a CCovD object \
to another using a Christoffel CTensor object. This subsection explains how \
that works.\
\>", "Text",
 CellChangeTimes->{{3.790867869454154*^9, 3.790868006811914*^9}},
 CellID->59022842,ExpressionUUID->"40c7b29c-084f-49de-96c5-041fc110be88"],

Cell[CellGroupData[{

Cell["\<\
This is the Levi-Civita connection associated to the metric. It has the \
structure CCovD[pd, christoffel, metric], meaning that it is the covariant \
derivative that can be obtained from the partial derivative pd using the \
given christoffel tensor. The metric tensor is stored to denote where the \
object came from:\
\>", "MathCaption",
 CellChangeTimes->{{3.728675671703569*^9, 3.7286756718621902`*^9}, {
  3.790868035675818*^9, 3.79086818064627*^9}},
 CellID->1299346037,ExpressionUUID->"87b55407-2b65-4b9f-896b-57988d7e36a0"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"cd", "=", 
  RowBox[{"CovDOfMetric", "[", "met", "]"}]}]], "Input",
 CellLabel->"In[50]:=",
 CellID->347637516,ExpressionUUID->"7651de6c-1489-4006-97db-263fa6885dd3"],

Cell[BoxData[
 RowBox[{"CCovD", "[", 
  RowBox[{"PDch", ",", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", 
           RowBox[{"-", 
            FractionBox[
             InterpretationBox[
              StyleBox["M",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`mass,
              Editable->False], 
             RowBox[{
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`mass,
                Editable->False], " ", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}], "-", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], "2"]}]]}], ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"-", 
            FractionBox[
             InterpretationBox[
              StyleBox["M",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`mass,
              Editable->False], 
             RowBox[{
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`mass,
                Editable->False], " ", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}], "-", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], "2"]}]]}], ",", "0", ",", "0", ",", "0"}], 
          "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           FractionBox[
            RowBox[{
             InterpretationBox[
              StyleBox["M",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`mass,
              Editable->False], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"-", "2"}], " ", 
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`mass,
                 Editable->False]}], "+", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}], ")"}]}], 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "3"]], ",", "0", ",", "0", ",", "0"}], "}"}], 
         ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", 
           FractionBox[
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`mass,
             Editable->False], 
            RowBox[{
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`mass,
               Editable->False], " ", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}], "-", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}]], ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           RowBox[{
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["M",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`mass,
              Editable->False]}], "-", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]}], ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           RowBox[{
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`mass,
                Editable->False]}], "-", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}], ")"}], " ", 
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["Sin",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Sin,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], "2"]}]}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           FractionBox["1", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]], ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", 
           FractionBox["1", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]], ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           RowBox[{
            RowBox[{"-", 
             RowBox[{
              InterpretationBox[
               StyleBox["Cos",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cos,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}]}], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["Sin",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Sin,
              Editable->False], "[", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False], "]"}]}]}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           FractionBox["1", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           RowBox[{
            InterpretationBox[
             StyleBox["Cot",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cot,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", 
           FractionBox["1", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]], ",", 
           RowBox[{
            InterpretationBox[
             StyleBox["Cot",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cot,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], ",", "0"}], "}"}]}], "}"}]}], "}"}], 
     ",", 
     RowBox[{"{", 
      RowBox[{"ch", ",", 
       RowBox[{"-", "ch"}], ",", 
       RowBox[{"-", "ch"}]}], "}"}], ",", "0"}], "]"}], ",", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "1"}], "+", 
          FractionBox[
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`mass,
             Editable->False]}], 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]]}], ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", 
         FractionBox["1", 
          RowBox[{"1", "-", 
           FractionBox[
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["M",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`mass,
              Editable->False]}], 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]]}]], ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"], ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Sin",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Sin,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "2"]}]}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "ch"}], ",", 
       RowBox[{"-", "ch"}]}], "}"}], ",", "0"}], "]"}]}], "]"}]], "Output",
 CellChangeTimes->{
  3.728675672950856*^9, 3.728753471438788*^9, 3.728843476183692*^9, 
   3.730955500954187*^9, 3.7309561947885942`*^9, 3.731115692263851*^9, 
   3.731115874223795*^9, 3.733367657693376*^9, 3.733367890951665*^9, 
   3.733368122893031*^9, 3.790868030837*^9, 3.792273814299039*^9, {
   3.794517912529595*^9, 3.7945179370467243`*^9}, 3.794518015577402*^9, 
   3.794518122845168*^9, 3.7945187808453903`*^9, 3.794519060847563*^9, 
   3.794519296239736*^9, 3.794520005079321*^9, 3.794520180887047*^9, 
   3.794520315540016*^9, 3.794520417219576*^9, 3.7945205865953712`*^9, 
   3.794520662619923*^9},
 CellLabel->"Out[50]=",
 CellID->581478645,ExpressionUUID->"f3f9faae-5483-4a8b-8aa0-f9c20b2fc6ff"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Now we can use it to compute covariant derivatives of tensor fields. The \
notation is identical to that used in ",
 StyleBox["xTensor",
  FontSlant->"Italic"],
 ":"
}], "MathCaption",
 CellChangeTimes->{{3.728675678933674*^9, 3.728675679078023*^9}},
 CellID->1997642041,ExpressionUUID->"1a0169ab-ee58-444a-8c7c-b14902e16745"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"cd", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"v", "[", "b", "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.73111569613713*^9, 3.7311156966490583`*^9}},
 CellLabel->"In[51]:=",
 CellID->610445889,ExpressionUUID->"e82d6b52-5d3b-46cb-ba3c-6099a652a8e7"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{"-", 
             FractionBox[
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`mass,
               Editable->False], 
              RowBox[{
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`mass,
                 Editable->False]}], "-", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}]]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{"-", 
                FractionBox[
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], $CellContext`mass, Editable -> False], 
                 RowBox[{
                   RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], "-", 
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}]]}]}]],
           
           Annotation[#, {1, 1} -> -$CellContext`mass/(
             2 $CellContext`mass - $CellContext`r[]), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{"1", "+", 
             FractionBox[
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`mass,
               Editable->False], 
              RowBox[{
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`mass,
                 Editable->False]}], "-", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}]]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{"1", "+", 
                FractionBox[
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], $CellContext`mass, Editable -> False], 
                 RowBox[{
                   RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], "-", 
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}]]}]}]],
           
           Annotation[#, {2, 2} -> 
            1 + $CellContext`mass/(2 $CellContext`mass - $CellContext`r[]), 
            "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {3, 3} -> 1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "4"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {4, 4} -> 1, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[0.6, 0.4, 0.2], 5 -> RGBColor[0.6, 0.4, 0.2]}, 
          "RowsIndexed" -> {
           1 -> RGBColor[0.6, 0.4, 0.2], 5 -> RGBColor[0.6, 0.4, 0.2]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"b", " "},
         {" ", "a"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{-$CellContext`mass/(
     2 $CellContext`mass - $CellContext`r[]), 0, 0, 0}, {
    0, 1 + $CellContext`mass/(2 $CellContext`mass - $CellContext`r[]), 0, 
     0}, {0, 0, 1, 0}, {0, 0, 0, 1}}, {$CellContext`ch, -$CellContext`ch}, 
   0][$CellContext`b, -$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.7286756801451397`*^9, 3.728753471664481*^9, 3.728843477578292*^9, 
   3.730955502921583*^9, 3.7309561949006968`*^9, 3.731115699800259*^9, 
   3.731115875569201*^9, 3.733367657921392*^9, 3.733367891085548*^9, 
   3.733368123016527*^9, 3.790868165151987*^9, 3.792273814382469*^9, {
   3.794517912587296*^9, 3.7945179370948057`*^9}, 3.794518015644855*^9, 
   3.794518122893846*^9, 3.794518780896983*^9, 3.79451906091503*^9, 
   3.794519296288472*^9, 3.794520005130928*^9, 3.7945201810003567`*^9, 
   3.7945203155959272`*^9, 3.7945204172884417`*^9, 3.794520586653779*^9, 
   3.794520662669187*^9},
 CellLabel->"Out[51]=",
 CellID->1860992665,ExpressionUUID->"0e73ae32-a018-4b9e-b3fd-f639d9c259cc"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The covariant derivative of the metric field is zero:", "MathCaption",
 CellChangeTimes->{{3.728675686110931*^9, 3.728675686285577*^9}},
 CellID->43269167,ExpressionUUID->"bb1b6577-128a-4ab0-8035-4ec8c11b1d54"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"cd", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"met", "[", 
   RowBox[{
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}]}], "]"}], "]"}]], "Input",
 CellLabel->"In[52]:=",
 CellID->1375417083,ExpressionUUID->"8bef311c-51d9-47f1-8b61-9533e7d9439d"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{
  3.728675687466888*^9, 3.72875347183173*^9, 3.7288434788769827`*^9, 
   3.7309555041695642`*^9, 3.7309561949898367`*^9, 3.73111570342976*^9, 
   3.731115876926914*^9, 3.733367658072053*^9, 3.733367891202523*^9, 
   3.733368123112833*^9, 3.7908681683174887`*^9, 3.7922738144382257`*^9, {
   3.7945179126431704`*^9, 3.794517937148679*^9}, 3.794518015750821*^9, 
   3.794518122945445*^9, 3.794518780949193*^9, 3.794519060967469*^9, 
   3.7945192963352623`*^9, 3.794520005189999*^9, 3.794520181062067*^9, 
   3.7945203156419353`*^9, 3.794520417336257*^9, 3.7945205867078238`*^9, 
   3.794520662718184*^9},
 CellLabel->"Out[52]=",
 CellID->1751885747,ExpressionUUID->"320422c9-7cef-4e49-82ef-28150e3d70d0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
This is the Christoffel tensor of the Levi-Civita connection with respect to \
the partial derivative of the chart:\
\>", "MathCaption",
 CellChangeTimes->{{3.731115795791019*^9, 3.731115819973092*^9}},
 CellID->766568751,ExpressionUUID->"ab6e4910-2c72-4976-aa09-ccbd973cfbed"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Christoffel", "[", 
  RowBox[{"cd", ",", "PDch"}], "]"}]], "Input",
 CellChangeTimes->{{3.731115821688629*^9, 3.7311158302762423`*^9}},
 CellLabel->"In[53]:=",
 CellID->1024019936,ExpressionUUID->"ca478766-7e5c-473e-86ca-3e1919bab2ea"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", 
         RowBox[{"-", 
          FractionBox[
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`mass,
            Editable->False], 
           RowBox[{
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["M",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`mass,
              Editable->False], " ", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]}], "-", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"]}]]}], ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", 
          FractionBox[
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`mass,
            Editable->False], 
           RowBox[{
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["M",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`mass,
              Editable->False], " ", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]}], "-", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"]}]]}], ",", "0", ",", "0", ",", "0"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         FractionBox[
          RowBox[{
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`mass,
            Editable->False], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "2"}], " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`mass,
               Editable->False]}], "+", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]}], ")"}]}], 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "3"]], ",", "0", ",", "0", ",", "0"}], "}"}], 
       ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", 
         FractionBox[
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`mass,
           Editable->False], 
          RowBox[{
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`mass,
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]}], "-", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}]], ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", 
         RowBox[{
          RowBox[{"2", " ", 
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`mass,
            Editable->False]}], "-", 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False]}], ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["M",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`mass,
              Editable->False]}], "-", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]}], ")"}], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Sin",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Sin,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "2"]}]}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", 
         FractionBox["1", 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False]], ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", 
         FractionBox["1", 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False]], ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", 
         RowBox[{
          RowBox[{"-", 
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}]}], " ", 
          RowBox[{
           InterpretationBox[
            StyleBox["Sin",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Sin,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}]}]}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", 
         FractionBox["1", 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False]]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", 
         RowBox[{
          InterpretationBox[
           StyleBox["Cot",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Cot,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", 
         FractionBox["1", 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False]], ",", 
         RowBox[{
          InterpretationBox[
           StyleBox["Cot",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Cot,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], ",", "0"}], "}"}]}], "}"}]}], "}"}], ",", 
   
   RowBox[{"{", 
    RowBox[{"ch", ",", 
     RowBox[{"-", "ch"}], ",", 
     RowBox[{"-", "ch"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{
  3.7311158316779747`*^9, 3.731115878652367*^9, 3.733367658218239*^9, 
   3.733367891318388*^9, 3.733368123233034*^9, 3.7908681843522787`*^9, 
   3.792273814490683*^9, {3.7945179126885443`*^9, 3.79451793724623*^9}, 
   3.794518015807645*^9, 3.7945181229863234`*^9, 3.794518780990553*^9, 
   3.7945190610132713`*^9, 3.794519296378108*^9, 3.794520005233053*^9, 
   3.794520181227426*^9, 3.794520315697006*^9, 3.794520417379716*^9, 
   3.79452058675739*^9, 3.794520662761939*^9},
 CellLabel->"Out[53]=",
 CellID->1942949459,ExpressionUUID->"30d33be7-e97c-4859-a4f3-2ddb27bbd1f9"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "[", 
  RowBox[{"a", ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.73111583524755*^9, 3.731115839004216*^9}},
 CellLabel->"In[54]:=",
 CellID->888420665,ExpressionUUID->"3cafbb67-84ef-437d-8e47-f432b1804284"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                FractionBox[
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`mass,
                  Editable->False], 
                 RowBox[{
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`mass,
                    Editable->False], " ", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], "-", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]]}]}]],
              
              Annotation[#, {1, 1, 2} -> -$CellContext`mass/(
                2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 3} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "4"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
          GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                FractionBox[
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`mass,
                  Editable->False], 
                 RowBox[{
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`mass,
                    Editable->False], " ", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], "-", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]]}]}]],
              Annotation[#, {1, 2, 1} -> -$CellContext`mass/(
                2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 3} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "4"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
          GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 3} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "4"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
          GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 3} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "4"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox[
               FractionBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`mass,
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`mass,
                    Editable->False]}], "+", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}]}], 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "3"]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}], 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "3"]]}]],
              
              Annotation[#, {2, 1, 
                1} -> $CellContext`mass $CellContext`r[]^(-3) ((-2) \
$CellContext`mass + $CellContext`r[]), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 3} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "4"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
          GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               FractionBox[
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`mass,
                 Editable->False], 
                RowBox[{
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`mass,
                   Editable->False], " ", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}], "-", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                 FractionBox[
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], 
                  RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]]}]],
              
              Annotation[#, {2, 2, 2} -> $CellContext`mass/(
                2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 3} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "4"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
          GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`mass,
                  Editable->False]}], "-", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "3"}], "}"}], "\[Rule]", 
                 RowBox[{
                   RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], "-", 
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}]}]],
              
              Annotation[#, {2, 3, 3} -> 
               2 $CellContext`mass - $CellContext`r[], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "4"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
          GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 3} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`mass,
                    Editable->False]}], "-", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}], ")"}], " ", 
                SuperscriptBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Sin",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Sin,
                   Editable->False], "[", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False], "]"}], "2"]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "4"}], "}"}], "\[Rule]", 
                 RowBox[{
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                    "2"]}]}]],
              
              Annotation[#, {2, 4, 
                4} -> (2 $CellContext`mass - $CellContext`r[]) Sin[
                   $CellContext`\[Theta][]]^2, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 3} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "4"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
          GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               FractionBox["1", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "3"}], "}"}], "\[Rule]", 
                 FractionBox["1", 
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False]]}]],
              Annotation[#, {3, 2, 3} -> $CellContext`r[]^(-1), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "4"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
          GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               FractionBox["1", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "2"}], "}"}], "\[Rule]", 
                 FractionBox["1", 
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False]]}]],
              Annotation[#, {3, 3, 2} -> $CellContext`r[]^(-1), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 3} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "4"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
          GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 3} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                RowBox[{"-", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False], "]"}]}], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Sin",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Sin,
                  Editable->False], "[", 
                 InterpretationBox[
                  StyleBox["\[Theta]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Theta][],
                  Editable->False], "]"}]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "4"}], "}"}], "\[Rule]", 
                 RowBox[{
                   RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Cos"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cos, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}]}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}]}]}]],
              Annotation[#, {3, 4, 4} -> -Cos[
                  $CellContext`\[Theta][]] Sin[
                  $CellContext`\[Theta][]], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 3} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "4"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
          GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 3} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               FractionBox["1", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "4"}], "}"}], "\[Rule]", 
                 FractionBox["1", 
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False]]}]],
              Annotation[#, {4, 2, 4} -> $CellContext`r[]^(-1), 
               "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
          GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 3} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["Cot",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 Cot,
                 Editable->False], "[", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "4"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Cot"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cot, Editable -> False], "[", 
                   InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}]}]],
              Annotation[#, {4, 3, 4} -> Cot[
                 $CellContext`\[Theta][]], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
          GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               FractionBox["1", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "2"}], "}"}], "\[Rule]", 
                 FractionBox["1", 
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False]]}]],
              Annotation[#, {4, 4, 2} -> $CellContext`r[]^(-1), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["Cot",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 Cot,
                 Editable->False], "[", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "3"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Cot"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cot, Editable -> False], "[", 
                   InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}]}]],
              Annotation[#, {4, 4, 3} -> Cot[
                 $CellContext`\[Theta][]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "4"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[0.6, 0.4, 0.2], 5 -> RGBColor[0.6, 0.4, 0.2]}, 
          "RowsIndexed" -> {
           1 -> RGBColor[0.6, 0.4, 0.2], 5 -> RGBColor[0.6, 0.4, 0.2]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a", " ", " "},
         {" ", "b", "c"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{{
     0, -$CellContext`mass/(
      2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2), 0, 
      0}, {-$CellContext`mass/(
      2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2), 0, 0, 0}, {
     0, 0, 0, 0}, {0, 0, 0, 
     0}}, {{$CellContext`mass $CellContext`r[]^(-3) ((-2) $CellContext`mass + \
$CellContext`r[]), 0, 0, 0}, {
     0, $CellContext`mass/(
      2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2), 0, 0}, {
     0, 0, 2 $CellContext`mass - $CellContext`r[], 0}, {
     0, 0, 0, (2 $CellContext`mass - $CellContext`r[]) Sin[
         $CellContext`\[Theta][]]^2}}, {{0, 0, 0, 0}, {
     0, 0, $CellContext`r[]^(-1), 0}, {0, $CellContext`r[]^(-1), 0, 0}, {
     0, 0, 0, -Cos[
        $CellContext`\[Theta][]] Sin[
        $CellContext`\[Theta][]]}}, {{0, 0, 0, 0}, {
     0, 0, 0, $CellContext`r[]^(-1)}, {0, 0, 0, 
      Cot[
       $CellContext`\[Theta][]]}, {0, $CellContext`r[]^(-1), 
      Cot[
       $CellContext`\[Theta][]], 
      0}}}, {$CellContext`ch, -$CellContext`ch, -$CellContext`ch}, 
   0][$CellContext`a, -$CellContext`b, -$CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.7311158399524813`*^9, 3.73111588057057*^9, 3.7333676585165586`*^9, 
   3.733367891536249*^9, 3.73336812333564*^9, 3.7908681871117496`*^9, 
   3.792273814556056*^9, {3.794517912752749*^9, 3.794517937393937*^9}, 
   3.794518015870791*^9, 3.794518123057809*^9, 3.79451878104897*^9, 
   3.7945190610845613`*^9, 3.7945192964368763`*^9, 3.794520005364032*^9, 
   3.794520181297667*^9, 3.7945203157564096`*^9, 3.794520417435657*^9, 
   3.794520586824445*^9, 3.7945206628318233`*^9},
 CellLabel->"Out[54]=",
 CellID->515221131,ExpressionUUID->"6d14b6e8-fdb9-4970-9fb3-3a2d32592f4e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToBasisExpand", "[", "%", "]"}]], "Input",
 CellChangeTimes->{{3.7908682035452347`*^9, 3.790868237092312*^9}},
 CellLabel->"In[55]:=",
 CellID->1369706427,ExpressionUUID->"44596de9-a3f2-4b36-8b45-3e92a627dfc5"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0.6, 0.4, 0.2]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`ch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0.6, 0.4, 0.2]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`ch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["3",
             FontColor->RGBColor[0.6, 0.4, 0.2]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`ch}, $CellContext`a],
    Editable->False], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Cot",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cot,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0.6, 0.4, 0.2]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`ch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0.6, 0.4, 0.2]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`ch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["3",
             FontColor->RGBColor[0.6, 0.4, 0.2]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`ch}, $CellContext`a],
    Editable->False], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Cot",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cot,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0.6, 0.4, 0.2]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`ch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0.6, 0.4, 0.2]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`ch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["1",
             FontColor->RGBColor[0.6, 0.4, 0.2]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`ch}, $CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`mass,
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}], ")"}]}], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0.6, 0.4, 0.2]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`ch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0.6, 0.4, 0.2]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`ch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["2",
              FontColor->RGBColor[0.6, 0.4, 0.2]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`ch}, $CellContext`a],
     Editable->False]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0.6, 0.4, 0.2]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`ch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0.6, 0.4, 0.2]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`ch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["2",
              FontColor->RGBColor[0.6, 0.4, 0.2]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`ch}, $CellContext`a],
     Editable->False]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0.6, 0.4, 0.2]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`ch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0.6, 0.4, 0.2]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`ch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["3",
              FontColor->RGBColor[0.6, 0.4, 0.2]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`ch}, $CellContext`a],
     Editable->False]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0.6, 0.4, 0.2]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`ch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0.6, 0.4, 0.2]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`ch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["3",
              FontColor->RGBColor[0.6, 0.4, 0.2]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`ch}, $CellContext`a],
     Editable->False]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mass,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[0.6, 0.4, 0.2]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`ch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[0.6, 0.4, 0.2]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`ch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["1",
              FontColor->RGBColor[0.6, 0.4, 0.2]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`ch}, $CellContext`a],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`mass,
        Editable->False]}], "+", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}]}], 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "3"]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mass,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0.6, 0.4, 0.2]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`ch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[0.6, 0.4, 0.2]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`ch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["0",
              FontColor->RGBColor[0.6, 0.4, 0.2]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`ch}, $CellContext`a],
     Editable->False]}], 
   RowBox[{
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`mass,
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}], "-", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mass,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[0.6, 0.4, 0.2]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`ch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0.6, 0.4, 0.2]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`ch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["0",
              FontColor->RGBColor[0.6, 0.4, 0.2]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`ch}, $CellContext`a],
     Editable->False]}], 
   RowBox[{
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`mass,
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}], "-", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mass,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0.6, 0.4, 0.2]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`ch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0.6, 0.4, 0.2]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`ch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["1",
              FontColor->RGBColor[0.6, 0.4, 0.2]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`ch}, $CellContext`a],
     Editable->False]}], 
   RowBox[{
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`mass,
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}], "-", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0.6, 0.4, 0.2]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`ch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0.6, 0.4, 0.2]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`ch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["2",
             FontColor->RGBColor[0.6, 0.4, 0.2]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`ch}, $CellContext`a],
    Editable->False], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Cos",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cos,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Sin",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Sin,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0.6, 0.4, 0.2]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`ch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0.6, 0.4, 0.2]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`ch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["1",
             FontColor->RGBColor[0.6, 0.4, 0.2]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`ch}, $CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`mass,
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}], ")"}], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}]}]], "Output",
 CellChangeTimes->{
  3.790868210393156*^9, 3.792273814668347*^9, {3.794517912851956*^9, 
   3.7945179374958677`*^9}, 3.794518015990697*^9, 3.794518123170982*^9, 
   3.7945187811307793`*^9, 3.794519061173709*^9, 3.7945192965316*^9, 
   3.7945200055228024`*^9, 3.7945201814007387`*^9, 3.794520315874297*^9, 
   3.7945204175389547`*^9, 3.794520586931527*^9, 3.794520662985681*^9},
 CellLabel->"Out[55]=",
 CellID->667484013,ExpressionUUID->"f06a09a7-36a2-4a82-9c9b-5c00db0ddb0e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Commuting covariant derivatives:", "MathCaption",
 CellChangeTimes->{{3.7286756924288807`*^9, 3.7286756925893307`*^9}},
 CellID->1872918627,ExpressionUUID->"38bfe98c-0c6d-4d93-a83b-0b04befe42dd"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"comm", "=", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"cd", "[", 
      RowBox[{"-", "b"}], "]"}], "@", 
     RowBox[{
      RowBox[{"cd", "[", 
       RowBox[{"-", "a"}], "]"}], "@", 
      RowBox[{"v", "[", "c", "]"}]}]}], "-", 
    RowBox[{
     RowBox[{"cd", "[", 
      RowBox[{"-", "a"}], "]"}], "@", 
     RowBox[{
      RowBox[{"cd", "[", 
       RowBox[{"-", "b"}], "]"}], "@", 
      RowBox[{"v", "[", "c", "]"}]}]}]}], "//", "Simplify"}]}]], "Input",
 CellChangeTimes->{{3.731115714728816*^9, 3.731115722824421*^9}},
 CellLabel->"In[56]:=",
 CellID->328066263,ExpressionUUID->"0485410d-1be7-4e51-be20-d1e140f2ff8a"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               FractionBox[
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`mass,
                  Editable->False]}], 
                RowBox[{
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`mass,
                   Editable->False], " ", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}], "-", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], 
                  RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]]}]],
              
              Annotation[#, {1, 1, 2} -> 
               2 $CellContext`mass/(
                2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 3} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "4"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
          GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                FractionBox[
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`mass,
                   Editable->False]}], 
                 RowBox[{
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`mass,
                    Editable->False], " ", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], "-", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]]}]}]],
              Annotation[#, {1, 2, 1} -> (-2) $CellContext`mass/(
                2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 3} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "4"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
          GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 3} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "4"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
          GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 3} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "4"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 3} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "4"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
          GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 3} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "4"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
          GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 3} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "4"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
          GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 3} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "4"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 3} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "4"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
          GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               FractionBox[
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`mass,
                 Editable->False], 
                RowBox[{
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`mass,
                   Editable->False], " ", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}], "-", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "3"}], "}"}], "\[Rule]", 
                 FractionBox[
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], 
                  RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]]}]],
              
              Annotation[#, {3, 2, 3} -> $CellContext`mass/(
                2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "4"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
          GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                FractionBox[
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`mass,
                  Editable->False], 
                 RowBox[{
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`mass,
                    Editable->False], " ", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], "-", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]]}]}]],
              
              Annotation[#, {3, 3, 2} -> -$CellContext`mass/(
                2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 3} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "4"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
          GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 3} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "4"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 3} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "4"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
          GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 3} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               FractionBox[
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`mass,
                 Editable->False], 
                RowBox[{
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`mass,
                   Editable->False], " ", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}], "-", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "4"}], "}"}], "\[Rule]", 
                 FractionBox[
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], 
                  RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]]}]],
              
              Annotation[#, {4, 2, 4} -> $CellContext`mass/(
                2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2), 
               "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
          GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 3} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "4"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
          GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                FractionBox[
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`mass,
                  Editable->False], 
                 RowBox[{
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`mass,
                    Editable->False], " ", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], "-", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]]}]}]],
              
              Annotation[#, {4, 4, 2} -> -$CellContext`mass/(
                2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 3} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "4"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[0.6, 0.4, 0.2], 5 -> RGBColor[0.6, 0.4, 0.2]}, 
          "RowsIndexed" -> {
           1 -> RGBColor[0.6, 0.4, 0.2], 5 -> RGBColor[0.6, 0.4, 0.2]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"c", " ", " "},
         {" ", "a", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{{
     0, 2 $CellContext`mass/(
      2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2), 0, 
      0}, {(-2) $CellContext`mass/(
      2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2), 0, 0, 0}, {
     0, 0, 0, 0}, {0, 0, 0, 0}}, {{0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0}, {
    0, 0, 0, 0}}, {{0, 0, 0, 0}, {
     0, 0, $CellContext`mass/(
      2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2), 0}, {
     0, -$CellContext`mass/(
      2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2), 0, 0}, {0, 
     0, 0, 0}}, {{0, 0, 0, 0}, {
     0, 0, 0, $CellContext`mass/(
      2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2)}, {0, 0, 0, 
     0}, {0, -$CellContext`mass/(
      2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2), 0, 
      0}}}, {$CellContext`ch, -$CellContext`ch, -$CellContext`ch}, 
   0][$CellContext`c, -$CellContext`a, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.728675693590768*^9, 3.7287534720122128`*^9, 3.728843480126709*^9, 
   3.7309555054550257`*^9, 3.730956195122137*^9, 3.731115723765023*^9, 
   3.731115882753975*^9, 3.733367658824644*^9, 3.7333678917193747`*^9, 
   3.733368123461193*^9, 3.790868258533386*^9, 3.792273814758573*^9, {
   3.7945179129354553`*^9, 3.79451793757367*^9}, 3.794518016068852*^9, 
   3.794518123255974*^9, 3.794518781214658*^9, 3.794519061258699*^9, 
   3.79451929661547*^9, 3.794520005612336*^9, 3.794520181484859*^9, 
   3.794520315959774*^9, 3.794520417621417*^9, 3.794520587024825*^9, 
   3.7945206630747633`*^9},
 CellLabel->"Out[56]=",
 CellID->2096718210,ExpressionUUID->"efc799bf-d859-4a71-aa02-f771e8c3c194"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We can check that the result coincides with the expected contraction with the \
Riemann tensor of the connection:\
\>", "MathCaption",
 CellChangeTimes->{{3.728675699132769*^9, 3.728675699277088*^9}},
 CellID->1287232317,ExpressionUUID->"4bef990f-c921-4e42-a3a1-ce13f8fc0aa6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"comm", "==", 
  RowBox[{
   RowBox[{
    RowBox[{"Riemann", "[", "cd", "]"}], "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "d"}], ",", "c"}], "]"}], 
   RowBox[{"v", "[", "d", "]"}]}]}]], "Input",
 CellChangeTimes->{{3.731115726704123*^9, 3.731115729951963*^9}},
 CellLabel->"In[57]:=",
 CellID->763291941,ExpressionUUID->"342d542e-f304-4dc0-aea0-aadf66e08c9a"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       StyleBox[GridBox[{
          {GridBox[{
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {1, 1, 1} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox[
                FractionBox[
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`mass,
                   Editable->False]}], 
                 RowBox[{
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`mass,
                    Editable->False], " ", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], "-", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]],
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                  FractionBox[
                   RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], 
                   RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]]}]],
               
               Annotation[#, {1, 1, 2} -> 
                2 $CellContext`mass/(
                 2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2), 
                "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "3"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {1, 1, 3} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "4"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {1, 1, 4} -> 0, "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
           GridBox[{
             {
              TagBox[
               TooltipBox[
                RowBox[{"-", 
                 FractionBox[
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`mass,
                    Editable->False]}], 
                  RowBox[{
                   RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`mass,
                    Editable->False], " ", 
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], "-", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]]}],
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                  RowBox[{"-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]]}]}]],
               
               Annotation[#, {1, 2, 1} -> (-2) $CellContext`mass/(
                 2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2), 
                "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {1, 2, 2} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "3"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {1, 2, 3} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "4"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {1, 2, 4} -> 0, "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
           GridBox[{
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "3", ",", "1"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {1, 3, 1} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "3", ",", "2"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {1, 3, 2} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "3", ",", "3"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {1, 3, 3} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "3", ",", "4"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {1, 3, 4} -> 0, "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
           GridBox[{
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "4", ",", "1"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {1, 4, 1} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "4", ",", "2"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {1, 4, 2} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "4", ",", "3"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {1, 4, 3} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "4", ",", "4"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {1, 4, 4} -> 0, "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}]},
          
          {GridBox[{
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {2, 1, 1} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {2, 1, 2} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "3"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {2, 1, 3} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "4"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {2, 1, 4} -> 0, "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
           GridBox[{
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {2, 2, 1} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {2, 2, 2} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "3"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {2, 2, 3} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "4"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {2, 2, 4} -> 0, "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
           GridBox[{
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "3", ",", "1"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {2, 3, 1} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "3", ",", "2"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {2, 3, 2} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "3", ",", "3"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {2, 3, 3} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "3", ",", "4"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {2, 3, 4} -> 0, "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
           GridBox[{
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "4", ",", "1"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {2, 4, 1} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "4", ",", "2"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {2, 4, 2} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "4", ",", "3"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {2, 4, 3} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "4", ",", "4"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {2, 4, 4} -> 0, "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> 
                RGBColor[0.6, 0.4, 0.2]}}]},
          {GridBox[{
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {3, 1, 1} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {3, 1, 2} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "1", ",", "3"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {3, 1, 3} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "1", ",", "4"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {3, 1, 4} -> 0, "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
           GridBox[{
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {3, 2, 1} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {3, 2, 2} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox[
                FractionBox[
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`mass,
                  Editable->False], 
                 RowBox[{
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`mass,
                    Editable->False], " ", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], "-", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]],
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "2", ",", "3"}], "}"}], "\[Rule]", 
                  FractionBox[
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], 
                   RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]]}]],
               
               Annotation[#, {3, 2, 3} -> $CellContext`mass/(
                 2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2), 
                "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "2", ",", "4"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {3, 2, 4} -> 0, "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
           GridBox[{
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "3", ",", "1"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {3, 3, 1} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox[
                RowBox[{"-", 
                 FractionBox[
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`mass,
                   Editable->False], 
                  RowBox[{
                   RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`mass,
                    Editable->False], " ", 
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], "-", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]]}],
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "3", ",", "2"}], "}"}], "\[Rule]", 
                  RowBox[{"-", 
                    FractionBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]]}]}]],
               
               Annotation[#, {3, 3, 2} -> -$CellContext`mass/(
                 2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2), 
                "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "3", ",", "3"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {3, 3, 3} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "3", ",", "4"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {3, 3, 4} -> 0, "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
           GridBox[{
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "4", ",", "1"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {3, 4, 1} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "4", ",", "2"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {3, 4, 2} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "4", ",", "3"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {3, 4, 3} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "4", ",", "4"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {3, 4, 4} -> 0, "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> 
                RGBColor[0.6, 0.4, 0.2]}}]},
          {GridBox[{
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {4, 1, 1} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {4, 1, 2} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "1", ",", "3"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {4, 1, 3} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "1", ",", "4"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {4, 1, 4} -> 0, "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
           GridBox[{
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {4, 2, 1} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {4, 2, 2} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "2", ",", "3"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {4, 2, 3} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox[
                FractionBox[
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`mass,
                  Editable->False], 
                 RowBox[{
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`mass,
                    Editable->False], " ", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], "-", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]],
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "2", ",", "4"}], "}"}], "\[Rule]", 
                  FractionBox[
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], 
                   RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]]}]],
               
               Annotation[#, {4, 2, 4} -> $CellContext`mass/(
                 2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2), 
                "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
           GridBox[{
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "3", ",", "1"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {4, 3, 1} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "3", ",", "2"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {4, 3, 2} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "3", ",", "3"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {4, 3, 3} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "3", ",", "4"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {4, 3, 4} -> 0, "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
           GridBox[{
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "4", ",", "1"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {4, 4, 1} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox[
                RowBox[{"-", 
                 FractionBox[
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`mass,
                   Editable->False], 
                  RowBox[{
                   RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`mass,
                    Editable->False], " ", 
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], "-", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]]}],
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "4", ",", "2"}], "}"}], "\[Rule]", 
                  RowBox[{"-", 
                    FractionBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]]}]}]],
               Annotation[#, {4, 4, 2} -> -$CellContext`mass/(
                 2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2), 
                "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "4", ",", "3"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {4, 4, 3} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "4", ",", "4"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {4, 4, 4} -> 0, "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}]}
         },
         AutoDelete->False,
         GridBoxDividers->{
          "ColumnsIndexed" -> {
            1 -> RGBColor[0.6, 0.4, 0.2], 5 -> RGBColor[0.6, 0.4, 0.2]}, 
           "RowsIndexed" -> {
            1 -> RGBColor[0.6, 0.4, 0.2], 5 -> RGBColor[0.6, 0.4, 0.2]}}],
        FontSize->Small], 
       StyleBox[GridBox[{
          {"c", " ", " "},
          {" ", "a", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xCoba`CTensor[{{{
      0, 2 $CellContext`mass/(
       2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2), 0, 
       0}, {(-2) $CellContext`mass/(
       2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2), 0, 0, 0}, {
      0, 0, 0, 0}, {0, 0, 0, 0}}, {{0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0}, {
     0, 0, 0, 0}}, {{0, 0, 0, 0}, {
      0, 0, $CellContext`mass/(
       2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2), 0}, {
      0, -$CellContext`mass/(
       2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2), 0, 0}, {0, 
      0, 0, 0}}, {{0, 0, 0, 0}, {
      0, 0, 0, $CellContext`mass/(
       2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2)}, {0, 0, 0, 
      0}, {0, -$CellContext`mass/(
       2 $CellContext`mass $CellContext`r[] - $CellContext`r[]^2), 0, 
       0}}}, {$CellContext`ch, -$CellContext`ch, -$CellContext`ch}, 
    0][$CellContext`c, -$CellContext`a, -$CellContext`b],
   Editable->False], "\[Equal]", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       StyleBox[GridBox[{
          {GridBox[{
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {1, 1, 1} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {1, 1, 2} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "3"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {1, 1, 3} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "4"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {1, 1, 4} -> 0, "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
           GridBox[{
             {
              TagBox[
               TooltipBox[
                FractionBox[
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`mass,
                   Editable->False]}], 
                 RowBox[{
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`mass,
                    Editable->False]}], "-", 
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}], " ", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}]],
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                  FractionBox[
                   RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], 
                   RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}]]}]],
               
               Annotation[#, {1, 2, 1} -> 
                2 $CellContext`mass (
                   2 $CellContext`mass - \
$CellContext`r[])^(-1)/$CellContext`r[], "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {1, 2, 2} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "3"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {1, 2, 3} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "4"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {1, 2, 4} -> 0, "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
           GridBox[{
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "3", ",", "1"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {1, 3, 1} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "3", ",", "2"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {1, 3, 2} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "3", ",", "3"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {1, 3, 3} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "3", ",", "4"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {1, 3, 4} -> 0, "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
           GridBox[{
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "4", ",", "1"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {1, 4, 1} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "4", ",", "2"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {1, 4, 2} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "4", ",", "3"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {1, 4, 3} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "4", ",", "4"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {1, 4, 4} -> 0, "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> 
                RGBColor[0.6, 0.4, 0.2]}}]},
          {GridBox[{
             {
              TagBox[
               TooltipBox[
                RowBox[{"-", 
                 FractionBox[
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`mass,
                    Editable->False]}], 
                  RowBox[{
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`mass,
                    Editable->False]}], "-", 
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}], " ", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}]]}],
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                  RowBox[{"-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}]]}]}]],
               
               Annotation[#, {2, 1, 
                 1} -> (-2) $CellContext`mass (
                   2 $CellContext`mass - \
$CellContext`r[])^(-1)/$CellContext`r[], "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {2, 1, 2} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "3"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {2, 1, 3} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "4"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {2, 1, 4} -> 0, "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
           GridBox[{
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {2, 2, 1} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {2, 2, 2} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "3"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {2, 2, 3} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "4"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {2, 2, 4} -> 0, "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
           GridBox[{
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "3", ",", "1"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {2, 3, 1} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "3", ",", "2"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {2, 3, 2} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox[
                FractionBox[
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`mass,
                  Editable->False], 
                 RowBox[{
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`mass,
                    Editable->False]}], "-", 
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}], " ", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}]],
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "3", ",", "3"}], "}"}], "\[Rule]", 
                  FractionBox[
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], 
                   RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}]]}]],
               
               Annotation[#, {2, 3, 
                 3} -> $CellContext`mass (
                   2 $CellContext`mass - \
$CellContext`r[])^(-1)/$CellContext`r[], "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "3", ",", "4"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {2, 3, 4} -> 0, "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
           GridBox[{
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "4", ",", "1"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {2, 4, 1} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "4", ",", "2"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {2, 4, 2} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "4", ",", "3"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {2, 4, 3} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox[
                FractionBox[
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`mass,
                  Editable->False], 
                 RowBox[{
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`mass,
                    Editable->False]}], "-", 
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}], " ", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}]],
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "4", ",", "4"}], "}"}], "\[Rule]", 
                  FractionBox[
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], 
                   RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}]]}]],
               
               Annotation[#, {2, 4, 
                 4} -> $CellContext`mass (
                   2 $CellContext`mass - \
$CellContext`r[])^(-1)/$CellContext`r[], "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> 
                RGBColor[0.6, 0.4, 0.2]}}]},
          {GridBox[{
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {3, 1, 1} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {3, 1, 2} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "1", ",", "3"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {3, 1, 3} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "1", ",", "4"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {3, 1, 4} -> 0, "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
           GridBox[{
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {3, 2, 1} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {3, 2, 2} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox[
                RowBox[{"-", 
                 FractionBox[
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`mass,
                   Editable->False], 
                  RowBox[{
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`mass,
                    Editable->False]}], "-", 
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}], " ", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}]]}],
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "2", ",", "3"}], "}"}], "\[Rule]", 
                  RowBox[{"-", 
                    FractionBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}]]}]}]],
               
               Annotation[#, {3, 2, 
                 3} -> -$CellContext`mass (
                   2 $CellContext`mass - \
$CellContext`r[])^(-1)/$CellContext`r[], "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "2", ",", "4"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {3, 2, 4} -> 0, "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
           GridBox[{
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "3", ",", "1"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {3, 3, 1} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "3", ",", "2"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {3, 3, 2} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "3", ",", "3"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {3, 3, 3} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "3", ",", "4"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {3, 3, 4} -> 0, "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
           GridBox[{
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "4", ",", "1"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {3, 4, 1} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "4", ",", "2"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {3, 4, 2} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "4", ",", "3"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {3, 4, 3} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"3", ",", "4", ",", "4"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {3, 4, 4} -> 0, "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> 
                RGBColor[0.6, 0.4, 0.2]}}]},
          {GridBox[{
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {4, 1, 1} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {4, 1, 2} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "1", ",", "3"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {4, 1, 3} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "1", ",", "4"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {4, 1, 4} -> 0, "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
           GridBox[{
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {4, 2, 1} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {4, 2, 2} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "2", ",", "3"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {4, 2, 3} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox[
                RowBox[{"-", 
                 FractionBox[
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`mass,
                   Editable->False], 
                  RowBox[{
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`mass,
                    Editable->False]}], "-", 
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}], " ", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}]]}],
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "2", ",", "4"}], "}"}], "\[Rule]", 
                  RowBox[{"-", 
                    FractionBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}]]}]}]],
               
               Annotation[#, {4, 2, 
                 4} -> -$CellContext`mass (
                   2 $CellContext`mass - \
$CellContext`r[])^(-1)/$CellContext`r[], "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
           GridBox[{
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "3", ",", "1"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {4, 3, 1} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "3", ",", "2"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {4, 3, 2} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "3", ",", "3"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {4, 3, 3} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "3", ",", "4"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {4, 3, 4} -> 0, "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}], 
           GridBox[{
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "4", ",", "1"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {4, 4, 1} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "4", ",", "2"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {4, 4, 2} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "4", ",", "3"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {4, 4, 3} -> 0, "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["0",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"4", ",", "4", ",", "4"}], "}"}], "\[Rule]", 
                  "0"}]],
               Annotation[#, {4, 4, 4} -> 0, "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[0.6, 0.4, 0.2], 2 -> RGBColor[0.6, 0.4, 0.2]}}]}
         },
         AutoDelete->False,
         GridBoxDividers->{
          "ColumnsIndexed" -> {
            1 -> RGBColor[0.6, 0.4, 0.2], 5 -> RGBColor[0.6, 0.4, 0.2]}, 
           "RowsIndexed" -> {
            1 -> RGBColor[0.6, 0.4, 0.2], 5 -> RGBColor[0.6, 0.4, 0.2]}}],
        FontSize->Small], 
       StyleBox[GridBox[{
          {" ", " ", "c"},
          {"a", "b", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xCoba`CTensor[{{{0, 0, 0, 0}, {
      2 $CellContext`mass (
         2 $CellContext`mass - $CellContext`r[])^(-1)/$CellContext`r[], 0, 0, 
       0}, {0, 0, 0, 0}, {0, 0, 0, 
      0}}, {{(-2) $CellContext`mass (
         2 $CellContext`mass - $CellContext`r[])^(-1)/$CellContext`r[], 0, 0, 
       0}, {0, 0, 0, 0}, {
      0, 0, $CellContext`mass (
         2 $CellContext`mass - $CellContext`r[])^(-1)/$CellContext`r[], 0}, {
      0, 0, 0, $CellContext`mass (
         2 $CellContext`mass - $CellContext`r[])^(-1)/$CellContext`r[]}}, {{0,
       0, 0, 0}, {
      0, 0, -$CellContext`mass (
         2 $CellContext`mass - $CellContext`r[])^(-1)/$CellContext`r[], 0}, {
      0, 0, 0, 0}, {0, 0, 0, 0}}, {{0, 0, 0, 0}, {
      0, 0, 0, -$CellContext`mass (
         2 $CellContext`mass - $CellContext`r[])^(-1)/$CellContext`r[]}, {0, 
      0, 0, 0}, {0, 0, 0, 
      0}}}, {-$CellContext`ch, -$CellContext`ch, $CellContext`ch}, 
    0][-$CellContext`a, -$CellContext`b, $CellContext`c],
   Editable->False]}]], "Output",
 CellChangeTimes->{
  3.7286757003468237`*^9, 3.728753472215955*^9, 3.728843482213537*^9, 
   3.730955506942773*^9, 3.730956195295621*^9, 3.7311157358274403`*^9, 
   3.731115889570994*^9, 3.7333676590894613`*^9, 3.733367891920555*^9, 
   3.733368123664611*^9, 3.7908682614970207`*^9, 3.7922738149155483`*^9, {
   3.794517913049202*^9, 3.7945179377101316`*^9}, 3.794518016209988*^9, 
   3.794518123399159*^9, 3.794518781354169*^9, 3.7945190614124537`*^9, 
   3.794519296757712*^9, 3.7945200057574663`*^9, 3.794520181627393*^9, 
   3.794520316110004*^9, 3.7945204177668867`*^9, 3.794520587247526*^9, 
   3.794520663221344*^9},
 CellLabel->"Out[57]=",
 CellID->1483791847,ExpressionUUID->"da65ce41-3670-4869-800e-c7f9fc4c5daf"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "Simplify"}]], "Input",
 CellLabel->"In[58]:=",
 CellID->1345708058,ExpressionUUID->"9ead0511-b09c-4a0c-a11b-924cd3b37aec"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728675702199052*^9, 3.7287534723606577`*^9, 
  3.728843484363851*^9, 3.730955508444231*^9, 3.730956195392414*^9, 
  3.7311157376425533`*^9, 3.731115891763739*^9, 3.733367659239753*^9, 
  3.7333678920731583`*^9, 3.733368123781913*^9, 3.790868263073756*^9, 
  3.7922738149612703`*^9, 3.79451793775528*^9, 3.7945180162529383`*^9, 
  3.7945181234678392`*^9, 3.7945187814000387`*^9, 3.794519061509489*^9, 
  3.7945192968044653`*^9, 3.794520005800316*^9, 3.794520181669861*^9, 
  3.794520316167967*^9, 3.79452041781396*^9, 3.794520587301579*^9, 
  3.794520663331367*^9},
 CellLabel->"Out[58]=",
 CellID->1509463565,ExpressionUUID->"7b9a9f9a-1939-413e-9011-063b64c1a7f9"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Being a Levi-Civita connection, it has zero torsion:", "MathCaption",
 CellChangeTimes->{{3.728675707955168*^9, 3.728675708116582*^9}},
 CellID->935778175,ExpressionUUID->"b60576e2-08b5-451c-88df-0ba95e3026ed"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Torsion", "[", "cd", "]"}], "[", 
  RowBox[{"a", ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}]}], "]"}]], "Input",
 CellLabel->"In[59]:=",
 CellID->1601040244,ExpressionUUID->"1477a6ac-0bfe-4f8f-84ce-b43904bcc1d1"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.7286757090860662`*^9, 3.728753472509335*^9, 
  3.728843485803751*^9, 3.730955509654377*^9, 3.730956195498877*^9, 
  3.7311157416078987`*^9, 3.7311158932313633`*^9, 3.733367659390051*^9, 
  3.733367892218713*^9, 3.733368123899638*^9, 3.790868268628714*^9, 
  3.792273815013587*^9, 3.794517937793839*^9, 3.794518016294058*^9, 
  3.794518123508265*^9, 3.794518781440752*^9, 3.794519061564622*^9, 
  3.794519296851201*^9, 3.794520005848229*^9, 3.7945201817155523`*^9, 
  3.794520316207363*^9, 3.794520417857044*^9, 3.794520587352583*^9, 
  3.794520663381649*^9},
 CellLabel->"Out[59]=",
 CellID->1329814944,ExpressionUUID->"da40d6e5-2ff1-4233-b352-2005f3c72d90"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The Schwarzschild metric is Ricci-flat:", "MathCaption",
 CellChangeTimes->{{3.7286757148598223`*^9, 3.7286757150122643`*^9}},
 CellID->910436832,ExpressionUUID->"ce4c71f3-33a1-4d6f-8304-3725f95ce08f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Ricci", "[", "cd", "]"}], "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}]}], "]"}]], "Input",
 CellLabel->"In[60]:=",
 CellID->890850549,ExpressionUUID->"41c8a63c-1573-471f-a4ab-88fa0876ebe2"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.7286757164231443`*^9, 3.728753472644946*^9, 
  3.728843486981102*^9, 3.7309555108455763`*^9, 3.730956195589477*^9, 
  3.7311157455325003`*^9, 3.73111589471483*^9, 3.733367659617169*^9, 
  3.733367892319808*^9, 3.733368123992247*^9, 3.790868269978217*^9, 
  3.792273815053762*^9, 3.794517937834826*^9, 3.794518016335333*^9, 
  3.794518123564542*^9, 3.794518781487836*^9, 3.794519061617629*^9, 
  3.794519296904171*^9, 3.794520005887376*^9, 3.794520181756959*^9, 
  3.794520316254019*^9, 3.794520417914494*^9, 3.7945205874043818`*^9, 
  3.794520663436561*^9},
 CellLabel->"Out[60]=",
 CellID->1337449479,ExpressionUUID->"3b2aeba0-03d9-4be3-b6e2-6262db6313d7"]
}, Open  ]]
}, Open  ]],

Cell["\<\
We will see later how to compute all curvature tensors in a single and faster \
computation, using the command MetricCompute.\
\>", "Text",
 CellChangeTimes->{{3.731115900154146*^9, 3.731115932985011*^9}},
 CellID->1710545020,ExpressionUUID->"78994f27-8620-4a23-ba9b-4e9e15f8567c"]
}, Closed]],

Cell[CellGroupData[{

Cell["3.1.5. Cleanup", "Subsubsection",
 CellChangeTimes->{{3.728674025410263*^9, 3.728674025658721*^9}, {
  3.733368006168591*^9, 3.733368006272663*^9}},
 CellID->1678441731,ExpressionUUID->"f1b15873-32a4-47e2-9a08-77dbec0815fa"],

Cell[CellGroupData[{

Cell["Remove the structures defined through Def commands:", "MathCaption",
 CellID->224985925,ExpressionUUID->"e3acd022-1202-471a-8b64-4ccd87e45194"],

Cell[BoxData[
 RowBox[{"UnsetCMetric", "[", "met", "]"}]], "Input",
 CellLabel->"In[61]:=",
 CellID->1767222755,ExpressionUUID->"51532a59-b35b-48b8-8454-210c435704e7"],

Cell[BoxData[
 RowBox[{"UndefConstantSymbol", "[", "mass", "]"}]], "Input",
 CellLabel->"In[62]:=",
 CellID->2014005275,ExpressionUUID->"2ba7cd89-015c-4110-abc7-b4c5f939cc99"],

Cell[BoxData[
 RowBox[{"UndefChart", "[", "ch", "]"}]], "Input",
 CellLabel->"In[63]:=",
 CellID->749412911,ExpressionUUID->"763bc2bb-6304-4f3c-813a-2f2ea6f65d7e"],

Cell[BoxData[
 RowBox[{"UndefBasis", "[", "red", "]"}]], "Input",
 CellLabel->"In[64]:=",
 CellID->391205522,ExpressionUUID->"51c7e33c-e4fb-410a-99be-84a7593906dc"],

Cell[BoxData[
 RowBox[{"UndefBasis", "[", "blue", "]"}]], "Input",
 CellLabel->"In[65]:=",
 CellID->1543764787,ExpressionUUID->"fabccde2-570c-4bcb-93d8-8226888993de"],

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M", "]"}]], "Input",
 CellLabel->"In[66]:=",
 CellID->157352055,ExpressionUUID->"9822dbd5-1f8d-4e43-a5ba-4464ee2309ec"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Remove the expressions we have defined interactively:", "MathCaption",
 CellID->1063618928,ExpressionUUID->"d0bde9da-416f-4626-bad5-185fadca1959"],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{"comm", ",", "cd", ",", "v", ",", "met", ",", "S", ",", "T"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.7311159523443947`*^9, 3.731115954448077*^9}},
 CellLabel->"In[67]:=",
 CellID->589445268,ExpressionUUID->"e006247d-5489-405e-bd51-8724703d20c1"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Clear the cache:", "MathCaption",
 CellID->464555665,ExpressionUUID->"43a411ee-b7b6-432a-af4a-8fe6522ecd06"],

Cell[BoxData[
 RowBox[{"ClearxCobaCache", "[", "All", "]"}]], "Input",
 CellLabel->"In[68]:=",
 CellID->1684004810,ExpressionUUID->"25d1baf4-0a34-4bf3-94ac-5ec29b94e60f"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.2. Tensors", "Subsection",
 CellChangeTimes->{{3.728674032033964*^9, 3.728674032242378*^9}, {
  3.728675275328267*^9, 3.728675275480215*^9}},
 CellID->1381857857,ExpressionUUID->"79b27aa2-b114-40c6-8e45-9fd7eb1892a0"],

Cell[CellGroupData[{

Cell["3.2.1. CTensor expressions", "Subsubsection",
 CellChangeTimes->{{3.733368013424474*^9, 3.733368017672427*^9}},
 CellID->1996843941,ExpressionUUID->"a15826e2-c0da-4abe-add5-46cc54ce5bb9"],

Cell["\<\
The head CTensor describes a tensor by giving an array of components in some \
bases. A third argument adds density weight.\
\>", "Text",
 CellChangeTimes->{{3.728675751170332*^9, 3.728675752954624*^9}},
 CellID->81681071,ExpressionUUID->"fd464a42-19cc-49da-a676-da71dbb88ccc"],

Cell[CellGroupData[{

Cell["Define a manifold with a basis field for its tangent bundle:", \
"MathCaption",
 CellChangeTimes->{3.728675793872196*^9},
 CellID->571709864,ExpressionUUID->"a65255cd-6832-4321-b1cc-672a7bf192ac"],

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M", ",", "4", ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f"}], "}"}]}], 
  "]"}]], "Input",
 CellLabel->"In[69]:=",
 CellID->885847570,ExpressionUUID->"99cd3fc8-8770-45aa-91c5-4293470059bc"],

Cell[BoxData[
 RowBox[{"DefBasis", "[", 
  RowBox[{"B", ",", "TangentM", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2", ",", "3"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[70]:=",
 CellID->2083467986,ExpressionUUID->"1725273d-ce82-4525-80dd-590d7c2f622e"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Define some scalar field on the manifold:", "MathCaption",
 CellChangeTimes->{3.728675797359973*^9},
 CellID->2125576367,ExpressionUUID->"e6c886bf-ea73-4ab1-84e8-a46b8c33d454"],

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"s", "[", "]"}], ",", "M"}], "]"}]], "Input",
 CellLabel->"In[71]:=",
 CellID->1171532324,ExpressionUUID->"336ef8dc-8322-4c5d-a458-6c86d55156a1"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
This is a vector field on M. It automatically adds the third argument, \
meaning that there is no added density weight:\
\>", "MathCaption",
 CellChangeTimes->{3.728675800191895*^9},
 CellID->1813693772,ExpressionUUID->"24964055-cb16-4769-8821-c9107df3f6b7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"v", "=", 
  RowBox[{"CTensor", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"0", ",", 
      RowBox[{"I", " ", 
       RowBox[{"s", "[", "]"}]}], ",", "1", ",", 
      RowBox[{"3", 
       RowBox[{
        RowBox[{"s", "[", "]"}], "^", "2"}]}]}], "}"}], ",", 
    RowBox[{"{", "B", "}"}]}], "]"}]}]], "Input",
 CellLabel->"In[72]:=",
 CellID->1368340486,ExpressionUUID->"34bd2b24-3386-44e1-b723-50ffe64d2598"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0", ",", 
     RowBox[{"\[ImaginaryI]", " ", 
      InterpretationBox[
       StyleBox["s",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`s[],
       Editable->False]}], ",", "1", ",", 
     RowBox[{"3", " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["s",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`s[],
        Editable->False], "2"]}]}], "}"}], ",", 
   RowBox[{"{", "B", "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.728675801386998*^9, 3.728753494282233*^9, 
  3.728843500879489*^9, 3.730955561345038*^9, 3.730956196640173*^9, 
  3.731115998431457*^9, 3.733367675139443*^9, 3.7333678935681057`*^9, 
  3.733368125192564*^9, 3.792273845997307*^9, 3.794517938658092*^9, 
  3.794518017142544*^9, 3.7945181699105873`*^9, 3.7945187822403193`*^9, 
  3.794519062373506*^9, 3.794519297656069*^9, 3.79452002291575*^9, 
  3.794520209038137*^9, 3.7945203172734737`*^9, 3.794520418749219*^9, 
  3.794520588395947*^9, 3.794520664242036*^9},
 CellLabel->"Out[72]=",
 CellID->802060404,ExpressionUUID->"90771a56-7b70-4e0b-8036-aab69d097fde"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "This object is now treated by ",
 StyleBox["xAct",
  FontSlant->"Italic"],
 " as a head v declared with DefTensor:"
}], "MathCaption",
 CellChangeTimes->{3.728675803000017*^9},
 CellID->2051244454,ExpressionUUID->"d4fc798e-61de-461c-b8f6-1538877694ef"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"xTensorQ", "[", "v", "]"}]], "Input",
 CellLabel->"In[73]:=",
 CellID->1710553087,ExpressionUUID->"d5cd3302-004c-4aff-b81a-578e3b3b2a32"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728675803975914*^9, 3.728753494399184*^9, 
  3.7288435012207603`*^9, 3.730955562779497*^9, 3.730956196745626*^9, 
  3.731116003971129*^9, 3.733367675256168*^9, 3.733367893669462*^9, 
  3.7333681253164167`*^9, 3.79227384605319*^9, 3.794517938708919*^9, 
  3.794518017194614*^9, 3.794518169960636*^9, 3.794518782288987*^9, 
  3.7945190624156837`*^9, 3.79451929770584*^9, 3.79452002296667*^9, 
  3.794520209079282*^9, 3.794520317319909*^9, 3.794520418799911*^9, 
  3.794520588452325*^9, 3.794520664347649*^9},
 CellLabel->"Out[73]=",
 CellID->474357809,ExpressionUUID->"05b94de3-26ce-484e-8e30-2fb3bde08a1d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SlotsOfTensor", "[", "v", "]"}]], "Input",
 CellLabel->"In[74]:=",
 CellID->1214283920,ExpressionUUID->"5954461d-4ebd-46ec-a236-a74c18548ac6"],

Cell[BoxData[
 RowBox[{"{", 
  InterpretationBox[
   StyleBox["\[DoubleStruckCapitalT]M",
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`TangentM,
   Editable->False], "}"}]], "Output",
 CellChangeTimes->{3.72867580453515*^9, 3.728753494515204*^9, 
  3.7288435017544527`*^9, 3.730955563357017*^9, 3.730956196848454*^9, 
  3.731116005699133*^9, 3.733367675406073*^9, 3.733367893769504*^9, 
  3.733368125407927*^9, 3.792273846103973*^9, 3.7945179387595787`*^9, 
  3.794518017245521*^9, 3.794518170016726*^9, 3.794518782327814*^9, 
  3.794519062464299*^9, 3.79451929774905*^9, 3.794520023011519*^9, 
  3.7945202091212807`*^9, 3.794520317372155*^9, 3.794520418844866*^9, 
  3.794520588515854*^9, 3.794520664400419*^9},
 CellLabel->"Out[74]=",
 CellID->1072398199,ExpressionUUID->"ed7c22ff-687d-418b-9d83-491012655440"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SymmetryGroupOfTensor", "[", "v", "]"}]], "Input",
 CellLabel->"In[75]:=",
 CellID->31382093,ExpressionUUID->"f20ae69d-1ea2-4cb9-95c4-1be13ff0558b"],

Cell[BoxData[
 RowBox[{"StrongGenSet", "[", 
  RowBox[{
   RowBox[{"{", "}"}], ",", 
   RowBox[{"GenSet", "[", "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728675805116156*^9, 3.7287534946318703`*^9, 
  3.728843502067689*^9, 3.730955563906987*^9, 3.7309561969376574`*^9, 
  3.731116006618064*^9, 3.7333676757598333`*^9, 3.7333678938672113`*^9, 
  3.7333681255335703`*^9, 3.792273846154921*^9, 3.794517938812109*^9, 
  3.7945180172884502`*^9, 3.7945181700633087`*^9, 3.794518782368041*^9, 
  3.7945190625608873`*^9, 3.794519297792974*^9, 3.7945200230560913`*^9, 
  3.794520209159984*^9, 3.794520317417214*^9, 3.794520418888289*^9, 
  3.7945205885722437`*^9, 3.794520664448674*^9},
 CellLabel->"Out[75]=",
 CellID->2113764969,ExpressionUUID->"6ccdecbb-5103-455a-ac08-34056ef53758"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DependenciesOfTensor", "[", "v", "]"}]], "Input",
 CellLabel->"In[76]:=",
 CellID->215871800,ExpressionUUID->"525f982e-cbdf-4edb-89b4-a3495070f9be"],

Cell[BoxData[
 RowBox[{"{", "M", "}"}]], "Output",
 CellChangeTimes->{3.728675805699164*^9, 3.728753494748464*^9, 
  3.728843502275702*^9, 3.730955564508172*^9, 3.730956197047958*^9, 
  3.7311160075825453`*^9, 3.733367675938718*^9, 3.733367893987175*^9, 
  3.733368125652482*^9, 3.792273846210004*^9, 3.7945179388748913`*^9, 
  3.79451801733876*^9, 3.794518170111496*^9, 3.794518782406125*^9, 
  3.794519062610342*^9, 3.794519297841854*^9, 3.794520023103237*^9, 
  3.794520209204336*^9, 3.794520317461412*^9, 3.7945204189834003`*^9, 
  3.794520588630686*^9, 3.794520664503824*^9},
 CellLabel->"Out[76]=",
 CellID->225861662,ExpressionUUID->"475b654f-0a6e-484b-87e1-a70c2c6caf62"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"WeightOfTensor", "[", "v", "]"}]], "Input",
 CellLabel->"In[77]:=",
 CellID->202902747,ExpressionUUID->"29a7ef17-0097-482e-aa0f-e9fc58e5f1fb"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.7286758064419527`*^9, 3.728753494865938*^9, 
  3.7288435026992607`*^9, 3.730955565057556*^9, 3.730956197144227*^9, 
  3.731116008421331*^9, 3.733367676091051*^9, 3.733367894085622*^9, 
  3.733368125765983*^9, 3.7922738462651997`*^9, 3.794517938926527*^9, 
  3.7945180173864107`*^9, 3.7945181701618032`*^9, 3.7945187824522953`*^9, 
  3.794519062647951*^9, 3.794519297881864*^9, 3.794520023146532*^9, 
  3.794520209245419*^9, 3.794520317509694*^9, 3.7945204190353403`*^9, 
  3.7945205886823463`*^9, 3.7945206645497303`*^9},
 CellLabel->"Out[77]=",
 CellID->1453348985,ExpressionUUID->"5cc0fde0-2f9b-4ca6-8d48-0e3bdd25de38"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Dagger", "[", "v", "]"}]], "Input",
 CellLabel->"In[78]:=",
 CellID->1178007437,ExpressionUUID->"a4560393-8272-4bdb-b2c5-ae7c42cdf4ba"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0", ",", 
     RowBox[{
      RowBox[{"-", "\[ImaginaryI]"}], " ", 
      InterpretationBox[
       StyleBox["s",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`s[],
       Editable->False]}], ",", "1", ",", 
     RowBox[{"3", " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["s",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`s[],
        Editable->False], "2"]}]}], "}"}], ",", 
   RowBox[{"{", "B", "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.728675807092182*^9, 3.728753494983902*^9, 
  3.728843503001474*^9, 3.730955565663677*^9, 3.730956197249689*^9, 
  3.731116009151289*^9, 3.733367676320389*^9, 3.733367894184924*^9, 
  3.733368125883531*^9, 3.792273846315753*^9, 3.794517938974965*^9, 
  3.794518017432497*^9, 3.7945181702122173`*^9, 3.794518782492087*^9, 
  3.7945190626885347`*^9, 3.7945192979851007`*^9, 3.794520023195302*^9, 
  3.794520209295616*^9, 3.7945203175601273`*^9, 3.7945204190820923`*^9, 
  3.794520588738132*^9, 3.794520664607123*^9},
 CellLabel->"Out[78]=",
 CellID->957389917,ExpressionUUID->"247b6aab-e149-4e4e-9bd2-88b562838797"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We can have a higher-rank tensor. Let us also make it a density of weight +2 \
in the basis B:\
\>", "MathCaption",
 CellChangeTimes->{3.728675809367618*^9},
 CellID->1458528489,ExpressionUUID->"23ad7362-8029-47ee-8280-80f4fc4804c1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"A", "=", 
  RowBox[{"CTensor", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"s", "[", "]"}], 
     RowBox[{"Array", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"Signature", "[", 
         RowBox[{"{", "##", "}"}], "]"}], "&"}], ",", 
       RowBox[{"{", 
        RowBox[{"4", ",", "4", ",", "4"}], "}"}]}], "]"}]}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "B"}], ",", 
      RowBox[{"-", "B"}], ",", 
      RowBox[{"-", "B"}]}], "}"}], ",", 
    RowBox[{"2", "B"}]}], "]"}]}]], "Input",
 CellLabel->"In[79]:=",
 CellID->2109967765,ExpressionUUID->"1fdc1000-cdac-461b-b6b1-85df819977d7"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", 
         InterpretationBox[
          StyleBox["s",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`s[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["s",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`s[],
          Editable->False]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", 
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["s",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`s[],
           Editable->False]}], ",", "0", ",", 
         InterpretationBox[
          StyleBox["s",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`s[],
          Editable->False]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", 
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["s",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`s[],
           Editable->False]}], ",", 
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["s",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`s[],
           Editable->False]}], ",", "0"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", 
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["s",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`s[],
           Editable->False]}], ",", 
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["s",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`s[],
           Editable->False]}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         InterpretationBox[
          StyleBox["s",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`s[],
          Editable->False], ",", "0", ",", "0", ",", 
         InterpretationBox[
          StyleBox["s",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`s[],
          Editable->False]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         InterpretationBox[
          StyleBox["s",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`s[],
          Editable->False], ",", "0", ",", 
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["s",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`s[],
           Editable->False]}], ",", "0"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", 
         InterpretationBox[
          StyleBox["s",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`s[],
          Editable->False], ",", "0", ",", 
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["s",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`s[],
           Editable->False]}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["s",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`s[],
           Editable->False]}], ",", "0", ",", "0", ",", 
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["s",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`s[],
           Editable->False]}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         InterpretationBox[
          StyleBox["s",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`s[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["s",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`s[],
          Editable->False], ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", 
         InterpretationBox[
          StyleBox["s",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`s[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["s",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`s[],
          Editable->False], ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["s",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`s[],
           Editable->False]}], ",", "0", ",", 
         InterpretationBox[
          StyleBox["s",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`s[],
          Editable->False], ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["s",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`s[],
           Editable->False]}], ",", 
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["s",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`s[],
           Editable->False]}], ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}]}], "}"}], 
   ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "B"}], ",", 
     RowBox[{"-", "B"}], ",", 
     RowBox[{"-", "B"}]}], "}"}], ",", 
   RowBox[{"2", " ", "B"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728675810649885*^9, 3.72875349510137*^9, 
  3.728843503213108*^9, 3.73095556676171*^9, 3.730956197345634*^9, 
  3.731116014637292*^9, 3.733367676553178*^9, 3.7333678943066196`*^9, 
  3.733368125998989*^9, 3.7922738463737698`*^9, 3.7945179390277977`*^9, 
  3.794518017479138*^9, 3.7945181702633657`*^9, 3.7945187825387993`*^9, 
  3.7945190627278147`*^9, 3.794519298030531*^9, 3.794520023239386*^9, 
  3.79452020939825*^9, 3.79452031761477*^9, 3.794520419129135*^9, 
  3.794520588795307*^9, 3.7945206646675243`*^9},
 CellLabel->"Out[79]=",
 CellID->1904851695,ExpressionUUID->"72e71561-1ce8-45e3-9e7b-3d61d0d6fe6f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"xTensorQ", "[", "A", "]"}]], "Input",
 CellLabel->"In[80]:=",
 CellID->1850014161,ExpressionUUID->"53738b58-a7eb-49d6-9748-0be16ca5b84d"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.72867581191368*^9, 3.7287534952177057`*^9, 
  3.728843503654882*^9, 3.7309555680192003`*^9, 3.730956197465554*^9, 
  3.731116015894034*^9, 3.7333676766957607`*^9, 3.733367894402117*^9, 
  3.733368126116955*^9, 3.792273846425371*^9, 3.79451793907892*^9, 
  3.7945180175307703`*^9, 3.7945181703116827`*^9, 3.794518782577776*^9, 
  3.7945190627724743`*^9, 3.794519298075383*^9, 3.7945200232853327`*^9, 
  3.794520209444064*^9, 3.7945203176687307`*^9, 3.794520419172906*^9, 
  3.79452058885111*^9, 3.794520664722048*^9},
 CellLabel->"Out[80]=",
 CellID->156228389,ExpressionUUID->"1228c308-f086-4444-b64a-bb8a5a9c2630"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SlotsOfTensor", "[", "A", "]"}]], "Input",
 CellLabel->"In[81]:=",
 CellID->1656730899,ExpressionUUID->"8e2f4e65-5e8f-45db-b7d4-81c10005d1ef"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox["\[DoubleStruckCapitalT]M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TangentM,
     Editable->False]}], ",", 
   RowBox[{"-", 
    InterpretationBox[
     StyleBox["\[DoubleStruckCapitalT]M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TangentM,
     Editable->False]}], ",", 
   RowBox[{"-", 
    InterpretationBox[
     StyleBox["\[DoubleStruckCapitalT]M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TangentM,
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.7286758125308332`*^9, 3.728753495333055*^9, 
  3.728843504386352*^9, 3.7309555686053247`*^9, 3.730956197562501*^9, 
  3.731116016729355*^9, 3.733367676838799*^9, 3.733367894502363*^9, 
  3.733368126215535*^9, 3.792273846478237*^9, 3.794517939125626*^9, 
  3.794518017573805*^9, 3.794518170360754*^9, 3.7945187826727552`*^9, 
  3.79451906281043*^9, 3.794519298119999*^9, 3.794520023327795*^9, 
  3.794520209499353*^9, 3.794520317720552*^9, 3.794520419217578*^9, 
  3.794520588903696*^9, 3.794520664765868*^9},
 CellLabel->"Out[81]=",
 CellID->1505517289,ExpressionUUID->"05c95fc2-1807-4bba-88b7-6ec343a2f1b4"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SymmetryGroupOfTensor", "[", "A", "]"}]], "Input",
 CellLabel->"In[82]:=",
 CellID->6226305,ExpressionUUID->"f4a66114-7b07-408e-a1bc-ebb071de9fd5"],

Cell[BoxData[
 RowBox[{"StrongGenSet", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "2", ",", "3"}], "}"}], ",", 
   RowBox[{"GenSet", "[", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"-", 
        RowBox[{"(", "1", ",", "2", ")"}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      -xAct`xPerm`Cycles[{1, 2}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[
       RowBox[{"-", 
        RowBox[{"(", "2", ",", "3", ")"}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      -xAct`xPerm`Cycles[{2, 3}],
      Editable->False]}], "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.7286758131931868`*^9, 3.728753495449922*^9, 
  3.728843504984178*^9, 3.7309555692205677`*^9, 3.730956197664969*^9, 
  3.731116017538414*^9, 3.733367676962845*^9, 3.733367894619318*^9, 
  3.733368126309451*^9, 3.792273846528997*^9, 3.794517939175157*^9, 
  3.794518017625333*^9, 3.794518170410902*^9, 3.794518782724119*^9, 
  3.794519062847125*^9, 3.7945192981663218`*^9, 3.794520023368754*^9, 
  3.7945202096738653`*^9, 3.7945203177686*^9, 3.794520419269055*^9, 
  3.794520588960307*^9, 3.794520664820745*^9},
 CellLabel->"Out[82]=",
 CellID->1576747269,ExpressionUUID->"7bf1f88c-86d3-4271-af4e-8fdcb0801fd8"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DependenciesOfTensor", "[", "A", "]"}]], "Input",
 CellLabel->"In[83]:=",
 CellID->833233827,ExpressionUUID->"63f5929f-6d18-424b-9980-9898fbf3b005"],

Cell[BoxData[
 RowBox[{"{", "M", "}"}]], "Output",
 CellChangeTimes->{3.7286758137661552`*^9, 3.7287534955672083`*^9, 
  3.728843505200938*^9, 3.7309555698873787`*^9, 3.730956197762004*^9, 
  3.731116018488386*^9, 3.7333676772930593`*^9, 3.733367894719739*^9, 
  3.733368126433669*^9, 3.792273846591063*^9, 3.7945179392541533`*^9, 
  3.794518017668707*^9, 3.794518170458654*^9, 3.794518782762517*^9, 
  3.794519062892503*^9, 3.794519298210204*^9, 3.7945200234126253`*^9, 
  3.794520209823771*^9, 3.7945203178124228`*^9, 3.794520419315362*^9, 
  3.794520589013274*^9, 3.794520664867193*^9},
 CellLabel->"Out[83]=",
 CellID->443146103,ExpressionUUID->"02a27b44-ee36-4f13-8d44-ecd8d1f092f8"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"WeightOfTensor", "[", "A", "]"}]], "Input",
 CellLabel->"In[84]:=",
 CellID->1309204025,ExpressionUUID->"0f8342cb-7fd0-4999-841e-b56d6955d661"],

Cell[BoxData[
 RowBox[{"2", " ", "B"}]], "Output",
 CellChangeTimes->{3.728675814367072*^9, 3.7287534956658077`*^9, 
  3.728843505621286*^9, 3.730955570538196*^9, 3.73095619786373*^9, 
  3.731116019163198*^9, 3.733367677470417*^9, 3.7333678948187943`*^9, 
  3.733368126587356*^9, 3.792273846634617*^9, 3.7945179393022747`*^9, 
  3.79451801771097*^9, 3.794518170511324*^9, 3.794518782804461*^9, 
  3.794519062928743*^9, 3.794519298258971*^9, 3.794520023468629*^9, 
  3.7945202098927317`*^9, 3.7945203178607187`*^9, 3.794520419357259*^9, 
  3.7945205890627613`*^9, 3.794520664921341*^9},
 CellLabel->"Out[84]=",
 CellID->483265325,ExpressionUUID->"efc270ca-d5a1-4c2c-9715-fc484ba71395"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Dagger", "[", "A", "]"}], "===", "A"}]], "Input",
 CellLabel->"In[85]:=",
 CellID->1453052978,ExpressionUUID->"5ad868df-6da0-4d90-a39a-67c96ac0411a"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728675814999905*^9, 3.728753495784649*^9, 
  3.7288435058498163`*^9, 3.7309555711966677`*^9, 3.7309561979657297`*^9, 
  3.731116022730914*^9, 3.733367677573138*^9, 3.733367894918247*^9, 
  3.73336812676588*^9, 3.792273846686431*^9, 3.7945179393518*^9, 
  3.794518017804405*^9, 3.794518170554831*^9, 3.794518782843354*^9, 
  3.794519062964851*^9, 3.7945192983003063`*^9, 3.794520023515848*^9, 
  3.7945202099705877`*^9, 3.794520317909913*^9, 3.794520419403448*^9, 
  3.794520589112885*^9, 3.7945206649660053`*^9},
 CellLabel->"Out[85]=",
 CellID->457164321,ExpressionUUID->"b3fb11be-cfc9-40a6-83cf-7ca7cdc2eba2"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
A CTensor with an array of zeros is replaced by the head Zero, which \
autoconverts to 0 when given indices:\
\>", "MathCaption",
 CellChangeTimes->{3.728675816647326*^9},
 CellID->1677100544,ExpressionUUID->"68a457c1-f64c-4f7c-bebb-35ace95cb117"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"0", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0"}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"B", ",", "B"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[86]:=",
 CellID->1305976133,ExpressionUUID->"1819b3b1-10ec-4607-a4a3-46142ab968c8"],

Cell[BoxData["Zero"], "Output",
 CellChangeTimes->{3.728675817865872*^9, 3.7287534958795567`*^9, 
  3.728843506386848*^9, 3.730955572363842*^9, 3.730956198062298*^9, 
  3.731116029656992*^9, 3.733367677710581*^9, 3.7333678950186253`*^9, 
  3.733368126981718*^9, 3.792273846728899*^9, 3.794517939395949*^9, 
  3.7945180178540382`*^9, 3.794518170595701*^9, 3.794518782885047*^9, 
  3.7945190630039873`*^9, 3.79451929834167*^9, 3.794520023570695*^9, 
  3.7945202100211363`*^9, 3.794520317954179*^9, 3.79452041945576*^9, 
  3.794520589167285*^9, 3.794520665020817*^9},
 CellLabel->"Out[86]=",
 CellID->2011737728,ExpressionUUID->"ee3dc255-89b8-49e1-8623-9307a0b8fc9a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Zero", "[", 
  RowBox[{"a", ",", "b"}], "]"}]], "Input",
 CellLabel->"In[87]:=",
 CellID->1938287600,ExpressionUUID->"c8854cfa-ab04-4832-8aff-a823ff9a92b6"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728675818476911*^9, 3.728753495982561*^9, 
  3.728843506867116*^9, 3.730955573054245*^9, 3.730956198166168*^9, 
  3.731116030800585*^9, 3.7333676778062773`*^9, 3.733367895136345*^9, 
  3.73336812711681*^9, 3.792273846783165*^9, 3.7945179394464684`*^9, 
  3.794518017903462*^9, 3.794518170639598*^9, 3.794518782926754*^9, 
  3.794519063043323*^9, 3.79451929838552*^9, 3.79452002361658*^9, 
  3.794520210072899*^9, 3.7945203180107517`*^9, 3.794520419499755*^9, 
  3.794520589219674*^9, 3.794520665069805*^9},
 CellLabel->"Out[87]=",
 CellID->1293456128,ExpressionUUID->"749b2a8f-9435-4655-807e-e5804cad801c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
A CTensor containing a scalar is automatically converted into the scalar:\
\>", "MathCaption",
 CellChangeTimes->{3.728675820822998*^9},
 CellID->1480316639,ExpressionUUID->"d901b97e-1afc-47ff-a483-e3495402f338"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CTensor", "[", 
   RowBox[{
    RowBox[{"s", "[", "]"}], ",", 
    RowBox[{"{", "}"}]}], "]"}], "[", "]"}]], "Input",
 CellLabel->"In[88]:=",
 CellID->1083701178,ExpressionUUID->"4c44df56-a018-49ac-9ad7-15c3e88cec40"],

Cell[BoxData[
 InterpretationBox[
  StyleBox["s",
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`s[],
  Editable->False]], "Output",
 CellChangeTimes->{3.728675822298213*^9, 3.728753496084051*^9, 
  3.728843507051812*^9, 3.7309555742627*^9, 3.730956198263032*^9, 
  3.7311160368834352`*^9, 3.7333676779300823`*^9, 3.733367895234746*^9, 
  3.733368127265984*^9, 3.792273846828051*^9, 3.7945179394904003`*^9, 
  3.7945180179519176`*^9, 3.794518170694635*^9, 3.794518782975079*^9, 
  3.794519063079698*^9, 3.794519298428089*^9, 3.794520023671934*^9, 
  3.79452021012444*^9, 3.794520318054908*^9, 3.794520419551499*^9, 
  3.794520589315357*^9, 3.794520665120596*^9},
 CellLabel->"Out[88]=",
 CellID->140184715,ExpressionUUID->"fb07ce59-2b94-4377-82e1-d34b85d396b7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellLabel->"In[89]:=",
 CellID->2066445537,ExpressionUUID->"806e2b9e-1abf-445e-849d-29546d7ce537"],

Cell["s[]", "Output",
 CellChangeTimes->{3.728675822878367*^9, 3.728753496184163*^9, 
  3.728843507225868*^9, 3.730955574913417*^9, 3.73095619836725*^9, 
  3.731116038086255*^9, 3.7333676780735903`*^9, 3.7333678953553877`*^9, 
  3.7333681274012833`*^9, 3.792273846870113*^9, 3.794517939536366*^9, 
  3.794518018000433*^9, 3.794518170737296*^9, 3.794518783021246*^9, 
  3.794519063122652*^9, 3.794519298472342*^9, 3.79452002371898*^9, 
  3.794520210170866*^9, 3.794520318108165*^9, 3.794520419600233*^9, 
  3.794520589371292*^9, 3.794520665171534*^9},
 CellLabel->"Out[89]//InputForm=",
 CellID->294944246,ExpressionUUID->"3370931e-a100-4abb-8d64-c398b7fe21cd"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["However, if it is a scalar density then we cannot do that:", \
"MathCaption",
 CellChangeTimes->{3.728675788672503*^9},
 CellID->548246988,ExpressionUUID->"cbe58b02-668a-41c4-ace6-f82af36d7ebf"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CTensor", "[", 
   RowBox[{
    RowBox[{"s", "[", "]"}], ",", 
    RowBox[{"{", "}"}], ",", "B"}], "]"}], "[", "]"}]], "Input",
 CellLabel->"In[90]:=",
 CellID->1682281686,ExpressionUUID->"7a09d8ed-5b45-4978-a6e9-13247dca452b"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   OverscriptBox[
    InterpretationBox[
     StyleBox["s",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`s[],
     Editable->False], 
    StyleBox["~",
     FontColor->RGBColor[1., 0., 0.]]],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[
   $CellContext`s[], {}, $CellContext`B][],
  Editable->False]], "Output",
 CellChangeTimes->{3.728675826882457*^9, 3.728753496284339*^9, 
  3.7288435074059973`*^9, 3.730955576073072*^9, 3.73095619846341*^9, 
  3.7311160403044643`*^9, 3.733367678191471*^9, 3.73336789545238*^9, 
  3.733368127533347*^9, 3.792273846945606*^9, 3.7945179395829897`*^9, 
  3.794518018044527*^9, 3.794518170776188*^9, 3.794518783073887*^9, 
  3.794519063160008*^9, 3.79451929852105*^9, 3.794520023770789*^9, 
  3.79452021021218*^9, 3.794520318152999*^9, 3.7945204196465893`*^9, 
  3.794520589421523*^9, 3.7945206652181673`*^9},
 CellLabel->"Out[90]=",
 CellID->735685246,ExpressionUUID->"11e8b11d-6bf6-4348-9f2d-9bc3e5a15e8c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellLabel->"In[91]:=",
 CellID->21896850,ExpressionUUID->"710b7220-32ec-4a34-b47f-2d75e920cb16"],

Cell["CTensor[s[], {}, B][]", "Output",
 CellChangeTimes->{3.728675827467602*^9, 3.728753496400544*^9, 
  3.728843507617235*^9, 3.730955576685845*^9, 3.730956198564337*^9, 
  3.731116042152417*^9, 3.733367678290937*^9, 3.733367895552188*^9, 
  3.7333681276669683`*^9, 3.792273847001934*^9, 3.794517939629512*^9, 
  3.794518018091673*^9, 3.794518170820318*^9, 3.7945187831198473`*^9, 
  3.79451906320522*^9, 3.7945192985646133`*^9, 3.794520023813222*^9, 
  3.794520210253811*^9, 3.794520318219912*^9, 3.794520419688793*^9, 
  3.794520589478383*^9, 3.79452066526865*^9},
 CellLabel->"Out[91]//InputForm=",
 CellID->1598573743,ExpressionUUID->"c2f40a5c-9dc4-4917-8fd5-08de489a3ff6"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["A self-consistency check:", "MathCaption",
 CellChangeTimes->{{3.728675785634433*^9, 3.728675785784861*^9}},
 CellID->464506489,ExpressionUUID->"688a89de-e370-4d47-b921-5757e24afe5b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CTensor", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"1", ",", "2"}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"3", ",", "4"}], "}"}]}], "}"}], ",", 
    RowBox[{"{", "B", "}"}]}], "]"}], "//", "Catch"}]], "Input",
 CellLabel->"In[92]:=",
 CellID->857503778,ExpressionUUID->"6c1f8af5-5a0d-4ebe-bfa6-751aad9ece2d"],

Cell[BoxData[
 TemplateBox[{
  "CTensor", "nobd", 
   "\"Incompatible number of bases \\!\\(\\*RowBox[{\\\"1\\\"}]\\) and array \
depth \\!\\(\\*RowBox[{\\\"2\\\"}]\\).\"", 2, 92, 3, 19683117308123869115, 
   "Local", "xAct`xCoba`CTensor"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728675829861899*^9, 3.7287534965554132`*^9, 
  3.7288435077900887`*^9, 3.730955578185288*^9, 3.7309561986797943`*^9, 
  3.731116055137779*^9, 3.733367678426105*^9, 3.733367895652803*^9, 
  3.73336812781695*^9, 3.792273847044771*^9, 3.794517939671582*^9, 
  3.79451801815693*^9, 3.794518170869191*^9, 3.794518783159541*^9, 
  3.7945190632430983`*^9, 3.794519298606731*^9, 3.7945200238882217`*^9, 
  3.794520210328401*^9, 3.79452031827024*^9, 3.794520419740497*^9, 
  3.794520589526615*^9, 3.794520665320722*^9},
 CellLabel->"During evaluation of In[92]:=",
 CellID->60990984,ExpressionUUID->"2fa175a2-15c2-4e8c-8144-3b6d27606722"]
}, Open  ]]
}, Open  ]],

Cell["\<\
Finally, note that there is no collection of the CTensor objects existing in \
a given session. In particular they do not appear in the list $Tensors.\
\>", "MathCaption",
 CellID->606725672,ExpressionUUID->"ed6e0153-ad5a-4741-9a99-b6c0ef41ff77"]
}, Closed]],

Cell[CellGroupData[{

Cell["3.2.2. Indexed CTensor expressions", "Subsubsection",
 CellChangeTimes->{{3.733368019248228*^9, 3.733368020440226*^9}},
 CellID->986013348,ExpressionUUID->"6373d80e-5081-4724-8f96-76962c4fbb01"],

Cell[TextData[{
 "Following the standard behaviour in ",
 StyleBox["xTensor",
  FontSlant->"Italic"],
 ", the CTensor expressions have a special typesetting when indices are \
added. This is just MatrixForm, which uses:\n\tT[a]\t\t(Column)\n\tT[a, b]\t\t\
(Column-Row)\n\tT[a, b, c]\t(Column-Row)-(Column)\n\t...\t\t..."
}], "Text",
 CellID->1942519584,ExpressionUUID->"28c38144-bcf1-4c8e-89aa-87ca284d4ce3"],

Cell[CellGroupData[{

Cell[TextData[{
 "Vectors (both covariant and contravariant) are represented as columns. In \
",
 StyleBox["xAct",
  FontSlant->"Italic"],
 " we use the index height to distinguish upper/lower, and never to \
distinguish rows from columns in a matrix."
}], "MathCaption",
 CellChangeTimes->{{3.728675849853898*^9, 3.728675850014153*^9}},
 CellID->1880246594,ExpressionUUID->"1d7fef68-d4f5-414c-bc49-6cfb3423b325"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"v", "[", "a", "]"}]], "Input",
 CellLabel->"In[93]:=",
 CellID->1082639753,ExpressionUUID->"e9e05e27-2a28-4280-936a-c7c92841ef3d"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{"\[ImaginaryI]", " ", 
             InterpretationBox[
              StyleBox["s",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`s[],
              Editable->False]}],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              RowBox[{"\[ImaginaryI]", " ", 
                InterpretationBox[
                 StyleBox["s", AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`s[], Editable -> False]}]}]],
           Annotation[#, {2} -> Complex[0, 1] $CellContext`s[], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", "1"}]],
           Annotation[#, {3} -> 1, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{"3", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["s",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`s[],
               Editable->False], "2"]}],
            RowBox[{
              RowBox[{"{", "4", "}"}], "\[Rule]", 
              RowBox[{"3", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                  "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`s[], Editable -> False], "2"]}]}]],
           Annotation[#, {4} -> 3 $CellContext`s[]^2, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{
   0, Complex[0, 1] $CellContext`s[], 1, 
    3 $CellContext`s[]^2}, {$CellContext`B}, 0][$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.728675913138433*^9, 3.7287534966682243`*^9, 
  3.728843513800911*^9, 3.730955591058522*^9, 3.730956198795639*^9, 
  3.731116067889845*^9, 3.733367895751829*^9, 3.7333681279506683`*^9, 
  3.792273852484748*^9, 3.7945179397168694`*^9, 3.7945180181949167`*^9, 
  3.794518188565864*^9, 3.7945187832036457`*^9, 3.794519063288198*^9, 
  3.794519298654831*^9, 3.79452002394384*^9, 3.7945202103879843`*^9, 
  3.794520318320261*^9, 3.794520419785736*^9, 3.794520589583146*^9, 
  3.7945206653767967`*^9},
 CellLabel->"Out[93]=",
 CellID->725768469,ExpressionUUID->"de17c893-3d51-4e9a-a990-abd059d93ad1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The color of the lines denote the basis associated to the components:\
\>", "MathCaption",
 CellID->999618911,ExpressionUUID->"cd29e58d-a01c-4384-be50-af1063129f9c"],

Cell[CellGroupData[{

Cell[BoxData["v"], "Input",
 CellLabel->"In[94]:=",
 CellID->830466696,ExpressionUUID->"ff5ff4fa-f150-4bc5-825b-dee35d74aabe"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0", ",", 
     RowBox[{"\[ImaginaryI]", " ", 
      InterpretationBox[
       StyleBox["s",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`s[],
       Editable->False]}], ",", "1", ",", 
     RowBox[{"3", " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["s",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`s[],
        Editable->False], "2"]}]}], "}"}], ",", 
   RowBox[{"{", "B", "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.7286759141856327`*^9, 3.7287534967835207`*^9, 
  3.728843513977191*^9, 3.730955592698852*^9, 3.730956198898223*^9, 
  3.7311160718540897`*^9, 3.733367895869598*^9, 3.7333681280829897`*^9, 
  3.792273852541779*^9, 3.794517939760235*^9, 3.7945180182338963`*^9, 
  3.7945181886171827`*^9, 3.794518783251163*^9, 3.7945190633257723`*^9, 
  3.794519298696288*^9, 3.7945200239890614`*^9, 3.794520210432962*^9, 
  3.7945203183674793`*^9, 3.794520419838374*^9, 3.794520589635605*^9, 
  3.794520665423484*^9},
 CellLabel->"Out[94]=",
 CellID->283543523,ExpressionUUID->"9b802ed8-e83c-4871-b4a6-750d790d668e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"BasisColor", "[", "B", "]"}]], "Input",
 CellLabel->"In[95]:=",
 CellID->683765760,ExpressionUUID->"9c48f53b-0344-4426-959c-610ff12b3050"],

Cell[BoxData[
 InterpretationBox[
  ButtonBox[
   TooltipBox[
    GraphicsBox[{
      {GrayLevel[0], RectangleBox[{0, 0}]}, 
      {GrayLevel[0], RectangleBox[{1, -1}]}, 
      {RGBColor[1, 0, 0], RectangleBox[{0, -1}, {2, 1}]}},
     AspectRatio->1,
     DefaultBaseStyle->"ColorSwatchGraphics",
     Frame->True,
     FrameStyle->RGBColor[0.6666666666666666, 0., 0.],
     FrameTicks->None,
     ImageSize->
      Dynamic[{Automatic, 1.35 CurrentValue["FontCapHeight"]/
        AbsoluteCurrentValue[Magnification]}],
     PlotRangePadding->None],
    StyleBox[
     RowBox[{"RGBColor", "[", 
       RowBox[{"1", ",", "0", ",", "0"}], "]"}], NumberMarks -> False]],
   Appearance->None,
   BaseStyle->{},
   BaselinePosition->Baseline,
   ButtonFunction:>With[{Typeset`box$ = EvaluationBox[]}, 
     If[
      Not[
       AbsoluteCurrentValue["Deployed"]], 
      SelectionMove[Typeset`box$, All, Expression]; 
      FrontEnd`Private`$ColorSelectorInitialAlpha = 1; 
      FrontEnd`Private`$ColorSelectorInitialColor = RGBColor[1, 0, 0]; 
      FrontEnd`Private`$ColorSelectorUseMakeBoxes = True; 
      MathLink`CallFrontEnd[
        FrontEnd`AttachCell[Typeset`box$, 
         FrontEndResource["RGBColorValueSelector"], {0, {Left, Bottom}}, {
         Left, Top}, 
         "ClosingActions" -> {
          "SelectionDeparture", "ParentChanged", "EvaluatorQuit"}]]]],
   DefaultBaseStyle->{},
   Evaluator->Automatic,
   Method->"Preemptive"],
  RGBColor[1, 0, 0],
  Editable->False,
  Selectable->False]], "Output",
 CellChangeTimes->{3.728675914812406*^9, 3.72875349690101*^9, 
  3.72884351420093*^9, 3.730955593400508*^9, 3.730956198995352*^9, 
  3.7311160727745323`*^9, 3.733367895962216*^9, 3.73336812821667*^9, 
  3.792273852587514*^9, 3.794517939814464*^9, 3.794518018270101*^9, 
  3.7945181886644783`*^9, 3.794518783294319*^9, 3.794519063362762*^9, 
  3.794519298741089*^9, 3.794520024041087*^9, 3.79452021047955*^9, 
  3.794520318415908*^9, 3.794520419882638*^9, 3.794520589690024*^9, 
  3.794520665470881*^9},
 CellLabel->"Out[95]=",
 CellID->249778368,ExpressionUUID->"6ae2fb0e-8dc8-4e5b-a7df-d4c98c00ac57"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Higher-order tensors. Note the double tilde, representing weight +2 in the \
basis B:\
\>", "MathCaption",
 CellChangeTimes->{3.728675916970722*^9},
 CellID->1456919595,ExpressionUUID->"211b2dfd-9318-4607-8d72-b47dc1d252b9"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"A", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}]}], "]"}]], "Input",
 CellLabel->"In[96]:=",
 CellID->839090071,ExpressionUUID->"cc90eeec-4f4b-4464-82d3-e1fe83abd639"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      OverscriptBox[
       OverscriptBox[
        StyleBox[GridBox[{
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {1, 1, 1} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {1, 1, 2} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "3"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {1, 1, 3} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "4"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {1, 1, 4} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {1, 2, 1} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {1, 2, 2} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 InterpretationBox[
                  StyleBox["s",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`s[],
                  Editable->False],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "3"}], "}"}], "\[Rule]", 
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}]],
                Annotation[#, {1, 2, 3} -> $CellContext`s[], "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 InterpretationBox[
                  StyleBox["s",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`s[],
                  Editable->False],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "4"}], "}"}], "\[Rule]", 
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}]],
                Annotation[#, {1, 2, 4} -> $CellContext`s[], "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "3", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {1, 3, 1} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 RowBox[{"-", 
                  InterpretationBox[
                   StyleBox["s",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`s[],
                   Editable->False]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "3", ",", "2"}], "}"}], "\[Rule]", 
                   RowBox[{"-", 
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}]}]],
                Annotation[#, {1, 3, 2} -> -$CellContext`s[], "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "3", ",", "3"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {1, 3, 3} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 InterpretationBox[
                  StyleBox["s",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`s[],
                  Editable->False],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "3", ",", "4"}], "}"}], "\[Rule]", 
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}]],
                Annotation[#, {1, 3, 4} -> $CellContext`s[], "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "4", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {1, 4, 1} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 RowBox[{"-", 
                  InterpretationBox[
                   StyleBox["s",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`s[],
                   Editable->False]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "4", ",", "2"}], "}"}], "\[Rule]", 
                   RowBox[{"-", 
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}]}]],
                Annotation[#, {1, 4, 2} -> -$CellContext`s[], "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 RowBox[{"-", 
                  InterpretationBox[
                   StyleBox["s",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`s[],
                   Editable->False]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "4", ",", "3"}], "}"}], "\[Rule]", 
                   RowBox[{"-", 
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}]}]],
                Annotation[#, {1, 4, 3} -> -$CellContext`s[], "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "4", ",", "4"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {1, 4, 4} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {2, 1, 1} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {2, 1, 2} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 RowBox[{"-", 
                  InterpretationBox[
                   StyleBox["s",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`s[],
                   Editable->False]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "3"}], "}"}], "\[Rule]", 
                   RowBox[{"-", 
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}]}]],
                Annotation[#, {2, 1, 3} -> -$CellContext`s[], "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 RowBox[{"-", 
                  InterpretationBox[
                   StyleBox["s",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`s[],
                   Editable->False]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "4"}], "}"}], "\[Rule]", 
                   RowBox[{"-", 
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}]}]],
                Annotation[#, {2, 1, 4} -> -$CellContext`s[], "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {2, 2, 1} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {2, 2, 2} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "3"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {2, 2, 3} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "4"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {2, 2, 4} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox[
                 InterpretationBox[
                  StyleBox["s",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`s[],
                  Editable->False],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "3", ",", "1"}], "}"}], "\[Rule]", 
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}]],
                Annotation[#, {2, 3, 1} -> $CellContext`s[], "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "3", ",", "2"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {2, 3, 2} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "3", ",", "3"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {2, 3, 3} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 InterpretationBox[
                  StyleBox["s",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`s[],
                  Editable->False],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "3", ",", "4"}], "}"}], "\[Rule]", 
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}]],
                Annotation[#, {2, 3, 4} -> $CellContext`s[], "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox[
                 InterpretationBox[
                  StyleBox["s",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`s[],
                  Editable->False],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "4", ",", "1"}], "}"}], "\[Rule]", 
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}]],
                Annotation[#, {2, 4, 1} -> $CellContext`s[], "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "4", ",", "2"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {2, 4, 2} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 RowBox[{"-", 
                  InterpretationBox[
                   StyleBox["s",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`s[],
                   Editable->False]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "4", ",", "3"}], "}"}], "\[Rule]", 
                   RowBox[{"-", 
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}]}]],
                Annotation[#, {2, 4, 3} -> -$CellContext`s[], "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "4", ",", "4"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {2, 4, 4} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {3, 1, 1} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 InterpretationBox[
                  StyleBox["s",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`s[],
                  Editable->False],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}]],
                Annotation[#, {3, 1, 2} -> $CellContext`s[], "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "1", ",", "3"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {3, 1, 3} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 RowBox[{"-", 
                  InterpretationBox[
                   StyleBox["s",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`s[],
                   Editable->False]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "1", ",", "4"}], "}"}], "\[Rule]", 
                   RowBox[{"-", 
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}]}]],
                Annotation[#, {3, 1, 4} -> -$CellContext`s[], "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox[
                 RowBox[{"-", 
                  InterpretationBox[
                   StyleBox["s",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`s[],
                   Editable->False]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                   RowBox[{"-", 
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}]}]],
                Annotation[#, {3, 2, 1} -> -$CellContext`s[], "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {3, 2, 2} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "2", ",", "3"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {3, 2, 3} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 RowBox[{"-", 
                  InterpretationBox[
                   StyleBox["s",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`s[],
                   Editable->False]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "2", ",", "4"}], "}"}], "\[Rule]", 
                   RowBox[{"-", 
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}]}]],
                Annotation[#, {3, 2, 4} -> -$CellContext`s[], "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "3", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {3, 3, 1} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "3", ",", "2"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {3, 3, 2} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "3", ",", "3"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {3, 3, 3} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "3", ",", "4"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {3, 3, 4} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox[
                 InterpretationBox[
                  StyleBox["s",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`s[],
                  Editable->False],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "4", ",", "1"}], "}"}], "\[Rule]", 
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}]],
                Annotation[#, {3, 4, 1} -> $CellContext`s[], "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 InterpretationBox[
                  StyleBox["s",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`s[],
                  Editable->False],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "4", ",", "2"}], "}"}], "\[Rule]", 
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}]],
                Annotation[#, {3, 4, 2} -> $CellContext`s[], "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "4", ",", "3"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {3, 4, 3} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "4", ",", "4"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {3, 4, 4} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {4, 1, 1} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 InterpretationBox[
                  StyleBox["s",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`s[],
                  Editable->False],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}]],
                Annotation[#, {4, 1, 2} -> $CellContext`s[], "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 InterpretationBox[
                  StyleBox["s",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`s[],
                  Editable->False],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "1", ",", "3"}], "}"}], "\[Rule]", 
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}]],
                Annotation[#, {4, 1, 3} -> $CellContext`s[], "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "1", ",", "4"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {4, 1, 4} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox[
                 RowBox[{"-", 
                  InterpretationBox[
                   StyleBox["s",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`s[],
                   Editable->False]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                   RowBox[{"-", 
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}]}]],
                Annotation[#, {4, 2, 1} -> -$CellContext`s[], "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {4, 2, 2} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 InterpretationBox[
                  StyleBox["s",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`s[],
                  Editable->False],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "2", ",", "3"}], "}"}], "\[Rule]", 
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}]],
                Annotation[#, {4, 2, 3} -> $CellContext`s[], "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "2", ",", "4"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {4, 2, 4} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox[
                 RowBox[{"-", 
                  InterpretationBox[
                   StyleBox["s",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`s[],
                   Editable->False]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "3", ",", "1"}], "}"}], "\[Rule]", 
                   RowBox[{"-", 
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}]}]],
                Annotation[#, {4, 3, 1} -> -$CellContext`s[], "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 RowBox[{"-", 
                  InterpretationBox[
                   StyleBox["s",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`s[],
                   Editable->False]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "3", ",", "2"}], "}"}], "\[Rule]", 
                   RowBox[{"-", 
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}]}]],
                Annotation[#, {4, 3, 2} -> -$CellContext`s[], "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "3", ",", "3"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {4, 3, 3} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "3", ",", "4"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {4, 3, 4} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "4", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {4, 4, 1} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "4", ",", "2"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {4, 4, 2} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "4", ",", "3"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {4, 4, 3} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "4", ",", "4"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {4, 4, 4} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}],
         FontSize->Small], 
        StyleBox["~",
         FontColor->RGBColor[1., 0., 0.]]], 
       StyleBox["~",
        FontColor->RGBColor[1., 0., 0.]]], 
      StyleBox[GridBox[{
         {" ", " ", " "},
         {"a", "b", "c"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{{0, 0, 0, 0}, {0, 0, 
      $CellContext`s[], 
      $CellContext`s[]}, {0, -$CellContext`s[], 0, 
      $CellContext`s[]}, {0, -$CellContext`s[], -$CellContext`s[], 0}}, {{
     0, 0, -$CellContext`s[], -$CellContext`s[]}, {0, 0, 0, 0}, {
      $CellContext`s[], 0, 0, 
      $CellContext`s[]}, {
      $CellContext`s[], 0, -$CellContext`s[], 0}}, {{0, 
      $CellContext`s[], 0, -$CellContext`s[]}, {-$CellContext`s[], 0, 
      0, -$CellContext`s[]}, {0, 0, 0, 0}, {
      $CellContext`s[], 
      $CellContext`s[], 0, 0}}, {{0, 
      $CellContext`s[], 
      $CellContext`s[], 0}, {-$CellContext`s[], 0, 
      $CellContext`s[], 0}, {-$CellContext`s[], -$CellContext`s[], 0, 0}, {0, 
     0, 0, 0}}}, {-$CellContext`B, -$CellContext`B, -$CellContext`B}, 
   2 $CellContext`B][-$CellContext`a, -$CellContext`b, -$CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{3.728675918301079*^9, 3.728753497050349*^9, 
  3.728843514524542*^9, 3.730955594512937*^9, 3.7309561991093683`*^9, 
  3.7311160740085163`*^9, 3.7333678960635*^9, 3.733368128344076*^9, 
  3.792273852646656*^9, 3.794517939868157*^9, 3.794518018324379*^9, 
  3.7945181887247066`*^9, 3.794518783350177*^9, 3.7945190634122257`*^9, 
  3.7945192988000174`*^9, 3.7945200240994787`*^9, 3.79452021053724*^9, 
  3.794520318478767*^9, 3.794520419937944*^9, 3.794520589755213*^9, 
  3.794520665542091*^9},
 CellLabel->"Out[96]=",
 CellID->1841984738,ExpressionUUID->"faa03992-7fa3-4da2-b1f6-4a3c26e0020d"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
All three indices use the same basis B, so all lines in the representation \
are red.\
\>", "MathCaption",
 CellID->2130802682,ExpressionUUID->"43b69aa9-1f4d-4f20-8ff6-0a454a45e8a7"],

Cell[CellGroupData[{

Cell[BoxData["A"], "Input",
 CellLabel->"In[97]:=",
 CellID->963379002,ExpressionUUID->"d5444b21-8a94-4f68-8d5a-7d0530a91678"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", 
         InterpretationBox[
          StyleBox["s",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`s[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["s",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`s[],
          Editable->False]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", 
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["s",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`s[],
           Editable->False]}], ",", "0", ",", 
         InterpretationBox[
          StyleBox["s",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`s[],
          Editable->False]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", 
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["s",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`s[],
           Editable->False]}], ",", 
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["s",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`s[],
           Editable->False]}], ",", "0"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", 
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["s",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`s[],
           Editable->False]}], ",", 
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["s",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`s[],
           Editable->False]}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         InterpretationBox[
          StyleBox["s",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`s[],
          Editable->False], ",", "0", ",", "0", ",", 
         InterpretationBox[
          StyleBox["s",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`s[],
          Editable->False]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         InterpretationBox[
          StyleBox["s",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`s[],
          Editable->False], ",", "0", ",", 
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["s",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`s[],
           Editable->False]}], ",", "0"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", 
         InterpretationBox[
          StyleBox["s",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`s[],
          Editable->False], ",", "0", ",", 
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["s",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`s[],
           Editable->False]}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["s",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`s[],
           Editable->False]}], ",", "0", ",", "0", ",", 
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["s",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`s[],
           Editable->False]}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         InterpretationBox[
          StyleBox["s",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`s[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["s",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`s[],
          Editable->False], ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", 
         InterpretationBox[
          StyleBox["s",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`s[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["s",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`s[],
          Editable->False], ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["s",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`s[],
           Editable->False]}], ",", "0", ",", 
         InterpretationBox[
          StyleBox["s",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`s[],
          Editable->False], ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["s",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`s[],
           Editable->False]}], ",", 
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["s",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`s[],
           Editable->False]}], ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}]}], "}"}], 
   ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "B"}], ",", 
     RowBox[{"-", "B"}], ",", 
     RowBox[{"-", "B"}]}], "}"}], ",", 
   RowBox[{"2", " ", "B"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728675920138648*^9, 3.728753497167416*^9, 
  3.728843514759267*^9, 3.730955595899736*^9, 3.730956199217504*^9, 
  3.731116075966908*^9, 3.733367896191687*^9, 3.7333681284670486`*^9, 
  3.792273852701557*^9, 3.7945179399154778`*^9, 3.794518018366503*^9, 
  3.794518188780449*^9, 3.794518783392379*^9, 3.7945190634555817`*^9, 
  3.794519298844075*^9, 3.794520024150322*^9, 3.794520210582876*^9, 
  3.794520318528282*^9, 3.794520420003034*^9, 3.794520589812488*^9, 
  3.794520665591525*^9},
 CellLabel->"Out[97]=",
 CellID->113236414,ExpressionUUID->"74f036dd-52b2-454f-80d4-9ee6c4087836"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Note that odd indices correspond to vertical lines and even indices \
correspond to horizontal lines. This is nonstandard, but as we said it helps \
in distinguishing a matrix {n, 1} from a vector, etc:\
\>", "MathCaption",
 CellID->143131777,ExpressionUUID->"bd5a66c1-76de-4fa5-9174-65adaae3c628"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"MatrixForm", "[", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", "1", "}"}], ",", 
      RowBox[{"{", "2", "}"}]}], "}"}], "]"}], ",", 
   RowBox[{"MatrixForm", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "2"}], "}"}], "]"}]}], "}"}]], "Input",
 CellLabel->"In[98]:=",
 CellID->434091399,ExpressionUUID->"6fa0ddfe-94ca-40cb-b899-89a0122a8d3f"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   TagBox[
    RowBox[{"(", "\[NoBreak]", GridBox[{
       {"1"},
       {"2"}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Baseline}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.7]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}], "\[NoBreak]", ")"}],
    Function[BoxForm`e$, 
     MatrixForm[BoxForm`e$]]], ",", 
   TagBox[
    RowBox[{"(", "\[NoBreak]", 
     TagBox[GridBox[{
        {"1"},
        {"2"}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Baseline}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.5599999999999999]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      Column], "\[NoBreak]", ")"}],
    Function[BoxForm`e$, 
     MatrixForm[BoxForm`e$]]]}], "}"}]], "Output",
 CellChangeTimes->{3.7286759218398952`*^9, 3.72875349726202*^9, 
  3.728843514896347*^9, 3.730955597692339*^9, 3.730956199311894*^9, 
  3.731116077519535*^9, 3.733367896313506*^9, 3.7333681286001368`*^9, 
  3.792273852748253*^9, 3.794517939966434*^9, 3.7945180184073887`*^9, 
  3.794518188826692*^9, 3.794518783443468*^9, 3.794519063496286*^9, 
  3.7945192988946*^9, 3.794520024195364*^9, 3.794520210626732*^9, 
  3.794520318576174*^9, 3.794520420053034*^9, 3.7945205898681602`*^9, 
  3.794520665638256*^9},
 CellLabel->"Out[98]=",
 CellID->493898856,ExpressionUUID->"dacdc655-71fc-43f8-bcf1-47d10435e886"]
}, Open  ]]
}, Open  ]],

Cell["\<\
This type of representation can get difficult to understand for ranks beyond \
2. To help with that you can hover over a given element with the mouse, and \
you will get positional information.\
\>", "MathCaption",
 CellChangeTimes->{3.728675924954261*^9},
 CellID->1023131317,ExpressionUUID->"cae99855-d5fc-49ab-98a8-1f461466032c"],

Cell[CellGroupData[{

Cell["\<\
To avoid formatted tensors getting too large, a given component will not be \
typeset if its ByteCount is larger than a given predefined limit (that you \
can configure):\
\>", "MathCaption",
 CellChangeTimes->{3.728675926738263*^9},
 CellID->866483290,ExpressionUUID->"53d2d327-f727-4e87-9da0-8448ccffeeec"],

Cell[CellGroupData[{

Cell[BoxData["$LargeComponentSize"], "Input",
 CellLabel->"In[99]:=",
 CellID->898044723,ExpressionUUID->"74cc9980-6f44-4f48-8810-9b79c98406b4"],

Cell[BoxData["1000"], "Output",
 CellChangeTimes->{3.72867592790958*^9, 3.728753497379475*^9, 
  3.72884351511936*^9, 3.730955599733864*^9, 3.730956199416418*^9, 
  3.7311160800640306`*^9, 3.733367896443303*^9, 3.7333681287336283`*^9, 
  3.792273852798915*^9, 3.794517940008329*^9, 3.794518018442936*^9, 
  3.794518188880432*^9, 3.794518783483783*^9, 3.794519063547027*^9, 
  3.7945192989542227`*^9, 3.794520024237585*^9, 3.794520210673935*^9, 
  3.794520318625661*^9, 3.794520420098598*^9, 3.794520589925107*^9, 
  3.794520665687207*^9},
 CellLabel->"Out[99]=",
 CellID->933777137,ExpressionUUID->"3596209e-22bd-4668-8e08-713c8f583ad1"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData["$LargeComponentString"], "Input",
 CellLabel->"In[100]:=",
 CellID->1110383075,ExpressionUUID->"7dcc2637-f47f-4100-9fe4-044497045efe"],

Cell[BoxData["\<\"\[GrayCircle]\"\>"], "Output",
 CellChangeTimes->{3.728675928461459*^9, 3.728753497446431*^9, 
  3.728843515718472*^9, 3.7309556003017797`*^9, 3.7309561995128508`*^9, 
  3.7311160807036037`*^9, 3.733367896575225*^9, 3.7333681288829603`*^9, 
  3.792273852846445*^9, 3.794517940061317*^9, 3.794518018479372*^9, 
  3.7945181889267893`*^9, 3.79451878352435*^9, 3.7945190635913*^9, 
  3.7945192989951353`*^9, 3.7945200242832527`*^9, 3.7945202107139997`*^9, 
  3.794520318670211*^9, 3.794520420147768*^9, 3.7945205899777937`*^9, 
  3.7945206657309303`*^9},
 CellLabel->"Out[100]=",
 CellID->181849283,ExpressionUUID->"1829d1d2-e0ba-4a31-9b66-fb8eb15a3bc3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"s", "[", "]"}], "+", "1"}], ")"}], "^", "4"}], ",", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"s", "[", "]"}], "+", "1"}], ")"}], "^", "5"}], ",", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"s", "[", "]"}], "+", "1"}], ")"}], "^", "6"}], ",", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"s", "[", "]"}], "+", "1"}], ")"}], "^", "7"}]}], "}"}], "//", 
  "Expand"}]], "Input",
 CellLabel->"In[101]:=",
 CellID->1454077224,ExpressionUUID->"4c50ad62-f3ef-49f1-877e-54bd1d7a4c62"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"1", "+", 
    RowBox[{"4", " ", 
     InterpretationBox[
      StyleBox["s",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`s[],
      Editable->False]}], "+", 
    RowBox[{"6", " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["s",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`s[],
       Editable->False], "2"]}], "+", 
    RowBox[{"4", " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["s",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`s[],
       Editable->False], "3"]}], "+", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["s",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`s[],
      Editable->False], "4"]}], ",", 
   RowBox[{"1", "+", 
    RowBox[{"5", " ", 
     InterpretationBox[
      StyleBox["s",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`s[],
      Editable->False]}], "+", 
    RowBox[{"10", " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["s",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`s[],
       Editable->False], "2"]}], "+", 
    RowBox[{"10", " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["s",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`s[],
       Editable->False], "3"]}], "+", 
    RowBox[{"5", " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["s",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`s[],
       Editable->False], "4"]}], "+", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["s",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`s[],
      Editable->False], "5"]}], ",", 
   RowBox[{"1", "+", 
    RowBox[{"6", " ", 
     InterpretationBox[
      StyleBox["s",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`s[],
      Editable->False]}], "+", 
    RowBox[{"15", " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["s",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`s[],
       Editable->False], "2"]}], "+", 
    RowBox[{"20", " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["s",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`s[],
       Editable->False], "3"]}], "+", 
    RowBox[{"15", " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["s",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`s[],
       Editable->False], "4"]}], "+", 
    RowBox[{"6", " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["s",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`s[],
       Editable->False], "5"]}], "+", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["s",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`s[],
      Editable->False], "6"]}], ",", 
   RowBox[{"1", "+", 
    RowBox[{"7", " ", 
     InterpretationBox[
      StyleBox["s",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`s[],
      Editable->False]}], "+", 
    RowBox[{"21", " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["s",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`s[],
       Editable->False], "2"]}], "+", 
    RowBox[{"35", " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["s",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`s[],
       Editable->False], "3"]}], "+", 
    RowBox[{"35", " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["s",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`s[],
       Editable->False], "4"]}], "+", 
    RowBox[{"21", " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["s",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`s[],
       Editable->False], "5"]}], "+", 
    RowBox[{"7", " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["s",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`s[],
       Editable->False], "6"]}], "+", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["s",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`s[],
      Editable->False], "7"]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728675929273237*^9, 3.728753497567059*^9, 
  3.728843515874641*^9, 3.730955600927574*^9, 3.730956199615027*^9, 
  3.731116081270927*^9, 3.733367896667039*^9, 3.7333681290166683`*^9, 
  3.792273852897808*^9, 3.7945179401075373`*^9, 3.794518018520784*^9, 
  3.794518188977257*^9, 3.7945187835653563`*^9, 3.7945190636301193`*^9, 
  3.794519299043744*^9, 3.794520024328355*^9, 3.794520210759976*^9, 
  3.794520318725844*^9, 3.794520420195907*^9, 3.794520590036405*^9, 
  3.79452066578542*^9},
 CellLabel->"Out[101]=",
 CellID->471669700,ExpressionUUID->"81f77b31-5f0f-4cc5-aacc-32d87a8831dc"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ByteCount", "/@", "%"}]], "Input",
 CellLabel->"In[102]:=",
 CellID->1791678429,ExpressionUUID->"f91e135a-ec55-4dc4-a959-6641be83e62a"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"688", ",", "880", ",", "1072", ",", "1264"}], "}"}]], "Output",
 CellChangeTimes->{3.7286759298316393`*^9, 3.728753497683734*^9, 
  3.728843516046851*^9, 3.730955601492302*^9, 3.7309561997149153`*^9, 
  3.7311160819297447`*^9, 3.733367896768787*^9, 3.733368129150403*^9, 
  3.792273852945785*^9, 3.7945179401529016`*^9, 3.7945180185578947`*^9, 
  3.794518189029158*^9, 3.79451878360614*^9, 3.7945190636707993`*^9, 
  3.794519299085966*^9, 3.794520024370125*^9, 3.794520210804132*^9, 
  3.794520318771241*^9, 3.79452042024162*^9, 3.7945205901076117`*^9, 
  3.794520665830723*^9},
 CellLabel->"Out[102]=",
 CellID->990506823,ExpressionUUID->"ad2fc02c-68b6-4bb4-ba9e-341905be4f8e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"s", "[", "]"}], "+", "1"}], ")"}], "^", "4"}], ",", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"s", "[", "]"}], "+", "1"}], ")"}], "^", "5"}], ",", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"s", "[", "]"}], "+", "1"}], ")"}], "^", "6"}], ",", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"s", "[", "]"}], "+", "1"}], ")"}], "^", "7"}]}], "}"}], 
     ",", 
     RowBox[{"{", "B", "}"}]}], "]"}], "[", "a", "]"}], "//", 
  "Expand"}]], "Input",
 CellLabel->"In[103]:=",
 CellID->1319592022,ExpressionUUID->"a4174643-6a44-4a5f-ae3f-ae317f651959"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{"1", "+", 
             RowBox[{"4", " ", 
              InterpretationBox[
               StyleBox["s",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`s[],
               Editable->False]}], "+", 
             RowBox[{"6", " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["s",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`s[],
                Editable->False], "2"]}], "+", 
             RowBox[{"4", " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["s",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`s[],
                Editable->False], "3"]}], "+", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["s",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`s[],
               Editable->False], "4"]}],
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", 
              RowBox[{"1", "+", 
                RowBox[{"4", " ", 
                  InterpretationBox[
                   StyleBox[
                   "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`s[], Editable -> False]}], "+", 
                RowBox[{"6", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "2"]}], "+", 
                RowBox[{"4", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "3"]}], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                  "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`s[], Editable -> False], "4"]}]}]],
           
           Annotation[#, {1} -> 
            1 + 4 $CellContext`s[] + 6 $CellContext`s[]^2 + 
             4 $CellContext`s[]^3 + $CellContext`s[]^4, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{"1", "+", 
             RowBox[{"5", " ", 
              InterpretationBox[
               StyleBox["s",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`s[],
               Editable->False]}], "+", 
             RowBox[{"10", " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["s",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`s[],
                Editable->False], "2"]}], "+", 
             RowBox[{"10", " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["s",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`s[],
                Editable->False], "3"]}], "+", 
             RowBox[{"5", " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["s",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`s[],
                Editable->False], "4"]}], "+", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["s",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`s[],
               Editable->False], "5"]}],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              RowBox[{"1", "+", 
                RowBox[{"5", " ", 
                  InterpretationBox[
                   StyleBox[
                   "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`s[], Editable -> False]}], "+", 
                RowBox[{"10", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "2"]}], "+", 
                RowBox[{"10", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "3"]}], "+", 
                RowBox[{"5", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "4"]}], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                  "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`s[], Editable -> False], "5"]}]}]],
           
           Annotation[#, {2} -> 
            1 + 5 $CellContext`s[] + 10 $CellContext`s[]^2 + 
             10 $CellContext`s[]^3 + 
             5 $CellContext`s[]^4 + $CellContext`s[]^5, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", 
              RowBox[{"1", "+", 
                RowBox[{"6", " ", 
                  InterpretationBox[
                   StyleBox[
                   "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`s[], Editable -> False]}], "+", 
                RowBox[{"15", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "2"]}], "+", 
                RowBox[{"20", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "3"]}], "+", 
                RowBox[{"15", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "4"]}], "+", 
                RowBox[{"6", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "5"]}], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                  "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`s[], Editable -> False], "6"]}]}]],
           
           Annotation[#, {3} -> 
            1 + 6 $CellContext`s[] + 15 $CellContext`s[]^2 + 
             20 $CellContext`s[]^3 + 15 $CellContext`s[]^4 + 
             6 $CellContext`s[]^5 + $CellContext`s[]^6, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", "4", "}"}], "\[Rule]", 
              RowBox[{"1", "+", 
                RowBox[{"7", " ", 
                  InterpretationBox[
                   StyleBox[
                   "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`s[], Editable -> False]}], "+", 
                RowBox[{"21", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "2"]}], "+", 
                RowBox[{"35", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "3"]}], "+", 
                RowBox[{"35", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "4"]}], "+", 
                RowBox[{"21", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "5"]}], "+", 
                RowBox[{"7", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "6"]}], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                  "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`s[], Editable -> False], "7"]}]}]],
           
           Annotation[#, {4} -> 
            1 + 7 $CellContext`s[] + 21 $CellContext`s[]^2 + 
             35 $CellContext`s[]^3 + 35 $CellContext`s[]^4 + 
             21 $CellContext`s[]^5 + 
             7 $CellContext`s[]^6 + $CellContext`s[]^7, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{
   1 + 4 $CellContext`s[] + 6 $CellContext`s[]^2 + 
    4 $CellContext`s[]^3 + $CellContext`s[]^4, 1 + 5 $CellContext`s[] + 
    10 $CellContext`s[]^2 + 10 $CellContext`s[]^3 + 
    5 $CellContext`s[]^4 + $CellContext`s[]^5, 1 + 6 $CellContext`s[] + 
    15 $CellContext`s[]^2 + 20 $CellContext`s[]^3 + 15 $CellContext`s[]^4 + 
    6 $CellContext`s[]^5 + $CellContext`s[]^6, 1 + 7 $CellContext`s[] + 
    21 $CellContext`s[]^2 + 35 $CellContext`s[]^3 + 35 $CellContext`s[]^4 + 
    21 $CellContext`s[]^5 + 
    7 $CellContext`s[]^6 + $CellContext`s[]^7}, {$CellContext`B}, 
   0][$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.728675930572823*^9, 3.7287534977778463`*^9, 
  3.728843516655938*^9, 3.730955602081316*^9, 3.730956199808443*^9, 
  3.731116082575889*^9, 3.733367896884974*^9, 3.733368129283475*^9, 
  3.792273852998394*^9, 3.794517940205872*^9, 3.794518018601645*^9, 
  3.794518189082541*^9, 3.79451878366427*^9, 3.794519063713676*^9, 
  3.794519299131714*^9, 3.794520024427278*^9, 3.794520210855833*^9, 
  3.794520318828532*^9, 3.794520420299015*^9, 3.794520590162527*^9, 
  3.794520665891461*^9},
 CellLabel->"Out[103]=",
 CellID->154942617,ExpressionUUID->"feaa1e17-6ac8-4309-9176-324714e567a7"]
}, Open  ]]
}, Open  ]],

Cell["\<\
Still, hovering over the \[OpenCurlyDoubleQuote]hidden component\
\[CloseCurlyDoubleQuote] will show you what it actually is.\
\>", "MathCaption",
 CellChangeTimes->{3.728675906251088*^9},
 CellID->1772207373,ExpressionUUID->"5df70d6f-e464-4e18-8bf0-f47266bca227"],

Cell[CellGroupData[{

Cell["\<\
As usual in xTensor, you can choose your own formatting for the tensor:\
\>", "MathCaption",
 CellChangeTimes->{3.728675904290742*^9},
 CellID->2038617768,ExpressionUUID->"18406536-ee1d-4f9b-81e0-794e8915b3c4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"v", "[", "a", "]"}]], "Input",
 CellLabel->"In[104]:=",
 CellID->2094722538,ExpressionUUID->"aa060a18-5781-4c90-9e02-81774e74fcf1"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{"\[ImaginaryI]", " ", 
             InterpretationBox[
              StyleBox["s",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`s[],
              Editable->False]}],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              RowBox[{"\[ImaginaryI]", " ", 
                InterpretationBox[
                 StyleBox["s", AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`s[], Editable -> False]}]}]],
           Annotation[#, {2} -> Complex[0, 1] $CellContext`s[], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", "1"}]],
           Annotation[#, {3} -> 1, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{"3", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["s",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`s[],
               Editable->False], "2"]}],
            RowBox[{
              RowBox[{"{", "4", "}"}], "\[Rule]", 
              RowBox[{"3", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                  "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`s[], Editable -> False], "2"]}]}]],
           Annotation[#, {4} -> 3 $CellContext`s[]^2, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{
   0, Complex[0, 1] $CellContext`s[], 1, 
    3 $CellContext`s[]^2}, {$CellContext`B}, 0][$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.72867593245676*^9, 3.728753497900971*^9, 
  3.728843516939042*^9, 3.7309556038000383`*^9, 3.730956199917529*^9, 
  3.73111608746346*^9, 3.733367896985474*^9, 3.733368129417417*^9, 
  3.7922738530498533`*^9, 3.794517940262871*^9, 3.7945180186443033`*^9, 
  3.794518189131404*^9, 3.794518783704965*^9, 3.79451906375419*^9, 
  3.794519299178604*^9, 3.79452002447379*^9, 3.794520210904949*^9, 
  3.794520318877454*^9, 3.794520420354424*^9, 3.794520590216382*^9, 
  3.7945206659404373`*^9},
 CellLabel->"Out[104]=",
 CellID->1370383755,ExpressionUUID->"40503d61-9c2b-4f27-b558-9ff7267ad0dd"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We need to use Evaluate because PrintAs has attribute HoldFirst:\
\>", "MathCaption",
 CellID->906580619,ExpressionUUID->"2dbb4005-ef00-48e1-99d8-ecdc425ca691"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", 
    RowBox[{"Evaluate", "[", "v", "]"}], "]"}], "^=", 
   "\"\<\[DoubleStruckCapitalV]\>\""}], ";"}]], "Input",
 CellLabel->"In[105]:=",
 CellID->503988068,ExpressionUUID->"f7c26616-a89a-4c91-a0df-3742f347415c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"v", "[", "a", "]"}]], "Input",
 CellLabel->"In[106]:=",
 CellID->851955108,ExpressionUUID->"d8117faa-3a9c-452b-a5ba-d190524b60a6"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"\[DoubleStruckCapitalV]", 
      StyleBox[GridBox[{
         {"a"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{
   0, Complex[0, 1] $CellContext`s[], 1, 
    3 $CellContext`s[]^2}, {$CellContext`B}, 0][$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.728675934087634*^9, 3.7287534980500383`*^9, 
  3.728843517161367*^9, 3.730955605533149*^9, 3.730956200024411*^9, 
  3.731116089252994*^9, 3.7333678971703672`*^9, 3.733368129609619*^9, 
  3.792273853134248*^9, 3.794517940340609*^9, 3.794518018706027*^9, 
  3.7945181892110567`*^9, 3.794518783772386*^9, 3.794519063814732*^9, 
  3.794519299242278*^9, 3.794520024545257*^9, 3.794520211002809*^9, 
  3.794520318956563*^9, 3.794520420416975*^9, 3.7945205903037577`*^9, 
  3.794520666005086*^9},
 CellLabel->"Out[106]=",
 CellID->277656210,ExpressionUUID->"59565aa5-6375-42a8-b197-b33674fb2aaf"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellLabel->"In[107]:=",
 CellID->1081456262,ExpressionUUID->"18fff9d7-a4cf-4266-b42f-ff9893892c12"],

Cell["CTensor[{0, I*s[], 1, 3*s[]^2}, {B}, 0][a]", "Output",
 CellChangeTimes->{3.728675934665162*^9, 3.728753498165447*^9, 
  3.7288435174532423`*^9, 3.730955607025741*^9, 3.7309562001329603`*^9, 
  3.7311160898388653`*^9, 3.733367897287362*^9, 3.7333681297338753`*^9, 
  3.792273853195631*^9, 3.794517940394384*^9, 3.7945180187457113`*^9, 
  3.794518189249949*^9, 3.7945187838163424`*^9, 3.794519063860194*^9, 
  3.794519299286377*^9, 3.794520024594658*^9, 3.7945202110513353`*^9, 
  3.794520319001039*^9, 3.7945204204670353`*^9, 3.794520590355315*^9, 
  3.79452066605578*^9},
 CellLabel->"Out[107]//InputForm=",
 CellID->85956539,ExpressionUUID->"9187f295-1b2b-4ba4-8ddf-04ee66f8228c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
That definition was actually stored as an upvalue of CTensor. You can remove \
it as follows:\
\>", "MathCaption",
 CellChangeTimes->{3.728675900995356*^9},
 CellID->1453967730,ExpressionUUID->"360fc500-ceef-42af-9ab6-25d7b03f6218"],

Cell[BoxData[
 RowBox[{"CTensor", "/:", 
  RowBox[{
   RowBox[{"PrintAs", "[", 
    RowBox[{"Evaluate", "[", "v", "]"}], "]"}], "=."}]}]], "Input",
 CellLabel->"In[108]:=",
 CellID->619218897,ExpressionUUID->"b05e1aac-d23f-47f6-b0fa-51d574d2202a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"v", "[", "a", "]"}]], "Input",
 CellLabel->"In[109]:=",
 CellID->506248517,ExpressionUUID->"55b5b03d-5f8a-418a-8db6-0c7a46f25a07"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{"\[ImaginaryI]", " ", 
             InterpretationBox[
              StyleBox["s",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`s[],
              Editable->False]}],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              RowBox[{"\[ImaginaryI]", " ", 
                InterpretationBox[
                 StyleBox["s", AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`s[], Editable -> False]}]}]],
           Annotation[#, {2} -> Complex[0, 1] $CellContext`s[], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", "1"}]],
           Annotation[#, {3} -> 1, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{"3", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["s",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`s[],
               Editable->False], "2"]}],
            RowBox[{
              RowBox[{"{", "4", "}"}], "\[Rule]", 
              RowBox[{"3", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                  "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`s[], Editable -> False], "2"]}]}]],
           Annotation[#, {4} -> 3 $CellContext`s[]^2, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{
   0, Complex[0, 1] $CellContext`s[], 1, 
    3 $CellContext`s[]^2}, {$CellContext`B}, 0][$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.7286759365071*^9, 3.728753498380707*^9, 
  3.7288435184347477`*^9, 3.730955608986121*^9, 3.730956200242787*^9, 
  3.731116091664873*^9, 3.7333678974407177`*^9, 3.733368129967042*^9, 
  3.792273853265984*^9, 3.794517940471054*^9, 3.794518018854219*^9, 
  3.794518189318657*^9, 3.7945187838795424`*^9, 3.794519063921028*^9, 
  3.79451929935163*^9, 3.794520024665001*^9, 3.794520211114496*^9, 
  3.794520319070142*^9, 3.7945204205381413`*^9, 3.794520590427369*^9, 
  3.794520666121702*^9},
 CellLabel->"Out[109]=",
 CellID->1562540189,ExpressionUUID->"2e6ee187-d10e-4953-bb4f-47175c2b1339"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Components of a CTensor may be other tensors, as long as there are no free \
indices:\
\>", "MathCaption",
 CellChangeTimes->{{3.728675898347484*^9, 3.728675898499796*^9}},
 CellID->379971663,ExpressionUUID->"e0ac1f65-ee18-40ea-8a18-ed9d87cc9174"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CTensor", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"1", ",", 
      RowBox[{"Basis", "[", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{
          RowBox[{"-", "a"}], ",", 
          RowBox[{"-", "B"}]}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{"b", ",", "B"}], "}"}]}], "]"}], ",", 
      RowBox[{"3", 
       RowBox[{"s", "[", "]"}]}], ",", "0"}], "}"}], ",", 
    RowBox[{"{", "B", "}"}]}], "]"}], "[", "c", "]"}]], "Input",
 CellLabel->"In[110]:=",
 CellID->1714857510,ExpressionUUID->"deb9c786-47f3-4b70-b950-b7ef1b29b8eb"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", "1"}]],
           Annotation[#, {1} -> 1, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            InterpretationBox[
             StyleBox[GridBox[{
                {"e", 
                 StyleBox[GridBox[{
                    {" ", 
                    StyleBox["b",
                    FontColor->RGBColor[1., 0., 0.]]},
                    {
                    StyleBox["a",
                    FontColor->RGBColor[1., 0., 0.]], " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "Rows" -> {{
                    Offset[0.]}}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "Rows" -> {{Center}}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             
             xAct`xTensor`Basis[{-$CellContext`a, -$CellContext`B}, \
{$CellContext`b, $CellContext`B}],
             Editable->False],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              InterpretationBox[
               StyleBox[
                GridBox[{{"e", 
                   StyleBox[
                    
                    GridBox[{{
                    " ", "\!\(\*StyleBox[\"b\",FontColor->RGBColor[1., 0., \
0.]]\)"}, {"\!\(\*StyleBox[\"a\",FontColor->RGBColor[1., 0., 0.]]\)", " "}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                 ColumnSpacings -> 0.05, RowAlignments -> Center], 
                AutoSpacing -> False, ShowAutoStyles -> False], 
               
               xAct`xTensor`Basis[{-$CellContext`a, -$CellContext`B}, \
{$CellContext`b, $CellContext`B}], Editable -> False]}]],
           
           Annotation[#, {2} -> 
            xAct`xTensor`Basis[{-$CellContext`a, -$CellContext`B}, \
{$CellContext`b, $CellContext`B}], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{"3", " ", 
             InterpretationBox[
              StyleBox["s",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`s[],
              Editable->False]}],
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", 
              RowBox[{"3", " ", 
                InterpretationBox[
                 StyleBox["s", AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`s[], Editable -> False]}]}]],
           Annotation[#, {3} -> 3 $CellContext`s[], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", "4", "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4} -> 0, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"c"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{1, 
    xAct`xTensor`Basis[{-$CellContext`a, -$CellContext`B}, {$CellContext`b, \
$CellContext`B}], 3 $CellContext`s[], 0}, {$CellContext`B}, 0][$CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{3.728675937607935*^9, 3.728753498606914*^9, 
  3.728843519027792*^9, 3.730955610167903*^9, 3.7309562003631496`*^9, 
  3.731116095826548*^9, 3.733367897531384*^9, 3.733368130065118*^9, 
  3.7922738533202353`*^9, 3.794517940527298*^9, 3.794518018907365*^9, 
  3.794518189376505*^9, 3.794518783929448*^9, 3.7945190639661713`*^9, 
  3.794519299404818*^9, 3.7945200247221937`*^9, 3.794520211167144*^9, 
  3.7945203191288567`*^9, 3.794520420599738*^9, 3.7945205905009737`*^9, 
  3.794520666176543*^9},
 CellLabel->"Out[110]=",
 CellID->35437360,ExpressionUUID->"f8e1020d-d79c-47eb-a884-bd6cfca5a523"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "Loading ",
 StyleBox["xTerior",
  FontSlant->"Italic"],
 ", the components can also be differential forms, but not tensor-valued \
differential forms."
}], "MathCaption",
 CellID->1816980729,ExpressionUUID->"a9fb137e-8383-4978-8306-646bcc9bd9ef"],

Cell[CellGroupData[{

Cell["Self-consistency checks:", "MathCaption",
 CellChangeTimes->{{3.7286758927076674`*^9, 3.7286758928679*^9}},
 CellID->777651963,ExpressionUUID->"5e64e359-6f90-4a32-b3b3-24f3a8881a54"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"v", "[", 
   RowBox[{"a", ",", "b"}], "]"}], "//", "Catch"}]], "Input",
 CellLabel->"In[111]:=",
 CellID->411066499,ExpressionUUID->"a1915fea-53a5-49cd-8bfd-656c57f59e06"],

Cell[BoxData[
 TemplateBox[{
  "CTensor", "nobi", 
   "\"Incompatible number of bases \\!\\(\\*RowBox[{\\\"1\\\"}]\\) and \
indices \\!\\(\\*RowBox[{\\\"2\\\"}]\\).\"", 2, 111, 4, 19683117308123869115, 
   "Local", "xAct`xCoba`CTensor"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728675939289867*^9, 3.7287534987443*^9, 
  3.72884351938206*^9, 3.730955611751458*^9, 3.7309562004655743`*^9, 
  3.731116099548719*^9, 3.733367897654101*^9, 3.7333681301831284`*^9, 
  3.792273853365871*^9, 3.7945179405761127`*^9, 3.794518018945627*^9, 
  3.794518189416831*^9, 3.794518783970592*^9, 3.7945190640065393`*^9, 
  3.794519299452444*^9, 3.79452002476889*^9, 3.794520211209453*^9, 
  3.794520319171747*^9, 3.794520420649097*^9, 3.794520590554871*^9, 
  3.794520666217251*^9},
 CellLabel->"During evaluation of In[111]:=",
 CellID->742739301,ExpressionUUID->"b4fd9280-16f2-40ef-97b7-38d1799a16a4"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"v", "[", 
   RowBox[{"-", "a"}], "]"}], "//", "Catch"}]], "Input",
 CellLabel->"In[112]:=",
 CellID->125015885,ExpressionUUID->"6555d131-84e5-4779-993d-05c7990ce012"],

Cell[BoxData[
 TemplateBox[{
  "MetricsOfVBundle", "missing", 
   "\"There is no \\!\\(\\*RowBox[{\\\"\\\\\\\"metric\\\\\\\"\\\"}]\\) in \\!\
\\(\\*InterpretationBox[StyleBox[\\\"\[DoubleStruckCapitalT]M\\\", \
Rule[AutoSpacing, False], Rule[ShowAutoStyles, False]], TangentM, \
Rule[Editable, False]]\\).\"", 2, 112, 5, 19683117308123869115, "Local", 
   "xAct`xTensor`MetricsOfVBundle"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728675939855215*^9, 3.728753498880011*^9, 
  3.72884351954768*^9, 3.730955612418495*^9, 3.730956200564581*^9, 
  3.73111610265126*^9, 3.733367897752858*^9, 3.73336813031643*^9, 
  3.792273853417695*^9, 3.794517940627033*^9, 3.7945180189865522`*^9, 
  3.794518189464032*^9, 3.794518784017988*^9, 3.794519064042598*^9, 
  3.7945192994942513`*^9, 3.794520024822939*^9, 3.7945202112552843`*^9, 
  3.794520319226327*^9, 3.794520420701173*^9, 3.794520590608726*^9, 
  3.794520666260206*^9},
 CellLabel->"During evaluation of In[112]:=",
 CellID->2044475543,ExpressionUUID->"65a078a6-b2a1-461a-af37-cd0ce34074b2"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.2.3. Basic operations", "Subsubsection",
 CellChangeTimes->{{3.73336802204809*^9, 3.733368023751897*^9}},
 CellID->2138164501,ExpressionUUID->"729b12b4-8f4a-4da1-8584-80022d84c090"],

Cell[TextData[{
 "Standard algebraic operations proceed with identical notations to those in ",
 StyleBox["xTensor",
  FontSlant->"Italic"],
 ". "
}], "Text",
 CellID->139419409,ExpressionUUID->"c974a48a-3dd4-4651-b425-45b2e303eaf5"],

Cell[CellGroupData[{

Cell["\<\
Scalar expressions are automatically multiplied by CTensor objects:\
\>", "MathCaption",
 CellChangeTimes->{{3.72867597784116*^9, 3.728675978007945*^9}},
 CellID->317348186,ExpressionUUID->"08fe3204-5c21-4569-a3d9-c77cc61a4a6b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"v", "[", "a", "]"}]], "Input",
 CellLabel->"In[113]:=",
 CellID->1469066601,ExpressionUUID->"1616ce59-cfe2-4b45-9c02-49bc57993e00"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{"\[ImaginaryI]", " ", 
             InterpretationBox[
              StyleBox["s",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`s[],
              Editable->False]}],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              RowBox[{"\[ImaginaryI]", " ", 
                InterpretationBox[
                 StyleBox["s", AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`s[], Editable -> False]}]}]],
           Annotation[#, {2} -> Complex[0, 1] $CellContext`s[], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", "1"}]],
           Annotation[#, {3} -> 1, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{"3", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["s",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`s[],
               Editable->False], "2"]}],
            RowBox[{
              RowBox[{"{", "4", "}"}], "\[Rule]", 
              RowBox[{"3", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                  "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`s[], Editable -> False], "2"]}]}]],
           Annotation[#, {4} -> 3 $CellContext`s[]^2, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{
   0, Complex[0, 1] $CellContext`s[], 1, 
    3 $CellContext`s[]^2}, {$CellContext`B}, 0][$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.7286759798711023`*^9, 3.728753499013833*^9, 
  3.728843525991601*^9, 3.730955622882945*^9, 3.730956200656519*^9, 
  3.7311161316804113`*^9, 3.733367684596599*^9, 3.7333678978454027`*^9, 
  3.73336813044915*^9, 3.792273853469377*^9, 3.7945179406781073`*^9, 
  3.794518019027166*^9, 3.7945182057759647`*^9, 3.794518784061626*^9, 
  3.7945190640885057`*^9, 3.794519299547443*^9, 3.794520024885169*^9, 
  3.7945202113083963`*^9, 3.794520319280143*^9, 3.794520420752955*^9, 
  3.7945205906656313`*^9, 3.794520666304517*^9},
 CellLabel->"Out[113]=",
 CellID->1762099201,ExpressionUUID->"e305c16b-6a84-4066-85dc-c6388a2fcaae"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"-", "7"}], 
  RowBox[{
   RowBox[{"s", "[", "]"}], "^", "2"}], 
  RowBox[{"v", "[", "a", "]"}]}]], "Input",
 CellLabel->"In[114]:=",
 CellID->709827812,ExpressionUUID->"5af2da47-a180-42a5-9c83-64bab7a6fa2d"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{"-", "7"}], " ", "\[ImaginaryI]", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["s",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`s[],
               Editable->False], "3"]}],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{"-", "7"}], " ", "\[ImaginaryI]", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                  "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`s[], Editable -> False], "3"]}]}]],
           
           Annotation[#, {2} -> Complex[0, -7] $CellContext`s[]^3, 
            "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{"-", "7"}], " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["s",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`s[],
               Editable->False], "2"]}],
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{"-", "7"}], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                  "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`s[], Editable -> False], "2"]}]}]],
           Annotation[#, {3} -> (-7) $CellContext`s[]^2, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{"-", "21"}], " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["s",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`s[],
               Editable->False], "4"]}],
            RowBox[{
              RowBox[{"{", "4", "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{"-", "21"}], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                  "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`s[], Editable -> False], "4"]}]}]],
           Annotation[#, {4} -> (-21) $CellContext`s[]^4, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{
   0, Complex[
     0, -7] $CellContext`s[]^3, (-7) $CellContext`s[]^2, (-21) \
$CellContext`s[]^4}, {$CellContext`B}, 0][$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.728675980950975*^9, 3.7287534991214333`*^9, 
  3.72884352618565*^9, 3.7309556229816504`*^9, 3.730956200779475*^9, 
  3.731116133013609*^9, 3.733367684707767*^9, 3.733367897969413*^9, 
  3.733368130599135*^9, 3.792273853518589*^9, 3.794517940726123*^9, 
  3.794518019070188*^9, 3.794518205832307*^9, 3.794518784107973*^9, 
  3.7945190641290703`*^9, 3.794519299592461*^9, 3.794520024935412*^9, 
  3.794520211354827*^9, 3.794520319326*^9, 3.794520420802387*^9, 
  3.794520590719898*^9, 3.794520666378212*^9},
 CellLabel->"Out[114]=",
 CellID->1801779077,ExpressionUUID->"44a12c9d-66a6-4d8a-a249-e1b5ec0dab70"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Add CTensor expressions with the same index configuration:", \
"MathCaption",
 CellChangeTimes->{3.728675982887513*^9},
 CellID->758025728,ExpressionUUID->"49312b70-6fb5-41c7-b898-bc4b7d6fa992"],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "=", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"Pi", ",", "E", ",", "I", ",", "1"}], "}"}], ",", 
     RowBox[{"{", "B", "}"}]}], "]"}]}], ";"}]], "Input",
 CellLabel->"In[115]:=",
 CellID->1265416492,ExpressionUUID->"a245922e-f79b-43cd-a9c1-dc7afe288f20"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"v", "[", "a", "]"}], "+", 
  RowBox[{"3", 
   RowBox[{"w", "[", "a", "]"}]}]}]], "Input",
 CellLabel->"In[116]:=",
 CellID->1324424803,ExpressionUUID->"cad607b5-9103-4b17-9379-14f8a69cd892"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{"3", " ", "\[Pi]"}],
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", 
              RowBox[{"3", " ", "\[Pi]"}]}]],
           Annotation[#, {1} -> 3 Pi, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{"3", " ", "\[ExponentialE]"}], "+", 
             RowBox[{"\[ImaginaryI]", " ", 
              InterpretationBox[
               StyleBox["s",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`s[],
               Editable->False]}]}],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{"3", " ", "\[ExponentialE]"}], "+", 
                RowBox[{"\[ImaginaryI]", " ", 
                  InterpretationBox[
                   StyleBox[
                   "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`s[], Editable -> False]}]}]}]],
           
           Annotation[#, {2} -> 3 E + Complex[0, 1] $CellContext`s[], 
            "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{"1", "+", 
             RowBox[{"3", " ", "\[ImaginaryI]"}]}],
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", 
              RowBox[{"1", "+", 
                RowBox[{"3", " ", "\[ImaginaryI]"}]}]}]],
           Annotation[#, {3} -> Complex[1, 3], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{"3", "+", 
             RowBox[{"3", " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["s",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`s[],
                Editable->False], "2"]}]}],
            RowBox[{
              RowBox[{"{", "4", "}"}], "\[Rule]", 
              RowBox[{"3", "+", 
                RowBox[{"3", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "2"]}]}]}]],
           Annotation[#, {4} -> 3 + 3 $CellContext`s[]^2, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{3 Pi, 3 E + Complex[0, 1] $CellContext`s[], 
    Complex[1, 3], 3 + 3 $CellContext`s[]^2}, {$CellContext`B}, 
   0][$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.728675984468026*^9, 3.728753499331847*^9, 
  3.728843526452808*^9, 3.730955623146592*^9, 3.7309562009507847`*^9, 
  3.7311161410700827`*^9, 3.7333676849069347`*^9, 3.733367898136023*^9, 
  3.733368130817367*^9, 3.7922738536007137`*^9, 3.794517940805188*^9, 
  3.794518019130911*^9, 3.7945182058988028`*^9, 3.794518784166045*^9, 
  3.79451906419447*^9, 3.7945192996608057`*^9, 3.794520025018033*^9, 
  3.7945202114183073`*^9, 3.7945203193973637`*^9, 3.7945204208715477`*^9, 
  3.794520590794734*^9, 3.7945206664505177`*^9},
 CellLabel->"Out[116]=",
 CellID->834040792,ExpressionUUID->"44157af1-03b3-4d7e-9abd-741759412272"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Sometimes it may happen that the tensors have the same indices, but in a \
different order. Then we reorder one of them:\
\>", "MathCaption",
 CellChangeTimes->{3.728675986047452*^9},
 CellID->319048824,ExpressionUUID->"856f49f9-7ed7-4f84-976f-af98309fb362"],

Cell[BoxData[
 RowBox[{
  RowBox[{"T", "=", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"1", ",", "2"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"3", ",", "4"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"B", ",", "B"}], "}"}]}], "]"}]}], ";"}]], "Input",
 CellLabel->"In[117]:=",
 CellID->2134402056,ExpressionUUID->"cc4c4da4-4e4e-4607-8939-e5b3125edf51"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"T", "[", 
   RowBox[{"a", ",", "b"}], "]"}], "+", 
  RowBox[{"T", "[", 
   RowBox[{"b", ",", "a"}], "]"}]}]], "Input",
 CellLabel->"In[118]:=",
 CellID->1328515704,ExpressionUUID->"9c98aa42-7e0c-42bc-b48b-0fc433a8ebe2"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["2",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "2"}]],
           Annotation[#, {1, 1} -> 2, "Tooltip"]& ], 
          TagBox[
           TooltipBox["5",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "5"}]],
           Annotation[#, {1, 2} -> 5, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["5",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "5"}]],
           Annotation[#, {2, 1} -> 5, "Tooltip"]& ], 
          TagBox[
           TooltipBox["8",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", "8"}]],
           Annotation[#, {2, 2} -> 8, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a", "b"},
         {" ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{2, 5}, {5, 8}}, {$CellContext`B, $CellContext`B}, 
   0][$CellContext`a, $CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.72867598770222*^9, 3.7287534994859257`*^9, 
  3.72884352715421*^9, 3.7309556233134604`*^9, 3.7309562010584097`*^9, 
  3.731116150920518*^9, 3.73336768510847*^9, 3.733367898304421*^9, 
  3.7333681310494204`*^9, 3.792273853670026*^9, 3.794517940884573*^9, 
  3.794518019193782*^9, 3.7945182059653063`*^9, 3.794518784228773*^9, 
  3.794519064257337*^9, 3.7945192997284317`*^9, 3.7945200250868883`*^9, 
  3.7945202114845*^9, 3.794520319468684*^9, 3.794520420947167*^9, 
  3.794520590879632*^9, 3.794520666557678*^9},
 CellLabel->"Out[118]=",
 CellID->1097827175,ExpressionUUID->"cdfbed00-ef03-4c52-bd56-b85f021c8358"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"T", "[", 
   RowBox[{"a", ",", "b"}], "]"}], "-", 
  RowBox[{"T", "[", 
   RowBox[{"b", ",", "a"}], "]"}]}]], "Input",
 CellLabel->"In[119]:=",
 CellID->467338297,ExpressionUUID->"db0cbad1-d546-4609-8f2c-d70513d0639d"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{"-", "1"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{"-", "1"}]}]],
           Annotation[#, {1, 2} -> -1, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {2, 1} -> 1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 2} -> 0, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a", "b"},
         {" ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{0, -1}, {1, 0}}, {$CellContext`B, $CellContext`B}, 
   0][$CellContext`a, $CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.7286759883010674`*^9, 3.728753499626791*^9, 
  3.728843527589587*^9, 3.7309556234250298`*^9, 3.730956201167386*^9, 
  3.731116152154662*^9, 3.73336768534634*^9, 3.7333678984206057`*^9, 
  3.733368131183099*^9, 3.792273853722719*^9, 3.794517940934044*^9, 
  3.794518019237609*^9, 3.7945182060116158`*^9, 3.794518784272498*^9, 
  3.7945190643009863`*^9, 3.7945192997765417`*^9, 3.794520025136056*^9, 
  3.794520211530758*^9, 3.794520319524233*^9, 3.794520421037683*^9, 
  3.7945205909297447`*^9, 3.794520666616962*^9},
 CellLabel->"Out[119]=",
 CellID->1835201,ExpressionUUID->"1b0fb8be-8f4e-4ffe-82b1-bc96f0f235c9"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
However, we do not move the indices by default, because that may interfere \
with the canonicalizer. The function ToCCanonical is needed then, to sort the \
indices as determined by IndexSort:\
\>", "MathCaption",
 CellChangeTimes->{3.728675989895308*^9},
 CellID->1936974638,ExpressionUUID->"5e814303-ccc2-4f04-8c2d-d94b91d27f6f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"T", "[", 
  RowBox[{"b", ",", "a"}], "]"}]], "Input",
 CellLabel->"In[120]:=",
 CellID->1028135993,ExpressionUUID->"514d53ae-284a-4a9b-b12c-bbc96993dd56"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {1, 1} -> 1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["2",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "2"}]],
           Annotation[#, {1, 2} -> 2, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["3",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "3"}]],
           Annotation[#, {2, 1} -> 3, "Tooltip"]& ], 
          TagBox[
           TooltipBox["4",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", "4"}]],
           Annotation[#, {2, 2} -> 4, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"b", "a"},
         {" ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{1, 2}, {3, 4}}, {$CellContext`B, $CellContext`B}, 
   0][$CellContext`b, $CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.7286759911305723`*^9, 3.728753499765074*^9, 
  3.7288435279553757`*^9, 3.730955623541301*^9, 3.730956201256146*^9, 
  3.7311161622078447`*^9, 3.733367685665058*^9, 3.733367898556796*^9, 
  3.733368131317539*^9, 3.792273853772128*^9, 3.794517940984375*^9, 
  3.794518019284738*^9, 3.794518206062016*^9, 3.794518784322707*^9, 
  3.7945190643428*^9, 3.7945192998225946`*^9, 3.794520025192115*^9, 
  3.79452021157905*^9, 3.794520319572259*^9, 3.794520421095263*^9, 
  3.794520590979821*^9, 3.7945206666641006`*^9},
 CellLabel->"Out[120]=",
 CellID->779452149,ExpressionUUID->"2095b021-29ca-4ebc-845a-d03bbdb1f9d8"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ToCCanonical"}]], "Input",
 CellLabel->"In[121]:=",
 CellID->922554359,ExpressionUUID->"b4baffb7-bfc5-414a-ba66-31b3d52c0a30"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {1, 1} -> 1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["3",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "3"}]],
           Annotation[#, {1, 2} -> 3, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["2",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "2"}]],
           Annotation[#, {2, 1} -> 2, "Tooltip"]& ], 
          TagBox[
           TooltipBox["4",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", "4"}]],
           Annotation[#, {2, 2} -> 4, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a", "b"},
         {" ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{1, 3}, {2, 4}}, {$CellContext`B, $CellContext`B}, 
   0][$CellContext`a, $CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.728675991713004*^9, 3.728753499897304*^9, 
  3.728843528168988*^9, 3.730955623657976*^9, 3.730956201368579*^9, 
  3.7311161646075153`*^9, 3.7333676858250103`*^9, 3.733367898669093*^9, 
  3.73336813144995*^9, 3.7922738538241367`*^9, 3.794517941032304*^9, 
  3.794518019325553*^9, 3.794518206111884*^9, 3.7945187843656673`*^9, 
  3.7945190643863993`*^9, 3.794519299868804*^9, 3.794520025239882*^9, 
  3.7945202116253433`*^9, 3.794520319629284*^9, 3.794520421148387*^9, 
  3.7945205910274057`*^9, 3.7945206667115707`*^9},
 CellLabel->"Out[121]=",
 CellID->602021543,ExpressionUUID->"ff2ff706-e373-42a1-a2bc-1ee6749f40f5"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Note that ToCanonical is a different operation, based on the existence of \
symmetry of the tensor. In this case there is no symmetry and hence \
ToCanonical does nothing:\
\>", "MathCaption",
 CellID->1771022318,ExpressionUUID->"0d0c51fc-a729-4c3f-8fd1-05d0aca8ed82"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%%", "//", "ToCanonical"}]], "Input",
 CellLabel->"In[122]:=",
 CellID->1448751206,ExpressionUUID->"39be0805-dff8-4f04-b229-c9df11bc6a00"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {1, 1} -> 1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["2",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "2"}]],
           Annotation[#, {1, 2} -> 2, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["3",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "3"}]],
           Annotation[#, {2, 1} -> 3, "Tooltip"]& ], 
          TagBox[
           TooltipBox["4",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", "4"}]],
           Annotation[#, {2, 2} -> 4, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"b", "a"},
         {" ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{1, 2}, {3, 4}}, {$CellContext`B, $CellContext`B}, 
   0][$CellContext`b, $CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.728675993581189*^9, 3.7287535000099697`*^9, 
  3.728843528368881*^9, 3.730955623774683*^9, 3.7309562014565153`*^9, 
  3.731116173662319*^9, 3.733367685919829*^9, 3.733367898787464*^9, 
  3.733368131583255*^9, 3.792273853875813*^9, 3.794517941088127*^9, 
  3.794518019374187*^9, 3.794518206160577*^9, 3.794518784405905*^9, 
  3.794519064426838*^9, 3.794519299911779*^9, 3.794520025290473*^9, 
  3.7945202116749363`*^9, 3.79452031967651*^9, 3.7945204212026157`*^9, 
  3.794520591103528*^9, 3.794520666754406*^9},
 CellLabel->"Out[122]=",
 CellID->1380612184,ExpressionUUID->"32436825-b64a-4174-a3cb-2bc779d78908"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Times multiplication of two CTensor objects triggers their immediate tensor \
product:\
\>", "MathCaption",
 CellChangeTimes->{3.728675995263026*^9},
 CellID->444496031,ExpressionUUID->"13e419a8-fb49-4460-b205-35f95dcc24da"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"v", "[", "a", "]"}], 
  RowBox[{"v", "[", "b", "]"}]}]], "Input",
 CellLabel->"In[123]:=",
 CellID->1696668751,ExpressionUUID->"47eb79c3-ed35-403c-a239-0fa53d758fd5"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{"-", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["s",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`s[],
               Editable->False], "2"]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{"-", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                  "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`s[], Editable -> False], "2"]}]}]],
           Annotation[#, {2, 2} -> -$CellContext`s[]^2, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{"\[ImaginaryI]", " ", 
             InterpretationBox[
              StyleBox["s",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`s[],
              Editable->False]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", 
              RowBox[{"\[ImaginaryI]", " ", 
                InterpretationBox[
                 StyleBox["s", AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`s[], Editable -> False]}]}]],
           
           Annotation[#, {2, 3} -> Complex[0, 1] $CellContext`s[], 
            "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{"3", " ", "\[ImaginaryI]", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["s",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`s[],
               Editable->False], "3"]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "4"}], "}"}], "\[Rule]", 
              RowBox[{"3", " ", "\[ImaginaryI]", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                  "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`s[], Editable -> False], "3"]}]}]],
           
           Annotation[#, {2, 4} -> Complex[0, 3] $CellContext`s[]^3, 
            "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{"\[ImaginaryI]", " ", 
             InterpretationBox[
              StyleBox["s",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`s[],
              Editable->False]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{"\[ImaginaryI]", " ", 
                InterpretationBox[
                 StyleBox["s", AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`s[], Editable -> False]}]}]],
           
           Annotation[#, {3, 2} -> Complex[0, 1] $CellContext`s[], 
            "Tooltip"]& ], 
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {3, 3} -> 1, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{"3", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["s",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`s[],
               Editable->False], "2"]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "4"}], "}"}], "\[Rule]", 
              RowBox[{"3", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                  "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`s[], Editable -> False], "2"]}]}]],
           Annotation[#, {3, 4} -> 3 $CellContext`s[]^2, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{"3", " ", "\[ImaginaryI]", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["s",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`s[],
               Editable->False], "3"]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{"3", " ", "\[ImaginaryI]", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                  "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`s[], Editable -> False], "3"]}]}]],
           
           Annotation[#, {4, 2} -> Complex[0, 3] $CellContext`s[]^3, 
            "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{"3", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["s",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`s[],
               Editable->False], "2"]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "3"}], "}"}], "\[Rule]", 
              RowBox[{"3", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                  "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`s[], Editable -> False], "2"]}]}]],
           Annotation[#, {4, 3} -> 3 $CellContext`s[]^2, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{"9", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["s",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`s[],
               Editable->False], "4"]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "4"}], "}"}], "\[Rule]", 
              RowBox[{"9", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                  "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`s[], Editable -> False], "4"]}]}]],
           Annotation[#, {4, 4} -> 9 $CellContext`s[]^4, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a", "b"},
         {" ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{0, 0, 0, 0}, {
    0, -$CellContext`s[]^2, Complex[0, 1] $CellContext`s[], 
     Complex[0, 3] $CellContext`s[]^3}, {
    0, Complex[0, 1] $CellContext`s[], 1, 3 $CellContext`s[]^2}, {
    0, Complex[0, 3] $CellContext`s[]^3, 3 $CellContext`s[]^2, 
     9 $CellContext`s[]^4}}, {$CellContext`B, $CellContext`B}, 
   0][$CellContext`a, $CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.728675996445676*^9, 3.728753500148738*^9, 
  3.72884352856975*^9, 3.730955623905298*^9, 3.730956201579135*^9, 
  3.731116189922639*^9, 3.7333676860225077`*^9, 3.733367898880072*^9, 
  3.733368131717576*^9, 3.792273853931795*^9, 3.794517941138159*^9, 
  3.7945180194188213`*^9, 3.7945182062173862`*^9, 3.794518784454373*^9, 
  3.7945190644734783`*^9, 3.7945192999733887`*^9, 3.79452002533992*^9, 
  3.794520211738329*^9, 3.7945203197302856`*^9, 3.794520421268508*^9, 
  3.794520591157618*^9, 3.7945206668077173`*^9},
 CellLabel->"Out[123]=",
 CellID->388170653,ExpressionUUID->"4aa24cd7-e65e-40ac-b62c-bbd46aa1b6a6"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"v", "[", "a", "]"}], 
  RowBox[{"T", "[", 
   RowBox[{"b", ",", "c"}], "]"}]}]], "Input",
 CellLabel->"In[124]:=",
 CellID->1542255061,ExpressionUUID->"7e9dafe1-ea5f-48de-a84e-432cd97842f9"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{"\[ImaginaryI]", " ", 
                InterpretationBox[
                 StyleBox["s",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`s[],
                 Editable->False]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{"\[ImaginaryI]", " ", 
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}]}]],
              
              Annotation[#, {1, 1, 2} -> Complex[0, 1] $CellContext`s[], 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["1",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "3"}], "}"}], "\[Rule]", "1"}]],
              Annotation[#, {1, 1, 3} -> 1, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{"3", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["s",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`s[],
                  Editable->False], "2"]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "4"}], "}"}], "\[Rule]", 
                 RowBox[{"3", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "2"]}]}]],
              Annotation[#, {1, 1, 4} -> 3 $CellContext`s[]^2, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{"2", " ", "\[ImaginaryI]", " ", 
                InterpretationBox[
                 StyleBox["s",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`s[],
                 Editable->False]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{"2", " ", "\[ImaginaryI]", " ", 
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}]}]],
              
              Annotation[#, {1, 2, 2} -> Complex[0, 2] $CellContext`s[], 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["2",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "3"}], "}"}], "\[Rule]", "2"}]],
              Annotation[#, {1, 2, 3} -> 2, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{"6", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["s",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`s[],
                  Editable->False], "2"]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "4"}], "}"}], "\[Rule]", 
                 RowBox[{"6", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "2"]}]}]],
              Annotation[#, {1, 2, 4} -> 6 $CellContext`s[]^2, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{"3", " ", "\[ImaginaryI]", " ", 
                InterpretationBox[
                 StyleBox["s",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`s[],
                 Editable->False]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{"3", " ", "\[ImaginaryI]", " ", 
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}]}]],
              Annotation[#, {2, 1, 2} -> Complex[0, 3] $CellContext`s[], 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["3",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "3"}], "}"}], "\[Rule]", "3"}]],
              Annotation[#, {2, 1, 3} -> 3, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{"9", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["s",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`s[],
                  Editable->False], "2"]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "4"}], "}"}], "\[Rule]", 
                 RowBox[{"9", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "2"]}]}]],
              Annotation[#, {2, 1, 4} -> 9 $CellContext`s[]^2, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{"4", " ", "\[ImaginaryI]", " ", 
                InterpretationBox[
                 StyleBox["s",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`s[],
                 Editable->False]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{"4", " ", "\[ImaginaryI]", " ", 
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}]}]],
              
              Annotation[#, {2, 2, 2} -> Complex[0, 4] $CellContext`s[], 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["4",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "3"}], "}"}], "\[Rule]", "4"}]],
              Annotation[#, {2, 2, 3} -> 4, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{"12", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["s",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`s[],
                  Editable->False], "2"]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "4"}], "}"}], "\[Rule]", 
                 RowBox[{"12", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "2"]}]}]],
              
              Annotation[#, {2, 2, 4} -> 12 $CellContext`s[]^2, 
               "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"b", "c", "a"},
         {" ", " ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{{
     0, Complex[0, 1] $CellContext`s[], 1, 3 $CellContext`s[]^2}, {
     0, Complex[0, 2] $CellContext`s[], 2, 6 $CellContext`s[]^2}}, {{
     0, Complex[0, 3] $CellContext`s[], 3, 9 $CellContext`s[]^2}, {
     0, Complex[0, 4] $CellContext`s[], 4, 
      12 $CellContext`s[]^2}}}, {$CellContext`B, $CellContext`B, \
$CellContext`B}, 0][$CellContext`b, $CellContext`c, $CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.728675997609291*^9, 3.728753500284565*^9, 
  3.728843528748417*^9, 3.730955624001012*^9, 3.7309562016865597`*^9, 
  3.731116191019553*^9, 3.73336768628106*^9, 3.733367899002871*^9, 
  3.733368131850626*^9, 3.792273853985403*^9, 3.79451794119515*^9, 
  3.794518019461863*^9, 3.794518206273571*^9, 3.7945187845035973`*^9, 
  3.7945190645152903`*^9, 3.7945193000249863`*^9, 3.7945200253970118`*^9, 
  3.794520211789645*^9, 3.794520319780246*^9, 3.794520421322847*^9, 
  3.794520591208034*^9, 3.7945206668580647`*^9},
 CellLabel->"Out[124]=",
 CellID->2076329032,ExpressionUUID->"54b5863f-ff8c-491d-9047-73b560deaa7e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Finally, contractions are denoted by repeated staggered indices:\
\>", "MathCaption",
 CellChangeTimes->{3.728675999318771*^9},
 CellID->1023370296,ExpressionUUID->"d2060bd4-f144-4a07-9bec-63da079bf586"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"A", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}]}], "]"}], 
  RowBox[{"v", "[", "b", "]"}]}]], "Input",
 CellLabel->"In[125]:=",
 CellID->588036049,ExpressionUUID->"d1846d1d-9375-4cb3-b949-4d3d77dda139"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      OverscriptBox[
       OverscriptBox[
        StyleBox[GridBox[{
           {
            TagBox[
             TooltipBox["0",
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
             Annotation[#, {1, 1} -> 0, "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              RowBox[{
               RowBox[{"-", 
                InterpretationBox[
                 StyleBox["s",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`s[],
                 Editable->False]}], "-", 
               RowBox[{"3", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["s",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`s[],
                  Editable->False], "3"]}]}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                RowBox[{
                  RowBox[{"-", 
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}], "-", 
                  RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "3"]}]}]}]],
             
             Annotation[#, {1, 2} -> -$CellContext`s[] - 3 $CellContext`s[]^3,
               "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              RowBox[{
               RowBox[{"\[ImaginaryI]", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["s",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`s[],
                  Editable->False], "2"]}], "-", 
               RowBox[{"3", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["s",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`s[],
                  Editable->False], "3"]}]}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", 
                RowBox[{
                  RowBox[{"\[ImaginaryI]", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "2"]}], "-", 
                  RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "3"]}]}]}]],
             
             Annotation[#, {1, 3} -> 
              Complex[0, 1] $CellContext`s[]^2 - 3 $CellContext`s[]^3, 
              "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["s",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`s[],
                Editable->False], "+", 
               RowBox[{"\[ImaginaryI]", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["s",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`s[],
                  Editable->False], "2"]}]}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "4"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`s[], Editable -> False], "+", 
                  RowBox[{"\[ImaginaryI]", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "2"]}]}]}]],
             
             Annotation[#, {1, 4} -> $CellContext`s[] + 
               Complex[0, 1] $CellContext`s[]^2, "Tooltip"]& ]},
           {
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["s",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`s[],
                Editable->False], "+", 
               RowBox[{"3", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["s",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`s[],
                  Editable->False], "3"]}]}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`s[], Editable -> False], "+", 
                  RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "3"]}]}]}]],
             Annotation[#, {2, 1} -> $CellContext`s[] + 3 $CellContext`s[]^3, 
              "Tooltip"]& ], 
            TagBox[
             TooltipBox["0",
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", "0"}]],
             Annotation[#, {2, 2} -> 0, "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              RowBox[{
               RowBox[{"-", "3"}], " ", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["s",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`s[],
                 Editable->False], "3"]}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", 
                RowBox[{
                  RowBox[{"-", "3"}], " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "3"]}]}]],
             Annotation[#, {2, 3} -> (-3) $CellContext`s[]^3, "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              InterpretationBox[
               StyleBox["s",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`s[],
               Editable->False],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "4"}], "}"}], "\[Rule]", 
                InterpretationBox[
                 StyleBox["s", AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`s[], Editable -> False]}]],
             Annotation[#, {2, 4} -> $CellContext`s[], "Tooltip"]& ]},
           {
            TagBox[
             TooltipBox[
              RowBox[{
               RowBox[{
                RowBox[{"-", "\[ImaginaryI]"}], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["s",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`s[],
                  Editable->False], "2"]}], "+", 
               RowBox[{"3", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["s",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`s[],
                  Editable->False], "3"]}]}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", 
                RowBox[{
                  RowBox[{
                    RowBox[{"-", "\[ImaginaryI]"}], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "2"]}], "+", 
                  RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "3"]}]}]}]],
             
             Annotation[#, {3, 1} -> 
              Complex[0, -1] $CellContext`s[]^2 + 3 $CellContext`s[]^3, 
              "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              RowBox[{"3", " ", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["s",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`s[],
                 Editable->False], "3"]}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", 
                RowBox[{"3", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "3"]}]}]],
             Annotation[#, {3, 2} -> 3 $CellContext`s[]^3, "Tooltip"]& ], 
            TagBox[
             TooltipBox["0",
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", "0"}]],
             Annotation[#, {3, 3} -> 0, "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              RowBox[{
               RowBox[{"-", "\[ImaginaryI]"}], " ", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["s",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`s[],
                 Editable->False], "2"]}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"3", ",", "4"}], "}"}], "\[Rule]", 
                RowBox[{
                  RowBox[{"-", "\[ImaginaryI]"}], " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "2"]}]}]],
             
             Annotation[#, {3, 4} -> Complex[0, -1] $CellContext`s[]^2, 
              "Tooltip"]& ]},
           {
            TagBox[
             TooltipBox[
              RowBox[{
               RowBox[{"-", 
                InterpretationBox[
                 StyleBox["s",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`s[],
                 Editable->False]}], "-", 
               RowBox[{"\[ImaginaryI]", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["s",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`s[],
                  Editable->False], "2"]}]}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"4", ",", "1"}], "}"}], "\[Rule]", 
                RowBox[{
                  RowBox[{"-", 
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}], "-", 
                  RowBox[{"\[ImaginaryI]", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "2"]}]}]}]],
             
             Annotation[#, {4, 1} -> -$CellContext`s[] + 
               Complex[0, -1] $CellContext`s[]^2, "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              RowBox[{"-", 
               InterpretationBox[
                StyleBox["s",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`s[],
                Editable->False]}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"4", ",", "2"}], "}"}], "\[Rule]", 
                RowBox[{"-", 
                  InterpretationBox[
                   StyleBox[
                   "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`s[], Editable -> False]}]}]],
             Annotation[#, {4, 2} -> -$CellContext`s[], "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              RowBox[{"\[ImaginaryI]", " ", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["s",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`s[],
                 Editable->False], "2"]}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"4", ",", "3"}], "}"}], "\[Rule]", 
                RowBox[{"\[ImaginaryI]", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "2"]}]}]],
             
             Annotation[#, {4, 3} -> Complex[0, 1] $CellContext`s[]^2, 
              "Tooltip"]& ], 
            TagBox[
             TooltipBox["0",
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"4", ",", "4"}], "}"}], "\[Rule]", "0"}]],
             Annotation[#, {4, 4} -> 0, "Tooltip"]& ]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}],
         FontSize->Small], 
        StyleBox["~",
         FontColor->RGBColor[1., 0., 0.]]], 
       StyleBox["~",
        FontColor->RGBColor[1., 0., 0.]]], 
      StyleBox[GridBox[{
         {" ", " "},
         {"a", "c"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{
    0, -$CellContext`s[] - 3 $CellContext`s[]^3, 
     Complex[0, 1] $CellContext`s[]^2 - 
     3 $CellContext`s[]^3, $CellContext`s[] + 
     Complex[0, 1] $CellContext`s[]^2}, {$CellContext`s[] + 
     3 $CellContext`s[]^3, 0, (-3) $CellContext`s[]^3, 
     $CellContext`s[]}, {
    Complex[0, -1] $CellContext`s[]^2 + 3 $CellContext`s[]^3, 
     3 $CellContext`s[]^3, 0, 
     Complex[0, -1] $CellContext`s[]^2}, {-$CellContext`s[] + 
     Complex[0, -1] $CellContext`s[]^2, -$CellContext`s[], 
     Complex[0, 1] $CellContext`s[]^2, 
     0}}, {-$CellContext`B, -$CellContext`B}, 
   2 $CellContext`B][-$CellContext`a, -$CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{3.7286760007982683`*^9, 3.728753500391059*^9, 
  3.7288435291408253`*^9, 3.7309556241339617`*^9, 3.73095620177635*^9, 
  3.73111619495582*^9, 3.733367686490965*^9, 3.733367899119672*^9, 
  3.733368131982992*^9, 3.792273854038821*^9, 3.79451794125222*^9, 
  3.794518019511633*^9, 3.794518206318368*^9, 3.79451878455515*^9, 
  3.794519064582415*^9, 3.794519300075903*^9, 3.794520025451942*^9, 
  3.7945202118412037`*^9, 3.794520319833453*^9, 3.79452042138138*^9, 
  3.794520591274797*^9, 3.7945206669123583`*^9},
 CellLabel->"Out[125]=",
 CellID->252777074,ExpressionUUID->"79ff31f9-cd4c-422c-993c-44e55eab4f5d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", " ", 
  RowBox[{"v", "[", "c", "]"}]}]], "Input",
 CellLabel->"In[126]:=",
 CellID->502310434,ExpressionUUID->"b1a413fc-45f3-4be8-b8ed-7fdedfdfab87"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      OverscriptBox[
       OverscriptBox[
        StyleBox[GridBox[{
           {
            TagBox[
             TooltipBox["\<\"\[GrayCircle]\"\>",
              RowBox[{
                RowBox[{"{", "1", "}"}], "\[Rule]", 
                RowBox[{
                  RowBox[{"\[ImaginaryI]", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "2"]}], "-", 
                  RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "3"]}], "+", 
                  RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "2"]}]}], ")"}]}], 
                  "+", 
                  RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False]}], "-", 
                    RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "s", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`s[], Editable -> False], "3"]}]}], 
                    ")"}]}]}]}]],
             
             Annotation[#, {1} -> 
              Complex[0, 1] $CellContext`s[]^2 - 3 $CellContext`s[]^3 + 
               3 $CellContext`s[]^2 ($CellContext`s[] + 
                 Complex[0, 1] $CellContext`s[]^2) + 
               Complex[0, 1] $CellContext`s[] (-$CellContext`s[] - 
                 3 $CellContext`s[]^3), "Tooltip"]& ]},
           {
            TagBox[
             TooltipBox["0",
              RowBox[{
                RowBox[{"{", "2", "}"}], "\[Rule]", "0"}]],
             Annotation[#, {2} -> 0, "Tooltip"]& ]},
           {
            TagBox[
             TooltipBox["0",
              RowBox[{
                RowBox[{"{", "3", "}"}], "\[Rule]", "0"}]],
             Annotation[#, {3} -> 0, "Tooltip"]& ]},
           {
            TagBox[
             TooltipBox["0",
              RowBox[{
                RowBox[{"{", "4", "}"}], "\[Rule]", "0"}]],
             Annotation[#, {4} -> 0, "Tooltip"]& ]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
         FontSize->Small], 
        StyleBox["~",
         FontColor->RGBColor[1., 0., 0.]]], 
       StyleBox["~",
        FontColor->RGBColor[1., 0., 0.]]], 
      StyleBox[GridBox[{
         {" "},
         {"a"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{
   Complex[0, 1] $CellContext`s[]^2 - 3 $CellContext`s[]^3 + 
    3 $CellContext`s[]^2 ($CellContext`s[] + 
      Complex[0, 1] $CellContext`s[]^2) + 
    Complex[0, 1] $CellContext`s[] (-$CellContext`s[] - 3 $CellContext`s[]^3),
     0, 0, 0}, {-$CellContext`B}, 2 $CellContext`B][-$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.7286760019153337`*^9, 3.728753500590955*^9, 
  3.7288435295047073`*^9, 3.730955624270019*^9, 3.7309562018839817`*^9, 
  3.73111619647886*^9, 3.7333676866382017`*^9, 3.733367899213386*^9, 
  3.733368132149581*^9, 3.792273854090776*^9, 3.794517941312722*^9, 
  3.794518019557159*^9, 3.7945182063782787`*^9, 3.7945187846013117`*^9, 
  3.794519064635043*^9, 3.794519300126273*^9, 3.794520025501175*^9, 
  3.794520211892906*^9, 3.794520319885172*^9, 3.7945204214419003`*^9, 
  3.7945205913424177`*^9, 3.794520666962126*^9},
 CellLabel->"Out[126]=",
 CellID->701628952,ExpressionUUID->"fa8204dc-8b15-4029-b4ac-cb86e6b5acf7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "Simplify"}]], "Input",
 CellLabel->"In[127]:=",
 CellID->965495525,ExpressionUUID->"3a07f6b7-ab66-4dd5-9db9-f11756839b2a"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728676002888247*^9, 3.728753500783278*^9, 
  3.728843529675853*^9, 3.730955624377348*^9, 3.7309562019782467`*^9, 
  3.731116197289336*^9, 3.733367686831113*^9, 3.733367899337524*^9, 
  3.733368132284062*^9, 3.792273854138331*^9, 3.7945179413584642`*^9, 
  3.794518019599409*^9, 3.794518206418672*^9, 3.794518784640791*^9, 
  3.7945190646715393`*^9, 3.794519300170335*^9, 3.794520025542664*^9, 
  3.794520211936223*^9, 3.794520319936102*^9, 3.7945204214861603`*^9, 
  3.794520591437497*^9, 3.7945206670046673`*^9},
 CellLabel->"Out[127]=",
 CellID->445090587,ExpressionUUID->"2a5c7680-24e4-48d7-8a02-c8612204af07"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.2.4. Metric", "Subsubsection",
 CellChangeTimes->{{3.733368025759824*^9, 3.7333680301198*^9}},
 CellID->1762278910,ExpressionUUID->"ddfdf62f-df37-415e-bf5e-3d6da1d78d9b"],

Cell["\<\
Once we have a metric field in our problem, there are many more things we can \
do. In this section we focus on how to set a metric field, and which other \
tensors are algebraically related to it. We can specify a metric field in \
several possible ways. We start with the simpler case of a metric given in a \
coordinate chart. We shall use the example of spherical angles on the \
2-sphere.\
\>", "Text",
 CellID->527180946,ExpressionUUID->"c88d1560-a46c-4bf6-ad94-5a6e6caf4c56"],

Cell[CellGroupData[{

Cell["Define the S2 manifold, with script indices:", "MathCaption",
 CellChangeTimes->{{3.728676036741083*^9, 3.728676036965098*^9}},
 CellID->1960115547,ExpressionUUID->"af00e35e-8906-4692-891b-66e8e3de7516"],

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"S2", ",", "2", ",", 
   RowBox[{"{", 
    RowBox[{
    "\[ScriptA]", ",", "\[ScriptB]", ",", "\[ScriptC]", ",", "\[ScriptD]", 
     ",", "\[ScriptE]", ",", "\[ScriptF]"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[128]:=",
 CellID->1262853449,ExpressionUUID->"f1e42cfa-e54c-4782-83cf-597fbf8fb106"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Define the chart:", "MathCaption",
 CellChangeTimes->{{3.728676039988744*^9, 3.728676040125195*^9}},
 CellID->1197034042,ExpressionUUID->"8c8f6f41-1bab-4d0a-b7cb-21c4d6980418"],

Cell[BoxData[
 RowBox[{"DefChart", "[", 
  RowBox[{"sph", ",", "S2", ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", "2"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"\[Theta]", "[", "]"}], ",", 
     RowBox[{"\[Phi]", "[", "]"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[129]:=",
 CellID->924338798,ExpressionUUID->"219344eb-131f-428f-9974-45327618e521"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Construct a metric field, that is a symmetric covariant 2-tensor:\
\>", "MathCaption",
 CellChangeTimes->{3.728676043028781*^9},
 CellID->465475588,ExpressionUUID->"913598fa-e4bd-4af8-9a30-676d1707c5e9"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"met", "=", 
  RowBox[{"CTensor", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"1", ",", "0"}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{
         RowBox[{"Sin", "[", 
          RowBox[{"\[Theta]", "[", "]"}], "]"}], "^", "2"}]}], "}"}]}], "}"}],
     ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "sph"}], ",", 
      RowBox[{"-", "sph"}]}], "}"}]}], "]"}]}]], "Input",
 CellLabel->"In[130]:=",
 CellID->2112765142,ExpressionUUID->"4c16c335-9138-4d67-8464-a0e6268bbdcf"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", 
       SuperscriptBox[
        RowBox[{
         InterpretationBox[
          StyleBox["Sin",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Sin,
          Editable->False], "[", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False], "]"}], "2"]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "sph"}], ",", 
     RowBox[{"-", "sph"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.728676044164261*^9, 3.728753501015791*^9, 
  3.728843537021455*^9, 3.730955640718032*^9, 3.730956202216313*^9, 
  3.731116268799754*^9, 3.733367691426537*^9, 3.733367899603252*^9, 
  3.733368132600327*^9, 3.7922738542670307`*^9, 3.7945179414838533`*^9, 
  3.79451801971314*^9, 3.794518231844029*^9, 3.7945187847562027`*^9, 
  3.794519064781056*^9, 3.79451930027597*^9, 3.794520025662698*^9, 
  3.794520212079557*^9, 3.794520320051586*^9, 3.794520421609456*^9, 
  3.794520591596633*^9, 3.79452066711339*^9},
 CellLabel->"Out[130]=",
 CellID->951953208,ExpressionUUID->"1b1e8275-3f20-4c2d-b706-abe1ec0590e8"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"met", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.731116276184556*^9, 3.731116278189012*^9}},
 CellLabel->"In[131]:=",
 CellID->806148330,ExpressionUUID->"ec2fc400-1233-4208-851f-30cbd3887689"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {1, 1} -> 1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 2} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["Sin",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Sin,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], "2"],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              SuperscriptBox[
               RowBox[{
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], Sin, Editable -> False], "[", 
                 InterpretationBox[
                  StyleBox[
                  "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                   False], 
                  $CellContext`\[Theta][], Editable -> False], "]"}], "2"]}]],
           
           Annotation[#, {2, 2} -> Sin[
               $CellContext`\[Theta][]]^2, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " "},
         {"a", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{1, 0}, {0, Sin[
       $CellContext`\[Theta][]]^2}}, {-$CellContext`sph, -$CellContext`sph}, 
   0][-$CellContext`a, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.73111627864228*^9, 3.733367691541107*^9, 
  3.7333678997038193`*^9, 3.7333681327334337`*^9, 3.792273854320054*^9, 
  3.7945179415379143`*^9, 3.794518019754678*^9, 3.794518231897315*^9, 
  3.794518784805181*^9, 3.7945190648310423`*^9, 3.794519300331036*^9, 
  3.794520025717127*^9, 3.794520212131483*^9, 3.7945203201037207`*^9, 
  3.7945204216639853`*^9, 3.794520591648546*^9, 3.794520667165765*^9},
 CellLabel->"Out[131]=",
 CellID->724396706,ExpressionUUID->"aca912ae-d165-430c-9b66-d6d84c8e9cd1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We have not specified that this is a metric, and there is actually no metric \
yet in TangentS2. Therefore this object is not yet defined:\
\>", "MathCaption",
 CellChangeTimes->{{3.731116282204754*^9, 3.731116331419374*^9}},
 CellID->2057601857,ExpressionUUID->"4adec57b-2706-46a9-abbd-f5dfb895a753"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"met", "[", 
   RowBox[{"a", ",", "b"}], "]"}], "//", "Catch"}]], "Input",
 CellChangeTimes->{{3.7311163328470984`*^9, 3.731116341250045*^9}},
 CellLabel->"In[132]:=",
 CellID->256560628,ExpressionUUID->"1e891def-d31f-4fee-be3a-0b553397c8d2"],

Cell[BoxData[
 TemplateBox[{
  "MetricsOfVBundle", "missing", 
   "\"There is no \\!\\(\\*RowBox[{\\\"\\\\\\\"metric\\\\\\\"\\\"}]\\) in \\!\
\\(\\*InterpretationBox[StyleBox[\\\"\[DoubleStruckCapitalT]S2\\\", \
Rule[AutoSpacing, False], Rule[ShowAutoStyles, False]], TangentS2, \
Rule[Editable, False]]\\).\"", 2, 132, 6, 19683117308123869115, "Local", 
   "xAct`xTensor`MetricsOfVBundle"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{{3.731116335567094*^9, 3.731116346039044*^9}, 
   3.733367691659993*^9, 3.733367899805286*^9, 3.733368132866482*^9, 
   3.7922738543662367`*^9, 3.794517941596458*^9, 3.794518019789688*^9, 
   3.7945182319499607`*^9, 3.794518784844122*^9, 3.794519064869279*^9, 
   3.7945193003734703`*^9, 3.794520025764144*^9, 3.794520212172657*^9, 
   3.7945203201514263`*^9, 3.794520421709944*^9, 3.7945205916908207`*^9, 
   3.794520667209309*^9},
 CellLabel->"During evaluation of In[132]:=",
 CellID->850642899,ExpressionUUID->"03e6c921-729d-4ed0-8d61-64ea36dc9b4f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Compute the inverse metric tensor with the Inv operator:", "MathCaption",
 CellChangeTimes->{
  3.7286760458043118`*^9, {3.731116356138235*^9, 3.731116364978272*^9}},
 CellID->1384402267,ExpressionUUID->"c83f6708-25ad-450f-8949-481ebbc1a280"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"imet", "=", 
  RowBox[{"Inv", "[", "met", "]"}]}]], "Input",
 CellLabel->"In[133]:=",
 CellID->1811069871,ExpressionUUID->"91d1ed1e-c38d-43b2-b87e-ade1dbb844d9"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", 
       SuperscriptBox[
        RowBox[{
         InterpretationBox[
          StyleBox["Csc",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Csc,
          Editable->False], "[", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False], "]"}], "2"]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"sph", ",", "sph"}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.728676046926742*^9, 3.7287535011382113`*^9, 
  3.7288435372382517`*^9, 3.730955641764958*^9, 3.730956202311513*^9, 
  3.731116367133135*^9, 3.7333676918802557`*^9, 3.73336789990238*^9, 
  3.7333681330010843`*^9, 3.792273854418632*^9, 3.794517941645451*^9, 
  3.7945180198444223`*^9, 3.7945182320063877`*^9, 3.794518784888112*^9, 
  3.7945190649102182`*^9, 3.794519300420781*^9, 3.794520025814056*^9, 
  3.794520212222931*^9, 3.7945203202018557`*^9, 3.794520421755866*^9, 
  3.794520591741722*^9, 3.794520667253887*^9},
 CellLabel->"Out[133]=",
 CellID->956565414,ExpressionUUID->"24d4afd4-7df0-48b0-b077-567a7c30f11d"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Check that they are indeed inverses:", "MathCaption",
 CellChangeTimes->{3.728676048468464*^9},
 CellID->8739004,ExpressionUUID->"e72fac13-eb6d-461f-97a7-dd76769b5b51"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"met", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}]}], "]"}], 
  RowBox[{"imet", "[", 
   RowBox[{"b", ",", "c"}], "]"}]}]], "Input",
 CellLabel->"In[134]:=",
 CellID->847552232,ExpressionUUID->"7f224a5a-7fb4-4cef-8cb4-47f872e52418"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {1, 1} -> 1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 2} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {2, 2} -> 1, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"c", " "},
         {" ", "a"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{1, 0}, {0, 1}}, {$CellContext`sph, -$CellContext`sph}, 
   0][$CellContext`c, -$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.7286760495636597`*^9, 3.728753501236162*^9, 
  3.728843537654858*^9, 3.730955643000383*^9, 3.730956202418186*^9, 
  3.731116370499518*^9, 3.7333676921930103`*^9, 3.733367900021085*^9, 
  3.733368133150084*^9, 3.792273854467676*^9, 3.794517941696413*^9, 
  3.794518019886047*^9, 3.7945182320541573`*^9, 3.794518784932646*^9, 
  3.79451906494993*^9, 3.7945193004686947`*^9, 3.794520025864814*^9, 
  3.794520212270587*^9, 3.794520320274291*^9, 3.7945204218032503`*^9, 
  3.794520591785541*^9, 3.794520667299449*^9},
 CellLabel->"Out[134]=",
 CellID->932331656,ExpressionUUID->"3da3b3e9-c0fa-4716-bcd1-87765f56292b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We can also compute the metric determinant, which is a density of weight +2 \
in the given chart (note the third argument of CTensor in the result):\
\>", "MathCaption",
 CellChangeTimes->{3.728676051596489*^9},
 CellID->2003781953,ExpressionUUID->"d3b79422-88fc-463c-ab21-8b578d578056"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Determinant", "[", 
  RowBox[{"met", ",", "sph"}], "]"}]], "Input",
 CellLabel->"In[135]:=",
 CellID->724731817,ExpressionUUID->"d3b497a4-b15f-47a7-991c-9d8d0cf1185a"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"], ",", 
   RowBox[{"{", "}"}], ",", 
   RowBox[{"2", " ", "sph"}]}], "]"}]], "Output",
 CellChangeTimes->{3.7286760528912*^9, 3.728753501352489*^9, 
  3.728843538095438*^9, 3.730955644284965*^9, 3.73095620251237*^9, 
  3.7311163808103247`*^9, 3.733367692425318*^9, 3.733367900119933*^9, 
  3.7333681332662373`*^9, 3.792273854518351*^9, 3.794517941748439*^9, 
  3.794518019924467*^9, 3.794518232102759*^9, 3.7945187849706793`*^9, 
  3.794519064989562*^9, 3.7945193005098143`*^9, 3.7945200259447527`*^9, 
  3.794520212314722*^9, 3.794520320329056*^9, 3.794520421857524*^9, 
  3.794520591828363*^9, 3.794520667348497*^9},
 CellLabel->"Out[135]=",
 CellID->1657453880,ExpressionUUID->"65aa2173-fd27-4ced-8fbe-567372aa234e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Or the epsilon tensor, which is a pure tensor, and not a density:\
\>", "MathCaption",
 CellChangeTimes->{3.728676054371952*^9},
 CellID->440573293,ExpressionUUID->"527978eb-cb87-4715-848e-211a12fa3129"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"epsilon", "[", "met", "]"}]], "Input",
 CellLabel->"In[136]:=",
 CellID->191720595,ExpressionUUID->"9737edba-9e4d-4cde-b355-4a2b3665cde3"],

Cell[BoxData[
 TemplateBox[{
  "SignDetOfMetric", "num", 
   "\"Cannot determine sign of determinant of matrix \\!\\(\\*RowBox[{\\\"{\\\
\", RowBox[{RowBox[{\\\"{\\\", RowBox[{\\\"1\\\", \\\",\\\", \\\"0\\\"}], \
\\\"}\\\"}], \\\",\\\", RowBox[{\\\"{\\\", RowBox[{\\\"0\\\", \\\",\\\", \
SuperscriptBox[RowBox[{InterpretationBox[StyleBox[RowBox[{\\\"Sin\\\"}], \
Rule[AutoSpacing, False], Rule[ShowAutoStyles, False]], Sin, Rule[Editable, \
False]], \\\"[\\\", InterpretationBox[StyleBox[\\\"\[Theta]\\\", \
Rule[AutoSpacing, False], Rule[ShowAutoStyles, False]], \\\\[Theta][], \
Rule[Editable, False]], \\\"]\\\"}], \\\"2\\\"]}], \\\"}\\\"}]}], \\\"}\\\"}]\
\\). Assuming +1.\"", 2, 136, 7, 19683117308123869115, "Local", 
   "xAct`xTensor`SignDetOfMetric"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728676055658989*^9, 3.728753501463854*^9, 
  3.728843538518857*^9, 3.73095564537047*^9, 3.730956202611326*^9, 
  3.731116394468582*^9, 3.73336769257514*^9, 3.733367900210175*^9, 
  3.733368133396494*^9, 3.7922738545661507`*^9, 3.794517941800221*^9, 
  3.794518020088299*^9, 3.7945182321520777`*^9, 3.7945187850216913`*^9, 
  3.794519065028139*^9, 3.7945193005603848`*^9, 3.7945200259951468`*^9, 
  3.79452021236193*^9, 3.794520320398699*^9, 3.794520421903373*^9, 
  3.794520591875139*^9, 3.7945206674147177`*^9},
 CellLabel->"During evaluation of In[136]:=",
 CellID->1816975923,ExpressionUUID->"da9d62e4-d83b-49ce-a120-f020ed262a30"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   InterpretationBox[
    RowBox[{
     TagBox["SparseArray",
      "SummaryHead"], "[", 
     DynamicModuleBox[{Typeset`open$$ = False, Typeset`embedState$$ = 
      "Ready"}, 
      TemplateBox[{
        PaneSelectorBox[{False -> GridBox[{{
              PaneBox[
               ButtonBox[
                DynamicBox[
                 FEPrivate`FrontEndResource[
                 "FEBitmaps", "SquarePlusIconMedium"]], 
                ButtonFunction :> (Typeset`open$$ = True), Appearance -> None,
                 Evaluator -> Automatic, Method -> "Preemptive"], 
               Alignment -> {Center, Center}, ImageSize -> 
               Dynamic[{
                 Automatic, 3.5 CurrentValue["FontCapHeight"]/
                  AbsoluteCurrentValue[Magnification]}]], 
              GraphicsBox[
               
               RasterBox[{{{0.5, 0., 0.}, {0.235431, 0.32765, 0.833291}}, {{
                0.235431, 0.32765, 0.833291}, {0.5, 0., 0.}}}, {{0, 0}, {2, 
                2}}, {0, 1}], AspectRatio -> 1, Axes -> False, Frame -> False,
                FrameLabel -> {None, None}, FrameStyle -> Directive[
                 Opacity[0.5], 
                 Thickness[Tiny], 
                 RGBColor[0.368417, 0.506779, 0.709798]], 
               FrameTicks -> {{None, None}, {None, None}}, GridLinesStyle -> 
               Directive[
                 GrayLevel[0.5, 0.4]], ImageSize -> 
               Dynamic[{
                 Automatic, 3.5 CurrentValue["FontCapHeight"]/
                  AbsoluteCurrentValue[Magnification]}], 
               Method -> {
                "AxisPadding" -> Scaled[0.02], "DefaultBoundaryStyle" -> 
                 Automatic, 
                 "DefaultGraphicsInteraction" -> {
                  "Version" -> 1.2, "TrackMousePosition" -> {True, False}, 
                   "Effects" -> {
                    "Highlight" -> {"ratio" -> 2}, 
                    "HighlightPoint" -> {"ratio" -> 2}, 
                    "Droplines" -> {
                    "freeformCursorMode" -> True, 
                    "placement" -> {"x" -> "All", "y" -> "None"}}}}, 
                 "DefaultPlotStyle" -> Automatic, "DomainPadding" -> 
                 Scaled[0.02], "RangePadding" -> Scaled[0.05]}], 
              GridBox[{{
                 RowBox[{
                   TagBox[
                   "\"Specified elements: \"", "SummaryItemAnnotation"], 
                   "\[InvisibleSpace]", 
                   TagBox["2", "SummaryItem"]}]}, {
                 RowBox[{
                   TagBox["\"Dimensions: \"", "SummaryItemAnnotation"], 
                   "\[InvisibleSpace]", 
                   TagBox[
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2"}], "}"}], "SummaryItem"]}]}}, 
               GridBoxAlignment -> {
                "Columns" -> {{Left}}, "Rows" -> {{Automatic}}}, AutoDelete -> 
               False, GridBoxItemSize -> {
                "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
               GridBoxSpacings -> {
                "Columns" -> {{2}}, "Rows" -> {{Automatic}}}, 
               BaseStyle -> {
                ShowStringCharacters -> False, NumberMarks -> False, 
                 PrintPrecision -> 3, ShowSyntaxStyles -> False}]}}, 
            GridBoxAlignment -> {"Rows" -> {{Top}}}, AutoDelete -> False, 
            GridBoxItemSize -> {
             "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
            BaselinePosition -> {1, 1}], True -> GridBox[{{
              PaneBox[
               ButtonBox[
                DynamicBox[
                 FEPrivate`FrontEndResource[
                 "FEBitmaps", "SquareMinusIconMedium"]], 
                ButtonFunction :> (Typeset`open$$ = False), Appearance -> 
                None, Evaluator -> Automatic, Method -> "Preemptive"], 
               Alignment -> {Center, Center}, ImageSize -> 
               Dynamic[{
                 Automatic, 3.5 CurrentValue["FontCapHeight"]/
                  AbsoluteCurrentValue[Magnification]}]], 
              GraphicsBox[
               
               RasterBox[{{{0.5, 0., 0.}, {0.235431, 0.32765, 0.833291}}, {{
                0.235431, 0.32765, 0.833291}, {0.5, 0., 0.}}}, {{0, 0}, {2, 
                2}}, {0, 1}], AspectRatio -> 1, Axes -> False, Frame -> False,
                FrameLabel -> {None, None}, FrameStyle -> Directive[
                 Opacity[0.5], 
                 Thickness[Tiny], 
                 RGBColor[0.368417, 0.506779, 0.709798]], 
               FrameTicks -> {{None, None}, {None, None}}, GridLinesStyle -> 
               Directive[
                 GrayLevel[0.5, 0.4]], ImageSize -> 
               Dynamic[{
                 Automatic, 3.5 CurrentValue["FontCapHeight"]/
                  AbsoluteCurrentValue[Magnification]}], 
               Method -> {
                "AxisPadding" -> Scaled[0.02], "DefaultBoundaryStyle" -> 
                 Automatic, 
                 "DefaultGraphicsInteraction" -> {
                  "Version" -> 1.2, "TrackMousePosition" -> {True, False}, 
                   "Effects" -> {
                    "Highlight" -> {"ratio" -> 2}, 
                    "HighlightPoint" -> {"ratio" -> 2}, 
                    "Droplines" -> {
                    "freeformCursorMode" -> True, 
                    "placement" -> {"x" -> "All", "y" -> "None"}}}}, 
                 "DefaultPlotStyle" -> Automatic, "DomainPadding" -> 
                 Scaled[0.02], "RangePadding" -> Scaled[0.05]}], 
              GridBox[{{
                 RowBox[{
                   TagBox[
                   "\"Specified elements: \"", "SummaryItemAnnotation"], 
                   "\[InvisibleSpace]", 
                   TagBox["2", "SummaryItem"]}]}, {
                 RowBox[{
                   TagBox["\"Dimensions: \"", "SummaryItemAnnotation"], 
                   "\[InvisibleSpace]", 
                   TagBox[
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2"}], "}"}], "SummaryItem"]}]}, {
                 RowBox[{
                   TagBox["\"Default: \"", "SummaryItemAnnotation"], 
                   "\[InvisibleSpace]", 
                   TagBox["0", "SummaryItem"]}]}, {
                 RowBox[{
                   TagBox["\"Density: \"", "SummaryItemAnnotation"], 
                   "\[InvisibleSpace]", 
                   TagBox["0.5`", "SummaryItem"]}]}, {
                 RowBox[{
                   TagBox["\"Elements:\"", "SummaryItemAnnotation"], 
                   "\[InvisibleSpace]", 
                   TagBox["\"\"", "SummaryItem"]}]}, {
                 TagBox[
                  TagBox[
                   GridBox[{{
                    RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                    StyleBox[
                    PaneBox[
                    SqrtBox[
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], "2"]],
                     ContentPadding -> False, FrameMargins -> 0, StripOnInput -> 
                    True, BaselinePosition -> Baseline, 
                    ImageSize -> {{1, 300}, Automatic}], LineBreakWithin -> 
                    False]}]}, {
                    RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                    StyleBox[
                    PaneBox[
                    RowBox[{"-", 
                    SqrtBox[
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                    "2"]]}], ContentPadding -> False, FrameMargins -> 0, 
                    StripOnInput -> True, BaselinePosition -> Baseline, 
                    ImageSize -> {{1, 300}, Automatic}], LineBreakWithin -> 
                    False]}]}}, GridBoxAlignment -> {"Columns" -> {{Left}}}, 
                    DefaultBaseStyle -> "Column", 
                    GridBoxItemSize -> {
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], 
                   "Column"], "SummaryItem"]}}, 
               GridBoxAlignment -> {
                "Columns" -> {{Left}}, "Rows" -> {{Automatic}}}, AutoDelete -> 
               False, GridBoxItemSize -> {
                "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
               GridBoxSpacings -> {
                "Columns" -> {{2}}, "Rows" -> {{Automatic}}}, 
               BaseStyle -> {
                ShowStringCharacters -> False, NumberMarks -> False, 
                 PrintPrecision -> 3, ShowSyntaxStyles -> False}]}}, 
            GridBoxAlignment -> {"Rows" -> {{Top}}}, AutoDelete -> False, 
            GridBoxItemSize -> {
             "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
            BaselinePosition -> {1, 1}]}, 
         Dynamic[Typeset`open$$], ImageSize -> Automatic]},
       "SummaryPanel"],
      DynamicModuleValues:>{}], "]"}],
    SparseArray[
    Automatic, {2, 2}, 0, {
     1, {{0, 1, 2}, {{2}, {1}}}, {(Sin[$CellContext`\[Theta][]]^2)^
       Rational[1, 2], -(Sin[$CellContext`\[Theta][]]^2)^Rational[1, 2]}}],
    Editable->False,
    SelectWithContents->True,
    Selectable->False], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "sph"}], ",", 
     RowBox[{"-", "sph"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.728676055731612*^9, 3.728753501537191*^9, 
  3.728843538597211*^9, 3.730955645464108*^9, 3.730956202676413*^9, 
  3.7311163945724373`*^9, 3.733367692645447*^9, 3.733367900284371*^9, 
  3.733368133472533*^9, 3.792273854643963*^9, 3.79451794188449*^9, 
  3.794518020158409*^9, 3.7945182322170267`*^9, 3.794518785096219*^9, 
  3.794519065102674*^9, 3.794519300633142*^9, 3.7945200260661173`*^9, 
  3.7945202124459553`*^9, 3.7945203204707613`*^9, 3.7945204219814377`*^9, 
  3.794520591952559*^9, 3.7945206674854317`*^9},
 CellLabel->"Out[136]=",
 CellID->696877521,ExpressionUUID->"3d2b855c-b4b2-41fb-8824-a94633fbe0e0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
It is constructed as a SparseArray. Use Normal to get the corresponding \
normal array:\
\>", "MathCaption",
 CellID->2053850088,ExpressionUUID->"76c09268-669c-4f14-a0e5-932b0387ae95"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "Normal"}]], "Input",
 CellLabel->"In[137]:=",
 CellID->406287138,ExpressionUUID->"42a5f6ef-fd0e-42d0-a4eb-ed4338bbe8a1"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"0", ",", 
       SqrtBox[
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Sin",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Sin,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"]]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", 
        SqrtBox[
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Sin",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Sin,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]]}], ",", "0"}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "sph"}], ",", 
     RowBox[{"-", "sph"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.728676057777096*^9, 3.728753501649724*^9, 
  3.728843538781906*^9, 3.730955650930112*^9, 3.7309562027997637`*^9, 
  3.7311164006285067`*^9, 3.7333676928978643`*^9, 3.733367900404236*^9, 
  3.7333681336005096`*^9, 3.792273854794142*^9, 3.7945179419446087`*^9, 
  3.794518020207347*^9, 3.794518232265193*^9, 3.794518785144413*^9, 
  3.7945190651522417`*^9, 3.794519300687296*^9, 3.794520026120175*^9, 
  3.794520212495071*^9, 3.7945203205235023`*^9, 3.794520422101943*^9, 
  3.794520592004999*^9, 3.7945206675379953`*^9},
 CellLabel->"Out[137]=",
 CellID->44684709,ExpressionUUID->"57422f81-3012-4e46-9e04-199daca22218"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Simplify", "[", 
  RowBox[{"%", ",", 
   RowBox[{
    RowBox[{"Sin", "[", 
     RowBox[{"\[Theta]", "[", "]"}], "]"}], ">", "0"}]}], "]"}]], "Input",
 CellLabel->"In[138]:=",
 CellID->1038388545,ExpressionUUID->"4a1967fa-c885-4322-96d2-98c864b1c8a9"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"0", ",", 
       RowBox[{
        InterpretationBox[
         StyleBox["Sin",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Sin,
         Editable->False], "[", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False], "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", 
        RowBox[{
         InterpretationBox[
          StyleBox["Sin",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Sin,
          Editable->False], "[", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False], "]"}]}], ",", "0"}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "sph"}], ",", 
     RowBox[{"-", "sph"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.728676058451811*^9, 3.728753501760151*^9, 
  3.72884353902472*^9, 3.730955653597658*^9, 3.730956202899005*^9, 
  3.731116404393714*^9, 3.733367693128508*^9, 3.733367900498461*^9, 
  3.733368133733858*^9, 3.792273854858161*^9, 3.7945179419987507`*^9, 
  3.794518020255103*^9, 3.79451823232724*^9, 3.794518785193671*^9, 
  3.794519065197981*^9, 3.7945193007331333`*^9, 3.7945200261747837`*^9, 
  3.7945202125439157`*^9, 3.7945203205740423`*^9, 3.794520422154052*^9, 
  3.794520592052568*^9, 3.7945206675944633`*^9},
 CellLabel->"Out[138]=",
 CellID->1544111360,ExpressionUUID->"b28d7da9-1ab9-4b97-870e-8860825ccdb1"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "So far the metric field we have is just a normal tensor field, and ",
 StyleBox["xCoba",
  FontSlant->"Italic"],
 " does not know yet that it is going to play any special role. In order to \
declare it as special we need to use SetCMetric, the component equivalent of \
DefMetric."
}], "Text",
 CellID->950262901,ExpressionUUID->"e28e9306-fece-4925-985a-9797a807b043"],

Cell[CellGroupData[{

Cell["\<\
Declare met as the first-metric of the tangent bundle TangentS2. This will \
precompute a number of objects associated to the metric.\
\>", "MathCaption",
 CellChangeTimes->{3.7286760606361933`*^9},
 CellID->232694368,ExpressionUUID->"2cdb493e-75e6-47c3-925c-a2b15578a003"],

Cell[BoxData[
 RowBox[{"SetCMetric", "[", 
  RowBox[{"met", ",", "sph", ",", 
   RowBox[{"SignatureOfMetric", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"2", ",", "0", ",", "0"}], "}"}]}]}], "]"}]], "Input",
 CellLabel->"In[139]:=",
 CellID->326521789,ExpressionUUID->"cdf01273-2ad2-4b66-9710-e93d92789e28"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Now we can raise and lower indices, because we know what the metric is.\
\>", "MathCaption",
 CellChangeTimes->{3.728676063403723*^9},
 CellID->1030989638,ExpressionUUID->"92c20e72-1098-4ce1-9e71-f786f0ad2678"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"v", "=", 
  RowBox[{"CTensor", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"1", ",", "2"}], "}"}], ",", 
    RowBox[{"{", "sph", "}"}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.731116450663233*^9, 3.731116450871149*^9}},
 CellLabel->"In[140]:=",
 CellID->1944781870,ExpressionUUID->"cf28bfdd-a561-411e-873b-6942866c7a43"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "2"}], "}"}], ",", 
   RowBox[{"{", "sph", "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.728676064793762*^9, 3.728753501916699*^9, 
  3.728843540004698*^9, 3.7309556618502197`*^9, 3.730956203048633*^9, 
  3.731116451750867*^9, 3.733367693406724*^9, 3.733367900703336*^9, 
  3.733368133899973*^9, 3.792273854942314*^9, 3.79451794209559*^9, 
  3.794518020321228*^9, 3.794518232401621*^9, 3.79451878525846*^9, 
  3.794519065263013*^9, 3.794519300804789*^9, 3.794520026248267*^9, 
  3.794520212612193*^9, 3.7945203206447144`*^9, 3.794520422230598*^9, 
  3.7945205921211767`*^9, 3.7945206676728563`*^9},
 CellLabel->"Out[140]=",
 CellID->1687237172,ExpressionUUID->"06b8b117-c613-449e-bde6-fcacd4a0a48c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"v", "[", "a", "]"}]], "Input",
 CellChangeTimes->{{3.731116453037887*^9, 3.7311164531911592`*^9}},
 CellLabel->"In[141]:=",
 CellID->1290059295,ExpressionUUID->"8e4ddb28-cd5f-4c2b-b7be-58993e16c9a2"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", "1"}]],
           Annotation[#, {1} -> 1, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["2",
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", "2"}]],
           Annotation[#, {2} -> 2, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{1, 2}, {$CellContext`sph}, 0][$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.728676065623397*^9, 3.728753502001403*^9, 
  3.728843540592276*^9, 3.730955662651278*^9, 3.73095620314038*^9, 
  3.731116453603775*^9, 3.733367693691441*^9, 3.733367900801897*^9, 
  3.733368134034402*^9, 3.792273854994358*^9, 3.794517942151114*^9, 
  3.794518020369358*^9, 3.794518232456089*^9, 3.7945187853051367`*^9, 
  3.794519065312443*^9, 3.7945193008531513`*^9, 3.7945200263020077`*^9, 
  3.794520212661154*^9, 3.794520320699121*^9, 3.794520422283245*^9, 
  3.79452059217549*^9, 3.794520667727046*^9},
 CellLabel->"Out[141]=",
 CellID->2021074816,ExpressionUUID->"ffc2d082-1c18-4310-9847-e6511d77fbca"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"v", "[", 
  RowBox[{"-", "a"}], "]"}]], "Input",
 CellChangeTimes->{{3.731116454567132*^9, 3.7311164546790743`*^9}},
 CellLabel->"In[142]:=",
 CellID->510741193,ExpressionUUID->"98c86999-7c32-4512-aa90-ea2b51283263"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", "1"}]],
           Annotation[#, {1} -> 1, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{"2", " ", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], "2"]}],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              RowBox[{"2", " ", 
                SuperscriptBox[
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                   InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                 "2"]}]}]],
           Annotation[#, {2} -> 2 Sin[
                $CellContext`\[Theta][]]^2, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" "},
         {"a"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{1, 2 Sin[
       $CellContext`\[Theta][]]^2}, {-$CellContext`sph}, 0][-$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.728676066219687*^9, 3.728753502115199*^9, 
  3.728843541176351*^9, 3.7309556634227133`*^9, 3.730956203250592*^9, 
  3.731116455062862*^9, 3.733367693888474*^9, 3.7333679009204283`*^9, 
  3.7333681342002897`*^9, 3.7922738550472193`*^9, 3.794517942203671*^9, 
  3.794518020415204*^9, 3.7945182325030117`*^9, 3.794518785349265*^9, 
  3.7945190653530483`*^9, 3.794519300906741*^9, 3.794520026355332*^9, 
  3.794520212708894*^9, 3.794520320744041*^9, 3.794520422334852*^9, 
  3.7945205922208853`*^9, 3.79452066776919*^9},
 CellLabel->"Out[142]=",
 CellID->1455587913,ExpressionUUID->"785051c1-38f0-403f-8e52-5b2f4d0faee4"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"met", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}]}], "]"}], 
  RowBox[{"v", "[", "b", "]"}]}]], "Input",
 CellChangeTimes->{{3.731116457567071*^9, 3.731116457687114*^9}},
 CellLabel->"In[143]:=",
 CellID->1805651455,ExpressionUUID->"1db299f2-d128-44a1-b0a6-fc6c51d45fff"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", "1"}]],
           Annotation[#, {1} -> 1, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{"2", " ", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], "2"]}],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              RowBox[{"2", " ", 
                SuperscriptBox[
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                   InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                 "2"]}]}]],
           Annotation[#, {2} -> 2 Sin[
                $CellContext`\[Theta][]]^2, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" "},
         {"a"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{1, 2 Sin[
       $CellContext`\[Theta][]]^2}, {-$CellContext`sph}, 0][-$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.728676066883497*^9, 3.728753502230887*^9, 
  3.728843541387692*^9, 3.730955664159463*^9, 3.7309562033406076`*^9, 
  3.731116458310327*^9, 3.733367694011248*^9, 3.733367901037712*^9, 
  3.733368134330778*^9, 3.792273855100029*^9, 3.794517942256661*^9, 
  3.794518020460882*^9, 3.794518232560233*^9, 3.7945187853910427`*^9, 
  3.79451906539327*^9, 3.794519300957837*^9, 3.794520026403234*^9, 
  3.794520212757547*^9, 3.794520320793478*^9, 3.794520422387203*^9, 
  3.794520592268255*^9, 3.7945206678205843`*^9},
 CellLabel->"Out[143]=",
 CellID->2079830814,ExpressionUUID->"c69088f3-8381-4c25-b213-0ecf9bbfa903"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", " ", 
  RowBox[{"imet", "[", 
   RowBox[{"a", ",", "b"}], "]"}]}]], "Input",
 CellLabel->"In[144]:=",
 CellID->1177228789,ExpressionUUID->"e2fa1e5b-67fa-4a73-aaf5-3edda9c20898"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", "1"}]],
           Annotation[#, {1} -> 1, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["2",
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", "2"}]],
           Annotation[#, {2} -> 2, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"b"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{1, 2}, {$CellContext`sph}, 0][$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.728676067471979*^9, 3.7287535023516493`*^9, 
  3.728843541587948*^9, 3.7309556649434757`*^9, 3.7309562034629793`*^9, 
  3.7311164612625628`*^9, 3.733367694155286*^9, 3.733367901129724*^9, 
  3.73336813443346*^9, 3.792273855149756*^9, 3.7945179423133183`*^9, 
  3.794518020508644*^9, 3.794518232603956*^9, 3.794518785443054*^9, 
  3.7945190654302063`*^9, 3.794519301077161*^9, 3.7945200264588013`*^9, 
  3.7945202128048697`*^9, 3.794520320843088*^9, 3.794520422435444*^9, 
  3.7945205923138237`*^9, 3.794520667868436*^9},
 CellLabel->"Out[144]=",
 CellID->1125168679,ExpressionUUID->"17f4e406-92a9-41f8-883b-8a9d10eba7a8"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
And there is a contravariant version of the metric, identical to the inverse \
metric tensor we constructed before:\
\>", "MathCaption",
 CellChangeTimes->{{3.731116475958256*^9, 3.7311165158445797`*^9}},
 CellID->1087857793,ExpressionUUID->"fbb4da9a-588e-458a-9906-20abbb057bee"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"met", "[", 
    RowBox[{"a", ",", "b"}], "]"}], ",", 
   RowBox[{"imet", "[", 
    RowBox[{"a", ",", "b"}], "]"}]}], "}"}]], "Input",
 CellChangeTimes->{{3.731116488882328*^9, 3.731116500293507*^9}},
 CellLabel->"In[145]:=",
 CellID->1786613944,ExpressionUUID->"84cc1f4c-246f-4de4-983e-77a654e8879a"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        StyleBox[GridBox[{
           {
            TagBox[
             TooltipBox["1",
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "1"}]],
             Annotation[#, {1, 1} -> 1, "Tooltip"]& ], 
            TagBox[
             TooltipBox["0",
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
             Annotation[#, {1, 2} -> 0, "Tooltip"]& ]},
           {
            TagBox[
             TooltipBox["0",
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
             Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["Csc",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 Csc,
                 Editable->False], "[", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False], "]"}], "2"],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                SuperscriptBox[
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Csc"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Csc, Editable -> False], "[", 
                   InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                 "2"]}]],
             Annotation[#, {2, 2} -> Csc[
                 $CellContext`\[Theta][]]^2, "Tooltip"]& ]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
         FontSize->Small], 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xCoba`CTensor[{{1, 0}, {0, Csc[
         $CellContext`\[Theta][]]^2}}, {$CellContext`sph, $CellContext`sph}, 
     0][$CellContext`a, $CellContext`b],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        StyleBox[GridBox[{
           {
            TagBox[
             TooltipBox["1",
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "1"}]],
             Annotation[#, {1, 1} -> 1, "Tooltip"]& ], 
            TagBox[
             TooltipBox["0",
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
             Annotation[#, {1, 2} -> 0, "Tooltip"]& ]},
           {
            TagBox[
             TooltipBox["0",
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
             Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["Csc",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 Csc,
                 Editable->False], "[", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False], "]"}], "2"],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                SuperscriptBox[
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Csc"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Csc, Editable -> False], "[", 
                   InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                 "2"]}]],
             Annotation[#, {2, 2} -> Csc[
                 $CellContext`\[Theta][]]^2, "Tooltip"]& ]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
         FontSize->Small], 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xCoba`CTensor[{{1, 0}, {0, Csc[
         $CellContext`\[Theta][]]^2}}, {$CellContext`sph, $CellContext`sph}, 
     0][$CellContext`a, $CellContext`b],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{{3.7311164907539577`*^9, 3.7311165231322317`*^9}, 
   3.7333676944596653`*^9, 3.733367901252689*^9, 3.7333681345714817`*^9, 
   3.792273855210062*^9, 3.7945179424421797`*^9, 3.7945180205539293`*^9, 
   3.794518232665267*^9, 3.794518785488153*^9, 3.794519065474362*^9, 
   3.794519301132717*^9, 3.7945200265188503`*^9, 3.794520212858026*^9, 
   3.794520320899633*^9, 3.794520422489249*^9, 3.794520592424769*^9, 
   3.794520667926489*^9},
 CellLabel->"Out[145]=",
 CellID->2116476118,ExpressionUUID->"faccadf6-dcee-4ac1-a864-2867f481ae8c"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.2.5. ToCTensor: changes of bases and height", "Subsubsection",
 CellChangeTimes->{{3.733368031399674*^9, 3.7333680329038687`*^9}},
 CellID->93454524,ExpressionUUID->"15ff0bd8-6a27-4801-94e2-215f648c4a35"],

Cell["\<\
One of the most useful features of xCoba in comparison with other tensor CASs \
is its ability to manipulate several bases simultaneously and tensor \
components in different, even mixed bases. We can also have CTensor objects \
expressed in different bases, and we can easily change those bases, \
effectively implementing the basic tensorial transformation law. Note that \
this changes simultaneously bases and components, such that the changes \
compensate mutually and the actual tensor is left unchanged. Hence we are \
referring to passive changes.\
\>", "Text",
 CellChangeTimes->{{3.730955764489324*^9, 3.730955767848778*^9}},
 CellID->879638911,ExpressionUUID->"fba372ca-b450-469f-88a5-0f6c50231f23"],

Cell[CellGroupData[{

Cell["\<\
Introduce a second frame on the sphere. To avoid discussing coordinate \
dependencies, we choose it to be a rigid transformation. Note how the \
transformation is provided: the new frame is associated to the first tensor \
slot, and in covariant form:\
\>", "MathCaption",
 CellChangeTimes->{{3.7286761027460337`*^9, 3.728676102874404*^9}},
 CellID->813751229,ExpressionUUID->"2fef869b-fbb6-4446-acea-f02c79cc4c03"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefBasis", "[", 
  RowBox[{"sph2", ",", "TangentS2", ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", "2"}], "}"}], ",", 
   RowBox[{"BasisChange", "\[Rule]", 
    RowBox[{"CTensor", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{"2", ",", "1"}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{"-", "3"}], ",", "2"}], "}"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "sph2"}], ",", "sph"}], "}"}]}], "]"}]}], ",", 
   RowBox[{"BasisColor", "\[Rule]", "Blue"}]}], "]"}]], "Input",
 CellLabel->"In[146]:=",
 CellID->1541041707,ExpressionUUID->"9fd032d3-5d29-4298-ab0c-b20ba21aca17"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`sph}, {1, $CellContext`sph2}],
     Editable->False], "\[Rule]", 
    FractionBox["2", "7"]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{1, -$CellContext`sph}, {1, $CellContext`sph2}] -> 
   Rational[2, 7], " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.731116559812228*^9, 3.733367699949583*^9, 
  3.733367901367055*^9, 3.7333681346984053`*^9, 3.7922738552800827`*^9, 
  3.7945179425126266`*^9, 3.79451802061865*^9, 3.794518255104044*^9, 
  3.7945187855417967`*^9, 3.794519065531867*^9, 3.7945193011928997`*^9, 
  3.794520026579459*^9, 3.7945202129167*^9, 3.794520320975061*^9, 
  3.794520422563588*^9, 3.794520592490206*^9, 3.794520667992527*^9},
 CellLabel->"During evaluation of In[146]:=",
 CellID->694986318,ExpressionUUID->"60fff894-fe95-45a0-91d0-af8acc69a4f4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`sph}, {2, $CellContext`sph2}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     FractionBox["1", "7"]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{1, -$CellContext`sph}, {2, $CellContext`sph2}] -> 
   Rational[-1, 7], " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.731116559812228*^9, 3.733367699949583*^9, 
  3.733367901367055*^9, 3.7333681346984053`*^9, 3.7922738552800827`*^9, 
  3.7945179425126266`*^9, 3.79451802061865*^9, 3.794518255104044*^9, 
  3.7945187855417967`*^9, 3.794519065531867*^9, 3.7945193011928997`*^9, 
  3.794520026579459*^9, 3.7945202129167*^9, 3.794520320975061*^9, 
  3.794520422563588*^9, 3.794520592490206*^9, 3.794520668012496*^9},
 CellLabel->"During evaluation of In[146]:=",
 CellID->471490593,ExpressionUUID->"bfe89d69-7c36-4533-89ee-fb67ebe89e92"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`sph}, {1, $CellContext`sph2}],
     Editable->False], "\[Rule]", 
    FractionBox["3", "7"]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{2, -$CellContext`sph}, {1, $CellContext`sph2}] -> 
   Rational[3, 7], " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.731116559812228*^9, 3.733367699949583*^9, 
  3.733367901367055*^9, 3.7333681346984053`*^9, 3.7922738552800827`*^9, 
  3.7945179425126266`*^9, 3.79451802061865*^9, 3.794518255104044*^9, 
  3.7945187855417967`*^9, 3.794519065531867*^9, 3.7945193011928997`*^9, 
  3.794520026579459*^9, 3.7945202129167*^9, 3.794520320975061*^9, 
  3.794520422563588*^9, 3.794520592490206*^9, 3.794520668027945*^9},
 CellLabel->"During evaluation of In[146]:=",
 CellID->127739121,ExpressionUUID->"79e4ea9b-b100-45b0-b36f-ccf76e78be62"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`sph}, {2, $CellContext`sph2}],
     Editable->False], "\[Rule]", 
    FractionBox["2", "7"]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{2, -$CellContext`sph}, {2, $CellContext`sph2}] -> 
   Rational[2, 7], " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.731116559812228*^9, 3.733367699949583*^9, 
  3.733367901367055*^9, 3.7333681346984053`*^9, 3.7922738552800827`*^9, 
  3.7945179425126266`*^9, 3.79451802061865*^9, 3.794518255104044*^9, 
  3.7945187855417967`*^9, 3.794519065531867*^9, 3.7945193011928997`*^9, 
  3.794520026579459*^9, 3.7945202129167*^9, 3.794520320975061*^9, 
  3.794520422563588*^9, 3.794520592490206*^9, 3.794520668061002*^9},
 CellLabel->"During evaluation of In[146]:=",
 CellID->1575243294,ExpressionUUID->"527b8224-38b9-4acf-99d6-c7664481c208"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`sph2}, {1, $CellContext`sph}],
     Editable->False], "\[Rule]", "2"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{1, -$CellContext`sph2}, {1, $CellContext`sph}] -> 2, 
   " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.731116559812228*^9, 3.733367699949583*^9, 
  3.733367901367055*^9, 3.7333681346984053`*^9, 3.7922738552800827`*^9, 
  3.7945179425126266`*^9, 3.79451802061865*^9, 3.794518255104044*^9, 
  3.7945187855417967`*^9, 3.794519065531867*^9, 3.7945193011928997`*^9, 
  3.794520026579459*^9, 3.7945202129167*^9, 3.794520320975061*^9, 
  3.794520422563588*^9, 3.794520592490206*^9, 3.794520668072935*^9},
 CellLabel->"During evaluation of In[146]:=",
 CellID->935642042,ExpressionUUID->"cd84e3e3-21f3-4585-8102-28cf0c483bf8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`sph2}, {2, $CellContext`sph}],
     Editable->False], "\[Rule]", "1"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{1, -$CellContext`sph2}, {2, $CellContext`sph}] -> 1, 
   " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.731116559812228*^9, 3.733367699949583*^9, 
  3.733367901367055*^9, 3.7333681346984053`*^9, 3.7922738552800827`*^9, 
  3.7945179425126266`*^9, 3.79451802061865*^9, 3.794518255104044*^9, 
  3.7945187855417967`*^9, 3.794519065531867*^9, 3.7945193011928997`*^9, 
  3.794520026579459*^9, 3.7945202129167*^9, 3.794520320975061*^9, 
  3.794520422563588*^9, 3.794520592490206*^9, 3.7945206681034517`*^9},
 CellLabel->"During evaluation of In[146]:=",
 CellID->1460536560,ExpressionUUID->"1bbe3d4d-f4d8-4c3f-bb13-9144e5f456fd"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`sph2}, {1, $CellContext`sph}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", "3"}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{2, -$CellContext`sph2}, {1, $CellContext`sph}] -> -3, 
   " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.731116559812228*^9, 3.733367699949583*^9, 
  3.733367901367055*^9, 3.7333681346984053`*^9, 3.7922738552800827`*^9, 
  3.7945179425126266`*^9, 3.79451802061865*^9, 3.794518255104044*^9, 
  3.7945187855417967`*^9, 3.794519065531867*^9, 3.7945193011928997`*^9, 
  3.794520026579459*^9, 3.7945202129167*^9, 3.794520320975061*^9, 
  3.794520422563588*^9, 3.794520592490206*^9, 3.7945206681142178`*^9},
 CellLabel->"During evaluation of In[146]:=",
 CellID->1499878710,ExpressionUUID->"edaa92c1-55e0-4c37-8c30-1682522f3dec"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`sph2}, {2, $CellContext`sph}],
     Editable->False], "\[Rule]", "2"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{2, -$CellContext`sph2}, {2, $CellContext`sph}] -> 2, 
   " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.731116559812228*^9, 3.733367699949583*^9, 
  3.733367901367055*^9, 3.7333681346984053`*^9, 3.7922738552800827`*^9, 
  3.7945179425126266`*^9, 3.79451802061865*^9, 3.794518255104044*^9, 
  3.7945187855417967`*^9, 3.794519065531867*^9, 3.7945193011928997`*^9, 
  3.794520026579459*^9, 3.7945202129167*^9, 3.794520320975061*^9, 
  3.794520422563588*^9, 3.794520592490206*^9, 3.794520668127924*^9},
 CellLabel->"During evaluation of In[146]:=",
 CellID->1298430484,ExpressionUUID->"ba7fe945-5931-4677-b721-e770a66b25e5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      UnderscriptBox[
       OverscriptBox["J", 
        StyleBox["~",
         FontColor->RGBColor[1, 0, 0]]], 
       StyleBox["~",
        FontColor->RGBColor[0, 0, 1]]],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Jacobiansphsph2[],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      UnderscriptBox[
       OverscriptBox[
        FractionBox["1", "7"], 
        StyleBox["~",
         FontColor->RGBColor[1., 0., 0.]]], 
       StyleBox["~",
        FontColor->RGBColor[0., 0., 1.]]],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xCoba`CTensor[
      Rational[1, 7], {}, $CellContext`sph - $CellContext`sph2][],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Jacobiansphsph2"}],
  SequenceForm[
  "Added independent rule ", $CellContext`Jacobiansphsph2[] -> 
   xAct`xCoba`CTensor[
     Rational[1, 7], {}, $CellContext`sph - $CellContext`sph2][], 
   " for tensor ", $CellContext`Jacobiansphsph2],
  Editable->False]], "Print",
 CellChangeTimes->{3.731116559812228*^9, 3.733367699949583*^9, 
  3.733367901367055*^9, 3.7333681346984053`*^9, 3.7922738552800827`*^9, 
  3.7945179425126266`*^9, 3.79451802061865*^9, 3.794518255104044*^9, 
  3.7945187855417967`*^9, 3.794519065531867*^9, 3.7945193011928997`*^9, 
  3.794520026579459*^9, 3.7945202129167*^9, 3.794520320975061*^9, 
  3.794520422563588*^9, 3.794520592490206*^9, 3.7945206681350927`*^9},
 CellLabel->"During evaluation of In[146]:=",
 CellID->2139807421,ExpressionUUID->"f0ea909e-07cc-4403-986e-257c1d424a9b"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["This is a vector field on the sphere:", "MathCaption",
 CellChangeTimes->{{3.72867610528189*^9, 3.7286761054100647`*^9}},
 CellID->1635186729,ExpressionUUID->"b0de54d5-6a1b-4eab-9f5f-0cca5eaecac2"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"v", "=", 
  RowBox[{"CTensor", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"1", ",", "1"}], "}"}], ",", 
    RowBox[{"{", "sph", "}"}]}], "]"}]}]], "Input",
 CellLabel->"In[147]:=",
 CellID->1235123644,ExpressionUUID->"de6dded8-0a70-486a-9974-541ee7bf40db"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "1"}], "}"}], ",", 
   RowBox[{"{", "sph", "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.72867610656839*^9, 3.728753502500987*^9, 
  3.728843548662925*^9, 3.7309556990424232`*^9, 3.730956203629511*^9, 
  3.731116575904091*^9, 3.733367700594612*^9, 3.7333679017909107`*^9, 
  3.7333681353177*^9, 3.792273855512083*^9, 3.794517942715129*^9, 
  3.79451802080079*^9, 3.7945182553754683`*^9, 3.794518785825021*^9, 
  3.794519065763659*^9, 3.79451930142778*^9, 3.794520026781599*^9, 
  3.79452021321241*^9, 3.79452032122537*^9, 3.794520422757805*^9, 
  3.794520592742738*^9, 3.794520668182948*^9},
 CellLabel->"Out[147]=",
 CellID->1982172085,ExpressionUUID->"39be02f8-cb6f-4156-9069-c742f7fbb547"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Change it to the new frame:", "MathCaption",
 CellChangeTimes->{{3.7286761075856953`*^9, 3.728676107713916*^9}},
 CellID->461896852,ExpressionUUID->"46c8960f-c0cc-46fd-b2de-4d68db8b175c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCTensor", "[", 
  RowBox[{"v", ",", 
   RowBox[{"{", "sph2", "}"}]}], "]"}]], "Input",
 CellLabel->"In[148]:=",
 CellID->2094491306,ExpressionUUID->"4f820507-1328-47d1-8df3-0c0660860dd8"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     FractionBox["5", "7"], ",", 
     FractionBox["1", "7"]}], "}"}], ",", 
   RowBox[{"{", "sph2", "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.7286761085920362`*^9, 3.728753502617202*^9, 
  3.728843548842682*^9, 3.730955700457118*^9, 3.7309562037336607`*^9, 
  3.731116577992168*^9, 3.7333677008659897`*^9, 3.733367901935218*^9, 
  3.733368135448498*^9, 3.792273855565474*^9, 3.794517942777186*^9, 
  3.794518020846373*^9, 3.794518255423668*^9, 3.7945187858746557`*^9, 
  3.794519065811317*^9, 3.794519301500861*^9, 3.7945200268338623`*^9, 
  3.794520213260922*^9, 3.7945203213363667`*^9, 3.794520422821391*^9, 
  3.794520592791296*^9, 3.7945206682316227`*^9},
 CellLabel->"Out[148]=",
 CellID->452416684,ExpressionUUID->"369bb546-b40b-454c-bdea-2bb7f1c3d7ee"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Change it back to the original frame:", "MathCaption",
 CellChangeTimes->{{3.7286761097535543`*^9, 3.7286761098819942`*^9}},
 CellID->1452411003,ExpressionUUID->"0e25f331-0ac8-4564-ad53-c974391e09db"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCTensor", "[", 
  RowBox[{"%", ",", 
   RowBox[{"{", "sph", "}"}]}], "]"}]], "Input",
 CellLabel->"In[149]:=",
 CellID->225111972,ExpressionUUID->"c6bf1f87-4479-4022-9483-5b77ab1d236c"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "1"}], "}"}], ",", 
   RowBox[{"{", "sph", "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.728676110767076*^9, 3.728753502755621*^9, 
  3.7288435490558777`*^9, 3.7309557021476727`*^9, 3.7309562038498163`*^9, 
  3.7311165798564157`*^9, 3.733367701111219*^9, 3.73336790206789*^9, 
  3.7333681355806837`*^9, 3.792273855615981*^9, 3.794517942831164*^9, 
  3.7945180209634113`*^9, 3.794518255473123*^9, 3.794518785919672*^9, 
  3.794519065859564*^9, 3.794519301563034*^9, 3.794520026884075*^9, 
  3.7945202133097553`*^9, 3.794520321433848*^9, 3.794520422866165*^9, 
  3.794520592837473*^9, 3.7945206682828283`*^9},
 CellLabel->"Out[149]=",
 CellID->1675097608,ExpressionUUID->"dd781b5f-f301-4d47-8dde-3aa31f6a2c89"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The identity tensor in the sph frame:", "MathCaption",
 CellChangeTimes->{{3.7286761118494663`*^9, 3.728676111985793*^9}},
 CellID->1892848718,ExpressionUUID->"390144d5-6ca7-4946-b5ae-1be03a560267"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCTensor", "[", 
  RowBox[{"delta", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "sph"}], ",", "sph"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[150]:=",
 CellID->285656768,ExpressionUUID->"d33eb0b0-af68-4a20-b544-fc54e1a294fb"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "1"}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "sph"}], ",", "sph"}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.7286761128675547`*^9, 3.728753502849312*^9, 
  3.7288435494091663`*^9, 3.7309557053153152`*^9, 3.730956203946227*^9, 
  3.7311165845404*^9, 3.733367701310601*^9, 3.7333679021656713`*^9, 
  3.733368135714673*^9, 3.792273855670068*^9, 3.794517942887891*^9, 
  3.794518021013277*^9, 3.7945182555214643`*^9, 3.794518785971491*^9, 
  3.794519065902011*^9, 3.794519301613161*^9, 3.794520027000415*^9, 
  3.794520213353772*^9, 3.794520321491478*^9, 3.79452042291367*^9, 
  3.7945205928882523`*^9, 3.794520668335985*^9},
 CellLabel->"Out[150]=",
 CellID->2002369133,ExpressionUUID->"20426501-18f9-49fd-b96f-326538e7e5f8"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Its components do not change under a change of basis:", "MathCaption",
 CellChangeTimes->{{3.728676114369464*^9, 3.728676114497685*^9}},
 CellID->925233859,ExpressionUUID->"b9137542-a102-4723-acfb-eca83836ed60"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCTensor", "[", 
  RowBox[{"%", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "sph2"}], ",", "sph2"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[151]:=",
 CellID->476380286,ExpressionUUID->"8d074e6b-da6f-46db-bfa1-27794f1062e0"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "1"}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "sph2"}], ",", "sph2"}], "}"}], ",", "0"}], "]"}]], "Output",\

 CellChangeTimes->{3.728676115580373*^9, 3.728753502967815*^9, 
  3.728843549992716*^9, 3.730955706473222*^9, 3.73095620404983*^9, 
  3.7311165891217337`*^9, 3.733367701427845*^9, 3.73336790226746*^9, 
  3.7333681358468237`*^9, 3.792273855785727*^9, 3.794517942942418*^9, 
  3.7945180210616198`*^9, 3.794518255574893*^9, 3.794518786016056*^9, 
  3.794519065946678*^9, 3.7945193016607323`*^9, 3.794520027056294*^9, 
  3.79452021339603*^9, 3.794520321540834*^9, 3.794520422965929*^9, 
  3.794520592941352*^9, 3.794520668386332*^9},
 CellLabel->"Out[151]=",
 CellID->601786194,ExpressionUUID->"8f31ad41-a3b1-4a41-9496-e31cf20d6544"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The components of the metric do change:", "MathCaption",
 CellChangeTimes->{{3.728676117057246*^9, 3.7286761171935167`*^9}},
 CellID->290874391,ExpressionUUID->"f297dc2a-0116-4f54-8d13-f2053d3421dd"],

Cell[CellGroupData[{

Cell[BoxData["met"], "Input",
 CellLabel->"In[152]:=",
 CellID->1053108965,ExpressionUUID->"4baaf4d7-1226-43d4-92fb-009afb7220b3"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", 
       SuperscriptBox[
        RowBox[{
         InterpretationBox[
          StyleBox["Sin",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Sin,
          Editable->False], "[", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False], "]"}], "2"]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "sph"}], ",", 
     RowBox[{"-", "sph"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.7286761180788507`*^9, 3.728753503197815*^9, 
  3.7288435502883167`*^9, 3.7309557076309843`*^9, 3.7309562041454487`*^9, 
  3.731116596367614*^9, 3.731116639077922*^9, 3.733367701693275*^9, 
  3.733367902359351*^9, 3.733368135980451*^9, 3.7922738558414907`*^9, 
  3.794517943004327*^9, 3.794518021117931*^9, 3.7945182556214933`*^9, 
  3.7945187860585318`*^9, 3.794519065988503*^9, 3.794519301710647*^9, 
  3.7945200271058073`*^9, 3.79452021344169*^9, 3.7945203215912943`*^9, 
  3.7945204230139513`*^9, 3.794520592994248*^9, 3.7945206684986343`*^9},
 CellLabel->"Out[152]=",
 CellID->1720068835,ExpressionUUID->"3743728b-37c0-4512-8a62-24310fa10372"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCTensor", "[", 
  RowBox[{"%", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "sph2"}], ",", 
     RowBox[{"-", "sph2"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[153]:=",
 CellID->1996814204,ExpressionUUID->"d035ad0d-37a0-4a93-9525-1b06045d818b"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"4", "+", 
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Sin",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Sin,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"]}], ",", 
       RowBox[{
        RowBox[{"-", "6"}], "+", 
        RowBox[{"2", " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Sin",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Sin,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}]}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "6"}], "+", 
        RowBox[{"2", " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Sin",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Sin,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}]}], ",", 
       RowBox[{"9", "+", 
        RowBox[{"4", " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Sin",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Sin,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}]}]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "sph2"}], ",", 
     RowBox[{"-", "sph2"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.7286761186619387`*^9, 3.728753503333844*^9, 
  3.728843550592276*^9, 3.7309557082714453`*^9, 3.7309562042499533`*^9, 
  3.731116600997437*^9, 3.731116641181046*^9, 3.7333677019937353`*^9, 
  3.73336790248314*^9, 3.733368136113861*^9, 3.7922738558994713`*^9, 
  3.794517943064328*^9, 3.7945180211606207`*^9, 3.7945182556777477`*^9, 
  3.79451878612206*^9, 3.794519066040182*^9, 3.794519301762455*^9, 
  3.7945200271733217`*^9, 3.794520213494689*^9, 3.794520321641142*^9, 
  3.794520423114047*^9, 3.794520593042297*^9, 3.794520668552074*^9},
 CellLabel->"Out[153]=",
 CellID->1919537578,ExpressionUUID->"07ea3ab9-d72c-4c71-bede-04d4c98f53bc"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["but they still form a symmetric 2-covariant tensor field:", \
"MathCaption",
 CellChangeTimes->{{3.731116656015995*^9, 3.731116669415678*^9}},
 CellID->139451525,ExpressionUUID->"ec790cea-789b-41ee-8075-28dc0a46f31b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SymmetryGroupOfTensor", "[", "%", "]"}]], "Input",
 CellChangeTimes->{{3.731116642903839*^9, 3.731116647337907*^9}},
 CellLabel->"In[154]:=",
 CellID->15590300,ExpressionUUID->"e6fe2094-8b84-408c-bfd7-5f29de64a910"],

Cell[BoxData[
 RowBox[{"StrongGenSet", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "2"}], "}"}], ",", 
   RowBox[{"GenSet", "[", 
    InterpretationBox[
     StyleBox[
      RowBox[{"(", "1", ",", "2", ")"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xPerm`Cycles[{1, 2}],
     Editable->False], "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.731116648021472*^9, 3.733367702199834*^9, 
  3.7333679025779533`*^9, 3.733368136248125*^9, 3.792273855952577*^9, 
  3.794517943123732*^9, 3.794518021205*^9, 3.7945182557275553`*^9, 
  3.7945187861662483`*^9, 3.794519066082656*^9, 3.794519301809924*^9, 
  3.794520027228299*^9, 3.794520213538178*^9, 3.7945203216914873`*^9, 
  3.794520423168989*^9, 3.794520593093286*^9, 3.794520668604909*^9},
 CellLabel->"Out[154]=",
 CellID->21244565,ExpressionUUID->"9c4904a5-0132-41bd-923e-c32cacb80199"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We can also use CTensor to change the height of the slots of a tensor. \
Effectively that absorbs metric factors:\
\>", "MathCaption",
 CellChangeTimes->{{3.7286761201452513`*^9, 3.7286761202734222`*^9}, {
  3.7309557557208967`*^9, 3.730955758168998*^9}},
 CellID->1561257415,ExpressionUUID->"ab8552c2-4b44-4df8-a1af-85a2c6722ad5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCTensor", "[", 
  RowBox[{"met", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "sph2"}], ",", "sph2"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[155]:=",
 CellID->2098174859,ExpressionUUID->"46935187-9e7e-4eb4-989b-e221c472609b"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "1"}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "sph2"}], ",", "sph2"}], "}"}], ",", "0"}], "]"}]], "Output",\

 CellChangeTimes->{3.7286761211829367`*^9, 3.728753503466366*^9, 
  3.728843551195188*^9, 3.730955709372058*^9, 3.730956204346127*^9, 
  3.7311166787080307`*^9, 3.733367702442789*^9, 3.733367902683323*^9, 
  3.733368136398306*^9, 3.7922738560116663`*^9, 3.7945179431839457`*^9, 
  3.79451802124679*^9, 3.7945182557767344`*^9, 3.7945187862095613`*^9, 
  3.794519066126589*^9, 3.7945193018586607`*^9, 3.794520027279765*^9, 
  3.7945202135934553`*^9, 3.794520321738276*^9, 3.794520423219964*^9, 
  3.794520593139408*^9, 3.794520668656852*^9},
 CellLabel->"Out[155]=",
 CellID->575118553,ExpressionUUID->"3669b687-7f90-4281-8b0e-3725ebe0f2b7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData["v"], "Input",
 CellLabel->"In[156]:=",
 CellID->194260731,ExpressionUUID->"d5ed94c4-1a08-4067-88a2-687919640ab6"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "1"}], "}"}], ",", 
   RowBox[{"{", "sph", "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.7286761217573347`*^9, 3.728753503600305*^9, 
  3.728843551709519*^9, 3.7309557099879503`*^9, 3.730956204450705*^9, 
  3.731116679731805*^9, 3.733367702649626*^9, 3.7333679028003407`*^9, 
  3.733368136530574*^9, 3.792273856069606*^9, 3.7945179432365427`*^9, 
  3.794518021290954*^9, 3.794518255824853*^9, 3.7945187862580833`*^9, 
  3.79451906616824*^9, 3.7945193019097147`*^9, 3.794520027334077*^9, 
  3.794520213639337*^9, 3.794520321786306*^9, 3.794520423267045*^9, 
  3.794520593193829*^9, 3.794520668716757*^9},
 CellLabel->"Out[156]=",
 CellID->193625124,ExpressionUUID->"af4593b6-0970-4f94-9f1b-70f141f03894"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCTensor", "[", 
  RowBox[{"v", ",", 
   RowBox[{"{", 
    RowBox[{"-", "sph"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[157]:=",
 CellID->1887239997,ExpressionUUID->"c16e1072-6ad6-4a8f-aba0-dae1953728f7"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"-", "sph"}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.7286761222916822`*^9, 3.728753503734795*^9, 
  3.728843551892592*^9, 3.73095571059761*^9, 3.730956204545691*^9, 
  3.731116681413877*^9, 3.733367702793211*^9, 3.7333679028932667`*^9, 
  3.7333681366641617`*^9, 3.7922738561240187`*^9, 3.794517943290929*^9, 
  3.79451802133317*^9, 3.794518255873672*^9, 3.794518786306093*^9, 
  3.794519066211969*^9, 3.7945193019560328`*^9, 3.794520027385281*^9, 
  3.7945202136923532`*^9, 3.794520321827422*^9, 3.794520423318544*^9, 
  3.794520593238728*^9, 3.794520668766094*^9},
 CellLabel->"Out[157]=",
 CellID->819733863,ExpressionUUID->"0f1228f9-df98-4cf0-9b74-6e1d89492cb6"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Finally, we can also use ToCTensor to expand a tensor into its components:\
\>", "MathCaption",
 CellChangeTimes->{{3.728676123993044*^9, 3.7286761241214113`*^9}},
 CellID->1934229275,ExpressionUUID->"96897e6f-4240-4c44-829d-734f69e2b231"],

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"S", "[", 
    RowBox[{"\[ScriptA]", ",", 
     RowBox[{"-", "\[ScriptB]"}]}], "]"}], ",", "S2"}], "]"}]], "Input",
 CellLabel->"In[158]:=",
 CellID->2131460286,ExpressionUUID->"0d6ce07e-9a57-41d4-bb16-711464c62053"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCTensor", "[", 
  RowBox[{"S", ",", 
   RowBox[{"{", 
    RowBox[{"sph", ",", 
     RowBox[{"-", "sph2"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[159]:=",
 CellID->1278311195,ExpressionUUID->"6f5d3fbd-1c61-47e4-91e9-50f3e21a36a9"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"S", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], " "},
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`S[{1, $CellContext`sph}, {1, -$CellContext`sph2}],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"S", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], " "},
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`S[{1, $CellContext`sph}, {2, -$CellContext`sph2}],
        Editable->False]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"S", 
            StyleBox[GridBox[{
               {
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]], " "},
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`S[{2, $CellContext`sph}, {1, -$CellContext`sph2}],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"S", 
            StyleBox[GridBox[{
               {
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]], " "},
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "Rows" -> {{
                   Offset[0.]}}}],
             FontSize->9]}
          },
          GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`S[{2, $CellContext`sph}, {2, -$CellContext`sph2}],
        Editable->False]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"sph", ",", 
     RowBox[{"-", "sph2"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.728676125546248*^9, 3.72875350396524*^9, 
  3.728843552196416*^9, 3.730955713197482*^9, 3.730956204712789*^9, 
  3.731116690075118*^9, 3.733367703224104*^9, 3.733367903100915*^9, 
  3.733368136880728*^9, 3.7922738562147303`*^9, 3.794517943420961*^9, 
  3.794518021400025*^9, 3.794518255957858*^9, 3.794518786379438*^9, 
  3.794519066276649*^9, 3.794519302064831*^9, 3.794520027462707*^9, 
  3.7945202137621317`*^9, 3.7945203219065*^9, 3.79452042338977*^9, 
  3.794520593317913*^9, 3.794520668856515*^9},
 CellLabel->"Out[159]=",
 CellID->1159439936,ExpressionUUID->"32fdf79b-164c-4473-a308-9e0043e85b84"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Compare with the result from ComponentArray:", "MathCaption",
 CellChangeTimes->{{3.728676126768794*^9, 3.7286761269210567`*^9}},
 CellID->319862686,ExpressionUUID->"053464b8-c7fb-4e89-b6da-38b1bc48886a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentArray", "[", 
  RowBox[{"S", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"\[ScriptA]", ",", "sph"}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "\[ScriptB]"}], ",", 
      RowBox[{"-", "sph2"}]}], "}"}]}], "]"}], "]"}]], "Input",
 CellLabel->"In[160]:=",
 CellID->780595161,ExpressionUUID->"fb4d4a04-5d86-4f1a-b2e4-0effda1e513f"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"S", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`S[{1, $CellContext`sph}, {1, -$CellContext`sph2}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"S", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`S[{1, $CellContext`sph}, {2, -$CellContext`sph2}],
      Editable->False]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"S", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`S[{2, $CellContext`sph}, {1, -$CellContext`sph2}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"S", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`S[{2, $CellContext`sph}, {2, -$CellContext`sph2}],
      Editable->False]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728676127912634*^9, 3.728753504087216*^9, 
  3.728843552421702*^9, 3.730955715144766*^9, 3.73095620482939*^9, 
  3.731116694353956*^9, 3.7333677033950233`*^9, 3.733367903217814*^9, 
  3.733368137012936*^9, 3.792273856272525*^9, 3.79451794348772*^9, 
  3.794518021452924*^9, 3.794518256007296*^9, 3.794518786435787*^9, 
  3.794519066330065*^9, 3.794519302121773*^9, 3.794520027526306*^9, 
  3.794520213819669*^9, 3.7945203220081*^9, 3.794520423441083*^9, 
  3.7945205933981524`*^9, 3.7945206689264936`*^9},
 CellLabel->"Out[160]=",
 CellID->440391474,ExpressionUUID->"6dcf4c65-8cbd-4a4d-bfe2-42fcd1194bf3"]
}, Open  ]]
}, Open  ]],

Cell["\<\
Instead of specifying the change of basis with the option BasisChange at the \
time of definition of the new basis, we can first define the two bases and \
then use SetBasisChange to specify the change.\
\>", "Text",
 CellChangeTimes->{{3.730955792729478*^9, 3.730955861389758*^9}},
 CellID->1179051796,ExpressionUUID->"26b4b618-ed42-47dd-aca8-b75197131f29"],

Cell[CellGroupData[{

Cell["\<\
Define another basis, this time without specifying the basis change:\
\>", "MathCaption",
 CellChangeTimes->{{3.730955877261319*^9, 3.730955892796626*^9}},
 CellID->1856836499,ExpressionUUID->"62902d7c-791a-4470-8fe4-9c5884100b3d"],

Cell[BoxData[
 RowBox[{"DefBasis", "[", 
  RowBox[{"sph3", ",", "TangentS2", ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", "2"}], "}"}], ",", 
   RowBox[{"BasisColor", "\[Rule]", "Orange"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.730955898764659*^9, 3.7309559138678417`*^9}},
 CellLabel->"In[161]:=",
 CellID->1318474035,ExpressionUUID->"81664d37-92e8-43c8-983d-ecf693d02258"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Now set the change of basis:", "MathCaption",
 CellChangeTimes->{{3.73095592549046*^9, 3.730955932683408*^9}},
 CellID->701825391,ExpressionUUID->"6060a072-d7f2-4330-8a15-3fd9e7c9055f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SetBasisChange", "[", 
  RowBox[{"CTensor", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"2", ",", "1"}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "3"}], ",", "2"}], "}"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "sph3"}], ",", "sph"}], "}"}]}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.730955934742961*^9, 3.730955960322322*^9}},
 CellLabel->"In[162]:=",
 CellID->1009691769,ExpressionUUID->"c3928df2-6ff2-44f7-8b3f-6948b6e43207"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0.5, 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`sph}, {1, $CellContext`sph3}],
     Editable->False], "\[Rule]", 
    FractionBox["2", "7"]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{1, -$CellContext`sph}, {1, $CellContext`sph3}] -> 
   Rational[2, 7], " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.731116701348281*^9, 3.733367703775797*^9, 
  3.7333679033977213`*^9, 3.733368137230505*^9, 3.792273856368531*^9, 
  3.794517943583797*^9, 3.7945180215434017`*^9, 3.79451825610491*^9, 
  3.7945187865240927`*^9, 3.7945190664129*^9, 3.79451930220175*^9, 
  3.794520027614374*^9, 3.7945202139105177`*^9, 3.794520322108407*^9, 
  3.794520423537464*^9, 3.794520593488057*^9, 3.794520669037397*^9},
 CellLabel->"During evaluation of In[162]:=",
 CellID->1430855956,ExpressionUUID->"e4e7cf8d-36f4-4fa5-93b6-15d5b5d7a12e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0.5, 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`sph}, {2, $CellContext`sph3}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     FractionBox["1", "7"]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{1, -$CellContext`sph}, {2, $CellContext`sph3}] -> 
   Rational[-1, 7], " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.731116701348281*^9, 3.733367703775797*^9, 
  3.7333679033977213`*^9, 3.733368137230505*^9, 3.792273856368531*^9, 
  3.794517943583797*^9, 3.7945180215434017`*^9, 3.79451825610491*^9, 
  3.7945187865240927`*^9, 3.7945190664129*^9, 3.79451930220175*^9, 
  3.794520027614374*^9, 3.7945202139105177`*^9, 3.794520322108407*^9, 
  3.794520423537464*^9, 3.794520593488057*^9, 3.7945206690657473`*^9},
 CellLabel->"During evaluation of In[162]:=",
 CellID->874173386,ExpressionUUID->"6d086ec4-0f75-44c3-a9d0-60e8065afc4f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0.5, 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`sph}, {1, $CellContext`sph3}],
     Editable->False], "\[Rule]", 
    FractionBox["3", "7"]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{2, -$CellContext`sph}, {1, $CellContext`sph3}] -> 
   Rational[3, 7], " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.731116701348281*^9, 3.733367703775797*^9, 
  3.7333679033977213`*^9, 3.733368137230505*^9, 3.792273856368531*^9, 
  3.794517943583797*^9, 3.7945180215434017`*^9, 3.79451825610491*^9, 
  3.7945187865240927`*^9, 3.7945190664129*^9, 3.79451930220175*^9, 
  3.794520027614374*^9, 3.7945202139105177`*^9, 3.794520322108407*^9, 
  3.794520423537464*^9, 3.794520593488057*^9, 3.794520669087426*^9},
 CellLabel->"During evaluation of In[162]:=",
 CellID->1296469375,ExpressionUUID->"77edd136-6347-449d-8e62-aeada9a009d4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0.5, 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`sph}, {2, $CellContext`sph3}],
     Editable->False], "\[Rule]", 
    FractionBox["2", "7"]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{2, -$CellContext`sph}, {2, $CellContext`sph3}] -> 
   Rational[2, 7], " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.731116701348281*^9, 3.733367703775797*^9, 
  3.7333679033977213`*^9, 3.733368137230505*^9, 3.792273856368531*^9, 
  3.794517943583797*^9, 3.7945180215434017`*^9, 3.79451825610491*^9, 
  3.7945187865240927`*^9, 3.7945190664129*^9, 3.79451930220175*^9, 
  3.794520027614374*^9, 3.7945202139105177`*^9, 3.794520322108407*^9, 
  3.794520423537464*^9, 3.794520593488057*^9, 3.794520669107389*^9},
 CellLabel->"During evaluation of In[162]:=",
 CellID->1391703163,ExpressionUUID->"402ab056-cedd-42a7-a0b5-431f83ad6f52"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0.5, 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`sph3}, {1, $CellContext`sph}],
     Editable->False], "\[Rule]", "2"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{1, -$CellContext`sph3}, {1, $CellContext`sph}] -> 2, 
   " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.731116701348281*^9, 3.733367703775797*^9, 
  3.7333679033977213`*^9, 3.733368137230505*^9, 3.792273856368531*^9, 
  3.794517943583797*^9, 3.7945180215434017`*^9, 3.79451825610491*^9, 
  3.7945187865240927`*^9, 3.7945190664129*^9, 3.79451930220175*^9, 
  3.794520027614374*^9, 3.7945202139105177`*^9, 3.794520322108407*^9, 
  3.794520423537464*^9, 3.794520593488057*^9, 3.794520669122862*^9},
 CellLabel->"During evaluation of In[162]:=",
 CellID->2110661064,ExpressionUUID->"d32a9e07-0320-49de-aae9-91d60d0f5784"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0.5, 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`sph3}, {2, $CellContext`sph}],
     Editable->False], "\[Rule]", "1"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{1, -$CellContext`sph3}, {2, $CellContext`sph}] -> 1, 
   " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.731116701348281*^9, 3.733367703775797*^9, 
  3.7333679033977213`*^9, 3.733368137230505*^9, 3.792273856368531*^9, 
  3.794517943583797*^9, 3.7945180215434017`*^9, 3.79451825610491*^9, 
  3.7945187865240927`*^9, 3.7945190664129*^9, 3.79451930220175*^9, 
  3.794520027614374*^9, 3.7945202139105177`*^9, 3.794520322108407*^9, 
  3.794520423537464*^9, 3.794520593488057*^9, 3.794520669158523*^9},
 CellLabel->"During evaluation of In[162]:=",
 CellID->1745621502,ExpressionUUID->"e63d5f98-6fcd-4fbe-b810-7a42a67cca99"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0.5, 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`sph3}, {1, $CellContext`sph}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", "3"}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{2, -$CellContext`sph3}, {1, $CellContext`sph}] -> -3, 
   " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.731116701348281*^9, 3.733367703775797*^9, 
  3.7333679033977213`*^9, 3.733368137230505*^9, 3.792273856368531*^9, 
  3.794517943583797*^9, 3.7945180215434017`*^9, 3.79451825610491*^9, 
  3.7945187865240927`*^9, 3.7945190664129*^9, 3.79451930220175*^9, 
  3.794520027614374*^9, 3.7945202139105177`*^9, 3.794520322108407*^9, 
  3.794520423537464*^9, 3.794520593488057*^9, 3.7945206691739187`*^9},
 CellLabel->"During evaluation of In[162]:=",
 CellID->1886482447,ExpressionUUID->"1fdc16b1-cc0f-400e-b577-6af6e448ba53"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0.5, 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`sph3}, {2, $CellContext`sph}],
     Editable->False], "\[Rule]", "2"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{2, -$CellContext`sph3}, {2, $CellContext`sph}] -> 2, 
   " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.731116701348281*^9, 3.733367703775797*^9, 
  3.7333679033977213`*^9, 3.733368137230505*^9, 3.792273856368531*^9, 
  3.794517943583797*^9, 3.7945180215434017`*^9, 3.79451825610491*^9, 
  3.7945187865240927`*^9, 3.7945190664129*^9, 3.79451930220175*^9, 
  3.794520027614374*^9, 3.7945202139105177`*^9, 3.794520322108407*^9, 
  3.794520423537464*^9, 3.794520593488057*^9, 3.7945206692034407`*^9},
 CellLabel->"During evaluation of In[162]:=",
 CellID->661922421,ExpressionUUID->"a9894a20-8fb8-4620-875d-18355110a5b9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      UnderscriptBox[
       OverscriptBox["J", 
        StyleBox["~",
         FontColor->RGBColor[1, 0, 0]]], 
       StyleBox["~",
        FontColor->RGBColor[1, 0.5, 0]]],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Jacobiansphsph3[],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      UnderscriptBox[
       OverscriptBox[
        FractionBox["1", "7"], 
        StyleBox["~",
         FontColor->RGBColor[1., 0., 0.]]], 
       StyleBox["~",
        FontColor->RGBColor[1., 0.5, 0.]]],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xCoba`CTensor[
      Rational[1, 7], {}, $CellContext`sph - $CellContext`sph3][],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Jacobiansphsph3"}],
  SequenceForm[
  "Added independent rule ", $CellContext`Jacobiansphsph3[] -> 
   xAct`xCoba`CTensor[
     Rational[1, 7], {}, $CellContext`sph - $CellContext`sph3][], 
   " for tensor ", $CellContext`Jacobiansphsph3],
  Editable->False]], "Print",
 CellChangeTimes->{3.731116701348281*^9, 3.733367703775797*^9, 
  3.7333679033977213`*^9, 3.733368137230505*^9, 3.792273856368531*^9, 
  3.794517943583797*^9, 3.7945180215434017`*^9, 3.79451825610491*^9, 
  3.7945187865240927`*^9, 3.7945190664129*^9, 3.79451930220175*^9, 
  3.794520027614374*^9, 3.7945202139105177`*^9, 3.794520322108407*^9, 
  3.794520423537464*^9, 3.794520593488057*^9, 3.794520669214313*^9},
 CellLabel->"During evaluation of In[162]:=",
 CellID->964838287,ExpressionUUID->"4d8efc83-7475-47bb-9cdf-0ba2f7996838"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Again, ToCTensor performs simple changes of basis with that information:\
\>", "MathCaption",
 CellChangeTimes->{{3.730955999744823*^9, 3.7309560265924253`*^9}},
 CellID->1606319613,ExpressionUUID->"f369526a-c35e-421c-a629-77290182e26d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCTensor", "[", 
  RowBox[{"v", ",", 
   RowBox[{"{", "sph3", "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.7309559866109037`*^9, 3.730955987081265*^9}},
 CellLabel->"In[163]:=",
 CellID->1812900123,ExpressionUUID->"c298d41b-3b49-41c1-84e0-b1ddd935186c"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     FractionBox["5", "7"], ",", 
     FractionBox["1", "7"]}], "}"}], ",", 
   RowBox[{"{", "sph3", "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.7309559898456182`*^9, 3.730956205049572*^9, 
  3.731116703850642*^9, 3.733367704544794*^9, 3.733367903767138*^9, 
  3.733368137797303*^9, 3.792273856633029*^9, 3.794517943802557*^9, 
  3.794518021729653*^9, 3.794518256325782*^9, 3.794518786766341*^9, 
  3.7945190665972223`*^9, 3.7945193024434013`*^9, 3.7945200278468018`*^9, 
  3.794520214160976*^9, 3.794520322364217*^9, 3.7945204237381477`*^9, 
  3.794520593688426*^9, 3.794520669262842*^9},
 CellLabel->"Out[163]=",
 CellID->388071112,ExpressionUUID->"c3e70acc-d37f-4a93-9f25-23c801a96cc0"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.2.6. Extracting components", "Subsubsection",
 CellChangeTimes->{{3.7333680347996473`*^9, 3.733368039055501*^9}},
 CellID->1961079369,ExpressionUUID->"526aaba5-f88e-4422-b189-ffe8f3a4dbec"],

Cell["\<\
Closely related to the problem of changes of bases we have that of extraction \
of particular components. If necessary, a change of basis will be performed \
automatically.\
\>", "Text",
 CellID->281940627,ExpressionUUID->"df4fcd98-f0e4-49a5-9920-af1889df2862"],

Cell[CellGroupData[{

Cell["Extract a component of the metric in the same bases:", "MathCaption",
 CellChangeTimes->{{3.728676167678884*^9, 3.728676167815362*^9}},
 CellID->1227192707,ExpressionUUID->"612e8563-4f91-4bcb-92e1-6e1a69045d72"],

Cell[CellGroupData[{

Cell[BoxData["met"], "Input",
 CellLabel->"In[164]:=",
 CellID->994880761,ExpressionUUID->"c1245b0c-51ad-45cc-9ed9-51325ead73e2"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", 
       SuperscriptBox[
        RowBox[{
         InterpretationBox[
          StyleBox["Sin",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Sin,
          Editable->False], "[", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False], "]"}], "2"]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "sph"}], ",", 
     RowBox[{"-", "sph"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.72867615026628*^9, 3.7287535042001266`*^9, 
  3.7288435590610743`*^9, 3.730956070294903*^9, 3.73095620514754*^9, 
  3.7311167184080753`*^9, 3.73336771020849*^9, 3.7333679038846407`*^9, 
  3.7333681379896507`*^9, 3.7922738566903973`*^9, 3.794517943861443*^9, 
  3.7945180217764673`*^9, 3.794518269242384*^9, 3.794518786816606*^9, 
  3.794519066645261*^9, 3.794519302493443*^9, 3.794520027897324*^9, 
  3.794520214213622*^9, 3.794520322409864*^9, 3.794520423789929*^9, 
  3.794520593740158*^9, 3.794520669321507*^9},
 CellLabel->"Out[164]=",
 CellID->403343445,ExpressionUUID->"ef15ec67-5611-4423-b449-6313e9d6063d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"met", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"2", ",", 
     RowBox[{"-", "sph"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"2", ",", 
     RowBox[{"-", "sph"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[165]:=",
 CellID->1649546631,ExpressionUUID->"ef71946e-cfa0-423f-a1ab-da230c81efb5"],

Cell[BoxData[
 SuperscriptBox[
  RowBox[{
   InterpretationBox[
    StyleBox["Sin",
     ShowAutoStyles->False,
     AutoSpacing->False],
    Sin,
    Editable->False], "[", 
   InterpretationBox[
    StyleBox["\[Theta]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Theta][],
    Editable->False], "]"}], "2"]], "Output",
 CellChangeTimes->{3.728676151418161*^9, 3.728753504332889*^9, 
  3.72884355926035*^9, 3.730956071452589*^9, 3.7309562052484407`*^9, 
  3.731116719507311*^9, 3.733367710377557*^9, 3.733367903999875*^9, 
  3.7333681381328907`*^9, 3.792273856750249*^9, 3.7945179439158382`*^9, 
  3.794518021820796*^9, 3.794518269294488*^9, 3.794518786858514*^9, 
  3.794519066690012*^9, 3.794519302538945*^9, 3.7945200279458103`*^9, 
  3.7945202142544603`*^9, 3.794520322459161*^9, 3.794520423831205*^9, 
  3.794520593787839*^9, 3.794520669367708*^9},
 CellLabel->"Out[165]=",
 CellID->1534890183,ExpressionUUID->"d769e564-959e-4261-a62c-ea10e6d0d961"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Extract a component in a different basis:", "MathCaption",
 CellChangeTimes->{3.72867617309473*^9},
 CellID->602909514,ExpressionUUID->"f186d316-a00a-432a-af88-43e6c99873b8"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"met", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"2", ",", 
     RowBox[{"-", "sph2"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"2", ",", 
     RowBox[{"-", "sph2"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[166]:=",
 CellID->1272355252,ExpressionUUID->"6e9ca28a-248d-42e7-abaa-82d8db1ca3c2"],

Cell[BoxData[
 RowBox[{"9", "+", 
  RowBox[{"4", " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}]}]], "Output",
 CellChangeTimes->{3.728676152385695*^9, 3.728753504467554*^9, 
  3.72884355942553*^9, 3.730956072769939*^9, 3.7309562053442793`*^9, 
  3.731116720624255*^9, 3.733367710511881*^9, 3.733367904134128*^9, 
  3.733368138301084*^9, 3.792273866221402*^9, 3.794517943974154*^9, 
  3.7945180218613157`*^9, 3.794518269342182*^9, 3.794518786903157*^9, 
  3.794519066741167*^9, 3.794519302587966*^9, 3.794520028013715*^9, 
  3.7945202143049173`*^9, 3.794520322502709*^9, 3.794520423882162*^9, 
  3.794520593834198*^9, 3.7945206694238653`*^9},
 CellLabel->"Out[166]=",
 CellID->1325737917,ExpressionUUID->"bb382c09-e9d0-44ce-88fb-7be6182ddb4c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToCTensor", "[", 
   RowBox[{"met", ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "sph2"}], ",", 
      RowBox[{"-", "sph2"}]}], "}"}]}], "]"}], "[", 
  RowBox[{
   RowBox[{"-", "\[ScriptA]"}], ",", 
   RowBox[{"-", "\[ScriptB]"}]}], "]"}]], "Input",
 CellLabel->"In[167]:=",
 CellID->1246444383,ExpressionUUID->"be007877-8bfa-4931-bb52-095d2851a0fb"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{"4", "+", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], "2"]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{"4", "+", 
                SuperscriptBox[
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                   InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                 "2"]}]}]],
           Annotation[#, {1, 1} -> 4 + Sin[
                $CellContext`\[Theta][]]^2, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{"-", "6"}], "+", 
             RowBox[{"2", " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["Sin",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 Sin,
                 Editable->False], "[", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False], "]"}], "2"]}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{"-", "6"}], "+", 
                RowBox[{"2", " ", 
                  SuperscriptBox[
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                   "2"]}]}]}]],
           Annotation[#, {1, 2} -> -6 + 2 Sin[
                 $CellContext`\[Theta][]]^2, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{"-", "6"}], "+", 
             RowBox[{"2", " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["Sin",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 Sin,
                 Editable->False], "[", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False], "]"}], "2"]}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{"-", "6"}], "+", 
                RowBox[{"2", " ", 
                  SuperscriptBox[
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                   "2"]}]}]}]],
           Annotation[#, {2, 1} -> -6 + 2 Sin[
                 $CellContext`\[Theta][]]^2, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{"9", "+", 
             RowBox[{"4", " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["Sin",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 Sin,
                 Editable->False], "[", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False], "]"}], "2"]}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{"9", "+", 
                RowBox[{"4", " ", 
                  SuperscriptBox[
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                   "2"]}]}]}]],
           Annotation[#, {2, 2} -> 9 + 4 Sin[
                 $CellContext`\[Theta][]]^2, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[0, 0, 1], 3 -> RGBColor[0, 0, 1]},
           "RowsIndexed" -> {1 -> RGBColor[0, 0, 1], 3 -> RGBColor[0, 0, 1]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " "},
         {"\[ScriptA]", "\[ScriptB]"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{4 + Sin[
        $CellContext`\[Theta][]]^2, -6 + 2 Sin[
         $CellContext`\[Theta][]]^2}, {-6 + 2 Sin[
         $CellContext`\[Theta][]]^2, 9 + 4 Sin[
         $CellContext`\[Theta][]]^2}}, {-$CellContext`sph2, \
-$CellContext`sph2}, 0][-$CellContext`\[ScriptA], -$CellContext`\[ScriptB]],
  Editable->False]], "Output",
 CellChangeTimes->{3.72867615293433*^9, 3.7287535045863943`*^9, 
  3.728843560079303*^9, 3.730956073563442*^9, 3.730956205452908*^9, 
  3.73111672128071*^9, 3.733367710630225*^9, 3.73336790422826*^9, 
  3.733368138455612*^9, 3.792273867335618*^9, 3.794517944033331*^9, 
  3.7945180219188147`*^9, 3.794518269392702*^9, 3.7945187869461803`*^9, 
  3.79451906679368*^9, 3.794519302644301*^9, 3.794520028070232*^9, 
  3.7945202143556023`*^9, 3.7945203225501537`*^9, 3.79452042395462*^9, 
  3.7945205938852882`*^9, 3.794520669487585*^9},
 CellLabel->"Out[167]=",
 CellID->1534333394,ExpressionUUID->"ef1a9e4e-275d-464f-bf15-68f92aa8d953"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["A mixture of bases:", "MathCaption",
 CellChangeTimes->{3.7286761762463017`*^9},
 CellID->393619928,ExpressionUUID->"3beab8ea-c782-45bb-8c68-49e22cb9e130"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"met", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", 
     RowBox[{"-", "sph"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"2", ",", 
     RowBox[{"-", "sph2"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[168]:=",
 CellID->1320497136,ExpressionUUID->"d36c6d43-00c0-4325-bc5b-23f029e1eb47"],

Cell[BoxData[
 RowBox[{"-", "3"}]], "Output",
 CellChangeTimes->{3.728676153887371*^9, 3.728753504698986*^9, 
  3.728843560423382*^9, 3.7309560751110373`*^9, 3.730956205547225*^9, 
  3.73111672240812*^9, 3.7333677109013577`*^9, 3.733367904351081*^9, 
  3.733368138599989*^9, 3.792273868405007*^9, 3.7945179440867167`*^9, 
  3.794518021964233*^9, 3.79451826943964*^9, 3.794518786990473*^9, 
  3.794519066838118*^9, 3.794519302686511*^9, 3.794520028121496*^9, 
  3.794520214400791*^9, 3.794520322594255*^9, 3.7945204239992933`*^9, 
  3.7945205939305487`*^9, 3.794520669536578*^9},
 CellLabel->"Out[168]=",
 CellID->557719782,ExpressionUUID->"cb6df262-0e60-4bce-9f8f-d2c6fe7661bb"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToCTensor", "[", 
   RowBox[{"met", ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "sph"}], ",", 
      RowBox[{"-", "sph2"}]}], "}"}]}], "]"}], "[", 
  RowBox[{
   RowBox[{"-", "\[ScriptA]"}], ",", 
   RowBox[{"-", "\[ScriptB]"}]}], "]"}]], "Input",
 CellLabel->"In[169]:=",
 CellID->1114169880,ExpressionUUID->"61f9a81f-85e7-4ae2-a7c4-b13772dc6be3"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["2",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "2"}]],
           Annotation[#, {1, 1} -> 2, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{"-", "3"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{"-", "3"}]}]],
           Annotation[#, {1, 2} -> -3, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["Sin",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Sin,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], "2"],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
              SuperscriptBox[
               RowBox[{
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], Sin, Editable -> False], "[", 
                 InterpretationBox[
                  StyleBox[
                  "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                   False], 
                  $CellContext`\[Theta][], Editable -> False], "]"}], "2"]}]],
           
           Annotation[#, {2, 1} -> Sin[
               $CellContext`\[Theta][]]^2, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{"2", " ", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], "2"]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{"2", " ", 
                SuperscriptBox[
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                   InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                 "2"]}]}]],
           Annotation[#, {2, 2} -> 2 Sin[
                $CellContext`\[Theta][]]^2, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[0, 0, 1], 3 -> RGBColor[0, 0, 1]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " "},
         {"\[ScriptA]", "\[ScriptB]"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{2, -3}, {Sin[
       $CellContext`\[Theta][]]^2, 2 Sin[
        $CellContext`\[Theta][]]^2}}, {-$CellContext`sph, -$CellContext`sph2},
    0][-$CellContext`\[ScriptA], -$CellContext`\[ScriptB]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728676154465638*^9, 3.7287535048348293`*^9, 
  3.728843560625567*^9, 3.7309560759641323`*^9, 3.730956205640316*^9, 
  3.731116723079719*^9, 3.733367711209918*^9, 3.733367904467661*^9, 
  3.733368138717321*^9, 3.792273868934956*^9, 3.794517944146864*^9, 
  3.794518022012059*^9, 3.79451826948984*^9, 3.794518787036248*^9, 
  3.794519066881344*^9, 3.7945193027412233`*^9, 3.7945200281781187`*^9, 
  3.794520214452921*^9, 3.794520322643379*^9, 3.794520424054429*^9, 
  3.79452059398307*^9, 3.79452066959299*^9},
 CellLabel->"Out[169]=",
 CellID->1977740566,ExpressionUUID->"482e077e-8a61-4736-b46e-a450f8ad931d"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We can extract a sub-CTensor expression:", "MathCaption",
 CellChangeTimes->{3.7286761795022078`*^9},
 CellID->1354926793,ExpressionUUID->"ddcecfa2-7427-49a6-b65e-b0bc78362f15"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"met", "[", 
  RowBox[{
   RowBox[{"-", "\[ScriptA]"}], ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", 
     RowBox[{"-", "sph2"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[170]:=",
 CellID->338017163,ExpressionUUID->"dcbad002-4fd1-4651-904b-d9793639b887"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["2",
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", "2"}]],
           Annotation[#, {1} -> 2, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["Sin",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Sin,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], "2"],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              SuperscriptBox[
               RowBox[{
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], Sin, Editable -> False], "[", 
                 InterpretationBox[
                  StyleBox[
                  "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                   False], 
                  $CellContext`\[Theta][], Editable -> False], "]"}], "2"]}]],
           
           Annotation[#, {2} -> Sin[
               $CellContext`\[Theta][]]^2, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" "},
         {"\[ScriptA]"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{2, Sin[
      $CellContext`\[Theta][]]^2}, {-$CellContext`sph}, 
   0][-$CellContext`\[ScriptA]],
  Editable->False]], "Output",
 CellChangeTimes->{3.7286761554657583`*^9, 3.728753504968709*^9, 
  3.7288435608127832`*^9, 3.730956077210737*^9, 3.730956205751337*^9, 
  3.7311167242458553`*^9, 3.733367711359771*^9, 3.7333679045668364`*^9, 
  3.7333681388388042`*^9, 3.79227387085375*^9, 3.794517944202299*^9, 
  3.7945180220561857`*^9, 3.794518269547967*^9, 3.79451878708084*^9, 
  3.794519066931431*^9, 3.794519302789768*^9, 3.7945200282325077`*^9, 
  3.7945202144990873`*^9, 3.794520322697138*^9, 3.794520424118733*^9, 
  3.794520594031258*^9, 3.794520669645368*^9},
 CellLabel->"Out[170]=",
 CellID->1133804496,ExpressionUUID->"509bc2dd-c7f8-4bb9-ba6f-ba00bc8c22cf"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"met", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"2", ",", 
     RowBox[{"-", "sph"}]}], "}"}], ",", 
   RowBox[{"-", "\[ScriptA]"}]}], "]"}]], "Input",
 CellLabel->"In[171]:=",
 CellID->2094758824,ExpressionUUID->"e9316541-8d42-41d5-bc92-dfc777533be4"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["Sin",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Sin,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], "2"],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              SuperscriptBox[
               RowBox[{
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], Sin, Editable -> False], "[", 
                 InterpretationBox[
                  StyleBox[
                  "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                   False], 
                  $CellContext`\[Theta][], Editable -> False], "]"}], "2"]}]],
           
           Annotation[#, {2} -> Sin[
               $CellContext`\[Theta][]]^2, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" "},
         {"\[ScriptA]"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{0, Sin[
      $CellContext`\[Theta][]]^2}, {-$CellContext`sph}, 
   0][-$CellContext`\[ScriptA]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728676156015222*^9, 3.728753505102223*^9, 
  3.7288435612492657`*^9, 3.7309560779826603`*^9, 3.730956205845302*^9, 
  3.73111672485462*^9, 3.733367711678566*^9, 3.733367904684621*^9, 
  3.733368138951*^9, 3.792273872804658*^9, 3.794517944259815*^9, 
  3.794518022105925*^9, 3.7945182695903378`*^9, 3.79451878712591*^9, 
  3.79451906697822*^9, 3.794519302838604*^9, 3.7945200282861*^9, 
  3.7945202145480223`*^9, 3.7945203227461567`*^9, 3.79452042417142*^9, 
  3.794520594079467*^9, 3.794520669701297*^9},
 CellLabel->"Out[171]=",
 CellID->1229764441,ExpressionUUID->"b3223866-21b2-4647-b26a-76550379a148"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Using b-indices will produce an expansion into basis objects:\
\>", "MathCaption",
 CellChangeTimes->{3.728676183838193*^9},
 CellID->864018702,ExpressionUUID->"02face2d-1d65-4346-8537-a6b804ba15cf"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"met", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"\[ScriptA]", ",", 
     RowBox[{"-", "sph"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"\[ScriptB]", ",", 
     RowBox[{"-", "sph"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[172]:=",
 CellID->277903500,ExpressionUUID->"7bc19b58-d17a-41ae-afbb-57367ef17c8e"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["\[ScriptA]",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{$CellContext`\[ScriptA], -$CellContext`sph}, {
     1, $CellContext`sph}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["\[ScriptB]",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{$CellContext`\[ScriptB], -$CellContext`sph}, {
     1, $CellContext`sph}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["\[ScriptA]",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{$CellContext`\[ScriptA], -$CellContext`sph}, {
     2, $CellContext`sph}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["\[ScriptB]",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{$CellContext`\[ScriptB], -$CellContext`sph}, {
     2, $CellContext`sph}],
    Editable->False], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}]}]], "Output",
 CellChangeTimes->{3.7286761891863747`*^9, 3.7287535052493477`*^9, 
  3.728843561645411*^9, 3.730956079444665*^9, 3.7309562059566*^9, 
  3.731116725879318*^9, 3.7333677118297663`*^9, 3.733367904816944*^9, 
  3.7333681390669727`*^9, 3.79227387445575*^9, 3.794517944319467*^9, 
  3.794518022151948*^9, 3.794518269652338*^9, 3.794518787178277*^9, 
  3.794519067034348*^9, 3.7945193028890343`*^9, 3.794520028350791*^9, 
  3.794520214599395*^9, 3.794520322801589*^9, 3.7945204242251368`*^9, 
  3.7945205941571407`*^9, 3.7945206697667217`*^9},
 CellLabel->"Out[172]=",
 CellID->1268341159,ExpressionUUID->"1eb397ff-d066-494e-a219-5f250524a12f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The same result can be obtained directly from the abstract-index expression \
using the following syntax:\
\>", "MathCaption",
 CellChangeTimes->{3.728676186334074*^9},
 CellID->1518795783,ExpressionUUID->"2f55925f-fb97-4f0b-b45b-e06dee9db503"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"met", "[", 
  RowBox[{
   RowBox[{"-", "\[ScriptA]"}], ",", 
   RowBox[{"-", "\[ScriptB]"}]}], "]"}]], "Input",
 CellLabel->"In[173]:=",
 CellID->1410096639,ExpressionUUID->"04673799-f987-40df-a745-4127362c5c90"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {1, 1} -> 1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 2} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["Sin",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Sin,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], "2"],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              SuperscriptBox[
               RowBox[{
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], Sin, Editable -> False], "[", 
                 InterpretationBox[
                  StyleBox[
                  "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                   False], 
                  $CellContext`\[Theta][], Editable -> False], "]"}], "2"]}]],
           
           Annotation[#, {2, 2} -> Sin[
               $CellContext`\[Theta][]]^2, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " "},
         {"\[ScriptA]", "\[ScriptB]"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{1, 0}, {0, Sin[
       $CellContext`\[Theta][]]^2}}, {-$CellContext`sph, -$CellContext`sph}, 
   0][-$CellContext`\[ScriptA], -$CellContext`\[ScriptB]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728676190352063*^9, 3.728753505369253*^9, 
  3.7288435618571043`*^9, 3.730956086914813*^9, 3.730956206066482*^9, 
  3.731116727112977*^9, 3.733367711943284*^9, 3.73336790498359*^9, 
  3.7333681391836777`*^9, 3.79227387725289*^9, 3.794517944379759*^9, 
  3.7945180221936493`*^9, 3.79451826969545*^9, 3.794518787221984*^9, 
  3.7945190670816517`*^9, 3.7945193029406967`*^9, 3.7945200284060793`*^9, 
  3.79452021464764*^9, 3.794520322850362*^9, 3.7945204242726383`*^9, 
  3.794520594218047*^9, 3.7945206698206778`*^9},
 CellLabel->"Out[173]=",
 CellID->1087179191,ExpressionUUID->"d291ecd7-b37b-4271-94bc-3159b42fb16b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ToBasisExpand"}]], "Input",
 CellLabel->"In[174]:=",
 CellID->2066250287,ExpressionUUID->"af6ca114-92bd-46eb-bc32-db8de7b18e11"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"\[ScriptA]", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`\[ScriptA], {1, $CellContext`sph}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"\[ScriptB]", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`\[ScriptB], {1, $CellContext`sph}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"\[ScriptA]", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`\[ScriptA], {2, $CellContext`sph}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"\[ScriptB]", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`\[ScriptB], {2, $CellContext`sph}],
    Editable->False], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}]}]], "Output",
 CellChangeTimes->{3.728676190952519*^9, 3.728753505514186*^9, 
  3.72884356205587*^9, 3.730956088443798*^9, 3.730956206160194*^9, 
  3.731116727828224*^9, 3.733367712076786*^9, 3.733367905138439*^9, 
  3.733368139335037*^9, 3.792273877824305*^9, 3.794517944452292*^9, 
  3.794518022242653*^9, 3.794518269750823*^9, 3.794518787274802*^9, 
  3.7945190671362047`*^9, 3.7945193029915113`*^9, 3.794520028472204*^9, 
  3.7945202147002077`*^9, 3.7945203229083347`*^9, 3.794520424327973*^9, 
  3.794520594272786*^9, 3.794520669883692*^9},
 CellLabel->"Out[174]=",
 CellID->1429794461,ExpressionUUID->"21807853-fe18-4c6b-9873-e084f2ca5b39"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "FromBasisExpand"}]], "Input",
 CellChangeTimes->{{3.792273882674738*^9, 3.79227388712003*^9}},
 CellLabel->"In[175]:=",
 CellID->1930809927,ExpressionUUID->"2bba8e08-d993-4528-8ebb-fd73ea30dacf"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {1, 1} -> 1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 2} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["Sin",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Sin,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], "2"],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              SuperscriptBox[
               RowBox[{
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], Sin, Editable -> False], "[", 
                 InterpretationBox[
                  StyleBox[
                  "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                   False], 
                  $CellContext`\[Theta][], Editable -> False], "]"}], "2"]}]],
           
           Annotation[#, {2, 2} -> Sin[
               $CellContext`\[Theta][]]^2, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " "},
         {"\[ScriptA]", "\[ScriptB]"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{1, 0}, {0, Sin[
       $CellContext`\[Theta][]]^2}}, {-$CellContext`sph, -$CellContext`sph}, 
   0][-$CellContext`\[ScriptA], -$CellContext`\[ScriptB]],
  Editable->False]], "Output",
 CellChangeTimes->{3.792273887537353*^9, 3.79451794451188*^9, 
  3.794518022294635*^9, 3.7945182698074627`*^9, 3.7945187873185062`*^9, 
  3.794519067182762*^9, 3.794519303040225*^9, 3.794520028527495*^9, 
  3.7945202147466288`*^9, 3.7945203229554033`*^9, 3.794520424375389*^9, 
  3.794520594323893*^9, 3.794520669937183*^9},
 CellLabel->"Out[175]=",
 CellID->213496097,ExpressionUUID->"6d4a7d53-a06f-4c33-b7e5-f166fc69f6f2"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.2.7. Cleanup", "Subsubsection",
 CellChangeTimes->{{3.733368036167544*^9, 3.733368040775607*^9}},
 CellID->1544931844,ExpressionUUID->"9219f830-e3e8-4579-bdbb-02abbf1b4098"],

Cell[BoxData[
 RowBox[{
  RowBox[{"UndefTensor", "/@", 
   RowBox[{"{", 
    RowBox[{"Jacobiansphsph3", ",", "Jacobiansphsph2"}], "}"}]}], 
  ";"}]], "Input",
 CellChangeTimes->{{3.7333677490020113`*^9, 3.733367821742803*^9}},
 CellLabel->"In[176]:=",
 CellID->309267742,ExpressionUUID->"d5fed17a-4ad1-4f5a-bfdd-72e34fdcb376"],

Cell[CellGroupData[{

Cell["Undefine S2:", "MathCaption",
 CellID->444726231,ExpressionUUID->"11f8b1df-d6d6-4dcd-b716-a32c7a95537f"],

Cell[BoxData[
 RowBox[{"UndefBasis", "[", "sph3", "]"}]], "Input",
 CellChangeTimes->{{3.730956113549554*^9, 3.730956119868907*^9}},
 CellLabel->"In[177]:=",
 CellID->2144554484,ExpressionUUID->"7e87df96-e388-4f29-a100-d5786daeb8dc"],

Cell[BoxData[
 RowBox[{"UndefBasis", "[", "sph2", "]"}]], "Input",
 CellLabel->"In[178]:=",
 CellID->1270672857,ExpressionUUID->"59f827e0-dbde-4fe9-a5a7-6c275e8a6887"],

Cell[BoxData[
 RowBox[{"UndefChart", "[", "sph", "]"}]], "Input",
 CellLabel->"In[179]:=",
 CellID->1621321831,ExpressionUUID->"7aee4012-f840-476a-9b08-a05d3ff8ad68"],

Cell[BoxData[
 RowBox[{"UndefTensor", "[", "S", "]"}]], "Input",
 CellLabel->"In[180]:=",
 CellID->136016834,ExpressionUUID->"8b27a859-5c79-4550-b010-52f9d40d36c1"],

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "S2", "]"}]], "Input",
 CellLabel->"In[181]:=",
 CellID->1112328630,ExpressionUUID->"9990bab1-c31b-4e75-ab66-4c61ce21646c"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Undefine M:", "MathCaption",
 CellID->1062137275,ExpressionUUID->"18c7abda-191f-4f10-b6bf-1cb3819bed1b"],

Cell[BoxData[
 RowBox[{"UndefBasis", "[", "B", "]"}]], "Input",
 CellLabel->"In[182]:=",
 CellID->1457107061,ExpressionUUID->"a367107f-1e83-4ff4-929d-53c8bf1b67bd"],

Cell[BoxData[
 RowBox[{"UndefTensor", "[", "s", "]"}]], "Input",
 CellLabel->"In[183]:=",
 CellID->1842347486,ExpressionUUID->"757e2dc0-c6cd-42af-a197-7f2fcb63f87c"],

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M", "]"}]], "Input",
 CellLabel->"In[184]:=",
 CellID->1914974756,ExpressionUUID->"33b676a7-8d85-486e-9251-de3765b20320"],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{"A", ",", "T", ",", "v", ",", "w", ",", "met", ",", "imet"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.731116742365241*^9, 3.731116742733302*^9}, {
  3.794518071650016*^9, 3.794518073842124*^9}},
 CellLabel->"In[185]:=",
 CellID->643306503,ExpressionUUID->"95ee7682-1e99-406e-9e54-cb8d7fb7e787"],

Cell[BoxData[
 RowBox[{"ClearxCobaCache", "[", "All", "]"}]], "Input",
 CellLabel->"In[186]:=",
 CellID->1619716923,ExpressionUUID->"9e8b6f7a-d5c9-4eb8-a9a9-d73fcce1ca33"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.3. Covariant derivatives", "Subsection",
 CellChangeTimes->{{3.728674033201705*^9, 3.72867403339421*^9}, {
  3.728675276464171*^9, 3.72867527661625*^9}},
 CellID->1079615489,ExpressionUUID->"303e3f2a-6f81-4934-ad85-b656a9fa266f"],

Cell[CellGroupData[{

Cell["3.3.1. CCovD expressions", "Subsubsection",
 CellChangeTimes->{{3.7333680532470837`*^9, 3.7333680536637363`*^9}},
 CellID->1008190120,ExpressionUUID->"a311af45-e61b-43f4-8098-ebc8efe9f940"],

Cell["\<\
The key idea is to represent a covariant derivative as another connection \
plus a Christoffel CTensor expressing the relation between those two. \
Optionally, we can also specify a metric CTensor, in case the connection is \
associated to it. Note that in this subsection we do not need to raise and \
lower indices automatically, and therefore there is no need to use SetCMetric.\
\>", "Text",
 CellChangeTimes->{{3.7311171338875713`*^9, 3.731117161142995*^9}},
 CellID->1005372010,ExpressionUUID->"f3d2a201-a843-436f-a567-17922544619f"],

Cell[CellGroupData[{

Cell["Define polar coordinates on the 2-plane:", "MathCaption",
 CellChangeTimes->{{3.728676251298978*^9, 3.7286762514513683`*^9}},
 CellID->886037474,ExpressionUUID->"0518d0c5-7325-4fe1-88bd-d80d5ab10d70"],

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"E2", ",", "2", ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f"}], "}"}]}], 
  "]"}]], "Input",
 CellLabel->"In[187]:=",
 CellID->980217220,ExpressionUUID->"1d878512-9fa2-4918-91d5-3ee3bc3b598c"],

Cell[BoxData[
 RowBox[{"DefChart", "[", 
  RowBox[{"polar", ",", "E2", ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", "2"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"r", "[", "]"}], ",", 
     RowBox[{"\[Theta]", "[", "]"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[188]:=",
 CellID->1163760403,ExpressionUUID->"540e6284-c53b-422d-b045-366cb90d08ae"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"met", "=", 
  RowBox[{"CTensor", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"1", ",", "0"}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{
         RowBox[{"r", "[", "]"}], "^", "2"}]}], "}"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "polar"}], ",", 
      RowBox[{"-", "polar"}]}], "}"}]}], "]"}]}]], "Input",
 CellLabel->"In[189]:=",
 CellID->40036345,ExpressionUUID->"0f79031b-f894-41fe-a68f-847113d5b085"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "polar"}], ",", 
     RowBox[{"-", "polar"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.72867625361474*^9, 3.7287535180929413`*^9, 
  3.72884359699434*^9, 3.7309562444929647`*^9, 3.7311168084497137`*^9, 
  3.733367958759943*^9, 3.733368140985314*^9, 3.7945179453786087`*^9, 
  3.794518023145912*^9, 3.794518313015523*^9, 3.79451842959923*^9, 
  3.794518788094989*^9, 3.794519068032642*^9, 3.7945206815400887`*^9},
 CellLabel->"Out[189]=",
 CellID->345888892,ExpressionUUID->"eafd919a-7318-443e-83f4-1bad751da22c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
This is the Levi-Civita connection of the metric. We see it has three \
arguments: the derivative of the coordinate basis, a Christoffel tensor and \
the metric itself.\
\>", "MathCaption",
 CellChangeTimes->{{3.728676255819002*^9, 3.728676255947146*^9}},
 CellID->780044460,ExpressionUUID->"a1b8f519-bad1-47ad-81cd-8a81346a5a68"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"cd", "=", 
  RowBox[{"CovDOfMetric", "[", "met", "]"}]}]], "Input",
 CellLabel->"In[190]:=",
 CellID->1967057932,ExpressionUUID->"635d181a-7fed-47fc-b646-47206876bf10"],

Cell[BoxData[
 RowBox[{"CCovD", "[", 
  RowBox[{"PDpolar", ",", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", 
           RowBox[{"-", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]}]}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", 
           FractionBox["1", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           FractionBox["1", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]], ",", "0"}], "}"}]}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"polar", ",", 
       RowBox[{"-", "polar"}], ",", 
       RowBox[{"-", "polar"}]}], "}"}], ",", "0"}], "]"}], ",", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"1", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"]}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "polar"}], ",", 
       RowBox[{"-", "polar"}]}], "}"}], ",", "0"}], "]"}]}], "]"}]], "Output",\

 CellChangeTimes->{3.728676257128284*^9, 3.728753518321683*^9, 
  3.7288435993772*^9, 3.7309562446591883`*^9, 3.731116869164805*^9, 
  3.733367959976047*^9, 3.733368141103273*^9, 3.794517945432181*^9, 
  3.7945180231938047`*^9, 3.7945183130682793`*^9, 3.7945184296440763`*^9, 
  3.794518788144382*^9, 3.7945190680863256`*^9, 3.794520681591814*^9},
 CellLabel->"Out[190]=",
 CellID->116375415,ExpressionUUID->"d0288911-1b53-4fcc-8a99-a8c20da04d6a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["This is the Christoffel tensor:", "MathCaption",
 CellChangeTimes->{{3.731116878751547*^9, 3.731116884320256*^9}},
 CellID->585314454,ExpressionUUID->"9250d531-2f51-470e-b70b-a574e90397c7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Christoffel", "[", 
  RowBox[{"cd", ",", "PDpolar"}], "]"}]], "Input",
 CellChangeTimes->{{3.731116886355649*^9, 3.731116891375905*^9}},
 CellLabel->"In[191]:=",
 CellID->1828426690,ExpressionUUID->"b5b9fe63-4b23-4b86-99a4-e04a53883d61"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", 
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False]}]}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", 
         FractionBox["1", 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False]]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         FractionBox["1", 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False]], ",", "0"}], "}"}]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"polar", ",", 
     RowBox[{"-", "polar"}], ",", 
     RowBox[{"-", "polar"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.731116893396291*^9, 3.7333679611443577`*^9, 
  3.733368141201846*^9, 3.794517945528283*^9, 3.794518023245412*^9, 
  3.7945183131107073`*^9, 3.794518429691092*^9, 3.794518788186665*^9, 
  3.794519068130102*^9, 3.794520681631995*^9},
 CellLabel->"Out[191]=",
 CellID->1711481135,ExpressionUUID->"9ce89680-c643-4711-8df1-9da04c901d58"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "[", 
  RowBox[{"a", ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.731116895916285*^9, 3.731116899383636*^9}},
 CellLabel->"In[192]:=",
 CellID->1416228891,ExpressionUUID->"9a8339cd-5b0c-47c2-9ef8-d1a98816d304"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{"-", 
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}]}]],
              Annotation[#, {1, 2, 2} -> -$CellContext`r[], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               FractionBox["1", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                 FractionBox["1", 
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False]]}]],
              
              Annotation[#, {2, 1, 2} -> $CellContext`r[]^(-1), 
               "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox[
               FractionBox["1", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                 FractionBox["1", 
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False]]}]],
              Annotation[#, {2, 2, 1} -> $CellContext`r[]^(-1), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a", " ", " "},
         {" ", "b", "c"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{{0, 0}, {0, -$CellContext`r[]}}, {{
     0, $CellContext`r[]^(-1)}, {$CellContext`r[]^(-1), 
      0}}}, {$CellContext`polar, -$CellContext`polar, -$CellContext`polar}, 
   0][$CellContext`a, -$CellContext`b, -$CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{3.731116900333343*^9, 3.7333679618166018`*^9, 
  3.7333681413184423`*^9, 3.794517945579503*^9, 3.7945180232910137`*^9, 
  3.794518313166615*^9, 3.79451842974271*^9, 3.7945187882405033`*^9, 
  3.794519068179557*^9, 3.794520681683372*^9},
 CellLabel->"Out[192]=",
 CellID->1467468027,ExpressionUUID->"30acdedf-54f3-46fc-90bf-dca3ce1e0562"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToBasisExpand", "[", "%", "]"}]], "Input",
 CellChangeTimes->{{3.7311169160021772`*^9, 3.7311169202403603`*^9}},
 CellLabel->"In[193]:=",
 CellID->1960374463,ExpressionUUID->"6a621bd8-023a-43c5-9186-01591447111a"],

Cell[BoxData[
 RowBox[{
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`polar}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`polar}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`polar}, $CellContext`a],
     Editable->False]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`polar}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`polar}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`polar}, $CellContext`a],
     Editable->False]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`polar}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`polar}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`polar}, $CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.731116923840803*^9, 3.733367962512498*^9, 
  3.733368141468001*^9, 3.794517945622877*^9, 3.794518023342811*^9, 
  3.794518313223154*^9, 3.794518429796343*^9, 3.794518788290374*^9, 
  3.79451906824833*^9, 3.794520681749497*^9},
 CellLabel->"Out[193]=",
 CellID->2032927641,ExpressionUUID->"5fbcb91b-8364-418f-a469-ff3ff42ad5f1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
This derivative operator can be now used normally to perform operations:\
\>", "MathCaption",
 CellChangeTimes->{{3.728676258242752*^9, 3.728676258371073*^9}},
 CellID->1126893565,ExpressionUUID->"380feb36-a0e0-4718-b325-664637eb0555"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"cd", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{
   RowBox[{"r", "[", "]"}], "^", "3"}], "]"}]], "Input",
 CellLabel->"In[194]:=",
 CellID->1041864062,ExpressionUUID->"18f5f0fb-a7d8-4065-8c76-15b8f31bd024"],

Cell[BoxData[
 RowBox[{"3", " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]]},
          {"a", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`polar}],
   Editable->False], " ", 
  SuperscriptBox[
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False], "2"]}]], "Output",
 CellChangeTimes->{3.72867625936546*^9, 3.7287535185013657`*^9, 
  3.728843600660553*^9, 3.730956244764226*^9, 3.73111696054456*^9, 
  3.733367964073243*^9, 3.733368141584462*^9, 3.794517945663233*^9, 
  3.794518023385962*^9, 3.79451831326504*^9, 3.7945184298428373`*^9, 
  3.794518788330023*^9, 3.794519068290103*^9, 3.794520681801689*^9},
 CellLabel->"Out[194]=",
 CellID->950038517,ExpressionUUID->"61239a10-8bf7-476e-9b94-7742de5c261c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"cd", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "0"}], "}"}], ",", 
     RowBox[{"{", "polar", "}"}]}], "]"}], "[", "b", "]"}], "]"}]], "Input",
 CellLabel->"In[195]:=",
 CellID->561211820,ExpressionUUID->"f3c48a0a-97d7-4769-8644-d5f2a2e34d9c"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 2} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox["1", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              FractionBox["1", 
               InterpretationBox[
                StyleBox[
                "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                $CellContext`r[], Editable -> False]]}]],
           Annotation[#, {2, 2} -> $CellContext`r[]^(-1), "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"b", " "},
         {" ", "a"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{0, 0}, {
    0, $CellContext`r[]^(-1)}}, {$CellContext`polar, -$CellContext`polar}, 
   0][$CellContext`b, -$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.728676260616864*^9, 3.728753519025847*^9, 
  3.7288436014784803`*^9, 3.730956244912465*^9, 3.731116963809124*^9, 
  3.733367964628858*^9, 3.733368141701929*^9, 3.794517945701223*^9, 
  3.794518023429537*^9, 3.794518313312235*^9, 3.79451842989627*^9, 
  3.7945187883772163`*^9, 3.794519068332987*^9, 3.794520681848521*^9},
 CellLabel->"Out[195]=",
 CellID->756439774,ExpressionUUID->"beffe58f-1617-47ad-9e4f-4011c1a766b3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"cd", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"met", "[", 
   RowBox[{
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}]}], "]"}], "]"}]], "Input",
 CellLabel->"In[196]:=",
 CellID->1036377213,ExpressionUUID->"b8c2f8a5-1953-4d93-8580-5bfeaef9fea4"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728676261217165*^9, 3.728753519338952*^9, 
  3.728843602261718*^9, 3.7309562450590153`*^9, 3.731116967006864*^9, 
  3.733367965344412*^9, 3.7333681418180227`*^9, 3.794517945740408*^9, 
  3.7945180234791613`*^9, 3.794518313355915*^9, 3.794518429940868*^9, 
  3.7945187884172983`*^9, 3.794519068372529*^9, 3.7945206818909607`*^9},
 CellLabel->"Out[196]=",
 CellID->380091378,ExpressionUUID->"2b479b1c-f48b-4820-bd73-0546431dbb3e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"cd", "[", 
   RowBox[{"-", "a"}], "]"}], "@", 
  RowBox[{
   RowBox[{"cd", "[", 
    RowBox[{"-", "b"}], "]"}], "@", 
   RowBox[{
    RowBox[{"CTensor", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"r", "[", "]"}]}], "}"}], ",", 
      RowBox[{"{", "polar", "}"}]}], "]"}], "[", "c", "]"}]}]}]], "Input",
 CellLabel->"In[197]:=",
 CellID->601293435,ExpressionUUID->"2d41266b-b549-4bf5-b5f2-8ff42df558c4"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{"-", 
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}]}]],
              Annotation[#, {1, 1, 2} -> -$CellContext`r[], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{"-", 
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}]}]],
              Annotation[#, {1, 2, 1} -> -$CellContext`r[], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox[
               FractionBox["2", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                 FractionBox["2", 
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False]]}]],
              Annotation[#, {2, 1, 1} -> 2/$CellContext`r[], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                 InterpretationBox[
                  StyleBox[
                  "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`r[], Editable -> False]}]],
              Annotation[#, {2, 2, 2} -> $CellContext`r[], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"c", " ", " "},
         {" ", "b", "a"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{{0, -$CellContext`r[]}, {-$CellContext`r[], 0}}, {{
     2/$CellContext`r[], 0}, {0, 
      $CellContext`r[]}}}, {$CellContext`polar, -$CellContext`polar, \
-$CellContext`polar}, 0][$CellContext`c, -$CellContext`b, -$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.728676261784699*^9, 3.728753519692444*^9, 
  3.7288436030119247`*^9, 3.7309562451651773`*^9, 3.731116975473469*^9, 
  3.7333679659226103`*^9, 3.7333681419682693`*^9, 3.794517945781642*^9, 
  3.7945180235248203`*^9, 3.794518313400462*^9, 3.79451842999354*^9, 
  3.7945187884620037`*^9, 3.794519068417754*^9, 3.79452068194427*^9},
 CellLabel->"Out[197]=",
 CellID->1579423868,ExpressionUUID->"3b86da22-7c13-414f-9d40-56d3451996e5"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ToCCanonical"}]], "Input",
 CellLabel->"In[198]:=",
 CellID->998379539,ExpressionUUID->"a05281f7-f839-4be5-ba0d-6508a6201bb8"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               FractionBox["2", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                 FractionBox["2", 
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False]]}]],
              Annotation[#, {1, 1, 2} -> 2/$CellContext`r[], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{"-", 
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}]}]],
              Annotation[#, {1, 2, 1} -> -$CellContext`r[], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{"-", 
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}]}]],
              Annotation[#, {2, 1, 1} -> -$CellContext`r[], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                 InterpretationBox[
                  StyleBox[
                  "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`r[], Editable -> False]}]],
              Annotation[#, {2, 2, 2} -> $CellContext`r[], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " ", "c"},
         {"a", "b", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{{
     0, 2/$CellContext`r[]}, {-$CellContext`r[], 0}}, {{-$CellContext`r[], 
      0}, {0, 
      $CellContext`r[]}}}, {-$CellContext`polar, -$CellContext`polar, \
$CellContext`polar}, 0][-$CellContext`a, -$CellContext`b, $CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{3.728676262661229*^9, 3.7287535199601107`*^9, 
  3.728843604463306*^9, 3.7309562454234467`*^9, 3.7311169826728277`*^9, 
  3.733367966628282*^9, 3.733368142084565*^9, 3.794517945830538*^9, 
  3.79451802357613*^9, 3.794518313448743*^9, 3.794518430042778*^9, 
  3.794518788506302*^9, 3.7945190684643507`*^9, 3.7945206819971447`*^9},
 CellLabel->"Out[198]=",
 CellID->331588909,ExpressionUUID->"fe96ea97-bf6e-4169-98f7-4aadae456a3d"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
CCovD derivatives are typeset with the \[Gradient] character, and have a \
tooltip:\
\>", "MathCaption",
 CellChangeTimes->{{3.728676264282476*^9, 3.7286762644093924`*^9}},
 CellID->617302719,ExpressionUUID->"9abb7009-df38-4644-89f9-09062863ab06"],

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"v", "[", "a", "]"}], ",", "E2"}], "]"}]], "Input",
 CellLabel->"In[199]:=",
 CellID->513299576,ExpressionUUID->"ddcae87a-6e7c-46c9-89c9-755774024044"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"cd", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"v", "[", "b", "]"}], "]"}]], "Input",
 CellLabel->"In[200]:=",
 CellID->584163833,ExpressionUUID->"289e6d71-6989-4300-b578-5f04be3ad456"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox[
     TooltipBox["\[Gradient]",
      RowBox[{"CCovD", "[", "PDpolar", ",", 
        StyleBox[
         GridBox[{{
            GridBox[{{
               TagBox[
                TooltipBox["0", 
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]], Annotation[#, {1, 1, 1} -> 0, "Tooltip"]& ]}, {
               TagBox[
                TooltipBox["0", 
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                   "0"}]], Annotation[#, {1, 1, 2} -> 0, "Tooltip"]& ]}}, 
             AutoDelete -> False, 
             GridBoxDividers -> {
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], 
            GridBox[{{
               TagBox[
                TooltipBox["0", 
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]], Annotation[#, {1, 2, 1} -> 0, "Tooltip"]& ]}, {
               TagBox[
                TooltipBox[
                 RowBox[{"-", 
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], 
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                   RowBox[{"-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}]}]], 
                Annotation[#, {1, 2, 2} -> -$CellContext`r[], "Tooltip"]& ]}},
              AutoDelete -> False, 
             GridBoxDividers -> {
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}, {
            GridBox[{{
               TagBox[
                TooltipBox["0", 
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]], Annotation[#, {2, 1, 1} -> 0, "Tooltip"]& ]}, {
               TagBox[
                TooltipBox[
                 FractionBox["1", 
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False]], 
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                   FractionBox["1", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]], 
                Annotation[#, {2, 1, 2} -> $CellContext`r[]^(-1), 
                 "Tooltip"]& ]}}, AutoDelete -> False, 
             GridBoxDividers -> {
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], 
            GridBox[{{
               TagBox[
                TooltipBox[
                 FractionBox["1", 
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False]], 
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                   FractionBox["1", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]], 
                Annotation[#, {2, 2, 1} -> $CellContext`r[]^(-1), 
                 "Tooltip"]& ]}, {
               TagBox[
                TooltipBox["0", 
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                   "0"}]], Annotation[#, {2, 2, 2} -> 0, "Tooltip"]& ]}}, 
             AutoDelete -> False, 
             GridBoxDividers -> {
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}}, 
          AutoDelete -> False, 
          GridBoxDividers -> {
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], Small], ",", 
        StyleBox[
         GridBox[{{
            TagBox[
             TooltipBox["1", 
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "1"}]], 
             Annotation[#, {1, 1} -> 1, "Tooltip"]& ], 
            TagBox[
             TooltipBox["0", 
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]], 
             Annotation[#, {1, 2} -> 0, "Tooltip"]& ]}, {
            TagBox[
             TooltipBox["0", 
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]], 
             Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              SuperscriptBox[
               InterpretationBox[
                StyleBox[
                "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                $CellContext`r[], Editable -> False], "2"], 
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                  "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`r[], Editable -> False], "2"]}]], 
             Annotation[#, {2, 2} -> $CellContext`r[]^2, "Tooltip"]& ]}}, 
          AutoDelete -> False, 
          GridBoxDividers -> {
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], Small], 
        "]"}]], "a"], GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"b"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CCovD[$CellContext`PDpolar, 
   xAct`xCoba`CTensor[{{{0, 0}, {0, -$CellContext`r[]}}, {{
      0, $CellContext`r[]^(-1)}, {$CellContext`r[]^(-1), 
       0}}}, {$CellContext`polar, -$CellContext`polar, -$CellContext`polar}, 
    0], 
   xAct`xCoba`CTensor[{{1, 0}, {
     0, $CellContext`r[]^2}}, {-$CellContext`polar, -$CellContext`polar}, 
    0]][-$CellContext`a][
   $CellContext`v[$CellContext`b]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728676266146119*^9, 3.728753520497919*^9, 
  3.728843606462552*^9, 3.7309562457493067`*^9, 3.73111701985529*^9, 
  3.733367968562771*^9, 3.733368142334927*^9, 3.794517945897682*^9, 
  3.7945180236452503`*^9, 3.7945183135131397`*^9, 3.79451843012082*^9, 
  3.7945187885723143`*^9, 3.79451906855208*^9, 3.794520682071867*^9},
 CellLabel->"Out[200]=",
 CellID->294729598,ExpressionUUID->"47921bcd-1696-4873-b04a-9ff3dea885d5"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Expansion is not automatic, as usual in xTensor, and must be asked for \
explicitly:\
\>", "MathCaption",
 CellChangeTimes->{{3.728676267874239*^9, 3.7286762680026903`*^9}},
 CellID->1086742571,ExpressionUUID->"ad1a2c82-4caa-429b-ac00-0248150af90c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ChangeCovD", "[", 
  RowBox[{"%", ",", "cd", ",", "PDpolar"}], "]"}]], "Input",
 CellLabel->"In[201]:=",
 CellID->533973062,ExpressionUUID->"c3694b3d-7a66-4720-a121-681ced691f24"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        StyleBox[GridBox[{
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {1, 1, 1} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {1, 1, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {1, 2, 1} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 RowBox[{"-", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                   RowBox[{"-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}]}]],
                Annotation[#, {1, 2, 2} -> -$CellContext`r[], "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {2, 1, 1} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 FractionBox["1", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                   FractionBox["1", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]],
                
                Annotation[#, {2, 1, 2} -> $CellContext`r[]^(-1), 
                 "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox[
                 FractionBox["1", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                   FractionBox["1", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]],
                
                Annotation[#, {2, 2, 1} -> $CellContext`r[]^(-1), 
                 "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {2, 2, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
         FontSize->Small], 
        StyleBox[GridBox[{
           {"b", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xCoba`CTensor[{{{0, 0}, {0, -$CellContext`r[]}}, {{
       0, $CellContext`r[]^(-1)}, {$CellContext`r[]^(-1), 
        0}}}, {$CellContext`polar, -$CellContext`polar, -$CellContext`polar}, 
     0][$CellContext`b, -$CellContext`a, -$CellContext`c],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox[
      StyleBox["\[ScriptCapitalD]",
       FontColor->RGBColor[1., 0., 0.]], "a"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"b"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`PDpolar[-$CellContext`a][
    $CellContext`v[$CellContext`b]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728676268928995*^9, 3.72875352082728*^9, 
  3.7288436079072933`*^9, 3.730956245892069*^9, 3.7311170324205227`*^9, 
  3.733367969623474*^9, 3.733368142450653*^9, 3.794517945947196*^9, 
  3.794518023701556*^9, 3.794518313563718*^9, 3.794518430177445*^9, 
  3.794518788624082*^9, 3.794519068598064*^9, 3.7945206821300087`*^9},
 CellLabel->"Out[201]=",
 CellID->1423063604,ExpressionUUID->"e50f1d88-cf2d-4bd8-9eff-99ae693811d8"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ChangeCovD", "[", 
  RowBox[{"%", ",", "PDpolar", ",", "cd"}], "]"}]], "Input",
 CellLabel->"In[202]:=",
 CellID->1081287593,ExpressionUUID->"60753910-c654-4c48-aaba-906fb3854db3"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        StyleBox[GridBox[{
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {1, 1, 1} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {1, 1, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {1, 2, 1} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 RowBox[{"-", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                   RowBox[{"-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}]}]],
                Annotation[#, {1, 2, 2} -> -$CellContext`r[], "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {2, 1, 1} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 FractionBox["1", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                   FractionBox["1", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]],
                
                Annotation[#, {2, 1, 2} -> $CellContext`r[]^(-1), 
                 "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox[
                 FractionBox["1", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                   FractionBox["1", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]],
                
                Annotation[#, {2, 2, 1} -> $CellContext`r[]^(-1), 
                 "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {2, 2, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
         FontSize->Small], 
        StyleBox[GridBox[{
           {"b", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xCoba`CTensor[{{{0, 0}, {0, -$CellContext`r[]}}, {{
       0, $CellContext`r[]^(-1)}, {$CellContext`r[]^(-1), 
        0}}}, {$CellContext`polar, -$CellContext`polar, -$CellContext`polar}, 
     0][$CellContext`b, -$CellContext`a, -$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        StyleBox[GridBox[{
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {1, 1, 1} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {1, 1, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {1, 2, 1} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}]],
                Annotation[#, {1, 2, 2} -> $CellContext`r[], "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {2, 1, 1} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 RowBox[{"-", 
                  FractionBox["1", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                   RowBox[{"-", 
                    FractionBox["1", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]}]],
                
                Annotation[#, {2, 1, 2} -> (-1)/$CellContext`r[], 
                 "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox[
                 RowBox[{"-", 
                  FractionBox["1", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                   RowBox[{"-", 
                    FractionBox["1", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]}]],
                Annotation[#, {2, 2, 1} -> (-1)/$CellContext`r[], 
                 "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {2, 2, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
         FontSize->Small], 
        StyleBox[GridBox[{
           {"b", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xCoba`CTensor[{{{0, 0}, {0, 
        $CellContext`r[]}}, {{
       0, (-1)/$CellContext`r[]}, {(-1)/$CellContext`r[], 
        0}}}, {$CellContext`polar, -$CellContext`polar, -$CellContext`polar}, 
     0][$CellContext`b, -$CellContext`a, -$CellContext`c],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox[
      TooltipBox["\[Gradient]",
       RowBox[{"CCovD", "[", "PDpolar", ",", 
         StyleBox[
          GridBox[{{
             GridBox[{{
                TagBox[
                 TooltipBox["0", 
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    "0"}]], Annotation[#, {1, 1, 1} -> 0, "Tooltip"]& ]}, {
                TagBox[
                 TooltipBox["0", 
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    "0"}]], Annotation[#, {1, 1, 2} -> 0, "Tooltip"]& ]}}, 
              AutoDelete -> False, 
              GridBoxDividers -> {
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], 
             GridBox[{{
                TagBox[
                 TooltipBox["0", 
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    "0"}]], Annotation[#, {1, 2, 1} -> 0, "Tooltip"]& ]}, {
                TagBox[
                 TooltipBox[
                  RowBox[{"-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], 
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    RowBox[{"-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}]}]], 
                 Annotation[#, {1, 2, 2} -> -$CellContext`r[], 
                  "Tooltip"]& ]}}, AutoDelete -> False, 
              GridBoxDividers -> {
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}, {
             GridBox[{{
                TagBox[
                 TooltipBox["0", 
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    "0"}]], Annotation[#, {2, 1, 1} -> 0, "Tooltip"]& ]}, {
                TagBox[
                 TooltipBox[
                  FractionBox["1", 
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]], 
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox["1", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]], 
                 Annotation[#, {2, 1, 2} -> $CellContext`r[]^(-1), 
                  "Tooltip"]& ]}}, AutoDelete -> False, 
              GridBoxDividers -> {
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], 
             GridBox[{{
                TagBox[
                 TooltipBox[
                  FractionBox["1", 
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]], 
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox["1", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]], 
                 Annotation[#, {2, 2, 1} -> $CellContext`r[]^(-1), 
                  "Tooltip"]& ]}, {
                TagBox[
                 TooltipBox["0", 
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    "0"}]], Annotation[#, {2, 2, 2} -> 0, "Tooltip"]& ]}}, 
              AutoDelete -> False, 
              GridBoxDividers -> {
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}}, 
           AutoDelete -> False, 
           GridBoxDividers -> {
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], Small], ",", 
         
         StyleBox[
          GridBox[{{
             TagBox[
              TooltipBox["1", 
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "1"}]], 
              Annotation[#, {1, 1} -> 1, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0", 
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]], 
              Annotation[#, {1, 2} -> 0, "Tooltip"]& ]}, {
             TagBox[
              TooltipBox["0", 
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]], 
              Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`r[], Editable -> False], "2"], 
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False], "2"]}]], 
              Annotation[#, {2, 2} -> $CellContext`r[]^2, "Tooltip"]& ]}}, 
           AutoDelete -> False, 
           GridBoxDividers -> {
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], Small], 
         "]"}]], "a"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"b"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xCoba`CCovD[$CellContext`PDpolar, 
    xAct`xCoba`CTensor[{{{0, 0}, {0, -$CellContext`r[]}}, {{
       0, $CellContext`r[]^(-1)}, {$CellContext`r[]^(-1), 
        0}}}, {$CellContext`polar, -$CellContext`polar, -$CellContext`polar}, 
     0], 
    xAct`xCoba`CTensor[{{1, 0}, {
      0, $CellContext`r[]^2}}, {-$CellContext`polar, -$CellContext`polar}, 
     0]][-$CellContext`a][
    $CellContext`v[$CellContext`b]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728676270262701*^9, 3.728753521259163*^9, 
  3.728843609123123*^9, 3.730956246086542*^9, 3.731117044789054*^9, 
  3.733367970427462*^9, 3.7333681425647707`*^9, 3.794517946005885*^9, 
  3.79451802376269*^9, 3.794518313622086*^9, 3.7945184302367*^9, 
  3.794518788680485*^9, 3.7945190686543293`*^9, 3.794520682195528*^9},
 CellLabel->"Out[202]=",
 CellID->904615789,ExpressionUUID->"4d9afb6d-4035-4aa1-8d8a-9f8f57016ac5"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ToCanonical"}]], "Input",
 CellLabel->"In[203]:=",
 CellID->708386491,ExpressionUUID->"3844f8cb-d448-48ef-bb3d-79e0405b15ab"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        StyleBox[GridBox[{
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {1, 1, 1} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {1, 1, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {1, 2, 1} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 RowBox[{"-", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                   RowBox[{"-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}]}]],
                Annotation[#, {1, 2, 2} -> -$CellContext`r[], "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {2, 1, 1} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 FractionBox["1", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                   FractionBox["1", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]],
                
                Annotation[#, {2, 1, 2} -> $CellContext`r[]^(-1), 
                 "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox[
                 FractionBox["1", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                   FractionBox["1", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]],
                
                Annotation[#, {2, 2, 1} -> $CellContext`r[]^(-1), 
                 "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {2, 2, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
         FontSize->Small], 
        StyleBox[GridBox[{
           {"b", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xCoba`CTensor[{{{0, 0}, {0, -$CellContext`r[]}}, {{
       0, $CellContext`r[]^(-1)}, {$CellContext`r[]^(-1), 
        0}}}, {$CellContext`polar, -$CellContext`polar, -$CellContext`polar}, 
     0][$CellContext`b, -$CellContext`a, -$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        StyleBox[GridBox[{
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {1, 1, 1} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {1, 1, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {1, 2, 1} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}]],
                Annotation[#, {1, 2, 2} -> $CellContext`r[], "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {2, 1, 1} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 RowBox[{"-", 
                  FractionBox["1", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                   RowBox[{"-", 
                    FractionBox["1", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]}]],
                
                Annotation[#, {2, 1, 2} -> (-1)/$CellContext`r[], 
                 "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox[
                 RowBox[{"-", 
                  FractionBox["1", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                   RowBox[{"-", 
                    FractionBox["1", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]}]],
                Annotation[#, {2, 2, 1} -> (-1)/$CellContext`r[], 
                 "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {2, 2, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
         FontSize->Small], 
        StyleBox[GridBox[{
           {"b", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xCoba`CTensor[{{{0, 0}, {0, 
        $CellContext`r[]}}, {{
       0, (-1)/$CellContext`r[]}, {(-1)/$CellContext`r[], 
        0}}}, {$CellContext`polar, -$CellContext`polar, -$CellContext`polar}, 
     0][$CellContext`b, -$CellContext`a, -$CellContext`c],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox[
      TooltipBox["\[Gradient]",
       RowBox[{"CCovD", "[", "PDpolar", ",", 
         StyleBox[
          GridBox[{{
             GridBox[{{
                TagBox[
                 TooltipBox["0", 
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    "0"}]], Annotation[#, {1, 1, 1} -> 0, "Tooltip"]& ]}, {
                TagBox[
                 TooltipBox["0", 
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    "0"}]], Annotation[#, {1, 1, 2} -> 0, "Tooltip"]& ]}}, 
              AutoDelete -> False, 
              GridBoxDividers -> {
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], 
             GridBox[{{
                TagBox[
                 TooltipBox["0", 
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    "0"}]], Annotation[#, {1, 2, 1} -> 0, "Tooltip"]& ]}, {
                TagBox[
                 TooltipBox[
                  RowBox[{"-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], 
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    RowBox[{"-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}]}]], 
                 Annotation[#, {1, 2, 2} -> -$CellContext`r[], 
                  "Tooltip"]& ]}}, AutoDelete -> False, 
              GridBoxDividers -> {
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}, {
             GridBox[{{
                TagBox[
                 TooltipBox["0", 
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    "0"}]], Annotation[#, {2, 1, 1} -> 0, "Tooltip"]& ]}, {
                TagBox[
                 TooltipBox[
                  FractionBox["1", 
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]], 
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox["1", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]], 
                 Annotation[#, {2, 1, 2} -> $CellContext`r[]^(-1), 
                  "Tooltip"]& ]}}, AutoDelete -> False, 
              GridBoxDividers -> {
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], 
             GridBox[{{
                TagBox[
                 TooltipBox[
                  FractionBox["1", 
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]], 
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox["1", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]], 
                 Annotation[#, {2, 2, 1} -> $CellContext`r[]^(-1), 
                  "Tooltip"]& ]}, {
                TagBox[
                 TooltipBox["0", 
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    "0"}]], Annotation[#, {2, 2, 2} -> 0, "Tooltip"]& ]}}, 
              AutoDelete -> False, 
              GridBoxDividers -> {
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}}, 
           AutoDelete -> False, 
           GridBoxDividers -> {
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], Small], ",", 
         
         StyleBox[
          GridBox[{{
             TagBox[
              TooltipBox["1", 
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "1"}]], 
              Annotation[#, {1, 1} -> 1, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0", 
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]], 
              Annotation[#, {1, 2} -> 0, "Tooltip"]& ]}, {
             TagBox[
              TooltipBox["0", 
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]], 
              Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`r[], Editable -> False], "2"], 
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False], "2"]}]], 
              Annotation[#, {2, 2} -> $CellContext`r[]^2, "Tooltip"]& ]}}, 
           AutoDelete -> False, 
           GridBoxDividers -> {
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], Small], 
         "]"}]], "a"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"b"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xCoba`CCovD[$CellContext`PDpolar, 
    xAct`xCoba`CTensor[{{{0, 0}, {0, -$CellContext`r[]}}, {{
       0, $CellContext`r[]^(-1)}, {$CellContext`r[]^(-1), 
        0}}}, {$CellContext`polar, -$CellContext`polar, -$CellContext`polar}, 
     0], 
    xAct`xCoba`CTensor[{{1, 0}, {
      0, $CellContext`r[]^2}}, {-$CellContext`polar, -$CellContext`polar}, 
     0]][-$CellContext`a][
    $CellContext`v[$CellContext`b]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7286762709076557`*^9, 3.728753521442565*^9, 
  3.7288436099758387`*^9, 3.730956246254142*^9, 3.731117049438489*^9, 
  3.7333679712169857`*^9, 3.73336814271542*^9, 3.794517946069676*^9, 
  3.7945180238333807`*^9, 3.794518313696665*^9, 3.7945184303231297`*^9, 
  3.794518788751108*^9, 3.794519068722014*^9, 3.7945206822742863`*^9},
 CellLabel->"Out[203]=",
 CellID->1840828610,ExpressionUUID->"045d9e57-cbce-4307-bde3-fd87322684a0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
[This is an interesting problem. The two CTensor expresions should cancel, \
but they don\[CloseCurlyQuote]t (TODO). We need to force them together:]\
\>", "MathCaption",
 CellID->379329235,ExpressionUUID->"357f43ed-9395-40fa-a355-f74dae9fb991"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Collect", "[", 
  RowBox[{"%", ",", "_v", ",", "Simplify"}], "]"}]], "Input",
 CellLabel->"In[204]:=",
 CellID->1654988349,ExpressionUUID->"a1780918-5795-4c0a-a85d-3c085fe2ec4d"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox[
     TooltipBox["\[Gradient]",
      RowBox[{"CCovD", "[", "PDpolar", ",", 
        StyleBox[
         GridBox[{{
            GridBox[{{
               TagBox[
                TooltipBox["0", 
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]], Annotation[#, {1, 1, 1} -> 0, "Tooltip"]& ]}, {
               TagBox[
                TooltipBox["0", 
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                   "0"}]], Annotation[#, {1, 1, 2} -> 0, "Tooltip"]& ]}}, 
             AutoDelete -> False, 
             GridBoxDividers -> {
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], 
            GridBox[{{
               TagBox[
                TooltipBox["0", 
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]], Annotation[#, {1, 2, 1} -> 0, "Tooltip"]& ]}, {
               TagBox[
                TooltipBox[
                 RowBox[{"-", 
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], 
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                   RowBox[{"-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}]}]], 
                Annotation[#, {1, 2, 2} -> -$CellContext`r[], "Tooltip"]& ]}},
              AutoDelete -> False, 
             GridBoxDividers -> {
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}, {
            GridBox[{{
               TagBox[
                TooltipBox["0", 
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]], Annotation[#, {2, 1, 1} -> 0, "Tooltip"]& ]}, {
               TagBox[
                TooltipBox[
                 FractionBox["1", 
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False]], 
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                   FractionBox["1", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]], 
                Annotation[#, {2, 1, 2} -> $CellContext`r[]^(-1), 
                 "Tooltip"]& ]}}, AutoDelete -> False, 
             GridBoxDividers -> {
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], 
            GridBox[{{
               TagBox[
                TooltipBox[
                 FractionBox["1", 
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False]], 
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                   FractionBox["1", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]], 
                Annotation[#, {2, 2, 1} -> $CellContext`r[]^(-1), 
                 "Tooltip"]& ]}, {
               TagBox[
                TooltipBox["0", 
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                   "0"}]], Annotation[#, {2, 2, 2} -> 0, "Tooltip"]& ]}}, 
             AutoDelete -> False, 
             GridBoxDividers -> {
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}}, 
          AutoDelete -> False, 
          GridBoxDividers -> {
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], Small], ",", 
        StyleBox[
         GridBox[{{
            TagBox[
             TooltipBox["1", 
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "1"}]], 
             Annotation[#, {1, 1} -> 1, "Tooltip"]& ], 
            TagBox[
             TooltipBox["0", 
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]], 
             Annotation[#, {1, 2} -> 0, "Tooltip"]& ]}, {
            TagBox[
             TooltipBox["0", 
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]], 
             Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              SuperscriptBox[
               InterpretationBox[
                StyleBox[
                "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                $CellContext`r[], Editable -> False], "2"], 
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                  "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`r[], Editable -> False], "2"]}]], 
             Annotation[#, {2, 2} -> $CellContext`r[]^2, "Tooltip"]& ]}}, 
          AutoDelete -> False, 
          GridBoxDividers -> {
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], Small], 
        "]"}]], "a"], GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"b"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CCovD[$CellContext`PDpolar, 
   xAct`xCoba`CTensor[{{{0, 0}, {0, -$CellContext`r[]}}, {{
      0, $CellContext`r[]^(-1)}, {$CellContext`r[]^(-1), 
       0}}}, {$CellContext`polar, -$CellContext`polar, -$CellContext`polar}, 
    0], 
   xAct`xCoba`CTensor[{{1, 0}, {
     0, $CellContext`r[]^2}}, {-$CellContext`polar, -$CellContext`polar}, 
    0]][-$CellContext`a][
   $CellContext`v[$CellContext`b]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728676272945978*^9, 3.728753521800397*^9, 
  3.728843611412796*^9, 3.730956246414035*^9, 3.731117058906221*^9, 
  3.733367972426701*^9, 3.7333681428512707`*^9, 3.794517946114802*^9, 
  3.7945180238867292`*^9, 3.7945183137459106`*^9, 3.794518430390588*^9, 
  3.794518788806041*^9, 3.794519068770747*^9, 3.794520682329751*^9},
 CellLabel->"Out[204]=",
 CellID->9545941,ExpressionUUID->"81b43565-00a6-426b-b3cb-cc2935df5838"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We can mix abstract and CTensor Christoffels:", "MathCaption",
 CellID->1782777519,ExpressionUUID->"e9c91a2d-6638-48a7-aa80-543285dcb2e2"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ChangeCovD", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"cd", "[", 
     RowBox[{"-", "a"}], "]"}], "[", 
    RowBox[{"v", "[", "b", "]"}], "]"}], ",", "cd"}], "]"}]], "Input",
 CellLabel->"In[205]:=",
 CellID->1449918889,ExpressionUUID->"84341d38-0f59-4af7-ab6f-dc53f7aa0abb"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]], "]"}], 
          StyleBox[GridBox[{
             {"b", " ", " "},
             {" ", "a", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelPDpolar[$CellContext`b, -$CellContext`a, \
-$CellContext`c],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox[GridBox[{
             {GridBox[{
                {
                 TagBox[
                  TooltipBox["0",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    "0"}]],
                  Annotation[#, {1, 1, 1} -> 0, "Tooltip"]& ]},
                {
                 TagBox[
                  TooltipBox["0",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    "0"}]],
                  Annotation[#, {1, 1, 2} -> 0, "Tooltip"]& ]}
               },
               AutoDelete->False,
               
               GridBoxDividers->{
                "ColumnsIndexed" -> {
                  1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], 
              GridBox[{
                {
                 TagBox[
                  TooltipBox["0",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    "0"}]],
                  Annotation[#, {1, 2, 1} -> 0, "Tooltip"]& ]},
                {
                 TagBox[
                  TooltipBox[
                   RowBox[{"-", 
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}],
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    RowBox[{"-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}]}]],
                  Annotation[#, {1, 2, 2} -> -$CellContext`r[], 
                   "Tooltip"]& ]}
               },
               AutoDelete->False,
               
               GridBoxDividers->{
                "ColumnsIndexed" -> {
                  1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
             {GridBox[{
                {
                 TagBox[
                  TooltipBox["0",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    "0"}]],
                  Annotation[#, {2, 1, 1} -> 0, "Tooltip"]& ]},
                {
                 TagBox[
                  TooltipBox[
                   FractionBox["1", 
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]],
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox["1", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]],
                  
                  Annotation[#, {2, 1, 2} -> $CellContext`r[]^(-1), 
                   "Tooltip"]& ]}
               },
               AutoDelete->False,
               
               GridBoxDividers->{
                "ColumnsIndexed" -> {
                  1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], 
              GridBox[{
                {
                 TagBox[
                  TooltipBox[
                   FractionBox["1", 
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]],
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox["1", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]],
                  
                  Annotation[#, {2, 2, 1} -> $CellContext`r[]^(-1), 
                   "Tooltip"]& ]},
                {
                 TagBox[
                  TooltipBox["0",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    "0"}]],
                  Annotation[#, {2, 2, 2} -> 0, "Tooltip"]& ]}
               },
               AutoDelete->False,
               
               GridBoxDividers->{
                "ColumnsIndexed" -> {
                  1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
              "RowsIndexed" -> {
               1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
           FontSize->Small], 
          StyleBox[GridBox[{
             {"b", " ", " "},
             {" ", "a", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xCoba`CTensor[{{{0, 0}, {0, -$CellContext`r[]}}, {{
         0, $CellContext`r[]^(-1)}, {$CellContext`r[]^(-1), 
          0}}}, {$CellContext`polar, -$CellContext`polar, \
-$CellContext`polar}, 0][$CellContext`b, -$CellContext`a, -$CellContext`c],
      Editable->False]}], ")"}]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "a"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"b"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`a][
    $CellContext`v[$CellContext`b]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728676274430126*^9, 3.728753521946528*^9, 
  3.728843612680067*^9, 3.730956246543079*^9, 3.731117071669788*^9, 
  3.7333679734623013`*^9, 3.733368142948976*^9, 3.794517946163196*^9, 
  3.7945180239409657`*^9, 3.7945183138036423`*^9, 3.794518430462347*^9, 
  3.794518788895556*^9, 3.7945190688210907`*^9, 3.794520682388164*^9},
 CellLabel->"Out[205]=",
 CellID->1524840509,ExpressionUUID->"3ae51338-8a03-4bfa-85af-1e941238cdf8"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.3.2. Torsion and curvature", "Subsubsection",
 CellChangeTimes->{{3.733368055790737*^9, 3.733368056638995*^9}},
 CellID->1387396780,ExpressionUUID->"b6fcbbfa-7618-49c4-9dc9-5a7da8e14817"],

Cell["\<\
Given a 2-covariant symmetric CTensor object we can use it as a metric field, \
from which we can compute curvature. This can be done step by step or with \
the command MetricCompute. In this subsection we use the simpler approach, \
and then we will give example of MetricCompute usage in the next section.\
\>", "Text",
 CellChangeTimes->{{3.7311171338875713`*^9, 3.7311173103695393`*^9}},
 CellID->644289339,ExpressionUUID->"0234e021-7ae8-4a3c-92d5-42e6d1e92560"],

Cell[CellGroupData[{

Cell["\<\
Compute the torsion of the covariant derivative defined in the previous \
subsection. It is a torsion-free derivative:\
\>", "MathCaption",
 CellChangeTimes->{{3.728676297616967*^9, 3.728676297745158*^9}, {
  3.731117327264917*^9, 3.731117333255575*^9}},
 CellID->600242334,ExpressionUUID->"838b5a17-192d-47a5-be44-e0ced7195f80"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Torsion", "[", "cd", "]"}]], "Input",
 CellLabel->"In[206]:=",
 CellID->1542217579,ExpressionUUID->"2f8de1ad-be6e-4255-91a7-6209674f180c"],

Cell[BoxData["Zero"], "Output",
 CellChangeTimes->{3.728676298737335*^9, 3.728753525444069*^9, 
  3.728843620702857*^9, 3.730956251332032*^9, 3.731117335166031*^9, 
  3.733367981310246*^9, 3.7333681430523453`*^9, 3.794517946203073*^9, 
  3.794518023983642*^9, 3.794518321401642*^9, 3.7945184372446012`*^9, 
  3.794518788941408*^9, 3.794519068863802*^9, 3.794520682434895*^9},
 CellLabel->"Out[206]=",
 CellID->2072075112,ExpressionUUID->"fa4ba825-32b0-4799-a388-77134e5c66ad"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Compute its curvature. It is a flat connection:", "MathCaption",
 CellChangeTimes->{{3.728676299784931*^9, 3.728676299905064*^9}, {
  3.731117348295805*^9, 3.7311173562720137`*^9}},
 CellID->380466988,ExpressionUUID->"1bc39641-97d7-4f54-91b0-b3c9713f9c71"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Riemann", "[", "cd", "]"}]], "Input",
 CellLabel->"In[207]:=",
 CellID->1449850197,ExpressionUUID->"512df925-f8d6-477d-bb1e-299e8e57561b"],

Cell[BoxData["Zero"], "Output",
 CellChangeTimes->{3.72867630082298*^9, 3.728753525624609*^9, 
  3.728843620892292*^9, 3.730956251475367*^9, 3.731117357651976*^9, 
  3.733367981479815*^9, 3.7333681431846*^9, 3.794517946242708*^9, 
  3.794518024058853*^9, 3.794518321457086*^9, 3.7945184373029423`*^9, 
  3.7945187889801073`*^9, 3.794519068948688*^9, 3.794520682503457*^9},
 CellLabel->"Out[207]=",
 CellID->1771587632,ExpressionUUID->"c84135a6-f3b7-402b-951f-b28c9f30dd68"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Suppose we introduce a new metric:", "MathCaption",
 CellChangeTimes->{{3.7286763018084927`*^9, 3.7286763019291487`*^9}},
 CellID->112343117,ExpressionUUID->"66895eb2-d23a-4f47-b675-34087eecb047"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"MET", "=", 
  RowBox[{"CTensor", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"1", ",", "0"}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{
         RowBox[{"r", "[", "]"}], "^", "4"}]}], "}"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "polar"}], ",", 
      RowBox[{"-", "polar"}]}], "}"}]}], "]"}]}]], "Input",
 CellLabel->"In[208]:=",
 CellID->512641706,ExpressionUUID->"abcc4cc3-cc43-4988-b119-dd27d5553be8"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "4"]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "polar"}], ",", 
     RowBox[{"-", "polar"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.728676302856349*^9, 3.72875352579456*^9, 
  3.728843621114656*^9, 3.730956251593794*^9, 3.731117374341206*^9, 
  3.733367981598672*^9, 3.733368143302142*^9, 3.794517946279902*^9, 
  3.79451802410563*^9, 3.794518321505754*^9, 3.794518437355341*^9, 
  3.794518789022373*^9, 3.794519068993701*^9, 3.79452068255473*^9},
 CellLabel->"Out[208]=",
 CellID->212565369,ExpressionUUID->"be8a52e0-e2e9-493b-a216-067de7708325"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"CD", "=", 
  RowBox[{"CovDOfMetric", "[", "MET", "]"}]}]], "Input",
 CellLabel->"In[209]:=",
 CellID->251640662,ExpressionUUID->"8b2fad4a-7c6f-482a-9fce-e6025be670d9"],

Cell[BoxData[
 RowBox[{"CCovD", "[", 
  RowBox[{"PDpolar", ",", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", 
           RowBox[{
            RowBox[{"-", "2"}], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "3"]}]}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", 
           FractionBox["2", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           FractionBox["2", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]], ",", "0"}], "}"}]}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"polar", ",", 
       RowBox[{"-", "polar"}], ",", 
       RowBox[{"-", "polar"}]}], "}"}], ",", "0"}], "]"}], ",", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"1", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "4"]}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "polar"}], ",", 
       RowBox[{"-", "polar"}]}], "}"}], ",", "0"}], "]"}]}], "]"}]], "Output",\

 CellChangeTimes->{3.728676303894205*^9, 3.728753525942878*^9, 
  3.72884362168851*^9, 3.730956251726863*^9, 3.731117389290864*^9, 
  3.733367981747467*^9, 3.7333681433958683`*^9, 3.79451794632364*^9, 
  3.7945180241508837`*^9, 3.794518321557249*^9, 3.794518437407995*^9, 
  3.7945187890659*^9, 3.7945190690351954`*^9, 3.794520682603026*^9},
 CellLabel->"Out[209]=",
 CellID->1320851822,ExpressionUUID->"0280e046-2761-4d32-955e-19a7513f2b8e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["It is still torsionless:", "MathCaption",
 CellChangeTimes->{{3.728676305527667*^9, 3.728676306024333*^9}},
 CellID->185461206,ExpressionUUID->"8494a88a-a005-487a-9f1e-4948cff37802"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Torsion", "[", "CD", "]"}], "[", 
  RowBox[{"a", ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}]}], "]"}]], "Input",
 CellLabel->"In[210]:=",
 CellID->1047717445,ExpressionUUID->"114c8fdd-d5f8-4223-b966-cd758bc14627"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.7286763069840393`*^9, 3.7287535261925917`*^9, 
  3.728843622298615*^9, 3.730956251842821*^9, 3.731117392860607*^9, 
  3.733367981864717*^9, 3.7333681435183067`*^9, 3.794517946361967*^9, 
  3.794518024191087*^9, 3.794518321609272*^9, 3.79451843745874*^9, 
  3.794518789105*^9, 3.7945190690727787`*^9, 3.794520682654591*^9},
 CellLabel->"Out[210]=",
 CellID->1018113315,ExpressionUUID->"93db569b-a90b-48df-b6f8-d23c1d076dd9"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["But now it has curvature:", "MathCaption",
 CellChangeTimes->{{3.728676308176428*^9, 3.728676308304661*^9}},
 CellID->674974360,ExpressionUUID->"a0744664-0ec4-421d-86ea-19dec982f701"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Riemann", "[", "CD", "]"}], "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}], ",", "d"}], "]"}]], "Input",
 CellLabel->"In[211]:=",
 CellID->1285445066,ExpressionUUID->"bba9eae9-5c27-426d-9d13-5c310600561b"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                FractionBox["2", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox["2", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]]}]}]],
              
              Annotation[#, {1, 2, 1, 2} -> (-2) $CellContext`r[]^(-2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{"2", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"2", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]}]],
              
              Annotation[#, {1, 2, 2, 1} -> 2 $CellContext`r[]^2, 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               FractionBox["2", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox["2", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]]}]],
              
              Annotation[#, {2, 1, 1, 2} -> 2 $CellContext`r[]^(-2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                RowBox[{"-", "2"}], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{
                   RowBox[{"-", "2"}], " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]}]],
              
              Annotation[#, {2, 1, 2, 1} -> (-2) $CellContext`r[]^2, 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " ", " ", "d"},
         {"a", "b", "c", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{{{0, 0}, {0, 
     0}}, {{0, (-2) $CellContext`r[]^(-2)}, {2 $CellContext`r[]^2, 0}}}, {{{
      0, 2 $CellContext`r[]^(-2)}, {(-2) $CellContext`r[]^2, 0}}, {{0, 0}, {0,
      0}}}}, {-$CellContext`polar, -$CellContext`polar, -$CellContext`polar, \
$CellContext`polar}, 
   0][-$CellContext`a, -$CellContext`b, -$CellContext`c, $CellContext`d],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.728676309240893*^9, 3.728753526575939*^9, 3.728843622581789*^9, 
   3.730956251982437*^9, 3.731117394440946*^9, {3.731117535933845*^9, 
   3.7311175553885*^9}, 3.733367981979405*^9, 3.7333681436348763`*^9, 
   3.794517946426304*^9, 3.794518024240102*^9, 3.794518321666507*^9, 
   3.794518437521472*^9, 3.794518789164372*^9, 3.794519069128928*^9, 
   3.794520682732163*^9},
 CellLabel->"Out[211]=",
 CellID->280785909,ExpressionUUID->"92576d3c-558a-4a85-8e08-2df7f8524f27"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SymmetryGroupOfTensor", "[", 
  RowBox[{"Riemann", "[", "CD", "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.731117539517272*^9, 3.731117552463374*^9}},
 CellLabel->"In[212]:=",
 CellID->2135112499,ExpressionUUID->"eabe65c0-b6ba-44da-b392-4a73abc619c7"],

Cell[BoxData[
 RowBox[{"StrongGenSet", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "2"}], "}"}], ",", 
   RowBox[{"GenSet", "[", 
    InterpretationBox[
     StyleBox[
      RowBox[{"-", 
       RowBox[{"(", "1", ",", "2", ")"}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     -xAct`xPerm`Cycles[{1, 2}],
     Editable->False], "]"}]}], "]"}]], "Output",
 CellChangeTimes->{{3.731117546680073*^9, 3.731117555969131*^9}, 
   3.733367982181612*^9, 3.733368143752595*^9, 3.794517946466547*^9, 
   3.794518024280781*^9, 3.794518321716298*^9, 3.794518437571731*^9, 
   3.794518789206686*^9, 3.7945190691719313`*^9, 3.794520682777141*^9},
 CellLabel->"Out[212]=",
 CellID->1147200104,ExpressionUUID->"f3a94b6b-87e7-4013-834b-7916ce9a66e8"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
That tensor is not Riemann-symmetric. We cannot automatically lower its last \
index (because we have not used SetCMetric), but we can manually do it:\
\>", "MathCaption",
 CellChangeTimes->{{3.731117440412856*^9, 3.731117495251219*^9}},
 CellID->652069640,ExpressionUUID->"711e4996-a13f-4841-86ad-132978480215"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"Riemann", "[", "CD", "]"}], "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}], ",", "e"}], "]"}], 
  RowBox[{"MET", "[", 
   RowBox[{
    RowBox[{"-", "e"}], ",", 
    RowBox[{"-", "d"}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.731117496363752*^9, 3.731117511698442*^9}},
 CellLabel->"In[213]:=",
 CellID->1096816598,ExpressionUUID->"cec7a0d4-bc92-4c60-a36b-3a0a4c019911"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                RowBox[{"-", "2"}], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{
                   RowBox[{"-", "2"}], " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]}]],
              
              Annotation[#, {1, 2, 1, 2} -> (-2) $CellContext`r[]^2, 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{"2", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"2", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]}]],
              
              Annotation[#, {1, 2, 2, 1} -> 2 $CellContext`r[]^2, 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{"2", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"2", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]}]],
              
              Annotation[#, {2, 1, 1, 2} -> 2 $CellContext`r[]^2, 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                RowBox[{"-", "2"}], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{
                   RowBox[{"-", "2"}], " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]}]],
              
              Annotation[#, {2, 1, 2, 1} -> (-2) $CellContext`r[]^2, 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " ", " ", " "},
         {"a", "b", "c", "d"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{{{0, 0}, {0, 
     0}}, {{0, (-2) $CellContext`r[]^2}, {2 $CellContext`r[]^2, 0}}}, {{{
      0, 2 $CellContext`r[]^2}, {(-2) $CellContext`r[]^2, 0}}, {{0, 0}, {0, 
     0}}}}, {-$CellContext`polar, -$CellContext`polar, -$CellContext`polar, \
-$CellContext`polar}, 
   0][-$CellContext`a, -$CellContext`b, -$CellContext`c, -$CellContext`d],
  Editable->False]], "Output",
 CellChangeTimes->{3.731117515180135*^9, 3.731117565538981*^9, 
  3.733367982475732*^9, 3.733368143885355*^9, 3.794517946518062*^9, 
  3.7945180243253927`*^9, 3.794518321768413*^9, 3.794518437632268*^9, 
  3.79451878925762*^9, 3.7945190692263002`*^9, 3.7945206828331137`*^9},
 CellLabel->"Out[213]=",
 CellID->883965162,ExpressionUUID->"bcf872d2-80ef-443f-8950-5dce06e19b75"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
There is also this alternative equivalent object, which is now fully \
Riemann-symmetric:\
\>", "MathCaption",
 CellChangeTimes->{{3.731117520657281*^9, 3.731117523778171*^9}, {
  3.7311175745522957`*^9, 3.731117592287133*^9}},
 CellID->191320769,ExpressionUUID->"5dbe9a51-caa2-43f3-a4bd-f2dd24c1e40d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"RiemannDown", "[", "CD", "]"}], "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}], ",", 
   RowBox[{"-", "d"}]}], "]"}]], "Input",
 CellLabel->"In[214]:=",
 CellID->1634635234,ExpressionUUID->"8f56b820-42f1-4252-8457-6b3a01a56569"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                RowBox[{"-", "2"}], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{
                   RowBox[{"-", "2"}], " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]}]],
              
              Annotation[#, {1, 2, 1, 2} -> (-2) $CellContext`r[]^2, 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{"2", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"2", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]}]],
              
              Annotation[#, {1, 2, 2, 1} -> 2 $CellContext`r[]^2, 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{"2", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"2", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]}]],
              
              Annotation[#, {2, 1, 1, 2} -> 2 $CellContext`r[]^2, 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                RowBox[{"-", "2"}], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{
                   RowBox[{"-", "2"}], " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]}]],
              
              Annotation[#, {2, 1, 2, 1} -> (-2) $CellContext`r[]^2, 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " ", " ", " "},
         {"a", "b", "c", "d"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{{{0, 0}, {0, 
     0}}, {{0, (-2) $CellContext`r[]^2}, {2 $CellContext`r[]^2, 0}}}, {{{
      0, 2 $CellContext`r[]^2}, {(-2) $CellContext`r[]^2, 0}}, {{0, 0}, {0, 
     0}}}}, {-$CellContext`polar, -$CellContext`polar, -$CellContext`polar, \
-$CellContext`polar}, 
   0][-$CellContext`a, -$CellContext`b, -$CellContext`c, -$CellContext`d],
  Editable->False]], "Output",
 CellChangeTimes->{3.728676310632845*^9, 3.728753526825882*^9, 
  3.728843622785537*^9, 3.7309562520785513`*^9, 3.731117593859332*^9, 
  3.733367982759177*^9, 3.7333681440007677`*^9, 3.794517946567864*^9, 
  3.794518024371344*^9, 3.794518321832139*^9, 3.794518437681529*^9, 
  3.794518789303792*^9, 3.794519069276494*^9, 3.794520682888756*^9},
 CellLabel->"Out[214]=",
 CellID->510607788,ExpressionUUID->"7cf83d29-7fd4-4d34-81a1-465c0cd681f3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SymmetryGroupOfTensor", "[", 
  RowBox[{"RiemannDown", "[", "CD", "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.7311175950436563`*^9, 3.731117605351108*^9}},
 CellLabel->"In[215]:=",
 CellID->195231078,ExpressionUUID->"50fe4117-7f05-4849-b738-409ae4ee0690"],

Cell[BoxData[
 RowBox[{"StrongGenSet", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"3", ",", "1"}], "}"}], ",", 
   RowBox[{"GenSet", "[", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"-", 
        RowBox[{"(", "3", ",", "4", ")"}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      -xAct`xPerm`Cycles[{3, 4}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[
       RowBox[{"-", 
        RowBox[{"(", "1", ",", "2", ")"}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      -xAct`xPerm`Cycles[{1, 2}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"(", "1", ",", "3", ")"}], 
        RowBox[{"(", "2", ",", "4", ")"}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xPerm`Cycles[{1, 3}, {2, 4}],
      Editable->False]}], "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.731117605785747*^9, 3.733367982911716*^9, 
  3.733368144095995*^9, 3.794517946608099*^9, 3.7945180244114733`*^9, 
  3.794518321876542*^9, 3.7945184377299128`*^9, 3.794518789344966*^9, 
  3.794519069322341*^9, 3.7945206829331017`*^9},
 CellLabel->"Out[215]=",
 CellID->440439504,ExpressionUUID->"acb7c167-1e16-49aa-8947-566ea0d69887"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Ricci and Ricci scalar:", "MathCaption",
 CellChangeTimes->{{3.731117618582675*^9, 3.7311176241514997`*^9}},
 CellID->44226057,ExpressionUUID->"f301d7b1-fba6-4107-942d-d4bb2753e163"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Ricci", "[", "CD", "]"}], "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}]}], "]"}]], "Input",
 CellLabel->"In[216]:=",
 CellID->217426225,ExpressionUUID->"2df2d09b-1e12-46c4-9b70-28b3492dabc9"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{"-", 
             FractionBox["2", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], "2"]]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{"-", 
                FractionBox["2", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False], "2"]]}]}]],
           Annotation[#, {1, 1} -> (-2) $CellContext`r[]^(-2), "Tooltip"]& ], 
          
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 2} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{"-", "2"}], " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{"-", "2"}], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                  "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`r[], Editable -> False], "2"]}]}]],
           Annotation[#, {2, 2} -> (-2) $CellContext`r[]^2, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " "},
         {"a", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{(-2) $CellContext`r[]^(-2), 0}, {
    0, (-2) $CellContext`r[]^2}}, {-$CellContext`polar, -$CellContext`polar}, 
   0][-$CellContext`a, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.728676311851449*^9, 3.7287535270447063`*^9, 
  3.7288436229815807`*^9, 3.7309562521789618`*^9, 3.7311176258118*^9, 
  3.733367983060844*^9, 3.733368144218804*^9, 3.794517946650181*^9, 
  3.7945180244564238`*^9, 3.794518321936947*^9, 3.7945184377760553`*^9, 
  3.794518789392125*^9, 3.7945190693722897`*^9, 3.794520682996418*^9},
 CellLabel->"Out[216]=",
 CellID->90612519,ExpressionUUID->"0fafe665-3696-4fe9-bc09-7352c7945006"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"RicciScalar", "[", "CD", "]"}], "[", "]"}]], "Input",
 CellLabel->"In[217]:=",
 CellID->1547945717,ExpressionUUID->"cdf9d9d9-77f8-47cd-9816-bb2cdfb6f81f"],

Cell[BoxData[
 RowBox[{"-", 
  FractionBox["4", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]]}]], "Output",
 CellChangeTimes->{3.728676312467813*^9, 3.728753527190517*^9, 
  3.728843623156201*^9, 3.730956252307591*^9, 3.7311176267349787`*^9, 
  3.733367983298564*^9, 3.733368144338439*^9, 3.7945179466865664`*^9, 
  3.794518024496538*^9, 3.794518321982563*^9, 3.7945184378323927`*^9, 
  3.794518789433094*^9, 3.794519069414546*^9, 3.794520683042774*^9},
 CellLabel->"Out[217]=",
 CellID->358701110,ExpressionUUID->"558920b3-1f10-405f-9d73-af071bdb29bd"]
}, Open  ]]
}, Open  ]],

Cell["\<\
As long as we do not use SetCMetric we can compute with many different metric \
objects simultaneously. Using SetCMetric would select one as being the \
first-metric and therefore xCoba might use it to raise and lower indices of \
some tensors.\
\>", "Text",
 CellChangeTimes->{{3.7311171338875713`*^9, 3.731117161142995*^9}, {
  3.7311177006297073`*^9, 3.731117739266671*^9}},
 CellID->1911163297,ExpressionUUID->"ff44fa7d-2406-4075-9176-04614a19e6da"],

Cell[CellGroupData[{

Cell["Take a list of different CTensor metrics:", "MathCaption",
 CellChangeTimes->{{3.7311176397819643`*^9, 3.731117678212915*^9}, {
  3.731117753530039*^9, 3.731117763537881*^9}},
 CellID->163945661,ExpressionUUID->"cfa7fc30-7b27-429b-8b7c-d85b71400bde"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Table", "[", 
  RowBox[{
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"1", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", 
         RowBox[{
          RowBox[{"r", "[", "]"}], "^", "n"}]}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "polar"}], ",", 
       RowBox[{"-", "polar"}]}], "}"}]}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"n", ",", "0", ",", "10", ",", "2"}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.731117767246419*^9, 3.7311178127361517`*^9}},
 CellLabel->"In[218]:=",
 CellID->934847067,ExpressionUUID->"240db47b-aca9-4e26-9ec2-5604bc04b910"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"1", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "1"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "polar"}], ",", 
       RowBox[{"-", "polar"}]}], "}"}], ",", "0"}], "]"}], ",", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"1", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"]}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "polar"}], ",", 
       RowBox[{"-", "polar"}]}], "}"}], ",", "0"}], "]"}], ",", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"1", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "4"]}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "polar"}], ",", 
       RowBox[{"-", "polar"}]}], "}"}], ",", "0"}], "]"}], ",", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"1", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "6"]}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "polar"}], ",", 
       RowBox[{"-", "polar"}]}], "}"}], ",", "0"}], "]"}], ",", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"1", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "8"]}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "polar"}], ",", 
       RowBox[{"-", "polar"}]}], "}"}], ",", "0"}], "]"}], ",", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"1", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "10"]}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "polar"}], ",", 
       RowBox[{"-", "polar"}]}], "}"}], ",", "0"}], "]"}]}], "}"}]], "Output",\

 CellChangeTimes->{3.731117815883647*^9, 3.733367983582884*^9, 
  3.7333681444689093`*^9, 3.7945179467258368`*^9, 3.794518024541794*^9, 
  3.794518322040368*^9, 3.794518437876338*^9, 3.794518789476042*^9, 
  3.7945190694576*^9, 3.794520683095386*^9},
 CellLabel->"Out[218]=",
 CellID->702958641,ExpressionUUID->"47139ffd-2c77-4cd5-8f6f-54e7b035f14a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Compute their respective Riemann tensors:", "MathCaption",
 CellChangeTimes->{{3.731117868534038*^9, 3.7311178785019197`*^9}},
 CellID->540626935,ExpressionUUID->"dfba7197-d26f-4bf3-bf3a-b2a0524bfa56"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"Riemann", "[", 
     RowBox[{"CovDOfMetric", "[", "#", "]"}], "]"}], "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], "&"}], "/@", "%"}]], "Input",
 CellChangeTimes->{{3.731117828851452*^9, 3.7311178571906233`*^9}},
 CellLabel->"In[219]:=",
 CellID->2016378214,ExpressionUUID->"62b3e6ab-c2e4-4a80-8161-7c78ec091f71"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0", ",", "0", ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        StyleBox[GridBox[{
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 1, 2} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 2, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 2, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 2, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox[
                 RowBox[{"-", 
                  FractionBox["2", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", 
                   RowBox[{"-", 
                    FractionBox["2", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]]}]}]],
                
                Annotation[#, {1, 2, 1, 2} -> (-2) $CellContext`r[]^(-2), 
                 "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 RowBox[{"2", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", 
                   RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]}]],
                
                Annotation[#, {1, 2, 2, 1} -> 2 $CellContext`r[]^2, 
                 "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 2, 2, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]},
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 1, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox[
                 FractionBox["2", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", 
                   FractionBox["2", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]]}]],
                
                Annotation[#, {2, 1, 1, 2} -> 2 $CellContext`r[]^(-2), 
                 "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 RowBox[{
                  RowBox[{"-", "2"}], " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", 
                   RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]}]],
                
                Annotation[#, {2, 1, 2, 1} -> (-2) $CellContext`r[]^2, 
                 "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 1, 2, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 1, 2} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 2, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 2, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
         FontSize->Small], 
        StyleBox[GridBox[{
           {" ", " ", " ", "d"},
           {"a", "b", "c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xCoba`CTensor[{{{{0, 0}, {0, 
       0}}, {{0, (-2) $CellContext`r[]^(-2)}, {2 $CellContext`r[]^2, 0}}}, {{{
        0, 2 $CellContext`r[]^(-2)}, {(-2) $CellContext`r[]^2, 0}}, {{0, 0}, {
       0, 0}}}}, {-$CellContext`polar, -$CellContext`polar, \
-$CellContext`polar, $CellContext`polar}, 
     0][-$CellContext`a, -$CellContext`b, -$CellContext`c, $CellContext`d],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        StyleBox[GridBox[{
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 1, 2} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 2, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 2, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 2, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox[
                 RowBox[{"-", 
                  FractionBox["6", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", 
                   RowBox[{"-", 
                    FractionBox["6", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]]}]}]],
                
                Annotation[#, {1, 2, 1, 2} -> (-6) $CellContext`r[]^(-2), 
                 "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 RowBox[{"6", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "4"]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", 
                   RowBox[{"6", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "4"]}]}]],
                
                Annotation[#, {1, 2, 2, 1} -> 6 $CellContext`r[]^4, 
                 "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 2, 2, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]},
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 1, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox[
                 FractionBox["6", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", 
                   FractionBox["6", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]]}]],
                
                Annotation[#, {2, 1, 1, 2} -> 6 $CellContext`r[]^(-2), 
                 "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 RowBox[{
                  RowBox[{"-", "6"}], " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "4"]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", 
                   RowBox[{
                    RowBox[{"-", "6"}], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "4"]}]}]],
                
                Annotation[#, {2, 1, 2, 1} -> (-6) $CellContext`r[]^4, 
                 "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 1, 2, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 1, 2} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 2, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 2, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
         FontSize->Small], 
        StyleBox[GridBox[{
           {" ", " ", " ", "d"},
           {"a", "b", "c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xCoba`CTensor[{{{{0, 0}, {0, 
       0}}, {{0, (-6) $CellContext`r[]^(-2)}, {6 $CellContext`r[]^4, 0}}}, {{{
        0, 6 $CellContext`r[]^(-2)}, {(-6) $CellContext`r[]^4, 0}}, {{0, 0}, {
       0, 0}}}}, {-$CellContext`polar, -$CellContext`polar, \
-$CellContext`polar, $CellContext`polar}, 
     0][-$CellContext`a, -$CellContext`b, -$CellContext`c, $CellContext`d],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        StyleBox[GridBox[{
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 1, 2} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 2, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 2, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 2, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox[
                 RowBox[{"-", 
                  FractionBox["12", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", 
                   RowBox[{"-", 
                    FractionBox["12", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]]}]}]],
                
                Annotation[#, {1, 2, 1, 2} -> (-12) $CellContext`r[]^(-2), 
                 "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 RowBox[{"12", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "6"]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", 
                   RowBox[{"12", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "6"]}]}]],
                
                Annotation[#, {1, 2, 2, 1} -> 12 $CellContext`r[]^6, 
                 "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 2, 2, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]},
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 1, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox[
                 FractionBox["12", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", 
                   FractionBox["12", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]]}]],
                
                Annotation[#, {2, 1, 1, 2} -> 12 $CellContext`r[]^(-2), 
                 "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 RowBox[{
                  RowBox[{"-", "12"}], " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "6"]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", 
                   RowBox[{
                    RowBox[{"-", "12"}], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "6"]}]}]],
                
                Annotation[#, {2, 1, 2, 1} -> (-12) $CellContext`r[]^6, 
                 "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 1, 2, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 1, 2} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 2, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 2, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
         FontSize->Small], 
        StyleBox[GridBox[{
           {" ", " ", " ", "d"},
           {"a", "b", "c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xCoba`CTensor[{{{{0, 0}, {0, 
       0}}, {{0, (-12) $CellContext`r[]^(-2)}, {
        12 $CellContext`r[]^6, 0}}}, {{{
        0, 12 $CellContext`r[]^(-2)}, {(-12) $CellContext`r[]^6, 0}}, {{0, 
       0}, {0, 0}}}}, {-$CellContext`polar, -$CellContext`polar, \
-$CellContext`polar, $CellContext`polar}, 
     0][-$CellContext`a, -$CellContext`b, -$CellContext`c, $CellContext`d],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        StyleBox[GridBox[{
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 1, 2} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 2, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 2, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 2, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox[
                 RowBox[{"-", 
                  FractionBox["20", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", 
                   RowBox[{"-", 
                    FractionBox["20", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]]}]}]],
                
                Annotation[#, {1, 2, 1, 2} -> (-20) $CellContext`r[]^(-2), 
                 "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 RowBox[{"20", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "8"]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", 
                   RowBox[{"20", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "8"]}]}]],
                
                Annotation[#, {1, 2, 2, 1} -> 20 $CellContext`r[]^8, 
                 "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 2, 2, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]},
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 1, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox[
                 FractionBox["20", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", 
                   FractionBox["20", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]]}]],
                
                Annotation[#, {2, 1, 1, 2} -> 20 $CellContext`r[]^(-2), 
                 "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 RowBox[{
                  RowBox[{"-", "20"}], " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "8"]}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", 
                   RowBox[{
                    RowBox[{"-", "20"}], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "8"]}]}]],
                
                Annotation[#, {2, 1, 2, 1} -> (-20) $CellContext`r[]^8, 
                 "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 1, 2, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 1, 2} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 2, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 2, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
         FontSize->Small], 
        StyleBox[GridBox[{
           {" ", " ", " ", "d"},
           {"a", "b", "c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xCoba`CTensor[{{{{0, 0}, {0, 
       0}}, {{0, (-20) $CellContext`r[]^(-2)}, {
        20 $CellContext`r[]^8, 0}}}, {{{
        0, 20 $CellContext`r[]^(-2)}, {(-20) $CellContext`r[]^8, 0}}, {{0, 
       0}, {0, 0}}}}, {-$CellContext`polar, -$CellContext`polar, \
-$CellContext`polar, $CellContext`polar}, 
     0][-$CellContext`a, -$CellContext`b, -$CellContext`c, $CellContext`d],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.731117861990246*^9, 3.7333679838256197`*^9, 
  3.733368144628632*^9, 3.7945179468296003`*^9, 3.794518024652484*^9, 
  3.7945183221546097`*^9, 3.7945184379966288`*^9, 3.794518789584503*^9, 
  3.794519069569909*^9, 3.794520683211824*^9},
 CellLabel->"Out[219]=",
 CellID->1318337140,ExpressionUUID->"47a25be4-3bdb-44ec-ad35-7c7b10055f4a"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.3.3. Cleanup", "Subsubsection",
 CellChangeTimes->{{3.7333680578707533`*^9, 3.7333680584869003`*^9}},
 CellID->1674752848,ExpressionUUID->"78e8b403-cf08-41dd-a413-ef7a35faf34f"],

Cell[BoxData[
 RowBox[{"UndefChart", "[", "polar", "]"}]], "Input",
 CellLabel->"In[220]:=",
 CellID->502955107,ExpressionUUID->"6b5d4cd4-d760-406e-9039-12d4030c623c"],

Cell[BoxData[
 RowBox[{"UndefTensor", "[", "v", "]"}]], "Input",
 CellLabel->"In[221]:=",
 CellID->670848757,ExpressionUUID->"fb077724-34c3-4da1-8531-2e161d064532"],

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "E2", "]"}]], "Input",
 CellLabel->"In[222]:=",
 CellID->665912149,ExpressionUUID->"75fb1a05-5fdf-4593-adf9-2e87ba3d6f05"],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{"cd", ",", "CD", ",", "met", ",", "MET", ",", "n"}], "]"}]], "Input",\

 CellChangeTimes->{{3.794518397111031*^9, 3.794518397375662*^9}},
 CellLabel->"In[223]:=",
 CellID->1448203726,ExpressionUUID->"a1b3f89f-50f6-4f39-b9ca-72eae35dadb7"],

Cell[BoxData[
 RowBox[{"ClearxCobaCache", "[", "All", "]"}]], "Input",
 CellLabel->"In[224]:=",
 CellID->380367784,ExpressionUUID->"c9aef380-7a4b-48dd-83f6-f49589c44ecd"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.4. Vector analysis in flat space", "Subsection",
 CellChangeTimes->{{3.728656510283742*^9, 3.728656511217067*^9}, {
  3.7286598570138693`*^9, 3.728659868855068*^9}, {3.728660932877496*^9, 
  3.7286609335000553`*^9}, {3.7908686321917143`*^9, 3.790868643182276*^9}, {
  3.7908767987022667`*^9, 3.7908768053721333`*^9}, {3.7908770227443533`*^9, 
  3.7908770262385197`*^9}},
 CellID->1994287226,ExpressionUUID->"5138cf00-354a-4b53-9b00-3a8ac2e7abeb"],

Cell["\<\
All of xAct can work in flat space situations, but it may feel overwhelming \
at times, because xAct was developed with curved space applications in mind. \
The default settings are prepared to handle the general case.\
\>", "Text",
 CellChangeTimes->{{3.790863109490128*^9, 3.790863258626314*^9}, {
  3.790866420158791*^9, 3.790866466799429*^9}, {3.790867225498068*^9, 
  3.79086723724268*^9}, {3.790868709189741*^9, 3.7908689078781157`*^9}, {
  3.791378020105124*^9, 3.791378021137537*^9}},
 CellID->2143659166,ExpressionUUID->"0be1e37b-5ae4-4021-b3c8-102d7f6f7ed7"],

Cell["\<\
This subsection explains how to construct trivial operators like Grad, Div, \
Curl and Laplacian in 3-dimensional flat space but with non-Cartesian \
coordinates. To distinguish them from the Wolfram Language builtins, we will \
prepend their names with x: xGrad, xDiv, xCurl and xLaplacian. To make it \
more interesting, we will work with coordinated bases, which are typically \
used in non-flat situations, and then compare to orthonormal bases, which are \
typically used in flat space computations (and what the WL builtins use). \
This will feel convoluted, but will allow us to explain a number of important \
concepts.\
\>", "Text",
 CellChangeTimes->{{3.790863109490128*^9, 3.790863258626314*^9}, {
  3.790866420158791*^9, 3.790866466799429*^9}, {3.790867225498068*^9, 
  3.79086723724268*^9}, {3.790868709189741*^9, 3.7908689326766863`*^9}, {
  3.7908702783898993`*^9, 3.790870374723096*^9}, {3.790871328490143*^9, 
  3.7908713410340357`*^9}, {3.791378067424333*^9, 3.791378175860508*^9}},
 CellID->602637783,ExpressionUUID->"186862f9-109f-458f-ba35-ad1c244da9af"],

Cell[CellGroupData[{

Cell["3.4.1. Set up bases and metric", "Subsubsection",
 CellChangeTimes->{{3.728676331871439*^9, 3.728676332031807*^9}, 
   3.790876746494073*^9, {3.790876969601294*^9, 3.790876983335988*^9}},
 CellID->2036502373,ExpressionUUID->"ab5241d3-290b-48d8-ae82-3d56aeeb9aae"],

Cell[CellGroupData[{

Cell["Define the Euclidean 3-dimensional space with indices i...n:", \
"MathCaption",
 CellChangeTimes->{{3.7908689409961576`*^9, 3.7908689699955482`*^9}},
 CellID->729583777,ExpressionUUID->"1c4793f9-95a9-40e4-ad47-e9b60adc40a2"],

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"E3", ",", "3", ",", 
   RowBox[{"{", 
    RowBox[{"i", ",", "j", ",", "k", ",", "l", ",", "m", ",", "n"}], "}"}]}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.790868971999699*^9, 3.790868983915318*^9}},
 CellLabel->"In[225]:=",
 CellID->2016996031,ExpressionUUID->"7d2f106e-9362-435c-9788-63a523c09c05"]
}, Open  ]],

Cell[CellGroupData[{

Cell["We will work with spherical coordinates:", "MathCaption",
 CellChangeTimes->{{3.79086902037632*^9, 3.790869028961493*^9}},
 CellID->827825506,ExpressionUUID->"eec7e3b9-27e7-484d-ab41-d31f48475588"],

Cell[BoxData[
 RowBox[{"DefChart", "[", 
  RowBox[{"sph", ",", "E3", ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", "2", ",", "3"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"r", "[", "]"}], ",", 
     RowBox[{"\[Theta]", "[", "]"}], ",", 
     RowBox[{"\[Phi]", "[", "]"}]}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.790869031725273*^9, 3.790869055120769*^9}},
 CellLabel->"In[226]:=",
 CellID->1747532039,ExpressionUUID->"b0176ee2-8ecb-4d0c-be84-23e05886ebba"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The previous step just declared that there are 3 coordinate fields on E3 and \
specified their names, but it did not yet declare them as being spherical \
coordinates. The coordinates are spherical because the metric has this form \
in their coordinate basis. It is fundamental to note that the coordinate \
fields are given as r[], \[Theta][], etc, not just as r, \[Theta]. The empty \
pair of brackets is what indicates that they are scalar fields:\
\>", "MathCaption",
 CellChangeTimes->{{3.7908690819182043`*^9, 3.790869122038159*^9}, {
  3.791378237170374*^9, 3.791378253272827*^9}, {3.7913782918563766`*^9, 
  3.7913783753725557`*^9}, {3.791378705123269*^9, 3.791378766393764*^9}},
 CellID->249981071,ExpressionUUID->"b8eb0a82-ed2c-42ef-a1f7-6356fe7d8abf"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"met", "=", 
  RowBox[{"CTensor", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"1", ",", "0", ",", "0"}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{
         RowBox[{"r", "[", "]"}], "^", "2"}], ",", "0"}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", "0", ",", 
        RowBox[{
         RowBox[{
          RowBox[{"r", "[", "]"}], "^", "2"}], 
         RowBox[{
          RowBox[{"Sin", "[", 
           RowBox[{"\[Theta]", "[", "]"}], "]"}], "^", "2"}]}]}], "}"}]}], 
     "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "sph"}], ",", 
      RowBox[{"-", "sph"}]}], "}"}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.790869077739669*^9, 3.790869078847818*^9}, {
  3.790869123589923*^9, 3.790869155494154*^9}},
 CellLabel->"In[227]:=",
 CellID->1520839681,ExpressionUUID->"b7b95916-e133-473e-8fc2-90c6f75a07ef"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "0", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"], ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", 
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"], " ", 
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Sin",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Sin,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"]}]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "sph"}], ",", 
     RowBox[{"-", "sph"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.790869156347746*^9, 3.790871239626603*^9, 
  3.790874171335096*^9, 3.790874278608671*^9, 3.7908750445114803`*^9, 
  3.7908758257736998`*^9, 3.7908765862346573`*^9, 3.790876663363796*^9, 
  3.790876702398246*^9, 3.7913784036169453`*^9, 3.791381687518841*^9, 
  3.791389641786085*^9, 3.7913897244849977`*^9, 3.7913897877685738`*^9, 
  3.791390670929275*^9, 3.7945179473294153`*^9, 3.794518454523644*^9, 
  3.794518555408395*^9, 3.794518683028182*^9, 3.794518790122414*^9, 
  3.7945190976060038`*^9, 3.7945206885885344`*^9},
 CellLabel->"Out[227]=",
 CellID->132570903,ExpressionUUID->"3d1aaf4e-ac79-44a1-b2dd-f13e29a5ab27"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Now declare that this tensor field will be used to raise and lower indices. \
Yes, this is important even if we are working on a flat space, because we are \
not using an orthonormal basis. The second argument specifies that the metric \
is given as field in terms of the coordinates of the chart sph. The signature \
of this metric has three +1s, zero -1s and zero 0s:\
\>", "MathCaption",
 CellChangeTimes->{{3.790869215498747*^9, 3.7908692719520073`*^9}, {
  3.7908693134895887`*^9, 3.790869342278859*^9}, {3.7913784216924047`*^9, 
  3.7913784383157043`*^9}, {3.7913785067936163`*^9, 3.7913785393928747`*^9}},
 CellID->737204209,ExpressionUUID->"1c8caed1-224b-4d6a-ac87-a09f3c66a102"],

Cell[BoxData[
 RowBox[{"SetCMetric", "[", 
  RowBox[{"met", ",", "sph", ",", 
   RowBox[{"SignatureOfMetric", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"3", ",", "0", ",", "0"}], "}"}]}]}], "]"}]], "Input",
 CellChangeTimes->{{3.7908691732696333`*^9, 3.790869209483489*^9}},
 CellLabel->"In[228]:=",
 CellID->766683321,ExpressionUUID->"57948b00-1fd7-4c1b-8a20-f8660164a6cc"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Already the metric has different forms depending on the index structure:\
\>", "MathCaption",
 CellChangeTimes->{{3.790869396004573*^9, 3.79086940854049*^9}},
 CellID->1533747171,ExpressionUUID->"6d4133b7-fce6-4f23-8197-803a5e50f085"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"met", "[", 
  RowBox[{
   RowBox[{"-", "i"}], ",", 
   RowBox[{"-", "j"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.790869410472232*^9, 3.790869414780115*^9}},
 CellLabel->"In[229]:=",
 CellID->1451294931,ExpressionUUID->"6344bee0-b29f-4509-b0ae-f232d2c31557"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {1, 1} -> 1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 3} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox[
                "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                $CellContext`r[], Editable -> False], "2"]}]],
           Annotation[#, {2, 2} -> $CellContext`r[]^2, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 3} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"], " ", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], "2"]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", 
              RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                  "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`r[], Editable -> False], "2"], " ", 
                SuperscriptBox[
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                   InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                 "2"]}]}]],
           Annotation[#, {3, 3} -> $CellContext`r[]^2 Sin[
                $CellContext`\[Theta][]]^2, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 4 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 4 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " "},
         {"i", "j"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{1, 0, 0}, {0, $CellContext`r[]^2, 0}, {
    0, 0, $CellContext`r[]^2 Sin[
        $CellContext`\[Theta][]]^2}}, {-$CellContext`sph, -$CellContext`sph}, 
   0][-$CellContext`i, -$CellContext`j],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.79086941512783*^9, 3.790871244013434*^9, 3.790874171458811*^9, 
   3.7908742787146807`*^9, 3.79087504461762*^9, 3.790875825878504*^9, 
   3.7908765863482122`*^9, 3.790876663471374*^9, 3.790876702513957*^9, {
   3.7913785838451*^9, 3.791378598093437*^9}, 3.791381687623144*^9, 
   3.791389641891571*^9, 3.7913897245899887`*^9, 3.791389787874174*^9, 
   3.791390671038239*^9, 3.7945179474075327`*^9, 3.7945184545990458`*^9, 
   3.794518558148847*^9, 3.794518685698453*^9, 3.794518790232278*^9, 
   3.7945191020303507`*^9, 3.7945206925081577`*^9},
 CellLabel->"Out[229]=",
 CellID->1339399285,ExpressionUUID->"cbff5f4f-09b5-47a0-b7a9-ba40e6e161af"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"met", "[", 
  RowBox[{
   RowBox[{"-", "i"}], ",", "j"}], "]"}]], "Input",
 CellChangeTimes->{{3.790869422015403*^9, 3.7908694249401617`*^9}},
 CellLabel->"In[230]:=",
 CellID->1652864636,ExpressionUUID->"004633b5-c551-42ee-a683-9681bd964c90"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {1, 1} -> 1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 3} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {2, 2} -> 1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 3} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {3, 3} -> 1, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 4 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 4 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", "j"},
         {"i", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{1, 0, 0}, {0, 1, 0}, {0, 0, 
   1}}, {-$CellContext`sph, $CellContext`sph}, 
   0][-$CellContext`i, $CellContext`j],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.790869425282598*^9, 3.79087124669241*^9, 3.790874171516157*^9, 
   3.79087427876756*^9, 3.790875044664661*^9, 3.790875825934332*^9, 
   3.790876586399847*^9, 3.790876663523863*^9, 3.790876702579411*^9, {
   3.7913785852367573`*^9, 3.791378598928595*^9}, 3.7913816876749268`*^9, 
   3.7913896419411297`*^9, 3.79138972464268*^9, 3.791389787927986*^9, 
   3.791390671094516*^9, 3.794517947450201*^9, 3.794518454648628*^9, 
   3.794518559360243*^9, 3.7945186868122387`*^9, 3.794518790299383*^9, 
   3.7945191027648773`*^9, 3.794520693806831*^9},
 CellLabel->"Out[230]=",
 CellID->1849594818,ExpressionUUID->"db26364b-fd67-4e37-9199-e5603d5aad17"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"met", "[", 
  RowBox[{"i", ",", 
   RowBox[{"-", "j"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.790869426431405*^9, 3.790869429115952*^9}},
 CellLabel->"In[231]:=",
 CellID->890035918,ExpressionUUID->"d9a13145-2320-4904-bac6-cb168c76d66b"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {1, 1} -> 1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 3} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {2, 2} -> 1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 3} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {3, 3} -> 1, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 4 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 4 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"i", " "},
         {" ", "j"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{1, 0, 0}, {0, 1, 0}, {0, 0, 
   1}}, {$CellContext`sph, -$CellContext`sph}, 
   0][$CellContext`i, -$CellContext`j],
  Editable->False]], "Output",
 CellChangeTimes->{3.79086942948144*^9, 3.790871247260108*^9, 
  3.790874171570157*^9, 3.790874278823196*^9, 3.790875044723179*^9, 
  3.7908758259895353`*^9, 3.790876586464603*^9, 3.7908766635751247`*^9, 
  3.790876702645328*^9, 3.791378599844163*^9, 3.791381687730671*^9, 
  3.7913896420009203`*^9, 3.7913897246948442`*^9, 3.791389787980959*^9, 
  3.791390671148796*^9, 3.794517947493987*^9, 3.794518454695848*^9, 
  3.794518560110903*^9, 3.794518687363063*^9, 3.794518790355686*^9, 
  3.7945191032550793`*^9, 3.794520695045355*^9},
 CellLabel->"Out[231]=",
 CellID->260753550,ExpressionUUID->"32800829-1542-4453-a32b-f22647bade22"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"met", "[", 
  RowBox[{"i", ",", "j"}], "]"}]], "Input",
 CellChangeTimes->{{3.7908694301747437`*^9, 3.790869432315819*^9}},
 CellLabel->"In[232]:=",
 CellID->370060660,ExpressionUUID->"8848ed1f-1495-4e8b-b8be-ea08dcc154f5"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {1, 1} -> 1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 3} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox["1", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              FractionBox["1", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`r[], Editable -> False], "2"]]}]],
           Annotation[#, {2, 2} -> $CellContext`r[]^(-2), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 3} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox[
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Csc",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Csc,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], "2"], 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", 
              FractionBox[
               SuperscriptBox[
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"Csc"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Csc, Editable -> False], "[", 
                  InterpretationBox[
                   StyleBox[
                   "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`\[Theta][], Editable -> False], "]"}], "2"], 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`r[], Editable -> False], "2"]]}]],
           Annotation[#, {3, 3} -> Csc[
                $CellContext`\[Theta][]]^2 $CellContext`r[]^(-2), 
            "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 4 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 4 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"i", "j"},
         {" ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{1, 0, 0}, {0, $CellContext`r[]^(-2), 0}, {0, 0, Csc[
        $CellContext`\[Theta][]]^2 $CellContext`r[]^(-2)}}, \
{$CellContext`sph, $CellContext`sph}, 0][$CellContext`i, $CellContext`j],
  Editable->False]], "Output",
 CellChangeTimes->{3.7908694326657743`*^9, 3.7908712478289337`*^9, 
  3.7908741716260157`*^9, 3.7908742788781357`*^9, 3.790875044768622*^9, 
  3.79087582604672*^9, 3.7908765865262947`*^9, 3.7908766636273737`*^9, 
  3.790876702709463*^9, 3.7913786005968513`*^9, 3.791381687779469*^9, 
  3.7913896420546722`*^9, 3.791389724750111*^9, 3.791389788038348*^9, 
  3.791390671204214*^9, 3.79451794757104*^9, 3.794518454736682*^9, 
  3.794518561195664*^9, 3.79451868830016*^9, 3.7945187904073277`*^9, 
  3.7945191040816402`*^9, 3.79452069610867*^9},
 CellLabel->"Out[232]=",
 CellID->1028464144,ExpressionUUID->"bb2cefd1-dabb-400b-8a42-42734a95e4c8"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Take a covariant vector field, given by its three components in the sph \
vector basis:\
\>", "MathCaption",
 CellChangeTimes->{{3.790869460465385*^9, 3.790869470578167*^9}, {
  3.790869608917232*^9, 3.790869609853649*^9}, {3.7908697797187033`*^9, 
  3.790869780182411*^9}, {3.791378642285108*^9, 3.7913786613729467`*^9}},
 CellID->84579354,ExpressionUUID->"2ed84eae-fb15-4418-a113-862a6985d2d7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"v", "=", 
  RowBox[{"CTensor", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       RowBox[{"r", "[", "]"}], "^", "2"}], ",", 
      RowBox[{
       RowBox[{"r", "[", "]"}], 
       RowBox[{"Cos", "[", 
        RowBox[{"\[Theta]", "[", "]"}], "]"}]}], ",", 
      RowBox[{"r", "[", "]"}]}], "}"}], ",", 
    RowBox[{"{", "sph", "}"}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.7908695426920147`*^9, 3.790869546400902*^9}, {
  3.7908695845679073`*^9, 3.790869623773601*^9}, {3.790869750143873*^9, 
  3.79086975599221*^9}},
 CellLabel->"In[233]:=",
 CellID->1416225960,ExpressionUUID->"915d2044-a6aa-437f-aab1-563e7e674603"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], ",", 
     RowBox[{
      RowBox[{
       InterpretationBox[
        StyleBox["Cos",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Cos,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}], "}"}], ",", 
   RowBox[{"{", "sph", "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{{3.790869602866747*^9, 3.79086962425572*^9}, {
   3.79086975671717*^9, 3.790869781629381*^9}, 3.79087124914589*^9, 
   3.790874171682027*^9, 3.790874278928089*^9, 3.790875044824243*^9, 
   3.7908758260992928`*^9, 3.790876586578642*^9, 3.790876663682158*^9, 
   3.790876702756239*^9, 3.791378622242655*^9, 3.791381687832123*^9, 
   3.791389642102334*^9, 3.791389724799189*^9, 3.791389788090667*^9, 
   3.7913906712538548`*^9, 3.794517947612186*^9, 3.794518454782208*^9, 
   3.794518562614547*^9, 3.7945186895409937`*^9, 3.794518790458153*^9, 
   3.794519105512512*^9, 3.794520697368621*^9},
 CellLabel->"Out[233]=",
 CellID->1183323500,ExpressionUUID->"cf8722f0-9085-4d67-963d-b6a8133c3657"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"v", "[", "i", "]"}]], "Input",
 CellChangeTimes->{{3.790869614600861*^9, 3.790869618445003*^9}},
 CellLabel->"In[234]:=",
 CellID->285499045,ExpressionUUID->"e9691046-644d-46bb-994c-e70c2a0650f8"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"],
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox[
                "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                $CellContext`r[], Editable -> False], "2"]}]],
           Annotation[#, {1} -> $CellContext`r[]^2, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox["Cos",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cos,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], " ", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]}],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"Cos"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cos, Editable -> False], "[", 
                  InterpretationBox[
                   StyleBox[
                   "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`\[Theta][], Editable -> False], "]"}], " ", 
                InterpretationBox[
                 StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`r[], Editable -> False]}]}]],
           Annotation[#, {2} -> Cos[
               $CellContext`\[Theta][]] $CellContext`r[], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False],
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", 
              InterpretationBox[
               StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> False], 
               $CellContext`r[], Editable -> False]}]],
           Annotation[#, {3} -> $CellContext`r[], "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"i"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{$CellContext`r[]^2, Cos[
      $CellContext`\[Theta][]] $CellContext`r[], 
    $CellContext`r[]}, {$CellContext`sph}, 0][$CellContext`i],
  Editable->False]], "Output",
 CellChangeTimes->{{3.790869615563468*^9, 3.790869625984188*^9}, {
   3.790869758117681*^9, 3.790869782804626*^9}, 3.790871249749167*^9, 
   3.790874171735365*^9, 3.790874278983658*^9, 3.790875044868422*^9, 
   3.790875826155775*^9, 3.790876586641716*^9, 3.7908766637489357`*^9, 
   3.7908767028261766`*^9, 3.791378627363595*^9, 3.79138168788031*^9, 
   3.791389642157257*^9, 3.7913897248538523`*^9, 3.791389788147129*^9, 
   3.791390671313725*^9, 3.79451794764968*^9, 3.794518454827907*^9, 
   3.79451856336007*^9, 3.794518690043014*^9, 3.794518790508876*^9, 
   3.794519106301848*^9, 3.794520697920062*^9},
 CellLabel->"Out[234]=",
 CellID->128632176,ExpressionUUID->"dd689ca1-471b-4b89-ab11-3724c24aaa2c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"v", "[", 
  RowBox[{"-", "i"}], "]"}]], "Input",
 CellChangeTimes->{{3.790869634807685*^9, 3.7908696359806023`*^9}},
 CellLabel->"In[235]:=",
 CellID->390796769,ExpressionUUID->"3ebf67a1-a823-48cd-8d30-78d8f48b8ada"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"],
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox[
                "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                $CellContext`r[], Editable -> False], "2"]}]],
           Annotation[#, {1} -> $CellContext`r[]^2, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox["Cos",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cos,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "3"]}],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"Cos"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cos, Editable -> False], "[", 
                  InterpretationBox[
                   StyleBox[
                   "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`\[Theta][], Editable -> False], "]"}], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                  "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`r[], Editable -> False], "3"]}]}]],
           Annotation[#, {2} -> Cos[
               $CellContext`\[Theta][]] $CellContext`r[]^3, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "3"], " ", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], "2"]}],
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", 
              RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                  "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`r[], Editable -> False], "3"], " ", 
                SuperscriptBox[
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                   InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                 "2"]}]}]],
           Annotation[#, {3} -> $CellContext`r[]^3 Sin[
                $CellContext`\[Theta][]]^2, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" "},
         {"i"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{$CellContext`r[]^2, Cos[
      $CellContext`\[Theta][]] $CellContext`r[]^3, $CellContext`r[]^3 Sin[
       $CellContext`\[Theta][]]^2}, {-$CellContext`sph}, 0][-$CellContext`i],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.790869636752451*^9, {3.790869759117381*^9, 3.7908697841076*^9}, 
   3.7908712508993*^9, 3.7908741719037457`*^9, 3.790874279038157*^9, 
   3.7908750449250107`*^9, 3.790875826216687*^9, 3.790876586703233*^9, 
   3.790876663814786*^9, 3.79087670287891*^9, 3.7913786295463667`*^9, 
   3.7913816879364147`*^9, 3.791389642210581*^9, 3.791389724904641*^9, 
   3.791389788204973*^9, 3.791390671369823*^9, 3.7945179476965113`*^9, 
   3.794518454877442*^9, 3.7945185644946136`*^9, 3.794518691040968*^9, 
   3.794518790561451*^9, 3.794519107398323*^9, 3.794520698976989*^9},
 CellLabel->"Out[235]=",
 CellID->107021857,ExpressionUUID->"9cd31672-443b-40b9-8c2e-0e435edc4021"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["This is its norm squared:", "MathCaption",
 CellChangeTimes->{{3.7908696500359983`*^9, 3.79086965392099*^9}, {
  3.7908720767360573`*^9, 3.790872078008497*^9}},
 CellID->1518995414,ExpressionUUID->"152a40b4-a216-4943-99e6-e73807e1856b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"met", "[", 
   RowBox[{
    RowBox[{"-", "i"}], ",", 
    RowBox[{"-", "j"}]}], "]"}], 
  RowBox[{"v", "[", "i", "]"}], 
  RowBox[{"v", "[", "j", "]"}]}]], "Input",
 CellChangeTimes->{{3.790869660271473*^9, 3.7908696915384893`*^9}, {
  3.790869770983953*^9, 3.790869776680518*^9}, {3.790871524220497*^9, 
  3.7908715339229727`*^9}},
 CellLabel->"In[236]:=",
 CellID->665169177,ExpressionUUID->"da2d510c-d6da-4d13-a4cc-7cd174ba7af1"],

Cell[BoxData[
 RowBox[{
  SuperscriptBox[
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False], "4"], "+", 
  RowBox[{
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Cos",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Cos,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "4"]}], "+", 
  RowBox[{
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "4"], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}]}]], "Output",
 CellChangeTimes->{{3.790869663855303*^9, 3.790869691873629*^9}, {
   3.790869761262836*^9, 3.790869790069371*^9}, 3.790871251970986*^9, 
   3.790871534386106*^9, 3.79087417195467*^9, 3.7908742790934057`*^9, 
   3.790875044975588*^9, 3.790875826281558*^9, 3.790876586760645*^9, 
   3.790876663872561*^9, 3.7908767029280663`*^9, 3.791378670264826*^9, 
   3.7913816879861927`*^9, 3.7913896422663927`*^9, 3.7913897249572477`*^9, 
   3.791389788257979*^9, 3.791390671422529*^9, 3.794517947732998*^9, 
   3.794518454924902*^9, 3.794518566030739*^9, 3.7945186924987097`*^9, 
   3.794518790613353*^9, 3.7945191086538267`*^9, 3.794520700200695*^9},
 CellLabel->"Out[236]=",
 CellID->1731000397,ExpressionUUID->"75cf4ea1-2fe8-43f4-8aed-4ad62d75e814"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["It can also be computed as", "MathCaption",
 CellChangeTimes->{{3.790871536426777*^9, 3.790871540930702*^9}},
 CellID->1863649492,ExpressionUUID->"b84729a8-4341-402a-bffe-3e41f736ba6f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"v", "[", "i", "]"}], 
  RowBox[{"v", "[", 
   RowBox[{"-", "i"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.790871542662118*^9, 3.7908715455387287`*^9}},
 CellLabel->"In[237]:=",
 CellID->678840868,ExpressionUUID->"b6d7e278-2e44-48cf-b03a-fc0d72c41f97"],

Cell[BoxData[
 RowBox[{
  SuperscriptBox[
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False], "4"], "+", 
  RowBox[{
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Cos",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Cos,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "4"]}], "+", 
  RowBox[{
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "4"], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}]}]], "Output",
 CellChangeTimes->{3.790871545959138*^9, 3.790874172008152*^9, 
  3.7908742791466017`*^9, 3.7908750450177393`*^9, 3.79087582635323*^9, 
  3.790876586817893*^9, 3.790876663932077*^9, 3.7908767029772377`*^9, 
  3.791378676432302*^9, 3.791381688036783*^9, 3.791389642318372*^9, 
  3.791389725004031*^9, 3.7913897883110228`*^9, 3.791390671474909*^9, 
  3.7945179477693653`*^9, 3.794518454966283*^9, 3.794518568550322*^9, 
  3.794518694051331*^9, 3.794518790668592*^9, 3.794519110668562*^9, 
  3.794520701776943*^9},
 CellLabel->"Out[237]=",
 CellID->394812275,ExpressionUUID->"95679bc6-ba94-435f-a6cb-0b40ee5c6189"]
}, Open  ]]
}, Open  ]],

Cell["\<\
To compare with standard results, let us introduce an orthonormal basis. The \
orthonormal basis is not a coordinated basis and therefore cannot be defined \
using DefChart via another chart (aka coordinate system). We need to do it \
using DefBasis and provide a transformation matrix from the (already \
declared) coordinated basis.\
\>", "Text",
 CellChangeTimes->{{3.790863109490128*^9, 3.790863258626314*^9}, {
  3.790866420158791*^9, 3.790866466799429*^9}, {3.790867225498068*^9, 
  3.79086723724268*^9}, {3.790868709189741*^9, 3.7908689326766863`*^9}, {
  3.790869830533938*^9, 3.79086991508281*^9}, {3.79087041083344*^9, 
  3.790870426871901*^9}, {3.7908714124562893`*^9, 3.790871493884232*^9}, {
  3.791378803296052*^9, 3.7913788502310553`*^9}},
 CellID->1033628766,ExpressionUUID->"21851168-dd41-4f1a-9abf-b2f48e97d2af"],

Cell[CellGroupData[{

Cell["\<\
These are the three vector fields of the coordinated basis sph. The \
contravariant abstract index i indicates they are vector fields. The red \
(meaning basis sph) basis indices number the vectors from 1 to 3:\
\>", "MathCaption",
 CellChangeTimes->{{3.7908715010360727`*^9, 3.790871510276013*^9}, {
  3.790871834096127*^9, 3.790871851120332*^9}, {3.79087206680023*^9, 
  3.790872069360618*^9}, {3.791378887604738*^9, 3.791378905477313*^9}, {
  3.791378954610086*^9, 3.791378974018628*^9}, {3.791379040984434*^9, 
  3.791379101814938*^9}},
 CellID->1802171964,ExpressionUUID->"eed4577e-2413-4ac6-96f4-2a7c269787d1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "sph"}]}], "}"}], ",", "i"}], "]"}], ",", 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"2", ",", 
       RowBox[{"-", "sph"}]}], "}"}], ",", "i"}], "]"}], ",", 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"3", ",", 
       RowBox[{"-", "sph"}]}], "}"}], ",", "i"}], "]"}]}], "}"}]], "Input",
 CellChangeTimes->{{3.791378909541957*^9, 3.791378938293564*^9}},
 CellLabel->"In[238]:=",
 CellID->1928346054,ExpressionUUID->"debbc475-15ef-42f1-b3d4-c3311830fc82"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "i"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`sph}, $CellContext`i],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "i"},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`sph}, $CellContext`i],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "i"},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`sph}, $CellContext`i],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.791378938740049*^9, 3.7913816880925827`*^9, 
  3.791389642375287*^9, 3.791389725056655*^9, 3.7913897883695593`*^9, 
  3.7913906715295486`*^9, 3.79451794781073*^9, 3.794518455016548*^9, 
  3.794518572514312*^9, 3.794518696333947*^9, 3.794518790749185*^9, 
  3.794519112751813*^9, 3.794520703790577*^9},
 CellLabel->"Out[238]=",
 CellID->413143220,ExpressionUUID->"3bdec94b-6278-43ab-9e48-caf34db68fee"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The (square of the) norms of those coordinated basis vectors are not always 1:\
\>", "MathCaption",
 CellChangeTimes->{{3.7908715010360727`*^9, 3.790871510276013*^9}, {
  3.790871834096127*^9, 3.790871851120332*^9}, {3.79087206680023*^9, 
  3.790872069360618*^9}, {3.791379112390127*^9, 3.791379117837096*^9}},
 CellID->127573517,ExpressionUUID->"4353cd9e-3d30-454a-a14d-b97665ac64bc"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"met", "[", 
    RowBox[{
     RowBox[{"-", "i"}], ",", 
     RowBox[{"-", "j"}]}], "]"}], 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "sph"}]}], "}"}], ",", "i"}], "]"}], 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "sph"}]}], "}"}], ",", "j"}], "]"}]}], "//", 
  "ContractBasis"}]], "Input",
 CellChangeTimes->{{3.7908718538022757`*^9, 3.790871884079403*^9}, {
  3.7908719149654016`*^9, 3.7908719433412857`*^9}},
 CellLabel->"In[239]:=",
 CellID->1964538901,ExpressionUUID->"d62e71e2-d59f-4622-954c-fb94ec5af85d"],

Cell[BoxData["1"], "Output",
 CellChangeTimes->{{3.7908718630768023`*^9, 3.790871884574935*^9}, {
   3.7908719220693274`*^9, 3.7908719436798553`*^9}, 3.7908741720750647`*^9, 
   3.790874279199078*^9, 3.790875053218047*^9, 3.790875826406767*^9, 
   3.7908765868757343`*^9, 3.790876663985952*^9, 3.7908767036219063`*^9, 
   3.791381688142832*^9, 3.791389642428533*^9, 3.7913897251084538`*^9, 
   3.7913897884219723`*^9, 3.791390671585137*^9, 3.794517947847468*^9, 
   3.79451845505956*^9, 3.7945185739426737`*^9, 3.79451869761071*^9, 
   3.7945187908043118`*^9, 3.7945191145589437`*^9, 3.7945207053441677`*^9},
 CellLabel->"Out[239]=",
 CellID->1720729634,ExpressionUUID->"a777605e-3c74-489d-b8d3-b58067c80c98"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"met", "[", 
    RowBox[{
     RowBox[{"-", "i"}], ",", 
     RowBox[{"-", "j"}]}], "]"}], 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"2", ",", 
       RowBox[{"-", "sph"}]}], "}"}], ",", "i"}], "]"}], 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"2", ",", 
       RowBox[{"-", "sph"}]}], "}"}], ",", "j"}], "]"}]}], "//", 
  "ContractBasis"}]], "Input",
 CellChangeTimes->{{3.790871892461402*^9, 3.790871957004917*^9}},
 CellLabel->"In[240]:=",
 CellID->1465014692,ExpressionUUID->"04700e20-ab5c-40f8-9943-47943e4d2e14"],

Cell[BoxData[
 SuperscriptBox[
  InterpretationBox[
   StyleBox["r",
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`r[],
   Editable->False], "2"]], "Output",
 CellChangeTimes->{
  3.790871894669067*^9, {3.790871953187689*^9, 3.790871957416471*^9}, 
   3.790874172165517*^9, 3.790874279249133*^9, 3.790875053276079*^9, 
   3.79087582645782*^9, 3.7908765869344273`*^9, 3.7908766640354156`*^9, 
   3.7908767036964703`*^9, 3.791381688199341*^9, 3.791389642477788*^9, 
   3.7913897251561937`*^9, 3.7913897884766817`*^9, 3.791390671637916*^9, 
   3.794517947883522*^9, 3.794518455108073*^9, 3.794518574549747*^9, 
   3.79451869815532*^9, 3.794518790861228*^9, 3.7945191151308327`*^9, 
   3.794520705903605*^9},
 CellLabel->"Out[240]=",
 CellID->943675289,ExpressionUUID->"6dc4405c-793d-498f-9855-de4a9b87ba39"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"met", "[", 
    RowBox[{
     RowBox[{"-", "i"}], ",", 
     RowBox[{"-", "j"}]}], "]"}], 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"3", ",", 
       RowBox[{"-", "sph"}]}], "}"}], ",", "i"}], "]"}], 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"3", ",", 
       RowBox[{"-", "sph"}]}], "}"}], ",", "j"}], "]"}]}], "//", 
  "ContractBasis"}]], "Input",
 CellChangeTimes->{{3.790871899150612*^9, 3.790871920685667*^9}, {
  3.790871960492383*^9, 3.7908719714920673`*^9}},
 CellLabel->"In[241]:=",
 CellID->1935060583,ExpressionUUID->"54e603f0-ac3e-4d8b-a228-39ba133e7ef0"],

Cell[BoxData[
 RowBox[{
  SuperscriptBox[
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False], "2"], " ", 
  SuperscriptBox[
   RowBox[{
    InterpretationBox[
     StyleBox["Sin",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Sin,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}], "2"]}]], "Output",
 CellChangeTimes->{3.7908719023307056`*^9, 3.790871971843936*^9, 
  3.790874172221922*^9, 3.790874279302759*^9, 3.790875053333082*^9, 
  3.7908758265367823`*^9, 3.790876587023424*^9, 3.790876664090671*^9, 
  3.790876703751281*^9, 3.791381688246961*^9, 3.79138964255825*^9, 
  3.791389725206388*^9, 3.791389788528633*^9, 3.79139067169278*^9, 
  3.7945179479289303`*^9, 3.794518455150956*^9, 3.794518575598235*^9, 
  3.794518699218802*^9, 3.7945187909078836`*^9, 3.794519116180705*^9, 
  3.7945207071521397`*^9},
 CellLabel->"Out[241]=",
 CellID->1577671198,ExpressionUUID->"22ad6c51-b5e0-486c-a693-7ff3451a8af4"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["They are orthogonal, though:", "MathCaption",
 CellChangeTimes->{{3.790871984787944*^9, 3.7908720012031107`*^9}},
 CellID->185335001,ExpressionUUID->"996e3394-2ed2-4d88-8699-13e27dfbd93b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"met", "[", 
    RowBox[{
     RowBox[{"-", "i"}], ",", 
     RowBox[{"-", "j"}]}], "]"}], 
   RowBox[{
    RowBox[{"BasisArray", "[", 
     RowBox[{
      RowBox[{"-", "sph"}], ",", 
      RowBox[{"-", "sph"}]}], "]"}], "[", 
    RowBox[{"i", ",", "j"}], "]"}]}], "//", "ContractBasis"}]], "Input",
 CellChangeTimes->{{3.7908720031823673`*^9, 3.7908720315865307`*^9}},
 CellLabel->"In[242]:=",
 CellID->405821759,ExpressionUUID->"635ef241-ca06-4a62-a2d1-6e9f813ead18"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       RowBox[{
        InterpretationBox[
         StyleBox["Sin",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Sin,
         Editable->False], "[", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False], "]"}], "2"]}]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{{3.790872017302188*^9, 3.790872031982006*^9}, 
   3.7908741722946177`*^9, 3.790874280001017*^9, 3.7908750533810863`*^9, 
   3.7908758265931273`*^9, 3.790876587078401*^9, 3.790876664152875*^9, 
   3.7908767038112926`*^9, 3.7913816883034163`*^9, 3.7913896426422577`*^9, 
   3.791389725261338*^9, 3.791389788587669*^9, 3.791390671747369*^9, 
   3.794517947979102*^9, 3.794518455202023*^9, 3.79451857689332*^9, 
   3.7945187003961143`*^9, 3.7945187909673367`*^9, 3.794519117183147*^9, 
   3.794520708221764*^9},
 CellLabel->"Out[242]=",
 CellID->844695776,ExpressionUUID->"39ce3c4f-4599-41d1-8311-aac7d369a7e6"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Hence, to construct an orthonormal basis we just have to divide the vectors \
by their norms. This time we sill represent the new basis with blue color. \
Note that the basis is defined on the vbundle TangentE3 of the manifold E3, \
not on the manifold E3 itself:\
\>", "MathCaption",
 CellChangeTimes->{{3.790872041930179*^9, 3.7908720914479*^9}, {
  3.7913791698043823`*^9, 3.791379229858451*^9}},
 CellID->2112124970,ExpressionUUID->"9933235f-f632-4c83-a8b5-32d5e0f3ea96"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefBasis", "[", 
  RowBox[{"ortho", ",", "TangentE3", ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", "2", ",", "3"}], "}"}], ",", 
   RowBox[{"BasisColor", "\[Rule]", "Blue"}], ",", 
   RowBox[{"BasisChange", "\[Rule]", 
    RowBox[{"CTensor", "[", 
     RowBox[{
      RowBox[{"DiagonalMatrix", "[", 
       RowBox[{"{", 
        RowBox[{"1", ",", 
         RowBox[{"1", "/", 
          RowBox[{"r", "[", "]"}]}], ",", 
         RowBox[{"1", "/", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"r", "[", "]"}], 
            RowBox[{"Sin", "[", 
             RowBox[{"\[Theta]", "[", "]"}], "]"}]}], ")"}]}]}], "}"}], "]"}],
       ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "ortho"}], ",", "sph"}], "}"}]}], "]"}]}]}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.790872092730699*^9, 3.790872115767293*^9}, {
  3.790872187957431*^9, 3.790872193101244*^9}, {3.790872223517397*^9, 
  3.7908722597624617`*^9}},
 CellLabel->"In[243]:=",
 CellID->406172060,ExpressionUUID->"ba109330-948a-45a3-a783-e8f573b397a6"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`sph}, {1, $CellContext`ortho}],
     Editable->False], "\[Rule]", "1"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{1, -$CellContext`sph}, {1, $CellContext`ortho}] -> 1, 
   " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.790872116781603*^9, 3.79087226858707*^9, 
  3.790874172372891*^9, 3.7908742800898237`*^9, 3.790875054520811*^9, 
  3.7908758266723757`*^9, 3.790876587146179*^9, 3.790876664256629*^9, 
  3.790876704392824*^9, 3.791379246448139*^9, 3.79138168836451*^9, 
  3.791389642708082*^9, 3.791389725325252*^9, 3.791389788681539*^9, 
  3.791390671847569*^9, 3.79451794803996*^9, 3.794518455264621*^9, 
  3.794518579497794*^9, 3.794518702089163*^9, 3.7945187911107893`*^9, 
  3.794519119146768*^9, 3.794520710320738*^9},
 CellLabel->"During evaluation of In[243]:=",
 CellID->2024881592,ExpressionUUID->"ad91bc7c-2055-48e7-8a73-0936814457a9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`sph}, {2, $CellContext`ortho}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{1, -$CellContext`sph}, {2, $CellContext`ortho}] -> 0, 
   " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.790872116781603*^9, 3.79087226858707*^9, 
  3.790874172372891*^9, 3.7908742800898237`*^9, 3.790875054520811*^9, 
  3.7908758266723757`*^9, 3.790876587146179*^9, 3.790876664256629*^9, 
  3.790876704392824*^9, 3.791379246448139*^9, 3.79138168836451*^9, 
  3.791389642708082*^9, 3.791389725325252*^9, 3.791389788681539*^9, 
  3.791390671847569*^9, 3.79451794803996*^9, 3.794518455264621*^9, 
  3.794518579497794*^9, 3.794518702089163*^9, 3.7945187911107893`*^9, 
  3.794519119146768*^9, 3.7945207103398867`*^9},
 CellLabel->"During evaluation of In[243]:=",
 CellID->1011562735,ExpressionUUID->"0cb826aa-9ec3-4931-89cb-df7c2d904e6b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`sph}, {3, $CellContext`ortho}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{1, -$CellContext`sph}, {3, $CellContext`ortho}] -> 0, 
   " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.790872116781603*^9, 3.79087226858707*^9, 
  3.790874172372891*^9, 3.7908742800898237`*^9, 3.790875054520811*^9, 
  3.7908758266723757`*^9, 3.790876587146179*^9, 3.790876664256629*^9, 
  3.790876704392824*^9, 3.791379246448139*^9, 3.79138168836451*^9, 
  3.791389642708082*^9, 3.791389725325252*^9, 3.791389788681539*^9, 
  3.791390671847569*^9, 3.79451794803996*^9, 3.794518455264621*^9, 
  3.794518579497794*^9, 3.794518702089163*^9, 3.7945187911107893`*^9, 
  3.794519119146768*^9, 3.7945207103835783`*^9},
 CellLabel->"During evaluation of In[243]:=",
 CellID->1854337493,ExpressionUUID->"84086254-e91b-4a6c-bfba-64339f00b534"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`sph}, {1, $CellContext`ortho}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{2, -$CellContext`sph}, {1, $CellContext`ortho}] -> 0, 
   " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.790872116781603*^9, 3.79087226858707*^9, 
  3.790874172372891*^9, 3.7908742800898237`*^9, 3.790875054520811*^9, 
  3.7908758266723757`*^9, 3.790876587146179*^9, 3.790876664256629*^9, 
  3.790876704392824*^9, 3.791379246448139*^9, 3.79138168836451*^9, 
  3.791389642708082*^9, 3.791389725325252*^9, 3.791389788681539*^9, 
  3.791390671847569*^9, 3.79451794803996*^9, 3.794518455264621*^9, 
  3.794518579497794*^9, 3.794518702089163*^9, 3.7945187911107893`*^9, 
  3.794519119146768*^9, 3.794520710419491*^9},
 CellLabel->"During evaluation of In[243]:=",
 CellID->1567559587,ExpressionUUID->"f24a3335-0353-434b-9f54-6bda18b705ef"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`sph}, {2, $CellContext`ortho}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{2, -$CellContext`sph}, {
     2, $CellContext`ortho}] -> $CellContext`r[], " for tensor ", 
   xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.790872116781603*^9, 3.79087226858707*^9, 
  3.790874172372891*^9, 3.7908742800898237`*^9, 3.790875054520811*^9, 
  3.7908758266723757`*^9, 3.790876587146179*^9, 3.790876664256629*^9, 
  3.790876704392824*^9, 3.791379246448139*^9, 3.79138168836451*^9, 
  3.791389642708082*^9, 3.791389725325252*^9, 3.791389788681539*^9, 
  3.791390671847569*^9, 3.79451794803996*^9, 3.794518455264621*^9, 
  3.794518579497794*^9, 3.794518702089163*^9, 3.7945187911107893`*^9, 
  3.794519119146768*^9, 3.794520710435763*^9},
 CellLabel->"During evaluation of In[243]:=",
 CellID->644597313,ExpressionUUID->"187ec5c3-f2b9-4e59-ae86-0f943c7433ac"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`sph}, {3, $CellContext`ortho}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{2, -$CellContext`sph}, {3, $CellContext`ortho}] -> 0, 
   " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.790872116781603*^9, 3.79087226858707*^9, 
  3.790874172372891*^9, 3.7908742800898237`*^9, 3.790875054520811*^9, 
  3.7908758266723757`*^9, 3.790876587146179*^9, 3.790876664256629*^9, 
  3.790876704392824*^9, 3.791379246448139*^9, 3.79138168836451*^9, 
  3.791389642708082*^9, 3.791389725325252*^9, 3.791389788681539*^9, 
  3.791390671847569*^9, 3.79451794803996*^9, 3.794518455264621*^9, 
  3.794518579497794*^9, 3.794518702089163*^9, 3.7945187911107893`*^9, 
  3.794519119146768*^9, 3.794520710471396*^9},
 CellLabel->"During evaluation of In[243]:=",
 CellID->1839524123,ExpressionUUID->"6873306b-40eb-4179-a646-210113ea8024"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`sph}, {1, $CellContext`ortho}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{3, -$CellContext`sph}, {1, $CellContext`ortho}] -> 0, 
   " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.790872116781603*^9, 3.79087226858707*^9, 
  3.790874172372891*^9, 3.7908742800898237`*^9, 3.790875054520811*^9, 
  3.7908758266723757`*^9, 3.790876587146179*^9, 3.790876664256629*^9, 
  3.790876704392824*^9, 3.791379246448139*^9, 3.79138168836451*^9, 
  3.791389642708082*^9, 3.791389725325252*^9, 3.791389788681539*^9, 
  3.791390671847569*^9, 3.79451794803996*^9, 3.794518455264621*^9, 
  3.794518579497794*^9, 3.794518702089163*^9, 3.7945187911107893`*^9, 
  3.794519119146768*^9, 3.7945207104860687`*^9},
 CellLabel->"During evaluation of In[243]:=",
 CellID->1707353049,ExpressionUUID->"ff00565a-58f8-4c1f-b79c-059126372ae2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`sph}, {2, $CellContext`ortho}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{3, -$CellContext`sph}, {2, $CellContext`ortho}] -> 0, 
   " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.790872116781603*^9, 3.79087226858707*^9, 
  3.790874172372891*^9, 3.7908742800898237`*^9, 3.790875054520811*^9, 
  3.7908758266723757`*^9, 3.790876587146179*^9, 3.790876664256629*^9, 
  3.790876704392824*^9, 3.791379246448139*^9, 3.79138168836451*^9, 
  3.791389642708082*^9, 3.791389725325252*^9, 3.791389788681539*^9, 
  3.791390671847569*^9, 3.79451794803996*^9, 3.794518455264621*^9, 
  3.794518579497794*^9, 3.794518702089163*^9, 3.7945187911107893`*^9, 
  3.794519119146768*^9, 3.794520710519693*^9},
 CellLabel->"During evaluation of In[243]:=",
 CellID->806203329,ExpressionUUID->"0ed60525-a3aa-493d-a881-dda77d11662e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`sph}, {3, $CellContext`ortho}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], " ", 
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}]}]}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{3, -$CellContext`sph}, {
     3, $CellContext`ortho}] -> $CellContext`r[] Sin[
      $CellContext`\[Theta][]], " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.790872116781603*^9, 3.79087226858707*^9, 
  3.790874172372891*^9, 3.7908742800898237`*^9, 3.790875054520811*^9, 
  3.7908758266723757`*^9, 3.790876587146179*^9, 3.790876664256629*^9, 
  3.790876704392824*^9, 3.791379246448139*^9, 3.79138168836451*^9, 
  3.791389642708082*^9, 3.791389725325252*^9, 3.791389788681539*^9, 
  3.791390671847569*^9, 3.79451794803996*^9, 3.794518455264621*^9, 
  3.794518579497794*^9, 3.794518702089163*^9, 3.7945187911107893`*^9, 
  3.794519119146768*^9, 3.7945207105361643`*^9},
 CellLabel->"During evaluation of In[243]:=",
 CellID->1385021580,ExpressionUUID->"84cdad14-0b86-449f-bf2d-55efd727d4ba"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`ortho}, {1, $CellContext`sph}],
     Editable->False], "\[Rule]", "1"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{1, -$CellContext`ortho}, {1, $CellContext`sph}] -> 1, 
   " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.790872116781603*^9, 3.79087226858707*^9, 
  3.790874172372891*^9, 3.7908742800898237`*^9, 3.790875054520811*^9, 
  3.7908758266723757`*^9, 3.790876587146179*^9, 3.790876664256629*^9, 
  3.790876704392824*^9, 3.791379246448139*^9, 3.79138168836451*^9, 
  3.791389642708082*^9, 3.791389725325252*^9, 3.791389788681539*^9, 
  3.791390671847569*^9, 3.79451794803996*^9, 3.794518455264621*^9, 
  3.794518579497794*^9, 3.794518702089163*^9, 3.7945187911107893`*^9, 
  3.794519119146768*^9, 3.794520710572627*^9},
 CellLabel->"During evaluation of In[243]:=",
 CellID->1482445227,ExpressionUUID->"5d04b2f5-26d7-437b-807f-7cb94099aa3c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`ortho}, {2, $CellContext`sph}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{1, -$CellContext`ortho}, {2, $CellContext`sph}] -> 0, 
   " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.790872116781603*^9, 3.79087226858707*^9, 
  3.790874172372891*^9, 3.7908742800898237`*^9, 3.790875054520811*^9, 
  3.7908758266723757`*^9, 3.790876587146179*^9, 3.790876664256629*^9, 
  3.790876704392824*^9, 3.791379246448139*^9, 3.79138168836451*^9, 
  3.791389642708082*^9, 3.791389725325252*^9, 3.791389788681539*^9, 
  3.791390671847569*^9, 3.79451794803996*^9, 3.794518455264621*^9, 
  3.794518579497794*^9, 3.794518702089163*^9, 3.7945187911107893`*^9, 
  3.794519119146768*^9, 3.794520710589724*^9},
 CellLabel->"During evaluation of In[243]:=",
 CellID->351657695,ExpressionUUID->"de6e9089-74c2-43d6-a76f-3ff77ebb6a09"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`ortho}, {3, $CellContext`sph}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{1, -$CellContext`ortho}, {3, $CellContext`sph}] -> 0, 
   " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.790872116781603*^9, 3.79087226858707*^9, 
  3.790874172372891*^9, 3.7908742800898237`*^9, 3.790875054520811*^9, 
  3.7908758266723757`*^9, 3.790876587146179*^9, 3.790876664256629*^9, 
  3.790876704392824*^9, 3.791379246448139*^9, 3.79138168836451*^9, 
  3.791389642708082*^9, 3.791389725325252*^9, 3.791389788681539*^9, 
  3.791390671847569*^9, 3.79451794803996*^9, 3.794518455264621*^9, 
  3.794518579497794*^9, 3.794518702089163*^9, 3.7945187911107893`*^9, 
  3.794519119146768*^9, 3.7945207106225863`*^9},
 CellLabel->"During evaluation of In[243]:=",
 CellID->435883321,ExpressionUUID->"c10f058f-d9c2-4c9f-b0d4-52906d77dcbe"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`ortho}, {1, $CellContext`sph}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{2, -$CellContext`ortho}, {1, $CellContext`sph}] -> 0, 
   " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.790872116781603*^9, 3.79087226858707*^9, 
  3.790874172372891*^9, 3.7908742800898237`*^9, 3.790875054520811*^9, 
  3.7908758266723757`*^9, 3.790876587146179*^9, 3.790876664256629*^9, 
  3.790876704392824*^9, 3.791379246448139*^9, 3.79138168836451*^9, 
  3.791389642708082*^9, 3.791389725325252*^9, 3.791389788681539*^9, 
  3.791390671847569*^9, 3.79451794803996*^9, 3.794518455264621*^9, 
  3.794518579497794*^9, 3.794518702089163*^9, 3.7945187911107893`*^9, 
  3.794519119146768*^9, 3.794520710637903*^9},
 CellLabel->"During evaluation of In[243]:=",
 CellID->824256130,ExpressionUUID->"7ef9366c-74cb-4afb-ad51-91052f458719"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`ortho}, {2, $CellContext`sph}],
     Editable->False], "\[Rule]", 
    FractionBox["1", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{2, -$CellContext`ortho}, {
     2, $CellContext`sph}] -> $CellContext`r[]^(-1), " for tensor ", 
   xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.790872116781603*^9, 3.79087226858707*^9, 
  3.790874172372891*^9, 3.7908742800898237`*^9, 3.790875054520811*^9, 
  3.7908758266723757`*^9, 3.790876587146179*^9, 3.790876664256629*^9, 
  3.790876704392824*^9, 3.791379246448139*^9, 3.79138168836451*^9, 
  3.791389642708082*^9, 3.791389725325252*^9, 3.791389788681539*^9, 
  3.791390671847569*^9, 3.79451794803996*^9, 3.794518455264621*^9, 
  3.794518579497794*^9, 3.794518702089163*^9, 3.7945187911107893`*^9, 
  3.794519119146768*^9, 3.7945207106708193`*^9},
 CellLabel->"During evaluation of In[243]:=",
 CellID->2003253760,ExpressionUUID->"1031d50e-5544-4c4d-b672-d3459e393c21"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`ortho}, {3, $CellContext`sph}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{2, -$CellContext`ortho}, {3, $CellContext`sph}] -> 0, 
   " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.790872116781603*^9, 3.79087226858707*^9, 
  3.790874172372891*^9, 3.7908742800898237`*^9, 3.790875054520811*^9, 
  3.7908758266723757`*^9, 3.790876587146179*^9, 3.790876664256629*^9, 
  3.790876704392824*^9, 3.791379246448139*^9, 3.79138168836451*^9, 
  3.791389642708082*^9, 3.791389725325252*^9, 3.791389788681539*^9, 
  3.791390671847569*^9, 3.79451794803996*^9, 3.794518455264621*^9, 
  3.794518579497794*^9, 3.794518702089163*^9, 3.7945187911107893`*^9, 
  3.794519119146768*^9, 3.794520710687244*^9},
 CellLabel->"During evaluation of In[243]:=",
 CellID->437326654,ExpressionUUID->"d66658df-f2a8-432c-8385-232351acae78"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`ortho}, {1, $CellContext`sph}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{3, -$CellContext`ortho}, {1, $CellContext`sph}] -> 0, 
   " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.790872116781603*^9, 3.79087226858707*^9, 
  3.790874172372891*^9, 3.7908742800898237`*^9, 3.790875054520811*^9, 
  3.7908758266723757`*^9, 3.790876587146179*^9, 3.790876664256629*^9, 
  3.790876704392824*^9, 3.791379246448139*^9, 3.79138168836451*^9, 
  3.791389642708082*^9, 3.791389725325252*^9, 3.791389788681539*^9, 
  3.791390671847569*^9, 3.79451794803996*^9, 3.794518455264621*^9, 
  3.794518579497794*^9, 3.794518702089163*^9, 3.7945187911107893`*^9, 
  3.794519119146768*^9, 3.7945207107236433`*^9},
 CellLabel->"During evaluation of In[243]:=",
 CellID->1460033744,ExpressionUUID->"24fc12ea-5860-4987-9b20-5542ee0120e6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`ortho}, {2, $CellContext`sph}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{3, -$CellContext`ortho}, {2, $CellContext`sph}] -> 0, 
   " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.790872116781603*^9, 3.79087226858707*^9, 
  3.790874172372891*^9, 3.7908742800898237`*^9, 3.790875054520811*^9, 
  3.7908758266723757`*^9, 3.790876587146179*^9, 3.790876664256629*^9, 
  3.790876704392824*^9, 3.791379246448139*^9, 3.79138168836451*^9, 
  3.791389642708082*^9, 3.791389725325252*^9, 3.791389788681539*^9, 
  3.791390671847569*^9, 3.79451794803996*^9, 3.794518455264621*^9, 
  3.794518579497794*^9, 3.794518702089163*^9, 3.7945187911107893`*^9, 
  3.794519119146768*^9, 3.794520710739216*^9},
 CellLabel->"During evaluation of In[243]:=",
 CellID->1638984786,ExpressionUUID->"bfe157fd-3170-45e5-bedf-3d89258fa507"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`ortho}, {3, $CellContext`sph}],
     Editable->False], "\[Rule]", 
    FractionBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Csc",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Csc,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{3, -$CellContext`ortho}, {3, $CellContext`sph}] -> 
   Csc[
      $CellContext`\[Theta][]]/$CellContext`r[], " for tensor ", 
   xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.790872116781603*^9, 3.79087226858707*^9, 
  3.790874172372891*^9, 3.7908742800898237`*^9, 3.790875054520811*^9, 
  3.7908758266723757`*^9, 3.790876587146179*^9, 3.790876664256629*^9, 
  3.790876704392824*^9, 3.791379246448139*^9, 3.79138168836451*^9, 
  3.791389642708082*^9, 3.791389725325252*^9, 3.791389788681539*^9, 
  3.791390671847569*^9, 3.79451794803996*^9, 3.794518455264621*^9, 
  3.794518579497794*^9, 3.794518702089163*^9, 3.7945187911107893`*^9, 
  3.794519119146768*^9, 3.7945207107796926`*^9},
 CellLabel->"During evaluation of In[243]:=",
 CellID->644558233,ExpressionUUID->"64c3e4d7-4568-4b0c-8c4a-874b61ee9b20"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      UnderscriptBox[
       OverscriptBox["J", 
        StyleBox["~",
         FontColor->RGBColor[0, 0, 1]]], 
       StyleBox["~",
        FontColor->RGBColor[1, 0, 0]]],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Jacobianorthosph[],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      UnderscriptBox[
       OverscriptBox[
        FractionBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Csc",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Csc,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"]], 
        StyleBox["~",
         FontColor->RGBColor[0., 0., 1.]]], 
       StyleBox["~",
        FontColor->RGBColor[1., 0., 0.]]],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xCoba`CTensor[Csc[
        $CellContext`\[Theta][]] $CellContext`r[]^(-2), {}, \
$CellContext`ortho - $CellContext`sph][],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Jacobianorthosph"}],
  SequenceForm[
  "Added independent rule ", $CellContext`Jacobianorthosph[] -> 
   xAct`xCoba`CTensor[Csc[
       $CellContext`\[Theta][]] $CellContext`r[]^(-2), {}, $CellContext`ortho - \
$CellContext`sph][], " for tensor ", $CellContext`Jacobianorthosph],
  Editable->False]], "Print",
 CellChangeTimes->{3.790872116781603*^9, 3.79087226858707*^9, 
  3.790874172372891*^9, 3.7908742800898237`*^9, 3.790875054520811*^9, 
  3.7908758266723757`*^9, 3.790876587146179*^9, 3.790876664256629*^9, 
  3.790876704392824*^9, 3.791379246448139*^9, 3.79138168836451*^9, 
  3.791389642708082*^9, 3.791389725325252*^9, 3.791389788681539*^9, 
  3.791390671847569*^9, 3.79451794803996*^9, 3.794518455264621*^9, 
  3.794518579497794*^9, 3.794518702089163*^9, 3.7945187911107893`*^9, 
  3.794519119146768*^9, 3.7945207107900248`*^9},
 CellLabel->"During evaluation of In[243]:=",
 CellID->1182706563,ExpressionUUID->"42500356-87c9-47f3-aeb6-938114088b1e"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["These are the three orthonormal basis vectors:", "MathCaption",
 CellChangeTimes->{{3.790872380246313*^9, 3.7908723947880096`*^9}, {
  3.791379274529063*^9, 3.7913792752654552`*^9}},
 CellID->1259963770,ExpressionUUID->"7563648f-b089-47e4-a528-4e7254525b92"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"BasisArray", "[", 
   RowBox[{"-", "ortho"}], "]"}], "[", "i", "]"}]], "Input",
 CellLabel->"In[244]:=",
 CellID->753808889,ExpressionUUID->"3155a783-7817-4e1f-8fd8-f0e6526c41fe"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "i"},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`ortho}, $CellContext`i],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "i"},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`ortho}, $CellContext`i],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "i"},
           {
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`ortho}, $CellContext`i],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{{3.7908723916742764`*^9, 3.790872395895187*^9}, 
   3.7908741728406982`*^9, 3.790874280705063*^9, 3.790875056387274*^9, 
   3.790875827290773*^9, 3.790876587627077*^9, 3.790876664871011*^9, 
   3.7908767050920563`*^9, 3.7913792780884647`*^9, 3.791381688953896*^9, 
   3.791389643272813*^9, 3.791389725801467*^9, 3.791389789223826*^9, 
   3.791390672420973*^9, 3.794517948387147*^9, 3.794518455736693*^9, 
   3.794518584101746*^9, 3.794518705716837*^9, 3.79451879165814*^9, 
   3.794518825544973*^9, 3.7945188762053347`*^9, 3.794519121751182*^9, 
   3.794519153774427*^9, 3.79452071223035*^9, 3.794520761605372*^9},
 CellLabel->"Out[244]=",
 CellID->1537918424,ExpressionUUID->"8528a086-1541-4883-9280-3bfaee26a865"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Express them in the coordinate basis sph:", "MathCaption",
 CellChangeTimes->{{3.790872397309156*^9, 3.790872417915724*^9}},
 CellID->1485286390,ExpressionUUID->"ec2a70d9-9349-433c-93f8-4b79b94814fa"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"BasisExpand", "[", 
   RowBox[{"%", ",", "sph"}], "]"}], "//", "ToValues"}]], "Input",
 CellChangeTimes->{{3.7908722964509087`*^9, 3.790872377496463*^9}, 
   3.790872408987669*^9},
 CellLabel->"In[245]:=",
 CellID->1806190466,ExpressionUUID->"f7944bd8-2a29-4141-8cb0-e639f6b113a9"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "i"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`sph}, $CellContext`i],
    Editable->False], ",", 
   FractionBox[
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "i"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`sph}, $CellContext`i],
     Editable->False], 
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False]], ",", 
   FractionBox[
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", "i"},
             {
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "Rows" -> {{
                 Offset[0.]}}}],
           FontSize->9]}
        },
        GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{3, -$CellContext`sph}, $CellContext`i],
      Editable->False], " ", 
     RowBox[{
      InterpretationBox[
       StyleBox["Csc",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Csc,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}]}], 
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False]]}], "}"}]], "Output",
 CellChangeTimes->{{3.790872302968313*^9, 3.790872306331728*^9}, {
   3.7908723440411053`*^9, 3.790872369484696*^9}, 3.790872409398649*^9, 
   3.7908741729090776`*^9, 3.790874280803846*^9, 3.790875057885212*^9, 
   3.7908758273493633`*^9, 3.790876587683836*^9, 3.790876664936668*^9, 
   3.790876705166576*^9, 3.791379294422832*^9, 3.791381689013585*^9, 
   3.791389643326027*^9, 3.791389725858224*^9, 3.7913897892790527`*^9, 
   3.7913906724776993`*^9, 3.794517948566976*^9, 3.794518455894149*^9, 
   3.794518711780047*^9, {3.794520749063146*^9, 3.794520762653347*^9}},
 CellLabel->"Out[245]=",
 CellID->1917932673,ExpressionUUID->"0fa52a0f-9e55-48d0-9126-a1ccae189439"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["They are now orthonormal:", "MathCaption",
 CellChangeTimes->{{3.790871984787944*^9, 3.7908720012031107`*^9}, {
  3.790872445186379*^9, 3.790872449074511*^9}, {3.791379298327911*^9, 
  3.791379300504218*^9}},
 CellID->677715185,ExpressionUUID->"7ec9b2dd-0b11-4e62-b361-3f8eaeacf0f7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"met", "[", 
    RowBox[{
     RowBox[{"-", "i"}], ",", 
     RowBox[{"-", "j"}]}], "]"}], 
   RowBox[{
    RowBox[{"BasisArray", "[", 
     RowBox[{
      RowBox[{"-", "ortho"}], ",", 
      RowBox[{"-", "ortho"}]}], "]"}], "[", 
    RowBox[{"i", ",", "j"}], "]"}]}], "//", "ContractBasis"}]], "Input",
 CellChangeTimes->{{3.7908720031823673`*^9, 3.7908720315865307`*^9}, {
  3.7908724378432913`*^9, 3.790872441042946*^9}},
 CellLabel->"In[246]:=",
 CellID->105850607,ExpressionUUID->"193da3bd-e7ab-4026-991d-dc912cb3af7a"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "1"}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{{3.790872017302188*^9, 3.790872031982006*^9}, 
   3.790872442030423*^9, 3.790874172975026*^9, 3.7908742808717012`*^9, 
   3.7908750594475107`*^9, 3.790875827417396*^9, 3.790876587726428*^9, 
   3.79087666499975*^9, 3.790876705236794*^9, 3.791379303061943*^9, 
   3.791381689086707*^9, 3.791389643394589*^9, 3.7913897259282417`*^9, 
   3.7913897893466*^9, 3.791390672539371*^9, 3.79451794864107*^9, 
   3.794518455948718*^9, 3.794518715769568*^9, 3.794520764211277*^9},
 CellLabel->"Out[246]=",
 CellID->870818342,ExpressionUUID->"2aec1e78-ad08-4e24-97bc-3719d56ff96a"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.4.2. Differentiation in flat space", "Subsubsection",
 CellChangeTimes->{{3.728676331871439*^9, 3.728676332031807*^9}, 
   3.790876746494073*^9, {3.790876952785844*^9, 3.790876986414445*^9}},
 CellID->534635677,ExpressionUUID->"c62b6cab-7dd4-4353-87ab-ba8d1dd1789f"],

Cell["\<\
Now let us discuss differentiation. We need to distinguish the partial \
derivative with respect to the coordinate fields (which is independent of the \
choice of metric) from covariant differentiation (which is determined by the \
choice of metric).\
\>", "Text",
 CellChangeTimes->{{3.790863109490128*^9, 3.790863258626314*^9}, {
  3.790866420158791*^9, 3.790866466799429*^9}, {3.790867225498068*^9, 
  3.79086723724268*^9}, {3.790868709189741*^9, 3.7908689326766863`*^9}, {
  3.790869830533938*^9, 3.79086991508281*^9}, {3.79087041083344*^9, 
  3.790870426871901*^9}},
 CellID->1463034371,ExpressionUUID->"a52485ca-9153-4935-abaa-6d47e3509721"],

Cell[CellGroupData[{

Cell["\<\
The first thing to observe is that all derivative operators coincide on \
scalars, and give the coordinated basis vectors:\
\>", "MathCaption",
 CellChangeTimes->{{3.790869976472934*^9, 3.7908700047594757`*^9}, {
  3.7908700698851957`*^9, 3.790870099612248*^9}},
 CellID->2072147979,ExpressionUUID->"7692223d-a03f-41c3-9a32-253413e57809"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"-", "i"}], "]"}], "/@", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"r", "[", "]"}], ",", 
    RowBox[{"\[Theta]", "[", "]"}], ",", 
    RowBox[{"\[Phi]", "[", "]"}]}], "}"}]}]], "Input",
 CellChangeTimes->{{3.790870006898445*^9, 3.790870048669271*^9}},
 CellLabel->"In[247]:=",
 CellID->1120704634,ExpressionUUID->"7603b50b-4c4f-42a7-80ac-96bff97e32c7"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"i", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`i, {1, $CellContext`sph}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"i", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`i, {2, $CellContext`sph}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"i", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`i, {3, $CellContext`sph}],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{{3.790870031124381*^9, 3.790870049074205*^9}, 
   3.790871253799255*^9, 3.790874173029567*^9, 3.790874280931679*^9, 
   3.790875061138954*^9, 3.790875827472735*^9, 3.7908765877922583`*^9, 
   3.790876665059642*^9, 3.790876705306933*^9, 3.7913817012790823`*^9, 
   3.791389643449603*^9, 3.791389725986205*^9, 3.7913897894049177`*^9, 
   3.791390672596541*^9, 3.794517948692202*^9, 3.7945207700393353`*^9},
 CellLabel->"Out[247]=",
 CellID->1890332435,ExpressionUUID->"247755fe-57de-4030-8cbe-971bbc145808"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PDsph", "[", 
   RowBox[{"-", "i"}], "]"}], "/@", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"r", "[", "]"}], ",", 
    RowBox[{"\[Theta]", "[", "]"}], ",", 
    RowBox[{"\[Phi]", "[", "]"}]}], "}"}]}]], "Input",
 CellChangeTimes->{{3.790870202280018*^9, 3.7908702030797377`*^9}},
 CellLabel->"In[248]:=",
 CellID->514969174,ExpressionUUID->"96aa78c9-de8a-4a1d-95fa-df83d7ea647b"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"i", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`i, {1, $CellContext`sph}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"i", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`i, {2, $CellContext`sph}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"i", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`i, {3, $CellContext`sph}],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.790870203562798*^9, 3.79087125436637*^9, 
  3.7908741730848713`*^9, 3.790874280994459*^9, 3.790875061649925*^9, 
  3.790875827552843*^9, 3.790876587846931*^9, 3.790876665109798*^9, 
  3.790876705381909*^9, 3.791381701330402*^9, 3.791389643502098*^9, 
  3.7913897260357428`*^9, 3.791389789460761*^9, 3.791390672654818*^9, 
  3.794517948735311*^9, 3.794520770096941*^9},
 CellLabel->"Out[248]=",
 CellID->944161327,ExpressionUUID->"3f280083-365f-41bf-9c25-6ae53074a0cd"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Therefore we have trivially that (again using any derivative):\
\>", "MathCaption",
 CellChangeTimes->{{3.790870112090465*^9, 3.790870127875556*^9}, {
  3.790870231720273*^9, 3.790870237775673*^9}},
 CellID->188737940,ExpressionUUID->"ac126853-2c53-4967-b59d-5399d807c1dc"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"{", 
    RowBox[{"1", ",", 
     RowBox[{"-", "sph"}]}], "}"}], "]"}], "/@", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"r", "[", "]"}], ",", 
    RowBox[{"\[Theta]", "[", "]"}], ",", 
    RowBox[{"\[Phi]", "[", "]"}]}], "}"}]}]], "Input",
 CellChangeTimes->{{3.790870129877727*^9, 3.7908701531546507`*^9}},
 CellLabel->"In[249]:=",
 CellID->852783625,ExpressionUUID->"46111333-18fd-46c9-8f2b-59858dd0b361"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"1", ",", "0", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.790870153573854*^9, 3.7908712555718517`*^9, 
  3.790874173132985*^9, 3.790874281047934*^9, 3.790875062965029*^9, 
  3.790875827605134*^9, 3.790876587889742*^9, 3.7908766651586657`*^9, 
  3.7908767054310837`*^9, 3.791381701372984*^9, 3.791389643550976*^9, 
  3.7913897260835743`*^9, 3.7913897895040083`*^9, 3.791390672705892*^9, 
  3.794517948777944*^9, 3.79452077014706*^9},
 CellLabel->"Out[249]=",
 CellID->762755206,ExpressionUUID->"c0842d25-0d03-4aaf-bf6c-0629001d210a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"{", 
    RowBox[{"2", ",", 
     RowBox[{"-", "sph"}]}], "}"}], "]"}], "/@", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"r", "[", "]"}], ",", 
    RowBox[{"\[Theta]", "[", "]"}], ",", 
    RowBox[{"\[Phi]", "[", "]"}]}], "}"}]}]], "Input",
 CellChangeTimes->{{3.79087015918497*^9, 3.790870159240861*^9}},
 CellLabel->"In[250]:=",
 CellID->1028307225,ExpressionUUID->"5dcc9711-26f5-4d62-aa0a-d7659426ba1c"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0", ",", "1", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.790870159504239*^9, 3.790871256057188*^9, 
  3.790874173213543*^9, 3.790874281100814*^9, 3.790875063412902*^9, 
  3.7908758276556673`*^9, 3.790876587935728*^9, 3.79087666522484*^9, 
  3.790876705491394*^9, 3.791381701415007*^9, 3.791389643598954*^9, 
  3.791389726129882*^9, 3.791389789546286*^9, 3.791390672757999*^9, 
  3.794517948818472*^9, 3.79452077019308*^9},
 CellLabel->"Out[250]=",
 CellID->1296858299,ExpressionUUID->"06f442b3-c06a-40c4-9a38-93626c3532ec"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"{", 
    RowBox[{"3", ",", 
     RowBox[{"-", "sph"}]}], "}"}], "]"}], "/@", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"r", "[", "]"}], ",", 
    RowBox[{"\[Theta]", "[", "]"}], ",", 
    RowBox[{"\[Phi]", "[", "]"}]}], "}"}]}]], "Input",
 CellChangeTimes->{{3.790870165457944*^9, 3.790870165577862*^9}},
 CellLabel->"In[251]:=",
 CellID->442214286,ExpressionUUID->"823ed00b-1ae5-49b8-84e5-bcee1a9c3533"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0", ",", "0", ",", "1"}], "}"}]], "Output",
 CellChangeTimes->{3.7908701658639936`*^9, 3.790871257074202*^9, 
  3.790874173268511*^9, 3.7908742811524973`*^9, 3.790875065959558*^9, 
  3.790875827709674*^9, 3.7908765880082703`*^9, 3.790876665293868*^9, 
  3.790876705545598*^9, 3.791381701452862*^9, 3.791389643675777*^9, 
  3.791389726178438*^9, 3.791389789592317*^9, 3.791390672824458*^9, 
  3.794517948863515*^9, 3.794520770240547*^9},
 CellLabel->"Out[251]=",
 CellID->1485897890,ExpressionUUID->"f603c589-3bac-4987-8aeb-b5eb561e27c3"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Derivatives of vectors do not all coincide, and this is where things get \
interesting. The PDsph derivative is the only one which gives zero on the \
basis vectors of the sph chart. Other derivatives, like the general PD \
derivative, give Christoffel tensors, with will be discussed further down:\
\>", "MathCaption",
 CellChangeTimes->{{3.79087044323223*^9, 3.7908705336456413`*^9}, {
  3.791379527384885*^9, 3.791379570647777*^9}},
 CellID->1437765897,ExpressionUUID->"38bc8d8d-379c-4e84-8517-abc0a1e2583d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"-", "i"}], "]"}], "[", 
  RowBox[{"Basis", "[", 
   RowBox[{
    RowBox[{"-", "j"}], ",", 
    RowBox[{"{", 
     RowBox[{"1", ",", "sph"}], "}"}]}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.7908705355046787`*^9, 3.790870549301176*^9}},
 CellLabel->"In[252]:=",
 CellID->814287199,ExpressionUUID->"2ab1f190-fac7-4295-9568-457ca94cf296"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"\[CapitalGamma]", "[", 
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[1., 0., 0.]], "]"}], 
      StyleBox[GridBox[{
         {
          StyleBox["1",
           FontColor->RGBColor[1., 0., 0.]], " ", " "},
         {" ", "i", "j"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`ChristoffelPDsph[{
   1, $CellContext`sph}, -$CellContext`i, -$CellContext`j],
  Editable->False]], "Output",
 CellChangeTimes->{3.7908705497854443`*^9, 3.79087125880101*^9, 
  3.7908741733216763`*^9, 3.7908742812043457`*^9, 3.790875067935498*^9, 
  3.79087582776375*^9, 3.790876588064928*^9, 3.790876665348075*^9, 
  3.7908767056064053`*^9, 3.7913795747841997`*^9, 3.791381701498637*^9, 
  3.7913896437257767`*^9, 3.791389726228217*^9, 3.791389789642528*^9, 
  3.791390672876459*^9, 3.794517948904546*^9, 3.794520770287665*^9},
 CellLabel->"Out[252]=",
 CellID->398506267,ExpressionUUID->"0301ad41-c026-46a3-838c-f11b0de30df9"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PDsph", "[", 
   RowBox[{"-", "i"}], "]"}], "[", 
  RowBox[{"Basis", "[", 
   RowBox[{
    RowBox[{"-", "j"}], ",", 
    RowBox[{"{", 
     RowBox[{"1", ",", "sph"}], "}"}]}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.790870561132412*^9, 3.79087056186069*^9}},
 CellLabel->"In[253]:=",
 CellID->1701931036,ExpressionUUID->"c1624c3a-868e-427a-a4fe-b7325888d413"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.7908705624848127`*^9, 3.7908712593294373`*^9, 
  3.79087417337352*^9, 3.790874281253138*^9, 3.790875068429811*^9, 
  3.7908758278237343`*^9, 3.79087658811205*^9, 3.790876665404338*^9, 
  3.790876705663768*^9, 3.791379576467486*^9, 3.791381701534153*^9, 
  3.791389643769074*^9, 3.791389726274293*^9, 3.791389789743087*^9, 
  3.7913906729258747`*^9, 3.794517948946154*^9, 3.794520770330069*^9},
 CellLabel->"Out[253]=",
 CellID->1840818123,ExpressionUUID->"413b145e-4344-46f8-aeae-4bc12fc268e3"]
}, Open  ]]
}, Open  ]],

Cell["\<\
If the result depends on the choice of derivative, which one is the important \
one to use? In general, it is the covariant derivative of the metric, which \
is the derivative with geometrical meaning (i.e. with meaning determined by \
the metric).\
\>", "Text",
 CellChangeTimes->{{3.790863109490128*^9, 3.790863258626314*^9}, {
  3.790866420158791*^9, 3.790866466799429*^9}, {3.790867225498068*^9, 
  3.79086723724268*^9}, {3.790868709189741*^9, 3.7908689326766863`*^9}, {
  3.790869830533938*^9, 3.79086991508281*^9}, {3.79087041083344*^9, 
  3.790870426871901*^9}, {3.790870648034322*^9, 3.7908707070637093`*^9}},
 CellID->1812968013,ExpressionUUID->"16c4582e-f8a8-473c-a89b-85d58ed21824"],

Cell[CellGroupData[{

Cell["\<\
The covariant derivative associated to the metric is the so-called \
Levi-Civita covariant derivative, computed with LC. The result is a CCovD \
object, which can be understood as playing the role for covariant derivatives \
that CTensor plays for tensors. It is a way of giving an *explicit* covariant \
derivative, as opposed to a general implicit covariant derivative defined \
with DefCovD:\
\>", "MathCaption",
 CellChangeTimes->{{3.79086980499693*^9, 3.790869809358234*^9}, {
  3.790869919226156*^9, 3.790869938577907*^9}, {3.7913796377172737`*^9, 
  3.791379768081592*^9}},
 CellID->777729093,ExpressionUUID->"691feddb-9532-4c95-9bab-5579f0af9930"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"cd", "=", 
  RowBox[{"LC", "[", "met", "]"}]}]], "Input",
 CellChangeTimes->{{3.790869715996189*^9, 3.790869724657208*^9}},
 CellLabel->"In[254]:=",
 CellID->1424946824,ExpressionUUID->"b7b7cdea-4576-4bba-83e3-6f9b38a1dd8d"],

Cell[BoxData[
 RowBox[{"CCovD", "[", 
  RowBox[{"PDsph", ",", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", 
           RowBox[{"-", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]}], ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           RowBox[{
            RowBox[{"-", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]}], " ", 
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["Sin",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Sin,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], "2"]}]}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", 
           FractionBox["1", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]], ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           FractionBox["1", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]], ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           RowBox[{
            RowBox[{"-", 
             RowBox[{
              InterpretationBox[
               StyleBox["Cos",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cos,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}]}], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["Sin",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Sin,
              Editable->False], "[", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False], "]"}]}]}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           FractionBox["1", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           RowBox[{
            InterpretationBox[
             StyleBox["Cot",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cot,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           FractionBox["1", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]], ",", 
           RowBox[{
            InterpretationBox[
             StyleBox["Cot",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cot,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], ",", "0"}], "}"}]}], "}"}]}], "}"}], 
     ",", 
     RowBox[{"{", 
      RowBox[{"sph", ",", 
       RowBox[{"-", "sph"}], ",", 
       RowBox[{"-", "sph"}]}], "}"}], ",", "0"}], "]"}], ",", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"1", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"], ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Sin",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Sin,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "2"]}]}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "sph"}], ",", 
       RowBox[{"-", "sph"}]}], "}"}], ",", "0"}], "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.790869726378018*^9, 3.790869799076977*^9, 
  3.790869940410087*^9, 3.790871260916308*^9, 3.790874173442821*^9, 
  3.790874281330914*^9, 3.790875069937804*^9, 3.7908758278871117`*^9, 
  3.790876588177883*^9, 3.7908766654663563`*^9, 3.790876705728362*^9, 
  3.791379771258552*^9, 3.791381701585209*^9, 3.791389643838696*^9, 
  3.79138972633972*^9, 3.791389789798292*^9, 3.7913906729999323`*^9, 
  3.794517949000298*^9, 3.794520770385419*^9},
 CellLabel->"Out[254]=",
 CellID->1980774579,ExpressionUUID->"22648db8-5422-4081-bb5d-ab98aab7219e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["It gives zero on its own metric tensor:", "MathCaption",
 CellChangeTimes->{{3.790869946104097*^9, 3.7908699517120447`*^9}, {
  3.79137978195266*^9, 3.791379785561068*^9}},
 CellID->21676925,ExpressionUUID->"afbe6405-521b-49df-8548-01a678f13a87"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"cd", "[", 
   RowBox[{"-", "i"}], "]"}], "[", 
  RowBox[{"met", "[", 
   RowBox[{
    RowBox[{"-", "j"}], ",", 
    RowBox[{"-", "k"}]}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.790869729324415*^9, 3.790869735401105*^9}},
 CellLabel->"In[255]:=",
 CellID->488081856,ExpressionUUID->"10765649-7433-42be-b578-3d63c7161d70"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.7908697359090652`*^9, 3.790871262197315*^9, 
  3.7908741734966087`*^9, 3.79087428138796*^9, 3.7908750712147923`*^9, 
  3.790875827935699*^9, 3.7908765882262573`*^9, 3.7908766655231237`*^9, 
  3.790876705788838*^9, 3.791379788443578*^9, 3.79138170162363*^9, 
  3.7913896438894873`*^9, 3.791389726387971*^9, 3.7913897898443947`*^9, 
  3.791390673052784*^9, 3.794517949042115*^9, 3.79452077042591*^9},
 CellLabel->"Out[255]=",
 CellID->586853459,ExpressionUUID->"c4bcb560-c8eb-4f6b-96a2-174e583c6009"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Partial derivatives of the metric are not zero:", "MathCaption",
 CellChangeTimes->{{3.7908707353904123`*^9, 3.790870742958466*^9}, {
   3.791379811391658*^9, 3.79137987355827*^9}, 3.791379932660872*^9},
 CellID->119153639,ExpressionUUID->"9abbdc40-8a76-42c5-a934-c1dbf595e99c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PDsph", "[", 
   RowBox[{"-", "i"}], "]"}], "[", 
  RowBox[{"met", "[", 
   RowBox[{
    RowBox[{"-", "j"}], ",", 
    RowBox[{"-", "k"}]}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.790870744418543*^9, 3.790870752326449*^9}},
 CellLabel->"In[256]:=",
 CellID->571661501,ExpressionUUID->"09349706-9595-4a24-ba21-be2f1a29dc3d"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 3} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 3} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 3} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 3} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}]}]],
              Annotation[#, {2, 2, 1} -> 2 $CellContext`r[], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 3} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 3} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 3} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 3} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False], " ", 
                SuperscriptBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Sin",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Sin,
                   Editable->False], "[", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False], "]"}], "2"]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                    "2"]}]}]],
              Annotation[#, {3, 3, 1} -> 2 $CellContext`r[] Sin[
                   $CellContext`\[Theta][]]^2, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{"2", " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 InterpretationBox[
                  StyleBox["\[Theta]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Theta][],
                  Editable->False], "]"}], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Sin",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Sin,
                  Editable->False], "[", 
                 InterpretationBox[
                  StyleBox["\[Theta]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Theta][],
                  Editable->False], "]"}]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Cos"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cos, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}]}]}]],
              Annotation[#, {3, 3, 2} -> 2 Cos[
                  $CellContext`\[Theta][]] $CellContext`r[]^2 Sin[
                  $CellContext`\[Theta][]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 3} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 4 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 4 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " ", " "},
         {"j", "k", "i"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{{0, 0, 0}, {0, 0, 0}, {0, 0, 
    0}}, {{0, 0, 0}, {2 $CellContext`r[], 0, 0}, {0, 0, 0}}, {{0, 0, 0}, {0, 
     0, 0}, {2 $CellContext`r[] Sin[
         $CellContext`\[Theta][]]^2, 2 Cos[
        $CellContext`\[Theta][]] $CellContext`r[]^2 Sin[
        $CellContext`\[Theta][]], 
      0}}}, {-$CellContext`sph, -$CellContext`sph, -$CellContext`sph}, 
   0][-$CellContext`j, -$CellContext`k, -$CellContext`i],
  Editable->False]], "Output",
 CellChangeTimes->{3.790870752722706*^9, 3.7908712638316307`*^9, 
  3.7908741735678453`*^9, 3.790874281453043*^9, 3.790875072728827*^9, 
  3.7908758279920063`*^9, 3.790876588284267*^9, 3.790876665587082*^9, 
  3.7908767058603697`*^9, 3.791379802884947*^9, 3.79137991644004*^9, 
  3.791381701671586*^9, 3.791389643942033*^9, 3.791389726444992*^9, 
  3.7913897899024878`*^9, 3.7913906731151447`*^9, 3.794517949088624*^9, 
  3.794520770477462*^9},
 CellLabel->"Out[256]=",
 CellID->2001178389,ExpressionUUID->"41143670-3862-422a-9cf9-6bfeaed19d13"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Again, the fact the three levels of this tensor are red means that the tensor \
components are given in the sph basis:\
\>", "MathCaption",
 CellChangeTimes->{{3.7908707353904123`*^9, 3.790870742958466*^9}, {
   3.791379811391658*^9, 3.79137987355827*^9}, 3.791379938307592*^9},
 CellID->1630978937,ExpressionUUID->"f37f562f-1f6f-4d88-beca-30f63168a663"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellChangeTimes->{{3.791379905743452*^9, 3.791379909349324*^9}},
 CellLabel->"In[257]:=",
 CellID->1724161398,ExpressionUUID->"c23ff10d-fccd-4ebc-9993-777a1f39338e"],

Cell["\<\
CTensor[{{{0, 0, 0}, {0, 0, 0}, {0, 0, 0}}, {{0, 0, 0}, {2*r[], 0, 0}, {0, 0, \
0}}, 
   {{0, 0, 0}, {0, 0, 0}, {2*r[]*Sin[\[Theta][]]^2, \
2*Cos[\[Theta][]]*r[]^2*Sin[\[Theta][]], 0}}}, {-sph, -sph, -sph}, 0][-j, -k, \
-i]\
\>", "Output",
 CellChangeTimes->{3.791379917120071*^9, 3.791381701708117*^9, 
  3.7913896439877787`*^9, 3.791389726496521*^9, 3.791389789952956*^9, 
  3.791390673160356*^9, 3.794517949129998*^9, 3.7945207705193377`*^9},
 CellLabel->"Out[257]//InputForm=",
 CellID->2063788256,ExpressionUUID->"754d3dfd-1d73-4fb4-bff6-57497e682cbb"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.4.3. xGrad", "Subsubsection",
 CellChangeTimes->{{3.728676331871439*^9, 3.728676332031807*^9}, 
   3.790876746494073*^9, {3.7908769396980658`*^9, 3.7908769402563057`*^9}, {
   3.790876988271463*^9, 3.7908770050309887`*^9}},
 CellID->381584480,ExpressionUUID->"2956bdc5-f871-4ae4-980e-ae38d5bb2ae8"],

Cell["\<\
Let us start with the first differential operator, the gradient.\
\>", "Text",
 CellChangeTimes->{{3.790863109490128*^9, 3.790863258626314*^9}, {
   3.790866420158791*^9, 3.790866466799429*^9}, {3.790867225498068*^9, 
   3.79086723724268*^9}, {3.790868709189741*^9, 3.7908689326766863`*^9}, {
   3.790869830533938*^9, 3.79086991508281*^9}, {3.79087041083344*^9, 
   3.790870426871901*^9}, {3.790870648034322*^9, 3.7908707070637093`*^9}, {
   3.7908727496977987`*^9, 3.7908727625849524`*^9}, {3.7913799872601547`*^9, 
   3.791380103390847*^9}, 3.7913801842837*^9},
 CellID->32869574,ExpressionUUID->"0f69733b-c7c9-4c4a-8d5b-504581296894"],

Cell[CellGroupData[{

Cell["\<\
So let us define the gradient of a scalar field. It is a vector field, so we \
need to provide the index that the final expression will have:\
\>", "MathCaption",
 CellChangeTimes->{{3.790870802708065*^9, 3.790870845371211*^9}, {
   3.790870953623087*^9, 3.790870955647373*^9}, 3.7913801932435427`*^9},
 CellID->1655986458,ExpressionUUID->"3aa9a903-e224-4f1e-8442-00857a6deab9"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"xGrad", "[", "scalar_", "]"}], "[", "i_", "]"}], ":=", 
  RowBox[{
   RowBox[{"cd", "[", "i", "]"}], "[", "scalar", "]"}]}]], "Input",
 CellChangeTimes->{{3.790870816350527*^9, 3.7908708597708797`*^9}, {
  3.7908709379425173`*^9, 3.790870991949873*^9}, {3.790871351058755*^9, 
  3.790871351593996*^9}},
 CellLabel->"In[258]:=",
 CellID->495600690,ExpressionUUID->"cf4b265b-8440-4deb-89d5-896fed8fdc6b"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Take a general scalar field. In xAct we need to declare all symbols, even \
functions of other things. This is not so much to know the names in advance, \
but to know their tensorial properties. In this case we are declaring that \
all arguments of f must be scalars and that the result of f should also be \
treated as a scalar:\
\>", "MathCaption",
 CellChangeTimes->{{3.790870878848419*^9, 3.790870893585466*^9}, {
  3.791380218858651*^9, 3.791380287993561*^9}},
 CellID->1451400784,ExpressionUUID->"de64321c-66df-4af6-bedc-3b65097e744a"],

Cell[BoxData[
 RowBox[{"DefScalarFunction", "[", "f", "]"}]], "Input",
 CellChangeTimes->{{3.79087089650746*^9, 3.790870899024523*^9}},
 CellLabel->"In[259]:=",
 CellID->146926218,ExpressionUUID->"b09cbd94-4d34-4736-8d2a-abd74a34681d"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Compute the gradient of our general scalar field:", "MathCaption",
 CellChangeTimes->{{3.7913803172155447`*^9, 3.791380329815634*^9}, {
  3.791380362815133*^9, 3.791380414668491*^9}},
 CellID->1907885024,ExpressionUUID->"f8ae5881-f6df-4fee-952d-682939bbbf35"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"grad", "=", 
  RowBox[{
   RowBox[{"xGrad", "[", 
    RowBox[{"f", "[", 
     RowBox[{
      RowBox[{"r", "[", "]"}], ",", 
      RowBox[{"\[Theta]", "[", "]"}], ",", 
      RowBox[{"\[Phi]", "[", "]"}]}], "]"}], "]"}], "[", "i", 
   "]"}]}]], "Input",
 CellChangeTimes->{{3.7908709051402493`*^9, 3.790870918568489*^9}, 
   3.7908709998936033`*^9, {3.790871356074007*^9, 3.790871356577907*^9}, {
   3.79087248959521*^9, 3.7908724911387053`*^9}},
 CellLabel->"In[260]:=",
 CellID->827776867,ExpressionUUID->"104563f0-119e-43e9-8d51-51fd6b7c8173"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["f",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`f,
               Editable->False], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"1", ",", "0", ",", "0"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["\[Phi]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Phi][],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", 
              RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"f"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], $CellContext`f, Editable -> False], 
                 TagBox[
                  RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                 MultilineFunction -> None], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`\[Theta][], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "\[Phi]", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`\[Phi][], Editable -> False]}], "]"}]}]],
           Annotation[#, {1} -> Derivative[1, 0, 0][$CellContext`f][
              $CellContext`r[], 
              $CellContext`\[Theta][], 
              $CellContext`\[Phi][]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["f",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`f,
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"0", ",", "1", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["\[Phi]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Phi][],
                Editable->False]}], "]"}], 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"f"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`f, Editable -> False], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                  MultilineFunction -> None], "[", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], ",", 
                   InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], ",", 
                   InterpretationBox[
                    StyleBox[
                    "\[Phi]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Phi][], Editable -> False]}], "]"}], 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`r[], Editable -> False], "2"]]}]],
           
           Annotation[#, {2} -> $CellContext`r[]^(-2) 
             Derivative[0, 1, 0][$CellContext`f][
               $CellContext`r[], 
               $CellContext`\[Theta][], 
               $CellContext`\[Phi][]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["Csc",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 Csc,
                 Editable->False], "[", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False], "]"}], "2"], " ", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["f",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`f,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"0", ",", "0", ",", "1"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["\[Phi]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Phi][],
                 Editable->False]}], "]"}]}], 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]],
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Csc"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Csc, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], "2"], 
                 " ", 
                 RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"f"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`f, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "\[Phi]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Phi][], Editable -> False]}], "]"}]}], 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`r[], Editable -> False], "2"]]}]],
           Annotation[#, {3} -> Csc[
                $CellContext`\[Theta][]]^2 $CellContext`r[]^(-2) 
             Derivative[0, 0, 1][$CellContext`f][
               $CellContext`r[], 
               $CellContext`\[Theta][], 
               $CellContext`\[Phi][]], "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"i"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{
    Derivative[1, 0, 0][$CellContext`f][
     $CellContext`r[], 
     $CellContext`\[Theta][], 
     $CellContext`\[Phi][]], $CellContext`r[]^(-2) 
    Derivative[0, 1, 0][$CellContext`f][
      $CellContext`r[], 
      $CellContext`\[Theta][], 
      $CellContext`\[Phi][]], Csc[
       $CellContext`\[Theta][]]^2 $CellContext`r[]^(-2) 
    Derivative[0, 0, 1][$CellContext`f][
      $CellContext`r[], 
      $CellContext`\[Theta][], 
      $CellContext`\[Phi][]]}, {$CellContext`sph}, 0][$CellContext`i],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.7908710003961067`*^9, 3.790871268489264*^9, {3.790871357268942*^9, 
   3.790871359914186*^9}, 3.790872491648898*^9, 3.7908741737221127`*^9, 
   3.790874281595845*^9, 3.7908750767391043`*^9, 3.790875828127178*^9, 
   3.790876588402272*^9, 3.7908766657240057`*^9, 3.790876705986526*^9, 
   3.7913805001285343`*^9, 3.791381719689392*^9, 3.791389644138133*^9, 
   3.79138972663722*^9, 3.791389790074669*^9, 3.791390673314467*^9, 
   3.794517949229849*^9, 3.794520777700791*^9},
 CellLabel->"Out[260]=",
 CellID->251697358,ExpressionUUID->"baaf87b4-502f-49b8-80c1-7d5d3facbda5"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
By default it is returned as a CTensor object, but we can easily translate it \
into a linear combination of basis vectors (of the same basis, i.e. color) of \
the CTensor itself:\
\>", "MathCaption",
 CellChangeTimes->{{3.7913804180124903`*^9, 3.7913804886986437`*^9}},
 CellID->1850360014,ExpressionUUID->"f9121996-fbba-4ca8-97ba-5cdfdeedd3ff"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToBasisExpand", "[", "grad", "]"}]], "Input",
 CellChangeTimes->{{3.79087136221903*^9, 3.790871367432551*^9}, {
  3.790872493938444*^9, 3.790872494737113*^9}},
 CellLabel->"In[261]:=",
 CellID->635491938,ExpressionUUID->"ed46ade5-da58-4f27-bf1b-d7bef1f01728"],

Cell[BoxData[
 RowBox[{
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "i"},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`sph}, $CellContext`i],
     Editable->False], " ", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Csc",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Csc,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"], " ", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["f",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`f,
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"0", ",", "0", ",", "1"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Phi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Phi][],
       Editable->False]}], "]"}]}], 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "i"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`sph}, $CellContext`i],
     Editable->False], " ", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["f",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`f,
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"0", ",", "1", ",", "0"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Phi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Phi][],
       Editable->False]}], "]"}]}], 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "i"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`sph}, $CellContext`i],
    Editable->False], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["f",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`f,
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"1", ",", "0", ",", "0"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Phi]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Phi][],
      Editable->False]}], "]"}]}]}]], "Output",
 CellChangeTimes->{
  3.790871367800509*^9, 3.7908724950090313`*^9, 3.790874173801556*^9, 
   3.790874281660061*^9, 3.790875078005682*^9, 3.7908758281870127`*^9, 
   3.790876588456644*^9, 3.790876665799911*^9, 3.790876706105959*^9, {
   3.7913804918325787`*^9, 3.791380505299494*^9}, 3.7913817197353277`*^9, 
   3.791389644188541*^9, 3.791389726710331*^9, 3.791389790127018*^9, 
   3.791390673365527*^9, 3.794517949281028*^9, 3.794520777750162*^9},
 CellLabel->"Out[261]=",
 CellID->1882194963,ExpressionUUID->"2e203cea-3ec4-4e24-b617-8f063e97cdfa"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Compare with WL's result. It's different because we are using a different \
basis:\
\>", "MathCaption",
 CellChangeTimes->{{3.790872498089777*^9, 3.790872518688676*^9}},
 CellID->1020070558,ExpressionUUID->"328a59c0-7439-4fb3-9d5b-9b13a1068042"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Grad", "[", 
  RowBox[{
   RowBox[{"f", "[", 
    RowBox[{"r", ",", "\[Theta]", ",", "\[Phi]"}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"r", ",", "\[Theta]", ",", "\[Phi]"}], "}"}], ",", 
   "\"\<Spherical\>\""}], "]"}]], "Input",
 CellChangeTimes->{{3.790871286653029*^9, 3.790871308067676*^9}},
 CellLabel->"In[262]:=",
 CellID->1254913691,ExpressionUUID->"dfb05cc6-4cd1-4773-9620-c26f9f930078"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["f",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`f,
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"1", ",", "0", ",", "0"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{"r", ",", "\[Theta]", ",", "\[Phi]"}], "]"}], ",", 
   FractionBox[
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["f",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`f,
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"0", ",", "1", ",", "0"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{"r", ",", "\[Theta]", ",", "\[Phi]"}], "]"}], "r"], ",", 
   FractionBox[
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["Csc",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Csc,
       Editable->False], "[", "\[Theta]", "]"}], " ", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["f",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`f,
        Editable->False], 
       TagBox[
        RowBox[{"(", 
         RowBox[{"0", ",", "0", ",", "1"}], ")"}],
        Derivative],
       MultilineFunction->None], "[", 
      RowBox[{"r", ",", "\[Theta]", ",", "\[Phi]"}], "]"}]}], "r"]}], 
  "}"}]], "Output",
 CellChangeTimes->{{3.790871301797434*^9, 3.790871308880621*^9}, 
   3.790874173890998*^9, 3.790874281763363*^9, 3.790875080383992*^9, 
   3.79087582827391*^9, 3.790876588533021*^9, 3.790876665883696*^9, 
   3.7908767061888*^9, 3.791380516236403*^9, 3.7913817198050737`*^9, 
   3.791389644271235*^9, 3.7913897267871857`*^9, 3.791389790203107*^9, 
   3.791390673446577*^9, 3.794517949359684*^9, 3.7945207778196898`*^9},
 CellLabel->"Out[262]=",
 CellID->1616622831,ExpressionUUID->"c630915a-ca80-471d-a56d-93f03fa849e0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Change the basis of our result. There are several possible ways to do it:\
\>", "MathCaption",
 CellChangeTimes->{{3.790872571639709*^9, 3.79087261052639*^9}, {
  3.790872714106262*^9, 3.7908727236010447`*^9}},
 CellID->2063232277,ExpressionUUID->"1a97b0d3-5783-43af-991a-c7fb12615af3"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToBasis", "[", "ortho", "]"}], "[", "grad", "]"}]], "Input",
 CellChangeTimes->{{3.790872523405451*^9, 3.790872547808359*^9}},
 CellLabel->"In[263]:=",
 CellID->1477722874,ExpressionUUID->"2b9795a6-739d-4cdb-9a86-122d65259d60"],

Cell[BoxData[
 RowBox[{
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["i",
              FontColor->RGBColor[0., 0., 1.]]},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{
      3, -$CellContext`ortho}, {$CellContext`i, $CellContext`ortho}],
     Editable->False], " ", 
    RowBox[{
     InterpretationBox[
      StyleBox["Csc",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Csc,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], " ", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["f",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`f,
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"0", ",", "0", ",", "1"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Phi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Phi][],
       Editable->False]}], "]"}]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["i",
              FontColor->RGBColor[0., 0., 1.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "Rows" -> {{
                Offset[0.]}}}],
          FontSize->9]}
       },
       GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{
      2, -$CellContext`ortho}, {$CellContext`i, $CellContext`ortho}],
     Editable->False], " ", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["f",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`f,
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"0", ",", "1", ",", "0"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Phi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Phi][],
       Editable->False]}], "]"}]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["i",
             FontColor->RGBColor[0., 0., 1.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{
     1, -$CellContext`ortho}, {$CellContext`i, $CellContext`ortho}],
    Editable->False], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["f",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`f,
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"1", ",", "0", ",", "0"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Phi]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Phi][],
      Editable->False]}], "]"}]}]}]], "Output",
 CellChangeTimes->{{3.790872531866478*^9, 3.790872548499736*^9}, 
   3.790874173986758*^9, 3.7908742818803663`*^9, 3.790875082449822*^9, 
   3.790875828370274*^9, 3.790876588620249*^9, 3.790876665972337*^9, 
   3.7908767062617617`*^9, 3.791380537730076*^9, 3.791381719868026*^9, 
   3.7913896443544493`*^9, 3.791389726884533*^9, 3.7913897902876377`*^9, 
   3.791390673534037*^9, 3.794517949426238*^9, 3.794520777875869*^9},
 CellLabel->"Out[263]=",
 CellID->616818773,ExpressionUUID->"dcf65ab4-3b87-4d67-a78d-73e687e67938"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCTensor", "[", 
  RowBox[{
   RowBox[{"Head", "[", "grad", "]"}], ",", 
   RowBox[{"{", "ortho", "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.7908726250705442`*^9, 3.7908726636854353`*^9}, {
  3.790872696396124*^9, 3.790872703027165*^9}, {3.7908727807691097`*^9, 
  3.79087278116057*^9}},
 CellLabel->"In[264]:=",
 CellID->89864978,ExpressionUUID->"5d102868-bf21-4b7a-a55b-d5dd76676e16"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["f",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`f,
        Editable->False], 
       TagBox[
        RowBox[{"(", 
         RowBox[{"1", ",", "0", ",", "0"}], ")"}],
        Derivative],
       MultilineFunction->None], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["\[Phi]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Phi][],
        Editable->False]}], "]"}], ",", 
     FractionBox[
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["f",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`f,
         Editable->False], 
        TagBox[
         RowBox[{"(", 
          RowBox[{"0", ",", "1", ",", "0"}], ")"}],
         Derivative],
        MultilineFunction->None], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Phi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Phi][],
         Editable->False]}], "]"}], 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]], ",", 
     FractionBox[
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox["Csc",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Csc,
         Editable->False], "[", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False], "]"}], " ", 
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["f",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`f,
          Editable->False], 
         TagBox[
          RowBox[{"(", 
           RowBox[{"0", ",", "0", ",", "1"}], ")"}],
          Derivative],
         MultilineFunction->None], "[", 
        RowBox[{
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["\[Phi]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Phi][],
          Editable->False]}], "]"}]}], 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]]}], "}"}], ",", 
   RowBox[{"{", "ortho", "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{
  3.790872664039483*^9, {3.79087269980494*^9, 3.790872703376974*^9}, 
   3.790872781574747*^9, 3.790874174035901*^9, 3.790874281929739*^9, 
   3.790875085878787*^9, 3.790875828413499*^9, 3.790876588662614*^9, 
   3.790876666046434*^9, 3.790876706304874*^9, 3.7913805541971197`*^9, 
   3.791381719909321*^9, 3.7913896444058247`*^9, 3.791389726927977*^9, 
   3.7913897903339033`*^9, 3.791390673587851*^9, 3.79451794945541*^9, 
   3.794520777916918*^9},
 CellLabel->"Out[264]=",
 CellID->969553742,ExpressionUUID->"f912ba0b-7ed4-413d-897c-7c2998fac8da"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "[", "i", "]"}]], "Input",
 CellChangeTimes->{{3.79087278548315*^9, 3.790872786728426*^9}},
 CellLabel->"In[265]:=",
 CellID->19739716,ExpressionUUID->"f9588d62-e91e-4584-ad61-75c239502537"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["f",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`f,
               Editable->False], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"1", ",", "0", ",", "0"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["\[Phi]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Phi][],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", 
              RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"f"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], $CellContext`f, Editable -> False], 
                 TagBox[
                  RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                 MultilineFunction -> None], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`\[Theta][], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "\[Phi]", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`\[Phi][], Editable -> False]}], "]"}]}]],
           Annotation[#, {1} -> Derivative[1, 0, 0][$CellContext`f][
              $CellContext`r[], 
              $CellContext`\[Theta][], 
              $CellContext`\[Phi][]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["f",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`f,
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"0", ",", "1", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["\[Phi]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Phi][],
                Editable->False]}], "]"}], 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"f"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`f, Editable -> False], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                  MultilineFunction -> None], "[", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], ",", 
                   InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], ",", 
                   InterpretationBox[
                    StyleBox[
                    "\[Phi]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Phi][], Editable -> False]}], "]"}], 
               InterpretationBox[
                StyleBox[
                "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                $CellContext`r[], Editable -> False]]}]],
           
           Annotation[#, {2} -> $CellContext`r[]^(-1) 
             Derivative[0, 1, 0][$CellContext`f][
               $CellContext`r[], 
               $CellContext`\[Theta][], 
               $CellContext`\[Phi][]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["Csc",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Csc,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], " ", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["f",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`f,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"0", ",", "0", ",", "1"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["\[Phi]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Phi][],
                 Editable->False]}], "]"}]}], 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]],
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Csc"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Csc, Editable -> False], "[", 
                   InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], " ", 
                 RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"f"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`f, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "\[Phi]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Phi][], Editable -> False]}], "]"}]}], 
               InterpretationBox[
                StyleBox[
                "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                $CellContext`r[], Editable -> False]]}]],
           Annotation[#, {3} -> Csc[
               $CellContext`\[Theta][]] $CellContext`r[]^(-1) 
             Derivative[0, 0, 1][$CellContext`f][
               $CellContext`r[], 
               $CellContext`\[Theta][], 
               $CellContext`\[Phi][]], "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[0, 0, 1], 2 -> RGBColor[0, 0, 1]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"i"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{
    Derivative[1, 0, 0][$CellContext`f][
     $CellContext`r[], 
     $CellContext`\[Theta][], 
     $CellContext`\[Phi][]], $CellContext`r[]^(-1) 
    Derivative[0, 1, 0][$CellContext`f][
      $CellContext`r[], 
      $CellContext`\[Theta][], 
      $CellContext`\[Phi][]], Csc[
      $CellContext`\[Theta][]] $CellContext`r[]^(-1) 
    Derivative[0, 0, 1][$CellContext`f][
      $CellContext`r[], 
      $CellContext`\[Theta][], 
      $CellContext`\[Phi][]]}, {$CellContext`ortho}, 0][$CellContext`i],
  Editable->False]], "Output",
 CellChangeTimes->{3.790872787062243*^9, 3.790874174096238*^9, 
  3.7908742819854603`*^9, 3.790875086487911*^9, 3.7908758284705153`*^9, 
  3.790876588713725*^9, 3.790876666130577*^9, 3.790876706360422*^9, 
  3.7913805576307077`*^9, 3.791381719951342*^9, 3.791389644466958*^9, 
  3.791389726972321*^9, 3.791389790381888*^9, 3.791390673643681*^9, 
  3.794517949504299*^9, 3.794520777967064*^9},
 CellLabel->"Out[265]=",
 CellID->183828358,ExpressionUUID->"046e5e4b-1244-4537-9a94-529b971a1a05"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.4.4. xDiv", "Subsubsection",
 CellChangeTimes->{{3.728676331871439*^9, 3.728676332031807*^9}, 
   3.790876746494073*^9, {3.7908769298745613`*^9, 3.790876930314081*^9}, {
   3.790876991383491*^9, 3.790877007445849*^9}},
 CellID->672074522,ExpressionUUID->"71d13d73-61a9-4519-b580-6c053d13b01a"],

Cell["\<\
Now we introduce the divergence of a vector, which is a scalar.\
\>", "Text",
 CellChangeTimes->{{3.790863109490128*^9, 3.790863258626314*^9}, {
  3.790866420158791*^9, 3.790866466799429*^9}, {3.790867225498068*^9, 
  3.79086723724268*^9}, {3.790868709189741*^9, 3.7908689326766863`*^9}, {
  3.790869830533938*^9, 3.79086991508281*^9}, {3.79087041083344*^9, 
  3.790870426871901*^9}, {3.790870648034322*^9, 3.7908707070637093`*^9}, {
  3.7908727496977987`*^9, 3.7908727625849524`*^9}, {3.7908727985522947`*^9, 
  3.790872812135479*^9}},
 CellID->281396522,ExpressionUUID->"f1d0a315-2f3a-4d52-a102-8126fc6d3f8c"],

Cell[CellGroupData[{

Cell["\<\
We provide the input as a vector field, and with a second argument we specify \
the index in which this object is a vector field. We will use that \
information to know where to differentiate:\
\>", "MathCaption",
 CellChangeTimes->{{3.790872884923176*^9, 3.790872931115179*^9}, {
  3.791380605718253*^9, 3.791380677612712*^9}, {3.791381756409584*^9, 
  3.791381814288299*^9}},
 CellID->756365217,ExpressionUUID->"c210bfeb-ade5-4273-bcef-735fbcb06102"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"xDiv", "[", 
    RowBox[{"v_", ",", "i_"}], "]"}], "[", "]"}], ":=", 
  RowBox[{
   RowBox[{"cd", "[", 
    RowBox[{"-", "i"}], "]"}], "[", "v", "]"}]}]], "Input",
 CellChangeTimes->{{3.790872818041565*^9, 3.790872880037217*^9}, {
  3.791381737874783*^9, 3.791381742043044*^9}, {3.7913817908083067`*^9, 
  3.791381793544559*^9}, {3.791381859615088*^9, 3.791381860015039*^9}},
 CellLabel->"In[266]:=",
 CellID->31867511,ExpressionUUID->"06b848d7-d3b8-4cfb-9e4f-6e813b1f712d"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Take our contravariant vector v and compute its divergence:", \
"MathCaption",
 CellChangeTimes->{{3.790872948459085*^9, 3.79087295164926*^9}, {
  3.791380689963608*^9, 3.7913807051716433`*^9}},
 CellID->1407314168,ExpressionUUID->"7e12d5a7-d34a-4a61-9928-9cc6cdd05759"],

Cell[CellGroupData[{

Cell[BoxData["v"], "Input",
 CellChangeTimes->{3.790872952668605*^9},
 CellLabel->"In[267]:=",
 CellID->243721534,ExpressionUUID->"c77794e0-8b97-49a8-b702-8ab080a4ccf0"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], ",", 
     RowBox[{
      RowBox[{
       InterpretationBox[
        StyleBox["Cos",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Cos,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}], "}"}], ",", 
   RowBox[{"{", "sph", "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.790872952991405*^9, 3.790873055205535*^9, 
  3.790874174167225*^9, 3.790874282053442*^9, 3.790875088781734*^9, 
  3.790875828588526*^9, 3.79087658877777*^9, 3.7908766662519007`*^9, 
  3.790876706419941*^9, 3.791380695200058*^9, 3.791381830294106*^9, 
  3.791381862115183*^9, 3.7913896445453463`*^9, 3.79138972703736*^9, 
  3.791389790452856*^9, 3.791390673719253*^9, 3.7945179495747337`*^9, 
  3.794520778022911*^9},
 CellLabel->"Out[267]=",
 CellID->1521728520,ExpressionUUID->"5e48e28f-1f46-4fc0-b08f-67b432e99f0e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"xDiv", "[", 
    RowBox[{
     RowBox[{"v", "[", "i", "]"}], ",", "i"}], "]"}], "[", "]"}], "//", 
  "Simplify"}]], "Input",
 CellChangeTimes->{{3.79087295714156*^9, 3.7908729608103647`*^9}, {
  3.79087305678308*^9, 3.79087305794307*^9}, {3.791381832927849*^9, 
  3.79138183399174*^9}},
 CellLabel->"In[268]:=",
 CellID->788203964,ExpressionUUID->"bb207e93-417f-47f6-8e64-7952805e2621"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{"4", "+", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["Cos",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Cos,
       Editable->False], "[", 
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False]}], "]"}], " ", 
     RowBox[{
      InterpretationBox[
       StyleBox["Csc",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Csc,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}]}]}], ")"}], " ", 
  InterpretationBox[
   StyleBox["r",
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`r[],
   Editable->False]}]], "Output",
 CellChangeTimes->{
  3.7908729614169493`*^9, 3.790873058342018*^9, 3.790874174258617*^9, 
   3.790874282124325*^9, 3.790875089290646*^9, 3.790875828653882*^9, 
   3.790876588833292*^9, 3.790876666322042*^9, 3.790876706485619*^9, 
   3.791380707189159*^9, {3.791381834793702*^9, 3.791381862720999*^9}, 
   3.791389644616765*^9, 3.791389727106812*^9, 3.79138979052761*^9, 
   3.7913906738045073`*^9, 3.7945179497016478`*^9, 3.794520778079114*^9},
 CellLabel->"Out[268]=",
 CellID->1308483517,ExpressionUUID->"77f6ea19-2061-4c86-b712-ca251bffe196"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
If we blindly try Div on the vector components we get a wrong result:\
\>", "MathCaption",
 CellChangeTimes->{{3.790872985623899*^9, 3.7908730039847717`*^9}},
 CellID->254136468,ExpressionUUID->"62eb2fe0-10db-4c69-b3a4-625b851148bc"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"First", "[", "v", "]"}]], "Input",
 CellChangeTimes->{{3.790873007672503*^9, 3.790873016608554*^9}},
 CellLabel->"In[269]:=",
 CellID->781127457,ExpressionUUID->"b7df79d4-3b7b-428e-8701-d9e9318dedd4"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox["Cos",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Cos,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], " ", 
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False]}], ",", 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{{3.790873012371587*^9, 3.7908730169266863`*^9}, {
   3.790873048340003*^9, 3.790873060643556*^9}, 3.790874174308228*^9, 
   3.790874282176819*^9, 3.7908750911168528`*^9, 3.790875828705125*^9, 
   3.790876588872017*^9, 3.7908766663745937`*^9, 3.7908767065265703`*^9, 
   3.791380719039098*^9, 3.79138187569027*^9, 3.7913896446902523`*^9, 
   3.7913897271510897`*^9, 3.79138979057465*^9, 3.791390673884508*^9, 
   3.79451794974571*^9, 3.794520778120406*^9},
 CellLabel->"Out[269]=",
 CellID->1449924939,ExpressionUUID->"ff9be134-f1e8-491c-a23b-3289bb85740b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Div", "[", 
   RowBox[{"%", ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"r", "[", "]"}], ",", 
      RowBox[{"\[Theta]", "[", "]"}], ",", 
      RowBox[{"\[Phi]", "[", "]"}]}], "}"}], ",", "\"\<Spherical\>\""}], 
   "]"}], "//", "Simplify"}]], "Input",
 CellChangeTimes->{{3.790873018955257*^9, 3.7908730631428556`*^9}},
 CellLabel->"In[270]:=",
 CellID->1238589112,ExpressionUUID->"064403b3-aa61-42d1-967b-77903d7923bc"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox["Cos",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cos,
     Editable->False], "[", 
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Csc",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Csc,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}]}], "+", 
  RowBox[{"4", " ", 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{{3.790873045506117*^9, 3.790873063584127*^9}, 
   3.790874174392727*^9, 3.790874282248171*^9, 3.790875091665347*^9, 
   3.790875828778037*^9, 3.790876588945066*^9, 3.790876666457883*^9, 
   3.790876706606179*^9, 3.791380723021941*^9, 3.7913818779569197`*^9, 
   3.791389644754417*^9, 3.7913897272170467`*^9, 3.791389790644026*^9, 
   3.791390673952017*^9, 3.794517949814039*^9, 3.7945207781835537`*^9},
 CellLabel->"Out[270]=",
 CellID->1391507264,ExpressionUUID->"deeb048f-2f91-48fa-8b2d-4c3a007a37bd"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The problem again is that WL's Div expects vector components in the \
orthonormal basis:\
\>", "MathCaption",
 CellChangeTimes->{{3.790873073901173*^9, 3.790873101156569*^9}},
 CellID->45102520,ExpressionUUID->"8ee10e57-f5b8-4ebe-90d0-675ead531f57"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCTensor", "[", 
  RowBox[{"v", ",", 
   RowBox[{"{", "ortho", "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.790873103776528*^9, 3.790873110949892*^9}},
 CellLabel->"In[271]:=",
 CellID->1102713718,ExpressionUUID->"edbc718e-4806-43c3-8c1b-41d38c791bfb"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], ",", 
     RowBox[{
      RowBox[{
       InterpretationBox[
        StyleBox["Cos",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Cos,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"]}], ",", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}]}]}], "}"}], ",", 
   RowBox[{"{", "ortho", "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.790873111291787*^9, 3.790874174448072*^9, 
  3.790874282299358*^9, 3.790875092989665*^9, 3.790875828822577*^9, 
  3.790876588988892*^9, 3.790876666506195*^9, 3.790876706654599*^9, 
  3.791380741723723*^9, 3.791381879516906*^9, 3.791389644797545*^9, 
  3.791389727263668*^9, 3.791389790690278*^9, 3.7913906740019827`*^9, 
  3.7945179498578377`*^9, 3.7945207782225723`*^9},
 CellLabel->"Out[271]=",
 CellID->965670218,ExpressionUUID->"aafedd50-7530-423c-84c2-546bdb9a426c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"First", "[", "%", "]"}]], "Input",
 CellChangeTimes->{{3.790873114959854*^9, 3.790873116805706*^9}},
 CellLabel->"In[272]:=",
 CellID->1028117429,ExpressionUUID->"2880ade8-f584-4f9a-90dc-e28e834b9227"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox["Cos",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Cos,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}], ",", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"], " ", 
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}]}]}], "}"}]], "Output",
 CellChangeTimes->{3.790873117497471*^9, 3.790874174503203*^9, 
  3.790874282365312*^9, 3.7908750934617043`*^9, 3.790875828873885*^9, 
  3.79087658906297*^9, 3.790876666564*^9, 3.7908767067006073`*^9, 
  3.7913807450787497`*^9, 3.791381880458764*^9, 3.791389644840887*^9, 
  3.791389727312318*^9, 3.791389790761036*^9, 3.791390674051629*^9, 
  3.794517949894265*^9, 3.794520778264778*^9},
 CellLabel->"Out[272]=",
 CellID->1485433547,ExpressionUUID->"af5f4593-78cc-4bee-bfb3-e83fabae5e6e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Div", "[", 
   RowBox[{"%", ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"r", "[", "]"}], ",", 
      RowBox[{"\[Theta]", "[", "]"}], ",", 
      RowBox[{"\[Phi]", "[", "]"}]}], "}"}], ",", "\"\<Spherical\>\""}], 
   "]"}], "//", "Simplify"}]], "Input",
 CellChangeTimes->{{3.7908731199516277`*^9, 3.790873131755287*^9}},
 CellLabel->"In[273]:=",
 CellID->1658101237,ExpressionUUID->"b5dd81f0-d39a-41c4-b833-18d1a167b4e8"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{"4", "+", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["Cos",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Cos,
       Editable->False], "[", 
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False]}], "]"}], " ", 
     RowBox[{
      InterpretationBox[
       StyleBox["Csc",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Csc,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}]}]}], ")"}], " ", 
  InterpretationBox[
   StyleBox["r",
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`r[],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.790873132444065*^9, 3.7908741745718946`*^9, 
  3.790874282442656*^9, 3.790875094860826*^9, 3.790875828949656*^9, 
  3.790876589129149*^9, 3.7908766666396713`*^9, 3.790876706769434*^9, 
  3.791380748449819*^9, 3.791381881084179*^9, 3.7913896449017487`*^9, 
  3.7913897273736153`*^9, 3.791389790827773*^9, 3.791390674115822*^9, 
  3.794517949947982*^9, 3.794520778319171*^9},
 CellLabel->"Out[273]=",
 CellID->1184228054,ExpressionUUID->"937b3f4e-1316-4787-bda3-28a4627e551f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Note that our notations contain the information of which basis is being used, \
and the system can handle all of them:\
\>", "MathCaption",
 CellChangeTimes->{{3.790873149195983*^9, 3.790873188074432*^9}},
 CellID->932149756,ExpressionUUID->"f420ea82-6a59-4953-9a02-81e11707d109"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCTensor", "[", 
  RowBox[{"v", ",", 
   RowBox[{"{", "ortho", "}"}]}], "]"}]], "Input",
 CellLabel->"In[274]:=",
 CellID->1684957388,ExpressionUUID->"901d0ffe-3c35-4380-b5e9-0227eb4cd761"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], ",", 
     RowBox[{
      RowBox[{
       InterpretationBox[
        StyleBox["Cos",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Cos,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"]}], ",", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}]}]}], "}"}], ",", 
   RowBox[{"{", "ortho", "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{{3.790873195159502*^9, 3.790873218951232*^9}, 
   3.790874174624702*^9, 3.7908742824966288`*^9, 3.790875096381819*^9, 
   3.790875829001535*^9, 3.790876589166114*^9, 3.7908766667003183`*^9, 
   3.790876706817623*^9, 3.791380758352806*^9, 3.791381884754155*^9, 
   3.791389644942305*^9, 3.791389727417461*^9, 3.791389790873932*^9, 
   3.7913906741647*^9, 3.794517949985853*^9, 3.794520778359961*^9},
 CellLabel->"Out[274]=",
 CellID->442463182,ExpressionUUID->"41c4acaa-b00b-4d52-95d9-e61b6db98dc6"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"xDiv", "[", 
     RowBox[{
      RowBox[{"%", "[", "i", "]"}], ",", "i"}], "]"}], "[", "]"}], "//", 
   "ToValues"}], "//", "Simplify"}]], "Input",
 CellChangeTimes->{{3.7908731967568607`*^9, 3.790873223497355*^9}, {
  3.79138188887002*^9, 3.791381893117593*^9}},
 CellLabel->"In[275]:=",
 CellID->1636349316,ExpressionUUID->"c7b873a3-8897-492e-9ab8-c4347073b017"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{"4", "+", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["Cos",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Cos,
       Editable->False], "[", 
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False]}], "]"}], " ", 
     RowBox[{
      InterpretationBox[
       StyleBox["Csc",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Csc,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}]}]}], ")"}], " ", 
  InterpretationBox[
   StyleBox["r",
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`r[],
   Editable->False]}]], "Output",
 CellChangeTimes->{{3.7908732037818737`*^9, 3.790873223882782*^9}, 
   3.7908741748758993`*^9, 3.790874282736944*^9, 3.7908750981585608`*^9, 
   3.790875829265834*^9, 3.790876589408643*^9, 3.7908766669490223`*^9, 
   3.790876707083037*^9, 3.79138076225126*^9, 3.791381894471303*^9, 
   3.791389645188518*^9, 3.791389727656536*^9, 3.7913897911205597`*^9, 
   3.791390674406698*^9, 3.7945179503308983`*^9, 3.794520778596122*^9},
 CellLabel->"Out[275]=",
 CellID->1890313960,ExpressionUUID->"d72a018c-6bc2-4764-880c-56a127e565c5"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We can even start with the covariant components of the tensor, and our xDiv \
will handle that too:\
\>", "MathCaption",
 CellChangeTimes->{{3.790873149195983*^9, 3.790873188074432*^9}, {
  3.791380784615819*^9, 3.7913808325512257`*^9}},
 CellID->976614439,ExpressionUUID->"a0784977-f726-4d10-8cce-8b6e52d48f0a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"v", "[", 
  RowBox[{"-", "i"}], "]"}]], "Input",
 CellChangeTimes->{{3.791380836160263*^9, 3.791380862647675*^9}, {
  3.7913819107398157`*^9, 3.791381912324749*^9}},
 CellLabel->"In[276]:=",
 CellID->1763483765,ExpressionUUID->"8d9675b8-5240-4269-97bd-329640ca2c56"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"],
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox[
                "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                $CellContext`r[], Editable -> False], "2"]}]],
           Annotation[#, {1} -> $CellContext`r[]^2, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox["Cos",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cos,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "3"]}],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"Cos"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cos, Editable -> False], "[", 
                  InterpretationBox[
                   StyleBox[
                   "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`\[Theta][], Editable -> False], "]"}], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                  "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`r[], Editable -> False], "3"]}]}]],
           Annotation[#, {2} -> Cos[
               $CellContext`\[Theta][]] $CellContext`r[]^3, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "3"], " ", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], "2"]}],
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", 
              RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                  "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`r[], Editable -> False], "3"], " ", 
                SuperscriptBox[
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                   InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                 "2"]}]}]],
           Annotation[#, {3} -> $CellContext`r[]^3 Sin[
                $CellContext`\[Theta][]]^2, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" "},
         {"i"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{$CellContext`r[]^2, Cos[
      $CellContext`\[Theta][]] $CellContext`r[]^3, $CellContext`r[]^3 Sin[
       $CellContext`\[Theta][]]^2}, {-$CellContext`sph}, 0][-$CellContext`i],
  Editable->False]], "Output",
 CellChangeTimes->{{3.791380836915844*^9, 3.79138089690539*^9}, 
   3.7913819127248917`*^9, 3.7913896452389507`*^9, 3.791389727705305*^9, 
   3.7913897911712914`*^9, 3.791390674458199*^9, 3.7945179503785057`*^9, 
   3.7945207786393023`*^9},
 CellLabel->"Out[276]=",
 CellID->633652002,ExpressionUUID->"c4b98d51-2a95-4e37-ba81-2e6e054dc534"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"xDiv", "[", 
    RowBox[{"%", ",", 
     RowBox[{"-", "i"}]}], "]"}], "[", "]"}], "//", "Simplify"}]], "Input",
 CellChangeTimes->{{3.791380865070588*^9, 3.791380879678364*^9}, {
  3.791381916069277*^9, 3.7913819173411007`*^9}},
 CellLabel->"In[277]:=",
 CellID->1364973140,ExpressionUUID->"7b5e6ba2-3e67-442a-b6d5-1164c0302750"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{"4", "+", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["Cos",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Cos,
       Editable->False], "[", 
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False]}], "]"}], " ", 
     RowBox[{
      InterpretationBox[
       StyleBox["Csc",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Csc,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}]}]}], ")"}], " ", 
  InterpretationBox[
   StyleBox["r",
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`r[],
   Editable->False]}]], "Output",
 CellChangeTimes->{{3.791380869819973*^9, 3.7913808975845957`*^9}, 
   3.791381918747724*^9, 3.791389645317328*^9, 3.791389727785117*^9, 
   3.791389791248803*^9, 3.791390674542272*^9, 3.794517950453847*^9, 
   3.794520778763309*^9},
 CellLabel->"Out[277]=",
 CellID->417792141,ExpressionUUID->"7e42e7b4-ca62-4f0a-8ae1-f01b290f609e"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.4.5. xCurl", "Subsubsection",
 CellChangeTimes->{{3.728676331871439*^9, 3.728676332031807*^9}, 
   3.790876746494073*^9, {3.790876916730966*^9, 3.790876918474334*^9}, {
   3.790876994119276*^9, 3.790877010006895*^9}},
 CellID->1060734641,ExpressionUUID->"d62a8fa2-188c-430c-871b-1f08c826b625"],

Cell["\<\
Now we show how to construct the Curl of a vector field. To do this, we first \
need to discuss another interesting object: the epsilon tensor associated \
with a metric:\
\>", "Text",
 CellChangeTimes->{{3.790863109490128*^9, 3.790863258626314*^9}, {
  3.790866420158791*^9, 3.790866466799429*^9}, {3.790867225498068*^9, 
  3.79086723724268*^9}, {3.790868709189741*^9, 3.7908689326766863`*^9}, {
  3.790869830533938*^9, 3.79086991508281*^9}, {3.79087041083344*^9, 
  3.790870426871901*^9}, {3.790870648034322*^9, 3.7908707070637093`*^9}, {
  3.7908727496977987`*^9, 3.7908727625849524`*^9}, {3.7908732630164347`*^9, 
  3.79087327540744*^9}, {3.790873390660242*^9, 3.790873414041915*^9}, {
  3.791380917388586*^9, 3.791380943579883*^9}},
 CellID->558804793,ExpressionUUID->"e20fb4d6-4d42-4ff4-af8f-715fed87e939"],

Cell[CellGroupData[{

Cell["\<\
The epsilon tensor is fully antisymmetric. Again it will have different forms \
in different bases, being particularly simple in the orthonormal basis:\
\>", "MathCaption",
 CellChangeTimes->{{3.790875378526244*^9, 3.790875449659071*^9}, {
  3.791380960994999*^9, 3.791380976683111*^9}},
 CellID->289463677,ExpressionUUID->"e78b9c5a-df05-41c3-9f71-7866b8b6ea31"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"epsilonmet", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"epsilon", "[", "met", "]"}], "//", "Normal"}], "//", 
   "PowerExpand"}]}]], "Input",
 CellChangeTimes->{{3.79087534307516*^9, 3.790875344831394*^9}, {
  3.790875385700342*^9, 3.790875429987072*^9}},
 CellLabel->"In[278]:=",
 CellID->229469696,ExpressionUUID->"46d584f0-0fca-47ff-b3ca-dc1a04344119"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"], " ", 
          RowBox[{
           InterpretationBox[
            StyleBox["Sin",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Sin,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}]}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", 
         RowBox[{
          RowBox[{"-", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}], " ", 
          RowBox[{
           InterpretationBox[
            StyleBox["Sin",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Sin,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}]}], ",", "0"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", 
         RowBox[{
          RowBox[{"-", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}], " ", 
          RowBox[{
           InterpretationBox[
            StyleBox["Sin",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Sin,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}]}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"], " ", 
          RowBox[{
           InterpretationBox[
            StyleBox["Sin",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Sin,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}]}], ",", "0", ",", "0"}], "}"}]}], "}"}], 
     ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"], " ", 
          RowBox[{
           InterpretationBox[
            StyleBox["Sin",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Sin,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}]}], ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{
          RowBox[{"-", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}], " ", 
          RowBox[{
           InterpretationBox[
            StyleBox["Sin",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Sin,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}]}], ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0"}], "}"}]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "sph"}], ",", 
     RowBox[{"-", "sph"}], ",", 
     RowBox[{"-", "sph"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.7908753452521563`*^9, 3.790875394342634*^9, 
  3.790875430333026*^9, 3.790875829322234*^9, 3.7908765894704647`*^9, 
  3.790876667001831*^9, 3.790876707156522*^9, 3.791380978886862*^9, 
  3.791381946149431*^9, 3.791389645367958*^9, 3.791389727832247*^9, 
  3.791389791300206*^9, 3.791390674591552*^9, 3.794517950493174*^9, 
  3.79452077880399*^9},
 CellLabel->"Out[278]=",
 CellID->1708471744,ExpressionUUID->"33d522ac-199d-4905-8b12-26e0df45df7b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCTensor", "[", 
  RowBox[{"epsilonmet", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "ortho"}], ",", 
     RowBox[{"-", "ortho"}], ",", 
     RowBox[{"-", "ortho"}]}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.790875451829699*^9, 3.7908754645081167`*^9}},
 CellLabel->"In[279]:=",
 CellID->500436437,ExpressionUUID->"49300fc6-f3ac-48c0-bcd2-863007062ccb"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "1"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", 
         RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", 
         RowBox[{"-", "1"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"1", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "1", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "1"}], ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0"}], "}"}]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "ortho"}], ",", 
     RowBox[{"-", "ortho"}], ",", 
     RowBox[{"-", "ortho"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.790875464953988*^9, 3.79087582937891*^9, 
  3.790876589517549*^9, 3.790876667057032*^9, 3.790876707206397*^9, 
  3.791380985808312*^9, 3.791381947426566*^9, 3.791389645415249*^9, 
  3.791389727896512*^9, 3.791389791347431*^9, 3.7913906746467752`*^9, 
  3.7945179505341673`*^9, 3.794520778845866*^9},
 CellLabel->"Out[279]=",
 CellID->280757371,ExpressionUUID->"f487881d-1261-472f-b8d6-9671a6713533"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Being associated to the metric, it also has zero covariant derivative with \
respect to the Levi-Civita connection:\
\>", "MathCaption",
 CellChangeTimes->{{3.791390698251346*^9, 3.791390736594597*^9}},
 CellID->1431860481,ExpressionUUID->"b53f5d01-5a32-4151-ab56-33ce1fc850fb"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"cd", "[", 
   RowBox[{"-", "i"}], "]"}], "[", 
  RowBox[{"epsilonmet", "[", 
   RowBox[{
    RowBox[{"-", "j"}], ",", 
    RowBox[{"-", "k"}], ",", 
    RowBox[{"-", "l"}]}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.79139068318122*^9, 3.791390694212048*^9}},
 CellLabel->"In[280]:=",
 CellID->156844470,ExpressionUUID->"5f8c272b-757d-4135-925f-8cc71865db3d"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.791390695056802*^9, 3.7945179505801497`*^9, 
  3.7945207788879967`*^9},
 CellLabel->"Out[280]=",
 CellID->667561938,ExpressionUUID->"39250d6d-4bd5-4d0f-8f44-5f4d45c2f80e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
This is the definition of Curl. Again, we need to provide the index in which \
the object is a vector. Do not use the same index for i and j in this \
definition. There are ways of avoiding this type of collision in xAct (for \
more robust code), but we will not discuss that in this intro section:\
\>", "MathCaption",
 CellChangeTimes->{{3.7908733781701937`*^9, 3.790873388235153*^9}, {
  3.79087566452384*^9, 3.7908756735479317`*^9}, {3.791381949867289*^9, 
  3.791382024591825*^9}, {3.791390930668908*^9, 3.791390943923842*^9}, {
  3.791391048113316*^9, 3.791391051016368*^9}},
 CellID->1790938121,ExpressionUUID->"dc238619-6cab-49da-8d0b-94919658b8cf"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"xCurl", "[", 
    RowBox[{"v_", ",", "j_"}], "]"}], "[", "i_", "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", "k", "}"}], ",", 
    RowBox[{
     RowBox[{"epsilonmet", "[", 
      RowBox[{"i", ",", "k", ",", 
       RowBox[{"-", "j"}]}], "]"}], 
     RowBox[{
      RowBox[{"cd", "[", 
       RowBox[{"-", "k"}], "]"}], "[", "v", "]"}]}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.790873324041769*^9, 3.790873374986455*^9}, 
   3.79087550869764*^9, {3.791382030209566*^9, 3.791382107518795*^9}},
 CellLabel->"In[281]:=",
 CellID->1098684169,ExpressionUUID->"9ffe7d13-e6e7-4505-8c91-344331278eea"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Apply it to our vector v:", "MathCaption",
 CellChangeTimes->{{3.7913810054096947`*^9, 3.79138101172195*^9}},
 CellID->1773697533,ExpressionUUID->"9635fec6-88e8-4225-a206-714d6206d967"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"xCurl", "[", 
   RowBox[{
    RowBox[{"v", "[", "j", "]"}], ",", "j"}], "]"}], "[", "i", "]"}]], "Input",\

 CellChangeTimes->{{3.790875510171784*^9, 3.790875516289357*^9}, {
  3.7908755851921387`*^9, 3.790875591038911*^9}, {3.791382128502475*^9, 
  3.7913821319022427`*^9}},
 CellLabel->"In[282]:=",
 CellID->1882649907,ExpressionUUID->"db5a8743-b3ea-44d2-9e8c-b367b914e825"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{"2", " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Cos",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cos,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], " ", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]}],
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", 
              RowBox[{"2", " ", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"Cos"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cos, Editable -> False], "[", 
                  InterpretationBox[
                   StyleBox[
                   "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`\[Theta][], Editable -> False], "]"}], " ", 
                InterpretationBox[
                 StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`r[], Editable -> False]}]}]],
           Annotation[#, {1} -> 2 Cos[
               $CellContext`\[Theta][]] $CellContext`r[], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{"-", "3"}], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Sin",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Sin,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}]}],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{"-", "3"}], " ", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                  InterpretationBox[
                   StyleBox[
                   "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`\[Theta][], Editable -> False], "]"}]}]}]],
           Annotation[#, {2} -> (-3) Sin[
               $CellContext`\[Theta][]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{"3", " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Cot",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cot,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}]}],
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", 
              RowBox[{"3", " ", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"Cot"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cot, Editable -> False], "[", 
                  InterpretationBox[
                   StyleBox[
                   "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`\[Theta][], Editable -> False], "]"}]}]}]],
           Annotation[#, {3} -> 3 Cot[
               $CellContext`\[Theta][]], "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"i"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{2 Cos[
      $CellContext`\[Theta][]] $CellContext`r[], (-3) Sin[
      $CellContext`\[Theta][]], 3 Cot[
      $CellContext`\[Theta][]]}, {$CellContext`sph}, 0][$CellContext`i],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.790875516745356*^9, {3.790875586805578*^9, 3.79087561700828*^9}, 
   3.79087567643274*^9, 3.790875829470278*^9, 3.7908765896057787`*^9, 
   3.790876667152864*^9, 3.790876707310792*^9, 3.79138101336425*^9, 
   3.791382136303969*^9, 3.7913896454840307`*^9, 3.791389727966875*^9, 
   3.791389791418635*^9, 3.7945179506963997`*^9, 3.794520778954526*^9},
 CellLabel->"Out[282]=",
 CellID->1258920038,ExpressionUUID->"7af7eadb-3546-4638-911d-a584dbc56e3a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
To compare to WL's builtin we need to convert the original vector to the \
orthonormal basis, perform the computation and then change back to the \
coordinate basis:\
\>", "MathCaption",
 CellChangeTimes->{{3.7908756828832493`*^9, 3.790875736625782*^9}},
 CellID->2040425438,ExpressionUUID->"845d3065-c7b6-4828-a25f-8c7fbb273271"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"First", "[", 
  RowBox[{"ToCTensor", "[", 
   RowBox[{"v", ",", 
    RowBox[{"{", "ortho", "}"}]}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.79087552274067*^9, 3.790875554288776*^9}},
 CellLabel->"In[283]:=",
 CellID->1551436921,ExpressionUUID->"7353ae65-3b36-43c4-a041-c8a273d4b433"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox["Cos",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Cos,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}], ",", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"], " ", 
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}]}]}], "}"}]], "Output",
 CellChangeTimes->{{3.7908755279782457`*^9, 3.790875554638979*^9}, 
   3.79087562732451*^9, 3.7908757386652403`*^9, 3.790875829528002*^9, 
   3.790876589656856*^9, 3.790876667205113*^9, 3.790876707362409*^9, 
   3.791381022520637*^9, 3.7913821431483*^9, 3.791389645531505*^9, 
   3.791389728012081*^9, 3.791389791471849*^9, 3.7945179507346563`*^9, 
   3.794520778997744*^9},
 CellLabel->"Out[283]=",
 CellID->1660188422,ExpressionUUID->"66e20a52-ad1f-484b-aeb5-fa803027db1b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Curl", "[", 
  RowBox[{"%", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"r", "[", "]"}], ",", 
     RowBox[{"\[Theta]", "[", "]"}], ",", 
     RowBox[{"\[Phi]", "[", "]"}]}], "}"}], ",", "\"\<Spherical\>\""}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.7908755292423058`*^9, 3.790875544329484*^9}},
 CellLabel->"In[284]:=",
 CellID->611133573,ExpressionUUID->"3b900027-ac68-42fa-bb95-009c0110c12a"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"2", " ", 
    RowBox[{
     InterpretationBox[
      StyleBox["Cos",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Cos,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], " ", 
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", "3"}], " ", 
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], " ", 
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}]}], ",", 
   RowBox[{"3", " ", 
    RowBox[{
     InterpretationBox[
      StyleBox["Cos",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Cos,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], " ", 
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{{3.790875544687262*^9, 3.790875555770805*^9}, 
   3.790875628338787*^9, 3.79087574112281*^9, 3.790875829619548*^9, 
   3.790876589720563*^9, 3.79087666730506*^9, 3.790876707433673*^9, 
   3.791381025511024*^9, 3.791382143894925*^9, 3.791389645598625*^9, 
   3.791389728071506*^9, 3.7913897915305862`*^9, 3.794517950790984*^9, 
   3.794520779054682*^9},
 CellLabel->"Out[284]=",
 CellID->531783083,ExpressionUUID->"9c10164e-b3ac-4d60-97cc-5057f02d1ba0"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{"%", ",", 
   RowBox[{"{", "ortho", "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.7908756305450983`*^9, 3.790875637190001*^9}},
 CellLabel->"In[285]:=",
 CellID->1078768095,ExpressionUUID->"f775012f-d390-4249-968c-eea04d5c0b8e"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"2", " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["Cos",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Cos,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ",", 
     RowBox[{
      RowBox[{"-", "3"}], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}]}], ",", 
     RowBox[{"3", " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["Cos",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Cos,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", "ortho", "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.790875637533606*^9, 3.790875743918594*^9, 
  3.790875829670701*^9, 3.790876589760578*^9, 3.7908766673616533`*^9, 
  3.790876707486742*^9, 3.791381029718627*^9, 3.791382144468796*^9, 
  3.791389645648835*^9, 3.791389728115597*^9, 3.791389791576664*^9, 
  3.794517950827877*^9, 3.79452077909578*^9},
 CellLabel->"Out[285]=",
 CellID->1998820814,ExpressionUUID->"58380817-3aa3-40cd-b6e9-b093f1123c1c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToCTensor", "[", 
   RowBox[{"%", ",", 
    RowBox[{"{", "sph", "}"}]}], "]"}], "[", "i", "]"}]], "Input",
 CellChangeTimes->{{3.7908756405822773`*^9, 3.7908756572205*^9}},
 CellLabel->"In[286]:=",
 CellID->1559514164,ExpressionUUID->"f881812f-144c-40e2-8b93-2d5ed7019e82"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{"2", " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Cos",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cos,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], " ", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]}],
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", 
              RowBox[{"2", " ", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"Cos"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cos, Editable -> False], "[", 
                  InterpretationBox[
                   StyleBox[
                   "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`\[Theta][], Editable -> False], "]"}], " ", 
                InterpretationBox[
                 StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`r[], Editable -> False]}]}]],
           Annotation[#, {1} -> 2 Cos[
               $CellContext`\[Theta][]] $CellContext`r[], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{"-", "3"}], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Sin",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Sin,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}]}],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{"-", "3"}], " ", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                  InterpretationBox[
                   StyleBox[
                   "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`\[Theta][], Editable -> False], "]"}]}]}]],
           Annotation[#, {2} -> (-3) Sin[
               $CellContext`\[Theta][]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{"3", " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Cot",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cot,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}]}],
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", 
              RowBox[{"3", " ", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"Cot"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cot, Editable -> False], "[", 
                  InterpretationBox[
                   StyleBox[
                   "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`\[Theta][], Editable -> False], "]"}]}]}]],
           Annotation[#, {3} -> 3 Cot[
               $CellContext`\[Theta][]], "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"i"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{2 Cos[
      $CellContext`\[Theta][]] $CellContext`r[], (-3) Sin[
      $CellContext`\[Theta][]], 3 Cot[
      $CellContext`\[Theta][]]}, {$CellContext`sph}, 0][$CellContext`i],
  Editable->False]], "Output",
 CellChangeTimes->{3.790875646771535*^9, 3.790875745024456*^9, 
  3.790875829727744*^9, 3.7908765898126373`*^9, 3.790876667421473*^9, 
  3.7908767075649147`*^9, 3.791381031142046*^9, 3.791382145058929*^9, 
  3.791389645715169*^9, 3.791389728166009*^9, 3.7913897916295853`*^9, 
  3.7945179508682947`*^9, 3.7945207791460457`*^9},
 CellLabel->"Out[286]=",
 CellID->1094730877,ExpressionUUID->"e37bc666-85b2-4b13-b576-e6d3d9776675"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "A final warning on xCurl vs Curl: our definition Hodge-dualizes only one \
index, leaving all other indices untouched, while WL's definition \
Hodge-dualizes all tensor indices, meaning that the original tensor needs to \
have less indices than the dimension (otherwise the result would be zero). In \
other words, for a tensor of rank n in dimension d (d is 3 in our example \
here), xCurl will produce a result of rank d + n - 3, while Curl will produce \
a result of rank d - n - 1 (see a discussion in [",
 ButtonBox["MSE",
  BaseStyle->"Hyperlink",
  ButtonData->{
    URL["https://mathematica.stackexchange.com/questions/191373/what-is-the-\
definition-of-curl-in-mathematica"], None},
  ButtonNote->
   "https://mathematica.stackexchange.com/questions/191373/what-is-the-\
definition-of-curl-in-mathematica"],
 "]). For n = 1 both formulas give the same result of rank d - 2, but will \
differ for other input ranks."
}], "MathCaption",
 CellChangeTimes->{{3.791391131374173*^9, 3.7913913562306557`*^9}, {
  3.791391400156659*^9, 3.791391468907185*^9}, {3.791391534520734*^9, 
  3.7913916177580843`*^9}},
 CellID->198307171,ExpressionUUID->"cbfb09a4-f913-44ff-a2f9-f6479a1c17df"]
}, Closed]],

Cell[CellGroupData[{

Cell["3.4.6. xLaplacian", "Subsubsection",
 CellChangeTimes->{{3.728676331871439*^9, 3.728676332031807*^9}, 
   3.790876746494073*^9, {3.790876916730966*^9, 3.790876918474334*^9}, {
   3.790876994119276*^9, 3.790877010006895*^9}, {3.79138104976902*^9, 
   3.791381053712916*^9}},
 CellID->916191242,ExpressionUUID->"e3e725b5-3945-4f84-98b4-2b59b83a4ebc"],

Cell["\<\
Finally we can construct the Laplacian, as a combination of xGrad and xDiv. \
It takes a scalar field and returns another scalar field:\
\>", "Text",
 CellChangeTimes->{{3.790863109490128*^9, 3.790863258626314*^9}, {
  3.790866420158791*^9, 3.790866466799429*^9}, {3.790867225498068*^9, 
  3.79086723724268*^9}, {3.790868709189741*^9, 3.7908689326766863`*^9}, {
  3.790869830533938*^9, 3.79086991508281*^9}, {3.79087041083344*^9, 
  3.790870426871901*^9}, {3.790870648034322*^9, 3.7908707070637093`*^9}, {
  3.7908727496977987`*^9, 3.7908727625849524`*^9}, {3.7908732630164347`*^9, 
  3.79087327540744*^9}, {3.790873390660242*^9, 3.790873414041915*^9}, {
  3.791380917388586*^9, 3.791380943579883*^9}, {3.791381060366744*^9, 
  3.7913811024869623`*^9}},
 CellID->968455304,ExpressionUUID->"b1d67a70-9242-4c8b-b348-161bced1e008"],

Cell[CellGroupData[{

Cell["\<\
The epsilon tensor is fully antisymmetric. Again it will have different forms \
in different bases, being particularly simple in the orthonormal basis:\
\>", "MathCaption",
 CellChangeTimes->{{3.790875378526244*^9, 3.790875449659071*^9}, {
  3.791380960994999*^9, 3.791380976683111*^9}},
 CellID->1049733715,ExpressionUUID->"8656ce63-b738-48b8-920a-654df330f3b9"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"xLaplacian", "[", "scalar_", "]"}], "[", "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", "i", "}"}], ",", 
    RowBox[{
     RowBox[{"xDiv", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"xGrad", "[", "scalar", "]"}], "[", "i", "]"}], ",", "i"}], 
      "]"}], "[", "]"}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.79087534307516*^9, 3.790875344831394*^9}, {
  3.790875385700342*^9, 3.790875429987072*^9}, {3.791381106670734*^9, 
  3.7913811669573917`*^9}, {3.7913821641236486`*^9, 3.791382193515623*^9}},
 CellLabel->"In[287]:=",
 CellID->1296387900,ExpressionUUID->"0b3ab37d-f8b0-4004-b4a1-98a9df45e43c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"xLaplacian", "[", 
    RowBox[{"f", "[", 
     RowBox[{
      RowBox[{"r", "[", "]"}], ",", 
      RowBox[{"\[Theta]", "[", "]"}], ",", 
      RowBox[{"\[Phi]", "[", "]"}]}], "]"}], "]"}], "[", "]"}], "//", 
  "Simplify"}]], "Input",
 CellChangeTimes->{{3.7913812252585173`*^9, 3.791381279081477*^9}},
 CellLabel->"In[288]:=",
 CellID->83753186,ExpressionUUID->"6d754ff9-6311-428a-a70e-ecc0503dc402"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   RowBox[{
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Csc",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Csc,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"], " ", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["f",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`f,
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"0", ",", "0", ",", "2"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Phi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Phi][],
       Editable->False]}], "]"}]}], "+", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox["Cot",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Cot,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], " ", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["f",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`f,
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"0", ",", "1", ",", "0"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Phi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Phi][],
       Editable->False]}], "]"}]}], "+", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["f",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`f,
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "2", ",", "0"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Phi]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Phi][],
      Editable->False]}], "]"}], "+", 
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], " ", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["f",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`f,
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"1", ",", "0", ",", "0"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Phi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Phi][],
       Editable->False]}], "]"}]}], "+", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"], " ", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["f",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`f,
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"2", ",", "0", ",", "0"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Phi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Phi][],
       Editable->False]}], "]"}]}]}], 
  SuperscriptBox[
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False], "2"]]], "Output",
 CellChangeTimes->{{3.791381240424595*^9, 3.79138127954283*^9}, 
   3.791382206238502*^9, 3.791389645921932*^9, 3.79138972837672*^9, 
   3.7913897918414593`*^9, 3.794517951076552*^9, 3.794520779353929*^9},
 CellLabel->"Out[288]=",
 CellID->1649445890,ExpressionUUID->"641b3c7d-b4a8-49bb-8907-1e3e0117aa93"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
In this case the comparison with WL's builtin is direct, because there are no \
vectors involved in input or output:\
\>", "MathCaption",
 CellChangeTimes->{{3.791381309664484*^9, 3.791381336591469*^9}},
 CellID->1964110742,ExpressionUUID->"1f03adee-8230-49ba-b565-b25fe89b3504"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Laplacian", "[", 
   RowBox[{
    RowBox[{"f", "[", 
     RowBox[{
      RowBox[{"r", "[", "]"}], ",", 
      RowBox[{"\[Theta]", "[", "]"}], ",", 
      RowBox[{"\[Phi]", "[", "]"}]}], "]"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"r", "[", "]"}], ",", 
      RowBox[{"\[Theta]", "[", "]"}], ",", 
      RowBox[{"\[Phi]", "[", "]"}]}], "}"}], ",", "\"\<Spherical\>\""}], 
   "]"}], "//", "Simplify"}]], "Input",
 CellChangeTimes->{{3.791381255784482*^9, 3.791381293409218*^9}},
 CellLabel->"In[289]:=",
 CellID->1499017740,ExpressionUUID->"bb1bbd52-c764-42cb-bb91-0f4029e795be"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   RowBox[{
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Csc",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Csc,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"], " ", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["f",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`f,
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"0", ",", "0", ",", "2"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Phi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Phi][],
       Editable->False]}], "]"}]}], "+", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox["Cot",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Cot,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], " ", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["f",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`f,
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"0", ",", "1", ",", "0"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Phi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Phi][],
       Editable->False]}], "]"}]}], "+", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["f",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`f,
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "2", ",", "0"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Phi]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Phi][],
      Editable->False]}], "]"}], "+", 
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], " ", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["f",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`f,
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"1", ",", "0", ",", "0"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Phi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Phi][],
       Editable->False]}], "]"}]}], "+", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"], " ", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["f",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`f,
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"2", ",", "0", ",", "0"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Phi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Phi][],
       Editable->False]}], "]"}]}]}], 
  SuperscriptBox[
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False], "2"]]], "Output",
 CellChangeTimes->{{3.791381271328425*^9, 3.791381293877088*^9}, 
   3.791382208089902*^9, 3.791389646001058*^9, 3.79138972845763*^9, 
   3.791389791938155*^9, 3.7945179511522923`*^9, 3.794520779425156*^9},
 CellLabel->"Out[289]=",
 CellID->370674107,ExpressionUUID->"d8d6b0ee-39d8-4210-8028-730405491035"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.4.7. Generalizations", "Subsubsection",
 CellChangeTimes->{{3.728676331871439*^9, 3.728676332031807*^9}, 
   3.790876746494073*^9, {3.790876916730966*^9, 3.790876918474334*^9}, {
   3.790876994119276*^9, 3.790877010006895*^9}, {3.79138104976902*^9, 
   3.791381053712916*^9}, {3.791382235563252*^9, 3.791382241882552*^9}},
 CellID->868674767,ExpressionUUID->"197e099a-255c-47bb-9218-1bc6b25bba11"],

Cell["\<\
We said that xGrad and xLaplacian act on scalars, and that xDiv and xCurl act \
on vectors. Actually, they can act on any tensor, and our definitions above \
already handle those extensions. Let us see how it works.\
\>", "Text",
 CellChangeTimes->{{3.790863109490128*^9, 3.790863258626314*^9}, {
  3.790866420158791*^9, 3.790866466799429*^9}, {3.790867225498068*^9, 
  3.79086723724268*^9}, {3.790868709189741*^9, 3.7908689326766863`*^9}, {
  3.790869830533938*^9, 3.79086991508281*^9}, {3.79087041083344*^9, 
  3.790870426871901*^9}, {3.790870648034322*^9, 3.7908707070637093`*^9}, {
  3.7908727496977987`*^9, 3.7908727625849524`*^9}, {3.7908732630164347`*^9, 
  3.79087327540744*^9}, {3.790873390660242*^9, 3.790873414041915*^9}, {
  3.791380917388586*^9, 3.791380943579883*^9}, {3.791381060366744*^9, 
  3.7913811024869623`*^9}, {3.791382254921803*^9, 3.791382331982356*^9}, {
  3.791382387197455*^9, 3.791382407805159*^9}},
 CellID->1222976663,ExpressionUUID->"b4985408-6528-42e8-9a37-1bb897c6961b"],

Cell[CellGroupData[{

Cell["Let us take, for example, the gradient of our vector field:", \
"MathCaption",
 CellChangeTimes->{{3.790875378526244*^9, 3.790875449659071*^9}, {
  3.791380960994999*^9, 3.791380976683111*^9}, {3.791382466675681*^9, 
  3.7913824826988707`*^9}},
 CellID->1002789604,ExpressionUUID->"0aada75f-aca7-4a33-990c-5c7b6c2e0d9b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"grad", "=", 
  RowBox[{
   RowBox[{"xGrad", "[", 
    RowBox[{"v", "[", "i", "]"}], "]"}], "[", "j", "]"}]}]], "Input",
 CellChangeTimes->{{3.791382483973365*^9, 3.791382492218658*^9}, {
  3.7913825453999157`*^9, 3.791382547176528*^9}, {3.791382699733615*^9, 
  3.791382700532157*^9}},
 CellLabel->"In[290]:=",
 CellID->587006693,ExpressionUUID->"f2ce655b-f526-49d8-8e4a-0d2a7607bad4"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`r[], Editable -> False]}]}]],
           Annotation[#, {1, 1} -> 2 $CellContext`r[], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{"-", 
             RowBox[{
              InterpretationBox[
               StyleBox["Cos",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cos,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{"-", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"Cos"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cos, Editable -> False], "[", 
                  InterpretationBox[
                   StyleBox[
                   "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`\[Theta][], Editable -> False], "]"}]}]}]],
           Annotation[#, {1, 2} -> -Cos[
               $CellContext`\[Theta][]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{"-", "1"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", 
              RowBox[{"-", "1"}]}]],
           Annotation[#, {1, 3} -> -1, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{"2", " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Cos",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cos,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{"2", " ", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"Cos"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cos, Editable -> False], "[", 
                  InterpretationBox[
                   StyleBox[
                   "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`\[Theta][], Editable -> False], "]"}]}]}]],
           Annotation[#, {2, 1} -> 2 Cos[
               $CellContext`\[Theta][]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{"1", "-", 
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}]}], 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{"1", "-", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                   InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}]}], 
               InterpretationBox[
                StyleBox[
                "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                $CellContext`r[], Editable -> False]]}]],
           Annotation[#, {2, 2} -> $CellContext`r[]^(-1) (1 - Sin[
               $CellContext`\[Theta][]]), "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{"-", 
             FractionBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Cot",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Cot,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", 
              RowBox[{"-", 
                FractionBox[
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Cot"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cot, Editable -> False], "[", 
                   InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                 InterpretationBox[
                  StyleBox[
                  "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`r[], Editable -> False]]}]}]],
           Annotation[#, {2, 3} -> -Cot[
               $CellContext`\[Theta][]]/$CellContext`r[], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["2",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "2"}]],
           Annotation[#, {3, 1} -> 2, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{
              InterpretationBox[
               StyleBox["Cot",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cot,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"Cot"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], Cot, Editable -> False], "[", 
                 InterpretationBox[
                  StyleBox[
                  "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                   False], 
                  $CellContext`\[Theta][], Editable -> False], "]"}], 
               InterpretationBox[
                StyleBox[
                "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                $CellContext`r[], Editable -> False]]}]],
           Annotation[#, {3, 2} -> Cot[
               $CellContext`\[Theta][]]/$CellContext`r[], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{
              RowBox[{"(", 
               RowBox[{"1", "+", 
                RowBox[{
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False], "]"}], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cot",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cot,
                   Editable->False], "[", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False], "]"}]}]}], ")"}], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["Csc",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 Csc,
                 Editable->False], "[", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False], "]"}], "2"]}], 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 RowBox[{"(", 
                   RowBox[{"1", "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Cos"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cos, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Cot"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cot, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}]}]}], 
                   ")"}], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Csc"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Csc, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                  "2"]}], 
               InterpretationBox[
                StyleBox[
                "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                $CellContext`r[], Editable -> False]]}]],
           Annotation[#, {3, 3} -> (1 + Cos[
                 $CellContext`\[Theta][]] Cot[
                 $CellContext`\[Theta][]]) Csc[
                $CellContext`\[Theta][]]^2/$CellContext`r[], "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 4 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 4 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"i", "j"},
         {" ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{2 $CellContext`r[], -Cos[
       $CellContext`\[Theta][]], -1}, {2 Cos[
       $CellContext`\[Theta][]], $CellContext`r[]^(-1) (1 - Sin[
       $CellContext`\[Theta][]]), -Cot[
       $CellContext`\[Theta][]]/$CellContext`r[]}, {2, Cot[
       $CellContext`\[Theta][]]/$CellContext`r[], (1 + Cos[
         $CellContext`\[Theta][]] Cot[
         $CellContext`\[Theta][]]) Csc[
        $CellContext`\[Theta][]]^2/$CellContext`r[]}}, {$CellContext`sph, \
$CellContext`sph}, 0][$CellContext`i, $CellContext`j],
  Editable->False]], "Output",
 CellChangeTimes->{{3.79138249281078*^9, 3.791382502042919*^9}, 
   3.791382547738996*^9, 3.7913827011423283`*^9, 3.791389646068644*^9, 
   3.791389728521595*^9, 3.7913897920048513`*^9, 3.7945179512116833`*^9, 
   3.794520779485361*^9},
 CellLabel->"Out[290]=",
 CellID->69656402,ExpressionUUID->"e1887f5e-0c8e-4a76-a52d-32211774579c"]
}, Open  ]]
}, Open  ]],

Cell["\<\
What happened here is that the new index (j) was added at the end of the \
expression. In other words, the general rule is that xGrad will add a new \
index, at the end. The particular case of differentiating a scalar adds a new \
index at the end, which just happens to be the only index, and hence returns \
a vector.\
\>", "MathCaption",
 CellChangeTimes->{{3.7913825134412813`*^9, 3.791382625206352*^9}},
 CellID->2134094243,ExpressionUUID->"ed9812dc-50b3-479c-9514-9342ec91209f"],

Cell[CellGroupData[{

Cell["\<\
The WL builtin can also do this, and as usual, the comparison involves just \
changes of basis:\
\>", "MathCaption",
 CellChangeTimes->{{3.7913826356542673`*^9, 3.791382662501233*^9}},
 CellID->1767026200,ExpressionUUID->"f643908d-c63f-4d63-9eb3-b5a5906336ff"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"res", "=", 
  RowBox[{"ToCTensor", "[", 
   RowBox[{
    RowBox[{"Head", "[", "grad", "]"}], ",", 
    RowBox[{"{", 
     RowBox[{"ortho", ",", "ortho"}], "}"}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.790872523405451*^9, 3.790872547808359*^9}, {
  3.791382724435629*^9, 3.791382744770197*^9}, {3.791382819825096*^9, 
  3.79138283696742*^9}},
 CellLabel->"In[291]:=",
 CellID->87906517,ExpressionUUID->"45e4d2ae-bc32-424d-be98-2ab81a55be39"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]}], ",", 
       RowBox[{
        RowBox[{"-", 
         RowBox[{
          InterpretationBox[
           StyleBox["Cos",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Cos,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}]}], " ", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]}], ",", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]}], " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["Sin",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Sin,
          Editable->False], "[", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False], "]"}]}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"2", " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["Cos",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Cos,
          Editable->False], "[", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False], "]"}], " ", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]}], ",", 
       RowBox[{
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], " ", 
         RowBox[{
          InterpretationBox[
           StyleBox["Sin",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Sin,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}]}]}], ",", 
       RowBox[{
        RowBox[{"-", 
         RowBox[{
          InterpretationBox[
           StyleBox["Cos",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Cos,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}]}], " ", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["Sin",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Sin,
          Editable->False], "[", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False], "]"}]}], ",", 
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["Cos",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Cos,
          Editable->False], "[", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False], "]"}], " ", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]}], ",", 
       RowBox[{
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "+", 
        RowBox[{
         RowBox[{
          InterpretationBox[
           StyleBox["Cos",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Cos,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], " ", 
         RowBox[{
          InterpretationBox[
           StyleBox["Cot",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Cot,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], " ", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]}]}]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"ortho", ",", "ortho"}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{
  3.79138270351648*^9, 3.791382745123427*^9, {3.791382821397653*^9, 
   3.7913828372093697`*^9}, 3.7913896461333513`*^9, 3.791389728576603*^9, 
   3.791389792073146*^9, 3.7945179512620287`*^9, 3.794520779542005*^9},
 CellLabel->"Out[291]=",
 CellID->336560799,ExpressionUUID->"06998c2a-9c0b-46c2-a77a-f102cf2e0c3c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"First", "@", 
  RowBox[{"ToCTensor", "[", 
   RowBox[{"v", ",", 
    RowBox[{"{", "ortho", "}"}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.791382755075986*^9, 3.791382777400447*^9}},
 CellLabel->"In[292]:=",
 CellID->649106798,ExpressionUUID->"a134847e-2eb4-4d78-90ee-5fbbb8bcb2b6"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox["Cos",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Cos,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}], ",", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"], " ", 
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}]}]}], "}"}]], "Output",
 CellChangeTimes->{{3.7913827604482822`*^9, 3.7913828379439697`*^9}, 
   3.791389646177856*^9, 3.7913897286199207`*^9, 3.7913897921169643`*^9, 
   3.7945179513117027`*^9, 3.7945207795801497`*^9},
 CellLabel->"Out[292]=",
 CellID->105422595,ExpressionUUID->"9a4fce62-a3aa-4369-a506-d3bf25b0cb8d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Grad", "[", 
   RowBox[{"%", ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"r", "[", "]"}], ",", 
      RowBox[{"\[Theta]", "[", "]"}], ",", 
      RowBox[{"\[Phi]", "[", "]"}]}], "}"}], ",", "\"\<Spherical\>\""}], 
   "]"}], "//", "Simplify"}]], "Input",
 CellChangeTimes->{{3.791382781627572*^9, 3.791382802463484*^9}},
 CellLabel->"In[293]:=",
 CellID->1707919238,ExpressionUUID->"494774e5-1a40-4baf-b655-c54b43c60397"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ",", 
     RowBox[{
      RowBox[{"-", 
       RowBox[{
        InterpretationBox[
         StyleBox["Cos",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Cos,
         Editable->False], "[", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False], "]"}]}], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ",", 
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"2", " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["Cos",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Cos,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], " ", 
       RowBox[{
        InterpretationBox[
         StyleBox["Sin",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Sin,
         Editable->False], "[", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False], "]"}]}]}], ",", 
     RowBox[{
      RowBox[{"-", 
       RowBox[{
        InterpretationBox[
         StyleBox["Cos",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Cos,
         Editable->False], "[", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False], "]"}]}], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}]}], ",", 
     RowBox[{
      RowBox[{
       InterpretationBox[
        StyleBox["Cos",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Cos,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "+", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox["Cos",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Cos,
         Editable->False], "[", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False], "]"}], " ", 
       RowBox[{
        InterpretationBox[
         StyleBox["Cot",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Cot,
         Editable->False], "[", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False], "]"}], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}]}]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{{3.791382795737595*^9, 3.791382838616259*^9}, 
   3.791389646237727*^9, 3.791389728687139*^9, 3.791389792180372*^9, 
   3.794517951368854*^9, 3.794520779638516*^9},
 CellLabel->"Out[293]=",
 CellID->163545390,ExpressionUUID->"a0bd25ce-a61f-4403-9db1-e5aa7c37a9c8"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "===", 
  RowBox[{"First", "[", "res", "]"}]}]], "Input",
 CellChangeTimes->{{3.791382828224594*^9, 3.791382840087631*^9}},
 CellLabel->"In[294]:=",
 CellID->1608572934,ExpressionUUID->"2614020d-9d36-4bab-b5ec-6a281421e3a8"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{{3.7913828313238277`*^9, 3.7913828403591022`*^9}, 
   3.7913896462793207`*^9, 3.791389728730928*^9, 3.791389792225252*^9, 
   3.7945179514097977`*^9, 3.794520779677754*^9},
 CellLabel->"Out[294]=",
 CellID->2075785217,ExpressionUUID->"1eba6922-7b2b-40e9-86ae-aec13edc7133"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Now that we have a two-tensor grad, let us compute its divergence for \
example. But there are two indices in a two-tensor. Which one should we \
differentiate? This is where the index argument of xDiv comes in handy. The \
two-tensor grad has indices i, j. Let us take divergences in both separately:\
\
\>", "MathCaption",
 CellChangeTimes->{{3.7913829066679783`*^9, 3.791383035497295*^9}},
 CellID->2129093082,ExpressionUUID->"fdd1e62c-efa7-47b1-9535-17c0c7f29a8e"],

Cell[CellGroupData[{

Cell[BoxData["grad"], "Input",
 CellChangeTimes->{{3.7913829922442083`*^9, 3.7913829930267267`*^9}},
 CellLabel->"In[295]:=",
 CellID->1096420752,ExpressionUUID->"2d209ba0-ff36-4574-95ce-8518dfc5114b"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`r[], Editable -> False]}]}]],
           Annotation[#, {1, 1} -> 2 $CellContext`r[], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{"-", 
             RowBox[{
              InterpretationBox[
               StyleBox["Cos",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cos,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{"-", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"Cos"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cos, Editable -> False], "[", 
                  InterpretationBox[
                   StyleBox[
                   "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`\[Theta][], Editable -> False], "]"}]}]}]],
           Annotation[#, {1, 2} -> -Cos[
               $CellContext`\[Theta][]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{"-", "1"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", 
              RowBox[{"-", "1"}]}]],
           Annotation[#, {1, 3} -> -1, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{"2", " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Cos",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cos,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{"2", " ", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"Cos"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cos, Editable -> False], "[", 
                  InterpretationBox[
                   StyleBox[
                   "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`\[Theta][], Editable -> False], "]"}]}]}]],
           Annotation[#, {2, 1} -> 2 Cos[
               $CellContext`\[Theta][]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{"1", "-", 
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}]}], 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{"1", "-", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                   InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}]}], 
               InterpretationBox[
                StyleBox[
                "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                $CellContext`r[], Editable -> False]]}]],
           Annotation[#, {2, 2} -> $CellContext`r[]^(-1) (1 - Sin[
               $CellContext`\[Theta][]]), "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{"-", 
             FractionBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Cot",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Cot,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", 
              RowBox[{"-", 
                FractionBox[
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Cot"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cot, Editable -> False], "[", 
                   InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                 InterpretationBox[
                  StyleBox[
                  "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`r[], Editable -> False]]}]}]],
           Annotation[#, {2, 3} -> -Cot[
               $CellContext`\[Theta][]]/$CellContext`r[], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["2",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "2"}]],
           Annotation[#, {3, 1} -> 2, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{
              InterpretationBox[
               StyleBox["Cot",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cot,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"Cot"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], Cot, Editable -> False], "[", 
                 InterpretationBox[
                  StyleBox[
                  "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                   False], 
                  $CellContext`\[Theta][], Editable -> False], "]"}], 
               InterpretationBox[
                StyleBox[
                "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                $CellContext`r[], Editable -> False]]}]],
           Annotation[#, {3, 2} -> Cot[
               $CellContext`\[Theta][]]/$CellContext`r[], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{
              RowBox[{"(", 
               RowBox[{"1", "+", 
                RowBox[{
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False], "]"}], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cot",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cot,
                   Editable->False], "[", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False], "]"}]}]}], ")"}], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["Csc",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 Csc,
                 Editable->False], "[", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False], "]"}], "2"]}], 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 RowBox[{"(", 
                   RowBox[{"1", "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Cos"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cos, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Cot"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cot, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}]}]}], 
                   ")"}], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Csc"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Csc, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                  "2"]}], 
               InterpretationBox[
                StyleBox[
                "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                $CellContext`r[], Editable -> False]]}]],
           Annotation[#, {3, 3} -> (1 + Cos[
                 $CellContext`\[Theta][]] Cot[
                 $CellContext`\[Theta][]]) Csc[
                $CellContext`\[Theta][]]^2/$CellContext`r[], "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 4 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 4 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"i", "j"},
         {" ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{2 $CellContext`r[], -Cos[
       $CellContext`\[Theta][]], -1}, {2 Cos[
       $CellContext`\[Theta][]], $CellContext`r[]^(-1) (1 - Sin[
       $CellContext`\[Theta][]]), -Cot[
       $CellContext`\[Theta][]]/$CellContext`r[]}, {2, Cot[
       $CellContext`\[Theta][]]/$CellContext`r[], (1 + Cos[
         $CellContext`\[Theta][]] Cot[
         $CellContext`\[Theta][]]) Csc[
        $CellContext`\[Theta][]]^2/$CellContext`r[]}}, {$CellContext`sph, \
$CellContext`sph}, 0][$CellContext`i, $CellContext`j],
  Editable->False]], "Output",
 CellChangeTimes->{3.791382993368802*^9, 3.79138303652911*^9, 
  3.7913896463257723`*^9, 3.791389728786764*^9, 3.791389792292618*^9, 
  3.7945179514743834`*^9, 3.794520779766967*^9},
 CellLabel->"Out[295]=",
 CellID->1850296604,ExpressionUUID->"88a54bea-9175-48be-af79-d5e996e723f8"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"gradDivi", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"xDiv", "[", 
     RowBox[{"grad", ",", "i"}], "]"}], "[", "]"}], "//", 
   "Simplify"}]}]], "Input",
 CellChangeTimes->{{3.79138303851099*^9, 3.7913830518970337`*^9}, {
  3.791383090519772*^9, 3.791383091631785*^9}},
 CellLabel->"In[296]:=",
 CellID->34089487,ExpressionUUID->"74170c86-1985-4f4f-8cbf-1c0c9da5181d"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{"4", "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["Cos",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Cos,
                Editable->False], "[", 
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["Csc",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Csc,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}]}]}],
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", 
              RowBox[{"4", "+", 
                RowBox[{
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Cos"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cos, Editable -> False], "[", 
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False]}], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Csc"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Csc, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}]}]}]}]],
           
           Annotation[#, {1} -> 4 + Cos[2 $CellContext`\[Theta][]] Csc[
                $CellContext`\[Theta][]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "3"}], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False], "]"}]}], "+", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"3", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], ")"}], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["Csc",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 Csc,
                 Editable->False], "[", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False], "]"}], "2"]}], 
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}]],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 RowBox[{"(", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"-", "3"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Cos"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cos, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Cos"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cos, Editable -> False], "[", 
                    RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False]}], "]"}]}], 
                   ")"}], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Csc"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Csc, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                  "2"]}], 
               RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox[
                  "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`r[], Editable -> False]}]]}]],
           Annotation[#, {2} -> Rational[1, 2] ((-3) Cos[
                 $CellContext`\[Theta][]] + Cos[3 $CellContext`\[Theta][]]) 
             Csc[
                $CellContext`\[Theta][]]^2/$CellContext`r[], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3} -> 0, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"j"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{4 + Cos[2 $CellContext`\[Theta][]] Csc[
       $CellContext`\[Theta][]], Rational[1, 2] ((-3) Cos[
        $CellContext`\[Theta][]] + Cos[3 $CellContext`\[Theta][]]) Csc[
       $CellContext`\[Theta][]]^2/$CellContext`r[], 0}, {$CellContext`sph}, 
   0][$CellContext`j],
  Editable->False]], "Output",
 CellChangeTimes->{{3.7913830438525887`*^9, 3.791383052617464*^9}, 
   3.7913830920951033`*^9, 3.7913896464524918`*^9, 3.7913897289214783`*^9, 
   3.791389792424614*^9, 3.794517951686717*^9, 3.7945207798992558`*^9},
 CellLabel->"Out[296]=",
 CellID->224331221,ExpressionUUID->"10d20048-385f-446f-9bbc-69e20d7a7671"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"gradDivj", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"xDiv", "[", 
     RowBox[{"grad", ",", "j"}], "]"}], "[", "]"}], "//", 
   "Simplify"}]}]], "Input",
 CellChangeTimes->{{3.7913830533705873`*^9, 3.791383059288473*^9}, {
  3.791383093911455*^9, 3.7913830984802923`*^9}},
 CellLabel->"In[297]:=",
 CellID->667712078,ExpressionUUID->"ea579ab8-e31d-4391-b62a-c5e7b28597cf"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{"4", "-", 
             RowBox[{"2", " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["Cos",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Cos,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["Cot",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Cot,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}]}], "+", 
             RowBox[{"2", " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}]}]}],
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", 
              RowBox[{"4", "-", 
                RowBox[{"2", " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Cos"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cos, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Cot"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cot, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}]}], "+", 
                RowBox[{"2", " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}]}]}]}]],
           
           Annotation[#, {1} -> 4 - 2 Cos[
               $CellContext`\[Theta][]] Cot[
               $CellContext`\[Theta][]] + 2 Sin[
                $CellContext`\[Theta][]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{"-", 
             FractionBox[
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["Cos",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 Cos,
                 Editable->False], "[", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False], "]"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "3"}], "+", 
                 SuperscriptBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["Cot",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    Cot,
                    Editable->False], "[", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False], "]"}], "2"]}], ")"}]}], 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]]}],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              RowBox[{"-", 
                FractionBox[
                 RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Cos"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cos, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Cot"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cot, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                    "2"]}], ")"}]}], 
                 InterpretationBox[
                  StyleBox[
                  "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`r[], Editable -> False]]}]}]],
           Annotation[#, {2} -> -Cos[
               $CellContext`\[Theta][]] (-3 + Cot[
                 $CellContext`\[Theta][]]^2)/$CellContext`r[], 
            "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            FractionBox["4", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]],
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", 
              FractionBox["4", 
               InterpretationBox[
                StyleBox[
                "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                $CellContext`r[], Editable -> False]]}]],
           Annotation[#, {3} -> 4/$CellContext`r[], "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"i"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{4 - 2 Cos[
      $CellContext`\[Theta][]] Cot[
      $CellContext`\[Theta][]] + 2 Sin[
       $CellContext`\[Theta][]], -Cos[
      $CellContext`\[Theta][]] (-3 + Cot[
        $CellContext`\[Theta][]]^2)/$CellContext`r[], 
    4/$CellContext`r[]}, {$CellContext`sph}, 0][$CellContext`i],
  Editable->False]], "Output",
 CellChangeTimes->{
  3.791383059644992*^9, {3.79138309551727*^9, 3.791383098975849*^9}, 
   3.7913896465120077`*^9, 3.7913897290082827`*^9, 3.791389792530278*^9, 
   3.794517951783696*^9, 3.794520779957803*^9},
 CellLabel->"Out[297]=",
 CellID->1501030862,ExpressionUUID->"4000a53f-3fa6-4743-aefb-7dba51d8b84d"]
}, Open  ]]
}, Open  ]],

Cell["\<\
The general rule is that xDiv[expr, i] will remove index i, leaving all \
others present. The particular case of differentiating a vector removes the \
index of that vector and returns a scalar.\
\>", "MathCaption",
 CellChangeTimes->{{3.7913831441420097`*^9, 3.791383192491046*^9}},
 CellID->1029605588,ExpressionUUID->"34796ba6-6c17-464d-b9cc-7b694288572e"],

Cell[CellGroupData[{

Cell["\<\
The WL builtin can also do this, but only with respect to the last (j in \
gradDivj in our case) index. Let us now compare by converting WL's result \
back to the coordinated basis:\
\>", "MathCaption",
 CellChangeTimes->{{3.791383202714188*^9, 3.79138325249701*^9}, {
  3.791383394684536*^9, 3.791383403669579*^9}},
 CellID->144830439,ExpressionUUID->"8bbaa18c-c261-4315-8986-b48094842e57"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"First", "@", 
  RowBox[{"ToCTensor", "[", 
   RowBox[{
    RowBox[{"Head", "[", "grad", "]"}], ",", 
    RowBox[{"{", 
     RowBox[{"ortho", ",", "ortho"}], "}"}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.791383273113359*^9, 3.7913832856415157`*^9}},
 CellLabel->"In[298]:=",
 CellID->1247079980,ExpressionUUID->"f881ee5d-1fc4-4dd6-84cd-399679570011"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ",", 
     RowBox[{
      RowBox[{"-", 
       RowBox[{
        InterpretationBox[
         StyleBox["Cos",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Cos,
         Editable->False], "[", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False], "]"}]}], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ",", 
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"2", " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["Cos",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Cos,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], " ", 
       RowBox[{
        InterpretationBox[
         StyleBox["Sin",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Sin,
         Editable->False], "[", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False], "]"}]}]}], ",", 
     RowBox[{
      RowBox[{"-", 
       RowBox[{
        InterpretationBox[
         StyleBox["Cos",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Cos,
         Editable->False], "[", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False], "]"}]}], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}]}], ",", 
     RowBox[{
      RowBox[{
       InterpretationBox[
        StyleBox["Cos",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Cos,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "+", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox["Cos",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Cos,
         Editable->False], "[", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False], "]"}], " ", 
       RowBox[{
        InterpretationBox[
         StyleBox["Cot",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Cot,
         Editable->False], "[", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False], "]"}], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}]}]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{{3.791383283006221*^9, 3.791383310029574*^9}, {
   3.7913833615690603`*^9, 3.791383380787634*^9}, 3.791389646553432*^9, 
   3.791389729054023*^9, 3.791389792577965*^9, 3.794517951830059*^9, 
   3.794520779996504*^9},
 CellLabel->"Out[298]=",
 CellID->1335296325,ExpressionUUID->"c53827de-e31d-4533-95ed-b5244f84ab0c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Div", "[", 
   RowBox[{"%", ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"r", "[", "]"}], ",", 
      RowBox[{"\[Theta]", "[", "]"}], ",", 
      RowBox[{"\[Phi]", "[", "]"}]}], "}"}], ",", "\"\<Spherical\>\""}], 
   "]"}], "//", "Simplify"}]], "Input",
 CellChangeTimes->{{3.791383292256611*^9, 3.7913833059685497`*^9}},
 CellLabel->"In[299]:=",
 CellID->873950029,ExpressionUUID->"a02e23d7-2d71-4a4c-bf55-3bd01e04e89f"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"4", "-", 
    RowBox[{"2", " ", 
     RowBox[{
      InterpretationBox[
       StyleBox["Cos",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Cos,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], " ", 
     RowBox[{
      InterpretationBox[
       StyleBox["Cot",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Cot,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}]}], "+", 
    RowBox[{"2", " ", 
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     RowBox[{
      InterpretationBox[
       StyleBox["Cos",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Cos,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}]}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "3"}], "+", 
      SuperscriptBox[
       RowBox[{
        InterpretationBox[
         StyleBox["Cot",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Cot,
         Editable->False], "[", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False], "]"}], "2"]}], ")"}]}], ",", 
   RowBox[{"4", " ", 
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}]}]}], "}"}]], "Output",
 CellChangeTimes->{{3.791383302409419*^9, 3.791383310765856*^9}, {
   3.791383362310768*^9, 3.791383381566711*^9}, 3.791389646622047*^9, 
   3.791389729128541*^9, 3.79138979265105*^9, 3.794517951894136*^9, 
   3.794520780061699*^9},
 CellLabel->"Out[299]=",
 CellID->1121223125,ExpressionUUID->"a4d8f1b7-64a9-43b5-b305-68f311409c04"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToCTensor", "[", 
   RowBox[{
    RowBox[{"CTensor", "[", 
     RowBox[{"%", ",", 
      RowBox[{"{", "ortho", "}"}]}], "]"}], ",", 
    RowBox[{"{", "sph", "}"}]}], "]"}], "[", "i", "]"}]], "Input",
 CellChangeTimes->{{3.791383318303651*^9, 3.7913833784137774`*^9}},
 CellLabel->"In[300]:=",
 CellID->1283630650,ExpressionUUID->"3efe70c2-cb00-4ec5-a6fe-075c3f34b4b5"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{"4", "-", 
             RowBox[{"2", " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["Cos",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Cos,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["Cot",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Cot,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}]}], "+", 
             RowBox[{"2", " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}]}]}],
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", 
              RowBox[{"4", "-", 
                RowBox[{"2", " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Cos"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cos, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Cot"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cot, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}]}], "+", 
                RowBox[{"2", " ", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}]}]}]}]],
           
           Annotation[#, {1} -> 4 - 2 Cos[
               $CellContext`\[Theta][]] Cot[
               $CellContext`\[Theta][]] + 2 Sin[
                $CellContext`\[Theta][]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{"-", 
             FractionBox[
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["Cos",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 Cos,
                 Editable->False], "[", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False], "]"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "3"}], "+", 
                 SuperscriptBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["Cot",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    Cot,
                    Editable->False], "[", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False], "]"}], "2"]}], ")"}]}], 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]]}],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              RowBox[{"-", 
                FractionBox[
                 RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Cos"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cos, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Cot"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cot, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                    "2"]}], ")"}]}], 
                 InterpretationBox[
                  StyleBox[
                  "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`r[], Editable -> False]]}]}]],
           Annotation[#, {2} -> -Cos[
               $CellContext`\[Theta][]] (-3 + Cot[
                 $CellContext`\[Theta][]]^2)/$CellContext`r[], 
            "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            FractionBox["4", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]],
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", 
              FractionBox["4", 
               InterpretationBox[
                StyleBox[
                "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                $CellContext`r[], Editable -> False]]}]],
           Annotation[#, {3} -> 4/$CellContext`r[], "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"i"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{4 - 2 Cos[
      $CellContext`\[Theta][]] Cot[
      $CellContext`\[Theta][]] + 2 Sin[
       $CellContext`\[Theta][]], -Cos[
      $CellContext`\[Theta][]] (-3 + Cot[
        $CellContext`\[Theta][]]^2)/$CellContext`r[], 
    4/$CellContext`r[]}, {$CellContext`sph}, 0][$CellContext`i],
  Editable->False]], "Output",
 CellChangeTimes->{{3.791383346901909*^9, 3.791383382187993*^9}, 
   3.791389646671796*^9, 3.791389729179614*^9, 3.79138979270153*^9, 
   3.794517951941877*^9, 3.794520780106717*^9},
 CellLabel->"Out[300]=",
 CellID->228059450,ExpressionUUID->"8e760cb9-8a10-4785-9c92-ad43eb4f767b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "===", "gradDivj"}]], "Input",
 CellChangeTimes->{{3.791383366376361*^9, 3.791383370534505*^9}},
 CellLabel->"In[301]:=",
 CellID->615280195,ExpressionUUID->"52d0ad8a-9745-410b-9a6c-163c139d44c4"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{{3.791383371308609*^9, 3.791383382780529*^9}, 
   3.791389646737399*^9, 3.7913897292259493`*^9, 3.791389792746449*^9, 
   3.794517951984136*^9, 3.794520780147377*^9},
 CellLabel->"Out[301]=",
 CellID->1884349902,ExpressionUUID->"1f663a81-1369-4873-8096-acc6ca5343b2"]
}, Open  ]]
}, Open  ]],

Cell["\<\
The same type of example could be done with xCurl (where we also have a \
freedom to choose the index to differentiate) and xLaplacian (where we don't \
have a freedom, because xGrad always places the new index at the end, and \
xDiv must differentiate the new index just added, and no other one).\
\>", "MathCaption",
 CellChangeTimes->{{3.791383436028668*^9, 3.791383540945201*^9}},
 CellID->1768506554,ExpressionUUID->"19016ad7-f785-4b92-b322-2990e85c46e6"],

Cell["\<\
Another interesting observation is that our definitions are valid for any \
type of xAct expression, not just CTensor expressions, though the results may \
not be as nice as expected. It is, however, interesting to explore what the \
possibilities are. We will see along the way how some new commands (like \
ContractBasis and BasisExpand) are needed to specify when to perform changes \
of basis.Those commands to manipulate Basis objects have been explained in \
detailed in previous sections.\
\>", "Text",
 CellChangeTimes->{{3.790863109490128*^9, 3.790863258626314*^9}, {
  3.790866420158791*^9, 3.790866466799429*^9}, {3.790867225498068*^9, 
  3.79086723724268*^9}, {3.790868709189741*^9, 3.7908689326766863`*^9}, {
  3.790869830533938*^9, 3.79086991508281*^9}, {3.79087041083344*^9, 
  3.790870426871901*^9}, {3.790870648034322*^9, 3.7908707070637093`*^9}, {
  3.7908727496977987`*^9, 3.7908727625849524`*^9}, {3.7908732630164347`*^9, 
  3.79087327540744*^9}, {3.790873390660242*^9, 3.790873414041915*^9}, {
  3.791380917388586*^9, 3.791380943579883*^9}, {3.791381060366744*^9, 
  3.7913811024869623`*^9}, {3.791382254921803*^9, 3.791382331982356*^9}, {
  3.791382387197455*^9, 3.791382407805159*^9}, {3.7913835886715803`*^9, 
  3.791383641013793*^9}, {3.7913840384245987`*^9, 3.7913840879669228`*^9}, {
  3.791387749683214*^9, 3.791387775129589*^9}},
 CellID->1520660747,ExpressionUUID->"4e48f271-2919-441b-8535-b176aa6b5ac5"],

Cell[CellGroupData[{

Cell["\<\
First, we can use linear combinations of basis vectors. We will need \
ContractBasis to perform contractions of Basis vectors:\
\>", "MathCaption",
 CellChangeTimes->{{3.791383728227152*^9, 3.791383744650626*^9}, {
  3.7913879042607193`*^9, 3.7913879214355183`*^9}},
 CellID->1481543602,ExpressionUUID->"f21e8996-27e2-4e3e-8478-6a467598122e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToBasisExpand", "[", 
  RowBox[{"v", "[", "i", "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.7913837496432257`*^9, 3.7913837551222467`*^9}},
 CellLabel->"In[302]:=",
 CellID->283055898,ExpressionUUID->"4e3846cd-e080-4b3a-8525-cdaa49fde47b"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "i"},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`sph}, $CellContext`i],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "i"},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`sph}, $CellContext`i],
    Editable->False], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Cos",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cos,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}], " ", 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "i"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`sph}, $CellContext`i],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}]}]], "Output",
 CellChangeTimes->{{3.7913837554484262`*^9, 3.791383798309782*^9}, {
   3.791383837279035*^9, 3.791383855833832*^9}, 3.7913840297717953`*^9, {
   3.791387865125386*^9, 3.791387885323963*^9}, 3.7913896467878227`*^9, 
   3.791389729273217*^9, 3.79138979281279*^9, 3.794517952033478*^9, 
   3.7945207801986322`*^9},
 CellLabel->"Out[302]=",
 CellID->1327460605,ExpressionUUID->"01de2fd7-23ad-4847-9554-b248db23c76f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"xDiv", "[", 
   RowBox[{"%", ",", "i"}], "]"}], "[", "]"}]], "Input",
 CellChangeTimes->{{3.7913837623693457`*^9, 3.791383796753147*^9}, {
   3.791383839119523*^9, 3.791383858407111*^9}, 3.791387868966724*^9},
 CellLabel->"In[303]:=",
 CellID->1809987634,ExpressionUUID->"e583d3e2-6d39-464e-8029-dd06c0f8a273"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"i", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`i, {1, $CellContext`sph}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "i"},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`sph}, $CellContext`i],
    Editable->False]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"i", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`i, {1, $CellContext`sph}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "i"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`sph}, $CellContext`i],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]}], "+", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox["Cos",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cos,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Cot",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cot,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}], " ", 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "i"},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`sph}, $CellContext`i],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {"i", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[-$CellContext`i, {1, $CellContext`sph}],
       Editable->False], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["Cos",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Cos,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {"i", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "Rows" -> {{
                  Offset[0.]}}}],
            FontSize->9]}
         },
         GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[-$CellContext`i, {2, $CellContext`sph}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}]}]}], ")"}]}]}]], "Output",
 CellChangeTimes->{{3.791383766972064*^9, 3.791383799206332*^9}, {
   3.79138384199214*^9, 3.79138385880844*^9}, {3.79138786972222*^9, 
   3.7913878860476017`*^9}, 3.791389646841593*^9, 3.791389729341318*^9, 
   3.791389792863912*^9, 3.794517952084691*^9, 3.794520780257307*^9},
 CellLabel->"Out[303]=",
 CellID->422779552,ExpressionUUID->"96288a4f-f4ac-482b-847b-c0d90d3afdcc"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "//", "ContractBasis"}], "//", "Simplify"}]], "Input",
 CellChangeTimes->{{3.79138787860821*^9, 3.791387889734098*^9}},
 CellLabel->"In[304]:=",
 CellID->768461419,ExpressionUUID->"8fbc87ab-1129-4a33-9538-45e83b8240f9"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{"4", "+", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["Cos",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Cos,
       Editable->False], "[", 
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False]}], "]"}], " ", 
     RowBox[{
      InterpretationBox[
       StyleBox["Csc",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Csc,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}]}]}], ")"}], " ", 
  InterpretationBox[
   StyleBox["r",
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`r[],
   Editable->False]}]], "Output",
 CellChangeTimes->{{3.791387881995378*^9, 3.791387890277248*^9}, 
   3.7913896468829403`*^9, 3.791389729384864*^9, 3.791389792940507*^9, 
   3.794517952127698*^9, 3.7945207803014402`*^9},
 CellLabel->"Out[304]=",
 CellID->1780026188,ExpressionUUID->"133e5b2b-2c02-48b1-a3a5-71856b239c5f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Let us now start with a different basis, as specified by BasisExpand. In this \
case we will find Basis change objects, like Basis[{3, -sph}, {1, ortho}]. \
Their values have been already given before, and can be inserted in the \
computation using ToValues:\
\>", "MathCaption",
 CellChangeTimes->{{3.791383884756711*^9, 3.791383905355645*^9}, {
  3.791387925884078*^9, 3.791388027241596*^9}, {3.791388127317099*^9, 
  3.79138812803021*^9}},
 CellID->2016765644,ExpressionUUID->"b9c1dfac-767d-44a0-92eb-728847748b36"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToBasisExpand", "[", 
  RowBox[{"v", "[", "i", "]"}], "]"}]], "Input",
 CellLabel->"In[305]:=",
 CellID->132922870,ExpressionUUID->"9e985458-b242-4b02-a1e0-4d172bad7e00"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "i"},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`sph}, $CellContext`i],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "i"},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`sph}, $CellContext`i],
    Editable->False], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Cos",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cos,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}], " ", 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "i"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`sph}, $CellContext`i],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}]}]], "Output",
 CellChangeTimes->{{3.791383915843968*^9, 3.791383988481852*^9}, {
   3.7913878215226297`*^9, 3.791387845309516*^9}, 3.791387893411951*^9, 
   3.791388028390335*^9, 3.791388060630868*^9, 3.791388296183728*^9, 
   3.791389646927375*^9, 3.791389729436529*^9, 3.791389792998763*^9, 
   3.7945179521780577`*^9, 3.7945207803643017`*^9},
 CellLabel->"Out[305]=",
 CellID->305455359,ExpressionUUID->"39f28d6b-dec6-4812-b2c2-17d8102f8f39"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"BasisExpand", "[", 
  RowBox[{"%", ",", "ortho"}], "]"}]], "Input",
 CellChangeTimes->{{3.791383917215803*^9, 3.7913839774431047`*^9}, 
   3.791387896885603*^9},
 CellLabel->"In[306]:=",
 CellID->512812450,ExpressionUUID->"5ac457bf-7577-446a-8ad9-fe8c33b5950d"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "i"},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`ortho}, $CellContext`i],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`sph}, {1, $CellContext`ortho}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "i"},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`ortho}, $CellContext`i],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`sph}, {2, $CellContext`ortho}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "i"},
           {
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`ortho}, $CellContext`i],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`sph}, {3, $CellContext`ortho}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "i"},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`ortho}, $CellContext`i],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`sph}, {1, $CellContext`ortho}],
    Editable->False], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Cos",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cos,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}], " ", 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "i"},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`ortho}, $CellContext`i],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`sph}, {2, $CellContext`ortho}],
    Editable->False], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Cos",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cos,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}], " ", 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`sph}, {3, $CellContext`ortho}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "i"},
           {
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`ortho}, $CellContext`i],
    Editable->False], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Cos",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cos,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}], " ", 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "i"},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`ortho}, $CellContext`i],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`sph}, {1, $CellContext`ortho}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`sph}, {2, $CellContext`ortho}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "i"},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`ortho}, $CellContext`i],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`sph}, {3, $CellContext`ortho}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "i"},
           {
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`ortho}, $CellContext`i],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}]}]], "Output",
 CellChangeTimes->{
  3.791383923792972*^9, {3.791383959513262*^9, 3.791383990082747*^9}, {
   3.791387822474118*^9, 3.7913878459580917`*^9}, 3.791387897829771*^9, 
   3.7913880297405653`*^9, 3.791388061325746*^9, 3.791388296975931*^9, 
   3.79138964699065*^9, 3.7913897295090933`*^9, 3.791389793065585*^9, 
   3.794517952243784*^9, 3.794520780442541*^9},
 CellLabel->"Out[306]=",
 CellID->1302302383,ExpressionUUID->"27c221b7-2297-4c65-9feb-885259204401"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ToValues"}]], "Input",
 CellChangeTimes->{{3.791388031570026*^9, 3.791388033529426*^9}},
 CellLabel->"In[307]:=",
 CellID->809087849,ExpressionUUID->"55c91845-d52a-41e1-9be2-4c8935bf6fee"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "i"},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`ortho}, $CellContext`i],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "i"},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`ortho}, $CellContext`i],
    Editable->False], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Cos",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cos,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "i"},
           {
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`ortho}, $CellContext`i],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Sin",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Sin,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}]}]}]], "Output",
 CellChangeTimes->{{3.791388033877593*^9, 3.791388061867621*^9}, 
   3.791388297743559*^9, 3.7913896470382767`*^9, 3.791389729558629*^9, 
   3.791389793117742*^9, 3.7945179523669786`*^9, 3.794520780496619*^9},
 CellLabel->"Out[307]=",
 CellID->642820857,ExpressionUUID->"fdd7d099-b8c0-4b9b-8d93-b70acd7ed5b9"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"xDiv", "[", 
     RowBox[{"%", ",", "i"}], "]"}], "[", "]"}], "//", "ContractBasis"}], "//",
   "ToValues"}]], "Input",
 CellChangeTimes->{{3.7913878252740793`*^9, 3.791387853422729*^9}, {
  3.791388041265252*^9, 3.7913880645443563`*^9}},
 CellLabel->"In[308]:=",
 CellID->344828528,ExpressionUUID->"35d66f65-29ac-46b2-83df-4f3e7485bcb0"],

Cell[BoxData[
 RowBox[{
  RowBox[{"4", " ", 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]}], "+", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox["Cos",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cos,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Cot",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cot,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}], " ", 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[0., 0., 1.]], ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
        StyleBox[GridBox[{
           {"i", " ", " "},
           {" ", "i", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelPDorthoPDsph[$CellContext`i, -$CellContext`i, {
     1, -$CellContext`ortho}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[0., 0., 1.]], ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
        StyleBox[GridBox[{
           {"i", " ", " "},
           {" ", "i", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelPDorthoPDsph[$CellContext`i, -$CellContext`i, {
     2, -$CellContext`ortho}],
    Editable->False], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Cos",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cos,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Sin",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Sin,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[0., 0., 1.]], ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
        StyleBox[GridBox[{
           {"i", " ", " "},
           {" ", "i", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelPDorthoPDsph[$CellContext`i, -$CellContext`i, {
     3, -$CellContext`ortho}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Sin",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Sin,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}]}]}]], "Output",
 CellChangeTimes->{{3.7913878324440937`*^9, 3.791387855092494*^9}, {
   3.791388046270076*^9, 3.791388064935549*^9}, 3.791388298365264*^9, 
   3.7913896470919523`*^9, 3.7913897296235228`*^9, 3.7913897931755943`*^9, 
   3.794517952558423*^9, 3.7945207805537024`*^9},
 CellLabel->"Out[308]=",
 CellID->2047399189,ExpressionUUID->"04cb0dc3-19ad-44a0-9a86-1f25afffe773"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ToValues"}]], "Input",
 CellChangeTimes->{{3.7913880478429537`*^9, 3.791388049728936*^9}},
 CellLabel->"In[309]:=",
 CellID->474557054,ExpressionUUID->"ddddc421-40e2-4a7e-b085-1e84dd79e5f2"],

Cell[BoxData[
 RowBox[{
  RowBox[{"4", " ", 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]}], "+", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox["Cos",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cos,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Cot",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cot,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}], " ", 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[0., 0., 1.]], ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
        StyleBox[GridBox[{
           {"i", " ", " "},
           {" ", "i", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelPDorthoPDsph[$CellContext`i, -$CellContext`i, {
     1, -$CellContext`ortho}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[0., 0., 1.]], ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
        StyleBox[GridBox[{
           {"i", " ", " "},
           {" ", "i", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelPDorthoPDsph[$CellContext`i, -$CellContext`i, {
     2, -$CellContext`ortho}],
    Editable->False], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Cos",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cos,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Sin",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Sin,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[0., 0., 1.]], ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
        StyleBox[GridBox[{
           {"i", " ", " "},
           {" ", "i", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "Rows" -> {{
               Offset[0.]}}}],
         FontSize->9]}
      },
      GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelPDorthoPDsph[$CellContext`i, -$CellContext`i, {
     3, -$CellContext`ortho}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Sin",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Sin,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}]}]}]], "Output",
 CellChangeTimes->{3.791388050049494*^9, 3.7913882988646317`*^9, 
  3.791389647142715*^9, 3.7913897296709633`*^9, 3.7913897932274933`*^9, 
  3.7945179527678843`*^9, 3.7945207806009283`*^9},
 CellLabel->"Out[309]=",
 CellID->1507693642,ExpressionUUID->"f0ec0301-19ff-4545-870a-390bff39195c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
There we see the Christoffel terms relating our two parallel derivatives. \
Currently we can evaluate them doing something like this [Jose: we should \
have an easier way to do this, perhaps via ToValues]:\
\>", "MathCaption",
 CellChangeTimes->{{3.791388371262669*^9, 3.791388445723987*^9}},
 CellID->1144158072,ExpressionUUID->"931179a4-0175-4e14-a857-76cde9f72101"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "/.", 
   RowBox[{"ChristoffelPDorthoPDsph", "\[Rule]", 
    RowBox[{"ToCTensor", "[", 
     RowBox[{"ChristoffelPDorthoPDsph", ",", 
      RowBox[{"{", 
       RowBox[{"ortho", ",", 
        RowBox[{"-", "ortho"}], ",", 
        RowBox[{"-", "ortho"}]}], "}"}]}], "]"}]}]}], "//", 
  "Simplify"}]], "Input",
 CellChangeTimes->{{3.791388259549371*^9, 3.791388288002022*^9}, {
  3.791388465547749*^9, 3.791388466819648*^9}},
 CellLabel->"In[310]:=",
 CellID->832807879,ExpressionUUID->"8ed4657e-bc5b-4e64-a84f-90bad2e79c1c"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{"4", "+", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["Cos",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Cos,
       Editable->False], "[", 
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False]}], "]"}], " ", 
     RowBox[{
      InterpretationBox[
       StyleBox["Csc",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Csc,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}]}]}], ")"}], " ", 
  InterpretationBox[
   StyleBox["r",
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`r[],
   Editable->False]}]], "Output",
 CellChangeTimes->{{3.791388292540847*^9, 3.791388299446712*^9}, 
   3.791388467228777*^9, 3.791389647186792*^9, 3.7913897297215776`*^9, 
   3.791389793279005*^9, 3.794517953082806*^9, 3.794520780638546*^9},
 CellLabel->"Out[310]=",
 CellID->1647490141,ExpressionUUID->"2bc55ad7-4073-4dc7-ad8e-9ee546ef0c02"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Yet another possibility is to act on general tensor fields. For example take \
a scalar field F[]:\
\>", "MathCaption",
 CellChangeTimes->{{3.7913884870027103`*^9, 3.7913885174740047`*^9}, {
  3.791388589950096*^9, 3.791388590230603*^9}},
 CellID->699828843,ExpressionUUID->"ab262ee2-26eb-4370-8ffb-a89ec321d920"],

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"F", "[", "]"}], ",", "E3"}], "]"}]], "Input",
 CellChangeTimes->{{3.791388518516498*^9, 3.7913885221546926`*^9}, {
  3.791388592821857*^9, 3.7913885929577017`*^9}},
 CellLabel->"In[311]:=",
 CellID->379832614,ExpressionUUID->"014e6b74-ce7c-45c9-a9c0-1a38a1df7f2e"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
This gives the derivative of the scalar field in terms of PDsph:\
\>", "MathCaption",
 CellChangeTimes->{{3.791388551552145*^9, 3.7913885748399363`*^9}},
 CellID->1890230017,ExpressionUUID->"9800838a-5e9a-48bc-94bb-8613e0b9442d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"xGrad", "[", 
   RowBox[{"F", "[", "]"}], "]"}], "[", "i", "]"}]], "Input",
 CellChangeTimes->{{3.7913885255000153`*^9, 3.79138853233622*^9}, {
  3.791388598303393*^9, 3.791388598478757*^9}},
 CellLabel->"In[312]:=",
 CellID->2109643964,ExpressionUUID->"883b194d-4ab5-474b-992d-b5c9543267e2"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            InterpretationBox[
             StyleBox[
              RowBox[{
               SubscriptBox[
                StyleBox["\[ScriptCapitalD]",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]], "F"}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`PDsph[{1, -$CellContext`sph}][
              $CellContext`F[]],
             Editable->False],
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", 
              InterpretationBox[
               StyleBox[
                RowBox[{
                  SubscriptBox[
                  "\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[1.\
, 0., 0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0., 0.]]\)"], 
                  "F"}], AutoSpacing -> False, ShowAutoStyles -> False], 
               $CellContext`PDsph[{1, -$CellContext`sph}][
                $CellContext`F[]], Editable -> False]}]],
           Annotation[#, {1} -> $CellContext`PDsph[{1, -$CellContext`sph}][
              $CellContext`F[]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            FractionBox[
             InterpretationBox[
              StyleBox[
               RowBox[{
                SubscriptBox[
                 StyleBox["\[ScriptCapitalD]",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]], "F"}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`PDsph[{2, -$CellContext`sph}][
               $CellContext`F[]],
              Editable->False], 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              FractionBox[
               InterpretationBox[
                StyleBox[
                 RowBox[{
                   SubscriptBox[
                   "\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[\
1., 0., 0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0., 0.]]\)"], 
                   "F"}], AutoSpacing -> False, ShowAutoStyles -> False], 
                $CellContext`PDsph[{2, -$CellContext`sph}][
                 $CellContext`F[]], Editable -> False], 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`r[], Editable -> False], "2"]]}]],
           
           Annotation[#, {
             2} -> $CellContext`r[]^(-2) $CellContext`PDsph[{
               2, -$CellContext`sph}][
               $CellContext`F[]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["Csc",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 Csc,
                 Editable->False], "[", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False], "]"}], "2"], " ", 
              RowBox[{"(", 
               InterpretationBox[
                StyleBox[
                 RowBox[{
                  SubscriptBox[
                   StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["3",
                    FontColor->RGBColor[1., 0., 0.]]], "F"}],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`PDsph[{3, -$CellContext`sph}][
                 $CellContext`F[]],
                Editable->False], ")"}]}], 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]],
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Csc"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Csc, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], "2"], 
                 " ", 
                 RowBox[{"(", 
                   InterpretationBox[
                    StyleBox[
                    RowBox[{
                    SubscriptBox[
                    "\!\(\*StyleBox[\"\[ScriptCapitalD]\",FontColor->RGBColor[\
1., 0., 0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0., 0.]]\)"], 
                    "F"}], AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`PDsph[{3, -$CellContext`sph}][
                    $CellContext`F[]], Editable -> False], ")"}]}], 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`r[], Editable -> False], "2"]]}]],
           Annotation[#, {3} -> Csc[
                $CellContext`\[Theta][]]^2 $CellContext`r[]^(-2) \
$CellContext`PDsph[{3, -$CellContext`sph}][
               $CellContext`F[]], "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"i"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{
    $CellContext`PDsph[{1, -$CellContext`sph}][
     $CellContext`F[]], $CellContext`r[]^(-2) $CellContext`PDsph[{
      2, -$CellContext`sph}][
      $CellContext`F[]], Csc[
       $CellContext`\[Theta][]]^2 $CellContext`r[]^(-2) $CellContext`PDsph[{
      3, -$CellContext`sph}][
      $CellContext`F[]]}, {$CellContext`sph}, 0][$CellContext`i],
  Editable->False]], "Output",
 CellChangeTimes->{3.79138853269479*^9, 3.791388598663389*^9, 
  3.791389647279838*^9, 3.791389729816532*^9, 3.7913897933827*^9, 
  3.794517953158101*^9, 3.7945207807060423`*^9},
 CellLabel->"Out[312]=",
 CellID->1860139016,ExpressionUUID->"0be61e83-8f01-4320-aeae-3bc501415e82"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Compare with an expression in terms of a scalar function. Note how F[] and \
the coordinates r[], \[Phi][] and \[Phi][] are scalar fields on E3, while f \
is a scalar function of 3 scalar arguments:\
\>", "MathCaption",
 CellChangeTimes->{{3.7913886082002697`*^9, 3.791388619126567*^9}, {
  3.791388713092705*^9, 3.791388765161974*^9}},
 CellID->457891937,ExpressionUUID->"7fc90934-e112-4664-b74a-637498525287"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", 
  RowBox[{
   RowBox[{"F", "[", "]"}], "\[Rule]", 
   RowBox[{"f", "[", 
    RowBox[{
     RowBox[{"r", "[", "]"}], ",", 
     RowBox[{"\[Theta]", "[", "]"}], ",", 
     RowBox[{"\[Phi]", "[", "]"}]}], "]"}]}]}]], "Input",
 CellChangeTimes->{{3.791388620590653*^9, 3.791388629054552*^9}},
 CellLabel->"In[313]:=",
 CellID->2117997876,ExpressionUUID->"3fd95a17-be19-4e10-a212-82d0142d56b0"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["f",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`f,
               Editable->False], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"1", ",", "0", ",", "0"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["\[Phi]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Phi][],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", 
              RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"f"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], $CellContext`f, Editable -> False], 
                 TagBox[
                  RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                 MultilineFunction -> None], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`\[Theta][], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "\[Phi]", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`\[Phi][], Editable -> False]}], "]"}]}]],
           Annotation[#, {1} -> Derivative[1, 0, 0][$CellContext`f][
              $CellContext`r[], 
              $CellContext`\[Theta][], 
              $CellContext`\[Phi][]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["f",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`f,
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"0", ",", "1", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["\[Phi]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Phi][],
                Editable->False]}], "]"}], 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"f"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`f, Editable -> False], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                  MultilineFunction -> None], "[", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], ",", 
                   InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], ",", 
                   InterpretationBox[
                    StyleBox[
                    "\[Phi]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Phi][], Editable -> False]}], "]"}], 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`r[], Editable -> False], "2"]]}]],
           
           Annotation[#, {2} -> $CellContext`r[]^(-2) 
             Derivative[0, 1, 0][$CellContext`f][
               $CellContext`r[], 
               $CellContext`\[Theta][], 
               $CellContext`\[Phi][]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["Csc",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 Csc,
                 Editable->False], "[", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False], "]"}], "2"], " ", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["f",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`f,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"0", ",", "0", ",", "1"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["\[Phi]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Phi][],
                 Editable->False]}], "]"}]}], 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]],
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Csc"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Csc, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], "2"], 
                 " ", 
                 RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"f"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`f, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "\[Phi]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Phi][], Editable -> False]}], "]"}]}], 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`r[], Editable -> False], "2"]]}]],
           Annotation[#, {3} -> Csc[
                $CellContext`\[Theta][]]^2 $CellContext`r[]^(-2) 
             Derivative[0, 0, 1][$CellContext`f][
               $CellContext`r[], 
               $CellContext`\[Theta][], 
               $CellContext`\[Phi][]], "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"i"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{
    Derivative[1, 0, 0][$CellContext`f][
     $CellContext`r[], 
     $CellContext`\[Theta][], 
     $CellContext`\[Phi][]], $CellContext`r[]^(-2) 
    Derivative[0, 1, 0][$CellContext`f][
      $CellContext`r[], 
      $CellContext`\[Theta][], 
      $CellContext`\[Phi][]], Csc[
       $CellContext`\[Theta][]]^2 $CellContext`r[]^(-2) 
    Derivative[0, 0, 1][$CellContext`f][
      $CellContext`r[], 
      $CellContext`\[Theta][], 
      $CellContext`\[Phi][]]}, {$CellContext`sph}, 0][$CellContext`i],
  Editable->False]], "Output",
 CellChangeTimes->{3.791388630694367*^9, 3.79138964732275*^9, 
  3.7913897298666983`*^9, 3.791389793433737*^9, 3.794517953208159*^9, 
  3.7945207808134537`*^9},
 CellLabel->"Out[313]=",
 CellID->289508886,ExpressionUUID->"487d157d-40ae-4ee8-96b6-73ff26788fd7"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.4.8. Clean up", "Subsubsection",
 CellChangeTimes->{{3.728676331871439*^9, 3.728676332031807*^9}, 
   3.790876746494073*^9, {3.7908769064909887`*^9, 3.79087690825729*^9}, {
   3.7908769966951647`*^9, 3.790876998551077*^9}, 3.79138104324002*^9, 
   3.791382245394075*^9, 3.791388889677356*^9, 3.7913889328447943`*^9, 
   3.791388981450349*^9, {3.791389589534655*^9, 3.791389589918775*^9}},
 CellID->1440932172,ExpressionUUID->"b6f6d361-19ff-4dd8-80e7-03d4b0e2ed85"],

Cell[CellGroupData[{

Cell["Clean up:", "MathCaption",
 CellChangeTimes->{{3.790875769287215*^9, 3.790875772872446*^9}},
 CellID->1946333281,ExpressionUUID->"f39c67df-5e2c-4e69-a3c1-3da363e73119"],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{
  "v", ",", "met", ",", "epsilonmet", ",", "cd", ",", "f", ",", "grad", ",", 
   "xCurl", ",", "xDiv", ",", "xGrad", ",", "xLaplacian", ",", "gradDivi", 
   ",", "gradDivj", ",", "res", ",", "scalar"}], "]"}]], "Input",
 CellChangeTimes->{{3.790875776033627*^9, 3.790875786240136*^9}, {
  3.7908759244431562`*^9, 3.790875924683341*^9}, {3.7908766039731483`*^9, 
  3.790876620852048*^9}, {3.791381442509543*^9, 3.791381482802071*^9}, {
  3.791389615542356*^9, 3.7913896174539623`*^9}, {3.791389750690187*^9, 
  3.791389762635083*^9}, {3.791389817969069*^9, 3.7913898216163816`*^9}},
 CellLabel->"In[314]:=",
 CellID->1143769821,ExpressionUUID->"f6e48f61-1a48-410e-8cfa-d2769f35edce"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefTensor", "/@", 
  RowBox[{"{", 
   RowBox[{"Jacobianorthosph", ",", "Jacobiansph", ",", "F"}], 
   "}"}]}]], "Input",
 CellChangeTimes->{{3.790875845129425*^9, 3.790875849086762*^9}, {
  3.790875909996818*^9, 3.790875918355557*^9}, {3.7908766771474037`*^9, 
  3.790876677425819*^9}, {3.79138969817316*^9, 3.791389698476242*^9}},
 CellLabel->"In[315]:=",
 CellID->781951820,ExpressionUUID->"5a5c565c-f24a-4dbc-8e3e-da65f302f5dc"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"Null", ",", "Null", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.7908765900101967`*^9, 3.7908766677751303`*^9, 
  3.790876707663782*^9, 3.7913814600346193`*^9, 3.79138968238778*^9, 
  3.791389729960596*^9, 3.791389793532407*^9, 3.7945179532769012`*^9, 
  3.794520785572171*^9},
 CellLabel->"Out[315]=",
 CellID->181151293,ExpressionUUID->"05141944-9eab-4bce-bdc8-324a17686a2e"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"UndefBasis", "[", "ortho", "]"}]], "Input",
 CellChangeTimes->{{3.79087579111639*^9, 3.790875794111793*^9}},
 CellLabel->"In[316]:=",
 CellID->745731941,ExpressionUUID->"4538f408-1f7a-466c-bde6-9db38392d719"],

Cell[BoxData[
 RowBox[{"UndefChart", "[", "sph", "]"}]], "Input",
 CellChangeTimes->{{3.790875858066002*^9, 3.790875862253334*^9}},
 CellLabel->"In[317]:=",
 CellID->765596817,ExpressionUUID->"9b0a6170-8ba0-4499-bec7-d3a1a544c6b5"],

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "E3", "]"}]], "Input",
 CellChangeTimes->{{3.790875893344412*^9, 3.7908758975329657`*^9}},
 CellLabel->"In[318]:=",
 CellID->1101425542,ExpressionUUID->"b6a1fd3c-4753-4546-b6e6-efccef3120ce"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.5. Curvature computations", "Subsection",
 CellChangeTimes->{{3.728674034137783*^9, 3.728674034322071*^9}, {
   3.728675277568149*^9, 3.728675277720169*^9}, 3.790876737753367*^9},
 CellID->1930012216,ExpressionUUID->"b5b1cc60-2f5e-4b76-b32d-092d58898b8a"],

Cell["\<\
This is probably the most important computation in GR, and we need to do it \
in a simple and fast enough way. Each example in this section is prepared to \
be evaluated independently of the rest of the notebook, and hence there are \
respective Needs commands.\
\>", "Text",
 CellID->2122310024,ExpressionUUID->"727cb295-9f98-4c56-8067-eee3787c03e4"],

Cell[CellGroupData[{

Cell["\<\
To have info messages from the caching system, activate this cell:\
\>", "Text",
 CellID->2108680345,ExpressionUUID->"a3efdba3-5116-4530-bd04-7f0ba8628f0d"],

Cell[BoxData[
 RowBox[{
  RowBox[{"$xCobaCacheVerbose", "=", "True"}], ";"}]], "Input",
 Evaluatable->False,
 CellChangeTimes->{{3.728843637541938*^9, 3.728843638564739*^9}},
 CellID->667931974,ExpressionUUID->"24a693e7-b373-4d6e-b250-f5de3f9c21e7"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
To have info messages from ComponentValue, activate this cell:\
\>", "Text",
 CellChangeTimes->{{3.731118219531498*^9, 3.731118222698004*^9}},
 CellID->985426970,ExpressionUUID->"3a5f7de8-aef3-4036-b978-11940f5ad11b"],

Cell[BoxData[
 RowBox[{
  RowBox[{"$CVVerbose", "=", "True"}], ";"}]], "Input",
 Evaluatable->False,
 CellChangeTimes->{{3.728843637541938*^9, 3.728843638564739*^9}, {
  3.731118226465748*^9, 3.7311182266821136`*^9}},
 CellID->677199126,ExpressionUUID->"cc65d2e4-bb59-4bb6-9858-3f3d687beb92"]
}, Open  ]],

Cell[CellGroupData[{

Cell["To see messages from MetricCompute activate this cell:", "Text",
 CellID->1035664149,ExpressionUUID->"94dcf799-c8ae-4b70-99e9-8b18740130da"],

Cell[BoxData[
 RowBox[{"SetOptions", "[", 
  RowBox[{"MetricCompute", ",", 
   RowBox[{"Verbose", "\[Rule]", "True"}]}], "]"}]], "Input",
 Evaluatable->False,
 CellID->1052393308,ExpressionUUID->"2c557be7-e801-4864-a6e0-ba379c140888"]
}, Open  ]],

Cell[CellGroupData[{

Cell["3.5.1. Kerr-Newman in the rotating frame", "Subsubsection",
 CellChangeTimes->{{3.728676331871439*^9, 3.728676332031807*^9}, 
   3.790876746494073*^9},
 CellID->899010864,ExpressionUUID->"7634dbcd-b0e0-4524-8be6-4034d738f711"],

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<xAct`xCoba`\>\"", "]"}]], "Input",
 CellLabel->"In[319]:=",
 CellID->2128134603,ExpressionUUID->"b94284bc-43fd-40ff-8db9-19980a3e85d1"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"$DefInfoQ", "=", 
   RowBox[{"$UndefInfoQ", "=", "False"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"$CVVerbose", "=", "False"}], ";"}]}], "Input",
 CellChangeTimes->{{3.7311179923634243`*^9, 3.731118003648952*^9}},
 CellLabel->"In[320]:=",
 CellID->1218131836,ExpressionUUID->"8558780a-d670-4277-adff-d566c1f941c6"],

Cell[CellGroupData[{

Cell["Define a 4d manifold:", "MathCaption",
 CellID->1180160801,ExpressionUUID->"9188e563-835e-4554-9a4b-1978a2be6e79"],

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M", ",", "4", ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f"}], "}"}]}], 
  "]"}]], "Input",
 CellLabel->"In[322]:=",
 CellID->803618715,ExpressionUUID->"40db2ab2-e942-4ed7-bb3d-2e58ae783d7e"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Boyer-Lindquist coordinates:", "MathCaption",
 CellID->895087022,ExpressionUUID->"86766bd2-e690-47e6-888e-a512e4609436"],

Cell[BoxData[
 RowBox[{"DefChart", "[", 
  RowBox[{"BL", ",", "M", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2", ",", "3"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"t", "[", "]"}], ",", 
     RowBox[{"r", "[", "]"}], ",", 
     RowBox[{"\[Theta]", "[", "]"}], ",", 
     RowBox[{"\[Phi]", "[", "]"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[323]:=",
 CellID->604631519,ExpressionUUID->"c7efe24b-10de-46f2-8d47-a655c9be2d6d"],

Cell[BoxData[{
 RowBox[{"DefConstantSymbol", "[", 
  RowBox[{"Mass", ",", 
   RowBox[{"PrintAs", "\[Rule]", "\"\<M\>\""}]}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"DefConstantSymbol", "[", 
  RowBox[{"Rotation", ",", 
   RowBox[{"PrintAs", "\[Rule]", "\"\<a\>\""}]}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"DefConstantSymbol", "[", 
  RowBox[{"Charge", ",", 
   RowBox[{"PrintAs", "\[Rule]", "\"\<Q\>\""}]}], "]"}]}], "Input",
 CellLabel->"In[324]:=",
 CellID->149322229,ExpressionUUID->"4ee1c6bf-c88e-449a-8795-e5b4b80de214"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Change of basis:", "MathCaption",
 CellID->1397803107,ExpressionUUID->"a9c05320-0e15-4849-8f9a-04ea88d4ce34"],

Cell[BoxData[
 RowBox[{
  RowBox[{"\[Rho]2", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"r", "[", "]"}], "^", "2"}], "+", 
    RowBox[{
     RowBox[{"Rotation", "^", "2"}], 
     RowBox[{
      RowBox[{"Cos", "[", 
       RowBox[{"\[Theta]", "[", "]"}], "]"}], "^", "2"}]}]}]}], 
  ";"}]], "Input",
 CellLabel->"In[327]:=",
 CellID->1553204816,ExpressionUUID->"f533c991-bdca-4306-9cc7-a7d93e1d2f82"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"change", "=", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      FractionBox[
       RowBox[{
        RowBox[{
         RowBox[{"r", "[", "]"}], "^", "2"}], "+", 
        RowBox[{"Rotation", "^", "2"}]}], "\[Rho]2"], ",", "0", ",", "0", ",", 
      FractionBox["Rotation", "\[Rho]2"]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{"0", ",", "1", ",", "0", ",", "0"}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{"0", ",", "0", ",", "1", ",", "0"}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"Rotation", " ", 
       RowBox[{
        RowBox[{"Sin", "[", 
         RowBox[{"\[Theta]", "[", "]"}], "]"}], "^", "2"}]}], ",", "0", ",", 
      "0", ",", "1"}], "}"}]}], "}"}]}]], "Input",
 CellLabel->"In[328]:=",
 CellID->2049613227,ExpressionUUID->"d408439f-f35c-49f9-b66b-7ad4a2ffe4fa"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], "2"], "+", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"]}], 
      RowBox[{
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], "2"], " ", 
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Cos",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Cos,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"]}], "+", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"]}]], ",", "0", ",", "0", ",", 
     FractionBox[
      InterpretationBox[
       StyleBox["a",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Rotation,
       Editable->False], 
      RowBox[{
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], "2"], " ", 
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Cos",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Cos,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"]}], "+", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"]}]]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "1", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["a",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Rotation,
       Editable->False], " ", 
      SuperscriptBox[
       RowBox[{
        InterpretationBox[
         StyleBox["Sin",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Sin,
         Editable->False], "[", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False], "]"}], "2"]}], ",", "0", ",", "0", ",", "1"}], 
    "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728676390065113*^9, 3.728753569859839*^9, 
  3.728843649856629*^9, 3.730956280966756*^9, 3.731117972140011*^9, 
  3.731118045115099*^9, 3.733368152535852*^9, 3.7945179539644003`*^9},
 CellLabel->"Out[328]=",
 CellID->624138270,ExpressionUUID->"5359e9bb-242b-4149-8278-647ecc4c5cb3"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Now define a frame in terms of the BL coordinates. Note that we need to \
provide the frame in a very specific way:\
\>", "MathCaption",
 CellID->920319793,ExpressionUUID->"6a666914-ba0b-47eb-ad52-575d5c224e6d"],

Cell[BoxData[
 RowBox[{"DefBasis", "[", 
  RowBox[{"frame", ",", "TangentM", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2", ",", "3"}], "}"}], ",", 
   RowBox[{"BasisColor", "\[Rule]", 
    RowBox[{"RGBColor", "[", 
     RowBox[{"0", ",", "0", ",", "1"}], "]"}]}], ",", 
   RowBox[{"BasisChange", "\[Rule]", 
    RowBox[{"CTensor", "[", 
     RowBox[{"change", ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "frame"}], ",", "BL"}], "}"}]}], "]"}]}]}], 
  "]"}]], "Input",
 CellLabel->"In[329]:=",
 CellID->1637535257,ExpressionUUID->"06123832-3ea0-4154-ab10-473165123e31"]
}, Open  ]],

Cell[CellGroupData[{

Cell["This explains the change we have provided:", "MathCaption",
 CellID->2052128094,ExpressionUUID->"1702e0ea-2c80-4c27-b274-dc076bd78b4d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCTensor", "[", 
  RowBox[{"Basis", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "frame"}], ",", "BL"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[330]:=",
 CellID->1529505038,ExpressionUUID->"c64995bc-d0d4-40c9-a992-ed36cc8d994a"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       FractionBox[
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], "+", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"]}], 
        RowBox[{
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "2"]}], "+", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"]}]], ",", "0", ",", "0", ",", 
       FractionBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], 
        RowBox[{
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "2"]}], "+", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"]}]]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "1", ",", "0", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", "1", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], " ", 
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Sin",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Sin,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"]}], ",", "0", ",", "0", ",", "1"}], 
      "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "frame"}], ",", "BL"}], "}"}], ",", "0"}], "]"}]], "Output",\

 CellChangeTimes->{3.728676390342963*^9, 3.728753570112019*^9, 
  3.728843652114235*^9, 3.730956281218519*^9, 3.7311180513983603`*^9, 
  3.733368160219096*^9, 3.7945179542153683`*^9},
 CellLabel->"Out[330]=",
 CellID->1581771942,ExpressionUUID->"5b372aeb-c6eb-40ef-8136-ec36ad371c0b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToCTensor", "[", 
   RowBox[{"Basis", ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "BL"}], ",", "frame"}], "}"}]}], "]"}], "//", 
  "Simplify"}]], "Input",
 CellLabel->"In[331]:=",
 CellID->1400909739,ExpressionUUID->"0fac8e23-5b34-47c3-826c-603dd175fa61"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "0", ",", "0", ",", 
       RowBox[{"-", 
        FractionBox[
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False]}], 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False]}], "]"}]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}]}]]}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "1", ",", "0", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", "1", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False]}], " ", 
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Sin",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Sin,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"]}], ",", "0", ",", "0", ",", 
       FractionBox[
        RowBox[{"2", " ", 
         RowBox[{"(", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], "+", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}], ")"}]}], 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], "+", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], " ", 
          RowBox[{
           InterpretationBox[
            StyleBox["Cos",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Cos,
            Editable->False], "[", 
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False]}], "]"}]}], "+", 
         RowBox[{"2", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]}]}]]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "BL"}], ",", "frame"}], "}"}], ",", "0"}], "]"}]], "Output",\

 CellChangeTimes->{3.7286763904903812`*^9, 3.728753570242834*^9, 
  3.728843652901079*^9, 3.730956281316536*^9, 3.731118052328496*^9, 
  3.7333681611855392`*^9, 3.7945179542634363`*^9},
 CellLabel->"Out[331]=",
 CellID->1140962981,ExpressionUUID->"a8d15157-8b0a-46cc-9b33-f3ca3bb4a50e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCTensor", "[", 
  RowBox[{
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"vt", ",", "vr", ",", "v\[Theta]", ",", "v\[Phi]"}], "}"}], ",", 
     RowBox[{"{", "BL", "}"}]}], "]"}], ",", 
   RowBox[{"{", "frame", "}"}]}], "]"}]], "Input",
 CellLabel->"In[332]:=",
 CellID->979854515,ExpressionUUID->"5d7fb1a6-5f20-4cdb-836f-a82c6f3ac1d6"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"vt", "-", 
      RowBox[{
       InterpretationBox[
        StyleBox["a",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Rotation,
        Editable->False], " ", "v\[Phi]", " ", 
       SuperscriptBox[
        RowBox[{
         InterpretationBox[
          StyleBox["Sin",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Sin,
          Editable->False], "[", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False], "]"}], "2"]}]}], ",", "vr", ",", "v\[Theta]", ",", 
     FractionBox[
      RowBox[{"2", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", 
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False]}], " ", "vt"}], "+", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], " ", "v\[Phi]"}], "+", 
         RowBox[{"v\[Phi]", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]}]}], ")"}]}], 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], "2"], "+", 
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], "2"], " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["Cos",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Cos,
          Editable->False], "[", 
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False]}], "]"}]}], "+", 
       RowBox[{"2", " ", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]}]}]]}], "}"}], ",", 
   RowBox[{"{", "frame", "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.728676390640329*^9, 3.7287535703598423`*^9, 
  3.728843653685162*^9, 3.73095628142439*^9, 3.731118052984367*^9, 
  3.7333681620033703`*^9, 3.7945179543216867`*^9},
 CellLabel->"Out[332]=",
 CellID->1283337604,ExpressionUUID->"6321522a-091f-4d8c-9a04-97fb602bb294"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{"vt", ",", "vr", ",", "v\[Theta]", ",", "v\[Phi]"}], "]"}]], "Input",\

 CellLabel->"In[333]:=",
 CellID->1299646170,ExpressionUUID->"d7336ef0-98c6-45dd-ae9c-48284018320f"]
}, Open  ]],

Cell[CellGroupData[{

Cell["The metric tensor is diagonal in this frame:", "MathCaption",
 CellID->1097005553,ExpressionUUID->"e7adb246-3135-4f7b-95f5-eceeeaca9295"],

Cell[BoxData[
 RowBox[{
  RowBox[{"factor", "=", 
   RowBox[{
    FractionBox[
     RowBox[{
      RowBox[{"r", "[", "]"}], "^", "2"}], "\[Rho]2"], 
    RowBox[{"(", 
     RowBox[{"1", "-", 
      FractionBox[
       RowBox[{"2", "Mass"}], 
       RowBox[{"r", "[", "]"}]], "+", 
      FractionBox[
       RowBox[{
        RowBox[{"Rotation", "^", "2"}], "+", 
        RowBox[{"Charge", "^", "2"}]}], 
       RowBox[{
        RowBox[{"r", "[", "]"}], "^", "2"}]]}], ")"}]}]}], ";"}]], "Input",
 CellLabel->"In[334]:=",
 CellID->1225219776,ExpressionUUID->"8ccd7f82-bf2f-40a8-9380-5c6055b218ed"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"metric", "=", 
  RowBox[{"CTensor", "[", 
   RowBox[{
    RowBox[{"DiagonalMatrix", "[", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "factor"}], ",", 
       RowBox[{"1", "/", "factor"}], ",", "\[Rho]2", ",", 
       RowBox[{"\[Rho]2", " ", 
        RowBox[{
         RowBox[{"Sin", "[", 
          RowBox[{"\[Theta]", "[", "]"}], "]"}], "^", "2"}]}]}], "}"}], "]"}],
     ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "frame"}], ",", 
      RowBox[{"-", "frame"}]}], "}"}], ",", "0"}], "]"}]}]], "Input",
 CellLabel->"In[335]:=",
 CellID->48084149,ExpressionUUID->"0a72d98e-bcdb-4da8-8451-1bbe859e2609"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", 
        FractionBox[
         RowBox[{
          RowBox[{"(", 
           RowBox[{"1", "+", 
            FractionBox[
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["Q",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Charge,
                Editable->False], "2"], "+", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["a",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Rotation,
                Editable->False], "2"]}], 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]], "-", 
            FractionBox[
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False]}], 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]]}], ")"}], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]}], 
         RowBox[{
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["Cos",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Cos,
              Editable->False], "[", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False], "]"}], "2"]}], "+", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]}]]}], ",", "0", ",", "0", ",", "0"}], 
      "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", 
       FractionBox[
        RowBox[{
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "2"]}], "+", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"]}], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"1", "+", 
           FractionBox[
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["Q",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Charge,
               Editable->False], "2"], "+", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"]}], 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"]], "-", 
           FractionBox[
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["M",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Mass,
              Editable->False]}], 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]]}], ")"}], " ", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"]}]], ",", "0", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", 
       RowBox[{
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Cos",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Cos,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}], "+", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]}], ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", "0", ",", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["Cos",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Cos,
              Editable->False], "[", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False], "]"}], "2"]}], "+", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]}], ")"}], " ", 
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Sin",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Sin,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"]}]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "frame"}], ",", 
     RowBox[{"-", "frame"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.7286763909144506`*^9, 3.728753570627632*^9, 
  3.728843656215597*^9, 3.730956281616479*^9, 3.7311180562945633`*^9, 
  3.733368164721225*^9, 3.794517954405154*^9},
 CellLabel->"Out[335]=",
 CellID->1876843607,ExpressionUUID->"3f4bea74-3b99-4c30-b951-fc6c08561543"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The coordinated metric tensor is nondiagonal:", "MathCaption",
 CellID->485428197,ExpressionUUID->"0297881f-8a84-4855-9dd0-d1fa2e3269dd"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCTensor", "[", 
  RowBox[{"metric", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "BL"}], ",", 
     RowBox[{"-", "BL"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[336]:=",
 CellID->301332026,ExpressionUUID->"2a2065f9-91df-4524-8380-c2903a15cded"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", 
        FractionBox[
         RowBox[{"2", " ", 
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["Q",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Charge,
              Editable->False], "2"], "+", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], "-", 
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["M",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Mass,
              Editable->False], " ", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]}], "+", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"], "-", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"], " ", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], "2"]}]}], ")"}]}], 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False]}], "]"}]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}]}]]}], ",", "0", ",", "0", ",", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["Q",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Charge,
             Editable->False], "2"], "-", 
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Mass,
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]}]}], ")"}], " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Sin",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Sin,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}], 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], "+", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], " ", 
          RowBox[{
           InterpretationBox[
            StyleBox["Cos",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Cos,
            Editable->False], "[", 
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False]}], "]"}]}], "+", 
         RowBox[{"2", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]}]}]]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", 
       FractionBox[
        RowBox[{
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "2"]}], "+", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"]}], 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["Q",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Charge,
           Editable->False], "2"], "+", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], "-", 
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Mass,
           Editable->False], " ", 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False]}], "+", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"]}]], ",", "0", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", 
       RowBox[{
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Cos",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Cos,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}], "+", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]}], ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["Q",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Charge,
             Editable->False], "2"], "-", 
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Mass,
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]}]}], ")"}], " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Sin",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Sin,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}], 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], "+", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], " ", 
          RowBox[{
           InterpretationBox[
            StyleBox["Cos",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Cos,
            Editable->False], "[", 
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False]}], "]"}]}], "+", 
         RowBox[{"2", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]}]}]], ",", "0", ",", "0", ",", 
       RowBox[{
        FractionBox[
         RowBox[{"2", " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"], "+", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}], ")"}], "2"], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Sin",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Sin,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "2"]}], 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False]}], "]"}]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}]}]], "-", 
        FractionBox[
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], " ", 
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["Q",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Charge,
              Editable->False], "2"], "+", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], "+", 
            RowBox[{
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"-", "2"}], " ", 
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False]}], "+", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}], ")"}]}]}], ")"}], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Sin",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Sin,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "4"]}], 
         RowBox[{
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["Cos",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Cos,
              Editable->False], "[", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False], "]"}], "2"]}], "+", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]}]]}]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "BL"}], ",", 
     RowBox[{"-", "BL"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.7286763924095907`*^9, 3.728753571529663*^9, 
  3.728843660755485*^9, 3.7309562825758657`*^9, 3.7311180665792227`*^9, 
  3.7333681682054768`*^9, 3.794517954604175*^9},
 CellLabel->"Out[336]=",
 CellID->1870695713,ExpressionUUID->"4bc6fb78-8be9-4107-8f55-0d2f9f4436ea"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Compute curvature tensors:", "MathCaption",
 CellID->422493918,ExpressionUUID->"02838182-e322-4598-a333-70aa676cd105"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"MetricCompute", "[", 
   RowBox[{"metric", ",", "BL", ",", "All", ",", 
    RowBox[{"Parallelize", "\[Rule]", "True"}]}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[337]:=",
 CellID->1274342844,ExpressionUUID->"43968ed5-a8b3-46d7-9c69-f4c3228e5040"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"5.508279`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.728676399957385*^9, 3.728753578676353*^9, 
  3.728843669563285*^9, 3.7309562896162357`*^9, 3.7311180812672462`*^9, 
  3.733368179091886*^9, 3.794517960151216*^9},
 CellLabel->"Out[337]=",
 CellID->809436665,ExpressionUUID->"b6675c96-fe31-47bb-a9fe-aed6a1fb8011"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
That computed and simplified everything, including the Kretschmann scalar and \
the first covariant derivative of the Riemann tensor. All those results are \
now cached. If we run it again the computation is much faster:\
\>", "MathCaption",
 CellChangeTimes->{{3.731118090646637*^9, 3.73111812174946*^9}},
 CellID->76948523,ExpressionUUID->"78bd1f94-0fa3-488a-b61c-c2dc2b57a9dc"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"MetricCompute", "[", 
   RowBox[{"metric", ",", "BL", ",", "All"}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[338]:=",
 CellID->488670567,ExpressionUUID->"56557837-f24b-40dc-b46a-9c8adaf3a33b"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.034553`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.728676400239497*^9, 3.728753578963751*^9, 
  3.728843672218933*^9, 3.730956289999316*^9, 3.731118124770753*^9, 
  3.7333681834474697`*^9, 3.7945179603402967`*^9},
 CellLabel->"Out[338]=",
 CellID->2040608557,ExpressionUUID->"30477f7e-c077-4cae-b28c-ee891b3dd8d4"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Extract information:", "MathCaption",
 CellID->1103477868,ExpressionUUID->"7ce0fdab-7593-4c15-b365-169771325d38"],

Cell[BoxData[
 RowBox[{
  RowBox[{"cd", "=", 
   RowBox[{"CovDOfMetric", "[", "metric", "]"}]}], ";"}]], "Input",
 CellLabel->"In[339]:=",
 CellID->483751984,ExpressionUUID->"599400d9-4807-4db7-8216-8604ba0a629e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"RicciScalar", "[", "cd", "]"}], "[", "]"}]], "Input",
 CellLabel->"In[340]:=",
 CellID->46507546,ExpressionUUID->"22f358ee-7b98-443a-ad1e-4cf3ad95159c"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.7286764004769363`*^9, 3.728753579229369*^9, 
  3.7288436765381727`*^9, 3.730956290147964*^9, 3.731118132810968*^9, 
  3.733368185739315*^9, 3.794517960412219*^9},
 CellLabel->"Out[340]=",
 CellID->1194174909,ExpressionUUID->"e179ed83-1b1d-45c2-97c6-4555e61d4a6d"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The Ricci tensor is again diagonal in this frame:", "MathCaption",
 CellID->1573690101,ExpressionUUID->"d6570d23-67c2-438b-8a5d-0da1be68683d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Ricci", "[", "cd", "]"}]], "Input",
 CellLabel->"In[341]:=",
 CellID->600931055,ExpressionUUID->"23542a43-1848-4492-93a6-574608833559"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       FractionBox[
        RowBox[{"8", " ", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["Q",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Charge,
           Editable->False], "2"], " ", 
         RowBox[{"(", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["Q",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Charge,
             Editable->False], "2"], "+", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "2"}], " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Mass,
                Editable->False]}], "+", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}], ")"}]}]}], ")"}]}], 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], "+", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["Cos",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Cos,
              Editable->False], "[", 
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False]}], "]"}]}], "+", 
           RowBox[{"2", " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"]}]}], ")"}], "3"]], ",", "0", ",", "0", 
       ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", 
       RowBox[{"-", 
        FractionBox[
         RowBox[{"2", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["Q",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Charge,
            Editable->False], "2"]}], 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], "+", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Cos",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cos,
               Editable->False], "[", 
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False]}], "]"}]}], "+", 
            RowBox[{"2", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}]}], ")"}], " ", 
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["Q",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Charge,
              Editable->False], "2"], "+", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], "+", 
            RowBox[{
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"-", "2"}], " ", 
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False]}], "+", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}], ")"}]}]}], ")"}]}]]}], ",", "0", ",", 
       "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", 
       FractionBox[
        RowBox[{"2", " ", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["Q",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Charge,
           Editable->False], "2"]}], 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], "+", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], " ", 
          RowBox[{
           InterpretationBox[
            StyleBox["Cos",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Cos,
            Editable->False], "[", 
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False]}], "]"}]}], "+", 
         RowBox[{"2", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]}]}]], ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", "0", ",", 
       FractionBox[
        RowBox[{"2", " ", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["Q",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Charge,
           Editable->False], "2"], " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Sin",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Sin,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}], 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], "+", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], " ", 
          RowBox[{
           InterpretationBox[
            StyleBox["Cos",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Cos,
            Editable->False], "[", 
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False]}], "]"}]}], "+", 
         RowBox[{"2", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]}]}]]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "frame"}], ",", 
     RowBox[{"-", "frame"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.7286764006288548`*^9, 3.72875357939991*^9, 
  3.7288436777925987`*^9, 3.730956290257471*^9, 3.7311181363347282`*^9, 
  3.733368186922182*^9, 3.794517960463688*^9},
 CellLabel->"Out[341]=",
 CellID->115530701,ExpressionUUID->"8b14c1b5-73a7-4a06-84f9-d7ec4f0ea222"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}]}], "]"}]], "Input",
 CellLabel->"In[342]:=",
 CellID->1720648468,ExpressionUUID->"74512ce5-bdf7-4f88-ad73-d4524f0c5016"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{"8", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"Q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Charge, Editable -> False], "2"], 
                 " ", 
                 RowBox[{"(", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Charge, Editable -> False], "2"], 
                    "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"a"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Rotation, Editable -> False], "2"], 
                    "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}], 
                   ")"}]}], 
               SuperscriptBox[
                RowBox[{"(", 
                  RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"a"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Rotation, Editable -> False], "2"], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"a"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Rotation, Editable -> False], "2"], 
                    " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Cos"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cos, Editable -> False], "[", 
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False]}], "]"}]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]}], ")"}], 
                "3"]]}]],
           
           Annotation[#, {1, 1} -> 
            8 $CellContext`Charge^2 ($CellContext`Rotation^2 + \
$CellContext`Rotation^2 Cos[2 $CellContext`\[Theta][]] + 
               2 $CellContext`r[]^2)^(-3) ($CellContext`Charge^2 + \
$CellContext`Rotation^2 + $CellContext`r[] ((-2) $CellContext`Mass + \
$CellContext`r[])), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{"-", 
                FractionBox[
                 RowBox[{"2", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Charge, Editable -> False], "2"]}], 
                 RowBox[{
                   RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"a"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Rotation, Editable -> False], "2"], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"a"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Rotation, Editable -> False], "2"], 
                    " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Cos"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cos, Editable -> False], "[", 
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False]}], "]"}]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]}], ")"}], 
                   " ", 
                   RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Charge, Editable -> False], "2"], 
                    "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"a"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Rotation, Editable -> False], "2"], 
                    "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}], 
                    ")"}]}]]}]}]],
           
           Annotation[#, {2, 
             2} -> (-2) $CellContext`Charge^2 ($CellContext`Rotation^2 + \
$CellContext`Rotation^2 Cos[2 $CellContext`\[Theta][]] + 
               2 $CellContext`r[]^2)^(-1)/($CellContext`Charge^2 + \
$CellContext`Rotation^2 + $CellContext`r[] ((-2) $CellContext`Mass + \
$CellContext`r[])), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{"2", " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["Q",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Charge,
                Editable->False], "2"]}], 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["a",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Rotation,
                Editable->False], "2"], "+", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["Cos",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 Cos,
                 Editable->False], "[", 
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["\[Theta]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Theta][],
                  Editable->False]}], "]"}]}], "+", 
              RowBox[{"2", " ", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False], "2"]}]}]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{"2", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"Q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Charge, Editable -> False], "2"]}], 
               RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"a"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Rotation, Editable -> False], "2"], 
                 "+", 
                 RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"a"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Rotation, Editable -> False], "2"], 
                   " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Cos"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cos, Editable -> False], "[", 
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False]}], "]"}]}], 
                 "+", 
                 RowBox[{"2", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]}]]}]],
           
           Annotation[#, {3, 3} -> 
            2 $CellContext`Charge^2/($CellContext`Rotation^2 + \
$CellContext`Rotation^2 Cos[2 $CellContext`\[Theta][]] + 
             2 $CellContext`r[]^2), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "4"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{"2", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"Q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Charge, Editable -> False], "2"], 
                 " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                  "2"]}], 
               RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"a"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Rotation, Editable -> False], "2"], 
                 "+", 
                 RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"a"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Rotation, Editable -> False], "2"], 
                   " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Cos"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Cos, Editable -> False], "[", 
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False]}], "]"}]}], 
                 "+", 
                 RowBox[{"2", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]}]]}]],
           
           Annotation[#, {4, 4} -> 
            2 $CellContext`Charge^2 ($CellContext`Rotation^2 + \
$CellContext`Rotation^2 Cos[2 $CellContext`\[Theta][]] + 
               2 $CellContext`r[]^2)^(-1) Sin[
                $CellContext`\[Theta][]]^2, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]},
           "RowsIndexed" -> {1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " "},
         {"a", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{
    8 $CellContext`Charge^2 ($CellContext`Rotation^2 + \
$CellContext`Rotation^2 Cos[2 $CellContext`\[Theta][]] + 
       2 $CellContext`r[]^2)^(-3) ($CellContext`Charge^2 + \
$CellContext`Rotation^2 + $CellContext`r[] ((-2) $CellContext`Mass + \
$CellContext`r[])), 0, 0, 0}, {
    0, (-2) $CellContext`Charge^2 ($CellContext`Rotation^2 + \
$CellContext`Rotation^2 Cos[2 $CellContext`\[Theta][]] + 
       2 $CellContext`r[]^2)^(-1)/($CellContext`Charge^2 + \
$CellContext`Rotation^2 + $CellContext`r[] ((-2) $CellContext`Mass + \
$CellContext`r[])), 0, 0}, {
    0, 0, 2 $CellContext`Charge^2/($CellContext`Rotation^2 + \
$CellContext`Rotation^2 Cos[2 $CellContext`\[Theta][]] + 
     2 $CellContext`r[]^2), 0}, {
    0, 0, 0, 2 $CellContext`Charge^2 ($CellContext`Rotation^2 + \
$CellContext`Rotation^2 Cos[2 $CellContext`\[Theta][]] + 
       2 $CellContext`r[]^2)^(-1) Sin[
        $CellContext`\[Theta][]]^2}}, {-$CellContext`frame, \
-$CellContext`frame}, 0][-$CellContext`a, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.728676400801085*^9, 3.7287535795840387`*^9, 
  3.728843679706231*^9, 3.7309562903735857`*^9, 3.73111813721097*^9, 
  3.733368188423842*^9, 3.794517960521007*^9},
 CellLabel->"Out[342]=",
 CellID->399274819,ExpressionUUID->"3feb056c-cd4b-4bbc-babb-10e916723d45"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Kretschmann", "[", "cd", "]"}], "[", "]"}]], "Input",
 CellLabel->"In[343]:=",
 CellID->1401549961,ExpressionUUID->"762bb786-dcc9-4794-9f18-63447e71b3c3"],

Cell[BoxData[
 RowBox[{"-", 
  RowBox[{
   FractionBox["1", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], "2"], "+", 
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], "2"], " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["Cos",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Cos,
          Editable->False], "[", 
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False]}], "]"}]}], "+", 
       RowBox[{"2", " ", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]}]}], ")"}], "6"]], 
   RowBox[{"32", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "42"}], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["Q",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Charge,
         Editable->False], "4"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], "4"]}], "+", 
      RowBox[{"30", " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Mass,
         Editable->False], "2"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], "6"]}], "+", 
      RowBox[{"3", " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Mass,
         Editable->False], "2"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], "6"], " ", 
       RowBox[{
        InterpretationBox[
         StyleBox["Cos",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Cos,
         Editable->False], "[", 
        RowBox[{"6", " ", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False]}], "]"}]}], "+", 
      RowBox[{"360", " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["Q",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Charge,
         Editable->False], "2"], " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Mass,
        Editable->False], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], "4"], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], "+", 
      RowBox[{"272", " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["Q",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Charge,
         Editable->False], "4"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], "2"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"]}], "-", 
      RowBox[{"540", " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Mass,
         Editable->False], "2"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], "4"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"]}], "-", 
      RowBox[{"960", " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["Q",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Charge,
         Editable->False], "2"], " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Mass,
        Editable->False], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], "2"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "3"]}], "-", 
      RowBox[{"112", " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["Q",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Charge,
         Editable->False], "4"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "4"]}], "+", 
      RowBox[{"720", " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Mass,
         Editable->False], "2"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], "2"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "4"]}], "+", 
      RowBox[{"192", " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["Q",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Charge,
         Editable->False], "2"], " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Mass,
        Editable->False], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "5"]}], "-", 
      RowBox[{"96", " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Mass,
         Editable->False], "2"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "6"]}], "+", 
      RowBox[{"2", " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], "4"], " ", 
       RowBox[{
        InterpretationBox[
         StyleBox["Cos",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Cos,
         Editable->False], "[", 
        RowBox[{"4", " ", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False]}], "]"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "7"}], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["Q",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Charge,
            Editable->False], "4"]}], "+", 
         RowBox[{"60", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["Q",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Charge,
            Editable->False], "2"], " ", 
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Mass,
           Editable->False], " ", 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False]}], "+", 
         RowBox[{"9", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False], "2"], " ", 
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], "-", 
            RowBox[{"10", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}]}], ")"}]}]}], ")"}]}], "+", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], "2"], " ", 
       RowBox[{
        InterpretationBox[
         StyleBox["Cos",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Cos,
         Editable->False], "[", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False]}], "]"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "8"}], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["Q",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Charge,
            Editable->False], "4"], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"7", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"]}], "-", 
            RowBox[{"34", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}]}], ")"}]}], "+", 
         RowBox[{"480", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["Q",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Charge,
            Editable->False], "2"], " ", 
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Mass,
           Editable->False], " ", 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], " ", 
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], "-", 
            RowBox[{"2", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}]}], ")"}]}], "+", 
         RowBox[{"45", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False], "2"], " ", 
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "4"], "-", 
            RowBox[{"16", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"], " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}], "+", 
            RowBox[{"16", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "4"]}]}], ")"}]}]}], ")"}]}]}], 
     ")"}]}]}]}]], "Output",
 CellChangeTimes->{3.7286764009605627`*^9, 3.72875357974678*^9, 
  3.728843681135004*^9, 3.73095629048492*^9, 3.731118139508012*^9, 
  3.733368189686901*^9, 3.7945179605697317`*^9},
 CellLabel->"Out[343]=",
 CellID->1872969705,ExpressionUUID->"1166aa24-6cc5-4436-b35e-f267476d87c0"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Riemann", "[", "cd", "]"}]], "Input",
 CellLabel->"In[344]:=",
 CellID->1808354982,ExpressionUUID->"a0d17ee4-dde4-4aa9-819f-84ddc64ae39e"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", 
           RowBox[{"-", 
            FractionBox[
             RowBox[{"8", " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}]}]}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "6"}], " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], "+", 
                RowBox[{"4", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "3"]}], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Charge,
                    Editable->False], "2"], "-", 
                   RowBox[{"6", " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False], " ", 
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}]}], ")"}]}], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Charge,
                   Editable->False], "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                   RowBox[{"6", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}]}], 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]}], ")"}], "4"]]}], ",", "0", ",", 
           "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"-", 
            FractionBox[
             RowBox[{"2", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "6"}], " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], "+", 
                RowBox[{"4", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "3"]}], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Charge,
                    Editable->False], "2"], "-", 
                   RowBox[{"6", " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False], " ", 
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}]}], ")"}]}], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Charge,
                   Editable->False], "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                   RowBox[{"6", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}]}], 
             RowBox[{
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], "+", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    Cos,
                    Editable->False], "[", 
                   RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                 RowBox[{"2", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}], "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}]}]}], ")"}]}]]}], ",", "0", ",", 
           "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           RowBox[{"-", 
            FractionBox[
             RowBox[{"8", " ", 
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["Cot",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Cot,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"4", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Charge,
                   Editable->False], "2"], " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], "+", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                   RowBox[{"6", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}]}], 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]}], ")"}], "3"]]}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           FractionBox[
            RowBox[{"4", " ", 
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"4", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], " ", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}], "+", 
               RowBox[{
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                  RowBox[{"6", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Sin",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Sin,
               Editable->False], "[", 
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False]}], "]"}]}], 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"2", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}]}], ")"}], "3"]], ",", "0"}], 
          "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           FractionBox[
            RowBox[{
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["Q",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Charge,
                 Editable->False], "2"], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{"-", "2"}], " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"-", "3"}], " ", 
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}], "+", 
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "3"]}], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}], " ", 
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Charge,
                    Editable->False], "2"], "-", 
                  RowBox[{"3", " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False], " ", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}]}], ")"}]}], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], " ", 
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                  RowBox[{"2", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}]}], 
            RowBox[{"2", " ", 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    Cos,
                    Editable->False], "[", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False], "]"}], "2"]}], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}], ")"}], "4"]}]], ",", "0"}], "}"}],
          ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           RowBox[{"-", 
            FractionBox[
             RowBox[{"4", " ", 
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["Cot",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Cot,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"4", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Charge,
                   Editable->False], "2"], " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], "+", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                   RowBox[{"6", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}]}], 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]}], ")"}], "3"]]}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           FractionBox[
            RowBox[{
             RowBox[{
              RowBox[{"-", "3"}], " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["a",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Rotation,
                Editable->False], "2"], " ", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}], "+", 
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], "3"]}], "+", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["a",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Rotation,
                Editable->False], "2"], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["Cos",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Cos,
                Editable->False], "[", 
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False]}], "]"}], " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], "-", 
                RowBox[{"3", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}]}], ")"}]}], "+", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["Q",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Charge,
                Editable->False], "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "-", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]}], ")"}]}]}], 
            RowBox[{"2", " ", 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    Cos,
                    Editable->False], "[", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False], "]"}], "2"]}], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}], ")"}], "2"]}]], ",", "0", ",", 
           "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", 
           FractionBox[
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["Q",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Charge,
                 Editable->False], "2"], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{"-", "2"}], " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"4", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], " ", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}], "+", 
               RowBox[{
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                  RowBox[{"6", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Sin",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Sin,
               Editable->False], "[", 
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False]}], "]"}]}], 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"2", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}]}], ")"}], "3"]], ",", "0", ",", 
           "0"}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           FractionBox[
            RowBox[{"8", " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["Q",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Charge,
                 Editable->False], "2"], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{"-", "2"}], " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"-", "3"}], " ", 
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}], "+", 
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "3"]}], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}], " ", 
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Charge,
                    Editable->False], "2"], "-", 
                  RowBox[{"3", " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False], " ", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}]}], ")"}]}], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], " ", 
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                  RowBox[{"2", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}]}], 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"2", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}]}], ")"}], "4"]]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           FractionBox[
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"4", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], " ", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}], "+", 
               RowBox[{
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                  RowBox[{"6", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Sin",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Sin,
               Editable->False], "[", 
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False]}], "]"}]}], 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"2", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}]}], ")"}], "3"]], ",", "0"}], "}"}],
          ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", 
           RowBox[{"-", 
            FractionBox[
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}]}]}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"4", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Charge,
                   Editable->False], "2"], " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], "+", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                   RowBox[{"6", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False]}], "]"}]}], 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]}], ")"}], "3"]]}], ",", "0", ",", 
           "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           FractionBox[
            RowBox[{"2", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"-", "3"}], " ", 
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}], "+", 
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "3"]}], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}], " ", 
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Charge,
                    Editable->False], "2"], "-", 
                  RowBox[{"3", " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False], " ", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}]}], ")"}]}], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], " ", 
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                  RowBox[{"2", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], "2"]}], 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"2", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}]}], ")"}], "2"]], ",", "0", ",", 
           "0", ",", "0"}], "}"}]}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", 
           FractionBox[
            RowBox[{"8", " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["Q",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Charge,
                 Editable->False], "2"], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{"-", "2"}], " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"-", "6"}], " ", 
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}], "+", 
               RowBox[{"4", " ", 
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "3"]}], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}], " ", 
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Charge,
                    Editable->False], "2"], "-", 
                  RowBox[{"6", " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False], " ", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}]}], ")"}]}], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], " ", 
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                  RowBox[{"6", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}]}], 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"2", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}]}], ")"}], "4"]], ",", "0", ",", 
           "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           FractionBox[
            RowBox[{"2", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"-", "6"}], " ", 
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}], "+", 
               RowBox[{"4", " ", 
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "3"]}], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}], " ", 
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Charge,
                    Editable->False], "2"], "-", 
                  RowBox[{"6", " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False], " ", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}]}], ")"}]}], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], " ", 
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                  RowBox[{"6", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}]}], 
            RowBox[{
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]}], ")"}], "2"], " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["Q",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Charge,
                 Editable->False], "2"], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{"-", "2"}], " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}], ")"}]}]}], ")"}]}]], ",", "0", ",", 
           "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           FractionBox[
            RowBox[{"8", " ", 
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Cot",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cot,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"4", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], " ", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}], "+", 
               RowBox[{
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                  RowBox[{"6", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}]}], 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"2", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}]}], ")"}], "3"]]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           RowBox[{"-", 
            FractionBox[
             RowBox[{"4", " ", 
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"4", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Charge,
                   Editable->False], "2"], " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], "+", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                   RowBox[{"6", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False]}], "]"}]}], 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]}], ")"}], "3"]]}], ",", "0"}], 
          "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           FractionBox[
            RowBox[{"4", " ", 
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Cot",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cot,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"4", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], " ", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}], "+", 
               RowBox[{
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                  RowBox[{"6", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}]}], 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"2", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}]}], ")"}], "3"]]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           RowBox[{"-", 
            FractionBox[
             RowBox[{
              RowBox[{
               RowBox[{"-", "3"}], " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Mass,
                Editable->False], " ", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], " ", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}], "+", 
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Mass,
                Editable->False], " ", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False], "3"]}], "+", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["Cos",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 Cos,
                 Editable->False], "[", 
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["\[Theta]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Theta][],
                  Editable->False]}], "]"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Charge,
                   Editable->False], "2"], "-", 
                 RowBox[{"3", " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Mass,
                   Editable->False], " ", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["Q",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Charge,
                 Editable->False], "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], "-", 
                 RowBox[{"2", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], 
             RowBox[{"2", " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], " ", 
                  SuperscriptBox[
                   RowBox[{
                    InterpretationBox[
                    StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    Cos,
                    Editable->False], "[", 
                    InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False], "]"}], "2"]}], "+", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}], ")"}], "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}]}]}], ")"}]}]]}], ",", "0"}], 
          "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", 
           FractionBox[
            RowBox[{
             RowBox[{
              RowBox[{"-", "3"}], " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["a",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Rotation,
                Editable->False], "2"], " ", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}], "+", 
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], "3"]}], "+", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["a",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Rotation,
                Editable->False], "2"], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["Cos",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Cos,
                Editable->False], "[", 
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False]}], "]"}], " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], "-", 
                RowBox[{"3", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}]}], ")"}]}], "+", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["Q",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Charge,
                Editable->False], "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "-", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]}], ")"}]}]}], 
            RowBox[{"2", " ", 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    Cos,
                    Editable->False], "[", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False], "]"}], "2"]}], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}], ")"}], "2"]}]], ",", "0", ",", 
           "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           FractionBox[
            RowBox[{
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"4", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], " ", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}], "+", 
               RowBox[{
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                  RowBox[{"6", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Sin",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Sin,
               Editable->False], "[", 
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False]}], "]"}]}], 
            RowBox[{"2", " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"2", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["Q",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Charge,
                 Editable->False], "2"], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{"-", "2"}], " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}], ")"}]}]}], ")"}]}]], ",", "0", ",", 
           "0", ",", "0"}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           RowBox[{"-", 
            FractionBox[
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"4", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Charge,
                   Editable->False], "2"], " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], "+", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                   RowBox[{"6", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False]}], "]"}]}], 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]}], ")"}], "3"]]}], ",", "0"}], 
          "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           FractionBox[
            RowBox[{
             RowBox[{"6", " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["a",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Rotation,
                Editable->False], "2"], " ", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}], "-", 
             RowBox[{"4", " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], "3"]}], "-", 
             RowBox[{"2", " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["a",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Rotation,
                Editable->False], "2"], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["Cos",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Cos,
                Editable->False], "[", 
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False]}], "]"}], " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], "-", 
                RowBox[{"3", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}]}], ")"}]}], "-", 
             RowBox[{"2", " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["Q",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Charge,
                Editable->False], "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "-", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]}], ")"}]}]}], 
            RowBox[{
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]}], ")"}], "2"], " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["Q",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Charge,
                 Editable->False], "2"], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{"-", "2"}], " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}], ")"}]}]}], ")"}]}]]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"-", 
            FractionBox[
             RowBox[{
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"4", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Charge,
                   Editable->False], "2"], " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], "+", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                   RowBox[{"6", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False]}], "]"}]}], 
             RowBox[{"2", " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}]}]}], ")"}]}]]}], ",", "0", ",", 
           "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", 
           FractionBox[
            RowBox[{"2", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"-", "3"}], " ", 
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}], "+", 
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "3"]}], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}], " ", 
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Charge,
                    Editable->False], "2"], "-", 
                  RowBox[{"3", " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False], " ", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}]}], ")"}]}], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], " ", 
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                  RowBox[{"2", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], "2"]}], 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"2", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}]}], ")"}], "2"]], ",", "0", ",", 
           "0"}], "}"}]}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           RowBox[{"-", 
            FractionBox[
             RowBox[{
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}]}]}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "3"}], " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], "+", 
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "3"]}], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Charge,
                    Editable->False], "2"], "-", 
                   RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False], " ", 
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}]}], ")"}]}], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Charge,
                   Editable->False], "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                   RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}]}], 
             RowBox[{"2", " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], " ", 
                  SuperscriptBox[
                   RowBox[{
                    InterpretationBox[
                    StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    Cos,
                    Editable->False], "[", 
                    InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False], "]"}], "2"]}], "+", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}], ")"}], "4"]}]]}], ",", "0"}], 
          "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           FractionBox[
            RowBox[{"4", " ", 
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Cot",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cot,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"4", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], " ", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}], "+", 
               RowBox[{
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                  RowBox[{"6", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}]}], 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"2", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}]}], ")"}], "3"]]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"-", 
            FractionBox[
             RowBox[{
              RowBox[{
               RowBox[{"-", "3"}], " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Mass,
                Editable->False], " ", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], " ", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}], "+", 
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Mass,
                Editable->False], " ", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False], "3"]}], "+", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["Cos",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 Cos,
                 Editable->False], "[", 
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["\[Theta]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Theta][],
                  Editable->False]}], "]"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Charge,
                   Editable->False], "2"], "-", 
                 RowBox[{"3", " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Mass,
                   Editable->False], " ", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["Q",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Charge,
                 Editable->False], "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], "-", 
                 RowBox[{"2", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], 
             RowBox[{"2", " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], " ", 
                  SuperscriptBox[
                   RowBox[{
                    InterpretationBox[
                    StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    Cos,
                    Editable->False], "[", 
                    InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False], "]"}], "2"]}], "+", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}], ")"}], "2"]}]]}], ",", "0", ",", 
           "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", 
           RowBox[{"-", 
            FractionBox[
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}]}]}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"4", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Charge,
                   Editable->False], "2"], " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], "+", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                   RowBox[{"6", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False]}], "]"}]}], 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]}], ")"}], "3"]]}], ",", "0", ",", 
           "0"}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           RowBox[{"-", 
            FractionBox[
             RowBox[{"4", " ", 
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["Cot",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Cot,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"4", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Charge,
                   Editable->False], "2"], " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], "+", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                   RowBox[{"6", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}]}], 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]}], ")"}], "3"]]}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           FractionBox[
            RowBox[{
             RowBox[{
              RowBox[{"-", "3"}], " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["a",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Rotation,
                Editable->False], "2"], " ", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}], "+", 
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], "3"]}], "+", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["a",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Rotation,
                Editable->False], "2"], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["Cos",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Cos,
                Editable->False], "[", 
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False]}], "]"}], " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], "-", 
                RowBox[{"3", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}]}], ")"}]}], "+", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["Q",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Charge,
                Editable->False], "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "-", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]}], ")"}]}]}], 
            RowBox[{"2", " ", 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    Cos,
                    Editable->False], "[", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False], "]"}], "2"]}], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}], ")"}], "2"], " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["Q",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Charge,
                 Editable->False], "2"], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{"-", "2"}], " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}], ")"}]}]}], ")"}]}]], ",", "0"}], "}"}],
          ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", 
           RowBox[{"-", 
            FractionBox[
             RowBox[{
              RowBox[{
               RowBox[{"-", "3"}], " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Mass,
                Editable->False], " ", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], " ", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}], "+", 
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Mass,
                Editable->False], " ", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False], "3"]}], "+", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["Cos",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 Cos,
                 Editable->False], "[", 
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["\[Theta]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Theta][],
                  Editable->False]}], "]"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Charge,
                   Editable->False], "2"], "-", 
                 RowBox[{"3", " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Mass,
                   Editable->False], " ", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["Q",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Charge,
                 Editable->False], "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], "-", 
                 RowBox[{"2", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], 
             RowBox[{"2", " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], " ", 
                  SuperscriptBox[
                   RowBox[{
                    InterpretationBox[
                    StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    Cos,
                    Editable->False], "[", 
                    InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False], "]"}], "2"]}], "+", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}], ")"}], "2"]}]]}], ",", "0", ",", 
           "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"-", 
            FractionBox[
             RowBox[{
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"4", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Charge,
                   Editable->False], "2"], " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], "+", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                   RowBox[{"6", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False]}], "]"}]}], 
             RowBox[{"2", " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}]}]}], ")"}]}]]}], ",", "0", ",", 
           "0", ",", "0"}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", 
           RowBox[{"-", 
            FractionBox[
             RowBox[{"4", " ", 
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}]}]}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"4", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Charge,
                   Editable->False], "2"], " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], "+", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                   RowBox[{"6", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False]}], "]"}]}], 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]}], ")"}], "3"]]}], ",", "0", ",", 
           "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"-", 
            FractionBox[
             RowBox[{
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"4", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Charge,
                   Editable->False], "2"], " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], "+", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                   RowBox[{"6", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False]}], "]"}]}], 
             RowBox[{
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}]}]}], ")"}]}]]}], ",", "0", ",", 
           "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           RowBox[{"-", 
            FractionBox[
             RowBox[{"2", " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], "-", 
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}]}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "3"}], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"]}], "-", 
                RowBox[{"3", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]}], ")"}]}], 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]}], ")"}], "2"]]}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           RowBox[{"-", 
            FractionBox[
             RowBox[{"2", " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], "-", 
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}]}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"3", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"]}], "+", 
                RowBox[{"3", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "-", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]}], ")"}], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["Sin",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 Sin,
                 Editable->False], "[", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False], "]"}], "2"]}], 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]}], ")"}], "2"]]}], ",", "0"}], 
          "}"}]}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           RowBox[{"-", 
            FractionBox[
             RowBox[{"8", " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}]}]}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "3"}], " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], "+", 
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "3"]}], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Charge,
                    Editable->False], "2"], "-", 
                   RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False], " ", 
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}]}], ")"}]}], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Charge,
                   Editable->False], "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                   RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}]}], 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]}], ")"}], "4"]]}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           RowBox[{"-", 
            FractionBox[
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"4", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Charge,
                   Editable->False], "2"], " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], "+", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                   RowBox[{"6", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False]}], "]"}]}], 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]}], ")"}], "3"]]}], ",", "0"}], 
          "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", 
           FractionBox[
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["Q",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Charge,
                 Editable->False], "2"], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{"-", "2"}], " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"4", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], " ", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}], "+", 
               RowBox[{
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                  RowBox[{"6", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Sin",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Sin,
               Editable->False], "[", 
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False]}], "]"}]}], 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"2", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}]}], ")"}], "3"]], ",", "0", ",", 
           "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"-", 
            FractionBox[
             RowBox[{"2", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "3"}], " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], "+", 
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "3"]}], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Charge,
                    Editable->False], "2"], "-", 
                   RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False], " ", 
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}]}], ")"}]}], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Charge,
                   Editable->False], "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                   RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["Sin",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 Sin,
                 Editable->False], "[", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False], "]"}], "2"]}], 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]}], ")"}], "2"]]}], ",", "0", ",", 
           "0", ",", "0"}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           FractionBox[
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"4", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], " ", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}], "+", 
               RowBox[{
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                  RowBox[{"6", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Sin",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Sin,
               Editable->False], "[", 
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False]}], "]"}]}], 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"2", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}]}], ")"}], "3"]], ",", "0"}], "}"}],
          ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           RowBox[{"-", 
            FractionBox[
             RowBox[{
              RowBox[{"6", " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Mass,
                Editable->False], " ", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], " ", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}], "-", 
              RowBox[{"4", " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Mass,
                Editable->False], " ", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False], "3"]}], "-", 
              RowBox[{"2", " ", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["Cos",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 Cos,
                 Editable->False], "[", 
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["\[Theta]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Theta][],
                  Editable->False]}], "]"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Charge,
                   Editable->False], "2"], "-", 
                 RowBox[{"3", " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Mass,
                   Editable->False], " ", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}]}], ")"}]}], "-", 
              RowBox[{"2", " ", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["Q",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Charge,
                 Editable->False], "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], "-", 
                 RowBox[{"2", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], 
             RowBox[{
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], "+", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    Cos,
                    Editable->False], "[", 
                   RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                 RowBox[{"2", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}], "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}]}]}], ")"}]}]]}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           FractionBox[
            RowBox[{
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"4", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], " ", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}], "+", 
               RowBox[{
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                  RowBox[{"6", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Sin",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Sin,
               Editable->False], "[", 
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False]}], "]"}]}], 
            RowBox[{"2", " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"2", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["Q",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Charge,
                 Editable->False], "2"], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{"-", "2"}], " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}], ")"}]}]}], ")"}]}]], ",", "0", ",", 
           "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", 
           RowBox[{"-", 
            FractionBox[
             RowBox[{"2", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "3"}], " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], "+", 
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "3"]}], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Charge,
                    Editable->False], "2"], "-", 
                   RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False], " ", 
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}]}], ")"}]}], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["Q",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Charge,
                   Editable->False], "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                   RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["Sin",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 Sin,
                 Editable->False], "[", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False], "]"}], "2"]}], 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Cos",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   Cos,
                   Editable->False], "[", 
                  RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]}], ")"}], "2"]]}], ",", "0", ",", 
           "0"}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", 
           FractionBox[
            RowBox[{"4", " ", 
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["Q",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Charge,
                 Editable->False], "2"], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{"-", "2"}], " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"4", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], " ", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}], "+", 
               RowBox[{
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                  RowBox[{"6", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Sin",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Sin,
               Editable->False], "[", 
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False]}], "]"}]}], 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"2", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}]}], ")"}], "3"]], ",", "0", ",", 
           "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           FractionBox[
            RowBox[{
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"4", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Charge,
                  Editable->False], "2"], " ", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}], "+", 
               RowBox[{
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], "-", 
                  RowBox[{"6", " ", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Sin",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Sin,
               Editable->False], "[", 
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False]}], "]"}]}], 
            RowBox[{
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"2", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["Q",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Charge,
                 Editable->False], "2"], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{"-", "2"}], " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}], ")"}]}]}], ")"}]}]], ",", "0", ",", 
           "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           FractionBox[
            RowBox[{"2", " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["Q",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Charge,
                 Editable->False], "2"], "-", 
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"-", "3"}], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"]}], "-", 
               RowBox[{"3", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"2", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}]}], ")"}]}], 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"2", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}]}], ")"}], "2"]]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           FractionBox[
            RowBox[{"2", " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["Q",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Charge,
                 Editable->False], "2"], "-", 
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], " ", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"3", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"]}], "+", 
               RowBox[{"3", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "-", 
               RowBox[{"2", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}]}], ")"}], " ", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], "2"]}], 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Cos",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  Cos,
                  Editable->False], "[", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Theta][],
                   Editable->False]}], "]"}]}], "+", 
               RowBox[{"2", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}]}], ")"}], "2"]], ",", "0"}], 
          "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}]}], 
      "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "frame"}], ",", 
     RowBox[{"-", "frame"}], ",", 
     RowBox[{"-", "frame"}], ",", "frame"}], "}"}], ",", "0"}], 
  "]"}]], "Output",
 CellChangeTimes->{3.728676401199925*^9, 3.728753580030881*^9, 
  3.7288436825033407`*^9, 3.730956290699951*^9, 3.731118141061708*^9, 
  3.733368191046858*^9, 3.794517960680543*^9},
 CellLabel->"Out[344]=",
 CellID->1186534209,ExpressionUUID->"103cc88e-ae96-4c3d-b92b-d6adcedd05fd"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Check the Bianchi identity:", "MathCaption",
 CellID->1095120460,ExpressionUUID->"5d53e608-4803-4897-9e43-c9ffc4fde434"],

Cell[BoxData[
 RowBox[{
  RowBox[{"cdriemdown", "=", 
   RowBox[{"TensorDerivative", "[", 
    RowBox[{
     RowBox[{"RiemannDown", "[", "cd", "]"}], ",", "cd"}], "]"}]}], 
  ";"}]], "Input",
 CellLabel->"In[345]:=",
 CellID->1754716430,ExpressionUUID->"63b17bc0-06e7-4af1-a2f8-2c5586e64cf5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"cdriemdown", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "d"}], ",", 
      RowBox[{"-", "e"}]}], "]"}], "+", 
    RowBox[{"cdriemdown", "[", 
     RowBox[{
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "d"}], ",", 
      RowBox[{"-", "e"}]}], "]"}], "+", 
    RowBox[{"cdriemdown", "[", 
     RowBox[{
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "d"}], ",", 
      RowBox[{"-", "e"}]}], "]"}]}], "//", "Simplify"}], "//", 
  "Timing"}]], "Input",
 CellLabel->"In[346]:=",
 CellID->126114759,ExpressionUUID->"25edd050-b25f-4dbc-9905-d9e6db0f9380"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.007116`", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.7286764036958714`*^9, 3.72875358054841*^9, 
  3.7288436849561577`*^9, 3.7309562909584293`*^9, 3.731118147350976*^9, 
  3.7333681939904127`*^9, 3.794517960755826*^9},
 CellLabel->"Out[346]=",
 CellID->547304621,ExpressionUUID->"81273cb6-bb72-49da-a041-e52a47920673"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Clean up:", "MathCaption",
 CellID->968662159,ExpressionUUID->"b848bb31-6e6b-4b2c-8fad-6b97ba4f0b4d"],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{
  "metric", ",", "cd", ",", "factor", ",", "\[Rho]2", ",", "change", ",", 
   "cdriemdown"}], "]"}]], "Input",
 CellLabel->"In[347]:=",
 CellID->1831468572,ExpressionUUID->"e40f37ec-7c95-4e4a-a4fe-bcdf24b55d84"],

Cell[BoxData[
 RowBox[{
  RowBox[{"UndefConstantSymbol", "[", 
   RowBox[{"{", 
    RowBox[{"Mass", ",", "Rotation", ",", "Charge"}], "}"}], "]"}], 
  ";"}]], "Input",
 CellLabel->"In[348]:=",
 CellID->1609177153,ExpressionUUID->"5aa63be9-440d-4598-aa5f-4e388e155335"],

Cell[BoxData[
 RowBox[{"UndefTensor", "[", "JacobianBLframe", "]"}]], "Input",
 CellChangeTimes->{{3.733368222468423*^9, 3.7333682350248117`*^9}},
 CellLabel->"In[349]:=",
 CellID->1102632757,ExpressionUUID->"233baefb-5d5d-498a-a4df-13ab7cd2d040"],

Cell[BoxData[
 RowBox[{"UndefBasis", "[", "frame", "]"}]], "Input",
 CellLabel->"In[350]:=",
 CellID->31360153,ExpressionUUID->"b051f718-c16e-4204-be0a-6a75be57b7d4"],

Cell[BoxData[
 RowBox[{"UndefChart", "[", "BL", "]"}]], "Input",
 CellLabel->"In[351]:=",
 CellID->936597956,ExpressionUUID->"53feb4b3-e44e-422c-aea4-b76b453485ed"],

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M", "]"}]], "Input",
 CellLabel->"In[352]:=",
 CellID->346713982,ExpressionUUID->"1ffc3542-fa9c-4efe-a98c-2d85e7029da9"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Clear the cache:", "MathCaption",
 CellID->1098761163,ExpressionUUID->"026fb734-2e18-4d57-b712-430bbfa9228f"],

Cell[BoxData[
 RowBox[{"ClearxCobaCache", "[", "All", "]"}]], "Input",
 CellLabel->"In[353]:=",
 CellID->666985988,ExpressionUUID->"64e65936-5ae4-429c-b534-85cd452a5e3a"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.5.2. General 2d metric connection with torsion", "Subsubsection",
 CellChangeTimes->{{3.72867641222936*^9, 3.728676412395969*^9}, 
   3.790876748402484*^9},
 CellID->168862976,ExpressionUUID->"ef267f85-b30c-4069-9067-7d160e846c9a"],

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<xAct`xCoba`\>\"", "]"}]], "Input",
 CellLabel->"In[354]:=",
 CellID->1254139945,ExpressionUUID->"d55c534e-e749-4b1b-a760-3d69b8ae8a7f"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"$DefInfoQ", "=", 
   RowBox[{"$UndefInfoQ", "=", "False"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"$CVVerbose", "=", "False"}], ";"}]}], "Input",
 CellChangeTimes->{{3.731118193314114*^9, 3.731118198362093*^9}},
 CellLabel->"In[355]:=",
 CellID->778060310,ExpressionUUID->"d82adbe8-28f1-41c3-ad03-4dec3f8ba65e"],

Cell[CellGroupData[{

Cell["Define a 2d manifold:", "MathCaption",
 CellID->831703587,ExpressionUUID->"a60d815c-71c1-48a8-96b3-fe0de2ab2564"],

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M", ",", "2", ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f"}], "}"}]}], 
  "]"}]], "Input",
 CellLabel->"In[357]:=",
 CellID->471128386,ExpressionUUID->"b6179f4b-a5bf-4c5f-a9b6-09fc793b7fac"]
}, Open  ]],

Cell[CellGroupData[{

Cell["with a general chart:", "MathCaption",
 CellID->482577013,ExpressionUUID->"d00f9eca-b9d7-4b90-b6e3-ccd9e991e9fb"],

Cell[BoxData[
 RowBox[{"DefChart", "[", 
  RowBox[{"gen", ",", "M", ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", "2"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"x", "[", "]"}], ",", 
     RowBox[{"y", "[", "]"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[358]:=",
 CellID->1773606419,ExpressionUUID->"b2c033bf-83c1-4ba0-bbe0-2c910cddad15"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Let us now work with a fully general metric:", "MathCaption",
 CellID->1930063773,ExpressionUUID->"83a6946e-836d-45f2-ae28-a7ddef20d0a0"],

Cell[BoxData[
 RowBox[{
  RowBox[{"DefScalarFunction", "/@", 
   RowBox[{"{", 
    RowBox[{"gxx", ",", "gxy", ",", "gyy"}], "}"}]}], ";"}]], "Input",
 CellLabel->"In[359]:=",
 CellID->841344876,ExpressionUUID->"8b7a6ef4-0997-481a-a53f-d6b46aa2859a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{"metric", "=", 
    RowBox[{"CTensor", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{
          RowBox[{"gxx", "[", 
           RowBox[{
            RowBox[{"x", "[", "]"}], ",", 
            RowBox[{"y", "[", "]"}]}], "]"}], ",", 
          RowBox[{"gxy", "[", 
           RowBox[{
            RowBox[{"x", "[", "]"}], ",", 
            RowBox[{"y", "[", "]"}]}], "]"}]}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{"gxy", "[", 
           RowBox[{
            RowBox[{"x", "[", "]"}], ",", 
            RowBox[{"y", "[", "]"}]}], "]"}], ",", 
          RowBox[{"gyy", "[", 
           RowBox[{
            RowBox[{"x", "[", "]"}], ",", 
            RowBox[{"y", "[", "]"}]}], "]"}]}], "}"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "gen"}], ",", 
        RowBox[{"-", "gen"}]}], "}"}]}], "]"}]}], ")"}], "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}]}], "]"}]], "Input",
 CellLabel->"In[360]:=",
 CellID->1450999084,ExpressionUUID->"fa7e6fe3-4895-447e-99f3-cbf64dfd637c"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{
             InterpretationBox[
              StyleBox["gxx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{
                InterpretationBox[
                 StyleBox[
                  RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                  False], $CellContext`gxx, Editable -> False], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`x[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`y[], Editable -> False]}], "]"}]}]],
           Annotation[#, {1, 1} -> $CellContext`gxx[
              $CellContext`x[], 
              $CellContext`y[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             InterpretationBox[
              StyleBox["gxy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{
                InterpretationBox[
                 StyleBox[
                  RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                  False], $CellContext`gxy, Editable -> False], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`x[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`y[], Editable -> False]}], "]"}]}]],
           Annotation[#, {1, 2} -> $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             InterpretationBox[
              StyleBox["gxy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{
                InterpretationBox[
                 StyleBox[
                  RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                  False], $CellContext`gxy, Editable -> False], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`x[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`y[], Editable -> False]}], "]"}]}]],
           Annotation[#, {2, 1} -> $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             InterpretationBox[
              StyleBox["gyy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gyy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{
                InterpretationBox[
                 StyleBox[
                  RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                  False], $CellContext`gyy, Editable -> False], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`x[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`y[], Editable -> False]}], "]"}]}]],
           Annotation[#, {2, 2} -> $CellContext`gyy[
              $CellContext`x[], 
              $CellContext`y[]], "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " "},
         {"a", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{
     $CellContext`gxx[
      $CellContext`x[], 
      $CellContext`y[]], 
     $CellContext`gxy[
      $CellContext`x[], 
      $CellContext`y[]]}, {
     $CellContext`gxy[
      $CellContext`x[], 
      $CellContext`y[]], 
     $CellContext`gyy[
      $CellContext`x[], 
      $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 
   0][-$CellContext`a, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.7286764976025057`*^9, 3.7287535840429564`*^9, 
  3.728843697930614*^9, 3.730956299426064*^9, 3.731118237327302*^9, 
  3.731118973685905*^9, 3.7333683124117107`*^9, 3.733368367887189*^9, 
  3.794517961560644*^9},
 CellLabel->"Out[360]=",
 CellID->876305046,ExpressionUUID->"b8e456bf-e4cb-462f-9e05-645a0c3e4c55"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SymmetryGroupOfTensor", "[", "metric", "]"}]], "Input",
 CellLabel->"In[361]:=",
 CellID->230039046,ExpressionUUID->"8c941de1-ca7f-401d-9f06-16e5588b5966"],

Cell[BoxData[
 RowBox[{"StrongGenSet", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "2"}], "}"}], ",", 
   RowBox[{"GenSet", "[", 
    InterpretationBox[
     StyleBox[
      RowBox[{"(", "1", ",", "2", ")"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xPerm`Cycles[{1, 2}],
     Editable->False], "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728676497717989*^9, 3.728753584284091*^9, 
  3.7288436984310703`*^9, 3.7309562995485077`*^9, 3.73111823850808*^9, 
  3.731118974391448*^9, 3.7333683130573807`*^9, 3.733368368011014*^9, 
  3.794517961613062*^9},
 CellLabel->"Out[361]=",
 CellID->616056996,ExpressionUUID->"3673aa49-cccd-4537-9858-93b819ed8284"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
In this case we will also have torsion, a general torsion tensor:\
\>", "MathCaption",
 CellID->638670385,ExpressionUUID->"4d2e63bd-cf66-4d6c-9f9b-313b6a562c1b"],

Cell[BoxData[
 RowBox[{
  RowBox[{"DefScalarFunction", "/@", 
   RowBox[{"{", 
    RowBox[{"T112", ",", "T212"}], "}"}]}], ";"}]], "Input",
 CellLabel->"In[362]:=",
 CellID->1009892989,ExpressionUUID->"0d6b9bd2-64c3-4ee8-aef9-c1de825ff8fc"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{"torsion", "=", 
    RowBox[{"CTensor", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{"0", ",", 
            RowBox[{"T112", "[", 
             RowBox[{
              RowBox[{"x", "[", "]"}], ",", 
              RowBox[{"y", "[", "]"}]}], "]"}]}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", 
             RowBox[{"T112", "[", 
              RowBox[{
               RowBox[{"x", "[", "]"}], ",", 
               RowBox[{"y", "[", "]"}]}], "]"}]}], ",", "0"}], "}"}]}], "}"}],
         ",", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{"0", ",", 
            RowBox[{"T212", "[", 
             RowBox[{
              RowBox[{"x", "[", "]"}], ",", 
              RowBox[{"y", "[", "]"}]}], "]"}]}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", 
             RowBox[{"T212", "[", 
              RowBox[{
               RowBox[{"x", "[", "]"}], ",", 
               RowBox[{"y", "[", "]"}]}], "]"}]}], ",", "0"}], "}"}]}], 
         "}"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"gen", ",", 
        RowBox[{"-", "gen"}], ",", 
        RowBox[{"-", "gen"}]}], "}"}]}], "]"}]}], ")"}], "[", 
  RowBox[{"a", ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}]}], "]"}]], "Input",
 CellLabel->"In[363]:=",
 CellID->1708203882,ExpressionUUID->"3ed046d9-b62e-40c5-a83f-b839b40c1761"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["T112",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T112,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 1, 2} -> $CellContext`T112[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["T112",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`T112,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{"-", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]],
              Annotation[#, {1, 2, 1} -> -$CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["T212",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T212,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 1, 2} -> $CellContext`T212[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["T212",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`T212,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{"-", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]],
              Annotation[#, {2, 2, 1} -> -$CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a", " ", " "},
         {" ", "b", "c"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{{0, 
      $CellContext`T112[
       $CellContext`x[], 
       $CellContext`y[]]}, {-$CellContext`T112[
        $CellContext`x[], 
        $CellContext`y[]], 0}}, {{0, 
      $CellContext`T212[
       $CellContext`x[], 
       $CellContext`y[]]}, {-$CellContext`T212[
        $CellContext`x[], 
        $CellContext`y[]], 
      0}}}, {$CellContext`gen, -$CellContext`gen, -$CellContext`gen}, 
   0][$CellContext`a, -$CellContext`b, -$CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{3.7286764979087057`*^9, 3.728753584842661*^9, 
  3.7288436988313017`*^9, 3.730956299679963*^9, 3.7311182408130007`*^9, 
  3.731118976752184*^9, 3.733368314689329*^9, 3.7333683699456882`*^9, 
  3.7945179616883497`*^9},
 CellLabel->"Out[363]=",
 CellID->1051269585,ExpressionUUID->"8066f187-6190-4f9a-b25d-c47058c9d770"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SymmetryGroupOfTensor", "[", "torsion", "]"}]], "Input",
 CellLabel->"In[364]:=",
 CellID->721232698,ExpressionUUID->"3b857b95-5f70-477d-a1dc-ed22e3c1cf4b"],

Cell[BoxData[
 RowBox[{"StrongGenSet", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"2", ",", "3"}], "}"}], ",", 
   RowBox[{"GenSet", "[", 
    InterpretationBox[
     StyleBox[
      RowBox[{"-", 
       RowBox[{"(", "2", ",", "3", ")"}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     -xAct`xPerm`Cycles[{2, 3}],
     Editable->False], "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728676498052425*^9, 3.728753584974671*^9, 
  3.728843699067399*^9, 3.730956299807645*^9, 3.73111824811966*^9, 
  3.731118978460829*^9, 3.733368315474828*^9, 3.733368370081435*^9, 
  3.794517961738491*^9},
 CellLabel->"Out[364]=",
 CellID->1339951226,ExpressionUUID->"ee18d6ac-87ea-4715-a7af-4eee0cb08973"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Construct a covariant derivative that is associated to the metric and has the \
given torsion tensor:\
\>", "MathCaption",
 CellChangeTimes->{{3.731118400467594*^9, 3.731118406187318*^9}, {
  3.73111900888839*^9, 3.731119036677868*^9}},
 CellID->1518900828,ExpressionUUID->"fa11c277-4c89-4afd-8a70-4cf718b71973"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"cd", "=", 
  RowBox[{"CovDOfMetric", "[", 
   RowBox[{"metric", ",", "torsion"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.731118984360132*^9, 3.731118985686317*^9}},
 CellLabel->"In[365]:=",
 CellID->1028104307,ExpressionUUID->"794976fd-c18a-4ac9-bc6a-5d644d62b137"],

Cell[BoxData[
 RowBox[{"CCovD", "[", 
  RowBox[{"PDgen", ",", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           RowBox[{"-", 
            FractionBox[
             RowBox[{
              RowBox[{"2", " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["T112",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T112,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}], "+", 
              RowBox[{"2", " ", 
               SuperscriptBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], "2"], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["T212",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T212,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}], "+", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["gxx",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxx,
                  Editable->False], 
                 TagBox[
                  RowBox[{"(", 
                   RowBox[{"1", ",", "0"}], ")"}],
                  Derivative],
                 MultilineFunction->None], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}], "+", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}]}], 
             RowBox[{"2", " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "-", 
                RowBox[{
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxx,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gyy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}]]}], ",", 
           FractionBox[
            RowBox[{
             RowBox[{
              RowBox[{"-", "2"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["T112",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`T112,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], "-", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"0", ",", "1"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "2"}], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gyy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`T212,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gyy,
                   Editable->False], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}]}], ")"}]}]}], 
            RowBox[{"2", " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], "2"], "-", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxx",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxx,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["gyy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gyy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}]}]}], ")"}]}]]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           FractionBox[
            RowBox[{
             RowBox[{
              RowBox[{"-", "2"}], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], "2"], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["T112",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`T112,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], "-", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"0", ",", "1"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "2"}], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gyy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`T212,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gyy,
                   Editable->False], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}]}], ")"}]}]}], 
            RowBox[{"2", " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], "2"], "-", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxx",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxx,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["gyy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gyy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}]}]}], ")"}]}]], ",", 
           FractionBox[
            RowBox[{
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "2"}], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gyy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`T112,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gyy,
                   Editable->False], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}]}], ")"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "2"}], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gyy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`T212,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                RowBox[{"2", " ", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gyy,
                   Editable->False], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}]}], ")"}]}]}], 
            RowBox[{"2", " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], "2"], "-", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxx",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxx,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["gyy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gyy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}]}]}], ")"}]}]]}], "}"}]}], "}"}], 
       ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           FractionBox[
            RowBox[{
             RowBox[{"2", " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], "2"], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["T112",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`T112,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"2", " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`T212,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxx,
                   Editable->False], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], "-", 
                RowBox[{"2", " ", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}]}], 
            RowBox[{"2", " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], "2"], "-", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxx",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxx,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["gyy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gyy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}]}]}], ")"}]}]], ",", 
           FractionBox[
            RowBox[{
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"2", " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`T212,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxx,
                   Editable->False], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}]}], ")"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"2", " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`T112,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gyy,
                   Editable->False], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}]}], ")"}]}]}], 
            RowBox[{"2", " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], "2"], "-", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxx",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxx,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["gyy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gyy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}]}]}], ")"}]}]]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           FractionBox[
            RowBox[{
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"0", ",", "1"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"2", " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`T112,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"2", " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gyy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`T212,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gyy,
                   Editable->False], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}]}], ")"}]}]}], 
            RowBox[{"2", " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], "2"], "-", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxx",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxx,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["gyy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gyy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}]}]}], ")"}]}]], ",", 
           FractionBox[
            RowBox[{
             RowBox[{"2", " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], "2"], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["T112",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`T112,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], "-", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"0", ",", "1"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"2", " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gyy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`T212,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{"2", " ", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gyy,
                   Editable->False], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}]}], ")"}]}]}], 
            RowBox[{"2", " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], "2"], "-", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxx",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxx,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["gyy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gyy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}]}]}], ")"}]}]]}], "}"}]}], 
        "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"gen", ",", 
       RowBox[{"-", "gen"}], ",", 
       RowBox[{"-", "gen"}]}], "}"}], ",", "0"}], "]"}], ",", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{
          InterpretationBox[
           StyleBox["gxx",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gxx,
           Editable->False], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["x",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["y",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`y[],
            Editable->False]}], "]"}], ",", 
         RowBox[{
          InterpretationBox[
           StyleBox["gxy",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gxy,
           Editable->False], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["x",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["y",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`y[],
            Editable->False]}], "]"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{
          InterpretationBox[
           StyleBox["gxy",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gxy,
           Editable->False], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["x",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["y",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`y[],
            Editable->False]}], "]"}], ",", 
         RowBox[{
          InterpretationBox[
           StyleBox["gyy",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gyy,
           Editable->False], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["x",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["y",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`y[],
            Editable->False]}], "]"}]}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "gen"}], ",", 
       RowBox[{"-", "gen"}]}], "}"}], ",", "0"}], "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.7286764985362453`*^9, 3.728753585386524*^9, 
  3.728843699850906*^9, 3.730956300372238*^9, 3.731118285249446*^9, 
  3.731118408310595*^9, 3.73111899744685*^9, 3.7333683167130547`*^9, 
  3.73336837034396*^9, 3.794517961941008*^9},
 CellLabel->"Out[365]=",
 CellID->2046732940,ExpressionUUID->"b18f5042-f5d4-456f-8831-bacb8b29d661"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["It has the torsion we provided:", "MathCaption",
 CellID->1123655241,ExpressionUUID->"6cdc36c6-13e7-4d42-b297-a5e914013ab2"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Torsion", "[", "cd", "]"}], "===", "torsion"}]], "Input",
 CellLabel->"In[366]:=",
 CellID->302443769,ExpressionUUID->"4de65dc3-8c14-42fc-b738-160e3be8d518"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728676498685892*^9, 3.7287535855434113`*^9, 
  3.728843700152853*^9, 3.730956300591868*^9, 3.731118418017288*^9, 
  3.7311190432005444`*^9, 3.733368318257494*^9, 3.733368370475739*^9, 
  3.794517961992578*^9},
 CellLabel->"Out[366]=",
 CellID->1047464934,ExpressionUUID->"18b5649e-e6ca-48b5-a846-9c8d78394058"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Therefore this covariant derivative is not the Levi-Civita connection of the \
metric. We can obtain the Levi-Civita connection specifying that torsion is \
Zero (which is actually the default):\
\>", "MathCaption",
 CellChangeTimes->{{3.73111843768226*^9, 3.7311184657936068`*^9}, {
  3.731119079062181*^9, 3.731119094164011*^9}},
 CellID->1902182538,ExpressionUUID->"aff740f3-ab19-41a7-9d09-15abb8539e53"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"lc", "=", 
  RowBox[{"CovDOfMetric", "[", 
   RowBox[{"metric", ",", "Zero"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.731118467732916*^9, 3.7311184983134212`*^9}, {
  3.7311190588302107`*^9, 3.731119067992818*^9}},
 CellLabel->"In[367]:=",
 CellID->1570432718,ExpressionUUID->"5de306fe-f0bb-4554-98b5-b4f846904337"],

Cell[BoxData[
 RowBox[{"CCovD", "[", 
  RowBox[{"PDgen", ",", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           RowBox[{"-", 
            FractionBox[
             RowBox[{
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["gxx",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxx,
                  Editable->False], 
                 TagBox[
                  RowBox[{"(", 
                   RowBox[{"1", ",", "0"}], ")"}],
                  Derivative],
                 MultilineFunction->None], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}], "+", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}]}], 
             RowBox[{"2", " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "-", 
                RowBox[{
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxx,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gyy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}]]}], ",", 
           FractionBox[
            RowBox[{
             RowBox[{
              RowBox[{"-", 
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}], " ", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"0", ",", "1"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}]}], 
            RowBox[{"2", " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], "2"], "-", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxx",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxx,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["gyy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gyy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}]}]}], ")"}]}]]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           FractionBox[
            RowBox[{
             RowBox[{
              RowBox[{"-", 
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}], " ", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"0", ",", "1"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}]}], 
            RowBox[{"2", " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], "2"], "-", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxx",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxx,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["gyy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gyy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}]}]}], ")"}]}]], ",", 
           FractionBox[
            RowBox[{
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"0", ",", "1"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "2"}], " ", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gyy,
                   Editable->False], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}]}], ")"}]}]}], 
            RowBox[{"2", " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], "2"], "-", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxx",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxx,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["gyy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gyy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}]}]}], ")"}]}]]}], "}"}]}], "}"}], 
       ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           FractionBox[
            RowBox[{
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxx,
                   Editable->False], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], "-", 
                RowBox[{"2", " ", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}]}], 
            RowBox[{"2", " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], "2"], "-", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxx",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxx,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], " ", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["gyy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gyy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}]}]}], ")"}]}]], ",", 
           FractionBox[
            RowBox[{
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"0", ",", "1"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], "-", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}]}], 
            RowBox[{
             RowBox[{"2", " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], "2"]}], "-", 
             RowBox[{"2", " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}]}]]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           FractionBox[
            RowBox[{
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"0", ",", "1"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], "-", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}]}], 
            RowBox[{
             RowBox[{"2", " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], "2"]}], "-", 
             RowBox[{"2", " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}]}]], ",", 
           RowBox[{"-", 
            FractionBox[
             RowBox[{
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["gyy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gyy,
                  Editable->False], 
                 TagBox[
                  RowBox[{"(", 
                   RowBox[{"0", ",", "1"}], ")"}],
                  Derivative],
                 MultilineFunction->None], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}], "+", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "2"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], 
             RowBox[{"2", " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "-", 
                RowBox[{
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxx,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gyy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}]]}]}], "}"}]}], 
        "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"gen", ",", 
       RowBox[{"-", "gen"}], ",", 
       RowBox[{"-", "gen"}]}], "}"}], ",", "0"}], "]"}], ",", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{
          InterpretationBox[
           StyleBox["gxx",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gxx,
           Editable->False], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["x",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["y",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`y[],
            Editable->False]}], "]"}], ",", 
         RowBox[{
          InterpretationBox[
           StyleBox["gxy",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gxy,
           Editable->False], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["x",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["y",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`y[],
            Editable->False]}], "]"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{
          InterpretationBox[
           StyleBox["gxy",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gxy,
           Editable->False], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["x",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["y",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`y[],
            Editable->False]}], "]"}], ",", 
         RowBox[{
          InterpretationBox[
           StyleBox["gyy",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gyy,
           Editable->False], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["x",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["y",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`y[],
            Editable->False]}], "]"}]}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "gen"}], ",", 
       RowBox[{"-", "gen"}]}], "}"}], ",", "0"}], "]"}]}], "]"}]], "Output",
 CellChangeTimes->{{3.731118484270021*^9, 3.731118508142869*^9}, 
   3.731119095662364*^9, 3.7333683196538973`*^9, 3.7333683706043177`*^9, 
   3.794517962051668*^9},
 CellLabel->"Out[367]=",
 CellID->1819691279,ExpressionUUID->"d31edbb2-3022-40ff-8bb9-6cdc40cc6dc8"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Torsion", "[", "lc", "]"}]], "Input",
 CellChangeTimes->{{3.73111848745179*^9, 3.731118510559671*^9}},
 CellLabel->"In[368]:=",
 CellID->1156953266,ExpressionUUID->"efcb8b99-4823-4f08-a16d-ee635b068388"],

Cell[BoxData["Zero"], "Output",
 CellChangeTimes->{{3.7311184915324507`*^9, 3.731118511478772*^9}, 
   3.731119100201552*^9, 3.733368321191917*^9, 3.7333683707087297`*^9, 
   3.794517962098194*^9},
 CellLabel->"Out[368]=",
 CellID->1236709226,ExpressionUUID->"74a3665e-44e4-4a74-ac2b-522db6e5a5b6"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["There is a shortcut to get the Levi-Civita connection:", "MathCaption",
 CellChangeTimes->{{3.731119103587391*^9, 3.731119115563191*^9}},
 CellID->562721375,ExpressionUUID->"8c33fccc-a5be-4644-9744-faa61970b62c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"LC", "[", "metric", "]"}], "===", "lc"}]], "Input",
 CellChangeTimes->{{3.731119117503272*^9, 3.7311191227787867`*^9}},
 CellLabel->"In[369]:=",
 CellID->253715799,ExpressionUUID->"4b1116eb-6427-4eb2-945b-6b5ac5ebb61b"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.7311191234578238`*^9, 3.733368322292493*^9, 
  3.733368370842986*^9, 3.794517962145445*^9},
 CellLabel->"Out[369]=",
 CellID->2025225716,ExpressionUUID->"1c79bee8-16a4-438d-a0b3-f17048d8d965"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Both covariant derivatives are associated to the metric, in the sense that:\
\>", "MathCaption",
 CellChangeTimes->{{3.731118518144848*^9, 3.731118530303355*^9}},
 CellID->519202859,ExpressionUUID->"12924cfb-fc9f-4830-b114-e5e29d3e86aa"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"cd", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"metric", "[", 
   RowBox[{
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}]}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.731118532378029*^9, 3.731118536511159*^9}},
 CellLabel->"In[370]:=",
 CellID->46692262,ExpressionUUID->"6367e556-f81a-4b5f-86cf-54921fbb7ba1"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.731118537767043*^9, 3.731119131280799*^9, 
  3.733368323324184*^9, 3.733368370975638*^9, 3.794517962195593*^9},
 CellLabel->"Out[370]=",
 CellID->352803862,ExpressionUUID->"17ea0895-722e-40d4-890b-f27bb87d6ffd"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"lc", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"metric", "[", 
   RowBox[{
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}]}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.731118539392825*^9, 3.731118544246849*^9}},
 CellLabel->"In[371]:=",
 CellID->1213071162,ExpressionUUID->"d612c7b2-cb77-4401-ad9a-11a150e9a2d5"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.731118544739238*^9, 3.731119133119946*^9, 
  3.733368323875317*^9, 3.7333683711580763`*^9, 3.7945179622490664`*^9},
 CellLabel->"Out[371]=",
 CellID->372159267,ExpressionUUID->"a2dc8b5b-9548-4fd2-ad2d-e2ffc1027320"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The Christoffel tensor relating cd to the partial derivative of the chart is \
not symmetric:\
\>", "MathCaption",
 CellChangeTimes->{{3.731119145011397*^9, 3.7311191574177113`*^9}},
 CellID->884640233,ExpressionUUID->"0dddd146-5faf-4114-9a02-a9823de4668c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"christoffel", "=", 
  RowBox[{"Christoffel", "[", 
   RowBox[{"cd", ",", "PDgen"}], "]"}]}]], "Input",
 CellLabel->"In[372]:=",
 CellID->1119107850,ExpressionUUID->"175fec35-3a6c-461e-83f9-c3523b4767a2"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", 
          FractionBox[
           RowBox[{
            RowBox[{"2", " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["gxx",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxx,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["gxy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["T112",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`T112,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}]}], "+", 
            RowBox[{"2", " ", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], "2"], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["T212",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`T212,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}]}], "+", 
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox["gyy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gyy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], " ", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"1", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}]}], "+", 
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox["gxy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["gxx",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxx,
                  Editable->False], 
                 TagBox[
                  RowBox[{"(", 
                   RowBox[{"0", ",", "1"}], ")"}],
                  Derivative],
                 MultilineFunction->None], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], "-", 
               RowBox[{"2", " ", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxy,
                   Editable->False], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}]}]}], ")"}]}]}], 
           RowBox[{"2", " ", 
            RowBox[{"(", 
             RowBox[{
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], "2"], "-", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}]}], ")"}]}]]}], ",", 
         FractionBox[
          RowBox[{
           RowBox[{
            RowBox[{"-", "2"}], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["gxx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["gyy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gyy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["T112",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`T112,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}]}], "-", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["gyy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gyy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["gxx",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxx,
               Editable->False], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"0", ",", "1"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}]}], "+", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["gxy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "2"}], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["T212",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T212,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}], "+", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], ")"}]}]}], 
          RowBox[{"2", " ", 
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], "2"], "-", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}]}], ")"}]}]]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         FractionBox[
          RowBox[{
           RowBox[{
            RowBox[{"-", "2"}], " ", 
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["gxy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], "2"], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["T112",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`T112,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}]}], "-", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["gyy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gyy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["gxx",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxx,
               Editable->False], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"0", ",", "1"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}]}], "+", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["gxy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "2"}], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["T212",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T212,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}], "+", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], ")"}]}]}], 
          RowBox[{"2", " ", 
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], "2"], "-", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}]}], ")"}]}]], ",", 
         FractionBox[
          RowBox[{
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["gxy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "2"}], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["T112",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T112,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}], "+", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"0", ",", "1"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], ")"}]}], "+", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["gyy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gyy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "2"}], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["T212",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T212,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}], "-", 
              RowBox[{"2", " ", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], 
                 TagBox[
                  RowBox[{"(", 
                   RowBox[{"0", ",", "1"}], ")"}],
                  Derivative],
                 MultilineFunction->None], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}], "+", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], ")"}]}]}], 
          RowBox[{"2", " ", 
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], "2"], "-", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}]}], ")"}]}]]}], "}"}]}], "}"}], 
     ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         FractionBox[
          RowBox[{
           RowBox[{"2", " ", 
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["gxx",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxx,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], "2"], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["T112",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`T112,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}]}], "+", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["gxy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["gxx",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxx,
               Editable->False], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"1", ",", "0"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}]}], "+", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["gxx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["T212",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T212,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}], "+", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"0", ",", "1"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], "-", 
              RowBox[{"2", " ", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], 
                 TagBox[
                  RowBox[{"(", 
                   RowBox[{"1", ",", "0"}], ")"}],
                  Derivative],
                 MultilineFunction->None], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}]}], ")"}]}]}], 
          RowBox[{"2", " ", 
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], "2"], "-", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}]}], ")"}]}]], ",", 
         FractionBox[
          RowBox[{
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["gxy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["T212",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T212,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}], "+", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"0", ",", "1"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], ")"}]}], "+", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["gxx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["T112",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T112,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}], "-", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], ")"}]}]}], 
          RowBox[{"2", " ", 
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], "2"], "-", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}]}], ")"}]}]]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         FractionBox[
          RowBox[{
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["gxy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["gxx",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxx,
               Editable->False], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"0", ",", "1"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}]}], "+", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["gxx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["T112",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T112,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}], "+", 
              RowBox[{"2", " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["T212",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T212,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}], "-", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], ")"}]}]}], 
          RowBox[{"2", " ", 
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], "2"], "-", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}]}], ")"}]}]], ",", 
         FractionBox[
          RowBox[{
           RowBox[{"2", " ", 
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["gxy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], "2"], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["T112",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`T112,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}]}], "-", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["gxx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["gyy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gyy,
               Editable->False], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"0", ",", "1"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}]}], "+", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["gxy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["T212",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T212,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}], "+", 
              RowBox[{"2", " ", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], 
                 TagBox[
                  RowBox[{"(", 
                   RowBox[{"0", ",", "1"}], ")"}],
                  Derivative],
                 MultilineFunction->None], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}], "-", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], ")"}]}]}], 
          RowBox[{"2", " ", 
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], "2"], "-", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}]}], ")"}]}]]}], "}"}]}], "}"}]}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"gen", ",", 
     RowBox[{"-", "gen"}], ",", 
     RowBox[{"-", "gen"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.7286764988200397`*^9, 3.728753585704132*^9, 
  3.728843700653661*^9, 3.730956300734391*^9, 3.731118590620656*^9, 
  3.731119159842699*^9, 3.733368324919039*^9, 3.733368371342729*^9, 
  3.794517962310359*^9},
 CellLabel->"Out[372]=",
 CellID->1408420405,ExpressionUUID->"fa34b121-9e26-4ec3-a9d4-147a94ac9f83"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Its antisymmetric part (actually twice it) is precisely given by the torsion:\
\
\>", "MathCaption",
 CellID->606995413,ExpressionUUID->"0b53b27d-9e1a-473a-aed8-9a71e411d651"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"christoffel", "[", 
     RowBox[{"a", ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}]}], "]"}], "-", 
    RowBox[{"christoffel", "[", 
     RowBox[{"a", ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "b"}]}], "]"}]}], "\[Equal]", 
   RowBox[{"torsion", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}]}], "//", "Simplify"}]], "Input",
 CellLabel->"In[373]:=",
 CellID->1224583041,ExpressionUUID->"0b270d5d-600a-4688-b522-ff0e5f63d89c"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.7286764989474792`*^9, 3.7287535858932867`*^9, 
  3.72884370321772*^9, 3.730956300876128*^9, 3.731118601710676*^9, 
  3.7311191659680033`*^9, 3.733368326357931*^9, 3.7333683716249647`*^9, 
  3.794517962400075*^9},
 CellLabel->"Out[373]=",
 CellID->1756989601,ExpressionUUID->"347b4a88-36d0-4189-a50c-f6252826648c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The Christoffel tensor relating the covariant derivative with torsion and the \
Levi-Civita connection is proportional to the torsion tensor:\
\>", "MathCaption",
 CellChangeTimes->{{3.731118664354686*^9, 3.731118700225721*^9}},
 CellID->1987924259,ExpressionUUID->"e87939c1-7354-4f3f-b660-6ae13d778dbd"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"chris", "=", 
  RowBox[{
   RowBox[{"Christoffel", "[", 
    RowBox[{"cd", ",", "lc"}], "]"}], "//", "Simplify"}]}]], "Input",
 CellChangeTimes->{{3.731118622591895*^9, 3.731118648377554*^9}, {
  3.73111893295452*^9, 3.731118934320038*^9}},
 CellLabel->"In[374]:=",
 CellID->1367457052,ExpressionUUID->"ff5f5cd3-6d7e-4a5a-8d63-02b4f9e70e22"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", 
          FractionBox[
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["gxy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["T112",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T112,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}], "+", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["T212",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T212,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}]}], ")"}]}], 
           RowBox[{
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["gxy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], "2"], "-", 
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox["gxx",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxx,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["gyy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gyy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}]}]}]]}], ",", 
         FractionBox[
          RowBox[{
           RowBox[{
            InterpretationBox[
             StyleBox["gyy",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`gyy,
             Editable->False], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["x",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`x[],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["y",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`y[],
              Editable->False]}], "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["T112",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`T112,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["T212",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`T212,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}]}], ")"}]}], 
          RowBox[{
           RowBox[{"-", 
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["gxy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], "2"]}], "+", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["gxx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["gyy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gyy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}]}]}]]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", 
          FractionBox[
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["gxy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["T112",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T112,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}], "+", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["T212",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T212,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}]}], ")"}]}], 
           RowBox[{
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["gxy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], "2"], "-", 
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox["gxx",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxx,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["gyy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gyy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}]}]}]]}], ",", 
         FractionBox[
          RowBox[{
           RowBox[{
            InterpretationBox[
             StyleBox["gyy",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`gyy,
             Editable->False], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["x",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`x[],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["y",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`y[],
              Editable->False]}], "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["T112",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`T112,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["T212",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`T212,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}]}], ")"}]}], 
          RowBox[{
           RowBox[{"-", 
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["gxy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], "2"]}], "+", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["gxx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["gyy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gyy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}]}]}]]}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         FractionBox[
          RowBox[{
           RowBox[{
            InterpretationBox[
             StyleBox["gxx",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`gxx,
             Editable->False], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["x",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`x[],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["y",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`y[],
              Editable->False]}], "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["T112",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`T112,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["T212",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`T212,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}]}], ")"}]}], 
          RowBox[{
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["gxy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], "2"], "-", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["gxx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["gyy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gyy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}]}]}]], ",", 
         FractionBox[
          RowBox[{
           RowBox[{
            InterpretationBox[
             StyleBox["gxy",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`gxy,
             Editable->False], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["x",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`x[],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["y",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`y[],
              Editable->False]}], "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["T112",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`T112,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["T212",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`T212,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}]}], ")"}]}], 
          RowBox[{
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["gxy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], "2"], "-", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["gxx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["gyy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gyy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}]}]}]]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         FractionBox[
          RowBox[{
           RowBox[{
            InterpretationBox[
             StyleBox["gxx",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`gxx,
             Editable->False], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["x",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`x[],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["y",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`y[],
              Editable->False]}], "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["T112",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`T112,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["T212",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`T212,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}]}], ")"}]}], 
          RowBox[{
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["gxy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], "2"], "-", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["gxx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["gyy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gyy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}]}]}]], ",", 
         FractionBox[
          RowBox[{
           RowBox[{
            InterpretationBox[
             StyleBox["gxy",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`gxy,
             Editable->False], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["x",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`x[],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["y",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`y[],
              Editable->False]}], "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["T112",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`T112,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["T212",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`T212,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}]}], ")"}]}], 
          RowBox[{
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["gxy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], "2"], "-", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["gxx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["gyy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gyy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}]}]}]]}], "}"}]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"gen", ",", 
     RowBox[{"-", "gen"}], ",", 
     RowBox[{"-", "gen"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{{3.731118630719843*^9, 3.7311186490191593`*^9}, 
   3.7311191846302013`*^9, 3.733368327509028*^9, 3.733368371809288*^9, 
   3.7945179624926434`*^9},
 CellLabel->"Out[374]=",
 CellID->1824077757,ExpressionUUID->"643dec68-cf47-4ee3-a451-446e8076613e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Its antisymmetric part is the original torsion tensor:", "MathCaption",
 CellID->280631018,ExpressionUUID->"e816dcce-a1ba-48ff-9db3-530a9ca9a4b2"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"chris", "[", 
     RowBox[{"a", ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}]}], "]"}], "-", 
    RowBox[{"chris", "[", 
     RowBox[{"a", ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "b"}]}], "]"}]}], "\[Equal]", 
   RowBox[{"torsion", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}]}], "//", "Simplify"}]], "Input",
 CellLabel->"In[375]:=",
 CellID->691430714,ExpressionUUID->"789d7b07-1535-422a-8d71-cb112955058b"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728676517094144*^9, 3.728753597536083*^9, 
  3.728843723967806*^9, 3.730956312427054*^9, 3.7311192219579067`*^9, 
  3.733368328992312*^9, 3.733368371946389*^9, 3.794517962543272*^9},
 CellLabel->"Out[375]=",
 CellID->1945800974,ExpressionUUID->"14e43d40-0f30-4f1e-ba34-a6aa2990ce04"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Its symmetric part is a complicated object:", "MathCaption",
 CellID->1661744684,ExpressionUUID->"a2a56dca-4f7b-411a-ad9b-6a59e681b65f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"chris", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "+", 
   RowBox[{"chris", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "b"}]}], "]"}]}], "//", "Simplify"}]], "Input",
 CellLabel->"In[376]:=",
 CellID->918074753,ExpressionUUID->"5227ff79-26c9-46da-a029-9c32f20b699b"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {GridBox[{
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]}]],
              Annotation[#, {1, 1, 1} -> (-2) $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]}]],
              Annotation[#, {1, 1, 2} -> -($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]}]],
              Annotation[#, {1, 2, 1} -> -($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                  RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
              Annotation[#, {1, 2, 2} -> 2 $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] (-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                  RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
              Annotation[#, {2, 1, 1} -> 2 $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                  RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
              Annotation[#, {2, 1, 2} -> ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                  RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
              Annotation[#, {2, 2, 1} -> ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                  RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
              Annotation[#, {2, 2, 2} -> 2 $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a", " ", " "},
         {" ", "b", "c"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{{(-2) $CellContext`gxy[
        $CellContext`x[], 
        $CellContext`y[]] ($CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[]]^2 - $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[]] $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[]])^(-1) ($CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[]] $CellContext`T112[
          $CellContext`x[], 
          $CellContext`y[]] + $CellContext`gxy[
          $CellContext`x[], 
          $CellContext`y[]] $CellContext`T212[
          $CellContext`x[], 
          $CellContext`y[]]), -($CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[]]^2 - $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[]] $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[]])^(-1) ($CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[]]^2 $CellContext`T112[
          $CellContext`x[], 
          $CellContext`y[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[]] $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[]] $CellContext`T112[
          $CellContext`x[], 
          $CellContext`y[]] + 2 $CellContext`gxy[
          $CellContext`x[], 
          $CellContext`y[]] $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[]] $CellContext`T212[
          $CellContext`x[], 
          $CellContext`y[]])}, {-($CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[]]^2 - $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[]] $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[]])^(-1) ($CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[]]^2 $CellContext`T112[
          $CellContext`x[], 
          $CellContext`y[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[]] $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[]] $CellContext`T112[
          $CellContext`x[], 
          $CellContext`y[]] + 2 $CellContext`gxy[
          $CellContext`x[], 
          $CellContext`y[]] $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[]] $CellContext`T212[
          $CellContext`x[], 
          $CellContext`y[]]), 2 $CellContext`gyy[
        $CellContext`x[], 
        $CellContext`y[]] (-$CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[]]^2 + $CellContext`gxx[
           $CellContext`x[], 
           $CellContext`y[]] $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[]])^(-1) ($CellContext`gxy[
          $CellContext`x[], 
          $CellContext`y[]] $CellContext`T112[
          $CellContext`x[], 
          $CellContext`y[]] + $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[]] $CellContext`T212[
          $CellContext`x[], 
          $CellContext`y[]])}}, {{2 $CellContext`gxx[
        $CellContext`x[], 
        $CellContext`y[]] ($CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[]]^2 - $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[]] $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[]])^(-1) ($CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[]] $CellContext`T112[
          $CellContext`x[], 
          $CellContext`y[]] + $CellContext`gxy[
          $CellContext`x[], 
          $CellContext`y[]] $CellContext`T212[
          $CellContext`x[], 
          $CellContext`y[]]), ($CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[]]^2 - $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[]] $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[]])^(-1) ($CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[]]^2 $CellContext`T212[
          $CellContext`x[], 
          $CellContext`y[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[]] (2 $CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[]] $CellContext`T112[
            $CellContext`x[], 
            $CellContext`y[]] + $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[]] $CellContext`T212[
            $CellContext`x[], 
            $CellContext`y[]]))}, {($CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[]]^2 - $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[]] $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[]])^(-1) ($CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[]]^2 $CellContext`T212[
          $CellContext`x[], 
          $CellContext`y[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[]] (2 $CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[]] $CellContext`T112[
            $CellContext`x[], 
            $CellContext`y[]] + $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[]] $CellContext`T212[
            $CellContext`x[], 
            $CellContext`y[]])), 2 $CellContext`gxy[
        $CellContext`x[], 
        $CellContext`y[]] ($CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[]]^2 - $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[]] $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[]])^(-1) ($CellContext`gxy[
          $CellContext`x[], 
          $CellContext`y[]] $CellContext`T112[
          $CellContext`x[], 
          $CellContext`y[]] + $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[]] $CellContext`T212[
          $CellContext`x[], 
          $CellContext`y[]])}}}, {$CellContext`gen, -$CellContext`gen, \
-$CellContext`gen}, 0][$CellContext`a, -$CellContext`b, -$CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{3.728676517434057*^9, 3.728753597734448*^9, 
  3.728843724428856*^9, 3.730956312571793*^9, 3.7311192267264767`*^9, 
  3.733368330169147*^9, 3.733368372094673*^9, 3.794517962622118*^9},
 CellLabel->"Out[376]=",
 CellID->1195993502,ExpressionUUID->"c0ad7ec7-df2b-471c-8630-7dbf194d1b40"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Head", "[", "%", "]"}]], "Input",
 CellLabel->"In[377]:=",
 CellID->772457763,ExpressionUUID->"24757151-2c51-47a8-b048-e8bb4742e127"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", 
          FractionBox[
           RowBox[{"2", " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["gxy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["T112",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T112,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}], "+", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["T212",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T212,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}]}], ")"}]}], 
           RowBox[{
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["gxy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], "2"], "-", 
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox["gxx",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxx,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["gyy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gyy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}]}]}]]}], ",", 
         RowBox[{"-", 
          FractionBox[
           RowBox[{
            RowBox[{
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], "2"], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["T112",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`T112,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}]}], "+", 
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox["gxx",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxx,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["gyy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gyy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["T112",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`T112,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}]}], "+", 
            RowBox[{"2", " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["gxy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["gyy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gyy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["T212",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`T212,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}]}]}], 
           RowBox[{
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["gxy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], "2"], "-", 
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox["gxx",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxx,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["gyy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gyy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}]}]}]]}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", 
          FractionBox[
           RowBox[{
            RowBox[{
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], "2"], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["T112",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`T112,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}]}], "+", 
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox["gxx",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxx,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["gyy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gyy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["T112",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`T112,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}]}], "+", 
            RowBox[{"2", " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["gxy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["gyy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gyy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["T212",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`T212,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}]}]}], 
           RowBox[{
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["gxy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], "2"], "-", 
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox["gxx",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxx,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["gyy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gyy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}]}]}]]}], ",", 
         FractionBox[
          RowBox[{"2", " ", 
           RowBox[{
            InterpretationBox[
             StyleBox["gyy",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`gyy,
             Editable->False], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["x",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`x[],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["y",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`y[],
              Editable->False]}], "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["T112",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`T112,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["T212",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`T212,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}]}], ")"}]}], 
          RowBox[{
           RowBox[{"-", 
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["gxy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], "2"]}], "+", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["gxx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["gyy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gyy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}]}]}]]}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         FractionBox[
          RowBox[{"2", " ", 
           RowBox[{
            InterpretationBox[
             StyleBox["gxx",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`gxx,
             Editable->False], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["x",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`x[],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["y",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`y[],
              Editable->False]}], "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["T112",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`T112,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["T212",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`T212,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}]}], ")"}]}], 
          RowBox[{
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["gxy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], "2"], "-", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["gxx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["gyy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gyy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}]}]}]], ",", 
         FractionBox[
          RowBox[{
           RowBox[{
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["gxy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], "2"], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["T212",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`T212,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}]}], "+", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["gxx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["T112",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T112,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}], "+", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["T212",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T212,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}]}], ")"}]}]}], 
          RowBox[{
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["gxy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], "2"], "-", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["gxx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["gyy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gyy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}]}]}]]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         FractionBox[
          RowBox[{
           RowBox[{
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["gxy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], "2"], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["T212",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`T212,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}]}], "+", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["gxx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["T112",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T112,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}], "+", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{
                InterpretationBox[
                 StyleBox["T212",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T212,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}]}]}], ")"}]}]}], 
          RowBox[{
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["gxy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], "2"], "-", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["gxx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["gyy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gyy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}]}]}]], ",", 
         FractionBox[
          RowBox[{"2", " ", 
           RowBox[{
            InterpretationBox[
             StyleBox["gxy",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`gxy,
             Editable->False], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["x",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`x[],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["y",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`y[],
              Editable->False]}], "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["T112",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`T112,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["T212",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`T212,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}]}], ")"}]}], 
          RowBox[{
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["gxy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], "2"], "-", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["gxx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["gyy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gyy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}]}]}]]}], "}"}]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"gen", ",", 
     RowBox[{"-", "gen"}], ",", 
     RowBox[{"-", "gen"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.728676517816462*^9, 3.728753597937533*^9, 
  3.728843724705454*^9, 3.730956312701585*^9, 3.731119232845522*^9, 
  3.733368331489684*^9, 3.7333683724661713`*^9, 3.7945179626764812`*^9},
 CellLabel->"Out[377]=",
 CellID->615852194,ExpressionUUID->"9856d164-2971-43f3-b9bc-6982d3569a82"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Let us now look at the curvature tensors:", "MathCaption",
 CellChangeTimes->{{3.731118736840169*^9, 3.731118748431528*^9}},
 CellID->1579180349,ExpressionUUID->"67222cac-ebb0-4fb2-82a2-b6dfb03b535b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"riemann", "=", 
  RowBox[{"Riemann", "[", "cd", "]"}]}]], "Input",
 CellLabel->"In[378]:=",
 CellID->533732370,ExpressionUUID->"fe7d54c8-2ea6-494c-b7f2-a50fde9f313c"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0"}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           RowBox[{"-", 
            RowBox[{
             FractionBox["1", 
              RowBox[{"4", " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "-", 
                  RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}], "2"]}]], 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "+", 
                RowBox[{
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gyy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "+", 
                   RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}], "+", 
                RowBox[{
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                   RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                   RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}], "+", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}]}], 
                  ")"}]}], "+", 
                RowBox[{"4", " ", 
                 SuperscriptBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "3"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "+", 
                RowBox[{
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxx,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                   SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "+", 
                   RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}]}], 
               ")"}]}]}]}], ",", 
           RowBox[{
            FractionBox["1", 
             RowBox[{"4", " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "-", 
                 RowBox[{
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}], "2"]}]], 
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox["gxx",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxx,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"2", " ", 
                SuperscriptBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxx,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "+", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["gyy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gyy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], " ", 
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "+", 
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}], "+", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{"-", "2"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                  RowBox[{"4", " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}], "+", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}]}], 
                 ")"}]}], "+", 
               RowBox[{"4", " ", 
                SuperscriptBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "3"], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "+", 
               RowBox[{"2", " ", 
                SuperscriptBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "+", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxx",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxx,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                  SuperscriptBox[
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "+", 
                  RowBox[{"4", " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}]}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"-", 
            RowBox[{
             FractionBox["1", 
              RowBox[{"4", " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "-", 
                  RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}], "2"]}]], 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "+", 
                RowBox[{
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gyy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "+", 
                   RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}], "+", 
                RowBox[{
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                   RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                   RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}], "+", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}]}], 
                  ")"}]}], "+", 
                RowBox[{"4", " ", 
                 SuperscriptBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "3"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "+", 
                RowBox[{
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxx,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                   SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "+", 
                   RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}]}], 
               ")"}]}]}]}], ",", 
           RowBox[{
            FractionBox["1", 
             RowBox[{"4", " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "-", 
                 RowBox[{
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}], "2"]}]], 
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox["gxy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"2", " ", 
                SuperscriptBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxx,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "+", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["gyy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gyy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], " ", 
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "+", 
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}], "+", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{"-", "2"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                  RowBox[{"4", " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}], "+", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}]}], 
                 ")"}]}], "+", 
               RowBox[{"4", " ", 
                SuperscriptBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "3"], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "+", 
               RowBox[{"2", " ", 
                SuperscriptBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "+", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxx",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxx,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                  SuperscriptBox[
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "+", 
                  RowBox[{"4", " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}]}]}], "}"}]}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           RowBox[{
            FractionBox["1", 
             RowBox[{"4", " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "-", 
                 RowBox[{
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}], "2"]}]], 
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox["gxy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"2", " ", 
                SuperscriptBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxx,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "+", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["gyy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gyy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], " ", 
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "+", 
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}], "+", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{"-", "2"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                  RowBox[{"4", " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}], "+", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}]}], 
                 ")"}]}], "+", 
               RowBox[{"4", " ", 
                SuperscriptBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "3"], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "+", 
               RowBox[{"2", " ", 
                SuperscriptBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "+", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxx",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxx,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                  SuperscriptBox[
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "+", 
                  RowBox[{"4", " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}]}], ",", 
           RowBox[{"-", 
            RowBox[{
             FractionBox["1", 
              RowBox[{"4", " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "-", 
                  RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}], "2"]}]], 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "+", 
                RowBox[{
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gyy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "+", 
                   RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}], "+", 
                RowBox[{
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                   RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                   RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}], "+", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}]}], 
                  ")"}]}], "+", 
                RowBox[{"4", " ", 
                 SuperscriptBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "3"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "+", 
                RowBox[{
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxx,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                   SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "+", 
                   RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}]}], 
               ")"}]}]}]}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{
            FractionBox["1", 
             RowBox[{"4", " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "-", 
                 RowBox[{
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}], "2"]}]], 
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox["gyy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gyy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"2", " ", 
                SuperscriptBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxx,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "+", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["gyy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gyy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], " ", 
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "+", 
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}], "+", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{"-", "2"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                  RowBox[{"4", " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}], "+", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}]}], 
                 ")"}]}], "+", 
               RowBox[{"4", " ", 
                SuperscriptBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "3"], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "+", 
               RowBox[{"2", " ", 
                SuperscriptBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "+", 
               RowBox[{
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxx",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxx,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                  SuperscriptBox[
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "+", 
                  RowBox[{"4", " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                  RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}]}], ",", 
           RowBox[{"-", 
            RowBox[{
             FractionBox["1", 
              RowBox[{"4", " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "-", 
                  RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}], "2"]}]], 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "+", 
                RowBox[{
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gyy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "+", 
                   RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}], "+", 
                RowBox[{
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                   RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                   RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}], "+", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}]}], 
                  ")"}]}], "+", 
                RowBox[{"4", " ", 
                 SuperscriptBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "3"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "+", 
                RowBox[{
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxx,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                   SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "+", 
                   RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}]}], 
               ")"}]}]}]}]}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0"}], "}"}]}], "}"}]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "gen"}], ",", 
     RowBox[{"-", "gen"}], ",", 
     RowBox[{"-", "gen"}], ",", "gen"}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.7286765001901083`*^9, 3.7287535871293507`*^9, 
  3.728843704863583*^9, 3.73095630216103*^9, 3.731118751294403*^9, 
  3.731119260500442*^9, 3.733368333951373*^9, 3.733368373635429*^9, 
  3.794517963684691*^9},
 CellLabel->"Out[378]=",
 CellID->1904129665,ExpressionUUID->"f0eda76f-267f-4953-b6d0-41afb7f8e069"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
To analyze the symmetries of the result we need the Riemann with four indices \
down:\
\>", "MathCaption",
 CellID->880919305,ExpressionUUID->"8c5dab29-33f0-4e20-87ff-eb47b43154e3"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"riemanndown", "=", 
  RowBox[{"RiemannDown", "[", "cd", "]"}]}]], "Input",
 CellLabel->"In[379]:=",
 CellID->858496717,ExpressionUUID->"b366e89a-7541-48ab-b24a-2ebff341af05"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0"}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", 
           RowBox[{
            FractionBox["1", 
             RowBox[{"4", " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "-", 
                RowBox[{
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxx,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gyy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}]], 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               SuperscriptBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxx",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxx,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "2"}], " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                 RowBox[{"4", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "+", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", 
                  SuperscriptBox[
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"]}], "+", 
                 RowBox[{
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}], "+", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"2", " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                 RowBox[{
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                 RowBox[{"4", " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                 RowBox[{
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}]}], 
                ")"}]}], "-", 
              RowBox[{"4", " ", 
               SuperscriptBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "-", 
              RowBox[{"2", " ", 
               SuperscriptBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "-", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                 SuperscriptBox[
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "+", 
                 RowBox[{"4", " ", 
                  SuperscriptBox[
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}]}], 
             ")"}]}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{
            FractionBox["1", 
             RowBox[{"4", " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "-", 
                RowBox[{
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxx,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gyy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}]], 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", 
               SuperscriptBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxx",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxx,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{
                   InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "+", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "+", 
                 RowBox[{
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}], "+", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "2"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                 RowBox[{
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                 RowBox[{"4", " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                 RowBox[{
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}], "+", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}]}], 
                ")"}]}], "+", 
              RowBox[{"4", " ", 
               SuperscriptBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "+", 
              RowBox[{"2", " ", 
               SuperscriptBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "+", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                 SuperscriptBox[
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "+", 
                 RowBox[{"4", " ", 
                  SuperscriptBox[
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}]}], 
             ")"}]}], ",", "0"}], "}"}]}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", 
           RowBox[{
            FractionBox["1", 
             RowBox[{"4", " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "-", 
                RowBox[{
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxx,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gyy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}]], 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", 
               SuperscriptBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxx",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxx,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{
                   InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "+", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "+", 
                 RowBox[{
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}], "+", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "2"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                 RowBox[{
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                 RowBox[{"4", " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                 RowBox[{
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}], "+", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}]}], 
                ")"}]}], "+", 
              RowBox[{"4", " ", 
               SuperscriptBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "+", 
              RowBox[{"2", " ", 
               SuperscriptBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "+", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                 SuperscriptBox[
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "+", 
                 RowBox[{"4", " ", 
                  SuperscriptBox[
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}]}], 
             ")"}]}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{
            FractionBox["1", 
             RowBox[{"4", " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "-", 
                RowBox[{
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gxx,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`gyy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}]], 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               SuperscriptBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxx",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxx,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "2"}], " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                 RowBox[{"4", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "+", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", 
                  SuperscriptBox[
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"]}], "+", 
                 RowBox[{
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}], "+", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"2", " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                 RowBox[{
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                 RowBox[{"4", " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                 RowBox[{
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}]}], 
                ")"}]}], "-", 
              RowBox[{"4", " ", 
               SuperscriptBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "-", 
              RowBox[{"2", " ", 
               SuperscriptBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}], "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "-", 
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                 SuperscriptBox[
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], "+", 
                 RowBox[{"4", " ", 
                  SuperscriptBox[
                   RowBox[{
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "2"], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T112,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxx",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxx,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`T212,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gxy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gxy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["gyy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`gyy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}]}], 
             ")"}]}], ",", "0"}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0"}], "}"}]}], "}"}]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "gen"}], ",", 
     RowBox[{"-", "gen"}], ",", 
     RowBox[{"-", "gen"}], ",", 
     RowBox[{"-", "gen"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.7286765018987617`*^9, 3.728753587516198*^9, 
  3.728843705974324*^9, 3.730956302918606*^9, 3.731118764290413*^9, 
  3.731119264924243*^9, 3.733368335453237*^9, 3.733368374237208*^9, 
  3.794517964002468*^9},
 CellLabel->"Out[379]=",
 CellID->905972061,ExpressionUUID->"eec44119-9fd7-40d4-b428-30ef3dee79b0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Symmetry is complete in this 2d case:", "MathCaption",
 CellID->606341460,ExpressionUUID->"e4c1b74b-9a13-4304-be65-55ff5e0bcb17"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SymmetryGroupOfTensor", "[", "riemann", "]"}]], "Input",
 CellLabel->"In[380]:=",
 CellID->971013205,ExpressionUUID->"a9cd576f-e219-4b42-9d59-e6eef9d9dd12"],

Cell[BoxData[
 RowBox[{"StrongGenSet", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "2"}], "}"}], ",", 
   RowBox[{"GenSet", "[", 
    InterpretationBox[
     StyleBox[
      RowBox[{"-", 
       RowBox[{"(", "1", ",", "2", ")"}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     -xAct`xPerm`Cycles[{1, 2}],
     Editable->False], "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728676506026121*^9, 3.728753587655241*^9, 
  3.728843709442172*^9, 3.730956303238474*^9, 3.731118767839576*^9, 
  3.731119266911165*^9, 3.7333683368863573`*^9, 3.733368374472212*^9, 
  3.794517964084609*^9},
 CellLabel->"Out[380]=",
 CellID->1013371997,ExpressionUUID->"49b170c3-db3f-4a2e-bbdb-87189aba8aea"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SymmetryGroupOfTensor", "[", "riemanndown", "]"}]], "Input",
 CellLabel->"In[381]:=",
 CellID->522926982,ExpressionUUID->"7ed1df38-e3d8-471e-ba29-462fcfcbb240"],

Cell[BoxData[
 RowBox[{"StrongGenSet", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"3", ",", "1"}], "}"}], ",", 
   RowBox[{"GenSet", "[", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"-", 
        RowBox[{"(", "3", ",", "4", ")"}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      -xAct`xPerm`Cycles[{3, 4}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[
       RowBox[{"-", 
        RowBox[{"(", "1", ",", "2", ")"}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      -xAct`xPerm`Cycles[{1, 2}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"(", "1", ",", "3", ")"}], 
        RowBox[{"(", "2", ",", "4", ")"}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xPerm`Cycles[{1, 3}, {2, 4}],
      Editable->False]}], "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728676506975709*^9, 3.728753587791044*^9, 
  3.728843711462768*^9, 3.730956303451198*^9, 3.73111876882552*^9, 
  3.7311192676705513`*^9, 3.733368337539291*^9, 3.733368374732679*^9, 
  3.79451796415355*^9},
 CellLabel->"Out[381]=",
 CellID->50349222,ExpressionUUID->"6f7634fe-b0f8-4940-a550-f28f774683e9"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Now compute everything. Most of the time is spent in the simplification of \
the results:\
\>", "MathCaption",
 CellChangeTimes->{{3.731118791598076*^9, 3.731118803533884*^9}},
 CellID->1229780792,ExpressionUUID->"3299ca2e-f0ce-4b7f-9bc1-d41e5efc1d37"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"MetricCompute", "[", 
   RowBox[{"metric", ",", "torsion", ",", "gen", ",", "All", ",", 
    RowBox[{"Parallelize", "\[Rule]", "True"}]}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[382]:=",
 CellID->1543152416,ExpressionUUID->"0c390ac8-0815-4b1e-be0c-e9fde1d06685"],

Cell[BoxData["$Aborted"], "Output",
 CellChangeTimes->{3.7286765154996157`*^9, 3.728753595664665*^9, 
  3.728843719409333*^9, 3.730956311261489*^9, 3.731118786615004*^9, 
  3.731119282909293*^9, 3.733368345826952*^9, 3.733368384575733*^9, 
  3.794517970179813*^9},
 CellLabel->"Out[382]=",
 CellID->110817236,ExpressionUUID->"7d3ce317-0136-4b1e-8eb2-078bc3650e17"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["There is Ricci-scalar curvature:", "MathCaption",
 CellID->1047379202,ExpressionUUID->"d25eba21-85c9-4f02-9066-b08cbdbc1bcd"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"RicciScalar", "[", "cd", "]"}], "[", "]"}]], "Input",
 CellLabel->"In[30]:=",
 CellID->378793555,ExpressionUUID->"85e79de6-05bc-4d85-8b71-6039aca6daae"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{
    RowBox[{"2", " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["gxx",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`gxx,
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["y",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`y[],
         Editable->False]}], "]"}], "2"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["T112",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T112,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}], " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["gyy",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gyy,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"0", ",", "1"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}]}], "-", 
       RowBox[{"2", " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["gyy",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`gyy,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}], " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["T112",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`T112,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"0", ",", "1"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}]}]}], ")"}]}], "+", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["gyy",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`gyy,
       Editable->False], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["y",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`y[],
        Editable->False]}], "]"}], " ", 
     RowBox[{"(", 
      RowBox[{
       SuperscriptBox[
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["gxx",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gxx,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"0", ",", "1"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}], "2"], "+", 
       RowBox[{
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["gxx",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gxx,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"1", ",", "0"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "2"}], " ", 
           RowBox[{
            InterpretationBox[
             StyleBox["gyy",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`gyy,
             Editable->False], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["x",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`x[],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["y",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`y[],
              Editable->False]}], "]"}], " ", 
           RowBox[{
            InterpretationBox[
             StyleBox["T212",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`T212,
             Editable->False], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["x",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`x[],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["y",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`y[],
              Editable->False]}], "]"}]}], "-", 
          RowBox[{"2", " ", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["gxy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxy,
              Editable->False], 
             TagBox[
              RowBox[{"(", 
               RowBox[{"0", ",", "1"}], ")"}],
              Derivative],
             MultilineFunction->None], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["x",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`x[],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["y",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`y[],
              Editable->False]}], "]"}]}], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["gyy",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`gyy,
             Editable->False], 
            TagBox[
             RowBox[{"(", 
              RowBox[{"1", ",", "0"}], ")"}],
             Derivative],
            MultilineFunction->None], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["y",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`y[],
             Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}], "+", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["gxy",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`gxy,
       Editable->False], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["y",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`y[],
        Editable->False]}], "]"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "2"}], " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["gxx",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gxx,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"0", ",", "1"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}], " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["gxy",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gxy,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"0", ",", "1"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}]}], "+", 
       RowBox[{
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["gyy",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gyy,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"0", ",", "1"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}], " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["gxx",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gxx,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"1", ",", "0"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}]}], "+", 
       RowBox[{"4", " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["gxy",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gxy,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"0", ",", "1"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}], " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["gxy",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gxy,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"1", ",", "0"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}]}], "-", 
       RowBox[{
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["gxx",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gxx,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"0", ",", "1"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}], " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["gyy",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gyy,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"1", ",", "0"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}]}], "-", 
       RowBox[{"2", " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["gxy",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gxy,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"1", ",", "0"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}], " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["gyy",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gyy,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"1", ",", "0"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}]}], "-", 
       RowBox[{"2", " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["gxx",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`gxx,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", 
            RowBox[{
             InterpretationBox[
              StyleBox["T212",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`T212,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}]}], " ", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["gyy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gyy,
              Editable->False], 
             TagBox[
              RowBox[{"(", 
               RowBox[{"0", ",", "1"}], ")"}],
              Derivative],
             MultilineFunction->None], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["x",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`x[],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["y",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`y[],
              Editable->False]}], "]"}]}], "+", 
          RowBox[{
           RowBox[{
            InterpretationBox[
             StyleBox["T112",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`T112,
             Editable->False], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["x",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`x[],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["y",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`y[],
              Editable->False]}], "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"0", ",", "1"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], "+", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"1", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}], "+", 
       RowBox[{"2", " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["gyy",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`gyy,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", 
            RowBox[{
             InterpretationBox[
              StyleBox["T112",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`T112,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}]}], " ", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["gxx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxx,
              Editable->False], 
             TagBox[
              RowBox[{"(", 
               RowBox[{"1", ",", "0"}], ")"}],
              Derivative],
             MultilineFunction->None], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["x",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`x[],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["y",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`y[],
              Editable->False]}], "]"}]}], "+", 
          RowBox[{
           RowBox[{
            InterpretationBox[
             StyleBox["T212",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`T212,
             Editable->False], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["x",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`x[],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["y",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`y[],
              Editable->False]}], "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"0", ",", "1"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], "+", 
             RowBox[{"2", " ", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}]}], ")"}]}], "-", 
          RowBox[{"2", " ", 
           RowBox[{
            InterpretationBox[
             StyleBox["gxx",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`gxx,
             Editable->False], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["x",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`x[],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["y",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`y[],
              Editable->False]}], "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["T212",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`T212,
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"0", ",", "1"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], "-", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["T112",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`T112,
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"1", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}]}], ")"}]}], 
    "+", 
    RowBox[{"4", " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["gxy",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`gxy,
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["y",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`y[],
         Editable->False]}], "]"}], "3"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["T212",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T212,
          Editable->False], 
         TagBox[
          RowBox[{"(", 
           RowBox[{"0", ",", "1"}], ")"}],
          Derivative],
         MultilineFunction->None], "[", 
        RowBox[{
         InterpretationBox[
          StyleBox["x",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["y",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`y[],
          Editable->False]}], "]"}], "-", 
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["T112",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T112,
          Editable->False], 
         TagBox[
          RowBox[{"(", 
           RowBox[{"1", ",", "0"}], ")"}],
          Derivative],
         MultilineFunction->None], "[", 
        RowBox[{
         InterpretationBox[
          StyleBox["x",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["y",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`y[],
          Editable->False]}], "]"}]}], ")"}]}], "+", 
    RowBox[{"2", " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["gxy",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`gxy,
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["y",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`y[],
         Editable->False]}], "]"}], "2"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"2", " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["T112",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T112,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}], " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["gxx",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gxx,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"0", ",", "1"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}]}], "+", 
       RowBox[{"2", " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["gxx",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`gxx,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}], " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["T112",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`T112,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"0", ",", "1"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}]}], "+", 
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["gxx",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`gxx,
          Editable->False], 
         TagBox[
          RowBox[{"(", 
           RowBox[{"0", ",", "2"}], ")"}],
          Derivative],
         MultilineFunction->None], "[", 
        RowBox[{
         InterpretationBox[
          StyleBox["x",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["y",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`y[],
          Editable->False]}], "]"}], "-", 
       RowBox[{"2", " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["T212",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T212,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}], " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["gyy",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gyy,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"1", ",", "0"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}]}], "-", 
       RowBox[{"2", " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["gyy",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`gyy,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}], " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["T212",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`T212,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"1", ",", "0"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}]}], "-", 
       RowBox[{"2", " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["gxy",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gxy,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"1", ",", "1"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}]}], "+", 
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["gyy",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`gyy,
          Editable->False], 
         TagBox[
          RowBox[{"(", 
           RowBox[{"2", ",", "0"}], ")"}],
          Derivative],
         MultilineFunction->None], "[", 
        RowBox[{
         InterpretationBox[
          StyleBox["x",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["y",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`y[],
          Editable->False]}], "]"}]}], ")"}]}], "+", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["gxx",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`gxx,
       Editable->False], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["y",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`y[],
        Editable->False]}], "]"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["gxx",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gxx,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"0", ",", "1"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}], " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["gyy",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gyy,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"0", ",", "1"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}]}], "-", 
       RowBox[{"2", " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["gyy",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gyy,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"0", ",", "1"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}], " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["gxy",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gxy,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"1", ",", "0"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}]}], "+", 
       SuperscriptBox[
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["gyy",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gyy,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"1", ",", "0"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}], "2"], "+", 
       RowBox[{"4", " ", 
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["gyy",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`gyy,
           Editable->False], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["x",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["y",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`y[],
            Editable->False]}], "]"}], "2"], " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["T212",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`T212,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"1", ",", "0"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}]}], "-", 
       RowBox[{"2", " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["gyy",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`gyy,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["gxx",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`gxx,
             Editable->False], 
            TagBox[
             RowBox[{"(", 
              RowBox[{"0", ",", "2"}], ")"}],
             Derivative],
            MultilineFunction->None], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["y",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`y[],
             Editable->False]}], "]"}], "+", 
          RowBox[{
           RowBox[{
            InterpretationBox[
             StyleBox["T112",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`T112,
             Editable->False], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["x",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`x[],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["y",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`y[],
              Editable->False]}], "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"0", ",", "1"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}], "-", 
             RowBox[{"2", " ", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}]}], ")"}]}], "+", 
          RowBox[{
           RowBox[{
            InterpretationBox[
             StyleBox["T212",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`T212,
             Editable->False], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["x",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`x[],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["y",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`y[],
              Editable->False]}], "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"0", ",", "1"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], "-", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"1", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}]}], ")"}]}], "-", 
          RowBox[{"2", " ", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["gxy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxy,
              Editable->False], 
             TagBox[
              RowBox[{"(", 
               RowBox[{"1", ",", "1"}], ")"}],
              Derivative],
             MultilineFunction->None], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["x",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`x[],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["y",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`y[],
              Editable->False]}], "]"}]}], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["gyy",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`gyy,
             Editable->False], 
            TagBox[
             RowBox[{"(", 
              RowBox[{"2", ",", "0"}], ")"}],
             Derivative],
            MultilineFunction->None], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["y",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`y[],
             Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}]}], ")"}], "/", 
  RowBox[{"(", 
   RowBox[{"2", " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       SuperscriptBox[
        RowBox[{
         InterpretationBox[
          StyleBox["gxy",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`gxy,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}], "2"], "-", 
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["gxx",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`gxx,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}], " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["gyy",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`gyy,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}]}]}], ")"}], "2"]}], ")"}]}]], "Output",
 CellChangeTimes->{3.728676515708261*^9, 3.728753595851736*^9, 
  3.728843719749156*^9, 3.730956311367432*^9, 3.731118815158545*^9, 
  3.731119286017075*^9, 3.733368384901527*^9},
 CellLabel->"Out[30]=",
 CellID->601619684,ExpressionUUID->"1bbfec0c-9c1e-4e64-99c2-896f149fb185"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["But there is no Einstein tensor, even having torsion:", "MathCaption",
 CellID->832336293,ExpressionUUID->"15b5cbd7-aac9-4a2f-8718-94482cf7bf49"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Einstein", "[", "cd", "]"}], "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}]}], "]"}]], "Input",
 CellLabel->"In[31]:=",
 CellID->367599893,ExpressionUUID->"3d95121f-fced-4fef-8417-d02f848b7718"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728676515940003*^9, 3.72875359604247*^9, 
  3.728843720051957*^9, 3.730956311453001*^9, 3.731118820122959*^9, 
  3.731119289888214*^9, 3.733368385133481*^9},
 CellLabel->"Out[31]=",
 CellID->1069018912,ExpressionUUID->"24804c50-8866-4b6a-bf2f-4c1497ca40e4"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The Levi-Civita connection has different curvature:", "MathCaption",
 CellChangeTimes->{{3.731119304500415*^9, 3.731119325771699*^9}},
 CellID->606168795,ExpressionUUID->"99efe8ec-725e-4081-afd4-55b780bfca23"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"RicciScalar", "[", "lc", "]"}]], "Input",
 CellChangeTimes->{{3.7311193274437037`*^9, 3.731119329419743*^9}},
 CellLabel->"In[32]:=",
 CellID->216516197,ExpressionUUID->"65778c9e-58fd-4880-8c5c-9d3430d3182d"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox["gxx",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`gxx,
         Editable->False], "[", 
        RowBox[{
         InterpretationBox[
          StyleBox["x",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["y",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`y[],
          Editable->False]}], "]"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["gxx",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`gxx,
             Editable->False], 
            TagBox[
             RowBox[{"(", 
              RowBox[{"0", ",", "1"}], ")"}],
             Derivative],
            MultilineFunction->None], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["y",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`y[],
             Editable->False]}], "]"}], " ", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["gyy",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`gyy,
             Editable->False], 
            TagBox[
             RowBox[{"(", 
              RowBox[{"0", ",", "1"}], ")"}],
             Derivative],
            MultilineFunction->None], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["y",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`y[],
             Editable->False]}], "]"}]}], "-", 
         RowBox[{"2", " ", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["gyy",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`gyy,
             Editable->False], 
            TagBox[
             RowBox[{"(", 
              RowBox[{"0", ",", "1"}], ")"}],
             Derivative],
            MultilineFunction->None], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["y",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`y[],
             Editable->False]}], "]"}], " ", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["gxy",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`gxy,
             Editable->False], 
            TagBox[
             RowBox[{"(", 
              RowBox[{"1", ",", "0"}], ")"}],
             Derivative],
            MultilineFunction->None], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["y",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`y[],
             Editable->False]}], "]"}]}], "+", 
         SuperscriptBox[
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["gyy",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`gyy,
             Editable->False], 
            TagBox[
             RowBox[{"(", 
              RowBox[{"1", ",", "0"}], ")"}],
             Derivative],
            MultilineFunction->None], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["y",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`y[],
             Editable->False]}], "]"}], "2"]}], ")"}]}], "+", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox["gxy",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`gxy,
         Editable->False], "[", 
        RowBox[{
         InterpretationBox[
          StyleBox["x",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["y",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`y[],
          Editable->False]}], "]"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["gyy",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`gyy,
             Editable->False], 
            TagBox[
             RowBox[{"(", 
              RowBox[{"0", ",", "1"}], ")"}],
             Derivative],
            MultilineFunction->None], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["y",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`y[],
             Editable->False]}], "]"}], " ", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["gxx",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`gxx,
             Editable->False], 
            TagBox[
             RowBox[{"(", 
              RowBox[{"1", ",", "0"}], ")"}],
             Derivative],
            MultilineFunction->None], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["y",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`y[],
             Editable->False]}], "]"}]}], "+", 
         RowBox[{"2", " ", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["gxy",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`gxy,
             Editable->False], 
            TagBox[
             RowBox[{"(", 
              RowBox[{"1", ",", "0"}], ")"}],
             Derivative],
            MultilineFunction->None], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["y",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`y[],
             Editable->False]}], "]"}], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"2", " ", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"0", ",", "1"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}]}], "-", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["gyy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gyy,
               Editable->False], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"1", ",", "0"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}]}], ")"}]}], "-", 
         RowBox[{
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["gxx",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`gxx,
             Editable->False], 
            TagBox[
             RowBox[{"(", 
              RowBox[{"0", ",", "1"}], ")"}],
             Derivative],
            MultilineFunction->None], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["y",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`y[],
             Editable->False]}], "]"}], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"2", " ", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"0", ",", "1"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}]}], "+", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["gyy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gyy,
               Editable->False], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"1", ",", "0"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}], "+", 
      RowBox[{"2", " ", 
       SuperscriptBox[
        RowBox[{
         InterpretationBox[
          StyleBox["gxy",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`gxy,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False]}], "]"}], "2"], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["gxx",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`gxx,
            Editable->False], 
           TagBox[
            RowBox[{"(", 
             RowBox[{"0", ",", "2"}], ")"}],
            Derivative],
           MultilineFunction->None], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["x",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["y",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`y[],
            Editable->False]}], "]"}], "-", 
         RowBox[{"2", " ", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["gxy",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`gxy,
             Editable->False], 
            TagBox[
             RowBox[{"(", 
              RowBox[{"1", ",", "1"}], ")"}],
             Derivative],
            MultilineFunction->None], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["y",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`y[],
             Editable->False]}], "]"}]}], "+", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["gyy",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`gyy,
            Editable->False], 
           TagBox[
            RowBox[{"(", 
             RowBox[{"2", ",", "0"}], ")"}],
            Derivative],
           MultilineFunction->None], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["x",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["y",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`y[],
            Editable->False]}], "]"}]}], ")"}]}], "+", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox["gyy",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`gyy,
         Editable->False], "[", 
        RowBox[{
         InterpretationBox[
          StyleBox["x",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["y",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`y[],
          Editable->False]}], "]"}], " ", 
       RowBox[{"(", 
        RowBox[{
         SuperscriptBox[
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["gxx",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`gxx,
             Editable->False], 
            TagBox[
             RowBox[{"(", 
              RowBox[{"0", ",", "1"}], ")"}],
             Derivative],
            MultilineFunction->None], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["y",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`y[],
             Editable->False]}], "]"}], "2"], "+", 
         RowBox[{
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["gxx",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`gxx,
             Editable->False], 
            TagBox[
             RowBox[{"(", 
              RowBox[{"1", ",", "0"}], ")"}],
             Derivative],
            MultilineFunction->None], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["y",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`y[],
             Editable->False]}], "]"}], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{
             RowBox[{"-", "2"}], " ", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"0", ",", "1"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}]}], "+", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["gyy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gyy,
               Editable->False], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"1", ",", "0"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}]}], ")"}]}], "-", 
         RowBox[{"2", " ", 
          RowBox[{
           InterpretationBox[
            StyleBox["gxx",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`gxx,
            Editable->False], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["y",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`y[],
             Editable->False]}], "]"}], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["gxx",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxx,
               Editable->False], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"0", ",", "2"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], "-", 
            RowBox[{"2", " ", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"1", ",", "1"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}]}], "+", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["gyy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gyy,
               Editable->False], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"2", ",", "0"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}]}], ")"}], "/", 
    
    RowBox[{"(", 
     RowBox[{"2", " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["gxy",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`gxy,
            Editable->False], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["y",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`y[],
             Editable->False]}], "]"}], "2"], "-", 
         RowBox[{
          RowBox[{
           InterpretationBox[
            StyleBox["gxx",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`gxx,
            Editable->False], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["y",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`y[],
             Editable->False]}], "]"}], " ", 
          RowBox[{
           InterpretationBox[
            StyleBox["gyy",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`gyy,
            Editable->False], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["y",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`y[],
             Editable->False]}], "]"}]}]}], ")"}], "2"]}], ")"}]}], ",", 
   RowBox[{"{", "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.7286765167799873`*^9, 3.7287535970088*^9, 
  3.728843722500791*^9, 3.730956312121624*^9, 3.731119330420312*^9, 
  3.7333683856430397`*^9},
 CellLabel->"Out[32]=",
 CellID->1152663182,ExpressionUUID->"1320d392-6b4f-4ad6-86cc-dd0d5a67c4bf"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Full cache table:", "MathCaption",
 CellID->142785295,ExpressionUUID->"77b9b101-27c2-4e91-be0f-bebe87af4bae"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"PrintxCobaCache", "[", "]"}]], "Input",
 CellLabel->"In[33]:=",
 CellID->1053219366,ExpressionUUID->"96512aa5-0eee-4214-a1c2-2d7e73793aea"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {"1", 
     RowBox[{"\<\"ReportCompute\"\>", "[", 
      RowBox[{"\<\"ADChristoffel\"\>", "[", 
       RowBox[{
        RowBox[{"-", "1"}], ",", 
        RowBox[{"-", "1"}], ",", 
        RowBox[{"-", "1"}], ",", 
        RowBox[{"-", "1"}]}], "]"}], "]"}], GridBox[{
       {
        RowBox[{"\[LeftSkeleton]", "4", "\[RightSkeleton]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.7]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.1]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]},
    {"2", 
     RowBox[{"\<\"ReportCompute\"\>", "[", 
      RowBox[{"\<\"CDRiemann\"\>", "[", 
       RowBox[{
        RowBox[{"-", "1"}], ",", 
        RowBox[{"-", "1"}], ",", 
        RowBox[{"-", "1"}], ",", 
        RowBox[{"-", "1"}], ",", 
        RowBox[{"-", "1"}]}], "]"}], "]"}], GridBox[{
       {
        RowBox[{"\[LeftSkeleton]", "2", "\[RightSkeleton]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.7]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.1]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]},
    {"3", 
     RowBox[{"\<\"ReportCompute\"\>", "[", 
      RowBox[{"\<\"Christoffel\"\>", "[", 
       RowBox[{
        RowBox[{"-", "1"}], ",", 
        RowBox[{"-", "1"}], ",", 
        RowBox[{"-", "1"}]}], "]"}], "]"}], GridBox[{
       {
        RowBox[{"\[LeftSkeleton]", "8", "\[RightSkeleton]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.7]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.1]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]},
    {"4", 
     RowBox[{"\<\"ReportCompute\"\>", "[", 
      RowBox[{"\<\"Christoffel\"\>", "[", 
       RowBox[{"1", ",", 
        RowBox[{"-", "1"}], ",", 
        RowBox[{"-", "1"}]}], "]"}], "]"}], GridBox[{
       {
        RowBox[{"\[LeftSkeleton]", "8", "\[RightSkeleton]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.7]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.1]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]},
    {"5", 
     RowBox[{"\<\"ReportCompute\"\>", "[", 
      RowBox[{"\<\"DetMetric\"\>", "[", "]"}], "]"}], GridBox[{
       {
        RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.7]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.1]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]},
    {"6", 
     RowBox[{"\<\"ReportCompute\"\>", "[", 
      RowBox[{"\<\"DMetric\"\>", "[", 
       RowBox[{
        RowBox[{"-", "1"}], ",", 
        RowBox[{"-", "1"}], ",", 
        RowBox[{"-", "1"}]}], "]"}], "]"}], GridBox[{
       {
        RowBox[{"\[LeftSkeleton]", "6", "\[RightSkeleton]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.7]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.1]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]},
    {"7", 
     RowBox[{"\<\"ReportCompute\"\>", "[", 
      RowBox[{"\<\"Einstein\"\>", "[", 
       RowBox[{
        RowBox[{"-", "1"}], ",", 
        RowBox[{"-", "1"}]}], "]"}], "]"}], GridBox[{
       {
        RowBox[{"\[LeftSkeleton]", "4", "\[RightSkeleton]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.7]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.1]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]},
    {"8", 
     RowBox[{"\<\"ReportCompute\"\>", "[", 
      RowBox[{"\<\"Kretschmann\"\>", "[", "]"}], "]"}], GridBox[{
       {
        RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.7]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.1]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]},
    {"9", 
     RowBox[{"\<\"ReportCompute\"\>", "[", 
      RowBox[{"\<\"Metric\"\>", "[", 
       RowBox[{"1", ",", "1"}], "]"}], "]"}], GridBox[{
       {
        RowBox[{"\[LeftSkeleton]", "3", "\[RightSkeleton]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.7]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.1]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]},
    {"10", 
     RowBox[{"\<\"ReportCompute\"\>", "[", 
      RowBox[{"\<\"Ricci\"\>", "[", 
       RowBox[{
        RowBox[{"-", "1"}], ",", 
        RowBox[{"-", "1"}]}], "]"}], "]"}], GridBox[{
       {
        RowBox[{"\[LeftSkeleton]", "4", "\[RightSkeleton]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.7]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.1]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]},
    {"11", 
     RowBox[{"\<\"ReportCompute\"\>", "[", 
      RowBox[{"\<\"RicciScalar\"\>", "[", "]"}], "]"}], GridBox[{
       {
        RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.7]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.1]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]},
    {"12", 
     RowBox[{"\<\"ReportCompute\"\>", "[", 
      RowBox[{"\<\"Riemann\"\>", "[", 
       RowBox[{
        RowBox[{"-", "1"}], ",", 
        RowBox[{"-", "1"}], ",", 
        RowBox[{"-", "1"}], ",", 
        RowBox[{"-", "1"}]}], "]"}], "]"}], GridBox[{
       {
        RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.7]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.1]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]},
    {"13", 
     RowBox[{"\<\"ReportCompute\"\>", "[", 
      RowBox[{"\<\"Riemann\"\>", "[", 
       RowBox[{
        RowBox[{"-", "1"}], ",", 
        RowBox[{"-", "1"}], ",", 
        RowBox[{"-", "1"}], ",", "1"}], "]"}], "]"}], GridBox[{
       {
        RowBox[{"\[LeftSkeleton]", "4", "\[RightSkeleton]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.7]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.1]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]},
    {"14", 
     RowBox[{"\<\"ReportCompute\"\>", "[", 
      RowBox[{"\<\"Riemann\"\>", "[", 
       RowBox[{
        RowBox[{"-", "1"}], ",", 
        RowBox[{"-", "1"}], ",", "1", ",", "1"}], "]"}], "]"}], GridBox[{
       {
        RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.7]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.1]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]},
    {"15", 
     RowBox[{"\<\"ReportCompute\"\>", "[", 
      RowBox[{"\<\"Torsion\"\>", "[", 
       RowBox[{
        RowBox[{"-", "1"}], ",", 
        RowBox[{"-", "1"}], ",", 
        RowBox[{"-", "1"}]}], "]"}], "]"}], GridBox[{
       {
        RowBox[{"\[LeftSkeleton]", "2", "\[RightSkeleton]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.7]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.1]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]},
    {"16", 
     RowBox[{"\<\"ChartsOf\"\>", "[", 
      RowBox[{"{", 
       RowBox[{"\[LeftSkeleton]", "2", "\[RightSkeleton]"}], "}"}], "]"}], 
     TagBox[
      RowBox[{"{", "gen", "}"}],
      Short[#, 1]& ]},
    {"17", 
     TagBox[
      RowBox[{"Christoffel", "[", 
       RowBox[{
        RowBox[{"CCovD", "[", 
         RowBox[{"PDgen", ",", 
          StyleBox[GridBox[{
             {GridBox[{
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                  Annotation[#, {1, 1, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                  Annotation[#, {1, 1, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
               },
               AutoDelete->False,
               
               GridBoxDividers->{
                "ColumnsIndexed" -> {
                  1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], 
              GridBox[{
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                  Annotation[#, {1, 2, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                  Annotation[#, {1, 2, 2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
               },
               AutoDelete->False,
               
               GridBoxDividers->{
                "ColumnsIndexed" -> {
                  1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
             {GridBox[{
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                  Annotation[#, {2, 1, 1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                  Annotation[#, {2, 1, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
               },
               AutoDelete->False,
               
               GridBoxDividers->{
                "ColumnsIndexed" -> {
                  1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], 
              GridBox[{
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                  Annotation[#, {2, 2, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                  Annotation[#, {2, 2, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
               },
               AutoDelete->False,
               
               GridBoxDividers->{
                "ColumnsIndexed" -> {
                  1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
              "RowsIndexed" -> {
               1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
           FontSize->Small], ",", 
          StyleBox[GridBox[{
             {
              TagBox[
               TooltipBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxx",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxx,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}],
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
               Annotation[#, {1, 1} -> $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]], "Tooltip"]& ], 
              TagBox[
               TooltipBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}],
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
               Annotation[#, {1, 2} -> $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]], "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}],
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
               Annotation[#, {2, 1} -> $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]], "Tooltip"]& ], 
              TagBox[
               TooltipBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gyy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gyy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}],
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
               Annotation[#, {2, 2} -> $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]], "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
              "RowsIndexed" -> {
               1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
           FontSize->Small]}], "]"}], ",", "PDgen"}], "]"}],
      Short[#, 1]& ], 
     TagBox[
      StyleBox[GridBox[{
         {GridBox[{
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                  RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
              Annotation[#, {1, 1, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                  RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
              Annotation[#, {1, 1, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                  RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
              Annotation[#, {1, 2, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                  RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
              Annotation[#, {1, 2, 2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                  RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
              Annotation[#, {2, 1, 1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                  RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
              Annotation[#, {2, 1, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                  RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
              Annotation[#, {2, 2, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                  RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
              Annotation[#, {2, 2, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small],
      Short[#, 1]& ]},
    {"18", 
     TagBox[
      RowBox[{"CovDOfMetric", "[", 
       StyleBox[GridBox[{
          {
           TagBox[
            TooltipBox[
             RowBox[{
              InterpretationBox[
               StyleBox["gxx",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxx,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}],
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
               RowBox[{
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], $CellContext`gxx, Editable -> False], "[", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                   InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
            Annotation[#, {1, 1} -> $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]], "Tooltip"]& ], 
           TagBox[
            TooltipBox[
             RowBox[{
              InterpretationBox[
               StyleBox["gxy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}],
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
               RowBox[{
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], $CellContext`gxy, Editable -> False], "[", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                   InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
            Annotation[#, {1, 2} -> $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]], "Tooltip"]& ]},
          {
           TagBox[
            TooltipBox[
             RowBox[{
              InterpretationBox[
               StyleBox["gxy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}],
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
               RowBox[{
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], $CellContext`gxy, Editable -> False], "[", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                   InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
            Annotation[#, {2, 1} -> $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]], "Tooltip"]& ], 
           TagBox[
            TooltipBox[
             RowBox[{
              InterpretationBox[
               StyleBox["gyy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gyy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}],
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
               RowBox[{
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], $CellContext`gyy, Editable -> False], "[", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                   InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
            Annotation[#, {2, 2} -> $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]], "Tooltip"]& ]}
         },
         AutoDelete->False,
         GridBoxDividers->{
          "ColumnsIndexed" -> {
            1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
        FontSize->Small], "]"}],
      Short[#, 1]& ], 
     TagBox[
      RowBox[{"CCovD", "[", 
       RowBox[{"PDgen", ",", 
        StyleBox[GridBox[{
           {GridBox[{
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                Annotation[#, {1, 1, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                Annotation[#, {1, 1, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                Annotation[#, {1, 2, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                Annotation[#, {1, 2, 2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
           {GridBox[{
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                Annotation[#, {2, 1, 1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                Annotation[#, {2, 1, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                Annotation[#, {2, 2, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                Annotation[#, {2, 2, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
         FontSize->Small], ",", 
        StyleBox[GridBox[{
           {
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {1, 1} -> $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {1, 2} -> $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ]},
           {
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {2, 1} -> $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {2, 2} -> $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
         FontSize->Small]}], "]"}],
      Short[#, 1]& ]},
    {"19", 
     TagBox[
      RowBox[{"CovDOfMetric", "[", 
       RowBox[{
        StyleBox[GridBox[{
           {
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {1, 1} -> $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {1, 2} -> $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ]},
           {
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {2, 1} -> $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {2, 2} -> $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
         FontSize->Small], ",", "Zero"}], "]"}],
      Short[#, 1]& ], 
     TagBox[
      RowBox[{"CCovD", "[", 
       RowBox[{"PDgen", ",", 
        StyleBox[GridBox[{
           {GridBox[{
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]}]],
                
                Annotation[#, {1, 1, 
                  1} -> (Rational[-1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                Annotation[#, {1, 1, 2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                Annotation[#, {1, 2, 1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                Annotation[#, {1, 2, 2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
           {GridBox[{
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                Annotation[#, {2, 1, 1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                Annotation[#, {2, 1, 2} -> (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                Annotation[#, {2, 2, 1} -> (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]}]],
                
                Annotation[#, {2, 2, 
                  2} -> (Rational[-1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
         FontSize->Small], ",", 
        StyleBox[GridBox[{
           {
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {1, 1} -> $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {1, 2} -> $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ]},
           {
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {2, 1} -> $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {2, 2} -> $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
         FontSize->Small]}], "]"}],
      Short[#, 1]& ]},
    {"20", 
     TagBox[
      RowBox[{"CovDOfMetric", "[", 
       RowBox[{
        StyleBox[GridBox[{
           {
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {1, 1} -> $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {1, 2} -> $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ]},
           {
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {2, 1} -> $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {2, 2} -> $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
         FontSize->Small], ",", 
        StyleBox[GridBox[{
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {1, 1, 1} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["T112",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`T112,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
                Annotation[#, {1, 1, 2} -> $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]], "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                   RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]],
                Annotation[#, {1, 2, 1} -> -$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]], "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {1, 2, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {2, 1, 1} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["T212",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`T212,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`y[],
                    Editable->False]}], "]"}],
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
                Annotation[#, {2, 1, 2} -> $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]], "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                   RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]],
                Annotation[#, {2, 2, 1} -> -$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]], "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                   "0"}]],
                Annotation[#, {2, 2, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
         FontSize->Small]}], "]"}],
      Short[#, 1]& ], 
     TagBox[
      RowBox[{"CCovD", "[", 
       RowBox[{"PDgen", ",", 
        StyleBox[GridBox[{
           {GridBox[{
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]}]],
                
                Annotation[#, {1, 1, 
                  1} -> (Rational[-1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                Annotation[#, {1, 1, 2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((((-2) $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                Annotation[#, {1, 2, 1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                Annotation[#, {1, 2, 2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
           {GridBox[{
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                Annotation[#, {2, 1, 1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                Annotation[#, {2, 1, 2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                Annotation[#, {2, 2, 1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                Annotation[#, {2, 2, 2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
         FontSize->Small], ",", 
        StyleBox[GridBox[{
           {
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {1, 1} -> $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {1, 2} -> $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ]},
           {
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {2, 1} -> $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {2, 2} -> $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
         FontSize->Small]}], "]"}],
      Short[#, 1]& ]},
    {"21", 
     TagBox[
      RowBox[{"Determinant", "[", 
       RowBox[{
        StyleBox[GridBox[{
           {
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {1, 1} -> $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {1, 2} -> $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ]},
           {
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {2, 1} -> $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {2, 2} -> $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
         FontSize->Small], ",", "gen"}], "]"}],
      Short[#, 1]& ], 
     TagBox[
      OverscriptBox[
       OverscriptBox[
        RowBox[{"CTensor", "[", 
         RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}], "]"}], 
        RowBox[{"\[LeftSkeleton]", "49", "\[RightSkeleton]"}]], 
       StyleBox["~",
        FontColor->RGBColor[1., 0., 0.]]],
      Short[#, 1]& ]},
    {"22", 
     TagBox[
      RowBox[{"Einstein", "[", 
       RowBox[{"CCovD", "[", 
        RowBox[{"PDgen", ",", 
         StyleBox[GridBox[{
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]}]],
                 Annotation[#, {1, 1, 1} -> (Rational[-1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 1, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((((-2) $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 2, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 2, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 1, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 1, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 2, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 2, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small], ",", 
         StyleBox[GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 1} -> $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 2} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 1} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 2} -> $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small]}], "]"}], "]"}],
      Short[#, 1]& ], 
     TagBox["Zero",
      Short[#, 1]& ]},
    {"23", 
     TagBox[
      RowBox[{"Einstein", "[", 
       RowBox[{"CCovD", "[", 
        RowBox[{"PDgen", ",", 
         StyleBox[GridBox[{
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {1, 1, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {1, 1, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {1, 2, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 2, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 1, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {2, 1, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {2, 2, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {2, 2, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small], ",", 
         StyleBox[GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 1} -> $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 2} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 1} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 2} -> $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small]}], "]"}], "]"}],
      Short[#, 1]& ], 
     TagBox["Zero",
      Short[#, 1]& ]},
    {"24", 
     TagBox[
      RowBox[{"Inv", "[", 
       StyleBox[GridBox[{
          {
           TagBox[
            TooltipBox[
             RowBox[{
              InterpretationBox[
               StyleBox["gxx",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxx,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}],
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
               RowBox[{
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], $CellContext`gxx, Editable -> False], "[", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                   InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
            Annotation[#, {1, 1} -> $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]], "Tooltip"]& ], 
           TagBox[
            TooltipBox[
             RowBox[{
              InterpretationBox[
               StyleBox["gxy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}],
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
               RowBox[{
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], $CellContext`gxy, Editable -> False], "[", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                   InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
            Annotation[#, {1, 2} -> $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]], "Tooltip"]& ]},
          {
           TagBox[
            TooltipBox[
             RowBox[{
              InterpretationBox[
               StyleBox["gxy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}],
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
               RowBox[{
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], $CellContext`gxy, Editable -> False], "[", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                   InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
            Annotation[#, {2, 1} -> $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]], "Tooltip"]& ], 
           TagBox[
            TooltipBox[
             RowBox[{
              InterpretationBox[
               StyleBox["gyy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gyy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}],
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
               RowBox[{
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], $CellContext`gyy, Editable -> False], "[", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                   InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
            Annotation[#, {2, 2} -> $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]], "Tooltip"]& ]}
         },
         AutoDelete->False,
         GridBoxDividers->{
          "ColumnsIndexed" -> {
            1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
        FontSize->Small], "]"}],
      Short[#, 1]& ], 
     TagBox[
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], $CellContext`gyy, Editable -> False], "[", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                   InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], 
               RowBox[{
                 RowBox[{"-", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                 "+", 
                 RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
           Annotation[#, {1, 1} -> $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]]/(-$CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]]^2 + $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]] $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]]), "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], $CellContext`gxy, Editable -> False], "[", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                   InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], 
               RowBox[{
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                 
                 RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
           Annotation[#, {1, 2} -> $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]]/($CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]]^2 - $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]] $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]]), "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], $CellContext`gxy, Editable -> False], "[", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                   InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], 
               RowBox[{
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                 
                 RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
           Annotation[#, {2, 1} -> $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]]/($CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]]^2 - $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]] $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]]), "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], $CellContext`gxx, Editable -> False], "[", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                   InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], 
               RowBox[{
                 RowBox[{"-", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                 "+", 
                 RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
           Annotation[#, {2, 2} -> $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]]/(-$CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]]^2 + $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]] $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]]), "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small],
      Short[#, 1]& ]},
    {"25", 
     TagBox[
      RowBox[{"Kretschmann", "[", 
       RowBox[{"CCovD", "[", 
        RowBox[{"PDgen", ",", 
         StyleBox[GridBox[{
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {1, 1, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {1, 1, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {1, 2, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 2, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 1, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {2, 1, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {2, 2, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {2, 2, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small], ",", 
         StyleBox[GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 1} -> $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 2} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 1} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 2} -> $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small]}], "]"}], "]"}],
      Short[#, 1]& ], 
     TagBox[
      RowBox[{"CTensor", "[", 
       RowBox[{
        FractionBox[
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}], ")"}], "2"], 
         
         RowBox[{"4", " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}], "2"], "-", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}], "]"}], 
              " ", 
              RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}]}]}], ")"}],
            "4"]}]], ",", 
        RowBox[{"{", "}"}], ",", "0"}], "]"}],
      Short[#, 1]& ]},
    {"26", 
     TagBox[
      RowBox[{"LC", "[", 
       StyleBox[GridBox[{
          {
           TagBox[
            TooltipBox[
             RowBox[{
              InterpretationBox[
               StyleBox["gxx",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxx,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}],
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
               RowBox[{
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], $CellContext`gxx, Editable -> False], "[", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                   InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
            Annotation[#, {1, 1} -> $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]], "Tooltip"]& ], 
           TagBox[
            TooltipBox[
             RowBox[{
              InterpretationBox[
               StyleBox["gxy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}],
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
               RowBox[{
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], $CellContext`gxy, Editable -> False], "[", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                   InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
            Annotation[#, {1, 2} -> $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]], "Tooltip"]& ]},
          {
           TagBox[
            TooltipBox[
             RowBox[{
              InterpretationBox[
               StyleBox["gxy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gxy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}],
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
               RowBox[{
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], $CellContext`gxy, Editable -> False], "[", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                   InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
            Annotation[#, {2, 1} -> $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]], "Tooltip"]& ], 
           TagBox[
            TooltipBox[
             RowBox[{
              InterpretationBox[
               StyleBox["gyy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`gyy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["y",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`y[],
                Editable->False]}], "]"}],
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
               RowBox[{
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], $CellContext`gyy, Editable -> False], "[", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                   InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
            Annotation[#, {2, 2} -> $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]], "Tooltip"]& ]}
         },
         AutoDelete->False,
         GridBoxDividers->{
          "ColumnsIndexed" -> {
            1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
        FontSize->Small], "]"}],
      Short[#, 1]& ], 
     TagBox[
      RowBox[{"CCovD", "[", 
       RowBox[{"PDgen", ",", 
        StyleBox[GridBox[{
           {GridBox[{
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]}]],
                
                Annotation[#, {1, 1, 
                  1} -> (Rational[-1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                Annotation[#, {1, 1, 2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                Annotation[#, {1, 2, 1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                Annotation[#, {1, 2, 2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
           {GridBox[{
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                Annotation[#, {2, 1, 1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                Annotation[#, {2, 1, 2} -> (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                Annotation[#, {2, 2, 1} -> (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]}]],
                
                Annotation[#, {2, 2, 
                  2} -> (Rational[-1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
         FontSize->Small], ",", 
        StyleBox[GridBox[{
           {
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {1, 1} -> $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {1, 2} -> $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ]},
           {
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {2, 1} -> $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {2, 2} -> $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
         FontSize->Small]}], "]"}],
      Short[#, 1]& ]},
    {"27", 
     TagBox[
      RowBox[{"Ricci", "[", 
       RowBox[{"CCovD", "[", 
        RowBox[{"PDgen", ",", 
         StyleBox[GridBox[{
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]}]],
                 Annotation[#, {1, 1, 1} -> (Rational[-1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 1, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 2, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 2, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 1, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {2, 1, 2} -> (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {2, 2, 1} -> (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]}]],
                 Annotation[#, {2, 2, 2} -> (Rational[-1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small], ",", 
         StyleBox[GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 1} -> $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 2} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 1} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 2} -> $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small]}], "]"}], "]"}],
      Short[#, 1]& ], 
     TagBox[
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                 RowBox[{"(", 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], 
               RowBox[{"4", " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                  "2"]}]]}]],
           Annotation[#, {1, 1} -> ((Rational[1, 4] $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]]) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])^(-2)) ($CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]^2) + $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] (Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) + (2 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2) (Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] ((-2) 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))), "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                 RowBox[{"(", 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], 
               RowBox[{"4", " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                  "2"]}]]}]],
           Annotation[#, {1, 2} -> ((Rational[1, 4] $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]]) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])^(-2)) ($CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]^2) + $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] (Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) + (2 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2) (Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] ((-2) 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))), "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                 RowBox[{"(", 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], 
               RowBox[{"4", " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                  "2"]}]]}]],
           Annotation[#, {2, 1} -> ((Rational[1, 4] $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]]) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])^(-2)) ($CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]^2) + $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] (Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) + (2 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2) (Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] ((-2) 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))), "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                 RowBox[{"(", 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], 
               RowBox[{"4", " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                  "2"]}]]}]],
           Annotation[#, {2, 2} -> ((Rational[1, 4] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]]) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])^(-2)) ($CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]^2) + $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] (Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) + (2 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2) (Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] ((-2) 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))), "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small],
      Short[#, 1]& ]},
    {"28", 
     TagBox[
      RowBox[{"Ricci", "[", 
       RowBox[{"CCovD", "[", 
        RowBox[{"PDgen", ",", 
         StyleBox[GridBox[{
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]}]],
                 Annotation[#, {1, 1, 1} -> (Rational[-1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 1, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((((-2) $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 2, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 2, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 1, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 1, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 2, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 2, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small], ",", 
         StyleBox[GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 1} -> $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 2} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 1} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 2} -> $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small]}], "]"}], "]"}],
      Short[#, 1]& ], 
     TagBox[
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                 RowBox[{"(", 
                   RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "3"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], 
               RowBox[{"4", " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                  "2"]}]]}]],
           Annotation[#, {1, 1} -> ((Rational[1, 4] $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]]) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]^2) ($CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]]) + $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] (((-2) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^3) (Derivative[0, 1][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]]) + (2 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))), "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                 RowBox[{"(", 
                   RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "3"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], 
               RowBox[{"4", " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                  "2"]}]]}]],
           Annotation[#, {1, 2} -> ((Rational[1, 4] $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]]) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]^2) ($CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]]) + $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] (((-2) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^3) (Derivative[0, 1][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]]) + (2 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))), "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                 RowBox[{"(", 
                   RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "3"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], 
               RowBox[{"4", " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                  "2"]}]]}]],
           Annotation[#, {2, 1} -> ((Rational[1, 4] $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]]) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]^2) ($CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]]) + $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] (((-2) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^3) (Derivative[0, 1][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]]) + (2 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))), "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                 RowBox[{"(", 
                   RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "3"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], 
               RowBox[{"4", " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                  "2"]}]]}]],
           Annotation[#, {2, 2} -> ((Rational[1, 4] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]]) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]^2) ($CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]]) + $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] (((-2) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^3) (Derivative[0, 1][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]]) + (2 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))), "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small],
      Short[#, 1]& ]},
    {"29", 
     TagBox[
      RowBox[{"Ricci", "[", 
       RowBox[{"CCovD", "[", 
        RowBox[{"PDgen", ",", 
         StyleBox[GridBox[{
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {1, 1, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {1, 1, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {1, 2, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 2, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 1, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {2, 1, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {2, 2, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {2, 2, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small], ",", 
         StyleBox[GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 1} -> $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 2} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 1} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 2} -> $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small]}], "]"}], "]"}],
      Short[#, 1]& ], 
     TagBox[
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                 RowBox[{"(", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                   ")"}]}], 
               RowBox[{
                 RowBox[{"-", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                 "+", 
                 RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
           Annotation[#, {1, 1} -> ($CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]]/(-$CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 + $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ((-$CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]] - $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                $CellContext`x[], 
                $CellContext`y[]] + 
              Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                 $CellContext`x[], 
                 $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + 
               Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + $CellContext`T212[
                 $CellContext`x[], 
                 $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + ((Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]] + $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                 $CellContext`x[], 
                 $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                $CellContext`x[], 
                $CellContext`y[]] + 
              Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]), "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                 RowBox[{"(", 
                   RowBox[{
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                   ")"}]}], 
               RowBox[{"4", " ", 
                 RowBox[{"(", 
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                   ")"}]}]]}]],
           Annotation[#, {1, 2} -> ((Rational[1, 4] $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]])/($CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]]^2 - $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]] $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]])) ((4 $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]] + (4 $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                 $CellContext`x[], 
                 $CellContext`y[]] + 2 Derivative[0, 2][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]) + (4 $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]) + ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + 
                Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) - (4 $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                $CellContext`x[], 
                $CellContext`y[]] - (4 $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                $CellContext`x[], 
                $CellContext`y[]] - 4 Derivative[1, 1][$CellContext`gxy][
                $CellContext`x[], 
                $CellContext`y[]] + 2 Derivative[2, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]), "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{"-", 
                FractionBox[
                 RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                 RowBox[{
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                   
                   RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]}]],
           Annotation[#, {2, 1} -> ((-$CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]])/($CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]]^2 - $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]] $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]])) ((-$CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]] - $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                $CellContext`x[], 
                $CellContext`y[]] + 
              Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                 $CellContext`x[], 
                 $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + 
               Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + $CellContext`T212[
                 $CellContext`x[], 
                 $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + ((Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]] + $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                 $CellContext`x[], 
                 $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                $CellContext`x[], 
                $CellContext`y[]] + 
              Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]), "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{"-", 
                FractionBox[
                 RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                 RowBox[{"4", " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]}]],
           Annotation[#, {2, 2} -> ((Rational[-1, 4] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])/(-$CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 + $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ((4 $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]] + (4 $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                 $CellContext`x[], 
                 $CellContext`y[]] + 2 Derivative[0, 2][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]) + (4 $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]) + ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + 
                Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) - (4 $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                $CellContext`x[], 
                $CellContext`y[]] - (4 $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                $CellContext`x[], 
                $CellContext`y[]] - 4 Derivative[1, 1][$CellContext`gxy][
                $CellContext`x[], 
                $CellContext`y[]] + 2 Derivative[2, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]), "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small],
      Short[#, 1]& ]},
    {"30", 
     TagBox[
      RowBox[{"RicciScalar", "[", 
       RowBox[{"CCovD", "[", 
        RowBox[{"PDgen", ",", 
         StyleBox[GridBox[{
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]}]],
                 Annotation[#, {1, 1, 1} -> (Rational[-1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 1, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 2, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 2, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 1, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {2, 1, 2} -> (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {2, 2, 1} -> (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]}]],
                 Annotation[#, {2, 2, 2} -> (Rational[-1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small], ",", 
         StyleBox[GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 1} -> $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 2} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 1} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 2} -> $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small]}], "]"}], "]"}],
      Short[#, 1]& ], 
     TagBox[
      RowBox[{"CTensor", "[", 
       RowBox[{
        FractionBox[
         RowBox[{
          RowBox[{
           RowBox[{
            InterpretationBox[
             StyleBox["gxx",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`gxx,
             Editable->False], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["x",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`x[],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["y",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`y[],
              Editable->False]}], "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"0", ",", "1"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox[
                  RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}],
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], 
                TagBox[
                 RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}]}], "-", 
             RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}], "+", 
             SuperscriptBox[
              RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}], "2"]}], 
            ")"}]}], "+", 
          RowBox[{"\[LeftSkeleton]", "3", "\[RightSkeleton]"}]}], 
         RowBox[{"2", " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}], "2"], "-", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}], "]"}], 
              " ", 
              RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}]}]}], ")"}],
            "2"]}]], ",", 
        RowBox[{"{", "}"}], ",", "0"}], "]"}],
      Short[#, 1]& ]},
    {"31", 
     TagBox[
      RowBox[{"RicciScalar", "[", 
       RowBox[{"CCovD", "[", 
        RowBox[{"PDgen", ",", 
         StyleBox[GridBox[{
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]}]],
                 Annotation[#, {1, 1, 1} -> (Rational[-1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 1, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((((-2) $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 2, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 2, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 1, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 1, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 2, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 2, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small], ",", 
         StyleBox[GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 1} -> $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 2} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 1} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 2} -> $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small]}], "]"}], "]"}],
      Short[#, 1]& ], 
     TagBox[
      RowBox[{"CTensor", "[", 
       RowBox[{
        FractionBox[
         RowBox[{
          RowBox[{"2", " ", 
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["gxx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], "2"], " ", 
           RowBox[{"(", 
            RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}], ")"}]}], 
          "+", 
          RowBox[{"\[LeftSkeleton]", "4", "\[RightSkeleton]"}], "+", 
          RowBox[{
           RowBox[{
            InterpretationBox[
             StyleBox["gxx",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`gxx,
             Editable->False], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["x",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`x[],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["y",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`y[],
              Editable->False]}], "]"}], " ", 
           RowBox[{"(", 
            RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}], ")"}]}]}], 
         RowBox[{"2", " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}], "2"], "-", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}], "]"}], 
              " ", 
              RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}]}]}], ")"}],
            "2"]}]], ",", 
        RowBox[{"{", "}"}], ",", "0"}], "]"}],
      Short[#, 1]& ]},
    {"32", 
     TagBox[
      RowBox[{"RicciScalar", "[", 
       RowBox[{"CCovD", "[", 
        RowBox[{"PDgen", ",", 
         StyleBox[GridBox[{
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {1, 1, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {1, 1, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {1, 2, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 2, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 1, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {2, 1, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {2, 2, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {2, 2, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small], ",", 
         StyleBox[GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 1} -> $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 2} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 1} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 2} -> $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small]}], "]"}], "]"}],
      Short[#, 1]& ], 
     TagBox[
      RowBox[{"CTensor", "[", 
       RowBox[{
        FractionBox[
         RowBox[{
          RowBox[{"2", " ", 
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["gxx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False]}], "]"}], "2"], " ", 
           RowBox[{"(", 
            RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}], ")"}]}], 
          "+", 
          RowBox[{"\[LeftSkeleton]", "4", "\[RightSkeleton]"}], "+", 
          RowBox[{
           RowBox[{
            InterpretationBox[
             StyleBox["gxx",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`gxx,
             Editable->False], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["x",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`x[],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["y",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`y[],
              Editable->False]}], "]"}], " ", 
           RowBox[{"(", 
            RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}], ")"}]}]}], 
         RowBox[{"2", " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}], "2"], "-", 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}], "]"}], 
              " ", 
              RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}]}]}], ")"}],
            "2"]}]], ",", 
        RowBox[{"{", "}"}], ",", "0"}], "]"}],
      Short[#, 1]& ]},
    {"33", 
     TagBox[
      RowBox[{"Riemann", "[", 
       RowBox[{"CCovD", "[", 
        RowBox[{"PDgen", ",", 
         StyleBox[GridBox[{
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]}]],
                 Annotation[#, {1, 1, 1} -> (Rational[-1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 1, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 2, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 2, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 1, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {2, 1, 2} -> (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {2, 2, 1} -> (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]}]],
                 Annotation[#, {2, 2, 2} -> (Rational[-1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small], ",", 
         StyleBox[GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 1} -> $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 2} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 1} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 2} -> $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small]}], "]"}], "]"}],
      Short[#, 1]& ], 
     TagBox[
      StyleBox[GridBox[{
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]}]]}]}]],
              Annotation[#, {1, 2, 1, 1} -> ((Rational[-1, 4] $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2)) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 
                    Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], 
                  RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]}]]}]],
              
              Annotation[#, {1, 2, 1, 
                2} -> ((Rational[1, 4] $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2)) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 
                    Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]}]]}]}]],
              Annotation[#, {1, 2, 2, 1} -> ((Rational[-1, 4] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2)) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 
                    Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], 
                  RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]}]]}]],
              
              Annotation[#, {1, 2, 2, 
                2} -> ((Rational[1, 4] $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2)) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 
                    Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], 
                  RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]}]]}]],
              
              Annotation[#, {2, 1, 1, 
                1} -> ((Rational[1, 4] $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2)) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 
                    Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], 
                  RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]}]]}]],
              
              Annotation[#, {2, 1, 1, 
                2} -> ((Rational[1, 4] $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2)) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((-Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 
                    Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-2) 
                    Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], 
                  RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]}]]}]],
              
              Annotation[#, {2, 1, 2, 
                1} -> ((Rational[1, 4] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2)) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 
                    Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]}]]}]}]],
              Annotation[#, {2, 1, 2, 2} -> ((Rational[-1, 4] $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2)) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 
                    Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small],
      Short[#, 1]& ]},
    {"34", 
     TagBox[
      RowBox[{"Riemann", "[", 
       RowBox[{"CCovD", "[", 
        RowBox[{"PDgen", ",", 
         StyleBox[GridBox[{
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]}]],
                 Annotation[#, {1, 1, 1} -> (Rational[-1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 1, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((((-2) $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 2, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 2, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 1, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 1, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 2, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 2, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small], ",", 
         StyleBox[GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 1} -> $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 2} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 1} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 2} -> $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small]}], "]"}], "]"}],
      Short[#, 1]& ], 
     TagBox[
      StyleBox[GridBox[{
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "3"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]}]]}]}]],
              Annotation[#, {1, 2, 1, 1} -> ((Rational[-1, 4] $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) 
                    Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                    Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                   Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                    Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "3"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], 
                  RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]}]]}]],
              
              Annotation[#, {1, 2, 1, 
                2} -> ((Rational[1, 4] $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) 
                    Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                    Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                   Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                    Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "3"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]}]]}]}]],
              Annotation[#, {1, 2, 2, 1} -> ((Rational[-1, 4] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) 
                    Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                    Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                   Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                    Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "3"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], 
                  RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]}]]}]],
              
              Annotation[#, {1, 2, 2, 
                2} -> ((Rational[1, 4] $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) 
                    Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                    Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                   Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                    Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "3"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], 
                  RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]}]]}]],
              
              Annotation[#, {2, 1, 1, 
                1} -> ((Rational[1, 4] $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) 
                    Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                    Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                   Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                    Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "3"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]}]]}]}]],
              Annotation[#, {2, 1, 1, 2} -> ((Rational[-1, 4] $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) 
                    Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                    Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                   Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                    Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "3"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], 
                  RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]}]]}]],
              Annotation[#, {2, 1, 2, 
                1} -> ((Rational[1, 4] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) 
                    Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                    Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                   Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                    Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "3"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]}]]}]}]],
              Annotation[#, {2, 1, 2, 2} -> ((Rational[-1, 4] $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) 
                    Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                    Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                   Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                    Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small],
      Short[#, 1]& ]},
    {"35", 
     TagBox[
      RowBox[{"Riemann", "[", 
       RowBox[{"CCovD", "[", 
        RowBox[{"PDgen", ",", 
         StyleBox[GridBox[{
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {1, 1, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {1, 1, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {1, 2, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 2, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 1, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {2, 1, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {2, 2, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {2, 2, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small], ",", 
         StyleBox[GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 1} -> $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 2} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 1} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 2} -> $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small]}], "]"}], "]"}],
      Short[#, 1]& ], 
     TagBox[
      StyleBox[GridBox[{
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                  RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
              Annotation[#, {1, 2, 1, 1} -> ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                  RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
              Annotation[#, {1, 2, 1, 2} -> ($CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]/(-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                  RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {1, 2, 2, 
                1} -> ((Rational[1, 4] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])/(-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                   Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] - (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - 4 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 2 Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                  RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {1, 2, 2, 
                2} -> ((Rational[1, 4] $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]])/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) ((4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                   Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] - (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - 4 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 2 Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]}]],
              Annotation[#, {2, 1, 1, 1} -> ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]])/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]}]],
              Annotation[#, {2, 1, 1, 2} -> ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]])/(-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]}]],
              Annotation[#, {2, 1, 2, 1} -> ((Rational[-1, 4] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])/(-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                   Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] - (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - 4 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 2 Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]}]],
              Annotation[#, {2, 1, 2, 2} -> ((Rational[-1, 4] $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]])/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) ((4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                   Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] - (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - 4 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 2 Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small],
      Short[#, 1]& ]},
    {"36", 
     TagBox[
      RowBox[{"RiemannDown", "[", 
       RowBox[{"CCovD", "[", 
        RowBox[{"PDgen", ",", 
         StyleBox[GridBox[{
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]}]],
                 Annotation[#, {1, 1, 1} -> (Rational[-1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 1, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((((-2) $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 2, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 2, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 1, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 1, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 2, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 2, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small], ",", 
         StyleBox[GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 1} -> $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 2} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 1} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 2} -> $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small]}], "]"}], "]"}],
      Short[#, 1]& ], 
     TagBox[
      StyleBox[GridBox[{
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], "-", 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "3"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], 
                  RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {1, 2, 1, 
                2} -> (Rational[1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 (((-2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (4 
                    Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 
                    Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) - (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                  Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]]) - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                   Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                   Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "3"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], 
                  RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {1, 2, 2, 
                1} -> (Rational[1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) 
                    Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                    Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                   Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                    Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "3"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], 
                  RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {2, 1, 1, 
                2} -> (Rational[1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) 
                    Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                    Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                   Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                    Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], "-", 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "3"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], 
                  RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {2, 1, 2, 
                1} -> (Rational[1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 (((-2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (4 
                    Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 
                    Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) - (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                  Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]]) - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                   Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                   Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small],
      Short[#, 1]& ]},
    {"37", 
     TagBox[
      RowBox[{"RiemannDown", "[", 
       RowBox[{"CCovD", "[", 
        RowBox[{"PDgen", ",", 
         StyleBox[GridBox[{
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {1, 1, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {1, 1, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {1, 2, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 2, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 1, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {2, 1, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {2, 2, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {2, 2, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small], ",", 
         StyleBox[GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 1} -> $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 2} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 1} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 2} -> $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small]}], "]"}], "]"}],
      Short[#, 1]& ], 
     TagBox[
      StyleBox[GridBox[{
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                   RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                   RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                   "-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                   RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                   "+", 
                   FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                   RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                   RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]}]],
              Annotation[#, {1, 2, 1, 2} -> (-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] + 
                Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                   Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + 
                Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                   RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                   RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                   "+", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                   FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                   RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                   "-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                   RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                   RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]}]],
              Annotation[#, {1, 2, 2, 1} -> $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                Rational[1, 2] Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + ((Rational[1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + ((Rational[1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + ((
                   Rational[-1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] - $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - Derivative[1, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + 
                Rational[1, 2] Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                   RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                   RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                   "+", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                   FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                   RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                   "-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                   RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                   RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]}]],
              Annotation[#, {2, 1, 1, 2} -> $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                Rational[1, 2] Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + ((Rational[1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + ((Rational[1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + ((
                   Rational[-1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] - $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - Derivative[1, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + 
                Rational[1, 2] Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                   RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                   RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                   "-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                   RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                   "+", 
                   FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                   RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                   RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]}]],
              Annotation[#, {2, 1, 2, 1} -> (-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] + 
                Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                   Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + 
                Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small],
      Short[#, 1]& ]},
    {"38", 
     TagBox[
      RowBox[{"TensorDerivative", "[", 
       RowBox[{
        StyleBox[GridBox[{
           {
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {1, 1} -> $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {1, 2} -> $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ]},
           {
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {2, 1} -> $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {2, 2} -> $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
         FontSize->Small], ",", 
        RowBox[{"CCovD", "[", 
         RowBox[{"PDgen", ",", 
          StyleBox[GridBox[{
             {GridBox[{
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]}]],
                  
                  Annotation[#, {1, 1, 
                    1} -> (Rational[-1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                  Annotation[#, {1, 1, 2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]}
               },
               AutoDelete->False,
               
               GridBoxDividers->{
                "ColumnsIndexed" -> {
                  1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], 
              GridBox[{
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                  Annotation[#, {1, 2, 1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]},
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                  Annotation[#, {1, 2, 2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
               },
               AutoDelete->False,
               
               GridBoxDividers->{
                "ColumnsIndexed" -> {
                  1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
             {GridBox[{
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                  Annotation[#, {2, 1, 1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                  Annotation[#, {2, 1, 2} -> (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]}
               },
               AutoDelete->False,
               
               GridBoxDividers->{
                "ColumnsIndexed" -> {
                  1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], 
              GridBox[{
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                  Annotation[#, {2, 2, 1} -> (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]},
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]}]],
                  
                  Annotation[#, {2, 2, 
                    2} -> (Rational[-1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
               },
               AutoDelete->False,
               
               GridBoxDividers->{
                "ColumnsIndexed" -> {
                  1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
              "RowsIndexed" -> {
               1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
           FontSize->Small], ",", 
          StyleBox[GridBox[{
             {
              TagBox[
               TooltipBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxx",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxx,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}],
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
               Annotation[#, {1, 1} -> $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]], "Tooltip"]& ], 
              TagBox[
               TooltipBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}],
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
               Annotation[#, {1, 2} -> $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]], "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}],
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
               Annotation[#, {2, 1} -> $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]], "Tooltip"]& ], 
              TagBox[
               TooltipBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gyy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gyy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}],
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
               Annotation[#, {2, 2} -> $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]], "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
              "RowsIndexed" -> {
               1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
           FontSize->Small]}], "]"}]}], "]"}],
      Short[#, 1]& ], 
     TagBox["Zero",
      Short[#, 1]& ]},
    {"39", 
     TagBox[
      RowBox[{"TensorDerivative", "[", 
       RowBox[{
        StyleBox[GridBox[{
           {
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {1, 1} -> $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {1, 2} -> $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ]},
           {
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gxy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gxy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {2, 1} -> $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ], 
            TagBox[
             TooltipBox[
              RowBox[{
               InterpretationBox[
                StyleBox["gyy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`gyy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["y",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`y[],
                 Editable->False]}], "]"}],
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
             Annotation[#, {2, 2} -> $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]], "Tooltip"]& ]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
         FontSize->Small], ",", 
        RowBox[{"CCovD", "[", 
         RowBox[{"PDgen", ",", 
          StyleBox[GridBox[{
             {GridBox[{
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]}]],
                  
                  Annotation[#, {1, 1, 
                    1} -> (Rational[-1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                  Annotation[#, {1, 1, 2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((((-2) $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
               },
               AutoDelete->False,
               
               GridBoxDividers->{
                "ColumnsIndexed" -> {
                  1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], 
              GridBox[{
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                  Annotation[#, {1, 2, 1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                  Annotation[#, {1, 2, 2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
               },
               AutoDelete->False,
               
               GridBoxDividers->{
                "ColumnsIndexed" -> {
                  1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
             {GridBox[{
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                  Annotation[#, {2, 1, 1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                  Annotation[#, {2, 1, 2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
               },
               AutoDelete->False,
               
               GridBoxDividers->{
                "ColumnsIndexed" -> {
                  1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], 
              GridBox[{
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                  Annotation[#, {2, 2, 1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                  Annotation[#, {2, 2, 2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
               },
               AutoDelete->False,
               
               GridBoxDividers->{
                "ColumnsIndexed" -> {
                  1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
              "RowsIndexed" -> {
               1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
           FontSize->Small], ",", 
          StyleBox[GridBox[{
             {
              TagBox[
               TooltipBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxx",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxx,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}],
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
               Annotation[#, {1, 1} -> $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]], "Tooltip"]& ], 
              TagBox[
               TooltipBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}],
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
               Annotation[#, {1, 2} -> $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]], "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}],
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
               Annotation[#, {2, 1} -> $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]], "Tooltip"]& ], 
              TagBox[
               TooltipBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gyy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gyy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}],
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
               Annotation[#, {2, 2} -> $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]], "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
              "RowsIndexed" -> {
               1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
           FontSize->Small]}], "]"}]}], "]"}],
      Short[#, 1]& ], 
     TagBox["Zero",
      Short[#, 1]& ]},
    {"40", 
     TagBox[
      RowBox[{"TensorDerivative", "[", 
       RowBox[{
        StyleBox[GridBox[{
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 1, 2} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 2, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 2, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 2, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]}]],
                Annotation[#, {1, 2, 1, 2} -> (-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                   Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]], "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]}]],
                Annotation[#, {1, 2, 2, 1} -> $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[-1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                  Rational[1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]], "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 2, 2, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]},
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 1, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]}]],
                Annotation[#, {2, 1, 1, 2} -> $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[-1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                  Rational[1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]], "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]}]],
                Annotation[#, {2, 1, 2, 1} -> (-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                   Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]], "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 1, 2, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 1, 2} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 2, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 2, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
         FontSize->Small], ",", 
        RowBox[{"CCovD", "[", 
         RowBox[{"PDgen", ",", 
          StyleBox[GridBox[{
             {GridBox[{
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                  Annotation[#, {1, 1, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                  Annotation[#, {1, 1, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
               },
               AutoDelete->False,
               
               GridBoxDividers->{
                "ColumnsIndexed" -> {
                  1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], 
              GridBox[{
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                  Annotation[#, {1, 2, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                  Annotation[#, {1, 2, 2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
               },
               AutoDelete->False,
               
               GridBoxDividers->{
                "ColumnsIndexed" -> {
                  1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
             {GridBox[{
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                  Annotation[#, {2, 1, 1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                  Annotation[#, {2, 1, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
               },
               AutoDelete->False,
               
               GridBoxDividers->{
                "ColumnsIndexed" -> {
                  1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], 
              GridBox[{
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                  Annotation[#, {2, 2, 1} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
                {
                 TagBox[
                  TooltipBox["\<\"\[GrayCircle]\"\>",
                   RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                  Annotation[#, {2, 2, 2} -> ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
               },
               AutoDelete->False,
               
               GridBoxDividers->{
                "ColumnsIndexed" -> {
                  1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
              "RowsIndexed" -> {
               1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
           FontSize->Small], ",", 
          StyleBox[GridBox[{
             {
              TagBox[
               TooltipBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxx",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxx,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}],
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
               Annotation[#, {1, 1} -> $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]], "Tooltip"]& ], 
              TagBox[
               TooltipBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}],
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
               Annotation[#, {1, 2} -> $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]], "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gxy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gxy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}],
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
               Annotation[#, {2, 1} -> $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]], "Tooltip"]& ], 
              TagBox[
               TooltipBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["gyy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`gyy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False]}], "]"}],
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
               Annotation[#, {2, 2} -> $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]], "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
              "RowsIndexed" -> {
               1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
           FontSize->Small]}], "]"}]}], "]"}],
      Short[#, 1]& ], 
     TagBox[
      StyleBox[GridBox[{
         {GridBox[{
            {GridBox[{
               {
                TagBox[
                 TooltipBox["0",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1", ",", "1", ",", "1"}], 
                    "}"}], "\[Rule]", "0"}]],
                 Annotation[#, {1, 1, 1, 1, 1} -> 0, "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["0",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1", ",", "1", ",", "2"}], 
                    "}"}], "\[Rule]", "0"}]],
                 Annotation[#, {1, 1, 1, 1, 2} -> 0, "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["0",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1", ",", "2", ",", "1"}], 
                    "}"}], "\[Rule]", "0"}]],
                 Annotation[#, {1, 1, 1, 2, 1} -> 0, "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["0",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1", ",", "2", ",", "2"}], 
                    "}"}], "\[Rule]", "0"}]],
                 Annotation[#, {1, 1, 1, 2, 2} -> 0, "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
            {GridBox[{
               {
                TagBox[
                 TooltipBox["0",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2", ",", "1", ",", "1"}], 
                    "}"}], "\[Rule]", "0"}]],
                 Annotation[#, {1, 1, 2, 1, 1} -> 0, "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["0",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2", ",", "1", ",", "2"}], 
                    "}"}], "\[Rule]", "0"}]],
                 Annotation[#, {1, 1, 2, 1, 2} -> 0, "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["0",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2", ",", "2", ",", "1"}], 
                    "}"}], "\[Rule]", "0"}]],
                 Annotation[#, {1, 1, 2, 2, 1} -> 0, "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["0",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2", ",", "2", ",", "2"}], 
                    "}"}], "\[Rule]", "0"}]],
                 Annotation[#, {1, 1, 2, 2, 2} -> 0, "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
            {GridBox[{
               {
                TagBox[
                 TooltipBox["0",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1", ",", "1", ",", "1"}], 
                    "}"}], "\[Rule]", "0"}]],
                 Annotation[#, {1, 2, 1, 1, 1} -> 0, "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["0",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1", ",", "1", ",", "2"}], 
                    "}"}], "\[Rule]", "0"}]],
                 Annotation[#, {1, 2, 1, 1, 2} -> 0, "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1", ",", "2", ",", "1"}], 
                    "}"}], "\[Rule]", 
                    RowBox[{
                    FractionBox["1", "4"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "4"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    FractionBox[
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    FractionBox[
                    RowBox[{"8", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    FractionBox[
                    RowBox[{"8", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"3", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}]],
                 
                 Annotation[#, {1, 2, 1, 2, 1} -> 
                  Rational[1, 4] (((-4) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (4 Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (
                    4 (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                    Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + (
                    4 (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[2, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[2, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[2, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[2, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[2, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 4 Derivative[2, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[3, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1", ",", "2", ",", "2"}], 
                    "}"}], "\[Rule]", 
                    RowBox[{
                    FractionBox["1", "4"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "4"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "3"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    FractionBox[
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{"8", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    FractionBox[
                    RowBox[{"8", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}]],
                 
                 Annotation[#, {1, 2, 1, 2, 2} -> 
                  Rational[1, 4] (((-4) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - (4 
                    Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 2][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 2][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 3][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 2][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 
                    Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 
                    Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                    Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[0, 2][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 4 Derivative[1, 2][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                    Derivative[2, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2", ",", "1", ",", "1"}], 
                    "}"}], "\[Rule]", 
                    RowBox[{
                    FractionBox["1", "4"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{"8", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{"8", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"3", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}]],
                 
                 Annotation[#, {1, 2, 2, 1, 1} -> 
                  Rational[1, 4] ((4 Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - (
                    4 (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - (4 
                    Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (
                    4 (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[2, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[2, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[2, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[2, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[2, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - 4 Derivative[2, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[3, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2", ",", "1", ",", "2"}], 
                    "}"}], "\[Rule]", 
                    RowBox[{
                    FractionBox["1", "4"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "3"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    FractionBox[
                    RowBox[{"8", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{"8", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}]],
                 
                 Annotation[#, {1, 2, 2, 1, 2} -> 
                  Rational[1, 4] ((4 Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                    Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 2][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 2][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 3][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 2][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (4 
                    Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 
                    Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - (4 
                    Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[0, 2][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - 4 Derivative[1, 2][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + 
                    2 Derivative[2, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["0",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2", ",", "2", ",", "1"}], 
                    "}"}], "\[Rule]", "0"}]],
                 Annotation[#, {1, 2, 2, 2, 1} -> 0, "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["0",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2", ",", "2", ",", "2"}], 
                    "}"}], "\[Rule]", "0"}]],
                 Annotation[#, {1, 2, 2, 2, 2} -> 0, "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]},
         {GridBox[{
            {GridBox[{
               {
                TagBox[
                 TooltipBox["0",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1", ",", "1", ",", "1"}], 
                    "}"}], "\[Rule]", "0"}]],
                 Annotation[#, {2, 1, 1, 1, 1} -> 0, "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["0",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1", ",", "1", ",", "2"}], 
                    "}"}], "\[Rule]", "0"}]],
                 Annotation[#, {2, 1, 1, 1, 2} -> 0, "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1", ",", "2", ",", "1"}], 
                    "}"}], "\[Rule]", 
                    RowBox[{
                    FractionBox["1", "4"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{"8", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{"8", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"3", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}]],
                 
                 Annotation[#, {2, 1, 1, 2, 1} -> 
                  Rational[1, 4] ((4 Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - (
                    4 (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - (4 
                    Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (
                    4 (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[2, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[2, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[2, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[2, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[2, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - 4 Derivative[2, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[3, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1", ",", "2", ",", "2"}], 
                    "}"}], "\[Rule]", 
                    RowBox[{
                    FractionBox["1", "4"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "3"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    FractionBox[
                    RowBox[{"8", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{"8", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}]],
                 
                 Annotation[#, {2, 1, 1, 2, 2} -> 
                  Rational[1, 4] ((4 Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                    Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 2][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 2][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 3][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 2][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (4 
                    Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 
                    Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - (4 
                    Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[0, 2][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - 4 Derivative[1, 2][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + 
                    2 Derivative[2, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2", ",", "1", ",", "1"}], 
                    "}"}], "\[Rule]", 
                    RowBox[{
                    FractionBox["1", "4"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "4"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    FractionBox[
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    FractionBox[
                    RowBox[{"8", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    FractionBox[
                    RowBox[{"8", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"3", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}]],
                 
                 Annotation[#, {2, 1, 2, 1, 1} -> 
                  Rational[1, 4] (((-4) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (4 Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (
                    4 (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                    Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + (
                    4 (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[2, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[2, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[2, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[2, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[2, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 4 Derivative[2, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[3, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2", ",", "1", ",", "2"}], 
                    "}"}], "\[Rule]", 
                    RowBox[{
                    FractionBox["1", "4"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "4"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "3"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    FractionBox[
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{"8", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    FractionBox[
                    RowBox[{"8", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}]],
                 
                 Annotation[#, {2, 1, 2, 1, 2} -> 
                  Rational[1, 4] (((-4) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - (4 
                    Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 2][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 2][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 3][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 2][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 
                    Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 
                    Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                    Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[0, 2][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 4 Derivative[1, 2][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                    Derivative[2, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["0",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2", ",", "2", ",", "1"}], 
                    "}"}], "\[Rule]", "0"}]],
                 Annotation[#, {2, 1, 2, 2, 1} -> 0, "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["0",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2", ",", "2", ",", "2"}], 
                    "}"}], "\[Rule]", "0"}]],
                 Annotation[#, {2, 1, 2, 2, 2} -> 0, "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
            {GridBox[{
               {
                TagBox[
                 TooltipBox["0",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1", ",", "1", ",", "1"}], 
                    "}"}], "\[Rule]", "0"}]],
                 Annotation[#, {2, 2, 1, 1, 1} -> 0, "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["0",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1", ",", "1", ",", "2"}], 
                    "}"}], "\[Rule]", "0"}]],
                 Annotation[#, {2, 2, 1, 1, 2} -> 0, "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["0",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1", ",", "2", ",", "1"}], 
                    "}"}], "\[Rule]", "0"}]],
                 Annotation[#, {2, 2, 1, 2, 1} -> 0, "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["0",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1", ",", "2", ",", "2"}], 
                    "}"}], "\[Rule]", "0"}]],
                 Annotation[#, {2, 2, 1, 2, 2} -> 0, "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
            {GridBox[{
               {
                TagBox[
                 TooltipBox["0",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2", ",", "1", ",", "1"}], 
                    "}"}], "\[Rule]", "0"}]],
                 Annotation[#, {2, 2, 2, 1, 1} -> 0, "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["0",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2", ",", "1", ",", "2"}], 
                    "}"}], "\[Rule]", "0"}]],
                 Annotation[#, {2, 2, 2, 1, 2} -> 0, "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["0",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2", ",", "2", ",", "1"}], 
                    "}"}], "\[Rule]", "0"}]],
                 Annotation[#, {2, 2, 2, 2, 1} -> 0, "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["0",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2", ",", "2", ",", "2"}], 
                    "}"}], "\[Rule]", "0"}]],
                 Annotation[#, {2, 2, 2, 2, 2} -> 0, "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small],
      Short[#, 1]& ]},
    {"41", 
     TagBox[
      RowBox[{"ToCTensorOneSlot", "[", 
       RowBox[{
        StyleBox[GridBox[{
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 1, 2} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 2, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 2, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                Annotation[#, {1, 2, 1, 1} -> ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                   Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                   Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ], 
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                Annotation[#, {1, 2, 1, 2} -> ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]/(-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                   Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                   Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                
                Annotation[#, {1, 2, 2, 
                  1} -> ((Rational[1, 4] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])/(-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - 4 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ], 
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                
                Annotation[#, {1, 2, 2, 
                  2} -> ((Rational[1, 4] $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]])/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - 4 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]},
           {GridBox[{
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", 
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]}]],
                Annotation[#, {2, 1, 1, 1} -> ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]])/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                   Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                   Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ], 
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", 
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]}]],
                Annotation[#, {2, 1, 1, 2} -> ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]])/(-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                   Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                   Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", 
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]}]],
                
                Annotation[#, {2, 1, 2, 
                  1} -> ((Rational[-1, 4] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])/(-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - 4 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ], 
               TagBox[
                TooltipBox["\<\"\[GrayCircle]\"\>",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", 
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    FractionBox[
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]}]],
                
                Annotation[#, {2, 1, 2, 
                  2} -> ((Rational[-1, 4] $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]])/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                    Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - 4 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 1, 2} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 2, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 2, 2} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
         FontSize->Small], ",", 
        RowBox[{"{", 
         RowBox[{"3", ",", 
          StyleBox[GridBox[{
             {
              TagBox[
               TooltipBox["\<\"\[GrayCircle]\"\>",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                  FractionBox[
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], 
                   RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
               Annotation[#, {1, 1} -> $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]/(-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ], 
              TagBox[
               TooltipBox["\<\"\[GrayCircle]\"\>",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                  FractionBox[
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], 
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
               Annotation[#, {1, 2} -> $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]), "Tooltip"]& ]},
             {
              TagBox[
               TooltipBox["\<\"\[GrayCircle]\"\>",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                  FractionBox[
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], 
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
               Annotation[#, {2, 1} -> $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]), "Tooltip"]& ], 
              TagBox[
               TooltipBox["\<\"\[GrayCircle]\"\>",
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                  FractionBox[
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], 
                   RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
               Annotation[#, {2, 2} -> $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]/(-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]}
            },
            AutoDelete->False,
            
            GridBoxDividers->{
             "ColumnsIndexed" -> {
               1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
              "RowsIndexed" -> {
               1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
           FontSize->Small]}], "}"}]}], "]"}],
      Short[#, 1]& ], 
     TagBox[
      StyleBox[GridBox[{
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "3"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], 
                  RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]}]]}]],
              
              Annotation[#, {1, 2, 1, 
                2} -> (Rational[1, 4] ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) 
                    Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                    Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                   Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                    Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "3"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]}]]}]}]],
              Annotation[#, {1, 2, 2, 1} -> (Rational[-1, 4] ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) 
                    Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                    Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                   Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                    Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "3"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]}]]}]}]],
              Annotation[#, {2, 1, 1, 2} -> (Rational[-1, 4] ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) 
                    Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                    Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                   Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                    Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "3"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "+", 
                    
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], 
                  RowBox[{"4", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"]}]]}]],
              
              Annotation[#, {2, 1, 2, 
                1} -> (Rational[1, 4] ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) 
                    Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                    Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                   Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                    Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                    Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small],
      Short[#, 1]& ]},
    {"42", 
     TagBox[
      RowBox[{"Torsion", "[", 
       RowBox[{"CCovD", "[", 
        RowBox[{"PDgen", ",", 
         StyleBox[GridBox[{
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]}]],
                 Annotation[#, {1, 1, 1} -> (Rational[-1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 1, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 2, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 2, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 1, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {2, 1, 2} -> (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]]}]],
                 Annotation[#, {2, 2, 1} -> (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]}]],
                 Annotation[#, {2, 2, 2} -> (Rational[-1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small], ",", 
         StyleBox[GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 1} -> $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 2} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 1} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 2} -> $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small]}], "]"}], "]"}],
      Short[#, 1]& ], 
     TagBox["Zero",
      Short[#, 1]& ]},
    {"43", 
     TagBox[
      RowBox[{"Torsion", "[", 
       RowBox[{"CCovD", "[", 
        RowBox[{"PDgen", ",", 
         StyleBox[GridBox[{
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]}]],
                 Annotation[#, {1, 1, 1} -> (Rational[-1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 1, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((((-2) $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 2, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {1, 2, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
            {GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 1, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 1, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 2, 
                   1} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]},
               {
                TagBox[
                 TooltipBox["\<\"\[GrayCircle]\"\>",
                  RowBox[{
                    RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]]}]],
                 
                 Annotation[#, {2, 2, 
                   2} -> (Rational[1, 2]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), "Tooltip"]& ]}
              },
              AutoDelete->False,
              
              GridBoxDividers->{
               "ColumnsIndexed" -> {
                 1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small], ",", 
         StyleBox[GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxx,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 1} -> $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 2} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gxy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gxy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 1} -> $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["gyy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`gyy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 2} -> $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
          FontSize->Small]}], "]"}], "]"}],
      Short[#, 1]& ], 
     TagBox[
      StyleBox[GridBox[{
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["T112",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T112,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 1, 2} -> $CellContext`T112[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{"-", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]],
              Annotation[#, {1, 2, 1} -> -$CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["T212",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T212,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 1, 2} -> $CellContext`T212[
                 $CellContext`x[], 
                 $CellContext`y[]], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{"-", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False]}], "]"}]}]}]],
              Annotation[#, {2, 2, 1} -> -$CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 3 -> RGBColor[1, 0, 0]}}],
       FontSize->Small],
      Short[#, 1]& ]}
   },
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}},
   GridBoxSpacings->{"Columns" -> {
       Offset[0.27999999999999997`], {
        Offset[2.0999999999999996`]}, 
       Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
       Offset[0.2], {
        Offset[0.4]}, 
       Offset[0.2]}, "RowsIndexed" -> {}}],
  TableForm[{{
     "ReportCompute"[
      "ADChristoffel"[-1, -1, -1, -1]], {
      Skeleton[4]}}, {
     "ReportCompute"[
      "CDRiemann"[-1, -1, -1, -1, -1]], {
      Skeleton[2]}}, {
     "ReportCompute"[
      "Christoffel"[-1, -1, -1]], {
      Skeleton[8]}}, {
     "ReportCompute"[
      "Christoffel"[1, -1, -1]], {
      Skeleton[8]}}, {
     "ReportCompute"[
      "DetMetric"[]], {
      Skeleton[1]}}, {
     "ReportCompute"[
      "DMetric"[-1, -1, -1]], {
      Skeleton[6]}}, {
     "ReportCompute"[
      "Einstein"[-1, -1]], {
      Skeleton[4]}}, {
     "ReportCompute"[
      "Kretschmann"[]], {
      Skeleton[1]}}, {
     "ReportCompute"[
      "Metric"[1, 1]], {
      Skeleton[3]}}, {
     "ReportCompute"[
      "Ricci"[-1, -1]], {
      Skeleton[4]}}, {
     "ReportCompute"[
      "RicciScalar"[]], {
      Skeleton[1]}}, {
     "ReportCompute"[
      "Riemann"[-1, -1, -1, -1]], {
      Skeleton[1]}}, {
     "ReportCompute"[
      "Riemann"[-1, -1, -1, 1]], {
      Skeleton[4]}}, {
     "ReportCompute"[
      "Riemann"[-1, -1, 1, 1]], {
      Skeleton[1]}}, {
     "ReportCompute"[
      "Torsion"[-1, -1, -1]], {
      Skeleton[2]}}, {
     "ChartsOf"[
      Shallow[{{
         $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[]], 
         $CellContext`gxy[
          $CellContext`x[], 
          $CellContext`y[]]}, {
         $CellContext`gxy[
          $CellContext`x[], 
          $CellContext`y[]], 
         $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[]]}}, 1]], 
     Short[
      HoldForm[{$CellContext`gen}], 1]}, {
     Short[
      HoldForm[
       xAct`xTensor`Christoffel[
        xAct`xCoba`CCovD[$CellContext`PDgen, 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{{((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}, {((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}, {{(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}, {((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}}, {$CellContext`gen, \
-$CellContext`gen, -$CellContext`gen}, 0], 
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{
              $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]]}, {
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 
            0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]], \
$CellContext`PDgen]], 1], 
     Short[
      HoldForm[
       RawBoxes[
        xAct`xCoba`Private`CTensorBoxes[
         xAct`xCoba`CTensor[{{{((-2) $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 + (2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]) $CellContext`T112[
                 $CellContext`x[], 
                 $CellContext`y[]] + (2 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2) $CellContext`T212[
                 $CellContext`x[], 
                 $CellContext`y[]] + $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] + $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]])), ((-2) $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 + (2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]]) $CellContext`T112[
                 $CellContext`x[], 
                 $CellContext`y[]] + $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] + $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]))}, {((-2) $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 + (2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2) $CellContext`T112[
                 $CellContext`x[], 
                 $CellContext`y[]] + $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] + $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]]^2 - $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]] $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]])) ($CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]))}}, {{(Rational[1, 2]/($CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]]^2 - $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]] $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]])) ((2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]^2) $CellContext`T112[
                 $CellContext`x[], 
                 $CellContext`y[]] + $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] + $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]])), ((-2) $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 + (2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]]) + $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]))}, {((-2) $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 + (2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] + $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])), ((-2) $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 + (2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])^(-1) (((-2) $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2) $CellContext`T112[
                 $CellContext`x[], 
                 $CellContext`y[]] + $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]] + $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]))}}}, {$CellContext`gen, \
-$CellContext`gen, -$CellContext`gen}, 0], 
         xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]], 1]}, {
     Short[
      HoldForm[
       xAct`xTensor`CovDOfMetric[
        RawBoxes[
         xAct`xCoba`Private`CTensorBoxes[
          xAct`xCoba`CTensor[{{
             $CellContext`gxx[
              $CellContext`x[], 
              $CellContext`y[]], 
             $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]]}, {
             $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]], 
             $CellContext`gyy[
              $CellContext`x[], 
              $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 0],
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]]], 1], 
     Short[
      HoldForm[
       xAct`xCoba`CCovD[$CellContext`PDgen, 
        RawBoxes[
         xAct`xCoba`Private`CTensorBoxes[
          xAct`xCoba`CTensor[{{{((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]])), ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]))}, {((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}}, {{(
               Rational[1, 2]/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]])), ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}, {((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])), ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-1) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}}}, {$CellContext`gen, \
-$CellContext`gen, -$CellContext`gen}, 0], 
          xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], 
        RawBoxes[
         xAct`xCoba`Private`CTensorBoxes[
          xAct`xCoba`CTensor[{{
             $CellContext`gxx[
              $CellContext`x[], 
              $CellContext`y[]], 
             $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]]}, {
             $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]], 
             $CellContext`gyy[
              $CellContext`x[], 
              $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 0],
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]]], 1]}, {
     Short[
      HoldForm[
       xAct`xTensor`CovDOfMetric[
        RawBoxes[
         xAct`xCoba`Private`CTensorBoxes[
          xAct`xCoba`CTensor[{{
             $CellContext`gxx[
              $CellContext`x[], 
              $CellContext`y[]], 
             $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]]}, {
             $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]], 
             $CellContext`gyy[
              $CellContext`x[], 
              $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 0],
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], 
        xAct`xTensor`Zero]], 1], 
     Short[
      HoldForm[
       xAct`xCoba`CCovD[$CellContext`PDgen, 
        RawBoxes[
         xAct`xCoba`Private`CTensorBoxes[
          xAct`xCoba`CTensor[{{{(Rational[-1, 2]/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ($CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ((-$CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])}, {(Rational[1, 2]/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ((-$CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]), (Rational[1, 2]/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}}, {{(
               Rational[1, 2]/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]])), (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - (2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])^(-1) ($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]])}, {(2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - (2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])^(-1) ($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]), (Rational[-1, 2]/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ($CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}}}, {$CellContext`gen, \
-$CellContext`gen, -$CellContext`gen}, 0], 
          xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], 
        RawBoxes[
         xAct`xCoba`Private`CTensorBoxes[
          xAct`xCoba`CTensor[{{
             $CellContext`gxx[
              $CellContext`x[], 
              $CellContext`y[]], 
             $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]]}, {
             $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]], 
             $CellContext`gyy[
              $CellContext`x[], 
              $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 0],
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]]], 1]}, {
     Short[
      HoldForm[
       xAct`xTensor`CovDOfMetric[
        RawBoxes[
         xAct`xCoba`Private`CTensorBoxes[
          xAct`xCoba`CTensor[{{
             $CellContext`gxx[
              $CellContext`x[], 
              $CellContext`y[]], 
             $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]]}, {
             $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]], 
             $CellContext`gyy[
              $CellContext`x[], 
              $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 0],
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], 
        RawBoxes[
         xAct`xCoba`Private`CTensorBoxes[
          xAct`xCoba`CTensor[{{{0, 
              $CellContext`T112[
               $CellContext`x[], 
               $CellContext`y[]]}, {-$CellContext`T112[
                $CellContext`x[], 
                $CellContext`y[]], 0}}, {{0, 
              $CellContext`T212[
               $CellContext`x[], 
               $CellContext`y[]]}, {-$CellContext`T212[
                $CellContext`x[], 
                $CellContext`y[]], 
              0}}}, {$CellContext`gen, -$CellContext`gen, -$CellContext`gen}, 
           0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]]], 1], 
     Short[
      HoldForm[
       xAct`xCoba`CCovD[$CellContext`PDgen, 
        RawBoxes[
         xAct`xCoba`Private`CTensorBoxes[
          xAct`xCoba`CTensor[{{{(Rational[-1, 2]/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ((((-2) $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] - $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}, {(Rational[1, 2]/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] - $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}}, {{(
               Rational[1, 2]/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]))}, {(Rational[1, 2]/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]))}}}, {$CellContext`gen, \
-$CellContext`gen, -$CellContext`gen}, 0], 
          xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], 
        RawBoxes[
         xAct`xCoba`Private`CTensorBoxes[
          xAct`xCoba`CTensor[{{
             $CellContext`gxx[
              $CellContext`x[], 
              $CellContext`y[]], 
             $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]]}, {
             $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]], 
             $CellContext`gyy[
              $CellContext`x[], 
              $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 0],
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]]], 1]}, {
     Short[
      HoldForm[
       xAct`xTensor`Determinant[
        RawBoxes[
         xAct`xCoba`Private`CTensorBoxes[
          xAct`xCoba`CTensor[{{
             $CellContext`gxx[
              $CellContext`x[], 
              $CellContext`y[]], 
             $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]]}, {
             $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]], 
             $CellContext`gyy[
              $CellContext`x[], 
              $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 0],
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], \
$CellContext`gen]], 1], 
     Short[
      HoldForm[
       RawBoxes[
        xAct`xCoba`Private`CTensorBoxes[
         xAct`xCoba`CTensor[-$CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]]^2 + $CellContext`gxx[
             $CellContext`x[], 
             $CellContext`y[]] $CellContext`gyy[
             $CellContext`x[], 
             $CellContext`y[]], {}, 2 $CellContext`gen], 
         xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]], 1]}, {
     Short[
      HoldForm[
       xAct`xTensor`Einstein[
        xAct`xCoba`CCovD[$CellContext`PDgen, 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{{(Rational[-1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((((-2) $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}, {(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}, {{(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}, {(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}}}, {$CellContext`gen, \
-$CellContext`gen, -$CellContext`gen}, 0], 
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{
              $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]]}, {
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 
            0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]]]], 1], 
     Short[
      HoldForm[xAct`xTensor`Zero], 1]}, {
     Short[
      HoldForm[
       xAct`xTensor`Einstein[
        xAct`xCoba`CCovD[$CellContext`PDgen, 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{{((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}, {((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}, {{(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}, {((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}}, {$CellContext`gen, \
-$CellContext`gen, -$CellContext`gen}, 0], 
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{
              $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]]}, {
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 
            0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]]]], 1], 
     Short[
      HoldForm[xAct`xTensor`Zero], 1]}, {
     Short[
      HoldForm[
       xAct`xTensor`Inv[
        RawBoxes[
         xAct`xCoba`Private`CTensorBoxes[
          xAct`xCoba`CTensor[{{
             $CellContext`gxx[
              $CellContext`x[], 
              $CellContext`y[]], 
             $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]]}, {
             $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]], 
             $CellContext`gyy[
              $CellContext`x[], 
              $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 0],
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]]], 1], 
     Short[
      HoldForm[
       RawBoxes[
        xAct`xCoba`Private`CTensorBoxes[
         xAct`xCoba`CTensor[{{$CellContext`gyy[
              $CellContext`x[], 
              $CellContext`y[]]/(-$CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]]^2 + $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]] $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]]), $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]]/($CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]]^2 - $CellContext`gxx[
              $CellContext`x[], 
              $CellContext`y[]] $CellContext`gyy[
              $CellContext`x[], 
              $CellContext`y[]])}, {$CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]]/($CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]]^2 - $CellContext`gxx[
              $CellContext`x[], 
              $CellContext`y[]] $CellContext`gyy[
              $CellContext`x[], 
              $CellContext`y[]]), $CellContext`gxx[
              $CellContext`x[], 
              $CellContext`y[]]/(-$CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]]^2 + $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]] $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]])}}, {$CellContext`gen, $CellContext`gen}, 0],
          xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]], 1]}, {
     Short[
      HoldForm[
       xAct`xTensor`Kretschmann[
        xAct`xCoba`CCovD[$CellContext`PDgen, 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{{((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}, {((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}, {{(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}, {((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}}, {$CellContext`gen, \
-$CellContext`gen, -$CellContext`gen}, 0], 
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{
              $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]]}, {
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 
            0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]]]], 1], 
     Short[
      HoldForm[
       RawBoxes[
        xAct`xCoba`Private`CTensorBoxes[
         xAct`xCoba`CTensor[(Rational[1, 4] ($CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]]^2 - $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]] $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]])^(-4)) ((2 $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]]^2) ($CellContext`T112[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]] - (2 $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                $CellContext`x[], 
                $CellContext`y[]]) + $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])) + $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]] (((-2) Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] + (4 Derivative[0, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                $CellContext`x[], 
                $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]] - (2 Derivative[1, 0][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]] - (2 $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]]) ((-$CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) + (2 $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]]))) + (4 $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]]^3) (Derivative[0, 1][$CellContext`T212][
                $CellContext`x[], 
                $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
               $CellContext`x[], 
               $CellContext`y[]]) + (2 $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]]^2) ((2 $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] + (2 $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                $CellContext`x[], 
                $CellContext`y[]] - (2 $CellContext`T212[
                 $CellContext`x[], 
                 $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]] - (2 $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                $CellContext`x[], 
                $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                $CellContext`x[], 
                $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]]) + $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]] - (2 Derivative[0, 1][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                $CellContext`x[], 
                $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]^2 + (4 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                 $CellContext`x[], 
                 $CellContext`y[]] - (2 $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] + $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) - 2 Derivative[1, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]])))^2, {}, 0], 
         xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]], 1]}, {
     Short[
      HoldForm[
       xAct`xTensor`LC[
        RawBoxes[
         xAct`xCoba`Private`CTensorBoxes[
          xAct`xCoba`CTensor[{{
             $CellContext`gxx[
              $CellContext`x[], 
              $CellContext`y[]], 
             $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]]}, {
             $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]], 
             $CellContext`gyy[
              $CellContext`x[], 
              $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 0],
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]]], 1], 
     Short[
      HoldForm[
       xAct`xCoba`CCovD[$CellContext`PDgen, 
        RawBoxes[
         xAct`xCoba`Private`CTensorBoxes[
          xAct`xCoba`CTensor[{{{(Rational[-1, 2]/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ($CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ((-$CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])}, {(Rational[1, 2]/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ((-$CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]), (Rational[1, 2]/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}}, {{(
               Rational[1, 2]/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]])), (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - (2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])^(-1) ($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]])}, {(2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - (2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])^(-1) ($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]), (Rational[-1, 2]/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ($CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}}}, {$CellContext`gen, \
-$CellContext`gen, -$CellContext`gen}, 0], 
          xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], 
        RawBoxes[
         xAct`xCoba`Private`CTensorBoxes[
          xAct`xCoba`CTensor[{{
             $CellContext`gxx[
              $CellContext`x[], 
              $CellContext`y[]], 
             $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]]}, {
             $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]], 
             $CellContext`gyy[
              $CellContext`x[], 
              $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 0],
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]]], 1]}, {
     Short[
      HoldForm[
       xAct`xTensor`Ricci[
        xAct`xCoba`CCovD[$CellContext`PDgen, 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{{(Rational[-1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])}, {(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((-2) 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}, {{(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-1) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])}, {(2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-1) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]), (Rational[-1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((-2) 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}}, {$CellContext`gen, \
-$CellContext`gen, -$CellContext`gen}, 0], 
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{
              $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]]}, {
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 
            0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]]]], 1], 
     Short[
      HoldForm[
       RawBoxes[
        xAct`xCoba`Private`CTensorBoxes[
         xAct`xCoba`CTensor[{{((Rational[1, 4] $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]]) ($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])^(-2)) ($CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]^2) + $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]] (Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + (2 Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])) + (2 $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2) (Derivative[0, 2][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]) + $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) - (2 $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]))), ((Rational[1, 4] $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]]) ($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])^(-2)) ($CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]^2) + $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]] (Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + (2 Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])) + (2 $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2) (Derivative[0, 2][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]) + $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) - (2 $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])))}, {((Rational[1, 4] $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]]) ($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])^(-2)) ($CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]^2) + $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]] (Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + (2 Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])) + (2 $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2) (Derivative[0, 2][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]) + $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) - (2 $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]))), ((Rational[1, 4] $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]]) ($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])^(-2)) ($CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]^2) + $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]] (Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + (2 Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])) + (2 $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2) (Derivative[0, 2][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]) + $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) - (2 $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])))}}, {-$CellContext`gen, \
-$CellContext`gen}, 0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]], 
      1]}, {
     Short[
      HoldForm[
       xAct`xTensor`Ricci[
        xAct`xCoba`CCovD[$CellContext`PDgen, 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{{(Rational[-1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((((-2) $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}, {(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}, {{(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}, {(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}}}, {$CellContext`gen, \
-$CellContext`gen, -$CellContext`gen}, 0], 
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{
              $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]]}, {
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 
            0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]]]], 1], 
     Short[
      HoldForm[
       RawBoxes[
        xAct`xCoba`Private`CTensorBoxes[
         xAct`xCoba`CTensor[{{((Rational[1, 4] $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]]) ($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]]^2) ($CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]]) + $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) + $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]] (((-2) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + (4 Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]] - (2 Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]] - (2 $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]]))) + (4 $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^3) (Derivative[0, 1][$CellContext`T212][
                 $CellContext`x[], 
                 $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                $CellContext`x[], 
                $CellContext`y[]]) + (2 $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2) ((2 $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + (2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] - (2 $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]] - (2 $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                 $CellContext`x[], 
                 $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]) + $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) - 2 Derivative[1, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]))), ((Rational[1, 4] $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]]) ($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]]^2) ($CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]]) + $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) + $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]] (((-2) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + (4 Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]] - (2 Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]] - (2 $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]]))) + (4 $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^3) (Derivative[0, 1][$CellContext`T212][
                 $CellContext`x[], 
                 $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                $CellContext`x[], 
                $CellContext`y[]]) + (2 $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2) ((2 $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + (2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] - (2 $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]] - (2 $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                 $CellContext`x[], 
                 $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]) + $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) - 2 Derivative[1, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])))}, {((Rational[1, 4] $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]]) ($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]]^2) ($CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]]) + $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) + $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]] (((-2) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + (4 Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]] - (2 Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]] - (2 $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]]))) + (4 $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^3) (Derivative[0, 1][$CellContext`T212][
                 $CellContext`x[], 
                 $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                $CellContext`x[], 
                $CellContext`y[]]) + (2 $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2) ((2 $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + (2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] - (2 $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]] - (2 $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                 $CellContext`x[], 
                 $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]) + $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) - 2 Derivative[1, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]))), ((Rational[1, 4] $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]]) ($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]]^2) ($CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]]) + $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) + $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]] (((-2) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + (4 Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]] - (2 Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]] - (2 $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]]))) + (4 $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^3) (Derivative[0, 1][$CellContext`T212][
                 $CellContext`x[], 
                 $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                $CellContext`x[], 
                $CellContext`y[]]) + (2 $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2) ((2 $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + (2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] - (2 $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]] - (2 $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                 $CellContext`x[], 
                 $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]) + $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) - 2 Derivative[1, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])))}}, {-$CellContext`gen, \
-$CellContext`gen}, 0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]], 
      1]}, {
     Short[
      HoldForm[
       xAct`xTensor`Ricci[
        xAct`xCoba`CCovD[$CellContext`PDgen, 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{{((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}, {((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}, {{(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}, {((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}}, {$CellContext`gen, \
-$CellContext`gen, -$CellContext`gen}, 0], 
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{
              $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]]}, {
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 
            0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]]]], 1], 
     Short[
      HoldForm[
       RawBoxes[
        xAct`xCoba`Private`CTensorBoxes[
         xAct`xCoba`CTensor[{{($CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]]/(-$CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]]^2 + $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]] $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]])) ((-$CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                $CellContext`x[], 
                $CellContext`y[]] - $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
               $CellContext`x[], 
               $CellContext`y[]] + 
             Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                $CellContext`x[], 
                $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]])) + $CellContext`T112[
                $CellContext`x[], 
                $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]))) ($CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`T112[
                 $CellContext`x[], 
                 $CellContext`y[]] + $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`T212[
                 $CellContext`x[], 
                 $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                $CellContext`x[], 
                $CellContext`y[]] + 
              Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]) + $CellContext`T212[
                $CellContext`x[], 
                $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]) + ((
                Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])) ((-$CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) + $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                $CellContext`x[], 
                $CellContext`y[]] + $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                $CellContext`x[], 
                $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
               $CellContext`x[], 
               $CellContext`y[]] + 
             Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]]), ((Rational[1, 4] $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]])/($CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]]^2 - $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]] $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]])) ((4 $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                $CellContext`x[], 
                $CellContext`y[]] + (4 $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                $CellContext`x[], 
                $CellContext`y[]] + 2 Derivative[0, 2][$CellContext`gxx][
                $CellContext`x[], 
                $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-1) 
               Derivative[0, 1][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]])) + (4 $CellContext`T112[
                 $CellContext`x[], 
                 $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                $CellContext`x[], 
                $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-1) 
               Derivative[0, 1][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-1) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))) ((2 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] + (2 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]]) + (4 $CellContext`T212[
                 $CellContext`x[], 
                 $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]]) + ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + 
               Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) - (4 $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
               $CellContext`x[], 
               $CellContext`y[]] - (4 $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
               $CellContext`x[], 
               $CellContext`y[]] - 4 Derivative[1, 1][$CellContext`gxy][
               $CellContext`x[], 
               $CellContext`y[]] + 2 Derivative[2, 0][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]])}, {((-$CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]])/($CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]]^2 - $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]] $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]])) ((-$CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                $CellContext`x[], 
                $CellContext`y[]] - $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
               $CellContext`x[], 
               $CellContext`y[]] + 
             Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                $CellContext`x[], 
                $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]])) + $CellContext`T112[
                $CellContext`x[], 
                $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]))) ($CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`T112[
                 $CellContext`x[], 
                 $CellContext`y[]] + $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`T212[
                 $CellContext`x[], 
                 $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                $CellContext`x[], 
                $CellContext`y[]] + 
              Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]) + $CellContext`T212[
                $CellContext`x[], 
                $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]) + ((
                Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])) ((-$CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) + $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                $CellContext`x[], 
                $CellContext`y[]] + $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                $CellContext`x[], 
                $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
               $CellContext`x[], 
               $CellContext`y[]] + 
             Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]]), ((Rational[-1, 4] $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]])/(-$CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]]^2 + $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]] $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]])) ((4 $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                $CellContext`x[], 
                $CellContext`y[]] + (4 $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                $CellContext`x[], 
                $CellContext`y[]] + 2 Derivative[0, 2][$CellContext`gxx][
                $CellContext`x[], 
                $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-1) 
               Derivative[0, 1][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]])) + (4 $CellContext`T112[
                 $CellContext`x[], 
                 $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                $CellContext`x[], 
                $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-1) 
               Derivative[0, 1][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-1) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))) ((2 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] + (2 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]]) + (4 $CellContext`T212[
                 $CellContext`x[], 
                 $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]]) + ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + 
               Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) - (4 $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
               $CellContext`x[], 
               $CellContext`y[]] - (4 $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
               $CellContext`x[], 
               $CellContext`y[]] - 4 Derivative[1, 1][$CellContext`gxy][
               $CellContext`x[], 
               $CellContext`y[]] + 2 Derivative[2, 0][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]])}}, {-$CellContext`gen, -$CellContext`gen}, 
          0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]], 1]}, {
     Short[
      HoldForm[
       xAct`xTensor`RicciScalar[
        xAct`xCoba`CCovD[$CellContext`PDgen, 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{{(Rational[-1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])}, {(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((-2) 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}, {{(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-1) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])}, {(2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-1) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]), (Rational[-1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((-2) 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}}, {$CellContext`gen, \
-$CellContext`gen, -$CellContext`gen}, 0], 
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{
              $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]]}, {
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 
            0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]]]], 1], 
     Short[
      HoldForm[
       RawBoxes[
        xAct`xCoba`Private`CTensorBoxes[
         xAct`xCoba`CTensor[(Rational[1, 2] ($CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]]^2 - $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]] $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]])^(-2)) ($CellContext`gxx[
              $CellContext`x[], 
              $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                $CellContext`x[], 
                $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]] - (2 Derivative[0, 1][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
               $CellContext`x[], 
               $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]]^2) + $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]] (Derivative[0, 1][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                $CellContext`x[], 
                $CellContext`y[]] + (2 Derivative[1, 0][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]]) (2 Derivative[0, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxx][
               $CellContext`x[], 
               $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]])) + (2 $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]]^2) (Derivative[0, 2][$CellContext`gxx][
               $CellContext`x[], 
               $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
               $CellContext`x[], 
               $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
               $CellContext`x[], 
               $CellContext`y[]]) + $CellContext`gyy[
              $CellContext`x[], 
              $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                $CellContext`x[], 
                $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                $CellContext`x[], 
                $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]) - (2 $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                $CellContext`x[], 
                $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                $CellContext`x[], 
                $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]]))), {}, 0], 
         xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]], 1]}, {
     Short[
      HoldForm[
       xAct`xTensor`RicciScalar[
        xAct`xCoba`CCovD[$CellContext`PDgen, 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{{(Rational[-1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((((-2) $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}, {(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}, {{(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}, {(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}}}, {$CellContext`gen, \
-$CellContext`gen, -$CellContext`gen}, 0], 
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{
              $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]]}, {
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 
            0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]]]], 1], 
     Short[
      HoldForm[
       RawBoxes[
        xAct`xCoba`Private`CTensorBoxes[
         xAct`xCoba`CTensor[(Rational[1, 2] ($CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]]^2 - $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]] $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]]^2) ($CellContext`T112[
                $CellContext`x[], 
                $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]] - (2 $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
               $CellContext`x[], 
               $CellContext`y[]]) + $CellContext`gyy[
              $CellContext`x[], 
              $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                $CellContext`x[], 
                $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                $CellContext`x[], 
                $CellContext`y[]] (((-2) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]])) + $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]] (((-2) Derivative[0, 1][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                $CellContext`x[], 
                $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                $CellContext`x[], 
                $CellContext`y[]] + (4 Derivative[0, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                $CellContext`x[], 
                $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
               $CellContext`x[], 
               $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
               $CellContext`x[], 
               $CellContext`y[]] - (2 Derivative[1, 0][$CellContext`gxy][
                $CellContext`x[], 
                $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
               $CellContext`x[], 
               $CellContext`y[]] - (2 $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]]) ((-$CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]] + $CellContext`T112[
                 $CellContext`x[], 
                 $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])) + (2 $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]]) ((-$CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]]))) + (4 $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]]^3) (Derivative[0, 1][$CellContext`T212][
               $CellContext`x[], 
               $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
              $CellContext`x[], 
              $CellContext`y[]]) + (2 $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]]^2) ((2 $CellContext`T112[
                 $CellContext`x[], 
                 $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                $CellContext`x[], 
                $CellContext`y[]] + (2 $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                $CellContext`x[], 
                $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
               $CellContext`x[], 
               $CellContext`y[]] - (2 $CellContext`T212[
                $CellContext`x[], 
                $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
               $CellContext`x[], 
               $CellContext`y[]] - (2 $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
               $CellContext`x[], 
               $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
               $CellContext`x[], 
               $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
               $CellContext`x[], 
               $CellContext`y[]]) + $CellContext`gxx[
              $CellContext`x[], 
              $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                $CellContext`x[], 
                $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]] - (2 Derivative[0, 1][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
               $CellContext`x[], 
               $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]]^2 + (4 $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                $CellContext`x[], 
                $CellContext`y[]] - (2 $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                $CellContext`x[], 
                $CellContext`y[]] + $CellContext`T112[
                 $CellContext`x[], 
                 $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + $CellContext`T212[
                 $CellContext`x[], 
                 $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]) - 2 Derivative[1, 1][$CellContext`gxy][
                $CellContext`x[], 
                $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]]))), {}, 0], 
         xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]], 1]}, {
     Short[
      HoldForm[
       xAct`xTensor`RicciScalar[
        xAct`xCoba`CCovD[$CellContext`PDgen, 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{{((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}, {((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}, {{(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}, {((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}}, {$CellContext`gen, \
-$CellContext`gen, -$CellContext`gen}, 0], 
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{
              $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]]}, {
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 
            0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]]]], 1], 
     Short[
      HoldForm[
       RawBoxes[
        xAct`xCoba`Private`CTensorBoxes[
         xAct`xCoba`CTensor[(Rational[1, 2] ($CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]]^2 - $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]] $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]]^2) ($CellContext`T112[
                $CellContext`x[], 
                $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]] - (2 $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
               $CellContext`x[], 
               $CellContext`y[]]) + $CellContext`gyy[
              $CellContext`x[], 
              $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                $CellContext`x[], 
                $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                $CellContext`x[], 
                $CellContext`y[]] (((-2) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]])) + $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]] (((-2) Derivative[0, 1][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                $CellContext`x[], 
                $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                $CellContext`x[], 
                $CellContext`y[]] + (4 Derivative[0, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                $CellContext`x[], 
                $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
               $CellContext`x[], 
               $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
               $CellContext`x[], 
               $CellContext`y[]] - (2 Derivative[1, 0][$CellContext`gxy][
                $CellContext`x[], 
                $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
               $CellContext`x[], 
               $CellContext`y[]] - (2 $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]]) ((-$CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]] + $CellContext`T112[
                 $CellContext`x[], 
                 $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])) + (2 $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]]) ((-$CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]]))) + (4 $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]]^3) (Derivative[0, 1][$CellContext`T212][
               $CellContext`x[], 
               $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
              $CellContext`x[], 
              $CellContext`y[]]) + (2 $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]]^2) ((2 $CellContext`T112[
                 $CellContext`x[], 
                 $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                $CellContext`x[], 
                $CellContext`y[]] + (2 $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                $CellContext`x[], 
                $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
               $CellContext`x[], 
               $CellContext`y[]] - (2 $CellContext`T212[
                $CellContext`x[], 
                $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
               $CellContext`x[], 
               $CellContext`y[]] - (2 $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
               $CellContext`x[], 
               $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
               $CellContext`x[], 
               $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
               $CellContext`x[], 
               $CellContext`y[]]) + $CellContext`gxx[
              $CellContext`x[], 
              $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                $CellContext`x[], 
                $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]] - (2 Derivative[0, 1][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
               $CellContext`x[], 
               $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]]^2 + (4 $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                $CellContext`x[], 
                $CellContext`y[]] - (2 $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                $CellContext`x[], 
                $CellContext`y[]] + $CellContext`T112[
                 $CellContext`x[], 
                 $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + $CellContext`T212[
                 $CellContext`x[], 
                 $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]) - 2 Derivative[1, 1][$CellContext`gxy][
                $CellContext`x[], 
                $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                $CellContext`x[], 
                $CellContext`y[]]))), {}, 0], 
         xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]], 1]}, {
     Short[
      HoldForm[
       xAct`xTensor`Riemann[
        xAct`xCoba`CCovD[$CellContext`PDgen, 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{{(Rational[-1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])}, {(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((-2) 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}, {{(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-1) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])}, {(2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-1) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]), (Rational[-1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((-2) 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}}, {$CellContext`gen, \
-$CellContext`gen, -$CellContext`gen}, 0], 
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{
              $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]]}, {
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 
            0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]]]], 1], 
     Short[
      HoldForm[
       RawBoxes[
        xAct`xCoba`Private`CTensorBoxes[
         xAct`xCoba`CTensor[{{{{0, 0}, {0, 
            0}}, {{((Rational[-1, 4] $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-2)) ($CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                  Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 
                   Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) (Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), ((Rational[1, 4] $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-2)) ($CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                  Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 
                   Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) (Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])))}, {((Rational[-1, 4] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-2)) ($CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                  Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 
                   Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) (Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), ((Rational[1, 4] $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-2)) ($CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                  Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 
                   Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) (Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])))}}}, {{{((
                Rational[1, 4] $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-2)) ($CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                  Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 
                   Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) (Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), ((Rational[1, 4] $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-2)) ((-$CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                  Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] ((-Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 
                   Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-2) 
                    Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (2 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2) (Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) - $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] ((-2) 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])))}, {((Rational[1, 4] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-2)) ($CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                  Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 
                   Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) (Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), ((Rational[-1, 4] $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-2)) ($CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                  Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 
                   Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) (Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] ((-2) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])))}}, {{0, 0}, {0, 
            0}}}}, {-$CellContext`gen, -$CellContext`gen, -$CellContext`gen, \
$CellContext`gen}, 0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]], 
      1]}, {
     Short[
      HoldForm[
       xAct`xTensor`Riemann[
        xAct`xCoba`CCovD[$CellContext`PDgen, 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{{(Rational[-1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((((-2) $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}, {(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}, {{(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}, {(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}}}, {$CellContext`gen, \
-$CellContext`gen, -$CellContext`gen}, 0], 
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{
              $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]]}, {
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 
            0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]]]], 1], 
     Short[
      HoldForm[
       RawBoxes[
        xAct`xCoba`Private`CTensorBoxes[
         xAct`xCoba`CTensor[{{{{0, 0}, {0, 
            0}}, {{((Rational[-1, 4] $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (((-2) 
                   Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                 Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]]) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                  Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                  Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), ((Rational[1, 4] $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (((-2) 
                   Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                 Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]]) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                  Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                  Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])))}, {((Rational[-1, 4] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (((-2) 
                   Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                 Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]]) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                  Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                  Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), ((Rational[1, 4] $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (((-2) 
                   Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                 Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]]) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                  Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                  Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])))}}}, {{{((
                Rational[1, 4] $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (((-2) 
                   Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                 Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]]) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                  Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                  Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), ((Rational[-1, 4] $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (((-2) 
                   Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                 Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]]) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                  Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                  Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])))}, {((
                Rational[1, 4] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (((-2) 
                   Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                 Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]]) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                  Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                  Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), ((Rational[-1, 4] $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (((-2) 
                   Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                 Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]]) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                  Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                  Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])))}}, {{0, 0}, {0, 
            0}}}}, {-$CellContext`gen, -$CellContext`gen, -$CellContext`gen, \
$CellContext`gen}, 0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]], 
      1]}, {
     Short[
      HoldForm[
       xAct`xTensor`Riemann[
        xAct`xCoba`CCovD[$CellContext`PDgen, 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{{((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}, {((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}, {{(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}, {((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}}, {$CellContext`gen, \
-$CellContext`gen, -$CellContext`gen}, 0], 
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{
              $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]]}, {
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 
            0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]]]], 1], 
     Short[
      HoldForm[
       RawBoxes[
        xAct`xCoba`Private`CTensorBoxes[
         xAct`xCoba`CTensor[{{{{0, 0}, {0, 0}}, {{($CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]]/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ((-$CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] - $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                 $CellContext`x[], 
                 $CellContext`y[]] + 
               Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + 
                Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + ((
                  Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + 
               Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]), ($CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]]/(-$CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((-$CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] - $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                 $CellContext`x[], 
                 $CellContext`y[]] + 
               Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + 
                Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + ((
                  Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + 
               Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])}, {((Rational[1, 4] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])/(-$CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((4 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] + (4 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] + 2 Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + (4 $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]] - (4 $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                 $CellContext`x[], 
                 $CellContext`y[]] - 4 Derivative[1, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + 2 Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]), ((Rational[1, 4] $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]])/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ((4 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] + (4 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] + 2 Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + (4 $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]] - (4 $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                 $CellContext`x[], 
                 $CellContext`y[]] - 4 Derivative[1, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + 2 Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])}}}, {{{((-$CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]])/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ((-$CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] - $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                 $CellContext`x[], 
                 $CellContext`y[]] + 
               Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + 
                Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + ((
                  Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + 
               Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]), ((-$CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]])/(-$CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((-$CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] - $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                 $CellContext`x[], 
                 $CellContext`y[]] + 
               Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + 
                Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + ((
                  Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + 
               Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])}, {((Rational[-1, 4] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])/(-$CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((4 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] + (4 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] + 2 Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + (4 $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]] - (4 $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                 $CellContext`x[], 
                 $CellContext`y[]] - 4 Derivative[1, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + 2 Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]), ((Rational[-1, 4] $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]])/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ((4 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] + (4 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] + 2 Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + (4 $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]] - (4 $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                 $CellContext`x[], 
                 $CellContext`y[]] - 4 Derivative[1, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + 2 Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])}}, {{0, 0}, {0, 
            0}}}}, {-$CellContext`gen, -$CellContext`gen, -$CellContext`gen, \
$CellContext`gen}, 0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]], 
      1]}, {
     Short[
      HoldForm[
       xAct`xTensor`RiemannDown[
        xAct`xCoba`CCovD[$CellContext`PDgen, 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{{(Rational[-1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((((-2) $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}, {(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}, {{(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}, {(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}}}, {$CellContext`gen, \
-$CellContext`gen, -$CellContext`gen}, 0], 
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{
              $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]]}, {
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 
            0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]]]], 1], 
     Short[
      HoldForm[
       RawBoxes[
        xAct`xCoba`Private`CTensorBoxes[
         xAct`xCoba`CTensor[{{{{0, 0}, {0, 
            0}}, {{0, (Rational[1, 4]/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ($CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 (((-2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] ((2 Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - (4 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 
                   Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) - (4 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^3) (Derivative[0, 1][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]]) - (2 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])))}, {(Rational[1, 4]/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (((-2) 
                   Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                 Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]]) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                  Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                  Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), 0}}}, {{{
             0, (Rational[1, 4]/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (((-2) 
                   Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                 Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]]) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                  Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                  Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])))}, {(
               Rational[1, 4]/($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])) ($CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 (((-2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] ((2 Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - (4 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 
                   Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) - (4 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^3) (Derivative[0, 1][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]]) - (2 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] - (2 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) - $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))), 0}}, {{0, 0}, {0, 
            0}}}}, {-$CellContext`gen, -$CellContext`gen, -$CellContext`gen, \
-$CellContext`gen}, 0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]], 
      1]}, {
     Short[
      HoldForm[
       xAct`xTensor`RiemannDown[
        xAct`xCoba`CCovD[$CellContext`PDgen, 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{{((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}, {((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}, {{(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}, {((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}}, {$CellContext`gen, \
-$CellContext`gen, -$CellContext`gen}, 0], 
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{
              $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]]}, {
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 
            0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]]]], 1], 
     Short[
      HoldForm[
       RawBoxes[
        xAct`xCoba`Private`CTensorBoxes[
         xAct`xCoba`CTensor[{{{{0, 0}, {0, 0}}, {{0, (-$CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]] - $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                $CellContext`x[], 
                $CellContext`y[]] + 
              Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                 $CellContext`x[], 
                 $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + 
               Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + $CellContext`T212[
                 $CellContext`x[], 
                 $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + ((Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]] + $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                 $CellContext`x[], 
                 $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                $CellContext`x[], 
                $CellContext`y[]] + 
              Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]}, {$CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[0, 1][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]] + $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                 $CellContext`x[], 
                 $CellContext`y[]] + 
              Rational[1, 2] Derivative[0, 2][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] + ((Rational[1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - $CellContext`T112[
                $CellContext`x[], 
                $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]]) + ((Rational[1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) + (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + 
                Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) - $CellContext`T212[
                $CellContext`x[], 
                $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]) + ((Rational[-1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                $CellContext`x[], 
                $CellContext`y[]] - $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                $CellContext`x[], 
                $CellContext`y[]] - Derivative[1, 1][$CellContext`gxy][
               $CellContext`x[], 
               $CellContext`y[]] + 
              Rational[1, 2] Derivative[2, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]], 0}}}, {{{0, $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[0, 1][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]] + $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                 $CellContext`x[], 
                 $CellContext`y[]] + 
              Rational[1, 2] Derivative[0, 2][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] + ((Rational[1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - $CellContext`T112[
                $CellContext`x[], 
                $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]]) + ((Rational[1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) + (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + 
                Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) - $CellContext`T212[
                $CellContext`x[], 
                $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]) + ((Rational[-1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                $CellContext`x[], 
                $CellContext`y[]] - $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                $CellContext`x[], 
                $CellContext`y[]] - Derivative[1, 1][$CellContext`gxy][
               $CellContext`x[], 
               $CellContext`y[]] + 
              Rational[1, 2] Derivative[2, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]]}, {(-$CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]] - $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                $CellContext`x[], 
                $CellContext`y[]] + 
              Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                 $CellContext`x[], 
                 $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                 $CellContext`x[], 
                 $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + 
               Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + $CellContext`T212[
                 $CellContext`x[], 
                 $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + ((Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]] + $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                 $CellContext`x[], 
                 $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                $CellContext`x[], 
                $CellContext`y[]] + 
              Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                 $CellContext`x[], 
                 $CellContext`y[]], 0}}, {{0, 0}, {0, 
            0}}}}, {-$CellContext`gen, -$CellContext`gen, -$CellContext`gen, \
-$CellContext`gen}, 0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]], 
      1]}, {
     Short[
      HoldForm[
       xAct`xTensor`TensorDerivative[
        RawBoxes[
         xAct`xCoba`Private`CTensorBoxes[
          xAct`xCoba`CTensor[{{
             $CellContext`gxx[
              $CellContext`x[], 
              $CellContext`y[]], 
             $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]]}, {
             $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]], 
             $CellContext`gyy[
              $CellContext`x[], 
              $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 0],
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], 
        xAct`xCoba`CCovD[$CellContext`PDgen, 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{{(Rational[-1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])}, {(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((-2) 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}, {{(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-1) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])}, {(2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-1) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]), (Rational[-1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((-2) 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}}, {$CellContext`gen, \
-$CellContext`gen, -$CellContext`gen}, 0], 
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{
              $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]]}, {
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 
            0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]]]], 1], 
     Short[
      HoldForm[xAct`xTensor`Zero], 1]}, {
     Short[
      HoldForm[
       xAct`xTensor`TensorDerivative[
        RawBoxes[
         xAct`xCoba`Private`CTensorBoxes[
          xAct`xCoba`CTensor[{{
             $CellContext`gxx[
              $CellContext`x[], 
              $CellContext`y[]], 
             $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]]}, {
             $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[]], 
             $CellContext`gyy[
              $CellContext`x[], 
              $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 0],
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], 
        xAct`xCoba`CCovD[$CellContext`PDgen, 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{{(Rational[-1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((((-2) $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}, {(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}, {{(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}, {(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}}}, {$CellContext`gen, \
-$CellContext`gen, -$CellContext`gen}, 0], 
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{
              $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]]}, {
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 
            0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]]]], 1], 
     Short[
      HoldForm[xAct`xTensor`Zero], 1]}, {
     Short[
      HoldForm[
       xAct`xTensor`TensorDerivative[
        RawBoxes[
         xAct`xCoba`Private`CTensorBoxes[
          xAct`xCoba`CTensor[{{{{0, 0}, {0, 0}}, {{0, (-$CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] - $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                 $CellContext`x[], 
                 $CellContext`y[]] + 
               Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + 
                Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + ((
                  Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + 
               Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]}, {$CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] + 
               Rational[1, 2] Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + ((Rational[1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - $CellContext`T112[
                 $CellContext`x[], 
                 $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + ((Rational[1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - $CellContext`T212[
                 $CellContext`x[], 
                 $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + ((
                  Rational[-1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]] - $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                 $CellContext`x[], 
                 $CellContext`y[]] - Derivative[1, 1][$CellContext`gxy][
                $CellContext`x[], 
                $CellContext`y[]] + 
               Rational[1, 2] Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]], 0}}}, {{{0, $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] + 
               Rational[1, 2] Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + ((Rational[1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - $CellContext`T112[
                 $CellContext`x[], 
                 $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + ((Rational[1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - $CellContext`T212[
                 $CellContext`x[], 
                 $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + ((
                  Rational[-1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                 $CellContext`x[], 
                 $CellContext`y[]] - $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                 $CellContext`x[], 
                 $CellContext`y[]] - Derivative[1, 1][$CellContext`gxy][
                $CellContext`x[], 
                $CellContext`y[]] + 
               Rational[1, 2] Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]}, {(-$CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] - $CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                 $CellContext`x[], 
                 $CellContext`y[]] + 
               Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + 
                Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + ((
                  Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                 $CellContext`x[], 
                 $CellContext`y[]] + 
               Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]], 0}}, {{0, 0}, {0, 
             0}}}}, {-$CellContext`gen, -$CellContext`gen, -$CellContext`gen, \
-$CellContext`gen}, 0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], 
        xAct`xCoba`CCovD[$CellContext`PDgen, 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{{((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}, {((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}, {{(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}, {((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), ((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}}, {$CellContext`gen, \
-$CellContext`gen, -$CellContext`gen}, 0], 
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{
              $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]]}, {
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 
            0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]]]], 1], 
     Short[
      HoldForm[
       RawBoxes[
        xAct`xCoba`Private`CTensorBoxes[
         xAct`xCoba`CTensor[{{{{{0, 0}, {0, 0}}, {{0, 0}, {0, 
            0}}}, {{{0, 0}, {
              Rational[1, 4] (((-4) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - (4 
                 Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[1, 0][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) Derivative[1, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) Derivative[1, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] - (4 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 1][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] - (4 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 1][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - 2 Derivative[1, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[2, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[2, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) (Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[2, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) - ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) - $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[2, 0][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[2, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + 4 Derivative[2, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[3, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]), 
               Rational[1, 4] (((-4) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] - (4 Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - (4 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 2][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] - (4 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 2][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - 2 Derivative[0, 3][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 2][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 
                  Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + (4 Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) (((4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) - ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[0, 2][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + 4 Derivative[1, 2][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] - ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                Derivative[2, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])}}, {{
              Rational[1, 4] ((4 Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 
                  Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] - (
                 4 (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] - (4 Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - (
                 4 (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[1, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) Derivative[1, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) Derivative[1, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + 2 Derivative[1, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[2, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[2, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[2, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (4 $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]]) (-Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                   Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                   Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]]) (-Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[2, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] - (4 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[2, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - 4 Derivative[2, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + 2 Derivative[3, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]), 
               Rational[1, 4] ((4 Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 2][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 2][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + 2 Derivative[0, 3][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 2][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (4 
                 Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]]) (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) - (4 
                 Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]]) (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) - (4 
                 Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] - (4 Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                  Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (4 $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]]) (-Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]]) (-Derivative[0, 2][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 1][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] - (4 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 1][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - 4 Derivative[1, 2][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                   Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                   Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + 
                2 Derivative[2, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])}, {0, 0}}}}, {{{{0, 0}, {
              Rational[1, 4] ((4 Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 
                  Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] - (
                 4 (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] - (4 Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - (
                 4 (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[1, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) Derivative[1, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) Derivative[1, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + 2 Derivative[1, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[2, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[2, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[2, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (4 $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]]) (-Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                   Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                   Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]]) (-Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[2, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] - (4 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[2, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - 4 Derivative[2, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + 2 Derivative[3, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]), 
               Rational[1, 4] ((4 Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 2][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 2][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + 2 Derivative[0, 3][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 2][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (4 
                 Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]]) (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) - (4 
                 Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]]) (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) - (4 
                 Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] - (4 Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                  Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (4 $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]]) (-Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]]) (-Derivative[0, 2][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 1][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] - (4 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 1][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - 4 Derivative[1, 2][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                   Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                   Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + 
                2 Derivative[2, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])}}, {{
              Rational[1, 4] (((-4) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - (4 
                 Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[1, 0][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) Derivative[1, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) Derivative[1, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] - (4 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 1][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] - (4 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 1][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - 2 Derivative[1, 2][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[2, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[2, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) (Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[2, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[2, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) - ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) - $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[2, 0][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[2, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + 4 Derivative[2, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[3, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]), 
               Rational[1, 4] (((-4) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] - (4 Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - (4 $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 2][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] - (4 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[0, 2][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - 2 Derivative[0, 3][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 2][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 
                  Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) + ((($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-2) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + (4 Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) (((4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]) - ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) (-Derivative[0, 2][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) - Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + 4 Derivative[1, 2][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] - ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - ((8/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] - $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                    Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                Derivative[2, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])}, {0, 0}}}, {{{0, 0}, {0, 0}}, {{0, 0}, {
            0, 0}}}}}, {-$CellContext`gen, -$CellContext`gen, \
-$CellContext`gen, -$CellContext`gen, -$CellContext`gen}, 0], 
         xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]], 1]}, {
     Short[
      HoldForm[
       xAct`xCoba`ToCTensorOneSlot[
        RawBoxes[
         xAct`xCoba`Private`CTensorBoxes[
          xAct`xCoba`CTensor[{{{{0, 0}, {0, 0}}, {{($CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] + 
                Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                   Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + 
                Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]), ($CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]]/(-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] + 
                Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                   Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + 
                Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])}, {((Rational[1, 4] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])/(-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) ((4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + 2 Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] - (4 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - 4 Derivative[1, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + 2 Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]), ((Rational[1, 4] $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]])/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + 2 Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] - (4 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - 4 Derivative[1, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + 2 Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])}}}, {{{((-$CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]])/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] + 
                Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                   Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + 
                Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]), ((-$CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]])/(-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] + 
                Rational[-1, 2] Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((Rational[-1, 4]/($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) - (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + 
                 Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] (-Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + ((
                   Rational[1, 2]/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + 
                Rational[-1, 2] Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])}, {((Rational[-1, 4] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])/(-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])) ((4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + 2 Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] - (4 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - 4 Derivative[1, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + 2 Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]), ((Rational[-1, 4] $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]])/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((4 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]] + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] + 2 Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (4 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])^(-1) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + (4 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + ((4/((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]))) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + 
                  Rational[-1, 2] Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (4 $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]] - (4 $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                  $CellContext`x[], 
                  $CellContext`y[]] - 4 Derivative[1, 1][$CellContext`gxy][
                  $CellContext`x[], 
                  $CellContext`y[]] + 2 Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])}}, {{0, 0}, {0, 
             0}}}}, {-$CellContext`gen, -$CellContext`gen, -$CellContext`gen, \
$CellContext`gen}, 0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], {3, 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{$CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]]/(-$CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 + $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]]), $CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]]/($CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]]^2 - $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]] $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]])}, {$CellContext`gxy[
                $CellContext`x[], 
                $CellContext`y[]]/($CellContext`gxy[
                 $CellContext`x[], 
                 $CellContext`y[]]^2 - $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]] $CellContext`gyy[
                $CellContext`x[], 
                $CellContext`y[]]), $CellContext`gxx[
                $CellContext`x[], 
                $CellContext`y[]]/(-$CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]]^2 + $CellContext`gxx[
                 $CellContext`x[], 
                 $CellContext`y[]] $CellContext`gyy[
                 $CellContext`x[], 
                 $CellContext`y[]])}}, {$CellContext`gen, $CellContext`gen}, 
            0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]}]], 1], 
     Short[
      HoldForm[
       RawBoxes[
        xAct`xCoba`Private`CTensorBoxes[
         xAct`xCoba`CTensor[{{{{0, 0}, {0, 
            0}}, {{0, (Rational[1, 4] ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (((-2) 
                   Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                 Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]]) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                  Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                  Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])))}, {(Rational[-1, 4] ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (((-2) 
                   Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                 Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]]) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                  Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                  Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), 0}}}, {{{
             0, (Rational[-1, 4] ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (((-2) 
                   Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                 Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]]) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                  Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                  Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])))}, {(
               Rational[1, 4] ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-2)) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + $CellContext`gxy[
                  $CellContext`x[], 
                  $CellContext`y[]] (((-2) 
                   Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (4 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])) + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) ((-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 1][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]]))) + (4 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^3) (
                 Derivative[0, 1][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - Derivative[1, 0][$CellContext`T112][
                  $CellContext`x[], 
                  $CellContext`y[]]) + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) ((2 $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`T112][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 2][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`T212][
                   $CellContext`x[], 
                   $CellContext`y[]] - 2 Derivative[1, 1][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]) + $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 
                  Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[1, 0][$CellContext`gxy][
                   $CellContext`x[], 
                   $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]^2 + (4 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) Derivative[1, 0][$CellContext`T212][
                    $CellContext`x[], 
                    $CellContext`y[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) (Derivative[0, 2][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] + $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] (2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) - 2 
                  Derivative[1, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[2, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))), 0}}, {{0, 0}, {0, 
            0}}}}, {-$CellContext`gen, -$CellContext`gen, $CellContext`gen, \
$CellContext`gen}, 0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]], 
      1]}, {
     Short[
      HoldForm[
       xAct`xTensor`Torsion[
        xAct`xCoba`CCovD[$CellContext`PDgen, 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{{(Rational[-1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])}, {(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((-$CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((-2) 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}, {{(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-1) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]])}, {(2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2 - (2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]])^(-1) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[1, 0][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]]), (Rational[-1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((-2) 
                   Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}}, {$CellContext`gen, \
-$CellContext`gen, -$CellContext`gen}, 0], 
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{
              $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]]}, {
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 
            0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]]]], 1], 
     Short[
      HoldForm[xAct`xTensor`Zero], 1]}, {
     Short[
      HoldForm[
       xAct`xTensor`Torsion[
        xAct`xCoba`CCovD[$CellContext`PDgen, 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{{(Rational[-1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T212[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((((-2) $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}, {(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gyy[
                   $CellContext`x[], 
                   $CellContext`y[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[]]))}}, {{(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[1, 0][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]] - 2 Derivative[1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + Derivative[0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[]]) + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}, {(Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] Derivative[0, 1][$CellContext`gxx][
                   $CellContext`x[], 
                   $CellContext`y[]] + $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]])), (Rational[1, 2]/($CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]]^2 - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] $CellContext`gyy[
                  $CellContext`x[], 
                  $CellContext`y[]])) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[]]^2) $CellContext`T112[
                   $CellContext`x[], 
                   $CellContext`y[]] - $CellContext`gxx[
                  $CellContext`x[], 
                  $CellContext`y[]] Derivative[0, 1][$CellContext`gyy][
                  $CellContext`x[], 
                  $CellContext`y[]] + $CellContext`gxy[
                   $CellContext`x[], 
                   $CellContext`y[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[]] + 2 Derivative[0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[]] - Derivative[1, 0][$CellContext`gyy][
                   $CellContext`x[], 
                   $CellContext`y[]]))}}}, {$CellContext`gen, \
-$CellContext`gen, -$CellContext`gen}, 0], 
           xAct`xCoba`Private`$xCobaCacheLargeComponentSize]], 
         RawBoxes[
          xAct`xCoba`Private`CTensorBoxes[
           xAct`xCoba`CTensor[{{
              $CellContext`gxx[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]]}, {
              $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[]], 
              $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[]]}}, {-$CellContext`gen, -$CellContext`gen}, 
            0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]]]], 1], 
     Short[
      HoldForm[
       RawBoxes[
        xAct`xCoba`Private`CTensorBoxes[
         xAct`xCoba`CTensor[{{{0, 
             $CellContext`T112[
              $CellContext`x[], 
              $CellContext`y[]]}, {-$CellContext`T112[
               $CellContext`x[], 
               $CellContext`y[]], 0}}, {{0, 
             $CellContext`T212[
              $CellContext`x[], 
              $CellContext`y[]]}, {-$CellContext`T212[
               $CellContext`x[], 
               $CellContext`y[]], 
             0}}}, {$CellContext`gen, -$CellContext`gen, -$CellContext`gen}, 
          0], xAct`xCoba`Private`$xCobaCacheLargeComponentSize]]], 1]}}, 
   TableHeadings -> {Automatic, None}]]], "Output",
 CellChangeTimes->{3.728676520782053*^9, 3.7287536005288486`*^9, 
  3.72884372805814*^9, 3.73095631488002*^9, 3.731119342738673*^9, 
  3.733368387841898*^9},
 CellLabel->"Out[33]//TableForm=",
 CellID->360120287,ExpressionUUID->"46294718-78e9-4133-ad74-4e16b13f6123"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Clean up:", "MathCaption",
 CellID->1502852846,ExpressionUUID->"c1681903-5d69-4f8d-a1c7-a8630340bc0e"],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{
  "metric", ",", "torsion", ",", "cd", ",", "christoffel", ",", "riemann", 
   ",", "riemanndown", ",", "lc", ",", "chris"}], "]"}]], "Input",
 CellChangeTimes->{{3.731119391858962*^9, 3.731119396281283*^9}},
 CellLabel->"In[34]:=",
 CellID->574457121,ExpressionUUID->"05046742-69c9-484e-af67-1af2d0bbb329"],

Cell[BoxData[
 RowBox[{
  RowBox[{"UndefScalarFunction", "[", "$ScalarFunctions", "]"}], 
  ";"}]], "Input",
 CellLabel->"In[35]:=",
 CellID->1961839382,ExpressionUUID->"c46076c3-27d1-45c0-8779-fa7cc22b7cde"],

Cell[BoxData[
 RowBox[{"UndefChart", "[", "gen", "]"}]], "Input",
 CellLabel->"In[36]:=",
 CellID->170577236,ExpressionUUID->"532b5357-dd19-4dd4-b933-b8c4a247e19f"],

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M", "]"}]], "Input",
 CellLabel->"In[37]:=",
 CellID->936707,ExpressionUUID->"7f9dc083-5f0f-41b0-934e-aa5772caa148"],

Cell[BoxData[
 RowBox[{"ClearxCobaCache", "[", "All", "]"}]], "Input",
 CellLabel->"In[38]:=",
 CellID->710849020,ExpressionUUID->"2f848caf-808f-49a0-86b6-ed3ad3d79aba"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.5.3. General 3d metric connection with torsion", "Subsubsection",
 CellChangeTimes->{{3.728676534413612*^9, 3.728676534622197*^9}, 
   3.7908767503278313`*^9},
 CellID->787950879,ExpressionUUID->"92193d4e-b29b-4099-b849-346405bd9932"],

Cell[CellGroupData[{

Cell["Now we repeat the same computation in the 3d case.", "MathCaption",
 CellID->1706255951,ExpressionUUID->"51b3c354-a94f-4a48-8133-728a127f4f06"],

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<xAct`xCoba`\>\"", "]"}]], "Input",
 CellLabel->"In[39]:=",
 CellID->1694630066,ExpressionUUID->"8475a7c5-bd1d-44b5-a6b3-e773281426dd"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"$DefInfoQ", "=", 
   RowBox[{"$UndefInfoQ", "=", "False"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"$CVVerbose", "=", "False"}], ";"}]}], "Input",
 CellChangeTimes->{{3.73111945645512*^9, 3.731119461847053*^9}},
 CellLabel->"In[40]:=",
 CellID->1719392748,ExpressionUUID->"015571dd-f029-4c5e-9d91-3090259789f7"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Define a 3d manifold:", "MathCaption",
 CellID->1293961604,ExpressionUUID->"c5911706-7556-475a-b85b-368a0fba0052"],

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M", ",", "3", ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f"}], "}"}]}], 
  "]"}]], "Input",
 CellLabel->"In[42]:=",
 CellID->45785259,ExpressionUUID->"ba70e460-4966-464c-b20d-691ff9a5a949"],

Cell[BoxData[
 RowBox[{"DefChart", "[", 
  RowBox[{"gen", ",", "M", ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", "2", ",", "3"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"x", "[", "]"}], ",", 
     RowBox[{"y", "[", "]"}], ",", 
     RowBox[{"z", "[", "]"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[43]:=",
 CellID->713702387,ExpressionUUID->"2babde1f-f6ca-4276-9929-82483e3928da"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Let us work with the general 3d case:", "MathCaption",
 CellID->1499849814,ExpressionUUID->"1d80b263-c725-4146-aec2-c94bb0be9ab3"],

Cell[BoxData[
 RowBox[{
  RowBox[{"DefScalarFunction", "/@", 
   RowBox[{"{", 
    RowBox[{
    "gxx", ",", "gxy", ",", "gxz", ",", "gyy", ",", "gyz", ",", "gzz"}], 
    "}"}]}], ";"}]], "Input",
 CellLabel->"In[44]:=",
 CellID->812775412,ExpressionUUID->"6b11e660-2802-4e79-8494-abc120747971"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{"metric", "=", 
    RowBox[{"CTensor", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{
          RowBox[{"gxx", "[", 
           RowBox[{
            RowBox[{"x", "[", "]"}], ",", 
            RowBox[{"y", "[", "]"}], ",", 
            RowBox[{"z", "[", "]"}]}], "]"}], ",", 
          RowBox[{"gxy", "[", 
           RowBox[{
            RowBox[{"x", "[", "]"}], ",", 
            RowBox[{"y", "[", "]"}], ",", 
            RowBox[{"z", "[", "]"}]}], "]"}], ",", 
          RowBox[{"gxz", "[", 
           RowBox[{
            RowBox[{"x", "[", "]"}], ",", 
            RowBox[{"y", "[", "]"}], ",", 
            RowBox[{"z", "[", "]"}]}], "]"}]}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{"gxy", "[", 
           RowBox[{
            RowBox[{"x", "[", "]"}], ",", 
            RowBox[{"y", "[", "]"}], ",", 
            RowBox[{"z", "[", "]"}]}], "]"}], ",", 
          RowBox[{"gyy", "[", 
           RowBox[{
            RowBox[{"x", "[", "]"}], ",", 
            RowBox[{"y", "[", "]"}], ",", 
            RowBox[{"z", "[", "]"}]}], "]"}], ",", 
          RowBox[{"gyz", "[", 
           RowBox[{
            RowBox[{"x", "[", "]"}], ",", 
            RowBox[{"y", "[", "]"}], ",", 
            RowBox[{"z", "[", "]"}]}], "]"}]}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{"gxz", "[", 
           RowBox[{
            RowBox[{"x", "[", "]"}], ",", 
            RowBox[{"y", "[", "]"}], ",", 
            RowBox[{"z", "[", "]"}]}], "]"}], ",", 
          RowBox[{"gyz", "[", 
           RowBox[{
            RowBox[{"x", "[", "]"}], ",", 
            RowBox[{"y", "[", "]"}], ",", 
            RowBox[{"z", "[", "]"}]}], "]"}], ",", 
          RowBox[{"gzz", "[", 
           RowBox[{
            RowBox[{"x", "[", "]"}], ",", 
            RowBox[{"y", "[", "]"}], ",", 
            RowBox[{"z", "[", "]"}]}], "]"}]}], "}"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "gen"}], ",", 
        RowBox[{"-", "gen"}]}], "}"}]}], "]"}]}], ")"}], "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}]}], "]"}]], "Input",
 CellLabel->"In[45]:=",
 CellID->172870917,ExpressionUUID->"c87c8daf-a6dc-4c38-a848-9840f7f3e1e8"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{
             InterpretationBox[
              StyleBox["gxx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{
                InterpretationBox[
                 StyleBox[
                  RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                  False], $CellContext`gxx, Editable -> False], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`x[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`y[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`z[], Editable -> False]}], "]"}]}]],
           Annotation[#, {1, 1} -> $CellContext`gxx[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             InterpretationBox[
              StyleBox["gxy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{
                InterpretationBox[
                 StyleBox[
                  RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                  False], $CellContext`gxy, Editable -> False], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`x[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`y[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`z[], Editable -> False]}], "]"}]}]],
           Annotation[#, {1, 2} -> $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             InterpretationBox[
              StyleBox["gxz",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxz,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", 
              RowBox[{
                InterpretationBox[
                 StyleBox[
                  RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                  False], $CellContext`gxz, Editable -> False], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`x[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`y[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`z[], Editable -> False]}], "]"}]}]],
           Annotation[#, {1, 3} -> $CellContext`gxz[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             InterpretationBox[
              StyleBox["gxy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{
                InterpretationBox[
                 StyleBox[
                  RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                  False], $CellContext`gxy, Editable -> False], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`x[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`y[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`z[], Editable -> False]}], "]"}]}]],
           Annotation[#, {2, 1} -> $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             InterpretationBox[
              StyleBox["gyy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gyy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{
                InterpretationBox[
                 StyleBox[
                  RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                  False], $CellContext`gyy, Editable -> False], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`x[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`y[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`z[], Editable -> False]}], "]"}]}]],
           Annotation[#, {2, 2} -> $CellContext`gyy[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             InterpretationBox[
              StyleBox["gyz",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gyz,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", 
              RowBox[{
                InterpretationBox[
                 StyleBox[
                  RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                  False], $CellContext`gyz, Editable -> False], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`x[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`y[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`z[], Editable -> False]}], "]"}]}]],
           Annotation[#, {2, 3} -> $CellContext`gyz[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             InterpretationBox[
              StyleBox["gxz",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gxz,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{
                InterpretationBox[
                 StyleBox[
                  RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                  False], $CellContext`gxz, Editable -> False], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`x[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`y[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`z[], Editable -> False]}], "]"}]}]],
           Annotation[#, {3, 1} -> $CellContext`gxz[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             InterpretationBox[
              StyleBox["gyz",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gyz,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{
                InterpretationBox[
                 StyleBox[
                  RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                  False], $CellContext`gyz, Editable -> False], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`x[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`y[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`z[], Editable -> False]}], "]"}]}]],
           Annotation[#, {3, 2} -> $CellContext`gyz[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             InterpretationBox[
              StyleBox["gzz",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`gzz,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["y",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`y[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", 
              RowBox[{
                InterpretationBox[
                 StyleBox[
                  RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                  False], $CellContext`gzz, Editable -> False], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`x[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`y[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`z[], Editable -> False]}], "]"}]}]],
           Annotation[#, {3, 3} -> $CellContext`gzz[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]], "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 4 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 4 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " "},
         {"a", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{
     $CellContext`gxx[
      $CellContext`x[], 
      $CellContext`y[], 
      $CellContext`z[]], 
     $CellContext`gxy[
      $CellContext`x[], 
      $CellContext`y[], 
      $CellContext`z[]], 
     $CellContext`gxz[
      $CellContext`x[], 
      $CellContext`y[], 
      $CellContext`z[]]}, {
     $CellContext`gxy[
      $CellContext`x[], 
      $CellContext`y[], 
      $CellContext`z[]], 
     $CellContext`gyy[
      $CellContext`x[], 
      $CellContext`y[], 
      $CellContext`z[]], 
     $CellContext`gyz[
      $CellContext`x[], 
      $CellContext`y[], 
      $CellContext`z[]]}, {
     $CellContext`gxz[
      $CellContext`x[], 
      $CellContext`y[], 
      $CellContext`z[]], 
     $CellContext`gyz[
      $CellContext`x[], 
      $CellContext`y[], 
      $CellContext`z[]], 
     $CellContext`gzz[
      $CellContext`x[], 
      $CellContext`y[], 
      $CellContext`z[]]}}, {-$CellContext`gen, -$CellContext`gen}, 
   0][-$CellContext`a, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.728676570002798*^9, 3.728753604035306*^9, 
  3.728843743546665*^9, 3.7309563179261837`*^9, 3.7311194665192223`*^9, 
  3.731119613145679*^9, 3.7311199296738358`*^9, 3.733368414793096*^9},
 CellLabel->"Out[45]=",
 CellID->1047119116,ExpressionUUID->"cb0618b0-7a3e-4130-9618-bbcd992801ee"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SymmetryGroupOfTensor", "[", "metric", "]"}]], "Input",
 CellLabel->"In[46]:=",
 CellID->121248807,ExpressionUUID->"9302139c-3731-42c2-a6a2-a1b5cc2bdff8"],

Cell[BoxData[
 RowBox[{"StrongGenSet", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "2"}], "}"}], ",", 
   RowBox[{"GenSet", "[", 
    InterpretationBox[
     StyleBox[
      RowBox[{"(", "1", ",", "2", ")"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xPerm`Cycles[{1, 2}],
     Editable->False], "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728676570198267*^9, 3.728753604163444*^9, 
  3.728843743772505*^9, 3.730956318076869*^9, 3.7311194679152412`*^9, 
  3.7311196142927217`*^9, 3.731119930541769*^9, 3.7333684149170923`*^9},
 CellLabel->"Out[46]=",
 CellID->302495447,ExpressionUUID->"204a8dfb-0c09-4998-9e03-3616d949860b"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"DefScalarFunction", "/@", 
   RowBox[{"{", 
    RowBox[{
    "T112", ",", "T113", ",", "T123", ",", "T212", ",", "T213", ",", "T223", 
     ",", "T312", ",", "T313", ",", "T323"}], "}"}]}], ";"}]], "Input",
 CellLabel->"In[47]:=",
 CellID->1906487990,ExpressionUUID->"850afa34-94cc-4755-8bae-3b6edd54ff33"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{"torsion", "=", 
    RowBox[{"CTensor", "[", 
     RowBox[{
      RowBox[{"{", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"{", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{"0", ",", 
            RowBox[{"T112", "[", 
             RowBox[{
              RowBox[{"x", "[", "]"}], ",", 
              RowBox[{"y", "[", "]"}], ",", 
              RowBox[{"z", "[", "]"}]}], "]"}], ",", 
            RowBox[{"T113", "[", 
             RowBox[{
              RowBox[{"x", "[", "]"}], ",", 
              RowBox[{"y", "[", "]"}], ",", 
              RowBox[{"z", "[", "]"}]}], "]"}]}], "}"}], ",", 
          "\[IndentingNewLine]", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", 
             RowBox[{"T112", "[", 
              RowBox[{
               RowBox[{"x", "[", "]"}], ",", 
               RowBox[{"y", "[", "]"}], ",", 
               RowBox[{"z", "[", "]"}]}], "]"}]}], ",", "0", ",", 
            RowBox[{"T123", "[", 
             RowBox[{
              RowBox[{"x", "[", "]"}], ",", 
              RowBox[{"y", "[", "]"}], ",", 
              RowBox[{"z", "[", "]"}]}], "]"}]}], "}"}], ",", 
          "\[IndentingNewLine]", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", 
             RowBox[{"T113", "[", 
              RowBox[{
               RowBox[{"x", "[", "]"}], ",", 
               RowBox[{"y", "[", "]"}], ",", 
               RowBox[{"z", "[", "]"}]}], "]"}]}], ",", 
            RowBox[{"-", 
             RowBox[{"T123", "[", 
              RowBox[{
               RowBox[{"x", "[", "]"}], ",", 
               RowBox[{"y", "[", "]"}], ",", 
               RowBox[{"z", "[", "]"}]}], "]"}]}], ",", "0"}], "}"}]}], 
         "\[IndentingNewLine]", "}"}], ",", "\[IndentingNewLine]", 
        RowBox[{"{", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{"0", ",", 
            RowBox[{"T212", "[", 
             RowBox[{
              RowBox[{"x", "[", "]"}], ",", 
              RowBox[{"y", "[", "]"}], ",", 
              RowBox[{"z", "[", "]"}]}], "]"}], ",", 
            RowBox[{"T213", "[", 
             RowBox[{
              RowBox[{"x", "[", "]"}], ",", 
              RowBox[{"y", "[", "]"}], ",", 
              RowBox[{"z", "[", "]"}]}], "]"}]}], "}"}], ",", 
          "\[IndentingNewLine]", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", 
             RowBox[{"T212", "[", 
              RowBox[{
               RowBox[{"x", "[", "]"}], ",", 
               RowBox[{"y", "[", "]"}], ",", 
               RowBox[{"z", "[", "]"}]}], "]"}]}], ",", "0", ",", 
            RowBox[{"T223", "[", 
             RowBox[{
              RowBox[{"x", "[", "]"}], ",", 
              RowBox[{"y", "[", "]"}], ",", 
              RowBox[{"z", "[", "]"}]}], "]"}]}], "}"}], ",", 
          "\[IndentingNewLine]", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", 
             RowBox[{"T213", "[", 
              RowBox[{
               RowBox[{"x", "[", "]"}], ",", 
               RowBox[{"y", "[", "]"}], ",", 
               RowBox[{"z", "[", "]"}]}], "]"}]}], ",", 
            RowBox[{"-", 
             RowBox[{"T223", "[", 
              RowBox[{
               RowBox[{"x", "[", "]"}], ",", 
               RowBox[{"y", "[", "]"}], ",", 
               RowBox[{"z", "[", "]"}]}], "]"}]}], ",", "0"}], "}"}]}], 
         "\[IndentingNewLine]", "}"}], ",", "\[IndentingNewLine]", 
        RowBox[{"{", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{"0", ",", 
            RowBox[{"T312", "[", 
             RowBox[{
              RowBox[{"x", "[", "]"}], ",", 
              RowBox[{"y", "[", "]"}], ",", 
              RowBox[{"z", "[", "]"}]}], "]"}], ",", 
            RowBox[{"T313", "[", 
             RowBox[{
              RowBox[{"x", "[", "]"}], ",", 
              RowBox[{"y", "[", "]"}], ",", 
              RowBox[{"z", "[", "]"}]}], "]"}]}], "}"}], ",", 
          "\[IndentingNewLine]", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", 
             RowBox[{"T312", "[", 
              RowBox[{
               RowBox[{"x", "[", "]"}], ",", 
               RowBox[{"y", "[", "]"}], ",", 
               RowBox[{"z", "[", "]"}]}], "]"}]}], ",", "0", ",", 
            RowBox[{"T323", "[", 
             RowBox[{
              RowBox[{"x", "[", "]"}], ",", 
              RowBox[{"y", "[", "]"}], ",", 
              RowBox[{"z", "[", "]"}]}], "]"}]}], "}"}], ",", 
          "\[IndentingNewLine]", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", 
             RowBox[{"T313", "[", 
              RowBox[{
               RowBox[{"x", "[", "]"}], ",", 
               RowBox[{"y", "[", "]"}], ",", 
               RowBox[{"z", "[", "]"}]}], "]"}]}], ",", 
            RowBox[{"-", 
             RowBox[{"T323", "[", 
              RowBox[{
               RowBox[{"x", "[", "]"}], ",", 
               RowBox[{"y", "[", "]"}], ",", 
               RowBox[{"z", "[", "]"}]}], "]"}]}], ",", "0"}], "}"}]}], 
         "\[IndentingNewLine]", "}"}]}], "\[IndentingNewLine]", "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"gen", ",", 
        RowBox[{"-", "gen"}], ",", 
        RowBox[{"-", "gen"}]}], "}"}]}], "]"}]}], ")"}], "[", 
  RowBox[{"a", ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}]}], "]"}]], "Input",
 CellLabel->"In[48]:=",
 CellID->1887634690,ExpressionUUID->"46190505-05e2-4a44-8a2f-6e85a89e1c5d"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["T112",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T112,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["z",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`z[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 1, 2} -> $CellContext`T112[
                 $CellContext`x[], 
                 $CellContext`y[], 
                 $CellContext`z[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["T113",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T113,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["z",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`z[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "3"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 1, 3} -> $CellContext`T113[
                 $CellContext`x[], 
                 $CellContext`y[], 
                 $CellContext`z[]], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["T112",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`T112,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`z[],
                   Editable->False]}], "]"}]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{"-", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}]],
              Annotation[#, {1, 2, 1} -> -$CellContext`T112[
                  $CellContext`x[], 
                  $CellContext`y[], 
                  $CellContext`z[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["T123",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T123,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["z",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`z[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "3"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]],
              Annotation[#, {1, 2, 3} -> $CellContext`T123[
                 $CellContext`x[], 
                 $CellContext`y[], 
                 $CellContext`z[]], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["T113",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`T113,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`z[],
                   Editable->False]}], "]"}]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{"-", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}]],
              Annotation[#, {1, 3, 1} -> -$CellContext`T113[
                  $CellContext`x[], 
                  $CellContext`y[], 
                  $CellContext`z[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["T123",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`T123,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`z[],
                   Editable->False]}], "]"}]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{"-", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}]],
              Annotation[#, {1, 3, 2} -> -$CellContext`T123[
                  $CellContext`x[], 
                  $CellContext`y[], 
                  $CellContext`z[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 3} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["T212",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T212,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["z",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`z[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 1, 2} -> $CellContext`T212[
                 $CellContext`x[], 
                 $CellContext`y[], 
                 $CellContext`z[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["T213",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T213,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["z",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`z[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "3"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 1, 3} -> $CellContext`T213[
                 $CellContext`x[], 
                 $CellContext`y[], 
                 $CellContext`z[]], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["T212",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`T212,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`z[],
                   Editable->False]}], "]"}]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{"-", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}]],
              Annotation[#, {2, 2, 1} -> -$CellContext`T212[
                  $CellContext`x[], 
                  $CellContext`y[], 
                  $CellContext`z[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["T223",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T223,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["z",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`z[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "3"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]],
              Annotation[#, {2, 2, 3} -> $CellContext`T223[
                 $CellContext`x[], 
                 $CellContext`y[], 
                 $CellContext`z[]], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["T213",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`T213,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`z[],
                   Editable->False]}], "]"}]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{"-", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}]],
              Annotation[#, {2, 3, 1} -> -$CellContext`T213[
                  $CellContext`x[], 
                  $CellContext`y[], 
                  $CellContext`z[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["T223",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`T223,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`z[],
                   Editable->False]}], "]"}]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{"-", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}]],
              Annotation[#, {2, 3, 2} -> -$CellContext`T223[
                  $CellContext`x[], 
                  $CellContext`y[], 
                  $CellContext`z[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 3} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 1} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["T312",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T312,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["z",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`z[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]],
              Annotation[#, {3, 1, 2} -> $CellContext`T312[
                 $CellContext`x[], 
                 $CellContext`y[], 
                 $CellContext`z[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["T313",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T313,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["z",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`z[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "3"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]],
              Annotation[#, {3, 1, 3} -> $CellContext`T313[
                 $CellContext`x[], 
                 $CellContext`y[], 
                 $CellContext`z[]], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["T312",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`T312,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`z[],
                   Editable->False]}], "]"}]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{"-", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}]],
              Annotation[#, {3, 2, 1} -> -$CellContext`T312[
                  $CellContext`x[], 
                  $CellContext`y[], 
                  $CellContext`z[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "2"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 2} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["T323",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`T323,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["y",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`y[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["z",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`z[],
                  Editable->False]}], "]"}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "3"}], "}"}], "\[Rule]", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]],
              Annotation[#, {3, 2, 3} -> $CellContext`T323[
                 $CellContext`x[], 
                 $CellContext`y[], 
                 $CellContext`z[]], "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["T313",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`T313,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`z[],
                   Editable->False]}], "]"}]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "1"}], "}"}], "\[Rule]", 
                 RowBox[{"-", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}]],
              Annotation[#, {3, 3, 1} -> -$CellContext`T313[
                  $CellContext`x[], 
                  $CellContext`y[], 
                  $CellContext`z[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["T323",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`T323,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["y",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`y[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`z[],
                   Editable->False]}], "]"}]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "2"}], "}"}], "\[Rule]", 
                 RowBox[{"-", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}]],
              Annotation[#, {3, 3, 2} -> -$CellContext`T323[
                  $CellContext`x[], 
                  $CellContext`y[], 
                  $CellContext`z[]], "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "3"}], "}"}], "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 3} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 4 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 4 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a", " ", " "},
         {" ", "b", "c"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{{0, 
      $CellContext`T112[
       $CellContext`x[], 
       $CellContext`y[], 
       $CellContext`z[]], 
      $CellContext`T113[
       $CellContext`x[], 
       $CellContext`y[], 
       $CellContext`z[]]}, {-$CellContext`T112[
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]], 0, 
      $CellContext`T123[
       $CellContext`x[], 
       $CellContext`y[], 
       $CellContext`z[]]}, {-$CellContext`T113[
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]], -$CellContext`T123[
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]], 0}}, {{0, 
      $CellContext`T212[
       $CellContext`x[], 
       $CellContext`y[], 
       $CellContext`z[]], 
      $CellContext`T213[
       $CellContext`x[], 
       $CellContext`y[], 
       $CellContext`z[]]}, {-$CellContext`T212[
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]], 0, 
      $CellContext`T223[
       $CellContext`x[], 
       $CellContext`y[], 
       $CellContext`z[]]}, {-$CellContext`T213[
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]], -$CellContext`T223[
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]], 0}}, {{0, 
      $CellContext`T312[
       $CellContext`x[], 
       $CellContext`y[], 
       $CellContext`z[]], 
      $CellContext`T313[
       $CellContext`x[], 
       $CellContext`y[], 
       $CellContext`z[]]}, {-$CellContext`T312[
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]], 0, 
      $CellContext`T323[
       $CellContext`x[], 
       $CellContext`y[], 
       $CellContext`z[]]}, {-$CellContext`T313[
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]], -$CellContext`T323[
        $CellContext`x[], 
        $CellContext`y[], 
        $CellContext`z[]], 
      0}}}, {$CellContext`gen, -$CellContext`gen, -$CellContext`gen}, 
   0][$CellContext`a, -$CellContext`b, -$CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{3.7286765704827833`*^9, 3.728753604355486*^9, 
  3.7288437440887823`*^9, 3.7309563182885*^9, 3.731119469745234*^9, 
  3.731119615646492*^9, 3.7311199316876173`*^9, 3.733368415084087*^9},
 CellLabel->"Out[48]=",
 CellID->947185280,ExpressionUUID->"6b272ca2-b0cf-423f-8007-37d672e31fb1"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SymmetryGroupOfTensor", "[", "torsion", "]"}]], "Input",
 CellLabel->"In[49]:=",
 CellID->2029181533,ExpressionUUID->"c1cdb529-97b7-4052-8e74-867871cb8422"],

Cell[BoxData[
 RowBox[{"StrongGenSet", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"2", ",", "3"}], "}"}], ",", 
   RowBox[{"GenSet", "[", 
    InterpretationBox[
     StyleBox[
      RowBox[{"-", 
       RowBox[{"(", "2", ",", "3", ")"}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     -xAct`xPerm`Cycles[{2, 3}],
     Editable->False], "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728676570681984*^9, 3.728753604480196*^9, 
  3.728843744847845*^9, 3.730956318447516*^9, 3.7311194716161013`*^9, 
  3.731119617141925*^9, 3.7311199332018957`*^9, 3.733368415198093*^9},
 CellLabel->"Out[49]=",
 CellID->46725749,ExpressionUUID->"603ec58e-72fd-408b-83d8-d99ff58cfb0e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Declare that metric as the first-metric of the vbundle, and add torsion. This \
precomputes some tensors:\
\>", "MathCaption",
 CellChangeTimes->{{3.731119936926437*^9, 3.731119951462109*^9}, {
  3.731120018594201*^9, 3.731120025515539*^9}},
 CellID->743660032,ExpressionUUID->"bc101896-db84-4c88-b09f-1695b615c67b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SetCMetric", "[", 
   RowBox[{"metric", ",", "torsion", ",", "gen", ",", 
    RowBox[{"SignatureOfMetric", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{"3", ",", "0", ",", "0"}], "}"}]}]}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellChangeTimes->{{3.731119952376243*^9, 3.731119994756598*^9}},
 CellLabel->"In[50]:=",
 CellID->1256506760,ExpressionUUID->"a93e5375-c72d-402f-86c7-faae4dc67634"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"8.011439`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.731119976183012*^9, 3.731120019070552*^9, 
  3.733368423300601*^9},
 CellLabel->"Out[50]=",
 CellID->1488132288,ExpressionUUID->"cd3b3551-3ca6-4663-9365-919b30b8a356"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
This is now the covariant derivative associated to the metric, by default \
with torsion:\
\>", "MathCaption",
 CellChangeTimes->{{3.73112004205068*^9, 3.731120047578763*^9}},
 CellID->1168661657,ExpressionUUID->"1b966922-3091-4455-82e7-b5db4df0e2bd"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{
    RowBox[{"cd", "=", 
     RowBox[{"CovDOfMetric", "[", "metric", "]"}]}], ";"}], ")"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellChangeTimes->{{3.7311194797334547`*^9, 3.731119494309866*^9}, {
  3.731120038740044*^9, 3.7311200389311657`*^9}},
 CellLabel->"In[51]:=",
 CellID->811033737,ExpressionUUID->"496ec91a-ac9f-41f6-bacf-999f3773ad11"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.000286`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.728676579419901*^9, 3.728753612547031*^9, 
  3.7288437536738167`*^9, 3.730956326927071*^9, 3.731119508416676*^9, 
  3.731119627896524*^9, 3.731120050225397*^9, 3.73336842350218*^9},
 CellLabel->"Out[51]=",
 CellID->1444811454,ExpressionUUID->"1888ecaf-85f8-4fba-aad8-e241bcef59f0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["It has the torsion we provided:", "MathCaption",
 CellID->1877902312,ExpressionUUID->"870ff78d-b19c-4422-b09d-9d1623aa8dc1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Torsion", "[", "cd", "]"}], "===", "torsion"}]], "Input",
 CellLabel->"In[52]:=",
 CellID->336997825,ExpressionUUID->"56df24df-365b-432a-b3bf-2b4accb6cb83"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728676579609042*^9, 3.728753612712714*^9, 
  3.7288437538751593`*^9, 3.730956327095848*^9, 3.731119510932192*^9, 
  3.73111962925498*^9, 3.7311200527941093`*^9, 3.733368423667666*^9},
 CellLabel->"Out[52]=",
 CellID->1348510749,ExpressionUUID->"655040cf-395c-4cb8-8a56-556a55426e5d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{"christoffel", "=", 
    RowBox[{"Christoffel", "[", 
     RowBox[{"cd", ",", "PDgen"}], "]"}]}], ")"}], "[", 
  RowBox[{"a", ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}]}], "]"}]], "Input",
 CellLabel->"In[53]:=",
 CellID->1094267069,ExpressionUUID->"d0d69b49-2553-412c-90dd-b38980e847f7"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {GridBox[{
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "-", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                  RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}]]}]],
              Annotation[#, {1, 1, 1} -> (Rational[1, 2]/($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))) (($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) Derivative[1, 0, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + ($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   Derivative[0, 1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - 2 
                   Derivative[1, 0, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) - ($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                  Derivative[0, 0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - 2 
                  Derivative[1, 0, 0][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                  RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}]]}]],
              Annotation[#, {1, 1, 2} -> (Rational[1, 2]/($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))) (($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   Derivative[0, 1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + ((-$CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) Derivative[1, 0, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + ($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) - Derivative[0, 0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   Derivative[0, 1, 0][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   Derivative[1, 0, 0][$CellContext`gyz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "3"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                  RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}]]}]],
              Annotation[#, {1, 1, 3} -> (Rational[1, 2]/($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))) (($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   Derivative[0, 0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + ((-$CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) - $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                   0, 1, 0][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   Derivative[1, 0, 0][$CellContext`gyz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + ($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) Derivative[1, 0, 0][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]), "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                  RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}]]}]],
              Annotation[#, {1, 2, 1} -> (Rational[1, 2]/($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))) (($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) Derivative[0, 1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + ($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                   1, 0, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) - ($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                  0, 1, 0][$CellContext`gxz][
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] - Derivative[1, 0, 0][$CellContext`gyz][
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]])), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], 
                  RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}]]}]],
              Annotation[#, {1, 2, 2} -> (Rational[1, 2]/($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))) (((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2) $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2) ($CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ($CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + (2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] Derivative[0, 1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                    Derivative[0, 0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - 2 
                    Derivative[0, 1, 0][$CellContext`gyz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])) + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ((2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2) $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] Derivative[0, 1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (2 Derivative[0, 1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                    1, 0, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ((2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - (2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ($CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) - $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (Derivative[0, 0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - 2 
                    Derivative[0, 1, 0][$CellContext`gyz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ((-2) 
                    Derivative[0, 1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                    Derivative[1, 0, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "3"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                  RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}]]}]],
              Annotation[#, {1, 2, 3} -> (Rational[1, 2]/($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))) (((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2) $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 ($CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (-$CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ((2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                    0, 1, 0][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])) + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) - $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                    Derivative[0, 0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ((-$CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + Derivative[0, 1, 0][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])) + ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   Derivative[0, 1, 0][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                   1, 0, 0][$CellContext`gyz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])), "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "1"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                  RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}]]}]],
              Annotation[#, {1, 3, 1} -> (Rational[1, 2]/($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))) (($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) Derivative[0, 0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + ((-$CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (-$CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) - $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                   0, 1, 0][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   Derivative[1, 0, 0][$CellContext`gyz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + ($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) (((-2) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - (2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - (2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   Derivative[1, 0, 0][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "2"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                  RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}]]}]],
              Annotation[#, {1, 3, 2} -> (Rational[1, 2]/($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))) ((((-2) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 ($CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (-$CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ((2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[0, 1, 0][$CellContext`gzz][
                    
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])) + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                    0, 0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ((-$CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + Derivative[0, 1, 0][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])) + ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   Derivative[0, 1, 0][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                   1, 0, 0][$CellContext`gyz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "3"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], 
                  RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}]]}]],
              Annotation[#, {1, 3, 3} -> (Rational[1, 2]/($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))) ((((-2) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2) $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^3) $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gyz[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] (((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (((-2) $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                    Derivative[0, 0, 1][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])) + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2) ($CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (((-2) $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                    Derivative[0, 0, 1][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (((-2) $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - 2 
                    Derivative[0, 0, 1][$CellContext`gyz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                    Derivative[0, 1, 0][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])) + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 ((2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   2 Derivative[0, 0, 1][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                   1, 0, 0][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ((2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                    2 Derivative[0, 0, 1][$CellContext`gyz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                    0, 1, 0][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (((-2) $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - 2 
                    Derivative[0, 0, 1][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                    Derivative[1, 0, 0][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))), "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                  RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}]]}]],
              Annotation[#, {2, 1, 1} -> (Rational[1, 2]/($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2) ($CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gxz[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] ((2 $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2) $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] Derivative[1, 0, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (Derivative[0, 1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - 2 
                    Derivative[1, 0, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (((-2) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2) ($CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] Derivative[1, 0, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (Derivative[0, 0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - 2 
                    Derivative[1, 0, 0][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (((-2) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ($CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + (2 $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) - $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] Derivative[0, 0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] Derivative[0, 1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - (2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) Derivative[1, 0, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) Derivative[1, 0, 0][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                  RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}]]}]],
              Annotation[#, {2, 1, 2} -> (Rational[1, 2]/($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))) ((-($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   Derivative[0, 1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + ($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) Derivative[1, 0, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) - Derivative[0, 0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   Derivative[0, 1, 0][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   Derivative[1, 0, 0][$CellContext`gyz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "3"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                  RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}]]}]],
              Annotation[#, {2, 1, 3} -> (Rational[1, 2]/($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))) ((-($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   Derivative[0, 0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + ($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) - $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                   0, 1, 0][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   Derivative[1, 0, 0][$CellContext`gyz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) Derivative[1, 0, 0][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]), "Tooltip"]& ]}
           },
           AutoDelete->False,
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                  RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}]]}]],
              Annotation[#, {2, 2, 1} -> (Rational[1, 2]/($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))) (((-$CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) Derivative[0, 1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + ($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - (2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   Derivative[1, 0, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                   0, 1, 0][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                   1, 0, 0][$CellContext`gyz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                  RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}]]}]],
              Annotation[#, {2, 2, 2} -> (Rational[1, 2]/($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2) ($CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 
                  Derivative[0, 1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] (((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2) $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] Derivative[0, 1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (Derivative[0, 0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - 2 
                    Derivative[0, 1, 0][$CellContext`gyz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ((((-2) $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - (2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - 2 
                    Derivative[0, 1, 0][$CellContext`gyz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                    2 Derivative[0, 1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                    1, 0, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])) + ($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - (2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - 2 
                   Derivative[0, 1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   Derivative[1, 0, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "3"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                  RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}]]}]],
              Annotation[#, {2, 2, 3} -> (Rational[1, 2]/($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))) (($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   Derivative[0, 0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) Derivative[0, 1, 0][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + ((-$CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   Derivative[0, 1, 0][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                   1, 0, 0][$CellContext`gyz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])), "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "1"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                  RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}]]}]],
              Annotation[#, {2, 3, 1} -> (Rational[1, 2]/($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))) (((-$CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) Derivative[0, 0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + ($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (-$CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) - $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                   0, 1, 0][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   Derivative[1, 0, 0][$CellContext`gyz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ((2 $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                   1, 0, 0][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "2"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"3", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                  RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}]]}]],
              Annotation[#, {2, 3, 2} -> (Rational[1, 2]/($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))) (($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ($CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) - $CellContext`gxx[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] ((2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2) $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] Derivative[0, 0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ((2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                    0, 1, 0][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])) + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ((-$CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ($CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - 3 $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) - Derivative[0, 1, 0][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) - $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                    Derivative[0, 1, 0][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                    1, 0, 0][$CellContext`gyz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])) + ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   Derivative[0, 1, 0][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                   1, 0, 0][$CellContext`gyz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "3"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], 
                  RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}]]}]],
              Annotation[#, {2, 3, 3} -> (Rational[1, 2]/($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))) ((2 $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^3) $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 (((-2) $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ($CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + (2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   2 Derivative[0, 0, 1][$CellContext`gyz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                   0, 1, 0][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (((-2) $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                    Derivative[0, 0, 1][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (((-2) $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - 2 
                    Derivative[0, 0, 1][$CellContext`gyz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                    Derivative[0, 1, 0][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])) + ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ((2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   2 Derivative[0, 0, 1][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                   1, 0, 0][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ((((-2) $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ((2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                    0, 0, 1][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (((-2) $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - (2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - 2 
                    Derivative[0, 0, 1][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                    Derivative[1, 0, 0][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))), "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "1"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "-", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                  RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}]]}]],
              Annotation[#, {3, 1, 1} -> (Rational[1, 2]/($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 (((-2) $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) - (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^3) $CellContext`T213[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] + ($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) Derivative[1, 0, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ((2 $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2) $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] Derivative[1, 0, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (Derivative[0, 1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - 2 
                    Derivative[1, 0, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 ((2 $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ($CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) - Derivative[0, 0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   2 Derivative[1, 0, 0][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2) $CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (((-2) $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] Derivative[0, 0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] Derivative[0, 1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) Derivative[1, 0, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) Derivative[1, 0, 0][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "2"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                  RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}]]}]],
              Annotation[#, {3, 1, 2} -> (Rational[1, 2]/($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))) (($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   Derivative[0, 1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) Derivative[1, 0, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) - Derivative[0, 0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   Derivative[0, 1, 0][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   Derivative[1, 0, 0][$CellContext`gyz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "3"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                  RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}]]}]],
              Annotation[#, {3, 1, 3} -> (Rational[1, 2]/($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))) (($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   Derivative[0, 0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) - $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                   0, 1, 0][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   Derivative[1, 0, 0][$CellContext`gyz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) Derivative[1, 0, 0][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]), "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "1"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                  RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}]]}]],
              Annotation[#, {3, 2, 1} -> (Rational[1, 2]/($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))) (($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) Derivative[0, 1, 0][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                   1, 0, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) - ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                  0, 1, 0][$CellContext`gxz][
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] - Derivative[1, 0, 0][$CellContext`gyz][
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]])), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "2"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], 
                  RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}]]}]],
              Annotation[#, {3, 2, 2} -> (Rational[1, 2]/($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))) (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^3) $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2) $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] Derivative[0, 1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ((2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                    Derivative[0, 0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - 2 
                    Derivative[0, 1, 0][$CellContext`gyz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])) - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - 2 
                  Derivative[0, 1, 0][$CellContext`gyz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + ($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   2 Derivative[0, 1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                   1, 0, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ((2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                    0, 1, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (((-2) $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - (2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - 2 
                    Derivative[0, 1, 0][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                    Derivative[1, 0, 0][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "3"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{"3", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                  RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}]]}]],
              Annotation[#, {3, 2, 3} -> (Rational[1, 2]/($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))) (($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ((2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                    Derivative[0, 0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ((2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                    0, 1, 0][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 ((-$CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ($CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + Derivative[0, 1, 0][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2) $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (-$CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 3 $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                    Derivative[0, 1, 0][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                    1, 0, 0][$CellContext`gyz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])) - ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                  Derivative[0, 1, 0][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                  1, 0, 0][$CellContext`gyz][
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]])), "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "1"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                  RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}]]}]],
              Annotation[#, {3, 3, 1} -> (Rational[1, 2]/($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))) (($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) Derivative[0, 0, 1][$CellContext`gxx][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + ((-$CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (-$CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) - $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (-$CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                   0, 1, 0][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   Derivative[1, 0, 0][$CellContext`gyz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) (((-2) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - (2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - (2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   Derivative[1, 0, 0][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "2"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["y", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T112"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T112, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T312"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T312, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T212"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T212, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                  RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}]]}]],
              Annotation[#, {3, 3, 2} -> (Rational[1, 2]/($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))) (($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] Derivative[0, 0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ((2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                    0, 1, 0][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])) - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + (2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                  0, 1, 0][$CellContext`gzz][
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]]) + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (((-2) $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2) $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (-$CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`T112[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gyy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                    Derivative[0, 1, 0][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                    1, 0, 0][$CellContext`gyz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])) - ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ((-$CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T312[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`T212[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                  Derivative[0, 1, 0][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                  1, 0, 0][$CellContext`gyz][
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]])), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "3"}], "}"}], "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T123"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T123, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T113"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T113, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T223"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T223, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T323"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T323, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], ")"}]}], 
                    "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T213"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T213, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"T313"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`T313, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                  RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gxx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gxx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], "-", 
                    
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gyy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gyy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"gzz"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`gzz, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "y", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`y[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], ")"}]}]]}]],
              Annotation[#, {3, 3, 3} -> (Rational[1, 2]/($CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - ((2 $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                   $CellContext`x[], 
                   $CellContext`y[], 
                   $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ($CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]))) ((2 $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2) ($CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2 
                  Derivative[0, 0, 1][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ((2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]^2) $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] Derivative[0, 0, 1][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ((2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                    2 Derivative[0, 0, 1][$CellContext`gyz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                    0, 1, 0][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])) + $CellContext`gxz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (((2 $CellContext`gxx[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T123[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] (((-2) $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) ($CellContext`T113[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + $CellContext`T223[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) - (2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T323[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - 2 
                    Derivative[0, 0, 1][$CellContext`gyz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                    Derivative[0, 1, 0][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) + $CellContext`gyy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] ((2 $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + (2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                    2 Derivative[0, 0, 1][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - Derivative[
                    1, 0, 0][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])) + ($CellContext`gxy[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) (((-2) $CellContext`gyz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T213[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - (2 $CellContext`gzz[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]]) $CellContext`T313[
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] - 2 
                   Derivative[0, 0, 1][$CellContext`gxz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]] + 
                   Derivative[1, 0, 0][$CellContext`gzz][
                    $CellContext`x[], 
                    $CellContext`y[], 
                    $CellContext`z[]])), "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 4 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 4 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a", " ", " "},
         {" ", "b", "c"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{{(Rational[1, 2]/($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - ((2 $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) $CellContext`gyz[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gzz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]))) (($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) Derivative[1, 0, 0][$CellContext`gxx][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + ($CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) ((2 $CellContext`gxx[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T112[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + (2 $CellContext`gxy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T212[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + (2 $CellContext`gxz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T312[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + Derivative[0, 1, 0][$CellContext`gxx][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - 2 Derivative[1, 0, 0][$CellContext`gxy][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) - ($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - $CellContext`gxy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] $CellContext`gyz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) ((2 $CellContext`gxx[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) $CellContext`T113[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + (2 $CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) $CellContext`T213[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + (2 $CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) $CellContext`T313[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + Derivative[0, 0, 1][$CellContext`gxx][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - 2 Derivative[1, 0, 0][$CellContext`gxz][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]])), (Rational[1, 2]/($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - ((2 $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) $CellContext`gyz[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gzz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]))) (($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) ((2 $CellContext`gxx[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T112[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + (2 $CellContext`gxy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T212[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + (2 $CellContext`gxz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T312[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + Derivative[0, 1, 0][$CellContext`gxx][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) + ((-$CellContext`gxz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) Derivative[1, 0, 0][$CellContext`gyy][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + ($CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) ((-$CellContext`gxx[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T123[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T212[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T213[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] ($CellContext`T113[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`T223[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) + $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T312[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T313[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ($CellContext`T112[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] - $CellContext`T323[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) - Derivative[0, 0, 1][$CellContext`gxy][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + Derivative[0, 1, 0][$CellContext`gxz][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + Derivative[1, 0, 0][$CellContext`gyz][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]])), (Rational[1, 2]/($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - ((2 $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) $CellContext`gyz[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gzz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]))) (($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) ((2 $CellContext`gxx[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T113[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + (2 $CellContext`gxy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T213[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + (2 $CellContext`gxz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T313[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + Derivative[0, 0, 1][$CellContext`gxx][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) + ((-$CellContext`gxz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) ($CellContext`gxx[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T123[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T212[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T213[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ($CellContext`T113[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + $CellContext`T223[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) - $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T312[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T313[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] (-$CellContext`T112[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + $CellContext`T323[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - Derivative[0, 1, 0][$CellContext`gxz][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + Derivative[1, 0, 0][$CellContext`gyz][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) + ($CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) Derivative[1, 0, 0][$CellContext`gzz][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]])}, {(Rational[1, 2]/($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - ((2 $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) $CellContext`gyz[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gzz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]))) (($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) Derivative[0, 1, 0][$CellContext`gxx][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + ($CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) ((2 $CellContext`gxy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T112[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + (2 $CellContext`gyy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T212[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + (2 $CellContext`gyz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T312[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - Derivative[1, 0, 0][$CellContext`gyy][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) - ($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - $CellContext`gxy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] $CellContext`gyz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) ($CellContext`gxx[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T123[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T212[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T213[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] ($CellContext`T113[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`T223[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) + $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T312[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T313[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] ($CellContext`T112[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`T323[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - Derivative[0, 1, 0][$CellContext`gxz][
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] - Derivative[1, 0, 0][$CellContext`gyz][
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]])), (Rational[1, 2]/($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - ((2 $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) $CellContext`gyz[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gzz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]))) (((2 $CellContext`gxy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]^2) $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`T123[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - (2 $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2) ($CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T212[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T223[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) + $CellContext`gxy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] (((-2) $CellContext`gyy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) ($CellContext`gzz[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] $CellContext`T112[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + $CellContext`gxz[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] $CellContext`T123[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]) + (2 $CellContext`gyz[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]^2) ($CellContext`T112[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + $CellContext`T323[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) + $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] Derivative[0, 1, 0][$CellContext`gyy][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ((2 $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) $CellContext`T223[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + Derivative[0, 0, 1][$CellContext`gyy][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] - 2 Derivative[0, 1, 0][$CellContext`gyz][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]])) + $CellContext`gyz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ((2 $CellContext`gyz[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]^2) $CellContext`T312[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] Derivative[0, 1, 0][$CellContext`gyy][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] (2 Derivative[0, 1, 0][$CellContext`gxy][
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] - Derivative[1, 0, 0][$CellContext`gyy][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]])) + $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ((2 $CellContext`gyz[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]^2) $CellContext`T212[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - (2 $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) ($CellContext`gzz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] $CellContext`T312[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + $CellContext`gxz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] $CellContext`T323[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) - $CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] (Derivative[0, 0, 1][$CellContext`gyy][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - 2 Derivative[0, 1, 0][$CellContext`gyz][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) + $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ((-2) Derivative[0, 1, 0][$CellContext`gxy][
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + Derivative[1, 0, 0][$CellContext`gyy][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]))), (Rational[1, 2]/($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - ((2 $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) $CellContext`gyz[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gzz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]))) (((2 $CellContext`gxy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]^2) $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`T123[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]^2 ($CellContext`T113[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + $CellContext`T223[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) + $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ($CellContext`gyy[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] (-$CellContext`T113[
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + $CellContext`T223[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gyy][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) + $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ((2 $CellContext`gzz[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) $CellContext`T323[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] - Derivative[0, 1, 0][$CellContext`gzz][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]])) + $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]^2 ($CellContext`T112[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] - $CellContext`T323[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) - $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] ((2 $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]) $CellContext`T123[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + (2 $CellContext`gyy[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]) $CellContext`T223[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + Derivative[0, 0, 1][$CellContext`gyy][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) + $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ((-$CellContext`gzz[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) ($CellContext`T112[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]] + $CellContext`T323[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) + Derivative[0, 1, 0][$CellContext`gzz][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]])) + ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) ($CellContext`gxx[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T123[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T213[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T312[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ($CellContext`T212[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + $CellContext`T313[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + Derivative[0, 1, 0][$CellContext`gxz][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - Derivative[1, 0, 0][$CellContext`gyz][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]))}, {(Rational[1, 2]/($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - ((2 $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) $CellContext`gyz[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gzz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]))) (($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) Derivative[0, 0, 1][$CellContext`gxx][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + ((-$CellContext`gxz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) ($CellContext`gxx[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T123[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T212[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T213[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] (-$CellContext`T113[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + $CellContext`T223[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) - $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T312[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T313[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] (-$CellContext`T112[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + $CellContext`T323[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - Derivative[0, 1, 0][$CellContext`gxz][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + Derivative[1, 0, 0][$CellContext`gyz][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) + ($CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) (((-2) $CellContext`gxz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T113[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - (2 $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) $CellContext`T213[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - (2 $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) $CellContext`T313[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + Derivative[1, 0, 0][$CellContext`gzz][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]])), (Rational[1, 2]/($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - ((2 $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) $CellContext`gyz[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gzz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]))) ((((-2) $CellContext`gxz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]^2) $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`T123[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]^2 ($CellContext`T113[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + $CellContext`T223[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) + $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ($CellContext`gyy[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] (-$CellContext`T113[
                $CellContext`x[], 
                $CellContext`y[], 
                $CellContext`z[]] + $CellContext`T223[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gyy][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) + $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ((2 $CellContext`gzz[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) $CellContext`T323[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] - Derivative[0, 1, 0][$CellContext`gzz][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]])) + $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]^2 ($CellContext`T112[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] - $CellContext`T323[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) + $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ((2 $CellContext`gxy[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) $CellContext`T123[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] - (2 $CellContext`gyy[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]) $CellContext`T223[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] - Derivative[0, 0, 1][$CellContext`gyy][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) + $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ((-$CellContext`gzz[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) ($CellContext`T112[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]] + $CellContext`T323[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) + Derivative[0, 1, 0][$CellContext`gzz][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]])) + ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) ((-$CellContext`gxx[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T123[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T213[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T312[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ($CellContext`T212[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + $CellContext`T313[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + Derivative[0, 1, 0][$CellContext`gxz][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - Derivative[1, 0, 0][$CellContext`gyz][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]])), (Rational[1, 2]/($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - ((2 $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) $CellContext`gyz[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gzz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]))) ((((-2) $CellContext`gxz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]^2) $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`T123[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + (2 $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^3) $CellContext`T213[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - $CellContext`gyz[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] (((2 $CellContext`gyy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) $CellContext`T213[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] (((-2) $CellContext`gzz[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]) $CellContext`T223[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + Derivative[0, 0, 1][$CellContext`gzz][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]])) + $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] (((2 $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]) $CellContext`gzz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T123[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + (2 $CellContext`gyz[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]^2) ($CellContext`T113[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] - $CellContext`T223[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) + $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] (((-2) $CellContext`gzz[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) $CellContext`T113[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + Derivative[0, 0, 1][$CellContext`gzz][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) + $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] (((-2) $CellContext`gzz[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) $CellContext`T323[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] - 2 Derivative[0, 0, 1][$CellContext`gyz][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + Derivative[0, 1, 0][$CellContext`gzz][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]])) + $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 ((2 $CellContext`gzz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T313[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + 2 Derivative[0, 0, 1][$CellContext`gxz][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - Derivative[1, 0, 0][$CellContext`gzz][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) + $CellContext`gzz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ((2 $CellContext`gzz[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) $CellContext`T323[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + 2 Derivative[0, 0, 1][$CellContext`gyz][
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] - Derivative[0, 1, 0][$CellContext`gzz][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) + $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] (((-2) $CellContext`gzz[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) $CellContext`T313[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] - 2 Derivative[0, 0, 1][$CellContext`gxz][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + Derivative[1, 0, 0][$CellContext`gzz][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]])))}}, {{(Rational[1, 2]/($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - ((2 $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) $CellContext`gyz[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gzz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]))) ((2 $CellContext`gxx[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2) ($CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T112[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T113[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) + ((2 $CellContext`gxy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]^2) $CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`T213[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - $CellContext`gxz[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] ((2 $CellContext`gxz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]^2) $CellContext`T312[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] Derivative[1, 0, 0][$CellContext`gxx][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] (Derivative[0, 1, 0][$CellContext`gxx][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - 2 Derivative[1, 0, 0][$CellContext`gxy][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]])) + $CellContext`gxy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] (((-2) $CellContext`gxz[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]^2) ($CellContext`T212[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] - $CellContext`T313[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) + $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] Derivative[1, 0, 0][$CellContext`gxx][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] (Derivative[0, 0, 1][$CellContext`gxx][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] - 2 Derivative[1, 0, 0][$CellContext`gxz][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]])) + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] (((-2) $CellContext`gxz[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]^2) $CellContext`T112[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + (2 $CellContext`gxy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) ($CellContext`gzz[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] $CellContext`T212[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] - $CellContext`gyz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] $CellContext`T213[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) + (2 $CellContext`gxz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) ($CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] $CellContext`T113[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + $CellContext`gzz[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] $CellContext`T312[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] - $CellContext`gyz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] $CellContext`T313[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) - $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] Derivative[0, 0, 1][$CellContext`gxx][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] Derivative[0, 1, 0][$CellContext`gxx][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - (2 $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) Derivative[1, 0, 0][$CellContext`gxy][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + (2 $CellContext`gyz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) Derivative[1, 0, 0][$CellContext`gxz][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]])), (Rational[1, 2]/($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - ((2 $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) $CellContext`gyz[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gzz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]))) ((-($CellContext`gxz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] $CellContext`gyz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] - $CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]])) ((2 $CellContext`gxx[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T112[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + (2 $CellContext`gxy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T212[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + (2 $CellContext`gxz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T312[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + Derivative[0, 1, 0][$CellContext`gxx][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) + ($CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gxx[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) Derivative[1, 0, 0][$CellContext`gyy][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + ((-$CellContext`gxy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gxx[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) ((-$CellContext`gxx[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T123[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T212[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T213[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] ($CellContext`T113[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`T223[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) + $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T312[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T313[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ($CellContext`T112[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] - $CellContext`T323[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) - Derivative[0, 0, 1][$CellContext`gxy][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + Derivative[0, 1, 0][$CellContext`gxz][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + Derivative[1, 0, 0][$CellContext`gyz][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]])), (Rational[1, 2]/($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - ((2 $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) $CellContext`gyz[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gzz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]))) ((-($CellContext`gxz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] $CellContext`gyz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] - $CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]])) ((2 $CellContext`gxx[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T113[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + (2 $CellContext`gxy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T213[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + (2 $CellContext`gxz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T313[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + Derivative[0, 0, 1][$CellContext`gxx][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) + ($CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gxx[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) ($CellContext`gxx[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T123[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T212[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T213[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ($CellContext`T113[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + $CellContext`T223[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) - $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T312[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T313[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] (-$CellContext`T112[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + $CellContext`T323[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - Derivative[0, 1, 0][$CellContext`gxz][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + Derivative[1, 0, 0][$CellContext`gyz][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) + ((-$CellContext`gxy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gxx[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) Derivative[1, 0, 0][$CellContext`gzz][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]])}, {(Rational[1, 2]/($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - ((2 $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) $CellContext`gyz[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gzz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]))) (((-$CellContext`gxz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) Derivative[0, 1, 0][$CellContext`gxx][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + ($CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gxx[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) (((-2) $CellContext`gxy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T112[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - (2 $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) $CellContext`T212[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - (2 $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) $CellContext`T312[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + Derivative[1, 0, 0][$CellContext`gyy][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) + ($CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gxx[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) ($CellContext`gxx[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T123[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T212[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T213[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ($CellContext`T113[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + $CellContext`T223[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) + $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T312[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T313[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ($CellContext`T112[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + $CellContext`T323[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - Derivative[0, 1, 0][$CellContext`gxz][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - Derivative[1, 0, 0][$CellContext`gyz][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]])), (Rational[1, 2]/($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - ((2 $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) $CellContext`gyz[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gzz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]))) ((2 $CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2) ($CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T112[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T123[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) + $CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 Derivative[0, 1, 0][$CellContext`gyy][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - $CellContext`gxx[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] (((2 $CellContext`gyy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) $CellContext`T223[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + (2 $CellContext`gyz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]^2) $CellContext`T323[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] Derivative[0, 1, 0][$CellContext`gyy][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] (Derivative[0, 0, 1][$CellContext`gyy][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - 2 Derivative[0, 1, 0][$CellContext`gyz][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]])) + $CellContext`gxy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ((((-2) $CellContext`gxx[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]) $CellContext`gyz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T123[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ((2 $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) $CellContext`T223[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] - (2 $CellContext`gyz[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]) ($CellContext`T112[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] - $CellContext`T323[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gyy][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] - 2 Derivative[0, 1, 0][$CellContext`gyz][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) + $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ((2 $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) $CellContext`T212[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + (2 $CellContext`gyz[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) $CellContext`T312[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + 2 Derivative[0, 1, 0][$CellContext`gxy][
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] - Derivative[1, 0, 0][$CellContext`gyy][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]])) + ($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) (((-2) $CellContext`gyy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T212[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - (2 $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) $CellContext`T312[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - 2 Derivative[0, 1, 0][$CellContext`gxy][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + Derivative[1, 0, 0][$CellContext`gyy][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]])), (Rational[1, 2]/($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - ((2 $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) $CellContext`gyz[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gzz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]))) (($CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gxx[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) ((2 $CellContext`gxy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T123[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + (2 $CellContext`gyy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T223[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + (2 $CellContext`gyz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T323[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + Derivative[0, 0, 1][$CellContext`gyy][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) + ((-$CellContext`gxy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gxx[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) Derivative[0, 1, 0][$CellContext`gzz][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + ((-$CellContext`gxz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) ($CellContext`gxx[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T123[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T212[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T213[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ($CellContext`T113[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + $CellContext`T223[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) + $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T312[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T313[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ($CellContext`T112[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + $CellContext`T323[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + Derivative[0, 1, 0][$CellContext`gxz][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - Derivative[1, 0, 0][$CellContext`gyz][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]))}, {(Rational[1, 2]/($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - ((2 $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) $CellContext`gyz[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gzz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]))) (((-$CellContext`gxz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) Derivative[0, 0, 1][$CellContext`gxx][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + ($CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gxx[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) ($CellContext`gxx[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T123[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T212[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T213[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] (-$CellContext`T113[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + $CellContext`T223[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) - $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T312[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T313[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] (-$CellContext`T112[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + $CellContext`T323[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - Derivative[0, 1, 0][$CellContext`gxz][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + Derivative[1, 0, 0][$CellContext`gyz][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) + ($CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gxx[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) ((2 $CellContext`gxz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T113[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + (2 $CellContext`gyz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T213[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + (2 $CellContext`gzz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T313[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - Derivative[1, 0, 0][$CellContext`gzz][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]])), (Rational[1, 2]/($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - ((2 $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) $CellContext`gyz[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gzz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]))) (($CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) ($CellContext`T113[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - $CellContext`T223[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) + $CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 ((2 $CellContext`gxy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T123[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] (-$CellContext`T112[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + $CellContext`T323[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gyy][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) - $CellContext`gxx[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] ((2 $CellContext`gyz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]^2) $CellContext`T223[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] Derivative[0, 0, 1][$CellContext`gyy][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] ((2 $CellContext`gzz[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]) $CellContext`T323[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] - Derivative[0, 1, 0][$CellContext`gzz][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]])) + $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ((-$CellContext`gyz[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) ($CellContext`T113[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]] - 3 $CellContext`T223[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) + $CellContext`gzz[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] ($CellContext`T112[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]] + $CellContext`T323[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) - Derivative[0, 1, 0][$CellContext`gzz][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) - $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] ($CellContext`gxx[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] $CellContext`T123[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + $CellContext`gyy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] $CellContext`T213[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + $CellContext`gzz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] $CellContext`T312[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + $CellContext`gyz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] ($CellContext`T212[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + $CellContext`T313[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + Derivative[0, 1, 0][$CellContext`gxz][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - Derivative[1, 0, 0][$CellContext`gyz][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]])) + ($CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) ((-$CellContext`gxx[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T123[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T213[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T312[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ($CellContext`T212[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + $CellContext`T313[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + Derivative[0, 1, 0][$CellContext`gxz][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - Derivative[1, 0, 0][$CellContext`gyz][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]])), (Rational[1, 2]/($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - ((2 $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) $CellContext`gyz[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gzz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]))) ((2 $CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^3) $CellContext`T123[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 (((-2) $CellContext`gyz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) ($CellContext`T113[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] - $CellContext`T223[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) + (2 $CellContext`gzz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T323[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + 2 Derivative[0, 0, 1][$CellContext`gyz][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - Derivative[0, 1, 0][$CellContext`gzz][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] (((-2) $CellContext`gzz[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) $CellContext`T223[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + Derivative[0, 0, 1][$CellContext`gzz][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) + $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] (((-2) $CellContext`gzz[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) $CellContext`T323[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] - 2 Derivative[0, 0, 1][$CellContext`gyz][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + Derivative[0, 1, 0][$CellContext`gzz][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]])) + ($CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) ((2 $CellContext`gyz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T213[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + (2 $CellContext`gzz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T313[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + 2 Derivative[0, 0, 1][$CellContext`gxz][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - Derivative[1, 0, 0][$CellContext`gzz][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) + $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ((((-2) $CellContext`gxx[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]) $CellContext`gzz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T123[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ((2 $CellContext`gzz[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) $CellContext`T113[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] - Derivative[0, 0, 1][$CellContext`gzz][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) + $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] (((-2) $CellContext`gyz[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) $CellContext`T213[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] - (2 $CellContext`gzz[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]) $CellContext`T313[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] - 2 Derivative[0, 0, 1][$CellContext`gxz][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + Derivative[1, 0, 0][$CellContext`gzz][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]])))}}, {{(Rational[1, 2]/($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - ((2 $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) $CellContext`gyz[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gzz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]))) ($CellContext`gxx[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 (((-2) $CellContext`gyz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T112[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + (2 $CellContext`gyy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T113[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) - (2 $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^3) $CellContext`T213[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] + ($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) Derivative[1, 0, 0][$CellContext`gxx][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ((2 $CellContext`gxz[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]^2) $CellContext`T312[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] Derivative[1, 0, 0][$CellContext`gxx][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] (Derivative[0, 1, 0][$CellContext`gxx][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] - 2 Derivative[1, 0, 0][$CellContext`gxy][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]])) + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 ((2 $CellContext`gxz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) ($CellContext`T212[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] - $CellContext`T313[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) - Derivative[0, 0, 1][$CellContext`gxx][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + 2 Derivative[1, 0, 0][$CellContext`gxz][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] (((-2) $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]^2) $CellContext`T113[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + (2 $CellContext`gxy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) ($CellContext`gxz[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] $CellContext`T112[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] - $CellContext`gyz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] $CellContext`T212[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + $CellContext`gyy[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] $CellContext`T213[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]) + $CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] (((-2) $CellContext`gyz[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) $CellContext`T312[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + (2 $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) $CellContext`T313[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]) + $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] Derivative[0, 0, 1][$CellContext`gxx][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] Derivative[0, 1, 0][$CellContext`gxx][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + (2 $CellContext`gyz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) Derivative[1, 0, 0][$CellContext`gxy][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - (2 $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) Derivative[1, 0, 0][$CellContext`gxz][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]])), (Rational[1, 2]/($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - ((2 $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) $CellContext`gyz[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gzz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]))) (($CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) ((2 $CellContext`gxx[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T112[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + (2 $CellContext`gxy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T212[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + (2 $CellContext`gxz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T312[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + Derivative[0, 1, 0][$CellContext`gxx][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) + ((-$CellContext`gxy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gxx[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) Derivative[1, 0, 0][$CellContext`gyy][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + ($CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gxx[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) ((-$CellContext`gxx[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T123[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T212[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T213[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] ($CellContext`T113[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`T223[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) + $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T312[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T313[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ($CellContext`T112[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] - $CellContext`T323[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) - Derivative[0, 0, 1][$CellContext`gxy][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + Derivative[0, 1, 0][$CellContext`gxz][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + Derivative[1, 0, 0][$CellContext`gyz][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]])), (Rational[1, 2]/($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - ((2 $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) $CellContext`gyz[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gzz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]))) (($CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) ((2 $CellContext`gxx[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T113[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + (2 $CellContext`gxy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T213[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + (2 $CellContext`gxz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T313[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + Derivative[0, 0, 1][$CellContext`gxx][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) + ((-$CellContext`gxy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gxx[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) ($CellContext`gxx[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T123[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T212[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T213[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ($CellContext`T113[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + $CellContext`T223[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) - $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T312[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T313[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] (-$CellContext`T112[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + $CellContext`T323[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - Derivative[0, 1, 0][$CellContext`gxz][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + Derivative[1, 0, 0][$CellContext`gyz][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) + ($CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gxx[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) Derivative[1, 0, 0][$CellContext`gzz][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]])}, {(Rational[1, 2]/($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - ((2 $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) $CellContext`gyz[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gzz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]))) (($CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) Derivative[0, 1, 0][$CellContext`gxx][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + ($CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gxx[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) ((2 $CellContext`gxy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T112[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + (2 $CellContext`gyy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T212[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + (2 $CellContext`gyz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T312[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - Derivative[1, 0, 0][$CellContext`gyy][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) - ($CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 - $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) ($CellContext`gxx[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T123[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T212[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T213[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] ($CellContext`T113[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`T223[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) + $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T312[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T313[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] ($CellContext`T112[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`T323[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - Derivative[0, 1, 0][$CellContext`gxz][
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] - Derivative[1, 0, 0][$CellContext`gyz][
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]])), (Rational[1, 2]/($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - ((2 $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) $CellContext`gyz[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gzz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]))) (((-2) $CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^3) $CellContext`T123[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ((2 $CellContext`gyy[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]^2) $CellContext`T223[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] Derivative[0, 1, 0][$CellContext`gyy][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ((2 $CellContext`gyz[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) $CellContext`T323[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + Derivative[0, 0, 1][$CellContext`gyy][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] - 2 Derivative[0, 1, 0][$CellContext`gyz][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]])) - $CellContext`gxy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]^2 ((2 $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) $CellContext`T223[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + (2 $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) ($CellContext`T112[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`T323[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gyy][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - 2 Derivative[0, 1, 0][$CellContext`gyz][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) + ($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) ((2 $CellContext`gyy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T212[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + (2 $CellContext`gyz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T312[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + 2 Derivative[0, 1, 0][$CellContext`gxy][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - Derivative[1, 0, 0][$CellContext`gyy][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) + $CellContext`gxy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] (((2 $CellContext`gxx[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]) $CellContext`gyy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T123[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ((2 $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) $CellContext`T112[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] - Derivative[0, 1, 0][$CellContext`gyy][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) + $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] (((-2) $CellContext`gyy[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) $CellContext`T212[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] - (2 $CellContext`gyz[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]) $CellContext`T312[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] - 2 Derivative[0, 1, 0][$CellContext`gxy][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + Derivative[1, 0, 0][$CellContext`gyy][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]))), (Rational[1, 2]/($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - ((2 $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) $CellContext`gyz[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gzz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]))) (($CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) ($CellContext`T112[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`T323[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ((2 $CellContext`gyz[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) $CellContext`T323[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + Derivative[0, 0, 1][$CellContext`gyy][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) + $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ((2 $CellContext`gyz[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) $CellContext`T223[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] - Derivative[0, 1, 0][$CellContext`gzz][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]])) + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 ((-$CellContext`gyz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) ($CellContext`T113[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + $CellContext`T223[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) + Derivative[0, 1, 0][$CellContext`gzz][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) + $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] (((-2) $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]^2) $CellContext`T123[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] ($CellContext`gyy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] (-$CellContext`T113[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]] + $CellContext`T223[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]) + $CellContext`gyz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] ($CellContext`T112[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + 3 $CellContext`T323[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gyy][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) + $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ($CellContext`gxx[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] $CellContext`T123[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + $CellContext`gyy[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] $CellContext`T213[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + $CellContext`gzz[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] $CellContext`T312[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + $CellContext`gyz[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] ($CellContext`T212[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]] + $CellContext`T313[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + Derivative[0, 1, 0][$CellContext`gxz][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] - Derivative[1, 0, 0][$CellContext`gyz][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]])) - ($CellContext`gxy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] $CellContext`gyz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) ((-$CellContext`gxx[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) $CellContext`T123[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T213[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T312[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] ($CellContext`T212[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`T313[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + Derivative[0, 1, 0][$CellContext`gxz][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - Derivative[1, 0, 0][$CellContext`gyz][
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]]))}, {(Rational[1, 2]/($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - ((2 $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) $CellContext`gyz[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gzz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]))) (($CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) Derivative[0, 0, 1][$CellContext`gxx][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + ((-$CellContext`gxy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gxx[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) ($CellContext`gxx[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T123[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T212[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T213[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] (-$CellContext`T113[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + $CellContext`T223[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) - $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T312[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T313[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] (-$CellContext`T112[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + $CellContext`T323[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - Derivative[0, 1, 0][$CellContext`gxz][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + Derivative[1, 0, 0][$CellContext`gyz][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) + ($CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gxx[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) (((-2) $CellContext`gxz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T113[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - (2 $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) $CellContext`T213[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - (2 $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) $CellContext`T313[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + Derivative[1, 0, 0][$CellContext`gzz][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]])), (Rational[1, 2]/($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - ((2 $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) $CellContext`gyz[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gzz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]))) (($CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) ($CellContext`T112[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - $CellContext`T323[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] Derivative[0, 0, 1][$CellContext`gyy][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ((2 $CellContext`gyz[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) $CellContext`T223[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + (2 $CellContext`gzz[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) $CellContext`T323[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] - Derivative[0, 1, 0][$CellContext`gzz][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]])) - $CellContext`gxy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]^2 ($CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] ($CellContext`T113[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`T223[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) + (2 $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) $CellContext`T323[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - Derivative[0, 1, 0][$CellContext`gzz][
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]]) + $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] (((-2) $CellContext`gxy[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]^2) $CellContext`T123[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] ($CellContext`gyy[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] (-$CellContext`T113[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]] + $CellContext`T223[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]) + $CellContext`gyz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] ($CellContext`T112[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] - $CellContext`T323[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gyy][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) + $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ($CellContext`gxx[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] $CellContext`T123[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + $CellContext`gyy[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] $CellContext`T213[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + $CellContext`gzz[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] $CellContext`T312[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + $CellContext`gyz[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] ($CellContext`T212[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]] + $CellContext`T313[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + Derivative[0, 1, 0][$CellContext`gxz][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] - Derivative[1, 0, 0][$CellContext`gyz][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]])) - ($CellContext`gxy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] $CellContext`gyz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) ((-$CellContext`gxx[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) $CellContext`T123[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T213[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T312[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] ($CellContext`T212[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`T313[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) + Derivative[0, 0, 1][$CellContext`gxy][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + Derivative[0, 1, 0][$CellContext`gxz][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - Derivative[1, 0, 0][$CellContext`gyz][
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]])), (Rational[1, 2]/($CellContext`gxz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gyy[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] - ((2 $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]]) $CellContext`gyz[
         $CellContext`x[], 
         $CellContext`y[], 
         $CellContext`z[]] + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 $CellContext`gzz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ($CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2 - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gzz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]))) ((2 $CellContext`gxz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]^2) ($CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] $CellContext`T113[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`T123[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) + $CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]^2 Derivative[0, 0, 1][$CellContext`gzz][
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] + $CellContext`gxx[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] ((2 $CellContext`gyz[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]^2) $CellContext`T223[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - $CellContext`gyy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] Derivative[0, 0, 1][$CellContext`gzz][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + $CellContext`gyz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ((2 $CellContext`gzz[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) $CellContext`T323[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + 2 Derivative[0, 0, 1][$CellContext`gyz][
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] - Derivative[0, 1, 0][$CellContext`gzz][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]])) + $CellContext`gxz[
          $CellContext`x[], 
          $CellContext`y[], 
          $CellContext`z[]] (((2 $CellContext`gxx[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]) $CellContext`gyz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T123[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] + $CellContext`gxy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] (((-2) $CellContext`gyz[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) ($CellContext`T113[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]] + $CellContext`T223[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) - (2 $CellContext`gzz[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]]) $CellContext`T323[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] - 2 Derivative[0, 0, 1][$CellContext`gyz][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]] + Derivative[0, 1, 0][$CellContext`gzz][
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) + $CellContext`gyy[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] ((2 $CellContext`gyz[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) $CellContext`T213[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + (2 $CellContext`gzz[
               $CellContext`x[], 
               $CellContext`y[], 
               $CellContext`z[]]) $CellContext`T313[
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] + 2 Derivative[0, 0, 1][$CellContext`gxz][
              $CellContext`x[], 
              $CellContext`y[], 
              $CellContext`z[]] - Derivative[1, 0, 0][$CellContext`gzz][
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]])) + ($CellContext`gxy[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] $CellContext`gyz[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]) (((-2) $CellContext`gyz[
             $CellContext`x[], 
             $CellContext`y[], 
             $CellContext`z[]]) $CellContext`T213[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]] - (2 $CellContext`gzz[
            $CellContext`x[], 
            $CellContext`y[], 
            $CellContext`z[]]) $CellContext`T313[
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] - 2 Derivative[0, 0, 1][$CellContext`gxz][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]] + Derivative[1, 0, 0][$CellContext`gzz][
           $CellContext`x[], 
           $CellContext`y[], 
           $CellContext`z[]]))}}}, {$CellContext`gen, -$CellContext`gen, \
-$CellContext`gen}, 0][$CellContext`a, -$CellContext`b, -$CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{3.728676580013177*^9, 3.728753613057827*^9, 
  3.7288437543014097`*^9, 3.7309563274587717`*^9, 3.7311195155036087`*^9, 
  3.73111963050414*^9, 3.7311200548232183`*^9, 3.733368424039878*^9},
 CellLabel->"Out[53]=",
 CellID->2025231652,ExpressionUUID->"ee1108bb-9a9e-44ef-b55b-42c1a941f7e4"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"christoffel", "[", 
     RowBox[{"a", ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}]}], "]"}], "-", 
    RowBox[{"christoffel", "[", 
     RowBox[{"a", ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "b"}]}], "]"}]}], "\[Equal]", 
   RowBox[{"torsion", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}]}], "//", "Simplify"}]], "Input",
 CellLabel->"In[54]:=",
 CellID->1920205738,ExpressionUUID->"0815256a-948d-4663-84df-2c49039ec206"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728676580170499*^9, 3.7287536131625147`*^9, 
  3.728843754526066*^9, 3.73095632760211*^9, 3.731119517902569*^9, 
  3.731119631542882*^9, 3.731120056524002*^9, 3.733368424186301*^9},
 CellLabel->"Out[54]=",
 CellID->452570214,ExpressionUUID->"b96a8c3d-bbfe-48cc-b6f2-77035e8078de"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The connection is associated to the metric:", "MathCaption",
 CellID->1194722857,ExpressionUUID->"d38eccff-39bb-4f25-a870-106e5d6317e7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TensorDerivative", "[", 
  RowBox[{"metric", ",", "cd"}], "]"}]], "Input",
 CellLabel->"In[55]:=",
 CellID->734958149,ExpressionUUID->"7b4da674-2b8b-4a6f-8267-7e7702dc5877"],

Cell[BoxData["Zero"], "Output",
 CellChangeTimes->{3.728676580378783*^9, 3.728753613357127*^9, 
  3.728843754869331*^9, 3.730956327766659*^9, 3.7311195212754583`*^9, 
  3.7311196329272327`*^9, 3.7311200580202637`*^9, 3.733368424393641*^9},
 CellLabel->"Out[55]=",
 CellID->2107992671,ExpressionUUID->"3d7e3082-4f27-42cc-904d-23d8cbe5a330"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Compute everything. We simplify using Together because Simplify is too slow \
in this case:\
\>", "MathCaption",
 CellID->638351558,ExpressionUUID->"696aae64-e798-4a06-abb6-2d939248a9b1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"MetricCompute", "[", 
   RowBox[{"metric", ",", "torsion", ",", "gen", ",", 
    RowBox[{"\"\<Riemann\>\"", "[", 
     RowBox[{
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}], ",", "1"}], "]"}], ",", 
    RowBox[{"CVSimplify", "\[Rule]", "Together"}], ",", 
    RowBox[{"Parallelize", "\[Rule]", "True"}]}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellChangeTimes->{{3.731119650962965*^9, 3.731119654528495*^9}, {
  3.731119824556274*^9, 3.731119824979342*^9}},
 CellLabel->"In[56]:=",
 CellID->2093818541,ExpressionUUID->"baf5ca67-ddf2-4eb4-9e37-bfb59464edf7"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"7.163495`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.728676588078182*^9, 3.728753620740806*^9, 
  3.728843762583496*^9, 3.7309563349748*^9, 3.731119533228957*^9, 
  3.731120077160934*^9, 3.733368431687449*^9},
 CellLabel->"Out[56]=",
 CellID->1995096938,ExpressionUUID->"6505fc3b-e6c5-4e3e-8a36-9c8e2b8f1677"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["It has curvature:", "MathCaption",
 CellChangeTimes->{3.731119544803987*^9},
 CellID->1409375136,ExpressionUUID->"30709b74-1cdd-4f95-9886-5849c4e3a081"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"(", 
    RowBox[{"riemann", "=", 
     RowBox[{"Riemann", "[", "cd", "]"}]}], ")"}], ";"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[57]:=",
 CellID->623284342,ExpressionUUID->"dd2ef8f7-6083-4d5c-a96f-c2e7b6efa54f"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.001722`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.728676588253688*^9, 3.728753620883922*^9, 
  3.7288437627693*^9, 3.730956335104103*^9, 3.7311200806493464`*^9, 
  3.7333684318384123`*^9},
 CellLabel->"Out[57]=",
 CellID->784572708,ExpressionUUID->"74833659-82f4-4a11-a031-5449210d1530"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"(", 
    RowBox[{"riemanndown", "=", 
     RowBox[{"RiemannDown", "[", "cd", "]"}]}], ")"}], ";"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[58]:=",
 CellID->1569700537,ExpressionUUID->"b58136a0-7722-4980-8d4b-de2e26211bec"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.000732`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.7286765884015217`*^9, 3.728753621013775*^9, 
  3.728843762944063*^9, 3.730956335220962*^9, 3.731120082464321*^9, 
  3.73336843198735*^9},
 CellLabel->"Out[58]=",
 CellID->763910776,ExpressionUUID->"ab02b11a-d9a7-484a-9775-cab0b1af3930"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Symmetry is NOT complete in this 3d case:", "MathCaption",
 CellID->1469137121,ExpressionUUID->"e247d33d-fc52-4208-9e5a-1ff5d2e98cd3"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SymmetryGroupOfTensor", "[", "riemann", "]"}]], "Input",
 CellLabel->"In[59]:=",
 CellID->20619441,ExpressionUUID->"5cffd40b-5662-4fb5-bb77-658076c0e057"],

Cell[BoxData[
 RowBox[{"StrongGenSet", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "2"}], "}"}], ",", 
   RowBox[{"GenSet", "[", 
    InterpretationBox[
     StyleBox[
      RowBox[{"-", 
       RowBox[{"(", "1", ",", "2", ")"}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     -xAct`xPerm`Cycles[{1, 2}],
     Editable->False], "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.7286765921495743`*^9, 3.728753624914178*^9, 
  3.7288437666877337`*^9, 3.7309563389446487`*^9, 3.7311200879541693`*^9, 
  3.7333684357927647`*^9},
 CellLabel->"Out[59]=",
 CellID->1391170466,ExpressionUUID->"0417c870-e907-4305-aa6f-505cc7f6c87c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SymmetryGroupOfTensor", "[", "riemanndown", "]"}]], "Input",
 CellLabel->"In[60]:=",
 CellID->1421394764,ExpressionUUID->"28ab8d1a-d150-40a6-8d3c-a76fff65953e"],

Cell[BoxData[
 RowBox[{"StrongGenSet", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"3", ",", "1"}], "}"}], ",", 
   RowBox[{"GenSet", "[", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"-", 
        RowBox[{"(", "3", ",", "4", ")"}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      -xAct`xPerm`Cycles[{3, 4}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[
       RowBox[{"-", 
        RowBox[{"(", "1", ",", "2", ")"}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      -xAct`xPerm`Cycles[{1, 2}],
      Editable->False]}], "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728676593239558*^9, 3.728753625985368*^9, 
  3.728843767823838*^9, 3.730956339951598*^9, 3.731120089661538*^9, 
  3.733368436841938*^9},
 CellLabel->"Out[60]=",
 CellID->1747557173,ExpressionUUID->"24e15609-53e2-4b78-b6a4-1f43925a7c3c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Clean up:", "MathCaption",
 CellID->719058655,ExpressionUUID->"55771f9d-b07f-4127-b6e1-669777713aa2"],

Cell[BoxData[
 RowBox[{"UnsetCMetric", "[", "metric", "]"}]], "Input",
 CellLabel->"In[61]:=",
 CellID->1328332916,ExpressionUUID->"580d1839-a2de-44ff-8ec6-be5399c9537b"],

Cell[BoxData[
 RowBox[{
  RowBox[{"UndefScalarFunction", "/@", "$ScalarFunctions"}], ";"}]], "Input",
 CellLabel->"In[62]:=",
 CellID->51325000,ExpressionUUID->"a69898e0-f11e-4b11-b2b1-85212891f342"],

Cell[BoxData[
 RowBox[{"UndefChart", "[", "gen", "]"}]], "Input",
 CellLabel->"In[63]:=",
 CellID->1478332553,ExpressionUUID->"e3b47781-5c33-4395-837f-bd6058e95ffd"],

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M", "]"}]], "Input",
 CellLabel->"In[64]:=",
 CellID->2096906128,ExpressionUUID->"d215e680-a434-4fb6-aff7-3dba2faf7ffa"],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{
  "metric", ",", "torsion", ",", "cd", ",", "christoffel", ",", "riemann", 
   ",", "riemanndown"}], "]"}]], "Input",
 CellLabel->"In[65]:=",
 CellID->347162379,ExpressionUUID->"61226aa2-351e-421f-9914-2d6a5c3c967e"],

Cell[BoxData[
 RowBox[{"ClearxCobaCache", "[", "All", "]"}]], "Input",
 CellLabel->"In[66]:=",
 CellID->45034954,ExpressionUUID->"1468d86d-eff6-4efc-a3b8-260e3a38a554"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.5.4. A hard example: The Pomeransky-Senkov metric", "Subsubsection",
 CellChangeTimes->{{3.7286766060360003`*^9, 3.7286766063545313`*^9}, 
   3.7908767520574102`*^9},
 CellID->988228134,ExpressionUUID->"d2ff033d-5806-4796-862a-95f90d52c195"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<xAct`xCoba`\>\"", "]"}]], "Input",
 CellLabel->"In[1]:=",
 CellID->511722240,ExpressionUUID->"d1192b7a-a9cc-44ad-9566-10e581a22e17"],

Cell[CellGroupData[{

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.7309563795388813`*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->127239746,ExpressionUUID->"421a3cfd-dc36-4922-a7fa-80123cbbdb09"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`xPerm`  version \"\>", 
   "\[InvisibleSpace]", "\<\"1.2.3\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2015", ",", "8", ",", "23"}], "}"}]}],
  SequenceForm["Package xAct`xPerm`  version ", "1.2.3", ", ", {2015, 8, 23}],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.730956379552923*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1772182257,ExpressionUUID->"4aec092c-ad01-493b-95b8-b20e0b8ae14d"],

Cell[BoxData["\<\"CopyRight (C) 2003-2018, Jose M. Martin-Garcia, under the \
General Public License.\"\>"], "Print",
 CellChangeTimes->{3.73095637957051*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1616493156,ExpressionUUID->"af5afd6a-5948-420b-a784-1e15bc43096f"],

Cell[BoxData["\<\"Connecting to external mac executable...\"\>"], "Print",
 CellChangeTimes->{3.730956379584647*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->584854695,ExpressionUUID->"6b903f41-d5d0-4df3-9026-e97c04933794"],

Cell[BoxData["\<\"Connection established.\"\>"], "Print",
 CellChangeTimes->{3.7309563795971622`*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->620206546,ExpressionUUID->"4c75b752-a665-44cd-bcc6-6a814fcab2aa"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.730956379611802*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->2073279731,ExpressionUUID->"e50dafbb-1cf7-48e7-a9fc-b29474f82ff7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`xTensor`  version \"\>", 
   "\[InvisibleSpace]", "\<\"1.1.3\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2018", ",", "2", ",", "28"}], "}"}]}],
  SequenceForm[
  "Package xAct`xTensor`  version ", "1.1.3", ", ", {2018, 2, 28}],
  Editable->False]], "Print",
 CellChangeTimes->{3.730956379625037*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->542782695,ExpressionUUID->"62a19878-f686-4e5d-9b0e-efb14ef37468"],

Cell[BoxData["\<\"CopyRight (C) 2002-2018, Jose M. Martin-Garcia, under the \
General Public License.\"\>"], "Print",
 CellChangeTimes->{3.7309563796384974`*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->38222409,ExpressionUUID->"2f73d07b-3711-4be0-90f5-fbed18457c8c"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.73095637971623*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->306417610,ExpressionUUID->"909a848e-c6bb-4e6c-90d5-fbb55c9c8914"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`xCoba`  version \"\>", 
   "\[InvisibleSpace]", "\<\"0.8.4\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2018", ",", "2", ",", "28"}], "}"}]}],
  SequenceForm["Package xAct`xCoba`  version ", "0.8.4", ", ", {2018, 2, 28}],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.730956379729575*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->294985764,ExpressionUUID->"f7b2637d-4dd7-4f41-a82a-72dda31f8212"],

Cell[BoxData["\<\"CopyRight (C) 2005-2018, David Yllanes and Jose M. \
Martin-Garcia, under the General Public License.\"\>"], "Print",
 CellChangeTimes->{3.730956379742036*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1433478272,ExpressionUUID->"4b0e0004-6e46-4ada-8e64-8c9c5f3a3431"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.730956379754175*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1997765458,ExpressionUUID->"97341e2a-bc0d-48d2-a661-64c5ff1ac1b1"],

Cell[BoxData["\<\"These packages come with ABSOLUTELY NO WARRANTY; for \
details type Disclaimer[]. This is free software, and you are welcome to \
redistribute it under certain conditions. See the General Public License for \
details.\"\>"], "Print",
 CellChangeTimes->{3.730956379767593*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->537231761,ExpressionUUID->"c574d2f4-f62c-436d-8356-f5a11f5f4977"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.7309563797809267`*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1232955139,ExpressionUUID->"888474dc-47e7-47aa-a861-7bf627aa7857"]
}, Open  ]]
}, Closed]],

Cell[BoxData[
 RowBox[{
  RowBox[{"$DefInfoQ", "=", 
   RowBox[{"$UndefInfoQ", "=", "False"}]}], ";"}]], "Input",
 CellLabel->"In[2]:=",
 CellID->1991165092,ExpressionUUID->"33ced122-c44d-4771-803b-297ed3a150e8"],

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M", ",", "5", ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f"}], "}"}]}], 
  "]"}]], "Input",
 CellLabel->"In[3]:=",
 CellID->1274030043,ExpressionUUID->"08253957-1249-4212-9bfd-b54df69cbdf0"],

Cell[BoxData[
 RowBox[{"DefChart", "[", 
  RowBox[{"PS", ",", "M", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2", ",", "3", ",", "4"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"t", "[", "]"}], ",", 
     RowBox[{"x", "[", "]"}], ",", 
     RowBox[{"y", "[", "]"}], ",", 
     RowBox[{"\[Phi]", "[", "]"}], ",", 
     RowBox[{"\[Psi]", "[", "]"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[4]:=",
 CellID->63488464,ExpressionUUID->"65b825f3-7a55-4f14-b2ef-8b6bdd7aecb2"],

Cell[CellGroupData[{

Cell[TextData[{
 "Formulas taken from [",
 ButtonBox["arXiv:0911.0802",
  BaseStyle->"Hyperlink",
  ButtonData->{
    URL["http://arxiv.org/pdf/0911.0802v2.pdf"], None},
  ButtonNote->"http://arxiv.org/pdf/0911.0802v2.pdf"],
 "]:"
}], "MathCaption",
 CellID->1898725536,ExpressionUUID->"8b1c0f79-8dca-494b-87ae-d0f07b4c879a"],

Cell[BoxData[
 RowBox[{"DefConstantSymbol", "[", 
  RowBox[{"{", 
   RowBox[{"\[Lambda]", ",", "\[Nu]", ",", "k"}], "}"}], "]"}]], "Input",
 CellLabel->"In[5]:=",
 CellID->273089132,ExpressionUUID->"c8dc1aa6-e9ac-4cf3-9885-ab4a87eb0d06"],

Cell[BoxData[
 RowBox[{
  RowBox[{"ds2", "=", 
   RowBox[{
    RowBox[{
     FractionBox[
      RowBox[{"2", " ", 
       RowBox[{"H", "[", 
        RowBox[{"x", ",", "y"}], "]"}], " ", 
       SuperscriptBox["k", "2"], " "}], 
      RowBox[{
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "\[Nu]"}], ")"}], "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"x", "-", "y"}], ")"}], "2"]}]], 
     RowBox[{"(", 
      RowBox[{
       FractionBox[
        SuperscriptBox["dx", "2"], 
        RowBox[{"G", "[", "x", "]"}]], "-", 
       FractionBox[
        SuperscriptBox["dy", "2"], 
        RowBox[{"G", "[", "y", "]"}]]}], ")"}]}], "-", 
    RowBox[{"2", 
     FractionBox[
      RowBox[{"J", "[", 
       RowBox[{"x", ",", "y"}], "]"}], 
      RowBox[{"H", "[", 
       RowBox[{"y", ",", "x"}], "]"}]], "d\[Phi]", " ", "d\[Psi]"}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{"H", "[", 
       RowBox[{"y", ",", "x"}], "]"}], 
      RowBox[{"H", "[", 
       RowBox[{"x", ",", "y"}], "]"}]], 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"dt", "+", "\[CapitalOmega]"}], ")"}], "2"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{"F", "[", 
       RowBox[{"x", ",", "y"}], "]"}], 
      RowBox[{"H", "[", 
       RowBox[{"y", ",", "x"}], "]"}]], " ", 
     SuperscriptBox["d\[Psi]", "2"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{"F", "[", 
       RowBox[{"y", ",", "x"}], "]"}], 
      RowBox[{"H", "[", 
       RowBox[{"y", ",", "x"}], "]"}]], " ", 
     SuperscriptBox["d\[Phi]", "2"]}]}]}], ";"}]], "Input",
 CellLabel->"In[6]:=",
 CellID->636487885,ExpressionUUID->"abb14f69-54f8-4549-b3b6-92503e86e821"],

Cell[BoxData[
 RowBox[{
  RowBox[{"\[CapitalOmega]", ":=", 
   RowBox[{
    RowBox[{
     RowBox[{"M", "[", 
      RowBox[{"x", ",", "y"}], "]"}], " ", "d\[Psi]"}], "+", 
    RowBox[{
     RowBox[{"P", "[", 
      RowBox[{"x", ",", "y"}], "]"}], " ", "d\[Phi]"}]}]}], ";"}]], "Input",
 CellLabel->"In[7]:=",
 CellID->835837488,ExpressionUUID->"dc66e6fc-e43d-476c-85b3-21b842575739"],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"M", "[", 
    RowBox[{"x_", ",", "y_"}], "]"}], ":=", 
   FractionBox[
    RowBox[{"2", " ", "k", " ", "\[Lambda]", " ", 
     SqrtBox[
      RowBox[{
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"\[Nu]", "+", "1"}], ")"}], "2"], "-", 
       SuperscriptBox["\[Lambda]", "2"]}]], " ", 
     RowBox[{"(", 
      RowBox[{"y", "+", "1"}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "\[Lambda]"}], "+", "\[Nu]", "-", 
       RowBox[{"2", " ", "\[Nu]", " ", "x"}], "+", 
       RowBox[{"\[Nu]", " ", "x", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"(", 
            RowBox[{"\[Lambda]", "+", "\[Nu]", "-", "1"}], ")"}], " ", "x"}], 
          "+", "2"}], ")"}], " ", "y"}], "-", "1"}], ")"}]}], 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"1", "-", "\[Lambda]", "+", "\[Nu]"}], ")"}], " ", 
     RowBox[{"H", "[", 
      RowBox[{"y", ",", "x"}], "]"}]}]]}], ";"}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{"P", "[", 
    RowBox[{"x_", ",", "y_"}], "]"}], ":=", 
   FractionBox[
    RowBox[{"2", " ", "k", " ", "\[Lambda]", " ", 
     SqrtBox["\[Nu]"], " ", 
     SqrtBox[
      RowBox[{
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"\[Nu]", "+", "1"}], ")"}], "2"], "-", 
       SuperscriptBox["\[Lambda]", "2"]}]], " ", 
     RowBox[{"(", 
      RowBox[{
       SuperscriptBox["x", "2"], "-", "1"}], ")"}], " ", "y"}], 
    RowBox[{"H", "[", 
     RowBox[{"y", ",", "x"}], "]"}]]}], ";"}]}], "Input",
 CellLabel->"In[8]:=",
 CellID->1230125789,ExpressionUUID->"9def3a93-8b4e-4195-ac40-49a51b230c9b"],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"H", "[", 
    RowBox[{"x_", ",", "y_"}], "]"}], ":=", 
   RowBox[{
    SuperscriptBox["\[Lambda]", "2"], "+", 
    RowBox[{"2", " ", "\[Nu]", " ", 
     RowBox[{"(", 
      RowBox[{"1", "-", 
       SuperscriptBox["x", "2"]}], ")"}], " ", "y", " ", "\[Lambda]"}], "+", 
    RowBox[{"2", " ", "x", " ", 
     RowBox[{"(", 
      RowBox[{"1", "-", 
       RowBox[{
        SuperscriptBox["\[Nu]", "2"], " ", 
        SuperscriptBox["y", "2"]}]}], ")"}], " ", "\[Lambda]"}], "-", 
    SuperscriptBox["\[Nu]", "2"], "+", 
    RowBox[{"\[Nu]", " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", 
        SuperscriptBox["\[Lambda]", "2"]}], "-", 
       SuperscriptBox["\[Nu]", "2"], "+", "1"}], ")"}], " ", 
     SuperscriptBox["x", "2"], " ", 
     SuperscriptBox["y", "2"]}], "+", "1"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{"F", "[", 
    RowBox[{"x_", ",", "y_"}], "]"}], ":=", 
   RowBox[{
    FractionBox[
     RowBox[{"2", " ", 
      SuperscriptBox["k", "2"]}], 
     RowBox[{
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"x", "-", "y"}], ")"}], "2"], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "-", "\[Nu]"}], ")"}], "2"]}]], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"(", 
        RowBox[{"1", "-", 
         SuperscriptBox["y", "2"]}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{"1", "-", "\[Nu]"}], ")"}], "2"], "-", 
            SuperscriptBox["\[Lambda]", "2"]}], ")"}], " ", 
          RowBox[{"(", 
           RowBox[{"\[Nu]", "+", "1"}], ")"}]}], "+", 
         RowBox[{"y", " ", "\[Lambda]", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", 
             SuperscriptBox["\[Lambda]", "2"]}], "-", 
            RowBox[{"3", " ", 
             SuperscriptBox["\[Nu]", "2"]}], "+", 
            RowBox[{"2", " ", "\[Nu]"}], "+", "1"}], ")"}]}]}], ")"}], " ", 
       RowBox[{"G", "[", "x", "]"}]}], "+", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", 
           RowBox[{"(", 
            RowBox[{"1", "-", "\[Nu]"}], ")"}]}], " ", "\[Nu]", " ", 
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox["\[Lambda]", "2"], "+", 
            SuperscriptBox["\[Nu]", "2"], "-", "1"}], ")"}], " ", 
          SuperscriptBox["x", "4"]}], "+", 
         RowBox[{"\[Lambda]", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"2", " ", 
             SuperscriptBox["\[Nu]", "3"]}], "-", 
            RowBox[{"3", " ", 
             SuperscriptBox["\[Nu]", "2"]}], "-", 
            SuperscriptBox["\[Lambda]", "2"], "+", "1"}], ")"}], " ", 
          SuperscriptBox["x", "3"]}], "+", 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{"1", "-", "\[Nu]"}], ")"}], "2"], "-", 
            SuperscriptBox["\[Lambda]", "2"]}], ")"}], " ", 
          RowBox[{"(", 
           RowBox[{"\[Nu]", "+", "1"}], ")"}], " ", 
          SuperscriptBox["x", "2"]}], "+", 
         RowBox[{"\[Lambda]", " ", 
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox["\[Lambda]", "2"], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{"1", "-", "\[Nu]"}], ")"}], "2"]}], ")"}], " ", "x"}], 
         "+", 
         RowBox[{"2", " ", 
          SuperscriptBox["\[Lambda]", "2"]}]}], ")"}], " ", 
       RowBox[{"G", "[", "y", "]"}]}]}], ")"}]}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{"J", "[", 
    RowBox[{"x_", ",", "y_"}], "]"}], ":=", 
   FractionBox[
    RowBox[{"2", " ", 
     SuperscriptBox["k", "2"], " ", 
     RowBox[{"(", 
      RowBox[{"1", "-", 
       SuperscriptBox["x", "2"]}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{"1", "-", 
       SuperscriptBox["y", "2"]}], ")"}], " ", "\[Lambda]", " ", 
     SqrtBox["\[Nu]"], " ", 
     RowBox[{"(", 
      RowBox[{
       SuperscriptBox["\[Lambda]", "2"], "+", 
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{"x", "+", "y"}], ")"}], " ", "\[Nu]", " ", "\[Lambda]"}], 
       "-", 
       SuperscriptBox["\[Nu]", "2"], "-", 
       RowBox[{"x", " ", "y", " ", "\[Nu]", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", 
           SuperscriptBox["\[Lambda]", "2"]}], "-", 
          SuperscriptBox["\[Nu]", "2"], "+", "1"}], ")"}]}], "+", "1"}], 
      ")"}]}], 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"x", "-", "y"}], ")"}], " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"1", "-", "\[Nu]"}], ")"}], "2"]}]]}], ";"}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{"G", "[", "x_", "]"}], ":=", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{"1", "-", 
      SuperscriptBox["x", "2"]}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"\[Nu]", " ", 
       SuperscriptBox["x", "2"]}], "+", 
      RowBox[{"\[Lambda]", " ", "x"}], "+", "1"}], ")"}]}]}], ";"}]}], "Input",\

 CellLabel->"In[10]:=",
 CellID->899066054,ExpressionUUID->"2ad47032-5f48-4164-a1ce-e32797ee3e25"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{
    RowBox[{"matrix", "=", 
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{"CoefficientArrays", "[", 
         RowBox[{"ds2", ",", 
          RowBox[{"{", 
           RowBox[{
           "dt", ",", "dx", ",", "dy", ",", "d\[Phi]", ",", "d\[Psi]"}], 
           "}"}], ",", 
          RowBox[{"\"\<Symmetric\>\"", "\[Rule]", "True"}]}], "]"}], "[", 
        RowBox[{"[", "3", "]"}], "]"}], "//", "Normal"}], "//", 
      "Simplify"}]}], ";"}], ")"}], "//", "AbsoluteTiming"}]], "Input",
 CellLabel->"In[14]:=",
 CellID->908229286,ExpressionUUID->"43079b39-5b79-41b4-b087-712ea7e0bfcf"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"4.101053`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.728676634807064*^9, 3.72867669628505*^9, 
  3.7287536304644613`*^9, 3.7288437763747396`*^9, 3.730956385280395*^9},
 CellLabel->"Out[14]=",
 CellID->1207311270,ExpressionUUID->"45b207d6-8b28-47d5-bf2f-3bd63a0d1cb0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["A check:", "MathCaption",
 CellID->874642477,ExpressionUUID->"914c2c82-9474-4baa-9fad-b9def86c8fae"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"ds2", "-", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"dt", ",", "dx", ",", "dy", ",", "d\[Phi]", ",", "d\[Psi]"}], 
      "}"}], ".", "matrix", ".", 
     RowBox[{"{", 
      RowBox[{"dt", ",", "dx", ",", "dy", ",", "d\[Phi]", ",", "d\[Psi]"}], 
      "}"}]}]}], "//", "Simplify"}], "//", "AbsoluteTiming"}]], "Input",
 CellLabel->"In[15]:=",
 CellID->590547224,ExpressionUUID->"59fa8d46-64bd-46cd-ae80-c694afac595e"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.819206`", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.7286766359290247`*^9, 3.728676697480486*^9, 
  3.728753631554785*^9, 3.728843777507987*^9, 3.7309563862260733`*^9},
 CellLabel->"Out[15]=",
 CellID->1855018034,ExpressionUUID->"0a4f4464-479e-4a1a-a082-72ccd2b0cb32"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"met", "=", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"matrix", "/.", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"x", "\[Rule]", 
         RowBox[{"x", "[", "]"}]}], ",", 
        RowBox[{"y", "\[Rule]", 
         RowBox[{"y", "[", "]"}]}]}], "}"}]}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "PS"}], ",", 
       RowBox[{"-", "PS"}]}], "}"}]}], "]"}]}], ";"}]], "Input",
 CellLabel->"In[16]:=",
 CellID->1372342851,ExpressionUUID->"3250959d-1182-47e1-a4c8-616626dc37df"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Simplify is way too slow for this metric. In this case it is more useful to \
simplify with Together:\
\>", "MathCaption",
 CellID->991927977,ExpressionUUID->"d55d703d-49dc-4dfd-8197-734176dabe08"],

Cell[BoxData[
 RowBox[{
  RowBox[{"$CVSimplify", "=", "Together"}], ";"}]], "Input",
 CellLabel->"In[17]:=",
 CellID->417957778,ExpressionUUID->"15699e43-b181-4529-98e3-b8c1609bf578"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Inverse has a hard time with this matrix. Specify explicitly the signature:\
\>", "MathCaption",
 CellID->197365509,ExpressionUUID->"3e594bed-dec5-427f-a434-ffac56d210aa"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SetCMetric", "[", 
   RowBox[{"met", ",", "PS", ",", 
    RowBox[{"SignatureOfMetric", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{"4", ",", "1", ",", "0"}], "}"}]}], ",", 
    RowBox[{"Verbose", "\[Rule]", "True"}]}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[18]:=",
 CellID->1407358795,ExpressionUUID->"f4531ce2-9861-44d1-9abd-0e35d4c37ab9"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Metric\"\>", "[", 
    RowBox[{"1", ",", "1"}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Metric"[1, 1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676636210846*^9, 3.7286766977737017`*^9, 
  3.728753631846854*^9, 3.7288437779703503`*^9, 3.730956386743865*^9},
 CellLabel->"During evaluation of In[18]:=",
 CellID->120510292,ExpressionUUID->"71fd4389-51e5-4da6-b2ab-e9009bc173ba"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Constructed in \"\>", "\[InvisibleSpace]", "28.625969`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Constructed in ", 28.625969, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676636210846*^9, 3.7286766977737017`*^9, 
  3.728753631846854*^9, 3.7288437779703503`*^9, 3.730956415335332*^9},
 CellLabel->"During evaluation of In[18]:=",
 CellID->339398523,ExpressionUUID->"ff30b7c5-9abc-4d4f-9ea9-c08e4b372d59"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Applied \"\>", "\[InvisibleSpace]", "Together", 
   "\[InvisibleSpace]", "\<\" in \"\>", "\[InvisibleSpace]", "1.533548`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Applied ", Together, " in ", 1.533548, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676636210846*^9, 3.7286766977737017`*^9, 
  3.728753631846854*^9, 3.7288437779703503`*^9, 3.730956416868148*^9},
 CellLabel->"During evaluation of In[18]:=",
 CellID->554409629,ExpressionUUID->"1f792a75-f092-4e6f-9cd6-155adffd023b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Stored in \"\>", "\[InvisibleSpace]", "0.001378`", 
   "\[InvisibleSpace]", "\<\" seconds and \"\>", "\[InvisibleSpace]", 
   "234848", "\[InvisibleSpace]", "\<\" bytes\"\>"}],
  SequenceForm["    Stored in ", 0.001378, " seconds and ", 234848, " bytes"],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.728676636210846*^9, 3.7286766977737017`*^9, 
  3.728753631846854*^9, 3.7288437779703503`*^9, 3.7309564168808823`*^9},
 CellLabel->"During evaluation of In[18]:=",
 CellID->31090270,ExpressionUUID->"6e4a4c19-f548-4f7c-a93b-bed51194d48f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"DMetric\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "DMetric"[-1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676636210846*^9, 3.7286766977737017`*^9, 
  3.728753631846854*^9, 3.7288437779703503`*^9, 3.730956416893446*^9},
 CellLabel->"During evaluation of In[18]:=",
 CellID->695641669,ExpressionUUID->"282296ef-b224-47bb-847e-e0fb3bfbcbb0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Constructed in \"\>", "\[InvisibleSpace]", "0.002695`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Constructed in ", 0.002695, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676636210846*^9, 3.7286766977737017`*^9, 
  3.728753631846854*^9, 3.7288437779703503`*^9, 3.730956416905376*^9},
 CellLabel->"During evaluation of In[18]:=",
 CellID->748556615,ExpressionUUID->"4a7b48b2-960c-4c64-8cef-da5fa0f4bea9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Applied \"\>", "\[InvisibleSpace]", "Together", 
   "\[InvisibleSpace]", "\<\" in \"\>", "\[InvisibleSpace]", "0.596092`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Applied ", Together, " in ", 0.596092, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676636210846*^9, 3.7286766977737017`*^9, 
  3.728753631846854*^9, 3.7288437779703503`*^9, 3.7309564174760513`*^9},
 CellLabel->"During evaluation of In[18]:=",
 CellID->1031824179,ExpressionUUID->"a70bbca8-15f2-4c98-b2c3-08e8378885b3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Christoffel\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Christoffel"[-1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676636210846*^9, 3.7286766977737017`*^9, 
  3.728753631846854*^9, 3.7288437779703503`*^9, 3.73095641749415*^9},
 CellLabel->"During evaluation of In[18]:=",
 CellID->938526928,ExpressionUUID->"d990a980-fbdb-4d53-a006-176257fa205e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Constructed in \"\>", "\[InvisibleSpace]", "0.003063`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Constructed in ", 0.003063, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676636210846*^9, 3.7286766977737017`*^9, 
  3.728753631846854*^9, 3.7288437779703503`*^9, 3.7309564175119762`*^9},
 CellLabel->"During evaluation of In[18]:=",
 CellID->554691681,ExpressionUUID->"5204f59f-6b26-4220-9ae6-0a02943e034e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Applied \"\>", "\[InvisibleSpace]", "Together", 
   "\[InvisibleSpace]", "\<\" in \"\>", "\[InvisibleSpace]", "0.66771`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Applied ", Together, " in ", 0.66771, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676636210846*^9, 3.7286766977737017`*^9, 
  3.728753631846854*^9, 3.7288437779703503`*^9, 3.730956418166299*^9},
 CellLabel->"During evaluation of In[18]:=",
 CellID->476570940,ExpressionUUID->"4128648a-937b-4e66-94a0-8864938c67fc"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Christoffel\"\>", "[", 
    RowBox[{"1", ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Christoffel"[1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676636210846*^9, 3.7286766977737017`*^9, 
  3.728753631846854*^9, 3.7288437779703503`*^9, 3.730956418221395*^9},
 CellLabel->"During evaluation of In[18]:=",
 CellID->1761268917,ExpressionUUID->"b6c2b842-6976-4021-81df-edf20f3e0916"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Constructed in \"\>", "\[InvisibleSpace]", "0.010721`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Constructed in ", 0.010721, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676636210846*^9, 3.7286766977737017`*^9, 
  3.728753631846854*^9, 3.7288437779703503`*^9, 3.730956418239599*^9},
 CellLabel->"During evaluation of In[18]:=",
 CellID->732097771,ExpressionUUID->"6f18d7a9-0fae-4742-837e-cc90dec8a5d8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Applied \"\>", "\[InvisibleSpace]", "Together", 
   "\[InvisibleSpace]", "\<\" in \"\>", "\[InvisibleSpace]", "2.757399`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Applied ", Together, " in ", 2.757399, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676636210846*^9, 3.7286766977737017`*^9, 
  3.728753631846854*^9, 3.7288437779703503`*^9, 3.730956421031508*^9},
 CellLabel->"During evaluation of In[18]:=",
 CellID->1433423534,ExpressionUUID->"b630ca4e-2643-470a-8910-e69b2bf8cd0e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", 
   "\[InvisibleSpace]", "\<\"CovDOfMetric\"\>"}],
  SequenceForm["** ReportCompute: ", "CovDOfMetric"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676636210846*^9, 3.7286766977737017`*^9, 
  3.728753631846854*^9, 3.7288437779703503`*^9, 3.7309564210430603`*^9},
 CellLabel->"During evaluation of In[18]:=",
 CellID->1077783892,ExpressionUUID->"74277c14-3d0f-4d75-b4b1-d784291351bf"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"34.408922`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.7286767328215933`*^9, 3.728753667495758*^9, 
  3.7288438135464087`*^9, 3.7309564211693487`*^9},
 CellLabel->"Out[18]=",
 CellID->289310399,ExpressionUUID->"d6be4381-1ae8-46c9-91f5-086136cdab71"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Now we compute all the way down to the Ricci tensor.", "MathCaption",
 CellID->1725377310,ExpressionUUID->"76f7ce72-3ee5-4788-b441-374f6b5a30c4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"MetricCompute", "[", 
   RowBox[{"met", ",", "PS", ",", 
    RowBox[{"\"\<Ricci\>\"", "[", 
     RowBox[{
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}]}], "]"}], ",", 
    RowBox[{"Verbose", "\[Rule]", "True"}], ",", 
    RowBox[{"Parallelize", "\[Rule]", "True"}]}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[19]:=",
 CellID->944763729,ExpressionUUID->"691708be-6030-40cc-9e1b-e08600b6a8b5"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Metric\"\>", "[", 
    RowBox[{"1", ",", "1"}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Metric"[1, 1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676732999867*^9, 3.728753667735461*^9, 
  3.728843813867194*^9, 3.730956421337194*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->536899185,ExpressionUUID->"5d92a489-8b65-48c6-a905-9d424cf73517"],

Cell[BoxData["\<\"    Already cached\"\>"], "Print",
 CellChangeTimes->{3.728676732999867*^9, 3.728753667735461*^9, 
  3.728843813867194*^9, 3.730956421357283*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->2085801248,ExpressionUUID->"8977fdc5-4290-4e83-9162-19abb0775fff"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"DMetric\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "DMetric"[-1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676732999867*^9, 3.728753667735461*^9, 
  3.728843813867194*^9, 3.730956421375717*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->538521508,ExpressionUUID->"ddff720b-4cfd-4123-add7-f58405422a1d"],

Cell[BoxData["\<\"    Already cached\"\>"], "Print",
 CellChangeTimes->{3.728676732999867*^9, 3.728753667735461*^9, 
  3.728843813867194*^9, 3.73095642138735*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1434896498,ExpressionUUID->"01b06207-7dc3-42a5-a3b5-f5469a16cd3c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"DDMetric\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "DDMetric"[-1, -1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676732999867*^9, 3.728753667735461*^9, 
  3.728843813867194*^9, 3.730956421398097*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->64086076,ExpressionUUID->"fa2d5371-cbc4-4d48-82dd-1c63e84052bb"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Constructed in \"\>", "\[InvisibleSpace]", "0.058546`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Constructed in ", 0.058546, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676732999867*^9, 3.728753667735461*^9, 
  3.728843813867194*^9, 3.730956421408826*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1205372185,ExpressionUUID->"81b7e4a3-682a-40aa-bd6b-44e695bdd24f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Applied \"\>", "\[InvisibleSpace]", "Together", 
   "\[InvisibleSpace]", "\<\" in \"\>", "\[InvisibleSpace]", "4.621189`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Applied ", Together, " in ", 4.621189, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676732999867*^9, 3.728753667735461*^9, 
  3.728843813867194*^9, 3.730956426008918*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->223757888,ExpressionUUID->"6742f342-aa69-454d-bedf-113c203b968c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Christoffel\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Christoffel"[-1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676732999867*^9, 3.728753667735461*^9, 
  3.728843813867194*^9, 3.73095642609972*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->454957719,ExpressionUUID->"bb164395-5360-4bac-8526-2fa68c9038b8"],

Cell[BoxData["\<\"    Already cached\"\>"], "Print",
 CellChangeTimes->{3.728676732999867*^9, 3.728753667735461*^9, 
  3.728843813867194*^9, 3.7309564261145763`*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1432599913,ExpressionUUID->"fd5b6e90-25a7-4855-96d8-2a28bbe952b6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Christoffel\"\>", "[", 
    RowBox[{"1", ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Christoffel"[1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676732999867*^9, 3.728753667735461*^9, 
  3.728843813867194*^9, 3.730956426129437*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->120087627,ExpressionUUID->"b6e5d160-6716-4771-944b-79d0769e5ffa"],

Cell[BoxData["\<\"    Already cached\"\>"], "Print",
 CellChangeTimes->{3.728676732999867*^9, 3.728753667735461*^9, 
  3.728843813867194*^9, 3.730956426144491*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1825602456,ExpressionUUID->"d292bca8-b96a-4dcb-bf4d-a43f5530dfaf"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", 
   "\[InvisibleSpace]", "\<\"CovDOfMetric\"\>"}],
  SequenceForm["** ReportCompute: ", "CovDOfMetric"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676732999867*^9, 3.728753667735461*^9, 
  3.728843813867194*^9, 3.73095642615851*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1801613303,ExpressionUUID->"7badd94e-aa3f-4497-8130-3113b5b320ce"],

Cell[BoxData["\<\"    Already cached\"\>"], "Print",
 CellChangeTimes->{3.728676732999867*^9, 3.728753667735461*^9, 
  3.728843813867194*^9, 3.7309564261735907`*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->195360821,ExpressionUUID->"c95421d7-6b76-4788-b4fc-0ed441d1d60f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Riemann\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Riemann"[-1, -1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676732999867*^9, 3.728753667735461*^9, 
  3.728843813867194*^9, 3.730956426188033*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->2074305931,ExpressionUUID->"74ad2dae-0348-442f-b0e9-20ea9d7ef8e3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Constructed in \"\>", "\[InvisibleSpace]", "0.030915`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Constructed in ", 0.030915, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676732999867*^9, 3.728753667735461*^9, 
  3.728843813867194*^9, 3.730956426202447*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->610256919,ExpressionUUID->"c35edef4-7f77-4501-917c-d5f8afb503c4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Applied \"\>", "\[InvisibleSpace]", "Together", 
   "\[InvisibleSpace]", "\<\" in \"\>", "\[InvisibleSpace]", "7.384309`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Applied ", Together, " in ", 7.384309, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676732999867*^9, 3.728753667735461*^9, 
  3.728843813867194*^9, 3.730956433548736*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->206139561,ExpressionUUID->"ffe0a830-01d9-4652-8b1c-d0514caaba16"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Stored in \"\>", "\[InvisibleSpace]", "0.404685`", 
   "\[InvisibleSpace]", "\<\" seconds and \"\>", "\[InvisibleSpace]", 
   "95170840", "\[InvisibleSpace]", "\<\" bytes\"\>"}],
  SequenceForm[
  "    Stored in ", 0.404685, " seconds and ", 95170840, " bytes"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676732999867*^9, 3.728753667735461*^9, 
  3.728843813867194*^9, 3.7309564341639223`*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->696479116,ExpressionUUID->"a7cc9aca-b8de-4cce-9d57-137ca50a9c9d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Riemann\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", "1"}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Riemann"[-1, -1, -1, 1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676732999867*^9, 3.728753667735461*^9, 
  3.728843813867194*^9, 3.7309564342248077`*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1938626462,ExpressionUUID->"ee94c78f-7587-4e43-823b-b70231c7b54c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Constructed in \"\>", "\[InvisibleSpace]", "0.075057`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Constructed in ", 0.075057, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676732999867*^9, 3.728753667735461*^9, 
  3.728843813867194*^9, 3.730956434291094*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->873054550,ExpressionUUID->"17f66a69-f34c-4f7c-99f3-4aaa33a5503d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Applied \"\>", "\[InvisibleSpace]", "Together", 
   "\[InvisibleSpace]", "\<\" in \"\>", "\[InvisibleSpace]", "13.622471`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Applied ", Together, " in ", 13.622471, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676732999867*^9, 3.728753667735461*^9, 
  3.728843813867194*^9, 3.73095644789159*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1819545546,ExpressionUUID->"06beb516-8aa4-40f9-9574-7ac35602e2c2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Stored in \"\>", "\[InvisibleSpace]", "0.442943`", 
   "\[InvisibleSpace]", "\<\" seconds and \"\>", "\[InvisibleSpace]", 
   "117519088", "\[InvisibleSpace]", "\<\" bytes\"\>"}],
  SequenceForm[
  "    Stored in ", 0.442943, " seconds and ", 117519088, " bytes"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676732999867*^9, 3.728753667735461*^9, 
  3.728843813867194*^9, 3.730956448934361*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->934858342,ExpressionUUID->"56180939-8371-44c2-a699-06782fa3a9a0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Ricci\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Ricci"[-1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676732999867*^9, 3.728753667735461*^9, 
  3.728843813867194*^9, 3.730956448995315*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1173792542,ExpressionUUID->"4087d0a7-3fbd-4aa2-9614-0b00313c9460"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Constructed in \"\>", "\[InvisibleSpace]", "0.00078`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Constructed in ", 0.00078, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676732999867*^9, 3.728753667735461*^9, 
  3.728843813867194*^9, 3.730956449009843*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1146472052,ExpressionUUID->"28811e61-9f5f-40ab-af4a-5be4246a2e94"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Applied \"\>", "\[InvisibleSpace]", "Together", 
   "\[InvisibleSpace]", "\<\" in \"\>", "\[InvisibleSpace]", "1.074412`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Applied ", Together, " in ", 1.074412, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676732999867*^9, 3.728753667735461*^9, 
  3.728843813867194*^9, 3.730956450032914*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1532497080,ExpressionUUID->"be7c0b94-c90f-4329-bee5-78469f856c30"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Stored in \"\>", "\[InvisibleSpace]", "0.015552`", 
   "\[InvisibleSpace]", "\<\" seconds and \"\>", "\[InvisibleSpace]", "640", 
   "\[InvisibleSpace]", "\<\" bytes\"\>"}],
  SequenceForm["    Stored in ", 0.015552, " seconds and ", 640, " bytes"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676732999867*^9, 3.728753667735461*^9, 
  3.728843813867194*^9, 3.730956450086808*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1907045089,ExpressionUUID->"9946078b-61a3-49e1-ac0e-bca256f6f7c0"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"28.743406`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.7286767631804523`*^9, 3.728753694062134*^9, 
  3.7288438406018267`*^9, 3.730956450101224*^9},
 CellLabel->"Out[19]=",
 CellID->513379348,ExpressionUUID->"8fc9d204-92c1-476f-9b5c-9e3e8b951d22"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["It is a Ricci-flat metric:", "MathCaption",
 CellID->277432000,ExpressionUUID->"701bd548-5abb-4f78-bc58-0192e45baaab"],

Cell[BoxData[
 RowBox[{
  RowBox[{"cd", "=", 
   RowBox[{"CovDOfMetric", "[", "met", "]"}]}], ";"}]], "Input",
 CellLabel->"In[20]:=",
 CellID->612626363,ExpressionUUID->"49cb1615-bf96-4a31-a6f9-577d4e955a2d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Ricci", "[", "cd", "]"}]], "Input",
 CellLabel->"In[21]:=",
 CellID->1374679223,ExpressionUUID->"acb4bb2a-3fe3-4058-ab0f-af37462f12e8"],

Cell[BoxData["Zero"], "Output",
 CellChangeTimes->{3.728676763522859*^9, 3.728753694310535*^9, 
  3.7288438408728647`*^9, 3.7309564504625177`*^9},
 CellLabel->"Out[21]=",
 CellID->1924891837,ExpressionUUID->"7e516e77-ceeb-469e-946b-ad2c434bb636"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{
  "met", ",", "cd", ",", "dt", ",", "dx", ",", "dy", ",", "d\[Phi]", ",", 
   "d\[Psi]", ",", "ds2", ",", "F", ",", "G", ",", "H", ",", "J", ",", "P", 
   ",", "\[CapitalOmega]", ",", "matrix"}], "]"}]], "Input",
 CellLabel->"In[22]:=",
 CellID->858534719,ExpressionUUID->"54417a2d-f882-4dc8-8d5a-3fc8b86eb538"],

Cell[BoxData[
 RowBox[{"UndefConstantSymbol", "[", 
  RowBox[{"{", 
   RowBox[{"\[Lambda]", ",", "\[Nu]", ",", "k"}], "}"}], "]"}]], "Input",
 CellLabel->"In[23]:=",
 CellID->450398040,ExpressionUUID->"0d103c7f-a147-4058-99ab-379fe8034401"],

Cell[BoxData[
 RowBox[{"UndefChart", "[", "PS", "]"}]], "Input",
 CellLabel->"In[24]:=",
 CellID->2107486252,ExpressionUUID->"c740d95f-ea7b-4365-a149-5758a6661051"],

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M", "]"}]], "Input",
 CellLabel->"In[25]:=",
 CellID->2073647393,ExpressionUUID->"89645d35-3094-4ff5-959b-30727bf8c0fd"],

Cell[BoxData[
 RowBox[{"ClearxCobaCache", "[", "All", "]"}]], "Input",
 CellLabel->"In[26]:=",
 CellID->279626988,ExpressionUUID->"08be4813-9aca-442d-b555-67e3d353e9e1"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
3.5.5. A hard example: Kerr-Newman-deSitter in Boyer-Lindquist-like \
\[Theta]-coordinates\
\>", "Subsubsection",
 CellChangeTimes->{{3.72867667503309*^9, 3.728676675199717*^9}, 
   3.7908767536130867`*^9},
 CellID->996567680,ExpressionUUID->"fcd30e6a-3b7a-4b2d-ac9f-d045bd6ba2b1"],

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<xAct`xCoba`\>\"", "]"}]], "Input",
 CellLabel->"In[27]:=",
 CellID->1088034807,ExpressionUUID->"7215fc00-a883-41b3-ab15-cf9c19a80c0b"],

Cell[BoxData[
 RowBox[{
  RowBox[{"$DefInfoQ", "=", 
   RowBox[{"$UndefInfoQ", "=", "False"}]}], ";"}]], "Input",
 CellLabel->"In[28]:=",
 CellID->777747044,ExpressionUUID->"062f5dd5-6f0a-407c-911a-4aa001fe9eba"],

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M", ",", "4", ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f"}], "}"}]}], 
  "]"}]], "Input",
 CellLabel->"In[29]:=",
 CellID->2009092866,ExpressionUUID->"54c95a90-8e6a-49cf-8d88-36fc9aff7f97"],

Cell[BoxData[
 RowBox[{"DefChart", "[", 
  RowBox[{"BL", ",", "M", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2", ",", "3"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"t", "[", "]"}], ",", 
     RowBox[{"r", "[", "]"}], ",", 
     RowBox[{"\[Theta]", "[", "]"}], ",", 
     RowBox[{"\[Phi]", "[", "]"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[30]:=",
 CellID->1938417081,ExpressionUUID->"8afcfaf9-12b8-4e06-9634-063904f81d6e"],

Cell[BoxData[{
 RowBox[{"DefConstantSymbol", "[", 
  RowBox[{"Mass", ",", 
   RowBox[{"PrintAs", "\[Rule]", "\"\<M\>\""}]}], "]"}], "\n", 
 RowBox[{"DefConstantSymbol", "[", 
  RowBox[{"Rotation", ",", 
   RowBox[{"PrintAs", "\[Rule]", "\"\<a\>\""}]}], "]"}], "\n", 
 RowBox[{"DefConstantSymbol", "[", 
  RowBox[{"Charge", ",", 
   RowBox[{"PrintAs", "\[Rule]", "\"\<Q\>\""}]}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"DefConstantSymbol", "[", 
  RowBox[{"CosmoC", ",", 
   RowBox[{"PrintAs", "\[Rule]", "\"\<\[CapitalLambda]\>\""}]}], 
  "]"}]}], "Input",
 CellLabel->"In[31]:=",
 CellID->1648955377,ExpressionUUID->"4509a127-ba1c-4c1d-ad42-a3e97ef50b43"],

Cell[CellGroupData[{

Cell["In Boyer\[Dash]Lindquist-like coordinates:", "MathCaption",
 CellID->716556704,ExpressionUUID->"d487aa7f-e3d8-49f9-91ab-bbef59c6a6c6"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[CapitalXi]", "=", 
   RowBox[{"1", "+", 
    RowBox[{"CosmoC", " ", 
     RowBox[{
      RowBox[{"Rotation", "^", "2"}], "/", "3"}]}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Deltar", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"1", "-", 
       RowBox[{"CosmoC", " ", 
        RowBox[{
         RowBox[{
          RowBox[{"r", "[", "]"}], "^", "2"}], "/", "3"}]}]}], ")"}], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"r", "[", "]"}], "^", "2"}], "+", 
       RowBox[{"Rotation", "^", "2"}]}], ")"}]}], "+", 
    RowBox[{"Charge", "^", "2"}], "-", 
    RowBox[{"2", "\[CapitalXi]", " ", "Mass", " ", 
     RowBox[{"r", "[", "]"}]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Delta\[Theta]", "=", 
   RowBox[{"1", "+", 
    RowBox[{"CosmoC", " ", 
     RowBox[{"Rotation", "^", "2"}], 
     RowBox[{
      RowBox[{
       RowBox[{"Cos", "[", 
        RowBox[{"\[Theta]", "[", "]"}], "]"}], "^", "2"}], "/", "3"}]}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Sigma", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"r", "[", "]"}], "^", "2"}], "+", 
    RowBox[{
     RowBox[{"Rotation", "^", "2"}], 
     RowBox[{
      RowBox[{"Cos", "[", 
       RowBox[{"\[Theta]", "[", "]"}], "]"}], "^", "2"}]}]}]}], 
  ";"}]}], "Input",
 CellLabel->"In[35]:=",
 CellID->80231752,ExpressionUUID->"e541e4d0-d285-4f1a-a4fe-bbdce46ae038"],

Cell[BoxData[
 RowBox[{
  RowBox[{"ds2", "=", 
   RowBox[{
    RowBox[{
     FractionBox["Sigma", "Deltar"], 
     RowBox[{"dr", "^", "2"}]}], "+", 
    RowBox[{
     FractionBox["Sigma", "Delta\[Theta]"], 
     RowBox[{"d\[Theta]", "^", "2"}]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{
       RowBox[{"Sin", "[", 
        RowBox[{"\[Theta]", "[", "]"}], "]"}], "^", "2"}], 
      RowBox[{
       RowBox[{"\[CapitalXi]", "^", "2"}], "Sigma"}]], "Delta\[Theta]", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"Rotation", " ", "dt"}], "-", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"r", "[", "]"}], "^", "2"}], "+", 
           RowBox[{"Rotation", "^", "2"}]}], ")"}], "d\[Phi]"}]}], ")"}], "^",
       "2"}]}], "-", 
    RowBox[{
     FractionBox["Deltar", 
      RowBox[{
       RowBox[{"\[CapitalXi]", "^", "2"}], "Sigma"}]], 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"dt", "-", 
        RowBox[{"Rotation", " ", 
         RowBox[{
          RowBox[{"Sin", "[", 
           RowBox[{"\[Theta]", "[", "]"}], "]"}], "^", "2"}], "d\[Phi]"}]}], 
       ")"}], "^", "2"}]}]}]}], ";"}]], "Input",
 CellLabel->"In[39]:=",
 CellID->1982809195,ExpressionUUID->"d599e52d-66d5-431c-8e23-16d056035a79"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"MatrixForm", "[", 
  RowBox[{"KNdSMatrix", "=", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"CoefficientArrays", "[", 
       RowBox[{"ds2", ",", 
        RowBox[{"{", 
         RowBox[{"dt", ",", "dr", ",", "d\[Theta]", ",", "d\[Phi]"}], "}"}], 
        ",", 
        RowBox[{"\"\<Symmetric\>\"", "\[Rule]", "True"}]}], "]"}], "[", 
      RowBox[{"[", "3", "]"}], "]"}], "//", "Normal"}], "//", "Simplify"}]}], 
  "]"}]], "Input",
 CellLabel->"In[40]:=",
 CellID->422906219,ExpressionUUID->"6465ba3a-a7b1-4a8f-a236-b409c7b39c42"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      FractionBox[
       RowBox[{"3", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "3"}], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["Q",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Charge,
             Editable->False], "2"]}], "-", 
          RowBox[{"3", " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"]}], "+", 
          RowBox[{"6", " ", 
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False], " ", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]}], "+", 
          RowBox[{"2", " ", 
           InterpretationBox[
            StyleBox["\[CapitalLambda]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`CosmoC,
            Editable->False], " ", 
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]}], "-", 
          RowBox[{"3", " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}], "+", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[CapitalLambda]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`CosmoC,
            Editable->False], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}], "+", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[CapitalLambda]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`CosmoC,
            Editable->False], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "4"]}], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           RowBox[{"(", 
            RowBox[{"3", "+", 
             RowBox[{
              InterpretationBox[
               StyleBox["\[CapitalLambda]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`CosmoC,
               Editable->False], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["a",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Rotation,
                Editable->False], "2"], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["Cos",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 Cos,
                 Editable->False], "[", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False], "]"}], "2"]}]}], ")"}], " ", 
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["Sin",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Sin,
              Editable->False], "[", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False], "]"}], "2"]}]}], ")"}]}], 
       RowBox[{
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"3", "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[CapitalLambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`CosmoC,
             Editable->False], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"]}]}], ")"}], "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["Cos",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Cos,
              Editable->False], "[", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False], "]"}], "2"]}], "+", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]}], ")"}]}]], "0", "0", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{"3", " ", 
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", "6"}], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["Q",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Charge,
              Editable->False], "2"]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[CapitalLambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`CosmoC,
             Editable->False], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "4"]}], "+", 
           RowBox[{"12", " ", 
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Mass,
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]}], "+", 
           RowBox[{"4", " ", 
            InterpretationBox[
             StyleBox["\[CapitalLambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`CosmoC,
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Mass,
             Editable->False], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], " ", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]}], "+", 
           RowBox[{"3", " ", 
            InterpretationBox[
             StyleBox["\[CapitalLambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`CosmoC,
             Editable->False], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"]}], "+", 
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["\[CapitalLambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`CosmoC,
             Editable->False], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "4"]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[CapitalLambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`CosmoC,
             Editable->False], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["Cos",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Cos,
              Editable->False], "[", 
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False]}], "]"}], " ", 
            RowBox[{"(", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["a",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Rotation,
                Editable->False], "2"], "+", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], "2"]}], ")"}]}]}], ")"}], " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Sin",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Sin,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}], 
        RowBox[{"2", " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"3", "+", 
            RowBox[{
             InterpretationBox[
              StyleBox["\[CapitalLambda]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`CosmoC,
              Editable->False], " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"]}]}], ")"}], "2"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], " ", 
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["Cos",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cos,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], "2"]}], "+", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}], ")"}]}]]}]},
     {"0", 
      FractionBox[
       RowBox[{
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Cos",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Cos,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}], "+", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]}], 
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["Q",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Charge,
          Editable->False], "2"], "-", 
        RowBox[{
         FractionBox["2", "3"], " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Mass,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{"3", "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[CapitalLambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`CosmoC,
             Editable->False], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"]}]}], ")"}], " ", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]}], "+", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], "+", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"1", "-", 
           FractionBox[
            RowBox[{
             InterpretationBox[
              StyleBox["\[CapitalLambda]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`CosmoC,
              Editable->False], " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}], "3"]}], ")"}]}]}]], "0", "0"},
     {"0", "0", 
      FractionBox[
       RowBox[{"3", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["Cos",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Cos,
              Editable->False], "[", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False], "]"}], "2"]}], "+", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]}], ")"}]}], 
       RowBox[{"3", "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["\[CapitalLambda]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`CosmoC,
          Editable->False], " ", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Cos",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Cos,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}]}]], "0"},
     {
      RowBox[{"-", 
       FractionBox[
        RowBox[{"3", " ", 
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", "6"}], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["Q",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Charge,
              Editable->False], "2"]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[CapitalLambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`CosmoC,
             Editable->False], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "4"]}], "+", 
           RowBox[{"12", " ", 
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Mass,
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]}], "+", 
           RowBox[{"4", " ", 
            InterpretationBox[
             StyleBox["\[CapitalLambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`CosmoC,
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Mass,
             Editable->False], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], " ", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]}], "+", 
           RowBox[{"3", " ", 
            InterpretationBox[
             StyleBox["\[CapitalLambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`CosmoC,
             Editable->False], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"]}], "+", 
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["\[CapitalLambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`CosmoC,
             Editable->False], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "4"]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[CapitalLambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`CosmoC,
             Editable->False], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["Cos",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Cos,
              Editable->False], "[", 
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False]}], "]"}], " ", 
            RowBox[{"(", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["a",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Rotation,
                Editable->False], "2"], "+", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], "2"]}], ")"}]}]}], ")"}], " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Sin",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Sin,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}], 
        RowBox[{"2", " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"3", "+", 
            RowBox[{
             InterpretationBox[
              StyleBox["\[CapitalLambda]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`CosmoC,
              Editable->False], " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"]}]}], ")"}], "2"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], " ", 
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["Cos",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cos,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], "2"]}], "+", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}], ")"}]}]]}], "0", "0", 
      FractionBox[
       RowBox[{"3", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "3"}], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["Q",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Charge,
             Editable->False], "2"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"]}], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"3", "+", 
             RowBox[{
              InterpretationBox[
               StyleBox["\[CapitalLambda]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`CosmoC,
               Editable->False], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["a",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Rotation,
                Editable->False], "2"]}]}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "4"], "+", 
             RowBox[{"2", " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], "4"]}], "+", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["a",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Rotation,
                Editable->False], "2"], " ", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False]}], "+", 
                RowBox[{"3", " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}]}], ")"}]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False]}], "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"3", " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["Q",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Charge,
                Editable->False], "2"]}], "+", 
             RowBox[{
              RowBox[{"(", 
               RowBox[{"3", "+", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["\[CapitalLambda]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`CosmoC,
                  Editable->False], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"]}]}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"], "-", 
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}], ")"}]}]}], ")"}]}]}], ")"}], " ", 
        
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Sin",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Sin,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"]}], 
       RowBox[{"2", " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"3", "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[CapitalLambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`CosmoC,
             Editable->False], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"]}]}], ")"}], "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["Cos",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Cos,
              Editable->False], "[", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False], "]"}], "2"]}], "+", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]}], ")"}]}]]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{3.728676783566958*^9},
 CellLabel->"Out[40]//MatrixForm=",
 CellID->98098026,ExpressionUUID->"41e8b6d4-3b4d-4769-9df6-b4211435f35c"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"met", "=", 
   RowBox[{"CTensor", "[", 
    RowBox[{"KNdSMatrix", ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "BL"}], ",", 
       RowBox[{"-", "BL"}]}], "}"}]}], "]"}]}], ";"}]], "Input",
 CellLabel->"In[41]:=",
 CellID->1372777775,ExpressionUUID->"6d84c684-fb5b-42a3-96e1-0dc497f28572"],

Cell[BoxData[
 RowBox[{
  RowBox[{"$CVSimplify", "=", "Together"}], ";"}]], "Input",
 CellLabel->"In[42]:=",
 CellID->1859244704,ExpressionUUID->"911feaa3-789a-450f-9ef0-663af62c0137"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SetCMetric", "[", 
   RowBox[{"met", ",", "BL", ",", 
    RowBox[{"SignatureOfMetric", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{"3", ",", "1", ",", "0"}], "}"}]}]}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[43]:=",
 CellID->291192631,ExpressionUUID->"2b9bd81a-be7d-462d-a1d4-cd9f29980464"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.890141`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.728676785142496*^9},
 CellLabel->"Out[43]=",
 CellID->1613565733,ExpressionUUID->"902e90f3-f156-490d-a971-c6c3bda4ebbb"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Compute Riemann tensor:", "MathCaption",
 CellID->1299934521,ExpressionUUID->"324fd376-8264-418b-941a-2e97e0ed5e69"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"MetricCompute", "[", 
   RowBox[{"met", ",", "BL", ",", 
    RowBox[{"\"\<Riemann\>\"", "[", 
     RowBox[{
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}], ",", "1"}], "]"}], ",", 
    RowBox[{"Verbose", "\[Rule]", "True"}]}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[44]:=",
 CellID->1648353516,ExpressionUUID->"c207b4b3-cdc9-4743-91a1-1b4411efbc4d"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Metric\"\>", "[", 
    RowBox[{"1", ",", "1"}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Metric"[1, 1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676785616259*^9},
 CellLabel->"During evaluation of In[44]:=",
 CellID->775780215,ExpressionUUID->"835e4f58-9813-4249-ba82-e4f20be89fbf"],

Cell[BoxData["\<\"    Already cached\"\>"], "Print",
 CellChangeTimes->{3.728676785639469*^9},
 CellLabel->"During evaluation of In[44]:=",
 CellID->1164852568,ExpressionUUID->"ff82a3fa-7396-4a8d-8133-241a3b234e8a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"DMetric\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "DMetric"[-1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286767856648617`*^9},
 CellLabel->"During evaluation of In[44]:=",
 CellID->1216184324,ExpressionUUID->"ddcdcbe8-ee1d-449a-a00c-acaec8eb0059"],

Cell[BoxData["\<\"    Already cached\"\>"], "Print",
 CellChangeTimes->{3.728676785684205*^9},
 CellLabel->"During evaluation of In[44]:=",
 CellID->660581322,ExpressionUUID->"1a4998d8-4506-4477-b5fe-d2c7f4a46331"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"DDMetric\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "DDMetric"[-1, -1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676785703679*^9},
 CellLabel->"During evaluation of In[44]:=",
 CellID->2047307511,ExpressionUUID->"3333447c-40c3-4ede-b1a4-de48fd407784"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Constructed in \"\>", "\[InvisibleSpace]", "0.0065`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Constructed in ", 0.0065, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676785723742*^9},
 CellLabel->"During evaluation of In[44]:=",
 CellID->2114345143,ExpressionUUID->"44330806-7c00-4c50-982b-b6c4dc5390c9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Applied \"\>", "\[InvisibleSpace]", "Together", 
   "\[InvisibleSpace]", "\<\" in \"\>", "\[InvisibleSpace]", "0.091191`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Applied ", Together, " in ", 0.091191, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286767857443867`*^9},
 CellLabel->"During evaluation of In[44]:=",
 CellID->1386911492,ExpressionUUID->"44ea9afc-22cb-4298-a481-4bb39d3a04e2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Christoffel\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Christoffel"[-1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286767857651052`*^9},
 CellLabel->"During evaluation of In[44]:=",
 CellID->1033941506,ExpressionUUID->"c68c512f-9f74-4f1b-8c24-01ebae4607ee"],

Cell[BoxData["\<\"    Already cached\"\>"], "Print",
 CellChangeTimes->{3.728676785786704*^9},
 CellLabel->"During evaluation of In[44]:=",
 CellID->17219818,ExpressionUUID->"3203bd73-1f69-4e5e-bdff-753bd6244108"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Christoffel\"\>", "[", 
    RowBox[{"1", ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Christoffel"[1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676785805435*^9},
 CellLabel->"During evaluation of In[44]:=",
 CellID->1650672428,ExpressionUUID->"52229f00-c2fd-42e1-953d-3156240aecc5"],

Cell[BoxData["\<\"    Already cached\"\>"], "Print",
 CellChangeTimes->{3.728676785824046*^9},
 CellLabel->"During evaluation of In[44]:=",
 CellID->1657745450,ExpressionUUID->"29da8afa-bef7-4213-b5e1-cf8463a2eb41"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", 
   "\[InvisibleSpace]", "\<\"CovDOfMetric\"\>"}],
  SequenceForm["** ReportCompute: ", "CovDOfMetric"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676785843587*^9},
 CellLabel->"During evaluation of In[44]:=",
 CellID->274941972,ExpressionUUID->"d3e9b296-41bf-4a4a-b566-d3c999f96453"],

Cell[BoxData["\<\"    Already cached\"\>"], "Print",
 CellChangeTimes->{3.728676785863058*^9},
 CellLabel->"During evaluation of In[44]:=",
 CellID->1712398707,ExpressionUUID->"8263e620-7ee4-4450-802b-c093693e94c2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Riemann\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Riemann"[-1, -1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676785882868*^9},
 CellLabel->"During evaluation of In[44]:=",
 CellID->1856715102,ExpressionUUID->"2b06e44b-1cd9-43da-b867-59c16874a299"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Constructed in \"\>", "\[InvisibleSpace]", "0.004297`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Constructed in ", 0.004297, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676785902348*^9},
 CellLabel->"During evaluation of In[44]:=",
 CellID->725844757,ExpressionUUID->"07f8618c-b29a-475e-9c8b-e7e04ede0c1b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Applied \"\>", "\[InvisibleSpace]", "Together", 
   "\[InvisibleSpace]", "\<\" in \"\>", "\[InvisibleSpace]", "4.295331`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Applied ", Together, " in ", 4.295331, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286767899630537`*^9},
 CellLabel->"During evaluation of In[44]:=",
 CellID->1565715939,ExpressionUUID->"559a4ba5-31e1-4ff7-a1fc-0f4747df76c1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Stored in \"\>", "\[InvisibleSpace]", "0.819687`", 
   "\[InvisibleSpace]", "\<\" seconds and \"\>", "\[InvisibleSpace]", 
   "177578632", "\[InvisibleSpace]", "\<\" bytes\"\>"}],
  SequenceForm[
  "    Stored in ", 0.819687, " seconds and ", 177578632, " bytes"],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286767912878017`*^9},
 CellLabel->"During evaluation of In[44]:=",
 CellID->1870660082,ExpressionUUID->"8d1726b4-ff60-4e3d-8d0d-c4e88f2e14d3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Riemann\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", "1"}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Riemann"[-1, -1, -1, 1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7286767913045607`*^9},
 CellLabel->"During evaluation of In[44]:=",
 CellID->1023406094,ExpressionUUID->"54fc00f8-d173-4558-9803-db38b6cc9e11"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Constructed in \"\>", "\[InvisibleSpace]", "0.112902`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Constructed in ", 0.112902, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676791455605*^9},
 CellLabel->"During evaluation of In[44]:=",
 CellID->1189693035,ExpressionUUID->"3e5fd03b-a991-4853-9c9d-c876d299e793"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Applied \"\>", "\[InvisibleSpace]", "Together", 
   "\[InvisibleSpace]", "\<\" in \"\>", "\[InvisibleSpace]", "23.287835`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Applied ", Together, " in ", 23.287835, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.72867681468386*^9},
 CellLabel->"During evaluation of In[44]:=",
 CellID->2098085675,ExpressionUUID->"b0eb6b10-659b-4d8e-8a47-e23cbb499d97"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Stored in \"\>", "\[InvisibleSpace]", "1.664889`", 
   "\[InvisibleSpace]", "\<\" seconds and \"\>", "\[InvisibleSpace]", 
   "341347392", "\[InvisibleSpace]", "\<\" bytes\"\>"}],
  SequenceForm[
  "    Stored in ", 1.664889, " seconds and ", 341347392, " bytes"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728676818751236*^9},
 CellLabel->"During evaluation of In[44]:=",
 CellID->844455985,ExpressionUUID->"70f5956b-c00d-48b2-ae6c-1ef160b4c9f8"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"33.208696`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.728676818768523*^9},
 CellLabel->"Out[44]=",
 CellID->52316872,ExpressionUUID->"46100676-3fd1-47e9-b0cd-4ed2587e0ed1"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"CD", "=", 
   RowBox[{"CovDOfMetric", "[", "met", "]"}]}], ";"}]], "Input",
 CellLabel->"In[45]:=",
 CellID->203862004,ExpressionUUID->"6b8bb9fa-99f4-4f24-9f04-b5d2275fddc9"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ByteCount", "[", 
  RowBox[{"Riemann", "[", "CD", "]"}], "]"}]], "Input",
 CellLabel->"In[46]:=",
 CellID->1072284263,ExpressionUUID->"0e804627-b88b-4fb1-a065-9ec70adb70f0"],

Cell[BoxData["341351568"], "Output",
 CellChangeTimes->{3.728676819284031*^9},
 CellLabel->"Out[46]=",
 CellID->968856081,ExpressionUUID->"97aada0b-255b-4e55-a09b-d3e1bcc231f0"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"$CVSimplify", "=", "Identity"}], ";"}]], "Input",
 CellLabel->"In[47]:=",
 CellID->463309187,ExpressionUUID->"81f0af1d-5eab-4b3a-861f-7170d3a9e20b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"ricci", "=", 
    RowBox[{"Ricci", "[", "CD", "]"}]}], ";"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[48]:=",
 CellID->61801887,ExpressionUUID->"ee993d98-67d4-4c4b-b415-140efed4ede1"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.919248`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.728676820376053*^9},
 CellLabel->"Out[48]=",
 CellID->1343141813,ExpressionUUID->"94ea17dd-e1dc-4a14-9651-199b30e6afd3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Map", "[", 
   RowBox[{"ByteCount", ",", 
    RowBox[{"ricci", "[", 
     RowBox[{"[", "1", "]"}], "]"}], ",", 
    RowBox[{"{", "2", "}"}]}], "]"}], "//", "MatrixForm"}]], "Input",
 CellLabel->"In[49]:=",
 CellID->1114659762,ExpressionUUID->"453df55f-aa90-4e8a-b654-5c0798cf7ce2"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {"5950696", "16", "16", "8132840"},
     {"16", "23938592", "13212304", "16"},
     {"16", "13203752", "33457720", "16"},
     {"8132960", "16", "16", "11835624"}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{3.728676820652398*^9},
 CellLabel->"Out[49]//MatrixForm=",
 CellID->1983837474,ExpressionUUID->"4f38b2af-1567-4a45-9ece-f53fb94be573"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"ricci", "=", 
    RowBox[{"Together", "[", "ricci", "]"}]}], ";"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[50]:=",
 CellID->1343612025,ExpressionUUID->"88b38de8-404d-4fc2-87ca-58bd447d2978"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"13.620722`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.728676834412208*^9},
 CellLabel->"Out[50]=",
 CellID->525322196,ExpressionUUID->"553c96cb-fbc5-46af-8477-cce4e6c0108f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Map", "[", 
   RowBox[{"ByteCount", ",", 
    RowBox[{"ricci", "[", 
     RowBox[{"[", "1", "]"}], "]"}], ",", 
    RowBox[{"{", "2", "}"}]}], "]"}], "//", "MatrixForm"}]], "Input",
 CellLabel->"In[51]:=",
 CellID->1913538034,ExpressionUUID->"570c0c58-e8b1-4b9d-acad-7e629dd7eec5"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {"3560176", "16", "16", "4646296"},
     {"16", "13661528", "8069952", "16"},
     {"16", "8069952", "19769624", "16"},
     {"4646296", "16", "16", "7050752"}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{3.728676834797142*^9},
 CellLabel->"Out[51]//MatrixForm=",
 CellID->784875355,ExpressionUUID->"6b69fd9b-ff74-46c4-8dbd-4c254fd47d40"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We separate components to see which ones are harder. We see that we are \
missing the terms proportional to the charge:\
\>", "MathCaption",
 CellID->1117084146,ExpressionUUID->"d2b3b6dc-ac60-4410-b7b0-8597903914e2"],

Cell[BoxData[
 RowBox[{
  RowBox[{"diff", "=", 
   RowBox[{"ricci", "-", 
    RowBox[{"CosmoC", " ", "met"}]}]}], ";"}]], "Input",
 CellLabel->"In[52]:=",
 CellID->1531199748,ExpressionUUID->"d1a5e2cc-bc54-43cf-b065-0103df98093a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"diff", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"0", ",", 
       RowBox[{"-", "BL"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", 
       RowBox[{"-", "BL"}]}], "}"}]}], "]"}], "//", "Simplify"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[53]:=",
 CellID->1577622635,ExpressionUUID->"dd2bea58-8aff-4a58-afca-dada75c0585e"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"4.811115`", ",", 
   FractionBox[
    RowBox[{"3", " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["Q",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Charge,
       Editable->False], "2"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"24", " ", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["Q",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Charge,
          Editable->False], "2"]}], "+", 
       RowBox[{"36", " ", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], "2"]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[CapitalLambda]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`CosmoC,
         Editable->False], " ", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], "4"]}], "-", 
       RowBox[{"12", " ", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], "2"], " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["Cos",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Cos,
          Editable->False], "[", 
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False]}], "]"}]}], "-", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[CapitalLambda]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`CosmoC,
         Editable->False], " ", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], "4"], " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["Cos",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Cos,
          Editable->False], "[", 
         RowBox[{"4", " ", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False]}], "]"}]}], "-", 
       RowBox[{"48", " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Mass,
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]}], "-", 
       RowBox[{"16", " ", 
        InterpretationBox[
         StyleBox["\[CapitalLambda]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`CosmoC,
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Mass,
         Editable->False], " ", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], "2"], " ", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]}], "+", 
       RowBox[{"24", " ", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]}], "-", 
       RowBox[{"8", " ", 
        InterpretationBox[
         StyleBox["\[CapitalLambda]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`CosmoC,
         Editable->False], " ", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], "2"], " ", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]}], "-", 
       RowBox[{"8", " ", 
        InterpretationBox[
         StyleBox["\[CapitalLambda]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`CosmoC,
         Editable->False], " ", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "4"]}]}], ")"}]}], 
    RowBox[{
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"3", "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["\[CapitalLambda]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`CosmoC,
          Editable->False], " ", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"]}]}], ")"}], "2"], " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], "2"], "+", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], " ", 
         RowBox[{
          InterpretationBox[
           StyleBox["Cos",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Cos,
           Editable->False], "[", 
          RowBox[{"2", " ", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False]}], "]"}]}], "+", 
        RowBox[{"2", " ", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"]}]}], ")"}], "3"]}]]}], "}"}]], "Output",
 CellChangeTimes->{3.728676839786233*^9},
 CellLabel->"Out[53]=",
 CellID->1555531611,ExpressionUUID->"ed5be0ca-426f-47d5-88a7-20ff517a7a0f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"diff", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"0", ",", 
       RowBox[{"-", "BL"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "BL"}]}], "}"}]}], "]"}], "//", "Simplify"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[54]:=",
 CellID->2002514938,ExpressionUUID->"615dd8e2-ddb2-49ee-b410-119a1628b8e8"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.096737`", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.728676840090295*^9},
 CellLabel->"Out[54]=",
 CellID->353867499,ExpressionUUID->"3239f048-0abc-4454-82b0-972ad38d908e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"diff", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"0", ",", 
       RowBox[{"-", "BL"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"2", ",", 
       RowBox[{"-", "BL"}]}], "}"}]}], "]"}], "//", "Simplify"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[55]:=",
 CellID->1434883467,ExpressionUUID->"8d18d307-07b4-4310-9e29-1efe41fadf66"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.092227`", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.7286768403368607`*^9},
 CellLabel->"Out[55]=",
 CellID->1068837150,ExpressionUUID->"1cce3183-4040-44a8-9e4e-895d8f052006"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"diff", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"0", ",", 
       RowBox[{"-", "BL"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"3", ",", 
       RowBox[{"-", "BL"}]}], "}"}]}], "]"}], "//", "Simplify"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[56]:=",
 CellID->1057373424,ExpressionUUID->"d1de3a38-e4aa-47ba-9575-b027ce849734"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"7.954464`", ",", 
   RowBox[{"-", 
    FractionBox[
     RowBox[{"12", " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["Q",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Charge,
        Editable->False], "2"], " ", 
      InterpretationBox[
       StyleBox["a",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Rotation,
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"6", " ", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["Q",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Charge,
           Editable->False], "2"]}], "+", 
        RowBox[{"12", " ", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["\[CapitalLambda]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`CosmoC,
          Editable->False], " ", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "4"]}], "-", 
        RowBox[{"12", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Mass,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]}], "-", 
        RowBox[{"4", " ", 
         InterpretationBox[
          StyleBox["\[CapitalLambda]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`CosmoC,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Mass,
          Editable->False], " ", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], " ", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]}], "+", 
        RowBox[{"12", " ", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["\[CapitalLambda]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`CosmoC,
          Editable->False], " ", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], " ", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"]}], "-", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[CapitalLambda]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`CosmoC,
          Editable->False], " ", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "4"]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["\[CapitalLambda]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`CosmoC,
          Editable->False], " ", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], " ", 
         RowBox[{
          InterpretationBox[
           StyleBox["Cos",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Cos,
           Editable->False], "[", 
          RowBox[{"2", " ", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False]}], "]"}], " ", 
         RowBox[{"(", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], "+", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}], ")"}]}]}], ")"}], " ", 
      SuperscriptBox[
       RowBox[{
        InterpretationBox[
         StyleBox["Sin",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Sin,
         Editable->False], "[", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False], "]"}], "2"]}], 
     RowBox[{
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"3", "+", 
         RowBox[{
          InterpretationBox[
           StyleBox["\[CapitalLambda]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`CosmoC,
           Editable->False], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"]}]}], ")"}], "2"], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], "+", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], " ", 
          RowBox[{
           InterpretationBox[
            StyleBox["Cos",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Cos,
            Editable->False], "[", 
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False]}], "]"}]}], "+", 
         RowBox[{"2", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]}]}], ")"}], "3"]}]]}]}], "}"}]], "Output",
 CellChangeTimes->{3.7286768484532413`*^9},
 CellLabel->"Out[56]=",
 CellID->245272895,ExpressionUUID->"060308be-880b-4151-8987-b6416dfec00e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"diff", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "BL"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "BL"}]}], "}"}]}], "]"}], "//", "Simplify"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[57]:=",
 CellID->932708609,ExpressionUUID->"9f726658-434d-4309-9ae5-0a9f6b0841d6"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"16.221321`", ",", 
   FractionBox[
    RowBox[{"6", " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["Q",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Charge,
       Editable->False], "2"]}], 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], "2"], "+", 
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], "2"], " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["Cos",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Cos,
          Editable->False], "[", 
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False]}], "]"}]}], "+", 
       RowBox[{"2", " ", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]}]}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "3"}], " ", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["Q",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Charge,
          Editable->False], "2"]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"6", " ", 
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False]}], "-", 
          RowBox[{"3", " ", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]}], "+", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[CapitalLambda]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`CosmoC,
            Editable->False], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "3"]}]}], ")"}]}], "+", 
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "3"}], "+", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[CapitalLambda]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`CosmoC,
            Editable->False], " ", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False]}], "+", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]}], ")"}]}]}], ")"}]}]}], ")"}]}]]}], 
  "}"}]], "Output",
 CellChangeTimes->{3.7286768648186893`*^9},
 CellLabel->"Out[57]=",
 CellID->1280427672,ExpressionUUID->"3bdd1b02-8137-4986-9b05-97e54eb7e28e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"diff", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "BL"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"2", ",", 
       RowBox[{"-", "BL"}]}], "}"}]}], "]"}], "//", "Simplify"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[58]:=",
 CellID->880359372,ExpressionUUID->"22598816-9e19-4982-9965-2bf5f223de14"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"24.599346`", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.728676889564389*^9},
 CellLabel->"Out[58]=",
 CellID->214256717,ExpressionUUID->"998db42f-e181-4c56-b825-e62f4f07a0fc"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"diff", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "BL"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"3", ",", 
       RowBox[{"-", "BL"}]}], "}"}]}], "]"}], "//", "Simplify"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[59]:=",
 CellID->1211152600,ExpressionUUID->"668ff03e-f79b-47f4-afae-63f50de3dd5d"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.097953`", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.728676889794855*^9},
 CellLabel->"Out[59]=",
 CellID->1830050247,ExpressionUUID->"8e11b76b-e669-482c-9552-b45328623fae"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"diff", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"2", ",", 
       RowBox[{"-", "BL"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"2", ",", 
       RowBox[{"-", "BL"}]}], "}"}]}], "]"}], "//", "Simplify"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[60]:=",
 CellID->231595302,ExpressionUUID->"0d79a4ba-8f07-4610-8e13-affd8fd401bf"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"122.222311`", ",", 
   FractionBox[
    RowBox[{"12", " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["Q",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Charge,
       Editable->False], "2"]}], 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"6", "+", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[CapitalLambda]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`CosmoC,
         Editable->False], " ", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], "2"]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[CapitalLambda]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`CosmoC,
         Editable->False], " ", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], "2"], " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["Cos",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Cos,
          Editable->False], "[", 
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False]}], "]"}]}]}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], "2"], "+", 
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], "2"], " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["Cos",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Cos,
          Editable->False], "[", 
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False]}], "]"}]}], "+", 
       RowBox[{"2", " ", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]}]}], ")"}]}]]}], "}"}]], "Output",
 CellChangeTimes->{3.7286770121535*^9},
 CellLabel->"Out[60]=",
 CellID->617410896,ExpressionUUID->"ed9c6fdc-d0b2-4445-a50f-fa7723756926"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"diff", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"2", ",", 
       RowBox[{"-", "BL"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"3", ",", 
       RowBox[{"-", "BL"}]}], "}"}]}], "]"}], "//", "Simplify"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[61]:=",
 CellID->1652335347,ExpressionUUID->"dfe632b7-6d62-4d53-a58e-14839d82a1dc"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.108582`", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.728677012386661*^9},
 CellLabel->"Out[61]=",
 CellID->183946768,ExpressionUUID->"e58704a8-6aa2-41fd-bdfe-f5356eedde0f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"diff", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"3", ",", 
       RowBox[{"-", "BL"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"3", ",", 
       RowBox[{"-", "BL"}]}], "}"}]}], "]"}], "//", "Simplify"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[62]:=",
 CellID->1750523611,ExpressionUUID->"e5c09111-7fc8-4abf-824c-1f22fec776c9"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"12.503307`", ",", 
   RowBox[{"-", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"12", " ", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["Q",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Charge,
          Editable->False], "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "3"}], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["Q",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Charge,
             Editable->False], "2"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"]}], "-", 
          RowBox[{"9", " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "4"]}], "-", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[CapitalLambda]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`CosmoC,
            Editable->False], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "6"]}], "+", 
          RowBox[{"6", " ", 
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]}], "+", 
          RowBox[{"2", " ", 
           InterpretationBox[
            StyleBox["\[CapitalLambda]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`CosmoC,
            Editable->False], " ", 
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "4"], " ", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]}], "-", 
          RowBox[{"15", " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}], "-", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[CapitalLambda]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`CosmoC,
            Editable->False], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "4"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}], "-", 
          RowBox[{"6", " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "4"]}], "-", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False]}], "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "3"}], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["Q",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Charge,
                Editable->False], "2"]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox["\[CapitalLambda]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`CosmoC,
               Editable->False], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["a",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Rotation,
                Editable->False], "4"]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"6", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False]}], "-", 
                RowBox[{"3", " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], "+", 
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["\[CapitalLambda]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`CosmoC,
                  Editable->False], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "3"]}]}], ")"}]}], "+", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["a",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Rotation,
                Editable->False], "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "3"}], "+", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["\[CapitalLambda]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`CosmoC,
                  Editable->False], " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                   RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}]}], ")"}]}]}], ")"}]}]}], ")"}]}]}], 
         ")"}], " ", 
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Sin",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Sin,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"]}], ")"}], "/", 
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"3", "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[CapitalLambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`CosmoC,
             Editable->False], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"]}]}], ")"}], "2"], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], "+", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["Cos",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Cos,
              Editable->False], "[", 
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False]}], "]"}]}], "+", 
           RowBox[{"2", " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"]}]}], ")"}], "3"]}], ")"}]}], ")"}]}]}], 
  "}"}]], "Output",
 CellChangeTimes->{3.72867702502107*^9},
 CellLabel->"Out[62]=",
 CellID->1364361743,ExpressionUUID->"c620b273-0f03-43bb-8d78-81e06af29e7a"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{
  "met", ",", "\[CapitalXi]", ",", "ricci", ",", "diff", ",", "ds2", ",", 
   "dt", ",", "dr", ",", "d\[Theta]", ",", "d\[Phi]", ",", "Deltar", ",", 
   "Delta\[Theta]", ",", "Sigma", ",", "KNdSMatrix"}], "]"}]], "Input",
 CellLabel->"In[63]:=",
 CellID->1896933198,ExpressionUUID->"1a72008c-69df-4e0c-8da5-a9be8d22925e"],

Cell[BoxData[
 RowBox[{"UndefConstantSymbol", "[", 
  RowBox[{"{", 
   RowBox[{"Mass", ",", "Rotation", ",", "Charge", ",", "CosmoC"}], "}"}], 
  "]"}]], "Input",
 CellLabel->"In[64]:=",
 CellID->214035424,ExpressionUUID->"d4f75154-58b2-442b-948f-532bbe9033fb"],

Cell[BoxData[
 RowBox[{"UndefChart", "[", "BL", "]"}]], "Input",
 CellLabel->"In[65]:=",
 CellID->258262443,ExpressionUUID->"3fdb6381-b5b5-4859-9a15-3c9573b21560"],

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M", "]"}]], "Input",
 CellLabel->"In[66]:=",
 CellID->552428914,ExpressionUUID->"6ad213a1-9331-44be-88cb-77c982ee09e8"],

Cell[BoxData[
 RowBox[{"ClearxCobaCache", "[", "All", "]"}]], "Input",
 CellLabel->"In[67]:=",
 CellID->1100951617,ExpressionUUID->"b8886a2a-8b57-4aca-bd35-d01211ab1bcf"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
3.5.6. A very (actually too) hard example: The black saturn metric\
\>", "Subsubsection",
 CellChangeTimes->{{3.7286769426988297`*^9, 3.728676943034114*^9}, 
   3.790876755047715*^9},
 CellID->974724367,ExpressionUUID->"b2e7d809-e79b-40eb-b5c7-67be3637aff2"],

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<xAct`xCoba`\>\"", "]"}]], "Input",
 CellLabel->"In[68]:=",
 CellID->1187176853,ExpressionUUID->"de286e65-fab5-43ce-968c-ead467d81ebc"],

Cell[BoxData[
 RowBox[{
  RowBox[{"$DefInfoQ", "=", 
   RowBox[{"$UndefInfoQ", "=", "False"}]}], ";"}]], "Input",
 CellLabel->"In[69]:=",
 CellID->1392415406,ExpressionUUID->"96c5f748-1e1a-45f3-b101-b8f49f646f92"],

Cell[CellGroupData[{

Cell[TextData[{
 "Solution taken from [",
 ButtonBox["arXiv:0701035v2",
  BaseStyle->"Hyperlink",
  ButtonData->{
    URL["http://arxiv.org/pdf/hep-th/0701035v2.pdf"], None},
  ButtonNote->"http://arxiv.org/pdf/hep-th/0701035v2.pdf"],
 "]. Thanks to Alfonso Garc\[IAcute]a-Parrado for the actual ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " formulas."
}], "MathCaption",
 CellID->1196427283,ExpressionUUID->"c6b5777d-d64a-4813-813a-6f74e12eff27"],

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M", ",", "5", ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f"}], "}"}]}], 
  "]"}]], "Input",
 CellLabel->"In[70]:=",
 CellID->1463958520,ExpressionUUID->"266015e3-b74d-46d0-950a-4b142c0021bd"],

Cell[BoxData[
 RowBox[{"DefChart", "[", 
  RowBox[{"BS", ",", "M", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2", ",", "3", ",", "4"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"t", "[", "]"}], ",", 
     RowBox[{"\[Rho]", "[", "]"}], ",", 
     RowBox[{"z", "[", "]"}], ",", 
     RowBox[{"\[Psi]", "[", "]"}], ",", 
     RowBox[{"\[Phi]", "[", "]"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[71]:=",
 CellID->1030701961,ExpressionUUID->"33098eef-56b6-45d1-8585-6ca0a8f5a5ef"],

Cell[BoxData[
 RowBox[{"DefConstantSymbol", "[", 
  RowBox[{"{", 
   RowBox[{
   "ks", ",", "q", ",", "a1", ",", "a2", ",", "a3", ",", "a4", ",", "a5", ",",
     "c1", ",", "c2"}], "}"}], "]"}]], "Input",
 CellLabel->"In[72]:=",
 CellID->1404761825,ExpressionUUID->"3167c475-8992-4e7a-aa9c-464ece0630e0"],

Cell[BoxData[
 RowBox[{"DefScalarFunction", "[", 
  RowBox[{"{", 
   RowBox[{
   "\[Mu]1", ",", "\[Mu]2", ",", "\[Mu]3", ",", "\[Mu]4", ",", "\[Mu]5", ",", 
    "M0", ",", "M1", ",", "M2", ",", "M3", ",", "M4", ",", "R1", ",", "R2", 
    ",", "Hx", ",", "Hy", ",", "Gx", ",", "Gy", ",", "\[Omega]\[Psi]", ",", 
    "P", ",", "F"}], "}"}], "]"}]], "Input",
 CellLabel->"In[73]:=",
 CellID->396777657,ExpressionUUID->"4962a93d-4458-4ae0-9de5-2250b5cd75b7"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
\[Mu] and R functions do not depend on any other functions. These functions \
are the only ones with explicit dependence in z:\
\>", "MathCaption",
 CellID->705357264,ExpressionUUID->"fce8ae7b-2aeb-4f91-8181-b723b829b173"],

Cell[BoxData[
 RowBox[{
  RowBox[{"\[Mu]Functions", "=", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"\[Mu]1", "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{"\[Rho]", ",", "z"}], "}"}], ",", 
        RowBox[{
         RowBox[{"Sqrt", "[", 
          RowBox[{
           RowBox[{"\[Rho]", "^", "2"}], "+", 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"z", "-", "a1"}], ")"}], "^", "2"}]}], "]"}], "-", 
         RowBox[{"(", 
          RowBox[{"z", "-", "a1"}], ")"}]}]}], "]"}]}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"\[Mu]2", "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{"\[Rho]", ",", "z"}], "}"}], ",", 
        RowBox[{
         RowBox[{"Sqrt", "[", 
          RowBox[{
           RowBox[{"\[Rho]", "^", "2"}], "+", 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"z", "-", "a2"}], ")"}], "^", "2"}]}], "]"}], "-", 
         RowBox[{"(", 
          RowBox[{"z", "-", "a2"}], ")"}]}]}], "]"}]}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"\[Mu]3", "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{"\[Rho]", ",", "z"}], "}"}], ",", 
        RowBox[{
         RowBox[{"Sqrt", "[", 
          RowBox[{
           RowBox[{"\[Rho]", "^", "2"}], "+", 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"z", "-", "a3"}], ")"}], "^", "2"}]}], "]"}], "-", 
         RowBox[{"(", 
          RowBox[{"z", "-", "a3"}], ")"}]}]}], "]"}]}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"\[Mu]4", "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{"\[Rho]", ",", "z"}], "}"}], ",", 
        RowBox[{
         RowBox[{"Sqrt", "[", 
          RowBox[{
           RowBox[{"\[Rho]", "^", "2"}], "+", 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"z", "-", "a4"}], ")"}], "^", "2"}]}], "]"}], "-", 
         RowBox[{"(", 
          RowBox[{"z", "-", "a4"}], ")"}]}]}], "]"}]}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"\[Mu]5", "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{"\[Rho]", ",", "z"}], "}"}], ",", 
        RowBox[{
         RowBox[{"Sqrt", "[", 
          RowBox[{
           RowBox[{"\[Rho]", "^", "2"}], "+", 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"z", "-", "a5"}], ")"}], "^", "2"}]}], "]"}], "-", 
         RowBox[{"(", 
          RowBox[{"z", "-", "a5"}], ")"}]}]}], "]"}]}]}], 
    "\[IndentingNewLine]", "}"}]}], ";"}]], "Input",
 CellLabel->"In[74]:=",
 CellID->829804608,ExpressionUUID->"eebe3900-afe0-4e53-8286-eeea3135ccdf"],

Cell[BoxData[
 RowBox[{
  RowBox[{"RFunctions", "=", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"R1", "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{"\[Rho]", ",", "z"}], "}"}], ",", 
        RowBox[{"Sqrt", "[", 
         RowBox[{
          RowBox[{"\[Rho]", "^", "2"}], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"z", "-", "a1"}], ")"}], "^", "2"}]}], "]"}]}], "]"}]}], 
     ",", "\[IndentingNewLine]", 
     RowBox[{"R2", "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{"\[Rho]", ",", "z"}], "}"}], ",", 
        RowBox[{"Sqrt", "[", 
         RowBox[{
          RowBox[{"\[Rho]", "^", "2"}], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"z", "-", "a2"}], ")"}], "^", "2"}]}], "]"}]}], "]"}]}]}],
     "\[IndentingNewLine]", "}"}]}], ";"}]], "Input",
 CellLabel->"In[75]:=",
 CellID->767243038,ExpressionUUID->"eadafc14-b002-4c4f-b506-e839b9eb2021"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The G, P, F and M functions are given in terms of \[Mu] functions:\
\>", "MathCaption",
 CellID->1561923078,ExpressionUUID->"78c63d67-c387-47bb-8f28-36872f52e9fd"],

Cell[BoxData[
 RowBox[{
  RowBox[{"GFunctions", "=", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"Gx", "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{"\[Rho]", ",", "z"}], "}"}], ",", 
        FractionBox[
         RowBox[{
          RowBox[{"\[Rho]", "^", "2"}], 
          RowBox[{"\[Mu]4", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}]}], 
         RowBox[{
          RowBox[{"\[Mu]3", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}], " ", 
          RowBox[{"\[Mu]5", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}]}]]}], "]"}]}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"Gy", "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{"\[Rho]", ",", "z"}], "}"}], ",", 
        FractionBox[
         RowBox[{
          RowBox[{"\[Mu]3", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}], 
          RowBox[{"\[Mu]5", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}]}], 
         RowBox[{"\[Mu]4", "[", 
          RowBox[{"\[Rho]", ",", "z"}], "]"}]]}], "]"}]}]}], 
    "\[IndentingNewLine]", "}"}]}], ";"}]], "Input",
 CellLabel->"In[76]:=",
 CellID->401071018,ExpressionUUID->"0ee4cd7e-0823-40ff-b1cf-01bf1ae77b37"],

Cell[BoxData[
 RowBox[{
  RowBox[{"PFunction", "=", 
   RowBox[{"P", "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"\[Rho]", ",", "z"}], "}"}], ",", 
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"\[Mu]3", "[", 
            RowBox[{"\[Rho]", ",", "z"}], "]"}], 
           RowBox[{"\[Mu]4", "[", 
            RowBox[{"\[Rho]", ",", "z"}], "]"}]}], "+", 
          RowBox[{"\[Rho]", "^", "2"}]}], ")"}], "^", "2"}], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"\[Mu]1", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}], 
          RowBox[{"\[Mu]5", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}]}], "+", 
         RowBox[{"\[Rho]", "^", "2"}]}], ")"}], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"\[Mu]4", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}], 
          RowBox[{"\[Mu]5", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}]}], "+", 
         RowBox[{"\[Rho]", "^", "2"}]}], ")"}]}]}], "]"}]}]}], ";"}]], "Input",\

 CellLabel->"In[77]:=",
 CellID->1996915240,ExpressionUUID->"73eb0a69-aae4-412d-8fbd-46f3ac7c6bf9"],

Cell[BoxData[
 RowBox[{
  RowBox[{"FFunction", "=", 
   RowBox[{"F", "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"\[Rho]", ",", "z"}], "}"}], ",", 
      RowBox[{
       RowBox[{"\[Mu]1", "[", 
        RowBox[{"\[Rho]", ",", "z"}], "]"}], 
       RowBox[{"\[Mu]5", "[", 
        RowBox[{"\[Rho]", ",", "z"}], "]"}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"\[Mu]1", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}], "-", 
          RowBox[{"\[Mu]3", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}]}], ")"}], "^", "2"}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"\[Mu]2", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}], "-", 
          RowBox[{"\[Mu]4", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}]}], ")"}], "^", "2"}], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"\[Rho]", "^", "2"}], "+", 
         RowBox[{
          RowBox[{"\[Mu]1", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}], 
          RowBox[{"\[Mu]3", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}]}]}], ")"}], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"\[Rho]", "^", "2"}], "+", 
         RowBox[{
          RowBox[{"\[Mu]2", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}], 
          RowBox[{"\[Mu]3", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}]}]}], ")"}], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"\[Rho]", "^", "2"}], "+", 
         RowBox[{
          RowBox[{"\[Mu]1", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}], 
          RowBox[{"\[Mu]4", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}]}]}], ")"}], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"\[Rho]", "^", "2"}], "+", 
         RowBox[{
          RowBox[{"\[Mu]2", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}], 
          RowBox[{"\[Mu]4", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}]}]}], ")"}], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"\[Rho]", "^", "2"}], "+", 
         RowBox[{
          RowBox[{"\[Mu]2", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}], 
          RowBox[{"\[Mu]5", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}]}]}], ")"}], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"\[Rho]", "^", "2"}], "+", 
         RowBox[{
          RowBox[{"\[Mu]3", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}], 
          RowBox[{"\[Mu]5", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}]}]}], ")"}], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"\[Rho]", "^", "2"}], "+", 
         RowBox[{
          RowBox[{"\[Mu]1", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}], "^", "2"}]}], ")"}], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"\[Rho]", "^", "2"}], "+", 
         RowBox[{
          RowBox[{"\[Mu]2", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}], "^", "2"}]}], ")"}], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"\[Rho]", "^", "2"}], "+", 
         RowBox[{
          RowBox[{"\[Mu]3", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}], "^", "2"}]}], ")"}], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"\[Rho]", "^", "2"}], "+", 
         RowBox[{
          RowBox[{"\[Mu]4", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}], "^", "2"}]}], ")"}], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"\[Rho]", "^", "2"}], "+", 
         RowBox[{
          RowBox[{"\[Mu]5", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}], "^", "2"}]}], ")"}]}]}], 
     "]"}]}]}], ";"}]], "Input",
 CellLabel->"In[78]:=",
 CellID->1280003364,ExpressionUUID->"c58770a4-d601-41e4-97bb-9e70fc59534c"],

Cell[BoxData[
 RowBox[{
  RowBox[{"MFunctions", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"M0", "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{"\[Rho]", ",", "z"}], "}"}], ",", 
        RowBox[{
         RowBox[{"\[Mu]2", "[", 
          RowBox[{"\[Rho]", ",", "z"}], "]"}], 
         RowBox[{
          RowBox[{"\[Mu]5", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}], "^", "2"}], 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{"\[Mu]1", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}], "-", 
            RowBox[{"\[Mu]3", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}]}], ")"}], "^", "2"}], 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{"\[Mu]2", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}], "-", 
            RowBox[{"\[Mu]4", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}]}], ")"}], "^", "2"}], 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{"\[Rho]", "^", "2"}], "+", 
            RowBox[{
             RowBox[{"\[Mu]1", "[", 
              RowBox[{"\[Rho]", ",", "z"}], "]"}], 
             RowBox[{"\[Mu]2", "[", 
              RowBox[{"\[Rho]", ",", "z"}], "]"}]}]}], ")"}], "^", "2"}], 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{"\[Rho]", "^", "2"}], "+", 
            RowBox[{
             RowBox[{"\[Mu]1", "[", 
              RowBox[{"\[Rho]", ",", "z"}], "]"}], 
             RowBox[{"\[Mu]4", "[", 
              RowBox[{"\[Rho]", ",", "z"}], "]"}]}]}], ")"}], "^", "2"}], 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{"\[Rho]", "^", "2"}], "+", 
            RowBox[{
             RowBox[{"\[Mu]2", "[", 
              RowBox[{"\[Rho]", ",", "z"}], "]"}], 
             RowBox[{"\[Mu]3", "[", 
              RowBox[{"\[Rho]", ",", "z"}], "]"}]}]}], ")"}], "^", "2"}]}]}], 
       "]"}]}], ",", 
     RowBox[{"M1", "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{"\[Rho]", ",", "z"}], "}"}], ",", 
        RowBox[{
         RowBox[{"\[Mu]2", "[", 
          RowBox[{"\[Rho]", ",", "z"}], "]"}], 
         RowBox[{"\[Mu]3", "[", 
          RowBox[{"\[Rho]", ",", "z"}], "]"}], 
         RowBox[{"\[Mu]4", "[", 
          RowBox[{"\[Rho]", ",", "z"}], "]"}], 
         RowBox[{"\[Mu]5", "[", 
          RowBox[{"\[Rho]", ",", "z"}], "]"}], 
         RowBox[{
          RowBox[{"\[Mu]1", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}], "^", "2"}], 
         RowBox[{"\[Rho]", "^", "2"}], 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{"\[Mu]1", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}], "-", 
            RowBox[{"\[Mu]2", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}]}], ")"}], "^", "2"}], 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{"\[Mu]2", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}], "-", 
            RowBox[{"\[Mu]4", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}]}], ")"}], "^", "2"}], 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{"\[Mu]1", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}], "-", 
            RowBox[{"\[Mu]5", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}]}], ")"}], "^", "2"}], 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{"\[Rho]", "^", "2"}], "+", 
            RowBox[{
             RowBox[{"\[Mu]2", "[", 
              RowBox[{"\[Rho]", ",", "z"}], "]"}], 
             RowBox[{"\[Mu]3", "[", 
              RowBox[{"\[Rho]", ",", "z"}], "]"}]}]}], ")"}], "^", "2"}]}]}], 
       "]"}]}], ",", 
     RowBox[{"M2", "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{"\[Rho]", ",", "z"}], "}"}], ",", 
        RowBox[{
         RowBox[{"\[Mu]2", "[", 
          RowBox[{"\[Rho]", ",", "z"}], "]"}], 
         RowBox[{"\[Mu]3", "[", 
          RowBox[{"\[Rho]", ",", "z"}], "]"}], 
         RowBox[{"\[Mu]4", "[", 
          RowBox[{"\[Rho]", ",", "z"}], "]"}], 
         RowBox[{"\[Mu]5", "[", 
          RowBox[{"\[Rho]", ",", "z"}], "]"}], 
         RowBox[{"\[Rho]", "^", "2"}], 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{"\[Mu]1", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}], "-", 
            RowBox[{"\[Mu]2", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}]}], ")"}], "^", "2"}], 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{"\[Mu]1", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}], "-", 
            RowBox[{"\[Mu]3", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}]}], ")"}], "^", "2"}], 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{"\[Rho]", "^", "2"}], "+", 
            RowBox[{
             RowBox[{"\[Mu]1", "[", 
              RowBox[{"\[Rho]", ",", "z"}], "]"}], 
             RowBox[{"\[Mu]4", "[", 
              RowBox[{"\[Rho]", ",", "z"}], "]"}]}]}], ")"}], "^", "2"}], 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{"\[Rho]", "^", "2"}], "+", 
            RowBox[{
             RowBox[{"\[Mu]2", "[", 
              RowBox[{"\[Rho]", ",", "z"}], "]"}], 
             RowBox[{"\[Mu]5", "[", 
              RowBox[{"\[Rho]", ",", "z"}], "]"}]}]}], ")"}], "^", "2"}]}]}], 
       "]"}]}], ",", 
     RowBox[{"M3", "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{"\[Rho]", ",", "z"}], "}"}], ",", 
        RowBox[{"2", 
         RowBox[{"\[Mu]1", "[", 
          RowBox[{"\[Rho]", ",", "z"}], "]"}], 
         RowBox[{"\[Mu]2", "[", 
          RowBox[{"\[Rho]", ",", "z"}], "]"}], 
         RowBox[{"\[Mu]3", "[", 
          RowBox[{"\[Rho]", ",", "z"}], "]"}], 
         RowBox[{"\[Mu]4", "[", 
          RowBox[{"\[Rho]", ",", "z"}], "]"}], 
         RowBox[{"\[Mu]5", "[", 
          RowBox[{"\[Rho]", ",", "z"}], "]"}], 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"\[Mu]1", "[", 
            RowBox[{"\[Rho]", ",", "z"}], "]"}], "-", 
           RowBox[{"\[Mu]3", "[", 
            RowBox[{"\[Rho]", ",", "z"}], "]"}]}], ")"}], 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"\[Mu]1", "[", 
            RowBox[{"\[Rho]", ",", "z"}], "]"}], "-", 
           RowBox[{"\[Mu]5", "[", 
            RowBox[{"\[Rho]", ",", "z"}], "]"}]}], ")"}], 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"\[Mu]2", "[", 
            RowBox[{"\[Rho]", ",", "z"}], "]"}], "-", 
           RowBox[{"\[Mu]4", "[", 
            RowBox[{"\[Rho]", ",", "z"}], "]"}]}], ")"}], 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"\[Rho]", "^", "2"}], "+", 
           RowBox[{
            RowBox[{"\[Mu]1", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}], "^", "2"}]}], ")"}], 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"\[Rho]", "^", "2"}], "+", 
           RowBox[{
            RowBox[{"\[Mu]2", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}], "^", "2"}]}], ")"}], 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"\[Rho]", "^", "2"}], "+", 
           RowBox[{
            RowBox[{"\[Mu]1", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}], 
            RowBox[{"\[Mu]4", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}]}]}], ")"}], 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"\[Rho]", "^", "2"}], "+", 
           RowBox[{
            RowBox[{"\[Mu]2", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}], 
            RowBox[{"\[Mu]3", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}]}]}], ")"}], 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"\[Rho]", "^", "2"}], "+", 
           RowBox[{
            RowBox[{"\[Mu]2", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}], 
            RowBox[{"\[Mu]5", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}]}]}], ")"}]}]}], "]"}]}], ",", 
     RowBox[{"M4", "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{"\[Rho]", ",", "z"}], "}"}], ",", 
        RowBox[{
         RowBox[{
          RowBox[{"\[Mu]1", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}], "^", "2"}], 
         RowBox[{"\[Mu]2", "[", 
          RowBox[{"\[Rho]", ",", "z"}], "]"}], 
         RowBox[{
          RowBox[{"\[Mu]3", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}], "^", "2"}], 
         RowBox[{
          RowBox[{"\[Mu]4", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}], "^", "2"}], 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{"\[Mu]1", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}], "-", 
            RowBox[{"\[Mu]5", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}]}], ")"}], "^", "2"}], 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{"\[Rho]", "^", "2"}], "+", 
            RowBox[{
             RowBox[{"\[Mu]1", "[", 
              RowBox[{"\[Rho]", ",", "z"}], "]"}], 
             RowBox[{"\[Mu]2", "[", 
              RowBox[{"\[Rho]", ",", "z"}], "]"}]}]}], ")"}], "^", "2"}], 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{"\[Rho]", "^", "2"}], "+", 
            RowBox[{
             RowBox[{"\[Mu]2", "[", 
              RowBox[{"\[Rho]", ",", "z"}], "]"}], 
             RowBox[{"\[Mu]5", "[", 
              RowBox[{"\[Rho]", ",", "z"}], "]"}]}]}], ")"}], "^", "2"}]}]}], 
       "]"}]}]}], "\[IndentingNewLine]", "}"}]}], ";"}]], "Input",
 CellLabel->"In[79]:=",
 CellID->1672846749,ExpressionUUID->"e32c445f-68b4-4919-aeda-1faec84002ab"]
}, Open  ]],

Cell[CellGroupData[{

Cell["The H functions are given in terms of M, \[Mu], F functions:", \
"MathCaption",
 CellID->679403919,ExpressionUUID->"7fd54161-df58-48d3-9032-561a14edf374"],

Cell[BoxData[
 RowBox[{
  RowBox[{"HFunctions", "=", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"Hx", "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{"\[Rho]", ",", "z"}], "}"}], ",", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"M0", "[", 
            RowBox[{"\[Rho]", ",", "z"}], "]"}], "+", 
           RowBox[{
            RowBox[{"c1", "^", "2"}], 
            RowBox[{"M1", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}]}], "+", 
           RowBox[{
            RowBox[{"c2", "^", "2"}], 
            RowBox[{"M2", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}]}], "+", 
           RowBox[{"c1", " ", "c2", " ", 
            RowBox[{"M3", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}]}], "+", 
           RowBox[{
            RowBox[{"c1", "^", "2"}], 
            RowBox[{"c2", "^", "2"}], 
            RowBox[{"M4", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}]}]}], ")"}], "/", 
         RowBox[{"F", "[", 
          RowBox[{"\[Rho]", ",", "z"}], "]"}]}]}], "]"}]}], ",", 
     RowBox[{"Hy", "\[Rule]", 
      RowBox[{"Function", "[", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{"\[Rho]", ",", "z"}], "}"}], ",", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"M0", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}], 
            RowBox[{
             RowBox[{"\[Mu]1", "[", 
              RowBox[{"\[Rho]", ",", "z"}], "]"}], "/", 
             RowBox[{"\[Mu]2", "[", 
              RowBox[{"\[Rho]", ",", "z"}], "]"}]}]}], "-", 
           RowBox[{
            RowBox[{"c1", "^", "2"}], 
            RowBox[{"M1", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}], 
            RowBox[{
             RowBox[{"\[Rho]", "^", "2"}], "/", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"\[Mu]1", "[", 
                RowBox[{"\[Rho]", ",", "z"}], "]"}], 
               RowBox[{"\[Mu]2", "[", 
                RowBox[{"\[Rho]", ",", "z"}], "]"}]}], ")"}]}]}], "-", 
           RowBox[{
            RowBox[{"c2", "^", "2"}], 
            RowBox[{"M2", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}], 
            RowBox[{
             RowBox[{"(", 
              RowBox[{
               RowBox[{"\[Mu]1", "[", 
                RowBox[{"\[Rho]", ",", "z"}], "]"}], 
               RowBox[{"\[Mu]2", "[", 
                RowBox[{"\[Rho]", ",", "z"}], "]"}]}], ")"}], "/", 
             RowBox[{"\[Rho]", "^", "2"}]}]}], "+", 
           RowBox[{"c1", " ", "c2", " ", 
            RowBox[{"M3", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}]}], "+", 
           RowBox[{
            RowBox[{"c1", "^", "2"}], 
            RowBox[{"c2", "^", "2"}], 
            RowBox[{"M4", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}], 
            RowBox[{
             RowBox[{"\[Mu]2", "[", 
              RowBox[{"\[Rho]", ",", "z"}], "]"}], "/", 
             RowBox[{"\[Mu]1", "[", 
              RowBox[{"\[Rho]", ",", "z"}], "]"}]}]}]}], ")"}], 
         RowBox[{
          RowBox[{"\[Mu]3", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}], "/", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"\[Mu]4", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}], 
            RowBox[{"F", "[", 
             RowBox[{"\[Rho]", ",", "z"}], "]"}]}], ")"}]}]}]}], "]"}]}]}], 
    "\[IndentingNewLine]", "}"}]}], ";"}]], "Input",
 CellLabel->"In[80]:=",
 CellID->571115534,ExpressionUUID->"46913206-0630-4731-9b01-c66e2d3145e9"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\[Omega]\[Psi] is given in terms of R, M, F, G functions:", \
"MathCaption",
 CellID->1634361495,ExpressionUUID->"ebccc0df-c8a5-4ccc-a07c-6d4284ae6898"],

Cell[BoxData[
 RowBox[{
  RowBox[{"\[Omega]\[Psi]Function", "=", 
   RowBox[{"\[Omega]\[Psi]", "\[Rule]", 
    RowBox[{"Function", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"\[Rho]", ",", "z"}], "}"}], ",", 
      RowBox[{"2", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"c1", " ", 
           RowBox[{"R1", "[", 
            RowBox[{"\[Rho]", ",", "z"}], "]"}], 
           RowBox[{"Sqrt", "[", 
            RowBox[{
             RowBox[{"M0", "[", 
              RowBox[{"\[Rho]", ",", "z"}], "]"}], 
             RowBox[{"M1", "[", 
              RowBox[{"\[Rho]", ",", "z"}], "]"}]}], "]"}]}], "-", 
          RowBox[{"c2", " ", 
           RowBox[{"R2", "[", 
            RowBox[{"\[Rho]", ",", "z"}], "]"}], 
           RowBox[{"Sqrt", "[", 
            RowBox[{
             RowBox[{"M0", "[", 
              RowBox[{"\[Rho]", ",", "z"}], "]"}], 
             RowBox[{"M2", "[", 
              RowBox[{"\[Rho]", ",", "z"}], "]"}]}], "]"}]}], "+", 
          RowBox[{
           RowBox[{"c1", "^", "2"}], "c2", " ", 
           RowBox[{"R2", "[", 
            RowBox[{"\[Rho]", ",", "z"}], "]"}], 
           RowBox[{"Sqrt", "[", 
            RowBox[{
             RowBox[{"M1", "[", 
              RowBox[{"\[Rho]", ",", "z"}], "]"}], 
             RowBox[{"M4", "[", 
              RowBox[{"\[Rho]", ",", "z"}], "]"}]}], "]"}]}], "-", 
          RowBox[{"c1", " ", 
           RowBox[{"c2", "^", "2"}], " ", 
           RowBox[{"R1", "[", 
            RowBox[{"\[Rho]", ",", "z"}], "]"}], 
           RowBox[{"Sqrt", "[", 
            RowBox[{
             RowBox[{"M2", "[", 
              RowBox[{"\[Rho]", ",", "z"}], "]"}], 
             RowBox[{"M4", "[", 
              RowBox[{"\[Rho]", ",", "z"}], "]"}]}], "]"}]}]}], ")"}], "/", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"F", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}], 
          RowBox[{"Sqrt", "@", 
           RowBox[{"Gx", "[", 
            RowBox[{"\[Rho]", ",", "z"}], "]"}]}]}], ")"}]}]}]}], "]"}]}]}], 
  ";"}]], "Input",
 CellLabel->"In[81]:=",
 CellID->250317688,ExpressionUUID->"84e19b4a-47c8-45af-a143-684cc5d77aea"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The metric is given in terms of H, \[Omega]\[Psi], F, P, G functions (note \
that P and F appear together), meaning that in the end all functions are \
replaced by elementary functions:\
\>", "MathCaption",
 CellID->168917743,ExpressionUUID->"a2db18f3-c19e-41ca-bbb6-8ed360bbc62c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"BlackSaturn", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"-", 
     FractionBox[
      RowBox[{"Hy", "[", 
       RowBox[{"\[Rho]", ",", "z"}], "]"}], 
      RowBox[{"Hx", "[", 
       RowBox[{"\[Rho]", ",", "z"}], "]"}]]}], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"Basis", "[", 
       RowBox[{
        RowBox[{"-", "a"}], ",", 
        RowBox[{"{", 
         RowBox[{"0", ",", "BS"}], "}"}]}], "]"}], "+", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         FractionBox[
          RowBox[{"\[Omega]\[Psi]", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}], 
          RowBox[{"Hy", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}]], "+", "q"}], ")"}], 
       RowBox[{"Basis", "[", 
        RowBox[{
         RowBox[{"-", "a"}], ",", 
         RowBox[{"{", 
          RowBox[{"3", ",", "BS"}], "}"}]}], "]"}]}]}], ")"}], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"Basis", "[", 
       RowBox[{
        RowBox[{"-", "b"}], ",", 
        RowBox[{"{", 
         RowBox[{"0", ",", "BS"}], "}"}]}], "]"}], "+", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         FractionBox[
          RowBox[{"\[Omega]\[Psi]", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}], 
          RowBox[{"Hy", "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}]], "+", "q"}], ")"}], 
       RowBox[{"Basis", "[", 
        RowBox[{
         RowBox[{"-", "b"}], ",", 
         RowBox[{"{", 
          RowBox[{"3", ",", "BS"}], "}"}]}], "]"}]}]}], ")"}]}], "+", 
   RowBox[{
    RowBox[{"Hx", "[", 
     RowBox[{"\[Rho]", ",", "z"}], "]"}], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"ks", "^", "2"}], 
       RowBox[{"P", "[", 
        RowBox[{"\[Rho]", ",", "z"}], "]"}], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"Basis", "[", 
           RowBox[{
            RowBox[{"-", "a"}], ",", 
            RowBox[{"{", 
             RowBox[{"1", ",", "BS"}], "}"}]}], "]"}], 
          RowBox[{"Basis", "[", 
           RowBox[{
            RowBox[{"-", "b"}], ",", 
            RowBox[{"{", 
             RowBox[{"1", ",", "BS"}], "}"}]}], "]"}]}], "+", 
         RowBox[{
          RowBox[{"Basis", "[", 
           RowBox[{
            RowBox[{"-", "a"}], ",", 
            RowBox[{"{", 
             RowBox[{"2", ",", "BS"}], "}"}]}], "]"}], 
          RowBox[{"Basis", "[", 
           RowBox[{
            RowBox[{"-", "b"}], ",", 
            RowBox[{"{", 
             RowBox[{"2", ",", "BS"}], "}"}]}], "]"}]}]}], ")"}]}], "+", 
      RowBox[{
       FractionBox[
        RowBox[{"Gy", "[", 
         RowBox[{"\[Rho]", ",", "z"}], "]"}], 
        RowBox[{"Hy", "[", 
         RowBox[{"\[Rho]", ",", "z"}], "]"}]], 
       RowBox[{"Basis", "[", 
        RowBox[{
         RowBox[{"-", "a"}], ",", 
         RowBox[{"{", 
          RowBox[{"3", ",", "BS"}], "}"}]}], "]"}], 
       RowBox[{"Basis", "[", 
        RowBox[{
         RowBox[{"-", "b"}], ",", 
         RowBox[{"{", 
          RowBox[{"3", ",", "BS"}], "}"}]}], "]"}]}], "+", 
      RowBox[{
       FractionBox[
        RowBox[{"Gx", "[", 
         RowBox[{"\[Rho]", ",", "z"}], "]"}], 
        RowBox[{"Hx", "[", 
         RowBox[{"\[Rho]", ",", "z"}], "]"}]], 
       RowBox[{"Basis", "[", 
        RowBox[{
         RowBox[{"-", "a"}], ",", 
         RowBox[{"{", 
          RowBox[{"4", ",", "BS"}], "}"}]}], "]"}], 
       RowBox[{"Basis", "[", 
        RowBox[{
         RowBox[{"-", "b"}], ",", 
         RowBox[{"{", 
          RowBox[{"4", ",", "BS"}], "}"}]}], "]"}]}]}], ")"}]}]}]}]], "Input",\

 CellLabel->"In[82]:=",
 CellID->27125267,ExpressionUUID->"aee4f8a8-513a-46e1-b591-94efdbcf2d2c"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox["Hx",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Hx,
     Editable->False], "[", 
    RowBox[{"\[Rho]", ",", "z"}], "]"}], " ", 
   RowBox[{"(", 
    RowBox[{
     FractionBox[
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["4",
                 FontColor->RGBColor[1., 0., 0.]]},
               {"a", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[-$CellContext`a, {4, $CellContext`BS}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["4",
                 FontColor->RGBColor[1., 0., 0.]]},
               {"b", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[-$CellContext`b, {4, $CellContext`BS}],
        Editable->False], " ", 
       RowBox[{
        InterpretationBox[
         StyleBox["Gx",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Gx,
         Editable->False], "[", 
        RowBox[{"\[Rho]", ",", "z"}], "]"}]}], 
      RowBox[{
       InterpretationBox[
        StyleBox["Hx",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Hx,
        Editable->False], "[", 
       RowBox[{"\[Rho]", ",", "z"}], "]"}]], "+", 
     FractionBox[
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["3",
                 FontColor->RGBColor[1., 0., 0.]]},
               {"a", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`BS}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["3",
                 FontColor->RGBColor[1., 0., 0.]]},
               {"b", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`BS}],
        Editable->False], " ", 
       RowBox[{
        InterpretationBox[
         StyleBox["Gy",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Gy,
         Editable->False], "[", 
        RowBox[{"\[Rho]", ",", "z"}], "]"}]}], 
      RowBox[{
       InterpretationBox[
        StyleBox["Hy",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Hy,
        Editable->False], "[", 
       RowBox[{"\[Rho]", ",", "z"}], "]"}]], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["ks",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ks,
        Editable->False], "2"], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox[GridBox[{
             {"e", 
              StyleBox[GridBox[{
                 {" ", 
                  StyleBox["1",
                   FontColor->RGBColor[1., 0., 0.]]},
                 {"a", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`BS}],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"e", 
              StyleBox[GridBox[{
                 {" ", 
                  StyleBox["1",
                   FontColor->RGBColor[1., 0., 0.]]},
                 {"b", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`BS}],
          Editable->False]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox[GridBox[{
             {"e", 
              StyleBox[GridBox[{
                 {" ", 
                  StyleBox["2",
                   FontColor->RGBColor[1., 0., 0.]]},
                 {"a", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`BS}],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"e", 
              StyleBox[GridBox[{
                 {" ", 
                  StyleBox["2",
                   FontColor->RGBColor[1., 0., 0.]]},
                 {"b", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`BS}],
          Editable->False]}]}], ")"}], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["P",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`P,
        Editable->False], "[", 
       RowBox[{"\[Rho]", ",", "z"}], "]"}]}]}], ")"}]}], "-", 
  FractionBox[
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox["Hy",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Hy,
      Editable->False], "[", 
     RowBox[{"\[Rho]", ",", "z"}], "]"}], " ", 
    RowBox[{"(", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {"a", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`BS}],
       Editable->False], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["3",
                 FontColor->RGBColor[1., 0., 0.]]},
               {"a", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`BS}],
        Editable->False], " ", 
       RowBox[{"(", 
        RowBox[{
         InterpretationBox[
          StyleBox["q",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`q,
          Editable->False], "+", 
         FractionBox[
          RowBox[{
           InterpretationBox[
            StyleBox["\[Omega]\[Psi]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Omega]\[Psi],
            Editable->False], "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}], 
          RowBox[{
           InterpretationBox[
            StyleBox["Hy",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Hy,
            Editable->False], "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}]]}], ")"}]}]}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {"b", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`BS}],
       Editable->False], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["3",
                 FontColor->RGBColor[1., 0., 0.]]},
               {"b", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`BS}],
        Editable->False], " ", 
       RowBox[{"(", 
        RowBox[{
         InterpretationBox[
          StyleBox["q",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`q,
          Editable->False], "+", 
         FractionBox[
          RowBox[{
           InterpretationBox[
            StyleBox["\[Omega]\[Psi]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Omega]\[Psi],
            Editable->False], "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}], 
          RowBox[{
           InterpretationBox[
            StyleBox["Hy",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Hy,
            Editable->False], "[", 
           RowBox[{"\[Rho]", ",", "z"}], "]"}]]}], ")"}]}]}], ")"}]}], 
   RowBox[{
    InterpretationBox[
     StyleBox["Hx",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Hx,
     Editable->False], "[", 
    RowBox[{"\[Rho]", ",", "z"}], "]"}]]}]], "Output",
 CellChangeTimes->{3.728677031403874*^9},
 CellLabel->"Out[82]=",
 CellID->880666121,ExpressionUUID->"1e3fbe4b-60e9-467d-aaf1-7ede1cdbd0bb"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"matrix", "=", 
  RowBox[{
   RowBox[{"Coefficient", "[", 
    RowBox[{"BlackSaturn", ",", 
     RowBox[{
      RowBox[{"BasisArray", "[", 
       RowBox[{"BS", ",", "BS"}], "]"}], "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "b"}]}], "]"}]}], "]"}], "/.", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"\[Rho]", "\[Rule]", 
      RowBox[{"\[Rho]", "[", "]"}]}], ",", 
     RowBox[{"z", "\[Rule]", 
      RowBox[{"z", "[", "]"}]}]}], "}"}]}]}]], "Input",
 CellLabel->"In[83]:=",
 CellID->1339971424,ExpressionUUID->"701307c1-2863-432b-b0f4-4a2540b82e59"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", 
      FractionBox[
       RowBox[{
        InterpretationBox[
         StyleBox["Hy",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Hy,
         Editable->False], "[", 
        RowBox[{
         InterpretationBox[
          StyleBox["\[Rho]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Rho][],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z[],
          Editable->False]}], "]"}], 
       RowBox[{
        InterpretationBox[
         StyleBox["Hx",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Hx,
         Editable->False], "[", 
        RowBox[{
         InterpretationBox[
          StyleBox["\[Rho]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Rho][],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z[],
          Editable->False]}], "]"}]]}], ",", "0", ",", "0", ",", 
     RowBox[{"-", 
      FractionBox[
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["Hy",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Hy,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["\[Rho]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Rho][],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z[],
           Editable->False]}], "]"}], " ", 
        RowBox[{"(", 
         RowBox[{
          InterpretationBox[
           StyleBox["q",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`q,
           Editable->False], "+", 
          FractionBox[
           RowBox[{
            InterpretationBox[
             StyleBox["\[Omega]\[Psi]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Omega]\[Psi],
             Editable->False], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["\[Rho]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Rho][],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["z",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`z[],
              Editable->False]}], "]"}], 
           RowBox[{
            InterpretationBox[
             StyleBox["Hy",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Hy,
             Editable->False], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["\[Rho]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Rho][],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["z",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`z[],
              Editable->False]}], "]"}]]}], ")"}]}], 
       RowBox[{
        InterpretationBox[
         StyleBox["Hx",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Hx,
         Editable->False], "[", 
        RowBox[{
         InterpretationBox[
          StyleBox["\[Rho]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Rho][],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z[],
          Editable->False]}], "]"}]]}], ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["ks",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ks,
        Editable->False], "2"], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["Hx",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Hx,
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho][],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["z",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`z[],
         Editable->False]}], "]"}], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["P",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`P,
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho][],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["z",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`z[],
         Editable->False]}], "]"}]}], ",", "0", ",", "0", ",", "0"}], "}"}], 
   ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["ks",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ks,
        Editable->False], "2"], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["Hx",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Hx,
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho][],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["z",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`z[],
         Editable->False]}], "]"}], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["P",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`P,
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho][],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["z",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`z[],
         Editable->False]}], "]"}]}], ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", 
      FractionBox[
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["Hy",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Hy,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["\[Rho]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Rho][],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z[],
           Editable->False]}], "]"}], " ", 
        RowBox[{"(", 
         RowBox[{
          InterpretationBox[
           StyleBox["q",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`q,
           Editable->False], "+", 
          FractionBox[
           RowBox[{
            InterpretationBox[
             StyleBox["\[Omega]\[Psi]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Omega]\[Psi],
             Editable->False], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["\[Rho]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Rho][],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["z",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`z[],
              Editable->False]}], "]"}], 
           RowBox[{
            InterpretationBox[
             StyleBox["Hy",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Hy,
             Editable->False], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["\[Rho]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Rho][],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["z",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`z[],
              Editable->False]}], "]"}]]}], ")"}]}], 
       RowBox[{
        InterpretationBox[
         StyleBox["Hx",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Hx,
         Editable->False], "[", 
        RowBox[{
         InterpretationBox[
          StyleBox["\[Rho]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Rho][],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z[],
          Editable->False]}], "]"}]]}], ",", "0", ",", "0", ",", 
     RowBox[{
      FractionBox[
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["Gy",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Gy,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["\[Rho]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Rho][],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z[],
           Editable->False]}], "]"}], " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["Hx",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Hx,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["\[Rho]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Rho][],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z[],
           Editable->False]}], "]"}]}], 
       RowBox[{
        InterpretationBox[
         StyleBox["Hy",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Hy,
         Editable->False], "[", 
        RowBox[{
         InterpretationBox[
          StyleBox["\[Rho]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Rho][],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z[],
          Editable->False]}], "]"}]], "-", 
      FractionBox[
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["Hy",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Hy,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["\[Rho]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Rho][],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z[],
           Editable->False]}], "]"}], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox["q",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`q,
            Editable->False], "+", 
           FractionBox[
            RowBox[{
             InterpretationBox[
              StyleBox["\[Omega]\[Psi]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Omega]\[Psi],
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["\[Rho]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Rho][],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}], 
            RowBox[{
             InterpretationBox[
              StyleBox["Hy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Hy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["\[Rho]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Rho][],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}]]}], ")"}], "2"]}], 
       RowBox[{
        InterpretationBox[
         StyleBox["Hx",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Hx,
         Editable->False], "[", 
        RowBox[{
         InterpretationBox[
          StyleBox["\[Rho]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Rho][],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z[],
          Editable->False]}], "]"}]]}], ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", 
     RowBox[{
      InterpretationBox[
       StyleBox["Gx",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Gx,
       Editable->False], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Rho]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Rho][],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["z",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`z[],
        Editable->False]}], "]"}]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728677032039473*^9},
 CellLabel->"Out[83]=",
 CellID->898200523,ExpressionUUID->"41b8ccc5-59b7-4c04-b42a-8388352a5717"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
If we expand the metric into its elementary functions then the result is too \
big to handle, even with Together:\
\>", "MathCaption",
 CellID->618494157,ExpressionUUID->"9487987c-cfc6-4407-aea0-c32b6fccfbb4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ByteCount", "[", 
  RowBox[{"ematrix", "=", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{
          RowBox[{"matrix", "/.", "\[Omega]\[Psi]Function"}], "/.", 
          "HFunctions"}], "/.", "MFunctions"}], "/.", "FFunction"}], "/.", 
       "PFunction"}], "/.", "GFunctions"}], "/.", "RFunctions"}], "/.", 
    "\[Mu]Functions"}]}], "]"}]], "Input",
 CellLabel->"In[84]:=",
 CellID->1210445219,ExpressionUUID->"57f154de-f93f-4db2-984b-34f2fc99de4b"],

Cell[BoxData["996320"], "Output",
 CellChangeTimes->{3.728677032325939*^9},
 CellLabel->"Out[84]=",
 CellID->1393635298,ExpressionUUID->"834da0fc-7cad-455e-af2b-ab2e6f8fef1c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Level", "[", 
   RowBox[{"ematrix", ",", 
    RowBox[{"{", 
     RowBox[{"-", "1"}], "}"}], ",", 
    RowBox[{"Heads", "\[Rule]", "True"}]}], "]"}], "//", "Union"}]], "Input",
 CellLabel->"In[85]:=",
 CellID->1665868489,ExpressionUUID->"8f68c92f-b78a-4d9e-8605-c3b702e46e58"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"-", "2"}], ",", 
   RowBox[{"-", "1"}], ",", 
   RowBox[{"-", 
    FractionBox["1", "2"]}], ",", "0", ",", 
   FractionBox["1", "2"], ",", "2", ",", "3", ",", "4", ",", 
   InterpretationBox[
    StyleBox["a1",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`a1,
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["a2",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`a2,
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["a3",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`a3,
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["a4",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`a4,
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["a5",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`a5,
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["c1",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`c1,
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["c2",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`c2,
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["ks",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ks,
    Editable->False], ",", "List", ",", "Plus", ",", 
   InterpretationBox[
    StyleBox["Power",
     ShowAutoStyles->False,
     AutoSpacing->False],
    Power,
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["q",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`q,
    Editable->False], ",", "Times", ",", "z", ",", "\[Rho]"}], 
  "}"}]], "Output",
 CellChangeTimes->{3.728677032752334*^9},
 CellLabel->"Out[85]=",
 CellID->224315657,ExpressionUUID->"8422051a-82b6-410b-a367-1ccbad920a1c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Instead, we will compute the Ricci tensor in the nonexpanded form and then \
replace the rules:\
\>", "MathCaption",
 CellID->817028329,ExpressionUUID->"430adfef-7997-46c4-8462-535419b12780"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{"met", "=", 
    RowBox[{"CTensor", "[", 
     RowBox[{"matrix", ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "BS"}], ",", 
        RowBox[{"-", "BS"}]}], "}"}]}], "]"}]}], ")"}], "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}]}], "]"}]], "Input",
 CellLabel->"In[86]:=",
 CellID->784686585,ExpressionUUID->"f80b0b53-ec19-4cb9-b604-45fec62f4168"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{"-", 
             FractionBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Hy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Hy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Rho]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Rho][],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], 
              RowBox[{
               InterpretationBox[
                StyleBox["Hx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Hx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Rho]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Rho][],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{"-", 
                FractionBox[
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}]}]],
           Annotation[#, {1, 1} -> ((-1)/$CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{"-", 
             FractionBox[
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["Hy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Hy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["\[Rho]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Rho][],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["z",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`z[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`q,
                  Editable->False], "+", 
                 FractionBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["\[Omega]\[Psi]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Omega]\[Psi],
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Rho][],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`z[],
                    Editable->False]}], "]"}], 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["Hy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Hy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Rho][],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`z[],
                    Editable->False]}], "]"}]]}], ")"}]}], 
              RowBox[{
               InterpretationBox[
                StyleBox["Hx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Hx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Rho]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Rho][],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "4"}], "}"}], "\[Rule]", 
              RowBox[{"-", 
                FractionBox[
                 RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}]}]],
           Annotation[#, {1, 4} -> (((-1)/$CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "5"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 5} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["ks",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ks,
               Editable->False], "2"], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Hx",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Hx,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Rho]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Rho][],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["P",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`P,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Rho]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Rho][],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], $CellContext`ks, Editable -> False], "2"], " ", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}]],
           Annotation[#, {2, 2} -> ($CellContext`ks^2 $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 4} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "5"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 5} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["ks",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`ks,
               Editable->False], "2"], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Hx",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Hx,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Rho]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Rho][],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["P",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`P,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Rho]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Rho][],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", 
              RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], $CellContext`ks, Editable -> False], "2"], " ", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}]],
           Annotation[#, {3, 3} -> ($CellContext`ks^2 $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 4} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "5"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 5} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{"-", 
             FractionBox[
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["Hy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Hy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["\[Rho]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Rho][],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["z",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`z[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`q,
                  Editable->False], "+", 
                 FractionBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["\[Omega]\[Psi]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Omega]\[Psi],
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Rho][],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`z[],
                    Editable->False]}], "]"}], 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["Hy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Hy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Rho][],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`z[],
                    Editable->False]}], "]"}]]}], ")"}]}], 
              RowBox[{
               InterpretationBox[
                StyleBox["Hx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Hx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Rho]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Rho][],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{"-", 
                FractionBox[
                 RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}]}]],
           Annotation[#, {4, 1} -> (((-1)/$CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "4"}], "}"}], "\[Rule]", 
              RowBox[{
                FractionBox[
                 RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                FractionBox[
                 RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}]}]],
           Annotation[#, {4, 4} -> $CellContext`Gy[
                $CellContext`\[Rho][], 
                $CellContext`z[]] ($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]/$CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) - ($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])^2, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "5"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 5} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {5, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {5, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {5, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {5, 4} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             InterpretationBox[
              StyleBox["Gx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Gx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["\[Rho]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Rho][],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "5"}], "}"}], "\[Rule]", 
              RowBox[{
                InterpretationBox[
                 StyleBox[
                  RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                  False], $CellContext`Gx, Editable -> False], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`\[Rho][], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`z[], Editable -> False]}], "]"}]}]],
           Annotation[#, {5, 5} -> $CellContext`Gx[
              $CellContext`\[Rho][], 
              $CellContext`z[]], "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 6 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 6 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " "},
         {"a", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{((-1)/$CellContext`Hx[
       $CellContext`\[Rho][], 
       $CellContext`z[]]) $CellContext`Hy[
       $CellContext`\[Rho][], 
       $CellContext`z[]], 0, 0, (((-1)/$CellContext`Hx[
        $CellContext`\[Rho][], 
        $CellContext`z[]]) $CellContext`Hy[
        $CellContext`\[Rho][], 
        $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
          $CellContext`\[Rho][], 
          $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
         $CellContext`\[Rho][], 
         $CellContext`z[]]), 0}, {0, ($CellContext`ks^2 $CellContext`Hx[
        $CellContext`\[Rho][], 
        $CellContext`z[]]) $CellContext`P[
       $CellContext`\[Rho][], 
       $CellContext`z[]], 0, 0, 0}, {
    0, 0, ($CellContext`ks^2 $CellContext`Hx[
        $CellContext`\[Rho][], 
        $CellContext`z[]]) $CellContext`P[
       $CellContext`\[Rho][], 
       $CellContext`z[]], 0, 0}, {(((-1)/$CellContext`Hx[
        $CellContext`\[Rho][], 
        $CellContext`z[]]) $CellContext`Hy[
        $CellContext`\[Rho][], 
        $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
          $CellContext`\[Rho][], 
          $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
         $CellContext`\[Rho][], 
         $CellContext`z[]]), 0, 0, $CellContext`Gy[
        $CellContext`\[Rho][], 
        $CellContext`z[]] ($CellContext`Hx[
         $CellContext`\[Rho][], 
         $CellContext`z[]]/$CellContext`Hy[
        $CellContext`\[Rho][], 
        $CellContext`z[]]) - ($CellContext`Hx[
         $CellContext`\[Rho][], 
         $CellContext`z[]]^(-1) $CellContext`Hy[
        $CellContext`\[Rho][], 
        $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
          $CellContext`\[Rho][], 
          $CellContext`z[]])^2, 0}, {0, 0, 0, 0, 
     $CellContext`Gx[
      $CellContext`\[Rho][], 
      $CellContext`z[]]}}, {-$CellContext`BS, -$CellContext`BS}, 
   0][-$CellContext`a, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.728677033207548*^9},
 CellLabel->"Out[86]=",
 CellID->1239446714,ExpressionUUID->"56ae631b-b6bd-4515-9ca3-27e98f05ee6b"]
}, Open  ]]
}, Open  ]],

Cell["\<\
Actually, it is possible to multiply the H and \[Omega]\[Psi] functions by F \
and then F appears only dividing P in the metric, so that F could be \
reabsorbed in the definition of P, simplifying the form of some intermediate \
results, but we prefer to follow the original.\
\>", "MathCaption",
 CellID->1667552426,ExpressionUUID->"be3e1cc2-e665-4ea6-bb9b-45b2238b8380"],

Cell[CellGroupData[{

Cell["We still use Simplify everywhere:", "MathCaption",
 CellID->1892030432,ExpressionUUID->"7d714ec5-837a-4ae4-b606-4c9279f78034"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SetCMetric", "[", 
   RowBox[{"met", ",", "BS", ",", 
    RowBox[{"SignatureOfMetric", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{"4", ",", "1", ",", "0"}], "}"}]}]}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[87]:=",
 CellID->1565109495,ExpressionUUID->"d36ba079-44c6-4447-ba46-9a610a3b921d"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.023599`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.728677033353104*^9},
 CellLabel->"Out[87]=",
 CellID->789546298,ExpressionUUID->"c87d4638-d419-4e6e-b5ce-8d07311ffb59"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"MetricCompute", "[", 
   RowBox[{"met", ",", "BS", ",", 
    RowBox[{"\"\<Ricci\>\"", "[", 
     RowBox[{
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}]}], "]"}]}], "]"}], "//", "AbsoluteTiming"}]], "Input",\

 CellLabel->"In[88]:=",
 CellID->374022728,ExpressionUUID->"c14e3612-7757-43aa-9001-abb8e2296c5e"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.16361`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.7286770336436043`*^9},
 CellLabel->"Out[88]=",
 CellID->603175673,ExpressionUUID->"63a5ef3e-2285-4fdf-a7a0-841e7b4accdc"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"cd", "=", 
   RowBox[{"CovDOfMetric", "[", "met", "]"}]}], ";"}]], "Input",
 CellLabel->"In[89]:=",
 CellID->765348338,ExpressionUUID->"08542a6b-b916-466f-9816-a65051593408"],

Cell[BoxData[
 RowBox[{
  RowBox[{"ricci", "=", 
   RowBox[{"Ricci", "[", "cd", "]"}]}], ";"}]], "Input",
 CellLabel->"In[90]:=",
 CellID->1929954883,ExpressionUUID->"c3cbe275-2ab9-4d08-91b5-5b195a1a285c"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Some components are already shown to be 0:", "MathCaption",
 CellID->1158714293,ExpressionUUID->"0a8c8845-244d-451d-af98-e1f79b7123e9"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ricci", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}]}], "]"}]], "Input",
 CellLabel->"In[91]:=",
 CellID->1129411455,ExpressionUUID->"cd667a72-cc9b-4186-b553-a8eab2ee2276"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{
                FractionBox[
                 RowBox[{
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                FractionBox[
                 RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    ")"}]}], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}],
                    "-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"]}], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                 RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                FractionBox[
                 RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                    FractionBox[
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], 
                 RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                FractionBox[
                 RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"]}], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    ")"}]}], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], 
                    ")"}]}]}], 
                 RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}]}]],
           Annotation[#, {1, 1} -> $CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) (((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + \
(($CellContext`ks^(-2)/$CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])/$CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) (((((
                    Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (Rational[1, 2] ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + (Rational[1, 2] (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((Rational[-1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + Rational[1, 2] (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 - ((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 2][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 2][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((((
                    Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((-$CellContext`ks^2) \
$CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])) + (($CellContext`Gy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-1)/$CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) ((((
                   Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))^2 + ((((
                    Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((-$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + (((
                   Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))^2 + ((((
                    Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((-$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + \
(($CellContext`ks^(-2)/$CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])/$CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) (((((
                    Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((-$CellContext`ks^2) \
$CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) + ((((
                    Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (Rational[1, 2] ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + (Rational[1, 2] (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((Rational[-1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + Rational[1, 2] (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 - ((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[2, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[2, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "4"}], "}"}], "\[Rule]", 
              RowBox[{
                FractionBox[
                 RowBox[{
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                FractionBox[
                 RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    ")"}]}], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}],
                    "-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"]}], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                 RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "-", 
                FractionBox[
                 RowBox[{
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                   " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "-", 
                    FractionBox[
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], 
                 RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                   
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]], 
                "+", 
                FractionBox[
                 RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"]}], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    ")"}]}], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], 
                    ")"}]}]}], 
                 RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}]}]],
           Annotation[#, {1, 4} -> $CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) (((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + ((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + \
(($CellContext`ks^(-2)/$CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])/$CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) (((((
                    Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + (Rational[1, 2] (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + (
                 Rational[1, 2] ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + 
               Rational[1, 2] ((((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 - (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ($CellContext`q + \
$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 2][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 2][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) (((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 - ((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                    Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 2][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 2][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + ((((
                    Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) - \
((((($CellContext`ks^(-4)/$CellContext`Gx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`Gy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-2)) $CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) $CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^2 + ((($CellContext`ks^4 \
$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])) ((((
                  Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])/$CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))^2 + ((((
                   Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) ((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((-$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Gy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]/$CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                 Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + (((
                  Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])/$CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))^2 + ((((
                   Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) ((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((-$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Gy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]/$CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                 Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + \
(($CellContext`ks^(-2)/$CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])/$CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) (((((
                    Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + ((((
                    Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + (Rational[1, 2] (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + (
                 Rational[1, 2] ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + 
               Rational[1, 2] ((((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 - (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ($CellContext`q + \
$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[2, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[2, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) (((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 - ((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                    Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[2, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[2, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "5"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 5} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{
                FractionBox[
                 RowBox[{
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"]}], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                   FractionBox[
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "-", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                FractionBox[
                 RowBox[{
                   RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                   " ", 
                   RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"]}], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], 
                    ")"}]}]}], ")"}]}], 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                FractionBox[
                 RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"]}], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "+", 
                   FractionBox[
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"]}], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                   FractionBox[
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                 RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                FractionBox[
                 RowBox[{
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                   " ", 
                   RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"]}], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], 
                    ")"}]}]}], ")"}]}], 
                 RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                   
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]], 
                "-", 
                FractionBox[
                 RowBox[{
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                   " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"]}], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], 
                    ")"}]}]}], ")"}]}], 
                 RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                   
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]], 
                "+", 
                FractionBox[
                 RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"]}], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], 
                    ")"}]}]}], ")"}]}], 
                 RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}]}]],
           Annotation[#, {2, 2} -> $CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) (((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) + (Rational[1, 4]/$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^2 + ((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + 
               Rational[-1, 2] Derivative[2, 0][$CellContext`Gx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) + ($CellContext`Gy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) ($CellContext`Gy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) (((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((-$CellContext`ks^2) \
$CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) + ((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (Rational[-1, 2] ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + (
                 Rational[-1, 2] (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((Rational[-1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + 
               Rational[1, 2] ((((-2) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                   Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[2, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[2, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])) + \
(($CellContext`ks^(-2)/$CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])/$CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) (((((
                    Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (((
                   Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2 + 
               Rational[
                 1, 2] ((((-2) $CellContext`ks^2) 
                   Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 2][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 2][$CellContext`P][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) + ((((
                    Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (((
                   Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2 + 
               Rational[
                 1, 2] ((((-2) $CellContext`ks^2) 
                   Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[2, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[2, 0][$CellContext`P][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])) + \
((((($CellContext`ks^(-4)/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])) (((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + ((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + (Rational[-1, 2] (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + (Rational[-1, 2] ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + 
               Rational[1, 2] (((((-2) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ($CellContext`q + \
$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[2, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[2, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) (((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 - ((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                   Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) 
                  Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[2, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[2, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) - \
((((($CellContext`ks^(-4)/$CellContext`Gx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`Gy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-2)) $CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) $CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^2 + ((($CellContext`ks^4 \
$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])) (((((
                   Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + ((((
                   Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + (Rational[1, 2] (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                 Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + (
                Rational[1, 2] ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + 
              Rational[1, 2] ((((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 - (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ($CellContext`q + \
$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] - (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                  Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[2, 0][$CellContext`Hx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[2, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) (((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 - ((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                   Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) 
                  Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[2, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[2, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + (($CellContext`Gy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-1)/$CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) (((((
                    Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((-$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + ((((
                    Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((-$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + (Rational[1, 2] (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + (((Rational[1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                   Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + (
                 Rational[1, 2] ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + ((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                   Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + 
               Rational[1, 2] ((((-2)/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                   Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (((2 $CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                   Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ($CellContext`q + \
$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (((2 $CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) 
                 Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 - ((((4 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (((4/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2 - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[2, 0][$CellContext`Gy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[2, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[2, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[2, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[2, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 - ((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                    Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[2, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[2, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))), "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", 
              RowBox[{
                FractionBox[
                 RowBox[{
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                   FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "-", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                FractionBox[
                 RowBox[{
                   RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                   " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox["1", "2"]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}],
                     "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], ")"}],
                     " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}],
                     "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], 
                    ")"}]}]}], ")"}]}], 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                FractionBox[
                 RowBox[{
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                   " ", 
                   RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}],
                     "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], ")"}],
                     " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}],
                     "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], 
                    ")"}]}]}], ")"}]}], 
                 RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                   
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]], 
                "-", 
                FractionBox[
                 RowBox[{
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                   " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], 
                    ")"}]}]}], ")"}]}], 
                 RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                   
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]], 
                "+", 
                FractionBox[
                 RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], 
                    ")"}]}]}], ")"}]}], 
                 RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}]}]],
           Annotation[#, {2, 3} -> $CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) (((Rational[1, 4]/$CellContext`Gx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] + ((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Gx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] + ((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + 
               Rational[-1, 2] Derivative[1, 1][$CellContext`Gx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) + ($CellContext`Gy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) ($CellContext`Gy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) ((
                 Rational[-1, 2] (((Rational[-1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])))) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) + ((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (
                 Rational[-1, 
                   2] (((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])))) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + 
               Rational[1, 2] (((((-2) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])) + \
((((($CellContext`ks^(-4)/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])) (((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (
                 Rational[-1, 2] (((Rational[-1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])))) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + ((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + (
                 Rational[-1, 
                   2] (((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])))) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                 Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + 
               Rational[1, 2] ((((((-2) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) 
                  Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) - \
((((($CellContext`ks^(-4)/$CellContext`Gx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`Gy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-2)) $CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) $CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^2 + ((($CellContext`ks^4 \
$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])) (((((
                   Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) + ((((
                   Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + (Rational[1, 2] (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                 Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + (
                Rational[1, 2] ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + 
              Rational[1, 2] (((((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                  Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 1][$CellContext`Hx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) 
                  Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + (($CellContext`Gy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-1)/$CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) (((((
                    Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (((-$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                   Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((((
                    Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((-$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + (Rational[1, 2] (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + (((Rational[1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                   Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + (
                 Rational[1, 2] ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + ((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                   Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + 
               Rational[1, 2] ((((-1)/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                   Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ((((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                   Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ((((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                   Derivative[0, 1][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                   Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ((((2 $CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) 
                  Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ((((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 1][$CellContext`Gy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 1][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 1][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[1, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 4} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "5"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 5} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{
                FractionBox[
                 RowBox[{
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                   FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "-", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                FractionBox[
                 RowBox[{
                   RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                   " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox["1", "2"]}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}],
                     "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], ")"}],
                     " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}],
                     "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], 
                    ")"}]}]}], ")"}]}], 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                FractionBox[
                 RowBox[{
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                   " ", 
                   RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    ")"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}],
                     "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], ")"}],
                     " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}],
                     "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], 
                    ")"}]}]}], ")"}]}], 
                 RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                   
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]], 
                "-", 
                FractionBox[
                 RowBox[{
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                   " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], 
                    ")"}]}]}], ")"}]}], 
                 RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                   
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]], 
                "+", 
                FractionBox[
                 RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], 
                    ")"}]}]}], ")"}]}], 
                 RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}]}]],
           Annotation[#, {3, 2} -> $CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) (((Rational[1, 4]/$CellContext`Gx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] + ((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Gx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] + ((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + 
               Rational[-1, 2] Derivative[1, 1][$CellContext`Gx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) + ($CellContext`Gy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) ($CellContext`Gy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) ((
                 Rational[-1, 2] (((Rational[-1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])))) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) + ((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (
                 Rational[-1, 
                   2] (((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])))) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + 
               Rational[1, 2] (((((-2) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])) + \
((((($CellContext`ks^(-4)/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])) (((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (
                 Rational[-1, 2] (((Rational[-1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])))) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + ((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + (
                 Rational[-1, 
                   2] (((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])))) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                 Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + 
               Rational[1, 2] ((((((-2) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) 
                  Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) - \
((((($CellContext`ks^(-4)/$CellContext`Gx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`Gy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-2)) $CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) $CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^2 + ((($CellContext`ks^4 \
$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])) (((((
                   Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) + ((((
                   Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + (Rational[1, 2] (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                 Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + (
                Rational[1, 2] ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + 
              Rational[1, 2] (((((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                  Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 1][$CellContext`Hx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) 
                  Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + (($CellContext`Gy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-1)/$CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) (((((
                    Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (((-$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                   Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((((
                    Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((-$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + (Rational[1, 2] (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + (((Rational[1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                   Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + (
                 Rational[1, 2] ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + ((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                   Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + 
               Rational[1, 2] ((((-1)/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                   Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ((((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                   Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ((((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                   Derivative[0, 1][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                   Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ((((2 $CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) 
                  Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ((((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 1][$CellContext`Gy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 1][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 1][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[1, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))), "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", 
              RowBox[{
                FractionBox[
                 RowBox[{
                   FractionBox[
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "-", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"]}], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                FractionBox[
                 RowBox[{
                   RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                   " ", 
                   RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}],
                     "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"]}], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                FractionBox[
                 RowBox[{
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                   " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}],
                     "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"]}], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], 
                 RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                   
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]], 
                "+", 
                FractionBox[
                 RowBox[{
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                   " ", 
                   RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], "-", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}],
                     "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"]}], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], 
                 RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                   
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]], 
                "+", 
                FractionBox[
                 RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}],
                     "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"]}], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], 
                 RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "-", 
                FractionBox[
                 RowBox[{
                   FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"]}], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "-", 
                   FractionBox[
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], "+", 
                   FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"]}], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "-", 
                   FractionBox[
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    "2"], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}]}], 
                 RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}]}]],
           Annotation[#, {3, 3} -> $CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) ((Rational[1, 4]/$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^2 + ((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + 
               Rational[-1, 2] Derivative[0, 2][$CellContext`Gx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] + ((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])) + ($CellContext`Gy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) ($CellContext`Gy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) (((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (Rational[-1, 2] ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + (
                 Rational[-1, 2] (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((Rational[-1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + 
               Rational[1, 2] ((((-2) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                   Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 2][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 2][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) + ((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((-$CellContext`ks^2) \
$CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])) - \
((((($CellContext`ks^(-4)/$CellContext`Gx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`Gy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-2)) $CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) $CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^2 + ((($CellContext`ks^4 \
$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])) (((((
                   Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + (Rational[1, 2] (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                 Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + (
                Rational[1, 2] ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + 
              Rational[1, 2] ((((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 - (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ($CellContext`q + \
$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] - (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                  Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 2][$CellContext`Hx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 2][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) (((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 - ((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                   Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) 
                  Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 2][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[0, 2][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + ((((
                   Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + \
((((($CellContext`ks^(-4)/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])) (((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + (Rational[-1, 2] (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + (Rational[-1, 2] ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + 
               Rational[1, 2] (((((-2) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ($CellContext`q + \
$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 2][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 2][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) (((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 - ((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                   Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) 
                  Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 2][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[0, 2][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + ((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + (($CellContext`Gy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-1)/$CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) (((((
                    Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((-$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + (Rational[1, 2] (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + (((Rational[1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                   Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + (
                 Rational[1, 2] ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + ((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                   Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + 
               Rational[1, 2] ((((-2)/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                   Derivative[0, 1][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (((2 $CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                   Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ($CellContext`q + \
$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (((2 $CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) 
                 Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 - ((((4 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (((4/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2 - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 2][$CellContext`Gy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 2][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 2][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[0, 2][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 2][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 - ((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                    Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 2][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 2][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + ((((
                    Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((-$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) - \
(($CellContext`ks^(-2)/$CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) (((((
                   Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) + (((
                  Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])/$CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2 + 
              Rational[
                1, 2] (((2 $CellContext`ks^2) 
                  Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 2][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 2][$CellContext`P][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) + ((((
                   Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) + (((
                  Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])/$CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2 + 
              Rational[
                1, 2] (((2 $CellContext`ks^2) 
                  Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[2, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[2, 0][$CellContext`P][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 4} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "5"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 5} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{
                FractionBox[
                 RowBox[{
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                FractionBox[
                 RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    ")"}]}], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}],
                    "-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"]}], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                 RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "-", 
                FractionBox[
                 RowBox[{
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                   " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "-", 
                    FractionBox[
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], 
                 RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                   
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]], 
                "+", 
                FractionBox[
                 RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"]}], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    ")"}]}], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], 
                    ")"}]}]}], 
                 RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}]}]],
           Annotation[#, {4, 1} -> $CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) (((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + ((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + \
(($CellContext`ks^(-2)/$CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])/$CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) (((((
                    Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + (Rational[1, 2] (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + (
                 Rational[1, 2] ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + 
               Rational[1, 2] ((((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 - (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ($CellContext`q + \
$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 2][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 2][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) (((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 - ((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                    Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 2][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 2][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + ((((
                    Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) - \
((((($CellContext`ks^(-4)/$CellContext`Gx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`Gy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-2)) $CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) $CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^2 + ((($CellContext`ks^4 \
$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])) ((((
                  Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])/$CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))^2 + ((((
                   Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) ((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((-$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Gy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]/$CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                 Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + (((
                  Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])/$CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))^2 + ((((
                   Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) ((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((-$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Gy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]/$CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                 Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + \
(($CellContext`ks^(-2)/$CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])/$CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) (((((
                    Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + ((((
                    Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + (Rational[1, 2] (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + (
                 Rational[1, 2] ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + 
               Rational[1, 2] ((((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 - (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ($CellContext`q + \
$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[2, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[2, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) (((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 - ((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                    Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[2, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[2, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "4"}], "}"}], "\[Rule]", 
              RowBox[{
                FractionBox[
                 RowBox[{
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                FractionBox[
                 RowBox[{
                   RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                   " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "-", 
                    FractionBox[
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                FractionBox[
                 RowBox[{
                   FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "-", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], "-", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    ")"}]}], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "+", 
                    FractionBox[
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}],
                    "+", 
                   FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"]}], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                 RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "-", 
                FractionBox[
                 RowBox[{
                   FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"]}], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                   FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "-", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], "-", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]]}], 
                    ")"}]}], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "+", 
                    FractionBox[
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["z", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "3"]], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], 
                    ")"}]}]}], 
                 RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}]}]],
           Annotation[#, {4, 4} -> $CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) (((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (((-$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + ((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (((-$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + ($CellContext`Gy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) ($CellContext`Gy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) ((((
                   Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))^2 + ((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((-$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + (((
                   Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))^2 + ((((
                    Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((-$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) - \
(($CellContext`ks^(-2)/$CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) (((((
                   Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((-$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Gy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]/$CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                 Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + (Rational[-1, 2] (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                 Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + (((Rational[1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + (Rational[-1, 2] ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + ((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + Rational[1, 2] (((2/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^2 - (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                 Derivative[0, 1][$CellContext`Gy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] - (((2 $CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                 Derivative[0, 1][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] + (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ($CellContext`q + \
$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (((2 $CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) 
                 Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((((4 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - (((4/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                  Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ((2/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2 + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 2][$CellContext`Gy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 2][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 2][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                Derivative[0, 2][$CellContext`Hy][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 2][$CellContext`Hy][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 - ((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) 
                 Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 2][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                  Derivative[0, 2][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + ((((
                   Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])) (((-$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Gy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]/$CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                 Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) - \
(($CellContext`ks^(-2)/$CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) (((((
                   Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])) (((-$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Gy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]/$CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                 Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + ((((
                   Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((-$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Gy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]/$CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                 Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + (Rational[-1, 2] (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                 Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + (((Rational[1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + (Rational[-1, 2] ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) ((($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + ((((((
                    Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) + Rational[1, 2] (((2/$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^2 - (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                 Derivative[1, 0][$CellContext`Gy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] - (((2 $CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                 Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] + (((2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) ($CellContext`q + \
$CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (((2 $CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) 
                 Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((((4 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - (((4/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                  Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) - ((2/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2 + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[2, 0][$CellContext`Gy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[2, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[2, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                Derivative[2, 0][$CellContext`Hy][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[2, 0][$CellContext`Hy][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] - (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) (((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 - ((2 $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                  Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) 
                 Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[2, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                  Derivative[2, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "5"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 5} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {5, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {5, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {5, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {5, 4} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "5"}], "}"}], "\[Rule]", 
              RowBox[{
                FractionBox[
                 RowBox[{
                   RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                   " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                FractionBox[
                 RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"]}], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                 RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                FractionBox[
                 RowBox[{"2", " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], ")"}], 
                   " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], 
                 RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "4"], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], " ", 
                   
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]}]], 
                "-", 
                FractionBox[
                 RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]], "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]], "+", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"q"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`q, Editable -> False], "+", 
                    FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"]]}], 
                    "+", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Omega]\[Psi]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Omega]\[Psi], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]]}], ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], 
                    ")"}]}], 
                 RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gy"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gy, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "-", 
                FractionBox[
                 RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"]}], " ", 
                    
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}], "-", 
                   FractionBox[
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], "2"], 
                    RowBox[{"4", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "-", 
                   FractionBox[
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                    RowBox[{"4", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Gx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Gx, Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]}], 
                 RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"ks"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`ks, Editable -> False], "2"], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Hx"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Hx, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"P"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`P, Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[Rho]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Rho][], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "z", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`z[], Editable -> False]}], "]"}]}]]}]}]],
           Annotation[#, {5, 5} -> ($CellContext`Gy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) ($CellContext`Gy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) - ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2)) (((((
                    Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) + ((((
                    Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) ((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) - \
(($CellContext`ks^(-2)/$CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ((Rational[-1, 4]/$CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^2 + ((((
                   Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) + 
              Rational[1, 2] Derivative[0, 2][$CellContext`Gx][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]] + ((((
                   Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])) + ((((((
                    2 $CellContext`ks^(-4))/$CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Gy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])) (((((
                    Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + ((((
                    Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`P[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) (((((-1) $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                    Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) - (($CellContext`Gy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1)/$CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) (((((
                   Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) (((-$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Gy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]/$CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                 Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) + ((((
                   Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) (((-$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])/$CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`Gy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]/$CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] - (($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]] + (($CellContext`Gy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) 
                 Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + (((2/$CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]])) ((((-1) $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]] + $CellContext`Hy[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]^(-1) 
                   Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]))) - \
(($CellContext`ks^(-2)/$CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) (((((
                   Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) + (Rational[-1, 4]/$CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^2 + ((((
                   Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]])/$CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
                    $CellContext`\[Rho][], 
                    $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) + 
              Rational[1, 2] Derivative[2, 0][$CellContext`Gx][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]), "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 6 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 6 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " "},
         {"a", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{$CellContext`Gx[
         $CellContext`\[Rho][], 
         $CellContext`z[]]^(-1) (((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ((((-1) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + ((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ((((-1) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]])) + (($CellContext`ks^(-2)/$CellContext`Hx[
         $CellContext`\[Rho][], 
         $CellContext`z[]])/$CellContext`P[
        $CellContext`\[Rho][], 
        $CellContext`z[]]) (((((
            Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) (($CellContext`ks^2 $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + (Rational[1, 2] ((($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) (((((((
                Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]])/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) + (Rational[1, 2] (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]])) (((Rational[-1, 2]/$CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]] ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]/$CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2)) (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + ((((((
                Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) + Rational[1, 2] (((2 $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-3)) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]^2 - ((2 $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 2][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[0, 2][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + ((((
            Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) (((-$CellContext`ks^2) $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
           $CellContext`\[Rho][], 
           $CellContext`z[]])) + (($CellContext`Gy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]^(-1)/$CellContext`Hx[
         $CellContext`\[Rho][], 
         $CellContext`z[]]) $CellContext`Hy[
         $CellContext`\[Rho][], 
         $CellContext`z[]]) ((((
           Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) (((((-1) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))^2 + ((((
            Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) (((-$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]])/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Gy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) + (((
           Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) (((((-1) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))^2 + ((((
            Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) (((-$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]])/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Gy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) + (($CellContext`ks^(-2)/$CellContext`Hx[
         $CellContext`\[Rho][], 
         $CellContext`z[]])/$CellContext`P[
        $CellContext`\[Rho][], 
        $CellContext`z[]]) (((((
            Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) (((-$CellContext`ks^2) $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) + ((((
            Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) (($CellContext`ks^2 $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + (Rational[1, 2] ((($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) (((((((
                Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]])/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) + (Rational[1, 2] (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]])) (((Rational[-1, 2]/$CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]] ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]/$CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2)) (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + ((((((
                Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) + Rational[1, 2] (((2 $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-3)) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]^2 - ((2 $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[2, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[2, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]])), 0, 0, $CellContext`Gx[
         $CellContext`\[Rho][], 
         $CellContext`z[]]^(-1) (((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) + ((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) + (($CellContext`ks^(-2)/$CellContext`Hx[
         $CellContext`\[Rho][], 
         $CellContext`z[]])/$CellContext`P[
        $CellContext`\[Rho][], 
        $CellContext`z[]]) (((((
            Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) (((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) + (Rational[1, 2] (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) (((((((
                Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]])/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) + (Rational[1, 2] ((($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]] ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]/$CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2)) (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + ((((((
                Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) + Rational[1, 2] ((((2 $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-3)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]^2 - (((2 $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (((2 $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + ((2/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) - (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[0, 2][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 2][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) (((2 $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]]^2 - ((2 $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 2][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 2][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) + ((((
            Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]])) (((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) - ((((($CellContext`ks^(-4)/$CellContext`Gx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`Gy[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) $CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]]^(-2)) $CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) $CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) $CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) $CellContext`P[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
          $CellContext`\[Rho][], 
          $CellContext`z[]])) ((((
          Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
          $CellContext`\[Rho][], 
          $CellContext`z[]])/$CellContext`P[
         $CellContext`\[Rho][], 
         $CellContext`z[]]) (((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))^2 + ((((
           Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) ((((-1) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]])) (((-$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]])/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Gy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]/$CellContext`Hy[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] - (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] + (($CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]] $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (((2/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) + (((
          Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
          $CellContext`\[Rho][], 
          $CellContext`z[]])/$CellContext`P[
         $CellContext`\[Rho][], 
         $CellContext`z[]]) (((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))^2 + ((((
           Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) ((((-1) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]])) (((-$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]])/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Gy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]/$CellContext`Hy[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] - (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] + (($CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]] $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (((2/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]))) + (($CellContext`ks^(-2)/$CellContext`Hx[
         $CellContext`\[Rho][], 
         $CellContext`z[]])/$CellContext`P[
        $CellContext`\[Rho][], 
        $CellContext`z[]]) (((((
            Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]])) (((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) + ((((
            Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) (((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) + (Rational[1, 2] (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) (((((((
                Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]])/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) + (Rational[1, 2] ((($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]] ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]/$CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2)) (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + ((((((
                Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) + Rational[1, 2] ((((2 $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-3)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]^2 - (((2 $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (((2 $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + ((2/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) - (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[2, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[2, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) (((2 $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]]^2 - ((2 $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[2, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[2, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))), 0}, {0, $CellContext`Gx[
         $CellContext`\[Rho][], 
         $CellContext`z[]]^(-1) (((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) + (Rational[1, 4]/$CellContext`Gx[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]^2 + ((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + 
       Rational[-1, 2] Derivative[2, 0][$CellContext`Gx][
          $CellContext`\[Rho][], 
          $CellContext`z[]]) + ($CellContext`Gy[
          $CellContext`\[Rho][], 
          $CellContext`z[]]^(-1) ($CellContext`Gy[
           $CellContext`\[Rho][], 
           $CellContext`z[]] ($CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) - ($CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^(-1) $CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
             $CellContext`\[Rho][], 
             $CellContext`z[]])^2)) (((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) (((-$CellContext`ks^2) $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) + ((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) (($CellContext`ks^2 $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + (Rational[-1, 2] ((($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) (((((((
                Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]])/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) + (Rational[-1, 2] (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]])) (((Rational[-1, 2]/$CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]] ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]/$CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2)) (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + ((((((
                Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) + Rational[1, 2] ((((-2) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-3)) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]^2 + ((2 $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[2, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^(-1) Derivative[2, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]])) + (($CellContext`ks^(-2)/$CellContext`Hx[
         $CellContext`\[Rho][], 
         $CellContext`z[]])/$CellContext`P[
        $CellContext`\[Rho][], 
        $CellContext`z[]]) (((((
            Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]])) (($CellContext`ks^2 $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + (((
           Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
             $CellContext`\[Rho][], 
             $CellContext`z[]])^2 + 
       Rational[
         1, 2] ((((-2) $CellContext`ks^2) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`ks^2 $CellContext`P[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 2][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 2][$CellContext`P][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) + ((((
            Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]])) (($CellContext`ks^2 $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + (((
           Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
             $CellContext`\[Rho][], 
             $CellContext`z[]])^2 + 
       Rational[
         1, 2] ((((-2) $CellContext`ks^2) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`ks^2 $CellContext`P[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[2, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[2, 0][$CellContext`P][
           $CellContext`\[Rho][], 
           $CellContext`z[]])) + ((((($CellContext`ks^(-4)/$CellContext`Gx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`Gy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^(-2)) $CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) $CellContext`P[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
           $CellContext`\[Rho][], 
           $CellContext`z[]])) (((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]])) (((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) + ((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) (((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) + (Rational[-1, 2] (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) (((((((
                Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]])/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) + (Rational[-1, 2] ((($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]] ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]/$CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2)) (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + ((((((
                Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) + 
       Rational[1, 2] (((((-2) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-3)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]^2 + (((2 $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (((2 $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) - ((2/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[2, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[2, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) (((2 $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]]^2 - ((2 $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) 
          Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[2, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[2, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]))) - \
((((($CellContext`ks^(-4)/$CellContext`Gx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`Gy[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) $CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]]^(-2)) $CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) $CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) $CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) $CellContext`P[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
          $CellContext`\[Rho][], 
          $CellContext`z[]])) (((((
           Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
           $CellContext`\[Rho][], 
           $CellContext`z[]])) (((((-1) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) + ((((
           Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]])) (((((-1) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) + (Rational[1, 2] (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (($CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]] $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (((2/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]))) (((((((
               Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
               $CellContext`\[Rho][], 
               $CellContext`z[]])/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]])/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) + (Rational[1, 2] ((($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]/$CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2)) (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + ((((((
               Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
               $CellContext`\[Rho][], 
               $CellContext`z[]])/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) ((($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) + Rational[1, 2] ((((2 $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-3)) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]]^2 - (((2 $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
          $CellContext`\[Rho][], 
          $CellContext`z[]] - (((2 $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ((((-1) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) 
          Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + ((2/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) - (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) Derivative[2, 0][$CellContext`Hx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[2, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) (((2 $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]]^2 - ((2 $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) 
          Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[2, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[2, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]))) + (($CellContext`Gy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]^(-1)/$CellContext`Hx[
         $CellContext`\[Rho][], 
         $CellContext`z[]]) $CellContext`Hy[
         $CellContext`\[Rho][], 
         $CellContext`z[]]) (((((
            Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]])) (((-$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]])/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Gy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) + ((((
            Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) (((-$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]])/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Gy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) + (Rational[1, 2] (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) (((((((
                Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) + (((Rational[1, 2]/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]])/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]/$CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + ($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]/$CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + (($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - (($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]] $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - (((2/$CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) + (Rational[1, 2] ((($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]] ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]/$CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2)) ((($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) + ((((((
                Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]/$CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + ($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]/$CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + (($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - (($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]] $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - (((2/$CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) + Rational[1, 2] ((((-2)/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (((2 $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-3)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]^2 + (((2 $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Gy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (((2 $CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (((2 $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (((2 $CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-3)) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]^2 - ((((4 $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + (((4/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) + ((2/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2 - ($CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[2, 0][$CellContext`Gy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Gy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[2, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[2, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[2, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[2, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) (((2 $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]]^2 - ((2 $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[2, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[2, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))), $CellContext`Gx[
         $CellContext`\[Rho][], 
         $CellContext`z[]]^(-1) (((Rational[1, 4]/$CellContext`Gx[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] + ((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) Derivative[1, 0][$CellContext`Gx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] + ((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + 
       Rational[-1, 2] Derivative[1, 1][$CellContext`Gx][
          $CellContext`\[Rho][], 
          $CellContext`z[]]) + ($CellContext`Gy[
          $CellContext`\[Rho][], 
          $CellContext`z[]]^(-1) ($CellContext`Gy[
           $CellContext`\[Rho][], 
           $CellContext`z[]] ($CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) - ($CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^(-1) $CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
             $CellContext`\[Rho][], 
             $CellContext`z[]])^2)) ((
         Rational[-1, 2] (((Rational[-1, 2]/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`Gy[
                $CellContext`\[Rho][], 
                $CellContext`z[]] ($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]/$CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) - ($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])^2)) (($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
               $CellContext`\[Rho][], 
               $CellContext`z[]] - $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) + ((((((
                 Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])/$CellContext`Gy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) $CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) $CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) ((($CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
               $CellContext`\[Rho][], 
               $CellContext`z[]] - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ((((-1) $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                $CellContext`\[Rho][], 
                $CellContext`z[]] + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) 
              Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                $CellContext`\[Rho][], 
                $CellContext`z[]])))) (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) + ((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) ((((-1) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + ((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) (($CellContext`ks^2 $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + (
         Rational[-1, 
           2] (((((((Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])/$CellContext`Gy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) $CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) $CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) (($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
               $CellContext`\[Rho][], 
               $CellContext`z[]] - $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]])/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((($CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
               $CellContext`\[Rho][], 
               $CellContext`z[]] - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ((((-1) $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                $CellContext`\[Rho][], 
                $CellContext`z[]] + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) 
              Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                $CellContext`\[Rho][], 
                $CellContext`z[]])))) ((($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) + Rational[1, 2] (((((-2) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-3)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^(-1) Derivative[1, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]])) + ((((($CellContext`ks^(-4)/$CellContext`Gx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`Gy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^(-2)) $CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) $CellContext`P[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
           $CellContext`\[Rho][], 
           $CellContext`z[]])) (((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (((((-1) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) (($CellContext`ks^2 $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + (
         Rational[-1, 2] (((Rational[-1, 2]/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`Gy[
                $CellContext`\[Rho][], 
                $CellContext`z[]] ($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]/$CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) - ($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])^2)) (($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
               $CellContext`\[Rho][], 
               $CellContext`z[]] - $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) + ((((((
                 Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])/$CellContext`Gy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) $CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) $CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) ((($CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
               $CellContext`\[Rho][], 
               $CellContext`z[]] - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ((((-1) $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                $CellContext`\[Rho][], 
                $CellContext`z[]] + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) 
              Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                $CellContext`\[Rho][], 
                $CellContext`z[]])))) ((($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) + ((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) (((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) + (
         Rational[-1, 
           2] (((((((Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])/$CellContext`Gy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) $CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) $CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) (($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
               $CellContext`\[Rho][], 
               $CellContext`z[]] - $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]])/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((($CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
               $CellContext`\[Rho][], 
               $CellContext`z[]] - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ((((-1) $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                $CellContext`\[Rho][], 
                $CellContext`z[]] + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) 
              Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                $CellContext`\[Rho][], 
                $CellContext`z[]])))) (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (($CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]] $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (((2/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) + Rational[1, 2] ((((((-2) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-3)) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[1, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((2 $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) 
           Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) 
          Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]))) - \
((((($CellContext`ks^(-4)/$CellContext`Gx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`Gy[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) $CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]]^(-2)) $CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) $CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) $CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) $CellContext`P[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
          $CellContext`\[Rho][], 
          $CellContext`z[]])) (((((
           Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) (((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) (($CellContext`ks^2 $CellContext`P[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) + ((((
           Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]])) (((((-1) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) + (Rational[1, 2] (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (($CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]] $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (((2/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]))) (((((((
               Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
               $CellContext`\[Rho][], 
               $CellContext`z[]])/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]])/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) + (Rational[1, 2] ((($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]/$CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2)) (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + ((((((
               Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
               $CellContext`\[Rho][], 
               $CellContext`z[]])/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) ((($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) + Rational[1, 2] (((((2 $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-3)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] - (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] - (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
          $CellContext`\[Rho][], 
          $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ((((-1) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) 
          Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) - (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) Derivative[1, 1][$CellContext`Hx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[1, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((2 $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) 
           Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) 
          Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]))) + (($CellContext`Gy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]^(-1)/$CellContext`Hx[
         $CellContext`\[Rho][], 
         $CellContext`z[]]) $CellContext`Hy[
         $CellContext`\[Rho][], 
         $CellContext`z[]]) (((((
            Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (((-$CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]])/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]] $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + (((2/$CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) (($CellContext`ks^2 $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + ((((
            Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) (((-$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]])/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Gy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) + (Rational[1, 2] (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) (((((((
                Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) + (((Rational[1, 2]/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]])/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]/$CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + ($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]/$CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + (($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - (($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]] $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - (((2/$CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) + (Rational[1, 2] ((($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]] ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]/$CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2)) ((($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) + ((((((
                Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]/$CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + ($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]/$CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + (($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - (($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]] $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - (((2/$CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) + Rational[1, 2] ((((-1)/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Gy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ((((2 $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-3)) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ((((2 $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Gy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ((((2 $CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-3)) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ((((2 $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) + (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) - ($CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 1][$CellContext`Gy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Gy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 1][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[1, 1][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[1, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[1, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((2 $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))), 0, 0}, {0, $CellContext`Gx[
         $CellContext`\[Rho][], 
         $CellContext`z[]]^(-1) (((Rational[1, 4]/$CellContext`Gx[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] + ((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) Derivative[1, 0][$CellContext`Gx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] + ((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + 
       Rational[-1, 2] Derivative[1, 1][$CellContext`Gx][
          $CellContext`\[Rho][], 
          $CellContext`z[]]) + ($CellContext`Gy[
          $CellContext`\[Rho][], 
          $CellContext`z[]]^(-1) ($CellContext`Gy[
           $CellContext`\[Rho][], 
           $CellContext`z[]] ($CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) - ($CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^(-1) $CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
             $CellContext`\[Rho][], 
             $CellContext`z[]])^2)) ((
         Rational[-1, 2] (((Rational[-1, 2]/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`Gy[
                $CellContext`\[Rho][], 
                $CellContext`z[]] ($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]/$CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) - ($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])^2)) (($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
               $CellContext`\[Rho][], 
               $CellContext`z[]] - $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) + ((((((
                 Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])/$CellContext`Gy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) $CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) $CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) ((($CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
               $CellContext`\[Rho][], 
               $CellContext`z[]] - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ((((-1) $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                $CellContext`\[Rho][], 
                $CellContext`z[]] + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) 
              Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                $CellContext`\[Rho][], 
                $CellContext`z[]])))) (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) + ((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) ((((-1) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + ((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) (($CellContext`ks^2 $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + (
         Rational[-1, 
           2] (((((((Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])/$CellContext`Gy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) $CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) $CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) (($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
               $CellContext`\[Rho][], 
               $CellContext`z[]] - $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]])/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((($CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
               $CellContext`\[Rho][], 
               $CellContext`z[]] - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ((((-1) $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                $CellContext`\[Rho][], 
                $CellContext`z[]] + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) 
              Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                $CellContext`\[Rho][], 
                $CellContext`z[]])))) ((($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) + Rational[1, 2] (((((-2) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-3)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^(-1) Derivative[1, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]])) + ((((($CellContext`ks^(-4)/$CellContext`Gx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`Gy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^(-2)) $CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) $CellContext`P[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
           $CellContext`\[Rho][], 
           $CellContext`z[]])) (((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (((((-1) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) (($CellContext`ks^2 $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + (
         Rational[-1, 2] (((Rational[-1, 2]/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`Gy[
                $CellContext`\[Rho][], 
                $CellContext`z[]] ($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]/$CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) - ($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])^2)) (($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
               $CellContext`\[Rho][], 
               $CellContext`z[]] - $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) + ((((((
                 Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])/$CellContext`Gy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) $CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) $CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) ((($CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
               $CellContext`\[Rho][], 
               $CellContext`z[]] - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ((((-1) $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                $CellContext`\[Rho][], 
                $CellContext`z[]] + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) 
              Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                $CellContext`\[Rho][], 
                $CellContext`z[]])))) ((($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) + ((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) (((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) + (
         Rational[-1, 
           2] (((((((Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])/$CellContext`Gy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 \
$CellContext`q) $CellContext`Gx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) $CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]) $CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`P[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) (($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
               $CellContext`\[Rho][], 
               $CellContext`z[]] - $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]])/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((($CellContext`Hx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
               $CellContext`\[Rho][], 
               $CellContext`z[]] - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ((((-1) $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
                $CellContext`\[Rho][], 
                $CellContext`z[]] + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) 
              Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
                $CellContext`\[Rho][], 
                $CellContext`z[]])))) (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (($CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]] $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (((2/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) + Rational[1, 2] ((((((-2) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-3)) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[1, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((2 $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) 
           Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) 
          Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]))) - \
((((($CellContext`ks^(-4)/$CellContext`Gx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`Gy[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) $CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]]^(-2)) $CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) $CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) $CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) $CellContext`P[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
          $CellContext`\[Rho][], 
          $CellContext`z[]])) (((((
           Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) (((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) (($CellContext`ks^2 $CellContext`P[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) + ((((
           Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]])) (((((-1) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) + (Rational[1, 2] (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (($CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]] $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (((2/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]))) (((((((
               Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
               $CellContext`\[Rho][], 
               $CellContext`z[]])/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]])/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) + (Rational[1, 2] ((($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]/$CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2)) (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + ((((((
               Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
               $CellContext`\[Rho][], 
               $CellContext`z[]])/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) ((($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) + Rational[1, 2] (((((2 $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-3)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] - (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] - (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
          $CellContext`\[Rho][], 
          $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ((((-1) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) 
          Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) - (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) Derivative[1, 1][$CellContext`Hx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[1, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((2 $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) 
           Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) 
          Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]))) + (($CellContext`Gy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]^(-1)/$CellContext`Hx[
         $CellContext`\[Rho][], 
         $CellContext`z[]]) $CellContext`Hy[
         $CellContext`\[Rho][], 
         $CellContext`z[]]) (((((
            Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (((-$CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]])/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]] $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + (((2/$CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) (($CellContext`ks^2 $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + ((((
            Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) (((-$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]])/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Gy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) + (Rational[1, 2] (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) (((((((
                Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) + (((Rational[1, 2]/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]])/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]/$CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + ($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]/$CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + (($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - (($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]] $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - (((2/$CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) + (Rational[1, 2] ((($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]] ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]/$CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2)) ((($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) + ((((((
                Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]/$CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + ($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]/$CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + (($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - (($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]] $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - (((2/$CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) + Rational[1, 2] ((((-1)/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Gy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ((((2 $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-3)) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ((((2 $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Gy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ((((2 $CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-3)) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ((((2 $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) + (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) - ($CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 1][$CellContext`Gy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Gy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 1][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[1, 1][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[1, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[1, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((2 $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))), $CellContext`Gx[
         $CellContext`\[Rho][], 
         $CellContext`z[]]^(-1) ((Rational[1, 4]/$CellContext`Gx[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]^2 + ((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + 
       Rational[-1, 2] Derivative[0, 2][$CellContext`Gx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] + ((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
           $CellContext`\[Rho][], 
           $CellContext`z[]])) + ($CellContext`Gy[
          $CellContext`\[Rho][], 
          $CellContext`z[]]^(-1) ($CellContext`Gy[
           $CellContext`\[Rho][], 
           $CellContext`z[]] ($CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) - ($CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^(-1) $CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
             $CellContext`\[Rho][], 
             $CellContext`z[]])^2)) (((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) (($CellContext`ks^2 $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + (Rational[-1, 2] ((($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) (((((((
                Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]])/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) + (Rational[-1, 2] (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]])) (((Rational[-1, 2]/$CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]] ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]/$CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2)) (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + ((((((
                Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) + Rational[1, 2] ((((-2) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-3)) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]^2 + ((2 $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 2][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^(-1) Derivative[0, 2][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) + ((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) (((-$CellContext`ks^2) $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
           $CellContext`\[Rho][], 
           $CellContext`z[]])) - ((((($CellContext`ks^(-4)/$CellContext`Gx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`Gy[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) $CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]]^(-2)) $CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) $CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) $CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) $CellContext`P[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
          $CellContext`\[Rho][], 
          $CellContext`z[]])) (((((
           Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]])) (((((-1) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) + (Rational[1, 2] (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (($CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]] $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (((2/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]))) (((((((
               Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
               $CellContext`\[Rho][], 
               $CellContext`z[]])/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]])/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) + (Rational[1, 2] ((($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]/$CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2)) (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + ((((((
               Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
               $CellContext`\[Rho][], 
               $CellContext`z[]])/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) ((($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) + Rational[1, 2] ((((2 $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-3)) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]]^2 - (((2 $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
          $CellContext`\[Rho][], 
          $CellContext`z[]] - (((2 $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ((((-1) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) 
          Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + ((2/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) - (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) Derivative[0, 2][$CellContext`Hx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[0, 2][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) (((2 $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]]^2 - ((2 $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) 
          Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 2][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[0, 2][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) + ((((
           Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
           $CellContext`\[Rho][], 
           $CellContext`z[]])) (((((-1) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]))) + \
((((($CellContext`ks^(-4)/$CellContext`Gx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`Gy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^(-2)) $CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) $CellContext`P[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
           $CellContext`\[Rho][], 
           $CellContext`z[]])) (((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) (((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) + (Rational[-1, 2] (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) (((((((
                Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]])/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) + (Rational[-1, 2] ((($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]] ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]/$CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2)) (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + ((((((
                Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) + 
       Rational[1, 2] (((((-2) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-3)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]^2 + (((2 $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (((2 $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) - ((2/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 2][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[0, 2][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) (((2 $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]]^2 - ((2 $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) 
          Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 2][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[0, 2][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) + ((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]])) (((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) + (($CellContext`Gy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]^(-1)/$CellContext`Hx[
         $CellContext`\[Rho][], 
         $CellContext`z[]]) $CellContext`Hy[
         $CellContext`\[Rho][], 
         $CellContext`z[]]) (((((
            Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) (((-$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]])/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Gy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) + (Rational[1, 2] (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) (((((((
                Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) + (((Rational[1, 2]/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]])/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]/$CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + ($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]/$CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + (($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - (($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]] $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - (((2/$CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) + (Rational[1, 2] ((($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]] ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]/$CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2)) ((($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) + ((((((
                Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]/$CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + ($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]/$CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + (($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                   $CellContext`\[Rho][], 
                   $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                  $CellContext`\[Rho][], 
                  $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - (($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]] $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - (((2/$CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) + Rational[1, 2] ((((-2)/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (((2 $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-3)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]^2 + (((2 $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Gy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (((2 $CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (((2 $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (((2 $CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-3)) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]^2 - ((((4 $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + (((4/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) + ((2/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2 - ($CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 2][$CellContext`Gy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Gy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 2][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[0, 2][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[0, 2][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[0, 2][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) (((2 $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]]^2 - ((2 $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 2][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 2][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) + ((((
            Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]])) (((-$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]])/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Gy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) - (($CellContext`ks^(-2)/$CellContext`Hx[
        $CellContext`\[Rho][], 
        $CellContext`z[]])/$CellContext`P[
       $CellContext`\[Rho][], 
       $CellContext`z[]]) (((((
           Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
           $CellContext`\[Rho][], 
           $CellContext`z[]])) (($CellContext`ks^2 $CellContext`P[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) + (((
          Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
          $CellContext`\[Rho][], 
          $CellContext`z[]])/$CellContext`P[
         $CellContext`\[Rho][], 
         $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]])^2 + 
      Rational[1, 2] (((2 $CellContext`ks^2) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`ks^2 $CellContext`P[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 2][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 2][$CellContext`P][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) + ((((
           Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
           $CellContext`\[Rho][], 
           $CellContext`z[]])) (($CellContext`ks^2 $CellContext`P[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) + (((
          Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
          $CellContext`\[Rho][], 
          $CellContext`z[]])/$CellContext`P[
         $CellContext`\[Rho][], 
         $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]])^2 + 
      Rational[1, 2] (((2 $CellContext`ks^2) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`ks^2 $CellContext`P[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[2, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[2, 0][$CellContext`P][
           $CellContext`\[Rho][], 
           $CellContext`z[]])), 0, 0}, {$CellContext`Gx[
         $CellContext`\[Rho][], 
         $CellContext`z[]]^(-1) (((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) + ((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) + (($CellContext`ks^(-2)/$CellContext`Hx[
         $CellContext`\[Rho][], 
         $CellContext`z[]])/$CellContext`P[
        $CellContext`\[Rho][], 
        $CellContext`z[]]) (((((
            Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) (((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) + (Rational[1, 2] (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) (((((((
                Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]])/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) + (Rational[1, 2] ((($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]] ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]/$CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2)) (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + ((((((
                Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) + Rational[1, 2] ((((2 $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-3)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]^2 - (((2 $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (((2 $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + ((2/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) - (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[0, 2][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 2][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) (((2 $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]]^2 - ((2 $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 2][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 2][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) + ((((
            Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]])) (((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) - ((((($CellContext`ks^(-4)/$CellContext`Gx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`Gy[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) $CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]]^(-2)) $CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) $CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) $CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) $CellContext`P[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
          $CellContext`\[Rho][], 
          $CellContext`z[]])) ((((
          Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
          $CellContext`\[Rho][], 
          $CellContext`z[]])/$CellContext`P[
         $CellContext`\[Rho][], 
         $CellContext`z[]]) (((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))^2 + ((((
           Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) ((((-1) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]])) (((-$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]])/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Gy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]/$CellContext`Hy[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] - (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] + (($CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]] $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (((2/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) + (((
          Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
          $CellContext`\[Rho][], 
          $CellContext`z[]])/$CellContext`P[
         $CellContext`\[Rho][], 
         $CellContext`z[]]) (((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))^2 + ((((
           Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) ((((-1) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]])) (((-$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]])/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Gy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]/$CellContext`Hy[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] - (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] + (($CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]] $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (((2/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]))) + (($CellContext`ks^(-2)/$CellContext`Hx[
         $CellContext`\[Rho][], 
         $CellContext`z[]])/$CellContext`P[
        $CellContext`\[Rho][], 
        $CellContext`z[]]) (((((
            Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]])) (((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) + ((((
            Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) (((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) + (Rational[1, 2] (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) (((((((
                Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + (((Rational[1, 2]/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]])/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) + (Rational[1, 2] ((($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]] ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]/$CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) - ($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2)) (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + ((((((
                Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
                $CellContext`\[Rho][], 
                $CellContext`z[]])/$CellContext`Gy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`P[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((($CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
              $CellContext`\[Rho][], 
              $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))) + Rational[1, 2] ((((2 $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-3)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]]^2 - (((2 $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (((2 $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) + ((2/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) - (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[2, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[2, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) (((2 $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]]^2 - ((2 $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[2, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[2, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))), 0, 0, $CellContext`Gx[
         $CellContext`\[Rho][], 
         $CellContext`z[]]^(-1) (((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (((-$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]])/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Gy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) + ((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (((-$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]])/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Gy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) + ($CellContext`Gy[
          $CellContext`\[Rho][], 
          $CellContext`z[]]^(-1) ($CellContext`Gy[
           $CellContext`\[Rho][], 
           $CellContext`z[]] ($CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) - ($CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^(-1) $CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
             $CellContext`\[Rho][], 
             $CellContext`z[]])^2)) ((((
           Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) (((((-1) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))^2 + ((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) (((-$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]])/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Gy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) + (((
           Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) (((((-1) $CellContext`Hx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ((((-1) $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
               $CellContext`\[Rho][], 
               $CellContext`z[]] + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) 
             Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
               $CellContext`\[Rho][], 
               $CellContext`z[]]))^2 + ((((
            Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) (((-$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]])/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Gy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) - (($CellContext`ks^(-2)/$CellContext`Hx[
        $CellContext`\[Rho][], 
        $CellContext`z[]])/$CellContext`P[
       $CellContext`\[Rho][], 
       $CellContext`z[]]) (((((
           Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]])) (((-$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]])/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Gy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]/$CellContext`Hy[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] - (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] + (($CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]] $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (((2/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) + (Rational[-1, 2] (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (($CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]] $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (((2/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]))) (((((((
               Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
               $CellContext`\[Rho][], 
               $CellContext`z[]])/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) ((($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) + (((Rational[1, 2]/$CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]])/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) + (Rational[-1, 2] ((($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]/$CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2)) ((($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) + ((((((
               Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
               $CellContext`\[Rho][], 
               $CellContext`z[]])/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) + Rational[1, 2] (((2/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (((2 $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-3)) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]^2 - (((2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Gy][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
          $CellContext`\[Rho][], 
          $CellContext`z[]] - (((2 $CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
          $CellContext`\[Rho][], 
          $CellContext`z[]] + (((2 $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (((2 $CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-3)) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]^2 + ((((4 $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) - (((4/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ((((-1) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) 
          Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) - ((2/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) $CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]])^2 + ($CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 2][$CellContext`Gy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Gy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 2][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[0, 2][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (($CellContext`Gy[
            $CellContext`\[Rho][], 
            $CellContext`z[]] $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^(-2)) Derivative[0, 2][$CellContext`Hy][
          $CellContext`\[Rho][], 
          $CellContext`z[]] - ($CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])^2) Derivative[0, 2][$CellContext`Hy][
          $CellContext`\[Rho][], 
          $CellContext`z[]] - (((2/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) (((2 $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]^2 - ((2 $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 2][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) 
          Derivative[0, 2][$CellContext`\[Omega]\[Psi]][
            $CellContext`\[Rho][], 
            $CellContext`z[]])) + ((((
           Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
           $CellContext`\[Rho][], 
           $CellContext`z[]])) (((-$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]])/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Gy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]/$CellContext`Hy[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] - (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] + (($CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]] $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (((2/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]))) - (($CellContext`ks^(-2)/$CellContext`Hx[
        $CellContext`\[Rho][], 
        $CellContext`z[]])/$CellContext`P[
       $CellContext`\[Rho][], 
       $CellContext`z[]]) (((((
           Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
           $CellContext`\[Rho][], 
           $CellContext`z[]])) (((-$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]])/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Gy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]/$CellContext`Hy[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] - (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] + (($CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]] $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (((2/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) + ((((
           Rational[1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
            $CellContext`\[Rho][], 
            $CellContext`z[]])) (((-$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]])/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Gy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]/$CellContext`Hy[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] - (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] + (($CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]] $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (((2/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) + (Rational[-1, 2] (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + (($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (($CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]] $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (((2/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]))) (((((((
               Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
               $CellContext`\[Rho][], 
               $CellContext`z[]])/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) ((($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) + (((Rational[1, 2]/$CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]])/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) + (Rational[-1, 2] ((($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]))) (((Rational[-1, 2]/$CellContext`Gy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]/$CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) - ($CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2)) ((($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) + ((((((
               Rational[-1, 2] $CellContext`ks^(-4))/$CellContext`Gx[
               $CellContext`\[Rho][], 
               $CellContext`z[]])/$CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`P[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) $CellContext`P[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]) $CellContext`P[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + ($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]/$CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + (($CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                  $CellContext`\[Rho][], 
                  $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                 $CellContext`\[Rho][], 
                 $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
             $CellContext`\[Rho][], 
             $CellContext`z[]] - (($CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]] $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] - (((2/$CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) + Rational[1, 2] (((2/$CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (((2 $CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-3)) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]^2 - (((2 $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Gy][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
          $CellContext`\[Rho][], 
          $CellContext`z[]] - (((2 $CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
          $CellContext`\[Rho][], 
          $CellContext`z[]] + (((2 $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) ($CellContext`q + $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
                $CellContext`\[Rho][], 
                $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (((2 $CellContext`Gy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-3)) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]^2 + ((((4 $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) - (((4/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ((((-1) $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) 
          Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) - ((2/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) $CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]])^2 + ($CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[2, 0][$CellContext`Gy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Gy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[2, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[2, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - (($CellContext`Gy[
            $CellContext`\[Rho][], 
            $CellContext`z[]] $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^(-2)) Derivative[2, 0][$CellContext`Hy][
          $CellContext`\[Rho][], 
          $CellContext`z[]] - ($CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])^2) Derivative[2, 0][$CellContext`Hy][
          $CellContext`\[Rho][], 
          $CellContext`z[]] - (((2/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) (((2 $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-3)) $CellContext`\[Omega]\[Psi][
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]]^2 - ((2 $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2) $CellContext`\[Omega]\[Psi][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[2, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) 
          Derivative[2, 0][$CellContext`\[Omega]\[Psi]][
            $CellContext`\[Rho][], 
            $CellContext`z[]]))), 0}, {0, 0, 0, 0, ($CellContext`Gy[
          $CellContext`\[Rho][], 
          $CellContext`z[]]^(-1) ($CellContext`Gy[
           $CellContext`\[Rho][], 
           $CellContext`z[]] ($CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) - ($CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^(-1) $CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
             $CellContext`\[Rho][], 
             $CellContext`z[]])^2)) (((((
            Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ((((-1) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]]) + ((((
            Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) ((((-1) $CellContext`Hx[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-2)) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]])) - (($CellContext`ks^(-2)/$CellContext`Hx[
        $CellContext`\[Rho][], 
        $CellContext`z[]])/$CellContext`P[
       $CellContext`\[Rho][], 
       $CellContext`z[]]) ((Rational[-1, 4]/$CellContext`Gx[
         $CellContext`\[Rho][], 
         $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
          $CellContext`\[Rho][], 
          $CellContext`z[]]^2 + ((((
           Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
          $CellContext`\[Rho][], 
          $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) + 
      Rational[1, 2] Derivative[0, 2][$CellContext`Gx][
         $CellContext`\[Rho][], 
         $CellContext`z[]] + ((((
           Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
          $CellContext`\[Rho][], 
          $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
          $CellContext`\[Rho][], 
          $CellContext`z[]])) + ((((((
            2 $CellContext`ks^(-4))/$CellContext`Gx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`Gy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^(-2)) $CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]^(-2)) ((((($CellContext`ks^4 $CellContext`q) \
$CellContext`Gx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) $CellContext`P[
            $CellContext`\[Rho][], 
            $CellContext`z[]]^2 + ((($CellContext`ks^4 $CellContext`Gx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`P[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^2) $CellContext`\[Omega]\[Psi][
           $CellContext`\[Rho][], 
           $CellContext`z[]])) (((((
            Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[0, 1][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) + ((((
            Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]])/$CellContext`P[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) (((((-1) $CellContext`Hx[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hx][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])) Derivative[1, 0][$CellContext`Hy][
            $CellContext`\[Rho][], 
            $CellContext`z[]] + ($CellContext`Hx[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ((((-1) $CellContext`Hy[
                 $CellContext`\[Rho][], 
                 $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
              $CellContext`\[Rho][], 
              $CellContext`z[]] + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) 
            Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
              $CellContext`\[Rho][], 
              $CellContext`z[]]))) - (($CellContext`Gy[
          $CellContext`\[Rho][], 
          $CellContext`z[]]^(-1)/$CellContext`Hx[
        $CellContext`\[Rho][], 
        $CellContext`z[]]) $CellContext`Hy[
        $CellContext`\[Rho][], 
        $CellContext`z[]]) (((((
           Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
          $CellContext`\[Rho][], 
          $CellContext`z[]]) (((-$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]])/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 1][$CellContext`Gy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Gy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]/$CellContext`Hy[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] - (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] + (($CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]] $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2)) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[0, 1][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (((2/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[0, 1][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[0, 1][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]])) + ((((
           Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
          $CellContext`\[Rho][], 
          $CellContext`z[]]) (((-$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]])/$CellContext`Hy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[1, 0][$CellContext`Gy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`Gy[
           $CellContext`\[Rho][], 
           $CellContext`z[]]/$CellContext`Hy[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] - (($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2) $CellContext`Hy[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hx][
          $CellContext`\[Rho][], 
          $CellContext`z[]] + (($CellContext`Gy[
             $CellContext`\[Rho][], 
             $CellContext`z[]] $CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-2)) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]^(-1) ($CellContext`q + $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
               $CellContext`\[Rho][], 
               $CellContext`z[]])^2) Derivative[1, 0][$CellContext`Hy][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + (((2/$CellContext`Hx[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) $CellContext`Hy[
             $CellContext`\[Rho][], 
             $CellContext`z[]]) ($CellContext`q + $CellContext`Hy[
               $CellContext`\[Rho][], 
               $CellContext`z[]]^(-1) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]])) ((((-1) $CellContext`Hy[
                $CellContext`\[Rho][], 
                $CellContext`z[]]^(-2)) $CellContext`\[Omega]\[Psi][
              $CellContext`\[Rho][], 
              $CellContext`z[]]) Derivative[1, 0][$CellContext`Hy][
             $CellContext`\[Rho][], 
             $CellContext`z[]] + $CellContext`Hy[
              $CellContext`\[Rho][], 
              $CellContext`z[]]^(-1) 
           Derivative[1, 0][$CellContext`\[Omega]\[Psi]][
             $CellContext`\[Rho][], 
             $CellContext`z[]]))) - (($CellContext`ks^(-2)/$CellContext`Hx[
        $CellContext`\[Rho][], 
        $CellContext`z[]])/$CellContext`P[
       $CellContext`\[Rho][], 
       $CellContext`z[]]) (((((
           Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) Derivative[0, 1][$CellContext`Gx][
          $CellContext`\[Rho][], 
          $CellContext`z[]]) (((-$CellContext`ks^2) $CellContext`P[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[0, 1][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] - ($CellContext`ks^2 $CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]]) Derivative[0, 1][$CellContext`P][
          $CellContext`\[Rho][], 
          $CellContext`z[]]) + (Rational[-1, 4]/$CellContext`Gx[
         $CellContext`\[Rho][], 
         $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
          $CellContext`\[Rho][], 
          $CellContext`z[]]^2 + ((((
           Rational[-1, 4] $CellContext`ks^(-2))/$CellContext`Hx[
           $CellContext`\[Rho][], 
           $CellContext`z[]])/$CellContext`P[
          $CellContext`\[Rho][], 
          $CellContext`z[]]) Derivative[1, 0][$CellContext`Gx][
          $CellContext`\[Rho][], 
          $CellContext`z[]]) (($CellContext`ks^2 $CellContext`P[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`Hx][
           $CellContext`\[Rho][], 
           $CellContext`z[]] + ($CellContext`ks^2 $CellContext`Hx[
            $CellContext`\[Rho][], 
            $CellContext`z[]]) Derivative[1, 0][$CellContext`P][
           $CellContext`\[Rho][], 
           $CellContext`z[]]) + 
      Rational[1, 2] Derivative[2, 0][$CellContext`Gx][
         $CellContext`\[Rho][], 
         $CellContext`z[]])}}, {-$CellContext`BS, -$CellContext`BS}, 
   0][-$CellContext`a, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.728677035114276*^9},
 CellLabel->"Out[91]=",
 CellID->1450722271,ExpressionUUID->"15d46692-20d7-4c91-9daa-fabb583c3f52"]
}, Open  ]]
}, Open  ]],

Cell["Now we need to show that the rest are also zero:", "MathCaption",
 CellID->1490901547,ExpressionUUID->"70ff9298-8303-4242-95ac-2d60bea7aacd"],

Cell[CellGroupData[{

Cell["This is easy:", "MathCaption",
 CellID->743107233,ExpressionUUID->"c9a0de83-7a7b-4f07-84b6-3a2335884936"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Numerator", "[", 
  RowBox[{"ricci", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"4", ",", 
      RowBox[{"-", "BS"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{"4", ",", 
      RowBox[{"-", "BS"}]}], "}"}]}], "]"}], "]"}]], "Input",
 CellLabel->"In[92]:=",
 CellID->1785857726,ExpressionUUID->"3e154b54-1ea6-4ab9-b8f1-389a94535ab5"],

Cell[BoxData[
 RowBox[{
  FractionBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      FractionBox[
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["Gy",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Gy,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["\[Rho]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Rho][],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z[],
           Editable->False]}], "]"}], " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["Hx",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Hx,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["\[Rho]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Rho][],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z[],
           Editable->False]}], "]"}]}], 
       RowBox[{
        InterpretationBox[
         StyleBox["Hy",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Hy,
         Editable->False], "[", 
        RowBox[{
         InterpretationBox[
          StyleBox["\[Rho]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Rho][],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z[],
          Editable->False]}], "]"}]], "-", 
      FractionBox[
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["Hy",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Hy,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["\[Rho]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Rho][],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z[],
           Editable->False]}], "]"}], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox["q",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`q,
            Editable->False], "+", 
           FractionBox[
            RowBox[{
             InterpretationBox[
              StyleBox["\[Omega]\[Psi]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Omega]\[Psi],
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["\[Rho]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Rho][],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}], 
            RowBox[{
             InterpretationBox[
              StyleBox["Hy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Hy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["\[Rho]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Rho][],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}]]}], ")"}], "2"]}], 
       RowBox[{
        InterpretationBox[
         StyleBox["Hx",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Hx,
         Editable->False], "[", 
        RowBox[{
         InterpretationBox[
          StyleBox["\[Rho]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Rho][],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z[],
          Editable->False]}], "]"}]]}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", 
       FractionBox[
        RowBox[{
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["Gx",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Gx,
            Editable->False], 
           TagBox[
            RowBox[{"(", 
             RowBox[{"0", ",", "1"}], ")"}],
            Derivative],
           MultilineFunction->None], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[Rho]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Rho][],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z[],
            Editable->False]}], "]"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            FractionBox[
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["Hy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Hy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Rho]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Rho][],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["Hx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Hx,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"0", ",", "1"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Rho]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Rho][],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Hx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Hx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Rho]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Rho][],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], "2"]]}], "+", 
           FractionBox[
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["Hy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Hy,
               Editable->False], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"0", ",", "1"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["\[Rho]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Rho][],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}], 
            RowBox[{
             InterpretationBox[
              StyleBox["Hx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Hx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["\[Rho]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Rho][],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}]]}], ")"}]}], 
        RowBox[{"4", " ", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["ks",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`ks,
           Editable->False], "2"], " ", 
         RowBox[{
          InterpretationBox[
           StyleBox["Hx",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Hx,
           Editable->False], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[Rho]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Rho][],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z[],
            Editable->False]}], "]"}], " ", 
         RowBox[{
          InterpretationBox[
           StyleBox["P",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`P,
           Editable->False], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[Rho]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Rho][],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z[],
            Editable->False]}], "]"}]}]]}], "-", 
      FractionBox[
       RowBox[{
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["Gx",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Gx,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"1", ",", "0"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["\[Rho]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Rho][],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z[],
           Editable->False]}], "]"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", 
           FractionBox[
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox["Hy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Hy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Rho]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Rho][],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}], " ", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["Hx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Hx,
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"1", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Rho]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Rho][],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]}], 
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["Hx",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Hx,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Rho]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Rho][],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}], "2"]]}], "+", 
          FractionBox[
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["Hy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Hy,
              Editable->False], 
             TagBox[
              RowBox[{"(", 
               RowBox[{"1", ",", "0"}], ")"}],
              Derivative],
             MultilineFunction->None], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["\[Rho]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Rho][],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["z",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`z[],
              Editable->False]}], "]"}], 
           RowBox[{
            InterpretationBox[
             StyleBox["Hx",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Hx,
             Editable->False], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["\[Rho]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Rho][],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["z",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`z[],
              Editable->False]}], "]"}]]}], ")"}]}], 
       RowBox[{"4", " ", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["ks",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`ks,
          Editable->False], "2"], " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["Hx",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Hx,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["\[Rho]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Rho][],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z[],
           Editable->False]}], "]"}], " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["P",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`P,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["\[Rho]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Rho][],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z[],
           Editable->False]}], "]"}]}]]}], ")"}]}], 
   RowBox[{
    InterpretationBox[
     StyleBox["Gy",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Gy,
     Editable->False], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Rho]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho][],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["z",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`z[],
      Editable->False]}], "]"}]], "-", 
  FractionBox[
   RowBox[{
    RowBox[{"-", 
     FractionBox[
      SuperscriptBox[
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["Gx",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Gx,
          Editable->False], 
         TagBox[
          RowBox[{"(", 
           RowBox[{"0", ",", "1"}], ")"}],
          Derivative],
         MultilineFunction->None], "[", 
        RowBox[{
         InterpretationBox[
          StyleBox["\[Rho]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Rho][],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z[],
          Editable->False]}], "]"}], "2"], 
      RowBox[{"4", " ", 
       RowBox[{
        InterpretationBox[
         StyleBox["Gx",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Gx,
         Editable->False], "[", 
        RowBox[{
         InterpretationBox[
          StyleBox["\[Rho]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Rho][],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z[],
          Editable->False]}], "]"}]}]]}], "-", 
    FractionBox[
     RowBox[{
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["Gx",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Gx,
         Editable->False], 
        TagBox[
         RowBox[{"(", 
          RowBox[{"0", ",", "1"}], ")"}],
         Derivative],
        MultilineFunction->None], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho][],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["z",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`z[],
         Editable->False]}], "]"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["ks",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`ks,
           Editable->False], "2"], " ", 
         RowBox[{
          InterpretationBox[
           StyleBox["P",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`P,
           Editable->False], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[Rho]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Rho][],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z[],
            Editable->False]}], "]"}], " ", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["Hx",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Hx,
            Editable->False], 
           TagBox[
            RowBox[{"(", 
             RowBox[{"0", ",", "1"}], ")"}],
            Derivative],
           MultilineFunction->None], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[Rho]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Rho][],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z[],
            Editable->False]}], "]"}]}], "+", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["ks",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`ks,
           Editable->False], "2"], " ", 
         RowBox[{
          InterpretationBox[
           StyleBox["Hx",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Hx,
           Editable->False], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[Rho]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Rho][],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z[],
            Editable->False]}], "]"}], " ", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["P",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`P,
            Editable->False], 
           TagBox[
            RowBox[{"(", 
             RowBox[{"0", ",", "1"}], ")"}],
            Derivative],
           MultilineFunction->None], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[Rho]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Rho][],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z[],
            Editable->False]}], "]"}]}]}], ")"}]}], 
     RowBox[{"4", " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["ks",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ks,
        Editable->False], "2"], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["Hx",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Hx,
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho][],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["z",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`z[],
         Editable->False]}], "]"}], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["P",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`P,
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho][],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["z",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`z[],
         Editable->False]}], "]"}]}]], "+", 
    RowBox[{
     FractionBox["1", "2"], " ", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["Gx",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Gx,
        Editable->False], 
       TagBox[
        RowBox[{"(", 
         RowBox[{"0", ",", "2"}], ")"}],
        Derivative],
       MultilineFunction->None], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Rho]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Rho][],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["z",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`z[],
        Editable->False]}], "]"}]}], "-", 
    FractionBox[
     RowBox[{
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["Gx",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Gx,
         Editable->False], 
        TagBox[
         RowBox[{"(", 
          RowBox[{"1", ",", "0"}], ")"}],
         Derivative],
        MultilineFunction->None], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho][],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["z",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`z[],
         Editable->False]}], "]"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["ks",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`ks,
            Editable->False], "2"]}], " ", 
         RowBox[{
          InterpretationBox[
           StyleBox["P",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`P,
           Editable->False], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[Rho]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Rho][],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z[],
            Editable->False]}], "]"}], " ", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["Hx",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Hx,
            Editable->False], 
           TagBox[
            RowBox[{"(", 
             RowBox[{"1", ",", "0"}], ")"}],
            Derivative],
           MultilineFunction->None], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[Rho]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Rho][],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z[],
            Editable->False]}], "]"}]}], "-", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["ks",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`ks,
           Editable->False], "2"], " ", 
         RowBox[{
          InterpretationBox[
           StyleBox["Hx",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Hx,
           Editable->False], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[Rho]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Rho][],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z[],
            Editable->False]}], "]"}], " ", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["P",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`P,
            Editable->False], 
           TagBox[
            RowBox[{"(", 
             RowBox[{"1", ",", "0"}], ")"}],
            Derivative],
           MultilineFunction->None], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[Rho]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Rho][],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z[],
            Editable->False]}], "]"}]}]}], ")"}]}], 
     RowBox[{"4", " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["ks",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ks,
        Editable->False], "2"], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["Hx",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Hx,
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho][],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["z",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`z[],
         Editable->False]}], "]"}], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["P",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`P,
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho][],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["z",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`z[],
         Editable->False]}], "]"}]}]]}], 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["ks",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ks,
      Editable->False], "2"], " ", 
    RowBox[{
     InterpretationBox[
      StyleBox["Hx",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Hx,
      Editable->False], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho][],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["z",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`z[],
       Editable->False]}], "]"}], " ", 
    RowBox[{
     InterpretationBox[
      StyleBox["P",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`P,
      Editable->False], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho][],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["z",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`z[],
       Editable->False]}], "]"}]}]], "+", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{"2", " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["ks",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`ks,
          Editable->False], "4"], " ", 
        InterpretationBox[
         StyleBox["q",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`q,
         Editable->False], " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["Gx",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Gx,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["\[Rho]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Rho][],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z[],
           Editable->False]}], "]"}], " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["Hx",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Hx,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["\[Rho]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Rho][],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z[],
           Editable->False]}], "]"}], " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["Hy",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Hy,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["\[Rho]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Rho][],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z[],
           Editable->False]}], "]"}], " ", 
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["P",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`P,
           Editable->False], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[Rho]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Rho][],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z[],
            Editable->False]}], "]"}], "2"]}], "+", 
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["ks",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`ks,
          Editable->False], "4"], " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["Gx",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Gx,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["\[Rho]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Rho][],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z[],
           Editable->False]}], "]"}], " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["Hx",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Hx,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["\[Rho]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Rho][],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z[],
           Editable->False]}], "]"}], " ", 
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["P",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`P,
           Editable->False], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[Rho]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Rho][],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z[],
            Editable->False]}], "]"}], "2"], " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["\[Omega]\[Psi]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Omega]\[Psi],
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["\[Rho]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Rho][],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z[],
           Editable->False]}], "]"}]}]}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", 
        FractionBox[
         RowBox[{
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["Gx",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Gx,
             Editable->False], 
            TagBox[
             RowBox[{"(", 
              RowBox[{"0", ",", "1"}], ")"}],
             Derivative],
            MultilineFunction->None], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Rho]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho][],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["z",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`z[],
             Editable->False]}], "]"}], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", 
             FractionBox[
              RowBox[{
               RowBox[{
                InterpretationBox[
                 StyleBox["Hy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Hy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["\[Rho]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Rho][],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["z",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`z[],
                  Editable->False]}], "]"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["q",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`q,
                  Editable->False], "+", 
                 FractionBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["\[Omega]\[Psi]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Omega]\[Psi],
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Rho][],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`z[],
                    Editable->False]}], "]"}], 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["Hy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Hy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Rho][],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`z[],
                    Editable->False]}], "]"}]]}], ")"}], " ", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["Hx",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Hx,
                  Editable->False], 
                 TagBox[
                  RowBox[{"(", 
                   RowBox[{"0", ",", "1"}], ")"}],
                  Derivative],
                 MultilineFunction->None], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["\[Rho]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Rho][],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["z",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`z[],
                  Editable->False]}], "]"}]}], 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["Hx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Hx,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["\[Rho]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Rho][],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["z",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`z[],
                  Editable->False]}], "]"}], "2"]]}], "+", 
            FractionBox[
             RowBox[{
              RowBox[{"(", 
               RowBox[{
                InterpretationBox[
                 StyleBox["q",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`q,
                 Editable->False], "+", 
                FractionBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["\[Omega]\[Psi]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Omega]\[Psi],
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Rho][],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`z[],
                    Editable->False]}], "]"}], 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Hy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Hy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Rho][],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`z[],
                    Editable->False]}], "]"}]]}], ")"}], " ", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["Hy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Hy,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"0", ",", "1"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Rho]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Rho][],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], 
             RowBox[{
              InterpretationBox[
               StyleBox["Hx",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Hx,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Rho]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Rho][],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]], "+", 
            FractionBox[
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["Hy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Hy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Rho]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Rho][],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", 
                 FractionBox[
                  RowBox[{
                   RowBox[{
                    InterpretationBox[
                    StyleBox["\[Omega]\[Psi]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Omega]\[Psi],
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Rho][],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`z[],
                    Editable->False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["Hy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Hy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Rho][],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`z[],
                    Editable->False]}], "]"}]}], 
                  SuperscriptBox[
                   RowBox[{
                    InterpretationBox[
                    StyleBox["Hy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Hy,
                    Editable->False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Rho][],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`z[],
                    Editable->False]}], "]"}], "2"]]}], "+", 
                FractionBox[
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["\[Omega]\[Psi]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Omega]\[Psi],
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Rho][],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`z[],
                    Editable->False]}], "]"}], 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Hy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Hy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Rho][],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`z[],
                    Editable->False]}], "]"}]]}], ")"}]}], 
             RowBox[{
              InterpretationBox[
               StyleBox["Hx",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Hx,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Rho]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Rho][],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]]}], ")"}]}], 
         RowBox[{"4", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["ks",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`ks,
            Editable->False], "2"], " ", 
          RowBox[{
           InterpretationBox[
            StyleBox["Hx",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Hx,
            Editable->False], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Rho]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho][],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["z",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`z[],
             Editable->False]}], "]"}], " ", 
          RowBox[{
           InterpretationBox[
            StyleBox["P",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`P,
            Editable->False], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Rho]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho][],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["z",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`z[],
             Editable->False]}], "]"}]}]]}], "-", 
       FractionBox[
        RowBox[{
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["Gx",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Gx,
            Editable->False], 
           TagBox[
            RowBox[{"(", 
             RowBox[{"1", ",", "0"}], ")"}],
            Derivative],
           MultilineFunction->None], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[Rho]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Rho][],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z[],
            Editable->False]}], "]"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            FractionBox[
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["Hy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Hy,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Rho]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Rho][],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{"(", 
               RowBox[{
                InterpretationBox[
                 StyleBox["q",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`q,
                 Editable->False], "+", 
                FractionBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["\[Omega]\[Psi]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Omega]\[Psi],
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Rho][],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`z[],
                    Editable->False]}], "]"}], 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Hy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Hy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Rho][],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`z[],
                    Editable->False]}], "]"}]]}], ")"}], " ", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["Hx",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Hx,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Rho]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Rho][],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Hx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Hx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Rho]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Rho][],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], "2"]]}], "+", 
           FractionBox[
            RowBox[{
             RowBox[{"(", 
              RowBox[{
               InterpretationBox[
                StyleBox["q",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`q,
                Editable->False], "+", 
               FractionBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["\[Omega]\[Psi]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Omega]\[Psi],
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Rho][],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`z[],
                   Editable->False]}], "]"}], 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Hy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Hy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Rho][],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`z[],
                   Editable->False]}], "]"}]]}], ")"}], " ", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["Hy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Hy,
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"1", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Rho]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Rho][],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]}], 
            RowBox[{
             InterpretationBox[
              StyleBox["Hx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Hx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["\[Rho]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Rho][],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}]], "+", 
           FractionBox[
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox["Hy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Hy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Rho]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Rho][],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", 
                FractionBox[
                 RowBox[{
                  RowBox[{
                   InterpretationBox[
                    StyleBox["\[Omega]\[Psi]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Omega]\[Psi],
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Rho][],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`z[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["Hy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Hy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Rho][],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`z[],
                    Editable->False]}], "]"}]}], 
                 SuperscriptBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["Hy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Hy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Rho][],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`z[],
                    Editable->False]}], "]"}], "2"]]}], "+", 
               FractionBox[
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["\[Omega]\[Psi]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Omega]\[Psi],
                   Editable->False], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Rho][],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`z[],
                   Editable->False]}], "]"}], 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Hy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Hy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Rho][],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`z[],
                   Editable->False]}], "]"}]]}], ")"}]}], 
            RowBox[{
             InterpretationBox[
              StyleBox["Hx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Hx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["\[Rho]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Rho][],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}]]}], ")"}]}], 
        RowBox[{"4", " ", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["ks",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`ks,
           Editable->False], "2"], " ", 
         RowBox[{
          InterpretationBox[
           StyleBox["Hx",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Hx,
           Editable->False], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[Rho]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Rho][],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z[],
            Editable->False]}], "]"}], " ", 
         RowBox[{
          InterpretationBox[
           StyleBox["P",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`P,
           Editable->False], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[Rho]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Rho][],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z[],
            Editable->False]}], "]"}]}]]}], ")"}]}], ")"}], "/", 
   RowBox[{"(", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["ks",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ks,
       Editable->False], "4"], " ", 
     RowBox[{
      InterpretationBox[
       StyleBox["Gx",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Gx,
       Editable->False], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Rho]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Rho][],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["z",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`z[],
        Editable->False]}], "]"}], " ", 
     RowBox[{
      InterpretationBox[
       StyleBox["Gy",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Gy,
       Editable->False], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Rho]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Rho][],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["z",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`z[],
        Editable->False]}], "]"}], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Hx",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Hx,
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho][],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["z",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`z[],
         Editable->False]}], "]"}], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["P",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`P,
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho][],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["z",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`z[],
         Editable->False]}], "]"}], "2"]}], ")"}]}], "-", 
  RowBox[{
   FractionBox["1", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["Gy",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Gy,
       Editable->False], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Rho]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Rho][],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["z",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`z[],
        Editable->False]}], "]"}], " ", 
     RowBox[{
      InterpretationBox[
       StyleBox["Hx",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Hx,
       Editable->False], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Rho]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Rho][],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["z",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`z[],
        Editable->False]}], "]"}]}]], 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox["Hy",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Hy,
      Editable->False], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho][],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["z",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`z[],
       Editable->False]}], "]"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", 
       RowBox[{
        FractionBox["1", 
         RowBox[{"4", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["ks",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`ks,
            Editable->False], "2"], " ", 
          RowBox[{
           InterpretationBox[
            StyleBox["Hx",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Hx,
            Editable->False], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Rho]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho][],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["z",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`z[],
             Editable->False]}], "]"}], " ", 
          RowBox[{
           InterpretationBox[
            StyleBox["P",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`P,
            Editable->False], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Rho]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho][],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["z",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`z[],
             Editable->False]}], "]"}]}]], 
        RowBox[{
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["Gx",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Gx,
            Editable->False], 
           TagBox[
            RowBox[{"(", 
             RowBox[{"0", ",", "1"}], ")"}],
            Derivative],
           MultilineFunction->None], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[Rho]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Rho][],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z[],
            Editable->False]}], "]"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            FractionBox[
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["Hx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Hx,
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Rho]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Rho][],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}], " ", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["Gy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Gy,
                 Editable->False], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"0", ",", "1"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Rho]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Rho][],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["z",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`z[],
                 Editable->False]}], "]"}]}], 
             RowBox[{
              InterpretationBox[
               StyleBox["Hy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Hy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Rho]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Rho][],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]]}], "-", 
           FractionBox[
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox["Gy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Gy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Rho]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Rho][],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}], " ", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["Hx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Hx,
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"0", ",", "1"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Rho]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Rho][],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]}], 
            RowBox[{
             InterpretationBox[
              StyleBox["Hy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Hy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["\[Rho]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Rho][],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}]], "-", 
           FractionBox[
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox["Hy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Hy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Rho]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Rho][],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}], " ", 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                InterpretationBox[
                 StyleBox["q",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`q,
                 Editable->False], "+", 
                FractionBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["\[Omega]\[Psi]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Omega]\[Psi],
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Rho][],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`z[],
                    Editable->False]}], "]"}], 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Hy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Hy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Rho][],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`z[],
                    Editable->False]}], "]"}]]}], ")"}], "2"], " ", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["Hx",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Hx,
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"0", ",", "1"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Rho]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Rho][],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]}], 
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["Hx",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Hx,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Rho]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Rho][],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}], "2"]], "+", 
           FractionBox[
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox["Gy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Gy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Rho]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Rho][],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Hx",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Hx,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Rho]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Rho][],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}], " ", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["Hy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Hy,
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"0", ",", "1"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Rho]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Rho][],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]}], 
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["Hy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Hy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Rho]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Rho][],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}], "2"]], "+", 
           FractionBox[
            RowBox[{
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                InterpretationBox[
                 StyleBox["q",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`q,
                 Editable->False], "+", 
                FractionBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["\[Omega]\[Psi]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Omega]\[Psi],
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Rho][],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`z[],
                    Editable->False]}], "]"}], 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Hy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Hy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Rho][],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`z[],
                    Editable->False]}], "]"}]]}], ")"}], "2"], " ", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["Hy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Hy,
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"0", ",", "1"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Rho]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Rho][],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]}], 
            RowBox[{
             InterpretationBox[
              StyleBox["Hx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Hx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["\[Rho]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Rho][],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}]], "+", 
           FractionBox[
            RowBox[{"2", " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Hy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Hy,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Rho]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Rho][],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}], " ", 
             RowBox[{"(", 
              RowBox[{
               InterpretationBox[
                StyleBox["q",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`q,
                Editable->False], "+", 
               FractionBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["\[Omega]\[Psi]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Omega]\[Psi],
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Rho][],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`z[],
                   Editable->False]}], "]"}], 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Hy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Hy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Rho][],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`z[],
                   Editable->False]}], "]"}]]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", 
                FractionBox[
                 RowBox[{
                  RowBox[{
                   InterpretationBox[
                    StyleBox["\[Omega]\[Psi]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Omega]\[Psi],
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Rho][],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`z[],
                    Editable->False]}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox["Hy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Hy,
                    Editable->False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Rho][],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`z[],
                    Editable->False]}], "]"}]}], 
                 SuperscriptBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["Hy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Hy,
                    Editable->False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Rho][],
                    Editable->False], ",", 
                    InterpretationBox[
                    StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`z[],
                    Editable->False]}], "]"}], "2"]]}], "+", 
               FractionBox[
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["\[Omega]\[Psi]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Omega]\[Psi],
                   Editable->False], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Rho][],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`z[],
                   Editable->False]}], "]"}], 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Hy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Hy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Rho][],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`z[],
                   Editable->False]}], "]"}]]}], ")"}]}], 
            RowBox[{
             InterpretationBox[
              StyleBox["Hx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Hx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["\[Rho]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Rho][],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}]]}], ")"}]}]}]}], "-", 
      RowBox[{
       FractionBox["1", 
        RowBox[{"4", " ", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["ks",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`ks,
           Editable->False], "2"], " ", 
         RowBox[{
          InterpretationBox[
           StyleBox["Hx",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Hx,
           Editable->False], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[Rho]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Rho][],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z[],
            Editable->False]}], "]"}], " ", 
         RowBox[{
          InterpretationBox[
           StyleBox["P",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`P,
           Editable->False], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[Rho]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Rho][],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z[],
            Editable->False]}], "]"}]}]], 
       RowBox[{
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["Gx",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Gx,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"1", ",", "0"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["\[Rho]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Rho][],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z[],
           Editable->False]}], "]"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", 
           FractionBox[
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox["Hx",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Hx,
               Editable->False], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Rho]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Rho][],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}], " ", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["Gy",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Gy,
                Editable->False], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"1", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Rho]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Rho][],
                Editable->False], ",", 
               InterpretationBox[
                StyleBox["z",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`z[],
                Editable->False]}], "]"}]}], 
            RowBox[{
             InterpretationBox[
              StyleBox["Hy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Hy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["\[Rho]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Rho][],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}]]}], "-", 
          FractionBox[
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["Gy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Gy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["\[Rho]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Rho][],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}], " ", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["Hx",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Hx,
               Editable->False], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"1", ",", "0"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["\[Rho]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Rho][],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}]}], 
           RowBox[{
            InterpretationBox[
             StyleBox["Hy",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Hy,
             Editable->False], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["\[Rho]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Rho][],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["z",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`z[],
              Editable->False]}], "]"}]], "-", 
          FractionBox[
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["Hy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Hy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["\[Rho]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Rho][],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}], " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               InterpretationBox[
                StyleBox["q",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`q,
                Editable->False], "+", 
               FractionBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["\[Omega]\[Psi]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Omega]\[Psi],
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Rho][],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`z[],
                   Editable->False]}], "]"}], 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Hy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Hy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Rho][],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`z[],
                   Editable->False]}], "]"}]]}], ")"}], "2"], " ", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["Hx",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Hx,
               Editable->False], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"1", ",", "0"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["\[Rho]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Rho][],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}]}], 
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["Hx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Hx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["\[Rho]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Rho][],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}], "2"]], "+", 
          FractionBox[
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["Gy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Gy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["\[Rho]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Rho][],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["Hx",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Hx,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["\[Rho]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Rho][],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}], " ", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["Hy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Hy,
               Editable->False], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"1", ",", "0"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["\[Rho]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Rho][],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}]}], 
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["Hy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Hy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["\[Rho]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Rho][],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}], "2"]], "+", 
          FractionBox[
           RowBox[{
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               InterpretationBox[
                StyleBox["q",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`q,
                Editable->False], "+", 
               FractionBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["\[Omega]\[Psi]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Omega]\[Psi],
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Rho][],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`z[],
                   Editable->False]}], "]"}], 
                RowBox[{
                 InterpretationBox[
                  StyleBox["Hy",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Hy,
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Rho][],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`z[],
                   Editable->False]}], "]"}]]}], ")"}], "2"], " ", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["Hy",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Hy,
               Editable->False], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"1", ",", "0"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["\[Rho]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Rho][],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}]}], 
           RowBox[{
            InterpretationBox[
             StyleBox["Hx",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Hx,
             Editable->False], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["\[Rho]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Rho][],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["z",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`z[],
              Editable->False]}], "]"}]], "+", 
          FractionBox[
           RowBox[{"2", " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["Hy",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Hy,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["\[Rho]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Rho][],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]}], "]"}], " ", 
            RowBox[{"(", 
             RowBox[{
              InterpretationBox[
               StyleBox["q",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`q,
               Editable->False], "+", 
              FractionBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Omega]\[Psi]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Omega]\[Psi],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["\[Rho]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Rho][],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["z",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`z[],
                  Editable->False]}], "]"}], 
               RowBox[{
                InterpretationBox[
                 StyleBox["Hy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Hy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["\[Rho]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Rho][],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["z",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`z[],
                  Editable->False]}], "]"}]]}], ")"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", 
               FractionBox[
                RowBox[{
                 RowBox[{
                  InterpretationBox[
                   StyleBox["\[Omega]\[Psi]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`\[Omega]\[Psi],
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Rho][],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`z[],
                    Editable->False]}], "]"}], " ", 
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["Hy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Hy,
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Rho][],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`z[],
                    Editable->False]}], "]"}]}], 
                SuperscriptBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["Hy",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Hy,
                   Editable->False], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["\[Rho]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Rho][],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["z",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`z[],
                    Editable->False]}], "]"}], "2"]]}], "+", 
              FractionBox[
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["\[Omega]\[Psi]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Omega]\[Psi],
                  Editable->False], 
                 TagBox[
                  RowBox[{"(", 
                   RowBox[{"1", ",", "0"}], ")"}],
                  Derivative],
                 MultilineFunction->None], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["\[Rho]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Rho][],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["z",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`z[],
                  Editable->False]}], "]"}], 
               RowBox[{
                InterpretationBox[
                 StyleBox["Hy",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Hy,
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["\[Rho]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Rho][],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["z",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`z[],
                  Editable->False]}], "]"}]]}], ")"}]}], 
           RowBox[{
            InterpretationBox[
             StyleBox["Hx",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Hx,
             Editable->False], "[", 
            RowBox[{
             InterpretationBox[
              StyleBox["\[Rho]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Rho][],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox["z",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`z[],
              Editable->False]}], "]"}]]}], ")"}]}]}]}], ")"}]}]}], "-", 
  FractionBox[
   RowBox[{
    RowBox[{"-", 
     FractionBox[
      RowBox[{
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["Gx",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Gx,
          Editable->False], 
         TagBox[
          RowBox[{"(", 
           RowBox[{"0", ",", "1"}], ")"}],
          Derivative],
         MultilineFunction->None], "[", 
        RowBox[{
         InterpretationBox[
          StyleBox["\[Rho]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Rho][],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z[],
          Editable->False]}], "]"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["ks",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`ks,
             Editable->False], "2"]}], " ", 
          RowBox[{
           InterpretationBox[
            StyleBox["P",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`P,
            Editable->False], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Rho]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho][],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["z",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`z[],
             Editable->False]}], "]"}], " ", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["Hx",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Hx,
             Editable->False], 
            TagBox[
             RowBox[{"(", 
              RowBox[{"0", ",", "1"}], ")"}],
             Derivative],
            MultilineFunction->None], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Rho]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho][],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["z",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`z[],
             Editable->False]}], "]"}]}], "-", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["ks",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`ks,
            Editable->False], "2"], " ", 
          RowBox[{
           InterpretationBox[
            StyleBox["Hx",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Hx,
            Editable->False], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Rho]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho][],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["z",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`z[],
             Editable->False]}], "]"}], " ", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["P",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`P,
             Editable->False], 
            TagBox[
             RowBox[{"(", 
              RowBox[{"0", ",", "1"}], ")"}],
             Derivative],
            MultilineFunction->None], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Rho]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho][],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["z",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`z[],
             Editable->False]}], "]"}]}]}], ")"}]}], 
      RowBox[{"4", " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["ks",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`ks,
         Editable->False], "2"], " ", 
       RowBox[{
        InterpretationBox[
         StyleBox["Hx",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Hx,
         Editable->False], "[", 
        RowBox[{
         InterpretationBox[
          StyleBox["\[Rho]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Rho][],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z[],
          Editable->False]}], "]"}], " ", 
       RowBox[{
        InterpretationBox[
         StyleBox["P",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`P,
         Editable->False], "[", 
        RowBox[{
         InterpretationBox[
          StyleBox["\[Rho]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Rho][],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z[],
          Editable->False]}], "]"}]}]]}], "-", 
    FractionBox[
     SuperscriptBox[
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["Gx",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Gx,
         Editable->False], 
        TagBox[
         RowBox[{"(", 
          RowBox[{"1", ",", "0"}], ")"}],
         Derivative],
        MultilineFunction->None], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho][],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["z",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`z[],
         Editable->False]}], "]"}], "2"], 
     RowBox[{"4", " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["Gx",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Gx,
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho][],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["z",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`z[],
         Editable->False]}], "]"}]}]], "-", 
    FractionBox[
     RowBox[{
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["Gx",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Gx,
         Editable->False], 
        TagBox[
         RowBox[{"(", 
          RowBox[{"1", ",", "0"}], ")"}],
         Derivative],
        MultilineFunction->None], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho][],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["z",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`z[],
         Editable->False]}], "]"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["ks",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`ks,
           Editable->False], "2"], " ", 
         RowBox[{
          InterpretationBox[
           StyleBox["P",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`P,
           Editable->False], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[Rho]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Rho][],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z[],
            Editable->False]}], "]"}], " ", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["Hx",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Hx,
            Editable->False], 
           TagBox[
            RowBox[{"(", 
             RowBox[{"1", ",", "0"}], ")"}],
            Derivative],
           MultilineFunction->None], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[Rho]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Rho][],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z[],
            Editable->False]}], "]"}]}], "+", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["ks",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`ks,
           Editable->False], "2"], " ", 
         RowBox[{
          InterpretationBox[
           StyleBox["Hx",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Hx,
           Editable->False], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[Rho]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Rho][],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z[],
            Editable->False]}], "]"}], " ", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["P",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`P,
            Editable->False], 
           TagBox[
            RowBox[{"(", 
             RowBox[{"1", ",", "0"}], ")"}],
            Derivative],
           MultilineFunction->None], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[Rho]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Rho][],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z[],
            Editable->False]}], "]"}]}]}], ")"}]}], 
     RowBox[{"4", " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["ks",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ks,
        Editable->False], "2"], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["Hx",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Hx,
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho][],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["z",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`z[],
         Editable->False]}], "]"}], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["P",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`P,
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho][],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["z",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`z[],
         Editable->False]}], "]"}]}]], "+", 
    RowBox[{
     FractionBox["1", "2"], " ", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["Gx",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Gx,
        Editable->False], 
       TagBox[
        RowBox[{"(", 
         RowBox[{"2", ",", "0"}], ")"}],
        Derivative],
       MultilineFunction->None], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Rho]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Rho][],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["z",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`z[],
        Editable->False]}], "]"}]}]}], 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["ks",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ks,
      Editable->False], "2"], " ", 
    RowBox[{
     InterpretationBox[
      StyleBox["Hx",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Hx,
      Editable->False], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho][],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["z",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`z[],
       Editable->False]}], "]"}], " ", 
    RowBox[{
     InterpretationBox[
      StyleBox["P",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`P,
      Editable->False], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho][],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["z",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`z[],
       Editable->False]}], "]"}]}]]}]], "Output",
 CellChangeTimes->{3.7286770357868834`*^9},
 CellLabel->"Out[92]=",
 CellID->1907364306,ExpressionUUID->"03fd2137-5c1c-4276-85b5-54bfb1c25cf8"]
}, Closed]],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"%", "/.", "GFunctions"}], "//", "Expand"}], ";"}]], "Input",
 CellLabel->"In[93]:=",
 CellID->464768349,ExpressionUUID->"70e0b9f1-33d9-44ed-b558-cdfb98b97f52"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"%", "/.", "\[Mu]Functions"}], "//", "Expand"}], ";"}]], "Input",
 CellLabel->"In[94]:=",
 CellID->963617504,ExpressionUUID->"334e1ca6-d49e-4a99-a9a7-24ac3979b87e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "//", "Together"}], "//", "AbsoluteTiming"}]], "Input",
 CellLabel->"In[95]:=",
 CellID->1345207701,ExpressionUUID->"77dfc849-340b-4be2-85e9-40db5a00b02c"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.576666`", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.728677036824142*^9},
 CellLabel->"Out[95]=",
 CellID->1583492173,ExpressionUUID->"79aaa6ef-a506-4a55-a7b8-367354dfae3b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "However the other six independent components are too hard for ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " on a laptop. We just show that results are correct for generic values of \
the coefficients and variables. Those random values must be exact values, in \
order to get an exact zero. Well, the truth is that not even with integers we \
can get a zero, so we do it numerically in the end."
}], "MathCaption",
 CellID->1251325234,ExpressionUUID->"e1b622f2-7fb8-43f0-aecd-df742f9b3953"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"SeedRandom", "[", "123", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"random", "=", 
  RowBox[{"Thread", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
     "q", ",", "a1", ",", "a2", ",", "a3", ",", "a4", ",", "a5", ",", "c1", 
      ",", "c2", ",", "ks", ",", 
      RowBox[{"z", "[", "]"}], ",", 
      RowBox[{"\[Rho]", "[", "]"}]}], "}"}], "\[Rule]", 
    RowBox[{"RandomInteger", "[", 
     RowBox[{"100", ",", "11"}], "]"}]}], "]"}]}]}], "Input",
 CellLabel->"In[96]:=",
 CellID->715852767,ExpressionUUID->"e4f1af23-68e3-4b32-ba51-027e34a93953"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox["q",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`q,
     Editable->False], "\[Rule]", "58"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox["a1",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`a1,
     Editable->False], "\[Rule]", "42"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox["a2",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`a2,
     Editable->False], "\[Rule]", "64"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox["a3",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`a3,
     Editable->False], "\[Rule]", "38"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox["a4",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`a4,
     Editable->False], "\[Rule]", "60"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox["a5",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`a5,
     Editable->False], "\[Rule]", "96"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox["c1",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`c1,
     Editable->False], "\[Rule]", "15"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox["c2",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`c2,
     Editable->False], "\[Rule]", "82"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox["ks",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ks,
     Editable->False], "\[Rule]", "75"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox["z",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`z[],
     Editable->False], "\[Rule]", "25"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox["\[Rho]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Rho][],
     Editable->False], "\[Rule]", "38"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728677036996626*^9},
 CellLabel->"Out[97]=",
 CellID->2135597987,ExpressionUUID->"6320df60-43ef-42ac-b4ec-6e3c8f9681b9"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"rules", "=", 
   RowBox[{"FoldedRule", "[", 
    RowBox[{
    "\[Omega]\[Psi]Function", ",", "HFunctions", ",", "MFunctions", ",", 
     "GFunctions", ",", "FFunction", ",", "PFunction", ",", "RFunctions", ",",
      "\[Mu]Functions"}], "]"}]}], ";"}]], "Input",
 CellLabel->"In[98]:=",
 CellID->1304264200,ExpressionUUID->"db45c0c8-3c5b-4534-a076-5e753cd280ca"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"r00", "=", 
   RowBox[{
    RowBox[{"ricci", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "BS"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "BS"}]}], "}"}]}], "]"}], "/.", "rules"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"r00n", "=", 
   RowBox[{"r00", "/.", "random"}]}], ";"}], "\n", 
 RowBox[{"N", "[", "r00n", "]"}]}], "Input",
 CellLabel->"In[99]:=",
 CellID->1402824471,ExpressionUUID->"e4c40c8b-cda8-47bf-90c5-6be056415730"],

Cell[BoxData[
 RowBox[{"-", "1.3318685146440419`*^-16"}]], "Output",
 CellChangeTimes->{3.728677046844404*^9},
 CellLabel->"Out[101]=",
 CellID->2101175198,ExpressionUUID->"ce0388cd-ee89-4876-bba1-72b217778c43"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"r11", "=", 
   RowBox[{
    RowBox[{"ricci", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "BS"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "BS"}]}], "}"}]}], "]"}], "/.", "rules"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"r11n", "=", 
   RowBox[{"r11", "/.", "random"}]}], ";"}], "\n", 
 RowBox[{"N", "[", "r11n", "]"}]}], "Input",
 CellLabel->"In[102]:=",
 CellID->146989363,ExpressionUUID->"220e1095-a8f3-48e6-9246-6139b73c863c"],

Cell[BoxData[
 RowBox[{"-", "2.6152263629414946`*^-8"}]], "Output",
 CellChangeTimes->{3.728677076269478*^9},
 CellLabel->"Out[104]=",
 CellID->1015765640,ExpressionUUID->"d5c77961-076d-4f3f-8f5d-83672ca7fe6c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"r22", "=", 
   RowBox[{
    RowBox[{"ricci", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"2", ",", 
        RowBox[{"-", "BS"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"2", ",", 
        RowBox[{"-", "BS"}]}], "}"}]}], "]"}], "/.", "rules"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"r22n", "=", 
   RowBox[{"r22", "/.", "random"}]}], ";"}], "\n", 
 RowBox[{"N", "[", "r22n", "]"}]}], "Input",
 CellLabel->"In[105]:=",
 CellID->1480889602,ExpressionUUID->"2e71067e-a939-403c-8acb-1be6522dfe2f"],

Cell[BoxData[
 RowBox[{"-", "2.484825115090482`*^-9"}]], "Output",
 CellChangeTimes->{3.728677105140905*^9},
 CellLabel->"Out[107]=",
 CellID->1625601477,ExpressionUUID->"46e70369-607a-42c6-86b6-8d3127db1376"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"r33", "=", 
   RowBox[{
    RowBox[{"ricci", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"3", ",", 
        RowBox[{"-", "BS"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"3", ",", 
        RowBox[{"-", "BS"}]}], "}"}]}], "]"}], "/.", "rules"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"r33n", "=", 
   RowBox[{"r33", "/.", "random"}]}], ";"}], "\n", 
 RowBox[{"N", "[", "r33n", "]"}]}], "Input",
 CellLabel->"In[108]:=",
 CellID->1570441424,ExpressionUUID->"60debcd5-14be-40e3-8dc3-fbda0c7ffab3"],

Cell[BoxData[
 RowBox[{"-", "4.564662033897859`*^-13"}]], "Output",
 CellChangeTimes->{3.728677128350608*^9},
 CellLabel->"Out[110]=",
 CellID->1238064068,ExpressionUUID->"230db245-d890-4607-8af0-c0506ae89988"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"r12", "=", 
   RowBox[{
    RowBox[{"ricci", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "BS"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"2", ",", 
        RowBox[{"-", "BS"}]}], "}"}]}], "]"}], "/.", "rules"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"r12n", "=", 
   RowBox[{"r12", "/.", "random"}]}], ";"}], "\n", 
 RowBox[{"N", "[", "r12n", "]"}]}], "Input",
 CellLabel->"In[111]:=",
 CellID->364957946,ExpressionUUID->"69bad4bd-e2b2-4050-8a3c-2daa7612c904"],

Cell[BoxData["8.34131654354098`*^-9"], "Output",
 CellChangeTimes->{3.7286771612022457`*^9},
 CellLabel->"Out[113]=",
 CellID->662439540,ExpressionUUID->"100dccf5-0c60-454d-b9d5-b74624f712c5"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"r03", "=", 
   RowBox[{
    RowBox[{"ricci", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "BS"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"3", ",", 
        RowBox[{"-", "BS"}]}], "}"}]}], "]"}], "/.", "rules"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"r03n", "=", 
   RowBox[{"r03", "/.", "random"}]}], ";"}], "\n", 
 RowBox[{"N", "[", "r03n", "]"}]}], "Input",
 CellLabel->"In[114]:=",
 CellID->2109071834,ExpressionUUID->"92a86a93-c3a7-41f9-9b73-17978e31ed5f"],

Cell[BoxData[
 RowBox[{"-", "7.68565321964731`*^-15"}]], "Output",
 CellChangeTimes->{3.7286771776983557`*^9},
 CellLabel->"Out[116]=",
 CellID->1066122732,ExpressionUUID->"51bb3aa7-4131-4543-8c5d-5fde59787a57"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{
  "BlackSaturn", ",", "matrix", ",", "ematrix", ",", "FFunction", ",", 
   "GFunctions", ",", "HFunctions", ",", "PFunction", ",", "RFunctions", ",", 
   "MFunctions", ",", "\[Mu]Functions", ",", "\[Omega]\[Psi]Function", ",", 
   "met", ",", "cd", ",", "ricci", ",", "r00", ",", "r11", ",", "r22", ",", 
   "r33", ",", "r12", ",", "r03", ",", "r00n", ",", "r11n", ",", "r22n", ",", 
   "r33n", ",", "r12n", ",", "r03n", ",", "random", ",", "rules"}], 
  "]"}]], "Input",
 CellLabel->"In[117]:=",
 CellID->170533998,ExpressionUUID->"7997daac-4313-4c63-8c3d-c3417471f872"],

Cell[BoxData[
 RowBox[{"UndefScalarFunction", "[", 
  RowBox[{"{", 
   RowBox[{
   "\[Mu]1", ",", "\[Mu]2", ",", "\[Mu]3", ",", "\[Mu]4", ",", "\[Mu]5", ",", 
    "M0", ",", "M1", ",", "M2", ",", "M3", ",", "M4", ",", "R1", ",", "R2", 
    ",", "Hx", ",", "Hy", ",", "Gx", ",", "Gy", ",", "\[Omega]\[Psi]", ",", 
    "P", ",", "F"}], "}"}], "]"}]], "Input",
 CellLabel->"In[118]:=",
 CellID->1421719288,ExpressionUUID->"a5f8eacb-4cf6-4d39-8734-18814ebdf7ce"],

Cell[BoxData[
 RowBox[{"UndefConstantSymbol", "[", 
  RowBox[{"{", 
   RowBox[{
   "ks", ",", "q", ",", "a1", ",", "a2", ",", "a3", ",", "a4", ",", "a5", ",",
     "c1", ",", "c2"}], "}"}], "]"}]], "Input",
 CellLabel->"In[119]:=",
 CellID->1933016894,ExpressionUUID->"1196794c-d6ab-4127-8694-67db0c4689b2"],

Cell[BoxData[
 RowBox[{"UndefChart", "[", "BS", "]"}]], "Input",
 CellLabel->"In[120]:=",
 CellID->1082325642,ExpressionUUID->"a80d6460-6d42-4607-8b2e-ced7e0950a96"],

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M", "]"}]], "Input",
 CellLabel->"In[121]:=",
 CellID->1460575004,ExpressionUUID->"08a737fb-26d8-44ae-920c-5e2b9cabce40"],

Cell[BoxData[
 RowBox[{"ClearxCobaCache", "[", "All", "]"}]], "Input",
 CellLabel->"In[122]:=",
 CellID->2146007018,ExpressionUUID->"90fef4e7-c30b-4e34-bf24-9f78ae378122"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.5.7. An example with two bases", "Subsubsection",
 CellChangeTimes->{{3.728676996119493*^9, 3.728676996311599*^9}, 
   3.7908767566464453`*^9},
 CellID->952100161,ExpressionUUID->"56d80cc5-0f35-4776-b875-c37d38b2db67"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<xAct`xCoba`\>\"", "]"}]], "Input",
 CellLabel->"In[1]:=",
 CellID->1046752130,ExpressionUUID->"faa52d1c-cdee-4d8e-8089-5afc813acbcf"],

Cell[CellGroupData[{

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.7333684689725113`*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->263638693,ExpressionUUID->"3c159864-2e8e-4126-87d5-61f5bb03fd86"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`xPerm`  version \"\>", 
   "\[InvisibleSpace]", "\<\"1.2.3\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2015", ",", "8", ",", "23"}], "}"}]}],
  SequenceForm["Package xAct`xPerm`  version ", "1.2.3", ", ", {2015, 8, 23}],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.7333684689852667`*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->716749490,ExpressionUUID->"8097e80d-55b1-4b05-a75e-48060719b801"],

Cell[BoxData["\<\"CopyRight (C) 2003-2018, Jose M. Martin-Garcia, under the \
General Public License.\"\>"], "Print",
 CellChangeTimes->{3.733368469002411*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1172902407,ExpressionUUID->"ad3af31f-d0cc-431b-aec5-4c9d4b027cb0"],

Cell[BoxData["\<\"Connecting to external mac executable...\"\>"], "Print",
 CellChangeTimes->{3.733368469013866*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1218358636,ExpressionUUID->"8b117629-37b1-4e63-a2d7-fbd3df66c258"],

Cell[BoxData["\<\"Connection established.\"\>"], "Print",
 CellChangeTimes->{3.733368469026567*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->697021107,ExpressionUUID->"e40b0a8e-958e-4c0a-837d-a0f6c7d0f168"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.733368469039213*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->310450964,ExpressionUUID->"882bc889-0e8e-4ba0-8c63-9e7e25ef0970"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`xTensor`  version \"\>", 
   "\[InvisibleSpace]", "\<\"1.1.3\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2018", ",", "2", ",", "28"}], "}"}]}],
  SequenceForm[
  "Package xAct`xTensor`  version ", "1.1.3", ", ", {2018, 2, 28}],
  Editable->False]], "Print",
 CellChangeTimes->{3.733368469051519*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1511573385,ExpressionUUID->"7b60ecce-49fb-4b1a-be6e-ad81ad6d5f34"],

Cell[BoxData["\<\"CopyRight (C) 2002-2018, Jose M. Martin-Garcia, under the \
General Public License.\"\>"], "Print",
 CellChangeTimes->{3.733368469063476*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->282337685,ExpressionUUID->"40c0aa33-7d3d-4ac3-924f-32868e46d9fd"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.73336846913603*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1458095572,ExpressionUUID->"83d26627-c635-43a8-acb4-46ba1d709bbf"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`xCoba`  version \"\>", 
   "\[InvisibleSpace]", "\<\"0.8.4\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2018", ",", "2", ",", "28"}], "}"}]}],
  SequenceForm["Package xAct`xCoba`  version ", "0.8.4", ", ", {2018, 2, 28}],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.733368469148177*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1276382687,ExpressionUUID->"4ece7435-28a2-4e7c-8f00-ef5ca9cc43c0"],

Cell[BoxData["\<\"CopyRight (C) 2005-2018, David Yllanes and Jose M. \
Martin-Garcia, under the General Public License.\"\>"], "Print",
 CellChangeTimes->{3.733368469160069*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->949670526,ExpressionUUID->"3e89e572-0f5b-4f3c-bebe-c7fab7fd8cfe"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.733368469172174*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1014065978,ExpressionUUID->"ff064016-153c-4e90-91aa-1c6a472ae061"],

Cell[BoxData["\<\"These packages come with ABSOLUTELY NO WARRANTY; for \
details type Disclaimer[]. This is free software, and you are welcome to \
redistribute it under certain conditions. See the General Public License for \
details.\"\>"], "Print",
 CellChangeTimes->{3.7333684691843557`*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->976669654,ExpressionUUID->"853374a3-e1d4-4762-be92-41c9b2161963"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.733368469196247*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1669360757,ExpressionUUID->"d64f7793-9f6f-4684-9e63-5f41b6fa71d8"]
}, Open  ]]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"$DefInfoQ", "=", 
   RowBox[{"$UndefInfoQ", "=", "False"}]}], ";"}]], "Input",
 CellLabel->"In[2]:=",
 CellID->577012686,ExpressionUUID->"9bd1ce9e-069e-42fa-badb-77b87918a05b"],

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M4", ",", "4", ",", 
   RowBox[{"{", 
    RowBox[{
    "\[Mu]", ",", "\[Nu]", ",", "\[Alpha]", ",", "\[Beta]", ",", "\[Gamma]", 
     ",", "\[Delta]", ",", "\[Lambda]"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[3]:=",
 CellID->1963547397,ExpressionUUID->"a66a252b-8335-4871-bb34-019d9e6d4559"],

Cell[BoxData[
 RowBox[{"DefConstantSymbol", "[", 
  RowBox[{"Mass", ",", 
   RowBox[{"PrintAs", "\[Rule]", "\"\<M\>\""}]}], "]"}]], "Input",
 CellLabel->"In[4]:=",
 CellID->959897930,ExpressionUUID->"915b3f72-68bc-4248-ad34-53a7e94424d1"],

Cell[CellGroupData[{

Cell["This is the metric in the standard coordinates:", "MathCaption",
 CellID->1107353246,ExpressionUUID->"6c23bcdf-2d3a-4372-9939-d28da3bec2bd"],

Cell[BoxData[
 RowBox[{"DefChart", "[", 
  RowBox[{"schw", ",", "M4", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2", ",", "3"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"t", "[", "]"}], ",", 
     RowBox[{"r", "[", "]"}], ",", 
     RowBox[{"\[Theta]", "[", "]"}], ",", 
     RowBox[{"\[Phi]", "[", "]"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[5]:=",
 CellID->1192720620,ExpressionUUID->"9402bb28-22a0-4c59-ad99-3b252a4d7abe"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"MatrixForm", "[", 
  RowBox[{"SchwarzschildMatrix", "=", 
   RowBox[{"DiagonalMatrix", "[", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "1"}], "+", 
       RowBox[{"2", 
        RowBox[{"Mass", "/", 
         RowBox[{"r", "[", "]"}]}]}]}], ",", 
      RowBox[{"1", "/", 
       RowBox[{"(", 
        RowBox[{"1", "-", 
         RowBox[{"2", 
          RowBox[{"Mass", "/", 
           RowBox[{"r", "[", "]"}]}]}]}], ")"}]}], ",", 
      RowBox[{
       RowBox[{"r", "[", "]"}], "^", "2"}], ",", 
      RowBox[{
       RowBox[{
        RowBox[{"r", "[", "]"}], "^", "2"}], 
       RowBox[{
        RowBox[{"Sin", "[", 
         RowBox[{"\[Theta]", "[", "]"}], "]"}], "^", "2"}]}]}], "}"}], 
    "]"}]}], "]"}]], "Input",
 CellLabel->"In[6]:=",
 CellID->1443099242,ExpressionUUID->"26a257d7-cde5-414e-9718-53ab684cc3b8"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      RowBox[{
       RowBox[{"-", "1"}], "+", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Mass,
          Editable->False]}], 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]]}], "0", "0", "0"},
     {"0", 
      FractionBox["1", 
       RowBox[{"1", "-", 
        FractionBox[
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Mass,
           Editable->False]}], 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]]}]], "0", "0"},
     {"0", "0", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"], "0"},
     {"0", "0", "0", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"], " ", 
       SuperscriptBox[
        RowBox[{
         InterpretationBox[
          StyleBox["Sin",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Sin,
          Editable->False], "[", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False], "]"}], "2"]}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{3.728677179039442*^9, 3.733368473641652*^9},
 CellLabel->"Out[6]//MatrixForm=",
 CellID->1906380320,ExpressionUUID->"25d210f7-2204-4082-be97-a08e1e93e8b2"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"$LargeComponentSize", "=", "2000"}], ";"}]], "Input",
 CellLabel->"In[7]:=",
 CellID->781168204,ExpressionUUID->"4fc36915-6394-4fda-8f8a-44fb88398de6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"metric", "=", 
  RowBox[{"CTensor", "[", 
   RowBox[{"SchwarzschildMatrix", ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "schw"}], ",", 
      RowBox[{"-", "schw"}]}], "}"}]}], "]"}]}]], "Input",
 CellLabel->"In[8]:=",
 CellID->1526470517,ExpressionUUID->"495665e6-f8b8-4a0c-830e-a37ae6f51b4c"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "1"}], "+", 
        FractionBox[
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Mass,
           Editable->False]}], 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]]}], ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", 
       FractionBox["1", 
        RowBox[{"1", "-", 
         FractionBox[
          RowBox[{"2", " ", 
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False]}], 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False]]}]], ",", "0", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"], ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", "0", ",", 
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"], " ", 
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Sin",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Sin,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"]}]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "schw"}], ",", 
     RowBox[{"-", "schw"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.728677179289358*^9, 3.733368475644968*^9},
 CellLabel->"Out[8]=",
 CellID->1146005907,ExpressionUUID->"7327228f-d07d-4e20-8ee7-4a8912ad84ea"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["This is now the metric in an adapted frame:", "MathCaption",
 CellID->1013473648,ExpressionUUID->"cc5e26c8-d717-4df6-8c0b-eac693cbae8b"],

Cell[BoxData[
 RowBox[{"DefBasis", "[", 
  RowBox[{"frame", ",", "TangentM4", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2", ",", "3"}], "}"}], ",", 
   RowBox[{"BasisColor", "\[Rule]", 
    RowBox[{"RGBColor", "[", 
     RowBox[{"0", ",", "0", ",", "1"}], "]"}]}], ",", 
   RowBox[{"BasisChange", "\[Rule]", 
    RowBox[{"CTensor", "[", 
     RowBox[{
      RowBox[{"DiagonalMatrix", "[", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"1", "/", 
          RowBox[{"Sqrt", "[", 
           RowBox[{"1", "-", 
            RowBox[{"2", 
             RowBox[{"Mass", "/", 
              RowBox[{"r", "[", "]"}]}]}]}], "]"}]}], ",", 
         RowBox[{"Sqrt", "[", 
          RowBox[{"1", "-", 
           RowBox[{"2", 
            RowBox[{"Mass", "/", 
             RowBox[{"r", "[", "]"}]}]}]}], "]"}], ",", 
         RowBox[{"1", "/", 
          RowBox[{"r", "[", "]"}]}], ",", 
         RowBox[{
          RowBox[{"1", "/", 
           RowBox[{"r", "[", "]"}]}], "/", 
          RowBox[{"Sin", "[", 
           RowBox[{"\[Theta]", "[", "]"}], "]"}]}]}], "}"}], "]"}], ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "frame"}], ",", "schw"}], "}"}]}], "]"}]}]}], 
  "]"}]], "Input",
 CellLabel->"In[9]:=",
 CellID->1758797579,ExpressionUUID->"7ce966ce-fc2b-48a1-a515-50d76c95ca0d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"metricF", "=", 
  RowBox[{"CTensor", "[", 
   RowBox[{
    RowBox[{"DiagonalMatrix", "[", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "1"}], ",", "1", ",", "1", ",", "1"}], "}"}], "]"}], ",", 
    
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "frame"}], ",", 
      RowBox[{"-", "frame"}]}], "}"}], ",", "0"}], "]"}]}]], "Input",
 CellLabel->"In[10]:=",
 CellID->2145992945,ExpressionUUID->"92492632-eddf-4da6-be59-664046b06a9d"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "1"}], ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "1", ",", "0", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", "1", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", "0", ",", "1"}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "frame"}], ",", 
     RowBox[{"-", "frame"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.728677179521884*^9, 3.7333684866106453`*^9},
 CellLabel->"Out[10]=",
 CellID->670912572,ExpressionUUID->"0c833855-7cbe-49ca-a7df-837e39930d50"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We start by precomputing everything in advance for both expressions of the \
metric:\
\>", "MathCaption",
 CellID->764803550,ExpressionUUID->"7ea71d3e-5e66-408f-bacf-bee42758599b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"MetricCompute", "[", 
   RowBox[{"metric", ",", "schw", ",", "All"}], "]"}], "//", 
  "Timing"}]], "Input",
 CellLabel->"In[11]:=",
 CellID->1363637555,ExpressionUUID->"13eb0201-8dab-4aa5-a2fb-143dce550da4"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.152459`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.7286771798424673`*^9, 3.733368488063292*^9},
 CellLabel->"Out[11]=",
 CellID->1696935750,ExpressionUUID->"4ec61e03-c8df-476f-aa03-f0d28a5e5669"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"MetricCompute", "[", 
   RowBox[{"metricF", ",", "schw", ",", "All"}], "]"}], "//", 
  "Timing"}]], "Input",
 CellLabel->"In[12]:=",
 CellID->1747311246,ExpressionUUID->"30aed816-b889-4944-98ad-0e0de7ce4194"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.124111`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.7286771800616302`*^9, 3.7333684890721397`*^9},
 CellLabel->"Out[12]=",
 CellID->1274418663,ExpressionUUID->"5d18c90c-e6a8-4e98-9947-a42f96f2141f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["These are the metrics:", "MathCaption",
 CellID->1066268222,ExpressionUUID->"6517bff2-9be7-4a0f-8667-37a79c57aeb7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"metric", "[", 
  RowBox[{
   RowBox[{"-", "\[Alpha]"}], ",", 
   RowBox[{"-", "\[Beta]"}]}], "]"}]], "Input",
 CellLabel->"In[13]:=",
 CellID->18972939,ExpressionUUID->"e5c05129-689d-4c0d-92c2-601742d2fc52"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{"-", "1"}], "+", 
             FractionBox[
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Mass,
                Editable->False]}], 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{"-", "1"}], "+", 
                FractionBox[
                 RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], 
                 InterpretationBox[
                  StyleBox[
                  "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`r[], Editable -> False]]}]}]],
           
           Annotation[#, {1, 1} -> -1 + 
             2 ($CellContext`Mass/$CellContext`r[]), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox["1", 
             RowBox[{"1", "-", 
              FractionBox[
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False]}], 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]]}]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              FractionBox["1", 
               RowBox[{"1", "-", 
                 FractionBox[
                  RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], 
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False]]}]]}]],
           
           Annotation[#, {2, 
             2} -> (1 - 2 ($CellContext`Mass/$CellContext`r[]))^(-1), 
            "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox[
                "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                $CellContext`r[], Editable -> False], "2"]}]],
           Annotation[#, {3, 3} -> $CellContext`r[]^2, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"], " ", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], "2"]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "4"}], "}"}], "\[Rule]", 
              RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                  "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`r[], Editable -> False], "2"], " ", 
                SuperscriptBox[
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                   InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                 "2"]}]}]],
           Annotation[#, {4, 4} -> $CellContext`r[]^2 Sin[
                $CellContext`\[Theta][]]^2, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " "},
         {"\[Alpha]", "\[Beta]"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{-1 + 2 ($CellContext`Mass/$CellContext`r[]), 0, 0, 
     0}, {0, (1 - 2 ($CellContext`Mass/$CellContext`r[]))^(-1), 0, 0}, {
    0, 0, $CellContext`r[]^2, 0}, {0, 0, 0, $CellContext`r[]^2 Sin[
        $CellContext`\[Theta][]]^2}}, {-$CellContext`schw, \
-$CellContext`schw}, 0][-$CellContext`\[Alpha], -$CellContext`\[Beta]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728677180216591*^9, 3.733368490360055*^9},
 CellLabel->"Out[13]=",
 CellID->659278950,ExpressionUUID->"bd628d10-672d-4746-adb7-d3fdc6d10cdb"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"metricF", "[", 
  RowBox[{
   RowBox[{"-", "\[Alpha]"}], ",", 
   RowBox[{"-", "\[Beta]"}]}], "]"}]], "Input",
 CellLabel->"In[14]:=",
 CellID->924782085,ExpressionUUID->"212483cd-2d70-4693-9135-aaeac95205dd"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{"-", "1"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{"-", "1"}]}]],
           Annotation[#, {1, 1} -> -1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {2, 2} -> 1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {3, 3} -> 1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "4"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {4, 4} -> 1, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]},
           "RowsIndexed" -> {1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " "},
         {"\[Alpha]", "\[Beta]"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{-1, 0, 0, 0}, {0, 1, 0, 0}, {0, 0, 1, 0}, {0, 0, 0, 
   1}}, {-$CellContext`frame, -$CellContext`frame}, 
   0][-$CellContext`\[Alpha], -$CellContext`\[Beta]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728677180689645*^9, 3.733368492278377*^9},
 CellLabel->"Out[14]=",
 CellID->1880112572,ExpressionUUID->"790ddc3e-d333-4db8-b160-a96c02c5da8b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["and their inverses:", "MathCaption",
 CellID->2094495682,ExpressionUUID->"597bda04-dc91-476d-b62d-e2e9d08bd00b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Inv", "[", "metric", "]"}], "[", 
  RowBox[{"\[Alpha]", ",", "\[Beta]"}], "]"}]], "Input",
 CellLabel->"In[15]:=",
 CellID->636985750,ExpressionUUID->"fe5c6a00-1ae0-4219-8d31-8858543f85c4"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            FractionBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], 
             RowBox[{
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Mass,
                Editable->False]}], "-", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
              FractionBox[
               InterpretationBox[
                StyleBox[
                "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                $CellContext`r[], Editable -> False], 
               RowBox[{
                 RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], "-", 
                 InterpretationBox[
                  StyleBox[
                  "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`r[], Editable -> False]}]]}]],
           
           Annotation[#, {1, 
             1} -> (2 $CellContext`Mass - $CellContext`r[])^(-1) \
$CellContext`r[], "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{"1", "-", 
             FractionBox[
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Mass,
                Editable->False]}], 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{"1", "-", 
                FractionBox[
                 RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], 
                 InterpretationBox[
                  StyleBox[
                  "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`r[], Editable -> False]]}]}]],
           
           Annotation[#, {2, 2} -> 1 - 2 ($CellContext`Mass/$CellContext`r[]),
             "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox["1", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", 
              FractionBox["1", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`r[], Editable -> False], "2"]]}]],
           Annotation[#, {3, 3} -> $CellContext`r[]^(-2), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox[
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Csc",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Csc,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], "2"], 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "4"}], "}"}], "\[Rule]", 
              FractionBox[
               SuperscriptBox[
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"Csc"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Csc, Editable -> False], "[", 
                  InterpretationBox[
                   StyleBox[
                   "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`\[Theta][], Editable -> False], "]"}], "2"], 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`r[], Editable -> False], "2"]]}]],
           Annotation[#, {4, 4} -> Csc[
                $CellContext`\[Theta][]]^2 $CellContext`r[]^(-2), 
            "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"\[Alpha]", "\[Beta]"},
         {" ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{(
       2 $CellContext`Mass - $CellContext`r[])^(-1) $CellContext`r[], 0, 0, 
     0}, {0, 1 - 2 ($CellContext`Mass/$CellContext`r[]), 0, 0}, {
    0, 0, $CellContext`r[]^(-2), 0}, {0, 0, 0, Csc[
        $CellContext`\[Theta][]]^2 $CellContext`r[]^(-2)}}, \
{$CellContext`schw, $CellContext`schw}, 
   0][$CellContext`\[Alpha], $CellContext`\[Beta]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728677181472666*^9, 3.733368493574376*^9},
 CellLabel->"Out[15]=",
 CellID->1563000292,ExpressionUUID->"aaf1e2cd-b499-4413-b14d-e737d4b1d6d0"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Inv", "[", "metricF", "]"}], "[", 
  RowBox[{"\[Alpha]", ",", "\[Beta]"}], "]"}]], "Input",
 CellLabel->"In[16]:=",
 CellID->1360894438,ExpressionUUID->"8c3079b8-f68b-4153-82d0-77a087067af5"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{"-", "1"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{"-", "1"}]}]],
           Annotation[#, {1, 1} -> -1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {2, 2} -> 1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {3, 3} -> 1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "4"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {4, 4} -> 1, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]},
           "RowsIndexed" -> {1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"\[Alpha]", "\[Beta]"},
         {" ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{-1, 0, 0, 0}, {0, 1, 0, 0}, {0, 0, 1, 0}, {0, 0, 0, 
   1}}, {$CellContext`frame, $CellContext`frame}, 
   0][$CellContext`\[Alpha], $CellContext`\[Beta]],
  Editable->False]], "Output",
 CellChangeTimes->{3.7286771816382513`*^9, 3.733368494879616*^9},
 CellLabel->"Out[16]=",
 CellID->1381377051,ExpressionUUID->"1473b186-5ccd-4039-8879-d3a94fb62ca4"]
}, Open  ]]
}, Open  ]],

Cell["\<\
Note that we do not use SetCMetric, and therefore we cannot raise and lower \
indices. In this way all metrics play an equivalent role.\
\>", "MathCaption",
 CellID->581661594,ExpressionUUID->"43ac862b-5903-40e3-8eba-14fc758cdbf4"],

Cell[CellGroupData[{

Cell["The LeviCivita connections:", "MathCaption",
 CellID->640849214,ExpressionUUID->"6e3ce83b-771d-4588-a4ff-7ad612c301fd"],

Cell[BoxData[
 RowBox[{
  RowBox[{"cd", "=", 
   RowBox[{"CovDOfMetric", "[", "metric", "]"}]}], ";"}]], "Input",
 CellLabel->"In[17]:=",
 CellID->1122249993,ExpressionUUID->"ca105fad-e10d-4665-965f-d7c0d40d4057"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Torsion", "[", "cd", "]"}]], "Input",
 CellLabel->"In[18]:=",
 CellID->306760750,ExpressionUUID->"303c57a0-6b1d-4be4-95e6-d21659a0de53"],

Cell[BoxData["Zero"], "Output",
 CellChangeTimes->{3.728677181900774*^9, 3.733368498513282*^9},
 CellLabel->"Out[18]=",
 CellID->2055329345,ExpressionUUID->"81321768-cd78-4d1b-9c52-50c5ecd6eac8"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"cdF", "=", 
   RowBox[{"CovDOfMetric", "[", "metricF", "]"}]}], ";"}]], "Input",
 CellLabel->"In[19]:=",
 CellID->1168529140,ExpressionUUID->"fcb29837-a65f-486d-bbb5-c65519f57027"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Torsion", "[", "cdF", "]"}]], "Input",
 CellLabel->"In[20]:=",
 CellID->731304165,ExpressionUUID->"741b615b-a030-45fc-b61e-4950a89bafa9"],

Cell[BoxData["Zero"], "Output",
 CellChangeTimes->{3.728677182106188*^9, 3.733368499896915*^9},
 CellLabel->"Out[20]=",
 CellID->1866202543,ExpressionUUID->"e8646506-ee06-400a-a3b9-b89feabdf16a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Note the difference between a coordinated and a non-coordinated basis:\
\>", "MathCaption",
 CellID->695752681,ExpressionUUID->"e61dbd88-e7f5-4caa-9c8e-21fff0feb3af"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCTensor", "[", 
  RowBox[{"TorsionPDschw", ",", 
   RowBox[{"{", 
    RowBox[{"schw", ",", 
     RowBox[{"-", "schw"}], ",", 
     RowBox[{"-", "schw"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[21]:=",
 CellID->466851652,ExpressionUUID->"48ed3467-55eb-497d-8a1d-46641487950c"],

Cell[BoxData["Zero"], "Output",
 CellChangeTimes->{3.728677182372334*^9, 3.7333685011639423`*^9},
 CellLabel->"Out[21]=",
 CellID->49490229,ExpressionUUID->"4ea02165-a640-475f-93e5-2e81ae4ccacb"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCTensor", "[", 
  RowBox[{"TorsionPDframe", ",", 
   RowBox[{"{", 
    RowBox[{"frame", ",", 
     RowBox[{"-", "frame"}], ",", 
     RowBox[{"-", "frame"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[22]:=",
 CellID->1492636797,ExpressionUUID->"f28db7b6-75dc-4812-abd9-ac7aa84d7d58"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", 
         RowBox[{"-", 
          FractionBox[
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False], 
           RowBox[{
            SqrtBox[
             RowBox[{"1", "-", 
              FractionBox[
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False]}], 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]]}]], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"]}]]}], ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         FractionBox[
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Mass,
           Editable->False], 
          RowBox[{
           SqrtBox[
            RowBox[{"1", "-", 
             FractionBox[
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Mass,
                Editable->False]}], 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]]}]], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}]], ",", "0", ",", "0", ",", "0"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", 
         FractionBox[
          SqrtBox[
           RowBox[{"1", "-", 
            FractionBox[
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False]}], 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]]}]], 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False]], ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", 
         RowBox[{"-", 
          FractionBox[
           SqrtBox[
            RowBox[{"1", "-", 
             FractionBox[
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Mass,
                Editable->False]}], 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]]}]], 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]]}], ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", 
         FractionBox[
          SqrtBox[
           RowBox[{"1", "-", 
            FractionBox[
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False]}], 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]]}]], 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False]]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", 
         FractionBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Cot",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Cot,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False]]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", 
         RowBox[{"-", 
          FractionBox[
           SqrtBox[
            RowBox[{"1", "-", 
             FractionBox[
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Mass,
                Editable->False]}], 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]]}]], 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]]}], ",", 
         RowBox[{"-", 
          FractionBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Cot",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cot,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]]}], ",", "0"}], "}"}]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"frame", ",", 
     RowBox[{"-", "frame"}], ",", 
     RowBox[{"-", "frame"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.7286771828063498`*^9, 3.733368501809525*^9},
 CellLabel->"Out[22]=",
 CellID->1252467486,ExpressionUUID->"cb74294e-6d80-4585-bde6-ffb08bb27165"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Metric compatibility. We can use any form of the metric or the derivative:\
\>", "MathCaption",
 CellID->1671773579,ExpressionUUID->"ba670c55-95e3-44d5-8c69-f61a70a12332"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"cd", "[", 
   RowBox[{"-", "\[Mu]"}], "]"}], "[", 
  RowBox[{"metric", "[", 
   RowBox[{
    RowBox[{"-", "\[Alpha]"}], ",", 
    RowBox[{"-", "\[Beta]"}]}], "]"}], "]"}]], "Input",
 CellLabel->"In[23]:=",
 CellID->1969988552,ExpressionUUID->"ca96705e-c2da-4175-8336-dd8a3b1379b5"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.7286771831441174`*^9, 3.733368506146636*^9},
 CellLabel->"Out[23]=",
 CellID->1867085882,ExpressionUUID->"4f504605-7124-41c4-93c0-a71294efc870"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"cdF", "[", 
   RowBox[{"-", "\[Mu]"}], "]"}], "[", 
  RowBox[{"metricF", "[", 
   RowBox[{
    RowBox[{"-", "\[Alpha]"}], ",", 
    RowBox[{"-", "\[Beta]"}]}], "]"}], "]"}]], "Input",
 CellLabel->"In[24]:=",
 CellID->1712486856,ExpressionUUID->"2085d3c3-326a-4bc5-8542-f382f0215506"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728677183282454*^9, 3.73336851551271*^9},
 CellLabel->"Out[24]=",
 CellID->782550271,ExpressionUUID->"1fcd15f7-6363-48ac-ae58-46340e9c3a51"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"cdF", "[", 
   RowBox[{"-", "\[Mu]"}], "]"}], "[", 
  RowBox[{"metric", "[", 
   RowBox[{
    RowBox[{"-", "\[Alpha]"}], ",", 
    RowBox[{"-", "\[Beta]"}]}], "]"}], "]"}]], "Input",
 CellLabel->"In[25]:=",
 CellID->563681893,ExpressionUUID->"e741c133-a068-40fe-9100-ed07c761ac0a"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728677183490357*^9, 3.733368518411378*^9},
 CellLabel->"Out[25]=",
 CellID->36108568,ExpressionUUID->"0d392747-cf75-4aa3-9221-637e02b01cfb"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"cd", "[", 
   RowBox[{"-", "\[Mu]"}], "]"}], "[", 
  RowBox[{"metricF", "[", 
   RowBox[{
    RowBox[{"-", "\[Alpha]"}], ",", 
    RowBox[{"-", "\[Beta]"}]}], "]"}], "]"}]], "Input",
 CellLabel->"In[26]:=",
 CellID->905320914,ExpressionUUID->"238b8407-a703-41fd-922e-faf089a1c531"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728677184109391*^9, 3.733368528247291*^9},
 CellLabel->"Out[26]=",
 CellID->43528326,ExpressionUUID->"be56381b-c435-46fc-894b-ad04eddec675"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Riemann:", "MathCaption",
 CellID->1261237816,ExpressionUUID->"f5a74e04-693b-4706-a1e8-4702e6c08a5b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{"riem", "=", 
    RowBox[{"Simplify", "@", 
     RowBox[{"Riemann", "[", "cd", "]"}]}]}], ")"}], "[", 
  RowBox[{
   RowBox[{"-", "\[Alpha]"}], ",", 
   RowBox[{"-", "\[Beta]"}], ",", 
   RowBox[{"-", "\[Gamma]"}], ",", "\[Delta]"}], "]"}]], "Input",
 CellLabel->"In[27]:=",
 CellID->1960377622,ExpressionUUID->"aa04a91a-6dd0-4f3d-a80b-239e10e2683d"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               FractionBox[
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "-", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}]}], 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "4"]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}], 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "4"]]}]],
              
              Annotation[#, {1, 2, 1, 
                2} -> ((2 $CellContext`Mass) (
                  2 $CellContext`Mass - $CellContext`r[])) \
$CellContext`r[]^(-4), "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               FractionBox[
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False]}], 
                RowBox[{
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "-", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], 
                  RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]]}]],
              
              Annotation[#, {1, 2, 2, 
                1} -> ((2 $CellContext`Mass)/(
                 2 $CellContext`Mass - $CellContext`r[])) \
$CellContext`r[]^(-2), "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               FractionBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}]}], 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "4"]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}], 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "4"]]}]],
              
              Annotation[#, {1, 3, 1, 
                3} -> ($CellContext`Mass $CellContext`r[]^(-4)) ((-2) \
$CellContext`Mass + $CellContext`r[]), "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               FractionBox[
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], 
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False]]}]],
              
              Annotation[#, {1, 3, 3, 
                1} -> $CellContext`Mass/$CellContext`r[], "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               FractionBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}]}], 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "4"]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}], 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "4"]]}]],
              
              Annotation[#, {1, 4, 1, 
                4} -> ($CellContext`Mass $CellContext`r[]^(-4)) ((-2) \
$CellContext`Mass + $CellContext`r[]), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               FractionBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 SuperscriptBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["Sin",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    Sin,
                    Editable->False], "[", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False], "]"}], "2"]}], 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                    "2"]}], 
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False]]}]],
              
              Annotation[#, {1, 4, 4, 
                1} -> ($CellContext`Mass/$CellContext`r[]) Sin[
                   $CellContext`\[Theta][]]^2, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               FractionBox[
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}]}], 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "4"]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}], 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "4"]]}]],
              
              Annotation[#, {2, 1, 1, 
                2} -> ((
                  2 $CellContext`Mass) $CellContext`r[]^(-4)) ((-2) \
$CellContext`Mass + $CellContext`r[]), "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                FractionBox[
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Mass,
                   Editable->False]}], 
                 RowBox[{
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "-", 
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}], " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]]}]}]],
              
              Annotation[#, {2, 1, 2, 
                1} -> (((-2) $CellContext`Mass)/(
                 2 $CellContext`Mass - $CellContext`r[])) \
$CellContext`r[]^(-2), "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               FractionBox[
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], 
                RowBox[{
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "-", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], 
                  RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]]}]],
              
              Annotation[#, {2, 3, 2, 
                3} -> ($CellContext`Mass/(
                 2 $CellContext`Mass - $CellContext`r[])) \
$CellContext`r[]^(-2), "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               FractionBox[
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], 
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False]]}]],
              
              Annotation[#, {2, 3, 3, 
                2} -> $CellContext`Mass/$CellContext`r[], "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               FractionBox[
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], 
                RowBox[{
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "-", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], 
                  RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]]}]],
              
              Annotation[#, {2, 4, 2, 
                4} -> ($CellContext`Mass/(
                 2 $CellContext`Mass - $CellContext`r[])) \
$CellContext`r[]^(-2), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               FractionBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 SuperscriptBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["Sin",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    Sin,
                    Editable->False], "[", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False], "]"}], "2"]}], 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                    "2"]}], 
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False]]}]],
              
              Annotation[#, {2, 4, 4, 
                2} -> ($CellContext`Mass/$CellContext`r[]) Sin[
                   $CellContext`\[Theta][]]^2, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               FractionBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "-", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}]}], 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "4"]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}], 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "4"]]}]],
              
              Annotation[#, {3, 1, 1, 
                3} -> ($CellContext`Mass (
                  2 $CellContext`Mass - $CellContext`r[])) \
$CellContext`r[]^(-4), "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                FractionBox[
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]}]],
              
              Annotation[#, {3, 1, 3, 
                1} -> (-$CellContext`Mass)/$CellContext`r[], "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                FractionBox[
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], 
                 RowBox[{
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "-", 
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}], " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]]}]}]],
              
              Annotation[#, {3, 2, 2, 
                3} -> ((-$CellContext`Mass)/(
                 2 $CellContext`Mass - $CellContext`r[])) \
$CellContext`r[]^(-2), "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                FractionBox[
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]}]],
              
              Annotation[#, {3, 2, 3, 
                2} -> (-$CellContext`Mass)/$CellContext`r[], "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               FractionBox[
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False]}], 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], 
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False]]}]],
              
              Annotation[#, {3, 4, 3, 4} -> 
               2 ($CellContext`Mass/$CellContext`r[]), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                FractionBox[
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Mass,
                   Editable->False], " ", 
                  SuperscriptBox[
                   RowBox[{
                    InterpretationBox[
                    StyleBox["Sin",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    Sin,
                    Editable->False], "[", 
                    InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False], "]"}], "2"]}], 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                    "2"]}], 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]}]],
              
              Annotation[#, {3, 4, 4, 
                3} -> (((-2) $CellContext`Mass)/$CellContext`r[]) Sin[
                   $CellContext`\[Theta][]]^2, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               FractionBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "-", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}]}], 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "4"]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}], 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "4"]]}]],
              
              Annotation[#, {4, 1, 1, 
                4} -> ($CellContext`Mass (
                  2 $CellContext`Mass - $CellContext`r[])) \
$CellContext`r[]^(-4), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                FractionBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Mass,
                   Editable->False], " ", 
                  SuperscriptBox[
                   RowBox[{
                    InterpretationBox[
                    StyleBox["Sin",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    Sin,
                    Editable->False], "[", 
                    InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False], "]"}], "2"]}], 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                    "2"]}], 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]}]],
              
              Annotation[#, {4, 1, 4, 
                1} -> ((-$CellContext`Mass)/$CellContext`r[]) Sin[
                   $CellContext`\[Theta][]]^2, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                FractionBox[
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], 
                 RowBox[{
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "-", 
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}], " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]]}]}]],
              
              Annotation[#, {4, 2, 2, 
                4} -> ((-$CellContext`Mass)/(
                 2 $CellContext`Mass - $CellContext`r[])) \
$CellContext`r[]^(-2), "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                FractionBox[
                 RowBox[{
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Mass,
                   Editable->False], " ", 
                  SuperscriptBox[
                   RowBox[{
                    InterpretationBox[
                    StyleBox["Sin",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    Sin,
                    Editable->False], "[", 
                    InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False], "]"}], "2"]}], 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                    "2"]}], 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]}]],
              
              Annotation[#, {4, 2, 4, 
                2} -> ((-$CellContext`Mass)/$CellContext`r[]) Sin[
                   $CellContext`\[Theta][]]^2, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                FractionBox[
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Mass,
                   Editable->False]}], 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]}]],
              
              Annotation[#, {4, 3, 3, 
                4} -> (-2) ($CellContext`Mass/$CellContext`r[]), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               FractionBox[
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 SuperscriptBox[
                  RowBox[{
                   InterpretationBox[
                    StyleBox["Sin",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    Sin,
                    Editable->False], "[", 
                   InterpretationBox[
                    StyleBox["\[Theta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Theta][],
                    Editable->False], "]"}], "2"]}], 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                    "2"]}], 
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False]]}]],
              
              Annotation[#, {4, 3, 4, 
                3} -> ((2 $CellContext`Mass)/$CellContext`r[]) Sin[
                   $CellContext`\[Theta][]]^2, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " ", " ", "\[Delta]"},
         {"\[Alpha]", "\[Beta]", "\[Gamma]", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{{{0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 
     0}}, {{0, ((2 $CellContext`Mass) (
         2 $CellContext`Mass - $CellContext`r[])) $CellContext`r[]^(-4), 0, 
       0}, {((2 $CellContext`Mass)/(
        2 $CellContext`Mass - $CellContext`r[])) $CellContext`r[]^(-2), 0, 0, 
       0}, {0, 0, 0, 0}, {0, 0, 0, 0}}, {{
      0, 0, ($CellContext`Mass $CellContext`r[]^(-4)) ((-2) $CellContext`Mass + \
$CellContext`r[]), 0}, {0, 0, 0, 
      0}, {$CellContext`Mass/$CellContext`r[], 0, 0, 0}, {0, 0, 0, 0}}, {{
      0, 0, 0, ($CellContext`Mass $CellContext`r[]^(-4)) ((-2) \
$CellContext`Mass + $CellContext`r[])}, {0, 0, 0, 0}, {0, 0, 0, 
      0}, {($CellContext`Mass/$CellContext`r[]) Sin[
          $CellContext`\[Theta][]]^2, 0, 0, 0}}}, {{{
      0, ((2 $CellContext`Mass) $CellContext`r[]^(-4)) ((-2) \
$CellContext`Mass + $CellContext`r[]), 0, 
       0}, {(((-2) $CellContext`Mass)/(
        2 $CellContext`Mass - $CellContext`r[])) $CellContext`r[]^(-2), 0, 0, 
       0}, {0, 0, 0, 0}, {0, 0, 0, 0}}, {{0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0,
      0}, {0, 0, 0, 
     0}}, {{0, 0, 0, 0}, {
      0, 0, ($CellContext`Mass/(
        2 $CellContext`Mass - $CellContext`r[])) $CellContext`r[]^(-2), 0}, {
      0, $CellContext`Mass/$CellContext`r[], 0, 0}, {0, 0, 0, 0}}, {{0, 0, 0, 
      0}, {0, 0, 
       0, ($CellContext`Mass/(
        2 $CellContext`Mass - $CellContext`r[])) $CellContext`r[]^(-2)}, {0, 
      0, 0, 0}, {0, ($CellContext`Mass/$CellContext`r[]) Sin[
          $CellContext`\[Theta][]]^2, 0, 0}}}, {{{
      0, 0, ($CellContext`Mass (
         2 $CellContext`Mass - $CellContext`r[])) $CellContext`r[]^(-4), 0}, {
      0, 0, 0, 0}, {(-$CellContext`Mass)/$CellContext`r[], 0, 0, 0}, {0, 0, 0,
       0}}, {{0, 0, 0, 0}, {
      0, 0, ((-$CellContext`Mass)/(
        2 $CellContext`Mass - $CellContext`r[])) $CellContext`r[]^(-2), 0}, {
      0, (-$CellContext`Mass)/$CellContext`r[], 0, 0}, {0, 0, 0, 0}}, {{0, 0, 
     0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 
     0}}, {{0, 0, 0, 0}, {0, 0, 0, 0}, {
      0, 0, 0, 2 ($CellContext`Mass/$CellContext`r[])}, {
      0, 0, (((-2) $CellContext`Mass)/$CellContext`r[]) Sin[
          $CellContext`\[Theta][]]^2, 0}}}, {{{
      0, 0, 0, ($CellContext`Mass (
         2 $CellContext`Mass - $CellContext`r[])) $CellContext`r[]^(-4)}, {0, 
      0, 0, 0}, {0, 0, 0, 0}, {((-$CellContext`Mass)/$CellContext`r[]) Sin[
          $CellContext`\[Theta][]]^2, 0, 0, 0}}, {{0, 0, 0, 0}, {
      0, 0, 0, ((-$CellContext`Mass)/(
        2 $CellContext`Mass - $CellContext`r[])) $CellContext`r[]^(-2)}, {0, 
      0, 0, 0}, {0, ((-$CellContext`Mass)/$CellContext`r[]) Sin[
          $CellContext`\[Theta][]]^2, 0, 0}}, {{0, 0, 0, 0}, {0, 0, 0, 0}, {
      0, 0, 0, (-2) ($CellContext`Mass/$CellContext`r[])}, {
      0, 0, ((2 $CellContext`Mass)/$CellContext`r[]) Sin[
          $CellContext`\[Theta][]]^2, 0}}, {{0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0,
      0, 0}, {0, 0, 0, 
     0}}}}, {-$CellContext`schw, -$CellContext`schw, -$CellContext`schw, \
$CellContext`schw}, 
   0][-$CellContext`\[Alpha], -$CellContext`\[Beta], -$CellContext`\[Gamma], \
$CellContext`\[Delta]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728677184361485*^9, 3.733368533547579*^9},
 CellLabel->"Out[27]=",
 CellID->1028024297,ExpressionUUID->"55f04e4a-64f1-4413-8ed9-c27863a555e5"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{"riemF", "=", 
    RowBox[{"Simplify", "@", 
     RowBox[{"Riemann", "[", "cdF", "]"}]}]}], ")"}], "[", 
  RowBox[{
   RowBox[{"-", "\[Alpha]"}], ",", 
   RowBox[{"-", "\[Beta]"}], ",", 
   RowBox[{"-", "\[Gamma]"}], ",", "\[Delta]"}], "]"}]], "Input",
 CellLabel->"In[28]:=",
 CellID->694152686,ExpressionUUID->"64716f40-7639-439c-adde-9b2c258a567f"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                FractionBox[
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Mass,
                   Editable->False]}], 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "3"]]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "3"]]}]}]],
              
              Annotation[#, {1, 2, 1, 
                2} -> ((-2) $CellContext`Mass) $CellContext`r[]^(-3), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                FractionBox[
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Mass,
                   Editable->False]}], 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "3"]]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "3"]]}]}]],
              
              Annotation[#, {1, 2, 2, 
                1} -> ((-2) $CellContext`Mass) $CellContext`r[]^(-3), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               FractionBox[
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "3"]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "3"]]}]],
              
              Annotation[#, {1, 3, 1, 
                3} -> $CellContext`Mass $CellContext`r[]^(-3), "Tooltip"]& ], 
             
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               FractionBox[
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "3"]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "3"]]}]],
              
              Annotation[#, {1, 3, 3, 
                1} -> $CellContext`Mass $CellContext`r[]^(-3), "Tooltip"]& ], 
             
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               FractionBox[
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "3"]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "3"]]}]],
              
              Annotation[#, {1, 4, 1, 
                4} -> $CellContext`Mass $CellContext`r[]^(-3), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               FractionBox[
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "3"]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "3"]]}]],
              
              Annotation[#, {1, 4, 4, 
                1} -> $CellContext`Mass $CellContext`r[]^(-3), "Tooltip"]& ], 
             
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               FractionBox[
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False]}], 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "3"]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "3"]]}]],
              
              Annotation[#, {2, 1, 1, 
                2} -> (2 $CellContext`Mass) $CellContext`r[]^(-3), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               FractionBox[
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False]}], 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "3"]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "3"]]}]],
              
              Annotation[#, {2, 1, 2, 
                1} -> (2 $CellContext`Mass) $CellContext`r[]^(-3), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                FractionBox[
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "3"]]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "3"]]}]}]],
              
              Annotation[#, {2, 3, 2, 
                3} -> (-$CellContext`Mass) $CellContext`r[]^(-3), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               FractionBox[
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "3"]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "3"]]}]],
              
              Annotation[#, {2, 3, 3, 
                2} -> $CellContext`Mass $CellContext`r[]^(-3), "Tooltip"]& ], 
             
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                FractionBox[
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "3"]]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "3"]]}]}]],
              
              Annotation[#, {2, 4, 2, 
                4} -> (-$CellContext`Mass) $CellContext`r[]^(-3), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               FractionBox[
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "3"]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "3"]]}]],
              
              Annotation[#, {2, 4, 4, 
                2} -> $CellContext`Mass $CellContext`r[]^(-3), "Tooltip"]& ], 
             
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                FractionBox[
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "3"]]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "3"]]}]}]],
              
              Annotation[#, {3, 1, 1, 
                3} -> (-$CellContext`Mass) $CellContext`r[]^(-3), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                FractionBox[
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "3"]]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "3"]]}]}]],
              
              Annotation[#, {3, 1, 3, 
                1} -> (-$CellContext`Mass) $CellContext`r[]^(-3), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               FractionBox[
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "3"]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "3"]]}]],
              
              Annotation[#, {3, 2, 2, 
                3} -> $CellContext`Mass $CellContext`r[]^(-3), "Tooltip"]& ], 
             
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                FractionBox[
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "3"]]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "3"]]}]}]],
              
              Annotation[#, {3, 2, 3, 
                2} -> (-$CellContext`Mass) $CellContext`r[]^(-3), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               FractionBox[
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False]}], 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "3"]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "3"]]}]],
              
              Annotation[#, {3, 4, 3, 
                4} -> (2 $CellContext`Mass) $CellContext`r[]^(-3), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                FractionBox[
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Mass,
                   Editable->False]}], 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "3"]]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "3"]]}]}]],
              
              Annotation[#, {3, 4, 4, 
                3} -> ((-2) $CellContext`Mass) $CellContext`r[]^(-3), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                FractionBox[
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "3"]]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "3"]]}]}]],
              
              Annotation[#, {4, 1, 1, 
                4} -> (-$CellContext`Mass) $CellContext`r[]^(-3), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                FractionBox[
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "3"]]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "3"]]}]}]],
              
              Annotation[#, {4, 1, 4, 
                1} -> (-$CellContext`Mass) $CellContext`r[]^(-3), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               FractionBox[
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False], 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "3"]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "3"]]}]],
              
              Annotation[#, {4, 2, 2, 
                4} -> $CellContext`Mass $CellContext`r[]^(-3), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                FractionBox[
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "3"]]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False], 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "3"]]}]}]],
              
              Annotation[#, {4, 2, 4, 
                2} -> (-$CellContext`Mass) $CellContext`r[]^(-3), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               RowBox[{"-", 
                FractionBox[
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Mass,
                   Editable->False]}], 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "3"]]}],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "3"]]}]}]],
              
              Annotation[#, {4, 3, 3, 
                4} -> ((-2) $CellContext`Mass) $CellContext`r[]^(-3), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox[
               FractionBox[
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False]}], 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "3"]],
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`Mass, Editable -> False]}], 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "3"]]}]],
              
              Annotation[#, {4, 3, 4, 
                3} -> (2 $CellContext`Mass) $CellContext`r[]^(-3), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]},
           "RowsIndexed" -> {1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " ", " ", "\[Delta]"},
         {"\[Alpha]", "\[Beta]", "\[Gamma]", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{{{0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 
     0}}, {{0, ((-2) $CellContext`Mass) $CellContext`r[]^(-3), 0, 
       0}, {((-2) $CellContext`Mass) $CellContext`r[]^(-3), 0, 0, 0}, {0, 0, 
      0, 0}, {0, 0, 0, 0}}, {{
      0, 0, $CellContext`Mass $CellContext`r[]^(-3), 0}, {0, 0, 0, 
      0}, {$CellContext`Mass $CellContext`r[]^(-3), 0, 0, 0}, {0, 0, 0, 
      0}}, {{0, 0, 0, $CellContext`Mass $CellContext`r[]^(-3)}, {0, 0, 0, 
      0}, {0, 0, 0, 
      0}, {$CellContext`Mass $CellContext`r[]^(-3), 0, 0, 0}}}, {{{
      0, (2 $CellContext`Mass) $CellContext`r[]^(-3), 0, 
       0}, {(2 $CellContext`Mass) $CellContext`r[]^(-3), 0, 0, 0}, {0, 0, 0, 
      0}, {0, 0, 0, 0}}, {{0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 
     0}}, {{0, 0, 0, 0}, {
      0, 0, (-$CellContext`Mass) $CellContext`r[]^(-3), 0}, {
      0, $CellContext`Mass $CellContext`r[]^(-3), 0, 0}, {0, 0, 0, 0}}, {{0, 
      0, 0, 0}, {0, 0, 0, (-$CellContext`Mass) $CellContext`r[]^(-3)}, {0, 0, 
      0, 0}, {0, $CellContext`Mass $CellContext`r[]^(-3), 0, 0}}}, {{{
      0, 0, (-$CellContext`Mass) $CellContext`r[]^(-3), 0}, {0, 0, 0, 
      0}, {(-$CellContext`Mass) $CellContext`r[]^(-3), 0, 0, 0}, {0, 0, 0, 
      0}}, {{0, 0, 0, 0}, {
      0, 0, $CellContext`Mass $CellContext`r[]^(-3), 0}, {
      0, (-$CellContext`Mass) $CellContext`r[]^(-3), 0, 0}, {0, 0, 0, 0}}, {{
     0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 
     0}}, {{0, 0, 0, 0}, {0, 0, 0, 0}, {
      0, 0, 0, (2 $CellContext`Mass) $CellContext`r[]^(-3)}, {
      0, 0, ((-2) $CellContext`Mass) $CellContext`r[]^(-3), 0}}}, {{{
      0, 0, 0, (-$CellContext`Mass) $CellContext`r[]^(-3)}, {0, 0, 0, 0}, {0, 
      0, 0, 0}, {(-$CellContext`Mass) $CellContext`r[]^(-3), 0, 0, 0}}, {{0, 
      0, 0, 0}, {0, 0, 0, $CellContext`Mass $CellContext`r[]^(-3)}, {0, 0, 0, 
      0}, {0, (-$CellContext`Mass) $CellContext`r[]^(-3), 0, 0}}, {{0, 0, 0, 
      0}, {0, 0, 0, 0}, {
      0, 0, 0, ((-2) $CellContext`Mass) $CellContext`r[]^(-3)}, {
      0, 0, (2 $CellContext`Mass) $CellContext`r[]^(-3), 0}}, {{0, 0, 0, 0}, {
     0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 
     0}}}}, {-$CellContext`frame, -$CellContext`frame, -$CellContext`frame, \
$CellContext`frame}, 
   0][-$CellContext`\[Alpha], -$CellContext`\[Beta], -$CellContext`\[Gamma], \
$CellContext`\[Delta]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728677184572197*^9, 3.7333685349074793`*^9},
 CellLabel->"Out[28]=",
 CellID->1514571602,ExpressionUUID->"402cb11a-31ec-412b-92a2-2e3c26c7254c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["They are the same tensor, modulo a change of basis:", "MathCaption",
 CellID->2040276538,ExpressionUUID->"5c37ed58-3014-4275-b198-eb2117c531bd"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToCTensor", "[", 
   RowBox[{"riem", ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "frame"}], ",", 
      RowBox[{"-", "frame"}], ",", 
      RowBox[{"-", "frame"}], ",", "frame"}], "}"}]}], "]"}], "==", 
  "riemF"}]], "Input",
 CellLabel->"In[29]:=",
 CellID->2056382517,ExpressionUUID->"90a53eef-2a71-4209-9583-6cf7f128e157"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728677184976886*^9, 3.733368540008596*^9},
 CellLabel->"Out[29]=",
 CellID->1100047659,ExpressionUUID->"feb3adcc-7438-4e7e-8161-eccd7ca185af"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Ricci tensors:", "MathCaption",
 CellID->1852112204,ExpressionUUID->"d6757a56-afb4-446d-8cf5-97048ef9d096"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Ricci", "[", "cd", "]"}]], "Input",
 CellLabel->"In[30]:=",
 CellID->674349363,ExpressionUUID->"e2b5c0cb-2d4a-4314-a2be-59fc2c98028e"],

Cell[BoxData["Zero"], "Output",
 CellChangeTimes->{3.72867718544275*^9, 3.7333685457319393`*^9},
 CellLabel->"Out[30]=",
 CellID->16490870,ExpressionUUID->"6a3575e5-251d-46ca-8b39-51a88bd182c5"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"UndefTensor", "[", "Jacobianframeschw", "]"}]], "Input",
 CellChangeTimes->{{3.733368566739477*^9, 3.7333685743970127`*^9}},
 CellLabel->"In[34]:=",
 CellID->2127000328,ExpressionUUID->"167c5b18-1e0b-440a-9861-3811c560d19c"],

Cell[BoxData[
 RowBox[{"UndefBasis", "[", "frame", "]"}]], "Input",
 CellLabel->"In[35]:=",
 CellID->649493526,ExpressionUUID->"b06347fc-a6a5-4acb-9d25-152ede547029"],

Cell[BoxData[
 RowBox[{"UndefChart", "[", "schw", "]"}]], "Input",
 CellLabel->"In[36]:=",
 CellID->378313781,ExpressionUUID->"28da1443-3736-4b8b-8e15-18c39a213d6d"],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{
  "SchwarzschildMatrix", ",", "cd", ",", "cdF", ",", "metric", ",", "metricF",
    ",", "riem", ",", "riemF"}], "]"}]], "Input",
 CellLabel->"In[37]:=",
 CellID->1768122147,ExpressionUUID->"f5e56003-7ebd-4f46-9171-8c37c6d67672"],

Cell[BoxData[
 RowBox[{"UndefConstantSymbol", "[", "Mass", "]"}]], "Input",
 CellLabel->"In[38]:=",
 CellID->2097701090,ExpressionUUID->"081656fe-2706-45bd-a18c-c99dd98bdae1"],

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M4", "]"}]], "Input",
 CellLabel->"In[39]:=",
 CellID->823186831,ExpressionUUID->"66f00f8f-835f-41e5-bc20-bd014179d2b2"],

Cell[BoxData[
 RowBox[{"ClearxCobaCache", "[", "All", "]"}]], "Input",
 CellLabel->"In[40]:=",
 CellID->1211414643,ExpressionUUID->"7b7289df-49d5-4ac6-acb8-90c8727545e3"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.6. Examples comparing with RGTC, by Bonanos", "Subsection",
 CellChangeTimes->{{3.728674035065624*^9, 3.7286740352900867`*^9}, {
   3.728675278624102*^9, 3.728675278760154*^9}, 3.7908767395227203`*^9},
 CellID->2100931341,ExpressionUUID->"cab48f9c-968b-4465-8703-4487b0c4d6ca"],

Cell[TextData[{
 "In order to understand why we do what we do you need to read Bonanos\
\[CloseCurlyQuote] examples in the documentation of RGTC [",
 ButtonBox["link",
  BaseStyle->"Hyperlink",
  ButtonData->{
    URL["http://www.inp.demokritos.gr/~sbonano/"], None},
  ButtonNote->"http://www.inp.demokritos.gr/~sbonano/"],
 "]."
}], "Text",
 CellID->1426924213,ExpressionUUID->"001e0087-9c76-4dbe-ba2a-c0614c0444d3"],

Cell[CellGroupData[{

Cell["3.6.0. Load xCoba", "Subsubsection",
 CellChangeTimes->{{3.7286771125390873`*^9, 3.7286771127222157`*^9}, 
   3.7908767620386047`*^9},
 CellID->1365361842,ExpressionUUID->"c50e8747-19c6-4348-8c1f-05ddea1300c5"],

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<xAct`xCoba`\>\"", "]"}]], "Input",
 CellLabel->"In[160]:=",
 CellID->1773710299,ExpressionUUID->"bee02021-9c7b-4ab6-a085-9a6dbc657429"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"$DefInfoQ", "=", 
   RowBox[{"$UndefInfoQ", "=", "False"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"$PrePrint", "=", "ScreenDollarIndices"}], ";"}]}], "Input",
 CellLabel->"In[161]:=",
 CellID->1528491286,ExpressionUUID->"e52adf5e-dad0-40dc-a10d-6fe1ca7f1b5f"]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
3.6.1. Flat spacetime in rotating prolate spheroidal coordinates\
\>", "Subsubsection",
 CellChangeTimes->{{3.728677114482016*^9, 3.728677114658161*^9}, 
   3.790876763423274*^9},
 CellID->1153658388,ExpressionUUID->"c69ee9a4-3b2d-4d75-ab06-81adee5a4efa"],

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M", ",", "4", ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f"}], "}"}]}], 
  "]"}]], "Input",
 CellLabel->"In[163]:=",
 CellID->2054162000,ExpressionUUID->"88af7186-ceb1-4bd6-9110-d14c58513b10"],

Cell[BoxData[
 RowBox[{"DefChart", "[", 
  RowBox[{"rps", ",", "M", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2", ",", "3"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"t", "[", "]"}], ",", 
     RowBox[{"\[Phi]", "[", "]"}], ",", 
     RowBox[{"\[Lambda]", "[", "]"}], ",", 
     RowBox[{"\[Mu]", "[", "]"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[164]:=",
 CellID->1353624828,ExpressionUUID->"ebc9095c-0632-4a34-9769-5625a994fdc6"],

Cell[BoxData[
 RowBox[{"DefConstantSymbol", "[", "\[CapitalOmega]", "]"}]], "Input",
 CellLabel->"In[165]:=",
 CellID->2078211525,ExpressionUUID->"52f9f81c-9e5c-48d0-b831-4c1555a51bf7"],

Cell[BoxData[
 RowBox[{
  RowBox[{"met", "=", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"1", "-", 
          RowBox[{
           RowBox[{"\[CapitalOmega]", "^", "2"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"\[Lambda]", "[", "]"}], "^", "2"}], "-", "1"}], ")"}], 
           
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{
              RowBox[{"\[Mu]", "[", "]"}], "^", "2"}]}], ")"}]}]}], ",", 
         RowBox[{
          RowBox[{"-", "\[CapitalOmega]"}], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{
             RowBox[{"\[Lambda]", "[", "]"}], "^", "2"}], "-", "1"}], ")"}], 
          RowBox[{"(", 
           RowBox[{"1", "-", 
            RowBox[{
             RowBox[{"\[Mu]", "[", "]"}], "^", "2"}]}], ")"}]}], ",", "0", 
         ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "\[CapitalOmega]"}], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{
             RowBox[{"\[Lambda]", "[", "]"}], "^", "2"}], "-", "1"}], ")"}], 
          RowBox[{"(", 
           RowBox[{"1", "-", 
            RowBox[{
             RowBox[{"\[Mu]", "[", "]"}], "^", "2"}]}], ")"}]}], ",", 
         RowBox[{
          RowBox[{"-", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"\[Lambda]", "[", "]"}], "^", "2"}], "-", "1"}], 
            ")"}]}], 
          RowBox[{"(", 
           RowBox[{"1", "-", 
            RowBox[{
             RowBox[{"\[Mu]", "[", "]"}], "^", "2"}]}], ")"}]}], ",", "0", 
         ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", 
         RowBox[{
          RowBox[{"-", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"\[Lambda]", "[", "]"}], "^", "2"}], "-", 
             RowBox[{
              RowBox[{"\[Mu]", "[", "]"}], "^", "2"}]}], ")"}]}], "/", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{
             RowBox[{"\[Lambda]", "[", "]"}], "^", "2"}], "-", "1"}], ")"}]}],
          ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", 
         RowBox[{
          RowBox[{"-", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"\[Lambda]", "[", "]"}], "^", "2"}], "-", 
             RowBox[{
              RowBox[{"\[Mu]", "[", "]"}], "^", "2"}]}], ")"}]}], "/", 
          RowBox[{"(", 
           RowBox[{"1", "-", 
            RowBox[{
             RowBox[{"\[Mu]", "[", "]"}], "^", "2"}]}], ")"}]}]}], "}"}]}], 
      "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "rps"}], ",", 
       RowBox[{"-", "rps"}]}], "}"}]}], "]"}]}], ";"}]], "Input",
 CellLabel->"In[166]:=",
 CellID->1827065867,ExpressionUUID->"391758a9-c29b-4a14-9589-ab34dd4d44c6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"met", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}]}], "]"}]], "Input",
 CellLabel->"In[167]:=",
 CellID->627358216,ExpressionUUID->"e2c09bf5-3fae-4f1e-818c-2d381b0b25de"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{"1", "-", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["\[CapitalOmega]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[CapitalOmega],
                Editable->False], "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["\[Lambda]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Lambda][],
                  Editable->False], "2"]}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{"1", "-", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["\[Mu]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Mu][],
                  Editable->False], "2"]}], ")"}]}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{"1", "-", 
                RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalOmega]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalOmega], 
                    Editable -> False], "2"], " ", 
                  RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "1"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "\[Lambda]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Lambda][], Editable -> False], "2"]}], 
                    ")"}], " ", 
                  RowBox[{"(", 
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "\[Mu]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`\[Mu][], Editable -> False], "2"]}], 
                    ")"}]}]}]}]],
           
           Annotation[#, {1, 1} -> 
            1 - ($CellContext`\[CapitalOmega]^2 (-1 + \
$CellContext`\[Lambda][]^2)) (1 - $CellContext`\[Mu][]^2), "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{"-", 
              InterpretationBox[
               StyleBox["\[CapitalOmega]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[CapitalOmega],
               Editable->False]}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Lambda]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Lambda][],
                 Editable->False], "2"]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "-", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Mu]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Mu][],
                 Editable->False], "2"]}], ")"}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{"-", 
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"\[CapitalOmega]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalOmega], 
                   Editable -> False]}], " ", 
                RowBox[{"(", 
                  RowBox[{
                    RowBox[{"-", "1"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "\[Lambda]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Lambda][], Editable -> False], "2"]}], 
                  ")"}], " ", 
                RowBox[{"(", 
                  RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "\[Mu]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`\[Mu][], Editable -> False], "2"]}], 
                  ")"}]}]}]],
           
           Annotation[#, {1, 
             2} -> ((-$CellContext`\[CapitalOmega]) (-1 + $CellContext`\
\[Lambda][]^2)) (1 - $CellContext`\[Mu][]^2), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{"-", 
              InterpretationBox[
               StyleBox["\[CapitalOmega]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[CapitalOmega],
               Editable->False]}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Lambda]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Lambda][],
                 Editable->False], "2"]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "-", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Mu]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Mu][],
                 Editable->False], "2"]}], ")"}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{"-", 
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"\[CapitalOmega]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalOmega], 
                   Editable -> False]}], " ", 
                RowBox[{"(", 
                  RowBox[{
                    RowBox[{"-", "1"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "\[Lambda]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Lambda][], Editable -> False], "2"]}], 
                  ")"}], " ", 
                RowBox[{"(", 
                  RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "\[Mu]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`\[Mu][], Editable -> False], "2"]}], 
                  ")"}]}]}]],
           
           Annotation[#, {2, 
             1} -> ((-$CellContext`\[CapitalOmega]) (-1 + $CellContext`\
\[Lambda][]^2)) (1 - $CellContext`\[Mu][]^2), "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{"(", 
              RowBox[{"1", "-", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Lambda]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Lambda][],
                 Editable->False], "2"]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "-", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Mu]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Mu][],
                 Editable->False], "2"]}], ")"}]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{"(", 
                  RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "\[Lambda]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Lambda][], Editable -> False], "2"]}], 
                  ")"}], " ", 
                RowBox[{"(", 
                  RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "\[Mu]", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`\[Mu][], Editable -> False], "2"]}], 
                  ")"}]}]}]],
           
           Annotation[#, {2, 
             2} -> (1 - $CellContext`\[Lambda][]^2) (
              1 - $CellContext`\[Mu][]^2), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{
              RowBox[{"-", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Lambda]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Lambda][],
                 Editable->False], "2"]}], "+", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["\[Mu]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Mu][],
                Editable->False], "2"]}], 
             RowBox[{
              RowBox[{"-", "1"}], "+", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["\[Lambda]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Lambda][],
                Editable->False], "2"]}]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 RowBox[{"-", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "\[Lambda]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Lambda][], Editable -> False], "2"]}], "+", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox[
                   "\[Mu]", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`\[Mu][], Editable -> False], "2"]}], 
               RowBox[{
                 RowBox[{"-", "1"}], "+", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox[
                   "\[Lambda]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                   $CellContext`\[Lambda][], Editable -> False], "2"]}]]}]],
           
           Annotation[#, {3, 
             3} -> (-1 + $CellContext`\[Lambda][]^2)^(-1) (-$CellContext`\
\[Lambda][]^2 + $CellContext`\[Mu][]^2), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{
              RowBox[{"-", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Lambda]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Lambda][],
                 Editable->False], "2"]}], "+", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["\[Mu]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Mu][],
                Editable->False], "2"]}], 
             RowBox[{"1", "-", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["\[Mu]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Mu][],
                Editable->False], "2"]}]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "4"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 RowBox[{"-", 
                   SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "\[Lambda]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Lambda][], Editable -> False], "2"]}], "+", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox[
                   "\[Mu]", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`\[Mu][], Editable -> False], "2"]}], 
               RowBox[{"1", "-", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox[
                   "\[Mu]", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`\[Mu][], Editable -> False], "2"]}]]}]],
           
           Annotation[#, {4, 
             4} -> (1 - $CellContext`\[Mu][]^2)^(-1) \
(-$CellContext`\[Lambda][]^2 + $CellContext`\[Mu][]^2), "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " "},
         {"a", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{
    1 - ($CellContext`\[CapitalOmega]^2 (-1 + $CellContext`\[Lambda][]^2)) (
      1 - $CellContext`\[Mu][]^2), ((-$CellContext`\[CapitalOmega]) (-1 + \
$CellContext`\[Lambda][]^2)) (1 - $CellContext`\[Mu][]^2), 0, 
     0}, {((-$CellContext`\[CapitalOmega]) (-1 + $CellContext`\[Lambda][]^2)) \
(1 - $CellContext`\[Mu][]^2), (1 - $CellContext`\[Lambda][]^2) (
      1 - $CellContext`\[Mu][]^2), 0, 0}, {
    0, 0, (-1 + $CellContext`\[Lambda][]^2)^(-1) (-$CellContext`\[Lambda][]^2 + \
$CellContext`\[Mu][]^2), 0}, {
    0, 0, 0, (
       1 - $CellContext`\[Mu][]^2)^(-1) (-$CellContext`\[Lambda][]^2 + \
$CellContext`\[Mu][]^2)}}, {-$CellContext`rps, -$CellContext`rps}, 
   0][-$CellContext`a, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.728677238429143*^9},
 CellLabel->"Out[167]=",
 CellID->1740023790,ExpressionUUID->"0215dc22-76d5-4bca-808e-c60fe4b9a4d7"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"$CVSimplify", "=", "Simplify"}], ";"}]], "Input",
 CellLabel->"In[168]:=",
 CellID->227810920,ExpressionUUID->"47b29c6c-df67-4651-b4a2-f3ff8e9ddd74"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SetCMetric", "[", 
   RowBox[{"met", ",", "rps", ",", 
    RowBox[{"SignatureOfMetric", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{"3", ",", "1", ",", "0"}], "}"}]}]}], "]"}], "//", 
  "Timing"}]], "Input",
 CellLabel->"In[169]:=",
 CellID->2002273486,ExpressionUUID->"7f57fad2-ff0e-4ebe-a03d-596779c4e2fc"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.07745`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.7286772387087803`*^9},
 CellLabel->"Out[169]=",
 CellID->147533684,ExpressionUUID->"3f685232-cf09-4b6f-a35a-3f08c5f3a762"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"MetricCompute", "[", 
   RowBox[{"met", ",", "rps", ",", 
    RowBox[{"\"\<Riemann\>\"", "[", 
     RowBox[{
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}]}], "]"}]}], "]"}], "//", "Timing"}]], "Input",
 CellLabel->"In[170]:=",
 CellID->1301690639,ExpressionUUID->"95c52da4-9879-4da4-a3e3-004c6e4ac249"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.034032`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.728677238870236*^9},
 CellLabel->"Out[170]=",
 CellID->1402498350,ExpressionUUID->"fff86c91-68fb-4d37-88c8-767a1908e2a3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"cd", "=", 
  RowBox[{"CovDOfMetric", "[", "met", "]"}]}]], "Input",
 CellLabel->"In[171]:=",
 CellID->1075298082,ExpressionUUID->"78d02b5e-23f2-495a-9918-7c7a9018dbe2"],

Cell[BoxData[
 RowBox[{"CCovD", "[", 
  RowBox[{"PDrps", ",", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           FractionBox[
            RowBox[{
             InterpretationBox[
              StyleBox["\[CapitalOmega]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[CapitalOmega],
              Editable->False], " ", 
             InterpretationBox[
              StyleBox["\[Lambda]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Lambda][],
              Editable->False]}], 
            RowBox[{
             RowBox[{"-", "1"}], "+", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Lambda]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Lambda][],
               Editable->False], "2"]}]], ",", 
           FractionBox[
            RowBox[{
             InterpretationBox[
              StyleBox["\[CapitalOmega]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[CapitalOmega],
              Editable->False], " ", 
             InterpretationBox[
              StyleBox["\[Mu]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Mu][],
              Editable->False]}], 
            RowBox[{
             RowBox[{"-", "1"}], "+", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Mu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Mu][],
               Editable->False], "2"]}]]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           FractionBox[
            InterpretationBox[
             StyleBox["\[Lambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Lambda][],
             Editable->False], 
            RowBox[{
             RowBox[{"-", "1"}], "+", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Lambda]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Lambda][],
               Editable->False], "2"]}]], ",", 
           FractionBox[
            InterpretationBox[
             StyleBox["\[Mu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu][],
             Editable->False], 
            RowBox[{
             RowBox[{"-", "1"}], "+", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Mu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Mu][],
               Editable->False], "2"]}]]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           FractionBox[
            RowBox[{
             InterpretationBox[
              StyleBox["\[CapitalOmega]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[CapitalOmega],
              Editable->False], " ", 
             InterpretationBox[
              StyleBox["\[Lambda]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Lambda][],
              Editable->False]}], 
            RowBox[{
             RowBox[{"-", "1"}], "+", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Lambda]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Lambda][],
               Editable->False], "2"]}]], ",", 
           FractionBox[
            InterpretationBox[
             StyleBox["\[Lambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Lambda][],
             Editable->False], 
            RowBox[{
             RowBox[{"-", "1"}], "+", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Lambda]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Lambda][],
               Editable->False], "2"]}]], ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           FractionBox[
            RowBox[{
             InterpretationBox[
              StyleBox["\[CapitalOmega]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[CapitalOmega],
              Editable->False], " ", 
             InterpretationBox[
              StyleBox["\[Mu]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Mu][],
              Editable->False]}], 
            RowBox[{
             RowBox[{"-", "1"}], "+", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Mu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Mu][],
               Editable->False], "2"]}]], ",", 
           FractionBox[
            InterpretationBox[
             StyleBox["\[Mu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu][],
             Editable->False], 
            RowBox[{
             RowBox[{"-", "1"}], "+", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Mu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Mu][],
               Editable->False], "2"]}]], ",", "0", ",", "0"}], "}"}]}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           FractionBox[
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[CapitalOmega]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[CapitalOmega],
               Editable->False], "2"], " ", 
             InterpretationBox[
              StyleBox["\[Lambda]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Lambda][],
              Editable->False], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Lambda]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Lambda][],
                 Editable->False], "2"]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Mu]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Mu][],
                 Editable->False], "2"]}], ")"}]}], 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Lambda]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Lambda][],
               Editable->False], "2"], "-", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Mu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Mu][],
               Editable->False], "2"]}]], ",", 
           FractionBox[
            RowBox[{
             InterpretationBox[
              StyleBox["\[CapitalOmega]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[CapitalOmega],
              Editable->False], " ", 
             InterpretationBox[
              StyleBox["\[Lambda]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Lambda][],
              Editable->False], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Lambda]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Lambda][],
                 Editable->False], "2"]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Mu]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Mu][],
                 Editable->False], "2"]}], ")"}]}], 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Lambda]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Lambda][],
               Editable->False], "2"], "-", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Mu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Mu][],
               Editable->False], "2"]}]], ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           FractionBox[
            RowBox[{
             InterpretationBox[
              StyleBox["\[CapitalOmega]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[CapitalOmega],
              Editable->False], " ", 
             InterpretationBox[
              StyleBox["\[Lambda]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Lambda][],
              Editable->False], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Lambda]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Lambda][],
                 Editable->False], "2"]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Mu]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Mu][],
                 Editable->False], "2"]}], ")"}]}], 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Lambda]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Lambda][],
               Editable->False], "2"], "-", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Mu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Mu][],
               Editable->False], "2"]}]], ",", 
           FractionBox[
            RowBox[{
             InterpretationBox[
              StyleBox["\[Lambda]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Lambda][],
              Editable->False], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Lambda]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Lambda][],
                 Editable->False], "2"]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Mu]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Mu][],
                 Editable->False], "2"]}], ")"}]}], 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Lambda]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Lambda][],
               Editable->False], "2"], "-", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Mu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Mu][],
               Editable->False], "2"]}]], ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           FractionBox[
            RowBox[{
             InterpretationBox[
              StyleBox["\[Lambda]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Lambda][],
              Editable->False], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Mu]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Mu][],
                 Editable->False], "2"]}], ")"}]}], 
            RowBox[{
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Lambda]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Lambda][],
                 Editable->False], "2"]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Lambda]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Lambda][],
                 Editable->False], "2"], "-", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Mu]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Mu][],
                 Editable->False], "2"]}], ")"}]}]], ",", 
           FractionBox[
            InterpretationBox[
             StyleBox["\[Mu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu][],
             Editable->False], 
            RowBox[{
             RowBox[{"-", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["\[Lambda]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Lambda][],
                Editable->False], "2"]}], "+", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Mu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Mu][],
               Editable->False], "2"]}]]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           FractionBox[
            InterpretationBox[
             StyleBox["\[Mu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu][],
             Editable->False], 
            RowBox[{
             RowBox[{"-", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["\[Lambda]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Lambda][],
                Editable->False], "2"]}], "+", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Mu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Mu][],
               Editable->False], "2"]}]], ",", 
           FractionBox[
            RowBox[{
             InterpretationBox[
              StyleBox["\[Lambda]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Lambda][],
              Editable->False], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Lambda]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Lambda][],
                 Editable->False], "2"]}], ")"}]}], 
            RowBox[{
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Lambda]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Lambda][],
                 Editable->False], "2"], "-", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Mu]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Mu][],
                 Editable->False], "2"]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Mu]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Mu][],
                 Editable->False], "2"]}], ")"}]}]]}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           RowBox[{"-", 
            FractionBox[
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["\[CapitalOmega]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[CapitalOmega],
                Editable->False], "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["\[Lambda]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Lambda][],
                  Editable->False], "2"]}], ")"}], " ", 
              InterpretationBox[
               StyleBox["\[Mu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Mu][],
               Editable->False], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["\[Mu]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Mu][],
                  Editable->False], "2"]}], ")"}]}], 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["\[Lambda]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Lambda][],
                Editable->False], "2"], "-", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["\[Mu]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Mu][],
                Editable->False], "2"]}]]}], ",", 
           RowBox[{"-", 
            FractionBox[
             RowBox[{
              InterpretationBox[
               StyleBox["\[CapitalOmega]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[CapitalOmega],
               Editable->False], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["\[Lambda]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Lambda][],
                  Editable->False], "2"]}], ")"}], " ", 
              InterpretationBox[
               StyleBox["\[Mu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Mu][],
               Editable->False], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["\[Mu]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Mu][],
                  Editable->False], "2"]}], ")"}]}], 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["\[Lambda]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Lambda][],
                Editable->False], "2"], "-", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["\[Mu]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Mu][],
                Editable->False], "2"]}]]}], ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"-", 
            FractionBox[
             RowBox[{
              InterpretationBox[
               StyleBox["\[CapitalOmega]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[CapitalOmega],
               Editable->False], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["\[Lambda]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Lambda][],
                  Editable->False], "2"]}], ")"}], " ", 
              InterpretationBox[
               StyleBox["\[Mu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Mu][],
               Editable->False], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["\[Mu]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Mu][],
                  Editable->False], "2"]}], ")"}]}], 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["\[Lambda]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Lambda][],
                Editable->False], "2"], "-", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["\[Mu]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Mu][],
                Editable->False], "2"]}]]}], ",", 
           FractionBox[
            RowBox[{
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Lambda]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Lambda][],
                 Editable->False], "2"]}], ")"}], " ", 
             InterpretationBox[
              StyleBox["\[Mu]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Mu][],
              Editable->False], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Mu]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Mu][],
                 Editable->False], "2"]}], ")"}]}], 
            RowBox[{
             RowBox[{"-", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["\[Lambda]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Lambda][],
                Editable->False], "2"]}], "+", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Mu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Mu][],
               Editable->False], "2"]}]], ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           FractionBox[
            RowBox[{
             InterpretationBox[
              StyleBox["\[Mu]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Mu][],
              Editable->False], "-", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Mu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Mu][],
               Editable->False], "3"]}], 
            RowBox[{
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Lambda]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Lambda][],
                 Editable->False], "2"]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Lambda]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Lambda][],
                 Editable->False], "2"], "-", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Mu]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Mu][],
                 Editable->False], "2"]}], ")"}]}]], ",", 
           FractionBox[
            InterpretationBox[
             StyleBox["\[Lambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Lambda][],
             Editable->False], 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Lambda]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Lambda][],
               Editable->False], "2"], "-", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Mu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Mu][],
               Editable->False], "2"]}]]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           FractionBox[
            InterpretationBox[
             StyleBox["\[Lambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Lambda][],
             Editable->False], 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Lambda]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Lambda][],
               Editable->False], "2"], "-", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Mu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Mu][],
               Editable->False], "2"]}]], ",", 
           RowBox[{"-", 
            FractionBox[
             RowBox[{
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["\[Lambda]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Lambda][],
                  Editable->False], "2"]}], ")"}], " ", 
              InterpretationBox[
               StyleBox["\[Mu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Mu][],
               Editable->False]}], 
             RowBox[{
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["\[Lambda]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Lambda][],
                  Editable->False], "2"], "-", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["\[Mu]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Mu][],
                  Editable->False], "2"]}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["\[Mu]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Mu][],
                  Editable->False], "2"]}], ")"}]}]]}]}], "}"}]}], "}"}]}], 
      "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"rps", ",", 
       RowBox[{"-", "rps"}], ",", 
       RowBox[{"-", "rps"}]}], "}"}], ",", "0"}], "]"}], ",", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"1", "-", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["\[CapitalOmega]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[CapitalOmega],
             Editable->False], "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "1"}], "+", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Lambda]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Lambda][],
               Editable->False], "2"]}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{"1", "-", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Mu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Mu][],
               Editable->False], "2"]}], ")"}]}]}], ",", 
         RowBox[{
          RowBox[{"-", 
           InterpretationBox[
            StyleBox["\[CapitalOmega]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[CapitalOmega],
            Editable->False]}], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "1"}], "+", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["\[Lambda]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Lambda][],
              Editable->False], "2"]}], ")"}], " ", 
          RowBox[{"(", 
           RowBox[{"1", "-", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["\[Mu]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Mu][],
              Editable->False], "2"]}], ")"}]}], ",", "0", ",", "0"}], "}"}], 
       ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{
          RowBox[{"-", 
           InterpretationBox[
            StyleBox["\[CapitalOmega]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[CapitalOmega],
            Editable->False]}], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "1"}], "+", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["\[Lambda]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Lambda][],
              Editable->False], "2"]}], ")"}], " ", 
          RowBox[{"(", 
           RowBox[{"1", "-", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["\[Mu]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Mu][],
              Editable->False], "2"]}], ")"}]}], ",", 
         RowBox[{
          RowBox[{"(", 
           RowBox[{"1", "-", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["\[Lambda]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Lambda][],
              Editable->False], "2"]}], ")"}], " ", 
          RowBox[{"(", 
           RowBox[{"1", "-", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["\[Mu]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Mu][],
              Editable->False], "2"]}], ")"}]}], ",", "0", ",", "0"}], "}"}], 
       ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", 
         FractionBox[
          RowBox[{
           RowBox[{"-", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["\[Lambda]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Lambda][],
              Editable->False], "2"]}], "+", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["\[Mu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu][],
             Editable->False], "2"]}], 
          RowBox[{
           RowBox[{"-", "1"}], "+", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["\[Lambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Lambda][],
             Editable->False], "2"]}]], ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", 
         FractionBox[
          RowBox[{
           RowBox[{"-", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["\[Lambda]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Lambda][],
              Editable->False], "2"]}], "+", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["\[Mu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu][],
             Editable->False], "2"]}], 
          RowBox[{"1", "-", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["\[Mu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu][],
             Editable->False], "2"]}]]}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "rps"}], ",", 
       RowBox[{"-", "rps"}]}], "}"}], ",", "0"}], "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728677239016295*^9},
 CellLabel->"Out[171]=",
 CellID->1009258978,ExpressionUUID->"cb1498ab-b4e9-4a58-b962-970af5890c9c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Riemann", "[", "cd", "]"}]], "Input",
 CellLabel->"In[172]:=",
 CellID->646111261,ExpressionUUID->"d1cd0e03-ac6a-4176-a280-630e1649898a"],

Cell[BoxData["Zero"], "Output",
 CellChangeTimes->{3.728677239220049*^9},
 CellLabel->"Out[172]=",
 CellID->1789799240,ExpressionUUID->"42eb32d1-b256-4ec5-8f87-6a36b143cd8e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TensorDerivative", "[", 
  RowBox[{"met", ",", "cd"}], "]"}]], "Input",
 CellLabel->"In[173]:=",
 CellID->357195227,ExpressionUUID->"e9dfd15f-74a8-44c4-8386-665c0dafc540"],

Cell[BoxData["Zero"], "Output",
 CellChangeTimes->{3.728677239463093*^9},
 CellLabel->"Out[173]=",
 CellID->1352383151,ExpressionUUID->"5f0aa0d2-0935-405f-ab19-f14705e93f86"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TensorDerivative", "[", 
  RowBox[{
   RowBox[{"Inv", "[", "met", "]"}], ",", "cd"}], "]"}]], "Input",
 CellLabel->"In[174]:=",
 CellID->203745033,ExpressionUUID->"7389ebdf-b398-4505-b633-02628f487f51"],

Cell[BoxData["Zero"], "Output",
 CellChangeTimes->{3.728677239610124*^9},
 CellLabel->"Out[174]=",
 CellID->536699009,ExpressionUUID->"b2a5d210-7868-4b44-b53b-b9f49d2a28a9"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{"met", ",", "cd"}], "]"}]], "Input",
 CellLabel->"In[175]:=",
 CellID->2117368227,ExpressionUUID->"161e63d5-b840-4c45-9aa7-393360289890"],

Cell[BoxData[
 RowBox[{"UndefChart", "[", "rps", "]"}]], "Input",
 CellLabel->"In[176]:=",
 CellID->656276752,ExpressionUUID->"3b8069d0-a689-46bd-8f77-4b583ac218d5"],

Cell[BoxData[
 RowBox[{"UndefConstantSymbol", "[", "\[CapitalOmega]", "]"}]], "Input",
 CellLabel->"In[177]:=",
 CellID->567359872,ExpressionUUID->"034a186e-5920-45be-ae35-50684d69fd25"],

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M", "]"}]], "Input",
 CellLabel->"In[178]:=",
 CellID->789976024,ExpressionUUID->"e52d83c7-1c66-48f6-82fc-fea259a3ae05"],

Cell[BoxData[
 RowBox[{"ClearxCobaCache", "[", "All", "]"}]], "Input",
 CellLabel->"In[179]:=",
 CellID->1383129614,ExpressionUUID->"1b9cf383-82d4-4314-8f59-54b2be2cb1df"]
}, Closed]],

Cell[CellGroupData[{

Cell["3.6.2. Schwarzschild", "Subsubsection",
 CellChangeTimes->{{3.728677116097837*^9, 3.728677116297902*^9}, 
   3.790876765359956*^9},
 CellID->1896427370,ExpressionUUID->"151cf641-9819-4eff-af98-759a6ca08fc8"],

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M", ",", "4", ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f"}], "}"}]}], 
  "]"}]], "Input",
 CellLabel->"In[180]:=",
 CellID->302518816,ExpressionUUID->"187f3e31-d974-4787-9a43-6e9c614b46f8"],

Cell[BoxData[
 RowBox[{"DefChart", "[", 
  RowBox[{"sch", ",", "M", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2", ",", "3"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"t", "[", "]"}], ",", 
     RowBox[{"r", "[", "]"}], ",", 
     RowBox[{"\[Theta]", "[", "]"}], ",", 
     RowBox[{"\[Phi]", "[", "]"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[181]:=",
 CellID->1339948063,ExpressionUUID->"d55e052f-6448-4736-99ec-5b83207ea047"],

Cell[BoxData[
 RowBox[{"DefConstantSymbol", "[", "Mass", "]"}]], "Input",
 CellLabel->"In[182]:=",
 CellID->1993572399,ExpressionUUID->"5e3fbd19-869e-407f-9769-e3c1214783b9"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"met", "=", 
  RowBox[{"CTensor", "[", 
   RowBox[{
    RowBox[{"DiagonalMatrix", "[", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"1", "-", 
        RowBox[{"2", 
         RowBox[{"Mass", "/", 
          RowBox[{"r", "[", "]"}]}]}]}], ",", 
       RowBox[{
        RowBox[{"-", "1"}], "/", 
        RowBox[{"(", 
         RowBox[{"1", "-", 
          RowBox[{"2", 
           RowBox[{"Mass", "/", 
            RowBox[{"r", "[", "]"}]}]}]}], ")"}]}], ",", 
       RowBox[{"-", 
        RowBox[{
         RowBox[{"r", "[", "]"}], "^", "2"}]}], ",", 
       RowBox[{
        RowBox[{"-", 
         RowBox[{
          RowBox[{"r", "[", "]"}], "^", "2"}]}], 
        RowBox[{
         RowBox[{"Sin", "[", 
          RowBox[{"\[Theta]", "[", "]"}], "]"}], "^", "2"}]}]}], "}"}], "]"}],
     ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "sch"}], ",", 
      RowBox[{"-", "sch"}]}], "}"}]}], "]"}]}]], "Input",
 CellLabel->"In[183]:=",
 CellID->931449499,ExpressionUUID->"737aeb28-7bfc-4716-bff5-b6d0ae682e3d"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"1", "-", 
        FractionBox[
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["Mass",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Mass,
           Editable->False]}], 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]]}], ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", 
       RowBox[{"-", 
        FractionBox["1", 
         RowBox[{"1", "-", 
          FractionBox[
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["Mass",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Mass,
             Editable->False]}], 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]]}]]}], ",", "0", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", 
       RowBox[{"-", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]}], ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", "0", ",", 
       RowBox[{
        RowBox[{"-", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"]}], " ", 
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Sin",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Sin,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"]}]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "sch"}], ",", 
     RowBox[{"-", "sch"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.728677246540413*^9},
 CellLabel->"Out[183]=",
 CellID->1325426988,ExpressionUUID->"c3d00fff-b6d9-4b67-8484-f479cf368016"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"$CVSimplify", "=", "Simplify"}], ";"}]], "Input",
 CellLabel->"In[184]:=",
 CellID->1858690769,ExpressionUUID->"809eef50-09e4-4c01-b251-ae18d110db32"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SetCMetric", "[", 
   RowBox[{"met", ",", "sch", ",", 
    RowBox[{"SignatureOfMetric", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{"3", ",", "1", ",", "0"}], "}"}]}]}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[185]:=",
 CellID->1718625321,ExpressionUUID->"81dcab0a-8033-4b9e-b7d2-f7b42923daa5"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.036636`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.728677246759623*^9},
 CellLabel->"Out[185]=",
 CellID->603516985,ExpressionUUID->"2f0cc529-1dfe-47aa-9573-666f878a1861"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"MetricCompute", "[", 
   RowBox[{"met", ",", "sch", ",", 
    RowBox[{"\"\<Weyl\>\"", "[", 
     RowBox[{
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}]}], "]"}]}], "]"}], "//", "AbsoluteTiming"}]], "Input",\

 CellLabel->"In[186]:=",
 CellID->797100998,ExpressionUUID->"f8448ed8-ddae-4ab2-9a96-c0ec6988af6b"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.065601`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.728677246938315*^9},
 CellLabel->"Out[186]=",
 CellID->744702114,ExpressionUUID->"86880898-bd71-4b40-a740-675432815dff"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"cd", "=", 
  RowBox[{"CovDOfMetric", "[", "met", "]"}]}]], "Input",
 CellLabel->"In[187]:=",
 CellID->437292908,ExpressionUUID->"65d2847a-d5bd-4335-97ac-89fdb9888f05"],

Cell[BoxData[
 RowBox[{"CCovD", "[", 
  RowBox[{"PDsch", ",", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", 
           RowBox[{"-", 
            FractionBox[
             InterpretationBox[
              StyleBox["Mass",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Mass,
              Editable->False], 
             RowBox[{
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["Mass",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Mass,
                Editable->False], " ", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}], "-", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], "2"]}]]}], ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"-", 
            FractionBox[
             InterpretationBox[
              StyleBox["Mass",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Mass,
              Editable->False], 
             RowBox[{
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["Mass",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Mass,
                Editable->False], " ", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}], "-", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], "2"]}]]}], ",", "0", ",", "0", ",", "0"}], 
          "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           FractionBox[
            RowBox[{
             InterpretationBox[
              StyleBox["Mass",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Mass,
              Editable->False], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"-", "2"}], " ", 
                InterpretationBox[
                 StyleBox["Mass",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False]}], "+", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}], ")"}]}], 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "3"]], ",", "0", ",", "0", ",", "0"}], "}"}], 
         ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", 
           FractionBox[
            InterpretationBox[
             StyleBox["Mass",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Mass,
             Editable->False], 
            RowBox[{
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["Mass",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False], " ", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}], "-", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}]], ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           RowBox[{
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["Mass",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Mass,
              Editable->False]}], "-", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]}], ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           RowBox[{
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["Mass",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Mass,
                Editable->False]}], "-", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}], ")"}], " ", 
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["Sin",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Sin,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], "2"]}]}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           FractionBox["1", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]], ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", 
           FractionBox["1", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]], ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           RowBox[{
            RowBox[{"-", 
             RowBox[{
              InterpretationBox[
               StyleBox["Cos",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cos,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}]}], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["Sin",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Sin,
              Editable->False], "[", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False], "]"}]}]}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           FractionBox["1", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           RowBox[{
            InterpretationBox[
             StyleBox["Cot",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cot,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", 
           FractionBox["1", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]], ",", 
           RowBox[{
            InterpretationBox[
             StyleBox["Cot",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cot,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], ",", "0"}], "}"}]}], "}"}]}], "}"}], 
     ",", 
     RowBox[{"{", 
      RowBox[{"sch", ",", 
       RowBox[{"-", "sch"}], ",", 
       RowBox[{"-", "sch"}]}], "}"}], ",", "0"}], "]"}], ",", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"1", "-", 
          FractionBox[
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["Mass",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Mass,
             Editable->False]}], 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]]}], ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", 
         RowBox[{"-", 
          FractionBox["1", 
           RowBox[{"1", "-", 
            FractionBox[
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["Mass",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False]}], 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]]}]]}], ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", 
         RowBox[{"-", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]}], ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", 
         RowBox[{
          RowBox[{"-", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Sin",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Sin,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "2"]}]}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "sch"}], ",", 
       RowBox[{"-", "sch"}]}], "}"}], ",", "0"}], "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728677247090959*^9},
 CellLabel->"Out[187]=",
 CellID->673735282,ExpressionUUID->"bbb4888b-eedb-471b-89aa-4a7cfd611da1"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"riemann", "=", 
  RowBox[{"Riemann", "[", "cd", "]"}]}]], "Input",
 CellLabel->"In[188]:=",
 CellID->1909141065,ExpressionUUID->"d4b1f57e-b6fd-40ad-b9fe-d66c35781458"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", 
           FractionBox[
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["Mass",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Mass,
              Editable->False], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["Mass",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False]}], "-", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}], ")"}]}], 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "4"]], ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           FractionBox[
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["Mass",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Mass,
              Editable->False]}], 
            RowBox[{
             RowBox[{"(", 
              RowBox[{
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["Mass",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False]}], "-", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}], ")"}], " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}]], ",", "0", ",", "0", ",", "0"}], 
          "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           FractionBox[
            RowBox[{
             InterpretationBox[
              StyleBox["Mass",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Mass,
              Editable->False], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"-", "2"}], " ", 
                InterpretationBox[
                 StyleBox["Mass",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False]}], "+", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}], ")"}]}], 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "4"]], ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           FractionBox[
            InterpretationBox[
             StyleBox["Mass",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Mass,
             Editable->False], 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]], ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           FractionBox[
            RowBox[{
             InterpretationBox[
              StyleBox["Mass",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Mass,
              Editable->False], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"-", "2"}], " ", 
                InterpretationBox[
                 StyleBox["Mass",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False]}], "+", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}], ")"}]}], 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "4"]]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           FractionBox[
            RowBox[{
             InterpretationBox[
              StyleBox["Mass",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Mass,
              Editable->False], " ", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], "2"]}], 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]], ",", "0", ",", "0", ",", "0"}], "}"}]}], 
        "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", 
           RowBox[{"-", 
            FractionBox[
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["Mass",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["Mass",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False]}], "-", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}], ")"}]}], 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "4"]]}], ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"-", 
            FractionBox[
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["Mass",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False]}], 
             RowBox[{
              RowBox[{"(", 
               RowBox[{
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["Mass",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False]}], "-", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}], ")"}], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], "2"]}]]}], ",", "0", ",", "0", ",", "0"}], 
          "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           FractionBox[
            InterpretationBox[
             StyleBox["Mass",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Mass,
             Editable->False], 
            RowBox[{
             RowBox[{"(", 
              RowBox[{
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["Mass",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False]}], "-", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}], ")"}], " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}]], ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", 
           FractionBox[
            InterpretationBox[
             StyleBox["Mass",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Mass,
             Editable->False], 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]], ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           FractionBox[
            InterpretationBox[
             StyleBox["Mass",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Mass,
             Editable->False], 
            RowBox[{
             RowBox[{"(", 
              RowBox[{
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["Mass",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False]}], "-", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}], ")"}], " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}]]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", 
           FractionBox[
            RowBox[{
             InterpretationBox[
              StyleBox["Mass",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Mass,
              Editable->False], " ", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], "2"]}], 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]], ",", "0", ",", "0"}], "}"}]}], "}"}]}], "}"}],
      ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           RowBox[{"-", 
            FractionBox[
             RowBox[{
              InterpretationBox[
               StyleBox["Mass",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "2"}], " ", 
                 InterpretationBox[
                  StyleBox["Mass",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False]}], "+", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}], ")"}]}], 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "4"]]}], ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"-", 
            FractionBox[
             InterpretationBox[
              StyleBox["Mass",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Mass,
              Editable->False], 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]]}], ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           RowBox[{"-", 
            FractionBox[
             InterpretationBox[
              StyleBox["Mass",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Mass,
              Editable->False], 
             RowBox[{
              RowBox[{"(", 
               RowBox[{
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["Mass",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False]}], "-", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}], ")"}], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], "2"]}]]}], ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", 
           RowBox[{"-", 
            FractionBox[
             InterpretationBox[
              StyleBox["Mass",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Mass,
              Editable->False], 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]]}], ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           FractionBox[
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["Mass",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Mass,
              Editable->False]}], 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           RowBox[{"-", 
            FractionBox[
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["Mass",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["Sin",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 Sin,
                 Editable->False], "[", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False], "]"}], "2"]}], 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]]}], ",", "0"}], "}"}]}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           RowBox[{"-", 
            FractionBox[
             RowBox[{
              InterpretationBox[
               StyleBox["Mass",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "2"}], " ", 
                 InterpretationBox[
                  StyleBox["Mass",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False]}], "+", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}], ")"}]}], 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "4"]]}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"-", 
            FractionBox[
             RowBox[{
              InterpretationBox[
               StyleBox["Mass",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["Sin",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 Sin,
                 Editable->False], "[", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False], "]"}], "2"]}], 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]]}], ",", "0", ",", "0", ",", "0"}], "}"}]}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           RowBox[{"-", 
            FractionBox[
             InterpretationBox[
              StyleBox["Mass",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Mass,
              Editable->False], 
             RowBox[{
              RowBox[{"(", 
               RowBox[{
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["Mass",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False]}], "-", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}], ")"}], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], "2"]}]]}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", 
           RowBox[{"-", 
            FractionBox[
             RowBox[{
              InterpretationBox[
               StyleBox["Mass",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["Sin",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 Sin,
                 Editable->False], "[", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False], "]"}], "2"]}], 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]]}], ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
       
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           RowBox[{"-", 
            FractionBox[
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["Mass",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False]}], 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]]}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           FractionBox[
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["Mass",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Mass,
              Editable->False], " ", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], "2"]}], 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]], ",", "0"}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}]}], 
      "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "sch"}], ",", 
     RowBox[{"-", "sch"}], ",", 
     RowBox[{"-", "sch"}], ",", "sch"}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.7286772472463408`*^9},
 CellLabel->"Out[188]=",
 CellID->1935097590,ExpressionUUID->"915153af-806e-4f61-a506-9d3d0c3fdd11"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"riemann", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", 
     RowBox[{"-", "sch"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", 
     RowBox[{"-", "sch"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", 
     RowBox[{"-", "sch"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "sch"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[189]:=",
 CellID->2025911961,ExpressionUUID->"d33b11bc-e329-4405-a673-12a541a7b56d"],

Cell[BoxData[
 RowBox[{"-", 
  FractionBox[
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["Mass",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Mass,
     Editable->False]}], 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["Mass",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Mass,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]]}]], "Output",
 CellChangeTimes->{3.728677247394298*^9},
 CellLabel->"Out[189]=",
 CellID->1692844962,ExpressionUUID->"7d6005ec-eaeb-40a8-8716-a32ca48540e5"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"MetricCompute", "[", 
   RowBox[{"met", ",", "sch", ",", 
    RowBox[{"\"\<CDRiemann\>\"", "[", 
     RowBox[{
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}]}], "]"}]}], "]"}], "//", "AbsoluteTiming"}]], "Input",\

 CellLabel->"In[190]:=",
 CellID->693045523,ExpressionUUID->"81cc7b82-d62a-436b-b80d-add2d1d1bd26"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.056146`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.728677247630097*^9},
 CellLabel->"Out[190]=",
 CellID->1781920878,ExpressionUUID->"94f1fdb2-8527-44bf-89f6-8f416538bdac"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"cdriemann", "=", 
   RowBox[{"TensorDerivative", "[", 
    RowBox[{"riemann", ",", "cd"}], "]"}]}], ";"}]], "Input",
 CellLabel->"In[191]:=",
 CellID->382131502,ExpressionUUID->"7ecb2479-c2bb-40ae-b15d-67fd600a5942"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"cdriemann", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "d"}], ",", 
      RowBox[{"-", "e"}]}], "]"}], "+", 
    RowBox[{"cdriemann", "[", 
     RowBox[{
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "d"}], ",", 
      RowBox[{"-", "e"}]}], "]"}], "+", 
    RowBox[{"cdriemann", "[", 
     RowBox[{
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "d"}], ",", 
      RowBox[{"-", "e"}]}], "]"}]}], "//", "Simplify"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[192]:=",
 CellID->1508049584,ExpressionUUID->"837c87eb-ffc4-4992-82b7-978c8e3773f1"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.061111`", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.7286772479534903`*^9},
 CellLabel->"Out[192]=",
 CellID->2084739131,ExpressionUUID->"b6e08297-3287-4ca5-b4fc-e3b0d19dbf7b"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{"met", ",", "cd", ",", "riemann", ",", "cdriemann"}], 
  "]"}]], "Input",
 CellLabel->"In[193]:=",
 CellID->1480866034,ExpressionUUID->"6038687a-b7f9-44c8-9c75-9e7403263ccd"],

Cell[BoxData[
 RowBox[{"UndefConstantSymbol", "[", "Mass", "]"}]], "Input",
 CellLabel->"In[194]:=",
 CellID->1727448355,ExpressionUUID->"cd25d69e-3f9e-4d5f-95bf-dac0b3538edd"],

Cell[BoxData[
 RowBox[{"UndefChart", "[", "sch", "]"}]], "Input",
 CellLabel->"In[195]:=",
 CellID->1393078692,ExpressionUUID->"e35cb880-17e6-483e-ab84-9c9f2b058a17"],

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M", "]"}]], "Input",
 CellLabel->"In[196]:=",
 CellID->1606856850,ExpressionUUID->"472161a3-3fbf-4790-ad46-bb4c29562987"],

Cell[BoxData[
 RowBox[{"ClearxCobaCache", "[", "All", "]"}]], "Input",
 CellLabel->"In[197]:=",
 CellID->1836216925,ExpressionUUID->"9eab08b6-0b49-4023-808c-68de83a87caf"]
}, Closed]],

Cell[CellGroupData[{

Cell["3.6.3. Schwarzschild with cosmological constant", "Subsubsection",
 CellChangeTimes->{{3.728677118025834*^9, 3.728677118274026*^9}, 
   3.7908767668967333`*^9},
 CellID->1957851085,ExpressionUUID->"e6e04444-d4ff-4ebe-8c93-6e4498667d65"],

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M", ",", "4", ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f"}], "}"}]}], 
  "]"}]], "Input",
 CellLabel->"In[198]:=",
 CellID->1858746261,ExpressionUUID->"1bd5e852-305f-464a-886f-df1cbe24b433"],

Cell[BoxData[
 RowBox[{"DefChart", "[", 
  RowBox[{"sch", ",", "M", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2", ",", "3"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"t", "[", "]"}], ",", 
     RowBox[{"r", "[", "]"}], ",", 
     RowBox[{"\[Theta]", "[", "]"}], ",", 
     RowBox[{"\[Phi]", "[", "]"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[199]:=",
 CellID->503795718,ExpressionUUID->"ff77e9df-96fe-4c97-a83f-ed2f0125c5d5"],

Cell[BoxData[
 RowBox[{"DefConstantSymbol", "[", "\[CapitalMu]", "]"}]], "Input",
 CellLabel->"In[200]:=",
 CellID->1914095952,ExpressionUUID->"366f814f-10d5-480f-84e9-0ff572682009"],

Cell[BoxData[
 RowBox[{"DefConstantSymbol", "[", "\[CapitalLambda]", "]"}]], "Input",
 CellLabel->"In[201]:=",
 CellID->905633204,ExpressionUUID->"59e68030-21ea-42dc-a07f-e1105daed564"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"met", "=", 
  RowBox[{"CTensor", "[", 
   RowBox[{
    RowBox[{"DiagonalMatrix", "[", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"1", "-", 
        RowBox[{"2", 
         RowBox[{"\[CapitalMu]", "/", 
          RowBox[{"r", "[", "]"}]}]}], "-", 
        RowBox[{"\[CapitalLambda]", " ", 
         RowBox[{
          RowBox[{
           RowBox[{"r", "[", "]"}], "^", "2"}], "/", "3"}]}]}], ",", 
       RowBox[{
        RowBox[{"-", "1"}], "/", 
        RowBox[{"(", 
         RowBox[{"1", "-", 
          RowBox[{"2", 
           RowBox[{"\[CapitalMu]", "/", 
            RowBox[{"r", "[", "]"}]}]}], "-", 
          RowBox[{"\[CapitalLambda]", " ", 
           RowBox[{
            RowBox[{
             RowBox[{"r", "[", "]"}], "^", "2"}], "/", "3"}]}]}], ")"}]}], 
       ",", 
       RowBox[{"-", 
        RowBox[{
         RowBox[{"r", "[", "]"}], "^", "2"}]}], ",", 
       RowBox[{
        RowBox[{"-", 
         RowBox[{
          RowBox[{"r", "[", "]"}], "^", "2"}]}], 
        RowBox[{
         RowBox[{"Sin", "[", 
          RowBox[{"\[Theta]", "[", "]"}], "]"}], "^", "2"}]}]}], "}"}], "]"}],
     ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "sch"}], ",", 
      RowBox[{"-", "sch"}]}], "}"}]}], "]"}]}]], "Input",
 CellLabel->"In[202]:=",
 CellID->1760653339,ExpressionUUID->"b06165c4-4722-4f48-8041-806a14043f86"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"1", "-", 
        FractionBox[
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["\[CapitalMu]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[CapitalMu],
           Editable->False]}], 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]], "-", 
        FractionBox[
         RowBox[{
          InterpretationBox[
           StyleBox["\[CapitalLambda]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[CapitalLambda],
           Editable->False], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]}], "3"]}], ",", "0", ",", "0", ",", "0"}], 
      "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", 
       RowBox[{"-", 
        FractionBox["1", 
         RowBox[{"1", "-", 
          FractionBox[
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["\[CapitalMu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[CapitalMu],
             Editable->False]}], 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]], "-", 
          FractionBox[
           RowBox[{
            InterpretationBox[
             StyleBox["\[CapitalLambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[CapitalLambda],
             Editable->False], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"]}], "3"]}]]}], ",", "0", ",", "0"}], 
      "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", 
       RowBox[{"-", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]}], ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", "0", ",", 
       RowBox[{
        RowBox[{"-", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"]}], " ", 
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Sin",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Sin,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"]}]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "sch"}], ",", 
     RowBox[{"-", "sch"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.7286772566837673`*^9},
 CellLabel->"Out[202]=",
 CellID->1657532209,ExpressionUUID->"fed72294-041c-4b2f-aa45-5b731a880e47"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"$CVSimplify", "=", "Simplify"}], ";"}]], "Input",
 CellLabel->"In[203]:=",
 CellID->1930459496,ExpressionUUID->"7eba7489-eeec-4d64-8250-6e5fbda55a4e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SetCMetric", "[", 
   RowBox[{"met", ",", "sch", ",", 
    RowBox[{"SignatureOfMetric", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{"3", ",", "1", ",", "0"}], "}"}]}]}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[204]:=",
 CellID->1049810396,ExpressionUUID->"a9191b72-e365-48c3-a584-50b47fcf90b4"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.059137`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.728677257282571*^9},
 CellLabel->"Out[204]=",
 CellID->985967779,ExpressionUUID->"2945a4d0-2cff-4e5f-96ca-06d08730276f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"MetricCompute", "[", 
   RowBox[{"met", ",", "sch", ",", 
    RowBox[{"\"\<CDRiemann\>\"", "[", 
     RowBox[{
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}]}], "]"}]}], "]"}], "//", "AbsoluteTiming"}]], "Input",\

 CellLabel->"In[205]:=",
 CellID->986857765,ExpressionUUID->"6a809c07-fb09-4a62-9b44-6d24ae92bbd2"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.125904`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.728677257785822*^9},
 CellLabel->"Out[205]=",
 CellID->1176641401,ExpressionUUID->"16125f6b-f53c-4139-b341-211aee777a86"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"cd", "=", 
   RowBox[{"CovDOfMetric", "[", "met", "]"}]}], ";"}]], "Input",
 CellLabel->"In[206]:=",
 CellID->1331458196,ExpressionUUID->"5c7ff7db-ec68-456e-ab0d-0de915a0dba5"],

Cell[BoxData[
 RowBox[{
  RowBox[{"riemann", "=", 
   RowBox[{"Riemann", "[", "cd", "]"}]}], ";"}]], "Input",
 CellLabel->"In[207]:=",
 CellID->1038298432,ExpressionUUID->"572d3290-8132-4518-905f-b23d49442b23"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"riemann", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", 
     RowBox[{"-", "sch"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", 
     RowBox[{"-", "sch"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", 
     RowBox[{"-", "sch"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "sch"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[208]:=",
 CellID->229631102,ExpressionUUID->"dfcea39b-8367-4fc8-a0fa-18f3cbb31283"],

Cell[BoxData[
 RowBox[{"-", 
  FractionBox[
   RowBox[{
    RowBox[{"6", " ", 
     InterpretationBox[
      StyleBox["\[CapitalMu]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalMu],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[CapitalLambda]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalLambda],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "3"]}]}], 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"6", " ", 
       InterpretationBox[
        StyleBox["\[CapitalMu]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalMu],
        Editable->False]}], "-", 
      RowBox[{"3", " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[CapitalLambda]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalLambda],
        Editable->False], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "3"]}]}], ")"}]}]]}]], "Output",
 CellChangeTimes->{3.728677258180497*^9},
 CellLabel->"Out[208]=",
 CellID->1782605031,ExpressionUUID->"1e591b60-c0b1-420a-99a5-1b4119b93feb"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Ricci", "[", "cd", "]"}], "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"2", ",", 
     RowBox[{"-", "sch"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"2", ",", 
     RowBox[{"-", "sch"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[209]:=",
 CellID->528190349,ExpressionUUID->"d76e814d-7d08-405d-8c29-d4635f465c23"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox["\[CapitalLambda]",
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`\[CapitalLambda],
   Editable->False], " ", 
  SuperscriptBox[
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False], "2"]}]], "Output",
 CellChangeTimes->{3.728677258345619*^9},
 CellLabel->"Out[209]=",
 CellID->698396720,ExpressionUUID->"fc9f3533-059d-4469-9647-32ab675aa8ce"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"RicciScalar", "[", "cd", "]"}], "[", "]"}]], "Input",
 CellLabel->"In[210]:=",
 CellID->485389371,ExpressionUUID->"2e839399-72d9-4456-9211-e887a6b6e80c"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", "4"}], " ", 
  InterpretationBox[
   StyleBox["\[CapitalLambda]",
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`\[CapitalLambda],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7286772585830107`*^9},
 CellLabel->"Out[210]=",
 CellID->1955542185,ExpressionUUID->"801d0581-1274-4e61-9dd8-9506caa20153"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"cdriemann", "=", 
   RowBox[{"TensorDerivative", "[", 
    RowBox[{"riemann", ",", "cd"}], "]"}]}], ";"}]], "Input",
 CellLabel->"In[211]:=",
 CellID->535888911,ExpressionUUID->"89a4dc42-314b-4245-bee4-0627525aa84b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"cdriemann", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "d"}], ",", 
      RowBox[{"-", "e"}]}], "]"}], "+", 
    RowBox[{"cdriemann", "[", 
     RowBox[{
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "d"}], ",", 
      RowBox[{"-", "e"}]}], "]"}], "+", 
    RowBox[{"cdriemann", "[", 
     RowBox[{
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "d"}], ",", 
      RowBox[{"-", "e"}]}], "]"}]}], "//", "Simplify"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[212]:=",
 CellID->138511659,ExpressionUUID->"7402756f-45bf-4424-b0a8-96367e5d10ef"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.082076`", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.7286772590094624`*^9},
 CellLabel->"Out[212]=",
 CellID->1707769607,ExpressionUUID->"37a24cc5-98a2-475f-bf6e-8d75cd512c93"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"weyl", "=", 
   RowBox[{"Weyl", "[", "cd", "]"}]}], ";"}]], "Input",
 CellLabel->"In[213]:=",
 CellID->551689592,ExpressionUUID->"e06673c1-f8c9-4c47-88e5-4ddd8c8a727f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"weyl", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", "c", ",", "d"}], "]"}], 
   RowBox[{"weyl", "[", 
    RowBox[{
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}], ",", "a", ",", "b"}], "]"}]}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[214]:=",
 CellID->1560068645,ExpressionUUID->"a45abe1b-f3fe-443b-832c-8772892bce28"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.038757`", ",", 
   FractionBox[
    RowBox[{"48", " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[CapitalMu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalMu],
       Editable->False], "2"]}], 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "6"]]}], "}"}]], "Output",
 CellChangeTimes->{3.728677259282609*^9},
 CellLabel->"Out[214]=",
 CellID->1364842694,ExpressionUUID->"ae21f2eb-8b7f-4dd7-985f-23f63abc651b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"weyl", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", "c", ",", "d"}], "]"}], 
   RowBox[{"weyl", "[", 
    RowBox[{
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}], ",", "e", ",", "f"}], "]"}], 
   RowBox[{"weyl", "[", 
    RowBox[{
     RowBox[{"-", "e"}], ",", 
     RowBox[{"-", "f"}], ",", "a", ",", "b"}], "]"}]}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[215]:=",
 CellID->1964531839,ExpressionUUID->"945870dc-f9ec-496e-b0c4-1dbf625e9009"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.067139`", ",", 
   RowBox[{"-", 
    FractionBox[
     RowBox[{"96", " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[CapitalMu]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalMu],
        Editable->False], "3"]}], 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "9"]]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728677259494957*^9},
 CellLabel->"Out[215]=",
 CellID->1638324451,ExpressionUUID->"395328f1-7e55-41c3-a877-3f809aeeb167"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{"met", ",", "cd", ",", "riemann", ",", "cdriemann", ",", "weyl"}], 
  "]"}]], "Input",
 CellLabel->"In[216]:=",
 CellID->1801495627,ExpressionUUID->"694cb379-4a73-4fc0-ad17-358b4f8c2447"],

Cell[BoxData[
 RowBox[{"UndefConstantSymbol", "[", "\[CapitalMu]", "]"}]], "Input",
 CellLabel->"In[217]:=",
 CellID->365063936,ExpressionUUID->"11f971b8-a089-485f-b480-5d369806dc2e"],

Cell[BoxData[
 RowBox[{"UndefConstantSymbol", "[", "\[CapitalLambda]", "]"}]], "Input",
 CellLabel->"In[218]:=",
 CellID->268792873,ExpressionUUID->"c0eb45b7-871d-41e9-9500-2bf1b3aeefa0"],

Cell[BoxData[
 RowBox[{"UndefChart", "[", "sch", "]"}]], "Input",
 CellLabel->"In[219]:=",
 CellID->1443706870,ExpressionUUID->"6d5af50a-0e74-42c4-b9f0-2cef4422ef7b"],

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M", "]"}]], "Input",
 CellLabel->"In[220]:=",
 CellID->2111120284,ExpressionUUID->"8e1ad2ac-a251-4e96-b671-8a778080d654"],

Cell[BoxData[
 RowBox[{"ClearxCobaCache", "[", "All", "]"}]], "Input",
 CellLabel->"In[221]:=",
 CellID->873797454,ExpressionUUID->"3e4df76b-a434-4926-88af-04e15d5c9fba"]
}, Closed]],

Cell[CellGroupData[{

Cell["3.6.4. Robertson-Walker", "Subsubsection",
 CellChangeTimes->{{3.728677119849506*^9, 3.728677120073847*^9}, 
   3.7908767685911827`*^9},
 CellID->1420631192,ExpressionUUID->"a368fb88-c5d8-4bfd-96f5-c6f267bdb37d"],

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M", ",", "4", ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f"}], "}"}]}], 
  "]"}]], "Input",
 CellLabel->"In[222]:=",
 CellID->825256598,ExpressionUUID->"2e73e135-9684-40c8-8478-53a497bc1833"],

Cell[BoxData[
 RowBox[{"DefChart", "[", 
  RowBox[{"com", ",", "M", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2", ",", "3"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"t", "[", "]"}], ",", 
     RowBox[{"r", "[", "]"}], ",", 
     RowBox[{"\[Theta]", "[", "]"}], ",", 
     RowBox[{"\[Phi]", "[", "]"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[223]:=",
 CellID->1638250260,ExpressionUUID->"4310c980-f951-4643-87b6-20f7bfafd4cb"],

Cell[BoxData[
 RowBox[{"DefConstantSymbol", "[", "\[Kappa]", "]"}]], "Input",
 CellLabel->"In[224]:=",
 CellID->1184281633,ExpressionUUID->"8623cc2a-cba8-4362-bcb6-b1346a5cda7d"],

Cell[BoxData[
 RowBox[{"DefScalarFunction", "[", "scale", "]"}]], "Input",
 CellLabel->"In[225]:=",
 CellID->1342965435,ExpressionUUID->"ea9a72fc-e839-4f4d-a5b8-dcec96a6cd00"],

Cell[BoxData[
 RowBox[{
  RowBox[{"met", "=", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"1", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", 
         RowBox[{
          RowBox[{"-", 
           RowBox[{
            RowBox[{"scale", "[", 
             RowBox[{"t", "[", "]"}], "]"}], "^", "2"}]}], "/", 
          RowBox[{"(", 
           RowBox[{"1", "-", 
            RowBox[{"\[Kappa]", " ", 
             RowBox[{
              RowBox[{"r", "[", "]"}], "^", "2"}]}]}], ")"}]}], ",", "0", ",",
          "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", 
         RowBox[{
          RowBox[{"-", 
           RowBox[{
            RowBox[{"scale", "[", 
             RowBox[{"t", "[", "]"}], "]"}], "^", "2"}]}], 
          RowBox[{
           RowBox[{"r", "[", "]"}], "^", "2"}]}], ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", 
         RowBox[{
          RowBox[{"-", 
           RowBox[{
            RowBox[{"r", "[", "]"}], "^", "2"}]}], 
          RowBox[{
           RowBox[{"scale", "[", 
            RowBox[{"t", "[", "]"}], "]"}], "^", "2"}], 
          RowBox[{
           RowBox[{"Sin", "[", 
            RowBox[{"\[Theta]", "[", "]"}], "]"}], "^", "2"}]}]}], "}"}]}], 
      "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "com"}], ",", 
       RowBox[{"-", "com"}]}], "}"}]}], "]"}]}], ";"}]], "Input",
 CellLabel->"In[226]:=",
 CellID->1166936075,ExpressionUUID->"ef83c15f-90c0-4641-b104-27bcf4a224c1"],

Cell[BoxData[
 RowBox[{
  RowBox[{"$CVSimplify", "=", "Simplify"}], ";"}]], "Input",
 CellLabel->"In[227]:=",
 CellID->1873948121,ExpressionUUID->"2bd5debe-8427-4162-b6cc-65a43a5e526c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SetCMetric", "[", 
   RowBox[{"met", ",", "com", ",", 
    RowBox[{"SignatureOfMetric", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{"3", ",", "1", ",", "0"}], "}"}]}]}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[228]:=",
 CellID->450902745,ExpressionUUID->"522c3bbe-1998-49c0-88d7-824d8bf8bca3"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.046001`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.7286772771741467`*^9},
 CellLabel->"Out[228]=",
 CellID->1412734845,ExpressionUUID->"074d7576-e084-49f9-aaa6-6a073e2a3537"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"MetricCompute", "[", 
   RowBox[{"met", ",", "com", ",", 
    RowBox[{"\"\<CDRiemann\>\"", "[", 
     RowBox[{
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}]}], "]"}]}], "]"}], "//", "AbsoluteTiming"}]], "Input",\

 CellLabel->"In[229]:=",
 CellID->1599836902,ExpressionUUID->"c75e921c-77fb-458d-8778-b38d146ca932"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.187247`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.728677277476108*^9},
 CellLabel->"Out[229]=",
 CellID->1255338683,ExpressionUUID->"4ee09b44-cfbb-4559-b317-ae5e528a7b35"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"cd", "=", 
   RowBox[{"CovDOfMetric", "[", "met", "]"}]}], ";"}]], "Input",
 CellLabel->"In[230]:=",
 CellID->1028420069,ExpressionUUID->"c99dda72-2311-45f8-8707-6d405cfc65af"],

Cell[CellGroupData[{

Cell["The space is conformally flat:", "MathCaption",
 CellID->1707708627,ExpressionUUID->"38bc0cd8-dfc4-4e18-9a2e-2f17ce865f79"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Weyl", "[", "cd", "]"}]], "Input",
 CellLabel->"In[231]:=",
 CellID->339951834,ExpressionUUID->"c46497bf-80a6-434a-a880-b3e144867dc0"],

Cell[BoxData["Zero"], "Output",
 CellChangeTimes->{3.728677277821261*^9},
 CellLabel->"Out[231]=",
 CellID->1597897053,ExpressionUUID->"08e8e660-49b8-476b-b866-9eb2d43e5f25"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Ricci", "[", "cd", "]"}], "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0", ",", 
     RowBox[{"-", "com"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", 
     RowBox[{"-", "com"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[232]:=",
 CellID->686277402,ExpressionUUID->"e71f1505-84d5-462b-975a-61cdda3844ee"],

Cell[BoxData[
 RowBox[{"-", 
  FractionBox[
   RowBox[{"3", " ", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["scale",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`scale,
       Editable->False], "\[Prime]\[Prime]",
      MultilineFunction->None], "[", 
     InterpretationBox[
      StyleBox["t",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`t[],
      Editable->False], "]"}]}], 
   RowBox[{
    InterpretationBox[
     StyleBox["scale",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`scale,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["t",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`t[],
     Editable->False], "]"}]]}]], "Output",
 CellChangeTimes->{3.728677277965638*^9},
 CellLabel->"Out[232]=",
 CellID->1572764036,ExpressionUUID->"f6a6ade0-c367-49e9-bb22-86c8628ad3dc"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Einstein", "[", "cd", "]"}], "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0", ",", 
     RowBox[{"-", "com"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", 
     RowBox[{"-", "com"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[233]:=",
 CellID->1098907181,ExpressionUUID->"bc9442dd-f05f-4e17-8cef-b7848b5f8e28"],

Cell[BoxData[
 FractionBox[
  RowBox[{"3", " ", 
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Kappa]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa],
      Editable->False], "+", 
     SuperscriptBox[
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["scale",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`scale,
         Editable->False], "\[Prime]",
        MultilineFunction->None], "[", 
       InterpretationBox[
        StyleBox["t",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`t[],
        Editable->False], "]"}], "2"]}], ")"}]}], 
  SuperscriptBox[
   RowBox[{
    InterpretationBox[
     StyleBox["scale",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`scale,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["t",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`t[],
     Editable->False], "]"}], "2"]]], "Output",
 CellChangeTimes->{3.7286772781134*^9},
 CellLabel->"Out[233]=",
 CellID->68942129,ExpressionUUID->"8b78a84f-89ed-4944-b482-e5eb3560f053"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"cdriemann", "=", 
   RowBox[{"TensorDerivative", "[", 
    RowBox[{
     RowBox[{"RiemannDown", "[", "cd", "]"}], ",", "cd"}], "]"}]}], 
  ";"}]], "Input",
 CellLabel->"In[234]:=",
 CellID->2068847176,ExpressionUUID->"286a6d25-ba84-48a7-90db-8b32e084b411"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"cdriemann", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}], ",", 
     RowBox[{"-", "e"}]}], "]"}], "+", 
   RowBox[{"cdriemann", "[", 
    RowBox[{
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "d"}], ",", 
     RowBox[{"-", "e"}]}], "]"}], "+", 
   RowBox[{"cdriemann", "[", 
    RowBox[{
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "d"}], ",", 
     RowBox[{"-", "e"}]}], "]"}]}], "//", "Simplify"}]], "Input",
 CellLabel->"In[235]:=",
 CellID->1206758455,ExpressionUUID->"572cbd85-3f97-4e63-87bc-332dc676f190"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728677278345656*^9},
 CellLabel->"Out[235]=",
 CellID->1216676370,ExpressionUUID->"2193d515-f3fd-44d1-8c3c-eedc6a65e466"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{"met", ",", "cd", ",", "cdriemann"}], "]"}]], "Input",
 CellLabel->"In[236]:=",
 CellID->2134093256,ExpressionUUID->"ef07c5ea-e29e-4cfe-a0fc-c33858f337f7"],

Cell[BoxData[
 RowBox[{"UndefConstantSymbol", "[", "\[Kappa]", "]"}]], "Input",
 CellLabel->"In[237]:=",
 CellID->433481398,ExpressionUUID->"c96f60aa-3766-4fc0-844d-cc64eee12483"],

Cell[BoxData[
 RowBox[{"UndefScalarFunction", "[", "scale", "]"}]], "Input",
 CellLabel->"In[238]:=",
 CellID->1257325562,ExpressionUUID->"10bd86a4-3baf-42dd-833b-17505b2b4606"],

Cell[BoxData[
 RowBox[{"UndefChart", "[", "com", "]"}]], "Input",
 CellLabel->"In[239]:=",
 CellID->631634054,ExpressionUUID->"82775c2b-76c9-4e88-b2f6-bf20a659cc99"],

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M", "]"}]], "Input",
 CellLabel->"In[240]:=",
 CellID->861878457,ExpressionUUID->"3fac0238-1fab-4bbf-ab4c-11478444c69e"],

Cell[BoxData[
 RowBox[{"ClearxCobaCache", "[", "All", "]"}]], "Input",
 CellLabel->"In[241]:=",
 CellID->814027784,ExpressionUUID->"72a9d524-b732-4591-bd6c-524c63bf0dce"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.6.5. Conformally flat 5-dimensional metric", "Subsubsection",
 CellChangeTimes->{{3.728677121441588*^9, 3.728677121697835*^9}, 
   3.790876770575539*^9},
 CellID->1894134298,ExpressionUUID->"11318086-eb63-457e-bc15-c56c1be55fdf"],

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M", ",", "5", ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f"}], "}"}]}], 
  "]"}]], "Input",
 CellLabel->"In[242]:=",
 CellID->417563358,ExpressionUUID->"92b85df9-87ad-4c0f-8d8c-2bf37398c172"],

Cell[BoxData[
 RowBox[{"DefChart", "[", 
  RowBox[{"xs", ",", "M", ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", "2", ",", "3", ",", "4", ",", "5"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"x1", "[", "]"}], ",", 
     RowBox[{"x2", "[", "]"}], ",", 
     RowBox[{"x3", "[", "]"}], ",", 
     RowBox[{"x4", "[", "]"}], ",", 
     RowBox[{"x5", "[", "]"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[243]:=",
 CellID->10799510,ExpressionUUID->"04452c7b-51f8-4b67-94d7-ac5616115d3f"],

Cell[BoxData[
 RowBox[{"DefScalarFunction", "[", "\[CapitalOmega]", "]"}]], "Input",
 CellLabel->"In[244]:=",
 CellID->1118738222,ExpressionUUID->"7d7a9790-ef67-4587-9f02-d3fd6fa8652a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"met", "=", 
  RowBox[{"CTensor", "[", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"\[CapitalOmega]", "[", 
       RowBox[{
        RowBox[{"x1", "[", "]"}], ",", 
        RowBox[{"x2", "[", "]"}], ",", 
        RowBox[{"x3", "[", "]"}], ",", 
        RowBox[{"x4", "[", "]"}], ",", 
        RowBox[{"x5", "[", "]"}]}], "]"}], "^", "2"}], 
     RowBox[{"IdentityMatrix", "[", "5", "]"}]}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "xs"}], ",", 
      RowBox[{"-", "xs"}]}], "}"}]}], "]"}]}]], "Input",
 CellLabel->"In[245]:=",
 CellID->1413967669,ExpressionUUID->"962b08e3-3adf-40b5-b78f-ffee4be371b6"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       SuperscriptBox[
        RowBox[{
         InterpretationBox[
          StyleBox["\[CapitalOmega]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalOmega],
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x1",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x1[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x2",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x2[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x3",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x3[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x4",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x4[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x5",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x5[],
           Editable->False]}], "]"}], "2"], ",", "0", ",", "0", ",", "0", ",",
        "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", 
       SuperscriptBox[
        RowBox[{
         InterpretationBox[
          StyleBox["\[CapitalOmega]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalOmega],
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x1",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x1[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x2",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x2[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x3",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x3[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x4",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x4[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x5",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x5[],
           Editable->False]}], "]"}], "2"], ",", "0", ",", "0", ",", "0"}], 
      "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", 
       SuperscriptBox[
        RowBox[{
         InterpretationBox[
          StyleBox["\[CapitalOmega]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalOmega],
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x1",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x1[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x2",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x2[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x3",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x3[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x4",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x4[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x5",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x5[],
           Editable->False]}], "]"}], "2"], ",", "0", ",", "0"}], "}"}], ",", 
     
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", "0", ",", 
       SuperscriptBox[
        RowBox[{
         InterpretationBox[
          StyleBox["\[CapitalOmega]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalOmega],
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x1",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x1[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x2",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x2[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x3",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x3[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x4",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x4[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x5",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x5[],
           Editable->False]}], "]"}], "2"], ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", 
       SuperscriptBox[
        RowBox[{
         InterpretationBox[
          StyleBox["\[CapitalOmega]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalOmega],
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["x1",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x1[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x2",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x2[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x3",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x3[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x4",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x4[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x5",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x5[],
           Editable->False]}], "]"}], "2"]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "xs"}], ",", 
     RowBox[{"-", "xs"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.728677296453094*^9},
 CellLabel->"Out[245]=",
 CellID->1204595430,ExpressionUUID->"a71f4608-0b1f-4008-9309-74d3b151347b"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"$CVSimplify", "=", "Simplify"}], ";"}]], "Input",
 CellLabel->"In[246]:=",
 CellID->1473621768,ExpressionUUID->"176c9ed7-3710-4cd7-9736-5a58d847a4e5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SetCMetric", "[", 
   RowBox[{"met", ",", "xs", ",", 
    RowBox[{"SignatureOfMetric", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{"4", ",", "1", ",", "0"}], "}"}]}]}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[247]:=",
 CellID->232153925,ExpressionUUID->"29ac6433-98fc-436a-b48a-7e364040ed4c"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.02573`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.728677296701766*^9},
 CellLabel->"Out[247]=",
 CellID->1307051741,ExpressionUUID->"a50a4089-bba7-49ba-bbf5-22f465b33755"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"MetricCompute", "[", 
   RowBox[{"met", ",", "xs", ",", 
    RowBox[{"\"\<Weyl\>\"", "[", 
     RowBox[{
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}]}], "]"}]}], "]"}], "//", "AbsoluteTiming"}]], "Input",\

 CellLabel->"In[248]:=",
 CellID->1724975914,ExpressionUUID->"1cc8e48c-d148-4847-9e46-35259af87a96"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.429549`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.7286772972573767`*^9},
 CellLabel->"Out[248]=",
 CellID->1008385824,ExpressionUUID->"0897a419-34c2-4e62-bde6-90029ee8435e"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"cd", "=", 
   RowBox[{"CovDOfMetric", "[", "met", "]"}]}], ";"}]], "Input",
 CellLabel->"In[249]:=",
 CellID->703937734,ExpressionUUID->"89b92ce6-d787-40fa-ae26-071a2f1767ca"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Weyl", "[", "cd", "]"}]], "Input",
 CellLabel->"In[250]:=",
 CellID->594084015,ExpressionUUID->"a37489ae-26eb-4507-831d-590ef8a266df"],

Cell[BoxData["Zero"], "Output",
 CellChangeTimes->{3.728677297501855*^9},
 CellLabel->"Out[250]=",
 CellID->2147085491,ExpressionUUID->"ee730209-a73b-4df5-a4e6-f11286449fba"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Nonzero components of RiemannDown:", "MathCaption",
 CellID->653504176,ExpressionUUID->"fcf2f4e9-3fd6-4091-878e-54791cd547e1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Length", "[", 
  RowBox[{"nonzero", "=", 
   RowBox[{"DeleteCases", "[", 
    RowBox[{
     RowBox[{"Flatten", "@", 
      RowBox[{"First", "@", 
       RowBox[{"RiemannDown", "[", "cd", "]"}]}]}], ",", "0"}], "]"}]}], 
  "]"}]], "Input",
 CellLabel->"In[251]:=",
 CellID->1213315274,ExpressionUUID->"8053bf52-1937-46d7-ab77-67b1137553d3"],

Cell[BoxData["280"], "Output",
 CellChangeTimes->{3.728677297692217*^9},
 CellLabel->"Out[251]=",
 CellID->1898126696,ExpressionUUID->"b2a2bf7c-afd8-4527-8cf5-3ac715e3ed12"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Length", "[", 
  RowBox[{"differentnonzero", "=", 
   RowBox[{"DeleteDuplicates", "[", "nonzero", "]"}]}], "]"}]], "Input",
 CellLabel->"In[252]:=",
 CellID->992738097,ExpressionUUID->"0eb4eb0b-5209-46fa-a7b0-10cff60ae974"],

Cell[BoxData["40"], "Output",
 CellChangeTimes->{3.72867729811996*^9},
 CellLabel->"Out[252]=",
 CellID->945803487,ExpressionUUID->"9c73a7c5-6e7e-4c5c-9c0a-80ef36ac766f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Apart from numerical factors:", "MathCaption",
 CellID->1041177422,ExpressionUUID->"a6e397c3-e107-4fb6-91d0-dcdb4ad0ab5e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Union", "[", 
   RowBox[{"differentnonzero", ",", 
    RowBox[{"SameTest", "\[Rule]", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"NumericQ", "[", 
        RowBox[{"Simplify", "[", 
         RowBox[{"#1", "/", "#2"}], "]"}], "]"}], "&"}], ")"}]}]}], "]"}], "//",
   "Length"}]], "Input",
 CellLabel->"In[253]:=",
 CellID->1372175798,ExpressionUUID->"ff1aee4e-64d0-47d8-9f14-fe7bbaf8a726"],

Cell[BoxData["20"], "Output",
 CellChangeTimes->{3.728677298963911*^9},
 CellLabel->"Out[253]=",
 CellID->445250915,ExpressionUUID->"a88b47ca-001b-4ee3-ae76-a04d8f4ea0c7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"RicciScalar", "[", "cd", "]"}], "[", "]"}]], "Input",
 CellLabel->"In[254]:=",
 CellID->885845759,ExpressionUUID->"30207a68-40f6-4e11-9430-9810503c6778"],

Cell[BoxData[
 RowBox[{"-", 
  RowBox[{
   FractionBox["1", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["\[CapitalOmega]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalOmega],
       Editable->False], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["x1",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x1[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["x2",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x2[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["x3",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x3[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["x4",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x4[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["x5",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x5[],
        Editable->False]}], "]"}], "4"]], 
   RowBox[{"4", " ", 
    RowBox[{"(", 
     RowBox[{
      SuperscriptBox[
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["\[CapitalOmega]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalOmega],
          Editable->False], 
         TagBox[
          RowBox[{"(", 
           RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", "1"}], ")"}],
          Derivative],
         MultilineFunction->None], "[", 
        RowBox[{
         InterpretationBox[
          StyleBox["x1",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x1[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["x2",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x2[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["x3",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x3[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["x4",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x4[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["x5",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x5[],
          Editable->False]}], "]"}], "2"], "+", 
      SuperscriptBox[
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["\[CapitalOmega]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalOmega],
          Editable->False], 
         TagBox[
          RowBox[{"(", 
           RowBox[{"0", ",", "0", ",", "0", ",", "1", ",", "0"}], ")"}],
          Derivative],
         MultilineFunction->None], "[", 
        RowBox[{
         InterpretationBox[
          StyleBox["x1",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x1[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["x2",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x2[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["x3",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x3[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["x4",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x4[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["x5",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x5[],
          Editable->False]}], "]"}], "2"], "+", 
      SuperscriptBox[
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["\[CapitalOmega]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalOmega],
          Editable->False], 
         TagBox[
          RowBox[{"(", 
           RowBox[{"0", ",", "0", ",", "1", ",", "0", ",", "0"}], ")"}],
          Derivative],
         MultilineFunction->None], "[", 
        RowBox[{
         InterpretationBox[
          StyleBox["x1",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x1[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["x2",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x2[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["x3",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x3[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["x4",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x4[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["x5",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x5[],
          Editable->False]}], "]"}], "2"], "+", 
      SuperscriptBox[
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["\[CapitalOmega]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalOmega],
          Editable->False], 
         TagBox[
          RowBox[{"(", 
           RowBox[{"0", ",", "1", ",", "0", ",", "0", ",", "0"}], ")"}],
          Derivative],
         MultilineFunction->None], "[", 
        RowBox[{
         InterpretationBox[
          StyleBox["x1",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x1[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["x2",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x2[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["x3",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x3[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["x4",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x4[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["x5",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x5[],
          Editable->False]}], "]"}], "2"], "+", 
      SuperscriptBox[
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["\[CapitalOmega]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalOmega],
          Editable->False], 
         TagBox[
          RowBox[{"(", 
           RowBox[{"1", ",", "0", ",", "0", ",", "0", ",", "0"}], ")"}],
          Derivative],
         MultilineFunction->None], "[", 
        RowBox[{
         InterpretationBox[
          StyleBox["x1",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x1[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["x2",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x2[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["x3",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x3[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["x4",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x4[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["x5",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x5[],
          Editable->False]}], "]"}], "2"], "+", 
      RowBox[{"2", " ", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[CapitalOmega]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[CapitalOmega],
         Editable->False], "[", 
        RowBox[{
         InterpretationBox[
          StyleBox["x1",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x1[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["x2",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x2[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["x3",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x3[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["x4",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x4[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["x5",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x5[],
          Editable->False]}], "]"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["\[CapitalOmega]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[CapitalOmega],
            Editable->False], 
           TagBox[
            RowBox[{"(", 
             RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", "2"}], ")"}],
            Derivative],
           MultilineFunction->None], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["x1",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x1[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["x2",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x2[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["x3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x3[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["x4",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x4[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["x5",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x5[],
            Editable->False]}], "]"}], "+", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["\[CapitalOmega]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[CapitalOmega],
            Editable->False], 
           TagBox[
            RowBox[{"(", 
             RowBox[{"0", ",", "0", ",", "0", ",", "2", ",", "0"}], ")"}],
            Derivative],
           MultilineFunction->None], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["x1",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x1[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["x2",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x2[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["x3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x3[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["x4",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x4[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["x5",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x5[],
            Editable->False]}], "]"}], "+", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["\[CapitalOmega]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[CapitalOmega],
            Editable->False], 
           TagBox[
            RowBox[{"(", 
             RowBox[{"0", ",", "0", ",", "2", ",", "0", ",", "0"}], ")"}],
            Derivative],
           MultilineFunction->None], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["x1",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x1[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["x2",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x2[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["x3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x3[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["x4",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x4[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["x5",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x5[],
            Editable->False]}], "]"}], "+", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["\[CapitalOmega]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[CapitalOmega],
            Editable->False], 
           TagBox[
            RowBox[{"(", 
             RowBox[{"0", ",", "2", ",", "0", ",", "0", ",", "0"}], ")"}],
            Derivative],
           MultilineFunction->None], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["x1",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x1[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["x2",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x2[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["x3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x3[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["x4",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x4[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["x5",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x5[],
            Editable->False]}], "]"}], "+", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["\[CapitalOmega]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[CapitalOmega],
            Editable->False], 
           TagBox[
            RowBox[{"(", 
             RowBox[{"2", ",", "0", ",", "0", ",", "0", ",", "0"}], ")"}],
            Derivative],
           MultilineFunction->None], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["x1",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x1[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["x2",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x2[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["x3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x3[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["x4",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x4[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["x5",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x5[],
            Editable->False]}], "]"}]}], ")"}]}]}], ")"}]}]}]}]], "Output",
 CellChangeTimes->{3.7286772991735086`*^9},
 CellLabel->"Out[254]=",
 CellID->1499102257,ExpressionUUID->"040233fb-2af7-4a0a-89f7-e45dbf1546e7"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Constant curvature case:", "MathCaption",
 CellID->1664008414,ExpressionUUID->"0cd90f33-8f4a-4163-8956-76d293bbc8b4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ccrule", "=", 
  RowBox[{"\[CapitalOmega]", "\[Rule]", 
   RowBox[{"Function", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"x1", ",", "x2", ",", "x3", ",", "x4", ",", "x5"}], "}"}], ",", 
     
     RowBox[{"1", "/", 
      RowBox[{"(", 
       RowBox[{"1", "+", 
        RowBox[{
         RowBox[{"\[Kappa]", "/", "4"}], 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"x1", "^", "2"}], "+", 
           RowBox[{"x2", "^", "2"}], "+", 
           RowBox[{"x3", "^", "2"}], "+", 
           RowBox[{"x4", "^", "2"}], "+", 
           RowBox[{"x5", "^", "2"}]}], ")"}]}]}], ")"}]}]}], 
    "]"}]}]}]], "Input",
 CellLabel->"In[255]:=",
 CellID->309879217,ExpressionUUID->"75bed984-cab3-4623-bbb8-84cb845427c0"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox["\[CapitalOmega]",
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`\[CapitalOmega],
   Editable->False], "\[Rule]", 
  RowBox[{"Function", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"x1", ",", "x2", ",", "x3", ",", "x4", ",", "x5"}], "}"}], ",", 
    FractionBox["1", 
     RowBox[{"1", "+", 
      RowBox[{
       FractionBox["1", "4"], " ", "\[Kappa]", " ", 
       RowBox[{"(", 
        RowBox[{
         SuperscriptBox["x1", "2"], "+", 
         SuperscriptBox["x2", "2"], "+", 
         SuperscriptBox["x3", "2"], "+", 
         SuperscriptBox["x4", "2"], "+", 
         SuperscriptBox["x5", "2"]}], ")"}]}]}]]}], "]"}]}]], "Output",
 CellChangeTimes->{3.72867729935394*^9},
 CellLabel->"Out[255]=",
 CellID->1133486795,ExpressionUUID->"ee5c9cbe-8a0c-491a-b6c9-e3aa0c7e6033"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"%%", "/.", "ccrule"}], "//", "Simplify"}]], "Input",
 CellLabel->"In[256]:=",
 CellID->1708787611,ExpressionUUID->"b97b6642-cd9b-47a5-a013-7d8a11f760db"],

Cell[BoxData[
 RowBox[{"20", " ", "\[Kappa]"}]], "Output",
 CellChangeTimes->{3.728677299519579*^9},
 CellLabel->"Out[256]=",
 CellID->185131028,ExpressionUUID->"6d688e24-ec73-4423-8c88-e1fcfbe00720"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"Ricci", "[", "cd", "]"}], "\[Equal]", 
    RowBox[{"4", "\[Kappa]", " ", "met"}]}], "/.", "ccrule"}], "//", 
  "Simplify"}]], "Input",
 CellLabel->"In[257]:=",
 CellID->1626472265,ExpressionUUID->"a416eaa5-f0b4-4e33-95f0-007d029c19bb"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.7286772997031517`*^9},
 CellLabel->"Out[257]=",
 CellID->1940866740,ExpressionUUID->"170dbacb-b437-416a-897e-ebf4fcb435fa"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{
  "ccrule", ",", "\[Kappa]", ",", "met", ",", "cd", ",", "nonzero", ",", 
   "differentnonzero"}], "]"}]], "Input",
 CellLabel->"In[258]:=",
 CellID->460218077,ExpressionUUID->"9fda21d1-380d-4e0d-b0a9-ca331e16e793"],

Cell[BoxData[
 RowBox[{"UndefScalarFunction", "[", "\[CapitalOmega]", "]"}]], "Input",
 CellLabel->"In[259]:=",
 CellID->1796607292,ExpressionUUID->"a4c4e8c4-4fa9-433f-831d-f4ed0c22c871"],

Cell[BoxData[
 RowBox[{"UndefChart", "[", "xs", "]"}]], "Input",
 CellLabel->"In[260]:=",
 CellID->1958679824,ExpressionUUID->"c85f93e1-7e3d-4df9-a839-19064992a182"],

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M", "]"}]], "Input",
 CellLabel->"In[261]:=",
 CellID->754977105,ExpressionUUID->"84da5be2-80f9-4241-bfeb-88c73a0c5913"],

Cell[BoxData[
 RowBox[{"ClearxCobaCache", "[", "All", "]"}]], "Input",
 CellLabel->"In[262]:=",
 CellID->1566817259,ExpressionUUID->"ce6deefc-4524-4c1e-b0f9-24f02e2c9643"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.6.6. Kerr in Boyer-Lindquist x-coordinates", "Subsubsection",
 CellChangeTimes->{{3.728677126241446*^9, 3.7286771264014463`*^9}, 
   3.790876772279477*^9},
 CellID->1053192499,ExpressionUUID->"3f6a2971-b706-4f65-94b2-201e63eba470"],

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M", ",", "4", ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c", ",", "e", ",", "d", ",", "f"}], "}"}]}], 
  "]"}]], "Input",
 CellLabel->"In[263]:=",
 CellID->1446112119,ExpressionUUID->"445f5659-83e6-48cc-9f30-0d6866222c36"],

Cell[BoxData[
 RowBox[{"DefChart", "[", 
  RowBox[{"BL", ",", "M", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2", ",", "3"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"t", "[", "]"}], ",", 
     RowBox[{"r", "[", "]"}], ",", 
     RowBox[{"x", "[", "]"}], ",", 
     RowBox[{"\[Phi]", "[", "]"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[264]:=",
 CellID->1733950934,ExpressionUUID->"a9862d3b-61da-4960-ab3b-5c2b347e2146"],

Cell[BoxData[
 RowBox[{"DefConstantSymbol", "[", 
  RowBox[{"{", 
   RowBox[{"\[CapitalMu]", ",", "\[Alpha]"}], "}"}], "]"}]], "Input",
 CellLabel->"In[265]:=",
 CellID->1160257948,ExpressionUUID->"e2ae665e-bbce-4b11-886c-856d268f2e27"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Sigma", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"r", "[", "]"}], "^", "2"}], "+", 
    RowBox[{
     RowBox[{"\[Alpha]", "^", "2"}], 
     RowBox[{
      RowBox[{"x", "[", "]"}], "^", "2"}]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Delta", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"r", "[", "]"}], "^", "2"}], "+", 
    RowBox[{"\[Alpha]", "^", "2"}], "-", 
    RowBox[{"2", "\[CapitalMu]", " ", 
     RowBox[{"r", "[", "]"}]}]}]}], ";"}]}], "Input",
 CellLabel->"In[266]:=",
 CellID->1303817276,ExpressionUUID->"2ed9ea66-f6dd-4ae4-8ba8-8eae756289d5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"met", "=", 
  RowBox[{"CTensor", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"1", "-", 
         RowBox[{"2", "\[CapitalMu]", " ", 
          RowBox[{
           RowBox[{"r", "[", "]"}], "/", "Sigma"}]}]}], ",", "0", ",", "0", 
        ",", 
        RowBox[{"2", "\[Alpha]", " ", "\[CapitalMu]", " ", 
         RowBox[{"r", "[", "]"}], 
         RowBox[{
          RowBox[{"(", 
           RowBox[{"1", "-", 
            RowBox[{
             RowBox[{"x", "[", "]"}], "^", "2"}]}], ")"}], "/", "Sigma"}]}]}],
        "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{
         RowBox[{"-", "Sigma"}], "/", "Delta"}], ",", "0", ",", "0"}], "}"}], 
      ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", "0", ",", 
        RowBox[{
         RowBox[{"-", "Sigma"}], "/", 
         RowBox[{"(", 
          RowBox[{"1", "-", 
           RowBox[{
            RowBox[{"x", "[", "]"}], "^", "2"}]}], ")"}]}], ",", "0"}], "}"}],
       ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"2", "\[Alpha]", " ", "\[CapitalMu]", " ", 
         RowBox[{"r", "[", "]"}], 
         RowBox[{
          RowBox[{"(", 
           RowBox[{"1", "-", 
            RowBox[{
             RowBox[{"x", "[", "]"}], "^", "2"}]}], ")"}], "/", "Sigma"}]}], 
        ",", "0", ",", "0", ",", 
        RowBox[{
         RowBox[{"-", 
          RowBox[{"(", 
           RowBox[{"1", "-", 
            RowBox[{
             RowBox[{"x", "[", "]"}], "^", "2"}]}], ")"}]}], 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"r", "[", "]"}], "^", "2"}], "+", 
           RowBox[{"\[Alpha]", "^", "2"}], "+", 
           RowBox[{"2", 
            RowBox[{"\[Alpha]", "^", "2"}], "\[CapitalMu]", " ", 
            RowBox[{"r", "[", "]"}], 
            RowBox[{
             RowBox[{"(", 
              RowBox[{"1", "-", 
               RowBox[{
                RowBox[{"x", "[", "]"}], "^", "2"}]}], ")"}], "/", 
             "Sigma"}]}]}], ")"}]}]}], "}"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "BL"}], ",", 
      RowBox[{"-", "BL"}]}], "}"}]}], "]"}]}]], "Input",
 CellLabel->"In[268]:=",
 CellID->333823744,ExpressionUUID->"f68c8677-5cda-4013-ac6f-8918f6d20775"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"1", "-", 
        FractionBox[
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["\[CapitalMu]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[CapitalMu],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False]}], 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha],
             Editable->False], "2"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], "2"]}]}]]}], ",", "0", ",", "0", ",", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[CapitalMu]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalMu],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{"1", "-", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], "2"]}], ")"}]}], 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"], "+", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["\[Alpha]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Alpha],
            Editable->False], "2"], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["x",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x[],
            Editable->False], "2"]}]}]]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", 
       FractionBox[
        RowBox[{
         RowBox[{"-", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]}], "-", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["\[Alpha]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Alpha],
            Editable->False], "2"], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["x",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x[],
            Editable->False], "2"]}]}], 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["\[Alpha]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Alpha],
           Editable->False], "2"], "-", 
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["\[CapitalMu]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[CapitalMu],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False]}], "+", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"]}]], ",", "0", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", 
       FractionBox[
        RowBox[{
         RowBox[{"-", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]}], "-", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["\[Alpha]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Alpha],
            Editable->False], "2"], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["x",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x[],
            Editable->False], "2"]}]}], 
        RowBox[{"1", "-", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], "2"]}]], ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[CapitalMu]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalMu],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{"1", "-", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], "2"]}], ")"}]}], 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"], "+", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["\[Alpha]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Alpha],
            Editable->False], "2"], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["x",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x[],
            Editable->False], "2"]}]}]], ",", "0", ",", "0", ",", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["x",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x[],
            Editable->False], "2"]}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["\[Alpha]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Alpha],
            Editable->False], "2"], "+", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"], "+", 
          FractionBox[
           RowBox[{"2", " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["\[Alpha]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Alpha],
              Editable->False], "2"], " ", 
            InterpretationBox[
             StyleBox["\[CapitalMu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[CapitalMu],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], " ", 
            RowBox[{"(", 
             RowBox[{"1", "-", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], "2"]}], ")"}]}], 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"], "+", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Alpha]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Alpha],
               Editable->False], "2"], " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], "2"]}]}]]}], ")"}]}]}], "}"}]}], "}"}], ",", 
   
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "BL"}], ",", 
     RowBox[{"-", "BL"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.728677320629352*^9},
 CellLabel->"Out[268]=",
 CellID->891901034,ExpressionUUID->"ca924408-d115-452f-a517-be25ee441f02"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"$CVSimplify", "=", "Simplify"}], ";"}]], "Input",
 CellLabel->"In[269]:=",
 CellID->1965872164,ExpressionUUID->"f5819f61-25f3-449f-9838-613b1a72e508"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SetCMetric", "[", 
   RowBox[{"met", ",", "BL", ",", 
    RowBox[{"SignatureOfMetric", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{"3", ",", "1", ",", "0"}], "}"}]}]}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[270]:=",
 CellID->1377551253,ExpressionUUID->"47eb9bb5-7cc0-4670-b1b3-8666c8f8d3ee"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.383434`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.7286773212099323`*^9},
 CellLabel->"Out[270]=",
 CellID->744822378,ExpressionUUID->"769c8ef6-7927-4089-8f2d-72d88b0e8bf7"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"MetricCompute", "[", 
  RowBox[{"met", ",", "BL", ",", 
   RowBox[{"\"\<Weyl\>\"", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}], "]"}]], "Input",
 CellLabel->"In[271]:=",
 CellID->1875756529,ExpressionUUID->"e3a96576-26e6-4a05-8544-d0998f51543d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Ricci", "[", 
  RowBox[{"CovDOfMetric", "[", "met", "]"}], "]"}]], "Input",
 CellLabel->"In[272]:=",
 CellID->780044943,ExpressionUUID->"b2acaca0-b155-4a42-9c52-199087c6df90"],

Cell[BoxData["Zero"], "Output",
 CellChangeTimes->{3.7286773225749826`*^9},
 CellLabel->"Out[272]=",
 CellID->547528683,ExpressionUUID->"317396df-5738-400a-8186-6db6946af311"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"MetricCompute", "[", 
   RowBox[{"met", ",", "BL", ",", 
    RowBox[{"\"\<CDRiemann\>\"", "[", 
     RowBox[{
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}]}], "]"}]}], "]"}], "//", "AbsoluteTiming"}]], "Input",\

 CellLabel->"In[273]:=",
 CellID->975993065,ExpressionUUID->"7d4c40b5-a713-45e1-9b4f-e0fad862722c"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"1.027731`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.7286773237610073`*^9},
 CellLabel->"Out[273]=",
 CellID->347147786,ExpressionUUID->"27b520e9-07ed-4572-8838-26e954637cd9"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"cd", "=", 
   RowBox[{"CovDOfMetric", "[", "met", "]"}]}], ";"}]], "Input",
 CellLabel->"In[274]:=",
 CellID->28281242,ExpressionUUID->"06577532-7635-491f-844a-ed35f14f0795"],

Cell[BoxData[
 RowBox[{
  RowBox[{"cdR", "=", 
   RowBox[{"TensorDerivative", "[", 
    RowBox[{
     RowBox[{"RiemannDown", "[", "cd", "]"}], ",", "cd"}], "]"}]}], 
  ";"}]], "Input",
 CellLabel->"In[275]:=",
 CellID->728948307,ExpressionUUID->"f59b0314-73bb-46e4-adae-779eeeec6aa6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"cdR", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}], ",", 
     RowBox[{"-", "e"}]}], "]"}], "+", 
   RowBox[{"cdR", "[", 
    RowBox[{
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "d"}], ",", 
     RowBox[{"-", "e"}]}], "]"}], "+", 
   RowBox[{"cdR", "[", 
    RowBox[{
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "d"}], ",", 
     RowBox[{"-", "e"}]}], "]"}]}], "//", "Simplify"}]], "Input",
 CellLabel->"In[276]:=",
 CellID->39989614,ExpressionUUID->"955c73ad-2d23-4435-8cb1-e4119b01e593"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728677324105303*^9},
 CellLabel->"Out[276]=",
 CellID->1947254050,ExpressionUUID->"3b7cd555-6c1f-4472-8acc-270a3443d4f9"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Find Killing vectors. Assume [r, x] dependence of the contravariant \
components of the Killings:\
\>", "MathCaption",
 CellID->803715071,ExpressionUUID->"914f8c4e-f516-4f9b-aa77-deff91aa9cc9"],

Cell[BoxData[
 RowBox[{"DefScalarFunction", "[", 
  RowBox[{"{", 
   RowBox[{"\[Xi]0", ",", "\[Xi]1", ",", "\[Xi]2", ",", "\[Xi]3"}], "}"}], 
  "]"}]], "Input",
 CellLabel->"In[277]:=",
 CellID->367058667,ExpressionUUID->"c62b1b22-c1ab-4f73-a0fa-76a3c9c535d0"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{"\[Xi]", "=", 
    RowBox[{"CTensor", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"\[Xi]0", "[", 
         RowBox[{
          RowBox[{"r", "[", "]"}], ",", 
          RowBox[{"x", "[", "]"}]}], "]"}], ",", 
        RowBox[{"\[Xi]1", "[", 
         RowBox[{
          RowBox[{"r", "[", "]"}], ",", 
          RowBox[{"x", "[", "]"}]}], "]"}], ",", 
        RowBox[{"\[Xi]2", "[", 
         RowBox[{
          RowBox[{"r", "[", "]"}], ",", 
          RowBox[{"x", "[", "]"}]}], "]"}], ",", 
        RowBox[{"\[Xi]3", "[", 
         RowBox[{
          RowBox[{"r", "[", "]"}], ",", 
          RowBox[{"x", "[", "]"}]}], "]"}]}], "}"}], ",", 
      RowBox[{"{", "BL", "}"}]}], "]"}]}], ")"}], "[", "a", "]"}]], "Input",
 CellLabel->"In[278]:=",
 CellID->1581982085,ExpressionUUID->"8c3b1d86-e094-40d3-86fd-dc88766da5ca"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{
             InterpretationBox[
              StyleBox["\[Xi]0",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Xi]0,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", 
              RowBox[{
                InterpretationBox[
                 StyleBox[
                  RowBox[{"\[Xi]0"}], AutoSpacing -> False, ShowAutoStyles -> 
                  False], $CellContext`\[Xi]0, Editable -> False], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`x[], Editable -> False]}], "]"}]}]],
           Annotation[#, {1} -> $CellContext`\[Xi]0[
              $CellContext`r[], 
              $CellContext`x[]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             InterpretationBox[
              StyleBox["\[Xi]1",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Xi]1,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", 
              RowBox[{
                InterpretationBox[
                 StyleBox[
                  RowBox[{"\[Xi]1"}], AutoSpacing -> False, ShowAutoStyles -> 
                  False], $CellContext`\[Xi]1, Editable -> False], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`x[], Editable -> False]}], "]"}]}]],
           Annotation[#, {2} -> $CellContext`\[Xi]1[
              $CellContext`r[], 
              $CellContext`x[]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             InterpretationBox[
              StyleBox["\[Xi]2",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Xi]2,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", 
              RowBox[{
                InterpretationBox[
                 StyleBox[
                  RowBox[{"\[Xi]2"}], AutoSpacing -> False, ShowAutoStyles -> 
                  False], $CellContext`\[Xi]2, Editable -> False], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`x[], Editable -> False]}], "]"}]}]],
           Annotation[#, {3} -> $CellContext`\[Xi]2[
              $CellContext`r[], 
              $CellContext`x[]], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            RowBox[{
             InterpretationBox[
              StyleBox["\[Xi]3",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Xi]3,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False]}], "]"}],
            RowBox[{
              RowBox[{"{", "4", "}"}], "\[Rule]", 
              RowBox[{
                InterpretationBox[
                 StyleBox[
                  RowBox[{"\[Xi]3"}], AutoSpacing -> False, ShowAutoStyles -> 
                  False], $CellContext`\[Xi]3, Editable -> False], "[", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                   "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`x[], Editable -> False]}], "]"}]}]],
           Annotation[#, {4} -> $CellContext`\[Xi]3[
              $CellContext`r[], 
              $CellContext`x[]], "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"a"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{
    $CellContext`\[Xi]0[
     $CellContext`r[], 
     $CellContext`x[]], 
    $CellContext`\[Xi]1[
     $CellContext`r[], 
     $CellContext`x[]], 
    $CellContext`\[Xi]2[
     $CellContext`r[], 
     $CellContext`x[]], 
    $CellContext`\[Xi]3[
     $CellContext`r[], 
     $CellContext`x[]]}, {$CellContext`BL}, 0][$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.7286773243729773`*^9},
 CellLabel->"Out[278]=",
 CellID->1995521717,ExpressionUUID->"274837b2-714a-4841-8ab4-682f016f3dc6"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We can compute the equations in two equivalent ways:", "MathCaption",
 CellID->736617837,ExpressionUUID->"3ec649e4-1abd-404c-9ad2-b50e7931f157"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"LieD", "[", 
     RowBox[{"\[Xi]", "[", "a", "]"}], "]"}], "[", 
    RowBox[{"met", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}]}], "]"}], "]"}], "==", 
   RowBox[{
    RowBox[{
     RowBox[{"cd", "[", 
      RowBox[{"-", "a"}], "]"}], "[", 
     RowBox[{"\[Xi]", "[", 
      RowBox[{"-", "b"}], "]"}], "]"}], "+", 
    RowBox[{
     RowBox[{"cd", "[", 
      RowBox[{"-", "b"}], "]"}], "[", 
     RowBox[{"\[Xi]", "[", 
      RowBox[{"-", "a"}], "]"}], "]"}]}]}], "//", "Simplify"}]], "Input",
 CellLabel->"In[279]:=",
 CellID->92761504,ExpressionUUID->"7b7a1063-463c-4ef5-b2a2-1ef4bfe23c57"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728677325948557*^9},
 CellLabel->"Out[279]=",
 CellID->1481382420,ExpressionUUID->"db57d1ce-67b3-4cc6-83b5-fecb514e6812"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Killing", "=", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"First", "@", 
      RowBox[{"Head", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"cd", "[", 
          RowBox[{"-", "a"}], "]"}], "[", 
         RowBox[{"\[Xi]", "[", 
          RowBox[{"-", "b"}], "]"}], "]"}], "+", 
        RowBox[{
         RowBox[{"cd", "[", 
          RowBox[{"-", "b"}], "]"}], "[", 
         RowBox[{"\[Xi]", "[", 
          RowBox[{"-", "a"}], "]"}], "]"}]}], "]"}]}], "//", "Simplify"}], "//",
     "Flatten"}], "//", "DeleteDuplicates"}]}]], "Input",
 CellLabel->"In[280]:=",
 CellID->130982475,ExpressionUUID->"4b9f3fb6-bc5e-416f-9ba0-b25ad17bc504"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   FractionBox[
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox["\[CapitalMu]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalMu],
      Editable->False], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"], "-", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha],
             Editable->False], "2"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], "2"]}]}], ")"}], " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["\[Xi]1",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Xi]1,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False]}], "]"}]}], "+", 
       RowBox[{"2", " ", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha],
          Editable->False], "2"], " ", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["\[Xi]2",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Xi]2,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False]}], "]"}]}]}], ")"}]}], 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"], "+", 
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha],
          Editable->False], "2"], " ", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["x",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x[],
          Editable->False], "2"]}]}], ")"}], "2"]], ",", 
   FractionBox[
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         InterpretationBox[
          StyleBox["\[CapitalMu]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalMu],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]}], "+", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"], "+", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["\[Alpha]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Alpha],
           Editable->False], "2"], " ", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], "2"]}]}], ")"}], " ", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["\[Xi]0",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Xi]0,
         Editable->False], 
        TagBox[
         RowBox[{"(", 
          RowBox[{"1", ",", "0"}], ")"}],
         Derivative],
        MultilineFunction->None], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False]}], "]"}]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[CapitalMu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalMu],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["x",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x[],
          Editable->False], "2"]}], ")"}], " ", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["\[Xi]3",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Xi]3,
         Editable->False], 
        TagBox[
         RowBox[{"(", 
          RowBox[{"1", ",", "0"}], ")"}],
         Derivative],
        MultilineFunction->None], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False]}], "]"}]}]}], 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}]], ",", 
   FractionBox[
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         InterpretationBox[
          StyleBox["\[CapitalMu]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalMu],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]}], "+", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"], "+", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["\[Alpha]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Alpha],
           Editable->False], "2"], " ", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], "2"]}]}], ")"}], " ", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["\[Xi]0",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Xi]0,
         Editable->False], 
        TagBox[
         RowBox[{"(", 
          RowBox[{"0", ",", "1"}], ")"}],
         Derivative],
        MultilineFunction->None], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False]}], "]"}]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[CapitalMu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalMu],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["x",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x[],
          Editable->False], "2"]}], ")"}], " ", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["\[Xi]3",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Xi]3,
         Editable->False], 
        TagBox[
         RowBox[{"(", 
          RowBox[{"0", ",", "1"}], ")"}],
         Derivative],
        MultilineFunction->None], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False]}], "]"}]}]}], 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}]], ",", 
   RowBox[{"-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[CapitalMu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalMu],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], "2"]}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"]}], "+", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["\[Alpha]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Alpha],
              Editable->False], "2"], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["x",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`x[],
              Editable->False], "2"]}]}], ")"}], " ", 
         RowBox[{
          InterpretationBox[
           StyleBox["\[Xi]1",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Xi]1,
           Editable->False], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["x",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x[],
            Editable->False]}], "]"}]}], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha],
             Editable->False], "2"], "+", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}], ")"}], " ", 
         InterpretationBox[
          StyleBox["x",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x[],
          Editable->False], " ", 
         RowBox[{
          InterpretationBox[
           StyleBox["\[Xi]2",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Xi]2,
           Editable->False], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["x",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x[],
            Editable->False]}], "]"}]}]}], ")"}]}], 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"], "+", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["\[Alpha]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Alpha],
           Editable->False], "2"], " ", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], "2"]}]}], ")"}], "2"]]}], ",", 
   RowBox[{"-", 
    FractionBox[
     RowBox[{"2", " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", 
             InterpretationBox[
              StyleBox["\[CapitalMu]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[CapitalMu],
              Editable->False]}], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"]}], "+", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["\[Alpha]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Alpha],
              Editable->False], "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "+", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[CapitalMu]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[CapitalMu],
                Editable->False], " ", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], "2"]}], "-", 
              RowBox[{
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], " ", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
         RowBox[{
          InterpretationBox[
           StyleBox["\[Xi]1",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Xi]1,
           Editable->False], "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["x",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x[],
            Editable->False]}], "]"}]}], "+", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha],
             Editable->False], "2"], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "2"}], " ", 
               InterpretationBox[
                StyleBox["\[CapitalMu]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[CapitalMu],
                Editable->False]}], "+", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}], ")"}]}]}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["\[Alpha]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Alpha],
              Editable->False], "2"], " ", 
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["\[Xi]2",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Xi]2,
              Editable->False], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False]}], "]"}]}], "+", 
           RowBox[{
            RowBox[{"(", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], "2"], "+", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Alpha]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Alpha],
                 Editable->False], "2"], " ", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["x",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x[],
                 Editable->False], "2"]}]}], ")"}], " ", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Xi]1",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Xi]1,
               Editable->False], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"1", ",", "0"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], ",", 
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False]}], "]"}]}]}], ")"}]}]}], ")"}]}], 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha],
          Editable->False], "2"], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", "2"}], " ", 
            InterpretationBox[
             StyleBox["\[CapitalMu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[CapitalMu],
             Editable->False]}], "+", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]}], ")"}]}]}], ")"}], "2"]]}], ",", 
   FractionBox[
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"], "+", 
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha],
          Editable->False], "2"], " ", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["x",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x[],
          Editable->False], "2"]}]}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], "2"]}], ")"}]}], " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["\[Xi]1",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Xi]1,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"0", ",", "1"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False]}], "]"}]}], "+", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["\[Alpha]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Alpha],
            Editable->False], "2"], "+", 
          RowBox[{
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "2"}], " ", 
              InterpretationBox[
               StyleBox["\[CapitalMu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[CapitalMu],
               Editable->False]}], "+", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]}], ")"}]}]}], ")"}], " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["\[Xi]2",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Xi]2,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"1", ",", "0"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False]}], "]"}]}]}], ")"}]}], 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha],
         Editable->False], "2"], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "2"}], " ", 
           InterpretationBox[
            StyleBox["\[CapitalMu]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[CapitalMu],
            Editable->False]}], "+", 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False]}], ")"}]}]}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], "2"]}], ")"}]}]], ",", 
   FractionBox[
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], "2"]}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "2"}], " ", 
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[CapitalMu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[CapitalMu],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["\[Xi]0",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Xi]0,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"1", ",", "0"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False]}], "]"}]}], "+", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "4"], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha],
             Editable->False], "4"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], "2"]}], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha],
             Editable->False], "2"], " ", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["\[CapitalMu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[CapitalMu],
               Editable->False]}], "+", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "-", 
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["\[CapitalMu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[CapitalMu],
               Editable->False], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], "2"]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["\[Xi]3",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Xi]3,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"1", ",", "0"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False]}], "]"}]}]}], ")"}]}], 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}]], ",", 
   FractionBox[
    RowBox[{"2", " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["x",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x[],
            Editable->False], "2"]}], ")"}], " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["\[Xi]1",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Xi]1,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False]}], "]"}]}], "-", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["\[Alpha]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Alpha],
            Editable->False], "2"], "+", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]}], ")"}], " ", 
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["\[Xi]2",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Xi]2,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False]}], "]"}]}], "+", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["x",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x[],
            Editable->False], "2"]}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha],
             Editable->False], "2"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], "2"]}]}], ")"}], " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["\[Xi]2",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Xi]2,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"0", ",", "1"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False]}], "]"}]}]}], ")"}]}], 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], "2"]}], ")"}], "2"]], ",", 
   FractionBox[
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], "2"]}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "2"}], " ", 
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[CapitalMu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[CapitalMu],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["\[Xi]0",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Xi]0,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"0", ",", "1"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False]}], "]"}]}], "+", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "4"], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha],
             Editable->False], "4"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], "2"]}], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha],
             Editable->False], "2"], " ", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["\[CapitalMu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[CapitalMu],
               Editable->False]}], "+", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "-", 
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["\[CapitalMu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[CapitalMu],
               Editable->False], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], "2"]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["\[Xi]3",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Xi]3,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"0", ",", "1"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False]}], "]"}]}]}], ")"}]}], 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}]], ",", 
   RowBox[{
    FractionBox["1", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"], "+", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["\[Alpha]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Alpha],
           Editable->False], "2"], " ", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], "2"]}]}], ")"}], "2"]], 
    RowBox[{"2", " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], "2"]}], ")"}]}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "5"]}], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha],
             Editable->False], "2"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             InterpretationBox[
              StyleBox["\[CapitalMu]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[CapitalMu],
              Editable->False], "-", 
             RowBox[{
              InterpretationBox[
               StyleBox["\[CapitalMu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[CapitalMu],
               Editable->False], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], "2"]}], "-", 
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], "2"]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha],
             Editable->False], "4"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], "4"]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox["\[CapitalMu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[CapitalMu],
               Editable->False], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], "2"]}], ")"}]}]}], ")"}]}]}], ")"}], " ", 
        
        RowBox[{
         InterpretationBox[
          StyleBox["\[Xi]1",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Xi]1,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False]}], "]"}]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], " ", 
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "6"], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha],
             Editable->False], "6"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], "4"]}], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha],
             Editable->False], "2"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "3"], " ", 
           RowBox[{"(", 
            RowBox[{
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "+", 
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], "2"]}], "-", 
             RowBox[{"4", " ", 
              InterpretationBox[
               StyleBox["\[CapitalMu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[CapitalMu],
               Editable->False], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], "2"]}], ")"}]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha],
             Editable->False], "4"], " ", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], "2"], " ", 
              RowBox[{"(", 
               RowBox[{"2", "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], "2"]}], ")"}]}], "-", 
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["\[CapitalMu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[CapitalMu],
               Editable->False], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["x",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x[],
                  Editable->False], "4"]}], ")"}]}]}], ")"}]}]}], ")"}], " ", 
        
        RowBox[{
         InterpretationBox[
          StyleBox["\[Xi]2",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Xi]2,
          Editable->False], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False]}], "]"}]}]}], ")"}]}]}]}], "}"}]], "Output",
 CellChangeTimes->{3.7286773261869087`*^9},
 CellLabel->"Out[280]=",
 CellID->1816413269,ExpressionUUID->"a9be3bfe-b70b-412f-9c54-045f8a0149bd"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Solve", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"Killing", "[", 
     RowBox[{"[", 
      RowBox[{"{", 
       RowBox[{"1", ",", "4"}], "}"}], "]"}], "]"}], "\[Equal]", 
    RowBox[{"{", 
     RowBox[{"0", ",", "0"}], "}"}]}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"\[Xi]1", "[", 
      RowBox[{
       RowBox[{"r", "[", "]"}], ",", 
       RowBox[{"x", "[", "]"}]}], "]"}], ",", 
     RowBox[{"\[Xi]2", "[", 
      RowBox[{
       RowBox[{"r", "[", "]"}], ",", 
       RowBox[{"x", "[", "]"}]}], "]"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[281]:=",
 CellID->100969460,ExpressionUUID->"80a55849-fd05-40a3-9696-25270f4168c2"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Xi]1",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Xi]1,
       Editable->False], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False]}], "]"}], "\[Rule]", "0"}], ",", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Xi]2",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Xi]2,
       Editable->False], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False]}], "]"}], "\[Rule]", "0"}]}], "}"}], "}"}]], "Output",\

 CellChangeTimes->{3.728677326404129*^9},
 CellLabel->"Out[281]=",
 CellID->1977769978,ExpressionUUID->"c63e4a9a-35b0-4ac3-8097-db7892170e31"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Killing", "=", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"Killing", "/.", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"\[Xi]1", "|", "\[Xi]2"}], ")"}], "\[Rule]", 
       RowBox[{"Function", "[", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"r", ",", "x"}], "}"}], ",", "0"}], "]"}]}]}], "//", 
     "Simplify"}], "//", "DeleteDuplicates"}], "//", 
   RowBox[{
    RowBox[{"DeleteCases", "[", 
     RowBox[{"#", ",", "0"}], "]"}], "&"}]}]}]], "Input",
 CellLabel->"In[282]:=",
 CellID->941478087,ExpressionUUID->"cb19ad0f-dbb1-41fc-9ee3-1abcfcc923df"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   FractionBox[
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         InterpretationBox[
          StyleBox["\[CapitalMu]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalMu],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]}], "+", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"], "+", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["\[Alpha]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Alpha],
           Editable->False], "2"], " ", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], "2"]}]}], ")"}], " ", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["\[Xi]0",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Xi]0,
         Editable->False], 
        TagBox[
         RowBox[{"(", 
          RowBox[{"1", ",", "0"}], ")"}],
         Derivative],
        MultilineFunction->None], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False]}], "]"}]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[CapitalMu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalMu],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["x",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x[],
          Editable->False], "2"]}], ")"}], " ", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["\[Xi]3",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Xi]3,
         Editable->False], 
        TagBox[
         RowBox[{"(", 
          RowBox[{"1", ",", "0"}], ")"}],
         Derivative],
        MultilineFunction->None], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False]}], "]"}]}]}], 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}]], ",", 
   FractionBox[
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         InterpretationBox[
          StyleBox["\[CapitalMu]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalMu],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]}], "+", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"], "+", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["\[Alpha]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Alpha],
           Editable->False], "2"], " ", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], "2"]}]}], ")"}], " ", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["\[Xi]0",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Xi]0,
         Editable->False], 
        TagBox[
         RowBox[{"(", 
          RowBox[{"0", ",", "1"}], ")"}],
         Derivative],
        MultilineFunction->None], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False]}], "]"}]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[CapitalMu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalMu],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["x",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x[],
          Editable->False], "2"]}], ")"}], " ", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["\[Xi]3",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Xi]3,
         Editable->False], 
        TagBox[
         RowBox[{"(", 
          RowBox[{"0", ",", "1"}], ")"}],
         Derivative],
        MultilineFunction->None], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False]}], "]"}]}]}], 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}]], ",", 
   FractionBox[
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], "2"]}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "2"}], " ", 
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[CapitalMu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[CapitalMu],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["\[Xi]0",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Xi]0,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"1", ",", "0"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False]}], "]"}]}], "+", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "4"], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha],
             Editable->False], "4"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], "2"]}], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha],
             Editable->False], "2"], " ", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["\[CapitalMu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[CapitalMu],
               Editable->False]}], "+", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "-", 
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["\[CapitalMu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[CapitalMu],
               Editable->False], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], "2"]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["\[Xi]3",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Xi]3,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"1", ",", "0"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False]}], "]"}]}]}], ")"}]}], 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}]], ",", 
   FractionBox[
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], "2"]}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "2"}], " ", 
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[CapitalMu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[CapitalMu],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["\[Xi]0",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Xi]0,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"0", ",", "1"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False]}], "]"}]}], "+", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "4"], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha],
             Editable->False], "4"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], "2"]}], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha],
             Editable->False], "2"], " ", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["\[CapitalMu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[CapitalMu],
               Editable->False]}], "+", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "-", 
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["\[CapitalMu]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[CapitalMu],
               Editable->False], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], "2"]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["x",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`x[],
                Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["\[Xi]3",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Xi]3,
           Editable->False], 
          TagBox[
           RowBox[{"(", 
            RowBox[{"0", ",", "1"}], ")"}],
           Derivative],
          MultilineFunction->None], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], ",", 
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False]}], "]"}]}]}], ")"}]}], 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}]]}], "}"}]], "Output",
 CellChangeTimes->{3.728677326588376*^9},
 CellLabel->"Out[282]=",
 CellID->1628131980,ExpressionUUID->"d942dc06-20a2-45cb-9f95-3a1110b91773"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Solve", "[", 
  RowBox[{
   RowBox[{"Killing", "==", 
    RowBox[{"{", 
     RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Xi]0",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Xi]0,
        Editable->False], 
       TagBox[
        RowBox[{"(", 
         RowBox[{"1", ",", "0"}], ")"}],
        Derivative],
       MultilineFunction->None], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False]}], "]"}], ",", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Xi]0",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Xi]0,
        Editable->False], 
       TagBox[
        RowBox[{"(", 
         RowBox[{"0", ",", "1"}], ")"}],
        Derivative],
       MultilineFunction->None], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False]}], "]"}], ",", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Xi]3",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Xi]3,
        Editable->False], 
       TagBox[
        RowBox[{"(", 
         RowBox[{"1", ",", "0"}], ")"}],
        Derivative],
       MultilineFunction->None], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False]}], "]"}], ",", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Xi]3",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Xi]3,
        Editable->False], 
       TagBox[
        RowBox[{"(", 
         RowBox[{"0", ",", "1"}], ")"}],
        Derivative],
       MultilineFunction->None], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False]}], "]"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[283]:=",
 CellID->320489248,ExpressionUUID->"85c2325b-fd4c-4e7b-95cb-021a93cb1e8b"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Xi]0",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Xi]0,
        Editable->False], 
       TagBox[
        RowBox[{"(", 
         RowBox[{"1", ",", "0"}], ")"}],
        Derivative],
       MultilineFunction->None], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False]}], "]"}], "\[Rule]", "0"}], ",", 
    RowBox[{
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Xi]0",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Xi]0,
        Editable->False], 
       TagBox[
        RowBox[{"(", 
         RowBox[{"0", ",", "1"}], ")"}],
        Derivative],
       MultilineFunction->None], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False]}], "]"}], "\[Rule]", "0"}], ",", 
    RowBox[{
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Xi]3",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Xi]3,
        Editable->False], 
       TagBox[
        RowBox[{"(", 
         RowBox[{"1", ",", "0"}], ")"}],
        Derivative],
       MultilineFunction->None], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False]}], "]"}], "\[Rule]", "0"}], ",", 
    RowBox[{
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Xi]3",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Xi]3,
        Editable->False], 
       TagBox[
        RowBox[{"(", 
         RowBox[{"0", ",", "1"}], ")"}],
        Derivative],
       MultilineFunction->None], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False]}], "]"}], "\[Rule]", "0"}]}], "}"}], "}"}]], "Output",\

 CellChangeTimes->{3.728677326793827*^9},
 CellLabel->"Out[283]=",
 CellID->777557786,ExpressionUUID->"d25fed79-b283-4f78-bd11-a99ab8b08b2a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We can check the result with a Lie derivative:", "MathCaption",
 CellID->1729984190,ExpressionUUID->"560351bf-b3fc-4b93-9f55-afd0b6e5e11d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"LieD", "[", 
   RowBox[{
    RowBox[{"CTensor", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"C0", ",", "0", ",", "0", ",", "C3"}], "}"}], ",", 
      RowBox[{"{", "BL", "}"}]}], "]"}], "[", "a", "]"}], "]"}], "[", 
  RowBox[{"met", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}]}], "]"}], "]"}]], "Input",
 CellLabel->"In[284]:=",
 CellID->1063184597,ExpressionUUID->"1da75305-040b-45e4-8158-ae7287757726"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728677326977882*^9},
 CellLabel->"Out[284]=",
 CellID->975717307,ExpressionUUID->"413378af-acf2-43ab-b37b-b131f9d3cde3"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Clean up:", "MathCaption",
 CellID->595419003,ExpressionUUID->"6c9f2426-4eb2-4f07-9d89-25d8c10c8ee8"],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{
  "met", ",", "cdR", ",", "Sigma", ",", "Delta", ",", "\[Xi]", ",", "C0", ",",
    "C3", ",", "cd", ",", "Killing"}], "]"}]], "Input",
 CellLabel->"In[285]:=",
 CellID->2080872048,ExpressionUUID->"b0b2259a-1b48-4c71-a8bf-16750c357372"],

Cell[BoxData[
 RowBox[{"UndefScalarFunction", "[", 
  RowBox[{"{", 
   RowBox[{"\[Xi]0", ",", "\[Xi]1", ",", "\[Xi]2", ",", "\[Xi]3"}], "}"}], 
  "]"}]], "Input",
 CellLabel->"In[286]:=",
 CellID->1445319341,ExpressionUUID->"bfb596e1-a7ca-4837-b7f9-fb981e52661d"],

Cell[BoxData[
 RowBox[{"UndefConstantSymbol", "[", 
  RowBox[{"{", 
   RowBox[{"\[Alpha]", ",", "\[CapitalMu]"}], "}"}], "]"}]], "Input",
 CellLabel->"In[287]:=",
 CellID->1294632970,ExpressionUUID->"fd8eadc9-9bdf-4f49-bcdf-573a2a2a2512"],

Cell[BoxData[
 RowBox[{"UndefChart", "[", "BL", "]"}]], "Input",
 CellLabel->"In[288]:=",
 CellID->592015386,ExpressionUUID->"70a39765-b714-4990-a889-12240bd8a7da"],

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M", "]"}]], "Input",
 CellLabel->"In[289]:=",
 CellID->710855951,ExpressionUUID->"ff9d74f5-b120-48d5-95be-90762f94294a"],

Cell[BoxData[
 RowBox[{"ClearxCobaCache", "[", "All", "]"}]], "Input",
 CellLabel->"In[290]:=",
 CellID->904440196,ExpressionUUID->"ba385016-2a3e-411c-a823-6c80f284fa95"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.6.7. Floratos-Kehagias 8-dim Ricci-flat metric", "Subsubsection",
 CellChangeTimes->{{3.728677130209395*^9, 3.728677130609417*^9}, 
   3.790876773855007*^9},
 CellID->1037120706,ExpressionUUID->"6b504b5f-1275-47b2-b42c-54c202b212da"],

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M", ",", "8", ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f"}], "}"}]}], 
  "]"}]], "Input",
 CellLabel->"In[291]:=",
 CellID->1910426690,ExpressionUUID->"7cf91e03-3e8a-489b-a896-6f6cca2f2cf0"],

Cell[BoxData[
 RowBox[{"DefChart", "[", 
  RowBox[{"FK", ",", "M", ",", 
   RowBox[{"Range", "[", "8", "]"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"x1", "[", "]"}], ",", 
     RowBox[{"x2", "[", "]"}], ",", 
     RowBox[{"x3", "[", "]"}], ",", 
     RowBox[{"x4", "[", "]"}], ",", 
     RowBox[{"x5", "[", "]"}], ",", 
     RowBox[{"x6", "[", "]"}], ",", 
     RowBox[{"x7", "[", "]"}], ",", 
     RowBox[{"x8", "[", "]"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[292]:=",
 CellID->1513372533,ExpressionUUID->"d5ad536c-6914-4007-9ce1-a5626f3908da"],

Cell[BoxData[
 RowBox[{"DefScalarFunction", "[", 
  RowBox[{"{", 
   RowBox[{
   "\[Alpha]", ",", "\[Beta]", ",", "\[Chi]", ",", "\[Upsilon]", ",", 
    "\[Gamma]", ",", "\[Eta]"}], "}"}], "]"}]], "Input",
 CellLabel->"In[293]:=",
 CellID->455847063,ExpressionUUID->"98d9d1ef-5f42-4606-aeb3-68caaa45ca3f"],

Cell[BoxData[
 RowBox[{
  RowBox[{"met", "=", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{
          RowBox[{
           RowBox[{"\[Upsilon]", "^", "2"}], " ", 
           RowBox[{"\[Gamma]", "^", "2"}], " ", 
           RowBox[{"\[Eta]", "^", "2"}]}], ",", "0", ",", "0", ",", "0", ",", 
          "0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{"0", ",", 
          RowBox[{
           RowBox[{"\[Upsilon]", "^", "2"}], " ", 
           RowBox[{"\[Gamma]", "^", "2"}], 
           RowBox[{"\[Eta]", "^", "2"}]}], ",", "0", ",", "0", ",", "0", ",", 
          "0", ",", "0", ",", "0"}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{"0", ",", "0", ",", 
          RowBox[{"1", "/", 
           RowBox[{"\[Upsilon]", "^", "2"}]}], ",", "0", ",", "0", ",", "0", 
          ",", "0", ",", 
          RowBox[{"2", 
           RowBox[{"\[Alpha]", "/", 
            RowBox[{"\[Upsilon]", "^", "2"}]}]}]}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{"0", ",", "0", ",", "0", ",", 
          RowBox[{"1", "/", 
           RowBox[{"\[Gamma]", "^", "2"}]}], ",", 
          RowBox[{"2", 
           RowBox[{"\[Beta]", "/", 
            RowBox[{"\[Gamma]", "^", "2"}]}]}], ",", "0", ",", "0", ",", 
          "0"}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{"0", ",", "0", ",", "0", ",", 
          RowBox[{"2", 
           RowBox[{"\[Beta]", "/", 
            RowBox[{"\[Gamma]", "^", "2"}]}]}], ",", 
          RowBox[{
           RowBox[{"\[Gamma]", "^", "2"}], "+", 
           RowBox[{"4", 
            RowBox[{
             RowBox[{"\[Beta]", "^", "2"}], "/", 
             RowBox[{"\[Gamma]", "^", "2"}]}]}]}], ",", "0", ",", "0", ",", 
          "0"}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", "0", ",", 
          RowBox[{
           RowBox[{"\[Eta]", "^", "2"}], "+", 
           RowBox[{"4", 
            RowBox[{
             RowBox[{"\[Chi]", "^", "2"}], "/", 
             RowBox[{"\[Eta]", "^", "2"}]}]}]}], ",", 
          RowBox[{"2", 
           RowBox[{"\[Chi]", "/", 
            RowBox[{"\[Eta]", "^", "2"}]}]}], ",", "0"}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", "0", ",", 
          RowBox[{"2", 
           RowBox[{"\[Chi]", "/", 
            RowBox[{"\[Eta]", "^", "2"}]}]}], ",", 
          RowBox[{"1", "/", 
           RowBox[{"\[Eta]", "^", "2"}]}], ",", "0"}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{"0", ",", "0", ",", 
          RowBox[{"2", 
           RowBox[{"\[Alpha]", "/", 
            RowBox[{"\[Upsilon]", "^", "2"}]}]}], ",", "0", ",", "0", ",", 
          "0", ",", "0", ",", 
          RowBox[{
           RowBox[{"4", 
            RowBox[{
             RowBox[{"\[Alpha]", "^", "2"}], "/", 
             RowBox[{"\[Upsilon]", "^", "2"}]}]}], "+", 
           RowBox[{"\[Upsilon]", "^", "2"}]}]}], "}"}]}], "}"}], "/.", 
      RowBox[{
       RowBox[{"f", ":", 
        RowBox[{"(", 
         RowBox[{
         "\[Alpha]", "|", "\[Beta]", "|", "\[Chi]", "|", "\[Upsilon]", "|", 
          "\[Gamma]", "|", "\[Eta]"}], ")"}]}], "\[Rule]", 
       RowBox[{"f", "[", 
        RowBox[{
         RowBox[{"x1", "[", "]"}], ",", 
         RowBox[{"x2", "[", "]"}]}], "]"}]}]}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "FK"}], ",", 
       RowBox[{"-", "FK"}]}], "}"}]}], "]"}]}], ";"}]], "Input",
 CellLabel->"In[294]:=",
 CellID->1347418992,ExpressionUUID->"a2256773-823c-4c34-b71b-cdca39a73e14"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"met", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}]}], "]"}]], "Input",
 CellLabel->"In[295]:=",
 CellID->855940929,ExpressionUUID->"66037e10-3c18-4fa3-8447-99c43519d93c"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["\[Gamma]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Gamma],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}], "2"], " ", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["\[Eta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Eta],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}], "2"], " ", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["\[Upsilon]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Upsilon],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}], "2"]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{
                SuperscriptBox[
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                SuperscriptBox[
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                SuperscriptBox[
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}]}]],
           
           Annotation[#, {1, 1} -> ($CellContext`\[Gamma][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^2 $CellContext`\[Eta][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^2) $CellContext`\[Upsilon][
                $CellContext`x1[], 
                $CellContext`x2[]]^2, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 4} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "5"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 5} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "6"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 6} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "7"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 7} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "8"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 8} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["\[Gamma]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Gamma],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}], "2"], " ", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["\[Eta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Eta],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}], "2"], " ", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["\[Upsilon]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Upsilon],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}], "2"]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{
                SuperscriptBox[
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                SuperscriptBox[
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                SuperscriptBox[
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}]}]],
           
           Annotation[#, {2, 2} -> ($CellContext`\[Gamma][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^2 $CellContext`\[Eta][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^2) $CellContext`\[Upsilon][
                $CellContext`x1[], 
                $CellContext`x2[]]^2, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 4} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "5"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 5} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "6"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 6} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "7"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 7} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "8"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 8} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox["1", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["\[Upsilon]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Upsilon],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}], "2"]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", 
              FractionBox["1", 
               SuperscriptBox[
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                   Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]]}]],
           Annotation[#, {3, 3} -> $CellContext`\[Upsilon][
               $CellContext`x1[], 
               $CellContext`x2[]]^(-2), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 4} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "5"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 5} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "6"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 6} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "7"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 7} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{"2", " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Alpha]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Alpha],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}]}], 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["\[Upsilon]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Upsilon],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}], "2"]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "8"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{"2", " ", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
               SuperscriptBox[
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                   Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]]}]],
           Annotation[#, {3, 8} -> (2 $CellContext`\[Alpha][
                $CellContext`x1[], 
                $CellContext`x2[]]) $CellContext`\[Upsilon][
                $CellContext`x1[], 
                $CellContext`x2[]]^(-2), "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox["1", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["\[Gamma]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Gamma],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}], "2"]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "4"}], "}"}], "\[Rule]", 
              FractionBox["1", 
               SuperscriptBox[
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                   False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]]}]],
           Annotation[#, {4, 4} -> $CellContext`\[Gamma][
               $CellContext`x1[], 
               $CellContext`x2[]]^(-2), "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{"2", " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Beta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Beta],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}]}], 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["\[Gamma]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Gamma],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}], "2"]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "5"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{"2", " ", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
               SuperscriptBox[
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                   False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]]}]],
           Annotation[#, {4, 5} -> (2 $CellContext`\[Beta][
                $CellContext`x1[], 
                $CellContext`x2[]]) $CellContext`\[Gamma][
                $CellContext`x1[], 
                $CellContext`x2[]]^(-2), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "6"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 6} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "7"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 7} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "8"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 8} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {5, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {5, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {5, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{"2", " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Beta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Beta],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}]}], 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["\[Gamma]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Gamma],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}], "2"]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "4"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{"2", " ", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
               SuperscriptBox[
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                   False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]]}]],
           Annotation[#, {5, 4} -> (2 $CellContext`\[Beta][
                $CellContext`x1[], 
                $CellContext`x2[]]) $CellContext`\[Gamma][
                $CellContext`x1[], 
                $CellContext`x2[]]^(-2), "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             FractionBox[
              RowBox[{"4", " ", 
               SuperscriptBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["\[Beta]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Beta],
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x1",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x1[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["x2",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x2[],
                   Editable->False]}], "]"}], "2"]}], 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Gamma]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Gamma],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"]], "+", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["\[Gamma]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Gamma],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}], "2"]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "5"}], "}"}], "\[Rule]", 
              RowBox[{
                FractionBox[
                 RowBox[{"4", " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]], 
                "+", 
                SuperscriptBox[
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}]}]],
           
           Annotation[#, {5, 5} -> (4 $CellContext`\[Beta][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^2) $CellContext`\[Gamma][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-2) + $CellContext`\[Gamma][
                $CellContext`x1[], 
                $CellContext`x2[]]^2, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "6"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {5, 6} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "7"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {5, 7} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "8"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {5, 8} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"6", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {6, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"6", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {6, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"6", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {6, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"6", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {6, 4} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"6", ",", "5"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {6, 5} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["\[Eta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Eta],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}], "2"], "+", 
             FractionBox[
              RowBox[{"4", " ", 
               SuperscriptBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["\[Chi]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Chi],
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x1",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x1[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["x2",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x2[],
                   Editable->False]}], "]"}], "2"]}], 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Eta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Eta],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"]]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"6", ",", "6"}], "}"}], "\[Rule]", 
              RowBox[{
                SuperscriptBox[
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                FractionBox[
                 RowBox[{"4", " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]]}]}]],
           Annotation[#, {6, 6} -> $CellContext`\[Eta][
                $CellContext`x1[], 
                $CellContext`x2[]]^2 + (4 $CellContext`\[Eta][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2)) $CellContext`\[Chi][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^2, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{"2", " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Chi]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Chi],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}]}], 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["\[Eta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Eta],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}], "2"]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"6", ",", "7"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{"2", " ", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
               SuperscriptBox[
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]]}]],
           Annotation[#, {6, 7} -> (2 $CellContext`\[Eta][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-2)) $CellContext`\[Chi][
               $CellContext`x1[], 
               $CellContext`x2[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"6", ",", "8"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {6, 8} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"7", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {7, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"7", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {7, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"7", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {7, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"7", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {7, 4} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"7", ",", "5"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {7, 5} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{"2", " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Chi]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Chi],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}]}], 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["\[Eta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Eta],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}], "2"]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"7", ",", "6"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{"2", " ", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
               SuperscriptBox[
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]]}]],
           Annotation[#, {7, 6} -> (2 $CellContext`\[Eta][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-2)) $CellContext`\[Chi][
               $CellContext`x1[], 
               $CellContext`x2[]], "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox["1", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["\[Eta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Eta],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}], "2"]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"7", ",", "7"}], "}"}], "\[Rule]", 
              FractionBox["1", 
               SuperscriptBox[
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]]}]],
           Annotation[#, {7, 7} -> $CellContext`\[Eta][
               $CellContext`x1[], 
               $CellContext`x2[]]^(-2), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"7", ",", "8"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {7, 8} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"8", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {8, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"8", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {8, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{"2", " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Alpha]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Alpha],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}]}], 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["\[Upsilon]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Upsilon],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}], "2"]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"8", ",", "3"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{"2", " ", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
               SuperscriptBox[
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                   Editable -> False], "[", 
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]]}]],
           Annotation[#, {8, 3} -> (2 $CellContext`\[Alpha][
                $CellContext`x1[], 
                $CellContext`x2[]]) $CellContext`\[Upsilon][
                $CellContext`x1[], 
                $CellContext`x2[]]^(-2), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"8", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {8, 4} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"8", ",", "5"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {8, 5} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"8", ",", "6"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {8, 6} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"8", ",", "7"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {8, 7} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             FractionBox[
              RowBox[{"4", " ", 
               SuperscriptBox[
                RowBox[{
                 InterpretationBox[
                  StyleBox["\[Alpha]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Alpha],
                  Editable->False], "[", 
                 RowBox[{
                  InterpretationBox[
                   StyleBox["x1",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x1[],
                   Editable->False], ",", 
                  InterpretationBox[
                   StyleBox["x2",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`x2[],
                   Editable->False]}], "]"}], "2"]}], 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Upsilon]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Upsilon],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"]], "+", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["\[Upsilon]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Upsilon],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}], "2"]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"8", ",", "8"}], "}"}], "\[Rule]", 
              RowBox[{
                FractionBox[
                 RowBox[{"4", " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]], 
                "+", 
                SuperscriptBox[
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}]}]],
           
           Annotation[#, {8, 8} -> (4 $CellContext`\[Alpha][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^2) $CellContext`\[Upsilon][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-2) + $CellContext`\[Upsilon][
                $CellContext`x1[], 
                $CellContext`x2[]]^2, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 9 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 9 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " "},
         {"a", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{($CellContext`\[Gamma][
         $CellContext`x1[], 
         $CellContext`x2[]]^2 $CellContext`\[Eta][
         $CellContext`x1[], 
         $CellContext`x2[]]^2) $CellContext`\[Upsilon][
        $CellContext`x1[], 
        $CellContext`x2[]]^2, 0, 0, 0, 0, 0, 0, 0}, {0, ($CellContext`\[Gamma][
         $CellContext`x1[], 
         $CellContext`x2[]]^2 $CellContext`\[Eta][
         $CellContext`x1[], 
         $CellContext`x2[]]^2) $CellContext`\[Upsilon][
        $CellContext`x1[], 
        $CellContext`x2[]]^2, 0, 0, 0, 0, 0, 0}, {
    0, 0, $CellContext`\[Upsilon][
       $CellContext`x1[], 
       $CellContext`x2[]]^(-2), 0, 0, 0, 0, (2 $CellContext`\[Alpha][
        $CellContext`x1[], 
        $CellContext`x2[]]) $CellContext`\[Upsilon][
        $CellContext`x1[], 
        $CellContext`x2[]]^(-2)}, {0, 0, 0, $CellContext`\[Gamma][
       $CellContext`x1[], 
       $CellContext`x2[]]^(-2), (2 $CellContext`\[Beta][
        $CellContext`x1[], 
        $CellContext`x2[]]) $CellContext`\[Gamma][
        $CellContext`x1[], 
        $CellContext`x2[]]^(-2), 0, 0, 0}, {0, 0, 0, (2 $CellContext`\[Beta][
        $CellContext`x1[], 
        $CellContext`x2[]]) $CellContext`\[Gamma][
        $CellContext`x1[], 
        $CellContext`x2[]]^(-2), (4 $CellContext`\[Beta][
          $CellContext`x1[], 
          $CellContext`x2[]]^2) $CellContext`\[Gamma][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-2) + $CellContext`\[Gamma][
        $CellContext`x1[], 
        $CellContext`x2[]]^2, 0, 0, 0}, {0, 0, 0, 0, 0, $CellContext`\[Eta][
        $CellContext`x1[], 
        $CellContext`x2[]]^2 + (4 $CellContext`\[Eta][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2)) $CellContext`\[Chi][
         $CellContext`x1[], 
         $CellContext`x2[]]^2, (2 $CellContext`\[Eta][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-2)) $CellContext`\[Chi][
       $CellContext`x1[], 
       $CellContext`x2[]], 0}, {0, 0, 0, 0, 0, (2 $CellContext`\[Eta][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-2)) $CellContext`\[Chi][
       $CellContext`x1[], 
       $CellContext`x2[]], $CellContext`\[Eta][
       $CellContext`x1[], 
       $CellContext`x2[]]^(-2), 0}, {0, 0, (2 $CellContext`\[Alpha][
        $CellContext`x1[], 
        $CellContext`x2[]]) $CellContext`\[Upsilon][
        $CellContext`x1[], 
        $CellContext`x2[]]^(-2), 0, 0, 0, 0, (4 $CellContext`\[Alpha][
          $CellContext`x1[], 
          $CellContext`x2[]]^2) $CellContext`\[Upsilon][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-2) + $CellContext`\[Upsilon][
        $CellContext`x1[], 
        $CellContext`x2[]]^2}}, {-$CellContext`FK, -$CellContext`FK}, 
   0][-$CellContext`a, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.728677355179558*^9},
 CellLabel->"Out[295]=",
 CellID->272741481,ExpressionUUID->"fb266273-8da4-4a93-af78-72470826eb79"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"$CVSimplify", "=", "Simplify"}], ";"}]], "Input",
 CellLabel->"In[296]:=",
 CellID->1249549649,ExpressionUUID->"54cfadc8-f25b-412b-ab1f-447592da1674"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SetCMetric", "[", 
   RowBox[{"met", ",", "FK", ",", 
    RowBox[{"SignatureOfMetric", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{"7", ",", "1", ",", "0"}], "}"}]}]}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[297]:=",
 CellID->938366730,ExpressionUUID->"21eede11-47c2-416c-9817-fff095bddc9f"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.365182`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.728677355749601*^9},
 CellLabel->"Out[297]=",
 CellID->1235968469,ExpressionUUID->"16fee66b-a18c-4c15-91b2-e17fd60c481a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"MetricCompute", "[", 
   RowBox[{"met", ",", "FK", ",", 
    RowBox[{"\"\<Weyl\>\"", "[", 
     RowBox[{
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}]}], "]"}], ",", 
    RowBox[{"Parallelize", "\[Rule]", "True"}]}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[298]:=",
 CellID->282041222,ExpressionUUID->"f50180be-36cf-4f8e-ba37-aaf178342d78"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"10.267368`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.728677366138257*^9},
 CellLabel->"Out[298]=",
 CellID->312741052,ExpressionUUID->"3180d89f-80f1-4584-8874-66fc22683bb0"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{"ricci", "=", 
    RowBox[{"Ricci", "[", 
     RowBox[{"CovDOfMetric", "[", "met", "]"}], "]"}]}], ")"}], "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}]}], "]"}]], "Input",
 CellLabel->"In[299]:=",
 CellID->1088936812,ExpressionUUID->"dd2efce4-19f2-40ac-95bf-e5194c109bd4"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{"-", 
                  FractionBox[
                   RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "4"]]}], 
                "+", 
                FractionBox[
                 RowBox[{
                   SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "-", 
                   SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]], 
                "+", 
                FractionBox[
                 RowBox[{
                   SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "-", 
                   SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]], 
                "-", 
                FractionBox[
                 RowBox[{"2", " ", 
                   SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "4"]], 
                "-", 
                FractionBox[
                 RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]], "-", 
                FractionBox[
                 RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]], "-", 
                FractionBox[
                 RowBox[{
                   RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                   "+", 
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "3"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "4"]]}]}]],
           Annotation[#, {1, 1} -> ((-2) $CellContext`\[Gamma][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-4)) 
              Derivative[1, 0][$CellContext`\[Beta]][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^2 + $CellContext`\[Gamma][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-2) (
               Derivative[0, 1][$CellContext`\[Gamma]][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^2 - 
               Derivative[1, 0][$CellContext`\[Gamma]][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^2) + $CellContext`\[Eta][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-2) (
               Derivative[0, 1][$CellContext`\[Eta]][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^2 - Derivative[1, 0][$CellContext`\[Eta]][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^2) - (2 $CellContext`\[Eta][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-4)) 
             Derivative[1, 0][$CellContext`\[Chi]][
                $CellContext`x1[], 
                $CellContext`x2[]]^2 - $CellContext`\[Gamma][
                $CellContext`x1[], 
                $CellContext`x2[]]^(-1) (
              Derivative[0, 2][$CellContext`\[Gamma]][
                $CellContext`x1[], 
                $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Gamma]][
                $CellContext`x1[], 
                $CellContext`x2[]]) - $CellContext`\[Eta][
                $CellContext`x1[], 
                $CellContext`x2[]]^(-1) (Derivative[0, 2][$CellContext`\[Eta]][
                $CellContext`x1[], 
                $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Eta]][
                $CellContext`x1[], 
                $CellContext`x2[]]) - $CellContext`\[Upsilon][
                $CellContext`x1[], 
                $CellContext`x2[]]^(-4) (
              2 Derivative[1, 0][$CellContext`\[Alpha]][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^2 + $CellContext`\[Upsilon][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^2 (-
                 Derivative[0, 1][$CellContext`\[Upsilon]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^2 + 
                Derivative[1, 0][$CellContext`\[Upsilon]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2) + $CellContext`\[Upsilon][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^3 (
                Derivative[0, 2][$CellContext`\[Upsilon]][
                  $CellContext`x1[], 
                  $CellContext`x2[]] + 
                Derivative[2, 0][$CellContext`\[Upsilon]][
                  $CellContext`x1[], 
                  $CellContext`x2[]])), "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{"-", 
                  FractionBox[
                   RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "4"]]}], 
                "-", 
                FractionBox[
                 RowBox[{"2", " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "4"]], 
                "-", 
                FractionBox[
                 RowBox[{"2", " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]], 
                "-", 
                FractionBox[
                 RowBox[{"2", " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]], 
                "-", 
                FractionBox[
                 RowBox[{"2", " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "4"]]}]}]],
           Annotation[#, {1, 2} -> (((-2) $CellContext`\[Upsilon][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^(-4)) 
               Derivative[0, 1][$CellContext`\[Alpha]][
                 $CellContext`x1[], 
                 $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Alpha]][
                $CellContext`x1[], 
                $CellContext`x2[]] - ((2 $CellContext`\[Gamma][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-4)) 
              Derivative[0, 1][$CellContext`\[Beta]][
                $CellContext`x1[], 
                $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Beta]][
               $CellContext`x1[], 
               $CellContext`x2[]] - ((2 $CellContext`\[Gamma][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2)) 
              Derivative[0, 1][$CellContext`\[Gamma]][
                $CellContext`x1[], 
                $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Gamma]][
               $CellContext`x1[], 
               $CellContext`x2[]] - ((2 $CellContext`\[Upsilon][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2)) 
              Derivative[0, 1][$CellContext`\[Upsilon]][
                $CellContext`x1[], 
                $CellContext`x2[]]) 
             Derivative[1, 0][$CellContext`\[Upsilon]][
               $CellContext`x1[], 
               $CellContext`x2[]] - (2 $CellContext`\[Eta][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-4)) (($CellContext`\[Eta][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2 
                Derivative[0, 1][$CellContext`\[Eta]][
                  $CellContext`x1[], 
                  $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Eta]][
                 $CellContext`x1[], 
                 $CellContext`x2[]] + Derivative[0, 1][$CellContext`\[Chi]][
                 $CellContext`x1[], 
                 $CellContext`x2[]] Derivative[1, 0][$CellContext`\[Chi]][
                 $CellContext`x1[], 
                 $CellContext`x2[]]), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 4} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "5"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 5} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "6"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 6} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "7"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 7} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "8"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 8} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{"-", 
                  FractionBox[
                   RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "4"]]}], 
                "-", 
                FractionBox[
                 RowBox[{"2", " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "4"]], 
                "-", 
                FractionBox[
                 RowBox[{"2", " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]], 
                "-", 
                FractionBox[
                 RowBox[{"2", " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]], 
                "-", 
                FractionBox[
                 RowBox[{"2", " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}]}], 
                    ")"}]}], 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "4"]]}]}]],
           Annotation[#, {2, 1} -> (((-2) $CellContext`\[Upsilon][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^(-4)) 
               Derivative[0, 1][$CellContext`\[Alpha]][
                 $CellContext`x1[], 
                 $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Alpha]][
                $CellContext`x1[], 
                $CellContext`x2[]] - ((2 $CellContext`\[Gamma][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-4)) 
              Derivative[0, 1][$CellContext`\[Beta]][
                $CellContext`x1[], 
                $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Beta]][
               $CellContext`x1[], 
               $CellContext`x2[]] - ((2 $CellContext`\[Gamma][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2)) 
              Derivative[0, 1][$CellContext`\[Gamma]][
                $CellContext`x1[], 
                $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Gamma]][
               $CellContext`x1[], 
               $CellContext`x2[]] - ((2 $CellContext`\[Upsilon][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2)) 
              Derivative[0, 1][$CellContext`\[Upsilon]][
                $CellContext`x1[], 
                $CellContext`x2[]]) 
             Derivative[1, 0][$CellContext`\[Upsilon]][
               $CellContext`x1[], 
               $CellContext`x2[]] - (2 $CellContext`\[Eta][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-4)) (($CellContext`\[Eta][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2 
                Derivative[0, 1][$CellContext`\[Eta]][
                  $CellContext`x1[], 
                  $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Eta]][
                 $CellContext`x1[], 
                 $CellContext`x2[]] + Derivative[0, 1][$CellContext`\[Chi]][
                 $CellContext`x1[], 
                 $CellContext`x2[]] Derivative[1, 0][$CellContext`\[Chi]][
                 $CellContext`x1[], 
                 $CellContext`x2[]]), "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{"-", 
                  FractionBox[
                   RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "4"]]}], 
                "-", 
                FractionBox[
                 RowBox[{"2", " ", 
                   SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "4"]], 
                "+", 
                FractionBox[
                 RowBox[{
                   RowBox[{"-", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                   "+", 
                   SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]], 
                "+", 
                FractionBox[
                 RowBox[{
                   RowBox[{"-", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                   "+", 
                   SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]], 
                "-", 
                FractionBox[
                 RowBox[{
                   RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                   "+", 
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "-", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "4"]], 
                "-", 
                FractionBox[
                 RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]], "-", 
                FractionBox[
                 RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                   RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]]}]}]],
           Annotation[#, {2, 2} -> ((-2) $CellContext`\[Upsilon][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-4)) 
              Derivative[0, 1][$CellContext`\[Alpha]][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^2 - (2 $CellContext`\[Eta][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-4)) 
             Derivative[0, 1][$CellContext`\[Chi]][
                $CellContext`x1[], 
                $CellContext`x2[]]^2 + $CellContext`\[Eta][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-2) (-
                Derivative[0, 1][$CellContext`\[Eta]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2 + 
               Derivative[1, 0][$CellContext`\[Eta]][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^2) + $CellContext`\[Upsilon][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-2) (-
                Derivative[0, 1][$CellContext`\[Upsilon]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2 + 
               Derivative[1, 0][$CellContext`\[Upsilon]][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^2) - $CellContext`\[Gamma][
                $CellContext`x1[], 
                $CellContext`x2[]]^(-4) (
              2 Derivative[0, 1][$CellContext`\[Beta]][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^2 + $CellContext`\[Gamma][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^2 (
                Derivative[0, 1][$CellContext`\[Gamma]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2 - 
                Derivative[1, 0][$CellContext`\[Gamma]][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^2 + $CellContext`\[Gamma][
                   $CellContext`x1[], 
                   $CellContext`x2[]] (Derivative[0, 2][$CellContext`\[Gamma]][
                    $CellContext`x1[], 
                    $CellContext`x2[]] + 
                  Derivative[2, 0][$CellContext`\[Gamma]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]))) - $CellContext`\[Eta][
                $CellContext`x1[], 
                $CellContext`x2[]]^(-1) (Derivative[0, 2][$CellContext`\[Eta]][
                $CellContext`x1[], 
                $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Eta]][
                $CellContext`x1[], 
                $CellContext`x2[]]) - $CellContext`\[Upsilon][
                $CellContext`x1[], 
                $CellContext`x2[]]^(-1) (
              Derivative[0, 2][$CellContext`\[Upsilon]][
                $CellContext`x1[], 
                $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Upsilon]][
                $CellContext`x1[], 
                $CellContext`x2[]]), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 4} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "5"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 5} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "6"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 6} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "7"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 7} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "8"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 8} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 RowBox[{"2", " ", 
                   SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                 "+", 
                 RowBox[{"2", " ", 
                   SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                 "-", 
                 RowBox[{
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                 RowBox[{
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "3"], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
               RowBox[{
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "8"]}]]}]],
           Annotation[#, {3, 3} -> (($CellContext`\[Gamma][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2) $CellContext`\[Eta][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-8)) (
              2 Derivative[0, 1][$CellContext`\[Alpha]][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^2 + 
              2 Derivative[1, 0][$CellContext`\[Alpha]][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^2 - $CellContext`\[Upsilon][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^2 (
               Derivative[0, 1][$CellContext`\[Upsilon]][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^2 + 
               Derivative[1, 0][$CellContext`\[Upsilon]][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^2) + $CellContext`\[Upsilon][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^3 (
                Derivative[0, 2][$CellContext`\[Upsilon]][
                  $CellContext`x1[], 
                  $CellContext`x2[]] + 
                Derivative[2, 0][$CellContext`\[Upsilon]][
                  $CellContext`x1[], 
                  $CellContext`x2[]])), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 4} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "5"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 5} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "6"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 6} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "7"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 7} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "8"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 RowBox[{
                   RowBox[{"-", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "3"]}], 
                   " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "4"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                 RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    "-", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "3"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], 
               RowBox[{
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "8"]}]]}]],
           Annotation[#, {3, 8} -> (($CellContext`\[Gamma][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2) $CellContext`\[Eta][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-8)) ((-$CellContext`\[Upsilon][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^3) (((-4) 
                  Derivative[0, 1][$CellContext`\[Alpha]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]) 
                 Derivative[0, 1][$CellContext`\[Upsilon]][
                   $CellContext`x1[], 
                   $CellContext`x2[]] - (4 
                 Derivative[1, 0][$CellContext`\[Alpha]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]) 
                Derivative[1, 0][$CellContext`\[Upsilon]][
                  $CellContext`x1[], 
                  $CellContext`x2[]] + $CellContext`\[Upsilon][
                   $CellContext`x1[], 
                   $CellContext`x2[]] (Derivative[0, 2][$CellContext`\[Alpha]][
                    $CellContext`x1[], 
                    $CellContext`x2[]] + 
                  Derivative[2, 0][$CellContext`\[Alpha]][
                    $CellContext`x1[], 
                    $CellContext`x2[]])) + (2 $CellContext`\[Alpha][
                  $CellContext`x1[], 
                  $CellContext`x2[]]) (
                2 Derivative[0, 1][$CellContext`\[Alpha]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^2 + 
                2 Derivative[1, 0][$CellContext`\[Alpha]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^2 - $CellContext`\[Upsilon][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2 (
                 Derivative[0, 1][$CellContext`\[Upsilon]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^2 + 
                 Derivative[1, 0][$CellContext`\[Upsilon]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^2) + $CellContext`\[Upsilon][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^3 (
                  Derivative[0, 2][$CellContext`\[Upsilon]][
                    $CellContext`x1[], 
                    $CellContext`x2[]] + 
                  Derivative[2, 0][$CellContext`\[Upsilon]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]))), "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "4"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 RowBox[{"2", " ", 
                   SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                 "+", 
                 RowBox[{"2", " ", 
                   SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                 "-", 
                 RowBox[{
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                 RowBox[{
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "3"], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
               RowBox[{
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "8"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}]]}]],
           Annotation[#, {4, 4} -> (($CellContext`\[Gamma][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-8) $CellContext`\[Eta][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-2)) (
              2 Derivative[0, 1][$CellContext`\[Beta]][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^2 + 
              2 Derivative[1, 0][$CellContext`\[Beta]][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^2 - $CellContext`\[Gamma][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^2 (Derivative[0, 1][$CellContext`\[Gamma]][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^2 + 
               Derivative[1, 0][$CellContext`\[Gamma]][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^2) + $CellContext`\[Gamma][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^3 (
                Derivative[0, 2][$CellContext`\[Gamma]][
                  $CellContext`x1[], 
                  $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Gamma]][
                  $CellContext`x1[], 
                  $CellContext`x2[]])), "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "5"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 RowBox[{
                   RowBox[{"-", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "3"]}], 
                   " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "4"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                 RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    "-", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "3"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], 
               RowBox[{
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "8"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}]]}]],
           Annotation[#, {4, 5} -> (($CellContext`\[Gamma][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-8) $CellContext`\[Eta][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-2)) ((-$CellContext`\[Gamma][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^3) (((-4) 
                  Derivative[0, 1][$CellContext`\[Beta]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]) 
                 Derivative[0, 1][$CellContext`\[Gamma]][
                   $CellContext`x1[], 
                   $CellContext`x2[]] - (4 
                 Derivative[1, 0][$CellContext`\[Beta]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Gamma]][
                  $CellContext`x1[], 
                  $CellContext`x2[]] + $CellContext`\[Gamma][
                   $CellContext`x1[], 
                   $CellContext`x2[]] (
                  Derivative[0, 2][$CellContext`\[Beta]][
                    $CellContext`x1[], 
                    $CellContext`x2[]] + 
                  Derivative[2, 0][$CellContext`\[Beta]][
                    $CellContext`x1[], 
                    $CellContext`x2[]])) + (2 $CellContext`\[Beta][
                  $CellContext`x1[], 
                  $CellContext`x2[]]) (
                2 Derivative[0, 1][$CellContext`\[Beta]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^2 + 
                2 Derivative[1, 0][$CellContext`\[Beta]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^2 - $CellContext`\[Gamma][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2 (
                 Derivative[0, 1][$CellContext`\[Gamma]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^2 + 
                 Derivative[1, 0][$CellContext`\[Gamma]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^2) + $CellContext`\[Gamma][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^3 (
                  Derivative[0, 2][$CellContext`\[Gamma]][
                    $CellContext`x1[], 
                    $CellContext`x2[]] + 
                  Derivative[2, 0][$CellContext`\[Gamma]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]))), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "6"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 6} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "7"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 7} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "8"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 8} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {5, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {5, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {5, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "4"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 RowBox[{
                   RowBox[{"-", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "3"]}], 
                   " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "4"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                 RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    "-", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "3"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], 
               RowBox[{
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "8"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}]]}]],
           Annotation[#, {5, 4} -> (($CellContext`\[Gamma][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-8) $CellContext`\[Eta][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-2)) ((-$CellContext`\[Gamma][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^3) (((-4) 
                  Derivative[0, 1][$CellContext`\[Beta]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]) 
                 Derivative[0, 1][$CellContext`\[Gamma]][
                   $CellContext`x1[], 
                   $CellContext`x2[]] - (4 
                 Derivative[1, 0][$CellContext`\[Beta]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Gamma]][
                  $CellContext`x1[], 
                  $CellContext`x2[]] + $CellContext`\[Gamma][
                   $CellContext`x1[], 
                   $CellContext`x2[]] (
                  Derivative[0, 2][$CellContext`\[Beta]][
                    $CellContext`x1[], 
                    $CellContext`x2[]] + 
                  Derivative[2, 0][$CellContext`\[Beta]][
                    $CellContext`x1[], 
                    $CellContext`x2[]])) + (2 $CellContext`\[Beta][
                  $CellContext`x1[], 
                  $CellContext`x2[]]) (
                2 Derivative[0, 1][$CellContext`\[Beta]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^2 + 
                2 Derivative[1, 0][$CellContext`\[Beta]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^2 - $CellContext`\[Gamma][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2 (
                 Derivative[0, 1][$CellContext`\[Gamma]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^2 + 
                 Derivative[1, 0][$CellContext`\[Gamma]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^2) + $CellContext`\[Gamma][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^3 (
                  Derivative[0, 2][$CellContext`\[Gamma]][
                    $CellContext`x1[], 
                    $CellContext`x2[]] + 
                  Derivative[2, 0][$CellContext`\[Gamma]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]))), "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "5"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 RowBox[{
                   RowBox[{"-", "4"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "3"], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "4"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                 RowBox[{"4", " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    "-", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "3"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "-", 
                 RowBox[{
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "4"], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    "-", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["x2", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "3"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], 
               RowBox[{
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "8"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}]]}]],
           Annotation[#, {5, 5} -> (($CellContext`\[Gamma][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-8) $CellContext`\[Eta][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-2)) ((((-4) $CellContext`\[Beta][
                   $CellContext`x1[], 
                   $CellContext`x2[]]) $CellContext`\[Gamma][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^3) (((-4) 
                  Derivative[0, 1][$CellContext`\[Beta]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]) 
                 Derivative[0, 1][$CellContext`\[Gamma]][
                   $CellContext`x1[], 
                   $CellContext`x2[]] - (4 
                 Derivative[1, 0][$CellContext`\[Beta]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Gamma]][
                  $CellContext`x1[], 
                  $CellContext`x2[]] + $CellContext`\[Gamma][
                   $CellContext`x1[], 
                   $CellContext`x2[]] (
                  Derivative[0, 2][$CellContext`\[Beta]][
                    $CellContext`x1[], 
                    $CellContext`x2[]] + 
                  Derivative[2, 0][$CellContext`\[Beta]][
                    $CellContext`x1[], 
                    $CellContext`x2[]])) + (4 $CellContext`\[Beta][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2) (
                2 Derivative[0, 1][$CellContext`\[Beta]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^2 + 
                2 Derivative[1, 0][$CellContext`\[Beta]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^2 - $CellContext`\[Gamma][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2 (
                 Derivative[0, 1][$CellContext`\[Gamma]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^2 + 
                 Derivative[1, 0][$CellContext`\[Gamma]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^2) + $CellContext`\[Gamma][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^3 (
                  Derivative[0, 2][$CellContext`\[Gamma]][
                    $CellContext`x1[], 
                    $CellContext`x2[]] + 
                  Derivative[2, 0][$CellContext`\[Gamma]][
                    $CellContext`x1[], 
                    $CellContext`x2[]])) - $CellContext`\[Gamma][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^4 (
               2 Derivative[0, 1][$CellContext`\[Beta]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2 + 
               2 Derivative[1, 0][$CellContext`\[Beta]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2 - $CellContext`\[Gamma][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^2 (
                Derivative[0, 1][$CellContext`\[Gamma]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2 + 
                Derivative[1, 0][$CellContext`\[Gamma]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2) + $CellContext`\[Gamma][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^3 (
                 Derivative[0, 2][$CellContext`\[Gamma]][
                   $CellContext`x1[], 
                   $CellContext`x2[]] + 
                 Derivative[2, 0][$CellContext`\[Gamma]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]))), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "6"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {5, 6} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "7"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {5, 7} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "8"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {5, 8} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"6", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {6, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"6", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {6, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"6", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {6, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"6", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {6, 4} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"6", ",", "5"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {6, 5} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"6", ",", "6"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 RowBox[{
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "6"], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "-", 
                 RowBox[{"4", " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                 RowBox[{"8", " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "-", 
                 RowBox[{
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "7"], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], ")"}]}],
                  "+", 
                 RowBox[{"4", " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "3"], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "-", 
                 RowBox[{"2", " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "4"], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], 
               RowBox[{
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "8"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}]]}]],
           Annotation[#, {6, 6} -> (($CellContext`\[Gamma][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2) $CellContext`\[Eta][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-8)) $CellContext`\[Upsilon][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-2)) ($CellContext`\[Eta][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^6 (
                Derivative[0, 1][$CellContext`\[Eta]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2 + 
                Derivative[1, 0][$CellContext`\[Eta]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2) - ((4 $CellContext`\[Eta][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2) $CellContext`\[Chi][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^2) (Derivative[0, 1][$CellContext`\[Eta]][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Eta]][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^2) + (8 $CellContext`\[Chi][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2) (
                Derivative[0, 1][$CellContext`\[Chi]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2 + 
                Derivative[1, 0][$CellContext`\[Chi]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2) - $CellContext`\[Eta][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^7 (Derivative[0, 2][$CellContext`\[Eta]][
                 $CellContext`x1[], 
                 $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Eta]][
                 $CellContext`x1[], 
                 $CellContext`x2[]]) + ((4 $CellContext`\[Eta][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^3) $CellContext`\[Chi][
                  $CellContext`x1[], 
                  $CellContext`x2[]]) ((4 
                  Derivative[0, 1][$CellContext`\[Eta]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]) 
                 Derivative[0, 1][$CellContext`\[Chi]][
                   $CellContext`x1[], 
                   $CellContext`x2[]] + (4 
                  Derivative[1, 0][$CellContext`\[Eta]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]) 
                 Derivative[1, 0][$CellContext`\[Chi]][
                   $CellContext`x1[], 
                   $CellContext`x2[]] + $CellContext`\[Chi][
                   $CellContext`x1[], 
                   $CellContext`x2[]] (Derivative[0, 2][$CellContext`\[Eta]][
                    $CellContext`x1[], 
                    $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Eta]][
                    $CellContext`x1[], 
                    $CellContext`x2[]])) - (2 $CellContext`\[Eta][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^4) (Derivative[0, 1][$CellContext`\[Chi]][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Chi]][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^2 + (2 $CellContext`\[Chi][
                   $CellContext`x1[], 
                   $CellContext`x2[]]) (
                 Derivative[0, 2][$CellContext`\[Chi]][
                   $CellContext`x1[], 
                   $CellContext`x2[]] + 
                 Derivative[2, 0][$CellContext`\[Chi]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]))), "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"6", ",", "7"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 RowBox[{
                   RowBox[{"-", "2"}], " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                 RowBox[{"4", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                 RowBox[{"2", " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "3"], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "-", 
                 RowBox[{
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "4"], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
               RowBox[{
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "8"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}]]}]],
           Annotation[#, {6, 7} -> (($CellContext`\[Gamma][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2) $CellContext`\[Eta][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-8)) $CellContext`\[Upsilon][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-2)) ((((-2) $CellContext`\[Eta][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^2) $CellContext`\[Chi][
                  $CellContext`x1[], 
                  $CellContext`x2[]]) (Derivative[0, 1][$CellContext`\[Eta]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2 + 
                Derivative[1, 0][$CellContext`\[Eta]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2) + (4 $CellContext`\[Chi][
                  $CellContext`x1[], 
                  $CellContext`x2[]]) (Derivative[0, 1][$CellContext`\[Chi]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2 + 
                Derivative[1, 0][$CellContext`\[Chi]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2) + (2 $CellContext`\[Eta][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^3) ((2 
                  Derivative[0, 1][$CellContext`\[Eta]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]) 
                 Derivative[0, 1][$CellContext`\[Chi]][
                   $CellContext`x1[], 
                   $CellContext`x2[]] + (2 
                  Derivative[1, 0][$CellContext`\[Eta]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]) 
                 Derivative[1, 0][$CellContext`\[Chi]][
                   $CellContext`x1[], 
                   $CellContext`x2[]] + $CellContext`\[Chi][
                   $CellContext`x1[], 
                   $CellContext`x2[]] (Derivative[0, 2][$CellContext`\[Eta]][
                    $CellContext`x1[], 
                    $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Eta]][
                    $CellContext`x1[], 
                    $CellContext`x2[]])) - $CellContext`\[Eta][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^4 (Derivative[0, 2][$CellContext`\[Chi]][
                 $CellContext`x1[], 
                 $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Chi]][
                 $CellContext`x1[], 
                 $CellContext`x2[]])), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"6", ",", "8"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {6, 8} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"7", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {7, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"7", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {7, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"7", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {7, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"7", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {7, 4} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"7", ",", "5"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {7, 5} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"7", ",", "6"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 RowBox[{
                   RowBox[{"-", "2"}], " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                 RowBox[{"4", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                 RowBox[{"2", " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "3"], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "-", 
                 RowBox[{
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "4"], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
               RowBox[{
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "8"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}]]}]],
           Annotation[#, {7, 6} -> (($CellContext`\[Gamma][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2) $CellContext`\[Eta][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-8)) $CellContext`\[Upsilon][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-2)) ((((-2) $CellContext`\[Eta][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^2) $CellContext`\[Chi][
                  $CellContext`x1[], 
                  $CellContext`x2[]]) (Derivative[0, 1][$CellContext`\[Eta]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2 + 
                Derivative[1, 0][$CellContext`\[Eta]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2) + (4 $CellContext`\[Chi][
                  $CellContext`x1[], 
                  $CellContext`x2[]]) (Derivative[0, 1][$CellContext`\[Chi]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2 + 
                Derivative[1, 0][$CellContext`\[Chi]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2) + (2 $CellContext`\[Eta][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^3) ((2 
                  Derivative[0, 1][$CellContext`\[Eta]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]) 
                 Derivative[0, 1][$CellContext`\[Chi]][
                   $CellContext`x1[], 
                   $CellContext`x2[]] + (2 
                  Derivative[1, 0][$CellContext`\[Eta]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]) 
                 Derivative[1, 0][$CellContext`\[Chi]][
                   $CellContext`x1[], 
                   $CellContext`x2[]] + $CellContext`\[Chi][
                   $CellContext`x1[], 
                   $CellContext`x2[]] (Derivative[0, 2][$CellContext`\[Eta]][
                    $CellContext`x1[], 
                    $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Eta]][
                    $CellContext`x1[], 
                    $CellContext`x2[]])) - $CellContext`\[Eta][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^4 (Derivative[0, 2][$CellContext`\[Chi]][
                 $CellContext`x1[], 
                 $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Chi]][
                 $CellContext`x1[], 
                 $CellContext`x2[]])), "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"7", ",", "7"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 RowBox[{
                   RowBox[{"-", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                   " ", 
                   RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                 RowBox[{"2", " ", 
                   RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                 RowBox[{
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "3"], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["x2", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], 
               RowBox[{
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "8"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}]]}]],
           Annotation[#, {7, 7} -> (($CellContext`\[Gamma][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2) $CellContext`\[Eta][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-8)) $CellContext`\[Upsilon][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-2)) ((-$CellContext`\[Eta][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2) (
                Derivative[0, 1][$CellContext`\[Eta]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2 + 
                Derivative[1, 0][$CellContext`\[Eta]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2) + 
              2 (Derivative[0, 1][$CellContext`\[Chi]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2 + 
                Derivative[1, 0][$CellContext`\[Chi]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2) + $CellContext`\[Eta][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^3 (
                Derivative[0, 2][$CellContext`\[Eta]][
                  $CellContext`x1[], 
                  $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Eta]][
                  $CellContext`x1[], 
                  $CellContext`x2[]])), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"7", ",", "8"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {7, 8} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"8", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {8, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"8", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {8, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"8", ",", "3"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 RowBox[{
                   RowBox[{"-", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "3"]}], 
                   " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "4"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                 RowBox[{"2", " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    "-", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "3"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], 
               RowBox[{
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "8"]}]]}]],
           Annotation[#, {8, 3} -> (($CellContext`\[Gamma][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2) $CellContext`\[Eta][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-8)) ((-$CellContext`\[Upsilon][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^3) (((-4) 
                  Derivative[0, 1][$CellContext`\[Alpha]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]) 
                 Derivative[0, 1][$CellContext`\[Upsilon]][
                   $CellContext`x1[], 
                   $CellContext`x2[]] - (4 
                 Derivative[1, 0][$CellContext`\[Alpha]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]) 
                Derivative[1, 0][$CellContext`\[Upsilon]][
                  $CellContext`x1[], 
                  $CellContext`x2[]] + $CellContext`\[Upsilon][
                   $CellContext`x1[], 
                   $CellContext`x2[]] (Derivative[0, 2][$CellContext`\[Alpha]][
                    $CellContext`x1[], 
                    $CellContext`x2[]] + 
                  Derivative[2, 0][$CellContext`\[Alpha]][
                    $CellContext`x1[], 
                    $CellContext`x2[]])) + (2 $CellContext`\[Alpha][
                  $CellContext`x1[], 
                  $CellContext`x2[]]) (
                2 Derivative[0, 1][$CellContext`\[Alpha]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^2 + 
                2 Derivative[1, 0][$CellContext`\[Alpha]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^2 - $CellContext`\[Upsilon][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2 (
                 Derivative[0, 1][$CellContext`\[Upsilon]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^2 + 
                 Derivative[1, 0][$CellContext`\[Upsilon]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^2) + $CellContext`\[Upsilon][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^3 (
                  Derivative[0, 2][$CellContext`\[Upsilon]][
                    $CellContext`x1[], 
                    $CellContext`x2[]] + 
                  Derivative[2, 0][$CellContext`\[Upsilon]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]))), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"8", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {8, 4} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"8", ",", "5"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {8, 5} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"8", ",", "6"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {8, 6} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"8", ",", "7"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {8, 7} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["\<\"\[GrayCircle]\"\>",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"8", ",", "8"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 RowBox[{
                   RowBox[{"-", "4"}], " ", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "3"], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "4"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], "-", 
                    RowBox[{"4", " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "+", 
                 RowBox[{"4", " ", 
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    "-", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "3"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox["x2", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}], "-", 
                 RowBox[{
                   SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "4"], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    "-", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                    ")"}]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "3"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}], 
                    ")"}]}]}], ")"}]}]}], 
               RowBox[{
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "8"]}]]}]],
           Annotation[#, {8, 8} -> (($CellContext`\[Gamma][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2) $CellContext`\[Eta][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-8)) ((((-4) $CellContext`\[Alpha][
                   $CellContext`x1[], 
                   $CellContext`x2[]]) $CellContext`\[Upsilon][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^3) (((-4) 
                  Derivative[0, 1][$CellContext`\[Alpha]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]) 
                 Derivative[0, 1][$CellContext`\[Upsilon]][
                   $CellContext`x1[], 
                   $CellContext`x2[]] - (4 
                 Derivative[1, 0][$CellContext`\[Alpha]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]) 
                Derivative[1, 0][$CellContext`\[Upsilon]][
                  $CellContext`x1[], 
                  $CellContext`x2[]] + $CellContext`\[Upsilon][
                   $CellContext`x1[], 
                   $CellContext`x2[]] (Derivative[0, 2][$CellContext`\[Alpha]][
                    $CellContext`x1[], 
                    $CellContext`x2[]] + 
                  Derivative[2, 0][$CellContext`\[Alpha]][
                    $CellContext`x1[], 
                    $CellContext`x2[]])) + (4 $CellContext`\[Alpha][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2) (
                2 Derivative[0, 1][$CellContext`\[Alpha]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^2 + 
                2 Derivative[1, 0][$CellContext`\[Alpha]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^2 - $CellContext`\[Upsilon][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2 (
                 Derivative[0, 1][$CellContext`\[Upsilon]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^2 + 
                 Derivative[1, 0][$CellContext`\[Upsilon]][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^2) + $CellContext`\[Upsilon][
                    $CellContext`x1[], 
                    $CellContext`x2[]]^3 (
                  Derivative[0, 2][$CellContext`\[Upsilon]][
                    $CellContext`x1[], 
                    $CellContext`x2[]] + 
                  Derivative[2, 0][$CellContext`\[Upsilon]][
                    $CellContext`x1[], 
                    $CellContext`x2[]])) - $CellContext`\[Upsilon][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^4 (
               2 Derivative[0, 1][$CellContext`\[Alpha]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2 + 
               2 Derivative[1, 0][$CellContext`\[Alpha]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2 - $CellContext`\[Upsilon][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^2 (
                Derivative[0, 1][$CellContext`\[Upsilon]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2 + 
                Derivative[1, 0][$CellContext`\[Upsilon]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^2) + $CellContext`\[Upsilon][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^3 (
                 Derivative[0, 2][$CellContext`\[Upsilon]][
                   $CellContext`x1[], 
                   $CellContext`x2[]] + 
                 Derivative[2, 0][$CellContext`\[Upsilon]][
                   $CellContext`x1[], 
                   $CellContext`x2[]]))), "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 9 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 9 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " "},
         {"a", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{((-2) $CellContext`\[Gamma][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-4)) Derivative[1, 0][$CellContext`\[Beta]][
         $CellContext`x1[], 
         $CellContext`x2[]]^2 + $CellContext`\[Gamma][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-2) (Derivative[0, 1][$CellContext`\[Gamma]][
          $CellContext`x1[], 
          $CellContext`x2[]]^2 - Derivative[1, 0][$CellContext`\[Gamma]][
         $CellContext`x1[], 
         $CellContext`x2[]]^2) + $CellContext`\[Eta][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-2) (Derivative[0, 1][$CellContext`\[Eta]][
          $CellContext`x1[], 
          $CellContext`x2[]]^2 - Derivative[1, 0][$CellContext`\[Eta]][
         $CellContext`x1[], 
         $CellContext`x2[]]^2) - (2 $CellContext`\[Eta][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-4)) Derivative[1, 0][$CellContext`\[Chi]][
        $CellContext`x1[], 
        $CellContext`x2[]]^2 - $CellContext`\[Gamma][
        $CellContext`x1[], 
        $CellContext`x2[]]^(-1) (Derivative[0, 2][$CellContext`\[Gamma]][
        $CellContext`x1[], 
        $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Gamma]][
        $CellContext`x1[], 
        $CellContext`x2[]]) - $CellContext`\[Eta][
        $CellContext`x1[], 
        $CellContext`x2[]]^(-1) (Derivative[0, 2][$CellContext`\[Eta]][
        $CellContext`x1[], 
        $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Eta]][
        $CellContext`x1[], 
        $CellContext`x2[]]) - $CellContext`\[Upsilon][
        $CellContext`x1[], 
        $CellContext`x2[]]^(-4) (2 Derivative[1, 0][$CellContext`\[Alpha]][
          $CellContext`x1[], 
          $CellContext`x2[]]^2 + $CellContext`\[Upsilon][
          $CellContext`x1[], 
          $CellContext`x2[]]^2 (-Derivative[0, 1][$CellContext`\[Upsilon]][
            $CellContext`x1[], 
            $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Upsilon]][
           $CellContext`x1[], 
           $CellContext`x2[]]^2) + $CellContext`\[Upsilon][
          $CellContext`x1[], 
          $CellContext`x2[]]^3 (Derivative[0, 2][$CellContext`\[Upsilon]][
          $CellContext`x1[], 
          $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Upsilon]][
          $CellContext`x1[], 
          $CellContext`x2[]])), (((-2) $CellContext`\[Upsilon][
           $CellContext`x1[], 
           $CellContext`x2[]]^(-4)) Derivative[0, 1][$CellContext`\[Alpha]][
         $CellContext`x1[], 
         $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Alpha]][
        $CellContext`x1[], 
        $CellContext`x2[]] - ((2 $CellContext`\[Gamma][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-4)) Derivative[0, 1][$CellContext`\[Beta]][
        $CellContext`x1[], 
        $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Beta]][
       $CellContext`x1[], 
       $CellContext`x2[]] - ((2 $CellContext`\[Gamma][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2)) Derivative[0, 1][$CellContext`\[Gamma]][
        $CellContext`x1[], 
        $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Gamma]][
       $CellContext`x1[], 
       $CellContext`x2[]] - ((2 $CellContext`\[Upsilon][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2)) Derivative[0, 1][$CellContext`\[Upsilon]][
        $CellContext`x1[], 
        $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Upsilon]][
       $CellContext`x1[], 
       $CellContext`x2[]] - (2 $CellContext`\[Eta][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-4)) (($CellContext`\[Eta][
           $CellContext`x1[], 
           $CellContext`x2[]]^2 Derivative[0, 1][$CellContext`\[Eta]][
          $CellContext`x1[], 
          $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Eta]][
         $CellContext`x1[], 
         $CellContext`x2[]] + Derivative[0, 1][$CellContext`\[Chi]][
         $CellContext`x1[], 
         $CellContext`x2[]] Derivative[1, 0][$CellContext`\[Chi]][
         $CellContext`x1[], 
         $CellContext`x2[]]), 0, 0, 0, 0, 0, 
     0}, {(((-2) $CellContext`\[Upsilon][
           $CellContext`x1[], 
           $CellContext`x2[]]^(-4)) Derivative[0, 1][$CellContext`\[Alpha]][
         $CellContext`x1[], 
         $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Alpha]][
        $CellContext`x1[], 
        $CellContext`x2[]] - ((2 $CellContext`\[Gamma][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-4)) Derivative[0, 1][$CellContext`\[Beta]][
        $CellContext`x1[], 
        $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Beta]][
       $CellContext`x1[], 
       $CellContext`x2[]] - ((2 $CellContext`\[Gamma][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2)) Derivative[0, 1][$CellContext`\[Gamma]][
        $CellContext`x1[], 
        $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Gamma]][
       $CellContext`x1[], 
       $CellContext`x2[]] - ((2 $CellContext`\[Upsilon][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2)) Derivative[0, 1][$CellContext`\[Upsilon]][
        $CellContext`x1[], 
        $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Upsilon]][
       $CellContext`x1[], 
       $CellContext`x2[]] - (2 $CellContext`\[Eta][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-4)) (($CellContext`\[Eta][
           $CellContext`x1[], 
           $CellContext`x2[]]^2 Derivative[0, 1][$CellContext`\[Eta]][
          $CellContext`x1[], 
          $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Eta]][
         $CellContext`x1[], 
         $CellContext`x2[]] + Derivative[0, 1][$CellContext`\[Chi]][
         $CellContext`x1[], 
         $CellContext`x2[]] Derivative[1, 0][$CellContext`\[Chi]][
         $CellContext`x1[], 
         $CellContext`x2[]]), ((-2) $CellContext`\[Upsilon][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-4)) Derivative[0, 1][$CellContext`\[Alpha]][
         $CellContext`x1[], 
         $CellContext`x2[]]^2 - (2 $CellContext`\[Eta][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-4)) Derivative[0, 1][$CellContext`\[Chi]][
        $CellContext`x1[], 
        $CellContext`x2[]]^2 + $CellContext`\[Eta][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-2) (-Derivative[0, 1][$CellContext`\[Eta]][
           $CellContext`x1[], 
           $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Eta]][
          $CellContext`x1[], 
          $CellContext`x2[]]^2) + $CellContext`\[Upsilon][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-2) (-Derivative[0, 1][$CellContext`\[Upsilon]][
           $CellContext`x1[], 
           $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Upsilon]][
          $CellContext`x1[], 
          $CellContext`x2[]]^2) - $CellContext`\[Gamma][
        $CellContext`x1[], 
        $CellContext`x2[]]^(-4) (2 Derivative[0, 1][$CellContext`\[Beta]][
          $CellContext`x1[], 
          $CellContext`x2[]]^2 + $CellContext`\[Gamma][
          $CellContext`x1[], 
          $CellContext`x2[]]^2 (Derivative[0, 1][$CellContext`\[Gamma]][
           $CellContext`x1[], 
           $CellContext`x2[]]^2 - Derivative[1, 0][$CellContext`\[Gamma]][
          $CellContext`x1[], 
          $CellContext`x2[]]^2 + $CellContext`\[Gamma][
           $CellContext`x1[], 
           $CellContext`x2[]] (Derivative[0, 2][$CellContext`\[Gamma]][
            $CellContext`x1[], 
            $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Gamma]][
            $CellContext`x1[], 
            $CellContext`x2[]]))) - $CellContext`\[Eta][
        $CellContext`x1[], 
        $CellContext`x2[]]^(-1) (Derivative[0, 2][$CellContext`\[Eta]][
        $CellContext`x1[], 
        $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Eta]][
        $CellContext`x1[], 
        $CellContext`x2[]]) - $CellContext`\[Upsilon][
        $CellContext`x1[], 
        $CellContext`x2[]]^(-1) (Derivative[0, 2][$CellContext`\[Upsilon]][
        $CellContext`x1[], 
        $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Upsilon]][
        $CellContext`x1[], 
        $CellContext`x2[]]), 0, 0, 0, 0, 0, 0}, {0, 0, (($CellContext`\[Gamma][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2) $CellContext`\[Eta][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-8)) (2 Derivative[0, 1][$CellContext`\[Alpha]][
          $CellContext`x1[], 
          $CellContext`x2[]]^2 + 2 Derivative[1, 0][$CellContext`\[Alpha]][
          $CellContext`x1[], 
          $CellContext`x2[]]^2 - $CellContext`\[Upsilon][
         $CellContext`x1[], 
         $CellContext`x2[]]^2 (Derivative[0, 1][$CellContext`\[Upsilon]][
          $CellContext`x1[], 
          $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Upsilon]][
          $CellContext`x1[], 
          $CellContext`x2[]]^2) + $CellContext`\[Upsilon][
          $CellContext`x1[], 
          $CellContext`x2[]]^3 (Derivative[0, 2][$CellContext`\[Upsilon]][
          $CellContext`x1[], 
          $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Upsilon]][
          $CellContext`x1[], 
          $CellContext`x2[]])), 0, 0, 0, 0, (($CellContext`\[Gamma][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2) $CellContext`\[Eta][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-8)) ((-$CellContext`\[Upsilon][
           $CellContext`x1[], 
           $CellContext`x2[]]^3) (((-4) 
          Derivative[0, 1][$CellContext`\[Alpha]][
            $CellContext`x1[], 
            $CellContext`x2[]]) Derivative[0, 1][$CellContext`\[Upsilon]][
           $CellContext`x1[], 
           $CellContext`x2[]] - (4 Derivative[1, 0][$CellContext`\[Alpha]][
           $CellContext`x1[], 
           $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Upsilon]][
          $CellContext`x1[], 
          $CellContext`x2[]] + $CellContext`\[Upsilon][
           $CellContext`x1[], 
           $CellContext`x2[]] (Derivative[0, 2][$CellContext`\[Alpha]][
            $CellContext`x1[], 
            $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Alpha]][
            $CellContext`x1[], 
            $CellContext`x2[]])) + (2 $CellContext`\[Alpha][
          $CellContext`x1[], 
          $CellContext`x2[]]) (2 Derivative[0, 1][$CellContext`\[Alpha]][
            $CellContext`x1[], 
            $CellContext`x2[]]^2 + 2 Derivative[1, 0][$CellContext`\[Alpha]][
            $CellContext`x1[], 
            $CellContext`x2[]]^2 - $CellContext`\[Upsilon][
           $CellContext`x1[], 
           $CellContext`x2[]]^2 (Derivative[0, 1][$CellContext`\[Upsilon]][
            $CellContext`x1[], 
            $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Upsilon]][
            $CellContext`x1[], 
            $CellContext`x2[]]^2) + $CellContext`\[Upsilon][
            $CellContext`x1[], 
            $CellContext`x2[]]^3 (Derivative[0, 2][$CellContext`\[Upsilon]][
            $CellContext`x1[], 
            $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Upsilon]][
            $CellContext`x1[], 
            $CellContext`x2[]])))}, {0, 0, 0, (($CellContext`\[Gamma][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-8) $CellContext`\[Eta][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-2)) (2 Derivative[0, 1][$CellContext`\[Beta]][
          $CellContext`x1[], 
          $CellContext`x2[]]^2 + 2 Derivative[1, 0][$CellContext`\[Beta]][
          $CellContext`x1[], 
          $CellContext`x2[]]^2 - $CellContext`\[Gamma][
         $CellContext`x1[], 
         $CellContext`x2[]]^2 (Derivative[0, 1][$CellContext`\[Gamma]][
          $CellContext`x1[], 
          $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Gamma]][
          $CellContext`x1[], 
          $CellContext`x2[]]^2) + $CellContext`\[Gamma][
          $CellContext`x1[], 
          $CellContext`x2[]]^3 (Derivative[0, 2][$CellContext`\[Gamma]][
          $CellContext`x1[], 
          $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Gamma]][
          $CellContext`x1[], 
          $CellContext`x2[]])), (($CellContext`\[Gamma][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-8) $CellContext`\[Eta][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-2)) ((-$CellContext`\[Gamma][
           $CellContext`x1[], 
           $CellContext`x2[]]^3) (((-4) Derivative[0, 1][$CellContext`\[Beta]][
            $CellContext`x1[], 
            $CellContext`x2[]]) Derivative[0, 1][$CellContext`\[Gamma]][
           $CellContext`x1[], 
           $CellContext`x2[]] - (4 Derivative[1, 0][$CellContext`\[Beta]][
           $CellContext`x1[], 
           $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Gamma]][
          $CellContext`x1[], 
          $CellContext`x2[]] + $CellContext`\[Gamma][
           $CellContext`x1[], 
           $CellContext`x2[]] (Derivative[0, 2][$CellContext`\[Beta]][
            $CellContext`x1[], 
            $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Beta]][
            $CellContext`x1[], 
            $CellContext`x2[]])) + (2 $CellContext`\[Beta][
          $CellContext`x1[], 
          $CellContext`x2[]]) (2 Derivative[0, 1][$CellContext`\[Beta]][
            $CellContext`x1[], 
            $CellContext`x2[]]^2 + 2 Derivative[1, 0][$CellContext`\[Beta]][
            $CellContext`x1[], 
            $CellContext`x2[]]^2 - $CellContext`\[Gamma][
           $CellContext`x1[], 
           $CellContext`x2[]]^2 (Derivative[0, 1][$CellContext`\[Gamma]][
            $CellContext`x1[], 
            $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Gamma]][
            $CellContext`x1[], 
            $CellContext`x2[]]^2) + $CellContext`\[Gamma][
            $CellContext`x1[], 
            $CellContext`x2[]]^3 (Derivative[0, 2][$CellContext`\[Gamma]][
            $CellContext`x1[], 
            $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Gamma]][
            $CellContext`x1[], 
            $CellContext`x2[]]))), 0, 0, 0}, {0, 0, 0, (($CellContext`\[Gamma][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-8) $CellContext`\[Eta][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-2)) ((-$CellContext`\[Gamma][
           $CellContext`x1[], 
           $CellContext`x2[]]^3) (((-4) Derivative[0, 1][$CellContext`\[Beta]][
            $CellContext`x1[], 
            $CellContext`x2[]]) Derivative[0, 1][$CellContext`\[Gamma]][
           $CellContext`x1[], 
           $CellContext`x2[]] - (4 Derivative[1, 0][$CellContext`\[Beta]][
           $CellContext`x1[], 
           $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Gamma]][
          $CellContext`x1[], 
          $CellContext`x2[]] + $CellContext`\[Gamma][
           $CellContext`x1[], 
           $CellContext`x2[]] (Derivative[0, 2][$CellContext`\[Beta]][
            $CellContext`x1[], 
            $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Beta]][
            $CellContext`x1[], 
            $CellContext`x2[]])) + (2 $CellContext`\[Beta][
          $CellContext`x1[], 
          $CellContext`x2[]]) (2 Derivative[0, 1][$CellContext`\[Beta]][
            $CellContext`x1[], 
            $CellContext`x2[]]^2 + 2 Derivative[1, 0][$CellContext`\[Beta]][
            $CellContext`x1[], 
            $CellContext`x2[]]^2 - $CellContext`\[Gamma][
           $CellContext`x1[], 
           $CellContext`x2[]]^2 (Derivative[0, 1][$CellContext`\[Gamma]][
            $CellContext`x1[], 
            $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Gamma]][
            $CellContext`x1[], 
            $CellContext`x2[]]^2) + $CellContext`\[Gamma][
            $CellContext`x1[], 
            $CellContext`x2[]]^3 (Derivative[0, 2][$CellContext`\[Gamma]][
            $CellContext`x1[], 
            $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Gamma]][
            $CellContext`x1[], 
            $CellContext`x2[]]))), (($CellContext`\[Gamma][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-8) $CellContext`\[Eta][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-2)) ((((-4) $CellContext`\[Beta][
           $CellContext`x1[], 
           $CellContext`x2[]]) $CellContext`\[Gamma][
           $CellContext`x1[], 
           $CellContext`x2[]]^3) (((-4) Derivative[0, 1][$CellContext`\[Beta]][
            $CellContext`x1[], 
            $CellContext`x2[]]) Derivative[0, 1][$CellContext`\[Gamma]][
           $CellContext`x1[], 
           $CellContext`x2[]] - (4 Derivative[1, 0][$CellContext`\[Beta]][
           $CellContext`x1[], 
           $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Gamma]][
          $CellContext`x1[], 
          $CellContext`x2[]] + $CellContext`\[Gamma][
           $CellContext`x1[], 
           $CellContext`x2[]] (Derivative[0, 2][$CellContext`\[Beta]][
            $CellContext`x1[], 
            $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Beta]][
            $CellContext`x1[], 
            $CellContext`x2[]])) + (4 $CellContext`\[Beta][
           $CellContext`x1[], 
           $CellContext`x2[]]^2) (2 Derivative[0, 1][$CellContext`\[Beta]][
            $CellContext`x1[], 
            $CellContext`x2[]]^2 + 2 Derivative[1, 0][$CellContext`\[Beta]][
            $CellContext`x1[], 
            $CellContext`x2[]]^2 - $CellContext`\[Gamma][
           $CellContext`x1[], 
           $CellContext`x2[]]^2 (Derivative[0, 1][$CellContext`\[Gamma]][
            $CellContext`x1[], 
            $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Gamma]][
            $CellContext`x1[], 
            $CellContext`x2[]]^2) + $CellContext`\[Gamma][
            $CellContext`x1[], 
            $CellContext`x2[]]^3 (Derivative[0, 2][$CellContext`\[Gamma]][
            $CellContext`x1[], 
            $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Gamma]][
            $CellContext`x1[], 
            $CellContext`x2[]])) - $CellContext`\[Gamma][
         $CellContext`x1[], 
         $CellContext`x2[]]^4 (2 Derivative[0, 1][$CellContext`\[Beta]][
           $CellContext`x1[], 
           $CellContext`x2[]]^2 + 2 Derivative[1, 0][$CellContext`\[Beta]][
           $CellContext`x1[], 
           $CellContext`x2[]]^2 - $CellContext`\[Gamma][
          $CellContext`x1[], 
          $CellContext`x2[]]^2 (Derivative[0, 1][$CellContext`\[Gamma]][
           $CellContext`x1[], 
           $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Gamma]][
           $CellContext`x1[], 
           $CellContext`x2[]]^2) + $CellContext`\[Gamma][
           $CellContext`x1[], 
           $CellContext`x2[]]^3 (Derivative[0, 2][$CellContext`\[Gamma]][
           $CellContext`x1[], 
           $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Gamma]][
           $CellContext`x1[], 
           $CellContext`x2[]]))), 0, 0, 0}, {
    0, 0, 0, 0, 0, (($CellContext`\[Gamma][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2) $CellContext`\[Eta][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-8)) $CellContext`\[Upsilon][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-2)) ($CellContext`\[Eta][
          $CellContext`x1[], 
          $CellContext`x2[]]^6 (Derivative[0, 1][$CellContext`\[Eta]][
           $CellContext`x1[], 
           $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Eta]][
           $CellContext`x1[], 
           $CellContext`x2[]]^2) - ((4 $CellContext`\[Eta][
           $CellContext`x1[], 
           $CellContext`x2[]]^2) $CellContext`\[Chi][
          $CellContext`x1[], 
          $CellContext`x2[]]^2) (Derivative[0, 1][$CellContext`\[Eta]][
          $CellContext`x1[], 
          $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Eta]][
          $CellContext`x1[], 
          $CellContext`x2[]]^2) + (8 $CellContext`\[Chi][
           $CellContext`x1[], 
           $CellContext`x2[]]^2) (Derivative[0, 1][$CellContext`\[Chi]][
           $CellContext`x1[], 
           $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Chi]][
           $CellContext`x1[], 
           $CellContext`x2[]]^2) - $CellContext`\[Eta][
         $CellContext`x1[], 
         $CellContext`x2[]]^7 (Derivative[0, 2][$CellContext`\[Eta]][
         $CellContext`x1[], 
         $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Eta]][
         $CellContext`x1[], 
         $CellContext`x2[]]) + ((4 $CellContext`\[Eta][
            $CellContext`x1[], 
            $CellContext`x2[]]^3) $CellContext`\[Chi][
          $CellContext`x1[], 
          $CellContext`x2[]]) ((4 Derivative[0, 1][$CellContext`\[Eta]][
            $CellContext`x1[], 
            $CellContext`x2[]]) Derivative[0, 1][$CellContext`\[Chi]][
           $CellContext`x1[], 
           $CellContext`x2[]] + (4 Derivative[1, 0][$CellContext`\[Eta]][
            $CellContext`x1[], 
            $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Chi]][
           $CellContext`x1[], 
           $CellContext`x2[]] + $CellContext`\[Chi][
           $CellContext`x1[], 
           $CellContext`x2[]] (Derivative[0, 2][$CellContext`\[Eta]][
            $CellContext`x1[], 
            $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Eta]][
            $CellContext`x1[], 
            $CellContext`x2[]])) - (2 $CellContext`\[Eta][
          $CellContext`x1[], 
          $CellContext`x2[]]^4) (Derivative[0, 1][$CellContext`\[Chi]][
          $CellContext`x1[], 
          $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Chi]][
          $CellContext`x1[], 
          $CellContext`x2[]]^2 + (2 $CellContext`\[Chi][
           $CellContext`x1[], 
           $CellContext`x2[]]) (Derivative[0, 2][$CellContext`\[Chi]][
           $CellContext`x1[], 
           $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Chi]][
           $CellContext`x1[], 
           $CellContext`x2[]]))), (($CellContext`\[Gamma][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2) $CellContext`\[Eta][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-8)) $CellContext`\[Upsilon][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-2)) ((((-2) $CellContext`\[Eta][
            $CellContext`x1[], 
            $CellContext`x2[]]^2) $CellContext`\[Chi][
          $CellContext`x1[], 
          $CellContext`x2[]]) (Derivative[0, 1][$CellContext`\[Eta]][
           $CellContext`x1[], 
           $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Eta]][
           $CellContext`x1[], 
           $CellContext`x2[]]^2) + (4 $CellContext`\[Chi][
          $CellContext`x1[], 
          $CellContext`x2[]]) (Derivative[0, 1][$CellContext`\[Chi]][
           $CellContext`x1[], 
           $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Chi]][
           $CellContext`x1[], 
           $CellContext`x2[]]^2) + (2 $CellContext`\[Eta][
           $CellContext`x1[], 
           $CellContext`x2[]]^3) ((2 Derivative[0, 1][$CellContext`\[Eta]][
            $CellContext`x1[], 
            $CellContext`x2[]]) Derivative[0, 1][$CellContext`\[Chi]][
           $CellContext`x1[], 
           $CellContext`x2[]] + (2 Derivative[1, 0][$CellContext`\[Eta]][
            $CellContext`x1[], 
            $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Chi]][
           $CellContext`x1[], 
           $CellContext`x2[]] + $CellContext`\[Chi][
           $CellContext`x1[], 
           $CellContext`x2[]] (Derivative[0, 2][$CellContext`\[Eta]][
            $CellContext`x1[], 
            $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Eta]][
            $CellContext`x1[], 
            $CellContext`x2[]])) - $CellContext`\[Eta][
         $CellContext`x1[], 
         $CellContext`x2[]]^4 (Derivative[0, 2][$CellContext`\[Chi]][
         $CellContext`x1[], 
         $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Chi]][
         $CellContext`x1[], 
         $CellContext`x2[]])), 0}, {0, 0, 0, 0, 0, (($CellContext`\[Gamma][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2) $CellContext`\[Eta][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-8)) $CellContext`\[Upsilon][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-2)) ((((-2) $CellContext`\[Eta][
            $CellContext`x1[], 
            $CellContext`x2[]]^2) $CellContext`\[Chi][
          $CellContext`x1[], 
          $CellContext`x2[]]) (Derivative[0, 1][$CellContext`\[Eta]][
           $CellContext`x1[], 
           $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Eta]][
           $CellContext`x1[], 
           $CellContext`x2[]]^2) + (4 $CellContext`\[Chi][
          $CellContext`x1[], 
          $CellContext`x2[]]) (Derivative[0, 1][$CellContext`\[Chi]][
           $CellContext`x1[], 
           $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Chi]][
           $CellContext`x1[], 
           $CellContext`x2[]]^2) + (2 $CellContext`\[Eta][
           $CellContext`x1[], 
           $CellContext`x2[]]^3) ((2 Derivative[0, 1][$CellContext`\[Eta]][
            $CellContext`x1[], 
            $CellContext`x2[]]) Derivative[0, 1][$CellContext`\[Chi]][
           $CellContext`x1[], 
           $CellContext`x2[]] + (2 Derivative[1, 0][$CellContext`\[Eta]][
            $CellContext`x1[], 
            $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Chi]][
           $CellContext`x1[], 
           $CellContext`x2[]] + $CellContext`\[Chi][
           $CellContext`x1[], 
           $CellContext`x2[]] (Derivative[0, 2][$CellContext`\[Eta]][
            $CellContext`x1[], 
            $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Eta]][
            $CellContext`x1[], 
            $CellContext`x2[]])) - $CellContext`\[Eta][
         $CellContext`x1[], 
         $CellContext`x2[]]^4 (Derivative[0, 2][$CellContext`\[Chi]][
         $CellContext`x1[], 
         $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Chi]][
         $CellContext`x1[], 
         $CellContext`x2[]])), (($CellContext`\[Gamma][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2) $CellContext`\[Eta][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-8)) $CellContext`\[Upsilon][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-2)) ((-$CellContext`\[Eta][
           $CellContext`x1[], 
           $CellContext`x2[]]^2) (Derivative[0, 1][$CellContext`\[Eta]][
           $CellContext`x1[], 
           $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Eta]][
           $CellContext`x1[], 
           $CellContext`x2[]]^2) + 2 (Derivative[0, 1][$CellContext`\[Chi]][
           $CellContext`x1[], 
           $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Chi]][
           $CellContext`x1[], 
           $CellContext`x2[]]^2) + $CellContext`\[Eta][
          $CellContext`x1[], 
          $CellContext`x2[]]^3 (Derivative[0, 2][$CellContext`\[Eta]][
          $CellContext`x1[], 
          $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Eta]][
          $CellContext`x1[], 
          $CellContext`x2[]])), 0}, {0, 0, (($CellContext`\[Gamma][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2) $CellContext`\[Eta][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-8)) ((-$CellContext`\[Upsilon][
           $CellContext`x1[], 
           $CellContext`x2[]]^3) (((-4) 
          Derivative[0, 1][$CellContext`\[Alpha]][
            $CellContext`x1[], 
            $CellContext`x2[]]) Derivative[0, 1][$CellContext`\[Upsilon]][
           $CellContext`x1[], 
           $CellContext`x2[]] - (4 Derivative[1, 0][$CellContext`\[Alpha]][
           $CellContext`x1[], 
           $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Upsilon]][
          $CellContext`x1[], 
          $CellContext`x2[]] + $CellContext`\[Upsilon][
           $CellContext`x1[], 
           $CellContext`x2[]] (Derivative[0, 2][$CellContext`\[Alpha]][
            $CellContext`x1[], 
            $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Alpha]][
            $CellContext`x1[], 
            $CellContext`x2[]])) + (2 $CellContext`\[Alpha][
          $CellContext`x1[], 
          $CellContext`x2[]]) (2 Derivative[0, 1][$CellContext`\[Alpha]][
            $CellContext`x1[], 
            $CellContext`x2[]]^2 + 2 Derivative[1, 0][$CellContext`\[Alpha]][
            $CellContext`x1[], 
            $CellContext`x2[]]^2 - $CellContext`\[Upsilon][
           $CellContext`x1[], 
           $CellContext`x2[]]^2 (Derivative[0, 1][$CellContext`\[Upsilon]][
            $CellContext`x1[], 
            $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Upsilon]][
            $CellContext`x1[], 
            $CellContext`x2[]]^2) + $CellContext`\[Upsilon][
            $CellContext`x1[], 
            $CellContext`x2[]]^3 (Derivative[0, 2][$CellContext`\[Upsilon]][
            $CellContext`x1[], 
            $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Upsilon]][
            $CellContext`x1[], 
            $CellContext`x2[]]))), 0, 0, 0, 0, (($CellContext`\[Gamma][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2) $CellContext`\[Eta][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-8)) ((((-4) $CellContext`\[Alpha][
           $CellContext`x1[], 
           $CellContext`x2[]]) $CellContext`\[Upsilon][
           $CellContext`x1[], 
           $CellContext`x2[]]^3) (((-4) 
          Derivative[0, 1][$CellContext`\[Alpha]][
            $CellContext`x1[], 
            $CellContext`x2[]]) Derivative[0, 1][$CellContext`\[Upsilon]][
           $CellContext`x1[], 
           $CellContext`x2[]] - (4 Derivative[1, 0][$CellContext`\[Alpha]][
           $CellContext`x1[], 
           $CellContext`x2[]]) Derivative[1, 0][$CellContext`\[Upsilon]][
          $CellContext`x1[], 
          $CellContext`x2[]] + $CellContext`\[Upsilon][
           $CellContext`x1[], 
           $CellContext`x2[]] (Derivative[0, 2][$CellContext`\[Alpha]][
            $CellContext`x1[], 
            $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Alpha]][
            $CellContext`x1[], 
            $CellContext`x2[]])) + (4 $CellContext`\[Alpha][
           $CellContext`x1[], 
           $CellContext`x2[]]^2) (2 Derivative[0, 1][$CellContext`\[Alpha]][
            $CellContext`x1[], 
            $CellContext`x2[]]^2 + 2 Derivative[1, 0][$CellContext`\[Alpha]][
            $CellContext`x1[], 
            $CellContext`x2[]]^2 - $CellContext`\[Upsilon][
           $CellContext`x1[], 
           $CellContext`x2[]]^2 (Derivative[0, 1][$CellContext`\[Upsilon]][
            $CellContext`x1[], 
            $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Upsilon]][
            $CellContext`x1[], 
            $CellContext`x2[]]^2) + $CellContext`\[Upsilon][
            $CellContext`x1[], 
            $CellContext`x2[]]^3 (Derivative[0, 2][$CellContext`\[Upsilon]][
            $CellContext`x1[], 
            $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Upsilon]][
            $CellContext`x1[], 
            $CellContext`x2[]])) - $CellContext`\[Upsilon][
         $CellContext`x1[], 
         $CellContext`x2[]]^4 (2 Derivative[0, 1][$CellContext`\[Alpha]][
           $CellContext`x1[], 
           $CellContext`x2[]]^2 + 2 Derivative[1, 0][$CellContext`\[Alpha]][
           $CellContext`x1[], 
           $CellContext`x2[]]^2 - $CellContext`\[Upsilon][
          $CellContext`x1[], 
          $CellContext`x2[]]^2 (Derivative[0, 1][$CellContext`\[Upsilon]][
           $CellContext`x1[], 
           $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Upsilon]][
           $CellContext`x1[], 
           $CellContext`x2[]]^2) + $CellContext`\[Upsilon][
           $CellContext`x1[], 
           $CellContext`x2[]]^3 (Derivative[0, 2][$CellContext`\[Upsilon]][
           $CellContext`x1[], 
           $CellContext`x2[]] + Derivative[2, 0][$CellContext`\[Upsilon]][
           $CellContext`x1[], 
           $CellContext`x2[]])))}}, {-$CellContext`FK, -$CellContext`FK}, 
   0][-$CellContext`a, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.728677366430746*^9},
 CellLabel->"Out[299]=",
 CellID->1346376755,ExpressionUUID->"e620f93f-da13-4af0-a16d-01718162c5f0"]
}, Open  ]],

Cell[CellGroupData[{

Cell["These are the analiticity rules used by Bonanos:", "MathCaption",
 CellID->1456317855,ExpressionUUID->"18b05fea-5955-4d62-938b-9d2db5eb50a7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"rules", "=", 
  RowBox[{"{", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"D", "[", 
      RowBox[{
       RowBox[{"\[Alpha]", "[", 
        RowBox[{
         RowBox[{"x1", "[", "]"}], ",", 
         RowBox[{"x2", "[", "]"}]}], "]"}], ",", 
       RowBox[{"x1", "[", "]"}]}], "]"}], "\[Rule]", 
     RowBox[{
      RowBox[{"-", 
       RowBox[{"\[Upsilon]", "[", 
        RowBox[{
         RowBox[{"x1", "[", "]"}], ",", 
         RowBox[{"x2", "[", "]"}]}], "]"}]}], 
      RowBox[{"D", "[", 
       RowBox[{
        RowBox[{"\[Upsilon]", "[", 
         RowBox[{
          RowBox[{"x1", "[", "]"}], ",", 
          RowBox[{"x2", "[", "]"}]}], "]"}], ",", 
        RowBox[{"x2", "[", "]"}]}], "]"}]}]}], ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"D", "[", 
      RowBox[{
       RowBox[{"\[Alpha]", "[", 
        RowBox[{
         RowBox[{"x1", "[", "]"}], ",", 
         RowBox[{"x2", "[", "]"}]}], "]"}], ",", 
       RowBox[{"x2", "[", "]"}]}], "]"}], "\[Rule]", "   ", 
     RowBox[{
      RowBox[{"\[Upsilon]", "[", 
       RowBox[{
        RowBox[{"x1", "[", "]"}], ",", 
        RowBox[{"x2", "[", "]"}]}], "]"}], 
      RowBox[{"D", "[", 
       RowBox[{
        RowBox[{"\[Upsilon]", "[", 
         RowBox[{
          RowBox[{"x1", "[", "]"}], ",", 
          RowBox[{"x2", "[", "]"}]}], "]"}], ",", 
        RowBox[{"x1", "[", "]"}]}], "]"}]}]}], ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"D", "[", 
      RowBox[{
       RowBox[{"\[Beta]", "[", 
        RowBox[{
         RowBox[{"x1", "[", "]"}], ",", 
         RowBox[{"x2", "[", "]"}]}], "]"}], ",", 
       RowBox[{"x1", "[", "]"}]}], "]"}], "\[Rule]", 
     RowBox[{
      RowBox[{"-", 
       RowBox[{"\[Gamma]", "[", 
        RowBox[{
         RowBox[{"x1", "[", "]"}], ",", 
         RowBox[{"x2", "[", "]"}]}], "]"}]}], 
      RowBox[{"D", "[", 
       RowBox[{
        RowBox[{"\[Gamma]", "[", 
         RowBox[{
          RowBox[{"x1", "[", "]"}], ",", 
          RowBox[{"x2", "[", "]"}]}], "]"}], ",", 
        RowBox[{"x2", "[", "]"}]}], "]"}]}]}], ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"D", "[", 
      RowBox[{
       RowBox[{"\[Beta]", "[", 
        RowBox[{
         RowBox[{"x1", "[", "]"}], ",", 
         RowBox[{"x2", "[", "]"}]}], "]"}], ",", 
       RowBox[{"x2", "[", "]"}]}], "]"}], "\[Rule]", "   ", 
     RowBox[{
      RowBox[{"\[Gamma]", "[", 
       RowBox[{
        RowBox[{"x1", "[", "]"}], ",", 
        RowBox[{"x2", "[", "]"}]}], "]"}], 
      RowBox[{"D", "[", 
       RowBox[{
        RowBox[{"\[Gamma]", "[", 
         RowBox[{
          RowBox[{"x1", "[", "]"}], ",", 
          RowBox[{"x2", "[", "]"}]}], "]"}], ",", 
        RowBox[{"x1", "[", "]"}]}], "]"}]}]}], ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"D", "[", 
      RowBox[{
       RowBox[{"\[Chi]", "[", 
        RowBox[{
         RowBox[{"x1", "[", "]"}], ",", 
         RowBox[{"x2", "[", "]"}]}], "]"}], ",", 
       RowBox[{"x1", "[", "]"}]}], "]"}], "\[Rule]", 
     RowBox[{
      RowBox[{"-", 
       RowBox[{"\[Eta]", "[", 
        RowBox[{
         RowBox[{"x1", "[", "]"}], ",", 
         RowBox[{"x2", "[", "]"}]}], "]"}]}], 
      RowBox[{"D", "[", 
       RowBox[{
        RowBox[{"\[Eta]", "[", 
         RowBox[{
          RowBox[{"x1", "[", "]"}], ",", 
          RowBox[{"x2", "[", "]"}]}], "]"}], ",", 
        RowBox[{"x2", "[", "]"}]}], "]"}]}]}], ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"D", "[", 
      RowBox[{
       RowBox[{"\[Chi]", "[", 
        RowBox[{
         RowBox[{"x1", "[", "]"}], ",", 
         RowBox[{"x2", "[", "]"}]}], "]"}], ",", 
       RowBox[{"x2", "[", "]"}]}], "]"}], "\[Rule]", "   ", 
     RowBox[{
      RowBox[{"\[Eta]", "[", 
       RowBox[{
        RowBox[{"x1", "[", "]"}], ",", 
        RowBox[{"x2", "[", "]"}]}], "]"}], 
      RowBox[{"D", "[", 
       RowBox[{
        RowBox[{"\[Eta]", "[", 
         RowBox[{
          RowBox[{"x1", "[", "]"}], ",", 
          RowBox[{"x2", "[", "]"}]}], "]"}], ",", 
        RowBox[{"x1", "[", "]"}]}], "]"}]}]}], ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{
      RowBox[{"Z_.", " ", 
       RowBox[{
        RowBox[{
         RowBox[{"Derivative", "[", 
          RowBox[{"0", ",", "2"}], "]"}], "[", "G_", "]"}], "[", 
        RowBox[{
         RowBox[{"x1", "[", "]"}], ",", 
         RowBox[{"x2", "[", "]"}]}], "]"}]}], "+", 
      RowBox[{"Z_.", " ", 
       RowBox[{
        RowBox[{
         RowBox[{"Derivative", "[", 
          RowBox[{"2", ",", "0"}], "]"}], "[", "G_", "]"}], "[", 
        RowBox[{
         RowBox[{"x1", "[", "]"}], ",", 
         RowBox[{"x2", "[", "]"}]}], "]"}]}]}], "\[RuleDelayed]", " ", 
     RowBox[{
      RowBox[{"-", "Z"}], 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{
           RowBox[{
            RowBox[{"Derivative", "[", 
             RowBox[{"1", ",", "0"}], "]"}], "[", "G", "]"}], "[", 
           RowBox[{
            RowBox[{"x1", "[", "]"}], ",", 
            RowBox[{"x2", "[", "]"}]}], "]"}], "^", "2"}], "+", 
         RowBox[{
          RowBox[{
           RowBox[{
            RowBox[{"Derivative", "[", 
             RowBox[{"0", ",", "1"}], "]"}], "[", "G", "]"}], "[", 
           RowBox[{
            RowBox[{"x1", "[", "]"}], ",", 
            RowBox[{"x2", "[", "]"}]}], "]"}], "^", "2"}]}], ")"}], "/", 
       RowBox[{"G", "[", 
        RowBox[{
         RowBox[{"x1", "[", "]"}], ",", 
         RowBox[{"x2", "[", "]"}]}], "]"}]}]}]}]}], "\[IndentingNewLine]", 
   "}"}]}]], "Input",
 CellLabel->"In[300]:=",
 CellID->1463064126,ExpressionUUID->"453d346a-7e35-4f08-b661-54846bec1690"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha],
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"1", ",", "0"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["x1",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x1[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["x2",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x2[],
       Editable->False]}], "]"}], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Upsilon]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Upsilon],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["x1",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x1[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["x2",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x2[],
         Editable->False]}], "]"}]}], " ", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Upsilon]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Upsilon],
        Editable->False], 
       TagBox[
        RowBox[{"(", 
         RowBox[{"0", ",", "1"}], ")"}],
        Derivative],
       MultilineFunction->None], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["x1",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x1[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["x2",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x2[],
        Editable->False]}], "]"}]}]}], ",", 
   RowBox[{
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha],
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"0", ",", "1"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["x1",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x1[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["x2",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x2[],
       Editable->False]}], "]"}], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Upsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Upsilon],
       Editable->False], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["x1",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x1[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["x2",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x2[],
        Editable->False]}], "]"}], " ", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Upsilon]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Upsilon],
        Editable->False], 
       TagBox[
        RowBox[{"(", 
         RowBox[{"1", ",", "0"}], ")"}],
        Derivative],
       MultilineFunction->None], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["x1",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x1[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["x2",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x2[],
        Editable->False]}], "]"}]}]}], ",", 
   RowBox[{
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta],
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"1", ",", "0"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["x1",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x1[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["x2",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x2[],
       Editable->False]}], "]"}], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Gamma]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["x1",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x1[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["x2",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x2[],
         Editable->False]}], "]"}]}], " ", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Gamma]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma],
        Editable->False], 
       TagBox[
        RowBox[{"(", 
         RowBox[{"0", ",", "1"}], ")"}],
        Derivative],
       MultilineFunction->None], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["x1",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x1[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["x2",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x2[],
        Editable->False]}], "]"}]}]}], ",", 
   RowBox[{
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta],
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"0", ",", "1"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["x1",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x1[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["x2",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x2[],
       Editable->False]}], "]"}], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma],
       Editable->False], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["x1",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x1[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["x2",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x2[],
        Editable->False]}], "]"}], " ", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Gamma]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma],
        Editable->False], 
       TagBox[
        RowBox[{"(", 
         RowBox[{"1", ",", "0"}], ")"}],
        Derivative],
       MultilineFunction->None], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["x1",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x1[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["x2",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x2[],
        Editable->False]}], "]"}]}]}], ",", 
   RowBox[{
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Chi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Chi],
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"1", ",", "0"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["x1",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x1[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["x2",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x2[],
       Editable->False]}], "]"}], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Eta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Eta],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["x1",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x1[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["x2",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x2[],
         Editable->False]}], "]"}]}], " ", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Eta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Eta],
        Editable->False], 
       TagBox[
        RowBox[{"(", 
         RowBox[{"0", ",", "1"}], ")"}],
        Derivative],
       MultilineFunction->None], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["x1",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x1[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["x2",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x2[],
        Editable->False]}], "]"}]}]}], ",", 
   RowBox[{
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Chi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Chi],
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"0", ",", "1"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["x1",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x1[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["x2",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x2[],
       Editable->False]}], "]"}], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Eta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta],
       Editable->False], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["x1",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x1[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["x2",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x2[],
        Editable->False]}], "]"}], " ", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Eta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Eta],
        Editable->False], 
       TagBox[
        RowBox[{"(", 
         RowBox[{"1", ",", "0"}], ")"}],
        Derivative],
       MultilineFunction->None], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["x1",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x1[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["x2",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x2[],
        Editable->False]}], "]"}]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"Z_.", " ", 
      RowBox[{
       SuperscriptBox["G_", 
        TagBox[
         RowBox[{"(", 
          RowBox[{"0", ",", "2"}], ")"}],
         Derivative],
        MultilineFunction->None], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["x1",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x1[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["x2",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x2[],
         Editable->False]}], "]"}]}], "+", 
     RowBox[{"Z_.", " ", 
      RowBox[{
       SuperscriptBox["G_", 
        TagBox[
         RowBox[{"(", 
          RowBox[{"2", ",", "0"}], ")"}],
         Derivative],
        MultilineFunction->None], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["x1",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x1[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["x2",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x2[],
         Editable->False]}], "]"}]}]}], "\[RuleDelayed]", 
    RowBox[{"-", 
     FractionBox[
      RowBox[{"Z", " ", 
       RowBox[{"(", 
        RowBox[{
         SuperscriptBox[
          RowBox[{
           SuperscriptBox["G", 
            TagBox[
             RowBox[{"(", 
              RowBox[{"1", ",", "0"}], ")"}],
             Derivative],
            MultilineFunction->None], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["x1",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x1[],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["x2",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x2[],
             Editable->False]}], "]"}], "2"], "+", 
         SuperscriptBox[
          RowBox[{
           SuperscriptBox["G", 
            TagBox[
             RowBox[{"(", 
              RowBox[{"0", ",", "1"}], ")"}],
             Derivative],
            MultilineFunction->None], "[", 
           RowBox[{
            InterpretationBox[
             StyleBox["x1",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x1[],
             Editable->False], ",", 
            InterpretationBox[
             StyleBox["x2",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x2[],
             Editable->False]}], "]"}], "2"]}], ")"}]}], 
      RowBox[{"G", "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["x1",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x1[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["x2",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x2[],
         Editable->False]}], "]"}]]}]}]}], "}"}]], "Output",
 CellChangeTimes->{3.7286773666088343`*^9},
 CellLabel->"Out[300]=",
 CellID->518932846,ExpressionUUID->"e49da1d5-7e5f-45cb-8e62-fe779075e3c0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Using the first 6 rules we get that the Laplacians of \[Alpha], \[Beta], \
\[Gamma] vanish:\
\>", "MathCaption",
 CellID->98042681,ExpressionUUID->"4c86f785-2dfc-46f8-9b93-dfdc303600c4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"D", "[", 
    RowBox[{
     RowBox[{"\[Alpha]", "[", 
      RowBox[{
       RowBox[{"x1", "[", "]"}], ",", 
       RowBox[{"x2", "[", "]"}]}], "]"}], ",", 
     RowBox[{"x1", "[", "]"}], ",", 
     RowBox[{"x1", "[", "]"}]}], "]"}], "+", 
   RowBox[{"D", "[", 
    RowBox[{
     RowBox[{"\[Alpha]", "[", 
      RowBox[{
       RowBox[{"x1", "[", "]"}], ",", 
       RowBox[{"x2", "[", "]"}]}], "]"}], ",", 
     RowBox[{"x2", "[", "]"}], ",", 
     RowBox[{"x2", "[", "]"}]}], "]"}]}], "==", 
  RowBox[{
   RowBox[{"D", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"D", "[", 
       RowBox[{
        RowBox[{"\[Alpha]", "[", 
         RowBox[{
          RowBox[{"x1", "[", "]"}], ",", 
          RowBox[{"x2", "[", "]"}]}], "]"}], ",", 
        RowBox[{"x1", "[", "]"}]}], "]"}], "/.", "rules"}], ",", 
     RowBox[{"x1", "[", "]"}]}], "]"}], "+", 
   RowBox[{"D", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"D", "[", 
       RowBox[{
        RowBox[{"\[Alpha]", "[", 
         RowBox[{
          RowBox[{"x1", "[", "]"}], ",", 
          RowBox[{"x2", "[", "]"}]}], "]"}], ",", 
        RowBox[{"x2", "[", "]"}]}], "]"}], "/.", "rules"}], ",", 
     RowBox[{"x2", "[", "]"}]}], "]"}]}]}]], "Input",
 CellLabel->"In[301]:=",
 CellID->683382318,ExpressionUUID->"f2d1b3fa-117e-4d09-b184-41d17f1d4f71"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["\[Alpha]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha],
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "2"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["x1",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`x1[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["x2",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`x2[],
      Editable->False]}], "]"}], "+", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["\[Alpha]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha],
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"2", ",", "0"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["x1",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`x1[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["x2",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`x2[],
      Editable->False]}], "]"}]}], "\[Equal]", "0"}]], "Output",
 CellChangeTimes->{3.728677366788821*^9},
 CellLabel->"Out[301]=",
 CellID->773252255,ExpressionUUID->"132063cc-7ceb-4f61-9bf4-57478884d458"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["But using the last rule we also get:", "MathCaption",
 CellID->1215196276,ExpressionUUID->"a224a5d9-a725-491a-9b74-c97eba961101"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"D", "[", 
     RowBox[{
      RowBox[{"\[Alpha]", "[", 
       RowBox[{
        RowBox[{"x1", "[", "]"}], ",", 
        RowBox[{"x2", "[", "]"}]}], "]"}], ",", 
      RowBox[{"x1", "[", "]"}], ",", 
      RowBox[{"x1", "[", "]"}]}], "]"}], "+", 
    RowBox[{"D", "[", 
     RowBox[{
      RowBox[{"\[Alpha]", "[", 
       RowBox[{
        RowBox[{"x1", "[", "]"}], ",", 
        RowBox[{"x2", "[", "]"}]}], "]"}], ",", 
      RowBox[{"x2", "[", "]"}], ",", 
      RowBox[{"x2", "[", "]"}]}], "]"}]}], "/.", "rules"}], "//", 
  "Simplify"}]], "Input",
 CellLabel->"In[302]:=",
 CellID->1234327762,ExpressionUUID->"07e8ab1d-0336-4103-aefb-5203b4117da1"],

Cell[BoxData[
 RowBox[{"-", 
  FractionBox[
   RowBox[{
    SuperscriptBox[
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], 
       TagBox[
        RowBox[{"(", 
         RowBox[{"0", ",", "1"}], ")"}],
        Derivative],
       MultilineFunction->None], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["x1",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x1[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["x2",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x2[],
        Editable->False]}], "]"}], "2"], "+", 
    SuperscriptBox[
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], 
       TagBox[
        RowBox[{"(", 
         RowBox[{"1", ",", "0"}], ")"}],
        Derivative],
       MultilineFunction->None], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["x1",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x1[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["x2",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x2[],
        Editable->False]}], "]"}], "2"]}], 
   RowBox[{
    InterpretationBox[
     StyleBox["\[Alpha]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Alpha],
     Editable->False], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["x1",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`x1[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["x2",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`x2[],
      Editable->False]}], "]"}]]}]], "Output",
 CellChangeTimes->{3.728677366938148*^9},
 CellLabel->"Out[302]=",
 CellID->1627101788,ExpressionUUID->"4f0fbff2-6867-45b3-adb5-94bf3f032099"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["And then with the other rules:", "MathCaption",
 CellID->1411293017,ExpressionUUID->"eb7f2f2a-a344-4eec-8307-70fafd0594a5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "/.", "rules"}], "//", "Simplify"}]], "Input",
 CellLabel->"In[303]:=",
 CellID->644855604,ExpressionUUID->"f3d9a494-8327-4e0b-848e-4072cf60769f"],

Cell[BoxData[
 RowBox[{"-", 
  FractionBox[
   RowBox[{
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["\[Upsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Upsilon],
       Editable->False], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["x1",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x1[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["x2",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x2[],
        Editable->False]}], "]"}], "2"], " ", 
    RowBox[{"(", 
     RowBox[{
      SuperscriptBox[
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["\[Upsilon]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Upsilon],
          Editable->False], 
         TagBox[
          RowBox[{"(", 
           RowBox[{"0", ",", "1"}], ")"}],
          Derivative],
         MultilineFunction->None], "[", 
        RowBox[{
         InterpretationBox[
          StyleBox["x1",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x1[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["x2",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x2[],
          Editable->False]}], "]"}], "2"], "+", 
      SuperscriptBox[
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["\[Upsilon]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Upsilon],
          Editable->False], 
         TagBox[
          RowBox[{"(", 
           RowBox[{"1", ",", "0"}], ")"}],
          Derivative],
         MultilineFunction->None], "[", 
        RowBox[{
         InterpretationBox[
          StyleBox["x1",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x1[],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox["x2",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x2[],
          Editable->False]}], "]"}], "2"]}], ")"}]}], 
   RowBox[{
    InterpretationBox[
     StyleBox["\[Alpha]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Alpha],
     Editable->False], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["x1",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`x1[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["x2",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`x2[],
      Editable->False]}], "]"}]]}]], "Output",
 CellChangeTimes->{3.7286773670864353`*^9},
 CellLabel->"Out[303]=",
 CellID->642990018,ExpressionUUID->"ba070409-d0b9-4350-8952-281cd0ac754b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
For some reason, for us the rules given by Bonanos are not enough:\
\>", "MathCaption",
 CellID->1764985683,ExpressionUUID->"6e23595d-c4b1-4ab1-9fde-8ba45905409f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"ricci", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "//.", "rules"}], "//", 
  "Simplify"}]], "Input",
 CellLabel->"In[304]:=",
 CellID->922758170,ExpressionUUID->"82cbe296-a963-4c78-90d8-89e3017c8369"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 4} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "5"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 5} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "6"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 6} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "7"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 7} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "8"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 8} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 4} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "5"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 5} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "6"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 6} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "7"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 7} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "8"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 8} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 4} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "5"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 5} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "6"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 6} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "7"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 7} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{
              SuperscriptBox[
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["\[Upsilon]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Upsilon],
                  Editable->False], 
                 TagBox[
                  RowBox[{"(", 
                   RowBox[{"0", ",", "1"}], ")"}],
                  Derivative],
                 MultilineFunction->None], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"], "+", 
              SuperscriptBox[
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["\[Upsilon]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Upsilon],
                  Editable->False], 
                 TagBox[
                  RowBox[{"(", 
                   RowBox[{"1", ",", "0"}], ")"}],
                  Derivative],
                 MultilineFunction->None], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"]}], 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["\[Alpha]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Alpha],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Gamma]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Gamma],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Eta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Eta],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Upsilon]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Upsilon],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"]}]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "8"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 SuperscriptBox[
                  RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+",
                  
                 SuperscriptBox[
                  RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
               RowBox[{
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}]]}]],
           Annotation[#, {3, 8} -> ((($CellContext`\[Alpha][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^(-1) $CellContext`\[Gamma][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^(-2)) $CellContext`\[Eta][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-2)) (
              Derivative[0, 1][$CellContext`\[Upsilon]][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^2 + 
              Derivative[1, 0][$CellContext`\[Upsilon]][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^2), "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 4} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{
              SuperscriptBox[
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["\[Gamma]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Gamma],
                  Editable->False], 
                 TagBox[
                  RowBox[{"(", 
                   RowBox[{"0", ",", "1"}], ")"}],
                  Derivative],
                 MultilineFunction->None], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"], "+", 
              SuperscriptBox[
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["\[Gamma]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Gamma],
                  Editable->False], 
                 TagBox[
                  RowBox[{"(", 
                   RowBox[{"1", ",", "0"}], ")"}],
                  Derivative],
                 MultilineFunction->None], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"]}], 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["\[Beta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Beta],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Gamma]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Gamma],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Eta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Eta],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Upsilon]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Upsilon],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"]}]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "5"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 SuperscriptBox[
                  RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                 SuperscriptBox[
                  RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
               RowBox[{
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}]]}]],
           Annotation[#, {4, 5} -> ((($CellContext`\[Beta][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^(-1) $CellContext`\[Gamma][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^(-2)) $CellContext`\[Eta][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-2)) (
              Derivative[0, 1][$CellContext`\[Gamma]][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^2 + 
              Derivative[1, 0][$CellContext`\[Gamma]][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^2), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "6"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 6} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "7"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 7} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "8"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 8} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {5, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {5, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {5, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{
              SuperscriptBox[
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["\[Gamma]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Gamma],
                  Editable->False], 
                 TagBox[
                  RowBox[{"(", 
                   RowBox[{"0", ",", "1"}], ")"}],
                  Derivative],
                 MultilineFunction->None], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"], "+", 
              SuperscriptBox[
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["\[Gamma]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Gamma],
                  Editable->False], 
                 TagBox[
                  RowBox[{"(", 
                   RowBox[{"1", ",", "0"}], ")"}],
                  Derivative],
                 MultilineFunction->None], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"]}], 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["\[Beta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Beta],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Gamma]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Gamma],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Eta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Eta],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Upsilon]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Upsilon],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"]}]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "4"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 SuperscriptBox[
                  RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                 SuperscriptBox[
                  RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
               RowBox[{
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Beta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Beta], Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}]]}]],
           Annotation[#, {5, 4} -> ((($CellContext`\[Beta][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^(-1) $CellContext`\[Gamma][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^(-2)) $CellContext`\[Eta][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-2)) (
              Derivative[0, 1][$CellContext`\[Gamma]][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^2 + 
              Derivative[1, 0][$CellContext`\[Gamma]][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^2), "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{"4", " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["\[Gamma]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Gamma],
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x1",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x1[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["x2",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x2[],
                    Editable->False]}], "]"}], "2"], "+", 
                SuperscriptBox[
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["\[Gamma]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Gamma],
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x1",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x1[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["x2",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x2[],
                    Editable->False]}], "]"}], "2"]}], ")"}]}], 
             RowBox[{
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Gamma]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Gamma],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Eta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Eta],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Upsilon]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Upsilon],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"]}]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "5"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{"4", " ", 
                 RowBox[{"(", 
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                   ")"}]}], 
               RowBox[{
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}]]}]],
           Annotation[#, {5, 5} -> (((4 $CellContext`\[Gamma][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^(-2)) $CellContext`\[Eta][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-2)) (
              Derivative[0, 1][$CellContext`\[Gamma]][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^2 + 
              Derivative[1, 0][$CellContext`\[Gamma]][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^2), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "6"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {5, 6} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "7"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {5, 7} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"5", ",", "8"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {5, 8} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"6", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {6, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"6", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {6, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"6", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {6, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"6", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {6, 4} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"6", ",", "5"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {6, 5} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{"4", " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["\[Eta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Eta],
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x1",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x1[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["x2",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x2[],
                    Editable->False]}], "]"}], "2"], "+", 
                SuperscriptBox[
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["\[Eta]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Eta],
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x1",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x1[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["x2",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x2[],
                    Editable->False]}], "]"}], "2"]}], ")"}]}], 
             RowBox[{
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Gamma]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Gamma],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Eta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Eta],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Upsilon]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Upsilon],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"]}]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"6", ",", "6"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{"4", " ", 
                 RowBox[{"(", 
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                   ")"}]}], 
               RowBox[{
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}]]}]],
           Annotation[#, {6, 6} -> (((4 $CellContext`\[Gamma][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^(-2)) $CellContext`\[Eta][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-2)) (
              Derivative[0, 1][$CellContext`\[Eta]][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^2 + 
              Derivative[1, 0][$CellContext`\[Eta]][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^2), "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{
              SuperscriptBox[
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["\[Eta]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Eta],
                  Editable->False], 
                 TagBox[
                  RowBox[{"(", 
                   RowBox[{"0", ",", "1"}], ")"}],
                  Derivative],
                 MultilineFunction->None], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"], "+", 
              SuperscriptBox[
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["\[Eta]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Eta],
                  Editable->False], 
                 TagBox[
                  RowBox[{"(", 
                   RowBox[{"1", ",", "0"}], ")"}],
                  Derivative],
                 MultilineFunction->None], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"]}], 
             RowBox[{
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Gamma]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Gamma],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Eta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Eta],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Upsilon]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Upsilon],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Chi]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Chi],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}]}]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"6", ",", "7"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 SuperscriptBox[
                  RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                 SuperscriptBox[
                  RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
               RowBox[{
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}]]}]],
           Annotation[#, {6, 7} -> ((($CellContext`\[Gamma][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^(-2) $CellContext`\[Eta][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2))/$CellContext`\[Chi][
               $CellContext`x1[], 
               $CellContext`x2[]]) (Derivative[0, 1][$CellContext`\[Eta]][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^2 + 
              Derivative[1, 0][$CellContext`\[Eta]][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^2), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"6", ",", "8"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {6, 8} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"7", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {7, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"7", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {7, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"7", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {7, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"7", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {7, 4} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"7", ",", "5"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {7, 5} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{
              SuperscriptBox[
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["\[Eta]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Eta],
                  Editable->False], 
                 TagBox[
                  RowBox[{"(", 
                   RowBox[{"0", ",", "1"}], ")"}],
                  Derivative],
                 MultilineFunction->None], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"], "+", 
              SuperscriptBox[
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["\[Eta]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Eta],
                  Editable->False], 
                 TagBox[
                  RowBox[{"(", 
                   RowBox[{"1", ",", "0"}], ")"}],
                  Derivative],
                 MultilineFunction->None], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"]}], 
             RowBox[{
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Gamma]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Gamma],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Eta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Eta],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Upsilon]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Upsilon],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"], " ", 
              RowBox[{
               InterpretationBox[
                StyleBox["\[Chi]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Chi],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}]}]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"7", ",", "6"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 SuperscriptBox[
                  RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                 SuperscriptBox[
                  RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
               RowBox[{
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Chi]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Chi], Editable -> False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}]}]]}]],
           Annotation[#, {7, 6} -> ((($CellContext`\[Gamma][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^(-2) $CellContext`\[Eta][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2))/$CellContext`\[Chi][
               $CellContext`x1[], 
               $CellContext`x2[]]) (Derivative[0, 1][$CellContext`\[Eta]][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^2 + 
              Derivative[1, 0][$CellContext`\[Eta]][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^2), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"7", ",", "7"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {7, 7} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"7", ",", "8"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {7, 8} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"8", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {8, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"8", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {8, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{
              SuperscriptBox[
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["\[Upsilon]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Upsilon],
                  Editable->False], 
                 TagBox[
                  RowBox[{"(", 
                   RowBox[{"0", ",", "1"}], ")"}],
                  Derivative],
                 MultilineFunction->None], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"], "+", 
              SuperscriptBox[
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["\[Upsilon]",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`\[Upsilon],
                  Editable->False], 
                 TagBox[
                  RowBox[{"(", 
                   RowBox[{"1", ",", "0"}], ")"}],
                  Derivative],
                 MultilineFunction->None], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"]}], 
             RowBox[{
              RowBox[{
               InterpretationBox[
                StyleBox["\[Alpha]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Alpha],
                Editable->False], "[", 
               RowBox[{
                InterpretationBox[
                 StyleBox["x1",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x1[],
                 Editable->False], ",", 
                InterpretationBox[
                 StyleBox["x2",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`x2[],
                 Editable->False]}], "]"}], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Gamma]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Gamma],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Eta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Eta],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Upsilon]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Upsilon],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"]}]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"8", ",", "3"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 SuperscriptBox[
                  RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                 SuperscriptBox[
                  RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
               RowBox[{
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "[", 
                   RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}]]}]],
           Annotation[#, {8, 3} -> ((($CellContext`\[Alpha][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^(-1) $CellContext`\[Gamma][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^(-2)) $CellContext`\[Eta][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-2)) (
              Derivative[0, 1][$CellContext`\[Upsilon]][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^2 + 
              Derivative[1, 0][$CellContext`\[Upsilon]][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^2), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"8", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {8, 4} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"8", ",", "5"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {8, 5} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"8", ",", "6"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {8, 6} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"8", ",", "7"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {8, 7} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{"4", " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["\[Upsilon]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Upsilon],
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x1",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x1[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["x2",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x2[],
                    Editable->False]}], "]"}], "2"], "+", 
                SuperscriptBox[
                 RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["\[Upsilon]",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`\[Upsilon],
                    Editable->False], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{
                   InterpretationBox[
                    StyleBox["x1",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x1[],
                    Editable->False], ",", 
                   InterpretationBox[
                    StyleBox["x2",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`x2[],
                    Editable->False]}], "]"}], "2"]}], ")"}]}], 
             RowBox[{
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Gamma]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Gamma],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Eta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Eta],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"], " ", 
              SuperscriptBox[
               RowBox[{
                InterpretationBox[
                 StyleBox["\[Upsilon]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Upsilon],
                 Editable->False], "[", 
                RowBox[{
                 InterpretationBox[
                  StyleBox["x1",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x1[],
                  Editable->False], ",", 
                 InterpretationBox[
                  StyleBox["x2",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`x2[],
                  Editable->False]}], "]"}], "2"]}]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"8", ",", "8"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{"4", " ", 
                 RowBox[{"(", 
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "1"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], "+", 
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}], Derivative], 
                    MultilineFunction -> None], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}], 
                   ")"}]}], 
               RowBox[{
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Gamma]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Gamma], Editable -> 
                    False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Eta]"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`\[Eta], Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Upsilon]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Upsilon], 
                    Editable -> False], "[", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "x1", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x1[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    "x2", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x2[], Editable -> False]}], "]"}], "2"]}]]}]],
           Annotation[#, {8, 8} -> (((4 $CellContext`\[Gamma][
                   $CellContext`x1[], 
                   $CellContext`x2[]]^(-2)) $CellContext`\[Eta][
                  $CellContext`x1[], 
                  $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^(-2)) (
              Derivative[0, 1][$CellContext`\[Upsilon]][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^2 + 
              Derivative[1, 0][$CellContext`\[Upsilon]][
                 $CellContext`x1[], 
                 $CellContext`x2[]]^2), "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 9 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 9 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " "},
         {"a", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{0, 0, 0, 0, 0, 0, 0, 0}, {0, 0, 0, 0, 0, 0, 0, 0}, {
    0, 0, 0, 0, 0, 0, 0, ((($CellContext`\[Alpha][
           $CellContext`x1[], 
           $CellContext`x2[]]^(-1) $CellContext`\[Gamma][
           $CellContext`x1[], 
           $CellContext`x2[]]^(-2)) $CellContext`\[Eta][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-2)) (Derivative[0, 1][$CellContext`\[Upsilon]][
         $CellContext`x1[], 
         $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Upsilon]][
         $CellContext`x1[], 
         $CellContext`x2[]]^2)}, {0, 0, 0, 0, ((($CellContext`\[Beta][
           $CellContext`x1[], 
           $CellContext`x2[]]^(-1) $CellContext`\[Gamma][
           $CellContext`x1[], 
           $CellContext`x2[]]^(-2)) $CellContext`\[Eta][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-2)) (Derivative[0, 1][$CellContext`\[Gamma]][
         $CellContext`x1[], 
         $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Gamma]][
         $CellContext`x1[], 
         $CellContext`x2[]]^2), 0, 0, 0}, {0, 0, 0, ((($CellContext`\[Beta][
           $CellContext`x1[], 
           $CellContext`x2[]]^(-1) $CellContext`\[Gamma][
           $CellContext`x1[], 
           $CellContext`x2[]]^(-2)) $CellContext`\[Eta][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-2)) (Derivative[0, 1][$CellContext`\[Gamma]][
         $CellContext`x1[], 
         $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Gamma]][
         $CellContext`x1[], 
         $CellContext`x2[]]^2), (((4 $CellContext`\[Gamma][
           $CellContext`x1[], 
           $CellContext`x2[]]^(-2)) $CellContext`\[Eta][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-2)) (Derivative[0, 1][$CellContext`\[Gamma]][
         $CellContext`x1[], 
         $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Gamma]][
         $CellContext`x1[], 
         $CellContext`x2[]]^2), 0, 0, 0}, {
    0, 0, 0, 0, 0, (((4 $CellContext`\[Gamma][
           $CellContext`x1[], 
           $CellContext`x2[]]^(-2)) $CellContext`\[Eta][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-2)) (Derivative[0, 1][$CellContext`\[Eta]][
         $CellContext`x1[], 
         $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Eta]][
         $CellContext`x1[], 
         $CellContext`x2[]]^2), ((($CellContext`\[Gamma][
           $CellContext`x1[], 
           $CellContext`x2[]]^(-2) $CellContext`\[Eta][
           $CellContext`x1[], 
           $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2))/$CellContext`\[Chi][
       $CellContext`x1[], 
       $CellContext`x2[]]) (Derivative[0, 1][$CellContext`\[Eta]][
         $CellContext`x1[], 
         $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Eta]][
         $CellContext`x1[], 
         $CellContext`x2[]]^2), 0}, {0, 0, 0, 0, 0, ((($CellContext`\[Gamma][
           $CellContext`x1[], 
           $CellContext`x2[]]^(-2) $CellContext`\[Eta][
           $CellContext`x1[], 
           $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2))/$CellContext`\[Chi][
       $CellContext`x1[], 
       $CellContext`x2[]]) (Derivative[0, 1][$CellContext`\[Eta]][
         $CellContext`x1[], 
         $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Eta]][
         $CellContext`x1[], 
         $CellContext`x2[]]^2), 0, 0}, {0, 0, ((($CellContext`\[Alpha][
           $CellContext`x1[], 
           $CellContext`x2[]]^(-1) $CellContext`\[Gamma][
           $CellContext`x1[], 
           $CellContext`x2[]]^(-2)) $CellContext`\[Eta][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-2)) (Derivative[0, 1][$CellContext`\[Upsilon]][
         $CellContext`x1[], 
         $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Upsilon]][
         $CellContext`x1[], 
         $CellContext`x2[]]^2), 0, 0, 0, 0, (((4 $CellContext`\[Gamma][
           $CellContext`x1[], 
           $CellContext`x2[]]^(-2)) $CellContext`\[Eta][
          $CellContext`x1[], 
          $CellContext`x2[]]^(-2)) $CellContext`\[Upsilon][
         $CellContext`x1[], 
         $CellContext`x2[]]^(-2)) (Derivative[0, 1][$CellContext`\[Upsilon]][
         $CellContext`x1[], 
         $CellContext`x2[]]^2 + Derivative[1, 0][$CellContext`\[Upsilon]][
         $CellContext`x1[], 
         $CellContext`x2[]]^2)}}, {-$CellContext`FK, -$CellContext`FK}, 
   0][-$CellContext`a, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.7286773672762527`*^9},
 CellLabel->"Out[304]=",
 CellID->595520822,ExpressionUUID->"9db0fdaf-3e6f-4be7-96cf-b4fc7ed3be01"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We need to add the implied rules:", "MathCaption",
 CellID->1444903834,ExpressionUUID->"c0c769b5-df94-47fc-be8c-46a427d16b94"],

Cell[BoxData[
 RowBox[{
  RowBox[{"rules2", "=", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"D", "[", 
        RowBox[{
         RowBox[{"\[Upsilon]", "[", 
          RowBox[{
           RowBox[{"x1", "[", "]"}], ",", 
           RowBox[{"x2", "[", "]"}]}], "]"}], ",", 
         RowBox[{"x1", "[", "]"}]}], "]"}], "^", "2"}], "\[RuleDelayed]", 
      RowBox[{"-", 
       RowBox[{
        RowBox[{"D", "[", 
         RowBox[{
          RowBox[{"\[Upsilon]", "[", 
           RowBox[{
            RowBox[{"x1", "[", "]"}], ",", 
            RowBox[{"x2", "[", "]"}]}], "]"}], ",", 
          RowBox[{"x2", "[", "]"}]}], "]"}], "^", "2"}]}]}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"D", "[", 
        RowBox[{
         RowBox[{"\[Gamma]", "[", 
          RowBox[{
           RowBox[{"x1", "[", "]"}], ",", 
           RowBox[{"x2", "[", "]"}]}], "]"}], ",", 
         RowBox[{"x1", "[", "]"}]}], "]"}], "^", "2"}], "\[RuleDelayed]", 
      RowBox[{"-", 
       RowBox[{
        RowBox[{"D", "[", 
         RowBox[{
          RowBox[{"\[Gamma]", "[", 
           RowBox[{
            RowBox[{"x1", "[", "]"}], ",", 
            RowBox[{"x2", "[", "]"}]}], "]"}], ",", 
          RowBox[{"x2", "[", "]"}]}], "]"}], "^", "2"}]}]}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"D", "[", 
        RowBox[{
         RowBox[{"\[Eta]", "[", 
          RowBox[{
           RowBox[{"x1", "[", "]"}], ",", 
           RowBox[{"x2", "[", "]"}]}], "]"}], ",", 
         RowBox[{"x1", "[", "]"}]}], "]"}], "^", "2"}], "\[RuleDelayed]", 
      RowBox[{"-", 
       RowBox[{
        RowBox[{"D", "[", 
         RowBox[{
          RowBox[{"\[Eta]", "[", 
           RowBox[{
            RowBox[{"x1", "[", "]"}], ",", 
            RowBox[{"x2", "[", "]"}]}], "]"}], ",", 
          RowBox[{"x2", "[", "]"}]}], "]"}], "^", "2"}]}]}]}], 
    "\[IndentingNewLine]", "}"}]}], ";"}]], "Input",
 CellLabel->"In[305]:=",
 CellID->753657762,ExpressionUUID->"e745e6d6-16f5-47f5-bf9c-51554d83ada6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"ricci", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}]}], "]"}], "//.", "rules"}], "//.", "rules2"}], "//", 
  "Simplify"}]], "Input",
 CellLabel->"In[306]:=",
 CellID->867455928,ExpressionUUID->"4de3d195-a751-4bb5-9906-e7ec279c08e0"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728677367568058*^9},
 CellLabel->"Out[306]=",
 CellID->1644660334,ExpressionUUID->"a39b68ee-3663-4e52-971b-b55d147f513f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
That is all a bit fishy, but in principle it is not trivial given that the \
functions are complex. However, if the functions are complex I\
\[CloseCurlyQuote]m not sure the operations are correct because I would \
expect a Hermitian metric, and not just symmetric. The algorithms are not \
prepared to handle Hermitian metrics anyway...\
\>", "MathCaption",
 CellID->173511674,ExpressionUUID->"302c25d6-e51f-4846-a6a6-b4174fc177e3"],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{
  "met", ",", "ricci", ",", "rules", ",", "rules2", ",", "G", ",", "Z"}], 
  "]"}]], "Input",
 CellLabel->"In[307]:=",
 CellID->2075395939,ExpressionUUID->"e8e0f8f8-b32e-4dca-bb63-1a47cf9af351"],

Cell[BoxData[
 RowBox[{"UndefScalarFunction", "[", 
  RowBox[{"{", 
   RowBox[{
   "\[Alpha]", ",", "\[Beta]", ",", "\[Chi]", ",", "\[Gamma]", ",", "\[Eta]", 
    ",", "\[Upsilon]"}], "}"}], "]"}]], "Input",
 CellLabel->"In[308]:=",
 CellID->1750152199,ExpressionUUID->"d257e816-21e4-4ee1-b9aa-eb62445a87ce"],

Cell[BoxData[
 RowBox[{"UndefChart", "[", "FK", "]"}]], "Input",
 CellLabel->"In[309]:=",
 CellID->724495634,ExpressionUUID->"42b82bac-ff56-4847-93d1-1e77a5a73a2a"],

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M", "]"}]], "Input",
 CellLabel->"In[310]:=",
 CellID->365842148,ExpressionUUID->"5e546700-0ac7-435b-a945-83e0a071d5cb"],

Cell[BoxData[
 RowBox[{"ClearxCobaCache", "[", "All", "]"}]], "Input",
 CellLabel->"In[311]:=",
 CellID->463300075,ExpressionUUID->"9033225a-e95e-4161-bb9c-fe6b3a17a678"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.6.8. Harrison metric II-A-7", "Subsubsection",
 CellChangeTimes->{{3.728677132089271*^9, 3.728677132577326*^9}, 
   3.790876775679221*^9},
 CellID->2056140150,ExpressionUUID->"d0300eb7-fb7f-40b1-b5ba-abd1300b4d73"],

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M", ",", "4", ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f"}], "}"}]}], 
  "]"}]], "Input",
 CellLabel->"In[312]:=",
 CellID->240653526,ExpressionUUID->"2705ea94-c0b3-4323-a390-d8285d0fe322"],

Cell[BoxData[
 RowBox[{"DefChart", "[", 
  RowBox[{"H", ",", "M", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2", ",", "3"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"t", "[", "]"}], ",", 
     RowBox[{"x", "[", "]"}], ",", 
     RowBox[{"y", "[", "]"}], ",", 
     RowBox[{"z", "[", "]"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[313]:=",
 CellID->952104181,ExpressionUUID->"b95ba3ff-b4f2-4631-ae8a-6778a5bb7e6b"],

Cell[BoxData[
 RowBox[{"DefScalarFunction", "[", 
  RowBox[{"{", 
   RowBox[{"v", ",", "F2"}], "}"}], "]"}]], "Input",
 CellLabel->"In[314]:=",
 CellID->1114316270,ExpressionUUID->"09500b5f-0ac1-4578-a9a4-6bd6cd6165a4"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"rt3", "=", 
   RowBox[{"Sqrt", "[", "3", "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"F1", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"x", "[", "]"}], "/", "rt3"}], "-", 
    RowBox[{
     RowBox[{
      RowBox[{"t", "[", "]"}], "^", "2"}], "/", "12"}]}]}], " ", 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    SuperscriptBox["F2", "\[Prime]",
     MultilineFunction->None], "[", 
    RowBox[{"z", "[", "]"}], "]"}], "=", 
   RowBox[{
    RowBox[{"F2", "[", 
     RowBox[{"z", "[", "]"}], "]"}], 
    RowBox[{
     RowBox[{"v", "[", 
      RowBox[{"z", "[", "]"}], "]"}], "/", 
     RowBox[{"z", "[", "]"}]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    SuperscriptBox["v", "\[Prime]",
     MultilineFunction->None], "[", 
    RowBox[{"z", "[", "]"}], "]"}], "=", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"v", "[", 
         RowBox[{"z", "[", "]"}], "]"}], "^", "2"}], "-", "1"}], ")"}], "/", 
     RowBox[{"(", 
      RowBox[{"4", 
       RowBox[{"z", "[", "]"}]}], ")"}]}], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", "  ", 
       RowBox[{
        RowBox[{"v", "[", 
         RowBox[{"z", "[", "]"}], "]"}], "/", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"z", "[", "]"}], "-", "1"}], ")"}]}]}], "+", 
      RowBox[{"4", "/", "rt3"}]}], ")"}]}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"n1t", "=", 
   RowBox[{"2", "+", "rt3"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"n1x", "=", 
   RowBox[{"1", "+", "rt3"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"n1y", "=", 
   RowBox[{"-", "rt3"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"n1z", "=", 
   RowBox[{"3", "+", "rt3"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"n2t", "=", 
   RowBox[{"1", "+", 
    RowBox[{"2", "/", "rt3"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"n2x", "=", 
   RowBox[{"1", "+", 
    RowBox[{"2", "/", "rt3"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"n2y", "=", 
   RowBox[{"-", "1"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"n2z", "=", 
   RowBox[{"1", "+", 
    RowBox[{"2", "/", "rt3"}]}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"n3t", "=", 
   RowBox[{"1", "/", "rt3"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"n3x", "=", 
   RowBox[{"1", "+", 
    RowBox[{"1", "/", "rt3"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"n3y", "=", 
   RowBox[{
    RowBox[{"-", "1"}], "/", "rt3"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"n3z", "=", 
   RowBox[{"2", "+", 
    RowBox[{"1", "/", "rt3"}]}]}], ";"}]}], "Input",
 CellLabel->"In[315]:=",
 CellID->108664403,ExpressionUUID->"3f7861af-8d90-4559-be01-bbf203c1dfd4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"met", "=", 
  RowBox[{"CTensor", "[", 
   RowBox[{
    RowBox[{"DiagonalMatrix", "[", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        RowBox[{"F1", "^", "n1t"}], " ", 
        RowBox[{
         RowBox[{"F2", "[", 
          RowBox[{"z", "[", "]"}], "]"}], "^", "n2t"}], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "/", 
           RowBox[{"z", "[", "]"}]}], ")"}], "^", "n3t"}]}], ",", 
       RowBox[{
        RowBox[{"-", 
         RowBox[{"F1", "^", "n1x"}]}], " ", 
        RowBox[{
         RowBox[{"F2", "[", 
          RowBox[{"z", "[", "]"}], "]"}], "^", "n2x"}], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "/", 
           RowBox[{"z", "[", "]"}]}], ")"}], "^", "n3x"}]}], ",", 
       RowBox[{
        RowBox[{"-", 
         RowBox[{"F1", "^", "n1y"}]}], " ", 
        RowBox[{
         RowBox[{"F2", "[", 
          RowBox[{"z", "[", "]"}], "]"}], "^", "n2y"}], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "/", 
           RowBox[{"z", "[", "]"}]}], ")"}], "^", "n3y"}]}], ",", 
       RowBox[{
        RowBox[{"-", 
         RowBox[{"F1", "^", "n1z"}]}], " ", 
        RowBox[{
         RowBox[{"F2", "[", 
          RowBox[{"z", "[", "]"}], "]"}], "^", "n2z"}], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "/", 
           RowBox[{"z", "[", "]"}]}], ")"}], "^", "n3z"}], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"v", "[", 
             RowBox[{"z", "[", "]"}], "]"}], "^", "2"}], "-", "1"}], ")"}], 
         "/", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"z", "[", "]"}], "-", "1"}], ")"}]}]}]}], "}"}], "]"}], 
    ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "H"}], ",", 
      RowBox[{"-", "H"}]}], "}"}]}], "]"}]}]], "Input",
 CellLabel->"In[331]:=",
 CellID->961521271,ExpressionUUID->"803b2cca-c567-41b8-969f-32268df529c3"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["F2",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`F2,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z[],
           Editable->False], "]"}], 
         RowBox[{"1", "+", 
          FractionBox["2", 
           SqrtBox["3"]]}]], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            FractionBox[
             SuperscriptBox[
              InterpretationBox[
               StyleBox["t",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`t[],
               Editable->False], "2"], "12"]}], "+", 
           FractionBox[
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], 
            SqrtBox["3"]]}], ")"}], 
         RowBox[{"2", "+", 
          SqrtBox["3"]}]], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"-", 
           FractionBox["1", 
            InterpretationBox[
             StyleBox["z",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`z[],
             Editable->False]]}], ")"}], 
         FractionBox["1", 
          SqrtBox["3"]]]}], ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", 
       RowBox[{
        RowBox[{"-", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["F2",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`F2,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z[],
            Editable->False], "]"}], 
          RowBox[{"1", "+", 
           FractionBox["2", 
            SqrtBox["3"]]}]]}], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            FractionBox[
             SuperscriptBox[
              InterpretationBox[
               StyleBox["t",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`t[],
               Editable->False], "2"], "12"]}], "+", 
           FractionBox[
            InterpretationBox[
             StyleBox["x",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`x[],
             Editable->False], 
            SqrtBox["3"]]}], ")"}], 
         RowBox[{"1", "+", 
          SqrtBox["3"]}]], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"-", 
           FractionBox["1", 
            InterpretationBox[
             StyleBox["z",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`z[],
             Editable->False]]}], ")"}], 
         RowBox[{"1", "+", 
          FractionBox["1", 
           SqrtBox["3"]]}]]}], ",", "0", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", 
       RowBox[{"-", 
        FractionBox[
         RowBox[{
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", 
              FractionBox[
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["t",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`t[],
                 Editable->False], "2"], "12"]}], "+", 
             FractionBox[
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], 
              SqrtBox["3"]]}], ")"}], 
           RowBox[{"-", 
            SqrtBox["3"]}]], " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{"-", 
             FractionBox["1", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]]}], ")"}], 
           RowBox[{"-", 
            FractionBox["1", 
             SqrtBox["3"]]}]]}], 
         RowBox[{
          InterpretationBox[
           StyleBox["F2",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`F2,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z[],
           Editable->False], "]"}]]}], ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", "0", ",", 
       RowBox[{"-", 
        FractionBox[
         RowBox[{
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["F2",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`F2,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["z",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`z[],
             Editable->False], "]"}], 
           RowBox[{"1", "+", 
            FractionBox["2", 
             SqrtBox["3"]]}]], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "1"}], "+", 
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["v",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`v,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False], "]"}], "2"]}], ")"}], " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", 
              FractionBox[
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["t",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`t[],
                 Editable->False], "2"], "12"]}], "+", 
             FractionBox[
              InterpretationBox[
               StyleBox["x",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`x[],
               Editable->False], 
              SqrtBox["3"]]}], ")"}], 
           RowBox[{"3", "+", 
            SqrtBox["3"]}]], " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{"-", 
             FractionBox["1", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z[],
               Editable->False]]}], ")"}], 
           RowBox[{"2", "+", 
            FractionBox["1", 
             SqrtBox["3"]]}]]}], 
         RowBox[{
          RowBox[{"-", "1"}], "+", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z[],
           Editable->False]}]]}]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "H"}], ",", 
     RowBox[{"-", "H"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.728677375949813*^9},
 CellLabel->"Out[331]=",
 CellID->1231498035,ExpressionUUID->"786906ca-5643-4169-85de-f13061c3ff20"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"$CVSimplify", "=", "Simplify"}], ";"}]], "Input",
 CellLabel->"In[332]:=",
 CellID->515279972,ExpressionUUID->"aa34b0c2-4bd7-45d5-a231-969cca7d0700"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SetCMetric", "[", 
   RowBox[{"met", ",", "H", ",", 
    RowBox[{"SignatureOfMetric", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{"3", ",", "1", ",", "0"}], "}"}]}]}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[333]:=",
 CellID->1656424700,ExpressionUUID->"9dd2b05d-20bd-4f4f-a0a0-08c925fd7d4a"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"3.254341`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.728677379444993*^9},
 CellLabel->"Out[333]=",
 CellID->340084459,ExpressionUUID->"43db0136-cb57-4cab-a6dc-948c11215378"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"MetricCompute", "[", 
   RowBox[{"met", ",", "H", ",", 
    RowBox[{"\"\<Weyl\>\"", "[", 
     RowBox[{
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}]}], "]"}], ",", 
    RowBox[{"Parallelize", "\[Rule]", "True"}]}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[334]:=",
 CellID->566307973,ExpressionUUID->"cf807f30-f1d9-40ba-abec-bd2a041cd21a"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"5.274123`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.7286773849134817`*^9},
 CellLabel->"Out[334]=",
 CellID->1345517746,ExpressionUUID->"1f3d6b04-ff28-4445-9ce4-d628b53ba2fd"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Ricci", "[", 
  RowBox[{"CovDOfMetric", "[", "met", "]"}], "]"}]], "Input",
 CellLabel->"In[335]:=",
 CellID->378515271,ExpressionUUID->"163012b9-de78-4294-9ce8-439ff4398e55"],

Cell[BoxData["Zero"], "Output",
 CellChangeTimes->{3.728677385077537*^9},
 CellLabel->"Out[335]=",
 CellID->1113954112,ExpressionUUID->"aa3fa102-b5d3-43d2-a382-411dcdc8e236"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Determinant", "[", 
   RowBox[{"met", ",", "H"}], "]"}], "[", "]"}]], "Input",
 CellLabel->"In[336]:=",
 CellID->1129042012,ExpressionUUID->"6adbdf9e-af35-4829-afc4-8656aa7e2050"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   OverscriptBox[
    OverscriptBox[
     FractionBox[
      RowBox[{
       SuperscriptBox[
        RowBox[{
         InterpretationBox[
          StyleBox["F2",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`F2,
          Editable->False], "[", 
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z[],
          Editable->False], "]"}], 
        RowBox[{"2", "+", 
         RowBox[{"2", " ", 
          SqrtBox["3"]}]}]], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["v",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`v,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z[],
            Editable->False], "]"}], "2"]}], ")"}], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", 
           FractionBox[
            SuperscriptBox[
             InterpretationBox[
              StyleBox["t",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`t[],
              Editable->False], "2"], "12"]}], "+", 
          FractionBox[
           InterpretationBox[
            StyleBox["x",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x[],
            Editable->False], 
           SqrtBox["3"]]}], ")"}], 
        RowBox[{"2", " ", 
         RowBox[{"(", 
          RowBox[{"3", "+", 
           SqrtBox["3"]}], ")"}]}]], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"-", 
          FractionBox["1", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z[],
            Editable->False]]}], ")"}], 
        RowBox[{"4", "+", 
         FractionBox["2", 
          SqrtBox["3"]]}]], " ", 
       InterpretationBox[
        StyleBox["z",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`z[],
        Editable->False]}], 
      RowBox[{
       RowBox[{"-", "1"}], "+", 
       InterpretationBox[
        StyleBox["z",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`z[],
        Editable->False]}]], 
     StyleBox["~",
      FontColor->RGBColor[1., 0., 0.]]], 
    StyleBox["~",
     FontColor->RGBColor[1., 0., 0.]]],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[(((($CellContext`F2[
          $CellContext`z[]]^(2 + 2 3^Rational[1, 2]) (-1 + $CellContext`v[
           $CellContext`z[]]^2)) (Rational[-1, 12] $CellContext`t[]^2 + 
        3^Rational[-1, 2] $CellContext`x[])^(
        2 (3 + 3^
          Rational[1, 2])))/(-1 + $CellContext`z[])) ((-1)/$CellContext`z[])^(
      4 + 2 3^Rational[-1, 2])) $CellContext`z[], {}, 2 $CellContext`H][],
  Editable->False]], "Output",
 CellChangeTimes->{3.728677385473961*^9},
 CellLabel->"Out[336]=",
 CellID->464325838,ExpressionUUID->"a70e37a2-04bb-4c57-b5d1-3755a50ff9b1"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{
  "met", ",", "F1", ",", "n1", ",", "n2", ",", "n3", ",", "n1t", ",", "n1x", 
   ",", "n1y", ",", "n1z", ",", "n2t", ",", "n2x", ",", "n2y", ",", "n2z", 
   ",", "n3t", ",", "n3x", ",", "n3y", ",", "n3z", ",", "rt3"}], 
  "]"}]], "Input",
 CellLabel->"In[337]:=",
 CellID->2102794852,ExpressionUUID->"30fab4bd-b899-4c9f-9e33-4afa57f900ce"],

Cell[BoxData[
 RowBox[{"UndefScalarFunction", "[", 
  RowBox[{"{", 
   RowBox[{"v", ",", "F2"}], "}"}], "]"}]], "Input",
 CellLabel->"In[338]:=",
 CellID->534698885,ExpressionUUID->"142cf5f1-02c5-4458-974f-a812d00823df"],

Cell[BoxData[
 RowBox[{"UndefChart", "[", "H", "]"}]], "Input",
 CellLabel->"In[339]:=",
 CellID->1612581147,ExpressionUUID->"f45ee040-c982-42d9-b751-eeeb1ae50196"],

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M", "]"}]], "Input",
 CellLabel->"In[340]:=",
 CellID->1407712489,ExpressionUUID->"64bcf3d0-bd6c-4b7f-88c4-6110f8f6d7b1"],

Cell[BoxData[
 RowBox[{"ClearxCobaCache", "[", "All", "]"}]], "Input",
 CellLabel->"In[341]:=",
 CellID->596463624,ExpressionUUID->"f2098fbf-d7b6-488b-8e2c-a559c51319a3"]
}, Closed]],

Cell[CellGroupData[{

Cell["3.6.9. Bondi metric", "Subsubsection",
 CellChangeTimes->{{3.7286771345529623`*^9, 3.728677134761394*^9}, 
   3.790876777231434*^9},
 CellID->310464372,ExpressionUUID->"ed11af34-3206-4e65-9265-efca14ea9494"],

Cell[CellGroupData[{

Cell["\<\
This is an example of a case in which using the default simplification \
provided by Simplify is hopelessly slow. Instead, we use Expand, bringing \
down timings from a few minutes to less than 1 second.\
\>", "Text",
 CellID->1736921641,ExpressionUUID->"ecfd0e7d-b29a-4a22-ac04-e39499f1857b"],

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M", ",", "4", ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f"}], "}"}]}], 
  "]"}]], "Input",
 CellLabel->"In[342]:=",
 CellID->1696806616,ExpressionUUID->"1f4bb601-52b7-45af-b2cd-c499250f19b3"],

Cell[BoxData[
 RowBox[{"DefChart", "[", 
  RowBox[{"bondi", ",", "M", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2", ",", "3"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"u", "[", "]"}], ",", 
     RowBox[{"r", "[", "]"}], ",", 
     RowBox[{"\[Theta]", "[", "]"}], ",", 
     RowBox[{"\[Phi]", "[", "]"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[343]:=",
 CellID->1793446599,ExpressionUUID->"f6eae2a1-2d48-4735-af80-1984aa9067b2"],

Cell[BoxData[
 RowBox[{"DefScalarFunction", "[", 
  RowBox[{"{", 
   RowBox[{"\[Beta]", ",", "V", ",", "\[Gamma]", ",", "U"}], "}"}], 
  "]"}]], "Input",
 CellLabel->"In[344]:=",
 CellID->376520101,ExpressionUUID->"fb4e5c33-240e-45a3-8140-d998013861cd"],

Cell[BoxData[
 RowBox[{
  RowBox[{"met", "=", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{
          RowBox[{
           RowBox[{"Exp", "[", 
            RowBox[{"2", " ", 
             RowBox[{"\[Beta]", "[", 
              RowBox[{
               RowBox[{"u", "[", "]"}], ",", 
               RowBox[{"r", "[", "]"}], ",", 
               RowBox[{"\[Theta]", "[", "]"}]}], "]"}]}], "]"}], 
           RowBox[{
            RowBox[{"V", "[", 
             RowBox[{
              RowBox[{"u", "[", "]"}], ",", 
              RowBox[{"r", "[", "]"}], ",", 
              RowBox[{"\[Theta]", "[", "]"}]}], "]"}], "/", 
            RowBox[{"r", "[", "]"}]}]}], "-", 
          RowBox[{
           RowBox[{
            RowBox[{"r", "[", "]"}], "^", "2"}], 
           RowBox[{"Exp", "[", 
            RowBox[{"2", 
             RowBox[{"\[Gamma]", "[", 
              RowBox[{
               RowBox[{"u", "[", "]"}], ",", 
               RowBox[{"r", "[", "]"}], ",", 
               RowBox[{"\[Theta]", "[", "]"}]}], "]"}]}], "]"}], 
           RowBox[{
            RowBox[{"U", "[", 
             RowBox[{
              RowBox[{"u", "[", "]"}], ",", 
              RowBox[{"r", "[", "]"}], ",", 
              RowBox[{"\[Theta]", "[", "]"}]}], "]"}], "^", "2"}]}]}], ",", 
         RowBox[{"Exp", "[", 
          RowBox[{"2", " ", 
           RowBox[{"\[Beta]", "[", 
            RowBox[{
             RowBox[{"u", "[", "]"}], ",", 
             RowBox[{"r", "[", "]"}], ",", 
             RowBox[{"\[Theta]", "[", "]"}]}], "]"}]}], "]"}], ",", 
         RowBox[{
          RowBox[{
           RowBox[{"r", "[", "]"}], "^", "2"}], 
          RowBox[{"Exp", "[", 
           RowBox[{"2", 
            RowBox[{"\[Gamma]", "[", 
             RowBox[{
              RowBox[{"u", "[", "]"}], ",", 
              RowBox[{"r", "[", "]"}], ",", 
              RowBox[{"\[Theta]", "[", "]"}]}], "]"}]}], "]"}], 
          RowBox[{"U", "[", 
           RowBox[{
            RowBox[{"u", "[", "]"}], ",", 
            RowBox[{"r", "[", "]"}], ",", 
            RowBox[{"\[Theta]", "[", "]"}]}], "]"}]}], ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"Exp", "[", 
          RowBox[{"2", " ", 
           RowBox[{"\[Beta]", "[", 
            RowBox[{
             RowBox[{"u", "[", "]"}], ",", 
             RowBox[{"r", "[", "]"}], ",", 
             RowBox[{"\[Theta]", "[", "]"}]}], "]"}]}], "]"}], ",", "0", ",", 
         "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{
          RowBox[{
           RowBox[{"r", "[", "]"}], "^", "2"}], 
          RowBox[{"Exp", "[", 
           RowBox[{"2", 
            RowBox[{"\[Gamma]", "[", 
             RowBox[{
              RowBox[{"u", "[", "]"}], ",", 
              RowBox[{"r", "[", "]"}], ",", 
              RowBox[{"\[Theta]", "[", "]"}]}], "]"}]}], "]"}], 
          RowBox[{"U", "[", 
           RowBox[{
            RowBox[{"u", "[", "]"}], ",", 
            RowBox[{"r", "[", "]"}], ",", 
            RowBox[{"\[Theta]", "[", "]"}]}], "]"}]}], ",", "0", ",", 
         RowBox[{
          RowBox[{"-", 
           RowBox[{
            RowBox[{"r", "[", "]"}], "^", "2"}]}], 
          RowBox[{"Exp", "[", 
           RowBox[{"2", 
            RowBox[{"\[Gamma]", "[", 
             RowBox[{
              RowBox[{"u", "[", "]"}], ",", 
              RowBox[{"r", "[", "]"}], ",", 
              RowBox[{"\[Theta]", "[", "]"}]}], "]"}]}], "]"}]}], ",", "0"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", 
         RowBox[{
          RowBox[{"-", 
           RowBox[{
            RowBox[{"r", "[", "]"}], "^", "2"}]}], 
          RowBox[{"Exp", "[", 
           RowBox[{
            RowBox[{"-", "2"}], 
            RowBox[{"\[Gamma]", "[", 
             RowBox[{
              RowBox[{"u", "[", "]"}], ",", 
              RowBox[{"r", "[", "]"}], ",", 
              RowBox[{"\[Theta]", "[", "]"}]}], "]"}]}], "]"}], 
          RowBox[{
           RowBox[{"Sin", "[", 
            RowBox[{"\[Theta]", "[", "]"}], "]"}], "^", "2"}]}]}], "}"}]}], 
      "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "bondi"}], ",", 
       RowBox[{"-", "bondi"}]}], "}"}]}], "]"}]}], ";"}]], "Input",
 CellLabel->"In[345]:=",
 CellID->1686513445,ExpressionUUID->"022b29fb-a5b4-42f1-b4dd-cbd756b12c9b"],

Cell[BoxData[
 RowBox[{
  RowBox[{"$CVSimplify", "=", "Expand"}], ";"}]], "Input",
 CellLabel->"In[346]:=",
 CellID->594813362,ExpressionUUID->"56b1ffab-cc49-4c60-8e85-b6046add68f4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SetCMetric", "[", 
   RowBox[{"met", ",", "bondi", ",", 
    RowBox[{"SignatureOfMetric", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{"3", ",", "1", ",", "0"}], "}"}]}]}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[347]:=",
 CellID->491798349,ExpressionUUID->"13b0d17e-8b99-44fb-82d9-113a8125126f"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.029858`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.7286773998207407`*^9},
 CellLabel->"Out[347]=",
 CellID->913687296,ExpressionUUID->"9d4e4827-4f4f-4b11-a667-de76921a0ff2"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"MetricCompute", "[", 
   RowBox[{"met", ",", "bondi", ",", 
    RowBox[{"\"\<Einstein\>\"", "[", 
     RowBox[{
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}]}], "]"}]}], "]"}], "//", "AbsoluteTiming"}]], "Input",\

 CellLabel->"In[348]:=",
 CellID->85729552,ExpressionUUID->"5544e523-f877-4b2c-8e4a-738dd1030aae"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.144877`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.728677400116602*^9},
 CellLabel->"Out[348]=",
 CellID->1834307822,ExpressionUUID->"5bda44a9-90ea-4daa-a0fa-5a94ce2bcb21"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"cd", "=", 
   RowBox[{"CovDOfMetric", "[", "met", "]"}]}], ";"}]], "Input",
 CellLabel->"In[349]:=",
 CellID->1392824723,ExpressionUUID->"5802d2ed-3bf8-4cf3-9551-99824f0c8b12"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Einstein", "[", "cd", "]"}], "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "bondi"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"2", ",", 
     RowBox[{"-", "bondi"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[350]:=",
 CellID->1633367161,ExpressionUUID->"d4cc5dc1-eea8-4181-a52f-400c15894626"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   SuperscriptBox["\[ExponentialE]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta],
       Editable->False], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["u",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`u[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False]}], "]"}]}]], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["U",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`U,
     Editable->False], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}]}], "-", 
  FractionBox[
   RowBox[{
    SuperscriptBox["\[ExponentialE]", 
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Beta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}]], " ", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["V",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`V,
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"0", ",", "0", ",", "1"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["u",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False]}], "]"}]}], 
   RowBox[{"2", " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "+", 
  RowBox[{
   SuperscriptBox["\[ExponentialE]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta],
       Editable->False], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["u",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`u[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False]}], "]"}]}]], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Cot",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cot,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["U",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`U,
     Editable->False], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta],
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "0", ",", "1"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}]}], "-", 
  FractionBox[
   RowBox[{"2", " ", 
    SuperscriptBox["\[ExponentialE]", 
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Beta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}]], " ", 
    RowBox[{
     InterpretationBox[
      StyleBox["V",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`V,
      Editable->False], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["u",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False]}], "]"}], " ", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta],
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"0", ",", "0", ",", "1"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["u",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False]}], "]"}]}], 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]], "+", 
  RowBox[{"3", " ", 
   SuperscriptBox["\[ExponentialE]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta],
       Editable->False], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["u",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`u[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False]}], "]"}]}]], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Cot",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cot,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["U",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`U,
     Editable->False], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["\[Gamma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma],
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "0", ",", "1"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}]}], "-", 
  RowBox[{"2", " ", 
   SuperscriptBox["\[ExponentialE]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta],
       Editable->False], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["u",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`u[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False]}], "]"}]}]], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["U",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`U,
     Editable->False], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}], " ", 
   SuperscriptBox[
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma],
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"0", ",", "0", ",", "1"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["u",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False]}], "]"}], "2"]}], "-", 
  RowBox[{
   SuperscriptBox["\[ExponentialE]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta],
       Editable->False], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["u",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`u[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False]}], "]"}]}]], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["U",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`U,
     Editable->False], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta],
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "0", ",", "2"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}]}], "+", 
  RowBox[{
   SuperscriptBox["\[ExponentialE]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta],
       Editable->False], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["u",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`u[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False]}], "]"}]}]], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["U",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`U,
     Editable->False], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["\[Gamma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma],
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "0", ",", "2"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}]}], "-", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   SuperscriptBox["\[ExponentialE]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "4"}], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Beta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}], "+", 
     RowBox[{"2", " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Gamma]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}]}]], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Cot",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cot,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["U",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`U,
     Editable->False], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["U",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`U,
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "1", ",", "0"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}]}], "-", 
  RowBox[{
   SuperscriptBox["\[ExponentialE]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "4"}], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Beta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}], "+", 
     RowBox[{"2", " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Gamma]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}]}]], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["U",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`U,
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "0", ",", "1"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["U",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`U,
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "1", ",", "0"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}]}], "+", 
  RowBox[{
   SuperscriptBox["\[ExponentialE]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "4"}], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Beta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}], "+", 
     RowBox[{"2", " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Gamma]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}]}]], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["U",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`U,
     Editable->False], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta],
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "0", ",", "1"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["U",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`U,
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "1", ",", "0"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}]}], "-", 
  RowBox[{
   SuperscriptBox["\[ExponentialE]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "4"}], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Beta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}], "+", 
     RowBox[{"2", " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Gamma]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}]}]], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["U",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`U,
     Editable->False], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["\[Gamma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma],
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "0", ",", "1"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["U",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`U,
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "1", ",", "0"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}]}], "+", 
  FractionBox[
   RowBox[{
    SuperscriptBox["\[ExponentialE]", 
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Beta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}]], " ", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["V",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`V,
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"0", ",", "0", ",", "1"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["u",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False]}], "]"}], " ", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta],
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"0", ",", "1", ",", "0"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["u",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False]}], "]"}]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "-", 
  FractionBox[
   RowBox[{"2", " ", 
    SuperscriptBox["\[ExponentialE]", 
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Beta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}]], " ", 
    RowBox[{
     InterpretationBox[
      StyleBox["Cot",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Cot,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], " ", 
    RowBox[{
     InterpretationBox[
      StyleBox["V",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`V,
      Editable->False], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["u",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False]}], "]"}], " ", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma],
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"0", ",", "1", ",", "0"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["u",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False]}], "]"}]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "-", 
  FractionBox[
   RowBox[{
    SuperscriptBox["\[ExponentialE]", 
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Beta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}]], " ", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["V",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`V,
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"0", ",", "0", ",", "1"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["u",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False]}], "]"}], " ", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma],
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"0", ",", "1", ",", "0"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["u",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False]}], "]"}]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{"2", " ", 
    SuperscriptBox["\[ExponentialE]", 
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Beta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}]], " ", 
    RowBox[{
     InterpretationBox[
      StyleBox["V",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`V,
      Editable->False], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["u",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False]}], "]"}], " ", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma],
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"0", ",", "0", ",", "1"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["u",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False]}], "]"}], " ", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma],
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"0", ",", "1", ",", "0"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["u",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False]}], "]"}]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "-", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   SuperscriptBox["\[ExponentialE]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "4"}], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Beta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}], "+", 
     RowBox[{"2", " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Gamma]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}]}]], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["U",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`U,
     Editable->False], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["U",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`U,
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "1", ",", "1"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}]}], "+", 
  FractionBox[
   RowBox[{
    SuperscriptBox["\[ExponentialE]", 
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Beta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}]], " ", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["V",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`V,
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"0", ",", "1", ",", "1"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["u",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False]}], "]"}]}], 
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False]}]], "+", 
  FractionBox[
   RowBox[{
    SuperscriptBox["\[ExponentialE]", 
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Beta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}]], " ", 
    RowBox[{
     InterpretationBox[
      StyleBox["V",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`V,
      Editable->False], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["u",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False]}], "]"}], " ", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta],
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"0", ",", "1", ",", "1"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["u",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False]}], "]"}]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "-", 
  FractionBox[
   RowBox[{
    SuperscriptBox["\[ExponentialE]", 
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Beta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}]], " ", 
    RowBox[{
     InterpretationBox[
      StyleBox["V",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`V,
      Editable->False], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["u",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False]}], "]"}], " ", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma],
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"0", ",", "1", ",", "1"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["u",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False]}], "]"}]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  RowBox[{
   SuperscriptBox["\[ExponentialE]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "4"}], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Beta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}], "+", 
     RowBox[{"2", " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Gamma]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}]}]], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["U",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`U,
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "1", ",", "0"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta],
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"1", ",", "0", ",", "0"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}]}], "+", 
  RowBox[{"2", " ", 
   SuperscriptBox["\[ExponentialE]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta],
       Editable->False], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["u",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`u[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False]}], "]"}]}]], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Cot",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cot,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["\[Gamma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma],
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"1", ",", "0", ",", "0"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}]}], "-", 
  RowBox[{"2", " ", 
   SuperscriptBox["\[ExponentialE]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta],
       Editable->False], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["u",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`u[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False]}], "]"}]}]], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["\[Gamma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma],
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "0", ",", "1"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["\[Gamma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma],
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"1", ",", "0", ",", "0"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}]}], "-", 
  RowBox[{
   SuperscriptBox["\[ExponentialE]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "4"}], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Beta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}], "+", 
     RowBox[{"2", " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Gamma]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}]}]], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["U",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`U,
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "1", ",", "0"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["\[Gamma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma],
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"1", ",", "0", ",", "0"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}]}], "-", 
  RowBox[{
   SuperscriptBox["\[ExponentialE]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta],
       Editable->False], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["u",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`u[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False]}], "]"}]}]], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta],
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"1", ",", "0", ",", "1"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}]}], "+", 
  RowBox[{
   SuperscriptBox["\[ExponentialE]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta],
       Editable->False], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox["u",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`u[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False]}], "]"}]}]], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["\[Gamma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma],
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"1", ",", "0", ",", "1"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}]}], "-", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   SuperscriptBox["\[ExponentialE]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "4"}], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Beta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}], "+", 
     RowBox[{"2", " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Gamma]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}]}]], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["U",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`U,
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"1", ",", "1", ",", "0"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}]}]}]], "Output",
 CellChangeTimes->{3.728677400301697*^9},
 CellLabel->"Out[350]=",
 CellID->2083811669,ExpressionUUID->"05dc5d83-257a-40ea-b14f-35704142f3b4"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Einstein", "[", "cd", "]"}], "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", 
     RowBox[{"-", "bondi"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"2", ",", 
     RowBox[{"-", "bondi"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[351]:=",
 CellID->1253320648,ExpressionUUID->"cd0c71a8-295f-49c5-a17b-148e51191d7e"],

Cell[BoxData[
 RowBox[{
  FractionBox[
   RowBox[{"2", " ", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta],
       Editable->False], 
      TagBox[
       RowBox[{"(", 
        RowBox[{"0", ",", "0", ",", "1"}], ")"}],
       Derivative],
      MultilineFunction->None], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["u",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False]}], "]"}]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  RowBox[{"2", " ", 
   SuperscriptBox["\[ExponentialE]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Beta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}], "+", 
     RowBox[{"2", " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Gamma]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}]}]], " ", 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["U",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`U,
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "1", ",", "0"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}]}], "-", 
  RowBox[{
   SuperscriptBox["\[ExponentialE]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Beta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}], "+", 
     RowBox[{"2", " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Gamma]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}]}]], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["U",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`U,
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "1", ",", "0"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta],
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "1", ",", "0"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}]}], "+", 
  RowBox[{"2", " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Cot",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cot,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["\[Gamma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma],
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "1", ",", "0"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}]}], "-", 
  RowBox[{"2", " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["\[Gamma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma],
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "0", ",", "1"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["\[Gamma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma],
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "1", ",", "0"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}]}], "+", 
  RowBox[{
   SuperscriptBox["\[ExponentialE]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Beta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}], "+", 
     RowBox[{"2", " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Gamma]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}]}]], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["U",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`U,
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "1", ",", "0"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["\[Gamma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma],
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "1", ",", "0"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}]}], "-", 
  RowBox[{
   SuperscriptBox[
    InterpretationBox[
     StyleBox["\[Beta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Beta],
     Editable->False], 
    TagBox[
     RowBox[{"(", 
      RowBox[{"0", ",", "1", ",", "1"}], ")"}],
     Derivative],
    MultilineFunction->None], "[", 
   RowBox[{
    InterpretationBox[
     StyleBox["u",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`u[],
     Editable->False], ",", 
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], ",", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False]}], "]"}], "+", 
  RowBox[{
   SuperscriptBox[
    InterpretationBox[
     StyleBox["\[Gamma]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Gamma],
     Editable->False], 
    TagBox[
     RowBox[{"(", 
      RowBox[{"0", ",", "1", ",", "1"}], ")"}],
     Derivative],
    MultilineFunction->None], "[", 
   RowBox[{
    InterpretationBox[
     StyleBox["u",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`u[],
     Editable->False], ",", 
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], ",", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False]}], "]"}], "+", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   SuperscriptBox["\[ExponentialE]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Beta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}], "+", 
     RowBox[{"2", " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Gamma]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma],
        Editable->False], "[", 
       RowBox[{
        InterpretationBox[
         StyleBox["u",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`u[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], ",", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}]}]], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["U",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`U,
      Editable->False], 
     TagBox[
      RowBox[{"(", 
       RowBox[{"0", ",", "2", ",", "0"}], ")"}],
      Derivative],
     MultilineFunction->None], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox["u",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`u[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False]}], "]"}]}]}]], "Output",
 CellChangeTimes->{3.728677400499357*^9},
 CellLabel->"Out[351]=",
 CellID->1459405607,ExpressionUUID->"21189b60-1247-472f-affc-033d54a4b0fd"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TensorDerivative", "[", 
  RowBox[{"met", ",", "cd"}], "]"}]], "Input",
 CellLabel->"In[352]:=",
 CellID->1872777177,ExpressionUUID->"f23aef84-559f-4771-ae6f-13acdcb78cff"],

Cell[BoxData["Zero"], "Output",
 CellChangeTimes->{3.728677400725721*^9},
 CellLabel->"Out[352]=",
 CellID->1819304462,ExpressionUUID->"56636b99-95ba-4aee-b787-835c72034724"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Check the Bianchi identities:", "MathCaption",
 CellID->1175925424,ExpressionUUID->"5f14c15b-0f67-4d7f-94d3-f5adfafcdf52"],

Cell[BoxData[
 RowBox[{
  RowBox[{"riemann", "=", 
   RowBox[{"RiemannDown", "[", "cd", "]"}]}], ";"}]], "Input",
 CellLabel->"In[353]:=",
 CellID->486597827,ExpressionUUID->"5fb186f5-4c38-4429-9a8a-66afa099e9bc"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"cdriemann", "=", 
    RowBox[{"TensorDerivative", "[", 
     RowBox[{"riemann", ",", "cd"}], "]"}]}], ";"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[354]:=",
 CellID->631950613,ExpressionUUID->"dd258dae-f143-4224-96e5-fb1d08933aa7"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"3.941703`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.7286774050629396`*^9},
 CellLabel->"Out[354]=",
 CellID->1177926848,ExpressionUUID->"05105290-c63d-49a0-92bf-be8e452a97c5"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"cdriemann", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "d"}], ",", 
      RowBox[{"-", "e"}]}], "]"}], "+", 
    RowBox[{"cdriemann", "[", 
     RowBox[{
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "d"}], ",", 
      RowBox[{"-", "e"}]}], "]"}], "+", 
    RowBox[{"cdriemann", "[", 
     RowBox[{
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "d"}], ",", 
      RowBox[{"-", "e"}]}], "]"}]}], "//", "Expand"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[355]:=",
 CellID->1824798542,ExpressionUUID->"a7b2914c-9ed3-489f-9902-66b659a3564b"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.286982`", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.7286774055183287`*^9},
 CellLabel->"Out[355]=",
 CellID->1912276290,ExpressionUUID->"d4a5d477-c54e-4da6-99da-4ddee78de0a3"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{"met", ",", "cd", ",", "riemann", ",", "cdriemann"}], 
  "]"}]], "Input",
 CellLabel->"In[356]:=",
 CellID->46288250,ExpressionUUID->"06e794a1-a292-4daa-b679-dbce70e5beb0"],

Cell[BoxData[
 RowBox[{"UndefScalarFunction", "[", "$ScalarFunctions", "]"}]], "Input",
 CellLabel->"In[357]:=",
 CellID->577537536,ExpressionUUID->"a65379ed-5b51-4d27-984f-1f44aefee5ae"],

Cell[BoxData[
 RowBox[{"UndefChart", "[", "bondi", "]"}]], "Input",
 CellLabel->"In[358]:=",
 CellID->47283112,ExpressionUUID->"40c8f778-45a8-46c1-ac8b-3ff4909d4dac"],

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M", "]"}]], "Input",
 CellLabel->"In[359]:=",
 CellID->118313994,ExpressionUUID->"0f336d37-1256-4b90-be46-bc4d1f06b853"],

Cell[BoxData[
 RowBox[{"ClearxCobaCache", "[", "All", "]"}]], "Input",
 CellLabel->"In[360]:=",
 CellID->2080923683,ExpressionUUID->"311f131b-6c98-4fff-86d1-610186acb37a"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.6.10. Kerr again in Boyer-Lindquist \[Theta]-coordinates", \
"Subsubsection",
 CellChangeTimes->{{3.728677137096867*^9, 3.728677137256919*^9}, 
   3.7908767789123983`*^9},
 CellID->165665139,ExpressionUUID->"a2b2b968-4a83-4f3c-9912-4d847259de35"],

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M", ",", "4", ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f"}], "}"}]}], 
  "]"}]], "Input",
 CellLabel->"In[361]:=",
 CellID->1195184661,ExpressionUUID->"3efa4ea1-4eee-45fe-971a-7427f70f6ca0"],

Cell[BoxData[
 RowBox[{"DefChart", "[", 
  RowBox[{"BL", ",", "M", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2", ",", "3"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"u", "[", "]"}], ",", 
     RowBox[{"r", "[", "]"}], ",", 
     RowBox[{"\[Theta]", "[", "]"}], ",", 
     RowBox[{"\[Phi]", "[", "]"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[362]:=",
 CellID->2054047036,ExpressionUUID->"efbc5179-355e-48c4-9913-1059ae0275a4"],

Cell[BoxData[
 RowBox[{"DefConstantSymbol", "[", 
  RowBox[{"{", 
   RowBox[{"\[Alpha]", ",", "\[CapitalMu]"}], "}"}], "]"}]], "Input",
 CellLabel->"In[363]:=",
 CellID->1568908527,ExpressionUUID->"dc137585-05f0-489a-96d1-fd4f21ddb4fe"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"met", "=", 
  RowBox[{"CTensor", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"1", "-", 
         FractionBox[
          RowBox[{"2", " ", "\[CapitalMu]", " ", 
           RowBox[{"r", "[", "]"}]}], 
          RowBox[{
           SuperscriptBox[
            RowBox[{"r", "[", "]"}], "2"], "+", 
           RowBox[{
            SuperscriptBox["\[Alpha]", "2"], " ", 
            SuperscriptBox[
             RowBox[{"Cos", "[", 
              RowBox[{"\[Theta]", "[", "]"}], "]"}], "2"]}]}]]}], ",", "1", 
        ",", "0", ",", 
        FractionBox[
         RowBox[{"2", "\[Alpha]", "  ", "\[CapitalMu]", " ", 
          RowBox[{"r", "[", "]"}], " ", 
          RowBox[{
           RowBox[{"Sin", "[", 
            RowBox[{"\[Theta]", "[", "]"}], "]"}], "^", "2"}]}], 
         RowBox[{
          SuperscriptBox[
           RowBox[{"r", "[", "]"}], "2"], "+", 
          RowBox[{
           SuperscriptBox["\[Alpha]", "2"], " ", 
           SuperscriptBox[
            RowBox[{"Cos", "[", 
             RowBox[{"\[Theta]", "[", "]"}], "]"}], "2"]}]}]]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", "0", ",", "0", ",", 
        RowBox[{
         RowBox[{"-", "\[Alpha]"}], "  ", 
         RowBox[{
          RowBox[{"Sin", "[", 
           RowBox[{"\[Theta]", "[", "]"}], "]"}], "^", "2"}]}]}], "}"}], ",", 
      
      RowBox[{"{", 
       RowBox[{"0", ",", "0", ",", 
        RowBox[{"-", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"r", "[", "]"}], "^", "2"}], "+", 
           RowBox[{
            RowBox[{"\[Alpha]", "^", "2"}], 
            RowBox[{
             RowBox[{"Cos", "[", 
              RowBox[{"\[Theta]", "[", "]"}], "]"}], "^", "2"}]}]}], ")"}]}], 
        ",", "0"}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{
        FractionBox[
         RowBox[{"2", "\[Alpha]", " ", "\[CapitalMu]", " ", 
          RowBox[{"r", "[", "]"}], " ", 
          RowBox[{
           RowBox[{"Sin", "[", 
            RowBox[{"\[Theta]", "[", "]"}], "]"}], "^", "2"}]}], 
         RowBox[{
          SuperscriptBox[
           RowBox[{"r", "[", "]"}], "2"], "+", 
          RowBox[{
           SuperscriptBox["\[Alpha]", "2"], " ", 
           SuperscriptBox[
            RowBox[{"Cos", "[", 
             RowBox[{"\[Theta]", "[", "]"}], "]"}], "2"]}]}]], ",", 
        RowBox[{
         RowBox[{"-", "\[Alpha]"}], "  ", 
         RowBox[{
          RowBox[{"Sin", "[", 
           RowBox[{"\[Theta]", "[", "]"}], "]"}], "^", "2"}]}], ",", "0", ",", 
        RowBox[{
         RowBox[{"-", 
          RowBox[{
           RowBox[{"Sin", "[", 
            RowBox[{"\[Theta]", "[", "]"}], "]"}], "^", "2"}]}], 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{"2", " ", 
             SuperscriptBox["\[Alpha]", "2"], " ", "\[CapitalMu]", " ", 
             RowBox[{"r", "[", "]"}]}], "+", 
            RowBox[{
             SuperscriptBox["\[Alpha]", "2"], " ", 
             SuperscriptBox[
              RowBox[{"r", "[", "]"}], "2"]}], "+", 
            SuperscriptBox[
             RowBox[{"r", "[", "]"}], "4"], "+", 
            RowBox[{
             SuperscriptBox["\[Alpha]", "2"], " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox["\[Alpha]", "2"], "-", 
               RowBox[{"2", " ", "\[CapitalMu]", " ", 
                RowBox[{"r", "[", "]"}]}], "+", 
               SuperscriptBox[
                RowBox[{"r", "[", "]"}], "2"]}], ")"}], " ", 
             SuperscriptBox[
              RowBox[{"Cos", "[", 
               RowBox[{"\[Theta]", "[", "]"}], "]"}], "2"]}]}], ")"}], "/", 
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox[
             RowBox[{"r", "[", "]"}], "2"], "+", 
            RowBox[{
             SuperscriptBox["\[Alpha]", "2"], " ", 
             SuperscriptBox[
              RowBox[{"Cos", "[", 
               RowBox[{"\[Theta]", "[", "]"}], "]"}], "2"]}]}], ")"}]}]}]}], 
       "}"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "BL"}], ",", 
      RowBox[{"-", "BL"}]}], "}"}]}], "]"}]}]], "Input",
 CellLabel->"In[364]:=",
 CellID->1841395423,ExpressionUUID->"7aac1b31-e191-44f0-9617-254db00b6d27"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"1", "-", 
        FractionBox[
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["\[CapitalMu]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[CapitalMu],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False]}], 
         RowBox[{
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha],
             Editable->False], "2"], " ", 
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["Cos",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Cos,
              Editable->False], "[", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False], "]"}], "2"]}], "+", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]}]]}], ",", "1", ",", "0", ",", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[CapitalMu]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalMu],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Sin",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Sin,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}], 
        RowBox[{
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["\[Alpha]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Alpha],
            Editable->False], "2"], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "2"]}], "+", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"]}]]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", "0", ",", "0", ",", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha],
          Editable->False]}], " ", 
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Sin",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Sin,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"]}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["\[Alpha]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Alpha],
            Editable->False], "2"]}], " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Cos",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Cos,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}], "-", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]}], ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[CapitalMu]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalMu],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Sin",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Sin,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}], 
        RowBox[{
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["\[Alpha]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Alpha],
            Editable->False], "2"], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "2"]}], "+", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"]}]], ",", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha],
          Editable->False]}], " ", 
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Sin",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Sin,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"]}], ",", "0", ",", 
       RowBox[{"-", 
        FractionBox[
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{"2", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Alpha]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Alpha],
               Editable->False], "2"], " ", 
             InterpretationBox[
              StyleBox["\[CapitalMu]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[CapitalMu],
              Editable->False], " ", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]}], "+", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Alpha]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Alpha],
               Editable->False], "2"], " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}], "+", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "4"], "+", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["\[Alpha]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Alpha],
               Editable->False], "2"], " ", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Cos",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Cos,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], "2"], " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["\[Alpha]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Alpha],
                 Editable->False], "2"], "-", 
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["\[CapitalMu]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[CapitalMu],
                 Editable->False], " ", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}], "+", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False], "2"]}], ")"}]}]}], ")"}], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Sin",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Sin,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "2"]}], 
         RowBox[{
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha],
             Editable->False], "2"], " ", 
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["Cos",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Cos,
              Editable->False], "[", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False], "]"}], "2"]}], "+", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]}]]}]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "BL"}], ",", 
     RowBox[{"-", "BL"}]}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.7286774121271143`*^9},
 CellLabel->"Out[364]=",
 CellID->496869296,ExpressionUUID->"d5e2380c-796b-4025-a856-c1d434719185"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"$CVSimplify", "=", "Simplify"}], ";"}]], "Input",
 CellLabel->"In[365]:=",
 CellID->82398920,ExpressionUUID->"e0ba803e-b83b-4d25-a6a4-91303dc9eb79"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SetCMetric", "[", 
   RowBox[{"met", ",", "BL", ",", 
    RowBox[{"SignatureOfMetric", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{"3", ",", "1", ",", "0"}], "}"}]}]}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[366]:=",
 CellID->1899179772,ExpressionUUID->"aaac1a50-131c-4c66-b6b4-ae9daee651db"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"1.588714`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.728677413948189*^9},
 CellLabel->"Out[366]=",
 CellID->1465838336,ExpressionUUID->"b65546a2-1b16-4ecd-a1d5-7fc394af5919"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"MetricCompute", "[", 
   RowBox[{"met", ",", "BL", ",", 
    RowBox[{"\"\<Ricci\>\"", "[", 
     RowBox[{
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}]}], "]"}]}], "]"}], "//", "AbsoluteTiming"}]], "Input",\

 CellLabel->"In[367]:=",
 CellID->1063268046,ExpressionUUID->"a10e1ac7-e102-4f6e-b440-66c782bc3e87"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"5.906369`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.728677420050071*^9},
 CellLabel->"Out[367]=",
 CellID->412514768,ExpressionUUID->"1ed179d4-3549-4671-b47f-eae6356ba670"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Ricci", "[", 
  RowBox[{"CovDOfMetric", "[", "met", "]"}], "]"}]], "Input",
 CellLabel->"In[368]:=",
 CellID->2091963332,ExpressionUUID->"dbe0e60c-71a1-41a0-89c6-5f6cee579a5c"],

Cell[BoxData["Zero"], "Output",
 CellChangeTimes->{3.728677420214003*^9},
 CellLabel->"Out[368]=",
 CellID->1058493260,ExpressionUUID->"6d26ee48-4532-4948-bf2d-280df12bea06"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", "met", "]"}]], "Input",
 CellLabel->"In[369]:=",
 CellID->1219530505,ExpressionUUID->"c93c3b26-b1fe-49ac-a0a5-b4d9c260f497"],

Cell[BoxData[
 RowBox[{"UndefConstantSymbol", "[", 
  RowBox[{"{", 
   RowBox[{"\[Alpha]", ",", "\[CapitalMu]"}], "}"}], "]"}]], "Input",
 CellLabel->"In[370]:=",
 CellID->1065270567,ExpressionUUID->"38a5583b-ac23-4762-a95e-c7316f5cbf7e"],

Cell[BoxData[
 RowBox[{"UndefChart", "[", "BL", "]"}]], "Input",
 CellLabel->"In[371]:=",
 CellID->431254530,ExpressionUUID->"95b40dd6-6965-48ce-83fb-7b883e34b8ca"],

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M", "]"}]], "Input",
 CellLabel->"In[372]:=",
 CellID->337577085,ExpressionUUID->"32fc32c7-d3ff-4af8-8fb2-362f3b2518e3"],

Cell[BoxData[
 RowBox[{"ClearxCobaCache", "[", "All", "]"}]], "Input",
 CellLabel->"In[373]:=",
 CellID->6272221,ExpressionUUID->"131027a4-8e90-49c7-b953-fef9056dc041"]
}, Closed]],

Cell[CellGroupData[{

Cell["3.6.11. Kerr in a rotating frame", "Subsubsection",
 CellChangeTimes->{{3.7286771388249893`*^9, 3.728677139216898*^9}, 
   3.790876780406602*^9},
 CellID->190519644,ExpressionUUID->"08a2783c-e9d4-4fb7-ae20-63478085d21e"],

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M", ",", "4", ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f"}], "}"}]}], 
  "]"}]], "Input",
 CellLabel->"In[374]:=",
 CellID->165539031,ExpressionUUID->"7269e2a0-ec32-48c3-8ffc-06834c3f3f01"],

Cell[BoxData[
 RowBox[{"DefChart", "[", 
  RowBox[{"BL", ",", "M", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2", ",", "3"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"t", "[", "]"}], ",", 
     RowBox[{"r", "[", "]"}], ",", 
     RowBox[{"x", "[", "]"}], ",", 
     RowBox[{"\[Phi]", "[", "]"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[375]:=",
 CellID->1223171152,ExpressionUUID->"ff1f6c37-1ae1-48e0-b53c-c9778ac0e942"],

Cell[BoxData[
 RowBox[{"DefConstantSymbol", "[", 
  RowBox[{"{", 
   RowBox[{"\[Alpha]", ",", "\[CapitalMu]"}], "}"}], "]"}]], "Input",
 CellLabel->"In[376]:=",
 CellID->2124082580,ExpressionUUID->"93e97590-ed37-4c36-bb68-3631deb916ff"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Sigma", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"r", "[", "]"}], "^", "2"}], "+", 
    RowBox[{
     RowBox[{"\[Alpha]", "^", "2"}], 
     RowBox[{
      RowBox[{"x", "[", "]"}], "^", "2"}]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Delta", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"r", "[", "]"}], "^", "2"}], "+", 
    RowBox[{"\[Alpha]", "^", "2"}], "-", 
    RowBox[{"2", "\[CapitalMu]", " ", 
     RowBox[{"r", "[", "]"}]}]}]}], ";"}]}], "Input",
 CellLabel->"In[377]:=",
 CellID->1835257979,ExpressionUUID->"e2dab97a-6aaa-4cf9-8962-8cd90528f704"],

Cell[BoxData[
 RowBox[{
  RowBox[{"$CVSimplify", "=", "Simplify"}], ";"}]], "Input",
 CellLabel->"In[379]:=",
 CellID->873231192,ExpressionUUID->"95ce0234-0fd6-40cd-87a2-358d377d7a40"],

Cell[BoxData[
 RowBox[{"DefBasis", "[", 
  RowBox[{"R", ",", "TangentM", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2", ",", "3"}], "}"}], ",", 
   RowBox[{"BasisColor", "\[Rule]", "Blue"}], ",", 
   RowBox[{"BasisChange", "\[Rule]", 
    RowBox[{"CTensor", "[", 
     RowBox[{
      RowBox[{"Transpose", "@", 
       RowBox[{"Inverse", "@", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{"1", ",", "0", ",", "0", ",", 
            RowBox[{"2", "\[Alpha]", " ", "\[CapitalMu]", " ", 
             RowBox[{"r", "[", "]"}], 
             RowBox[{
              RowBox[{"(", 
               RowBox[{"1", "-", 
                RowBox[{
                 RowBox[{"x", "[", "]"}], "^", "2"}]}], ")"}], "/", 
              RowBox[{"(", 
               RowBox[{"Sigma", "-", 
                RowBox[{"2", "\[CapitalMu]", " ", 
                 RowBox[{"r", "[", "]"}]}]}], ")"}]}]}]}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{"0", ",", "1", ",", "0", ",", "0"}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{"0", ",", "0", ",", "1", ",", "0"}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{"0", ",", "0", ",", "0", ",", "1"}], "}"}]}], "}"}]}]}], 
      ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "R"}], ",", "BL"}], "}"}]}], "]"}]}]}], "]"}]], "Input",
 CellLabel->"In[380]:=",
 CellID->493400346,ExpressionUUID->"ec978f37-fac3-49de-97df-1f14a84d9536"],

Cell[BoxData[
 RowBox[{
  RowBox[{"met", "=", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"DiagonalMatrix", "[", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"1", "-", 
         RowBox[{"2", "\[CapitalMu]", " ", 
          RowBox[{
           RowBox[{"r", "[", "]"}], "/", "Sigma"}]}]}], ",", 
        RowBox[{
         RowBox[{"-", "Sigma"}], "/", "Delta"}], ",", 
        RowBox[{
         RowBox[{"-", "Sigma"}], "/", 
         RowBox[{"(", 
          RowBox[{"1", "-", 
           RowBox[{
            RowBox[{"x", "[", "]"}], "^", "2"}]}], ")"}]}], ",", 
        RowBox[{
         RowBox[{"-", "Delta"}], " ", "Sigma", 
         RowBox[{
          RowBox[{"(", 
           RowBox[{"1", "-", 
            RowBox[{
             RowBox[{"x", "[", "]"}], "^", "2"}]}], ")"}], "/", 
          RowBox[{"(", 
           RowBox[{"Sigma", "-", 
            RowBox[{"2", "\[CapitalMu]", " ", 
             RowBox[{"r", "[", "]"}]}]}], ")"}]}]}]}], "}"}], "]"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "R"}], ",", 
       RowBox[{"-", "R"}]}], "}"}]}], "]"}]}], ";"}]], "Input",
 CellLabel->"In[381]:=",
 CellID->1145254155,ExpressionUUID->"f80260dd-4b34-4204-b935-f46220eb3e85"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SetCMetric", "[", 
   RowBox[{"met", ",", "BL", ",", 
    RowBox[{"SignatureOfMetric", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{"3", ",", "1", ",", "0"}], "}"}]}]}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[382]:=",
 CellID->390655024,ExpressionUUID->"ebeef5f6-a2fc-46b8-b894-daf14a643d22"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.49581`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.728677426439706*^9},
 CellLabel->"Out[382]=",
 CellID->1639756400,ExpressionUUID->"6b14676f-decc-473c-b7eb-e50b8cee305e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"MetricCompute", "[", 
   RowBox[{"met", ",", "BL", ",", 
    RowBox[{"\"\<Weyl\>\"", "[", 
     RowBox[{
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}], ",", 
      RowBox[{"-", "1"}]}], "]"}]}], "]"}], "//", "AbsoluteTiming"}]], "Input",\

 CellLabel->"In[383]:=",
 CellID->1683574617,ExpressionUUID->"66638003-673a-496c-a79d-103fbce568ae"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"1.672115`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.728677428300818*^9},
 CellLabel->"Out[383]=",
 CellID->728172849,ExpressionUUID->"692b21e2-472c-4dd0-876b-3861f4157350"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"cd", "=", 
   RowBox[{"CovDOfMetric", "[", "met", "]"}]}], ";"}]], "Input",
 CellLabel->"In[384]:=",
 CellID->631700026,ExpressionUUID->"9d780fd6-4f8d-4c95-b30e-07263e4092e5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Ricci", "[", "cd", "]"}]], "Input",
 CellLabel->"In[385]:=",
 CellID->1970669861,ExpressionUUID->"a717a385-1f44-44ba-b46c-b2446154fe41"],

Cell[BoxData["Zero"], "Output",
 CellChangeTimes->{3.728677428563377*^9},
 CellLabel->"Out[385]=",
 CellID->2039252537,ExpressionUUID->"d0d7d372-bcba-444d-9bfd-e62e07a676f2"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{"met", ",", "Delta", ",", "Sigma", ",", "cd"}], "]"}]], "Input",
 CellLabel->"In[386]:=",
 CellID->1452430951,ExpressionUUID->"0f9c5bf7-782f-43b0-84d9-07e5d2243c3d"],

Cell[BoxData[
 RowBox[{"UndefConstantSymbol", "[", 
  RowBox[{"{", 
   RowBox[{"\[Alpha]", ",", "\[CapitalMu]"}], "}"}], "]"}]], "Input",
 CellLabel->"In[387]:=",
 CellID->560378008,ExpressionUUID->"674a8da7-471a-4e35-8f3f-6a788114402f"],

Cell[BoxData[
 RowBox[{"UndefChart", "[", "BL", "]"}]], "Input",
 CellLabel->"In[388]:=",
 CellID->1557388361,ExpressionUUID->"f2dd6a5a-fbd9-4af6-82ba-bad372c0bc7a"],

Cell[BoxData[
 RowBox[{"UndefBasis", "[", "R", "]"}]], "Input",
 CellLabel->"In[389]:=",
 CellID->1495493691,ExpressionUUID->"c91fcb19-7c9a-4f9e-b76a-5e0be5f3c3c6"],

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M", "]"}]], "Input",
 CellLabel->"In[390]:=",
 CellID->613860149,ExpressionUUID->"859fbdc4-3188-4d1e-83c4-3f1da4dd9e76"],

Cell[BoxData[
 RowBox[{"ClearxCobaCache", "[", "All", "]"}]], "Input",
 CellLabel->"In[391]:=",
 CellID->636084055,ExpressionUUID->"357e4aa7-7547-46e8-b0b5-decdcca16496"]
}, Closed]],

Cell[CellGroupData[{

Cell["3.6.12. Kerr in a null frame", "Subsubsection",
 CellChangeTimes->{{3.728677140768674*^9, 3.728677141137021*^9}, 
   3.790876784222685*^9},
 CellID->491117619,ExpressionUUID->"01a3bdd2-3d25-4873-8344-57453f334564"],

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M", ",", "4", ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f"}], "}"}]}], 
  "]"}]], "Input",
 CellLabel->"In[392]:=",
 CellID->1434167112,ExpressionUUID->"20c63041-c165-4a88-a7ac-378107eafb88"],

Cell[BoxData[
 RowBox[{"DefChart", "[", 
  RowBox[{"BL", ",", "M", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2", ",", "3"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"t", "[", "]"}], ",", 
     RowBox[{"r", "[", "]"}], ",", 
     RowBox[{"x", "[", "]"}], ",", 
     RowBox[{"\[Phi]", "[", "]"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[393]:=",
 CellID->690197456,ExpressionUUID->"86d8ad8b-e3e5-48af-b3b0-f9d4e251a158"],

Cell[BoxData[
 RowBox[{"DefConstantSymbol", "[", 
  RowBox[{"{", 
   RowBox[{"\[Alpha]", ",", "\[CapitalMu]"}], "}"}], "]"}]], "Input",
 CellLabel->"In[394]:=",
 CellID->1589465545,ExpressionUUID->"637cc882-bf88-403d-b0b0-a42dbb83325d"],

Cell[BoxData[
 RowBox[{"DefBasis", "[", 
  RowBox[{"NP", ",", "TangentM", ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", "2", ",", "3", ",", "4"}], "}"}], ",", 
   RowBox[{"BasisColor", "\[Rule]", "Blue"}], ",", 
   RowBox[{"BasisChange", "\[Rule]", 
    RowBox[{"CTensor", "[", 
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{"{", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{
            FractionBox[
             RowBox[{
              SuperscriptBox[
               RowBox[{"r", "[", "]"}], "2"], "+", 
              RowBox[{
               SuperscriptBox["\[Alpha]", "2"], " ", 
               SuperscriptBox[
                RowBox[{"x", "[", "]"}], "2"]}]}], 
             RowBox[{" ", 
              SqrtBox[
               RowBox[{
                RowBox[{
                 RowBox[{"-", "2"}], " ", "\[CapitalMu]", " ", 
                 RowBox[{"r", "[", "]"}]}], "+", 
                SuperscriptBox[
                 RowBox[{"r", "[", "]"}], "2"], "+", 
                RowBox[{
                 SuperscriptBox["\[Alpha]", "2"], " ", 
                 SuperscriptBox[
                  RowBox[{"x", "[", "]"}], "2"]}]}]]}]], ",", 
            RowBox[{"-", 
             SqrtBox[
              RowBox[{
               SuperscriptBox["\[Alpha]", "2"], "+", 
               RowBox[{
                RowBox[{"r", "[", "]"}], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{"-", "2"}], " ", "\[CapitalMu]"}], "+", 
                  RowBox[{"r", "[", "]"}]}], ")"}]}]}]]}], ",", "0", ",", 
            "0"}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            FractionBox[
             RowBox[{
              SuperscriptBox[
               RowBox[{"r", "[", "]"}], "2"], "+", 
              RowBox[{
               SuperscriptBox["\[Alpha]", "2"], " ", 
               SuperscriptBox[
                RowBox[{"x", "[", "]"}], "2"]}]}], 
             SqrtBox[
              RowBox[{
               RowBox[{
                RowBox[{"-", "2"}], " ", "\[CapitalMu]", " ", 
                RowBox[{"r", "[", "]"}]}], "+", 
               SuperscriptBox[
                RowBox[{"r", "[", "]"}], "2"], "+", 
               RowBox[{
                SuperscriptBox["\[Alpha]", "2"], " ", 
                SuperscriptBox[
                 RowBox[{"x", "[", "]"}], "2"]}]}]]], ",", 
            SqrtBox[
             RowBox[{
              SuperscriptBox["\[Alpha]", "2"], "+", 
              RowBox[{
               RowBox[{"r", "[", "]"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "2"}], " ", "\[CapitalMu]"}], "+", 
                 RowBox[{"r", "[", "]"}]}], ")"}]}]}]], ",", "0", ",", "0"}], 
           "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            FractionBox[
             RowBox[{
             "2", "\[ImaginaryI]", " ", "\[Alpha]", " ", "\[CapitalMu]", " ", 
              
              RowBox[{"r", "[", "]"}], " ", 
              SqrtBox[
               RowBox[{"1", "-", 
                SuperscriptBox[
                 RowBox[{"x", "[", "]"}], "2"]}]]}], 
             RowBox[{
              SqrtBox[
               RowBox[{
                SuperscriptBox["\[Alpha]", "2"], "+", 
                RowBox[{
                 RowBox[{"r", "[", "]"}], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{"-", "2"}], " ", "\[CapitalMu]"}], "+", 
                   RowBox[{"r", "[", "]"}]}], ")"}]}]}]], " ", 
              SqrtBox[
               RowBox[{
                RowBox[{
                 RowBox[{"-", "2"}], " ", "\[CapitalMu]", " ", 
                 RowBox[{"r", "[", "]"}]}], "+", 
                SuperscriptBox[
                 RowBox[{"r", "[", "]"}], "2"], "+", 
                RowBox[{
                 SuperscriptBox["\[Alpha]", "2"], " ", 
                 SuperscriptBox[
                  RowBox[{"x", "[", "]"}], "2"]}]}]]}]], ",", "0", ",", 
            SqrtBox[
             RowBox[{"1", "-", 
              SuperscriptBox[
               RowBox[{"x", "[", "]"}], "2"]}]], ",", 
            RowBox[{"-", 
             FractionBox[
              RowBox[{"\[ImaginaryI]", " ", 
               SqrtBox[
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "2"}], " ", "\[CapitalMu]", " ", 
                  RowBox[{"r", "[", "]"}]}], "+", 
                 SuperscriptBox[
                  RowBox[{"r", "[", "]"}], "2"], "+", 
                 RowBox[{
                  SuperscriptBox["\[Alpha]", "2"], " ", 
                  SuperscriptBox[
                   RowBox[{"x", "[", "]"}], "2"]}]}]]}], 
              RowBox[{
               SqrtBox[
                RowBox[{
                 SuperscriptBox["\[Alpha]", "2"], "+", 
                 RowBox[{
                  RowBox[{"r", "[", "]"}], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", "\[CapitalMu]"}], "+", 
                    RowBox[{"r", "[", "]"}]}], ")"}]}]}]], " ", 
               SqrtBox[
                RowBox[{"1", "-", 
                 SuperscriptBox[
                  RowBox[{"x", "[", "]"}], "2"]}]]}]]}]}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", 
             FractionBox[
              RowBox[{
              "2", "\[ImaginaryI]", " ", "\[Alpha]", " ", "\[CapitalMu]", " ", 
               RowBox[{"r", "[", "]"}], " ", 
               SqrtBox[
                RowBox[{"1", "-", 
                 SuperscriptBox[
                  RowBox[{"x", "[", "]"}], "2"]}]]}], 
              RowBox[{
               SqrtBox[
                RowBox[{
                 SuperscriptBox["\[Alpha]", "2"], "+", 
                 RowBox[{
                  RowBox[{"r", "[", "]"}], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", "\[CapitalMu]"}], "+", 
                    RowBox[{"r", "[", "]"}]}], ")"}]}]}]], 
               SqrtBox[
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "2"}], " ", "\[CapitalMu]", " ", 
                  RowBox[{"r", "[", "]"}]}], "+", 
                 SuperscriptBox[
                  RowBox[{"r", "[", "]"}], "2"], "+", 
                 RowBox[{
                  SuperscriptBox["\[Alpha]", "2"], " ", 
                  SuperscriptBox[
                   RowBox[{"x", "[", "]"}], "2"]}]}]]}]]}], ",", "0", ",", 
            SqrtBox[
             RowBox[{"1", "-", 
              SuperscriptBox[
               RowBox[{"x", "[", "]"}], "2"]}]], ",", 
            FractionBox[
             RowBox[{"\[ImaginaryI]", " ", 
              SqrtBox[
               RowBox[{
                RowBox[{
                 RowBox[{"-", "2"}], " ", "\[CapitalMu]", " ", 
                 RowBox[{"r", "[", "]"}]}], "+", 
                SuperscriptBox[
                 RowBox[{"r", "[", "]"}], "2"], "+", 
                RowBox[{
                 SuperscriptBox["\[Alpha]", "2"], " ", 
                 SuperscriptBox[
                  RowBox[{"x", "[", "]"}], "2"]}]}]]}], 
             RowBox[{
              SqrtBox[
               RowBox[{
                SuperscriptBox["\[Alpha]", "2"], "+", 
                RowBox[{
                 RowBox[{"r", "[", "]"}], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{"-", "2"}], " ", "\[CapitalMu]"}], "+", 
                   RowBox[{"r", "[", "]"}]}], ")"}]}]}]], " ", 
              SqrtBox[
               RowBox[{"1", "-", 
                SuperscriptBox[
                 RowBox[{"x", "[", "]"}], "2"]}]]}]]}], "}"}]}], "}"}], "/", 
        SqrtBox[
         RowBox[{
          SuperscriptBox[
           RowBox[{"r", "[", "]"}], "2"], "+", 
          RowBox[{
           SuperscriptBox["\[Alpha]", "2"], " ", 
           SuperscriptBox[
            RowBox[{"x", "[", "]"}], "2"]}]}]]}], "/", 
       SqrtBox["2"]}], ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "NP"}], ",", "BL"}], "}"}]}], "]"}]}]}], "]"}]], "Input",\

 CellLabel->"In[395]:=",
 CellID->1598329782,ExpressionUUID->"09062ecf-76ac-4966-8054-34512cd20410"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{"met", "=", 
    RowBox[{"CTensor", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{"0", ",", "1", ",", "0", ",", "0"}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{"1", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{"0", ",", "0", ",", "0", ",", 
          RowBox[{"-", "1"}]}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{"0", ",", "0", ",", 
          RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "NP"}], ",", 
        RowBox[{"-", "NP"}]}], "}"}]}], "]"}]}], ")"}], "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}]}], "]"}]], "Input",
 CellLabel->"In[396]:=",
 CellID->255705751,ExpressionUUID->"997ba79f-02cc-4c39-9037-5ed67456391f"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {1, 2} -> 1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["1",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "1"}]],
           Annotation[#, {2, 1} -> 1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{"-", "1"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "4"}], "}"}], "\[Rule]", 
              RowBox[{"-", "1"}]}]],
           Annotation[#, {3, 4} -> -1, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{"-", "1"}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "3"}], "}"}], "\[Rule]", 
              RowBox[{"-", "1"}]}]],
           Annotation[#, {4, 3} -> -1, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 4} -> 0, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]},
           "RowsIndexed" -> {1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " "},
         {"a", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{0, 1, 0, 0}, {1, 0, 0, 0}, {0, 0, 0, -1}, {0, 0, -1, 
   0}}, {-$CellContext`NP, -$CellContext`NP}, 
   0][-$CellContext`a, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.728677437351392*^9, 3.815707636716153*^9, 
  3.815707720168676*^9},
 CellLabel->"Out[396]=",
 CellID->367396622,ExpressionUUID->"83581830-1c23-45eb-9bd0-c3f2d5ab0f29"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"$CVSimplify", "=", "Simplify"}], ";"}]], "Input",
 CellLabel->"In[397]:=",
 CellID->1937088703,ExpressionUUID->"0f34b516-613c-4fd4-b104-2a765c42f952"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"MetricCompute", "[", 
   RowBox[{"met", ",", "BL", ",", "All", ",", 
    RowBox[{"Parallelize", "\[Rule]", "True"}]}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[398]:=",
 CellID->797742801,ExpressionUUID->"7e8f3fde-8c9e-4c62-bb58-8c540b5a96d2"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"8.315597`", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{3.728677445876313*^9},
 CellLabel->"Out[398]=",
 CellID->258296431,ExpressionUUID->"3ff0a464-69d0-448e-b6f9-1b8989d83b93"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"cd", "=", 
   RowBox[{"CovDOfMetric", "[", "met", "]"}]}], ";"}]], "Input",
 CellLabel->"In[399]:=",
 CellID->1182563468,ExpressionUUID->"8a1d5299-5622-4b58-95ed-737278362d5d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Einstein", "[", "cd", "]"}], "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}]}], "]"}]], "Input",
 CellLabel->"In[400]:=",
 CellID->383053691,ExpressionUUID->"13a7e0eb-7b0d-4510-8ed4-fcbee5c26924"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728677446073563*^9, 3.815707731964554*^9},
 CellLabel->"Out[400]=",
 CellID->370118598,ExpressionUUID->"92119922-3da2-4378-852e-d8a7fc48b820"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Kretschmann", "[", "cd", "]"}], "[", "]"}]], "Input",
 CellLabel->"In[401]:=",
 CellID->545924414,ExpressionUUID->"b56d0877-80e9-4da5-b1f6-3107534cd3f0"],

Cell[BoxData[
 FractionBox[
  RowBox[{"48", " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["\[CapitalMu]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalMu],
     Editable->False], "2"], " ", 
   RowBox[{"(", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "6"], "-", 
     RowBox[{"15", " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "4"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}], "+", 
     RowBox[{"15", " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "4"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "4"]}], "-", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "6"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "6"]}]}], ")"}]}], 
  SuperscriptBox[
   RowBox[{"(", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}], ")"}], "6"]]], "Output",
 CellChangeTimes->{3.7286774462458057`*^9, 3.815707733598167*^9},
 CellLabel->"Out[401]=",
 CellID->291909415,ExpressionUUID->"03f671b0-03c3-493f-89fc-f1d32f3812cd"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"cdRiemann", "=", 
   RowBox[{"TensorDerivative", "[", 
    RowBox[{
     RowBox[{"RiemannDown", "[", "cd", "]"}], ",", "cd"}], "]"}]}], 
  ";"}]], "Input",
 CellLabel->"In[402]:=",
 CellID->1956173303,ExpressionUUID->"f9abdf03-5b87-449b-9ac7-d274a3a4bf30"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"cdRiemann", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}], ",", 
     RowBox[{"-", "e"}]}], "]"}], "+", 
   RowBox[{"cdRiemann", "[", 
    RowBox[{
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "d"}], ",", 
     RowBox[{"-", "e"}]}], "]"}], "+", 
   RowBox[{"cdRiemann", "[", 
    RowBox[{
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "d"}], ",", 
     RowBox[{"-", "e"}]}], "]"}]}], "//", "Simplify"}]], "Input",
 CellLabel->"In[403]:=",
 CellID->1577551598,ExpressionUUID->"e4b2aa93-ab32-42b8-ac5b-68fd1345ccbe"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728677446503951*^9, 3.815707736458012*^9},
 CellLabel->"Out[403]=",
 CellID->1523842790,ExpressionUUID->"f2ad4e0a-ae31-41c9-89af-91f852d046d6"]
}, Open  ]],

Cell[CellGroupData[{

Cell["The NP scalars for Kerr (with some choice of signs...):", "MathCaption",
 CellID->1895449293,ExpressionUUID->"7d22f626-4efb-43d0-9305-d7ce1c103b4f"],

Cell[BoxData[
 RowBox[{"SetCMetric", "[", 
  RowBox[{"met", ",", "BL"}], "]"}]], "Input",
 CellLabel->"In[404]:=",
 CellID->1043847853,ExpressionUUID->"274df80c-49c1-4b5f-b389-857d53bb2074"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"l", "=", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
     RowBox[{"{", "NP", "}"}]}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"n", "=", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"0", ",", "1", ",", "0", ",", "0"}], "}"}], ",", 
     RowBox[{"{", "NP", "}"}]}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"m", "=", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", "1", ",", "0"}], "}"}], ",", 
     RowBox[{"{", "NP", "}"}]}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"mbar", "=", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", "0", ",", "1"}], "}"}], ",", 
     RowBox[{"{", "NP", "}"}]}], "]"}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.815707743128133*^9, 3.815707746711977*^9}},
 CellLabel->"In[405]:=",
 CellID->1156953028,ExpressionUUID->"5ab446db-68a5-433b-82ec-533da9636ca9"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"met", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}]}], "]"}], "\[Equal]", 
  RowBox[{
   RowBox[{
    RowBox[{"l", "[", 
     RowBox[{"-", "a"}], "]"}], 
    RowBox[{"n", "[", 
     RowBox[{"-", "b"}], "]"}]}], "+", 
   RowBox[{
    RowBox[{"l", "[", 
     RowBox[{"-", "b"}], "]"}], 
    RowBox[{"n", "[", 
     RowBox[{"-", "a"}], "]"}]}], "-", 
   RowBox[{
    RowBox[{"m", "[", 
     RowBox[{"-", "a"}], "]"}], 
    RowBox[{"mbar", "[", 
     RowBox[{"-", "b"}], "]"}]}], "-", 
   RowBox[{
    RowBox[{"m", "[", 
     RowBox[{"-", "b"}], "]"}], 
    RowBox[{"mbar", "[", 
     RowBox[{"-", "a"}], "]"}]}]}]}]], "Input",
 CellLabel->"In[409]:=",
 CellID->1592601658,ExpressionUUID->"a67174db-b61b-4a8e-9e82-2eaeb525d88d"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728677446790934*^9, 3.815707749499505*^9},
 CellLabel->"Out[409]=",
 CellID->1968718156,ExpressionUUID->"e9d127a3-3935-4570-a857-9a12bd9cb127"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NP\[Kappa]", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"-", 
     RowBox[{"m", "[", "a", "]"}]}], 
    RowBox[{"l", "[", "b", "]"}], 
    RowBox[{
     RowBox[{"cd", "[", 
      RowBox[{"-", "b"}], "]"}], "[", 
     RowBox[{"l", "[", 
      RowBox[{"-", "a"}], "]"}], "]"}]}], "//", "Simplify"}]}]], "Input",
 CellLabel->"In[410]:=",
 CellID->718095,ExpressionUUID->"5629262a-7539-4d40-a868-da903c2e28be"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   InterpretationBox[
    StyleBox["\[Alpha]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Alpha],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["x",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x[],
       Editable->False], "2"]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalMu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalMu],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "-", 
        RowBox[{"\[ImaginaryI]", " ", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["x",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x[],
          Editable->False]}]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["x",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "\[ImaginaryI]"}], " ", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["x",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x[],
          Editable->False]}]}], ")"}]}]}], ")"}]}], 
  RowBox[{"2", " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "\[ImaginaryI]"}], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["x",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x[],
       Editable->False]}]}], ")"}], " ", 
   SqrtBox[
    RowBox[{"2", "-", 
     RowBox[{"2", " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}]], " ", 
   SqrtBox[
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}]], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalMu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalMu],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}], ")"}]}]]], "Output",
 CellChangeTimes->{3.8157077545907173`*^9},
 CellLabel->"Out[410]=",
 CellID->1081645726,ExpressionUUID->"9e925a96-a837-4d8c-8fcb-1802b880315c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NP\[Tau]", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"-", 
     RowBox[{"m", "[", "a", "]"}]}], 
    RowBox[{"n", "[", "b", "]"}], 
    RowBox[{
     RowBox[{"cd", "[", 
      RowBox[{"-", "b"}], "]"}], "[", 
     RowBox[{"l", "[", 
      RowBox[{"-", "a"}], "]"}], "]"}]}], "//", "Simplify"}]}]], "Input",
 CellLabel->"In[411]:=",
 CellID->1166521844,ExpressionUUID->"7220d04c-079c-4e9c-971c-7725a5864c73"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   SuperscriptBox[
    InterpretationBox[
     StyleBox["\[Alpha]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Alpha],
     Editable->False], "2"], " ", 
   InterpretationBox[
    StyleBox["x",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`x[],
    Editable->False], " ", 
   SqrtBox[
    RowBox[{"1", "-", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["x",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x[],
       Editable->False], "2"]}]]}], 
  RowBox[{"2", " ", 
   SqrtBox["2"], " ", 
   SqrtBox[
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}]], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalMu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalMu],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}], ")"}]}]]], "Output",
 CellChangeTimes->{3.728677447883045*^9, 3.815707767828603*^9},
 CellLabel->"Out[411]=",
 CellID->1874226486,ExpressionUUID->"d83c3bec-9d53-40fe-941f-22b3cabd4ee7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NP\[Sigma]", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"-", 
     RowBox[{"m", "[", "a", "]"}]}], 
    RowBox[{"m", "[", "b", "]"}], 
    RowBox[{
     RowBox[{"cd", "[", 
      RowBox[{"-", "b"}], "]"}], "[", 
     RowBox[{"l", "[", 
      RowBox[{"-", "a"}], "]"}], "]"}]}], "//", "Simplify"}]}]], "Input",
 CellLabel->"In[412]:=",
 CellID->387508315,ExpressionUUID->"d190f182-6b96-4632-b8a0-5be84b76276a"],

Cell[BoxData[
 RowBox[{"-", 
  FractionBox[
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["\[Alpha]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha],
      Editable->False], "2"], " ", 
    RowBox[{"(", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[CapitalMu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalMu],
       Editable->False], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}], ")"}]}], 
   RowBox[{"2", " ", 
    SqrtBox["2"], " ", 
    SqrtBox[
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "2"}], " ", 
          InterpretationBox[
           StyleBox["\[CapitalMu]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[CapitalMu],
           Editable->False]}], "+", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]}], ")"}]}]}]], " ", 
    SqrtBox[
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"], "+", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha],
         Editable->False], "2"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], "2"]}]}]], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], " ", 
       InterpretationBox[
        StyleBox["\[CapitalMu]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalMu],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], "+", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"], "+", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha],
         Editable->False], "2"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], "2"]}]}], ")"}]}]]}]], "Output",
 CellChangeTimes->{3.728677448170978*^9, 3.8157077698624763`*^9},
 CellLabel->"Out[412]=",
 CellID->1643886604,ExpressionUUID->"c852abf8-2fef-40d5-8ec7-3624ae724adc"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NP\[Rho]", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"-", 
     RowBox[{"m", "[", "a", "]"}]}], 
    RowBox[{"mbar", "[", "b", "]"}], 
    RowBox[{
     RowBox[{"cd", "[", 
      RowBox[{"-", "b"}], "]"}], "[", 
     RowBox[{"l", "[", 
      RowBox[{"-", "a"}], "]"}], "]"}]}], "//", "Simplify"}]}]], "Input",
 CellLabel->"In[413]:=",
 CellID->1379082493,ExpressionUUID->"4c6507b6-4c63-4d74-a96d-8a8064a5af16"],

Cell[BoxData[
 RowBox[{"-", 
  FractionBox[
   RowBox[{
    RowBox[{
     RowBox[{"-", "2"}], " ", "\[ImaginaryI]", " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         InterpretationBox[
          StyleBox["\[CapitalMu]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalMu],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]}], ")"}], "2"]}], "+", 
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox["\[Alpha]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "2"}], " ", 
        InterpretationBox[
         StyleBox["\[CapitalMu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[CapitalMu],
         Editable->False]}], "+", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], ")"}], " ", 
     InterpretationBox[
      StyleBox["x",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`x[],
      Editable->False]}], "+", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha],
       Editable->False], "3"], " ", 
     InterpretationBox[
      StyleBox["x",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`x[],
      Editable->False], " ", 
     RowBox[{"(", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[CapitalMu]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalMu],
        Editable->False], "+", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "-", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[CapitalMu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[CapitalMu],
         Editable->False], " ", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["x",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x[],
          Editable->False], "2"]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], " ", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["x",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x[],
          Editable->False], "2"]}]}], ")"}]}], "+", 
    RowBox[{"\[ImaginaryI]", " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha],
       Editable->False], "2"], " ", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]}], " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["x",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x[],
            Editable->False], "2"]}], ")"}]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[CapitalMu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[CapitalMu],
         Editable->False], " ", 
        RowBox[{"(", 
         RowBox[{"3", "+", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["x",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`x[],
            Editable->False], "2"]}], ")"}]}]}], ")"}]}]}], 
   RowBox[{"2", " ", 
    SqrtBox["2"], " ", 
    SqrtBox[
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "2"}], " ", 
          InterpretationBox[
           StyleBox["\[CapitalMu]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[CapitalMu],
           Editable->False]}], "+", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]}], ")"}]}]}]], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "\[ImaginaryI]"}], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False]}]}], ")"}], " ", 
    SqrtBox[
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"], "+", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha],
         Editable->False], "2"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], "2"]}]}]], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], " ", 
       InterpretationBox[
        StyleBox["\[CapitalMu]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalMu],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], "+", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"], "+", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha],
         Editable->False], "2"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], "2"]}]}], ")"}]}]]}]], "Output",
 CellChangeTimes->{3.728677448395739*^9, 3.815707771600445*^9},
 CellLabel->"Out[413]=",
 CellID->1503569625,ExpressionUUID->"519aab8f-d26d-46e6-9423-59cba49c703b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NP\[Pi]", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"-", 
     RowBox[{"mbar", "[", "a", "]"}]}], 
    RowBox[{"l", "[", "b", "]"}], 
    RowBox[{
     RowBox[{"cd", "[", 
      RowBox[{"-", "b"}], "]"}], "[", 
     RowBox[{"n", "[", 
      RowBox[{"-", "a"}], "]"}], "]"}]}], "//", "Simplify"}]}]], "Input",
 CellLabel->"In[414]:=",
 CellID->1709791137,ExpressionUUID->"f655f87d-248d-44a1-bd90-1a6ee2c33917"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   SuperscriptBox[
    InterpretationBox[
     StyleBox["\[Alpha]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Alpha],
     Editable->False], "2"], " ", 
   InterpretationBox[
    StyleBox["x",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`x[],
    Editable->False], " ", 
   SqrtBox[
    RowBox[{"1", "-", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["x",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x[],
       Editable->False], "2"]}]]}], 
  RowBox[{"2", " ", 
   SqrtBox["2"], " ", 
   SqrtBox[
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}]], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalMu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalMu],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}], ")"}]}]]], "Output",
 CellChangeTimes->{3.728677449150773*^9, 3.815707774412341*^9},
 CellLabel->"Out[414]=",
 CellID->1648042301,ExpressionUUID->"bafd8cd9-95dd-4146-be2d-5f5fcdfaf963"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NP\[Nu]", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"-", 
     RowBox[{"mbar", "[", "a", "]"}]}], 
    RowBox[{"n", "[", "b", "]"}], 
    RowBox[{
     RowBox[{"cd", "[", 
      RowBox[{"-", "b"}], "]"}], "[", 
     RowBox[{"n", "[", 
      RowBox[{"-", "a"}], "]"}], "]"}]}], "//", "Simplify"}]}]], "Input",
 CellLabel->"In[415]:=",
 CellID->746828945,ExpressionUUID->"356d23d8-2994-4dfc-bf32-5307c79a7848"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   InterpretationBox[
    StyleBox["\[Alpha]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Alpha],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["x",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x[],
       Editable->False], "2"]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalMu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalMu],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "-", 
        RowBox[{"\[ImaginaryI]", " ", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["x",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x[],
          Editable->False]}]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["x",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "\[ImaginaryI]"}], " ", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["x",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x[],
          Editable->False]}]}], ")"}]}]}], ")"}]}], 
  RowBox[{"2", " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "\[ImaginaryI]"}], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["x",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x[],
       Editable->False]}]}], ")"}], " ", 
   SqrtBox[
    RowBox[{"2", "-", 
     RowBox[{"2", " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}]], " ", 
   SqrtBox[
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}]], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalMu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalMu],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}], ")"}]}]]], "Output",
 CellChangeTimes->{3.728677449286951*^9, 3.815707776146328*^9},
 CellLabel->"Out[415]=",
 CellID->118923088,ExpressionUUID->"69df6ecd-35a8-4cbe-9155-ba8ad7cab630"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NP\[Mu]", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"-", 
     RowBox[{"mbar", "[", "a", "]"}]}], 
    RowBox[{"m", "[", "b", "]"}], 
    RowBox[{
     RowBox[{"cd", "[", 
      RowBox[{"-", "b"}], "]"}], "[", 
     RowBox[{"n", "[", 
      RowBox[{"-", "a"}], "]"}], "]"}]}], "//", "Simplify"}]}]], "Input",
 CellLabel->"In[416]:=",
 CellID->1173802436,ExpressionUUID->"a143c83f-179c-40af-91d3-ff6cec28ec03"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   RowBox[{
    RowBox[{"-", "2"}], " ", "\[ImaginaryI]", " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"], " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "2"}], " ", 
        InterpretationBox[
         StyleBox["\[CapitalMu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[CapitalMu],
         Editable->False]}], "+", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], ")"}], "2"]}], "+", 
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["\[Alpha]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Alpha],
     Editable->False], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], " ", 
       InterpretationBox[
        StyleBox["\[CapitalMu]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalMu],
        Editable->False]}], "+", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}], " ", 
    InterpretationBox[
     StyleBox["x",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`x[],
     Editable->False]}], "+", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["\[Alpha]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha],
      Editable->False], "3"], " ", 
    InterpretationBox[
     StyleBox["x",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`x[],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[CapitalMu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalMu],
       Editable->False], "+", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[CapitalMu]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalMu],
        Editable->False], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], "2"]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], "2"]}]}], ")"}]}], "+", 
   RowBox[{"\[ImaginaryI]", " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["\[Alpha]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha],
      Editable->False], "2"], " ", 
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]}], " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], "2"]}], ")"}]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[CapitalMu]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalMu],
        Editable->False], " ", 
       RowBox[{"(", 
        RowBox[{"3", "+", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], "2"]}], ")"}]}]}], ")"}]}]}], 
  RowBox[{"2", " ", 
   SqrtBox["2"], " ", 
   SqrtBox[
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha],
       Editable->False], "2"], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         InterpretationBox[
          StyleBox["\[CapitalMu]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalMu],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]}], ")"}]}]}]], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "\[ImaginaryI]"}], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["x",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x[],
       Editable->False]}]}], ")"}], " ", 
   SqrtBox[
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}]], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalMu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalMu],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}], ")"}]}]]], "Output",
 CellChangeTimes->{3.728677449669304*^9, 3.815707777632897*^9},
 CellLabel->"Out[416]=",
 CellID->535563402,ExpressionUUID->"7f8adb4a-7e3c-456b-89e3-91193f207be0"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NP\[Lambda]", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"-", 
     RowBox[{"mbar", "[", "a", "]"}]}], 
    RowBox[{"mbar", "[", "b", "]"}], 
    RowBox[{
     RowBox[{"cd", "[", 
      RowBox[{"-", "b"}], "]"}], "[", 
     RowBox[{"n", "[", 
      RowBox[{"-", "a"}], "]"}], "]"}]}], "//", "Simplify"}]}]], "Input",
 CellLabel->"In[417]:=",
 CellID->318902292,ExpressionUUID->"16c5bdc1-55f7-42ee-9204-c4c7be497345"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   SuperscriptBox[
    InterpretationBox[
     StyleBox["\[Alpha]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Alpha],
     Editable->False], "2"], " ", 
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[CapitalMu]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalMu],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["x",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x[],
       Editable->False], "2"]}], ")"}]}], 
  RowBox[{"2", " ", 
   SqrtBox["2"], " ", 
   SqrtBox[
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha],
       Editable->False], "2"], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         InterpretationBox[
          StyleBox["\[CapitalMu]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalMu],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]}], ")"}]}]}]], " ", 
   SqrtBox[
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}]], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalMu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalMu],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}], ")"}]}]]], "Output",
 CellChangeTimes->{3.7286774500257196`*^9, 3.815707779024898*^9},
 CellLabel->"Out[417]=",
 CellID->1520517115,ExpressionUUID->"91fadf47-f91f-4697-afa7-966fce5818f7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NP\[Epsilon]", "=", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"-", "1"}], "/", "2"}], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"n", "[", "a", "]"}], 
       RowBox[{"l", "[", "b", "]"}], 
       RowBox[{
        RowBox[{"cd", "[", 
         RowBox[{"-", "b"}], "]"}], "[", 
        RowBox[{"l", "[", 
         RowBox[{"-", "a"}], "]"}], "]"}]}], "-", 
      RowBox[{
       RowBox[{"mbar", "[", "a", "]"}], 
       RowBox[{"l", "[", "b", "]"}], 
       RowBox[{
        RowBox[{"cd", "[", 
         RowBox[{"-", "b"}], "]"}], "[", 
        RowBox[{"m", "[", 
         RowBox[{"-", "a"}], "]"}], "]"}]}]}], ")"}]}], "//", 
   "Simplify"}]}]], "Input",
 CellLabel->"In[418]:=",
 CellID->265141514,ExpressionUUID->"663bcc13-e732-4a2e-97a9-4141945fbf55"],

Cell[BoxData[
 RowBox[{"-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["\[CapitalMu]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalMu],
     Editable->False], " ", 
    SqrtBox[
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "2"}], " ", 
          InterpretationBox[
           StyleBox["\[CapitalMu]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[CapitalMu],
           Editable->False]}], "+", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]}], ")"}]}]}]], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"\[ImaginaryI]", " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False]}]}], ")"}]}], 
   RowBox[{"2", " ", 
    SqrtBox["2"], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "\[ImaginaryI]"}], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False]}]}], ")"}], " ", 
    SqrtBox[
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"], "+", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha],
         Editable->False], "2"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], "2"]}]}]], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], " ", 
       InterpretationBox[
        StyleBox["\[CapitalMu]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalMu],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], "+", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"], "+", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha],
         Editable->False], "2"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], "2"]}]}], ")"}]}]]}]], "Output",
 CellChangeTimes->{3.728677450262916*^9, 3.815707780377964*^9},
 CellLabel->"Out[418]=",
 CellID->282781427,ExpressionUUID->"abb12fa8-4f13-48d9-b543-2fada08947ee"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NP\[Gamma]", "=", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"-", "1"}], "/", "2"}], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"n", "[", "a", "]"}], 
       RowBox[{"n", "[", "b", "]"}], 
       RowBox[{
        RowBox[{"cd", "[", 
         RowBox[{"-", "b"}], "]"}], "[", 
        RowBox[{"l", "[", 
         RowBox[{"-", "a"}], "]"}], "]"}]}], "-", 
      RowBox[{
       RowBox[{"mbar", "[", "a", "]"}], 
       RowBox[{"n", "[", "b", "]"}], 
       RowBox[{
        RowBox[{"cd", "[", 
         RowBox[{"-", "b"}], "]"}], "[", 
        RowBox[{"m", "[", 
         RowBox[{"-", "a"}], "]"}], "]"}]}]}], ")"}]}], "//", 
   "Simplify"}]}]], "Input",
 CellLabel->"In[419]:=",
 CellID->636708170,ExpressionUUID->"3a3bce0d-503e-43df-aa7c-e2d424342613"],

Cell[BoxData[
 RowBox[{"-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["\[CapitalMu]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalMu],
     Editable->False], " ", 
    SqrtBox[
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "2"}], " ", 
          InterpretationBox[
           StyleBox["\[CapitalMu]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[CapitalMu],
           Editable->False]}], "+", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]}], ")"}]}]}]], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"\[ImaginaryI]", " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False]}]}], ")"}]}], 
   RowBox[{"2", " ", 
    SqrtBox["2"], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "\[ImaginaryI]"}], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False]}]}], ")"}], " ", 
    SqrtBox[
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"], "+", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha],
         Editable->False], "2"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], "2"]}]}]], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], " ", 
       InterpretationBox[
        StyleBox["\[CapitalMu]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalMu],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], "+", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"], "+", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha],
         Editable->False], "2"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], "2"]}]}], ")"}]}]]}]], "Output",
 CellChangeTimes->{3.728677450409799*^9, 3.815707781705719*^9},
 CellLabel->"Out[419]=",
 CellID->176308395,ExpressionUUID->"21b057e6-d07d-4035-bf52-1c6254673134"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NP\[Beta]", "=", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"-", "1"}], "/", "2"}], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"n", "[", "a", "]"}], 
       RowBox[{"m", "[", "b", "]"}], 
       RowBox[{
        RowBox[{"cd", "[", 
         RowBox[{"-", "b"}], "]"}], "[", 
        RowBox[{"l", "[", 
         RowBox[{"-", "a"}], "]"}], "]"}]}], "-", 
      RowBox[{
       RowBox[{"mbar", "[", "a", "]"}], 
       RowBox[{"m", "[", "b", "]"}], 
       RowBox[{
        RowBox[{"cd", "[", 
         RowBox[{"-", "b"}], "]"}], "[", 
        RowBox[{"m", "[", 
         RowBox[{"-", "a"}], "]"}], "]"}]}]}], ")"}]}], "//", 
   "Simplify"}]}]], "Input",
 CellLabel->"In[420]:=",
 CellID->804196282,ExpressionUUID->"bd7a9a67-040b-4505-973d-6d2794fcf332"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   RowBox[{"\[ImaginaryI]", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["\[CapitalMu]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalMu],
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"], " ", 
    InterpretationBox[
     StyleBox["x",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`x[],
     Editable->False]}], "+", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["\[Alpha]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha],
      Editable->False], "3"], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["x",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`x[],
      Editable->False], "4"]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox["\[Alpha]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Alpha],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[CapitalMu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalMu],
       Editable->False], "-", 
      RowBox[{"3", " ", 
       InterpretationBox[
        StyleBox["\[CapitalMu]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalMu],
        Editable->False], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], "2"]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], "2"]}]}], ")"}]}], "-", 
   RowBox[{"\[ImaginaryI]", " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["\[Alpha]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha],
      Editable->False], "2"], " ", 
    InterpretationBox[
     StyleBox["x",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`x[],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[CapitalMu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalMu],
       Editable->False], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[CapitalMu]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalMu],
        Editable->False], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], "2"]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], "2"]}]}], ")"}]}]}], 
  RowBox[{"2", " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "\[ImaginaryI]"}], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["x",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x[],
       Editable->False]}]}], ")"}], " ", 
   SqrtBox[
    RowBox[{"2", "-", 
     RowBox[{"2", " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}]], " ", 
   SqrtBox[
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}]], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalMu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalMu],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}], ")"}]}]]], "Output",
 CellChangeTimes->{3.728677450617874*^9, 3.815707783010069*^9},
 CellLabel->"Out[420]=",
 CellID->1276841733,ExpressionUUID->"766ae90d-99f7-4068-8a9a-f842cd3153a4"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NP\[Alpha]", "=", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"-", "1"}], "/", "2"}], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"n", "[", "a", "]"}], 
       RowBox[{"mbar", "[", "b", "]"}], 
       RowBox[{
        RowBox[{"cd", "[", 
         RowBox[{"-", "b"}], "]"}], "[", 
        RowBox[{"l", "[", 
         RowBox[{"-", "a"}], "]"}], "]"}]}], "-", 
      RowBox[{
       RowBox[{"mbar", "[", "a", "]"}], 
       RowBox[{"mbar", "[", "b", "]"}], 
       RowBox[{
        RowBox[{"cd", "[", 
         RowBox[{"-", "b"}], "]"}], "[", 
        RowBox[{"m", "[", 
         RowBox[{"-", "a"}], "]"}], "]"}]}]}], ")"}]}], "//", 
   "Simplify"}]}]], "Input",
 CellLabel->"In[421]:=",
 CellID->359140818,ExpressionUUID->"16e8e266-d384-4c1c-9de1-6586e8560990"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   RowBox[{
    RowBox[{"-", "\[ImaginaryI]"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["\[CapitalMu]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalMu],
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"], " ", 
    InterpretationBox[
     StyleBox["x",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`x[],
     Editable->False]}], "-", 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["\[Alpha]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha],
      Editable->False], "3"], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["x",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`x[],
      Editable->False], "4"]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox["\[Alpha]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Alpha],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[CapitalMu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalMu],
       Editable->False], "-", 
      RowBox[{"3", " ", 
       InterpretationBox[
        StyleBox["\[CapitalMu]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalMu],
        Editable->False], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], "2"]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], "2"]}]}], ")"}]}], "+", 
   RowBox[{"\[ImaginaryI]", " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["\[Alpha]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha],
      Editable->False], "2"], " ", 
    InterpretationBox[
     StyleBox["x",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`x[],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[CapitalMu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalMu],
       Editable->False], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[CapitalMu]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalMu],
        Editable->False], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], "2"]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], "2"]}]}], ")"}]}]}], 
  RowBox[{"2", " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "\[ImaginaryI]"}], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["x",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x[],
       Editable->False]}]}], ")"}], " ", 
   SqrtBox[
    RowBox[{"2", "-", 
     RowBox[{"2", " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}]], " ", 
   SqrtBox[
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}]], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalMu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalMu],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}], ")"}]}]]], "Output",
 CellChangeTimes->{3.7286774507793283`*^9, 3.8157077842225313`*^9},
 CellLabel->"Out[421]=",
 CellID->907952202,ExpressionUUID->"3e5dd349-3f42-4ddb-8239-b53ae8666379"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["NP propagation equations (16):", "MathCaption",
 CellID->1785949246,ExpressionUUID->"ebc4aa06-a4b7-4e32-8f58-bc68068e795f"],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"NPDl", "[", "expr_", "]"}], ":=", 
   RowBox[{"Module", "[", 
    RowBox[{
     RowBox[{"{", "a", "}"}], ",", 
     RowBox[{
      RowBox[{"l", "[", "a", "]"}], 
      RowBox[{
       RowBox[{"cd", "[", 
        RowBox[{"-", "a"}], "]"}], "[", "expr", "]"}]}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"NPDn", "[", "expr_", "]"}], ":=", 
   RowBox[{"Module", "[", 
    RowBox[{
     RowBox[{"{", "a", "}"}], ",", 
     RowBox[{
      RowBox[{"n", "[", "a", "]"}], 
      RowBox[{
       RowBox[{"cd", "[", 
        RowBox[{"-", "a"}], "]"}], "[", "expr", "]"}]}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"NPDm", "[", "expr_", "]"}], ":=", 
   RowBox[{"Module", "[", 
    RowBox[{
     RowBox[{"{", "a", "}"}], ",", 
     RowBox[{
      RowBox[{"m", "[", "a", "]"}], 
      RowBox[{
       RowBox[{"cd", "[", 
        RowBox[{"-", "a"}], "]"}], "[", "expr", "]"}]}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"NPDb", "[", "expr_", "]"}], ":=", 
   RowBox[{"Module", "[", 
    RowBox[{
     RowBox[{"{", "a", "}"}], ",", 
     RowBox[{
      RowBox[{"mbar", "[", "a", "]"}], 
      RowBox[{
       RowBox[{"cd", "[", 
        RowBox[{"-", "a"}], "]"}], "[", "expr", "]"}]}]}], "]"}]}], 
  ";"}]}], "Input",
 CellLabel->"In[422]:=",
 CellID->1070043403,ExpressionUUID->"b27915b6-48e5-4f6b-a87d-39f3f9684be4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"NPDl", "[", 
    RowBox[{"l", "[", "a", "]"}], "]"}], "\[Equal]", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      RowBox[{"(", 
       RowBox[{"NP\[Epsilon]", "+", 
        RowBox[{"Dagger", "[", "NP\[Epsilon]", "]"}]}], ")"}]}], 
     RowBox[{"l", "[", "a", "]"}]}], "+", " ", 
    RowBox[{
     RowBox[{"Dagger", "[", "NP\[Kappa]", "]"}], 
     RowBox[{"m", "[", "a", "]"}]}], "+", " ", 
    RowBox[{"NP\[Kappa]", " ", 
     RowBox[{"mbar", "[", "a", "]"}]}]}]}], "//", "Simplify"}]], "Input",
 CellLabel->"In[426]:=",
 CellID->1191414243,ExpressionUUID->"1f7e4e33-fec1-4c69-a1e3-d1585e815388"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728677451130063*^9, 3.815707793323604*^9},
 CellLabel->"Out[426]=",
 CellID->743261221,ExpressionUUID->"c8149c53-0760-46be-b5cc-a2474f665173"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"NPDn", "[", 
    RowBox[{"l", "[", "a", "]"}], "]"}], "\[Equal]", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      RowBox[{"(", 
       RowBox[{"NP\[Gamma]", "+", 
        RowBox[{"Dagger", "[", "NP\[Gamma]", "]"}]}], ")"}]}], 
     RowBox[{"l", "[", "a", "]"}]}], "+", " ", 
    RowBox[{
     RowBox[{"Dagger", "[", "NP\[Tau]", "]"}], 
     RowBox[{"m", "[", "a", "]"}]}], "+", " ", 
    RowBox[{"NP\[Tau]", " ", 
     RowBox[{"mbar", "[", "a", "]"}]}]}]}], "//", "Simplify"}]], "Input",
 CellLabel->"In[427]:=",
 CellID->615221389,ExpressionUUID->"e29a9280-e192-4dd1-a9f2-ead9efa4f7ad"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728677451259626*^9, 3.815707794629033*^9},
 CellLabel->"Out[427]=",
 CellID->1955537703,ExpressionUUID->"6ad750e3-f647-41aa-bebc-47c89cd8f37e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"NPDm", "[", 
    RowBox[{"l", "[", "a", "]"}], "]"}], "\[Equal]", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      RowBox[{"(", 
       RowBox[{"NP\[Beta]", "+", 
        RowBox[{"Dagger", "[", "NP\[Alpha]", "]"}]}], ")"}]}], 
     RowBox[{"l", "[", "a", "]"}]}], "+", " ", 
    RowBox[{
     RowBox[{"Dagger", "[", "NP\[Rho]", "]"}], 
     RowBox[{"m", "[", "a", "]"}]}], "+", " ", 
    RowBox[{"NP\[Sigma]", " ", 
     RowBox[{"mbar", "[", "a", "]"}]}]}]}], "//", "Simplify"}]], "Input",
 CellLabel->"In[428]:=",
 CellID->1562136441,ExpressionUUID->"c834fd11-ed09-4d2b-ae04-5ba2fa78bc18"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728677451492898*^9, 3.815707795516983*^9},
 CellLabel->"Out[428]=",
 CellID->288538542,ExpressionUUID->"62f3504b-77e8-4834-a9f9-a2c4355ab30e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"NPDb", "[", 
    RowBox[{"l", "[", "a", "]"}], "]"}], "\[Equal]", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      RowBox[{"(", 
       RowBox[{"NP\[Alpha]", "+", 
        RowBox[{"Dagger", "[", "NP\[Beta]", "]"}]}], ")"}]}], 
     RowBox[{"l", "[", "a", "]"}]}], "+", " ", 
    RowBox[{
     RowBox[{"Dagger", "[", "NP\[Sigma]", "]"}], 
     RowBox[{"m", "[", "a", "]"}]}], "+", " ", 
    RowBox[{"NP\[Rho]", " ", 
     RowBox[{"mbar", "[", "a", "]"}]}]}]}], "//", "Simplify"}]], "Input",
 CellLabel->"In[429]:=",
 CellID->487430925,ExpressionUUID->"2ce2b122-f698-41c3-83a8-77bc5ddf3245"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728677451707449*^9, 3.815707796232148*^9},
 CellLabel->"Out[429]=",
 CellID->1822781056,ExpressionUUID->"f11db48b-e969-448e-81a0-e0ad282b7d71"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"NPDl", "[", 
    RowBox[{"n", "[", "a", "]"}], "]"}], "\[Equal]", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"NP\[Epsilon]", "+", 
       RowBox[{"Dagger", "[", "NP\[Epsilon]", "]"}]}], ")"}], 
     RowBox[{"n", "[", "a", "]"}]}], "+", 
    RowBox[{"NP\[Pi]", " ", 
     RowBox[{"m", "[", "a", "]"}]}], "+", 
    RowBox[{
     RowBox[{"Dagger", "[", "NP\[Pi]", "]"}], " ", 
     RowBox[{"mbar", "[", "a", "]"}]}]}]}], "//", "Simplify"}]], "Input",
 CellLabel->"In[430]:=",
 CellID->1604336830,ExpressionUUID->"db6bf27a-7e3d-4ec9-b228-5e86732e2351"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728677451878398*^9, 3.815707796949766*^9},
 CellLabel->"Out[430]=",
 CellID->290758316,ExpressionUUID->"95db6d1d-5afe-479c-aaa8-ba8b7e041bf0"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"NPDn", "[", 
    RowBox[{"n", "[", "a", "]"}], "]"}], "\[Equal]", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"NP\[Gamma]", "+", 
       RowBox[{"Dagger", "[", "NP\[Gamma]", "]"}]}], ")"}], 
     RowBox[{"n", "[", "a", "]"}]}], "+", " ", 
    RowBox[{"NP\[Nu]", " ", 
     RowBox[{"m", "[", "a", "]"}]}], "+", 
    RowBox[{
     RowBox[{"Dagger", "[", "NP\[Nu]", "]"}], " ", 
     RowBox[{"mbar", "[", "a", "]"}]}]}]}], "//", "Simplify"}]], "Input",
 CellLabel->"In[431]:=",
 CellID->117722724,ExpressionUUID->"a8b12a89-bee0-424c-bf56-f69d1ba58d3b"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728677452045485*^9, 3.815707797631378*^9},
 CellLabel->"Out[431]=",
 CellID->501336594,ExpressionUUID->"6503d0a2-1500-4fb8-a565-d72f6a5ce5d2"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"NPDm", "[", 
    RowBox[{"n", "[", "a", "]"}], "]"}], "\[Equal]", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"NP\[Beta]", "+", 
       RowBox[{"Dagger", "[", "NP\[Alpha]", "]"}]}], ")"}], 
     RowBox[{"n", "[", "a", "]"}]}], "+", 
    RowBox[{"NP\[Mu]", " ", 
     RowBox[{"m", "[", "a", "]"}]}], "+", 
    RowBox[{
     RowBox[{"Dagger", "[", "NP\[Lambda]", "]"}], " ", 
     RowBox[{"mbar", "[", "a", "]"}]}]}]}], "//", "Simplify"}]], "Input",
 CellLabel->"In[432]:=",
 CellID->1565834334,ExpressionUUID->"62b68547-4c94-47d7-943f-e2289a0258af"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728677452243724*^9, 3.8157077982908077`*^9},
 CellLabel->"Out[432]=",
 CellID->2021936104,ExpressionUUID->"9c119f2d-1770-477d-99de-c69cf4a67f99"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"NPDb", "[", 
    RowBox[{"n", "[", "a", "]"}], "]"}], "\[Equal]", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"NP\[Alpha]", "+", 
       RowBox[{"Dagger", "[", "NP\[Beta]", "]"}]}], ")"}], 
     RowBox[{"n", "[", "a", "]"}]}], "+", " ", 
    RowBox[{"NP\[Lambda]", " ", 
     RowBox[{"m", "[", "a", "]"}]}], "+", 
    RowBox[{
     RowBox[{"Dagger", "[", "NP\[Mu]", "]"}], " ", 
     RowBox[{"mbar", "[", "a", "]"}]}]}]}], "//", "Simplify"}]], "Input",
 CellLabel->"In[433]:=",
 CellID->1917676718,ExpressionUUID->"b3a76cd2-741b-45b5-99fa-bf593eece7e6"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.7286774524637814`*^9, 3.8157077989568367`*^9},
 CellLabel->"Out[433]=",
 CellID->298992140,ExpressionUUID->"f257c8d8-6bc5-4b8f-ba00-3a9ebb1e4f65"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"NPDl", "[", 
    RowBox[{"m", "[", "a", "]"}], "]"}], "\[Equal]", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      RowBox[{"(", 
       RowBox[{"NP\[Epsilon]", "-", 
        RowBox[{"Dagger", "[", "NP\[Epsilon]", "]"}]}], ")"}]}], 
     RowBox[{"m", "[", "a", "]"}]}], "+", 
    RowBox[{
     RowBox[{"Dagger", "[", "NP\[Pi]", "]"}], " ", 
     RowBox[{"l", "[", "a", "]"}]}], "+", 
    RowBox[{"NP\[Kappa]", " ", 
     RowBox[{"n", "[", "a", "]"}]}]}]}], "//", "Simplify"}]], "Input",
 CellLabel->"In[434]:=",
 CellID->1845475191,ExpressionUUID->"47ef814e-9dd7-4a53-806d-9d761f5bff62"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.7286774526855383`*^9, 3.815707799645241*^9},
 CellLabel->"Out[434]=",
 CellID->935771448,ExpressionUUID->"174117bc-c31a-45d0-8415-d4f20acbe32e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"NPDn", "[", 
    RowBox[{"m", "[", "a", "]"}], "]"}], "\[Equal]", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      RowBox[{"(", 
       RowBox[{"NP\[Gamma]", "-", 
        RowBox[{"Dagger", "[", "NP\[Gamma]", "]"}]}], ")"}]}], 
     RowBox[{"m", "[", "a", "]"}]}], "+", 
    RowBox[{
     RowBox[{"Dagger", "[", "NP\[Nu]", "]"}], " ", 
     RowBox[{"l", "[", "a", "]"}]}], "+", 
    RowBox[{"NP\[Tau]", " ", 
     RowBox[{"n", "[", "a", "]"}]}]}]}], "//", "Simplify"}]], "Input",
 CellLabel->"In[435]:=",
 CellID->1497284131,ExpressionUUID->"d3c59194-8532-4839-aff1-2d31171e4608"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.7286774528245068`*^9, 3.815707800268673*^9},
 CellLabel->"Out[435]=",
 CellID->429808897,ExpressionUUID->"a914edef-e0b8-4844-a0fa-51c36e91c68f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"NPDm", "[", 
    RowBox[{"m", "[", "a", "]"}], "]"}], "\[Equal]", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      RowBox[{"(", 
       RowBox[{"NP\[Beta]", "-", 
        RowBox[{"Dagger", "[", "NP\[Alpha]", "]"}]}], ")"}]}], 
     RowBox[{"m", "[", "a", "]"}]}], "+", 
    RowBox[{
     RowBox[{"Dagger", "[", "NP\[Lambda]", "]"}], 
     RowBox[{"l", "[", "a", "]"}]}], "+", 
    RowBox[{"NP\[Sigma]", " ", 
     RowBox[{"n", "[", "a", "]"}]}]}]}], "//", "Simplify"}]], "Input",
 CellLabel->"In[436]:=",
 CellID->1747198645,ExpressionUUID->"10486367-08ab-40ee-821e-e07bb923b90c"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728677453026083*^9, 3.815707800962681*^9},
 CellLabel->"Out[436]=",
 CellID->1438516804,ExpressionUUID->"70750d13-d204-45c9-91d4-d565ceb284cf"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"NPDb", "[", 
    RowBox[{"m", "[", "a", "]"}], "]"}], "\[Equal]", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      RowBox[{"(", 
       RowBox[{"NP\[Alpha]", "-", 
        RowBox[{"Dagger", "[", "NP\[Beta]", "]"}]}], ")"}]}], 
     RowBox[{"m", "[", "a", "]"}]}], "+", " ", 
    RowBox[{
     RowBox[{"Dagger", "[", "NP\[Mu]", "]"}], " ", 
     RowBox[{"l", "[", "a", "]"}]}], "+", 
    RowBox[{"NP\[Rho]", " ", 
     RowBox[{"n", "[", "a", "]"}]}]}]}], "//", "Simplify"}]], "Input",
 CellLabel->"In[437]:=",
 CellID->451051290,ExpressionUUID->"542aec4c-1c0a-4344-a58e-2dfe4245a8bc"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728677453268561*^9, 3.8157078016088743`*^9},
 CellLabel->"Out[437]=",
 CellID->1370229639,ExpressionUUID->"5accd619-92c6-4d2e-8a8f-42673047e055"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"NPDl", "[", 
    RowBox[{"mbar", "[", "a", "]"}], "]"}], "\[Equal]", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"NP\[Epsilon]", "-", 
       RowBox[{"Dagger", "[", "NP\[Epsilon]", "]"}]}], ")"}], 
     RowBox[{"mbar", "[", "a", "]"}]}], "+", 
    RowBox[{"NP\[Pi]", " ", 
     RowBox[{"l", "[", "a", "]"}]}], "+", 
    RowBox[{
     RowBox[{"Dagger", "[", "NP\[Kappa]", "]"}], " ", 
     RowBox[{"n", "[", "a", "]"}]}]}]}], "//", "Simplify"}]], "Input",
 CellLabel->"In[438]:=",
 CellID->55237227,ExpressionUUID->"823c2131-b013-4e7c-a27c-4dd387a7292a"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728677453625033*^9, 3.815707802258973*^9},
 CellLabel->"Out[438]=",
 CellID->838171297,ExpressionUUID->"09bd042e-561c-4b81-9a05-bad8a4092ccd"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"NPDn", "[", 
    RowBox[{"mbar", "[", "a", "]"}], "]"}], "\[Equal]", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"NP\[Gamma]", "-", 
       RowBox[{"Dagger", "[", "NP\[Gamma]", "]"}]}], ")"}], 
     RowBox[{"mbar", "[", "a", "]"}]}], "+", 
    RowBox[{"NP\[Nu]", " ", 
     RowBox[{"l", "[", "a", "]"}]}], "+", 
    RowBox[{
     RowBox[{"Dagger", "[", "NP\[Tau]", "]"}], " ", 
     RowBox[{"n", "[", "a", "]"}]}]}]}], "//", "Simplify"}]], "Input",
 CellLabel->"In[439]:=",
 CellID->1423227569,ExpressionUUID->"ff150099-be54-4dcf-a694-2bbcdb49b515"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.7286774538333187`*^9, 3.8157078028479767`*^9},
 CellLabel->"Out[439]=",
 CellID->817900928,ExpressionUUID->"f5c26b7a-39e2-4678-8b86-ed497009ed0b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"NPDm", "[", 
    RowBox[{"mbar", "[", "a", "]"}], "]"}], "\[Equal]", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"NP\[Beta]", "-", 
       RowBox[{"Dagger", "[", "NP\[Alpha]", "]"}]}], ")"}], 
     RowBox[{"mbar", "[", "a", "]"}]}], "+", 
    RowBox[{"NP\[Mu]", " ", 
     RowBox[{"l", "[", "a", "]"}]}], "+", 
    RowBox[{
     RowBox[{"Dagger", "[", "NP\[Rho]", "]"}], " ", 
     RowBox[{"n", "[", "a", "]"}]}]}]}], "//", "Simplify"}]], "Input",
 CellLabel->"In[440]:=",
 CellID->967732399,ExpressionUUID->"63c71ee8-7276-408d-ad18-7da6076a1fa1"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.7286774544681587`*^9, 3.815707803527749*^9},
 CellLabel->"Out[440]=",
 CellID->1905697441,ExpressionUUID->"ea797666-9f84-49fb-b750-c36e7c6bf80a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"NPDb", "[", 
    RowBox[{"mbar", "[", "a", "]"}], "]"}], "\[Equal]", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"NP\[Alpha]", "-", 
       RowBox[{"Dagger", "[", "NP\[Beta]", "]"}]}], ")"}], 
     RowBox[{"mbar", "[", "a", "]"}]}], "+", 
    RowBox[{"NP\[Lambda]", " ", 
     RowBox[{"l", "[", "a", "]"}]}], "+", 
    RowBox[{
     RowBox[{"Dagger", "[", "NP\[Sigma]", "]"}], " ", 
     RowBox[{"n", "[", "a", "]"}]}]}]}], "//", "Simplify"}]], "Input",
 CellLabel->"In[441]:=",
 CellID->599019138,ExpressionUUID->"1a125594-193a-4c5f-9336-1e5f124fa8b5"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.7286774550621634`*^9, 3.81570780421533*^9},
 CellLabel->"Out[441]=",
 CellID->2115293610,ExpressionUUID->"83c4e491-cfdd-4ae8-90f0-6cb0aaf38251"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"weyl", "=", 
   RowBox[{"Weyl", "[", "cd", "]"}]}], ";"}]], "Input",
 CellLabel->"In[442]:=",
 CellID->1004906518,ExpressionUUID->"db10e987-2240-41a4-a582-31f187340eda"],

Cell[BoxData[
 RowBox[{
  RowBox[{"$LargeComponentSize", "=", "1000"}], ";"}]], "Input",
 CellLabel->"In[443]:=",
 CellID->1652448018,ExpressionUUID->"2539f084-af8c-488d-9871-8ce6571e391f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"weyl", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}], ",", 
   RowBox[{"-", "d"}]}], "]"}]], "Input",
 CellLabel->"In[444]:=",
 CellID->1435315522,ExpressionUUID->"56887868-e2ef-4169-9fe5-89172be5ebea"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "1", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 1, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}], "+", 
                    RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                  RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    "3"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {1, 2, 1, 
                2} -> $CellContext`\[CapitalMu] $CellContext`r[] \
($CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (-$CellContext`r[]^2 + 
                 3 $CellContext`\[Alpha]^2 $CellContext`x[]^2) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", "\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              Annotation[#, {1, 2, 1, 3} -> Complex[0, 
                  Rational[
                  3, 2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] ($CellContext`r[] + 
                  Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
                  1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {1, 2, 1, 4} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}], "+", 
                    RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    "3"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {1, 2, 2, 
                1} -> -$CellContext`\[CapitalMu] $CellContext`r[] \
($CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (-$CellContext`r[]^2 + 
                 3 $CellContext`\[Alpha]^2 $CellContext`x[]^2) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {1, 2, 2, 3} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", "\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              Annotation[#, {1, 2, 2, 4} -> Complex[0, 
                  Rational[
                  3, 2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] ($CellContext`r[] + 
                  Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
                  1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", "\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              Annotation[#, {1, 2, 3, 1} -> Complex[0, 
                  
                  Rational[-3, 
                   2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] ($CellContext`r[] + 
                  Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
                  1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {1, 2, 3, 2} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "3"}], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                  RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    "3"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {1, 2, 3, 4} -> 
               Complex[
                 0, 1] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
$CellContext`x[] ((-3) $CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2) ($CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2)^(-3) ((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {1, 2, 4, 1} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", "\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              Annotation[#, {1, 2, 4, 2} -> Complex[0, 
                  
                  Rational[-3, 
                   2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] ($CellContext`r[] + 
                  Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
                  1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "3"}], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    "3"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {1, 2, 4, 3} -> 
               Complex[
                 0, -1] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
$CellContext`x[] ((-3) $CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2) ($CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2)^(-3) ((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "2", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 2, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", "\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              Annotation[#, {1, 3, 1, 2} -> Complex[0, 
                  Rational[
                  3, 2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] ($CellContext`r[] + 
                  Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
                  1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "1"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {1, 3, 1, 3} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
                  Complex[
                    0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", "\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              Annotation[#, {1, 3, 2, 1} -> Complex[0, 
                  
                  Rational[-3, 
                   2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] ($CellContext`r[] + 
                  Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
                  1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {1, 3, 2, 4} -> 
               Rational[
                 1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
                  Complex[
                    0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "1"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {1, 3, 3, 1} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
                  Complex[
                    0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "\[ImaginaryI]"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {1, 3, 3, 4} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {1, 3, 4, 2} -> 
               Rational[-1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
                  Complex[
                    0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "\[ImaginaryI]"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {1, 3, 4, 3} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "3", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 3, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {1, 4, 1, 2} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "1"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "-", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {1, 4, 1, 4} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
                  Complex[
                    0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              Annotation[#, {1, 4, 2, 1} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "-", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {1, 4, 2, 3} -> 
               Rational[
                 1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
                  Complex[
                    0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "-", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {1, 4, 3, 2} -> 
               Rational[-1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
                  Complex[
                    0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {1, 4, 3, 4} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "1"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "-", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {1, 4, 4, 1} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
                  Complex[
                    0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {1, 4, 4, 3} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"1", ",", "4", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {1, 4, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}], "+", 
                    RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    "3"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {2, 1, 1, 
                2} -> -$CellContext`\[CapitalMu] $CellContext`r[] \
($CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (-$CellContext`r[]^2 + 
                 3 $CellContext`\[Alpha]^2 $CellContext`x[]^2) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", "\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              Annotation[#, {2, 1, 1, 3} -> Complex[0, 
                  
                  Rational[-3, 
                   2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] ($CellContext`r[] + 
                  Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
                  1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {2, 1, 1, 4} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}], "+", 
                    RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                  RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    "3"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {2, 1, 2, 
                1} -> $CellContext`\[CapitalMu] $CellContext`r[] \
($CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (-$CellContext`r[]^2 + 
                 3 $CellContext`\[Alpha]^2 $CellContext`x[]^2) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {2, 1, 2, 3} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", "\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              Annotation[#, {2, 1, 2, 4} -> Complex[0, 
                  
                  Rational[-3, 
                   2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] ($CellContext`r[] + 
                  Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
                  1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", "\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              Annotation[#, {2, 1, 3, 1} -> Complex[0, 
                  Rational[
                  3, 2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] ($CellContext`r[] + 
                  Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
                  1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {2, 1, 3, 2} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "3"}], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    "3"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {2, 1, 3, 4} -> 
               Complex[
                 0, -1] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
$CellContext`x[] ((-3) $CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2) ($CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2)^(-3) ((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {2, 1, 4, 1} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", "\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              Annotation[#, {2, 1, 4, 2} -> Complex[0, 
                  Rational[
                  3, 2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] ($CellContext`r[] + 
                  Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
                  1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "3"}], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                  RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    "3"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {2, 1, 4, 3} -> 
               Complex[
                 0, 1] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
$CellContext`x[] ((-3) $CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2) ($CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2)^(-3) ((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "1", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 1, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "2", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 2, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {2, 3, 1, 2} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "-", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {2, 3, 1, 4} -> 
               Rational[
                 1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
                  Complex[
                    0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {2, 3, 2, 1} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "1"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "-", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {2, 3, 2, 3} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
                  Complex[
                    0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "1"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "-", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {2, 3, 3, 2} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
                  Complex[
                    0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {2, 3, 3, 4} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "-", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {2, 3, 4, 1} -> 
               Rational[-1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
                  Complex[
                    0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {2, 3, 4, 3} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "3", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 3, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", "\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              Annotation[#, {2, 4, 1, 2} -> Complex[0, 
                  Rational[
                  3, 2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] ($CellContext`r[] + 
                  Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
                  1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {2, 4, 1, 3} -> 
               Rational[
                 1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
                  Complex[
                    0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", "\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              Annotation[#, {2, 4, 2, 1} -> Complex[0, 
                  
                  Rational[-3, 
                   2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] ($CellContext`r[] + 
                  Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
                  1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "1"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {2, 4, 2, 4} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
                  Complex[
                    0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {2, 4, 3, 1} -> 
               Rational[-1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
                  Complex[
                    0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "\[ImaginaryI]"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {2, 4, 3, 4} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "1"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {2, 4, 4, 2} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
                  Complex[
                    0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "\[ImaginaryI]"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {2, 4, 4, 3} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"2", ",", "4", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {2, 4, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", "\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              Annotation[#, {3, 1, 1, 2} -> Complex[0, 
                  
                  Rational[-3, 
                   2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] ($CellContext`r[] + 
                  Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
                  1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "1"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {3, 1, 1, 3} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
                  Complex[
                    0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", "\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              Annotation[#, {3, 1, 2, 1} -> Complex[0, 
                  Rational[
                  3, 2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] ($CellContext`r[] + 
                  Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
                  1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {3, 1, 2, 4} -> 
               Rational[-1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
                  Complex[
                    0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "1"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {3, 1, 3, 1} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
                  Complex[
                    0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "\[ImaginaryI]"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {3, 1, 3, 4} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {3, 1, 4, 2} -> 
               Rational[
                 1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
                  Complex[
                    0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "\[ImaginaryI]"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {3, 1, 4, 3} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "1", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 1, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {3, 2, 1, 2} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "-", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {3, 2, 1, 4} -> 
               Rational[-1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
                  Complex[
                    0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {3, 2, 2, 1} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "1"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "-", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {3, 2, 2, 3} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
                  Complex[
                    0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "1"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "-", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {3, 2, 3, 2} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + Complex[
                    0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {3, 2, 3, 4} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "-", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {3, 2, 4, 1} -> 
               Rational[
                 1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
                  Complex[
                    0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {3, 2, 4, 3} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "2", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 2, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "3", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 3, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "3"}], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                  RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    "3"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {3, 4, 1, 2} -> 
               Complex[
                 0, 1] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
$CellContext`x[] ((-3) $CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2) ($CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2)^(-3) ((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "\[ImaginaryI]"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {3, 4, 1, 3} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {3, 4, 1, 4} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "3"}], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    "3"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {3, 4, 2, 1} -> 
               Complex[
                 0, -1] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
$CellContext`x[] ((-3) $CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2) ($CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2)^(-3) ((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {3, 4, 2, 3} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "\[ImaginaryI]"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {3, 4, 2, 4} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "\[ImaginaryI]"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {3, 4, 3, 1} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {3, 4, 3, 2} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}], "+", 
                    RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                  RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    "3"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {3, 4, 3, 
                4} -> $CellContext`\[CapitalMu] $CellContext`r[] \
($CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (-$CellContext`r[]^2 + 
                 3 $CellContext`\[Alpha]^2 $CellContext`x[]^2) (
                 2 (
                   2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {3, 4, 4, 1} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "\[ImaginaryI]"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {3, 4, 4, 2} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}], "+", 
                    RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    "3"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {3, 4, 4, 
                3} -> -$CellContext`\[CapitalMu] $CellContext`r[] \
($CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (-$CellContext`r[]^2 + 
                 3 $CellContext`\[Alpha]^2 $CellContext`x[]^2) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"3", ",", "4", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {3, 4, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}]},
         {GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {4, 1, 1, 2} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "1"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "-", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {4, 1, 1, 4} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
                  Complex[
                    0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {4, 1, 2, 1} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "-", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {4, 1, 2, 3} -> 
               Rational[-1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
                  Complex[
                    0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "-", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {4, 1, 3, 2} -> 
               Rational[
                 1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
                  Complex[
                    0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {4, 1, 3, 4} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "1"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "-", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {4, 1, 4, 1} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + Complex[
                    0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {4, 1, 4, 3} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "1", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 1, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", "\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              Annotation[#, {4, 2, 1, 2} -> Complex[0, 
                  
                  Rational[-3, 
                   2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] ($CellContext`r[] + 
                  Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
                  1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {4, 2, 1, 3} -> 
               Rational[-1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
                  Complex[
                    0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", "\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              Annotation[#, {4, 2, 2, 1} -> Complex[0, 
                  Rational[
                  3, 2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] ($CellContext`r[] + 
                  Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
                  1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "1"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {4, 2, 2, 4} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
                  Complex[
                    0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {4, 2, 3, 1} -> 
               Rational[
                 1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
                  Complex[
                    0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "\[ImaginaryI]"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox["x", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {4, 2, 3, 4} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "1"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "+", 
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {4, 2, 4, 2} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
                  Complex[
                    0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "\[ImaginaryI]"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {4, 2, 4, 3} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "2", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 2, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "3"}], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    "3"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {4, 3, 1, 2} -> 
               Complex[
                 0, -1] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
$CellContext`x[] ((-3) $CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2) ($CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2)^(-3) ((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "\[ImaginaryI]"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {4, 3, 1, 3} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {4, 3, 1, 4} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "3"}], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                  RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    "3"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {4, 3, 2, 1} -> 
               Complex[
                 0, 1] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
$CellContext`x[] ((-3) $CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2) ($CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2)^(-3) ((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {4, 3, 2, 3} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "\[ImaginaryI]"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {4, 3, 2, 4} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "\[ImaginaryI]"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {4, 3, 3, 1} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {4, 3, 3, 2} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}], "+", 
                    RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    "3"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox["r", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {4, 3, 3, 
                4} -> -$CellContext`\[CapitalMu] $CellContext`r[] \
($CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (-$CellContext`r[]^2 + 
                 3 $CellContext`\[Alpha]^2 $CellContext`x[]^2) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", 
                 RowBox[{"-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                    RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"\[ImaginaryI]", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]}]],
              
              Annotation[#, {4, 3, 4, 1} -> 
               Rational[-3, 
                  2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, 1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    SqrtBox[
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}]], " ", 
                    SqrtBox[
                    RowBox[{"1", "-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]]}], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "\[ImaginaryI]"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False]}]}], ")"}], "3"], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {4, 3, 4, 2} -> 
               Rational[
                 3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
                 Rational[1, 2] (
                  Complex[
                    0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
                 Rational[
                  1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["\<\"\[GrayCircle]\"\>",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", 
                 FractionBox[
                  RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}], "+", 
                    RowBox[{"3", " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}], ")"}]}]}], 
                    ")"}]}], 
                  RowBox[{
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], ")"}], 
                    "3"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[CapitalMu]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[CapitalMu], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], "+", 
                    RowBox[{
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"\[Alpha]"}], AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`\[Alpha], Editable -> 
                    False], "2"], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "x", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`x[], Editable -> False], "2"]}]}], 
                    ")"}]}]]}]],
              
              Annotation[#, {4, 3, 4, 
                3} -> $CellContext`\[CapitalMu] $CellContext`r[] \
($CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (-$CellContext`r[]^2 + 
                 3 $CellContext`\[Alpha]^2 $CellContext`x[]^2) (
                 2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) \
$CellContext`r[] + $CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
               "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "3", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 3, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}], GridBox[{
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "1", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 1, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "1", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 1, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "1", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 1, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "1", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 1, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "2", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 2, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "2", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 2, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "2", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 2, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "2", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 2, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "3", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 3, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "3", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 3, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "3", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 3, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "3", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 3, 4} -> 0, "Tooltip"]& ]},
            {
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "4", ",", "1"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 4, 1} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "4", ",", "2"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 4, 2} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "4", ",", "3"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 4, 3} -> 0, "Tooltip"]& ], 
             TagBox[
              TooltipBox["0",
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"4", ",", "4", ",", "4", ",", "4"}], "}"}], 
                 "\[Rule]", "0"}]],
              Annotation[#, {4, 4, 4, 4} -> 0, "Tooltip"]& ]}
           },
           AutoDelete->False,
           
           GridBoxDividers->{
            "ColumnsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}, 
             "RowsIndexed" -> {
              1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]},
           "RowsIndexed" -> {1 -> RGBColor[0, 0, 1], 5 -> RGBColor[0, 0, 1]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " ", " ", " "},
         {"a", "b", "c", "d"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{{{0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 
     0}}, {{0, $CellContext`\[CapitalMu] $CellContext`r[] ($CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-3) ((-2) $CellContext`\
\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2)^(-1) (-$CellContext`r[]^2 + 
        3 $CellContext`\[Alpha]^2 $CellContext`x[]^2) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), Complex[0, 
         Rational[
         3, 2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] ($CellContext`r[] + 
         Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
         1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
       Rational[-3, 
         2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)}, \
{-$CellContext`\[CapitalMu] $CellContext`r[] ($CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-3) ((-2) $CellContext`\
\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2)^(-1) (-$CellContext`r[]^2 + 
        3 $CellContext`\[Alpha]^2 $CellContext`x[]^2) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 0, 
       Rational[-3, 
         2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), Complex[0, 
         Rational[
         3, 2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] ($CellContext`r[] + 
         Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
         1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)}, {Complex[0, 
         Rational[-3, 
          2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] ($CellContext`r[] + 
         Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
         1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
       Rational[
        3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0, 
       Complex[0, 
         1] $CellContext`\[Alpha] $CellContext`\[CapitalMu] $CellContext`x[] \
((-3) $CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2) \
($CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2))}, {
      Rational[3, 
         2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 
          2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), Complex[0, 
         Rational[-3, 
          2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] ($CellContext`r[] + 
         Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
         1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
       Complex[0, -1] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
$CellContext`x[] ((-3) $CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2) ($CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2)^(-3) ((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 0}}, {{0, Complex[0, 
         Rational[
         3, 2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] ($CellContext`r[] + 
         Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
         1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
       Rational[
        3, 2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
         Complex[0, 
            1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0}, {
      Complex[0, 
         Rational[-3, 
          2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] ($CellContext`r[] + 
         Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
         1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0, 0, 
       Rational[
        1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
         Complex[0, 
            1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2))}, {
      Rational[-3, 
         2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
         Complex[0, 
            1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0, 0, 
       Rational[
        3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)}, {
      0, Rational[-1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
         Complex[0, 
            1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
       Rational[-3, 
         2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0}}, {{
      0, Rational[-3, 
         2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0, 
       Rational[
        3, 2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
         Complex[0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)}, {
      Rational[3, 
         2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0, 
       Rational[
        1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
         Complex[0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 0}, {
      0, Rational[-1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
         Complex[
           0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 0, 
       Rational[-3, 
         2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)}, {
      Rational[-3, 
         2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
         Complex[0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0, 
       Rational[
        3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0}}}, {{{
      0, -$CellContext`\[CapitalMu] $CellContext`r[] ($CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-3) ((-2) $CellContext`\
\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2)^(-1) (-$CellContext`r[]^2 + 
        3 $CellContext`\[Alpha]^2 $CellContext`x[]^2) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), Complex[0, 
         Rational[-3, 
          2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] ($CellContext`r[] + 
         Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
         1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
       Rational[
        3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)}, \
{$CellContext`\[CapitalMu] $CellContext`r[] ($CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-3) ((-2) $CellContext`\
\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2)^(-1) (-$CellContext`r[]^2 + 
        3 $CellContext`\[Alpha]^2 $CellContext`x[]^2) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 0, 
       Rational[
        3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), Complex[0, 
         Rational[-3, 
          2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] ($CellContext`r[] + 
         Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
         1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)}, {Complex[0, 
         Rational[
         3, 2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] ($CellContext`r[] + 
         Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
         1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
       Rational[-3, 
         2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0, 
       Complex[0, -1] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
$CellContext`x[] ((-3) $CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2) ($CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2)^(-3) ((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2))}, {
      Rational[-3, 
         2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), Complex[0, 
         Rational[
         3, 2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] ($CellContext`r[] + 
         Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
         1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
       Complex[0, 
         1] $CellContext`\[Alpha] $CellContext`\[CapitalMu] $CellContext`x[] \
((-3) $CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2) \
($CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 0}}, {{0, 0, 0, 0}, {0, 0,
      0, 0}, {0, 0, 0, 0}, {0, 0, 0, 
     0}}, {{0, 
       Rational[-3, 
         2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0, 
       Rational[
        1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
         Complex[0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2))}, {
      Rational[
        3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0, 
       Rational[
        3, 2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
         Complex[0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0}, {
      0, Rational[-3, 
         2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
         Complex[0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0, 
       Rational[-3, 
         2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)}, {
      Rational[-1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
         Complex[0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 0, 
       Rational[
        3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0}}, {{
      0, Complex[0, 
         Rational[
         3, 2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] ($CellContext`r[] + 
         Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
         1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
       Rational[
        1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
         Complex[0, 
            1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 0}, {Complex[0, 
         Rational[-3, 
          2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] ($CellContext`r[] + 
         Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
         1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0, 0, 
       Rational[
        3, 2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
         Complex[0, 
            1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)}, {
      Rational[-1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
         Complex[0, 
            1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 0, 0, 
       Rational[
        3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)}, {
      0, Rational[-3, 
         2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
         Complex[0, 
            1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
       Rational[-3, 
         2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0}}}, {{{
      0, Complex[0, 
         Rational[-3, 
          2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] ($CellContext`r[] + 
         Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
         1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
       Rational[-3, 
         2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
         Complex[0, 
            1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0}, {
      Complex[0, 
         Rational[
         3, 2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] ($CellContext`r[] + 
         Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
         1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0, 0, 
       Rational[-1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
         Complex[0, 
            1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2))}, {
      Rational[3, 
         2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
         Complex[0, 
            1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0, 0, 
       Rational[-3, 
         2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)}, {
      0, Rational[
        1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
         Complex[0, 
            1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
       Rational[
        3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0}}, {{
      0, Rational[
        3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0, 
       Rational[-1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
         Complex[0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2))}, {
      Rational[-3, 
         2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0, 
       Rational[-3, 
         2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
         Complex[0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0}, {
      0, Rational[
        3, 2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
         Complex[0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0, 
       Rational[
        3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)}, {
      Rational[1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
         Complex[0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 0, 
       Rational[-3, 
         2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0}}, {{0, 0,
      0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 
     0}}, {{0, 
       Complex[0, 
         1] $CellContext`\[Alpha] $CellContext`\[CapitalMu] $CellContext`x[] \
((-3) $CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2) \
($CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
       Rational[
        3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
       Rational[-3, 
         2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)}, {
      Complex[0, -1] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
$CellContext`x[] ((-3) $CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2) ($CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2)^(-3) ((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 0, 
       Rational[-3, 
         2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
       Rational[
        3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)}, {
      Rational[-3, 
         2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
       Rational[
        3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
       0, $CellContext`\[CapitalMu] $CellContext`r[] ($CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-3) ((-2) $CellContext`\
\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2)^(-1) (-$CellContext`r[]^2 + 
        3 $CellContext`\[Alpha]^2 $CellContext`x[]^2) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2))}, {
      Rational[3, 
         2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
       Rational[-3, 
         2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), \
-$CellContext`\[CapitalMu] $CellContext`r[] ($CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-3) ((-2) $CellContext`\
\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2)^(-1) (-$CellContext`r[]^2 + 
        3 $CellContext`\[Alpha]^2 $CellContext`x[]^2) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 0}}}, {{{
      0, Rational[
        3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0, 
       Rational[-3, 
         2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
         Complex[0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)}, {
      Rational[-3, 
         2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0, 
       Rational[-1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
         Complex[0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 0}, {
      0, Rational[
        1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
         Complex[0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 0, 
       Rational[
        3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)}, {
      Rational[
        3, 2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
         Complex[0, -1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0, 
       Rational[-3, 
         2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0}}, {{
      0, Complex[0, 
         Rational[-3, 
          2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] ($CellContext`r[] + 
         Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
         1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
       Rational[-1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
         Complex[0, 
            1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 0}, {Complex[0, 
         Rational[
         3, 2]] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
($CellContext`\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] ($CellContext`r[] + 
         Complex[0, 1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (
         1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0, 0, 
       Rational[-3, 
         2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
         Complex[0, 
            1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)}, {
      Rational[1, 2] $CellContext`\[CapitalMu] ($CellContext`r[] + 
         Complex[0, 
            1] $CellContext`\[Alpha] $CellContext`x[])^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 0, 0, 
       Rational[-3, 
         2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)}, {
      0, Rational[
        3, 2] $CellContext`\[Alpha]^2 $CellContext`\[CapitalMu] \
($CellContext`r[] + 
         Complex[0, 
            1] $CellContext`\[Alpha] $CellContext`x[])^(-3) (-1 + \
$CellContext`x[]^2)/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
       Rational[
        3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 0}}, {{
      0, Complex[
        0, -1] $CellContext`\[Alpha] $CellContext`\[CapitalMu] \
$CellContext`x[] ((-3) $CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2) ($CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2)^(-3) ((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 
       Rational[-3, 
         2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
       Rational[
        3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)}, {
      Complex[0, 
         1] $CellContext`\[Alpha] $CellContext`\[CapitalMu] $CellContext`x[] \
((-3) $CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2) \
($CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-3) ((-2) \
$CellContext`\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-1) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 0, 
       Rational[
        3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
       Rational[-3, 
         2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2)}, {
      Rational[3, 
         2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
       Rational[-3, 
         2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
       0, -$CellContext`\[CapitalMu] $CellContext`r[] ($CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-3) ((-2) $CellContext`\
\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2)^(-1) (-$CellContext`r[]^2 + 
        3 $CellContext`\[Alpha]^2 $CellContext`x[]^2) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2))}, {
      Rational[-3, 
         2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, 
            1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), 
       Rational[
        3, 2] $CellContext`\[Alpha] $CellContext`\[CapitalMu] ($CellContext`\
\[Alpha]^2 + $CellContext`r[] ((-2) $CellContext`\[CapitalMu] + \
$CellContext`r[]))^
        Rational[1, 2] (
         Complex[0, -1] $CellContext`r[] + $CellContext`\[Alpha] \
$CellContext`x[])^(-3) (1 - $CellContext`x[]^2)^
        Rational[
         1, 2]/((-2) $CellContext`\[CapitalMu] $CellContext`r[] + \
$CellContext`r[]^2 + $CellContext`\[Alpha]^2 $CellContext`x[]^2), \
$CellContext`\[CapitalMu] $CellContext`r[] ($CellContext`r[]^2 + \
$CellContext`\[Alpha]^2 $CellContext`x[]^2)^(-3) ((-2) $CellContext`\
\[CapitalMu] $CellContext`r[] + $CellContext`r[]^2 + $CellContext`\[Alpha]^2 \
$CellContext`x[]^2)^(-1) (-$CellContext`r[]^2 + 
        3 $CellContext`\[Alpha]^2 $CellContext`x[]^2) (
        2 (2 $CellContext`\[CapitalMu] - $CellContext`r[]) $CellContext`r[] + \
$CellContext`\[Alpha]^2 (-3 + $CellContext`x[]^2)), 0}}, {{0, 0, 0, 0}, {0, 0,
      0, 0}, {0, 0, 0, 0}, {0, 0, 0, 
     0}}}}, {-$CellContext`NP, -$CellContext`NP, -$CellContext`NP, \
-$CellContext`NP}, 
   0][-$CellContext`a, -$CellContext`b, -$CellContext`c, -$CellContext`d],
  Editable->False]], "Output",
 CellChangeTimes->{3.72867745592373*^9, 3.815707809072237*^9},
 CellLabel->"Out[444]=",
 CellID->1638369278,ExpressionUUID->"6f9c32fe-6212-4dd9-94d6-2bb2d9d79ce5"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Psi0", "=", 
  RowBox[{
   RowBox[{"weyl", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], 
   RowBox[{"l", "[", "a", "]"}], 
   RowBox[{"m", "[", "b", "]"}], 
   RowBox[{"l", "[", "c", "]"}], 
   RowBox[{"m", "[", "d", "]"}]}]}]], "Input",
 CellLabel->"In[445]:=",
 CellID->110453815,ExpressionUUID->"5916c837-983c-4870-bd7e-95e2e4002e8f"],

Cell[BoxData[
 FractionBox[
  RowBox[{"3", " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["\[Alpha]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Alpha],
     Editable->False], "2"], " ", 
   InterpretationBox[
    StyleBox["\[CapitalMu]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[CapitalMu],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["x",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x[],
       Editable->False], "2"]}], ")"}]}], 
  RowBox[{"2", " ", 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "+", 
      RowBox[{"\[ImaginaryI]", " ", 
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False]}]}], ")"}], "3"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalMu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalMu],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}], ")"}]}]]], "Output",
 CellChangeTimes->{3.728677457045207*^9, 3.815707811003642*^9},
 CellLabel->"Out[445]=",
 CellID->877458086,ExpressionUUID->"7d0b3902-8761-4a79-9806-ee061d02445b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Psi1", "=", 
  RowBox[{
   RowBox[{"weyl", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], 
   RowBox[{"l", "[", "a", "]"}], 
   RowBox[{"n", "[", "b", "]"}], 
   RowBox[{"l", "[", "c", "]"}], 
   RowBox[{"m", "[", "d", "]"}]}]}]], "Input",
 CellLabel->"In[446]:=",
 CellID->1110408985,ExpressionUUID->"59691ddb-3eed-4035-a3ec-874adec6c94c"],

Cell[BoxData[
 FractionBox[
  RowBox[{"3", " ", "\[ImaginaryI]", " ", 
   InterpretationBox[
    StyleBox["\[Alpha]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Alpha],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox["\[CapitalMu]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[CapitalMu],
    Editable->False], " ", 
   SqrtBox[
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha],
       Editable->False], "2"], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         InterpretationBox[
          StyleBox["\[CapitalMu]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalMu],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]}], ")"}]}]}]], " ", 
   SqrtBox[
    RowBox[{"1", "-", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["x",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x[],
       Editable->False], "2"]}]]}], 
  RowBox[{"2", " ", 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "+", 
      RowBox[{"\[ImaginaryI]", " ", 
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False]}]}], ")"}], "3"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalMu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalMu],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}], ")"}]}]]], "Output",
 CellChangeTimes->{3.728677458155011*^9, 3.815707812182891*^9},
 CellLabel->"Out[446]=",
 CellID->2002328293,ExpressionUUID->"08bef24e-29ad-48f2-aa08-5dd14b1921b0"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Psi2", "=", 
  RowBox[{
   RowBox[{"weyl", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], 
   RowBox[{"l", "[", "a", "]"}], 
   RowBox[{"m", "[", "b", "]"}], 
   RowBox[{"mbar", "[", "c", "]"}], 
   RowBox[{"n", "[", "d", "]"}]}]}]], "Input",
 CellLabel->"In[447]:=",
 CellID->1763509838,ExpressionUUID->"10833a3d-c788-4036-8155-3185fb38d308"],

Cell[BoxData[
 RowBox[{"-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["\[CapitalMu]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalMu],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["\[CapitalMu]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[CapitalMu],
           Editable->False]}], "-", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]}], ")"}], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], "+", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha],
         Editable->False], "2"], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "3"}], "+", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False], "2"]}], ")"}]}]}], ")"}]}], 
   RowBox[{"2", " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "+", 
       RowBox[{"\[ImaginaryI]", " ", 
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False]}]}], ")"}], "3"], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], " ", 
       InterpretationBox[
        StyleBox["\[CapitalMu]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalMu],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], "+", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"], "+", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha],
         Editable->False], "2"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], "2"]}]}], ")"}]}]]}]], "Output",
 CellChangeTimes->{3.728677459241023*^9, 3.815707814013479*^9},
 CellLabel->"Out[447]=",
 CellID->1372414586,ExpressionUUID->"719e2803-5b72-418f-b64f-7120d8d9290f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Psi3", "=", 
  RowBox[{
   RowBox[{"weyl", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], 
   RowBox[{"l", "[", "a", "]"}], 
   RowBox[{"n", "[", "b", "]"}], 
   RowBox[{"mbar", "[", "c", "]"}], 
   RowBox[{"n", "[", "d", "]"}]}]}]], "Input",
 CellLabel->"In[448]:=",
 CellID->986196261,ExpressionUUID->"6a9bc01d-fdbb-4f67-bf97-bcd0563823fa"],

Cell[BoxData[
 RowBox[{"-", 
  FractionBox[
   RowBox[{"3", " ", "\[ImaginaryI]", " ", 
    InterpretationBox[
     StyleBox["\[Alpha]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Alpha],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox["\[CapitalMu]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalMu],
     Editable->False], " ", 
    SqrtBox[
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "2"}], " ", 
          InterpretationBox[
           StyleBox["\[CapitalMu]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[CapitalMu],
           Editable->False]}], "+", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]}], ")"}]}]}]], " ", 
    SqrtBox[
     RowBox[{"1", "-", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]]}], 
   RowBox[{"2", " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "+", 
       RowBox[{"\[ImaginaryI]", " ", 
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False]}]}], ")"}], "3"], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], " ", 
       InterpretationBox[
        StyleBox["\[CapitalMu]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalMu],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], "+", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"], "+", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha],
         Editable->False], "2"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], "2"]}]}], ")"}]}]]}]], "Output",
 CellChangeTimes->{3.728677460240052*^9, 3.815707814914288*^9},
 CellLabel->"Out[448]=",
 CellID->198692634,ExpressionUUID->"93cbe16b-3184-4d5b-9bc8-a43904acb3fc"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Psi4", "=", 
  RowBox[{
   RowBox[{"weyl", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], 
   RowBox[{"n", "[", "a", "]"}], 
   RowBox[{"mbar", "[", "b", "]"}], 
   RowBox[{"n", "[", "c", "]"}], 
   RowBox[{"mbar", "[", "d", "]"}]}]}]], "Input",
 CellLabel->"In[449]:=",
 CellID->1818920111,ExpressionUUID->"b40f221d-04e4-4320-93de-32d30edb8399"],

Cell[BoxData[
 FractionBox[
  RowBox[{"3", " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["\[Alpha]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Alpha],
     Editable->False], "2"], " ", 
   InterpretationBox[
    StyleBox["\[CapitalMu]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[CapitalMu],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["x",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x[],
       Editable->False], "2"]}], ")"}]}], 
  RowBox[{"2", " ", 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "+", 
      RowBox[{"\[ImaginaryI]", " ", 
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False]}]}], ")"}], "3"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalMu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalMu],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "2"]}]}], ")"}]}]]], "Output",
 CellChangeTimes->{3.728677461230929*^9, 3.815707815719355*^9},
 CellLabel->"Out[449]=",
 CellID->1383135580,ExpressionUUID->"0f5181cb-8bff-467b-8638-a82c1a854829"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"UndefBasis", "[", "NP", "]"}]], "Input",
 CellLabel->"In[450]:=",
 CellID->451105623,ExpressionUUID->"dff9e2d2-a4ee-4bf4-afd1-d3e5fabb0129"],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{
  "met", ",", "cd", ",", "cdRiemann", ",", "weyl", ",", "l", ",", "n", ",", 
   "m", ",", "mbar", ",", "expr", ",", "\"\<Global`NP*\>\"", ",", 
   "\"\<Global`Psi*\>\""}], "]"}]], "Input",
 CellChangeTimes->{{3.815707889203801*^9, 3.815707912131452*^9}},
 CellLabel->"In[451]:=",
 CellID->1114931851,ExpressionUUID->"606845a6-7d80-4613-addf-1f699a6dde97"],

Cell[BoxData[
 RowBox[{"UndefConstantSymbol", "[", 
  RowBox[{"{", 
   RowBox[{"\[Alpha]", ",", "\[CapitalMu]"}], "}"}], "]"}]], "Input",
 CellLabel->"In[452]:=",
 CellID->1864799682,ExpressionUUID->"24589f80-daac-426f-a834-526803b6a9d0"],

Cell[BoxData[
 RowBox[{"UndefChart", "[", "BL", "]"}]], "Input",
 CellLabel->"In[453]:=",
 CellID->1311911056,ExpressionUUID->"ded61440-2e85-46ea-944c-d1f4981bf859"],

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M", "]"}]], "Input",
 CellLabel->"In[454]:=",
 CellID->1074514443,ExpressionUUID->"a32d4cf1-30c1-48f4-a1f0-47a4d17ecac4"],

Cell[BoxData[
 RowBox[{"ClearxCobaCache", "[", "All", "]"}]], "Input",
 CellLabel->"In[455]:=",
 CellID->1174255878,ExpressionUUID->"d8229c44-ffef-4f14-a77d-89f85f694a6c"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.7. Extrinsic curvature computations do not yet work. TODO", \
"Subsection",
 CellChangeTimes->{{3.7286740381774397`*^9, 3.728674041505783*^9}, {
   3.728675279743945*^9, 3.7286752802320137`*^9}, {3.728677480572613*^9, 
   3.7286774831690197`*^9}, 3.790876741567401*^9, 3.794517818091647*^9, {
   3.79452092816298*^9, 3.794520928642911*^9}},
 CellID->343825616,ExpressionUUID->"c268a446-9dcb-46d7-ba09-c8a81639130c"],

Cell[BoxData["Exit"], "Input",
 CellChangeTimes->{{3.72867754435923*^9, 3.728677544873953*^9}},
 CellLabel->"In[464]:=",
 CellID->865471439,ExpressionUUID->"d410987b-4f4d-4e09-8645-b7905bdb34c5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"<<", "xAct`xTensor`"}]], "Input",
 CellLabel->"In[1]:=",
 CellID->1946117682,ExpressionUUID->"6d6f1eee-7370-463d-b623-ecf7b3770279"],

Cell[CellGroupData[{

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.728677507225314*^9, 3.72867754814367*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1525917970,ExpressionUUID->"40af1e79-c0bb-409b-a594-2b574f2ced05"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`xPerm`  version \"\>", 
   "\[InvisibleSpace]", "\<\"1.2.3\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2015", ",", "8", ",", "23"}], "}"}]}],
  SequenceForm["Package xAct`xPerm`  version ", "1.2.3", ", ", {2015, 8, 23}],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.728677507225314*^9, 3.728677548169468*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->853609316,ExpressionUUID->"7ce648ba-c6ff-4875-90ea-966233199868"],

Cell[BoxData["\<\"CopyRight (C) 2003-2015, Jose M. Martin-Garcia, under the \
General Public License.\"\>"], "Print",
 CellChangeTimes->{3.728677507225314*^9, 3.728677548195059*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1421012052,ExpressionUUID->"bcfc8227-68a8-422d-afd6-b3043675caba"],

Cell[BoxData["\<\"Connecting to external mac executable...\"\>"], "Print",
 CellChangeTimes->{3.728677507225314*^9, 3.7286775482171288`*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->783919677,ExpressionUUID->"369c58fe-1940-4c91-b024-da91c1bcc9a4"],

Cell[BoxData["\<\"Connection established.\"\>"], "Print",
 CellChangeTimes->{3.728677507225314*^9, 3.728677548237307*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->525836994,ExpressionUUID->"dfa4586a-022f-4904-9b95-0c067ff906a3"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.728677507225314*^9, 3.728677548257061*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->847150753,ExpressionUUID->"3a7a15e2-d099-4895-be9e-b201842c50e7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`xTensor`  version \"\>", 
   "\[InvisibleSpace]", "\<\"1.1.3\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2018", ",", "2", ",", "28"}], "}"}]}],
  SequenceForm[
  "Package xAct`xTensor`  version ", "1.1.3", ", ", {2018, 2, 28}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728677507225314*^9, 3.728677548277197*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->246126061,ExpressionUUID->"e024280e-4d84-4e38-9be3-ed9202086296"],

Cell[BoxData["\<\"CopyRight (C) 2002-2018, Jose M. Martin-Garcia, under the \
General Public License.\"\>"], "Print",
 CellChangeTimes->{3.728677507225314*^9, 3.728677548296917*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->587203105,ExpressionUUID->"79ac6af8-5570-4331-b69f-e845b7acb2af"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.728677507225314*^9, 3.728677548317053*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1435227770,ExpressionUUID->"a095beae-91c5-4e6a-bd59-5957238024e4"],

Cell[BoxData["\<\"These packages come with ABSOLUTELY NO WARRANTY; for \
details type Disclaimer[]. This is free software, and you are welcome to \
redistribute it under certain conditions. See the General Public License for \
details.\"\>"], "Print",
 CellChangeTimes->{3.728677507225314*^9, 3.728677548337117*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1336807286,ExpressionUUID->"d6bd787b-fbee-4228-a189-a1efc7f09ac3"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.728677507225314*^9, 3.728677548355753*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->746140888,ExpressionUUID->"6aa337d0-ba87-4a91-bfaf-afeb45eebd23"]
}, Open  ]]
}, Open  ]],

Cell[BoxData[{
 RowBox[{
  RowBox[{"$PrePrint", "=", "ScreenDollarIndices"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"$DefInfoQ", "=", "False"}], ";"}]}], "Input",
 CellLabel->"In[2]:=",
 CellID->3730788,ExpressionUUID->"5d31a923-3016-47dd-8642-0c27c79f033e"],

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M", ",", "4", ",", 
   RowBox[{"{", 
    RowBox[{
    "\[Mu]", ",", "\[Nu]", ",", "\[Alpha]", ",", "\[Beta]", ",", "\[Gamma]", 
     ",", "\[Delta]", ",", "\[Lambda]", ",", "\[Rho]", ",", "\[Sigma]"}], 
    "}"}]}], "]"}]], "Input",
 CellLabel->"In[4]:=",
 CellID->426328153,ExpressionUUID->"e0d2ce62-4b0e-4009-b74d-3d9593bbabe1"],

Cell[BoxData[
 RowBox[{"DefMetric", "[", 
  RowBox[{
   RowBox[{"-", "1"}], ",", 
   RowBox[{"g", "[", 
    RowBox[{
     RowBox[{"-", "\[Mu]"}], ",", 
     RowBox[{"-", "\[Nu]"}]}], "]"}], ",", "CD"}], "]"}]], "Input",
 CellLabel->"In[5]:=",
 CellID->1291821079,ExpressionUUID->"7c8e66da-476b-44bb-a46b-0d239d6a8a90"],

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"u", "[", "\[Mu]", "]"}], ",", "M"}], "]"}]], "Input",
 CellLabel->"In[6]:=",
 CellID->577203583,ExpressionUUID->"0a82e2dc-6bf6-4ae5-997c-28246f283fd5"],

Cell[BoxData[
 RowBox[{"u", "/:", 
  RowBox[{
   RowBox[{"u", "[", "\[Mu]_", "]"}], 
   RowBox[{"u", "[", 
    RowBox[{"-", "\[Mu]_"}], "]"}]}], ":=", 
  RowBox[{"-", "1"}]}]], "Input",
 CellLabel->"In[7]:=",
 CellID->876985982,ExpressionUUID->"33d2c4ff-7978-4da6-ab9f-4e3b827ade3e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefMetric", "[", 
  RowBox[{"1", ",", 
   RowBox[{"h", "[", 
    RowBox[{
     RowBox[{"-", "\[Mu]"}], ",", 
     RowBox[{"-", "\[Nu]"}]}], "]"}], ",", "cd", ",", 
   RowBox[{"InducedFrom", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"g", ",", "u"}], "}"}]}]}], "]"}]], "Input",
 CellLabel->"In[8]:=",
 CellID->1323130113,ExpressionUUID->"eb0c5da7-ca7f-4f98-8f63-5e2c3642d4dd"],

Cell[BoxData[
 TemplateBox[{
  "DefMetric", "old", 
   "\"There are already metrics \\!\\(\\*RowBox[{\\\"{\\\", \\\"g\\\", \
\\\"}\\\"}]\\) in vbundle \\!\\(\\*InterpretationBox[StyleBox[\\\"\
\[DoubleStruckCapitalT]M\\\", Rule[AutoSpacing, False], Rule[ShowAutoStyles, \
False]], TangentM, Rule[Editable, False]]\\).\"", 2, 8, 1, 
   19251607901465226019, "Local", "xAct`xTensor`DefMetric"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.7286775249612503`*^9, 3.7286775550422287`*^9},
 CellLabel->"During evaluation of In[8]:=",
 CellID->172368675,ExpressionUUID->"04ab7dd0-624c-4395-bb73-7f31e0bb73e3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CD", "[", 
   RowBox[{"-", "\[Mu]"}], "]"}], "[", 
  RowBox[{"u", "[", 
   RowBox[{"-", "\[Nu]"}], "]"}], "]"}]], "Input",
 CellLabel->"In[9]:=",
 CellID->1051488761,ExpressionUUID->"6242f4c4-a431-4bc8-a05f-008e58296351"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[EmptyDownTriangle]", "\[Mu]"], GridBox[{
      {"u", 
       StyleBox[GridBox[{
          {" "},
          {"\[Nu]"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`CD[-$CellContext`\[Mu]][
   $CellContext`u[-$CellContext`\[Nu]]],
  Editable->False]], "Output",
 CellChangeTimes->{3.7286775563114147`*^9},
 CellLabel->"Out[9]=",
 CellID->1709710596,ExpressionUUID->"6a8b6b27-0fc3-4d1d-87f3-62c449a7ec76"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "GradNormalToExtrinsicK"}]], "Input",
 CellLabel->"In[10]:=",
 CellID->1888364754,ExpressionUUID->"7f5f9b5c-a115-4608-a694-34425188f96d"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"Kh", 
       StyleBox[GridBox[{
          {" ", " "},
          {"\[Mu]", "\[Nu]"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`ExtrinsicKh[-$CellContext`\[Mu], -$CellContext`\[Nu]],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Au", 
        StyleBox[GridBox[{
           {" "},
           {"\[Nu]"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationu[-$CellContext`\[Nu]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"u", 
        StyleBox[GridBox[{
           {" "},
           {"\[Mu]"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`u[-$CellContext`\[Mu]],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728677556918067*^9},
 CellLabel->"Out[10]=",
 CellID->11238503,ExpressionUUID->"f8464135-b460-4031-9574-506ce7af03d0"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"%", " ", 
   RowBox[{"u", "[", "\[Mu]", "]"}]}], "//", "Expand"}]], "Input",
 CellLabel->"In[11]:=",
 CellID->2037262412,ExpressionUUID->"0baccec6-4dd8-4cf3-aea2-7f28744d41e3"],

Cell[BoxData[
 RowBox[{"-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"Au", 
       StyleBox[GridBox[{
          {" "},
          {"\[Nu]"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Accelerationu[-$CellContext`\[Nu]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7286775575937567`*^9},
 CellLabel->"Out[11]=",
 CellID->420817408,ExpressionUUID->"f85078d8-6366-4db9-8cd0-2ce6ba752dba"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr", "=", 
  RowBox[{
   RowBox[{"GaussCodazzi", "[", 
    RowBox[{
     RowBox[{"RicciCD", "[", 
      RowBox[{
       RowBox[{"-", "\[Mu]"}], ",", 
       RowBox[{"-", "\[Nu]"}]}], "]"}], ",", "h"}], "]"}], "//", 
   "ProjectDerivative"}]}]], "Input",
 CellLabel->"In[12]:=",
 CellID->1163274411,ExpressionUUID->"f8006af2-b451-421c-8c4b-45c3c5b8a36c"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"Au", 
         StyleBox[GridBox[{
            {" "},
            {"\[Mu]"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Accelerationu[-$CellContext`\[Mu]],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Au", 
        StyleBox[GridBox[{
           {" "},
           {"\[Nu]"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationu[-$CellContext`\[Nu]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"\[Alpha]", "\[Nu]"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKh[-$CellContext`\[Alpha], -$CellContext`\[Nu]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "\[Alpha]"},
           {"\[Mu]", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKh[-$CellContext`\[Mu], $CellContext`\[Alpha]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "\[Alpha]"},
           {"\[Alpha]", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKh[-$CellContext`\[Alpha], $CellContext`\[Alpha]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"\[Mu]", "\[Nu]"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKh[-$CellContext`\[Mu], -$CellContext`\[Nu]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "\[Alpha]"},
           {"\[Mu]", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKh[-$CellContext`\[Mu], $CellContext`\[Alpha]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"\[Nu]", "\[Alpha]"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKh[-$CellContext`\[Nu], -$CellContext`\[Alpha]],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     UnderscriptBox["P", "h"], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"u", 
           StyleBox[GridBox[{
              {"\[Alpha]"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`u[$CellContext`\[Alpha]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", "\[Alpha]"], GridBox[{
            {"Kh", 
             StyleBox[GridBox[{
                {" ", " "},
                {"\[Mu]", "\[Nu]"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CD[-$CellContext`\[Alpha]][
         $CellContext`ExtrinsicKh[-$CellContext`\[Mu], -$CellContext`\[Nu]]],
        Editable->False], ")"}]}], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Projectorh[$CellContext`u[$CellContext`\[Alpha]] \
$CellContext`CD[-$CellContext`\[Alpha]][
      $CellContext`ExtrinsicKh[-$CellContext`\[Mu], -$CellContext`\[Nu]]]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     UnderscriptBox["P", "h"], "[", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", "\[Nu]"], GridBox[{
          {"Au", 
           StyleBox[GridBox[{
              {" "},
              {"\[Mu]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CD[-$CellContext`\[Nu]][
       $CellContext`Accelerationu[-$CellContext`\[Mu]]],
      Editable->False], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Projectorh[
    $CellContext`CD[-$CellContext`\[Nu]][
     $CellContext`Accelerationu[-$CellContext`\[Mu]]]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " "},
          {"\[Mu]", "\[Nu]"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Riccicd[-$CellContext`\[Mu], -$CellContext`\[Nu]],
   Editable->False], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"h", 
        StyleBox[GridBox[{
           {"\[Alpha]", " "},
           {" ", "\[Beta]"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`h[$CellContext`\[Alpha], -$CellContext`\[Beta]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SuperscriptBox["\[EmptyDownTriangle]", "\[Beta]"], GridBox[{
           {"Kh", 
            StyleBox[GridBox[{
               {" ", " "},
               {"\[Alpha]", "\[Nu]"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`CD[$CellContext`\[Beta]][
        $CellContext`ExtrinsicKh[-$CellContext`\[Alpha], -$CellContext`\[Nu]]],
       Editable->False], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectorh[
     $CellContext`CD[$CellContext`\[Beta]][
      $CellContext`ExtrinsicKh[-$CellContext`\[Alpha], -$CellContext`\[Nu]]]],
    
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"u", 
        StyleBox[GridBox[{
           {" "},
           {"\[Mu]"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`u[-$CellContext`\[Mu]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SubscriptBox["\[EmptyDownTriangle]", "\[Nu]"], GridBox[{
           {"Kh", 
            StyleBox[GridBox[{
               {" ", "\[Alpha]"},
               {"\[Alpha]", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`CD[-$CellContext`\[Nu]][
        $CellContext`ExtrinsicKh[-$CellContext`\[Alpha], \
$CellContext`\[Alpha]]],
       Editable->False], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectorh[
     $CellContext`CD[-$CellContext`\[Nu]][
      $CellContext`ExtrinsicKh[-$CellContext`\[Alpha], \
$CellContext`\[Alpha]]]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"u", 
        StyleBox[GridBox[{
           {" "},
           {"\[Mu]"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`u[-$CellContext`\[Mu]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"h", 
        StyleBox[GridBox[{
           {" ", "\[Beta]"},
           {"\[Alpha]", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`h[-$CellContext`\[Alpha], $CellContext`\[Beta]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SubscriptBox["\[EmptyDownTriangle]", "\[Beta]"], GridBox[{
           {"Kh", 
            StyleBox[GridBox[{
               {" ", "\[Alpha]"},
               {"\[Mu]", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`CD[-$CellContext`\[Beta]][
        $CellContext`ExtrinsicKh[-$CellContext`\[Mu], $CellContext`\[Alpha]]],
       
       Editable->False], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectorh[
     $CellContext`CD[-$CellContext`\[Beta]][
      $CellContext`ExtrinsicKh[-$CellContext`\[Mu], $CellContext`\[Alpha]]]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"u", 
        StyleBox[GridBox[{
           {" "},
           {"\[Nu]"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`u[-$CellContext`\[Nu]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SubscriptBox["\[EmptyDownTriangle]", "\[Mu]"], GridBox[{
           {"Kh", 
            StyleBox[GridBox[{
               {" ", "\[Alpha]"},
               {"\[Alpha]", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`CD[-$CellContext`\[Mu]][
        $CellContext`ExtrinsicKh[-$CellContext`\[Alpha], \
$CellContext`\[Alpha]]],
       Editable->False], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectorh[
     $CellContext`CD[-$CellContext`\[Mu]][
      $CellContext`ExtrinsicKh[-$CellContext`\[Alpha], \
$CellContext`\[Alpha]]]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"u", 
        StyleBox[GridBox[{
           {" "},
           {"\[Nu]"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`u[-$CellContext`\[Nu]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Au", 
        StyleBox[GridBox[{
           {" "},
           {"\[Alpha]"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationu[-$CellContext`\[Alpha]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Au", 
        StyleBox[GridBox[{
           {"\[Alpha]"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationu[$CellContext`\[Alpha]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"u", 
        StyleBox[GridBox[{
           {" "},
           {"\[Mu]"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`u[-$CellContext`\[Mu]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"u", 
        StyleBox[GridBox[{
           {" "},
           {"\[Nu]"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`u[-$CellContext`\[Nu]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "\[Beta]"},
           {"\[Alpha]", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKh[-$CellContext`\[Alpha], $CellContext`\[Beta]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {"\[Alpha]", " "},
           {" ", "\[Beta]"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKh[$CellContext`\[Alpha], -$CellContext`\[Beta]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"u", 
        StyleBox[GridBox[{
           {" "},
           {"\[Mu]"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`u[-$CellContext`\[Mu]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"u", 
        StyleBox[GridBox[{
           {" "},
           {"\[Nu]"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`u[-$CellContext`\[Nu]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"h", 
        StyleBox[GridBox[{
           {"\[Alpha]", " "},
           {" ", "\[Beta]"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`h[$CellContext`\[Alpha], -$CellContext`\[Beta]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SuperscriptBox["\[EmptyDownTriangle]", "\[Beta]"], GridBox[{
           {"Au", 
            StyleBox[GridBox[{
               {" "},
               {"\[Alpha]"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`CD[$CellContext`\[Beta]][
        $CellContext`Accelerationu[-$CellContext`\[Alpha]]],
       Editable->False], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectorh[
     $CellContext`CD[$CellContext`\[Beta]][
      $CellContext`Accelerationu[-$CellContext`\[Alpha]]]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"u", 
        StyleBox[GridBox[{
           {" "},
           {"\[Mu]"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`u[-$CellContext`\[Mu]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"u", 
        StyleBox[GridBox[{
           {" "},
           {"\[Nu]"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`u[-$CellContext`\[Nu]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[
     RowBox[{
      StyleBox["(",
       FontColor->RGBColor[1, 0, 0]], 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"u", 
            StyleBox[GridBox[{
               {"\[Beta]"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`u[$CellContext`\[Beta]],
        Editable->False], " ", 
       RowBox[{"(", 
        InterpretationBox[
         StyleBox[
          RowBox[{
           SubscriptBox["\[EmptyDownTriangle]", "\[Beta]"], GridBox[{
             {"Kh", 
              StyleBox[GridBox[{
                 {" ", "\[Alpha]"},
                 {"\[Alpha]", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`CD[-$CellContext`\[Beta]][
          $CellContext`ExtrinsicKh[-$CellContext`\[Alpha], $CellContext`\
\[Alpha]]],
         Editable->False], ")"}]}], 
      StyleBox[")",
       FontColor->RGBColor[1, 0, 0]]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Scalar[$CellContext`u[$CellContext`\[Beta]] \
$CellContext`CD[-$CellContext`\[Beta]][
       $CellContext`ExtrinsicKh[-$CellContext`\[Alpha], \
$CellContext`\[Alpha]]]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"u", 
        StyleBox[GridBox[{
           {" "},
           {"\[Mu]"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`u[-$CellContext`\[Mu]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"u", 
        StyleBox[GridBox[{
           {" "},
           {"\[Nu]"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`u[-$CellContext`\[Nu]],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728677558372163*^9},
 CellLabel->"Out[12]=",
 CellID->1965183071,ExpressionUUID->"0f63456f-eef5-42d9-a65e-1dad8df7a3a6"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Now we construct the corresponding objects explicitly using xCoba:\
\>", "MathCaption",
 CellID->56875018,ExpressionUUID->"fd7f372d-5b65-413b-b0e4-2818d5df66b2"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"<<", "xAct`xCoba`"}]], "Input",
 CellLabel->"In[13]:=",
 CellID->268204611,ExpressionUUID->"fda09583-bb07-4426-af71-ecd43708a98e"],

Cell[CellGroupData[{

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.728677559895384*^9},
 CellLabel->"During evaluation of In[13]:=",
 CellID->2021331521,ExpressionUUID->"105ab4c3-4e2b-4421-8d43-e5b3a94d43e9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`xCoba`  version \"\>", 
   "\[InvisibleSpace]", "\<\"0.8.4\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2018", ",", "2", ",", "28"}], "}"}]}],
  SequenceForm["Package xAct`xCoba`  version ", "0.8.4", ", ", {2018, 2, 28}],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.728677559916027*^9},
 CellLabel->"During evaluation of In[13]:=",
 CellID->1475061554,ExpressionUUID->"a0133a86-ea02-4da9-b6ee-8f9f72f58322"],

Cell[BoxData["\<\"CopyRight (C) 2005-2018, David Yllanes and Jose M. \
Martin-Garcia, under the General Public License.\"\>"], "Print",
 CellChangeTimes->{3.728677559939972*^9},
 CellLabel->"During evaluation of In[13]:=",
 CellID->808317110,ExpressionUUID->"4316ad34-e8ab-472e-9369-d5e5a1d98df7"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.728677559955687*^9},
 CellLabel->"During evaluation of In[13]:=",
 CellID->404116848,ExpressionUUID->"5a694b80-2c7c-441c-993c-9130b87995d2"],

Cell[BoxData["\<\"These packages come with ABSOLUTELY NO WARRANTY; for \
details type Disclaimer[]. This is free software, and you are welcome to \
redistribute it under certain conditions. See the General Public License for \
details.\"\>"], "Print",
 CellChangeTimes->{3.7286775599705667`*^9},
 CellLabel->"During evaluation of In[13]:=",
 CellID->1648572522,ExpressionUUID->"af219891-98f5-4702-9c9e-ce729eab76c9"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.7286775599908*^9},
 CellLabel->"During evaluation of In[13]:=",
 CellID->724363660,ExpressionUUID->"cf7f3e23-b79f-47bc-89ec-81e66daf09ea"]
}, Open  ]]
}, Open  ]],

Cell[BoxData[
 RowBox[{"DefChart", "[", 
  RowBox[{"schw", ",", "M", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2", ",", "3"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"t", "[", "]"}], ",", 
     RowBox[{"r", "[", "]"}], ",", 
     RowBox[{"\[Theta]", "[", "]"}], ",", 
     RowBox[{"\[Phi]", "[", "]"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[14]:=",
 CellID->2129123830,ExpressionUUID->"c88e175d-7824-46c1-ae5e-c7dcb521cf60"],

Cell[BoxData[
 RowBox[{"DefConstantSymbol", "[", 
  RowBox[{"mass", ",", 
   RowBox[{"PrintAs", "\[Rule]", "\"\<M\>\""}]}], "]"}]], "Input",
 CellLabel->"In[15]:=",
 CellID->664200849,ExpressionUUID->"7e5810e0-91ce-48dd-881b-1c4b1df32444"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{"g", "=", 
    RowBox[{"CTensor", "[", 
     RowBox[{
      RowBox[{"DiagonalMatrix", "[", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "1"}], "+", 
          RowBox[{"2", 
           RowBox[{"mass", "/", 
            RowBox[{"r", "[", "]"}]}]}]}], ",", 
         RowBox[{"1", "/", 
          RowBox[{"(", 
           RowBox[{"1", "-", 
            RowBox[{"2", 
             RowBox[{"mass", "/", 
              RowBox[{"r", "[", "]"}]}]}]}], ")"}]}], ",", 
         RowBox[{
          RowBox[{"r", "[", "]"}], "^", "2"}], ",", 
         RowBox[{
          RowBox[{
           RowBox[{"r", "[", "]"}], "^", "2"}], 
          RowBox[{
           RowBox[{"Sin", "[", 
            RowBox[{"\[Theta]", "[", "]"}], "]"}], "^", "2"}]}]}], "}"}], 
       "]"}], ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "schw"}], ",", 
        RowBox[{"-", "schw"}]}], "}"}]}], "]"}]}], ")"}], "[", 
  RowBox[{
   RowBox[{"-", "\[Mu]"}], ",", 
   RowBox[{"-", "\[Nu]"}]}], "]"}]], "Input",
 CellLabel->"In[16]:=",
 CellID->603672037,ExpressionUUID->"1b3f2831-bf39-4605-a0dc-e7b5f83b7a84"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{
             RowBox[{"-", "1"}], "+", 
             FractionBox[
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`mass,
                Editable->False]}], 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
              RowBox[{
                RowBox[{"-", "1"}], "+", 
                FractionBox[
                 RowBox[{"2", " ", 
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], 
                 InterpretationBox[
                  StyleBox[
                  "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`r[], Editable -> False]]}]}]],
           
           Annotation[#, {1, 1} -> -1 + 
             2 ($CellContext`mass/$CellContext`r[]), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            FractionBox["1", 
             RowBox[{"1", "-", 
              FractionBox[
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`mass,
                 Editable->False]}], 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]]}]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
              FractionBox["1", 
               RowBox[{"1", "-", 
                 FractionBox[
                  RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], 
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False]]}]]}]],
           
           Annotation[#, {2, 
             2} -> (1 - 2 ($CellContext`mass/$CellContext`r[]))^(-1), 
            "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox[
                "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                $CellContext`r[], Editable -> False], "2"]}]],
           Annotation[#, {3, 3} -> $CellContext`r[]^2, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox[
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"], " ", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], "2"]}],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "4"}], "}"}], "\[Rule]", 
              RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                  "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`r[], Editable -> False], "2"], " ", 
                SuperscriptBox[
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    RowBox[{"Sin"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Sin, Editable -> False], "[", 
                   InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                 "2"]}]}]],
           Annotation[#, {4, 4} -> $CellContext`r[]^2 Sin[
                $CellContext`\[Theta][]]^2, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " "},
         {"\[Mu]", "\[Nu]"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{-1 + 2 ($CellContext`mass/$CellContext`r[]), 0, 0, 
     0}, {0, (1 - 2 ($CellContext`mass/$CellContext`r[]))^(-1), 0, 0}, {
    0, 0, $CellContext`r[]^2, 0}, {0, 0, 0, $CellContext`r[]^2 Sin[
        $CellContext`\[Theta][]]^2}}, {-$CellContext`schw, \
-$CellContext`schw}, 0][-$CellContext`\[Mu], -$CellContext`\[Nu]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728677563096695*^9},
 CellLabel->"Out[16]=",
 CellID->951592667,ExpressionUUID->"13ab58a5-eb42-4334-b94a-a664b8649277"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{"u", "=", 
    RowBox[{"CTensor", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"1", "/", 
         RowBox[{"Sqrt", "[", 
          RowBox[{"1", "-", 
           RowBox[{"2", 
            RowBox[{"mass", "/", 
             RowBox[{"r", "[", "]"}]}]}]}], "]"}]}], ",", "0", ",", "0", ",", 
        "0"}], "}"}], ",", 
      RowBox[{"{", "schw", "}"}]}], "]"}]}], ")"}], "[", "\[Mu]", 
  "]"}]], "Input",
 CellLabel->"In[17]:=",
 CellID->1097062665,ExpressionUUID->"80f49d09-fc77-41c7-9328-79a406e1e695"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            FractionBox["1", 
             SqrtBox[
              RowBox[{"1", "-", 
               FractionBox[
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`mass,
                  Editable->False]}], 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]]}]]],
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", 
              FractionBox["1", 
               SqrtBox[
                RowBox[{"1", "-", 
                  FractionBox[
                   RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], 
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]]]}]],
           
           Annotation[#, {1} -> (1 - 2 ($CellContext`mass/$CellContext`r[]))^
             Rational[-1, 2], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", "4", "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4} -> 0, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {"\[Mu]"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{(1 - 2 ($CellContext`mass/$CellContext`r[]))^
    Rational[-1, 2], 0, 0, 0}, {$CellContext`schw}, 0][$CellContext`\[Mu]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728677564327098*^9},
 CellLabel->"Out[17]=",
 CellID->728573098,ExpressionUUID->"8a602fc0-6398-46bd-8632-1ef24a9740ed"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The vector is normalized:", "MathCaption",
 CellID->877348773,ExpressionUUID->"a6ecf38a-2be8-4b9a-9a9c-05a0cfb8ca98"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"g", "[", 
    RowBox[{
     RowBox[{"-", "\[Mu]"}], ",", 
     RowBox[{"-", "\[Nu]"}]}], "]"}], 
   RowBox[{"u", "[", "\[Mu]", "]"}], 
   RowBox[{"u", "[", "\[Nu]", "]"}]}], "//", "Simplify"}]], "Input",
 CellLabel->"In[18]:=",
 CellID->978107509,ExpressionUUID->"8d7916ab-6176-43e0-96e3-adf027011cc3"],

Cell[BoxData[
 RowBox[{"-", "1"}]], "Output",
 CellChangeTimes->{3.728677566413361*^9},
 CellLabel->"Out[18]=",
 CellID->1848222812,ExpressionUUID->"0345e2f1-5198-4909-9b50-b6f678cda4fa"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"CD", "=", 
  RowBox[{"LC", "[", "g", "]"}]}]], "Input",
 CellLabel->"In[19]:=",
 CellID->644421335,ExpressionUUID->"bf41478f-90fb-4cfe-ae48-cbb336498f25"],

Cell[BoxData[
 RowBox[{"CCovD", "[", 
  RowBox[{"PDschw", ",", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", 
           RowBox[{"-", 
            FractionBox[
             InterpretationBox[
              StyleBox["M",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`mass,
              Editable->False], 
             RowBox[{
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`mass,
                Editable->False], " ", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}], "-", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], "2"]}]]}], ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"-", 
            FractionBox[
             InterpretationBox[
              StyleBox["M",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`mass,
              Editable->False], 
             RowBox[{
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`mass,
                Editable->False], " ", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}], "-", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], "2"]}]]}], ",", "0", ",", "0", ",", "0"}], 
          "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           FractionBox[
            RowBox[{
             InterpretationBox[
              StyleBox["M",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`mass,
              Editable->False], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"-", "2"}], " ", 
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`mass,
                 Editable->False]}], "+", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}], ")"}]}], 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "3"]], ",", "0", ",", "0", ",", "0"}], "}"}], 
         ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", 
           FractionBox[
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`mass,
             Editable->False], 
            RowBox[{
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`mass,
               Editable->False], " ", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}], "-", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}]], ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           RowBox[{
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["M",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`mass,
              Editable->False]}], "-", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]}], ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           RowBox[{
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`mass,
                Editable->False]}], "-", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}], ")"}], " ", 
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["Sin",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Sin,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], "2"]}]}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", 
           FractionBox["1", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]], ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", 
           FractionBox["1", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]], ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           RowBox[{
            RowBox[{"-", 
             RowBox[{
              InterpretationBox[
               StyleBox["Cos",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cos,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}]}], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["Sin",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Sin,
              Editable->False], "[", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False], "]"}]}]}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           FractionBox["1", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "0", ",", "0", ",", 
           RowBox[{
            InterpretationBox[
             StyleBox["Cot",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cot,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", 
           FractionBox["1", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]], ",", 
           RowBox[{
            InterpretationBox[
             StyleBox["Cot",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cot,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], ",", "0"}], "}"}]}], "}"}]}], "}"}], 
     ",", 
     RowBox[{"{", 
      RowBox[{"schw", ",", 
       RowBox[{"-", "schw"}], ",", 
       RowBox[{"-", "schw"}]}], "}"}], ",", "0"}], "]"}], ",", 
   RowBox[{"CTensor", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "1"}], "+", 
          FractionBox[
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`mass,
             Editable->False]}], 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]]}], ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", 
         FractionBox["1", 
          RowBox[{"1", "-", 
           FractionBox[
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["M",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`mass,
              Editable->False]}], 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]]}]], ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"], ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Sin",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Sin,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "2"]}]}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "schw"}], ",", 
       RowBox[{"-", "schw"}]}], "}"}], ",", "0"}], "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728677579530292*^9},
 CellLabel->"Out[19]=",
 CellID->750481805,ExpressionUUID->"b8a83ae6-cb41-4ad8-9c62-b1f19be33845"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CD", "[", 
   RowBox[{"-", "\[Mu]"}], "]"}], "[", 
  RowBox[{"g", "[", 
   RowBox[{
    RowBox[{"-", "\[Alpha]"}], ",", 
    RowBox[{"-", "\[Beta]"}]}], "]"}], "]"}]], "Input",
 CellLabel->"In[20]:=",
 CellID->81961636,ExpressionUUID->"7237a8e1-83ec-4c9e-8005-c156f6a071bd"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728677580201586*^9},
 CellLabel->"Out[20]=",
 CellID->2100127951,ExpressionUUID->"656b7f4d-0a89-4ea0-ba62-713404515210"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"u", "[", 
  RowBox[{"-", "\[Nu]"}], "]"}]], "Input",
 CellLabel->"In[21]:=",
 CellID->670090720,ExpressionUUID->"44e82d7e-b04d-4072-9bc8-48d95d477b8b"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox[
            RowBox[{"-", 
             SqrtBox[
              RowBox[{"1", "-", 
               FractionBox[
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`mass,
                  Editable->False]}], 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]]}]]}],
            RowBox[{
              RowBox[{"{", "1", "}"}], "\[Rule]", 
              RowBox[{"-", 
                SqrtBox[
                 RowBox[{"1", "-", 
                   FractionBox[
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]]}]}]],
           
           Annotation[#, {
             1} -> -(1 - 2 ($CellContext`mass/$CellContext`r[]))^
              Rational[1, 2], "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", "2", "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", "3", "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", "4", "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4} -> 0, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {
           1 -> RGBColor[1, 0, 0], 2 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" "},
         {"\[Nu]"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{-(1 - 2 ($CellContext`mass/$CellContext`r[]))^
     Rational[1, 2], 0, 0, 0}, {-$CellContext`schw}, 0][-$CellContext`\[Nu]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728677580882276*^9},
 CellLabel->"Out[21]=",
 CellID->1430974628,ExpressionUUID->"0a6c1d4c-71b2-4b9d-80e0-557037e8e36d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"CD", "[", 
    RowBox[{"-", "\[Mu]"}], "]"}], "[", 
   RowBox[{"u", "[", 
    RowBox[{"-", "\[Nu]"}], "]"}], "]"}], "//", "Simplify"}]], "Input",
 CellLabel->"In[22]:=",
 CellID->352319883,ExpressionUUID->"49f5d0b5-d8f2-49a8-adbb-1280a47efceb"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      StyleBox[GridBox[{
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {1, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox[
            FractionBox[
             RowBox[{
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`mass,
               Editable->False], " ", 
              SqrtBox[
               RowBox[{"1", "-", 
                FractionBox[
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`mass,
                   Editable->False]}], 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]]}]]}], 
             RowBox[{
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "2"}], " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`mass,
                  Editable->False]}], "+", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}], ")"}]}]],
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", 
              FractionBox[
               RowBox[{
                 InterpretationBox[
                  StyleBox[
                   RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                   False], $CellContext`mass, Editable -> False], " ", 
                 SqrtBox[
                  RowBox[{"1", "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]]}]]}], 
               RowBox[{
                 InterpretationBox[
                  StyleBox[
                  "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`r[], Editable -> False], " ", 
                 RowBox[{"(", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]]}]],
           
           Annotation[#, {2, 
             1} -> ($CellContext`mass (1 - 
                2 ($CellContext`mass/$CellContext`r[]))^
               Rational[
                1, 2]) ($CellContext`r[]^(-1)/((-2) $CellContext`mass + \
$CellContext`r[])), "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"2", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {2, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"3", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {3, 4} -> 0, "Tooltip"]& ]},
         {
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "1"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 1} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "2"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 2} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "3"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 3} -> 0, "Tooltip"]& ], 
          TagBox[
           TooltipBox["0",
            RowBox[{
              RowBox[{"{", 
                RowBox[{"4", ",", "4"}], "}"}], "\[Rule]", "0"}]],
           Annotation[#, {4, 4} -> 0, "Tooltip"]& ]}
        },
        AutoDelete->False,
        GridBoxDividers->{
         "ColumnsIndexed" -> {1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]},
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}],
       FontSize->Small], 
      StyleBox[GridBox[{
         {" ", " "},
         {"\[Nu]", "\[Mu]"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`CTensor[{{0, 0, 0, 
    0}, {($CellContext`mass (1 - 2 ($CellContext`mass/$CellContext`r[]))^
       Rational[
        1, 2]) ($CellContext`r[]^(-1)/((-2) $CellContext`mass + \
$CellContext`r[])), 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 
    0}}, {-$CellContext`schw, -$CellContext`schw}, 
   0][-$CellContext`\[Nu], -$CellContext`\[Mu]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728677581514394*^9},
 CellLabel->"Out[22]=",
 CellID->786980887,ExpressionUUID->"25b2220a-9fa9-4512-8f17-afac37ecc9b2"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Accelerationu", "=", 
  RowBox[{"-", 
   RowBox[{"Head", "[", 
    RowBox[{
     RowBox[{"u", "[", "\[Mu]", "]"}], 
     RowBox[{
      RowBox[{"CD", "[", 
       RowBox[{"-", "\[Mu]"}], "]"}], "[", 
      RowBox[{"u", "[", 
       RowBox[{"-", "\[Nu]"}], "]"}], "]"}]}], "]"}]}]}]], "Input",
 CellLabel->"In[23]:=",
 CellID->276336301,ExpressionUUID->"c5eeecca-c9fb-4e03-81aa-f7242ee792c9"],

Cell[BoxData[
 RowBox[{"CTensor", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0", ",", 
     RowBox[{"-", 
      FractionBox[
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`mass,
        Editable->False], 
       RowBox[{
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "2"}], " ", 
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`mass,
            Editable->False]}], "+", 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False]}], ")"}]}]]}], ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"-", "schw"}], "}"}], ",", "0"}], "]"}]], "Output",
 CellChangeTimes->{3.728677582122859*^9},
 CellLabel->"Out[23]=",
 CellID->135653359,ExpressionUUID->"3ddf51bd-e15d-4535-933a-aa856222549d"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The extrinsic curvature vanishes, because this is a time independent solution:\
\>", "MathCaption",
 CellID->1754932694,ExpressionUUID->"9f2ee235-cd64-41ef-a160-f289eb47d489"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"CD", "[", 
    RowBox[{"-", "\[Mu]"}], "]"}], "[", 
   RowBox[{"u", "[", 
    RowBox[{"-", "\[Nu]"}], "]"}], "]"}], "-", 
  RowBox[{
   RowBox[{"u", "[", 
    RowBox[{"-", "\[Mu]"}], "]"}], 
   RowBox[{"Accelerationu", "[", 
    RowBox[{"-", "\[Nu]"}], "]"}]}]}]], "Input",
 CellLabel->"In[24]:=",
 CellID->1613380160,ExpressionUUID->"d12f2ad3-5192-4e0a-8e40-dd0ae4eaa674"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728677587988696*^9},
 CellLabel->"Out[24]=",
 CellID->687433615,ExpressionUUID->"aa048c02-49a6-41e0-9a55-b16da77065dd"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ExtrinsicKh", "=", "Zero"}]], "Input",
 CellLabel->"In[25]:=",
 CellID->1461293161,ExpressionUUID->"29fa3ca7-514b-46c3-b1f5-d204aa2d754a"],

Cell[BoxData["Zero"], "Output",
 CellChangeTimes->{3.728677588701255*^9},
 CellLabel->"Out[25]=",
 CellID->2078357721,ExpressionUUID->"2f39849f-adf9-43ae-90ea-6b9892a877db"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We still have problems on how to replace the symbolic metric by the explicit \
metric.\
\>", "MathCaption",
 CellID->486782399,ExpressionUUID->"da04e4a8-3146-4683-8bb9-f04170c81e8d"],

Cell[CellGroupData[{

Cell[BoxData["expr"], "Input",
 CellLabel->"In[26]:=",
 CellID->1586568570,ExpressionUUID->"33210f84-34e9-4ed2-b698-ac507149334c"],

Cell[BoxData[
 TemplateBox[{
  "InducedFrom", "unknown", 
   "\"Unknown \\!\\(\\*RowBox[{\\\"\\\\\\\"metric\\\\\\\"\\\"}]\\) \
\\!\\(\\*RowBox[{\\\"CTensor\\\", \\\"[\\\", RowBox[{RowBox[{\\\"{\\\", \
RowBox[{RowBox[{\\\"{\\\", RowBox[{RowBox[{RowBox[{\\\"-\\\", \\\"1\\\"}], \\\
\"+\\\", FractionBox[RowBox[{\\\"2\\\", \\\" \\\", \
InterpretationBox[StyleBox[RowBox[{\\\"M\\\"}], Rule[AutoSpacing, False], \
Rule[ShowAutoStyles, False]], mass, Rule[Editable, False]]}], \
InterpretationBox[StyleBox[\\\"r\\\", Rule[AutoSpacing, False], \
Rule[ShowAutoStyles, False]], r[], Rule[Editable, False]]]}], \\\",\\\", \
\\\"0\\\", \\\",\\\", \\\"0\\\", \\\",\\\", \\\"0\\\"}], \\\"}\\\"}], \\\",\\\
\", RowBox[{\\\"{\\\", RowBox[{\\\"0\\\", \\\",\\\", FractionBox[\\\"1\\\", \
RowBox[{\\\"1\\\", \\\"-\\\", FractionBox[RowBox[{\\\"2\\\", \\\" \\\", \
InterpretationBox[StyleBox[RowBox[{\\\"M\\\"}], Rule[AutoSpacing, False], \
Rule[ShowAutoStyles, False]], mass, Rule[Editable, False]]}], \
InterpretationBox[StyleBox[\\\"r\\\", Rule[AutoSpacing, False], \
Rule[ShowAutoStyles, False]], r[], Rule[Editable, False]]]}]], \\\",\\\", \
\\\"0\\\", \\\",\\\", \\\"0\\\"}], \\\"}\\\"}], \\\",\\\", RowBox[{\\\"{\\\", \
RowBox[{\\\"0\\\", \\\",\\\", \\\"0\\\", \\\",\\\", \
SuperscriptBox[InterpretationBox[StyleBox[\\\"r\\\", Rule[AutoSpacing, \
False], Rule[ShowAutoStyles, False]], r[], Rule[Editable, False]], \
\\\"2\\\"], \\\",\\\", \\\"0\\\"}], \\\"}\\\"}], \\\",\\\", \
RowBox[{\\\"{\\\", RowBox[{\\\"0\\\", \\\",\\\", \\\"0\\\", \\\",\\\", \
\\\"0\\\", \\\",\\\", RowBox[{SuperscriptBox[InterpretationBox[StyleBox[\\\"r\
\\\", Rule[AutoSpacing, False], Rule[ShowAutoStyles, False]], r[], \
Rule[Editable, False]], \\\"2\\\"], \\\" \\\", \
SuperscriptBox[RowBox[{InterpretationBox[StyleBox[RowBox[{\\\"Sin\\\"}], \
Rule[AutoSpacing, False], Rule[ShowAutoStyles, False]], Sin, Rule[Editable, \
False]], \\\"[\\\", InterpretationBox[StyleBox[\\\"\[Theta]\\\", \
Rule[AutoSpacing, False], Rule[ShowAutoStyles, False]], \\\\[Theta][], \
Rule[Editable, False]], \\\"]\\\"}], \\\"2\\\"]}]}], \\\"}\\\"}]}], \
\\\"}\\\"}], \\\",\\\", RowBox[{\\\"{\\\", RowBox[{RowBox[{\\\"-\\\", \
\\\"schw\\\"}], \\\",\\\", RowBox[{\\\"-\\\", \\\"schw\\\"}]}], \\\"}\\\"}], \
\\\",\\\", \\\"0\\\"}], \\\"]\\\"}]\\).\"", 2, 26, 2, 19251607901465226019, 
   "Local", "xAct`xTensor`InducedFrom"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728677594903592*^9},
 CellLabel->"During evaluation of In[26]:=",
 CellID->731989197,ExpressionUUID->"5abc99a0-31d8-49b8-843a-e365cc7f1778"],

Cell[BoxData[
 TemplateBox[{
  "Throw", "nocatch", 
   "\"Uncaught \\!\\(\\*RowBox[{\\\"Throw\\\", \\\"[\\\", \\\"Null\\\", \\\"]\
\\\"}]\\) returned to top level.\"", 2, 26, 3, 19251607901465226019, "Local"},
  
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.728677594947353*^9},
 CellLabel->"During evaluation of In[26]:=",
 CellID->1756103963,ExpressionUUID->"7b5e3b88-71c7-48ea-9472-517c9bb48e89"],

Cell[BoxData[
 RowBox[{"Hold", "[", 
  RowBox[{"Throw", "[", "Null", "]"}], "]"}]], "Output",
 CellChangeTimes->{3.728677594964718*^9},
 CellLabel->"Out[26]=",
 CellID->1971545420,ExpressionUUID->"a1a799e1-1311-447d-a9fc-bc7a820286f4"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"GaussCodazzi", "[", 
   RowBox[{
    RowBox[{"RicciCD", "[", 
     RowBox[{
      RowBox[{"-", "\[Mu]"}], ",", 
      RowBox[{"-", "\[Nu]"}]}], "]"}], ",", "h"}], "]"}], "//", 
  "Simplify"}]], "Input",
 CellLabel->"In[27]:=",
 CellID->2010255759,ExpressionUUID->"5e1fed67-9814-41b7-b822-747c1938f019"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"\[Alpha]", "\[Beta]"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Riccicd[-$CellContext`\[Alpha], -$CellContext`\[Beta]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        StyleBox[GridBox[{
           {GridBox[{
              {
               TagBox[
                TooltipBox["1",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "1"}]],
                Annotation[#, {1, 1, 1, 1} -> 1, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 1, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 1, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "1", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 1, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 2, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["1",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "1"}]],
                Annotation[#, {1, 1, 2, 2} -> 1, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 2, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "2", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 2, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "3", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 3, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "3", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 3, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["1",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "3", ",", "3"}], "}"}], 
                   "\[Rule]", "1"}]],
                Annotation[#, {1, 1, 3, 3} -> 1, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "3", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 3, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "4", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 4, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "4", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 4, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "4", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 1, 4, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["1",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "1", ",", "4", ",", "4"}], "}"}], 
                   "\[Rule]", "1"}]],
                Annotation[#, {1, 1, 4, 4} -> 1, "Tooltip"]& ]}
             },
             AutoDelete->False,
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 2, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 2, 1, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 2, 1, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "1", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 2, 1, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 2, 2, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 2, 2, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 2, 2, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "2", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 2, 2, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "3", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 2, 3, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "3", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 2, 3, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "3", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 2, 3, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "3", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 2, 3, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "4", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 2, 4, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "4", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 2, 4, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "4", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 2, 4, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "2", ",", "4", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 2, 4, 4} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "3", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 3, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "3", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 3, 1, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "3", ",", "1", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 3, 1, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "3", ",", "1", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 3, 1, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "3", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 3, 2, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "3", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 3, 2, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "3", ",", "2", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 3, 2, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "3", ",", "2", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 3, 2, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "3", ",", "3", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 3, 3, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "3", ",", "3", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 3, 3, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "3", ",", "3", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 3, 3, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "3", ",", "3", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 3, 3, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "3", ",", "4", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 3, 4, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "3", ",", "4", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 3, 4, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "3", ",", "4", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 3, 4, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "3", ",", "4", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 3, 4, 4} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "4", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 4, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "4", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 4, 1, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "4", ",", "1", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 4, 1, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "4", ",", "1", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 4, 1, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "4", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 4, 2, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "4", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 4, 2, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "4", ",", "2", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 4, 2, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "4", ",", "2", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 4, 2, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "4", ",", "3", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 4, 3, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "4", ",", "3", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 4, 3, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "4", ",", "3", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 4, 3, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "4", ",", "3", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 4, 3, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "4", ",", "4", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 4, 4, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "4", ",", "4", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 4, 4, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "4", ",", "4", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 4, 4, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"1", ",", "4", ",", "4", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {1, 4, 4, 4} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}]},
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 1, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 1, 1, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 1, 1, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "1", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 1, 1, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 1, 2, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 1, 2, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 1, 2, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "2", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 1, 2, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "3", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 1, 3, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "3", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 1, 3, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "3", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 1, 3, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "3", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 1, 3, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "4", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 1, 4, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "4", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 1, 4, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "4", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 1, 4, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "1", ",", "4", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 1, 4, 4} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["1",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "1"}]],
                Annotation[#, {2, 2, 1, 1} -> 1, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 1, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 1, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "1", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 1, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 2, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["1",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "1"}]],
                Annotation[#, {2, 2, 2, 2} -> 1, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 2, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "2", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 2, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "3", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 3, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "3", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 3, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["1",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "3", ",", "3"}], "}"}], 
                   "\[Rule]", "1"}]],
                Annotation[#, {2, 2, 3, 3} -> 1, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "3", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 3, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "4", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 4, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "4", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 4, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "4", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 2, 4, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["1",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "2", ",", "4", ",", "4"}], "}"}], 
                   "\[Rule]", "1"}]],
                Annotation[#, {2, 2, 4, 4} -> 1, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "3", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 3, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "3", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 3, 1, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "3", ",", "1", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 3, 1, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "3", ",", "1", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 3, 1, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "3", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 3, 2, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "3", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 3, 2, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "3", ",", "2", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 3, 2, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "3", ",", "2", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 3, 2, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "3", ",", "3", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 3, 3, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "3", ",", "3", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 3, 3, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "3", ",", "3", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 3, 3, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "3", ",", "3", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 3, 3, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "3", ",", "4", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 3, 4, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "3", ",", "4", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 3, 4, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "3", ",", "4", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 3, 4, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "3", ",", "4", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 3, 4, 4} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "4", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 4, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "4", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 4, 1, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "4", ",", "1", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 4, 1, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "4", ",", "1", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 4, 1, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "4", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 4, 2, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "4", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 4, 2, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "4", ",", "2", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 4, 2, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "4", ",", "2", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 4, 2, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "4", ",", "3", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 4, 3, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "4", ",", "3", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 4, 3, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "4", ",", "3", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 4, 3, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "4", ",", "3", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 4, 3, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "4", ",", "4", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 4, 4, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "4", ",", "4", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 4, 4, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "4", ",", "4", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 4, 4, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"2", ",", "4", ",", "4", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {2, 4, 4, 4} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}]},
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "1", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 1, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "1", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 1, 1, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "1", ",", "1", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 1, 1, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "1", ",", "1", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 1, 1, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "1", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 1, 2, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "1", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 1, 2, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "1", ",", "2", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 1, 2, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "1", ",", "2", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 1, 2, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "1", ",", "3", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 1, 3, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "1", ",", "3", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 1, 3, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "1", ",", "3", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 1, 3, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "1", ",", "3", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 1, 3, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "1", ",", "4", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 1, 4, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "1", ",", "4", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 1, 4, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "1", ",", "4", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 1, 4, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "1", ",", "4", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 1, 4, 4} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "2", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 2, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "2", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 2, 1, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "2", ",", "1", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 2, 1, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "2", ",", "1", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 2, 1, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "2", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 2, 2, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "2", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 2, 2, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "2", ",", "2", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 2, 2, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "2", ",", "2", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 2, 2, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "2", ",", "3", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 2, 3, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "2", ",", "3", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 2, 3, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "2", ",", "3", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 2, 3, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "2", ",", "3", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 2, 3, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "2", ",", "4", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 2, 4, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "2", ",", "4", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 2, 4, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "2", ",", "4", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 2, 4, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "2", ",", "4", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 2, 4, 4} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["1",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "3", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "1"}]],
                Annotation[#, {3, 3, 1, 1} -> 1, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "3", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 3, 1, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "3", ",", "1", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 3, 1, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "3", ",", "1", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 3, 1, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "3", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 3, 2, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["1",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "3", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "1"}]],
                Annotation[#, {3, 3, 2, 2} -> 1, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "3", ",", "2", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 3, 2, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "3", ",", "2", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 3, 2, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "3", ",", "3", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 3, 3, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "3", ",", "3", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 3, 3, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["1",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "3", ",", "3", ",", "3"}], "}"}], 
                   "\[Rule]", "1"}]],
                Annotation[#, {3, 3, 3, 3} -> 1, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "3", ",", "3", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 3, 3, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "3", ",", "4", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 3, 4, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "3", ",", "4", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 3, 4, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "3", ",", "4", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 3, 4, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["1",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "3", ",", "4", ",", "4"}], "}"}], 
                   "\[Rule]", "1"}]],
                Annotation[#, {3, 3, 4, 4} -> 1, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "4", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 4, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "4", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 4, 1, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "4", ",", "1", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 4, 1, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "4", ",", "1", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 4, 1, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "4", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 4, 2, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "4", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 4, 2, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "4", ",", "2", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 4, 2, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "4", ",", "2", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 4, 2, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "4", ",", "3", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 4, 3, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "4", ",", "3", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 4, 3, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "4", ",", "3", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 4, 3, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "4", ",", "3", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 4, 3, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "4", ",", "4", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 4, 4, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "4", ",", "4", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 4, 4, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "4", ",", "4", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 4, 4, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"3", ",", "4", ",", "4", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {3, 4, 4, 4} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}]},
           {GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "1", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 1, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "1", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 1, 1, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "1", ",", "1", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 1, 1, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "1", ",", "1", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 1, 1, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "1", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 1, 2, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "1", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 1, 2, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "1", ",", "2", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 1, 2, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "1", ",", "2", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 1, 2, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "1", ",", "3", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 1, 3, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "1", ",", "3", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 1, 3, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "1", ",", "3", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 1, 3, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "1", ",", "3", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 1, 3, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "1", ",", "4", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 1, 4, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "1", ",", "4", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 1, 4, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "1", ",", "4", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 1, 4, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "1", ",", "4", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 1, 4, 4} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "2", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 2, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "2", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 2, 1, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "2", ",", "1", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 2, 1, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "2", ",", "1", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 2, 1, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "2", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 2, 2, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "2", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 2, 2, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "2", ",", "2", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 2, 2, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "2", ",", "2", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 2, 2, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "2", ",", "3", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 2, 3, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "2", ",", "3", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 2, 3, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "2", ",", "3", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 2, 3, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "2", ",", "3", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 2, 3, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "2", ",", "4", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 2, 4, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "2", ",", "4", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 2, 4, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "2", ",", "4", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 2, 4, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "2", ",", "4", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 2, 4, 4} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "3", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 3, 1, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "3", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 3, 1, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "3", ",", "1", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 3, 1, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "3", ",", "1", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 3, 1, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "3", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 3, 2, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "3", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 3, 2, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "3", ",", "2", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 3, 2, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "3", ",", "2", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 3, 2, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "3", ",", "3", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 3, 3, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "3", ",", "3", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 3, 3, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "3", ",", "3", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 3, 3, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "3", ",", "3", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 3, 3, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "3", ",", "4", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 3, 4, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "3", ",", "4", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 3, 4, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "3", ",", "4", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 3, 4, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "3", ",", "4", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 3, 4, 4} -> 0, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}], GridBox[{
              {
               TagBox[
                TooltipBox["1",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "4", ",", "1", ",", "1"}], "}"}], 
                   "\[Rule]", "1"}]],
                Annotation[#, {4, 4, 1, 1} -> 1, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "4", ",", "1", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 4, 1, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "4", ",", "1", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 4, 1, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "4", ",", "1", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 4, 1, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "4", ",", "2", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 4, 2, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["1",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "4", ",", "2", ",", "2"}], "}"}], 
                   "\[Rule]", "1"}]],
                Annotation[#, {4, 4, 2, 2} -> 1, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "4", ",", "2", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 4, 2, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "4", ",", "2", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 4, 2, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "4", ",", "3", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 4, 3, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "4", ",", "3", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 4, 3, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["1",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "4", ",", "3", ",", "3"}], "}"}], 
                   "\[Rule]", "1"}]],
                Annotation[#, {4, 4, 3, 3} -> 1, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "4", ",", "3", ",", "4"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 4, 3, 4} -> 0, "Tooltip"]& ]},
              {
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "4", ",", "4", ",", "1"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 4, 4, 1} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "4", ",", "4", ",", "2"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 4, 4, 2} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["0",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "4", ",", "4", ",", "3"}], "}"}], 
                   "\[Rule]", "0"}]],
                Annotation[#, {4, 4, 4, 3} -> 0, "Tooltip"]& ], 
               TagBox[
                TooltipBox["1",
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{"4", ",", "4", ",", "4", ",", "4"}], "}"}], 
                   "\[Rule]", "1"}]],
                Annotation[#, {4, 4, 4, 4} -> 1, "Tooltip"]& ]}
             },
             AutoDelete->False,
             
             GridBoxDividers->{
              "ColumnsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
               "RowsIndexed" -> {
                1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}]}
          },
          AutoDelete->False,
          
          GridBoxDividers->{
           "ColumnsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
            "RowsIndexed" -> {
             1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}],
         FontSize->Small], 
        StyleBox[GridBox[{
           {" ", "\[Alpha]", " ", "\[Beta]"},
           {"\[Mu]", " ", "\[Nu]", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xCoba`CTensor[{{{{1, 0, 0, 0}, {0, 1, 0, 0}, {0, 0, 1, 0}, {0, 0, 0, 
     1}}, {{0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0}}, {{0, 0, 0,
      0}, {0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0}}, {{0, 0, 0, 0}, {0, 0, 0, 
     0}, {0, 0, 0, 0}, {0, 0, 0, 0}}}, {{{0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0,
      0}, {0, 0, 0, 0}}, {{1, 0, 0, 0}, {0, 1, 0, 0}, {0, 0, 1, 0}, {0, 0, 0, 
     1}}, {{0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0}}, {{0, 0, 0,
      0}, {0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0}}}, {{{0, 0, 0, 0}, {0, 0, 
     0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0}}, {{0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 
     0, 0}, {0, 0, 0, 0}}, {{1, 0, 0, 0}, {0, 1, 0, 0}, {0, 0, 1, 0}, {0, 0, 
     0, 1}}, {{0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0}}}, {{{0, 
     0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0}}, {{0, 0, 0, 0}, {0, 
     0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0}}, {{0, 0, 0, 0}, {0, 0, 0, 0}, {0, 
     0, 0, 0}, {0, 0, 0, 0}}, {{1, 0, 0, 0}, {0, 1, 0, 0}, {0, 0, 1, 0}, {0, 
     0, 0, 1}}}}, {-$CellContext`schw, $CellContext`schw, -$CellContext`schw, \
$CellContext`schw}, 
     0][-$CellContext`\[Mu], $CellContext`\[Alpha], -$CellContext`\[Nu], \
$CellContext`\[Beta]],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       StyleBox[GridBox[{
          {
           TagBox[
            TooltipBox["\<\"\[GrayCircle]\"\>",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"1", ",", "1"}], "}"}], "\[Rule]", 
               FractionBox[
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], " ", 
                  RowBox[{"(", 
                    RowBox[{
                    RowBox[{"3", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], "-", 
                    InterpretationBox[
                    StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\[EmptyDownTriangle],\!\(\*StyleBox[\"\
\[ScriptCapitalD]\",FontColor->RGBColor[1., 0., 0.]]\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0., 0.]]\)",
                     " ", " "}, {
                    " ", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0., \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0., 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`ChristoffelcdPDschw[{1, $CellContext`schw}, {
                    2, -$CellContext`schw}, {2, -$CellContext`schw}], 
                    Editable -> False], "-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\[EmptyDownTriangle],\!\(\*StyleBox[\"\
\[ScriptCapitalD]\",FontColor->RGBColor[1., 0., 0.]]\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0., 0.]]\)",
                     " ", " "}, {
                    " ", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0., \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0., 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`ChristoffelcdPDschw[{1, $CellContext`schw}, {
                    3, -$CellContext`schw}, {3, -$CellContext`schw}], 
                    Editable -> False], " ", 
                    SuperscriptBox[
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"Csc"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], Csc, Editable -> False], "[", 
                    InterpretationBox[
                    StyleBox[
                    "\[Theta]", AutoSpacing -> False, ShowAutoStyles -> 
                    False], 
                    $CellContext`\[Theta][], Editable -> False], "]"}], 
                    "2"]}], "-", 
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\[EmptyDownTriangle],\!\(\*StyleBox[\"\
\[ScriptCapitalD]\",FontColor->RGBColor[1., 0., 0.]]\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0., 0.]]\)",
                     " ", " "}, {
                    " ", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0., \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0., 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`ChristoffelcdPDschw[{1, $CellContext`schw}, {
                    1, -$CellContext`schw}, {1, -$CellContext`schw}], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "-", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\[EmptyDownTriangle],\!\(\*StyleBox[\"\
\[ScriptCapitalD]\",FontColor->RGBColor[1., 0., 0.]]\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0., 0.]]\)",
                     " ", " "}, {
                    " ", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0., \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0., 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`ChristoffelcdPDschw[{1, $CellContext`schw}, {
                    1, -$CellContext`schw}, {1, -$CellContext`schw}], 
                    Editable -> False], " ", 
                    SuperscriptBox[
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"]}]}], ")"}]}], 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox[
                  "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                  $CellContext`r[], Editable -> False], "4"]]}]],
            
            Annotation[#, {1, 
              1} -> ($CellContext`mass $CellContext`r[]^(-4)) (
               3 $CellContext`mass - $CellContext`ChristoffelcdPDschw[{
                1, $CellContext`schw}, {2, -$CellContext`schw}, {
                2, -$CellContext`schw}] - $CellContext`ChristoffelcdPDschw[{
                 1, $CellContext`schw}, {3, -$CellContext`schw}, {
                 3, -$CellContext`schw}] Csc[
                  $CellContext`\[Theta][]]^2 - 
               2 $CellContext`r[] + ((
                  2 $CellContext`mass) $CellContext`ChristoffelcdPDschw[{
                   1, $CellContext`schw}, {1, -$CellContext`schw}, {
                   1, -$CellContext`schw}]) $CellContext`r[] - \
$CellContext`ChristoffelcdPDschw[{1, $CellContext`schw}, {
                 1, -$CellContext`schw}, {
                 1, -$CellContext`schw}] $CellContext`r[]^2), "Tooltip"]& ], 
           TagBox[
            TooltipBox["0",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"1", ",", "2"}], "}"}], "\[Rule]", "0"}]],
            Annotation[#, {1, 2} -> 0, "Tooltip"]& ], 
           TagBox[
            TooltipBox["0",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"1", ",", "3"}], "}"}], "\[Rule]", "0"}]],
            Annotation[#, {1, 3} -> 0, "Tooltip"]& ], 
           TagBox[
            TooltipBox["0",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"1", ",", "4"}], "}"}], "\[Rule]", "0"}]],
            Annotation[#, {1, 4} -> 0, "Tooltip"]& ]},
          {
           TagBox[
            TooltipBox["0",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"2", ",", "1"}], "}"}], "\[Rule]", "0"}]],
            Annotation[#, {2, 1} -> 0, "Tooltip"]& ], 
           TagBox[
            TooltipBox["\<\"\[GrayCircle]\"\>",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"2", ",", "2"}], "}"}], "\[Rule]", 
               FractionBox[
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], " ", 
                  RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "3"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], "+", 
                    RowBox[{"2", " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], "+", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\[EmptyDownTriangle],\!\(\*StyleBox[\"\
\[ScriptCapitalD]\",FontColor->RGBColor[1., 0., 0.]]\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0., 0.]]\)",
                     " ", " "}, {
                    " ", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0., \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0., 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`ChristoffelcdPDschw[{1, $CellContext`schw}, {
                    1, -$CellContext`schw}, {1, -$CellContext`schw}], 
                    Editable -> False], " ", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]}], 
                    ")"}]}], 
                RowBox[{
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False], "2"], " ", 
                  SuperscriptBox[
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}], "2"]}]]}]],
            
            
            Annotation[#, {2, 
              2} -> (($CellContext`mass $CellContext`r[]^(-2)) ((-2) \
$CellContext`mass + $CellContext`r[])^(-2)) ((-3) $CellContext`mass + 
               2 $CellContext`r[] + ($CellContext`ChristoffelcdPDschw[{
                   1, $CellContext`schw}, {1, -$CellContext`schw}, {
                   1, -$CellContext`schw}] $CellContext`r[]) ((-2) \
$CellContext`mass + $CellContext`r[])), "Tooltip"]& ], 
           TagBox[
            TooltipBox[
             FractionBox[
              RowBox[{
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`mass,
                Editable->False], " ", 
               InterpretationBox[
                StyleBox[GridBox[{
                   {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{"\[EmptyDownTriangle]", ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                    FontSize->9]}
                  },
                  
                  GridBoxAlignment->{
                   "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`ChristoffelcdPDschw[{1, $CellContext`schw}, {
                 2, -$CellContext`schw}, {1, -$CellContext`schw}],
                Editable->False]}], 
              RowBox[{
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "2"}], " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`mass,
                   Editable->False]}], "+", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], ")"}]}]],
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"2", ",", "3"}], "}"}], "\[Rule]", 
               FractionBox[
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\[EmptyDownTriangle],\!\(\*StyleBox[\"\
\[ScriptCapitalD]\",FontColor->RGBColor[1., 0., 0.]]\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0., 0.]]\)",
                     " ", " "}, {
                    " ", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0., \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0., 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelcdPDschw[{1, $CellContext`schw}, {
                    2, -$CellContext`schw}, {1, -$CellContext`schw}], 
                   Editable -> False]}], 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False], " ", 
                  RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]]}]],
            
            Annotation[#, {2, 
              3} -> ($CellContext`mass $CellContext`ChristoffelcdPDschw[{
                 1, $CellContext`schw}, {2, -$CellContext`schw}, {
                 1, -$CellContext`schw}]) ($CellContext`r[]^(-1)/((-2) \
$CellContext`mass + $CellContext`r[])), "Tooltip"]& ], 
           TagBox[
            TooltipBox[
             FractionBox[
              RowBox[{
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`mass,
                Editable->False], " ", 
               InterpretationBox[
                StyleBox[GridBox[{
                   {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{"\[EmptyDownTriangle]", ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0., 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                    FontSize->9]}
                  },
                  
                  GridBoxAlignment->{
                   "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`ChristoffelcdPDschw[{1, $CellContext`schw}, {
                 3, -$CellContext`schw}, {1, -$CellContext`schw}],
                Editable->False]}], 
              RowBox[{
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "2"}], " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`mass,
                   Editable->False]}], "+", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], ")"}]}]],
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"2", ",", "4"}], "}"}], "\[Rule]", 
               FractionBox[
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\[EmptyDownTriangle],\!\(\*StyleBox[\"\
\[ScriptCapitalD]\",FontColor->RGBColor[1., 0., 0.]]\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0., 0.]]\)",
                     " ", " "}, {
                    " ", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0., \
0.]]\)", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0., 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelcdPDschw[{1, $CellContext`schw}, {
                    3, -$CellContext`schw}, {1, -$CellContext`schw}], 
                   Editable -> False]}], 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False], " ", 
                  RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]]}]],
            
            Annotation[#, {2, 
              4} -> ($CellContext`mass $CellContext`ChristoffelcdPDschw[{
                 1, $CellContext`schw}, {3, -$CellContext`schw}, {
                 1, -$CellContext`schw}]) ($CellContext`r[]^(-1)/((-2) \
$CellContext`mass + $CellContext`r[])), "Tooltip"]& ]},
          {
           TagBox[
            TooltipBox["0",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"3", ",", "1"}], "}"}], "\[Rule]", "0"}]],
            Annotation[#, {3, 1} -> 0, "Tooltip"]& ], 
           TagBox[
            TooltipBox[
             FractionBox[
              RowBox[{
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`mass,
                Editable->False], " ", 
               InterpretationBox[
                StyleBox[GridBox[{
                   {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{"\[EmptyDownTriangle]", ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], " ", " "},
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                    FontSize->9]}
                  },
                  
                  GridBoxAlignment->{
                   "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`ChristoffelcdPDschw[{1, $CellContext`schw}, {
                 1, -$CellContext`schw}, {2, -$CellContext`schw}],
                Editable->False]}], 
              RowBox[{
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "2"}], " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`mass,
                   Editable->False]}], "+", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], ")"}]}]],
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"3", ",", "2"}], "}"}], "\[Rule]", 
               FractionBox[
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\[EmptyDownTriangle],\!\(\*StyleBox[\"\
\[ScriptCapitalD]\",FontColor->RGBColor[1., 0., 0.]]\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0., 0.]]\)",
                     " ", " "}, {
                    " ", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0., \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0., 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelcdPDschw[{1, $CellContext`schw}, {
                    1, -$CellContext`schw}, {2, -$CellContext`schw}], 
                   Editable -> False]}], 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False], " ", 
                  RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]]}]],
            
            Annotation[#, {3, 
              2} -> ($CellContext`mass $CellContext`ChristoffelcdPDschw[{
                 1, $CellContext`schw}, {1, -$CellContext`schw}, {
                 2, -$CellContext`schw}]) ($CellContext`r[]^(-1)/((-2) \
$CellContext`mass + $CellContext`r[])), "Tooltip"]& ], 
           TagBox[
            TooltipBox[
             FractionBox[
              RowBox[{
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`mass,
                Editable->False], " ", 
               InterpretationBox[
                StyleBox[GridBox[{
                   {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{"\[EmptyDownTriangle]", ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                    FontSize->9]}
                  },
                  
                  GridBoxAlignment->{
                   "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`ChristoffelcdPDschw[{1, $CellContext`schw}, {
                 2, -$CellContext`schw}, {2, -$CellContext`schw}],
                Editable->False]}], 
              RowBox[{
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "2"}], " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`mass,
                   Editable->False]}], "+", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], ")"}]}]],
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"3", ",", "3"}], "}"}], "\[Rule]", 
               FractionBox[
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\[EmptyDownTriangle],\!\(\*StyleBox[\"\
\[ScriptCapitalD]\",FontColor->RGBColor[1., 0., 0.]]\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0., 0.]]\)",
                     " ", " "}, {
                    " ", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0., \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0., 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelcdPDschw[{1, $CellContext`schw}, {
                    2, -$CellContext`schw}, {2, -$CellContext`schw}], 
                   Editable -> False]}], 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False], " ", 
                  RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]]}]],
            
            Annotation[#, {3, 
              3} -> ($CellContext`mass $CellContext`ChristoffelcdPDschw[{
                 1, $CellContext`schw}, {2, -$CellContext`schw}, {
                 2, -$CellContext`schw}]) ($CellContext`r[]^(-1)/((-2) \
$CellContext`mass + $CellContext`r[])), "Tooltip"]& ], 
           TagBox[
            TooltipBox[
             FractionBox[
              RowBox[{
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`mass,
                Editable->False], " ", 
               InterpretationBox[
                StyleBox[GridBox[{
                   {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{"\[EmptyDownTriangle]", ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0., 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                    FontSize->9]}
                  },
                  
                  GridBoxAlignment->{
                   "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`ChristoffelcdPDschw[{1, $CellContext`schw}, {
                 3, -$CellContext`schw}, {2, -$CellContext`schw}],
                Editable->False]}], 
              RowBox[{
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "2"}], " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`mass,
                   Editable->False]}], "+", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], ")"}]}]],
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"3", ",", "4"}], "}"}], "\[Rule]", 
               FractionBox[
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\[EmptyDownTriangle],\!\(\*StyleBox[\"\
\[ScriptCapitalD]\",FontColor->RGBColor[1., 0., 0.]]\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0., 0.]]\)",
                     " ", " "}, {
                    " ", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0., \
0.]]\)", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0., 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelcdPDschw[{1, $CellContext`schw}, {
                    3, -$CellContext`schw}, {2, -$CellContext`schw}], 
                   Editable -> False]}], 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False], " ", 
                  RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]]}]],
            
            Annotation[#, {3, 
              4} -> ($CellContext`mass $CellContext`ChristoffelcdPDschw[{
                 1, $CellContext`schw}, {3, -$CellContext`schw}, {
                 2, -$CellContext`schw}]) ($CellContext`r[]^(-1)/((-2) \
$CellContext`mass + $CellContext`r[])), "Tooltip"]& ]},
          {
           TagBox[
            TooltipBox["0",
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"4", ",", "1"}], "}"}], "\[Rule]", "0"}]],
            Annotation[#, {4, 1} -> 0, "Tooltip"]& ], 
           TagBox[
            TooltipBox[
             FractionBox[
              RowBox[{
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`mass,
                Editable->False], " ", 
               InterpretationBox[
                StyleBox[GridBox[{
                   {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{"\[EmptyDownTriangle]", ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], " ", " "},
                    {" ", 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0., 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                    FontSize->9]}
                  },
                  
                  GridBoxAlignment->{
                   "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`ChristoffelcdPDschw[{1, $CellContext`schw}, {
                 1, -$CellContext`schw}, {3, -$CellContext`schw}],
                Editable->False]}], 
              RowBox[{
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "2"}], " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`mass,
                   Editable->False]}], "+", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], ")"}]}]],
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"4", ",", "2"}], "}"}], "\[Rule]", 
               FractionBox[
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\[EmptyDownTriangle],\!\(\*StyleBox[\"\
\[ScriptCapitalD]\",FontColor->RGBColor[1., 0., 0.]]\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0., 0.]]\)",
                     " ", " "}, {
                    " ", "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0., \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0., 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelcdPDschw[{1, $CellContext`schw}, {
                    1, -$CellContext`schw}, {3, -$CellContext`schw}], 
                   Editable -> False]}], 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False], " ", 
                  RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]]}]],
            Annotation[#, {4, 
              2} -> ($CellContext`mass $CellContext`ChristoffelcdPDschw[{
                 1, $CellContext`schw}, {1, -$CellContext`schw}, {
                 3, -$CellContext`schw}]) ($CellContext`r[]^(-1)/((-2) \
$CellContext`mass + $CellContext`r[])), "Tooltip"]& ], 
           TagBox[
            TooltipBox[
             FractionBox[
              RowBox[{
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`mass,
                Editable->False], " ", 
               InterpretationBox[
                StyleBox[GridBox[{
                   {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{"\[EmptyDownTriangle]", ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], " ", " "},
                    {" ", 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0., 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                    FontSize->9]}
                  },
                  
                  GridBoxAlignment->{
                   "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`ChristoffelcdPDschw[{1, $CellContext`schw}, {
                 2, -$CellContext`schw}, {3, -$CellContext`schw}],
                Editable->False]}], 
              RowBox[{
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "2"}], " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`mass,
                   Editable->False]}], "+", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], ")"}]}]],
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"4", ",", "3"}], "}"}], "\[Rule]", 
               FractionBox[
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\[EmptyDownTriangle],\!\(\*StyleBox[\"\
\[ScriptCapitalD]\",FontColor->RGBColor[1., 0., 0.]]\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0., 0.]]\)",
                     " ", " "}, {
                    " ", "\!\(\*StyleBox[\"2\",FontColor->RGBColor[1., 0., \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0., 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelcdPDschw[{1, $CellContext`schw}, {
                    2, -$CellContext`schw}, {3, -$CellContext`schw}], 
                   Editable -> False]}], 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False], " ", 
                  RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]]}]],
            
            Annotation[#, {4, 
              3} -> ($CellContext`mass $CellContext`ChristoffelcdPDschw[{
                 1, $CellContext`schw}, {2, -$CellContext`schw}, {
                 3, -$CellContext`schw}]) ($CellContext`r[]^(-1)/((-2) \
$CellContext`mass + $CellContext`r[])), "Tooltip"]& ], 
           TagBox[
            TooltipBox[
             FractionBox[
              RowBox[{
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`mass,
                Editable->False], " ", 
               InterpretationBox[
                StyleBox[GridBox[{
                   {
                    RowBox[{"\[CapitalGamma]", "[", 
                    RowBox[{"\[EmptyDownTriangle]", ",", 
                    StyleBox["\[ScriptCapitalD]",
                    FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
                    StyleBox[GridBox[{
                    {
                    StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], " ", " "},
                    {" ", 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0., 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0., 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                    FontSize->9]}
                  },
                  
                  GridBoxAlignment->{
                   "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`ChristoffelcdPDschw[{1, $CellContext`schw}, {
                 3, -$CellContext`schw}, {3, -$CellContext`schw}],
                Editable->False]}], 
              RowBox[{
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "2"}], " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`mass,
                   Editable->False]}], "+", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], ")"}]}]],
             RowBox[{
               RowBox[{"{", 
                 RowBox[{"4", ",", "4"}], "}"}], "\[Rule]", 
               FractionBox[
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False], " ", 
                  InterpretationBox[
                   StyleBox[
                    
                    GridBox[{{
                    "\[CapitalGamma][\[EmptyDownTriangle],\!\(\*StyleBox[\"\
\[ScriptCapitalD]\",FontColor->RGBColor[1., 0., 0.]]\)]", 
                    StyleBox[
                    
                    GridBox[{{
                    "\!\(\*StyleBox[\"1\",FontColor->RGBColor[1., 0., 0.]]\)",
                     " ", " "}, {
                    " ", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0., \
0.]]\)", "\!\(\*StyleBox[\"3\",FontColor->RGBColor[1., 0., 0.]]\)"}}, 
                    RowSpacings -> 0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`ChristoffelcdPDschw[{1, $CellContext`schw}, {
                    3, -$CellContext`schw}, {3, -$CellContext`schw}], 
                   Editable -> False]}], 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                   "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                   $CellContext`r[], Editable -> False], " ", 
                  RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{"M"}], AutoSpacing -> False, ShowAutoStyles -> 
                    False], $CellContext`mass, Editable -> False]}], "+", 
                    InterpretationBox[
                    StyleBox[
                    "r", AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`r[], Editable -> False]}], ")"}]}]]}]],
            
            Annotation[#, {4, 
              4} -> ($CellContext`mass $CellContext`ChristoffelcdPDschw[{
                 1, $CellContext`schw}, {3, -$CellContext`schw}, {
                 3, -$CellContext`schw}]) ($CellContext`r[]^(-1)/((-2) \
$CellContext`mass + $CellContext`r[])), "Tooltip"]& ]}
         },
         AutoDelete->False,
         GridBoxDividers->{
          "ColumnsIndexed" -> {
            1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}, 
           "RowsIndexed" -> {1 -> RGBColor[1, 0, 0], 5 -> RGBColor[1, 0, 0]}}],
        FontSize->Small], 
       StyleBox[GridBox[{
          {" ", " "},
          {"\[Mu]", "\[Nu]"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xCoba`CTensor[{{($CellContext`mass $CellContext`r[]^(-4)) (
       3 $CellContext`mass - $CellContext`ChristoffelcdPDschw[{
        1, $CellContext`schw}, {2, -$CellContext`schw}, {
        2, -$CellContext`schw}] - $CellContext`ChristoffelcdPDschw[{
         1, $CellContext`schw}, {3, -$CellContext`schw}, {
         3, -$CellContext`schw}] Csc[
          $CellContext`\[Theta][]]^2 - 
       2 $CellContext`r[] + ((
          2 $CellContext`mass) $CellContext`ChristoffelcdPDschw[{
           1, $CellContext`schw}, {1, -$CellContext`schw}, {
           1, -$CellContext`schw}]) $CellContext`r[] - \
$CellContext`ChristoffelcdPDschw[{1, $CellContext`schw}, {
         1, -$CellContext`schw}, {1, -$CellContext`schw}] $CellContext`r[]^2),
       0, 0, 0}, {
     0, (($CellContext`mass $CellContext`r[]^(-2)) ((-2) $CellContext`mass + \
$CellContext`r[])^(-2)) ((-3) $CellContext`mass + 
       2 $CellContext`r[] + ($CellContext`ChristoffelcdPDschw[{
           1, $CellContext`schw}, {1, -$CellContext`schw}, {
           1, -$CellContext`schw}] $CellContext`r[]) ((-2) $CellContext`mass + \
$CellContext`r[])), ($CellContext`mass $CellContext`ChristoffelcdPDschw[{
         1, $CellContext`schw}, {2, -$CellContext`schw}, {
         1, -$CellContext`schw}]) ($CellContext`r[]^(-1)/((-2) \
$CellContext`mass + $CellContext`r[])), ($CellContext`mass \
$CellContext`ChristoffelcdPDschw[{1, $CellContext`schw}, {
         3, -$CellContext`schw}, {
         1, -$CellContext`schw}]) ($CellContext`r[]^(-1)/((-2) \
$CellContext`mass + $CellContext`r[]))}, {
     0, ($CellContext`mass $CellContext`ChristoffelcdPDschw[{
         1, $CellContext`schw}, {1, -$CellContext`schw}, {
         2, -$CellContext`schw}]) ($CellContext`r[]^(-1)/((-2) \
$CellContext`mass + $CellContext`r[])), ($CellContext`mass \
$CellContext`ChristoffelcdPDschw[{1, $CellContext`schw}, {
         2, -$CellContext`schw}, {
         2, -$CellContext`schw}]) ($CellContext`r[]^(-1)/((-2) \
$CellContext`mass + $CellContext`r[])), ($CellContext`mass \
$CellContext`ChristoffelcdPDschw[{1, $CellContext`schw}, {
         3, -$CellContext`schw}, {
         2, -$CellContext`schw}]) ($CellContext`r[]^(-1)/((-2) \
$CellContext`mass + $CellContext`r[]))}, {
     0, ($CellContext`mass $CellContext`ChristoffelcdPDschw[{
         1, $CellContext`schw}, {1, -$CellContext`schw}, {
         3, -$CellContext`schw}]) ($CellContext`r[]^(-1)/((-2) \
$CellContext`mass + $CellContext`r[])), ($CellContext`mass \
$CellContext`ChristoffelcdPDschw[{1, $CellContext`schw}, {
         2, -$CellContext`schw}, {
         3, -$CellContext`schw}]) ($CellContext`r[]^(-1)/((-2) \
$CellContext`mass + $CellContext`r[])), ($CellContext`mass \
$CellContext`ChristoffelcdPDschw[{1, $CellContext`schw}, {
         3, -$CellContext`schw}, {
         3, -$CellContext`schw}]) ($CellContext`r[]^(-1)/((-2) \
$CellContext`mass + $CellContext`r[]))}}, {-$CellContext`schw, \
-$CellContext`schw}, 0][-$CellContext`\[Mu], -$CellContext`\[Nu]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728677596276428*^9},
 CellLabel->"Out[27]=",
 CellID->1816803118,ExpressionUUID->"7d4b1263-61c8-47d7-86df-c87dc3de5e6f"]
}, Open  ]]
}, Open  ]]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{

Cell["4. TensorValues framework", "Section",
 CellFrame->{{0, 0}, {0, 0.5}},
 CellChangeTimes->{{3.7286627354220657`*^9, 3.728662748172317*^9}, {
  3.728664037492599*^9, 3.72866403885896*^9}, {3.7286641258241453`*^9, 
  3.7286641646769876`*^9}, {3.728670489894273*^9, 3.728670492830665*^9}, {
  3.728670919300784*^9, 3.728670923186561*^9}, {3.728673377395557*^9, 
  3.728673379040707*^9}, {3.7286735058905783`*^9, 3.72867350875427*^9}, {
  3.728675282031928*^9, 3.728675282407818*^9}},
 FontSize->18,
 FontColor->GrayLevel[0],
 CellTags->{"S5.18.1", "1.4"},
 CellID->1577395714,ExpressionUUID->"805539ef-7108-4523-9d24-2b919bf16083"],

Cell[CellGroupData[{

Cell["4.0. Setup", "Subsection",
 CellChangeTimes->{
  3.728670949025199*^9, {3.728673515985249*^9, 3.728673518098227*^9}, {
   3.728675283503668*^9, 3.728675283623868*^9}, {3.728753096363196*^9, 
   3.728753104197418*^9}},
 CellID->62024482,ExpressionUUID->"bfe53780-fd39-4e4e-893e-588c67b819d0"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<xAct`xCoba`\>\"", "]"}]], "Input",
 CellChangeTimes->{{3.728752132661909*^9, 3.7287521438570213`*^9}},
 CellLabel->"In[1]:=",
 CellID->950675144,ExpressionUUID->"ab18394d-e13a-4e76-96c3-fb28eaecbdd8"],

Cell[CellGroupData[{

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.728752145889915*^9, 3.7287522713321753`*^9, 
  3.728752604122591*^9, 3.728752649879531*^9, 3.7287530733683043`*^9, 
  3.728753131845132*^9, 3.7309131203844347`*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->2091747396,ExpressionUUID->"b316b630-1e29-407b-8914-39b46f136e77"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`xPerm`  version \"\>", 
   "\[InvisibleSpace]", "\<\"1.2.3\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2015", ",", "8", ",", "23"}], "}"}]}],
  SequenceForm["Package xAct`xPerm`  version ", "1.2.3", ", ", {2015, 8, 23}],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.728752145889915*^9, 3.7287522713321753`*^9, 
  3.728752604122591*^9, 3.728752649879531*^9, 3.7287530733683043`*^9, 
  3.728753131845132*^9, 3.730913120399169*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1691183182,ExpressionUUID->"c9bb89b0-c65a-4642-9e96-70aeea598ee5"],

Cell[BoxData["\<\"CopyRight (C) 2003-2018, Jose M. Martin-Garcia, under the \
General Public License.\"\>"], "Print",
 CellChangeTimes->{3.728752145889915*^9, 3.7287522713321753`*^9, 
  3.728752604122591*^9, 3.728752649879531*^9, 3.7287530733683043`*^9, 
  3.728753131845132*^9, 3.730913120415394*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1769671829,ExpressionUUID->"b94f223c-38be-451b-8918-a9cd71aeba81"],

Cell[BoxData["\<\"Connecting to external mac executable...\"\>"], "Print",
 CellChangeTimes->{3.728752145889915*^9, 3.7287522713321753`*^9, 
  3.728752604122591*^9, 3.728752649879531*^9, 3.7287530733683043`*^9, 
  3.728753131845132*^9, 3.730913120426551*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->225919053,ExpressionUUID->"1850d674-e5d4-409f-926b-4d15bcb316e9"],

Cell[BoxData["\<\"Connection established.\"\>"], "Print",
 CellChangeTimes->{3.728752145889915*^9, 3.7287522713321753`*^9, 
  3.728752604122591*^9, 3.728752649879531*^9, 3.7287530733683043`*^9, 
  3.728753131845132*^9, 3.730913120438143*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->270435875,ExpressionUUID->"480143ef-9b16-457f-9e1c-76b2dfc316f7"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.728752145889915*^9, 3.7287522713321753`*^9, 
  3.728752604122591*^9, 3.728752649879531*^9, 3.7287530733683043`*^9, 
  3.728753131845132*^9, 3.730913120448965*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1140354073,ExpressionUUID->"94ae02de-dbb5-4b90-a739-c8db3d7b420a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`xTensor`  version \"\>", 
   "\[InvisibleSpace]", "\<\"1.1.3\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2018", ",", "2", ",", "28"}], "}"}]}],
  SequenceForm[
  "Package xAct`xTensor`  version ", "1.1.3", ", ", {2018, 2, 28}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752145889915*^9, 3.7287522713321753`*^9, 
  3.728752604122591*^9, 3.728752649879531*^9, 3.7287530733683043`*^9, 
  3.728753131845132*^9, 3.730913120459993*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1786839083,ExpressionUUID->"20a8c96d-a1bc-44a3-90bf-40981c504c2d"],

Cell[BoxData["\<\"CopyRight (C) 2002-2018, Jose M. Martin-Garcia, under the \
General Public License.\"\>"], "Print",
 CellChangeTimes->{3.728752145889915*^9, 3.7287522713321753`*^9, 
  3.728752604122591*^9, 3.728752649879531*^9, 3.7287530733683043`*^9, 
  3.728753131845132*^9, 3.730913120470769*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->993148333,ExpressionUUID->"90b29920-e2a0-4ef8-8828-399ff621624f"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.728752145889915*^9, 3.7287522713321753`*^9, 
  3.728752604122591*^9, 3.728752649879531*^9, 3.7287530733683043`*^9, 
  3.728753131845132*^9, 3.730913120551237*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1601128247,ExpressionUUID->"59af40eb-cfc7-4486-b227-fb4f06c43d5c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`xCoba`  version \"\>", 
   "\[InvisibleSpace]", "\<\"0.8.4\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2018", ",", "2", ",", "28"}], "}"}]}],
  SequenceForm["Package xAct`xCoba`  version ", "0.8.4", ", ", {2018, 2, 28}],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.728752145889915*^9, 3.7287522713321753`*^9, 
  3.728752604122591*^9, 3.728752649879531*^9, 3.7287530733683043`*^9, 
  3.728753131845132*^9, 3.730913120562801*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->44866772,ExpressionUUID->"f11b693d-710c-4f31-bf50-51980d5fa69c"],

Cell[BoxData["\<\"CopyRight (C) 2005-2018, David Yllanes and Jose M. \
Martin-Garcia, under the General Public License.\"\>"], "Print",
 CellChangeTimes->{3.728752145889915*^9, 3.7287522713321753`*^9, 
  3.728752604122591*^9, 3.728752649879531*^9, 3.7287530733683043`*^9, 
  3.728753131845132*^9, 3.730913120573729*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->309076907,ExpressionUUID->"bb96e10d-0e58-4266-bc07-648da98ac212"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.728752145889915*^9, 3.7287522713321753`*^9, 
  3.728752604122591*^9, 3.728752649879531*^9, 3.7287530733683043`*^9, 
  3.728753131845132*^9, 3.730913120584551*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->2040117503,ExpressionUUID->"e89f4839-648f-4622-b014-05182bb97456"],

Cell[BoxData["\<\"These packages come with ABSOLUTELY NO WARRANTY; for \
details type Disclaimer[]. This is free software, and you are welcome to \
redistribute it under certain conditions. See the General Public License for \
details.\"\>"], "Print",
 CellChangeTimes->{3.728752145889915*^9, 3.7287522713321753`*^9, 
  3.728752604122591*^9, 3.728752649879531*^9, 3.7287530733683043`*^9, 
  3.728753131845132*^9, 3.730913120595944*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1530964919,ExpressionUUID->"2da79110-5701-4d23-8ca2-f5429d1a31f0"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.728752145889915*^9, 3.7287522713321753`*^9, 
  3.728752604122591*^9, 3.728752649879531*^9, 3.7287530733683043`*^9, 
  3.728753131845132*^9, 3.730913120606131*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1925818926,ExpressionUUID->"c2a5344f-ace1-444b-a0a4-e28a88b55eef"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M3", ",", " ", "3", ",", " ", 
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f"}], "}"}]}], 
  "]"}]], "Input",
 CellLabel->"In[2]:=",
 CellID->380146108,ExpressionUUID->"6c536c3e-df76-424e-8cdb-6747624148af"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefManifold", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"manifold \"\>", "\[InvisibleSpace]", "M3", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefManifold, ": Defining ", 
   "manifold ", $CellContext`M3, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752149133504*^9, 3.728752274742299*^9, 
  3.728752604666127*^9, 3.7287526512910624`*^9, 3.7287530738516808`*^9, 
  3.728753132922365*^9, 3.730913123707604*^9},
 CellLabel->"During evaluation of In[2]:=",
 CellID->584682180,ExpressionUUID->"d852119a-aa18-499b-aeb8-0a2498028da0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefVBundle", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle \"\>", "\[InvisibleSpace]", "TangentM3", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefVBundle, ": Defining ", 
   "vbundle ", $CellContext`TangentM3, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752149133504*^9, 3.728752274742299*^9, 
  3.728752604666127*^9, 3.7287526512910624`*^9, 3.7287530738516808`*^9, 
  3.728753132922365*^9, 3.730913123721814*^9},
 CellLabel->"During evaluation of In[2]:=",
 CellID->2015714235,ExpressionUUID->"ec161548-a0fb-416e-8ff3-b620cdfc1d59"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"v", "[", "a", "]"}], ",", "M3"}], "]"}], "\[IndentingNewLine]", 
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", "M3", ",", 
   RowBox[{"Antisymmetric", "[", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}]}], "}"}], "]"}]}], "]"}]}], "Input",
 CellLabel->"In[3]:=",
 CellID->139030448,ExpressionUUID->"0eafa0e3-dea9-414a-8891-a1680a1273aa"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"v", "[", "a", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`v[$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752229864522*^9, 3.728752275092703*^9, 
  3.728752604859099*^9, 3.728752653222728*^9, 3.7287530740831614`*^9, 
  3.728753133294174*^9, 3.730913123989715*^9},
 CellLabel->"During evaluation of In[3]:=",
 CellID->2126631414,ExpressionUUID->"d602889d-eae9-498d-8271-2b215fcc01c8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"T", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`T[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752229864522*^9, 3.728752275092703*^9, 
  3.728752604859099*^9, 3.728752653222728*^9, 3.7287530740831614`*^9, 
  3.728753133294174*^9, 3.730913124005471*^9},
 CellLabel->"During evaluation of In[3]:=",
 CellID->1723838077,ExpressionUUID->"c59fc2a5-dc8e-4b8a-8a29-24f7b9545473"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefBasis", "[", 
  RowBox[{"polar", ",", "TangentM3", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[5]:=",
 CellID->1700707167,ExpressionUUID->"4e3d7d18-3836-4177-a83f-b6a43f4649fa"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefBasis", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"basis \"\>", "\[InvisibleSpace]", "polar", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xCoba`DefBasis, ": Defining ", "basis ", $CellContext`polar, 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752180301803*^9, 3.728752276928277*^9, 
  3.72875260692347*^9, 3.7287526536873693`*^9, 3.72875307427592*^9, 
  3.7287531353102293`*^9, 3.730913124754068*^9},
 CellLabel->"During evaluation of In[5]:=",
 CellID->62698460,ExpressionUUID->"44062c32-4269-477a-829e-ec76593bcac8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"parallel derivative \"\>", "\[InvisibleSpace]", 
   RowBox[{"PDpolar", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "parallel derivative ", 
   $CellContext`PDpolar[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752180301803*^9, 3.728752276928277*^9, 
  3.72875260692347*^9, 3.7287526536873693`*^9, 3.72875307427592*^9, 
  3.7287531353102293`*^9, 3.7309131247700577`*^9},
 CellLabel->"During evaluation of In[5]:=",
 CellID->1506030190,ExpressionUUID->"15aaf5e3-b3f3-4817-a75d-63ae51949296"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"TorsionPDpolar", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "torsion tensor ", 
   $CellContext`TorsionPDpolar[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752180301803*^9, 3.728752276928277*^9, 
  3.72875260692347*^9, 3.7287526536873693`*^9, 3.72875307427592*^9, 
  3.7287531353102293`*^9, 3.730913124783428*^9},
 CellLabel->"During evaluation of In[5]:=",
 CellID->1607139788,ExpressionUUID->"8e696d68-8f2b-4c47-822d-1d6aff5b193e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelPDpolar", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "non-symmetric Christoffel tensor ", 
   $CellContext`ChristoffelPDpolar[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752180301803*^9, 3.728752276928277*^9, 
  3.72875260692347*^9, 3.7287526536873693`*^9, 3.72875307427592*^9, 
  3.7287531353102293`*^9, 3.7309131247919407`*^9},
 CellLabel->"During evaluation of In[5]:=",
 CellID->149104950,ExpressionUUID->"44534c5f-f066-4aa3-9f1a-5a2349399510"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Riemann tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RiemannPDpolar", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Riemann tensor ", 
   $CellContext`RiemannPDpolar[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, $CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752180301803*^9, 3.728752276928277*^9, 
  3.72875260692347*^9, 3.7287526536873693`*^9, 3.72875307427592*^9, 
  3.7287531353102293`*^9, 3.730913124800398*^9},
 CellLabel->"During evaluation of In[5]:=",
 CellID->2040962248,ExpressionUUID->"48023003-f168-49a3-9528-569f8af55091"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciPDpolar", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Ricci tensor ", 
   $CellContext`RicciPDpolar[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752180301803*^9, 3.728752276928277*^9, 
  3.72875260692347*^9, 3.7287526536873693`*^9, 3.72875307427592*^9, 
  3.7287531353102293`*^9, 3.7309131248082943`*^9},
 CellLabel->"During evaluation of In[5]:=",
 CellID->696208605,ExpressionUUID->"b88eb820-29af-419d-8c16-098209a06686"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaUppolar", "[", 
    RowBox[{"a", ",", "b", ",", "c"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric +1 density ", 
   $CellContext`etaUppolar[$CellContext`a, $CellContext`b, $CellContext`c], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752180301803*^9, 3.728752276928277*^9, 
  3.72875260692347*^9, 3.7287526536873693`*^9, 3.72875307427592*^9, 
  3.7287531353102293`*^9, 3.730913124816264*^9},
 CellLabel->"During evaluation of In[5]:=",
 CellID->1687115379,ExpressionUUID->"74ebb7b9-84e0-46cd-96f8-765a10706b24"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaDownpolar", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric -1 density ", 
   $CellContext`etaDownpolar[-$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752180301803*^9, 3.728752276928277*^9, 
  3.72875260692347*^9, 3.7287526536873693`*^9, 3.72875307427592*^9, 
  3.7287531353102293`*^9, 3.7309131248239822`*^9},
 CellLabel->"During evaluation of In[5]:=",
 CellID->2121583512,ExpressionUUID->"fa9e1df8-6e93-4c2a-b8eb-2ef5c3537dda"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefBasis", "[", 
  RowBox[{"cartesian", ",", "TangentM3", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2"}], "}"}], ",", 
   RowBox[{"BasisColor", "\[Rule]", "Green"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728752257742723*^9, 3.728752258563834*^9}},
 CellLabel->"In[6]:=",
 CellID->384010265,ExpressionUUID->"586d67d7-298f-43dd-80f4-0dd0d12a0c9b"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefBasis", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"basis \"\>", "\[InvisibleSpace]", "cartesian", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xCoba`DefBasis, ": Defining ", "basis ", $CellContext`cartesian,
    ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752188467963*^9, 3.7287522782410183`*^9, 
  3.7287526072424183`*^9, 3.7287526539742203`*^9, 3.7287530761769753`*^9, 
  3.728753135642808*^9, 3.730913125871146*^9},
 CellLabel->"During evaluation of In[6]:=",
 CellID->1347433818,ExpressionUUID->"e64e745c-c74d-4cce-87a2-f0edc3330010"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"parallel derivative \"\>", "\[InvisibleSpace]", 
   RowBox[{"PDcartesian", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "parallel derivative ", 
   $CellContext`PDcartesian[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752188467963*^9, 3.7287522782410183`*^9, 
  3.7287526072424183`*^9, 3.7287526539742203`*^9, 3.7287530761769753`*^9, 
  3.728753135642808*^9, 3.730913125885684*^9},
 CellLabel->"During evaluation of In[6]:=",
 CellID->1105121511,ExpressionUUID->"b2de2025-2e55-4853-b54c-bb8d1d32c508"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"TorsionPDcartesian", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "torsion tensor ", 
   $CellContext`TorsionPDcartesian[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752188467963*^9, 3.7287522782410183`*^9, 
  3.7287526072424183`*^9, 3.7287526539742203`*^9, 3.7287530761769753`*^9, 
  3.728753135642808*^9, 3.730913125898635*^9},
 CellLabel->"During evaluation of In[6]:=",
 CellID->191172474,ExpressionUUID->"0944c7fa-fbb3-48bb-ae5a-da9a0c093898"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelPDcartesian", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "non-symmetric Christoffel tensor ", 
   $CellContext`ChristoffelPDcartesian[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752188467963*^9, 3.7287522782410183`*^9, 
  3.7287526072424183`*^9, 3.7287526539742203`*^9, 3.7287530761769753`*^9, 
  3.728753135642808*^9, 3.730913125906476*^9},
 CellLabel->"During evaluation of In[6]:=",
 CellID->400234606,ExpressionUUID->"71609261-83b7-41ab-99f5-ed70f536996e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Riemann tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RiemannPDcartesian", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Riemann tensor ", 
   $CellContext`RiemannPDcartesian[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, $CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752188467963*^9, 3.7287522782410183`*^9, 
  3.7287526072424183`*^9, 3.7287526539742203`*^9, 3.7287530761769753`*^9, 
  3.728753135642808*^9, 3.730913125914369*^9},
 CellLabel->"During evaluation of In[6]:=",
 CellID->824240638,ExpressionUUID->"e308e820-f449-4ee1-850d-063010d89392"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciPDcartesian", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Ricci tensor ", 
   $CellContext`RicciPDcartesian[-$CellContext`a, -$CellContext`b], ". ", ""],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.728752188467963*^9, 3.7287522782410183`*^9, 
  3.7287526072424183`*^9, 3.7287526539742203`*^9, 3.7287530761769753`*^9, 
  3.728753135642808*^9, 3.730913125924347*^9},
 CellLabel->"During evaluation of In[6]:=",
 CellID->1742879716,ExpressionUUID->"7872fd95-f223-458f-a88d-3b918f45f731"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaUpcartesian", "[", 
    RowBox[{"a", ",", "b", ",", "c"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric +1 density ", 
   $CellContext`etaUpcartesian[$CellContext`a, $CellContext`b, \
$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752188467963*^9, 3.7287522782410183`*^9, 
  3.7287526072424183`*^9, 3.7287526539742203`*^9, 3.7287530761769753`*^9, 
  3.728753135642808*^9, 3.730913125934352*^9},
 CellLabel->"During evaluation of In[6]:=",
 CellID->902441732,ExpressionUUID->"ba88b171-611e-4640-bae3-888ed74739e9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaDowncartesian", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric -1 density ", 
   $CellContext`etaDowncartesian[-$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752188467963*^9, 3.7287522782410183`*^9, 
  3.7287526072424183`*^9, 3.7287526539742203`*^9, 3.7287530761769753`*^9, 
  3.728753135642808*^9, 3.730913125944542*^9},
 CellLabel->"During evaluation of In[6]:=",
 CellID->1808674610,ExpressionUUID->"429ffe24-7887-44a1-b4bb-5ac38fd58627"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["4.1. Assigning values to components", "Subsection",
 CellChangeTimes->{
  3.728670949025199*^9, {3.728673515985249*^9, 3.728673518098227*^9}, {
   3.728675283503668*^9, 3.728675283623868*^9}},
 CellID->25383259,ExpressionUUID->"304f8903-4ce4-4b85-8882-0ad638cee8a5"],

Cell[TextData[{
 "We need some convenient way to store all the values of the components of a \
tensor (or, more generally, an arbitrary expression). ",
 StyleBox["xCoba`", "MR"],
 " provides several tools to accomplish this, optimised to make full use of \
the symmetries. This section describes how to store and use these values, but \
not how to compute them. The latter is the object of Section 9 (not yet fully \
implemented).\n\nGiven a tensor there are several things we need to worry \
about when working with values:\n\t- 1. Are we giving values to the \
components of the tensor itself or to some derivative of it?\n\t- 2. The \
slot-symmetry of the expression.\n\t- 3. What is the character of the \
c-indices involved?\n\t- 4. Which bases are we using to form the component?\n\
\t- 5. When did we construct a particular set of components?\n\t- 6. Have we \
already computed values for all components of the expression?\n\nThis section \
describes a series of functions for the storage of tensor values."
}], "Text",
 CellChangeTimes->{3.730912909788946*^9},
 CellID->1414671907,ExpressionUUID->"ef1e14bf-3778-459c-a757-ad199c2d89ca"]
}, Closed]],

Cell[CellGroupData[{

Cell["4.2. FoldedRule and ValID", "Subsection",
 CellChangeTimes->{
  3.728670949025199*^9, {3.728673515985249*^9, 3.728673518098227*^9}, {
   3.728673590976366*^9, 3.728673596854003*^9}, {3.728675284543667*^9, 
   3.728675284695709*^9}},
 CellID->28781190,ExpressionUUID->"1f61dbb7-79fd-47da-ba7f-326b8fbe6e9d"],

Cell[TextData[{
 StyleBox["FoldedRule\t\t\t", "DisplayMath"],
 StyleBox["List of lists of rules to be applied sequentially\n", "Text"],
 StyleBox["ValID\t\t\t\t", "DisplayMath"],
 StyleBox["Identifier for each independent set of components stored for a \
tensor\n", "Text"],
 StyleBox["TensorValIDs\t\t", "DisplayMath"],
 StyleBox["List of all the ValIDs of a given tensor", "Text"],
 StyleBox["\t\nDateOfValID\t\t\t", "DisplayMath"],
 StyleBox["Time a given ValID was generated", "Text"]
}], "DefinitionBox",
 CellID->1277870832,ExpressionUUID->"a7487b0d-4d4f-44a3-815b-af6141d42a63"],

Cell["Identifying sets of component values.", "Text",
 CellChangeTimes->{3.728751806688648*^9},
 CellID->191119777,ExpressionUUID->"9c2310bc-14b9-4b43-af44-b64cbf2fb2cc"],

Cell[CellGroupData[{

Cell[TextData[{
 "Before introducing the general functions, let us consider a simple example. \
Let us suppose we have a rank two tensor ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["U", "ab"], TraditionalForm]],ExpressionUUID->
  "1c8a6b15-cba8-46dd-aace-45547e4fb444"],
 ", with no symmetries and only one basis. Then, the straightforward way of \
storing its components would be with a list of replacement rules:\n\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      SubscriptBox["U", 
       StyleBox["11",
        FontColor->RGBColor[1, 0, 0]]], "\[Rule]", " ", "a"}], ",", "  ", 
     RowBox[{
      SubscriptBox["U", 
       StyleBox["12",
        FontColor->RGBColor[1, 0, 0]]], "\[Rule]", " ", "b"}], ",", " ", 
     RowBox[{
      SubscriptBox["U", 
       StyleBox["13",
        FontColor->RGBColor[1, 0, 0]]], "\[Rule]", " ", "c"}], ",", " ", 
     "..."}], "}"}], TraditionalForm]],ExpressionUUID->
  "3e18e6ca-31b0-49d9-92b0-7430557a4a92"],
 "\nBut now consider, still with just one basis in play and no metric, that \
our tensor ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["T", 
     StyleBox["ab",
      FontSlant->"Italic"]], " "}], TraditionalForm]],ExpressionUUID->
  "f31e6c77-f018-44b0-9330-bd7b735dafb8"],
 "is symmetric. Then it would be a waste to store twice the component ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["T", 
      StyleBox["12",
       FontColor->RGBColor[1, 0, 0]]], "=", 
     SubscriptBox["T", 
      StyleBox["21",
       FontColor->RGBColor[1, 0, 0]]]}], ",", " ", 
    RowBox[{"for", " ", 
     RowBox[{"example", "."}]}]}], TraditionalForm]],ExpressionUUID->
  "a9abef13-7e51-4012-918f-d85adc4d03ee"],
 " With higher rank tensors and higher symmetries the redundancy makes the \
process very inefficient. Instead, we could think of storing components as a \
pair of lists. The first one would give the relations due to symmetry and the \
second one the values of the independent components. We could have something \
like\n\t{ ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       SubscriptBox["T", 
        StyleBox["21",
         FontColor->RGBColor[1, 0, 0]]], "\[Rule]", " ", 
       SubscriptBox["T", 
        StyleBox["12",
         FontColor->RGBColor[1, 0, 0]]]}], ",", " ", 
      RowBox[{
       SubscriptBox["T", 
        StyleBox["31",
         FontColor->RGBColor[1, 0, 0]]], "\[Rule]", " ", 
       SubscriptBox["T", 
        StyleBox["13",
         FontColor->RGBColor[1, 0, 0]]]}], ",", " ", 
      RowBox[{
       SubscriptBox["T", 
        StyleBox["32",
         FontColor->RGBColor[1, 0, 0]]], "\[Rule]", " ", 
       SubscriptBox["T", 
        StyleBox["23",
         FontColor->RGBColor[1, 0, 0]]]}]}], "}"}], ",", " ", 
    RowBox[{"{", " ", 
     RowBox[{
      RowBox[{
       SubscriptBox["T", 
        StyleBox["11",
         FontColor->RGBColor[1, 0, 0]]], "\[Rule]", " ", "a"}], ",", " ", 
      RowBox[{
       SubscriptBox["T", 
        StyleBox["12",
         FontColor->RGBColor[1, 0, 0]]], "\[Rule]", " ", "b"}], ",", " ", 
      RowBox[{
       SubscriptBox["T", 
        StyleBox["13",
         FontColor->RGBColor[1, 0, 0]]], "\[Rule]", " ", "c"}], ",", " ", 
      RowBox[{
       SubscriptBox["T", 
        StyleBox["22",
         FontColor->RGBColor[1, 0, 0]]], "\[Rule]", " ", "d"}], ",", " ", 
      RowBox[{
       SubscriptBox["T", 
        StyleBox["23",
         FontColor->RGBColor[1, 0, 0]]], "\[Rule]", " ", "e"}]}]}]}], 
   TraditionalForm]],ExpressionUUID->"882b0032-9410-4232-b030-1a6472179f1e"],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      SubscriptBox["T", 
       StyleBox["33",
        FontColor->RGBColor[1, 0, 0]]], "\[Rule]", " ", "f"}], "}"}], " ", 
    "}"}], TraditionalForm]],ExpressionUUID->
  "bb9b6e73-2ea1-453b-bbbc-1897a8151a67"],
 "\nIt may appear we have gained nothing, after all we still have 9 rules. \
However, all the information of the first list is determined by the symmetry \
group of the tensor, which ",
 StyleBox["xCoba`", "MR"],
 " already knows since its definition. Also, the independent values may be \
very large expressions while the dependent rules are always very simple. This \
structure is represented in ",
 StyleBox["xCoba`", "MR"],
 " with the concept of FoldedRule (defined in the auxiliary package ",
 StyleBox["xCore`", "MR"],
 "):"
}], "Text",
 CellID->1662039774,ExpressionUUID->"8fafaf2f-c0e3-46ad-a85a-0ccc8ac4a180"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "FoldedRule"}]], "Input",
 CellLabel->"In[7]:=",
 CellID->1582685089,ExpressionUUID->"a51aae78-aeab-45e4-9bd7-9bfd07bb4852"],

Cell[BoxData["\<\"FoldedRule[rules1, rules2, ...] contains a number of lists \
of rules which are applied sequentially (first rules1, then rules2, etc.) \
when called by ReplaceAll and ReplaceRepeated.\"\>"], "Print", "PrintUsage",
 CellChangeTimes->{3.730913176392028*^9},
 CellTags->"Info73730895176-7882880",
 CellID->745676275,ExpressionUUID->"37a7b42a-eb42-47c4-a943-ae80389a898c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Example:", "MathCaption",
 CellID->2006244096,ExpressionUUID->"3b39fbbc-d0cf-4779-9052-22d06eb19d4e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"x", "+", "t"}], "/.", 
  RowBox[{"FoldedRule", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"x", "\[Rule]", "y"}], ",", 
      RowBox[{"t", "\[Rule]", "x"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{"y", "\[Rule]", "z"}], "}"}]}], "]"}]}]], "Input",
 CellLabel->"In[8]:=",
 CellID->1303779925,ExpressionUUID->"7f01e485-86a2-4c52-b8bf-68896674c4c1"],

Cell[BoxData[
 RowBox[{"x", "+", "z"}]], "Output",
 CellChangeTimes->{3.7286776310400753`*^9, 3.728753072922711*^9, 
  3.72875314492383*^9, 3.7309131781139317`*^9},
 CellLabel->"Out[8]=",
 CellID->1127715014,ExpressionUUID->"ec26cdab-3871-4210-8735-986b6a6a786f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "In the next subsection we shall see how ",
 StyleBox["xCoba`", "MR"],
 "  generates these lists. The next problem comes when we consider a more \
general case where we have a metric and several bases. In this situation we \
may want to store the values for a given tensor in different bases and with \
different index characters. However, not all those combinations are \
independent. Recalling our symmetric tensor ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["T", 
     StyleBox["ab",
      FontSlant->"Italic"]], ",", " ", 
    RowBox[{"we", "  ", "see", " "}]}], TraditionalForm]],ExpressionUUID->
  "516a11be-5977-42a9-a78f-2ad80dc81f12"],
 "that the FoldedRules that would contain its values for ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox[
    SubscriptBox["T", 
     StyleBox["a",
      FontColor->RGBColor[1, 0, 0]]], 
    StyleBox["b",
     FontColor->RGBColor[1, 0, 0]]], TraditionalForm]],ExpressionUUID->
  "c815d5b7-42dc-44d5-9cd6-a42e83674e74"],
 "and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox[
    SuperscriptBox["T", 
     StyleBox["b",
      FontColor->RGBColor[1, 0, 0]]], 
    StyleBox["a",
     FontColor->RGBColor[1, 0, 0]]], TraditionalForm]],ExpressionUUID->
  "74bc1049-0d83-4815-8f75-a7eab1269cda"],
 " would have the same independent components (likewise for ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["T", 
      StyleBox[
       RowBox[{
        StyleBox["a",
         FontColor->RGBColor[1, 0, 0]], 
        StyleBox["b",
         FontColor->RGBColor[0, 1, 0]]}],
       FontSlant->"Italic"]], " ", "and", " ", 
     SubscriptBox["T", 
      StyleBox[
       RowBox[{
        StyleBox["b",
         FontColor->RGBColor[0, 1, 0]], 
        StyleBox["a",
         FontColor->RGBColor[1, 0, 0]]}],
       FontSlant->"Italic"]]}], ")"}], TraditionalForm]],ExpressionUUID->
  "21429fd2-9c42-4957-915b-527fc4ebf34f"],
 ". So among all sets of values for a tensor, only some will be independent. ",
 StyleBox["xCoba`", "MR"],
 "  identifies these independent sets with the concept of ",
 StyleBox["ValID", "MR"],
 ":"
}], "Text",
 CellID->1202511756,ExpressionUUID->"ff79b072-bbc4-4fcb-9ec5-7b23a3ae6995"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "ValID"}]], "Input",
 CellLabel->"In[9]:=",
 CellID->646481175,ExpressionUUID->"da1d17a4-a397-4373-b561-9ce49be87046"],

Cell[BoxData["\<\"ValID[tensor, {{basis1, basis2, ...}}] identifies a set of \
values for the tensor with c-indices respectively in basis1, basis2, ... \
(with sign indicating character as usual; LI and -LI are accepted). For a \
tensor with symmetry, ValID[tensor, {bases1, bases2, ...}] identifies a set \
of values related by that symmetry, where each of the basesi is a list of \
possible reorderings of the bases of the c-indices. ValID[tensor, der1, der2, \
..., bases] identifies a set of values for the given derivatives of the \
tensor in Postfix order (in Prefix notation it would mean \
...[der2[der1[tensor]]), where bases is a list of lists as before.\"\>"], \
"Print", "PrintUsage",
 CellChangeTimes->{3.7309131828268414`*^9},
 CellTags->"Info93730895182-7882880",
 CellID->1068687520,ExpressionUUID->"cd30eeb0-4a22-42fb-9239-de444f399b1e"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "In our example, we would have for our tensor ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["U", 
    StyleBox["ab",
     FontSlant->"Italic"]], TraditionalForm]],ExpressionUUID->
  "e65ceb65-3b7b-4338-8784-9ed270910293"],
 StyleBox[" ",
  FontSlant->"Italic"],
 "a single ValID, which would be\n\t",
 StyleBox["ValID[U, { {polar, polar}} ]", "MR"],
 "\n(both indices in the base polar). We have several for ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["T", 
    StyleBox["ab",
     FontSlant->"Italic"]], TraditionalForm]],ExpressionUUID->
  "f21a2bce-4140-436f-b098-1dca270867f2"],
 "\n\t",
 StyleBox["ValID[T, { {polar, -polar} , {-polar, polar}} ]\n\t ValID[T, { \
{polar, cartesian}, {cartesian, polar}]", "MR"]
}], "Text",
 CellChangeTimes->{{3.728751900582726*^9, 3.728751903004575*^9}},
 CellID->1794641734,ExpressionUUID->"b0ab2621-cff1-4048-9103-de5412f7619c"],

Cell[TextData[{
 "The first would represent the fact that ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox[
    SubscriptBox["T", 
     StyleBox["a",
      FontColor->RGBColor[1, 0, 0]]], 
    StyleBox["b",
     FontColor->RGBColor[1, 0, 0]]], TraditionalForm]],ExpressionUUID->
  "58707a5c-e5f3-4f09-acea-4d93555f7d94"],
 "and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox[
    SuperscriptBox["T", 
     StyleBox["b",
      FontColor->RGBColor[1, 0, 0]]], 
    StyleBox["a",
     FontColor->RGBColor[1, 0, 0]]], TraditionalForm]],ExpressionUUID->
  "f12947b9-cdd8-4d28-80a7-62a3811633b0"],
 " have the same independent rules and the second would represent the \
situation for ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["T", 
     StyleBox[
      RowBox[{
       StyleBox["a",
        FontColor->RGBColor[1, 0, 0]], 
       StyleBox["b",
        FontColor->RGBColor[0, 1, 0]]}],
      FontSlant->"Italic"]], " ", "and", " ", 
    SubscriptBox["T", 
     StyleBox[
      RowBox[{
       StyleBox["b",
        FontColor->RGBColor[0, 1, 0]], 
       StyleBox["a",
        FontColor->RGBColor[1, 0, 0]]}],
      FontSlant->"Italic"]]}], TraditionalForm]],ExpressionUUID->
  "0d238825-9946-4b57-9ab5-9755156cf11c"],
 ". Notice how even in the case of ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["U", 
    StyleBox["ab",
     FontSlant->"Italic"]], TraditionalForm]],ExpressionUUID->
  "f77424ac-de69-444e-93a0-60f38b90eb98"],
 ", where there is only one possible basis configuration in the ",
 StyleBox["ValID,", "MR"],
 " we give a list of lists of bases. The command ",
 StyleBox["VTensorValIDs", "MR"],
 " returns all the ",
 StyleBox["ValIDs ", "MR"],
 " we have stored for a given tensors (and its derivatives, as we shall see \
below).\n\nAll of this will be much clearer once we see how ",
 StyleBox["xCoba`", "MR"],
 "  generates the FoldedRules and their corresponding ",
 StyleBox["ValIDs", "MR"],
 " for a tensor in the next subsection."
}], "Text",
 CellID->1135801223,ExpressionUUID->"3dd0f2b6-1328-4b26-9036-552d3f9151ab"]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
4.3. ComponentValue and TensorValues without independent values\
\>", "Subsection",
 CellChangeTimes->{{3.7286739487605047`*^9, 3.728673950949669*^9}, {
  3.7286752855437737`*^9, 3.728675285703857*^9}},
 CellID->429016778,ExpressionUUID->"e248971e-3f48-4520-a35f-05ae5e6a11e1"],

Cell[TextData[{
 StyleBox["ComponentValue\t\t", "DisplayMath"],
 StyleBox["Give the value of a single component\n", "Text"],
 StyleBox["TensorValues\t\t", "DisplayMath"],
 StyleBox["List of all values for a given tensor\n", "Text"],
 StyleBox["TensorValIDs\t\t", "DisplayMath"],
 StyleBox["List of all the ValIDs of a given tensor", "Text"],
 StyleBox["\t\nDateOfValID\t\t\t", "DisplayMath"],
 StyleBox["Time a given ValID was generated\n", "Text"],
 StyleBox["$CVVerbose\t\t\t", "DisplayMath"],
 StyleBox["Switch verbose output on/off for ", "Text"],
 StyleBox["ComponentValue", "DisplayMath"]
}], "DefinitionBox",
 CellID->1382384867,ExpressionUUID->"abdb9eb6-d49a-4292-bb2b-4e69f1dc3fea"],

Cell["Storing values for components.", "Text",
 CellChangeTimes->{3.728751813712224*^9},
 CellID->116882776,ExpressionUUID->"1b172dc7-54f7-432e-8d34-421622e26724"],

Cell[CellGroupData[{

Cell[TextData[{
 "We have seen how ",
 StyleBox["xCoba`", "MR"],
 " stores and labels sets of tensor values for different basis \
configurations; this section describes how these sets are generated. The \
basic function is ",
 StyleBox["ComponentValue", "MR"],
 ", which generates the rule for a single component. ",
 StyleBox["ComponentValue", "MR"],
 " can be called with one or two arguments:"
}], "Text",
 CellID->1648762874,ExpressionUUID->"5000064a-3ee1-4d8c-a958-239e64756c83"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "ComponentValue"}]], "Input",
 CellLabel->"In[10]:=",
 CellID->1776126077,ExpressionUUID->"8f33011f-a881-4417-a49d-39532004b9e1"],

Cell[BoxData["\<\"ComponentValue[expr], for a component tensor expression (a \
tensor or a derivative of a tensor), returns a value rule of the form expr -> \
canon where canon is the canonical form of expr under ToCanonical. This value \
rule is automatically stored in the TensorValues list of that tensor so that \
it can be used from that list without having to recompute it. \
ComponentValue[expr, value] returns a value rule expr -> value, and stores \
internally both the dependent rule expr -> canon and the independent rule \
canon -> value.\"\>"], "Print", "PrintUsage",
 CellChangeTimes->{3.730913193509823*^9},
 CellTags->"Info103730895193-7882880",
 CellID->1879930841,ExpressionUUID->"0c931954-88f4-4d3a-83e2-46600457c59c"]
}, Open  ]]
}, Open  ]],

Cell["\<\
We begin by examining how it works with just one argument (i.e. without \
specifying an independent value). \
\>", "Text",
 CellChangeTimes->{3.728751937180395*^9},
 CellID->1357164508,ExpressionUUID->"f50445c3-dcf7-495a-bfeb-9f7926b3cfc0"],

Cell[CellGroupData[{

Cell["Let us define a totally antisymmetric tensor:", "MathCaption",
 CellChangeTimes->{{3.728751941002009*^9, 3.7287519426984863`*^9}},
 CellID->1315450794,ExpressionUUID->"09cd64c3-fffe-4730-b2b6-aa0686f27313"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"W", "[", 
    RowBox[{"a", ",", "b", ",", "c"}], "]"}], ",", "M3", ",", 
   RowBox[{"Antisymmetric", "[", 
    RowBox[{"{", 
     RowBox[{"a", ",", "b", ",", "c"}], "}"}], "]"}]}], "]"}]], "Input",
 CellLabel->"In[11]:=",
 CellID->1440071169,ExpressionUUID->"475b3701-8659-4464-9be4-529e753e79f3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"W", "[", 
    RowBox[{"a", ",", "b", ",", "c"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`W[$CellContext`a, $CellContext`b, $CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752192552044*^9, 3.7287522798292303`*^9, 
  3.7287526075249653`*^9, 3.7287526576407127`*^9, 3.7287530766924553`*^9, 
  3.728753149713359*^9, 3.730913195232541*^9},
 CellLabel->"During evaluation of In[11]:=",
 CellID->2059169265,ExpressionUUID->"2ae9d5ab-b66b-4aa1-82ee-31977acccb1e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SetOptions", "[", 
  RowBox[{"CanonicalPerm", ",", " ", 
   RowBox[{"MathLink", "\[Rule]", " ", "True"}]}], "]"}]], "Input",
 CellLabel->"In[12]:=",
 CellID->728005009,ExpressionUUID->"eb908b1a-97b1-4217-b6fb-bdff64fec3a2"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"MathLink", "\[Rule]", "True"}], ",", 
   RowBox[{"TimeVerbose", "\[Rule]", "False"}], ",", 
   RowBox[{"xPermVerbose", "\[Rule]", "False"}], ",", 
   RowBox[{"OrderedBase", "\[Rule]", "True"}]}], "}"}]], "Output",
 CellChangeTimes->{3.7287521933046923`*^9, 3.7287522804713697`*^9, 
  3.7287526076757107`*^9, 3.7287526582222157`*^9, 3.728753076824161*^9, 
  3.7287531498618507`*^9, 3.730913195914637*^9},
 CellLabel->"Out[12]=",
 CellID->1456202210,ExpressionUUID->"1d2f6bc7-e4bf-4876-b514-24a1642dfb5e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Suppose we want to generate a rule for one of its components", \
"MathCaption",
 CellID->1376549545,ExpressionUUID->"01adfdeb-ec38-4b7e-aaa6-2000c015bcea"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"W", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0", ",", "polar"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", "polar"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"2", ",", "polar"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[13]:=",
 CellID->826873684,ExpressionUUID->"cd9779f8-53b8-4ad1-bda8-cea63f1e34fb"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"W", 
      StyleBox[GridBox[{
         {
          StyleBox["0",
           FontColor->RGBColor[1., 0., 0.]], 
          StyleBox["1",
           FontColor->RGBColor[1., 0., 0.]], 
          StyleBox["2",
           FontColor->RGBColor[1., 0., 0.]]},
         {" ", " ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
   2, $CellContext`polar}],
  Editable->False]], "Output",
 CellChangeTimes->{3.7287521945281763`*^9, 3.728752282395976*^9, 
  3.728752608065337*^9, 3.728752659188092*^9, 3.7287530769594088`*^9, 
  3.7287531500027122`*^9, 3.730913197119123*^9},
 CellLabel->"Out[13]=",
 CellID->911245061,ExpressionUUID->"820b7dc3-d2c4-4989-8575-61d0dc282677"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentValue", "[", "%", "]"}]], "Input",
 CellLabel->"In[14]:=",
 CellID->1003199971,ExpressionUUID->"77daa12f-2b7b-4dd3-a1b9-28116b219f35"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
      2, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
      2, $CellContext`polar}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "Added independent rule ", $CellContext`W[{0, $CellContext`polar}, {
     1, $CellContext`polar}, {2, $CellContext`polar}] -> $CellContext`W[{
     0, $CellContext`polar}, {1, $CellContext`polar}, {
     2, $CellContext`polar}], " for tensor ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287521951509047`*^9, 3.7287522832416553`*^9, 
  3.728752608531705*^9, 3.728752660606514*^9, 3.728753077105896*^9, 
  3.728753150389906*^9, 3.730913197763578*^9},
 CellLabel->"During evaluation of In[14]:=",
 CellID->1620176819,ExpressionUUID->"d87ff284-11e2-4227-9248-8901a284ed20"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"W", 
       StyleBox[GridBox[{
          {
           StyleBox["0",
            FontColor->RGBColor[1., 0., 0.]], 
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], 
           StyleBox["2",
            FontColor->RGBColor[1., 0., 0.]]},
          {" ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
    2, $CellContext`polar}],
   Editable->False], "\[Rule]", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"W", 
       StyleBox[GridBox[{
          {
           StyleBox["0",
            FontColor->RGBColor[1., 0., 0.]], 
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], 
           StyleBox["2",
            FontColor->RGBColor[1., 0., 0.]]},
          {" ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
    2, $CellContext`polar}],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.72875219521898*^9, 3.728752283293723*^9, 
  3.728752608573062*^9, 3.7287526606560087`*^9, 3.728753077138187*^9, 
  3.728753150559952*^9, 3.7309131978094063`*^9},
 CellLabel->"Out[14]=",
 CellID->219152973,ExpressionUUID->"68cde602-955c-4465-bfc1-81a942c6736c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TensorValues", "[", "W", "]"}]], "Input",
 CellLabel->"In[15]:=",
 CellID->673242544,ExpressionUUID->"4c541fce-7a3f-4427-8ef8-05f8f0155eef"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"W", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
       2, $CellContext`polar}],
      Editable->False], "\[Rule]", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"W", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
       2, $CellContext`polar}],
      Editable->False]}], "}"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728752196017828*^9, 3.7287522842119293`*^9, 
  3.728752608691506*^9, 3.728752661506044*^9, 3.728753077255026*^9, 
  3.728753150896326*^9, 3.7309131985124397`*^9},
 CellLabel->"Out[15]=",
 CellID->2138622509,ExpressionUUID->"bf39ca3d-ca62-401e-91d2-ed7687bbf5d0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "As we can see,",
 StyleBox[" xCoba`", "MR"],
 " has recognised that ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["W", 
    StyleBox["012",
     FontColor->RGBColor[1, 0, 0]]], TraditionalForm]],ExpressionUUID->
  "a17027fd-550e-4403-bcc6-e1377ac59885"],
 " is an independent component and has added a rule for it in the second list \
of the FoldedRule. This rule is trivial, as we have not specified the \
numerical value of the component. But consider now the following "
}], "MathCaption",
 CellID->261216211,ExpressionUUID->"b343f231-2ece-447a-8947-70083c1288f7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentValue", "[", 
  RowBox[{"W", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"2", ",", "polar"}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{"1", ",", "polar"}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{"0", ",", "polar"}], "}"}]}], "]"}], "]"}]], "Input",
 CellLabel->"In[16]:=",
 CellID->804283809,ExpressionUUID->"c5766ee7-55b3-4727-b9c7-923175f710fc"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{2, $CellContext`polar}, {1, $CellContext`polar}, {
      0, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"W", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
       2, $CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`W[{2, $CellContext`polar}, {
     1, $CellContext`polar}, {
     0, $CellContext`polar}] -> -$CellContext`W[{0, $CellContext`polar}, {
      1, $CellContext`polar}, {2, $CellContext`polar}], 
   " for tensor ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752197280981*^9, 3.728752285460311*^9, 
  3.7287526088466883`*^9, 3.728752662922867*^9, 3.728753077403554*^9, 
  3.728753151108042*^9, 3.730913199736549*^9},
 CellLabel->"During evaluation of In[16]:=",
 CellID->360939735,ExpressionUUID->"e3343d74-3d9d-4319-836f-84b983d22a23"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"W", 
       StyleBox[GridBox[{
          {
           StyleBox["2",
            FontColor->RGBColor[1., 0., 0.]], 
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], 
           StyleBox["0",
            FontColor->RGBColor[1., 0., 0.]]},
          {" ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`W[{2, $CellContext`polar}, {1, $CellContext`polar}, {
    0, $CellContext`polar}],
   Editable->False], "\[Rule]", 
  RowBox[{"-", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
     2, $CellContext`polar}],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728752197335946*^9, 3.728752285512424*^9, 
  3.72875260889564*^9, 3.72875266297261*^9, 3.728753077438566*^9, 
  3.728753151173711*^9, 3.730913199764284*^9},
 CellLabel->"Out[16]=",
 CellID->1820307017,ExpressionUUID->"22ee4409-db33-4be2-ace0-0273715da9d3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TensorValues", "[", "W", "]"}]], "Input",
 CellLabel->"In[17]:=",
 CellID->1365141535,ExpressionUUID->"ea88911a-18dd-4196-b4de-d9c511497a73"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"W", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`W[{2, $CellContext`polar}, {1, $CellContext`polar}, {
       0, $CellContext`polar}],
      Editable->False], "\[Rule]", 
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"W", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
        2, $CellContext`polar}],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"W", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
       2, $CellContext`polar}],
      Editable->False], "\[Rule]", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"W", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
       2, $CellContext`polar}],
      Editable->False]}], "}"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728752198009251*^9, 3.7287522861454773`*^9, 
  3.728752609057831*^9, 3.728752663789487*^9, 3.7287530775534697`*^9, 
  3.728753151358754*^9, 3.73091320050698*^9},
 CellLabel->"Out[17]=",
 CellID->1479780630,ExpressionUUID->"022d0826-a20e-4032-8e86-2275b9bcfe41"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["xCoba`", "MR"],
 "  sees that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SuperscriptBox["W", 
     StyleBox["210",
      FontColor->RGBColor[1, 0, 0]]], " "}], TraditionalForm]],ExpressionUUID->
  "719fde74-786d-4f2d-8180-6daa0d8740c9"],
 "is not an independent component and generates for it a rule in the first \
list of the FoldedRule. We can easily generate the full set of component \
values:"
}], "MathCaption",
 CellID->34932609,ExpressionUUID->"f022ee5d-1202-4045-aab3-166dbab1ea72"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"W", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"a", ",", "polar"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"b", ",", "polar"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"c", ",", "polar"}], "}"}]}], "]"}], "//", "ComponentArray"}], "//",
   "Flatten"}]], "Input",
 CellLabel->"In[18]:=",
 CellID->849323832,ExpressionUUID->"5164f5e8-9871-4b42-8278-f16b74efbd02"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[{0, $CellContext`polar}, {0, $CellContext`polar}, {
     0, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[{0, $CellContext`polar}, {0, $CellContext`polar}, {
     1, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[{0, $CellContext`polar}, {0, $CellContext`polar}, {
     2, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
     0, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
     1, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
     2, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[{0, $CellContext`polar}, {2, $CellContext`polar}, {
     0, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[{0, $CellContext`polar}, {2, $CellContext`polar}, {
     1, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[{0, $CellContext`polar}, {2, $CellContext`polar}, {
     2, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[{1, $CellContext`polar}, {0, $CellContext`polar}, {
     0, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[{1, $CellContext`polar}, {0, $CellContext`polar}, {
     1, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[{1, $CellContext`polar}, {0, $CellContext`polar}, {
     2, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[{1, $CellContext`polar}, {1, $CellContext`polar}, {
     0, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[{1, $CellContext`polar}, {1, $CellContext`polar}, {
     1, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[{1, $CellContext`polar}, {1, $CellContext`polar}, {
     2, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[{1, $CellContext`polar}, {2, $CellContext`polar}, {
     0, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[{1, $CellContext`polar}, {2, $CellContext`polar}, {
     1, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[{1, $CellContext`polar}, {2, $CellContext`polar}, {
     2, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[{2, $CellContext`polar}, {0, $CellContext`polar}, {
     0, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[{2, $CellContext`polar}, {0, $CellContext`polar}, {
     1, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[{2, $CellContext`polar}, {0, $CellContext`polar}, {
     2, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[{2, $CellContext`polar}, {1, $CellContext`polar}, {
     0, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[{2, $CellContext`polar}, {1, $CellContext`polar}, {
     1, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[{2, $CellContext`polar}, {1, $CellContext`polar}, {
     2, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[{2, $CellContext`polar}, {2, $CellContext`polar}, {
     0, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[{2, $CellContext`polar}, {2, $CellContext`polar}, {
     1, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[{2, $CellContext`polar}, {2, $CellContext`polar}, {
     2, $CellContext`polar}],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728752201083424*^9, 3.728752287663561*^9, 
  3.7287526092699842`*^9, 3.72875266503242*^9, 3.728753077778809*^9, 
  3.728753152050249*^9, 3.730913201753416*^9},
 CellLabel->"Out[18]=",
 CellID->1791544785,ExpressionUUID->"b09810c7-0af4-423d-b8aa-7fdc143ad890"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentValue", "/@", " ", "%"}]], "Input",
 CellLabel->"In[19]:=",
 CellID->598199002,ExpressionUUID->"967611ed-f59a-47a0-b013-4ffd36a01d9e"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{0, $CellContext`polar}, {0, $CellContext`polar}, {
      0, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`W[{0, $CellContext`polar}, {
     0, $CellContext`polar}, {0, $CellContext`polar}] -> 0, 
   " for tensor ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752201780789*^9, 3.728752288392632*^9, 
  3.728752609442985*^9, 3.728752665637739*^9, 3.728753077932866*^9, 
  3.728753152239118*^9, 3.730913202438281*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1727926895,ExpressionUUID->"5a36791a-8da6-4927-bc09-70ac4f86d2b1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{0, $CellContext`polar}, {0, $CellContext`polar}, {
      1, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`W[{0, $CellContext`polar}, {
     0, $CellContext`polar}, {1, $CellContext`polar}] -> 0, 
   " for tensor ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752201780789*^9, 3.728752288392632*^9, 
  3.728752609442985*^9, 3.728752665637739*^9, 3.728753077932866*^9, 
  3.728753152239118*^9, 3.730913202486411*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->845965884,ExpressionUUID->"82359f14-8d2a-4bf2-a481-329480350d57"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{0, $CellContext`polar}, {0, $CellContext`polar}, {
      2, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`W[{0, $CellContext`polar}, {
     0, $CellContext`polar}, {2, $CellContext`polar}] -> 0, 
   " for tensor ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752201780789*^9, 3.728752288392632*^9, 
  3.728752609442985*^9, 3.728752665637739*^9, 3.728753077932866*^9, 
  3.728753152239118*^9, 3.730913202564357*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1806869248,ExpressionUUID->"a700f923-1f9e-4eaa-a7c1-2e77d65ff4ad"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
      0, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`W[{0, $CellContext`polar}, {
     1, $CellContext`polar}, {0, $CellContext`polar}] -> 0, 
   " for tensor ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752201780789*^9, 3.728752288392632*^9, 
  3.728752609442985*^9, 3.728752665637739*^9, 3.728753077932866*^9, 
  3.728753152239118*^9, 3.7309132026238194`*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->485198909,ExpressionUUID->"ae2f955b-c270-4343-8a8b-bb72105933a1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
      1, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`W[{0, $CellContext`polar}, {
     1, $CellContext`polar}, {1, $CellContext`polar}] -> 0, 
   " for tensor ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752201780789*^9, 3.728752288392632*^9, 
  3.728752609442985*^9, 3.728752665637739*^9, 3.728753077932866*^9, 
  3.728753152239118*^9, 3.730913202697097*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->742655904,ExpressionUUID->"f3bc0856-40e8-4c8a-94f2-84b4cfbf0f3f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{0, $CellContext`polar}, {2, $CellContext`polar}, {
      0, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`W[{0, $CellContext`polar}, {
     2, $CellContext`polar}, {0, $CellContext`polar}] -> 0, 
   " for tensor ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752201780789*^9, 3.728752288392632*^9, 
  3.728752609442985*^9, 3.728752665637739*^9, 3.728753077932866*^9, 
  3.728753152239118*^9, 3.730913202758346*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->642822164,ExpressionUUID->"3de0d4dd-ec52-436d-8834-7140194cfc5d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{0, $CellContext`polar}, {2, $CellContext`polar}, {
      1, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"W", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
       2, $CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`W[{0, $CellContext`polar}, {
     2, $CellContext`polar}, {
     1, $CellContext`polar}] -> -$CellContext`W[{0, $CellContext`polar}, {
      1, $CellContext`polar}, {2, $CellContext`polar}], 
   " for tensor ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752201780789*^9, 3.728752288392632*^9, 
  3.728752609442985*^9, 3.728752665637739*^9, 3.728753077932866*^9, 
  3.728753152239118*^9, 3.730913202830428*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->2069365938,ExpressionUUID->"0b8b82ee-695a-43c2-b8e7-4ae284d9c6d0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{0, $CellContext`polar}, {2, $CellContext`polar}, {
      2, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`W[{0, $CellContext`polar}, {
     2, $CellContext`polar}, {2, $CellContext`polar}] -> 0, 
   " for tensor ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752201780789*^9, 3.728752288392632*^9, 
  3.728752609442985*^9, 3.728752665637739*^9, 3.728753077932866*^9, 
  3.728753152239118*^9, 3.730913202891016*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->691311822,ExpressionUUID->"302d5d6e-c83b-48a0-b10f-8635747d850d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{1, $CellContext`polar}, {0, $CellContext`polar}, {
      0, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`W[{1, $CellContext`polar}, {
     0, $CellContext`polar}, {0, $CellContext`polar}] -> 0, 
   " for tensor ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752201780789*^9, 3.728752288392632*^9, 
  3.728752609442985*^9, 3.728752665637739*^9, 3.728753077932866*^9, 
  3.728753152239118*^9, 3.7309132030923986`*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->991262084,ExpressionUUID->"60069d13-db73-4ac4-b9fe-c270d8dcb531"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{1, $CellContext`polar}, {0, $CellContext`polar}, {
      1, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`W[{1, $CellContext`polar}, {
     0, $CellContext`polar}, {1, $CellContext`polar}] -> 0, 
   " for tensor ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752201780789*^9, 3.728752288392632*^9, 
  3.728752609442985*^9, 3.728752665637739*^9, 3.728753077932866*^9, 
  3.728753152239118*^9, 3.730913203180779*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->2111196009,ExpressionUUID->"b0cd6779-a498-4202-b7f1-c652950753c9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{1, $CellContext`polar}, {0, $CellContext`polar}, {
      2, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"W", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
       2, $CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`W[{1, $CellContext`polar}, {
     0, $CellContext`polar}, {
     2, $CellContext`polar}] -> -$CellContext`W[{0, $CellContext`polar}, {
      1, $CellContext`polar}, {2, $CellContext`polar}], 
   " for tensor ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752201780789*^9, 3.728752288392632*^9, 
  3.728752609442985*^9, 3.728752665637739*^9, 3.728753077932866*^9, 
  3.728753152239118*^9, 3.7309132032646713`*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->2146507529,ExpressionUUID->"1dc5c7c9-2293-486e-b105-231d6288769e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{1, $CellContext`polar}, {1, $CellContext`polar}, {
      0, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`W[{1, $CellContext`polar}, {
     1, $CellContext`polar}, {0, $CellContext`polar}] -> 0, 
   " for tensor ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752201780789*^9, 3.728752288392632*^9, 
  3.728752609442985*^9, 3.728752665637739*^9, 3.728753077932866*^9, 
  3.728753152239118*^9, 3.730913203348971*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->774253950,ExpressionUUID->"fbcc3b2d-f5e4-4946-b4c7-feb44c1c5b7b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{1, $CellContext`polar}, {1, $CellContext`polar}, {
      1, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`W[{1, $CellContext`polar}, {
     1, $CellContext`polar}, {1, $CellContext`polar}] -> 0, 
   " for tensor ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752201780789*^9, 3.728752288392632*^9, 
  3.728752609442985*^9, 3.728752665637739*^9, 3.728753077932866*^9, 
  3.728753152239118*^9, 3.730913203430984*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1282720477,ExpressionUUID->"61959597-ca53-464f-9999-f97754775411"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{1, $CellContext`polar}, {1, $CellContext`polar}, {
      2, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`W[{1, $CellContext`polar}, {
     1, $CellContext`polar}, {2, $CellContext`polar}] -> 0, 
   " for tensor ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752201780789*^9, 3.728752288392632*^9, 
  3.728752609442985*^9, 3.728752665637739*^9, 3.728753077932866*^9, 
  3.728753152239118*^9, 3.730913203513465*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->94724706,ExpressionUUID->"c201c77a-67f4-43b8-bac0-855d90537900"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{1, $CellContext`polar}, {2, $CellContext`polar}, {
      0, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
      2, $CellContext`polar}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`W[{1, $CellContext`polar}, {
     2, $CellContext`polar}, {0, $CellContext`polar}] -> $CellContext`W[{
     0, $CellContext`polar}, {1, $CellContext`polar}, {
     2, $CellContext`polar}], " for tensor ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752201780789*^9, 3.728752288392632*^9, 
  3.728752609442985*^9, 3.728752665637739*^9, 3.728753077932866*^9, 
  3.728753152239118*^9, 3.730913203592211*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->752751315,ExpressionUUID->"08468eee-9868-4c3c-9c3e-c26e62ca0616"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{1, $CellContext`polar}, {2, $CellContext`polar}, {
      1, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`W[{1, $CellContext`polar}, {
     2, $CellContext`polar}, {1, $CellContext`polar}] -> 0, 
   " for tensor ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752201780789*^9, 3.728752288392632*^9, 
  3.728752609442985*^9, 3.728752665637739*^9, 3.728753077932866*^9, 
  3.728753152239118*^9, 3.730913203663158*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->665186728,ExpressionUUID->"5eb96227-18c9-4b2f-94fa-76cd011daa89"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{1, $CellContext`polar}, {2, $CellContext`polar}, {
      2, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`W[{1, $CellContext`polar}, {
     2, $CellContext`polar}, {2, $CellContext`polar}] -> 0, 
   " for tensor ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752201780789*^9, 3.728752288392632*^9, 
  3.728752609442985*^9, 3.728752665637739*^9, 3.728753077932866*^9, 
  3.728753152239118*^9, 3.730913203724612*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->2140954861,ExpressionUUID->"c3c92aff-a9b5-4269-b99e-3f529ff81f81"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{2, $CellContext`polar}, {0, $CellContext`polar}, {
      0, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`W[{2, $CellContext`polar}, {
     0, $CellContext`polar}, {0, $CellContext`polar}] -> 0, 
   " for tensor ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752201780789*^9, 3.728752288392632*^9, 
  3.728752609442985*^9, 3.728752665637739*^9, 3.728753077932866*^9, 
  3.728753152239118*^9, 3.730913203813415*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1939843342,ExpressionUUID->"daef8708-9b35-46c0-8f39-0f4d91940bde"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{2, $CellContext`polar}, {0, $CellContext`polar}, {
      1, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
      2, $CellContext`polar}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`W[{2, $CellContext`polar}, {
     0, $CellContext`polar}, {1, $CellContext`polar}] -> $CellContext`W[{
     0, $CellContext`polar}, {1, $CellContext`polar}, {
     2, $CellContext`polar}], " for tensor ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752201780789*^9, 3.728752288392632*^9, 
  3.728752609442985*^9, 3.728752665637739*^9, 3.728753077932866*^9, 
  3.728753152239118*^9, 3.73091320387781*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1997803887,ExpressionUUID->"a92e1a87-2e0a-4871-a78e-2ff6661c9c3e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{2, $CellContext`polar}, {0, $CellContext`polar}, {
      2, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`W[{2, $CellContext`polar}, {
     0, $CellContext`polar}, {2, $CellContext`polar}] -> 0, 
   " for tensor ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752201780789*^9, 3.728752288392632*^9, 
  3.728752609442985*^9, 3.728752665637739*^9, 3.728753077932866*^9, 
  3.728753152239118*^9, 3.730913203934816*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1183107964,ExpressionUUID->"595c4ab2-0641-420a-8ef2-f02aad16eb7b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{2, $CellContext`polar}, {1, $CellContext`polar}, {
      1, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`W[{2, $CellContext`polar}, {
     1, $CellContext`polar}, {1, $CellContext`polar}] -> 0, 
   " for tensor ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752201780789*^9, 3.728752288392632*^9, 
  3.728752609442985*^9, 3.728752665637739*^9, 3.728753077932866*^9, 
  3.728753152239118*^9, 3.7309132040126247`*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->944463375,ExpressionUUID->"b887ed34-231b-45cf-ad3b-b2cde09d20b1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{2, $CellContext`polar}, {1, $CellContext`polar}, {
      2, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`W[{2, $CellContext`polar}, {
     1, $CellContext`polar}, {2, $CellContext`polar}] -> 0, 
   " for tensor ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752201780789*^9, 3.728752288392632*^9, 
  3.728752609442985*^9, 3.728752665637739*^9, 3.728753077932866*^9, 
  3.728753152239118*^9, 3.7309132040748043`*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1544308648,ExpressionUUID->"8354e3a4-0622-42b9-b54a-0446d967c360"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{2, $CellContext`polar}, {2, $CellContext`polar}, {
      0, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`W[{2, $CellContext`polar}, {
     2, $CellContext`polar}, {0, $CellContext`polar}] -> 0, 
   " for tensor ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752201780789*^9, 3.728752288392632*^9, 
  3.728752609442985*^9, 3.728752665637739*^9, 3.728753077932866*^9, 
  3.728753152239118*^9, 3.7309132041408777`*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->708089294,ExpressionUUID->"84fc46d4-3e53-4942-881c-b47a7bcab337"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{2, $CellContext`polar}, {2, $CellContext`polar}, {
      1, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`W[{2, $CellContext`polar}, {
     2, $CellContext`polar}, {1, $CellContext`polar}] -> 0, 
   " for tensor ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752201780789*^9, 3.728752288392632*^9, 
  3.728752609442985*^9, 3.728752665637739*^9, 3.728753077932866*^9, 
  3.728753152239118*^9, 3.730913204214439*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->291464229,ExpressionUUID->"d811d90b-231d-4979-98c2-0688b3223dc5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{2, $CellContext`polar}, {2, $CellContext`polar}, {
      2, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`W[{2, $CellContext`polar}, {
     2, $CellContext`polar}, {2, $CellContext`polar}] -> 0, 
   " for tensor ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752201780789*^9, 3.728752288392632*^9, 
  3.728752609442985*^9, 3.728752665637739*^9, 3.728753077932866*^9, 
  3.728753152239118*^9, 3.7309132042914867`*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->229097996,ExpressionUUID->"aa284a8a-1a63-4df4-88bf-2016a5d6b2e0"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{0, $CellContext`polar}, {0, $CellContext`polar}, {
      0, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{0, $CellContext`polar}, {0, $CellContext`polar}, {
      1, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{0, $CellContext`polar}, {0, $CellContext`polar}, {
      2, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
      0, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
      1, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
      2, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
      2, $CellContext`polar}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{0, $CellContext`polar}, {2, $CellContext`polar}, {
      0, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{0, $CellContext`polar}, {2, $CellContext`polar}, {
      1, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"W", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
       2, $CellContext`polar}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{0, $CellContext`polar}, {2, $CellContext`polar}, {
      2, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{1, $CellContext`polar}, {0, $CellContext`polar}, {
      0, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{1, $CellContext`polar}, {0, $CellContext`polar}, {
      1, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{1, $CellContext`polar}, {0, $CellContext`polar}, {
      2, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"W", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
       2, $CellContext`polar}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{1, $CellContext`polar}, {1, $CellContext`polar}, {
      0, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{1, $CellContext`polar}, {1, $CellContext`polar}, {
      1, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{1, $CellContext`polar}, {1, $CellContext`polar}, {
      2, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{1, $CellContext`polar}, {2, $CellContext`polar}, {
      0, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
      2, $CellContext`polar}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{1, $CellContext`polar}, {2, $CellContext`polar}, {
      1, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{1, $CellContext`polar}, {2, $CellContext`polar}, {
      2, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{2, $CellContext`polar}, {0, $CellContext`polar}, {
      0, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{2, $CellContext`polar}, {0, $CellContext`polar}, {
      1, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
      2, $CellContext`polar}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{2, $CellContext`polar}, {0, $CellContext`polar}, {
      2, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{2, $CellContext`polar}, {1, $CellContext`polar}, {
      0, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"W", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
       2, $CellContext`polar}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{2, $CellContext`polar}, {1, $CellContext`polar}, {
      1, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{2, $CellContext`polar}, {1, $CellContext`polar}, {
      2, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{2, $CellContext`polar}, {2, $CellContext`polar}, {
      0, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{2, $CellContext`polar}, {2, $CellContext`polar}, {
      1, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{2, $CellContext`polar}, {2, $CellContext`polar}, {
      2, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728752204874895*^9, 3.728752290909425*^9, 
  3.7287526116136217`*^9, 3.728752668347601*^9, 3.728753079807949*^9, 
  3.728753156034359*^9, 3.730913204489889*^9},
 CellLabel->"Out[19]=",
 CellID->175699453,ExpressionUUID->"1e4eecfb-f1cc-4aeb-838e-95d5564163f9"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "As we can see by reading the messages, the rules for ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["W", 
    StyleBox["012",
     FontColor->RGBColor[1, 0, 0]]], TraditionalForm]],ExpressionUUID->
  "49194c0a-92b2-4d04-9784-9882c14b9dd8"],
 " and ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["W", 
    StyleBox["210",
     FontColor->RGBColor[1, 0, 0]]], TraditionalForm]],ExpressionUUID->
  "89d9585b-9480-4d0d-a8f7-3a8b54ff29c3"],
 " have not been recomputed. Now:"
}], "MathCaption",
 CellChangeTimes->{3.728751963728772*^9},
 CellID->1868763863,ExpressionUUID->"ec014675-e171-40eb-83d3-50d600ca97c4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TensorValues", "[", "W", "]"}]], "Input",
 CellLabel->"In[20]:=",
 CellID->1365908066,ExpressionUUID->"f7e9ac07-ba86-4722-8d7d-3da38f1fe8f9"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"W", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`W[{2, $CellContext`polar}, {1, $CellContext`polar}, {
        0, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"W", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
         2, $CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"W", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`W[{0, $CellContext`polar}, {0, $CellContext`polar}, {
        0, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"W", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`W[{0, $CellContext`polar}, {0, $CellContext`polar}, {
        1, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"W", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`W[{0, $CellContext`polar}, {0, $CellContext`polar}, {
        2, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"W", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
        0, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"W", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
        1, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"W", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`W[{0, $CellContext`polar}, {2, $CellContext`polar}, {
        0, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"W", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`W[{0, $CellContext`polar}, {2, $CellContext`polar}, {
        1, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"W", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
         2, $CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"W", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`W[{0, $CellContext`polar}, {2, $CellContext`polar}, {
        2, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"W", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`W[{1, $CellContext`polar}, {0, $CellContext`polar}, {
        0, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"W", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`W[{1, $CellContext`polar}, {0, $CellContext`polar}, {
        1, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"W", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`W[{1, $CellContext`polar}, {0, $CellContext`polar}, {
        2, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"W", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
         2, $CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"W", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`W[{1, $CellContext`polar}, {1, $CellContext`polar}, {
        0, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"W", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`W[{1, $CellContext`polar}, {1, $CellContext`polar}, {
        1, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"W", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`W[{1, $CellContext`polar}, {1, $CellContext`polar}, {
        2, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"W", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`W[{1, $CellContext`polar}, {2, $CellContext`polar}, {
        0, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"W", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
        2, $CellContext`polar}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"W", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`W[{1, $CellContext`polar}, {2, $CellContext`polar}, {
        1, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"W", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`W[{1, $CellContext`polar}, {2, $CellContext`polar}, {
        2, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"W", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`W[{2, $CellContext`polar}, {0, $CellContext`polar}, {
        0, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"W", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`W[{2, $CellContext`polar}, {0, $CellContext`polar}, {
        1, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"W", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
        2, $CellContext`polar}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"W", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`W[{2, $CellContext`polar}, {0, $CellContext`polar}, {
        2, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"W", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`W[{2, $CellContext`polar}, {1, $CellContext`polar}, {
        1, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"W", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`W[{2, $CellContext`polar}, {1, $CellContext`polar}, {
        2, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"W", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`W[{2, $CellContext`polar}, {2, $CellContext`polar}, {
        0, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"W", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`W[{2, $CellContext`polar}, {2, $CellContext`polar}, {
        1, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"W", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`W[{2, $CellContext`polar}, {2, $CellContext`polar}, {
        2, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"W", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
       2, $CellContext`polar}],
      Editable->False], "\[Rule]", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"W", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`W[{0, $CellContext`polar}, {1, $CellContext`polar}, {
       2, $CellContext`polar}],
      Editable->False]}], "}"}]}], "]"}]], "Output",
 CellChangeTimes->{3.7287522073727694`*^9, 3.72875229192168*^9, 
  3.728752611823175*^9, 3.728752668641378*^9, 3.728753079975749*^9, 
  3.728753156287462*^9, 3.7309132053757772`*^9},
 CellLabel->"Out[20]=",
 CellID->1704148742,ExpressionUUID->"bf6c0915-6afc-4d9f-8dc1-3b4b13553e42"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "There is only one independent rule. We have one ",
 StyleBox["ValID", "MR"],
 " for ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["W", "abc"], TraditionalForm]],ExpressionUUID->
  "c0a8fb6d-265c-49ac-af86-331134cc7840"],
 ", which consists of only one basis configuration:"
}], "MathCaption",
 CellChangeTimes->{3.728751972361135*^9},
 CellID->1736347759,ExpressionUUID->"68fc9241-7d70-428a-aef9-c820be3c4055"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TensorValIDs", "[", "W", "]"}]], "Input",
 CellLabel->"In[21]:=",
 CellID->641506430,ExpressionUUID->"c3eb6f7f-5d3a-4cc4-beba-23d261671224"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"ValID", "[", 
   RowBox[{"W", ",", 
    RowBox[{"{", 
     RowBox[{"{", 
      RowBox[{"polar", ",", "polar", ",", "polar"}], "}"}], "}"}]}], "]"}], 
  "}"}]], "Output",
 CellChangeTimes->{3.7287522089491463`*^9, 3.728752293130131*^9, 
  3.728752611976907*^9, 3.7287526708737*^9, 3.728753080110289*^9, 
  3.728753156498126*^9, 3.7309132067324038`*^9},
 CellLabel->"Out[21]=",
 CellID->969769461,ExpressionUUID->"3ff14788-a495-4497-8e75-da68b6115b34"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["ValIDs", "MR"],
 " are timestamped:"
}], "MathCaption",
 CellID->1893691678,ExpressionUUID->"a8d0318c-7004-4a95-a800-4516d730231e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DateOfValID", "/@", " ", "%"}]], "Input",
 CellLabel->"In[22]:=",
 CellID->1568383606,ExpressionUUID->"806a1216-f75b-441b-bcff-38d7775ad7e5"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"{", 
   RowBox[{
   "2018", ",", "3", ",", "24", ",", "15", ",", "46", ",", 
    "44.218906`8.398182970850623"}], "}"}], "}"}]], "Output",
 CellChangeTimes->{3.728752209961405*^9, 3.728752294263762*^9, 
  3.728752612127315*^9, 3.728752671923506*^9, 3.72875308024487*^9, 
  3.7287531566800222`*^9, 3.730913211010572*^9},
 CellLabel->"Out[22]=",
 CellID->72603651,ExpressionUUID->"fc77f001-bc28-4c09-97c8-e3220f168385"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Let us now consider an example with several bases. Recall that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["T", "ab"], " ", "is", " ", 
    StyleBox["a",
     FontSlant->"Plain"], " ", "symmetric", " ", "tensor"}], 
   TraditionalForm]],ExpressionUUID->"53c8a91e-276d-4e2b-b752-63a2c1568400"],
 ":"
}], "MathCaption",
 CellChangeTimes->{3.7287519789207153`*^9},
 CellID->1059231524,ExpressionUUID->"0aa51283-12be-4090-bcea-067baa0999e7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentValue", "[", 
  RowBox[{"T", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"2", ",", 
      RowBox[{"-", "polar"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{"1", ",", 
      RowBox[{"-", "cartesian"}]}], "}"}]}], "]"}], "]"}]], "Input",
 CellLabel->"In[23]:=",
 CellID->2012273885,ExpressionUUID->"b82cddca-96b9-4d3c-a193-2a4bcb13009e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, -$CellContext`polar}, {1, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 1., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`cartesian}, {2, -$CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`T[{2, -$CellContext`polar}, {
     1, -$CellContext`cartesian}] -> -$CellContext`T[{
      1, -$CellContext`cartesian}, {2, -$CellContext`polar}], 
   " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752234875812*^9, 3.728752296093755*^9, 
  3.7287526122924223`*^9, 3.728752673073785*^9, 3.728753080377666*^9, 
  3.72875315685863*^9, 3.730913213210129*^9},
 CellLabel->"During evaluation of In[23]:=",
 CellID->375628720,ExpressionUUID->"9698d864-b653-4199-9f9d-efde0416f750"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {" ", " "},
          {
           StyleBox["2",
            FontColor->RGBColor[1., 0., 0.]], 
           StyleBox["1",
            FontColor->RGBColor[0., 1., 0.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[{2, -$CellContext`polar}, {1, -$CellContext`cartesian}],
   Editable->False], "\[Rule]", 
  RowBox[{"-", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 1., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{1, -$CellContext`cartesian}, {2, -$CellContext`polar}],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{{3.728752212844256*^9, 3.728752234977742*^9}, 
   3.7287522961469927`*^9, 3.728752612325783*^9, 3.7287526731253157`*^9, 
   3.7287530804285393`*^9, 3.7287531569246473`*^9, 3.730913213259541*^9},
 CellLabel->"Out[23]=",
 CellID->1737458507,ExpressionUUID->"d833bb2c-1d25-412e-9d04-18e25e6b5dba"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{
     RowBox[{"-", 
      RowBox[{"{", 
       RowBox[{"a", ",", "polar"}], "}"}]}], ",", 
     RowBox[{"-", 
      RowBox[{"{", 
       RowBox[{"b", ",", "cartesian"}], "}"}]}]}], "]"}], " ", "//", 
   "ComponentArray"}], "//", "Flatten"}]], "Input",
 CellLabel->"In[24]:=",
 CellID->472610126,ExpressionUUID->"0cc380eb-975f-40bb-87aa-2126b3ed4689"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[0., 1., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{0, -$CellContext`polar}, {0, -$CellContext`cartesian}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[0., 1., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{0, -$CellContext`polar}, {1, -$CellContext`cartesian}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[0., 1., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{0, -$CellContext`polar}, {2, -$CellContext`cartesian}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[0., 1., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{1, -$CellContext`polar}, {0, -$CellContext`cartesian}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[0., 1., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{1, -$CellContext`polar}, {1, -$CellContext`cartesian}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[0., 1., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{1, -$CellContext`polar}, {2, -$CellContext`cartesian}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[0., 1., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{2, -$CellContext`polar}, {0, -$CellContext`cartesian}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[0., 1., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{2, -$CellContext`polar}, {1, -$CellContext`cartesian}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[0., 1., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{2, -$CellContext`polar}, {2, -$CellContext`cartesian}],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.7287522360944567`*^9, 3.728752297191334*^9, 
  3.72875261252804*^9, 3.7287526740081053`*^9, 3.7287530807840157`*^9, 
  3.7287531576311073`*^9, 3.7309132141254387`*^9},
 CellLabel->"Out[24]=",
 CellID->1916977214,ExpressionUUID->"27cc14bf-57dc-4a02-9703-b2024c6bd0af"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentValue", " ", "/@", " ", "%"}]], "Input",
 CellLabel->"In[25]:=",
 CellID->1348841144,ExpressionUUID->"590e6f78-b634-4f51-9980-55855a609a4a"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`polar}, {0, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[0., 1., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`cartesian}, {0, -$CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`T[{0, -$CellContext`polar}, {
     0, -$CellContext`cartesian}] -> -$CellContext`T[{
      0, -$CellContext`cartesian}, {0, -$CellContext`polar}], 
   " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287522408437023`*^9, 3.728752299227724*^9, 
  3.728752612682852*^9, 3.728752674587267*^9, 3.7287530809989033`*^9, 
  3.728753157880414*^9, 3.730913214924382*^9},
 CellLabel->"During evaluation of In[25]:=",
 CellID->1041221882,ExpressionUUID->"eb307d47-ce33-4881-b439-5b6550346075"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`polar}, {1, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`polar}, {1, -$CellContext`cartesian}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added independent rule ", $CellContext`T[{0, -$CellContext`polar}, {
     1, -$CellContext`cartesian}] -> $CellContext`T[{
     0, -$CellContext`polar}, {1, -$CellContext`cartesian}], 
   " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287522408437023`*^9, 3.728752299227724*^9, 
  3.728752612682852*^9, 3.728752674587267*^9, 3.7287530809989033`*^9, 
  3.728753157880414*^9, 3.73091321497766*^9},
 CellLabel->"During evaluation of In[25]:=",
 CellID->2119415018,ExpressionUUID->"202b303b-18ac-4c41-a272-cf8dcbfede79"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`polar}, {2, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`polar}, {2, -$CellContext`cartesian}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added independent rule ", $CellContext`T[{0, -$CellContext`polar}, {
     2, -$CellContext`cartesian}] -> $CellContext`T[{
     0, -$CellContext`polar}, {2, -$CellContext`cartesian}], 
   " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287522408437023`*^9, 3.728752299227724*^9, 
  3.728752612682852*^9, 3.728752674587267*^9, 3.7287530809989033`*^9, 
  3.728753157880414*^9, 3.7309132150655823`*^9},
 CellLabel->"During evaluation of In[25]:=",
 CellID->696954737,ExpressionUUID->"b12f95cf-c0f1-4ca4-ae37-e5190ec7f03c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`polar}, {0, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[0., 1., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`cartesian}, {1, -$CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`T[{1, -$CellContext`polar}, {
     0, -$CellContext`cartesian}] -> -$CellContext`T[{
      0, -$CellContext`cartesian}, {1, -$CellContext`polar}], 
   " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287522408437023`*^9, 3.728752299227724*^9, 
  3.728752612682852*^9, 3.728752674587267*^9, 3.7287530809989033`*^9, 
  3.728753157880414*^9, 3.73091321514869*^9},
 CellLabel->"During evaluation of In[25]:=",
 CellID->1342454021,ExpressionUUID->"251149aa-04d7-417b-b2cd-0284e0e12a8a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`polar}, {1, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 1., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`cartesian}, {1, -$CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`T[{1, -$CellContext`polar}, {
     1, -$CellContext`cartesian}] -> -$CellContext`T[{
      1, -$CellContext`cartesian}, {1, -$CellContext`polar}], 
   " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287522408437023`*^9, 3.728752299227724*^9, 
  3.728752612682852*^9, 3.728752674587267*^9, 3.7287530809989033`*^9, 
  3.728753157880414*^9, 3.730913215232177*^9},
 CellLabel->"During evaluation of In[25]:=",
 CellID->1161535565,ExpressionUUID->"05ac1580-a4c3-46f6-9dd4-9b3950d22280"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`polar}, {2, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`polar}, {2, -$CellContext`cartesian}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added independent rule ", $CellContext`T[{1, -$CellContext`polar}, {
     2, -$CellContext`cartesian}] -> $CellContext`T[{
     1, -$CellContext`polar}, {2, -$CellContext`cartesian}], 
   " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287522408437023`*^9, 3.728752299227724*^9, 
  3.728752612682852*^9, 3.728752674587267*^9, 3.7287530809989033`*^9, 
  3.728753157880414*^9, 3.730913215415935*^9},
 CellLabel->"During evaluation of In[25]:=",
 CellID->1788456507,ExpressionUUID->"cee429dd-7af5-43f0-9254-cf0a38908f92"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, -$CellContext`polar}, {0, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[0., 1., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`cartesian}, {2, -$CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`T[{2, -$CellContext`polar}, {
     0, -$CellContext`cartesian}] -> -$CellContext`T[{
      0, -$CellContext`cartesian}, {2, -$CellContext`polar}], 
   " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287522408437023`*^9, 3.728752299227724*^9, 
  3.728752612682852*^9, 3.728752674587267*^9, 3.7287530809989033`*^9, 
  3.728753157880414*^9, 3.7309132154783087`*^9},
 CellLabel->"During evaluation of In[25]:=",
 CellID->259906090,ExpressionUUID->"594ea64e-5327-4ff9-9f64-323f5d28312c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, -$CellContext`polar}, {2, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[0., 1., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, -$CellContext`cartesian}, {2, -$CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`T[{2, -$CellContext`polar}, {
     2, -$CellContext`cartesian}] -> -$CellContext`T[{
      2, -$CellContext`cartesian}, {2, -$CellContext`polar}], 
   " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287522408437023`*^9, 3.728752299227724*^9, 
  3.728752612682852*^9, 3.728752674587267*^9, 3.7287530809989033`*^9, 
  3.728753157880414*^9, 3.730913215542655*^9},
 CellLabel->"During evaluation of In[25]:=",
 CellID->259916632,ExpressionUUID->"c215447a-468b-4e89-a3a5-e4f70f444c5f"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`polar}, {0, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[0., 1., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`cartesian}, {0, -$CellContext`polar}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`polar}, {1, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`polar}, {1, -$CellContext`cartesian}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`polar}, {2, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`polar}, {2, -$CellContext`cartesian}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`polar}, {0, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[0., 1., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`cartesian}, {1, -$CellContext`polar}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`polar}, {1, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 1., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`cartesian}, {1, -$CellContext`polar}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`polar}, {2, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`polar}, {2, -$CellContext`cartesian}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, -$CellContext`polar}, {0, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[0., 1., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`cartesian}, {2, -$CellContext`polar}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, -$CellContext`polar}, {1, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 1., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`cartesian}, {2, -$CellContext`polar}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, -$CellContext`polar}, {2, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[0., 1., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, -$CellContext`cartesian}, {2, -$CellContext`polar}],
      Editable->False]}]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728752241628936*^9, 3.7287523000938053`*^9, 
  3.728752613359529*^9, 3.728752675430757*^9, 3.7287530817982073`*^9, 
  3.728753158729156*^9, 3.7309132156079273`*^9},
 CellLabel->"Out[25]=",
 CellID->437096026,ExpressionUUID->"e9b63519-d193-4960-8222-b2e43cdda579"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TensorValIDs", "[", "T", "]"}]], "Input",
 CellLabel->"In[26]:=",
 CellID->106298051,ExpressionUUID->"9f45667a-aa14-4ecc-a11b-940827b4d299"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"ValID", "[", 
   RowBox[{"T", ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "cartesian"}], ",", 
        RowBox[{"-", "polar"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "polar"}], ",", 
        RowBox[{"-", "cartesian"}]}], "}"}]}], "}"}]}], "]"}], 
  "}"}]], "Output",
 CellChangeTimes->{3.728752302117049*^9, 3.7287526135117292`*^9, 
  3.7287526762081137`*^9, 3.7287530819589157`*^9, 3.7287531589250307`*^9, 
  3.730913216366838*^9},
 CellLabel->"Out[26]=",
 CellID->1193783197,ExpressionUUID->"dfa6b67c-6c0e-44fe-a655-22fdc91a3299"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "We still do not have all the rules, because our ValID mixes the cases ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["T", 
     StyleBox[
      RowBox[{
       StyleBox["a",
        FontColor->RGBColor[1, 0, 0]], 
       StyleBox["b",
        FontColor->RGBColor[0, 1, 0]]}],
      FontSlant->"Italic"]], " ", "and", " ", 
    SubscriptBox["T", 
     StyleBox[
      RowBox[{
       StyleBox["b",
        FontColor->RGBColor[0, 1, 0]], 
       StyleBox["a",
        FontColor->RGBColor[1, 0, 0]]}],
      FontSlant->"Italic"]]}], TraditionalForm]],ExpressionUUID->
  "23036cd4-a311-4d9f-a19e-88021a11f02d"],
 ":"
}], "MathCaption",
 CellID->1100519829,ExpressionUUID->"972e561a-9898-4142-8e3a-2e3dee0d72c3"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{
     RowBox[{"-", 
      RowBox[{"{", 
       RowBox[{"a", ",", "cartesian"}], "}"}]}], ",", 
     RowBox[{"-", 
      RowBox[{"{", 
       RowBox[{"b", ",", "polar"}], "}"}]}]}], "]"}], " ", "//", 
   "ComponentArray"}], "//", "Flatten"}]], "Input",
 CellLabel->"In[27]:=",
 CellID->160114392,ExpressionUUID->"4d2d8a56-2935-4ba0-934d-8d5eae44db6e"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["0",
             FontColor->RGBColor[0., 1., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{0, -$CellContext`cartesian}, {0, -$CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["0",
             FontColor->RGBColor[0., 1., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{0, -$CellContext`cartesian}, {1, -$CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["0",
             FontColor->RGBColor[0., 1., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{0, -$CellContext`cartesian}, {2, -$CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 1., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{1, -$CellContext`cartesian}, {0, -$CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 1., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{1, -$CellContext`cartesian}, {1, -$CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 1., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{1, -$CellContext`cartesian}, {2, -$CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 1., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{2, -$CellContext`cartesian}, {0, -$CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 1., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{2, -$CellContext`cartesian}, {1, -$CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 1., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{2, -$CellContext`cartesian}, {2, -$CellContext`polar}],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728752304668634*^9, 3.728752613690351*^9, 
  3.728752677459176*^9, 3.728753082106586*^9, 3.728753159125602*^9, 
  3.730913217440287*^9},
 CellLabel->"Out[27]=",
 CellID->150839986,ExpressionUUID->"c4d9b66a-1be9-4de4-af50-295dc646c78c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentValue", " ", "/@", " ", "%"}]], "Input",
 CellLabel->"In[28]:=",
 CellID->2090721084,ExpressionUUID->"eae75449-16db-48db-ac0b-ec783ac7334d"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`cartesian}, {0, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`cartesian}, {0, -$CellContext`polar}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added independent rule ", $CellContext`T[{0, -$CellContext`cartesian}, {
     0, -$CellContext`polar}] -> $CellContext`T[{
     0, -$CellContext`cartesian}, {0, -$CellContext`polar}], 
   " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752305719392*^9, 3.728752613908391*^9, 
  3.728752678036072*^9, 3.728753082231434*^9, 3.728753159318799*^9, 
  3.730913218040577*^9},
 CellLabel->"During evaluation of In[28]:=",
 CellID->1358305486,ExpressionUUID->"520d5762-bfc4-4d48-a0d3-3f326e336f43"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`cartesian}, {1, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`cartesian}, {1, -$CellContext`polar}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added independent rule ", $CellContext`T[{0, -$CellContext`cartesian}, {
     1, -$CellContext`polar}] -> $CellContext`T[{
     0, -$CellContext`cartesian}, {1, -$CellContext`polar}], 
   " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752305719392*^9, 3.728752613908391*^9, 
  3.728752678036072*^9, 3.728753082231434*^9, 3.728753159318799*^9, 
  3.730913218094417*^9},
 CellLabel->"During evaluation of In[28]:=",
 CellID->1289135592,ExpressionUUID->"fc9a5c20-dcf5-4e69-b904-7fccbb198b48"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`cartesian}, {2, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`cartesian}, {2, -$CellContext`polar}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added independent rule ", $CellContext`T[{0, -$CellContext`cartesian}, {
     2, -$CellContext`polar}] -> $CellContext`T[{
     0, -$CellContext`cartesian}, {2, -$CellContext`polar}], 
   " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752305719392*^9, 3.728752613908391*^9, 
  3.728752678036072*^9, 3.728753082231434*^9, 3.728753159318799*^9, 
  3.730913218182303*^9},
 CellLabel->"During evaluation of In[28]:=",
 CellID->539380161,ExpressionUUID->"da775947-2fbc-4c58-b0b5-285b2fd81061"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`cartesian}, {0, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 1., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`polar}, {1, -$CellContext`cartesian}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`T[{1, -$CellContext`cartesian}, {
     0, -$CellContext`polar}] -> -$CellContext`T[{0, -$CellContext`polar}, {
      1, -$CellContext`cartesian}], " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752305719392*^9, 3.728752613908391*^9, 
  3.728752678036072*^9, 3.728753082231434*^9, 3.728753159318799*^9, 
  3.730913218261077*^9},
 CellLabel->"During evaluation of In[28]:=",
 CellID->833891940,ExpressionUUID->"47e8c304-e785-46a3-b124-322bd958ace3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`cartesian}, {1, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`cartesian}, {1, -$CellContext`polar}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added independent rule ", $CellContext`T[{1, -$CellContext`cartesian}, {
     1, -$CellContext`polar}] -> $CellContext`T[{
     1, -$CellContext`cartesian}, {1, -$CellContext`polar}], 
   " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752305719392*^9, 3.728752613908391*^9, 
  3.728752678036072*^9, 3.728753082231434*^9, 3.728753159318799*^9, 
  3.730913218326755*^9},
 CellLabel->"During evaluation of In[28]:=",
 CellID->743555258,ExpressionUUID->"b3002e9f-c309-4990-a753-8d3ff085ba40"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`cartesian}, {2, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`cartesian}, {2, -$CellContext`polar}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added independent rule ", $CellContext`T[{1, -$CellContext`cartesian}, {
     2, -$CellContext`polar}] -> $CellContext`T[{
     1, -$CellContext`cartesian}, {2, -$CellContext`polar}], 
   " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752305719392*^9, 3.728752613908391*^9, 
  3.728752678036072*^9, 3.728753082231434*^9, 3.728753159318799*^9, 
  3.730913218393056*^9},
 CellLabel->"During evaluation of In[28]:=",
 CellID->725879313,ExpressionUUID->"1fa19b9e-a93e-40b6-b253-45da3530220f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, -$CellContext`cartesian}, {0, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 1., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`polar}, {2, -$CellContext`cartesian}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`T[{2, -$CellContext`cartesian}, {
     0, -$CellContext`polar}] -> -$CellContext`T[{0, -$CellContext`polar}, {
      2, -$CellContext`cartesian}], " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752305719392*^9, 3.728752613908391*^9, 
  3.728752678036072*^9, 3.728753082231434*^9, 3.728753159318799*^9, 
  3.730913218464748*^9},
 CellLabel->"During evaluation of In[28]:=",
 CellID->2146498631,ExpressionUUID->"8b579b3a-246e-4f34-9353-d4a74cca346d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, -$CellContext`cartesian}, {1, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 1., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`polar}, {2, -$CellContext`cartesian}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`T[{2, -$CellContext`cartesian}, {
     1, -$CellContext`polar}] -> -$CellContext`T[{1, -$CellContext`polar}, {
      2, -$CellContext`cartesian}], " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752305719392*^9, 3.728752613908391*^9, 
  3.728752678036072*^9, 3.728753082231434*^9, 3.728753159318799*^9, 
  3.730913218550145*^9},
 CellLabel->"During evaluation of In[28]:=",
 CellID->624705264,ExpressionUUID->"7cd88ef7-12c4-4778-8a39-df7d4b643c41"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, -$CellContext`cartesian}, {2, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, -$CellContext`cartesian}, {2, -$CellContext`polar}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added independent rule ", $CellContext`T[{2, -$CellContext`cartesian}, {
     2, -$CellContext`polar}] -> $CellContext`T[{
     2, -$CellContext`cartesian}, {2, -$CellContext`polar}], 
   " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752305719392*^9, 3.728752613908391*^9, 
  3.728752678036072*^9, 3.728753082231434*^9, 3.728753159318799*^9, 
  3.730913218633333*^9},
 CellLabel->"During evaluation of In[28]:=",
 CellID->373821865,ExpressionUUID->"596fe376-6224-4098-b443-f5597212e90f"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`cartesian}, {0, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`cartesian}, {0, -$CellContext`polar}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`cartesian}, {1, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`cartesian}, {1, -$CellContext`polar}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`cartesian}, {2, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`cartesian}, {2, -$CellContext`polar}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`cartesian}, {0, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 1., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`polar}, {1, -$CellContext`cartesian}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`cartesian}, {1, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`cartesian}, {1, -$CellContext`polar}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`cartesian}, {2, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`cartesian}, {2, -$CellContext`polar}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, -$CellContext`cartesian}, {0, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 1., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`polar}, {2, -$CellContext`cartesian}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, -$CellContext`cartesian}, {1, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 1., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`polar}, {2, -$CellContext`cartesian}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, -$CellContext`cartesian}, {2, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, -$CellContext`cartesian}, {2, -$CellContext`polar}],
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728752306645707*^9, 3.728752614683621*^9, 
  3.72875267903281*^9, 3.72875308325251*^9, 3.728753161144342*^9, 
  3.730913218708816*^9},
 CellLabel->"Out[28]=",
 CellID->1434417580,ExpressionUUID->"ab69764c-9f01-48ef-a555-442ce3d27aeb"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TensorValues", "[", "T", "]"}]], "Input",
 CellLabel->"In[29]:=",
 CellID->356162247,ExpressionUUID->"99abb788-1ea3-4e18-9477-399b236b9302"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`polar}, {1, -$CellContext`cartesian}],
       
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 1., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{1, -$CellContext`cartesian}, {2, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`polar}, {0, -$CellContext`cartesian}],
       
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 1., 0.]], 
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`cartesian}, {0, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`polar}, {0, -$CellContext`cartesian}],
       
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 1., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`cartesian}, {1, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`polar}, {1, -$CellContext`cartesian}],
       
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 1., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{1, -$CellContext`cartesian}, {1, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`polar}, {0, -$CellContext`cartesian}],
       
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 1., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`cartesian}, {2, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`polar}, {2, -$CellContext`cartesian}],
       
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["2",
                 FontColor->RGBColor[0., 1., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{2, -$CellContext`cartesian}, {2, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`cartesian}, {0, -$CellContext`polar}],
       
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[0., 1., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`polar}, {1, -$CellContext`cartesian}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`cartesian}, {0, -$CellContext`polar}],
       
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[0., 1., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`polar}, {2, -$CellContext`cartesian}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`cartesian}, {1, -$CellContext`polar}],
       
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[0., 1., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{1, -$CellContext`polar}, {2, -$CellContext`cartesian}],
        Editable->False]}]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`polar}, {1, -$CellContext`cartesian}],
       
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`polar}, {1, -$CellContext`cartesian}],
       
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`polar}, {2, -$CellContext`cartesian}],
       
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`polar}, {2, -$CellContext`cartesian}],
       
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`polar}, {2, -$CellContext`cartesian}],
       
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`polar}, {2, -$CellContext`cartesian}],
       
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`cartesian}, {0, -$CellContext`polar}],
       
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`cartesian}, {0, -$CellContext`polar}],
       
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`cartesian}, {1, -$CellContext`polar}],
       
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`cartesian}, {1, -$CellContext`polar}],
       
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`cartesian}, {2, -$CellContext`polar}],
       
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`cartesian}, {2, -$CellContext`polar}],
       
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`cartesian}, {1, -$CellContext`polar}],
       
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`cartesian}, {1, -$CellContext`polar}],
       
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`cartesian}, {2, -$CellContext`polar}],
       
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`cartesian}, {2, -$CellContext`polar}],
       
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`cartesian}, {2, -$CellContext`polar}],
       
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`cartesian}, {2, -$CellContext`polar}],
       
       Editable->False]}]}], "}"}]}], "]"}]], "Output",
 CellChangeTimes->{3.7287523089617147`*^9, 3.728752614858613*^9, 
  3.7287526798365107`*^9, 3.72875308342527*^9, 3.728753161346964*^9, 
  3.730913220102344*^9},
 CellLabel->"Out[29]=",
 CellID->841919323,ExpressionUUID->"dd762f07-93b9-4cd0-a0b2-aaca22f05a75"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TensorValIDs", "[", "T", "]"}]], "Input",
 CellLabel->"In[30]:=",
 CellID->409052785,ExpressionUUID->"327f707a-a81f-4fe6-97bb-091622f4886f"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"ValID", "[", 
   RowBox[{"T", ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "cartesian"}], ",", 
        RowBox[{"-", "polar"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "polar"}], ",", 
        RowBox[{"-", "cartesian"}]}], "}"}]}], "}"}]}], "]"}], 
  "}"}]], "Output",
 CellChangeTimes->{3.728752310181149*^9, 3.72875261499522*^9, 
  3.728752680891336*^9, 3.728753083558694*^9, 3.7287531615206547`*^9, 
  3.730913220767129*^9},
 CellLabel->"Out[30]=",
 CellID->1800952333,ExpressionUUID->"986f8825-5f65-4b64-b8ad-d214edfe2a76"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Now we still only have one ",
 StyleBox["ValID", "MR"],
 " for ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["T", 
     StyleBox["ab",
      FontSlant->"Italic"]], ",", " ", 
    RowBox[{
    "but", " ", "it", " ", "consists", " ", "of", " ", "two", " ", "basis", 
     " ", 
     RowBox[{"configurations", ".", "  "}]}]}], TraditionalForm]],
  ExpressionUUID->"ee8cf952-3582-47c3-a3ba-d5256d82678d"],
 "As we can see, the FoldedRule returned by ",
 StyleBox["TensorValues", "MR"],
 " mixes the cases ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["T", 
     StyleBox[
      RowBox[{
       StyleBox["a",
        FontColor->RGBColor[1, 0, 0]], 
       StyleBox["b",
        FontColor->RGBColor[0, 1, 0]]}],
      FontSlant->"Italic"]], " ", "and", " ", 
    SubscriptBox["T", 
     StyleBox[
      RowBox[{
       StyleBox["b",
        FontColor->RGBColor[0, 1, 0]], 
       StyleBox["a",
        FontColor->RGBColor[1, 0, 0]]}],
      FontSlant->"Italic"]]}], TraditionalForm]],ExpressionUUID->
  "2679ccdd-d372-44bd-9800-fc28b28191d8"],
 ". Another example, now mixing index characters:"
}], "MathCaption",
 CellID->1229868459,ExpressionUUID->"03068d6a-8a39-4d9b-ba3e-4543cd89f9eb"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ComponentArray", "[", 
   RowBox[{"T", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"a", ",", "polar"}], "}"}], ",", 
     RowBox[{"-", 
      RowBox[{"{", 
       RowBox[{"b", ",", "cartesian"}], "}"}]}]}], "]"}], "]"}], " ", "//", 
  "Flatten"}]], "Input",
 CellLabel->"In[31]:=",
 CellID->869088038,ExpressionUUID->"35723d94-8072-420e-b892-49fe7600370b"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], " "},
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 1., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{0, $CellContext`polar}, {0, -$CellContext`cartesian}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], " "},
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 1., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{0, $CellContext`polar}, {1, -$CellContext`cartesian}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], " "},
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 1., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{0, $CellContext`polar}, {2, -$CellContext`cartesian}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "},
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 1., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{1, $CellContext`polar}, {0, -$CellContext`cartesian}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "},
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 1., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{1, $CellContext`polar}, {1, -$CellContext`cartesian}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "},
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 1., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{1, $CellContext`polar}, {2, -$CellContext`cartesian}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "},
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 1., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{2, $CellContext`polar}, {0, -$CellContext`cartesian}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "},
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 1., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{2, $CellContext`polar}, {1, -$CellContext`cartesian}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "},
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 1., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{2, $CellContext`polar}, {2, -$CellContext`cartesian}],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728752311727783*^9, 3.728752615152969*^9, 
  3.728752682542748*^9, 3.728753083701407*^9, 3.728753161678153*^9, 
  3.7309132220577908`*^9},
 CellLabel->"Out[31]=",
 CellID->355857753,ExpressionUUID->"0dfd7c9e-b3b2-435d-bab3-253751f5017b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentValue", " ", "/@", " ", "%"}]], "Input",
 CellLabel->"In[32]:=",
 CellID->472884516,ExpressionUUID->"99656b06-9b33-47c6-9a98-dadc428aa399"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, $CellContext`polar}, {0, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["0",
               FontColor->RGBColor[0., 1., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`cartesian}, {0, $CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`T[{0, $CellContext`polar}, {
     0, -$CellContext`cartesian}] -> -$CellContext`T[{
      0, -$CellContext`cartesian}, {0, $CellContext`polar}], 
   " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752312458556*^9, 3.728752615294054*^9, 
  3.7287526831231956`*^9, 3.728753084150743*^9, 3.728753162460909*^9, 
  3.7309132226239634`*^9},
 CellLabel->"During evaluation of In[32]:=",
 CellID->2099191193,ExpressionUUID->"8a4934c1-847e-4602-8f2b-c4cd10a8e83f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, $CellContext`polar}, {1, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, $CellContext`polar}, {1, -$CellContext`cartesian}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added independent rule ", $CellContext`T[{0, $CellContext`polar}, {
     1, -$CellContext`cartesian}] -> $CellContext`T[{0, $CellContext`polar}, {
     1, -$CellContext`cartesian}], " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752312458556*^9, 3.728752615294054*^9, 
  3.7287526831231956`*^9, 3.728753084150743*^9, 3.728753162460909*^9, 
  3.730913222678569*^9},
 CellLabel->"During evaluation of In[32]:=",
 CellID->97544188,ExpressionUUID->"9a908228-7833-4c5d-a185-ae5067070fe2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, $CellContext`polar}, {2, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, $CellContext`polar}, {2, -$CellContext`cartesian}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added independent rule ", $CellContext`T[{0, $CellContext`polar}, {
     2, -$CellContext`cartesian}] -> $CellContext`T[{0, $CellContext`polar}, {
     2, -$CellContext`cartesian}], " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752312458556*^9, 3.728752615294054*^9, 
  3.7287526831231956`*^9, 3.728753084150743*^9, 3.728753162460909*^9, 
  3.7309132227651777`*^9},
 CellLabel->"During evaluation of In[32]:=",
 CellID->895145055,ExpressionUUID->"c3bc196a-727b-4af8-b845-3e8c02b5591b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, $CellContext`polar}, {0, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["0",
               FontColor->RGBColor[0., 1., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`cartesian}, {1, $CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`T[{1, $CellContext`polar}, {
     0, -$CellContext`cartesian}] -> -$CellContext`T[{
      0, -$CellContext`cartesian}, {1, $CellContext`polar}], 
   " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752312458556*^9, 3.728752615294054*^9, 
  3.7287526831231956`*^9, 3.728753084150743*^9, 3.728753162460909*^9, 
  3.7309132228497047`*^9},
 CellLabel->"During evaluation of In[32]:=",
 CellID->1579992150,ExpressionUUID->"56ce7bf8-3d30-4fab-b90b-acef8d75ff65"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, $CellContext`polar}, {1, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 1., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`cartesian}, {1, $CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`T[{1, $CellContext`polar}, {
     1, -$CellContext`cartesian}] -> -$CellContext`T[{
      1, -$CellContext`cartesian}, {1, $CellContext`polar}], 
   " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752312458556*^9, 3.728752615294054*^9, 
  3.7287526831231956`*^9, 3.728753084150743*^9, 3.728753162460909*^9, 
  3.730913222933041*^9},
 CellLabel->"During evaluation of In[32]:=",
 CellID->1292769895,ExpressionUUID->"c747e527-a9da-42f9-a80a-bd360e94e5e4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, $CellContext`polar}, {2, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, $CellContext`polar}, {2, -$CellContext`cartesian}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added independent rule ", $CellContext`T[{1, $CellContext`polar}, {
     2, -$CellContext`cartesian}] -> $CellContext`T[{1, $CellContext`polar}, {
     2, -$CellContext`cartesian}], " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752312458556*^9, 3.728752615294054*^9, 
  3.7287526831231956`*^9, 3.728753084150743*^9, 3.728753162460909*^9, 
  3.730913223060772*^9},
 CellLabel->"During evaluation of In[32]:=",
 CellID->76266324,ExpressionUUID->"b4404af1-1d48-432b-ba1c-9761f1183b11"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, $CellContext`polar}, {0, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["0",
               FontColor->RGBColor[0., 1., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`cartesian}, {2, $CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`T[{2, $CellContext`polar}, {
     0, -$CellContext`cartesian}] -> -$CellContext`T[{
      0, -$CellContext`cartesian}, {2, $CellContext`polar}], 
   " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752312458556*^9, 3.728752615294054*^9, 
  3.7287526831231956`*^9, 3.728753084150743*^9, 3.728753162460909*^9, 
  3.730913223127694*^9},
 CellLabel->"During evaluation of In[32]:=",
 CellID->1907853856,ExpressionUUID->"e566e3a0-a7ba-441b-a27d-50c2070b2516"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, $CellContext`polar}, {1, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 1., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`cartesian}, {2, $CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`T[{2, $CellContext`polar}, {
     1, -$CellContext`cartesian}] -> -$CellContext`T[{
      1, -$CellContext`cartesian}, {2, $CellContext`polar}], 
   " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752312458556*^9, 3.728752615294054*^9, 
  3.7287526831231956`*^9, 3.728753084150743*^9, 3.728753162460909*^9, 
  3.7309132231962337`*^9},
 CellLabel->"During evaluation of In[32]:=",
 CellID->1221424435,ExpressionUUID->"85651165-83cb-432f-9a6b-27a51aa1ab21"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, $CellContext`polar}, {2, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["2",
               FontColor->RGBColor[0., 1., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, -$CellContext`cartesian}, {2, $CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`T[{2, $CellContext`polar}, {
     2, -$CellContext`cartesian}] -> -$CellContext`T[{
      2, -$CellContext`cartesian}, {2, $CellContext`polar}], 
   " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752312458556*^9, 3.728752615294054*^9, 
  3.7287526831231956`*^9, 3.728753084150743*^9, 3.728753162460909*^9, 
  3.7309132232819777`*^9},
 CellLabel->"During evaluation of In[32]:=",
 CellID->684297372,ExpressionUUID->"23400bde-2282-4148-abdd-7533e1734029"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, $CellContext`polar}, {0, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["0",
               FontColor->RGBColor[0., 1., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`cartesian}, {0, $CellContext`polar}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, $CellContext`polar}, {1, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, $CellContext`polar}, {1, -$CellContext`cartesian}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, $CellContext`polar}, {2, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, $CellContext`polar}, {2, -$CellContext`cartesian}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, $CellContext`polar}, {0, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["0",
               FontColor->RGBColor[0., 1., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`cartesian}, {1, $CellContext`polar}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, $CellContext`polar}, {1, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 1., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`cartesian}, {1, $CellContext`polar}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, $CellContext`polar}, {2, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, $CellContext`polar}, {2, -$CellContext`cartesian}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, $CellContext`polar}, {0, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["0",
               FontColor->RGBColor[0., 1., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`cartesian}, {2, $CellContext`polar}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, $CellContext`polar}, {1, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 1., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`cartesian}, {2, $CellContext`polar}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, $CellContext`polar}, {2, -$CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["2",
               FontColor->RGBColor[0., 1., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{2, -$CellContext`cartesian}, {2, $CellContext`polar}],
      Editable->False]}]}]}], "}"}]], "Output",
 CellChangeTimes->{3.7287523138757772`*^9, 3.728752616001738*^9, 
  3.7287526840968723`*^9, 3.728753084968835*^9, 3.728753163480247*^9, 
  3.730913223360587*^9},
 CellLabel->"Out[32]=",
 CellID->1211001519,ExpressionUUID->"7aee3a74-2357-4e72-9be7-bc86b1b4b9d3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ComponentArray", "[", 
   RowBox[{"T", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "a"}], ",", "cartesian"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"a", ",", "polar"}], "}"}]}], "]"}], "]"}], " ", "//", 
  "Flatten"}]], "Input",
 CellLabel->"In[33]:=",
 CellID->641639522,ExpressionUUID->"d87201b3-cacc-4cbd-9892-ced8b530673e"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["0",
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{0, -$CellContext`cartesian}, {0, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["0",
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{0, -$CellContext`cartesian}, {1, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["0",
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{0, -$CellContext`cartesian}, {2, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{1, -$CellContext`cartesian}, {0, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{1, -$CellContext`cartesian}, {1, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{1, -$CellContext`cartesian}, {2, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{2, -$CellContext`cartesian}, {0, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{2, -$CellContext`cartesian}, {1, $CellContext`polar}],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 1., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{2, -$CellContext`cartesian}, {2, $CellContext`polar}],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728752314945395*^9, 3.728752616169639*^9, 
  3.728752684791965*^9, 3.728753085135903*^9, 3.728753163698567*^9, 
  3.730913224126588*^9},
 CellLabel->"Out[33]=",
 CellID->2024367663,ExpressionUUID->"8beeee48-ca51-4ca2-8ffe-5380ac3d7b32"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentValue", " ", "/@", " ", "%"}]], "Input",
 CellLabel->"In[34]:=",
 CellID->1093323557,ExpressionUUID->"342f5ca0-100c-435e-91a9-6ae297c5e9fa"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`cartesian}, {0, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`cartesian}, {0, $CellContext`polar}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added independent rule ", $CellContext`T[{0, -$CellContext`cartesian}, {
     0, $CellContext`polar}] -> $CellContext`T[{0, -$CellContext`cartesian}, {
     0, $CellContext`polar}], " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752315621111*^9, 3.728752616316856*^9, 
  3.728752685438339*^9, 3.728753085258987*^9, 3.728753163930303*^9, 
  3.7309132246918907`*^9},
 CellLabel->"During evaluation of In[34]:=",
 CellID->693721341,ExpressionUUID->"6bea4520-88b1-4400-a3fd-926d646164a1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`cartesian}, {1, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`cartesian}, {1, $CellContext`polar}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added independent rule ", $CellContext`T[{0, -$CellContext`cartesian}, {
     1, $CellContext`polar}] -> $CellContext`T[{0, -$CellContext`cartesian}, {
     1, $CellContext`polar}], " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752315621111*^9, 3.728752616316856*^9, 
  3.728752685438339*^9, 3.728753085258987*^9, 3.728753163930303*^9, 
  3.730913224745248*^9},
 CellLabel->"During evaluation of In[34]:=",
 CellID->685858624,ExpressionUUID->"e56c202a-d5c2-4859-94e9-6279d85935b8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`cartesian}, {2, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`cartesian}, {2, $CellContext`polar}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added independent rule ", $CellContext`T[{0, -$CellContext`cartesian}, {
     2, $CellContext`polar}] -> $CellContext`T[{0, -$CellContext`cartesian}, {
     2, $CellContext`polar}], " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752315621111*^9, 3.728752616316856*^9, 
  3.728752685438339*^9, 3.728753085258987*^9, 3.728753163930303*^9, 
  3.730913224833591*^9},
 CellLabel->"During evaluation of In[34]:=",
 CellID->1681160119,ExpressionUUID->"b305cd9e-7f27-47da-a2cf-0e62e73aa05b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`cartesian}, {0, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 1., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, $CellContext`polar}, {1, -$CellContext`cartesian}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`T[{1, -$CellContext`cartesian}, {
     0, $CellContext`polar}] -> -$CellContext`T[{0, $CellContext`polar}, {
      1, -$CellContext`cartesian}], " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752315621111*^9, 3.728752616316856*^9, 
  3.728752685438339*^9, 3.728753085258987*^9, 3.728753163930303*^9, 
  3.730913224916624*^9},
 CellLabel->"During evaluation of In[34]:=",
 CellID->815776454,ExpressionUUID->"9d0163bd-7dc5-4639-8857-3666273ba017"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`cartesian}, {1, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`cartesian}, {1, $CellContext`polar}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added independent rule ", $CellContext`T[{1, -$CellContext`cartesian}, {
     1, $CellContext`polar}] -> $CellContext`T[{1, -$CellContext`cartesian}, {
     1, $CellContext`polar}], " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752315621111*^9, 3.728752616316856*^9, 
  3.728752685438339*^9, 3.728753085258987*^9, 3.728753163930303*^9, 
  3.73091322499986*^9},
 CellLabel->"During evaluation of In[34]:=",
 CellID->1219259930,ExpressionUUID->"4756dd63-49f2-4a43-b83f-1b128d27393c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`cartesian}, {2, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`cartesian}, {2, $CellContext`polar}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added independent rule ", $CellContext`T[{1, -$CellContext`cartesian}, {
     2, $CellContext`polar}] -> $CellContext`T[{1, -$CellContext`cartesian}, {
     2, $CellContext`polar}], " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752315621111*^9, 3.728752616316856*^9, 
  3.728752685438339*^9, 3.728753085258987*^9, 3.728753163930303*^9, 
  3.730913225129939*^9},
 CellLabel->"During evaluation of In[34]:=",
 CellID->530736966,ExpressionUUID->"5c49f10c-d645-44c0-975c-a5b651c25c27"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, -$CellContext`cartesian}, {0, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 1., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, $CellContext`polar}, {2, -$CellContext`cartesian}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`T[{2, -$CellContext`cartesian}, {
     0, $CellContext`polar}] -> -$CellContext`T[{0, $CellContext`polar}, {
      2, -$CellContext`cartesian}], " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752315621111*^9, 3.728752616316856*^9, 
  3.728752685438339*^9, 3.728753085258987*^9, 3.728753163930303*^9, 
  3.7309132251922693`*^9},
 CellLabel->"During evaluation of In[34]:=",
 CellID->1596601571,ExpressionUUID->"68b01eeb-faf9-4807-b7a5-b9b114ebb9f5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, -$CellContext`cartesian}, {1, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 1., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, $CellContext`polar}, {2, -$CellContext`cartesian}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`T[{2, -$CellContext`cartesian}, {
     1, $CellContext`polar}] -> -$CellContext`T[{1, $CellContext`polar}, {
      2, -$CellContext`cartesian}], " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752315621111*^9, 3.728752616316856*^9, 
  3.728752685438339*^9, 3.728753085258987*^9, 3.728753163930303*^9, 
  3.730913225267345*^9},
 CellLabel->"During evaluation of In[34]:=",
 CellID->1274434532,ExpressionUUID->"a9ce7a5a-572d-4dd1-b12f-22bfe5de972e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, -$CellContext`cartesian}, {2, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, -$CellContext`cartesian}, {2, $CellContext`polar}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added independent rule ", $CellContext`T[{2, -$CellContext`cartesian}, {
     2, $CellContext`polar}] -> $CellContext`T[{2, -$CellContext`cartesian}, {
     2, $CellContext`polar}], " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752315621111*^9, 3.728752616316856*^9, 
  3.728752685438339*^9, 3.728753085258987*^9, 3.728753163930303*^9, 
  3.730913225331307*^9},
 CellLabel->"During evaluation of In[34]:=",
 CellID->419284146,ExpressionUUID->"b601d8a1-c0fe-4846-93bf-cf9de3b8b324"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`cartesian}, {0, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`cartesian}, {0, $CellContext`polar}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`cartesian}, {1, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`cartesian}, {1, $CellContext`polar}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`cartesian}, {2, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`cartesian}, {2, $CellContext`polar}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`cartesian}, {0, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 1., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, $CellContext`polar}, {1, -$CellContext`cartesian}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`cartesian}, {1, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`cartesian}, {1, $CellContext`polar}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`cartesian}, {2, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`cartesian}, {2, $CellContext`polar}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, -$CellContext`cartesian}, {0, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 1., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, $CellContext`polar}, {2, -$CellContext`cartesian}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, -$CellContext`cartesian}, {1, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 1., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, $CellContext`polar}, {2, -$CellContext`cartesian}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, -$CellContext`cartesian}, {2, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{2, -$CellContext`cartesian}, {2, $CellContext`polar}],
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728752317061481*^9, 3.728752617207735*^9, 
  3.7287526864049807`*^9, 3.728753085866696*^9, 3.728753165861928*^9, 
  3.730913225399885*^9},
 CellLabel->"Out[34]=",
 CellID->1014796015,ExpressionUUID->"3ab851e9-124d-4751-9f6c-a9fc063c3e0f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TensorValIDs", "[", "T", "]"}]], "Input",
 CellLabel->"In[35]:=",
 CellID->1026717895,ExpressionUUID->"53f239b8-9ba6-44f3-8e2e-d2b4c4381418"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"ValID", "[", 
    RowBox[{"T", ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "cartesian"}], ",", "polar"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"polar", ",", 
         RowBox[{"-", "cartesian"}]}], "}"}]}], "}"}]}], "]"}], ",", 
   RowBox[{"ValID", "[", 
    RowBox[{"T", ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "cartesian"}], ",", 
         RowBox[{"-", "polar"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "polar"}], ",", 
         RowBox[{"-", "cartesian"}]}], "}"}]}], "}"}]}], "]"}]}], 
  "}"}]], "Output",
 CellChangeTimes->{3.728752318121585*^9, 3.728752617364406*^9, 
  3.728752686905304*^9, 3.728753085979094*^9, 3.7287531660360937`*^9, 
  3.7309132260019093`*^9},
 CellLabel->"Out[35]=",
 CellID->1881863052,ExpressionUUID->"a7280395-7d2a-41cf-bfd4-47ca5325f297"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "We have two ",
 StyleBox["ValIDs", "MR"],
 ", both consisting of two basis configurations. Notice how newer ",
 StyleBox["ValIDs", "MR"],
 " appear first.We can retrieve the ",
 StyleBox["TensorValues", "MR"],
 " for either one of them:"
}], "MathCaption",
 CellChangeTimes->{3.728752017449646*^9},
 CellID->2051507455,ExpressionUUID->"a7417fde-1046-4c5d-868b-ff0b612a90ad"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"TensorValues", "[", 
   RowBox[{"T", ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "cartesian"}], ",", "polar"}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"polar", ",", 
        RowBox[{"-", "cartesian"}]}], "}"}]}], "}"}]}], "]"}], " "}]], "Input",\

 CellLabel->"In[36]:=",
 CellID->1835843600,ExpressionUUID->"71a9234b-b600-4777-a11d-0ae6440c3f15"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, $CellContext`polar}, {0, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]]},
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 1., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`cartesian}, {0, $CellContext`polar}],
        
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, $CellContext`polar}, {0, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]},
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 1., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`cartesian}, {1, $CellContext`polar}],
        
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, $CellContext`polar}, {1, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]},
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 1., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{1, -$CellContext`cartesian}, {1, $CellContext`polar}],
        
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, $CellContext`polar}, {0, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 1., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`cartesian}, {2, $CellContext`polar}],
        
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, $CellContext`polar}, {1, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 1., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{1, -$CellContext`cartesian}, {2, $CellContext`polar}],
        
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, $CellContext`polar}, {2, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {
                StyleBox["2",
                 FontColor->RGBColor[0., 1., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{2, -$CellContext`cartesian}, {2, $CellContext`polar}],
        
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`cartesian}, {0, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], " "},
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[0., 1., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, $CellContext`polar}, {1, -$CellContext`cartesian}],
        
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`cartesian}, {0, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], " "},
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[0., 1., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, $CellContext`polar}, {2, -$CellContext`cartesian}],
        
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`cartesian}, {1, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], " "},
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[0., 1., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{1, $CellContext`polar}, {2, -$CellContext`cartesian}],
        
        Editable->False]}]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, $CellContext`polar}, {1, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, $CellContext`polar}, {1, -$CellContext`cartesian}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, $CellContext`polar}, {2, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, $CellContext`polar}, {2, -$CellContext`cartesian}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, $CellContext`polar}, {2, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, $CellContext`polar}, {2, -$CellContext`cartesian}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`cartesian}, {0, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`cartesian}, {0, $CellContext`polar}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`cartesian}, {1, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`cartesian}, {1, $CellContext`polar}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`cartesian}, {2, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`cartesian}, {2, $CellContext`polar}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`cartesian}, {1, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`cartesian}, {1, $CellContext`polar}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`cartesian}, {2, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`cartesian}, {2, $CellContext`polar}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`cartesian}, {2, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`cartesian}, {2, $CellContext`polar}],
       Editable->False]}]}], "}"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728752319398776*^9, 3.728752617539462*^9, 
  3.7287526882691393`*^9, 3.728753086171112*^9, 3.728753166215345*^9, 
  3.730913227231992*^9},
 CellLabel->"Out[36]=",
 CellID->2133326197,ExpressionUUID->"92c64f40-dfed-4422-9247-52179e83909c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["or for all at the same time:", "MathCaption",
 CellID->1118382918,ExpressionUUID->"169747f2-ae56-4bdf-9c19-50e2b4400040"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TensorValues", "[", "T", "]"}]], "Input",
 CellLabel->"In[37]:=",
 CellID->1884824282,ExpressionUUID->"97e51d4e-b578-49ae-8148-42f09b5ed989"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, $CellContext`polar}, {0, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]]},
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 1., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`cartesian}, {0, $CellContext`polar}],
        
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, $CellContext`polar}, {0, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]},
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 1., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`cartesian}, {1, $CellContext`polar}],
        
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, $CellContext`polar}, {1, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]},
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 1., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{1, -$CellContext`cartesian}, {1, $CellContext`polar}],
        
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, $CellContext`polar}, {0, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 1., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`cartesian}, {2, $CellContext`polar}],
        
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, $CellContext`polar}, {1, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 1., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{1, -$CellContext`cartesian}, {2, $CellContext`polar}],
        
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, $CellContext`polar}, {2, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {
                StyleBox["2",
                 FontColor->RGBColor[0., 1., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{2, -$CellContext`cartesian}, {2, $CellContext`polar}],
        
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`cartesian}, {0, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], " "},
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[0., 1., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, $CellContext`polar}, {1, -$CellContext`cartesian}],
        
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`cartesian}, {0, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], " "},
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[0., 1., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, $CellContext`polar}, {2, -$CellContext`cartesian}],
        
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`cartesian}, {1, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], " "},
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[0., 1., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{1, $CellContext`polar}, {2, -$CellContext`cartesian}],
        
        Editable->False]}]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, $CellContext`polar}, {1, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, $CellContext`polar}, {1, -$CellContext`cartesian}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, $CellContext`polar}, {2, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, $CellContext`polar}, {2, -$CellContext`cartesian}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, $CellContext`polar}, {2, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, $CellContext`polar}, {2, -$CellContext`cartesian}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`cartesian}, {0, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`cartesian}, {0, $CellContext`polar}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`cartesian}, {1, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`cartesian}, {1, $CellContext`polar}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`cartesian}, {2, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`cartesian}, {2, $CellContext`polar}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`cartesian}, {1, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`cartesian}, {1, $CellContext`polar}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`cartesian}, {2, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`cartesian}, {2, $CellContext`polar}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`cartesian}, {2, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`cartesian}, {2, $CellContext`polar}],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`polar}, {1, -$CellContext`cartesian}],
       
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 1., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{1, -$CellContext`cartesian}, {2, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`polar}, {0, -$CellContext`cartesian}],
       
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 1., 0.]], 
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`cartesian}, {0, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`polar}, {0, -$CellContext`cartesian}],
       
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 1., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`cartesian}, {1, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`polar}, {1, -$CellContext`cartesian}],
       
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 1., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{1, -$CellContext`cartesian}, {1, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`polar}, {0, -$CellContext`cartesian}],
       
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[0., 1., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`cartesian}, {2, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`polar}, {2, -$CellContext`cartesian}],
       
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["2",
                 FontColor->RGBColor[0., 1., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{2, -$CellContext`cartesian}, {2, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`cartesian}, {0, -$CellContext`polar}],
       
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[0., 1., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`polar}, {1, -$CellContext`cartesian}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`cartesian}, {0, -$CellContext`polar}],
       
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[0., 1., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`polar}, {2, -$CellContext`cartesian}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`cartesian}, {1, -$CellContext`polar}],
       
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[0., 1., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{1, -$CellContext`polar}, {2, -$CellContext`cartesian}],
        Editable->False]}]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`polar}, {1, -$CellContext`cartesian}],
       
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`polar}, {1, -$CellContext`cartesian}],
       
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`polar}, {2, -$CellContext`cartesian}],
       
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`polar}, {2, -$CellContext`cartesian}],
       
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`polar}, {2, -$CellContext`cartesian}],
       
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`polar}, {2, -$CellContext`cartesian}],
       
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`cartesian}, {0, -$CellContext`polar}],
       
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`cartesian}, {0, -$CellContext`polar}],
       
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`cartesian}, {1, -$CellContext`polar}],
       
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`cartesian}, {1, -$CellContext`polar}],
       
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`cartesian}, {2, -$CellContext`polar}],
       
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`cartesian}, {2, -$CellContext`polar}],
       
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`cartesian}, {1, -$CellContext`polar}],
       
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`cartesian}, {1, -$CellContext`polar}],
       
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`cartesian}, {2, -$CellContext`polar}],
       
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`cartesian}, {2, -$CellContext`polar}],
       
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`cartesian}, {2, -$CellContext`polar}],
       
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`cartesian}, {2, -$CellContext`polar}],
       
       Editable->False]}]}], "}"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728752320466778*^9, 3.7287526178032007`*^9, 
  3.728752689433504*^9, 3.728753086406266*^9, 3.728753166404232*^9, 
  3.730913228378834*^9},
 CellLabel->"Out[37]=",
 CellID->213005585,ExpressionUUID->"ca0d79af-6688-441c-980e-275a4ff21e75"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We can delete single rules:", "MathCaption",
 CellChangeTimes->{3.728752011734829*^9},
 CellID->199311019,ExpressionUUID->"9345b3bc-7590-41e6-8678-84f705a10e40"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentValue", "[", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "cartesian"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"2", ",", 
       RowBox[{"-", "polar"}]}], "}"}]}], "]"}], ",", "Null"}], 
  "]"}]], "Input",
 CellLabel->"In[38]:=",
 CellID->1893290660,ExpressionUUID->"976690c4-4194-4a98-a03a-0f34e109a0be"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Dropped independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`cartesian}, {2, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`cartesian}, {2, -$CellContext`polar}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Dropped independent rule ", $CellContext`T[{1, -$CellContext`cartesian}, {
     2, -$CellContext`polar}] -> $CellContext`T[{
     1, -$CellContext`cartesian}, {2, -$CellContext`polar}], 
   " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752321479081*^9, 3.728752618292912*^9, 
  3.728752690559114*^9, 3.728753086558134*^9, 3.728753166597722*^9, 
  3.7309132297617683`*^9},
 CellLabel->"During evaluation of In[38]:=",
 CellID->673504539,ExpressionUUID->"e0966399-a6b3-40ed-9e69-d74072af37c5"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {" ", " "},
          {
           StyleBox["1",
            FontColor->RGBColor[0., 1., 0.]], 
           StyleBox["2",
            FontColor->RGBColor[1., 0., 0.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[{1, -$CellContext`cartesian}, {2, -$CellContext`polar}],
   Editable->False], "\[Rule]", "Null"}]], "Output",
 CellChangeTimes->{3.728752321532337*^9, 3.7287526184063263`*^9, 
  3.728752690608697*^9, 3.728753086591634*^9, 3.7287531666985273`*^9, 
  3.7309132297890873`*^9},
 CellLabel->"Out[38]=",
 CellID->204483636,ExpressionUUID->"b59666d3-a62d-4807-9c8d-6fddda04cb7c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "We can also easily delete all ",
 StyleBox["TensorValues", "MR"],
 " for a given tensor:"
}], "MathCaption",
 CellChangeTimes->{3.728752009410314*^9},
 CellID->339618327,ExpressionUUID->"7ffedb75-042b-49bb-ae6d-ecab9469bc40"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DeleteTensorValues", "[", "W", "]"}]], "Input",
 CellLabel->"In[39]:=",
 CellID->2101171810,ExpressionUUID->"ef0f8cd9-b5a6-4136-9a49-866a8d3392a0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Deleted values for tensor \"\>", "\[InvisibleSpace]", "W", 
   "\[InvisibleSpace]", "\<\", derivatives \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", "}"}], "\[InvisibleSpace]", "\<\" and bases \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"{", 
     RowBox[{"polar", ",", "polar", ",", "polar"}], "}"}], "}"}], 
   "\[InvisibleSpace]", "\<\".\"\>"}],
  SequenceForm[
  "Deleted values for tensor ", $CellContext`W, ", derivatives ", {}, 
   " and bases ", {{$CellContext`polar, $CellContext`polar, \
$CellContext`polar}}, "."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752322849203*^9, 3.72875261890987*^9, 
  3.728752691800647*^9, 3.728753086710965*^9, 3.728753166864861*^9, 
  3.730913230996602*^9},
 CellLabel->"During evaluation of In[39]:=",
 CellID->2120000322,ExpressionUUID->"4d13ca9f-33dc-4dd4-9add-411571941ab1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "or only those for a single ",
 StyleBox["ValID:", "MR"]
}], "MathCaption",
 CellChangeTimes->{3.728752006463127*^9},
 CellID->850775366,ExpressionUUID->"a3a725ab-52c5-452e-b216-9edc5116fcb8"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DeleteTensorValues", "[", 
  RowBox[{"T", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "cartesian"}], ",", "polar"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"polar", ",", 
       RowBox[{"-", "cartesian"}]}], "}"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[40]:=",
 CellID->1035454636,ExpressionUUID->"63d5bcab-202d-441e-9ae7-4e61609b6fec"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Deleted values for tensor \"\>", "\[InvisibleSpace]", "T", 
   "\[InvisibleSpace]", "\<\", derivatives \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", "}"}], "\[InvisibleSpace]", "\<\" and bases \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "cartesian"}], ",", "polar"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"polar", ",", 
       RowBox[{"-", "cartesian"}]}], "}"}]}], "}"}], 
   "\[InvisibleSpace]", "\<\".\"\>"}],
  SequenceForm[
  "Deleted values for tensor ", $CellContext`T, ", derivatives ", {}, 
   " and bases ", {{-$CellContext`cartesian, $CellContext`polar}, \
{$CellContext`polar, -$CellContext`cartesian}}, "."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752325521688*^9, 3.728752619318912*^9, 
  3.728752692838235*^9, 3.728753086845047*^9, 3.7287531670332747`*^9, 
  3.730913232715077*^9},
 CellLabel->"During evaluation of In[40]:=",
 CellID->229910653,ExpressionUUID->"99e19f52-e553-4e38-907f-1513bfbed6b3"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["4.4. ComponentValue and TensorValues with independent values", \
"Subsection",
 CellChangeTimes->{{3.728673952037448*^9, 3.7286739530573797`*^9}, {
  3.728675286575617*^9, 3.7286752867677383`*^9}},
 CellID->1163994800,ExpressionUUID->"8ba8375c-3190-4711-9a10-32582d770382"],

Cell[CellGroupData[{

Cell[TextData[{
 "Until now, the second list of the FoldedRules contained only trivial \
replacement rules such as ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["T", 
     StyleBox["11",
      FontColor->RGBColor[1, 0, 0]]], "\[Rule]", " ", 
    SubscriptBox["T", 
     StyleBox["11",
      FontColor->RGBColor[1, 0, 0]]]}], TraditionalForm]],ExpressionUUID->
  "27fce985-1420-4feb-ad62-b4133df7169a"],
 ". It is very simple to specify values for the LHS, we just have to call ",
 StyleBox["ComponentValue", "MR"],
 " with two arguments:"
}], "Text",
 CellID->1832063077,ExpressionUUID->"8bf8be01-0f69-43ad-bbd4-681b9392ff3c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentValue", "[", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "polar"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"2", ",", 
       RowBox[{"-", "polar"}]}], "}"}]}], "]"}], ",", "5"}], "]"}]], "Input",
 CellLabel->"In[41]:=",
 CellID->863604453,ExpressionUUID->"76257141-3971-4d43-ab65-b129a7a3258b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
     Editable->False], "\[Rule]", "5"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added independent rule ", $CellContext`T[{1, -$CellContext`polar}, {
     2, -$CellContext`polar}] -> 5, " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.72875233325113*^9, 3.728752697163075*^9, 
  3.728753086979504*^9, 3.7287531696816797`*^9, 3.7309132411790333`*^9},
 CellLabel->"During evaluation of In[41]:=",
 CellID->77600358,ExpressionUUID->"ec3c37f8-8801-4ce2-ab08-0401f8d4435b"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {" ", " "},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], 
           StyleBox["2",
            FontColor->RGBColor[1., 0., 0.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
   Editable->False], "\[Rule]", "5"}]], "Output",
 CellChangeTimes->{3.728752333340056*^9, 3.7287526972479887`*^9, 
  3.7287530870068607`*^9, 3.7287531697658873`*^9, 3.730913241250617*^9},
 CellLabel->"Out[41]=",
 CellID->706643177,ExpressionUUID->"2ff9247c-8e65-4b57-90b1-ccdebee3bfab"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TensorValues", "[", 
  RowBox[{"T", ",", 
   RowBox[{"{", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "polar"}], ",", 
      RowBox[{"-", "polar"}]}], "}"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[42]:=",
 CellID->2057569429,ExpressionUUID->"12df3aa3-194e-4fa3-97e2-6c1df1bf49f2"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
      Editable->False], "\[Rule]", "5"}], "}"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728752333522002*^9, 3.728752697446907*^9, 
  3.7287530871300583`*^9, 3.7287531699648733`*^9, 3.730913241378682*^9},
 CellLabel->"Out[42]=",
 CellID->967307929,ExpressionUUID->"714e74e9-2fe0-46b9-8a91-4fa13364aa78"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Notice what happens if we specify a dependent component", "Text",
 CellID->1377949064,ExpressionUUID->"844c275a-0412-4e33-8797-880f81cee20d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentValue", "[", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "polar"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", 
       RowBox[{"-", "polar"}]}], "}"}]}], "]"}], ",", "6"}], "]"}]], "Input",
 CellLabel->"In[43]:=",
 CellID->726269400,ExpressionUUID->"2e67a1f1-dc58-4dfa-9f78-b0c83a967d57"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{1, -$CellContext`polar}, {0, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`T[{1, -$CellContext`polar}, {
     0, -$CellContext`polar}] -> -$CellContext`T[{0, -$CellContext`polar}, {
      1, -$CellContext`polar}], " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287523337040777`*^9, 3.728752697649206*^9, 
  3.728753087259014*^9, 3.728753170148881*^9, 3.730913241513988*^9},
 CellLabel->"During evaluation of In[43]:=",
 CellID->1093154913,ExpressionUUID->"b44cf555-ea86-4d53-a56a-f5369cab6621"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", "6"}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "Added independent rule ", $CellContext`T[{0, -$CellContext`polar}, {
     1, -$CellContext`polar}] -> -6, " for tensor ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287523337040777`*^9, 3.728752697649206*^9, 
  3.728753087259014*^9, 3.728753170148881*^9, 3.730913241558259*^9},
 CellLabel->"During evaluation of In[43]:=",
 CellID->932022706,ExpressionUUID->"8f7e9e3d-cf6d-4b29-8cbd-07fdbf6bd733"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {" ", " "},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], 
           StyleBox["0",
            FontColor->RGBColor[1., 0., 0.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[{1, -$CellContext`polar}, {0, -$CellContext`polar}],
   Editable->False], "\[Rule]", "6"}]], "Output",
 CellChangeTimes->{3.72875233389044*^9, 3.72875269783197*^9, 
  3.728753087314776*^9, 3.7287531703310347`*^9, 3.730913241636497*^9},
 CellLabel->"Out[43]=",
 CellID->759898524,ExpressionUUID->"3edaf57e-4fd7-4382-948d-260be6c78e08"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
First the dependent rule is added and then the value (with a possible sign) \
is assigned to the corresponding independent component:\
\>", "Text",
 CellID->261916468,ExpressionUUID->"a4011bca-ea36-4b9e-b218-bd40ca109b01"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TensorValues", "[", 
  RowBox[{"T", ",", 
   RowBox[{"{", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "polar"}], ",", 
      RowBox[{"-", "polar"}]}], "}"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[44]:=",
 CellID->1836084973,ExpressionUUID->"0228f996-fadd-43eb-92d8-82fbbec2ba66"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{1, -$CellContext`polar}, {0, -$CellContext`polar}],
      Editable->False], "\[Rule]", 
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", "5"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", "6"}]}]}], "}"}]}], "]"}]], "Output",
 CellChangeTimes->{3.72875233405853*^9, 3.7287526983793497`*^9, 
  3.728753087473195*^9, 3.728753170688373*^9, 3.7309132417803507`*^9},
 CellLabel->"Out[44]=",
 CellID->529244159,ExpressionUUID->"67de174e-f312-4f7e-adba-e28afa6a00a9"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["ComponentValue", "MR"],
 " does not allow the user to introduce an inconsistent value"
}], "Text",
 CellID->410867545,ExpressionUUID->"069cd779-6951-4532-b393-bebd20211aff"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentValue", "[", 
  RowBox[{
   RowBox[{"W", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "polar"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", "polar"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", "polar"}], "}"}]}], "]"}], ",", "5"}], "]"}]], "Input",\

 CellLabel->"In[45]:=",
 CellID->1290903710,ExpressionUUID->"daf15d97-5701-43dd-9638-2fe1124714dd"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`W[{1, $CellContext`polar}, {1, $CellContext`polar}, {
      1, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`W[{1, $CellContext`polar}, {
     1, $CellContext`polar}, {1, $CellContext`polar}] -> 0, 
   " for tensor ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287523342505617`*^9, 3.728752699014263*^9, 
  3.728753087614195*^9, 3.728753171465045*^9, 3.730913241929893*^9},
 CellLabel->"During evaluation of In[45]:=",
 CellID->129314439,ExpressionUUID->"dcd810f3-f14f-4b9b-973b-9243aec7b9b3"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"W", 
       StyleBox[GridBox[{
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], 
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], 
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]]},
          {" ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`W[{1, $CellContext`polar}, {1, $CellContext`polar}, {
    1, $CellContext`polar}],
   Editable->False], "\[Rule]", "0"}]], "Output",
 CellChangeTimes->{3.728752334318542*^9, 3.72875269909829*^9, 
  3.7287530876584053`*^9, 3.728753171631096*^9, 3.7309132419655523`*^9},
 CellLabel->"Out[45]=",
 CellID->134325509,ExpressionUUID->"b9c986fb-2374-4aec-83f8-69662261f01c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We switch the rule generation messages off", "Text",
 CellID->1540419519,ExpressionUUID->"8ea2b9af-d35e-4f6b-a027-ba8de21fbbbc"],

Cell[BoxData[
 RowBox[{
  RowBox[{"$CVVerbose", "=", "False"}], ";"}]], "Input",
 CellLabel->"In[46]:=",
 CellID->1915491248,ExpressionUUID->"2b0fe639-da89-4e49-ac04-6992e97ee071"]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["ComponentValue", "MR"],
 " is threaded on pairs of lists when it has two arguments. This allows us to \
generate all independent rules with just one command. "
}], "Text",
 CellID->734052959,ExpressionUUID->"b459a90d-aaf6-4d71-bde4-9e81b35da2e1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"values", "=", 
  RowBox[{"Table", "[", 
   RowBox[{
    RowBox[{"i", "+", "j"}], ",", 
    RowBox[{"{", 
     RowBox[{"i", ",", "1", ",", "3"}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{"j", ",", "1", ",", "3"}], "}"}]}], "]"}]}]], "Input",
 CellLabel->"In[47]:=",
 CellID->2082013545,ExpressionUUID->"d8ba2e27-4f24-4b7f-b131-0e90c2b75821"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"2", ",", "3", ",", "4"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"3", ",", "4", ",", "5"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"4", ",", "5", ",", "6"}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.7287523350223627`*^9, 3.728752699333867*^9, 
  3.728753087861059*^9, 3.728753171883588*^9, 3.730913242148531*^9},
 CellLabel->"Out[47]=",
 CellID->603516766,ExpressionUUID->"8b2dee4c-541a-453e-b0f8-5064fd079c35"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentValue", "[", 
  RowBox[{
   RowBox[{"ComponentArray", "[", 
    RowBox[{"T", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"a", ",", "polar"}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"b", ",", "polar"}], "}"}]}], "]"}], "]"}], ",", "values"}], 
  "]"}]], "Input",
 CellLabel->"In[48]:=",
 CellID->1274343673,ExpressionUUID->"dfd56df5-54bd-4848-9a6a-4c004b0b24ff"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, $CellContext`polar}, {0, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, $CellContext`polar}, {1, $CellContext`polar}],
       Editable->False], "\[Rule]", "3"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, $CellContext`polar}, {2, $CellContext`polar}],
       Editable->False], "\[Rule]", "4"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, $CellContext`polar}, {0, $CellContext`polar}],
       Editable->False], "\[Rule]", "3"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, $CellContext`polar}, {1, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, $CellContext`polar}, {2, $CellContext`polar}],
       Editable->False], "\[Rule]", "5"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, $CellContext`polar}, {0, $CellContext`polar}],
       Editable->False], "\[Rule]", "4"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, $CellContext`polar}, {1, $CellContext`polar}],
       Editable->False], "\[Rule]", "5"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, $CellContext`polar}, {2, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.7287523355397882`*^9, 3.72875269953244*^9, 
  3.728753088051127*^9, 3.728753172082469*^9, 3.730913242413907*^9},
 CellLabel->"Out[48]=",
 CellID->741021381,ExpressionUUID->"43b5caeb-1ee6-4804-a3f6-3f4af1c7fe67"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ColumnForm", "/@", 
  RowBox[{"TensorValues", "[", 
   RowBox[{"T", ",", 
    RowBox[{"{", 
     RowBox[{"{", 
      RowBox[{"polar", ",", "polar"}], "}"}], "}"}]}], "]"}]}]], "Input",
 CellLabel->"In[49]:=",
 CellID->173212421,ExpressionUUID->"56298b9f-d091-4841-bd4e-132adfa2c029"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   InterpretationBox[GridBox[{
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, $CellContext`polar}, {0, $CellContext`polar}],
         Editable->False], "\[Rule]", "0"}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, $CellContext`polar}, {0, $CellContext`polar}],
         Editable->False], "\[Rule]", 
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"T", 
              StyleBox[GridBox[{
                 {
                  StyleBox["0",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["1",
                   FontColor->RGBColor[1., 0., 0.]]},
                 {" ", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T[{0, $CellContext`polar}, {1, $CellContext`polar}],
          Editable->False]}]}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, $CellContext`polar}, {1, $CellContext`polar}],
         Editable->False], "\[Rule]", "0"}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{2, $CellContext`polar}, {0, $CellContext`polar}],
         Editable->False], "\[Rule]", 
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"T", 
              StyleBox[GridBox[{
                 {
                  StyleBox["0",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["2",
                   FontColor->RGBColor[1., 0., 0.]]},
                 {" ", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T[{0, $CellContext`polar}, {2, $CellContext`polar}],
          Editable->False]}]}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{2, $CellContext`polar}, {1, $CellContext`polar}],
         Editable->False], "\[Rule]", 
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"T", 
              StyleBox[GridBox[{
                 {
                  StyleBox["1",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["2",
                   FontColor->RGBColor[1., 0., 0.]]},
                 {" ", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T[{1, $CellContext`polar}, {2, $CellContext`polar}],
          Editable->False]}]}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{2, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False], "\[Rule]", "0"}]}
     },
     BaselinePosition->{Baseline, {1, 1}},
     GridBoxAlignment->{
      "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
       "RowsIndexed" -> {}}],
    ColumnForm[{$CellContext`T[{0, $CellContext`polar}, {
        0, $CellContext`polar}] -> 
      0, $CellContext`T[{1, $CellContext`polar}, {
        0, $CellContext`polar}] -> -$CellContext`T[{0, $CellContext`polar}, {
         1, $CellContext`polar}], $CellContext`T[{1, $CellContext`polar}, {
        1, $CellContext`polar}] -> 
      0, $CellContext`T[{2, $CellContext`polar}, {
        0, $CellContext`polar}] -> -$CellContext`T[{0, $CellContext`polar}, {
         2, $CellContext`polar}], $CellContext`T[{2, $CellContext`polar}, {
        1, $CellContext`polar}] -> -$CellContext`T[{1, $CellContext`polar}, {
         2, $CellContext`polar}], $CellContext`T[{2, $CellContext`polar}, {
        2, $CellContext`polar}] -> 0}],
    Editable->False], ",", 
   InterpretationBox[GridBox[{
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, $CellContext`polar}, {1, $CellContext`polar}],
         Editable->False], "\[Rule]", 
        RowBox[{"-", "3"}]}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False], "\[Rule]", 
        RowBox[{"-", "4"}]}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False], "\[Rule]", 
        RowBox[{"-", "5"}]}]}
     },
     BaselinePosition->{Baseline, {1, 1}},
     GridBoxAlignment->{
      "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
       "RowsIndexed" -> {}}],
    ColumnForm[{$CellContext`T[{0, $CellContext`polar}, {
        1, $CellContext`polar}] -> -3, $CellContext`T[{
        0, $CellContext`polar}, {
        2, $CellContext`polar}] -> -4, $CellContext`T[{
        1, $CellContext`polar}, {2, $CellContext`polar}] -> -5}],
    Editable->False]}], "]"}]], "Output",
 CellChangeTimes->{3.72875233570413*^9, 3.72875269971419*^9, 
  3.728753088224227*^9, 3.728753172260994*^9, 3.730913242782494*^9},
 CellLabel->"Out[49]=",
 CellID->1088681514,ExpressionUUID->"a6240618-adf7-43c2-9c28-4900298790f7"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["4.5. AllComponentValues", "Subsection",
 CellChangeTimes->{{3.72867395511727*^9, 3.728673957205443*^9}, {
  3.728675287623538*^9, 3.728675287767529*^9}},
 CellID->1583260063,ExpressionUUID->"3fdef841-975b-4a9d-b578-b710322b4adb"],

Cell[TextData[{
 StyleBox["AllComponentValues\t\t", "DisplayMath"],
 StyleBox["Generation of the list of all the rules for a given ValID, taking \
the different configurations into account", "Text"],
 StyleBox[" ", "DisplayMath"]
}], "DefinitionBox",
 CellID->382008864,ExpressionUUID->"e357dd9f-ef1c-4a0e-8dfb-4a1f7e8f9089"],

Cell["Generating all value rules.", "Text",
 CellChangeTimes->{3.7287518243037233`*^9},
 CellID->1594356365,ExpressionUUID->"caca1ce9-89a7-4b4d-a9ac-9c30eb072b26"],

Cell[CellGroupData[{

Cell[TextData[{
 "We have just described an easy way to generate all the rules for a ",
 StyleBox["ValID", "MR"],
 " which consisted on only one index configuration. But as we saw before, a \
single ",
 StyleBox["ComponentArray", "MR"],
 " will not generate all the independent rules for a ",
 StyleBox["ValID", "MR"],
 " with several index configurations. One such case was the ",
 StyleBox["ValID", "MR"],
 " for ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["T", 
     StyleBox[
      RowBox[{
       StyleBox["a",
        FontColor->RGBColor[1, 0, 0]], 
       StyleBox["b",
        FontColor->RGBColor[0, 1, 0]]}],
      FontSlant->"Italic"]], " ", "and", " ", 
    SubscriptBox["T", 
     StyleBox[
      RowBox[{
       StyleBox["b",
        FontColor->RGBColor[0, 1, 0]], 
       StyleBox["a",
        FontColor->RGBColor[1, 0, 0]]}],
      FontSlant->"Italic"]]}], TraditionalForm]],ExpressionUUID->
  "57150005-3b7b-45c7-9fb1-6e74d32c4a71"],
 ". The following command (which only works for tensors and not for \
derivatives as we shall see in Section 7.4.) applies ",
 StyleBox["ComponentArray", "MR"],
 " as many times as necessary in order to generate all the rules."
}], "Text",
 CellID->46479289,ExpressionUUID->"20c4f55d-460e-45be-91af-997b7e52697c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"values", "=", 
  RowBox[{"RandomInteger", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"0", ",", "5"}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{"3", ",", "3"}], "}"}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.730913256192206*^9, 3.730913270254148*^9}},
 CellLabel->"In[50]:=",
 CellID->1640846362,ExpressionUUID->"49e35c66-8780-47c9-868f-02ac42a4b7e0"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"5", ",", "5", ",", "3"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"3", ",", "3", ",", "3"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"2", ",", "2", ",", "0"}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728752341091782*^9, 3.7287527051782513`*^9, 
  3.728753088361156*^9, 3.728753176195808*^9, 3.73091324729808*^9, 
  3.730913281719982*^9},
 CellLabel->"Out[50]=",
 CellID->665679041,ExpressionUUID->"d5d03b7e-1381-4ddd-97fb-e465969ea1b1"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ColumnForm", "/@", " ", 
  RowBox[{"AllComponentValues", "[", 
   RowBox[{
    RowBox[{"T", "[", 
     RowBox[{
      RowBox[{"-", 
       RowBox[{"{", 
        RowBox[{"a", ",", "polar"}], "}"}]}], ",", 
      RowBox[{"-", 
       RowBox[{"{", 
        RowBox[{"b", ",", "cartesian"}], "}"}]}]}], "]"}], ",", "values"}], 
   "]"}]}]], "Input",
 CellLabel->"In[51]:=",
 CellID->301165391,ExpressionUUID->"5f571067-a16d-42bf-b0d8-db15231eda4e"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   InterpretationBox[GridBox[{
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[0., 1., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{2, -$CellContext`polar}, {
          1, -$CellContext`cartesian}],
         Editable->False], "\[Rule]", 
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"T", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["1",
                   FontColor->RGBColor[0., 1., 0.]], 
                  StyleBox["2",
                   FontColor->RGBColor[1., 0., 0.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T[{1, -$CellContext`cartesian}, {
           2, -$CellContext`polar}],
          Editable->False]}]}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[0., 1., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`polar}, {
          0, -$CellContext`cartesian}],
         Editable->False], "\[Rule]", 
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"T", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["0",
                   FontColor->RGBColor[0., 1., 0.]], 
                  StyleBox["0",
                   FontColor->RGBColor[1., 0., 0.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T[{0, -$CellContext`cartesian}, {
           0, -$CellContext`polar}],
          Editable->False]}]}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[0., 1., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, -$CellContext`polar}, {
          0, -$CellContext`cartesian}],
         Editable->False], "\[Rule]", 
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"T", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["0",
                   FontColor->RGBColor[0., 1., 0.]], 
                  StyleBox["1",
                   FontColor->RGBColor[1., 0., 0.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T[{0, -$CellContext`cartesian}, {
           1, -$CellContext`polar}],
          Editable->False]}]}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[0., 1., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, -$CellContext`polar}, {
          1, -$CellContext`cartesian}],
         Editable->False], "\[Rule]", 
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"T", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["1",
                   FontColor->RGBColor[0., 1., 0.]], 
                  StyleBox["1",
                   FontColor->RGBColor[1., 0., 0.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T[{1, -$CellContext`cartesian}, {
           1, -$CellContext`polar}],
          Editable->False]}]}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[0., 1., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{2, -$CellContext`polar}, {
          0, -$CellContext`cartesian}],
         Editable->False], "\[Rule]", 
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"T", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["0",
                   FontColor->RGBColor[0., 1., 0.]], 
                  StyleBox["2",
                   FontColor->RGBColor[1., 0., 0.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T[{0, -$CellContext`cartesian}, {
           2, -$CellContext`polar}],
          Editable->False]}]}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[0., 1., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{2, -$CellContext`polar}, {
          2, -$CellContext`cartesian}],
         Editable->False], "\[Rule]", 
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"T", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["2",
                   FontColor->RGBColor[0., 1., 0.]], 
                  StyleBox["2",
                   FontColor->RGBColor[1., 0., 0.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T[{2, -$CellContext`cartesian}, {
           2, -$CellContext`polar}],
          Editable->False]}]}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[0., 1., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, -$CellContext`cartesian}, {
          0, -$CellContext`polar}],
         Editable->False], "\[Rule]", 
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"T", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["0",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["1",
                   FontColor->RGBColor[0., 1., 0.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T[{0, -$CellContext`polar}, {
           1, -$CellContext`cartesian}],
          Editable->False]}]}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["2",
                  FontColor->RGBColor[0., 1., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{2, -$CellContext`cartesian}, {
          0, -$CellContext`polar}],
         Editable->False], "\[Rule]", 
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"T", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["0",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["2",
                   FontColor->RGBColor[0., 1., 0.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T[{0, -$CellContext`polar}, {
           2, -$CellContext`cartesian}],
          Editable->False]}]}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["2",
                  FontColor->RGBColor[0., 1., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{2, -$CellContext`cartesian}, {
          1, -$CellContext`polar}],
         Editable->False], "\[Rule]", 
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"T", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["1",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["2",
                   FontColor->RGBColor[0., 1., 0.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T[{1, -$CellContext`polar}, {
           2, -$CellContext`cartesian}],
          Editable->False]}]}]}
     },
     BaselinePosition->{Baseline, {1, 1}},
     GridBoxAlignment->{
      "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
       "RowsIndexed" -> {}}],
    ColumnForm[{$CellContext`T[{2, -$CellContext`polar}, {
        1, -$CellContext`cartesian}] -> -$CellContext`T[{
         1, -$CellContext`cartesian}, {
         2, -$CellContext`polar}], $CellContext`T[{0, -$CellContext`polar}, {
        0, -$CellContext`cartesian}] -> -$CellContext`T[{
         0, -$CellContext`cartesian}, {
         0, -$CellContext`polar}], $CellContext`T[{1, -$CellContext`polar}, {
        0, -$CellContext`cartesian}] -> -$CellContext`T[{
         0, -$CellContext`cartesian}, {
         1, -$CellContext`polar}], $CellContext`T[{1, -$CellContext`polar}, {
        1, -$CellContext`cartesian}] -> -$CellContext`T[{
         1, -$CellContext`cartesian}, {
         1, -$CellContext`polar}], $CellContext`T[{2, -$CellContext`polar}, {
        0, -$CellContext`cartesian}] -> -$CellContext`T[{
         0, -$CellContext`cartesian}, {
         2, -$CellContext`polar}], $CellContext`T[{2, -$CellContext`polar}, {
        2, -$CellContext`cartesian}] -> -$CellContext`T[{
         2, -$CellContext`cartesian}, {
         2, -$CellContext`polar}], $CellContext`T[{
        1, -$CellContext`cartesian}, {
        0, -$CellContext`polar}] -> -$CellContext`T[{
         0, -$CellContext`polar}, {
         1, -$CellContext`cartesian}], $CellContext`T[{
        2, -$CellContext`cartesian}, {
        0, -$CellContext`polar}] -> -$CellContext`T[{
         0, -$CellContext`polar}, {
         2, -$CellContext`cartesian}], $CellContext`T[{
        2, -$CellContext`cartesian}, {
        1, -$CellContext`polar}] -> -$CellContext`T[{
         1, -$CellContext`polar}, {2, -$CellContext`cartesian}]}],
    Editable->False], ",", 
   InterpretationBox[GridBox[{
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[0., 1., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`polar}, {
          1, -$CellContext`cartesian}],
         Editable->False], "\[Rule]", "5"}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[0., 1., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`polar}, {
          2, -$CellContext`cartesian}],
         Editable->False], "\[Rule]", "3"}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[0., 1., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, -$CellContext`polar}, {
          2, -$CellContext`cartesian}],
         Editable->False], "\[Rule]", "3"}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[0., 1., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`cartesian}, {
          0, -$CellContext`polar}],
         Editable->False], "\[Rule]", 
        RowBox[{"-", "5"}]}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[0., 1., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`cartesian}, {
          1, -$CellContext`polar}],
         Editable->False], "\[Rule]", 
        RowBox[{"-", "3"}]}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[0., 1., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, -$CellContext`cartesian}, {
          2, -$CellContext`polar}],
         Editable->False], "\[Rule]", 
        RowBox[{"-", "2"}]}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[0., 1., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, -$CellContext`cartesian}, {
          1, -$CellContext`polar}],
         Editable->False], "\[Rule]", 
        RowBox[{"-", "3"}]}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["2",
                  FontColor->RGBColor[0., 1., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{2, -$CellContext`cartesian}, {
          2, -$CellContext`polar}],
         Editable->False], "\[Rule]", "0"}]},
      {
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[0., 1., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, -$CellContext`cartesian}, {
          2, -$CellContext`polar}],
         Editable->False], "\[Rule]", 
        RowBox[{"-", "2"}]}]}
     },
     BaselinePosition->{Baseline, {1, 1}},
     GridBoxAlignment->{
      "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
       "RowsIndexed" -> {}}],
    ColumnForm[{$CellContext`T[{0, -$CellContext`polar}, {
        1, -$CellContext`cartesian}] -> 
      5, $CellContext`T[{0, -$CellContext`polar}, {
        2, -$CellContext`cartesian}] -> 
      3, $CellContext`T[{1, -$CellContext`polar}, {
        2, -$CellContext`cartesian}] -> 
      3, $CellContext`T[{0, -$CellContext`cartesian}, {
        0, -$CellContext`polar}] -> -5, $CellContext`T[{
        0, -$CellContext`cartesian}, {
        1, -$CellContext`polar}] -> -3, $CellContext`T[{
        0, -$CellContext`cartesian}, {
        2, -$CellContext`polar}] -> -2, $CellContext`T[{
        1, -$CellContext`cartesian}, {
        1, -$CellContext`polar}] -> -3, $CellContext`T[{
        2, -$CellContext`cartesian}, {2, -$CellContext`polar}] -> 
      0, $CellContext`T[{1, -$CellContext`cartesian}, {
        2, -$CellContext`polar}] -> -2}],
    Editable->False]}], "]"}]], "Output",
 CellChangeTimes->{3.728752341321649*^9, 3.728752705415299*^9, 
  3.72875308857108*^9, 3.7287531764139433`*^9, 3.730913247441519*^9, 
  3.7309132829664173`*^9},
 CellLabel->"Out[51]=",
 CellID->940372741,ExpressionUUID->"ab3dcc0d-a2c6-4662-b4fe-d66fd176de7c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We turn the messages back on:", "Text",
 CellID->853435396,ExpressionUUID->"1458d9cc-a4c3-40b7-b878-e7994b0aaa37"],

Cell[BoxData[
 RowBox[{
  RowBox[{"$CVVerbose", "=", "True"}], ";"}]], "Input",
 CellLabel->"In[52]:=",
 CellID->645422259,ExpressionUUID->"0fedbdf4-f638-47f3-96d1-4a70d15ee8f4"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["4.6. Components and derivatives", "Subsection",
 CellChangeTimes->{{3.7286739583093576`*^9, 3.7286739598774347`*^9}, {
  3.728675288647456*^9, 3.728675288799488*^9}},
 CellID->672006016,ExpressionUUID->"6fd6bc1d-d135-46c8-9fd9-3d5ce2aa6fcb"],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["ComponentValue", "MR"],
 " also works with derivatives of tensors (but not for more general \
expressions, you can use ",
 StyleBox["Coba", "MR"],
 " as a placeholder and then substitute the expression). Remember that with \
our definitions the components of the derivative are not the derivatives of \
the components."
}], "Text",
 CellID->788611339,ExpressionUUID->"2820ac29-660f-4765-9a93-974137f76716"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"{", 
    RowBox[{"2", ",", 
     RowBox[{"-", "cartesian"}]}], "}"}], "]"}], "[", 
  RowBox[{"v", "[", 
   RowBox[{"{", 
    RowBox[{"1", ",", "cartesian"}], "}"}], "]"}], "]"}]], "Input",
 CellLabel->"In[53]:=",
 CellID->2069188936,ExpressionUUID->"ef0bfd45-0d0f-40e8-bef0-15e2ebade0e8"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[PartialD]", 
     StyleBox["2",
      FontColor->RGBColor[0., 1., 0.]]], GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {
           StyleBox["1",
            FontColor->RGBColor[0., 1., 0.]]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`PD[{2, -$CellContext`cartesian}][
   $CellContext`v[{1, $CellContext`cartesian}]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728752345261511*^9, 3.728752708729188*^9, 
  3.7287530887759666`*^9, 3.7287531788976793`*^9, 3.73091330606384*^9},
 CellLabel->"Out[53]=",
 CellID->1500896426,ExpressionUUID->"8df95361-3e92-4661-9d03-0c244dfa7db6"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentValue", "[", "%", "]"}]], "Input",
 CellLabel->"In[54]:=",
 CellID->1915021919,ExpressionUUID->"15fb2b1d-9a65-44ef-9ae6-b477981bb63b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[PartialD]", 
        StyleBox["2",
         FontColor->RGBColor[0., 1., 0.]]], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 1., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`PD[{2, -$CellContext`cartesian}][
      $CellContext`v[{1, $CellContext`cartesian}]],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[PartialD]", 
        StyleBox["2",
         FontColor->RGBColor[0., 1., 0.]]], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 1., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`PD[{2, -$CellContext`cartesian}][
      $CellContext`v[{1, $CellContext`cartesian}]],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "v"}],
  SequenceForm[
  "Added independent rule ", xAct`xTensor`PD[{2, -$CellContext`cartesian}][
     $CellContext`v[{1, $CellContext`cartesian}]] -> 
   xAct`xTensor`PD[{2, -$CellContext`cartesian}][
     $CellContext`v[{1, $CellContext`cartesian}]], 
   " for tensor ", $CellContext`v],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752345437352*^9, 3.728752708933154*^9, 
  3.728753088912881*^9, 3.728753179047255*^9, 3.730913306188192*^9},
 CellLabel->"During evaluation of In[54]:=",
 CellID->1223314120,ExpressionUUID->"5769b9a1-2702-4a56-b8f3-873a38155052"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", 
      StyleBox["2",
       FontColor->RGBColor[0., 1., 0.]]], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[0., 1., 0.]]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[{2, -$CellContext`cartesian}][
    $CellContext`v[{1, $CellContext`cartesian}]],
   Editable->False], "\[Rule]", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", 
      StyleBox["2",
       FontColor->RGBColor[0., 1., 0.]]], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[0., 1., 0.]]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[{2, -$CellContext`cartesian}][
    $CellContext`v[{1, $CellContext`cartesian}]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7287523455234528`*^9, 3.728752709015527*^9, 
  3.7287530889419937`*^9, 3.728753179132832*^9, 3.730913306237533*^9},
 CellLabel->"Out[54]=",
 CellID->231978462,ExpressionUUID->"560c1f93-f269-4217-81c4-63193e22f9aa"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"{", 
    RowBox[{"1", ",", 
     RowBox[{"-", "polar"}]}], "}"}], "]"}], "[", 
  RowBox[{
   RowBox[{"PDpolar", "[", 
    RowBox[{"{", 
     RowBox[{"2", ",", 
      RowBox[{"-", "cartesian"}]}], "}"}], "]"}], "[", 
   RowBox[{"v", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "cartesian"}], "}"}], "]"}], "]"}], "]"}]], "Input",
 CellLabel->"In[55]:=",
 CellID->250280249,ExpressionUUID->"daac4602-9b13-47a3-92bc-9979edfd4d5f"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[PartialD]", 
     StyleBox["1",
      FontColor->RGBColor[1., 0., 0.]]], 
    SubscriptBox[
     StyleBox["\[ScriptCapitalD]",
      FontColor->RGBColor[1., 0., 0.]], 
     StyleBox["2",
      FontColor->RGBColor[0., 1., 0.]]], GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {
           StyleBox["1",
            FontColor->RGBColor[0., 1., 0.]]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`PD[{1, -$CellContext`polar}][
   $CellContext`PDpolar[{2, -$CellContext`cartesian}][
    $CellContext`v[{1, $CellContext`cartesian}]]],
  Editable->False]], "Output",
 CellChangeTimes->{3.7287523457044363`*^9, 3.728752709216308*^9, 
  3.7287530890628357`*^9, 3.728753179328436*^9, 3.7309133063559017`*^9},
 CellLabel->"Out[55]=",
 CellID->1179191529,ExpressionUUID->"4dd977b0-227e-400f-b38f-bb765ae01c09"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentValue", "[", "%", "]"}]], "Input",
 CellLabel->"In[56]:=",
 CellID->793239431,ExpressionUUID->"8aa3d5c1-d409-47c8-9709-5be22fc2ff90"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[PartialD]", 
        StyleBox["1",
         FontColor->RGBColor[1., 0., 0.]]], 
       SubscriptBox[
        StyleBox["\[ScriptCapitalD]",
         FontColor->RGBColor[1., 0., 0.]], 
        StyleBox["2",
         FontColor->RGBColor[0., 1., 0.]]], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 1., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`PD[{1, -$CellContext`polar}][
      $CellContext`PDpolar[{2, -$CellContext`cartesian}][
       $CellContext`v[{1, $CellContext`cartesian}]]],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[PartialD]", 
        StyleBox["1",
         FontColor->RGBColor[1., 0., 0.]]], 
       SubscriptBox[
        StyleBox["\[ScriptCapitalD]",
         FontColor->RGBColor[1., 0., 0.]], 
        StyleBox["2",
         FontColor->RGBColor[0., 1., 0.]]], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 1., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`PD[{1, -$CellContext`polar}][
      $CellContext`PDpolar[{2, -$CellContext`cartesian}][
       $CellContext`v[{1, $CellContext`cartesian}]]],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "v"}],
  SequenceForm[
  "Added independent rule ", xAct`xTensor`PD[{1, -$CellContext`polar}][
     $CellContext`PDpolar[{2, -$CellContext`cartesian}][
      $CellContext`v[{1, $CellContext`cartesian}]]] -> 
   xAct`xTensor`PD[{1, -$CellContext`polar}][
     $CellContext`PDpolar[{2, -$CellContext`cartesian}][
      $CellContext`v[{1, $CellContext`cartesian}]]], 
   " for tensor ", $CellContext`v],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752345906848*^9, 3.728752709395109*^9, 
  3.7287530892254333`*^9, 3.728753179532754*^9, 3.730913306458996*^9},
 CellLabel->"During evaluation of In[56]:=",
 CellID->1692933387,ExpressionUUID->"aefaaf7a-826a-44c8-896d-124b88f14e0d"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", 
      StyleBox["1",
       FontColor->RGBColor[1., 0., 0.]]], 
     SubscriptBox[
      StyleBox["\[ScriptCapitalD]",
       FontColor->RGBColor[1., 0., 0.]], 
      StyleBox["2",
       FontColor->RGBColor[0., 1., 0.]]], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[0., 1., 0.]]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[{1, -$CellContext`polar}][
    $CellContext`PDpolar[{2, -$CellContext`cartesian}][
     $CellContext`v[{1, $CellContext`cartesian}]]],
   Editable->False], "\[Rule]", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", 
      StyleBox["1",
       FontColor->RGBColor[1., 0., 0.]]], 
     SubscriptBox[
      StyleBox["\[ScriptCapitalD]",
       FontColor->RGBColor[1., 0., 0.]], 
      StyleBox["2",
       FontColor->RGBColor[0., 1., 0.]]], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[0., 1., 0.]]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[{1, -$CellContext`polar}][
    $CellContext`PDpolar[{2, -$CellContext`cartesian}][
     $CellContext`v[{1, $CellContext`cartesian}]]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728752345973772*^9, 3.7287527094658957`*^9, 
  3.7287530892468157`*^9, 3.72875317959892*^9, 3.7309133065050163`*^9},
 CellLabel->"Out[56]=",
 CellID->1120596206,ExpressionUUID->"45cc6bbe-47f2-4239-abbe-626854bdc8cc"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Derivatives are represented by  additional central arguments in ",
 StyleBox["ValID", "MR"],
 ":"
}], "Text",
 CellID->1800793518,ExpressionUUID->"3b4b6709-01b0-43f4-98be-bf8d198f7443"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"TensorValIDs", "[", "v", "]"}], "//", "ColumnForm"}]], "Input",
 CellLabel->"In[57]:=",
 CellID->1449619855,ExpressionUUID->"42204b37-2f7c-4360-8cb9-350297425c11"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {
     RowBox[{"ValID", "[", 
      RowBox[{"v", ",", 
       RowBox[{"PDpolar", "[", 
        RowBox[{"-", "cartesian"}], "]"}], ",", 
       RowBox[{"PD", "[", 
        RowBox[{"-", "polar"}], "]"}], ",", 
       RowBox[{"{", 
        RowBox[{"{", 
         RowBox[{"cartesian", ",", 
          RowBox[{"-", "cartesian"}], ",", 
          RowBox[{"-", "polar"}]}], "}"}], "}"}]}], "]"}]},
    {
     RowBox[{"ValID", "[", 
      RowBox[{"v", ",", 
       RowBox[{"PD", "[", 
        RowBox[{"-", "cartesian"}], "]"}], ",", 
       RowBox[{"{", 
        RowBox[{"{", 
         RowBox[{"cartesian", ",", 
          RowBox[{"-", "cartesian"}]}], "}"}], "}"}]}], "]"}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  ColumnForm[{
    xAct`xCoba`ValID[$CellContext`v, 
     $CellContext`PDpolar[-$CellContext`cartesian], 
     xAct`xTensor`PD[-$CellContext`polar], {{$CellContext`cartesian, \
-$CellContext`cartesian, -$CellContext`polar}}], 
    xAct`xCoba`ValID[$CellContext`v, 
     xAct`xTensor`PD[-$CellContext`cartesian], {{$CellContext`cartesian, \
-$CellContext`cartesian}}]}],
  Editable->False]], "Output",
 CellChangeTimes->{3.728752346138287*^9, 3.728752709661289*^9, 
  3.728753089375183*^9, 3.728753179761516*^9, 3.730913306593102*^9},
 CellLabel->"Out[57]=",
 CellID->428850052,ExpressionUUID->"2fb99f5f-1004-4823-9b8b-6becae5afff7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DeleteTensorValues", "[", "v", "]"}]], "Input",
 CellLabel->"In[58]:=",
 CellID->805558412,ExpressionUUID->"f6431eb0-869a-40df-b6f2-3d61c6df8ca1"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Deleted values for tensor \"\>", "\[InvisibleSpace]", "v", 
   "\[InvisibleSpace]", "\<\", derivatives \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"PDpolar", "[", 
      RowBox[{"-", "cartesian"}], "]"}], ",", 
     RowBox[{"PD", "[", 
      RowBox[{"-", "polar"}], "]"}]}], "}"}], 
   "\[InvisibleSpace]", "\<\" and bases \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"{", 
     RowBox[{"cartesian", ",", 
      RowBox[{"-", "cartesian"}], ",", 
      RowBox[{"-", "polar"}]}], "}"}], "}"}], 
   "\[InvisibleSpace]", "\<\".\"\>"}],
  SequenceForm[
  "Deleted values for tensor ", $CellContext`v, ", derivatives ", {
    $CellContext`PDpolar[-$CellContext`cartesian], 
    xAct`xTensor`PD[-$CellContext`polar]}, 
   " and bases ", {{$CellContext`cartesian, -$CellContext`cartesian, \
-$CellContext`polar}}, "."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752346290156*^9, 3.728752709822103*^9, 
  3.728753089525257*^9, 3.728753179920081*^9, 3.730913306728994*^9},
 CellLabel->"During evaluation of In[58]:=",
 CellID->727704632,ExpressionUUID->"f6c4b5d4-d13d-4ebd-83ac-34254ef66040"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Deleted values for tensor \"\>", "\[InvisibleSpace]", "v", 
   "\[InvisibleSpace]", "\<\", derivatives \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"PD", "[", 
     RowBox[{"-", "cartesian"}], "]"}], "}"}], 
   "\[InvisibleSpace]", "\<\" and bases \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"{", 
     RowBox[{"cartesian", ",", 
      RowBox[{"-", "cartesian"}]}], "}"}], "}"}], 
   "\[InvisibleSpace]", "\<\".\"\>"}],
  SequenceForm[
  "Deleted values for tensor ", $CellContext`v, ", derivatives ", {
    xAct`xTensor`PD[-$CellContext`cartesian]}, 
   " and bases ", {{$CellContext`cartesian, -$CellContext`cartesian}}, "."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752346290156*^9, 3.728752709822103*^9, 
  3.728753089525257*^9, 3.728753179920081*^9, 3.730913306744897*^9},
 CellLabel->"During evaluation of In[58]:=",
 CellID->612380359,ExpressionUUID->"23f3f05e-7454-4bfe-ba86-3e3faeaeea67"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["4.7. BasisValues", "Subsection",
 CellChangeTimes->{{3.728673960981289*^9, 3.728673962557238*^9}, {
  3.7286752897672043`*^9, 3.728675289951446*^9}},
 CellID->958162795,ExpressionUUID->"f240955e-7d0e-482c-aec9-aeb322a0d79b"],

Cell[TextData[{
 StyleBox["BasisValues\t\t", "DisplayMath"],
 StyleBox["Values for the components of basis changes", "Text"],
 StyleBox[" ", "DisplayMath"]
}], "DefinitionBox",
 CellID->2009710359,ExpressionUUID->"e4d775ac-cfa8-4251-9c57-bfca6dca0753"],

Cell["Components of the basis changes.", "Text",
 CellChangeTimes->{3.7287518325993767`*^9},
 CellID->1579203697,ExpressionUUID->"754f4847-1dd6-4ee2-8c03-a540510b3728"],

Cell[CellGroupData[{

Cell[TextData[{
 "The components of the basis change matrices require separate treatment and \
are represented by the ",
 StyleBox["BasisValues", "MR"],
 " head. Here we do not allow derivatives, because the derivatives of these \
objects are automatically replaced by Christoffel tensors."
}], "Text",
 CellID->926142206,ExpressionUUID->"01d395e1-be21-4c2a-8b6e-6c73b2a49392"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "BasisValues"}]], "Input",
 CellLabel->"In[59]:=",
 CellID->710582786,ExpressionUUID->"3036805d-59cb-4d54-916a-ac172c7254d5"],

Cell[BoxData["\<\"BasisValues[basis1, -basis2] and BasisValues[-basis2, \
basis1] give the list of component values for the matrix of change between \
those two bases. There is no automatic computation of the inverse \
matrix.\"\>"], "Print", "PrintUsage",
 CellChangeTimes->{3.7309133098529997`*^9},
 CellTags->"Info593730895309-7554244",
 CellID->1005248081,ExpressionUUID->"24c2fc93-e784-4bb3-aaff-e4997b69698a"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"$CVVerbose", "=", "True"}], ";"}]], "Input",
 CellLabel->"In[60]:=",
 CellID->2068131625,ExpressionUUID->"641bdb87-61c8-47cc-8dfc-fcd75d47e46e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentValue", "[", 
  RowBox[{"Basis", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"1", ",", 
      RowBox[{"-", "polar"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{"2", ",", "cartesian"}], "}"}]}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.728752363951956*^9, 3.728752365838929*^9}},
 CellLabel->"In[61]:=",
 CellID->2065394680,ExpressionUUID->"0527cb35-1e85-4901-a6e9-bdc853eb7ac2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`polar}, {2, $CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`polar}, {2, $CellContext`cartesian}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{1, -$CellContext`polar}, {2, $CellContext`cartesian}] -> 
   xAct`xTensor`Basis[{1, -$CellContext`polar}, {2, $CellContext`cartesian}], 
   " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752366644792*^9, 3.7287527131786346`*^9, 
  3.7287530900569677`*^9, 3.728753182896132*^9, 3.7309133100562143`*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->950646418,ExpressionUUID->"c54f5c9f-5ad4-4216-922c-e2552ea2f6fe"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["2",
            FontColor->RGBColor[0., 1., 0.]]},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{1, -$CellContext`polar}, {2, $CellContext`cartesian}],
   Editable->False], "\[Rule]", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"e", 
       StyleBox[GridBox[{
          {" ", 
           StyleBox["2",
            FontColor->RGBColor[0., 1., 0.]]},
          {
           StyleBox["1",
            FontColor->RGBColor[1., 0., 0.]], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Basis[{1, -$CellContext`polar}, {2, $CellContext`cartesian}],
   Editable->False]}]], "Output",
 CellChangeTimes->{{3.728752349928907*^9, 3.728752366727981*^9}, 
   3.728752713266439*^9, 3.728753090075593*^9, 3.728753183033451*^9, 
   3.730913310107793*^9},
 CellLabel->"Out[61]=",
 CellID->236013646,ExpressionUUID->"d536c68d-188d-44d3-81a6-188680bef451"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AllComponentValues", "[", 
  RowBox[{
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "polar"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"b", ",", "cartesian"}], "}"}]}], "]"}], ",", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"Random", "[", "]"}], ",", 
     RowBox[{"{", "3", "}"}], ",", 
     RowBox[{"{", "3", "}"}]}], "]"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.728752370975893*^9, 3.72875237526265*^9}},
 CellLabel->"In[62]:=",
 CellID->1377197476,ExpressionUUID->"160d7db7-c30d-433b-a0e5-c08ac8061ab4"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]]},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`polar}, {0, $CellContext`cartesian}],
     Editable->False], "\[Rule]", "0.762651722655143`"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{0, -$CellContext`polar}, {0, $CellContext`cartesian}] -> 
   0.762651722655143, " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752375891617*^9, 3.728752713801128*^9, 
  3.728753090213911*^9, 3.7287531836674347`*^9, 3.730913310256626*^9},
 CellLabel->"During evaluation of In[62]:=",
 CellID->279052313,ExpressionUUID->"91f6cb09-d043-491e-b73a-5cedc3422967"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]]},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`polar}, {1, $CellContext`cartesian}],
     Editable->False], "\[Rule]", "0.3802645436601525`"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{0, -$CellContext`polar}, {1, $CellContext`cartesian}] -> 
   0.3802645436601525, " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752375891617*^9, 3.728752713801128*^9, 
  3.728753090213911*^9, 3.7287531836674347`*^9, 3.7309133102931423`*^9},
 CellLabel->"During evaluation of In[62]:=",
 CellID->2075423190,ExpressionUUID->"8a26fe8d-d341-4f39-b170-2e7fc33e3138"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]]},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`polar}, {2, $CellContext`cartesian}],
     Editable->False], "\[Rule]", "0.18745379107937912`"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{0, -$CellContext`polar}, {2, $CellContext`cartesian}] -> 
   0.18745379107937912`, " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752375891617*^9, 3.728752713801128*^9, 
  3.728753090213911*^9, 3.7287531836674347`*^9, 3.730913310361738*^9},
 CellLabel->"During evaluation of In[62]:=",
 CellID->60676109,ExpressionUUID->"d50a1931-94d1-4346-95f8-6b4947ccf8a2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`polar}, {0, $CellContext`cartesian}],
     Editable->False], "\[Rule]", "0.876479881154559`"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{1, -$CellContext`polar}, {0, $CellContext`cartesian}] -> 
   0.876479881154559, " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752375891617*^9, 3.728752713801128*^9, 
  3.728753090213911*^9, 3.7287531836674347`*^9, 3.730913310426976*^9},
 CellLabel->"During evaluation of In[62]:=",
 CellID->1876115285,ExpressionUUID->"64b34971-16ef-4ea4-9f49-976b7893718f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`polar}, {1, $CellContext`cartesian}],
     Editable->False], "\[Rule]", "0.02262175371824628`"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{1, -$CellContext`polar}, {1, $CellContext`cartesian}] -> 
   0.02262175371824628, " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752375891617*^9, 3.728752713801128*^9, 
  3.728753090213911*^9, 3.7287531836674347`*^9, 3.730913310494492*^9},
 CellLabel->"During evaluation of In[62]:=",
 CellID->367206094,ExpressionUUID->"f45abf14-47c3-4e38-9d80-90bf652a4f09"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent value \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`polar}, {2, $CellContext`cartesian}],
     Editable->False], "\[Rule]", "0.49550564865844027`"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent value ", 
   xAct`xTensor`Basis[{1, -$CellContext`polar}, {2, $CellContext`cartesian}] -> 
   0.49550564865844027`, " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752375891617*^9, 3.728752713801128*^9, 
  3.728753090213911*^9, 3.7287531836674347`*^9, 3.730913310561825*^9},
 CellLabel->"During evaluation of In[62]:=",
 CellID->1912304027,ExpressionUUID->"35b61a2d-e620-45c7-9fe7-72b6ea122930"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`polar}, {0, $CellContext`cartesian}],
     Editable->False], "\[Rule]", "0.8432436973888893`"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{2, -$CellContext`polar}, {0, $CellContext`cartesian}] -> 
   0.8432436973888893, " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752375891617*^9, 3.728752713801128*^9, 
  3.728753090213911*^9, 3.7287531836674347`*^9, 3.730913310627737*^9},
 CellLabel->"During evaluation of In[62]:=",
 CellID->853520787,ExpressionUUID->"efefcc0d-69c7-458f-af29-669487c82e27"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`polar}, {1, $CellContext`cartesian}],
     Editable->False], "\[Rule]", "0.8685328357541247`"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{2, -$CellContext`polar}, {1, $CellContext`cartesian}] -> 
   0.8685328357541247, " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752375891617*^9, 3.728752713801128*^9, 
  3.728753090213911*^9, 3.7287531836674347`*^9, 3.7309133106947603`*^9},
 CellLabel->"During evaluation of In[62]:=",
 CellID->2080588796,ExpressionUUID->"201f056d-5a77-4fcf-a8ad-ded5b0712338"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`polar}, {2, $CellContext`cartesian}],
     Editable->False], "\[Rule]", "0.5714435070376495`"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Basis"}],
  SequenceForm[
  "Added independent rule ", 
   xAct`xTensor`Basis[{2, -$CellContext`polar}, {2, $CellContext`cartesian}] -> 
   0.5714435070376495, " for tensor ", xAct`xTensor`Basis],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752375891617*^9, 3.728752713801128*^9, 
  3.728753090213911*^9, 3.7287531836674347`*^9, 3.730913310761847*^9},
 CellLabel->"During evaluation of In[62]:=",
 CellID->979670271,ExpressionUUID->"5b999428-062f-41b3-b810-83109e3e8af6"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{1, -$CellContext`polar}, {
        2, $CellContext`cartesian}],
       Editable->False], "\[Rule]", "0.49550564865844027`"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{0, -$CellContext`polar}, {
        0, $CellContext`cartesian}],
       Editable->False], "\[Rule]", "0.762651722655143`"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{0, -$CellContext`polar}, {
        1, $CellContext`cartesian}],
       Editable->False], "\[Rule]", "0.3802645436601525`"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{0, -$CellContext`polar}, {
        2, $CellContext`cartesian}],
       Editable->False], "\[Rule]", "0.18745379107937912`"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{1, -$CellContext`polar}, {
        0, $CellContext`cartesian}],
       Editable->False], "\[Rule]", "0.876479881154559`"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{1, -$CellContext`polar}, {
        1, $CellContext`cartesian}],
       Editable->False], "\[Rule]", "0.02262175371824628`"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{2, -$CellContext`polar}, {
        0, $CellContext`cartesian}],
       Editable->False], "\[Rule]", "0.8432436973888893`"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{2, -$CellContext`polar}, {
        1, $CellContext`cartesian}],
       Editable->False], "\[Rule]", "0.8685328357541247`"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{2, -$CellContext`polar}, {
        2, $CellContext`cartesian}],
       Editable->False], "\[Rule]", "0.5714435070376495`"}]}], "}"}]}], 
  "]"}]], "Output",
 CellChangeTimes->{{3.7287523501451597`*^9, 3.7287523766299257`*^9}, 
   3.72875271470156*^9, 3.728753090807899*^9, 3.728753184517663*^9, 
   3.73091331082251*^9},
 CellLabel->"Out[62]=",
 CellID->561123148,ExpressionUUID->"e09b7de6-d6c9-488e-80ce-e225a4e0aa2d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"BasisValues", "[", 
  RowBox[{
   RowBox[{"-", "polar"}], ",", "cartesian"}], "]"}]], "Input",
 CellChangeTimes->{{3.728752387501363*^9, 3.728752389389709*^9}},
 CellLabel->"In[63]:=",
 CellID->343148269,ExpressionUUID->"18f09ca9-5167-4e95-86f8-8e94d6f1ccb9"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{1, -$CellContext`polar}, {
        2, $CellContext`cartesian}],
       Editable->False], "\[Rule]", "0.49550564865844027`"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{0, -$CellContext`polar}, {
        0, $CellContext`cartesian}],
       Editable->False], "\[Rule]", "0.762651722655143`"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{0, -$CellContext`polar}, {
        1, $CellContext`cartesian}],
       Editable->False], "\[Rule]", "0.3802645436601525`"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{0, -$CellContext`polar}, {
        2, $CellContext`cartesian}],
       Editable->False], "\[Rule]", "0.18745379107937912`"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{1, -$CellContext`polar}, {
        0, $CellContext`cartesian}],
       Editable->False], "\[Rule]", "0.876479881154559`"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{1, -$CellContext`polar}, {
        1, $CellContext`cartesian}],
       Editable->False], "\[Rule]", "0.02262175371824628`"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{2, -$CellContext`polar}, {
        0, $CellContext`cartesian}],
       Editable->False], "\[Rule]", "0.8432436973888893`"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{2, -$CellContext`polar}, {
        1, $CellContext`cartesian}],
       Editable->False], "\[Rule]", "0.8685328357541247`"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{2, -$CellContext`polar}, {
        2, $CellContext`cartesian}],
       Editable->False], "\[Rule]", "0.5714435070376495`"}]}], "}"}]}], 
  "]"}]], "Output",
 CellChangeTimes->{{3.7287523505013*^9, 3.728752389832226*^9}, 
   3.728752714900261*^9, 3.728753090953401*^9, 3.728753184767418*^9, 
   3.730913310949575*^9},
 CellLabel->"Out[63]=",
 CellID->1196054459,ExpressionUUID->"f62a4166-b33c-4b7d-9a87-0d163e8c4859"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The inverse basis change is not automatically computed, because it may take a \
very long time\
\>", "Text",
 CellID->664466273,ExpressionUUID->"ba410e32-c41c-4497-8cf4-8eb66fa8b999"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"BasisValues", "[", 
  RowBox[{"cartesian", ",", 
   RowBox[{"-", "polar"}]}], "]"}]], "Input",
 CellLabel->"In[64]:=",
 CellID->769685911,ExpressionUUID->"39348cbe-d002-4e7c-9935-680f20679912"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{1, -$CellContext`polar}, {
        2, $CellContext`cartesian}],
       Editable->False], "\[Rule]", "0.49550564865844027`"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{0, -$CellContext`polar}, {
        0, $CellContext`cartesian}],
       Editable->False], "\[Rule]", "0.762651722655143`"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{0, -$CellContext`polar}, {
        1, $CellContext`cartesian}],
       Editable->False], "\[Rule]", "0.3802645436601525`"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{0, -$CellContext`polar}, {
        2, $CellContext`cartesian}],
       Editable->False], "\[Rule]", "0.18745379107937912`"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{1, -$CellContext`polar}, {
        0, $CellContext`cartesian}],
       Editable->False], "\[Rule]", "0.876479881154559`"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{1, -$CellContext`polar}, {
        1, $CellContext`cartesian}],
       Editable->False], "\[Rule]", "0.02262175371824628`"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{2, -$CellContext`polar}, {
        0, $CellContext`cartesian}],
       Editable->False], "\[Rule]", "0.8432436973888893`"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{2, -$CellContext`polar}, {
        1, $CellContext`cartesian}],
       Editable->False], "\[Rule]", "0.8685328357541247`"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{2, -$CellContext`polar}, {
        2, $CellContext`cartesian}],
       Editable->False], "\[Rule]", "0.5714435070376495`"}]}], "}"}]}], 
  "]"}]], "Output",
 CellChangeTimes->{{3.7287523507211313`*^9, 3.7287523806274967`*^9}, 
   3.7287527150832653`*^9, 3.728753091126257*^9, 3.728753184963105*^9, 
   3.7309133110966177`*^9},
 CellLabel->"Out[64]=",
 CellID->1640356493,ExpressionUUID->"7fc3a84b-427c-403c-b5a7-210b2250e832"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["4.8. Components of the metric, MetricInBasis", "Subsection",
 CellChangeTimes->{{3.728673963645653*^9, 3.7286739650851603`*^9}, {
  3.728675290855358*^9, 3.728675291047469*^9}},
 CellID->639730885,ExpressionUUID->"f64727aa-5c80-4f73-b728-49ecb2185f2b"],

Cell[TextData[{
 StyleBox["MetricInBasis\t\t", "DisplayMath"],
 StyleBox["Values for the components of the metric", "Text"]
}], "DefinitionBox",
 CellID->1653847313,ExpressionUUID->"fcd44251-b238-4cd0-b8d0-1b3fd17d75cd"],

Cell["Components of the metric.", "Text",
 CellChangeTimes->{3.72875183805508*^9},
 CellID->1330946470,ExpressionUUID->"88b92c65-80e2-4ca1-a848-b550cbc336a1"],

Cell[CellGroupData[{

Cell["\<\
It is very common in General Relativity to start working from a known line \
element. Inded, the typical approach is to start from the components of a \
metric in a coordinated basis, which determine both the metric and the \
coordinate system unambiguously. xCoba` provides a specialised function to \
define the values for the components of a metric, which can be used on its \
own or as an option for DefBasis.\
\>", "Text",
 CellID->1027654138,ExpressionUUID->"3e96739b-3d0e-43df-8d83-d7b3e871363f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "MetricInBasis"}]], "Input",
 CellLabel->"In[65]:=",
 CellID->1719608192,ExpressionUUID->"bc76d497-72bf-48a7-ba4a-f0f921098953"],

Cell[BoxData["\<\"MetricInBasis[metric, -basis, matrix] stores the values in \
matrix for the covariant components of metric in the given basis. \
MetricInBasis[metric, -basis, diagonal] stores the given values for the \
diagonal list of components, and zero everywhere else. MetricInBasis[metric, \
-basis, \\\"Orthonormal\\\"] takes diagonal values from \
SignatureOfMetric[metric] and zero everywhere else. MetricInBasis[metric, \
-basis, \\\"Orthogonal\\\"] stores zero off-diagonal values, but does not \
define values on the diagonal. MetricInBasis[metric, basis, values] stores \
values for the contravariant components of the inverse metric, where values \
is any of the previous possibilities. Finally, MetricInBasis[metric, basis, \
expr] computes the components from the abstract expression expr, using the \
given basis. In this case we check that the character of basis coincides with \
that of the two indices of expr.\\n\\nMetricInBasis is also an option for \
DefBasis and DefChart specifying that information at definition time, that is \
MetricInBasis->{metric, -basis, matrix}. Its default value is {}.\"\>"], \
"Print", "PrintUsage",
 CellChangeTimes->{3.7309133161044893`*^9},
 CellTags->"Info653730895316-7554244",
 CellID->876129877,ExpressionUUID->"c16536ba-9f45-451f-924f-29b7afe430d3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefMetric", "[", 
  RowBox[{
   RowBox[{"-", "1"}], ",", 
   RowBox[{"metric", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", " ", "cd", ",", 
   RowBox[{"{", 
    RowBox[{"\"\<|\>\"", ",", "\"\<D\>\""}], "}"}], ",", 
   RowBox[{"PrintAs", "\[Rule]", " ", "\"\<g\>\""}]}], "]"}]], "Input",
 CellChangeTimes->{{3.7287524432070827`*^9, 3.728752452234885*^9}, {
  3.728752491252006*^9, 3.728752495752884*^9}},
 CellLabel->"In[66]:=",
 CellID->304471059,ExpressionUUID->"db60b5f2-3336-443a-a1c8-e86ab574260b"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric metric tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"metric", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric metric tensor ", 
   $CellContext`metric[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287524974844646`*^9, 3.728752718483694*^9, 
  3.728753091477098*^9, 3.728753188144212*^9, 3.730913316225176*^9},
 CellLabel->"During evaluation of In[66]:=",
 CellID->1966958907,ExpressionUUID->"3a304b51-f050-4a70-8fd8-f3ee365822a6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric tensor \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"epsilonmetric", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric tensor ", 
   $CellContext`epsilonmetric[-$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287524974844646`*^9, 3.728752718483694*^9, 
  3.728753091477098*^9, 3.728753188144212*^9, 3.7309133162397127`*^9},
 CellLabel->"During evaluation of In[66]:=",
 CellID->315088031,ExpressionUUID->"5282bc52-f937-479b-be8f-c9719c05573a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"cd", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "covariant derivative ", 
   $CellContext`cd[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287524974844646`*^9, 3.728752718483694*^9, 
  3.728753091477098*^9, 3.728753188144212*^9, 3.730913316253883*^9},
 CellLabel->"During evaluation of In[66]:=",
 CellID->2137347162,ExpressionUUID->"d9d3a2c2-4f04-4a0a-94eb-4f3387ee7228"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing torsion tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"Torsioncd", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing torsion tensor ", 
   $CellContext`Torsioncd[$CellContext`a, -$CellContext`b, -$CellContext`c], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287524974844646`*^9, 3.728752718483694*^9, 
  3.728753091477098*^9, 3.728753188144212*^9, 3.7309133162623034`*^9},
 CellLabel->"During evaluation of In[66]:=",
 CellID->1434788365,ExpressionUUID->"994953e3-3314-460e-87a2-c682731a243f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"Christoffelcd", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Christoffel tensor ", 
   $CellContext`Christoffelcd[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287524974844646`*^9, 3.728752718483694*^9, 
  3.728753091477098*^9, 3.728753188144212*^9, 3.730913316270488*^9},
 CellLabel->"During evaluation of In[66]:=",
 CellID->394178392,ExpressionUUID->"79dc15bb-ff3f-433a-a2ff-4ece90a0da0a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"Riemanncd", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Riemann tensor ", 
   $CellContext`Riemanncd[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287524974844646`*^9, 3.728752718483694*^9, 
  3.728753091477098*^9, 3.728753188144212*^9, 3.73091331627886*^9},
 CellLabel->"During evaluation of In[66]:=",
 CellID->2116666996,ExpressionUUID->"095a1c65-d3c5-482e-8f00-cf49ffe17baf"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"Riccicd", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric Ricci tensor ", 
   $CellContext`Riccicd[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287524974844646`*^9, 3.728752718483694*^9, 
  3.728753091477098*^9, 3.728753188144212*^9, 3.7309133162872868`*^9},
 CellLabel->"During evaluation of In[66]:=",
 CellID->1755589217,ExpressionUUID->"79ccfb33-a200-43f7-9b82-ea91a76e3079"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Riemann automatically replaced \
by Ricci.\"\>"], "Print",
 CellChangeTimes->{3.7287524974844646`*^9, 3.728752718483694*^9, 
  3.728753091477098*^9, 3.728753188144212*^9, 3.7309133162957*^9},
 CellLabel->"During evaluation of In[66]:=",
 CellID->1910143986,ExpressionUUID->"928e9872-373d-4d1c-8b85-e894cda2f769"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"RicciScalarcd", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Ricci scalar ", 
   $CellContext`RicciScalarcd[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287524974844646`*^9, 3.728752718483694*^9, 
  3.728753091477098*^9, 3.728753188144212*^9, 3.730913316303938*^9},
 CellLabel->"During evaluation of In[66]:=",
 CellID->1056933699,ExpressionUUID->"28907ee7-bc12-40a0-8269-a6c224811b3c"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Ricci automatically replaced \
by RicciScalar.\"\>"], "Print",
 CellChangeTimes->{3.7287524974844646`*^9, 3.728752718483694*^9, 
  3.728753091477098*^9, 3.728753188144212*^9, 3.730913316312277*^9},
 CellLabel->"During evaluation of In[66]:=",
 CellID->2047434756,ExpressionUUID->"bd236b7a-a78c-474c-a251-031d6e41981d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Einstein tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"Einsteincd", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Einstein tensor ", 
   $CellContext`Einsteincd[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287524974844646`*^9, 3.728752718483694*^9, 
  3.728753091477098*^9, 3.728753188144212*^9, 3.730913316320073*^9},
 CellLabel->"During evaluation of In[66]:=",
 CellID->2105468211,ExpressionUUID->"35f261dd-d048-4fad-8520-28bde6496ad1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Weyl tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"Weylcd", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Weyl tensor ", 
   $CellContext`Weylcd[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287524974844646`*^9, 3.728752718483694*^9, 
  3.728753091477098*^9, 3.728753188144212*^9, 3.730913316328292*^9},
 CellLabel->"During evaluation of In[66]:=",
 CellID->368687011,ExpressionUUID->"2be8e081-def4-4f70-849a-bf2abd7d5463"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric TFRicci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TFRiccicd", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric TFRicci tensor ", 
   $CellContext`TFRiccicd[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287524974844646`*^9, 3.728752718483694*^9, 
  3.728753091477098*^9, 3.728753188144212*^9, 3.730913316339821*^9},
 CellLabel->"During evaluation of In[66]:=",
 CellID->1032272282,ExpressionUUID->"999b64e0-93b4-480f-b782-1a0eb65d17a1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Kretschmann scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"Kretschmanncd", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Kretschmann scalar ", 
   $CellContext`Kretschmanncd[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287524974844646`*^9, 3.728752718483694*^9, 
  3.728753091477098*^9, 3.728753188144212*^9, 3.7309133163516293`*^9},
 CellLabel->"During evaluation of In[66]:=",
 CellID->2027343108,ExpressionUUID->"0c5d9a82-ad40-45f4-aca9-39973954a4fb"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RiemannToWeylRules for dim \"\>", 
   "\[InvisibleSpace]", "3"}],
  SequenceForm["** DefCovD:  Computing RiemannToWeylRules for dim ", 3],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287524974844646`*^9, 3.728752718483694*^9, 
  3.728753091477098*^9, 3.728753188144212*^9, 3.730913316362158*^9},
 CellLabel->"During evaluation of In[66]:=",
 CellID->1580072854,ExpressionUUID->"1f00eeed-25e8-451d-a84c-ed11fcfa1d3e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RicciToTFRicci for dim \"\>", 
   "\[InvisibleSpace]", "3"}],
  SequenceForm["** DefCovD:  Computing RicciToTFRicci for dim ", 3],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287524974844646`*^9, 3.728752718483694*^9, 
  3.728753091477098*^9, 3.728753188144212*^9, 3.7309133163720713`*^9},
 CellLabel->"During evaluation of In[66]:=",
 CellID->501678376,ExpressionUUID->"ef978c55-14d4-4200-ba37-28f0422701e6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RicciToEinsteinRules for dim \"\>", 
   "\[InvisibleSpace]", "3"}],
  SequenceForm["** DefCovD:  Computing RicciToEinsteinRules for dim ", 3],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287524974844646`*^9, 3.728752718483694*^9, 
  3.728753091477098*^9, 3.728753188144212*^9, 3.730913316381966*^9},
 CellLabel->"During evaluation of In[66]:=",
 CellID->687667310,ExpressionUUID->"6573b891-a9ce-4973-83d8-8fb6167332a0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density \"\>", "\[InvisibleSpace]", 
   RowBox[{"Detmetric", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Determinant.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "weight +2 density ", 
   $CellContext`Detmetric[], ". ", "Determinant."],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287524974844646`*^9, 3.728752718483694*^9, 
  3.728753091477098*^9, 3.728753188144212*^9, 3.730913316392027*^9},
 CellLabel->"During evaluation of In[66]:=",
 CellID->2014974663,ExpressionUUID->"153ac3d1-1593-4e88-864a-07faab65692b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData["$Metrics"], "Input",
 CellLabel->"In[67]:=",
 CellID->104258482,ExpressionUUID->"d2985e35-a4c1-4b9f-ad31-2e070cd816ac"],

Cell[BoxData[
 RowBox[{"{", "metric", "}"}]], "Output",
 CellChangeTimes->{3.728752500353754*^9, 3.7287527188853283`*^9, 
  3.7287530917761106`*^9, 3.7287531887996798`*^9, 3.73091331654132*^9},
 CellLabel->"Out[67]=",
 CellID->2009531498,ExpressionUUID->"07e9964a-c2d0-4453-a2c1-52d63a561ba3"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"$CVVerbose", "=", "False"}], ";"}]], "Input",
 CellLabel->"In[68]:=",
 CellID->252209308,ExpressionUUID->"f1ef9782-5cb4-43ec-b915-0423e6eca365"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"MetricInBasis", "[", 
  RowBox[{"metric", ",", " ", 
   RowBox[{"-", "polar"}], ",", " ", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"i", "+", "j"}], ",", 
     RowBox[{"{", 
      RowBox[{"i", ",", "3"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"j", ",", "3"}], "}"}]}], "]"}]}], "]"}]], "Input",
 CellLabel->"In[69]:=",
 CellID->1896604025,ExpressionUUID->"04f66819-547b-4331-a7e4-37b2bb154e63"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", "2"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", "3"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", "4"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", "3"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", "4"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", "5"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{2, -$CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", "4"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{2, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", "5"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{2, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", "6"}]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728752501958098*^9, 3.7287527190934343`*^9, 
  3.728753092058606*^9, 3.728753189328863*^9, 3.730913316675584*^9},
 CellLabel->"Out[69]=",
 CellID->720570517,ExpressionUUID->"af4ef070-0aba-4c3c-adfe-43b697833815"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TensorValues", "[", "metric", "]"}]], "Input",
 CellLabel->"In[70]:=",
 CellID->927554819,ExpressionUUID->"c061e3bb-a5c9-44f5-99b0-22802acbf4fe"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{2, -$CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{2, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", "2"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", "3"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", "4"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", "4"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", "5"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{2, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", "6"}]}], "}"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728752502723617*^9, 3.728752719222382*^9, 
  3.7287530921895227`*^9, 3.728753189528697*^9, 3.730913316782445*^9},
 CellLabel->"Out[70]=",
 CellID->968393786,ExpressionUUID->"3f60ad0b-2698-43b4-84c4-b3c100088993"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["An example with a diagonal metric", "Text",
 CellID->585822608,ExpressionUUID->"1fc1843f-c177-4023-a458-3060b281f830"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DeleteTensorValues", "[", "metric", "]"}]], "Input",
 CellLabel->"In[71]:=",
 CellID->1855534808,ExpressionUUID->"629e18de-32db-4849-a24c-635e32820dc7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Deleted values for tensor \"\>", "\[InvisibleSpace]", "metric",
    "\[InvisibleSpace]", "\<\", derivatives \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", "}"}], "\[InvisibleSpace]", "\<\" and bases \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "polar"}], ",", 
      RowBox[{"-", "polar"}]}], "}"}], "}"}], 
   "\[InvisibleSpace]", "\<\".\"\>"}],
  SequenceForm[
  "Deleted values for tensor ", $CellContext`metric, ", derivatives ", {}, 
   " and bases ", {{-$CellContext`polar, -$CellContext`polar}}, "."],
  Editable->False]], "Print",
 CellChangeTimes->{3.72875250395888*^9, 3.7287527193399897`*^9, 
  3.728753092325674*^9, 3.728753189705442*^9, 3.7309133169071817`*^9},
 CellLabel->"During evaluation of In[71]:=",
 CellID->15800689,ExpressionUUID->"789c037c-9e04-4b6d-841e-06c44bff9cb7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"MetricInBasis", "[", 
  RowBox[{"metric", ",", 
   RowBox[{"-", "polar"}], ",", " ", 
   RowBox[{"{", 
    RowBox[{"1", ",", "2", ",", "3"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[72]:=",
 CellID->1002820424,ExpressionUUID->"5008830a-33fe-4733-ba75-d551df0a27f8"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", "1"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", "2"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{2, -$CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{2, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{2, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", "3"}]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728752504719063*^9, 3.728752719515377*^9, 
  3.728753092450733*^9, 3.72875318988873*^9, 3.730913317036084*^9},
 CellLabel->"Out[72]=",
 CellID->1421691442,ExpressionUUID->"a6ac495f-0711-494a-8b88-b29dfe3a4cfc"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["And know with an orthonormal one", "Text",
 CellID->1025964916,ExpressionUUID->"573aa27c-ca22-41f5-a86a-56d70de6755d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "SignatureOfMetric"}]], "Input",
 CellLabel->"In[73]:=",
 CellID->1246141449,ExpressionUUID->"72060461-9b16-406c-9454-22f935f114b9"],

Cell[BoxData["\<\"SignatureOfMetric[metric] gives the signature of the \
metric, in the form of a list of three elements: {p1s, m1s, zeros} giving the \
numbers of +1's, -1's and zeros, respectively, always in this order.\"\>"], \
"Print", "PrintUsage",
 CellChangeTimes->{3.730913317240342*^9},
 CellTags->"Info733730895317-7554244",
 CellID->1465009172,ExpressionUUID->"b001b759-6b67-43a6-85bf-2c61385ba6fd"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SignatureOfMetric", "[", "metric", "]"}], "^=", 
  RowBox[{"{", 
   RowBox[{"2", ",", "1", ",", "0"}], "}"}]}]], "Input",
 CellLabel->"In[74]:=",
 CellID->1964905018,ExpressionUUID->"91600d3a-1b2f-4e33-b40b-94efd0a65389"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"2", ",", "1", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.728752507240575*^9, 3.7287527202560177`*^9, 
  3.728753092775496*^9, 3.728753190590436*^9, 3.730913317389542*^9},
 CellLabel->"Out[74]=",
 CellID->1509140949,ExpressionUUID->"65fe807b-11b8-4244-b358-0c0d15456cb5"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"MetricInBasis", "[", 
  RowBox[{"metric", ",", 
   RowBox[{"-", "cartesian"}], ",", " ", "\"\<Orthonormal\>\""}], 
  "]"}]], "Input",
 CellLabel->"In[75]:=",
 CellID->2102661909,ExpressionUUID->"e8abf8ae-86ab-483f-ba45-3c5032f91549"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`cartesian}, {
        0, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", "1"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`cartesian}, {
        1, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`cartesian}, {
        2, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`cartesian}, {
        0, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`cartesian}, {
        1, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", "1"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`cartesian}, {
        2, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{2, -$CellContext`cartesian}, {
        0, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{2, -$CellContext`cartesian}, {
        1, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{2, -$CellContext`cartesian}, {
        2, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", "1"}]}]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.7287527204386787`*^9, 3.728753092925641*^9, 
  3.728753190966591*^9, 3.7309133175218563`*^9},
 CellLabel->"Out[75]=",
 CellID->202575126,ExpressionUUID->"47d43d34-e78a-469b-91e5-96b415dd83ab"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TensorValues", "[", "metric", "]"}]], "Input",
 CellLabel->"In[76]:=",
 CellID->1552088429,ExpressionUUID->"dc93bbe8-137e-40cd-8ece-1238eb42d99d"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`cartesian}, {
        0, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`cartesian}, {
        1, -$CellContext`cartesian}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{2, -$CellContext`cartesian}, {
        0, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`cartesian}, {
        2, -$CellContext`cartesian}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{2, -$CellContext`cartesian}, {
        1, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`cartesian}, {
        2, -$CellContext`cartesian}],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`cartesian}, {
        0, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", "1"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`cartesian}, {
        1, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`cartesian}, {
        2, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`cartesian}, {
        1, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", "1"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`cartesian}, {
        2, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{2, -$CellContext`cartesian}, {
        2, -$CellContext`cartesian}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", "1"}]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{2, -$CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{2, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", "1"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", "2"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{2, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", "3"}]}], "}"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728752720634378*^9, 3.7287530930641003`*^9, 
  3.728753191179035*^9, 3.730913317672484*^9},
 CellLabel->"Out[76]=",
 CellID->918562187,ExpressionUUID->"64a63038-bda9-4bbe-a695-3049153ac511"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["4.9. Changing bases or index characters", "Subsection",
 CellChangeTimes->{{3.728643013625104*^9, 3.7286430139538403`*^9}, {
  3.728673966125646*^9, 3.7286739677487698`*^9}, {3.728675291999371*^9, 
  3.7286752921592827`*^9}},
 CellID->2047113955,ExpressionUUID->"5601da75-08fc-42b3-a99b-511271499d04"],

Cell[TextData[{
 StyleBox["ChangeComponents\t\t", "DisplayMath"],
 StyleBox["Compute new TensorValues from an already stored index \
configuration", "Text"]
}], "DefinitionBox",
 CellID->1008733098,ExpressionUUID->"a12d9678-6dc2-4dea-9417-401f0f22f541"],

Cell["Changing bases or index characters.", "Text",
 CellChangeTimes->{3.728751843854821*^9},
 CellID->1541268668,ExpressionUUID->"92c53221-0337-4636-bf68-e9fc0aff8c3b"],

Cell[TextData[{
 "Suppose we already have the tensor values for a given ",
 StyleBox["ValID", "MR"],
 " and we want to generate the corresponding set for an (inequivalent) index \
configuration. We may want to perform a typical change of basis like ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["T", 
    StyleBox["ab",
     FontSlant->"Italic",
     FontColor->RGBColor[1, 0, 0]]], TraditionalForm]],ExpressionUUID->
  "77b416c8-72a4-4b15-b809-8565be6c5fad"],
 " to ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["T", 
    StyleBox["ab",
     FontSlant->"Italic",
     FontColor->RGBColor[0, 1, 0]]], TraditionalForm]],ExpressionUUID->
  "0fde216b-1940-4e91-9bd9-24c7cfa14fce"],
 ", for example. Or we may want to lower or raise some indices. Given \
starting and finishing index configurations, ",
 StyleBox["xCoba`", "MR"],
 " can compute the best route between them with the function ",
 StyleBox["ChangeComponents", "MR"]
}], "Text",
 CellID->1154752962,ExpressionUUID->"d284178e-ec27-43ee-9927-60a5daeda301"],

Cell[CellGroupData[{

Cell["A simple example", "Text",
 CellID->1123314395,ExpressionUUID->"2316770c-9622-4ffb-82ba-6050909bee86"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ChangeComponents", "[", 
  RowBox[{
   RowBox[{"v", "[", 
    RowBox[{"{", 
     RowBox[{"a", ",", "polar"}], "}"}], "]"}], ",", 
   RowBox[{"v", "[", 
    RowBox[{"-", 
     RowBox[{"{", 
      RowBox[{"a", ",", "polar"}], "}"}]}], "]"}]}], "]"}]], "Input",
 CellLabel->"In[77]:=",
 CellID->1016295988,ExpressionUUID->"772359a1-40e2-40ba-a10c-34e2b4052f7b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Computed \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {
             StyleBox["a",
              FontColor->RGBColor[1., 0., 0.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[{$CellContext`a, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {
              StyleBox["a",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["b",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`metric[{$CellContext`a, $CellContext`polar}, \
{$CellContext`b, $CellContext`polar}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {
              StyleBox["b",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{-$CellContext`b, -$CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" in \"\>", 
   "\[InvisibleSpace]", "0.013855`4.593151523614222", 
   "\[InvisibleSpace]", "\<\" Seconds\"\>"}],
  SequenceForm[
  "Computed ", $CellContext`v[{$CellContext`a, $CellContext`polar}] -> \
$CellContext`metric[{$CellContext`a, $CellContext`polar}, {$CellContext`b, \
$CellContext`polar}] $CellContext`v[{-$CellContext`b, -$CellContext`polar}], 
   " in ", 0.013855`4.593151523614222, " Seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.72875272514581*^9, 3.728753093181885*^9, 
  3.728753194351253*^9, 3.730913323812784*^9},
 CellLabel->"During evaluation of In[77]:=",
 CellID->567600032,ExpressionUUID->"188b3c53-3278-4d22-ad2a-255f816fc4c0"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{0, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, $CellContext`polar}, {0, $CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {" "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{0, -$CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, $CellContext`polar}, {1, $CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {" "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{1, -$CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {" "},
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{2, -$CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{1, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{1, $CellContext`polar}, {0, $CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {" "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{0, -$CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{1, $CellContext`polar}, {1, $CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {" "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{1, -$CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{1, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {" "},
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{2, -$CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{2, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{2, $CellContext`polar}, {0, $CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {" "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{0, -$CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{2, $CellContext`polar}, {1, $CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {" "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{1, -$CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{2, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {" "},
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{2, -$CellContext`polar}],
         Editable->False]}]}]}]}], "}"}]}], "]"}]], "Output",
 CellChangeTimes->{3.7287527252571077`*^9, 3.728753093213374*^9, 
  3.72875319447541*^9, 3.730913323871957*^9},
 CellLabel->"Out[77]=",
 CellID->531377850,ExpressionUUID->"843e134d-05dc-4629-bc3a-413fe76ba799"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TensorValIDs", "[", "v", "]"}]], "Input",
 CellLabel->"In[78]:=",
 CellID->984105045,ExpressionUUID->"45b67b8d-e04c-449e-b62f-877f93b32d49"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"ValID", "[", 
    RowBox[{"v", ",", 
     RowBox[{"{", 
      RowBox[{"{", "polar", "}"}], "}"}]}], "]"}], ",", 
   RowBox[{"ValID", "[", 
    RowBox[{"v", ",", 
     RowBox[{"{", 
      RowBox[{"{", 
       RowBox[{"-", "polar"}], "}"}], "}"}]}], "]"}]}], "}"}]], "Output",
 CellChangeTimes->{3.72875272539921*^9, 3.72875309337459*^9, 
  3.728753194647818*^9, 3.730913324816741*^9},
 CellLabel->"Out[78]=",
 CellID->357594777,ExpressionUUID->"5cbf25c4-c62b-4433-937a-6c31bfff933c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DeleteTensorValues", "[", "v", "]"}]], "Input",
 CellLabel->"In[79]:=",
 CellID->1975322205,ExpressionUUID->"e5e7dde7-f86f-42bf-b8a8-cb897ed938e3"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Deleted values for tensor \"\>", "\[InvisibleSpace]", "v", 
   "\[InvisibleSpace]", "\<\", derivatives \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", "}"}], "\[InvisibleSpace]", "\<\" and bases \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"{", "polar", "}"}], "}"}], "\[InvisibleSpace]", "\<\".\"\>"}],
  SequenceForm[
  "Deleted values for tensor ", $CellContext`v, ", derivatives ", {}, 
   " and bases ", {{$CellContext`polar}}, "."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752725736408*^9, 3.728753093493264*^9, 
  3.7287531948518553`*^9, 3.730913325475657*^9},
 CellLabel->"During evaluation of In[79]:=",
 CellID->977772303,ExpressionUUID->"68a92604-4d41-4134-8def-f350ffed9e22"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Deleted values for tensor \"\>", "\[InvisibleSpace]", "v", 
   "\[InvisibleSpace]", "\<\", derivatives \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", "}"}], "\[InvisibleSpace]", "\<\" and bases \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"{", 
     RowBox[{"-", "polar"}], "}"}], "}"}], "\[InvisibleSpace]", "\<\".\"\>"}],
  
  SequenceForm[
  "Deleted values for tensor ", $CellContext`v, ", derivatives ", {}, 
   " and bases ", {{-$CellContext`polar}}, "."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752725736408*^9, 3.728753093493264*^9, 
  3.7287531948518553`*^9, 3.730913325492299*^9},
 CellLabel->"During evaluation of In[79]:=",
 CellID->874229677,ExpressionUUID->"d801f1a2-64bc-4564-8360-2488161c5d99"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "A more complicated example, with an antisymmetric tensor ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["Y", "ab"], TraditionalForm]],ExpressionUUID->
  "bb5b246a-7ef9-472d-bebc-e4be5e674aa7"],
 ". To keep it shorter we can change the cnumbers of polar to make ",
 StyleBox["xCoba` ", "MR"],
 "think the manifold is two dimensional"
}], "Text",
 CellID->317888728,ExpressionUUID->"8f4b7272-0844-40fb-aa53-8e9226aa2ab0"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"Y", "[", 
    RowBox[{"a", ",", "b"}], "]"}], ",", "M3", ",", 
   RowBox[{"Antisymmetric", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "2"}], "}"}], "]"}]}], "]"}]], "Input",
 CellLabel->"In[80]:=",
 CellID->1964021407,ExpressionUUID->"3b81f1f8-8772-4354-8b73-87af195e5534"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"Y", "[", 
    RowBox[{"a", ",", "b"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`Y[$CellContext`a, $CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752726000536*^9, 3.728753093660549*^9, 
  3.728753195101433*^9, 3.730913326752368*^9},
 CellLabel->"During evaluation of In[80]:=",
 CellID->714658521,ExpressionUUID->"4703eae4-62fd-4842-952d-424749bfd346"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CNumbersOf", "[", "polar", "]"}], " ", "^=", 
  RowBox[{"{", 
   RowBox[{"1", ",", "2"}], "}"}]}]], "Input",
 CellLabel->"In[81]:=",
 CellID->1594741286,ExpressionUUID->"0d7f1732-1399-4c5a-b846-a8fc2c085de6"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"1", ",", "2"}], "}"}]], "Output",
 CellChangeTimes->{3.7287527261641617`*^9, 3.728753093779228*^9, 
  3.728753195435459*^9, 3.730913327325933*^9},
 CellLabel->"Out[81]=",
 CellID->1989078215,ExpressionUUID->"ec176970-451e-472c-a294-98774fc30040"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We set some configuration global variables", "Text",
 CellID->904151787,ExpressionUUID->"bc18e416-c471-4c1f-b7d6-c97268ffbeb2"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"$CVVerbose", "=", "True"}], ";"}], "\n", 
 RowBox[{
  RowBox[{"$CVReplace", "=", "False"}], ";"}], "\n", 
 RowBox[{
  RowBox[{"$CCSimplify", "=", "Expand"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"$TUseValues", "=", "ToCanonical"}], ";"}]}], "Input",
 CellLabel->"In[82]:=",
 CellID->484797937,ExpressionUUID->"96302cfa-192d-4119-91b4-da1c7a9ad25f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ChangeComponents", "[", 
  RowBox[{
   RowBox[{"Y", "[", 
    RowBox[{
     RowBox[{"-", 
      RowBox[{"{", 
       RowBox[{"a", ",", "polar"}], "}"}]}], ",", 
     RowBox[{"-", 
      RowBox[{"{", 
       RowBox[{"b", ",", "polar"}], "}"}]}]}], "]"}], ",", 
   RowBox[{"Y", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"a", ",", "polar"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"b", ",", "polar"}], "}"}]}], "]"}]}], "]"}]], "Input",
 CellLabel->"In[86]:=",
 CellID->445970441,ExpressionUUID->"3949617e-ef69-43b2-a456-71b9a8f3eca5"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{0, $CellContext`polar}, {0, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`Y[{0, $CellContext`polar}, {
     0, $CellContext`polar}] -> 0, " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.730913329025522*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->798952431,ExpressionUUID->"2b74f52f-d1a9-4679-bae2-e0e1c3416105"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added independent rule ", $CellContext`Y[{0, $CellContext`polar}, {
     1, $CellContext`polar}] -> $CellContext`Y[{0, $CellContext`polar}, {
     1, $CellContext`polar}], " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.7309133290875883`*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->1925746773,ExpressionUUID->"f38eaf3f-218f-4a58-9f9e-858eb9190d80"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added independent rule ", $CellContext`Y[{0, $CellContext`polar}, {
     2, $CellContext`polar}] -> $CellContext`Y[{0, $CellContext`polar}, {
     2, $CellContext`polar}], " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.73091332916409*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->1299969874,ExpressionUUID->"3996d130-49c3-462d-b19a-3c33ed527126"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{1, $CellContext`polar}, {0, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`Y[{1, $CellContext`polar}, {
     0, $CellContext`polar}] -> -$CellContext`Y[{0, $CellContext`polar}, {
      1, $CellContext`polar}], " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.730913329263636*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->1047127517,ExpressionUUID->"ef109891-4304-41e4-b905-adb550fed93e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{1, $CellContext`polar}, {1, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`Y[{1, $CellContext`polar}, {
     1, $CellContext`polar}] -> 0, " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.730913329346261*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->443691054,ExpressionUUID->"5d7997fc-3b19-402c-ba51-dc894d1be8a4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added independent rule ", $CellContext`Y[{1, $CellContext`polar}, {
     2, $CellContext`polar}] -> $CellContext`Y[{1, $CellContext`polar}, {
     2, $CellContext`polar}], " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.730913329407629*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->1052322194,ExpressionUUID->"247fa0ec-b4bf-47b6-ab28-4d601fa0524f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{2, $CellContext`polar}, {0, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`Y[{2, $CellContext`polar}, {
     0, $CellContext`polar}] -> -$CellContext`Y[{0, $CellContext`polar}, {
      2, $CellContext`polar}], " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.7309133294752693`*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->1851281513,ExpressionUUID->"2a807cb8-2b07-454a-b2e7-2c9625c0d93a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{2, $CellContext`polar}, {1, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`Y[{2, $CellContext`polar}, {
     1, $CellContext`polar}] -> -$CellContext`Y[{1, $CellContext`polar}, {
      2, $CellContext`polar}], " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.730913329540771*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->1745311255,ExpressionUUID->"2a65be83-b8ae-4006-a2b0-0399c3244616"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{2, $CellContext`polar}, {2, $CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`Y[{2, $CellContext`polar}, {
     2, $CellContext`polar}] -> 0, " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.730913329599515*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->967128412,ExpressionUUID->"09c52b0a-f67d-4406-9576-de1512b89df1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{0, $CellContext`polar}, {0, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
       Editable->False]}]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added independent rule ", $CellContext`Y[{0, $CellContext`polar}, {
     0, -$CellContext`polar}] -> $CellContext`metric[{
       0, -$CellContext`polar}, {1, -$CellContext`polar}] $CellContext`Y[{
       0, $CellContext`polar}, {
       1, $CellContext`polar}] + $CellContext`metric[{
       0, -$CellContext`polar}, {2, -$CellContext`polar}] $CellContext`Y[{
       0, $CellContext`polar}, {2, $CellContext`polar}], 
   " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.730913329681231*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->223496341,ExpressionUUID->"661dc0d9-264a-4957-93ed-c3859012bf33"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{1, $CellContext`polar}, {0, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{0, -$CellContext`polar}, {1, $CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`Y[{1, $CellContext`polar}, {
     0, -$CellContext`polar}] -> -$CellContext`Y[{0, -$CellContext`polar}, {
      1, $CellContext`polar}], " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.730913329742652*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->354104584,ExpressionUUID->"2bc5502c-49dd-45a2-a83a-de7f5f987040"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{0, -$CellContext`polar}, {1, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
       Editable->False]}]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added independent rule ", $CellContext`Y[{0, -$CellContext`polar}, {
     1, $CellContext`polar}] -> $CellContext`metric[{
       0, -$CellContext`polar}, {0, -$CellContext`polar}] $CellContext`Y[{
       0, $CellContext`polar}, {
       1, $CellContext`polar}] - $CellContext`metric[{
      0, -$CellContext`polar}, {2, -$CellContext`polar}] $CellContext`Y[{
      1, $CellContext`polar}, {2, $CellContext`polar}], 
   " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.730913329810092*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->758557451,ExpressionUUID->"724b7161-e19d-47e5-8c0d-b3a25b858ac3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{2, $CellContext`polar}, {0, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{0, -$CellContext`polar}, {2, $CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`Y[{2, $CellContext`polar}, {
     0, -$CellContext`polar}] -> -$CellContext`Y[{0, -$CellContext`polar}, {
      2, $CellContext`polar}], " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.73091332988028*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->1612150365,ExpressionUUID->"a781fb83-0c57-49a8-9586-943d46febda3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{0, -$CellContext`polar}, {2, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
       Editable->False]}]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added independent rule ", $CellContext`Y[{0, -$CellContext`polar}, {
     2, $CellContext`polar}] -> $CellContext`metric[{
       0, -$CellContext`polar}, {0, -$CellContext`polar}] $CellContext`Y[{
       0, $CellContext`polar}, {
       2, $CellContext`polar}] + $CellContext`metric[{
       0, -$CellContext`polar}, {1, -$CellContext`polar}] $CellContext`Y[{
       1, $CellContext`polar}, {2, $CellContext`polar}], 
   " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.730913329939548*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->610010356,ExpressionUUID->"9e5118c3-8082-4615-8a1c-09eb394a073c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{0, $CellContext`polar}, {1, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
       Editable->False]}]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added independent rule ", $CellContext`Y[{0, $CellContext`polar}, {
     1, -$CellContext`polar}] -> $CellContext`metric[{
       1, -$CellContext`polar}, {1, -$CellContext`polar}] $CellContext`Y[{
       0, $CellContext`polar}, {
       1, $CellContext`polar}] + $CellContext`metric[{
       1, -$CellContext`polar}, {2, -$CellContext`polar}] $CellContext`Y[{
       0, $CellContext`polar}, {2, $CellContext`polar}], 
   " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.730913330019596*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->357122514,ExpressionUUID->"6cd71785-287a-40ba-a8be-84c4d5d0d521"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{1, $CellContext`polar}, {1, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{0, -$CellContext`polar}, {
         1, -$CellContext`polar}],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
       Editable->False]}]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added independent rule ", $CellContext`Y[{1, $CellContext`polar}, {
     1, -$CellContext`polar}] -> (-$CellContext`metric[{
        0, -$CellContext`polar}, {1, -$CellContext`polar}]) $CellContext`Y[{
       0, $CellContext`polar}, {
       1, $CellContext`polar}] + $CellContext`metric[{
       1, -$CellContext`polar}, {2, -$CellContext`polar}] $CellContext`Y[{
       1, $CellContext`polar}, {2, $CellContext`polar}], 
   " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.730913330096685*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->2088783080,ExpressionUUID->"f29d6dc0-2c08-494d-98ee-b9be7e9fd7fe"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{2, $CellContext`polar}, {1, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{1, -$CellContext`polar}, {2, $CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`Y[{2, $CellContext`polar}, {
     1, -$CellContext`polar}] -> -$CellContext`Y[{1, -$CellContext`polar}, {
      2, $CellContext`polar}], " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.7309133301580267`*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->1262048051,ExpressionUUID->"29dc73f1-10fb-463c-95a5-1a987dfba9a0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{1, -$CellContext`polar}, {2, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
       Editable->False]}]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added independent rule ", $CellContext`Y[{1, -$CellContext`polar}, {
     2, $CellContext`polar}] -> $CellContext`metric[{
       0, -$CellContext`polar}, {1, -$CellContext`polar}] $CellContext`Y[{
       0, $CellContext`polar}, {
       2, $CellContext`polar}] + $CellContext`metric[{
       1, -$CellContext`polar}, {1, -$CellContext`polar}] $CellContext`Y[{
       1, $CellContext`polar}, {2, $CellContext`polar}], 
   " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.730913330224804*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->545611738,ExpressionUUID->"eeaf2f80-3e9e-496f-b533-d0475ebcaf2b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{0, $CellContext`polar}, {2, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{2, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
       Editable->False]}]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added independent rule ", $CellContext`Y[{0, $CellContext`polar}, {
     2, -$CellContext`polar}] -> $CellContext`metric[{
       1, -$CellContext`polar}, {2, -$CellContext`polar}] $CellContext`Y[{
       0, $CellContext`polar}, {
       1, $CellContext`polar}] + $CellContext`metric[{
       2, -$CellContext`polar}, {2, -$CellContext`polar}] $CellContext`Y[{
       0, $CellContext`polar}, {2, $CellContext`polar}], 
   " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.730913330293008*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->341307876,ExpressionUUID->"bbf2b4d7-cda1-4d8c-bd2f-5d624f6c81cb"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{1, $CellContext`polar}, {2, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{0, -$CellContext`polar}, {
         2, -$CellContext`polar}],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{2, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
       Editable->False]}]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added independent rule ", $CellContext`Y[{1, $CellContext`polar}, {
     2, -$CellContext`polar}] -> (-$CellContext`metric[{
        0, -$CellContext`polar}, {2, -$CellContext`polar}]) $CellContext`Y[{
       0, $CellContext`polar}, {
       1, $CellContext`polar}] + $CellContext`metric[{
       2, -$CellContext`polar}, {2, -$CellContext`polar}] $CellContext`Y[{
       1, $CellContext`polar}, {2, $CellContext`polar}], 
   " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.730913330357313*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->432650795,ExpressionUUID->"82de3e5e-7675-4306-882c-532bf0d9dbfb"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{2, $CellContext`polar}, {2, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{0, -$CellContext`polar}, {
         2, -$CellContext`polar}],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
       Editable->False]}]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added independent rule ", $CellContext`Y[{2, $CellContext`polar}, {
     2, -$CellContext`polar}] -> (-$CellContext`metric[{
        0, -$CellContext`polar}, {2, -$CellContext`polar}]) $CellContext`Y[{
       0, $CellContext`polar}, {
       2, $CellContext`polar}] - $CellContext`metric[{
      1, -$CellContext`polar}, {2, -$CellContext`polar}] $CellContext`Y[{
      1, $CellContext`polar}, {2, $CellContext`polar}], 
   " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.7309133304240417`*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->182282823,ExpressionUUID->"fb90fe3d-ccb1-439c-acee-9edd4e97d627"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{1, -$CellContext`polar}, {1, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{1, $CellContext`polar}, {1, -$CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`Y[{1, -$CellContext`polar}, {
     1, $CellContext`polar}] -> -$CellContext`Y[{1, $CellContext`polar}, {
      1, -$CellContext`polar}], " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.73091333049294*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->1315725539,ExpressionUUID->"1d6173d6-76a6-4a0b-86b1-c9e0023e6eb3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{2, -$CellContext`polar}, {1, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{1, $CellContext`polar}, {2, -$CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`Y[{2, -$CellContext`polar}, {
     1, $CellContext`polar}] -> -$CellContext`Y[{1, $CellContext`polar}, {
      2, -$CellContext`polar}], " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.7309133305583153`*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->760020814,ExpressionUUID->"61eef69e-ff16-450b-9095-ad1a4f12f54d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{2, -$CellContext`polar}, {2, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{2, $CellContext`polar}, {2, -$CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`Y[{2, -$CellContext`polar}, {
     2, $CellContext`polar}] -> -$CellContext`Y[{2, $CellContext`polar}, {
      2, -$CellContext`polar}], " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.7309133306245947`*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->437088542,ExpressionUUID->"518a9d0c-6058-4c99-99ea-7e797441d229"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Computed \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {
             StyleBox["a",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["b",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{$CellContext`a, $CellContext`polar}, {-$CellContext`b, \
-$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["c",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["b",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`metric[{-$CellContext`c, -$CellContext`polar}, \
{-$CellContext`b, -$CellContext`polar}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {
              StyleBox["a",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["c",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{$CellContext`a, $CellContext`polar}, {$CellContext`c, \
$CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" in \"\>", 
   "\[InvisibleSpace]", "1.062454`6.477855129414904", 
   "\[InvisibleSpace]", "\<\" Seconds\"\>"}],
  SequenceForm[
  "Computed ", $CellContext`Y[{$CellContext`a, $CellContext`polar}, \
{-$CellContext`b, -$CellContext`polar}] -> \
$CellContext`metric[{-$CellContext`c, -$CellContext`polar}, {-$CellContext`b, \
-$CellContext`polar}] $CellContext`Y[{$CellContext`a, $CellContext`polar}, \
{$CellContext`c, $CellContext`polar}], " in ", 1.062454`6.477855129414904, 
   " Seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.730913330714394*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->1753052654,ExpressionUUID->"c9319005-26c8-40f8-ad52-10e96b833eb1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{0, -$CellContext`polar}, {0, -$CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`Y[{0, -$CellContext`polar}, {
     0, -$CellContext`polar}] -> 0, " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.7309133307971373`*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->80318938,ExpressionUUID->"74fc2a1f-70f0-4eba-85b9-8000cc9d5db7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, -$CellContext`polar}, {2, $CellContext`polar}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, $CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False]}]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added independent rule ", $CellContext`Y[{0, -$CellContext`polar}, {
     1, -$CellContext`polar}] -> $CellContext`metric[{
       0, -$CellContext`polar}, {0, -$CellContext`polar}] $CellContext`Y[{
       0, $CellContext`polar}, {
       1, -$CellContext`polar}] - $CellContext`metric[{
      0, -$CellContext`polar}, {2, -$CellContext`polar}] $CellContext`Y[{
      1, -$CellContext`polar}, {
      2, $CellContext`polar}] + $CellContext`metric[{
       0, -$CellContext`polar}, {1, -$CellContext`polar}] $CellContext`Y[{
       1, $CellContext`polar}, {1, -$CellContext`polar}], 
   " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.7309133308815727`*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->1300660650,ExpressionUUID->"aa51c2e4-5ff3-4856-8964-717f06b59607"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{0, -$CellContext`polar}, {2, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, $CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{2, $CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False]}]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added independent rule ", $CellContext`Y[{0, -$CellContext`polar}, {
     2, -$CellContext`polar}] -> $CellContext`metric[{
       0, -$CellContext`polar}, {0, -$CellContext`polar}] $CellContext`Y[{
       0, $CellContext`polar}, {
       2, -$CellContext`polar}] + $CellContext`metric[{
       0, -$CellContext`polar}, {1, -$CellContext`polar}] $CellContext`Y[{
       1, $CellContext`polar}, {
       2, -$CellContext`polar}] + $CellContext`metric[{
       0, -$CellContext`polar}, {2, -$CellContext`polar}] $CellContext`Y[{
       2, $CellContext`polar}, {2, -$CellContext`polar}], 
   " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.730913330971043*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->967501737,ExpressionUUID->"63609b57-b9de-414b-b4cd-618396796203"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{1, -$CellContext`polar}, {0, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`Y[{1, -$CellContext`polar}, {
     0, -$CellContext`polar}] -> -$CellContext`Y[{0, -$CellContext`polar}, {
      1, -$CellContext`polar}], " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.7309133310478973`*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->55518675,ExpressionUUID->"cae46f48-5231-4dd5-88e5-a5f1e5c66e81"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Kept old independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, -$CellContext`polar}, {2, $CellContext`polar}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, $CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False]}]}]}], "\[InvisibleSpace]", "\<\" vs. new \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, -$CellContext`polar}, {1, $CellContext`polar}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, -$CellContext`polar}, {2, $CellContext`polar}],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False]}]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Kept old independent rule ", $CellContext`Y[{0, -$CellContext`polar}, {
     1, -$CellContext`polar}] -> $CellContext`metric[{
       0, -$CellContext`polar}, {0, -$CellContext`polar}] $CellContext`Y[{
       0, $CellContext`polar}, {
       1, -$CellContext`polar}] - $CellContext`metric[{
      0, -$CellContext`polar}, {2, -$CellContext`polar}] $CellContext`Y[{
      1, -$CellContext`polar}, {
      2, $CellContext`polar}] + $CellContext`metric[{
       0, -$CellContext`polar}, {1, -$CellContext`polar}] $CellContext`Y[{
       1, $CellContext`polar}, {1, -$CellContext`polar}], 
   " vs. new ", $CellContext`Y[{0, -$CellContext`polar}, {
     1, -$CellContext`polar}] -> $CellContext`metric[{
       1, -$CellContext`polar}, {1, -$CellContext`polar}] $CellContext`Y[{
       0, -$CellContext`polar}, {
       1, $CellContext`polar}] + $CellContext`metric[{
       1, -$CellContext`polar}, {2, -$CellContext`polar}] $CellContext`Y[{
       0, -$CellContext`polar}, {
       2, $CellContext`polar}] - $CellContext`metric[{
      0, -$CellContext`polar}, {1, -$CellContext`polar}] $CellContext`Y[{
      0, $CellContext`polar}, {0, -$CellContext`polar}], 
   " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.7309133311305313`*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->1094467927,ExpressionUUID->"5aaf199a-2b41-4c81-b976-6bc052bd1221"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{1, -$CellContext`polar}, {1, -$CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`Y[{1, -$CellContext`polar}, {
     1, -$CellContext`polar}] -> 0, " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.730913331213888*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->1525012534,ExpressionUUID->"e503dd77-6175-4100-b395-7ce11f58a412"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, $CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{2, $CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False]}]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added independent rule ", $CellContext`Y[{1, -$CellContext`polar}, {
     2, -$CellContext`polar}] -> $CellContext`metric[{
       0, -$CellContext`polar}, {1, -$CellContext`polar}] $CellContext`Y[{
       0, $CellContext`polar}, {
       2, -$CellContext`polar}] + $CellContext`metric[{
       1, -$CellContext`polar}, {1, -$CellContext`polar}] $CellContext`Y[{
       1, $CellContext`polar}, {
       2, -$CellContext`polar}] + $CellContext`metric[{
       1, -$CellContext`polar}, {2, -$CellContext`polar}] $CellContext`Y[{
       2, $CellContext`polar}, {2, -$CellContext`polar}], 
   " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.7309133312915077`*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->668331993,ExpressionUUID->"d5e33779-cc3f-41d4-b963-b7a30cafb337"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{2, -$CellContext`polar}, {0, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{0, -$CellContext`polar}, {2, -$CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`Y[{2, -$CellContext`polar}, {
     0, -$CellContext`polar}] -> -$CellContext`Y[{0, -$CellContext`polar}, {
      2, -$CellContext`polar}], " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.730913331359523*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->1316332541,ExpressionUUID->"7d994d69-d40c-4f86-ad14-79856c4fa169"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Kept old independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{0, -$CellContext`polar}, {2, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, $CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{2, $CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False]}]}]}], "\[InvisibleSpace]", "\<\" vs. new \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{0, -$CellContext`polar}, {2, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, -$CellContext`polar}, {1, $CellContext`polar}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{2, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, -$CellContext`polar}, {2, $CellContext`polar}],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False]}]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Kept old independent rule ", $CellContext`Y[{0, -$CellContext`polar}, {
     2, -$CellContext`polar}] -> $CellContext`metric[{
       0, -$CellContext`polar}, {0, -$CellContext`polar}] $CellContext`Y[{
       0, $CellContext`polar}, {
       2, -$CellContext`polar}] + $CellContext`metric[{
       0, -$CellContext`polar}, {1, -$CellContext`polar}] $CellContext`Y[{
       1, $CellContext`polar}, {
       2, -$CellContext`polar}] + $CellContext`metric[{
       0, -$CellContext`polar}, {2, -$CellContext`polar}] $CellContext`Y[{
       2, $CellContext`polar}, {2, -$CellContext`polar}], 
   " vs. new ", $CellContext`Y[{0, -$CellContext`polar}, {
     2, -$CellContext`polar}] -> $CellContext`metric[{
       1, -$CellContext`polar}, {2, -$CellContext`polar}] $CellContext`Y[{
       0, -$CellContext`polar}, {
       1, $CellContext`polar}] + $CellContext`metric[{
       2, -$CellContext`polar}, {2, -$CellContext`polar}] $CellContext`Y[{
       0, -$CellContext`polar}, {
       2, $CellContext`polar}] - $CellContext`metric[{
      0, -$CellContext`polar}, {2, -$CellContext`polar}] $CellContext`Y[{
      0, $CellContext`polar}, {0, -$CellContext`polar}], 
   " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.730913331441821*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->224485193,ExpressionUUID->"35009c62-b6a5-4c8f-b132-32717a7a57bc"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{2, -$CellContext`polar}, {1, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`Y[{2, -$CellContext`polar}, {
     1, -$CellContext`polar}] -> -$CellContext`Y[{1, -$CellContext`polar}, {
      2, -$CellContext`polar}], " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.730913331525361*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->1941354619,ExpressionUUID->"3a7c0e3c-db60-47f7-bbfc-19b0674dab02"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Kept old independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, $CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{2, $CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False]}]}]}], "\[InvisibleSpace]", "\<\" vs. new \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{0, -$CellContext`polar}, {
         2, -$CellContext`polar}],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{2, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, -$CellContext`polar}, {2, $CellContext`polar}],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, $CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False]}]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Kept old independent rule ", $CellContext`Y[{1, -$CellContext`polar}, {
     2, -$CellContext`polar}] -> $CellContext`metric[{
       0, -$CellContext`polar}, {1, -$CellContext`polar}] $CellContext`Y[{
       0, $CellContext`polar}, {
       2, -$CellContext`polar}] + $CellContext`metric[{
       1, -$CellContext`polar}, {1, -$CellContext`polar}] $CellContext`Y[{
       1, $CellContext`polar}, {
       2, -$CellContext`polar}] + $CellContext`metric[{
       1, -$CellContext`polar}, {2, -$CellContext`polar}] $CellContext`Y[{
       2, $CellContext`polar}, {2, -$CellContext`polar}], 
   " vs. new ", $CellContext`Y[{1, -$CellContext`polar}, {
     2, -$CellContext`polar}] -> (-$CellContext`metric[{
        0, -$CellContext`polar}, {2, -$CellContext`polar}]) $CellContext`Y[{
       0, $CellContext`polar}, {
       1, -$CellContext`polar}] + $CellContext`metric[{
       2, -$CellContext`polar}, {2, -$CellContext`polar}] $CellContext`Y[{
       1, -$CellContext`polar}, {
       2, $CellContext`polar}] - $CellContext`metric[{
      1, -$CellContext`polar}, {2, -$CellContext`polar}] $CellContext`Y[{
      1, $CellContext`polar}, {1, -$CellContext`polar}], 
   " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.730913331642336*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->1628032662,ExpressionUUID->"bac42445-8436-4e49-85b2-8a8a3dcf3d1b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{2, -$CellContext`polar}, {2, -$CellContext`polar}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`Y[{2, -$CellContext`polar}, {
     2, -$CellContext`polar}] -> 0, " for tensor ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.73091333173125*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->1255803807,ExpressionUUID->"ddd78133-278d-4224-a77d-eda45372d177"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Computed \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"Y", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["a",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["b",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Y[{-$CellContext`a, -$CellContext`polar}, {-$CellContext`b, \
-$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["c",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["a",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`metric[{-$CellContext`c, -$CellContext`polar}, \
{-$CellContext`a, -$CellContext`polar}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {
              StyleBox["c",
               FontColor->RGBColor[1., 0., 0.]], " "},
             {" ", 
              StyleBox["b",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{$CellContext`c, $CellContext`polar}, {-$CellContext`b, \
-$CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" in \"\>", 
   "\[InvisibleSpace]", "0.968667`6.437719498185803", 
   "\[InvisibleSpace]", "\<\" Seconds\"\>"}],
  SequenceForm[
  "Computed ", $CellContext`Y[{-$CellContext`a, -$CellContext`polar}, \
{-$CellContext`b, -$CellContext`polar}] -> \
$CellContext`metric[{-$CellContext`c, -$CellContext`polar}, {-$CellContext`a, \
-$CellContext`polar}] $CellContext`Y[{$CellContext`c, $CellContext`polar}, \
{-$CellContext`b, -$CellContext`polar}], " in ", 0.968667`6.437719498185803, 
   " Seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527264379463`*^9, 3.728753093984487*^9, 
  3.728753195927783*^9, 3.730913331813739*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->825301533,ExpressionUUID->"7c686282-8183-4d17-abc1-db618bca0829"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, -$CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, -$CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{2, -$CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{0, -$CellContext`polar}, {2, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{2, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{2, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          0, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], " "},
                {" ", 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {1, -$CellContext`polar}],
         Editable->False]}], "-", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{1, -$CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          1, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], " "},
                {" ", 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{1, $CellContext`polar}, {1, -$CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          0, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], " "},
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {2, -$CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          1, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], " "},
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{1, $CellContext`polar}, {2, -$CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]], " "},
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{2, $CellContext`polar}, {2, -$CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          1, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], " "},
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {2, -$CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{1, -$CellContext`polar}, {
          1, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], " "},
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{1, $CellContext`polar}, {2, -$CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{1, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]], " "},
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{2, $CellContext`polar}, {2, -$CellContext`polar}],
         Editable->False]}]}]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, $CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{0, -$CellContext`polar}, {1, $CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{2, $CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{0, -$CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{2, $CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{1, -$CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, -$CellContext`polar}, {1, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], " "},
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{1, $CellContext`polar}, {1, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{2, -$CellContext`polar}, {1, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], " "},
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{1, $CellContext`polar}, {2, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{2, -$CellContext`polar}, {2, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]], " "},
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{2, $CellContext`polar}, {2, -$CellContext`polar}],
        Editable->False]}]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          1, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, -$CellContext`polar}, {1, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          0, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
         Editable->False]}], "-", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, -$CellContext`polar}, {2, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          0, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          1, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{1, -$CellContext`polar}, {
          1, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{1, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, $CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"g", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["0",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["1",
                   FontColor->RGBColor[1., 0., 0.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`metric[{0, -$CellContext`polar}, {
           1, -$CellContext`polar}],
          Editable->False]}], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{1, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, -$CellContext`polar}, {2, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          1, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{1, -$CellContext`polar}, {
          1, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{1, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{2, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, $CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"g", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["0",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["2",
                   FontColor->RGBColor[1., 0., 0.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`metric[{0, -$CellContext`polar}, {
           2, -$CellContext`polar}],
          Editable->False]}], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{2, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{2, $CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"g", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["0",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["2",
                   FontColor->RGBColor[1., 0., 0.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`metric[{0, -$CellContext`polar}, {
           2, -$CellContext`polar}],
          Editable->False]}], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}], "-", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{1, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}]}]}]}], "}"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728752733023264*^9, 3.728753097087516*^9, 
  3.7287532015780582`*^9, 3.730913331935648*^9},
 CellLabel->"Out[86]=",
 CellID->245989158,ExpressionUUID->"c0084cef-ac58-47a8-9d81-7adb8abdea5e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The system prints ",
 Cell[BoxData[
  InterpretationBox[
   RowBox[{
    RowBox[{"\<\"Value \"\>", "\[InvisibleSpace]", 
     InterpretationBox[
      StyleBox[
       SubsuperscriptBox["g", 
        StyleBox[
         RowBox[{"1", "2"}],
         FontColor->RGBColor[1, 0, 0]], "  "],
       AutoSpacing->False],
      metric[{1, -polar}, {2, -polar}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[
       SubsuperscriptBox["Y", 
        RowBox[{" ", 
         StyleBox["2",
          FontColor->RGBColor[1, 0, 0]]}], 
        RowBox[{
         StyleBox["1",
          FontColor->RGBColor[1, 0, 0]], " "}]],
       AutoSpacing->False],
      Y[{1, polar}, {2, -polar}],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubsuperscriptBox["g", 
        StyleBox[
         RowBox[{"2", "2"}],
         FontColor->RGBColor[1, 0, 0]], "  "],
       AutoSpacing->False],
      metric[{2, -polar}, {2, -polar}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[
       SubsuperscriptBox["Y", 
        RowBox[{" ", 
         StyleBox["2",
          FontColor->RGBColor[1, 0, 0]]}], 
        RowBox[{
         StyleBox["2",
          FontColor->RGBColor[1, 0, 0]], " "}]],
       AutoSpacing->False],
      Y[{2, polar}, {2, -polar}],
      Editable->False], 
     "\[InvisibleSpace]", "\<\" expected to be 0 by symmetry.\"\>"}]}],
   SequenceForm[
   "Value ", metric[{1, -polar}, {2, -polar}] Y[{1, polar}, {2, -polar}] + 
    metric[{2, -polar}, {2, -polar}] Y[{2, polar}, {2, -polar}], 
    " expected to be 0 by symmetry."],
   Editable->False]], "Print",ExpressionUUID->
  "d3c0050a-f813-499a-a18d-cb62dc3a5e06"],
 "This is because of ",
 Cell[BoxData[
  RowBox[{"$UseValues", " ", "=", "ToCanonical"}]], "Print",ExpressionUUID->
  "cbe9c4b1-4725-430e-ac9f-28e71a3e3fe5"],
 " which tells the system only to use the independent rules at each step. If \
we had written ",
 Cell[BoxData[
  RowBox[{"$UseValues", " ", "=", "All"}]], "Print",ExpressionUUID->
  "bb293f9d-4a3d-4a1b-867e-2f40f7c6f6f6"],
 " instead that message would not have appeared. However, whenever a \
component is 0 because of the symmetry, the symmetry value is used, so our \
result is correct."
}], "Text",
 CellID->210155605,ExpressionUUID->"bd6a602b-98c4-4589-9ec9-59db2724c21c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TensorValues", "[", "Y", "]"}]], "Input",
 CellLabel->"In[87]:=",
 CellID->1622137567,ExpressionUUID->"8056cdde-7fd6-464d-b695-62c140cb2980"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, -$CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, -$CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{2, -$CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{0, -$CellContext`polar}, {2, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{2, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{2, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          0, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], " "},
                {" ", 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {1, -$CellContext`polar}],
         Editable->False]}], "-", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{1, -$CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          1, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], " "},
                {" ", 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{1, $CellContext`polar}, {1, -$CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          0, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], " "},
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {2, -$CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          1, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], " "},
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{1, $CellContext`polar}, {2, -$CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]], " "},
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{2, $CellContext`polar}, {2, -$CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          1, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], " "},
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {2, -$CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{1, -$CellContext`polar}, {
          1, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], " "},
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{1, $CellContext`polar}, {2, -$CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{1, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]], " "},
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{2, $CellContext`polar}, {2, -$CellContext`polar}],
         Editable->False]}]}]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, $CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{0, -$CellContext`polar}, {1, $CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{2, $CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{0, -$CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{2, $CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{1, -$CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, -$CellContext`polar}, {1, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], " "},
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{1, $CellContext`polar}, {1, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{2, -$CellContext`polar}, {1, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], " "},
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{1, $CellContext`polar}, {2, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{2, -$CellContext`polar}, {2, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]], " "},
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{2, $CellContext`polar}, {2, -$CellContext`polar}],
        Editable->False]}]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          1, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, -$CellContext`polar}, {1, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          0, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
         Editable->False]}], "-", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, -$CellContext`polar}, {2, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          0, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          1, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{1, -$CellContext`polar}, {
          1, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{1, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, $CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"g", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["0",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["1",
                   FontColor->RGBColor[1., 0., 0.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`metric[{0, -$CellContext`polar}, {
           1, -$CellContext`polar}],
          Editable->False]}], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{1, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, -$CellContext`polar}, {2, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          1, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{1, -$CellContext`polar}, {
          1, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{1, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{2, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, $CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"g", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["0",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["2",
                   FontColor->RGBColor[1., 0., 0.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`metric[{0, -$CellContext`polar}, {
           2, -$CellContext`polar}],
          Editable->False]}], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{2, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{2, $CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"g", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["0",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["2",
                   FontColor->RGBColor[1., 0., 0.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`metric[{0, -$CellContext`polar}, {
           2, -$CellContext`polar}],
          Editable->False]}], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}], "-", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{1, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"Y", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}]}]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {0, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, $CellContext`polar}, {0, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, $CellContext`polar}, {1, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{2, $CellContext`polar}, {0, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{2, $CellContext`polar}, {1, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{2, $CellContext`polar}, {2, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
       Editable->False]}]}], "}"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728752734326337*^9, 3.728753097312365*^9, 
  3.728753201917984*^9, 3.730913333750079*^9},
 CellLabel->"Out[87]=",
 CellID->285575129,ExpressionUUID->"9cf0b7e9-6673-40d2-8e7a-03be4ac56308"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We can always collapse the FoldedRules, but this may generate very long \
expressions\
\>", "Text",
 CellID->1346723794,ExpressionUUID->"2246b99f-b656-43a8-8919-92d90bb0ca29"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "CollapseFoldedRule"}]], "Input",
 CellLabel->"In[88]:=",
 CellID->2056650025,ExpressionUUID->"b949a72c-4b42-4966-b344-4b8c676c31f7"],

Cell[BoxData["\<\"CollapseFoldedRule[frule, {n, m}] converts the elements n \
to m of the foldedrule frule into a single list of replacements in terms of \
the independent objects at level m. The positional argument follows the \
notation of Take: both n, m or just m can be negative, counting from the end; \
a single positive integer m represents {1, m}; a single negative integer -n \
represents {-n, -1}. A third argument can be given specifying a function to \
be mapped on all rules after collapsing a new level. \
CollapseFoldedRule[frule] or CollapseFoldedRule[frule, All] are interpreted \
as collapse of all elements of frule.\"\>"], "Print", "PrintUsage",
 CellChangeTimes->{3.730913336012438*^9},
 CellTags->"Info883730895335-7554244",
 CellID->696624812,ExpressionUUID->"c78b4dc7-d8d5-444f-8eb7-d91f40dd7a9d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"CollapseFoldedRule", "[", 
  RowBox[{"%%", ",", "All"}], "]"}]], "Input",
 CellLabel->"In[89]:=",
 CellID->586451060,ExpressionUUID->"067f70b0-8b82-4dbf-bf73-384d01d22817"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{0, -$CellContext`polar}, {0, -$CellContext`polar}],
      Editable->False], "\[Rule]", "0"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{1, -$CellContext`polar}, {0, -$CellContext`polar}],
      Editable->False], "\[Rule]", 
     RowBox[{
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          0, -$CellContext`polar}],
         Editable->False]}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"g", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`metric[{1, -$CellContext`polar}, {
            1, -$CellContext`polar}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
           Editable->False]}], "+", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"g", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`metric[{1, -$CellContext`polar}, {
            2, -$CellContext`polar}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
           Editable->False]}]}], ")"}]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{0, -$CellContext`polar}, {
         2, -$CellContext`polar}],
        Editable->False], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"g", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`metric[{0, -$CellContext`polar}, {
            1, -$CellContext`polar}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
           Editable->False]}], "+", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"g", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`metric[{1, -$CellContext`polar}, {
            1, -$CellContext`polar}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
           Editable->False]}]}], ")"}]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{0, -$CellContext`polar}, {
         1, -$CellContext`polar}],
        Editable->False], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", 
           InterpretationBox[
            StyleBox[GridBox[{
               {"g", 
                StyleBox[GridBox[{
                   {" ", " "},
                   {
                    StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]]}
                  },
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                 FontSize->9]}
              },
              
              GridBoxAlignment->{
               "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                "Rows" -> {{Center}}, "RowsIndexed" -> {}},
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.27999999999999997`], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                "Rows" -> {
                  Offset[0.2], {
                   Offset[0.4]}, 
                  Offset[0.2]}, "RowsIndexed" -> {}}],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`metric[{0, -$CellContext`polar}, {
             1, -$CellContext`polar}],
            Editable->False]}], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
           Editable->False]}], "+", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"g", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`metric[{1, -$CellContext`polar}, {
            2, -$CellContext`polar}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
           Editable->False]}]}], ")"}]}]}]}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{1, -$CellContext`polar}, {1, -$CellContext`polar}],
      Editable->False], "\[Rule]", "0"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{2, -$CellContext`polar}, {0, -$CellContext`polar}],
      Editable->False], "\[Rule]", 
     RowBox[{
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          0, -$CellContext`polar}],
         Editable->False]}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"g", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`metric[{1, -$CellContext`polar}, {
            2, -$CellContext`polar}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
           Editable->False]}], "+", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"g", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`metric[{2, -$CellContext`polar}, {
            2, -$CellContext`polar}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
           Editable->False]}]}], ")"}]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{0, -$CellContext`polar}, {
         2, -$CellContext`polar}],
        Editable->False], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", 
           InterpretationBox[
            StyleBox[GridBox[{
               {"g", 
                StyleBox[GridBox[{
                   {" ", " "},
                   {
                    StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                  },
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                 FontSize->9]}
              },
              
              GridBoxAlignment->{
               "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                "Rows" -> {{Center}}, "RowsIndexed" -> {}},
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.27999999999999997`], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                "Rows" -> {
                  Offset[0.2], {
                   Offset[0.4]}, 
                  Offset[0.2]}, "RowsIndexed" -> {}}],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`metric[{0, -$CellContext`polar}, {
             2, -$CellContext`polar}],
            Editable->False]}], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
           Editable->False]}], "-", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"g", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`metric[{1, -$CellContext`polar}, {
            2, -$CellContext`polar}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
           Editable->False]}]}], ")"}]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{0, -$CellContext`polar}, {
         1, -$CellContext`polar}],
        Editable->False], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", 
           InterpretationBox[
            StyleBox[GridBox[{
               {"g", 
                StyleBox[GridBox[{
                   {" ", " "},
                   {
                    StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                  },
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                 FontSize->9]}
              },
              
              GridBoxAlignment->{
               "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                "Rows" -> {{Center}}, "RowsIndexed" -> {}},
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.27999999999999997`], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                "Rows" -> {
                  Offset[0.2], {
                   Offset[0.4]}, 
                  Offset[0.2]}, "RowsIndexed" -> {}}],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`metric[{0, -$CellContext`polar}, {
             2, -$CellContext`polar}],
            Editable->False]}], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
           Editable->False]}], "+", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"g", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`metric[{2, -$CellContext`polar}, {
            2, -$CellContext`polar}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
           Editable->False]}]}], ")"}]}]}]}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{2, -$CellContext`polar}, {1, -$CellContext`polar}],
      Editable->False], "\[Rule]", 
     RowBox[{
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          1, -$CellContext`polar}],
         Editable->False]}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"g", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`metric[{1, -$CellContext`polar}, {
            2, -$CellContext`polar}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
           Editable->False]}], "+", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"g", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`metric[{2, -$CellContext`polar}, {
            2, -$CellContext`polar}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
           Editable->False]}]}], ")"}]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{1, -$CellContext`polar}, {
         2, -$CellContext`polar}],
        Editable->False], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", 
           InterpretationBox[
            StyleBox[GridBox[{
               {"g", 
                StyleBox[GridBox[{
                   {" ", " "},
                   {
                    StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                  },
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                 FontSize->9]}
              },
              
              GridBoxAlignment->{
               "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                "Rows" -> {{Center}}, "RowsIndexed" -> {}},
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.27999999999999997`], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                "Rows" -> {
                  Offset[0.2], {
                   Offset[0.4]}, 
                  Offset[0.2]}, "RowsIndexed" -> {}}],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`metric[{0, -$CellContext`polar}, {
             2, -$CellContext`polar}],
            Editable->False]}], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
           Editable->False]}], "-", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"g", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`metric[{1, -$CellContext`polar}, {
            2, -$CellContext`polar}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
           Editable->False]}]}], ")"}]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{1, -$CellContext`polar}, {
         1, -$CellContext`polar}],
        Editable->False], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", 
           InterpretationBox[
            StyleBox[GridBox[{
               {"g", 
                StyleBox[GridBox[{
                   {" ", " "},
                   {
                    StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                  },
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                 FontSize->9]}
              },
              
              GridBoxAlignment->{
               "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                "Rows" -> {{Center}}, "RowsIndexed" -> {}},
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.27999999999999997`], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                "Rows" -> {
                  Offset[0.2], {
                   Offset[0.4]}, 
                  Offset[0.2]}, "RowsIndexed" -> {}}],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`metric[{0, -$CellContext`polar}, {
             2, -$CellContext`polar}],
            Editable->False]}], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
           Editable->False]}], "+", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"g", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`metric[{2, -$CellContext`polar}, {
            2, -$CellContext`polar}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
           Editable->False]}]}], ")"}]}]}]}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{2, -$CellContext`polar}, {2, -$CellContext`polar}],
      Editable->False], "\[Rule]", "0"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
      Editable->False], "\[Rule]", 
     RowBox[{
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{0, -$CellContext`polar}, {
         0, -$CellContext`polar}],
        Editable->False], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"g", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`metric[{1, -$CellContext`polar}, {
            1, -$CellContext`polar}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
           Editable->False]}], "+", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"g", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`metric[{1, -$CellContext`polar}, {
            2, -$CellContext`polar}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
           Editable->False]}]}], ")"}]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{0, -$CellContext`polar}, {
         2, -$CellContext`polar}],
        Editable->False], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"g", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`metric[{0, -$CellContext`polar}, {
            1, -$CellContext`polar}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
           Editable->False]}], "+", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"g", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`metric[{1, -$CellContext`polar}, {
            1, -$CellContext`polar}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
           Editable->False]}]}], ")"}]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{0, -$CellContext`polar}, {
         1, -$CellContext`polar}],
        Editable->False], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", 
           InterpretationBox[
            StyleBox[GridBox[{
               {"g", 
                StyleBox[GridBox[{
                   {" ", " "},
                   {
                    StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                    StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]]}
                  },
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                 FontSize->9]}
              },
              
              GridBoxAlignment->{
               "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                "Rows" -> {{Center}}, "RowsIndexed" -> {}},
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.27999999999999997`], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                "Rows" -> {
                  Offset[0.2], {
                   Offset[0.4]}, 
                  Offset[0.2]}, "RowsIndexed" -> {}}],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`metric[{0, -$CellContext`polar}, {
             1, -$CellContext`polar}],
            Editable->False]}], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
           Editable->False]}], "+", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"g", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`metric[{1, -$CellContext`polar}, {
            2, -$CellContext`polar}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
           Editable->False]}]}], ")"}]}]}]}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{0, -$CellContext`polar}, {2, -$CellContext`polar}],
      Editable->False], "\[Rule]", 
     RowBox[{
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{0, -$CellContext`polar}, {
         0, -$CellContext`polar}],
        Editable->False], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"g", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`metric[{1, -$CellContext`polar}, {
            2, -$CellContext`polar}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
           Editable->False]}], "+", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"g", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`metric[{2, -$CellContext`polar}, {
            2, -$CellContext`polar}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
           Editable->False]}]}], ")"}]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{0, -$CellContext`polar}, {
         2, -$CellContext`polar}],
        Editable->False], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", 
           InterpretationBox[
            StyleBox[GridBox[{
               {"g", 
                StyleBox[GridBox[{
                   {" ", " "},
                   {
                    StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                  },
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                 FontSize->9]}
              },
              
              GridBoxAlignment->{
               "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                "Rows" -> {{Center}}, "RowsIndexed" -> {}},
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.27999999999999997`], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                "Rows" -> {
                  Offset[0.2], {
                   Offset[0.4]}, 
                  Offset[0.2]}, "RowsIndexed" -> {}}],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`metric[{0, -$CellContext`polar}, {
             2, -$CellContext`polar}],
            Editable->False]}], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
           Editable->False]}], "-", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"g", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`metric[{1, -$CellContext`polar}, {
            2, -$CellContext`polar}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
           Editable->False]}]}], ")"}]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{0, -$CellContext`polar}, {
         1, -$CellContext`polar}],
        Editable->False], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", 
           InterpretationBox[
            StyleBox[GridBox[{
               {"g", 
                StyleBox[GridBox[{
                   {" ", " "},
                   {
                    StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                  },
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                 FontSize->9]}
              },
              
              GridBoxAlignment->{
               "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                "Rows" -> {{Center}}, "RowsIndexed" -> {}},
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.27999999999999997`], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                "Rows" -> {
                  Offset[0.2], {
                   Offset[0.4]}, 
                  Offset[0.2]}, "RowsIndexed" -> {}}],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`metric[{0, -$CellContext`polar}, {
             2, -$CellContext`polar}],
            Editable->False]}], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
           Editable->False]}], "+", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"g", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`metric[{2, -$CellContext`polar}, {
            2, -$CellContext`polar}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
           Editable->False]}]}], ")"}]}]}]}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
      Editable->False], "\[Rule]", 
     RowBox[{
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{0, -$CellContext`polar}, {
         1, -$CellContext`polar}],
        Editable->False], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"g", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`metric[{1, -$CellContext`polar}, {
            2, -$CellContext`polar}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
           Editable->False]}], "+", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"g", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`metric[{2, -$CellContext`polar}, {
            2, -$CellContext`polar}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
           Editable->False]}]}], ")"}]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{1, -$CellContext`polar}, {
         2, -$CellContext`polar}],
        Editable->False], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", 
           InterpretationBox[
            StyleBox[GridBox[{
               {"g", 
                StyleBox[GridBox[{
                   {" ", " "},
                   {
                    StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                  },
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                 FontSize->9]}
              },
              
              GridBoxAlignment->{
               "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                "Rows" -> {{Center}}, "RowsIndexed" -> {}},
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.27999999999999997`], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                "Rows" -> {
                  Offset[0.2], {
                   Offset[0.4]}, 
                  Offset[0.2]}, "RowsIndexed" -> {}}],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`metric[{0, -$CellContext`polar}, {
             2, -$CellContext`polar}],
            Editable->False]}], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
           Editable->False]}], "-", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"g", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`metric[{1, -$CellContext`polar}, {
            2, -$CellContext`polar}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
           Editable->False]}]}], ")"}]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{1, -$CellContext`polar}, {
         1, -$CellContext`polar}],
        Editable->False], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", 
           InterpretationBox[
            StyleBox[GridBox[{
               {"g", 
                StyleBox[GridBox[{
                   {" ", " "},
                   {
                    StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                    StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                  },
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                 FontSize->9]}
              },
              
              GridBoxAlignment->{
               "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                "Rows" -> {{Center}}, "RowsIndexed" -> {}},
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.27999999999999997`], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                "Rows" -> {
                  Offset[0.2], {
                   Offset[0.4]}, 
                  Offset[0.2]}, "RowsIndexed" -> {}}],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`metric[{0, -$CellContext`polar}, {
             2, -$CellContext`polar}],
            Editable->False]}], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["0",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
           Editable->False]}], "+", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"g", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`metric[{2, -$CellContext`polar}, {
            2, -$CellContext`polar}],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"Y", 
               StyleBox[GridBox[{
                  {
                   StyleBox["1",
                    FontColor->RGBColor[1., 0., 0.]], 
                   StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]]},
                  {" ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
           Editable->False]}]}], ")"}]}]}]}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{1, $CellContext`polar}, {0, -$CellContext`polar}],
      Editable->False], "\[Rule]", 
     RowBox[{
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          0, -$CellContext`polar}],
         Editable->False]}], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
        Editable->False]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{0, -$CellContext`polar}, {
         2, -$CellContext`polar}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}]}]}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{2, $CellContext`polar}, {0, -$CellContext`polar}],
      Editable->False], "\[Rule]", 
     RowBox[{
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          0, -$CellContext`polar}],
         Editable->False]}], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{0, -$CellContext`polar}, {
         1, -$CellContext`polar}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}]}]}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{2, $CellContext`polar}, {1, -$CellContext`polar}],
      Editable->False], "\[Rule]", 
     RowBox[{
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          1, -$CellContext`polar}],
         Editable->False]}], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{1, -$CellContext`polar}, {
         1, -$CellContext`polar}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}]}]}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{1, -$CellContext`polar}, {1, $CellContext`polar}],
      Editable->False], "\[Rule]", 
     RowBox[{
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{0, -$CellContext`polar}, {
         1, -$CellContext`polar}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
        Editable->False]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{1, -$CellContext`polar}, {
         2, -$CellContext`polar}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}]}]}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{2, -$CellContext`polar}, {1, $CellContext`polar}],
      Editable->False], "\[Rule]", 
     RowBox[{
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{0, -$CellContext`polar}, {
         2, -$CellContext`polar}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
        Editable->False]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{2, -$CellContext`polar}, {
         2, -$CellContext`polar}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}]}]}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{2, -$CellContext`polar}, {2, $CellContext`polar}],
      Editable->False], "\[Rule]", 
     RowBox[{
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{0, -$CellContext`polar}, {
         2, -$CellContext`polar}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{1, -$CellContext`polar}, {
         2, -$CellContext`polar}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}]}]}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{0, $CellContext`polar}, {0, -$CellContext`polar}],
      Editable->False], "\[Rule]", 
     RowBox[{
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{0, -$CellContext`polar}, {
         1, -$CellContext`polar}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
        Editable->False]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{0, -$CellContext`polar}, {
         2, -$CellContext`polar}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}]}]}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{0, -$CellContext`polar}, {1, $CellContext`polar}],
      Editable->False], "\[Rule]", 
     RowBox[{
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{0, -$CellContext`polar}, {
         0, -$CellContext`polar}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
        Editable->False]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{0, -$CellContext`polar}, {
         2, -$CellContext`polar}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}]}]}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{0, -$CellContext`polar}, {2, $CellContext`polar}],
      Editable->False], "\[Rule]", 
     RowBox[{
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{0, -$CellContext`polar}, {
         0, -$CellContext`polar}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{0, -$CellContext`polar}, {
         1, -$CellContext`polar}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}]}]}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{0, $CellContext`polar}, {1, -$CellContext`polar}],
      Editable->False], "\[Rule]", 
     RowBox[{
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{1, -$CellContext`polar}, {
         1, -$CellContext`polar}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
        Editable->False]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{1, -$CellContext`polar}, {
         2, -$CellContext`polar}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}]}]}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{1, $CellContext`polar}, {1, -$CellContext`polar}],
      Editable->False], "\[Rule]", 
     RowBox[{
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          1, -$CellContext`polar}],
         Editable->False]}], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
        Editable->False]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{1, -$CellContext`polar}, {
         2, -$CellContext`polar}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}]}]}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{1, -$CellContext`polar}, {2, $CellContext`polar}],
      Editable->False], "\[Rule]", 
     RowBox[{
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{0, -$CellContext`polar}, {
         1, -$CellContext`polar}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{1, -$CellContext`polar}, {
         1, -$CellContext`polar}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}]}]}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{0, $CellContext`polar}, {2, -$CellContext`polar}],
      Editable->False], "\[Rule]", 
     RowBox[{
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{1, -$CellContext`polar}, {
         2, -$CellContext`polar}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
        Editable->False]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{2, -$CellContext`polar}, {
         2, -$CellContext`polar}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}]}]}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{1, $CellContext`polar}, {2, -$CellContext`polar}],
      Editable->False], "\[Rule]", 
     RowBox[{
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False]}], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
        Editable->False]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{2, -$CellContext`polar}, {
         2, -$CellContext`polar}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}]}]}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{2, $CellContext`polar}, {2, -$CellContext`polar}],
      Editable->False], "\[Rule]", 
     RowBox[{
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False]}], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`metric[{1, -$CellContext`polar}, {
         2, -$CellContext`polar}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Y", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}]}]}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{0, $CellContext`polar}, {0, $CellContext`polar}],
      Editable->False], "\[Rule]", "0"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{1, $CellContext`polar}, {0, $CellContext`polar}],
      Editable->False], "\[Rule]", 
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
       Editable->False]}]}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{1, $CellContext`polar}, {1, $CellContext`polar}],
      Editable->False], "\[Rule]", "0"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{2, $CellContext`polar}, {0, $CellContext`polar}],
      Editable->False], "\[Rule]", 
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
       Editable->False]}]}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{2, $CellContext`polar}, {1, $CellContext`polar}],
      Editable->False], "\[Rule]", 
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Y", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
       Editable->False]}]}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{2, $CellContext`polar}, {2, $CellContext`polar}],
      Editable->False], "\[Rule]", "0"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
      Editable->False], "\[Rule]", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{0, $CellContext`polar}, {1, $CellContext`polar}],
      Editable->False]}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
      Editable->False], "\[Rule]", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{0, $CellContext`polar}, {2, $CellContext`polar}],
      Editable->False]}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
      Editable->False], "\[Rule]", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"Y", 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Y[{1, $CellContext`polar}, {2, $CellContext`polar}],
      Editable->False]}]}], "}"}], "]"}]], "Output",
 CellChangeTimes->{3.7287527350282784`*^9, 3.728753097807616*^9, 
  3.7287532025489397`*^9, 3.730913337481655*^9},
 CellLabel->"Out[89]=",
 CellID->17621519,ExpressionUUID->"ecda3e8c-2d5a-4436-a78a-2867ef1e557b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CNumbersOf", "[", "polar", "]"}], " ", "^=", 
  RowBox[{"{", 
   RowBox[{"0", ",", "1", ",", "2"}], "}"}]}]], "Input",
 CellLabel->"In[90]:=",
 CellID->615501615,ExpressionUUID->"cc83a7c6-99b7-4648-8d64-23ad9581eafc"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0", ",", "1", ",", "2"}], "}"}]], "Output",
 CellChangeTimes->{3.7287527353103437`*^9, 3.728753097936363*^9, 
  3.728753202826962*^9, 3.7309133388489847`*^9},
 CellLabel->"Out[90]=",
 CellID->1121161639,ExpressionUUID->"db260a15-2e8c-488b-9c70-42b5c9f08189"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We can also change bases", "Text",
 CellID->76550648,ExpressionUUID->"d6d2aa8a-64a0-4d53-91f9-3400715db483"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ChangeComponents", "[", 
  RowBox[{
   RowBox[{"v", "[", 
    RowBox[{"{", 
     RowBox[{"a", ",", "cartesian"}], "}"}], "]"}], ",", 
   RowBox[{"v", "[", 
    RowBox[{"{", 
     RowBox[{"a", ",", "polar"}], "}"}], "]"}]}], "]"}]], "Input",
 CellLabel->"In[91]:=",
 CellID->244036383,ExpressionUUID->"9a8b3092-f620-4daf-9d53-262c60e460ff"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[{0, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[{0, $CellContext`polar}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "v"}],
  SequenceForm[
  "Added independent rule ", $CellContext`v[{
     0, $CellContext`polar}] -> $CellContext`v[{0, $CellContext`polar}], 
   " for tensor ", $CellContext`v],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752736172009*^9, 3.728753098104485*^9, 
  3.72875320393391*^9, 3.730913339872089*^9},
 CellLabel->"During evaluation of In[91]:=",
 CellID->1577509860,ExpressionUUID->"cd70a4ad-b925-4302-9629-3f7de89bee79"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[{1, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[{1, $CellContext`polar}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "v"}],
  SequenceForm[
  "Added independent rule ", $CellContext`v[{
     1, $CellContext`polar}] -> $CellContext`v[{1, $CellContext`polar}], 
   " for tensor ", $CellContext`v],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752736172009*^9, 3.728753098104485*^9, 
  3.72875320393391*^9, 3.730913339909061*^9},
 CellLabel->"During evaluation of In[91]:=",
 CellID->375111158,ExpressionUUID->"3dfe7013-5084-4d51-836a-e17e235ac1f3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[{2, $CellContext`polar}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[{2, $CellContext`polar}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "v"}],
  SequenceForm[
  "Added independent rule ", $CellContext`v[{
     2, $CellContext`polar}] -> $CellContext`v[{2, $CellContext`polar}], 
   " for tensor ", $CellContext`v],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752736172009*^9, 3.728753098104485*^9, 
  3.72875320393391*^9, 3.730913340017459*^9},
 CellLabel->"During evaluation of In[91]:=",
 CellID->1628728561,ExpressionUUID->"1c5c90ea-a2ec-4af1-ba24-51d0a0ecaa4b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[0., 1., 0.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[{0, $CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{0, -$CellContext`polar}, {
        0, $CellContext`cartesian}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{0, $CellContext`polar}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{1, -$CellContext`polar}, {
        0, $CellContext`cartesian}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{1, $CellContext`polar}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{2, -$CellContext`polar}, {
        0, $CellContext`cartesian}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{2, $CellContext`polar}],
       Editable->False]}]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "v"}],
  SequenceForm[
  "Added independent rule ", $CellContext`v[{0, $CellContext`cartesian}] -> 
   xAct`xTensor`Basis[{0, -$CellContext`polar}, {
       0, $CellContext`cartesian}] $CellContext`v[{0, $CellContext`polar}] + 
    xAct`xTensor`Basis[{1, -$CellContext`polar}, {
       0, $CellContext`cartesian}] $CellContext`v[{1, $CellContext`polar}] + 
    xAct`xTensor`Basis[{2, -$CellContext`polar}, {
       0, $CellContext`cartesian}] $CellContext`v[{2, $CellContext`polar}], 
   " for tensor ", $CellContext`v],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752736172009*^9, 3.728753098104485*^9, 
  3.72875320393391*^9, 3.730913340108189*^9},
 CellLabel->"During evaluation of In[91]:=",
 CellID->66876418,ExpressionUUID->"356f188d-bba5-4070-9e39-378dcbb0f209"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 1., 0.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[{1, $CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{0, -$CellContext`polar}, {
        1, $CellContext`cartesian}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{0, $CellContext`polar}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{1, -$CellContext`polar}, {
        1, $CellContext`cartesian}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{1, $CellContext`polar}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{2, -$CellContext`polar}, {
        1, $CellContext`cartesian}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{2, $CellContext`polar}],
       Editable->False]}]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "v"}],
  SequenceForm[
  "Added independent rule ", $CellContext`v[{1, $CellContext`cartesian}] -> 
   xAct`xTensor`Basis[{0, -$CellContext`polar}, {
       1, $CellContext`cartesian}] $CellContext`v[{0, $CellContext`polar}] + 
    xAct`xTensor`Basis[{1, -$CellContext`polar}, {
       1, $CellContext`cartesian}] $CellContext`v[{1, $CellContext`polar}] + 
    xAct`xTensor`Basis[{2, -$CellContext`polar}, {
       1, $CellContext`cartesian}] $CellContext`v[{2, $CellContext`polar}], 
   " for tensor ", $CellContext`v],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752736172009*^9, 3.728753098104485*^9, 
  3.72875320393391*^9, 3.730913340215366*^9},
 CellLabel->"During evaluation of In[91]:=",
 CellID->858009945,ExpressionUUID->"7b5563a3-3256-4d09-a3aa-ea3a70ffaa16"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 1., 0.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[{2, $CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{0, -$CellContext`polar}, {
        2, $CellContext`cartesian}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{0, $CellContext`polar}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{1, -$CellContext`polar}, {
        2, $CellContext`cartesian}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{1, $CellContext`polar}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{2, -$CellContext`polar}, {
        2, $CellContext`cartesian}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{2, $CellContext`polar}],
       Editable->False]}]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "v"}],
  SequenceForm[
  "Added independent rule ", $CellContext`v[{2, $CellContext`cartesian}] -> 
   xAct`xTensor`Basis[{0, -$CellContext`polar}, {
       2, $CellContext`cartesian}] $CellContext`v[{0, $CellContext`polar}] + 
    xAct`xTensor`Basis[{1, -$CellContext`polar}, {
       2, $CellContext`cartesian}] $CellContext`v[{1, $CellContext`polar}] + 
    xAct`xTensor`Basis[{2, -$CellContext`polar}, {
       2, $CellContext`cartesian}] $CellContext`v[{2, $CellContext`polar}], 
   " for tensor ", $CellContext`v],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752736172009*^9, 3.728753098104485*^9, 
  3.72875320393391*^9, 3.730913340308655*^9},
 CellLabel->"During evaluation of In[91]:=",
 CellID->220456010,ExpressionUUID->"de4e363c-9266-4b1a-ac3c-97bbf3cee9bc"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Computed \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {
             StyleBox["a",
              FontColor->RGBColor[0., 1., 0.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[{$CellContext`a, $CellContext`cartesian}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["a",
               FontColor->RGBColor[0., 1., 0.]]},
             {
              StyleBox["b",
               FontColor->RGBColor[1., 0., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{-$CellContext`b, -$CellContext`polar}, \
{$CellContext`a, $CellContext`cartesian}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {
              StyleBox["b",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[{$CellContext`b, $CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" in \"\>", 
   "\[InvisibleSpace]", "0.304724`5.935451654010397", 
   "\[InvisibleSpace]", "\<\" Seconds\"\>"}],
  SequenceForm[
  "Computed ", $CellContext`v[{$CellContext`a, $CellContext`cartesian}] -> 
   xAct`xTensor`Basis[{-$CellContext`b, -$CellContext`polar}, \
{$CellContext`a, $CellContext`cartesian}] $CellContext`v[{$CellContext`b, \
$CellContext`polar}], " in ", 0.304724`5.935451654010397, " Seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728752736172009*^9, 3.728753098104485*^9, 
  3.72875320393391*^9, 3.730913340408618*^9},
 CellLabel->"During evaluation of In[91]:=",
 CellID->794868995,ExpressionUUID->"d3aa05a0-8368-4ffe-b9e9-cab317408d22"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[0., 1., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{0, $CellContext`cartesian}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["0",
                  FontColor->RGBColor[0., 1., 0.]]},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{0, -$CellContext`polar}, {
          0, $CellContext`cartesian}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{0, $CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["0",
                  FontColor->RGBColor[0., 1., 0.]]},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{1, -$CellContext`polar}, {
          0, $CellContext`cartesian}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{1, $CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["0",
                  FontColor->RGBColor[0., 1., 0.]]},
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{2, -$CellContext`polar}, {
          0, $CellContext`cartesian}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{2, $CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 1., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{1, $CellContext`cartesian}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["1",
                  FontColor->RGBColor[0., 1., 0.]]},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{0, -$CellContext`polar}, {
          1, $CellContext`cartesian}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{0, $CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["1",
                  FontColor->RGBColor[0., 1., 0.]]},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{1, -$CellContext`polar}, {
          1, $CellContext`cartesian}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{1, $CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["1",
                  FontColor->RGBColor[0., 1., 0.]]},
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{2, -$CellContext`polar}, {
          1, $CellContext`cartesian}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{2, $CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 1., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[{2, $CellContext`cartesian}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[0., 1., 0.]]},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{0, -$CellContext`polar}, {
          2, $CellContext`cartesian}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{0, $CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[0., 1., 0.]]},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{1, -$CellContext`polar}, {
          2, $CellContext`cartesian}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{1, $CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"e", 
             StyleBox[GridBox[{
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[0., 1., 0.]]},
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]], " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`Basis[{2, -$CellContext`polar}, {
          2, $CellContext`cartesian}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[{2, $CellContext`polar}],
         Editable->False]}]}]}]}], "}"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728752737351026*^9, 3.7287530986351233`*^9, 
  3.728753204766474*^9, 3.730913340511403*^9},
 CellLabel->"Out[91]=",
 CellID->2014914625,ExpressionUUID->"c8a97213-c1bc-4fcc-bfc6-283593b5682d"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["4.10. Replacing TensorValues in expressions", "Subsection",
 CellChangeTimes->{{3.7286430157211113`*^9, 3.728643016017661*^9}, {
  3.728673968932828*^9, 3.7286739705964937`*^9}, {3.7286752930712357`*^9, 
  3.728675293199355*^9}},
 CellID->1858146570,ExpressionUUID->"cccf106d-4bf8-4bca-a08e-ed5bdb322dcb"],

Cell[TextData[{
 StyleBox["ToValues\t\t", "DisplayMath"],
 StyleBox["Replace tensor values in a given exp", "Text"],
 StyleBox["ression", "Text"]
}], "DefinitionBox",
 CellID->1162207316,ExpressionUUID->"f55e0c65-2e83-4c83-aa1b-666b2bfa2854"],

Cell["Replacing TensorValues in expressions.", "Text",
 CellChangeTimes->{3.728751849118587*^9},
 CellID->1854590547,ExpressionUUID->"efe7940c-cd3b-43fb-aff6-a6326f0f4bd4"],

Cell[CellGroupData[{

Cell[TextData[{
 "As we have seen, ",
 StyleBox["xCoba`", "MR"],
 " does not provide an automatic tool for setting values (use ",
 StyleBox["Set", "MR"],
 " or ",
 StyleBox["SetDelayed", "MR"],
 " instead of rules), because this is potentially dangerous and removing the \
definitions would not be trivial. However, it is very simple to apply all the \
rules in ",
 StyleBox["TensorValues", "MR"],
 " to a given expression"
}], "Text",
 CellID->1525670871,ExpressionUUID->"f8c2c11c-8a93-49ef-8504-bd7e25d8e28f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "ToValues"}]], "Input",
 CellLabel->"In[92]:=",
 CellID->995583717,ExpressionUUID->"87fd8e9c-b191-469d-ac01-67adc708b70f"],

Cell[BoxData["\<\"ToValues[expr] returns expr after replacing all known \
tensor-values for the tensors in the expression. ToValues[expr, tensor] \
replaces only values for the given tensor. ToValues[expr, list] replaces \
values for the tensors in the list. ToValues[expr, list, f] applies the \
function f after replacement of each of the tensors in the list.\"\>"], \
"Print", "PrintUsage",
 CellChangeTimes->{3.7309133472411127`*^9},
 CellTags->"Info923730895347-7554244",
 CellID->821685044,ExpressionUUID->"57169d80-85a2-4839-8e86-b95b76f73bed"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TensorValues", "[", 
  RowBox[{"T", ",", 
   RowBox[{"{", 
    RowBox[{"{", 
     RowBox[{"polar", ",", "polar"}], "}"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[93]:=",
 CellID->97779173,ExpressionUUID->"d6ac9cb7-8eaf-4d4c-85ff-7ee2b1b11987"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, $CellContext`polar}, {0, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, $CellContext`polar}, {0, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, $CellContext`polar}, {1, $CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, $CellContext`polar}, {1, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, $CellContext`polar}, {0, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, $CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, $CellContext`polar}, {1, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{1, $CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, $CellContext`polar}, {2, $CellContext`polar}],
       Editable->False], "\[Rule]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, $CellContext`polar}, {1, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", "3"}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, $CellContext`polar}, {2, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", "4"}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, $CellContext`polar}, {2, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", "5"}]}]}], "}"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728752741567914*^9, 3.728753098989168*^9, 
  3.728753208501568*^9, 3.730913347376828*^9},
 CellLabel->"Out[93]=",
 CellID->1740684298,ExpressionUUID->"f30665af-b36b-4f1b-9881-8a75ae1963da"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"$CVVerbose", "=", "False"}], ";"}]], "Input",
 CellLabel->"In[94]:=",
 CellID->306849645,ExpressionUUID->"17272b76-fb79-421c-b796-7e9625041952"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ChangeComponents", "[", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{
     RowBox[{"-", 
      RowBox[{"{", 
       RowBox[{"a", ",", "polar"}], "}"}]}], ",", 
     RowBox[{"-", 
      RowBox[{"{", 
       RowBox[{"b", ",", "polar"}], "}"}]}]}], "]"}], ",", 
   RowBox[{"T", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"a", ",", "polar"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"b", ",", "polar"}], "}"}]}], "]"}]}], "]"}]], "Input",
 CellLabel->"In[95]:=",
 CellID->1113747068,ExpressionUUID->"7dd535fa-b9ec-4bd4-a89a-714bdfb79604"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Computed \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {
             StyleBox["a",
              FontColor->RGBColor[1., 0., 0.]], " "},
            {" ", 
             StyleBox["b",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{$CellContext`a, $CellContext`polar}, {-$CellContext`b, \
-$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["c",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["b",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`metric[{-$CellContext`c, -$CellContext`polar}, \
{-$CellContext`b, -$CellContext`polar}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {
              StyleBox["a",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["c",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{$CellContext`a, $CellContext`polar}, {$CellContext`c, \
$CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" in \"\>", 
   "\[InvisibleSpace]", "0.040299`5.056839262965366", 
   "\[InvisibleSpace]", "\<\" Seconds\"\>"}],
  SequenceForm[
  "Computed ", $CellContext`T[{$CellContext`a, $CellContext`polar}, \
{-$CellContext`b, -$CellContext`polar}] -> \
$CellContext`metric[{-$CellContext`c, -$CellContext`polar}, {-$CellContext`b, \
-$CellContext`polar}] $CellContext`T[{$CellContext`a, $CellContext`polar}, \
{$CellContext`c, $CellContext`polar}], " in ", 0.040299`5.056839262965366, 
   " Seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527421845217`*^9, 3.7287530991773252`*^9, 
  3.7287532087962914`*^9, 3.730913347626896*^9},
 CellLabel->"During evaluation of In[95]:=",
 CellID->511273405,ExpressionUUID->"514f4e98-8b53-4303-933c-37a11f120432"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Computed \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["a",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["b",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[{-$CellContext`a, -$CellContext`polar}, {-$CellContext`b, \
-$CellContext`polar}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["c",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["a",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`metric[{-$CellContext`c, -$CellContext`polar}, \
{-$CellContext`a, -$CellContext`polar}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {
              StyleBox["c",
               FontColor->RGBColor[1., 0., 0.]], " "},
             {" ", 
              StyleBox["b",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[{$CellContext`c, $CellContext`polar}, {-$CellContext`b, \
-$CellContext`polar}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" in \"\>", 
   "\[InvisibleSpace]", "0.041522`5.069823257605274", 
   "\[InvisibleSpace]", "\<\" Seconds\"\>"}],
  SequenceForm[
  "Computed ", $CellContext`T[{-$CellContext`a, -$CellContext`polar}, \
{-$CellContext`b, -$CellContext`polar}] -> \
$CellContext`metric[{-$CellContext`c, -$CellContext`polar}, {-$CellContext`a, \
-$CellContext`polar}] $CellContext`T[{$CellContext`c, $CellContext`polar}, \
{-$CellContext`b, -$CellContext`polar}], " in ", 0.041522`5.069823257605274, 
   " Seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.7287527421845217`*^9, 3.7287530991773252`*^9, 
  3.7287532087962914`*^9, 3.730913347675336*^9},
 CellLabel->"During evaluation of In[95]:=",
 CellID->890046188,ExpressionUUID->"dca44d0d-b6de-428f-8f46-77c16f54d5be"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`polar}, {2, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", " "},
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", "5"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", "6"}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          0, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], " "},
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, $CellContext`polar}, {2, -$CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          1, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], " "},
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, $CellContext`polar}, {2, -$CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]], " "},
                {" ", 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{2, $CellContext`polar}, {2, -$CellContext`polar}],
         Editable->False]}]}]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, $CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`polar}, {1, $CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, $CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, -$CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, $CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]},
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{1, -$CellContext`polar}, {2, $CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`polar}, {0, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], " "},
               {" ", 
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, $CellContext`polar}, {0, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`polar}, {0, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], " "},
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, $CellContext`polar}, {1, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`polar}, {1, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], " "},
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{1, $CellContext`polar}, {1, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`polar}, {0, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], " "},
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{0, $CellContext`polar}, {2, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`polar}, {1, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], " "},
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{1, $CellContext`polar}, {2, -$CellContext`polar}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, -$CellContext`polar}, {2, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]], " "},
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[{2, $CellContext`polar}, {2, -$CellContext`polar}],
        Editable->False]}]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, $CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          1, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, $CellContext`polar}, {1, $CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`polar}, {1, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          0, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, $CellContext`polar}, {1, $CellContext`polar}],
         Editable->False]}], "-", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, -$CellContext`polar}, {2, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          0, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          1, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, $CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{1, -$CellContext`polar}, {
          1, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, $CellContext`polar}, {1, $CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{1, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, $CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"g", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["0",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["1",
                   FontColor->RGBColor[1., 0., 0.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`metric[{0, -$CellContext`polar}, {
           1, -$CellContext`polar}],
          Editable->False]}], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, $CellContext`polar}, {1, $CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{1, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, -$CellContext`polar}, {2, $CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{0, -$CellContext`polar}, {
          1, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{1, -$CellContext`polar}, {
          1, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{0, $CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{1, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, $CellContext`polar}, {1, $CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{2, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{1, $CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"g", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["0",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["2",
                   FontColor->RGBColor[1., 0., 0.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`metric[{0, -$CellContext`polar}, {
           2, -$CellContext`polar}],
          Editable->False]}], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, $CellContext`polar}, {1, $CellContext`polar}],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{2, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[{2, $CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"g", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["0",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["2",
                   FontColor->RGBColor[1., 0., 0.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`metric[{0, -$CellContext`polar}, {
           2, -$CellContext`polar}],
          Editable->False]}], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{0, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}], "-", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"g", 
             StyleBox[GridBox[{
                {" ", " "},
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`metric[{1, -$CellContext`polar}, {
          2, -$CellContext`polar}],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["2",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[{1, $CellContext`polar}, {2, $CellContext`polar}],
         Editable->False]}]}]}]}], "}"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728752742448732*^9, 3.7287530993316193`*^9, 
  3.72875320899403*^9, 3.730913347784564*^9},
 CellLabel->"Out[95]=",
 CellID->1782680361,ExpressionUUID->"a6699ddd-d6b0-41e7-89ef-7f55c8ccf000"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"T", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"a", ",", "polar"}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{"b", ",", "polar"}], "}"}]}], "]"}], 
  RowBox[{"T", "[", 
   RowBox[{
    RowBox[{"-", 
     RowBox[{"{", 
      RowBox[{"a", ",", "polar"}], "}"}]}], ",", 
    RowBox[{"-", 
     RowBox[{"{", 
      RowBox[{"b", ",", "polar"}], "}"}]}]}], "]"}]}]], "Input",
 CellLabel->"In[96]:=",
 CellID->1433563111,ExpressionUUID->"201afdce-a6d8-48bc-bd23-d13ba4c40741"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {" ", " "},
          {
           StyleBox["a",
            FontColor->RGBColor[1., 0., 0.]], 
           StyleBox["b",
            FontColor->RGBColor[1., 0., 0.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[{-$CellContext`a, -$CellContext`polar}, {-$CellContext`b, \
-$CellContext`polar}],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {
           StyleBox["a",
            FontColor->RGBColor[1., 0., 0.]], 
           StyleBox["b",
            FontColor->RGBColor[1., 0., 0.]]},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[{$CellContext`a, $CellContext`polar}, {$CellContext`b, \
$CellContext`polar}],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728752742650653*^9, 3.7287530994752817`*^9, 
  3.728753209661407*^9, 3.730913347943734*^9},
 CellLabel->"Out[96]=",
 CellID->1964468030,ExpressionUUID->"1826a711-6d02-4668-84a2-ce202379bd80"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TraceBasisDummy", "[", "%", "]"}]], "Input",
 CellLabel->"In[97]:=",
 CellID->1684864799,ExpressionUUID->"3d67136d-aa76-4b71-b347-56754672769a"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{0, -$CellContext`polar}, {0, -$CellContext`polar}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{0, $CellContext`polar}, {0, $CellContext`polar}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{0, $CellContext`polar}, {1, $CellContext`polar}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{0, -$CellContext`polar}, {2, -$CellContext`polar}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{0, $CellContext`polar}, {2, $CellContext`polar}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{1, -$CellContext`polar}, {0, -$CellContext`polar}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{1, $CellContext`polar}, {0, $CellContext`polar}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{1, -$CellContext`polar}, {1, -$CellContext`polar}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{1, $CellContext`polar}, {1, $CellContext`polar}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{1, -$CellContext`polar}, {2, -$CellContext`polar}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{1, $CellContext`polar}, {2, $CellContext`polar}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{2, -$CellContext`polar}, {0, -$CellContext`polar}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{2, $CellContext`polar}, {0, $CellContext`polar}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{2, -$CellContext`polar}, {1, -$CellContext`polar}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{2, $CellContext`polar}, {1, $CellContext`polar}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{2, -$CellContext`polar}, {2, -$CellContext`polar}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[{2, $CellContext`polar}, {2, $CellContext`polar}],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728752742866024*^9, 3.728753210151045*^9, 
  3.730913348067165*^9},
 CellLabel->"Out[97]=",
 CellID->1400591232,ExpressionUUID->"f24d3c0e-d620-4f8a-89b0-90d1a3941c85"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "That expression has only the tensor ",
 StyleBox["T", "MR"],
 " and so we get"
}], "MathCaption",
 CellID->2106123915,ExpressionUUID->"13c59503-32d8-4b8a-898d-be9911676ef0"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToValues", "[", "%", "]"}]], "Input",
 CellLabel->"In[98]:=",
 CellID->651685904,ExpressionUUID->"90be1009-d2c0-4d45-846e-684b9e0c22f4"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", "14"}], "-", 
  RowBox[{"8", " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {2, -$CellContext`polar}],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"4", " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"g", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["0",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["2",
                   FontColor->RGBColor[1., 0., 0.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`metric[{0, -$CellContext`polar}, {
           2, -$CellContext`polar}],
          Editable->False]}], "+", 
        RowBox[{"5", " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"g", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["1",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["2",
                   FontColor->RGBColor[1., 0., 0.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`metric[{1, -$CellContext`polar}, {
           2, -$CellContext`polar}],
          Editable->False]}]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {1, -$CellContext`polar}],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"3", " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"g", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["0",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["2",
                   FontColor->RGBColor[1., 0., 0.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`metric[{0, -$CellContext`polar}, {
           2, -$CellContext`polar}],
          Editable->False]}], "-", 
        RowBox[{"5", " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"g", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["2",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["2",
                   FontColor->RGBColor[1., 0., 0.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`metric[{2, -$CellContext`polar}, {
           2, -$CellContext`polar}],
          Editable->False]}]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[{0, -$CellContext`polar}, {0, -$CellContext`polar}],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "3"}], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"g", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["1",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["2",
                   FontColor->RGBColor[1., 0., 0.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`metric[{1, -$CellContext`polar}, {
           2, -$CellContext`polar}],
          Editable->False]}], "-", 
        RowBox[{"4", " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"g", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {
                  StyleBox["2",
                   FontColor->RGBColor[1., 0., 0.]], 
                  StyleBox["2",
                   FontColor->RGBColor[1., 0., 0.]]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`metric[{2, -$CellContext`polar}, {
           2, -$CellContext`polar}],
          Editable->False]}]}], ")"}]}]}], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728752743041409*^9, 3.7287532103876667`*^9, 
  3.730913348225395*^9},
 CellLabel->"Out[98]=",
 CellID->1168581457,ExpressionUUID->"66d45648-72aa-4912-80f3-529399e04e20"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Now the metric tensors are explicit and hence:", "MathCaption",
 CellID->807968762,ExpressionUUID->"3735e57a-635a-40df-a0cc-5ac128bda534"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToValues", "[", "%", "]"}]], "Input",
 CellLabel->"In[99]:=",
 CellID->991882953,ExpressionUUID->"6cc39a29-c4ea-4835-bd84-55f7659d3348"],

Cell[BoxData["82"], "Output",
 CellChangeTimes->{3.7287527432323513`*^9, 3.728753210599719*^9, 
  3.730913348342381*^9},
 CellLabel->"Out[99]=",
 CellID->1610595984,ExpressionUUID->"84b40fd2-71bd-4026-b45a-4dabeb2c0721"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["4.11. Clean up", "Subsection",
 CellChangeTimes->{{3.7286430157211113`*^9, 3.728643016017661*^9}, {
  3.728673968932828*^9, 3.7286739705964937`*^9}, {3.7286752930712357`*^9, 
  3.728675293199355*^9}, {3.728752965093289*^9, 3.728752968211376*^9}},
 CellID->941636722,ExpressionUUID->"81303af2-2721-422a-8bc1-c362230067a0"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefMetric", "[", "metric", "]"}]], "Input",
 CellChangeTimes->{{3.7287532189549913`*^9, 3.728753224623734*^9}},
 CellLabel->"In[100]:=",
 CellID->1413614805,ExpressionUUID->"2d852c30-2fe0-4104-97d1-570a2ea4c50a"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Detmetric"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "weight +2 density", 
   " ", $CellContext`Detmetric],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753225674099*^9, 3.730913353912586*^9},
 CellLabel->"During evaluation of In[100]:=",
 CellID->894097473,ExpressionUUID->"4ec1466e-af06-46ad-8a2e-cf9b8bd6b75a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Christoffelcd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Christoffel tensor", " ", $CellContext`Christoffelcd],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753225674099*^9, 3.730913353926775*^9},
 CellLabel->"During evaluation of In[100]:=",
 CellID->1153150652,ExpressionUUID->"afde95b3-19a4-4d39-8afc-bb8b96ca9947"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Einstein tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Einsteincd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Einstein tensor", " ", $CellContext`Einsteincd],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753225674099*^9, 3.730913353940855*^9},
 CellLabel->"During evaluation of In[100]:=",
 CellID->2110577636,ExpressionUUID->"df914530-7842-4cb1-bc71-6bc42b4b7876"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Kretschmann scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Kretschmanncd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Kretschmann scalar", 
   " ", $CellContext`Kretschmanncd],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753225674099*^9, 3.7309133539516573`*^9},
 CellLabel->"During evaluation of In[100]:=",
 CellID->343447106,ExpressionUUID->"6fec2091-e9e5-4514-90e1-c2d48d4e1a0f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Riccicd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric Ricci tensor", 
   " ", $CellContext`Riccicd],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753225674099*^9, 3.730913353962109*^9},
 CellLabel->"During evaluation of In[100]:=",
 CellID->2026951053,ExpressionUUID->"f2e54da9-05bb-4da4-8bd6-ac612049b1fb"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciScalarcd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Ricci scalar", 
   " ", $CellContext`RicciScalarcd],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753225674099*^9, 3.7309133539721737`*^9},
 CellLabel->"During evaluation of In[100]:=",
 CellID->1433255410,ExpressionUUID->"6de2586e-353c-4ca9-b9ed-686219efdfe7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Riemanncd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`Riemanncd],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753225674099*^9, 3.73091335398216*^9},
 CellLabel->"During evaluation of In[100]:=",
 CellID->399607471,ExpressionUUID->"7151d2f5-0ebd-4615-82a1-b6470b719733"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric TFRicci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TFRiccicd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric TFRicci tensor",
    " ", $CellContext`TFRiccicd],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753225674099*^9, 3.73091335399216*^9},
 CellLabel->"During evaluation of In[100]:=",
 CellID->1021400928,ExpressionUUID->"22087298-e88e-49d3-8219-74f317cac93c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Torsioncd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`Torsioncd],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753225674099*^9, 3.73091335400253*^9},
 CellLabel->"During evaluation of In[100]:=",
 CellID->1848606425,ExpressionUUID->"f192623d-d32f-43ae-8a37-7f284f13db2f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Weyl tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Weylcd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Weyl tensor", 
   " ", $CellContext`Weylcd],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753225674099*^9, 3.730913354012598*^9},
 CellLabel->"During evaluation of In[100]:=",
 CellID->1882263826,ExpressionUUID->"1e8e7517-234e-4741-943f-47f9eb8be6d3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "cd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "covariant derivative", 
   " ", $CellContext`cd],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753225674099*^9, 3.7309133540224133`*^9},
 CellLabel->"During evaluation of In[100]:=",
 CellID->626519625,ExpressionUUID->"ccb0354b-baec-4a68-b7f4-506ed81147cd"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "epsilonmetric"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric tensor", 
   " ", $CellContext`epsilonmetric],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753225674099*^9, 3.730913354031824*^9},
 CellLabel->"During evaluation of In[100]:=",
 CellID->803016134,ExpressionUUID->"099d829c-4650-4e63-97c6-aea2896d1b75"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric metric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "metric"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric metric tensor", 
   " ", $CellContext`metric],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753225674099*^9, 3.730913354041662*^9},
 CellLabel->"During evaluation of In[100]:=",
 CellID->1504926390,ExpressionUUID->"6e2ef28e-c182-438b-a2a5-5e4a85c88dd6"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"UndefTensor", "/@", 
   RowBox[{"{", 
    RowBox[{"v", ",", "T", ",", "W", ",", "Y"}], "}"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.728752987719248*^9, 3.7287530033859253`*^9}},
 CellLabel->"In[101]:=",
 CellID->654448890,ExpressionUUID->"338b4ffa-ea39-4859-ac7a-e68a16addb55"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "v"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`v],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753003887903*^9, 3.7287532281231833`*^9, 
  3.7309133542266703`*^9},
 CellLabel->"During evaluation of In[101]:=",
 CellID->235442364,ExpressionUUID->"b771ac61-6a37-492c-95be-8245c3c8a434"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753003887903*^9, 3.7287532281231833`*^9, 
  3.730913354241859*^9},
 CellLabel->"During evaluation of In[101]:=",
 CellID->1414933974,ExpressionUUID->"6e8ec85b-7850-493d-8613-b3fdfa0f9bfa"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753003887903*^9, 3.7287532281231833`*^9, 
  3.730913354256259*^9},
 CellLabel->"During evaluation of In[101]:=",
 CellID->1708877787,ExpressionUUID->"cc52f977-3e6e-4470-bd39-822d81ee65be"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753003887903*^9, 3.7287532281231833`*^9, 
  3.730913354267712*^9},
 CellLabel->"During evaluation of In[101]:=",
 CellID->662671131,ExpressionUUID->"870d2000-b5bc-43bf-a1e3-a2ccdde15962"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"UndefBasis", "/@", 
   RowBox[{"{", 
    RowBox[{"polar", ",", "cartesian"}], "}"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.728753233467108*^9, 3.728753245646999*^9}},
 CellLabel->"In[102]:=",
 CellID->538418245,ExpressionUUID->"d75fa733-1d3c-427f-856f-fe5bca57dd76"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "ChristoffelPDpolar"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "non-symmetric Christoffel tensor", " ", $CellContext`ChristoffelPDpolar],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753246150777*^9, 3.730913354383564*^9},
 CellLabel->"During evaluation of In[102]:=",
 CellID->663505375,ExpressionUUID->"069c4f47-d436-41a9-93f0-1af4b64b117c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciPDpolar"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Ricci tensor", 
   " ", $CellContext`RicciPDpolar],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753246150777*^9, 3.73091335439925*^9},
 CellLabel->"During evaluation of In[102]:=",
 CellID->1799742401,ExpressionUUID->"1c83e161-510a-4b4f-ab50-59e93011014f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannPDpolar"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannPDpolar],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753246150777*^9, 3.730913354413699*^9},
 CellLabel->"During evaluation of In[102]:=",
 CellID->963365612,ExpressionUUID->"d10abd13-3f0f-406f-9393-351e894921bd"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TorsionPDpolar"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionPDpolar],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753246150777*^9, 3.730913354421852*^9},
 CellLabel->"During evaluation of In[102]:=",
 CellID->1406626151,ExpressionUUID->"d3de7c50-4506-48b0-a745-f259e90a84f7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"parallel derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "PDpolar"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "parallel derivative", 
   " ", $CellContext`PDpolar],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753246150777*^9, 3.730913354432023*^9},
 CellLabel->"During evaluation of In[102]:=",
 CellID->1665733681,ExpressionUUID->"50a875e1-047a-4187-9143-8e0100556ddd"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "etaUppolar"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric +1 density",
    " ", $CellContext`etaUppolar],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753246150777*^9, 3.7309133544399557`*^9},
 CellLabel->"During evaluation of In[102]:=",
 CellID->1389154275,ExpressionUUID->"1881041a-b205-4893-905f-493cc5163967"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "etaDownpolar"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric -1 density",
    " ", $CellContext`etaDownpolar],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753246150777*^9, 3.7309133544477787`*^9},
 CellLabel->"During evaluation of In[102]:=",
 CellID->73689741,ExpressionUUID->"d139ea8f-89c4-4337-a581-65034f8341f6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefBasis", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"basis\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "polar"}],
  SequenceForm[
  "** ", xAct`xCoba`UndefBasis, ": Undefined ", "basis", 
   " ", $CellContext`polar],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753246150777*^9, 3.730913354455235*^9},
 CellLabel->"During evaluation of In[102]:=",
 CellID->1552896315,ExpressionUUID->"9cb7030b-3910-4b1a-905a-24e9dafadda6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "ChristoffelPDcartesian"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "non-symmetric Christoffel tensor", 
   " ", $CellContext`ChristoffelPDcartesian],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753246150777*^9, 3.730913354463126*^9},
 CellLabel->"During evaluation of In[102]:=",
 CellID->1159604214,ExpressionUUID->"1fa70996-23e2-4d2a-b87e-62175a6936b6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciPDcartesian"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Ricci tensor", 
   " ", $CellContext`RicciPDcartesian],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753246150777*^9, 3.730913354471382*^9},
 CellLabel->"During evaluation of In[102]:=",
 CellID->1053928173,ExpressionUUID->"ab566967-3753-4cce-a33f-28de5671b144"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "RiemannPDcartesian"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannPDcartesian],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753246150777*^9, 3.730913354479043*^9},
 CellLabel->"During evaluation of In[102]:=",
 CellID->925683155,ExpressionUUID->"35348090-bb3b-430d-bf45-6815d1b3f564"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "TorsionPDcartesian"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionPDcartesian],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753246150777*^9, 3.730913354488799*^9},
 CellLabel->"During evaluation of In[102]:=",
 CellID->1585247217,ExpressionUUID->"bfc13289-bda0-47ad-80fd-4a1e60bae190"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"parallel derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "PDcartesian"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "parallel derivative", 
   " ", $CellContext`PDcartesian],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753246150777*^9, 3.730913354498846*^9},
 CellLabel->"During evaluation of In[102]:=",
 CellID->1567851437,ExpressionUUID->"6f4374f8-d602-4bd2-a791-bb41b95d24a4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "etaUpcartesian"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric +1 density",
    " ", $CellContext`etaUpcartesian],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753246150777*^9, 3.730913354508615*^9},
 CellLabel->"During evaluation of In[102]:=",
 CellID->1254752969,ExpressionUUID->"c3fca97a-a07c-4f06-a193-7e60af97db22"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "etaDowncartesian"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric -1 density",
    " ", $CellContext`etaDowncartesian],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753246150777*^9, 3.730913354518367*^9},
 CellLabel->"During evaluation of In[102]:=",
 CellID->343736443,ExpressionUUID->"04fb5b27-ad26-46ed-b518-13bd5d76de44"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefBasis", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"basis\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "cartesian"}],
  SequenceForm[
  "** ", xAct`xCoba`UndefBasis, ": Undefined ", "basis", 
   " ", $CellContext`cartesian],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753246150777*^9, 3.730913354528143*^9},
 CellLabel->"During evaluation of In[102]:=",
 CellID->447387071,ExpressionUUID->"528459a7-d166-4db8-9c8f-73b972ab9459"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M3", "]"}]], "Input",
 CellChangeTimes->{{3.7287532658340673`*^9, 3.728753270575151*^9}},
 CellLabel->"In[103]:=",
 CellID->457034461,ExpressionUUID->"ce1ea951-59b5-4d3a-9936-2537a0343628"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefVBundle", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]M3",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentM3,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefVBundle, ": Undefined ", "vbundle", 
   " ", $CellContext`TangentM3],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753271057877*^9, 3.730913354751047*^9},
 CellLabel->"During evaluation of In[103]:=",
 CellID->894254150,ExpressionUUID->"136cf734-e002-43ac-a9dc-187a06294d55"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefManifold", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"manifold\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "M3"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefManifold, ": Undefined ", "manifold", 
   " ", $CellContext`M3],
  Editable->False]], "Print",
 CellChangeTimes->{3.728753271057877*^9, 3.730913354768711*^9},
 CellLabel->"During evaluation of In[103]:=",
 CellID->1638937344,ExpressionUUID->"81ac4b33-f27e-4008-b347-b3d5d6460da6"]
}, Open  ]]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{
  "i", ",", "j", ",", "t", ",", "x", ",", "y", ",", "z", ",", "values"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.728753274824967*^9, 3.72875328424469*^9}},
 CellLabel->"In[104]:=",
 CellID->705238580,ExpressionUUID->"472e946e-30b0-4deb-aea8-9464e6871e86"]
}, Closed]],

Cell[CellGroupData[{

Cell["4.12. Example: Schwarzschild", "Subsection",
 CellChangeTimes->{{3.7286430157211113`*^9, 3.728643016017661*^9}, {
  3.728673968932828*^9, 3.7286739705964937`*^9}, {3.7286752930712357`*^9, 
  3.728675293199355*^9}, {3.728752965093289*^9, 3.728752968211376*^9}, {
  3.730913541982428*^9, 3.7309135505884933`*^9}},
 CellID->1994684186,ExpressionUUID->"3f0bcd93-51a6-4945-89d2-6e79a73667cc"],

Cell["\<\
This section is based on notes from Alfonso Garc\[IAcute]a-Parrado, \
circulated as a separate notebook since 2006, and now included here. This \
section can be executed separately from the rest of the notebook.\
\>", "Text",
 CellChangeTimes->{{3.730913577717367*^9, 3.7309136688806963`*^9}},
 CellID->543172681,ExpressionUUID->"def1e7ea-b590-44cd-95e7-6f4dfbd3bb0c"],

Cell[CellGroupData[{

Cell["4.12.1. Setup", "Subsubsection",
 CellChangeTimes->{{3.7286771125390873`*^9, 3.7286771127222157`*^9}, {
  3.730913913338131*^9, 3.730913919728261*^9}},
 CellID->807974394,ExpressionUUID->"3cb29779-1323-4284-93c3-f1a40c2f5634"],

Cell[CellGroupData[{

Cell["Load xCoba`, if it is not loaded yet:", "MathCaption",
 CellChangeTimes->{
  3.7309135645093603`*^9, {3.730913672392335*^9, 3.730913680200206*^9}},
 CellID->1713956350,ExpressionUUID->"f0ce7fef-0d0c-48c0-8e4e-02d723cef4dd"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<xAct`xCoba`\>\"", "]"}]], "Input",
 CellChangeTimes->{{3.7309136378429327`*^9, 3.730913647113517*^9}},
 CellLabel->"In[1]:=",
 CellID->611041600,ExpressionUUID->"596501f8-4c93-41b0-807f-551a497f2419"],

Cell[CellGroupData[{

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.730914316712738*^9, 3.7309149518956003`*^9, 
  3.730957251489663*^9, 3.730957334870701*^9, 3.73095756928409*^9, 
  3.730957688938222*^9, 3.7309577517026033`*^9, 3.730958180599801*^9, 
  3.7309588847543364`*^9, 3.7309599726626987`*^9, 3.7309604109907093`*^9, 
  3.731003121502275*^9, 3.731003282488576*^9, 3.731003927544757*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->796657213,ExpressionUUID->"d16713c7-abbb-4f49-a1c8-3e7a7aef6c67"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`xPerm`  version \"\>", 
   "\[InvisibleSpace]", "\<\"1.2.3\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2015", ",", "8", ",", "23"}], "}"}]}],
  SequenceForm["Package xAct`xPerm`  version ", "1.2.3", ", ", {2015, 8, 23}],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.730914316712738*^9, 3.7309149518956003`*^9, 
  3.730957251489663*^9, 3.730957334870701*^9, 3.73095756928409*^9, 
  3.730957688938222*^9, 3.7309577517026033`*^9, 3.730958180599801*^9, 
  3.7309588847543364`*^9, 3.7309599726626987`*^9, 3.7309604109907093`*^9, 
  3.731003121502275*^9, 3.731003282488576*^9, 3.731003927560172*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->2004723493,ExpressionUUID->"d2225d54-eadf-40f4-bd04-e25a5519b69f"],

Cell[BoxData["\<\"CopyRight (C) 2003-2018, Jose M. Martin-Garcia, under the \
General Public License.\"\>"], "Print",
 CellChangeTimes->{3.730914316712738*^9, 3.7309149518956003`*^9, 
  3.730957251489663*^9, 3.730957334870701*^9, 3.73095756928409*^9, 
  3.730957688938222*^9, 3.7309577517026033`*^9, 3.730958180599801*^9, 
  3.7309588847543364`*^9, 3.7309599726626987`*^9, 3.7309604109907093`*^9, 
  3.731003121502275*^9, 3.731003282488576*^9, 3.7310039275863123`*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1516809068,ExpressionUUID->"d989d15d-b10e-44ae-bcde-c2044f2c0fdf"],

Cell[BoxData["\<\"Connecting to external mac executable...\"\>"], "Print",
 CellChangeTimes->{3.730914316712738*^9, 3.7309149518956003`*^9, 
  3.730957251489663*^9, 3.730957334870701*^9, 3.73095756928409*^9, 
  3.730957688938222*^9, 3.7309577517026033`*^9, 3.730958180599801*^9, 
  3.7309588847543364`*^9, 3.7309599726626987`*^9, 3.7309604109907093`*^9, 
  3.731003121502275*^9, 3.731003282488576*^9, 3.731003927595768*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->231632339,ExpressionUUID->"d52c0e07-51a5-4b23-a651-29a3f34511ed"],

Cell[BoxData["\<\"Connection established.\"\>"], "Print",
 CellChangeTimes->{3.730914316712738*^9, 3.7309149518956003`*^9, 
  3.730957251489663*^9, 3.730957334870701*^9, 3.73095756928409*^9, 
  3.730957688938222*^9, 3.7309577517026033`*^9, 3.730958180599801*^9, 
  3.7309588847543364`*^9, 3.7309599726626987`*^9, 3.7309604109907093`*^9, 
  3.731003121502275*^9, 3.731003282488576*^9, 3.731003927606206*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1804353209,ExpressionUUID->"12f08b62-75bf-4103-bcf9-875f0de12be5"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.730914316712738*^9, 3.7309149518956003`*^9, 
  3.730957251489663*^9, 3.730957334870701*^9, 3.73095756928409*^9, 
  3.730957688938222*^9, 3.7309577517026033`*^9, 3.730958180599801*^9, 
  3.7309588847543364`*^9, 3.7309599726626987`*^9, 3.7309604109907093`*^9, 
  3.731003121502275*^9, 3.731003282488576*^9, 3.731003927616213*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1147804376,ExpressionUUID->"7eb45eb9-7fd0-4711-88f1-fb742b19677b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`xTensor`  version \"\>", 
   "\[InvisibleSpace]", "\<\"1.1.3\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2018", ",", "2", ",", "28"}], "}"}]}],
  SequenceForm[
  "Package xAct`xTensor`  version ", "1.1.3", ", ", {2018, 2, 28}],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914316712738*^9, 3.7309149518956003`*^9, 
  3.730957251489663*^9, 3.730957334870701*^9, 3.73095756928409*^9, 
  3.730957688938222*^9, 3.7309577517026033`*^9, 3.730958180599801*^9, 
  3.7309588847543364`*^9, 3.7309599726626987`*^9, 3.7309604109907093`*^9, 
  3.731003121502275*^9, 3.731003282488576*^9, 3.731003927626431*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->302546431,ExpressionUUID->"7f3c6944-64e7-4968-90af-0b44498ff29a"],

Cell[BoxData["\<\"CopyRight (C) 2002-2018, Jose M. Martin-Garcia, under the \
General Public License.\"\>"], "Print",
 CellChangeTimes->{3.730914316712738*^9, 3.7309149518956003`*^9, 
  3.730957251489663*^9, 3.730957334870701*^9, 3.73095756928409*^9, 
  3.730957688938222*^9, 3.7309577517026033`*^9, 3.730958180599801*^9, 
  3.7309588847543364`*^9, 3.7309599726626987`*^9, 3.7309604109907093`*^9, 
  3.731003121502275*^9, 3.731003282488576*^9, 3.731003927636532*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1804073368,ExpressionUUID->"113267a6-8468-4fe6-a994-e4895ec2026a"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.730914316712738*^9, 3.7309149518956003`*^9, 
  3.730957251489663*^9, 3.730957334870701*^9, 3.73095756928409*^9, 
  3.730957688938222*^9, 3.7309577517026033`*^9, 3.730958180599801*^9, 
  3.7309588847543364`*^9, 3.7309599726626987`*^9, 3.7309604109907093`*^9, 
  3.731003121502275*^9, 3.731003282488576*^9, 3.731003927736761*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->739483983,ExpressionUUID->"7bab0fb0-9be5-41d7-828f-8f41624ca7d6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`xCoba`  version \"\>", 
   "\[InvisibleSpace]", "\<\"0.8.4\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2018", ",", "2", ",", "28"}], "}"}]}],
  SequenceForm["Package xAct`xCoba`  version ", "0.8.4", ", ", {2018, 2, 28}],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.730914316712738*^9, 3.7309149518956003`*^9, 
  3.730957251489663*^9, 3.730957334870701*^9, 3.73095756928409*^9, 
  3.730957688938222*^9, 3.7309577517026033`*^9, 3.730958180599801*^9, 
  3.7309588847543364`*^9, 3.7309599726626987`*^9, 3.7309604109907093`*^9, 
  3.731003121502275*^9, 3.731003282488576*^9, 3.731003927747116*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1019437092,ExpressionUUID->"90ccfa99-dd99-4b13-9c30-2b4f212982bf"],

Cell[BoxData["\<\"CopyRight (C) 2005-2018, David Yllanes and Jose M. \
Martin-Garcia, under the General Public License.\"\>"], "Print",
 CellChangeTimes->{3.730914316712738*^9, 3.7309149518956003`*^9, 
  3.730957251489663*^9, 3.730957334870701*^9, 3.73095756928409*^9, 
  3.730957688938222*^9, 3.7309577517026033`*^9, 3.730958180599801*^9, 
  3.7309588847543364`*^9, 3.7309599726626987`*^9, 3.7309604109907093`*^9, 
  3.731003121502275*^9, 3.731003282488576*^9, 3.7310039277576933`*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1054669478,ExpressionUUID->"6c0cc4b0-3e6d-4fde-adbf-d2fd50a00c37"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.730914316712738*^9, 3.7309149518956003`*^9, 
  3.730957251489663*^9, 3.730957334870701*^9, 3.73095756928409*^9, 
  3.730957688938222*^9, 3.7309577517026033`*^9, 3.730958180599801*^9, 
  3.7309588847543364`*^9, 3.7309599726626987`*^9, 3.7309604109907093`*^9, 
  3.731003121502275*^9, 3.731003282488576*^9, 3.73100392776783*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->499894408,ExpressionUUID->"260cbd49-c36b-4a43-8aa0-bcd238b08e18"],

Cell[BoxData["\<\"These packages come with ABSOLUTELY NO WARRANTY; for \
details type Disclaimer[]. This is free software, and you are welcome to \
redistribute it under certain conditions. See the General Public License for \
details.\"\>"], "Print",
 CellChangeTimes->{3.730914316712738*^9, 3.7309149518956003`*^9, 
  3.730957251489663*^9, 3.730957334870701*^9, 3.73095756928409*^9, 
  3.730957688938222*^9, 3.7309577517026033`*^9, 3.730958180599801*^9, 
  3.7309588847543364`*^9, 3.7309599726626987`*^9, 3.7309604109907093`*^9, 
  3.731003121502275*^9, 3.731003282488576*^9, 3.731003927777965*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->769288192,ExpressionUUID->"04fa8bce-bab1-4a9e-a131-d325eadc0e2d"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.730914316712738*^9, 3.7309149518956003`*^9, 
  3.730957251489663*^9, 3.730957334870701*^9, 3.73095756928409*^9, 
  3.730957688938222*^9, 3.7309577517026033`*^9, 3.730958180599801*^9, 
  3.7309588847543364`*^9, 3.7309599726626987`*^9, 3.7309604109907093`*^9, 
  3.731003121502275*^9, 3.731003282488576*^9, 3.731003927787683*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1575088004,ExpressionUUID->"f529f93a-1710-4c66-85b5-50f71aca8b82"]
}, Open  ]]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"$PrePrint", "=", "ScreenDollarIndices"}], ";"}]], "Input",
 CellChangeTimes->{{3.730914497190155*^9, 3.73091450386228*^9}},
 CellLabel->"In[2]:=",
 CellID->1529876904,ExpressionUUID->"955ad36d-1f04-421c-b8f7-776aa247dde9"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Define 4d manifold with Lorentzian metric:", "MathCaption",
 CellID->1211465336,ExpressionUUID->"6f142790-0bda-4488-b70a-aacb88c0ffa0"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M4", ",", "4", ",", 
   RowBox[{"IndexRange", "[", 
    RowBox[{"a", ",", "p"}], "]"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.730913717183175*^9, 3.7309137173431253`*^9}},
 CellLabel->"In[3]:=",
 CellID->805444945,ExpressionUUID->"70cc8692-3f10-4953-8d44-3a549bf270e3"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefManifold", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"manifold \"\>", "\[InvisibleSpace]", "M4", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefManifold, ": Defining ", 
   "manifold ", $CellContext`M4, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913718089259*^9, 3.730914319979447*^9, 
  3.730914952454001*^9, 3.73095725223207*^9, 3.730957335423812*^9, 
  3.730957570269212*^9, 3.730957689602813*^9, 3.7309577526954203`*^9, 
  3.730958181285823*^9, 3.730958887038375*^9, 3.730959973317644*^9, 
  3.7309604115576477`*^9, 3.731003122084354*^9, 3.7310032830524282`*^9, 
  3.731003928571966*^9},
 CellLabel->"During evaluation of In[3]:=",
 CellID->473166659,ExpressionUUID->"15520097-f741-4bc2-9ce5-7fb02237dbf0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefVBundle", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle \"\>", "\[InvisibleSpace]", "TangentM4", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefVBundle, ": Defining ", 
   "vbundle ", $CellContext`TangentM4, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913718089259*^9, 3.730914319979447*^9, 
  3.730914952454001*^9, 3.73095725223207*^9, 3.730957335423812*^9, 
  3.730957570269212*^9, 3.730957689602813*^9, 3.7309577526954203`*^9, 
  3.730958181285823*^9, 3.730958887038375*^9, 3.730959973317644*^9, 
  3.7309604115576477`*^9, 3.731003122084354*^9, 3.7310032830524282`*^9, 
  3.7310039285865602`*^9},
 CellLabel->"During evaluation of In[3]:=",
 CellID->380092972,ExpressionUUID->"8b9ce7ce-1dda-4d70-882e-e7853429d7c7"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefMetric", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"3", ",", "1", ",", "0"}], "}"}], ",", 
   RowBox[{"metricg", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", "CD", ",", 
   RowBox[{"PrintAs", "->", "\"\<g\>\""}]}], "]"}]], "Input",
 CellChangeTimes->{{3.730913729384461*^9, 3.730913731039002*^9}},
 CellLabel->"In[4]:=",
 CellID->380288896,ExpressionUUID->"0990b25e-14d6-49d7-84a5-4baaaeedc45e"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric metric tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"metricg", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric metric tensor ", 
   $CellContext`metricg[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913740004959*^9, 3.730914320250893*^9, 
  3.730914952599628*^9, 3.730957252393077*^9, 3.730957335604322*^9, 
  3.7309575704510193`*^9, 3.730957689749982*^9, 3.7309577528421717`*^9, 
  3.73095818144633*^9, 3.730958887221574*^9, 3.730959973463541*^9, 
  3.730960411702566*^9, 3.731003122614393*^9, 3.7310032835561533`*^9, 
  3.731003928696597*^9},
 CellLabel->"During evaluation of In[4]:=",
 CellID->1606214226,ExpressionUUID->"9d045771-60cb-461d-ad1f-28fd0260612d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric tensor \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"epsilonmetricg", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric tensor ", 
   $CellContext`epsilonmetricg[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, -$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913740004959*^9, 3.730914320250893*^9, 
  3.730914952599628*^9, 3.730957252393077*^9, 3.730957335604322*^9, 
  3.7309575704510193`*^9, 3.730957689749982*^9, 3.7309577528421717`*^9, 
  3.73095818144633*^9, 3.730958887221574*^9, 3.730959973463541*^9, 
  3.730960411702566*^9, 3.731003122614393*^9, 3.7310032835561533`*^9, 
  3.731003928710929*^9},
 CellLabel->"During evaluation of In[4]:=",
 CellID->1773053592,ExpressionUUID->"6d48a9a7-66d9-4c5d-a924-4a1d8d0688c5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tetrametric \"\>", "\[InvisibleSpace]", 
   RowBox[{"Tetrametricg", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tetrametric ", 
   $CellContext`Tetrametricg[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, -$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913740004959*^9, 3.730914320250893*^9, 
  3.730914952599628*^9, 3.730957252393077*^9, 3.730957335604322*^9, 
  3.7309575704510193`*^9, 3.730957689749982*^9, 3.7309577528421717`*^9, 
  3.73095818144633*^9, 3.730958887221574*^9, 3.730959973463541*^9, 
  3.730960411702566*^9, 3.731003122614393*^9, 3.7310032835561533`*^9, 
  3.731003928724916*^9},
 CellLabel->"During evaluation of In[4]:=",
 CellID->1228634972,ExpressionUUID->"cc140eda-6185-4bed-91ea-4ef29771795f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tetrametric \"\>", "\[InvisibleSpace]", 
   RowBox[{"Tetrametricg\[Dagger]", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tetrametric ", 
   $CellContext`Tetrametricg\[Dagger][-$CellContext`a, -$CellContext`b, \
-$CellContext`c, -$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913740004959*^9, 3.730914320250893*^9, 
  3.730914952599628*^9, 3.730957252393077*^9, 3.730957335604322*^9, 
  3.7309575704510193`*^9, 3.730957689749982*^9, 3.7309577528421717`*^9, 
  3.73095818144633*^9, 3.730958887221574*^9, 3.730959973463541*^9, 
  3.730960411702566*^9, 3.731003122614393*^9, 3.7310032835561533`*^9, 
  3.731003928733017*^9},
 CellLabel->"During evaluation of In[4]:=",
 CellID->1612708499,ExpressionUUID->"7a65374f-1b00-445a-a9e7-ddff5c5b1c53"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"CD", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "covariant derivative ", 
   $CellContext`CD[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913740004959*^9, 3.730914320250893*^9, 
  3.730914952599628*^9, 3.730957252393077*^9, 3.730957335604322*^9, 
  3.7309575704510193`*^9, 3.730957689749982*^9, 3.7309577528421717`*^9, 
  3.73095818144633*^9, 3.730958887221574*^9, 3.730959973463541*^9, 
  3.730960411702566*^9, 3.731003122614393*^9, 3.7310032835561533`*^9, 
  3.731003928741104*^9},
 CellLabel->"During evaluation of In[4]:=",
 CellID->1400633906,ExpressionUUID->"a601eb96-ef30-4a1d-a787-24ce847a2605"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing torsion tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TorsionCD", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing torsion tensor ", 
   $CellContext`TorsionCD[$CellContext`a, -$CellContext`b, -$CellContext`c], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913740004959*^9, 3.730914320250893*^9, 
  3.730914952599628*^9, 3.730957252393077*^9, 3.730957335604322*^9, 
  3.7309575704510193`*^9, 3.730957689749982*^9, 3.7309577528421717`*^9, 
  3.73095818144633*^9, 3.730958887221574*^9, 3.730959973463541*^9, 
  3.730960411702566*^9, 3.731003122614393*^9, 3.7310032835561533`*^9, 
  3.7310039287490387`*^9},
 CellLabel->"During evaluation of In[4]:=",
 CellID->1885803375,ExpressionUUID->"3cb12b08-1ecd-412d-8392-183f1d55f7d1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelCD", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Christoffel tensor ", 
   $CellContext`ChristoffelCD[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913740004959*^9, 3.730914320250893*^9, 
  3.730914952599628*^9, 3.730957252393077*^9, 3.730957335604322*^9, 
  3.7309575704510193`*^9, 3.730957689749982*^9, 3.7309577528421717`*^9, 
  3.73095818144633*^9, 3.730958887221574*^9, 3.730959973463541*^9, 
  3.730960411702566*^9, 3.731003122614393*^9, 3.7310032835561533`*^9, 
  3.731003928757019*^9},
 CellLabel->"During evaluation of In[4]:=",
 CellID->2133661199,ExpressionUUID->"3a5b7f30-79b7-4f74-9b46-a9209f54a2d4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Riemann tensor ", 
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913740004959*^9, 3.730914320250893*^9, 
  3.730914952599628*^9, 3.730957252393077*^9, 3.730957335604322*^9, 
  3.7309575704510193`*^9, 3.730957689749982*^9, 3.7309577528421717`*^9, 
  3.73095818144633*^9, 3.730958887221574*^9, 3.730959973463541*^9, 
  3.730960411702566*^9, 3.731003122614393*^9, 3.7310032835561533`*^9, 
  3.731003928764965*^9},
 CellLabel->"During evaluation of In[4]:=",
 CellID->1761833987,ExpressionUUID->"90b1cbd5-ec94-469e-85ba-39e9af9fbc45"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric Ricci tensor ", 
   $CellContext`RicciCD[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913740004959*^9, 3.730914320250893*^9, 
  3.730914952599628*^9, 3.730957252393077*^9, 3.730957335604322*^9, 
  3.7309575704510193`*^9, 3.730957689749982*^9, 3.7309577528421717`*^9, 
  3.73095818144633*^9, 3.730958887221574*^9, 3.730959973463541*^9, 
  3.730960411702566*^9, 3.731003122614393*^9, 3.7310032835561533`*^9, 
  3.731003928772744*^9},
 CellLabel->"During evaluation of In[4]:=",
 CellID->1639294673,ExpressionUUID->"18e3c0a3-f864-420c-91e1-a8891373c423"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Riemann automatically replaced \
by Ricci.\"\>"], "Print",
 CellChangeTimes->{3.730913740004959*^9, 3.730914320250893*^9, 
  3.730914952599628*^9, 3.730957252393077*^9, 3.730957335604322*^9, 
  3.7309575704510193`*^9, 3.730957689749982*^9, 3.7309577528421717`*^9, 
  3.73095818144633*^9, 3.730958887221574*^9, 3.730959973463541*^9, 
  3.730960411702566*^9, 3.731003122614393*^9, 3.7310032835561533`*^9, 
  3.731003928780305*^9},
 CellLabel->"During evaluation of In[4]:=",
 CellID->1557861748,ExpressionUUID->"ec491262-9278-409f-b8b8-fb67752a1558"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"RicciScalarCD", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Ricci scalar ", 
   $CellContext`RicciScalarCD[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913740004959*^9, 3.730914320250893*^9, 
  3.730914952599628*^9, 3.730957252393077*^9, 3.730957335604322*^9, 
  3.7309575704510193`*^9, 3.730957689749982*^9, 3.7309577528421717`*^9, 
  3.73095818144633*^9, 3.730958887221574*^9, 3.730959973463541*^9, 
  3.730960411702566*^9, 3.731003122614393*^9, 3.7310032835561533`*^9, 
  3.731003928788004*^9},
 CellLabel->"During evaluation of In[4]:=",
 CellID->861774412,ExpressionUUID->"83368d89-0753-4141-a5fa-3377bf1e424e"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Ricci automatically replaced \
by RicciScalar.\"\>"], "Print",
 CellChangeTimes->{3.730913740004959*^9, 3.730914320250893*^9, 
  3.730914952599628*^9, 3.730957252393077*^9, 3.730957335604322*^9, 
  3.7309575704510193`*^9, 3.730957689749982*^9, 3.7309577528421717`*^9, 
  3.73095818144633*^9, 3.730958887221574*^9, 3.730959973463541*^9, 
  3.730960411702566*^9, 3.731003122614393*^9, 3.7310032835561533`*^9, 
  3.73100392879574*^9},
 CellLabel->"During evaluation of In[4]:=",
 CellID->1212827116,ExpressionUUID->"1c139746-fe84-4c8e-86b9-2add99baacbb"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Einstein tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"EinsteinCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Einstein tensor ", 
   $CellContext`EinsteinCD[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913740004959*^9, 3.730914320250893*^9, 
  3.730914952599628*^9, 3.730957252393077*^9, 3.730957335604322*^9, 
  3.7309575704510193`*^9, 3.730957689749982*^9, 3.7309577528421717`*^9, 
  3.73095818144633*^9, 3.730958887221574*^9, 3.730959973463541*^9, 
  3.730960411702566*^9, 3.731003122614393*^9, 3.7310032835561533`*^9, 
  3.731003928803314*^9},
 CellLabel->"During evaluation of In[4]:=",
 CellID->70561288,ExpressionUUID->"2c76a262-52e6-4aba-aee6-3ef478521c4e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Weyl tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"WeylCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "Weyl tensor ", 
   $CellContext`WeylCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913740004959*^9, 3.730914320250893*^9, 
  3.730914952599628*^9, 3.730957252393077*^9, 3.730957335604322*^9, 
  3.7309575704510193`*^9, 3.730957689749982*^9, 3.7309577528421717`*^9, 
  3.73095818144633*^9, 3.730958887221574*^9, 3.730959973463541*^9, 
  3.730960411702566*^9, 3.731003122614393*^9, 3.7310032835561533`*^9, 
  3.731003928810954*^9},
 CellLabel->"During evaluation of In[4]:=",
 CellID->512594272,ExpressionUUID->"6a63368e-ec76-4368-a270-536d0ff0992a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric TFRicci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TFRicciCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric TFRicci tensor ", 
   $CellContext`TFRicciCD[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913740004959*^9, 3.730914320250893*^9, 
  3.730914952599628*^9, 3.730957252393077*^9, 3.730957335604322*^9, 
  3.7309575704510193`*^9, 3.730957689749982*^9, 3.7309577528421717`*^9, 
  3.73095818144633*^9, 3.730958887221574*^9, 3.730959973463541*^9, 
  3.730960411702566*^9, 3.731003122614393*^9, 3.7310032835561533`*^9, 
  3.73100392889373*^9},
 CellLabel->"During evaluation of In[4]:=",
 CellID->1615021387,ExpressionUUID->"6cf7efc1-f296-40be-8add-21a5451d00b6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Kretschmann scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"KretschmannCD", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Kretschmann scalar ", 
   $CellContext`KretschmannCD[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913740004959*^9, 3.730914320250893*^9, 
  3.730914952599628*^9, 3.730957252393077*^9, 3.730957335604322*^9, 
  3.7309575704510193`*^9, 3.730957689749982*^9, 3.7309577528421717`*^9, 
  3.73095818144633*^9, 3.730958887221574*^9, 3.730959973463541*^9, 
  3.730960411702566*^9, 3.731003122614393*^9, 3.7310032835561533`*^9, 
  3.731003928901322*^9},
 CellLabel->"During evaluation of In[4]:=",
 CellID->1618636876,ExpressionUUID->"db84b139-501b-4b2b-a6fc-ff2e9499e157"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RiemannToWeylRules for dim \"\>", 
   "\[InvisibleSpace]", "4"}],
  SequenceForm["** DefCovD:  Computing RiemannToWeylRules for dim ", 4],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913740004959*^9, 3.730914320250893*^9, 
  3.730914952599628*^9, 3.730957252393077*^9, 3.730957335604322*^9, 
  3.7309575704510193`*^9, 3.730957689749982*^9, 3.7309577528421717`*^9, 
  3.73095818144633*^9, 3.730958887221574*^9, 3.730959973463541*^9, 
  3.730960411702566*^9, 3.731003122614393*^9, 3.7310032835561533`*^9, 
  3.73100392890868*^9},
 CellLabel->"During evaluation of In[4]:=",
 CellID->107134736,ExpressionUUID->"378afdea-0fbe-407e-bb7c-db0a15e80c74"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RicciToTFRicci for dim \"\>", 
   "\[InvisibleSpace]", "4"}],
  SequenceForm["** DefCovD:  Computing RicciToTFRicci for dim ", 4],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913740004959*^9, 3.730914320250893*^9, 
  3.730914952599628*^9, 3.730957252393077*^9, 3.730957335604322*^9, 
  3.7309575704510193`*^9, 3.730957689749982*^9, 3.7309577528421717`*^9, 
  3.73095818144633*^9, 3.730958887221574*^9, 3.730959973463541*^9, 
  3.730960411702566*^9, 3.731003122614393*^9, 3.7310032835561533`*^9, 
  3.7310039290414658`*^9},
 CellLabel->"During evaluation of In[4]:=",
 CellID->1022211766,ExpressionUUID->"615429d4-f7b5-4e27-bea3-bd366de56296"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RicciToEinsteinRules for dim \"\>", 
   "\[InvisibleSpace]", "4"}],
  SequenceForm["** DefCovD:  Computing RicciToEinsteinRules for dim ", 4],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913740004959*^9, 3.730914320250893*^9, 
  3.730914952599628*^9, 3.730957252393077*^9, 3.730957335604322*^9, 
  3.7309575704510193`*^9, 3.730957689749982*^9, 3.7309577528421717`*^9, 
  3.73095818144633*^9, 3.730958887221574*^9, 3.730959973463541*^9, 
  3.730960411702566*^9, 3.731003122614393*^9, 3.7310032835561533`*^9, 
  3.731003929049419*^9},
 CellLabel->"During evaluation of In[4]:=",
 CellID->443975162,ExpressionUUID->"cecb06a9-9c90-4e52-b354-93d6b76a2085"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density \"\>", "\[InvisibleSpace]", 
   RowBox[{"Detmetricg", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Determinant.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "weight +2 density ", 
   $CellContext`Detmetricg[], ". ", "Determinant."],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913740004959*^9, 3.730914320250893*^9, 
  3.730914952599628*^9, 3.730957252393077*^9, 3.730957335604322*^9, 
  3.7309575704510193`*^9, 3.730957689749982*^9, 3.7309577528421717`*^9, 
  3.73095818144633*^9, 3.730958887221574*^9, 3.730959973463541*^9, 
  3.730960411702566*^9, 3.731003122614393*^9, 3.7310032835561533`*^9, 
  3.731003929057513*^9},
 CellLabel->"During evaluation of In[4]:=",
 CellID->1532358056,ExpressionUUID->"2d7678db-4795-4974-be6c-45afa465769b"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Define the coordinates and their associated frame:", "MathCaption",
 CellChangeTimes->{3.7309137553177547`*^9},
 CellID->1148260682,ExpressionUUID->"05d2bfcf-17c7-4a6a-a904-5b76fe82f67d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefChart", "[", 
  RowBox[{"Sch", ",", "M4", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2", ",", "3"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"\[Tau]", "[", "]"}], ",", 
     RowBox[{"r", "[", "]"}], ",", 
     RowBox[{"\[Theta]", "[", "]"}], ",", 
     RowBox[{"\[Phi]", "[", "]"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[5]:=",
 CellID->1529230432,ExpressionUUID->"fc58fb07-e6ee-40f0-bab4-046b65ec5e45"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefChart", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"chart \"\>", "\[InvisibleSpace]", "Sch", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xCoba`DefChart, ": Defining ", "chart ", $CellContext`Sch, ". ",
    ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913758172369*^9, 3.730914323017124*^9, 
  3.73091495310096*^9, 3.7309572528759336`*^9, 3.7309573383205023`*^9, 
  3.730957572700955*^9, 3.730957691865986*^9, 3.73095775502427*^9, 
  3.7309581819639683`*^9, 3.730958887755672*^9, 3.730959973979632*^9, 
  3.7309604122194157`*^9, 3.731003123190838*^9, 3.731003284180112*^9, 
  3.731003930888709*^9},
 CellLabel->"During evaluation of In[5]:=",
 CellID->1717952368,ExpressionUUID->"2abc5ac0-00ad-4312-ba54-51c6ccc21a2c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Tau]", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "coordinate scalar ", 
   $CellContext`\[Tau][], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913758172369*^9, 3.730914323017124*^9, 
  3.73091495310096*^9, 3.7309572528759336`*^9, 3.7309573383205023`*^9, 
  3.730957572700955*^9, 3.730957691865986*^9, 3.73095775502427*^9, 
  3.7309581819639683`*^9, 3.730958887755672*^9, 3.730959973979632*^9, 
  3.7309604122194157`*^9, 3.731003123190838*^9, 3.731003284180112*^9, 
  3.731003930902793*^9},
 CellLabel->"During evaluation of In[5]:=",
 CellID->707815061,ExpressionUUID->"8e5ad076-0b4f-4160-8643-97283938683b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"r", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "coordinate scalar ", 
   $CellContext`r[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913758172369*^9, 3.730914323017124*^9, 
  3.73091495310096*^9, 3.7309572528759336`*^9, 3.7309573383205023`*^9, 
  3.730957572700955*^9, 3.730957691865986*^9, 3.73095775502427*^9, 
  3.7309581819639683`*^9, 3.730958887755672*^9, 3.730959973979632*^9, 
  3.7309604122194157`*^9, 3.731003123190838*^9, 3.731003284180112*^9, 
  3.7310039309172363`*^9},
 CellLabel->"During evaluation of In[5]:=",
 CellID->1335809494,ExpressionUUID->"1b1d1287-e566-4472-af3c-47786dedebca"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Theta]", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "coordinate scalar ", 
   $CellContext`\[Theta][], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913758172369*^9, 3.730914323017124*^9, 
  3.73091495310096*^9, 3.7309572528759336`*^9, 3.7309573383205023`*^9, 
  3.730957572700955*^9, 3.730957691865986*^9, 3.73095775502427*^9, 
  3.7309581819639683`*^9, 3.730958887755672*^9, 3.730959973979632*^9, 
  3.7309604122194157`*^9, 3.731003123190838*^9, 3.731003284180112*^9, 
  3.7310039309248962`*^9},
 CellLabel->"During evaluation of In[5]:=",
 CellID->1156119920,ExpressionUUID->"5d76ac16-420b-4587-aa43-e1561264fa64"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Phi]", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "coordinate scalar ", 
   $CellContext`\[Phi][], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913758172369*^9, 3.730914323017124*^9, 
  3.73091495310096*^9, 3.7309572528759336`*^9, 3.7309573383205023`*^9, 
  3.730957572700955*^9, 3.730957691865986*^9, 3.73095775502427*^9, 
  3.7309581819639683`*^9, 3.730958887755672*^9, 3.730959973979632*^9, 
  3.7309604122194157`*^9, 3.731003123190838*^9, 3.731003284180112*^9, 
  3.731003930932701*^9},
 CellLabel->"During evaluation of In[5]:=",
 CellID->475457384,ExpressionUUID->"8f3a7288-4e99-46ba-9788-724da2a6d67b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefMapping", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"mapping \"\>", "\[InvisibleSpace]", "Sch", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefMapping, ": Defining ", "mapping ", $CellContext`Sch,
    ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913758172369*^9, 3.730914323017124*^9, 
  3.73091495310096*^9, 3.7309572528759336`*^9, 3.7309573383205023`*^9, 
  3.730957572700955*^9, 3.730957691865986*^9, 3.73095775502427*^9, 
  3.7309581819639683`*^9, 3.730958887755672*^9, 3.730959973979632*^9, 
  3.7309604122194157`*^9, 3.731003123190838*^9, 3.731003284180112*^9, 
  3.731003930940515*^9},
 CellLabel->"During evaluation of In[5]:=",
 CellID->225602401,ExpressionUUID->"7e40402e-1fa8-4a39-b292-2ec12b32d304"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefMapping", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"inverse mapping \"\>", "\[InvisibleSpace]", 
   "iSch", "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefMapping, ": Defining ", 
   "inverse mapping ", $CellContext`iSch, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913758172369*^9, 3.730914323017124*^9, 
  3.73091495310096*^9, 3.7309572528759336`*^9, 3.7309573383205023`*^9, 
  3.730957572700955*^9, 3.730957691865986*^9, 3.73095775502427*^9, 
  3.7309581819639683`*^9, 3.730958887755672*^9, 3.730959973979632*^9, 
  3.7309604122194157`*^9, 3.731003123190838*^9, 3.731003284180112*^9, 
  3.7310039309486*^9},
 CellLabel->"During evaluation of In[5]:=",
 CellID->1390916652,ExpressionUUID->"ad0bb392-fa70-43db-ac0a-5b033424db40"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"mapping differential tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"diSch", "[", 
    RowBox[{
     RowBox[{"-", "\[FormalA]"}], ",", "iScha"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "mapping differential tensor ", 
   $CellContext`diSch[-\[FormalA], $CellContext`iScha], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913758172369*^9, 3.730914323017124*^9, 
  3.73091495310096*^9, 3.7309572528759336`*^9, 3.7309573383205023`*^9, 
  3.730957572700955*^9, 3.730957691865986*^9, 3.73095775502427*^9, 
  3.7309581819639683`*^9, 3.730958887755672*^9, 3.730959973979632*^9, 
  3.7309604122194157`*^9, 3.731003123190838*^9, 3.731003284180112*^9, 
  3.7310039309567213`*^9},
 CellLabel->"During evaluation of In[5]:=",
 CellID->1776672569,ExpressionUUID->"a501c618-180f-4bbd-a816-84ee34530d5d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"mapping differential tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"dSch", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", "Sch\[FormalA]"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "mapping differential tensor ", 
   $CellContext`dSch[-$CellContext`a, $CellContext`Sch\[FormalA]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913758172369*^9, 3.730914323017124*^9, 
  3.73091495310096*^9, 3.7309572528759336`*^9, 3.7309573383205023`*^9, 
  3.730957572700955*^9, 3.730957691865986*^9, 3.73095775502427*^9, 
  3.7309581819639683`*^9, 3.730958887755672*^9, 3.730959973979632*^9, 
  3.7309604122194157`*^9, 3.731003123190838*^9, 3.731003284180112*^9, 
  3.731003930965384*^9},
 CellLabel->"During evaluation of In[5]:=",
 CellID->1923239295,ExpressionUUID->"998e9b9a-0ee2-4973-8718-ee0c38ed2c0e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefBasis", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"basis \"\>", "\[InvisibleSpace]", "Sch", 
   "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Coordinated basis.\"\>"}],
  SequenceForm[
  "** ", xAct`xCoba`DefBasis, ": Defining ", "basis ", $CellContext`Sch, ". ",
    "Coordinated basis."],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913758172369*^9, 3.730914323017124*^9, 
  3.73091495310096*^9, 3.7309572528759336`*^9, 3.7309573383205023`*^9, 
  3.730957572700955*^9, 3.730957691865986*^9, 3.73095775502427*^9, 
  3.7309581819639683`*^9, 3.730958887755672*^9, 3.730959973979632*^9, 
  3.7309604122194157`*^9, 3.731003123190838*^9, 3.731003284180112*^9, 
  3.73100393097351*^9},
 CellLabel->"During evaluation of In[5]:=",
 CellID->2106460068,ExpressionUUID->"33125115-cbf2-436c-a5f5-c3fb35b0ff98"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"parallel derivative \"\>", "\[InvisibleSpace]", 
   RowBox[{"PDSch", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "parallel derivative ", 
   $CellContext`PDSch[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913758172369*^9, 3.730914323017124*^9, 
  3.73091495310096*^9, 3.7309572528759336`*^9, 3.7309573383205023`*^9, 
  3.730957572700955*^9, 3.730957691865986*^9, 3.73095775502427*^9, 
  3.7309581819639683`*^9, 3.730958887755672*^9, 3.730959973979632*^9, 
  3.7309604122194157`*^9, 3.731003123190838*^9, 3.731003284180112*^9, 
  3.7310039309812803`*^9},
 CellLabel->"During evaluation of In[5]:=",
 CellID->2006442081,ExpressionUUID->"f0bfdb22-ef49-400a-aa24-3fc9010c79d9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing torsion tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TorsionPDSch", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing torsion tensor ", 
   $CellContext`TorsionPDSch[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913758172369*^9, 3.730914323017124*^9, 
  3.73091495310096*^9, 3.7309572528759336`*^9, 3.7309573383205023`*^9, 
  3.730957572700955*^9, 3.730957691865986*^9, 3.73095775502427*^9, 
  3.7309581819639683`*^9, 3.730958887755672*^9, 3.730959973979632*^9, 
  3.7309604122194157`*^9, 3.731003123190838*^9, 3.731003284180112*^9, 
  3.731003930990047*^9},
 CellLabel->"During evaluation of In[5]:=",
 CellID->336346526,ExpressionUUID->"fe0903d3-d680-45fd-a1ea-3cc5cc197dcd"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelPDSch", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Christoffel tensor ", 
   $CellContext`ChristoffelPDSch[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913758172369*^9, 3.730914323017124*^9, 
  3.73091495310096*^9, 3.7309572528759336`*^9, 3.7309573383205023`*^9, 
  3.730957572700955*^9, 3.730957691865986*^9, 3.73095775502427*^9, 
  3.7309581819639683`*^9, 3.730958887755672*^9, 3.730959973979632*^9, 
  3.7309604122194157`*^9, 3.731003123190838*^9, 3.731003284180112*^9, 
  3.731003930998917*^9},
 CellLabel->"During evaluation of In[5]:=",
 CellID->799103578,ExpressionUUID->"2df21e08-2f09-4775-8ff4-947ce47e6357"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Riemann tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RiemannPDSch", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Riemann tensor ", 
   $CellContext`RiemannPDSch[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, $CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913758172369*^9, 3.730914323017124*^9, 
  3.73091495310096*^9, 3.7309572528759336`*^9, 3.7309573383205023`*^9, 
  3.730957572700955*^9, 3.730957691865986*^9, 3.73095775502427*^9, 
  3.7309581819639683`*^9, 3.730958887755672*^9, 3.730959973979632*^9, 
  3.7309604122194157`*^9, 3.731003123190838*^9, 3.731003284180112*^9, 
  3.7310039310066357`*^9},
 CellLabel->"During evaluation of In[5]:=",
 CellID->171107108,ExpressionUUID->"fa6e3254-0d5d-4a7d-92c4-a756970869a3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciPDSch", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Ricci tensor ", 
   $CellContext`RicciPDSch[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913758172369*^9, 3.730914323017124*^9, 
  3.73091495310096*^9, 3.7309572528759336`*^9, 3.7309573383205023`*^9, 
  3.730957572700955*^9, 3.730957691865986*^9, 3.73095775502427*^9, 
  3.7309581819639683`*^9, 3.730958887755672*^9, 3.730959973979632*^9, 
  3.7309604122194157`*^9, 3.731003123190838*^9, 3.731003284180112*^9, 
  3.7310039310140123`*^9},
 CellLabel->"During evaluation of In[5]:=",
 CellID->316560823,ExpressionUUID->"ab76495d-73d9-4bcd-928a-b70a46d4bff7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaUpSch", "[", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric +1 density ", 
   $CellContext`etaUpSch[$CellContext`a, $CellContext`b, $CellContext`c, \
$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913758172369*^9, 3.730914323017124*^9, 
  3.73091495310096*^9, 3.7309572528759336`*^9, 3.7309573383205023`*^9, 
  3.730957572700955*^9, 3.730957691865986*^9, 3.73095775502427*^9, 
  3.7309581819639683`*^9, 3.730958887755672*^9, 3.730959973979632*^9, 
  3.7309604122194157`*^9, 3.731003123190838*^9, 3.731003284180112*^9, 
  3.731003931022048*^9},
 CellLabel->"During evaluation of In[5]:=",
 CellID->1120758542,ExpressionUUID->"de43e0af-632b-4b34-b208-a7b970d673fa"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaDownSch", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric -1 density ", 
   $CellContext`etaDownSch[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913758172369*^9, 3.730914323017124*^9, 
  3.73091495310096*^9, 3.7309572528759336`*^9, 3.7309573383205023`*^9, 
  3.730957572700955*^9, 3.730957691865986*^9, 3.73095775502427*^9, 
  3.7309581819639683`*^9, 3.730958887755672*^9, 3.730959973979632*^9, 
  3.7309604122194157`*^9, 3.731003123190838*^9, 3.731003284180112*^9, 
  3.731003931030861*^9},
 CellLabel->"During evaluation of In[5]:=",
 CellID->1009747557,ExpressionUUID->"03cdaf01-0baa-4c5d-85f2-d8bbfdeefd88"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The mass of the spacetime:", "MathCaption",
 CellID->97390299,ExpressionUUID->"09105b1e-f0e8-46cd-8a11-1b8b9bf82cc8"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefConstantSymbol", "[", "M", "]"}]], "Input",
 CellLabel->"In[6]:=",
 CellID->456216279,ExpressionUUID->"8182bc83-c25c-406d-9517-0c2df8eb7e84"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefConstantSymbol", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"constant symbol \"\>", "\[InvisibleSpace]", "M", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefConstantSymbol, ": Defining ", 
   "constant symbol ", $CellContext`M, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730913765887748*^9, 3.730914324582122*^9, 
  3.730914953359746*^9, 3.730957253129841*^9, 3.730957338887485*^9, 
  3.7309575730343*^9, 3.730957692182702*^9, 3.730957755493804*^9, 
  3.7309581838838053`*^9, 3.730958888072901*^9, 3.730959975846377*^9, 
  3.730960414173512*^9, 3.731003124948893*^9, 3.731003285982576*^9, 
  3.731003931123908*^9},
 CellLabel->"During evaluation of In[6]:=",
 CellID->1411411905,ExpressionUUID->"3c7e1500-2d1f-429e-ab2e-27a7ba999d2c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The covariant derivative of the elements of the basis will be described by \
this Christoffel tensor:\
\>", "MathCaption",
 CellChangeTimes->{{3.73091378324617*^9, 3.730913827731214*^9}},
 CellID->1178915150,ExpressionUUID->"20518fc6-47af-4312-bc5f-a0cce6caa19c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CD", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"Basis", "[", 
   RowBox[{"b", ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "Sch"}]}], "}"}]}], "]"}], "]"}]], "Input",
 CellLabel->"In[7]:=",
 CellID->1307502051,ExpressionUUID->"d045dfa4-69e3-46bc-81aa-15430dc7e2df"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"ChristoffelCDPDSch", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`ChristoffelCDPDSch[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914325729656*^9, 3.730914953506021*^9, 
  3.7309572532425756`*^9, 3.7309573391095457`*^9, 3.730957573184032*^9, 
  3.730957692317792*^9, 3.730957755658186*^9, 3.730958184046651*^9, 
  3.7309588882211933`*^9, 3.730959975965467*^9, 3.730960414306851*^9, 
  3.731003125127863*^9, 3.731003286097335*^9, 3.731003931204772*^9},
 CellLabel->"During evaluation of In[7]:=",
 CellID->495522573,ExpressionUUID->"f6954cdb-cfc8-4d95-857d-1c29a605053d"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"\[CapitalGamma]", "[", 
       RowBox[{"\[EmptyDownTriangle]", ",", 
        StyleBox["\[ScriptCapitalD]",
         FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
      StyleBox[GridBox[{
         {"b", " ", " "},
         {" ", "a", 
          StyleBox["c",
           FontColor->RGBColor[1., 0., 0.]]}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`ChristoffelCDPDSch[$CellContext`b, -$CellContext`a, \
{-$CellContext`c, -$CellContext`Sch}],
  Editable->False]], "Output",
 CellChangeTimes->{3.730913824189146*^9, 3.730914325809108*^9, 
  3.730914953547511*^9, 3.7309572532761183`*^9, 3.730957339154127*^9, 
  3.730957573302845*^9, 3.7309576923503313`*^9, 3.730957755759489*^9, 
  3.7309581841164207`*^9, 3.730958888303697*^9, 3.730959975995798*^9, 
  3.730960414370409*^9, 3.7310031251572323`*^9, 3.731003286134523*^9, 
  3.731003931226742*^9},
 CellLabel->"Out[7]=",
 CellID->306301345,ExpressionUUID->"28cfd954-42b3-4ee5-bdb2-47537247b25a"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["4.12.2. Input metric", "Subsubsection",
 CellChangeTimes->{{3.7286771125390873`*^9, 3.7286771127222157`*^9}, {
  3.730913913338131*^9, 3.730913941103621*^9}},
 CellID->1896835986,ExpressionUUID->"6a438a47-4bf7-47d2-9665-8ec7d2f30988"],

Cell["\<\
Specify the values of the metric components in the coordinate basis just \
defined. We will store those components as tensor values for the metricg \
tensor.\
\>", "Text",
 CellChangeTimes->{{3.730913577717367*^9, 3.7309136688806963`*^9}, {
  3.7310031751218443`*^9, 3.731003229439324*^9}},
 CellID->1679298556,ExpressionUUID->"c71be471-1d2d-4450-bb4f-bbe458623bd4"],

Cell[CellGroupData[{

Cell["\<\
This is the matrix of scalar components of the Schwarzschild metric in the \
Sch coordinate basis:\
\>", "MathCaption",
 CellChangeTimes->{{3.7310032385758467`*^9, 3.7310032703986692`*^9}, {
  3.731003306493703*^9, 3.731003308397543*^9}},
 CellID->2064139895,ExpressionUUID->"23286258-8bf2-474a-8a91-1782df0dd7c7"],

Cell[BoxData[
 RowBox[{
  RowBox[{"matrix", "=", 
   RowBox[{"(", GridBox[{
      {
       RowBox[{"-", 
        RowBox[{"(", 
         RowBox[{"1", "-", 
          RowBox[{"2", 
           RowBox[{"M", "/", 
            RowBox[{"r", "[", "]"}]}]}]}], ")"}]}], "0", "0", "0"},
      {"0", 
       RowBox[{"1", "/", 
        RowBox[{"(", 
         RowBox[{"1", "-", 
          RowBox[{"2", 
           RowBox[{"M", "/", 
            RowBox[{"r", "[", "]"}]}]}]}], ")"}]}], "0", "0"},
      {"0", "0", 
       RowBox[{
        RowBox[{"r", "[", "]"}], "^", "2"}], "0"},
      {"0", "0", "0", 
       RowBox[{
        RowBox[{
         RowBox[{"r", "[", "]"}], "^", "2"}], 
        RowBox[{
         RowBox[{"Sin", "[", 
          RowBox[{"\[Theta]", "[", "]"}], "]"}], "^", "2"}]}]}
     }], ")"}]}], ";"}]], "Input",
 CellChangeTimes->{
  3.730913953632914*^9, {3.731003131685643*^9, 3.731003139893063*^9}},
 CellLabel->"In[8]:=",
 CellID->1548381627,ExpressionUUID->"033f9dff-bb6e-4ee4-893b-3e027d0c8c91"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Set those scalars as tensor values for the metricg tensor, reporting each \
individual assignment. We will later see how to deactivate them:\
\>", "MathCaption",
 CellChangeTimes->{{3.730914346681881*^9, 3.73091437470514*^9}, {
  3.731003299486094*^9, 3.731003345756345*^9}},
 CellID->1246174314,ExpressionUUID->"78f554ec-f58e-420d-be3a-f07bcacfa6ba"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"MetricInBasis", "[", 
  RowBox[{"metricg", ",", 
   RowBox[{"-", "Sch"}], ",", "matrix"}], "]"}]], "Input",
 CellChangeTimes->{{3.731003151843067*^9, 3.731003153962832*^9}},
 CellLabel->"In[9]:=",
 CellID->895709105,ExpressionUUID->"8db67e94-be57-459f-83ed-74882d1a5436"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "metricg"}],
  SequenceForm[
  "Added independent rule ", $CellContext`metricg[{0, -$CellContext`Sch}, {
     0, -$CellContext`Sch}] -> -1 + 2 ($CellContext`M/$CellContext`r[]), 
   " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914339339313*^9, 3.730914956546998*^9, 
  3.73095733937106*^9, 3.730957573517251*^9, 3.730957692549412*^9, 
  3.730957756021146*^9, 3.730958184697094*^9, 3.73095888847186*^9, 
  3.730959976146467*^9, 3.7309604145859423`*^9, 3.731003349194276*^9, 
  3.731003931329116*^9},
 CellLabel->"During evaluation of In[9]:=",
 CellID->488403556,ExpressionUUID->"874e064b-3f6b-49d2-b6fc-73bf4a442732"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", 
   "metricg"}],
  SequenceForm[
  "Added independent rule ", $CellContext`metricg[{0, -$CellContext`Sch}, {
     1, -$CellContext`Sch}] -> 0, " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914339339313*^9, 3.730914956546998*^9, 
  3.73095733937106*^9, 3.730957573517251*^9, 3.730957692549412*^9, 
  3.730957756021146*^9, 3.730958184697094*^9, 3.73095888847186*^9, 
  3.730959976146467*^9, 3.7309604145859423`*^9, 3.731003349194276*^9, 
  3.731003931370633*^9},
 CellLabel->"During evaluation of In[9]:=",
 CellID->916066858,ExpressionUUID->"f80235dc-e71d-4509-87dc-211382edad71"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", 
   "metricg"}],
  SequenceForm[
  "Added independent rule ", $CellContext`metricg[{0, -$CellContext`Sch}, {
     2, -$CellContext`Sch}] -> 0, " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914339339313*^9, 3.730914956546998*^9, 
  3.73095733937106*^9, 3.730957573517251*^9, 3.730957692549412*^9, 
  3.730957756021146*^9, 3.730958184697094*^9, 3.73095888847186*^9, 
  3.730959976146467*^9, 3.7309604145859423`*^9, 3.731003349194276*^9, 
  3.731003931441677*^9},
 CellLabel->"During evaluation of In[9]:=",
 CellID->1025417063,ExpressionUUID->"41e293bb-a4a1-4054-aac1-1cddc8d2aee5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", 
   "metricg"}],
  SequenceForm[
  "Added independent rule ", $CellContext`metricg[{0, -$CellContext`Sch}, {
     3, -$CellContext`Sch}] -> 0, " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914339339313*^9, 3.730914956546998*^9, 
  3.73095733937106*^9, 3.730957573517251*^9, 3.730957692549412*^9, 
  3.730957756021146*^9, 3.730958184697094*^9, 3.73095888847186*^9, 
  3.730959976146467*^9, 3.7309604145859423`*^9, 3.731003349194276*^9, 
  3.7310039315081053`*^9},
 CellLabel->"During evaluation of In[9]:=",
 CellID->628310666,ExpressionUUID->"e82fd881-2bae-4411-958f-aa95f5e61892"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{1, -$CellContext`Sch}, {0, -$CellContext`Sch}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "metricg"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`metricg[{1, -$CellContext`Sch}, {
     0, -$CellContext`Sch}] -> $CellContext`metricg[{0, -$CellContext`Sch}, {
     1, -$CellContext`Sch}], " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914339339313*^9, 3.730914956546998*^9, 
  3.73095733937106*^9, 3.730957573517251*^9, 3.730957692549412*^9, 
  3.730957756021146*^9, 3.730958184697094*^9, 3.73095888847186*^9, 
  3.730959976146467*^9, 3.7309604145859423`*^9, 3.731003349194276*^9, 
  3.731003931574939*^9},
 CellLabel->"During evaluation of In[9]:=",
 CellID->840271163,ExpressionUUID->"d8e9e2b6-a82f-42c0-9092-e1ca833c19d0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
     Editable->False], "\[Rule]", 
    FractionBox["1", 
     RowBox[{"1", "-", 
      FractionBox[
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False]}], 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]]}]]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "metricg"}],
  SequenceForm[
  "Added independent rule ", $CellContext`metricg[{1, -$CellContext`Sch}, {
     1, -$CellContext`Sch}] -> (1 - 2 ($CellContext`M/$CellContext`r[]))^(-1),
    " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914339339313*^9, 3.730914956546998*^9, 
  3.73095733937106*^9, 3.730957573517251*^9, 3.730957692549412*^9, 
  3.730957756021146*^9, 3.730958184697094*^9, 3.73095888847186*^9, 
  3.730959976146467*^9, 3.7309604145859423`*^9, 3.731003349194276*^9, 
  3.7310039316412888`*^9},
 CellLabel->"During evaluation of In[9]:=",
 CellID->584973869,ExpressionUUID->"151177a8-063f-46ed-94bf-f148bf4b3cf1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", 
   "metricg"}],
  SequenceForm[
  "Added independent rule ", $CellContext`metricg[{1, -$CellContext`Sch}, {
     2, -$CellContext`Sch}] -> 0, " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914339339313*^9, 3.730914956546998*^9, 
  3.73095733937106*^9, 3.730957573517251*^9, 3.730957692549412*^9, 
  3.730957756021146*^9, 3.730958184697094*^9, 3.73095888847186*^9, 
  3.730959976146467*^9, 3.7309604145859423`*^9, 3.731003349194276*^9, 
  3.7310039317081537`*^9},
 CellLabel->"During evaluation of In[9]:=",
 CellID->475680317,ExpressionUUID->"b252307c-a691-42e3-8189-936d81352b30"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", 
   "metricg"}],
  SequenceForm[
  "Added independent rule ", $CellContext`metricg[{1, -$CellContext`Sch}, {
     3, -$CellContext`Sch}] -> 0, " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914339339313*^9, 3.730914956546998*^9, 
  3.73095733937106*^9, 3.730957573517251*^9, 3.730957692549412*^9, 
  3.730957756021146*^9, 3.730958184697094*^9, 3.73095888847186*^9, 
  3.730959976146467*^9, 3.7309604145859423`*^9, 3.731003349194276*^9, 
  3.731003931774733*^9},
 CellLabel->"During evaluation of In[9]:=",
 CellID->1824953685,ExpressionUUID->"39a6f7ad-5d39-4886-8770-3e90c389a03d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{2, -$CellContext`Sch}, {0, -$CellContext`Sch}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "metricg"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`metricg[{2, -$CellContext`Sch}, {
     0, -$CellContext`Sch}] -> $CellContext`metricg[{0, -$CellContext`Sch}, {
     2, -$CellContext`Sch}], " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914339339313*^9, 3.730914956546998*^9, 
  3.73095733937106*^9, 3.730957573517251*^9, 3.730957692549412*^9, 
  3.730957756021146*^9, 3.730958184697094*^9, 3.73095888847186*^9, 
  3.730959976146467*^9, 3.7309604145859423`*^9, 3.731003349194276*^9, 
  3.7310039318425207`*^9},
 CellLabel->"During evaluation of In[9]:=",
 CellID->1635137341,ExpressionUUID->"316fa9a9-527b-4134-9ff7-4e884ce035af"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{2, -$CellContext`Sch}, {1, -$CellContext`Sch}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "metricg"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`metricg[{2, -$CellContext`Sch}, {
     1, -$CellContext`Sch}] -> $CellContext`metricg[{1, -$CellContext`Sch}, {
     2, -$CellContext`Sch}], " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914339339313*^9, 3.730914956546998*^9, 
  3.73095733937106*^9, 3.730957573517251*^9, 3.730957692549412*^9, 
  3.730957756021146*^9, 3.730958184697094*^9, 3.73095888847186*^9, 
  3.730959976146467*^9, 3.7309604145859423`*^9, 3.731003349194276*^9, 
  3.731003931908744*^9},
 CellLabel->"During evaluation of In[9]:=",
 CellID->405059528,ExpressionUUID->"889534eb-ddbb-49f4-956a-660d754c1f0c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "\[Rule]", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "metricg"}],
  SequenceForm[
  "Added independent rule ", $CellContext`metricg[{2, -$CellContext`Sch}, {
     2, -$CellContext`Sch}] -> $CellContext`r[]^2, 
   " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914339339313*^9, 3.730914956546998*^9, 
  3.73095733937106*^9, 3.730957573517251*^9, 3.730957692549412*^9, 
  3.730957756021146*^9, 3.730958184697094*^9, 3.73095888847186*^9, 
  3.730959976146467*^9, 3.7309604145859423`*^9, 3.731003349194276*^9, 
  3.731003931974237*^9},
 CellLabel->"During evaluation of In[9]:=",
 CellID->312326106,ExpressionUUID->"6f5b1187-02fc-40f4-a3be-273a2e4ee04e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", 
   "metricg"}],
  SequenceForm[
  "Added independent rule ", $CellContext`metricg[{2, -$CellContext`Sch}, {
     3, -$CellContext`Sch}] -> 0, " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914339339313*^9, 3.730914956546998*^9, 
  3.73095733937106*^9, 3.730957573517251*^9, 3.730957692549412*^9, 
  3.730957756021146*^9, 3.730958184697094*^9, 3.73095888847186*^9, 
  3.730959976146467*^9, 3.7309604145859423`*^9, 3.731003349194276*^9, 
  3.731003932024712*^9},
 CellLabel->"During evaluation of In[9]:=",
 CellID->814830274,ExpressionUUID->"85a6c695-f8ba-486a-9325-43de5d226b54"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{3, -$CellContext`Sch}, {0, -$CellContext`Sch}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "metricg"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`metricg[{3, -$CellContext`Sch}, {
     0, -$CellContext`Sch}] -> $CellContext`metricg[{0, -$CellContext`Sch}, {
     3, -$CellContext`Sch}], " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914339339313*^9, 3.730914956546998*^9, 
  3.73095733937106*^9, 3.730957573517251*^9, 3.730957692549412*^9, 
  3.730957756021146*^9, 3.730958184697094*^9, 3.73095888847186*^9, 
  3.730959976146467*^9, 3.7309604145859423`*^9, 3.731003349194276*^9, 
  3.731003932091436*^9},
 CellLabel->"During evaluation of In[9]:=",
 CellID->349198199,ExpressionUUID->"05c6c188-6d25-41c9-be08-05a64892f994"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{3, -$CellContext`Sch}, {1, -$CellContext`Sch}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "metricg"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`metricg[{3, -$CellContext`Sch}, {
     1, -$CellContext`Sch}] -> $CellContext`metricg[{1, -$CellContext`Sch}, {
     3, -$CellContext`Sch}], " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914339339313*^9, 3.730914956546998*^9, 
  3.73095733937106*^9, 3.730957573517251*^9, 3.730957692549412*^9, 
  3.730957756021146*^9, 3.730958184697094*^9, 3.73095888847186*^9, 
  3.730959976146467*^9, 3.7309604145859423`*^9, 3.731003349194276*^9, 
  3.731003932156702*^9},
 CellLabel->"During evaluation of In[9]:=",
 CellID->2048806989,ExpressionUUID->"17e88f77-e430-48ca-9e2d-72ab261fd99b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{3, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "metricg"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`metricg[{3, -$CellContext`Sch}, {
     2, -$CellContext`Sch}] -> $CellContext`metricg[{2, -$CellContext`Sch}, {
     3, -$CellContext`Sch}], " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914339339313*^9, 3.730914956546998*^9, 
  3.73095733937106*^9, 3.730957573517251*^9, 3.730957692549412*^9, 
  3.730957756021146*^9, 3.730958184697094*^9, 3.73095888847186*^9, 
  3.730959976146467*^9, 3.7309604145859423`*^9, 3.731003349194276*^9, 
  3.731003932207671*^9},
 CellLabel->"During evaluation of In[9]:=",
 CellID->1546268125,ExpressionUUID->"225c5473-98f9-4750-ba1b-0e39db9f2343"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", 
   "metricg"}],
  SequenceForm[
  "Added independent rule ", $CellContext`metricg[{3, -$CellContext`Sch}, {
     3, -$CellContext`Sch}] -> $CellContext`r[]^2 Sin[
       $CellContext`\[Theta][]]^2, " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914339339313*^9, 3.730914956546998*^9, 
  3.73095733937106*^9, 3.730957573517251*^9, 3.730957692549412*^9, 
  3.730957756021146*^9, 3.730958184697094*^9, 3.73095888847186*^9, 
  3.730959976146467*^9, 3.7309604145859423`*^9, 3.731003349194276*^9, 
  3.731003932256949*^9},
 CellLabel->"During evaluation of In[9]:=",
 CellID->1428764528,ExpressionUUID->"ba47ef64-3872-4f89-ada1-ad7ff0199308"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{"-", "1"}], "+", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{1, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      FractionBox["1", 
       RowBox[{"1", "-", 
        FractionBox[
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`M,
           Editable->False]}], 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]]}]]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{2, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{2, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{3, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{3, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{3, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"], " ", 
       SuperscriptBox[
        RowBox[{
         InterpretationBox[
          StyleBox["Sin",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Sin,
          Editable->False], "[", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False], "]"}], "2"]}]}]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.730913971648592*^9, 3.730914340386692*^9, 
  3.730914957591959*^9, 3.7309573400721817`*^9, 3.730957574454195*^9, 
  3.730957693665423*^9, 3.730957757161455*^9, 3.730958185799819*^9, 
  3.730958889937704*^9, 3.730959977037339*^9, 3.730960415668604*^9, 
  3.7310033503705893`*^9, 3.731003932320017*^9},
 CellLabel->"Out[9]=",
 CellID->83281342,ExpressionUUID->"337ef31c-2790-46fb-8276-68ab12722b65"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
This is the object stored internally, that separates values due to symmetry \
from truly independent values:\
\>", "MathCaption",
 CellChangeTimes->{{3.731003378748698*^9, 3.731003405794421*^9}},
 CellID->688445837,ExpressionUUID->"db9f21eb-cbda-432e-b4e4-429ef7a6e87f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TensorValues", "[", "metricg", "]"}]], "Input",
 CellChangeTimes->{{3.731003367796548*^9, 3.731003372515414*^9}},
 CellLabel->"In[10]:=",
 CellID->266995355,ExpressionUUID->"ae1f1244-b072-44f1-9f7f-4be92dd310ab"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{1, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{2, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{2, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{3, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{3, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{3, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{"-", "1"}], "+", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      FractionBox["1", 
       RowBox[{"1", "-", 
        FractionBox[
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`M,
           Editable->False]}], 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]]}]]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"], " ", 
       SuperscriptBox[
        RowBox[{
         InterpretationBox[
          StyleBox["Sin",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Sin,
          Editable->False], "[", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False], "]"}], "2"]}]}]}], "}"}]}], "]"}]], "Output",
 CellChangeTimes->{3.731003372976207*^9, 3.7310039325920353`*^9},
 CellLabel->"Out[10]=",
 CellID->1697660797,ExpressionUUID->"9ea4b071-f1cb-4772-b6a5-7e982c22b50e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["In terms of a basis the metric takes the form:", "MathCaption",
 CellChangeTimes->{
  3.7309140688767357`*^9, {3.73100341370597*^9, 3.7310034149060802`*^9}},
 CellID->1033605800,ExpressionUUID->"a35c88ae-789a-4ea7-923c-148df8e07865"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TraceBasisDummy", "[", 
  RowBox[{
   RowBox[{"metricg", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "Sch"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "Sch"}]}], "}"}]}], "]"}], 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"-", "c"}], ",", 
     RowBox[{"{", 
      RowBox[{"a", ",", "Sch"}], "}"}]}], "]"}], 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"-", "d"}], ",", 
     RowBox[{"{", 
      RowBox[{"b", ",", "Sch"}], "}"}]}], "]"}]}], "]"}]], "Input",
 CellLabel->"In[11]:=",
 CellID->817460017,ExpressionUUID->"f4c55fc9-9d7a-4bcb-8803-d5cd1c98a54d"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {1, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{1, -$CellContext`Sch}, {0, -$CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {1, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{2, -$CellContext`Sch}, {0, -$CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {1, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{2, -$CellContext`Sch}, {1, -$CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{3, -$CellContext`Sch}, {0, -$CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {1, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{3, -$CellContext`Sch}, {1, -$CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{3, -$CellContext`Sch}, {2, -$CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{
  3.730913992199342*^9, {3.7309140266490192`*^9, 3.730914041673592*^9}, 
   3.730914105808708*^9, 3.730914385078683*^9, 3.730914957773316*^9, 
   3.7309573402454042`*^9, 3.7309575746510277`*^9, 3.730957693902625*^9, 
   3.730957757321456*^9, 3.7309581859934807`*^9, 3.730958890357711*^9, 
   3.730959977228107*^9, 3.730960415844363*^9, 3.7310034180033817`*^9, 
   3.7310039327473392`*^9},
 CellLabel->"Out[11]=",
 CellID->1117281137,ExpressionUUID->"19287cd2-9614-4ce9-8a5f-d9ef2a1137b4"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"downmetric", "[", 
   RowBox[{
    RowBox[{"-", "c_"}], ",", 
    RowBox[{"-", "d_"}]}], "]"}], "=", 
  RowBox[{"%", "/.", 
   RowBox[{"TensorValues", "[", "metricg", "]"}]}]}]], "Input",
 CellChangeTimes->{{3.7309140280375433`*^9, 3.730914038834407*^9}, 
   3.730914103218519*^9},
 CellLabel->"In[12]:=",
 CellID->1141720579,ExpressionUUID->"9037ccd1-e733-4e54-bee7-c26659a12f3c"],

Cell[BoxData[
 RowBox[{
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"d", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`d, {1, $CellContext`Sch}],
     Editable->False]}], 
   RowBox[{"1", "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}]], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {0, $CellContext`Sch}],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {2, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {3, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}]}]], "Output",
 CellChangeTimes->{3.730914109323456*^9, 3.730914385978706*^9, 
  3.730914957879756*^9, 3.73095734037306*^9, 3.730957574779385*^9, 
  3.730957694015346*^9, 3.7309577574248133`*^9, 3.7309581860979023`*^9, 
  3.730958890501206*^9, 3.730959977330749*^9, 3.730960415953459*^9, 
  3.731003419941785*^9, 3.731003932827561*^9},
 CellLabel->"Out[12]=",
 CellID->547060191,ExpressionUUID->"d65f6e42-6c36-4662-8040-665af2a08848"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We need now the expression of the inverse metric:", "MathCaption",
 CellChangeTimes->{{3.730914074763009*^9, 3.730914075139142*^9}},
 CellID->1654533724,ExpressionUUID->"dea562ea-196f-4a98-9f4d-af866b4a7184"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Inverse", "[", "matrix", "]"}], "//", "Simplify"}]], "Input",
 CellChangeTimes->{{3.731003424225897*^9, 3.731003425124744*^9}},
 CellLabel->"In[13]:=",
 CellID->2130037912,ExpressionUUID->"5ccff746-b94b-41b1-88c9-394f88189a47"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     FractionBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], 
      RowBox[{
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False]}], "-", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}]], ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", 
     RowBox[{"1", "-", 
      FractionBox[
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False]}], 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]]}], ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", 
     FractionBox["1", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"]], ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0", ",", 
     FractionBox[
      SuperscriptBox[
       RowBox[{
        InterpretationBox[
         StyleBox["Csc",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Csc,
         Editable->False], "[", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False], "]"}], "2"], 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"]]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.7309140767604*^9, 3.730914111570616*^9, 
  3.730914387745187*^9, 3.730914957983149*^9, 3.730957340495442*^9, 
  3.730957574886998*^9, 3.730957694138166*^9, 3.730957757527442*^9, 
  3.730958186216241*^9, 3.7309588906574287`*^9, 3.7309599774334173`*^9, 
  3.730960416088706*^9, 3.731003425927126*^9, 3.731003932913373*^9},
 CellLabel->"Out[13]=",
 CellID->1063137330,ExpressionUUID->"88c686ee-b966-4c1c-8236-36decb3730d2"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AllComponentValues", "[", 
  RowBox[{
   RowBox[{"metricg", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"a", ",", "Sch"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"b", ",", "Sch"}], "}"}]}], "]"}], ",", "%"}], "]"}]], "Input",
 CellLabel->"In[14]:=",
 CellID->1300447763,ExpressionUUID->"e9890cf1-ade1-41f1-a960-f8b3da51bf83"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{0, $CellContext`Sch}, {0, $CellContext`Sch}],
     Editable->False], "\[Rule]", 
    FractionBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}]]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "metricg"}],
  SequenceForm[
  "Added independent rule ", $CellContext`metricg[{0, $CellContext`Sch}, {
     0, $CellContext`Sch}] -> (
      2 $CellContext`M - $CellContext`r[])^(-1) $CellContext`r[], 
   " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.7309143885292263`*^9, 3.73091495810636*^9, 
  3.730957340657146*^9, 3.7309575750176573`*^9, 3.730957694286817*^9, 
  3.730957757643265*^9, 3.730958186349585*^9, 3.7309588908048677`*^9, 
  3.730959977562368*^9, 3.730960416222013*^9, 3.731003448133547*^9, 
  3.731003932992507*^9},
 CellLabel->"During evaluation of In[14]:=",
 CellID->1085162270,ExpressionUUID->"e9ccd3cf-0b7b-4173-a95a-9a04b6490b94"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{0, $CellContext`Sch}, {1, $CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", 
   "metricg"}],
  SequenceForm[
  "Added independent rule ", $CellContext`metricg[{0, $CellContext`Sch}, {
     1, $CellContext`Sch}] -> 0, " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.7309143885292263`*^9, 3.73091495810636*^9, 
  3.730957340657146*^9, 3.7309575750176573`*^9, 3.730957694286817*^9, 
  3.730957757643265*^9, 3.730958186349585*^9, 3.7309588908048677`*^9, 
  3.730959977562368*^9, 3.730960416222013*^9, 3.731003448133547*^9, 
  3.731003933014757*^9},
 CellLabel->"During evaluation of In[14]:=",
 CellID->1891141566,ExpressionUUID->"d197abeb-0fb9-4c29-baba-12865df68114"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{0, $CellContext`Sch}, {2, $CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", 
   "metricg"}],
  SequenceForm[
  "Added independent rule ", $CellContext`metricg[{0, $CellContext`Sch}, {
     2, $CellContext`Sch}] -> 0, " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.7309143885292263`*^9, 3.73091495810636*^9, 
  3.730957340657146*^9, 3.7309575750176573`*^9, 3.730957694286817*^9, 
  3.730957757643265*^9, 3.730958186349585*^9, 3.7309588908048677`*^9, 
  3.730959977562368*^9, 3.730960416222013*^9, 3.731003448133547*^9, 
  3.731003933077058*^9},
 CellLabel->"During evaluation of In[14]:=",
 CellID->1658324162,ExpressionUUID->"7e5edb88-3a98-42da-a1b0-4791173bdceb"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{0, $CellContext`Sch}, {3, $CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", 
   "metricg"}],
  SequenceForm[
  "Added independent rule ", $CellContext`metricg[{0, $CellContext`Sch}, {
     3, $CellContext`Sch}] -> 0, " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.7309143885292263`*^9, 3.73091495810636*^9, 
  3.730957340657146*^9, 3.7309575750176573`*^9, 3.730957694286817*^9, 
  3.730957757643265*^9, 3.730958186349585*^9, 3.7309588908048677`*^9, 
  3.730959977562368*^9, 3.730960416222013*^9, 3.731003448133547*^9, 
  3.731003933143196*^9},
 CellLabel->"During evaluation of In[14]:=",
 CellID->1411419968,ExpressionUUID->"ce0c42e4-b0e5-43a7-aa9f-d5b07676703d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{1, $CellContext`Sch}, {0, $CellContext`Sch}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{0, $CellContext`Sch}, {1, $CellContext`Sch}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "metricg"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`metricg[{1, $CellContext`Sch}, {
     0, $CellContext`Sch}] -> $CellContext`metricg[{0, $CellContext`Sch}, {
     1, $CellContext`Sch}], " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.7309143885292263`*^9, 3.73091495810636*^9, 
  3.730957340657146*^9, 3.7309575750176573`*^9, 3.730957694286817*^9, 
  3.730957757643265*^9, 3.730958186349585*^9, 3.7309588908048677`*^9, 
  3.730959977562368*^9, 3.730960416222013*^9, 3.731003448133547*^9, 
  3.731003933211135*^9},
 CellLabel->"During evaluation of In[14]:=",
 CellID->631037414,ExpressionUUID->"8f1c470e-9926-48e9-8bc8-c88435430962"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{1, $CellContext`Sch}, {1, $CellContext`Sch}],
     Editable->False], "\[Rule]", 
    RowBox[{"1", "-", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]]}]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "metricg"}],
  SequenceForm[
  "Added independent rule ", $CellContext`metricg[{1, $CellContext`Sch}, {
     1, $CellContext`Sch}] -> 1 - 2 ($CellContext`M/$CellContext`r[]), 
   " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.7309143885292263`*^9, 3.73091495810636*^9, 
  3.730957340657146*^9, 3.7309575750176573`*^9, 3.730957694286817*^9, 
  3.730957757643265*^9, 3.730958186349585*^9, 3.7309588908048677`*^9, 
  3.730959977562368*^9, 3.730960416222013*^9, 3.731003448133547*^9, 
  3.73100393327673*^9},
 CellLabel->"During evaluation of In[14]:=",
 CellID->1571568212,ExpressionUUID->"931149c6-557b-4fa8-9eb9-d65c32fb82f3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{1, $CellContext`Sch}, {2, $CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", 
   "metricg"}],
  SequenceForm[
  "Added independent rule ", $CellContext`metricg[{1, $CellContext`Sch}, {
     2, $CellContext`Sch}] -> 0, " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.7309143885292263`*^9, 3.73091495810636*^9, 
  3.730957340657146*^9, 3.7309575750176573`*^9, 3.730957694286817*^9, 
  3.730957757643265*^9, 3.730958186349585*^9, 3.7309588908048677`*^9, 
  3.730959977562368*^9, 3.730960416222013*^9, 3.731003448133547*^9, 
  3.731003933343315*^9},
 CellLabel->"During evaluation of In[14]:=",
 CellID->1445564631,ExpressionUUID->"4431e7ca-7565-47a6-9337-f8757e02e1aa"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{1, $CellContext`Sch}, {3, $CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", 
   "metricg"}],
  SequenceForm[
  "Added independent rule ", $CellContext`metricg[{1, $CellContext`Sch}, {
     3, $CellContext`Sch}] -> 0, " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.7309143885292263`*^9, 3.73091495810636*^9, 
  3.730957340657146*^9, 3.7309575750176573`*^9, 3.730957694286817*^9, 
  3.730957757643265*^9, 3.730958186349585*^9, 3.7309588908048677`*^9, 
  3.730959977562368*^9, 3.730960416222013*^9, 3.731003448133547*^9, 
  3.731003933409721*^9},
 CellLabel->"During evaluation of In[14]:=",
 CellID->1122254706,ExpressionUUID->"f1d6c299-882f-4a4e-97ce-f887e74899c4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{2, $CellContext`Sch}, {0, $CellContext`Sch}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{0, $CellContext`Sch}, {2, $CellContext`Sch}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "metricg"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`metricg[{2, $CellContext`Sch}, {
     0, $CellContext`Sch}] -> $CellContext`metricg[{0, $CellContext`Sch}, {
     2, $CellContext`Sch}], " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.7309143885292263`*^9, 3.73091495810636*^9, 
  3.730957340657146*^9, 3.7309575750176573`*^9, 3.730957694286817*^9, 
  3.730957757643265*^9, 3.730958186349585*^9, 3.7309588908048677`*^9, 
  3.730959977562368*^9, 3.730960416222013*^9, 3.731003448133547*^9, 
  3.731003933477273*^9},
 CellLabel->"During evaluation of In[14]:=",
 CellID->385972141,ExpressionUUID->"6e2beb36-a95d-4958-b5b3-45fe68644a37"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{2, $CellContext`Sch}, {1, $CellContext`Sch}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{1, $CellContext`Sch}, {2, $CellContext`Sch}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "metricg"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`metricg[{2, $CellContext`Sch}, {
     1, $CellContext`Sch}] -> $CellContext`metricg[{1, $CellContext`Sch}, {
     2, $CellContext`Sch}], " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.7309143885292263`*^9, 3.73091495810636*^9, 
  3.730957340657146*^9, 3.7309575750176573`*^9, 3.730957694286817*^9, 
  3.730957757643265*^9, 3.730958186349585*^9, 3.7309588908048677`*^9, 
  3.730959977562368*^9, 3.730960416222013*^9, 3.731003448133547*^9, 
  3.731003933543029*^9},
 CellLabel->"During evaluation of In[14]:=",
 CellID->1633159847,ExpressionUUID->"945977cc-869e-4cb3-8857-c2e025dca29d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{2, $CellContext`Sch}, {2, $CellContext`Sch}],
     Editable->False], "\[Rule]", 
    FractionBox["1", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]]}], "\[InvisibleSpace]", "\<\" for tensor \"\>",
    "\[InvisibleSpace]", "metricg"}],
  SequenceForm[
  "Added independent rule ", $CellContext`metricg[{2, $CellContext`Sch}, {
     2, $CellContext`Sch}] -> $CellContext`r[]^(-2), 
   " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.7309143885292263`*^9, 3.73091495810636*^9, 
  3.730957340657146*^9, 3.7309575750176573`*^9, 3.730957694286817*^9, 
  3.730957757643265*^9, 3.730958186349585*^9, 3.7309588908048677`*^9, 
  3.730959977562368*^9, 3.730960416222013*^9, 3.731003448133547*^9, 
  3.731003933610505*^9},
 CellLabel->"During evaluation of In[14]:=",
 CellID->1529056222,ExpressionUUID->"1ce04239-f242-4110-8d47-4ee653c9a746"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{2, $CellContext`Sch}, {3, $CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], 
   "\[InvisibleSpace]", "\<\" for tensor \"\>", "\[InvisibleSpace]", 
   "metricg"}],
  SequenceForm[
  "Added independent rule ", $CellContext`metricg[{2, $CellContext`Sch}, {
     3, $CellContext`Sch}] -> 0, " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.7309143885292263`*^9, 3.73091495810636*^9, 
  3.730957340657146*^9, 3.7309575750176573`*^9, 3.730957694286817*^9, 
  3.730957757643265*^9, 3.730958186349585*^9, 3.7309588908048677`*^9, 
  3.730959977562368*^9, 3.730960416222013*^9, 3.731003448133547*^9, 
  3.731003933677045*^9},
 CellLabel->"During evaluation of In[14]:=",
 CellID->2058603152,ExpressionUUID->"88083b96-439c-4d7e-a3c7-4eba612d10dd"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{3, $CellContext`Sch}, {0, $CellContext`Sch}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{0, $CellContext`Sch}, {3, $CellContext`Sch}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "metricg"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`metricg[{3, $CellContext`Sch}, {
     0, $CellContext`Sch}] -> $CellContext`metricg[{0, $CellContext`Sch}, {
     3, $CellContext`Sch}], " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.7309143885292263`*^9, 3.73091495810636*^9, 
  3.730957340657146*^9, 3.7309575750176573`*^9, 3.730957694286817*^9, 
  3.730957757643265*^9, 3.730958186349585*^9, 3.7309588908048677`*^9, 
  3.730959977562368*^9, 3.730960416222013*^9, 3.731003448133547*^9, 
  3.731003933743004*^9},
 CellLabel->"During evaluation of In[14]:=",
 CellID->834138847,ExpressionUUID->"6ada885c-c343-49ec-b74c-f3726a81acdb"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{3, $CellContext`Sch}, {1, $CellContext`Sch}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{1, $CellContext`Sch}, {3, $CellContext`Sch}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "metricg"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`metricg[{3, $CellContext`Sch}, {
     1, $CellContext`Sch}] -> $CellContext`metricg[{1, $CellContext`Sch}, {
     3, $CellContext`Sch}], " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.7309143885292263`*^9, 3.73091495810636*^9, 
  3.730957340657146*^9, 3.7309575750176573`*^9, 3.730957694286817*^9, 
  3.730957757643265*^9, 3.730958186349585*^9, 3.7309588908048677`*^9, 
  3.730959977562368*^9, 3.730960416222013*^9, 3.731003448133547*^9, 
  3.731003933810371*^9},
 CellLabel->"During evaluation of In[14]:=",
 CellID->598353913,ExpressionUUID->"aeea2c5f-5c98-4684-94fd-0b1d556b007d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added dependent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{3, $CellContext`Sch}, {2, $CellContext`Sch}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{2, $CellContext`Sch}, {3, $CellContext`Sch}],
     Editable->False]}], "\[InvisibleSpace]", "\<\" for tensor \"\>", 
   "\[InvisibleSpace]", "metricg"}],
  SequenceForm[
  "Added dependent rule ", $CellContext`metricg[{3, $CellContext`Sch}, {
     2, $CellContext`Sch}] -> $CellContext`metricg[{2, $CellContext`Sch}, {
     3, $CellContext`Sch}], " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.7309143885292263`*^9, 3.73091495810636*^9, 
  3.730957340657146*^9, 3.7309575750176573`*^9, 3.730957694286817*^9, 
  3.730957757643265*^9, 3.730958186349585*^9, 3.7309588908048677`*^9, 
  3.730959977562368*^9, 3.730960416222013*^9, 3.731003448133547*^9, 
  3.731003933876618*^9},
 CellLabel->"During evaluation of In[14]:=",
 CellID->2015780537,ExpressionUUID->"4144fcd5-eabe-4960-a485-87dcd80e2732"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Added independent rule \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[{3, $CellContext`Sch}, {3, $CellContext`Sch}],
     Editable->False], "\[Rule]", 
    FractionBox[
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Csc",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Csc,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"], 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]]}], "\[InvisibleSpace]", "\<\" for tensor \"\>",
    "\[InvisibleSpace]", "metricg"}],
  SequenceForm[
  "Added independent rule ", $CellContext`metricg[{3, $CellContext`Sch}, {
     3, $CellContext`Sch}] -> Csc[
       $CellContext`\[Theta][]]^2 $CellContext`r[]^(-2), 
   " for tensor ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.7309143885292263`*^9, 3.73091495810636*^9, 
  3.730957340657146*^9, 3.7309575750176573`*^9, 3.730957694286817*^9, 
  3.730957757643265*^9, 3.730958186349585*^9, 3.7309588908048677`*^9, 
  3.730959977562368*^9, 3.730960416222013*^9, 3.731003448133547*^9, 
  3.7310039339436617`*^9},
 CellLabel->"During evaluation of In[14]:=",
 CellID->521470287,ExpressionUUID->"cd14d3f2-d6a7-45df-9557-b8b7307a8700"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{1, $CellContext`Sch}, {0, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{0, $CellContext`Sch}, {1, $CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{2, $CellContext`Sch}, {0, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{0, $CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{2, $CellContext`Sch}, {1, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{1, $CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{3, $CellContext`Sch}, {0, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{0, $CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{3, $CellContext`Sch}, {1, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{1, $CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{3, $CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{2, $CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{0, $CellContext`Sch}, {0, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      FractionBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], 
       RowBox[{
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], "-", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]}]]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{0, $CellContext`Sch}, {1, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{0, $CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{0, $CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{1, $CellContext`Sch}, {1, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      RowBox[{"1", "-", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{1, $CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{1, $CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{2, $CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      FractionBox["1", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"]]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{2, $CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{3, $CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      FractionBox[
       SuperscriptBox[
        RowBox[{
         InterpretationBox[
          StyleBox["Csc",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Csc,
          Editable->False], "[", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False], "]"}], "2"], 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"]]}]}], "}"}]}], "]"}]], "Output",
 CellChangeTimes->{3.7309140801021957`*^9, 3.730914112540588*^9, 
  3.730914389776279*^9, 3.730914959044139*^9, 3.730957342306266*^9, 
  3.7309575771038113`*^9, 3.730957695407596*^9, 3.7309577586761*^9, 
  3.7309581873314657`*^9, 3.730958892340783*^9, 3.730959978397502*^9, 
  3.730960417391567*^9, 3.7310034494644117`*^9, 3.731003934006259*^9},
 CellLabel->"Out[14]=",
 CellID->470350202,ExpressionUUID->"293f1f55-9d80-4b8a-9404-844e03a7ea6f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["In terms of a basis this is:", "MathCaption",
 CellChangeTimes->{3.7309140900424757`*^9},
 CellID->391610265,ExpressionUUID->"47a75909-a629-40f3-98b6-27b877bf984a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TraceBasisDummy", "[", 
  RowBox[{
   RowBox[{"metricg", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"a", ",", "Sch"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"b", ",", "Sch"}], "}"}]}], "]"}], 
   RowBox[{"Basis", "[", 
    RowBox[{"c", ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "Sch"}]}], "}"}]}], "]"}], 
   RowBox[{"Basis", "[", 
    RowBox[{"d", ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "Sch"}]}], "}"}]}], "]"}]}], "]"}]], "Input",
 CellLabel->"In[15]:=",
 CellID->1903829447,ExpressionUUID->"e315f4e6-a244-40b9-acdd-5ed984fee8db"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "c"},
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{0, $CellContext`Sch}, {0, $CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "c"},
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{0, $CellContext`Sch}, {1, $CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "c"},
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{0, $CellContext`Sch}, {2, $CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "c"},
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{0, $CellContext`Sch}, {3, $CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "c"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{1, $CellContext`Sch}, {0, $CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "c"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{1, $CellContext`Sch}, {1, $CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "c"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{1, $CellContext`Sch}, {2, $CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "c"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{1, $CellContext`Sch}, {3, $CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "c"},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{2, $CellContext`Sch}, {0, $CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "c"},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{2, $CellContext`Sch}, {1, $CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "c"},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{2, $CellContext`Sch}, {2, $CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "c"},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{2, $CellContext`Sch}, {3, $CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "c"},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{3, $CellContext`Sch}, {0, $CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "c"},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{3, $CellContext`Sch}, {1, $CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "c"},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{3, $CellContext`Sch}, {2, $CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "c"},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{3, $CellContext`Sch}, {3, $CellContext`Sch}],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{{3.73091409163104*^9, 3.7309141143914633`*^9}, 
   3.730914393034286*^9, 3.730914959202827*^9, 3.730957342456093*^9, 
   3.730957577304572*^9, 3.730957695569672*^9, 3.7309577588537827`*^9, 
   3.730958187513464*^9, 3.7309588925351753`*^9, 3.730959978596595*^9, 
   3.730960417583912*^9, 3.731003456321423*^9, 3.731003934130014*^9},
 CellLabel->"Out[15]=",
 CellID->288540407,ExpressionUUID->"ba691a27-b34b-47ae-b29c-16e3d257d1f3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"upmetric", "[", 
   RowBox[{"c_", ",", "d_"}], "]"}], "=", 
  RowBox[{"%", "/.", 
   RowBox[{"TensorValues", "[", "metricg", "]"}]}]}]], "Input",
 CellChangeTimes->{{3.730914093066718*^9, 3.730914099514428*^9}},
 CellLabel->"In[16]:=",
 CellID->988287745,ExpressionUUID->"4f43098e-8051-42f3-9647-ec8d9e212cb6"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "c"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{"1", "-", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]]}], ")"}]}], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "c"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`c],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "c"},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`c],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Csc",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Csc,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"]}], 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "c"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`c],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False]}], 
   RowBox[{
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`M,
      Editable->False]}], "-", 
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False]}]]}]], "Output",
 CellChangeTimes->{3.730914117357334*^9, 3.730914393779914*^9, 
  3.7309149593279552`*^9, 3.730957342573893*^9, 3.730957577497908*^9, 
  3.730957695695129*^9, 3.73095775897215*^9, 3.7309581876140757`*^9, 
  3.7309588926688213`*^9, 3.730959978829813*^9, 3.730960417762561*^9, 
  3.7310034573852177`*^9, 3.7310039342094793`*^9},
 CellLabel->"Out[16]=",
 CellID->480545523,ExpressionUUID->"6f7ce65f-05af-4a7e-9748-fe4f42682e7d"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Alternatively we can also compute:", "MathCaption",
 CellChangeTimes->{{3.731003850484971*^9, 3.7310038564995193`*^9}},
 CellID->1214767293,ExpressionUUID->"b8daa658-60fe-4744-a16f-44a28f731661"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"downmetric", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.731003857531158*^9, 3.7310038608754377`*^9}},
 CellLabel->"In[17]:=",
 CellID->186812753,ExpressionUUID->"68ccf8a3-140b-4064-92ae-23d2e2ab8d4b"],

Cell[BoxData[
 RowBox[{
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False]}], 
   RowBox[{"1", "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}]], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Sch}],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}]}]], "Output",
 CellChangeTimes->{3.731003861256324*^9, 3.731003934295456*^9},
 CellLabel->"Out[17]=",
 CellID->1990092341,ExpressionUUID->"f7803d46-d930-48f3-9536-4fdf12d40f17"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"InverseMetric", "[", 
   RowBox[{"%", ",", "Sch"}], "]"}], "[", 
  RowBox[{"c", ",", "d"}], "]"}]], "Input",
 CellChangeTimes->{{3.731003865884714*^9, 3.731003879802853*^9}},
 CellLabel->"In[18]:=",
 CellID->672927078,ExpressionUUID->"c7f0b36a-94cf-49a4-ac8a-294a5f5e6156"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density \"\>", "\[InvisibleSpace]", 
   RowBox[{"DetmetricgSch", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Determinant.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "weight +2 density ", 
   $CellContext`DetmetricgSch[], ". ", "Determinant."],
  Editable->False]], "Print",
 CellChangeTimes->{3.731003934376285*^9},
 CellLabel->"During evaluation of In[18]:=",
 CellID->424173162,ExpressionUUID->"fde3957a-907b-4933-bd02-db9450d5e44b"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "c"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{"1", "-", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]]}], ")"}]}], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "c"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`c],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "c"},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`c],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Csc",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Csc,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"]}], 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "c"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`c],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False]}], 
   RowBox[{
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`M,
      Editable->False]}], "-", 
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False]}]]}]], "Output",
 CellChangeTimes->{3.7310038811092577`*^9, 3.731003934442889*^9},
 CellLabel->"Out[18]=",
 CellID->1580109341,ExpressionUUID->"6ff22049-0e7b-4114-b12e-158014aee0d1"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "\[Equal]", 
  RowBox[{"upmetric", "[", 
   RowBox[{"c", ",", "d"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.731003893482373*^9, 3.731003897490139*^9}},
 CellLabel->"In[19]:=",
 CellID->2077978809,ExpressionUUID->"825e09ba-c183-473f-8b98-1aacb88b5d68"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.7310038978858624`*^9, 3.731003934520006*^9},
 CellLabel->"Out[19]=",
 CellID->1114915385,ExpressionUUID->"3983fb94-1150-42d6-9880-f7de7211397f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Consistency check:", "MathCaption",
 CellChangeTimes->{{3.730914131578691*^9, 3.730914134721081*^9}},
 CellID->1443006693,ExpressionUUID->"6533f8c7-0eef-44a0-a072-900d70a60381"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"upmetric", "[", 
   RowBox[{"a", ",", "b"}], "]"}], 
  RowBox[{"downmetric", "[", 
   RowBox[{
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.730914138287372*^9, 3.730914146152965*^9}},
 CellLabel->"In[20]:=",
 CellID->504774293,ExpressionUUID->"f97f4cb7-999c-4d0f-9ac1-f82a461c5964"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", "a"},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", "b"},
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`b],
      Editable->False], " ", 
     RowBox[{"(", 
      RowBox[{"1", "-", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]]}], ")"}]}], "+", 
    FractionBox[
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", "a"},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", "b"},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`b],
       Editable->False]}], 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]], "+", 
    FractionBox[
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", "a"},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", "b"},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`b],
       Editable->False], " ", 
      SuperscriptBox[
       RowBox[{
        InterpretationBox[
         StyleBox["Csc",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Csc,
         Editable->False], "[", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False], "]"}], "2"]}], 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]], "+", 
    FractionBox[
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", "a"},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", "b"},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}]]}], ")"}], " ", 
  RowBox[{"(", 
   RowBox[{
    FractionBox[
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {"b", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {"c", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
       Editable->False]}], 
     RowBox[{"1", "-", 
      FractionBox[
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False]}], 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]]}]], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Sch}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]},
             {"c", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
      Editable->False], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]]}], ")"}]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]},
             {"c", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]]},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]]},
             {"c", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], ")"}]}]], "Output",
 CellChangeTimes->{{3.730914146639073*^9, 3.730914180683028*^9}, 
   3.730914395088993*^9, 3.730914959433223*^9, 3.730957342685528*^9, 
   3.730957577937005*^9, 3.730957695805067*^9, 3.730957759084074*^9, 
   3.730958187721189*^9, 3.730958892797443*^9, 3.730959979013028*^9, 
   3.730960417940198*^9, 3.7310034605631647`*^9, 3.731003934597843*^9},
 CellLabel->"Out[20]=",
 CellID->1588055620,ExpressionUUID->"6df7bdcb-98fa-4d3e-af7e-0683ae3413b9"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ContractBasis"}]], "Input",
 CellChangeTimes->{{3.7309141515388536`*^9, 3.730914183431643*^9}},
 CellLabel->"In[21]:=",
 CellID->1994969649,ExpressionUUID->"c87e92d4-639d-4160-8660-feadd6cb5ea0"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`a],
    Editable->False]}], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`a],
     Editable->False]}], 
   RowBox[{"1", "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}]], "+", 
  FractionBox[
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`a],
     Editable->False]}], 
   RowBox[{
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`M,
      Editable->False]}], "-", 
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False]}]], "-", 
  FractionBox[
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`a],
     Editable->False]}], 
   RowBox[{
    RowBox[{"(", 
     RowBox[{"1", "-", 
      FractionBox[
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False]}], 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]]}], ")"}], " ", 
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False]}]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`a],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False]}], 
   RowBox[{
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`M,
      Editable->False]}], "-", 
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False]}]]}]], "Output",
 CellChangeTimes->{{3.7309141545709*^9, 3.730914183856633*^9}, 
   3.7309143958304977`*^9, 3.7309149595686197`*^9, 3.730957342816224*^9, 
   3.730957578073122*^9, 3.730957695938731*^9, 3.730957759214123*^9, 
   3.730958187852816*^9, 3.730958892944668*^9, 3.730959979151284*^9, 
   3.730960418078347*^9, 3.73100346231586*^9, 3.731003934702231*^9},
 CellLabel->"Out[21]=",
 CellID->1524040424,ExpressionUUID->"702c0622-2f8a-401b-bf9b-ed57e064661a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "Simplify"}]], "Input",
 CellChangeTimes->{{3.730914188025715*^9, 3.7309141897914543`*^9}},
 CellLabel->"In[22]:=",
 CellID->170760014,ExpressionUUID->"59179072-ce57-40db-9984-fbc3ea9c6b3b"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`a],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.730914190344841*^9, 3.730914396542982*^9, 
  3.73091495970018*^9, 3.7309573429923162`*^9, 3.7309575782042513`*^9, 
  3.730957696068388*^9, 3.730957759337339*^9, 3.730958187983204*^9, 
  3.730958893085223*^9, 3.730959979283805*^9, 3.730960418238213*^9, 
  3.731003463714053*^9, 3.7310039347803183`*^9},
 CellLabel->"Out[22]=",
 CellID->1509948399,ExpressionUUID->"258f703b-c675-4255-8830-c64d7f1bb0ec"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["From now on we do not output the ComponentValue messages:", \
"MathCaption",
 CellID->195861224,ExpressionUUID->"31eda655-aedb-4748-8039-bb53d0bb8a81"],

Cell[BoxData[
 RowBox[{
  RowBox[{"$CVVerbose", "=", "False"}], ";"}]], "Input",
 CellLabel->"In[23]:=",
 CellID->1894644564,ExpressionUUID->"045692e0-1376-44dc-b294-2b4d58667611"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["4.12.3. The Kl\[ODoubleDot]sch-Strobl coordinate system", \
"Subsubsection",
 CellChangeTimes->{{3.7286771125390873`*^9, 3.7286771127222157`*^9}, {
  3.730913913338131*^9, 3.730913941103621*^9}, {3.730914255302225*^9, 
  3.730914260915913*^9}},
 CellID->1072254693,ExpressionUUID->"e2af22ec-2645-4aa6-ac12-6ff329d0baad"],

Cell["\<\
As an example of basis change, we introduce an alternative coordinate system. \
See \"Explicit global coordinates for Schwarzschild and Reissner-Nordstr\
\[OSlash]m solutions\", Thomas Kl\[ODoubleDot]sch and Thomas Strobl Class. \
Quantum Grav. 13 (1996) 1191--1199.\
\>", "Text",
 CellChangeTimes->{{3.730914282068097*^9, 3.73091428533913*^9}, {
  3.7310034816880493`*^9, 3.7310034971593103`*^9}},
 CellID->1810798472,ExpressionUUID->"76f1c64b-eadc-4587-ae4e-ddc5e619bc66"],

Cell[CellGroupData[{

Cell["Define the new chart:", "MathCaption",
 CellChangeTimes->{{3.7309142880998707`*^9, 3.730914291475875*^9}},
 CellID->103387443,ExpressionUUID->"4c92cc5d-73f1-4cc7-88df-11bd4d5428ab"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefChart", "[", 
  RowBox[{"Klosch", ",", "M4", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2", ",", "3"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"x", "[", "]"}], ",", 
     RowBox[{"y", "[", "]"}], ",", 
     RowBox[{"\[CapitalTheta]", "[", "]"}], ",", 
     RowBox[{"\[CapitalPhi]", "[", "]"}]}], "}"}], ",", 
   RowBox[{"ChartColor", "\[Rule]", "Blue"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.730914293900119*^9, 3.730914296116062*^9}},
 CellLabel->"In[24]:=",
 CellID->2099157267,ExpressionUUID->"f6b2ba2f-1eed-4e3d-94ba-c704196f5c70"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefChart", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"chart \"\>", "\[InvisibleSpace]", "Klosch", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xCoba`DefChart, ": Defining ", "chart ", $CellContext`Klosch, 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914420030834*^9, 3.730914963852853*^9, 
  3.730957343179661*^9, 3.730957578486018*^9, 3.730957696250476*^9, 
  3.730957759493087*^9, 3.730958190431542*^9, 3.730958893305811*^9, 
  3.7309599794790688`*^9, 3.730960418437182*^9, 3.731003508441923*^9, 
  3.7310039567554817`*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->1125860365,ExpressionUUID->"63f668cc-96d7-4259-b2aa-ca8d5448b785"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"x", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "coordinate scalar ", 
   $CellContext`x[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914420030834*^9, 3.730914963852853*^9, 
  3.730957343179661*^9, 3.730957578486018*^9, 3.730957696250476*^9, 
  3.730957759493087*^9, 3.730958190431542*^9, 3.730958893305811*^9, 
  3.7309599794790688`*^9, 3.730960418437182*^9, 3.731003508441923*^9, 
  3.73100395676964*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->856740838,ExpressionUUID->"2133245e-400d-4ee8-8a9f-0c55be745469"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"y", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "coordinate scalar ", 
   $CellContext`y[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914420030834*^9, 3.730914963852853*^9, 
  3.730957343179661*^9, 3.730957578486018*^9, 3.730957696250476*^9, 
  3.730957759493087*^9, 3.730958190431542*^9, 3.730958893305811*^9, 
  3.7309599794790688`*^9, 3.730960418437182*^9, 3.731003508441923*^9, 
  3.731003956783559*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->365411131,ExpressionUUID->"78bbea23-bcba-4852-9bd6-6e3bc2b2d3db"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[CapitalTheta]", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "coordinate scalar ", 
   $CellContext`\[CapitalTheta][], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914420030834*^9, 3.730914963852853*^9, 
  3.730957343179661*^9, 3.730957578486018*^9, 3.730957696250476*^9, 
  3.730957759493087*^9, 3.730958190431542*^9, 3.730958893305811*^9, 
  3.7309599794790688`*^9, 3.730960418437182*^9, 3.731003508441923*^9, 
  3.731003956793428*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->1809131944,ExpressionUUID->"88d36984-02fe-47ca-9fba-48e4ac357eb5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[CapitalPhi]", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "coordinate scalar ", 
   $CellContext`\[CapitalPhi][], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914420030834*^9, 3.730914963852853*^9, 
  3.730957343179661*^9, 3.730957578486018*^9, 3.730957696250476*^9, 
  3.730957759493087*^9, 3.730958190431542*^9, 3.730958893305811*^9, 
  3.7309599794790688`*^9, 3.730960418437182*^9, 3.731003508441923*^9, 
  3.731003956802993*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->2007108582,ExpressionUUID->"729b3a0e-0f32-4d1c-bbc1-ab884d082d71"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefMapping", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"mapping \"\>", "\[InvisibleSpace]", "Klosch", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefMapping, ": Defining ", 
   "mapping ", $CellContext`Klosch, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914420030834*^9, 3.730914963852853*^9, 
  3.730957343179661*^9, 3.730957578486018*^9, 3.730957696250476*^9, 
  3.730957759493087*^9, 3.730958190431542*^9, 3.730958893305811*^9, 
  3.7309599794790688`*^9, 3.730960418437182*^9, 3.731003508441923*^9, 
  3.731003956812639*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->800810598,ExpressionUUID->"beb3f829-5e6a-4245-a869-fff699f10aac"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefMapping", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"inverse mapping \"\>", "\[InvisibleSpace]", 
   "iKlosch", "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefMapping, ": Defining ", 
   "inverse mapping ", $CellContext`iKlosch, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914420030834*^9, 3.730914963852853*^9, 
  3.730957343179661*^9, 3.730957578486018*^9, 3.730957696250476*^9, 
  3.730957759493087*^9, 3.730958190431542*^9, 3.730958893305811*^9, 
  3.7309599794790688`*^9, 3.730960418437182*^9, 3.731003508441923*^9, 
  3.7310039568226137`*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->943704837,ExpressionUUID->"8d474cbe-254a-4b72-8e88-0f4f7b6370c3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"mapping differential tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"diKlosch", "[", 
    RowBox[{
     RowBox[{"-", "\[FormalA]"}], ",", "iKloscha"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "mapping differential tensor ", 
   $CellContext`diKlosch[-\[FormalA], $CellContext`iKloscha], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914420030834*^9, 3.730914963852853*^9, 
  3.730957343179661*^9, 3.730957578486018*^9, 3.730957696250476*^9, 
  3.730957759493087*^9, 3.730958190431542*^9, 3.730958893305811*^9, 
  3.7309599794790688`*^9, 3.730960418437182*^9, 3.731003508441923*^9, 
  3.731003956833219*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->1680462364,ExpressionUUID->"473156ca-3379-4b66-a6eb-d33c248e41c7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"mapping differential tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"dKlosch", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", "Klosch\[FormalA]"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "mapping differential tensor ", 
   $CellContext`dKlosch[-$CellContext`a, $CellContext`Klosch\[FormalA]], ". ",
    ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914420030834*^9, 3.730914963852853*^9, 
  3.730957343179661*^9, 3.730957578486018*^9, 3.730957696250476*^9, 
  3.730957759493087*^9, 3.730958190431542*^9, 3.730958893305811*^9, 
  3.7309599794790688`*^9, 3.730960418437182*^9, 3.731003508441923*^9, 
  3.731003956842938*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->1973510051,ExpressionUUID->"21dcc1f6-1e2b-4b00-a35a-4746829d27f4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefBasis", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"basis \"\>", "\[InvisibleSpace]", "Klosch", 
   "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Coordinated basis.\"\>"}],
  SequenceForm[
  "** ", xAct`xCoba`DefBasis, ": Defining ", "basis ", $CellContext`Klosch, 
   ". ", "Coordinated basis."],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914420030834*^9, 3.730914963852853*^9, 
  3.730957343179661*^9, 3.730957578486018*^9, 3.730957696250476*^9, 
  3.730957759493087*^9, 3.730958190431542*^9, 3.730958893305811*^9, 
  3.7309599794790688`*^9, 3.730960418437182*^9, 3.731003508441923*^9, 
  3.731003956852449*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->898499242,ExpressionUUID->"190d0943-f739-4717-b8a5-b272efcc2590"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"parallel derivative \"\>", "\[InvisibleSpace]", 
   RowBox[{"PDKlosch", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "parallel derivative ", 
   $CellContext`PDKlosch[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914420030834*^9, 3.730914963852853*^9, 
  3.730957343179661*^9, 3.730957578486018*^9, 3.730957696250476*^9, 
  3.730957759493087*^9, 3.730958190431542*^9, 3.730958893305811*^9, 
  3.7309599794790688`*^9, 3.730960418437182*^9, 3.731003508441923*^9, 
  3.7310039568620243`*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->2119238237,ExpressionUUID->"435580d6-da6a-42af-b32f-75aee4f41944"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing torsion tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TorsionPDKlosch", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing torsion tensor ", 
   $CellContext`TorsionPDKlosch[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914420030834*^9, 3.730914963852853*^9, 
  3.730957343179661*^9, 3.730957578486018*^9, 3.730957696250476*^9, 
  3.730957759493087*^9, 3.730958190431542*^9, 3.730958893305811*^9, 
  3.7309599794790688`*^9, 3.730960418437182*^9, 3.731003508441923*^9, 
  3.731003956871571*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->2020763764,ExpressionUUID->"22674112-98f3-42ae-ba2f-07a6e857c108"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelPDKlosch", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Christoffel tensor ", 
   $CellContext`ChristoffelPDKlosch[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914420030834*^9, 3.730914963852853*^9, 
  3.730957343179661*^9, 3.730957578486018*^9, 3.730957696250476*^9, 
  3.730957759493087*^9, 3.730958190431542*^9, 3.730958893305811*^9, 
  3.7309599794790688`*^9, 3.730960418437182*^9, 3.731003508441923*^9, 
  3.731003956881139*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->522704243,ExpressionUUID->"c1d03025-7b88-48ad-a481-36501e0ff4e5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Riemann tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RiemannPDKlosch", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Riemann tensor ", 
   $CellContext`RiemannPDKlosch[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, $CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914420030834*^9, 3.730914963852853*^9, 
  3.730957343179661*^9, 3.730957578486018*^9, 3.730957696250476*^9, 
  3.730957759493087*^9, 3.730958190431542*^9, 3.730958893305811*^9, 
  3.7309599794790688`*^9, 3.730960418437182*^9, 3.731003508441923*^9, 
  3.7310039568908443`*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->1882095871,ExpressionUUID->"85b390f1-4fec-464b-800c-4174300236b4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciPDKlosch", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Ricci tensor ", 
   $CellContext`RicciPDKlosch[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914420030834*^9, 3.730914963852853*^9, 
  3.730957343179661*^9, 3.730957578486018*^9, 3.730957696250476*^9, 
  3.730957759493087*^9, 3.730958190431542*^9, 3.730958893305811*^9, 
  3.7309599794790688`*^9, 3.730960418437182*^9, 3.731003508441923*^9, 
  3.731003956900363*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->2098667234,ExpressionUUID->"2de08f74-2426-4ff3-96f1-2a18c3864789"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaUpKlosch", "[", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric +1 density ", 
   $CellContext`etaUpKlosch[$CellContext`a, $CellContext`b, $CellContext`c, \
$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914420030834*^9, 3.730914963852853*^9, 
  3.730957343179661*^9, 3.730957578486018*^9, 3.730957696250476*^9, 
  3.730957759493087*^9, 3.730958190431542*^9, 3.730958893305811*^9, 
  3.7309599794790688`*^9, 3.730960418437182*^9, 3.731003508441923*^9, 
  3.731003956911173*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->2028783510,ExpressionUUID->"c87d2623-b57b-4ebc-8fad-1ad605f6ea1d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaDownKlosch", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric -1 density ", 
   $CellContext`etaDownKlosch[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, -$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914420030834*^9, 3.730914963852853*^9, 
  3.730957343179661*^9, 3.730957578486018*^9, 3.730957696250476*^9, 
  3.730957759493087*^9, 3.730958190431542*^9, 3.730958893305811*^9, 
  3.7309599794790688`*^9, 3.730960418437182*^9, 3.731003508441923*^9, 
  3.731003956921817*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->19105504,ExpressionUUID->"0b4e6b73-504c-4c6b-9908-15bee4c979c8"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Basis", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", "Klosch"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[25]:=",
 CellID->1352115308,ExpressionUUID->"e2918619-d982-4dcf-9112-144c85bf2445"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"e", 
      StyleBox[GridBox[{
         {" ", 
          StyleBox["1",
           FontColor->RGBColor[0., 0., 1.]]},
         {"a", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Klosch}],
  Editable->False]], "Output",
 CellChangeTimes->{3.730914422811267*^9, 3.7309149642347727`*^9, 
  3.730957343448621*^9, 3.730957578738003*^9, 3.7309576965201607`*^9, 
  3.730957759793912*^9, 3.7309581908355827`*^9, 3.730958893675405*^9, 
  3.730959979900881*^9, 3.7309604188724117`*^9, 3.731003509959075*^9, 
  3.731003957135182*^9},
 CellLabel->"Out[25]=",
 CellID->1131142325,ExpressionUUID->"a3063a87-a30a-4ee3-a7c7-fb961383adf9"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["This is the coordinate change:", "MathCaption",
 CellID->1994660060,ExpressionUUID->"1fde3037-47a5-4d67-a297-60c2c656c8c7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"change", "=", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{
     RowBox[{"\[Tau]", "[", "]"}], "\[Rule]", 
     RowBox[{
      RowBox[{
       RowBox[{"x", "[", "]"}], 
       RowBox[{"y", "[", "]"}]}], "+", 
      RowBox[{"2", "M", 
       RowBox[{"(", 
        RowBox[{"1", "+", 
         RowBox[{"Log", "[", 
          RowBox[{
           RowBox[{"y", "[", "]"}], "/", 
           RowBox[{"x", "[", "]"}]}], "]"}]}], ")"}]}]}]}], ",", 
    RowBox[{
     RowBox[{"r", "[", "]"}], "\[Rule]", 
     RowBox[{
      RowBox[{
       RowBox[{"x", "[", "]"}], 
       RowBox[{"y", "[", "]"}]}], "+", 
      RowBox[{"2", "M"}]}]}], ",", 
    RowBox[{
     RowBox[{"\[Theta]", "[", "]"}], "\[Rule]", 
     RowBox[{"\[CapitalTheta]", "[", "]"}]}], ",", 
    RowBox[{
     RowBox[{"\[Phi]", "[", "]"}], "\[Rule]", 
     RowBox[{"\[CapitalPhi]", "[", "]"}]}]}], "}"}]}]], "Input",
 CellLabel->"In[26]:=",
 CellID->2031898750,ExpressionUUID->"225f231b-987f-4c6e-89f4-bb4cc02f8489"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox["\[Tau]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Tau][],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{"1", "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["Log",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Log,
          Editable->False], "[", 
         FractionBox[
          InterpretationBox[
           StyleBox["y",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`y[],
           Editable->False], 
          InterpretationBox[
           StyleBox["x",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`x[],
           Editable->False]], "]"}]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["x",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["y",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`y[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["x",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["y",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`y[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["\[CapitalTheta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox["\[Phi]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Phi][],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["\[CapitalPhi]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalPhi][],
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.73091443529468*^9, 3.7309149643645144`*^9, 
  3.730957343549404*^9, 3.730957578908259*^9, 3.730957696669477*^9, 
  3.730957759910203*^9, 3.730958190987938*^9, 3.730958893860227*^9, 
  3.73095998003263*^9, 3.730960419004855*^9, 3.731003518126664*^9, 
  3.731003957271627*^9},
 CellLabel->"Out[26]=",
 CellID->1583813031,ExpressionUUID->"4faa870e-b21c-484b-a0f5-9dc07d0bc991"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Jacobian matrix:", "MathCaption",
 CellID->656838601,ExpressionUUID->"1238f26f-661e-49ee-b834-cd0eeec6af60"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Outer", "[", 
  RowBox[{"D", ",", 
   RowBox[{"Last", "/@", "change"}], ",", 
   RowBox[{"ScalarsOfChart", "[", "Klosch", "]"}]}], "]"}]], "Input",
 CellLabel->"In[27]:=",
 CellID->1420421427,ExpressionUUID->"672ca59a-e028-4289-b08c-7139f74a6967"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], 
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False]]}], "+", 
      InterpretationBox[
       StyleBox["y",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`y[],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox["x",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x[],
       Editable->False], "+", 
      FractionBox[
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False]}], 
       InterpretationBox[
        StyleBox["y",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`y[],
        Editable->False]]}], ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     InterpretationBox[
      StyleBox["y",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`y[],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox["x",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`x[],
      Editable->False], ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "1", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0", ",", "1"}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.730914450011993*^9, 3.730914964504043*^9, 
  3.730957343661394*^9, 3.73095757905483*^9, 3.7309576967686253`*^9, 
  3.7309577600592623`*^9, 3.730958191098102*^9, 3.730958893994252*^9, 
  3.730959980164978*^9, 3.730960419121464*^9, 3.7310035197282457`*^9, 
  3.731003957411509*^9},
 CellLabel->"Out[27]=",
 CellID->1085807452,ExpressionUUID->"0ce47ce7-1ab9-445e-8032-61c3310f99d3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AllComponentValues", "[", 
  RowBox[{
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"a", ",", "Sch"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "Klosch"}]}], "}"}]}], "]"}], ",", "%"}], "]"}]], "Input",\

 CellLabel->"In[28]:=",
 CellID->1586431799,ExpressionUUID->"b47e879f-2102-45a6-8afb-f77970761766"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 0., 1.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {0, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{"-", 
        FractionBox[
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`M,
           Editable->False]}], 
         InterpretationBox[
          StyleBox["x",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`x[],
          Editable->False]]}], "+", 
       InterpretationBox[
        StyleBox["y",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`y[],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {0, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      RowBox[{
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], "+", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], 
        InterpretationBox[
         StyleBox["y",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`y[],
         Editable->False]]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {0, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {0, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 0., 1.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {1, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox["y",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`y[],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {1, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox["x",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x[],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {1, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {1, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 0., 1.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {2, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {2, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {2, $CellContext`Sch}],
       Editable->False], "\[Rule]", "1"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {2, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["0",
                FontColor->RGBColor[0., 0., 1.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {3, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {3, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {3, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {3, $CellContext`Sch}],
       Editable->False], "\[Rule]", "1"}]}], "}"}]}], "]"}]], "Output",
 CellChangeTimes->{3.730914454253119*^9, 3.7309149646471853`*^9, 
  3.7309573437955627`*^9, 3.730957579182527*^9, 3.730957696920947*^9, 
  3.730957760210476*^9, 3.730958191245761*^9, 3.730958894174663*^9, 
  3.730959980331491*^9, 3.730960419286312*^9, 3.7310035358883343`*^9, 
  3.731003957553762*^9},
 CellLabel->"Out[28]=",
 CellID->1997297205,ExpressionUUID->"ba4bfa49-2742-4e78-af7c-d3afb3e8a553"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Let us see the metric form in the new coordinates:", "MathCaption",
 CellChangeTimes->{3.730914462190181*^9},
 CellID->1890739288,ExpressionUUID->"a4b1cbdb-2098-40ef-b9aa-fec23a694b7d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"downmetric", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.730914473491747*^9, 3.7309144789257717`*^9}},
 CellLabel->"In[29]:=",
 CellID->954075472,ExpressionUUID->"da16bf41-e076-4932-afd5-dcbffb337efc"],

Cell[BoxData[
 RowBox[{
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False]}], 
   RowBox[{"1", "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}]], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Sch}],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}]}]], "Output",
 CellChangeTimes->{3.730914479323927*^9, 3.730914509329275*^9, 
  3.7309149647793093`*^9, 3.730957343926091*^9, 3.730957579285203*^9, 
  3.730957697045706*^9, 3.730957760337954*^9, 3.730958191381641*^9, 
  3.730958894325066*^9, 3.7309599804650173`*^9, 3.730960419402527*^9, 
  3.7310035385780163`*^9, 3.731003957686862*^9},
 CellLabel->"Out[29]=",
 CellID->1257238178,ExpressionUUID->"b3c809b6-7887-4fc1-8657-0c74d1b23e4a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SeparateBasis", "[", "Klosch", "]"}], "[", 
  RowBox[{"%", ",", 
   RowBox[{"IndicesOf", "[", "Sch", "]"}]}], "]"}]], "Input",
 CellLabel->"In[30]:=",
 CellID->1877005023,ExpressionUUID->"e4f9c2ec-8b36-4c0a-b0cd-c07233147727"],

Cell[BoxData[
 RowBox[{
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["c",
              FontColor->RGBColor[0., 0., 1.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {$CellContext`c, \
$CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {$CellContext`d, \
$CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["c",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{-$CellContext`c, -$CellContext`Klosch}, {
      1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{-$CellContext`d, -$CellContext`Klosch}, {
      1, $CellContext`Sch}],
     Editable->False]}], 
   RowBox[{"1", "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}]], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["c",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {$CellContext`c, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["d",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {$CellContext`d, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["c",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`c, -$CellContext`Klosch}, {
     0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["d",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`d, -$CellContext`Klosch}, {
     0, $CellContext`Sch}],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["c",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {$CellContext`c, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["d",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {$CellContext`d, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["c",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`c, -$CellContext`Klosch}, {
     2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["d",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`d, -$CellContext`Klosch}, {
     2, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["c",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {$CellContext`c, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["d",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {$CellContext`d, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["c",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`c, -$CellContext`Klosch}, {
     3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["d",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`d, -$CellContext`Klosch}, {
     3, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}]}]], "Output",
 CellChangeTimes->{{3.730914486955118*^9, 3.7309145105109053`*^9}, 
   3.7309149648999662`*^9, 3.730957344051302*^9, 3.730957579419586*^9, 
   3.730957697167828*^9, 3.7309577604782763`*^9, 3.730958191515798*^9, 
   3.730958894495764*^9, 3.7309599806327047`*^9, 3.7309604195373783`*^9, 
   3.7310035406966543`*^9, 3.731003957826577*^9},
 CellLabel->"Out[30]=",
 CellID->1310454813,ExpressionUUID->"b8731b58-e324-4b91-99e5-b0c15fd2c902"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TraceBasisDummy", "[", "%", "]"}]], "Input",
 CellLabel->"In[31]:=",
 CellID->126772144,ExpressionUUID->"ad7666f4-0f8d-440b-b5bf-0dffeffaf462"],

Cell[BoxData[
 RowBox[{
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[0., 0., 1.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[0., 0., 1.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Klosch}],
     Editable->False], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["0",
               FontColor->RGBColor[0., 0., 1.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {1, $CellContext`Sch}],
      Editable->False], "2"]}], 
   RowBox[{"1", "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[0., 0., 1.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {1, $CellContext`Sch}],
     Editable->False]}], 
   RowBox[{"1", "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[0., 0., 1.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {1, $CellContext`Sch}],
     Editable->False]}], 
   RowBox[{"1", "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Klosch}],
     Editable->False], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {1, $CellContext`Sch}],
      Editable->False], "2"]}], 
   RowBox[{"1", "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[0., 0., 1.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {1, $CellContext`Sch}],
     Editable->False]}], 
   RowBox[{"1", "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[0., 0., 1.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {1, $CellContext`Sch}],
     Editable->False]}], 
   RowBox[{"1", "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {1, $CellContext`Sch}],
     Editable->False]}], 
   RowBox[{"1", "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {1, $CellContext`Sch}],
     Editable->False]}], 
   RowBox[{"1", "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Klosch}],
     Editable->False], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {1, $CellContext`Sch}],
      Editable->False], "2"]}], 
   RowBox[{"1", "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[0., 0., 1.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {1, $CellContext`Sch}],
     Editable->False]}], 
   RowBox[{"1", "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[0., 0., 1.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {1, $CellContext`Sch}],
     Editable->False]}], 
   RowBox[{"1", "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {1, $CellContext`Sch}],
     Editable->False]}], 
   RowBox[{"1", "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {1, $CellContext`Sch}],
     Editable->False]}], 
   RowBox[{"1", "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {1, $CellContext`Sch}],
     Editable->False]}], 
   RowBox[{"1", "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {1, $CellContext`Sch}],
     Editable->False]}], 
   RowBox[{"1", "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Klosch}],
     Editable->False], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {1, $CellContext`Sch}],
      Editable->False], "2"]}], 
   RowBox[{"1", "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}]], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Klosch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {0, $CellContext`Sch}],
     Editable->False], "2"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {0, $CellContext`Sch}],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {0, $CellContext`Sch}],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Klosch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {0, $CellContext`Sch}],
     Editable->False], "2"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {0, $CellContext`Sch}],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {0, $CellContext`Sch}],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {0, $CellContext`Sch}],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {0, $CellContext`Sch}],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Klosch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {0, $CellContext`Sch}],
     Editable->False], "2"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {0, $CellContext`Sch}],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {0, $CellContext`Sch}],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {0, $CellContext`Sch}],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {0, $CellContext`Sch}],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {0, $CellContext`Sch}],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {0, $CellContext`Sch}],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Klosch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {0, $CellContext`Sch}],
     Editable->False], "2"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Klosch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {2, $CellContext`Sch}],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {2, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {2, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Klosch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {2, $CellContext`Sch}],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {2, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {2, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {2, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {2, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Klosch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {2, $CellContext`Sch}],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {2, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {2, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {2, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {2, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {2, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {2, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Klosch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {2, $CellContext`Sch}],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Klosch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["0",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {3, $CellContext`Sch}],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {3, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {3, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Klosch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {3, $CellContext`Sch}],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {3, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {3, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {3, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {3, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Klosch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {3, $CellContext`Sch}],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {3, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{0, -$CellContext`Klosch}, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {3, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {3, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Klosch}, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {3, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {3, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Klosch}, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {3, $CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Klosch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Klosch}, {3, $CellContext`Sch}],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}]}]], "Output",
 CellChangeTimes->{3.730914518409881*^9, 3.73091496520474*^9, 
  3.730957344400772*^9, 3.7309575796924753`*^9, 3.73095769745757*^9, 
  3.730957760847259*^9, 3.730958191864318*^9, 3.7309588952031193`*^9, 
  3.730959980956252*^9, 3.730960419849494*^9, 3.7310035417894163`*^9, 
  3.7310039580977383`*^9},
 CellLabel->"Out[31]=",
 CellID->221758459,ExpressionUUID->"b8b66085-f948-4785-9ad3-21ad1642c448"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "/.", 
   RowBox[{"BasisValues", "[", 
    RowBox[{"Sch", ",", 
     RowBox[{"-", "Klosch"}]}], "]"}]}], "/.", "change"}]], "Input",
 CellLabel->"In[32]:=",
 CellID->901786287,ExpressionUUID->"80ab3be1-07a9-41db-a2be-a4f7986c2158"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Klosch}],
    Editable->False], " ", 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["y",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`y[],
        Editable->False]}]}], ")"}], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Klosch}],
    Editable->False], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[CapitalTheta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][],
      Editable->False], "]"}], "2"], " ", 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["y",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`y[],
        Editable->False]}]}], ")"}], "2"]}], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Klosch}],
     Editable->False], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["x",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`x[],
      Editable->False], "2"]}], 
   RowBox[{"1", "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["y",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`y[],
        Editable->False]}]}]]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[0., 0., 1.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox["x",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`x[],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox["y",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`y[],
     Editable->False]}], 
   RowBox[{"1", "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["y",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`y[],
        Editable->False]}]}]]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[0., 0., 1.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox["x",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`x[],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox["y",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`y[],
     Editable->False]}], 
   RowBox[{"1", "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["y",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`y[],
        Editable->False]}]}]]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[0., 0., 1.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[0., 0., 1.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Klosch}],
     Editable->False], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["y",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`y[],
      Editable->False], "2"]}], 
   RowBox[{"1", "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["y",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`y[],
        Editable->False]}]}]]}]], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Klosch}],
    Editable->False], " ", 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{
      InterpretationBox[
       StyleBox["x",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`x[],
       Editable->False], "+", 
      FractionBox[
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False]}], 
       InterpretationBox[
        StyleBox["y",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`y[],
        Editable->False]]}], ")"}], "2"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      RowBox[{
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["y",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`y[],
         Editable->False]}]}]]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Klosch}],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox["x",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`x[],
      Editable->False], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      InterpretationBox[
       StyleBox["y",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`y[],
       Editable->False]]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      FractionBox[
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False]}], 
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False]]}], "+", 
     InterpretationBox[
      StyleBox["y",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`y[],
      Editable->False]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      RowBox[{
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["y",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`y[],
         Editable->False]}]}]]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Klosch}],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox["x",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`x[],
      Editable->False], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      InterpretationBox[
       StyleBox["y",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`y[],
       Editable->False]]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      FractionBox[
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False]}], 
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False]]}], "+", 
     InterpretationBox[
      StyleBox["y",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`y[],
      Editable->False]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      RowBox[{
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["y",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`y[],
         Editable->False]}]}]]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Klosch}],
    Editable->False], " ", 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], 
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False]]}], "+", 
      InterpretationBox[
       StyleBox["y",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`y[],
       Editable->False]}], ")"}], "2"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      RowBox[{
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox["x",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`x[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["y",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`y[],
         Editable->False]}]}]]}], ")"}]}]}]], "Output",
 CellChangeTimes->{3.730914524975033*^9, 3.7309149653784122`*^9, 
  3.730957344556356*^9, 3.7309575798248167`*^9, 3.730957697608325*^9, 
  3.730957761047563*^9, 3.7309581920062857`*^9, 3.730958895471797*^9, 
  3.730959981119595*^9, 3.730960420026711*^9, 3.731003546043722*^9, 
  3.731003958280616*^9},
 CellLabel->"Out[32]=",
 CellID->1210292868,ExpressionUUID->"f2726499-6382-4b81-adf7-2d4d232d5519"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"BasisCollect", "[", 
  RowBox[{"%", ",", "Simplify"}], "]"}]], "Input",
 CellChangeTimes->{{3.730914537611861*^9, 3.730914547515664*^9}, {
  3.730958870540021*^9, 3.73095887556082*^9}},
 CellLabel->"In[33]:=",
 CellID->1223675932,ExpressionUUID->"39ad73e1-3394-49a9-98eb-e0220977daed"],

Cell[BoxData[
 RowBox[{
  RowBox[{"4", " ", 
   InterpretationBox[
    StyleBox["M",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`M,
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Klosch}],
    Editable->False]}], "+", 
  RowBox[{"4", " ", 
   InterpretationBox[
    StyleBox["M",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`M,
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Klosch}],
    Editable->False]}], "+", 
  FractionBox[
   RowBox[{"8", " ", 
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[0., 0., 1.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Klosch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[0., 0., 1.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Klosch}],
     Editable->False], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["y",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`y[],
      Editable->False], "2"]}], 
   RowBox[{
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`M,
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox["x",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`x[],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["y",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`y[],
      Editable->False]}]}]], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Klosch}],
    Editable->False], " ", 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["y",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`y[],
        Editable->False]}]}], ")"}], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Klosch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Klosch}],
    Editable->False], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[CapitalTheta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][],
      Editable->False], "]"}], "2"], " ", 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox["x",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`x[],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["y",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`y[],
        Editable->False]}]}], ")"}], "2"]}]}]], "Output",
 CellChangeTimes->{3.73091455149441*^9, 3.730914965510994*^9, 
  3.7309573446519814`*^9, 3.730957579935967*^9, 3.7309576977517548`*^9, 
  3.7309577611759253`*^9, 3.730958192122428*^9, 3.730958895671298*^9, 
  3.730959981238611*^9, 3.730960420170499*^9, 3.731003547637092*^9, 
  3.731003958399209*^9},
 CellLabel->"Out[33]=",
 CellID->427128092,ExpressionUUID->"68671053-cf43-46b2-882b-9aa853715e8d"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["4.12.4. Compute Christoffel", "Subsubsection",
 CellChangeTimes->{{3.7286771125390873`*^9, 3.7286771127222157`*^9}, {
  3.730913913338131*^9, 3.730913941103621*^9}, {3.730914255302225*^9, 
  3.730914260915913*^9}, {3.730914607546776*^9, 3.730914613449122*^9}},
 CellID->1221044385,ExpressionUUID->"85fdac02-de01-46ac-8725-54464977baec"],

Cell["\<\
Next we calculate the Christoffel symbols. We will first do it directly from \
the metric coefficients, and then with a dedicated function:\
\>", "Text",
 CellChangeTimes->{{3.730914618888854*^9, 3.730914636928822*^9}, {
  3.731003562757186*^9, 3.7310035798765993`*^9}},
 CellID->889724663,ExpressionUUID->"c59e6737-267f-4913-95d2-537d6c03c4c0"],

Cell[CellGroupData[{

Cell["\<\
 First we compute all Christoffel values given by its symmetry:\
\>", "MathCaption",
 CellChangeTimes->{
  3.730914639167156*^9, {3.7310035855403433`*^9, 3.731003596756053*^9}},
 CellID->1397892159,ExpressionUUID->"e28239af-5828-4a88-962d-292297081d49"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AllComponentValues", "[", 
  RowBox[{
   RowBox[{"Christoffel", "[", 
    RowBox[{"CD", ",", "PDSch"}], "]"}], "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"a", ",", "Sch"}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "Sch"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "Sch"}]}], "}"}]}], "]"}], "]"}]], "Input",
 CellLabel->"In[34]:=",
 CellID->673776980,ExpressionUUID->"7100fcce-873b-4bb2-aaea-82335e3cd7f3"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}]}], "}"}]}], "]"}]], "Output",
 CellChangeTimes->{3.730914648972837*^9, 3.730914968525236*^9, 
  3.730957580191414*^9, 3.730957698033513*^9, 3.730957761486492*^9, 
  3.73095819516892*^9, 3.730958910311788*^9, 3.73095998151103*^9, 
  3.7309604204613857`*^9, 3.731003603793675*^9, 3.7310039632019653`*^9},
 CellLabel->"Out[34]=",
 CellID->1233106143,ExpressionUUID->"2f834320-15aa-4806-b54b-46d1886cc3eb"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["This must be zero:", "MathCaption",
 CellID->1640641540,ExpressionUUID->"8710df2c-2b51-4ec1-8347-7af67596cb81"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CD", "[", 
   RowBox[{"-", "c"}], "]"}], "[", 
  RowBox[{"downmetric", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}]}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.730914664303537*^9, 3.7309146704952374`*^9}},
 CellLabel->"In[35]:=",
 CellID->572335300,ExpressionUUID->"d720de4c-73f6-4c97-bf4a-17a82bfde98f"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Sch}],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", 
             RowBox[{"\[EmptyDownTriangle]", ",", 
              StyleBox["\[ScriptCapitalD]",
               FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]], " ", " "},
               {" ", "c", "b"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCDPDSch[{
         0, $CellContext`Sch}, -$CellContext`c, -$CellContext`b],
        Editable->False]}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", 
        FractionBox[
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`M,
           Editable->False]}], 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]]}], ")"}]}], "-", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["0",
                 FontColor->RGBColor[1., 0., 0.]]},
               {"b", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Sch}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]},
               {"c", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
        Editable->False]}], 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"]]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Sch}],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      FractionBox[
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", 
             RowBox[{"\[EmptyDownTriangle]", ",", 
              StyleBox["\[ScriptCapitalD]",
               FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], " ", " "},
               {" ", "c", "b"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCDPDSch[{
         1, $CellContext`Sch}, -$CellContext`c, -$CellContext`b],
        Editable->False], 
       RowBox[{"1", "-", 
        FractionBox[
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`M,
           Editable->False]}], 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]]}]]}], "-", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]},
               {"b", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]},
               {"c", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
        Editable->False]}], 
      RowBox[{
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", 
          FractionBox[
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`M,
             Editable->False]}], 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]]}], ")"}], "2"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"]}]]}], ")"}]}], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", "c", "a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{
      1, $CellContext`Sch}, -$CellContext`c, -$CellContext`a],
     Editable->False]}], 
   RowBox[{"1", "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}]], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], " ", " "},
           {" ", "c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCDPDSch[{
     0, $CellContext`Sch}, -$CellContext`c, -$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]]}], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
        StyleBox[GridBox[{
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " ", " "},
           {" ", "c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCDPDSch[{
     2, $CellContext`Sch}, -$CellContext`c, -$CellContext`a],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Sch}],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {"b", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {"c", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", "c", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{
        2, $CellContext`Sch}, -$CellContext`c, -$CellContext`b],
       Editable->False], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"]}]}], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
        StyleBox[GridBox[{
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " ", " "},
           {" ", "c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCDPDSch[{
     3, $CellContext`Sch}, -$CellContext`c, -$CellContext`a],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Sch}],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", 
             RowBox[{"\[EmptyDownTriangle]", ",", 
              StyleBox["\[ScriptCapitalD]",
               FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["3",
                 FontColor->RGBColor[1., 0., 0.]], " ", " "},
               {" ", "c", "b"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCDPDSch[{
         3, $CellContext`Sch}, -$CellContext`c, -$CellContext`b],
        Editable->False]}], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"], " ", 
      SuperscriptBox[
       RowBox[{
        InterpretationBox[
         StyleBox["Sin",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Sin,
         Editable->False], "[", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False], "]"}], "2"]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"e", 
           StyleBox[GridBox[{
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {"b", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"e", 
              StyleBox[GridBox[{
                 {" ", 
                  StyleBox["2",
                   FontColor->RGBColor[1., 0., 0.]]},
                 {"c", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
          Editable->False], " ", 
         RowBox[{
          InterpretationBox[
           StyleBox["Cos",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Cos,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], " ", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"], " ", 
         RowBox[{
          InterpretationBox[
           StyleBox["Sin",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Sin,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}]}], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"e", 
              StyleBox[GridBox[{
                 {" ", 
                  StyleBox["1",
                   FontColor->RGBColor[1., 0., 0.]]},
                 {"c", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Sin",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Sin,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}]}], ")"}]}]}], ")"}]}]}]], "Output",\

 CellChangeTimes->{3.730914679565292*^9, 3.73091497007798*^9, 
  3.730957580300445*^9, 3.7309576981750526`*^9, 3.7309577616362123`*^9, 
  3.730958195310823*^9, 3.7309589106092653`*^9, 3.7309599816509113`*^9, 
  3.7309604206438303`*^9, 3.7310036155352*^9, 3.7310039633178053`*^9},
 CellLabel->"Out[35]=",
 CellID->1597751851,ExpressionUUID->"d5456605-20cc-4739-9da7-b8ba28869df8"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", 
  RowBox[{"ToBasis", "[", "Sch", "]"}]}]], "Input",
 CellLabel->"In[36]:=",
 CellID->941146507,ExpressionUUID->"3060b31a-89df-46ba-a0f1-b8f7aa56f946"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["b",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`b, -$CellContext`Sch}, {
     0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], " ", " "},
           {" ", 
            StyleBox["c",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["a",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCDPDSch[{
     0, $CellContext`Sch}, {-$CellContext`c, -$CellContext`Sch}, \
{-$CellContext`a, -$CellContext`Sch}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["a",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`a, -$CellContext`Sch}, {
     0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
        StyleBox[GridBox[{
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], " ", " "},
           {" ", 
            StyleBox["c",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["b",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCDPDSch[{
     0, $CellContext`Sch}, {-$CellContext`c, -$CellContext`Sch}, \
{-$CellContext`b, -$CellContext`Sch}],
    Editable->False]}], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["b",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{-$CellContext`b, -$CellContext`Sch}, {
      1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["c",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["a",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{
      1, $CellContext`Sch}, {-$CellContext`c, -$CellContext`Sch}, \
{-$CellContext`a, -$CellContext`Sch}],
     Editable->False]}], 
   RowBox[{"1", "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["a",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{-$CellContext`a, -$CellContext`Sch}, {
      1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["c",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["b",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{
      1, $CellContext`Sch}, {-$CellContext`c, -$CellContext`Sch}, \
{-$CellContext`b, -$CellContext`Sch}],
     Editable->False]}], 
   RowBox[{"1", "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}]], "-", 
  FractionBox[
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["a",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{-$CellContext`a, -$CellContext`Sch}, {
      0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["b",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{-$CellContext`b, -$CellContext`Sch}, {
      0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["c",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{-$CellContext`c, -$CellContext`Sch}, {
      1, $CellContext`Sch}],
     Editable->False]}], 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]], "-", 
  FractionBox[
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["a",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{-$CellContext`a, -$CellContext`Sch}, {
      1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["b",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{-$CellContext`b, -$CellContext`Sch}, {
      1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["c",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{-$CellContext`c, -$CellContext`Sch}, {
      1, $CellContext`Sch}],
     Editable->False]}], 
   RowBox[{
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{"1", "-", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]]}], ")"}], "2"], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "-", 
  FractionBox[
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["b",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{-$CellContext`b, -$CellContext`Sch}, {
      0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["c",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["a",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{
      0, $CellContext`Sch}, {-$CellContext`c, -$CellContext`Sch}, \
{-$CellContext`a, -$CellContext`Sch}],
     Editable->False]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "-", 
  FractionBox[
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["a",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{-$CellContext`a, -$CellContext`Sch}, {
      0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["c",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["b",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{
      0, $CellContext`Sch}, {-$CellContext`c, -$CellContext`Sch}, \
{-$CellContext`b, -$CellContext`Sch}],
     Editable->False]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["a",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`a, -$CellContext`Sch}, {
     2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["b",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`b, -$CellContext`Sch}, {
     2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["c",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`c, -$CellContext`Sch}, {
     1, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["b",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`b, -$CellContext`Sch}, {
     2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
        StyleBox[GridBox[{
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " ", " "},
           {" ", 
            StyleBox["c",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["a",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCDPDSch[{
     2, $CellContext`Sch}, {-$CellContext`c, -$CellContext`Sch}, \
{-$CellContext`a, -$CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["a",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`a, -$CellContext`Sch}, {
     2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
        StyleBox[GridBox[{
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " ", " "},
           {" ", 
            StyleBox["c",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["b",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCDPDSch[{
     2, $CellContext`Sch}, {-$CellContext`c, -$CellContext`Sch}, \
{-$CellContext`b, -$CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["a",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`a, -$CellContext`Sch}, {
     3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["b",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`b, -$CellContext`Sch}, {
     3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["c",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`c, -$CellContext`Sch}, {
     2, $CellContext`Sch}],
    Editable->False], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Cos",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cos,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Sin",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Sin,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["a",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`a, -$CellContext`Sch}, {
     3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["b",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`b, -$CellContext`Sch}, {
     3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["c",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`c, -$CellContext`Sch}, {
     1, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["b",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`b, -$CellContext`Sch}, {
     3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
        StyleBox[GridBox[{
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " ", " "},
           {" ", 
            StyleBox["c",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["a",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCDPDSch[{
     3, $CellContext`Sch}, {-$CellContext`c, -$CellContext`Sch}, \
{-$CellContext`a, -$CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {
            StyleBox["a",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{-$CellContext`a, -$CellContext`Sch}, {
     3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
        StyleBox[GridBox[{
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " ", " "},
           {" ", 
            StyleBox["c",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["b",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCDPDSch[{
     3, $CellContext`Sch}, {-$CellContext`c, -$CellContext`Sch}, \
{-$CellContext`b, -$CellContext`Sch}],
    Editable->False], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}]}]], "Output",
 CellChangeTimes->{3.7309147000002127`*^9, 3.730914971153158*^9, 
  3.730957580511217*^9, 3.730957698353936*^9, 3.7309577618191223`*^9, 
  3.730958195539926*^9, 3.7309589108803587`*^9, 3.7309599818271427`*^9, 
  3.730960420918662*^9, 3.731003626360078*^9, 3.731003963455873*^9},
 CellLabel->"Out[36]=",
 CellID->1279837669,ExpressionUUID->"1aba94a4-c9e3-42fa-9b68-8b45812f89b8"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Flatten", "[", 
   RowBox[{"%", "//", "ComponentArray"}], "]"}], "/.", 
  RowBox[{"TensorValues", "[", "ChristoffelCDPDSch", "]"}]}]], "Input",
 CellLabel->"In[37]:=",
 CellID->1824173380,ExpressionUUID->"bc4e2281-5b32-4a8e-bac2-865167eb8815"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
      Editable->False]}], "-", 
    FractionBox[
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}], ",", 
   RowBox[{
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
      Editable->False]}], "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]], "-", 
    FractionBox[
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}], ",", 
   RowBox[{
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
      Editable->False]}], "-", 
    FractionBox[
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}], ",", 
   RowBox[{
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
      Editable->False]}], "-", 
    FractionBox[
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
     Editable->False], "-", 
    FractionBox[
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
      Editable->False], 
     RowBox[{"1", "-", 
      FractionBox[
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False]}], 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]]}]], "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
     Editable->False], "-", 
    FractionBox[
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
      Editable->False], 
     RowBox[{"1", "-", 
      FractionBox[
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False]}], 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]]}]], "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "-", 
    FractionBox[
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
      Editable->False], 
     RowBox[{"1", "-", 
      FractionBox[
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False]}], 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]]}]], "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "-", 
    FractionBox[
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
      Editable->False], 
     RowBox[{"1", "-", 
      FractionBox[
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False]}], 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]]}]], "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
     Editable->False], "-", 
    FractionBox[
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
      Editable->False], 
     RowBox[{"1", "-", 
      FractionBox[
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False]}], 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]]}]], "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
     Editable->False], "-", 
    FractionBox[
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
      Editable->False], 
     RowBox[{"1", "-", 
      FractionBox[
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False]}], 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]]}]], "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "-", 
    FractionBox[
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
      Editable->False], 
     RowBox[{"1", "-", 
      FractionBox[
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False]}], 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]]}]], "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "-", 
    FractionBox[
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
      Editable->False], 
     RowBox[{"1", "-", 
      FractionBox[
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False]}], 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]]}]], "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}], ",", 
   RowBox[{"-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], 
     RowBox[{"1", "-", 
      FractionBox[
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False]}], 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]]}]]}], ",", 
   RowBox[{
    RowBox[{"-", 
     FractionBox[
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", 
             RowBox[{"\[EmptyDownTriangle]", ",", 
              StyleBox["\[ScriptCapitalD]",
               FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], " ", " "},
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]], 
                StyleBox["1",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
         1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
        Editable->False]}], 
      RowBox[{"1", "-", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]]}]]}], "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], 
     RowBox[{
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "-", 
         FractionBox[
          RowBox[{"2", " ", 
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`M,
            Editable->False]}], 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False]]}], ")"}], "2"], " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"]}]]}], ",", 
   RowBox[{"-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], 
     RowBox[{"1", "-", 
      FractionBox[
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False]}], 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]]}]]}], ",", 
   RowBox[{"-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], 
     RowBox[{"1", "-", 
      FractionBox[
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False]}], 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]]}]]}], ",", 
   RowBox[{
    RowBox[{"-", 
     FractionBox[
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], 
      RowBox[{"1", "-", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]]}]]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     FractionBox[
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], 
      RowBox[{"1", "-", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]]}]]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
       1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     FractionBox[
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], 
      RowBox[{"1", "-", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]]}]]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
       1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     FractionBox[
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], 
      RowBox[{"1", "-", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]]}]]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
       1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     FractionBox[
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], 
      RowBox[{"1", "-", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]]}]]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     FractionBox[
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], 
      RowBox[{"1", "-", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]]}]]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
       1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     FractionBox[
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], 
      RowBox[{"1", "-", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]]}]]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
       1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     FractionBox[
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], 
      RowBox[{"1", "-", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]]}]]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
       1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     FractionBox[
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], 
      RowBox[{"1", "-", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]]}]]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     FractionBox[
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], 
      RowBox[{"1", "-", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]]}]]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
       1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     FractionBox[
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], 
      RowBox[{"1", "-", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]]}]]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
       1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     FractionBox[
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], 
      RowBox[{"1", "-", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]]}]]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
       1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]}]}], ",", 
   RowBox[{
    RowBox[{"-", "2"}], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}], ",", 
   RowBox[{
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}], "-", 
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
       1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]}]}], ",", 
   RowBox[{
    RowBox[{"-", "2"}], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
      2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}], ",", 
   RowBox[{
    RowBox[{"-", "2"}], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
      2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
       1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
       2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
       2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "-", 
    FractionBox[
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     FractionBox[
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], 
      RowBox[{"1", "-", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]]}]]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     FractionBox[
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], 
      RowBox[{"1", "-", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]]}]]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
       1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     FractionBox[
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], 
      RowBox[{"1", "-", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]]}]]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
       1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     FractionBox[
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], 
      RowBox[{"1", "-", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]]}]]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
       1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
       0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
       1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
       2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
       2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], ",", 
   RowBox[{
    RowBox[{"-", "2"}], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"], " ", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"]}], ",", 
   RowBox[{
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}], "-", 
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
       1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], ",", 
   RowBox[{
    RowBox[{"2", " ", 
     RowBox[{
      InterpretationBox[
       StyleBox["Cos",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Cos,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}]}], "-", 
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["3",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
       2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], ",", 
   RowBox[{
    RowBox[{"-", "2"}], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
      3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"], " ", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"]}]}], "}"}]], "Output",
 CellChangeTimes->{3.730914717977869*^9, 3.7309149725384703`*^9, 
  3.730957580766262*^9, 3.7309576986358547`*^9, 3.7309577621137047`*^9, 
  3.730958195810238*^9, 3.730958911621358*^9, 3.7309599821145573`*^9, 
  3.730960421249844*^9, 3.7310036276614637`*^9, 3.7310039636953278`*^9},
 CellLabel->"Out[37]=",
 CellID->1035494531,ExpressionUUID->"5011a6b7-b941-4ab5-84f5-ff890566cc8c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We solve the system of equations for the independent components of the \
Christoffel tensor:\
\>", "MathCaption",
 CellID->1277416663,ExpressionUUID->"d53b529b-2e83-4d63-bc8b-86b7f09c5d53"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Simplify", "/@", 
  RowBox[{"First", "@", 
   RowBox[{"Solve", "[", 
    RowBox[{
     RowBox[{"Map", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"#", "\[Equal]", "0"}], "&"}], ",", "%"}], "]"}], ",", 
     RowBox[{"First", "/@", 
      RowBox[{"IndependentRules", "[", 
       RowBox[{"TensorValues", "[", "ChristoffelCDPDSch", "]"}], "]"}]}]}], 
    "]"}]}]}]], "Input",
 CellChangeTimes->{{3.73091493801723*^9, 3.730914938470613*^9}},
 CellLabel->"In[38]:=",
 CellID->234373083,ExpressionUUID->"22216ff6-f151-4e6f-a204-7ddb898c4f78"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     FractionBox[
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], 
      RowBox[{
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]}], "-", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"]}]]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
     Editable->False], "\[Rule]", 
    FractionBox[
     RowBox[{
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]}], ")"}]}], 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "3"]]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
     Editable->False], "\[Rule]", 
    FractionBox[
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`M,
      Editable->False], 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], "-", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"]}]]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
      2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
      2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
      3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False]}], "-", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], ")"}], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "\[Rule]", 
    FractionBox["1", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
      2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
      2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
      3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      RowBox[{
       InterpretationBox[
        StyleBox["Cos",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Cos,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}]}], " ", 
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", 
    FractionBox["1", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
      2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
      2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox["Cot",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Cot,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
      3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}]}], "}"}]], "Output",
 CellChangeTimes->{3.7309149174158773`*^9, 3.730914974274811*^9, 
  3.730957580935176*^9, 3.730957698873293*^9, 3.730957762327335*^9, 
  3.730958195999484*^9, 3.7309589119207487`*^9, 3.730959982304628*^9, 
  3.7309604214613743`*^9, 3.731003640210477*^9, 3.731003963836548*^9},
 CellLabel->"Out[38]=",
 CellID->349107229,ExpressionUUID->"199c7454-b167-4d4c-bbed-b10f46dc51e5"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["and store them:", "MathCaption",
 CellID->479789036,ExpressionUUID->"cde347d0-0b25-4f9a-ae99-cd22956f859b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", 
  RowBox[{"Rule", "\[Rule]", "ComponentValue"}]}]], "Input",
 CellLabel->"In[39]:=",
 CellID->13233567,ExpressionUUID->"4a675e87-1f4d-4212-9e4a-5da5cfee009d"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     FractionBox[
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], 
      RowBox[{
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]}], "-", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"]}]]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
      3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
     Editable->False], "\[Rule]", 
    FractionBox[
     RowBox[{
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]}], ")"}]}], 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "3"]]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
     Editable->False], "\[Rule]", 
    FractionBox[
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`M,
      Editable->False], 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], "-", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"]}]]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
      2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
      2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
      3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False]}], "-", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], ")"}], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "\[Rule]", 
    FractionBox["1", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
      2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
      2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
      3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      RowBox[{
       InterpretationBox[
        StyleBox["Cos",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Cos,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}]}], " ", 
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
      0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
      1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", 
    FractionBox["1", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
      2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
      2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox["Cot",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Cot,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
      3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
     Editable->False], "\[Rule]", "0"}]}], "}"}]], "Output",
 CellChangeTimes->{3.730914994641488*^9, 3.730957581097755*^9, 
  3.730957699038553*^9, 3.730957762557869*^9, 3.730958196170323*^9, 
  3.73095891213649*^9, 3.73095998246417*^9, 3.730960421661899*^9, 
  3.7310036438780727`*^9, 3.7310039640151176`*^9},
 CellLabel->"Out[39]=",
 CellID->519874199,ExpressionUUID->"506a1d70-25dc-465c-848c-af82b80ea85b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["This is the Christoffel tensor:", "MathCaption",
 CellID->284379225,ExpressionUUID->"21661525-6ef4-44e5-8bd6-244d93130807"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"christoffel", "=", 
  RowBox[{
   RowBox[{"BasisExpand", "[", 
    RowBox[{
     RowBox[{"ChristoffelCDPDSch", "[", 
      RowBox[{"a", ",", 
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "c"}]}], "]"}], ",", "Sch"}], "]"}], "//", 
   "ToValues"}]}]], "Input",
 CellChangeTimes->{{3.730957839957473*^9, 3.730957911943934*^9}, {
  3.730957996431459*^9, 3.730958002006164*^9}},
 CellLabel->"In[40]:=",
 CellID->239099082,ExpressionUUID->"be365d0c-4598-4922-a401-d05a57077863"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`a],
    Editable->False], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Cot",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cot,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`a],
    Editable->False], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Cot",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cot,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}], ")"}]}], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`a],
     Editable->False]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`a],
     Editable->False]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`a],
     Editable->False]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`a],
     Editable->False]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`a],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "+", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}]}], 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "3"]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`a],
     Editable->False]}], 
   RowBox[{
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`M,
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}], "-", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`a],
     Editable->False]}], 
   RowBox[{
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`M,
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}], "-", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`a],
     Editable->False]}], 
   RowBox[{
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`M,
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}], "-", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`a],
    Editable->False], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Cos",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cos,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Sin",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Sin,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}], ")"}], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}]}]], "Output",
 CellChangeTimes->{{3.7309578869303303`*^9, 3.730957912798296*^9}, 
   3.730958002749752*^9, 3.730958197810392*^9, 3.73095891468334*^9, 
   3.730958961111272*^9, 3.73095998261794*^9, 3.730960421855035*^9, 
   3.7310036484400806`*^9, 3.731003964173135*^9},
 CellLabel->"Out[40]=",
 CellID->1562186597,ExpressionUUID->"9f737625-1131-4879-9f2d-42ab66d754ab"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We may calculate the Christoffel symbols of different kind. Note that by \
symmetry we compute simultaneously {up,down,up} and {up,up,down}:\
\>", "MathCaption",
 CellID->2000808040,ExpressionUUID->"68b8a124-ad6a-47bc-b642-1fb1535bb3bd"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"$BMUseValues", "=", "All"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"$TUseValues", "=", "All"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"$CCSimplify", "=", "Simplify"}], ";"}]}], "Input",
 CellLabel->"In[41]:=",
 CellID->134299907,ExpressionUUID->"c5daf762-08eb-451c-bafd-2c2b1946b0ce"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ChangeComponents", "[", 
  RowBox[{
   RowBox[{"ChristoffelCDPDSch", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"a", ",", "Sch"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "Sch"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"c", ",", "Sch"}], "}"}]}], "]"}], ",", 
   RowBox[{"ChristoffelCDPDSch", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"a", ",", "Sch"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "Sch"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "c"}], ",", 
       RowBox[{"-", "Sch"}]}], "}"}]}], "]"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.7309150561628437`*^9, 3.7309150821058283`*^9}},
 CellLabel->"In[44]:=",
 CellID->159286465,ExpressionUUID->"991f9245-05e8-497d-a02e-00ce230e776b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Computed \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["a",
              FontColor->RGBColor[1., 0., 0.]], " ", 
             StyleBox["c",
              FontColor->RGBColor[1., 0., 0.]]},
            {" ", 
             StyleBox["b",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{$CellContext`a, $CellContext`Sch}, \
{-$CellContext`b, -$CellContext`Sch}, {$CellContext`c, $CellContext`Sch}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["a",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["b",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["d",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{$CellContext`a, $CellContext`Sch}, \
{-$CellContext`b, -$CellContext`Sch}, {-$CellContext`d, -$CellContext`Sch}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {
              StyleBox["c",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["d",
               FontColor->RGBColor[1., 0., 0.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`metricg[{$CellContext`c, $CellContext`Sch}, \
{$CellContext`d, $CellContext`Sch}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" in \"\>", 
   "\[InvisibleSpace]", "0.175736`5.696405730516842", 
   "\[InvisibleSpace]", "\<\" Seconds\"\>"}],
  SequenceForm[
  "Computed ", $CellContext`ChristoffelCDPDSch[{$CellContext`a, \
$CellContext`Sch}, {-$CellContext`b, -$CellContext`Sch}, {$CellContext`c, \
$CellContext`Sch}] -> $CellContext`ChristoffelCDPDSch[{$CellContext`a, \
$CellContext`Sch}, {-$CellContext`b, -$CellContext`Sch}, {-$CellContext`d, \
-$CellContext`Sch}] $CellContext`metricg[{$CellContext`c, $CellContext`Sch}, \
{$CellContext`d, $CellContext`Sch}], " in ", 0.175736`5.696405730516842, 
   " Seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{{3.730915087105476*^9, 3.7309151052710648`*^9}, 
   3.730957581557006*^9, 3.730957699496936*^9, 3.7309577631858797`*^9, 
   3.730958196674505*^9, 3.730958912703986*^9, 3.730959982986858*^9, 
   3.730960422260198*^9, 3.731003656873822*^9, 3.731003964610814*^9},
 CellLabel->"During evaluation of In[44]:=",
 CellID->535067557,ExpressionUUID->"36dc944b-b68a-466d-9864-2996d9a28992"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {0, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, $CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {0, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, $CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {1, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, $CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {0, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, $CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {1, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, $CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        2, $CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {0, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {1, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        2, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        3, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {0, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, $CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {0, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, $CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {1, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        1, $CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {0, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, $CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {1, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        1, $CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        2, $CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {0, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {1, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        1, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        2, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        3, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {0, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        0, $CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {0, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        0, $CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {1, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        1, $CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {0, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        0, $CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {1, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        1, $CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        2, $CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {0, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        0, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {1, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        1, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        2, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        3, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {0, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        0, $CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {0, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        0, $CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {1, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        1, $CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {0, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        0, $CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {1, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        1, $CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        2, $CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {0, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        0, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {1, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        1, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        2, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        3, -$CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        3, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, $CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, $CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        2, $CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        2, $CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        3, $CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        3, $CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        3, $CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        1, $CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, $CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        2, $CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        2, $CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        3, $CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        3, $CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        3, $CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        1, $CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, $CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        2, $CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        2, $CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        3, $CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        3, $CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        3, $CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        1, $CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, $CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        2, $CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        2, $CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        3, $CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        3, $CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        3, $CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, $CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      FractionBox[
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False], 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"]]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, $CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       FractionBox[
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False], 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", "2"}], " ", 
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`M,
             Editable->False]}], "+", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]}], ")"}], "2"]]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, $CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, $CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, $CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        2, $CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        0, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        1, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        2, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, $CellContext`Sch}, {
        3, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, $CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       FractionBox[
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False], 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, $CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        1, $CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       FractionBox[
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False], 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, $CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        1, $CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        2, $CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      FractionBox[
       RowBox[{
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], "-", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]}], 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"]]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        0, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        1, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        2, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, $CellContext`Sch}, {
        3, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      FractionBox[
       RowBox[{
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], "-", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]}], 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"]]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        0, $CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        0, $CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        1, $CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      FractionBox["1", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "3"]]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        0, $CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        1, $CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      FractionBox[
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]}], 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"]]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        2, $CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        0, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        1, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        2, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, $CellContext`Sch}, {
        3, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{
         InterpretationBox[
          StyleBox["Cot",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Cot,
          Editable->False], "[", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False], "]"}], 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        0, $CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        0, $CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        1, $CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, $CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      FractionBox[
       SuperscriptBox[
        RowBox[{
         InterpretationBox[
          StyleBox["Csc",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Csc,
          Editable->False], "[", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False], "]"}], "2"], 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "3"]]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        0, $CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        1, $CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        2, $CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, $CellContext`Sch}],
       Editable->False], "\[Rule]", 
      FractionBox[
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["Cot",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Cot,
          Editable->False], "[", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False], "]"}], " ", 
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Csc",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Csc,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"]}], 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"]]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        0, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        1, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      FractionBox[
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]}], 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"]]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        2, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      FractionBox[
       RowBox[{
        InterpretationBox[
         StyleBox["Cot",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Cot,
         Editable->False], "[", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False], "]"}], 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"]]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], " "},
              {" ", " ", 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, $CellContext`Sch}, {
        3, $CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}]}], "}"}]}], "]"}]], "Output",
 CellChangeTimes->{{3.730915087744018*^9, 3.730915105593409*^9}, 
   3.730957581802367*^9, 3.730957699660759*^9, 3.7309577633942547`*^9, 
   3.730958196910221*^9, 3.730958912897944*^9, 3.7309599831587467`*^9, 
   3.730960422428759*^9, 3.731003657090084*^9, 3.731003964819895*^9},
 CellLabel->"Out[44]=",
 CellID->1366400510,ExpressionUUID->"90fcfae9-f337-43f3-a2a9-3d67ca4fdd8b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ChangeComponents", "[", 
  RowBox[{
   RowBox[{"ChristoffelCDPDSch", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "Sch"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "Sch"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "c"}], ",", 
       RowBox[{"-", "Sch"}]}], "}"}]}], "]"}], ",", 
   RowBox[{"ChristoffelCDPDSch", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"a", ",", "Sch"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "Sch"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "c"}], ",", 
       RowBox[{"-", "Sch"}]}], "}"}]}], "]"}]}], "]"}]], "Input",
 CellLabel->"In[45]:=",
 CellID->949306026,ExpressionUUID->"5527432a-bd88-4fd0-aadb-d7fb29e916b0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Computed \"\>", "\[InvisibleSpace]", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " "},
            {
             StyleBox["a",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["b",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["c",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDSch[{-$CellContext`a, -$CellContext`Sch}, \
{-$CellContext`b, -$CellContext`Sch}, {-$CellContext`c, -$CellContext`Sch}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["d",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["b",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["c",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDSch[{$CellContext`d, $CellContext`Sch}, \
{-$CellContext`b, -$CellContext`Sch}, {-$CellContext`c, -$CellContext`Sch}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["d",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["a",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`metricg[{-$CellContext`d, -$CellContext`Sch}, \
{-$CellContext`a, -$CellContext`Sch}],
      Editable->False]}]}], "\[InvisibleSpace]", "\<\" in \"\>", 
   "\[InvisibleSpace]", "0.174315`5.692879753738121", 
   "\[InvisibleSpace]", "\<\" Seconds\"\>"}],
  SequenceForm[
  "Computed ", $CellContext`ChristoffelCDPDSch[{-$CellContext`a, \
-$CellContext`Sch}, {-$CellContext`b, -$CellContext`Sch}, {-$CellContext`c, \
-$CellContext`Sch}] -> $CellContext`ChristoffelCDPDSch[{$CellContext`d, \
$CellContext`Sch}, {-$CellContext`b, -$CellContext`Sch}, {-$CellContext`c, \
-$CellContext`Sch}] $CellContext`metricg[{-$CellContext`d, \
-$CellContext`Sch}, {-$CellContext`a, -$CellContext`Sch}], " in ", 
   0.174315`5.692879753738121, " Seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.73091515132101*^9, 3.730957582258922*^9, 
  3.730957700003861*^9, 3.7309577637600527`*^9, 3.730958197206205*^9, 
  3.73095891332686*^9, 3.7309599835240583`*^9, 3.730960422773184*^9, 
  3.731003661514682*^9, 3.731003965076845*^9},
 CellLabel->"During evaluation of In[45]:=",
 CellID->2050734199,ExpressionUUID->"eec716a5-d95e-4f9e-929d-cb61e14f7ef7"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, -$CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, -$CellContext`Sch}, {
        2, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, -$CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, -$CellContext`Sch}, {
        2, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, -$CellContext`Sch}, {
        3, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, -$CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, -$CellContext`Sch}, {
        3, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, -$CellContext`Sch}, {
        3, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, -$CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, -$CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, -$CellContext`Sch}, {
        2, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, -$CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, -$CellContext`Sch}, {
        2, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, -$CellContext`Sch}, {
        3, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, -$CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, -$CellContext`Sch}, {
        3, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, -$CellContext`Sch}, {
        3, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, -$CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, -$CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, -$CellContext`Sch}, {
        2, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, -$CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, -$CellContext`Sch}, {
        2, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, -$CellContext`Sch}, {
        3, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, -$CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, -$CellContext`Sch}, {
        3, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, -$CellContext`Sch}, {
        3, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, -$CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, -$CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, -$CellContext`Sch}, {
        2, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, -$CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, -$CellContext`Sch}, {
        2, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, -$CellContext`Sch}, {
        3, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, -$CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, -$CellContext`Sch}, {
        3, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, -$CellContext`Sch}, {
        3, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, -$CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, -$CellContext`Sch}, {
        0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, -$CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       FractionBox[
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False], 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, -$CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, -$CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, -$CellContext`Sch}, {
        2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, -$CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{0, -$CellContext`Sch}, {
        3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, -$CellContext`Sch}, {
        0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      FractionBox[
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False], 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"]]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, -$CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, -$CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, -$CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       FractionBox[
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False], 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", "2"}], " ", 
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`M,
             Editable->False]}], "+", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]}], ")"}], "2"]]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, -$CellContext`Sch}, {
        2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, -$CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{1, -$CellContext`Sch}, {
        3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]}], " ", 
       SuperscriptBox[
        RowBox[{
         InterpretationBox[
          StyleBox["Sin",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Sin,
          Editable->False], "[", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False], "]"}], "2"]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, -$CellContext`Sch}, {
        0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, -$CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, -$CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, -$CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, -$CellContext`Sch}, {
        2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, -$CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{2, -$CellContext`Sch}, {
        3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{"-", 
        RowBox[{
         InterpretationBox[
          StyleBox["Cos",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Cos,
          Editable->False], "[", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False], "]"}]}], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"], " ", 
       RowBox[{
        InterpretationBox[
         StyleBox["Sin",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Sin,
         Editable->False], "[", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False], "]"}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, -$CellContext`Sch}, {
        0, -$CellContext`Sch}, {0, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, -$CellContext`Sch}, {
        0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, -$CellContext`Sch}, {
        0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, -$CellContext`Sch}, {
        0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      RowBox[{
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], " ", 
       SuperscriptBox[
        RowBox[{
         InterpretationBox[
          StyleBox["Sin",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Sin,
          Editable->False], "[", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False], "]"}], "2"]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, -$CellContext`Sch}, {
        2, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, -$CellContext`Sch}, {
        2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox["Cos",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Cos,
         Editable->False], "[", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False], "]"}], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"], " ", 
       RowBox[{
        InterpretationBox[
         StyleBox["Sin",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Sin,
         Editable->False], "[", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False], "]"}]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDSch[{3, -$CellContext`Sch}, {
        3, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}]}], "}"}]}], "]"}]], "Output",
 CellChangeTimes->{3.730915151489182*^9, 3.730957582374569*^9, 
  3.7309577001191797`*^9, 3.7309577638728533`*^9, 3.730958197347495*^9, 
  3.730958913730983*^9, 3.73095998365405*^9, 3.7309604228964853`*^9, 
  3.731003661625278*^9, 3.7310039651838837`*^9},
 CellLabel->"Out[45]=",
 CellID->1724631890,ExpressionUUID->"95405406-43aa-483e-b74d-0a503bfd82e0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
These are the values we have computed, taking into account the symmetries:\
\>", "MathCaption",
 CellID->1250815119,ExpressionUUID->"a43ab713-ba1f-407b-ac5c-2bd499b5f197"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"TensorValIDs", "[", "ChristoffelCDPDSch", "]"}], "//", 
  "ColumnForm"}]], "Input",
 CellLabel->"In[46]:=",
 CellID->732284550,ExpressionUUID->"10414202-7102-4db4-87ab-b655dc251b22"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {
     RowBox[{"ValID", "[", 
      RowBox[{"ChristoffelCDPDSch", ",", 
       RowBox[{"{", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{"-", "Sch"}], ",", 
          RowBox[{"-", "Sch"}], ",", 
          RowBox[{"-", "Sch"}]}], "}"}], "}"}]}], "]"}]},
    {
     RowBox[{"ValID", "[", 
      RowBox[{"ChristoffelCDPDSch", ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"Sch", ",", 
           RowBox[{"-", "Sch"}], ",", "Sch"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"Sch", ",", "Sch", ",", 
           RowBox[{"-", "Sch"}]}], "}"}]}], "}"}]}], "]"}]},
    {
     RowBox[{"ValID", "[", 
      RowBox[{"ChristoffelCDPDSch", ",", 
       RowBox[{"{", 
        RowBox[{"{", 
         RowBox[{"Sch", ",", 
          RowBox[{"-", "Sch"}], ",", 
          RowBox[{"-", "Sch"}]}], "}"}], "}"}]}], "]"}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  ColumnForm[{
    xAct`xCoba`ValID[$CellContext`ChristoffelCDPDSch, {{-$CellContext`Sch, \
-$CellContext`Sch, -$CellContext`Sch}}], 
    xAct`xCoba`ValID[$CellContext`ChristoffelCDPDSch, {{$CellContext`Sch, \
-$CellContext`Sch, $CellContext`Sch}, {$CellContext`Sch, $CellContext`Sch, \
-$CellContext`Sch}}], 
    xAct`xCoba`ValID[$CellContext`ChristoffelCDPDSch, {{$CellContext`Sch, \
-$CellContext`Sch, -$CellContext`Sch}}]}],
  Editable->False]], "Output",
 CellChangeTimes->{3.7309151640134172`*^9, 3.730957582729422*^9, 
  3.730957700252664*^9, 3.7309577640444593`*^9, 3.730958197503948*^9, 
  3.730958913909342*^9, 3.730959983768639*^9, 3.730960423038514*^9, 
  3.7310036645433483`*^9, 3.7310039652658443`*^9},
 CellLabel->"Out[46]=",
 CellID->1301391671,ExpressionUUID->"ea0468b6-cd9f-4545-af5d-1c80e1f4c47b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["and this is when we did it:", "MathCaption",
 CellID->2144077367,ExpressionUUID->"9f7201f0-e74e-4786-a751-3286470d89cf"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"DateOfValID", "/@", 
   RowBox[{"TensorValIDs", "[", "ChristoffelCDPDSch", "]"}]}], "//", 
  "ColumnForm"}]], "Input",
 CellLabel->"In[47]:=",
 CellID->732710509,ExpressionUUID->"f6c19781-c59b-4a9b-ac56-aaa192fe499b"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {
     RowBox[{"{", 
      RowBox[{
      "2018", ",", "3", ",", "25", ",", "16", ",", "59", ",", 
       "25.05134`8.151405943643837"}], "}"}]},
    {
     RowBox[{"{", 
      RowBox[{
      "2018", ",", "3", ",", "25", ",", "16", ",", "59", ",", 
       "24.49587`8.141667851165957"}], "}"}]},
    {
     RowBox[{"{", 
      RowBox[{
      "2018", ",", "3", ",", "25", ",", "16", ",", "59", ",", 
       "23.928911`8.13149791729395"}], "}"}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  ColumnForm[{{2018, 3, 25, 16, 59, 25.05134`8.151405943643837}, {
    2018, 3, 25, 16, 59, 24.49587`8.141667851165957}, {
    2018, 3, 25, 16, 59, 23.928911`8.13149791729395}}],
  Editable->False]], "Output",
 CellChangeTimes->{3.730915168770265*^9, 3.730957583074251*^9, 
  3.73095770038631*^9, 3.730957764194343*^9, 3.730958197638166*^9, 
  3.730958914076089*^9, 3.730959983902007*^9, 3.73096042318944*^9, 
  3.7310036663944407`*^9, 3.7310039653123426`*^9},
 CellLabel->"Out[47]=",
 CellID->368869802,ExpressionUUID->"7d7eff90-0eaa-44b1-aacd-5553d162331d"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We can compute the same result directly with this function:", \
"MathCaption",
 CellChangeTimes->{{3.7309579403599253`*^9, 3.730957954871543*^9}},
 CellID->911802371,ExpressionUUID->"ecfac672-a095-48d0-ac02-d169c0e8b5f7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"christoffelB", "=", 
  RowBox[{
   RowBox[{"ChristoffelFromMetric", "[", 
    RowBox[{
     RowBox[{"metricg", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "b"}]}], "]"}], ",", "Sch"}], "]"}], "[", 
   RowBox[{"a", ",", 
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.7309575936941643`*^9, 3.7309576078191223`*^9}, {
  3.7309579883516407`*^9, 3.7309580061501837`*^9}},
 CellLabel->"In[48]:=",
 CellID->1275362256,ExpressionUUID->"adc6e788-a132-42e2-b12f-d2ca6a948400"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`a],
    Editable->False], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Cot",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cot,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`a],
    Editable->False], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Cot",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cot,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}], ")"}]}], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`a],
     Editable->False]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`a],
     Editable->False]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`a],
     Editable->False]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`a],
     Editable->False]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`a],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "+", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}]}], 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "3"]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`a],
     Editable->False]}], 
   RowBox[{
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`M,
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}], "-", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`a],
     Editable->False]}], 
   RowBox[{
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`M,
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}], "-", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`a],
     Editable->False]}], 
   RowBox[{
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`M,
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}], "-", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`a],
    Editable->False], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Cos",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Cos,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}], " ", 
   RowBox[{
    InterpretationBox[
     StyleBox["Sin",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Sin,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "a"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["M",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`M,
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}], ")"}], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}]}]], "Output",
 CellChangeTimes->{3.730957612441921*^9, 3.730957703499352*^9, 
  3.730957767262211*^9, 3.7309579590286016`*^9, 3.730958008542713*^9, 
  3.730958200171132*^9, 3.730958917134115*^9, 3.730959986182342*^9, 
  3.730960425599478*^9, 3.731003674612968*^9, 3.731003967648591*^9},
 CellLabel->"Out[48]=",
 CellID->1956956972,ExpressionUUID->"d9f5f813-cf75-4378-ae12-514bbad6febe"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"christoffel", "-", "christoffelB"}]], "Input",
 CellChangeTimes->{{3.730957973936082*^9, 3.7309579762227592`*^9}, {
  3.730958016086714*^9, 3.730958025715716*^9}},
 CellLabel->"In[49]:=",
 CellID->197703657,ExpressionUUID->"3f9c398e-e80c-414b-a790-2e063aac65a1"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.730957977594729*^9, 3.730958026497992*^9, 
  3.730958200304182*^9, 3.73095891742727*^9, 3.730959986336918*^9, 
  3.730960425745729*^9, 3.7310036767625237`*^9, 3.7310039677632017`*^9},
 CellLabel->"Out[49]=",
 CellID->1863774599,ExpressionUUID->"158a7bbd-e1b7-4d5a-b3f4-4bca8f872cfb"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["4.12.5. Compute Riemann", "Subsubsection",
 CellChangeTimes->{{3.7286771125390873`*^9, 3.7286771127222157`*^9}, {
   3.730913913338131*^9, 3.730913941103621*^9}, {3.730914255302225*^9, 
   3.730914260915913*^9}, {3.730914607546776*^9, 3.730914613449122*^9}, {
   3.730915185063478*^9, 3.730915186325802*^9}, {3.7309155101408854`*^9, 
   3.7309155279226313`*^9}, 3.730958995454458*^9},
 CellID->346755344,ExpressionUUID->"392ba1ff-32ee-4089-a98d-f55575f12edf"],

Cell["\<\
We now calculate the Riemann tensor. This time we use first a dedicated \
function and then we use the formula in terms of derivatives of Christoffel.\
\>", "Text",
 CellChangeTimes->{
  3.730915204727722*^9, {3.730959077602152*^9, 3.730959103513144*^9}, {
   3.7310037006233*^9, 3.73100370560054*^9}},
 CellID->490983037,ExpressionUUID->"2ee7dbad-069a-4faf-adc4-9ed7ccbdf8b1"],

Cell[CellGroupData[{

Cell["Compute the Riemann tensor directly:", "MathCaption",
 CellChangeTimes->{{3.730958227478345*^9, 3.730958236734091*^9}, {
  3.7309591076569557`*^9, 3.730959110441019*^9}},
 CellID->984582989,ExpressionUUID->"f64e8db9-bb0c-4a54-95b9-63128068dbe9"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"riemann", "=", 
  RowBox[{
   RowBox[{"RiemannFromChristoffel", "[", 
    RowBox[{
     RowBox[{"ChristoffelCDPDSch", "[", 
      RowBox[{"a", ",", 
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "c"}]}], "]"}], ",", "Sch"}], "]"}], "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}], ",", "d"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.730958064387971*^9, 3.730958094578826*^9}, {
  3.7309592354233503`*^9, 3.7309592365331383`*^9}, {3.730959798891603*^9, 
  3.730959820770982*^9}},
 CellLabel->"In[50]:=",
 CellID->1706034063,ExpressionUUID->"2ce53276-9ed0-43fe-a6a9-8942c5fb96d3"],

Cell[BoxData[
 RowBox[{
  FractionBox[
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}]}], 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "4"]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}]}], 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "4"]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}]}], 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "4"]], "-", 
  FractionBox[
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "+", 
  FractionBox[
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "-", 
  FractionBox[
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "+", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}]}], 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "4"]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "+", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}]}], 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "4"]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "+", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}]}], 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "4"]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "-", 
  FractionBox[
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]]}]], "Output",
 CellChangeTimes->{
  3.7309581062810307`*^9, 3.730958211417862*^9, 3.730958366784244*^9, 
   3.730959006780151*^9, 3.730959158107045*^9, 3.730959240495644*^9, {
   3.730959807566533*^9, 3.730959824691669*^9}, 3.7309599894812603`*^9, 
   3.7309604288846493`*^9, 3.731003718188533*^9, 3.7310039710892677`*^9},
 CellLabel->"Out[50]=",
 CellID->1263210251,ExpressionUUID->"f9f63f23-65f4-4216-a8ed-d752b3f894a4"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Check that the corresponding Ricci tensor vanishes:", "MathCaption",
 CellChangeTimes->{{3.7309583779771633`*^9, 3.7309583855371*^9}, {
  3.730959114168536*^9, 3.73095911687269*^9}},
 CellID->1663828751,ExpressionUUID->"5a4d012b-9501-4eda-b0df-2957b7ff24d2"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"riemann", " ", 
   RowBox[{"delta", "[", 
    RowBox[{
     RowBox[{"-", "d"}], ",", "b"}], "]"}]}], "//", 
  "ContractBasis"}]], "Input",
 CellChangeTimes->{{3.730958246895479*^9, 3.730958262549251*^9}, {
  3.730958355532761*^9, 3.730958370232649*^9}, {3.7309592406716003`*^9, 
  3.7309592417728252`*^9}},
 CellLabel->"In[51]:=",
 CellID->72553858,ExpressionUUID->"4b87912c-0cb7-4042-b6e1-0910eeedc719"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{
  3.730958269814745*^9, 3.73095837093034*^9, 3.730959024498209*^9, 
   3.730959158310833*^9, {3.73095924072255*^9, 3.730959242508627*^9}, 
   3.7309598271318417`*^9, 3.730959989874803*^9, 3.73096042902667*^9, 
   3.7310037192938347`*^9, 3.73100397122121*^9},
 CellLabel->"Out[51]=",
 CellID->751354315,ExpressionUUID->"9e6032a9-1a24-48ba-8abf-c4adfeafb75e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Alternatively, we can start from the formula of the Riemann tensor in terms \
of derivatives of the Christoffel tensor:\
\>", "MathCaption",
 CellChangeTimes->{{3.7309591238813972`*^9, 3.730959150839679*^9}},
 CellID->417221997,ExpressionUUID->"600273b2-62d5-4693-a574-70a1f6a85ea6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ChangeCurvature", "[", 
  RowBox[{
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], ",", "CD", ",", "PDSch"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.730958433271613*^9, 3.730958491119947*^9}},
 CellLabel->"In[52]:=",
 CellID->42034799,ExpressionUUID->"904aa7ce-1b36-4d1f-9ad5-3bc5e44cfead"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCDPDSch[$CellContext`d, -$CellContext`b, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCDPDSch[$CellContext`e, -$CellContext`a, \
-$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCDPDSch[$CellContext`d, -$CellContext`a, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCDPDSch[$CellContext`e, -$CellContext`b, \
-$CellContext`c],
    Editable->False]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox[
      StyleBox["\[ScriptCapitalD]",
       FontColor->RGBColor[1., 0., 0.]], "a"], GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`PDSch[-$CellContext`a][
    $CellContext`ChristoffelCDPDSch[$CellContext`d, -$CellContext`b, \
-$CellContext`c]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox[
      StyleBox["\[ScriptCapitalD]",
       FontColor->RGBColor[1., 0., 0.]], "b"], GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`PDSch[-$CellContext`b][
    $CellContext`ChristoffelCDPDSch[$CellContext`d, -$CellContext`a, \
-$CellContext`c]],
   Editable->False]}]], "Output",
 CellChangeTimes->{{3.7309584381343803`*^9, 3.730958518711958*^9}, 
   3.7309585626069736`*^9, 3.730958605391366*^9, {3.7309590257420483`*^9, 
   3.73095903806115*^9}, 3.7309591584948473`*^9, 3.7309592442577047`*^9, 
   3.730959828464398*^9, 3.730959990102323*^9, 3.730960429189394*^9, 
   3.731003727383985*^9, 3.731003971297125*^9},
 CellLabel->"Out[52]=",
 CellID->808127845,ExpressionUUID->"93e1d808-70ec-4465-8051-3486d94da4c7"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Expand into all actual components, insert the actual values, and compute the \
derivatives:\
\>", "MathCaption",
 CellChangeTimes->{{3.73095916757423*^9, 3.730959196894225*^9}},
 CellID->1364939662,ExpressionUUID->"1d009b24-8d61-4d65-a800-7aa7df520be1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"BasisExpand", "[", 
   RowBox[{"%", ",", "Sch"}], "]"}], "//", "ToValues"}]], "Input",
 CellChangeTimes->{{3.730958491896838*^9, 3.7309585400222073`*^9}},
 CellLabel->"In[53]:=",
 CellID->2039489065,ExpressionUUID->"1aa1cd16-5036-405e-b4b0-9e788175c311"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Sch}],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Cos",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Cos,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Cos",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Cos,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Cot",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Cot,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Cot",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Cot,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Csc",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Csc,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Csc",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Csc,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "+", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}]}], 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "4"]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "+", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}]}], 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "4"]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "+", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}]}], 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "4"]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "+", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}]}], 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "4"]], "-", 
  FractionBox[
   RowBox[{"2", " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`M,
      Editable->False], "2"], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], "-", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"]}], ")"}], "2"]], "+", 
  FractionBox[
   RowBox[{"2", " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`M,
      Editable->False], "2"], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], "-", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"]}], ")"}], "2"]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}]}], ")"}]}], 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], "-", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"]}], ")"}], "2"]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}]}], ")"}]}], 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], "-", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"]}], ")"}], "2"]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}]}], 
   RowBox[{
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`M,
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}], "-", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}]}], 
   RowBox[{
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`M,
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}], "-", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}]}], 
   RowBox[{
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`M,
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}], "-", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}]}], 
   RowBox[{
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`M,
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}], "-", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   RowBox[{
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], "-", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"]}], ")"}]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   RowBox[{
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], "-", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"]}], ")"}]}]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   RowBox[{
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], "-", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"]}], ")"}]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   RowBox[{
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], "-", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"]}], ")"}]}]], "-", 
  FractionBox[
   RowBox[{"2", " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`M,
      Editable->False], "2"], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "+", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}]}], 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "3"], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], "-", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"]}], ")"}]}]], "+", 
  FractionBox[
   RowBox[{"2", " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`M,
      Editable->False], "2"], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "+", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}]}], 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "3"], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], "-", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"]}], ")"}]}]], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox["M",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`M,
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     FractionBox["1", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "3"]], "-", 
     FractionBox[
      RowBox[{"3", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "2"}], " ", 
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`M,
           Editable->False]}], "+", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]}], ")"}]}], 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "4"]]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox["M",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`M,
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     FractionBox["1", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "3"]], "-", 
     FractionBox[
      RowBox[{"3", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "2"}], " ", 
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`M,
           Editable->False]}], "+", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]}], ")"}]}], 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "4"]]}], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["Sin",
       ShowAutoStyles->False,
       AutoSpacing->False],
      Sin,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox["\[Theta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Theta][],
      Editable->False], "]"}], "2"]}], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}], " ", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}], " ", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}], " ", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}], " ", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}], " ", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"]}], 
   RowBox[{
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`M,
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}], "-", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}], " ", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"]}], 
   RowBox[{
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`M,
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}], "-", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}], " ", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"]}], 
   RowBox[{
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`M,
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}], "-", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}], " ", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"]}], 
   RowBox[{
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`M,
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}], "-", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      SuperscriptBox[
       RowBox[{
        InterpretationBox[
         StyleBox["Cos",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Cos,
         Editable->False], "[", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False], "]"}], "2"]}], "+", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", "d"},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      SuperscriptBox[
       RowBox[{
        InterpretationBox[
         StyleBox["Cos",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Cos,
         Editable->False], "[", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False], "]"}], "2"]}], "+", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Sin",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Sin,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}], ")"}]}]}]], "Output",
 CellChangeTimes->{{3.730958504437758*^9, 3.730958521440362*^9}, 
   3.730958566032119*^9, 3.7309586081661587`*^9, 3.7309590407454576`*^9, 
   3.73095924737*^9, 3.730959831278563*^9, 3.730959991758605*^9, 
   3.730960430858406*^9, 3.731003733746972*^9, 3.731003973006432*^9},
 CellLabel->"Out[53]=",
 CellID->1518107287,ExpressionUUID->"85391685-8a06-404b-a678-96ce5ba05862"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"riemannB", "=", 
  RowBox[{"BasisCollect", "[", 
   RowBox[{"%", ",", "Simplify"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.7309586083549547`*^9, 3.730958614792557*^9}, {
  3.7309590354368267`*^9, 3.730959035699492*^9}, {3.730959247569965*^9, 
  3.730959248894054*^9}},
 CellLabel->"In[54]:=",
 CellID->875107252,ExpressionUUID->"37831916-acf2-4903-a995-09b51fa8d395"],

Cell[BoxData[
 RowBox[{
  FractionBox[
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}]}], 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "4"]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}]}], 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "4"]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}]}], 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "4"]], "-", 
  FractionBox[
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "+", 
  FractionBox[
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "-", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "-", 
  FractionBox[
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`d],
     Editable->False]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "+", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}]}], 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "4"]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "+", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}]}], 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "4"]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False]}], "+", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False]}], ")"}]}], 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "4"]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {0, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "+", 
  FractionBox[
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]], "-", 
  FractionBox[
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`M,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`b, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`Sch}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "d"},
            {
             StyleBox["2",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`Sch}, $CellContext`d],
     Editable->False], " ", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"]}], 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]]}]], "Output",
 CellChangeTimes->{3.730958615568624*^9, 3.7309590431228733`*^9, 
  3.7309592511896772`*^9, 3.730959832635169*^9, 3.7309599933582172`*^9, 
  3.730960432121738*^9, 3.731003738084366*^9, 3.7310039743221483`*^9},
 CellLabel->"Out[54]=",
 CellID->575962602,ExpressionUUID->"07433b5b-c5e7-4b3c-bd71-0fd7fba0a765"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"riemann", "-", "riemannB"}]], "Input",
 CellChangeTimes->{{3.7309592534923983`*^9, 3.730959256316256*^9}},
 CellLabel->"In[55]:=",
 CellID->452783367,ExpressionUUID->"f30bd70c-119c-49cb-91fd-820ef14ef05c"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.730959257144738*^9, 3.730959832831428*^9, 
  3.7309599935709667`*^9, 3.730960432277549*^9, 3.731003739925111*^9, 
  3.731003974397814*^9},
 CellLabel->"Out[55]=",
 CellID->1494654785,ExpressionUUID->"623de4dd-87d2-44a8-8575-32b56fbb2f08"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Store the results in the TensorValues framework:", "MathCaption",
 CellChangeTimes->{{3.730960135526763*^9, 3.730960147022396*^9}},
 CellID->523132918,ExpressionUUID->"fda1508c-63b5-479a-862f-a4d36c45fd63"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentArray", "[", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"riemann", " ", 
       RowBox[{"downmetric", "[", 
        RowBox[{
         RowBox[{"-", "d"}], ",", 
         RowBox[{"-", "e"}]}], "]"}]}], "//", 
      RowBox[{"ToBasis", "[", "Sch", "]"}]}], "//", "ContractBasis"}], "//", 
    "ToValues"}], "//", "Simplify"}], "]"}]], "Input",
 CellChangeTimes->{{3.730959268997254*^9, 3.730959284699109*^9}, {
  3.7309593892182302`*^9, 3.730959392447698*^9}, {3.7309595873384323`*^9, 
  3.730959605696604*^9}, {3.730959869735909*^9, 3.730959924678236*^9}},
 CellLabel->"In[56]:=",
 CellID->664211117,ExpressionUUID->"1c1521ce-bd09-4364-b7bd-000795caf175"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", 
         RowBox[{"-", 
          FractionBox[
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`M,
             Editable->False]}], 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "3"]]}], ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         FractionBox[
          RowBox[{"2", " ", 
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`M,
            Editable->False]}], 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "3"]], ",", "0", ",", "0", ",", "0"}], "}"}], 
       ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", 
         FractionBox[
          RowBox[{
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`M,
            Editable->False], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "2"}], " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`M,
               Editable->False]}], "+", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]}], ")"}]}], 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]], ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", 
          FractionBox[
           RowBox[{
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`M,
             Editable->False], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "2"}], " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`M,
                Editable->False]}], "+", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}], ")"}]}], 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]]}], ",", "0", ",", "0", ",", "0"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", 
         FractionBox[
          RowBox[{
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`M,
            Editable->False], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "2"}], " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`M,
               Editable->False]}], "+", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]}], ")"}], " ", 
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["Sin",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Sin,
              Editable->False], "[", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False], "]"}], "2"]}], 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", 
          FractionBox[
           RowBox[{
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`M,
             Editable->False], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "2"}], " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`M,
                Editable->False]}], "+", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}], ")"}], " ", 
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["Sin",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Sin,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], "2"]}], 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]]}], ",", "0", ",", "0", ",", "0"}], 
        "}"}]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", 
         FractionBox[
          RowBox[{"2", " ", 
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`M,
            Editable->False]}], 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "3"]], ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", 
          FractionBox[
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`M,
             Editable->False]}], 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "3"]]}], ",", "0", ",", "0", ",", "0"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", 
         FractionBox[
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`M,
           Editable->False], 
          RowBox[{
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`M,
             Editable->False]}], "-", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]}]], ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", 
         RowBox[{"-", 
          FractionBox[
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`M,
            Editable->False], 
           RowBox[{
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["M",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`M,
              Editable->False]}], "-", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]}]]}], ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", 
         FractionBox[
          RowBox[{
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`M,
            Editable->False], " ", 
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["Sin",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Sin,
              Editable->False], "[", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False], "]"}], "2"]}], 
          RowBox[{
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`M,
             Editable->False]}], "-", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]}]]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", 
         RowBox[{"-", 
          FractionBox[
           RowBox[{
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`M,
             Editable->False], " ", 
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["Sin",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Sin,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], "2"]}], 
           RowBox[{
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["M",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`M,
              Editable->False]}], "-", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]}]]}], ",", "0", ",", "0"}], "}"}]}], "}"}]}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", 
         FractionBox[
          RowBox[{
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`M,
            Editable->False], " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "2"}], " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`M,
                Editable->False]}], "+", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}], ")"}], "2"]}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`M,
               Editable->False]}], "-", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]}], ")"}], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}]], ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", 
          FractionBox[
           RowBox[{
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`M,
             Editable->False], " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"-", "2"}], " ", 
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`M,
                 Editable->False]}], "+", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}], ")"}], "2"]}], 
           RowBox[{
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`M,
                Editable->False]}], "-", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}], ")"}], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"]}]]}], ",", "0", ",", "0", ",", "0"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", 
         RowBox[{"-", 
          FractionBox[
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`M,
            Editable->False], 
           RowBox[{
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["M",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`M,
              Editable->False]}], "-", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]}]]}], ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", 
         FractionBox[
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`M,
           Editable->False], 
          RowBox[{
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`M,
             Editable->False]}], "-", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]}]], ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", 
         RowBox[{"-", 
          FractionBox[
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`M,
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "2"}], " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`M,
                Editable->False]}], "+", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}], ")"}], " ", 
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["Sin",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Sin,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], "2"]}], 
           RowBox[{
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["M",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`M,
              Editable->False]}], "-", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]}]]}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", 
         FractionBox[
          RowBox[{"2", " ", 
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`M,
            Editable->False], " ", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "2"}], " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`M,
               Editable->False]}], "+", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]}], ")"}], " ", 
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["Sin",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Sin,
              Editable->False], "[", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False], "]"}], "2"]}], 
          RowBox[{
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`M,
             Editable->False]}], "-", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]}]], ",", "0"}], "}"}]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", 
         FractionBox[
          RowBox[{
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`M,
            Editable->False], " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "2"}], " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`M,
                Editable->False]}], "+", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}], ")"}], "2"], " ", 
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["Sin",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Sin,
              Editable->False], "[", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False], "]"}], "2"]}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`M,
               Editable->False]}], "-", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]}], ")"}], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}]]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", 
          FractionBox[
           RowBox[{
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`M,
             Editable->False], " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"-", "2"}], " ", 
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`M,
                 Editable->False]}], "+", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}], ")"}], "2"], " ", 
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["Sin",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Sin,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], "2"]}], 
           RowBox[{
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`M,
                Editable->False]}], "-", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}], ")"}], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"]}]]}], ",", "0", ",", "0", ",", "0"}], 
        "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", 
         RowBox[{"-", 
          FractionBox[
           RowBox[{
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`M,
             Editable->False], " ", 
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["Sin",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Sin,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], "2"]}], 
           RowBox[{
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["M",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`M,
              Editable->False]}], "-", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]}]]}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", 
         FractionBox[
          RowBox[{
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`M,
            Editable->False], " ", 
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["Sin",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Sin,
              Editable->False], "[", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False], "]"}], "2"]}], 
          RowBox[{
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`M,
             Editable->False]}], "-", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]}]], ",", "0", ",", "0"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", 
         RowBox[{
          RowBox[{"-", "2"}], " ", 
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`M,
           Editable->False], " ", 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Sin",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Sin,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "2"]}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", 
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`M,
           Editable->False], " ", 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Sin",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Sin,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "2"]}], ",", "0"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}]}], 
    "}"}]}], "}"}]], "Output",
 CellChangeTimes->{
  3.730959285289893*^9, 3.730959394289113*^9, {3.730959595333889*^9, 
   3.730959606229347*^9}, {3.730959908854148*^9, 3.7309599307398033`*^9}, 
   3.730959999167808*^9, 3.730960437848219*^9, 3.7310037540935993`*^9, 
   3.731003979953491*^9},
 CellLabel->"Out[56]=",
 CellID->1546752536,ExpressionUUID->"fca560cf-16de-4f26-8c6e-7f14484a00e4"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"AllComponentValues", "[", 
   RowBox[{
    RowBox[{"RiemannCD", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "a"}], ",", 
        RowBox[{"-", "Sch"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "b"}], ",", 
        RowBox[{"-", "Sch"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "c"}], ",", 
        RowBox[{"-", "Sch"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "d"}], ",", 
        RowBox[{"-", "Sch"}]}], "}"}]}], "]"}], ",", "%"}], "]"}], 
  ";"}]], "Input",
 CellChangeTimes->{{3.730959690965798*^9, 3.7309597427185707`*^9}, {
  3.730959939455212*^9, 3.7309599410207653`*^9}},
 CellLabel->"In[57]:=",
 CellID->1318404106,ExpressionUUID->"5a398ef2-5732-401d-aed7-3c1dcb9ee78c"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Print the independent components only:", "MathCaption",
 CellChangeTimes->{
  3.730959756531416*^9, {3.730960104409368*^9, 3.7309601138555326`*^9}, {
   3.730960160974958*^9, 3.7309601616298943`*^9}},
 CellID->1376471820,ExpressionUUID->"d70976d4-bff3-4435-b204-45d2494da542"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IndependentRules", "[", 
   RowBox[{"TensorValues", "[", "RiemannCD", "]"}], "]"}], "//", 
  "Column"}]], "Input",
 CellChangeTimes->{{3.730959766796323*^9, 3.730959773171053*^9}, {
  3.730959954022172*^9, 3.730959955909058*^9}, {3.730960078882218*^9, 
  3.730960091248302*^9}},
 CellLabel->"In[58]:=",
 CellID->1668033256,ExpressionUUID->"7bc2c9c1-a605-4db0-aa03-1a99c53e8434"],

Cell[BoxData[
 TagBox[GridBox[{
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {0, -$CellContext`Sch}, {1, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "3"]]}]}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`Sch}, {
        1, -$CellContext`Sch}, {2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`Sch}, {
        2, -$CellContext`Sch}, {0, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      FractionBox[
       RowBox[{
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "2"}], " ", 
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`M,
            Editable->False]}], "+", 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False]}], ")"}]}], 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"]]}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`Sch}, {
        2, -$CellContext`Sch}, {0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`Sch}, {
        2, -$CellContext`Sch}, {1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`Sch}, {
        2, -$CellContext`Sch}, {1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`Sch}, {
        2, -$CellContext`Sch}, {2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`Sch}, {
        3, -$CellContext`Sch}, {0, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`M,
             Editable->False]}], "-", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]}], ")"}], " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Sin",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Sin,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}], 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]]}]}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`Sch}, {
        3, -$CellContext`Sch}, {1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`Sch}, {
        3, -$CellContext`Sch}, {1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`Sch}, {
        3, -$CellContext`Sch}, {2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{1, -$CellContext`Sch}, {
        2, -$CellContext`Sch}, {1, -$CellContext`Sch}, {2, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      FractionBox[
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False], 
       RowBox[{
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], "-", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]}]]}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{1, -$CellContext`Sch}, {
        2, -$CellContext`Sch}, {1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{1, -$CellContext`Sch}, {
        2, -$CellContext`Sch}, {2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{1, -$CellContext`Sch}, {
        3, -$CellContext`Sch}, {1, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      FractionBox[
       RowBox[{
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`M,
         Editable->False], " ", 
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Sin",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Sin,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"]}], 
       RowBox[{
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`M,
          Editable->False]}], "-", 
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False]}]]}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{1, -$CellContext`Sch}, {
        3, -$CellContext`Sch}, {2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", "0"}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{2, -$CellContext`Sch}, {
        3, -$CellContext`Sch}, {2, -$CellContext`Sch}, {3, -$CellContext`Sch}],
       Editable->False], "\[Rule]", 
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`M,
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], " ", 
       SuperscriptBox[
        RowBox[{
         InterpretationBox[
          StyleBox["Sin",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Sin,
          Editable->False], "[", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False], "]"}], "2"]}]}]}
   },
   DefaultBaseStyle->"Column",
   GridBoxAlignment->{"Columns" -> {{Left}}},
   GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
  "Column"]], "Output",
 CellChangeTimes->{{3.7309597584815693`*^9, 3.7309597743027973`*^9}, {
   3.7309599495904207`*^9, 3.730959957082068*^9}, 3.730960001437522*^9, 
   3.730960091862981*^9, 3.730960439535017*^9, 3.731003759102469*^9, 
   3.73100398150698*^9},
 CellLabel->"Out[58]=",
 CellID->980325617,ExpressionUUID->"77fa5dc4-8d7c-4af7-8841-bcad716dd4d7"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["4.12.6. Clean up", "Subsubsection",
 CellChangeTimes->{{3.7286771125390873`*^9, 3.7286771127222157`*^9}, {
   3.730913913338131*^9, 3.730913941103621*^9}, {3.730914255302225*^9, 
   3.730914260915913*^9}, {3.730914607546776*^9, 3.730914613449122*^9}, {
   3.730915185063478*^9, 3.730915186325802*^9}, {3.7309155101408854`*^9, 
   3.7309155279226313`*^9}, 3.730958995454458*^9, {3.730960223229556*^9, 
   3.7309602268348713`*^9}},
 CellID->1364148720,ExpressionUUID->"1600e4f9-c0a8-4191-aced-c8c6a812c098"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefMetric", "[", "metricg", "]"}]], "Input",
 CellChangeTimes->{{3.730960232386859*^9, 3.73096023843542*^9}},
 CellLabel->"In[59]:=",
 CellID->268720204,ExpressionUUID->"e09043d3-7842-401d-99cd-f1df58d2da64"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "DetmetricgSch"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "weight +2 density", 
   " ", $CellContext`DetmetricgSch],
  Editable->False]], "Print",
 CellChangeTimes->{3.73096023883737*^9, 3.730960545888906*^9, 
  3.7310040292164707`*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->1480324350,ExpressionUUID->"c4467cd6-9876-4b72-a92d-36fab95ddaeb"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Detmetricg"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "weight +2 density", 
   " ", $CellContext`Detmetricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.73096023883737*^9, 3.730960545888906*^9, 
  3.731004029231159*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->1213299955,ExpressionUUID->"56d4f122-0720-40f2-83f2-5789e28228a9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "ChristoffelCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Christoffel tensor", " ", $CellContext`ChristoffelCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.73096023883737*^9, 3.730960545888906*^9, 
  3.731004029244244*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->1575115420,ExpressionUUID->"749ec44c-22db-4aaa-9645-a5a5ef2df16f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "ChristoffelCDPDSch"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`ChristoffelCDPDSch],
  Editable->False]], "Print",
 CellChangeTimes->{3.73096023883737*^9, 3.730960545888906*^9, 
  3.7310040292539167`*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->933947371,ExpressionUUID->"ae61ed66-33d1-4a5b-becf-afc75e0bb75d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Einstein tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "EinsteinCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Einstein tensor", " ", $CellContext`EinsteinCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.73096023883737*^9, 3.730960545888906*^9, 
  3.731004029263711*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->895488474,ExpressionUUID->"11db8726-5c77-462a-97b0-6bf957eec1fd"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Kretschmann scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "KretschmannCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Kretschmann scalar", 
   " ", $CellContext`KretschmannCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.73096023883737*^9, 3.730960545888906*^9, 
  3.7310040292738743`*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->909384342,ExpressionUUID->"16f8f05e-58ee-4601-b046-a20f35eee716"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric Ricci tensor", 
   " ", $CellContext`RicciCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.73096023883737*^9, 3.730960545888906*^9, 
  3.731004029283555*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->381520295,ExpressionUUID->"307e3fe4-9acc-4715-a2e4-c1219fda9984"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciScalarCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Ricci scalar", 
   " ", $CellContext`RicciScalarCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.73096023883737*^9, 3.730960545888906*^9, 
  3.7310040292929697`*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->1972911770,ExpressionUUID->"838ed2cb-f866-4da2-b98a-2f82e13b4a96"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.73096023883737*^9, 3.730960545888906*^9, 
  3.731004029302433*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->1580608710,ExpressionUUID->"4830fafb-cf39-497a-bb0a-34e6bddd0fa6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric TFRicci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TFRicciCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric TFRicci tensor",
    " ", $CellContext`TFRicciCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.73096023883737*^9, 3.730960545888906*^9, 
  3.7310040293118134`*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->917875580,ExpressionUUID->"9f9481c5-25af-4482-8c9f-45ff396c41d5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TorsionCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.73096023883737*^9, 3.730960545888906*^9, 
  3.731004029321364*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->1359836659,ExpressionUUID->"7377b964-2cb8-470a-811b-2364038c245a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Weyl tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "WeylCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Weyl tensor", 
   " ", $CellContext`WeylCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.73096023883737*^9, 3.730960545888906*^9, 
  3.731004029330699*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->1253756439,ExpressionUUID->"652d7cc1-eb47-4489-a678-e6915290127a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "CD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "covariant derivative", 
   " ", $CellContext`CD],
  Editable->False]], "Print",
 CellChangeTimes->{3.73096023883737*^9, 3.730960545888906*^9, 
  3.7310040293402033`*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->774925909,ExpressionUUID->"710f5e63-69bf-49e4-8cc8-8a8a2e7d1b51"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tetrametric\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "Tetrametricg\[Dagger]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tetrametric", 
   " ", $CellContext`Tetrametricg\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{3.73096023883737*^9, 3.730960545888906*^9, 
  3.731004029349553*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->1838917155,ExpressionUUID->"c9e8cbf3-18ec-46a7-9a82-d0185a5ec905"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tetrametric\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Tetrametricg"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tetrametric", 
   " ", $CellContext`Tetrametricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.73096023883737*^9, 3.730960545888906*^9, 
  3.731004029358982*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->95976461,ExpressionUUID->"3dbebf0a-6f4a-47ba-aa63-99c91a4ccdc1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "epsilonmetricg"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric tensor", 
   " ", $CellContext`epsilonmetricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.73096023883737*^9, 3.730960545888906*^9, 
  3.7310040293686123`*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->316161130,ExpressionUUID->"859a5819-91ab-476f-be02-405b23ff59ba"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric metric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "metricg"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric metric tensor", 
   " ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.73096023883737*^9, 3.730960545888906*^9, 
  3.73100402937836*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->845569232,ExpressionUUID->"7c12dac8-2577-47db-8885-73ff69290bb5"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefConstantSymbol", "[", "M", "]"}]], "Input",
 CellChangeTimes->{{3.730960318146244*^9, 3.730960329864518*^9}},
 CellLabel->"In[60]:=",
 CellID->22553876,ExpressionUUID->"5b5c42b8-c279-4c08-93f8-87f64be4096f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefConstantSymbol", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"constant symbol\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["M",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`M,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefConstantSymbol, ": Undefined ", "constant symbol", 
   " ", $CellContext`M],
  Editable->False]], "Print",
 CellChangeTimes->{3.7309603306998167`*^9, 3.730960547207429*^9, 
  3.7310040296144733`*^9},
 CellLabel->"During evaluation of In[60]:=",
 CellID->1582790025,ExpressionUUID->"bd4be96b-0d7b-4edb-9d84-0edfb2e0fd72"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefBasis", "/@", 
  RowBox[{"{", 
   RowBox[{"Klosch", ",", "Sch"}], "}"}]}]], "Input",
 CellChangeTimes->{{3.730960249970537*^9, 3.730960266978402*^9}},
 CellLabel->"In[61]:=",
 CellID->1704242589,ExpressionUUID->"15e94ce6-a49e-4f49-beee-a6276d28a78a"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "x"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "coordinate scalar", 
   " ", $CellContext`x],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.73100402974002*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->1072014786,ExpressionUUID->"7aead2fb-f053-42b3-ad89-188c6fac6126"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "y"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "coordinate scalar", 
   " ", $CellContext`y],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.7310040297548237`*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->2112193619,ExpressionUUID->"160ea8f3-8b6e-4085-ba21-09d716a728dc"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "\[CapitalTheta]"}],
  
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "coordinate scalar", 
   " ", $CellContext`\[CapitalTheta]],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.7310040297685833`*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->1733469623,ExpressionUUID->"41b951a0-d4ca-4070-8843-60e84f79cec2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "\[CapitalPhi]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "coordinate scalar", 
   " ", $CellContext`\[CapitalPhi]],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.731004029776483*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->1660798559,ExpressionUUID->"d0aca02b-de3b-4807-973b-50b182f299f8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"mapping differential tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "diKlosch"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "mapping differential tensor", " ", $CellContext`diKlosch],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.731004029784536*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->48937407,ExpressionUUID->"169da2fc-82b4-4a95-8bd5-f86e57e07612"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefMapping", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"inverse mapping\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "iKlosch"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefMapping, ": Undefined ", "inverse mapping", 
   " ", $CellContext`iKlosch],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.731004029916943*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->4762648,ExpressionUUID->"1180dc5a-010c-408c-b320-9092c4e5d882"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"mapping differential tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "dKlosch"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "mapping differential tensor", " ", $CellContext`dKlosch],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.731004029927948*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->588929997,ExpressionUUID->"3ef44509-2b35-4a9c-91de-013ecdaade92"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "ChristoffelPDKlosch"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Christoffel tensor", " ", $CellContext`ChristoffelPDKlosch],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.731004029938568*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->394330556,ExpressionUUID->"0dda059f-07b8-4a97-970a-bdebe1062733"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciPDKlosch"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Ricci tensor", 
   " ", $CellContext`RicciPDKlosch],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.73100402994948*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->379120050,ExpressionUUID->"72f990af-0eab-4760-af46-ddd65a38d8b1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannPDKlosch"}],
  
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannPDKlosch],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.731004029960186*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->286699701,ExpressionUUID->"32da0bfb-e76c-4d4c-94aa-8c6b54cc424b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TorsionPDKlosch"}],
  
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionPDKlosch],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.7310040299739*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->1756054486,ExpressionUUID->"1ca91084-6cbc-48f7-a6c5-1a6c358cc03b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"parallel derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "PDKlosch"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "parallel derivative", 
   " ", $CellContext`PDKlosch],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.731004029984591*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->1139985481,ExpressionUUID->"34c8d1d6-5f1a-4bcd-b44b-bb488d875673"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "etaUpKlosch"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric +1 density",
    " ", $CellContext`etaUpKlosch],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.731004029994562*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->2034364280,ExpressionUUID->"a65c812b-c13f-43ec-bd3e-cbd697a4e36b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "etaDownKlosch"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric -1 density",
    " ", $CellContext`etaDownKlosch],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.731004030004335*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->1473310073,ExpressionUUID->"c295ac7c-6a13-4630-b4cd-ba0a677003cf"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefBasis", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"basis\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "Klosch"}],
  SequenceForm[
  "** ", xAct`xCoba`UndefBasis, ": Undefined ", "basis", 
   " ", $CellContext`Klosch],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.7310040300140753`*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->1435947001,ExpressionUUID->"041ada7e-6d51-42d4-bd77-39e2bb69f906"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "\[Tau]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "coordinate scalar", 
   " ", $CellContext`\[Tau]],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.731004030024034*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->540020821,ExpressionUUID->"89a3edc4-63fa-45ab-b2cc-e3adc62c7110"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "r"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "coordinate scalar", 
   " ", $CellContext`r],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.73100403003371*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->864046325,ExpressionUUID->"01de40d3-6270-49e1-99c8-95eb29dc51c5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "\[Theta]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "coordinate scalar", 
   " ", $CellContext`\[Theta]],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.731004030043355*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->742163261,ExpressionUUID->"d54a7889-a704-4d53-8b6f-ec33b020d94b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "\[Phi]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "coordinate scalar", 
   " ", $CellContext`\[Phi]],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.7310040300535307`*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->952905851,ExpressionUUID->"857af552-24f9-4797-bf4b-20ad964aaf82"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"mapping differential tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "diSch"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "mapping differential tensor", " ", $CellContext`diSch],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.731004030063115*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->1703082078,ExpressionUUID->"70e53f76-ac1c-44cc-97fd-1c56bdbab74a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefMapping", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"inverse mapping\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "iSch"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefMapping, ": Undefined ", "inverse mapping", 
   " ", $CellContext`iSch],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.731004030128927*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->2038437730,ExpressionUUID->"3460c327-12e5-4849-826b-cd0d6c23d027"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"mapping differential tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "dSch"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "mapping differential tensor", " ", $CellContext`dSch],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.73100403013888*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->1216307519,ExpressionUUID->"42cb6630-ad89-40fd-811f-219b8c50e4a1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "ChristoffelPDSch"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Christoffel tensor", " ", $CellContext`ChristoffelPDSch],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.731004030146433*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->583636040,ExpressionUUID->"aeaa498b-8c91-4954-9ea7-fa39b80d986a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciPDSch"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Ricci tensor", 
   " ", $CellContext`RicciPDSch],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.731004030154003*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->1613108431,ExpressionUUID->"1b6fad96-6614-4da9-a63f-b001828dfb46"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannPDSch"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannPDSch],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.731004030161529*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->1679840089,ExpressionUUID->"df84f3b9-0ff0-4acc-baa4-53fcce6faf38"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TorsionPDSch"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionPDSch],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.731004030168723*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->104868714,ExpressionUUID->"b8beb026-aa03-4aed-ac5b-834151f0a518"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"parallel derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "PDSch"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "parallel derivative", 
   " ", $CellContext`PDSch],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.731004030175763*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->1591885658,ExpressionUUID->"2febfcf8-09a3-4b8c-82f0-f4df99fb25fc"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "etaUpSch"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric +1 density",
    " ", $CellContext`etaUpSch],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.73100403018264*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->1470979680,ExpressionUUID->"fe6758be-7307-479e-8693-48e6a28d3e9b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "etaDownSch"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric -1 density",
    " ", $CellContext`etaDownSch],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.731004030190033*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->923737496,ExpressionUUID->"52796659-f7a8-40a2-9101-951ec0520869"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefBasis", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"basis\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "Sch"}],
  SequenceForm[
  "** ", xAct`xCoba`UndefBasis, ": Undefined ", "basis", 
   " ", $CellContext`Sch],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.731004030197925*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->1096811109,ExpressionUUID->"40e1095a-ba17-40b6-ae0f-63fcdc0cdd39"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"Null", ",", "Null"}], "}"}]], "Output",
 CellChangeTimes->{{3.730960260942861*^9, 3.7309602678202*^9}, 
   3.7309605482422028`*^9, 3.7310040302049837`*^9},
 CellLabel->"Out[61]=",
 CellID->104683115,ExpressionUUID->"a8ce465f-d5cf-43a8-9a5f-73792a2aee9a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M4", "]"}]], "Input",
 CellChangeTimes->{{3.730960243755797*^9, 3.7309602878662977`*^9}},
 CellLabel->"In[62]:=",
 CellID->755307977,ExpressionUUID->"1c73492a-126a-479c-b2c9-500368e3b893"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefVBundle", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]M4",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentM4,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefVBundle, ": Undefined ", "vbundle", 
   " ", $CellContext`TangentM4],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960288299341*^9, 3.7309605493598957`*^9, 
  3.731004030540015*^9},
 CellLabel->"During evaluation of In[62]:=",
 CellID->620006912,ExpressionUUID->"86347138-8c19-4f51-b0ca-292254efbb95"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefManifold", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"manifold\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "M4"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefManifold, ": Undefined ", "manifold", 
   " ", $CellContext`M4],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960288299341*^9, 3.7309605493598957`*^9, 
  3.731004030554305*^9},
 CellLabel->"During evaluation of In[62]:=",
 CellID->865514740,ExpressionUUID->"243f0019-c2b2-4e73-ae81-aaf5c7fbcdc2"]
}, Open  ]]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{
  "change", ",", "G", ",", "downmetric", ",", "upmetric", ",", "christoffel", 
   ",", "christoffelB", ",", "riemann", ",", "riemannB"}], "]"}]], "Input",
 CellChangeTimes->{{3.730960348152565*^9, 3.730960389822296*^9}, {
  3.730960565026269*^9, 3.730960566881948*^9}},
 CellLabel->"In[63]:=",
 CellID->234831011,ExpressionUUID->"232968c4-4f59-4d94-8028-98f21d96af23"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["4.13. Example: Kerr-Newman", "Subsection",
 CellChangeTimes->{{3.7286430157211113`*^9, 3.728643016017661*^9}, {
  3.728673968932828*^9, 3.7286739705964937`*^9}, {3.7286752930712357`*^9, 
  3.728675293199355*^9}, {3.728752965093289*^9, 3.728752968211376*^9}, {
  3.730913541982428*^9, 3.7309135505884933`*^9}, {3.7310047712709923`*^9, 
  3.731004783260107*^9}},
 CellID->590673241,ExpressionUUID->"4f9fcfd4-2025-4a43-a25b-6521aa90dea4"],

Cell["\<\
This section can be evaluated separately from the rest of the notebook.\
\>", "Text",
 CellChangeTimes->{{3.730913577717367*^9, 3.7309136688806963`*^9}, {
  3.7310089507091627`*^9, 3.7310089618753223`*^9}},
 CellID->1973862211,ExpressionUUID->"c9b98765-d104-4605-9a17-34896f964a37"],

Cell[CellGroupData[{

Cell["4.13.1. Setup", "Subsubsection",
 CellChangeTimes->{{3.7286771125390873`*^9, 3.7286771127222157`*^9}, {
  3.730913913338131*^9, 3.730913919728261*^9}, {3.731004803517726*^9, 
  3.731004803724196*^9}},
 CellID->784013520,ExpressionUUID->"8d7ca4fe-498a-449e-b596-a97ad0950afd"],

Cell[CellGroupData[{

Cell["Load xCoba`, if it is not loaded yet:", "MathCaption",
 CellChangeTimes->{
  3.7309135645093603`*^9, {3.730913672392335*^9, 3.730913680200206*^9}},
 CellID->2103056295,ExpressionUUID->"3594f763-a25b-474e-ac20-29034dfd7399"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<xAct`xCoba`\>\"", "]"}]], "Input",
 CellChangeTimes->{{3.7309136378429327`*^9, 3.730913647113517*^9}},
 CellLabel->"In[1]:=",
 CellID->633804092,ExpressionUUID->"d809aff5-3a2e-4325-b050-00508222595f"],

Cell[CellGroupData[{

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.730914316712738*^9, 3.7309149518956003`*^9, 
  3.730957251489663*^9, 3.730957334870701*^9, 3.73095756928409*^9, 
  3.730957688938222*^9, 3.7309577517026033`*^9, 3.730958180599801*^9, 
  3.7309588847543364`*^9, 3.7309599726626987`*^9, 3.7309604109907093`*^9, 
  3.731003121502275*^9, 3.731003282488576*^9, 3.731003927544757*^9, 
  3.731004819081128*^9, 3.731004855024767*^9, 3.7310053688052263`*^9, 
  3.731006601145508*^9, 3.731006672320314*^9, 3.731007288679187*^9, 
  3.731007351436225*^9, 3.731007512119136*^9, 3.7310076030410643`*^9, 
  3.731007650776442*^9, 3.731008144763761*^9, 3.7310086387473297`*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1896641082,ExpressionUUID->"bff068ff-c48c-453a-8c9c-062adbfa862a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`xPerm`  version \"\>", 
   "\[InvisibleSpace]", "\<\"1.2.3\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2015", ",", "8", ",", "23"}], "}"}]}],
  SequenceForm["Package xAct`xPerm`  version ", "1.2.3", ", ", {2015, 8, 23}],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.730914316712738*^9, 3.7309149518956003`*^9, 
  3.730957251489663*^9, 3.730957334870701*^9, 3.73095756928409*^9, 
  3.730957688938222*^9, 3.7309577517026033`*^9, 3.730958180599801*^9, 
  3.7309588847543364`*^9, 3.7309599726626987`*^9, 3.7309604109907093`*^9, 
  3.731003121502275*^9, 3.731003282488576*^9, 3.731003927544757*^9, 
  3.731004819081128*^9, 3.731004855024767*^9, 3.7310053688052263`*^9, 
  3.731006601145508*^9, 3.731006672320314*^9, 3.731007288679187*^9, 
  3.731007351436225*^9, 3.731007512119136*^9, 3.7310076030410643`*^9, 
  3.731007650776442*^9, 3.731008144763761*^9, 3.7310086387645073`*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1746892753,ExpressionUUID->"4ec7aaaf-bd23-4e0d-88d6-2d7a1607f774"],

Cell[BoxData["\<\"CopyRight (C) 2003-2018, Jose M. Martin-Garcia, under the \
General Public License.\"\>"], "Print",
 CellChangeTimes->{3.730914316712738*^9, 3.7309149518956003`*^9, 
  3.730957251489663*^9, 3.730957334870701*^9, 3.73095756928409*^9, 
  3.730957688938222*^9, 3.7309577517026033`*^9, 3.730958180599801*^9, 
  3.7309588847543364`*^9, 3.7309599726626987`*^9, 3.7309604109907093`*^9, 
  3.731003121502275*^9, 3.731003282488576*^9, 3.731003927544757*^9, 
  3.731004819081128*^9, 3.731004855024767*^9, 3.7310053688052263`*^9, 
  3.731006601145508*^9, 3.731006672320314*^9, 3.731007288679187*^9, 
  3.731007351436225*^9, 3.731007512119136*^9, 3.7310076030410643`*^9, 
  3.731007650776442*^9, 3.731008144763761*^9, 3.7310086387820587`*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1777561534,ExpressionUUID->"c0ddb2c8-902a-4bc5-aef4-7155462e7456"],

Cell[BoxData["\<\"Connecting to external mac executable...\"\>"], "Print",
 CellChangeTimes->{3.730914316712738*^9, 3.7309149518956003`*^9, 
  3.730957251489663*^9, 3.730957334870701*^9, 3.73095756928409*^9, 
  3.730957688938222*^9, 3.7309577517026033`*^9, 3.730958180599801*^9, 
  3.7309588847543364`*^9, 3.7309599726626987`*^9, 3.7309604109907093`*^9, 
  3.731003121502275*^9, 3.731003282488576*^9, 3.731003927544757*^9, 
  3.731004819081128*^9, 3.731004855024767*^9, 3.7310053688052263`*^9, 
  3.731006601145508*^9, 3.731006672320314*^9, 3.731007288679187*^9, 
  3.731007351436225*^9, 3.731007512119136*^9, 3.7310076030410643`*^9, 
  3.731007650776442*^9, 3.731008144763761*^9, 3.7310086387938538`*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1831688521,ExpressionUUID->"c12210cf-b647-4922-a2f1-4ef6aeacafd0"],

Cell[BoxData["\<\"Connection established.\"\>"], "Print",
 CellChangeTimes->{3.730914316712738*^9, 3.7309149518956003`*^9, 
  3.730957251489663*^9, 3.730957334870701*^9, 3.73095756928409*^9, 
  3.730957688938222*^9, 3.7309577517026033`*^9, 3.730958180599801*^9, 
  3.7309588847543364`*^9, 3.7309599726626987`*^9, 3.7309604109907093`*^9, 
  3.731003121502275*^9, 3.731003282488576*^9, 3.731003927544757*^9, 
  3.731004819081128*^9, 3.731004855024767*^9, 3.7310053688052263`*^9, 
  3.731006601145508*^9, 3.731006672320314*^9, 3.731007288679187*^9, 
  3.731007351436225*^9, 3.731007512119136*^9, 3.7310076030410643`*^9, 
  3.731007650776442*^9, 3.731008144763761*^9, 3.7310086388078747`*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1010892702,ExpressionUUID->"735948ea-d414-4293-8aa1-06c5e1e480e1"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.730914316712738*^9, 3.7309149518956003`*^9, 
  3.730957251489663*^9, 3.730957334870701*^9, 3.73095756928409*^9, 
  3.730957688938222*^9, 3.7309577517026033`*^9, 3.730958180599801*^9, 
  3.7309588847543364`*^9, 3.7309599726626987`*^9, 3.7309604109907093`*^9, 
  3.731003121502275*^9, 3.731003282488576*^9, 3.731003927544757*^9, 
  3.731004819081128*^9, 3.731004855024767*^9, 3.7310053688052263`*^9, 
  3.731006601145508*^9, 3.731006672320314*^9, 3.731007288679187*^9, 
  3.731007351436225*^9, 3.731007512119136*^9, 3.7310076030410643`*^9, 
  3.731007650776442*^9, 3.731008144763761*^9, 3.731008638820578*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1330511097,ExpressionUUID->"3c99a295-7990-4874-aa1d-0a27e3ea74de"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`xTensor`  version \"\>", 
   "\[InvisibleSpace]", "\<\"1.1.3\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2018", ",", "2", ",", "28"}], "}"}]}],
  SequenceForm[
  "Package xAct`xTensor`  version ", "1.1.3", ", ", {2018, 2, 28}],
  Editable->False]], "Print",
 CellChangeTimes->{3.730914316712738*^9, 3.7309149518956003`*^9, 
  3.730957251489663*^9, 3.730957334870701*^9, 3.73095756928409*^9, 
  3.730957688938222*^9, 3.7309577517026033`*^9, 3.730958180599801*^9, 
  3.7309588847543364`*^9, 3.7309599726626987`*^9, 3.7309604109907093`*^9, 
  3.731003121502275*^9, 3.731003282488576*^9, 3.731003927544757*^9, 
  3.731004819081128*^9, 3.731004855024767*^9, 3.7310053688052263`*^9, 
  3.731006601145508*^9, 3.731006672320314*^9, 3.731007288679187*^9, 
  3.731007351436225*^9, 3.731007512119136*^9, 3.7310076030410643`*^9, 
  3.731007650776442*^9, 3.731008144763761*^9, 3.731008638834755*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->152791068,ExpressionUUID->"347e17a4-bbcd-4382-a55e-a4d2f56c41bb"],

Cell[BoxData["\<\"CopyRight (C) 2002-2018, Jose M. Martin-Garcia, under the \
General Public License.\"\>"], "Print",
 CellChangeTimes->{3.730914316712738*^9, 3.7309149518956003`*^9, 
  3.730957251489663*^9, 3.730957334870701*^9, 3.73095756928409*^9, 
  3.730957688938222*^9, 3.7309577517026033`*^9, 3.730958180599801*^9, 
  3.7309588847543364`*^9, 3.7309599726626987`*^9, 3.7309604109907093`*^9, 
  3.731003121502275*^9, 3.731003282488576*^9, 3.731003927544757*^9, 
  3.731004819081128*^9, 3.731004855024767*^9, 3.7310053688052263`*^9, 
  3.731006601145508*^9, 3.731006672320314*^9, 3.731007288679187*^9, 
  3.731007351436225*^9, 3.731007512119136*^9, 3.7310076030410643`*^9, 
  3.731007650776442*^9, 3.731008144763761*^9, 3.73100863884757*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1716521711,ExpressionUUID->"648eec4e-d467-4913-aa9f-c5ca610830f2"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.730914316712738*^9, 3.7309149518956003`*^9, 
  3.730957251489663*^9, 3.730957334870701*^9, 3.73095756928409*^9, 
  3.730957688938222*^9, 3.7309577517026033`*^9, 3.730958180599801*^9, 
  3.7309588847543364`*^9, 3.7309599726626987`*^9, 3.7309604109907093`*^9, 
  3.731003121502275*^9, 3.731003282488576*^9, 3.731003927544757*^9, 
  3.731004819081128*^9, 3.731004855024767*^9, 3.7310053688052263`*^9, 
  3.731006601145508*^9, 3.731006672320314*^9, 3.731007288679187*^9, 
  3.731007351436225*^9, 3.731007512119136*^9, 3.7310076030410643`*^9, 
  3.731007650776442*^9, 3.731008144763761*^9, 3.731008638943099*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->285082979,ExpressionUUID->"4d5e1b62-3f1d-4333-8d8c-9d79f4926a24"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`xCoba`  version \"\>", 
   "\[InvisibleSpace]", "\<\"0.8.4\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2018", ",", "2", ",", "28"}], "}"}]}],
  SequenceForm["Package xAct`xCoba`  version ", "0.8.4", ", ", {2018, 2, 28}],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.730914316712738*^9, 3.7309149518956003`*^9, 
  3.730957251489663*^9, 3.730957334870701*^9, 3.73095756928409*^9, 
  3.730957688938222*^9, 3.7309577517026033`*^9, 3.730958180599801*^9, 
  3.7309588847543364`*^9, 3.7309599726626987`*^9, 3.7309604109907093`*^9, 
  3.731003121502275*^9, 3.731003282488576*^9, 3.731003927544757*^9, 
  3.731004819081128*^9, 3.731004855024767*^9, 3.7310053688052263`*^9, 
  3.731006601145508*^9, 3.731006672320314*^9, 3.731007288679187*^9, 
  3.731007351436225*^9, 3.731007512119136*^9, 3.7310076030410643`*^9, 
  3.731007650776442*^9, 3.731008144763761*^9, 3.731008638956394*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->353611596,ExpressionUUID->"39cd9dd9-55a6-4a5b-9cb0-bc82959b243b"],

Cell[BoxData["\<\"CopyRight (C) 2005-2018, David Yllanes and Jose M. \
Martin-Garcia, under the General Public License.\"\>"], "Print",
 CellChangeTimes->{3.730914316712738*^9, 3.7309149518956003`*^9, 
  3.730957251489663*^9, 3.730957334870701*^9, 3.73095756928409*^9, 
  3.730957688938222*^9, 3.7309577517026033`*^9, 3.730958180599801*^9, 
  3.7309588847543364`*^9, 3.7309599726626987`*^9, 3.7309604109907093`*^9, 
  3.731003121502275*^9, 3.731003282488576*^9, 3.731003927544757*^9, 
  3.731004819081128*^9, 3.731004855024767*^9, 3.7310053688052263`*^9, 
  3.731006601145508*^9, 3.731006672320314*^9, 3.731007288679187*^9, 
  3.731007351436225*^9, 3.731007512119136*^9, 3.7310076030410643`*^9, 
  3.731007650776442*^9, 3.731008144763761*^9, 3.731008638968711*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1062842723,ExpressionUUID->"4b1fc357-c0b0-4e37-92a3-63cd47b38790"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.730914316712738*^9, 3.7309149518956003`*^9, 
  3.730957251489663*^9, 3.730957334870701*^9, 3.73095756928409*^9, 
  3.730957688938222*^9, 3.7309577517026033`*^9, 3.730958180599801*^9, 
  3.7309588847543364`*^9, 3.7309599726626987`*^9, 3.7309604109907093`*^9, 
  3.731003121502275*^9, 3.731003282488576*^9, 3.731003927544757*^9, 
  3.731004819081128*^9, 3.731004855024767*^9, 3.7310053688052263`*^9, 
  3.731006601145508*^9, 3.731006672320314*^9, 3.731007288679187*^9, 
  3.731007351436225*^9, 3.731007512119136*^9, 3.7310076030410643`*^9, 
  3.731007650776442*^9, 3.731008144763761*^9, 3.731008638981106*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->428669715,ExpressionUUID->"20a4312b-6db2-42b2-b4e4-6ebe13105a04"],

Cell[BoxData["\<\"These packages come with ABSOLUTELY NO WARRANTY; for \
details type Disclaimer[]. This is free software, and you are welcome to \
redistribute it under certain conditions. See the General Public License for \
details.\"\>"], "Print",
 CellChangeTimes->{3.730914316712738*^9, 3.7309149518956003`*^9, 
  3.730957251489663*^9, 3.730957334870701*^9, 3.73095756928409*^9, 
  3.730957688938222*^9, 3.7309577517026033`*^9, 3.730958180599801*^9, 
  3.7309588847543364`*^9, 3.7309599726626987`*^9, 3.7309604109907093`*^9, 
  3.731003121502275*^9, 3.731003282488576*^9, 3.731003927544757*^9, 
  3.731004819081128*^9, 3.731004855024767*^9, 3.7310053688052263`*^9, 
  3.731006601145508*^9, 3.731006672320314*^9, 3.731007288679187*^9, 
  3.731007351436225*^9, 3.731007512119136*^9, 3.7310076030410643`*^9, 
  3.731007650776442*^9, 3.731008144763761*^9, 3.731008638994403*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1991747967,ExpressionUUID->"cfc266a7-e276-453f-b9d6-5730d55009cc"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{3.730914316712738*^9, 3.7309149518956003`*^9, 
  3.730957251489663*^9, 3.730957334870701*^9, 3.73095756928409*^9, 
  3.730957688938222*^9, 3.7309577517026033`*^9, 3.730958180599801*^9, 
  3.7309588847543364`*^9, 3.7309599726626987`*^9, 3.7309604109907093`*^9, 
  3.731003121502275*^9, 3.731003282488576*^9, 3.731003927544757*^9, 
  3.731004819081128*^9, 3.731004855024767*^9, 3.7310053688052263`*^9, 
  3.731006601145508*^9, 3.731006672320314*^9, 3.731007288679187*^9, 
  3.731007351436225*^9, 3.731007512119136*^9, 3.7310076030410643`*^9, 
  3.731007650776442*^9, 3.731008144763761*^9, 3.731008639007503*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->332991933,ExpressionUUID->"25ef8328-7041-4e1f-b2f2-c6cec6998ee0"]
}, Open  ]]
}, Open  ]],

Cell[BoxData[{
 RowBox[{
  RowBox[{"$DefInfoQ", "=", "False"}], ";"}], "\n", 
 RowBox[{
  RowBox[{"$CVVerbose", "=", "False"}], ";"}]}], "Input",
 CellChangeTimes->{{3.458061670223096*^9, 3.4580616749787188`*^9}, 
   3.458280552972406*^9, {3.5081452631080103`*^9, 3.508145265389002*^9}},
 CellLabel->"In[2]:=",
 CellID->2006047368,ExpressionUUID->"012040cb-213b-485f-a513-78a5817b41d3"],

Cell[BoxData[
 RowBox[{
  RowBox[{"$PrePrint", "=", "ScreenDollarIndices"}], ";"}]], "Input",
 CellChangeTimes->{{3.730914497190155*^9, 3.73091450386228*^9}},
 CellLabel->"In[4]:=",
 CellID->120754463,ExpressionUUID->"db4fc597-7708-4b87-aa0c-76fba75fe1f0"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Define 4d manifold with Lorentzian metric:", "MathCaption",
 CellID->148213226,ExpressionUUID->"2e8e88be-359f-4d55-8b5b-9b672dd86ce9"],

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M4", ",", "4", ",", 
   RowBox[{"IndexRange", "[", 
    RowBox[{"a", ",", "p"}], "]"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.730913717183175*^9, 3.7309137173431253`*^9}},
 CellLabel->"In[5]:=",
 CellID->2086142509,ExpressionUUID->"a9bdc773-311a-4646-b41f-95edfd0a5a71"],

Cell[BoxData[
 RowBox[{"DefMetric", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"3", ",", "1", ",", "0"}], "}"}], ",", 
   RowBox[{"metricg", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", "CD", ",", 
   RowBox[{"PrintAs", "->", "\"\<g\>\""}]}], "]"}]], "Input",
 CellChangeTimes->{{3.730913729384461*^9, 3.730913731039002*^9}},
 CellLabel->"In[6]:=",
 CellID->1412400003,ExpressionUUID->"d0219ec1-54cf-4f51-8080-ba6b11f8309b"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Declare the Boyer-Lindquist chart and their associated frame:\
\>", "MathCaption",
 CellChangeTimes->{
  3.7309137553177547`*^9, {3.731004876881217*^9, 3.731004885297583*^9}},
 CellID->1238377332,ExpressionUUID->"81095bee-25c6-48ca-835c-c4a7665d96e8"],

Cell[BoxData[
 RowBox[{"DefChart", "[", 
  RowBox[{"B", ",", "M4", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "2", ",", "3"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"t", "[", "]"}], ",", 
     RowBox[{"r", "[", "]"}], ",", 
     RowBox[{"\[Theta]", "[", "]"}], ",", 
     RowBox[{"\[Phi]", "[", "]"}]}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.731004892243355*^9, 3.731004896763291*^9}},
 CellLabel->"In[7]:=",
 CellID->2143795847,ExpressionUUID->"6c7b9365-7476-4b38-ba05-7824c772c60c"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Define the Kerr-Newman constants:", "MathCaption",
 CellChangeTimes->{{3.7310049119464684`*^9, 3.731004921320256*^9}},
 CellID->1350532130,ExpressionUUID->"9a87f3d3-536f-4f58-8eac-04f82676e28b"],

Cell[BoxData[{
 RowBox[{"DefConstantSymbol", "[", 
  RowBox[{"Mass", ",", 
   RowBox[{"PrintAs", "\[Rule]", "\"\<M\>\""}]}], "]"}], "\n", 
 RowBox[{"DefConstantSymbol", "[", 
  RowBox[{"Rotation", ",", 
   RowBox[{"PrintAs", "\[Rule]", "\"\<a\>\""}]}], "]"}], "\n", 
 RowBox[{"DefConstantSymbol", "[", 
  RowBox[{"Charge", ",", 
   RowBox[{"PrintAs", "\[Rule]", "\"\<Q\>\""}]}], "]"}]}], "Input",
 CellChangeTimes->{{3.458062300336478*^9, 3.458062325157645*^9}, 
   3.458311328980466*^9},
 CellLabel->"In[8]:=",
 CellID->1985924282,ExpressionUUID->"5fe73ad1-9924-4f5d-b43c-0b8ddeacab07"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The covariant derivative of the elements of the basis will be described by \
this Christoffel tensor:\
\>", "MathCaption",
 CellChangeTimes->{{3.73091378324617*^9, 3.730913827731214*^9}},
 CellID->479736035,ExpressionUUID->"3859844e-7fd5-4bcf-9860-f9244333a45f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CD", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"Basis", "[", 
   RowBox[{"b", ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "B"}]}], "}"}]}], "]"}], "]"}]], "Input",
 CellChangeTimes->{3.7310049503513737`*^9},
 CellLabel->"In[11]:=",
 CellID->1269929407,ExpressionUUID->"eadba501-79ae-4805-88b1-6d1e20d0792f"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"\[CapitalGamma]", "[", 
       RowBox[{"\[EmptyDownTriangle]", ",", 
        StyleBox["\[ScriptCapitalD]",
         FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
      StyleBox[GridBox[{
         {"b", " ", " "},
         {" ", "a", 
          StyleBox["c",
           FontColor->RGBColor[1., 0., 0.]]}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`ChristoffelCDPDB[$CellContext`b, -$CellContext`a, \
{-$CellContext`c, -$CellContext`B}],
  Editable->False]], "Output",
 CellChangeTimes->{3.730913824189146*^9, 3.730914325809108*^9, 
  3.730914953547511*^9, 3.7309572532761183`*^9, 3.730957339154127*^9, 
  3.730957573302845*^9, 3.7309576923503313`*^9, 3.730957755759489*^9, 
  3.7309581841164207`*^9, 3.730958888303697*^9, 3.730959975995798*^9, 
  3.730960414370409*^9, 3.7310031251572323`*^9, 3.731003286134523*^9, 
  3.731003931226742*^9, 3.731004954106793*^9, 3.731005372157915*^9, 
  3.731006602460855*^9, 3.7310066758368473`*^9, 3.731007290199917*^9, 
  3.731007354877017*^9, 3.731007515687945*^9, 3.731007606463485*^9, 
  3.731007654198999*^9, 3.73100816549592*^9, 3.7310086487086077`*^9},
 CellLabel->"Out[11]=",
 CellID->1625239968,ExpressionUUID->"9bb51acd-53fa-4bf2-a129-95c2c0355400"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["4.13.2. Input metric", "Subsubsection",
 CellChangeTimes->{{3.7286771125390873`*^9, 3.7286771127222157`*^9}, {
  3.730913913338131*^9, 3.730913941103621*^9}, {3.731004967632353*^9, 
  3.7310049678228483`*^9}},
 CellID->1899572791,ExpressionUUID->"9f8bffcf-9bfe-4428-a0ee-4c58a7237e26"],

Cell["\<\
Specify the values of the metric components in the coordinate basis just \
defined. We will store those components as tensor values for the metricg \
tensor.\
\>", "Text",
 CellChangeTimes->{{3.730913577717367*^9, 3.7309136688806963`*^9}, {
  3.7310031751218443`*^9, 3.731003229439324*^9}},
 CellID->1628002035,ExpressionUUID->"81b1532c-b33a-4242-9bf4-38944101df69"],

Cell[CellGroupData[{

Cell["\<\
This is the matrix of scalar components of the Kerr-Newman metric in the B \
coordinate basis:\
\>", "MathCaption",
 CellChangeTimes->{{3.4582213740661182`*^9, 3.458221395832098*^9}, {
  3.731004998964964*^9, 3.731005008045374*^9}},
 CellID->360157534,ExpressionUUID->"309f6f49-7cf0-4cbf-b167-6f9131046394"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Delta", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"r", "[", "]"}], "^", "2"}], "+", 
    RowBox[{"Rotation", "^", "2"}], "+", 
    RowBox[{"Charge", "^", "2"}], "-", 
    RowBox[{"2", "Mass", " ", 
     RowBox[{"r", "[", "]"}]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Sigma", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"r", "[", "]"}], "^", "2"}], "+", 
    RowBox[{
     RowBox[{"Rotation", "^", "2"}], 
     RowBox[{
      RowBox[{"Cos", "[", 
       RowBox[{"\[Theta]", "[", "]"}], "]"}], "^", "2"}]}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"MatrixForm", "[", "\[IndentingNewLine]", 
  RowBox[{"matrix", "=", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         RowBox[{"(", 
          RowBox[{"Delta", "-", 
           RowBox[{
            RowBox[{"Rotation", "^", "2"}], 
            RowBox[{
             RowBox[{"Sin", "[", 
              RowBox[{"\[Theta]", "[", "]"}], "]"}], "^", "2"}]}]}], ")"}]}], 
        "/", "Sigma"}], ",", "0", ",", "0", ",", 
       RowBox[{
        RowBox[{"-", "Rotation"}], " ", 
        RowBox[{
         RowBox[{"Sin", "[", 
          RowBox[{"\[Theta]", "[", "]"}], "]"}], "^", "2"}], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"r", "[", "]"}], "^", "2"}], "+", 
           RowBox[{"Rotation", "^", "2"}], "-", "Delta"}], ")"}], "/", 
         "Sigma"}]}]}], "}"}], ",", "\[IndentingNewLine]", 
     RowBox[{"{", 
      RowBox[{"0", ",", 
       RowBox[{"Sigma", "/", "Delta"}], ",", "0", ",", "0"}], "}"}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", "Sigma", ",", "0"}], "}"}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "Rotation"}], " ", 
        RowBox[{
         RowBox[{"Sin", "[", 
          RowBox[{"\[Theta]", "[", "]"}], "]"}], "^", "2"}], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"r", "[", "]"}], "^", "2"}], "+", 
           RowBox[{"Rotation", "^", "2"}], "-", "Delta"}], ")"}], "/", 
         "Sigma"}]}], ",", "0", ",", "0", ",", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"r", "[", "]"}], "^", "2"}], "+", 
              RowBox[{"Rotation", "^", "2"}]}], ")"}], "^", "2"}], "-", 
           RowBox[{"Delta", " ", 
            RowBox[{"Rotation", "^", "2"}], 
            RowBox[{
             RowBox[{"Sin", "[", 
              RowBox[{"\[Theta]", "[", "]"}], "]"}], "^", "2"}]}]}], ")"}], 
         "/", "Sigma"}], " ", 
        RowBox[{
         RowBox[{"Sin", "[", 
          RowBox[{"\[Theta]", "[", "]"}], "]"}], "^", "2"}]}]}], "}"}]}], 
    "\[IndentingNewLine]", "}"}]}], "\[IndentingNewLine]", "]"}]}], "Input",
 CellChangeTimes->{{3.458209471423118*^9, 3.458209514619155*^9}, {
  3.4582099113745127`*^9, 3.4582099515188932`*^9}, {3.458224763913744*^9, 
  3.458224767223271*^9}, {3.4582366252815*^9, 3.4582366418126383`*^9}, {
  3.731005014405198*^9, 3.7310050212378397`*^9}},
 CellLabel->"In[12]:=",
 CellID->1558756692,ExpressionUUID->"68166242-18f5-48f2-83f1-79a9bbc327d5"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      FractionBox[
       RowBox[{
        RowBox[{"-", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["Q",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Charge,
           Editable->False], "2"]}], "-", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], "2"], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Mass,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]}], "-", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"], "+", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Sin",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Sin,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}]}], 
       RowBox[{
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Cos",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Cos,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}], "+", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]}]], "0", "0", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["Q",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Charge,
              Editable->False], "2"]}], "+", 
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Mass,
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]}]}], ")"}], " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Sin",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Sin,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}], 
        RowBox[{
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "2"]}], "+", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"]}]]}]},
     {"0", 
      FractionBox[
       RowBox[{
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Cos",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Cos,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}], "+", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]}], 
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["Q",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Charge,
          Editable->False], "2"], "+", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], "2"], "-", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Mass,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]}], "+", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]}]], "0", "0"},
     {"0", "0", 
      RowBox[{
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], "2"], " ", 
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Cos",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Cos,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"]}], "+", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"]}], "0"},
     {
      RowBox[{"-", 
       FractionBox[
        RowBox[{
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["Q",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Charge,
              Editable->False], "2"]}], "+", 
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Mass,
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]}]}], ")"}], " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Sin",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Sin,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}], 
        RowBox[{
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "2"]}], "+", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"]}]]}], "0", "0", 
      FractionBox[
       RowBox[{
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Sin",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Sin,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"], "+", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}], ")"}], "2"], "-", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["Q",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Charge,
               Editable->False], "2"], "+", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"], "-", 
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False], " ", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}], "+", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}], ")"}], " ", 
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["Sin",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Sin,
              Editable->False], "[", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False], "]"}], "2"]}]}], ")"}]}], 
       RowBox[{
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Cos",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Cos,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}], "+", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]}]]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{
  3.731005030373361*^9, 3.7310053722883387`*^9, 3.7310066026331778`*^9, 
   3.7310066759629707`*^9, 3.73100729487096*^9, {3.73100733987294*^9, 
   3.731007355146669*^9}, 3.731007515788048*^9, 3.7310076065926657`*^9, 
   3.731007654308161*^9, 3.731008171647073*^9, 3.7310086559817266`*^9},
 CellLabel->"Out[14]//MatrixForm=",
 CellID->1427830721,ExpressionUUID->"1fad0e2f-8890-4eaa-9032-3f0b2c42a934"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Set those scalars as tensor values for the metricg tensor:", \
"MathCaption",
 CellChangeTimes->{{3.730914346681881*^9, 3.73091437470514*^9}, {
   3.731003299486094*^9, 3.731003345756345*^9}, 3.7310050503498087`*^9},
 CellID->408762261,ExpressionUUID->"2d79acda-5f04-4d81-846e-17c33809bdfe"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"MetricInBasis", "[", 
  RowBox[{"metricg", ",", 
   RowBox[{"-", "B"}], ",", "matrix"}], "]"}]], "Input",
 CellChangeTimes->{{3.731003151843067*^9, 3.731003153962832*^9}, {
  3.731005053444412*^9, 3.731005054012368*^9}},
 CellLabel->"In[15]:=",
 CellID->267133874,ExpressionUUID->"54106a6f-11e0-468a-892a-1022bba25642"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{0, -$CellContext`B}, {0, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["Q",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Charge,
           Editable->False], "2"], "+", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], "-", 
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Mass,
           Editable->False], " ", 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False]}], "+", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"], "-", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Sin",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Sin,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "2"]}]}], 
        RowBox[{
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "2"]}], "+", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"]}]]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{0, -$CellContext`B}, {1, -$CellContext`B}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{0, -$CellContext`B}, {2, -$CellContext`B}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{0, -$CellContext`B}, {3, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      FractionBox[
       RowBox[{
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], " ", 
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["Q",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Charge,
            Editable->False], "2"], "-", 
          RowBox[{"2", " ", 
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False], " ", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]}]}], ")"}], " ", 
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Sin",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Sin,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"]}], 
       RowBox[{
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Cos",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Cos,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}], "+", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]}]]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{1, -$CellContext`B}, {0, -$CellContext`B}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{1, -$CellContext`B}, {1, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      FractionBox[
       RowBox[{
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Cos",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Cos,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}], "+", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]}], 
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["Q",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Charge,
          Editable->False], "2"], "+", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], "2"], "-", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Mass,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]}], "+", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]}]]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{1, -$CellContext`B}, {2, -$CellContext`B}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{1, -$CellContext`B}, {3, -$CellContext`B}],
       Editable->False], "\[Rule]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{2, -$CellContext`B}, {0, -$CellContext`B}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{2, -$CellContext`B}, {1, -$CellContext`B}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{2, -$CellContext`B}, {2, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], "2"], " ", 
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Cos",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Cos,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"]}], "+", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{2, -$CellContext`B}, {3, -$CellContext`B}],
       Editable->False], "\[Rule]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{3, -$CellContext`B}, {0, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      FractionBox[
       RowBox[{
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], " ", 
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["Q",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Charge,
            Editable->False], "2"], "-", 
          RowBox[{"2", " ", 
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False], " ", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]}]}], ")"}], " ", 
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Sin",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Sin,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"]}], 
       RowBox[{
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Cos",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Cos,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}], "+", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]}]]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{3, -$CellContext`B}, {1, -$CellContext`B}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{3, -$CellContext`B}, {2, -$CellContext`B}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{3, -$CellContext`B}, {3, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      FractionBox[
       RowBox[{
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Sin",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Sin,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"], "+", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}], ")"}], "2"], "-", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["Q",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Charge,
               Editable->False], "2"], "+", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"], "-", 
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False], " ", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}], "+", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}], ")"}], " ", 
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["Sin",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Sin,
              Editable->False], "[", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False], "]"}], "2"]}]}], ")"}]}], 
       RowBox[{
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Cos",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Cos,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}], "+", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]}]]}]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{
  3.730913971648592*^9, 3.730914340386692*^9, 3.730914957591959*^9, 
   3.7309573400721817`*^9, 3.730957574454195*^9, 3.730957693665423*^9, 
   3.730957757161455*^9, 3.730958185799819*^9, 3.730958889937704*^9, 
   3.730959977037339*^9, 3.730960415668604*^9, 3.7310033503705893`*^9, 
   3.731003932320017*^9, 3.731005055126691*^9, 3.731005372559147*^9, 
   3.73100660458666*^9, 3.731006676234416*^9, 3.731007295182538*^9, {
   3.731007340022509*^9, 3.7310073554204493`*^9}, 3.731007516070697*^9, 
   3.73100760687929*^9, 3.731007654593628*^9, 3.731008173154725*^9, 
   3.731008656361755*^9},
 CellLabel->"Out[15]=",
 CellID->1568656269,ExpressionUUID->"c722a884-4c0c-49ad-8e0c-35db5f0106c5"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
This is the object stored internally, that separates values due to symmetry \
from truly independent values:\
\>", "MathCaption",
 CellChangeTimes->{{3.731003378748698*^9, 3.731003405794421*^9}},
 CellID->655749584,ExpressionUUID->"5b98dbdb-8323-4b5c-b2bd-7f48d03acd9a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TensorValues", "[", "metricg", "]"}]], "Input",
 CellChangeTimes->{{3.731003367796548*^9, 3.731003372515414*^9}},
 CellLabel->"In[16]:=",
 CellID->297181884,ExpressionUUID->"7d0367da-d792-44f0-ba7b-4737637c0d41"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{1, -$CellContext`B}, {0, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{0, -$CellContext`B}, {1, -$CellContext`B}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{2, -$CellContext`B}, {0, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{0, -$CellContext`B}, {2, -$CellContext`B}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{2, -$CellContext`B}, {1, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{1, -$CellContext`B}, {2, -$CellContext`B}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{3, -$CellContext`B}, {0, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{0, -$CellContext`B}, {3, -$CellContext`B}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{3, -$CellContext`B}, {1, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{1, -$CellContext`B}, {3, -$CellContext`B}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{3, -$CellContext`B}, {2, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{2, -$CellContext`B}, {3, -$CellContext`B}],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{0, -$CellContext`B}, {0, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["Q",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Charge,
           Editable->False], "2"], "+", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], "-", 
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Mass,
           Editable->False], " ", 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False]}], "+", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"], "-", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Sin",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Sin,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "2"]}]}], 
        RowBox[{
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "2"]}], "+", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"]}]]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{0, -$CellContext`B}, {1, -$CellContext`B}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{0, -$CellContext`B}, {2, -$CellContext`B}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{0, -$CellContext`B}, {3, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      FractionBox[
       RowBox[{
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], " ", 
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["Q",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Charge,
            Editable->False], "2"], "-", 
          RowBox[{"2", " ", 
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False], " ", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]}]}], ")"}], " ", 
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Sin",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Sin,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"]}], 
       RowBox[{
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Cos",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Cos,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}], "+", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]}]]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{1, -$CellContext`B}, {1, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      FractionBox[
       RowBox[{
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Cos",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Cos,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}], "+", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]}], 
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["Q",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Charge,
          Editable->False], "2"], "+", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], "2"], "-", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["M",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Mass,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False]}], "+", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]}]]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{1, -$CellContext`B}, {2, -$CellContext`B}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{1, -$CellContext`B}, {3, -$CellContext`B}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{2, -$CellContext`B}, {2, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], "2"], " ", 
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Cos",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Cos,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"]}], "+", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{2, -$CellContext`B}, {3, -$CellContext`B}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[{3, -$CellContext`B}, {3, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      FractionBox[
       RowBox[{
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Sin",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Sin,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"], "+", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}], ")"}], "2"], "-", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["Q",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Charge,
               Editable->False], "2"], "+", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"], "-", 
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False], " ", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}], "+", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}], ")"}], " ", 
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["Sin",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Sin,
              Editable->False], "[", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False], "]"}], "2"]}]}], ")"}]}], 
       RowBox[{
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Cos",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Cos,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}], "+", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]}]]}]}], "}"}]}], "]"}]], "Output",
 CellChangeTimes->{
  3.731003372976207*^9, 3.7310039325920353`*^9, 3.731005067066492*^9, 
   3.731005372697289*^9, 3.731006604751392*^9, 3.731006676363699*^9, 
   3.731007295352789*^9, {3.73100734015522*^9, 3.7310073555709257`*^9}, 
   3.731007516214222*^9, 3.731007607027974*^9, 3.73100765476688*^9, 
   3.7310081746163073`*^9, 3.7310086565619802`*^9},
 CellLabel->"Out[16]=",
 CellID->1896662511,ExpressionUUID->"1045434f-633e-43f5-a2b0-30e40e310dcc"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["In terms of a basis the metric takes the form:", "MathCaption",
 CellChangeTimes->{
  3.7309140688767357`*^9, {3.73100341370597*^9, 3.7310034149060802`*^9}},
 CellID->1498039507,ExpressionUUID->"a9ce3e45-fbf3-4450-a5e3-b89aad94e014"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TraceBasisDummy", "[", 
  RowBox[{
   RowBox[{"metricg", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "B"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "B"}]}], "}"}]}], "]"}], 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"-", "c"}], ",", 
     RowBox[{"{", 
      RowBox[{"a", ",", "B"}], "}"}]}], "]"}], 
   RowBox[{"Basis", "[", 
    RowBox[{
     RowBox[{"-", "d"}], ",", 
     RowBox[{"{", 
      RowBox[{"b", ",", "B"}], "}"}]}], "]"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.731005087405284*^9, 3.7310050918105183`*^9}},
 CellLabel->"In[17]:=",
 CellID->451716721,ExpressionUUID->"b9fa1e5f-cdc1-4767-a6b4-15773cdd297f"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {0, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{0, -$CellContext`B}, {0, -$CellContext`B}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {1, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{0, -$CellContext`B}, {1, -$CellContext`B}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {2, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{0, -$CellContext`B}, {2, -$CellContext`B}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {3, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{0, -$CellContext`B}, {3, -$CellContext`B}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {0, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{1, -$CellContext`B}, {0, -$CellContext`B}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {1, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{1, -$CellContext`B}, {1, -$CellContext`B}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {2, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{1, -$CellContext`B}, {2, -$CellContext`B}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {3, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{1, -$CellContext`B}, {3, -$CellContext`B}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {0, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{2, -$CellContext`B}, {0, -$CellContext`B}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {1, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{2, -$CellContext`B}, {1, -$CellContext`B}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {2, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{2, -$CellContext`B}, {2, -$CellContext`B}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {3, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{2, -$CellContext`B}, {3, -$CellContext`B}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {0, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["0",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{3, -$CellContext`B}, {0, -$CellContext`B}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {1, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["1",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{3, -$CellContext`B}, {1, -$CellContext`B}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {2, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{3, -$CellContext`B}, {2, -$CellContext`B}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {3, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["3",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{3, -$CellContext`B}, {3, -$CellContext`B}],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{
  3.730913992199342*^9, {3.7309140266490192`*^9, 3.730914041673592*^9}, 
   3.730914105808708*^9, 3.730914385078683*^9, 3.730914957773316*^9, 
   3.7309573402454042`*^9, 3.7309575746510277`*^9, 3.730957693902625*^9, 
   3.730957757321456*^9, 3.7309581859934807`*^9, 3.730958890357711*^9, 
   3.730959977228107*^9, 3.730960415844363*^9, 3.7310034180033817`*^9, 
   3.7310039327473392`*^9, 3.7310050940557537`*^9, 3.7310053728374777`*^9, 
   3.731006604915716*^9, 3.731006676509124*^9, 3.7310072955636463`*^9, {
   3.7310073402722483`*^9, 3.7310073557456427`*^9}, 3.731007516421929*^9, 
   3.731007607293427*^9, 3.7310076550010767`*^9, 3.731008175972864*^9, 
   3.731008656803891*^9},
 CellLabel->"Out[17]=",
 CellID->2107049956,ExpressionUUID->"ac629944-5638-4c52-a08a-fd37918f6939"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", 
  RowBox[{"TensorValues", "[", "metricg", "]"}]}]], "Input",
 CellChangeTimes->{{3.7309140280375433`*^9, 3.730914038834407*^9}, 
   3.730914103218519*^9, 3.731005099042862*^9},
 CellLabel->"In[18]:=",
 CellID->746516283,ExpressionUUID->"68590c4b-e27f-4bde-a7ab-4a5f8d931941"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`c, {2, $CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"e", 
        StyleBox[GridBox[{
           {" ", 
            StyleBox["2",
             FontColor->RGBColor[1., 0., 0.]]},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Basis[-$CellContext`d, {2, $CellContext`B}],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["a",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Rotation,
        Editable->False], "2"], " ", 
      SuperscriptBox[
       RowBox[{
        InterpretationBox[
         StyleBox["Cos",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Cos,
         Editable->False], "[", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False], "]"}], "2"]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]}], ")"}]}], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {1, $CellContext`B}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]},
            {"d", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`d, {1, $CellContext`B}],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], "2"], " ", 
       SuperscriptBox[
        RowBox[{
         InterpretationBox[
          StyleBox["Cos",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Cos,
          Editable->False], "[", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False], "]"}], "2"]}], "+", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"]}], ")"}]}], 
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["Q",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Charge,
      Editable->False], "2"], "+", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["a",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Rotation,
      Editable->False], "2"], "-", 
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox["M",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Mass,
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False]}], "+", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["a",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Rotation,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`B}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"d", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`d, {0, $CellContext`B}],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["Q",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Charge,
        Editable->False], "2"], "-", 
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Mass,
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}]}], ")"}], " ", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"]}], 
   RowBox[{
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["a",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Rotation,
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Cos",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Cos,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}], "+", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox["a",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Rotation,
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`B}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"d", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`d, {3, $CellContext`B}],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["Q",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Charge,
        Editable->False], "2"], "-", 
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Mass,
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}]}], ")"}], " ", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"]}], 
   RowBox[{
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["a",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Rotation,
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Cos",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Cos,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}], "+", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "-", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {0, $CellContext`B}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]},
            {"d", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`d, {0, $CellContext`B}],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["Q",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Charge,
        Editable->False], "2"], "+", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["a",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Rotation,
        Editable->False], "2"], "-", 
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Mass,
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], "+", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"], "-", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], "2"], " ", 
       SuperscriptBox[
        RowBox[{
         InterpretationBox[
          StyleBox["Sin",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Sin,
          Editable->False], "[", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False], "]"}], "2"]}]}], ")"}]}], 
   RowBox[{
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["a",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Rotation,
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Cos",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Cos,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}], "+", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]], "+", 
  FractionBox[
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`c, {3, $CellContext`B}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[1., 0., 0.]]},
            {"d", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`d, {3, $CellContext`B}],
     Editable->False], " ", 
    SuperscriptBox[
     RowBox[{
      InterpretationBox[
       StyleBox["Sin",
        ShowAutoStyles->False,
        AutoSpacing->False],
       Sin,
       Editable->False], "[", 
      InterpretationBox[
       StyleBox["\[Theta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Theta][],
       Editable->False], "]"}], "2"], " ", 
    RowBox[{"(", 
     RowBox[{
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], "+", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"]}], ")"}], "2"], "-", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], "2"], " ", 
       RowBox[{"(", 
        RowBox[{
         SuperscriptBox[
          InterpretationBox[
           StyleBox["Q",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Charge,
           Editable->False], "2"], "+", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], "-", 
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Mass,
           Editable->False], " ", 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False]}], "+", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "2"]}], ")"}], " ", 
       SuperscriptBox[
        RowBox[{
         InterpretationBox[
          StyleBox["Sin",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Sin,
          Editable->False], "[", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False], "]"}], "2"]}]}], ")"}]}], 
   RowBox[{
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["a",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Rotation,
       Editable->False], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       InterpretationBox[
        StyleBox["Cos",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Cos,
        Editable->False], "[", 
       InterpretationBox[
        StyleBox["\[Theta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Theta][],
        Editable->False], "]"}], "2"]}], "+", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]]}]], "Output",
 CellChangeTimes->{
  3.730914109323456*^9, 3.730914385978706*^9, 3.730914957879756*^9, 
   3.73095734037306*^9, 3.730957574779385*^9, 3.730957694015346*^9, 
   3.7309577574248133`*^9, 3.7309581860979023`*^9, 3.730958890501206*^9, 
   3.730959977330749*^9, 3.730960415953459*^9, 3.731003419941785*^9, 
   3.731003932827561*^9, 3.731005100231434*^9, 3.731005372972686*^9, 
   3.731006605050823*^9, 3.731006676703875*^9, 3.7310072961240664`*^9, {
   3.731007340413926*^9, 3.731007355887594*^9}, 3.7310075165481443`*^9, 
   3.73100760751134*^9, 3.731007655129301*^9, 3.731008176773849*^9, 
   3.731008656948326*^9},
 CellLabel->"Out[18]=",
 CellID->1183271462,ExpressionUUID->"532d6bde-8954-464c-bb1c-17545c135edb"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["4.13.3. Compute metric-related tensors", "Subsubsection",
 CellChangeTimes->{{3.7286771125390873`*^9, 3.7286771127222157`*^9}, {
  3.730913913338131*^9, 3.730913941103621*^9}, {3.731004967632353*^9, 
  3.7310049678228483`*^9}, {3.7310051409369698`*^9, 3.731005159992453*^9}},
 CellID->424015098,ExpressionUUID->"d514aac7-9a8d-4000-af71-aac01ae83808"],

Cell["\<\
Rather than following the flexible, but potentially slow, steps we used with \
the Schwarzschild metric, here we shall use MetricCompute, which is simpler \
and faster, as we did in the CTensor framework.\
\>", "Text",
 CellChangeTimes->{
  3.730915204727722*^9, {3.730959077602152*^9, 3.730959103513144*^9}, {
   3.7310037006233*^9, 3.73100370560054*^9}, {3.731005176695373*^9, 
   3.731005255757349*^9}},
 CellID->1294809613,ExpressionUUID->"642fc84e-a609-4cd1-ad04-b13f631e4dca"],

Cell[CellGroupData[{

Cell["\<\
Compute the down-Riemann tensor. The use of Simplify simplification produces \
nicer results, but slower:\
\>", "MathCaption",
 CellChangeTimes->{{3.73100528729373*^9, 3.731005338393602*^9}},
 CellID->1799817068,ExpressionUUID->"8ee56a8e-222b-4c3c-b37e-35ab6e782ec9"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"MetricCompute", "[", 
  RowBox[{"metricg", ",", "B", ",", 
   RowBox[{"\"\<Riemann\>\"", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}], ",", 
   RowBox[{"CVSimplify", "\[Rule]", "Simplify"}], ",", 
   RowBox[{"Verbose", "\[Rule]", "True"}], ",", 
   RowBox[{"Parallelize", "\[Rule]", "True"}]}], "]"}]], "Input",
 CellChangeTimes->{
  3.7310053011561003`*^9, {3.731005349426632*^9, 3.7310053576462193`*^9}},
 CellLabel->"In[19]:=",
 CellID->472400955,ExpressionUUID->"8582ced9-75c2-49af-99a0-983fc8860f43"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Metric\"\>", "[", 
    RowBox[{"1", ",", "1"}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Metric"[1, 1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7310073559532757`*^9, 3.7310075167045317`*^9, 
  3.731007607653213*^9, 3.7310076552919493`*^9, 3.7310081861639843`*^9, 
  3.73100866091131*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->619974268,ExpressionUUID->"811310bd-0d6a-4b79-8c63-dd1fc0e94ea6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Constructed in \"\>", "\[InvisibleSpace]", "0.02336`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Constructed in ", 0.02336, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.7310073559532757`*^9, 3.7310075167045317`*^9, 
  3.731007607653213*^9, 3.7310076552919493`*^9, 3.7310081861639843`*^9, 
  3.731008660928277*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->316867032,ExpressionUUID->"5fd1a74f-1516-451f-a3be-20e4799963dd"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Applied \"\>", "\[InvisibleSpace]", "Simplify", 
   "\[InvisibleSpace]", "\<\" in \"\>", "\[InvisibleSpace]", "4.637506`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Applied ", Simplify, " in ", 4.637506, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.7310073559532757`*^9, 3.7310075167045317`*^9, 
  3.731007607653213*^9, 3.7310076552919493`*^9, 3.7310081861639843`*^9, 
  3.731008665513525*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1126934732,ExpressionUUID->"228ddc58-2686-430b-a9fa-86136c838338"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Stored in \"\>", "\[InvisibleSpace]", "0.000664`", 
   "\[InvisibleSpace]", "\<\" seconds and \"\>", "\[InvisibleSpace]", "14072",
    "\[InvisibleSpace]", "\<\" bytes\"\>"}],
  SequenceForm["    Stored in ", 0.000664, " seconds and ", 14072, " bytes"],
  Editable->False]], "Print",
 CellChangeTimes->{3.7310073559532757`*^9, 3.7310075167045317`*^9, 
  3.731007607653213*^9, 3.7310076552919493`*^9, 3.7310081861639843`*^9, 
  3.731008665527069*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1733757831,ExpressionUUID->"2affbb7e-862a-4d45-93d2-63388fbddd96"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"DMetric\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "DMetric"[-1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7310073559532757`*^9, 3.7310075167045317`*^9, 
  3.731007607653213*^9, 3.7310076552919493`*^9, 3.7310081861639843`*^9, 
  3.731008665540092*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->62259474,ExpressionUUID->"2d934f90-9c78-4387-bbd8-147bf101d9c5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Constructed in \"\>", "\[InvisibleSpace]", "0.000645`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Constructed in ", 0.000645, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.7310073559532757`*^9, 3.7310075167045317`*^9, 
  3.731007607653213*^9, 3.7310076552919493`*^9, 3.7310081861639843`*^9, 
  3.7310086655536213`*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->570443429,ExpressionUUID->"9e48ada5-a905-44a3-85be-933ca8746308"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Applied \"\>", "\[InvisibleSpace]", "Simplify", 
   "\[InvisibleSpace]", "\<\" in \"\>", "\[InvisibleSpace]", "0.363813`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Applied ", Simplify, " in ", 0.363813, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.7310073559532757`*^9, 3.7310075167045317`*^9, 
  3.731007607653213*^9, 3.7310076552919493`*^9, 3.7310081861639843`*^9, 
  3.731008665885715*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1075710995,ExpressionUUID->"8b9e8991-0414-4b82-8631-1ace45f7a7b4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Stored in \"\>", "\[InvisibleSpace]", "0.00125`", 
   "\[InvisibleSpace]", "\<\" seconds and \"\>", "\[InvisibleSpace]", "64376",
    "\[InvisibleSpace]", "\<\" bytes\"\>"}],
  SequenceForm["    Stored in ", 0.00125, " seconds and ", 64376, " bytes"],
  Editable->False]], "Print",
 CellChangeTimes->{3.7310073559532757`*^9, 3.7310075167045317`*^9, 
  3.731007607653213*^9, 3.7310076552919493`*^9, 3.7310081861639843`*^9, 
  3.731008665899043*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1957352519,ExpressionUUID->"3ec77afa-e320-4e4a-9bf5-39cbc153332b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"DDMetric\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "DDMetric"[-1, -1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7310073559532757`*^9, 3.7310075167045317`*^9, 
  3.731007607653213*^9, 3.7310076552919493`*^9, 3.7310081861639843`*^9, 
  3.731008665912184*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->2069282087,ExpressionUUID->"373854e1-f7c4-42d0-96da-358283cd3857"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Constructed in \"\>", "\[InvisibleSpace]", "0.001239`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Constructed in ", 0.001239, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.7310073559532757`*^9, 3.7310075167045317`*^9, 
  3.731007607653213*^9, 3.7310076552919493`*^9, 3.7310081861639843`*^9, 
  3.7310086659261713`*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->2073856691,ExpressionUUID->"aa02c315-766b-4bc6-9509-ebddb6172690"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Applied \"\>", "\[InvisibleSpace]", "Simplify", 
   "\[InvisibleSpace]", "\<\" in \"\>", "\[InvisibleSpace]", "1.404028`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Applied ", Simplify, " in ", 1.404028, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.7310073559532757`*^9, 3.7310075167045317`*^9, 
  3.731007607653213*^9, 3.7310076552919493`*^9, 3.7310081861639843`*^9, 
  3.731008667302079*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1649814827,ExpressionUUID->"2e5ed5bc-e176-43ee-91a2-01f7a8f949b6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Stored in \"\>", "\[InvisibleSpace]", "0.004691`", 
   "\[InvisibleSpace]", "\<\" seconds and \"\>", "\[InvisibleSpace]", 
   "325896", "\[InvisibleSpace]", "\<\" bytes\"\>"}],
  SequenceForm["    Stored in ", 0.004691, " seconds and ", 325896, " bytes"],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.7310073559532757`*^9, 3.7310075167045317`*^9, 
  3.731007607653213*^9, 3.7310076552919493`*^9, 3.7310081861639843`*^9, 
  3.73100866731581*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->128947325,ExpressionUUID->"d9d110a5-d781-4489-ad1a-c7ae62c79e7c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Christoffel\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Christoffel"[-1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7310073559532757`*^9, 3.7310075167045317`*^9, 
  3.731007607653213*^9, 3.7310076552919493`*^9, 3.7310081861639843`*^9, 
  3.731008667328659*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->406691452,ExpressionUUID->"1a35c1bf-9804-4a28-984a-66844f22963f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Constructed in \"\>", "\[InvisibleSpace]", "0.000757`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Constructed in ", 0.000757, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.7310073559532757`*^9, 3.7310075167045317`*^9, 
  3.731007607653213*^9, 3.7310076552919493`*^9, 3.7310081861639843`*^9, 
  3.731008667341434*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1970269010,ExpressionUUID->"da007344-82fc-4b53-b603-1d98cde97f90"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Applied \"\>", "\[InvisibleSpace]", "Simplify", 
   "\[InvisibleSpace]", "\<\" in \"\>", "\[InvisibleSpace]", "0.45699`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Applied ", Simplify, " in ", 0.45699, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.7310073559532757`*^9, 3.7310075167045317`*^9, 
  3.731007607653213*^9, 3.7310076552919493`*^9, 3.7310081861639843`*^9, 
  3.731008667768631*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->113414330,ExpressionUUID->"1b353ff6-b42b-4d50-a706-abf3f2225a81"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Stored in \"\>", "\[InvisibleSpace]", "0.001532`", 
   "\[InvisibleSpace]", "\<\" seconds and \"\>", "\[InvisibleSpace]", 
   "114384", "\[InvisibleSpace]", "\<\" bytes\"\>"}],
  SequenceForm["    Stored in ", 0.001532, " seconds and ", 114384, " bytes"],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.7310073559532757`*^9, 3.7310075167045317`*^9, 
  3.731007607653213*^9, 3.7310076552919493`*^9, 3.7310081861639843`*^9, 
  3.731008667782795*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1336649421,ExpressionUUID->"1528b829-dd4e-4d42-b2f6-cf4c9626e058"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Christoffel\"\>", "[", 
    RowBox[{"1", ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Christoffel"[1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7310073559532757`*^9, 3.7310075167045317`*^9, 
  3.731007607653213*^9, 3.7310076552919493`*^9, 3.7310081861639843`*^9, 
  3.731008667796228*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->240387480,ExpressionUUID->"25a8a703-0772-45ca-afff-56df57287b6f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Constructed in \"\>", "\[InvisibleSpace]", "0.001298`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Constructed in ", 0.001298, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.7310073559532757`*^9, 3.7310075167045317`*^9, 
  3.731007607653213*^9, 3.7310076552919493`*^9, 3.7310081861639843`*^9, 
  3.731008667809071*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->775636353,ExpressionUUID->"fc711523-9fd4-4c30-b325-8139cc5dd753"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Applied \"\>", "\[InvisibleSpace]", "Simplify", 
   "\[InvisibleSpace]", "\<\" in \"\>", "\[InvisibleSpace]", "0.597588`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Applied ", Simplify, " in ", 0.597588, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.7310073559532757`*^9, 3.7310075167045317`*^9, 
  3.731007607653213*^9, 3.7310076552919493`*^9, 3.7310081861639843`*^9, 
  3.731008668372225*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->894447581,ExpressionUUID->"8ceb9784-ad88-439c-8ea5-4cd1bc630cf3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Stored in \"\>", "\[InvisibleSpace]", "0.001432`", 
   "\[InvisibleSpace]", "\<\" seconds and \"\>", "\[InvisibleSpace]", "83184",
    "\[InvisibleSpace]", "\<\" bytes\"\>"}],
  SequenceForm["    Stored in ", 0.001432, " seconds and ", 83184, " bytes"],
  Editable->False]], "Print",
 CellChangeTimes->{3.7310073559532757`*^9, 3.7310075167045317`*^9, 
  3.731007607653213*^9, 3.7310076552919493`*^9, 3.7310081861639843`*^9, 
  3.731008668385569*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->594964867,ExpressionUUID->"651f865b-b920-4277-b8cb-a50613eef1e7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Riemann\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Riemann"[-1, -1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7310073559532757`*^9, 3.7310075167045317`*^9, 
  3.731007607653213*^9, 3.7310076552919493`*^9, 3.7310081861639843`*^9, 
  3.731008668399329*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->890659237,ExpressionUUID->"5549a295-689c-4a2e-817d-46c9c1976f31"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Constructed in \"\>", "\[InvisibleSpace]", "0.002102`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Constructed in ", 0.002102, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.7310073559532757`*^9, 3.7310075167045317`*^9, 
  3.731007607653213*^9, 3.7310076552919493`*^9, 3.7310081861639843`*^9, 
  3.731008668412369*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1348776669,ExpressionUUID->"ceb79d6d-7d41-4e92-a3e3-2d1e0685b081"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Applied \"\>", "\[InvisibleSpace]", "Simplify", 
   "\[InvisibleSpace]", "\<\" in \"\>", "\[InvisibleSpace]", "1.297734`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Applied ", Simplify, " in ", 1.297734, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.7310073559532757`*^9, 3.7310075167045317`*^9, 
  3.731007607653213*^9, 3.7310076552919493`*^9, 3.7310081861639843`*^9, 
  3.7310086696778507`*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->52098151,ExpressionUUID->"4b91d6e2-554f-4cbf-a64a-dae06858e1e0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Stored in \"\>", "\[InvisibleSpace]", "0.012314`", 
   "\[InvisibleSpace]", "\<\" seconds and \"\>", "\[InvisibleSpace]", 
   "328496", "\[InvisibleSpace]", "\<\" bytes\"\>"}],
  SequenceForm["    Stored in ", 0.012314, " seconds and ", 328496, " bytes"],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.7310073559532757`*^9, 3.7310075167045317`*^9, 
  3.731007607653213*^9, 3.7310076552919493`*^9, 3.7310081861639843`*^9, 
  3.7310086697251253`*^9},
 CellLabel->"During evaluation of In[19]:=",
 CellID->1587779528,ExpressionUUID->"4793a181-9bc1-45da-9e37-d1ad6a18d207"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Check the independent components of the result:", "MathCaption",
 CellChangeTimes->{{3.7310054096655407`*^9, 3.7310054457110233`*^9}},
 CellID->933872350,ExpressionUUID->"9bffa71c-51f6-4436-b340-7fe884827cf1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IndependentRules", "[", 
   RowBox[{"TensorValues", "[", "RiemannCD", "]"}], "]"}], "//", 
  "Column"}]], "Input",
 CellChangeTimes->{{3.461425067096795*^9, 3.461425087110055*^9}, {
  3.73100542364931*^9, 3.731005431007599*^9}},
 CellLabel->"In[20]:=",
 CellID->172003964,ExpressionUUID->"2f8d462c-c1f9-4b0b-9d8b-aeef33324e0f"],

Cell[BoxData[
 TagBox[GridBox[{
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`B}, {1, -$CellContext`B}, {
        0, -$CellContext`B}, {1, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      RowBox[{
       FractionBox["1", 
        RowBox[{
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], "+", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Cos",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cos,
               Editable->False], "[", 
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False]}], "]"}]}], "+", 
            RowBox[{"2", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}]}], ")"}], "3"], " ", 
         RowBox[{"(", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["Q",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Charge,
             Editable->False], "2"], "+", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "2"}], " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Mass,
                Editable->False]}], "+", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}], ")"}]}]}], ")"}]}]], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "4"}], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["Q",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Charge,
            Editable->False], "4"], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"]}], "-", 
         RowBox[{"5", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["Q",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Charge,
            Editable->False], "2"], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "4"]}], "+", 
         RowBox[{"32", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["Q",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Charge,
            Editable->False], "2"], " ", 
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Mass,
           Editable->False], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], " ", 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False]}], "+", 
         RowBox[{"27", " ", 
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Mass,
           Editable->False], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "4"], " ", 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False]}], "+", 
         RowBox[{"24", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["Q",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Charge,
            Editable->False], "4"], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]}], "+", 
         RowBox[{"24", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["Q",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Charge,
            Editable->False], "2"], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]}], "-", 
         RowBox[{"48", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False], "2"], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]}], "-", 
         RowBox[{"64", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["Q",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Charge,
            Editable->False], "2"], " ", 
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Mass,
           Editable->False], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "3"]}], "+", 
         RowBox[{"4", " ", 
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Mass,
           Editable->False], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "3"]}], "+", 
         RowBox[{"24", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["Q",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Charge,
            Editable->False], "2"], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "4"]}], "+", 
         RowBox[{"32", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False], "2"], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "4"]}], "-", 
         RowBox[{"16", " ", 
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Mass,
           Editable->False], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "5"]}], "+", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "4"], " ", 
          RowBox[{
           InterpretationBox[
            StyleBox["Cos",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Cos,
            Editable->False], "[", 
           RowBox[{"4", " ", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False]}], "]"}], " ", 
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["Q",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Charge,
              Editable->False], "2"], "-", 
            RowBox[{"3", " ", 
             InterpretationBox[
              StyleBox["M",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Mass,
              Editable->False], " ", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]}]}], ")"}]}], "-", 
         RowBox[{"4", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], " ", 
          RowBox[{
           InterpretationBox[
            StyleBox["Cos",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Cos,
            Editable->False], "[", 
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False]}], "]"}], " ", 
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["Q",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Charge,
              Editable->False], "4"], "+", 
            RowBox[{
             InterpretationBox[
              StyleBox["M",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Mass,
              Editable->False], " ", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"-", "6"}], " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "2"]}], "+", 
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"12", " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "-", 
                  RowBox[{"7", " ", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}]}], ")"}], " ", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False]}]}], ")"}]}], "+", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["Q",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Charge,
               Editable->False], "2"], " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "+", 
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{"-", "4"}], " ", 
                   InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}], ")"}]}]}], ")"}]}]}], ")"}]}]}], 
        ")"}]}]}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`B}, {1, -$CellContext`B}, {
        0, -$CellContext`B}, {2, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{"3", " ", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Mass,
             Editable->False], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["Cos",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Cos,
              Editable->False], "[", 
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False]}], "]"}]}], "+", 
           RowBox[{"4", " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["Q",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Charge,
              Editable->False], "2"], " ", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Mass,
             Editable->False], " ", 
            RowBox[{"(", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["a",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Rotation,
                Editable->False], "2"], "-", 
              RowBox[{"6", " ", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
         RowBox[{
          InterpretationBox[
           StyleBox["Sin",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Sin,
           Editable->False], "[", 
          RowBox[{"2", " ", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False]}], "]"}]}], 
        RowBox[{"4", " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"], " ", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Cos",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Cos,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], "2"]}], "+", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"]}], ")"}], "3"]}]]}]}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`B}, {1, -$CellContext`B}, {
        0, -$CellContext`B}, {3, -$CellContext`B}],
       Editable->False], "\[Rule]", "0"}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`B}, {1, -$CellContext`B}, {
        1, -$CellContext`B}, {2, -$CellContext`B}],
       Editable->False], "\[Rule]", "0"}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`B}, {1, -$CellContext`B}, {
        1, -$CellContext`B}, {3, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{"4", " ", 
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["Q",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Charge,
              Editable->False], "4"], " ", 
            RowBox[{"(", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["a",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Rotation,
                Editable->False], "2"], "-", 
              RowBox[{"6", " ", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False], "2"]}]}], ")"}]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Mass,
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"3", " ", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"]}], "-", 
              RowBox[{"2", " ", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False], "2"]}]}], ")"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"3", " ", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"]}], "+", 
              RowBox[{
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "4"}], " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Mass,
                   Editable->False]}], "+", 
                 RowBox[{"3", " ", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}]}], ")"}]}]}], ")"}]}], "+", 
           RowBox[{"2", " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["Q",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Charge,
              Editable->False], "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["a",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Rotation,
                Editable->False], "4"], "+", 
              RowBox[{"4", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Mass,
                   Editable->False]}], "-", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], ")"}], " ", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False], "3"]}], "-", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], " ", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"4", " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Mass,
                   Editable->False]}], "+", 
                 RowBox[{"3", " ", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}]}], ")"}]}]}], ")"}]}], "+", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["Cos",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Cos,
              Editable->False], "[", 
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False]}], "]"}], " ", 
            RowBox[{"(", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["Q",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Charge,
                Editable->False], "4"], "+", 
              RowBox[{"3", " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Mass,
                Editable->False], " ", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "3"}], " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"]}], "+", 
                 RowBox[{"4", " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Mass,
                   Editable->False], " ", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}], "-", 
                 RowBox[{"3", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}]}], ")"}]}], "+", 
              RowBox[{"2", " ", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["Q",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Charge,
                 Editable->False], "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], "-", 
                 RowBox[{"4", " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Mass,
                   Editable->False], " ", 
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False]}], "+", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}], ")"}]}]}], ")"}]}]}], ")"}], " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Sin",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Sin,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}], 
        RowBox[{
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], "+", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Cos",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cos,
               Editable->False], "[", 
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False]}], "]"}]}], "+", 
            RowBox[{"2", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}]}], ")"}], "3"], " ", 
         RowBox[{"(", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["Q",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Charge,
             Editable->False], "2"], "+", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "2"}], " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Mass,
                Editable->False]}], "+", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}], ")"}]}]}], ")"}]}]]}]}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`B}, {1, -$CellContext`B}, {
        2, -$CellContext`B}, {3, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      FractionBox[
       RowBox[{
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "5"}], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"]}], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False]}], "]"}]}], "-", 
          RowBox[{"4", " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}]}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False]}], "]"}]}], "+", 
          RowBox[{"4", " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["Q",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Charge,
             Editable->False], "2"], " ", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]}], "+", 
          RowBox[{
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False], " ", 
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"], "-", 
             RowBox[{"6", " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["Sin",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Sin,
          Editable->False], "[", 
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False]}], "]"}]}], 
       RowBox[{"8", " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], " ", 
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["Cos",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cos,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], "2"]}], "+", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}], ")"}], "3"]}]]}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`B}, {2, -$CellContext`B}, {
        0, -$CellContext`B}, {2, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      RowBox[{
       FractionBox["1", 
        RowBox[{"2", " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"], " ", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Cos",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Cos,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], "2"]}], "+", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"]}], ")"}], "3"]}]], 
       RowBox[{"(", 
        RowBox[{
         FractionBox[
          RowBox[{"4", " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["Cos",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Cos,
              Editable->False], "[", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False], "]"}], "2"], " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["Q",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Charge,
                Editable->False], "2"], "-", 
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Mass,
                Editable->False], " ", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}]}], ")"}], "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"], "+", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}], ")"}]}], 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], "+", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["Cos",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Cos,
              Editable->False], "[", 
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False]}], "]"}]}], "+", 
           RowBox[{"2", " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"]}]}]], "-", 
         RowBox[{
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], " ", 
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["Q",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Charge,
              Editable->False], "2"], "+", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], "+", 
            RowBox[{
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"-", "2"}], " ", 
                InterpretationBox[
                 StyleBox["M",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Mass,
                 Editable->False]}], "+", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}], ")"}]}]}], ")"}], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox["M",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Mass,
              Editable->False], " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Cos",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cos,
               Editable->False], "[", 
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False]}], "]"}]}], "+", 
            RowBox[{"2", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["Q",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Charge,
               Editable->False], "2"], " ", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]}], "+", 
            RowBox[{
             InterpretationBox[
              StyleBox["M",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Mass,
              Editable->False], " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["a",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Rotation,
                 Editable->False], "2"], "-", 
               RowBox[{"2", " ", 
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False], "2"]}]}], ")"}]}]}], ")"}]}], "+", 
         RowBox[{"2", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "4"], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "2"], " ", 
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["Q",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Charge,
              Editable->False], "2"], "-", 
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["M",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Mass,
              Editable->False], " ", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]}]}], ")"}], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Sin",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Sin,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "2"]}], "-", 
         FractionBox[
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "4"], " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["Q",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Charge,
                Editable->False], "2"], "-", 
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Mass,
                Editable->False], " ", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}]}], ")"}], "2"], " ", 
           SuperscriptBox[
            RowBox[{
             InterpretationBox[
              StyleBox["Sin",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Sin,
              Editable->False], "[", 
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False]}], "]"}], "2"]}], 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], "+", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["Cos",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Cos,
              Editable->False], "[", 
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False]}], "]"}]}], "+", 
           RowBox[{"2", " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"]}]}]], "+", 
         RowBox[{"2", " ", 
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["Q",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Charge,
              Editable->False], "2"], "-", 
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["M",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Mass,
              Editable->False], " ", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]}]}], ")"}], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "4"], " ", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Cos",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Cos,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], "2"], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Cos",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cos,
               Editable->False], "[", 
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False]}], "]"}]}], "+", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Cos",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cos,
               Editable->False], "[", 
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False]}], "]"}], " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}], "+", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "4"], " ", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Sin",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Sin,
                Editable->False], "[", 
               RowBox[{"2", " ", 
                InterpretationBox[
                 StyleBox["\[Theta]",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`\[Theta][],
                 Editable->False]}], "]"}], "2"]}]}], ")"}]}]}], ")"}]}]}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`B}, {2, -$CellContext`B}, {
        0, -$CellContext`B}, {3, -$CellContext`B}],
       Editable->False], "\[Rule]", "0"}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`B}, {2, -$CellContext`B}, {
        1, -$CellContext`B}, {2, -$CellContext`B}],
       Editable->False], "\[Rule]", "0"}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`B}, {2, -$CellContext`B}, {
        1, -$CellContext`B}, {3, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      FractionBox[
       RowBox[{
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "2"}], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"]}], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False]}], "]"}]}], "-", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False]}], "]"}]}], "+", 
          RowBox[{"4", " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["Q",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Charge,
             Editable->False], "2"], " ", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]}], "+", 
          RowBox[{
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False], " ", 
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"], "-", 
             RowBox[{"6", " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["Sin",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Sin,
          Editable->False], "[", 
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False]}], "]"}]}], 
       RowBox[{"4", " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], " ", 
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["Cos",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cos,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], "2"]}], "+", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}], ")"}], "3"]}]]}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`B}, {2, -$CellContext`B}, {
        2, -$CellContext`B}, {3, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      FractionBox[
       RowBox[{"4", " ", 
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["Q",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Charge,
             Editable->False], "4"], " ", 
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"], "-", 
             RowBox[{"2", " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], "2"]}]}], ")"}]}], "-", 
          RowBox[{
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False], " ", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"3", " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["a",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Rotation,
                Editable->False], "2"]}], "-", 
             RowBox[{"2", " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], "2"]}]}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"3", " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["a",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Rotation,
                Editable->False], "2"]}], "-", 
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False], " ", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}], "+", 
             RowBox[{"3", " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], "2"]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["Q",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Charge,
             Editable->False], "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"4", " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["a",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Rotation,
                Editable->False], "4"]}], "-", 
             RowBox[{"5", " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["a",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Rotation,
                Editable->False], "2"], " ", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}], "+", 
             RowBox[{"6", " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], "3"]}], "-", 
             RowBox[{"4", " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], "4"]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False]}], "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["Q",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Charge,
               Editable->False], "4"], "+", 
             RowBox[{"3", " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False], " ", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "3"}], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"]}], "+", 
                RowBox[{"2", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], "-", 
                RowBox[{"3", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]}], ")"}]}], "+", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["Q",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Charge,
                Editable->False], "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"4", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"]}], "-", 
                RowBox[{"5", " ", 
                 InterpretationBox[
                  StyleBox["M",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Mass,
                  Editable->False], " ", 
                 InterpretationBox[
                  StyleBox["r",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`r[],
                  Editable->False]}], "+", 
                RowBox[{"4", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}]}], ")"}]}]}], ")"}]}]}], ")"}], 
        " ", 
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Sin",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Sin,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"]}], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False]}], "]"}]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}]}], ")"}], "3"]]}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`B}, {3, -$CellContext`B}, {
        0, -$CellContext`B}, {3, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      FractionBox[
       RowBox[{"4", " ", 
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["Q",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Charge,
            Editable->False], "2"], "+", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], "+", 
          RowBox[{
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "2"}], " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False]}], "+", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]}], ")"}]}]}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "3"}], " ", 
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]}], "+", 
          RowBox[{"2", " ", 
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "3"]}], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False]}], "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["Q",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Charge,
               Editable->False], "2"], "-", 
             RowBox[{"3", " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False], " ", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["Q",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Charge,
             Editable->False], "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"], "-", 
             RowBox[{"2", " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Sin",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Sin,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"]}], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False]}], "]"}]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}]}], ")"}], "3"]]}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`B}, {3, -$CellContext`B}, {
        1, -$CellContext`B}, {2, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      FractionBox[
       RowBox[{
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False]}], "]"}]}], "+", 
          RowBox[{"4", " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["Q",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Charge,
             Editable->False], "2"], " ", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]}], "+", 
          RowBox[{
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False], " ", 
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"], "-", 
             RowBox[{"6", " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["Sin",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Sin,
          Editable->False], "[", 
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False]}], "]"}]}], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False]}], "]"}]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}]}], ")"}], "2"]]}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`B}, {3, -$CellContext`B}, {
        1, -$CellContext`B}, {3, -$CellContext`B}],
       Editable->False], "\[Rule]", "0"}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{0, -$CellContext`B}, {3, -$CellContext`B}, {
        2, -$CellContext`B}, {3, -$CellContext`B}],
       Editable->False], "\[Rule]", "0"}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{1, -$CellContext`B}, {2, -$CellContext`B}, {
        1, -$CellContext`B}, {2, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{
         RowBox[{
          RowBox[{"-", "3"}], " ", 
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Mass,
           Editable->False], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], " ", 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False]}], "+", 
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Mass,
           Editable->False], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "3"]}], "+", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], " ", 
          RowBox[{
           InterpretationBox[
            StyleBox["Cos",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Cos,
            Editable->False], "[", 
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False]}], "]"}], " ", 
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["Q",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Charge,
              Editable->False], "2"], "-", 
            RowBox[{"3", " ", 
             InterpretationBox[
              StyleBox["M",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Mass,
              Editable->False], " ", 
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False]}]}], ")"}]}], "+", 
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["Q",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Charge,
            Editable->False], "2"], " ", 
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], "-", 
            RowBox[{"2", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}]}], ")"}]}]}], 
        RowBox[{"2", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], " ", 
            SuperscriptBox[
             RowBox[{
              InterpretationBox[
               StyleBox["Cos",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cos,
               Editable->False], "[", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False], "]"}], "2"]}], "+", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["Q",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Charge,
             Editable->False], "2"], "+", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "2"}], " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Mass,
                Editable->False]}], "+", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}], ")"}]}]}], ")"}]}]]}]}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{1, -$CellContext`B}, {2, -$CellContext`B}, {
        1, -$CellContext`B}, {3, -$CellContext`B}],
       Editable->False], "\[Rule]", "0"}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{1, -$CellContext`B}, {2, -$CellContext`B}, {
        2, -$CellContext`B}, {3, -$CellContext`B}],
       Editable->False], "\[Rule]", "0"}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{1, -$CellContext`B}, {3, -$CellContext`B}, {
        1, -$CellContext`B}, {3, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      RowBox[{
       FractionBox["1", 
        RowBox[{
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], "+", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Cos",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cos,
               Editable->False], "[", 
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False]}], "]"}]}], "+", 
            RowBox[{"2", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}]}], ")"}], "3"], " ", 
         RowBox[{"(", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["Q",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Charge,
             Editable->False], "2"], "+", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "2"}], " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Mass,
                Editable->False]}], "+", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}], ")"}]}]}], ")"}]}]], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["Q",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Charge,
              Editable->False], "4"]}], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "4"]}], "-", 
          RowBox[{"5", " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["Q",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Charge,
             Editable->False], "2"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "6"]}], "+", 
          RowBox[{"8", " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["Q",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Charge,
             Editable->False], "2"], " ", 
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "4"], " ", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]}], "+", 
          RowBox[{"18", " ", 
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "6"], " ", 
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False]}], "+", 
          RowBox[{"12", " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["Q",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Charge,
             Editable->False], "4"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}], "+", 
          RowBox[{"11", " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["Q",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Charge,
             Editable->False], "2"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "4"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}], "-", 
          RowBox[{"12", " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Mass,
             Editable->False], "2"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "4"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}], "-", 
          RowBox[{"32", " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["Q",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Charge,
             Editable->False], "2"], " ", 
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "3"]}], "+", 
          RowBox[{"14", " ", 
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "4"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "3"]}], "+", 
          RowBox[{"24", " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["Q",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Charge,
             Editable->False], "2"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "4"]}], "+", 
          RowBox[{"16", " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Mass,
             Editable->False], "2"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "4"]}], "-", 
          RowBox[{"12", " ", 
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "5"]}], "+", 
          RowBox[{"8", " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["Q",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Charge,
             Editable->False], "2"], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "6"]}], "-", 
          RowBox[{"8", " ", 
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False], " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "7"]}], "+", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "4"], " ", 
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            RowBox[{"4", " ", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False]}], "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["Q",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Charge,
               Editable->False], "2"], "-", 
             RowBox[{"6", " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False], " ", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}]}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["Q",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Charge,
               Editable->False], "2"], "+", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"], "-", 
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False], " ", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False]}], "+", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}], ")"}]}], "-", 
          RowBox[{"4", " ", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], " ", 
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            RowBox[{"2", " ", 
             InterpretationBox[
              StyleBox["\[Theta]",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`\[Theta][],
              Editable->False]}], "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"3", " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["Q",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Charge,
                Editable->False], "4"], " ", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], "2"]}], "+", 
             RowBox[{
              InterpretationBox[
               StyleBox["M",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Mass,
               Editable->False], " ", 
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "3"}], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "4"]}], "-", 
                RowBox[{"8", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}], "+", 
                RowBox[{
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"4", " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "-", 
                   RowBox[{"5", " ", 
                    InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}]}], ")"}], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "3"]}]}], ")"}]}], "+", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["Q",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Charge,
                Editable->False], "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox[
                 InterpretationBox[
                  StyleBox["a",
                   ShowAutoStyles->False,
                   AutoSpacing->False],
                  $CellContext`Rotation,
                  Editable->False], "4"], "+", 
                RowBox[{"5", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Rotation,
                   Editable->False], "2"], " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "2"]}], "+", 
                RowBox[{"4", " ", 
                 SuperscriptBox[
                  InterpretationBox[
                   StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`r[],
                   Editable->False], "3"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    InterpretationBox[
                    StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Mass,
                    Editable->False]}], "+", 
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False]}], ")"}]}]}], ")"}]}]}], ")"}]}]}], 
         ")"}], " ", 
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Sin",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Sin,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"]}]}]}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{1, -$CellContext`B}, {3, -$CellContext`B}, {
        2, -$CellContext`B}, {3, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{"3", " ", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["a",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Rotation,
           Editable->False], "2"], " ", 
         RowBox[{
          InterpretationBox[
           StyleBox["Cos",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Cos,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], " ", 
         RowBox[{"(", 
          RowBox[{
           SuperscriptBox[
            InterpretationBox[
             StyleBox["a",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Rotation,
             Editable->False], "2"], "+", 
           SuperscriptBox[
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False], "2"]}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Mass,
             Editable->False], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["Cos",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Cos,
              Editable->False], "[", 
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False]}], "]"}]}], "+", 
           RowBox[{"4", " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["Q",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Charge,
              Editable->False], "2"], " ", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Mass,
             Editable->False], " ", 
            RowBox[{"(", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["a",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Rotation,
                Editable->False], "2"], "-", 
              RowBox[{"6", " ", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False], "2"]}]}], ")"}]}]}], ")"}], " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Sin",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Sin,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "3"]}], 
        RowBox[{"2", " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"], " ", 
             SuperscriptBox[
              RowBox[{
               InterpretationBox[
                StyleBox["Cos",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                Cos,
                Editable->False], "[", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False], "]"}], "2"]}], "+", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"]}], ")"}], "3"]}]]}]}]},
    {
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["2",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["3",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[{2, -$CellContext`B}, {3, -$CellContext`B}, {
        2, -$CellContext`B}, {3, -$CellContext`B}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       RowBox[{
        FractionBox["1", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], "+", 
            RowBox[{
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"], " ", 
             RowBox[{
              InterpretationBox[
               StyleBox["Cos",
                ShowAutoStyles->False,
                AutoSpacing->False],
               Cos,
               Editable->False], "[", 
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["\[Theta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Theta][],
                Editable->False]}], "]"}]}], "+", 
            RowBox[{"2", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}]}], ")"}], "3"]], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["Q",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Charge,
               Editable->False], "4"]}], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "4"]}], "-", 
           RowBox[{"13", " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["Q",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Charge,
              Editable->False], "2"], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "6"]}], "+", 
           RowBox[{"5", " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["Q",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Charge,
              Editable->False], "2"], " ", 
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Mass,
             Editable->False], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "4"], " ", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]}], "+", 
           RowBox[{"27", " ", 
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Mass,
             Editable->False], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "6"], " ", 
            InterpretationBox[
             StyleBox["r",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`r[],
             Editable->False]}], "+", 
           RowBox[{"4", " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["Q",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Charge,
              Editable->False], "4"], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"]}], "-", 
           RowBox[{"13", " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["Q",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Charge,
              Editable->False], "2"], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "4"], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"]}], "-", 
           RowBox[{"6", " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["M",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Mass,
              Editable->False], "2"], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "4"], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "2"]}], "-", 
           RowBox[{"12", " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["Q",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Charge,
              Editable->False], "2"], " ", 
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Mass,
             Editable->False], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "3"]}], "+", 
           RowBox[{"31", " ", 
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Mass,
             Editable->False], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "4"], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "3"]}], "+", 
           RowBox[{"8", " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["Q",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Charge,
              Editable->False], "2"], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "4"]}], "+", 
           RowBox[{"8", " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["M",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Mass,
              Editable->False], "2"], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "4"]}], "-", 
           RowBox[{"12", " ", 
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Mass,
             Editable->False], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "5"]}], "+", 
           RowBox[{"8", " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["Q",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Charge,
              Editable->False], "2"], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "6"]}], "-", 
           RowBox[{"16", " ", 
            InterpretationBox[
             StyleBox["M",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`Mass,
             Editable->False], " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["r",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`r[],
              Editable->False], "7"]}], "+", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "4"], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["Cos",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Cos,
              Editable->False], "[", 
             RowBox[{"4", " ", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False]}], "]"}], " ", 
            RowBox[{"(", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["Q",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Charge,
                Editable->False], "2"], "-", 
              RowBox[{"3", " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Mass,
                Editable->False], " ", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}]}], ")"}], " ", 
            RowBox[{"(", 
             RowBox[{
              SuperscriptBox[
               InterpretationBox[
                StyleBox["Q",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Charge,
                Editable->False], "2"], "+", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["a",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Rotation,
                Editable->False], "2"], "-", 
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Mass,
                Editable->False], " ", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False]}], "+", 
              SuperscriptBox[
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], "2"]}], ")"}]}], "-", 
           RowBox[{"4", " ", 
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], " ", 
            RowBox[{
             InterpretationBox[
              StyleBox["Cos",
               ShowAutoStyles->False,
               AutoSpacing->False],
              Cos,
              Editable->False], "[", 
             RowBox[{"2", " ", 
              InterpretationBox[
               StyleBox["\[Theta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Theta][],
               Editable->False]}], "]"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["Q",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Charge,
                 Editable->False], "4"], " ", 
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["r",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`r[],
                 Editable->False], "2"]}], "+", 
              RowBox[{
               InterpretationBox[
                StyleBox["M",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`Mass,
                Editable->False], " ", 
               InterpretationBox[
                StyleBox["r",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`r[],
                Editable->False], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "6"}], " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "4"]}], "-", 
                 RowBox[{"13", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}], "+", 
                 RowBox[{"2", " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Mass,
                   Editable->False], " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "3"]}], "-", 
                 RowBox[{"7", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "4"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox[
                InterpretationBox[
                 StyleBox["Q",
                  ShowAutoStyles->False,
                  AutoSpacing->False],
                 $CellContext`Charge,
                 Editable->False], "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"3", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "4"]}], "+", 
                 RowBox[{"7", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["a",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`Rotation,
                    Editable->False], "2"], " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "2"]}], "-", 
                 RowBox[{"3", " ", 
                  InterpretationBox[
                   StyleBox["M",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                   $CellContext`Mass,
                   Editable->False], " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "3"]}], "+", 
                 RowBox[{"4", " ", 
                  SuperscriptBox[
                   InterpretationBox[
                    StyleBox["r",
                    ShowAutoStyles->False,
                    AutoSpacing->False],
                    $CellContext`r[],
                    Editable->False], "4"]}]}], ")"}]}]}], ")"}]}]}], ")"}], 
         " ", 
         SuperscriptBox[
          RowBox[{
           InterpretationBox[
            StyleBox["Sin",
             ShowAutoStyles->False,
             AutoSpacing->False],
            Sin,
            Editable->False], "[", 
           InterpretationBox[
            StyleBox["\[Theta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Theta][],
            Editable->False], "]"}], "2"]}]}]}]}]}
   },
   DefaultBaseStyle->"Column",
   GridBoxAlignment->{"Columns" -> {{Left}}},
   GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
  "Column"]], "Output",
 CellChangeTimes->{3.731005431549613*^9, 3.7310066141420507`*^9, 
  3.731006684716689*^9, 3.731007309849278*^9, 3.731007366608286*^9, 
  3.7310075249898233`*^9, 3.731007615477026*^9, 3.731007663827312*^9, 
  3.731008203797518*^9, 3.731008670030842*^9},
 CellLabel->"Out[20]=",
 CellID->1178729474,ExpressionUUID->"6d31d707-d5fa-4bb8-8ffc-a8e59075811a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Compute the Kretschmann scalar. Note how previous computations were cached:\
\>", "MathCaption",
 CellChangeTimes->{{3.7310054801513557`*^9, 3.731005511372871*^9}},
 CellID->891591804,ExpressionUUID->"d744620d-cec3-45f7-bc4d-a14f29979bf5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"MetricCompute", "[", 
  RowBox[{"metricg", ",", "B", ",", 
   RowBox[{"\"\<Kretschmann\>\"", "[", "]"}], ",", 
   RowBox[{"CVSimplify", "\[Rule]", "Simplify"}], ",", 
   RowBox[{"Verbose", "\[Rule]", "True"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.458495181034781*^9, 3.458495196545216*^9}, {
  3.4614250284183807`*^9, 3.46142503744906*^9}, {3.4614251212794952`*^9, 
  3.4614251254447803`*^9}, {3.731005488845776*^9, 3.7310054976008472`*^9}},
 CellLabel->"In[21]:=",
 CellID->1253380061,ExpressionUUID->"43e2d24a-90ba-490b-81ae-1f6cab55b8f4"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Metric\"\>", "[", 
    RowBox[{"1", ",", "1"}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Metric"[1, 1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.731005498316283*^9, 3.73100661428561*^9, 
  3.731006684828423*^9, 3.73100731002862*^9, 3.731007369995331*^9, 
  3.731007525173118*^9, 3.731007615634243*^9, 3.731007664059927*^9, 
  3.73100820817942*^9, 3.731008670181942*^9},
 CellLabel->"During evaluation of In[21]:=",
 CellID->369625943,ExpressionUUID->"7fdc86a3-9da7-4548-ab71-bdafb91bd9f1"],

Cell[BoxData["\<\"    Already cached\"\>"], "Print",
 CellChangeTimes->{3.731005498316283*^9, 3.73100661428561*^9, 
  3.731006684828423*^9, 3.73100731002862*^9, 3.731007369995331*^9, 
  3.731007525173118*^9, 3.731007615634243*^9, 3.731007664059927*^9, 
  3.73100820817942*^9, 3.731008670198039*^9},
 CellLabel->"During evaluation of In[21]:=",
 CellID->1780516877,ExpressionUUID->"def63c57-3a39-4cb7-b9f7-e227382c2470"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"DMetric\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "DMetric"[-1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.731005498316283*^9, 3.73100661428561*^9, 
  3.731006684828423*^9, 3.73100731002862*^9, 3.731007369995331*^9, 
  3.731007525173118*^9, 3.731007615634243*^9, 3.731007664059927*^9, 
  3.73100820817942*^9, 3.7310086702143383`*^9},
 CellLabel->"During evaluation of In[21]:=",
 CellID->2110014441,ExpressionUUID->"61c9b278-e884-4818-9f45-6ebbf4fa5a7b"],

Cell[BoxData["\<\"    Already cached\"\>"], "Print",
 CellChangeTimes->{3.731005498316283*^9, 3.73100661428561*^9, 
  3.731006684828423*^9, 3.73100731002862*^9, 3.731007369995331*^9, 
  3.731007525173118*^9, 3.731007615634243*^9, 3.731007664059927*^9, 
  3.73100820817942*^9, 3.731008670224375*^9},
 CellLabel->"During evaluation of In[21]:=",
 CellID->1615246976,ExpressionUUID->"0934395f-3a38-4d36-8ac9-39fc6af222e4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"DDMetric\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "DDMetric"[-1, -1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.731005498316283*^9, 3.73100661428561*^9, 
  3.731006684828423*^9, 3.73100731002862*^9, 3.731007369995331*^9, 
  3.731007525173118*^9, 3.731007615634243*^9, 3.731007664059927*^9, 
  3.73100820817942*^9, 3.731008670234149*^9},
 CellLabel->"During evaluation of In[21]:=",
 CellID->1115143905,ExpressionUUID->"4a44770e-9a01-4991-9af0-35ef67789eed"],

Cell[BoxData["\<\"    Already cached\"\>"], "Print",
 CellChangeTimes->{3.731005498316283*^9, 3.73100661428561*^9, 
  3.731006684828423*^9, 3.73100731002862*^9, 3.731007369995331*^9, 
  3.731007525173118*^9, 3.731007615634243*^9, 3.731007664059927*^9, 
  3.73100820817942*^9, 3.7310086702439747`*^9},
 CellLabel->"During evaluation of In[21]:=",
 CellID->1133823966,ExpressionUUID->"07d5c611-b7ff-4e75-863f-be7315075735"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Christoffel\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Christoffel"[-1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.731005498316283*^9, 3.73100661428561*^9, 
  3.731006684828423*^9, 3.73100731002862*^9, 3.731007369995331*^9, 
  3.731007525173118*^9, 3.731007615634243*^9, 3.731007664059927*^9, 
  3.73100820817942*^9, 3.731008670253848*^9},
 CellLabel->"During evaluation of In[21]:=",
 CellID->1556996731,ExpressionUUID->"dc0e1f29-ce59-4575-9265-ca82e13b4fda"],

Cell[BoxData["\<\"    Already cached\"\>"], "Print",
 CellChangeTimes->{3.731005498316283*^9, 3.73100661428561*^9, 
  3.731006684828423*^9, 3.73100731002862*^9, 3.731007369995331*^9, 
  3.731007525173118*^9, 3.731007615634243*^9, 3.731007664059927*^9, 
  3.73100820817942*^9, 3.7310086702639847`*^9},
 CellLabel->"During evaluation of In[21]:=",
 CellID->1355819222,ExpressionUUID->"b14f0848-ae87-407c-81a6-2594ba028c99"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Christoffel\"\>", "[", 
    RowBox[{"1", ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Christoffel"[1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.731005498316283*^9, 3.73100661428561*^9, 
  3.731006684828423*^9, 3.73100731002862*^9, 3.731007369995331*^9, 
  3.731007525173118*^9, 3.731007615634243*^9, 3.731007664059927*^9, 
  3.73100820817942*^9, 3.731008670273946*^9},
 CellLabel->"During evaluation of In[21]:=",
 CellID->304685837,ExpressionUUID->"9196ec60-2f83-42fc-bfae-31e79649ce7c"],

Cell[BoxData["\<\"    Already cached\"\>"], "Print",
 CellChangeTimes->{3.731005498316283*^9, 3.73100661428561*^9, 
  3.731006684828423*^9, 3.73100731002862*^9, 3.731007369995331*^9, 
  3.731007525173118*^9, 3.731007615634243*^9, 3.731007664059927*^9, 
  3.73100820817942*^9, 3.731008670283453*^9},
 CellLabel->"During evaluation of In[21]:=",
 CellID->1253848011,ExpressionUUID->"87591ade-45f6-47f3-902c-1d77489139ab"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Riemann\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Riemann"[-1, -1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.731005498316283*^9, 3.73100661428561*^9, 
  3.731006684828423*^9, 3.73100731002862*^9, 3.731007369995331*^9, 
  3.731007525173118*^9, 3.731007615634243*^9, 3.731007664059927*^9, 
  3.73100820817942*^9, 3.73100867029263*^9},
 CellLabel->"During evaluation of In[21]:=",
 CellID->166492866,ExpressionUUID->"c7656a4d-7e15-4297-b60a-398f63bc38e6"],

Cell[BoxData["\<\"    Already cached\"\>"], "Print",
 CellChangeTimes->{3.731005498316283*^9, 3.73100661428561*^9, 
  3.731006684828423*^9, 3.73100731002862*^9, 3.731007369995331*^9, 
  3.731007525173118*^9, 3.731007615634243*^9, 3.731007664059927*^9, 
  3.73100820817942*^9, 3.731008670302568*^9},
 CellLabel->"During evaluation of In[21]:=",
 CellID->74406821,ExpressionUUID->"ce53fdb8-773e-4115-939d-bf8a1a8814ad"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Riemann\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", "1"}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Riemann"[-1, -1, -1, 1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.731005498316283*^9, 3.73100661428561*^9, 
  3.731006684828423*^9, 3.73100731002862*^9, 3.731007369995331*^9, 
  3.731007525173118*^9, 3.731007615634243*^9, 3.731007664059927*^9, 
  3.73100820817942*^9, 3.731008670312134*^9},
 CellLabel->"During evaluation of In[21]:=",
 CellID->719957993,ExpressionUUID->"95bb3adb-0482-4121-b475-965f827811b1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Constructed in \"\>", "\[InvisibleSpace]", "0.004125`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Constructed in ", 0.004125, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.731005498316283*^9, 3.73100661428561*^9, 
  3.731006684828423*^9, 3.73100731002862*^9, 3.731007369995331*^9, 
  3.731007525173118*^9, 3.731007615634243*^9, 3.731007664059927*^9, 
  3.73100820817942*^9, 3.731008670321327*^9},
 CellLabel->"During evaluation of In[21]:=",
 CellID->1411121153,ExpressionUUID->"33cfefea-e224-4241-b962-cfc5d21614ef"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Applied \"\>", "\[InvisibleSpace]", "Simplify", 
   "\[InvisibleSpace]", "\<\" in \"\>", "\[InvisibleSpace]", "5.136589`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Applied ", Simplify, " in ", 5.136589, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.731005498316283*^9, 3.73100661428561*^9, 
  3.731006684828423*^9, 3.73100731002862*^9, 3.731007369995331*^9, 
  3.731007525173118*^9, 3.731007615634243*^9, 3.731007664059927*^9, 
  3.73100820817942*^9, 3.731008675338022*^9},
 CellLabel->"During evaluation of In[21]:=",
 CellID->2051304650,ExpressionUUID->"14d4da98-0e1a-40cc-b544-69822dfe4432"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Stored in \"\>", "\[InvisibleSpace]", "0.011748`", 
   "\[InvisibleSpace]", "\<\" seconds and \"\>", "\[InvisibleSpace]", 
   "390368", "\[InvisibleSpace]", "\<\" bytes\"\>"}],
  SequenceForm["    Stored in ", 0.011748, " seconds and ", 390368, " bytes"],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.731005498316283*^9, 3.73100661428561*^9, 
  3.731006684828423*^9, 3.73100731002862*^9, 3.731007369995331*^9, 
  3.731007525173118*^9, 3.731007615634243*^9, 3.731007664059927*^9, 
  3.73100820817942*^9, 3.731008675391788*^9},
 CellLabel->"During evaluation of In[21]:=",
 CellID->574783612,ExpressionUUID->"f3cf2a43-7539-4af1-82b2-481689ff7623"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Riemann\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", "1", ",", "1"}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Riemann"[-1, -1, 1, 1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.731005498316283*^9, 3.73100661428561*^9, 
  3.731006684828423*^9, 3.73100731002862*^9, 3.731007369995331*^9, 
  3.731007525173118*^9, 3.731007615634243*^9, 3.731007664059927*^9, 
  3.73100820817942*^9, 3.731008675403983*^9},
 CellLabel->"During evaluation of In[21]:=",
 CellID->1006722678,ExpressionUUID->"c09da613-3fa0-4165-b22c-d4b475034e37"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Constructed in \"\>", "\[InvisibleSpace]", "0.002175`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Constructed in ", 0.002175, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.731005498316283*^9, 3.73100661428561*^9, 
  3.731006684828423*^9, 3.73100731002862*^9, 3.731007369995331*^9, 
  3.731007525173118*^9, 3.731007615634243*^9, 3.731007664059927*^9, 
  3.73100820817942*^9, 3.731008675415964*^9},
 CellLabel->"During evaluation of In[21]:=",
 CellID->2111878080,ExpressionUUID->"4fe44c4f-cf4a-46ed-9b3d-3c86adf75180"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Applied \"\>", "\[InvisibleSpace]", "Simplify", 
   "\[InvisibleSpace]", "\<\" in \"\>", "\[InvisibleSpace]", "1.964168`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Applied ", Simplify, " in ", 1.964168, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.731005498316283*^9, 3.73100661428561*^9, 
  3.731006684828423*^9, 3.73100731002862*^9, 3.731007369995331*^9, 
  3.731007525173118*^9, 3.731007615634243*^9, 3.731007664059927*^9, 
  3.73100820817942*^9, 3.731008677328076*^9},
 CellLabel->"During evaluation of In[21]:=",
 CellID->725172944,ExpressionUUID->"9a1f76c3-8bc7-4a76-b9db-e46d4524cf34"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Stored in \"\>", "\[InvisibleSpace]", "0.012077`", 
   "\[InvisibleSpace]", "\<\" seconds and \"\>", "\[InvisibleSpace]", 
   "268648", "\[InvisibleSpace]", "\<\" bytes\"\>"}],
  SequenceForm["    Stored in ", 0.012077, " seconds and ", 268648, " bytes"],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.731005498316283*^9, 3.73100661428561*^9, 
  3.731006684828423*^9, 3.73100731002862*^9, 3.731007369995331*^9, 
  3.731007525173118*^9, 3.731007615634243*^9, 3.731007664059927*^9, 
  3.73100820817942*^9, 3.731008677398834*^9},
 CellLabel->"During evaluation of In[21]:=",
 CellID->1011973083,ExpressionUUID->"76e05a46-0102-4416-9b87-731effa48322"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Kretschmann\"\>", "[", "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Kretschmann"[]],
  Editable->False]], "Print",
 CellChangeTimes->{3.731005498316283*^9, 3.73100661428561*^9, 
  3.731006684828423*^9, 3.73100731002862*^9, 3.731007369995331*^9, 
  3.731007525173118*^9, 3.731007615634243*^9, 3.731007664059927*^9, 
  3.73100820817942*^9, 3.731008677413028*^9},
 CellLabel->"During evaluation of In[21]:=",
 CellID->161121741,ExpressionUUID->"f4e1af7b-1245-4d66-8ed0-e4441c76f8b0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Constructed in \"\>", "\[InvisibleSpace]", "0.002853`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Constructed in ", 0.002853, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.731005498316283*^9, 3.73100661428561*^9, 
  3.731006684828423*^9, 3.73100731002862*^9, 3.731007369995331*^9, 
  3.731007525173118*^9, 3.731007615634243*^9, 3.731007664059927*^9, 
  3.73100820817942*^9, 3.73100867742583*^9},
 CellLabel->"During evaluation of In[21]:=",
 CellID->2143705767,ExpressionUUID->"6abd2108-a14e-453a-99d6-ef585cf1a51f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Applied \"\>", "\[InvisibleSpace]", "Simplify", 
   "\[InvisibleSpace]", "\<\" in \"\>", "\[InvisibleSpace]", "1.260936`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Applied ", Simplify, " in ", 1.260936, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.731005498316283*^9, 3.73100661428561*^9, 
  3.731006684828423*^9, 3.73100731002862*^9, 3.731007369995331*^9, 
  3.731007525173118*^9, 3.731007615634243*^9, 3.731007664059927*^9, 
  3.73100820817942*^9, 3.731008678613307*^9},
 CellLabel->"During evaluation of In[21]:=",
 CellID->929680250,ExpressionUUID->"978d45eb-b1ec-49b4-9456-2a297de44734"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Stored in \"\>", "\[InvisibleSpace]", "0.000193`", 
   "\[InvisibleSpace]", "\<\" seconds and \"\>", "\[InvisibleSpace]", "7792", 
   "\[InvisibleSpace]", "\<\" bytes\"\>"}],
  SequenceForm["    Stored in ", 0.000193, " seconds and ", 7792, " bytes"],
  Editable->False]], "Print",
 CellChangeTimes->{3.731005498316283*^9, 3.73100661428561*^9, 
  3.731006684828423*^9, 3.73100731002862*^9, 3.731007369995331*^9, 
  3.731007525173118*^9, 3.731007615634243*^9, 3.731007664059927*^9, 
  3.73100820817942*^9, 3.731008678625905*^9},
 CellLabel->"During evaluation of In[21]:=",
 CellID->44946178,ExpressionUUID->"0585785f-ce86-459d-aa6e-8f48ae17693c"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["This is the result:", "MathCaption",
 CellChangeTimes->{{3.731005587627713*^9, 3.731005590082395*^9}},
 CellID->336090536,ExpressionUUID->"d0fc5947-e88b-487f-9916-a12f37f743e8"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"KretschmannCD", "[", "]"}], "//", "ToValues"}]], "Input",
 CellChangeTimes->{{3.46142520037217*^9, 3.461425217439412*^9}},
 CellLabel->"In[22]:=",
 CellID->1680049480,ExpressionUUID->"15087f24-7176-4f9c-9859-688200ebe04c"],

Cell[BoxData[
 RowBox[{"-", 
  RowBox[{
   FractionBox["1", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], "2"], "+", 
       RowBox[{
        SuperscriptBox[
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], "2"], " ", 
        RowBox[{
         InterpretationBox[
          StyleBox["Cos",
           ShowAutoStyles->False,
           AutoSpacing->False],
          Cos,
          Editable->False], "[", 
         RowBox[{"2", " ", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False]}], "]"}]}], "+", 
       RowBox[{"2", " ", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]}]}], ")"}], "6"]], 
   RowBox[{"32", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "42"}], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["Q",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Charge,
         Editable->False], "4"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], "4"]}], "+", 
      RowBox[{"30", " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Mass,
         Editable->False], "2"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], "6"]}], "+", 
      RowBox[{"3", " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Mass,
         Editable->False], "2"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], "6"], " ", 
       RowBox[{
        InterpretationBox[
         StyleBox["Cos",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Cos,
         Editable->False], "[", 
        RowBox[{"6", " ", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False]}], "]"}]}], "+", 
      RowBox[{"360", " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["Q",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Charge,
         Editable->False], "2"], " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Mass,
        Editable->False], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], "4"], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False]}], "+", 
      RowBox[{"272", " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["Q",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Charge,
         Editable->False], "4"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], "2"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"]}], "-", 
      RowBox[{"540", " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Mass,
         Editable->False], "2"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], "4"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "2"]}], "-", 
      RowBox[{"960", " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["Q",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Charge,
         Editable->False], "2"], " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Mass,
        Editable->False], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], "2"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "3"]}], "-", 
      RowBox[{"112", " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["Q",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Charge,
         Editable->False], "4"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "4"]}], "+", 
      RowBox[{"720", " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Mass,
         Editable->False], "2"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], "2"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "4"]}], "+", 
      RowBox[{"192", " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["Q",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Charge,
         Editable->False], "2"], " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Mass,
        Editable->False], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "5"]}], "-", 
      RowBox[{"96", " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Mass,
         Editable->False], "2"], " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["r",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`r[],
         Editable->False], "6"]}], "+", 
      RowBox[{"2", " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], "4"], " ", 
       RowBox[{
        InterpretationBox[
         StyleBox["Cos",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Cos,
         Editable->False], "[", 
        RowBox[{"4", " ", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False]}], "]"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "7"}], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["Q",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Charge,
            Editable->False], "4"]}], "+", 
         RowBox[{"60", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["Q",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Charge,
            Editable->False], "2"], " ", 
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Mass,
           Editable->False], " ", 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False]}], "+", 
         RowBox[{"9", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False], "2"], " ", 
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], "-", 
            RowBox[{"10", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}]}], ")"}]}]}], ")"}]}], "+", 
      RowBox[{
       SuperscriptBox[
        InterpretationBox[
         StyleBox["a",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Rotation,
         Editable->False], "2"], " ", 
       RowBox[{
        InterpretationBox[
         StyleBox["Cos",
          ShowAutoStyles->False,
          AutoSpacing->False],
         Cos,
         Editable->False], "[", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Theta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Theta][],
          Editable->False]}], "]"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "8"}], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["Q",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Charge,
            Editable->False], "4"], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"7", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"]}], "-", 
            RowBox[{"34", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}]}], ")"}]}], "+", 
         RowBox[{"480", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["Q",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Charge,
            Editable->False], "2"], " ", 
          InterpretationBox[
           StyleBox["M",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`Mass,
           Editable->False], " ", 
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], " ", 
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "2"], "-", 
            RowBox[{"2", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}]}], ")"}]}], "+", 
         RowBox[{"45", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["M",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Mass,
            Editable->False], "2"], " ", 
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox[
             InterpretationBox[
              StyleBox["a",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`Rotation,
              Editable->False], "4"], "-", 
            RowBox[{"16", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["a",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`Rotation,
               Editable->False], "2"], " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "2"]}], "+", 
            RowBox[{"16", " ", 
             SuperscriptBox[
              InterpretationBox[
               StyleBox["r",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`r[],
               Editable->False], "4"]}]}], ")"}]}]}], ")"}]}]}], 
     ")"}]}]}]}]], "Output",
 CellChangeTimes->{3.731005525800476*^9, 3.7310066229358587`*^9, 
  3.731006693461583*^9, 3.7310073822456503`*^9, 3.731007533898209*^9, 
  3.7310076243162727`*^9, 3.7310076725392103`*^9, 3.731008219781784*^9, 
  3.73100867895091*^9},
 CellLabel->"Out[22]=",
 CellID->1643653549,ExpressionUUID->"20faf592-16f0-457a-9ded-c5a28807eb45"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Collect powers of the charge parameter:", "MathCaption",
 CellChangeTimes->{{3.73100559510603*^9, 3.7310056030180407`*^9}},
 CellID->934268211,ExpressionUUID->"95f6e10e-e82c-41af-928a-ea149e87afe6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"Collect", "[", 
    RowBox[{"#", ",", "Charge", ",", "FullSimplify"}], "]"}], "&"}], "/@", 
  "%"}]], "Input",
 CellChangeTimes->{{3.461425232910903*^9, 3.461425247553841*^9}, {
  3.7310055559072437`*^9, 3.731005556498513*^9}},
 CellLabel->"In[23]:=",
 CellID->376564360,ExpressionUUID->"c2f80f50-46fb-4e58-bfd9-a56669de8fd9"],

Cell[BoxData[
 RowBox[{"-", 
  RowBox[{
   FractionBox["1", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       RowBox[{"2", " ", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["a",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`Rotation,
          Editable->False], "2"], " ", 
        SuperscriptBox[
         RowBox[{
          InterpretationBox[
           StyleBox["Cos",
            ShowAutoStyles->False,
            AutoSpacing->False],
           Cos,
           Editable->False], "[", 
          InterpretationBox[
           StyleBox["\[Theta]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Theta][],
           Editable->False], "]"}], "2"]}], "+", 
       RowBox[{"2", " ", 
        SuperscriptBox[
         InterpretationBox[
          StyleBox["r",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`r[],
          Editable->False], "2"]}]}], ")"}], "6"]], 
   RowBox[{"32", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"192", " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["Q",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Charge,
         Editable->False], "2"], " ", 
       InterpretationBox[
        StyleBox["M",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Mass,
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"5", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "4"], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "4"]}], "-", 
         RowBox[{"10", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "2"], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]}], "+", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "4"]}], ")"}]}], "-", 
      RowBox[{"16", " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["Q",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Charge,
         Editable->False], "4"], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"7", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "4"], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "4"]}], "-", 
         RowBox[{"34", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "2"], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]}], "+", 
         RowBox[{"7", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "4"]}]}], ")"}]}], "+", 
      RowBox[{"96", " ", 
       SuperscriptBox[
        InterpretationBox[
         StyleBox["M",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`Mass,
         Editable->False], "2"], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "6"], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "6"]}], "-", 
         RowBox[{"15", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "4"], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "4"], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "2"]}], "+", 
         RowBox[{"15", " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["a",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`Rotation,
            Editable->False], "2"], " ", 
          SuperscriptBox[
           RowBox[{
            InterpretationBox[
             StyleBox["Cos",
              ShowAutoStyles->False,
              AutoSpacing->False],
             Cos,
             Editable->False], "[", 
            InterpretationBox[
             StyleBox["\[Theta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Theta][],
             Editable->False], "]"}], "2"], " ", 
          SuperscriptBox[
           InterpretationBox[
            StyleBox["r",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`r[],
            Editable->False], "4"]}], "-", 
         SuperscriptBox[
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], "6"]}], ")"}]}]}], ")"}]}]}]}]], "Output",
 CellChangeTimes->{3.731005560192649*^9, 3.73100662377489*^9, 
  3.731006694307373*^9, 3.7310073841820297`*^9, 3.731007534718196*^9, 
  3.731007625136578*^9, 3.731007673362965*^9, 3.731008221843556*^9, 
  3.731008679949238*^9},
 CellLabel->"Out[23]=",
 CellID->1803046682,ExpressionUUID->"4dda3aaf-de82-4e52-b227-6ab0526b8048"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["4.13.4. Bianchi identity", "Subsubsection",
 CellChangeTimes->{{3.7286771125390873`*^9, 3.7286771127222157`*^9}, {
  3.730913913338131*^9, 3.730913941103621*^9}, {3.731004967632353*^9, 
  3.7310049678228483`*^9}, {3.7310051409369698`*^9, 3.731005159992453*^9}, {
  3.7310056594734583`*^9, 3.7310056652799463`*^9}},
 CellID->184667122,ExpressionUUID->"d00c094b-e312-4fda-bdfc-b90d5707f82e"],

Cell[CellGroupData[{

Cell["Request computation of derivatives of the Riemann tensor:", \
"MathCaption",
 CellChangeTimes->{{3.731005751692844*^9, 3.731005759620018*^9}},
 CellID->460893304,ExpressionUUID->"a61889c1-4790-496f-ae19-85d63df3feef"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"MetricCompute", "[", 
  RowBox[{"metricg", ",", "B", ",", "All", ",", 
   RowBox[{"CVSimplify", "\[Rule]", "Simplify"}], ",", 
   RowBox[{"Verbose", "\[Rule]", "True"}], ",", 
   RowBox[{"Parallelize", "\[Rule]", "True"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.458495181034781*^9, 3.458495196545216*^9}, {
  3.4614250284183807`*^9, 3.46142503744906*^9}, {3.4614251212794952`*^9, 
  3.4614251254447803`*^9}, {3.461429803725239*^9, 3.461429804207926*^9}, {
  3.731005678520875*^9, 3.7310056917404633`*^9}, {3.731006581897505*^9, 
  3.731006586913571*^9}, {3.731006711576078*^9, 3.7310067295969763`*^9}, {
  3.731007397936996*^9, 3.7310073993109083`*^9}, {3.731007555961917*^9, 
  3.731007564573452*^9}},
 CellLabel->"In[24]:=",
 CellID->106417567,ExpressionUUID->"fb9c75f2-586b-4e94-8136-46ab68fc4f8e"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"DetMetric\"\>", "[", "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "DetMetric"[]],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008687761652*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->183170943,ExpressionUUID->"bbf47d44-7861-426c-9fea-1409257ac8b7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Constructed in \"\>", "\[InvisibleSpace]", "0.090088`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Constructed in ", 0.090088, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.7310086878493633`*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->1209972850,ExpressionUUID->"76296028-388f-4de8-b91c-bfe3d00594c3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Applied \"\>", "\[InvisibleSpace]", "Simplify", 
   "\[InvisibleSpace]", "\<\" in \"\>", "\[InvisibleSpace]", "0.01355`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Applied ", Simplify, " in ", 0.01355, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008687865837*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->1491036507,ExpressionUUID->"f150aba2-17c9-4436-a290-21a70df61d60"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Stored in \"\>", "\[InvisibleSpace]", "0.000188`", 
   "\[InvisibleSpace]", "\<\" seconds and \"\>", "\[InvisibleSpace]", "1200", 
   "\[InvisibleSpace]", "\<\" bytes\"\>"}],
  SequenceForm["    Stored in ", 0.000188, " seconds and ", 1200, " bytes"],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008687878306*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->863896306,ExpressionUUID->"ff721be8-1985-463f-884b-8ab6b21042dd"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Metric\"\>", "[", 
    RowBox[{"1", ",", "1"}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Metric"[1, 1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008687890749*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->368277575,ExpressionUUID->"cff8f03a-eb09-496a-bee7-c6da57b64512"],

Cell[BoxData["\<\"    Already cached\"\>"], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008687903535*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->593332371,ExpressionUUID->"54841b90-153b-4142-b2e2-685c80bb8614"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"DMetric\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "DMetric"[-1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008687915371*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->1370586376,ExpressionUUID->"61ac913b-c832-469e-94f8-8e60ce798575"],

Cell[BoxData["\<\"    Already cached\"\>"], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008687928183*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->1535623710,ExpressionUUID->"9f364b11-1b2a-4ea6-aa61-0548dd5197cc"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"DDMetric\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "DDMetric"[-1, -1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008687941115*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->769224324,ExpressionUUID->"0d5f0487-adc0-4513-b18c-f98489b9f1b2"],

Cell[BoxData["\<\"    Already cached\"\>"], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008687954035*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->501658528,ExpressionUUID->"7969149a-0473-48d5-8746-f967498d8c66"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Christoffel\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Christoffel"[-1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008687966855*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->354081974,ExpressionUUID->"b2b5d23e-c13b-4969-923a-9e4638061cd6"],

Cell[BoxData["\<\"    Already cached\"\>"], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008687979257*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->378551181,ExpressionUUID->"d21c5c89-4b03-4d91-9a5b-b576bbf42ed9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Christoffel\"\>", "[", 
    RowBox[{"1", ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Christoffel"[1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.7310086879912663`*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->1463136653,ExpressionUUID->"37dcb28d-1f37-4e4a-b525-58071f684dc2"],

Cell[BoxData["\<\"    Already cached\"\>"], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008688003385*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->102485174,ExpressionUUID->"6da4def3-938a-434d-8ff0-baea367481e3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Riemann\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Riemann"[-1, -1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008688015633*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->186434524,ExpressionUUID->"6c3fcb47-fa05-4fdb-8251-c07f42b81c22"],

Cell[BoxData["\<\"    Already cached\"\>"], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008688027628*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->226403895,ExpressionUUID->"046b0df9-804c-4f4d-9395-0c27ee2e44b4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Riemann\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", "1"}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Riemann"[-1, -1, -1, 1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008688039983*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->1976724428,ExpressionUUID->"660252cf-ce03-48e9-b3f4-a8ed3b7026dc"],

Cell[BoxData["\<\"    Already cached\"\>"], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008688051797*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->1235442306,ExpressionUUID->"d5c09eef-55c3-473b-bafa-86a80cb74c95"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Ricci\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Ricci"[-1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008688064061*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->59454099,ExpressionUUID->"ad819efc-86ea-4685-bdc6-9764e6b3858c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Constructed in \"\>", "\[InvisibleSpace]", "0.000298`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Constructed in ", 0.000298, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008688078176*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->665146038,ExpressionUUID->"60c8d46c-4879-4f41-b384-54dcbb782737"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Applied \"\>", "\[InvisibleSpace]", "Simplify", 
   "\[InvisibleSpace]", "\<\" in \"\>", "\[InvisibleSpace]", "0.162379`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Applied ", Simplify, " in ", 0.162379, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008688091194*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->1457078031,ExpressionUUID->"ae23918b-a1aa-4685-9824-2e6a463aa3a1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Stored in \"\>", "\[InvisibleSpace]", "0.000491`", 
   "\[InvisibleSpace]", "\<\" seconds and \"\>", "\[InvisibleSpace]", "17312",
    "\[InvisibleSpace]", "\<\" bytes\"\>"}],
  SequenceForm["    Stored in ", 0.000491, " seconds and ", 17312, " bytes"],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008688104176*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->1354440276,ExpressionUUID->"d1379018-181d-415e-8ab0-6f9551348ded"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"RicciScalar\"\>", "[", "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "RicciScalar"[]],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008688116313*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->751060532,ExpressionUUID->"0a0c82b4-7944-41e6-8f85-709cfe1b6288"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Constructed in \"\>", "\[InvisibleSpace]", "0.000193`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Constructed in ", 0.000193, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.7310086881287537`*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->165484258,ExpressionUUID->"bd55c4a4-1e50-4e93-9a30-fed5ea6403be"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Applied \"\>", "\[InvisibleSpace]", "Simplify", 
   "\[InvisibleSpace]", "\<\" in \"\>", "\[InvisibleSpace]", "0.070497`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Applied ", Simplify, " in ", 0.070497, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008688140767*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->302601341,ExpressionUUID->"adff7d3e-99ed-457f-bd1e-11662a806cf7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Stored in \"\>", "\[InvisibleSpace]", "0.000178`", 
   "\[InvisibleSpace]", "\<\" seconds and \"\>", "\[InvisibleSpace]", "256", 
   "\[InvisibleSpace]", "\<\" bytes\"\>"}],
  SequenceForm["    Stored in ", 0.000178, " seconds and ", 256, " bytes"],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008688154325*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->579462091,ExpressionUUID->"417ccc52-f3ec-4c20-a74a-677a05712b5e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Einstein\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Einstein"[-1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.7310086881674232`*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->181024292,ExpressionUUID->"aaf09a23-1c6a-49a4-9b4c-6f62ab76aba9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Constructed in \"\>", "\[InvisibleSpace]", "0.000159`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Constructed in ", 0.000159, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008688176998*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->1638431492,ExpressionUUID->"7593b95c-989e-4f8e-baae-057bc3c23c7f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Applied \"\>", "\[InvisibleSpace]", "Simplify", 
   "\[InvisibleSpace]", "\<\" in \"\>", "\[InvisibleSpace]", "0.076986`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Applied ", Simplify, " in ", 0.076986, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008688187879*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->2082764210,ExpressionUUID->"24f510a5-201c-44cc-978f-426eff2d9fcd"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Stored in \"\>", "\[InvisibleSpace]", "0.000426`", 
   "\[InvisibleSpace]", "\<\" seconds and \"\>", "\[InvisibleSpace]", "17312",
    "\[InvisibleSpace]", "\<\" bytes\"\>"}],
  SequenceForm["    Stored in ", 0.000426, " seconds and ", 17312, " bytes"],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.73100868819835*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->1034631370,ExpressionUUID->"7fe4f1ab-1641-48cd-ad19-b8e79a5662e1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Weyl\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Weyl"[-1, -1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008688210845*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->874345831,ExpressionUUID->"fecab347-f586-4753-bb4d-437a8ed00cdb"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Constructed in \"\>", "\[InvisibleSpace]", "0.000994`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Constructed in ", 0.000994, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008688278968*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->2047268843,ExpressionUUID->"4ae53884-ac08-462a-b68e-47e990e2a881"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Applied \"\>", "\[InvisibleSpace]", "Simplify", 
   "\[InvisibleSpace]", "\<\" in \"\>", "\[InvisibleSpace]", "0.652728`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Applied ", Simplify, " in ", 0.652728, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008688807992*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->1464452067,ExpressionUUID->"3beb5447-cb9e-4232-aff2-da488c640995"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Stored in \"\>", "\[InvisibleSpace]", "0.017185`", 
   "\[InvisibleSpace]", "\<\" seconds and \"\>", "\[InvisibleSpace]", 
   "323080", "\[InvisibleSpace]", "\<\" bytes\"\>"}],
  SequenceForm["    Stored in ", 0.017185, " seconds and ", 323080, " bytes"],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008688854603*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->735811802,ExpressionUUID->"f26473c3-ef50-4e73-8e32-272233e5c0f7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Riemann\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", "1", ",", "1"}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Riemann"[-1, -1, 1, 1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008688865206*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->1578116788,ExpressionUUID->"03c16e84-9896-4947-89fe-cbd0169f4084"],

Cell[BoxData["\<\"    Already cached\"\>"], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.7310086888753843`*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->1985815466,ExpressionUUID->"846a0ab8-6924-4882-9858-add350f7891a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"Kretschmann\"\>", "[", "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "Kretschmann"[]],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008688887294*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->1557179035,ExpressionUUID->"ffe3e429-790d-4ee5-aa9e-fdc30b567734"],

Cell[BoxData["\<\"    Already cached\"\>"], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008688896963*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->124835256,ExpressionUUID->"3d24aad0-2f6d-4280-b366-2659a7c8e3c0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** ReportCompute: \"\>", "\[InvisibleSpace]", 
   RowBox[{"\<\"CDRiemann\"\>", "[", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}],
  SequenceForm["** ReportCompute: ", 
   "CDRiemann"[-1, -1, -1, -1, -1]],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008688909873*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->14624473,ExpressionUUID->"04b0b9eb-ad0c-4eb0-8360-ad0280b09131"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Constructed in \"\>", "\[InvisibleSpace]", "0.01464`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Constructed in ", 0.01464, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.7310086889217663`*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->980381953,ExpressionUUID->"7d99c010-e004-4cd4-953d-cf8b7107877d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Applied \"\>", "\[InvisibleSpace]", "Simplify", 
   "\[InvisibleSpace]", "\<\" in \"\>", "\[InvisibleSpace]", "7.95971`", 
   "\[InvisibleSpace]", "\<\" seconds\"\>"}],
  SequenceForm["    Applied ", Simplify, " in ", 7.95971, " seconds"],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008696827359*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->1784944287,ExpressionUUID->"240bdd32-2d51-4940-8bb6-9c9dee2a3b3a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"    Stored in \"\>", "\[InvisibleSpace]", "0.015598`", 
   "\[InvisibleSpace]", "\<\" seconds and \"\>", "\[InvisibleSpace]", 
   "1557584", "\[InvisibleSpace]", "\<\" bytes\"\>"}],
  SequenceForm["    Stored in ", 0.015598, " seconds and ", 1557584, " bytes"],
  Editable->False]], "Print",
 CellChangeTimes->{3.731007625242915*^9, 3.7310076736059303`*^9, 
  3.73100822881746*^9, 3.731008696878819*^9},
 CellLabel->"During evaluation of In[24]:=",
 CellID->1359566666,ExpressionUUID->"dc2357c1-e12a-4780-a8ab-fea1944e6aef"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
In particular now we have computed the determinant of the metric as well:\
\>", "MathCaption",
 CellChangeTimes->{{3.731005796643592*^9, 3.73100580578724*^9}},
 CellID->2126404645,ExpressionUUID->"70fac69f-28f0-4e4a-a769-d6faf4f599a4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"DetmetricgB", "[", "]"}], "//", "ToValues"}]], "Input",
 CellChangeTimes->{3.73100580922717*^9},
 CellLabel->"In[25]:=",
 CellID->60015141,ExpressionUUID->"cca0254f-4ce7-4efa-8e5f-7f903d9faca2"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   FractionBox["1", "4"]}], " ", 
  SuperscriptBox[
   RowBox[{"(", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["a",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Rotation,
       Editable->False], "2"], "+", 
     RowBox[{
      SuperscriptBox[
       InterpretationBox[
        StyleBox["a",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Rotation,
        Editable->False], "2"], " ", 
      RowBox[{
       InterpretationBox[
        StyleBox["Cos",
         ShowAutoStyles->False,
         AutoSpacing->False],
        Cos,
        Editable->False], "[", 
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox["\[Theta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Theta][],
         Editable->False]}], "]"}]}], "+", 
     RowBox[{"2", " ", 
      SuperscriptBox[
       InterpretationBox[
        StyleBox["r",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`r[],
        Editable->False], "2"]}]}], ")"}], "2"], " ", 
  SuperscriptBox[
   RowBox[{
    InterpretationBox[
     StyleBox["Sin",
      ShowAutoStyles->False,
      AutoSpacing->False],
     Sin,
     Editable->False], "[", 
    InterpretationBox[
     StyleBox["\[Theta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Theta][],
     Editable->False], "]"}], "2"]}]], "Output",
 CellChangeTimes->{
  3.7310058125264177`*^9, {3.731007696228694*^9, 3.731007709664495*^9}, 
   3.731008263856223*^9, 3.731008701348543*^9},
 CellLabel->"Out[25]=",
 CellID->1521614842,ExpressionUUID->"5d3bc757-95f1-4cb0-8d83-ca017107d521"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Check of the contracted Bianchi identity. Start from the uncontracted \
derivative of the Einstein tensor:\
\>", "MathCaption",
 CellChangeTimes->{{3.458490528768036*^9, 3.458490539212646*^9}, {
  3.731005961030731*^9, 3.731005993789019*^9}},
 CellID->1937546183,ExpressionUUID->"bba588bf-90b0-4d97-accb-8cd0578c2db9"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "a"}], "]"}], "[", 
    RowBox[{"EinsteinCD", "[", 
     RowBox[{
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}]}], "]"}], "]"}], "//", 
   RowBox[{"ToBasis", "[", "B", "]"}]}], "//", 
  RowBox[{"ToBasis", "[", "B", "]"}]}]], "Input",
 CellChangeTimes->{{3.461429579768877*^9, 3.461429598313718*^9}, {
  3.461429646043256*^9, 3.46142964741238*^9}, {3.731005824827737*^9, 
  3.731005865665436*^9}},
 CellLabel->"In[26]:=",
 CellID->2044158220,ExpressionUUID->"8418701f-4e9a-4387-baf7-792198aabd02"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["d",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["a",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["c",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDB[{$CellContext`d, $CellContext`B}, \
{-$CellContext`a, -$CellContext`B}, {-$CellContext`c, -$CellContext`B}],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"G", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["b",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["d",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`EinsteinCD[{-$CellContext`b, -$CellContext`B}, \
{-$CellContext`d, -$CellContext`B}],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
        StyleBox[GridBox[{
           {
            StyleBox["d",
             FontColor->RGBColor[1., 0., 0.]], " ", " "},
           {" ", 
            StyleBox["a",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["b",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCDPDB[{$CellContext`d, $CellContext`B}, \
{-$CellContext`a, -$CellContext`B}, {-$CellContext`b, -$CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"G", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["d",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["c",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`EinsteinCD[{-$CellContext`d, -$CellContext`B}, \
{-$CellContext`c, -$CellContext`B}],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", 
      StyleBox["a",
       FontColor->RGBColor[1., 0., 0.]]], GridBox[{
       {
        RowBox[{"G", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["b",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["c",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[{-$CellContext`a, -$CellContext`B}][
    $CellContext`EinsteinCD[{-$CellContext`b, -$CellContext`B}, \
{-$CellContext`c, -$CellContext`B}]],
   Editable->False]}]], "Output",
 CellChangeTimes->{{3.731005826760985*^9, 3.731005865979822*^9}, {
   3.731007697627531*^9, 3.7310077121600513`*^9}, 3.7310082653237953`*^9, 
   3.7310087030320807`*^9},
 CellLabel->"Out[26]=",
 CellID->511626195,ExpressionUUID->"775e4cf9-b85d-477f-a185-0d193c17b61b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Now perform the contraction:", "MathCaption",
 CellChangeTimes->{{3.7310059997566347`*^9, 3.7310060042366133`*^9}},
 CellID->592187185,ExpressionUUID->"f895cd67-a333-4279-8a13-972b21825409"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"%", 
   RowBox[{"metricg", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"c", ",", "B"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"b", ",", "B"}], "}"}]}], "]"}]}], "//", "Expand"}]], "Input",
 CellChangeTimes->{{3.461429600724888*^9, 3.461429651143211*^9}, {
   3.461429729651132*^9, 3.461429731548937*^9}, 3.73100583884242*^9},
 CellLabel->"In[27]:=",
 CellID->1875766430,ExpressionUUID->"4209e6b8-cfa3-4fd1-ada4-f7b19e45456e"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["d",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["a",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["c",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDB[{$CellContext`d, $CellContext`B}, \
{-$CellContext`a, -$CellContext`B}, {-$CellContext`c, -$CellContext`B}],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"G", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["b",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["d",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`EinsteinCD[{-$CellContext`b, -$CellContext`B}, \
{-$CellContext`d, -$CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {
            StyleBox["c",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["b",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{$CellContext`c, $CellContext`B}, {$CellContext`b, \
$CellContext`B}],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
        StyleBox[GridBox[{
           {
            StyleBox["d",
             FontColor->RGBColor[1., 0., 0.]], " ", " "},
           {" ", 
            StyleBox["a",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["b",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCDPDB[{$CellContext`d, $CellContext`B}, \
{-$CellContext`a, -$CellContext`B}, {-$CellContext`b, -$CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"G", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["d",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["c",
             FontColor->RGBColor[1., 0., 0.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`EinsteinCD[{-$CellContext`d, -$CellContext`B}, \
{-$CellContext`c, -$CellContext`B}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {
            StyleBox["c",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["b",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{$CellContext`c, $CellContext`B}, {$CellContext`b, \
$CellContext`B}],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {
            StyleBox["c",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["b",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[{$CellContext`c, $CellContext`B}, {$CellContext`b, \
$CellContext`B}],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", 
        StyleBox["a",
         FontColor->RGBColor[1., 0., 0.]]], GridBox[{
         {
          RowBox[{"G", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["b",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["c",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[{-$CellContext`a, -$CellContext`B}][
      $CellContext`EinsteinCD[{-$CellContext`b, -$CellContext`B}, \
{-$CellContext`c, -$CellContext`B}]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.731005844227751*^9, 3.7310058969219027`*^9, 
  3.7310077133949003`*^9, 3.7310082666661997`*^9, 3.7310087044173098`*^9},
 CellLabel->"Out[27]=",
 CellID->849032740,ExpressionUUID->"8fa903ec-cce7-4202-b89a-3ac0edac79d9"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Expand into components. Each expression has 140 terms:", "MathCaption",
 CellChangeTimes->{{3.7310060070685377`*^9, 3.731006022700121*^9}},
 CellID->1222276682,ExpressionUUID->"53082d73-54e9-46a8-a194-4594b57b09c8"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"array", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"%", "//", "TraceBasisDummy"}], "//", "ComponentArray"}], "//", 
   "Expand"}]}]], "Input",
 CellChangeTimes->{{3.458490167290288*^9, 3.45849041795218*^9}, {
   3.4584905709888287`*^9, 3.4584905879398127`*^9}, 3.4584941645836973`*^9},
 CellLabel->"In[28]:=",
 CellID->591424021,ExpressionUUID->"4072f733-ce48-4585-a1b1-d6613f10533d"],

Cell[BoxData[
 InterpretationBox[
  TagBox[
   FrameBox[GridBox[{
      {
       ItemBox[
        TagBox[
         RowBox[{"{", 
          RowBox[{
           RowBox[{
            TemplateBox[{"260"},
             "OutputSizeLimit`Skeleton"], "+", 
            RowBox[{
             InterpretationBox[
              StyleBox[GridBox[{
                 {"g", 
                  StyleBox[GridBox[{
                    {
                    TemplateBox[{"49"},
                    "OutputSizeLimit`Skeleton"], 
                    TemplateBox[{"49"},
                    "OutputSizeLimit`Skeleton"]},
                    {" ", " "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`metricg[{1, $CellContext`B}, {3, $CellContext`B}],
              Editable->False], " ", 
             RowBox[{"(", 
              InterpretationBox[
               StyleBox[
                RowBox[{
                 SubscriptBox["\[EmptyDownTriangle]", 
                  TemplateBox[{"49"},
                   "OutputSizeLimit`Skeleton"]], GridBox[{
                   {
                    RowBox[{"G", "[", "\[EmptyDownTriangle]", "]"}], 
                    StyleBox[
                    TemplateBox[{"1"},
                    "OutputSizeLimit`Skeleton"],
                    FontSize->9]}
                  },
                  
                  GridBoxAlignment->{
                   "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}]}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`CD[{0, -$CellContext`B}][
                $CellContext`EinsteinCD[{3, -$CellContext`B}, {
                 1, -$CellContext`B}]],
               Editable->False], ")"}]}], "+", 
            RowBox[{
             InterpretationBox[
              StyleBox[GridBox[{
                 {"g", 
                  StyleBox[GridBox[{
                    {
                    StyleBox["2",
                    FontColor->RGBColor[1., 0., 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0., 0.]]},
                    {" ", " "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`metricg[{2, $CellContext`B}, {3, $CellContext`B}],
              Editable->False], " ", 
             RowBox[{"(", 
              InterpretationBox[
               StyleBox[
                RowBox[{
                 SubscriptBox["\[EmptyDownTriangle]", 
                  StyleBox["0",
                   FontColor->RGBColor[1., 0., 0.]]], GridBox[{
                   {
                    RowBox[{"G", "[", "\[EmptyDownTriangle]", "]"}], 
                    StyleBox[GridBox[{
                    {" ", " "},
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0., 0.]], 
                    TemplateBox[{"49"},
                    "OutputSizeLimit`Skeleton"]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                    FontSize->9]}
                  },
                  
                  GridBoxAlignment->{
                   "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}]}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`CD[{0, -$CellContext`B}][
                $CellContext`EinsteinCD[{3, -$CellContext`B}, {
                 2, -$CellContext`B}]],
               Editable->False], ")"}]}], "+", 
            RowBox[{
             InterpretationBox[
              StyleBox[GridBox[{
                 {"g", 
                  StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0., 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0., 0.]]},
                    {" ", " "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`metricg[{3, $CellContext`B}, {3, $CellContext`B}],
              Editable->False], " ", 
             RowBox[{"(", 
              InterpretationBox[
               StyleBox[
                RowBox[{
                 SubscriptBox["\[EmptyDownTriangle]", 
                  StyleBox["0",
                   FontColor->RGBColor[1., 0., 0.]]], GridBox[{
                   {
                    RowBox[{"G", "[", "\[EmptyDownTriangle]", "]"}], 
                    StyleBox[GridBox[{
                    {" ", " "},
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0., 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0., 0.]]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                    FontSize->9]}
                  },
                  
                  GridBoxAlignment->{
                   "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}]}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`CD[{0, -$CellContext`B}][
                $CellContext`EinsteinCD[{3, -$CellContext`B}, {
                 3, -$CellContext`B}]],
               Editable->False], ")"}]}]}], ",", 
           RowBox[{
            TemplateBox[{"261"},
             "OutputSizeLimit`Skeleton"], "+", 
            TemplateBox[{"1"},
             "OutputSizeLimit`Skeleton"], "+", 
            RowBox[{
             InterpretationBox[
              StyleBox[GridBox[{
                 {"g", 
                  StyleBox[
                   TemplateBox[{"1"},
                    "OutputSizeLimit`Skeleton"],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`metricg[{3, $CellContext`B}, {3, $CellContext`B}],
              Editable->False], " ", 
             RowBox[{"(", 
              InterpretationBox[
               StyleBox[
                RowBox[{
                 SubscriptBox["\[EmptyDownTriangle]", 
                  TemplateBox[{"49"},
                   "OutputSizeLimit`Skeleton"]], GridBox[{
                   {
                    RowBox[{"G", "[", "\[EmptyDownTriangle]", "]"}], 
                    StyleBox[GridBox[{
                    {" ", " "},
                    {
                    TemplateBox[{"49"},
                    "OutputSizeLimit`Skeleton"], 
                    TemplateBox[{"49"},
                    "OutputSizeLimit`Skeleton"]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                    FontSize->9]}
                  },
                  
                  GridBoxAlignment->{
                   "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}]}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`CD[{1, -$CellContext`B}][
                $CellContext`EinsteinCD[{3, -$CellContext`B}, {
                 3, -$CellContext`B}]],
               Editable->False], ")"}]}]}], ",", 
           TemplateBox[{"1"},
            "OutputSizeLimit`Skeleton"], ",", 
           RowBox[{
            TemplateBox[{"261"},
             "OutputSizeLimit`Skeleton"], "+", 
            RowBox[{
             InterpretationBox[
              StyleBox[GridBox[{
                 {"g", 
                  StyleBox[GridBox[{
                    {
                    TemplateBox[{"49"},
                    "OutputSizeLimit`Skeleton"], 
                    TemplateBox[{"49"},
                    "OutputSizeLimit`Skeleton"]},
                    {" ", " "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`metricg[{2, $CellContext`B}, {3, $CellContext`B}],
              Editable->False], " ", 
             RowBox[{"(", 
              InterpretationBox[
               StyleBox[
                RowBox[{
                 SubscriptBox["\[EmptyDownTriangle]", 
                  TemplateBox[{"49"},
                   "OutputSizeLimit`Skeleton"]], GridBox[{
                   {
                    RowBox[{"G", "[", "\[EmptyDownTriangle]", "]"}], 
                    StyleBox[
                    TemplateBox[{"1"},
                    "OutputSizeLimit`Skeleton"],
                    FontSize->9]}
                  },
                  
                  GridBoxAlignment->{
                   "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}]}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`CD[{3, -$CellContext`B}][
                $CellContext`EinsteinCD[{3, -$CellContext`B}, {
                 2, -$CellContext`B}]],
               Editable->False], ")"}]}], "+", 
            RowBox[{
             InterpretationBox[
              StyleBox[GridBox[{
                 {"g", 
                  StyleBox[GridBox[{
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0., 0.]], 
                    StyleBox["3",
                    FontColor->RGBColor[1., 0., 0.]]},
                    {" ", " "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`metricg[{3, $CellContext`B}, {3, $CellContext`B}],
              Editable->False], " ", 
             RowBox[{"(", 
              InterpretationBox[
               StyleBox[
                RowBox[{
                 SubscriptBox["\[EmptyDownTriangle]", 
                  StyleBox["3",
                   FontColor->RGBColor[1., 0., 0.]]], GridBox[{
                   {
                    RowBox[{"G", "[", "\[EmptyDownTriangle]", "]"}], 
                    StyleBox[GridBox[{
                    {" ", " "},
                    {
                    StyleBox["3",
                    FontColor->RGBColor[1., 0., 0.]], 
                    TemplateBox[{"49"},
                    "OutputSizeLimit`Skeleton"]}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                    FontSize->9]}
                  },
                  
                  GridBoxAlignment->{
                   "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}]}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`CD[{3, -$CellContext`B}][
                $CellContext`EinsteinCD[{3, -$CellContext`B}, {
                 3, -$CellContext`B}]],
               Editable->False], ")"}]}]}]}], "}"}],
         Short[#, 5]& ],
        BaseStyle->{Deployed -> False},
        StripOnInput->False]},
      {GridBox[{
         {
          TagBox[
           TooltipBox[
            StyleBox[
             StyleBox[
              DynamicBox[ToBoxes[
                FEPrivate`FrontEndResource[
                "FEStrings", "sizeBriefExplanation"], StandardForm],
               ImageSizeCache->{58., {2., 8.}}],
              StripOnInput->False,
              DynamicUpdating->True], "OSLText",
             StripOnInput->False],
            StyleBox[
             DynamicBox[
              ToBoxes[
               FEPrivate`FrontEndResource["FEStrings", "sizeExplanation"], 
               StandardForm]], DynamicUpdating -> True, StripOnInput -> 
             False]],
           Annotation[#, 
            Style[
             Dynamic[
              FEPrivate`FrontEndResource["FEStrings", "sizeExplanation"]], 
             DynamicUpdating -> True], "Tooltip"]& ], 
          ButtonBox[
           PaneSelectorBox[{False->
            StyleBox[
             StyleBox[
              DynamicBox[ToBoxes[
                FEPrivate`FrontEndResource["FEStrings", "sizeShowLess"], 
                StandardForm],
               ImageSizeCache->{50., {0., 8.}}],
              StripOnInput->False,
              DynamicUpdating->True], "OSLControl",
             StripOnInput->False], True->
            StyleBox[
             StyleBox[
              DynamicBox[ToBoxes[
                FEPrivate`FrontEndResource["FEStrings", "sizeShowLess"], 
                StandardForm]],
              StripOnInput->False,
              DynamicUpdating->True], "OSLControlActive",
             StripOnInput->False]}, Dynamic[
             CurrentValue["MouseOver"]],
            Alignment->Center,
            FrameMargins->0,
            ImageSize->{Automatic, 25}],
           Appearance->None,
           BaselinePosition->Baseline,
           
           ButtonFunction:>OutputSizeLimit`ButtonFunction[
            OutputSizeLimit`Defer, 28, 19266884935402643381, 5/2],
           Enabled->True,
           Evaluator->Automatic,
           Method->"Queued"], 
          ButtonBox[
           PaneSelectorBox[{False->
            StyleBox[
             StyleBox[
              DynamicBox[ToBoxes[
                FEPrivate`FrontEndResource["FEStrings", "sizeShowMore"], 
                StandardForm],
               ImageSizeCache->{56., {0., 8.}}],
              StripOnInput->False,
              DynamicUpdating->True], "OSLControl",
             StripOnInput->False], True->
            StyleBox[
             StyleBox[
              DynamicBox[ToBoxes[
                FEPrivate`FrontEndResource["FEStrings", "sizeShowMore"], 
                StandardForm]],
              StripOnInput->False,
              DynamicUpdating->True], "OSLControlActive",
             StripOnInput->False]}, Dynamic[
             CurrentValue["MouseOver"]],
            Alignment->Center,
            FrameMargins->0,
            ImageSize->{Automatic, 25}],
           Appearance->None,
           BaselinePosition->Baseline,
           
           ButtonFunction:>OutputSizeLimit`ButtonFunction[
            OutputSizeLimit`Defer, 28, 19266884935402643381, 5 2],
           Enabled->True,
           Evaluator->Automatic,
           Method->"Queued"], 
          ButtonBox[
           PaneSelectorBox[{False->
            StyleBox[
             StyleBox[
              DynamicBox[ToBoxes[
                FEPrivate`FrontEndResource["FEStrings", "sizeShowAll"], 
                StandardForm],
               ImageSizeCache->{42., {0., 8.}}],
              StripOnInput->False,
              DynamicUpdating->True], "OSLControl",
             StripOnInput->False], True->
            StyleBox[
             StyleBox[
              DynamicBox[ToBoxes[
                FEPrivate`FrontEndResource["FEStrings", "sizeShowAll"], 
                StandardForm]],
              StripOnInput->False,
              DynamicUpdating->True], "OSLControlActive",
             StripOnInput->False]}, Dynamic[
             CurrentValue["MouseOver"]],
            Alignment->Center,
            FrameMargins->0,
            ImageSize->{Automatic, 25}],
           Appearance->None,
           BaselinePosition->Baseline,
           
           ButtonFunction:>OutputSizeLimit`ButtonFunction[
            OutputSizeLimit`Defer, 28, 19266884935402643381, Infinity],
           Enabled->True,
           Evaluator->Automatic,
           Method->"Queued"], 
          ButtonBox[
           PaneSelectorBox[{False->
            StyleBox[
             StyleBox[
              DynamicBox[ToBoxes[
                FEPrivate`FrontEndResource["FEStrings", "sizeChangeLimit"], 
                StandardForm],
               ImageSizeCache->{74., {0., 8.}}],
              StripOnInput->False,
              DynamicUpdating->True], "OSLControl",
             StripOnInput->False], True->
            StyleBox[
             StyleBox[
              DynamicBox[ToBoxes[
                FEPrivate`FrontEndResource["FEStrings", "sizeChangeLimit"], 
                StandardForm]],
              StripOnInput->False,
              DynamicUpdating->True], "OSLControlActive",
             StripOnInput->False]}, Dynamic[
             CurrentValue["MouseOver"]],
            Alignment->Center,
            FrameMargins->0,
            ImageSize->{Automatic, 25}],
           Appearance->None,
           BaselinePosition->Baseline,
           ButtonFunction:>FrontEndExecute[{
              FrontEnd`SetOptions[
              FrontEnd`$FrontEnd, 
               FrontEnd`PreferencesSettings -> {"Page" -> "Advanced"}], 
              FrontEnd`FrontEndToken["PreferencesDialog"]}],
           Evaluator->None,
           Method->"Preemptive"]}
        },
        AutoDelete->False,
        FrameStyle->GrayLevel[0.85],
        GridBoxDividers->{"Columns" -> {False, {True}}},
        GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}},
        GridBoxSpacings->{"Columns" -> {{2}}}]}
     },
     DefaultBaseStyle->"Column",
     GridBoxAlignment->{
      "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
       "RowsIndexed" -> {}},
     GridBoxDividers->{
      "Columns" -> {{False}}, "ColumnsIndexed" -> {}, "Rows" -> {{False}}, 
       "RowsIndexed" -> {}},
     GridBoxItemSize->{
      "Columns" -> {{Automatic}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.5599999999999999]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], 
         Offset[1.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    BaseStyle->"OutputSizeLimit",
    FrameMargins->{{12, 12}, {0, 15}},
    FrameStyle->GrayLevel[0.85],
    RoundingRadius->5,
    StripOnInput->False],
   Deploy,
   DefaultBaseStyle->"Deploy"],
  If[19266884935402643381 === $SessionID, 
   Out[28], Message[
     MessageName[Syntax, "noinfoker"]]; Missing["NotAvailable"]; 
   Null]]], "Output",
 CellChangeTimes->{3.7310059110438557`*^9, 3.731007717013115*^9, 
  3.731008269411189*^9, 3.731008708340123*^9},
 CellLabel->"Out[28]=",
 CellID->1115212920,ExpressionUUID->"3a9aba09-a70d-4833-8421-4e4a0c0c047a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Length", "/@", "array"}]], "Input",
 CellChangeTimes->{{3.4584905806267157`*^9, 3.458490583643784*^9}, {
  3.731005925967264*^9, 3.731005927199164*^9}},
 CellLabel->"In[29]:=",
 CellID->1455885125,ExpressionUUID->"e22ed8d8-15c7-4d25-9bf0-c5921aeedd6e"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"140", ",", "140", ",", "140", ",", "140"}], "}"}]], "Output",
 CellChangeTimes->{3.731005928865163*^9, 3.7310077178622437`*^9, 
  3.731008271538402*^9, 3.731008708532445*^9},
 CellLabel->"Out[29]=",
 CellID->146275424,ExpressionUUID->"ffa81a00-4754-4b5f-aee3-cd892881cfb0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Replace the actual values of the tensors and simplify:", "MathCaption",
 CellChangeTimes->{{3.731006032819583*^9, 3.7310060634027452`*^9}},
 CellID->565555491,ExpressionUUID->"4aede06f-33af-4081-8929-d4bac1cad013"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"array", "//", "ToValues"}], "//", "Simplify"}]], "Input",
 CellChangeTimes->{{3.4584904206567163`*^9, 3.458490503420165*^9}, {
  3.461429836369713*^9, 3.4614298414929256`*^9}},
 CellLabel->"In[30]:=",
 CellID->926375781,ExpressionUUID->"ed466231-1bdf-4d74-9166-5ba5b3dd1c4e"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.7310059405776443`*^9, 3.731007720983687*^9, 
  3.7310082740356483`*^9, 3.7310087098725863`*^9},
 CellLabel->"Out[30]=",
 CellID->1891364216,ExpressionUUID->"a5bf000c-8c8b-49de-b1f5-15804e810755"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["4.13.5. Clean up", "Subsubsection",
 CellChangeTimes->{{3.7286771125390873`*^9, 3.7286771127222157`*^9}, {
   3.730913913338131*^9, 3.730913941103621*^9}, {3.730914255302225*^9, 
   3.730914260915913*^9}, {3.730914607546776*^9, 3.730914613449122*^9}, {
   3.730915185063478*^9, 3.730915186325802*^9}, {3.7309155101408854`*^9, 
   3.7309155279226313`*^9}, 3.730958995454458*^9, {3.730960223229556*^9, 
   3.7309602268348713`*^9}, {3.731006140857567*^9, 3.731006143391964*^9}},
 CellID->1043706123,ExpressionUUID->"487ae3fb-e0b4-4ed9-8f5b-b84c92a6c7a9"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefMetric", "[", "metricg", "]"}]], "Input",
 CellChangeTimes->{{3.730960232386859*^9, 3.73096023843542*^9}},
 CellLabel->"In[31]:=",
 CellID->922380565,ExpressionUUID->"577d2eb1-767b-423b-a8a1-c8fc8ff0004a"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "ChristoffelCDPDB"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`ChristoffelCDPDB],
  Editable->False]], "Print",
 CellChangeTimes->{3.73096023883737*^9, 3.730960545888906*^9, 
  3.7310040292164707`*^9, 3.731006152886045*^9, 3.731008751944386*^9},
 CellLabel->"During evaluation of In[31]:=",
 CellID->1372880448,ExpressionUUID->"7dcc6e35-cc92-416f-ba1c-6b307e7170b5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "CD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "covariant derivative", 
   " ", $CellContext`CD],
  Editable->False]], "Print",
 CellChangeTimes->{3.73096023883737*^9, 3.730960545888906*^9, 
  3.7310040292164707`*^9, 3.731006152886045*^9, 3.731008751961137*^9},
 CellLabel->"During evaluation of In[31]:=",
 CellID->1437386168,ExpressionUUID->"1e28365d-a6c5-42b5-bf81-29cd38e3e573"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tetrametric\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "Tetrametricg\[Dagger]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tetrametric", 
   " ", $CellContext`Tetrametricg\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{3.73096023883737*^9, 3.730960545888906*^9, 
  3.7310040292164707`*^9, 3.731006152886045*^9, 3.731008751978241*^9},
 CellLabel->"During evaluation of In[31]:=",
 CellID->1105780473,ExpressionUUID->"7969d275-c801-4273-8458-29e159a0fa15"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tetrametric\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Tetrametricg"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tetrametric", 
   " ", $CellContext`Tetrametricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.73096023883737*^9, 3.730960545888906*^9, 
  3.7310040292164707`*^9, 3.731006152886045*^9, 3.7310087519907913`*^9},
 CellLabel->"During evaluation of In[31]:=",
 CellID->528928573,ExpressionUUID->"1dd2ee7b-c23e-417b-8559-ac8fa0eb814a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "epsilonmetricg"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric tensor", 
   " ", $CellContext`epsilonmetricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.73096023883737*^9, 3.730960545888906*^9, 
  3.7310040292164707`*^9, 3.731006152886045*^9, 3.731008752003137*^9},
 CellLabel->"During evaluation of In[31]:=",
 CellID->1267711478,ExpressionUUID->"ef1c9ecf-4e2c-4b45-88c3-9f66ce715e29"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric metric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "metricg"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric metric tensor", 
   " ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.73096023883737*^9, 3.730960545888906*^9, 
  3.7310040292164707`*^9, 3.731006152886045*^9, 3.7310087520155993`*^9},
 CellLabel->"During evaluation of In[31]:=",
 CellID->1241548859,ExpressionUUID->"d214a4dc-1ac7-4265-a0d6-600d09bceeb0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"UndefConstantSymbol", "[", "Mass", "]"}], "\[IndentingNewLine]", 
 RowBox[{"UndefConstantSymbol", "[", "Rotation", "]"}], "\[IndentingNewLine]", 
 RowBox[{"UndefConstantSymbol", "[", "Charge", "]"}]}], "Input",
 CellChangeTimes->{{3.730960318146244*^9, 3.730960329864518*^9}, {
  3.731008770739306*^9, 3.731008818344096*^9}},
 CellLabel->"In[32]:=",
 CellID->744990806,ExpressionUUID->"7a2b0459-5b1b-420c-a3c4-3709aeb5e014"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefConstantSymbol", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"constant symbol\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["M",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Mass,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefConstantSymbol, ": Undefined ", "constant symbol", 
   " ", $CellContext`Mass],
  Editable->False]], "Print",
 CellChangeTimes->{3.7309603306998167`*^9, 3.730960547207429*^9, 
  3.7310040296144733`*^9, 3.731008819203187*^9},
 CellLabel->"During evaluation of In[32]:=",
 CellID->22444187,ExpressionUUID->"05b9661b-b10d-4750-b68e-af0f00965b20"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefConstantSymbol", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"constant symbol\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["a",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Rotation,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefConstantSymbol, ": Undefined ", "constant symbol", 
   " ", $CellContext`Rotation],
  Editable->False]], "Print",
 CellChangeTimes->{3.7309603306998167`*^9, 3.730960547207429*^9, 
  3.7310040296144733`*^9, 3.731008819215604*^9},
 CellLabel->"During evaluation of In[32]:=",
 CellID->1409812684,ExpressionUUID->"e492630c-59b4-4671-a0a7-0257aaad14e8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefConstantSymbol", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"constant symbol\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["Q",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Charge,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefConstantSymbol, ": Undefined ", "constant symbol", 
   " ", $CellContext`Charge],
  Editable->False]], "Print",
 CellChangeTimes->{3.7309603306998167`*^9, 3.730960547207429*^9, 
  3.7310040296144733`*^9, 3.731008819232615*^9},
 CellLabel->"During evaluation of In[32]:=",
 CellID->1534702637,ExpressionUUID->"03564a3f-1b3d-4fd9-9347-9141bd823c12"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefChart", "[", "B", "]"}]], "Input",
 CellChangeTimes->{{3.730960249970537*^9, 3.730960266978402*^9}, {
  3.7310088248718567`*^9, 3.731008825792068*^9}, {3.7310088581509733`*^9, 
  3.731008859472332*^9}},
 CellLabel->"In[35]:=",
 CellID->1735670720,ExpressionUUID->"44cf8def-0543-4cdf-9a24-aeef4013dddd"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "t"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "coordinate scalar", 
   " ", $CellContext`t],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.73100402974002*^9, 3.731008862698151*^9},
 CellLabel->"During evaluation of In[35]:=",
 CellID->42574735,ExpressionUUID->"e59d184e-219d-4962-a542-249c1295f5f5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "r"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "coordinate scalar", 
   " ", $CellContext`r],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.73100402974002*^9, 3.731008862714867*^9},
 CellLabel->"During evaluation of In[35]:=",
 CellID->441516694,ExpressionUUID->"2ee1a4cd-dc93-4feb-8544-af87076be995"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "\[Theta]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "coordinate scalar", 
   " ", $CellContext`\[Theta]],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.73100402974002*^9, 3.731008862730893*^9},
 CellLabel->"During evaluation of In[35]:=",
 CellID->1015075673,ExpressionUUID->"c2b24ac5-ef33-4f49-a7c9-068cdf0b4b2f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"coordinate scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "\[Phi]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "coordinate scalar", 
   " ", $CellContext`\[Phi]],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.73100402974002*^9, 3.731008862739972*^9},
 CellLabel->"During evaluation of In[35]:=",
 CellID->766744343,ExpressionUUID->"7c460714-d990-4947-bc7e-e0000db6c9b7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"mapping differential tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "diB"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "mapping differential tensor", " ", $CellContext`diB],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.73100402974002*^9, 3.731008862750071*^9},
 CellLabel->"During evaluation of In[35]:=",
 CellID->1770290801,ExpressionUUID->"1f8d0d4e-fc67-4b9a-95d3-187f4842ce53"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefMapping", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"inverse mapping\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "iB"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefMapping, ": Undefined ", "inverse mapping", 
   " ", $CellContext`iB],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.73100402974002*^9, 3.73100886298946*^9},
 CellLabel->"During evaluation of In[35]:=",
 CellID->2039046869,ExpressionUUID->"dc8fe303-7750-4a14-a2c8-38bf1ff1d2cd"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"mapping differential tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "dB"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "mapping differential tensor", " ", $CellContext`dB],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.73100402974002*^9, 3.731008863002747*^9},
 CellLabel->"During evaluation of In[35]:=",
 CellID->716688457,ExpressionUUID->"6fc22b14-2226-4e65-8bb2-ae9100e3c9d4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"parallel derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "PDB"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "parallel derivative", 
   " ", $CellContext`PDB],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.73100402974002*^9, 3.731008863015359*^9},
 CellLabel->"During evaluation of In[35]:=",
 CellID->147120776,ExpressionUUID->"b6f07ceb-344e-4212-8802-9c6e5e40bf07"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "etaUpB"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric +1 density",
    " ", $CellContext`etaUpB],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.73100402974002*^9, 3.731008863027986*^9},
 CellLabel->"During evaluation of In[35]:=",
 CellID->905204535,ExpressionUUID->"708c70c2-c938-4463-b2b9-c8588853fe04"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "etaDownB"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric -1 density",
    " ", $CellContext`etaDownB],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960267372757*^9, 3.7309605478696823`*^9, 
  3.73100402974002*^9, 3.731008863040379*^9},
 CellLabel->"During evaluation of In[35]:=",
 CellID->998464397,ExpressionUUID->"e719c1b2-fe93-402f-9457-5c0d49b66dd9"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M4", "]"}]], "Input",
 CellChangeTimes->{{3.730960243755797*^9, 3.7309602878662977`*^9}},
 CellLabel->"In[36]:=",
 CellID->694636180,ExpressionUUID->"7ae20ff5-279a-489c-af61-2341149aeeed"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefVBundle", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]M4",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentM4,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefVBundle, ": Undefined ", "vbundle", 
   " ", $CellContext`TangentM4],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960288299341*^9, 3.7309605493598957`*^9, 
  3.731004030540015*^9, 3.731008872955274*^9},
 CellLabel->"During evaluation of In[36]:=",
 CellID->849232217,ExpressionUUID->"70caf522-4986-4bcb-8c72-db390f134a35"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefManifold", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"manifold\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "M4"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefManifold, ": Undefined ", "manifold", 
   " ", $CellContext`M4],
  Editable->False]], "Print",
 CellChangeTimes->{3.730960288299341*^9, 3.7309605493598957`*^9, 
  3.731004030540015*^9, 3.731008872972279*^9},
 CellLabel->"During evaluation of In[36]:=",
 CellID->869553157,ExpressionUUID->"dfa57b20-5a52-4423-94f8-66e72145b017"]
}, Open  ]]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{"matrix", ",", "array", ",", "Delta", ",", "Sigma"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.730960348152565*^9, 3.730960389822296*^9}, {
  3.730960565026269*^9, 3.730960566881948*^9}, {3.731008888055463*^9, 
  3.731008905621435*^9}},
 CellLabel->"In[38]:=",
 CellID->1282535422,ExpressionUUID->"9aa015eb-d9d0-485b-bbcd-6f4e10e4fca0"]
}, Closed]]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\[Infinity]. Final comments", "Section",
 CellChangeTimes->{{3.728674059280712*^9, 
  3.72867406708753*^9}},ExpressionUUID->"c2047bc9-e93b-4a5a-88aa-\
30d128fe093b"],

Cell[TextData[{
 StyleBox["Note:", "TB"],
 " For further information about ",
 StyleBox["xCoba`", "MR"],
 ", and to be kept informed about new releases, you may contact the authors \
electronically at ",
 StyleBox["yllanes@xact.es ",
  FontFamily->"Helvetica"],
 "or ",
 StyleBox["jose@xact.es",
  FontFamily->"Helvetica"],
 ". Suggestions and comments are welcome and very much appreciated!"
}], "Text",
 CellChangeTimes->{{3.728646210871155*^9, 
  3.728646211973754*^9}},ExpressionUUID->"dcf454f5-b13e-4e4a-8dd7-\
2e8c4243eba0"],

Cell[CellGroupData[{

Cell[TextData[{
 "This is the doc file xCobaDoc.nb of version 0.8.4 of ",
 StyleBox["xCoba`", "MR"],
 ". Last update on 16 February 2020."
}], "Text",
 CellChangeTimes->{{3.790877653323028*^9, 3.7908776575052023`*^9}},
 CellID->982851045,ExpressionUUID->"74575b00-067d-46dd-b842-ff31b347101c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "xAct`xCoba`*"}]], "Input",
 CellLabel->"In[66]:=",ExpressionUUID->"16502e6b-c879-4721-8730-f2a6166014f5"],

Cell[BoxData[
 StyleBox[
  FrameBox[GridBox[{
     {
      DynamicModuleBox[{Typeset`open$$ = True}, 
       PaneSelectorBox[{False->
        ButtonBox[
         RowBox[{
          
          DynamicBox[FEPrivate`FrontEndResource[
           "FEBitmaps", "RightPointerOpener"]], " ", 
          StyleBox["xAct`xCoba`", "InformationGridGroupHeader"]}],
         Appearance->None,
         BaseStyle->"InformationGridLink",
         ButtonFunction:>FEPrivate`Set[Typeset`open$$, True],
         Evaluator->Automatic,
         Method->"Preemptive"], True->
        PaneBox[GridBox[{
           {
            ButtonBox[
             RowBox[{
              
              DynamicBox[FEPrivate`FrontEndResource[
               "FEBitmaps", "DownPointerOpener"],
               ImageSizeCache->{10., {3., 7.}}], " ", 
              StyleBox["xAct`xCoba`", "InformationGridGroupHeader"]}],
             Appearance->None,
             BaseStyle->"InformationGridLink",
             ButtonFunction:>FEPrivate`Set[Typeset`open$$, False],
             Evaluator->Automatic,
             Method->"Preemptive"]},
           {
            PaneBox[GridBox[{
               {
                ButtonBox[
                 StyleBox["AllComponentValues", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "AllComponentValues", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["CVSimplify", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "CVSimplify", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["PrecomposeBasis", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "PrecomposeBasis", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["UndefChart", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "UndefChart", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic]},
               {
                ButtonBox[
                 
                 StyleBox["AutomaticBasisContractionStart", 
                  "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "AutomaticBasisContractionStart", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["DaggerCIndex", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "DaggerCIndex", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["PrintxCobaCache", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "PrintxCobaCache", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["UnsetCMetric", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "UnsetCMetric", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic]},
               {
                ButtonBox[
                 
                 StyleBox["AutomaticBasisContractionStop", 
                  "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "AutomaticBasisContractionStop", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["DateOfValID", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "DateOfValID", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["RicciRotation", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "RicciRotation", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["ValID", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "ValID", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic]},
               {
                ButtonBox[
                 StyleBox["BasesOfVBundle", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "BasesOfVBundle", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["DefBasis", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "DefBasis", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["RiemannFromChristoffel", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "RiemannFromChristoffel", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["VBundleOfBasis", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "VBundleOfBasis", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic]},
               {
                ButtonBox[
                 StyleBox["BasisArray", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "BasisArray", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["DefChart", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "DefChart", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["RuleToSet", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "RuleToSet", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["VectorD", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "VectorD", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic]},
               {
                ButtonBox[
                 StyleBox["BasisChange", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "BasisChange", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["DeleteTensorValues", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "DeleteTensorValues", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["ScalarsOfChart", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "ScalarsOfChart", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["WithSetValues", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "WithSetValues", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic]},
               {
                ButtonBox[
                 StyleBox["BasisCollect", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "BasisCollect", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["DependenciesOfBasis", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "DependenciesOfBasis", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["SeparateBasis", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "SeparateBasis", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["xCobaCache", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "xCobaCache", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic]},
               {
                ButtonBox[
                 StyleBox["BasisColor", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "BasisColor", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["Disclaimer", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "Disclaimer", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["SetBasisChange", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "SetBasisChange", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["xCobaCacheTable", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "xCobaCacheTable", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic]},
               {
                ButtonBox[
                 StyleBox["BasisExpand", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "BasisExpand", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["Dual", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "Dual", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["SetCMetric", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "SetCMetric", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["\[DoubleStruckX]", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "\[DoubleStruckX]", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic]},
               {
                ButtonBox[
                 StyleBox["BasisOfCovD", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "BasisOfCovD", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["DummyToBasis", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "DummyToBasis", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["SetDaggerMatrix", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "SetDaggerMatrix", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 
                 StyleBox["$AutomaticCTensorMetricChange", 
                  "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "$AutomaticCTensorMetricChange", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic]},
               {
                ButtonBox[
                 StyleBox["BasisValues", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "BasisValues", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 
                 StyleBox["epsilonOrientationOfMetric", 
                  "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "epsilonOrientationOfMetric", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["SetToRule", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "SetToRule", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["$BMUseValues", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "$BMUseValues", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic]},
               {
                ButtonBox[
                 StyleBox["CBasis", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "CBasis", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["epsilonToetaDown", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "epsilonToetaDown", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["SplitBasis", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "SplitBasis", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["$CCSimplify", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "$CCSimplify", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic]},
               {
                ButtonBox[
                 StyleBox["CCovD", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "CCovD", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["epsilonToetaUp", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "epsilonToetaUp", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["TableOfComponents", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "TableOfComponents", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["$CheckZeroValue", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "$CheckZeroValue", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic]},
               {
                ButtonBox[
                 StyleBox["ChangeChart", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "ChangeChart", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["etaDown", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "etaDown", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["TangentReals", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "TangentReals", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["$CTensorRepresentation", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "$CTensorRepresentation", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic]},
               {
                ButtonBox[
                 StyleBox["ChangeComponents", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "ChangeComponents", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["etaDownToepsilon", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "etaDownToepsilon", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["TensorValIDs", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "TensorValIDs", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["$CVReplace", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "$CVReplace", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic]},
               {
                ButtonBox[
                 StyleBox["ChartColor", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "ChartColor", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["etaOrientation", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "etaOrientation", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["TensorValues", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "TensorValues", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["$CVSimplify", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "$CVSimplify", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic]},
               {
                ButtonBox[
                 StyleBox["ChartOfScalar", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "ChartOfScalar", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["etaUp", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "etaUp", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["ThreadComponentArray", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "ThreadComponentArray", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["$CVVerbose", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "$CVVerbose", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic]},
               {
                ButtonBox[
                 StyleBox["ChartsOfManifold", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "ChartsOfManifold", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["etaUpToepsilon", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "etaUpToepsilon", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["ToBasis", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "ToBasis", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["$IdentityScalarSymbol", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "$IdentityScalarSymbol", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic]},
               {
                ButtonBox[
                 StyleBox["ChristoffelFromMetric", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "ChristoffelFromMetric", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 
                 StyleBox["ExtendedCoordinateDerivatives", 
                  "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "ExtendedCoordinateDerivatives", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["ToBasisExpand", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "ToBasisExpand", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["$LargeComponentSize", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "$LargeComponentSize", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic]},
               {
                ButtonBox[
                 StyleBox["ClearxCobaCache", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "ClearxCobaCache", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["FormatBasis", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "FormatBasis", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["ToCCanonical", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "ToCCanonical", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["$LargeComponentString", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "$LargeComponentString", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic]},
               {
                ButtonBox[
                 StyleBox["CMapping", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "CMapping", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["FreeToBasis", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "FreeToBasis", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["ToCCovD", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "ToCCovD", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["$PDPostfixSymbol", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "$PDPostfixSymbol", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic]},
               {
                ButtonBox[
                 StyleBox["CNumbersOf", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "CNumbersOf", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["FromBasisExpand", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "FromBasisExpand", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["ToCTensor", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "ToCTensor", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["$PDPrefixSymbol", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "$PDPrefixSymbol", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic]},
               {
                ButtonBox[
                 StyleBox["ColorBasisForm", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "ColorBasisForm", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["HeldComponentValue", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "HeldComponentValue", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["ToCTensorOneSlot", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "ToCTensorOneSlot", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["$TUseValues", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "$TUseValues", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic]},
               {
                ButtonBox[
                 StyleBox["ComponentArray", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "ComponentArray", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["InChart", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "InChart", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["ToInChart", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "ToInChart", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["$Version", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "$Version", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic]},
               {
                ButtonBox[
                 StyleBox["ComponentValue", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "ComponentValue", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["InverseMetric", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "InverseMetric", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["ToTensorRules", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "ToTensorRules", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["$xCobaCache", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "$xCobaCache", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic]},
               {
                ButtonBox[
                 StyleBox["ContractBasis", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "ContractBasis", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["ManifoldOfChart", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "ManifoldOfChart", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["ToValues", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "ToValues", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["$xCobaCacheVerbose", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "$xCobaCacheVerbose", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic]},
               {
                ButtonBox[
                 StyleBox["Coordinate", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "Coordinate", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["MetricCompute", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "MetricCompute", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["TraceBasisDummy", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "TraceBasisDummy", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["$xTensorVersionExpected", "InformationGridButton"],
                 
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "$xTensorVersionExpected", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic]},
               {
                ButtonBox[
                 StyleBox["CTensor", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "CTensor", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["MetricInBasis", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "MetricInBasis", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["Transition", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "Transition", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], ""},
               {
                ButtonBox[
                 StyleBox["CurveTangent", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "CurveTangent", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["PDOfBasis", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "PDOfBasis", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["UndefBasis", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-b15e5932-69eb-4aaf-bfe9-42f5608059bf", {
                   "UndefBasis", "xAct`xCoba`"}, False},
                 ButtonNote->"xAct`xCoba`",
                 Evaluator->Automatic], ""}
              },
              DefaultBaseStyle->"Text",
              
              GridBoxAlignment->{
               "Columns" -> {{Left}}, "Rows" -> {{Baseline}}},
              GridBoxItemSize->{"Columns" -> {{
                   Scaled[0.2375]}}}],
             ImageMargins->{{10, 0}, {0, 2}}]}
          },
          GridBoxAlignment->{"Columns" -> {{Left}}, "Rows" -> {{Baseline}}}],
         FrameMargins->{{0, 0}, {8, 0}}]}, Dynamic[Typeset`open$$],
        ImageSize->Automatic]]}
    },
    GridBoxAlignment->{"Columns" -> {{Left}}, "Rows" -> {{Baseline}}},
    GridBoxDividers->{"ColumnsIndexed" -> {{False}}, "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.5599999999999999]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.8]}, 
        Offset[0.2]}}],
   BaseStyle->"InformationTitleFrame"], "InformationGridPlain"]], "Output",
 CellChangeTimes->{3.7908776691368837`*^9},
 CellLabel->"Out[66]=",
 CellID->1863535942,ExpressionUUID->"77f7f330-1d53-4e72-a91f-a53f445fc8f4"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Names", "[", "\"\<xAct`xCoba`*\>\"", "]"}], "//", 
  "Length"}]], "Input",
 CellChangeTimes->{{3.728753336103764*^9, 3.7287533516900187`*^9}},
 CellLabel->"In[67]:=",
 CellID->1079795302,ExpressionUUID->"05c6bb80-47db-407c-8464-bdf51e5d25a3"],

Cell[BoxData["114"], "Output",
 CellChangeTimes->{{3.728753347921014*^9, 3.7287533523779507`*^9}, 
   3.790877673862124*^9},
 CellLabel->"Out[67]=",
 CellID->1883203041,ExpressionUUID->"255a084f-d647-4497-8139-4bf66fea0871"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["1.2. Euclidean spaces (Where should this go?)", "Subsection",
 CellChangeTimes->{{3.7908777594612827`*^9, 
  3.790877765237441*^9}},ExpressionUUID->"63cba538-ce7d-4ea2-9e48-\
05b69d04f250"],

Cell[TextData[{
 "Currently the only predefined manifolds are the Euclidean spaces ",
 StyleBox["Reals[dim]", "MR"],
 " for nonnegative integer ",
 StyleBox["dim", "MR"],
 ". Its elements (\[OpenCurlyDoubleQuote]points\[CloseCurlyDoubleQuote]) are \
lists of real numbers. There are no complex manifolds in ",
 StyleBox["xTensor`", "MR"],
 " yet."
}], "Text",ExpressionUUID->"a0b443eb-98b9-4fe2-bea3-7d76e8a684e8"],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox[ButtonBox["Reals",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "Reals.nb", CharacterEncoding -> 
     "iso8859-1"], None}], "MR"],
 StyleBox["[d]\t\t", "DisplayMath"],
 StyleBox["Manifold of d-dimensional real tuples", "Text"]
}], "DefinitionBox",ExpressionUUID->"1a198fb4-af4f-4333-89ba-5465ad122fc9"],

Cell["Euclidean manifold.", "SmallText",ExpressionUUID->"744d701d-fdf5-429a-9d2b-59053d4bb4d1"]
}, Open  ]],

Cell[CellGroupData[{

Cell["This is the manifold of 3-tuples:", "MathCaption",ExpressionUUID->"c11240fa-1024-4f38-b1c9-c73d8dccddd0"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Reals", "[", "3", "]"}]], "Input",
 CellLabel->
  "In[399]:=",ExpressionUUID->"e5a482e4-b44f-47ef-86cf-bf1e1b4ce279"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   SuperscriptBox["\[DoubleStruckCapitalR]", "3"],
   ShowAutoStyles->False,
   AutoSpacing->False],
  Reals[3],
  Editable->False]], "Output",
 CellLabel->
  "Out[399]=",ExpressionUUID->"e28c4509-5b96-4c50-a7b9-bbd2d7fb3e3e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Points are considered 0-dimensional manifolds", "MathCaption",ExpressionUUID->"70466b7e-7da9-464d-877e-21d1965bda80"],

Cell[BoxData[
 RowBox[{
  RowBox[{"P", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "1"}], ",", "3", ",", "Pi"}], "}"}]}], ";"}]], "Input",
 CellLabel->
  "In[400]:=",ExpressionUUID->"90bea40b-ab5e-4d18-a81f-cade07620092"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ManifoldQ", "[", "P", "]"}]], "Input",
 CellLabel->
  "In[401]:=",ExpressionUUID->"395255ae-98d9-4a48-badd-79511ac570cf"],

Cell[BoxData["True"], "Output",
 CellLabel->
  "Out[401]=",ExpressionUUID->"da44a243-f384-42be-9d22-c52b1076ce54"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DimOfManifold", "[", "P", "]"}]], "Input",
 CellLabel->
  "In[402]:=",ExpressionUUID->"94faed76-0eb1-4960-ad06-4b086d74171b"],

Cell[BoxData["0"], "Output",
 CellLabel->
  "Out[402]=",ExpressionUUID->"0ae4b48e-26ec-4111-b69d-9b893705a5b7"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
They are submanifolds of the Euclidean space of the corresponding dimension:\
\>", "MathCaption",ExpressionUUID->"c6531a46-9c91-4613-9e9d-c35508d7ee46"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SubmanifoldQ", "[", 
  RowBox[{
   RowBox[{"Reals", "[", "3", "]"}], ",", "P"}], "]"}]], "Input",
 CellLabel->
  "In[403]:=",ExpressionUUID->"c1c1fce2-4756-440c-9fb1-ae5ddf6e4f16"],

Cell[BoxData["True"], "Output",
 CellLabel->
  "Out[403]=",ExpressionUUID->"e0885eae-b96b-4a4b-a91b-c2bfb1dae124"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SubmanifoldQ", "[", 
  RowBox[{
   RowBox[{"Reals", "[", "5", "]"}], ",", "P"}], "]"}]], "Input",
 CellLabel->
  "In[404]:=",ExpressionUUID->"a1eef84b-546e-4797-bc56-eeba76c8c561"],

Cell[BoxData["False"], "Output",
 CellLabel->
  "Out[404]=",ExpressionUUID->"2eede9cd-3050-49ed-ab3c-b1c514274054"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "The Euclidean manifolds have tangent spaces, denoted as ",
 StyleBox["TangentReals[dim]", "MR"],
 ". Its elements are vectors, and we will NOT use a notation based on lists. \
Euclidean vectors will be handled in the twin package ",
 StyleBox["xCoba`", "MR"],
 ", using the ",
 StyleBox["CTensor[list, {basis}]", "MR"],
 " construction."
}], "Text",ExpressionUUID->"f901778d-682a-459d-8459-80ad56ccf2dc"],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox[ButtonBox["TangentReals",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "TangentReals.nb", CharacterEncoding -> 
     "iso8859-1"], None}], "MR"],
 StyleBox["[d]\t\t", "DisplayMath"],
 StyleBox["Tangent bundle of the manifold of d-dimensional real tuples", 
  "Text"]
}], "DefinitionBox",ExpressionUUID->"31812a06-4c72-433c-bc8b-b0ed98a1a39b"],

Cell["Tangent bundle of an Euclidean manifold.", "SmallText",ExpressionUUID->"91508877-8b79-48d8-8847-ce29be63c7c5"]
}, Open  ]],

Cell[CellGroupData[{

Cell["This is the tangent vbundle of the Euclidean 4-space:", "MathCaption",ExpressionUUID->"e7d5da21-7545-4750-b73a-496903da954f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Tangent", "[", 
  RowBox[{"Reals", "[", "4", "]"}], "]"}]], "Input",
 CellLabel->
  "In[405]:=",ExpressionUUID->"805e3127-7679-4c87-8c48-549af65fcfec"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   SuperscriptBox["\[DoubleStruckCapitalT]\[DoubleStruckCapitalR]", "4"],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xCoba`TangentReals[4],
  Editable->False]], "Output",
 CellLabel->
  "Out[405]=",ExpressionUUID->"e3aceb00-7931-484f-ac64-f9cf19d5d200"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellLabel->
  "In[406]:=",ExpressionUUID->"b118f388-d9af-4300-b81e-e5cd16e40e63"],

Cell["TangentReals[4]", "Output",
 CellLabel->
  "Out[406]//InputForm=",ExpressionUUID->"f30a699b-d010-47c8-aaeb-\
3c072e37bab0"]
}, Open  ]]
}, Open  ]]
}, Closed]]
}, Closed]]
},
ScreenStyleEnvironment->"Brackets",
PrintingStyleEnvironment->"Printout",
WindowSize->{1099, 850},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
PrintingCopies->1,
PrintingPageRange->{Automatic, Automatic},
PageHeaders->{{Inherited, Inherited, 
   Cell[
    TextData[{
      ValueBox["FileName"]}], "Header"]}, {
   Cell[
    TextData[{
      ValueBox["FileName"]}], "Header"], Inherited, Inherited}},
PageFooters->{{Inherited, 
   Cell[
    TextData[{
     "\[Copyright]2003-2004 Jos\[EAcute] M. Mart\[IAcute]n-Garc\[IAcute]a"}], 
    "PageFooter"], Inherited}, {Inherited, 
   Cell[
    TextData[{
     "\[Copyright]2003-2004 Jos\[EAcute] M. Mart\[IAcute]n-Garc\[IAcute]a"}], 
    "PageFooter"], Inherited}},
PrintingOptions->{"Magnification"->1,
"PaperOrientation"->"Portrait",
"PaperSize"->{597.562, 842.375},
"PostScriptOutputFile":>FrontEnd`FileName[{$RootDirectory, "home", "jmm"}, 
  "example.ps", CharacterEncoding -> "iso8859-1"],
"PrintCellBrackets"->False,
"PrintMultipleHorizontalPages"->False,
"PrintRegistrationMarks"->True,
"PrintingMargins"->{{54, 54}, {108, 108}}},
ShowSelection->True,
TaggingRules->{
 "DocuToolsSettingsInternal" -> {
   "$PacletVersion" -> "0.10.2122", "$MVersion" -> "12", "$FlaggedVersion" -> 
    11.3}, ScrollPosition -> {0., 24.}, "TryRealOnly" -> False},
Magnification->1,
FrontEndVersion->"12.3 for Mac OS X x86 (64-bit) (April 13, 2021)",
StyleDefinitions->FrontEnd`FileName[{"Wolfram"}, "TutorialPageStyles.nb", 
  CharacterEncoding -> "UTF-8"],
ExpressionUUID->"b7dda2b5-80c1-4047-ad0f-c806086b74ae"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "S5.18.1"->{
  Cell[580, 22, 134, 3, 53, "Section",ExpressionUUID->"dbbba50a-4c94-44b0-a4d2-1993a97edb6b",
   CellTags->{"S5.18.1", "1.1"}],
  Cell[7378, 170, 810, 20, 58, "MathCaption",ExpressionUUID->"905e69e0-0b65-4680-a63d-32011e1e7f2e",
   CellTags->{"S5.18.1", "1.3"}],
  Cell[8213, 194, 130, 2, 27, "Input",ExpressionUUID->"230b9d8c-8eb7-4fe3-9ce6-d08664c83251",
   CellTags->"S5.18.1"],
  Cell[8368, 200, 1514, 23, 22, "Print",ExpressionUUID->"ba9e778d-7b75-4d49-badd-8ace32f2ff65",
   CellTags->"S5.18.1",
   CellID->1179299757],
  Cell[9885, 225, 1787, 31, 22, "Print",ExpressionUUID->"81c1f1fe-1d92-4100-b475-ed767052ad80",
   CellTags->"S5.18.1",
   CellID->1364437736],
  Cell[11675, 258, 1536, 23, 22, "Print",ExpressionUUID->"375f14ff-36e7-4c99-b124-e9243169232e",
   CellTags->"S5.18.1",
   CellID->1274646286],
  Cell[13214, 283, 1493, 22, 22, "Print",ExpressionUUID->"5f20e29b-4673-4992-a466-a9e4352a4854",
   CellTags->"S5.18.1",
   CellID->1830749977],
  Cell[14710, 307, 1475, 22, 22, "Print",ExpressionUUID->"29c5c7c7-d6f2-43fa-82bd-1ac7b2c2935e",
   CellTags->"S5.18.1",
   CellID->269169223],
  Cell[16188, 331, 1515, 23, 22, "Print",ExpressionUUID->"a1fdd9a4-466b-4b62-96b9-9d9112a747f6",
   CellTags->"S5.18.1",
   CellID->1326570379],
  Cell[17706, 356, 1790, 31, 22, "Print",ExpressionUUID->"d6981527-ad0a-4b72-b1de-9d2aaccab228",
   CellTags->"S5.18.1",
   CellID->493136699],
  Cell[19499, 389, 1536, 23, 22, "Print",ExpressionUUID->"605e2495-5085-4665-913a-70e8c1c4e280",
   CellTags->"S5.18.1",
   CellID->1009106320],
  Cell[21038, 414, 1515, 23, 22, "Print",ExpressionUUID->"0cc75c7a-dfed-4d24-a707-c5d025a7c427",
   CellTags->"S5.18.1",
   CellID->1371559881],
  Cell[22556, 439, 1788, 31, 22, "Print",ExpressionUUID->"6f2083df-3a6f-45b1-993a-3ba0cc722b11",
   CellTags->"S5.18.1",
   CellID->717453069],
  Cell[24347, 472, 1555, 23, 22, "Print",ExpressionUUID->"12c7c00d-d500-4221-8f3f-9b085dda45e2",
   CellTags->"S5.18.1",
   CellID->113052778],
  Cell[25905, 497, 1515, 23, 22, "Print",ExpressionUUID->"4ac41860-f22b-4a85-8cf3-1e92c262da90",
   CellTags->"S5.18.1",
   CellID->1702495898],
  Cell[27423, 522, 1671, 25, 42, "Print",ExpressionUUID->"2dcd86c4-fb1e-4819-a823-74f6e956c7b2",
   CellTags->"S5.18.1",
   CellID->776445058],
  Cell[29097, 549, 1515, 23, 22, "Print",ExpressionUUID->"76c3b7cc-ab4f-4624-a2ae-cf7a524c3985",
   CellTags->"S5.18.1",
   CellID->1601450634],
  Cell[549833, 13843, 480, 9, 54, "Section",ExpressionUUID->"9f2804d8-0164-470b-a13e-c7b68c3d7339",
   CellTags->{"S5.18.1", "1.4"},
   CellID->853073756],
  Cell[550338, 13856, 238, 4, 39, "Subsection",ExpressionUUID->"86385e57-5f81-479d-b8c0-73ad6467ad57",
   CellTags->{"S5.18.1", "1.4"}],
  Cell[550769, 13869, 592, 10, 57, "Text",ExpressionUUID->"6ea21273-ba95-4a75-a84b-683255bd3833",
   CellTags->{"S5.18.1", "1.5"}],
  Cell[565477, 14196, 465, 8, 56, "MathCaption",ExpressionUUID->"1b913b5b-14e7-4071-af65-bf81fe188bf6",
   CellTags->{"S5.18.1", "1.6"}],
  Cell[576936, 14437, 453, 9, 56, "MathCaption",ExpressionUUID->"f50dbadd-96cf-49cb-b613-f6c7131614b5",
   CellTags->{"S5.18.1", "1.7"}],
  Cell[1851873, 50220, 541, 10, 54, "Section",ExpressionUUID->"2c215748-a3b5-4b64-a43d-34271bae9190",
   CellTags->{"S5.18.1", "1.4"},
   CellID->256747671],
  Cell[2850668, 73822, 561, 10, 54, "Section",ExpressionUUID->"c70e460c-3a92-47d2-8cd3-6c82c14c37b0",
   CellTags->{"S5.18.1", "1.4"},
   CellID->1107613592],
  Cell[28969106, 655825, 633, 11, 54, "Section",ExpressionUUID->"805539ef-7108-4523-9d24-2b919bf16083",
   CellTags->{"S5.18.1", "1.4"},
   CellID->1577395714]},
 "1.1"->{
  Cell[580, 22, 134, 3, 53, "Section",ExpressionUUID->"dbbba50a-4c94-44b0-a4d2-1993a97edb6b",
   CellTags->{"S5.18.1", "1.1"}]},
 "1.3"->{
  Cell[7378, 170, 810, 20, 58, "MathCaption",ExpressionUUID->"905e69e0-0b65-4680-a63d-32011e1e7f2e",
   CellTags->{"S5.18.1", "1.3"}]},
 "1.4"->{
  Cell[549833, 13843, 480, 9, 54, "Section",ExpressionUUID->"9f2804d8-0164-470b-a13e-c7b68c3d7339",
   CellTags->{"S5.18.1", "1.4"},
   CellID->853073756],
  Cell[550338, 13856, 238, 4, 39, "Subsection",ExpressionUUID->"86385e57-5f81-479d-b8c0-73ad6467ad57",
   CellTags->{"S5.18.1", "1.4"}],
  Cell[1851873, 50220, 541, 10, 54, "Section",ExpressionUUID->"2c215748-a3b5-4b64-a43d-34271bae9190",
   CellTags->{"S5.18.1", "1.4"},
   CellID->256747671],
  Cell[2850668, 73822, 561, 10, 54, "Section",ExpressionUUID->"c70e460c-3a92-47d2-8cd3-6c82c14c37b0",
   CellTags->{"S5.18.1", "1.4"},
   CellID->1107613592],
  Cell[28969106, 655825, 633, 11, 54, "Section",ExpressionUUID->"805539ef-7108-4523-9d24-2b919bf16083",
   CellTags->{"S5.18.1", "1.4"},
   CellID->1577395714]},
 "1.5"->{
  Cell[550769, 13869, 592, 10, 57, "Text",ExpressionUUID->"6ea21273-ba95-4a75-a84b-683255bd3833",
   CellTags->{"S5.18.1", "1.5"}]},
 "1.6"->{
  Cell[565477, 14196, 465, 8, 56, "MathCaption",ExpressionUUID->"1b913b5b-14e7-4071-af65-bf81fe188bf6",
   CellTags->{"S5.18.1", "1.6"}]},
 "1.7"->{
  Cell[576936, 14437, 453, 9, 56, "MathCaption",ExpressionUUID->"f50dbadd-96cf-49cb-b613-f6c7131614b5",
   CellTags->{"S5.18.1", "1.7"}]},
 "Info73730895176-7882880"->{
  Cell[29003739, 656655, 386, 5, 67, "Print",ExpressionUUID->"37a7b42a-eb42-47c4-a943-ae80389a898c",
   CellTags->"Info73730895176-7882880",
   CellID->745676275]},
 "Info93730895182-7882880"->{
  Cell[29007396, 656769, 854, 12, 111, "Print",ExpressionUUID->"cd30eeb0-4a22-42fb-9239-de444f399b1e",
   CellTags->"Info93730895182-7882880",
   CellID->1068687520]},
 "Info103730895193-7882880"->{
  Cell[29013088, 656927, 737, 10, 111, "Print",ExpressionUUID->"0c931954-88f4-4d3a-83e2-46600457c59c",
   CellTags->"Info103730895193-7882880",
   CellID->1879930841]},
 "Info593730895309-7554244"->{
  Cell[29795883, 678083, 415, 6, 67, "Print",ExpressionUUID->"24c2fc93-e784-4bb3-aaff-e4997b69698a",
   CellTags->"Info593730895309-7554244",
   CellID->1005248081]},
 "Info653730895316-7554244"->{
  Cell[29858694, 679740, 1311, 18, 243, "Print",ExpressionUUID->"c16536ba-9f45-451f-924f-29b7afe430d3",
   CellTags->"Info653730895316-7554244",
   CellID->876129877]},
 "Info733730895317-7554244"->{
  Cell[29916459, 681218, 410, 6, 67, "Print",ExpressionUUID->"b001b759-6b67-43a6-85bf-2c61385ba6fd",
   CellTags->"Info733730895317-7554244",
   CellID->1465009172]},
 "Info883730895335-7554244"->{
  Cell[30466956, 695591, 822, 11, 133, "Print",ExpressionUUID->"c78b4dc7-d8d5-444f-8eb7-d91f40dd7a9d",
   CellTags->"Info883730895335-7554244",
   CellID->696624812]},
 "Info923730895347-7554244"->{
  Cell[30790401, 704064, 551, 8, 89, "Print",ExpressionUUID->"57169d80-85a2-4839-8e86-b95b76f73bed",
   CellTags->"Info923730895347-7554244",
   CellID->821685044]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"S5.18.1", 34263102, 799850},
 {"1.1", 34266697, 799919},
 {"1.3", 34266836, 799922},
 {"1.4", 34266982, 799925},
 {"1.5", 34267764, 799940},
 {"1.6", 34267907, 799943},
 {"1.7", 34268056, 799946},
 {"Info73730895176-7882880", 34268225, 799949},
 {"Info93730895182-7882880", 34268420, 799953},
 {"Info103730895193-7882880", 34268619, 799957},
 {"Info593730895309-7554244", 34268819, 799961},
 {"Info653730895316-7554244", 34269017, 799965},
 {"Info733730895317-7554244", 34269217, 799969},
 {"Info883730895335-7554244", 34269415, 799973},
 {"Info923730895347-7554244", 34269614, 799977}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[580, 22, 134, 3, 53, "Section",ExpressionUUID->"dbbba50a-4c94-44b0-a4d2-1993a97edb6b",
 CellTags->{"S5.18.1", "1.1"}],
Cell[717, 27, 1168, 22, 76, "Text",ExpressionUUID->"3ea432db-0842-4a82-933f-4fc79004767f"],
Cell[1888, 51, 917, 15, 76, "Text",ExpressionUUID->"0338f2fc-a904-4892-9b19-3ff6702765bf",
 CellID->981026496],
Cell[2808, 68, 695, 18, 24, "Text",ExpressionUUID->"fa9ac34b-3607-4207-92e9-185e063bce35",
 CellID->1715514340],
Cell[3506, 88, 861, 16, 57, "Text",ExpressionUUID->"6474d4f7-8d4a-4dea-b3d1-8af5c08e2840",
 CellID->2099645976],
Cell[4370, 106, 363, 7, 23, "Text",ExpressionUUID->"5f424d0a-16ab-4859-b710-5e7bb8907abd",
 CellID->711920896],
Cell[4736, 115, 257, 3, 23, "Text",ExpressionUUID->"87731305-ca71-46f2-ab1e-543f7913297c",
 CellID->519866281],
Cell[4996, 120, 321, 4, 23, "Text",ExpressionUUID->"8e592d57-865d-451c-9c9e-525017076f79",
 CellID->1569558782],
Cell[5320, 126, 216, 2, 23, "Text",ExpressionUUID->"57b046f0-6234-4e17-b5ad-ed61f3702da2",
 CellID->1902448593],
Cell[5539, 130, 928, 14, 74, "Text",ExpressionUUID->"ca3b484e-a3a6-4023-91b1-a60eef8264fe",
 CellID->166422542],
Cell[6470, 146, 643, 12, 42, "Text",ExpressionUUID->"d824ff7a-9884-45ba-bc77-126663bb22d5",
 CellID->1209417041],
Cell[CellGroupData[{
Cell[7138, 162, 215, 4, 39, "Subsection",ExpressionUUID->"c400feeb-afb2-4fb8-b75a-bc95a6dfa74c"],
Cell[CellGroupData[{
Cell[7378, 170, 810, 20, 58, "MathCaption",ExpressionUUID->"905e69e0-0b65-4680-a63d-32011e1e7f2e",
 CellTags->{"S5.18.1", "1.3"}],
Cell[CellGroupData[{
Cell[8213, 194, 130, 2, 27, "Input",ExpressionUUID->"230b9d8c-8eb7-4fe3-9ce6-d08664c83251",
 CellTags->"S5.18.1"],
Cell[CellGroupData[{
Cell[8368, 200, 1514, 23, 22, "Print",ExpressionUUID->"ba9e778d-7b75-4d49-badd-8ace32f2ff65",
 CellTags->"S5.18.1",
 CellID->1179299757],
Cell[9885, 225, 1787, 31, 22, "Print",ExpressionUUID->"81c1f1fe-1d92-4100-b475-ed767052ad80",
 CellTags->"S5.18.1",
 CellID->1364437736],
Cell[11675, 258, 1536, 23, 22, "Print",ExpressionUUID->"375f14ff-36e7-4c99-b124-e9243169232e",
 CellTags->"S5.18.1",
 CellID->1274646286],
Cell[13214, 283, 1493, 22, 22, "Print",ExpressionUUID->"5f20e29b-4673-4992-a466-a9e4352a4854",
 CellTags->"S5.18.1",
 CellID->1830749977],
Cell[14710, 307, 1475, 22, 22, "Print",ExpressionUUID->"29c5c7c7-d6f2-43fa-82bd-1ac7b2c2935e",
 CellTags->"S5.18.1",
 CellID->269169223],
Cell[16188, 331, 1515, 23, 22, "Print",ExpressionUUID->"a1fdd9a4-466b-4b62-96b9-9d9112a747f6",
 CellTags->"S5.18.1",
 CellID->1326570379],
Cell[17706, 356, 1790, 31, 22, "Print",ExpressionUUID->"d6981527-ad0a-4b72-b1de-9d2aaccab228",
 CellTags->"S5.18.1",
 CellID->493136699],
Cell[19499, 389, 1536, 23, 22, "Print",ExpressionUUID->"605e2495-5085-4665-913a-70e8c1c4e280",
 CellTags->"S5.18.1",
 CellID->1009106320],
Cell[21038, 414, 1515, 23, 22, "Print",ExpressionUUID->"0cc75c7a-dfed-4d24-a707-c5d025a7c427",
 CellTags->"S5.18.1",
 CellID->1371559881],
Cell[22556, 439, 1788, 31, 22, "Print",ExpressionUUID->"6f2083df-3a6f-45b1-993a-3ba0cc722b11",
 CellTags->"S5.18.1",
 CellID->717453069],
Cell[24347, 472, 1555, 23, 22, "Print",ExpressionUUID->"12c7c00d-d500-4221-8f3f-9b085dda45e2",
 CellTags->"S5.18.1",
 CellID->113052778],
Cell[25905, 497, 1515, 23, 22, "Print",ExpressionUUID->"4ac41860-f22b-4a85-8cf3-1e92c262da90",
 CellTags->"S5.18.1",
 CellID->1702495898],
Cell[27423, 522, 1671, 25, 42, "Print",ExpressionUUID->"2dcd86c4-fb1e-4819-a823-74f6e956c7b2",
 CellTags->"S5.18.1",
 CellID->776445058],
Cell[29097, 549, 1515, 23, 22, "Print",ExpressionUUID->"76c3b7cc-ab4f-4624-a2ae-cf7a524c3985",
 CellTags->"S5.18.1",
 CellID->1601450634]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[30673, 579, 289, 6, 41, "MathCaption",ExpressionUUID->"07b7acc4-9757-4c71-ab8a-974855d85ede",
 CellID->899063648],
Cell[30965, 587, 258, 5, 27, "Input",ExpressionUUID->"542ec53a-331d-4856-a7d7-e70cc5cb75bc",
 CellID->867791644]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[31272, 598, 365, 6, 39, "Subsection",ExpressionUUID->"1d1a3bbf-f661-4f3d-8fe4-621aa6bd670b"],
Cell[31640, 606, 374, 11, 24, "Text",ExpressionUUID->"f0fe6d4f-06bd-416b-9c68-4dbe13a5980f"],
Cell[CellGroupData[{
Cell[32039, 621, 303, 11, 41, "MathCaption",ExpressionUUID->"d73d0932-aee1-47f0-97a5-1d9b24e8bb84"],
Cell[CellGroupData[{
Cell[32367, 636, 278, 6, 27, "Input",ExpressionUUID->"2227dbb2-4f25-4016-a854-ecb257ec0cc6"],
Cell[CellGroupData[{
Cell[32670, 646, 1050, 19, 22, "Print",ExpressionUUID->"90952141-66c3-40b6-9e8b-4a485d2388ca",
 CellID->293029645],
Cell[33723, 667, 1058, 19, 22, "Print",ExpressionUUID->"ddb35440-7399-42e7-b857-c7fce7c011ee",
 CellID->769240944]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[34842, 693, 321, 8, 40, "MathCaption",ExpressionUUID->"2f4e63a1-9698-4648-be5f-74b3a9d261cf"],
Cell[CellGroupData[{
Cell[35188, 705, 334, 7, 27, "Input",ExpressionUUID->"1c61b0aa-a0ba-45a1-b437-0e5a59a551ed"],
Cell[35525, 714, 1216, 21, 22, "Print",ExpressionUUID->"9fe35c8a-fa01-4731-b10d-4e8e8ceaf3c2",
 CellID->1034147936],
Cell[36744, 737, 1236, 21, 22, "Message",ExpressionUUID->"7695401f-4718-42c0-b04f-8e56c8635877",
 CellID->1339205181],
Cell[37983, 760, 1238, 21, 22, "Message",ExpressionUUID->"e3d52473-9aff-4678-a8ae-d3522aa85672",
 CellID->245372860],
Cell[39224, 783, 1397, 24, 22, "Print",ExpressionUUID->"726ea625-8b95-43ae-9d74-519d49f6d3d6",
 CellID->813055780]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[40670, 813, 212, 6, 41, "MathCaption",ExpressionUUID->"144e099c-20ee-46c4-b579-e9eda0162447"],
Cell[CellGroupData[{
Cell[40907, 823, 524, 15, 47, "Input",ExpressionUUID->"fa47a144-19e3-4f11-8851-55e3f7cfe830"],
Cell[CellGroupData[{
Cell[41456, 842, 1236, 21, 22, "Print",ExpressionUUID->"73315d18-a979-4e26-a785-4c75b58f02e3",
 CellID->211920697],
Cell[42695, 865, 1324, 24, 22, "Print",ExpressionUUID->"10f2167a-ccaa-4478-86bd-151a6a2d36fd",
 CellID->1223267753]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[44080, 896, 230, 6, 41, "MathCaption",ExpressionUUID->"9e87e110-73d4-42a6-a9b0-9820a7c58a00"],
Cell[CellGroupData[{
Cell[44335, 906, 498, 13, 27, "Input",ExpressionUUID->"1aa374b7-92b8-4bb2-b014-cf39b8e0f0d2"],
Cell[CellGroupData[{
Cell[44858, 923, 1400, 26, 22, "Print",ExpressionUUID->"f17adc8a-1554-437a-b866-8a0058efb006",
 CellID->2123278630],
Cell[46261, 951, 1522, 28, 22, "Print",ExpressionUUID->"430e5b8f-b19e-4f46-bc80-bddc73bf21bd",
 CellID->1208172211]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[47832, 985, 264, 6, 27, "Input",ExpressionUUID->"09246897-571f-4402-a8c3-88dac122abee"],
Cell[CellGroupData[{
Cell[48121, 995, 1260, 22, 22, "Print",ExpressionUUID->"9cb4191c-52e6-4546-aa2d-de81d64e02a0",
 CellID->370703793],
Cell[49384, 1019, 1298, 22, 22, "Print",ExpressionUUID->"f5d1b321-f009-4aeb-a074-92b6b7447389",
 CellID->568589004]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[50743, 1048, 268, 8, 41, "MathCaption",ExpressionUUID->"8df2b989-60fa-483e-92c7-b9c8d36c5750"],
Cell[CellGroupData[{
Cell[51036, 1060, 242, 7, 27, "Input",ExpressionUUID->"3e510aa4-9ca5-4a69-b048-4e760800cdd1"],
Cell[51281, 1069, 2547, 68, 32, "Output",ExpressionUUID->"e0637ed9-f69f-441c-b979-9f0b0300d0da",
 CellID->1751099190]
}, Open  ]]
}, Open  ]],
Cell[53855, 1141, 685, 12, 58, "Text",ExpressionUUID->"f5748fd6-e0f0-4e93-ad14-30ba9184170e"],
Cell[CellGroupData[{
Cell[54565, 1157, 593, 13, 57, "MathCaption",ExpressionUUID->"e67021f0-f9cd-4b75-947a-e00601bd1833"],
Cell[CellGroupData[{
Cell[55183, 1174, 360, 7, 27, "Input",ExpressionUUID->"087432d1-54fc-4625-bf20-fe69ae86adb8"],
Cell[CellGroupData[{
Cell[55568, 1185, 1350, 23, 22, "Print",ExpressionUUID->"07145505-31b4-4972-8458-6a60642c1427",
 CellID->1009788355],
Cell[56921, 1210, 1448, 25, 22, "Print",ExpressionUUID->"a1676383-3aae-43cd-8644-a9771f721ad4",
 CellID->2097103891],
Cell[58372, 1237, 1550, 28, 22, "Print",ExpressionUUID->"d2d6785c-0e39-4a2b-bc5b-b3eb57923967",
 CellID->1523481473],
Cell[59925, 1267, 1599, 30, 22, "Print",ExpressionUUID->"8838fbb9-2a5c-4beb-95b7-1ec14e0063be",
 CellID->705593530],
Cell[61527, 1299, 1619, 30, 22, "Print",ExpressionUUID->"1b918961-8ade-4cf5-82cd-352e6e27a8bf",
 CellID->505759976],
Cell[63149, 1331, 1535, 28, 22, "Print",ExpressionUUID->"1769bf0c-6ad3-486d-9254-70352e1c8bb2",
 CellID->567641806],
Cell[64687, 1361, 1522, 27, 22, "Print",ExpressionUUID->"af12c497-c6c3-4f87-9f79-d34ab399bb97",
 CellID->1233473468],
Cell[66212, 1390, 1590, 30, 22, "Print",ExpressionUUID->"6baa3bcf-ed20-447d-b484-08e577adc695",
 CellID->1328932185]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[67863, 1427, 464, 7, 40, "MathCaption",ExpressionUUID->"bfbb7878-6d87-45eb-9b1f-8134d0c9e7df",
 CellID->1847620262],
Cell[CellGroupData[{
Cell[68352, 1438, 850, 24, 27, "Input",ExpressionUUID->"184ddbd8-4220-455a-bebc-cb681598f84f",
 CellID->1127027992],
Cell[69205, 1464, 3665, 99, 32, "Output",ExpressionUUID->"867453c7-31ee-48dc-8797-b58f5cef4fe1",
 CellID->1870663923]
}, Open  ]],
Cell[CellGroupData[{
Cell[72907, 1568, 313, 6, 27, "Input",ExpressionUUID->"cde85017-b1a2-4b16-af78-9a532fb54ba8",
 CellID->1216174007],
Cell[73223, 1576, 3640, 100, 32, "Output",ExpressionUUID->"773b7ba5-ac1b-4482-8aa3-a01368a66620",
 CellID->823175240]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[76912, 1682, 377, 7, 40, "MathCaption",ExpressionUUID->"6c7c536b-1f29-4165-95e7-d244a9731e07",
 CellID->1297931365],
Cell[CellGroupData[{
Cell[77314, 1693, 823, 23, 27, "Input",ExpressionUUID->"e5a12cc0-15d1-49e0-a791-2092a633badc",
 CellID->659237437],
Cell[78140, 1718, 3669, 99, 32, "Output",ExpressionUUID->"9b553423-f706-4774-919b-b3b99757f0e5",
 CellID->467570674]
}, Open  ]],
Cell[CellGroupData[{
Cell[81846, 1822, 375, 8, 27, "Input",ExpressionUUID->"08b1c38d-b280-4ba5-9542-498233de0309",
 CellID->1565089844],
Cell[82224, 1832, 3697, 100, 32, "Output",ExpressionUUID->"e113f92b-1c0b-45c5-abd3-69bddb4ade84",
 CellID->605431428]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[85970, 1938, 745, 16, 57, "MathCaption",ExpressionUUID->"30041a36-35a2-422b-b446-4df5385dfd6c",
 CellID->1988507208],
Cell[CellGroupData[{
Cell[86740, 1958, 374, 8, 27, "Input",ExpressionUUID->"080a21f4-b11e-4035-8174-656b65cd33df",
 CellID->1795804312],
Cell[87117, 1968, 921, 15, 26, "Output",ExpressionUUID->"7b18a9c0-66e5-43fe-9f12-0f3afd5acc02",
 CellID->892606450]
}, Open  ]],
Cell[CellGroupData[{
Cell[88075, 1988, 376, 8, 27, "Input",ExpressionUUID->"e38a628e-0602-41d9-af35-a7b5321a68a3",
 CellID->1945905506],
Cell[88454, 1998, 920, 15, 26, "Output",ExpressionUUID->"0bb558b1-33dd-4e0d-8955-07f90454dab9",
 CellID->993820116]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[89423, 2019, 531, 9, 57, "MathCaption",ExpressionUUID->"e3936a74-132d-44fc-b228-0fe43db46f04",
 CellID->1341226455],
Cell[CellGroupData[{
Cell[89979, 2032, 460, 11, 27, "Input",ExpressionUUID->"23d79355-2836-42ce-a4e2-308af96ae7ff",
 CellID->1937961720],
Cell[90442, 2045, 844, 13, 26, "Output",ExpressionUUID->"a2c4d7de-cf2f-4c7c-8e2a-c17c918d58e7",
 CellID->294662927]
}, Open  ]],
Cell[CellGroupData[{
Cell[91323, 2063, 432, 10, 27, "Input",ExpressionUUID->"9a072d86-7c51-46a6-a794-ea1ce43b63f9",
 CellID->1385272999],
Cell[91758, 2075, 1698, 41, 32, "Output",ExpressionUUID->"acc7e043-d07c-4b5b-8f6f-ea6ceaccec25",
 CellID->1257509919]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[93505, 2122, 343, 7, 40, "MathCaption",ExpressionUUID->"9291b8b6-e93e-46b0-889f-10bfad8faacc"],
Cell[CellGroupData[{
Cell[93873, 2133, 485, 9, 27, "Input",ExpressionUUID->"01f5cb31-2e45-4c6f-a741-9ffabb0fc8e2"],
Cell[CellGroupData[{
Cell[94383, 2146, 1293, 23, 22, "Print",ExpressionUUID->"38372027-d060-4742-b26b-76a3b75a5413",
 CellID->175165393],
Cell[95679, 2171, 1390, 25, 22, "Print",ExpressionUUID->"75914c48-15b9-42b6-8572-8aac4d274b9c",
 CellID->1952163761],
Cell[97072, 2198, 1489, 28, 22, "Print",ExpressionUUID->"f8cff853-766e-4995-9938-2f77693c6975",
 CellID->761172261],
Cell[98564, 2228, 1541, 30, 22, "Print",ExpressionUUID->"b6999010-c6fe-4aa8-859a-375782b264df",
 CellID->462105448],
Cell[100108, 2260, 1561, 30, 22, "Print",ExpressionUUID->"cf96c5f8-8290-430f-8ec7-a5aeac738910",
 CellID->1305396984],
Cell[101672, 2292, 1478, 28, 22, "Print",ExpressionUUID->"80b02b7a-4e63-4408-ada1-22c125276511",
 CellID->1114172336],
Cell[103153, 2322, 1466, 27, 22, "Print",ExpressionUUID->"25c1bc30-3e76-4558-9191-872f42012ae4",
 CellID->1984532959],
Cell[104622, 2351, 1532, 30, 22, "Print",ExpressionUUID->"bfdf68ce-aaaf-4e5a-8413-d7db42039835",
 CellID->1476920356]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[106215, 2388, 373, 7, 40, "MathCaption",ExpressionUUID->"391bcf87-e02f-46a0-ae84-04de15fddfa1",
 CellID->1984038762],
Cell[CellGroupData[{
Cell[106613, 2399, 413, 11, 27, "Input",ExpressionUUID->"f14ab5d7-76cc-4ee3-958e-22377b581561",
 CellID->1760885965],
Cell[107029, 2412, 1807, 44, 32, "Output",ExpressionUUID->"e25a6674-5c55-45a7-b6e2-a963129aa111",
 CellID->1154765805]
}, Open  ]],
Cell[CellGroupData[{
Cell[108873, 2461, 483, 12, 27, "Input",ExpressionUUID->"1a2d5ac5-51d4-43f0-888a-8404a262a17e",
 CellID->320834710],
Cell[109359, 2475, 1785, 44, 32, "Output",ExpressionUUID->"d0bed620-799c-42e4-b848-a09a51795b7f",
 CellID->1691980261]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[111193, 2525, 443, 12, 41, "MathCaption",ExpressionUUID->"b08378bc-0f01-42bb-8aa5-c2e25a22ee2e"],
Cell[CellGroupData[{
Cell[111661, 2541, 501, 10, 27, "Input",ExpressionUUID->"4116992a-0df4-4d2e-9008-e0fb7d61d8fd"],
Cell[CellGroupData[{
Cell[112187, 2555, 1273, 22, 22, "Print",ExpressionUUID->"6591214d-41a9-4dbe-b28a-16b494d0bf93",
 CellID->94044733],
Cell[113463, 2579, 1370, 24, 22, "Print",ExpressionUUID->"877fe4cc-7525-490d-8268-30a02429f64c",
 CellID->61863339],
Cell[114836, 2605, 1471, 27, 22, "Print",ExpressionUUID->"2d346f63-5728-446f-a198-415a4ad27ea8",
 CellID->355053425],
Cell[116310, 2634, 1525, 29, 22, "Print",ExpressionUUID->"fbf8c4c1-c8a7-4985-a24c-274ce27758b7",
 CellID->1673342609],
Cell[117838, 2665, 1546, 29, 22, "Print",ExpressionUUID->"a4b76c2e-1ce9-4892-8bed-93608127618b",
 CellID->423388351],
Cell[119387, 2696, 1460, 27, 22, "Print",ExpressionUUID->"08d1647f-09ff-4012-afe7-469fc346d5ca",
 CellID->1274692746],
Cell[120850, 2725, 1515, 29, 22, "Print",ExpressionUUID->"69c6d938-4cd9-4a55-ad85-28d5d297fe0c",
 CellID->464119550],
Cell[122368, 2756, 1582, 29, 22, "Print",ExpressionUUID->"19502156-09e8-4f67-a721-f2862d1f3b09",
 CellID->796670946],
Cell[123953, 2787, 1527, 30, 22, "Print",ExpressionUUID->"6253fa69-9408-4fcb-bef8-27bd65716fa3",
 CellID->88050377],
Cell[125483, 2819, 1606, 30, 22, "Print",ExpressionUUID->"4bd00293-c989-41f8-9630-c31ed377a1c2",
 CellID->246453456],
Cell[127092, 2851, 1295, 23, 22, "Print",ExpressionUUID->"b3e02462-15be-444c-a299-3476d3805d80",
 CellID->1794763576],
Cell[128390, 2876, 1415, 25, 22, "Print",ExpressionUUID->"dc918e01-1f1f-4b35-87ba-a2e99dfbfefe",
 CellID->1036116070],
Cell[129808, 2903, 1471, 26, 22, "Print",ExpressionUUID->"f1fb491a-d09d-4286-bc60-e89eff9af817",
 CellID->57894967],
Cell[131282, 2931, 1464, 27, 22, "Print",ExpressionUUID->"3fd0c586-9037-4ff6-8575-cdfb0d270a5d",
 CellID->230618278],
Cell[132749, 2960, 1521, 28, 22, "Print",ExpressionUUID->"b3e11ff3-0898-461b-bd7d-bca31dda6d13",
 CellID->1940899158]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[134309, 2993, 438, 8, 41, "Text",ExpressionUUID->"41b7ddbd-e6b9-4e91-b209-172a0ce81027",
 CellID->1031311492],
Cell[CellGroupData[{
Cell[134772, 3005, 212, 2, 40, "MathCaption",ExpressionUUID->"8cf80a25-7610-4ee8-bf5f-1a09bd97c61b",
 CellID->567493020],
Cell[CellGroupData[{
Cell[135009, 3011, 499, 11, 27, "Input",ExpressionUUID->"939b2cc8-339b-4dbf-aef5-bc4fc457a978",
 CellID->798724419],
Cell[135511, 3024, 3818, 104, 49, "Output",ExpressionUUID->"acf2015e-6d38-40d4-9196-ed53dd263eb3",
 CellID->951812572]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[139378, 3134, 214, 2, 40, "MathCaption",ExpressionUUID->"00a20cd5-14c5-4389-b2e8-c7fec37e866d",
 CellID->1224072815],
Cell[CellGroupData[{
Cell[139617, 3140, 444, 11, 27, "Input",ExpressionUUID->"0fb10a9e-1a6e-46b7-ace1-db410f868873",
 CellID->146109740],
Cell[140064, 3153, 4863, 134, 56, "Output",ExpressionUUID->"cafb6616-daab-42b6-bb4c-c077fc2b5f1a",
 CellID->407381183]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[144976, 3293, 268, 5, 40, "MathCaption",ExpressionUUID->"276f5939-64c7-4551-81d1-62b1cfe2aefe",
 CellID->1517941996],
Cell[CellGroupData[{
Cell[145269, 3302, 229, 4, 27, "Input",ExpressionUUID->"b40a2a60-a516-441e-b0b1-92132c68a089",
 CellID->873925861],
Cell[145501, 3308, 920, 15, 40, "Output",ExpressionUUID->"69738db6-24e3-4311-8947-118ab01abcf0",
 CellID->896289178]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[146470, 3329, 298, 7, 40, "MathCaption",ExpressionUUID->"893ff924-3a92-4a50-bdcb-73ecee828f56"],
Cell[CellGroupData[{
Cell[146793, 3340, 465, 12, 27, "Input",ExpressionUUID->"88eb222a-0a82-4e42-b087-30b334c8c6b3",
 CellID->495245388],
Cell[147261, 3354, 2608, 70, 32, "Output",ExpressionUUID->"ece8551b-c1c4-421a-8c1a-ff37b29d4dd9",
 CellID->29071767]
}, Open  ]],
Cell[CellGroupData[{
Cell[149906, 3429, 226, 4, 27, "Input",ExpressionUUID->"e352fa8f-f991-40f4-9f4a-696eb0edc24f",
 CellID->354917875],
Cell[150135, 3435, 1768, 43, 32, "Output",ExpressionUUID->"5b33e57d-752f-4b2f-8f8b-45257cd01ecd",
 CellID->2128758162]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[151952, 3484, 336, 6, 40, "MathCaption",ExpressionUUID->"01922e37-a050-4631-968d-54ca32a1b501",
 CellID->1926011727],
Cell[CellGroupData[{
Cell[152313, 3494, 259, 5, 27, "Input",ExpressionUUID->"2fd3c73e-8945-46f6-9f02-1723d21557c5",
 CellID->1726618588],
Cell[152575, 3501, 1127, 23, 26, "Output",ExpressionUUID->"039476bc-8c05-46fb-9df3-55717c3c1d3a",
 CellID->734337630]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[153751, 3530, 223, 2, 40, "MathCaption",ExpressionUUID->"f315b260-ffa6-404c-8aba-49421fe2cd0d",
 CellID->2040884250],
Cell[CellGroupData[{
Cell[153999, 3536, 579, 16, 27, "Input",ExpressionUUID->"b852019f-6038-4c28-98ce-8b2f112ff081",
 CellID->1532779866],
Cell[154581, 3554, 10427, 291, 32, "Output",ExpressionUUID->"0e33bec5-87a2-48d0-b3e5-c0bcca9df27e",
 CellID->1881806026]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[165057, 3851, 168, 3, 40, "MathCaption",ExpressionUUID->"31d51e36-85ae-433d-b70d-5fef2a5210bc",
 CellID->217378450],
Cell[CellGroupData[{
Cell[165250, 3858, 199, 4, 27, "Input",ExpressionUUID->"95c84cfd-f695-49df-bf33-878fbd0f1e50",
 CellID->609925603],
Cell[165452, 3864, 7909, 212, 82, "Output",ExpressionUUID->"0dd1d415-ef67-429d-a517-377fc94aee8a",
 CellID->2026104355]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[173410, 4082, 226, 3, 40, "MathCaption",ExpressionUUID->"9b6da2d7-1380-4536-96d6-bf6760d56437",
 CellID->467947572],
Cell[CellGroupData[{
Cell[173661, 4089, 306, 7, 27, "Input",ExpressionUUID->"c683d903-b92a-41aa-b2c1-fddf615ddfe1",
 CellID->39003799],
Cell[173970, 4098, 19873, 542, 82, "Output",ExpressionUUID->"b9b7c6be-1a3a-4740-8de9-5c594af361aa",
 CellID->432518983]
}, Open  ]],
Cell[CellGroupData[{
Cell[193880, 4645, 230, 4, 27, "Input",ExpressionUUID->"bc7e61e9-0384-4762-83c1-7f23e3b88429",
 CellID->1702767342],
Cell[194113, 4651, 1126, 22, 26, "Output",ExpressionUUID->"c79d527e-e5d6-4fd2-821f-4965540672c0",
 CellID->1505882069]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[195288, 4679, 438, 8, 40, "MathCaption",ExpressionUUID->"20f59a56-cf63-49ad-ba67-2f2baef4bca1",
 CellID->242311782],
Cell[CellGroupData[{
Cell[195751, 4691, 785, 17, 27, "Input",ExpressionUUID->"b64f669d-7de5-4b8d-b960-22ad52163c34",
 CellID->1584502302],
Cell[196539, 4710, 1254, 23, 40, "Output",ExpressionUUID->"94fe6a1d-0e46-4ec6-aa6c-9d9252c0296c",
 CellID->183612468]
}, Open  ]],
Cell[CellGroupData[{
Cell[197830, 4738, 219, 4, 27, "Input",ExpressionUUID->"6b084ed5-c180-42e2-9db7-0e32ed98bca7",
 CellID->1820321917],
Cell[198052, 4744, 2651, 72, 56, "Output",ExpressionUUID->"47b89554-787b-481f-9f5b-0aac11630c57",
 CellID->1733239118]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[200752, 4822, 208, 2, 40, "MathCaption",ExpressionUUID->"26b0cced-a85b-4400-ba9d-162cfaf7c110",
 CellID->632631793],
Cell[CellGroupData[{
Cell[200985, 4828, 228, 4, 27, "Input",ExpressionUUID->"1104ab5c-dc5f-4a57-b1ff-8f6b8500797e",
 CellID->1842212869],
Cell[201216, 4834, 3803, 104, 49, "Output",ExpressionUUID->"eb5a6790-4533-4e5c-9712-8d06838a7ae7",
 CellID->671775273]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[205068, 4944, 272, 6, 41, "MathCaption",ExpressionUUID->"a3116752-c77a-4c3e-a713-23368952637d",
 CellID->174587974],
Cell[CellGroupData[{
Cell[205365, 4954, 226, 4, 27, "Input",ExpressionUUID->"efe69f1c-6189-4eee-a396-338e69fd47ae",
 CellID->777404065],
Cell[205594, 4960, 896, 14, 26, "Output",ExpressionUUID->"cf8ae64f-a76c-4633-ac77-a64d36eb2d63",
 CellID->1028234363]
}, Open  ]],
Cell[CellGroupData[{
Cell[206527, 4979, 263, 6, 27, "Input",ExpressionUUID->"f7dd621f-2615-4f58-a013-ea6a4f177477",
 CellID->870075231],
Cell[206793, 4987, 5386, 148, 76, "Output",ExpressionUUID->"33403d98-cb72-4542-b224-17b6450c97fb",
 CellID->2121223330]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[212228, 5141, 233, 3, 40, "MathCaption",ExpressionUUID->"e0f55b9e-3dcd-4936-9ae4-6a3451845cfe"],
Cell[CellGroupData[{
Cell[212486, 5148, 350, 8, 27, "Input",ExpressionUUID->"d4c9e50d-7192-45ff-8f8d-7cc00c86e048"],
Cell[212839, 5158, 1783, 42, 32, "Output",ExpressionUUID->"00e4b102-90f5-4370-8a91-349e7401633d",
 CellID->1277912894]
}, Open  ]],
Cell[CellGroupData[{
Cell[214659, 5205, 381, 9, 27, "Input",ExpressionUUID->"a78b18f2-7258-4c1a-bc84-9fb076646eaf"],
Cell[215043, 5216, 1336, 21, 22, "Message",ExpressionUUID->"549e158b-4fa1-4245-9e29-d7e06fd9582d",
 CellID->407528486]
}, Open  ]],
Cell[CellGroupData[{
Cell[216416, 5242, 447, 11, 27, "Input",ExpressionUUID->"574efbd0-14d4-469e-a113-a3a4d8886874"],
Cell[216866, 5255, 1825, 45, 32, "Output",ExpressionUUID->"1b59040e-08cc-44cb-a994-4b7e08a720a6",
 CellID->1751994733]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[218740, 5306, 214, 2, 40, "MathCaption",ExpressionUUID->"63abdba6-c302-4ed1-9c3c-74bd5ee4104a",
 CellID->1417706722],
Cell[CellGroupData[{
Cell[218979, 5312, 453, 9, 27, "Input",ExpressionUUID->"faa06ec4-e845-446a-9328-1060eea2ebad",
 CellID->1090855304],
Cell[219435, 5323, 11592, 310, 49, "Output",ExpressionUUID->"5f4fec61-8038-4013-841b-cd34297ac554",
 CellID->754733583]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[231076, 5639, 271, 5, 40, "MathCaption",ExpressionUUID->"1c30ac6f-acd3-4454-b9d3-6a09f8f82b70",
 CellID->709935661],
Cell[CellGroupData[{
Cell[231372, 5648, 219, 4, 27, "Input",ExpressionUUID->"15f54362-4420-4f40-8fc2-7d59c9ba8288",
 CellID->1422333728],
Cell[231594, 5654, 11999, 263, 48, "Output",ExpressionUUID->"a8ab833a-16d1-4f97-a83a-ae7075a4d5e7",
 CellID->635845039]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[243642, 5923, 292, 6, 40, "MathCaption",ExpressionUUID->"732e82b6-b346-4d4a-8cb7-4804076352ee"],
Cell[CellGroupData[{
Cell[243959, 5933, 906, 27, 27, "Input",ExpressionUUID->"644c1d17-43de-4b30-933f-75681cb99f09"],
Cell[244868, 5962, 4475, 127, 32, "Output",ExpressionUUID->"65cd8321-03db-4867-aad4-cadea8023d4a",
 CellID->1421076894]
}, Open  ]],
Cell[CellGroupData[{
Cell[249380, 6094, 148, 2, 27, "Input",ExpressionUUID->"4023367d-aa79-4d9f-abeb-f2900830add1"],
Cell[249531, 6098, 1826, 44, 32, "Output",ExpressionUUID->"4dccd27d-c450-4d22-8572-f86837b10363",
 CellID->2075856550]
}, Open  ]],
Cell[CellGroupData[{
Cell[251394, 6147, 291, 5, 27, "Input",ExpressionUUID->"b3f1b95a-f13a-49eb-83bd-261e42316d5b"],
Cell[251688, 6154, 3489, 97, 32, "Output",ExpressionUUID->"91944c9a-d826-4f99-a471-80f305b052aa",
 CellID->875006227]
}, Open  ]],
Cell[CellGroupData[{
Cell[255214, 6256, 171, 3, 27, "Input",ExpressionUUID->"185b485b-f451-4e4c-b55c-c113a375f0c1"],
Cell[255388, 6261, 3665, 102, 32, "Output",ExpressionUUID->"32da72cf-b3aa-4c46-8928-aa1bb46437f1",
 CellID->595358560]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[259102, 6369, 236, 3, 40, "MathCaption",ExpressionUUID->"f191f0cd-83f9-4208-b040-9a8523661f0f"],
Cell[CellGroupData[{
Cell[259363, 6376, 175, 3, 27, "Input",ExpressionUUID->"9562eb05-9c3e-4c35-a176-2bcd9d36f40c"],
Cell[259541, 6381, 1834, 44, 32, "Output",ExpressionUUID->"f9e38686-cad3-4fef-9d05-3e9dcb22de92",
 CellID->2130434817]
}, Open  ]],
Cell[CellGroupData[{
Cell[261412, 6430, 184, 4, 27, "Input",ExpressionUUID->"c0e15c0e-7246-4677-83f1-5e4102bfd863"],
Cell[261599, 6436, 3478, 97, 32, "Output",ExpressionUUID->"f0fed273-4c5d-46ef-b7b7-031066f8ea49",
 CellID->2018184273]
}, Open  ]],
Cell[CellGroupData[{
Cell[265114, 6538, 323, 6, 27, "Input",ExpressionUUID->"5f27a659-edaa-4817-9ca6-f92b5b99c917"],
Cell[265440, 6546, 5830, 169, 32, "Output",ExpressionUUID->"26da2d4c-7e0a-4ead-9100-485ae7d95a59",
 CellID->2072055645]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[271319, 6721, 227, 5, 40, "MathCaption",ExpressionUUID->"600bb412-90c2-4297-af29-984398da8d42"],
Cell[CellGroupData[{
Cell[271571, 6730, 591, 16, 27, "Input",ExpressionUUID->"2b19ef62-7e0e-4584-ac0f-0a8cd5e21383"],
Cell[272165, 6748, 6265, 185, 32, "Output",ExpressionUUID->"6b47f00a-c810-4918-b610-c9100445e6a5",
 CellID->454205248]
}, Open  ]]
}, Open  ]],
Cell[278457, 6937, 340, 6, 23, "Text",ExpressionUUID->"afbccaae-ae0f-4f20-8cca-5c862826a72a",
 CellID->1541888545],
Cell[CellGroupData[{
Cell[278822, 6947, 172, 3, 40, "MathCaption",ExpressionUUID->"deb17fcf-f095-41f1-a81c-cd04b791fe2d"],
Cell[278997, 6952, 351, 10, 27, "Input",ExpressionUUID->"33350187-25c2-431b-b909-dc8b7a9fb8a4"],
Cell[CellGroupData[{
Cell[279373, 6966, 563, 16, 27, "Input",ExpressionUUID->"1b4d9a3d-c4f5-483c-be9a-93895f145913"],
Cell[CellGroupData[{
Cell[279961, 6986, 2254, 51, 28, "Print",ExpressionUUID->"f40e06a7-3aa6-4c4b-98df-2e976f11afc6",
 CellID->944573012],
Cell[282218, 7039, 2280, 52, 28, "Print",ExpressionUUID->"0ded7113-e553-4f40-9e8f-b69fc788e421",
 CellID->1239494060],
Cell[284501, 7093, 2280, 52, 28, "Print",ExpressionUUID->"b375287e-5673-4831-892e-0cdd341ceb22",
 CellID->1585290520],
Cell[286784, 7147, 3377, 83, 28, "Print",ExpressionUUID->"117862d2-3d6e-41f6-8750-2c856039f956",
 CellID->148161311],
Cell[290164, 7232, 2253, 51, 28, "Print",ExpressionUUID->"61a662ad-2b6c-403a-852d-0fc368e7de74",
 CellID->1213607104],
Cell[292420, 7285, 2277, 52, 28, "Print",ExpressionUUID->"94dccdfb-5140-42e6-8dad-f0c21f3083d8",
 CellID->294917722],
Cell[294700, 7339, 3377, 83, 28, "Print",ExpressionUUID->"745c9ebc-aecf-46b3-8510-3b5bd218c97f",
 CellID->289896378],
Cell[298080, 7424, 3378, 83, 28, "Print",ExpressionUUID->"267332f1-8507-493f-af6a-97b1e2fcec76",
 CellID->1795513650],
Cell[301461, 7509, 2254, 51, 28, "Print",ExpressionUUID->"f79d217b-b6e5-41ae-b9ba-1d1a39fb69dd",
 CellID->757707906]
}, Open  ]],
Cell[303730, 7563, 14351, 394, 32, "Output",ExpressionUUID->"bb8f41fd-7280-4b15-8402-87665ef58891",
 CellID->218749432]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[318130, 7963, 240, 4, 40, "MathCaption",ExpressionUUID->"4bdb2f65-538f-4d0c-ac66-cb62449142e9",
 CellID->1885112655],
Cell[CellGroupData[{
Cell[318395, 7971, 177, 3, 27, "Input",ExpressionUUID->"00f32037-3bee-42f0-a788-d81b78c5e3ff"],
Cell[318575, 7976, 16429, 429, 158, "Output",ExpressionUUID->"5701f1c6-5f93-4f66-8f51-66aba91a25dd",
 CellID->2010904542]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[335053, 8411, 193, 3, 40, "MathCaption",ExpressionUUID->"201596bc-06bb-4556-922e-4c117a9b7035"],
Cell[335249, 8416, 159, 3, 27, "Input",ExpressionUUID->"26b472bf-3a6f-4dd6-a1a9-016ae202193e"],
Cell[CellGroupData[{
Cell[335433, 8423, 760, 22, 27, "Input",ExpressionUUID->"a88a7724-b78c-4527-b786-30607311eee0"],
Cell[CellGroupData[{
Cell[336218, 8449, 4717, 120, 28, "Print",ExpressionUUID->"b7e58551-71c7-4f5a-95ec-1dd7aeb8cc65",
 CellID->1593342428],
Cell[340938, 8571, 4716, 120, 28, "Print",ExpressionUUID->"d8326ce1-787c-4f52-ac74-31037b577a5a",
 CellID->185109306]
}, Open  ]],
Cell[345669, 8694, 106718, 2921, 128, "Output",ExpressionUUID->"721f3b7c-2116-4398-8363-82c4c54b088e",
 CellID->1575785086]
}, Open  ]],
Cell[CellGroupData[{
Cell[452424, 11620, 513, 12, 27, "Input",ExpressionUUID->"bdde1cee-8adf-4e1d-aaf1-a71df7604a34"],
Cell[452940, 11634, 39851, 1026, 158, "Output",ExpressionUUID->"1e9cb0c5-48c4-4d68-9765-c1690ca4190e",
 CellID->50503431]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[492840, 12666, 181, 3, 40, "MathCaption",ExpressionUUID->"22be9d13-4b88-494d-afe2-0b5ea855fdca"],
Cell[CellGroupData[{
Cell[493046, 12673, 150, 2, 27, "Input",ExpressionUUID->"d394f28a-2df6-4f92-80a6-e57a1477b212"],
Cell[CellGroupData[{
Cell[493221, 12679, 1480, 28, 22, "Print",ExpressionUUID->"ba0bb955-80e9-47ee-9fec-a0d468038c04",
 CellID->345137860],
Cell[494704, 12709, 1660, 34, 22, "Print",ExpressionUUID->"e59a450a-0dc6-4238-806d-87eb4d5235d9",
 CellID->378682473],
Cell[496367, 12745, 1479, 28, 22, "Print",ExpressionUUID->"387379d4-cd80-41ec-b4f1-67f669fa4421",
 CellID->1522277650]
}, Open  ]]
}, Open  ]],
Cell[497873, 12777, 158, 3, 27, "Input",ExpressionUUID->"12dc5e47-bb9e-4b22-82f4-8c45d98d3b38"],
Cell[CellGroupData[{
Cell[498056, 12784, 356, 7, 27, "Input",ExpressionUUID->"68321ca2-7f23-414c-afb7-abf188a74b94"],
Cell[CellGroupData[{
Cell[498437, 12795, 1305, 23, 22, "Print",ExpressionUUID->"20f30a94-0934-46fd-8929-a5a2e99cdfb1",
 CellID->668658622],
Cell[499745, 12820, 1249, 22, 22, "Print",ExpressionUUID->"248387c7-e78f-4d05-a0b9-51691f210971",
 CellID->363535203],
Cell[500997, 12844, 1259, 22, 22, "Print",ExpressionUUID->"3a692f01-3260-412e-b8d1-c7b37b3a5ece",
 CellID->345181106],
Cell[502259, 12868, 1258, 22, 22, "Print",ExpressionUUID->"beaafcbc-10ea-4767-8fbe-18c113cf1607",
 CellID->1105517995],
Cell[503520, 12892, 1251, 22, 22, "Print",ExpressionUUID->"4e9ac448-47ff-41ae-bfcf-fe983ff89c68",
 CellID->412428121],
Cell[504774, 12916, 1272, 22, 22, "Print",ExpressionUUID->"02f628a7-617b-437f-b33a-242efa0f90ee",
 CellID->1749700778],
Cell[506049, 12940, 1274, 22, 22, "Print",ExpressionUUID->"2fcee940-f0a3-41db-9456-a0d0f7637d7b",
 CellID->1727397475],
Cell[507326, 12964, 1219, 22, 22, "Print",ExpressionUUID->"e1a9d93f-cb91-4b8d-b880-c6072502cebf",
 CellID->537938532],
Cell[508548, 12988, 1309, 23, 22, "Print",ExpressionUUID->"c4c06884-8f62-47df-be2f-38528d36d8a9",
 CellID->226753454],
Cell[509860, 13013, 1252, 22, 22, "Print",ExpressionUUID->"06cf6d5b-0259-4d12-b4c3-f9e9f5001c9f",
 CellID->1409311600],
Cell[511115, 13037, 1260, 22, 22, "Print",ExpressionUUID->"6d55ee9c-dee6-4f87-9475-0d996d30c3bf",
 CellID->1702918437],
Cell[512378, 13061, 1260, 22, 22, "Print",ExpressionUUID->"3aa4c10f-7e90-48df-88dd-8b44919948ef",
 CellID->1435607090],
Cell[513641, 13085, 1254, 22, 22, "Print",ExpressionUUID->"a91eee8c-adc2-438b-aeea-29b261414380",
 CellID->1269587585],
Cell[514898, 13109, 1273, 22, 22, "Print",ExpressionUUID->"1edd1ea1-3722-4b18-b71c-67793ef58a57",
 CellID->561024503],
Cell[516174, 13133, 1276, 22, 22, "Print",ExpressionUUID->"01f1e567-7657-4842-84b1-3c72151fde10",
 CellID->1685611591],
Cell[517453, 13157, 1221, 22, 22, "Print",ExpressionUUID->"c9d2e7e6-026e-4677-970f-b7095935672b",
 CellID->469738713]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[518735, 13186, 140, 1, 40, "MathCaption",ExpressionUUID->"4fbaadb2-6371-4cd5-811a-1b7897c8e955"],
Cell[CellGroupData[{
Cell[518900, 13191, 281, 5, 27, "Input",ExpressionUUID->"a590ad3f-c9d9-49c4-bf1d-7488a8bfff9d"],
Cell[CellGroupData[{
Cell[519206, 13200, 1329, 24, 22, "Print",ExpressionUUID->"58c7e838-4903-439e-9fb2-a114e17da76c",
 CellID->908849206],
Cell[520538, 13226, 1309, 24, 22, "Print",ExpressionUUID->"75e6d2cd-bc1d-4eb4-819c-bfabfc64903d",
 CellID->2124907778],
Cell[521850, 13252, 1303, 23, 22, "Print",ExpressionUUID->"fd077c86-fbf7-4501-adfc-a0a64b2aeb62",
 CellID->671975236],
Cell[523156, 13277, 1281, 23, 22, "Print",ExpressionUUID->"9a71faa4-0069-4eca-ac0a-35762dfdbd84",
 CellID->271291879],
Cell[524440, 13302, 1262, 23, 22, "Print",ExpressionUUID->"280b1ad0-faba-4ec0-9170-733c59f08bde",
 CellID->494827772],
Cell[525705, 13327, 1250, 22, 22, "Print",ExpressionUUID->"7916fa80-7af1-46ba-a398-be16b67faa9c",
 CellID->1513682820],
Cell[526958, 13351, 1256, 22, 22, "Print",ExpressionUUID->"ac404752-7bb4-45f2-bd4f-e13dce0e2894",
 CellID->1375873378],
Cell[528217, 13375, 1257, 22, 22, "Print",ExpressionUUID->"08e5e904-64cf-4446-8929-9d98c375bfe9",
 CellID->199953150],
Cell[529477, 13399, 1248, 22, 22, "Print",ExpressionUUID->"135da206-eff4-4f71-9c57-c702f6732246",
 CellID->1948168142],
Cell[530728, 13423, 1233, 22, 22, "Print",ExpressionUUID->"a335989a-acd0-4c4e-a100-aa9806c2f73b",
 CellID->178796785],
Cell[531964, 13447, 1288, 23, 22, "Print",ExpressionUUID->"aed33cd4-05ac-4471-965f-d8ba2f029330",
 CellID->707943342],
Cell[533255, 13472, 1268, 22, 22, "Print",ExpressionUUID->"a9132ed3-6fb9-47c3-932e-719611ee79da",
 CellID->1344344614],
Cell[534526, 13496, 1293, 23, 22, "Print",ExpressionUUID->"7acfa328-d399-4e22-b87f-326da21303c5",
 CellID->724477411],
Cell[535822, 13521, 1272, 22, 22, "Print",ExpressionUUID->"d781ce1f-bf37-42fc-9cae-fb17be06238d",
 CellID->59568187],
Cell[537097, 13545, 1215, 22, 22, "Print",ExpressionUUID->"f3fb8537-e3a5-4b4e-bb32-66165ebb774d",
 CellID->435139407]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[538361, 13573, 312, 7, 27, "Input",ExpressionUUID->"624217e7-ddf7-407e-b693-f792e56ca870",
 CellID->891057657],
Cell[CellGroupData[{
Cell[538698, 13584, 972, 18, 22, "Print",ExpressionUUID->"72b19dfe-b5bc-4474-a567-bcf8a3ea3795",
 CellID->457359816],
Cell[539673, 13604, 954, 18, 22, "Print",ExpressionUUID->"d70573e5-0ebd-41ec-98be-4d74fa53556a",
 CellID->1002454899],
Cell[540630, 13624, 955, 18, 22, "Print",ExpressionUUID->"95773e13-5f2d-4533-80ed-72963d3056a2",
 CellID->1250276778],
Cell[541588, 13644, 972, 18, 22, "Print",ExpressionUUID->"4a7d9b88-d68a-477f-9e07-98b40b2f9e9c",
 CellID->274321951],
Cell[542563, 13664, 955, 18, 22, "Print",ExpressionUUID->"61b7f67c-4cc6-4d40-93cb-0ad1c804ea4e",
 CellID->2029123995],
Cell[543521, 13684, 955, 18, 22, "Print",ExpressionUUID->"cb85afb9-2b5e-489e-bd3c-73d1ef5cc3ce",
 CellID->1446669605]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[544525, 13708, 238, 4, 27, "Input",ExpressionUUID->"caf0ef7e-cbbb-4142-809d-3a6f314d6837",
 CellID->1401790690],
Cell[CellGroupData[{
Cell[544788, 13716, 1127, 24, 22, "Print",ExpressionUUID->"6dfdefbf-aa77-42dd-a57a-1ab493d7024b",
 CellID->1941333954],
Cell[545918, 13742, 1080, 24, 22, "Print",ExpressionUUID->"59a72445-1349-44df-a013-dcd667e416a6",
 CellID->1280716007]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[547047, 13772, 233, 4, 27, "Input",ExpressionUUID->"014fba38-4fd9-4108-a282-a171ba8b66e5",
 CellID->1216022970],
Cell[CellGroupData[{
Cell[547305, 13780, 1108, 24, 22, "Print",ExpressionUUID->"8dc4848f-120b-4c67-8a45-21cc4bda9bed",
 CellID->2059295894],
Cell[548416, 13806, 942, 18, 22, "Print",ExpressionUUID->"b86d2204-c8b1-4aab-a275-3b6bc4ddf341",
 CellID->1739635406]
}, Open  ]]
}, Open  ]],
Cell[549385, 13828, 387, 8, 27, "Input",ExpressionUUID->"5241413d-cefc-4d83-9d9f-25c95dd98a30",
 CellID->94413090]
}, Open  ]]
}, Closed]]
}, Open  ]],
Cell[CellGroupData[{
Cell[549833, 13843, 480, 9, 54, "Section",ExpressionUUID->"9f2804d8-0164-470b-a13e-c7b68c3d7339",
 CellTags->{"S5.18.1", "1.4"},
 CellID->853073756],
Cell[CellGroupData[{
Cell[550338, 13856, 238, 4, 39, "Subsection",ExpressionUUID->"86385e57-5f81-479d-b8c0-73ad6467ad57",
 CellTags->{"S5.18.1", "1.4"}],
Cell[CellGroupData[{
Cell[550601, 13864, 165, 3, 41, "Subsubsection",ExpressionUUID->"a8c121ed-a412-44f3-a13a-a9fd7d2138af"],
Cell[550769, 13869, 592, 10, 57, "Text",ExpressionUUID->"6ea21273-ba95-4a75-a84b-683255bd3833",
 CellTags->{"S5.18.1", "1.5"}],
Cell[551364, 13881, 676, 14, 109, "DefinitionBox",ExpressionUUID->"a204ebe7-8a70-466f-b93e-31f2286ed32e"],
Cell[552043, 13897, 93, 0, 23, "Text",ExpressionUUID->"061818b3-e6ff-40f8-b9bb-0589f86c83f6"],
Cell[CellGroupData[{
Cell[552161, 13901, 252, 3, 40, "MathCaption",ExpressionUUID->"f33a8b9b-febb-4147-996a-db46f665cd76",
 CellID->1431971475],
Cell[CellGroupData[{
Cell[552438, 13908, 355, 8, 27, "Input",ExpressionUUID->"ba46731a-9074-4342-a71d-f8619fbc0b8d",
 CellID->571909325],
Cell[CellGroupData[{
Cell[552818, 13920, 923, 17, 22, "Print",ExpressionUUID->"5d07142e-6384-4562-9e1e-d025fad18b87",
 CellID->60992304],
Cell[553744, 13939, 934, 17, 22, "Print",ExpressionUUID->"0f02e7b7-61d9-469b-bf54-2a3dd697325e",
 CellID->747953709]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[554727, 13962, 281, 6, 27, "Input",ExpressionUUID->"10df15cf-8da1-4d20-86a2-afc8aa52cda7",
 CellID->96268193],
Cell[555011, 13970, 929, 16, 22, "Print",ExpressionUUID->"7d02a7aa-abf8-489c-a5aa-ab4bb15a0e9b",
 CellID->550003913]
}, Open  ]],
Cell[CellGroupData[{
Cell[555977, 13991, 449, 12, 27, "Input",ExpressionUUID->"cea71f12-8a6f-4852-9035-62fd44fd7cd3",
 CellID->1818717525],
Cell[556429, 14005, 1039, 20, 22, "Print",ExpressionUUID->"bb212a4a-ecac-4a84-9bd7-94724e98603a",
 CellID->633278035]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[557517, 14031, 226, 3, 40, "MathCaption",ExpressionUUID->"a4b54d25-5faf-45cd-a737-f5b18e4b5737",
 CellID->1079830948],
Cell[CellGroupData[{
Cell[557768, 14038, 356, 8, 27, "Input",ExpressionUUID->"9fd9142d-7e83-4041-9d92-9bddd7058dc4",
 CellID->1538122804],
Cell[558127, 14048, 1234, 21, 22, "Print",ExpressionUUID->"7bf9964c-4d03-4b5e-9c1c-06b204350078",
 CellID->151982334],
Cell[559364, 14071, 1244, 20, 22, "Message",ExpressionUUID->"323404f2-2d8c-420a-837f-915d9896e2cd",
 CellID->1009031255],
Cell[560611, 14093, 1244, 20, 22, "Message",ExpressionUUID->"dd418d3b-7032-4c51-bb3d-679661bd6bc2",
 CellID->109942426],
Cell[561858, 14115, 1416, 24, 22, "Print",ExpressionUUID->"f4166d09-5e98-4f98-b63a-fa5b279c6a5a",
 CellID->966417362]
}, Open  ]],
Cell[CellGroupData[{
Cell[563311, 14144, 285, 7, 27, "Input",ExpressionUUID->"41fee987-76df-4dd9-899d-b2e054ca23ea",
 CellID->593059525],
Cell[CellGroupData[{
Cell[563621, 14155, 878, 16, 22, "Print",ExpressionUUID->"c67147a3-a54d-464a-a700-9bb268ac1e6e",
 CellID->300424984],
Cell[564502, 14173, 914, 16, 22, "Print",ExpressionUUID->"4551fc82-a1c5-4198-9f78-b4c56e9941f5",
 CellID->508651991]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[565477, 14196, 465, 8, 56, "MathCaption",ExpressionUUID->"1b913b5b-14e7-4071-af65-bf81fe188bf6",
 CellTags->{"S5.18.1", "1.6"}],
Cell[CellGroupData[{
Cell[565967, 14208, 362, 7, 27, "Input",ExpressionUUID->"c2ec987c-3aab-44cd-bc09-7bfa61defc86"],
Cell[CellGroupData[{
Cell[566354, 14219, 1137, 20, 22, "Print",ExpressionUUID->"a8d5bf5e-654f-497c-9a04-76d29474aa88",
 CellID->575210575],
Cell[567494, 14241, 1232, 22, 22, "Print",ExpressionUUID->"eb85d954-50b7-4900-b351-f4d20e285f43",
 CellID->1740238878],
Cell[568729, 14265, 1334, 25, 22, "Print",ExpressionUUID->"03a370a9-fb76-4e76-b865-9ea0dbdb6c32",
 CellID->1615034053],
Cell[570066, 14292, 1386, 27, 22, "Print",ExpressionUUID->"adccf27a-8b56-42cb-8fee-198c0d5b1f37",
 CellID->1831514338],
Cell[571455, 14321, 1405, 27, 22, "Print",ExpressionUUID->"1a770054-12c1-453b-910d-b5ab710aab7e",
 CellID->230882668],
Cell[572863, 14350, 1322, 25, 22, "Print",ExpressionUUID->"ca77aa55-0b76-4fc8-9afc-7a6f9c6a0e0e",
 CellID->2089534594],
Cell[574188, 14377, 1308, 24, 22, "Print",ExpressionUUID->"b4441d93-bff9-4ca3-95fd-c5a3278d2942",
 CellID->1007841967],
Cell[575499, 14403, 1376, 27, 22, "Print",ExpressionUUID->"42829db3-b635-4c4b-9dab-be7eaa8913ec",
 CellID->1596415480]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[576936, 14437, 453, 9, 56, "MathCaption",ExpressionUUID->"f50dbadd-96cf-49cb-b613-f6c7131614b5",
 CellTags->{"S5.18.1", "1.7"}],
Cell[CellGroupData[{
Cell[577414, 14450, 259, 4, 27, "Input",ExpressionUUID->"64d49cf8-473b-4b55-aed2-ceadd1bec358"],
Cell[577676, 14456, 706, 11, 26, "Output",ExpressionUUID->"8842468a-5a67-4cf0-897b-27804e44b84a",
 CellID->251193709]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[578431, 14473, 353, 8, 40, "MathCaption",ExpressionUUID->"f4b5fb15-fa95-40f5-af81-04cddf29abc6"],
Cell[CellGroupData[{
Cell[578809, 14485, 540, 10, 27, "Input",ExpressionUUID->"5b5205ac-cdcd-453f-b226-1215b1249018"],
Cell[CellGroupData[{
Cell[579374, 14499, 1126, 20, 22, "Print",ExpressionUUID->"02efbdff-6e89-493d-ae2c-f05edf8adb62",
 CellID->731076079],
Cell[580503, 14521, 1226, 22, 22, "Print",ExpressionUUID->"a7f64574-17de-47b5-a533-f16476dda035",
 CellID->1441715266],
Cell[581732, 14545, 1325, 25, 22, "Print",ExpressionUUID->"6ed28012-b6c2-442e-a554-54363d70a86a",
 CellID->850486561],
Cell[583060, 14572, 1377, 27, 22, "Print",ExpressionUUID->"5b03e026-057f-46f2-9000-f2a2ad0d9e12",
 CellID->476476295],
Cell[584440, 14601, 1397, 27, 22, "Print",ExpressionUUID->"2650f2e1-4809-43fc-b8b3-9948068612ed",
 CellID->170570402],
Cell[585840, 14630, 1314, 25, 22, "Print",ExpressionUUID->"c93d6a16-03ca-4db1-a236-7c10768f8686",
 CellID->2033561316],
Cell[587157, 14657, 1299, 24, 22, "Print",ExpressionUUID->"6734124c-c4b0-46b9-a37e-9c6ff3d3c624",
 CellID->862796007],
Cell[588459, 14683, 1368, 27, 22, "Print",ExpressionUUID->"63cd069e-036f-42ff-ab7e-b11dda6e66a4",
 CellID->1202905105]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[589888, 14717, 229, 3, 40, "MathCaption",ExpressionUUID->"588ab9ea-12b6-4cd9-be95-2b53ccebebd7"],
Cell[CellGroupData[{
Cell[590142, 14724, 240, 4, 27, "Input",ExpressionUUID->"9bfb9365-41e0-450e-9e89-c458ee1bfa00"],
Cell[590385, 14730, 12731, 305, 282, "Output",ExpressionUUID->"0a613360-10e5-4875-84c5-2b42866c06b7",
 CellID->835398157]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[603165, 15041, 218, 2, 40, "MathCaption",ExpressionUUID->"2c95fb17-daae-496b-90bd-14b7cceced56",
 CellID->1062134855],
Cell[CellGroupData[{
Cell[603408, 15047, 111, 1, 27, "Input",ExpressionUUID->"284de3dd-8b5c-47e4-a943-90d6d74f0e55"],
Cell[603522, 15050, 762, 13, 26, "Output",ExpressionUUID->"0a0e6d75-ec94-4060-aaf6-6f57961a95f1",
 CellID->1057914556]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[604345, 15070, 293, 5, 41, "Subsubsection",ExpressionUUID->"7883ce9d-6f1e-4b4c-826d-d82119e15c0c"],
Cell[604641, 15077, 1916, 46, 282, "Text",ExpressionUUID->"225c0d65-6dec-4c7f-aeb8-0a7353325878"],
Cell[CellGroupData[{
Cell[606582, 15127, 198, 3, 40, "MathCaption",ExpressionUUID->"536dc41c-e7b4-4863-968d-339931e28a80"],
Cell[CellGroupData[{
Cell[606805, 15134, 156, 2, 27, "Input",ExpressionUUID->"587060d8-a9c0-4eb1-8df3-ad17ce14a4b4"],
Cell[606964, 15138, 887, 17, 26, "Output",ExpressionUUID->"d0f83860-d6e6-4dee-995d-326572fc3736",
 CellID->2021201992]
}, Open  ]],
Cell[607866, 15158, 1071, 31, 27, "Input",ExpressionUUID->"d4b0cb4a-45b6-4991-a4c1-ca8aae86f333"],
Cell[CellGroupData[{
Cell[608962, 15193, 111, 1, 27, "Input",ExpressionUUID->"73543f66-ce0c-41c3-8759-deacdeff54c5"],
Cell[609076, 15196, 735, 12, 26, "Output",ExpressionUUID->"dc41ff6e-c6f1-43cd-bef7-b553e32bd387",
 CellID->1364513160]
}, Open  ]],
Cell[CellGroupData[{
Cell[609848, 15213, 148, 2, 27, "Input",ExpressionUUID->"32dc7866-a948-4db3-8202-b25604fd70b5"],
Cell[609999, 15217, 858, 15, 26, "Output",ExpressionUUID->"66ab0f5a-1c2e-4039-b052-ea86545a4864",
 CellID->1429685618]
}, Open  ]],
Cell[CellGroupData[{
Cell[610894, 15237, 153, 2, 27, "Input",ExpressionUUID->"58c5b5e2-6c70-47e7-b949-f203cb85b8c0"],
Cell[611050, 15241, 858, 15, 26, "Output",ExpressionUUID->"5654b148-4d9f-41e5-b225-0b18652bf811",
 CellID->149829661]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[611957, 15262, 212, 4, 40, "MathCaption",ExpressionUUID->"8e17a1d0-fcfa-45a7-88e1-e8335e2d68e4"],
Cell[CellGroupData[{
Cell[612194, 15270, 154, 2, 27, "Input",ExpressionUUID->"0c9a0486-7d17-47ea-a628-990f4e38afee"],
Cell[612351, 15274, 855, 14, 26, "Output",ExpressionUUID->"6117b2df-5ad9-475d-a5f1-ed2323824370",
 CellID->440272853]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[613255, 15294, 487, 12, 58, "MathCaption",ExpressionUUID->"129c6ba2-51dd-4a92-9ef7-b79a233f7ea6"],
Cell[CellGroupData[{
Cell[613767, 15310, 154, 2, 27, "Input",ExpressionUUID->"ea5d79e8-f11c-4409-8f0f-490801d63df1"],
Cell[613924, 15314, 850, 15, 26, "Output",ExpressionUUID->"50245daf-a32c-4598-9962-e0ce436a60dd",
 CellID->1691975529]
}, Open  ]]
}, Open  ]],
Cell[614801, 15333, 276, 5, 41, "MathCaption",ExpressionUUID->"444eeeea-77d5-4d25-a3ad-21940ae1a5b2"],
Cell[CellGroupData[{
Cell[615102, 15342, 329, 7, 41, "MathCaption",ExpressionUUID->"eebf6268-a2c4-4344-8477-950d6fc8c617"],
Cell[CellGroupData[{
Cell[615456, 15353, 559, 16, 27, "Input",ExpressionUUID->"268191a3-370e-44fb-995b-3764026a42ee"],
Cell[616018, 15371, 2497, 69, 32, "Output",ExpressionUUID->"27651885-7304-4ec0-813a-a8321c7c48fc",
 CellID->148472211]
}, Open  ]]
}, Open  ]],
Cell[618542, 15444, 965, 18, 159, "Text",ExpressionUUID->"d34d2147-8c98-4d66-b59f-1f302fc13ae7"],
Cell[CellGroupData[{
Cell[619532, 15466, 211, 4, 40, "MathCaption",ExpressionUUID->"76a5a604-83dc-4878-831c-146276a3514a"],
Cell[CellGroupData[{
Cell[619768, 15474, 1416, 44, 47, "Input",ExpressionUUID->"468fa3bd-ce86-4aa7-ad63-104127e19477"],
Cell[621187, 15520, 4491, 128, 32, "Output",ExpressionUUID->"0a565396-33c1-43ce-96a1-a91ac2a6c7ac",
 CellID->842480991]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[625727, 15654, 379, 7, 56, "MathCaption",ExpressionUUID->"ddd22766-b3c0-42af-a36b-3184ffae9202"],
Cell[CellGroupData[{
Cell[626131, 15665, 228, 6, 27, "Input",ExpressionUUID->"c578bbd3-765d-4ce0-adc2-c97919d73d2b"],
Cell[626362, 15673, 1086, 18, 22, "Message",ExpressionUUID->"3cec4acc-8b23-4480-b915-4b8c9e34d618",
 CellID->585816200]
}, Open  ]],
Cell[CellGroupData[{
Cell[627485, 15696, 477, 12, 27, "Input",ExpressionUUID->"adf56673-76fc-41bd-93bf-ad8c5a2b94f4"],
Cell[CellGroupData[{
Cell[627987, 15712, 1258, 24, 22, "Print",ExpressionUUID->"20864c41-33b6-4a0f-95c9-c285e9fcee58",
 CellID->1085112334],
Cell[629248, 15738, 1318, 26, 22, "Print",ExpressionUUID->"730608cc-55a0-4046-b25d-11ea92fada14",
 CellID->1051867214],
Cell[630569, 15766, 1175, 22, 22, "Print",ExpressionUUID->"085a7d4b-6ee8-44a3-bb4f-0b7713915333",
 CellID->646883594],
Cell[631747, 15790, 1294, 25, 22, "Print",ExpressionUUID->"24824b51-6c3b-4f1c-b8ad-d70855c9d57b",
 CellID->1610343244],
Cell[633044, 15817, 1311, 26, 22, "Print",ExpressionUUID->"b2b85861-ec88-4330-a61c-904933638c6a",
 CellID->302498767],
Cell[634358, 15845, 1338, 26, 22, "Print",ExpressionUUID->"a3ca23a7-b1dd-404b-b157-bf8fc6796c80",
 CellID->1002984520],
Cell[635699, 15873, 1256, 24, 22, "Print",ExpressionUUID->"83a4a250-88ed-450e-a417-71a9d511e8b3",
 CellID->1561682337],
Cell[636958, 15899, 762, 11, 22, "Print",ExpressionUUID->"90d12eb2-3f31-40e3-ac3f-3dd78d9c7516",
 CellID->649904325],
Cell[637723, 15912, 1141, 20, 22, "Print",ExpressionUUID->"69fb9bd7-0128-4d12-8f6d-626f92afc21a",
 CellID->840161633],
Cell[638867, 15934, 768, 11, 22, "Print",ExpressionUUID->"28354677-d7f6-4b2d-bce1-2bb26931fe8c",
 CellID->891586806],
Cell[639638, 15947, 1274, 25, 22, "Print",ExpressionUUID->"64173909-59c4-4153-996e-5da798772610",
 CellID->1906043323],
Cell[640915, 15974, 1345, 26, 22, "Print",ExpressionUUID->"f46be9e9-b6fa-4e8e-81b9-9edeb1cb37f5",
 CellID->856567362],
Cell[642263, 16002, 1264, 24, 22, "Print",ExpressionUUID->"d8c0ab6f-ac98-4a67-bc2e-352c2fd26079",
 CellID->1744767293],
Cell[643530, 16028, 1150, 20, 22, "Print",ExpressionUUID->"2d333f68-4b95-4dd0-9ae8-9d758b94bf28",
 CellID->414693666],
Cell[644683, 16050, 898, 15, 22, "Print",ExpressionUUID->"d85048b7-1dbb-4b7e-bfdb-0d10114c74f3",
 CellID->1172009947],
Cell[645584, 16067, 890, 15, 22, "Print",ExpressionUUID->"be26bf36-91cf-408c-b230-e70b976bad1e",
 CellID->1251168945],
Cell[646477, 16084, 901, 15, 22, "Print",ExpressionUUID->"7f4ec9d4-004d-43b9-b4ea-e93fd7c48ff6",
 CellID->236587191],
Cell[647381, 16101, 1168, 20, 22, "Print",ExpressionUUID->"59ba8c7e-b9f9-4809-8a9e-0544a8ce1967",
 CellID->1328688540]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[648598, 16127, 782, 22, 27, "Input",ExpressionUUID->"030e90da-d725-49ed-9995-ef89684cd6c2"],
Cell[649383, 16151, 3554, 100, 32, "Output",ExpressionUUID->"6976e279-f426-4d9d-8788-c3acb4c0b61f",
 CellID->1499578617]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[652986, 16257, 202, 3, 40, "MathCaption",ExpressionUUID->"eef2cdde-1048-4cb2-bd37-94b0ab1e03a8"],
Cell[CellGroupData[{
Cell[653213, 16264, 587, 16, 27, "Input",ExpressionUUID->"08565ed1-7cb9-4ead-a740-a5523f686629"],
Cell[653803, 16282, 2548, 69, 32, "Output",ExpressionUUID->"eafae9e1-9734-41d3-a6fd-34b24cdc25fe",
 CellID->1069677556]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[656400, 16357, 260, 5, 40, "MathCaption",ExpressionUUID->"95ca4695-08b7-4804-943d-72d469c0fe01",
 CellID->1116390537],
Cell[CellGroupData[{
Cell[656685, 16366, 480, 11, 27, "Input",ExpressionUUID->"ba477352-b254-44b7-bfd8-56f7ed5f643b"],
Cell[657168, 16379, 1024, 23, 26, "Output",ExpressionUUID->"d06c8e45-c4de-4288-a3e0-739ab116cb2f",
 CellID->1181479492]
}, Open  ]],
Cell[CellGroupData[{
Cell[658229, 16407, 149, 2, 27, "Input",ExpressionUUID->"824892bd-03f6-421f-811c-8a38c1b6e85c"],
Cell[CellGroupData[{
Cell[658403, 16413, 1074, 19, 22, "Print",ExpressionUUID->"86b6cda9-7f37-4e8b-8382-f637182db9fb",
 CellID->25712005],
Cell[659480, 16434, 1103, 19, 22, "Print",ExpressionUUID->"6f0e02c4-dd25-4aaa-b37f-b0bfcdd644e5",
 CellID->497454988],
Cell[660586, 16455, 1093, 19, 22, "Print",ExpressionUUID->"8d8d962a-ce87-460b-a5d4-2fc73a144f5e",
 CellID->572145545],
Cell[661682, 16476, 1084, 19, 22, "Print",ExpressionUUID->"7ae9f1cf-7849-4987-969b-58ca16e587b3",
 CellID->1765887196],
Cell[662769, 16497, 1080, 19, 22, "Print",ExpressionUUID->"c4324fc8-77d0-49e2-9f2f-914cc234a4fe",
 CellID->1042101632],
Cell[663852, 16518, 1072, 19, 22, "Print",ExpressionUUID->"cc17e843-3040-4928-9075-5f5208017b0c",
 CellID->1872987739],
Cell[664927, 16539, 1068, 19, 22, "Print",ExpressionUUID->"36b9b1d6-7bbe-4b9c-8b47-d80a532b9b09",
 CellID->1489631647],
Cell[665998, 16560, 1087, 19, 22, "Print",ExpressionUUID->"a15af2ba-cb1c-434c-8594-3e4729bdf953",
 CellID->874734403],
Cell[667088, 16581, 1068, 19, 22, "Print",ExpressionUUID->"8abc0fef-1867-4b06-bf48-37fbfa2a04e0",
 CellID->2135547506],
Cell[668159, 16602, 1058, 19, 22, "Print",ExpressionUUID->"83e698ab-3f75-469d-826e-d588eca2e1e2",
 CellID->1252741031],
Cell[669220, 16623, 1060, 19, 22, "Print",ExpressionUUID->"1a73b07e-d5a8-4dd5-8a4d-3815091c9092",
 CellID->888832829],
Cell[670283, 16644, 1091, 19, 22, "Print",ExpressionUUID->"2b3c0178-692c-4b1f-b092-37441bde7ea6",
 CellID->717108471],
Cell[671377, 16665, 1080, 19, 22, "Print",ExpressionUUID->"2ec132e4-e810-40f8-a567-f4ec448c5324",
 CellID->763845133]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[672496, 16689, 185, 2, 23, "Text",ExpressionUUID->"45cdac70-2586-456c-8eba-efb49ce35c18",
 CellID->608198706],
Cell[CellGroupData[{
Cell[672706, 16695, 181, 3, 40, "MathCaption",ExpressionUUID->"559dd06d-60d4-449f-9b2b-128865ba1876"],
Cell[CellGroupData[{
Cell[672912, 16702, 1411, 44, 85, "Input",ExpressionUUID->"6c5caee3-0d93-45be-8241-a5e44aed8f86"],
Cell[674326, 16748, 2557, 69, 32, "Output",ExpressionUUID->"b96e7312-cc11-40be-a003-87053f5eefad",
 CellID->1166833561]
}, Open  ]]
}, Open  ]],
Cell[676910, 16821, 246, 5, 41, "MathCaption",ExpressionUUID->"ae04f47c-7779-45d2-a8ff-6c1756179d6c"],
Cell[677159, 16828, 352, 5, 58, "DefinitionBox",ExpressionUUID->"3c54f885-109b-4734-90ee-89cccb8c4639"],
Cell[CellGroupData[{
Cell[677536, 16837, 147, 3, 23, "Text",ExpressionUUID->"62a19e41-3d53-4baf-b6d8-2693966b5396"],
Cell[677686, 16842, 157, 2, 27, "Input",ExpressionUUID->"c4f766fd-2ed4-4910-97f5-65ffeb5631e6"],
Cell[CellGroupData[{
Cell[677868, 16848, 1364, 44, 66, "Input",ExpressionUUID->"2fbe94f4-4d1b-4784-a55c-b1cbdd7c364b"],
Cell[679235, 16894, 2595, 70, 32, "Output",ExpressionUUID->"c3808519-8e25-4b73-a60c-6e99495e05b8",
 CellID->2082240699]
}, Open  ]],
Cell[681845, 16967, 156, 2, 27, "Input",ExpressionUUID->"0af93778-d4a6-4e12-b277-a1c839f8ca7a"]
}, Open  ]],
Cell[CellGroupData[{
Cell[682038, 16974, 246, 4, 40, "MathCaption",ExpressionUUID->"5781a117-9b27-41d3-b919-37216dd0c452"],
Cell[CellGroupData[{
Cell[682309, 16982, 1570, 50, 85, "Input",ExpressionUUID->"c210dfb2-afb9-46f3-a745-60fff0813869"],
Cell[683882, 17034, 6184, 181, 32, "Output",ExpressionUUID->"5cd7c505-3574-49db-8ba4-46d017cedfc2",
 CellID->287309547]
}, Open  ]]
}, Open  ]],
Cell[690093, 17219, 306, 4, 56, "MathCaption",ExpressionUUID->"afc732bd-e4d0-4f84-86fc-025f87d783b6"],
Cell[690402, 17225, 242, 6, 23, "Text",ExpressionUUID->"97a4ad94-1860-4290-af3b-c6e824da2462"],
Cell[CellGroupData[{
Cell[690669, 17235, 217, 2, 40, "MathCaption",ExpressionUUID->"d496bdab-0792-4b4d-9968-c0b15f004228",
 CellID->1298301299],
Cell[690889, 17239, 342, 7, 27, "Input",ExpressionUUID->"642d0f71-619f-4f53-b444-a151fe89bc4c"],
Cell[CellGroupData[{
Cell[691256, 17250, 1144, 35, 27, "Input",ExpressionUUID->"f4b7b0ee-0d87-4a63-9e33-8eed6897eb48"],
Cell[692403, 17287, 5262, 153, 32, "Output",ExpressionUUID->"6a19051b-3782-46e5-9dea-9304defd5093",
 CellID->672679339]
}, Open  ]],
Cell[697680, 17443, 345, 7, 27, "Input",ExpressionUUID->"7a64673b-5605-4ea8-916c-bdfaed8ea27b"],
Cell[CellGroupData[{
Cell[698050, 17454, 1146, 35, 27, "Input",ExpressionUUID->"de4efe2a-1ed1-42c8-99e1-d829a5090ffc"],
Cell[699199, 17491, 5106, 149, 32, "Output",ExpressionUUID->"141fdadf-d032-42ca-bbd2-df91a85cc813",
 CellID->565320818]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[704354, 17646, 106, 0, 40, "MathCaption",ExpressionUUID->"3961005f-d594-472d-a0a8-f641e0c1f219"],
Cell[704463, 17648, 427, 9, 47, "Input",ExpressionUUID->"c2d9f1d9-d136-44b0-8428-cecbe9730339"]
}, Open  ]],
Cell[704905, 17660, 175, 2, 23, "Text",ExpressionUUID->"1955df8b-2777-4c89-abe1-1b77490b65ad",
 CellID->2035023769],
Cell[CellGroupData[{
Cell[705105, 17666, 284, 5, 40, "MathCaption",ExpressionUUID->"d2ca40a6-8be7-49b9-90cc-ebe5e6d31bfa",
 CellID->1177372917],
Cell[CellGroupData[{
Cell[705414, 17675, 166, 3, 27, "Input",ExpressionUUID->"dac3c695-c2c7-4a8a-b803-c698bd4c3fa0",
 CellID->1192292561],
Cell[705583, 17680, 676, 10, 26, "Output",ExpressionUUID->"918b136f-d475-4861-99fb-967c71819511",
 CellID->580559433]
}, Open  ]],
Cell[CellGroupData[{
Cell[706296, 17695, 165, 3, 27, "Input",ExpressionUUID->"c69d5739-836b-4e1c-a08d-588bbde19b69",
 CellID->691162570],
Cell[706464, 17700, 663, 10, 26, "Output",ExpressionUUID->"59bcfa76-d232-4b84-a09d-0dda13e0d380",
 CellID->586231575]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[707176, 17716, 203, 6, 41, "MathCaption",ExpressionUUID->"f2ae76fc-7560-46ee-a7b6-829e61f9f246",
 CellID->630267367],
Cell[CellGroupData[{
Cell[707404, 17726, 405, 10, 27, "Input",ExpressionUUID->"71005db8-5dd4-4d14-b9af-e20758159a12",
 CellID->1514290165],
Cell[707812, 17738, 692, 12, 26, "Output",ExpressionUUID->"6483f584-b440-4cb1-a3c6-23150536e660",
 CellID->1901460078]
}, Open  ]],
Cell[CellGroupData[{
Cell[708541, 17755, 412, 10, 27, "Input",ExpressionUUID->"a2eedcf1-0da5-400a-93c9-8ea6c4278b6d",
 CellID->1956465519],
Cell[708956, 17767, 1444, 38, 32, "Output",ExpressionUUID->"9605cf6b-c627-400f-9946-e6ac066aa69d",
 CellID->1729986636]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[710461, 17812, 198, 3, 41, "Subsubsection",ExpressionUUID->"ad588e59-8951-4a3e-ae93-8d65e3ea4266"],
Cell[710662, 17817, 569, 17, 42, "Text",ExpressionUUID->"0358f8a9-76cc-4ae8-b6ae-4d70b6cd4a68"],
Cell[CellGroupData[{
Cell[711256, 17838, 212, 2, 40, "MathCaption",ExpressionUUID->"363948c7-d854-4309-8b3c-4c45d486a48c",
 CellID->2018963650],
Cell[CellGroupData[{
Cell[711493, 17844, 262, 5, 27, "Input",ExpressionUUID->"116011f7-42f4-407e-b277-46c132541d53"],
Cell[711758, 17851, 610, 9, 26, "Output",ExpressionUUID->"ca4e5d1f-ed96-45b4-b670-9efd0a82ca40",
 CellID->1047717905]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[712417, 17866, 228, 5, 40, "MathCaption",ExpressionUUID->"9fcdcbff-e456-4f29-a900-bb73000f05da"],
Cell[CellGroupData[{
Cell[712670, 17875, 115, 2, 27, "Input",ExpressionUUID->"6e4346ad-9575-41b4-8d31-d5fcedfe1699"],
Cell[712788, 17879, 689, 12, 26, "Output",ExpressionUUID->"67d2a4c7-7224-45f3-a8c4-e2be128e51a6",
 CellID->423408786]
}, Open  ]],
Cell[CellGroupData[{
Cell[713514, 17896, 529, 13, 27, "Input",ExpressionUUID->"1ab94e36-e437-4575-ba98-2bcc75ddb04f"],
Cell[CellGroupData[{
Cell[714068, 17913, 1033, 18, 22, "Print",ExpressionUUID->"1fe6afb1-6fea-4550-b90b-9be5cd4a70b8",
 CellID->457973117],
Cell[715104, 17933, 1131, 20, 22, "Print",ExpressionUUID->"f04e4bf4-9f57-4464-9ced-98b50742825e",
 CellID->572786571],
Cell[716238, 17955, 1234, 23, 22, "Print",ExpressionUUID->"907c69a4-1708-4938-9db6-bfada43db1fb",
 CellID->1802192943],
Cell[717475, 17980, 1286, 25, 22, "Print",ExpressionUUID->"a2f9a11a-b4d9-4715-bc6e-a335fbba43e1",
 CellID->1347435713],
Cell[718764, 18007, 1304, 25, 22, "Print",ExpressionUUID->"f37883bf-afd3-4da4-9162-78a3bff78891",
 CellID->1136970776],
Cell[720071, 18034, 1219, 23, 22, "Print",ExpressionUUID->"2dd05365-f479-407b-9293-aa89f3c4b10b",
 CellID->758341226],
Cell[721293, 18059, 1274, 25, 22, "Print",ExpressionUUID->"89fc858c-e1d2-4591-94c7-2f7392997ed6",
 CellID->125540437],
Cell[722570, 18086, 1340, 25, 22, "Print",ExpressionUUID->"29ef0b62-c1d3-4d21-a3ca-3b01a13cad7c",
 CellID->1129183305],
Cell[723913, 18113, 1287, 26, 22, "Print",ExpressionUUID->"f4e15e2e-bfd3-4a97-89c1-303eb6ba9bb7",
 CellID->870818596],
Cell[725203, 18141, 1365, 26, 22, "Print",ExpressionUUID->"9f04e2e3-b9b3-4946-a0f0-b8ac8bfed560",
 CellID->747088667],
Cell[726571, 18169, 1056, 19, 22, "Print",ExpressionUUID->"f251092d-e7c1-460a-a962-b24da7424004",
 CellID->2129825907],
Cell[727630, 18190, 1178, 21, 22, "Print",ExpressionUUID->"a30acda1-8557-4aa1-9ef3-c2dafa6648b1",
 CellID->1726590753],
Cell[728811, 18213, 1232, 22, 22, "Print",ExpressionUUID->"06551c1b-5de5-495f-bbde-b719b93b5835",
 CellID->2011747407],
Cell[730046, 18237, 1224, 23, 22, "Print",ExpressionUUID->"a4c6dfa7-f6b5-42f5-89d3-0ec85a88d69b",
 CellID->1456131081],
Cell[731273, 18262, 1281, 24, 22, "Print",ExpressionUUID->"232a1dd5-83a2-4ef5-9b78-2969fe6ecd84",
 CellID->471437155]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[732603, 18292, 115, 2, 27, "Input",ExpressionUUID->"f096b0d0-eb4b-4c4a-90ca-80a3e343c1ed"],
Cell[732721, 18296, 709, 12, 26, "Output",ExpressionUUID->"7fb28bc8-daaa-4dee-9c4e-755017aafaf8",
 CellID->1466307460]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[733479, 18314, 148, 2, 40, "MathCaption",ExpressionUUID->"48eceb47-8d82-4491-9731-a9bb32511f7d"],
Cell[CellGroupData[{
Cell[733652, 18320, 319, 7, 27, "Input",ExpressionUUID->"8e18fc1a-a81f-41ee-b09f-83a64b62a11e"],
Cell[733974, 18329, 670, 11, 26, "Output",ExpressionUUID->"0230dce8-6f63-485a-b7b2-737fcb3524aa",
 CellID->922155191]
}, Open  ]],
Cell[CellGroupData[{
Cell[734681, 18345, 323, 7, 27, "Input",ExpressionUUID->"0b33ed36-0ed6-4b22-bfd8-24aa09f7868d"],
Cell[735007, 18354, 961, 23, 26, "Output",ExpressionUUID->"bddbd679-9d17-4954-a94c-53dcc07bedcb",
 CellID->1233431399]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[736017, 18383, 147, 2, 40, "MathCaption",ExpressionUUID->"fad86ace-6ac0-438b-90ff-51ac842864bd"],
Cell[CellGroupData[{
Cell[736189, 18389, 602, 19, 27, "Input",ExpressionUUID->"5f4077eb-97ac-4911-aeb8-548b4e7365c5"],
Cell[736794, 18410, 2535, 69, 32, "Output",ExpressionUUID->"abf35b60-cdb0-421f-bde7-d59c5d7dec6e",
 CellID->400987425]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[739378, 18485, 133, 0, 40, "MathCaption",ExpressionUUID->"aec10345-992f-4fb8-8bd2-a8516d0afd05"],
Cell[CellGroupData[{
Cell[739536, 18489, 146, 3, 27, "Input",ExpressionUUID->"e8d7e545-e76e-49fd-aeff-5c74e2b64b91"],
Cell[739685, 18494, 613, 9, 26, "Output",ExpressionUUID->"4d52fca8-9aca-40a2-bb98-9bbfe18442ad",
 CellID->862925962]
}, Open  ]],
Cell[CellGroupData[{
Cell[740335, 18508, 388, 11, 27, "Input",ExpressionUUID->"cc27bc8f-e58e-4c31-8834-5742d763e399"],
Cell[740726, 18521, 1530, 39, 32, "Output",ExpressionUUID->"ed9ba09d-085e-42cb-8b14-4b30ac14095c",
 CellID->1859674700]
}, Open  ]],
Cell[CellGroupData[{
Cell[742293, 18565, 145, 3, 27, "Input",ExpressionUUID->"e84e2d6c-619b-4f6e-aa83-69e0a2ed03d7"],
Cell[742441, 18570, 647, 9, 26, "Output",ExpressionUUID->"1647086e-e1f9-4937-bf8d-073506ad91da",
 CellID->1593238590]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[743137, 18585, 371, 10, 57, "MathCaption",ExpressionUUID->"1de26540-5893-43fa-820e-af6c3d3d8df5"],
Cell[CellGroupData[{
Cell[743533, 18599, 210, 5, 27, "Input",ExpressionUUID->"c3bdfedb-e251-433b-8db0-dabdca1732d1"],
Cell[743746, 18606, 700, 12, 26, "Output",ExpressionUUID->"a80ca744-ea41-4a25-a079-4c9d3637d6bd",
 CellID->1702226930]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[744495, 18624, 479, 11, 58, "MathCaption",ExpressionUUID->"22b71b56-9c3d-4ee4-a1c7-04b8a788662e"],
Cell[CellGroupData[{
Cell[744999, 18639, 247, 6, 27, "Input",ExpressionUUID->"8695ff25-fb91-4c86-b442-70e743784f59"],
Cell[CellGroupData[{
Cell[745271, 18649, 1023, 18, 22, "Print",ExpressionUUID->"b201662e-eebe-4b95-85a6-ec15854e888b",
 CellID->1921187312],
Cell[746297, 18669, 1121, 20, 22, "Print",ExpressionUUID->"a0fa3a78-07ed-4240-9c37-9cd7be9ba9d5",
 CellID->1978520139],
Cell[747421, 18691, 1222, 23, 22, "Print",ExpressionUUID->"0f147443-fec2-47dc-b334-fca4e7f13c3c",
 CellID->1351186025],
Cell[748646, 18716, 1276, 25, 22, "Print",ExpressionUUID->"1e0eaa49-6100-47bc-aa3c-a60e317716fe",
 CellID->1891438797],
Cell[749925, 18743, 1292, 25, 22, "Print",ExpressionUUID->"ddf3c89c-36f6-4c37-863e-a1d366313c2e",
 CellID->211434638],
Cell[751220, 18770, 1210, 23, 22, "Print",ExpressionUUID->"73e3eab5-ce14-4da6-926a-c17ef6a89f9d",
 CellID->1644008728],
Cell[752433, 18795, 1196, 22, 22, "Print",ExpressionUUID->"fe01269b-06ca-4693-bffe-6dda69c1177e",
 CellID->1330089194],
Cell[753632, 18819, 1264, 25, 22, "Print",ExpressionUUID->"f2ff2998-b67c-42f3-aec8-034a8fa6a80d",
 CellID->1651801935]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[754945, 18850, 138, 3, 27, "Input",ExpressionUUID->"33bacd2f-f679-4de7-812e-cc96471a0dd4"],
Cell[755086, 18855, 14675, 349, 112, "Output",ExpressionUUID->"70d492b0-9c07-41e9-bb22-859d844a16e4",
 CellID->1287333306]
}, Open  ]],
Cell[769776, 19207, 1153, 38, 85, "Input",ExpressionUUID->"6990f634-9fbb-423e-90c8-0366ca3959de"]
}, Open  ]],
Cell[CellGroupData[{
Cell[770966, 19250, 133, 3, 40, "MathCaption",ExpressionUUID->"2a3e7300-524a-47ae-9d03-2d4426b648ff"],
Cell[CellGroupData[{
Cell[771124, 19257, 469, 16, 27, "Input",ExpressionUUID->"5107a35c-cee4-4852-bf67-585e2e2e102e"],
Cell[771596, 19275, 2459, 67, 32, "Output",ExpressionUUID->"a7a6821f-d72a-4815-aced-d68b5a5658b7",
 CellID->1835792792]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[774104, 19348, 148, 2, 40, "MathCaption",ExpressionUUID->"c74d384d-0b63-428c-a277-971b4d88f69b"],
Cell[CellGroupData[{
Cell[774277, 19354, 254, 6, 27, "Input",ExpressionUUID->"346fa2a8-49aa-40ec-8727-9793e6fab7f1"],
Cell[CellGroupData[{
Cell[774556, 19364, 1068, 18, 22, "Print",ExpressionUUID->"28e6af69-97a6-4888-ad68-8277aba051bd",
 CellID->1706961440],
Cell[775627, 19384, 1085, 18, 22, "Print",ExpressionUUID->"014c562f-73b8-4614-b07b-5e8df844cb66",
 CellID->666881807]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[776761, 19408, 222, 6, 27, "Input",ExpressionUUID->"9ca6c30f-993a-40e0-b2e8-04159f7b92ff"],
Cell[776986, 19416, 1444, 37, 32, "Output",ExpressionUUID->"20064156-7601-419d-b885-3c07ebe18bea",
 CellID->2138518791]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[778479, 19459, 201, 3, 40, "MathCaption",ExpressionUUID->"b998c5c5-aab8-4ba3-aca9-9822439a4e2d"],
Cell[778683, 19464, 884, 29, 47, "Input",ExpressionUUID->"f29c6c77-f41d-452f-af3e-304caf504daa"],
Cell[CellGroupData[{
Cell[779592, 19497, 222, 6, 27, "Input",ExpressionUUID->"40b02f95-2a4e-48df-86a9-6cadabd90bd2"],
Cell[779817, 19505, 1504, 38, 32, "Output",ExpressionUUID->"3947039e-d028-4533-b693-5b72b9db7ef1",
 CellID->711058643]
}, Open  ]],
Cell[CellGroupData[{
Cell[781358, 19548, 167, 4, 27, "Input",ExpressionUUID->"ac129649-f83e-453c-babf-02995386158b"],
Cell[CellGroupData[{
Cell[781550, 19556, 1087, 19, 22, "Print",ExpressionUUID->"c36167c3-2d5a-42ae-8ed3-44c817f6f55f",
 CellID->1866498034],
Cell[782640, 19577, 1033, 18, 22, "Print",ExpressionUUID->"7097503f-b615-40d6-b6e3-38ca750d199a",
 CellID->1938825709],
Cell[783676, 19597, 1041, 18, 22, "Print",ExpressionUUID->"4c7f35cd-420f-4809-800a-abb39e54752e",
 CellID->2074315232],
Cell[784720, 19617, 1038, 18, 22, "Print",ExpressionUUID->"f861fcff-37b1-4048-9d3e-d264b52c6d15",
 CellID->766578826],
Cell[785761, 19637, 1031, 18, 22, "Print",ExpressionUUID->"2ad2bbb2-1fd7-466f-b734-af4e57cc52e2",
 CellID->1135930229],
Cell[786795, 19657, 1050, 18, 22, "Print",ExpressionUUID->"76106789-5b14-4fda-958e-43fc1a3cd5f2",
 CellID->449736973],
Cell[787848, 19677, 1057, 18, 22, "Print",ExpressionUUID->"f2f2eaf2-db82-4ae3-9505-fdc9d78f79dc",
 CellID->1734351418],
Cell[788908, 19697, 998, 18, 22, "Print",ExpressionUUID->"bb1a7bb2-4c97-4b61-aab9-4a3c853a3aae",
 CellID->1448741595]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[789979, 19723, 168, 3, 41, "Subsubsection",ExpressionUUID->"2293933f-3cd3-45d8-9482-8786a1dadffa"],
Cell[790150, 19728, 180, 3, 41, "DefinitionBox",ExpressionUUID->"ae30a1d2-ed84-443a-b53c-1ab209af4509"],
Cell[790333, 19733, 92, 0, 23, "Text",ExpressionUUID->"71caf07e-a74c-4e54-b6d5-a6d4918e2d7d"],
Cell[CellGroupData[{
Cell[790450, 19737, 243, 7, 41, "MathCaption",ExpressionUUID->"36f5ef7f-e782-4794-96de-a953abadd8ac"],
Cell[CellGroupData[{
Cell[790718, 19748, 322, 6, 27, "Input",ExpressionUUID->"3c419301-418c-4d4c-9a96-ce367f785028"],
Cell[791043, 19756, 3368, 95, 32, "Output",ExpressionUUID->"19a17f7a-49ba-4bf5-81ee-6f1138d02cea",
 CellID->627601471]
}, Open  ]],
Cell[CellGroupData[{
Cell[794448, 19856, 269, 6, 27, "Input",ExpressionUUID->"1a09cf95-6c5f-4394-b6e3-bab07722f3da"],
Cell[794720, 19864, 2490, 68, 32, "Output",ExpressionUUID->"bf559b56-6d36-45a5-9c5a-3f4aceada8a1",
 CellID->1805732680]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[797259, 19938, 186, 4, 40, "MathCaption",ExpressionUUID->"3dcea435-b176-45a1-886a-4e555c34ee13"],
Cell[CellGroupData[{
Cell[797470, 19946, 346, 8, 27, "Input",ExpressionUUID->"96e1e910-c066-44ab-a78c-bc9ba62e0a4c"],
Cell[797819, 19956, 18793, 530, 32, "Output",ExpressionUUID->"68815388-709c-4091-8ba1-78b25af8ae7f",
 CellID->710253356]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[816661, 20492, 142, 3, 40, "MathCaption",ExpressionUUID->"7cfbeaad-18f5-437b-8e09-4ca35d55f9ef"],
Cell[CellGroupData[{
Cell[816828, 20499, 448, 10, 27, "Input",ExpressionUUID->"b397e55b-f385-4af2-9549-8033b90016e2"],
Cell[817279, 20511, 69705, 1798, 160, "Output",ExpressionUUID->"bac0b3ea-0565-4a8d-8420-9d5c09c4e000",
 CellID->1514044266]
}, Open  ]]
}, Open  ]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[887057, 22317, 215, 4, 39, "Subsection",ExpressionUUID->"72218c58-dd15-4055-929e-32e81a9d0e94"],
Cell[CellGroupData[{
Cell[887297, 22325, 167, 3, 41, "Subsubsection",ExpressionUUID->"36e48e02-82e5-4221-87be-c7807dcd35b4"],
Cell[887467, 22330, 361, 9, 24, "Text",ExpressionUUID->"54d297ff-67ce-4972-a3f7-040388e78001"],
Cell[CellGroupData[{
Cell[887853, 22343, 187, 2, 40, "MathCaption",ExpressionUUID->"b30e1f89-0463-43bb-95c2-da41d44b9bab",
 CellID->266904421],
Cell[CellGroupData[{
Cell[888065, 22349, 738, 22, 66, "Input",ExpressionUUID->"12c43a21-58ec-4b82-9c1a-a7c05b17d05a"],
Cell[CellGroupData[{
Cell[888828, 22375, 1143, 22, 22, "Print",ExpressionUUID->"53693de6-650a-4baf-b4ee-a251cd53ac4c",
 CellID->920137883],
Cell[889974, 22399, 1095, 21, 22, "Print",ExpressionUUID->"ded9321e-8b15-4c15-ad44-83e7972f6b4a",
 CellID->727657534],
Cell[891072, 22422, 1016, 19, 22, "Print",ExpressionUUID->"4ba7f0ca-08a0-42d3-b61b-7593946a5ab2",
 CellID->1968287920],
Cell[892091, 22443, 1078, 20, 22, "Print",ExpressionUUID->"10d789d3-6d8e-4f77-8192-91530ffedbec",
 CellID->1176973052]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[893230, 22470, 206, 2, 40, "MathCaption",ExpressionUUID->"bc2c03a8-6f62-45fc-ae7a-4e54c61d7dad",
 CellID->617006311],
Cell[CellGroupData[{
Cell[893461, 22476, 1722, 56, 47, "Input",ExpressionUUID->"75f1bb86-54a6-4869-b103-46477260b76c"],
Cell[895186, 22534, 8170, 241, 32, "Output",ExpressionUUID->"fce32b15-e43a-46a0-9615-96117e84f21d",
 CellID->914212447]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[903405, 22781, 242, 4, 40, "MathCaption",ExpressionUUID->"fd814381-e900-4682-a039-e257f3076cd5",
 CellID->1719940314],
Cell[CellGroupData[{
Cell[903672, 22789, 179, 4, 27, "Input",ExpressionUUID->"563f9246-0de7-43ec-be51-90801d6f26e9"],
Cell[903854, 22795, 948, 23, 26, "Output",ExpressionUUID->"97a1ee93-2ef3-4e72-80c9-58a69a31ff16",
 CellID->2070586091]
}, Open  ]],
Cell[CellGroupData[{
Cell[904839, 22823, 184, 4, 27, "Input",ExpressionUUID->"62e747e6-e704-464b-bb1f-b730bf6e975b"],
Cell[905026, 22829, 1828, 48, 26, "Output",ExpressionUUID->"95bd047b-053c-4fcf-ac10-2ff84c0d45ed",
 CellID->367931802]
}, Open  ]],
Cell[CellGroupData[{
Cell[906891, 22882, 181, 4, 27, "Input",ExpressionUUID->"59e57e41-bcdb-4c5a-9aa8-72cd878ffcb4"],
Cell[907075, 22888, 1281, 32, 26, "Output",ExpressionUUID->"4c94c0b5-6bbd-4c47-a71e-2d47ccce7641",
 CellID->1224376501]
}, Open  ]],
Cell[CellGroupData[{
Cell[908393, 22925, 203, 5, 27, "Input",ExpressionUUID->"f28ab0e5-e513-4521-b6db-96ee845e594e"],
Cell[908599, 22932, 1172, 29, 26, "Output",ExpressionUUID->"1a11710d-b6d0-43d4-b395-a359a7a706c1",
 CellID->907546753]
}, Open  ]],
Cell[CellGroupData[{
Cell[909808, 22966, 203, 5, 27, "Input",ExpressionUUID->"6696e00e-8955-4351-bcf8-101b3ef78416"],
Cell[910014, 22973, 887, 21, 26, "Output",ExpressionUUID->"2e00e1a5-bb06-4b0d-a32d-34cd52619aa7",
 CellID->1709416677]
}, Open  ]],
Cell[CellGroupData[{
Cell[910938, 22999, 228, 6, 27, "Input",ExpressionUUID->"571c8d35-a5c3-4086-8f79-78a9067155db"],
Cell[911169, 23007, 1312, 32, 26, "Output",ExpressionUUID->"3d57c3fe-bf7f-4b17-8abd-bba27b51330c",
 CellID->1129860773]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[912530, 23045, 358, 10, 41, "MathCaption",ExpressionUUID->"8fd8c5cb-8915-4e3c-8568-8993147aff59"],
Cell[CellGroupData[{
Cell[912913, 23059, 238, 7, 27, "Input",ExpressionUUID->"5011d870-ae08-4b36-b633-1db8d8650a02"],
Cell[913154, 23068, 1742, 46, 26, "Output",ExpressionUUID->"abbd73f1-d10a-435d-a708-cbeaa6fc6875",
 CellID->1519605037]
}, Open  ]],
Cell[CellGroupData[{
Cell[914933, 23119, 294, 6, 27, "Input",ExpressionUUID->"fe18a01b-35ef-4ad2-b5e0-e0ca9eb665c1"],
Cell[915230, 23127, 1321, 35, 26, "Output",ExpressionUUID->"9e8a4101-d72e-4af7-9de9-7de849439cb2",
 CellID->2126326735]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[916600, 23168, 112, 0, 40, "MathCaption",ExpressionUUID->"02557eb9-26fd-4a8a-b9f5-6532564d2409"],
Cell[CellGroupData[{
Cell[916737, 23172, 323, 7, 27, "Input",ExpressionUUID->"dc0547a7-f11c-4b82-922a-ee1c47f03739"],
Cell[917063, 23181, 1151, 30, 26, "Output",ExpressionUUID->"06c35b40-d3f0-45f3-98f1-d04cd820da48",
 CellID->898065418]
}, Open  ]],
Cell[CellGroupData[{
Cell[918251, 23216, 331, 7, 27, "Input",ExpressionUUID->"277857e2-9a66-4a12-8f05-5ddea2f68d9d"],
Cell[918585, 23225, 1265, 32, 26, "Output",ExpressionUUID->"0b50deb8-c36b-4b3b-85b3-7ad2f4a5934c",
 CellID->1270409210]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[919899, 23263, 381, 7, 57, "MathCaption",ExpressionUUID->"8416f90a-b65f-49bd-ba1a-fa0b3e1edeaf"],
Cell[CellGroupData[{
Cell[920305, 23274, 558, 15, 27, "Input",ExpressionUUID->"018e1f6f-3e95-40bc-ade0-65018c63ae4b"],
Cell[920866, 23291, 1173, 30, 26, "Output",ExpressionUUID->"a40e4244-2a8e-4398-8b7d-bd2eaaebb36b",
 CellID->1485173996]
}, Open  ]]
}, Open  ]],
Cell[922066, 23325, 183, 3, 40, "MathCaption",ExpressionUUID->"3a797d89-c1c6-4c23-b19b-e006f338f34a"]
}, Closed]],
Cell[CellGroupData[{
Cell[922286, 23333, 171, 3, 41, "Subsubsection",ExpressionUUID->"dd869560-b307-4804-b0a3-2322cee31c1d"],
Cell[922460, 23338, 215, 5, 24, "Text",ExpressionUUID->"b51bd2f9-84ff-4d6d-8263-772889b99dba"],
Cell[922678, 23345, 306, 5, 42, "DefinitionBox",ExpressionUUID->"a175050a-6be7-4d1d-be7f-847bee268215"],
Cell[922987, 23352, 113, 0, 23, "Text",ExpressionUUID->"293181a8-1fe7-4235-90ab-19e6b5ae9c1e"],
Cell[923103, 23354, 1481, 36, 296, "Text",ExpressionUUID->"f27ff5fb-0bdc-4e46-b3a9-8c578d8f3cda"],
Cell[CellGroupData[{
Cell[924609, 23394, 195, 2, 40, "MathCaption",ExpressionUUID->"08360204-917c-4a99-af4d-cb1071d41802",
 CellID->1402959324],
Cell[CellGroupData[{
Cell[924829, 23400, 113, 2, 27, "Input",ExpressionUUID->"1165e8c9-df89-4549-85d6-a51c3503f62c"],
Cell[924945, 23404, 8169, 241, 32, "Output",ExpressionUUID->"d4801b87-9bd0-40c0-91ca-1e0023a879b4",
 CellID->1639099083]
}, Open  ]],
Cell[CellGroupData[{
Cell[933151, 23650, 787, 24, 27, "Input",ExpressionUUID->"decbdad5-9c12-40a9-98d0-ed0418969392"],
Cell[933941, 23676, 4044, 121, 32, "Output",ExpressionUUID->"38100f1d-fe38-4c4b-8933-453d4db65395",
 CellID->632927560]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[938034, 23803, 243, 6, 40, "MathCaption",ExpressionUUID->"5e0ded7b-692e-4726-a6b0-de8c852fadc3"],
Cell[CellGroupData[{
Cell[938302, 23813, 152, 3, 27, "Input",ExpressionUUID->"beec4951-1537-41a4-83b8-a728fc481f7e"],
Cell[938457, 23818, 4602, 131, 32, "Output",ExpressionUUID->"20d4c8fe-a444-4a40-8384-9188bad474b1",
 CellID->1423645728]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[943108, 23955, 221, 6, 40, "MathCaption",ExpressionUUID->"599a2e5f-bd4a-4bd9-8bcd-9dd0c1cc6739"],
Cell[CellGroupData[{
Cell[943354, 23965, 1000, 28, 47, "Input",ExpressionUUID->"b34949bf-5c43-4c79-b7fa-befbb412bf18"],
Cell[944357, 23995, 12556, 357, 58, "Output",ExpressionUUID->"66168d59-6dd8-4555-8956-a12873168123",
 CellID->461810322]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[956962, 24358, 171, 3, 40, "MathCaption",ExpressionUUID->"68e86b2c-1ddd-43cf-ac86-bc034db2766b"],
Cell[CellGroupData[{
Cell[957158, 24365, 176, 4, 27, "Input",ExpressionUUID->"bacd5051-b924-46f6-9450-99424ec88845"],
Cell[957337, 24371, 4051, 121, 32, "Output",ExpressionUUID->"81f251a8-e610-4c77-a851-255407ceed5b",
 CellID->1960327329]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[961437, 24498, 143, 2, 40, "MathCaption",ExpressionUUID->"2c3c5f61-9573-4923-a647-fac8d8887e51"],
Cell[CellGroupData[{
Cell[961605, 24504, 217, 5, 27, "Input",ExpressionUUID->"b09fb82c-77e5-447b-ba60-68ae1b883283"],
Cell[961825, 24511, 5628, 162, 32, "Output",ExpressionUUID->"86303904-5c21-4f9d-bfb7-6b089adfd258",
 CellID->936333049]
}, Open  ]],
Cell[CellGroupData[{
Cell[967490, 24678, 238, 6, 27, "Input",ExpressionUUID->"99d0a71c-0244-4532-9286-09bccdf841fc"],
Cell[967731, 24686, 6467, 189, 32, "Output",ExpressionUUID->"84af7c29-383b-4271-8029-f214c3813ee5",
 CellID->1746856133]
}, Open  ]],
Cell[CellGroupData[{
Cell[974235, 24880, 306, 8, 27, "Input",ExpressionUUID->"a779f61a-671e-4dee-93fb-92ea30e20d03"],
Cell[974544, 24890, 4603, 131, 32, "Output",ExpressionUUID->"b4b29332-f625-4201-87cc-d6e56764fcc8",
 CellID->289727438]
}, Open  ]],
Cell[CellGroupData[{
Cell[979184, 25026, 181, 4, 27, "Input",ExpressionUUID->"cab229d0-47cc-4953-8742-1d727eb4a378"],
Cell[979368, 25032, 3241, 95, 32, "Output",ExpressionUUID->"de69afdd-43a9-4e89-af8e-32e0f9a11fc9",
 CellID->1601519298]
}, Open  ]],
Cell[CellGroupData[{
Cell[982646, 25132, 326, 8, 27, "Input",ExpressionUUID->"0c58b5b6-43b9-49d8-a8b3-dff1eba80588"],
Cell[982975, 25142, 5624, 162, 32, "Output",ExpressionUUID->"5436f05b-e70d-41e7-843f-f253ca6bf695",
 CellID->123988376]
}, Open  ]],
Cell[CellGroupData[{
Cell[988636, 25309, 295, 6, 27, "Input",ExpressionUUID->"cdc7ec7d-f32f-4eba-87ea-2b6961f5e2f6"],
Cell[988934, 25317, 7157, 210, 32, "Output",ExpressionUUID->"b3de59a9-ebab-4297-a0b8-01fe1a58ebaf",
 CellID->1187002912]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[996140, 25533, 590, 13, 57, "MathCaption",ExpressionUUID->"48d0d17c-29ba-4a7d-979a-86078e7b6c5d"],
Cell[CellGroupData[{
Cell[996755, 25550, 370, 8, 27, "Input",ExpressionUUID->"c82642e5-1e49-4451-bfe6-e5a36e43835f"],
Cell[997128, 25560, 5443, 158, 32, "Output",ExpressionUUID->"aace69ec-f9c7-456c-8ca8-c05f165d18ef",
 CellID->1585944226]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1002620, 25724, 338, 10, 41, "MathCaption",ExpressionUUID->"569c28bb-8349-40a3-b0d1-b6b865edb553"],
Cell[CellGroupData[{
Cell[1002983, 25738, 403, 10, 27, "Input",ExpressionUUID->"b9a6929d-0c6a-49b0-946b-da0a9ec9aaaf"],
Cell[1003389, 25750, 6462, 189, 32, "Output",ExpressionUUID->"8f52a453-6e25-4200-8bd6-c7af3faded62",
 CellID->1243667434]
}, Open  ]]
}, Open  ]],
Cell[1009878, 25943, 441, 12, 58, "MathCaption",ExpressionUUID->"38deea00-fb20-49eb-bf64-c902147e552d"],
Cell[CellGroupData[{
Cell[1010344, 25959, 226, 6, 41, "MathCaption",ExpressionUUID->"813e3dcc-82d7-422c-bd47-760da955b048"],
Cell[CellGroupData[{
Cell[1010595, 25969, 552, 16, 27, "Input",ExpressionUUID->"f1bb2e5f-6686-4721-b0f9-77d625b4c071"],
Cell[1011150, 25987, 2376, 66, 32, "Output",ExpressionUUID->"d99cedcd-a65c-4246-9ff3-df70a31031b6",
 CellID->1016337740]
}, Open  ]],
Cell[CellGroupData[{
Cell[1013563, 26058, 149, 3, 27, "Input",ExpressionUUID->"f53d44cd-fdad-49ff-b468-16a992b8d49f"],
Cell[1013715, 26063, 1536, 40, 32, "Output",ExpressionUUID->"fe08c6a8-73df-4f90-97bb-bcabb7244d6b",
 CellID->1903012753]
}, Open  ]],
Cell[1015266, 26106, 161, 3, 27, "Input",ExpressionUUID->"aebd8276-7a99-409a-9470-ce6aecc5708d"],
Cell[CellGroupData[{
Cell[1015452, 26113, 549, 16, 27, "Input",ExpressionUUID->"6d53468b-5e7c-4b13-9b1c-7528bc2f41ef"],
Cell[1016004, 26131, 1540, 40, 32, "Output",ExpressionUUID->"787a4d7d-3b0e-4527-b665-ed5a6eac435e",
 CellID->144068692]
}, Open  ]],
Cell[1017559, 26174, 160, 3, 27, "Input",ExpressionUUID->"88fed2ab-7212-47f8-a927-88b042fc7695"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1017756, 26182, 270, 6, 41, "MathCaption",ExpressionUUID->"fcac0608-b255-4d8e-bc6d-905276cafd3f"],
Cell[CellGroupData[{
Cell[1018051, 26192, 319, 9, 27, "Input",ExpressionUUID->"a509d96c-81ed-492a-a246-819a74a5c111"],
Cell[CellGroupData[{
Cell[1018395, 26205, 1108, 21, 22, "Print",ExpressionUUID->"c50841bb-4de4-4204-8883-245644f2340c",
 CellID->1934053673],
Cell[1019506, 26228, 1228, 24, 22, "Print",ExpressionUUID->"77d2a23d-4b82-48b0-8a1f-210a52c4287c",
 CellID->1327601119],
Cell[1020737, 26254, 1245, 25, 22, "Print",ExpressionUUID->"ee3f8f7e-19c8-4a88-a81d-7cb9077e1618",
 CellID->636914703],
Cell[1021985, 26281, 1323, 24, 22, "Print",ExpressionUUID->"b751e93b-4111-4347-8024-a11eaa173b43",
 CellID->1566916673],
Cell[1023311, 26307, 1197, 23, 22, "Print",ExpressionUUID->"cdde46d8-f613-40ab-83f1-20a19ec3a244",
 CellID->576559950],
Cell[1024511, 26332, 696, 10, 22, "Print",ExpressionUUID->"cc36fc9a-4372-4e87-b9a5-e99e1a5987d1",
 CellID->796864386]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1025256, 26348, 531, 15, 27, "Input",ExpressionUUID->"4e2226a7-a0bc-494a-be9f-4f39fa95d6d0"],
Cell[1025790, 26365, 2438, 68, 32, "Output",ExpressionUUID->"714d8fc7-1a51-4741-acba-885aa17a77f9",
 CellID->1119621641]
}, Open  ]],
Cell[CellGroupData[{
Cell[1028265, 26438, 149, 3, 27, "Input",ExpressionUUID->"a78ed31d-cde4-49e7-8fe8-9d86255c992b"],
Cell[1028417, 26443, 2439, 68, 32, "Output",ExpressionUUID->"46e5ca8e-6f5f-44d1-8769-a1f6f69d908f",
 CellID->1336700114]
}, Open  ]],
Cell[CellGroupData[{
Cell[1030893, 26516, 224, 5, 27, "Input",ExpressionUUID->"906d7d83-6d13-4c13-8ba4-eef6383ade75"],
Cell[1031120, 26523, 1103, 21, 22, "Print",ExpressionUUID->"9b2c4b76-8f19-4d8d-b408-bae86cd577fa",
 CellID->2113793198],
Cell[1032226, 26546, 3535, 101, 32, "Output",ExpressionUUID->"a7f97c6a-5a08-4795-8ec5-ad4bf39529aa",
 CellID->700306465]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1035810, 26653, 345, 7, 56, "MathCaption",ExpressionUUID->"5c1fa4a1-08a8-4b37-9e9a-ebfb7eabe0b3"],
Cell[CellGroupData[{
Cell[1036180, 26664, 744, 23, 27, "Input",ExpressionUUID->"695685a0-4fa9-4847-a927-b748e0277606"],
Cell[1036927, 26689, 3348, 97, 32, "Output",ExpressionUUID->"04e9e040-b068-4a24-afe4-3ec3ca15dca1",
 CellID->444699952]
}, Open  ]],
Cell[CellGroupData[{
Cell[1040312, 26791, 405, 11, 27, "Input",ExpressionUUID->"c35b6d88-8db8-43f9-b960-14330ddf37b3"],
Cell[1040720, 26804, 10204, 275, 58, "Output",ExpressionUUID->"2ac7433a-d156-4835-aa9d-c4caf79623e3",
 CellID->1942107452]
}, Open  ]],
Cell[CellGroupData[{
Cell[1050961, 27084, 691, 21, 27, "Input",ExpressionUUID->"a3b7e8ea-ba13-437a-8cd7-83b01c5608d3"],
Cell[1051655, 27107, 3214, 93, 35, "Output",ExpressionUUID->"b0cfb436-f367-4a19-9a9b-941d7ab9e511",
 CellID->2060783373]
}, Open  ]],
Cell[CellGroupData[{
Cell[1054906, 27205, 409, 11, 27, "Input",ExpressionUUID->"12d1ebb5-23a2-436d-b259-9600e185e618"],
Cell[1055318, 27218, 9601, 259, 64, "Output",ExpressionUUID->"43454c6d-c37a-4568-903c-d898194c0f15",
 CellID->1155650882]
}, Open  ]],
Cell[1064934, 27480, 240, 5, 66, "Input",ExpressionUUID->"89bd548d-8fbf-49f7-9f6d-9118b9fbd62c"],
Cell[CellGroupData[{
Cell[1065199, 27489, 146, 3, 27, "Input",ExpressionUUID->"6e7a8336-94b9-420f-85e2-5aa421faf43b"],
Cell[CellGroupData[{
Cell[1065370, 27496, 1044, 18, 22, "Print",ExpressionUUID->"f95facd7-cef4-4f6a-85ec-5a840cfe5520",
 CellID->1169328056],
Cell[1066417, 27516, 1012, 18, 22, "Print",ExpressionUUID->"c68201fc-d704-4ce8-9020-5db364fafca1",
 CellID->1273783495],
Cell[1067432, 27536, 1027, 18, 22, "Print",ExpressionUUID->"013cacbb-9ee9-4079-b240-2825dca0872f",
 CellID->214723522],
Cell[1068462, 27556, 1006, 18, 22, "Print",ExpressionUUID->"819540d9-5a4b-4888-b299-67394baec7e2",
 CellID->1085707294],
Cell[1069471, 27576, 1007, 18, 22, "Print",ExpressionUUID->"d157dd89-ab67-490f-afed-b3cdc243a4bc",
 CellID->313941699],
Cell[1070481, 27596, 998, 18, 22, "Print",ExpressionUUID->"288f7f26-6d97-4546-b49b-d68e3c6cd5f9",
 CellID->50814414]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[1071518, 27619, 552, 13, 71, "Text",ExpressionUUID->"7cf8717d-6da8-422d-acdd-5ec0734cc807"]
}, Closed]],
Cell[CellGroupData[{
Cell[1072107, 27637, 170, 3, 41, "Subsubsection",ExpressionUUID->"2f7db809-afa0-432a-9b5f-d4dd5996996f"],
Cell[1072280, 27642, 257, 4, 41, "DefinitionBox",ExpressionUUID->"cf26f1ac-868d-4dad-bc22-c41330818fa7"],
Cell[1072540, 27648, 115, 0, 23, "Text",ExpressionUUID->"a3d63aae-bc1a-4150-a064-1f924728add3"],
Cell[1072658, 27650, 431, 9, 42, "Text",ExpressionUUID->"63e7b682-5352-437f-8a38-97b2f00904ca"],
Cell[CellGroupData[{
Cell[1073114, 27663, 199, 2, 40, "MathCaption",ExpressionUUID->"35b53f9e-4c74-4fa9-a59c-ecf125a3deff",
 CellID->1734699235],
Cell[CellGroupData[{
Cell[1073338, 27669, 1413, 45, 47, "Input",ExpressionUUID->"1272182b-62ed-4834-a933-57c104d4b164"],
Cell[1074754, 27716, 6322, 185, 32, "Output",ExpressionUUID->"1d216f50-cf0d-46e1-b395-af3ba862f6cb",
 CellID->2077686682]
}, Open  ]],
Cell[CellGroupData[{
Cell[1081113, 27906, 179, 4, 27, "Input",ExpressionUUID->"548a95c6-39fa-4cdd-b119-5937f69198f7"],
Cell[1081295, 27912, 2721, 74, 32, "Output",ExpressionUUID->"fcf22ba5-7f9c-47a6-83e3-5d3a430f0158",
 CellID->1669015989]
}, Open  ]],
Cell[CellGroupData[{
Cell[1084053, 27991, 253, 6, 27, "Input",ExpressionUUID->"5bece986-0760-4793-a3a9-7b5708f5f7d4"],
Cell[1084309, 27999, 7005, 207, 32, "Output",ExpressionUUID->"f69450ac-f1a0-4236-b82d-569c68fcc130",
 CellID->1409327831]
}, Open  ]],
Cell[CellGroupData[{
Cell[1091351, 28211, 216, 4, 27, "Input",ExpressionUUID->"8164679e-d859-4574-a001-ab58fa03c245"],
Cell[1091570, 28217, 5287, 152, 32, "Output",ExpressionUUID->"a3421b9e-a7cb-48fe-b88d-a990b05485b5",
 CellID->279075165]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1096906, 28375, 160, 2, 40, "MathCaption",ExpressionUUID->"af622d27-d7d8-4beb-86f0-1a2ab685a91b"],
Cell[CellGroupData[{
Cell[1097091, 28381, 328, 7, 27, "Input",ExpressionUUID->"20a5e2fb-3ad2-413c-a702-f83641f675d6"],
Cell[1097422, 28390, 12007, 355, 32, "Output",ExpressionUUID->"7a34c1e8-71cc-4ff5-b36a-9ce932cd333b",
 CellID->1165567163]
}, Open  ]],
Cell[CellGroupData[{
Cell[1109466, 28750, 216, 4, 27, "Input",ExpressionUUID->"e1fcc168-85f5-446b-b0b9-2731f2001ca8"],
Cell[1109685, 28756, 8201, 237, 32, "Output",ExpressionUUID->"1e212e50-f90c-420a-ae1b-804b79d0d977",
 CellID->966044531]
}, Open  ]]
}, Open  ]],
Cell[1117913, 28997, 217, 2, 40, "MathCaption",ExpressionUUID->"3b71e45c-0a1c-4907-a514-80a783e8c4b0",
 CellID->2091127534],
Cell[CellGroupData[{
Cell[1118155, 29003, 258, 7, 41, "MathCaption",ExpressionUUID->"0150f988-a717-4c0e-aad7-d10771d03726"],
Cell[CellGroupData[{
Cell[1118438, 29014, 401, 9, 27, "Input",ExpressionUUID->"c13c403d-395d-4efb-94d6-7973939f84f4"],
Cell[1118842, 29025, 2720, 74, 32, "Output",ExpressionUUID->"8f4c9473-cf3a-46cc-a0d3-97e07d926ee7",
 CellID->133944636]
}, Open  ]],
Cell[CellGroupData[{
Cell[1121599, 29104, 394, 8, 27, "Input",ExpressionUUID->"117cbe02-7573-4b99-86f4-cce3bdb2e2ae"],
Cell[1121996, 29114, 4471, 130, 32, "Output",ExpressionUUID->"b94be748-8b20-466a-b603-ec8f8ec7ad47",
 CellID->1477250969]
}, Open  ]],
Cell[CellGroupData[{
Cell[1126504, 29249, 251, 6, 27, "Input",ExpressionUUID->"a9359986-8cb3-41b6-ac27-22a1fc133872"],
Cell[1126758, 29257, 6976, 206, 32, "Output",ExpressionUUID->"7317b7cf-663b-4ce1-8324-f6f25fcca5b9",
 CellID->1543138871]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1133783, 29469, 260, 6, 40, "MathCaption",ExpressionUUID->"750d3378-e6a2-4a32-b351-2b3ab9c43cbd"],
Cell[CellGroupData[{
Cell[1134068, 29479, 569, 16, 27, "Input",ExpressionUUID->"d59cbb71-0bfe-491e-9023-bb6ba33d23cb"],
Cell[1134640, 29497, 4760, 146, 32, "Output",ExpressionUUID->"4205d0f8-5153-4df4-9f84-8653bbc7422b",
 CellID->1668739797]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1139449, 29649, 268, 6, 40, "MathCaption",ExpressionUUID->"98de541f-0321-403b-8282-8fd22f924a10"],
Cell[CellGroupData[{
Cell[1139742, 29659, 375, 9, 27, "Input",ExpressionUUID->"3c6ec5dd-41b6-44f6-9482-a090eb9545f4"],
Cell[1140120, 29670, 2260, 64, 32, "Output",ExpressionUUID->"c07842ee-433d-498e-8820-509b2b6eb15c",
 CellID->2116524650]
}, Open  ]],
Cell[CellGroupData[{
Cell[1142417, 29739, 481, 13, 27, "Input",ExpressionUUID->"614ccb8a-3133-45a4-9207-1a9346de3694"],
Cell[1142901, 29754, 1454, 38, 34, "Output",ExpressionUUID->"c836bac3-1cd4-4be4-85cc-333860fe808d",
 CellID->1493055970]
}, Open  ]],
Cell[CellGroupData[{
Cell[1144392, 29797, 138, 3, 27, "Input",ExpressionUUID->"df320ab0-6523-4844-9657-18b498e9fd6e"],
Cell[1144533, 29802, 2377, 66, 34, "Output",ExpressionUUID->"231ee5ee-48b5-4210-902f-90db21d99bf3",
 CellID->941986412]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1146959, 29874, 261, 7, 41, "MathCaption",ExpressionUUID->"4a041000-bef7-4041-aadd-e82e849a574e"],
Cell[CellGroupData[{
Cell[1147245, 29885, 476, 13, 27, "Input",ExpressionUUID->"a0d93d66-1d28-4f23-b1dc-77e13cf8d6b1"],
Cell[1147724, 29900, 1853, 50, 33, "Output",ExpressionUUID->"6a1c9a0e-7e2c-475c-bfa5-5068437f7e28",
 CellID->300289431]
}, Open  ]],
Cell[CellGroupData[{
Cell[1149614, 29955, 316, 6, 27, "Input",ExpressionUUID->"b6f6df80-bd50-4a40-b68d-5042d28062f9"],
Cell[1149933, 29963, 1842, 50, 33, "Output",ExpressionUUID->"0e74474e-5fb6-471a-a734-c8f051915c28",
 CellID->1241121440]
}, Open  ]],
Cell[CellGroupData[{
Cell[1151812, 30018, 475, 11, 27, "Input",ExpressionUUID->"3cf7be0a-3efb-4bca-bda3-295dd56715b3"],
Cell[1152290, 30031, 3112, 84, 33, "Output",ExpressionUUID->"6f3bc8d0-92a0-42b0-b48d-111d698ee5e9",
 CellID->1893679720]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1155451, 30121, 273, 7, 41, "MathCaption",ExpressionUUID->"d622fb7a-a083-4ed9-a722-d6bc7b296e1e"],
Cell[CellGroupData[{
Cell[1155749, 30132, 455, 13, 27, "Input",ExpressionUUID->"3a47f130-7b80-48af-bfe4-fdffe2b8e5b5"],
Cell[1156207, 30147, 1504, 39, 32, "Output",ExpressionUUID->"d14cd1a0-6d50-4705-89b7-491163652aaf",
 CellID->1292102491]
}, Open  ]],
Cell[CellGroupData[{
Cell[1157748, 30191, 178, 4, 27, "Input",ExpressionUUID->"9ba56b7c-cab6-4c3a-ac3c-0938578f2d7e"],
Cell[1157929, 30197, 3195, 93, 32, "Output",ExpressionUUID->"a582cc06-2e82-4fbc-8d90-4c9079aadc00",
 CellID->1252954152]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1161173, 30296, 187, 3, 40, "MathCaption",ExpressionUUID->"56557915-6da7-4235-bb14-4ca4a504aab7"],
Cell[CellGroupData[{
Cell[1161385, 30303, 375, 10, 27, "Input",ExpressionUUID->"a184a7d2-2e33-4725-8f83-aa8413e16b2c"],
Cell[1161763, 30315, 2357, 66, 32, "Output",ExpressionUUID->"aa3a2aca-4d78-4b02-a2ef-e78e4760f824",
 CellID->1244693224]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1164169, 30387, 400, 9, 41, "MathCaption",ExpressionUUID->"c919089f-7209-47b9-9346-895425605ddc"],
Cell[CellGroupData[{
Cell[1164594, 30400, 393, 10, 27, "Input",ExpressionUUID->"ba14f68a-792e-46b1-a095-539345b5bc0e"],
Cell[1164990, 30412, 3157, 92, 32, "Output",ExpressionUUID->"ef6c99e0-b9d5-45c6-a200-910077606d9c",
 CellID->928540201]
}, Open  ]],
Cell[CellGroupData[{
Cell[1168184, 30509, 548, 15, 27, "Input",ExpressionUUID->"d7cb9dd0-f551-47bf-a944-0ea92fd31017"],
Cell[1168735, 30526, 3163, 92, 32, "Output",ExpressionUUID->"e68c0503-91a2-45c4-8f9a-78603ababc17",
 CellID->1530963271]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[1171959, 30625, 167, 3, 41, "Subsubsection",ExpressionUUID->"f4d26f62-4c68-41c1-a3de-23359605fc6a"],
Cell[1172129, 30630, 510, 8, 75, "DefinitionBox",ExpressionUUID->"49cac7fa-2261-4dc6-a325-a1061ea3c0a4"],
Cell[1172642, 30640, 105, 0, 23, "Text",ExpressionUUID->"d3bf3c50-64a0-4062-a9c8-417ec7e88489"],
Cell[CellGroupData[{
Cell[1172772, 30644, 251, 6, 40, "MathCaption",ExpressionUUID->"e3fdd907-e478-40d1-ab6c-1cbc7e57fee9"],
Cell[CellGroupData[{
Cell[1173048, 30654, 391, 10, 27, "Input",ExpressionUUID->"7794ffa4-c11e-4cc8-b842-54c07da3ad4d"],
Cell[1173442, 30666, 4118, 122, 32, "Output",ExpressionUUID->"8aad4887-a5c7-46e9-86ee-cb29054b7018",
 CellID->1009678446]
}, Open  ]],
Cell[CellGroupData[{
Cell[1177597, 30793, 145, 3, 27, "Input",ExpressionUUID->"4e8f5d8c-8453-45f7-b668-f4c988a37246"],
Cell[1177745, 30798, 2324, 66, 32, "Output",ExpressionUUID->"7e3607ea-24a5-4bb0-b76e-5105cba73936",
 CellID->249863328]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1180118, 30870, 304, 8, 41, "MathCaption",ExpressionUUID->"4b6b7cbd-5696-4509-b12e-a02c39c81ccc"],
Cell[CellGroupData[{
Cell[1180447, 30882, 774, 24, 27, "Input",ExpressionUUID->"138037b0-7585-4eb4-bb77-896672055746"],
Cell[1181224, 30908, 2448, 69, 32, "Output",ExpressionUUID->"729c95e3-047f-4326-8ca0-d6c20cd22a3d",
 CellID->1122423811]
}, Open  ]]
}, Open  ]],
Cell[1183699, 30981, 1518, 53, 67, "MathCaption",ExpressionUUID->"44236b51-0290-4270-863a-bbeeda07f574"],
Cell[CellGroupData[{
Cell[1185242, 31038, 317, 7, 41, "MathCaption",ExpressionUUID->"cabba59a-a17b-4596-930b-c32af474c8d6",
 CellID->1618843451],
Cell[CellGroupData[{
Cell[1185584, 31049, 474, 13, 27, "Input",ExpressionUUID->"783dabf5-b623-4e01-b026-a817ba1b992c"],
Cell[1186061, 31064, 8694, 249, 32, "Output",ExpressionUUID->"dfa840b2-95ad-440d-a0a5-12df50654ada",
 CellID->1080636229]
}, Open  ]],
Cell[CellGroupData[{
Cell[1194792, 31318, 149, 3, 27, "Input",ExpressionUUID->"8af79e68-da8d-42a9-9a27-7a792b8acd3b"],
Cell[1194944, 31323, 5466, 161, 32, "Output",ExpressionUUID->"bfdbadb0-90c3-42eb-b57f-14d63fcc6667",
 CellID->983898124]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1200459, 31490, 170, 3, 40, "MathCaption",ExpressionUUID->"9b0daaf3-a596-4e65-b3ea-140a1c50b392"],
Cell[CellGroupData[{
Cell[1200654, 31497, 521, 15, 27, "Input",ExpressionUUID->"a0c224ae-9878-4336-834a-43c7737e4589"],
Cell[1201178, 31514, 5531, 162, 32, "Output",ExpressionUUID->"407abdcd-be0b-4152-8eb1-13fa5e644893",
 CellID->751688168]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1206758, 31682, 201, 3, 40, "MathCaption",ExpressionUUID->"b8cb3dc7-e44f-403f-91ef-f6e1b1b75d23"],
Cell[CellGroupData[{
Cell[1206984, 31689, 345, 11, 27, "Input",ExpressionUUID->"35176443-6dda-4458-b171-3ebd84f30d7d"],
Cell[1207332, 31702, 3746, 114, 32, "Output",ExpressionUUID->"1a803500-a9e0-467b-9fb9-ba5de17fe454",
 CellID->2124227922]
}, Open  ]],
Cell[CellGroupData[{
Cell[1211115, 31821, 323, 6, 27, "Input",ExpressionUUID->"3f141297-d6f5-49e2-b848-0750c29390f3"],
Cell[1211441, 31829, 3827, 116, 32, "Output",ExpressionUUID->"a716f73a-252b-4341-aa72-574323edbaad",
 CellID->2130537326]
}, Open  ]],
Cell[CellGroupData[{
Cell[1215305, 31950, 317, 6, 27, "Input",ExpressionUUID->"42648f61-fcd0-42eb-b575-711558801ea7"],
Cell[1215625, 31958, 4012, 120, 32, "Output",ExpressionUUID->"d406b170-b4d6-4948-9f77-ac39b1200ec7",
 CellID->798508004]
}, Open  ]],
Cell[CellGroupData[{
Cell[1219674, 32083, 272, 6, 27, "Input",ExpressionUUID->"c51abc58-6e66-4aad-a733-e35a4def8f49"],
Cell[1219949, 32091, 4195, 125, 32, "Output",ExpressionUUID->"de3382db-0fd3-4069-8170-f505c22e79a3",
 CellID->898713125]
}, Open  ]],
Cell[CellGroupData[{
Cell[1224181, 32221, 392, 9, 27, "Input",ExpressionUUID->"510cb4de-9c2e-4923-8076-a8d443655978"],
Cell[1224576, 32232, 4175, 125, 32, "Output",ExpressionUUID->"26885443-f6cf-4901-a218-565f8431b25f",
 CellID->1419002524]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1228800, 32363, 164, 3, 40, "MathCaption",ExpressionUUID->"4d150a0a-98fe-44f6-9332-b686ccab10ca"],
Cell[CellGroupData[{
Cell[1228989, 32370, 654, 21, 27, "Input",ExpressionUUID->"d030aa1d-3ed5-4714-a87c-a14bd742e5c3"],
Cell[1229646, 32393, 3310, 97, 32, "Output",ExpressionUUID->"1c8bea4e-fe21-48fc-a0e8-eb5dff17bbd8",
 CellID->1369520933]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1233005, 32496, 296, 7, 42, "MathCaption",ExpressionUUID->"7805f1c1-5858-47c3-8ad5-c4afd7924bf9"],
Cell[CellGroupData[{
Cell[1233326, 32507, 150, 3, 27, "Input",ExpressionUUID->"a0fbd442-4b22-4416-b9dd-3b93f462f04e"],
Cell[1233479, 32512, 596, 9, 26, "Output",ExpressionUUID->"1de95ba4-e756-40e8-9ab8-488ef849c095",
 CellID->808592385]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1234124, 32527, 147, 3, 40, "MathCaption",ExpressionUUID->"4f19cd8f-8d6a-4354-a8ec-b69554592185"],
Cell[CellGroupData[{
Cell[1234296, 32534, 570, 18, 27, "Input",ExpressionUUID->"cae7f8f3-de97-45a6-8b50-bf5248677cc3"],
Cell[1234869, 32554, 4013, 120, 32, "Output",ExpressionUUID->"7821570f-f7bd-4ff2-8c62-c7420f0d720e",
 CellID->735696479]
}, Open  ]],
Cell[CellGroupData[{
Cell[1238919, 32679, 150, 3, 27, "Input",ExpressionUUID->"ecb5c111-4734-47cb-8cb6-9909b5fb5b65"],
Cell[1239072, 32684, 4017, 120, 32, "Output",ExpressionUUID->"d058777a-d008-455f-a1c9-d37a46a3f4c6",
 CellID->1787366774]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1243138, 32810, 193, 4, 40, "MathCaption",ExpressionUUID->"adc861ab-91dc-4aa3-b123-da287061da6e"],
Cell[CellGroupData[{
Cell[1243356, 32818, 176, 4, 27, "Input",ExpressionUUID->"c5d8897b-1b09-4de0-a41a-8be3149078ba"],
Cell[1243535, 32824, 595, 9, 26, "Output",ExpressionUUID->"4f83d7d0-375e-40dc-a8e8-6d755dcc56d1",
 CellID->1272116382]
}, Open  ]]
}, Open  ]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1244203, 32841, 214, 3, 39, "Subsection",ExpressionUUID->"991f8040-8df9-46b6-8ade-ffc5320a5f72",
 CellID->431306534],
Cell[1244420, 32846, 502, 12, 41, "Text",ExpressionUUID->"2df945df-a6d7-48c6-b307-6b14d39e4deb",
 CellID->1211608313],
Cell[1244925, 32860, 240, 4, 41, "DefinitionBox",ExpressionUUID->"f6d10f94-e232-4aa3-b888-8933390cc274",
 CellID->1060975737],
Cell[1245168, 32866, 119, 1, 23, "Text",ExpressionUUID->"7e4350b2-9c7c-4a3b-b021-569d2d7eeacd",
 CellID->1262490718],
Cell[1245290, 32869, 483, 9, 141, "Text",ExpressionUUID->"1cc9e05a-d0e6-4d57-9f23-19b2e24166b1",
 CellID->1594783279],
Cell[CellGroupData[{
Cell[1245798, 32882, 195, 2, 40, "MathCaption",ExpressionUUID->"f45ab70c-e978-404d-876f-9767a6147762",
 CellID->471215995],
Cell[CellGroupData[{
Cell[1246018, 32888, 518, 14, 27, "Input",ExpressionUUID->"4cfad6c5-a962-4aa3-82c5-805a365eb0ce",
 CellID->1543065066],
Cell[1246539, 32904, 5923, 181, 32, "Output",ExpressionUUID->"f134bd0a-8241-43b0-9978-3458cc23a22f",
 CellID->1483245410]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1252511, 33091, 216, 2, 40, "MathCaption",ExpressionUUID->"0f268c5c-6ec9-49fb-aa5e-f7edbac73476",
 CellID->75238861],
Cell[CellGroupData[{
Cell[1252752, 33097, 946, 30, 27, "Input",ExpressionUUID->"b2c72ff6-9813-4ac2-948d-73d07a4eaff0",
 CellID->779171946],
Cell[1253701, 33129, 4272, 128, 32, "Output",ExpressionUUID->"865ea503-f017-4505-81b6-fbd9169f1ef9",
 CellID->915431494]
}, Open  ]],
Cell[CellGroupData[{
Cell[1258010, 33262, 210, 5, 27, "Input",ExpressionUUID->"6bdaf458-95d9-43db-bfa8-c9f8513b195c",
 CellID->773194116],
Cell[1258223, 33269, 29117, 859, 32, "Output",ExpressionUUID->"a37ce3c7-2d08-4ff1-90fe-8641328c07cd",
 CellID->1084272550]
}, Open  ]],
Cell[CellGroupData[{
Cell[1287377, 34133, 243, 6, 27, "Input",ExpressionUUID->"a6765f3c-d1d0-4f93-ad97-7e1d954c995f",
 CellID->1769420663],
Cell[1287623, 34141, 27967, 826, 32, "Output",ExpressionUUID->"8e78d937-5a83-49bc-b6c7-2894f278cc14",
 CellID->1970003055]
}, Open  ]],
Cell[CellGroupData[{
Cell[1315627, 34972, 302, 6, 27, "Input",ExpressionUUID->"669472bb-3a73-4d05-950e-eb920b8035dd",
 CellID->656775127],
Cell[1315932, 34980, 8510, 251, 32, "Output",ExpressionUUID->"cde15701-7247-4360-a018-850f6a1ca19d",
 CellID->2085959586]
}, Open  ]],
Cell[CellGroupData[{
Cell[1324479, 35236, 329, 7, 27, "Input",ExpressionUUID->"fa37887c-091e-41cc-821a-5b5f2ef981b4",
 CellID->1101717314],
Cell[1324811, 35245, 8520, 251, 32, "Output",ExpressionUUID->"c1da8ff8-47f2-4cc4-bdb2-8b6e91b4f5fd",
 CellID->1052224627]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1333380, 35502, 184, 2, 40, "MathCaption",ExpressionUUID->"1bd3ccea-d4be-438c-83ce-b0aee7b42768",
 CellID->1379951860],
Cell[CellGroupData[{
Cell[1333589, 35508, 474, 13, 27, "Input",ExpressionUUID->"d0e6eeb6-355d-47c0-ac4d-28d98ef14f38",
 CellID->27523420],
Cell[1334066, 35523, 3662, 109, 32, "Output",ExpressionUUID->"f87cbf0a-34ba-4a07-bda3-9a54f355232b",
 CellID->211472944]
}, Open  ]],
Cell[CellGroupData[{
Cell[1337765, 35637, 1684, 56, 47, "Input",ExpressionUUID->"15ae4422-299c-4bbd-b911-ae6871e39793",
 CellID->1694225837],
Cell[1339452, 35695, 8336, 247, 32, "Output",ExpressionUUID->"baf485b2-cdb8-45aa-876b-0701c23368dd",
 CellID->140334023]
}, Open  ]],
Cell[CellGroupData[{
Cell[1347825, 35947, 163, 3, 27, "Input",ExpressionUUID->"c016f05a-e9a1-4621-9174-3ee8d407c4eb",
 CellID->459528220],
Cell[1347991, 35952, 22281, 658, 32, "Output",ExpressionUUID->"da51204e-3115-4125-b770-d20c0df386fa",
 CellID->656880155]
}, Open  ]],
Cell[CellGroupData[{
Cell[1370309, 36615, 162, 3, 27, "Input",ExpressionUUID->"b398098e-6222-4089-acc5-c6ec0bb51ef1",
 CellID->1391926505],
Cell[1370474, 36620, 20605, 605, 32, "Output",ExpressionUUID->"0a9150e0-e700-4399-ad72-3049ec6d28d7",
 CellID->2118819505]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1391128, 37231, 157, 2, 40, "MathCaption",ExpressionUUID->"71ed5452-40e5-4df4-95c2-e2ea371592e0",
 CellID->1558102094],
Cell[CellGroupData[{
Cell[1391310, 37237, 530, 14, 27, "Input",ExpressionUUID->"8b967ba1-e237-4ef9-a33f-b37a6e022d0d",
 CellID->606302340],
Cell[1391843, 37253, 3873, 115, 32, "Output",ExpressionUUID->"1440bf6b-660b-41ab-aa26-12381fb44d9e",
 CellID->1425585981]
}, Open  ]],
Cell[CellGroupData[{
Cell[1395753, 37373, 233, 6, 27, "Input",ExpressionUUID->"80849fe7-b459-4f78-a114-20adf5e0c1f8",
 CellID->1609809464],
Cell[1395989, 37381, 1089, 20, 22, "Print",ExpressionUUID->"9401ed46-fc0a-4328-829f-3c4f0645ee71",
 CellID->348533088],
Cell[1397081, 37403, 17992, 547, 32, "Output",ExpressionUUID->"2dfd0cc7-a164-4cb8-b1d6-8b99cc06e35c",
 CellID->1607704647]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[1415134, 37957, 286, 4, 39, "Subsection",ExpressionUUID->"91c17308-5319-495d-97c0-9bf71d378eba",
 CellID->307506738],
Cell[CellGroupData[{
Cell[1415445, 37965, 234, 3, 41, "Subsubsection",ExpressionUUID->"256f3d06-bb3e-426b-a9bd-fc1f6559166b",
 CellID->334234630],
Cell[1415682, 37970, 343, 6, 58, "DefinitionBox",ExpressionUUID->"6ce642fe-f734-4948-8fd7-f924ee187d73",
 CellID->1811689505],
Cell[1416028, 37978, 112, 1, 23, "Text",ExpressionUUID->"dea6c7aa-d716-456e-b91e-d9cb04e92dd5",
 CellID->2042363369],
Cell[1416143, 37981, 402, 9, 24, "Text",ExpressionUUID->"f0d186fb-8c5a-4b75-b6c7-098431a50e20",
 CellID->622529135],
Cell[CellGroupData[{
Cell[1416570, 37994, 239, 4, 40, "MathCaption",ExpressionUUID->"9a6f8d2c-b8e1-4c5b-aabc-8fca5f369f95",
 CellID->1942919801],
Cell[CellGroupData[{
Cell[1416834, 38002, 313, 7, 27, "Input",ExpressionUUID->"fbe10e4b-8c78-431f-9403-531838f0b8aa",
 CellID->306843761],
Cell[1417150, 38011, 3185, 94, 32, "Output",ExpressionUUID->"b3dfb33d-7d85-470f-8a1a-ecc5b9d24bb8",
 CellID->1698417242]
}, Open  ]]
}, Open  ]],
Cell[1420362, 38109, 267, 6, 41, "MathCaption",ExpressionUUID->"553bb5fe-d911-4ff7-b4ab-8d080b46dd11",
 CellID->1347482286],
Cell[CellGroupData[{
Cell[1420654, 38119, 242, 5, 40, "MathCaption",ExpressionUUID->"2ed23899-9c44-4230-9e5f-0f4e08aa5f9f",
 CellID->1919723187],
Cell[CellGroupData[{
Cell[1420921, 38128, 353, 9, 27, "Input",ExpressionUUID->"2d38aaff-2117-4c3d-95df-c21801dd1d99",
 CellID->993291381],
Cell[1421277, 38139, 10057, 286, 32, "Output",ExpressionUUID->"118dfd7e-7320-444e-8ba6-790cde7bf667",
 CellID->945950370]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1431383, 38431, 181, 2, 40, "MathCaption",ExpressionUUID->"6c5d0be9-ecf8-47b9-96ae-2d17aec35761",
 CellID->511983772],
Cell[CellGroupData[{
Cell[1431589, 38437, 381, 10, 27, "Input",ExpressionUUID->"b7c939b0-aea5-4e57-90e6-7f9e9e2da16d",
 CellID->2091766122],
Cell[1431973, 38449, 10071, 286, 32, "Output",ExpressionUUID->"069b0125-5679-40ea-af91-526e32de6654",
 CellID->766429849]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1442093, 38741, 173, 2, 40, "MathCaption",ExpressionUUID->"b6bb5788-32c2-4569-9371-f56d936d44f5",
 CellID->1478395],
Cell[CellGroupData[{
Cell[1442291, 38747, 459, 13, 27, "Input",ExpressionUUID->"b1d13d53-7038-4cf0-8cf5-dd071c81d4de",
 CellID->1225064648],
Cell[1442753, 38762, 3496, 100, 32, "Output",ExpressionUUID->"fc66019e-b4e6-4e29-8aa0-ed717e68d75b",
 CellID->779557437]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1446298, 38868, 179, 2, 40, "MathCaption",ExpressionUUID->"689bb334-967c-4dd7-86ab-5e512751f968",
 CellID->279831134],
Cell[CellGroupData[{
Cell[1446502, 38874, 859, 26, 27, "Input",ExpressionUUID->"0a486f8c-c973-4305-949d-cb8b149c4f2b",
 CellID->132682208],
Cell[1447364, 38902, 19358, 556, 32, "Output",ExpressionUUID->"d02bb1e6-0567-454f-a791-20266acfbe65",
 CellID->224113312]
}, Open  ]],
Cell[CellGroupData[{
Cell[1466759, 39463, 653, 19, 27, "Input",ExpressionUUID->"3057850f-0f17-4feb-807b-ca34a1f9ab76",
 CellID->357065364],
Cell[1467415, 39484, 19302, 548, 32, "Output",ExpressionUUID->"a91dd48a-16c7-4dc6-a925-e3a5196d97a0",
 CellID->2133601593]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1486766, 40038, 196, 2, 40, "MathCaption",ExpressionUUID->"8d58c171-da7d-45d8-b0a6-7912e456cb50",
 CellID->436142200],
Cell[CellGroupData[{
Cell[1486987, 40044, 432, 11, 27, "Input",ExpressionUUID->"67ea0e34-d820-4e63-8a77-573160d1b72e",
 CellID->880268189],
Cell[1487422, 40057, 1336, 36, 32, "Output",ExpressionUUID->"74ebc574-a49b-4abf-838a-0943aa930244",
 CellID->760035945]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1488807, 40099, 128, 1, 40, "MathCaption",ExpressionUUID->"f0abf66e-b1a5-4810-83c6-c23c22ef2a8d",
 CellID->1189385151],
Cell[CellGroupData[{
Cell[1488960, 40104, 394, 10, 27, "Input",ExpressionUUID->"bfaaaab2-2530-4843-aaeb-76765b2a8291",
 CellID->1218446581],
Cell[1489357, 40116, 10022, 286, 32, "Output",ExpressionUUID->"65074ebc-72ed-4bd5-8614-f735d2b96a01",
 CellID->691395259]
}, Open  ]],
Cell[CellGroupData[{
Cell[1499416, 40407, 392, 10, 27, "Input",ExpressionUUID->"4b73b2d4-a383-4978-9f40-9b49cda636fc",
 CellID->505395773],
Cell[1499811, 40419, 10047, 286, 32, "Output",ExpressionUUID->"1ac7588d-5955-434c-91b4-98438a2b4f99",
 CellID->246104096]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1509907, 40711, 210, 4, 41, "MathCaption",ExpressionUUID->"39b01977-996d-4a31-8694-545b4e1eac8b",
 CellID->251955103],
Cell[CellGroupData[{
Cell[1510142, 40719, 659, 20, 27, "Input",ExpressionUUID->"e4ad6a50-5f85-4073-a814-57f064618d51",
 CellID->1919147884],
Cell[1510804, 40741, 10738, 304, 32, "Output",ExpressionUUID->"7441d518-0d03-4be5-9660-16ec035aae00",
 CellID->218037350]
}, Open  ]]
}, Open  ]],
Cell[1521569, 41049, 954, 22, 78, "Text",ExpressionUUID->"17311fc4-58bc-40ff-b377-8e9c024377d4",
 CellID->1677179967],
Cell[CellGroupData[{
Cell[1522548, 41075, 188, 2, 40, "MathCaption",ExpressionUUID->"6f5d28cd-880a-48d9-ac6e-f60137683129",
 CellID->1292931946],
Cell[CellGroupData[{
Cell[1522761, 41081, 627, 19, 27, "Input",ExpressionUUID->"4fa29e41-8938-4aee-bf26-95bd3ff19ed3",
 CellID->493354847],
Cell[1523391, 41102, 10054, 286, 32, "Output",ExpressionUUID->"be918f10-a1e2-4492-8e73-0f5f98995554",
 CellID->946639668]
}, Open  ]],
Cell[CellGroupData[{
Cell[1533482, 41393, 513, 15, 27, "Input",ExpressionUUID->"240e0b44-c50a-420e-bf9a-537cba251d9c",
 CellID->738928068],
Cell[1533998, 41410, 10056, 286, 32, "Output",ExpressionUUID->"f86f11db-9eb2-438c-85dc-3a7ddc1bb92a",
 CellID->1614043100]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1544103, 41702, 381, 8, 56, "MathCaption",ExpressionUUID->"4799cf53-93c2-4548-b41c-02cfdba862f4",
 CellID->1420824871],
Cell[CellGroupData[{
Cell[1544509, 41714, 350, 9, 27, "Input",ExpressionUUID->"c0a5f5b7-a144-4d2f-8e81-4657f1b78b56",
 CellID->1932935904],
Cell[1544862, 41725, 1469, 39, 32, "Output",ExpressionUUID->"56698c26-ab0f-4d5b-8dcf-d83f178043e6",
 CellID->208952796]
}, Open  ]],
Cell[CellGroupData[{
Cell[1546368, 41769, 167, 3, 27, "Input",ExpressionUUID->"917eaa03-5db4-4ca2-9623-69cdf0b8e81b",
 CellID->933449985],
Cell[1546538, 41774, 10060, 286, 32, "Output",ExpressionUUID->"7ec7bc3e-19d3-4d0e-90f0-5e292ecdbcd4",
 CellID->733879527]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1556647, 42066, 222, 4, 40, "MathCaption",ExpressionUUID->"33fc9932-6e5c-4c8e-92a2-a825703ad7cc",
 CellID->792790203],
Cell[CellGroupData[{
Cell[1556894, 42074, 562, 17, 27, "Input",ExpressionUUID->"2407c5a8-5327-46f5-b5e1-7f08c87af4f0",
 CellID->1436395833],
Cell[1557459, 42093, 1376, 36, 32, "Output",ExpressionUUID->"559049d9-77dc-42cf-b4d7-0261c413ac6d",
 CellID->1314625568]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1558884, 42135, 239, 6, 41, "MathCaption",ExpressionUUID->"ab0f5ff5-09f9-4ff7-8007-ef3c2c6c2119",
 CellID->1655453907],
Cell[CellGroupData[{
Cell[1559148, 42145, 321, 7, 27, "Input",ExpressionUUID->"065bcdc8-d017-464e-adf7-deb839def043",
 CellID->415684417],
Cell[1559472, 42154, 1354, 36, 32, "Output",ExpressionUUID->"00746fdf-45ef-44ba-adde-1508d1b7117e",
 CellID->1600161040]
}, Open  ]],
Cell[CellGroupData[{
Cell[1560863, 42195, 337, 8, 27, "Input",ExpressionUUID->"cb10128b-ed67-43b4-b609-bb86f72e1731",
 CellID->1670230257],
Cell[1561203, 42205, 3192, 94, 32, "Output",ExpressionUUID->"29e66191-ea04-4afa-9e86-e5b0c6269509",
 CellID->968887899]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1564444, 42305, 170, 2, 40, "MathCaption",ExpressionUUID->"8c2b2c07-3538-4cd4-8d47-454bd16e15a3",
 CellID->1493729722],
Cell[CellGroupData[{
Cell[1564639, 42311, 430, 11, 27, "Input",ExpressionUUID->"b88d5f8e-c836-49de-9a87-be58be26e1dd",
 CellID->1031604224],
Cell[1565072, 42324, 10035, 286, 32, "Output",ExpressionUUID->"e11567ce-a08a-40be-8258-f1cf71fe7c9b",
 CellID->1877906939]
}, Open  ]],
Cell[CellGroupData[{
Cell[1575144, 42615, 428, 11, 27, "Input",ExpressionUUID->"58a23fdb-c11a-43b0-9fb9-c07e8f192fd9",
 CellID->1855656537],
Cell[1575575, 42628, 10045, 286, 32, "Output",ExpressionUUID->"af24a6a9-25d2-4280-882b-05c919f9b520",
 CellID->414763814]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1585669, 42920, 304, 7, 41, "MathCaption",ExpressionUUID->"96741730-4511-4908-81d1-46844edb4f98",
 CellID->924250181],
Cell[CellGroupData[{
Cell[1585998, 42931, 635, 18, 27, "Input",ExpressionUUID->"a01c276b-3c48-433e-8d5c-d316b1fec3d7",
 CellID->1123573316],
Cell[1586636, 42951, 34888, 933, 82, "Output",ExpressionUUID->"ebe3a2eb-9c86-4e02-a6ac-b18fd376157b",
 CellID->203503736]
}, Open  ]],
Cell[CellGroupData[{
Cell[1621561, 43889, 169, 3, 27, "Input",ExpressionUUID->"674c249b-e869-4535-aec4-6678bee89a06",
 CellID->1725122528],
Cell[1621733, 43894, 76405, 2086, 80, "Output",ExpressionUUID->"9ae8bc88-ba16-4153-8a72-3e3fd9fe41ae",
 CellID->1493324928]
}, Open  ]]
}, Open  ]],
Cell[1698165, 45984, 622, 17, 42, "MathCaption",ExpressionUUID->"4d8a8a9f-a481-4b6e-add1-68202ae98216",
 CellID->1307962225],
Cell[CellGroupData[{
Cell[1698812, 46005, 158, 2, 40, "MathCaption",ExpressionUUID->"9ed37ed2-f8da-44e5-9c2a-5bf264ca1685",
 CellID->1278504883],
Cell[CellGroupData[{
Cell[1698995, 46011, 489, 13, 27, "Input",ExpressionUUID->"00771092-e168-4e7e-9fa9-56e7de2b0166",
 CellID->110196699],
Cell[1699487, 46026, 12391, 359, 32, "Output",ExpressionUUID->"7f05672e-95a7-4ac1-8bef-7dc2a4793217",
 CellID->670682407]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1711927, 46391, 265, 5, 40, "MathCaption",ExpressionUUID->"be925756-e373-4e37-906a-7fad107dbc08",
 CellID->101228641],
Cell[CellGroupData[{
Cell[1712217, 46400, 479, 13, 27, "Input",ExpressionUUID->"bdf73f85-7693-432f-be78-2125d0961215",
 CellID->542720863],
Cell[1712699, 46415, 7549, 207, 82, "Output",ExpressionUUID->"049e18c5-a65e-4ee0-b410-e703b2943f67",
 CellID->17551346]
}, Open  ]]
}, Open  ]],
Cell[1720275, 46626, 558, 11, 40, "Text",ExpressionUUID->"9a804f81-43a4-4c2e-a320-bbe5d64d36c8",
 CellID->1928806140]
}, Closed]],
Cell[1720848, 46640, 303, 4, 41, "Subsubsection",ExpressionUUID->"1370a689-69b9-4350-86c5-fcf7ccf3e772",
 CellID->1179143515]
}, Closed]],
Cell[CellGroupData[{
Cell[1721188, 46649, 316, 5, 39, "Subsection",ExpressionUUID->"bf8a23c1-c6a3-4c9a-9947-82b3277d77db"],
Cell[CellGroupData[{
Cell[1721529, 46658, 290, 4, 41, "Subsubsection",ExpressionUUID->"fb3b1d67-18c9-47cc-8570-50ccfec54332"],
Cell[1721822, 46664, 524, 10, 92, "DefinitionBox",ExpressionUUID->"74ab3fbb-d0e0-409d-8631-e0e5123e9fac"],
Cell[1722349, 46676, 93, 0, 23, "Text",ExpressionUUID->"04f2cbef-3c0a-4be6-9e9d-d48ec264fe11"],
Cell[1722445, 46678, 2470, 79, 93, "Text",ExpressionUUID->"d47a1a03-58bf-4b62-ae1c-5c235726fe35"],
Cell[CellGroupData[{
Cell[1724940, 46761, 268, 3, 40, "MathCaption",ExpressionUUID->"14b3003d-f554-43e0-977f-89e94c4e3efe",
 CellID->1161546665],
Cell[CellGroupData[{
Cell[1725233, 46768, 430, 12, 27, "Input",ExpressionUUID->"ed4d8f17-15ea-47dd-bcab-170938f72991"],
Cell[1725666, 46782, 2527, 70, 32, "Output",ExpressionUUID->"e7c34ff0-99e6-4d6d-b313-41f914f3529f",
 CellID->964835428]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1728242, 46858, 229, 3, 40, "MathCaption",ExpressionUUID->"185d76a5-30d5-4edc-a9f5-f96b6ac9c6aa"],
Cell[CellGroupData[{
Cell[1728496, 46865, 1283, 40, 85, "Input",ExpressionUUID->"4396c0c7-a62d-4ef6-a1c0-45f107b5a961"],
Cell[1729782, 46907, 629, 11, 26, "Output",ExpressionUUID->"8d31b179-551a-41b0-8873-0d11a4a2a404",
 CellID->340019899]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1730460, 46924, 134, 0, 40, "MathCaption",ExpressionUUID->"37fba9d6-71e7-452b-a456-01a40c8cdd3a"],
Cell[CellGroupData[{
Cell[1730619, 46928, 400, 11, 27, "Input",ExpressionUUID->"e0b6311e-131d-4b16-861f-a2732904b3a0"],
Cell[1731022, 46941, 1424, 35, 32, "Output",ExpressionUUID->"f5d7bc51-3def-45a2-85b2-1a52b54ef786",
 CellID->260457826]
}, Open  ]]
}, Open  ]],
Cell[1732473, 46980, 977, 16, 92, "Text",ExpressionUUID->"e3602400-dc7f-4031-b615-65c120129823"],
Cell[CellGroupData[{
Cell[1733475, 47000, 189, 2, 40, "MathCaption",ExpressionUUID->"e59b8230-2ee5-42ce-a6c0-1632b86a9e9f",
 CellID->1185430083],
Cell[CellGroupData[{
Cell[1733689, 47006, 429, 12, 27, "Input",ExpressionUUID->"a54f58ac-59ae-4bb6-ad3e-b65d3bdb6919"],
Cell[1734121, 47020, 2434, 69, 32, "Output",ExpressionUUID->"25cb10e8-1285-4442-acef-570e0f7ca5b7",
 CellID->943008996]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1736604, 47095, 381, 7, 56, "MathCaption",ExpressionUUID->"7bb252c9-8ba4-4561-8a4e-846ece9e4d4d"],
Cell[CellGroupData[{
Cell[1737010, 47106, 144, 3, 27, "Input",ExpressionUUID->"02facd93-f3a9-43b6-979b-1f7a27c3d8a4"],
Cell[1737157, 47111, 619, 11, 26, "Output",ExpressionUUID->"4314cf9e-c8ee-43ae-964f-16f9212cb173",
 CellID->722880912]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1737825, 47128, 311, 9, 41, "MathCaption",ExpressionUUID->"34dddd89-5f25-4f76-b7bc-a8ee870a622e"],
Cell[CellGroupData[{
Cell[1738161, 47141, 379, 9, 27, "Input",ExpressionUUID->"78ae2b9b-dbd0-42d4-b2c3-18e9aa6b2481"],
Cell[1738543, 47152, 568, 9, 26, "Output",ExpressionUUID->"a23da994-ae82-4ffb-83dc-9c4ff32800cf",
 CellID->1144496905]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1739160, 47167, 101, 0, 23, "Text",ExpressionUUID->"9ed8ea1c-31fa-49ab-8bcb-9e158bf253da"],
Cell[CellGroupData[{
Cell[1739286, 47171, 289, 7, 27, "Input",ExpressionUUID->"5d5b0b0a-de0a-4fd2-b045-31f1c0ba3e54"],
Cell[1739578, 47180, 902, 16, 22, "Print",ExpressionUUID->"21665ddb-161d-406c-a456-be506dc15de6",
 CellID->548456156],
Cell[1740483, 47198, 835, 20, 28, "Output",ExpressionUUID->"8241cd92-b3c7-46bd-9c43-6717a3fdb1d3",
 CellID->902800968]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1741367, 47224, 216, 3, 40, "MathCaption",ExpressionUUID->"431b34bd-cb1f-478e-8796-fb15ee589f08"],
Cell[CellGroupData[{
Cell[1741608, 47231, 144, 3, 27, "Input",ExpressionUUID->"271545d4-e9c0-43fa-94bd-a37c24f9b26d"],
Cell[1741755, 47236, 574, 9, 26, "Output",ExpressionUUID->"501e1d33-d5f6-4510-9a62-54e7b00f34dc",
 CellID->1403903121]
}, Open  ]],
Cell[CellGroupData[{
Cell[1742366, 47250, 431, 11, 27, "Input",ExpressionUUID->"24596909-d185-49f0-976b-5dc0cf7df19e"],
Cell[1742800, 47263, 545, 8, 26, "Output",ExpressionUUID->"652af091-a720-43ff-8b23-d266bbbf4dd2",
 CellID->996925058]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1743394, 47277, 167, 3, 41, "MathCaption",ExpressionUUID->"b99d7649-2df5-4466-97e6-6e483720fc64"],
Cell[CellGroupData[{
Cell[1743586, 47284, 420, 12, 27, "Input",ExpressionUUID->"c4351d2b-cfa6-4687-b65f-6ab086b02c07"],
Cell[1744009, 47298, 1998, 55, 32, "Output",ExpressionUUID->"450246e7-fcfc-418b-a043-a43f526c1b7b",
 CellID->652035912]
}, Open  ]],
Cell[CellGroupData[{
Cell[1746044, 47358, 432, 12, 27, "Input",ExpressionUUID->"b84656c0-60d8-43ee-a023-996654ffbd93"],
Cell[1746479, 47372, 5338, 152, 32, "Output",ExpressionUUID->"cfcf1ec8-4ca0-47af-bc5b-05c0a835f9b9",
 CellID->158522343]
}, Open  ]],
Cell[CellGroupData[{
Cell[1751854, 47529, 255, 7, 27, "Input",ExpressionUUID->"5a98f0b8-352a-4f23-af7c-bf6e646fdc12"],
Cell[1752112, 47538, 3410, 95, 32, "Output",ExpressionUUID->"ccb9b423-b386-413e-9da1-1329d0aa03e8",
 CellID->1405296521]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[1755583, 47640, 266, 4, 41, "Subsubsection",ExpressionUUID->"5517a51f-3f93-4b9d-b5fc-8bd9fd3d7328"],
Cell[1755852, 47646, 250, 6, 41, "DefinitionBox",ExpressionUUID->"ce361ebb-e766-46f2-9ed6-359ba6ef7034"],
Cell[1756105, 47654, 83, 0, 23, "Text",ExpressionUUID->"fed624a6-ada2-4fc2-9ac7-d778d9a990c1"],
Cell[CellGroupData[{
Cell[1756213, 47658, 176, 3, 40, "MathCaption",ExpressionUUID->"1ee6647d-fcf0-431d-89d3-dd6e56aa114a"],
Cell[CellGroupData[{
Cell[1756414, 47665, 293, 7, 27, "Input",ExpressionUUID->"046b7ec2-b105-4604-b965-62d269f2b606"],
Cell[1756710, 47674, 4835, 146, 40, "Output",ExpressionUUID->"5c8574a2-f98f-458b-9f6a-b7bc131c45a5",
 CellID->1137660123]
}, Open  ]],
Cell[CellGroupData[{
Cell[1761582, 47825, 144, 3, 27, "Input",ExpressionUUID->"68c42f31-701f-4bee-b53c-795b37a11c55"],
Cell[1761729, 47830, 567, 10, 26, "Output",ExpressionUUID->"7777eebe-9070-4b80-a118-ee765acb7cbb",
 CellID->1606778020]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1762345, 47846, 279, 6, 40, "MathCaption",ExpressionUUID->"f78e620b-acda-4245-b00f-2488b34148f7"],
Cell[CellGroupData[{
Cell[1762649, 47856, 304, 8, 27, "Input",ExpressionUUID->"cda11638-b2eb-4771-ba40-f008d61c2e6f"],
Cell[1762956, 47866, 1040, 19, 22, "Print",ExpressionUUID->"94fdcefb-5765-447b-8ea6-2d75f4a67c55",
 CellID->511321115]
}, Open  ]],
Cell[CellGroupData[{
Cell[1764033, 47890, 319, 8, 27, "Input",ExpressionUUID->"e0ad47c0-69fe-45f1-92a4-f060b74ff0cd"],
Cell[1764355, 47900, 10667, 329, 40, "Output",ExpressionUUID->"c3a75e22-732d-4004-9bd3-8dd35fb07671",
 CellID->1691908158]
}, Open  ]],
Cell[CellGroupData[{
Cell[1775059, 48234, 144, 3, 27, "Input",ExpressionUUID->"57e16bda-76e0-483c-a625-10b95c98af0f"],
Cell[1775206, 48239, 524, 8, 26, "Output",ExpressionUUID->"81515935-6e43-4529-a547-22f014e03517",
 CellID->1069441579]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1775779, 48253, 201, 5, 40, "MathCaption",ExpressionUUID->"c4608ec9-3c78-44ef-abf4-25d9a8472eeb"],
Cell[CellGroupData[{
Cell[1776005, 48262, 293, 7, 27, "Input",ExpressionUUID->"e86cb3df-05dd-41d8-ac6a-a89ad460ef59"],
Cell[1776301, 48271, 1049, 19, 22, "Print",ExpressionUUID->"1b44471d-711a-4483-a5b9-c83d386f498c",
 CellID->1813976510]
}, Open  ]],
Cell[CellGroupData[{
Cell[1777387, 48295, 335, 9, 27, "Input",ExpressionUUID->"ce66d50b-96ce-4366-88e7-6b95d92ec609"],
Cell[1777725, 48306, 5861, 178, 40, "Output",ExpressionUUID->"d8aaef10-c90d-4fbe-af02-3a7945b42916",
 CellID->1825350758]
}, Open  ]],
Cell[CellGroupData[{
Cell[1783623, 48489, 147, 3, 27, "Input",ExpressionUUID->"717aa808-c21c-473c-af3f-f0f8e2fe6458"],
Cell[1783773, 48494, 518, 8, 26, "Output",ExpressionUUID->"dfdfc557-41c2-4c38-a24a-04cf5e19d0d7",
 CellID->1866292311]
}, Open  ]],
Cell[CellGroupData[{
Cell[1784328, 48507, 324, 8, 27, "Input",ExpressionUUID->"63ef81ea-9788-45f6-bc02-779cf7eff6b9"],
Cell[1784655, 48517, 1096, 20, 22, "Print",ExpressionUUID->"0038d2a2-a361-465b-a662-665bbe416ee8",
 CellID->628722895]
}, Open  ]],
Cell[CellGroupData[{
Cell[1785788, 48542, 407, 11, 27, "Input",ExpressionUUID->"dbedcf26-963c-4331-b9fa-fe6e6c6933ea"],
Cell[1786198, 48555, 14432, 438, 40, "Output",ExpressionUUID->"c4c9d5c7-dc02-463e-8f98-a55a051b5f74",
 CellID->1326153025]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1800679, 48999, 204, 3, 40, "MathCaption",ExpressionUUID->"89958e6f-5fbd-4853-a72e-6a4f7ed97fb2"],
Cell[CellGroupData[{
Cell[1800908, 49006, 498, 14, 27, "Input",ExpressionUUID->"e6c42b3a-8b82-4028-bdcd-f312501f4739"],
Cell[1801409, 49022, 1461, 37, 32, "Output",ExpressionUUID->"5634f899-488f-47d1-b3e6-23cbd878590c",
 CellID->2068602609]
}, Open  ]],
Cell[CellGroupData[{
Cell[1802907, 49064, 233, 6, 27, "Input",ExpressionUUID->"cfc3cd91-deba-4600-88a6-33745a56dfd6"],
Cell[CellGroupData[{
Cell[1803165, 49074, 906, 17, 22, "Print",ExpressionUUID->"bbc3bfd4-4779-4636-8405-7b21fd588dbc",
 CellID->1391288180],
Cell[1804074, 49093, 908, 17, 22, "Print",ExpressionUUID->"665c9837-2fbb-4aeb-ba7a-00ef22965ea5",
 CellID->1578492724],
Cell[1804985, 49112, 906, 17, 22, "Print",ExpressionUUID->"2be4aebe-ae6e-4589-9dc7-04764a8bd818",
 CellID->1861720877]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[1805964, 49137, 313, 5, 41, "Subsubsection",ExpressionUUID->"08ed6c8f-96d6-4476-b7e7-d08c642f69c5"],
Cell[1806280, 49144, 751, 11, 109, "DefinitionBox",ExpressionUUID->"8c8d68fd-236d-44bf-af7b-3c24667ca857"],
Cell[1807034, 49157, 190, 3, 23, "Text",ExpressionUUID->"5105a549-923b-4ea5-a757-3a48359d7237"],
Cell[CellGroupData[{
Cell[1807249, 49164, 379, 6, 57, "MathCaption",ExpressionUUID->"ce172b65-2844-4ea8-b50c-3d50b2e93fcf"],
Cell[CellGroupData[{
Cell[1807653, 49174, 652, 17, 27, "Input",ExpressionUUID->"dc93e8bb-70c6-43e2-a7b8-02251ad53bf8"],
Cell[CellGroupData[{
Cell[1808330, 49195, 1044, 21, 22, "Print",ExpressionUUID->"12cf3852-20ca-4b93-a629-98596fe330a0",
 CellID->220771680],
Cell[1809377, 49218, 1102, 23, 22, "Print",ExpressionUUID->"1089b79d-0f1e-40f7-8e01-7e507598e02a",
 CellID->1805367391],
Cell[1810482, 49243, 963, 19, 22, "Print",ExpressionUUID->"8a0a04c4-67af-4a95-b85b-89a1c813c827",
 CellID->1013292074],
Cell[1811448, 49264, 1082, 22, 22, "Print",ExpressionUUID->"acd942e6-516d-4aa8-bd06-32ea0e885eaf",
 CellID->854567008],
Cell[1812533, 49288, 1100, 23, 22, "Print",ExpressionUUID->"54f2a242-e079-4505-9005-8aab8226bb34",
 CellID->337072320],
Cell[1813636, 49313, 1126, 23, 22, "Print",ExpressionUUID->"a22bf354-1fe1-4bf2-9fb8-91f4211cb35f",
 CellID->112621454],
Cell[1814765, 49338, 1046, 21, 22, "Print",ExpressionUUID->"d88ea636-1de4-4542-88d9-a7e380e683c0",
 CellID->895684371],
Cell[1815814, 49361, 554, 8, 22, "Print",ExpressionUUID->"f779c985-a204-420c-b48b-630f38aaf9f9",
 CellID->2054341574],
Cell[1816371, 49371, 927, 17, 22, "Print",ExpressionUUID->"65de0311-db07-41fd-a4d6-206e365601ee",
 CellID->76757752],
Cell[1817301, 49390, 556, 8, 22, "Print",ExpressionUUID->"b387ee00-c2b7-48f9-a4a0-a06511077084",
 CellID->1577349664],
Cell[1817860, 49400, 1063, 22, 22, "Print",ExpressionUUID->"a07d72c8-2dd7-4bbf-b03c-ca868444d488",
 CellID->1977664280],
Cell[1818926, 49424, 1135, 23, 22, "Print",ExpressionUUID->"0d721454-298f-446e-b758-17cb9bb2bedb",
 CellID->1999149341],
Cell[1820064, 49449, 1054, 21, 22, "Print",ExpressionUUID->"541f3685-75f9-4cd6-94c9-76b8b8428b78",
 CellID->173833225],
Cell[1821121, 49472, 943, 17, 22, "Print",ExpressionUUID->"c79da98e-8db3-4437-8f75-a2df17e40c4f",
 CellID->1037252655],
Cell[1822067, 49491, 687, 12, 22, "Print",ExpressionUUID->"668fe0d6-1e96-4567-9ffe-8d1e1d63497e",
 CellID->1980609886],
Cell[1822757, 49505, 679, 12, 22, "Print",ExpressionUUID->"e50dd1e1-627d-47b8-a5fb-6d3f16646765",
 CellID->2113825502],
Cell[1823439, 49519, 693, 12, 22, "Print",ExpressionUUID->"1c26a392-4e91-4378-b46e-660f8aa71e26",
 CellID->1243239793],
Cell[1824135, 49533, 954, 17, 22, "Print",ExpressionUUID->"1dececb5-a6c8-46ee-910a-3be27b28079f",
 CellID->115315641]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1825138, 49556, 140, 3, 27, "Input",ExpressionUUID->"5e412601-0c66-493f-b76b-c632608e689d"],
Cell[1825281, 49561, 836, 20, 29, "Output",ExpressionUUID->"86b01d45-8954-45c9-a3f3-6a446eada101",
 CellID->1079728693]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1826166, 49587, 97, 0, 40, "MathCaption",ExpressionUUID->"52d194c2-576a-4d83-907e-d0f691c74315"],
Cell[CellGroupData[{
Cell[1826288, 49591, 268, 6, 27, "Input",ExpressionUUID->"a1372ca5-53aa-45be-aaf1-202a458cd20b"],
Cell[1826559, 49599, 965, 17, 22, "Print",ExpressionUUID->"27c93d52-f301-4620-b6c9-5ec74655f7b6",
 CellID->763326101],
Cell[1827527, 49618, 830, 20, 29, "Output",ExpressionUUID->"ae418ec8-30cf-4a7f-ad0f-060878bd1755",
 CellID->154472329]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1828406, 49644, 292, 6, 40, "MathCaption",ExpressionUUID->"da6a3990-a9a3-402b-8954-32ea402bca5e"],
Cell[CellGroupData[{
Cell[1828723, 49654, 182, 4, 27, "Input",ExpressionUUID->"22500a6b-149e-4031-aaca-fb4f124bfce5"],
Cell[1828908, 49660, 1361, 34, 32, "Output",ExpressionUUID->"417e2e6b-0351-489f-a732-904772e59c2b",
 CellID->2056187927]
}, Open  ]],
Cell[CellGroupData[{
Cell[1830306, 49699, 144, 3, 27, "Input",ExpressionUUID->"f971d30c-b3be-4291-9a5a-bc71f2a15b84"],
Cell[1830453, 49704, 554, 8, 26, "Output",ExpressionUUID->"38476243-64d2-413f-9e6d-527e966c4786",
 CellID->1043641565]
}, Open  ]],
Cell[CellGroupData[{
Cell[1831044, 49717, 422, 10, 27, "Input",ExpressionUUID->"4316b4ea-0107-4db2-9dc5-626be467ebdb"],
Cell[1831469, 49729, 959, 17, 22, "Print",ExpressionUUID->"da78f669-2bdc-4065-a824-380a655aa1c0",
 CellID->1984837906],
Cell[1832431, 49748, 1890, 52, 53, "Output",ExpressionUUID->"739959c9-3395-4e0c-af25-62471b9f2540",
 CellID->159713644]
}, Open  ]],
Cell[CellGroupData[{
Cell[1834358, 49805, 301, 7, 27, "Input",ExpressionUUID->"0539c1c0-5a28-485e-9b91-d30ca7209aa5"],
Cell[1834662, 49814, 1793, 50, 33, "Output",ExpressionUUID->"bd7ad3bb-31be-45e5-a910-0af47c6e5c96",
 CellID->2100139605]
}, Open  ]],
Cell[CellGroupData[{
Cell[1836492, 49869, 300, 7, 27, "Input",ExpressionUUID->"27c8ea32-d4df-4a30-9d91-66111c5675b4"],
Cell[1836795, 49878, 1434, 37, 32, "Output",ExpressionUUID->"d506648c-4660-4180-b745-8dc38590bb0b",
 CellID->1005119143]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1838278, 49921, 211, 3, 40, "MathCaption",ExpressionUUID->"c4e154b2-0aeb-4571-b328-ffadd6d548a0"],
Cell[CellGroupData[{
Cell[1838514, 49928, 277, 6, 27, "Input",ExpressionUUID->"5becb215-6358-4923-b8b9-22e02bced121"],
Cell[1838794, 49936, 551, 8, 26, "Output",ExpressionUUID->"cf83ba9f-5a92-4588-8364-f04d6f579fcd",
 CellID->931069099]
}, Open  ]],
Cell[CellGroupData[{
Cell[1839382, 49949, 235, 4, 27, "Input",ExpressionUUID->"6d715067-b856-4a86-8dd0-72127094ca28",
 CellID->859527540],
Cell[CellGroupData[{
Cell[1839642, 49957, 811, 15, 22, "Print",ExpressionUUID->"8ead0fa0-a7b3-423e-b0ff-0d1e795c0fc3",
 CellID->1912494851],
Cell[1840456, 49974, 809, 15, 22, "Print",ExpressionUUID->"70f2e64e-6b0f-440a-9afd-f117c0027aa0",
 CellID->1913216108],
Cell[1841268, 49991, 803, 15, 22, "Print",ExpressionUUID->"b1f51af3-05ea-41af-9daa-cea7066b3010",
 CellID->1150480625],
Cell[1842074, 50008, 830, 15, 22, "Print",ExpressionUUID->"85513c16-34a6-4d9d-846f-4d00831b9c6a",
 CellID->185106787],
Cell[1842907, 50025, 819, 15, 22, "Print",ExpressionUUID->"bd7a2902-aec0-4e91-b06c-ac89a501f889",
 CellID->1533448253],
Cell[1843729, 50042, 810, 15, 22, "Print",ExpressionUUID->"6b9d1ba8-79b7-4a83-8a4a-bd7a3bd1761a",
 CellID->713540524],
Cell[1844542, 50059, 808, 15, 22, "Print",ExpressionUUID->"6ec40b58-4209-4998-8190-91056e8bbd79",
 CellID->926902020],
Cell[1845353, 50076, 800, 15, 22, "Print",ExpressionUUID->"0055860d-d7a8-4ea7-84ac-413edc4db234",
 CellID->595874802],
Cell[1846156, 50093, 797, 15, 22, "Print",ExpressionUUID->"7d5052ec-b0d7-4750-b76b-0400efe8a8ab",
 CellID->1151551253],
Cell[1846956, 50110, 816, 15, 22, "Print",ExpressionUUID->"c447ee50-8a51-4bdd-b6b6-74c5f23167fe",
 CellID->999282407],
Cell[1847775, 50127, 797, 15, 22, "Print",ExpressionUUID->"a25e2713-996c-4177-8fb3-b38b9f1e2685",
 CellID->1617294909],
Cell[1848575, 50144, 783, 15, 22, "Print",ExpressionUUID->"853d1766-95bf-4d37-92bd-b299f7a71590",
 CellID->1175015484],
Cell[1849361, 50161, 788, 15, 22, "Print",ExpressionUUID->"8a3a4415-e547-4bbd-a098-6cd416d90f81",
 CellID->225621776],
Cell[1850152, 50178, 815, 15, 22, "Print",ExpressionUUID->"40cb1ffb-5755-4a82-9a17-f77b235bcb17",
 CellID->1716552277],
Cell[1850970, 50195, 806, 15, 22, "Print",ExpressionUUID->"c74045aa-c41f-485e-8dad-06c98e5ab153",
 CellID->798805741]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Closed]]
}, Closed]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1851873, 50220, 541, 10, 54, "Section",ExpressionUUID->"2c215748-a3b5-4b64-a43d-34271bae9190",
 CellTags->{"S5.18.1", "1.4"},
 CellID->256747671],
Cell[CellGroupData[{
Cell[1852439, 50234, 313, 4, 39, "Subsection",ExpressionUUID->"7d2ec21e-6cc6-48a6-86ff-67582dbed596",
 CellID->2126653677],
Cell[CellGroupData[{
Cell[1852777, 50242, 363, 4, 41, "Subsubsection",ExpressionUUID->"d90e3d53-1dd9-4a77-84ed-9cf49e0d4bec",
 CellID->1988076142],
Cell[1853143, 50248, 1130, 16, 108, "Text",ExpressionUUID->"689c8572-10e6-4141-bb7a-9591821d06e3",
 CellID->205791133],
Cell[1854276, 50266, 610, 9, 57, "Text",ExpressionUUID->"a335faf2-7e92-4082-a8c9-6b68db5c1fb2",
 CellID->1280456169],
Cell[1854889, 50277, 688, 10, 57, "Text",ExpressionUUID->"9bf3975a-c17c-403c-b3bd-5288c83dc305",
 CellID->415868790],
Cell[1855580, 50289, 413, 7, 40, "Text",ExpressionUUID->"e3dace28-2d9a-44f7-96f3-83baf0ba3a82",
 CellID->133220817],
Cell[1855996, 50298, 335, 6, 23, "Text",ExpressionUUID->"2e52efe7-11c6-4a61-a1e9-e3d4a0c7f884",
 CellID->624875189],
Cell[CellGroupData[{
Cell[1856356, 50308, 178, 3, 40, "MathCaption",ExpressionUUID->"ce5ba03b-5b2f-4862-b94c-877bd9c79e3d",
 CellID->2098012318],
Cell[CellGroupData[{
Cell[1856559, 50315, 570, 17, 27, "Input",ExpressionUUID->"3a064422-9148-449e-9be6-89df4a5ab782",
 CellID->1759028533],
Cell[1857132, 50334, 2529, 72, 32, "Output",ExpressionUUID->"93c3e456-f8bf-4ec2-8949-c8baeff63ee5",
 CellID->643272850]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1859710, 50412, 293, 5, 40, "MathCaption",ExpressionUUID->"fdef649d-074a-4c51-ad0f-7cd8f41a5a14",
 CellID->1098520583],
Cell[CellGroupData[{
Cell[1860028, 50421, 404, 11, 27, "Input",ExpressionUUID->"5cc9e161-76d5-4b61-9513-888fa51ea295",
 CellID->918525222],
Cell[1860435, 50434, 502, 8, 26, "Output",ExpressionUUID->"5d551068-876d-4e8e-830a-94fcdfcc0eae",
 CellID->2098680252]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1860986, 50448, 220, 2, 40, "MathCaption",ExpressionUUID->"7ab848b3-00c2-4524-85bf-00b7037475b7",
 CellID->119380206],
Cell[CellGroupData[{
Cell[1861231, 50454, 353, 9, 27, "Input",ExpressionUUID->"add5d5c5-6bef-48d4-b90c-b4c8d6d0f3bb",
 CellID->673395944],
Cell[1861587, 50465, 498, 8, 26, "Output",ExpressionUUID->"55cad4f8-66a9-4863-b987-bd72ed215b64",
 CellID->394869652]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1862134, 50479, 217, 2, 40, "MathCaption",ExpressionUUID->"263d1081-aecd-4950-aa5f-93539cbf0f42",
 CellID->39383645],
Cell[CellGroupData[{
Cell[1862376, 50485, 354, 8, 27, "Input",ExpressionUUID->"230d09b7-c761-454f-9243-7effafd41658",
 CellID->495958239],
Cell[1862733, 50495, 1391, 36, 32, "Output",ExpressionUUID->"22333f5d-76a5-4d5d-b367-fe0782a39e89",
 CellID->810790651]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1864173, 50537, 245, 4, 40, "MathCaption",ExpressionUUID->"94cbf179-96e8-4b59-abef-6b2f5b088226",
 CellID->181064170],
Cell[CellGroupData[{
Cell[1864443, 50545, 718, 21, 27, "Input",ExpressionUUID->"bd534649-f103-42c2-9546-84860f0b898b",
 CellID->1652359408],
Cell[1865164, 50568, 1648, 43, 32, "Output",ExpressionUUID->"f15c97fb-e47a-4646-93b3-b7a2ca881b48",
 CellID->2003355641]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1866861, 50617, 446, 11, 57, "MathCaption",ExpressionUUID->"b823a468-74ae-499b-b2ce-d8811f2d93da",
 CellID->1601747045],
Cell[CellGroupData[{
Cell[1867332, 50632, 438, 10, 27, "Input",ExpressionUUID->"1b02c939-8220-40fd-8277-1be5bb3b3fe2",
 CellID->1855070170],
Cell[1867773, 50644, 1573, 41, 32, "Output",ExpressionUUID->"39f806ad-0b7e-4b61-befa-21273b143f0e",
 CellID->589546690]
}, Open  ]],
Cell[CellGroupData[{
Cell[1869383, 50690, 228, 4, 27, "Input",ExpressionUUID->"47e62dc5-946c-470b-8a9d-2b81ad0d5d11",
 CellID->381890607],
Cell[1869614, 50696, 563, 8, 26, "Output",ExpressionUUID->"87225785-11fa-42bf-9274-4420a3e7df2d",
 CellID->766197533]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1870226, 50710, 295, 5, 40, "MathCaption",ExpressionUUID->"0f0b984e-cf9a-4287-9f11-ee05fd1856b0",
 CellID->1447927465],
Cell[CellGroupData[{
Cell[1870546, 50719, 443, 10, 27, "Input",ExpressionUUID->"94f0faea-ca78-4f0f-bf70-7cdbb9cdadcf",
 CellID->1934179853],
Cell[1870992, 50731, 1590, 42, 32, "Output",ExpressionUUID->"0a9bbfcc-42f5-475c-9bdb-52f4fb1500dd",
 CellID->703900149]
}, Open  ]],
Cell[CellGroupData[{
Cell[1872619, 50778, 224, 4, 27, "Input",ExpressionUUID->"9c102dd0-45e6-4d19-9bd6-6a4255bdf6aa",
 CellID->157390978],
Cell[1872846, 50784, 537, 8, 26, "Output",ExpressionUUID->"a5fa7ce9-bad0-41c4-be1f-04b08bdf91a0",
 CellID->1710995331]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1873432, 50798, 348, 7, 41, "MathCaption",ExpressionUUID->"9402fcd7-8e53-49d0-9b7e-296a5cc65a55",
 CellID->1067825639],
Cell[CellGroupData[{
Cell[1873805, 50809, 359, 9, 27, "Input",ExpressionUUID->"c21c3b2e-fd13-464f-aafd-723aed0d1cfe",
 CellID->424676694],
Cell[1874167, 50820, 1426, 38, 32, "Output",ExpressionUUID->"181d16c0-8a0b-4608-8d05-aafcf3286a4e",
 CellID->1613886747]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1875642, 50864, 445, 9, 57, "MathCaption",ExpressionUUID->"19b003db-2a0b-421c-9aa2-4feae25b1e29",
 CellID->10119455],
Cell[CellGroupData[{
Cell[1876112, 50877, 398, 11, 27, "Input",ExpressionUUID->"43768d32-c248-4828-ad5f-3132364c879b",
 CellID->426472072],
Cell[1876513, 50890, 1509, 40, 32, "Output",ExpressionUUID->"31ead0a5-1954-4ac9-9b78-bdd62ae8f740",
 CellID->661105440]
}, Open  ]],
Cell[CellGroupData[{
Cell[1878059, 50935, 505, 15, 27, "Input",ExpressionUUID->"92e47895-2c66-4af9-bebc-03607ccf188e",
 CellID->119643057],
Cell[1878567, 50952, 1721, 45, 32, "Output",ExpressionUUID->"f7745f14-70be-4fa9-bb56-b723d583e0df",
 CellID->1947674454]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1880337, 51003, 198, 2, 40, "MathCaption",ExpressionUUID->"57c1cfa8-53cb-43e2-8292-9c1f7f321955",
 CellID->1882859465],
Cell[CellGroupData[{
Cell[1880560, 51009, 407, 11, 27, "Input",ExpressionUUID->"be81e70c-e5dc-4837-95c1-5b26f43d10fd",
 CellID->2047489710],
Cell[1880970, 51022, 499, 8, 26, "Output",ExpressionUUID->"44753240-9e8a-4c91-8aee-420af15d78fd",
 CellID->734409410]
}, Open  ]],
Cell[CellGroupData[{
Cell[1881506, 51035, 431, 12, 27, "Input",ExpressionUUID->"e309059a-55a6-4953-9ba8-a284c870ded0",
 CellID->1501720041],
Cell[1881940, 51049, 1679, 44, 32, "Output",ExpressionUUID->"47993797-d38e-42a2-80f5-b391cc78e27d",
 CellID->1063436316]
}, Open  ]],
Cell[CellGroupData[{
Cell[1883656, 51098, 479, 12, 27, "Input",ExpressionUUID->"aaf95caf-d338-4cb0-b9ee-244e582e83e4",
 CellID->331410073],
Cell[1884138, 51112, 1722, 44, 32, "Output",ExpressionUUID->"e8e27757-f9ad-4c5f-b72e-262a66baf916",
 CellID->513451764]
}, Open  ]],
Cell[CellGroupData[{
Cell[1885897, 51161, 585, 16, 27, "Input",ExpressionUUID->"82109728-ed67-43e0-acab-4dfb1eeb1763",
 CellID->2084745468],
Cell[1886485, 51179, 1766, 45, 32, "Output",ExpressionUUID->"ab3cb86f-c295-4133-b79a-5270fefce4dd",
 CellID->1826832301]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[1888312, 51231, 405, 5, 41, "Subsubsection",ExpressionUUID->"2eb07e5a-875d-4e3b-b26f-439bd095966f",
 CellID->1300805695],
Cell[CellGroupData[{
Cell[1888742, 51240, 271, 5, 40, "MathCaption",ExpressionUUID->"02b878e0-a54e-47e8-a3d2-bc1fb60150a7",
 CellID->1295802205],
Cell[CellGroupData[{
Cell[1889038, 51249, 203, 4, 27, "Input",ExpressionUUID->"28927013-dfbb-49f7-b3a4-87a2b787d4d8",
 CellID->915995208],
Cell[1889244, 51255, 14611, 351, 284, "Output",ExpressionUUID->"ed10973b-578b-41d0-9210-b2e73506fb02",
 CellID->470484424]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1903904, 51612, 123, 1, 40, "MathCaption",ExpressionUUID->"49a61b71-c68e-47cb-a0d0-a8670ab0e5a7",
 CellID->579705646],
Cell[CellGroupData[{
Cell[1904052, 51617, 330, 9, 27, "Input",ExpressionUUID->"b140e9d1-d653-4799-92e6-ae367a8be29b",
 CellID->2126969530],
Cell[1904385, 51628, 504, 8, 26, "Output",ExpressionUUID->"8bda260c-48ed-4366-a9fe-16b5ce611d6f",
 CellID->396925914]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1904938, 51642, 188, 4, 40, "MathCaption",ExpressionUUID->"3afa6f5b-6a5e-4e06-8963-9eccd57600c6",
 CellID->910769021],
Cell[CellGroupData[{
Cell[1905151, 51650, 398, 12, 27, "Input",ExpressionUUID->"e9367cb9-4529-4a81-8893-b026ecfe8de2",
 CellID->31500131],
Cell[1905552, 51664, 1544, 41, 32, "Output",ExpressionUUID->"f54b3353-f9de-4a44-aa2c-bb5067d30fa3",
 CellID->1049629016]
}, Open  ]],
Cell[CellGroupData[{
Cell[1907133, 51710, 421, 13, 27, "Input",ExpressionUUID->"5345e42e-6496-44e7-b4aa-77f1521b4896",
 CellID->1198209940],
Cell[1907557, 51725, 1636, 42, 32, "Output",ExpressionUUID->"a75083be-1580-4d8c-8e2b-9022b880888e",
 CellID->1684195454]
}, Open  ]],
Cell[CellGroupData[{
Cell[1909230, 51772, 399, 12, 27, "Input",ExpressionUUID->"55b92aa5-5d43-4250-b633-90a82bd0106c",
 CellID->512724465],
Cell[1909632, 51786, 1575, 41, 32, "Output",ExpressionUUID->"2aac526e-6fda-431c-9476-442f7e01ef0d",
 CellID->291044271]
}, Open  ]],
Cell[CellGroupData[{
Cell[1911244, 51832, 427, 12, 27, "Input",ExpressionUUID->"edf9c63b-bf73-4ad2-8ebb-2e2b049ecfb3",
 CellID->113791411],
Cell[CellGroupData[{
Cell[1911696, 51848, 1110, 21, 22, "Print",ExpressionUUID->"bf126c11-bf2c-4775-bd11-5636046bc134",
 CellID->1760798478],
Cell[1912809, 51871, 1174, 21, 22, "Print",ExpressionUUID->"808f03f8-4da9-4de7-88df-db0fc52af4f1",
 CellID->1415364086]
}, Open  ]],
Cell[1913998, 51895, 1664, 44, 32, "Output",ExpressionUUID->"8173b221-428d-423e-9eb0-5f58ed8515cb",
 CellID->347835583]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1915711, 51945, 150, 1, 40, "MathCaption",ExpressionUUID->"d1a0bdf6-5132-45f4-bf5d-f6c74f709798",
 CellID->1432168523],
Cell[CellGroupData[{
Cell[1915886, 51950, 437, 13, 27, "Input",ExpressionUUID->"cb45c973-6612-4d7f-bc9e-37996d2db8b9",
 CellID->1459650485],
Cell[1916326, 51965, 738, 13, 22, "Message",ExpressionUUID->"1047f1e4-7d75-4d73-8a6d-26cf33ebccfc",
 CellID->622156347]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[1917125, 51985, 251, 3, 41, "Subsubsection",ExpressionUUID->"807c75e1-46b5-4a17-93cf-4e90f5048288",
 CellID->556197953],
Cell[1917379, 51990, 464, 10, 41, "Text",ExpressionUUID->"559bb3f7-da02-49cb-a073-b44f096c2e4b",
 CellID->22680680],
Cell[CellGroupData[{
Cell[1917868, 52004, 271, 5, 40, "MathCaption",ExpressionUUID->"6dde03ad-f4cb-456c-8dd3-537e34b2f126",
 CellID->1494094561],
Cell[CellGroupData[{
Cell[1918164, 52013, 391, 10, 27, "Input",ExpressionUUID->"65855468-2aef-4fe0-910f-2d4049df56ec",
 CellID->1089583241],
Cell[1918558, 52025, 1593, 42, 32, "Output",ExpressionUUID->"257d3d18-77de-458d-b7b5-732bf3d44b7a",
 CellID->65755722]
}, Open  ]],
Cell[CellGroupData[{
Cell[1920188, 52072, 366, 9, 27, "Input",ExpressionUUID->"f52ae658-260a-489e-9e40-d7e302c6845e",
 CellID->1114549266],
Cell[1920557, 52083, 1440, 38, 32, "Output",ExpressionUUID->"86029925-ea73-4413-9626-355929b7ecec",
 CellID->1349825096]
}, Open  ]]
}, Open  ]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1922070, 52129, 187, 4, 39, "Subsection",ExpressionUUID->"5d1d6849-a99b-4de9-a3d0-d37ac7639baf"],
Cell[CellGroupData[{
Cell[1922282, 52137, 381, 5, 41, "Subsubsection",ExpressionUUID->"d93f6ed4-496a-4c48-9f24-6fb16b80c08d",
 CellID->2089523020],
Cell[1922666, 52144, 658, 10, 57, "Text",ExpressionUUID->"0eaf4081-2e47-47bf-ae56-ae4e7f43f715",
 CellID->840881529],
Cell[1923327, 52156, 377, 7, 75, "DefinitionBox",ExpressionUUID->"82eb9af8-9886-4026-b35c-e76f693a5e10"],
Cell[1923707, 52165, 142, 3, 23, "Text",ExpressionUUID->"44455346-3c8a-4578-b8d3-77884d5e069c"],
Cell[CellGroupData[{
Cell[1923874, 52172, 388, 7, 40, "MathCaption",ExpressionUUID->"6bc23be5-6940-43bb-8453-b172c78be34b",
 CellID->1009234187],
Cell[CellGroupData[{
Cell[1924287, 52183, 714, 17, 27, "Input",ExpressionUUID->"fe27af13-9dff-48e3-81e1-f935a4a3101f"],
Cell[CellGroupData[{
Cell[1925026, 52204, 1010, 18, 22, "Print",ExpressionUUID->"f64b1fff-b928-42e9-8a18-5600e25f4154",
 CellID->1224112186],
Cell[1926039, 52224, 1057, 19, 22, "Print",ExpressionUUID->"5921a264-7dd4-47c7-bac0-8d78834a6fae",
 CellID->760171842],
Cell[1927099, 52245, 1057, 19, 22, "Print",ExpressionUUID->"9241dc63-3e65-4665-adcb-5eef1df6b6d9",
 CellID->827932491],
Cell[1928159, 52266, 1058, 19, 22, "Print",ExpressionUUID->"062018ed-ec1b-44a9-959a-766b646cab39",
 CellID->1514827324],
Cell[1929220, 52287, 1020, 18, 22, "Print",ExpressionUUID->"eed51d51-187e-4d69-a316-4b2f93dc3415",
 CellID->1284199283],
Cell[1930243, 52307, 1042, 19, 22, "Print",ExpressionUUID->"1d115aca-6210-453d-8672-359dad2c9dea",
 CellID->1326697031],
Cell[1931288, 52328, 1194, 23, 24, "Print",ExpressionUUID->"e7ac1d4d-864f-4c4f-9020-cbe363c68666",
 CellID->474293216],
Cell[1932485, 52353, 1202, 23, 24, "Print",ExpressionUUID->"fec38bb0-48a0-4f57-ba22-c2002229749a",
 CellID->2134267295],
Cell[1933690, 52378, 1052, 19, 22, "Print",ExpressionUUID->"fdffe5de-3496-4f0e-b023-0f9571fcfc92",
 CellID->1211071224],
Cell[1934745, 52399, 1107, 20, 22, "Print",ExpressionUUID->"c3a1c054-e44c-4c77-bdc1-6e0d27a2c80b",
 CellID->624055502],
Cell[1935855, 52421, 1231, 24, 22, "Print",ExpressionUUID->"a9068d7b-ae68-4e3c-9c58-eb0f20075b7b",
 CellID->190690166],
Cell[1937089, 52447, 1251, 25, 22, "Print",ExpressionUUID->"02fecf1e-c5c5-4e74-9a22-4925f6d98dec",
 CellID->884018638],
Cell[1938343, 52474, 1282, 25, 22, "Print",ExpressionUUID->"30416c28-3b39-4e5a-86f8-c77c14a7c760",
 CellID->1409378920],
Cell[1939628, 52501, 1195, 23, 22, "Print",ExpressionUUID->"005c0e2a-e836-424f-8716-cbd90e18bc18",
 CellID->686882030],
Cell[1940826, 52526, 1181, 22, 22, "Print",ExpressionUUID->"f62814ab-0dda-41ee-a3bf-7c494d966fa2",
 CellID->1701675585],
Cell[1942010, 52550, 1250, 25, 22, "Print",ExpressionUUID->"4360915f-4179-4f2c-90af-3c3b2b33cc8c",
 CellID->2021829996]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1943321, 52582, 251, 5, 40, "MathCaption",ExpressionUUID->"0149e487-24df-4112-88c1-01a3d4f80003",
 CellID->1206102974],
Cell[CellGroupData[{
Cell[1943597, 52591, 282, 5, 27, "Input",ExpressionUUID->"4dc55368-35fb-462e-8fad-26a59c8f7e6d",
 CellID->968864843],
Cell[1943882, 52598, 574, 9, 26, "Output",ExpressionUUID->"0f2bcd17-63ce-42fa-8ee7-81cc57a09b74",
 CellID->1262996291]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1944505, 52613, 165, 3, 40, "MathCaption",ExpressionUUID->"fdbf253a-036e-4b02-8002-2fad677fc882"],
Cell[CellGroupData[{
Cell[1944695, 52620, 368, 9, 27, "Input",ExpressionUUID->"97dc6d95-07e3-4293-a9e6-80739e43d581"],
Cell[1945066, 52631, 571, 9, 26, "Output",ExpressionUUID->"937b8699-50bf-4341-994f-647c82fc212f",
 CellID->1450295889]
}, Open  ]],
Cell[CellGroupData[{
Cell[1945674, 52645, 598, 18, 27, "Input",ExpressionUUID->"57395fc5-5ba1-4375-8214-746988fc9497"],
Cell[1946275, 52665, 575, 9, 26, "Output",ExpressionUUID->"056773b8-cdd3-40f1-a6eb-44f0357f42fc",
 CellID->1183612973]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1946899, 52680, 243, 3, 40, "MathCaption",ExpressionUUID->"29ef79f2-7c73-47dc-ba08-6a7ecaf4c298",
 CellID->1917134628],
Cell[CellGroupData[{
Cell[1947167, 52687, 153, 3, 27, "Input",ExpressionUUID->"5f515c29-fc9a-4e38-bf4e-a667c28835e0"],
Cell[1947323, 52692, 597, 10, 26, "Output",ExpressionUUID->"f4182371-c175-4a18-b48a-445f9ff692ef",
 CellID->842215050]
}, Open  ]],
Cell[CellGroupData[{
Cell[1947957, 52707, 315, 7, 27, "Input",ExpressionUUID->"ab9e62f9-5c1d-46df-9c3d-4d7781053935"],
Cell[1948275, 52716, 624, 11, 26, "Output",ExpressionUUID->"aa76bf58-6c95-41ba-8d5e-ddaf1ba1b910",
 CellID->1043627973]
}, Open  ]],
Cell[CellGroupData[{
Cell[1948936, 52732, 322, 7, 27, "Input",ExpressionUUID->"da1a54cb-6658-4366-b610-770d278eb8de"],
Cell[1949261, 52741, 627, 11, 26, "Output",ExpressionUUID->"f63bb3da-6350-4d8f-8af0-c1c3ed30abcd",
 CellID->1756397162]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1949937, 52758, 246, 4, 40, "MathCaption",ExpressionUUID->"0642d060-3083-49ee-a7c1-1b43c16925d2",
 CellID->1747104672],
Cell[CellGroupData[{
Cell[1950208, 52766, 208, 6, 27, "Input",ExpressionUUID->"05271e5b-d22a-4176-90a6-b2e7d4f66d74"],
Cell[1950419, 52774, 1386, 36, 32, "Output",ExpressionUUID->"9d63c6e7-e948-4110-99e6-d5b4385fd01d",
 CellID->456890072]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1951854, 52816, 284, 5, 40, "MathCaption",ExpressionUUID->"033280e2-a975-4a84-bd5d-0f10c7d41c0b",
 CellID->1681265147],
Cell[CellGroupData[{
Cell[1952163, 52825, 345, 8, 27, "Input",ExpressionUUID->"d1de9970-be80-44cf-8f94-b24ef9909869",
 CellID->1960971562],
Cell[1952511, 52835, 1413, 37, 32, "Output",ExpressionUUID->"3705f40e-9737-4961-b1eb-d946a0a0ccfc",
 CellID->1697402079]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1953973, 52878, 369, 6, 40, "MathCaption",ExpressionUUID->"571ced46-780a-4e73-a257-3c042132bf5e",
 CellID->810815228],
Cell[CellGroupData[{
Cell[1954367, 52888, 235, 4, 27, "Input",ExpressionUUID->"645abf26-050f-492f-baf9-1a48454f0cf5",
 CellID->1169243091],
Cell[CellGroupData[{
Cell[1954627, 52896, 830, 16, 22, "Print",ExpressionUUID->"e05f4522-a60a-4b5c-bcd6-f19eaf6fdb4d",
 CellID->1063975005],
Cell[1955460, 52914, 825, 16, 22, "Print",ExpressionUUID->"f497c4a5-d407-49ed-8674-1d50a83fa9eb",
 CellID->131580466],
Cell[1956288, 52932, 830, 16, 22, "Print",ExpressionUUID->"c162a5c9-c0ae-4157-890a-271f1e5e25ba",
 CellID->1712219299],
Cell[1957121, 52950, 857, 16, 22, "Print",ExpressionUUID->"823e7470-850c-4326-b016-e0923c53207c",
 CellID->988888493],
Cell[1957981, 52968, 832, 16, 22, "Print",ExpressionUUID->"7719fc84-fe53-43b1-b90c-2754ab8f45e1",
 CellID->872957718],
Cell[1958816, 52986, 855, 16, 22, "Print",ExpressionUUID->"571f9ae2-3175-408d-841e-6c895dd78c19",
 CellID->192090122],
Cell[1959674, 53004, 885, 17, 22, "Print",ExpressionUUID->"047f0038-92a6-4c30-89f0-99235577ceb3",
 CellID->790759013],
Cell[1960562, 53023, 838, 16, 22, "Print",ExpressionUUID->"993a695d-a0a0-4bcd-9225-9313770a5368",
 CellID->1657923855],
Cell[1961403, 53041, 848, 16, 22, "Print",ExpressionUUID->"11d08324-3f6c-4ab8-ac47-068f2a499603",
 CellID->1125911160],
Cell[1962254, 53059, 846, 16, 22, "Print",ExpressionUUID->"b8e9b156-5c3e-4729-87ee-b730b3aca8da",
 CellID->1710452403],
Cell[1963103, 53077, 838, 16, 22, "Print",ExpressionUUID->"6f788af2-b1c7-4987-9f35-ced71ba62bba",
 CellID->1357438479],
Cell[1963944, 53095, 857, 16, 22, "Print",ExpressionUUID->"7a082e34-c8bf-418d-9e3b-f079b3575780",
 CellID->1749734472],
Cell[1964804, 53113, 861, 16, 22, "Print",ExpressionUUID->"2536e67a-3350-4c03-95a3-8087cd59528d",
 CellID->146208886]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1965714, 53135, 764, 19, 27, "Input",ExpressionUUID->"46ec0d42-03d5-4767-800e-ea425709de77",
 CellID->1189915789],
Cell[CellGroupData[{
Cell[1966503, 53158, 812, 15, 22, "Print",ExpressionUUID->"d0310852-a8fc-4fc7-86cb-da2588576e75",
 CellID->611454834],
Cell[1967318, 53175, 863, 16, 22, "Print",ExpressionUUID->"2e073ffe-30ad-4594-9c29-b60cbe2d6df9",
 CellID->1012144143],
Cell[1968184, 53193, 862, 16, 22, "Print",ExpressionUUID->"9aa75be1-6ae7-4026-afb2-586c4b7dede6",
 CellID->888643514],
Cell[1969049, 53211, 861, 16, 22, "Print",ExpressionUUID->"cdd6c079-0b49-4477-b718-a42dbd888b05",
 CellID->1830258560],
Cell[1969913, 53229, 822, 15, 22, "Print",ExpressionUUID->"39db6774-a18c-47e8-95f8-c18c05a90538",
 CellID->599858292],
Cell[1970738, 53246, 845, 16, 22, "Print",ExpressionUUID->"0c729604-30f8-4f63-acc1-8d2c0db30ea9",
 CellID->1530074626],
Cell[1971586, 53264, 996, 20, 24, "Print",ExpressionUUID->"dd09ccdc-8dab-4c61-976e-c39f6856f7b9",
 CellID->1995050004],
Cell[1972585, 53286, 1005, 20, 24, "Print",ExpressionUUID->"267fa9be-bdc8-4602-8489-a39bbba78c40",
 CellID->2143236617],
Cell[1973593, 53308, 855, 16, 22, "Print",ExpressionUUID->"152e6ef6-67d1-4ae5-9391-144e3cb3a6b8",
 CellID->1634610788],
Cell[1974451, 53326, 910, 17, 22, "Print",ExpressionUUID->"fbfdd177-fba2-4890-89ab-cc47a721fa7a",
 CellID->145217845],
Cell[1975364, 53345, 1035, 21, 22, "Print",ExpressionUUID->"e3ef4c78-e613-4e63-be2e-3b2f8e401c26",
 CellID->1134897923],
Cell[1976402, 53368, 1054, 22, 22, "Print",ExpressionUUID->"0a86ff13-d8af-47d3-a264-21e77869d924",
 CellID->279639207],
Cell[1977459, 53392, 1085, 22, 22, "Print",ExpressionUUID->"7233746a-8c44-42b7-bf23-bf180043fc81",
 CellID->1201932413],
Cell[1978547, 53416, 999, 20, 22, "Print",ExpressionUUID->"da51a13a-7e8b-4352-aa03-75fde2fba632",
 CellID->1646881609],
Cell[1979549, 53438, 984, 19, 22, "Print",ExpressionUUID->"e358d9e9-7a39-4c69-be48-be28126ef9d7",
 CellID->222476280],
Cell[1980536, 53459, 1052, 22, 22, "Print",ExpressionUUID->"252e6a61-3522-4418-a9c9-3bf07af1f9ff",
 CellID->389808533]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1981649, 53488, 360, 6, 40, "MathCaption",ExpressionUUID->"63652355-9688-4749-83b3-9c5e73407244",
 CellID->1470402736],
Cell[CellGroupData[{
Cell[1982034, 53498, 337, 7, 27, "Input",ExpressionUUID->"c32bdf03-3459-4454-a7d1-dd381c6f730f",
 CellID->1556741925],
Cell[1982374, 53507, 3160, 93, 33, "Output",ExpressionUUID->"05302c71-fdb5-4b09-b086-1751ae0d7ec0",
 CellID->1686890781]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1985583, 53606, 225, 3, 40, "MathCaption",ExpressionUUID->"712b40e7-02fd-4eea-adc3-8e858e94d9b0",
 CellID->1922101666],
Cell[CellGroupData[{
Cell[1985833, 53613, 361, 8, 27, "Input",ExpressionUUID->"4f8374b5-f626-47aa-bb7f-1fd38eee6997",
 CellID->1196988032],
Cell[1986197, 53623, 2945, 87, 32, "Output",ExpressionUUID->"1eed1a2d-e1df-4794-a291-55e40587c278",
 CellID->494248573]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1989191, 53716, 224, 3, 40, "MathCaption",ExpressionUUID->"de30eb0d-09dd-4b2e-bdde-d3357f5e2e27",
 CellID->1981624549],
Cell[CellGroupData[{
Cell[1989440, 53723, 397, 11, 27, "Input",ExpressionUUID->"1a865b04-696f-44ce-b33f-19ec534977ed",
 CellID->79371859],
Cell[1989840, 53736, 2955, 87, 32, "Output",ExpressionUUID->"564650e2-d4b1-4877-9823-8c019fe8ff1b",
 CellID->369019815]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1992844, 53829, 276, 4, 40, "MathCaption",ExpressionUUID->"662820fe-7e52-4032-aedd-ddbc79e45f2b",
 CellID->342164511],
Cell[CellGroupData[{
Cell[1993145, 53837, 240, 4, 27, "Input",ExpressionUUID->"75ab1fbd-2e11-4d3e-b62b-b24d39ba103b",
 CellID->1115327372],
Cell[1993388, 53843, 815, 16, 22, "Print",ExpressionUUID->"675753f9-b01b-494b-ad75-00deee28831a",
 CellID->183258169]
}, Open  ]],
Cell[CellGroupData[{
Cell[1994240, 53864, 492, 13, 27, "Input",ExpressionUUID->"dcebf824-8a2d-4ae4-a5c1-4fd39b00dec5",
 CellID->1740807702],
Cell[1994735, 53879, 5397, 184, 32, "Output",ExpressionUUID->"ee975225-9b93-4155-b3e1-8449ba8ad760",
 CellID->528496585]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[2000193, 54070, 439, 6, 41, "Subsubsection",ExpressionUUID->"06d8f700-ca08-4499-b174-8fbdcbd1bca2",
 CellID->618036468],
Cell[2000635, 54078, 580, 9, 57, "Text",ExpressionUUID->"16c82575-a5f5-4d92-ab74-5bb9b2b9e00c",
 CellID->1917193779],
Cell[2001218, 54089, 801, 12, 57, "Text",ExpressionUUID->"d4a04d56-b5bd-439e-9477-c68fc3ace364",
 CellID->177552500],
Cell[CellGroupData[{
Cell[2002044, 54105, 193, 2, 40, "MathCaption",ExpressionUUID->"da01fe8f-18a0-46df-a371-0a10461d0da8",
 CellID->582647065],
Cell[CellGroupData[{
Cell[2002262, 54111, 266, 5, 27, "Input",ExpressionUUID->"332a3e92-f1cf-4176-b233-1551e1a92613",
 CellID->1224311109],
Cell[2002531, 54118, 576, 13, 26, "Output",ExpressionUUID->"403beafd-f0e7-40c6-9307-50d58f5295a7",
 CellID->833008401]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2003156, 54137, 358, 6, 40, "MathCaption",ExpressionUUID->"dd68cd28-2904-45eb-846b-9117ed84270c",
 CellID->188171900],
Cell[CellGroupData[{
Cell[2003539, 54147, 254, 5, 27, "Input",ExpressionUUID->"f157f915-70f8-4b3d-b640-3c27f0119c65",
 CellID->1431482550],
Cell[2003796, 54154, 1102, 36, 26, "Output",ExpressionUUID->"b0c3b5f0-9c1a-4d6c-b1ab-822b67dda615",
 CellID->1784566159]
}, Open  ]],
Cell[CellGroupData[{
Cell[2004935, 54195, 227, 4, 27, "Input",ExpressionUUID->"37d6cbac-c31a-425b-ab20-704fa8613a08",
 CellID->2028784213],
Cell[2005165, 54201, 399, 6, 26, "Output",ExpressionUUID->"0472a165-0158-4aac-84a2-dd01fa605ee5",
 CellID->1094520716]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2005613, 54213, 269, 3, 40, "MathCaption",ExpressionUUID->"76c2dc90-6cb8-46e9-8d3d-59bf99af9c95",
 CellID->2105785373],
Cell[CellGroupData[{
Cell[2005907, 54220, 361, 9, 27, "Input",ExpressionUUID->"e1cdf93c-5e88-453c-a556-8366dd461e03",
 CellID->19999940],
Cell[2006271, 54231, 436, 8, 26, "Output",ExpressionUUID->"47bf97f0-e7a6-4e2d-bd33-63cfca1e7f78",
 CellID->1057668632]
}, Open  ]],
Cell[CellGroupData[{
Cell[2006744, 54244, 368, 9, 27, "Input",ExpressionUUID->"e3ec0f34-fa8a-4428-892e-756f312e2bb4",
 CellID->87718649],
Cell[2007115, 54255, 439, 8, 26, "Output",ExpressionUUID->"5d7fe42c-3200-4f47-9288-06fc1704757c",
 CellID->361095595]
}, Open  ]]
}, Open  ]],
Cell[2007581, 54267, 344, 6, 40, "MathCaption",ExpressionUUID->"ffaf3618-4430-4a6a-be7c-f66250c7de20",
 CellID->1622491755],
Cell[2007928, 54275, 313, 5, 23, "Text",ExpressionUUID->"b1430c5d-7ac8-453c-981d-f7c559b4f1c8",
 CellID->799280702],
Cell[CellGroupData[{
Cell[2008266, 54284, 258, 3, 40, "MathCaption",ExpressionUUID->"a313a714-5727-41ad-b943-37c58125bd94",
 CellID->2013765975],
Cell[CellGroupData[{
Cell[2008549, 54291, 439, 10, 27, "Input",ExpressionUUID->"3d8e0e67-93d3-4378-b698-1a0aa95a1a12",
 CellID->920876781],
Cell[2008991, 54303, 7870, 199, 82, "Output",ExpressionUUID->"0cb62bed-8bba-4e7d-b533-d7877623588b",
 CellID->196066505]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2016910, 54508, 286, 5, 40, "MathCaption",ExpressionUUID->"a830cae6-a0f0-46d6-8986-f18ab42e36dd",
 CellID->237444473],
Cell[CellGroupData[{
Cell[2017221, 54517, 327, 7, 27, "Input",ExpressionUUID->"1129d404-c72b-4020-be0d-f655bf6f5a07",
 CellID->1975108063],
Cell[2017551, 54526, 26323, 630, 88, "Output",ExpressionUUID->"ceadf765-97e7-4461-afe0-49789ec3a977",
 CellID->325228131]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2043923, 55162, 286, 5, 40, "MathCaption",ExpressionUUID->"cd40ec7e-cdcd-43ed-bc8a-a554d3ceeb01",
 CellID->1221432916],
Cell[2044212, 55169, 254, 5, 27, "Input",ExpressionUUID->"ddf03eb3-189a-4fcb-b242-efa467c9c420",
 CellID->187134635],
Cell[CellGroupData[{
Cell[2044491, 55178, 425, 8, 27, "Input",ExpressionUUID->"e36cd802-d6fc-476d-8923-fac85c609a25",
 CellID->974452277],
Cell[2044919, 55188, 879, 17, 22, "Print",ExpressionUUID->"2a8366f6-e586-4262-bc74-9541ad0acb7e",
 CellID->948957470],
Cell[2045801, 55207, 166578, 3772, 88, "Output",ExpressionUUID->"1bbf0498-83e7-47d6-ba42-041db35cbec4",
 CellID->1914635668]
}, Open  ]],
Cell[2212394, 58982, 256, 5, 27, "Input",ExpressionUUID->"7257562e-d25c-4c8d-b2b3-ac9b30e6d313",
 CellID->707327288]
}, Open  ]],
Cell[CellGroupData[{
Cell[2212687, 58992, 291, 5, 40, "MathCaption",ExpressionUUID->"b7842a20-c47e-4c2e-a794-033c358db9df",
 CellID->1891731754],
Cell[CellGroupData[{
Cell[2213003, 59001, 290, 6, 27, "Input",ExpressionUUID->"108353b5-ea70-46cd-95b7-6ceb2ec6efde",
 CellID->642506035],
Cell[2213296, 59009, 13255, 334, 88, "Output",ExpressionUUID->"2e95a346-7095-4948-96f3-833d2d797722",
 CellID->666835233]
}, Open  ]],
Cell[CellGroupData[{
Cell[2226588, 59348, 279, 6, 27, "Input",ExpressionUUID->"f041a4f8-6ae6-47fe-b39d-f06b36fbb170",
 CellID->272106681],
Cell[2226870, 59356, 42298, 1050, 104, "Output",ExpressionUUID->"57eb0879-8175-4bf1-92e0-b58c233809ea",
 CellID->1920945982]
}, Open  ]],
Cell[2269183, 60409, 254, 5, 27, "Input",ExpressionUUID->"100a758a-a268-4ea6-9907-d6fca04654e0",
 CellID->1092795901],
Cell[CellGroupData[{
Cell[2269462, 60418, 279, 6, 27, "Input",ExpressionUUID->"347ee434-edd7-457e-8e13-f90f11524c9f",
 CellID->408925064],
Cell[2269744, 60426, 238365, 5460, 104, "Output",ExpressionUUID->"93713607-6a67-4cc1-b8e0-6c3808b12271",
 CellID->813641839]
}, Open  ]],
Cell[CellGroupData[{
Cell[2508146, 65891, 256, 5, 27, "Input",ExpressionUUID->"c71a330d-14b5-4c5d-8f2d-264cf926b500",
 CellID->2033365647],
Cell[2508405, 65898, 295167, 6817, 104, "Output",ExpressionUUID->"637b4e77-2b5b-477b-84f9-c661f90c87d8",
 CellID->1732991312]
}, Open  ]],
Cell[2803587, 72718, 189, 4, 27, "Input",ExpressionUUID->"be026b14-5f83-4716-b938-6e1495299782",
 CellID->1259202711]
}, Open  ]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[2803837, 72729, 255, 4, 39, "Subsection",ExpressionUUID->"a6f3d01d-d3d5-4e38-9231-6295fbe85441",
 CellID->969574339],
Cell[CellGroupData[{
Cell[2804117, 72737, 233, 4, 27, "Input",ExpressionUUID->"44b8029e-b934-4298-a4b9-efd86e20f302",
 CellID->99866145],
Cell[CellGroupData[{
Cell[2804375, 72745, 662, 13, 22, "Print",ExpressionUUID->"3d8e4379-848c-48be-840a-dac9f0762408",
 CellID->378126311],
Cell[2805040, 72760, 662, 13, 22, "Print",ExpressionUUID->"c082fa0a-b726-4a60-9769-2e08f28ec031",
 CellID->764797504],
Cell[2805705, 72775, 663, 13, 22, "Print",ExpressionUUID->"f16f9c8e-1db8-4179-9f31-ae410beaf31d",
 CellID->1262022433],
Cell[2806371, 72790, 694, 13, 22, "Print",ExpressionUUID->"5e4ea8db-3d98-4a2a-96fe-a0c481de03ea",
 CellID->115050012],
Cell[2807068, 72805, 670, 13, 22, "Print",ExpressionUUID->"e1bee64c-a166-40f2-8545-021845ade5c9",
 CellID->910269384],
Cell[2807741, 72820, 690, 13, 22, "Print",ExpressionUUID->"de24e7c8-7777-4364-a244-bbfc5584ab5f",
 CellID->224035660],
Cell[2808434, 72835, 686, 13, 22, "Print",ExpressionUUID->"71d2dd17-006c-4e9b-9d57-96452b3a00fa",
 CellID->828504429],
Cell[2809123, 72850, 720, 14, 22, "Print",ExpressionUUID->"b0f58539-2f5d-49c9-a1f1-33bdeb8f3b33",
 CellID->389971055],
Cell[2809846, 72866, 672, 13, 22, "Print",ExpressionUUID->"5f8fbb76-e854-4a5c-944c-38a59116d387",
 CellID->123430741],
Cell[2810521, 72881, 679, 13, 22, "Print",ExpressionUUID->"46888d52-80a7-4c1e-82ed-dbe11f5793e2",
 CellID->31260985],
Cell[2811203, 72896, 681, 13, 22, "Print",ExpressionUUID->"9166e738-7b29-41c5-9c0f-c3d111674cb2",
 CellID->1417365027],
Cell[2811887, 72911, 675, 13, 22, "Print",ExpressionUUID->"ff11bf5b-4de3-43ff-a44e-0d314adadbbd",
 CellID->1765396265],
Cell[2812565, 72926, 695, 13, 22, "Print",ExpressionUUID->"566c7e23-7fbb-4a16-9b55-5cfb6a58fe29",
 CellID->1380958903],
Cell[2813263, 72941, 697, 13, 22, "Print",ExpressionUUID->"9e7f8367-fd62-4738-b271-57c4ba2e3d88",
 CellID->1920509592]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2814009, 72960, 239, 4, 40, "MathCaption",ExpressionUUID->"9cf0465d-82eb-4df9-8609-4ea9f9b2e166",
 CellID->2108054115],
Cell[CellGroupData[{
Cell[2814273, 72968, 246, 4, 27, "Input",ExpressionUUID->"f06ecfab-57d9-4895-9908-9edb0ac82f65",
 CellID->243573842],
Cell[2814522, 72974, 663, 13, 22, "Print",ExpressionUUID->"7b20dc61-927e-4deb-9215-a82c1609ec6a",
 CellID->1325498334]
}, Open  ]],
Cell[CellGroupData[{
Cell[2815222, 72992, 344, 8, 27, "Input",ExpressionUUID->"ac1c6c0e-5fe9-4853-a5cb-92b5102b8670",
 CellID->1956939754],
Cell[CellGroupData[{
Cell[2815591, 73004, 706, 13, 22, "Print",ExpressionUUID->"31d76366-1b55-49cf-ac6d-118371752bf1",
 CellID->1802629207],
Cell[2816300, 73019, 688, 13, 22, "Print",ExpressionUUID->"4fa6e652-b96c-4216-803a-9012a310ab76",
 CellID->89790173],
Cell[2816991, 73034, 686, 13, 22, "Print",ExpressionUUID->"1b9adb9d-fbe1-40dd-b6ad-897c036701ac",
 CellID->1569917417],
Cell[2817680, 73049, 728, 14, 22, "Print",ExpressionUUID->"0ad3bf7a-1460-4647-84c8-b3fd01d6b6f8",
 CellID->1617940477],
Cell[2818411, 73065, 672, 13, 22, "Print",ExpressionUUID->"2d98467d-854e-4d5f-b0eb-022054bcd636",
 CellID->1287658625],
Cell[2819086, 73080, 682, 13, 22, "Print",ExpressionUUID->"4314aba7-a4c3-486a-8e01-2addb0c65a4a",
 CellID->1485799556],
Cell[2819771, 73095, 679, 13, 22, "Print",ExpressionUUID->"a9a3ed9e-a4a1-4a50-bf13-418522c67c39",
 CellID->893290376],
Cell[2820453, 73110, 673, 13, 22, "Print",ExpressionUUID->"3b27cab7-c6e8-49d6-8bf5-8a740973a1cc",
 CellID->473173255],
Cell[2821129, 73125, 691, 13, 22, "Print",ExpressionUUID->"beedab78-6764-4d35-b04d-d97a51a01560",
 CellID->490951944],
Cell[2821823, 73140, 697, 13, 22, "Print",ExpressionUUID->"e47ed9db-9876-4699-ba83-67944cbe107b",
 CellID->797151034],
Cell[2822523, 73155, 640, 13, 22, "Print",ExpressionUUID->"c6b91a09-bdfc-46e8-8760-be5c91ddb5f2",
 CellID->1714438452],
Cell[2823166, 73170, 730, 14, 22, "Print",ExpressionUUID->"de2ad920-57d7-43ec-9d9c-526ba4b5f0ce",
 CellID->1778970965],
Cell[2823899, 73186, 674, 13, 22, "Print",ExpressionUUID->"3d5e47df-a34a-40dc-a8c3-eb69fa2f8928",
 CellID->1897675221],
Cell[2824576, 73201, 682, 13, 22, "Print",ExpressionUUID->"561ebe83-fda0-49e3-8209-d1ce5ff71b8a",
 CellID->1947797750],
Cell[2825261, 73216, 681, 13, 22, "Print",ExpressionUUID->"ae6ecc2e-7122-493f-af1c-a921e1267319",
 CellID->396069382],
Cell[2825945, 73231, 674, 13, 22, "Print",ExpressionUUID->"32cbf2a8-30f2-45b0-8ff9-0a2c8eaef468",
 CellID->1686281221],
Cell[2826622, 73246, 692, 13, 22, "Print",ExpressionUUID->"9c186931-f868-44fa-ae80-b5aed1c0291d",
 CellID->986791888],
Cell[2827317, 73261, 698, 13, 22, "Print",ExpressionUUID->"0782ec81-69ee-4ac3-a41a-48bfc566f1cf",
 CellID->2142539482],
Cell[2828018, 73276, 641, 13, 22, "Print",ExpressionUUID->"895b11f0-2b36-461c-a6af-020f9aab0f77",
 CellID->655281078],
Cell[2828662, 73291, 754, 15, 22, "Print",ExpressionUUID->"a451ace1-2b26-452a-9763-fdb595b55be1",
 CellID->1012936130],
Cell[2829419, 73308, 737, 15, 22, "Print",ExpressionUUID->"a6391ba2-4ddf-4129-9891-32c82b77f749",
 CellID->1314667141],
Cell[2830159, 73325, 729, 14, 22, "Print",ExpressionUUID->"6d3ad999-44d3-45b6-beb4-a3b3bb2f4a91",
 CellID->151394804],
Cell[2830891, 73341, 710, 14, 22, "Print",ExpressionUUID->"5c2503e7-4ab3-4958-b3e8-9ece8900033e",
 CellID->1871832780],
Cell[2831604, 73357, 689, 14, 22, "Print",ExpressionUUID->"c070b5b2-bf54-4286-8a3d-44642d7128f5",
 CellID->1405588557],
Cell[2832296, 73373, 676, 13, 22, "Print",ExpressionUUID->"05f9f6b7-a5bd-4ba0-b358-c4c869da0a64",
 CellID->1406760499],
Cell[2832975, 73388, 682, 13, 22, "Print",ExpressionUUID->"803327b6-6b6d-4188-b626-d5da29f08034",
 CellID->1776236870],
Cell[2833660, 73403, 681, 13, 22, "Print",ExpressionUUID->"fe003b6b-eaca-4424-9110-54177131e550",
 CellID->1042897143],
Cell[2834344, 73418, 673, 13, 22, "Print",ExpressionUUID->"ebe2e56b-a789-4631-99e1-dc7d8ee8cbfa",
 CellID->210755587],
Cell[2835020, 73433, 659, 13, 22, "Print",ExpressionUUID->"5afdfb2f-a144-4024-9d1d-5707a7f74d59",
 CellID->968616806],
Cell[2835682, 73448, 716, 14, 22, "Print",ExpressionUUID->"94e0e5bd-333c-4c43-9695-1e484d4db09b",
 CellID->1616614182],
Cell[2836401, 73464, 693, 13, 22, "Print",ExpressionUUID->"dd4c70ba-ef36-4135-bc61-9d363526874f",
 CellID->471455030],
Cell[2837097, 73479, 720, 14, 22, "Print",ExpressionUUID->"f43191a0-945a-4c20-968f-f5fb12ab8e7b",
 CellID->1677515427],
Cell[2837820, 73495, 698, 13, 22, "Print",ExpressionUUID->"1b63a560-eb93-433a-8530-b701ede44fd1",
 CellID->1836463773],
Cell[2838521, 73510, 644, 13, 22, "Print",ExpressionUUID->"fc2fce22-f260-4425-9b6a-64a8c48ea6c6",
 CellID->1811658127]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2839214, 73529, 357, 9, 27, "Input",ExpressionUUID->"4c0d419c-480d-4aec-a43e-0b6a465060f6",
 CellID->1884771571],
Cell[CellGroupData[{
Cell[2839596, 73542, 689, 14, 22, "Print",ExpressionUUID->"0aee5ed0-b372-4dbf-b4cc-e7f3969e558a",
 CellID->1514993323],
Cell[2840288, 73558, 673, 14, 22, "Print",ExpressionUUID->"1e5351e4-9807-4fb2-97fc-9f833ea36f62",
 CellID->1926660829],
Cell[2840964, 73574, 671, 14, 22, "Print",ExpressionUUID->"405d48d5-32b8-417b-88a4-0c29246bb9cc",
 CellID->1630123537],
Cell[2841638, 73590, 691, 14, 22, "Print",ExpressionUUID->"c17cbc38-8ac1-4df5-b76f-b025d06cd6e3",
 CellID->2037401580],
Cell[2842332, 73606, 670, 14, 22, "Print",ExpressionUUID->"f743b2d6-27bb-49bc-b74b-9c6b7c9ea840",
 CellID->981603645],
Cell[2843005, 73622, 670, 14, 22, "Print",ExpressionUUID->"68926355-1ed8-41c2-8727-3577f0eab204",
 CellID->843085261],
Cell[2843678, 73638, 674, 14, 22, "Print",ExpressionUUID->"3975f6c0-c659-43c0-bcec-2f4a59def999",
 CellID->636878721],
Cell[2844355, 73654, 672, 14, 22, "Print",ExpressionUUID->"8d2649c1-e8b4-47fd-9dd9-329c9265faa9",
 CellID->962167199],
Cell[2845030, 73670, 688, 14, 22, "Print",ExpressionUUID->"ff314803-898d-498a-b26a-65f10e408ea8",
 CellID->592251683],
Cell[2845721, 73686, 670, 14, 22, "Print",ExpressionUUID->"a4a9349c-2d5a-43f3-8b5c-67d0d178771b",
 CellID->387332336]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2846440, 73706, 241, 4, 27, "Input",ExpressionUUID->"671ea665-b64c-4921-b7b1-8e211f238e61",
 CellID->1036853606],
Cell[CellGroupData[{
Cell[2846706, 73714, 843, 19, 22, "Print",ExpressionUUID->"9b26f3ae-b264-470e-b857-33c466e6303e",
 CellID->871637095],
Cell[2847552, 73735, 793, 19, 22, "Print",ExpressionUUID->"af73adc1-b161-415a-a8d9-451e8784ceed",
 CellID->1618699478]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2848394, 73760, 236, 4, 27, "Input",ExpressionUUID->"8d318d61-9dce-42d1-ae6d-d70f7b7729f1",
 CellID->2131866533],
Cell[CellGroupData[{
Cell[2848655, 73768, 816, 19, 22, "Print",ExpressionUUID->"9e3342ee-d13e-4ca3-bd13-30c7027c7991",
 CellID->1159286550],
Cell[2849474, 73789, 646, 13, 22, "Print",ExpressionUUID->"55882254-fda9-413d-bc0b-6e7a024951ca",
 CellID->9916619]
}, Open  ]]
}, Open  ]],
Cell[2850147, 73806, 460, 9, 27, "Input",ExpressionUUID->"f528cfe3-6a20-4b86-af2b-20f4763fc3b1",
 CellID->710107198]
}, Open  ]]
}, Closed]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2850668, 73822, 561, 10, 54, "Section",ExpressionUUID->"c70e460c-3a92-47d2-8cd3-6c82c14c37b0",
 CellTags->{"S5.18.1", "1.4"},
 CellID->1107613592],
Cell[CellGroupData[{
Cell[2851254, 73836, 232, 3, 39, "Subsection",ExpressionUUID->"cfab1a5a-5365-45b1-864f-ba0610d8efba",
 CellID->946395098],
Cell[CellGroupData[{
Cell[2851511, 73843, 227, 3, 41, "Subsubsection",ExpressionUUID->"472e654a-d1cf-4d81-a709-584acdb0da3e",
 CellID->1798522404],
Cell[2851741, 73848, 305, 5, 23, "Text",ExpressionUUID->"027b15b6-0fa6-42f9-94ca-278bdab4349b",
 CellID->1386590022],
Cell[CellGroupData[{
Cell[2852071, 73857, 227, 3, 40, "MathCaption",ExpressionUUID->"80694448-4b61-4cd3-b8f6-a785cd1fd441",
 CellID->214212395],
Cell[CellGroupData[{
Cell[2852323, 73864, 245, 4, 27, "Input",ExpressionUUID->"4f49134d-a716-4b08-9e2b-586e1bb6a149",
 CellID->1933712503],
Cell[CellGroupData[{
Cell[2852593, 73872, 732, 11, 22, "Print",ExpressionUUID->"238bf73a-58f0-47d2-8191-1e3f9a4d6be9",
 CellID->2064527593],
Cell[2853328, 73885, 1004, 19, 22, "Print",ExpressionUUID->"06b4e268-5bc1-4bfd-8093-6c49e848a057",
 CellID->1621970972],
Cell[2854335, 73906, 754, 11, 22, "Print",ExpressionUUID->"2c21ad01-669e-4913-9ef7-200c07bcb297",
 CellID->308711386],
Cell[2855092, 73919, 711, 10, 22, "Print",ExpressionUUID->"ebf6df46-287d-4f05-a00b-ff382be2dc77",
 CellID->191773350],
Cell[2855806, 73931, 693, 10, 22, "Print",ExpressionUUID->"973b7c77-470a-4775-8ac6-b27808953bc9",
 CellID->1909222950],
Cell[2856502, 73943, 730, 11, 22, "Print",ExpressionUUID->"cc4276cf-3902-4566-afb8-477a66374115",
 CellID->609467176],
Cell[2857235, 73956, 1007, 19, 22, "Print",ExpressionUUID->"becf4ddd-2a56-45d2-9ef9-06ad42303e0e",
 CellID->1965314489],
Cell[2858245, 73977, 752, 11, 22, "Print",ExpressionUUID->"ee7be762-1375-4e16-b6ef-8115af8119a5",
 CellID->594642116],
Cell[2859000, 73990, 732, 11, 22, "Print",ExpressionUUID->"59a9b218-0d29-445d-b8b2-bdb0b4e8f0f0",
 CellID->1908634121],
Cell[2859735, 74003, 1004, 19, 22, "Print",ExpressionUUID->"ae4cb934-042e-4f05-9fe2-240c31aa8971",
 CellID->1450436408],
Cell[2860742, 74024, 773, 11, 22, "Print",ExpressionUUID->"7942d65e-0060-46cb-8910-619832f5857a",
 CellID->1397832159],
Cell[2861518, 74037, 734, 11, 22, "Print",ExpressionUUID->"f5f7a00c-13a4-423d-9400-2f581a9251e3",
 CellID->2041481780],
Cell[2862255, 74050, 888, 13, 42, "Print",ExpressionUUID->"41a801ce-9042-4abd-b585-27c286a67312",
 CellID->838048341],
Cell[2863146, 74065, 732, 11, 22, "Print",ExpressionUUID->"aea99583-d203-4040-90e6-18176664c3f4",
 CellID->1880710161]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2863939, 74083, 128, 1, 40, "MathCaption",ExpressionUUID->"ba4cfccf-3bea-427b-948f-b95e4af8a646",
 CellID->250012957],
Cell[2864070, 74086, 359, 9, 66, "Input",ExpressionUUID->"45ba71ad-414f-42a4-96f6-e583ec3c5bdd",
 CellID->1489754767]
}, Open  ]],
Cell[CellGroupData[{
Cell[2864466, 74100, 117, 1, 40, "MathCaption",ExpressionUUID->"fb2335a8-373e-42b1-90b8-c948ac480124",
 CellID->796528596],
Cell[2864586, 74103, 259, 6, 27, "Input",ExpressionUUID->"4ad7960d-7007-4692-863f-930f5a0d7f32",
 CellID->1990541221]
}, Open  ]],
Cell[CellGroupData[{
Cell[2864882, 74114, 246, 4, 40, "MathCaption",ExpressionUUID->"e804e41b-5d36-4e6c-9caa-f12377cbd1fd",
 CellID->1526048381],
Cell[2865131, 74120, 321, 7, 27, "Input",ExpressionUUID->"327ea09f-44ec-4a84-b8be-f0f742efd06c",
 CellID->897943346],
Cell[2865455, 74129, 324, 7, 27, "Input",ExpressionUUID->"badef58d-46ec-4a80-9d47-da2d4681e4c4",
 CellID->1366228753]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[2865828, 74142, 230, 3, 41, "Subsubsection",ExpressionUUID->"9b46d7d3-8a5f-4551-986f-82ef61f7d1f9",
 CellID->1311497947],
Cell[2866061, 74147, 428, 11, 40, "Text",ExpressionUUID->"ba489325-b663-4a95-be0d-1a95805dc5ef",
 CellID->573842421],
Cell[CellGroupData[{
Cell[2866514, 74162, 331, 7, 40, "MathCaption",ExpressionUUID->"dfcfc4c4-9698-43b3-a3c3-8a540aaf0ec7",
 CellID->220511070],
Cell[CellGroupData[{
Cell[2866870, 74173, 304, 8, 27, "Input",ExpressionUUID->"bb010995-42b4-4f80-9e4f-bafffbc77444",
 CellID->1720262415],
Cell[2867177, 74183, 881, 16, 70, "Output",ExpressionUUID->"e4ba63b3-8498-4684-a5e6-b7f721134bf9",
 CellID->283875734]
}, Open  ]],
Cell[CellGroupData[{
Cell[2868095, 74204, 159, 3, 27, "Input",ExpressionUUID->"f2682149-1957-4a40-a01b-88c5b9f66701",
 CellID->203842424],
Cell[2868257, 74209, 725, 11, 70, "Output",ExpressionUUID->"19618354-fc1a-4c08-ad4b-ba3144d1e511",
 CellID->1538281351]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2869031, 74226, 322, 8, 40, "MathCaption",ExpressionUUID->"550badcb-df8b-483d-95cf-f86f2f3e17c1",
 CellID->1251341869],
Cell[CellGroupData[{
Cell[2869378, 74238, 153, 3, 27, "Input",ExpressionUUID->"369f4b69-df35-47ed-afca-7d62dc9483b4",
 CellID->995437212],
Cell[2869534, 74243, 2449, 68, 70, "Output",ExpressionUUID->"57650730-1273-43ad-a0b8-ca4c4a2ca047",
 CellID->339040024]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2872032, 74317, 203, 4, 40, "MathCaption",ExpressionUUID->"82388376-0b05-4bde-9135-6b747fb6cb4e",
 CellID->218785765],
Cell[CellGroupData[{
Cell[2872260, 74325, 164, 3, 27, "Input",ExpressionUUID->"364cddd3-42df-4ac6-aa5b-3ee5b54c2668",
 CellID->640348691],
Cell[2872427, 74330, 2204, 53, 70, "Output",ExpressionUUID->"86145cdf-2d12-403c-987f-705e1f7bdf4a",
 CellID->971567877]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2874680, 74389, 187, 2, 40, "MathCaption",ExpressionUUID->"1a1231b2-1997-466a-871c-b6bc12482ad7",
 CellID->976056709],
Cell[CellGroupData[{
Cell[2874892, 74395, 499, 14, 47, "Input",ExpressionUUID->"ee75c62f-41e4-40c7-a7d0-3058eb522f64",
 CellID->2105012377],
Cell[2875394, 74411, 1218, 26, 70, "Output",ExpressionUUID->"0d40a737-723d-4c4c-a389-faa57a8fddcc",
 CellID->1823140598]
}, Open  ]],
Cell[CellGroupData[{
Cell[2876649, 74442, 195, 5, 27, "Input",ExpressionUUID->"9bae844a-c14c-465f-88ca-1865faf2359c",
 CellID->181610089],
Cell[2876847, 74449, 5354, 146, 70, "Output",ExpressionUUID->"0691241f-6c09-402f-a8de-aa49f903c14f",
 CellID->1227355190]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2882250, 74601, 248, 4, 40, "MathCaption",ExpressionUUID->"ff66743e-e05c-43ac-921f-ce20d5926555",
 CellID->883783078],
Cell[CellGroupData[{
Cell[2882523, 74609, 404, 12, 27, "Input",ExpressionUUID->"06197b6d-cc5b-4a64-a46d-c7ed6b9eca4c",
 CellID->1745289294],
Cell[2882930, 74623, 1771, 45, 70, "Output",ExpressionUUID->"9fdf83f2-ec29-4867-9a7f-a6822de4b643",
 CellID->1578248895]
}, Open  ]],
Cell[CellGroupData[{
Cell[2884738, 74673, 635, 20, 27, "Input",ExpressionUUID->"3c8b5549-4c31-4254-a61d-cbc6895a198b",
 CellID->864809235],
Cell[2885376, 74695, 3899, 110, 70, "Output",ExpressionUUID->"781377a5-b169-4d13-a56a-88b51089b5bd",
 CellID->2058654168]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2889324, 74811, 372, 6, 56, "MathCaption",ExpressionUUID->"8d332b9e-5919-446c-8474-ae15637e990c",
 CellID->2013660344],
Cell[CellGroupData[{
Cell[2889721, 74821, 701, 21, 47, "Input",ExpressionUUID->"a876c047-9e06-4ecf-8305-a95f794dffe3",
 CellID->1394608535],
Cell[2890425, 74844, 5690, 146, 70, "Output",ExpressionUUID->"f41a813f-75f2-41c0-b18e-8e6994936e81",
 CellID->1691564759]
}, Open  ]],
Cell[CellGroupData[{
Cell[2896152, 74995, 703, 21, 47, "Input",ExpressionUUID->"90abb9a3-0c97-46ef-9ffa-c83640251cad",
 CellID->41836081],
Cell[2896858, 75018, 3714, 100, 70, "Output",ExpressionUUID->"0368f66e-0e1b-444c-be15-14ad4e1feb44",
 CellID->1740385527]
}, Open  ]]
}, Open  ]],
Cell[2900599, 75122, 259, 6, 23, "Text",ExpressionUUID->"53a71cf1-904b-4dc7-8010-0f08167ca4b5",
 CellID->1201126994],
Cell[CellGroupData[{
Cell[2900883, 75132, 196, 2, 40, "MathCaption",ExpressionUUID->"7852e2f9-ab1f-4064-83e2-5ff9c0c028ae",
 CellID->995106358],
Cell[CellGroupData[{
Cell[2901104, 75138, 195, 5, 27, "Input",ExpressionUUID->"a67eab13-4a6b-409a-a45a-12babe0aa5c3",
 CellID->149916070],
Cell[2901302, 75145, 717, 11, 70, "Output",ExpressionUUID->"67f4a88b-b26b-4e10-bb6d-3709e494b8bf",
 CellID->1932280149]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2902068, 75162, 185, 2, 40, "MathCaption",ExpressionUUID->"8dde0237-1f17-4f2f-86a5-39612bfa4cd7",
 CellID->547428114],
Cell[CellGroupData[{
Cell[2902278, 75168, 243, 7, 27, "Input",ExpressionUUID->"9f49d0df-c75f-48b9-8b60-273887d52513",
 CellID->804368250],
Cell[2902524, 75177, 2460, 69, 70, "Output",ExpressionUUID->"933eb438-a6f8-4ecb-b412-2529e3835d47",
 CellID->1584754309]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2905033, 75252, 210, 2, 40, "MathCaption",ExpressionUUID->"df974585-84f4-45e0-99d6-73c491c73ac9",
 CellID->982665710],
Cell[CellGroupData[{
Cell[2905268, 75258, 279, 8, 27, "Input",ExpressionUUID->"aee76edd-edf5-4fcf-9c9c-a5d705996605",
 CellID->1513342540],
Cell[2905550, 75268, 5358, 146, 70, "Output",ExpressionUUID->"ab6c91cd-e15e-4611-b7fb-d012567b613c",
 CellID->1873481269]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2910957, 75420, 193, 2, 40, "MathCaption",ExpressionUUID->"80575164-8a5b-4a91-8827-7f2e3f950d02",
 CellID->890263150],
Cell[CellGroupData[{
Cell[2911175, 75426, 260, 7, 27, "Input",ExpressionUUID->"4bc57f11-a773-40dd-86d9-811d78347615",
 CellID->1134193401],
Cell[2911438, 75435, 10334, 284, 70, "Output",ExpressionUUID->"20c3eb96-c8c9-4da7-859b-375de3ef604d",
 CellID->1194455769]
}, Open  ]],
Cell[CellGroupData[{
Cell[2921809, 75724, 159, 3, 27, "Input",ExpressionUUID->"377e1db9-7891-4652-8842-8ffa96e79bea",
 CellID->631607572],
Cell[2921971, 75729, 718, 11, 70, "Output",ExpressionUUID->"9bfd89fc-258f-4e95-bc6e-1830c51128d1",
 CellID->540977427]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2922738, 75746, 216, 2, 40, "MathCaption",ExpressionUUID->"78ebf05d-2dd5-4d24-b5ee-7147c7357c83",
 CellID->2137509914],
Cell[CellGroupData[{
Cell[2922979, 75752, 253, 7, 27, "Input",ExpressionUUID->"cdc353ac-6316-466e-b211-c5be91430bc5",
 CellID->2022677582],
Cell[2923235, 75761, 721, 11, 70, "Output",ExpressionUUID->"76d22231-c9b0-473a-855a-eb3047ac4937",
 CellID->1909337838]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2924005, 75778, 363, 8, 40, "MathCaption",ExpressionUUID->"1bde43d6-a4c5-4029-a6b4-6ab1fcdd0a78",
 CellID->564890773],
Cell[CellGroupData[{
Cell[2924393, 75790, 299, 9, 27, "Input",ExpressionUUID->"824d6f7a-d54c-4fd9-8067-ef883f205eaf",
 CellID->94018424],
Cell[2924695, 75801, 704, 11, 70, "Output",ExpressionUUID->"d7437574-ba2a-40fb-81b3-0f4ebc75ea44",
 CellID->392906145]
}, Open  ]]
}, Open  ]],
Cell[2925426, 75816, 331, 7, 23, "Text",ExpressionUUID->"747ae260-a1b6-4609-9b86-15aef1a6daf6",
 CellID->1154629751],
Cell[CellGroupData[{
Cell[2925782, 75827, 252, 4, 40, "MathCaption",ExpressionUUID->"c551201f-38e5-4db8-8a2a-6c017e559b11",
 CellID->1199990765],
Cell[CellGroupData[{
Cell[2926059, 75835, 154, 3, 27, "Input",ExpressionUUID->"26b0e0d3-c027-4600-9700-fa68f8132440",
 CellID->1225858378],
Cell[2926216, 75840, 2449, 68, 70, "Output",ExpressionUUID->"da597dc5-0162-4028-9e2c-616f923cc33c",
 CellID->61364490]
}, Open  ]],
Cell[CellGroupData[{
Cell[2928702, 75913, 213, 5, 27, "Input",ExpressionUUID->"0e1e1d1a-5b59-41c8-aa07-2e59074d1599",
 CellID->1329054869],
Cell[2928918, 75920, 4398, 127, 70, "Output",ExpressionUUID->"e18cd6e0-bbe7-45a7-b1d3-f11537557b80",
 CellID->560832870]
}, Open  ]],
Cell[CellGroupData[{
Cell[2933353, 76052, 163, 3, 27, "Input",ExpressionUUID->"6012be2c-d682-43c6-bc7c-15395d03a97b",
 CellID->1694720142],
Cell[2933519, 76057, 4402, 127, 70, "Output",ExpressionUUID->"47fc1ad9-9bf0-48bc-b6b5-b1a568e2e7f6",
 CellID->612319407]
}, Open  ]],
Cell[CellGroupData[{
Cell[2937958, 76189, 230, 4, 27, "Input",ExpressionUUID->"41f49a22-28d4-46bb-a046-07ffd40d91f8",
 CellID->757225],
Cell[2938191, 76195, 2454, 68, 70, "Output",ExpressionUUID->"35827a2f-00f0-434e-ba68-3ad3eb9ab176",
 CellID->510637025]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[2940706, 76270, 237, 3, 41, "Subsubsection",ExpressionUUID->"5fe8e3e7-f13c-43e2-bdf9-ad5e84894f10",
 CellID->1506731884],
Cell[2940946, 76275, 268, 5, 23, "Text",ExpressionUUID->"0a56d410-e643-4c32-8dac-2a8047ae7755",
 CellID->1773480832],
Cell[CellGroupData[{
Cell[2941239, 76284, 193, 2, 40, "MathCaption",ExpressionUUID->"33924e69-beea-43f0-b611-01a8304e95f8",
 CellID->515608964],
Cell[2941435, 76288, 511, 13, 27, "Input",ExpressionUUID->"44cc2207-fb49-42f2-8bc4-a7a4563ee3b2",
 CellID->743342303]
}, Open  ]],
Cell[CellGroupData[{
Cell[2941983, 76306, 219, 2, 40, "MathCaption",ExpressionUUID->"12226d3a-a740-4439-9528-f8a71024138d",
 CellID->1441711922],
Cell[2942205, 76310, 239, 5, 27, "Input",ExpressionUUID->"9016df49-04ad-4bd0-83e6-214869a2d228",
 CellID->794283953],
Cell[2942447, 76317, 1038, 33, 27, "Input",ExpressionUUID->"33a14312-4241-4671-9b30-785890809b72",
 CellID->772048319],
Cell[CellGroupData[{
Cell[2943510, 76354, 216, 6, 27, "Input",ExpressionUUID->"a1c20b71-2eb1-47cf-af1a-cee79ce5d739",
 CellID->719902259],
Cell[2943729, 76362, 9950, 264, 70, "Output",ExpressionUUID->"12567fc9-ea55-40a2-8bcc-5f4494002338",
 CellID->931825763]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2953728, 76632, 301, 5, 40, "MathCaption",ExpressionUUID->"f89f3030-a151-4409-af11-20385c37d108",
 CellID->1485801604],
Cell[CellGroupData[{
Cell[2954054, 76641, 274, 6, 27, "Input",ExpressionUUID->"f4208aa9-ac67-4566-b81e-067df1103fc0",
 CellID->1196101862],
Cell[2954331, 76649, 1050, 17, 70, "Message",ExpressionUUID->"5026d175-a6a7-4eff-abf8-6b64b4309f11",
 CellID->1744844817]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2955430, 76672, 285, 5, 40, "MathCaption",ExpressionUUID->"1b18e42a-51fe-4579-8ede-1a32ab800da7",
 CellID->1766165890],
Cell[CellGroupData[{
Cell[2955740, 76681, 446, 11, 27, "Input",ExpressionUUID->"c81140ce-3664-4751-bd18-ead8f98a983e",
 CellID->2064427616],
Cell[2956189, 76694, 936, 21, 70, "Output",ExpressionUUID->"25645139-4869-4123-b364-8c9d80a14b13",
 CellID->1717639534]
}, Open  ]],
Cell[CellGroupData[{
Cell[2957162, 76720, 222, 4, 27, "Input",ExpressionUUID->"ae362dcc-94e1-4324-be24-5c32f923de5c",
 CellID->1367265467],
Cell[2957387, 76726, 2750, 77, 70, "Output",ExpressionUUID->"78630d43-01fb-44ff-95b6-6cbdb0b55bf4",
 CellID->1892464964]
}, Open  ]],
Cell[CellGroupData[{
Cell[2960174, 76808, 266, 6, 27, "Input",ExpressionUUID->"7aae0d79-4f5d-4dbc-a5c2-2e203db308ed",
 CellID->206460103],
Cell[2960443, 76816, 1052, 18, 70, "Message",ExpressionUUID->"ba3b9d83-cfa0-4ed8-8735-72a75d52249e",
 CellID->278354231]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2961544, 76840, 278, 4, 40, "MathCaption",ExpressionUUID->"917465d0-14df-486c-b94e-dade6b7360e3",
 CellID->1338376509],
Cell[CellGroupData[{
Cell[2961847, 76848, 430, 11, 27, "Input",ExpressionUUID->"311a57e5-e35b-453e-b9e3-ef8341ea4996",
 CellID->1350268085],
Cell[2962280, 76861, 4045, 111, 70, "Output",ExpressionUUID->"558e2c03-3c8b-46e3-a4ee-e1ec20ee136a",
 CellID->1080177251]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2966374, 76978, 637, 10, 72, "MathCaption",ExpressionUUID->"3047ac7b-30a6-464c-a599-83c73e27ebc7",
 CellID->1778393041],
Cell[2967014, 76990, 305, 7, 27, "Input",ExpressionUUID->"8035bd06-daf9-49bf-b046-36cfda00156e",
 CellID->774441141]
}, Open  ]],
Cell[CellGroupData[{
Cell[2967356, 77002, 323, 6, 40, "MathCaption",ExpressionUUID->"20fba6c2-8ca2-4fb7-86b3-a1c42ab48ed1",
 CellID->133872320],
Cell[CellGroupData[{
Cell[2967704, 77012, 316, 8, 27, "Input",ExpressionUUID->"79d081b3-782e-4756-9af2-ca0a547a5ae2",
 CellID->1576703831],
Cell[2968023, 77022, 6663, 184, 70, "Output",ExpressionUUID->"f63760fc-bf78-43f0-9a89-214d5b6b8ab0",
 CellID->1859789466]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2974735, 77212, 120, 1, 40, "MathCaption",ExpressionUUID->"e3d392ed-88f0-4fe7-ad8d-f8e953eafb82",
 CellID->556313662],
Cell[CellGroupData[{
Cell[2974880, 77217, 285, 7, 27, "Input",ExpressionUUID->"ac8d6240-9f51-4643-9779-8389ff9fd036",
 CellID->1967401843],
Cell[2975168, 77226, 1251, 34, 70, "Output",ExpressionUUID->"2fe4ac5f-0c4d-4980-95a3-3995b940bcb4",
 CellID->857072977]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2976468, 77266, 155, 1, 40, "MathCaption",ExpressionUUID->"1e38c929-950f-4da9-bc9b-11bb38f71e5f",
 CellID->1754513065],
Cell[CellGroupData[{
Cell[2976648, 77271, 198, 5, 27, "Input",ExpressionUUID->"f674f4b9-10b0-423d-a6a0-2675dd1f329d",
 CellID->1003126498],
Cell[2976849, 77278, 5399, 148, 70, "Output",ExpressionUUID->"d54654a7-fc53-4fe3-a8db-230fb0001128",
 CellID->1652704010]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2982297, 77432, 191, 2, 40, "MathCaption",ExpressionUUID->"c5604c5c-bf1a-40ff-b5d2-facbc74fc77c",
 CellID->831538599],
Cell[CellGroupData[{
Cell[2982513, 77438, 179, 4, 27, "Input",ExpressionUUID->"b0042f75-63be-46c9-9578-e315b3c8e7cd",
 CellID->1663699122],
Cell[2982695, 77444, 10209, 270, 70, "Output",ExpressionUUID->"2f3378e6-ad73-46a8-bb3e-cc1d967f658c",
 CellID->1442805772]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2992953, 77720, 247, 4, 40, "MathCaption",ExpressionUUID->"93a60267-1455-4185-a87d-ed2027600ec4",
 CellID->1163447446],
Cell[CellGroupData[{
Cell[2993225, 77728, 225, 4, 27, "Input",ExpressionUUID->"c860eb0c-9a8b-4802-b522-03b09d55ad95",
 CellID->892729497],
Cell[2993453, 77734, 2976, 90, 70, "Output",ExpressionUUID->"a07ca585-c570-493b-9633-633e220431c8",
 CellID->1998751727]
}, Open  ]]
}, Open  ]],
Cell[2996456, 77828, 260, 3, 23, "Text",ExpressionUUID->"430ba7bc-1b38-47a7-8273-eefbe1cf70d3",
 CellID->1924939986],
Cell[CellGroupData[{
Cell[2996741, 77835, 363, 6, 40, "MathCaption",ExpressionUUID->"c320da86-3313-4ba5-9a5a-16f5c9bba7fc",
 CellID->1382136984],
Cell[CellGroupData[{
Cell[2997129, 77845, 189, 4, 27, "Input",ExpressionUUID->"84701a0f-e630-4b41-b650-e6f9c2d642fa",
 CellID->245680489],
Cell[2997321, 77851, 1300, 37, 70, "Output",ExpressionUUID->"cd975ac3-843b-4d1e-ae72-5d226663d733",
 CellID->2081069315]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2998670, 77894, 252, 5, 40, "MathCaption",ExpressionUUID->"775aa667-ed41-4cc8-b564-9907d1035f62",
 CellID->57914906],
Cell[CellGroupData[{
Cell[2998947, 77903, 172, 3, 27, "Input",ExpressionUUID->"36cbc39b-7a6b-4c65-9ce0-08d1e04bfa3b",
 CellID->1483626726],
Cell[2999122, 77908, 776, 13, 70, "Output",ExpressionUUID->"c61daeaa-30e5-41ab-9cf3-f2cedeefa15f",
 CellID->465818516]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2999947, 77927, 282, 5, 40, "MathCaption",ExpressionUUID->"9ef2b72b-f452-45e2-98a4-862fcdd745c3",
 CellID->2095173567],
Cell[CellGroupData[{
Cell[3000254, 77936, 169, 3, 27, "Input",ExpressionUUID->"ad867df0-2c12-4615-8ea4-b4dd3b19e4a6",
 CellID->242647365],
Cell[3000426, 77941, 737, 12, 70, "Output",ExpressionUUID->"7a98a1e7-44ef-472d-ab50-8cbf8ee11518",
 CellID->1433905097]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3001224, 77960, 248, 3, 41, "Subsubsection",ExpressionUUID->"d30fc830-5584-44b4-8bae-d349dd7fcddc",
 CellID->2079037655],
Cell[3001475, 77965, 556, 9, 57, "Text",ExpressionUUID->"40c7b29c-084f-49de-96c5-041fc110be88",
 CellID->59022842],
Cell[CellGroupData[{
Cell[3002056, 77978, 542, 9, 56, "MathCaption",ExpressionUUID->"87b55407-2b65-4b9f-896b-57988d7e36a0",
 CellID->1299346037],
Cell[CellGroupData[{
Cell[3002623, 77991, 190, 4, 27, "Input",ExpressionUUID->"7651de6c-1489-4006-97db-263fa6885dd3",
 CellID->347637516],
Cell[3002816, 77997, 13184, 385, 70, "Output",ExpressionUUID->"f3f9faae-5483-4a8b-8aa0-f9c20b2fc6ff",
 CellID->581478645]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3016049, 78388, 344, 8, 40, "MathCaption",ExpressionUUID->"1a0169ab-ee58-444a-8c7c-b14902e16745",
 CellID->1997642041],
Cell[CellGroupData[{
Cell[3016418, 78400, 296, 7, 27, "Input",ExpressionUUID->"e82d6b52-5d3b-46cb-ba3c-6099a652a8e7",
 CellID->610445889],
Cell[3016717, 78409, 8747, 232, 70, "Output",ExpressionUUID->"0e73ae32-a018-4b9e-b3fd-f639d9c259cc",
 CellID->1860992665]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3025513, 78647, 216, 2, 40, "MathCaption",ExpressionUUID->"bb1b6577-128a-4ab0-8035-4ec8c11b1d54",
 CellID->43269167],
Cell[CellGroupData[{
Cell[3025754, 78653, 296, 9, 27, "Input",ExpressionUUID->"8bef311c-51d9-47f1-8b61-9533e7d9439d",
 CellID->1375417083],
Cell[3026053, 78664, 751, 12, 70, "Output",ExpressionUUID->"320422c9-7cef-4e49-82ef-28150e3d70d0",
 CellID->1751885747]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3026853, 78682, 287, 5, 40, "MathCaption",ExpressionUUID->"ab6e4910-2c72-4976-aa09-ccbd973cfbed",
 CellID->766568751],
Cell[CellGroupData[{
Cell[3027165, 78691, 259, 5, 27, "Input",ExpressionUUID->"ca478766-7e5c-473e-86ca-3e1919bab2ea",
 CellID->1024019936],
Cell[3027427, 78698, 9971, 304, 70, "Output",ExpressionUUID->"30d33be7-e97c-4859-a4f3-2ddb27bbd1f9",
 CellID->1942949459]
}, Open  ]],
Cell[CellGroupData[{
Cell[3037435, 79007, 289, 7, 27, "Input",ExpressionUUID->"3cafbb67-84ef-437d-8e47-f432b1804284",
 CellID->888420665],
Cell[3037727, 79016, 41393, 1073, 70, "Output",ExpressionUUID->"6d14b6e8-fdb9-4970-9fb3-3a2d32592f4e",
 CellID->515221131]
}, Open  ]],
Cell[CellGroupData[{
Cell[3079157, 80094, 234, 4, 27, "Input",ExpressionUUID->"44596de9-a3f2-4b36-8b45-3e92a627dfc5",
 CellID->1369706427],
Cell[3079394, 80100, 42769, 1352, 70, "Output",ExpressionUUID->"f06a09a7-36a2-4a82-9c9b-5c00db0ddb0e",
 CellID->667484013]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3122212, 81458, 201, 2, 40, "MathCaption",ExpressionUUID->"38bfe98c-0c6d-4d93-a83b-0b04befe42dd",
 CellID->1872918627],
Cell[CellGroupData[{
Cell[3122438, 81464, 658, 20, 27, "Input",ExpressionUUID->"0485410d-1be7-4e51-be20-d1e140f2ff8a",
 CellID->328066263],
Cell[3123099, 81486, 36151, 940, 70, "Output",ExpressionUUID->"efc799bf-d859-4a71-aa02-f771e8c3c194",
 CellID->2096718210]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3159299, 82432, 286, 5, 40, "MathCaption",ExpressionUUID->"4bef990f-c921-4e42-a3a1-ce13f8fc0aa6",
 CellID->1287232317],
Cell[CellGroupData[{
Cell[3159610, 82441, 439, 12, 27, "Input",ExpressionUUID->"342d542e-f304-4dc0-aea0-aadf66e08c9a",
 CellID->763291941],
Cell[3160052, 82455, 75856, 2011, 70, "Output",ExpressionUUID->"da65ce41-3670-4869-800e-c7f9fc4c5daf",
 CellID->1483791847]
}, Open  ]],
Cell[CellGroupData[{
Cell[3235945, 84471, 157, 3, 27, "Input",ExpressionUUID->"9ead0511-b09c-4a0c-a11b-924cd3b37aec",
 CellID->1345708058],
Cell[3236105, 84476, 717, 11, 70, "Output",ExpressionUUID->"7b9a9f9a-1939-413e-9011-063b64c1a7f9",
 CellID->1509463565]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3236871, 84493, 216, 2, 40, "MathCaption",ExpressionUUID->"b60576e2-08b5-451c-88df-0ba95e3026ed",
 CellID->935778175],
Cell[CellGroupData[{
Cell[3237112, 84499, 260, 7, 27, "Input",ExpressionUUID->"1477a6ac-0bfe-4f8f-84ce-b43904bcc1d1",
 CellID->1601040244],
Cell[3237375, 84508, 708, 11, 70, "Output",ExpressionUUID->"da40d6e5-2ff1-4233-b352-2005f3c72d90",
 CellID->1329814944]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3238132, 84525, 207, 2, 40, "MathCaption",ExpressionUUID->"ce4c71f3-33a1-4d6f-8304-3725f95ce08f",
 CellID->910436832],
Cell[CellGroupData[{
Cell[3238364, 84531, 247, 7, 27, "Input",ExpressionUUID->"41c8a63c-1573-471f-a4ab-88fa0876ebe2",
 CellID->890850549],
Cell[3238614, 84540, 707, 11, 70, "Output",ExpressionUUID->"3b2aeba0-03d9-4be3-b6e2-6262db6313d7",
 CellID->1337449479]
}, Open  ]]
}, Open  ]],
Cell[3239348, 84555, 291, 5, 23, "Text",ExpressionUUID->"78994f27-8620-4a23-ba9b-4e9e15f8567c",
 CellID->1710545020]
}, Closed]],
Cell[CellGroupData[{
Cell[3239676, 84565, 230, 3, 41, "Subsubsection",ExpressionUUID->"f1b15873-32a4-47e2-9a08-77dbec0815fa",
 CellID->1678441731],
Cell[CellGroupData[{
Cell[3239931, 84572, 149, 1, 40, "MathCaption",ExpressionUUID->"e3acd022-1202-471a-8b64-4ccd87e45194",
 CellID->224985925],
Cell[3240083, 84575, 167, 3, 27, "Input",ExpressionUUID->"51532a59-b35b-48b8-8454-210c435704e7",
 CellID->1767222755],
Cell[3240253, 84580, 175, 3, 27, "Input",ExpressionUUID->"2ba7cd89-015c-4110-abc7-b4c5f939cc99",
 CellID->2014005275],
Cell[3240431, 84585, 163, 3, 27, "Input",ExpressionUUID->"763bc2bb-6304-4f3c-813a-2f2ea6f65d7e",
 CellID->749412911],
Cell[3240597, 84590, 164, 3, 27, "Input",ExpressionUUID->"51c7e33c-e4fb-410a-99be-84a7593906dc",
 CellID->391205522],
Cell[3240764, 84595, 166, 3, 27, "Input",ExpressionUUID->"fabccde2-570c-4bcb-93d8-8226888993de",
 CellID->1543764787],
Cell[3240933, 84600, 165, 3, 27, "Input",ExpressionUUID->"9822dbd5-1f8d-4e43-a5ba-4464ee2309ec",
 CellID->157352055]
}, Open  ]],
Cell[CellGroupData[{
Cell[3241135, 84608, 152, 1, 40, "MathCaption",ExpressionUUID->"d0bde9da-416f-4626-bad5-185fadca1959",
 CellID->1063618928],
Cell[3241290, 84611, 298, 6, 27, "Input",ExpressionUUID->"e006247d-5489-405e-bd51-8724703d20c1",
 CellID->589445268]
}, Open  ]],
Cell[CellGroupData[{
Cell[3241625, 84622, 114, 1, 40, "MathCaption",ExpressionUUID->"43a411ee-b7b6-432a-af4a-8fe6522ecd06",
 CellID->464555665],
Cell[3241742, 84625, 170, 3, 27, "Input",ExpressionUUID->"25d1baf4-0a34-4bf3-94ac-5ec29b94e60f",
 CellID->1684004810]
}, Open  ]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[3241973, 84635, 225, 3, 39, "Subsection",ExpressionUUID->"79b27aa2-b114-40c6-8e45-9fd7eb1892a0",
 CellID->1381857857],
Cell[CellGroupData[{
Cell[3242223, 84642, 193, 2, 41, "Subsubsection",ExpressionUUID->"a15826e2-c0da-4abe-add5-46cc54ce5bb9",
 CellID->1996843941],
Cell[3242419, 84646, 287, 5, 23, "Text",ExpressionUUID->"fd464a42-19cc-49da-a676-da71dbb88ccc",
 CellID->81681071],
Cell[CellGroupData[{
Cell[3242731, 84655, 202, 3, 40, "MathCaption",ExpressionUUID->"a65255cd-6832-4321-b1cc-672a7bf192ac",
 CellID->571709864],
Cell[3242936, 84660, 288, 7, 27, "Input",ExpressionUUID->"99cd3fc8-8770-45aa-91c5-4293470059bc",
 CellID->885847570],
Cell[3243227, 84669, 270, 6, 27, "Input",ExpressionUUID->"1725273d-ce82-4525-80dd-590d7c2f622e",
 CellID->2083467986]
}, Open  ]],
Cell[CellGroupData[{
Cell[3243534, 84680, 182, 2, 40, "MathCaption",ExpressionUUID->"e6c886bf-ea73-4ab1-84e8-a46b8c33d454",
 CellID->2125576367],
Cell[3243719, 84684, 209, 5, 27, "Input",ExpressionUUID->"336ef8dc-8322-4c5d-a458-6c86d55156a1",
 CellID->1171532324]
}, Open  ]],
Cell[CellGroupData[{
Cell[3243965, 84694, 268, 5, 40, "MathCaption",ExpressionUUID->"24964055-cb16-4769-8821-c9107df3f6b7",
 CellID->1813693772],
Cell[CellGroupData[{
Cell[3244258, 84703, 437, 13, 27, "Input",ExpressionUUID->"34bd2b24-3386-44e1-b723-50ffe64d2598",
 CellID->1368340486],
Cell[3244698, 84718, 1220, 30, 70, "Output",ExpressionUUID->"90771a56-7b70-4e0b-8036-aab69d097fde",
 CellID->802060404]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3245967, 84754, 270, 7, 40, "MathCaption",ExpressionUUID->"d4fc798e-61de-461c-b8f6-1538877694ef",
 CellID->2051244454],
Cell[CellGroupData[{
Cell[3246262, 84765, 161, 3, 27, "Input",ExpressionUUID->"d5cd3302-004c-4aff-b81a-578e3b3b2a32",
 CellID->1710553087],
Cell[3246426, 84770, 658, 10, 70, "Output",ExpressionUUID->"05b94de3-26ce-484e-8e30-2fb3bde08a1d",
 CellID->474357809]
}, Open  ]],
Cell[CellGroupData[{
Cell[3247121, 84785, 166, 3, 27, "Input",ExpressionUUID->"5954461d-4ebd-46ec-a236-a74c18548ac6",
 CellID->1214283920],
Cell[3247290, 84790, 835, 17, 70, "Output",ExpressionUUID->"ed7c22ff-687d-418b-9d83-491012655440",
 CellID->1072398199]
}, Open  ]],
Cell[CellGroupData[{
Cell[3248162, 84812, 172, 3, 27, "Input",ExpressionUUID->"f20ae69d-1ea2-4cb9-95c4-1be13ff0558b",
 CellID->31382093],
Cell[3248337, 84817, 782, 14, 70, "Output",ExpressionUUID->"6ccdecbb-5103-455a-ac08-34056ef53758",
 CellID->2113764969]
}, Open  ]],
Cell[CellGroupData[{
Cell[3249156, 84836, 172, 3, 27, "Input",ExpressionUUID->"525f982e-cbdf-4edb-89b4-a3495070f9be",
 CellID->215871800],
Cell[3249331, 84841, 679, 11, 70, "Output",ExpressionUUID->"475b654f-0a6e-484b-87e1-a70c2c6caf62",
 CellID->225861662]
}, Open  ]],
Cell[CellGroupData[{
Cell[3250047, 84857, 166, 3, 27, "Input",ExpressionUUID->"29a7ef17-0097-482e-aa0f-e9fc58e5f1fb",
 CellID->202902747],
Cell[3250216, 84862, 671, 10, 70, "Output",ExpressionUUID->"5cc0fde0-2f9b-4ca6-8d48-0e3bdd25de38",
 CellID->1453348985]
}, Open  ]],
Cell[CellGroupData[{
Cell[3250924, 84877, 159, 3, 27, "Input",ExpressionUUID->"a4560393-8272-4bdb-b2c5-ae7c42cdf4ba",
 CellID->1178007437],
Cell[3251086, 84882, 1245, 31, 70, "Output",ExpressionUUID->"247b6aab-e149-4e4e-9bd2-88b562838797",
 CellID->957389917]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3252380, 84919, 243, 5, 40, "MathCaption",ExpressionUUID->"23ad7362-8029-47ee-8280-80f4fc4804c1",
 CellID->1458528489],
Cell[CellGroupData[{
Cell[3252648, 84928, 635, 20, 27, "Input",ExpressionUUID->"1fdc1000-cdac-461b-b6b1-85df819977d7",
 CellID->2109967765],
Cell[3253286, 84950, 6924, 217, 70, "Output",ExpressionUUID->"72e71561-1ce8-45e3-9e7b-3d61d0d6fe6f",
 CellID->1904851695]
}, Open  ]],
Cell[CellGroupData[{
Cell[3260247, 85172, 161, 3, 27, "Input",ExpressionUUID->"53738b58-a7eb-49d6-9748-0be16ca5b84d",
 CellID->1850014161],
Cell[3260411, 85177, 668, 10, 70, "Output",ExpressionUUID->"1228c308-f086-4444-b64a-bb8a5a9c2630",
 CellID->156228389]
}, Open  ]],
Cell[CellGroupData[{
Cell[3261116, 85192, 166, 3, 27, "Input",ExpressionUUID->"8e2f4e65-5e8f-45db-b7d4-81c10005d1ef",
 CellID->1656730899],
Cell[3261285, 85197, 1274, 33, 70, "Output",ExpressionUUID->"05c95fc2-1807-4bba-88b7-6ec343a2f1b4",
 CellID->1505517289]
}, Open  ]],
Cell[CellGroupData[{
Cell[3262596, 85235, 171, 3, 27, "Input",ExpressionUUID->"f4a66114-7b07-408e-a1bc-ebb071de9fd5",
 CellID->6226305],
Cell[3262770, 85240, 1274, 32, 70, "Output",ExpressionUUID->"7bf1f88c-86d3-4271-af4e-8fdcb0801fd8",
 CellID->1576747269]
}, Open  ]],
Cell[CellGroupData[{
Cell[3264081, 85277, 172, 3, 27, "Input",ExpressionUUID->"63f5929f-6d18-424b-9980-9898fbf3b005",
 CellID->833233827],
Cell[3264256, 85282, 688, 11, 70, "Output",ExpressionUUID->"02a27b44-ee36-4f13-8d44-ecd8d1f092f8",
 CellID->443146103]
}, Open  ]],
Cell[CellGroupData[{
Cell[3264981, 85298, 167, 3, 27, "Input",ExpressionUUID->"0f8342cb-7fd0-4999-841e-b56d6955d661",
 CellID->1309204025],
Cell[3265151, 85303, 684, 11, 70, "Output",ExpressionUUID->"efc270ca-d5a1-4c2c-9715-fc484ba71395",
 CellID->483265325]
}, Open  ]],
Cell[CellGroupData[{
Cell[3265872, 85319, 184, 4, 27, "Input",ExpressionUUID->"5ad868df-6da0-4d90-a39a-67c96ac0411a",
 CellID->1453052978],
Cell[3266059, 85325, 664, 10, 70, "Output",ExpressionUUID->"b3fb11be-cfc9-40a6-83cf-7ca7cdc2eba2",
 CellID->457164321]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3266772, 85341, 257, 5, 40, "MathCaption",ExpressionUUID->"68a457c1-f64c-4f7c-bebb-35ace95cb117",
 CellID->1677100544],
Cell[CellGroupData[{
Cell[3267054, 85350, 387, 12, 27, "Input",ExpressionUUID->"1819b3b1-10ec-4607-a4a3-46142ab968c8",
 CellID->1305976133],
Cell[3267444, 85364, 664, 10, 70, "Output",ExpressionUUID->"ee3dc255-89b8-49e1-8623-9307a0b8fc9a",
 CellID->2011737728]
}, Open  ]],
Cell[CellGroupData[{
Cell[3268145, 85379, 180, 4, 27, "Input",ExpressionUUID->"c8854cfa-ab04-4832-8aff-a823ff9a92b6",
 CellID->1938287600],
Cell[3268328, 85385, 656, 10, 70, "Output",ExpressionUUID->"749b2a8f-9435-4655-807e-e5804cad801c",
 CellID->1293456128]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3269033, 85401, 222, 4, 40, "MathCaption",ExpressionUUID->"d901b97e-1afc-47ff-a483-e3495402f338",
 CellID->1480316639],
Cell[CellGroupData[{
Cell[3269280, 85409, 252, 7, 27, "Input",ExpressionUUID->"4c44df56-a018-49ac-9ad7-15c3e88cec40",
 CellID->1083701178],
Cell[3269535, 85418, 779, 16, 70, "Output",ExpressionUUID->"fb07ce59-2b94-4377-82e1-d34b85d396b7",
 CellID->140184715]
}, Open  ]],
Cell[CellGroupData[{
Cell[3270351, 85439, 158, 3, 27, "Input",ExpressionUUID->"806e2b9e-1abf-445e-849d-29546d7ce537",
 CellID->2066445537],
Cell[3270512, 85444, 660, 10, 70, "Output",ExpressionUUID->"3370931e-a100-4abb-8d64-c398b7fe21cd",
 CellID->294944246]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3271221, 85460, 200, 3, 40, "MathCaption",ExpressionUUID->"cbe58b02-668a-41c4-ace6-f82af36d7ebf",
 CellID->548246988],
Cell[CellGroupData[{
Cell[3271446, 85467, 262, 7, 27, "Input",ExpressionUUID->"7a09d8ed-5b45-4978-a6e9-13247dca452b",
 CellID->1682281686],
Cell[3271711, 85476, 1043, 26, 70, "Output",ExpressionUUID->"11e8b11d-6bf6-4348-9f2d-9bc3e5a15e8c",
 CellID->735685246]
}, Open  ]],
Cell[CellGroupData[{
Cell[3272791, 85507, 156, 3, 27, "Input",ExpressionUUID->"710b7220-32ec-4a34-b47f-2d75e920cb16",
 CellID->21896850],
Cell[3272950, 85512, 679, 10, 70, "Output",ExpressionUUID->"c2f40a5c-9dc4-4917-8fd5-08de489a3ff6",
 CellID->1598573743]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3273678, 85528, 189, 2, 40, "MathCaption",ExpressionUUID->"688a89de-e370-4d47-b921-5757e24afe5b",
 CellID->464506489],
Cell[CellGroupData[{
Cell[3273892, 85534, 397, 12, 27, "Input",ExpressionUUID->"6c1f8af5-5a0d-4ebe-bfa6-751aad9ece2d",
 CellID->857503778],
Cell[3274292, 85548, 933, 16, 70, "Message",ExpressionUUID->"2fa175a2-15c2-4e8c-8144-3b6d27606722",
 CellID->60990984]
}, Open  ]]
}, Open  ]],
Cell[3275252, 85568, 256, 4, 40, "MathCaption",ExpressionUUID->"ed6e0153-ad5a-4741-9a99-b6c0ef41ff77",
 CellID->606725672]
}, Closed]],
Cell[CellGroupData[{
Cell[3275545, 85577, 200, 2, 41, "Subsubsection",ExpressionUUID->"6373d80e-5081-4724-8f96-76962c4fbb01",
 CellID->986013348],
Cell[3275748, 85581, 408, 8, 139, "Text",ExpressionUUID->"28c38144-bcf1-4c8e-89aa-87ca284d4ce3",
 CellID->1942519584],
Cell[CellGroupData[{
Cell[3276181, 85593, 413, 9, 56, "MathCaption",ExpressionUUID->"1d7fef68-d4f5-414c-bc49-6cfb3423b325",
 CellID->1880246594],
Cell[CellGroupData[{
Cell[3276619, 85606, 154, 3, 27, "Input",ExpressionUUID->"e9e05e27-2a28-4280-936a-c7c92841ef3d",
 CellID->1082639753],
Cell[3276776, 85611, 3455, 94, 70, "Output",ExpressionUUID->"de17c893-3d51-4e9a-a990-abd059d93ad1",
 CellID->725768469]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3280280, 85711, 175, 3, 40, "MathCaption",ExpressionUUID->"cd29e58d-a01c-4384-be50-af1063129f9c",
 CellID->999618911],
Cell[CellGroupData[{
Cell[3280480, 85718, 126, 2, 27, "Input",ExpressionUUID->"ff5ff4fa-f150-4bc5-825b-dee35d74aabe",
 CellID->830466696],
Cell[3280609, 85722, 1209, 30, 70, "Output",ExpressionUUID->"9b802ed8-e83c-4871-b4a6-750d790d668e",
 CellID->283543523]
}, Open  ]],
Cell[CellGroupData[{
Cell[3281855, 85757, 162, 3, 27, "Input",ExpressionUUID->"9c48f53b-0344-4426-959c-610ff12b3050",
 CellID->683765760],
Cell[3282020, 85762, 2120, 52, 70, "Output",ExpressionUUID->"6ae2fb0e-8dc8-4e5b-a7df-d4c98c00ac57",
 CellID->249778368]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3284189, 85820, 234, 5, 40, "MathCaption",ExpressionUUID->"211b2dfd-9318-4607-8d72-b47dc1d252b9",
 CellID->1456919595],
Cell[CellGroupData[{
Cell[3284448, 85829, 243, 7, 27, "Input",ExpressionUUID->"cc90eeec-4f4b-4464-82d3-e1fe83abd639",
 CellID->839090071],
Cell[3284694, 85838, 35548, 920, 70, "Output",ExpressionUUID->"faa03992-7fa3-4da2-b1f6-4a3c26e0020d",
 CellID->1841984738]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3320291, 86764, 192, 4, 40, "MathCaption",ExpressionUUID->"43b69aa9-1f4d-4f20-8ff6-0a454a45e8a7",
 CellID->2130802682],
Cell[CellGroupData[{
Cell[3320508, 86772, 126, 2, 27, "Input",ExpressionUUID->"d5444b21-8a94-4f68-8d5a-7d0530a91678",
 CellID->963379002],
Cell[3320637, 86776, 6897, 217, 70, "Output",ExpressionUUID->"74f036dd-52b2-454f-80d4-9ee6c4087836",
 CellID->113236414]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3327583, 86999, 308, 5, 56, "MathCaption",ExpressionUUID->"bd5a66c1-76de-4fa5-9174-65adaae3c628",
 CellID->143131777],
Cell[CellGroupData[{
Cell[3327916, 87008, 403, 12, 27, "Input",ExpressionUUID->"6fa0ddfe-94ca-40cb-b899-89a0122a8d3f",
 CellID->434091399],
Cell[3328322, 87022, 1682, 44, 70, "Output",ExpressionUUID->"dacdc655-71fc-43f8-bcf1-47d10435e886",
 CellID->493898856]
}, Open  ]]
}, Open  ]],
Cell[3330031, 87070, 342, 6, 56, "MathCaption",ExpressionUUID->"cae99855-d5fc-49ab-98a8-1f461466032c",
 CellID->1023131317],
Cell[CellGroupData[{
Cell[3330398, 87080, 318, 6, 56, "MathCaption",ExpressionUUID->"53d2d327-f727-4e87-9da0-8448ccffeeec",
 CellID->866483290],
Cell[CellGroupData[{
Cell[3330741, 87090, 144, 2, 27, "Input",ExpressionUUID->"74cc9980-6f44-4f48-8810-9b79c98406b4",
 CellID->898044723],
Cell[3330888, 87094, 637, 10, 70, "Output",ExpressionUUID->"3596209e-22bd-4668-8e08-713c8f583ad1",
 CellID->933777137]
}, Open  ]],
Cell[CellGroupData[{
Cell[3331562, 87109, 148, 2, 27, "Input",ExpressionUUID->"7dcc2637-f47f-4100-9fe4-044497045efe",
 CellID->1110383075],
Cell[3331713, 87113, 668, 10, 70, "Output",ExpressionUUID->"1829d1d2-e0ba-4a31-9b66-fb8eb15a3bc3",
 CellID->181849283]
}, Open  ]],
Cell[CellGroupData[{
Cell[3332418, 87128, 650, 22, 27, "Input",ExpressionUUID->"4c50ad62-f3ef-49f1-877e-54bd1d7a4c62",
 CellID->1454077224],
Cell[3333071, 87152, 5216, 184, 70, "Output",ExpressionUUID->"81f77b31-5f0f-4cc5-aacc-32d87a8831dc",
 CellID->471669700]
}, Open  ]],
Cell[CellGroupData[{
Cell[3338324, 87341, 159, 3, 27, "Input",ExpressionUUID->"f91e135a-ec55-4dc4-a959-6641be83e62a",
 CellID->1791678429],
Cell[3338486, 87346, 718, 12, 70, "Output",ExpressionUUID->"ad2fc02c-68b6-4bb4-ba9e-341905be4f8e",
 CellID->990506823]
}, Open  ]],
Cell[CellGroupData[{
Cell[3339241, 87363, 824, 27, 27, "Input",ExpressionUUID->"a4174643-6a44-4a5f-ae3f-ae317f651959",
 CellID->1319592022],
Cell[3340068, 87392, 12083, 291, 70, "Output",ExpressionUUID->"feaa1e17-6ac8-4309-9176-324714e567a7",
 CellID->154942617]
}, Open  ]]
}, Open  ]],
Cell[3352178, 87687, 274, 5, 40, "MathCaption",ExpressionUUID->"5df70d6f-e464-4e18-8bf0-f47266bca227",
 CellID->1772207373],
Cell[CellGroupData[{
Cell[3352477, 87696, 220, 4, 40, "MathCaption",ExpressionUUID->"18406536-ee1d-4f9b-81e0-794e8915b3c4",
 CellID->2038617768],
Cell[CellGroupData[{
Cell[3352722, 87704, 155, 3, 27, "Input",ExpressionUUID->"aa060a18-5781-4c90-9e02-81774e74fcf1",
 CellID->2094722538],
Cell[3352880, 87709, 3448, 94, 70, "Output",ExpressionUUID->"40503d61-9c2b-4f27-b558-9ff7267ad0dd",
 CellID->1370383755]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3356377, 87809, 170, 3, 40, "MathCaption",ExpressionUUID->"2dbb4005-ef00-48e1-99d8-ecdc425ca691",
 CellID->906580619],
Cell[3356550, 87814, 274, 7, 27, "Input",ExpressionUUID->"f7c26616-a89a-4c91-a0df-3742f347415c",
 CellID->503988068],
Cell[CellGroupData[{
Cell[3356849, 87825, 154, 3, 27, "Input",ExpressionUUID->"d8117faa-3a9c-452b-a5ba-d190524b60a6",
 CellID->851955108],
Cell[3357006, 87830, 1497, 38, 70, "Output",ExpressionUUID->"59565aa5-6375-42a8-b197-b33674fb2aaf",
 CellID->277656210]
}, Open  ]],
Cell[CellGroupData[{
Cell[3358540, 87873, 159, 3, 27, "Input",ExpressionUUID->"18fff9d7-a4cf-4266-b42f-ff9893892c12",
 CellID->1081456262],
Cell[3358702, 87878, 688, 10, 70, "Output",ExpressionUUID->"9187f295-1b2b-4ba4-8ddf-04ee66f8228c",
 CellID->85956539]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3359439, 87894, 242, 5, 40, "MathCaption",ExpressionUUID->"360fc500-ceef-42af-9ab6-25d7b03f6218",
 CellID->1453967730],
Cell[3359684, 87901, 247, 6, 27, "Input",ExpressionUUID->"b05e1aac-d23f-47f6-b0fa-51d574d2202a",
 CellID->619218897],
Cell[CellGroupData[{
Cell[3359956, 87911, 154, 3, 27, "Input",ExpressionUUID->"55b5b03d-5f8a-418a-8db6-0c7a46f25a07",
 CellID->506248517],
Cell[3360113, 87916, 3449, 94, 70, "Output",ExpressionUUID->"2e6ee187-d10e-4953-bb4f-47175c2b1339",
 CellID->1562540189]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3363611, 88016, 257, 5, 40, "MathCaption",ExpressionUUID->"e0ac1f65-ee18-40ea-8a18-ed9d87cc9174",
 CellID->379971663],
Cell[CellGroupData[{
Cell[3363893, 88025, 593, 18, 27, "Input",ExpressionUUID->"deb9c786-47f3-4b70-b950-b7ef1b29b8eb",
 CellID->1714857510],
Cell[3364489, 88045, 5074, 132, 70, "Output",ExpressionUUID->"f8e1020d-d79c-47eb-a884-bd6cfca5a523",
 CellID->35437360]
}, Open  ]]
}, Open  ]],
Cell[3369590, 88181, 264, 7, 40, "MathCaption",ExpressionUUID->"a9fb137e-8383-4978-8306-646bcc9bd9ef",
 CellID->1816980729],
Cell[CellGroupData[{
Cell[3369879, 88192, 188, 2, 40, "MathCaption",ExpressionUUID->"5e64e359-6f90-4a32-b3b3-24f3a8881a54",
 CellID->777651963],
Cell[CellGroupData[{
Cell[3370092, 88198, 206, 5, 27, "Input",ExpressionUUID->"a1915fea-53a5-49cd-8bfd-656c57f59e06",
 CellID->411066499],
Cell[3370301, 88205, 907, 16, 70, "Message",ExpressionUUID->"b4fd9280-16f2-40ef-97b7-38d1799a16a4",
 CellID->742739301]
}, Open  ]],
Cell[CellGroupData[{
Cell[3371245, 88226, 201, 5, 27, "Input",ExpressionUUID->"6555d131-84e5-4779-993d-05c7990ce012",
 CellID->125015885],
Cell[3371449, 88233, 1059, 18, 70, "Message",ExpressionUUID->"65a078a6-b2a1-461a-af37-cd0ce34074b2",
 CellID->2044475543]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3372569, 88258, 189, 2, 41, "Subsubsection",ExpressionUUID->"729b12b4-8f4a-4da1-8584-80022d84c090",
 CellID->2138164501],
Cell[3372761, 88262, 233, 6, 23, "Text",ExpressionUUID->"c974a48a-3dd4-4651-b425-45b2e303eaf5",
 CellID->139419409],
Cell[CellGroupData[{
Cell[3373019, 88272, 238, 4, 40, "MathCaption",ExpressionUUID->"08fe3204-5c21-4569-a3d9-c77cc61a4a6b",
 CellID->317348186],
Cell[CellGroupData[{
Cell[3373282, 88280, 155, 3, 27, "Input",ExpressionUUID->"1616ce59-cfe2-4b45-9c02-49bc57993e00",
 CellID->1469066601],
Cell[3373440, 88285, 3481, 94, 70, "Output",ExpressionUUID->"e305c16b-6a84-4066-85dc-c6388a2fcaae",
 CellID->1762099201]
}, Open  ]],
Cell[CellGroupData[{
Cell[3376958, 88384, 242, 7, 27, "Input",ExpressionUUID->"5af2da47-a180-42a5-9c83-64bab7a6fa2d",
 CellID->709827812],
Cell[3377203, 88393, 4389, 120, 70, "Output",ExpressionUUID->"44a12c9d-66a6-4d8a-a249-e1b5ec0dab70",
 CellID->1801779077]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3381641, 88519, 200, 3, 40, "MathCaption",ExpressionUUID->"49312b70-6fb5-41c7-b898-bc4b7d6fa992",
 CellID->758025728],
Cell[3381844, 88524, 328, 9, 27, "Input",ExpressionUUID->"a245922e-f79b-43cd-a9c1-dc7afe288f20",
 CellID->1265416492],
Cell[CellGroupData[{
Cell[3382197, 88537, 225, 6, 27, "Input",ExpressionUUID->"cad607b5-9103-4b17-9379-14f8a69cd892",
 CellID->1324424803],
Cell[3382425, 88545, 4088, 109, 70, "Output",ExpressionUUID->"44157af1-03b3-4d7e-9abd-741759412272",
 CellID->834040792]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3386562, 88660, 268, 5, 40, "MathCaption",ExpressionUUID->"856f49f9-7ed7-4f84-976f-af98309fb362",
 CellID->319048824],
Cell[3386833, 88667, 448, 14, 27, "Input",ExpressionUUID->"cc4c4da4-4e4e-4607-8939-e5b3125edf51",
 CellID->2134402056],
Cell[CellGroupData[{
Cell[3387306, 88685, 254, 7, 27, "Input",ExpressionUUID->"9c98aa42-7e0c-42bc-b48b-0fc433a8ebe2",
 CellID->1328515704],
Cell[3387563, 88694, 2643, 70, 70, "Output",ExpressionUUID->"cdfbed00-ef03-4c52-bd56-b85f021c8358",
 CellID->1097827175]
}, Open  ]],
Cell[CellGroupData[{
Cell[3390243, 88769, 253, 7, 27, "Input",ExpressionUUID->"db0cbad1-d546-4609-8f2c-d70513d0639d",
 CellID->467338297],
Cell[3390499, 88778, 2704, 72, 70, "Output",ExpressionUUID->"1b0fb8be-8f4e-4ffe-82b1-bc96f0f235c9",
 CellID->1835201]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3393252, 88856, 341, 6, 56, "MathCaption",ExpressionUUID->"5e814303-ccc2-4f04-8c2d-d94b91d27f6f",
 CellID->1936974638],
Cell[CellGroupData[{
Cell[3393618, 88866, 178, 4, 27, "Input",ExpressionUUID->"514d53ae-284a-4a9b-b12c-bbc96993dd56",
 CellID->1028135993],
Cell[3393799, 88872, 2640, 70, 70, "Output",ExpressionUUID->"2095b021-29ca-4ebc-845a-d03bbdb1f9d8",
 CellID->779452149]
}, Open  ]],
Cell[CellGroupData[{
Cell[3396476, 88947, 161, 3, 27, "Input",ExpressionUUID->"b4baffb7-bfc5-414a-ba66-31b3d52c0a30",
 CellID->922554359],
Cell[3396640, 88952, 2648, 70, 70, "Output",ExpressionUUID->"ff2ff706-e373-42a1-a2bc-1ee6749f40f5",
 CellID->602021543]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3399337, 89028, 278, 5, 56, "MathCaption",ExpressionUUID->"0d0c51fc-a729-4c3f-8fd1-05d0aca8ed82",
 CellID->1771022318],
Cell[CellGroupData[{
Cell[3399640, 89037, 162, 3, 27, "Input",ExpressionUUID->"39be0805-dff8-4f04-b229-c9df11bc6a00",
 CellID->1448751206],
Cell[3399805, 89042, 2641, 70, 70, "Output",ExpressionUUID->"32436825-b64a-4174-a3cb-2bc779d78908",
 CellID->1380612184]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3402495, 89118, 234, 5, 40, "MathCaption",ExpressionUUID->"13e419a8-fb49-4460-b205-35f95dcc24da",
 CellID->444496031],
Cell[CellGroupData[{
Cell[3402754, 89127, 201, 5, 27, "Input",ExpressionUUID->"47eb79c3-ed35-403c-a239-0fa53d758fd5",
 CellID->1696668751],
Cell[3402958, 89134, 9960, 263, 70, "Output",ExpressionUUID->"4aa24cd7-e65e-40ac-b62c-bbd46aa1b6a6",
 CellID->388170653]
}, Open  ]],
Cell[CellGroupData[{
Cell[3412955, 89402, 225, 6, 27, "Input",ExpressionUUID->"7e9dafe1-ea5f-48de-a84e-432cd97842f9",
 CellID->1542255061],
Cell[3413183, 89410, 11719, 300, 70, "Output",ExpressionUUID->"54b5863f-ff8c-491d-9047-73b560deaa7e",
 CellID->2076329032]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3424951, 89716, 213, 4, 40, "MathCaption",ExpressionUUID->"d2060bd4-f144-4a07-9bec-63da079bf586",
 CellID->1023370296],
Cell[CellGroupData[{
Cell[3425189, 89724, 294, 9, 27, "Input",ExpressionUUID->"d1846d1d-9375-4cb3-b949-4d3d77dda139",
 CellID->588036049],
Cell[3425486, 89735, 17025, 433, 70, "Output",ExpressionUUID->"79ff31f9-cd4c-422c-993c-44e55eab4f5d",
 CellID->252777074]
}, Open  ]],
Cell[CellGroupData[{
Cell[3442548, 90173, 177, 4, 27, "Input",ExpressionUUID->"b1a413fc-45f3-4be8-b8ed-7fdedfdfab87",
 CellID->502310434],
Cell[3442728, 90179, 5593, 136, 70, "Output",ExpressionUUID->"fa8204dc-8b15-4029-b4ac-cb86e6b5acf7",
 CellID->701628952]
}, Open  ]],
Cell[CellGroupData[{
Cell[3448358, 90320, 157, 3, 27, "Input",ExpressionUUID->"3a07f6b7-ab66-4dd5-9db9-f11756839b2a",
 CellID->965495525],
Cell[3448518, 90325, 663, 10, 70, "Output",ExpressionUUID->"2a5c7680-24e4-48d7-8a02-c8612204af07",
 CellID->445090587]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3449242, 90342, 178, 2, 41, "Subsubsection",ExpressionUUID->"ddfdf62f-df37-415e-bf5e-3d6da1d78d9b",
 CellID->1762278910],
Cell[3449423, 90346, 492, 8, 57, "Text",ExpressionUUID->"c88d1560-a46c-4bf6-ad94-5a6e6caf4c56",
 CellID->527180946],
Cell[CellGroupData[{
Cell[3449940, 90358, 209, 2, 40, "MathCaption",ExpressionUUID->"af00e35e-8906-4692-891b-66e8e3de7516",
 CellID->1960115547],
Cell[3450152, 90362, 353, 8, 27, "Input",ExpressionUUID->"f1e42cfa-e54c-4782-83cf-597fbf8fb106",
 CellID->1262853449]
}, Open  ]],
Cell[CellGroupData[{
Cell[3450542, 90375, 182, 2, 40, "MathCaption",ExpressionUUID->"8c8f6f41-1bab-4d0a-b7cb-21c4d6980418",
 CellID->1197034042],
Cell[3450727, 90379, 369, 10, 27, "Input",ExpressionUUID->"219344eb-131f-428f-9974-45327618e521",
 CellID->924338798]
}, Open  ]],
Cell[CellGroupData[{
Cell[3451133, 90394, 213, 4, 40, "MathCaption",ExpressionUUID->"913598fa-e4bd-4af8-9a30-676d1707c5e9",
 CellID->465475588],
Cell[CellGroupData[{
Cell[3451371, 90402, 579, 19, 27, "Input",ExpressionUUID->"4c16c335-9138-4d67-8464-a0e6268bbdcf",
 CellID->2112765142],
Cell[3451953, 90423, 1378, 36, 70, "Output",ExpressionUUID->"1b1e8275-3f20-4c2d-b706-abe1ec0590e8",
 CellID->951953208]
}, Open  ]],
Cell[CellGroupData[{
Cell[3453368, 90464, 283, 7, 27, "Input",ExpressionUUID->"ec2fc400-1233-4208-851f-30cbd3887689",
 CellID->806148330],
Cell[3453654, 90473, 3594, 96, 70, "Output",ExpressionUUID->"aca912ae-d165-430c-9b66-d6d84c8e9cd1",
 CellID->724396706]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3457297, 90575, 311, 5, 40, "MathCaption",ExpressionUUID->"4adec57b-2706-46a9-abbd-f5dfb895a753",
 CellID->2057601857],
Cell[CellGroupData[{
Cell[3457633, 90584, 276, 6, 27, "Input",ExpressionUUID->"1e891def-d31f-4fee-be3a-0b553397c8d2",
 CellID->256560628],
Cell[3457912, 90592, 1006, 17, 70, "Message",ExpressionUUID->"03e6c921-729d-4ed0-8d61-64ea36dc9b4f",
 CellID->850642899]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3458967, 90615, 248, 3, 40, "MathCaption",ExpressionUUID->"c83f6708-25ad-450f-8949-481ebbc1a280",
 CellID->1384402267],
Cell[CellGroupData[{
Cell[3459240, 90622, 185, 4, 27, "Input",ExpressionUUID->"91d1ed1e-c38d-43b2-b87e-ade1dbb844d9",
 CellID->1811069871],
Cell[3459428, 90628, 1348, 34, 70, "Output",ExpressionUUID->"24d4afd4-7df0-48b0-b077-567a7c30f11d",
 CellID->956565414]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3460825, 90668, 174, 2, 40, "MathCaption",ExpressionUUID->"e72fac13-eb6d-461f-97a7-dd76769b5b51",
 CellID->8739004],
Cell[CellGroupData[{
Cell[3461024, 90674, 293, 9, 27, "Input",ExpressionUUID->"7f224a5a-7fb4-4cef-8cb4-47f872e52418",
 CellID->847552232],
Cell[3461320, 90685, 2648, 70, 70, "Output",ExpressionUUID->"3da3b3e9-c0fa-4716-bcd1-87765f56292b",
 CellID->932331656]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3464017, 90761, 297, 5, 40, "MathCaption",ExpressionUUID->"d3b79422-88fc-463c-ab21-8b578d578056",
 CellID->2003781953],
Cell[CellGroupData[{
Cell[3464339, 90770, 191, 4, 27, "Input",ExpressionUUID->"d3b497a4-b15f-47a7-991c-9d8d0cf1185a",
 CellID->724731817],
Cell[3464533, 90776, 1122, 28, 70, "Output",ExpressionUUID->"65aa2173-fd27-4ced-8fbe-567372aa234e",
 CellID->1657453880]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3465704, 90810, 213, 4, 40, "MathCaption",ExpressionUUID->"527978eb-cb87-4715-848e-211a12fa3129",
 CellID->440573293],
Cell[CellGroupData[{
Cell[3465942, 90818, 162, 3, 27, "Input",ExpressionUUID->"9737edba-9e4d-4cde-b355-4a2b3665cde3",
 CellID->191720595],
Cell[3466107, 90823, 1458, 23, 70, "Message",ExpressionUUID->"da9d62e4-d83b-49ce-a120-f020ed262a30",
 CellID->1816975923],
Cell[3467568, 90848, 10870, 224, 70, "Output",ExpressionUUID->"3d2b855c-b4b2-41fb-8824-a94633fbe0e0",
 CellID->696877521]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3478487, 91078, 194, 4, 40, "MathCaption",ExpressionUUID->"76c09268-669c-4f14-a0e5-932b0387ae95",
 CellID->2053850088],
Cell[CellGroupData[{
Cell[3478706, 91086, 155, 3, 27, "Input",ExpressionUUID->"42a5f6ef-fd0e-42d0-a4eb-ed4338bbe8a1",
 CellID->406287138],
Cell[3478864, 91091, 1896, 53, 70, "Output",ExpressionUUID->"57422f81-3012-4e46-9e04-199daca22218",
 CellID->44684709]
}, Open  ]],
Cell[CellGroupData[{
Cell[3480797, 91149, 274, 7, 27, "Input",ExpressionUUID->"4a1967fa-c885-4322-96d2-98c864b1c8a9",
 CellID->1038388545],
Cell[3481074, 91158, 1744, 49, 70, "Output",ExpressionUUID->"b28d7da9-1ab9-4b97-870e-8860825ccdb1",
 CellID->1544111360]
}, Open  ]]
}, Open  ]],
Cell[3482845, 91211, 386, 8, 40, "Text",ExpressionUUID->"e28e9306-fece-4925-985a-9797a807b043",
 CellID->950262901],
Cell[CellGroupData[{
Cell[3483256, 91223, 283, 5, 40, "MathCaption",ExpressionUUID->"2cdb493e-75e6-47c3-925c-a2b15578a003",
 CellID->232694368],
Cell[3483542, 91230, 307, 7, 27, "Input",ExpressionUUID->"cdf01273-2ad2-4b66-9710-e93d92789e28",
 CellID->326521789]
}, Open  ]],
Cell[CellGroupData[{
Cell[3483886, 91242, 220, 4, 40, "MathCaption",ExpressionUUID->"92c20e72-1098-4ce1-9e71-f786f0ad2678",
 CellID->1030989638],
Cell[CellGroupData[{
Cell[3484131, 91250, 352, 9, 27, "Input",ExpressionUUID->"cf28bfdd-a561-411e-873b-6942866c7a43",
 CellID->1944781870],
Cell[3484486, 91261, 801, 15, 70, "Output",ExpressionUUID->"06b8b117-c613-449e-bde6-fcacd4a0a48c",
 CellID->1687237172]
}, Open  ]],
Cell[CellGroupData[{
Cell[3485324, 91281, 223, 4, 27, "Input",ExpressionUUID->"8e4ddb28-cd5f-4c2b-b7be-58993e16c9a2",
 CellID->1290059295],
Cell[3485550, 91287, 2008, 55, 70, "Output",ExpressionUUID->"ffc2d082-1c18-4310-9847-e6511d77fbca",
 CellID->2021074816]
}, Open  ]],
Cell[CellGroupData[{
Cell[3487595, 91347, 240, 5, 27, "Input",ExpressionUUID->"98c86999-7c32-4512-aa90-ea2b51283263",
 CellID->510741193],
Cell[3487838, 91354, 3189, 85, 70, "Output",ExpressionUUID->"785051c1-38f0-403f-8e52-5b2f4d0faee4",
 CellID->1455587913]
}, Open  ]],
Cell[CellGroupData[{
Cell[3491064, 91444, 333, 9, 27, "Input",ExpressionUUID->"1db299f2-d128-44a1-b0a6-fc6c51d45fff",
 CellID->1805651455],
Cell[3491400, 91455, 3181, 85, 70, "Output",ExpressionUUID->"c69088f3-8381-4c25-b213-0ecf9bbfa903",
 CellID->2079830814]
}, Open  ]],
Cell[CellGroupData[{
Cell[3494618, 91545, 205, 5, 27, "Input",ExpressionUUID->"e2fa1e5b-67fa-4a73-aaf5-3edda9c20898",
 CellID->1177228789],
Cell[3494826, 91552, 2021, 55, 70, "Output",ExpressionUUID->"17f4e406-92a9-41f8-883b-8a9d10eba7a8",
 CellID->1125168679]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3496896, 91613, 290, 5, 40, "MathCaption",ExpressionUUID->"fbb4da9a-588e-458a-9906-20abbb057bee",
 CellID->1087857793],
Cell[CellGroupData[{
Cell[3497211, 91622, 352, 9, 27, "Input",ExpressionUUID->"84cc1f4c-246f-4de4-983e-77a654e8879a",
 CellID->1786613944],
Cell[3497566, 91633, 7146, 193, 70, "Output",ExpressionUUID->"faccadf6-dcee-4ac1-a864-2867f481ae8c",
 CellID->2116476118]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3504773, 91833, 212, 2, 41, "Subsubsection",ExpressionUUID->"15ff0bd8-6a27-4801-94e2-215f648c4a35",
 CellID->93454524],
Cell[3504988, 91837, 720, 11, 57, "Text",ExpressionUUID->"fba372ca-b450-469f-88a5-0f6c50231f23",
 CellID->879638911],
Cell[CellGroupData[{
Cell[3505733, 91852, 425, 7, 56, "MathCaption",ExpressionUUID->"2fef869b-fbb6-4446-acea-f02c79cc4c03",
 CellID->813751229],
Cell[CellGroupData[{
Cell[3506183, 91863, 702, 20, 27, "Input",ExpressionUUID->"9fd032d3-5d29-4298-ab0c-b20ba21aca17",
 CellID->1541041707],
Cell[CellGroupData[{
Cell[3506910, 91887, 1987, 48, 36, "Print",ExpressionUUID->"60fff894-fe95-45a0-91d0-af8acc69a4f4",
 CellID->694986318],
Cell[3508900, 91937, 2009, 49, 36, "Print",ExpressionUUID->"bfe89d69-7c36-4533-89ee-fb67ebe89e92",
 CellID->471490593],
Cell[3510912, 91988, 1987, 48, 36, "Print",ExpressionUUID->"79e4ea9b-b100-45b0-b36f-ccf76e78be62",
 CellID->127739121],
Cell[3512902, 92038, 1988, 48, 36, "Print",ExpressionUUID->"527b8224-38b9-4acf-99d6-c7664481c208",
 CellID->1575243294],
Cell[3514893, 92088, 1951, 47, 28, "Print",ExpressionUUID->"cd84e3e3-21f3-4585-8102-28cf0c483bf8",
 CellID->935642042],
Cell[3516847, 92137, 1954, 47, 28, "Print",ExpressionUUID->"1bbe3d4d-f4d8-4c3f-bb13-9144e5f456fd",
 CellID->1460536560],
Cell[3518804, 92186, 1975, 48, 28, "Print",ExpressionUUID->"edaa92c1-55e0-4c37-8c30-1682522f3dec",
 CellID->1499878710],
Cell[3520782, 92236, 1952, 47, 28, "Print",ExpressionUUID->"ba7fe945-5931-4677-b721-e770a66b25e5",
 CellID->1298430484],
Cell[3522737, 92285, 1751, 44, 44, "Print",ExpressionUUID->"f0ea909e-07cc-4403-986e-257c1d424a9b",
 CellID->2139807421]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3524549, 92336, 203, 2, 40, "MathCaption",ExpressionUUID->"b0de54d5-6a1b-4eab-9f5f-0cca5eaecac2",
 CellID->1635186729],
Cell[CellGroupData[{
Cell[3524777, 92342, 286, 8, 27, "Input",ExpressionUUID->"de6dded8-0a70-486a-9974-541ee7bf40db",
 CellID->1235123644],
Cell[3525066, 92352, 794, 15, 26, "Output",ExpressionUUID->"39be02f8-cb6f-4156-9069-c742f7fbb547",
 CellID->1982172085]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3525909, 92373, 193, 2, 40, "MathCaption",ExpressionUUID->"46c8960f-c0cc-46fd-b2de-4d68db8b175c",
 CellID->461896852],
Cell[CellGroupData[{
Cell[3526127, 92379, 213, 5, 27, "Input",ExpressionUUID->"4f820507-1328-47d1-8df3-0c0660860dd8",
 CellID->2094491306],
Cell[3526343, 92386, 857, 17, 40, "Output",ExpressionUUID->"369bb546-b40b-454c-bdea-2bb7f1c3d7ee",
 CellID->452416684]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3527249, 92409, 206, 2, 40, "MathCaption",ExpressionUUID->"0e25f331-0ac8-4564-ad53-c974391e09db",
 CellID->1452411003],
Cell[CellGroupData[{
Cell[3527480, 92415, 211, 5, 27, "Input",ExpressionUUID->"c6bf1f87-4479-4022-9483-5b77ab1d236c",
 CellID->225111972],
Cell[3527694, 92422, 810, 15, 26, "Output",ExpressionUUID->"dd781b5f-f301-4d47-8dde-3aa31f6a2c89",
 CellID->1675097608]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3528553, 92443, 204, 2, 40, "MathCaption",ExpressionUUID->"390144d5-6ca7-4946-b5ae-1be03a560267",
 CellID->1892848718],
Cell[CellGroupData[{
Cell[3528782, 92449, 263, 7, 27, "Input",ExpressionUUID->"d33eb0b0-af68-4a20-b544-fc54e1a294fb",
 CellID->285656768],
Cell[3529048, 92458, 958, 21, 26, "Output",ExpressionUUID->"20426501-18f9-49fd-b96f-326538e7e5f8",
 CellID->2002369133]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3530055, 92485, 217, 2, 40, "MathCaption",ExpressionUUID->"b9137542-a102-4723-acfb-eca83836ed60",
 CellID->925233859],
Cell[CellGroupData[{
Cell[3530297, 92491, 261, 7, 27, "Input",ExpressionUUID->"8d074e6b-da6f-46db-bfa1-27794f1062e0",
 CellID->476380286],
Cell[3530561, 92500, 957, 22, 26, "Output",ExpressionUUID->"8f31ad41-a3b1-4a41-9496-e31cf20d6544",
 CellID->601786194]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3531567, 92528, 205, 2, 40, "MathCaption",ExpressionUUID->"f297dc2a-0116-4f54-8d13-f2053d3421dd",
 CellID->290874391],
Cell[CellGroupData[{
Cell[3531797, 92534, 130, 2, 27, "Input",ExpressionUUID->"4baaf4d7-1226-43d4-92fb-009afb7220b3",
 CellID->1053108965],
Cell[3531930, 92538, 1419, 36, 29, "Output",ExpressionUUID->"3743728b-37c0-4512-8a62-24310fa10372",
 CellID->1720068835]
}, Open  ]],
Cell[CellGroupData[{
Cell[3533386, 92579, 283, 8, 27, "Input",ExpressionUUID->"d035ad0d-37a0-4a93-9525-1b06045d818b",
 CellID->1996814204],
Cell[3533672, 92589, 2977, 87, 29, "Output",ExpressionUUID->"07ea3ab9-d72c-4c71-bede-04d4c98f53bc",
 CellID->1919537578]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3536698, 92682, 223, 3, 40, "MathCaption",ExpressionUUID->"ec790cea-789b-41ee-8075-28dc0a46f31b",
 CellID->139451525],
Cell[CellGroupData[{
Cell[3536946, 92689, 239, 4, 27, "Input",ExpressionUUID->"e6fe2094-8b84-408c-bfd7-5f29de64a910",
 CellID->15590300],
Cell[3537188, 92695, 871, 20, 26, "Output",ExpressionUUID->"9c4904a5-0132-41bd-923e-c32cacb80199",
 CellID->21244565]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3538108, 92721, 341, 6, 40, "MathCaption",ExpressionUUID->"ab8552c2-4b44-4df8-a1af-85a2c6722ad5",
 CellID->1561257415],
Cell[CellGroupData[{
Cell[3538474, 92731, 264, 7, 27, "Input",ExpressionUUID->"46935187-9e7e-4eb4-989b-e221c472609b",
 CellID->2098174859],
Cell[3538741, 92740, 967, 22, 26, "Output",ExpressionUUID->"3669b687-7f90-4281-8b0e-3725ebe0f2b7",
 CellID->575118553]
}, Open  ]],
Cell[CellGroupData[{
Cell[3539745, 92767, 127, 2, 27, "Input",ExpressionUUID->"d5ed94c4-1a08-4067-88a2-687919640ab6",
 CellID->194260731],
Cell[3539875, 92771, 805, 15, 26, "Output",ExpressionUUID->"af4593b6-0970-4f94-9f1b-70f141f03894",
 CellID->193625124]
}, Open  ]],
Cell[CellGroupData[{
Cell[3540717, 92791, 232, 6, 27, "Input",ExpressionUUID->"c16e1072-6ad6-4a8f-aba0-dae1953728f7",
 CellID->1887239997],
Cell[3540952, 92799, 1202, 30, 29, "Output",ExpressionUUID->"0f1228f9-df98-4cf0-9b74-6e1d89492cb6",
 CellID->819733863]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3542203, 92835, 249, 4, 40, "MathCaption",ExpressionUUID->"96897e6f-4240-4c44-829d-734f69e2b231",
 CellID->1934229275],
Cell[3542455, 92841, 280, 7, 27, "Input",ExpressionUUID->"0d6ce07e-9a57-41d4-bb16-711464c62053",
 CellID->2131460286],
Cell[CellGroupData[{
Cell[3542760, 92852, 261, 7, 27, "Input",ExpressionUUID->"6f5d3fbd-1c61-47e4-91e9-50f3e21a36a9",
 CellID->1278311195],
Cell[3543024, 92861, 5318, 141, 32, "Output",ExpressionUUID->"32fdf79b-164c-4473-a308-9e0043e85b84",
 CellID->1159439936]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3548391, 93008, 210, 2, 40, "MathCaption",ExpressionUUID->"053464b8-c7fb-4e89-b6da-38b1bc48886a",
 CellID->319862686],
Cell[CellGroupData[{
Cell[3548626, 93014, 388, 11, 27, "Input",ExpressionUUID->"fb4d4a04-5d86-4f1a-b2e4-0effda1e513f",
 CellID->780595161],
Cell[3549017, 93027, 4926, 136, 32, "Output",ExpressionUUID->"6dcf4c65-8cbd-4a4d-bfe2-42fcd1194bf3",
 CellID->440391474]
}, Open  ]]
}, Open  ]],
Cell[3553970, 93167, 368, 6, 40, "Text",ExpressionUUID->"26b4b618-ed42-47dd-aca8-b75197131f29",
 CellID->1179051796],
Cell[CellGroupData[{
Cell[3554363, 93177, 241, 4, 40, "MathCaption",ExpressionUUID->"62902d7c-791a-4470-8fe4-9c5884100b3d",
 CellID->1856836499],
Cell[3554607, 93183, 377, 8, 27, "Input",ExpressionUUID->"81664d37-92e8-43c8-983d-ecf693d02258",
 CellID->1318474035]
}, Open  ]],
Cell[CellGroupData[{
Cell[3555021, 93196, 191, 2, 40, "MathCaption",ExpressionUUID->"6060a072-d7f2-4330-8a15-3fd9e7c9055f",
 CellID->701825391],
Cell[CellGroupData[{
Cell[3555237, 93202, 555, 16, 27, "Input",ExpressionUUID->"c3928df2-6ff2-44f7-8b3f-6948b6e43207",
 CellID->1009691769],
Cell[CellGroupData[{
Cell[3555817, 93222, 1986, 48, 36, "Print",ExpressionUUID->"e4e7cf8d-36f4-4fa5-93b6-15d5b5d7a12e",
 CellID->1430855956],
Cell[3557806, 93272, 2009, 49, 36, "Print",ExpressionUUID->"6d086ec4-0f75-44c3-a9d0-60e8065afc4f",
 CellID->874173386],
Cell[3559818, 93323, 1986, 48, 36, "Print",ExpressionUUID->"77edd136-6347-449d-8e62-aeada9a009d4",
 CellID->1296469375],
Cell[3561807, 93373, 1986, 48, 36, "Print",ExpressionUUID->"402ab056-cedd-42a7-a0b5-431f83ad6f52",
 CellID->1391703163],
Cell[3563796, 93423, 1950, 47, 28, "Print",ExpressionUUID->"d32a9e07-0320-49de-aae9-91d60d0f5784",
 CellID->2110661064],
Cell[3565749, 93472, 1950, 47, 28, "Print",ExpressionUUID->"e63d5f98-6fcd-4fbe-b810-7a42a67cca99",
 CellID->1745621502],
Cell[3567702, 93521, 1973, 48, 28, "Print",ExpressionUUID->"1fdc16b1-cc0f-400e-b577-6af6e448ba53",
 CellID->1886482447],
Cell[3569678, 93571, 1951, 47, 28, "Print",ExpressionUUID->"a9894a20-8fb8-4620-875d-18355110a5b9",
 CellID->661922421],
Cell[3571632, 93620, 1748, 44, 44, "Print",ExpressionUUID->"4d8efc83-7475-47bb-9cdf-0ba2f7996838",
 CellID->964838287]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3573441, 93671, 247, 4, 40, "MathCaption",ExpressionUUID->"f369526a-c35e-421c-a629-77290182e26d",
 CellID->1606319613],
Cell[CellGroupData[{
Cell[3573713, 93679, 281, 6, 27, "Input",ExpressionUUID->"c298d41b-3b49-41c1-84e0-b1ddd935186c",
 CellID->1812900123],
Cell[3573997, 93687, 784, 16, 40, "Output",ExpressionUUID->"c3e70acc-d37f-4a93-9f25-23c801a96cc0",
 CellID->388071112]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3574842, 93710, 197, 2, 41, "Subsubsection",ExpressionUUID->"526aaba5-f88e-4422-b189-ffe8f3a4dbec",
 CellID->1961079369],
Cell[3575042, 93714, 271, 5, 23, "Text",ExpressionUUID->"df4fcd98-f0e4-49a5-9920-af1889df2862",
 CellID->281940627],
Cell[CellGroupData[{
Cell[3575338, 93723, 217, 2, 40, "MathCaption",ExpressionUUID->"612e8563-4f91-4bcb-92e1-6e1a69045d72",
 CellID->1227192707],
Cell[CellGroupData[{
Cell[3575580, 93729, 129, 2, 27, "Input",ExpressionUUID->"c1245b0c-51ad-45cc-9ed9-51325ead73e2",
 CellID->994880761],
Cell[3575712, 93733, 1386, 36, 70, "Output",ExpressionUUID->"ef15ec67-5611-4423-b449-6313e9d6063d",
 CellID->403343445]
}, Open  ]],
Cell[CellGroupData[{
Cell[3577135, 93774, 324, 10, 27, "Input",ExpressionUUID->"ef71946e-cfa0-423f-a1ab-da230c81efb5",
 CellID->1649546631],
Cell[3577462, 93786, 985, 24, 70, "Output",ExpressionUUID->"d769e564-959e-4261-a62c-ea10e6d0d961",
 CellID->1534890183]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3578496, 93816, 180, 2, 40, "MathCaption",ExpressionUUID->"f186d316-a00a-432a-af88-43e6c99873b8",
 CellID->602909514],
Cell[CellGroupData[{
Cell[3578701, 93822, 326, 10, 27, "Input",ExpressionUUID->"6e9ca28a-248d-42e7-abaa-82d8db1ca3c2",
 CellID->1272355252],
Cell[3579030, 93834, 1056, 26, 70, "Output",ExpressionUUID->"bb382c09-e9d0-44ce-88fb-7be6182ddb4c",
 CellID->1325737917]
}, Open  ]],
Cell[CellGroupData[{
Cell[3580123, 93865, 397, 12, 27, "Input",ExpressionUUID->"be007877-8bfa-4931-bb52-095d2851a0fb",
 CellID->1246444383],
Cell[3580523, 93879, 7808, 200, 70, "Output",ExpressionUUID->"ef1a9e4e-275d-464f-bf15-68f92aa8d953",
 CellID->1534333394]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3588380, 94085, 161, 2, 40, "MathCaption",ExpressionUUID->"3beab8ea-c782-45bb-8c68-49e22cb9e130",
 CellID->393619928],
Cell[CellGroupData[{
Cell[3588566, 94091, 325, 10, 27, "Input",ExpressionUUID->"d36c6d43-00c0-4325-bc5b-23f029e1eb47",
 CellID->1320497136],
Cell[3588894, 94103, 678, 11, 70, "Output",ExpressionUUID->"cb6df262-0e60-4bce-9f8f-d2c6fe7661bb",
 CellID->557719782]
}, Open  ]],
Cell[CellGroupData[{
Cell[3589609, 94119, 396, 12, 27, "Input",ExpressionUUID->"61f9a81f-85e7-4ae2-a7c4-b13772dc6be3",
 CellID->1114169880],
Cell[3590008, 94133, 4973, 130, 70, "Output",ExpressionUUID->"482e077e-8a61-4736-b46e-a450f8ad931d",
 CellID->1977740566]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3595030, 94269, 183, 2, 40, "MathCaption",ExpressionUUID->"ddcecfa2-7427-49a6-b65e-b0bc78362f15",
 CellID->1354926793],
Cell[CellGroupData[{
Cell[3595238, 94275, 280, 8, 27, "Input",ExpressionUUID->"dcbad002-4fd1-4651-904b-d9793639b887",
 CellID->338017163],
Cell[3595521, 94285, 3099, 84, 70, "Output",ExpressionUUID->"509bc2dd-c7f8-4bb9-ba6f-ba00bc8c22cf",
 CellID->1133804496]
}, Open  ]],
Cell[CellGroupData[{
Cell[3598657, 94374, 280, 8, 27, "Input",ExpressionUUID->"e9316541-8d42-41d5-bc92-dfc777533be4",
 CellID->2094758824],
Cell[3598940, 94384, 3086, 84, 70, "Output",ExpressionUUID->"b3223866-21b2-4647-b26a-76550379a148",
 CellID->1229764441]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3602075, 94474, 209, 4, 40, "MathCaption",ExpressionUUID->"02face2d-1d65-4346-8537-a6b804ba15cf",
 CellID->864018702],
Cell[CellGroupData[{
Cell[3602309, 94482, 341, 10, 27, "Input",ExpressionUUID->"7bc19b58-d17a-41ae-afbb-57367ef17c8e",
 CellID->277903500],
Cell[3602653, 94494, 5143, 151, 70, "Output",ExpressionUUID->"1eb397ff-d066-494e-a219-5f250524a12f",
 CellID->1268341159]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3607845, 94651, 254, 5, 40, "MathCaption",ExpressionUUID->"2f55925f-fb97-4f0b-b45b-e06dee9db503",
 CellID->1518795783],
Cell[CellGroupData[{
Cell[3608124, 94660, 236, 6, 27, "Input",ExpressionUUID->"04673799-f987-40df-a745-4127362c5c90",
 CellID->1410096639],
Cell[3608363, 94668, 3748, 98, 70, "Output",ExpressionUUID->"d291ecd7-b37b-4271-94bc-3159b42fb16b",
 CellID->1087179191]
}, Open  ]],
Cell[CellGroupData[{
Cell[3612148, 94771, 163, 3, 27, "Input",ExpressionUUID->"af6ca114-92bd-46eb-bc32-db8de7b18e11",
 CellID->2066250287],
Cell[3612314, 94776, 4757, 139, 70, "Output",ExpressionUUID->"21807853-fe18-4c6b-9873-e084f2ca5b39",
 CellID->1429794461]
}, Open  ]],
Cell[CellGroupData[{
Cell[3617108, 94920, 230, 4, 27, "Input",ExpressionUUID->"2bba8e08-d993-4528-8ebb-fd73ea30dacf",
 CellID->1930809927],
Cell[3617341, 94926, 3535, 95, 70, "Output",ExpressionUUID->"6d4a7d53-a06f-4c33-b7e5-f166fc69f6f2",
 CellID->213496097]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3620937, 95028, 181, 2, 41, "Subsubsection",ExpressionUUID->"9219f830-e3e8-4579-bdbb-02abbf1b4098",
 CellID->1544931844],
Cell[3621121, 95032, 326, 8, 27, "Input",ExpressionUUID->"d5fed17a-4ad1-4f5a-bfdd-72e34fdcb376",
 CellID->309267742],
Cell[CellGroupData[{
Cell[3621472, 95044, 110, 1, 40, "MathCaption",ExpressionUUID->"11f8b1df-d6d6-4dcd-b716-a32c7a95537f",
 CellID->444726231],
Cell[3621585, 95047, 233, 4, 27, "Input",ExpressionUUID->"7e87df96-e388-4f29-a100-d5786daeb8dc",
 CellID->2144554484],
Cell[3621821, 95053, 167, 3, 27, "Input",ExpressionUUID->"59f827e0-dbde-4fe9-a5a7-6c275e8a6887",
 CellID->1270672857],
Cell[3621991, 95058, 166, 3, 27, "Input",ExpressionUUID->"7aee4012-f840-476a-9b08-a05d3ff8ad68",
 CellID->1621321831],
Cell[3622160, 95063, 164, 3, 27, "Input",ExpressionUUID->"8b27a859-5c79-4550-b010-52f9d40d36c1",
 CellID->136016834],
Cell[3622327, 95068, 168, 3, 27, "Input",ExpressionUUID->"9990bab1-c31b-4e75-ab66-4c61ce21646c",
 CellID->1112328630]
}, Open  ]],
Cell[CellGroupData[{
Cell[3622532, 95076, 110, 1, 40, "MathCaption",ExpressionUUID->"18c7abda-191f-4f10-b6bf-1cb3819bed1b",
 CellID->1062137275],
Cell[3622645, 95079, 164, 3, 27, "Input",ExpressionUUID->"a367107f-1e83-4ff4-929d-53c8bf1b67bd",
 CellID->1457107061],
Cell[3622812, 95084, 165, 3, 27, "Input",ExpressionUUID->"757e2dc0-c6cd-42af-a197-7f2fcb63f87c",
 CellID->1842347486],
Cell[3622980, 95089, 167, 3, 27, "Input",ExpressionUUID->"33b676a7-8d85-486e-9251-de3765b20320",
 CellID->1914974756],
Cell[3623150, 95094, 345, 7, 27, "Input",ExpressionUUID->"95ee7682-1e99-406e-9e54-cb8d7fb7e787",
 CellID->643306503],
Cell[3623498, 95103, 171, 3, 27, "Input",ExpressionUUID->"9e8b6f7a-d5c9-4eb8-a9a9-d73fcce1ca33",
 CellID->1619716923]
}, Open  ]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[3623730, 95113, 237, 3, 39, "Subsection",ExpressionUUID->"303e3f2a-6f81-4934-ad85-b656a9fa266f",
 CellID->1079615489],
Cell[CellGroupData[{
Cell[3623992, 95120, 195, 2, 41, "Subsubsection",ExpressionUUID->"a311af45-e61b-43f4-8098-ebc8efe9f940",
 CellID->1008190120],
Cell[3624190, 95124, 549, 8, 57, "Text",ExpressionUUID->"f3d2a201-a843-436f-a567-17922544619f",
 CellID->1005372010],
Cell[CellGroupData[{
Cell[3624764, 95136, 206, 2, 40, "MathCaption",ExpressionUUID->"0518d0c5-7325-4fe1-88bd-d80d5ab10d70",
 CellID->886037474],
Cell[3624973, 95140, 290, 7, 27, "Input",ExpressionUUID->"1d878512-9fa2-4918-91d5-3ee3bc3b598c",
 CellID->980217220],
Cell[3625266, 95149, 367, 10, 27, "Input",ExpressionUUID->"540e6284-c53b-422d-b045-366cb90d08ae",
 CellID->1163760403],
Cell[CellGroupData[{
Cell[3625658, 95163, 531, 17, 27, "Input",ExpressionUUID->"0f79031b-f894-41fe-a68f-847113d5b085",
 CellID->40036345],
Cell[3626192, 95182, 988, 26, 70, "Output",ExpressionUUID->"eafd919a-7318-443e-83f4-1bad751da22c",
 CellID->345888892]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3627229, 95214, 340, 6, 40, "MathCaption",ExpressionUUID->"a1b8f519-bad1-47ad-81cd-8a81346a5a68",
 CellID->780044460],
Cell[CellGroupData[{
Cell[3627594, 95224, 192, 4, 27, "Input",ExpressionUUID->"635d181a-7fed-47fc-b646-47206876bf10",
 CellID->1967057932],
Cell[3627789, 95230, 2352, 70, 70, "Output",ExpressionUUID->"d0288911-1b53-4fcc-8a99-a8c20da04d6a",
 CellID->116375415]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3630190, 95306, 195, 2, 40, "MathCaption",ExpressionUUID->"9250d531-2f51-470e-b70b-a574e90397c7",
 CellID->585314454],
Cell[CellGroupData[{
Cell[3630410, 95312, 261, 5, 27, "Input",ExpressionUUID->"b5b9fe63-4b23-4b86-99a4-e04a53883d61",
 CellID->1828426690],
Cell[3630674, 95319, 1553, 47, 70, "Output",ExpressionUUID->"9ce89680-c643-4711-8df1-9da04c901d58",
 CellID->1711481135]
}, Open  ]],
Cell[CellGroupData[{
Cell[3632264, 95371, 292, 7, 27, "Input",ExpressionUUID->"9a8339cd-5b0c-47c2-9ef8-d1a98816d304",
 CellID->1416228891],
Cell[3632559, 95380, 6015, 162, 70, "Output",ExpressionUUID->"30acdedf-54f3-46fc-90bf-dca3ce1e0562",
 CellID->1467468027]
}, Open  ]],
Cell[CellGroupData[{
Cell[3638611, 95547, 237, 4, 27, "Input",ExpressionUUID->"6a621bd8-023a-43c5-9186-01591447111a",
 CellID->1960374463],
Cell[3638851, 95553, 9241, 282, 70, "Output",ExpressionUUID->"5fbcb91b-8364-418f-a469-ff3ff42ad5f1",
 CellID->2032927641]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3648141, 95841, 245, 4, 40, "MathCaption",ExpressionUUID->"380feb36-a0e0-4718-b325-664637eb0555",
 CellID->1126893565],
Cell[CellGroupData[{
Cell[3648411, 95849, 250, 7, 27, "Input",ExpressionUUID->"18f5f0fb-a7d8-4065-8c76-15b8f31bd024",
 CellID->1041864062],
Cell[3648664, 95858, 1532, 43, 70, "Output",ExpressionUUID->"61239a10-8bf7-476e-9b94-7742de5c261c",
 CellID->950038517]
}, Open  ]],
Cell[CellGroupData[{
Cell[3650233, 95906, 373, 11, 27, "Input",ExpressionUUID->"f3c48a0a-97d7-4769-8644-d5f2a2e34d9c",
 CellID->561211820],
Cell[3650609, 95919, 2944, 80, 70, "Output",ExpressionUUID->"beffe58f-1617-47ad-9e4f-4011c1a766b3",
 CellID->756439774]
}, Open  ]],
Cell[CellGroupData[{
Cell[3653590, 96004, 297, 9, 27, "Input",ExpressionUUID->"b8c2f8a5-1953-4d93-8580-5bfeaef9fea4",
 CellID->1036377213],
Cell[3653890, 96015, 478, 7, 70, "Output",ExpressionUUID->"2b479b1c-f48b-4820-bd73-0546431dbb3e",
 CellID->380091378]
}, Open  ]],
Cell[CellGroupData[{
Cell[3654405, 96027, 476, 15, 27, "Input",ExpressionUUID->"2d41266b-b549-4bf5-b5f2-8ff42df558c4",
 CellID->601293435],
Cell[3654884, 96044, 6484, 171, 70, "Output",ExpressionUUID->"3b86da22-7c13-414f-9d40-56d3451996e5",
 CellID->1579423868]
}, Open  ]],
Cell[CellGroupData[{
Cell[3661405, 96220, 161, 3, 27, "Input",ExpressionUUID->"a05281f7-f839-4be5-ba0d-6508a6201bb8",
 CellID->998379539],
Cell[3661569, 96225, 6489, 172, 70, "Output",ExpressionUUID->"fe96ea97-bf6e-4169-98f7-4aadae456a3d",
 CellID->331588909]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3668107, 96403, 257, 5, 40, "MathCaption",ExpressionUUID->"9abb7009-df38-4644-89f9-09062863ab06",
 CellID->617302719],
Cell[3668367, 96410, 215, 5, 27, "Input",ExpressionUUID->"ddcae87a-6e7c-46c9-89c9-755774024044",
 CellID->513299576],
Cell[CellGroupData[{
Cell[3668607, 96419, 230, 6, 27, "Input",ExpressionUUID->"289e6d71-6989-4300-b578-5f04be3ad456",
 CellID->584163833],
Cell[3668840, 96427, 8089, 195, 70, "Output",ExpressionUUID->"47921bcd-1696-4873-b04a-9ff3dea885d5",
 CellID->294729598]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3676978, 96628, 259, 5, 40, "MathCaption",ExpressionUUID->"ad1a2c82-4caa-429b-ac00-0248150af90c",
 CellID->1086742571],
Cell[CellGroupData[{
Cell[3677262, 96637, 203, 4, 27, "Input",ExpressionUUID->"c3694b3d-7a66-4720-a121-681ced691f24",
 CellID->533973062],
Cell[3677468, 96643, 8354, 232, 70, "Output",ExpressionUUID->"e50f1d88-cf2d-4bd8-9eff-99ae693811d8",
 CellID->1423063604]
}, Open  ]],
Cell[CellGroupData[{
Cell[3685859, 96880, 204, 4, 27, "Input",ExpressionUUID->"60753910-c654-4c48-aaba-906fb3854db3",
 CellID->1081287593],
Cell[3686066, 96886, 22161, 584, 70, "Output",ExpressionUUID->"4d9afb6d-4035-4aa1-8d8a-9f8f57016ac5",
 CellID->904615789]
}, Open  ]],
Cell[CellGroupData[{
Cell[3708264, 97475, 160, 3, 27, "Input",ExpressionUUID->"3844f8cb-d448-48ef-bb3d-79e0405b15ab",
 CellID->708386491],
Cell[3708427, 97480, 22172, 584, 70, "Output",ExpressionUUID->"045d9e57-cbce-4307-bde3-fd87322684a0",
 CellID->1840828610]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3730648, 98070, 255, 4, 40, "MathCaption",ExpressionUUID->"357f43ed-9395-40fa-a355-f74dae9fb991",
 CellID->379329235],
Cell[CellGroupData[{
Cell[3730928, 98078, 202, 4, 27, "Input",ExpressionUUID->"a1780918-5795-4c0a-a85d-3c085fe2ec4d",
 CellID->1654988349],
Cell[3731133, 98084, 8088, 195, 70, "Output",ExpressionUUID->"81b43565-00a6-426b-b3cb-cc2935df5838",
 CellID->9545941]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3739270, 98285, 144, 1, 40, "MathCaption",ExpressionUUID->"e9c91a2d-6638-48a7-aa80-543285dcb2e2",
 CellID->1782777519],
Cell[CellGroupData[{
Cell[3739439, 98290, 299, 8, 27, "Input",ExpressionUUID->"84341d38-0f59-4af7-ab6f-dc53f7aa0abb",
 CellID->1449918889],
Cell[3739741, 98300, 9687, 265, 70, "Output",ExpressionUUID->"3ae51338-8a03-4bfa-85af-1e941238cdf8",
 CellID->1524840509]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3749489, 98572, 195, 2, 41, "Subsubsection",ExpressionUUID->"b6fcbbfa-7618-49c4-9dc9-5a7da8e14817",
 CellID->1387396780],
Cell[3749687, 98576, 476, 7, 40, "Text",ExpressionUUID->"0234e021-7ae8-4a3c-92d5-42e6d1e92560",
 CellID->644289339],
Cell[CellGroupData[{
Cell[3750188, 98587, 339, 6, 40, "MathCaption",ExpressionUUID->"838b5a17-192d-47a5-be44-e0ced7195f80",
 CellID->600242334],
Cell[CellGroupData[{
Cell[3750552, 98597, 162, 3, 27, "Input",ExpressionUUID->"2f8de1ad-be6e-4255-91a7-6209674f180c",
 CellID->1542217579],
Cell[3750717, 98602, 476, 7, 70, "Output",ExpressionUUID->"fa4ba825-32b0-4799-a388-77134e5c66ad",
 CellID->2072075112]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3751242, 98615, 262, 3, 40, "MathCaption",ExpressionUUID->"1bc39641-97d7-4f54-91b0-b3c9713f9c71",
 CellID->380466988],
Cell[CellGroupData[{
Cell[3751529, 98622, 162, 3, 27, "Input",ExpressionUUID->"512df925-f8d6-477d-bb1e-299e8e57561b",
 CellID->1449850197],
Cell[3751694, 98627, 473, 7, 70, "Output",ExpressionUUID->"c84135a6-f3b7-402b-951f-b28c9f30dd68",
 CellID->1771587632]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3752216, 98640, 202, 2, 40, "MathCaption",ExpressionUUID->"66895eb2-d23a-4f47-b675-34087eecb047",
 CellID->112343117],
Cell[CellGroupData[{
Cell[3752443, 98646, 532, 17, 27, "Input",ExpressionUUID->"abcc4cc3-cc43-4988-b119-dd27d5553be8",
 CellID->512641706],
Cell[3752978, 98665, 978, 26, 70, "Output",ExpressionUUID->"be8a52e0-e2e9-493b-a216-067de7708325",
 CellID->212565369]
}, Open  ]],
Cell[CellGroupData[{
Cell[3753993, 98696, 191, 4, 27, "Input",ExpressionUUID->"8b2fad4a-7c6f-482a-9fce-e6025be670d9",
 CellID->251640662],
Cell[3754187, 98702, 2420, 72, 70, "Output",ExpressionUUID->"0280e046-2761-4d32-955e-19a7513f2b8e",
 CellID->1320851822]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3756656, 98780, 188, 2, 40, "MathCaption",ExpressionUUID->"8494a88a-a005-487a-9f1e-4948cff37802",
 CellID->185461206],
Cell[CellGroupData[{
Cell[3756869, 98786, 261, 7, 27, "Input",ExpressionUUID->"114c8fdd-d5f8-4223-b966-cd758bc14627",
 CellID->1047717445],
Cell[3757133, 98795, 473, 7, 70, "Output",ExpressionUUID->"93db569b-a90b-48df-b6f8-d23c1d076dd9",
 CellID->1018113315]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3757655, 98808, 189, 2, 40, "MathCaption",ExpressionUUID->"a0744664-0ec4-421d-86ea-19dec982f701",
 CellID->674974360],
Cell[CellGroupData[{
Cell[3757869, 98814, 290, 8, 27, "Input",ExpressionUUID->"bba9eae9-5c27-426d-9d13-5c310600561b",
 CellID->1285445066],
Cell[3758162, 98824, 10228, 268, 70, "Output",ExpressionUUID->"92576d3c-558a-4a85-8e08-2df7f8524f27",
 CellID->280785909]
}, Open  ]],
Cell[CellGroupData[{
Cell[3768427, 99097, 276, 5, 27, "Input",ExpressionUUID->"eabe65c0-b6ba-44da-b392-4a73abc619c7",
 CellID->2135112499],
Cell[3768706, 99104, 761, 19, 70, "Output",ExpressionUUID->"f3a94b6b-87e7-4013-834b-7916ce9a66e8",
 CellID->1147200104]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3769516, 99129, 322, 5, 40, "MathCaption",ExpressionUUID->"711e4996-a13f-4841-86ad-132978480215",
 CellID->652069640],
Cell[CellGroupData[{
Cell[3769863, 99138, 473, 14, 27, "Input",ExpressionUUID->"cec7a0d4-bc92-4c60-a36b-3a0a4c019911",
 CellID->1096816598],
Cell[3770339, 99154, 10099, 265, 70, "Output",ExpressionUUID->"bcf872d2-80ef-443f-8950-5dce06e19b75",
 CellID->883965162]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3780487, 99425, 312, 6, 40, "MathCaption",ExpressionUUID->"5dbe9a51-caa2-43f3-a4bd-f2dd24c1e40d",
 CellID->191320769],
Cell[CellGroupData[{
Cell[3780824, 99435, 313, 9, 27, "Input",ExpressionUUID->"8f56b820-42f1-4252-8457-6b3a01a56569",
 CellID->1634635234],
Cell[3781140, 99446, 10167, 266, 70, "Output",ExpressionUUID->"7cf83d29-7fd4-4d34-81a1-465c0cd681f3",
 CellID->510607788]
}, Open  ]],
Cell[CellGroupData[{
Cell[3791344, 99717, 281, 5, 27, "Input",ExpressionUUID->"50fe4117-7f05-4849-b738-409ae4ee0690",
 CellID->195231078],
Cell[3791628, 99724, 1262, 37, 70, "Output",ExpressionUUID->"acb7c167-1e16-49aa-8947-566ea0d69887",
 CellID->440439504]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3792939, 99767, 188, 2, 40, "MathCaption",ExpressionUUID->"f301d7b1-fba6-4107-942d-d4bb2753e163",
 CellID->44226057],
Cell[CellGroupData[{
Cell[3793152, 99773, 248, 7, 27, "Input",ExpressionUUID->"2df2d09b-1e12-46c4-9b70-28b3492dabc9",
 CellID->217426225],
Cell[3793403, 99782, 3767, 101, 70, "Output",ExpressionUUID->"0fafe665-3696-4fe9-bc09-7352c7945006",
 CellID->90612519]
}, Open  ]],
Cell[CellGroupData[{
Cell[3797207, 99888, 189, 4, 27, "Input",ExpressionUUID->"cdf9d9d9-77f8-47cd-9816-bb2cdfb6f81f",
 CellID->1547945717],
Cell[3797399, 99894, 677, 16, 70, "Output",ExpressionUUID->"558920b3-1f10-405f-9d73-af071bdb29bd",
 CellID->358701110]
}, Open  ]]
}, Open  ]],
Cell[3798103, 99914, 463, 8, 40, "Text",ExpressionUUID->"ff44fa7d-2406-4075-9176-04614a19e6da",
 CellID->1911163297],
Cell[CellGroupData[{
Cell[3798591, 99926, 256, 3, 40, "MathCaption",ExpressionUUID->"cfa7fc30-7b27-429b-8b7c-d85b71400bde",
 CellID->163945661],
Cell[CellGroupData[{
Cell[3798872, 99933, 714, 21, 27, "Input",ExpressionUUID->"240db47b-aca9-4e26-9ec2-5604bc04b910",
 CellID->934847067],
Cell[3799589, 99956, 3566, 116, 70, "Output",ExpressionUUID->"47139ffd-2c77-4cd5-8f6f-54e7b035f14a",
 CellID->702958641]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3803204, 100078, 207, 2, 40, "MathCaption",ExpressionUUID->"dfba7197-d26f-4bf3-bf3a-b2a0524bfa56",
 CellID->540626935],
Cell[CellGroupData[{
Cell[3803436, 100084, 452, 12, 27, "Input",ExpressionUUID->"62b3e6ab-c2e4-4a80-8161-7c78ec091f71",
 CellID->2016378214],
Cell[3803891, 100098, 41345, 1062, 70, "Output",ExpressionUUID->"47a25be4-3bdb-44ec-ad35-7c7b10055f4a",
 CellID->1318337140]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3845297, 101167, 185, 2, 41, "Subsubsection",ExpressionUUID->"78e8b403-cf08-41dd-a413-ef7a35faf34f",
 CellID->1674752848],
Cell[3845485, 101171, 167, 3, 27, "Input",ExpressionUUID->"6b5d4cd4-d760-406e-9039-12d4030c623c",
 CellID->502955107],
Cell[3845655, 101176, 164, 3, 27, "Input",ExpressionUUID->"fb077724-34c3-4da1-8531-2e161d064532",
 CellID->670848757],
Cell[3845822, 101181, 167, 3, 27, "Input",ExpressionUUID->"75fb1a05-5fdf-4593-adf9-2e87ba3d6f05",
 CellID->665912149],
Cell[3845992, 101186, 287, 6, 27, "Input",ExpressionUUID->"a1b3f89f-50f6-4f39-b9ca-72eae35dadb7",
 CellID->1448203726],
Cell[3846282, 101194, 170, 3, 27, "Input",ExpressionUUID->"c9aef380-7a4b-48dd-83f6-f49589c44ecd",
 CellID->380367784]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[3846501, 101203, 454, 6, 39, "Subsection",ExpressionUUID->"5138cf00-354a-4b53-9b00-3a8ac2e7abeb",
 CellID->1994287226],
Cell[3846958, 101211, 578, 9, 40, "Text",ExpressionUUID->"0be1e37b-5ae4-4021-b3c8-102d7f6f7ed7",
 CellID->2143659166],
Cell[3847539, 101222, 1085, 16, 74, "Text",ExpressionUUID->"186862f9-109f-458f-ba35-ad1c244da9af",
 CellID->602637783],
Cell[CellGroupData[{
Cell[3848649, 101242, 269, 3, 41, "Subsubsection",ExpressionUUID->"ab5241d3-290b-48d8-ae82-3d56aeeb9aae",
 CellID->2036502373],
Cell[CellGroupData[{
Cell[3848943, 101249, 230, 3, 40, "MathCaption",ExpressionUUID->"1c4793f9-95a9-40e4-ad47-e9b60adc40a2",
 CellID->729583777],
Cell[3849176, 101254, 357, 8, 27, "Input",ExpressionUUID->"7d2f106e-9362-435c-9788-63a523c09c05",
 CellID->2016996031]
}, Open  ]],
Cell[CellGroupData[{
Cell[3849570, 101267, 203, 2, 40, "MathCaption",ExpressionUUID->"eec7e3b9-27e7-484d-ab41-d31f48475588",
 CellID->827825506],
Cell[3849776, 101271, 482, 12, 27, "Input",ExpressionUUID->"b0176ee2-8ecb-4d0c-be84-23e05886ebba",
 CellID->1747532039]
}, Open  ]],
Cell[CellGroupData[{
Cell[3850295, 101288, 772, 11, 72, "MathCaption",ExpressionUUID->"b8eb0a82-ed2c-42ef-a1f7-6356fe7d8abf",
 CellID->249981071],
Cell[CellGroupData[{
Cell[3851092, 101303, 937, 28, 27, "Input",ExpressionUUID->"b7b95916-e133-473e-8fc2-90c6f75a07ef",
 CellID->1520839681],
Cell[3852032, 101333, 1921, 53, 29, "Output",ExpressionUUID->"3d1aaf4e-ac79-44a1-b2dd-f13e29a5ab27",
 CellID->132570903]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3854002, 101392, 697, 10, 72, "MathCaption",ExpressionUUID->"1c8caed1-224b-4d6a-ac87-a09f3c66a102",
 CellID->737204209],
Cell[3854702, 101404, 375, 8, 27, "Input",ExpressionUUID->"57948b00-1fd7-4c1b-8a20-f8660164a6cc",
 CellID->766683321]
}, Open  ]],
Cell[CellGroupData[{
Cell[3855114, 101417, 244, 4, 40, "MathCaption",ExpressionUUID->"6d4133b7-fce6-4f23-8197-803a5e50f085",
 CellID->1533747171],
Cell[CellGroupData[{
Cell[3855383, 101425, 284, 7, 27, "Input",ExpressionUUID->"6344bee0-b29f-4509-b0ae-f232d2c31557",
 CellID->1451294931],
Cell[3855670, 101434, 5938, 157, 60, "Output",ExpressionUUID->"cbff5f4f-09b5-47a0-b7a9-ba40e6e161af",
 CellID->1339399285]
}, Open  ]],
Cell[CellGroupData[{
Cell[3861645, 101596, 267, 6, 27, "Input",ExpressionUUID->"004633b5-c551-42ee-a683-9681bd964c90",
 CellID->1652864636],
Cell[3861915, 101604, 3788, 103, 56, "Output",ExpressionUUID->"db26364b-fd67-4e37-9199-e5603d5aad17",
 CellID->1849594818]
}, Open  ]],
Cell[CellGroupData[{
Cell[3865740, 101712, 264, 6, 27, "Input",ExpressionUUID->"d9a13145-2320-4904-bac6-cb168c76d66b",
 CellID->890035918],
Cell[3866007, 101720, 3755, 102, 56, "Output",ExpressionUUID->"32800829-1542-4453-a32b-f22647bade22",
 CellID->260753550]
}, Open  ]],
Cell[CellGroupData[{
Cell[3869799, 101827, 247, 5, 27, "Input",ExpressionUUID->"8848ed1f-1495-4e8b-b8be-ea08dcc154f5",
 CellID->370060660],
Cell[3870049, 101834, 5908, 154, 70, "Output",ExpressionUUID->"bb2cefd1-dabb-400b-8a42-42734a95e4c8",
 CellID->1028464144]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3876006, 101994, 406, 7, 40, "MathCaption",ExpressionUUID->"2ed84eae-fb15-4418-a113-862a6985d2d7",
 CellID->84579354],
Cell[CellGroupData[{
Cell[3876437, 102005, 666, 18, 27, "Input",ExpressionUUID->"915d2044-a6aa-437f-aab1-563e7e674603",
 CellID->1416225960],
Cell[3877106, 102025, 1743, 49, 29, "Output",ExpressionUUID->"cf8722f0-9085-4d67-963d-b6a8133c3657",
 CellID->1183323500]
}, Open  ]],
Cell[CellGroupData[{
Cell[3878886, 102079, 220, 4, 27, "Input",ExpressionUUID->"e9691046-644d-46bb-994c-e70c2a0650f8",
 CellID->285499045],
Cell[3879109, 102085, 4559, 119, 50, "Output",ExpressionUUID->"dd689ca1-471b-4b89-ab11-3724c24aaa2c",
 CellID->128632176]
}, Open  ]],
Cell[CellGroupData[{
Cell[3883705, 102209, 240, 5, 27, "Input",ExpressionUUID->"3ebf67a1-a823-48cd-8d30-78d8f48b8ada",
 CellID->390796769],
Cell[3883948, 102216, 5927, 154, 54, "Output",ExpressionUUID->"9cd31672-443b-40b9-8c2e-0e435edc4021",
 CellID->107021857]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3889924, 102376, 242, 3, 40, "MathCaption",ExpressionUUID->"152a40b4-a216-4943-99e6-e73807e1856b",
 CellID->1518995414],
Cell[CellGroupData[{
Cell[3890191, 102383, 467, 12, 27, "Input",ExpressionUUID->"da2d510c-d6da-4d13-a4cc-7cd174ba7af1",
 CellID->665169177],
Cell[3890661, 102397, 2036, 63, 27, "Output",ExpressionUUID->"75cf4ea1-2fe8-43f4-8aed-4ad62d75e814",
 CellID->1731000397]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3892746, 102466, 191, 2, 40, "MathCaption",ExpressionUUID->"b84729a8-4341-402a-bffe-3e41f736ba6f",
 CellID->1863649492],
Cell[CellGroupData[{
Cell[3892962, 102472, 287, 7, 27, "Input",ExpressionUUID->"b6d7e278-2e44-48cf-b03a-fc0d72c41f97",
 CellID->678840868],
Cell[3893252, 102481, 1906, 62, 27, "Output",ExpressionUUID->"95679bc6-ba94-435f-a6cb-0b40ee5c6189",
 CellID->394812275]
}, Open  ]]
}, Open  ]],
Cell[3895185, 102547, 840, 13, 40, "Text",ExpressionUUID->"21851168-dd41-4f1a-9abf-b2f48e97d2af",
 CellID->1033628766],
Cell[CellGroupData[{
Cell[3896050, 102564, 626, 10, 56, "MathCaption",ExpressionUUID->"eed4577e-2413-4ac6-96f4-2a7c269787d1",
 CellID->1802171964],
Cell[CellGroupData[{
Cell[3896701, 102578, 657, 20, 27, "Input",ExpressionUUID->"debbc475-15ef-42f1-b3d4-c3311830fc82",
 CellID->1928346054],
Cell[3897361, 102600, 3203, 93, 32, "Output",ExpressionUUID->"3bdec94b-6278-43ab-9e48-caf34db68fee",
 CellID->413143220]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3900613, 102699, 395, 6, 40, "MathCaption",ExpressionUUID->"4353cd9e-3d30-454a-a14d-b97665ac64bc",
 CellID->127573517],
Cell[CellGroupData[{
Cell[3901033, 102709, 680, 21, 27, "Input",ExpressionUUID->"d62e71e2-d59f-4622-954c-fb94ec5af85d",
 CellID->1964538901],
Cell[3901716, 102732, 709, 10, 26, "Output",ExpressionUUID->"a777605e-3c74-489d-b8d3-b58067c80c98",
 CellID->1720729634]
}, Open  ]],
Cell[CellGroupData[{
Cell[3902462, 102747, 625, 20, 27, "Input",ExpressionUUID->"04700e20-ab5c-40f8-9943-47943e4d2e14",
 CellID->1465014692],
Cell[3903090, 102769, 826, 18, 27, "Output",ExpressionUUID->"6dc4405c-793d-498f-9855-de4a9b87ba39",
 CellID->943675289]
}, Open  ]],
Cell[CellGroupData[{
Cell[3903953, 102792, 676, 21, 27, "Input",ExpressionUUID->"54e603f0-ac3e-4d8b-a228-39ba133e7ef0",
 CellID->1935060583],
Cell[3904632, 102815, 1156, 32, 27, "Output",ExpressionUUID->"22ad6c51-b5e0-486c-a693-7ff3451a8af4",
 CellID->1577671198]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3905837, 102853, 194, 2, 40, "MathCaption",ExpressionUUID->"996e3394-2ed2-4d88-8699-13e27dfbd93b",
 CellID->185335001],
Cell[CellGroupData[{
Cell[3906056, 102859, 515, 15, 27, "Input",ExpressionUUID->"635ef241-ca06-4a62-a2d1-6e9f813ead18",
 CellID->405821759],
Cell[3906574, 102876, 1670, 47, 29, "Output",ExpressionUUID->"39ce3c4f-4599-41d1-8311-aac7d369a7e6",
 CellID->844695776]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3908293, 102929, 485, 8, 56, "MathCaption",ExpressionUUID->"9933235f-f632-4c83-a8b5-32d5e0f3ea96",
 CellID->2112124970],
Cell[CellGroupData[{
Cell[3908803, 102941, 1065, 29, 27, "Input",ExpressionUUID->"ba109330-948a-45a3-a783-e8f573b397a6",
 CellID->406172060],
Cell[CellGroupData[{
Cell[3909893, 102974, 2066, 49, 28, "Print",ExpressionUUID->"ad91bc7c-2055-48e7-8a73-0936814457a9",
 CellID->2024881592],
Cell[3911962, 103025, 2068, 49, 28, "Print",ExpressionUUID->"0cb826aa-9ec3-4931-89cb-df7c2d904e6b",
 CellID->1011562735],
Cell[3914033, 103076, 2068, 49, 28, "Print",ExpressionUUID->"84086254-e91b-4a6c-bfba-64339f00b534",
 CellID->1854337493],
Cell[3916104, 103127, 2066, 49, 28, "Print",ExpressionUUID->"f24a3335-0353-434b-9f54-6bda18b705ef",
 CellID->1567559587],
Cell[3918173, 103178, 2226, 56, 28, "Print",ExpressionUUID->"187ec5c3-f2b9-4e59-ae86-0f943c7433ac",
 CellID->644597313],
Cell[3920402, 103236, 2066, 49, 28, "Print",ExpressionUUID->"6873306b-40eb-4179-a646-210113ea8024",
 CellID->1839524123],
Cell[3922471, 103287, 2068, 49, 28, "Print",ExpressionUUID->"ff00565a-58f8-4c1f-b79c-059126372ae2",
 CellID->1707353049],
Cell[3924542, 103338, 2065, 49, 28, "Print",ExpressionUUID->"0ed60525-a3aa-493d-a881-dda77d11662e",
 CellID->806203329],
Cell[3926610, 103389, 2630, 70, 28, "Print",ExpressionUUID->"84cdad14-0b86-449f-bf2d-55efd727d4ba",
 CellID->1385021580],
Cell[3929243, 103461, 2066, 49, 28, "Print",ExpressionUUID->"5d04b2f5-26d7-437b-807f-7cb94099aa3c",
 CellID->1482445227],
Cell[3931312, 103512, 2065, 49, 28, "Print",ExpressionUUID->"de6e9089-74c2-43d6-a76f-3ff77ebb6a09",
 CellID->351657695],
Cell[3933380, 103563, 2067, 49, 28, "Print",ExpressionUUID->"c10f058f-d9c2-4c9f-b0d4-52906d77dcbe",
 CellID->435883321],
Cell[3935450, 103614, 2065, 49, 28, "Print",ExpressionUUID->"7ef9366c-74cb-4afb-ad51-91052f458719",
 CellID->824256130],
Cell[3937518, 103665, 2263, 57, 36, "Print",ExpressionUUID->"1031d50e-5544-4c4d-b672-d3459e393c21",
 CellID->2003253760],
Cell[3939784, 103724, 2065, 49, 28, "Print",ExpressionUUID->"d66658df-f2a8-432c-8385-232351acae78",
 CellID->437326654],
Cell[3941852, 103775, 2068, 49, 28, "Print",ExpressionUUID->"24fc12ea-5860-4987-9b20-5542ee0120e6",
 CellID->1460033744],
Cell[3943923, 103826, 2066, 49, 28, "Print",ExpressionUUID->"bfe157fd-3170-45e5-bedf-3d89258fa507",
 CellID->1638984786],
Cell[3945992, 103877, 2629, 71, 36, "Print",ExpressionUUID->"64c3e4d7-4568-4b0c-8c4a-874b61ee9b20",
 CellID->644558233],
Cell[3948624, 103950, 2542, 67, 45, "Print",ExpressionUUID->"42500356-87c9-47f3-aeb6-938114088b1e",
 CellID->1182706563]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3951227, 104024, 264, 3, 40, "MathCaption",ExpressionUUID->"7563648f-b089-47e4-a528-4e7254525b92",
 CellID->1259963770],
Cell[CellGroupData[{
Cell[3951516, 104031, 214, 5, 27, "Input",ExpressionUUID->"3155a783-7817-4e1f-8fd8-f0e6526c41fe",
 CellID->753808889],
Cell[3951733, 104038, 3521, 97, 32, "Output",ExpressionUUID->"8528a086-1541-4883-9280-3bfaee26a865",
 CellID->1537918424]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3955303, 104141, 206, 2, 40, "MathCaption",ExpressionUUID->"ec2a70d9-9349-433c-93f8-4b79b94814fa",
 CellID->1485286390],
Cell[CellGroupData[{
Cell[3955534, 104147, 316, 7, 27, "Input",ExpressionUUID->"f7944bd8-2a29-4141-8cb0-e639f6b113a9",
 CellID->1806190466],
Cell[3955853, 104156, 4222, 124, 49, "Output",ExpressionUUID->"0fa52a0f-9e55-48d0-9126-a1ccae189439",
 CellID->1917932673]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3960124, 104286, 289, 4, 40, "MathCaption",ExpressionUUID->"7ec9b2dd-0b11-4e62-b361-3f8eaeacf0f7",
 CellID->677715185],
Cell[CellGroupData[{
Cell[3960438, 104294, 570, 16, 27, "Input",ExpressionUUID->"193da3bd-e7ab-4026-991d-dc912cb3af7a",
 CellID->105850607],
Cell[3961011, 104312, 848, 17, 26, "Output",ExpressionUUID->"2aec1e78-ad08-4e24-97bc-3719d56ff96a",
 CellID->870818342]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3961920, 104336, 274, 3, 41, "Subsubsection",ExpressionUUID->"c62b6cab-7dd4-4353-87ab-ba8d1dd1789f",
 CellID->534635677],
Cell[3962197, 104341, 657, 11, 40, "Text",ExpressionUUID->"a52485ca-9153-4935-abaa-6d47e3509721",
 CellID->1463034371],
Cell[CellGroupData[{
Cell[3962879, 104356, 348, 6, 40, "MathCaption",ExpressionUUID->"7692223d-a03f-41c3-9a32-253413e57809",
 CellID->2072147979],
Cell[CellGroupData[{
Cell[3963252, 104366, 409, 11, 27, "Input",ExpressionUUID->"7603b50b-4c4f-42a7-80ac-96bff97e32c7",
 CellID->1120704634],
Cell[3963664, 104379, 3305, 94, 32, "Output",ExpressionUUID->"247755fe-57de-4030-8cbe-971bbc145808",
 CellID->1890332435]
}, Open  ]],
Cell[CellGroupData[{
Cell[3967006, 104478, 413, 11, 27, "Input",ExpressionUUID->"96aa78c9-de8a-4a1d-95fa-df83d7ea647b",
 CellID->514969174],
Cell[3967422, 104491, 3270, 94, 32, "Output",ExpressionUUID->"3f280083-365f-41bf-9c25-6ae53074a0cd",
 CellID->944161327]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3970741, 104591, 283, 5, 40, "MathCaption",ExpressionUUID->"ac126853-2c53-4967-b59d-5399d807c1dc",
 CellID->188737940],
Cell[CellGroupData[{
Cell[3971049, 104600, 463, 13, 27, "Input",ExpressionUUID->"46111333-18fd-46c9-8f2b-59858dd0b361",
 CellID->852783625],
Cell[3971515, 104615, 579, 10, 26, "Output",ExpressionUUID->"c0842d25-0d03-4aaf-bf6c-0629001d210a",
 CellID->762755206]
}, Open  ]],
Cell[CellGroupData[{
Cell[3972131, 104630, 461, 13, 27, "Input",ExpressionUUID->"5dcc9711-26f5-4d62-aa0a-d7659426ba1c",
 CellID->1028307225],
Cell[3972595, 104645, 571, 10, 26, "Output",ExpressionUUID->"06f442b3-c06a-40c4-9a38-93626c3532ec",
 CellID->1296858299]
}, Open  ]],
Cell[CellGroupData[{
Cell[3973203, 104660, 461, 13, 27, "Input",ExpressionUUID->"823ed00b-1ae5-49b8-84e5-bcee1a9c3533",
 CellID->442214286],
Cell[3973667, 104675, 577, 10, 26, "Output",ExpressionUUID->"f603c589-3bac-4987-8aeb-b5eb561e27c3",
 CellID->1485897890]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3974293, 104691, 521, 8, 56, "MathCaption",ExpressionUUID->"38bc8d8d-379c-4e84-8517-abc0a1e2583d",
 CellID->1437765897],
Cell[CellGroupData[{
Cell[3974839, 104703, 399, 11, 27, "Input",ExpressionUUID->"2ab1f190-fac7-4295-9568-457ca94cf296",
 CellID->814287199],
Cell[3975241, 104716, 1553, 40, 32, "Output",ExpressionUUID->"0301ad41-c026-46a3-838c-f11b0de30df9",
 CellID->398506267]
}, Open  ]],
Cell[CellGroupData[{
Cell[3976831, 104761, 400, 11, 27, "Input",ExpressionUUID->"c1624c3a-868e-427a-a4fe-b7325888d413",
 CellID->1701931036],
Cell[3977234, 104774, 544, 8, 26, "Output",ExpressionUUID->"413b145e-4344-46f8-aeae-4bc12fc268e3",
 CellID->1840818123]
}, Open  ]]
}, Open  ]],
Cell[3977805, 104786, 703, 11, 40, "Text",ExpressionUUID->"16c4582e-f8a8-473c-a89b-85d58ed21824",
 CellID->1812968013],
Cell[CellGroupData[{
Cell[3978533, 104801, 665, 11, 72, "MathCaption",ExpressionUUID->"691feddb-9532-4c95-9bab-5579f0af9930",
 CellID->777729093],
Cell[CellGroupData[{
Cell[3979223, 104816, 248, 5, 27, "Input",ExpressionUUID->"b7b7cdea-4576-4bba-83e3-6f9b38a1dd8d",
 CellID->1424946824],
Cell[3979474, 104823, 6668, 198, 94, "Output",ExpressionUUID->"22648db8-5422-4081-bb5d-ab98aab7219e",
 CellID->1980774579]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3986191, 105027, 252, 3, 40, "MathCaption",ExpressionUUID->"afbe6405-521b-49df-8548-01a678f13a87",
 CellID->21676925],
Cell[CellGroupData[{
Cell[3986468, 105034, 362, 10, 27, "Input",ExpressionUUID->"10765649-7433-42be-b578-3d63c7161d70",
 CellID->488081856],
Cell[3986833, 105046, 548, 8, 26, "Output",ExpressionUUID->"c4bcb560-c8eb-4f6b-96a2-174e583c6009",
 CellID->586853459]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3987430, 105060, 284, 3, 40, "MathCaption",ExpressionUUID->"9abbdc40-8a76-42c5-a934-c1dbf595e99c",
 CellID->119153639],
Cell[CellGroupData[{
Cell[3987739, 105067, 365, 10, 27, "Input",ExpressionUUID->"09349706-9595-4a24-ba21-be2f1a29dc3d",
 CellID->571661501],
Cell[3988107, 105079, 15558, 407, 140, "Output",ExpressionUUID->"41143670-3862-422a-9cf9-6bfeaed19d13",
 CellID->2001178389]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4003714, 105492, 364, 6, 40, "MathCaption",ExpressionUUID->"f37f562f-1f6f-4d88-beca-30f63168a663",
 CellID->1630978937],
Cell[CellGroupData[{
Cell[4004103, 105502, 225, 4, 27, "Input",ExpressionUUID->"c23ff10d-fccd-4ebc-9993-777a1f39338e",
 CellID->1724161398],
Cell[4004331, 105508, 567, 11, 43, "Output",ExpressionUUID->"754d3dfd-1d73-4fb4-bff6-57497e682cbb",
 CellID->2063788256]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[4004959, 105526, 306, 4, 41, "Subsubsection",ExpressionUUID->"2956bdc5-f871-4ae4-980e-ae38d5bb2ae8",
 CellID->381584480],
Cell[4005268, 105532, 647, 10, 23, "Text",ExpressionUUID->"0f69733b-c7c9-4c4a-8d5b-504581296894",
 CellID->32869574],
Cell[CellGroupData[{
Cell[4005940, 105546, 388, 6, 40, "MathCaption",ExpressionUUID->"3aa9a903-e224-4f1e-8442-00857a6deab9",
 CellID->1655986458],
Cell[4006331, 105554, 449, 10, 27, "Input",ExpressionUUID->"cf4b265b-8440-4deb-89d5-896fed8fdc6b",
 CellID->495600690]
}, Open  ]],
Cell[CellGroupData[{
Cell[4006817, 105569, 551, 9, 56, "MathCaption",ExpressionUUID->"de64321c-66df-4af6-bedc-3b65097e744a",
 CellID->1451400784],
Cell[4007371, 105580, 235, 4, 27, "Input",ExpressionUUID->"b09cbd94-4d34-4736-8d2a-abd74a34681d",
 CellID->146926218]
}, Open  ]],
Cell[CellGroupData[{
Cell[4007643, 105589, 265, 3, 40, "MathCaption",ExpressionUUID->"f8ae5881-f6df-4fee-952d-682939bbbf35",
 CellID->1907885024],
Cell[CellGroupData[{
Cell[4007933, 105596, 569, 14, 27, "Input",ExpressionUUID->"104563f0-119e-43e9-8d51-51fd6b7c8173",
 CellID->827776867],
Cell[4008505, 105612, 12165, 310, 70, "Output",ExpressionUUID->"baaf87b4-502f-49b8-80c1-7d5d3facbda5",
 CellID->251697358]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4020719, 105928, 356, 6, 56, "MathCaption",ExpressionUUID->"f9121996-fbba-4ca8-97ba-5cdfdeedd3ff",
 CellID->1850360014],
Cell[CellGroupData[{
Cell[4021100, 105938, 283, 5, 27, "Input",ExpressionUUID->"ed46ade5-da58-4f27-bf1b-d7bef1f01728",
 CellID->635491938],
Cell[4021386, 105945, 6713, 223, 70, "Output",ExpressionUUID->"2e203cea-3ec4-4e24-b617-8f063e97cdfa",
 CellID->1882194963]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4028148, 106174, 255, 5, 40, "MathCaption",ExpressionUUID->"328a59c0-7439-4fb3-9d5b-9b13a1068042",
 CellID->1020070558],
Cell[CellGroupData[{
Cell[4028428, 106183, 424, 10, 27, "Input",ExpressionUUID->"dfb05cc6-4cd1-4773-9620-c26f9f930078",
 CellID->1254913691],
Cell[4028855, 106195, 2015, 63, 70, "Output",ExpressionUUID->"c630915a-ca80-471d-a56d-93f03fa849e0",
 CellID->1616622831]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4030919, 106264, 296, 5, 40, "MathCaption",ExpressionUUID->"1a97b0d3-5783-43af-991a-c7fb12615af3",
 CellID->2063232277],
Cell[CellGroupData[{
Cell[4031240, 106273, 262, 5, 27, "Input",ExpressionUUID->"2b9795a6-739d-4cdb-9a86-122d65259d60",
 CellID->1477722874],
Cell[4031505, 106280, 6889, 228, 70, "Output",ExpressionUUID->"dcf65ab4-3b87-4d67-a78d-73e687e67938",
 CellID->616818773]
}, Open  ]],
Cell[CellGroupData[{
Cell[4038431, 106513, 416, 9, 27, "Input",ExpressionUUID->"5d102868-bf21-4b7a-a55b-d5dd76676e16",
 CellID->89864978],
Cell[4038850, 106524, 4252, 139, 70, "Output",ExpressionUUID->"f912ba0b-7ed4-413d-897c-7c2998fac8da",
 CellID->969553742]
}, Open  ]],
Cell[CellGroupData[{
Cell[4043139, 106668, 218, 4, 27, "Input",ExpressionUUID->"f9588d62-e91e-4584-ad61-75c239502537",
 CellID->19739716],
Cell[4043360, 106674, 11819, 303, 70, "Output",ExpressionUUID->"046e5e4b-1244-4537-9a94-529b971a1a05",
 CellID->183828358]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[4055240, 106984, 301, 4, 41, "Subsubsection",ExpressionUUID->"71d13d73-61a9-4519-b580-6c053d13b01a",
 CellID->672074522],
Cell[4055544, 106990, 621, 10, 23, "Text",ExpressionUUID->"f1d0a315-2f3a-4d52-a102-8126fc6d3f8c",
 CellID->281396522],
Cell[CellGroupData[{
Cell[4056190, 107004, 462, 8, 56, "MathCaption",ExpressionUUID->"c210bfeb-ade5-4273-bcef-735fbcb06102",
 CellID->756365217],
Cell[4056655, 107014, 521, 12, 27, "Input",ExpressionUUID->"06b848d7-d3b8-4cfb-9e4f-6e813b1f712d",
 CellID->31867511]
}, Open  ]],
Cell[CellGroupData[{
Cell[4057213, 107031, 276, 4, 40, "MathCaption",ExpressionUUID->"7e12d5a7-d34a-4a61-9928-9cc6cdd05759",
 CellID->1407314168],
Cell[CellGroupData[{
Cell[4057514, 107039, 169, 3, 27, "Input",ExpressionUUID->"c77794e0-8b97-49a8-b702-8ab080a4ccf0",
 CellID->243721534],
Cell[4057686, 107044, 1572, 47, 70, "Output",ExpressionUUID->"5e48e28f-1f46-4fc0-b08f-67b432e99f0e",
 CellID->1521728520]
}, Open  ]],
Cell[CellGroupData[{
Cell[4059295, 107096, 433, 11, 27, "Input",ExpressionUUID->"bb207e93-417f-47f6-8e64-7952805e2621",
 CellID->788203964],
Cell[4059731, 107109, 1506, 46, 70, "Output",ExpressionUUID->"77f6ea19-2061-4c86-b712-ca251bffe196",
 CellID->1308483517]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4061286, 107161, 243, 4, 40, "MathCaption",ExpressionUUID->"62eb2fe0-10db-4c69-b3a4-625b851148bc",
 CellID->254136468],
Cell[CellGroupData[{
Cell[4061554, 107169, 224, 4, 27, "Input",ExpressionUUID->"b7df79d4-3b7b-428e-8701-d9e9318dedd4",
 CellID->781127457],
Cell[4061781, 107175, 1447, 44, 70, "Output",ExpressionUUID->"ff9be134-f1e8-491c-a23b-3289bb85740b",
 CellID->1449924939]
}, Open  ]],
Cell[CellGroupData[{
Cell[4063265, 107224, 464, 12, 27, "Input",ExpressionUUID->"064403b3-aa61-42d1-967b-77903d7923bc",
 CellID->1238589112],
Cell[4063732, 107238, 1408, 44, 70, "Output",ExpressionUUID->"deeb048f-2f91-48fa-8b2d-4c3a007a37bd",
 CellID->1391507264]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4065189, 107288, 259, 5, 40, "MathCaption",ExpressionUUID->"8ee10e57-f5b8-4ebe-90d0-675ead531f57",
 CellID->45102520],
Cell[CellGroupData[{
Cell[4065473, 107297, 280, 6, 27, "Input",ExpressionUUID->"edbc718e-4806-43c3-8c1b-41d38c791bfb",
 CellID->1102713718],
Cell[4065756, 107305, 1979, 62, 70, "Output",ExpressionUUID->"aafedd50-7530-423c-84c2-546bdb9a426c",
 CellID->965670218]
}, Open  ]],
Cell[CellGroupData[{
Cell[4067772, 107372, 225, 4, 27, "Input",ExpressionUUID->"2880ade8-f584-4f9a-90dc-e28e834b9227",
 CellID->1028117429],
Cell[4068000, 107378, 1780, 59, 70, "Output",ExpressionUUID->"af5f4593-78cc-4bee-bfb3-e83fabae5e6e",
 CellID->1485433547]
}, Open  ]],
Cell[CellGroupData[{
Cell[4069817, 107442, 464, 12, 27, "Input",ExpressionUUID->"b5dd81f0-d39a-41c4-b833-18d1a167b4e8",
 CellID->1658101237],
Cell[4070284, 107456, 1455, 45, 70, "Output",ExpressionUUID->"937b3f4e-1316-4787-bda3-28a4627e551f",
 CellID->1184228054]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4071788, 107507, 290, 5, 40, "MathCaption",ExpressionUUID->"f420ea82-6a59-4953-9a02-81e11707d109",
 CellID->932149756],
Cell[CellGroupData[{
Cell[4072103, 107516, 214, 5, 27, "Input",ExpressionUUID->"901d0ffe-3c35-4380-b5e9-0227eb4cd761",
 CellID->1684957388],
Cell[4072320, 107523, 2004, 62, 70, "Output",ExpressionUUID->"41c4acaa-b00b-4d52-95d9-e61b6db98dc6",
 CellID->442463182]
}, Open  ]],
Cell[CellGroupData[{
Cell[4074361, 107590, 424, 11, 27, "Input",ExpressionUUID->"c7b873a3-8897-492e-9ab8-c4347073b017",
 CellID->1636349316],
Cell[4074788, 107603, 1487, 45, 70, "Output",ExpressionUUID->"d72a018c-6bc2-4764-880c-56a127e565c5",
 CellID->1890313960]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4076324, 107654, 322, 6, 40, "MathCaption",ExpressionUUID->"a0784977-f726-4d10-8cce-8b6e52d48f0a",
 CellID->976614439],
Cell[CellGroupData[{
Cell[4076671, 107664, 290, 6, 27, "Input",ExpressionUUID->"8d9675b8-5240-4269-97bd-329640ca2c56",
 CellID->1763483765],
Cell[4076964, 107672, 5580, 149, 70, "Output",ExpressionUUID->"c4b98d51-2a95-4e37-ba81-2e6e054dc534",
 CellID->633652002]
}, Open  ]],
Cell[CellGroupData[{
Cell[4082581, 107826, 376, 9, 27, "Input",ExpressionUUID->"7b5e6ba2-3e67-442a-b6d5-1164c0302750",
 CellID->1364973140],
Cell[4082960, 107837, 1293, 43, 70, "Output",ExpressionUUID->"7e42e7b4-ca62-4f0a-8ae1-f01b290f609e",
 CellID->417792141]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[4084314, 107887, 301, 4, 41, "Subsubsection",ExpressionUUID->"d62a8fa2-188c-430c-871b-1f08c826b625",
 CellID->1060734641],
Cell[4084618, 107893, 822, 13, 23, "Text",ExpressionUUID->"e20fb4d6-4d42-4ff4-af8f-715fed87e939",
 CellID->558804793],
Cell[CellGroupData[{
Cell[4085465, 107910, 372, 6, 40, "MathCaption",ExpressionUUID->"e78b9c5a-df05-41c3-9f71-7866b8b6ea31",
 CellID->289463677],
Cell[CellGroupData[{
Cell[4085862, 107920, 378, 9, 27, "Input",ExpressionUUID->"46d584f0-0fca-47ff-b3ca-dc1a04344119",
 CellID->229469696],
Cell[4086243, 107931, 5529, 171, 70, "Output",ExpressionUUID->"33d522ac-199d-4905-8b12-26e0df45df7b",
 CellID->1708471744]
}, Open  ]],
Cell[CellGroupData[{
Cell[4091809, 108107, 396, 10, 27, "Input",ExpressionUUID->"49300fc6-f3ac-48c0-bcd2-863007062ccb",
 CellID->500436437],
Cell[4092208, 108119, 1575, 43, 70, "Output",ExpressionUUID->"f487881d-1261-472f-b8d6-9671a6713533",
 CellID->280757371]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4093832, 108168, 288, 5, 40, "MathCaption",ExpressionUUID->"b53f5d01-5a32-4151-ab56-33ce1fc850fb",
 CellID->1431860481],
Cell[CellGroupData[{
Cell[4094145, 108177, 398, 11, 27, "Input",ExpressionUUID->"5f8c272b-757d-4135-925f-8cc71865db3d",
 CellID->156844470],
Cell[4094546, 108190, 221, 4, 70, "Output",ExpressionUUID->"39250d6d-4bd5-4d0f-8f44-5f4d45c2f80e",
 CellID->667561938]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4094816, 108200, 667, 10, 56, "MathCaption",ExpressionUUID->"dc238619-6cab-49da-8d0b-94919658b8cf",
 CellID->1790938121],
Cell[4095486, 108212, 666, 18, 27, "Input",ExpressionUUID->"9ffe7d13-e6e7-4505-8c91-344331278eea",
 CellID->1098684169]
}, Open  ]],
Cell[CellGroupData[{
Cell[4096189, 108235, 191, 2, 40, "MathCaption",ExpressionUUID->"9635fec6-88e8-4225-a206-714d6206d967",
 CellID->1773697533],
Cell[CellGroupData[{
Cell[4096405, 108241, 411, 10, 27, "Input",ExpressionUUID->"db5a8743-b3ea-44d2-9e8c-b367b914e825",
 CellID->1882649907],
Cell[4096819, 108253, 5619, 148, 70, "Output",ExpressionUUID->"7af7eadb-3546-4638-911d-a584dbc56e3a",
 CellID->1258920038]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4102487, 108407, 340, 6, 56, "MathCaption",ExpressionUUID->"845d3065-c7b6-4828-a25f-8c7fbb273271",
 CellID->2040425438],
Cell[CellGroupData[{
Cell[4102852, 108417, 313, 7, 27, "Input",ExpressionUUID->"7353ae65-3b36-43c4-a041-c8a273d4b433",
 CellID->1551436921],
Cell[4103168, 108426, 1766, 59, 70, "Output",ExpressionUUID->"66e20a52-ad1f-484b-aeb5-fa803027db1b",
 CellID->1660188422]
}, Open  ]],
Cell[CellGroupData[{
Cell[4104971, 108490, 426, 11, 27, "Input",ExpressionUUID->"3b900027-ac68-42fa-bb95-009c0110c12a",
 CellID->611133573],
Cell[4105400, 108503, 2072, 71, 70, "Output",ExpressionUUID->"9c10164e-b3ac-4d60-97cc-5057f02d1ba0",
 CellID->531783083]
}, Open  ]],
Cell[CellGroupData[{
Cell[4107509, 108579, 280, 6, 27, "Input",ExpressionUUID->"f775012f-d390-4249-968c-eea04d5c0b8e",
 CellID->1078768095],
Cell[4107792, 108587, 2240, 73, 70, "Output",ExpressionUUID->"58380817-3aa3-40cd-b6e9-b093f1123c1c",
 CellID->1998820814]
}, Open  ]],
Cell[CellGroupData[{
Cell[4110069, 108665, 308, 7, 27, "Input",ExpressionUUID->"f881812f-144c-40e2-8b93-2d5ed7019e82",
 CellID->1559514164],
Cell[4110380, 108674, 5573, 147, 70, "Output",ExpressionUUID->"e37bc666-85b2-4b13-b576-e6d3d9776675",
 CellID->1094730877]
}, Open  ]]
}, Open  ]],
Cell[4115980, 108825, 1204, 22, 88, "MathCaption",ExpressionUUID->"cbfb09a4-f913-44ff-a2f9-f6479a1c17df",
 CellID->198307171]
}, Closed]],
Cell[CellGroupData[{
Cell[4117221, 108852, 354, 5, 41, "Subsubsection",ExpressionUUID->"e3e725b5-3945-4f84-98b4-2b59b83a4ebc",
 CellID->916191242],
Cell[4117578, 108859, 838, 13, 23, "Text",ExpressionUUID->"b1d67a70-9242-4c8b-b348-161bced1e008",
 CellID->968455304],
Cell[CellGroupData[{
Cell[4118441, 108876, 373, 6, 40, "MathCaption",ExpressionUUID->"8656ce63-b738-48b8-920a-654df330f3b9",
 CellID->1049733715],
Cell[4118817, 108884, 682, 17, 27, "Input",ExpressionUUID->"0b3ab37d-f8b0-4004-b4a1-98a9df45e43c",
 CellID->1296387900],
Cell[CellGroupData[{
Cell[4119524, 108905, 447, 12, 27, "Input",ExpressionUUID->"6d754ff9-6311-428a-a70e-ecc0503dc402",
 CellID->83753186],
Cell[4119974, 108919, 5916, 218, 70, "Output",ExpressionUUID->"641b3c7d-b4a8-49bb-8907-1e3e0117aa93",
 CellID->1649445890]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4125939, 109143, 289, 5, 40, "MathCaption",ExpressionUUID->"1f03adee-8230-49ba-b565-b25fe89b3504",
 CellID->1964110742],
Cell[CellGroupData[{
Cell[4126253, 109152, 627, 17, 27, "Input",ExpressionUUID->"bb1bbd52-c764-42cb-bb91-0f4029e795be",
 CellID->1499017740],
Cell[4126883, 109171, 5916, 218, 70, "Output",ExpressionUUID->"d8d6b0ee-39d8-4210-8028-730405491035",
 CellID->370674107]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[4132860, 109396, 405, 5, 41, "Subsubsection",ExpressionUUID->"197e099a-255c-47bb-9218-1bc6b25bba11",
 CellID->868674767],
Cell[4133268, 109403, 1014, 15, 40, "Text",ExpressionUUID->"b4985408-6528-42e8-9a37-1bb897c6961b",
 CellID->1222976663],
Cell[CellGroupData[{
Cell[4134307, 109422, 326, 5, 40, "MathCaption",ExpressionUUID->"0aada75f-aca7-4a33-990c-5c7b6c2e0d9b",
 CellID->1002789604],
Cell[CellGroupData[{
Cell[4134658, 109431, 409, 9, 27, "Input",ExpressionUUID->"f2ce655b-f526-49d8-8e4a-0d2a7607bad4",
 CellID->587006693],
Cell[4135070, 109442, 14613, 376, 70, "Output",ExpressionUUID->"e1887f5e-0c8e-4a76-a52d-32211774579c",
 CellID->69656402]
}, Open  ]]
}, Open  ]],
Cell[4149710, 109822, 494, 8, 56, "MathCaption",ExpressionUUID->"ed9812dc-50b3-479c-9514-9342ec91209f",
 CellID->2134094243],
Cell[CellGroupData[{
Cell[4150229, 109834, 270, 5, 40, "MathCaption",ExpressionUUID->"f643908d-c63f-4d63-9eb3-b5a5906336ff",
 CellID->1767026200],
Cell[CellGroupData[{
Cell[4150524, 109843, 468, 11, 27, "Input",ExpressionUUID->"45e4d2ae-bc32-424d-be98-2ab81a55be39",
 CellID->87906517],
Cell[4150995, 109856, 6427, 215, 70, "Output",ExpressionUUID->"06998c2a-9c0b-46c2-a77a-f102cf2e0c3c",
 CellID->336560799]
}, Open  ]],
Cell[CellGroupData[{
Cell[4157459, 110076, 308, 7, 27, "Input",ExpressionUUID->"a134847e-2eb4-4d78-90ee-5fbbb8bcb2b6",
 CellID->649106798],
Cell[4157770, 110085, 1586, 56, 70, "Output",ExpressionUUID->"9a4fce62-a3aa-4369-a506-d3bf25b0cb8d",
 CellID->105422595]
}, Open  ]],
Cell[CellGroupData[{
Cell[4159393, 110146, 463, 12, 27, "Input",ExpressionUUID->"494774e5-1a40-4baf-b655-c54b43c60397",
 CellID->1707919238],
Cell[4159859, 110160, 5842, 210, 70, "Output",ExpressionUUID->"a0bd25ce-a61f-4403-9db1-e5aa7c37a9c8",
 CellID->163545390]
}, Open  ]],
Cell[CellGroupData[{
Cell[4165738, 110375, 252, 5, 27, "Input",ExpressionUUID->"2614020d-9d36-4bab-b5ec-6a281421e3a8",
 CellID->1608572934],
Cell[4165993, 110382, 324, 5, 70, "Output",ExpressionUUID->"1eba6922-7b2b-40e9-86ae-aec13edc7133",
 CellID->2075785217]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4166366, 110393, 478, 8, 56, "MathCaption",ExpressionUUID->"fdd1e62c-efa7-47b1-9535-17c0c7f29a8e",
 CellID->2129093082],
Cell[CellGroupData[{
Cell[4166869, 110405, 201, 3, 27, "Input",ExpressionUUID->"2d209ba0-ff36-4574-95ce-8518dfc5114b",
 CellID->1096420752],
Cell[4167073, 110410, 14561, 375, 70, "Output",ExpressionUUID->"88a54bea-9175-48be-af79-d5e996e723f8",
 CellID->1850296604]
}, Open  ]],
Cell[CellGroupData[{
Cell[4181671, 110790, 392, 10, 27, "Input",ExpressionUUID->"74170c86-1985-4f4f-8cbf-1c0c9da5181d",
 CellID->34089487],
Cell[4182066, 110802, 8400, 216, 70, "Output",ExpressionUUID->"10d20048-385f-446f-9bbc-69e20d7a7671",
 CellID->224331221]
}, Open  ]],
Cell[CellGroupData[{
Cell[4190503, 111023, 396, 10, 27, "Input",ExpressionUUID->"ea579ab8-e31d-4391-b62a-c5e7b28597cf",
 CellID->667712078],
Cell[4190902, 111035, 8719, 228, 70, "Output",ExpressionUUID->"4000a53f-3fa6-4743-aefb-7dba51d8b84d",
 CellID->1501030862]
}, Open  ]]
}, Open  ]],
Cell[4199648, 111267, 369, 6, 56, "MathCaption",ExpressionUUID->"34796ba6-6c17-464d-b9cc-7b694288572e",
 CellID->1029605588],
Cell[CellGroupData[{
Cell[4200042, 111277, 401, 7, 56, "MathCaption",ExpressionUUID->"8bbaa18c-c261-4315-8986-b48094842e57",
 CellID->144830439],
Cell[CellGroupData[{
Cell[4200468, 111288, 377, 9, 27, "Input",ExpressionUUID->"f881ee5d-1fc4-4dd6-84cd-399679570011",
 CellID->1247079980],
Cell[4200848, 111299, 5895, 211, 70, "Output",ExpressionUUID->"c53827de-e31d-4533-95ed-b5244f84ab0c",
 CellID->1335296325]
}, Open  ]],
Cell[CellGroupData[{
Cell[4206780, 111515, 463, 12, 27, "Input",ExpressionUUID->"a02e23d7-2d71-4a4c-bf55-3bd01e04e89f",
 CellID->873950029],
Cell[4207246, 111529, 2721, 96, 70, "Output",ExpressionUUID->"a4d8f1b7-64a9-43b5-b305-68f311409c04",
 CellID->1121223125]
}, Open  ]],
Cell[CellGroupData[{
Cell[4210004, 111630, 403, 10, 27, "Input",ExpressionUUID->"3efe70c2-cb00-4ec5-a6fe-075c3f34b4b5",
 CellID->1283630650],
Cell[4210410, 111642, 8689, 227, 70, "Output",ExpressionUUID->"8e760cb9-8a10-4785-9c92-ad43eb4f767b",
 CellID->228059450]
}, Open  ]],
Cell[CellGroupData[{
Cell[4219136, 111874, 224, 4, 27, "Input",ExpressionUUID->"52d0ad8a-9745-410b-9a6c-163c139d44c4",
 CellID->615280195],
Cell[4219363, 111880, 318, 5, 70, "Output",ExpressionUUID->"1f663a81-1369-4873-8096-acc6ca5343b2",
 CellID->1884349902]
}, Open  ]]
}, Open  ]],
Cell[4219708, 111889, 470, 7, 56, "MathCaption",ExpressionUUID->"19016ad7-f785-4b92-b322-2990e85c46e6",
 CellID->1768506554],
Cell[4220181, 111898, 1444, 21, 57, "Text",ExpressionUUID->"4e48f271-2919-441b-8535-b176aa6b5ac5",
 CellID->1520660747],
Cell[CellGroupData[{
Cell[4221650, 111923, 352, 6, 40, "MathCaption",ExpressionUUID->"f21e8996-27e2-4e3e-8478-6a467598122e",
 CellID->1481543602],
Cell[CellGroupData[{
Cell[4222027, 111933, 264, 5, 27, "Input",ExpressionUUID->"4e3846cd-e080-4b3a-8525-cdaa49fde47b",
 CellID->283055898],
Cell[4222294, 111940, 3995, 127, 70, "Output",ExpressionUUID->"01de2fd7-23ad-4847-9554-b248db23c76f",
 CellID->1327460605]
}, Open  ]],
Cell[CellGroupData[{
Cell[4226326, 112072, 345, 7, 27, "Input",ExpressionUUID->"e583d3e2-6d39-464e-8029-dd06c0f8a273",
 CellID->1809987634],
Cell[4226674, 112081, 9056, 288, 70, "Output",ExpressionUUID->"96288a4f-f4ac-482b-847b-c0d90d3afdcc",
 CellID->422779552]
}, Open  ]],
Cell[CellGroupData[{
Cell[4235767, 112374, 258, 5, 27, "Input",ExpressionUUID->"8fbc87ab-1129-4a33-9538-45e83b8240f9",
 CellID->768461419],
Cell[4236028, 112381, 1248, 42, 70, "Output",ExpressionUUID->"133e5b2b-2c02-48b1-a3a5-71856b239c5f",
 CellID->1780026188]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4237325, 112429, 528, 9, 56, "MathCaption",ExpressionUUID->"b9c1dfac-767d-44a0-92eb-728847748b36",
 CellID->2016765644],
Cell[CellGroupData[{
Cell[4237878, 112442, 194, 4, 27, "Input",ExpressionUUID->"9e985458-b242-4b02-a1e0-4d172bad7e00",
 CellID->132922870],
Cell[4238075, 112448, 4013, 127, 70, "Output",ExpressionUUID->"39f28d6b-dec6-4812-b2c2-17d8102f8f39",
 CellID->305455359]
}, Open  ]],
Cell[CellGroupData[{
Cell[4242125, 112580, 285, 6, 27, "Input",ExpressionUUID->"5ac457bf-7577-446a-8ad9-fe8c33b5950d",
 CellID->512812450],
Cell[4242413, 112588, 20041, 636, 70, "Output",ExpressionUUID->"27c221b7-2297-4c65-9feb-885259204401",
 CellID->1302302383]
}, Open  ]],
Cell[CellGroupData[{
Cell[4262491, 113229, 223, 4, 27, "Input",ExpressionUUID->"55c91845-d52a-41e1-9be2-4c8935bf6fee",
 CellID->809087849],
Cell[4262717, 113235, 4281, 140, 70, "Output",ExpressionUUID->"fdd7d099-b8c0-4b9b-8d93-b70acd7ed5b9",
 CellID->642820857]
}, Open  ]],
Cell[CellGroupData[{
Cell[4267035, 113380, 398, 10, 27, "Input",ExpressionUUID->"35d66f65-29ac-46b2-83df-4f3e7485bcb0",
 CellID->344828528],
Cell[4267436, 113392, 6658, 222, 70, "Output",ExpressionUUID->"04cb0dc3-19ad-44a0-9a86-1f25afffe773",
 CellID->2047399189]
}, Open  ]],
Cell[CellGroupData[{
Cell[4274131, 113619, 225, 4, 27, "Input",ExpressionUUID->"ddddc421-40e2-4a7e-b085-1e84dd79e5f2",
 CellID->474557054],
Cell[4274359, 113625, 6582, 221, 70, "Output",ExpressionUUID->"f0ec0301-19ff-4545-870a-390bff39195c",
 CellID->1507693642]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4280990, 113852, 378, 6, 56, "MathCaption",ExpressionUUID->"931179a4-0175-4e14-a857-76cde9f72101",
 CellID->1144158072],
Cell[CellGroupData[{
Cell[4281393, 113862, 562, 14, 27, "Input",ExpressionUUID->"8ed4657e-bc5b-4e64-a84f-90bad2e79c1c",
 CellID->832807879],
Cell[4281958, 113878, 1268, 42, 70, "Output",ExpressionUUID->"2bc55ad7-4073-4dc7-ad8e-9ee546ef0c02",
 CellID->1647490141]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4283275, 113926, 323, 6, 40, "MathCaption",ExpressionUUID->"ab262ee2-26eb-4370-8ffb-a89ec321d920",
 CellID->699828843],
Cell[4283601, 113934, 329, 7, 27, "Input",ExpressionUUID->"014e6b74-ce7c-45c9-a9c0-1a38a1df7f2e",
 CellID->379832614]
}, Open  ]],
Cell[CellGroupData[{
Cell[4283967, 113946, 239, 4, 40, "MathCaption",ExpressionUUID->"9800838a-5e9a-48bc-94bb-8613e0b9442d",
 CellID->1890230017],
Cell[CellGroupData[{
Cell[4284231, 113954, 327, 7, 27, "Input",ExpressionUUID->"883b194d-4ab5-474b-992d-b5c9543267e2",
 CellID->2109643964],
Cell[4284561, 113963, 7586, 193, 70, "Output",ExpressionUUID->"0be61e83-8f01-4320-aeae-3bc501415e82",
 CellID->1860139016]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4292196, 114162, 421, 7, 56, "MathCaption",ExpressionUUID->"7fc90934-e112-4664-b74a-637498525287",
 CellID->457891937],
Cell[CellGroupData[{
Cell[4292642, 114173, 424, 11, 27, "Input",ExpressionUUID->"3fd95a17-be19-4e10-a212-82d0142d56b0",
 CellID->2117997876],
Cell[4293069, 114186, 11828, 305, 70, "Output",ExpressionUUID->"487d157d-40ae-4ee8-96b6-73ff26788fd7",
 CellID->289508886]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[4304958, 114498, 472, 6, 41, "Subsubsection",ExpressionUUID->"b6f6d361-19ff-4dd8-80e7-03d4b0e2ed85",
 CellID->1440932172],
Cell[CellGroupData[{
Cell[4305455, 114508, 174, 2, 40, "MathCaption",ExpressionUUID->"f39c67df-5e2c-4e69-a3c1-3da363e73119",
 CellID->1946333281],
Cell[4305632, 114512, 732, 12, 27, "Input",ExpressionUUID->"f6e48f61-1a48-410e-8cfa-d2769f35edce",
 CellID->1143769821],
Cell[CellGroupData[{
Cell[4306389, 114528, 457, 9, 27, "Input",ExpressionUUID->"5a5c565c-f24a-4dbc-8e3e-da65f302f5dc",
 CellID->781951820],
Cell[4306849, 114539, 426, 8, 26, "Output",ExpressionUUID->"05141944-9eab-4bce-bdc8-324a17686a2e",
 CellID->181151293]
}, Open  ]],
Cell[4307290, 114550, 232, 4, 27, "Input",ExpressionUUID->"4538f408-1f7a-466c-bde6-9db38392d719",
 CellID->745731941],
Cell[4307525, 114556, 231, 4, 27, "Input",ExpressionUUID->"9b0a6170-8ba0-4499-bec7-d3a1a544c6b5",
 CellID->765596817],
Cell[4307759, 114562, 236, 4, 27, "Input",ExpressionUUID->"b6a1fd3c-4753-4546-b6e6-efccef3120ce",
 CellID->1101425542]
}, Open  ]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[4308056, 114573, 263, 3, 39, "Subsection",ExpressionUUID->"b5b1cc60-2f5e-4b76-b32d-092d58898b8a",
 CellID->1930012216],
Cell[4308322, 114578, 361, 6, 40, "Text",ExpressionUUID->"727cb295-9f98-4c56-8067-eee3787c03e4",
 CellID->2122310024],
Cell[CellGroupData[{
Cell[4308708, 114588, 166, 3, 23, "Text",ExpressionUUID->"a3efdba3-5116-4530-bd04-7f0ba8628f0d",
 CellID->2108680345],
Cell[4308877, 114593, 249, 5, 27, "Input",ExpressionUUID->"24a693e7-b373-4d6e-b250-f5de3f9c21e7",
 Evaluatable->False,
 CellID->667931974]
}, Open  ]],
Cell[CellGroupData[{
Cell[4309163, 114603, 227, 4, 23, "Text",ExpressionUUID->"3a5f7de8-aef3-4036-b978-11940f5ad11b",
 CellID->985426970],
Cell[4309393, 114609, 292, 6, 27, "Input",ExpressionUUID->"cc65d2e4-bb59-4bb6-9858-3f3d687beb92",
 Evaluatable->False,
 CellID->677199126]
}, Open  ]],
Cell[CellGroupData[{
Cell[4309722, 114620, 146, 1, 23, "Text",ExpressionUUID->"94dcf799-c8ae-4b70-99e9-8b18740130da",
 CellID->1035664149],
Cell[4309871, 114623, 234, 5, 27, "Input",ExpressionUUID->"2c557be7-e801-4864-a6e0-ba379c140888",
 Evaluatable->False,
 CellID->1052393308]
}, Open  ]],
Cell[CellGroupData[{
Cell[4310142, 114633, 232, 3, 41, "Subsubsection",ExpressionUUID->"7634dbcd-b0e0-4524-8be6-4034d738f711",
 CellID->899010864],
Cell[4310377, 114638, 177, 3, 27, "Input",ExpressionUUID->"b94284bc-43fd-40ff-8db9-19980a3e85d1",
 CellID->2128134603],
Cell[4310557, 114643, 365, 8, 47, "Input",ExpressionUUID->"8558780a-d670-4277-adff-d566c1f941c6",
 CellID->1218131836],
Cell[CellGroupData[{
Cell[4310947, 114655, 120, 1, 40, "MathCaption",ExpressionUUID->"9188e563-835e-4554-9a4b-1978a2be6e79",
 CellID->1180160801],
Cell[4311070, 114658, 289, 7, 27, "Input",ExpressionUUID->"40db2ab2-e942-4ed7-bb3d-2e58ae783d7e",
 CellID->803618715]
}, Open  ]],
Cell[CellGroupData[{
Cell[4311396, 114670, 126, 1, 40, "MathCaption",ExpressionUUID->"86766bd2-e690-47e6-888e-a512e4609436",
 CellID->895087022],
Cell[4311525, 114673, 459, 12, 27, "Input",ExpressionUUID->"c7efe24b-10de-46f2-8d47-a655c9be2d6d",
 CellID->604631519],
Cell[4311987, 114687, 536, 13, 66, "Input",ExpressionUUID->"4ee1c6bf-c88e-449a-8795-e5b4b80de214",
 CellID->149322229]
}, Open  ]],
Cell[CellGroupData[{
Cell[4312560, 114705, 115, 1, 40, "MathCaption",ExpressionUUID->"a9c05320-0e15-4849-8f9a-04ea88d4ce34",
 CellID->1397803107],
Cell[4312678, 114708, 400, 13, 27, "Input",ExpressionUUID->"f533c991-bdca-4306-9cc7-a7d93e1d2f82",
 CellID->1553204816],
Cell[CellGroupData[{
Cell[4313103, 114725, 840, 24, 45, "Input",ExpressionUUID->"d408439f-f35c-49f9-b66b-7ad4a2ffe4fa",
 CellID->2049613227],
Cell[4313946, 114751, 3622, 120, 70, "Output",ExpressionUUID->"5359e9bb-242b-4149-8278-647ecc4c5cb3",
 CellID->624138270]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4317617, 114877, 221, 4, 40, "MathCaption",ExpressionUUID->"6a666914-ba0b-47eb-ad52-575d5c224e6d",
 CellID->920319793],
Cell[4317841, 114883, 601, 16, 27, "Input",ExpressionUUID->"06123832-3ea0-4154-ab10-473165123e31",
 CellID->1637535257]
}, Open  ]],
Cell[CellGroupData[{
Cell[4318479, 114904, 141, 1, 40, "MathCaption",ExpressionUUID->"1702e0ea-2c80-4c27-b274-dc076bd78b4d",
 CellID->2052128094],
Cell[CellGroupData[{
Cell[4318645, 114909, 265, 7, 27, "Input",ExpressionUUID->"c64995bc-d0d4-40c9-a992-ed36cc8d994a",
 CellID->1529505038],
Cell[4318913, 114918, 3976, 126, 70, "Output",ExpressionUUID->"5b372aeb-c6eb-40ef-8136-ec36ad371c0b",
 CellID->1581771942]
}, Open  ]],
Cell[CellGroupData[{
Cell[4322926, 115049, 303, 9, 27, "Input",ExpressionUUID->"0fac8e23-5b34-47c3-826c-603dd175fa61",
 CellID->1400909739],
Cell[4323232, 115060, 4698, 146, 70, "Output",ExpressionUUID->"a8d15157-8b0a-46cc-9b33-f3ca3bb4a50e",
 CellID->1140962981]
}, Open  ]],
Cell[CellGroupData[{
Cell[4327967, 115211, 389, 10, 27, "Input",ExpressionUUID->"5d7fb1a6-5f20-4cdb-836f-a82c6f3ac1d6",
 CellID->979854515],
Cell[4328359, 115223, 3101, 98, 70, "Output",ExpressionUUID->"6321522a-091f-4d8c-9a04-97fb602bb294",
 CellID->1283337604]
}, Open  ]],
Cell[4331475, 115324, 221, 5, 27, "Input",ExpressionUUID->"d7336ef0-98c6-45dd-ae9c-48284018320f",
 CellID->1299646170]
}, Open  ]],
Cell[CellGroupData[{
Cell[4331733, 115334, 143, 1, 40, "MathCaption",ExpressionUUID->"e7adb246-3135-4f7b-95f5-eceeeaca9295",
 CellID->1097005553],
Cell[4331879, 115337, 594, 19, 46, "Input",ExpressionUUID->"8ccd7f82-bf2f-40a8-9380-5c6055b218ed",
 CellID->1225219776],
Cell[CellGroupData[{
Cell[4332498, 115360, 650, 19, 27, "Input",ExpressionUUID->"0a72d98e-bcdb-4da8-8451-1bbe859e2609",
 CellID->48084149],
Cell[4333151, 115381, 8252, 255, 70, "Output",ExpressionUUID->"3f4bea74-3b99-4c30-b951-fc6c08561543",
 CellID->1876843607]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4341452, 115642, 143, 1, 40, "MathCaption",ExpressionUUID->"0297881f-8a84-4855-9dd0-d1fa2e3269dd",
 CellID->485428197],
Cell[CellGroupData[{
Cell[4341620, 115647, 283, 8, 27, "Input",ExpressionUUID->"2a2065f9-91df-4524-8380-c2903a15cded",
 CellID->301332026],
Cell[4341906, 115657, 17554, 544, 70, "Output",ExpressionUUID->"4bc6fb78-8be9-4107-8f55-0d2f9f4436ea",
 CellID->1870695713]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4359509, 116207, 124, 1, 40, "MathCaption",ExpressionUUID->"02838182-e322-4598-a333-70aa676cd105",
 CellID->422493918],
Cell[CellGroupData[{
Cell[4359658, 116212, 303, 7, 27, "Input",ExpressionUUID->"43968ed5-a8b3-46d7-9c69-f4c3228e5040",
 CellID->1274342844],
Cell[4359964, 116221, 368, 7, 70, "Output",ExpressionUUID->"b6675c96-fe31-47bb-a9fe-aed6a1fb8011",
 CellID->809436665]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4360381, 116234, 390, 6, 56, "MathCaption",ExpressionUUID->"78bd1f94-0fa3-488a-b61c-c2dc2b57a9dc",
 CellID->76948523],
Cell[CellGroupData[{
Cell[4360796, 116244, 248, 6, 27, "Input",ExpressionUUID->"56557837-f24b-40dc-b46a-9c8adaf3a33b",
 CellID->488670567],
Cell[4361047, 116252, 369, 7, 70, "Output",ExpressionUUID->"30477f7e-c077-4cae-b28c-ee891b3dd8d4",
 CellID->2040608557]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4361465, 116265, 119, 1, 40, "MathCaption",ExpressionUUID->"7ce0fdab-7593-4c15-b365-169771325d38",
 CellID->1103477868],
Cell[4361587, 116268, 213, 5, 27, "Input",ExpressionUUID->"599400d9-4807-4db7-8216-8604ba0a629e",
 CellID->483751984],
Cell[CellGroupData[{
Cell[4361825, 116277, 187, 4, 27, "Input",ExpressionUUID->"22f358ee-7b98-443a-ad1e-4cf3ad95159c",
 CellID->46507546],
Cell[4362015, 116283, 313, 5, 70, "Output",ExpressionUUID->"e179ed83-1b1d-45c2-97c6-4555e61d4a6d",
 CellID->1194174909]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4362377, 116294, 148, 1, 40, "MathCaption",ExpressionUUID->"d6570d23-67c2-438b-8a5d-0da1be68683d",
 CellID->1573690101],
Cell[CellGroupData[{
Cell[4362550, 116299, 159, 3, 27, "Input",ExpressionUUID->"23542a43-1848-4492-93a6-574608833559",
 CellID->600931055],
Cell[4362712, 116304, 10053, 308, 70, "Output",ExpressionUUID->"8b14c1b5-73a7-4a06-84f9-d7ec4f0ea222",
 CellID->115530701]
}, Open  ]],
Cell[CellGroupData[{
Cell[4372802, 116617, 216, 6, 27, "Input",ExpressionUUID->"74512ce5-bdf7-4f88-ad73-d4524f0c5016",
 CellID->1720648468],
Cell[4373021, 116625, 19303, 465, 70, "Output",ExpressionUUID->"3feb056c-cd4b-4bbc-babb-10e916723d45",
 CellID->399274819]
}, Open  ]],
Cell[CellGroupData[{
Cell[4392361, 117095, 189, 4, 27, "Input",ExpressionUUID->"762bb786-dcc9-4794-9f18-63447e71b3c3",
 CellID->1401549961],
Cell[4392553, 117101, 14933, 490, 70, "Output",ExpressionUUID->"1166aa24-6cc5-4436-b35e-f267476d87c0",
 CellID->1872969705]
}, Open  ]],
Cell[CellGroupData[{
Cell[4407523, 117596, 162, 3, 27, "Input",ExpressionUUID->"a0d17ee4-dde4-4aa9-819f-84ddc64ae39e",
 CellID->1808354982],
Cell[4407688, 117601, 286755, 7686, 70, "Output",ExpressionUUID->"103cc88e-ae96-4c3d-b92b-d6adcedd05fd",
 CellID->1186534209]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4694492, 125293, 126, 1, 40, "MathCaption",ExpressionUUID->"5d53e608-4803-4897-9e43-c9ffc4fde434",
 CellID->1095120460],
Cell[4694621, 125296, 292, 8, 27, "Input",ExpressionUUID->"63b17bc0-06e7-4af1-a2f8-2c5586e64cf5",
 CellID->1754716430],
Cell[CellGroupData[{
Cell[4694938, 125308, 838, 27, 27, "Input",ExpressionUUID->"25edd050-b25f-4dbc-9905-d9e6db0f9380",
 CellID->126114759],
Cell[4695779, 125337, 368, 7, 70, "Output",ExpressionUUID->"81273cb6-bb72-49da-a041-e52a47920673",
 CellID->547304621]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4696196, 125350, 107, 1, 40, "MathCaption",ExpressionUUID->"b848bb31-6e6b-4b2c-8fad-6b97ba4f0b4d",
 CellID->968662159],
Cell[4696306, 125353, 261, 6, 27, "Input",ExpressionUUID->"e40f37ec-7c95-4e4a-a4fe-bcdf24b55d84",
 CellID->1831468572],
Cell[4696570, 125361, 268, 7, 27, "Input",ExpressionUUID->"5aa63be9-440d-4598-aa5f-4e388e155335",
 CellID->1609177153],
Cell[4696841, 125370, 247, 4, 27, "Input",ExpressionUUID->"233baefb-5d5d-498a-a4df-13ab7cd2d040",
 CellID->1102632757],
Cell[4697091, 125376, 166, 3, 27, "Input",ExpressionUUID->"b051f718-c16e-4204-be0a-6a75be57b7d4",
 CellID->31360153],
Cell[4697260, 125381, 164, 3, 27, "Input",ExpressionUUID->"53feb4b3-e44e-422c-aea4-b76b453485ed",
 CellID->936597956],
Cell[4697427, 125386, 166, 3, 27, "Input",ExpressionUUID->"1ffc3542-fa9c-4efe-a98c-2d85e7029da9",
 CellID->346713982]
}, Open  ]],
Cell[CellGroupData[{
Cell[4697630, 125394, 115, 1, 40, "MathCaption",ExpressionUUID->"026fb734-2e18-4d57-b712-430bbfa9228f",
 CellID->1098761163],
Cell[4697748, 125397, 170, 3, 27, "Input",ExpressionUUID->"64e65936-5ae4-429c-b534-85cd452a5e3a",
 CellID->666985988]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[4697967, 125406, 239, 3, 41, "Subsubsection",ExpressionUUID->"ef267f85-b30c-4069-9067-7d160e846c9a",
 CellID->168862976],
Cell[4698209, 125411, 177, 3, 27, "Input",ExpressionUUID->"d55c534e-e749-4b1b-a760-3d69b8ae8a7f",
 CellID->1254139945],
Cell[4698389, 125416, 362, 8, 47, "Input",ExpressionUUID->"d82adbe8-28f1-41c3-ad03-4dec3f8ba65e",
 CellID->778060310],
Cell[CellGroupData[{
Cell[4698776, 125428, 119, 1, 40, "MathCaption",ExpressionUUID->"a60d815c-71c1-48a8-96b3-fe0de2ab2564",
 CellID->831703587],
Cell[4698898, 125431, 289, 7, 27, "Input",ExpressionUUID->"b6179f4b-a5bf-4c5f-a9b6-09fc793b7fac",
 CellID->471128386]
}, Open  ]],
Cell[CellGroupData[{
Cell[4699224, 125443, 119, 1, 40, "MathCaption",ExpressionUUID->"d00f9eca-b9d7-4b90-b6e3-ccd9e991e9fb",
 CellID->482577013],
Cell[4699346, 125446, 357, 10, 27, "Input",ExpressionUUID->"b2c033bf-83c1-4ba0-bbe0-2c910cddad15",
 CellID->1773606419]
}, Open  ]],
Cell[CellGroupData[{
Cell[4699740, 125461, 143, 1, 40, "MathCaption",ExpressionUUID->"83a6946e-836d-45f2-ae28-a7ddef20d0a0",
 CellID->1930063773],
Cell[4699886, 125464, 249, 6, 27, "Input",ExpressionUUID->"8b7a6ef4-0997-481a-a53f-d6b46aa2859a",
 CellID->841344876],
Cell[CellGroupData[{
Cell[4700160, 125474, 1154, 36, 27, "Input",ExpressionUUID->"fa7e6fe3-4895-447e-99f3-cbf64dfd637c",
 CellID->1450999084],
Cell[4701317, 125512, 8265, 222, 70, "Output",ExpressionUUID->"b8e456bf-e4cb-462f-9e05-645a0c3e4c55",
 CellID->876305046]
}, Open  ]],
Cell[CellGroupData[{
Cell[4709619, 125739, 179, 3, 27, "Input",ExpressionUUID->"8c941de1-ca7f-401d-9f06-16e5588b5966",
 CellID->230039046],
Cell[4709801, 125744, 690, 18, 70, "Output",ExpressionUUID->"3673aa49-cccd-4537-9858-93b819ed8284",
 CellID->616056996]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4710540, 125768, 171, 3, 40, "MathCaption",ExpressionUUID->"4d2e63bd-cf66-4d6c-9f9b-313b6a562c1b",
 CellID->638670385],
Cell[4710714, 125773, 240, 6, 27, "Input",ExpressionUUID->"0d6b9bd2-64c3-4ee8-aef9-c1de825ff8fc",
 CellID->1009892989],
Cell[CellGroupData[{
Cell[4710979, 125783, 1548, 48, 47, "Input",ExpressionUUID->"3ed046d9-b62e-40c5-a83f-b839b40c1761",
 CellID->1708203882],
Cell[4712530, 125833, 10915, 281, 70, "Output",ExpressionUUID->"8066f187-6190-4f9a-b25d-c47058c9d770",
 CellID->1051269585]
}, Open  ]],
Cell[CellGroupData[{
Cell[4723482, 126119, 180, 3, 27, "Input",ExpressionUUID->"3b857b95-5f70-477d-a1dc-ed22e3c1cf4b",
 CellID->721232698],
Cell[4723665, 126124, 709, 19, 70, "Output",ExpressionUUID->"ee18d6ac-87ea-4715-a7af-4eee0cb08973",
 CellID->1339951226]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4724423, 126149, 322, 6, 40, "MathCaption",ExpressionUUID->"fa11c277-4c89-4afd-8a70-4cf718b71973",
 CellID->1518900828],
Cell[CellGroupData[{
Cell[4724770, 126159, 291, 6, 27, "Input",ExpressionUUID->"794976fd-c18a-4ac9-bc6a-5d644d62b137",
 CellID->1028104307],
Cell[4725064, 126167, 82460, 2245, 70, "Output",ExpressionUUID->"b18f5042-f5d4-456f-8831-bacb8b29d661",
 CellID->2046732940]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4807573, 128418, 130, 1, 40, "MathCaption",ExpressionUUID->"6cdc36c6-13e7-4d42-b297-a5e914013ab2",
 CellID->1123655241],
Cell[CellGroupData[{
Cell[4807728, 128423, 192, 4, 27, "Input",ExpressionUUID->"4de65dc3-8c14-42fc-b738-160e3be8d518",
 CellID->302443769],
Cell[4807923, 128429, 363, 6, 70, "Output",ExpressionUUID->"18b5649e-e6ca-48b5-a846-9c8d78394058",
 CellID->1047464934]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4808335, 128441, 417, 7, 56, "MathCaption",ExpressionUUID->"aff740f3-ab19-41a7-9d09-15abb8539e53",
 CellID->1902182538],
Cell[CellGroupData[{
Cell[4808777, 128452, 341, 7, 27, "Input",ExpressionUUID->"5de306fe-f0bb-4554-98b5-b4f846904337",
 CellID->1570432718],
Cell[4809121, 128461, 55131, 1529, 70, "Output",ExpressionUUID->"d31edbb2-3022-40ff-8bb9-6cdc40cc6dc8",
 CellID->1819691279]
}, Open  ]],
Cell[CellGroupData[{
Cell[4864289, 129995, 227, 4, 27, "Input",ExpressionUUID->"efcb8b99-4823-4f08-a16d-ee635b068388",
 CellID->1156953266],
Cell[4864519, 130001, 298, 5, 70, "Output",ExpressionUUID->"74a3665e-44e4-4a74-ac2b-522db6e5a5b6",
 CellID->1236709226]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4864866, 130012, 218, 2, 40, "MathCaption",ExpressionUUID->"8c33fccc-a5be-4644-9744-faa61970b62c",
 CellID->562721375],
Cell[CellGroupData[{
Cell[4865109, 130018, 254, 5, 27, "Input",ExpressionUUID->"4b1116eb-6427-4eb2-945b-6b5ac5ebb61b",
 CellID->253715799],
Cell[4865366, 130025, 245, 4, 70, "Output",ExpressionUUID->"1c79bee8-16a4-438d-a0b3-f17048d8d965",
 CellID->2025225716]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4865660, 130035, 247, 4, 40, "MathCaption",ExpressionUUID->"12924cfb-fc9f-4830-b114-e5e29d3e86aa",
 CellID->519202859],
Cell[CellGroupData[{
Cell[4865932, 130043, 364, 10, 27, "Input",ExpressionUUID->"6367e556-f81a-4b5f-86cf-54921fbb7ba1",
 CellID->46692262],
Cell[4866299, 130055, 261, 4, 70, "Output",ExpressionUUID->"17ea0895-722e-40d4-890b-f27bb87d6ffd",
 CellID->352803862]
}, Open  ]],
Cell[CellGroupData[{
Cell[4866597, 130064, 366, 10, 27, "Input",ExpressionUUID->"d612c7b2-cb77-4401-ad9a-11a150e9a2d5",
 CellID->1213071162],
Cell[4866966, 130076, 265, 4, 70, "Output",ExpressionUUID->"a2dc8b5b-9548-4fd2-ad2d-e2ffc1027320",
 CellID->372159267]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4867280, 130086, 267, 5, 40, "MathCaption",ExpressionUUID->"0dddd146-5faf-4114-9a02-a9823de4668c",
 CellID->884640233],
Cell[CellGroupData[{
Cell[4867572, 130095, 227, 5, 27, "Input",ExpressionUUID->"175fec35-3a6c-461e-83f9-c3523b4767a2",
 CellID->1119107850],
Cell[4867802, 130102, 75451, 2151, 70, "Output",ExpressionUUID->"fa34b121-9e26-4ec3-a9d4-147a94ac9f83",
 CellID->1408420405]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4943302, 132259, 185, 4, 40, "MathCaption",ExpressionUUID->"0b53b27d-9e1a-473a-aed8-9a71e411d651",
 CellID->606995413],
Cell[CellGroupData[{
Cell[4943512, 132267, 564, 17, 27, "Input",ExpressionUUID->"0b270d5d-600a-4688-b522-ff0e5f63d89c",
 CellID->1224583041],
Cell[4944079, 132286, 366, 6, 70, "Output",ExpressionUUID->"347b4a88-36d0-4189-a50c-f6252826648c",
 CellID->1756989601]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4944494, 132298, 314, 5, 40, "MathCaption",ExpressionUUID->"e87939c1-7354-4f3f-b660-6ae13d778dbd",
 CellID->1987924259],
Cell[CellGroupData[{
Cell[4944833, 132307, 365, 8, 27, "Input",ExpressionUUID->"ff5f5cd3-6d7e-4a5a-8d63-02b4f9e70e22",
 CellID->1367457052],
Cell[4945201, 132317, 47545, 1381, 70, "Output",ExpressionUUID->"643dec68-cf47-4ee3-a451-446e8076613e",
 CellID->1824077757]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4992795, 133704, 152, 1, 40, "MathCaption",ExpressionUUID->"e816dcce-a1ba-48ff-9db3-530a9ca9a4b2",
 CellID->280631018],
Cell[CellGroupData[{
Cell[4992972, 133709, 551, 17, 27, "Input",ExpressionUUID->"789d7b07-1535-422a-8d71-cb112955058b",
 CellID->691430714],
Cell[4993526, 133728, 336, 5, 70, "Output",ExpressionUUID->"14e43d40-0f30-4f1e-ba34-a6aa2990ce04",
 CellID->1945800974]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4993911, 133739, 142, 1, 40, "MathCaption",ExpressionUUID->"a2a56dca-4f7b-411a-ad9b-6a59e681b65f",
 CellID->1661744684],
Cell[CellGroupData[{
Cell[4994078, 133744, 400, 12, 27, "Input",ExpressionUUID->"5227ff79-26c9-46da-a029-9c32f20b699b",
 CellID->918074753],
Cell[4994481, 133758, 72390, 1553, 70, "Output",ExpressionUUID->"c0ad7ec7-df2b-471c-8630-7dbf194d1b40",
 CellID->1195993502]
}, Open  ]],
Cell[CellGroupData[{
Cell[5066908, 135316, 157, 3, 27, "Input",ExpressionUUID->"24757151-2c51-47a8-b048-e8bb4742e127",
 CellID->772457763],
Cell[5067068, 135321, 54874, 1587, 70, "Output",ExpressionUUID->"9856d164-2971-43f3-b9bc-6982d3569a82",
 CellID->615852194]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[5121991, 136914, 206, 2, 40, "MathCaption",ExpressionUUID->"67222cac-ebb0-4fb2-82a2-b6dfb03b535b",
 CellID->1579180349],
Cell[CellGroupData[{
Cell[5122222, 136920, 190, 4, 27, "Input",ExpressionUUID->"fe7d54c8-2ea6-494c-b7f2-a50fde9f313c",
 CellID->533732370],
Cell[5122415, 136926, 581992, 14991, 70, "Output",ExpressionUUID->"f0eda76f-267f-4953-b6d0-41afb7f8e069",
 CellID->1904129665]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[5704456, 151923, 191, 4, 40, "MathCaption",ExpressionUUID->"8c5dab29-33f0-4e20-87ff-eb47b43154e3",
 CellID->880919305],
Cell[CellGroupData[{
Cell[5704672, 151931, 198, 4, 27, "Input",ExpressionUUID->"b366e89a-7541-48ab-b24a-2ebff341af05",
 CellID->858496717],
Cell[5704873, 151937, 286929, 7428, 70, "Output",ExpressionUUID->"eec44119-9fd7-40d4-b428-30ef3dee79b0",
 CellID->905972061]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[5991851, 159371, 135, 1, 40, "MathCaption",ExpressionUUID->"e4c1b74b-9a13-4304-be65-55ff5e0bcb17",
 CellID->606341460],
Cell[CellGroupData[{
Cell[5992011, 159376, 180, 3, 27, "Input",ExpressionUUID->"a9cd576f-e219-4b42-9d59-e6eef9d9dd12",
 CellID->971013205],
Cell[5992194, 159381, 712, 19, 70, "Output",ExpressionUUID->"49b170c3-db3f-4a2e-bbdb-87189aba8aea",
 CellID->1013371997]
}, Open  ]],
Cell[CellGroupData[{
Cell[5992943, 159405, 184, 3, 27, "Input",ExpressionUUID->"7ed1df38-e3d8-471e-ba29-462fcfcbb240",
 CellID->522926982],
Cell[5993130, 159410, 1233, 37, 70, "Output",ExpressionUUID->"6f7634fe-b0f8-4940-a550-f28f774683e9",
 CellID->50349222]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[5994412, 159453, 262, 5, 40, "MathCaption",ExpressionUUID->"3299ca2e-f0ce-4b7f-9bc1-d41e5efc1d37",
 CellID->1229780792],
Cell[CellGroupData[{
Cell[5994699, 159462, 320, 7, 27, "Input",ExpressionUUID->"0c390ac8-0815-4b1e-be0c-e9fde1d06685",
 CellID->1543152416],
Cell[5995022, 159471, 364, 6, 70, "Output",ExpressionUUID->"7d3ce317-0136-4b1e-8eb2-078bc3650e17",
 CellID->110817236]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[5995435, 159483, 131, 1, 40, "MathCaption",ExpressionUUID->"d25eba21-85c9-4f02-9066-b08cbdbc1bcd",
 CellID->1047379202],
Cell[CellGroupData[{
Cell[5995591, 159488, 187, 4, 27, "Input",ExpressionUUID->"85e79de6-05bc-4d85-8b71-6039aca6daae",
 CellID->378793555],
Cell[5995781, 159494, 57095, 1851, 251, "Output",ExpressionUUID->"1bbfec0c-9c1e-4e64-99c2-896f149fb185",
 CellID->601619684]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6052925, 161351, 151, 1, 40, "MathCaption",ExpressionUUID->"15b5cbd7-aac9-4a2f-8718-94482cf7bf49",
 CellID->832336293],
Cell[CellGroupData[{
Cell[6053101, 161356, 250, 7, 27, "Input",ExpressionUUID->"3d95121f-fced-4fef-8417-d02f848b7718",
 CellID->367599893],
Cell[6053354, 161365, 307, 5, 26, "Output",ExpressionUUID->"24804c50-8866-4b6a-bf2f-4c1497ca40e4",
 CellID->1069018912]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6053710, 161376, 215, 2, 40, "MathCaption",ExpressionUUID->"99efe8ec-725e-4081-afd4-55b780bfca23",
 CellID->606168795],
Cell[CellGroupData[{
Cell[6053950, 161382, 232, 4, 27, "Input",ExpressionUUID->"65778c9e-58fd-4880-8c5c-9d3430d3182d",
 CellID->216516197],
Cell[6054185, 161388, 26001, 813, 126, "Output",ExpressionUUID->"1320d392-6b4f-4ad6-86cc-dd0d5a67c4bf",
 CellID->1152663182]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6080235, 162207, 115, 1, 40, "MathCaption",ExpressionUUID->"77b9b101-27c2-4e91-be0f-bebe87af4bae",
 CellID->142785295],
Cell[CellGroupData[{
Cell[6080375, 162212, 163, 3, 27, "Input",ExpressionUUID->"96512aa5-0eee-4214-a1c2-2d7e73793aea",
 CellID->1053219366],
Cell[6080541, 162217, 10982925, 228385, 2098, "Output",ExpressionUUID->"46294718-78e9-4133-ad74-4e16b13f6123",
 CellID->360120287]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[17063515, 390608, 108, 1, 40, "MathCaption",ExpressionUUID->"c1681903-5d69-4f8d-a1c7-a8630340bc0e",
 CellID->1502852846],
Cell[17063626, 390611, 357, 7, 27, "Input",ExpressionUUID->"05046742-69c9-484e-af67-1af2d0bbb329",
 CellID->574457121],
Cell[17063986, 390620, 208, 5, 27, "Input",ExpressionUUID->"c46076c3-27d1-45c0-8779-fa7cc22b7cde",
 CellID->1961839382],
Cell[17064197, 390627, 164, 3, 27, "Input",ExpressionUUID->"532b5357-dd19-4dd4-b933-b8c4a247e19f",
 CellID->170577236],
Cell[17064364, 390632, 162, 3, 27, "Input",ExpressionUUID->"7f9dc083-5f0f-41b0-934e-aa5772caa148",
 CellID->936707],
Cell[17064529, 390637, 169, 3, 27, "Input",ExpressionUUID->"2f848caf-808f-49a0-86b6-ed3ad3d79aba",
 CellID->710849020]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[17064747, 390646, 242, 3, 41, "Subsubsection",ExpressionUUID->"92193d4e-b29b-4099-b849-346405bd9932",
 CellID->787950879],
Cell[CellGroupData[{
Cell[17065014, 390653, 149, 1, 40, "MathCaption",ExpressionUUID->"51b3c354-a94f-4a48-8133-728a127f4f06",
 CellID->1706255951],
Cell[17065166, 390656, 176, 3, 27, "Input",ExpressionUUID->"8475a7c5-bd1d-44b5-a6b3-e773281426dd",
 CellID->1694630066],
Cell[17065345, 390661, 361, 8, 47, "Input",ExpressionUUID->"015571dd-f029-4c5e-9d91-3090259789f7",
 CellID->1719392748]
}, Open  ]],
Cell[CellGroupData[{
Cell[17065743, 390674, 120, 1, 40, "MathCaption",ExpressionUUID->"c5911706-7556-475a-b85b-368a0fba0052",
 CellID->1293961604],
Cell[17065866, 390677, 287, 7, 27, "Input",ExpressionUUID->"ba70e460-4966-464c-b20d-691ff9a5a949",
 CellID->45785259],
Cell[17066156, 390686, 401, 11, 27, "Input",ExpressionUUID->"2babde1f-f6ca-4276-9929-82483e3928da",
 CellID->713702387]
}, Open  ]],
Cell[CellGroupData[{
Cell[17066594, 390702, 136, 1, 40, "MathCaption",ExpressionUUID->"1d80b263-c725-4146-aec2-c94bb0be9ab3",
 CellID->1499849814],
Cell[17066733, 390705, 294, 8, 27, "Input",ExpressionUUID->"6b11e660-2802-4e79-8494-abc120747971",
 CellID->812775412],
Cell[CellGroupData[{
Cell[17067052, 390717, 2318, 67, 66, "Input",ExpressionUUID->"c87c8daf-a6dc-4c38-a848-9840f7f3e1e8",
 CellID->172870917],
Cell[17069373, 390786, 21116, 557, 56, "Output",ExpressionUUID->"cb0618b0-7a3e-4130-9618-bbcd992801ee",
 CellID->1047119116]
}, Open  ]],
Cell[CellGroupData[{
Cell[17090526, 391348, 178, 3, 27, "Input",ExpressionUUID->"9302139c-3731-42c2-a6a2-a1b5cc2bdff8",
 CellID->121248807],
Cell[17090707, 391353, 665, 17, 26, "Output",ExpressionUUID->"204a8dfb-0c09-4998-9e03-3616d949860b",
 CellID->302495447]
}, Open  ]],
Cell[17091387, 391373, 341, 8, 27, "Input",ExpressionUUID->"850afa34-94cc-4755-8bae-3b6edd54ff33",
 CellID->1906487990],
Cell[CellGroupData[{
Cell[17091753, 391385, 5569, 148, 332, "Input",ExpressionUUID->"46190505-05e2-4a44-8a2f-6e85a89e1c5d",
 CellID->1887634690],
Cell[17097325, 391535, 48767, 1215, 136, "Output",ExpressionUUID->"6b272ca2-b0cf-423f-8007-37d672e31fb1",
 CellID->947185280]
}, Open  ]],
Cell[CellGroupData[{
Cell[17146129, 392755, 180, 3, 27, "Input",ExpressionUUID->"c1cdb529-97b7-4052-8e74-867871cb8422",
 CellID->2029181533],
Cell[17146312, 392760, 686, 18, 26, "Output",ExpressionUUID->"603ec58e-72fd-408b-83d8-d99ff58cfb0e",
 CellID->46725749]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[17147047, 392784, 326, 6, 40, "MathCaption",ExpressionUUID->"bc101896-db84-4c88-b09f-1695b615c67b",
 CellID->743660032],
Cell[CellGroupData[{
Cell[17147398, 392794, 436, 10, 27, "Input",ExpressionUUID->"a93e5375-c72d-402f-86c7-faae4dc67634",
 CellID->1256506760],
Cell[17147837, 392806, 273, 6, 26, "Output",ExpressionUUID->"cd3b3551-3ca6-4663-9365-919b30b8a356",
 CellID->1488132288]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[17148159, 392818, 261, 5, 40, "MathCaption",ExpressionUUID->"1b966922-3091-4455-82e7-b5db4df0e2bd",
 CellID->1168661657],
Cell[CellGroupData[{
Cell[17148445, 392827, 399, 10, 27, "Input",ExpressionUUID->"496ec91a-ac9f-41f6-bacf-999f3773ad11",
 CellID->811033737],
Cell[17148847, 392839, 387, 7, 26, "Output",ExpressionUUID->"1888ecaf-85f8-4fba-aad8-e241bcef59f0",
 CellID->1444811454]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[17149283, 392852, 130, 1, 40, "MathCaption",ExpressionUUID->"870ff78d-b19c-4422-b09d-9d1623aa8dc1",
 CellID->1877902312],
Cell[CellGroupData[{
Cell[17149438, 392857, 191, 4, 27, "Input",ExpressionUUID->"56df24df-365b-432a-b3bf-2b4accb6cb83",
 CellID->336997825],
Cell[17149632, 392863, 336, 5, 26, "Output",ExpressionUUID->"655040cf-395c-4cb8-8a56-556a55426e5d",
 CellID->1348510749]
}, Open  ]],
Cell[CellGroupData[{
Cell[17150005, 392873, 353, 10, 27, "Input",ExpressionUUID->"d0d69b49-2553-412c-90dd-b38980e847f7",
 CellID->1094267069],
Cell[17150361, 392885, 1795035, 37919, 136, "Output",ExpressionUUID->"ee1108bb-9a9e-44ef-b55b-42c1a941f7e4",
 CellID->2025231652]
}, Open  ]],
Cell[CellGroupData[{
Cell[18945433, 430809, 563, 17, 27, "Input",ExpressionUUID->"0815256a-948d-4663-84df-2c49039ec206",
 CellID->1920205738],
Cell[18945999, 430828, 333, 5, 26, "Output",ExpressionUUID->"b96a8c3d-bbfe-48cc-b6f2-77035e8078de",
 CellID->452570214]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[18946381, 430839, 142, 1, 40, "MathCaption",ExpressionUUID->"d38eccff-39bb-4f25-a870-106e5d6317e7",
 CellID->1194722857],
Cell[CellGroupData[{
Cell[18946548, 430844, 197, 4, 27, "Input",ExpressionUUID->"7b4da674-2b8b-4a6f-8267-7e7702dc5877",
 CellID->734958149],
Cell[18946748, 430850, 339, 5, 26, "Output",ExpressionUUID->"3d7e3082-4f27-42cc-904d-23d8cbe5a330",
 CellID->2107992671]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[18947136, 430861, 197, 4, 40, "MathCaption",ExpressionUUID->"696aae64-e798-4a06-abb6-2d939248a9b1",
 CellID->638351558],
Cell[CellGroupData[{
Cell[18947358, 430869, 645, 15, 47, "Input",ExpressionUUID->"baf5ca67-ddf2-4eb4-9e37-bfb59464edf7",
 CellID->2093818541],
Cell[18948006, 430886, 362, 7, 26, "Output",ExpressionUUID->"6505fc3b-e6c5-4e3e-8a36-9c8e2b8f1677",
 CellID->1995096938]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[18948417, 430899, 158, 2, 40, "MathCaption",ExpressionUUID->"30709b74-1cdd-4f95-9886-5849c4e3a081",
 CellID->1409375136],
Cell[CellGroupData[{
Cell[18948600, 430905, 276, 8, 27, "Input",ExpressionUUID->"dd2ef8f7-6083-4d5c-a96f-c2e7b6efa54f",
 CellID->623284342],
Cell[18948879, 430915, 343, 7, 26, "Output",ExpressionUUID->"74833659-82f4-4a11-a031-5449210d1530",
 CellID->784572708]
}, Open  ]],
Cell[CellGroupData[{
Cell[18949259, 430927, 285, 8, 27, "Input",ExpressionUUID->"b58136a0-7722-4980-8d4b-de2e26211bec",
 CellID->1569700537],
Cell[18949547, 430937, 342, 7, 26, "Output",ExpressionUUID->"ab02b11a-d9a7-484a-9775-cab0b1af3930",
 CellID->763910776]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[18949938, 430950, 140, 1, 40, "MathCaption",ExpressionUUID->"e247d33d-fc52-4208-9e5a-1ff5d2e98cd3",
 CellID->1469137121],
Cell[CellGroupData[{
Cell[18950103, 430955, 178, 3, 27, "Input",ExpressionUUID->"5cffd40b-5662-4fb5-bb77-658076c0e057",
 CellID->20619441],
Cell[18950284, 430960, 650, 18, 26, "Output",ExpressionUUID->"0417c870-e907-4305-aa6f-505cc7f6c87c",
 CellID->1391170466]
}, Open  ]],
Cell[CellGroupData[{
Cell[18950971, 430983, 184, 3, 27, "Input",ExpressionUUID->"28ab8d1a-d150-40a6-8d3c-a76fff65953e",
 CellID->1421394764],
Cell[18951158, 430988, 891, 27, 26, "Output",ExpressionUUID->"24e15609-53e2-4b78-b6a4-1f43925a7c3c",
 CellID->1747557173]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[18952098, 431021, 107, 1, 40, "MathCaption",ExpressionUUID->"55771f9d-b07f-4127-b6e1-669777713aa2",
 CellID->719058655],
Cell[18952208, 431024, 170, 3, 27, "Input",ExpressionUUID->"580d1839-a2de-44ff-8ec6-be5399c9537b",
 CellID->1328332916],
Cell[18952381, 431029, 199, 4, 27, "Input",ExpressionUUID->"a69898e0-f11e-4b11-b2b1-85212891f342",
 CellID->51325000],
Cell[18952583, 431035, 165, 3, 27, "Input",ExpressionUUID->"e3b47781-5c33-4395-837f-bd6058e95ffd",
 CellID->1478332553],
Cell[18952751, 431040, 166, 3, 27, "Input",ExpressionUUID->"d215e680-a434-4fb6-aff7-3dba2faf7ffa",
 CellID->2096906128],
Cell[18952920, 431045, 266, 6, 27, "Input",ExpressionUUID->"61226aa2-351e-421f-9914-2d6a5c3c967e",
 CellID->347162379],
Cell[18953189, 431053, 168, 3, 27, "Input",ExpressionUUID->"1468d86d-eff6-4efc-a3b8-260e3a38a554",
 CellID->45034954]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[18953406, 431062, 249, 3, 41, "Subsubsection",ExpressionUUID->"d2ff033d-5806-4796-862a-95f90d52c195",
 CellID->988228134],
Cell[CellGroupData[{
Cell[18953680, 431069, 174, 3, 27, "Input",ExpressionUUID->"d1192b7a-a9cc-44ad-9566-10e581a22e17",
 CellID->511722240],
Cell[CellGroupData[{
Cell[18953879, 431076, 259, 4, 22, "Print",ExpressionUUID->"421a3cfd-dc36-4922-a7fa-80123cbbdb09",
 CellID->127239746],
Cell[18954141, 431082, 530, 12, 22, "Print",ExpressionUUID->"4aec092c-ad01-493b-95b8-b20e0b8ae14d",
 CellID->1772182257],
Cell[18954674, 431096, 278, 4, 22, "Print",ExpressionUUID->"af5afd6a-5948-420b-a784-1e15bc43096f",
 CellID->1616493156],
Cell[18954955, 431102, 235, 3, 22, "Print",ExpressionUUID->"6b903f41-d5d0-4df3-9026-e97c04933794",
 CellID->584854695],
Cell[18955193, 431107, 220, 3, 22, "Print",ExpressionUUID->"4c75b752-a665-44cd-bcc6-6a814fcab2aa",
 CellID->620206546],
Cell[18955416, 431112, 258, 4, 22, "Print",ExpressionUUID->"e50dafbb-1cf7-48e7-a9fc-b29474f82ff7",
 CellID->2073279731],
Cell[18955677, 431118, 533, 12, 22, "Print",ExpressionUUID->"62a19878-f686-4e5d-9b0e-efb14ef37468",
 CellID->542782695],
Cell[18956213, 431132, 279, 4, 22, "Print",ExpressionUUID->"2f73d07b-3711-4be0-90f5-fbed18457c8c",
 CellID->38222409],
Cell[18956495, 431138, 256, 4, 22, "Print",ExpressionUUID->"909a848e-c6bb-4e6c-90d5-fbb55c9c8914",
 CellID->306417610],
Cell[18956754, 431144, 529, 12, 22, "Print",ExpressionUUID->"f7b2637d-4dd7-4f41-a82a-72dda31f8212",
 CellID->294985764],
Cell[18957286, 431158, 297, 4, 22, "Print",ExpressionUUID->"4b0e0004-6e46-4ada-8e64-8c9c5f3a3431",
 CellID->1433478272],
Cell[18957586, 431164, 258, 4, 22, "Print",ExpressionUUID->"97341e2a-bc0d-48d2-a661-64c5ff1ac1b1",
 CellID->1997765458],
Cell[18957847, 431170, 412, 6, 42, "Print",ExpressionUUID->"c574d2f4-f62c-436d-8356-f5a11f5f4977",
 CellID->537231761],
Cell[18958262, 431178, 260, 4, 22, "Print",ExpressionUUID->"888474dc-47e7-47aa-a861-7bf627aa7857",
 CellID->1232955139]
}, Open  ]]
}, Closed]],
Cell[18958549, 431186, 212, 5, 23, "Input",ExpressionUUID->"33ced122-c44d-4771-803b-297ed3a150e8",
 CellID->1991165092],
Cell[18958764, 431193, 288, 7, 27, "Input",ExpressionUUID->"08253957-1249-4212-9bfd-b54df69cbdf0",
 CellID->1274030043],
Cell[18959055, 431202, 500, 13, 27, "Input",ExpressionUUID->"65b825f3-7a55-4f14-b2ef-8b6bdd7aecb2",
 CellID->63488464],
Cell[CellGroupData[{
Cell[18959580, 431219, 325, 9, 40, "MathCaption",ExpressionUUID->"8b1c0f79-8dca-494b-87ae-d0f07b4c879a",
 CellID->1898725536],
Cell[18959908, 431230, 237, 5, 27, "Input",ExpressionUUID->"c8dc1aa6-e9ac-4cf3-9885-ab4a87eb0d06",
 CellID->273089132],
Cell[18960148, 431237, 1696, 55, 50, "Input",ExpressionUUID->"abb14f69-54f8-4549-b3b6-92503e86e821",
 CellID->636487885],
Cell[18961847, 431294, 382, 11, 27, "Input",ExpressionUUID->"dc66e6fc-e43d-476c-85b3-21b842575739",
 CellID->835837488],
Cell[18962232, 431307, 1643, 50, 106, "Input",ExpressionUUID->"9def3a93-8b4e-4195-ac40-49a51b230c9b",
 CellID->1230125789],
Cell[18963878, 431359, 5217, 158, 210, "Input",ExpressionUUID->"2ad47032-5f48-4164-a1ce-e32797ee3e25",
 CellID->899066054],
Cell[CellGroupData[{
Cell[18969120, 431521, 650, 18, 47, "Input",ExpressionUUID->"43079b39-5b79-41b4-b087-712ea7e0bfcf",
 CellID->908229286],
Cell[18969773, 431541, 320, 6, 26, "Output",ExpressionUUID->"45b207d6-8b28-47d5-bf2f-3bd63a0d1cb0",
 CellID->1207311270]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[18970142, 431553, 106, 1, 40, "MathCaption",ExpressionUUID->"914c2c82-9474-4baa-9fad-b9def86c8fae",
 CellID->874642477],
Cell[CellGroupData[{
Cell[18970273, 431558, 469, 12, 27, "Input",ExpressionUUID->"59fa8d46-64bd-46cd-ae80-c694afac595e",
 CellID->590547224],
Cell[18970745, 431572, 318, 6, 26, "Output",ExpressionUUID->"0a4f4464-479e-4a1a-a082-72ccd2b0cb32",
 CellID->1855018034]
}, Open  ]],
Cell[18971078, 431581, 547, 17, 27, "Input",ExpressionUUID->"3250959d-1182-47e1-a4c8-616626dc37df",
 CellID->1372342851]
}, Open  ]],
Cell[CellGroupData[{
Cell[18971662, 431603, 207, 4, 40, "MathCaption",ExpressionUUID->"d55d703d-49dc-4dfd-8197-734176dabe08",
 CellID->991927977],
Cell[18971872, 431609, 183, 4, 27, "Input",ExpressionUUID->"15699e43-b181-4529-98e3-b8c1609bf578",
 CellID->417957778]
}, Open  ]],
Cell[CellGroupData[{
Cell[18972092, 431618, 181, 3, 40, "MathCaption",ExpressionUUID->"3e594bed-dec5-427f-a434-ffac56d210aa",
 CellID->197365509],
Cell[CellGroupData[{
Cell[18972298, 431625, 400, 10, 27, "Input",ExpressionUUID->"f4531ce2-9861-44d1-9abd-0e35d4c37ab9",
 CellID->1407358795],
Cell[CellGroupData[{
Cell[18972723, 431639, 513, 11, 22, "Print",ExpressionUUID->"71fd4389-51e5-4da6-b2ab-e9009bc173ba",
 CellID->120510292],
Cell[18973239, 431652, 505, 9, 22, "Print",ExpressionUUID->"ff30b7c5-9abc-4d4f-9ea9-c08e4b372d59",
 CellID->339398523],
Cell[18973747, 431663, 581, 10, 22, "Print",ExpressionUUID->"1f792a75-f092-4e6f-9cd6-155adffd023b",
 CellID->554409629],
Cell[18974331, 431675, 599, 11, 22, "Print",ExpressionUUID->"6e4a4c19-f548-4f7c-a93b-bed51194d48f",
 CellID->31090270],
Cell[18974933, 431688, 594, 14, 22, "Print",ExpressionUUID->"282296ef-b224-47bb-847e-e0fb3bfbcbb0",
 CellID->695641669],
Cell[18975530, 431704, 503, 9, 22, "Print",ExpressionUUID->"4a7b48b2-960c-4c64-8cef-da5fa0f4bea9",
 CellID->748556615],
Cell[18976036, 431715, 584, 10, 22, "Print",ExpressionUUID->"a70bbca8-15f2-4c98-b2c3-08e8378885b3",
 CellID->1031824179],
Cell[18976623, 431727, 601, 14, 22, "Print",ExpressionUUID->"d990a980-fbdb-4d53-a006-176257fa205e",
 CellID->938526928],
Cell[18977227, 431743, 505, 9, 22, "Print",ExpressionUUID->"5204f59f-6b26-4220-9ae6-0a02943e034e",
 CellID->554691681],
Cell[18977735, 431754, 579, 10, 22, "Print",ExpressionUUID->"4128648a-937b-4e66-94a0-8864938c67fc",
 CellID->476570940],
Cell[18978317, 431766, 581, 13, 22, "Print",ExpressionUUID->"b6c2b842-6976-4021-81df-edf20f3e0916",
 CellID->1761268917],
Cell[18978901, 431781, 503, 9, 22, "Print",ExpressionUUID->"6f18d7a9-0fae-4742-837e-cc90dec8a5d8",
 CellID->732097771],
Cell[18979407, 431792, 582, 10, 22, "Print",ExpressionUUID->"b630ca4e-2643-470a-8910-e69b2bf8cd0e",
 CellID->1433423534],
Cell[18979992, 431804, 468, 9, 22, "Print",ExpressionUUID->"74277c14-3d0f-4d75-b4b1-d784291351bf",
 CellID->1077783892]
}, Open  ]],
Cell[18980475, 431816, 301, 6, 26, "Output",ExpressionUUID->"d6be4381-1ae8-46c9-91f5-086136cdab71",
 CellID->289310399]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[18980825, 431828, 151, 1, 40, "MathCaption",ExpressionUUID->"76f7ce72-3ee5-4788-b441-374f6b5a30c4",
 CellID->1725377310],
Cell[CellGroupData[{
Cell[18981001, 431833, 458, 12, 27, "Input",ExpressionUUID->"691708be-6030-40cc-9e1b-e08600b6a8b5",
 CellID->944763729],
Cell[CellGroupData[{
Cell[18981484, 431849, 487, 11, 22, "Print",ExpressionUUID->"5d92a489-8b65-48c6-a905-9d424cf73517",
 CellID->536899185],
Cell[18981974, 431862, 284, 4, 22, "Print",ExpressionUUID->"8977fdc5-4290-4e83-9162-19abb0775fff",
 CellID->2085801248],
Cell[18982261, 431868, 568, 14, 22, "Print",ExpressionUUID->"ddff720b-4cfd-4123-add7-f58405422a1d",
 CellID->538521508],
Cell[18982832, 431884, 283, 4, 22, "Print",ExpressionUUID->"01b06207-7dc3-42a5-a3b5-f5469a16cd3c",
 CellID->1434896498],
Cell[18983118, 431890, 604, 15, 22, "Print",ExpressionUUID->"fa2d5371-cbc4-4d48-82dd-1c63e84052bb",
 CellID->64086076],
Cell[18983725, 431907, 478, 9, 22, "Print",ExpressionUUID->"81b7e4a3-682a-40aa-bd6b-44e695bdd24f",
 CellID->1205372185],
Cell[18984206, 431918, 555, 10, 22, "Print",ExpressionUUID->"6742f342-aa69-454d-bedf-113c203b968c",
 CellID->223757888],
Cell[18984764, 431930, 575, 14, 22, "Print",ExpressionUUID->"bb164395-5360-4bac-8526-2fa68c9038b8",
 CellID->454957719],
Cell[18985342, 431946, 286, 4, 22, "Print",ExpressionUUID->"fd5b6e90-25a7-4855-96d8-2a28bbe952b6",
 CellID->1432599913],
Cell[18985631, 431952, 554, 13, 22, "Print",ExpressionUUID->"b6e5d160-6716-4771-944b-79d0769e5ffa",
 CellID->120087627],
Cell[18986188, 431967, 284, 4, 22, "Print",ExpressionUUID->"d292bca8-b96a-4dcb-bf4d-a43f5530dfaf",
 CellID->1825602456],
Cell[18986475, 431973, 439, 9, 22, "Print",ExpressionUUID->"7badd94e-aa3f-4497-8130-3113b5b320ce",
 CellID->1801613303],
Cell[18986917, 431984, 285, 4, 22, "Print",ExpressionUUID->"c95421d7-6b76-4788-b4fc-0ed441d1d60f",
 CellID->195360821],
Cell[18987205, 431990, 604, 15, 22, "Print",ExpressionUUID->"74ad2dae-0348-442f-b0e9-20ea9d7ef8e3",
 CellID->2074305931],
Cell[18987812, 432007, 477, 9, 22, "Print",ExpressionUUID->"c35edef4-7f77-4501-917c-d5f8afb503c4",
 CellID->610256919],
Cell[18988292, 432018, 555, 10, 22, "Print",ExpressionUUID->"ffe0a830-01d9-4652-8b1c-d0514caaba16",
 CellID->206139561],
Cell[18988850, 432030, 578, 11, 22, "Print",ExpressionUUID->"a7cc9aca-b8de-4cce-9d57-137ca50a9c9d",
 CellID->696479116],
Cell[18989431, 432043, 584, 14, 22, "Print",ExpressionUUID->"ee94c78f-7587-4e43-823b-b70231c7b54c",
 CellID->1938626462],
Cell[18990018, 432059, 477, 9, 22, "Print",ExpressionUUID->"17f66a69-f34c-4f7c-99f3-4aaa33a5503d",
 CellID->873054550],
Cell[18990498, 432070, 557, 10, 22, "Print",ExpressionUUID->"06beb516-8aa4-40f9-9574-7ac35602e2c2",
 CellID->1819545546],
Cell[18991058, 432082, 578, 11, 22, "Print",ExpressionUUID->"56180939-8371-44c2-a699-06782fa3a9a0",
 CellID->934858342],
Cell[18991639, 432095, 530, 13, 22, "Print",ExpressionUUID->"4087d0a7-3fbd-4aa2-9614-0b00313c9460",
 CellID->1173792542],
Cell[18992172, 432110, 476, 9, 22, "Print",ExpressionUUID->"28811e61-9f5f-40ab-af4a-5be4246a2e94",
 CellID->1146472052],
Cell[18992651, 432121, 556, 10, 22, "Print",ExpressionUUID->"be7c0b94-c90f-4329-bee5-78469f856c30",
 CellID->1532497080],
Cell[18993210, 432133, 564, 10, 22, "Print",ExpressionUUID->"9946078b-61a3-49e1-ac0e-bca256f6f7c0",
 CellID->1907045089]
}, Open  ]],
Cell[18993789, 432146, 299, 6, 26, "Output",ExpressionUUID->"8fc9d204-92c1-476f-9b5c-9e3e8b951d22",
 CellID->513379348]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[18994137, 432158, 124, 1, 40, "MathCaption",ExpressionUUID->"701bd548-5abb-4f78-bc58-0192e45baaab",
 CellID->277432000],
Cell[18994264, 432161, 209, 5, 27, "Input",ExpressionUUID->"49cb1615-bf96-4a31-a6f9-577d4e955a2d",
 CellID->612626363],
Cell[CellGroupData[{
Cell[18994498, 432170, 159, 3, 27, "Input",ExpressionUUID->"acb4bb2a-3fe3-4058-ab0f-af37462f12e8",
 CellID->1374679223],
Cell[18994660, 432175, 246, 4, 26, "Output",ExpressionUUID->"7e516e77-ceeb-469e-946b-ad2c434bb636",
 CellID->1924891837]
}, Open  ]],
Cell[18994921, 432182, 361, 7, 27, "Input",ExpressionUUID->"54417a2d-f882-4dc8-8d5a-3fc8b86eb538",
 CellID->858534719],
Cell[18995285, 432191, 240, 5, 27, "Input",ExpressionUUID->"0d103c7f-a147-4058-99ab-379fe8034401",
 CellID->450398040],
Cell[18995528, 432198, 164, 3, 27, "Input",ExpressionUUID->"c740d95f-ea7b-4365-a149-5758a6661051",
 CellID->2107486252],
Cell[18995695, 432203, 166, 3, 27, "Input",ExpressionUUID->"89645d35-3094-4ff5-959b-30727bf8c0fd",
 CellID->2073647393],
Cell[18995864, 432208, 169, 3, 27, "Input",ExpressionUUID->"08be4813-9aca-442d-b555-67e3d353e9e1",
 CellID->279626988]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[18996082, 432217, 291, 6, 41, "Subsubsection",ExpressionUUID->"fcd30e6a-3b7a-4b2d-ac9f-d045bd6ba2b1",
 CellID->996567680],
Cell[18996376, 432225, 176, 3, 27, "Input",ExpressionUUID->"7215fc00-a883-41b3-ab15-cf9c19a80c0b",
 CellID->1088034807],
Cell[18996555, 432230, 212, 5, 27, "Input",ExpressionUUID->"062f5dd5-6f0a-407c-911a-4aa001fe9eba",
 CellID->777747044],
Cell[18996770, 432237, 289, 7, 27, "Input",ExpressionUUID->"54c95a90-8e6a-49cf-8d88-36fc9aff7f97",
 CellID->2009092866],
Cell[18997062, 432246, 459, 12, 27, "Input",ExpressionUUID->"8afcfaf9-12b8-4e06-9634-063904f81d6e",
 CellID->1938417081],
Cell[18997524, 432260, 660, 16, 85, "Input",ExpressionUUID->"4509a127-ba1c-4c1d-ad42-a3e97ef50b43",
 CellID->1648955377],
Cell[CellGroupData[{
Cell[18998209, 432280, 140, 1, 40, "MathCaption",ExpressionUUID->"d487aa7f-e3d8-49f9-91ab-bbef59c6a6c6",
 CellID->716556704],
Cell[18998352, 432283, 1450, 48, 85, "Input",ExpressionUUID->"e541e4d0-d285-4f1a-a4fe-bbdce46ae038",
 CellID->80231752],
Cell[18999805, 432333, 1289, 41, 87, "Input",ExpressionUUID->"d599e52d-66d5-431c-8e23-16d056035a79",
 CellID->1982809195],
Cell[CellGroupData[{
Cell[19001119, 432378, 564, 15, 27, "Input",ExpressionUUID->"6465ba3a-a7b1-4a8f-a236-b409c7b39c42",
 CellID->422906219],
Cell[19001686, 432395, 34916, 1047, 142, "Output",ExpressionUUID->"41e8b6d4-3b4d-4769-9df6-b4211435f35c",
 CellID->98098026]
}, Open  ]],
Cell[19036617, 433445, 339, 10, 27, "Input",ExpressionUUID->"6d84c684-fb5b-42a3-96e1-0dc497f28572",
 CellID->1372777775],
Cell[19036959, 433457, 184, 4, 27, "Input",ExpressionUUID->"911feaa3-789a-450f-9ef0-663af62c0137",
 CellID->1859244704],
Cell[CellGroupData[{
Cell[19037168, 433465, 349, 9, 27, "Input",ExpressionUUID->"2b9bd81a-be7d-462d-a1d4-cd9f29980464",
 CellID->291192631],
Cell[19037520, 433476, 226, 5, 26, "Output",ExpressionUUID->"902e90f3-f156-490d-a971-c6c3bda4ebbb",
 CellID->1613565733]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[19037795, 433487, 122, 1, 40, "MathCaption",ExpressionUUID->"324fd376-8264-418b-941a-2e97e0ed5e69",
 CellID->1299934521],
Cell[CellGroupData[{
Cell[19037942, 433492, 449, 12, 27, "Input",ExpressionUUID->"c207b4b3-cdc9-4743-91a1-1b4411efbc4d",
 CellID->1648353516],
Cell[CellGroupData[{
Cell[19038416, 433508, 418, 10, 22, "Print",ExpressionUUID->"835e4f58-9813-4249-ba82-e4f20be89fbf",
 CellID->775780215],
Cell[19038837, 433520, 215, 3, 22, "Print",ExpressionUUID->"ff82a3fa-7396-4a8d-8133-241a3b234e8a",
 CellID->1164852568],
Cell[19039055, 433525, 502, 13, 22, "Print",ExpressionUUID->"ddcdcbe8-ee1d-449a-a00c-acaec8eb0059",
 CellID->1216184324],
Cell[19039560, 433540, 214, 3, 22, "Print",ExpressionUUID->"1a4998d8-4506-4477-b5fe-d2c7f4a46331",
 CellID->660581322],
Cell[19039777, 433545, 537, 14, 22, "Print",ExpressionUUID->"3333447c-40c3-4ede-b1a4-de48fd407784",
 CellID->2047307511],
Cell[19040317, 433561, 405, 8, 22, "Print",ExpressionUUID->"44330806-7c00-4c50-982b-b6c4dc5390c9",
 CellID->2114345143],
Cell[19040725, 433571, 489, 9, 22, "Print",ExpressionUUID->"44ea9afc-22cb-4298-a481-4bb39d3a04e2",
 CellID->1386911492],
Cell[19041217, 433582, 510, 13, 22, "Print",ExpressionUUID->"c68c512f-9f74-4f1b-8c24-01ebae4607ee",
 CellID->1033941506],
Cell[19041730, 433597, 213, 3, 22, "Print",ExpressionUUID->"3203bd73-1f69-4e5e-bdff-753bd6244108",
 CellID->17219818],
Cell[19041946, 433602, 486, 12, 22, "Print",ExpressionUUID->"52229f00-c2fd-42e1-953d-3156240aecc5",
 CellID->1650672428],
Cell[19042435, 433616, 215, 3, 22, "Print",ExpressionUUID->"29da8afa-bef7-4213-b5e1-cf8463a2eb41",
 CellID->1657745450],
Cell[19042653, 433621, 370, 8, 22, "Print",ExpressionUUID->"d3e9b296-41bf-4a4a-b566-d3c999f96453",
 CellID->274941972],
Cell[19043026, 433631, 215, 3, 22, "Print",ExpressionUUID->"8263e620-7ee4-4450-802b-c093693e94c2",
 CellID->1712398707],
Cell[19043244, 433636, 535, 14, 22, "Print",ExpressionUUID->"2b06e44b-1cd9-43da-b867-59c16874a299",
 CellID->1856715102],
Cell[19043782, 433652, 408, 8, 22, "Print",ExpressionUUID->"07f8618c-b29a-475e-9c8b-e7e04ede0c1b",
 CellID->725844757],
Cell[19044193, 433662, 489, 9, 22, "Print",ExpressionUUID->"559a4ba5-31e1-4ff7-a1fc-0f4747df76c1",
 CellID->1565715939],
Cell[19044685, 433673, 512, 10, 22, "Print",ExpressionUUID->"8d1726b4-ff60-4e3d-8d0d-c4e88f2e14d3",
 CellID->1870660082],
Cell[19045200, 433685, 515, 13, 22, "Print",ExpressionUUID->"54fc00f8-d173-4558-9803-db38b6cc9e11",
 CellID->1023406094],
Cell[19045718, 433700, 409, 8, 22, "Print",ExpressionUUID->"3e5fd03b-a991-4853-9c9d-c876d299e793",
 CellID->1189693035],
Cell[19046130, 433710, 488, 9, 22, "Print",ExpressionUUID->"b0eb6b10-659b-4d8e-8a47-e23cbb499d97",
 CellID->2098085675],
Cell[19046621, 433721, 509, 10, 22, "Print",ExpressionUUID->"70f5956b-c00d-48b2-ae6c-1ef160b4c9f8",
 CellID->844455985]
}, Open  ]],
Cell[19047145, 433734, 225, 5, 26, "Output",ExpressionUUID->"46100676-3fd1-47e9-b0cd-4ed2587e0ed1",
 CellID->52316872]
}, Open  ]],
Cell[19047385, 433742, 209, 5, 27, "Input",ExpressionUUID->"6b8bb9fa-99f4-4f24-9f04-b5d2275fddc9",
 CellID->203862004],
Cell[CellGroupData[{
Cell[19047619, 433751, 197, 4, 27, "Input",ExpressionUUID->"0e804627-b88b-4fb1-a065-9ec70adb70f0",
 CellID->1072284263],
Cell[19047819, 433757, 177, 3, 26, "Output",ExpressionUUID->"97aada0b-255b-4e55-a09b-d3e1bcc231f0",
 CellID->968856081]
}, Open  ]],
Cell[19048011, 433763, 183, 4, 27, "Input",ExpressionUUID->"81f0af1d-5eab-4b3a-861f-7170d3a9e20b",
 CellID->463309187],
Cell[CellGroupData[{
Cell[19048219, 433771, 245, 7, 27, "Input",ExpressionUUID->"ee993d98-67d4-4c4b-b415-140efed4ede1",
 CellID->61801887],
Cell[19048467, 433780, 226, 5, 26, "Output",ExpressionUUID->"94ea17dd-e1dc-4a14-9651-199b30e6afd3",
 CellID->1343141813]
}, Open  ]],
Cell[CellGroupData[{
Cell[19048730, 433790, 316, 8, 27, "Input",ExpressionUUID->"453df55f-aa90-4e8a-b654-5c0798cf7ce2",
 CellID->1114659762],
Cell[19049049, 433800, 874, 22, 76, "Output",ExpressionUUID->"4f38b2af-1567-4a45-9ece-f53fb94be573",
 CellID->1983837474]
}, Open  ]],
Cell[CellGroupData[{
Cell[19049960, 433827, 253, 7, 27, "Input",ExpressionUUID->"88b38de8-404d-4fc2-87ca-58bd447d2978",
 CellID->1343612025],
Cell[19050216, 433836, 226, 5, 26, "Output",ExpressionUUID->"553c96cb-fbc5-46af-8477-cce4e6c0108f",
 CellID->525322196]
}, Open  ]],
Cell[CellGroupData[{
Cell[19050479, 433846, 316, 8, 27, "Input",ExpressionUUID->"570c0c58-e8b1-4b9d-acad-7e629dd7eec5",
 CellID->1913538034],
Cell[19050798, 433856, 870, 22, 76, "Output",ExpressionUUID->"6b69fd9b-ff74-46c4-8dbd-4c254fd47d40",
 CellID->784875355]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[19051717, 433884, 226, 4, 40, "MathCaption",ExpressionUUID->"d2b3b6dc-ac60-4410-b7b0-8597903914e2",
 CellID->1117084146],
Cell[19051946, 433890, 230, 6, 27, "Input",ExpressionUUID->"d1a5e2cc-bc54-43cf-b065-0103df98093a",
 CellID->1531199748],
Cell[CellGroupData[{
Cell[19052201, 433900, 408, 13, 27, "Input",ExpressionUUID->"dd2bea58-8aff-4a58-afca-dada75c0585e",
 CellID->1577622635],
Cell[19052612, 433915, 7135, 236, 54, "Output",ExpressionUUID->"ed5be0ca-426f-47d5-88a7-20ff517a7a0f",
 CellID->1555531611]
}, Open  ]],
Cell[CellGroupData[{
Cell[19059784, 434156, 408, 13, 27, "Input",ExpressionUUID->"615dd8e2-ddb2-49ee-b410-119a1628b8e8",
 CellID->2002514938],
Cell[19060195, 434171, 222, 5, 26, "Output",ExpressionUUID->"3239f048-0abc-4454-82b0-972ad38d908e",
 CellID->353867499]
}, Open  ]],
Cell[CellGroupData[{
Cell[19060454, 434181, 408, 13, 27, "Input",ExpressionUUID->"8d18d307-07b4-4310-9e29-1efe41fadf66",
 CellID->1434883467],
Cell[19060865, 434196, 225, 5, 26, "Output",ExpressionUUID->"1cce3183-4040-44a8-9e4e-895d8f052006",
 CellID->1068837150]
}, Open  ]],
Cell[CellGroupData[{
Cell[19061127, 434206, 408, 13, 27, "Input",ExpressionUUID->"d1de3a38-e4aa-47ba-9575-b027ce849734",
 CellID->1057373424],
Cell[19061538, 434221, 7773, 251, 54, "Output",ExpressionUUID->"060308be-880b-4151-8987-b6416dfec00e",
 CellID->245272895]
}, Open  ]],
Cell[CellGroupData[{
Cell[19069348, 434477, 407, 13, 27, "Input",ExpressionUUID->"9f726658-434d-4309-9ae5-0a9f6b0841d6",
 CellID->932708609],
Cell[19069758, 434492, 4311, 142, 50, "Output",ExpressionUUID->"3bdd1b02-8137-4986-9b05-97e54eb7e28e",
 CellID->1280427672]
}, Open  ]],
Cell[CellGroupData[{
Cell[19074106, 434639, 407, 13, 27, "Input",ExpressionUUID->"22598816-9e19-4982-9965-2bf5f223de14",
 CellID->880359372],
Cell[19074516, 434654, 223, 5, 26, "Output",ExpressionUUID->"998db42f-e181-4c56-b825-e62f4f07a0fc",
 CellID->214256717]
}, Open  ]],
Cell[CellGroupData[{
Cell[19074776, 434664, 408, 13, 27, "Input",ExpressionUUID->"668ff03e-f79b-47f4-afae-63f50de3dd5d",
 CellID->1211152600],
Cell[19075187, 434679, 223, 5, 26, "Output",ExpressionUUID->"8e11b76b-e669-482c-9552-b45328623fae",
 CellID->1830050247]
}, Open  ]],
Cell[CellGroupData[{
Cell[19075447, 434689, 407, 13, 27, "Input",ExpressionUUID->"0d79a4ba-8f07-4610-8e13-affd8fd401bf",
 CellID->231595302],
Cell[19075857, 434704, 2928, 98, 50, "Output",ExpressionUUID->"ed9c6fdc-d0b2-4445-a50f-fa7723756926",
 CellID->617410896]
}, Open  ]],
Cell[CellGroupData[{
Cell[19078822, 434807, 408, 13, 27, "Input",ExpressionUUID->"dfe632b7-6d62-4d53-a58e-14839d82a1dc",
 CellID->1652335347],
Cell[19079233, 434822, 222, 5, 26, "Output",ExpressionUUID->"e58704a8-6aa2-41fd-bdfe-f5356eedde0f",
 CellID->183946768]
}, Open  ]],
Cell[CellGroupData[{
Cell[19079492, 434832, 408, 13, 27, "Input",ExpressionUUID->"e5c09111-7fc8-4abf-824c-1f22fec776c9",
 CellID->1750523611],
Cell[19079903, 434847, 11715, 346, 60, "Output",ExpressionUUID->"c620b273-0f03-43bb-8d78-81e06af29e7a",
 CellID->1364361743]
}, Open  ]],
Cell[19091633, 435196, 371, 7, 27, "Input",ExpressionUUID->"1a72008c-69df-4e0c-8da5-a9be8d22925e",
 CellID->1896933198],
Cell[19092007, 435205, 261, 6, 27, "Input",ExpressionUUID->"d4f75154-58b2-442b-948f-532bbe9033fb",
 CellID->214035424],
Cell[19092271, 435213, 163, 3, 27, "Input",ExpressionUUID->"3fdb6381-b5b5-4859-9a15-3c9573b21560",
 CellID->258262443],
Cell[19092437, 435218, 165, 3, 27, "Input",ExpressionUUID->"6ad213a1-9331-44be-88cb-77c982ee09e8",
 CellID->552428914],
Cell[19092605, 435223, 170, 3, 27, "Input",ExpressionUUID->"b8886a2a-8b57-4aca-bd35-d01211ab1bcf",
 CellID->1100951617]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[19092824, 435232, 268, 5, 41, "Subsubsection",ExpressionUUID->"b2e7d809-e79b-40eb-b5c7-67be3637aff2",
 CellID->974724367],
Cell[19093095, 435239, 176, 3, 27, "Input",ExpressionUUID->"de286e65-fab5-43ce-968c-ead467d81ebc",
 CellID->1187176853],
Cell[19093274, 435244, 213, 5, 27, "Input",ExpressionUUID->"96c5f748-1e1a-45f3-b101-b8f49f646f92",
 CellID->1392415406],
Cell[CellGroupData[{
Cell[19093512, 435253, 456, 12, 40, "MathCaption",ExpressionUUID->"c6b5777d-d64a-4813-813a-6f74e12eff27",
 CellID->1196427283],
Cell[19093971, 435267, 289, 7, 27, "Input",ExpressionUUID->"266015e3-b74d-46d0-950a-4b142c0021bd",
 CellID->1463958520],
Cell[19094263, 435276, 508, 13, 27, "Input",ExpressionUUID->"33098eef-56b6-45d1-8585-6ca0a8f5a5ef",
 CellID->1030701961],
Cell[19094774, 435291, 304, 7, 27, "Input",ExpressionUUID->"3167c475-8992-4e7a-aa9c-464ece0630e0",
 CellID->1404761825],
Cell[19095081, 435300, 454, 9, 27, "Input",ExpressionUUID->"4962a93d-4458-4ae0-9de5-2250b5cd75b7",
 CellID->396777657]
}, Open  ]],
Cell[CellGroupData[{
Cell[19095572, 435314, 232, 4, 40, "MathCaption",ExpressionUUID->"fce8ae7b-2aeb-4f91-8181-b723b829b173",
 CellID->705357264],
Cell[19095807, 435320, 2750, 81, 142, "Input",ExpressionUUID->"eebe3900-afe0-4e53-8286-eeea3135ccdf",
 CellID->829804608],
Cell[19098560, 435403, 1032, 30, 85, "Input",ExpressionUUID->"eadafc14-b002-4c4f-b506-e839b9eb2021",
 CellID->767243038]
}, Open  ]],
Cell[CellGroupData[{
Cell[19099629, 435438, 173, 3, 40, "MathCaption",ExpressionUUID->"78c63d67-c387-47bb-8f28-36872f52e9fd",
 CellID->1561923078],
Cell[19099805, 435443, 1268, 36, 138, "Input",ExpressionUUID->"0ee4cd7e-0823-40ff-b1cf-01bf1ae77b37",
 CellID->401071018],
Cell[19101076, 435481, 1212, 36, 27, "Input",ExpressionUUID->"73eb0a69-aae4-412d-8fbd-46f3ac7c6bf9",
 CellID->1996915240],
Cell[19102291, 435519, 3689, 107, 85, "Input",ExpressionUUID->"c58770a4-d601-41e4-97bb-9e70fc59534c",
 CellID->1280003364],
Cell[19105983, 435628, 9811, 269, 256, "Input",ExpressionUUID->"e32c445f-68b4-4919-aeda-1faec84002ab",
 CellID->1672846749]
}, Open  ]],
Cell[CellGroupData[{
Cell[19115831, 435902, 160, 2, 40, "MathCaption",ExpressionUUID->"7fd54161-df58-48d3-9032-561a14edf374",
 CellID->679403919],
Cell[19115994, 435906, 3612, 97, 142, "Input",ExpressionUUID->"46913206-0630-4731-9b01-c66e2d3145e9",
 CellID->571115534]
}, Open  ]],
Cell[CellGroupData[{
Cell[19119643, 436008, 158, 2, 40, "MathCaption",ExpressionUUID->"ebccc0df-c8a5-4ccc-a07c-6d4284ae6898",
 CellID->1634361495],
Cell[19119804, 436012, 2167, 59, 104, "Input",ExpressionUUID->"84e19b4a-47c8-45af-a143-684cc5d77aea",
 CellID->250317688]
}, Open  ]],
Cell[CellGroupData[{
Cell[19122008, 436076, 291, 5, 56, "MathCaption",ExpressionUUID->"a2db18f3-c19e-41ca-bbb6-8ed360bbc62c",
 CellID->168917743],
Cell[CellGroupData[{
Cell[19122324, 436085, 3662, 117, 144, "Input",ExpressionUUID->"aee4f8a8-513a-46e1-b591-94efdbcf2d2c",
 CellID->27125267],
Cell[19125989, 436204, 18579, 515, 108, "Output",ExpressionUUID->"1e3fbe4b-60e9-467d-aaf1-7ede1cdbd0bb",
 CellID->880666121]
}, Open  ]],
Cell[CellGroupData[{
Cell[19144605, 436724, 604, 18, 27, "Input",ExpressionUUID->"701307c1-2863-432b-b0f4-4a2540b82e59",
 CellID->1339971424],
Cell[19145212, 436744, 15478, 513, 105, "Output",ExpressionUUID->"41b8ccc5-59b7-4c04-b42a-8388352a5717",
 CellID->898200523]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[19160739, 437263, 219, 4, 40, "MathCaption",ExpressionUUID->"9487987c-cfc6-4407-aea0-c32b6fccfbb4",
 CellID->618494157],
Cell[CellGroupData[{
Cell[19160983, 437271, 534, 15, 47, "Input",ExpressionUUID->"57f154de-f93f-4db2-984b-34f2fc99de4b",
 CellID->1210445219],
Cell[19161520, 437288, 175, 3, 26, "Output",ExpressionUUID->"834da0fc-7cad-455e-af2b-ab2e6f8fef1c",
 CellID->1393635298]
}, Open  ]],
Cell[CellGroupData[{
Cell[19161732, 437296, 310, 8, 27, "Input",ExpressionUUID->"8f68c92f-b78a-4d9e-8605-c3b702e46e58",
 CellID->1665868489],
Cell[19162045, 437306, 1875, 71, 40, "Output",ExpressionUUID->"8422051a-82b6-410b-a367-1ccbad920a1c",
 CellID->224315657]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[19163969, 437383, 201, 4, 40, "MathCaption",ExpressionUUID->"430adfef-7997-46c4-8462-535419b12780",
 CellID->817028329],
Cell[CellGroupData[{
Cell[19164195, 437391, 435, 14, 27, "Input",ExpressionUUID->"f80b0b53-ec19-4cb9-b604-45fec62f4168",
 CellID->784686585],
Cell[19164633, 437407, 41249, 1005, 116, "Output",ExpressionUUID->"56ae631b-b6bd-4515-9ca3-27e98f05ee6b",
 CellID->1239446714]
}, Open  ]]
}, Open  ]],
Cell[19205909, 438416, 382, 6, 56, "MathCaption",ExpressionUUID->"be3e1cc2-e665-4ea6-bb9b-45b2238b8380",
 CellID->1667552426],
Cell[CellGroupData[{
Cell[19206316, 438426, 132, 1, 40, "MathCaption",ExpressionUUID->"7d714ec5-837a-4ae4-b606-4c9279f78034",
 CellID->1892030432],
Cell[CellGroupData[{
Cell[19206473, 438431, 350, 9, 27, "Input",ExpressionUUID->"d36ba079-44c6-4447-ba46-9a610a3b921d",
 CellID->1565109495],
Cell[19206826, 438442, 225, 5, 26, "Output",ExpressionUUID->"c87d4638-d419-4e6e-b5ce-8d07311ffb59",
 CellID->789546298]
}, Open  ]],
Cell[CellGroupData[{
Cell[19207088, 438452, 353, 10, 27, "Input",ExpressionUUID->"c14e3612-7757-43aa-9001-abb8e2296c5e",
 CellID->374022728],
Cell[19207444, 438464, 226, 5, 26, "Output",ExpressionUUID->"63a5ef3e-2285-4fdf-a7a0-841e7b4accdc",
 CellID->603175673]
}, Open  ]],
Cell[19207685, 438472, 209, 5, 27, "Input",ExpressionUUID->"08542a6b-b916-466f-9816-a65051593408",
 CellID->765348338],
Cell[19207897, 438479, 205, 5, 27, "Input",ExpressionUUID->"c3cbe275-2ab9-4d08-91b5-5b195a1a285c",
 CellID->1929954883]
}, Open  ]],
Cell[CellGroupData[{
Cell[19208139, 438489, 141, 1, 40, "MathCaption",ExpressionUUID->"0a8c8845-244d-451d-af98-e1f79b7123e9",
 CellID->1158714293],
Cell[CellGroupData[{
Cell[19208305, 438494, 219, 6, 27, "Input",ExpressionUUID->"cd667a72-cc9b-4186-b553-a8eab2ee2276",
 CellID->1129411455],
Cell[19208527, 438502, 7165786, 149710, 82, "Output",ExpressionUUID->"15d46692-20d7-4c91-9daa-fabb583c3f52",
 CellID->1450722271]
}, Open  ]]
}, Open  ]],
Cell[26374340, 588216, 147, 1, 40, "MathCaption",ExpressionUUID->"70ff9298-8303-4242-95ac-2d60bea7aacd",
 CellID->1490901547],
Cell[CellGroupData[{
Cell[26374512, 588221, 111, 1, 40, "MathCaption",ExpressionUUID->"c9a0de83-7a7b-4f07-84b6-3a2335884936",
 CellID->743107233],
Cell[CellGroupData[{
Cell[26374648, 588226, 366, 11, 27, "Input",ExpressionUUID->"3e154b54-1ea6-4ab9-b8f1-389a94535ab5",
 CellID->1785857726],
Cell[26375017, 588239, 142191, 4301, 796, "Output",ExpressionUUID->"03fd2137-5c1c-4276-85b5-54bfb1c25cf8",
 CellID->1907364306]
}, Closed]],
Cell[26517223, 592543, 206, 5, 23, "Input",ExpressionUUID->"70e0b9f1-33d9-44ed-b558-cdfb98b97f52",
 CellID->464768349],
Cell[26517432, 592550, 210, 5, 27, "Input",ExpressionUUID->"334e1ca6-d49e-4a99-a9a7-24ac3979b87e",
 CellID->963617504],
Cell[CellGroupData[{
Cell[26517667, 592559, 194, 4, 27, "Input",ExpressionUUID->"77dfc849-340b-4be2-85e9-40db5a00b02c",
 CellID->1345207701],
Cell[26517864, 592565, 223, 5, 26, "Output",ExpressionUUID->"79aaa6ef-a506-4a55-a7b8-367354dfae3b",
 CellID->1583492173]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[26518136, 592576, 516, 9, 72, "MathCaption",ExpressionUUID->"e1b622f2-7fb8-43f0-aecd-df742f9b3953",
 CellID->1251325234],
Cell[CellGroupData[{
Cell[26518677, 592589, 594, 15, 47, "Input",ExpressionUUID->"e4f1af23-68e3-4b32-ba51-027e34a93953",
 CellID->715852767],
Cell[26519274, 592606, 2194, 82, 26, "Output",ExpressionUUID->"6320df60-43ef-42ac-b4ec-6e3c8f9681b9",
 CellID->2135597987]
}, Open  ]],
Cell[26521483, 592691, 397, 9, 27, "Input",ExpressionUUID->"db45c0c8-3c5b-4534-a076-5e753cd280ca",
 CellID->1304264200],
Cell[CellGroupData[{
Cell[26521905, 592704, 547, 17, 66, "Input",ExpressionUUID->"e4c40c8b-cda8-47bf-90c5-6be056415730",
 CellID->1402824471],
Cell[26522455, 592723, 211, 4, 27, "Output",ExpressionUUID->"ce0388cd-ee89-4876-bba1-72b217778c43",
 CellID->2101175198]
}, Open  ]],
Cell[CellGroupData[{
Cell[26522703, 592732, 547, 17, 66, "Input",ExpressionUUID->"220e1095-a8f3-48e6-9246-6139b73c863c",
 CellID->146989363],
Cell[26523253, 592751, 210, 4, 27, "Output",ExpressionUUID->"d5c77961-076d-4f3f-8f5d-83672ca7fe6c",
 CellID->1015765640]
}, Open  ]],
Cell[CellGroupData[{
Cell[26523500, 592760, 548, 17, 66, "Input",ExpressionUUID->"2e71067e-a939-403c-8acb-1be6522dfe2f",
 CellID->1480889602],
Cell[26524051, 592779, 209, 4, 27, "Output",ExpressionUUID->"46e70369-607a-42c6-86b6-8d3127db1376",
 CellID->1625601477]
}, Open  ]],
Cell[CellGroupData[{
Cell[26524297, 592788, 548, 17, 66, "Input",ExpressionUUID->"60debcd5-14be-40e3-8dc3-fbda0c7ffab3",
 CellID->1570441424],
Cell[26524848, 592807, 210, 4, 27, "Output",ExpressionUUID->"230db245-d890-4607-8af0-c0506ae89988",
 CellID->1238064068]
}, Open  ]],
Cell[CellGroupData[{
Cell[26525095, 592816, 547, 17, 66, "Input",ExpressionUUID->"69bad4bd-e2b2-4050-8a3c-2daa7612c904",
 CellID->364957946],
Cell[26525645, 592835, 192, 3, 27, "Output",ExpressionUUID->"100dccf5-0c60-454d-b9d5-b74624f712c5",
 CellID->662439540]
}, Open  ]],
Cell[CellGroupData[{
Cell[26525874, 592843, 548, 17, 66, "Input",ExpressionUUID->"92a86a93-c3a7-41f9-9b73-17978e31ed5f",
 CellID->2109071834],
Cell[26526425, 592862, 211, 4, 70, "Output",ExpressionUUID->"51bb3aa7-4131-4543-8c5d-5fde59787a57",
 CellID->1066122732]
}, Open  ]],
Cell[26526651, 592869, 621, 11, 66, "Input",ExpressionUUID->"7997daac-4313-4c63-8c3d-c3417471f872",
 CellID->170533998],
Cell[26527275, 592882, 458, 9, 27, "Input",ExpressionUUID->"a5f8eacb-4cf6-4d39-8734-18814ebdf7ce",
 CellID->1421719288],
Cell[26527736, 592893, 307, 7, 27, "Input",ExpressionUUID->"1196794c-d6ab-4127-8694-67db0c4689b2",
 CellID->1933016894],
Cell[26528046, 592902, 165, 3, 27, "Input",ExpressionUUID->"a80d6460-6d42-4607-8b2e-ced7e0950a96",
 CellID->1082325642],
Cell[26528214, 592907, 167, 3, 27, "Input",ExpressionUUID->"08a737fb-26d8-44ae-920c-5e2b9cabce40",
 CellID->1460575004],
Cell[26528384, 592912, 171, 3, 27, "Input",ExpressionUUID->"90fef4e7-c30b-4e34-bf24-9f78ae378122",
 CellID->2146007018]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[26528604, 592921, 226, 3, 41, "Subsubsection",ExpressionUUID->"56d80cc5-0f35-4776-b875-c37d38b2db67",
 CellID->952100161],
Cell[CellGroupData[{
Cell[26528855, 592928, 175, 3, 27, "Input",ExpressionUUID->"faa52d1c-cdee-4d8e-8089-5afc813acbcf",
 CellID->1046752130],
Cell[CellGroupData[{
Cell[26529055, 592935, 259, 4, 22, "Print",ExpressionUUID->"3c159864-2e8e-4126-87d5-61f5bb03fd86",
 CellID->263638693],
Cell[26529317, 592941, 531, 12, 22, "Print",ExpressionUUID->"8097e80d-55b1-4b05-a75e-48060719b801",
 CellID->716749490],
Cell[26529851, 592955, 279, 4, 22, "Print",ExpressionUUID->"ad3af31f-d0cc-431b-aec5-4c9d4b027cb0",
 CellID->1172902407],
Cell[26530133, 592961, 236, 3, 22, "Print",ExpressionUUID->"8b117629-37b1-4e63-a2d7-fbd3df66c258",
 CellID->1218358636],
Cell[26530372, 592966, 218, 3, 22, "Print",ExpressionUUID->"e40b0a8e-958e-4c0a-837d-a0f6c7d0f168",
 CellID->697021107],
Cell[26530593, 592971, 257, 4, 22, "Print",ExpressionUUID->"882bc889-0e8e-4ba0-8c63-9e7e25ef0970",
 CellID->310450964],
Cell[26530853, 592977, 534, 12, 22, "Print",ExpressionUUID->"7b60ecce-49fb-4b1a-be6e-ad81ad6d5f34",
 CellID->1511573385],
Cell[26531390, 592991, 278, 4, 22, "Print",ExpressionUUID->"40c0aa33-7d3d-4ac3-924f-32868e46d9fd",
 CellID->282337685],
Cell[26531671, 592997, 257, 4, 22, "Print",ExpressionUUID->"83d26627-c635-43a8-acb4-46ba1d709bbf",
 CellID->1458095572],
Cell[26531931, 593003, 530, 12, 22, "Print",ExpressionUUID->"4ece7435-28a2-4e7c-8f00-ef5ca9cc43c0",
 CellID->1276382687],
Cell[26532464, 593017, 296, 4, 22, "Print",ExpressionUUID->"3e89e572-0f5b-4f3c-bebe-c7fab7fd8cfe",
 CellID->949670526],
Cell[26532763, 593023, 258, 4, 22, "Print",ExpressionUUID->"ff064016-153c-4e90-91aa-1c6a472ae061",
 CellID->1014065978],
Cell[26533024, 593029, 414, 6, 42, "Print",ExpressionUUID->"853374a3-e1d4-4762-be92-41c9b2161963",
 CellID->976669654],
Cell[26533441, 593037, 258, 4, 22, "Print",ExpressionUUID->"d64f7793-9f6f-4684-9e63-5f41b6fa71d8",
 CellID->1669360757]
}, Open  ]]
}, Open  ]],
Cell[26533726, 593045, 211, 5, 27, "Input",ExpressionUUID->"9bd1ce9e-069e-42fa-badb-77b87918a05b",
 CellID->577012686],
Cell[26533940, 593052, 350, 8, 27, "Input",ExpressionUUID->"a66a252b-8335-4871-bb34-019d9e6d4559",
 CellID->1963547397],
Cell[26534293, 593062, 238, 5, 27, "Input",ExpressionUUID->"915b3f72-68bc-4248-ad34-53a7e94424d1",
 CellID->959897930],
Cell[CellGroupData[{
Cell[26534556, 593071, 146, 1, 40, "MathCaption",ExpressionUUID->"6c23bcdf-2d3a-4372-9939-d28da3bec2bd",
 CellID->1107353246],
Cell[26534705, 593074, 461, 12, 27, "Input",ExpressionUUID->"9402bb28-22a0-4c59-ad99-3b252a4d7abe",
 CellID->1192720620],
Cell[CellGroupData[{
Cell[26535191, 593090, 866, 27, 27, "Input",ExpressionUUID->"26a257d7-cde5-414e-9718-53ab684cc3b8",
 CellID->1443099242],
Cell[26536060, 593119, 2543, 83, 98, "Output",ExpressionUUID->"25d210f7-2204-4082-be97-a08e1e93e8b2",
 CellID->1906380320]
}, Open  ]],
Cell[26538618, 593205, 186, 4, 27, "Input",ExpressionUUID->"4fc36915-6394-4fda-8f8a-44fb88398de6",
 CellID->781168204],
Cell[CellGroupData[{
Cell[26538829, 593213, 330, 9, 27, "Input",ExpressionUUID->"495665e6-f8b8-4a0c-830e-a37ae6f51b4c",
 CellID->1526470517],
Cell[26539162, 593224, 2485, 80, 48, "Output",ExpressionUUID->"7327228f-d07d-4e20-8ee7-4a8912ad84ea",
 CellID->1146005907]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[26541696, 593310, 142, 1, 40, "MathCaption",ExpressionUUID->"cc5e26c8-d717-4df6-8c0b-eac693cbae8b",
 CellID->1013473648],
Cell[26541841, 593313, 1317, 37, 47, "Input",ExpressionUUID->"7ce966ce-fc2b-48a1-a515-50d76c95ca0d",
 CellID->1758797579],
Cell[CellGroupData[{
Cell[26543183, 593354, 470, 14, 27, "Input",ExpressionUUID->"92492632-eddf-4da6-be59-664046b06a9d",
 CellID->2145992945],
Cell[26543656, 593370, 742, 20, 26, "Output",ExpressionUUID->"0c833855-7cbe-49ca-a7df-837e39930d50",
 CellID->670912572]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[26544447, 593396, 190, 4, 40, "MathCaption",ExpressionUUID->"7ea71d3e-5e66-408f-bacf-bee42758599b",
 CellID->764803550],
Cell[CellGroupData[{
Cell[26544662, 593404, 242, 6, 27, "Input",ExpressionUUID->"13eb0201-8dab-4aa5-a2fb-143dce550da4",
 CellID->1363637555],
Cell[26544907, 593412, 250, 5, 26, "Output",ExpressionUUID->"4ec61e03-c8df-476f-aa03-f0d28a5e5669",
 CellID->1696935750]
}, Open  ]],
Cell[CellGroupData[{
Cell[26545194, 593422, 243, 6, 27, "Input",ExpressionUUID->"30aed816-b889-4944-98ad-0e0de7ce4194",
 CellID->1747311246],
Cell[26545440, 593430, 252, 5, 26, "Output",ExpressionUUID->"5d18c90c-e6a8-4e98-9947-a42f96f2141f",
 CellID->1274418663]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[26545741, 593441, 121, 1, 40, "MathCaption",ExpressionUUID->"6517bff2-9be7-4a0f-8667-37a79c57aeb7",
 CellID->1066268222],
Cell[CellGroupData[{
Cell[26545887, 593446, 231, 6, 27, "Input",ExpressionUUID->"e5c05129-689d-4c0d-92c2-601742d2fc52",
 CellID->18972939],
Cell[26546121, 593454, 9563, 256, 90, "Output",ExpressionUUID->"bd628d10-672d-4746-adb7-d3fdc6d10cdb",
 CellID->659278950]
}, Open  ]],
Cell[CellGroupData[{
Cell[26555721, 593715, 233, 6, 27, "Input",ExpressionUUID->"212483cd-2d70-4693-9135-aaeac95205dd",
 CellID->924782085],
Cell[26555957, 593723, 5064, 142, 68, "Output",ExpressionUUID->"790ddc3e-d333-4db8-b160-a96c02c5da8b",
 CellID->1880112572]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[26561070, 593871, 118, 1, 40, "MathCaption",ExpressionUUID->"597bda04-dc91-476d-b62d-e2e9d08bd00b",
 CellID->2094495682],
Cell[CellGroupData[{
Cell[26561213, 593876, 224, 5, 27, "Input",ExpressionUUID->"fe5c6a00-1ae0-4219-8d31-8858543f85c4",
 CellID->636985750],
Cell[26561440, 593883, 9818, 263, 94, "Output",ExpressionUUID->"aaf1e2cd-b499-4413-b14d-e737d4b1d6d0",
 CellID->1563000292]
}, Open  ]],
Cell[CellGroupData[{
Cell[26571295, 594151, 226, 5, 27, "Input",ExpressionUUID->"8c3079b8-f68b-4153-82d0-77a087067af5",
 CellID->1360894438],
Cell[26571524, 594158, 5062, 142, 68, "Output",ExpressionUUID->"1473b186-5ccd-4039-8879-d3a94fb62ca4",
 CellID->1381377051]
}, Open  ]]
}, Open  ]],
Cell[26576613, 594304, 241, 4, 40, "MathCaption",ExpressionUUID->"43ac862b-5903-40e3-8eba-14fc758cdbf4",
 CellID->581661594],
Cell[CellGroupData[{
Cell[26576879, 594312, 125, 1, 40, "MathCaption",ExpressionUUID->"6e3ce83b-771d-4588-a4ff-7ad612c301fd",
 CellID->640849214],
Cell[26577007, 594315, 213, 5, 27, "Input",ExpressionUUID->"ca105fad-e10d-4665-965f-d7c0d40d4057",
 CellID->1122249993],
Cell[CellGroupData[{
Cell[26577245, 594324, 160, 3, 27, "Input",ExpressionUUID->"303c57a0-6b1d-4be4-95e6-d21659a0de53",
 CellID->306760750],
Cell[26577408, 594329, 195, 3, 26, "Output",ExpressionUUID->"81321768-cd78-4d1b-9c52-50c5ecd6eac8",
 CellID->2055329345]
}, Open  ]],
Cell[26577618, 594335, 215, 5, 27, "Input",ExpressionUUID->"fcb29837-a65f-486d-bbb5-c65519f57027",
 CellID->1168529140],
Cell[CellGroupData[{
Cell[26577858, 594344, 161, 3, 27, "Input",ExpressionUUID->"741b615b-a030-45fc-b61e-4950a89bafa9",
 CellID->731304165],
Cell[26578022, 594349, 195, 3, 26, "Output",ExpressionUUID->"e8646506-ee06-400a-a3b9-b89feabdf16a",
 CellID->1866202543]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[26578266, 594358, 176, 3, 40, "MathCaption",ExpressionUUID->"e61dbd88-e7f5-4caa-9c8e-21fff0feb3af",
 CellID->695752681],
Cell[CellGroupData[{
Cell[26578467, 594365, 306, 8, 27, "Input",ExpressionUUID->"48ed3467-55eb-497d-8a1d-46641487950c",
 CellID->466851652],
Cell[26578776, 594375, 195, 3, 26, "Output",ExpressionUUID->"4ea02165-a640-475f-93e5-2e81ae4ccacb",
 CellID->49490229]
}, Open  ]],
Cell[CellGroupData[{
Cell[26579008, 594383, 311, 8, 27, "Input",ExpressionUUID->"f28db7b6-75dc-4812-abd9-ac7aa84d7d58",
 CellID->1492636797],
Cell[26579322, 594393, 8219, 250, 197, "Output",ExpressionUUID->"cb74294e-6d80-4585-bde6-ffb08bb27165",
 CellID->1252467486]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[26587590, 594649, 181, 3, 40, "MathCaption",ExpressionUUID->"ba670c55-95e3-44d5-8c69-f61a70a12332",
 CellID->1671773579],
Cell[CellGroupData[{
Cell[26587796, 594656, 316, 9, 27, "Input",ExpressionUUID->"ca96705e-c2da-4175-8336-dd8a3b1379b5",
 CellID->1969988552],
Cell[26588115, 594667, 194, 3, 26, "Output",ExpressionUUID->"4f504605-7124-41c4-93c0-a71294efc870",
 CellID->1867085882]
}, Open  ]],
Cell[CellGroupData[{
Cell[26588346, 594675, 318, 9, 27, "Input",ExpressionUUID->"2085d3c3-326a-4bc5-8542-f382f0215506",
 CellID->1712486856],
Cell[26588667, 594686, 190, 3, 26, "Output",ExpressionUUID->"1fcd15f7-6363-48ac-ae58-46340e9c3a51",
 CellID->782550271]
}, Open  ]],
Cell[CellGroupData[{
Cell[26588894, 594694, 316, 9, 27, "Input",ExpressionUUID->"e741c133-a068-40fe-9100-ed07c761ac0a",
 CellID->563681893],
Cell[26589213, 594705, 190, 3, 26, "Output",ExpressionUUID->"0d392747-cf75-4aa3-9221-637e02b01cfb",
 CellID->36108568]
}, Open  ]],
Cell[CellGroupData[{
Cell[26589440, 594713, 316, 9, 27, "Input",ExpressionUUID->"238b8407-a703-41fd-922e-faf089a1c531",
 CellID->905320914],
Cell[26589759, 594724, 190, 3, 26, "Output",ExpressionUUID->"be56381b-c435-46fc-894b-ad04eddec675",
 CellID->43528326]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[26589998, 594733, 107, 1, 40, "MathCaption",ExpressionUUID->"f5a74e04-693b-4706-a1e8-4702e6c08a5b",
 CellID->1261237816],
Cell[CellGroupData[{
Cell[26590130, 594738, 401, 11, 27, "Input",ExpressionUUID->"aa04a91a-6dd0-4f3d-a80b-239e10e2683d",
 CellID->1960377622],
Cell[26590534, 594751, 128901, 3244, 318, "Output",ExpressionUUID->"55f04e4a-64f1-4413-8ed9-c27863a555e5",
 CellID->1028024297]
}, Open  ]],
Cell[CellGroupData[{
Cell[26719472, 598000, 402, 11, 27, "Input",ExpressionUUID->"64716f40-7639-439c-adde-9b2c258a567f",
 CellID->694152686],
Cell[26719877, 598013, 107570, 2750, 312, "Output",ExpressionUUID->"402cb11a-31ec-412b-92a2-2e3c26c7254c",
 CellID->1514571602]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[26827496, 600769, 150, 1, 40, "MathCaption",ExpressionUUID->"5c37ed58-3014-4275-b198-eb2117c531bd",
 CellID->2040276538],
Cell[CellGroupData[{
Cell[26827671, 600774, 373, 11, 27, "Input",ExpressionUUID->"90a53eef-2a71-4209-9583-6cf7f128e157",
 CellID->2056382517],
Cell[26828047, 600787, 195, 3, 26, "Output",ExpressionUUID->"feb3adcc-7438-4e7e-8161-eccd7ca185af",
 CellID->1100047659]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[26828291, 600796, 113, 1, 40, "MathCaption",ExpressionUUID->"d6757a56-afb4-446d-8cf5-97048ef9d096",
 CellID->1852112204],
Cell[CellGroupData[{
Cell[26828429, 600801, 158, 3, 27, "Input",ExpressionUUID->"e2b5c0cb-2d4a-4314-a2be-59fc2c98028e",
 CellID->674349363],
Cell[26828590, 600806, 194, 3, 26, "Output",ExpressionUUID->"6a3575e5-251d-46ca-8b39-51a88bd182c5",
 CellID->16490870]
}, Open  ]],
Cell[26828799, 600812, 248, 4, 27, "Input",ExpressionUUID->"167c5b18-1e0b-440a-9861-3811c560d19c",
 CellID->2127000328],
Cell[26829050, 600818, 166, 3, 27, "Input",ExpressionUUID->"b06347fc-a6a5-4acb-9d25-152ede547029",
 CellID->649493526],
Cell[26829219, 600823, 165, 3, 27, "Input",ExpressionUUID->"28da1443-3736-4b8b-8e15-18c39a213d6d",
 CellID->378313781],
Cell[26829387, 600828, 278, 6, 27, "Input",ExpressionUUID->"f5e56003-7ebd-4f46-9171-8c37c6d67672",
 CellID->1768122147],
Cell[26829668, 600836, 175, 3, 27, "Input",ExpressionUUID->"081656fe-2706-45bd-a18c-c99dd98bdae1",
 CellID->2097701090],
Cell[26829846, 600841, 166, 3, 27, "Input",ExpressionUUID->"66f00f8f-835f-41e5-bc20-bd014179d2b2",
 CellID->823186831],
Cell[26830015, 600846, 170, 3, 27, "Input",ExpressionUUID->"7b7289df-49d5-4ac6-acb8-90c8727545e3",
 CellID->1211414643]
}, Open  ]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[26830246, 600856, 285, 3, 39, "Subsection",ExpressionUUID->"cab48f9c-968b-4465-8703-4487b0c4d6ca",
 CellID->2100931341],
Cell[26830534, 600861, 418, 10, 23, "Text",ExpressionUUID->"001e0087-9c76-4dbe-ba2a-c0614c0444d3",
 CellID->1426924213],
Cell[CellGroupData[{
Cell[26830977, 600875, 216, 3, 41, "Subsubsection",ExpressionUUID->"c50e8747-19c6-4348-8c1f-05ddea1300c5",
 CellID->1365361842],
Cell[26831196, 600880, 177, 3, 27, "Input",ExpressionUUID->"bee02021-9c7b-4ab6-a085-9a6dbc657429",
 CellID->1773710299],
Cell[26831376, 600885, 310, 7, 47, "Input",ExpressionUUID->"e52adf5e-dad0-40dc-a10d-6fe1ca7f1b5f",
 CellID->1528491286]
}, Closed]],
Cell[CellGroupData[{
Cell[26831723, 600897, 265, 5, 41, "Subsubsection",ExpressionUUID->"c69ee9a4-3b2d-4d75-ab06-81adee5a4efa",
 CellID->1153658388],
Cell[26831991, 600904, 290, 7, 27, "Input",ExpressionUUID->"88af7186-ceb1-4bd6-9110-d14c58513b10",
 CellID->2054162000],
Cell[26832284, 600913, 466, 12, 27, "Input",ExpressionUUID->"ebc9095c-0632-4a34-9769-5625a994fdc6",
 CellID->1353624828],
Cell[26832753, 600927, 185, 3, 27, "Input",ExpressionUUID->"52f9f81c-9e5c-48d0-b831-4c1555a51bf7",
 CellID->2078211525],
Cell[26832941, 600932, 3003, 91, 66, "Input",ExpressionUUID->"391758a9-c29b-4a14-9589-ab34dd4d44c6",
 CellID->1827065867],
Cell[CellGroupData[{
Cell[26835969, 601027, 217, 6, 27, "Input",ExpressionUUID->"e2c09bf5-3fae-4f1e-818c-2d381b0b25de",
 CellID->627358216],
Cell[26836189, 601035, 17666, 456, 94, "Output",ExpressionUUID->"0215dc22-76d5-4bca-808e-c60fe4b9a4d7",
 CellID->1740023790]
}, Open  ]],
Cell[26853870, 601494, 184, 4, 27, "Input",ExpressionUUID->"47b29c6c-df67-4651-b4a2-f3ff8e9ddd74",
 CellID->227810920],
Cell[CellGroupData[{
Cell[26854079, 601502, 344, 9, 27, "Input",ExpressionUUID->"7f57fad2-ff0e-4ebe-a03d-596779c4e2fc",
 CellID->2002273486],
Cell[26854426, 601513, 227, 5, 26, "Output",ExpressionUUID->"3f685232-cf09-4b6f-a35a-3f08c5f3a762",
 CellID->147533684]
}, Open  ]],
Cell[CellGroupData[{
Cell[26854690, 601523, 412, 11, 27, "Input",ExpressionUUID->"95c52da4-9879-4da4-a3e3-004c6e4ac249",
 CellID->1301690639],
Cell[26855105, 601536, 227, 5, 26, "Output",ExpressionUUID->"fff86c91-68fb-4d37-88c8-767a1908e2a3",
 CellID->1402498350]
}, Open  ]],
Cell[CellGroupData[{
Cell[26855369, 601546, 192, 4, 27, "Input",ExpressionUUID->"78d02b5e-23f2-495a-9918-7c7a9018dbe2",
 CellID->1075298082],
Cell[26855564, 601552, 36600, 1035, 331, "Output",ExpressionUUID->"cb1498ab-b4e9-4a58-b962-970af5890c9c",
 CellID->1009258978]
}, Open  ]],
Cell[CellGroupData[{
Cell[26892201, 602592, 161, 3, 27, "Input",ExpressionUUID->"d1cd0e03-ac6a-4176-a280-630e1649898a",
 CellID->646111261],
Cell[26892365, 602597, 174, 3, 26, "Output",ExpressionUUID->"42eb32d1-b256-4ec5-8f87-6a36b143cd8e",
 CellID->1789799240]
}, Open  ]],
Cell[CellGroupData[{
Cell[26892576, 602605, 195, 4, 27, "Input",ExpressionUUID->"e9dfd15f-74a8-44c4-8386-665c0dafc540",
 CellID->357195227],
Cell[26892774, 602611, 174, 3, 26, "Output",ExpressionUUID->"5f0aa0d2-0935-405f-ab19-f14705e93f86",
 CellID->1352383151]
}, Open  ]],
Cell[CellGroupData[{
Cell[26892985, 602619, 226, 5, 27, "Input",ExpressionUUID->"7389ebdf-b398-4505-b633-02628f487f51",
 CellID->203745033],
Cell[26893214, 602626, 173, 3, 26, "Output",ExpressionUUID->"b2a5d210-7868-4b44-b53b-b9f49d2a28a9",
 CellID->536699009]
}, Open  ]],
Cell[26893402, 602632, 186, 4, 27, "Input",ExpressionUUID->"161e63d5-b840-4c45-9aa7-393360289890",
 CellID->2117368227],
Cell[26893591, 602638, 165, 3, 27, "Input",ExpressionUUID->"3b8069d0-a689-46bd-8f77-4b583ac218d5",
 CellID->656276752],
Cell[26893759, 602643, 186, 3, 27, "Input",ExpressionUUID->"034a186e-5920-45be-ae35-50684d69fd25",
 CellID->567359872],
Cell[26893948, 602648, 166, 3, 27, "Input",ExpressionUUID->"e52d83c7-1c66-48f6-82fc-fea259a3ae05",
 CellID->789976024],
Cell[26894117, 602653, 171, 3, 27, "Input",ExpressionUUID->"1b9cf383-82d4-4314-8f59-54b2be2cb1df",
 CellID->1383129614]
}, Closed]],
Cell[CellGroupData[{
Cell[26894325, 602661, 213, 3, 41, "Subsubsection",ExpressionUUID->"151cf641-9819-4eff-af98-759a6ca08fc8",
 CellID->1896427370],
Cell[26894541, 602666, 289, 7, 27, "Input",ExpressionUUID->"187f3e31-d974-4787-9a43-6e9c614b46f8",
 CellID->302518816],
Cell[26894833, 602675, 461, 12, 27, "Input",ExpressionUUID->"d55e052f-6448-4736-99ec-5b83207ea047",
 CellID->1339948063],
Cell[26895297, 602689, 174, 3, 27, "Input",ExpressionUUID->"5e3fbd19-869e-407f-9769-e3c1214783b9",
 CellID->1993572399],
Cell[CellGroupData[{
Cell[26895496, 602696, 1044, 34, 27, "Input",ExpressionUUID->"737aeb28-7bfc-4716-bff5-b6d0ae682e3d",
 CellID->931449499],
Cell[26896543, 602732, 2545, 82, 48, "Output",ExpressionUUID->"c3d00fff-b6d9-4b67-8484-f479cf368016",
 CellID->1325426988]
}, Open  ]],
Cell[26899103, 602817, 185, 4, 27, "Input",ExpressionUUID->"809eef50-09e4-4c01-b251-ae18d110db32",
 CellID->1858690769],
Cell[CellGroupData[{
Cell[26899313, 602825, 352, 9, 27, "Input",ExpressionUUID->"81dcab0a-8033-4b9e-b7d2-f7b42923daa5",
 CellID->1718625321],
Cell[26899668, 602836, 226, 5, 26, "Output",ExpressionUUID->"2f0cc529-1dfe-47aa-9573-666f878a1861",
 CellID->603516985]
}, Open  ]],
Cell[CellGroupData[{
Cell[26899931, 602846, 418, 12, 27, "Input",ExpressionUUID->"f8448ed8-ddae-4ab2-9a96-c0ec6988af6b",
 CellID->797100998],
Cell[26900352, 602860, 226, 5, 26, "Output",ExpressionUUID->"86880898-bd71-4b40-a740-675432815dff",
 CellID->744702114]
}, Open  ]],
Cell[CellGroupData[{
Cell[26900615, 602870, 191, 4, 27, "Input",ExpressionUUID->"65d2847a-d5bd-4335-97ac-89fdb9888f05",
 CellID->437292908],
Cell[26900809, 602876, 12737, 378, 193, "Output",ExpressionUUID->"bbb4888b-eedb-471b-89aa-4a7cfd611da1",
 CellID->673735282]
}, Open  ]],
Cell[CellGroupData[{
Cell[26913583, 603259, 191, 4, 27, "Input",ExpressionUUID->"d4b1f57e-b6fd-40ad-b9fe-d66c35781458",
 CellID->1909141065],
Cell[26913777, 603265, 28464, 813, 539, "Output",ExpressionUUID->"915153af-806e-4f61-a506-9d3d0c3fdd11",
 CellID->1935097590]
}, Open  ]],
Cell[CellGroupData[{
Cell[26942278, 604083, 471, 15, 27, "Input",ExpressionUUID->"d33b11bc-e329-4405-a673-12a541a7b56d",
 CellID->2025911961],
Cell[26942752, 604100, 965, 35, 44, "Output",ExpressionUUID->"7d6005ec-eaeb-40a8-8716-a32ca48540e5",
 CellID->1692844962]
}, Open  ]],
Cell[CellGroupData[{
Cell[26943754, 604140, 455, 13, 27, "Input",ExpressionUUID->"81cc7b82-d62a-436b-b80d-add2d1d1bd26",
 CellID->693045523],
Cell[26944212, 604155, 227, 5, 26, "Output",ExpressionUUID->"94f1fdb2-8527-44bf-89f6-8f416538bdac",
 CellID->1781920878]
}, Open  ]],
Cell[26944454, 604163, 251, 6, 27, "Input",ExpressionUUID->"7ecb2479-c2bb-40ae-b15d-67fd600a5942",
 CellID->382131502],
Cell[CellGroupData[{
Cell[26944730, 604173, 844, 27, 47, "Input",ExpressionUUID->"837c87eb-ffc4-4992-82b7-978c8e3773f1",
 CellID->1508049584],
Cell[26945577, 604202, 226, 5, 26, "Output",ExpressionUUID->"b6e08297-3287-4ca5-b4fc-e3b0d19dbf7b",
 CellID->2084739131]
}, Open  ]],
Cell[26945818, 604210, 223, 5, 27, "Input",ExpressionUUID->"6038687a-b7f9-44c8-9c75-9e7403263ccd",
 CellID->1480866034],
Cell[26946044, 604217, 176, 3, 27, "Input",ExpressionUUID->"cd25d69e-3f9e-4d5f-95bf-dac0b3538edd",
 CellID->1727448355],
Cell[26946223, 604222, 166, 3, 27, "Input",ExpressionUUID->"e35cb880-17e6-483e-ab84-9c9f2b058a17",
 CellID->1393078692],
Cell[26946392, 604227, 167, 3, 27, "Input",ExpressionUUID->"472161a3-3fbf-4790-ad46-bb4c29562987",
 CellID->1606856850],
Cell[26946562, 604232, 171, 3, 27, "Input",ExpressionUUID->"9eab08b6-0b49-4023-808c-68de83a87caf",
 CellID->1836216925]
}, Closed]],
Cell[CellGroupData[{
Cell[26946770, 604240, 242, 3, 41, "Subsubsection",ExpressionUUID->"e6e04444-d4ff-4ebe-8c93-6e4498667d65",
 CellID->1957851085],
Cell[26947015, 604245, 290, 7, 27, "Input",ExpressionUUID->"1bd5e852-305f-464a-886f-df1cbe24b433",
 CellID->1858746261],
Cell[26947308, 604254, 460, 12, 27, "Input",ExpressionUUID->"ff77e9df-96fe-4c97-a83f-ed2f0125c5d5",
 CellID->503795718],
Cell[26947771, 604268, 182, 3, 27, "Input",ExpressionUUID->"366f814f-10d5-480f-84e9-0ff572682009",
 CellID->1914095952],
Cell[26947956, 604273, 185, 3, 27, "Input",ExpressionUUID->"59e68030-21ea-42dc-a07f-e1105daed564",
 CellID->905633204],
Cell[CellGroupData[{
Cell[26948166, 604280, 1371, 43, 27, "Input",ExpressionUUID->"b06165c4-4722-4f48-8041-806a14043f86",
 CellID->1760653339],
Cell[26949540, 604325, 3593, 114, 53, "Output",ExpressionUUID->"fed72294-041c-4b2f-aa45-5b731a880e47",
 CellID->1657532209]
}, Open  ]],
Cell[26953148, 604442, 185, 4, 27, "Input",ExpressionUUID->"7eba7489-eeec-4d64-8250-6e5fbda55a4e",
 CellID->1930459496],
Cell[CellGroupData[{
Cell[26953358, 604450, 352, 9, 27, "Input",ExpressionUUID->"a9191b72-e365-48c3-a584-50b47fcf90b4",
 CellID->1049810396],
Cell[26953713, 604461, 226, 5, 26, "Output",ExpressionUUID->"2945a4d0-2cff-4e5f-96ca-06d08730276f",
 CellID->985967779]
}, Open  ]],
Cell[CellGroupData[{
Cell[26953976, 604471, 455, 13, 27, "Input",ExpressionUUID->"6a809c07-fb09-4a62-9b44-6d24ae92bbd2",
 CellID->986857765],
Cell[26954434, 604486, 227, 5, 26, "Output",ExpressionUUID->"16125f6b-f53c-4139-b341-211aee777a86",
 CellID->1176641401]
}, Open  ]],
Cell[26954676, 604494, 211, 5, 27, "Input",ExpressionUUID->"5c7ff7db-ec68-456e-ab0d-0de915a0dba5",
 CellID->1331458196],
Cell[26954890, 604501, 210, 5, 27, "Input",ExpressionUUID->"572d3290-8132-4518-905f-b23d49442b23",
 CellID->1038298432],
Cell[CellGroupData[{
Cell[26955125, 604510, 470, 15, 27, "Input",ExpressionUUID->"dfcea39b-8367-4fc8-a0fa-18f3cbb31283",
 CellID->229631102],
Cell[26955598, 604527, 1855, 65, 50, "Output",ExpressionUUID->"1e591b60-c0b1-420a-99a5-1b4119b93feb",
 CellID->1782605031]
}, Open  ]],
Cell[CellGroupData[{
Cell[26957490, 604597, 354, 11, 27, "Input",ExpressionUUID->"d76e814d-7d08-405d-8c29-d4635f465c23",
 CellID->528190349],
Cell[26957847, 604610, 506, 17, 27, "Output",ExpressionUUID->"fc9f3533-059d-4469-9647-32ab675aa8ce",
 CellID->698396720]
}, Open  ]],
Cell[CellGroupData[{
Cell[26958390, 604632, 188, 4, 27, "Input",ExpressionUUID->"2e839399-72d9-4456-9211-e887a6b6e80c",
 CellID->485389371],
Cell[26958581, 604638, 369, 11, 26, "Output",ExpressionUUID->"801d0581-1274-4e61-9dd8-9506caa20153",
 CellID->1955542185]
}, Open  ]],
Cell[26958965, 604652, 251, 6, 27, "Input",ExpressionUUID->"89a4dc42-314b-4245-bee4-0627525aa84b",
 CellID->535888911],
Cell[CellGroupData[{
Cell[26959241, 604662, 843, 27, 47, "Input",ExpressionUUID->"7402756f-45bf-4424-b0a8-96367e5d10ef",
 CellID->138511659],
Cell[26960087, 604691, 226, 5, 26, "Output",ExpressionUUID->"37a24cc5-98a2-475f-bf6e-8d75cd512c93",
 CellID->1707769607]
}, Open  ]],
Cell[26960328, 604699, 203, 5, 27, "Input",ExpressionUUID->"e06673c1-f8c9-4c47-88e5-4ddd8c8a727f",
 CellID->551689592],
Cell[CellGroupData[{
Cell[26960556, 604708, 423, 13, 27, "Input",ExpressionUUID->"a45abe1b-f3fe-443b-832c-8772892bce28",
 CellID->1560068645],
Cell[26960982, 604723, 643, 21, 45, "Output",ExpressionUUID->"ae21f2eb-8b7f-4dd7-985f-23f63abc651b",
 CellID->1364842694]
}, Open  ]],
Cell[CellGroupData[{
Cell[26961662, 604749, 547, 17, 27, "Input",ExpressionUUID->"945870dc-f9ec-496e-b0c4-1dbf625e9009",
 CellID->1964531839],
Cell[26962212, 604768, 678, 22, 45, "Output",ExpressionUUID->"395328f1-7e55-41c3-a877-3f809aeeb167",
 CellID->1638324451]
}, Open  ]],
Cell[26962905, 604793, 236, 5, 27, "Input",ExpressionUUID->"694cb379-4a73-4fc0-ad17-358b4f8c2447",
 CellID->1801495627],
Cell[26963144, 604800, 183, 3, 27, "Input",ExpressionUUID->"11f971b8-a089-485f-b480-5d369806dc2e",
 CellID->365063936],
Cell[26963330, 604805, 187, 3, 27, "Input",ExpressionUUID->"c0eb45b7-871d-41e9-9500-2bf1b3aeefa0",
 CellID->268792873],
Cell[26963520, 604810, 166, 3, 27, "Input",ExpressionUUID->"6d5af50a-0e74-42c4-b9f0-2cef4422ef7b",
 CellID->1443706870],
Cell[26963689, 604815, 167, 3, 27, "Input",ExpressionUUID->"8e1ad2ac-a251-4e96-b671-8a778080d654",
 CellID->2111120284],
Cell[26963859, 604820, 170, 3, 27, "Input",ExpressionUUID->"3e4df76b-a434-4926-88af-04e15d5c9fba",
 CellID->873797454]
}, Closed]],
Cell[CellGroupData[{
Cell[26964066, 604828, 218, 3, 41, "Subsubsection",ExpressionUUID->"a368fb88-c5d8-4bfd-96f5-c6f267bdb37d",
 CellID->1420631192],
Cell[26964287, 604833, 289, 7, 27, "Input",ExpressionUUID->"2e73e135-9684-40c8-8478-53a497bc1833",
 CellID->825256598],
Cell[26964579, 604842, 461, 12, 27, "Input",ExpressionUUID->"4310c980-f951-4643-87b6-20f7bfafd4cb",
 CellID->1638250260],
Cell[26965043, 604856, 178, 3, 27, "Input",ExpressionUUID->"8623cc2a-cba8-4362-bcb6-b1346a5cda7d",
 CellID->1184281633],
Cell[26965224, 604861, 175, 3, 27, "Input",ExpressionUUID->"ea9a72fc-e839-4f4d-a5b8-dcec96a6cd00",
 CellID->1342965435],
Cell[26965402, 604866, 1625, 49, 47, "Input",ExpressionUUID->"ef83c15f-90c0-4641-b104-27bcf4a224c1",
 CellID->1166936075],
Cell[26967030, 604917, 185, 4, 27, "Input",ExpressionUUID->"2bd5debe-8427-4162-b6cc-65a43a5e526c",
 CellID->1873948121],
Cell[CellGroupData[{
Cell[26967240, 604925, 351, 9, 27, "Input",ExpressionUUID->"522c3bbe-1998-49c0-88d7-824d8bf8bca3",
 CellID->450902745],
Cell[26967594, 604936, 229, 5, 26, "Output",ExpressionUUID->"074d7576-e084-49f9-aaa6-6a073e2a3537",
 CellID->1412734845]
}, Open  ]],
Cell[CellGroupData[{
Cell[26967860, 604946, 456, 13, 27, "Input",ExpressionUUID->"c75e921c-77fb-458d-8778-b38d146ca932",
 CellID->1599836902],
Cell[26968319, 604961, 227, 5, 26, "Output",ExpressionUUID->"4ee09b44-cfbb-4559-b317-ae5e528a7b35",
 CellID->1255338683]
}, Open  ]],
Cell[26968561, 604969, 211, 5, 27, "Input",ExpressionUUID->"c99dda72-2311-45f8-8707-6d405cfc65af",
 CellID->1028420069],
Cell[CellGroupData[{
Cell[26968797, 604978, 129, 1, 40, "MathCaption",ExpressionUUID->"38bc0cd8-dfc4-4e18-9a2e-2f17ce865f79",
 CellID->1707708627],
Cell[CellGroupData[{
Cell[26968951, 604983, 158, 3, 27, "Input",ExpressionUUID->"c46497bf-80a6-434a-a880-b3e144867dc0",
 CellID->339951834],
Cell[26969112, 604988, 174, 3, 26, "Output",ExpressionUUID->"08e8e660-49b8-476b-b866-9eb2d43e5f25",
 CellID->1597897053]
}, Open  ]],
Cell[CellGroupData[{
Cell[26969323, 604996, 354, 11, 27, "Input",ExpressionUUID->"e71f1505-84d5-462b-975a-61cdda3844ee",
 CellID->686277402],
Cell[26969680, 605009, 955, 34, 45, "Output",ExpressionUUID->"f6a6ade0-c367-49e9-bb22-86c8628ad3dc",
 CellID->1572764036]
}, Open  ]],
Cell[CellGroupData[{
Cell[26970672, 605048, 358, 11, 27, "Input",ExpressionUUID->"bc9442dd-f05f-4e17-8cef-b7848b5f8e28",
 CellID->1098907181],
Cell[26971033, 605061, 1212, 43, 50, "Output",ExpressionUUID->"8b78a84f-89ed-4944-b482-e5eb3560f053",
 CellID->68942129]
}, Open  ]],
Cell[26972260, 605107, 291, 8, 27, "Input",ExpressionUUID->"286a6d25-ba84-48a7-90db-8b32e084b411",
 CellID->2068847176],
Cell[CellGroupData[{
Cell[26972576, 605119, 782, 25, 27, "Input",ExpressionUUID->"572cbd85-3f97-4e63-87bc-332dc676f190",
 CellID->1206758455],
Cell[26973361, 605146, 171, 3, 26, "Output",ExpressionUUID->"2193d515-f3fd-44d1-8c3c-eedc6a65e466",
 CellID->1216676370]
}, Open  ]],
Cell[26973547, 605152, 204, 4, 27, "Input",ExpressionUUID->"ef07c5ea-e29e-4cfe-a0fc-c33858f337f7",
 CellID->2134093256],
Cell[26973754, 605158, 179, 3, 27, "Input",ExpressionUUID->"c96f60aa-3766-4fc0-844d-cc64eee12483",
 CellID->433481398],
Cell[26973936, 605163, 177, 3, 27, "Input",ExpressionUUID->"10bd86a4-3baf-42dd-833b-17505b2b4606",
 CellID->1257325562],
Cell[26974116, 605168, 165, 3, 27, "Input",ExpressionUUID->"82775c2b-76c9-4e88-b2f6-bf20a659cc99",
 CellID->631634054],
Cell[26974284, 605173, 166, 3, 27, "Input",ExpressionUUID->"3fac0238-1fab-4bbf-ab4c-11478444c69e",
 CellID->861878457],
Cell[26974453, 605178, 170, 3, 27, "Input",ExpressionUUID->"72a9d524-b732-4591-bd6c-524c63bf0dce",
 CellID->814027784]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[26974672, 605187, 237, 3, 41, "Subsubsection",ExpressionUUID->"11318086-eb63-457e-bc15-c56c1be55fdf",
 CellID->1894134298],
Cell[26974912, 605192, 289, 7, 27, "Input",ExpressionUUID->"92b85df9-87ad-4c0f-8d8c-2bf37398c172",
 CellID->417563358],
Cell[26975204, 605201, 497, 13, 27, "Input",ExpressionUUID->"04452c7b-51f8-4b67-94d7-ac5616115d3f",
 CellID->10799510],
Cell[26975704, 605216, 185, 3, 27, "Input",ExpressionUUID->"7d7a9790-ef67-4587-9f02-d3fd6fa8652a",
 CellID->1118738222],
Cell[CellGroupData[{
Cell[26975914, 605223, 651, 19, 27, "Input",ExpressionUUID->"962b08e3-3adf-40b5-b78f-ffee4be371b6",
 CellID->1413967669],
Cell[26976568, 605244, 6890, 219, 51, "Output",ExpressionUUID->"a71f4608-0b1f-4008-9309-74d3b151347b",
 CellID->1204595430]
}, Open  ]],
Cell[26983473, 605466, 185, 4, 27, "Input",ExpressionUUID->"176c9ed7-3710-4cd7-9736-5a58d847a4e5",
 CellID->1473621768],
Cell[CellGroupData[{
Cell[26983683, 605474, 350, 9, 27, "Input",ExpressionUUID->"29ac6433-98fc-436a-b48a-7e364040ed4c",
 CellID->232153925],
Cell[26984036, 605485, 226, 5, 26, "Output",ExpressionUUID->"a50a4089-bba7-49ba-bbf5-22f465b33755",
 CellID->1307051741]
}, Open  ]],
Cell[CellGroupData[{
Cell[26984299, 605495, 418, 12, 27, "Input",ExpressionUUID->"1cc8e48c-d148-4847-9e46-35259af87a96",
 CellID->1724975914],
Cell[26984720, 605509, 229, 5, 26, "Output",ExpressionUUID->"0897a419-34c2-4e62-bde6-90029ee8435e",
 CellID->1008385824]
}, Open  ]],
Cell[26984964, 605517, 210, 5, 27, "Input",ExpressionUUID->"89b92ce6-d787-40fa-ae26-071a2f1767ca",
 CellID->703937734],
Cell[CellGroupData[{
Cell[26985199, 605526, 158, 3, 27, "Input",ExpressionUUID->"a37489ae-26eb-4507-831d-590ef8a266df",
 CellID->594084015],
Cell[26985360, 605531, 174, 3, 26, "Output",ExpressionUUID->"ee730209-a73b-4df5-a4e6-f11286449fba",
 CellID->2147085491]
}, Open  ]],
Cell[CellGroupData[{
Cell[26985571, 605539, 132, 1, 40, "MathCaption",ExpressionUUID->"fcf2f4e9-3fd6-4091-878e-54791cd547e1",
 CellID->653504176],
Cell[CellGroupData[{
Cell[26985728, 605544, 363, 10, 27, "Input",ExpressionUUID->"8053bf52-1937-46d7-ab77-67b1137553d3",
 CellID->1213315274],
Cell[26986094, 605556, 173, 3, 26, "Output",ExpressionUUID->"b2a2bf7c-afd8-4527-8cf5-3ac715e3ed12",
 CellID->1898126696]
}, Open  ]],
Cell[CellGroupData[{
Cell[26986304, 605564, 247, 5, 27, "Input",ExpressionUUID->"0eb4eb0b-5209-46fa-a7b0-10cff60ae974",
 CellID->992738097],
Cell[26986554, 605571, 170, 3, 26, "Output",ExpressionUUID->"9c73a7c5-6e7e-4c5c-9c0a-80ef36ac766f",
 CellID->945803487]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[26986773, 605580, 128, 1, 40, "MathCaption",ExpressionUUID->"a6e397c3-e107-4fb6-91d0-dcdb4ad0ab5e",
 CellID->1041177422],
Cell[CellGroupData[{
Cell[26986926, 605585, 429, 12, 27, "Input",ExpressionUUID->"ff1aee4e-64d0-47d8-9f14-fe7bbaf8a726",
 CellID->1372175798],
Cell[26987358, 605599, 171, 3, 26, "Output",ExpressionUUID->"a88b47ca-001b-4ee3-ae76-a04d8f4ea0c7",
 CellID->445250915]
}, Open  ]],
Cell[CellGroupData[{
Cell[26987566, 605607, 188, 4, 27, "Input",ExpressionUUID->"30207a68-40f6-4e11-9430-9810503c6778",
 CellID->885845759],
Cell[26987757, 605613, 16733, 534, 113, "Output",ExpressionUUID->"040233fb-2af7-4a0a-89f7-e45dbf1546e7",
 CellID->1499102257]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[27004539, 606153, 123, 1, 40, "MathCaption",ExpressionUUID->"0cd90f33-8f4a-4163-8956-76d293bbc8b4",
 CellID->1664008414],
Cell[CellGroupData[{
Cell[27004687, 606158, 755, 22, 27, "Input",ExpressionUUID->"75bed984-cab3-4623-bbb8-84cb845427c0",
 CellID->309879217],
Cell[27005445, 606182, 863, 25, 48, "Output",ExpressionUUID->"ee5c9cbe-8a0c-491a-b6c9-e3aa0c7e6033",
 CellID->1133486795]
}, Open  ]],
Cell[CellGroupData[{
Cell[27006345, 606212, 188, 4, 27, "Input",ExpressionUUID->"b97b6642-cd9b-47a5-a013-7d8a11f760db",
 CellID->1708787611],
Cell[27006536, 606218, 200, 4, 26, "Output",ExpressionUUID->"6d688e24-ec73-4423-8c88-e1fcfbe00720",
 CellID->185131028]
}, Open  ]],
Cell[CellGroupData[{
Cell[27006773, 606227, 295, 8, 27, "Input",ExpressionUUID->"a416eaa5-f0b4-4e33-95f0-007d029c19bb",
 CellID->1626472265],
Cell[27007071, 606237, 176, 3, 26, "Output",ExpressionUUID->"170dbacb-b437-416a-897e-ebf4fcb435fa",
 CellID->1940866740]
}, Open  ]],
Cell[27007262, 606243, 265, 6, 27, "Input",ExpressionUUID->"9fda21d1-380d-4e0d-b0a9-ca331e16e793",
 CellID->460218077],
Cell[27007530, 606251, 187, 3, 27, "Input",ExpressionUUID->"a4c4e8c4-4fa9-433f-831d-f4ed0c22c871",
 CellID->1796607292],
Cell[27007720, 606256, 165, 3, 27, "Input",ExpressionUUID->"c85f93e1-7e3d-4df9-a839-19064992a182",
 CellID->1958679824],
Cell[27007888, 606261, 166, 3, 27, "Input",ExpressionUUID->"84da5be2-80f9-4241-bfeb-88c73a0c5913",
 CellID->754977105],
Cell[27008057, 606266, 171, 3, 27, "Input",ExpressionUUID->"ce6deefc-4524-4c1e-b0f9-24f02e2c9643",
 CellID->1566817259]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[27008277, 606275, 239, 3, 41, "Subsubsection",ExpressionUUID->"3f6a2971-b706-4f65-94b2-201e63eba470",
 CellID->1053192499],
Cell[27008519, 606280, 290, 7, 27, "Input",ExpressionUUID->"445f5659-83e6-48cc-9f30-0d6866222c36",
 CellID->1446112119],
Cell[27008812, 606289, 453, 12, 27, "Input",ExpressionUUID->"a9862d3b-61da-4960-ab3b-5c2b347e2146",
 CellID->1733950934],
Cell[27009268, 606303, 236, 5, 27, "Input",ExpressionUUID->"e2ae665e-bbce-4b11-886c-856d268f2e27",
 CellID->1160257948],
Cell[27009507, 606310, 611, 19, 47, "Input",ExpressionUUID->"2ed9ea66-f6dd-4ae4-8ba8-8eae756289d5",
 CellID->1303817276],
Cell[CellGroupData[{
Cell[27010143, 606333, 2305, 70, 47, "Input",ExpressionUUID->"f68c8677-5cda-4013-ac6f-8918f6d20775",
 CellID->333823744],
Cell[27012451, 606405, 10408, 328, 94, "Output",ExpressionUUID->"ca924408-d115-452f-a517-be25ee441f02",
 CellID->891901034]
}, Open  ]],
Cell[27022874, 606736, 185, 4, 27, "Input",ExpressionUUID->"f5819f61-25f3-449f-9838-613b1a72e508",
 CellID->1965872164],
Cell[CellGroupData[{
Cell[27023084, 606744, 351, 9, 27, "Input",ExpressionUUID->"47eb9bb5-7cc0-4670-b1b3-8666c8f8d3ee",
 CellID->1377551253],
Cell[27023438, 606755, 228, 5, 26, "Output",ExpressionUUID->"769c8ef6-7927-4089-8f2d-72d88b0e8bf7",
 CellID->744822378]
}, Open  ]],
Cell[27023681, 606763, 372, 10, 27, "Input",ExpressionUUID->"e3a96576-26e6-4a05-8544-d0998f51543d",
 CellID->1875756529],
Cell[CellGroupData[{
Cell[27024078, 606777, 199, 4, 27, "Input",ExpressionUUID->"b2acaca0-b155-4a42-9c52-199087c6df90",
 CellID->780044943],
Cell[27024280, 606783, 175, 3, 26, "Output",ExpressionUUID->"317396df-5738-400a-8186-6db6946af311",
 CellID->547528683]
}, Open  ]],
Cell[CellGroupData[{
Cell[27024492, 606791, 454, 13, 27, "Input",ExpressionUUID->"7d4c40b5-a713-45e1-9b4f-e0fad862722c",
 CellID->975993065],
Cell[27024949, 606806, 228, 5, 26, "Output",ExpressionUUID->"27b520e9-07ed-4572-8838-26e954637cd9",
 CellID->347147786]
}, Open  ]],
Cell[27025192, 606814, 209, 5, 27, "Input",ExpressionUUID->"06577532-7635-491f-844a-ed35f14f0795",
 CellID->28281242],
Cell[27025404, 606821, 284, 8, 27, "Input",ExpressionUUID->"f59b0314-73bb-46e4-adae-779eeeec6aa6",
 CellID->728948307],
Cell[CellGroupData[{
Cell[27025713, 606833, 762, 25, 27, "Input",ExpressionUUID->"955c73ad-2d23-4435-8cb1-e4119b01e593",
 CellID->39989614],
Cell[27026478, 606860, 171, 3, 26, "Output",ExpressionUUID->"3b7cd555-6c1f-4472-8acc-270a3443d4f9",
 CellID->1947254050]
}, Open  ]],
Cell[CellGroupData[{
Cell[27026686, 606868, 203, 4, 40, "MathCaption",ExpressionUUID->"914f8c4e-f516-4f9b-aa77-deff91aa9cc9",
 CellID->803715071],
Cell[27026892, 606874, 260, 6, 27, "Input",ExpressionUUID->"c62b1b22-c1ab-4f73-a0fa-76a3c9c535d0",
 CellID->367058667],
Cell[CellGroupData[{
Cell[27027177, 606884, 895, 26, 27, "Input",ExpressionUUID->"8c3b1d86-e094-40d3-86fd-dc88766da5ca",
 CellID->1581982085],
Cell[27028075, 606912, 8024, 218, 62, "Output",ExpressionUUID->"274837b2-714a-4841-8ab4-682f016f3dc6",
 CellID->1995521717]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[27036148, 607136, 150, 1, 40, "MathCaption",ExpressionUUID->"3ec649e4-1abd-404c-9ad2-b50e7931f157",
 CellID->736617837],
Cell[CellGroupData[{
Cell[27036323, 607141, 685, 22, 27, "Input",ExpressionUUID->"7b7a1063-463c-4ef5-b2a2-1ef4bfe23c57",
 CellID->92761504],
Cell[27037011, 607165, 174, 3, 26, "Output",ExpressionUUID->"db57d1ce-67b3-4cc6-83b5-fecb514e6812",
 CellID->1481382420]
}, Open  ]],
Cell[CellGroupData[{
Cell[27037222, 607173, 678, 20, 27, "Input",ExpressionUUID->"4b9f3fb6-bc5e-416f-9ba0-b25ad17bc504",
 CellID->130982475],
Cell[27037903, 607195, 57044, 1843, 422, "Output",ExpressionUUID->"a9be3bfe-b70b-412f-9c54-045f8a0149bd",
 CellID->1816413269]
}, Open  ]],
Cell[CellGroupData[{
Cell[27094984, 609043, 667, 21, 27, "Input",ExpressionUUID->"80a55849-fd05-40a3-9696-25270f4168c2",
 CellID->100969460],
Cell[27095654, 609066, 1372, 49, 26, "Output",ExpressionUUID->"c63e4a9a-35b0-4ac3-8097-db7892170e31",
 CellID->1977769978]
}, Open  ]],
Cell[CellGroupData[{
Cell[27097063, 609120, 606, 18, 28, "Input",ExpressionUUID->"cb19ad0f-dbb1-41fc-9ee3-1abcfcc923df",
 CellID->941478087],
Cell[27097672, 609140, 20252, 677, 136, "Output",ExpressionUUID->"d942dc06-20a2-45cb-9f95-3a1110b91773",
 CellID->1628131980]
}, Open  ]],
Cell[CellGroupData[{
Cell[27117961, 609822, 3132, 113, 30, "Input",ExpressionUUID->"85c2325b-fd4c-4e7b-95cb-021a93cb1e8b",
 CellID->320489248],
Cell[27121096, 609937, 3165, 115, 29, "Output",ExpressionUUID->"d25fed79-b283-4f78-bd11-a99ab8b08b2a",
 CellID->777557786]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[27124310, 610058, 145, 1, 40, "MathCaption",ExpressionUUID->"560351bf-b3fc-4b93-9f55-afd0b6e5e11d",
 CellID->1729984190],
Cell[CellGroupData[{
Cell[27124480, 610063, 476, 14, 27, "Input",ExpressionUUID->"1da75305-040b-45e4-8158-ae7287757726",
 CellID->1063184597],
Cell[27124959, 610079, 170, 3, 26, "Output",ExpressionUUID->"413378af-acf2-43ab-b37b-b131f9d3cde3",
 CellID->975717307]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[27125178, 610088, 107, 1, 40, "MathCaption",ExpressionUUID->"6c9f2426-4eb2-4f07-9d89-25d8c10c8ee8",
 CellID->595419003],
Cell[27125288, 610091, 285, 6, 27, "Input",ExpressionUUID->"b0b2259a-1b48-4c71-a8bf-16750c357372",
 CellID->2080872048],
Cell[27125576, 610099, 263, 6, 27, "Input",ExpressionUUID->"bfb596e1-a7ca-4837-b7f9-fb981e52661d",
 CellID->1445319341],
Cell[27125842, 610107, 238, 5, 27, "Input",ExpressionUUID->"fd8eadc9-9bdf-4f49-bcdf-573a2a2a2512",
 CellID->1294632970],
Cell[27126083, 610114, 164, 3, 27, "Input",ExpressionUUID->"70a39765-b714-4990-a889-12240bd8a7da",
 CellID->592015386],
Cell[27126250, 610119, 166, 3, 27, "Input",ExpressionUUID->"ff9d74f5-b120-48d5-95be-90762f94294a",
 CellID->710855951],
Cell[27126419, 610124, 170, 3, 27, "Input",ExpressionUUID->"ba385016-2a3e-411c-a823-6c80f284fa95",
 CellID->904440196]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[27126638, 610133, 241, 3, 41, "Subsubsection",ExpressionUUID->"6b504b5f-1275-47b2-b42c-54c202b212da",
 CellID->1037120706],
Cell[27126882, 610138, 290, 7, 27, "Input",ExpressionUUID->"7cf91e03-3e8a-489b-a896-6f6cca2f2cf0",
 CellID->1910426690],
Cell[27127175, 610147, 564, 15, 27, "Input",ExpressionUUID->"d5ad536c-6914-4007-9ce1-a5626f3908da",
 CellID->1513372533],
Cell[27127742, 610164, 305, 7, 27, "Input",ExpressionUUID->"98d9d1ef-5f42-4606-aeb3-68caaa45ca3f",
 CellID->455847063],
Cell[27128050, 610173, 3613, 95, 105, "Input",ExpressionUUID->"a2256773-823c-4c34-b71b-cdca39a73e14",
 CellID->1347418992],
Cell[CellGroupData[{
Cell[27131688, 610272, 217, 6, 27, "Input",ExpressionUUID->"66037e10-3c18-4fa3-8447-99c43519d93c",
 CellID->855940929],
Cell[27131908, 610280, 67608, 1695, 188, "Output",ExpressionUUID->"fb266273-8da4-4a93-af78-72470826eb79",
 CellID->272741481]
}, Open  ]],
Cell[27199531, 611978, 185, 4, 27, "Input",ExpressionUUID->"54cfadc8-f25b-412b-ab1f-447592da1674",
 CellID->1249549649],
Cell[CellGroupData[{
Cell[27199741, 611986, 350, 9, 27, "Input",ExpressionUUID->"21eede11-47c2-416c-9817-fff095bddc9f",
 CellID->938366730],
Cell[27200094, 611997, 227, 5, 26, "Output",ExpressionUUID->"16fee66b-a18c-4c15-91b2-e17fd60c481a",
 CellID->1235968469]
}, Open  ]],
Cell[CellGroupData[{
Cell[27200358, 612007, 472, 13, 27, "Input",ExpressionUUID->"f50180be-36cf-4f8e-ba37-aaf178342d78",
 CellID->282041222],
Cell[27200833, 612022, 227, 5, 26, "Output",ExpressionUUID->"3180d89f-80f1-4584-8874-66fc22683bb0",
 CellID->312741052]
}, Open  ]],
Cell[CellGroupData[{
Cell[27201097, 612032, 345, 10, 27, "Input",ExpressionUUID->"dd2efce4-19f2-40ac-95bf-e5194c109bd4",
 CellID->1088936812],
Cell[27201445, 612044, 390514, 8333, 120, "Output",ExpressionUUID->"e620f93f-da13-4af0-a16d-01718162c5f0",
 CellID->1346376755]
}, Open  ]],
Cell[CellGroupData[{
Cell[27591996, 620382, 147, 1, 40, "MathCaption",ExpressionUUID->"18b05fea-5955-4d62-938b-9d2db5eb50a7",
 CellID->1456317855],
Cell[CellGroupData[{
Cell[27592168, 620387, 5678, 172, 199, "Input",ExpressionUUID->"453d346a-7e35-4f08-b661-54846bec1690",
 CellID->1463064126],
Cell[27597849, 620561, 15314, 557, 96, "Output",ExpressionUUID->"e49da1d5-7e5f-45cb-8e62-fe779075e3c0",
 CellID->518932846]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[27613212, 621124, 196, 4, 40, "MathCaption",ExpressionUUID->"4c86f785-2dfc-46f8-9b93-dfdc303600c4",
 CellID->98042681],
Cell[CellGroupData[{
Cell[27613433, 621132, 1352, 43, 47, "Input",ExpressionUUID->"f2d1b3fa-117e-4d09-b184-41d17f1d4f71",
 CellID->683382318],
Cell[27614788, 621177, 1524, 57, 27, "Output",ExpressionUUID->"132063cc-7ceb-4f61-9bf4-57478884d458",
 CellID->773252255]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[27616361, 621240, 135, 1, 40, "MathCaption",ExpressionUUID->"a224a5d9-a725-491a-9b74-c97eba961101",
 CellID->1215196276],
Cell[CellGroupData[{
Cell[27616521, 621245, 711, 22, 27, "Input",ExpressionUUID->"07e8ab1d-0336-4103-aefb-5203b4117da1",
 CellID->1234327762],
Cell[27617235, 621269, 2193, 80, 47, "Output",ExpressionUUID->"4f0fbff2-6867-45b3-adb5-94bf3f032099",
 CellID->1627101788]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[27619477, 621355, 129, 1, 40, "MathCaption",ExpressionUUID->"eb7f2f2a-a344-4eec-8307-70fafd0594a5",
 CellID->1411293017],
Cell[CellGroupData[{
Cell[27619631, 621360, 185, 4, 27, "Input",ExpressionUUID->"f3d9a494-8327-4e0b-848e-4072cf60769f",
 CellID->644855604],
Cell[27619819, 621366, 2933, 103, 50, "Output",ExpressionUUID->"ba070409-d0b9-4350-8952-281cd0ac754b",
 CellID->642990018]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[27622801, 621475, 173, 3, 40, "MathCaption",ExpressionUUID->"6e23595d-c4b1-4ab1-9fde-8ba45905409f",
 CellID->1764985683],
Cell[CellGroupData[{
Cell[27622999, 621482, 289, 9, 27, "Input",ExpressionUUID->"82cbe296-a963-4c78-90d8-89e3017c8369",
 CellID->922758170],
Cell[27623291, 621493, 114318, 2777, 190, "Output",ExpressionUUID->"9db0fdaf-3e6f-4be7-96cf-b4fc7ed3be01",
 CellID->595520822]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[27737658, 624276, 132, 1, 40, "MathCaption",ExpressionUUID->"c0c769b5-df94-47fc-be8c-46a427d16b94",
 CellID->1444903834],
Cell[27737793, 624279, 2101, 63, 104, "Input",ExpressionUUID->"e745e6d6-16f5-47f5-bf9c-51554d83ada6",
 CellID->753657762],
Cell[CellGroupData[{
Cell[27739919, 624346, 324, 10, 27, "Input",ExpressionUUID->"4de3d195-a751-4bb5-9906-e7ec279c08e0",
 CellID->867455928],
Cell[27740246, 624358, 171, 3, 26, "Output",ExpressionUUID->"a39b68ee-3663-4e52-971b-b55d147f513f",
 CellID->1644660334]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[27740466, 624367, 443, 7, 72, "MathCaption",ExpressionUUID->"302c25d6-e51f-4846-a6a6-b4174fc177e3",
 CellID->173511674],
Cell[27740912, 624376, 244, 6, 27, "Input",ExpressionUUID->"e8e0f8f8-b32e-4dca-bb63-1a47cf9af351",
 CellID->2075395939],
Cell[27741159, 624384, 308, 7, 27, "Input",ExpressionUUID->"d257e816-21e4-4ee1-b9aa-eb62445a87ce",
 CellID->1750152199],
Cell[27741470, 624393, 164, 3, 27, "Input",ExpressionUUID->"42b82bac-ff56-4847-93d1-1e77a5a73a2a",
 CellID->724495634],
Cell[27741637, 624398, 166, 3, 27, "Input",ExpressionUUID->"5e546700-0ac7-435b-a945-83e0a071d5cb",
 CellID->365842148],
Cell[27741806, 624403, 170, 3, 27, "Input",ExpressionUUID->"9033225a-e95e-4161-bb9c-fe6b3a17a678",
 CellID->463300075]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[27742025, 624412, 222, 3, 41, "Subsubsection",ExpressionUUID->"d0300eb7-fb7f-40b1-b5ba-abd1300b4d73",
 CellID->2056140150],
Cell[27742250, 624417, 289, 7, 27, "Input",ExpressionUUID->"2705ea94-c0b3-4323-a390-d8285d0fe322",
 CellID->240653526],
Cell[27742542, 624426, 446, 12, 27, "Input",ExpressionUUID->"b95ba3ff-b4f2-4631-ae8a-6778a5bb7e6b",
 CellID->952104181],
Cell[27742991, 624440, 219, 5, 27, "Input",ExpressionUUID->"09500b5f-0ac1-4578-a9a4-6bd6cd6165a4",
 CellID->1114316270],
Cell[27743213, 624447, 2781, 95, 313, "Input",ExpressionUUID->"3f7861af-8d90-4559-be01-bbf203c1dfd4",
 CellID->108664403],
Cell[CellGroupData[{
Cell[27746019, 624546, 2058, 66, 66, "Input",ExpressionUUID->"803b2cca-c567-41b8-969f-32268df529c3",
 CellID->961521271],
Cell[27748080, 624614, 8026, 256, 112, "Output",ExpressionUUID->"786906ca-5643-4169-85de-f13061c3ff20",
 CellID->1231498035]
}, Open  ]],
Cell[27756121, 624873, 184, 4, 27, "Input",ExpressionUUID->"aa34b0c2-4bd7-45d5-a231-969cca7d0700",
 CellID->515279972],
Cell[CellGroupData[{
Cell[27756330, 624881, 350, 9, 27, "Input",ExpressionUUID->"9dd2b05d-20bd-4f4f-a0a0-08c925fd7d4a",
 CellID->1656424700],
Cell[27756683, 624892, 226, 5, 26, "Output",ExpressionUUID->"43db0136-cb57-4cab-a6dc-948c11215378",
 CellID->340084459]
}, Open  ]],
Cell[CellGroupData[{
Cell[27756946, 624902, 471, 13, 27, "Input",ExpressionUUID->"cf807f30-f1d9-40ba-abec-bd2a041cd21a",
 CellID->566307973],
Cell[27757420, 624917, 229, 5, 26, "Output",ExpressionUUID->"1f3d6b04-ff28-4445-9ce4-d628b53ba2fd",
 CellID->1345517746]
}, Open  ]],
Cell[CellGroupData[{
Cell[27757686, 624927, 199, 4, 27, "Input",ExpressionUUID->"163012b9-de78-4294-9ce8-439ff4398e55",
 CellID->378515271],
Cell[27757888, 624933, 174, 3, 26, "Output",ExpressionUUID->"aa3fa102-b5d3-43d2-a382-411dcdc8e236",
 CellID->1113954112]
}, Open  ]],
Cell[CellGroupData[{
Cell[27758099, 624941, 214, 5, 27, "Input",ExpressionUUID->"6adbdf9e-af35-4829-afc4-8656aa7e2050",
 CellID->1129042012],
Cell[27758316, 624948, 3364, 108, 65, "Output",ExpressionUUID->"a70e37a2-04bb-4c57-b5d1-3755a50ff9b1",
 CellID->464325838]
}, Open  ]],
Cell[27761695, 625059, 389, 8, 27, "Input",ExpressionUUID->"30fab4bd-b899-4c9f-9e33-4afa57f900ce",
 CellID->2102794852],
Cell[27762087, 625069, 220, 5, 27, "Input",ExpressionUUID->"142cf5f1-02c5-4458-974f-a812d00823df",
 CellID->534698885],
Cell[27762310, 625076, 164, 3, 27, "Input",ExpressionUUID->"f45ee040-c982-42d9-b751-eeeb1ae50196",
 CellID->1612581147],
Cell[27762477, 625081, 167, 3, 27, "Input",ExpressionUUID->"64bcf3d0-bd6c-4b7f-88c4-6110f8f6d7b1",
 CellID->1407712489],
Cell[27762647, 625086, 170, 3, 27, "Input",ExpressionUUID->"f2098fbf-d7b6-488b-8e2c-a559c51319a3",
 CellID->596463624]
}, Closed]],
Cell[CellGroupData[{
Cell[27762854, 625094, 213, 3, 41, "Subsubsection",ExpressionUUID->"ed11af34-3206-4e65-9265-efca14ea9494",
 CellID->310464372],
Cell[CellGroupData[{
Cell[27763092, 625101, 303, 5, 40, "Text",ExpressionUUID->"ecfd0e7d-b29a-4a22-ac04-e39499f1857b",
 CellID->1736921641],
Cell[27763398, 625108, 290, 7, 27, "Input",ExpressionUUID->"1f4bb601-52b7-45af-b2cd-c499250f19b3",
 CellID->1696806616],
Cell[27763691, 625117, 463, 12, 27, "Input",ExpressionUUID->"f6eae2a1-2d48-4735-af80-1984aa9067b2",
 CellID->1793446599],
Cell[27764157, 625131, 253, 6, 27, "Input",ExpressionUUID->"fb4e5c33-240e-45a3-8140-d998013861cd",
 CellID->376520101],
Cell[27764413, 625139, 4450, 125, 104, "Input",ExpressionUUID->"022b29fb-a5b4-42f1-b4dd-cbd756b12c9b",
 CellID->1686513445],
Cell[27768866, 625266, 182, 4, 27, "Input",ExpressionUUID->"56b1ffab-cc49-4c60-8e85-b6046add68f4",
 CellID->594813362],
Cell[CellGroupData[{
Cell[27769073, 625274, 353, 9, 27, "Input",ExpressionUUID->"13b0d17e-8b99-44fb-82d9-113a8125126f",
 CellID->491798349],
Cell[27769429, 625285, 228, 5, 26, "Output",ExpressionUUID->"9d4e4827-4f4f-4b11-a667-de76921a0ff2",
 CellID->913687296]
}, Open  ]],
Cell[CellGroupData[{
Cell[27769694, 625295, 359, 10, 27, "Input",ExpressionUUID->"5544e523-f877-4b2c-8e4a-738dd1030aae",
 CellID->85729552],
Cell[27770056, 625307, 227, 5, 26, "Output",ExpressionUUID->"5bda44a9-90ea-4daa-a0fa-5a94ce2bcb21",
 CellID->1834307822]
}, Open  ]],
Cell[27770298, 625315, 211, 5, 27, "Input",ExpressionUUID->"5802d2ed-3bf8-4cf3-9551-99824f0c8b12",
 CellID->1392824723],
Cell[CellGroupData[{
Cell[27770534, 625324, 341, 10, 27, "Input",ExpressionUUID->"d4cc5dc1-eea8-4181-a52f-400c15894626",
 CellID->1633367161],
Cell[27770878, 625336, 73680, 2740, 351, "Output",ExpressionUUID->"05dc5d83-257a-40ea-b14f-35704142f3b4",
 CellID->2083811669]
}, Open  ]],
Cell[CellGroupData[{
Cell[27844595, 628081, 362, 11, 27, "Input",ExpressionUUID->"cd0c71a8-295f-49c5-a17b-148e51191d7e",
 CellID->1253320648],
Cell[27844960, 628094, 17985, 671, 99, "Output",ExpressionUUID->"21189b60-1247-472f-affc-033d54a4b0fd",
 CellID->1459405607]
}, Open  ]],
Cell[CellGroupData[{
Cell[27862982, 628770, 196, 4, 27, "Input",ExpressionUUID->"f23aef84-559f-4771-ae6f-13acdcb78cff",
 CellID->1872777177],
Cell[27863181, 628776, 174, 3, 26, "Output",ExpressionUUID->"56636b99-95ba-4aee-b787-835c72034724",
 CellID->1819304462]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[27863404, 628785, 128, 1, 40, "MathCaption",ExpressionUUID->"5f14c15b-0f67-4d7f-94d3-f5adfafcdf52",
 CellID->1175925424],
Cell[27863535, 628788, 213, 5, 27, "Input",ExpressionUUID->"5fb186f5-4c38-4429-9a8a-66afa099e9bc",
 CellID->486597827],
Cell[CellGroupData[{
Cell[27863773, 628797, 294, 8, 27, "Input",ExpressionUUID->"dd258dae-f143-4224-96e5-fb1d08933aa7",
 CellID->631950613],
Cell[27864070, 628807, 229, 5, 26, "Output",ExpressionUUID->"05105290-c63d-49a0-92bf-be8e452a97c5",
 CellID->1177926848]
}, Open  ]],
Cell[CellGroupData[{
Cell[27864336, 628817, 842, 27, 47, "Input",ExpressionUUID->"a7b2914c-9ed3-489f-9902-66b659a3564b",
 CellID->1824798542],
Cell[27865181, 628846, 226, 5, 26, "Output",ExpressionUUID->"d4a5d477-c54e-4da6-99da-4ddee78de0a3",
 CellID->1912276290]
}, Open  ]],
Cell[27865422, 628854, 221, 5, 27, "Input",ExpressionUUID->"06e794a1-a292-4daa-b679-dbce70e5beb0",
 CellID->46288250],
Cell[27865646, 628861, 187, 3, 27, "Input",ExpressionUUID->"a65379ed-5b51-4d27-984f-1f44aefee5ae",
 CellID->577537536],
Cell[27865836, 628866, 166, 3, 27, "Input",ExpressionUUID->"40c8f778-45a8-46c1-ac8b-3ff4909d4dac",
 CellID->47283112],
Cell[27866005, 628871, 166, 3, 27, "Input",ExpressionUUID->"0f336d37-1256-4b90-be46-bc4d1f06b853",
 CellID->118313994],
Cell[27866174, 628876, 171, 3, 27, "Input",ExpressionUUID->"311f131b-6c98-4fff-86d1-610186acb37a",
 CellID->2080923683]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[27866394, 628885, 254, 4, 41, "Subsubsection",ExpressionUUID->"a2b2b968-4a83-4f3c-9912-4d847259de35",
 CellID->165665139],
Cell[27866651, 628891, 290, 7, 27, "Input",ExpressionUUID->"3efa4ea1-4eee-45fe-971a-7427f70f6ca0",
 CellID->1195184661],
Cell[27866944, 628900, 460, 12, 27, "Input",ExpressionUUID->"efbc5179-355e-48c4-9913-1059ae0275a4",
 CellID->2054047036],
Cell[27867407, 628914, 236, 5, 27, "Input",ExpressionUUID->"dc137585-05f0-489a-96d1-fd4f21ddb4fe",
 CellID->1568908527],
Cell[CellGroupData[{
Cell[27867668, 628923, 4298, 121, 117, "Input",ExpressionUUID->"7aac1b31-e191-44f0-9617-254db00b6d27",
 CellID->1841395423],
Cell[27871969, 629046, 13272, 409, 94, "Output",ExpressionUUID->"d5e2380c-796b-4025-a856-c1d434719185",
 CellID->496869296]
}, Open  ]],
Cell[27885256, 629458, 183, 4, 27, "Input",ExpressionUUID->"e0ba803e-b83b-4d25-a6a4-91303dc9eb79",
 CellID->82398920],
Cell[CellGroupData[{
Cell[27885464, 629466, 351, 9, 27, "Input",ExpressionUUID->"aaac1a50-131c-4c66-b6b4-ae9daee651db",
 CellID->1899179772],
Cell[27885818, 629477, 227, 5, 26, "Output",ExpressionUUID->"b65546a2-1b16-4ecd-a1d5-7fc394af5919",
 CellID->1465838336]
}, Open  ]],
Cell[CellGroupData[{
Cell[27886082, 629487, 355, 10, 27, "Input",ExpressionUUID->"a10e1ac7-e102-4f6e-b440-66c782bc3e87",
 CellID->1063268046],
Cell[27886440, 629499, 226, 5, 26, "Output",ExpressionUUID->"1ed179d4-3549-4671-b47f-eae6356ba670",
 CellID->412514768]
}, Open  ]],
Cell[CellGroupData[{
Cell[27886703, 629509, 200, 4, 27, "Input",ExpressionUUID->"dbe0e60c-71a1-41a0-89c6-5f6cee579a5c",
 CellID->2091963332],
Cell[27886906, 629515, 174, 3, 26, "Output",ExpressionUUID->"6d26ee48-4532-4948-bf2d-280df12bea06",
 CellID->1058493260]
}, Open  ]],
Cell[27887095, 629521, 162, 3, 27, "Input",ExpressionUUID->"c93c3b26-b1fe-49ac-a0a5-b4d9c260f497",
 CellID->1219530505],
Cell[27887260, 629526, 238, 5, 27, "Input",ExpressionUUID->"38a5583b-ac23-4762-a95e-c7316f5cbf7e",
 CellID->1065270567],
Cell[27887501, 629533, 164, 3, 27, "Input",ExpressionUUID->"95b40dd6-6965-48ce-83fb-7b883e34b8ca",
 CellID->431254530],
Cell[27887668, 629538, 166, 3, 27, "Input",ExpressionUUID->"32fc32c7-d3ff-4af8-8fb2-362f3b2518e3",
 CellID->337577085],
Cell[27887837, 629543, 168, 3, 27, "Input",ExpressionUUID->"131027a4-8e90-49c7-b953-fef9056dc041",
 CellID->6272221]
}, Closed]],
Cell[CellGroupData[{
Cell[27888042, 629551, 226, 3, 41, "Subsubsection",ExpressionUUID->"08a2783c-e9d4-4fb7-ae20-63478085d21e",
 CellID->190519644],
Cell[27888271, 629556, 289, 7, 27, "Input",ExpressionUUID->"7269e2a0-ec32-48c3-8ffc-06834c3f3f01",
 CellID->165539031],
Cell[27888563, 629565, 453, 12, 27, "Input",ExpressionUUID->"ff1f6c37-1ae1-48e0-b53c-c9778ac0e942",
 CellID->1223171152],
Cell[27889019, 629579, 236, 5, 27, "Input",ExpressionUUID->"93e97590-ed37-4c36-bb68-3631deb916ff",
 CellID->2124082580],
Cell[27889258, 629586, 611, 19, 47, "Input",ExpressionUUID->"e2dab97a-6aaa-4cf9-8962-8cd90528f704",
 CellID->1835257979],
Cell[27889872, 629607, 184, 4, 27, "Input",ExpressionUUID->"95ce0234-0fd6-40cd-87a2-358d377d7a40",
 CellID->873231192],
Cell[27890059, 629613, 1452, 37, 66, "Input",ExpressionUUID->"ec978f37-fac3-49de-97df-1f14a84d9536",
 CellID->493400346],
Cell[27891514, 629652, 1213, 36, 47, "Input",ExpressionUUID->"f80260dd-4b34-4204-b935-f46220eb3e85",
 CellID->1145254155],
Cell[CellGroupData[{
Cell[27892752, 629692, 350, 9, 27, "Input",ExpressionUUID->"ebeef5f6-a2fc-46b8-b894-daf14a643d22",
 CellID->390655024],
Cell[27893105, 629703, 226, 5, 26, "Output",ExpressionUUID->"6b14676f-decc-473c-b7eb-e50b8cee305e",
 CellID->1639756400]
}, Open  ]],
Cell[CellGroupData[{
Cell[27893368, 629713, 418, 12, 27, "Input",ExpressionUUID->"66638003-673a-496c-a79d-103fbce568ae",
 CellID->1683574617],
Cell[27893789, 629727, 226, 5, 26, "Output",ExpressionUUID->"692b21e2-472c-4dd0-876b-3861f4157350",
 CellID->728172849]
}, Open  ]],
Cell[27894030, 629735, 210, 5, 27, "Input",ExpressionUUID->"9d780fd6-4f8d-4c95-b30e-07263e4092e5",
 CellID->631700026],
Cell[CellGroupData[{
Cell[27894265, 629744, 160, 3, 27, "Input",ExpressionUUID->"a717a385-1f44-44ba-b46c-b2446154fe41",
 CellID->1970669861],
Cell[27894428, 629749, 174, 3, 26, "Output",ExpressionUUID->"d0d7d372-bcba-444d-9bfd-e62e07a676f2",
 CellID->2039252537]
}, Open  ]],
Cell[27894617, 629755, 214, 4, 27, "Input",ExpressionUUID->"0f9c5bf7-782f-43b0-84d9-07e5d2243c3d",
 CellID->1452430951],
Cell[27894834, 629761, 237, 5, 27, "Input",ExpressionUUID->"674a8da7-471a-4e35-8f3f-6a788114402f",
 CellID->560378008],
Cell[27895074, 629768, 165, 3, 27, "Input",ExpressionUUID->"f2dd6a5a-fbd9-4af6-82ba-bad372c0bc7a",
 CellID->1557388361],
Cell[27895242, 629773, 164, 3, 27, "Input",ExpressionUUID->"c91fcb19-7c9a-4f9e-b76a-5e0be5f3c3c6",
 CellID->1495493691],
Cell[27895409, 629778, 166, 3, 27, "Input",ExpressionUUID->"859fbdc4-3188-4d1e-83c4-3f1da4dd9e76",
 CellID->613860149],
Cell[27895578, 629783, 170, 3, 27, "Input",ExpressionUUID->"357e4aa7-7547-46e8-b0b5-decdcca16496",
 CellID->636084055]
}, Closed]],
Cell[CellGroupData[{
Cell[27895785, 629791, 220, 3, 41, "Subsubsection",ExpressionUUID->"01a3bdd2-3d25-4873-8344-57453f334564",
 CellID->491117619],
Cell[27896008, 629796, 290, 7, 27, "Input",ExpressionUUID->"20c63041-c165-4a88-a7ac-378107eafb88",
 CellID->1434167112],
Cell[27896301, 629805, 452, 12, 27, "Input",ExpressionUUID->"86d8ad8b-e3e5-48af-b3b0-f9d4e251a158",
 CellID->690197456],
Cell[27896756, 629819, 236, 5, 27, "Input",ExpressionUUID->"637cc882-bf88-403d-b0b0-a42dbb83325d",
 CellID->1589465545],
Cell[27896995, 629826, 8362, 224, 239, "Input",ExpressionUUID->"09062ecf-76ac-4966-8054-34512cd20410",
 CellID->1598329782],
Cell[CellGroupData[{
Cell[27905382, 630054, 874, 26, 27, "Input",ExpressionUUID->"997ba79f-02cc-4c39-9037-5ed67456391f",
 CellID->255705751],
Cell[27906259, 630082, 4969, 143, 68, "Output",ExpressionUUID->"83581830-1c23-45eb-9bd0-c3f2d5ab0f29",
 CellID->367396622]
}, Open  ]],
Cell[27911243, 630228, 185, 4, 27, "Input",ExpressionUUID->"0f34b516-613c-4fd4-b104-2a765c42f952",
 CellID->1937088703],
Cell[CellGroupData[{
Cell[27911453, 630236, 299, 7, 27, "Input",ExpressionUUID->"7e8f3fde-8c9e-4c62-bb58-8c540b5a96d2",
 CellID->797742801],
Cell[27911755, 630245, 226, 5, 26, "Output",ExpressionUUID->"3ff0a464-69d0-448e-b6f9-1b8989d83b93",
 CellID->258296431]
}, Open  ]],
Cell[27911996, 630253, 211, 5, 27, "Input",ExpressionUUID->"8a1d5299-5622-4b58-95ed-737278362d5d",
 CellID->1182563468],
Cell[CellGroupData[{
Cell[27912232, 630262, 251, 7, 27, "Input",ExpressionUUID->"13a7e0eb-7b0d-4510-8ed4-fcbee5c26924",
 CellID->383053691],
Cell[27912486, 630271, 192, 3, 26, "Output",ExpressionUUID->"92119922-3da2-4378-852e-d8a7fc48b820",
 CellID->370118598]
}, Open  ]],
Cell[CellGroupData[{
Cell[27912715, 630279, 188, 4, 27, "Input",ExpressionUUID->"b56d0877-80e9-4da5-b1f6-3107534cd3f0",
 CellID->545924414],
Cell[27912906, 630285, 2987, 105, 54, "Output",ExpressionUUID->"03f671b0-03c3-493f-89fc-f1d32f3812cd",
 CellID->291909415]
}, Open  ]],
Cell[27915908, 630393, 291, 8, 27, "Input",ExpressionUUID->"f9abdf03-5b87-449b-9ac7-d274a3a4bf30",
 CellID->1956173303],
Cell[CellGroupData[{
Cell[27916224, 630405, 782, 25, 27, "Input",ExpressionUUID->"e4b2aa93-ab32-42b8-ac5b-68fd1345ccbe",
 CellID->1577551598],
Cell[27917009, 630432, 193, 3, 26, "Output",ExpressionUUID->"f2ad4e0a-ae31-41c9-89af-91f852d046d6",
 CellID->1523842790]
}, Open  ]],
Cell[CellGroupData[{
Cell[27917239, 630440, 154, 1, 40, "MathCaption",ExpressionUUID->"7d22f626-4efb-43d0-9305-d7ce1c103b4f",
 CellID->1895449293],
Cell[27917396, 630443, 190, 4, 27, "Input",ExpressionUUID->"274df80c-49c1-4b5f-b389-857d53bb2074",
 CellID->1043847853],
Cell[27917589, 630449, 1083, 31, 85, "Input",ExpressionUUID->"5ab446db-68a5-433b-82ec-533da9636ca9",
 CellID->1156953028],
Cell[CellGroupData[{
Cell[27918697, 630484, 784, 28, 27, "Input",ExpressionUUID->"a67174db-b61b-4a8e-9e82-2eaeb525d88d",
 CellID->1592601658],
Cell[27919484, 630514, 196, 3, 26, "Output",ExpressionUUID->"e9d127a3-3935-4570-a857-9a12bd9cb127",
 CellID->1968718156]
}, Open  ]],
Cell[CellGroupData[{
Cell[27919717, 630522, 428, 13, 27, "Input",ExpressionUUID->"5629262a-7539-4d40-a868-da903c2e28be",
 CellID->718095],
Cell[27920148, 630537, 5109, 184, 54, "Output",ExpressionUUID->"9e925a96-a837-4d8c-8fcb-1802b880315c",
 CellID->1081645726]
}, Open  ]],
Cell[CellGroupData[{
Cell[27925294, 630726, 430, 13, 27, "Input",ExpressionUUID->"7220d04c-079c-4e9c-971c-7725a5864c73",
 CellID->1166521844],
Cell[27925727, 630741, 2487, 91, 54, "Output",ExpressionUUID->"d83c3bec-9d53-40fe-941f-22b3cabd4ee7",
 CellID->1874226486]
}, Open  ]],
Cell[CellGroupData[{
Cell[27928251, 630837, 431, 13, 27, "Input",ExpressionUUID->"d190f182-6b96-4632-b8a0-5be84b76276a",
 CellID->387508315],
Cell[27928685, 630852, 3774, 133, 54, "Output",ExpressionUUID->"c852abf8-2fef-40d5-8ec7-3624ae724adc",
 CellID->1643886604]
}, Open  ]],
Cell[CellGroupData[{
Cell[27932496, 630990, 433, 13, 27, "Input",ExpressionUUID->"4c6507b6-4c63-4d74-a96d-8a8064a5af16",
 CellID->1379082493],
Cell[27932932, 631005, 8344, 292, 54, "Output",ExpressionUUID->"519aab8f-d26d-46e6-9423-59cba49c703b",
 CellID->1503569625]
}, Open  ]],
Cell[CellGroupData[{
Cell[27941313, 631302, 432, 13, 27, "Input",ExpressionUUID->"f655f87d-248d-44a1-bd90-1a6ee2c33917",
 CellID->1709791137],
Cell[27941748, 631317, 2487, 91, 54, "Output",ExpressionUUID->"bafd8cd9-95dd-4146-be2d-5f5fcdfaf963",
 CellID->1648042301]
}, Open  ]],
Cell[CellGroupData[{
Cell[27944272, 631413, 431, 13, 27, "Input",ExpressionUUID->"356d23d8-2994-4dfc-bf32-5307c79a7848",
 CellID->746828945],
Cell[27944706, 631428, 5128, 184, 54, "Output",ExpressionUUID->"69df6ecd-35a8-4cbe-9155-ba8ad7cab630",
 CellID->118923088]
}, Open  ]],
Cell[CellGroupData[{
Cell[27949871, 631617, 432, 13, 27, "Input",ExpressionUUID->"a143c83f-179c-40af-91d3-ff6cec28ec03",
 CellID->1173802436],
Cell[27950306, 631632, 8038, 291, 54, "Output",ExpressionUUID->"7f8adb4a-7e3c-456b-89e3-91193f207be0",
 CellID->535563402]
}, Open  ]],
Cell[CellGroupData[{
Cell[27958381, 631928, 438, 13, 27, "Input",ExpressionUUID->"16c5bdc1-55f7-42ee-9204-c4c7be497345",
 CellID->318902292],
Cell[27958822, 631943, 3628, 132, 54, "Output",ExpressionUUID->"91fadf47-f91f-4697-afa7-966fce5818f7",
 CellID->1520517115]
}, Open  ]],
Cell[CellGroupData[{
Cell[27962487, 632080, 804, 26, 27, "Input",ExpressionUUID->"663bcc13-e732-4a2e-97a9-4141945fbf55",
 CellID->265141514],
Cell[27963294, 632108, 4342, 153, 54, "Output",ExpressionUUID->"abb12fa8-4f13-48d9-b543-2fada08947ee",
 CellID->282781427]
}, Open  ]],
Cell[CellGroupData[{
Cell[27967673, 632266, 802, 26, 27, "Input",ExpressionUUID->"3a3bce0d-503e-43df-aa7c-e2d424342613",
 CellID->636708170],
Cell[27968478, 632294, 4342, 153, 54, "Output",ExpressionUUID->"21b057e6-d07d-4035-bf52-1c6254673134",
 CellID->176308395]
}, Open  ]],
Cell[CellGroupData[{
Cell[27972857, 632452, 801, 26, 27, "Input",ExpressionUUID->"bd7a9a67-040b-4505-973d-6d2794fcf332",
 CellID->804196282],
Cell[27973661, 632480, 6709, 244, 54, "Output",ExpressionUUID->"766ae90d-99f7-4068-8a9a-f842cd3153a4",
 CellID->1276841733]
}, Open  ]],
Cell[CellGroupData[{
Cell[27980407, 632729, 808, 26, 27, "Input",ExpressionUUID->"16e8e266-d384-4c1c-9de1-6586e8560990",
 CellID->359140818],
Cell[27981218, 632757, 6732, 245, 54, "Output",ExpressionUUID->"3e5dd349-3f42-4ddb-8239-b53ae8666379",
 CellID->907952202]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[27987999, 633008, 129, 1, 40, "MathCaption",ExpressionUUID->"ebc4aa06-a4b7-4e32-8f58-bc68068e795f",
 CellID->1785949246],
Cell[27988131, 633011, 1461, 50, 85, "Input",ExpressionUUID->"b27915b6-48e5-4f6b-a87d-39f3f9684be4",
 CellID->1070043403],
Cell[CellGroupData[{
Cell[27989617, 633065, 646, 18, 27, "Input",ExpressionUUID->"1f7e4e33-fec1-4c69-a1e3-d1585e815388",
 CellID->1191414243],
Cell[27990266, 633085, 195, 3, 26, "Output",ExpressionUUID->"c8149c53-0760-46be-b5cc-a2474f665173",
 CellID->743261221]
}, Open  ]],
Cell[CellGroupData[{
Cell[27990498, 633093, 637, 18, 27, "Input",ExpressionUUID->"e29a9280-e192-4dd1-a9f2-ead9efa4f7ad",
 CellID->615221389],
Cell[27991138, 633113, 196, 3, 26, "Output",ExpressionUUID->"6ad750e3-f647-41aa-bebc-47c89cd8f37e",
 CellID->1955537703]
}, Open  ]],
Cell[CellGroupData[{
Cell[27991371, 633121, 639, 18, 27, "Input",ExpressionUUID->"c834fd11-ed09-4d2b-ae04-5ba2fa78bc18",
 CellID->1562136441],
Cell[27992013, 633141, 195, 3, 26, "Output",ExpressionUUID->"62f3504b-77e8-4834-a9f9-a2c4355ab30e",
 CellID->288538542]
}, Open  ]],
Cell[CellGroupData[{
Cell[27992245, 633149, 638, 18, 27, "Input",ExpressionUUID->"2ce2b122-f698-41c3-83a8-77bc5ddf3245",
 CellID->487430925],
Cell[27992886, 633169, 196, 3, 26, "Output",ExpressionUUID->"f11db48b-e969-448e-81a0-e0ad282b7d71",
 CellID->1822781056]
}, Open  ]],
Cell[CellGroupData[{
Cell[27993119, 633177, 611, 17, 27, "Input",ExpressionUUID->"db6bf27a-7e3d-4ec9-b228-5e86732e2351",
 CellID->1604336830],
Cell[27993733, 633196, 195, 3, 26, "Output",ExpressionUUID->"95db6d1d-5afe-479c-aaa8-ba8b7e041bf0",
 CellID->290758316]
}, Open  ]],
Cell[CellGroupData[{
Cell[27993965, 633204, 611, 17, 27, "Input",ExpressionUUID->"a8b12a89-bee0-424c-bf56-f69d1ba58d3b",
 CellID->117722724],
Cell[27994579, 633223, 195, 3, 26, "Output",ExpressionUUID->"6503d0a2-1500-4fb8-a565-d72f6a5ce5d2",
 CellID->501336594]
}, Open  ]],
Cell[CellGroupData[{
Cell[27994811, 633231, 610, 17, 27, "Input",ExpressionUUID->"62b68547-4c94-47d7-943f-e2289a0258af",
 CellID->1565834334],
Cell[27995424, 633250, 198, 3, 26, "Output",ExpressionUUID->"9c119f2d-1770-477d-99de-c69cf4a67f99",
 CellID->2021936104]
}, Open  ]],
Cell[CellGroupData[{
Cell[27995659, 633258, 615, 17, 27, "Input",ExpressionUUID->"b3a76cd2-741b-45b5-99fa-bf593eece7e6",
 CellID->1917676718],
Cell[27996277, 633277, 199, 3, 26, "Output",ExpressionUUID->"f257c8d8-6bc5-4b8f-ba00-3a9ebb1e4f65",
 CellID->298992140]
}, Open  ]],
Cell[CellGroupData[{
Cell[27996513, 633285, 635, 18, 27, "Input",ExpressionUUID->"47ef814e-9dd7-4a53-806d-9d761f5bff62",
 CellID->1845475191],
Cell[27997151, 633305, 197, 3, 26, "Output",ExpressionUUID->"174117bc-c31a-45d0-8415-d4f20acbe32e",
 CellID->935771448]
}, Open  ]],
Cell[CellGroupData[{
Cell[27997385, 633313, 629, 18, 27, "Input",ExpressionUUID->"d3c59194-8532-4839-aff1-2d31171e4608",
 CellID->1497284131],
Cell[27998017, 633333, 197, 3, 26, "Output",ExpressionUUID->"a914edef-e0b8-4844-a0fa-51c36e91c68f",
 CellID->429808897]
}, Open  ]],
Cell[CellGroupData[{
Cell[27998251, 633341, 629, 18, 27, "Input",ExpressionUUID->"10486367-08ab-40ee-821e-e07bb923b90c",
 CellID->1747198645],
Cell[27998883, 633361, 196, 3, 26, "Output",ExpressionUUID->"70750d13-d204-45c9-91d4-d565ceb284cf",
 CellID->1438516804]
}, Open  ]],
Cell[CellGroupData[{
Cell[27999116, 633369, 632, 18, 27, "Input",ExpressionUUID->"542aec4c-1c0a-4344-a58e-2dfe4245a8bc",
 CellID->451051290],
Cell[27999751, 633389, 198, 3, 26, "Output",ExpressionUUID->"5accd619-92c6-4d2e-8a8f-42673047e055",
 CellID->1370229639]
}, Open  ]],
Cell[CellGroupData[{
Cell[27999986, 633397, 615, 17, 27, "Input",ExpressionUUID->"823c2131-b013-4e7c-a27c-4dd387a7292a",
 CellID->55237227],
Cell[28000604, 633416, 195, 3, 26, "Output",ExpressionUUID->"09bd042e-561c-4b81-9a05-bad8a4092ccd",
 CellID->838171297]
}, Open  ]],
Cell[CellGroupData[{
Cell[28000836, 633424, 611, 17, 27, "Input",ExpressionUUID->"ff150099-be54-4dcf-a694-2bbcdb49b515",
 CellID->1423227569],
Cell[28001450, 633443, 199, 3, 26, "Output",ExpressionUUID->"f5c26b7a-39e2-4678-8b86-ed497009ed0b",
 CellID->817900928]
}, Open  ]],
Cell[CellGroupData[{
Cell[28001686, 633451, 609, 17, 27, "Input",ExpressionUUID->"63c71ee8-7276-408d-ad18-7da6076a1fa1",
 CellID->967732399],
Cell[28002298, 633470, 198, 3, 26, "Output",ExpressionUUID->"ea797666-9f84-49fb-b750-c36e7c6bf80a",
 CellID->1905697441]
}, Open  ]],
Cell[CellGroupData[{
Cell[28002533, 633478, 615, 17, 27, "Input",ExpressionUUID->"1a125594-193a-4c5f-9336-1e5f124fa8b5",
 CellID->599019138],
Cell[28003151, 633497, 197, 3, 26, "Output",ExpressionUUID->"83c4e491-cfdd-4ae8-90f0-6cb0aaf38251",
 CellID->2115293610]
}, Open  ]],
Cell[28003363, 633503, 204, 5, 27, "Input",ExpressionUUID->"db10e987-2240-41a4-a582-31f187340eda",
 CellID->1004906518],
Cell[28003570, 633510, 189, 4, 27, "Input",ExpressionUUID->"2539f084-af8c-488d-9871-8ce6571e391f",
 CellID->1652448018],
Cell[CellGroupData[{
Cell[28003784, 633518, 277, 8, 27, "Input",ExpressionUUID->"56887868-e2ef-4169-9fe5-89172be5ebea",
 CellID->1435315522],
Cell[28004064, 633528, 691466, 15071, 256, "Output",ExpressionUUID->"6f9c32fe-6212-4dd9-94d6-2bb2d9d79ce5",
 CellID->1638369278]
}, Open  ]],
Cell[CellGroupData[{
Cell[28695567, 648604, 462, 14, 27, "Input",ExpressionUUID->"5916c837-983c-4870-bd7e-95e2e4002e8f",
 CellID->110453815],
Cell[28696032, 648620, 2503, 89, 52, "Output",ExpressionUUID->"7d0b3902-8761-4a79-9806-ee061d02445b",
 CellID->877458086]
}, Open  ]],
Cell[CellGroupData[{
Cell[28698572, 648714, 463, 14, 27, "Input",ExpressionUUID->"59691ddb-3eed-4035-a3ec-874adec6c94c",
 CellID->1110408985],
Cell[28699038, 648730, 3343, 119, 52, "Output",ExpressionUUID->"08bef24e-29ad-48f2-aa08-5dd14b1921b0",
 CellID->2002328293]
}, Open  ]],
Cell[CellGroupData[{
Cell[28702418, 648854, 466, 14, 27, "Input",ExpressionUUID->"10833a3d-c788-4036-8155-3185fb38d308",
 CellID->1763509838],
Cell[28702887, 648870, 3364, 115, 52, "Output",ExpressionUUID->"719e2803-5b72-418f-b64f-7120d8d9290f",
 CellID->1372414586]
}, Open  ]],
Cell[CellGroupData[{
Cell[28706288, 648990, 465, 14, 27, "Input",ExpressionUUID->"6a9bc01d-fdbb-4f67-bf97-bcd0563823fa",
 CellID->986196261],
Cell[28706756, 649006, 3475, 120, 52, "Output",ExpressionUUID->"93cbe16b-3184-4d5b-9bc8-a43904acb3fc",
 CellID->198692634]
}, Open  ]],
Cell[CellGroupData[{
Cell[28710268, 649131, 469, 14, 27, "Input",ExpressionUUID->"b40f221d-04e4-4320-93de-32d30edb8399",
 CellID->1818920111],
Cell[28710740, 649147, 2504, 89, 52, "Output",ExpressionUUID->"0f5181cb-8bff-467b-8638-a82c1a854829",
 CellID->1383135580]
}, Open  ]],
Cell[28713259, 649239, 164, 3, 27, "Input",ExpressionUUID->"dff9e2d2-a4ee-4bf4-afd1-d3e5fabb0129",
 CellID->451105623],
Cell[28713426, 649244, 405, 8, 27, "Input",ExpressionUUID->"606845a6-7d80-4613-addf-1f699a6dde97",
 CellID->1114931851],
Cell[28713834, 649254, 238, 5, 27, "Input",ExpressionUUID->"24589f80-daac-426f-a834-526803b6a9d0",
 CellID->1864799682],
Cell[28714075, 649261, 165, 3, 27, "Input",ExpressionUUID->"ded61440-2e85-46ea-944c-d1f4981bf859",
 CellID->1311911056],
Cell[28714243, 649266, 167, 3, 27, "Input",ExpressionUUID->"a32d4cf1-30c1-48f4-a1f0-47a4d17ecac4",
 CellID->1074514443],
Cell[28714413, 649271, 171, 3, 27, "Input",ExpressionUUID->"d8229c44-ffef-4f14-a77d-89f85f694a6c",
 CellID->1174255878]
}, Open  ]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[28714645, 649281, 423, 6, 39, "Subsection",ExpressionUUID->"c268a446-9dcb-46d7-ba09-c8a81639130c",
 CellID->343825616],
Cell[28715071, 649289, 195, 3, 27, "Input",ExpressionUUID->"d410987b-4f4d-4e09-8645-b7905bdb34c5",
 CellID->865471439],
Cell[CellGroupData[{
Cell[28715291, 649296, 156, 3, 27, "Input",ExpressionUUID->"6d6f1eee-7370-463d-b623-ecf7b3770279",
 CellID->1946117682],
Cell[CellGroupData[{
Cell[28715472, 649303, 279, 4, 22, "Print",ExpressionUUID->"40af1e79-c0bb-409b-a594-2b574f2ced05",
 CellID->1525917970],
Cell[28715754, 649309, 551, 12, 22, "Print",ExpressionUUID->"7ce648ba-c6ff-4875-90ea-966233199868",
 CellID->853609316],
Cell[28716308, 649323, 301, 4, 22, "Print",ExpressionUUID->"bcfc8227-68a8-422d-afd6-b3043675caba",
 CellID->1421012052],
Cell[28716612, 649329, 259, 3, 22, "Print",ExpressionUUID->"369c58fe-1940-4c91-b024-da91c1bcc9a4",
 CellID->783919677],
Cell[28716874, 649334, 240, 3, 22, "Print",ExpressionUUID->"dfa4586a-022f-4904-9b95-0c067ff906a3",
 CellID->525836994],
Cell[28717117, 649339, 279, 4, 22, "Print",ExpressionUUID->"3a7a15e2-d099-4895-be9e-b201842c50e7",
 CellID->847150753],
Cell[28717399, 649345, 555, 12, 22, "Print",ExpressionUUID->"e024280e-4d84-4e38-9be3-ed9202086296",
 CellID->246126061],
Cell[28717957, 649359, 300, 4, 22, "Print",ExpressionUUID->"79ac6af8-5570-4331-b69f-e845b7acb2af",
 CellID->587203105],
Cell[28718260, 649365, 280, 4, 22, "Print",ExpressionUUID->"a095beae-91c5-4e6a-bd59-5957238024e4",
 CellID->1435227770],
Cell[28718543, 649371, 435, 6, 42, "Print",ExpressionUUID->"d6bd787b-fbee-4228-a189-a1efc7f09ac3",
 CellID->1336807286],
Cell[28718981, 649379, 279, 4, 22, "Print",ExpressionUUID->"6aa337d0-ba87-4a91-bfaf-afeb45eebd23",
 CellID->746140888]
}, Open  ]]
}, Open  ]],
Cell[28719287, 649387, 274, 7, 47, "Input",ExpressionUUID->"5d31a923-3016-47dd-8642-0c27c79f033e",
 CellID->3730788],
Cell[28719564, 649396, 385, 9, 27, "Input",ExpressionUUID->"e0d2ce62-4b0e-4009-b74d-3d9593bbabe1",
 CellID->426328153],
Cell[28719952, 649407, 318, 9, 27, "Input",ExpressionUUID->"7c8e66da-476b-44bb-a46b-0d239d6a8a90",
 CellID->1291821079],
Cell[28720273, 649418, 216, 5, 27, "Input",ExpressionUUID->"0a82e2dc-6bf6-4ae5-997c-28246f283fd5",
 CellID->577203583],
Cell[28720492, 649425, 283, 8, 27, "Input",ExpressionUUID->"33d2c4ff-7978-4da6-ab9f-4e3b827ade3e",
 CellID->876985982],
Cell[CellGroupData[{
Cell[28720800, 649437, 400, 11, 27, "Input",ExpressionUUID->"eb0c5da7-ca7f-4f98-8f63-5e2c3642d4dd",
 CellID->1323130113],
Cell[28721203, 649450, 621, 11, 22, "Message",ExpressionUUID->"04ab7dd0-624c-4395-bb73-7f31e0bb73e3",
 CellID->172368675]
}, Open  ]],
Cell[CellGroupData[{
Cell[28721861, 649466, 256, 7, 27, "Input",ExpressionUUID->"6242f4c4-a431-4bc8-a05f-008e58296351",
 CellID->1051488761],
Cell[28722120, 649475, 1198, 34, 32, "Output",ExpressionUUID->"6a8b6b27-0fc3-4d1d-87f3-62c449a7ec76",
 CellID->1709710596]
}, Open  ]],
Cell[CellGroupData[{
Cell[28723355, 649514, 171, 3, 27, "Input",ExpressionUUID->"7f5f9b5c-a115-4608-a694-34425188f96d",
 CellID->1888364754],
Cell[28723529, 649519, 3103, 89, 32, "Output",ExpressionUUID->"f8464135-b460-4031-9574-506ce7af03d0",
 CellID->11238503]
}, Open  ]],
Cell[CellGroupData[{
Cell[28726669, 649613, 211, 5, 27, "Input",ExpressionUUID->"0baccec6-4dd8-4cf3-aea2-7f28744d41e3",
 CellID->2037262412],
Cell[28726883, 649620, 1126, 32, 32, "Output",ExpressionUUID->"f85078d8-6366-4db9-8cd0-2ce6ba752dba",
 CellID->420817408]
}, Open  ]],
Cell[CellGroupData[{
Cell[28728046, 649657, 379, 11, 27, "Input",ExpressionUUID->"f8006af2-b451-421c-8c4b-45c3c5b8a36c",
 CellID->1163274411],
Cell[28728428, 649670, 42991, 1209, 63, "Output",ExpressionUUID->"0f63456f-eef5-42d9-a65e-1dad8df7a3a6",
 CellID->1965183071]
}, Open  ]],
Cell[CellGroupData[{
Cell[28771456, 650884, 171, 3, 40, "MathCaption",ExpressionUUID->"fd7f372d-5b65-413b-b0e4-2818d5df66b2",
 CellID->56875018],
Cell[CellGroupData[{
Cell[28771652, 650891, 154, 3, 27, "Input",ExpressionUUID->"fda09583-bb07-4426-af71-ecd43708a98e",
 CellID->268204611],
Cell[CellGroupData[{
Cell[28771831, 650898, 259, 4, 22, "Print",ExpressionUUID->"105ab4c3-4e2b-4421-8d43-e5b3a94d43e9",
 CellID->2021331521],
Cell[28772093, 650904, 531, 12, 22, "Print",ExpressionUUID->"a0133a86-ea02-4da9-b6ee-8f9f72f58322",
 CellID->1475061554],
Cell[28772627, 650918, 297, 4, 22, "Print",ExpressionUUID->"4316ad34-e8ab-472e-9369-d5e5a1d98df7",
 CellID->808317110],
Cell[28772927, 650924, 258, 4, 22, "Print",ExpressionUUID->"5a694b80-2c7c-441c-993c-9130b87995d2",
 CellID->404116848],
Cell[28773188, 650930, 416, 6, 42, "Print",ExpressionUUID->"af219891-98f5-4702-9c9e-ce729eab76c9",
 CellID->1648572522],
Cell[28773607, 650938, 256, 4, 22, "Print",ExpressionUUID->"cf7f3e23-b79f-47bc-89ec-81e66daf09ea",
 CellID->724363660]
}, Open  ]]
}, Open  ]],
Cell[28773890, 650946, 461, 12, 27, "Input",ExpressionUUID->"c88e175d-7824-46c1-ae5e-c7dcb521cf60",
 CellID->2129123830],
Cell[28774354, 650960, 239, 5, 27, "Input",ExpressionUUID->"7e5810e0-91ce-48dd-881b-1c4b1df32444",
 CellID->664200849],
Cell[CellGroupData[{
Cell[28774618, 650969, 1172, 37, 27, "Input",ExpressionUUID->"1b3f2831-bf39-4605-a0dc-e7b5f83b7a84",
 CellID->603672037],
Cell[28775793, 651008, 9531, 256, 90, "Output",ExpressionUUID->"13ab58a5-eb42-4334-b94a-a664b8649277",
 CellID->951592667]
}, Open  ]],
Cell[CellGroupData[{
Cell[28785361, 651269, 579, 18, 27, "Input",ExpressionUUID->"80f49d09-fc77-41c7-9328-79a406e1e695",
 CellID->1097062665],
Cell[28785943, 651289, 3370, 95, 74, "Output",ExpressionUUID->"8a602fc0-6398-46bd-8632-1ef24a9740ed",
 CellID->728573098]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[28789362, 651390, 123, 1, 40, "MathCaption",ExpressionUUID->"a6ecf38a-2be8-4b9a-9a9c-05a0cfb8ca98",
 CellID->877348773],
Cell[CellGroupData[{
Cell[28789510, 651395, 349, 10, 27, "Input",ExpressionUUID->"8d7916ab-6176-43e0-96e3-adf027011cc3",
 CellID->978107509],
Cell[28789862, 651407, 187, 4, 26, "Output",ExpressionUUID->"0345e2f1-5198-4909-9b50-b6f678cda4fa",
 CellID->1848222812]
}, Open  ]],
Cell[CellGroupData[{
Cell[28790086, 651416, 178, 4, 27, "Input",ExpressionUUID->"bf41478f-90fb-4cfe-ae48-cbb336498f25",
 CellID->644421335],
Cell[28790267, 651422, 12626, 376, 158, "Output",ExpressionUUID->"b8a83ae6-cb41-4ad8-9c62-b1f19be33845",
 CellID->750481805]
}, Open  ]],
Cell[CellGroupData[{
Cell[28802930, 651803, 309, 9, 27, "Input",ExpressionUUID->"7237a8e1-83ec-4c9e-8005-c156f6a071bd",
 CellID->81961636],
Cell[28803242, 651814, 170, 3, 26, "Output",ExpressionUUID->"656b7f4d-0a89-4ea0-ba62-713404515210",
 CellID->2100127951]
}, Open  ]],
Cell[CellGroupData[{
Cell[28803449, 651822, 175, 4, 27, "Input",ExpressionUUID->"44e82d7e-b04d-4072-9bc8-48d95d477b8b",
 CellID->670090720],
Cell[28803627, 651828, 3388, 96, 70, "Output",ExpressionUUID->"0a6c1d4c-71b2-4b9d-80e0-557037e8e36d",
 CellID->1430974628]
}, Open  ]],
Cell[CellGroupData[{
Cell[28807052, 651929, 291, 8, 27, "Input",ExpressionUUID->"49f5d0b5-d8f2-49a8-adbb-1280a47efceb",
 CellID->352319883],
Cell[28807346, 651939, 8617, 232, 82, "Output",ExpressionUUID->"25b2220a-9fa9-4512-8f17-afac37ecc9b2",
 CellID->786980887]
}, Open  ]],
Cell[CellGroupData[{
Cell[28816000, 652176, 415, 12, 27, "Input",ExpressionUUID->"c5eeecca-c9fb-4e03-81aa-f7242ee792c9",
 CellID->276336301],
Cell[28816418, 652190, 1224, 40, 43, "Output",ExpressionUUID->"3ddf51bd-e15d-4535-933a-aa856222549d",
 CellID->135653359]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[28817691, 652236, 185, 3, 40, "MathCaption",ExpressionUUID->"9f2ee235-cd64-41ef-a160-f289eb47d489",
 CellID->1754932694],
Cell[CellGroupData[{
Cell[28817901, 652243, 420, 13, 27, "Input",ExpressionUUID->"d12f2ad3-5192-4e0a-8e40-dd0ae4eaa674",
 CellID->1613380160],
Cell[28818324, 652258, 169, 3, 26, "Output",ExpressionUUID->"aa048c02-49a6-41e0-9a55-b16da77065dd",
 CellID->687433615]
}, Open  ]],
Cell[CellGroupData[{
Cell[28818530, 652266, 162, 3, 27, "Input",ExpressionUUID->"29fa3ca7-514b-46c3-b1f5-d204aa2d754a",
 CellID->1461293161],
Cell[28818695, 652271, 173, 3, 26, "Output",ExpressionUUID->"2f39849f-adf9-43ae-90ea-6b9892a877db",
 CellID->2078357721]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[28818917, 652280, 192, 4, 40, "MathCaption",ExpressionUUID->"da04e4a8-3146-4683-8bb9-f04170c81e8d",
 CellID->486782399],
Cell[CellGroupData[{
Cell[28819134, 652288, 130, 2, 27, "Input",ExpressionUUID->"33210f84-34e9-4ed2-b698-ac507149334c",
 CellID->1586568570],
Cell[28819267, 652292, 2570, 39, 44, "Message",ExpressionUUID->"5abc99a0-31d8-49b8-843a-e365cc7f1778",
 CellID->731989197],
Cell[28821840, 652333, 416, 9, 22, "Message",ExpressionUUID->"7b5e3b88-71c7-48ea-9472-517c9bb48e89",
 CellID->1756103963],
Cell[28822259, 652344, 235, 5, 26, "Output",ExpressionUUID->"a1a799e1-1311-447d-a9fc-bc7a820286f4",
 CellID->1971545420]
}, Open  ]],
Cell[CellGroupData[{
Cell[28822531, 652354, 337, 10, 27, "Input",ExpressionUUID->"5e1fed67-9814-41b7-b822-747c1938f019",
 CellID->2010255759],
Cell[28822871, 652366, 146162, 3451, 256, "Output",ExpressionUUID->"7d4b1263-61c8-47d7-86df-c87dc3de5e6f",
 CellID->1816803118]
}, Open  ]]
}, Open  ]]
}, Closed]]
}, Open  ]],
Cell[CellGroupData[{
Cell[28969106, 655825, 633, 11, 54, "Section",ExpressionUUID->"805539ef-7108-4523-9d24-2b919bf16083",
 CellTags->{"S5.18.1", "1.4"},
 CellID->1577395714],
Cell[CellGroupData[{
Cell[28969764, 655840, 297, 5, 39, "Subsection",ExpressionUUID->"bfe53780-fd39-4e4e-893e-588c67b819d0",
 CellID->62024482],
Cell[CellGroupData[{
Cell[28970086, 655849, 242, 4, 27, "Input",ExpressionUUID->"ab18394d-e13a-4e76-96c3-fb28eaecbdd8",
 CellID->950675144],
Cell[CellGroupData[{
Cell[28970353, 655857, 402, 6, 22, "Print",ExpressionUUID->"b316b630-1e29-407b-8914-39b46f136e77",
 CellID->2091747396],
Cell[28970758, 655865, 672, 14, 22, "Print",ExpressionUUID->"c9bb89b0-c65a-4642-9e96-70aeea598ee5",
 CellID->1691183182],
Cell[28971433, 655881, 421, 6, 22, "Print",ExpressionUUID->"b94f223c-38be-451b-8918-a9cd71aeba81",
 CellID->1769671829],
Cell[28971857, 655889, 377, 5, 22, "Print",ExpressionUUID->"1850d674-e5d4-409f-926b-4d15bcb316e9",
 CellID->225919053],
Cell[28972237, 655896, 360, 5, 22, "Print",ExpressionUUID->"480143ef-9b16-457f-9e1c-76b2dfc316f7",
 CellID->270435875],
Cell[28972600, 655903, 400, 6, 22, "Print",ExpressionUUID->"94ae02de-dbb5-4b90-a739-c8db3d7b420a",
 CellID->1140354073],
Cell[28973003, 655911, 676, 14, 22, "Print",ExpressionUUID->"20a8c96d-a1bc-44a3-90bf-40981c504c2d",
 CellID->1786839083],
Cell[28973682, 655927, 420, 6, 22, "Print",ExpressionUUID->"90b29920-e2a0-4ef8-8828-399ff621624f",
 CellID->993148333],
Cell[28974105, 655935, 400, 6, 22, "Print",ExpressionUUID->"59af40eb-cfc7-4486-b227-fb4f06c43d5c",
 CellID->1601128247],
Cell[28974508, 655943, 670, 14, 22, "Print",ExpressionUUID->"f11b693d-710c-4f31-bf50-51980d5fa69c",
 CellID->44866772],
Cell[28975181, 655959, 438, 6, 22, "Print",ExpressionUUID->"bb96e10d-0e58-4266-bc07-648da98ac212",
 CellID->309076907],
Cell[28975622, 655967, 400, 6, 22, "Print",ExpressionUUID->"e89f4839-648f-4622-b014-05182bb97456",
 CellID->2040117503],
Cell[28976025, 655975, 555, 8, 42, "Print",ExpressionUUID->"2da79110-5701-4d23-8ca2-f5429d1a31f0",
 CellID->1530964919],
Cell[28976583, 655985, 400, 6, 22, "Print",ExpressionUUID->"c2a5344f-ace1-444b-a0a4-e28a88b55eef",
 CellID->1925818926]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[28977032, 655997, 298, 7, 27, "Input",ExpressionUUID->"6c536c3e-df76-424e-8cdb-6747624148af",
 CellID->380146108],
Cell[CellGroupData[{
Cell[28977355, 656008, 733, 14, 22, "Print",ExpressionUUID->"d852119a-aa18-499b-aeb8-0a2498028da0",
 CellID->584682180],
Cell[28978091, 656024, 744, 14, 22, "Print",ExpressionUUID->"ec161548-a0fb-416e-8ff3-b620cdfc1d59",
 CellID->2015714235]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[28978884, 656044, 544, 16, 47, "Input",ExpressionUUID->"0eafa0e3-dea9-414a-8891-a1680a1273aa",
 CellID->139030448],
Cell[CellGroupData[{
Cell[28979453, 656064, 764, 14, 22, "Print",ExpressionUUID->"d602889d-eae9-498d-8271-2b215fcc01c8",
 CellID->2126631414],
Cell[28980220, 656080, 849, 17, 22, "Print",ExpressionUUID->"c59fc2a5-dc8e-4b8a-8a29-24f7b9545473",
 CellID->1723838077]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[28981118, 656103, 264, 6, 27, "Input",ExpressionUUID->"4e3d7d18-3836-4177-a83f-b6a43f4649fa",
 CellID->1700707167],
Cell[CellGroupData[{
Cell[28981407, 656113, 722, 14, 22, "Print",ExpressionUUID->"44062c32-4269-477a-829e-ec76593bcac8",
 CellID->62698460],
Cell[28982132, 656129, 824, 16, 22, "Print",ExpressionUUID->"15aaf5e3-b3f3-4817-a75d-63ae51949296",
 CellID->1506030190],
Cell[28982959, 656147, 922, 19, 22, "Print",ExpressionUUID->"8e696d68-8f2b-4c47-822d-1d6aff5b193e",
 CellID->1607139788],
Cell[28983884, 656168, 975, 21, 22, "Print",ExpressionUUID->"44534c5f-f066-4aa3-9f1a-5a2349399510",
 CellID->149104950],
Cell[28984862, 656191, 994, 21, 22, "Print",ExpressionUUID->"48023003-f168-49a3-9528-569f8af55091",
 CellID->2040962248],
Cell[28985859, 656214, 911, 19, 22, "Print",ExpressionUUID->"b88eb820-29af-419d-8c16-098209a06686",
 CellID->696208605],
Cell[28986773, 656235, 896, 18, 22, "Print",ExpressionUUID->"74ebb7b9-84e0-46cd-96f8-765a10706b24",
 CellID->1687115379],
Cell[28987672, 656255, 966, 21, 22, "Print",ExpressionUUID->"fa9e1df8-6e93-4c2a-b8eb-2ef5c3537dda",
 CellID->2121583512]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[28988687, 656282, 386, 8, 27, "Input",ExpressionUUID->"586d67d7-298f-43dd-80f4-0dd0d12a0c9b",
 CellID->384010265],
Cell[CellGroupData[{
Cell[28989098, 656294, 738, 14, 22, "Print",ExpressionUUID->"e64e745c-c74d-4cce-87a2-f0edc3330010",
 CellID->1347433818],
Cell[28989839, 656310, 836, 16, 22, "Print",ExpressionUUID->"b2de2025-2e55-4853-b54c-bb8d1d32c508",
 CellID->1105121511],
Cell[28990678, 656328, 935, 19, 22, "Print",ExpressionUUID->"0944c7fa-fbb3-48bb-ae5a-da9a0c093898",
 CellID->191172474],
Cell[28991616, 656349, 987, 21, 22, "Print",ExpressionUUID->"71609261-83b7-41ab-99f5-ed70f536996e",
 CellID->400234606],
Cell[28992606, 656372, 1007, 21, 22, "Print",ExpressionUUID->"e308e820-f449-4ee1-850d-063010d89392",
 CellID->824240638],
Cell[28993616, 656395, 927, 20, 22, "Print",ExpressionUUID->"7872fd95-f223-458f-a88d-3b918f45f731",
 CellID->1742879716],
Cell[28994546, 656417, 907, 18, 22, "Print",ExpressionUUID->"ba88b171-611e-4640-bae3-888ed74739e9",
 CellID->902441732],
Cell[28995456, 656437, 978, 21, 22, "Print",ExpressionUUID->"429ffe24-7887-44a1-b4bb-5ac38fd58627",
 CellID->1808674610]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[28996495, 656465, 272, 4, 39, "Subsection",ExpressionUUID->"304f8903-4ce4-4b85-8882-0ad638cee8a5",
 CellID->25383259],
Cell[28996770, 656471, 1144, 17, 348, "Text",ExpressionUUID->"ef1e14bf-3778-459c-a757-ad199c2d89ca",
 CellID->1414671907]
}, Closed]],
Cell[CellGroupData[{
Cell[28997951, 656493, 312, 5, 39, "Subsection",ExpressionUUID->"1f61dbb7-79fd-47da-ba7f-326b8fbe6e9d",
 CellID->28781190],
Cell[28998266, 656500, 585, 11, 92, "DefinitionBox",ExpressionUUID->"a7487b0d-4d4f-44a3-815b-af6141d42a63",
 CellID->1277870832],
Cell[28998854, 656513, 170, 2, 23, "Text",ExpressionUUID->"9c2310bc-14b9-4b43-af44-b64cbf2fb2cc",
 CellID->191119777],
Cell[CellGroupData[{
Cell[28999049, 656519, 4510, 127, 235, "Text",ExpressionUUID->"8fafaf2f-c0e3-46ad-a85a-0ccc8ac4a180",
 CellID->1662039774],
Cell[CellGroupData[{
Cell[29003584, 656650, 152, 3, 27, "Input",ExpressionUUID->"a51aae78-aeab-45e4-9bd7-9bfd07bb4852",
 CellID->1582685089],
Cell[29003739, 656655, 386, 5, 67, "Print",ExpressionUUID->"37a7b42a-eb42-47c4-a943-ae80389a898c",
 CellTags->"Info73730895176-7882880",
 CellID->745676275]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[29004174, 656666, 107, 1, 40, "MathCaption",ExpressionUUID->"3b39fbbc-d0cf-4779-9052-22d06eb19d4e",
 CellID->2006244096],
Cell[CellGroupData[{
Cell[29004306, 656671, 407, 12, 27, "Input",ExpressionUUID->"7f01e485-86a2-4c52-b8bf-68896674c4c1",
 CellID->1303779925],
Cell[29004716, 656685, 263, 5, 26, "Output",ExpressionUUID->"ec26cdab-3871-4210-8735-986b6a6a786f",
 CellID->1127715014]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[29005028, 656696, 2194, 64, 78, "Text",ExpressionUUID->"ff79b072-bbc4-4fcb-9ec5-7b23a3ae6995",
 CellID->1202511756],
Cell[CellGroupData[{
Cell[29007247, 656764, 146, 3, 27, "Input",ExpressionUUID->"da1d17a4-a397-4373-b561-9ce49be87046",
 CellID->646481175],
Cell[29007396, 656769, 854, 12, 111, "Print",ExpressionUUID->"cd30eeb0-4a22-42fb-9239-de444f399b1e",
 CellTags->"Info93730895182-7882880",
 CellID->1068687520]
}, Open  ]]
}, Open  ]],
Cell[29008277, 656785, 886, 24, 145, "Text",ExpressionUUID->"b0ab2621-cff1-4048-9103-de5412f7619c",
 CellID->1794641734],
Cell[29009166, 656811, 2044, 63, 119, "Text",ExpressionUUID->"3dd0f2b6-1328-4b26-9036-552d3f9151ab",
 CellID->1135801223]
}, Closed]],
Cell[CellGroupData[{
Cell[29011247, 656879, 287, 5, 39, "Subsection",ExpressionUUID->"e248971e-3f48-4520-a35f-05ae5e6a11e1",
 CellID->429016778],
Cell[29011537, 656886, 691, 13, 109, "DefinitionBox",ExpressionUUID->"abdb9eb6-d49a-4292-bb2b-4e69f1dc3fea",
 CellID->1382384867],
Cell[29012231, 656901, 163, 2, 23, "Text",ExpressionUUID->"1b172dc7-54f7-432e-8d34-421622e26724",
 CellID->116882776],
Cell[CellGroupData[{
Cell[29012419, 656907, 484, 11, 42, "Text",ExpressionUUID->"5000064a-3ee1-4d8c-a958-239e64756c83",
 CellID->1648762874],
Cell[CellGroupData[{
Cell[29012928, 656922, 157, 3, 27, "Input",ExpressionUUID->"8f33011f-a881-4417-a49d-39532004b9e1",
 CellID->1776126077],
Cell[29013088, 656927, 737, 10, 111, "Print",ExpressionUUID->"0c931954-88f4-4d3a-83e2-46600457c59c",
 CellTags->"Info103730895193-7882880",
 CellID->1879930841]
}, Open  ]]
}, Open  ]],
Cell[29013852, 656941, 250, 5, 23, "Text",ExpressionUUID->"f50445c3-dcf7-495a-bfeb-9f7926b3cfc0",
 CellID->1357164508],
Cell[CellGroupData[{
Cell[29014127, 656950, 212, 2, 40, "MathCaption",ExpressionUUID->"09cd64c3-fffe-4730-b2b6-aa0686f27313",
 CellID->1315450794],
Cell[CellGroupData[{
Cell[29014364, 656956, 362, 9, 27, "Input",ExpressionUUID->"475b3701-8659-4464-9be4-529e753e79f3",
 CellID->1440071169],
Cell[29014729, 656967, 838, 15, 22, "Print",ExpressionUUID->"2ae9d5ab-b66b-4aa1-82ee-31977acccb1e",
 CellID->2059169265]
}, Open  ]],
Cell[CellGroupData[{
Cell[29015604, 656987, 247, 5, 27, "Input",ExpressionUUID->"eb908b1a-97b1-4217-b6fb-bdff64fec3a2",
 CellID->728005009],
Cell[29015854, 656994, 555, 11, 26, "Output",ExpressionUUID->"1d2f6bc7-e4bf-4876-b514-24a1642dfb5e",
 CellID->1456202210]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[29016458, 657011, 161, 2, 40, "MathCaption",ExpressionUUID->"01adfdeb-ec38-4b7e-aaa6-2000c015bcea",
 CellID->1376549545],
Cell[CellGroupData[{
Cell[29016644, 657017, 345, 10, 27, "Input",ExpressionUUID->"cd9779f8-53b8-4ad1-bda8-cea63f1e34fb",
 CellID->826873684],
Cell[29016992, 657029, 1479, 40, 32, "Output",ExpressionUUID->"820b7dc3-d2c4-4989-8575-61d0dc282677",
 CellID->911245061]
}, Open  ]],
Cell[CellGroupData[{
Cell[29018508, 657074, 167, 3, 27, "Input",ExpressionUUID->"77daa12f-2b7b-4dd3-a1b9-28116b219f35",
 CellID->1003199971],
Cell[29018678, 657079, 3368, 85, 28, "Print",ExpressionUUID->"d87ff284-11e2-4227-9248-8901a284ed20",
 CellID->1620176819],
Cell[29022049, 657166, 2740, 76, 32, "Output",ExpressionUUID->"68cde602-955c-4465-bfc1-81a942c6736c",
 CellID->219152973]
}, Open  ]],
Cell[CellGroupData[{
Cell[29024826, 657247, 164, 3, 27, "Input",ExpressionUUID->"4c541fce-7a3f-4427-8ef8-05f8f0155eef",
 CellID->673242544],
Cell[29024993, 657252, 3079, 82, 32, "Output",ExpressionUUID->"bf39ca3d-ca62-401e-91d2-ed7687bbf5d0",
 CellID->2138622509]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[29028121, 657340, 586, 14, 57, "MathCaption",ExpressionUUID->"b343f231-2ece-447a-8947-70083c1288f7",
 CellID->261216211],
Cell[CellGroupData[{
Cell[29028732, 657358, 393, 11, 27, "Input",ExpressionUUID->"c5766ee7-55b3-4727-b9c7-923175f710fc",
 CellID->804283809],
Cell[29029128, 657371, 3433, 88, 28, "Print",ExpressionUUID->"e3343d74-3d9d-4319-836f-84b983d22a23",
 CellID->360939735],
Cell[29032564, 657461, 2789, 77, 32, "Output",ExpressionUUID->"22ee4409-db33-4be2-ace0-0273715da9d3",
 CellID->1820307017]
}, Open  ]],
Cell[CellGroupData[{
Cell[29035390, 657543, 165, 3, 27, "Input",ExpressionUUID->"ea88911a-18dd-4196-b4de-d9c511497a73",
 CellID->1365141535],
Cell[29035558, 657548, 5793, 155, 32, "Output",ExpressionUUID->"022d0826-a20e-4032-8e86-2275b9bcfe41",
 CellID->1479780630]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[29041400, 657709, 520, 14, 57, "MathCaption",ExpressionUUID->"f022ee5d-1202-4045-aab3-166dbab1ea72",
 CellID->34932609],
Cell[CellGroupData[{
Cell[29041945, 657727, 430, 13, 27, "Input",ExpressionUUID->"5164f5e8-9871-4b42-8278-f16b74efbd02",
 CellID->849323832],
Cell[29042378, 657742, 33976, 952, 56, "Output",ExpressionUUID->"b09810c7-0af4-423d-b8aa-7fdc143ad890",
 CellID->1791544785]
}, Open  ]],
Cell[CellGroupData[{
Cell[29076391, 658699, 167, 3, 27, "Input",ExpressionUUID->"967611ed-f59a-47a0-b013-4ffd36a01d9e",
 CellID->598199002],
Cell[CellGroupData[{
Cell[29076583, 658706, 1993, 49, 28, "Print",ExpressionUUID->"5a36791a-8da6-4927-bc09-70ac4f86d2b1",
 CellID->1727926895],
Cell[29078579, 658757, 1992, 49, 28, "Print",ExpressionUUID->"82359f14-8d2a-4bf2-a481-329480350d57",
 CellID->845965884],
Cell[29080574, 658808, 1993, 49, 28, "Print",ExpressionUUID->"a700f923-1f9e-4eaa-a7c1-2e77d65ff4ad",
 CellID->1806869248],
Cell[29082570, 658859, 1994, 49, 28, "Print",ExpressionUUID->"ae2f955b-c270-4343-8a8b-bb72105933a1",
 CellID->485198909],
Cell[29084567, 658910, 1992, 49, 28, "Print",ExpressionUUID->"f3bc0856-40e8-4c8a-94f2-84b4cfbf0f3f",
 CellID->742655904],
Cell[29086562, 658961, 1992, 49, 28, "Print",ExpressionUUID->"3de0d4dd-ec52-436d-8834-7140194cfc5d",
 CellID->642822164],
Cell[29088557, 659012, 3432, 88, 28, "Print",ExpressionUUID->"0b8b82ee-695a-43c2-b8e7-4ae284d9c6d0",
 CellID->2069365938],
Cell[29091992, 659102, 1992, 49, 28, "Print",ExpressionUUID->"302d5d6e-c83b-48a0-b10f-8635747d850d",
 CellID->691311822],
Cell[29093987, 659153, 1994, 49, 28, "Print",ExpressionUUID->"60069d13-db73-4ac4-b9fe-c270d8dcb531",
 CellID->991262084],
Cell[29095984, 659204, 1993, 49, 28, "Print",ExpressionUUID->"b0cd6779-a498-4202-b7f1-c652950753c9",
 CellID->2111196009],
Cell[29097980, 659255, 3434, 88, 28, "Print",ExpressionUUID->"1dc5c7c9-2293-486e-b105-231d6288769e",
 CellID->2146507529],
Cell[29101417, 659345, 1992, 49, 28, "Print",ExpressionUUID->"fbcc3b2d-f5e4-4946-b4c7-feb44c1c5b7b",
 CellID->774253950],
Cell[29103412, 659396, 1993, 49, 28, "Print",ExpressionUUID->"61959597-ca53-464f-9999-f97754775411",
 CellID->1282720477],
Cell[29105408, 659447, 1991, 49, 28, "Print",ExpressionUUID->"c201c77a-67f4-43b8-bac0-855d90537900",
 CellID->94724706],
Cell[29107402, 659498, 3359, 85, 28, "Print",ExpressionUUID->"08468eee-9868-4c3c-9c3e-c26e62ca0616",
 CellID->752751315],
Cell[29110764, 659585, 1992, 49, 28, "Print",ExpressionUUID->"5eb96227-18c9-4b2f-94fa-76cd011daa89",
 CellID->665186728],
Cell[29112759, 659636, 1993, 49, 28, "Print",ExpressionUUID->"c3c92aff-a9b5-4269-b99e-3f529ff81f81",
 CellID->2140954861],
Cell[29114755, 659687, 1993, 49, 28, "Print",ExpressionUUID->"daef8708-9b35-46c0-8f39-0f4d91940bde",
 CellID->1939843342],
Cell[29116751, 659738, 3359, 85, 28, "Print",ExpressionUUID->"a92e1a87-2e0a-4871-a78e-2ff6661c9c3e",
 CellID->1997803887],
Cell[29120113, 659825, 1993, 49, 28, "Print",ExpressionUUID->"595c4ab2-0641-420a-8ef2-f02aad16eb7b",
 CellID->1183107964],
Cell[29122109, 659876, 1994, 49, 28, "Print",ExpressionUUID->"b887ed34-231b-45cf-ad3b-b2cde09d20b1",
 CellID->944463375],
Cell[29124106, 659927, 1995, 49, 28, "Print",ExpressionUUID->"8354e3a4-0622-42b9-b54a-0446d967c360",
 CellID->1544308648],
Cell[29126104, 659978, 1994, 49, 28, "Print",ExpressionUUID->"84fc46d4-3e53-4942-881c-b47a7bcab337",
 CellID->708089294],
Cell[29128101, 660029, 1992, 49, 28, "Print",ExpressionUUID->"d811d90b-231d-4979-98c2-0688b3223dc5",
 CellID->291464229],
Cell[29130096, 660080, 1994, 49, 28, "Print",ExpressionUUID->"aa284a8a-1a63-4df4-88bf-2016a5d6b2e0",
 CellID->229097996]
}, Open  ]],
Cell[29132105, 660132, 43555, 1195, 56, "Output",ExpressionUUID->"1e4eecfb-f1cc-4aeb-838e-95d5564163f9",
 CellID->175699453]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[29175709, 661333, 625, 18, 41, "MathCaption",ExpressionUUID->"ec014675-e171-40eb-83d3-50d600ca97c4",
 CellID->1868763863],
Cell[CellGroupData[{
Cell[29176359, 661355, 165, 3, 27, "Input",ExpressionUUID->"f7e9ac07-ba86-4722-8d7d-3da38f1fe8f9",
 CellID->1365908066],
Cell[29176527, 661360, 45955, 1200, 56, "Output",ExpressionUUID->"bf6c0915-6afc-4d9f-8dc1-3b4b13553e42",
 CellID->1704148742]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[29222531, 662566, 429, 11, 41, "MathCaption",ExpressionUUID->"68fc9241-7d70-428a-aef9-c820be3c4055",
 CellID->1736347759],
Cell[CellGroupData[{
Cell[29222985, 662581, 164, 3, 27, "Input",ExpressionUUID->"c3eb6f7f-5d3a-4cc4-beba-23d261671224",
 CellID->641506430],
Cell[29223152, 662586, 490, 12, 26, "Output",ExpressionUUID->"3ff14788-a495-4497-8e75-da68b6115b34",
 CellID->969769461]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[29223691, 662604, 158, 4, 41, "MathCaption",ExpressionUUID->"a8d0318c-7004-4a95-a800-4516d730231e",
 CellID->1893691678],
Cell[CellGroupData[{
Cell[29223874, 662612, 165, 3, 27, "Input",ExpressionUUID->"806a1216-f75b-441b-bcff-38d7775ad7e5",
 CellID->1568383606],
Cell[29224042, 662617, 457, 10, 26, "Output",ExpressionUUID->"fc77f001-bc28-4c09-97c8-e3220f168385",
 CellID->72603651]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[29224548, 662633, 469, 12, 42, "MathCaption",ExpressionUUID->"0aa51283-12be-4090-bcea-067baa0999e7",
 CellID->1059231524],
Cell[CellGroupData[{
Cell[29225042, 662649, 377, 11, 27, "Input",ExpressionUUID->"b82cddca-96b9-4d3c-a193-2a4bcb13009e",
 CellID->2012273885],
Cell[29225422, 662662, 3171, 81, 28, "Print",ExpressionUUID->"9698d864-b653-4199-9f9d-efde0416f750",
 CellID->375628720],
Cell[29228596, 662745, 2620, 71, 32, "Output",ExpressionUUID->"d833bb2c-1d25-412e-9d04-18e25e6b5dba",
 CellID->1737458507]
}, Open  ]],
Cell[CellGroupData[{
Cell[29231253, 662821, 419, 13, 27, "Input",ExpressionUUID->"0cc380eb-975f-40bb-87aa-2126b3ed4689",
 CellID->472610126],
Cell[29231675, 662836, 10620, 295, 32, "Output",ExpressionUUID->"27cc14bf-57dc-4a02-9703-b2024c6bd0af",
 CellID->1916977214]
}, Open  ]],
Cell[CellGroupData[{
Cell[29242332, 663136, 173, 3, 27, "Input",ExpressionUUID->"590e6f78-b634-4f51-9980-55855a609a4a",
 CellID->1348841144],
Cell[CellGroupData[{
Cell[29242530, 663143, 3175, 81, 28, "Print",ExpressionUUID->"eb307d47-ce33-4881-b439-5b6550346075",
 CellID->1041221882],
Cell[29245708, 663226, 3113, 79, 28, "Print",ExpressionUUID->"202b303b-18ac-4c41-a272-cf8dcbfede79",
 CellID->2119415018],
Cell[29248824, 663307, 3115, 79, 28, "Print",ExpressionUUID->"b12f95cf-c0f1-4ca4-ae37-e5190ec7f03c",
 CellID->696954737],
Cell[29251942, 663388, 3174, 81, 28, "Print",ExpressionUUID->"251149aa-04d7-417b-b2cd-0284e0e12a8a",
 CellID->1342454021],
Cell[29255119, 663471, 3175, 81, 28, "Print",ExpressionUUID->"05ac1580-a4c3-46f6-9dd4-9b3950d22280",
 CellID->1161535565],
Cell[29258297, 663554, 3114, 79, 28, "Print",ExpressionUUID->"cee429dd-7af5-43f0-9254-cf0a38908f92",
 CellID->1788456507],
Cell[29261414, 663635, 3176, 81, 28, "Print",ExpressionUUID->"594ea64e-5327-4ff9-9f64-323f5d28312c",
 CellID->259906090],
Cell[29264593, 663718, 3174, 81, 28, "Print",ExpressionUUID->"c215447a-468b-4e89-a3a5-e4f70f444c5f",
 CellID->259916632]
}, Open  ]],
Cell[29267782, 663802, 22028, 604, 32, "Output",ExpressionUUID->"e9b63519-d193-4960-8222-b2e43cdda579",
 CellID->437096026]
}, Open  ]],
Cell[CellGroupData[{
Cell[29289847, 664411, 164, 3, 27, "Input",ExpressionUUID->"9f45667a-aa14-4ecc-a11b-940827b4d299",
 CellID->106298051],
Cell[29290014, 664416, 656, 19, 26, "Output",ExpressionUUID->"dfa6b67c-6c0e-44fe-a655-22fdc91a3299",
 CellID->1193783197]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[29290719, 664441, 736, 24, 42, "MathCaption",ExpressionUUID->"972e561a-9898-4142-8e3a-2e3dee0d72c3",
 CellID->1100519829],
Cell[CellGroupData[{
Cell[29291480, 664469, 419, 13, 27, "Input",ExpressionUUID->"4d2d8a56-2935-4ba0-934d-8d5eae44db6e",
 CellID->160114392],
Cell[29291902, 664484, 10588, 295, 32, "Output",ExpressionUUID->"c4d9b66a-1be9-4de4-af50-295dc646c78c",
 CellID->150839986]
}, Open  ]],
Cell[CellGroupData[{
Cell[29302527, 664784, 173, 3, 27, "Input",ExpressionUUID->"eae75449-16db-48db-ac0b-ec783ac7334d",
 CellID->2090721084],
Cell[CellGroupData[{
Cell[29302725, 664791, 3088, 79, 28, "Print",ExpressionUUID->"520d5762-bfc4-4d48-a0d3-3f326e336f43",
 CellID->1358305486],
Cell[29305816, 664872, 3088, 79, 28, "Print",ExpressionUUID->"fc9a5c20-dcf5-4e69-b904-7fccbb198b48",
 CellID->1289135592],
Cell[29308907, 664953, 3087, 79, 28, "Print",ExpressionUUID->"da775947-2fbc-4c58-b0b5-285b2fd81061",
 CellID->539380161],
Cell[29311997, 665034, 3144, 80, 28, "Print",ExpressionUUID->"47e8c304-e785-46a3-b124-322bd958ace3",
 CellID->833891940],
Cell[29315144, 665116, 3087, 79, 28, "Print",ExpressionUUID->"b3002e9f-c309-4990-a753-8d3ff085ba40",
 CellID->743555258],
Cell[29318234, 665197, 3087, 79, 28, "Print",ExpressionUUID->"1fa19b9e-a93e-40b6-b253-45da3530220f",
 CellID->725879313],
Cell[29321324, 665278, 3145, 80, 28, "Print",ExpressionUUID->"8b579b3a-246e-4f34-9353-d4a74cca346d",
 CellID->2146498631],
Cell[29324472, 665360, 3144, 80, 28, "Print",ExpressionUUID->"7cd88ef7-12c4-4778-8a39-df7d4b643c41",
 CellID->624705264],
Cell[29327619, 665442, 3087, 79, 28, "Print",ExpressionUUID->"596fe376-6224-4098-b443-f5597212e90f",
 CellID->373821865]
}, Open  ]],
Cell[29330721, 665524, 21810, 598, 32, "Output",ExpressionUUID->"ab69764c-9f01-48ef-a555-442ce3d27aeb",
 CellID->1434417580]
}, Open  ]],
Cell[CellGroupData[{
Cell[29352568, 666127, 164, 3, 27, "Input",ExpressionUUID->"99abb788-1ea3-4e18-9477-399b236b9302",
 CellID->356162247],
Cell[29352735, 666132, 46170, 1226, 56, "Output",ExpressionUUID->"dd762f07-93b9-4cd0-a0b2-aaca22f05a75",
 CellID->841919323]
}, Open  ]],
Cell[CellGroupData[{
Cell[29398942, 667363, 164, 3, 27, "Input",ExpressionUUID->"327f707a-a81f-4fe6-97bb-091622f4886f",
 CellID->409052785],
Cell[29399109, 667368, 649, 19, 26, "Output",ExpressionUUID->"986f8825-5f65-4b64-b8ad-d214edfe2a76",
 CellID->1800952333]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[29399807, 667393, 1225, 40, 60, "MathCaption",ExpressionUUID->"03068d6a-8a39-4d9b-ba3e-4543cd89f9eb",
 CellID->1229868459],
Cell[CellGroupData[{
Cell[29401057, 667437, 399, 12, 27, "Input",ExpressionUUID->"35723d94-8072-420e-b892-49fe7600370b",
 CellID->869088038],
Cell[29401459, 667451, 10581, 295, 32, "Output",ExpressionUUID->"0dfd7c9e-b3b2-435d-bab3-253751f5017b",
 CellID->355857753]
}, Open  ]],
Cell[CellGroupData[{
Cell[29412077, 667751, 172, 3, 27, "Input",ExpressionUUID->"99656b06-9b33-47c6-9a98-dadc428aa399",
 CellID->472884516],
Cell[CellGroupData[{
Cell[29412274, 667758, 3149, 81, 28, "Print",ExpressionUUID->"8a4934c1-847e-4602-8f2b-c4cd10a8e83f",
 CellID->2099191193],
Cell[29415426, 667841, 3080, 78, 28, "Print",ExpressionUUID->"9a908228-7833-4c5d-a185-ae5067070fe2",
 CellID->97544188],
Cell[29418509, 667921, 3083, 78, 28, "Print",ExpressionUUID->"c3bc196a-727b-4af8-b845-3e8c02b5591b",
 CellID->895145055],
Cell[29421595, 668001, 3149, 81, 28, "Print",ExpressionUUID->"56ce7bf8-3d30-4fab-b90b-acef8d75ff65",
 CellID->1579992150],
Cell[29424747, 668084, 3147, 81, 28, "Print",ExpressionUUID->"c747e527-a9da-42f9-a80a-bd360e94e5e4",
 CellID->1292769895],
Cell[29427897, 668167, 3080, 78, 28, "Print",ExpressionUUID->"b4404af1-1d48-432b-ba1c-9761f1183b11",
 CellID->76266324],
Cell[29430980, 668247, 3147, 81, 28, "Print",ExpressionUUID->"e566e3a0-a7ba-441b-a27d-50c2070b2516",
 CellID->1907853856],
Cell[29434130, 668330, 3149, 81, 28, "Print",ExpressionUUID->"85651165-83cb-432f-9a6b-27a51aa1ab21",
 CellID->1221424435],
Cell[29437282, 668413, 3148, 81, 28, "Print",ExpressionUUID->"23400bde-2282-4148-abdd-7533e1734029",
 CellID->684297372]
}, Open  ]],
Cell[29440445, 668497, 21987, 604, 32, "Output",ExpressionUUID->"7aee3a74-2357-4e72-9be7-bc86b1b4b9d3",
 CellID->1211001519]
}, Open  ]],
Cell[CellGroupData[{
Cell[29462469, 669106, 399, 12, 27, "Input",ExpressionUUID->"d87201b3-cacc-4cbd-9892-ced8b530673e",
 CellID->641639522],
Cell[29462871, 669120, 10580, 295, 32, "Output",ExpressionUUID->"8beeee48-ca51-4ca2-8ffe-5380ac3d7b32",
 CellID->2024367663]
}, Open  ]],
Cell[CellGroupData[{
Cell[29473488, 669420, 173, 3, 27, "Input",ExpressionUUID->"342f5ca0-100c-435e-91a9-6ae297c5e9fa",
 CellID->1093323557],
Cell[CellGroupData[{
Cell[29473686, 669427, 3081, 78, 28, "Print",ExpressionUUID->"6bea4520-88b1-4400-a3fd-926d646164a1",
 CellID->693721341],
Cell[29476770, 669507, 3079, 78, 28, "Print",ExpressionUUID->"e56c202a-d5c2-4859-94e9-6279d85935b8",
 CellID->685858624],
Cell[29479852, 669587, 3080, 78, 28, "Print",ExpressionUUID->"b305cd9e-7f27-47da-a2cf-0e62e73aa05b",
 CellID->1681160119],
Cell[29482935, 669667, 3140, 80, 28, "Print",ExpressionUUID->"9d0163bd-7dc5-4639-8857-3666273ba017",
 CellID->815776454],
Cell[29486078, 669749, 3079, 78, 28, "Print",ExpressionUUID->"4756dd63-49f2-4a43-b83f-1b128d27393c",
 CellID->1219259930],
Cell[29489160, 669829, 3079, 78, 28, "Print",ExpressionUUID->"5c49f10c-d645-44c0-975c-a5b651c25c27",
 CellID->530736966],
Cell[29492242, 669909, 3143, 80, 28, "Print",ExpressionUUID->"68b01eeb-faf9-4807-b7a5-b9b114ebb9f5",
 CellID->1596601571],
Cell[29495388, 669991, 3141, 80, 28, "Print",ExpressionUUID->"a9ce7a5a-572d-4dd1-b12f-22bfe5de972e",
 CellID->1274434532],
Cell[29498532, 670073, 3079, 78, 28, "Print",ExpressionUUID->"b601d8a1-c0fe-4846-93bf-cf9de3b8b324",
 CellID->419284146]
}, Open  ]],
Cell[29501626, 670154, 21796, 598, 32, "Output",ExpressionUUID->"3ab851e9-124d-4751-9f6c-a9fc063c3e0f",
 CellID->1014796015]
}, Open  ]],
Cell[CellGroupData[{
Cell[29523459, 670757, 165, 3, 27, "Input",ExpressionUUID->"53f239b8-9ba6-44f3-8e2e-d2b4c4381418",
 CellID->1026717895],
Cell[29523627, 670762, 986, 30, 26, "Output",ExpressionUUID->"a7280395-7d2a-41cf-bfd4-47ca5325f297",
 CellID->1881863052]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[29524662, 670798, 392, 10, 41, "MathCaption",ExpressionUUID->"a7417fde-1046-4c5d-868b-ff0b612a90ad",
 CellID->2051507455],
Cell[CellGroupData[{
Cell[29525079, 670812, 442, 14, 27, "Input",ExpressionUUID->"71a9234b-b600-4777-a11d-0ae6440c3f15",
 CellID->1835843600],
Cell[29525524, 670828, 45999, 1208, 56, "Output",ExpressionUUID->"92c64f40-dfed-4422-9247-52179e83909c",
 CellID->2133326197]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[29571572, 672042, 127, 1, 40, "MathCaption",ExpressionUUID->"169747f2-ae56-4bdf-9c19-50e2b4400040",
 CellID->1118382918],
Cell[CellGroupData[{
Cell[29571724, 672047, 165, 3, 27, "Input",ExpressionUUID->"97e51d4e-b578-49ae-8148-42f09b5ed989",
 CellID->1884824282],
Cell[29571892, 672052, 91841, 2427, 104, "Output",ExpressionUUID->"ca0d79af-6688-441c-980e-275a4ff21e75",
 CellID->213005585]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[29663782, 674485, 167, 2, 40, "MathCaption",ExpressionUUID->"9345b3bc-7590-41e6-8678-84f705a10e40",
 CellID->199311019],
Cell[CellGroupData[{
Cell[29663974, 674491, 414, 13, 27, "Input",ExpressionUUID->"976690c4-4194-4a98-a03a-0f34e109a0be",
 CellID->1893290660],
Cell[29664391, 674506, 3093, 79, 28, "Print",ExpressionUUID->"e0966399-a6b3-40ed-9e69-d74072af37c5",
 CellID->673504539],
Cell[29667487, 674587, 1422, 38, 32, "Output",ExpressionUUID->"b59666d3-a62d-4807-9c8d-6fddda04cb7c",
 CellID->204483636]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[29668958, 674631, 243, 6, 41, "MathCaption",ExpressionUUID->"7ffedb75-042b-49bb-ae6d-ecab9469bc40",
 CellID->339618327],
Cell[CellGroupData[{
Cell[29669226, 674641, 171, 3, 27, "Input",ExpressionUUID->"ef0f8cd9-b5a6-4136-9a49-866a8d3392a0",
 CellID->2101171810],
Cell[29669400, 674646, 899, 19, 22, "Print",ExpressionUUID->"4d13ca9f-33dc-4dd4-9add-411571941ab1",
 CellID->2120000322]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[29670348, 674671, 209, 5, 41, "MathCaption",ExpressionUUID->"a3a725ab-52c5-452e-b216-9edc5116fcb8",
 CellID->850775366],
Cell[CellGroupData[{
Cell[29670582, 674680, 419, 12, 27, "Input",ExpressionUUID->"63d5bcab-202d-441e-9ae7-4e61609b6fec",
 CellID->1035454636],
Cell[29671004, 674694, 1061, 24, 22, "Print",ExpressionUUID->"99e19f52-e553-4e38-907f-1513bfbed6b3",
 CellID->229910653]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[29672126, 674725, 279, 4, 39, "Subsection",ExpressionUUID->"8ba8375c-3190-4711-9a10-32582d770382",
 CellID->1163994800],
Cell[CellGroupData[{
Cell[29672430, 674733, 638, 17, 43, "Text",ExpressionUUID->"8bf8be01-0f69-43ad-bbd4-681b9392ff3c",
 CellID->1832063077],
Cell[CellGroupData[{
Cell[29673093, 674754, 403, 12, 27, "Input",ExpressionUUID->"76257141-3971-4d43-ab65-b129a7a3258b",
 CellID->863604453],
Cell[29673499, 674768, 1813, 44, 28, "Print",ExpressionUUID->"ec3c37f8-8801-4ce2-ab08-0401f8d4435b",
 CellID->77600358],
Cell[29675315, 674814, 1390, 37, 32, "Output",ExpressionUUID->"2ff9247c-8e65-4b57-90b1-ccdebee3bfab",
 CellID->706643177]
}, Open  ]],
Cell[CellGroupData[{
Cell[29676742, 674856, 315, 9, 27, "Input",ExpressionUUID->"12df3aa3-194e-4fa3-97e2-6c1df1bf49f2",
 CellID->2057569429],
Cell[29677060, 674867, 1600, 42, 32, "Output",ExpressionUUID->"714e74e9-2fe0-46b9-8a91-4fa13364aa78",
 CellID->967307929]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[29678709, 674915, 147, 1, 23, "Text",ExpressionUUID->"844c275a-0412-4e33-8797-880f81cee20d",
 CellID->1377949064],
Cell[CellGroupData[{
Cell[29678881, 674920, 403, 12, 27, "Input",ExpressionUUID->"2e67a1f1-dc58-4dfa-9f78-b0c83a967d57",
 CellID->726269400],
Cell[CellGroupData[{
Cell[29679309, 674936, 3106, 79, 28, "Print",ExpressionUUID->"b44cf555-ea86-4d53-a56a-f5369cab6621",
 CellID->1093154913],
Cell[29682418, 675017, 1834, 45, 28, "Print",ExpressionUUID->"8f7e9e3d-cf6d-4b29-8cbd-07fdbf6bd733",
 CellID->932022706]
}, Open  ]],
Cell[29684267, 675065, 1384, 37, 32, "Output",ExpressionUUID->"3edaf57e-4fd7-4382-948d-260be6c78e08",
 CellID->759898524]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[29685700, 675108, 232, 4, 23, "Text",ExpressionUUID->"a4011bca-ea36-4b9e-b218-bd40ca109b01",
 CellID->261916468],
Cell[CellGroupData[{
Cell[29685957, 675116, 315, 9, 27, "Input",ExpressionUUID->"0228f996-fadd-43eb-92d8-82fbbec2ba66",
 CellID->1836084973],
Cell[29686275, 675127, 5408, 143, 32, "Output",ExpressionUUID->"67de174e-f312-4f7e-adba-e28afa6a00a9",
 CellID->529244159]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[29691732, 675276, 200, 4, 24, "Text",ExpressionUUID->"069cd779-6951-4532-b393-bebd20211aff",
 CellID->410867545],
Cell[CellGroupData[{
Cell[29691957, 675284, 427, 13, 27, "Input",ExpressionUUID->"daf15d97-5701-43dd-9638-2fe1124714dd",
 CellID->1290903710],
Cell[29692387, 675299, 1947, 48, 28, "Print",ExpressionUUID->"dcd810f3-f14f-4b9b-973b-9243aec7b9b3",
 CellID->129314439],
Cell[29694337, 675349, 1492, 40, 32, "Output",ExpressionUUID->"b9c986fb-2374-4aec-83f8-69662261f01c",
 CellID->134325509]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[29695878, 675395, 134, 1, 23, "Text",ExpressionUUID->"8ea2b9af-d35e-4f6b-a027-ba8de21fbbbc",
 CellID->1540419519],
Cell[29696015, 675398, 180, 4, 27, "Input",ExpressionUUID->"2b0fe639-da89-4e49-ac04-6992e97ee071",
 CellID->1915491248]
}, Open  ]],
Cell[CellGroupData[{
Cell[29696232, 675407, 273, 5, 24, "Text",ExpressionUUID->"b459a90d-aaf6-4d71-bde4-9e81b35da2e1",
 CellID->734052959],
Cell[CellGroupData[{
Cell[29696530, 675416, 367, 10, 27, "Input",ExpressionUUID->"d8ba2e27-4f24-4b7f-b131-0e90c2b75821",
 CellID->2082013545],
Cell[29696900, 675428, 494, 12, 26, "Output",ExpressionUUID->"8b2dee4c-541a-453e-b0f8-5064fd079c35",
 CellID->603516766]
}, Open  ]],
Cell[CellGroupData[{
Cell[29697431, 675445, 414, 12, 27, "Input",ExpressionUUID->"dfd56df5-54bd-4848-9a6a-4c004b0b24ff",
 CellID->1274343673],
Cell[29697848, 675459, 11779, 309, 32, "Output",ExpressionUUID->"43b5caeb-1ee6-4804-a3f6-3f4af1c7fe67",
 CellID->741021381]
}, Open  ]],
Cell[CellGroupData[{
Cell[29709664, 675773, 307, 8, 27, "Input",ExpressionUUID->"56298b9f-d091-4841-bd4e-132adfa2c029",
 CellID->173212421],
Cell[29709974, 675783, 18010, 456, 158, "Output",ExpressionUUID->"a6240618-adf7-43c2-9c28-4900298790f7",
 CellID->1088681514]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[29728045, 676246, 235, 3, 39, "Subsection",ExpressionUUID->"3fdef841-975b-4a9d-b578-b710322b4adb",
 CellID->1583260063],
Cell[29728283, 676251, 325, 6, 41, "DefinitionBox",ExpressionUUID->"e357dd9f-ef1c-4a0e-8dfb-4a1f7e8f9089",
 CellID->382008864],
Cell[29728611, 676259, 163, 2, 23, "Text",ExpressionUUID->"caca1ce9-89a7-4b4d-a9ac-9c30eb072b26",
 CellID->1594356365],
Cell[CellGroupData[{
Cell[29728799, 676265, 1277, 36, 78, "Text",ExpressionUUID->"20c4f55d-460e-45be-91af-997b7e52697c",
 CellID->46479289],
Cell[CellGroupData[{
Cell[29730101, 676305, 386, 10, 27, "Input",ExpressionUUID->"49e35c66-8780-47c9-868f-02ac42a4b7e0",
 CellID->1640846362],
Cell[29730490, 676317, 518, 13, 26, "Output",ExpressionUUID->"d5d03b7e-1381-4ddd-97fb-e465969ea1b1",
 CellID->665679041]
}, Open  ]],
Cell[CellGroupData[{
Cell[29731045, 676335, 467, 14, 27, "Input",ExpressionUUID->"5f571067-a16d-42bf-b0d8-db15231eda4e",
 CellID->301165391],
Cell[29731515, 676351, 40257, 1036, 233, "Output",ExpressionUUID->"ab3dcc0d-a2c6-4662-b4fe-d66fd176de7c",
 CellID->940372741]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[29771821, 677393, 120, 1, 23, "Text",ExpressionUUID->"1458d9cc-a4c3-40b7-b878-e7994b0aaa37",
 CellID->853435396],
Cell[29771944, 677396, 178, 4, 27, "Input",ExpressionUUID->"0fedbdf4-f638-47f3-96d1-4a70d15ee8f4",
 CellID->645422259]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[29772171, 677406, 247, 3, 39, "Subsection",ExpressionUUID->"6fd6bc1d-d135-46c8-9fd9-3d5ce2aa6fcb",
 CellID->672006016],
Cell[CellGroupData[{
Cell[29772443, 677413, 433, 9, 41, "Text",ExpressionUUID->"2820ac29-660f-4765-9a93-974137f76716",
 CellID->788611339],
Cell[CellGroupData[{
Cell[29772901, 677426, 346, 10, 27, "Input",ExpressionUUID->"ef0bfd45-0d0f-40e8-bef0-15e2ebade0e8",
 CellID->2069188936],
Cell[29773250, 677438, 1413, 39, 32, "Output",ExpressionUUID->"8df95361-3e92-4661-9d03-0c244dfa7db6",
 CellID->1500896426]
}, Open  ]],
Cell[CellGroupData[{
Cell[29774700, 677482, 167, 3, 27, "Input",ExpressionUUID->"15fb2b1d-9a65-44ef-9ae6-b477981bb63b",
 CellID->1915021919],
Cell[29774870, 677487, 3309, 87, 28, "Print",ExpressionUUID->"5769b9a1-2702-4a56-b8f3-873a38155052",
 CellID->1223314120],
Cell[29778182, 677576, 2660, 75, 32, "Output",ExpressionUUID->"560c1f93-f269-4217-81c4-63193e22f9aa",
 CellID->231978462]
}, Open  ]],
Cell[CellGroupData[{
Cell[29780879, 677656, 488, 15, 27, "Input",ExpressionUUID->"daac4602-9b13-47a3-92bc-9979edfd4d5f",
 CellID->250280249],
Cell[29781370, 677673, 1624, 45, 32, "Output",ExpressionUUID->"4dd977b0-227e-400f-b38f-bb765ae01c09",
 CellID->1179191529]
}, Open  ]],
Cell[CellGroupData[{
Cell[29783031, 677723, 166, 3, 27, "Input",ExpressionUUID->"8aa3d5c1-d409-47c8-9709-5be22fc2ff90",
 CellID->793239431],
Cell[29783200, 677728, 3873, 101, 28, "Print",ExpressionUUID->"aefaaf7a-826a-44c8-896d-124b88f14e0d",
 CellID->1692933387],
Cell[29787076, 677831, 3090, 87, 32, "Output",ExpressionUUID->"45cc6bbe-47f2-4239-abbe-626854bdc8cc",
 CellID->1120596206]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[29790215, 677924, 203, 5, 24, "Text",ExpressionUUID->"3b4b6709-01b0-43f4-98be-bf8d198f7443",
 CellID->1800793518],
Cell[CellGroupData[{
Cell[29790443, 677933, 198, 4, 27, "Input",ExpressionUUID->"42204b37-2f7c-4360-8cb9-350297425c11",
 CellID->1449619855],
Cell[29790644, 677939, 1521, 40, 44, "Output",ExpressionUUID->"2fb99f5f-1004-4823-9b8b-6becae5afff7",
 CellID->428850052]
}, Open  ]],
Cell[CellGroupData[{
Cell[29792202, 677984, 170, 3, 27, "Input",ExpressionUUID->"f6431eb0-869a-40df-b6f2-3d61c6df8ca1",
 CellID->805558412],
Cell[CellGroupData[{
Cell[29792397, 677991, 1184, 27, 22, "Print",ExpressionUUID->"f6c4b5d4-d13d-4ebd-83ac-34254ef66040",
 CellID->727704632],
Cell[29793584, 678020, 987, 21, 22, "Print",ExpressionUUID->"23f3f05e-7454-4bfe-ba86-3e3faeaeea67",
 CellID->612380359]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[29794644, 678049, 230, 3, 39, "Subsection",ExpressionUUID->"f240955e-7d0e-482c-aec9-aeb322a0d79b",
 CellID->958162795],
Cell[29794877, 678054, 252, 5, 41, "DefinitionBox",ExpressionUUID->"e4d775ac-cfa8-4251-9c57-bfca6dca0753",
 CellID->2009710359],
Cell[29795132, 678061, 168, 2, 23, "Text",ExpressionUUID->"754f4847-1dd6-4ee2-8c03-a540510b3728",
 CellID->1579203697],
Cell[CellGroupData[{
Cell[29795325, 678067, 377, 7, 41, "Text",ExpressionUUID->"01d395e1-be21-4c2a-8b6e-6c73b2a49392",
 CellID->926142206],
Cell[CellGroupData[{
Cell[29795727, 678078, 153, 3, 27, "Input",ExpressionUUID->"3036805d-59cb-4d54-916a-ac172c7254d5",
 CellID->710582786],
Cell[29795883, 678083, 415, 6, 67, "Print",ExpressionUUID->"24c2fc93-e784-4bb3-aaff-e4997b69698a",
 CellTags->"Info593730895309-7554244",
 CellID->1005248081]
}, Open  ]],
Cell[29796313, 678092, 179, 4, 27, "Input",ExpressionUUID->"641bdb87-61c8-47cc-8dfc-fcd75d47e46e",
 CellID->2068131625],
Cell[CellGroupData[{
Cell[29796517, 678100, 425, 11, 27, "Input",ExpressionUUID->"0527cb35-1e85-4901-a6e9-bdc853eb7ac2",
 CellID->2065394680],
Cell[29796945, 678113, 3084, 78, 28, "Print",ExpressionUUID->"c54f5c9f-5ad4-4216-922c-e2552ea2f6fe",
 CellID->950646418],
Cell[29800032, 678193, 2523, 70, 32, "Output",ExpressionUUID->"d536c68d-188d-44d3-81a6-188680bef451",
 CellID->236013646]
}, Open  ]],
Cell[CellGroupData[{
Cell[29802592, 678268, 632, 18, 27, "Input",ExpressionUUID->"160d7db7-c30d-433b-a0e5-c08ac8061ab4",
 CellID->1377197476],
Cell[CellGroupData[{
Cell[29803249, 678290, 1870, 45, 28, "Print",ExpressionUUID->"91f6cb09-d043-491e-b73a-5cedc3422967",
 CellID->279052313],
Cell[29805122, 678337, 1875, 45, 28, "Print",ExpressionUUID->"8a26fe8d-d341-4f39-b170-2e7fc33e3138",
 CellID->2075423190],
Cell[29807000, 678384, 1874, 45, 28, "Print",ExpressionUUID->"d50a1931-94d1-4346-95f8-6b4947ccf8a2",
 CellID->60676109],
Cell[29808877, 678431, 1871, 45, 28, "Print",ExpressionUUID->"64b34971-16ef-4ea4-9f49-976b7893718f",
 CellID->1876115285],
Cell[29810751, 678478, 1874, 45, 28, "Print",ExpressionUUID->"f45abf14-47c3-4e38-9d80-90bf652a4f09",
 CellID->367206094],
Cell[29812628, 678525, 1878, 45, 28, "Print",ExpressionUUID->"35b61a2d-e620-45c7-9fe7-72b6ea122930",
 CellID->1912304027],
Cell[29814509, 678572, 1872, 45, 28, "Print",ExpressionUUID->"efefcc0d-69c7-458f-af29-669487c82e27",
 CellID->853520787],
Cell[29816384, 678619, 1875, 45, 28, "Print",ExpressionUUID->"201f056d-5a77-4fcf-a8ad-ded5b0712338",
 CellID->2080588796],
Cell[29818262, 678666, 1872, 45, 28, "Print",ExpressionUUID->"5b999428-062f-41b3-b810-83109e3e8af6",
 CellID->979670271]
}, Open  ]],
Cell[29820149, 678714, 12102, 317, 53, "Output",ExpressionUUID->"e09b7de6-d6c9-488e-80ce-e225a4e0aa2d",
 CellID->561123148]
}, Open  ]],
Cell[CellGroupData[{
Cell[29832288, 679036, 283, 6, 27, "Input",ExpressionUUID->"18f09ca9-5167-4e95-86f8-8e94d6f1ccb9",
 CellID->343148269],
Cell[29832574, 679044, 12099, 317, 53, "Output",ExpressionUUID->"f62a4166-b33c-4b7d-9a87-0d163e8c4859",
 CellID->1196054459]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[29844722, 679367, 193, 4, 23, "Text",ExpressionUUID->"ba410e32-c41c-4497-8cf4-8eb66fa8b999",
 CellID->664466273],
Cell[CellGroupData[{
Cell[29844940, 679375, 217, 5, 27, "Input",ExpressionUUID->"39348cbe-d002-4e7c-9935-680f20679912",
 CellID->769685911],
Cell[29845160, 679382, 12109, 317, 53, "Output",ExpressionUUID->"7fc3a84b-427c-403c-b5a7-210b2250e832",
 CellID->1640356493]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[29857330, 679706, 258, 3, 39, "Subsection",ExpressionUUID->"f64727aa-5c80-4f73-b728-49ecb2185f2b",
 CellID->639730885],
Cell[29857591, 679711, 220, 4, 41, "DefinitionBox",ExpressionUUID->"fcd44251-b238-4cd0-b8d0-1b3fd17d75cd",
 CellID->1653847313],
Cell[29857814, 679717, 158, 2, 23, "Text",ExpressionUUID->"88b92c65-80e2-4ca1-a848-b550cbc336a1",
 CellID->1330946470],
Cell[CellGroupData[{
Cell[29857997, 679723, 513, 8, 57, "Text",ExpressionUUID->"3e96739b-3d0e-43df-8d83-d7b3e871363f",
 CellID->1027654138],
Cell[CellGroupData[{
Cell[29858535, 679735, 156, 3, 27, "Input",ExpressionUUID->"bc76d497-72bf-48a7-ba4a-f0f921098953",
 CellID->1719608192],
Cell[29858694, 679740, 1311, 18, 243, "Print",ExpressionUUID->"c16536ba-9f45-451f-924f-29b7afe430d3",
 CellTags->"Info653730895316-7554244",
 CellID->876129877]
}, Open  ]],
Cell[CellGroupData[{
Cell[29860042, 679763, 571, 14, 27, "Input",ExpressionUUID->"db60b5f2-3336-443a-a1c8-e86ab574260b",
 CellID->304471059],
Cell[CellGroupData[{
Cell[29860638, 679781, 854, 18, 22, "Print",ExpressionUUID->"3a304b51-f050-4a70-8fd8-f3ee365822a6",
 CellID->1966958907],
Cell[29861495, 679801, 913, 20, 22, "Print",ExpressionUUID->"5282bc52-f937-479b-be8f-c9719c05573a",
 CellID->315088031],
Cell[29862411, 679823, 772, 16, 22, "Print",ExpressionUUID->"d9d3a2c2-4f04-4a0a-94eb-4f3387ee7228",
 CellID->2137347162],
Cell[29863186, 679841, 894, 19, 22, "Print",ExpressionUUID->"994953e3-3314-460e-87a2-c682731a243f",
 CellID->1434788365],
Cell[29864083, 679862, 909, 20, 22, "Print",ExpressionUUID->"79dc15bb-ff3f-433a-a2ff-4ece90a0da0a",
 CellID->394178392],
Cell[29864995, 679884, 935, 20, 22, "Print",ExpressionUUID->"095a1c65-d3c5-482e-8f00-cf49ffe17baf",
 CellID->2116666996],
Cell[29865933, 679906, 856, 18, 22, "Print",ExpressionUUID->"79ccfb33-a200-43f7-9b82-ea91a76e3079",
 CellID->1755589217],
Cell[29866792, 679926, 359, 5, 22, "Print",ExpressionUUID->"928e9872-373d-4d1c-8b85-e894cda2f769",
 CellID->1910143986],
Cell[29867154, 679933, 738, 14, 22, "Print",ExpressionUUID->"28907ee7-bc12-40a0-8269-a6c224811b3c",
 CellID->1056933699],
Cell[29867895, 679949, 365, 5, 22, "Print",ExpressionUUID->"bd236b7a-a78c-474c-a251-031d6e41981d",
 CellID->2047434756],
Cell[29868263, 679956, 870, 19, 22, "Print",ExpressionUUID->"35f261dd-d048-4fad-8520-28bde6496ad1",
 CellID->2105468211],
Cell[29869136, 679977, 943, 20, 22, "Print",ExpressionUUID->"2be8e081-def4-4f70-849a-bf2abd7d5463",
 CellID->368687011],
Cell[29870082, 679999, 862, 18, 22, "Print",ExpressionUUID->"999b64e0-93b4-480f-b782-1a0eb65d17a1",
 CellID->1032272282],
Cell[29870947, 680019, 752, 14, 22, "Print",ExpressionUUID->"0c5d9a82-ad40-45f4-aca9-39973954a4fb",
 CellID->2027343108],
Cell[29871702, 680035, 496, 9, 22, "Print",ExpressionUUID->"1f00eeed-25e8-451d-a84c-ed11fcfa1d3e",
 CellID->1580072854],
Cell[29872201, 680046, 489, 9, 22, "Print",ExpressionUUID->"ef978c55-14d4-4200-ba37-28f0422701e6",
 CellID->501678376],
Cell[29872693, 680057, 499, 9, 22, "Print",ExpressionUUID->"6573b891-a9ce-4973-83d8-8fb6167332a0",
 CellID->687667310],
Cell[29873195, 680068, 764, 14, 22, "Print",ExpressionUUID->"153ac3d1-1593-4e88-864a-07faab65692b",
 CellID->2014974663]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[29874008, 680088, 133, 2, 27, "Input",ExpressionUUID->"d2985e35-a4c1-4b9f-ad31-2e070cd816ac",
 CellID->104258482],
Cell[29874144, 680092, 293, 5, 26, "Output",ExpressionUUID->"07e9964a-c2d0-4453-a2c1-52d63a561ba3",
 CellID->2009531498]
}, Open  ]],
Cell[29874452, 680100, 179, 4, 27, "Input",ExpressionUUID->"f1ef9782-5cb4-43ec-b915-0423e6eca365",
 CellID->252209308],
Cell[CellGroupData[{
Cell[29874656, 680108, 437, 12, 27, "Input",ExpressionUUID->"04f66819-547b-4331-a7e4-37b2bb154e63",
 CellID->1896604025],
Cell[29875096, 680122, 11843, 309, 32, "Output",ExpressionUUID->"af4ef070-0aba-4c3c-adfe-43b697833815",
 CellID->720570517]
}, Open  ]],
Cell[CellGroupData[{
Cell[29886976, 680436, 169, 3, 27, "Input",ExpressionUUID->"c061e3bb-a5c9-44f5-99b0-22802acbf4fe",
 CellID->927554819],
Cell[29887148, 680441, 15505, 403, 32, "Output",ExpressionUUID->"3f60ad0b-2698-43b4-84c4-b3c100088993",
 CellID->968393786]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[29902702, 680850, 124, 1, 23, "Text",ExpressionUUID->"1fc1843f-c177-4023-a458-3060b281f830",
 CellID->585822608],
Cell[CellGroupData[{
Cell[29902851, 680855, 176, 3, 27, "Input",ExpressionUUID->"629e18de-32db-4849-a24c-635e32820dc7",
 CellID->1855534808],
Cell[29903030, 680860, 896, 19, 22, "Print",ExpressionUUID->"789c037c-9e04-4b6d-841e-06c44bff9cb7",
 CellID->15800689]
}, Open  ]],
Cell[CellGroupData[{
Cell[29903963, 680884, 291, 7, 27, "Input",ExpressionUUID->"5008830a-33fe-4733-ba75-d551df0a27f8",
 CellID->1002820424],
Cell[29904257, 680893, 11841, 309, 32, "Output",ExpressionUUID->"a6ac495f-0711-494a-8b88-b29dfe3a4cfc",
 CellID->1421691442]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[29916147, 681208, 124, 1, 23, "Text",ExpressionUUID->"573aa27c-ca22-41f5-a86a-56d70de6755d",
 CellID->1025964916],
Cell[CellGroupData[{
Cell[29916296, 681213, 160, 3, 27, "Input",ExpressionUUID->"72060461-9b16-406c-9454-22f935f114b9",
 CellID->1246141449],
Cell[29916459, 681218, 410, 6, 67, "Print",ExpressionUUID->"b001b759-6b67-43a6-85bf-2c61385ba6fd",
 CellTags->"Info733730895317-7554244",
 CellID->1465009172]
}, Open  ]],
Cell[CellGroupData[{
Cell[29916906, 681229, 256, 6, 27, "Input",ExpressionUUID->"91600d3a-1b2f-4e33-b40b-94efd0a65389",
 CellID->1964905018],
Cell[29917165, 681237, 318, 6, 26, "Output",ExpressionUUID->"65fe807b-11b8-4244-b358-0c0d15456cb5",
 CellID->1509140949]
}, Open  ]],
Cell[CellGroupData[{
Cell[29917520, 681248, 257, 6, 27, "Input",ExpressionUUID->"e8abf8ae-86ab-483f-ba45-3c5032f91549",
 CellID->2102661909],
Cell[29917780, 681256, 11998, 319, 32, "Output",ExpressionUUID->"47d43d34-e78a-469b-91e5-96b415dd83ab",
 CellID->202575126]
}, Open  ]],
Cell[CellGroupData[{
Cell[29929815, 681580, 170, 3, 27, "Input",ExpressionUUID->"dc93bbe8-137e-40cd-8ece-1238eb42d99d",
 CellID->1552088429],
Cell[29929988, 681585, 30913, 813, 56, "Output",ExpressionUUID->"64a63038-bda9-4bbe-a695-3049153ac511",
 CellID->918562187]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[29960962, 682405, 307, 4, 39, "Subsection",ExpressionUUID->"5601da75-08fc-42b3-a99b-511271499d04",
 CellID->2047113955],
Cell[29961272, 682411, 253, 5, 41, "DefinitionBox",ExpressionUUID->"a12d9678-6dc2-4dea-9417-401f0f22f541",
 CellID->1008733098],
Cell[29961528, 682418, 169, 2, 23, "Text",ExpressionUUID->"92c53221-0337-4636-bf68-e9fc0aff8c3b",
 CellID->1541268668],
Cell[29961700, 682422, 1011, 26, 61, "Text",ExpressionUUID->"d284178e-ec27-43ee-9927-60a5daeda301",
 CellID->1154752962],
Cell[CellGroupData[{
Cell[29962736, 682452, 108, 1, 23, "Text",ExpressionUUID->"2316770c-9622-4ffb-82ba-6050909bee86",
 CellID->1123314395],
Cell[CellGroupData[{
Cell[29962869, 682457, 382, 11, 27, "Input",ExpressionUUID->"772359a1-40e2-40ba-a10c-34e2b4052f7b",
 CellID->1016295988],
Cell[29963254, 682470, 4247, 111, 28, "Print",ExpressionUUID->"188b3c53-3278-4d22-ad2a-255f816fc4c0",
 CellID->567600032],
Cell[29967504, 682583, 26563, 690, 32, "Output",ExpressionUUID->"843e134d-05dc-4629-bc3a-413fe76ba799",
 CellID->531377850]
}, Open  ]],
Cell[CellGroupData[{
Cell[29994104, 683278, 164, 3, 27, "Input",ExpressionUUID->"45b67b8d-e04c-449e-b62f-877f93b32d49",
 CellID->984105045],
Cell[29994271, 683283, 534, 15, 26, "Output",ExpressionUUID->"5cbf25c4-c62b-4433-937a-6c31bfff933c",
 CellID->357594777]
}, Open  ]],
Cell[CellGroupData[{
Cell[29994842, 683303, 171, 3, 27, "Input",ExpressionUUID->"e5e7dde7-f86f-42bf-b8a8-cb897ed938e3",
 CellID->1975322205],
Cell[CellGroupData[{
Cell[29995038, 683310, 764, 15, 22, "Print",ExpressionUUID->"68a92604-4d41-4134-8def-f350ffed9e22",
 CellID->977772303],
Cell[29995805, 683327, 789, 17, 22, "Print",ExpressionUUID->"d801f1a2-64bc-4564-8360-2488161c5d99",
 CellID->874229677]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[29996655, 683351, 435, 10, 42, "Text",ExpressionUUID->"8f4b7272-0844-40fb-aa53-8e9226aa2ab0",
 CellID->317888728],
Cell[CellGroupData[{
Cell[29997115, 683365, 342, 9, 27, "Input",ExpressionUUID->"3b81f1f8-8772-4354-8b73-87af195e5534",
 CellID->1964021407],
Cell[29997460, 683376, 734, 14, 22, "Print",ExpressionUUID->"4703eae4-62fd-4842-952d-424749bfd346",
 CellID->714658521]
}, Open  ]],
Cell[CellGroupData[{
Cell[29998231, 683395, 243, 6, 27, "Input",ExpressionUUID->"0d7f1732-1399-4c5a-b846-a8fc2c085de6",
 CellID->1594741286],
Cell[29998477, 683403, 286, 6, 26, "Output",ExpressionUUID->"ec176970-451e-472c-a294-98774fc30040",
 CellID->1989078215]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[29998812, 683415, 133, 1, 23, "Text",ExpressionUUID->"bc18e416-c471-4c1f-b7d6-c97268ffbeb2",
 CellID->904151787],
Cell[29998948, 683418, 398, 10, 85, "Input",ExpressionUUID->"96302cfa-192d-4119-91b4-da1c7a9ad25f",
 CellID->484797937],
Cell[CellGroupData[{
Cell[29999371, 683432, 573, 18, 27, "Input",ExpressionUUID->"3949617e-ef69-43b2-a456-71b9a8f3eca5",
 CellID->445970441],
Cell[CellGroupData[{
Cell[29999969, 683454, 1783, 44, 28, "Print",ExpressionUUID->"2b74f52f-d1a9-4679-bae2-e0e1c3416105",
 CellID->798952431],
Cell[30001755, 683500, 3017, 77, 28, "Print",ExpressionUUID->"f38eaf3f-218f-4a58-9f9e-858eb9190d80",
 CellID->1925746773],
Cell[30004775, 683579, 3014, 77, 28, "Print",ExpressionUUID->"3996d130-49c3-462d-b19a-3c33ed527126",
 CellID->1299969874],
Cell[30007792, 683658, 3076, 79, 28, "Print",ExpressionUUID->"ef109891-4304-41e4-b905-adb550fed93e",
 CellID->1047127517],
Cell[30010871, 683739, 1783, 44, 28, "Print",ExpressionUUID->"5d7997fc-3b19-402c-ba51-dc894d1be8a4",
 CellID->443691054],
Cell[30012657, 683785, 3015, 77, 28, "Print",ExpressionUUID->"247fa0ec-b4bf-47b6-ab28-4d601fa0524f",
 CellID->1052322194],
Cell[30015675, 683864, 3078, 79, 28, "Print",ExpressionUUID->"2a807cb8-2b07-454a-b2e7-2c9625c0d93a",
 CellID->1851281513],
Cell[30018756, 683945, 3076, 79, 28, "Print",ExpressionUUID->"2a65be83-b8ae-4006-a2b0-0399c3244616",
 CellID->1745311255],
Cell[30021835, 684026, 1783, 44, 28, "Print",ExpressionUUID->"09c52b0a-f67d-4406-9576-de1512b89df1",
 CellID->967128412],
Cell[30023621, 684072, 7086, 181, 28, "Print",ExpressionUUID->"661dc0d9-264a-4957-93ed-c3859012bf33",
 CellID->223496341],
Cell[30030710, 684255, 3079, 79, 28, "Print",ExpressionUUID->"2bc5502c-49dd-45a2-a83a-de7f5f987040",
 CellID->354104584],
Cell[30033792, 684336, 7084, 181, 28, "Print",ExpressionUUID->"724b7161-e19d-47e5-8c0d-b3a25b858ac3",
 CellID->758557451],
Cell[30040879, 684519, 3079, 79, 28, "Print",ExpressionUUID->"a781fb83-0c57-49a8-9586-943d46febda3",
 CellID->1612150365],
Cell[30043961, 684600, 7086, 181, 28, "Print",ExpressionUUID->"9e5118c3-8082-4615-8a1c-09eb394a073c",
 CellID->610010356],
Cell[30051050, 684783, 7086, 181, 28, "Print",ExpressionUUID->"6cd71785-287a-40ba-a8be-84c4d5d0d521",
 CellID->357122514],
Cell[30058139, 684966, 7166, 184, 28, "Print",ExpressionUUID->"f29d6dc0-2c08-494d-98ee-b9be7e9fd7fe",
 CellID->2088783080],
Cell[30065308, 685152, 3082, 79, 28, "Print",ExpressionUUID->"29dc73f1-10fb-463c-95a5-1a987dfba9a0",
 CellID->1262048051],
Cell[30068393, 685233, 7086, 181, 28, "Print",ExpressionUUID->"eeaf2f80-3e9e-496f-b533-d0475ebcaf2b",
 CellID->545611738],
Cell[30075482, 685416, 7086, 181, 28, "Print",ExpressionUUID->"bbf2b4d7-cda1-4d8c-bd2f-5d624f6c81cb",
 CellID->341307876],
Cell[30082571, 685599, 7165, 184, 28, "Print",ExpressionUUID->"82de3e5e-7675-4306-882c-532bf0d9dbfb",
 CellID->432650795],
Cell[30089739, 685785, 7165, 184, 28, "Print",ExpressionUUID->"fb90fe3d-ccb1-439c-acee-9edd4e97d627",
 CellID->182282823],
Cell[30096907, 685971, 3079, 79, 28, "Print",ExpressionUUID->"1d6173d6-76a6-4a0b-86b1-c9e0023e6eb3",
 CellID->1315725539],
Cell[30099989, 686052, 3081, 79, 28, "Print",ExpressionUUID->"61eef69e-ff16-450b-9095-ad1a4f12f54d",
 CellID->760020814],
Cell[30103073, 686133, 3081, 79, 28, "Print",ExpressionUUID->"518a9d0c-6058-4c99-99ea-7e797441d229",
 CellID->437088542],
Cell[30106157, 686214, 4583, 119, 28, "Print",ExpressionUUID->"c9319005-26c8-40f8-ad52-10e96b833eb1",
 CellID->1753052654],
Cell[30110743, 686335, 1788, 44, 28, "Print",ExpressionUUID->"74fc2a1f-70f0-4eba-85b9-8000cc9d5db7",
 CellID->80318938],
Cell[30112534, 686381, 9744, 249, 28, "Print",ExpressionUUID->"aa51c2e4-5ff3-4856-8964-717f06b59607",
 CellID->1300660650],
Cell[30122281, 686632, 9744, 249, 28, "Print",ExpressionUUID->"63609b57-b9de-414b-b4cd-618396796203",
 CellID->967501737],
Cell[30132028, 686883, 3084, 79, 28, "Print",ExpressionUUID->"cae46f48-5231-4dd5-88e5-a5f1e5c66e81",
 CellID->55518675],
Cell[30135115, 686964, 19055, 489, 28, "Print",ExpressionUUID->"5aaf199a-2b41-4c81-b976-6bc052bd1221",
 CellID->1094467927],
Cell[30154173, 687455, 1788, 44, 28, "Print",ExpressionUUID->"e503dd77-6175-4100-b395-7ce11f58a412",
 CellID->1525012534],
Cell[30155964, 687501, 9746, 249, 28, "Print",ExpressionUUID->"d5e33779-cc3f-41d4-b963-b7a30cafb337",
 CellID->668331993],
Cell[30165713, 687752, 3084, 79, 28, "Print",ExpressionUUID->"7d994d69-d40c-4f86-ad14-79856c4fa169",
 CellID->1316332541],
Cell[30168800, 687833, 19055, 489, 28, "Print",ExpressionUUID->"35009c62-b6a5-4c8f-b132-32717a7a57bc",
 CellID->224485193],
Cell[30187858, 688324, 3084, 79, 28, "Print",ExpressionUUID->"3a7c0e3c-db60-47f7-bbfc-19b0674dab02",
 CellID->1941354619],
Cell[30190945, 688405, 19135, 492, 28, "Print",ExpressionUUID->"bac42445-8436-4e49-85b2-8a8a3dcf3d1b",
 CellID->1628032662],
Cell[30210083, 688899, 1787, 44, 28, "Print",ExpressionUUID->"ddd78133-278d-4224-a77d-eda45372d177",
 CellID->1255803807],
Cell[30211873, 688945, 4590, 119, 28, "Print",ExpressionUUID->"7c686282-8183-4d17-abc1-db618bca0829",
 CellID->825301533]
}, Open  ]],
Cell[30216478, 689067, 114240, 2961, 128, "Output",ExpressionUUID->"c0084cef-ac58-47a8-9d81-7adb8abdea5e",
 CellID->245989158]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[30330767, 692034, 2354, 69, 57, "Text",ExpressionUUID->"bd6a602b-98c4-4589-9ec9-59db2724c21c",
 CellID->210155605],
Cell[CellGroupData[{
Cell[30333146, 692107, 165, 3, 27, "Input",ExpressionUUID->"8056cdde-7fd6-464d-b695-62c140cb2980",
 CellID->1622137567],
Cell[30333314, 692112, 133219, 3460, 152, "Output",ExpressionUUID->"9cf0b7e9-6673-40d2-8e7a-03be4ac56308",
 CellID->285575129]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[30466582, 695578, 185, 4, 23, "Text",ExpressionUUID->"2246b99f-b656-43a8-8919-92d90bb0ca29",
 CellID->1346723794],
Cell[CellGroupData[{
Cell[30466792, 695586, 161, 3, 27, "Input",ExpressionUUID->"b949a72c-4b42-4966-b344-4b8c676c31f7",
 CellID->2056650025],
Cell[30466956, 695591, 822, 11, 133, "Print",ExpressionUUID->"c78b4dc7-d8d5-444f-8eb7-d91f40dd7a9d",
 CellTags->"Info883730895335-7554244",
 CellID->696624812]
}, Open  ]],
Cell[CellGroupData[{
Cell[30467815, 695607, 196, 4, 27, "Input",ExpressionUUID->"067f70b0-8b82-4dbf-bf73-384d01d22817",
 CellID->586451060],
Cell[30468014, 695613, 252846, 6592, 248, "Output",ExpressionUUID->"ecda3e8c-2d5a-4436-a78a-2867ef1e557b",
 CellID->17621519]
}, Open  ]],
Cell[CellGroupData[{
Cell[30720897, 702210, 252, 6, 27, "Input",ExpressionUUID->"cc83a7c6-99b7-4648-8d64-23ad9581eafc",
 CellID->615501615],
Cell[30721152, 702218, 298, 6, 26, "Output",ExpressionUUID->"db260a15-2e8c-488b-9c70-42b5c9f08189",
 CellID->1121161639]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[30721499, 702230, 114, 1, 23, "Text",ExpressionUUID->"d6d2aa8a-64a0-4d53-91f9-3400715db483",
 CellID->76550648],
Cell[CellGroupData[{
Cell[30721638, 702235, 363, 10, 27, "Input",ExpressionUUID->"9a8b3092-f620-4daf-9d53-262c60e460ff",
 CellID->244036383],
Cell[CellGroupData[{
Cell[30722026, 702249, 2748, 73, 28, "Print",ExpressionUUID->"cd70a4ad-b925-4302-9629-3f7de89bee79",
 CellID->1577509860],
Cell[30724777, 702324, 2747, 73, 28, "Print",ExpressionUUID->"3dfe7013-5084-4d51-836a-e17e235ac1f3",
 CellID->375111158],
Cell[30727527, 702399, 2748, 73, 28, "Print",ExpressionUUID->"1c5c90ea-a2ec-4af1-ba24-51d0a0ecaa4b",
 CellID->1628728561],
Cell[30730278, 702474, 9209, 241, 28, "Print",ExpressionUUID->"356f188d-bba5-4070-9e39-378dcbb0f209",
 CellID->66876418],
Cell[30739490, 702717, 9210, 241, 28, "Print",ExpressionUUID->"7b5563a3-3256-4d09-a3aa-ea3a70ffaa16",
 CellID->858009945],
Cell[30748703, 702960, 9210, 241, 28, "Print",ExpressionUUID->"de4e363c-9266-4b1a-ac3c-97bbf3cee9bc",
 CellID->220456010],
Cell[30757916, 703203, 4261, 111, 28, "Print",ExpressionUUID->"d3aa05a0-8368-4ffe-b9e9-cab317408d22",
 CellID->794868995]
}, Open  ]],
Cell[30762192, 703317, 26703, 699, 32, "Output",ExpressionUUID->"c8a97213-c1bc-4fcc-bfc6-283593b5682d",
 CellID->2014914625]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[30788956, 704023, 311, 4, 39, "Subsection",ExpressionUUID->"cccf106d-4bf8-4bca-a08e-ed5bdb322dcb",
 CellID->1858146570],
Cell[30789270, 704029, 242, 5, 41, "DefinitionBox",ExpressionUUID->"f55e0c65-2e83-4c83-aa1b-666b2bfa2854",
 CellID->1162207316],
Cell[30789515, 704036, 172, 2, 23, "Text",ExpressionUUID->"efe7940c-cd3b-43fb-aff6-a6326f0f4bd4",
 CellID->1854590547],
Cell[CellGroupData[{
Cell[30789712, 704042, 511, 13, 42, "Text",ExpressionUUID->"f8c2c11c-8a93-49ef-8504-bd7e25d8e28f",
 CellID->1525670871],
Cell[CellGroupData[{
Cell[30790248, 704059, 150, 3, 27, "Input",ExpressionUUID->"87fd8e9c-b191-469d-ac01-67adc708b70f",
 CellID->995583717],
Cell[30790401, 704064, 551, 8, 89, "Print",ExpressionUUID->"57169d80-85a2-4839-8e86-b95b76f73bed",
 CellTags->"Info923730895347-7554244",
 CellID->821685044]
}, Open  ]],
Cell[CellGroupData[{
Cell[30790989, 704077, 269, 7, 27, "Input",ExpressionUUID->"d6ac9cb7-8eaf-4d4c-85ff-7ee2b1b11987",
 CellID->97779173],
Cell[30791261, 704086, 15659, 412, 32, "Output",ExpressionUUID->"f30665af-b36b-4f1b-9881-8a75ae1963da",
 CellID->1740684298]
}, Open  ]],
Cell[30806935, 704501, 179, 4, 27, "Input",ExpressionUUID->"17272b76-fb79-421c-b796-7e9625041952",
 CellID->306849645],
Cell[CellGroupData[{
Cell[30807139, 704509, 574, 18, 27, "Input",ExpressionUUID->"7dd535fa-b9ec-4bd4-a89a-714bdfb79604",
 CellID->1113747068],
Cell[CellGroupData[{
Cell[30807738, 704531, 4586, 119, 28, "Print",ExpressionUUID->"514f4e98-8b53-4303-933c-37a11f120432",
 CellID->511273405],
Cell[30812327, 704652, 4594, 119, 28, "Print",ExpressionUUID->"dca44d0d-b6de-428f-8f46-77c16f54d5be",
 CellID->890046188]
}, Open  ]],
Cell[30816936, 704774, 106001, 2754, 104, "Output",ExpressionUUID->"a6699ddd-d6b0-41e7-89ef-7f55c8ccf000",
 CellID->1782680361]
}, Open  ]],
Cell[CellGroupData[{
Cell[30922974, 707533, 512, 17, 27, "Input",ExpressionUUID->"201afdce-a6d8-48bc-bd23-d13ba4c40741",
 CellID->1433563111],
Cell[30923489, 707552, 2511, 71, 32, "Output",ExpressionUUID->"1826a711-6d02-4668-84a2-ce202379bd80",
 CellID->1964468030]
}, Open  ]],
Cell[CellGroupData[{
Cell[30926037, 707628, 168, 3, 27, "Input",ExpressionUUID->"3d67136d-aa76-4b71-b347-56754672769a",
 CellID->1684864799],
Cell[30926208, 707633, 20802, 590, 32, "Output",ExpressionUUID->"f24d3c0e-d620-4f8a-89b0-90d1a3941c85",
 CellID->1400591232]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[30947059, 708229, 191, 5, 41, "MathCaption",ExpressionUUID->"13c59503-32d8-4b8a-898d-be9911676ef0",
 CellID->2106123915],
Cell[CellGroupData[{
Cell[30947275, 708238, 160, 3, 27, "Input",ExpressionUUID->"90be1009-d2c0-4d45-846e-684b9e0c22f4",
 CellID->651685904],
Cell[30947438, 708243, 12559, 325, 32, "Output",ExpressionUUID->"66d45648-72aa-4912-80f3-529399e04e20",
 CellID->1168581457]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[30960046, 708574, 144, 1, 40, "MathCaption",ExpressionUUID->"3735e57a-635a-40df-a0cc-5ac128bda534",
 CellID->807968762],
Cell[CellGroupData[{
Cell[30960215, 708579, 160, 3, 27, "Input",ExpressionUUID->"6cc39a29-c4ea-4835-bd84-55f7659d3348",
 CellID->991882953],
Cell[30960378, 708584, 220, 4, 26, "Output",ExpressionUUID->"84b40fd2-71bd-4026-b45a-4dabeb2c0721",
 CellID->1610595984]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[30960659, 708595, 327, 4, 39, "Subsection",ExpressionUUID->"81303af2-2721-422a-8bc1-c362230067a0",
 CellID->941636722],
Cell[CellGroupData[{
Cell[30961011, 708603, 238, 4, 27, "Input",ExpressionUUID->"2d852c30-2fe0-4104-97d1-570a2ea4c50a",
 CellID->1413614805],
Cell[CellGroupData[{
Cell[30961274, 708611, 608, 12, 22, "Print",ExpressionUUID->"4ec1466e-af06-46ad-8a2e-cf9b8bd6b75a",
 CellID->894097473],
Cell[30961885, 708625, 639, 12, 22, "Print",ExpressionUUID->"afde95b3-19a4-4d39-8afc-bb8b96ca9947",
 CellID->1153150652],
Cell[30962527, 708639, 627, 12, 22, "Print",ExpressionUUID->"df914530-7842-4cb1-bc71-6bc42b4b7876",
 CellID->2110577636],
Cell[30963157, 708653, 620, 12, 22, "Print",ExpressionUUID->"6fec2091-e9e5-4514-90e1-c2d48d4e1a0f",
 CellID->343447106],
Cell[30963780, 708667, 615, 12, 22, "Print",ExpressionUUID->"f2e54da9-05bb-4da4-8bd6-ac612049b1fb",
 CellID->2026951053],
Cell[30964398, 708681, 609, 12, 22, "Print",ExpressionUUID->"6de2586e-353c-4ca9-b9ed-686219efdfe7",
 CellID->1433255410],
Cell[30965010, 708695, 601, 12, 22, "Print",ExpressionUUID->"7151d2f5-0ebd-4615-82a1-b6470b719733",
 CellID->399607471],
Cell[30965614, 708709, 622, 12, 22, "Print",ExpressionUUID->"22087298-e88e-49d3-8219-74f317cac93c",
 CellID->1021400928],
Cell[30966239, 708723, 602, 12, 22, "Print",ExpressionUUID->"f192623d-d32f-43ae-8a37-7f284f13db2f",
 CellID->1848606425],
Cell[30966844, 708737, 591, 12, 22, "Print",ExpressionUUID->"1e8e7517-234e-4741-943f-47f9eb8be6d3",
 CellID->1882263826],
Cell[30967438, 708751, 598, 12, 22, "Print",ExpressionUUID->"ccb0354b-baec-4a68-b7f4-506ed81147cd",
 CellID->626519625],
Cell[30968039, 708765, 622, 12, 22, "Print",ExpressionUUID->"099d829c-4650-4e63-97c6-aea2896d1b75",
 CellID->803016134],
Cell[30968664, 708779, 615, 12, 22, "Print",ExpressionUUID->"6e2ef28e-c182-438b-a2a5-5e4a85c88dd6",
 CellID->1504926390]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[30969328, 708797, 315, 7, 27, "Input",ExpressionUUID->"338b4ffa-ea39-4859-ac7a-e68a16addb55",
 CellID->654448890],
Cell[CellGroupData[{
Cell[30969668, 708808, 599, 13, 22, "Print",ExpressionUUID->"b771ac61-6a37-492c-95be-8245c3c8a434",
 CellID->235442364],
Cell[30970270, 708823, 598, 13, 22, "Print",ExpressionUUID->"6e8ec85b-7850-493d-8613-b3fdfa0f9bfa",
 CellID->1414933974],
Cell[30970871, 708838, 598, 13, 22, "Print",ExpressionUUID->"cc52f977-3e6e-4470-bd39-822d81ee65be",
 CellID->1708877787],
Cell[30971472, 708853, 597, 13, 22, "Print",ExpressionUUID->"870d2000-b5bc-43bf-a1e3-a2ccdde15962",
 CellID->662671131]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[30972118, 708872, 304, 7, 27, "Input",ExpressionUUID->"d75fa733-1d3c-427f-856f-fe5bca57dd76",
 CellID->538418245],
Cell[CellGroupData[{
Cell[30972447, 708883, 660, 13, 22, "Print",ExpressionUUID->"069c4f47-d436-41a9-93f0-1af4b64b117c",
 CellID->663505375],
Cell[30973110, 708898, 604, 12, 22, "Print",ExpressionUUID->"1c83e161-510a-4b4f-ab50-59e93011014f",
 CellID->1799742401],
Cell[30973717, 708912, 612, 12, 22, "Print",ExpressionUUID->"d10abd13-3f0f-406f-9393-351e894921bd",
 CellID->963365612],
Cell[30974332, 708926, 613, 12, 22, "Print",ExpressionUUID->"d3de7c50-4506-48b0-a745-f259e90a84f7",
 CellID->1406626151],
Cell[30974948, 708940, 605, 12, 22, "Print",ExpressionUUID->"50a875e1-047a-4187-9143-8e0100556ddd",
 CellID->1665733681],
Cell[30975556, 708954, 627, 12, 22, "Print",ExpressionUUID->"1881041a-b205-4893-905f-493cc5163967",
 CellID->1389154275],
Cell[30976186, 708968, 629, 12, 22, "Print",ExpressionUUID->"d139ea8f-89c4-4337-a581-65034f8341f6",
 CellID->73689741],
Cell[30976818, 708982, 573, 12, 22, "Print",ExpressionUUID->"9cb7030b-3910-4b1a-905a-24e9dafadda6",
 CellID->1552896315],
Cell[30977394, 708996, 673, 14, 22, "Print",ExpressionUUID->"1fa70996-23e2-4d2a-b87e-62175a6936b6",
 CellID->1159604214],
Cell[30978070, 709012, 613, 12, 22, "Print",ExpressionUUID->"ab566967-3753-4cce-a33f-28de5671b144",
 CellID->1053928173],
Cell[30978686, 709026, 624, 13, 22, "Print",ExpressionUUID->"35348090-bb3b-430d-bf45-6815d1b3f564",
 CellID->925683155],
Cell[30979313, 709041, 625, 13, 22, "Print",ExpressionUUID->"bfc13289-bda0-47ad-80fd-4a1e60bae190",
 CellID->1585247217],
Cell[30979941, 709056, 613, 12, 22, "Print",ExpressionUUID->"6f4374f8-d602-4bd2-a791-bb41b95d24a4",
 CellID->1567851437],
Cell[30980557, 709070, 633, 12, 22, "Print",ExpressionUUID->"c3fca97a-a07c-4f06-a193-7e60af97db22",
 CellID->1254752969],
Cell[30981193, 709084, 636, 12, 22, "Print",ExpressionUUID->"04fb5b27-ad26-46ed-b518-13bd5d76de44",
 CellID->343736443],
Cell[30981832, 709098, 580, 12, 22, "Print",ExpressionUUID->"528459a7-d166-4db8-9c8f-73b972ab9459",
 CellID->447387071]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[30982461, 709116, 235, 4, 27, "Input",ExpressionUUID->"ce1ea951-59b5-4d3a-9936-2537a0343628",
 CellID->457034461],
Cell[CellGroupData[{
Cell[30982721, 709124, 746, 18, 22, "Print",ExpressionUUID->"136cf734-e002-43ac-a9dc-187a06294d55",
 CellID->894254150],
Cell[30983470, 709144, 581, 12, 22, "Print",ExpressionUUID->"81ac4b33-f27e-4008-b347-b3d5d6460da6",
 CellID->1638937344]
}, Open  ]]
}, Open  ]],
Cell[30984078, 709160, 308, 7, 27, "Input",ExpressionUUID->"472e946e-30b0-4deb-aea8-9464e6871e86",
 CellID->705238580]
}, Closed]],
Cell[CellGroupData[{
Cell[30984423, 709172, 393, 5, 39, "Subsection",ExpressionUUID->"3f0bcd93-51a6-4945-89d2-6e79a73667cc",
 CellID->1994684186],
Cell[30984819, 709179, 378, 6, 40, "Text",ExpressionUUID->"def1e7ea-b590-44cd-95e7-6f4dfbd3bb0c",
 CellID->543172681],
Cell[CellGroupData[{
Cell[30985222, 709189, 232, 3, 41, "Subsubsection",ExpressionUUID->"3cb29779-1323-4284-93c3-f1a40c2f5634",
 CellID->807974394],
Cell[CellGroupData[{
Cell[30985479, 709196, 229, 3, 40, "MathCaption",ExpressionUUID->"f0ce7fef-0d0c-48c0-8e4e-02d723cef4dd",
 CellID->1713956350],
Cell[CellGroupData[{
Cell[30985733, 709203, 242, 4, 27, "Input",ExpressionUUID->"596501f8-4c93-41b0-807f-551a497f2419",
 CellID->611041600],
Cell[CellGroupData[{
Cell[30986000, 709211, 564, 8, 22, "Print",ExpressionUUID->"d16713c7-abbb-4f49-a1c8-3e7a7aef6c67",
 CellID->796657213],
Cell[30986567, 709221, 837, 16, 22, "Print",ExpressionUUID->"d2225d54-eadf-40f4-bd04-e25a5519b69f",
 CellID->2004723493],
Cell[30987407, 709239, 588, 8, 22, "Print",ExpressionUUID->"d989d15d-b10e-44ae-bcde-c2044f2c0fdf",
 CellID->1516809068],
Cell[30987998, 709249, 542, 7, 22, "Print",ExpressionUUID->"d52c0e07-51a5-4b23-a651-29a3f34511ed",
 CellID->231632339],
Cell[30988543, 709258, 526, 7, 22, "Print",ExpressionUUID->"12f08b62-75bf-4103-bcf9-875f0de12be5",
 CellID->1804353209],
Cell[30989072, 709267, 565, 8, 22, "Print",ExpressionUUID->"7eb45eb9-7fd0-4711-88f1-fb742b19677b",
 CellID->1147804376],
Cell[30989640, 709277, 840, 16, 22, "Print",ExpressionUUID->"7f3c6944-64e7-4968-90af-0b44498ff29a",
 CellID->302546431],
Cell[30990483, 709295, 586, 8, 22, "Print",ExpressionUUID->"113267a6-8468-4fe6-a994-e4895ec2026a",
 CellID->1804073368],
Cell[30991072, 709305, 564, 8, 22, "Print",ExpressionUUID->"7bab0fb0-9be5-41d7-828f-8f41624ca7d6",
 CellID->739483983],
Cell[30991639, 709315, 837, 16, 22, "Print",ExpressionUUID->"90ccfa99-dd99-4b13-9c30-2b4f212982bf",
 CellID->1019437092],
Cell[30992479, 709333, 606, 8, 22, "Print",ExpressionUUID->"6c0cc4b0-3e6d-4fde-adbf-d2fd50a00c37",
 CellID->1054669478],
Cell[30993088, 709343, 563, 8, 22, "Print",ExpressionUUID->"260cbd49-c36b-4a43-8aa0-bcd238b08e18",
 CellID->499894408],
Cell[30993654, 709353, 719, 10, 42, "Print",ExpressionUUID->"04fa8bce-bab1-4a9e-a131-d325eadc0e2d",
 CellID->769288192],
Cell[30994376, 709365, 565, 8, 22, "Print",ExpressionUUID->"f529f93a-1710-4c66-85b5-50f71aca8b82",
 CellID->1575088004]
}, Open  ]]
}, Open  ]],
Cell[30994968, 709377, 257, 5, 27, "Input",ExpressionUUID->"955ad36d-1f04-421c-b8f7-776aa247dde9",
 CellID->1529876904]
}, Open  ]],
Cell[CellGroupData[{
Cell[30995262, 709387, 141, 1, 40, "MathCaption",ExpressionUUID->"6f142790-0bda-4488-b70a-aacb88c0ffa0",
 CellID->1211465336],
Cell[CellGroupData[{
Cell[30995428, 709392, 327, 7, 27, "Input",ExpressionUUID->"70cc8692-3f10-4953-8d44-3a549bf270e3",
 CellID->805444945],
Cell[CellGroupData[{
Cell[30995780, 709403, 919, 17, 22, "Print",ExpressionUUID->"15520097-f741-4bc2-9ce5-7fb02237dbf0",
 CellID->473166659],
Cell[30996702, 709422, 931, 17, 22, "Print",ExpressionUUID->"8b9ce7ce-1dda-4d70-882e-e7853429d7c7",
 CellID->380092972]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[30997682, 709445, 470, 12, 27, "Input",ExpressionUUID->"0990b25e-14d6-49d7-84a5-4baaaeedc45e",
 CellID->380288896],
Cell[CellGroupData[{
Cell[30998177, 709461, 1090, 22, 22, "Print",ExpressionUUID->"9d045771-60cb-461d-ad1f-28fd0260612d",
 CellID->1606214226],
Cell[30999270, 709485, 1196, 25, 22, "Print",ExpressionUUID->"6d48a9a7-66d9-4c5d-a924-4a1d8d0688c5",
 CellID->1773053592],
Cell[31000469, 709512, 1167, 23, 22, "Print",ExpressionUUID->"cc140eda-6185-4bed-91ea-4ef29771795f",
 CellID->1228634972],
Cell[31001639, 709537, 1185, 23, 22, "Print",ExpressionUUID->"7a65374f-1b00-445a-a9e7-ddff5c5b1c53",
 CellID->1612708499],
Cell[31002827, 709562, 1006, 20, 22, "Print",ExpressionUUID->"a601eb96-ef30-4a1d-a787-24ce847a2605",
 CellID->1400633906],
Cell[31003836, 709584, 1128, 23, 22, "Print",ExpressionUUID->"3cb12b08-1ecd-412d-8392-183f1d55f7d1",
 CellID->1885803375],
Cell[31004967, 709609, 1144, 24, 22, "Print",ExpressionUUID->"3a5b7f30-79b7-4f74-9b46-a9209f54a2d4",
 CellID->2133661199],
Cell[31006114, 709635, 1170, 24, 22, "Print",ExpressionUUID->"90b1cbd5-ec94-469e-85ba-39e9af9fbc45",
 CellID->1761833987],
Cell[31007287, 709661, 1088, 22, 22, "Print",ExpressionUUID->"18e3c0a3-f864-420c-91e1-a8891373c423",
 CellID->1639294673],
Cell[31008378, 709685, 595, 9, 22, "Print",ExpressionUUID->"ec491262-9278-409f-b8b8-fb67752a1558",
 CellID->1557861748],
Cell[31008976, 709696, 971, 18, 22, "Print",ExpressionUUID->"83368d89-0753-4141-a5fa-3377bf1e424e",
 CellID->861774412],
Cell[31009950, 709716, 598, 9, 22, "Print",ExpressionUUID->"1c139746-fe84-4c8e-86b9-2add99baacbb",
 CellID->1212827116],
Cell[31010551, 709727, 1102, 23, 22, "Print",ExpressionUUID->"2c76a262-52e6-4aba-aee6-3ef478521c4e",
 CellID->70561288],
Cell[31011656, 709752, 1154, 23, 22, "Print",ExpressionUUID->"6a63368e-ec76-4368-a270-536d0ff0992a",
 CellID->512594272],
Cell[31012813, 709777, 1095, 22, 22, "Print",ExpressionUUID->"6cf7efc1-f296-40be-8add-21a5451d00b6",
 CellID->1615021387],
Cell[31013911, 709801, 984, 18, 22, "Print",ExpressionUUID->"db84b139-501b-4b2b-a6fc-ff2e9499e157",
 CellID->1618636876],
Cell[31014898, 709821, 728, 13, 22, "Print",ExpressionUUID->"378afdea-0fbe-407e-bb7c-db0a15e80c74",
 CellID->107134736],
Cell[31015629, 709836, 724, 13, 22, "Print",ExpressionUUID->"615429d4-f7b5-4e27-bea3-bd366de56296",
 CellID->1022211766],
Cell[31016356, 709851, 733, 13, 22, "Print",ExpressionUUID->"cecb06a9-9c90-4e52-b354-93d6b76a2085",
 CellID->443975162],
Cell[31017092, 709866, 1000, 18, 22, "Print",ExpressionUUID->"2d7678db-4795-4974-be6c-45afa465769b",
 CellID->1532358056]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[31018153, 709891, 193, 2, 40, "MathCaption",ExpressionUUID->"05d2bfcf-17c7-4a6a-a904-5b76fe82f67d",
 CellID->1148260682],
Cell[CellGroupData[{
Cell[31018371, 709897, 465, 12, 27, "Input",ExpressionUUID->"fc58fb07-e6ee-40f0-bab4-046b65ec5e45",
 CellID->1529230432],
Cell[CellGroupData[{
Cell[31018861, 709913, 909, 17, 22, "Print",ExpressionUUID->"2abc5ac0-00ad-4312-ba54-51c6ccc21a2c",
 CellID->1717952368],
Cell[31019773, 709932, 968, 18, 22, "Print",ExpressionUUID->"8e5ad076-0b4f-4160-8643-97283938683b",
 CellID->707815061],
Cell[31020744, 709952, 961, 18, 22, "Print",ExpressionUUID->"1b1d1287-e566-4472-af3c-47786dedebca",
 CellID->1335809494],
Cell[31021708, 709972, 975, 18, 22, "Print",ExpressionUUID->"5d76ac16-420b-4587-aa43-e1561264fa64",
 CellID->1156119920],
Cell[31022686, 709992, 968, 18, 22, "Print",ExpressionUUID->"8f3a7288-4e99-46ba-9788-724da2a6d67b",
 CellID->475457384],
Cell[31023657, 710012, 918, 17, 22, "Print",ExpressionUUID->"7e40402e-1fa8-4a39-b292-2ec12b32d304",
 CellID->225602401],
Cell[31024578, 710031, 939, 18, 22, "Print",ExpressionUUID->"ad0bb392-fa70-43db-ac0a-5b033424db40",
 CellID->1390916652],
Cell[31025520, 710051, 1092, 22, 24, "Print",ExpressionUUID->"a501c618-180f-4bbd-a816-84ee34530d5d",
 CellID->1776672569],
Cell[31026615, 710075, 1099, 22, 24, "Print",ExpressionUUID->"998e9b9a-0ee2-4973-8718-ee0c38ed2c0e",
 CellID->1923239295],
Cell[31027717, 710099, 948, 18, 22, "Print",ExpressionUUID->"33125115-cbf2-436c-a5f5-c3fb35b0ff98",
 CellID->2106460068],
Cell[31028668, 710119, 1009, 19, 22, "Print",ExpressionUUID->"f0bfdb22-ef49-400a-aa24-3fc9010c79d9",
 CellID->2006442081],
Cell[31029680, 710140, 1130, 23, 22, "Print",ExpressionUUID->"fe0903d3-d680-45fd-a1ea-3cc5cc197dcd",
 CellID->336346526],
Cell[31030813, 710165, 1150, 24, 22, "Print",ExpressionUUID->"2df21e08-2f09-4775-8ff4-947ce47e6357",
 CellID->799103578],
Cell[31031966, 710191, 1180, 24, 22, "Print",ExpressionUUID->"fa6e3254-0d5d-4a7d-92c4-a756970869a3",
 CellID->171107108],
Cell[31033149, 710217, 1096, 22, 22, "Print",ExpressionUUID->"ab76495d-73d9-4bcd-928a-b70a46d4bff7",
 CellID->316560823],
Cell[31034248, 710241, 1105, 21, 22, "Print",ExpressionUUID->"de43e0af-632b-4b34-b208-a7b970d673fa",
 CellID->1120758542],
Cell[31035356, 710264, 1197, 25, 22, "Print",ExpressionUUID->"03cdaf01-0baa-4c5d-85f2-d8bbfdeefd88",
 CellID->1009747557]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[31036614, 710296, 123, 1, 40, "MathCaption",ExpressionUUID->"09105b1e-f0e8-46cd-8a11-1b8b9bf82cc8",
 CellID->97390299],
Cell[CellGroupData[{
Cell[31036762, 710301, 168, 3, 27, "Input",ExpressionUUID->"8182bc83-c25c-406d-9517-0c2df8eb7e84",
 CellID->456216279],
Cell[31036933, 710306, 939, 17, 22, "Print",ExpressionUUID->"3c7e1500-2d1f-429e-ab2e-27a7ba999d2c",
 CellID->1411411905]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[31037921, 710329, 273, 5, 40, "MathCaption",ExpressionUUID->"20518fc6-47af-4312-bc5f-a0cce6caa19c",
 CellID->1178915150],
Cell[CellGroupData[{
Cell[31038219, 710338, 354, 11, 27, "Input",ExpressionUUID->"d045dfa4-69e3-46bc-81aa-15430dc7e2df",
 CellID->1307502051],
Cell[31038576, 710351, 1074, 20, 22, "Print",ExpressionUUID->"f6954cdb-cfc8-4d95-857d-1c29a605053d",
 CellID->495522573],
Cell[31039653, 710373, 1716, 43, 32, "Output",ExpressionUUID->"28cfd954-42b3-4ee5-bdb2-47537247b25a",
 CellID->306301345]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[31041430, 710423, 240, 3, 41, "Subsubsection",ExpressionUUID->"6a438a47-4bf7-47d2-9665-8ec7d2f30988",
 CellID->1896835986],
Cell[31041673, 710428, 376, 7, 23, "Text",ExpressionUUID->"c71be471-1d2d-4450-bb4f-bbe458623bd4",
 CellID->1679298556],
Cell[CellGroupData[{
Cell[31042074, 710439, 324, 6, 40, "MathCaption",ExpressionUUID->"23286258-8bf2-474a-8a91-1782df0dd7c7",
 CellID->2064139895],
Cell[31042401, 710447, 1005, 32, 77, "Input",ExpressionUUID->"033f9dff-bb6e-4ee4-893b-3e027d0c8c91",
 CellID->1548381627]
}, Open  ]],
Cell[CellGroupData[{
Cell[31043443, 710484, 361, 6, 40, "MathCaption",ExpressionUUID->"78f554ec-f58e-420d-be3a-f07bcacfa6ba",
 CellID->1246174314],
Cell[CellGroupData[{
Cell[31043829, 710494, 295, 6, 27, "Input",ExpressionUUID->"8db67e94-be57-459f-83ed-74882d1a5436",
 CellID->895709105],
Cell[CellGroupData[{
Cell[31044149, 710504, 2437, 64, 36, "Print",ExpressionUUID->"874e064b-3f6b-49d2-b6fc-73bf4a442732",
 CellID->488403556],
Cell[31046589, 710570, 1993, 48, 28, "Print",ExpressionUUID->"f80235dc-e71d-4509-87dc-211382edad71",
 CellID->916066858],
Cell[31048585, 710620, 1994, 48, 28, "Print",ExpressionUUID->"41e293bb-a4a1-4054-aac1-1cddc8d2aee5",
 CellID->1025417063],
Cell[31050582, 710670, 1995, 48, 28, "Print",ExpressionUUID->"e82fd881-2bae-4411-958f-aa95f5e61892",
 CellID->628310666],
Cell[31052580, 710720, 3220, 80, 28, "Print",ExpressionUUID->"d8e9e2b6-a82f-42c0-9092-e1ca833c19d0",
 CellID->840271163],
Cell[31055803, 710802, 2462, 64, 44, "Print",ExpressionUUID->"151177a8-063f-46ed-94bf-f148bf4b3cf1",
 CellID->584973869],
Cell[31058268, 710868, 1995, 48, 28, "Print",ExpressionUUID->"b252307c-a691-42e3-8189-936d81352b30",
 CellID->475680317],
Cell[31060266, 710918, 1994, 48, 28, "Print",ExpressionUUID->"39a6f7ad-5d39-4886-8770-3e90c389a03d",
 CellID->1824953685],
Cell[31062263, 710968, 3223, 80, 28, "Print",ExpressionUUID->"316fa9a9-527b-4134-9ff7-4e884ce035af",
 CellID->1635137341],
Cell[31065489, 711050, 3220, 80, 28, "Print",ExpressionUUID->"889534eb-ddbb-49f4-956a-660d754c1f0c",
 CellID->405059528],
Cell[31068712, 711132, 2182, 55, 28, "Print",ExpressionUUID->"6f5b1187-02fc-40f4-a3be-273a2e4ee04e",
 CellID->312326106],
Cell[31070897, 711189, 1993, 48, 28, "Print",ExpressionUUID->"85a6c695-f8ba-486a-9325-43de5d226b54",
 CellID->814830274],
Cell[31072893, 711239, 3220, 80, 28, "Print",ExpressionUUID->"05c6c188-6d25-41c9-be08-05a64892f994",
 CellID->349198199],
Cell[31076116, 711321, 3221, 80, 28, "Print",ExpressionUUID->"17e88f77-e430-48ca-9e2d-72ab261fd99b",
 CellID->2048806989],
Cell[31079340, 711403, 3221, 80, 28, "Print",ExpressionUUID->"225c5473-98f9-4750-ba1b-0e39db9f2343",
 CellID->1546268125],
Cell[31082564, 711485, 2632, 71, 28, "Print",ExpressionUUID->"ba47ef64-3872-4f89-ada1-ad7ff0199308",
 CellID->1428764528]
}, Open  ]],
Cell[31085211, 711559, 22619, 606, 77, "Output",ExpressionUUID->"337ef31c-2790-46fb-8276-68ab12722b65",
 CellID->83281342]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[31107879, 712171, 280, 5, 40, "MathCaption",ExpressionUUID->"db9f21eb-cbda-432e-b4e4-429ef7a6e87f",
 CellID->688445837],
Cell[CellGroupData[{
Cell[31108184, 712180, 236, 4, 27, "Input",ExpressionUUID->"ae1f1244-b072-44f1-9f7f-4be92dd310ab",
 CellID->266995355],
Cell[31108423, 712186, 29658, 790, 76, "Output",ExpressionUUID->"9ea4b071-f1cb-4772-b6a5-7e982c22b50e",
 CellID->1697660797]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[31138130, 712982, 239, 3, 40, "MathCaption",ExpressionUUID->"a35c88ae-789a-4ea7-923c-148df8e07865",
 CellID->1033605800],
Cell[CellGroupData[{
Cell[31138394, 712989, 720, 24, 27, "Input",ExpressionUUID->"f4c55fc9-9d7a-4bcb-8803-d5cd1c98a54d",
 CellID->817460017],
Cell[31139117, 713015, 52939, 1498, 56, "Output",ExpressionUUID->"19287cd2-9614-4ce9-8a5f-d9ef2a1137b4",
 CellID->1117281137]
}, Open  ]],
Cell[CellGroupData[{
Cell[31192093, 714518, 416, 11, 27, "Input",ExpressionUUID->"9037ccd1-e733-4e54-bee7-c26659a12f3c",
 CellID->1141720579],
Cell[31192512, 714531, 10590, 313, 57, "Output",ExpressionUUID->"d65f6e42-6c36-4662-8040-665af2a08848",
 CellID->547060191]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[31203151, 714850, 214, 2, 40, "MathCaption",ExpressionUUID->"dea562ea-196f-4a98-9f4d-af866b4a7184",
 CellID->1654533724],
Cell[CellGroupData[{
Cell[31203390, 714856, 262, 5, 27, "Input",ExpressionUUID->"5ccff746-b94b-41b1-88c9-394f88189a47",
 CellID->2130037912],
Cell[31203655, 714863, 2573, 83, 45, "Output",ExpressionUUID->"88c686ee-b966-4c1c-8236-36decb3730d2",
 CellID->1063137330]
}, Open  ]],
Cell[CellGroupData[{
Cell[31206265, 714951, 364, 10, 27, "Input",ExpressionUUID->"e9890cf1-ade1-41f1-a960-f8b3da51bf83",
 CellID->1300447763],
Cell[CellGroupData[{
Cell[31206654, 714965, 2591, 70, 34, "Print",ExpressionUUID->"e9ccd3cf-0b7b-4173-a95a-9a04b6490b94",
 CellID->1085162270],
Cell[31209248, 715037, 1996, 48, 28, "Print",ExpressionUUID->"d197abeb-0fb9-4c29-baba-12865df68114",
 CellID->1891141566],
Cell[31211247, 715087, 1996, 48, 28, "Print",ExpressionUUID->"7e5edb88-3a98-42da-a1b0-4791173bdceb",
 CellID->1658324162],
Cell[31213246, 715137, 1996, 48, 28, "Print",ExpressionUUID->"ce0c42e4-b0e5-43a7-aa9f-d5b07676703d",
 CellID->1411419968],
Cell[31215245, 715187, 3218, 80, 28, "Print",ExpressionUUID->"8f1c470e-9926-48e9-8bc8-c88435430962",
 CellID->631037414],
Cell[31218466, 715269, 2417, 63, 36, "Print",ExpressionUUID->"931149c6-557b-4fa8-9eb9-d65c32fb82f3",
 CellID->1571568212],
Cell[31220886, 715334, 1996, 48, 28, "Print",ExpressionUUID->"4431e7ca-7565-47a6-9337-f8757e02e1aa",
 CellID->1445564631],
Cell[31222885, 715384, 1996, 48, 28, "Print",ExpressionUUID->"f1d6c299-882f-4a4e-97ce-f887e74899c4",
 CellID->1122254706],
Cell[31224884, 715434, 3218, 80, 28, "Print",ExpressionUUID->"6e2beb36-a95d-4958-b5b3-45fe68644a37",
 CellID->385972141],
Cell[31228105, 715516, 3219, 80, 28, "Print",ExpressionUUID->"945977cc-869e-4cb3-8857-c2e025dca29d",
 CellID->1633159847],
Cell[31231327, 715598, 2218, 56, 38, "Print",ExpressionUUID->"1ce04239-f242-4110-8d47-4ee653c9a746",
 CellID->1529056222],
Cell[31233548, 715656, 1996, 48, 28, "Print",ExpressionUUID->"88083b96-439c-4d7e-a3c7-4eba612d10dd",
 CellID->2058603152],
Cell[31235547, 715706, 3218, 80, 28, "Print",ExpressionUUID->"6ada885c-c343-49ec-b74c-f3726a81acdb",
 CellID->834138847],
Cell[31238768, 715788, 3218, 80, 28, "Print",ExpressionUUID->"aeea2c5f-5c98-4684-94fd-0b1d556b007d",
 CellID->598353913],
Cell[31241989, 715870, 3219, 80, 28, "Print",ExpressionUUID->"4144fcd5-eabe-4960-a485-87dcd80e2732",
 CellID->2015780537],
Cell[31245211, 715952, 2636, 71, 41, "Print",ExpressionUUID->"cd14d3f2-d6a7-45df-9557-b8b7307a8700",
 CellID->521470287]
}, Open  ]],
Cell[31247862, 716026, 30018, 799, 73, "Output",ExpressionUUID->"293f1f55-9d80-4b8a-9404-844e03a7ea6f",
 CellID->470350202]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[31277929, 716831, 170, 2, 40, "MathCaption",ExpressionUUID->"47a75909-a629-40f3-98b6-27b877bf984a",
 CellID->391610265],
Cell[CellGroupData[{
Cell[31278124, 716837, 679, 22, 27, "Input",ExpressionUUID->"e315f4e6-a244-40b9-acdd-5ed984fee8db",
 CellID->1903829447],
Cell[31278806, 716861, 52848, 1496, 56, "Output",ExpressionUUID->"ba691a27-b34b-47ae-b29c-16e3d257d1f3",
 CellID->288540407]
}, Open  ]],
Cell[CellGroupData[{
Cell[31331691, 718362, 345, 8, 27, "Input",ExpressionUUID->"4f43098e-8051-42f3-9647-ec8d9e212cb6",
 CellID->988287745],
Cell[31332039, 718372, 10847, 319, 50, "Output",ExpressionUUID->"6f7ce65f-05af-4a7e-9748-fe4f42682e7d",
 CellID->480545523]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[31342935, 718697, 201, 2, 40, "MathCaption",ExpressionUUID->"b8daa658-60fe-4744-a16f-44a28f731661",
 CellID->1214767293],
Cell[CellGroupData[{
Cell[31343161, 718703, 291, 7, 27, "Input",ExpressionUUID->"68ccf8a3-140b-4064-92ae-23d2e2ab8d4b",
 CellID->186812753],
Cell[31343455, 718712, 10334, 309, 57, "Output",ExpressionUUID->"f7803d46-d930-48f3-9536-4fdf12d40f17",
 CellID->1990092341]
}, Open  ]],
Cell[CellGroupData[{
Cell[31353826, 719026, 308, 7, 27, "Input",ExpressionUUID->"c7f0b36a-94cf-49a4-ac8a-294a5f5e6156",
 CellID->672927078],
Cell[31354137, 719035, 678, 13, 22, "Print",ExpressionUUID->"fde3957a-907b-4933-bd02-db9450d5e44b",
 CellID->424173162],
Cell[31354818, 719050, 10587, 315, 50, "Output",ExpressionUUID->"6ff22049-0e7b-4114-b12e-158014aee0d1",
 CellID->1580109341]
}, Open  ]],
Cell[CellGroupData[{
Cell[31365442, 719370, 281, 6, 27, "Input",ExpressionUUID->"825e09ba-c183-473f-8b98-1aacb88b5d68",
 CellID->2077978809],
Cell[31365726, 719378, 197, 3, 26, "Output",ExpressionUUID->"3983fb94-1150-42d6-9880-f7de7211397f",
 CellID->1114915385]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[31365972, 719387, 183, 2, 40, "MathCaption",ExpressionUUID->"6533f8c7-0eef-44a0-a072-900d70a60381",
 CellID->1443006693],
Cell[CellGroupData[{
Cell[31366180, 719393, 369, 10, 27, "Input",ExpressionUUID->"f97f4cb7-999c-4d0f-9ac1-f82a461c5964",
 CellID->504774293],
Cell[31366552, 719405, 22403, 636, 57, "Output",ExpressionUUID->"6df7bdcb-98fa-4d3e-af7e-0683ae3413b9",
 CellID->1588055620]
}, Open  ]],
Cell[CellGroupData[{
Cell[31388992, 720046, 230, 4, 27, "Input",ExpressionUUID->"c87e92d4-639d-4160-8660-feadd6cb5ea0",
 CellID->1994969649],
Cell[31389225, 720052, 15704, 462, 57, "Output",ExpressionUUID->"702c0622-2f8a-401b-bf9b-ed57e064661a",
 CellID->1524040424]
}, Open  ]],
Cell[CellGroupData[{
Cell[31404966, 720519, 224, 4, 27, "Input",ExpressionUUID->"59179072-ce57-40db-9984-fbc3ea9c6b3b",
 CellID->170760014],
Cell[31405193, 720525, 8993, 252, 32, "Output",ExpressionUUID->"258f703b-c675-4255-8830-c64d7f1bb0ec",
 CellID->1509948399]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[31414235, 720783, 157, 2, 40, "MathCaption",ExpressionUUID->"31eda655-aedb-4748-8039-bb53d0bb8a81",
 CellID->195861224],
Cell[31414395, 720787, 180, 4, 27, "Input",ExpressionUUID->"045692e0-1376-44dc-b294-2b4d58667611",
 CellID->1894644564]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[31414624, 720797, 326, 5, 41, "Subsubsection",ExpressionUUID->"e2af22ec-2645-4aa6-ac12-6ff329d0baad",
 CellID->1072254693],
Cell[31414953, 720804, 484, 8, 40, "Text",ExpressionUUID->"76f1c64b-eadc-4587-ae4e-ddc5e619bc66",
 CellID->1810798472],
Cell[CellGroupData[{
Cell[31415462, 720816, 187, 2, 40, "MathCaption",ExpressionUUID->"4c92cc5d-73f1-4cc7-88df-11bd4d5428ab",
 CellID->103387443],
Cell[CellGroupData[{
Cell[31415674, 720822, 596, 14, 27, "Input",ExpressionUUID->"f6b2ba2f-1eed-4e3d-94ba-c704196f5c70",
 CellID->2099157267],
Cell[CellGroupData[{
Cell[31416295, 720840, 845, 16, 22, "Print",ExpressionUUID->"63f668cc-96d7-4259-b2aa-ca8d5448b785",
 CellID->1125860365],
Cell[31417143, 720858, 885, 17, 22, "Print",ExpressionUUID->"2133245e-400d-4ee8-8a9f-0c55be745469",
 CellID->856740838],
Cell[31418031, 720877, 886, 17, 22, "Print",ExpressionUUID->"78bbea23-bcba-4852-9bd6-6e3bc2b2d3db",
 CellID->365411131],
Cell[31418920, 720896, 915, 17, 22, "Print",ExpressionUUID->"88d36984-02fe-47ca-9fba-48e4ac357eb5",
 CellID->1809131944],
Cell[31419838, 720915, 911, 17, 22, "Print",ExpressionUUID->"729b3a0e-0f32-4d1c-bbc1-ab884d082d71",
 CellID->2007108582],
Cell[31420752, 720934, 852, 16, 22, "Print",ExpressionUUID->"beb3f829-5e6a-4245-a869-fff699f10aac",
 CellID->800810598],
Cell[31421607, 720952, 876, 17, 22, "Print",ExpressionUUID->"8d474cbe-254a-4b72-8e88-0f4f7b6370c3",
 CellID->943704837],
Cell[31422486, 720971, 1030, 21, 24, "Print",ExpressionUUID->"473156ca-3379-4b66-a6eb-d33c248e41c7",
 CellID->1680462364],
Cell[31423519, 720994, 1043, 22, 24, "Print",ExpressionUUID->"21dcc1f6-1e2b-4b00-a35a-4746829d27f4",
 CellID->1973510051],
Cell[31424565, 721018, 882, 17, 22, "Print",ExpressionUUID->"190d0943-f739-4717-b8a5-b272efcc2590",
 CellID->898499242],
Cell[31425450, 721037, 943, 18, 22, "Print",ExpressionUUID->"435580d6-da6a-42af-b32f-75aee4f41944",
 CellID->2119238237],
Cell[31426396, 721057, 1065, 22, 22, "Print",ExpressionUUID->"22674112-98f3-42ae-ba2f-07a6e857c108",
 CellID->2020763764],
Cell[31427464, 721081, 1084, 23, 22, "Print",ExpressionUUID->"c1d03025-7b88-48ad-a481-36501e0ff4e5",
 CellID->522704243],
Cell[31428551, 721106, 1115, 23, 22, "Print",ExpressionUUID->"85b390f1-4fec-464b-800c-4174300236b4",
 CellID->1882095871],
Cell[31429669, 721131, 1029, 21, 22, "Print",ExpressionUUID->"2de08f74-2426-4ff3-96f1-2a18c3864789",
 CellID->2098667234],
Cell[31430701, 721154, 1039, 20, 22, "Print",ExpressionUUID->"c87d2623-b57b-4ebc-8fad-1ad605f6ea1d",
 CellID->2028783510],
Cell[31431743, 721176, 1129, 24, 22, "Print",ExpressionUUID->"0b4e6b73-504c-4c6b-9908-15bee4c979c8",
 CellID->19105504]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[31432921, 721206, 254, 7, 27, "Input",ExpressionUUID->"e2918619-d982-4dcf-9112-144c85bf2445",
 CellID->1352115308],
Cell[31433178, 721215, 1426, 37, 32, "Output",ExpressionUUID->"a3063a87-a30a-4ee3-a7c7-fb961383adf9",
 CellID->1131142325]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[31434653, 721258, 129, 1, 40, "MathCaption",ExpressionUUID->"1fde3037-47a5-4d67-a297-60c2c656c8c7",
 CellID->1994660060],
Cell[CellGroupData[{
Cell[31434807, 721263, 990, 31, 27, "Input",ExpressionUUID->"225f231b-987f-4c6e-89f4-bb4cc02f8489",
 CellID->2031898750],
Cell[31435800, 721296, 3235, 113, 38, "Output",ExpressionUUID->"4faa870e-b21c-484b-a0f5-9dc07d0bc991",
 CellID->1583813031]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[31439084, 721415, 114, 1, 40, "MathCaption",ExpressionUUID->"1238f26f-661e-49ee-b834-cd0eeec6af60",
 CellID->656838601],
Cell[CellGroupData[{
Cell[31439223, 721420, 271, 6, 27, "Input",ExpressionUUID->"672ca59a-e028-4289-b08c-7139f74a6967",
 CellID->1420421427],
Cell[31439497, 721428, 2200, 72, 43, "Output",ExpressionUUID->"0ce47ce7-1ab9-445e-8032-61c3310f99d3",
 CellID->1085807452]
}, Open  ]],
Cell[CellGroupData[{
Cell[31441734, 721505, 413, 13, 27, "Input",ExpressionUUID->"b47e879f-2102-45a6-8afb-f77970761766",
 CellID->1586431799],
Cell[31442150, 721520, 22310, 595, 71, "Output",ExpressionUUID->"ba4bfa49-2742-4e78-af7c-d3afb3e8a553",
 CellID->1997297205]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[31464509, 722121, 191, 2, 40, "MathCaption",ExpressionUUID->"a4b1cbdb-2098-40ef-b9aa-fec23a694b7d",
 CellID->1890739288],
Cell[CellGroupData[{
Cell[31464725, 722127, 291, 7, 27, "Input",ExpressionUUID->"da16bf41-e076-4932-afd5-dcbffb337efc",
 CellID->954075472],
Cell[31465019, 722136, 10594, 313, 57, "Output",ExpressionUUID->"b3c809b6-7887-4fc1-8657-0c74d1b23e4a",
 CellID->1257238178]
}, Open  ]],
Cell[CellGroupData[{
Cell[31475650, 722454, 261, 6, 27, "Input",ExpressionUUID->"e4f9c2ec-8b36-4c0a-b0cd-c07233147727",
 CellID->1877005023],
Cell[31475914, 722462, 20084, 579, 57, "Output",ExpressionUUID->"b8731b58-e324-4b91-99e5-b0c15fd2c902",
 CellID->1310454813]
}, Open  ]],
Cell[CellGroupData[{
Cell[31496035, 723046, 167, 3, 27, "Input",ExpressionUUID->"ad7666f4-0f8d-440b-b5bf-0dffeffaf462",
 CellID->126772144],
Cell[31496205, 723051, 292591, 8492, 478, "Output",ExpressionUUID->"b8b66085-f948-4785-9ad3-21ad1642c448",
 CellID->221758459]
}, Open  ]],
Cell[CellGroupData[{
Cell[31788833, 731548, 270, 7, 27, "Input",ExpressionUUID->"80ab3be1-07a9-41db-a2be-a4f7986c2158",
 CellID->901786287],
Cell[31789106, 731557, 34856, 1105, 101, "Output",ExpressionUUID->"f2726499-6382-4b81-adf7-2d4d232d5519",
 CellID->1210292868]
}, Open  ]],
Cell[CellGroupData[{
Cell[31823999, 732667, 309, 6, 27, "Input",ExpressionUUID->"39ad73e1-3394-49a9-98eb-e0220977daed",
 CellID->1223675932],
Cell[31824311, 732675, 13969, 420, 52, "Output",ExpressionUUID->"68671053-cf43-46b2-882b-9aa853715e8d",
 CellID->427128092]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[31838341, 733102, 342, 4, 41, "Subsubsection",ExpressionUUID->"85fdac02-de01-46ac-8725-54464977baec",
 CellID->1221044385],
Cell[31838686, 733108, 355, 6, 23, "Text",ExpressionUUID->"c59e6737-267f-4913-95d2-537d6c03c4c0",
 CellID->889724663],
Cell[CellGroupData[{
Cell[31839066, 733118, 263, 5, 40, "MathCaption",ExpressionUUID->"e28239af-5828-4a88-962d-292297081d49",
 CellID->1397892159],
Cell[CellGroupData[{
Cell[31839354, 733127, 548, 17, 27, "Input",ExpressionUUID->"7100fcce-873b-4bb2-aaea-82335e3cd7f3",
 CellID->673776980],
Cell[31839905, 733146, 200838, 5068, 70, "Output",ExpressionUUID->"2f834320-15aa-4806-b54b-46d1886cc3eb",
 CellID->1233106143]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[32040792, 738220, 117, 1, 40, "MathCaption",ExpressionUUID->"8710df2c-2b51-4ec1-8347-7af67596cb81",
 CellID->1640641540],
Cell[CellGroupData[{
Cell[32040934, 738225, 370, 10, 27, "Input",ExpressionUUID->"d720de4c-73f6-4c97-bf4a-17a82bfde98f",
 CellID->572335300],
Cell[32041307, 738237, 37484, 1064, 70, "Output",ExpressionUUID->"d5456605-20cc-4739-9da7-b8ba28869df8",
 CellID->1597751851]
}, Open  ]],
Cell[CellGroupData[{
Cell[32078828, 739306, 185, 4, 27, "Input",ExpressionUUID->"3060b31a-89df-46ba-a0f1-b8f7aa56f946",
 CellID->941146507],
Cell[32079016, 739312, 50199, 1466, 70, "Output",ExpressionUUID->"1aba94a4-c9e3-42fa-9b68-8b45812f89b8",
 CellID->1279837669]
}, Open  ]],
Cell[CellGroupData[{
Cell[32129252, 740783, 278, 6, 27, "Input",ExpressionUUID->"bc4e2281-5b32-4a8e-bac2-865167eb8815",
 CellID->1824173380],
Cell[32129533, 740791, 263195, 7411, 70, "Output",ExpressionUUID->"5011a6b7-b941-4ab5-84f5-ff890566cc8c",
 CellID->1035494531]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[32392777, 748208, 199, 4, 40, "MathCaption",ExpressionUUID->"d53b529b-2e83-4d63-bc8b-86b7f09c5d53",
 CellID->1277416663],
Cell[CellGroupData[{
Cell[32393001, 748216, 569, 15, 27, "Input",ExpressionUUID->"22216ff6-f151-4e6f-a204-7ddb898c4f78",
 CellID->234373083],
Cell[32393573, 748233, 65545, 1795, 70, "Output",ExpressionUUID->"199c7454-b167-4d4c-bbed-b10f46dc51e5",
 CellID->349107229]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[32459167, 750034, 113, 1, 40, "MathCaption",ExpressionUUID->"cde347d0-0b25-4f9a-ae99-cd22956f859b",
 CellID->479789036],
Cell[CellGroupData[{
Cell[32459305, 750039, 193, 4, 27, "Input",ExpressionUUID->"4a675e87-1f4d-4212-9e4a-5da5cfee009d",
 CellID->13233567],
Cell[32459501, 750045, 65519, 1795, 70, "Output",ExpressionUUID->"506a1d70-25dc-465c-848c-af82b80ea85b",
 CellID->519874199]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[32525069, 751846, 129, 1, 40, "MathCaption",ExpressionUUID->"21661525-6ef4-44e5-8bd6-244d93130807",
 CellID->284379225],
Cell[CellGroupData[{
Cell[32525223, 751851, 505, 13, 27, "Input",ExpressionUUID->"be365d0c-4598-4922-a401-d05a57077863",
 CellID->239099082],
Cell[32525731, 751866, 48515, 1428, 70, "Output",ExpressionUUID->"9f737625-1131-4879-9f2d-42ab66d754ab",
 CellID->1562186597]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[32574295, 753300, 247, 4, 56, "MathCaption",ExpressionUUID->"68b8a124-ad6a-47bc-b642-1fb1535bb3bd",
 CellID->2000808040],
Cell[32574545, 753306, 346, 8, 66, "Input",ExpressionUUID->"c5daf762-08eb-451c-bafd-2c2b1946b0ce",
 CellID->134299907],
Cell[CellGroupData[{
Cell[32574916, 753318, 891, 27, 27, "Input",ExpressionUUID->"991f9245-05e8-497d-a02e-00ce230e776b",
 CellID->159286465],
Cell[32575810, 753347, 5473, 134, 70, "Print",ExpressionUUID->"36dc944b-b68a-466d-9864-2996d9a28992",
 CellID->535067557],
Cell[32581286, 753483, 309650, 7896, 70, "Output",ExpressionUUID->"90fcfae9-f337-43f3-a2a9-3d67ca4fdd8b",
 CellID->1366400510]
}, Open  ]],
Cell[CellGroupData[{
Cell[32890973, 761384, 913, 30, 47, "Input",ExpressionUUID->"5527432a-bd88-4fd0-aadb-d7fb29e916b0",
 CellID->949306026],
Cell[32891889, 761416, 5461, 134, 70, "Print",ExpressionUUID->"eec716a5-d95e-4f9e-929d-cb61e14f7ef7",
 CellID->2050734199],
Cell[32897353, 761552, 144157, 3687, 70, "Output",ExpressionUUID->"95405406-43aa-483e-b74d-0a503bfd82e0",
 CellID->1724631890]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[33041559, 765245, 181, 3, 40, "MathCaption",ExpressionUUID->"a43ab713-ba1f-407b-ac5c-2bd499b5f197",
 CellID->1250815119],
Cell[CellGroupData[{
Cell[33041765, 765252, 217, 5, 27, "Input",ExpressionUUID->"10414202-7102-4db4-87ab-b655dc251b22",
 CellID->732284550],
Cell[33041985, 765259, 1903, 49, 70, "Output",ExpressionUUID->"ea0468b6-cd9f-4545-af5d-1c80e1f4c47b",
 CellID->1301391671]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[33043937, 765314, 126, 1, 40, "MathCaption",ExpressionUUID->"9f7201f0-e74e-4786-a751-3286470d89cf",
 CellID->2144077367],
Cell[CellGroupData[{
Cell[33044088, 765319, 252, 6, 27, "Input",ExpressionUUID->"f6c19781-c59b-4a9b-ac56-aaa192fe499b",
 CellID->732710509],
Cell[33044343, 765327, 1225, 31, 70, "Output",ExpressionUUID->"7d7eff90-0eaa-44b1-aacd-5553d162331d",
 CellID->368869802]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[33045617, 765364, 227, 3, 40, "MathCaption",ExpressionUUID->"ecfac672-a095-48d0-ac02-d169c0e8b5f7",
 CellID->911802371],
Cell[CellGroupData[{
Cell[33045869, 765371, 571, 15, 27, "Input",ExpressionUUID->"adc6e788-a132-42e2-b12f-d2ca6a948400",
 CellID->1275362256],
Cell[33046443, 765388, 48532, 1428, 70, "Output",ExpressionUUID->"d9f5f813-cf75-4378-ae12-514bbad6febe",
 CellID->1956956972]
}, Open  ]],
Cell[CellGroupData[{
Cell[33095012, 766821, 286, 5, 27, "Input",ExpressionUUID->"3f9c398e-e80c-414b-a790-2e063aac65a1",
 CellID->197703657],
Cell[33095301, 766828, 333, 5, 70, "Output",ExpressionUUID->"158a7bbd-e1b7-4d5a-b3f4-4bca8f872cfb",
 CellID->1863774599]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[33095695, 766840, 465, 6, 41, "Subsubsection",ExpressionUUID->"392ba1ff-32ee-4089-a98d-f55575f12edf",
 CellID->346755344],
Cell[33096163, 766848, 387, 7, 23, "Text",ExpressionUUID->"2ee7dbad-069a-4faf-adc4-9ed7ccbdf8b1",
 CellID->490983037],
Cell[CellGroupData[{
Cell[33096575, 766859, 251, 3, 40, "MathCaption",ExpressionUUID->"f64e8db9-bb0c-4a54-95b9-63128068dbe9",
 CellID->984582989],
Cell[CellGroupData[{
Cell[33096851, 766866, 663, 17, 27, "Input",ExpressionUUID->"2ce53276-9ed0-43fe-a6a9-8942c5fb96d3",
 CellID->1706034063],
Cell[33097517, 766885, 120426, 3509, 242, "Output",ExpressionUUID->"f9f63f23-65f4-4216-a8ed-d752b3f894a4",
 CellID->1263210251]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[33217992, 770400, 264, 3, 40, "MathCaption",ExpressionUUID->"5a4d012b-9501-4eda-b0df-2957b7ff24d2",
 CellID->1663828751],
Cell[CellGroupData[{
Cell[33218281, 770407, 439, 11, 27, "Input",ExpressionUUID->"4b87912c-0cb7-4042-b6e1-0910eeedc719",
 CellID->72553858],
Cell[33218723, 770420, 406, 7, 26, "Output",ExpressionUUID->"9e6032a9-1a24-48ba-8abf-c4adfeafb75e",
 CellID->751354315]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[33219178, 770433, 293, 5, 40, "MathCaption",ExpressionUUID->"600273b2-62d5-4693-a574-70a1f6a85ea6",
 CellID->417221997],
Cell[CellGroupData[{
Cell[33219496, 770442, 418, 11, 27, "Input",ExpressionUUID->"904aa7ce-1b36-4d1f-9ad5-3bc5e44cfead",
 CellID->42034799],
Cell[33219917, 770455, 8028, 218, 32, "Output",ExpressionUUID->"93e1d808-70ec-4465-8051-3486d94da4c7",
 CellID->808127845]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[33227994, 770679, 263, 5, 40, "MathCaption",ExpressionUUID->"1d009b24-8d61-4d65-a800-7aa7df520be1",
 CellID->1364939662],
Cell[CellGroupData[{
Cell[33228282, 770688, 289, 6, 27, "Input",ExpressionUUID->"1aa1cd16-5036-405e-b4b0-9e788175c311",
 CellID->2039489065],
Cell[33228574, 770696, 227865, 6746, 523, "Output",ExpressionUUID->"85391685-8a06-404b-a678-96ce5ba05862",
 CellID->1518107287]
}, Open  ]],
Cell[CellGroupData[{
Cell[33456476, 777447, 393, 8, 27, "Input",ExpressionUUID->"37831916-acf2-4903-a995-09b51fa8d395",
 CellID->875107252],
Cell[33456872, 777457, 120326, 3507, 242, "Output",ExpressionUUID->"07433b5b-c5e7-4b3c-bd71-0fd7fba0a765",
 CellID->575962602]
}, Open  ]],
Cell[CellGroupData[{
Cell[33577235, 780969, 229, 4, 27, "Input",ExpressionUUID->"f30bd70c-119c-49cb-91fd-820ef14ef05c",
 CellID->452783367],
Cell[33577467, 780975, 288, 5, 26, "Output",ExpressionUUID->"623de4dd-87d2-44a8-8575-32b56fbb2f08",
 CellID->1494654785]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[33577804, 780986, 212, 2, 40, "MathCaption",ExpressionUUID->"fda1508c-63b5-479a-862f-a4d36c45fd63",
 CellID->523132918],
Cell[CellGroupData[{
Cell[33578041, 780992, 707, 17, 27, "Input",ExpressionUUID->"1c1521ce-bd09-4364-b7bd-000795caf175",
 CellID->664211117],
Cell[33578751, 781011, 33868, 1027, 458, "Output",ExpressionUUID->"fca560cf-16de-4f26-8c6e-7f14484a00e4",
 CellID->1546752536]
}, Open  ]],
Cell[33612634, 782041, 839, 26, 27, "Input",ExpressionUUID->"5a398ef2-5732-401d-aed7-3c1dcb9ee78c",
 CellID->1318404106]
}, Open  ]],
Cell[CellGroupData[{
Cell[33613510, 782072, 282, 4, 40, "MathCaption",ExpressionUUID->"d70976d4-bff3-4435-b204-45d2494da542",
 CellID->1376471820],
Cell[CellGroupData[{
Cell[33613817, 782080, 413, 9, 27, "Input",ExpressionUUID->"7bc2c9c1-a605-4db0-aa03-1a99c53e8434",
 CellID->1668033256],
Cell[33614233, 782091, 38222, 1028, 536, "Output",ExpressionUUID->"77fa5dc4-8d7c-4af7-8841-bcad716dd4d7",
 CellID->980325617]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[33652516, 783126, 511, 7, 41, "Subsubsection",ExpressionUUID->"1600e4f9-c0a8-4191-aced-c8c6a812c098",
 CellID->1364148720],
Cell[CellGroupData[{
Cell[33653052, 783137, 234, 4, 27, "Input",ExpressionUUID->"e09043d3-7842-401d-99cd-f1df58d2da64",
 CellID->268720204],
Cell[CellGroupData[{
Cell[33653311, 783145, 642, 13, 22, "Print",ExpressionUUID->"c4467cd6-9876-4b72-a92d-36fab95ddaeb",
 CellID->1480324350],
Cell[33653956, 783160, 634, 13, 22, "Print",ExpressionUUID->"56d4f122-0720-40f2-83f2-5789e28228a9",
 CellID->1213299955],
Cell[33654593, 783175, 662, 13, 22, "Print",ExpressionUUID->"749ec44c-22db-4aaa-9645-a5a5ef2df16f",
 CellID->1575115420],
Cell[33655258, 783190, 629, 13, 22, "Print",ExpressionUUID->"ae61ed66-33d1-4a5b-becf-afc75e0bb75d",
 CellID->933947371],
Cell[33655890, 783205, 649, 13, 22, "Print",ExpressionUUID->"11db8726-5c77-462a-97b0-6bf957eec1fd",
 CellID->895488474],
Cell[33656542, 783220, 643, 13, 22, "Print",ExpressionUUID->"16f8f05e-58ee-4601-b046-a20f35eee716",
 CellID->909384342],
Cell[33657188, 783235, 637, 13, 22, "Print",ExpressionUUID->"307e3fe4-9acc-4715-a2e4-c1219fda9984",
 CellID->381520295],
Cell[33657828, 783250, 632, 13, 22, "Print",ExpressionUUID->"838ed2cb-f866-4da2-b98a-2f82e13b4a96",
 CellID->1972911770],
Cell[33658463, 783265, 626, 13, 22, "Print",ExpressionUUID->"4830fafb-cf39-497a-bb0a-34e6bddd0fa6",
 CellID->1580608710],
Cell[33659092, 783280, 647, 13, 22, "Print",ExpressionUUID->"9f9481c5-25af-4482-8c9f-45ff396c41d5",
 CellID->917875580],
Cell[33659742, 783295, 626, 13, 22, "Print",ExpressionUUID->"7377b964-2cb8-470a-811b-2364038c245a",
 CellID->1359836659],
Cell[33660371, 783310, 614, 13, 22, "Print",ExpressionUUID->"652d7cc1-eb47-4489-a678-e6915290127a",
 CellID->1253756439],
Cell[33660988, 783325, 621, 13, 22, "Print",ExpressionUUID->"710f5e63-69bf-49e4-8cc8-8a8a2e7d1b51",
 CellID->774925909],
Cell[33661612, 783340, 648, 14, 22, "Print",ExpressionUUID->"c9e8cbf3-18ec-46a7-9a82-d0185a5ec905",
 CellID->1838917155],
Cell[33662263, 783356, 624, 13, 22, "Print",ExpressionUUID->"3dbebf0a-6f4a-47ba-aa63-99c91a4ccdc1",
 CellID->95976461],
Cell[33662890, 783371, 649, 13, 22, "Print",ExpressionUUID->"859a5819-91ab-476f-be02-405b23ff59ba",
 CellID->316161130],
Cell[33663542, 783386, 638, 13, 22, "Print",ExpressionUUID->"7c12dac8-2577-47db-8885-73ff69290bb5",
 CellID->845569232]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[33664229, 783405, 236, 4, 27, "Input",ExpressionUUID->"5b5c42b8-c279-4c08-93f8-87f64be4096f",
 CellID->22553876],
Cell[33664468, 783411, 765, 19, 22, "Print",ExpressionUUID->"bd4be96b-0d7b-4edb-9d84-0edfb2e0fd72",
 CellID->1582790025]
}, Open  ]],
Cell[CellGroupData[{
Cell[33665270, 783435, 279, 6, 27, "Input",ExpressionUUID->"15e94ce6-a49e-4f49-beee-a6276d28a78a",
 CellID->1704242589],
Cell[CellGroupData[{
Cell[33665574, 783445, 618, 13, 22, "Print",ExpressionUUID->"7aead2fb-f053-42b3-ad89-188c6fac6126",
 CellID->1072014786],
Cell[33666195, 783460, 621, 13, 22, "Print",ExpressionUUID->"160ea8f3-8b6e-4085-ba21-09d716a728dc",
 CellID->2112193619],
Cell[33666819, 783475, 652, 14, 22, "Print",ExpressionUUID->"41b951a0-d4ca-4070-8843-60e84f79cec2",
 CellID->1733469623],
Cell[33667474, 783491, 643, 13, 22, "Print",ExpressionUUID->"d0aca02b-de3b-4807-973b-50b182f299f8",
 CellID->1660798559],
Cell[33668120, 783506, 651, 13, 22, "Print",ExpressionUUID->"169da2fc-82b4-4a95-8bd5-f86e57e07612",
 CellID->48937407],
Cell[33668774, 783521, 626, 13, 22, "Print",ExpressionUUID->"1180dc5a-010c-408c-b320-9092c4e5d882",
 CellID->4762648],
Cell[33669403, 783536, 650, 13, 22, "Print",ExpressionUUID->"3ef44509-2b35-4a9c-91de-013ecdaade92",
 CellID->588929997],
Cell[33670056, 783551, 680, 14, 22, "Print",ExpressionUUID->"0dda059f-07b8-4a97-970a-bdebe1062733",
 CellID->394330556],
Cell[33670739, 783567, 631, 13, 22, "Print",ExpressionUUID->"72f990af-0eab-4760-af46-ddd65a38d8b1",
 CellID->379120050],
Cell[33671373, 783582, 643, 14, 22, "Print",ExpressionUUID->"32da0bfb-e76c-4d4c-94aa-8c6b54cc424b",
 CellID->286699701],
Cell[33672019, 783598, 642, 14, 22, "Print",ExpressionUUID->"1ca91084-6cbc-48f7-a6c5-1a6c358cc03b",
 CellID->1756054486],
Cell[33672664, 783614, 633, 13, 22, "Print",ExpressionUUID->"34c8d1d6-5f1a-4bcd-b44b-bb488d875673",
 CellID->1139985481],
Cell[33673300, 783629, 653, 13, 22, "Print",ExpressionUUID->"a65c812b-c13f-43ec-bd3e-cbd697a4e36b",
 CellID->2034364280],
Cell[33673956, 783644, 657, 13, 22, "Print",ExpressionUUID->"c295ac7c-6a13-4630-b4cd-ba0a677003cf",
 CellID->1473310073],
Cell[33674616, 783659, 603, 13, 22, "Print",ExpressionUUID->"041ada7e-6d51-42d4-bd77-39e2bb69f906",
 CellID->1435947001],
Cell[33675222, 783674, 628, 13, 22, "Print",ExpressionUUID->"89a3edc4-63fa-45ab-b2cc-e3adc62c7110",
 CellID->540020821],
Cell[33675853, 783689, 617, 13, 22, "Print",ExpressionUUID->"01de40d3-6270-49e1-99c8-95eb29dc51c5",
 CellID->864046325],
Cell[33676473, 783704, 632, 13, 22, "Print",ExpressionUUID->"d54a7889-a704-4d53-8b6f-ec33b020d94b",
 CellID->742163261],
Cell[33677108, 783719, 630, 13, 22, "Print",ExpressionUUID->"857af552-24f9-4797-bf4b-20ad964aaf82",
 CellID->952905851],
Cell[33677741, 783734, 647, 13, 22, "Print",ExpressionUUID->"70e53f76-ac1c-44cc-97fd-1c56bdbab74a",
 CellID->1703082078],
Cell[33678391, 783749, 623, 13, 22, "Print",ExpressionUUID->"3460c327-12e5-4849-826b-cd0d6c23d027",
 CellID->2038437730],
Cell[33679017, 783764, 644, 13, 22, "Print",ExpressionUUID->"42cb6630-ad89-40fd-811f-219b8c50e4a1",
 CellID->1216307519],
Cell[33679664, 783779, 670, 13, 22, "Print",ExpressionUUID->"aeaa498b-8c91-4954-9ea7-fa39b80d986a",
 CellID->583636040],
Cell[33680337, 783794, 627, 13, 22, "Print",ExpressionUUID->"1b6fad96-6614-4da9-a63f-b001828dfb46",
 CellID->1613108431],
Cell[33680967, 783809, 635, 13, 22, "Print",ExpressionUUID->"df84f3b9-0ff0-4acc-baa4-53fcce6faf38",
 CellID->1679840089],
Cell[33681605, 783824, 634, 13, 22, "Print",ExpressionUUID->"b8beb026-aa03-4aed-ac5b-834151f0a518",
 CellID->104868714],
Cell[33682242, 783839, 627, 13, 22, "Print",ExpressionUUID->"2febfcf8-09a3-4b8c-82f0-f4df99fb25fc",
 CellID->1591885658],
Cell[33682872, 783854, 646, 13, 22, "Print",ExpressionUUID->"fe6758be-7307-479e-8693-48e6a28d3e9b",
 CellID->1470979680],
Cell[33683521, 783869, 650, 13, 22, "Print",ExpressionUUID->"52796659-f7a8-40a2-9101-951ec0520869",
 CellID->923737496],
Cell[33684174, 783884, 595, 13, 22, "Print",ExpressionUUID->"40e1095a-ba17-40b6-ae0f-63fcdc0cdd39",
 CellID->1096811109]
}, Open  ]],
Cell[33684784, 783900, 294, 6, 26, "Output",ExpressionUUID->"a8ce465f-d5cf-43a8-9a5f-73792a2aee9a",
 CellID->104683115]
}, Open  ]],
Cell[CellGroupData[{
Cell[33685115, 783911, 234, 4, 27, "Input",ExpressionUUID->"1c73492a-126a-479c-b2c9-500368e3b893",
 CellID->755307977],
Cell[CellGroupData[{
Cell[33685374, 783919, 772, 19, 22, "Print",ExpressionUUID->"86347138-8c19-4f51-b0ca-292254efbb95",
 CellID->620006912],
Cell[33686149, 783940, 606, 13, 22, "Print",ExpressionUUID->"243f0019-c2b2-4e73-ae81-aaf5c7fbcdc2",
 CellID->865514740]
}, Open  ]]
}, Open  ]],
Cell[33686782, 783957, 418, 8, 27, "Input",ExpressionUUID->"232968c4-4f59-4d94-8028-98f21d96af23",
 CellID->234831011]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[33687249, 783971, 441, 6, 39, "Subsection",ExpressionUUID->"4f9fcfd4-2025-4a43-a25b-6521aa90dea4",
 CellID->590673241],
Cell[33687693, 783979, 292, 5, 23, "Text",ExpressionUUID->"c9b98765-d104-4605-9a17-34896f964a37",
 CellID->1973862211],
Cell[CellGroupData[{
Cell[33688010, 783988, 281, 4, 41, "Subsubsection",ExpressionUUID->"8d7ca4fe-498a-449e-b596-a97ad0950afd",
 CellID->784013520],
Cell[CellGroupData[{
Cell[33688316, 783996, 229, 3, 40, "MathCaption",ExpressionUUID->"3594f763-a25b-474e-ac20-29034dfd7399",
 CellID->2103056295],
Cell[CellGroupData[{
Cell[33688570, 784003, 242, 4, 27, "Input",ExpressionUUID->"d809aff5-3a2e-4325-b050-00508222595f",
 CellID->633804092],
Cell[CellGroupData[{
Cell[33688837, 784011, 847, 12, 22, "Print",ExpressionUUID->"bff068ff-c48c-453a-8c9c-062adbfa862a",
 CellID->1896641082],
Cell[33689687, 784025, 1119, 20, 22, "Print",ExpressionUUID->"4ec7aaaf-bd23-4e0d-88d6-2d7a1607f774",
 CellID->1746892753],
Cell[33690809, 784047, 868, 12, 22, "Print",ExpressionUUID->"c0ddb2c8-902a-4bc5-aef4-7155462e7456",
 CellID->1777561534],
Cell[33691680, 784061, 825, 11, 22, "Print",ExpressionUUID->"c12210cf-b647-4922-a2f1-4ef6aeacafd0",
 CellID->1831688521],
Cell[33692508, 784074, 808, 11, 22, "Print",ExpressionUUID->"735948ea-d414-4293-8aa1-06c5e1e480e1",
 CellID->1010892702],
Cell[33693319, 784087, 845, 12, 22, "Print",ExpressionUUID->"3c99a295-7990-4874-aa1d-0a27e3ea74de",
 CellID->1330511097],
Cell[33694167, 784101, 1120, 20, 22, "Print",ExpressionUUID->"347e17a4-bbcd-4382-a55e-a4d2f56c41bb",
 CellID->152791068],
Cell[33695290, 784123, 865, 12, 22, "Print",ExpressionUUID->"648eec4e-d467-4913-aa9f-c5ca610830f2",
 CellID->1716521711],
Cell[33696158, 784137, 844, 12, 22, "Print",ExpressionUUID->"4d5e1b62-3f1d-4333-8d8c-9d79f4926a24",
 CellID->285082979],
Cell[33697005, 784151, 1116, 20, 22, "Print",ExpressionUUID->"39cd9dd9-55a6-4a5b-9cb0-bc82959b243b",
 CellID->353611596],
Cell[33698124, 784173, 884, 12, 22, "Print",ExpressionUUID->"4b1fc357-c0b0-4e37-92a3-63cd47b38790",
 CellID->1062842723],
Cell[33699011, 784187, 844, 12, 22, "Print",ExpressionUUID->"20a4312b-6db2-42b2-b4e4-6ebe13105a04",
 CellID->428669715],
Cell[33699858, 784201, 1000, 14, 42, "Print",ExpressionUUID->"cfc266a7-e276-453f-b9d6-5730d55009cc",
 CellID->1991747967],
Cell[33700861, 784217, 844, 12, 22, "Print",ExpressionUUID->"25ef8328-7041-4e1f-b2f2-c6cec6998ee0",
 CellID->332991933]
}, Open  ]]
}, Open  ]],
Cell[33701732, 784233, 386, 8, 47, "Input",ExpressionUUID->"012040cb-213b-485f-a513-78a5817b41d3",
 CellID->2006047368],
Cell[33702121, 784243, 256, 5, 27, "Input",ExpressionUUID->"db4fc597-7708-4b87-aa0c-76fba75fe1f0",
 CellID->120754463]
}, Open  ]],
Cell[CellGroupData[{
Cell[33702414, 784253, 140, 1, 40, "MathCaption",ExpressionUUID->"2e8e88be-359f-4d55-8b5b-9b672dd86ce9",
 CellID->148213226],
Cell[33702557, 784256, 328, 7, 27, "Input",ExpressionUUID->"a9bdc773-311a-4646-b41f-95edfd0a5a71",
 CellID->2086142509],
Cell[33702888, 784265, 471, 12, 27, "Input",ExpressionUUID->"d0219ec1-54cf-4f51-8080-ba6b11f8309b",
 CellID->1412400003]
}, Open  ]],
Cell[CellGroupData[{
Cell[33703396, 784282, 261, 5, 40, "MathCaption",ExpressionUUID->"81095bee-25c6-48ca-835c-c4a7665d96e8",
 CellID->1238377332],
Cell[33703660, 784289, 524, 13, 27, "Input",ExpressionUUID->"6c7b9365-7476-4b38-ba05-7824c772c60c",
 CellID->2143795847]
}, Open  ]],
Cell[CellGroupData[{
Cell[33704221, 784307, 200, 2, 40, "MathCaption",ExpressionUUID->"9a87f3d3-536f-4f58-8eac-04f82676e28b",
 CellID->1350532130],
Cell[33704424, 784311, 587, 13, 66, "Input",ExpressionUUID->"5fe73ad1-9924-4f5d-b43c-0b8ddeacab07",
 CellID->1985924282]
}, Open  ]],
Cell[CellGroupData[{
Cell[33705048, 784329, 272, 5, 40, "MathCaption",ExpressionUUID->"3859844e-7fd5-4bcf-9860-f9244333a45f",
 CellID->479736035],
Cell[CellGroupData[{
Cell[33705345, 784338, 397, 12, 27, "Input",ExpressionUUID->"eadba501-79ae-4805-88b1-6d1e20d0792f",
 CellID->1269929407],
Cell[33705745, 784352, 1968, 46, 32, "Output",ExpressionUUID->"9bb51acd-53fa-4bf2-a129-95c2c0355400",
 CellID->1625239968]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[33707774, 784405, 291, 4, 41, "Subsubsection",ExpressionUUID->"9f8bffcf-9bfe-4428-a0ee-4c58a7237e26",
 CellID->1899572791],
Cell[33708068, 784411, 376, 7, 23, "Text",ExpressionUUID->"81b1532c-b33a-4242-9bf4-38944101df69",
 CellID->1628002035],
Cell[CellGroupData[{
Cell[33708469, 784422, 317, 6, 40, "MathCaption",ExpressionUUID->"309f6f49-7cf0-4cbf-b167-6f9131046394",
 CellID->360157534],
Cell[CellGroupData[{
Cell[33708811, 784432, 3361, 95, 218, "Input",ExpressionUUID->"68166242-18f5-48f2-83f1-79a9bbc327d5",
 CellID->1558756692],
Cell[33712175, 784529, 15310, 486, 120, "Output",ExpressionUUID->"1fad0e2f-8890-4eaa-9032-3f0b2c42a934",
 CellID->1427830721]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[33727534, 785021, 297, 4, 40, "MathCaption",ExpressionUUID->"2d79acda-5f04-4d81-846e-17c33809bdfe",
 CellID->408762261],
Cell[CellGroupData[{
Cell[33727856, 785029, 343, 7, 27, "Input",ExpressionUUID->"54106a6f-11e0-468a-892a-1022bba25642",
 CellID->267133874],
Cell[33728202, 785038, 35193, 1006, 146, "Output",ExpressionUUID->"c722a884-4c0c-49ad-8e0c-35db5f0106c5",
 CellID->1568656269]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[33763444, 786050, 280, 5, 40, "MathCaption",ExpressionUUID->"5b98dbdb-8323-4b5c-b2bd-7f48d03acd9a",
 CellID->655749584],
Cell[CellGroupData[{
Cell[33763749, 786059, 236, 4, 27, "Input",ExpressionUUID->"7d0367da-d792-44f0-ba7b-4737637c0d41",
 CellID->297181884],
Cell[33763988, 786065, 39968, 1116, 130, "Output",ExpressionUUID->"1045434f-633e-43f5-a2b0-30e40e310dcc",
 CellID->1896662511]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[33804005, 787187, 239, 3, 40, "MathCaption",ExpressionUUID->"a9ce3e45-fbf3-4450-a5e3-b89aad94e014",
 CellID->1498039507],
Cell[CellGroupData[{
Cell[33804269, 787194, 780, 25, 27, "Input",ExpressionUUID->"b9fa1e5f-cdc1-4767-a6b4-15773cdd297f",
 CellID->451716721],
Cell[33805052, 787221, 53105, 1502, 56, "Output",ExpressionUUID->"ac629944-5638-4c52-a08a-fd37918f6939",
 CellID->2107049956]
}, Open  ]],
Cell[CellGroupData[{
Cell[33858194, 788728, 310, 6, 27, "Input",ExpressionUUID->"68590c4b-e27f-4bde-a7ab-4a5f8d931941",
 CellID->746516283],
Cell[33858507, 788736, 26279, 834, 105, "Output",ExpressionUUID->"532d6bde-8954-464c-bb1c-17545c135edb",
 CellID->1183271462]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[33884847, 789577, 356, 4, 41, "Subsubsection",ExpressionUUID->"d514aac7-9a8d-4000-af71-aac01ae83808",
 CellID->424015098],
Cell[33885206, 789583, 492, 9, 40, "Text",ExpressionUUID->"642fc84e-a609-4cd1-ad04-b13f631e4dca",
 CellID->1294809613],
Cell[CellGroupData[{
Cell[33885723, 789596, 277, 5, 40, "MathCaption",ExpressionUUID->"8ee56a8e-222b-4c3c-b37e-35ab6e782ec9",
 CellID->1799817068],
Cell[CellGroupData[{
Cell[33886025, 789605, 629, 15, 27, "Input",ExpressionUUID->"8582ced9-75c2-49af-99a0-983fc8860f43",
 CellID->472400955],
Cell[CellGroupData[{
Cell[33886679, 789624, 541, 12, 22, "Print",ExpressionUUID->"811310bd-0d6a-4b79-8c63-dd1fc0e94ea6",
 CellID->619974268],
Cell[33887223, 789638, 530, 10, 22, "Print",ExpressionUUID->"5fd1a74f-1516-451f-a3be-20e4799963dd",
 CellID->316867032],
Cell[33887756, 789650, 611, 11, 22, "Print",ExpressionUUID->"228ddc58-2686-430b-a9fa-86136c838338",
 CellID->1126934732],
Cell[33888370, 789663, 623, 11, 22, "Print",ExpressionUUID->"2affbb7e-862a-4d45-93d2-63388fbddd96",
 CellID->1733757831],
Cell[33888996, 789676, 622, 15, 22, "Print",ExpressionUUID->"2d934f90-9c78-4387-bbd8-147bf101d9c5",
 CellID->62259474],
Cell[33889621, 789693, 534, 10, 22, "Print",ExpressionUUID->"9e48ada5-a905-44a3-85be-933ca8746308",
 CellID->570443429],
Cell[33890158, 789705, 611, 11, 22, "Print",ExpressionUUID->"8b9e8991-0414-4b82-8631-1ace45f7a7b4",
 CellID->1075710995],
Cell[33890772, 789718, 621, 11, 22, "Print",ExpressionUUID->"3ec77afa-e320-4e4a-9bf5-39cbc153332b",
 CellID->1957352519],
Cell[33891396, 789731, 661, 16, 22, "Print",ExpressionUUID->"373854e1-f7c4-42d0-96da-358283cd3857",
 CellID->2069282087],
Cell[33892060, 789749, 535, 10, 22, "Print",ExpressionUUID->"aa02c315-766b-4bc6-9509-ebddb6172690",
 CellID->2073856691],
Cell[33892598, 789761, 611, 11, 22, "Print",ExpressionUUID->"2e5ed5bc-e176-43ee-91a2-01f7a8f949b6",
 CellID->1649814827],
Cell[33893212, 789774, 626, 12, 22, "Print",ExpressionUUID->"d9d110a5-d781-4489-ad1a-c7ae62c79e7c",
 CellID->128947325],
Cell[33893841, 789788, 631, 15, 22, "Print",ExpressionUUID->"1a35c1bf-9804-4a28-984a-66844f22963f",
 CellID->406691452],
Cell[33894475, 789805, 533, 10, 22, "Print",ExpressionUUID->"da007344-82fc-4b53-b603-1d98cde97f90",
 CellID->1970269010],
Cell[33895011, 789817, 608, 11, 22, "Print",ExpressionUUID->"1b353ff6-b42b-4d50-a706-abf3f2225a81",
 CellID->113414330],
Cell[33895622, 789830, 628, 12, 22, "Print",ExpressionUUID->"1528b829-dd4e-4d42-b2f6-cf4c9626e058",
 CellID->1336649421],
Cell[33896253, 789844, 609, 14, 22, "Print",ExpressionUUID->"25a8a703-0772-45ca-afff-56df57287b6f",
 CellID->240387480],
Cell[33896865, 789860, 532, 10, 22, "Print",ExpressionUUID->"fc711523-9fd4-4c30-b325-8139cc5dd753",
 CellID->775636353],
Cell[33897400, 789872, 610, 11, 22, "Print",ExpressionUUID->"8ceb9784-ad88-439c-8ea5-4cd1bc630cf3",
 CellID->894447581],
Cell[33898013, 789885, 622, 11, 22, "Print",ExpressionUUID->"651f865b-b920-4277-b8cb-a50613eef1e7",
 CellID->594964867],
Cell[33898638, 789898, 658, 16, 22, "Print",ExpressionUUID->"5549a295-689c-4a2e-817d-46c9c1976f31",
 CellID->890659237],
Cell[33899299, 789916, 533, 10, 22, "Print",ExpressionUUID->"ceb79d6d-7d41-4e92-a3e3-2d1e0685b081",
 CellID->1348776669],
Cell[33899835, 789928, 611, 11, 22, "Print",ExpressionUUID->"4b91d6e2-554f-4cbf-a64a-dae06858e1e0",
 CellID->52098151],
Cell[33900449, 789941, 630, 12, 22, "Print",ExpressionUUID->"4793a181-9bc1-45da-9e37-d1ad6a18d207",
 CellID->1587779528]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[33901140, 789960, 215, 2, 40, "MathCaption",ExpressionUUID->"9bffa71c-51f6-4436-b340-7fe884827cf1",
 CellID->933872350],
Cell[CellGroupData[{
Cell[33901380, 789966, 362, 8, 27, "Input",ExpressionUUID->"2f8d462c-c1f9-4b0b-9d8b-aeef33324e0f",
 CellID->172003964],
Cell[33901745, 789976, 168002, 4857, 852, "Output",ExpressionUUID->"6d31d707-d5fa-4bb8-8ffc-a8e59075811a",
 CellID->1178729474]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[34069796, 794839, 249, 4, 40, "MathCaption",ExpressionUUID->"d744620d-cec3-45f7-bc4d-a14f29979bf5",
 CellID->891591804],
Cell[CellGroupData[{
Cell[34070070, 794847, 569, 10, 27, "Input",ExpressionUUID->"43e2d24a-90ba-490b-81ae-1f6cab55b8f4",
 CellID->1253380061],
Cell[CellGroupData[{
Cell[34070664, 794861, 622, 13, 22, "Print",ExpressionUUID->"7fdc86a3-9da7-4548-ab71-bdafb91bd9f1",
 CellID->369625943],
Cell[34071289, 794876, 419, 6, 22, "Print",ExpressionUUID->"def63c57-3a39-4cb7-b9f7-e227382c2470",
 CellID->1780516877],
Cell[34071711, 794884, 706, 16, 22, "Print",ExpressionUUID->"61c9b278-e884-4818-9f45-6ebbf4fa5a7b",
 CellID->2110014441],
Cell[34072420, 794902, 419, 6, 22, "Print",ExpressionUUID->"0934395f-3a38-4d36-8ac9-39fc6af222e4",
 CellID->1615246976],
Cell[34072842, 794910, 741, 17, 22, "Print",ExpressionUUID->"4a44770e-9a01-4991-9af0-35ef67789eed",
 CellID->1115143905],
Cell[34073586, 794929, 421, 6, 22, "Print",ExpressionUUID->"07d5c611-b7ff-4e75-863f-be7315075735",
 CellID->1133823966],
Cell[34074010, 794937, 712, 16, 22, "Print",ExpressionUUID->"dc0e1f29-ce59-4575-9265-ca82e13b4fda",
 CellID->1556996731],
Cell[34074725, 794955, 421, 6, 22, "Print",ExpressionUUID->"b14f0848-ae87-407c-81a6-2594ba028c99",
 CellID->1355819222],
Cell[34075149, 794963, 689, 15, 22, "Print",ExpressionUUID->"9196ec60-2f83-42fc-bfae-31e79649ce7c",
 CellID->304685837],
Cell[34075841, 794980, 419, 6, 22, "Print",ExpressionUUID->"87591ade-45f6-47f3-902c-1d77489139ab",
 CellID->1253848011],
Cell[34076263, 794988, 737, 17, 22, "Print",ExpressionUUID->"c7656a4d-7e15-4297-b60a-398f63bc38e6",
 CellID->166492866],
Cell[34077003, 795007, 417, 6, 22, "Print",ExpressionUUID->"ce53fdb8-773e-4115-939d-bf8a1a8814ad",
 CellID->74406821],
Cell[34077423, 795015, 716, 16, 22, "Print",ExpressionUUID->"95bb3adb-0482-4121-b475-965f827811b1",
 CellID->719957993],
Cell[34078142, 795033, 613, 11, 22, "Print",ExpressionUUID->"33cfefea-e224-4241-b962-cfc5d21614ef",
 CellID->1411121153],
Cell[34078758, 795046, 691, 12, 22, "Print",ExpressionUUID->"14d4da98-0e1a-40cc-b544-69822dfe4432",
 CellID->2051304650],
Cell[34079452, 795060, 707, 13, 22, "Print",ExpressionUUID->"f3cf2a43-7539-4af1-82b2-481689ff7623",
 CellID->574783612],
Cell[34080162, 795075, 695, 15, 22, "Print",ExpressionUUID->"c09da613-3fa0-4165-b22c-d4b475034e37",
 CellID->1006722678],
Cell[34080860, 795092, 613, 11, 22, "Print",ExpressionUUID->"4fe44c4f-cf4a-46ed-9b3d-3c86adf75180",
 CellID->2111878080],
Cell[34081476, 795105, 690, 12, 22, "Print",ExpressionUUID->"9a1f76c3-8bc7-4a76-b9db-e46d4524cf34",
 CellID->725172944],
Cell[34082169, 795119, 708, 13, 22, "Print",ExpressionUUID->"76e05a46-0102-4416-9b87-731effa48322",
 CellID->1011973083],
Cell[34082880, 795134, 598, 12, 22, "Print",ExpressionUUID->"f4e1af7b-1245-4d66-8ed0-e4441c76f8b0",
 CellID->161121741],
Cell[34083481, 795148, 612, 11, 22, "Print",ExpressionUUID->"6abd2108-a14e-453a-99d6-ef585cf1a51f",
 CellID->2143705767],
Cell[34084096, 795161, 690, 12, 22, "Print",ExpressionUUID->"978d45eb-b1ec-49b4-9456-2a297de44734",
 CellID->929680250],
Cell[34084789, 795175, 699, 12, 22, "Print",ExpressionUUID->"0585785f-ce86-459d-aa6e-8f48ae17693c",
 CellID->44946178]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[34085549, 795194, 183, 2, 40, "MathCaption",ExpressionUUID->"d0fc5947-e88b-487f-9916-a12f37f743e8",
 CellID->336090536],
Cell[CellGroupData[{
Cell[34085757, 795200, 257, 5, 27, "Input",ExpressionUUID->"15087f24-7176-4f9c-9859-688200ebe04c",
 CellID->1680049480],
Cell[34086017, 795207, 14984, 491, 117, "Output",ExpressionUUID->"20faf592-16f0-457a-9ded-c5a28807eb45",
 CellID->1643653549]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[34101050, 795704, 204, 2, 40, "MathCaption",ExpressionUUID->"95f6e10e-e82c-41af-928a-ea149e87afe6",
 CellID->934268211],
Cell[CellGroupData[{
Cell[34101279, 795710, 373, 9, 27, "Input",ExpressionUUID->"c2f80f50-46fb-4e58-bfd9-a56669de8fd9",
 CellID->376564360],
Cell[34101655, 795721, 9215, 291, 73, "Output",ExpressionUUID->"4dda3aaf-de82-4e52-b227-6ab0526b8048",
 CellID->1803046682]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[34110931, 796019, 395, 5, 41, "Subsubsection",ExpressionUUID->"d00c094b-e312-4fda-bdfc-b90d5707f82e",
 CellID->184667122],
Cell[CellGroupData[{
Cell[34111351, 796028, 223, 3, 40, "MathCaption",ExpressionUUID->"a61889c1-4790-496f-ae19-85d63df3feef",
 CellID->460893304],
Cell[CellGroupData[{
Cell[34111599, 796035, 827, 14, 27, "Input",ExpressionUUID->"fb9c75f2-586b-4e94-8136-46ab68fc4f8e",
 CellID->106417567],
Cell[CellGroupData[{
Cell[34112451, 796053, 460, 10, 22, "Print",ExpressionUUID->"bbf47d44-7861-426c-9fea-1409257ac8b7",
 CellID->183170943],
Cell[34112914, 796065, 481, 9, 22, "Print",ExpressionUUID->"76296028-388f-4de8-b91c-bfe3d00594c3",
 CellID->1209972850],
Cell[34113398, 796076, 555, 10, 22, "Print",ExpressionUUID->"f150aba2-17c9-4436-a290-21a70df61d60",
 CellID->1491036507],
Cell[34113956, 796088, 566, 10, 22, "Print",ExpressionUUID->"ff721be8-1985-463f-884b-8ab6b21042dd",
 CellID->863896306],
Cell[34114525, 796100, 488, 11, 22, "Print",ExpressionUUID->"cff8f03a-eb09-496a-bee7-c6da57b64512",
 CellID->368277575],
Cell[34115016, 796113, 284, 4, 22, "Print",ExpressionUUID->"54841b90-153b-4142-b2e2-685c80bb8614",
 CellID->593332371],
Cell[34115303, 796119, 570, 14, 22, "Print",ExpressionUUID->"61ac913b-c832-469e-94f8-8e60ce798575",
 CellID->1370586376],
Cell[34115876, 796135, 285, 4, 22, "Print",ExpressionUUID->"9f364b11-1b2a-4ea6-aa61-0548dd5197cc",
 CellID->1535623710],
Cell[34116164, 796141, 606, 15, 22, "Print",ExpressionUUID->"0d5f0487-adc0-4513-b18c-f98489b9f1b2",
 CellID->769224324],
Cell[34116773, 796158, 284, 4, 22, "Print",ExpressionUUID->"7969149a-0473-48d5-8746-f967498d8c66",
 CellID->501658528],
Cell[34117060, 796164, 577, 14, 22, "Print",ExpressionUUID->"b2b5d23e-c13b-4969-923a-9e4638061cd6",
 CellID->354081974],
Cell[34117640, 796180, 284, 4, 22, "Print",ExpressionUUID->"d21c5c89-4b03-4d91-9a5b-b576bbf42ed9",
 CellID->378551181],
Cell[34117927, 796186, 558, 13, 22, "Print",ExpressionUUID->"37dcb28d-1f37-4e4a-b525-58071f684dc2",
 CellID->1463136653],
Cell[34118488, 796201, 284, 4, 22, "Print",ExpressionUUID->"6da4def3-938a-434d-8ff0-baea367481e3",
 CellID->102485174],
Cell[34118775, 796207, 604, 15, 22, "Print",ExpressionUUID->"6c3fcb47-fa05-4fdb-8251-c07f42b81c22",
 CellID->186434524],
Cell[34119382, 796224, 284, 4, 22, "Print",ExpressionUUID->"046b0df9-804c-4f4d-9395-0c27ee2e44b4",
 CellID->226403895],
Cell[34119669, 796230, 583, 14, 22, "Print",ExpressionUUID->"660252cf-ce03-48e9-b3f4-a8ed3b7026dc",
 CellID->1976724428],
Cell[34120255, 796246, 285, 4, 22, "Print",ExpressionUUID->"d5c09eef-55c3-473b-bafa-86a80cb74c95",
 CellID->1235442306],
Cell[34120543, 796252, 529, 13, 22, "Print",ExpressionUUID->"ad819efc-86ea-4685-bdc6-9764e6b3858c",
 CellID->59454099],
Cell[34121075, 796267, 478, 9, 22, "Print",ExpressionUUID->"60c8d46c-4879-4f41-b384-54dcbb782737",
 CellID->665146038],
Cell[34121556, 796278, 557, 10, 22, "Print",ExpressionUUID->"ae23918b-a1aa-4685-9824-2e6a463aa3a1",
 CellID->1457078031],
Cell[34122116, 796290, 569, 10, 22, "Print",ExpressionUUID->"d1379018-181d-415e-8ab0-6f9551348ded",
 CellID->1354440276],
Cell[34122688, 796302, 464, 10, 22, "Print",ExpressionUUID->"0a0c82b4-7944-41e6-8f85-709cfe1b6288",
 CellID->751060532],
Cell[34123155, 796314, 480, 9, 22, "Print",ExpressionUUID->"bd55c4a4-1e50-4e93-9a30-fed5ea6403be",
 CellID->165484258],
Cell[34123638, 796325, 556, 10, 22, "Print",ExpressionUUID->"adff7d3e-99ed-457f-bd1e-11662a806cf7",
 CellID->302601341],
Cell[34124197, 796337, 564, 10, 22, "Print",ExpressionUUID->"417ccc52-f3ec-4c20-a74a-677a05712b5e",
 CellID->579462091],
Cell[34124764, 796349, 538, 13, 22, "Print",ExpressionUUID->"aaf09a23-1c6a-49a4-9b4c-6f62ab76aba9",
 CellID->181024292],
Cell[34125305, 796364, 479, 9, 22, "Print",ExpressionUUID->"7593b95c-989e-4f8e-baae-057bc3c23c7f",
 CellID->1638431492],
Cell[34125787, 796375, 557, 10, 22, "Print",ExpressionUUID->"24f510a5-201c-44cc-978f-426eff2d9fcd",
 CellID->2082764210],
Cell[34126347, 796387, 568, 10, 22, "Print",ExpressionUUID->"7fe4f1ab-1641-48cd-ad19-b8e79a5662e1",
 CellID->1034631370],
Cell[34126918, 796399, 598, 15, 22, "Print",ExpressionUUID->"fecab347-f586-4753-bb4d-437a8ed00cdb",
 CellID->874345831],
Cell[34127519, 796416, 479, 9, 22, "Print",ExpressionUUID->"4ae53884-ac08-462a-b68e-47e990e2a881",
 CellID->2047268843],
Cell[34128001, 796427, 557, 10, 22, "Print",ExpressionUUID->"3beb5447-cb9e-4232-aff2-da488c640995",
 CellID->1464452067],
Cell[34128561, 796439, 573, 11, 22, "Print",ExpressionUUID->"f26473c3-ef50-4e73-8e32-272233e5c0f7",
 CellID->735811802],
Cell[34129137, 796452, 561, 13, 22, "Print",ExpressionUUID->"03c16e84-9896-4947-89fe-cbd0169f4084",
 CellID->1578116788],
Cell[34129701, 796467, 287, 4, 22, "Print",ExpressionUUID->"846a0ab8-6924-4882-9858-add350f7891a",
 CellID->1985815466],
Cell[34129991, 796473, 465, 10, 22, "Print",ExpressionUUID->"ffe3e429-790d-4ee5-aa9e-fdc30b567734",
 CellID->1557179035],
Cell[34130459, 796485, 284, 4, 22, "Print",ExpressionUUID->"3d24aad0-2f6d-4280-b366-2659a7c8e3c0",
 CellID->124835256],
Cell[34130746, 796491, 642, 16, 22, "Print",ExpressionUUID->"04b0b9eb-ad0c-4eb0-8360-ad0280b09131",
 CellID->14624473],
Cell[34131391, 796509, 478, 9, 22, "Print",ExpressionUUID->"7d99c010-e004-4cd4-953d-cf8b7107877d",
 CellID->980381953],
Cell[34131872, 796520, 555, 10, 22, "Print",ExpressionUUID->"240bdd32-2d51-4940-8bb6-9c9dee2a3b3a",
 CellID->1784944287],
Cell[34132430, 796532, 573, 10, 22, "Print",ExpressionUUID->"dc2357c1-e12a-4780-a8ab-fea1944e6aef",
 CellID->1359566666]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[34133064, 796549, 245, 4, 40, "MathCaption",ExpressionUUID->"70fac69f-28f0-4e4a-a769-d6faf4f599a4",
 CellID->2126404645],
Cell[CellGroupData[{
Cell[34133334, 796557, 229, 5, 27, "Input",ExpressionUUID->"cca0254f-4ce7-4efa-8e5f-7f903d9faca2",
 CellID->60015141],
Cell[34133566, 796564, 1758, 62, 40, "Output",ExpressionUUID->"5d3bc757-95f1-4cb0-8d83-ca017107d521",
 CellID->1521614842]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[34135373, 796632, 328, 6, 40, "MathCaption",ExpressionUUID->"bba588bf-90b0-4d97-accb-8cd0578c2db9",
 CellID->1937546183],
Cell[CellGroupData[{
Cell[34135726, 796642, 596, 16, 27, "Input",ExpressionUUID->"8418701f-4e9a-4387-baf7-792198aabd02",
 CellID->2044158220],
Cell[34136325, 796660, 7153, 194, 32, "Output",ExpressionUUID->"775e4cf9-b85d-477f-a185-0d193c17b61b",
 CellID->511626195]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[34143527, 796860, 196, 2, 40, "MathCaption",ExpressionUUID->"f895cd67-a333-4279-8a13-972b21825409",
 CellID->592187185],
Cell[CellGroupData[{
Cell[34143748, 796866, 476, 12, 27, "Input",ExpressionUUID->"4209e6b8-cfa3-4fd1-ada4-f7b19e45456e",
 CellID->1875766430],
Cell[34144227, 796880, 10734, 295, 32, "Output",ExpressionUUID->"8fa903ec-cce7-4202-b89a-3ac0edac79d9",
 CellID->849032740]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[34155010, 797181, 221, 2, 40, "MathCaption",ExpressionUUID->"53082d73-54e9-46a8-a194-4594b57b09c8",
 CellID->1222276682],
Cell[CellGroupData[{
Cell[34155256, 797187, 406, 9, 27, "Input",ExpressionUUID->"4072f733-ce48-4585-a1b1-d6613f10533d",
 CellID->591424021],
Cell[34155665, 797198, 26509, 645, 146, "Output",ExpressionUUID->"3a9aba09-a70d-4833-8421-4e4a0c0c047a",
 CellID->1115212920]
}, Open  ]],
Cell[CellGroupData[{
Cell[34182211, 797848, 276, 5, 27, "Input",ExpressionUUID->"e22ed8d8-15c7-4d25-9bf0-c5921aeedd6e",
 CellID->1455885125],
Cell[34182490, 797855, 313, 6, 26, "Output",ExpressionUUID->"ffa81a00-4754-4b5f-aee3-cd892881cfb0",
 CellID->146275424]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[34182852, 797867, 220, 2, 40, "MathCaption",ExpressionUUID->"4aede06f-33af-4081-8929-d4bac1cad013",
 CellID->565555491],
Cell[CellGroupData[{
Cell[34183097, 797873, 310, 6, 27, "Input",ExpressionUUID->"ed466231-1bdf-4d74-9166-5ba5b3dd1c4e",
 CellID->926375781],
Cell[34183410, 797881, 310, 6, 26, "Output",ExpressionUUID->"a5bf000c-8c8b-49de-b1f5-15804e810755",
 CellID->1891364216]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[34183781, 797894, 557, 7, 41, "Subsubsection",ExpressionUUID->"487ae3fb-e0b4-4ed9-8f5b-b84c92a6c7a9",
 CellID->1043706123],
Cell[CellGroupData[{
Cell[34184363, 797905, 234, 4, 27, "Input",ExpressionUUID->"577d2eb1-767b-423b-a8a1-c8fc8ff0004a",
 CellID->922380565],
Cell[CellGroupData[{
Cell[34184622, 797913, 670, 13, 22, "Print",ExpressionUUID->"7dcc6e35-cc92-416f-ba1c-6b307e7170b5",
 CellID->1372880448],
Cell[34185295, 797928, 666, 13, 22, "Print",ExpressionUUID->"1e28365d-a6c5-42b5-bf81-29cd38e3e573",
 CellID->1437386168],
Cell[34185964, 797943, 694, 14, 22, "Print",ExpressionUUID->"7969d275-c801-4273-8458-29e159a0fa15",
 CellID->1105780473],
Cell[34186661, 797959, 673, 13, 22, "Print",ExpressionUUID->"1dd2ee7b-c23e-417b-8559-ac8fa0eb814a",
 CellID->528928573],
Cell[34187337, 797974, 694, 13, 22, "Print",ExpressionUUID->"ef1c9ecf-4e2c-4b45-88c3-9f66ce715e29",
 CellID->1267711478],
Cell[34188034, 797989, 688, 13, 22, "Print",ExpressionUUID->"d214a4dc-1ac7-4265-a0d6-600d09bceeb0",
 CellID->1241548859]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[34188771, 798008, 449, 7, 66, "Input",ExpressionUUID->"7a2b0459-5b1b-420c-a3c4-3709aeb5e014",
 CellID->744990806],
Cell[CellGroupData[{
Cell[34189245, 798019, 791, 19, 22, "Print",ExpressionUUID->"05b9661b-b10d-4750-b68e-af0f00965b20",
 CellID->22444187],
Cell[34190039, 798040, 801, 19, 22, "Print",ExpressionUUID->"e492630c-59b4-4671-a0a7-0257aaad14e8",
 CellID->1409812684],
Cell[34190843, 798061, 797, 19, 22, "Print",ExpressionUUID->"03564a3f-1b3d-4fd9-9347-9141bd823c12",
 CellID->1534702637]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[34191689, 798086, 331, 6, 27, "Input",ExpressionUUID->"44cf8def-0543-4cdf-9a24-aeef4013dddd",
 CellID->1735670720],
Cell[CellGroupData[{
Cell[34192045, 798096, 638, 13, 22, "Print",ExpressionUUID->"e59d184e-219d-4962-a542-249c1295f5f5",
 CellID->42574735],
Cell[34192686, 798111, 639, 13, 22, "Print",ExpressionUUID->"2ee1a4cd-dc93-4feb-8544-af87076be995",
 CellID->441516694],
Cell[34193328, 798126, 654, 13, 22, "Print",ExpressionUUID->"c2b24ac5-ef33-4f49-a7c9-068cdf0b4b2f",
 CellID->1015075673],
Cell[34193985, 798141, 649, 13, 22, "Print",ExpressionUUID->"7c460714-d990-4947-bc7e-e0000db6c9b7",
 CellID->766744343],
Cell[34194637, 798156, 664, 13, 22, "Print",ExpressionUUID->"1f8d0d4e-fc67-4b9a-95d3-187f4842ce53",
 CellID->1770290801],
Cell[34195304, 798171, 639, 13, 22, "Print",ExpressionUUID->"dc8fe303-7750-4a14-a2c8-38bf1ff1d2cd",
 CellID->2039046869],
Cell[34195946, 798186, 661, 13, 22, "Print",ExpressionUUID->"6fc22b14-2226-4e65-8bb2-ae9100e3c9d4",
 CellID->716688457],
Cell[34196610, 798201, 643, 13, 22, "Print",ExpressionUUID->"b6f07ceb-344e-4212-8802-9c6e5e40bf07",
 CellID->147120776],
Cell[34197256, 798216, 663, 13, 22, "Print",ExpressionUUID->"708c70c2-c938-4463-b2b9-c8588853fe04",
 CellID->905204535],
Cell[34197922, 798231, 667, 13, 22, "Print",ExpressionUUID->"e719c1b2-fe93-402f-9457-5c0d49b66dd9",
 CellID->998464397]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[34198638, 798250, 234, 4, 27, "Input",ExpressionUUID->"7ae20ff5-279a-489c-af61-2341149aeeed",
 CellID->694636180],
Cell[CellGroupData[{
Cell[34198897, 798258, 794, 19, 22, "Print",ExpressionUUID->"70caf522-4986-4bcb-8c72-db390f134a35",
 CellID->849232217],
Cell[34199694, 798279, 628, 13, 22, "Print",ExpressionUUID->"dfa57b20-5a52-4423-94f8-66e72145b017",
 CellID->869553157]
}, Open  ]]
}, Open  ]],
Cell[34200349, 798296, 386, 8, 27, "Input",ExpressionUUID->"9aa015eb-d9d0-485b-bbcd-6f4e10e4fca0",
 CellID->1282535422]
}, Closed]]
}, Closed]]
}, Open  ]],
Cell[CellGroupData[{
Cell[34200796, 798311, 171, 3, 43, "Section",ExpressionUUID->"c2047bc9-e93b-4a5a-88aa-30d128fe093b"],
Cell[34200970, 798316, 530, 15, 41, "Text",ExpressionUUID->"dcf454f5-b13e-4e4a-8dd7-2e8c4243eba0"],
Cell[CellGroupData[{
Cell[34201525, 798335, 293, 6, 24, "Text",ExpressionUUID->"74575b00-067d-46dd-b842-ff31b347101c",
 CellID->982851045],
Cell[CellGroupData[{
Cell[34201843, 798345, 134, 2, 27, "Input",ExpressionUUID->"16502e6b-c879-4721-8730-f2a6166014f5"],
Cell[34201980, 798349, 53235, 1236, 644, "Output",ExpressionUUID->"77f7f330-1d53-4e72-a91f-a53f445fc8f4",
 CellID->1863535942]
}, Open  ]],
Cell[CellGroupData[{
Cell[34255252, 799590, 277, 6, 27, "Input",ExpressionUUID->"05c6bb80-47db-407c-8464-bdf51e5d25a3",
 CellID->1079795302],
Cell[34255532, 799598, 224, 4, 26, "Output",ExpressionUUID->"255a084f-d647-4497-8139-4bf66fea0871",
 CellID->1883203041]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[34255805, 799608, 195, 3, 39, "Subsection",ExpressionUUID->"63cba538-ce7d-4ea2-9e48-05b69d04f250"],
Cell[34256003, 799613, 414, 9, 42, "Text",ExpressionUUID->"a0b443eb-98b9-4fe2-bea3-7d76e8a684e8"],
Cell[CellGroupData[{
Cell[34256442, 799626, 361, 8, 42, "DefinitionBox",ExpressionUUID->"1a198fb4-af4f-4333-89ba-5465ad122fc9"],
Cell[34256806, 799636, 95, 0, 30, "SmallText",ExpressionUUID->"744d701d-fdf5-429a-9d2b-59053d4bb4d1"]
}, Open  ]],
Cell[CellGroupData[{
Cell[34256938, 799641, 111, 0, 40, "MathCaption",ExpressionUUID->"c11240fa-1024-4f38-b1c9-c73d8dccddd0"],
Cell[CellGroupData[{
Cell[34257074, 799645, 141, 3, 27, "Input",ExpressionUUID->"e5a482e4-b44f-47ef-86cf-bf1e1b4ce279"],
Cell[34257218, 799650, 272, 9, 27, "Output",ExpressionUUID->"e28c4509-5b96-4c50-a7b9-bbd2d7fb3e3e"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[34257539, 799665, 123, 0, 40, "MathCaption",ExpressionUUID->"70466b7e-7da9-464d-877e-21d1965bda80"],
Cell[34257665, 799667, 231, 7, 27, "Input",ExpressionUUID->"90bea40b-ab5e-4d18-a81f-cade07620092"],
Cell[CellGroupData[{
Cell[34257921, 799678, 145, 3, 27, "Input",ExpressionUUID->"395255ae-98d9-4a48-badd-79511ac570cf"],
Cell[34258069, 799683, 114, 2, 26, "Output",ExpressionUUID->"da44a243-f384-42be-9d22-c52b1076ce54"]
}, Open  ]],
Cell[CellGroupData[{
Cell[34258220, 799690, 149, 3, 27, "Input",ExpressionUUID->"94faed76-0eb1-4960-ad06-4b086d74171b"],
Cell[34258372, 799695, 111, 2, 26, "Output",ExpressionUUID->"0ae4b48e-26ec-4111-b69d-9b893705a5b7"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[34258532, 799703, 162, 2, 40, "MathCaption",ExpressionUUID->"c6531a46-9c91-4613-9e9d-c35508d7ee46"],
Cell[CellGroupData[{
Cell[34258719, 799709, 204, 5, 27, "Input",ExpressionUUID->"c1c1fce2-4756-440c-9fb1-ae5ddf6e4f16"],
Cell[34258926, 799716, 114, 2, 26, "Output",ExpressionUUID->"e0885eae-b96b-4a4b-a91b-c2bfb1dae124"]
}, Open  ]],
Cell[CellGroupData[{
Cell[34259077, 799723, 204, 5, 27, "Input",ExpressionUUID->"a1eef84b-546e-4797-bc56-eeba76c8c561"],
Cell[34259284, 799730, 115, 2, 26, "Output",ExpressionUUID->"2eede9cd-3050-49ed-ab3c-b1c514274054"]
}, Open  ]]
}, Open  ]],
Cell[34259426, 799736, 422, 9, 42, "Text",ExpressionUUID->"f901778d-682a-459d-8459-80ad56ccf2dc"],
Cell[CellGroupData[{
Cell[34259873, 799749, 400, 9, 42, "DefinitionBox",ExpressionUUID->"31812a06-4c72-433c-bc8b-b0ed98a1a39b"],
Cell[34260276, 799760, 116, 0, 30, "SmallText",ExpressionUUID->"91508877-8b79-48d8-8847-ce29be63c7c5"]
}, Open  ]],
Cell[CellGroupData[{
Cell[34260429, 799765, 131, 0, 40, "MathCaption",ExpressionUUID->"e7d5da21-7545-4750-b73a-496903da954f"],
Cell[CellGroupData[{
Cell[34260585, 799769, 175, 4, 27, "Input",ExpressionUUID->"805e3127-7679-4c87-8c48-549af65fcfec"],
Cell[34260763, 799775, 313, 9, 27, "Output",ExpressionUUID->"e3aceb00-7931-484f-ac64-f9cf19d5d200"]
}, Open  ]],
Cell[CellGroupData[{
Cell[34261113, 799789, 141, 3, 27, "Input",ExpressionUUID->"b118f388-d9af-4300-b81e-e5cd16e40e63"],
Cell[34261257, 799794, 129, 3, 26, "Output",ExpressionUUID->"f30a699b-d010-47c8-aaeb-3c072e37bab0"]
}, Open  ]]
}, Open  ]]
}, Closed]]
}, Closed]]
}
]
*)

